/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** dosutil.c
**
** help routines for dos
**
** Creation date: 03.04.1995
** Last modified: 28.12.2020
*/

#include "system.h"
#include "dosutil.h"
#include "util.h"
#include "codeset.h"
#include "memory.h"
#include <dirent.h>
#include <dos/dosextens.h>
#include <dos/dostags.h>
#include <proto/wb.h>
#include <unistd.h>
#include <sys/stat.h>
#define MKDIR(dir, perm) mkdir(dir, perm)
#ifdef __MORPHOS__
    #define rmdir(dir) unlink(dir)	// some libc's don't have it
#endif

const char *GetTmpDir(void)
{
    return "T:";
}

FILE *FOpen(const char *filename, const char *mode)
{
    return fopen(filename, mode);
}

int MkDir(const char *dirname)
{
    return MKDIR(dirname, 0777);
}

int RmDir(const char *dirname)
{
    return rmdir(dirname);
}

int RmFile(const char *filename)
{
    return unlink(filename);
}

int PushDir(const char *dirname, Cwd *cwd)
{
    BPTR lock, old;

    if (cwd == NULL || dirname == NULL) {
	return -1;
    }
    if (dirname[0] == 0) {
	return 0;		/* nothing to do */
    }

    lock = Lock((CONST_STRPTR)dirname, ACCESS_READ);
    if (lock == (BPTR)0) {
	return -1;
    }

    old = CurrentDir(lock);
    if (*cwd != NULL) {
	UnLock(old);		/* we only keep oldest */
    } else if (old != (BPTR)0) {
	*cwd = (Cwd)old;
    } else {
	*cwd = (Cwd)-1;		/* encode NULL lock as -1 */
    }
    return 0;
}

int PopDir(Cwd *cwd)
{
    if (cwd && *cwd) {
	BPTR old = (BPTR)*cwd, lock;

	if (old == (BPTR)-1) {
	    old = (BPTR)0;
	}
	lock = CurrentDir(old);
	UnLock(lock);
	*cwd = NULL;
    }
    return 0;
}

int RenamePath(const char *oldpath, const char *newpath) 
{
    return rename(oldpath, newpath);
}

bool SameFile(const char *filename1, const char *filename2)
{
    BPTR l1, l2;
    bool result = FALSE;

    if(filename1 && filename2)
    {
        l1 = Lock((STRPTR)filename1, ACCESS_READ);
        l2 = Lock((STRPTR)filename2, ACCESS_READ);

        if(l1 && l2 && SameLock(l1, l2) == LOCK_SAME)
            result = TRUE;

        UnLock(l2);
        UnLock(l1);
   }
   return(result);
}

bool FileLocked(const char *filename)
{
    bool result = FALSE;
    BOOL success;

    if (filename) {
        success = DeleteFile((STRPTR)filename);
        if (!success && IoErr() != ERROR_OBJECT_NOT_FOUND) {
            result = TRUE;
        }
    }
    return result;
}

struct Lock *FileLockExclusive(const char *filename, bool *pIsLocked)
{
    BPTR file = (BPTR)0;
    bool isLocked = FALSE;

    if (filename) {
        file = Open((STRPTR)filename, MODE_NEWFILE);
        if (file == (BPTR)0 && IoErr() == ERROR_OBJECT_IN_USE) {
            isLocked = TRUE;
        }
    }
    if (pIsLocked != NULL) {
        *pIsLocked = isLocked;
    }
    return (struct Lock *)file;
}

struct Lock *FileLockShared(const char *filename)
{
    BPTR file = (BPTR)0;

    if (filename) {
        file = Open((STRPTR)filename, MODE_READWRITE);
    }
    return (struct Lock *)file;
}

void FileUnlock(struct Lock *file)
{
    char filename[80];

    if (file) {
        if (!NameFromFH((BPTR)file, (STRPTR)filename, sizeof(filename))) {
            filename[0] = 0;
        }
        Close((BPTR)file);
        if (filename[0] != 0) {
            DeleteFile((STRPTR)filename);
        }
    }
}

static struct Process *FindProcess(const char *name)
{
    struct Process *process = NULL;
    struct Task *task;

    Forbid();
    task = FindTask((CONST_STRPTR)name);
    Permit();

    if (task != NULL && task->tc_Node.ln_Type == NT_PROCESS) {
        process = (struct Process *)task;
    }

    return process;
}

static struct CommandLineInterface *FindProcessCli(const char *name)
{
    struct CommandLineInterface *cli = NULL;
    struct Process *process = FindProcess(name);

    if (process != NULL) {
        cli = BADDR(process->pr_CLI);
    }

    return cli;
}

#if (!defined __amigaos4__)
static APTR SetProcWindow(APTR win)
{
    struct Process *this = FindProcess(NULL);
    APTR oldwin = NULL;

    if (this != NULL) {
        oldwin = this->pr_WindowPtr;
        this->pr_WindowPtr = win;
    }
    return oldwin;
}
#endif

static BPTR LockNoReq(const char *filename, LONG accessMode) 
{
    APTR oldwin = SetProcWindow((APTR)-1);
    BPTR lock = Lock((STRPTR)filename, accessMode);
    SetProcWindow(oldwin);

    return lock;
}

#define ALIGN(class, type, var) \
    class char var ## mem[sizeof(type) + 4]; \
    type *var = (type *)((((long)var ## mem + 3) / 4) * 4)

short FileStat(const char *filename)
{
    BPTR lock;
    BOOL rc;
    ALIGN(static, struct FileInfoBlock, fib);

    if (filename == NULL) {
        return -1;                              /* does not exist */
    }

    lock = LockNoReq(filename, SHARED_LOCK);
    if (lock == (BPTR)0) {
        if (IoErr() == ERROR_OBJECT_IN_USE) {
	    return 127;			/* can't examine but file/dir exists */
	}
        return -1;                      /* no such file or directory */
    }

    rc = Examine(lock, fib);
    UnLock(lock);
     
    if (!rc) {
        return -1;                             /* can't examine */
    }

    if (fib->fib_DirEntryType > 0) {
        return 0;                              /* directory */
    }

    return 1;                                  /* file */
}

time_t FileModificationTime(const char *filename)
{
    BPTR lock;
    BOOL rc;
    APTR oldwin;
    ALIGN(static, struct FileInfoBlock, fib);
    LONG days;
    LONG mins;
    LONG secs;

    if (filename == NULL) {
        return TIME_T_INVALID;                  /* does not exist */
    }

    oldwin = SetProcWindow((APTR)-1);
    lock = Lock((STRPTR)filename, SHARED_LOCK);
    SetProcWindow(oldwin);

    if (lock == (BPTR)0) {
        return TIME_T_INVALID;                  /* can't lock */
    }

     rc = Examine(lock, fib);
     UnLock(lock);
     
     if (!rc) {
         return TIME_T_INVALID;                 /* can't examine */
     }

     days = 8 * 365 + 2;                        /* 1.1.1970 to 1.1.1978 */
     days += fib->fib_Date.ds_Days;
     mins = days * 24 * 60 + fib->fib_Date.ds_Minute;
     secs = mins * 60 + fib->fib_Date.ds_Tick / TICKS_PER_SECOND;

     return (time_t)secs;
}

void UpdateFileModificationTime(const char *filename)
{
    if (filename != NULL) {
        APTR oldwin;
        struct DateStamp ds;

        DateStamp(&ds);
        oldwin = SetProcWindow((APTR)-1);
        SetFileDate((CONST_STRPTR)filename, &ds);
        SetProcWindow(oldwin);
    }
}

bool FileIsWritable(const char *filename)
{
    BPTR lock;
    BOOL rc;
    ALIGN(static, struct FileInfoBlock, fib);

    if (filename == NULL) {
        return FALSE;                           /* does not exist */
    }

    lock = LockNoReq(filename, SHARED_LOCK);
    if (lock == (BPTR)0) {
        return FALSE;
    }

    rc = Examine(lock, fib);
    UnLock(lock);
     
    if (!rc) {
        return FALSE;                           /* can't examine */
    }

    return (fib->fib_Protection & FIBF_WRITE) != 0;
}

bool IsPathSeparator(int c)
{
    return (bool)(c == '/');
}

bool SplitPathname(const char *path, struct MemPool *memPool, 
        char **pdir, char **pfile)
{
    const char *s;

    if(pfile != NULL)
        *pfile = NULL;
    if(pdir != NULL)
        *pdir = NULL;

    if(path == NULL)
        return(TRUE);                   /* nothing to do */

    s = CodesetStrrchr(path, '/');
    if(s == NULL)
	s = CodesetStrrchr(path, ':');
    if(s != NULL)
	s++;
    else
	s = path;

    if(pdir != NULL)
    {
        long len = s - path;

	if(len > 0 && *(s - 1) == '/')
	    len--;
        *pdir = AllocStringN(memPool, path, len);
        if(*pdir == NULL)
            return(FALSE);
    }

    if(pfile != NULL)
    {
        *pfile = AllocString(memPool, s);
        if(*pfile == NULL)
        {
            if(pdir != NULL && *pdir != NULL)
            {
                FreeString(memPool, *pdir);
                *pdir = NULL;
            }
            return(FALSE);
        }
    }

    return(TRUE);
}

char *JoinPathname(struct MemPool *memPool, const char *dir, const char *file)
{
    const char *mid = "";
    long dirlen, filelen;
    char *s;

    dir = EmptyNullStr(dir);
    file = EmptyNullStr(file);

    dirlen = strlen(dir);
    filelen = strlen(file);

    if(filelen > 0 && dirlen > 0 && dir[dirlen-1] != '/' && dir[dirlen-1] != ':')
        mid = "/";
    else if(filelen == 0 && dirlen > 0 && dir[dirlen-1] == '/')
        dirlen--;

    s = PoolAllocMem(memPool, dirlen + strlen(mid) + filelen + 1);
    if(s == NULL)
        return(NULL);

    sprintf(s, "%.*s%s%s", (int)dirlen, dir, mid, file);
    return(s);
}

Dir *DirOpen(const char *filename)
{
    return (Dir *)opendir(filename);
}

const char *DirReadName(Dir *dir)
{
    struct dirent *entry = readdir((DIR *)dir);

    if (entry) {
        return entry->d_name;
    }
    return NULL;
}

void DirClose(Dir *dir)
{
    closedir((DIR *)dir);
}

/* DosProcess */

/* From YAM / Thore Böckelmann */

#if (!defined __amigaos4__)

#ifndef MEMF_SHARED
#define MEMF_SHARED MEMF_ANY
#endif

struct PathNode
{
    BPTR pn_Next;
    BPTR pn_Lock;
};

static void FreePathList(BPTR path)
{
    struct PathNode *pn = BADDR(path), *next;

    while (pn != NULL) {
        next = BADDR(pn->pn_Next);
        UnLock(pn->pn_Lock);
        FreeVec(pn);
        pn = next;
    }
}

static BPTR CopyPathList(const BPTR path)
{
    struct PathNode *src = BADDR(path);
    struct PathNode *this;
    BPTR first = (BPTR)0;
    BPTR *prev = &first;
    APTR oldwin;

    /*
    ** prevent any error requesters from popping up during the copy process,
    ** the DupLock() call might cause this
    */
    oldwin = SetProcWindow((APTR)-1);

    for(; src; src = BADDR(src->pn_Next)) {
        this = AllocVec(sizeof(*this), MEMF_SHARED);
        if (this != NULL) {
            // duplicate the path node, a failed DupLock() causes *NO* harm
            this->pn_Lock = DupLock(src->pn_Lock);
            this->pn_Next = (BPTR)0;
            *prev = MKBADDR(this);
            prev = &this->pn_Next;
        } else {
            FreePathList(first);
            first = (BPTR)0;
            break;
        }
    }

    /* restore the old error requester behaviour */
    SetProcWindow(oldwin);

    return first;
}

#endif

static BPTR ObtainSearchPath(void)
{
    BPTR path = (BPTR)0;

    if (WorkbenchBase != NULL && WorkbenchBase->lib_Version >= 44) {
        if (!WorkbenchControl(NULL, 
                    WBCTRLA_DuplicateSearchPath, (IPTR)&path, 
                    TAG_DONE)) {
            /*  eliminate any possibly modified path pointer if call failed */
            path = (BPTR)0;
        }
    }

#if !defined(__amigaos4__)
    /*
    ** If we couldn't obtain a duplicate copy of the workbench search
    ** path here it is very likely that we are running on a system with
    ** workbench.library < 44 or on MorphOS with an old workbench.lib.
    */
    if (path == (BPTR)0) {
        struct CommandLineInterface *cli;

        cli = FindProcessCli(NULL);
        if (cli == NULL || cli->cli_CommandDir == (BPTR)0) {
            cli = FindProcessCli("Workbench");
        }
        if (cli != NULL) {
            path = CopyPathList(cli->cli_CommandDir);
        }
    }
#endif

    return path;
}

static void ReleaseSearchPath(BPTR path)
{
    if (path != (BPTR)0) {
        if (WorkbenchBase != NULL && WorkbenchBase->lib_Version >= 44) {
            if (WorkbenchControl(NULL, WBCTRLA_FreeSearchPath, path, TAG_DONE)){
                path = (BPTR)0;
            }
        }

#if !defined(__amigaos4__)
        FreePathList(path);
#endif
    }
}

struct DosProcess
{
    char *cmd;
    long stackSize;
    struct Process *process;
    bool exited;
    int rc;
    bool destroyOnExit;
};

static long dosProcessesRunning = 0;
static struct DosProcess *newProcess = NULL;

static void HOOK_PROTO(processExited, GuiObject app, ulong *params)
{
    struct DosProcess *process = (struct DosProcess *)params[0];
    int rc = params[1];

    /* printf("processExited: %d\n", rc); */

    process->rc = rc;
    process->exited = TRUE;

    if (process->destroyOnExit) {
        CloseDosProcess(process);
    }

    dosProcessesRunning--;
    if (GuiApplication != NULL) {
        DoMethod(GuiApplication, MUIM_Application_ReturnID, ID_PROCESS_EXIT);
    }
}

static HOOK_STRUCT(processExited)

/*
** Here we need the INTERRUPT keyword do turn off stack checking
** Alternatively, we could compile this file without stack checking.
*/
static int INTERRUPT SAVEDS processEntry(void)
{
    struct DosProcess *process = newProcess;
    int rc;

    /* printf("cmd: %s\n", cmd); */
#if defined(__MORPHOS__)
    rc = SystemTags((STRPTR)process->cmd, 
            SYS_UserShell, TRUE,
            NP_Name, (IPTR)"BeeBase-sync-launcher",
            NP_PPCStackSize, process->stackSize,
            TAG_END);
#else
    rc = SystemTags((STRPTR)process->cmd, 
            SYS_UserShell, TRUE,
            NP_Name, (IPTR)"BeeBase-sync-launcher",
            NP_StackSize, process->stackSize,
            TAG_END);
#endif
    if (GuiApplication != NULL) {
        DoMethod(GuiApplication, MUIM_Application_PushMethod, GuiApplication, 4,
                MUIM_CallHook, &processExitedHook, process, rc);
    }
    return 0;
}

struct DosProcess *SpawnDosProcess(const char *cmdline, bool useShell, 
        bool needRC)
{
    bool quiet = TRUE;
    struct CommandLineInterface *cli;
    BPTR path = (BPTR)0;
    static const long minStackSize = STACK_SIZE;

    if (cmdline == NULL) {
        return NULL;
    }

#if 0
    printf("SpawnDosProcess: <%s>, %s %s\n", cmdline,
            useShell? "shell" : "direct", needRC? "rc" : "ignore");
#endif

    newProcess = calloc(1, sizeof(struct DosProcess));
    if (newProcess == NULL) {
        return NULL;
    }

    newProcess->rc = -1;

    /* we pass the cmdline through the newProcess variable */
    newProcess->cmd = strdup(cmdline);
    if (newProcess->cmd == NULL) {
        free(newProcess);
        return NULL;
    }

    cli = FindProcessCli(NULL);

#if defined(__MORPHOS__)
     NewGetTaskAttrsA(NULL, 
             &newProcess->stackSize, sizeof(newProcess->stackSize), 
             TASKINFOTYPE_STACKSIZE, NULL);
#else
    if (cli != NULL) {
        newProcess->stackSize = cli->cli_DefaultStack * 4;
    } else {
        struct Process *this = FindProcess(NULL);

        if (this != NULL) {
            newProcess->stackSize = this->pr_StackSize;
        }
    }
#endif

    if (newProcess->stackSize < minStackSize) {
        newProcess->stackSize = minStackSize;
    }

    if (cli == NULL || cli->cli_CommandDir == (BPTR)0) {
        path = ObtainSearchPath();
    }

    if (needRC) {
#if defined(__MORPHOS__)
        newProcess->process = CreateNewProcTags(
                NP_Entry, (IPTR)processEntry,
                NP_Name, (IPTR)"BeeBase-launcher",
                NP_Output, quiet? Open((STRPTR)"NIL:", MODE_OLDFILE) : Output(),
                NP_CloseOutput, quiet,
                NP_Input, quiet? Open((STRPTR)"NIL:", MODE_OLDFILE) : Input(),
                NP_CloseInput, quiet,
                NP_StackSize, minStackSize, /* for MUI */
                NP_Cli, TRUE,       /* needed or some commands are not found */
                path? NP_Path : TAG_IGNORE, path,
                NP_CommandName, (IPTR)"BeeBase-launcher",
                NP_CodeType, CODETYPE_PPC,
                TAG_END);
#else
        newProcess->process = CreateNewProcTags(
                NP_Entry, (IPTR)processEntry,
                NP_Name, (IPTR)"BeeBase-launcher",
                NP_Output, quiet? Open((STRPTR)"NIL:", MODE_OLDFILE) : Output(),
                NP_CloseOutput, quiet,
                NP_Input, quiet? Open((STRPTR)"NIL:", MODE_OLDFILE) : Input(),
                NP_CloseInput, quiet,
                NP_StackSize, minStackSize, /* for MUI */
                NP_Cli, TRUE,       /* needed or some commands are not found */
                path? NP_Path : TAG_IGNORE, path,
                NP_CommandName, (IPTR)"BeeBase-launcher",
                TAG_END);
#endif
        if (newProcess->process) {
            dosProcessesRunning++;
        } else {
            ReleaseSearchPath(path);
            free(newProcess->cmd);
            free(newProcess);
            newProcess = NULL;
        }
    } else {
#if defined(__MORPHOS__)
        long rc = SystemTags((STRPTR)newProcess->cmd, 
            SYS_UserShell, TRUE,
            SYS_Input, quiet? Open((STRPTR)"NIL:", MODE_READWRITE) : (BPTR)0,
            SYS_Output, quiet? Open((STRPTR)"NIL:", MODE_READWRITE) : (BPTR)0,
            SYS_Asynch, TRUE,
            NP_Name, (IPTR)"BeeBase-async-launcher",
            path? NP_Path : TAG_IGNORE, path,
            NP_PPCStackSize, newProcess->stackSize,
            TAG_END);
#else
        long rc = SystemTags((STRPTR)newProcess->cmd, 
            SYS_UserShell, TRUE,
            SYS_Input, quiet? Open((STRPTR)"NIL:", MODE_READWRITE) : (BPTR)0,
            SYS_Output, quiet? Open((STRPTR)"NIL:", MODE_READWRITE) : (BPTR)0,
            SYS_Asynch, TRUE,
            NP_Name, (IPTR)"BeeBase-async-launcher",
            path? NP_Path : TAG_IGNORE, path,
            NP_StackSize, newProcess->stackSize,
            TAG_END);
#endif
        if (rc == 0) {
            newProcess->exited = TRUE;
        } else {
            free(newProcess->cmd);
            free(newProcess);
            newProcess = NULL;
            if (rc == -1) {
                ReleaseSearchPath(path);
            }
        }
    }

    return newProcess;
}

bool DosProcessIsExited(struct DosProcess *process)
{
    return (bool)(process? process->exited : TRUE);
}

int DosProcessExitCode(struct DosProcess *process)
{
    return (process? process->rc : -1);
}

void CloseDosProcess(struct DosProcess *process)
{
    if (process) {
        if (process->exited) {
            free(process->cmd);
            free(process);
        } else {
            process->destroyOnExit = TRUE;
        }
    }
}

long DosProcessesRunning(void)
{
    return dosProcessesRunning;
}

