/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** codeset.c
**
** Codeset stands for the default encoding of a character string.
** On Amiga this is an 8bit code set using the system's default encoding.
**
** Creation date: 25.04.2007
** Last modified: 29.01.2009
*/

#ifdef __AROS__
#undef NO_INLINE_STDARG
#endif

#include <codeset.h>
#include <util.h>
#include <memory.h>
#include <error.h>
#include <libraries/locale.h>
#include <libraries/codesets.h>
#include <proto/locale.h>
#include <proto/codesets.h>

#include "codeset_amiga.h"

#if defined(__MORPHOS__)
#include <proto/keymap.h>

/* Conversion from ISO-8859-1 to UTF-8 does not work properly in MorphOS 2.0/2.1.
 * Bug is fixed in MorphOS 2.2 (keymap.library V51.5 or newer).
 */
#define MORPHOS_CAN_UTF8 (KeymapBase->lib_Version > 51 || (KeymapBase->lib_Version >= 51 && KeymapBase->lib_Revision >= 5))

static struct KeyMap *keymap = NULL;
#endif

#define CODESETS_MIN 6		/* Ver 6 introduces CSA_ErrPtr */
#define CODESETS_MIN_STR "6"

struct Library *CodesetsBase = NULL;
static struct codeset *DefaultCS = NULL;

#ifdef __amigaos4__
struct CodesetsIFace *ICodesets = NULL;
#endif

static cschar upchar[256];       /* contains upper case letters */
static cschar lowchar[256];      /* contains lower case letters */

const char *CodesetInit(void)
{
    int i;

    for (i = 0; i < 256; i++) {
	upchar[i] = (cschar)ConvToUpper(Locale, i);
	lowchar[i] = (cschar)ConvToLower(Locale, i);
    }

#if defined(__MORPHOS__)
    if (MORPHOS_CAN_UTF8) {
	keymap = AskKeyMapDefault();
	return NULL;		/* handle UTF-8 conversion internally */
    }
#endif

    CodesetsBase = OpenLibrary((CONST_STRPTR)CODESETSNAME, CODESETS_MIN);
    if (CodesetsBase == NULL) {
        return "Failed to open " CODESETSNAME " V" CODESETS_MIN_STR ".";
    }
#ifdef __amigaos4__
    if( NULL == (ICodesets = (struct CodesetsIFace*) GetInterface( CodesetsBase, "main", 1L, NULL )) ) {
	return "Failed to interface " CODESETSNAME " V" CODESETS_MIN_STR ".";
    }
#endif

    DefaultCS = CodesetsFindA(NULL, NULL);

    return NULL;
}

void CodesetExit(void)
{
    DefaultCS = NULL;

#ifdef __amigaos4__
    if( ICodesets ) {
        DropInterface( (struct Interface*)ICodesets );
        ICodesets = NULL;
    }
#endif

    CloseLibrary(CodesetsBase);   /* NULL ok here */
    CodesetsBase = NULL;
}

bool StrIsCodeset(const char *str)
{
    return TRUE;		/* native 8bit is always valid */
}

#if defined(__MORPHOS__)
static ULONG _UTF8_DecodeSafe(CONST_STRPTR ptr, WCHAR *ucs4)
{
    ULONG octets = UTF8_Decode(ptr, ucs4);

    /* Convert invalid UTF-8 sequences to '?' characters */
    switch (octets) {
	case 0:
	case 1: // Always OK
	    break;

	case 2: // Valid range is 0x80...0x7ff
	    if (*ucs4 < 0x80) {
     		*ucs4 = '?';
	    }
	    break;

	case 3: // Valid range is 0x800...0xffff
	    if (*ucs4 < 0x800) {
		*ucs4 = '?';
	    }

	    // Code points U+D800 - U+DFFF may never appear in UTF-8
	    if (*ucs4 >= 0xd800 && *ucs4 <= 0xdfff) {
		*ucs4 = '?';
	    }

	    // Code points U+FFFE and U+FFFF are legal but should be removed
	    if (*ucs4 == 0xfffe || *ucs4 == 0xffff) {
		*ucs4 = ' ';
	    }
      	    break;

	case 4: // Valid range is 0x10000...0x10ffff
	    if (*ucs4 < 0x10000 || *ucs4 > 0x10ffff) {
		*ucs4 = '?';
	    }
	    break;
    }

    return octets;
}
#endif

static const char *StrGetUTF8Trail(const char *in)
{
    const char *end, *s;
    signed char mask = 0x80;            /* -128 */
    int count;

    if (in == NULL) {
        return NULL;
    }
    end = in + strlen(in);
    if (end == in) {
        return end;
    }
    s = end;
    if ((s[-1] & mask) == 0) {          /* 7bit ASCII? */
        return end;
    }
    /* count up to 4 trailing 10xxxxxx characters */
    for (count = 0; count < 4; count++) {
        mask >>= 1;                     /* right shift with sign extension */
        if ((*--s & 0xC0) != 0x80) {
            break;
        }
        if (s == in) {
            return end;                 /* not valid UTF-8 */
        }
    }
    if ((*s & 0xFE) == 0xFE) {          /* 1111111x byte is illegal */
        return end;                     /* not valid UTF-8 */
    }
    if ((*s & mask) != mask) {          /* mask is 1...10...0 */
        return end;                     /* not partial UTF-8 sequence */
    }

    return s;                           /* start of partial UTF-8 sequence */
}

bool StrIsUTF8AllowTrail(const char *in, const char **pend)
{
    const char *end = StrGetUTF8Trail(in);

    if (pend) {
        *pend = end;
    }
    if (in == NULL) {
        return FALSE;
    }
    if (end == in) {
        return TRUE;
    }

#if defined(__MORPHOS__)
    if (MORPHOS_CAN_UTF8) {
        CONST_STRPTR ptr;
        ULONG octets;
	WCHAR wc;

	ptr = in;
        do {
            octets = _UTF8_DecodeSafe(ptr, &wc);
            if (wc == '?' && *ptr != '?') {
                return FALSE;           /* invalid UTF-8 */
            }
            ptr += octets;
        } while (octets > 0 && ptr < (CONST_STRPTR)end);
        return (bool)(ptr == (CONST_STRPTR)end);
    }
#endif
    /*
    ** There is no function in codesets.library that tests for UTF8.
    ** The CodesetsIsLegalUTF8* functions only test the first unicode 
    ** character (represented in UTF-8) of the input string.
    ** CodesetsIsValidUTF8 does not allow ASCII characters in the string.
    */ 
    do {
        UTF32 ucs, *ucsp = &ucs;
        ULONG res;

        res = CodesetsConvertUTF8toUTF32((const UTF8 **)&in, (const UTF8 *)end,
                &ucsp, ucsp+1, CSF_StrictConversion);
        if (res != CSR_ConversionOK && res != CSR_TargetExhausted) {
            return FALSE;
        }
    } while (in < end);

    return TRUE;
}

static char *CodesetConvertFromUTF8Help(const char *in, struct MemPool *memPool,
        bool allowTrail, long *perr, size_t *pwarnings, const char **pend)
{
    const char *end;
    long error = OK;
    ULONG warnings = 0;
    char *out, *tmp;
    ULONG size;

    if (allowTrail) {
        end = StrGetUTF8Trail(in);
    } else if (in != NULL) {
        end = in + strlen(in);
    } else {
        end = NULL;
    }

    if (StrIsASCII(in)) {
        out = (char *)in;
#if defined(__MORPHOS__)
    } else if (MORPHOS_CAN_UTF8) {
        CONST_STRPTR ptr;
	STRPTR t;
        ULONG octets;
	WCHAR wc;
	char ansi;

	ptr = in;
        size = 0;
        do {
            size++;
            if (ptr >= (CONST_STRPTR)end) {
                break;
            }
            octets = UTF8_Decode(ptr, &wc);
            ptr += octets;
        } while (octets > 0);

    	out = PoolAllocMem(memPool, size);
	if (out == NULL) {
	    error = ERR_MEM;
	} else {
            ptr = in;
	    t = out;

            do {
                if (ptr >= (CONST_STRPTR)end) {
                    *t++ = 0;
                    break;
                }
		octets = _UTF8_DecodeSafe(ptr, &wc);
		ansi = ToANSI(wc, keymap);
		if (ansi == '?') {
		    if (wc != '?') {
                        warnings++;     /* conversion error */
		    } else if (*ptr != '?') {
			warnings++;     /* invalid utf-8 */
		    }
		}
                *t++ = ansi;
                ptr += octets;
            } while (octets > 0);
        }
#endif
    } else {
	size = CodesetsUTF8Len((unsigned char *)in) + 1;
	out = PoolAllocMem(memPool, size);
	if (out == NULL) {
	    error = ERR_MEM;
	} else {
	    CodesetsUTF8ToStr(CSA_Source, (IPTR)in, 
                    CSA_SourceLen, (IPTR)(end - in),
		    CSA_Dest, (IPTR)out,
		    CSA_DestLen, size,
		    CSA_DestCodeset, (IPTR)DefaultCS,
		    CSA_AllocIfNeeded, FALSE,
		    CSA_ErrPtr, (IPTR)&warnings,
		    TAG_DONE);
            out[size-1] = 0;                /* make sure we are terminated */
	    if (strlen(out) + 1 != size) {  /* check memory size */
                tmp = out;
                out = AllocString(memPool, tmp);
                if (out == NULL) {
                    error = ERR_MEM;
                }
                PoolFreeMem(memPool, tmp, size);
            }
	}
    }

    if (pend) {
        *pend = end;
    }
    if (perr) {
        *perr = error;
    }
    if (pwarnings) {
        *pwarnings = warnings;
    }
    return out;
}

char *CodesetConvertFromUTF8(const char *in, struct MemPool *memPool, 
        long *perr, size_t *pwarnings)
{
    return CodesetConvertFromUTF8Help(in, memPool, FALSE, perr, pwarnings,NULL);
}

char *CodesetConvertFromUTF8AllowTrail(const char *in, struct MemPool *memPool, 
        long *perr, size_t *pwarnings, const char **end)
{
    return CodesetConvertFromUTF8Help(in, memPool, TRUE, perr, pwarnings, end);
}

char *CodesetConvertFromLocale(const char *in, struct MemPool *memPool, 
        long *perr, size_t *pwarnings)
{
    if (pwarnings) {
        *pwarnings = 0;
    }
    return CodesetConvertFromDefault8Bit(in, memPool, perr);
}

char *CodesetConvertFromLocaleAllowTrail(const char *in,struct MemPool *memPool,
        long *perr, size_t *pwarnings, const char **end)
{
    if (end) {
        *end = in? (in + strlen(in)) : NULL;
    }
    return CodesetConvertFromLocale(in, memPool, perr, pwarnings);
}

char *CodesetConvertFromDefault8Bit(const char *in, struct MemPool *memPool, 
        long *perr)
{
    if (perr) {
        *perr = OK;
    }
    return (char *)in;
}

#if defined(__MORPHOS__)
/* Compute how many octets are needed when converting 8bit string to UTF-8
 */
static int utf8_octets(const char *p)
{
    ULONG octets, len = 0;
    TEXT dummy[8];

    do {
        WCHAR wc = ToUCS4(*p++, keymap);
        octets = UTF8_Encode(wc, dummy);
        len += octets;
    } while (octets);

    return len;
}
#endif

char *CodesetConvertToUTF8(const char *in, struct MemPool *memPool, 
        long *perr)
{
    long error = OK;
    char *out;
    ULONG size;

    if (StrIsASCII(in)) {
        out = (char *)in;
#if defined(__MORPHOS__)
    } else if (MORPHOS_CAN_UTF8) {
        size = utf8_octets(in) + 1;
    	out = PoolAllocMem(memPool, size);

        if (out == NULL) {
	    error = ERR_MEM;
	} else {
            STRPTR ptr = out;
            ULONG octets;

            do {
                WCHAR wc = ToUCS4(*in++, keymap);
                octets = UTF8_Encode(wc, ptr);
                ptr += octets;
            } while (octets);

				out[size-1] = '\0';	// Workaround for a bug in MorphOS versions <= 2.3
        }
#endif
    } else {
	size = CodesetsStrLen((STRPTR)in, 
		CSA_SourceCodeset, (IPTR)DefaultCS, 
		TAG_DONE) + 1;
	out = PoolAllocMem(memPool, size);
	if (out == NULL) {
	    error = ERR_MEM;
	} else {
	    CodesetsUTF8Create(CSA_Source, (IPTR)in, 
	    	    CSA_SourceCodeset, (IPTR)DefaultCS,
		    CSA_Dest, (IPTR)out,
		    CSA_DestLen, size,
		    CSA_AllocIfNeeded, FALSE,
		    TAG_DONE);
	}
    }

    if (perr) {
        *perr = error;
    }
    return out;
}

char *CodesetConvertToLocale(const char *in, struct MemPool *memPool, 
        long *perr, size_t *pwarnings)
{
    return CodesetConvertToDefault8Bit(in, memPool, perr, pwarnings);
}

char *CodesetConvertToDefault8Bit(const char *in, struct MemPool *memPool, 
        long *perr, size_t *pwarnings)
{
    if (pwarnings) {
        *pwarnings = 0;
    }
    if (perr) {
        *perr = OK;
    }
    return (char *)in;
}

size_t CodesetUTF8Len(const char *s)
{
    size_t len;

    if (s == NULL) {
	len = 0;
#if defined(__MORPHOS__)
    } else if (MORPHOS_CAN_UTF8) {
        len = utf8_octets(s);
#endif
    } else {
	len = CodesetsStrLen((STRPTR)s, 
		CSA_SourceCodeset, (IPTR)DefaultCS, 
		TAG_DONE);
    }
		
    return len;
}

static FILE *utf8File = NULL;
static void HOOK_PROTO(writeUTF8, struct convertMsg *msg, STRPTR buf)
{
    fputs((const char *)buf, utf8File);
}
static HOOK_STRUCT(writeUTF8)

void CodesetWriteUTF8(const char *s, FILE *file)
{
    if (s != NULL && file != NULL) {
#if defined(__MORPHOS__)
	if (MORPHOS_CAN_UTF8) {
            ULONG octets;
            TEXT buffer[8];
	    WCHAR wc;

            while (*s) {
                wc = ToUCS4(*s++, keymap);
                octets = UTF8_Encode(wc, (STRPTR)&buffer);
                buffer[octets] = '\0';
                fputs(buffer, file);
            }
            return;
        }
#endif
	utf8File = file;
	CodesetsUTF8Create(CSA_Source, (IPTR)s,
		CSA_SourceCodeset, (IPTR)DefaultCS,
		CSA_DestHook, (IPTR)&writeUTF8Hook,
		TAG_DONE);
    }
}

short CodesetCharToStr(cschar c, char *out)
{
    if (out) {
	*out = c;
    }
    return 1;
}

bool CodesetIsAlNum(cschar c)
{
    return (bool)IsAlNum(Locale, (unsigned char)c);
}

bool CodesetIsDigit(cschar c)
{
    return (bool)IsDigit(Locale, (unsigned char)c);
}

bool CodesetIsLower(cschar c)
{
    return (bool)IsLower(Locale, (unsigned char)c);
}

bool CodesetIsPunct(cschar c)
{
    return (bool)IsPunct(Locale, (unsigned char)c);
}

bool CodesetIsSpace(cschar c)
{
    return (bool)IsSpace(Locale, (unsigned char)c);
}

bool CodesetIsBrkSpace(cschar c)
{
    return (bool)(IsSpace(Locale, (unsigned char)c) && c != (cschar)0xA0);
}

bool CodesetIsXDigit(cschar c)
{
    return (bool)IsXDigit(Locale, (unsigned char)c);
}

short CodesetDigitValue(cschar c)
{
    if (IsDigit(Locale, (unsigned char)c)) {
	return (short)(c - '0');
    }
    return -1;
}

short CodesetXDigitValue(cschar c)
{
    if (IsDigit(Locale, (unsigned char)c)) {
	return (short)(c - '0');
    }
    if (IsXDigit(Locale, (unsigned char)c)) {
	return (short)(10 + lowchar[(unsigned char)c] - 'a');
    }
    return -1;
}

cschar CodesetToUpper(cschar c)
{
    return upchar[(unsigned char)c];
}

cschar CodesetToLower(cschar c)
{
    return lowchar[(unsigned char)c];
}

static char *codesetStrUpDown(const char *str, struct MemPool *memPool, 
	const cschar *which)
{
    char *out = PoolAllocMem(memPool, strlen(str) + 1), *t;

    if (out) {
	for (t = out; (*t = which[(unsigned char)*str]) != 0; t++, str++) {
	}
    }
    return out;
}

char *CodesetStrUp(const char *str, struct MemPool *memPool)
{
    return codesetStrUpDown(str, memPool, upchar);
}

char *CodesetStrDown(const char *str, struct MemPool *memPool)
{
    return codesetStrUpDown(str, memPool, lowchar);
}

char *CodesetStrNBytesCopy(char *dst, const char *src, size_t nbytes)
{
#if defined(__MORPHOS__)
    stccpy(dst, EmptyNullStr(src), nbytes + 1);
#else
    strncpy(dst, EmptyNullStr(src), nbytes);
    dst[nbytes] = 0;
#endif
    return dst;
}

char *CodesetTrimStrChars(char *s, const char *front, const char *back)
{
   char *t = s;

   /* trailing chars */
   s += strlen(s) - 1;
   if (back) {
       while (s >= t && strchr(back, *s) != NULL) {
	   *s-- = 0;
       }
   } else {
       while (s >= t && IsSpace(Locale, (unsigned char)*s)) {
	   *s-- = 0;
       }
   }

   /* leading chars */
   s = t;
   if (front) {
       while (*s && strchr(front, *s) != NULL) {
	   s++;
       }
   } else {
       while (*s && IsSpace(Locale, (unsigned char)*s)) {
	   s++;
       }
   }
   if (s > t) {
	strcpy(t, s);
   }
   return t;
}

char *CodesetTrimStr(char *s)
{
    return CodesetTrimStrChars(s, NULL, NULL);
}


long CodesetStrNCollate(const char *s1, const char *s2, size_t n, short collate)
{
    return StrnCmp(Locale, (STRPTR)s1, (STRPTR)s2, n, collate);
}

long CodesetStrCollate(const char *s1, const char *s2, short collate)
{
    return StrnCmp(Locale, (STRPTR)s1, (STRPTR)s2, -1, collate);
}

long CodesetStrCmp(const char *s1, const char *s2)
{
    return StrnCmp(Locale, (STRPTR)s1, (STRPTR)s2, -1, SC_ASCII);
}

char *CodesetStrStr(const char *haystack, const char *needle)
{
    size_t l1 = strlen(haystack), l2 = strlen(needle);
    const char *p, *end = haystack + l1 - l2;

    for (p = haystack; p <= end; p++) {
	if (StrnCmp(Locale, (STRPTR)p, (STRPTR)needle, l2, 
		    CS_COLLATE_SIMPLE) == 0) {
	    return (char *)p;
	}
    }
   return NULL;
}

char *CodesetStrPBrk(const char *s, const char *accept)
{
    const char *a;
    char c;

    for (; *s; s++) {
	c = upchar[(unsigned char)*s];
	for (a = accept; *a; a++) {
	    if (c == upchar[(unsigned char)*a]) {
		return (char *)s;
	    }
	}
    }
    return NULL;
}

