/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** Exec.c
**
** Recover from software failure or stack overflow.
** Provide additional OS functions.
**
** Creation date: 02.04.1995
** Last modified: 12.12.2007
*/

#include "system.h"
#include "menuproject.h"
#include "main.h"
#include "project.h"
#include "requesters.h"
#include "appstrings_cat.h"
#include <exec/execbase.h>
#include <setjmp.h>

#if defined(__MORPHOS__)
#include <exec/tasks.h>
#endif

#if (!defined __AROS__) && (!defined __amigaos4__) && (!defined __MORPHOS__)
#define ENABLE_INSTALLEXECTRAP
#endif

long TrapNum, TrapReturn;            /* used by ExecTrap.[as] */

#ifdef ENABLE_INSTALLEXECTRAP

extern struct ExecBase *SysBase;
extern APTR ExecTrapFunc;
static APTR OldTrapCode;
static short ExecTrapInstalled = FALSE;

#endif

#define STR(x) #x

long __stack = STACK_SIZE;
#ifdef __GNUC__
const char *__stk_limit = NULL;
#endif

static jmp_buf StackOverflowJmpBuf;

static void Panic(bool isStackOverflow)
{
    struct Project *pr;
    const char *name;
    int result;

    if (isStackOverflow) {
        fprintf(stderr, "!!! Stack Overflow !!!\n");
    } else {
        fprintf(stderr, "!!! Software Failure $%08lx !!!\n", TrapNum);
    }
    if (MUIMasterBase == NULL) {
	fail(NULL, 30);
    }

    GuiSetSleep(TRUE);             /* disable all user input */

    FORNODE (pr, &ProjectList) {
        name = ProjectName(pr);
        if (name == NULL) {
            name = GCS(MSG_Unnamed);
        }

        if (isStackOverflow) {
            result = Ask(AT_ERROR, GCS(MSG_SoftwareFailureAnswers),
                    GCS(MSG_StackOverflow), name);
        } else {
            result = Ask(AT_ERROR, GCS(MSG_SoftwareFailureAnswers),
                    GCS(MSG_SoftwareFailure), TrapNum, name);
        }

        if (result) {
      	    MenuProjectSaveAs(pr);
        }
    }

    fail(NULL, 20);
}

#ifdef ENABLE_INSTALLEXECTRAP
void ASM ExecTrap(void)
{
    Panic(0);
}
#endif

#if (defined __GNUC__)
void __stkovf(void)
{
    longjmp(StackOverflowJmpBuf, 1);    /* restore stack to smaller size */
}
#endif

const char *InstallExecTrap(void)
{
    struct Task *this = FindTask(NULL);
    int ret;

#ifdef ENABLE_INSTALLEXECTRAP
    if (!ExecTrapInstalled) {
        OldTrapCode = this->tc_TrapCode;
        this->tc_TrapCode = ExecTrapFunc;
        ExecTrapInstalled = TRUE;
    }
#endif

#if defined(__MORPHOS__)
    {
        ULONG stack;

        NewGetTaskAttrsA(NULL, &stack, sizeof(ULONG), TASKINFOTYPE_STACKSIZE, NULL);

        if (stack < __stack)
            return "Stack size too small!  Minimum is " STACK_SIZE_STR " bytes.";
    }
    __stk_limit = (const char *)this->tc_ETask->PPCSPLower + 1024;
#elif defined(__GNUC__)
    if ((char *)this->tc_SPUpper - (char *)this->tc_SPLower < __stack) {
        return "Stack size too small!  Minimum is " STACK_SIZE_STR " bytes.";
    }
    __stk_limit = (const char *)this->tc_SPLower + 1024;
#endif

    ret = setjmp(StackOverflowJmpBuf);
    if (ret != 0) {		/* long jmp from a stack overflow */
	Panic(1);
    }

    return NULL;
}

void RestoreExecTrap(void)
{
#ifdef ENABLE_INSTALLEXECTRAP
    if (ExecTrapInstalled) {
        SysBase->ThisTask->tc_TrapCode = OldTrapCode;
        ExecTrapInstalled = FALSE;
    }
#endif
}

bool ExecIsMorphOS(void)
{
#if defined(__MORPHOS__)
    return TRUE;
#else
    return (bool)(SysBase->LibNode.lib_Version >= 50 && 
	    FindResident((CONST_STRPTR)"MorphOS") != NULL);
#endif
}

