/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** font.c
**
** Creation date: 13.04.1999
** Last modified: 11.07.2020
*/

#include "system.h"
#include <proto/graphics.h>
#include <proto/diskfont.h>
#include <intuition/intuitionbase.h>
#include "font.h"
#include "codeset.h"
#include "memory.h"

struct Library *DiskfontBase = NULL;

static struct MemPool *FontMemPool = NULL;
static struct List FontList;

#define FONT_NAME_MAXLEN 30

struct FontEntry
{
    struct Node fe_Node;
    struct TextAttr fe_TextAttr;
    struct TextFont *fe_TextFont;
};

struct FontSize
{
    bool isRelSize;
    WORD size;
};

static char BaseFontName[FONT_NAME_MAXLEN + 1] = "topaz.font";
static UWORD BaseFontSize = 11;

void FontInit(void)
{
    NewList(&FontList);
    FontMemPool = CreatePool(0L, 4096, 1024);
}

void FontExit(void)
{
    if (FontMemPool != NULL) {
	struct FontEntry *fe;

	while ((fe = (struct FontEntry *)REMHEAD(&FontList)) != NULL) {
	    if (fe->fe_TextFont != NULL) {
		CloseFont(fe->fe_TextFont);
	    }
	    PoolFreeMem(FontMemPool, fe, sizeof(struct FontEntry));
	}
        DeletePool(FontMemPool);
	FontMemPool = NULL;
    }
}

void SetBaseFont(struct TextFont *tf)
{
    if (tf != NULL) {
        const char *name = tf->tf_Message.mn_Node.ln_Name;

        /* printf("BaseFont %s %d\n", name, tf->tf_YSize); */
        if (name != NULL && strlen(name) <= FONT_NAME_MAXLEN) {
            strcpy(BaseFontName, name);
            BaseFontSize = tf->tf_YSize;
        } 
    }
}

static struct TextFont *openTextFontFromTextAttr(struct TextAttr *ta)
{
    struct FontEntry *fe;
    struct TextFont *tf;

    /* printf("Requested font '%s', size %d, style 0x%02x, flags 0x%02x\n", 
            ta->ta_Name, ta->ta_YSize, ta->ta_Style, ta->ta_Flags); */
    FORNODE(fe, &FontList) {
        if (!strcmp((const char *)fe->fe_TextAttr.ta_Name,
                    (const char *)ta->ta_Name) && 
                fe->fe_TextAttr.ta_YSize == ta->ta_YSize &&
                fe->fe_TextAttr.ta_Style == ta->ta_Style &&
                fe->fe_TextAttr.ta_Flags == ta->ta_Flags) {
            /* put entry at head of list */
            REMOVE(fe);
            ADDHEAD(&FontList, fe);
            /* printf("Returning cached font\n"); */
            return fe->fe_TextFont;
        }
    }


    /* printf("Opening font\n"); */
    tf = OpenFont(ta);
    if (tf == NULL) {
        /* printf("Opening diskfont\n"); */
        /* let's try diskfont.library */

        DiskfontBase = OpenLibrary((STRPTR)"diskfont.library", 0);
        if (DiskfontBase != NULL) {
            tf = OpenDiskFont(ta);
            CloseLibrary(DiskfontBase);
            DiskfontBase = NULL;
        }
    }

    if (tf == NULL) {
        /* printf("Font not found\n"); */
        return NULL;            /* no luck */
    }

    fe = PoolAllocMem(FontMemPool, sizeof(struct FontEntry));
    if (fe == NULL) {
        CloseFont(tf);
        return NULL;
    }

    fe->fe_TextAttr = *ta;
    fe->fe_TextAttr.ta_Name = (STRPTR)tf->tf_Message.mn_Node.ln_Name;
    fe->fe_TextFont = tf;
    ADDHEAD(&FontList, fe);

    /* printf("Opened font '%s', size %d, style 0x%02x, flags 0x%02x\n", 
            tf->tf_Message.mn_Node.ln_Name, tf->tf_YSize,
            tf->tf_Style, tf->tf_Flags); */

    return tf;
}

static struct TextFont *openTextFont(const char *fontname, size_t len, 
        const struct FontSize *size, UBYTE style)
{
    struct TextFont *tf = NULL;

    if (len > 0 || size->size != 0) {
        char name[FONT_NAME_MAXLEN + 1];
        struct TextAttr ta;

        memset(&ta, 0, sizeof(ta));
        if (len > 0) {
            sprintf(name, "%.*s.font",
                    (int)MIN(FONT_NAME_MAXLEN - 5, len), fontname);
            ta.ta_Name = (STRPTR)name;
        } else {
            ta.ta_Name = (STRPTR)BaseFontName;
        }
        if (size->isRelSize) {
            ta.ta_YSize = BaseFontSize + size->size;
        } else {
            ta.ta_YSize = size->size;
        }
        ta.ta_Style = style;

        tf = openTextFontFromTextAttr(&ta);
    }

    return tf;
}

static bool getFontSizeAndStyleSpec(const char *s,
        struct FontSize *size, struct FontStyleSpec *spec)
{
    char c = *s;
    char *t;
    long n;
    bool isSign;

    if (c == 0 || c == '/') {
        return FALSE;
    }
    isSign = (c == '+' || c == '-');
    if (isSign || CodesetIsDigit(c)) {
        n = strtol(s, &t, 10);
        if (*t == 0 || *t == '/') {
            size->isRelSize = isSign;
            size->size = n;
            return TRUE;
        }
        return FALSE;
    } else if (s[1] == 0 || s[1] == '/') {
        switch (c) {
            case 'u':
                spec->fss_Style |= FSF_UNDERLINED;
                break;
            case 'b':
                spec->fss_Style |= FSF_BOLD;
                break;
            case 'i':
                spec->fss_Style |= FSF_ITALIC;
                break;
            case 'e':
                spec->fss_Style |= FSF_EXTENDED;
                break;
            case 's':
                spec->fss_Shadow = TRUE;
                break;
            case 'o':
                spec->fss_Outline = TRUE;
                break;
            case 'g':
                spec->fss_Glow = TRUE;
                break;
            case 'n':
                spec->fss_Plain = TRUE;
                break;
            default:
                return FALSE;
        }
        return TRUE;
    } else if (c == 'c' || c == 'C') {
        ++s;
        n = strtol(s, &t, 16);
        if (*t == 0 || *t == '/') {
            if (c == 'c') {
                spec->fss_FrontColor = n;
            } else {
                spec->fss_BackColor = n;
            }
            return TRUE;
        }
        return FALSE;
    }
    return FALSE;
}

struct TextFont *FontTextFont(const char *fontspec, struct FontStyleSpec *fss)
{
    struct TextFont *tf = NULL;
    struct FontSize size;
    struct FontStyleSpec buffer;

    if (fss == NULL) {
        fss = &buffer;
    }
    memset(&size, 0, sizeof(size));
    memset(fss, 0, sizeof(*fss));
    fss->fss_FrontColor = -1;
    fss->fss_BackColor = -1;

    if (fontspec != NULL && fontspec[0] != 0) {
        const char *s = fontspec;
        bool endfound = FALSE;
        size_t len = strlen(fontspec);

        while ((s = CodesetStrchr(s, '/')) != NULL) {
            if (getFontSizeAndStyleSpec(s+1, &size, fss)) {
                if (!endfound) {
                    len = s - fontspec;
                    endfound = TRUE;
                }
            }
            s++;
        }

        tf = openTextFont(fontspec, len, &size, fss->fss_Style);
    }

    return tf;
}

