/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** ansiprint_mui.c
**
** Creation date: 28.05.1995
** Last modified: 29.01.2022
*/

#include "system.h"

#ifdef USE_ANSI_PRINT

#include "queryeditor.h"
#include "project.h"
#include "util.h"
#include "dosutil.h"
#include "memory.h"
#include "requesters.h"
#include "appstrings_cat.h"
#include "extclasses/ExtApplication_mcc.h"
#include <libraries/asl.h>

static const char *delimiterContent[4];
static const char *fontContent[3];
static const char *sizeContent[4];
static const char *afterPrintingContent[4];

#define OUTPUT_LEN 128
#define INIT_LEN 128
#define DELIMITER_LEN 16

struct PrintSettings
{
    char init[INIT_LEN + 1];
    short delimiter;
    char customdelimiter[DELIMITER_LEN + 1];
    short ident;
    bool head, esc, quotes, nlq, ff;
    short size, numlf;
    char output[OUTPUT_LEN + 1];
};

enum
{
    DELIMITER_SPACES, DELIMITER_TABS, DELIMITER_CUSTOM,
};

static struct PrintSettings printerSettings =
{
    "\\33c", DELIMITER_SPACES, "|", 0, 
    TRUE, TRUE, FALSE, TRUE, TRUE,
    1, 0, "PRT:",
};

static char *font[]= { ANSI_NLQON, ANSI_NLQOFF };
static char *pitch[] =
{
    ANSI_CONDENSEDOFF ANSI_ELITEOFF,
    ANSI_CONDENSEDOFF ANSI_ELITEON,
    ANSI_ELITEOFF ANSI_CONDENSEDON,
};

static void printEntry(FILE *file, struct QueryEntry **qes,
        const struct QueryFieldInfo *infos, ushort numFields,
        struct PrintSettings *ps)
{
    ushort i;
    char *pattern = ps->quotes? "\"%s\"" : "%s";

    fprintf(file, "%*s", ps->ident, "");
    for(i = 0; i < numFields && qes[i] != NULL && !ferror(file); i++)
    {
        if(ps->delimiter != DELIMITER_SPACES)
            fprintf(file, pattern, qes[i]->qe_String);
        else if (infos[i].qfi_Format == 0)
        {
            fprintf(file, pattern, qes[i]->qe_String);
            fprintf(file, "%*s",
                    (int)(strlen(qes[i]->qe_String) - infos[i].qfi_Width), "");
        }
        else
        {
            fprintf(file, "%*s", 
                    (int)(strlen(qes[i]->qe_String) - infos[i].qfi_Width), "");
            fprintf(file, pattern, qes[i]->qe_String);
        }
        if(i < numFields - 1)
	{
	    switch(ps->delimiter)
	    {
	    case DELIMITER_SPACES:
		fputc(' ', file);
		break;

	    case DELIMITER_TABS:
		fputc('\t', file);
		break;

	    case DELIMITER_CUSTOM:
		fputs(ps->customdelimiter, file);
		break;
	    }
	}
    }
    fputc('\n', file);
}

static void printEntries(struct Project *pr, Object *sfw, 
			 FILE *file, struct PrintSettings *ps)
{
    ushort numFields = nget(sfw, MUIA_SFWList_NumFields);
    struct QueryLine *ql = pget(sfw, MUIA_SFWList_Titles);
    const struct QueryFieldInfo *infos = pget(sfw, MUIA_SFWList_FieldInfos);
    long i, j;

    if (ql == NULL || infos == NULL) {
        return;                                            /* internal error */
    }

    /* initialize printer */
    if (ps->esc) {
        struct MemPool *memPool = ProjectMemPool(pr);
        char *init = NULL;

        init = StrCompress(memPool, ps->init);
        fprintf(file, "%s%s%s", init? init : ANSI_RESET, 
		font[ps->nlq? 0 : 1], pitch[ps->size]);
        if (init) {
            FreeString(memPool, init);
        }
    }

    /* print header */
    if (ps->head) {
        printEntry(file, ql->ql_Entries, infos, numFields, ps);

        if (ps->delimiter == DELIMITER_SPACES) {
            fprintf(file, "%*s", ps->ident, "");
            for (j = 0; j < numFields; j++) {
                for (i=0; i < infos[j].qfi_Width; i++) {
                    fputc('-', file);
                }
                if (j < numFields - 1) {
                    fputc('-', file);
                }
            }
            fputc('\n', file);
        }
    }

    /* print lines */
    for (i=0; !ferror(file); i++) {
        DoMethod(sfw, MUIM_List_GetEntry, i, &ql);
        if (ql == NULL) {
            break;
        }
        printEntry(file, ql->ql_Entries, infos, numFields, ps);
    }

    /* after printing */
    for (i=0; i < ps->numlf; i++) {
        fputc('\n', file);
    }
    if (ps->ff) {
        fputc('\f', file);
    }
}

struct PrintSFWData
{
    struct Project *d_Project;
    Object *d_SFWList;
    Object *d_Delimiter, *d_CustomDelimiter, *d_Font, *d_Init, *d_Indent;
    Object *d_Headline, *d_EscapeCodes, *d_Quotes;
    Object *d_Size, *d_AfterPrinting, *d_NumLineFeeds, *d_Output;
    Object *d_Ok, *d_Cancel;
};

static void disposePrintSFW(Object *app, Object *obj)
{
    struct PrintSFWData *d = pget(obj, MUIA_UserData);

    if(obj && d)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, 
		    sizeof(struct PrintSFWData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void HOOK_PROTO(closePrintSFW, Object *obj, ulong *ok)
{
    struct PrintSFWData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    struct MemPool *memPool = ProjectMemPool(pr);
    Object *app = pget(obj, MUIA_ApplicationObject);

    if(*ok)
    {
        FILE *file;
        struct PrintSettings ps;
        bool ok = FALSE;

	ps.delimiter = nget(d->d_Delimiter, MUIA_Radio_Active);
	CodesetStrNBytesCopy(ps.customdelimiter, 
		(const char *)pget(d->d_CustomDelimiter, MUIA_String_Contents),
		DELIMITER_LEN);
        ps.nlq = (nget(d->d_Font, MUIA_Radio_Active) == 0);
	CodesetStrNBytesCopy(ps.init, 
		(const char *)pget(d->d_Init, MUIA_String_Contents), INIT_LEN);
        ps.ident = nget(d->d_Indent, MUIA_Numeric_Value);
        ps.head = nget(d->d_Headline, MUIA_Selected);
        ps.esc = nget(d->d_EscapeCodes, MUIA_Selected);
        ps.quotes = nget(d->d_Quotes, MUIA_Selected);
        ps.size = nget(d->d_Size, MUIA_Radio_Active);
        switch(nget(d->d_AfterPrinting, MUIA_Radio_Active))
        {
	case 0:     ps.ff = TRUE; ps.numlf = 0; break;
	case 1:     ps.ff = FALSE;
	    ps.numlf = nget(d->d_NumLineFeeds, MUIA_String_Integer);
	    break;
	default:    ps.ff = FALSE; ps.numlf = 0;
	    break;
        }
	CodesetStrNBytesCopy(ps.output, 
		(const char *)pget(d->d_Output, MUIA_String_Contents), 
		OUTPUT_LEN);

        ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
 	if(FileStat(ps.output) >= 0)
	{
	    if(Ask(AT_WARN, GCS(MSG_AskReplaceAnswers),
		   GCS(MSG_AskReplaceFile), ps.output) == 0)
	    {
		return;			/* user canceled */
	    }
	}

        file = IFOpen(memPool, ps.output, "w");
        if(file == NULL)
	{
            Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_OpenFile), ps.output);
	    return;
	}

	GuiSetSleep(TRUE);

	printEntries(pr, d->d_SFWList, file, &ps);
        if(!ferror(file))
	    fflush(file);
	if(!ferror(file))
	{
	    printerSettings = ps;
	    ok = TRUE;
	}
	else
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_WriteFile), ps.output);
 	fclose(file);		/* fclose == 0 seems not to be enough */

	GuiSetSleep(FALSE);

        if(!ok)
            return;
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposePrintSFW(app, obj);
}

static Object *createPrintSFW(struct Project *pr, Object *sfw)
{
    Object *app = GuiApplication;
    struct MemPool *memPool = ProjectMemPool(pr);
    struct PrintSFWData *d;
    struct PrintSettings *ps = &printerSettings;
    Object *delimiter, *font, *misc, *afterPrinting, *top, *settings;
    Object *size, *popup, *buttons;
    Object *obj;

    if(app == NULL)
        return(NULL);

    d = PoolAllocMemZero(memPool, sizeof(struct PrintSFWData));
    if(d == NULL)
        return(NULL);
    d->d_Project = pr;
    d->d_SFWList = sfw;

    delimiterContent[0] = GCS(MSG_QueryPrint_Spaces);
    delimiterContent[1] = GCS(MSG_QueryPrint_Tabs);
    delimiterContent[2] = GCS(MSG_QueryPrint_Custom);
    delimiterContent[3] = NULL;
    fontContent[0] = GCS(MSG_QueryPrint_NLQ);
    fontContent[1] = GCS(MSG_QueryPrint_Draft);
    fontContent[2] = NULL;
    sizeContent[0] = GCS(MSG_QueryPrint_Pica);
    sizeContent[1] = GCS(MSG_QueryPrint_Elite);
    sizeContent[2] = GCS(MSG_QueryPrint_Condensed);
    sizeContent[3] = NULL;
    afterPrintingContent[0] = GCS(MSG_QueryPrint_FormFeed);
    afterPrintingContent[1] = GCS(MSG_QueryPrint_LineFeeds);
    afterPrintingContent[2] = GCS(MSG_QueryPrint_Nothing);
    afterPrintingContent[3] = NULL;

    d->d_Delimiter = RadioObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_QueryPrint_DelimiterBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_QueryPrint_Delimiter)),
	MUIA_Radio_Entries, delimiterContent,
	End;

    d->d_CustomDelimiter = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, NULL,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, 0,
	MUIA_String_MaxLen, DELIMITER_LEN,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    delimiter = HGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_QueryPrint_Delimiter),
        MUIA_Background, MUII_GroupBack,
	Child, HSpace(0),
	Child, d->d_Delimiter,
	Child, VGroup,
	    Child, VSpace(0),
	    Child, d->d_CustomDelimiter,
	    End,
	End;

    d->d_Font = RadioObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_QueryPrint_FontBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_Font)),
	MUIA_Radio_Entries, fontContent,
	End;

    font = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_Font),
        MUIA_Background, MUII_GroupBack,
	Child, HCenter(d->d_Font),
	End;

    d->d_Init = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_QueryPrint_InitSequenceBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_QueryPrint_InitSequence)),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_String_MaxLen, 128,
	End;
    d->d_Indent = SliderObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_QueryPrint_IndentBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_QueryPrint_Indent)),
	MUIA_Numeric_Min, 0,
	MUIA_Numeric_Max, 20,
	MUIA_Numeric_Value, 0,
	End;
    d->d_Headline = MUI_MakeObject(MUIO_Checkmark,
            (IPTR)GCS(MSG_QueryPrint_Headline));
    d->d_EscapeCodes = MUI_MakeObject(MUIO_Checkmark, 
	    (IPTR)GCS(MSG_QueryPrint_EscapeCodes));
    d->d_Quotes = MUI_MakeObject(MUIO_Checkmark,
            (IPTR)GCS(MSG_QueryPrint_Quotes));

    misc = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_QueryPrint_Misc),
        MUIA_Background, MUII_GroupBack,
	MUIA_Group_Columns, 2,
	Child, VSpace(0),
	Child, VSpace(0),
	Child, Label(GCS(MSG_QueryPrint_InitSequence)),
	Child, d->d_Init,
	Child, Label(GCS(MSG_QueryPrint_Indent)),
	Child, d->d_Indent,
	Child, Label(GCS(MSG_QueryPrint_Headline)),
	Child, 	HGroup,
	    Child, d->d_Headline,
	    Child, HSpace(0),
            End,
	Child, Label(GCS(MSG_QueryPrint_EscapeCodes)),
	Child, HGroup,
	    Child, d->d_EscapeCodes,
	    Child, HSpace(0),
	    End,
	Child, Label(GCS(MSG_QueryPrint_Quotes)),
	Child, HGroup,
	    Child, d->d_Quotes,
	    Child, HSpace(0),
            End,
	Child, VSpace(0),
	Child, VSpace(0),
	End;

    d->d_Size = RadioObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_QueryPrint_SizeBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_Size)),
	MUIA_Radio_Entries, sizeContent,
	End;

    size = HGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_Size),
        MUIA_Background, MUII_GroupBack,
	Child, HSpace(0),
	Child, d->d_Size,
	Child, HSpace(0),
	End;

    d->d_AfterPrinting = RadioObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_QueryPrint_AfterPrintingBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_QueryPrint_AfterPrinting)),
	MUIA_Radio_Entries, afterPrintingContent,
	End;

    d->d_NumLineFeeds = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_QueryPrint_NumLineFeedsBH),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_QueryPrint_NumLineFeeds)),
	MUIA_String_Accept, "-0123456789",
	MUIA_String_MaxLen, 5,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    afterPrinting = HGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_QueryPrint_AfterPrinting),
        MUIA_Background, MUII_GroupBack,
	Child, HSpace(0),
	Child, d->d_AfterPrinting,
	Child, HGroup,
	    Child, Label2(GCS(MSG_QueryPrint_NumLineFeeds)),
	    Child, d->d_NumLineFeeds,
            End,
	Child, HSpace(0),
	End;

    top = VGroup,
	MUIA_Group_Columns, 2,
	Child, VGroup,
	    Child, delimiter,
	    Child, font,
	    End,
	Child, misc,
	Child, size,
	Child, afterPrinting,
	End;

    d->d_Output = StringObject,
        MUIA_CycleChain, TRUE,
	MUIA_ControlChar, Shortcut(GCS(MSG_QueryPrint_Output)),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_String_MaxLen, 80,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    popup = PopaslObject,
        MUIA_ShortHelp, GCS(MSG_QueryPrint_OutputBH),
	MUIA_Popasl_Type, ASL_FileRequest,
	MUIA_Popstring_String, d->d_Output,
	MUIA_Popstring_Button, GuiPopButtonSetCWD(MUII_PopFile, pr),
	End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, d->d_Ok,
	Child, d->d_Cancel,
	End;

    settings = VGroup,
	Child, top,
        Child, HGroup,
	    Child, Label(GCS(MSG_QueryPrint_Output)),
	    Child, popup,
	    End,
	Child, VSpace(8),
	Child, buttons,
	End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
        MUIA_Window_NoMenus, TRUE,
        MUIA_Window_AppWindow, TRUE,
	MUIA_Window_Title, GCS(MSG_Print),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'P','R','Q'),
        MUIA_HelpNode, "Printing queries",
        MUIA_UserData, d,
	WindowContents, settings,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct PrintSFWData));
        return(NULL);
    }

    MUIAddAppMessageNotification(obj, popup);

    SetAttrs(d->d_Headline,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_QueryPrint_HeadlineBH),
	MUIA_Selected, TRUE,
        TAG_DONE);
    SetAttrs(d->d_EscapeCodes,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_QueryPrint_EscapeCodesBH),
	MUIA_Selected, TRUE,
        TAG_DONE);
    SetAttrs(d->d_Quotes,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_QueryPrint_QuotesBH),
        TAG_DONE);
    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE,
        obj, closePrintSFW, FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE,
        obj, closePrintSFW, FALSE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE,
        obj, closePrintSFW, TRUE);

    set(d->d_Delimiter, MUIA_Radio_Active, ps->delimiter);
    set(d->d_CustomDelimiter, MUIA_String_Contents, ps->customdelimiter);
    set(d->d_Font, MUIA_Radio_Active, ps->nlq? 0 : 1);
    set(d->d_Init, MUIA_String_Contents, ps->init);
    set(d->d_Indent, MUIA_Numeric_Value, ps->ident);
    set(d->d_Headline, MUIA_Selected, ps->head);
    set(d->d_EscapeCodes, MUIA_Selected, ps->esc);
    set(d->d_Quotes, MUIA_Selected, ps->quotes);
    set(d->d_Size, MUIA_Radio_Active, ps->size);
    set(d->d_AfterPrinting, MUIA_Radio_Active, ps->ff? 0 : ((ps->numlf > 0)? 1 : 2));
    set(d->d_NumLineFeeds, MUIA_String_Integer, ps->numlf);
    set(d->d_Output, MUIA_String_Contents, ps->output);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposePrintSFW(app, obj);
        return(NULL);
    }

    return(obj);
}

void GuiPrintSFWList(struct Project *pr, GuiObject sfw)
{
    createPrintSFW(pr, sfw);
}

#endif


