/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** Break_mui.c
**
** Creation date: 14.08.1995
** Last modified: 29.12.2020
*/

#include "break.h"
#include "codeset.h"
#include "appstrings_cat.h"
#include <time.h>

#define BREAK_SHORTDELAY (CLOCKS_PER_SEC * 1)
#define BREAK_LONGDELAY (CLOCKS_PER_SEC * 2)

#define BREAK_TITLE_LEN 200

static ulong breakCount = 0;
static bool breakGui = FALSE;

static struct {
    bool present;
    bool progressVisible;
    long abort;
    long pause;
    short percentage;
    GuiObject win;
    GuiObject progressGroup;
    GuiObject progressGauge;
    GuiObject pauseButton;
    GuiObject abortButton;
    clock_t popupTime;
    ulong sigs;
} breakData = { 0 };

static struct BreakSettings breakSettingsInit, breakSettingsCurrent;
static char breakTitleBuffer[BREAK_TITLE_LEN + 1];

static void breakUpdateProgress(void)
{
    if (breakData.progressGauge) {
        const struct BreakSettings *bs = &breakSettingsCurrent;
        bool visible = (bs->steps != 0 && bs->current != 0);

        if (breakData.progressVisible != visible) {
            set(breakData.progressGroup, MUIA_ShowMe, visible);
            breakData.progressVisible = visible;
        }
        if (visible) {
            short percentage = bs->current * 100 / bs->steps;
            if (percentage != breakData.percentage) {
                set(breakData.progressGauge, MUIA_Gauge_Current, percentage);
                breakData.percentage = percentage;
            }
        }
    }
}

static void breakDoSettings(void)
{
    const struct BreakSettings *bs = &breakSettingsCurrent;
        
#if 0
    if (breakData.pauseButton) {
        set(breakData.pauseButton, MUIA_Disabled, !bs->pause);
    }
    if (breakData.abortButton) {
        set(breakData.abortButton, MUIA_Disabled, !bs->abort);
    }
#endif
    if (breakData.win) {
        set(breakData.win, MUIA_Window_Title, bs->title);
    }
    breakUpdateProgress();
}

void BreakInit(struct BreakSettings *bs)
{
    if (breakCount == 0) {
        /* 
        ** We do not GuiSetSleep here as it interfers with browsing
        ** records using the arrow buttons of a panel in a table that
        ** has virtual fields causing program execution.
        */  
	breakGui = (GuiAnyOpenedWindow() != NULL);
        memset(&breakData, 0, sizeof(breakData));
        if (bs) {
            breakSettingsInit = *bs;
            breakGui |= bs->needGui;
        } else {
            memset(&breakSettingsInit, 0, sizeof(breakSettingsInit));
        }
        breakSettingsCurrent = breakSettingsInit;
        if (breakGui) {
            breakData.popupTime = clock() + BREAK_LONGDELAY;
        } else {
            breakData.popupTime = LONG_MAX;
        }
    } else if (bs) {
        struct BreakSettings old = breakSettingsCurrent;
        breakSettingsCurrent = *bs;
        *bs = old;
	if (old.title == breakTitleBuffer) {
	    breakSettingsCurrent.title = breakTitleBuffer;   /* keep message */
	}
        breakDoSettings();
    }

    breakCount++;
}

void BreakExit(const struct BreakSettings *bs)
{
    breakCount--;
    if (breakCount == 0) {
	if (breakData.win) {
	    GuiCloseWindow(breakData.win);
	    MUIRemMember(GuiApplication, breakData.win);
	    MUI_AddDisposeObject(GuiApplication, breakData.win);
            GuiSetSleep(FALSE);
	}
        memset(&breakData, 0, sizeof(breakData));
	breakGui = FALSE;
    } else if (bs) {
        breakSettingsCurrent = *bs;
        breakDoSettings();
    }
}

void BreakHide(void)
{
    GuiCloseWindow(breakData.win);
    breakData.present = FALSE;
    breakData.popupTime = LONG_MAX;
}

void BreakReInit(void)
{
    if (breakGui && !breakData.present) {
        breakData.popupTime = clock() + BREAK_SHORTDELAY;
    }
}

static void breakCreateGui(void)
{
    // 
    // Make sure BeeBase is the active application.
    // We don't want to popup the break window 
    // in front of another active application 
    // except when BeeBase does not have any open window.
    //
    if (!GuiActiveWindow() && GuiAnyOpenedWindow()) {   
        return;
    }

    if (breakData.win == NULL) {
        const struct BreakSettings *bs = &breakSettingsInit;

        breakData.progressGauge = GaugeObject,
            GaugeFrame,
            MUIA_Background, MUII_BACKGROUND,
            MUIA_Gauge_Horiz, TRUE,
            MUIA_Gauge_Max, 100,
            MUIA_Gauge_InfoText, "%ld%%",
            End;
        breakData.progressGroup = VGroup,
            MUIA_ShowMe, FALSE,
            Child, breakData.progressGauge,
            Child, ScaleObject, End,
            End;
        breakData.progressVisible = FALSE;

        if (bs->abort) {
            breakData.abortButton = TextObject,
                ButtonFrame,
                MUIA_InputMode, MUIV_InputMode_RelVerify,
                MUIA_Text_Contents, GCS(MSG_Abort),
                MUIA_Text_PreParse, "\33c",
                MUIA_Text_SetVMax, FALSE,
                MUIA_CycleChain, TRUE,
                End;
        }

        if (bs->pause) {
            breakData.pauseButton = TextObject,
                ButtonFrame,
                MUIA_InputMode, MUIV_InputMode_RelVerify,
                MUIA_Text_Contents, GCS(MSG_Pause),
                MUIA_Text_PreParse, "\33c",
                MUIA_Text_SetVMax, FALSE,
                MUIA_CycleChain, TRUE,
                End;
        }

        breakData.win = WindowObject,
            MUIA_Window_Width, MUIV_Window_Width_Visible(30),
            MUIA_Window_Height, MUIV_Window_Height_Visible(15),
            MUIA_Window_Title, bs->title,
            MUIA_Window_ScreenTitle, GuiScreenTitle,
            MUIA_Window_ID, MAKE_ID('_', 'B', 'R', 'K'),
            MUIA_Window_CloseGadget, FALSE,
            MUIA_Window_NoMenus, TRUE,
            WindowContents, VGroup,
                Child, breakData.progressGroup,
                Child, HGroup,
                    bs->pause? Child : TAG_IGNORE, breakData.pauseButton,
                    bs->abort? Child : TAG_IGNORE, breakData.abortButton,
                    End,
                End,
            End;

        if (breakData.win == NULL) {
            memset(&breakData, 0, sizeof(breakData));
            BreakReInit();          /* set new timer */
            return;
        }
        if (breakData.abortButton) {
            DoMethod(breakData.abortButton, MUIM_Notify, 
                    MUIA_Pressed, FALSE,
                    breakData.abortButton, 3, 
                    MUIM_WriteLong, TRUE, &breakData.abort);
        }
        if (breakData.pauseButton) {
            DoMethod(breakData.pauseButton, MUIM_Notify, 
                    MUIA_Pressed, FALSE,
                    breakData.pauseButton, 3, 
                    MUIM_WriteLong, TRUE, &breakData.pause);
        }

        GuiSetSleep(TRUE);
        MUIAddMember(GuiApplication, breakData.win);
    }

    breakDoSettings();
    GuiOpenWindow(breakData.win);
    DoMethod(GuiApplication, MUIM_Application_NewInput, &breakData.sigs);

    breakData.present = TRUE;
}
    
bool BreakCheck(void)
{
    if (breakGui) {
        if (!breakData.present) {
            if (clock() >= breakData.popupTime) {
                breakCreateGui();
            }
        } else {
            ulong sigs = SetSignal(0, breakData.sigs);

            if ((sigs & breakData.sigs) != 0) {
                breakData.sigs = sigs;
                DoMethod(GuiApplication, MUIM_Application_NewInput, 
                        &breakData.sigs);
            }

            if (breakData.pause) {
                SetAttrs(breakData.pauseButton,
                        MUIA_Text_Contents, GCS(MSG_Continue),
                        MUIA_Text_SetVMax, FALSE,
                        TAG_DONE);

                breakData.sigs = 0;
                breakData.pause = FALSE;
                do {
                    if (breakData.sigs) {
                        UpdateProjects(FALSE);
                        Wait(breakData.sigs);
                    }
                    DoMethod(GuiApplication, MUIM_Application_NewInput, 
                            &breakData.sigs);
                } while (!breakData.pause && !breakData.abort);
                SetAttrs(breakData.pauseButton,
                        MUIA_Text_Contents, GCS(MSG_Pause),
                        MUIA_Text_SetVMax, FALSE,
                        TAG_DONE);
                breakData.pause = FALSE;
            }
        }
    }

    return (bool)breakData.abort;
}

void BreakMessage(const char *str)
{
    struct BreakSettings *bs = &breakSettingsCurrent;
    char *s;

    CodesetStrNBytesCopy(breakTitleBuffer, str, BREAK_TITLE_LEN);
    s = CodesetStrchr(breakTitleBuffer, '\n');
    if(s != NULL) {
        *s = 0;
    }
    bs->title = breakTitleBuffer;
    if (breakData.win != NULL) {
        set(breakData.win, MUIA_Window_Title, bs->title);
    }
}

void BreakSetMaxSteps(ulong steps)
{
    breakSettingsCurrent.steps = steps;
    breakUpdateProgress();
}

ulong BreakCurrentSteps(void)
{
    return breakSettingsCurrent.current;
}

void BreakSetCurrentStep(ulong current)
{
    BreakCheck();
    breakSettingsCurrent.current = current;
    breakUpdateProgress();
}

void BreakAddCurrentStep(long add)
{
    BreakCheck();
    breakSettingsCurrent.current += add;
    breakUpdateProgress();
}

