/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** buildgui_mui.c
**
** Creation date: 11.12.1994
** Last modified: 21.06.2020
*/

#include "system.h"
#include "buildgui_mui.h"
#include "buildguiwindow_mui.h"
#include "buildguitable_mui.h"
#include "buildguifield_mui.h"
#include "preparse_mui.h"
#include "buildgui.h"
#include "structureeditor.h"
#include "display.h"
#include "gui.h"
#include "extclasses/DataTypes_mcc.h"
#include "util.h"
#include "memory.h"
#include "appstrings_cat.h"
#include "amiga/font.h"

/* #define PREACTIVATE_GUIOBJECTS */

static void FreeProjectGuiRec(struct MemPool *memPool, 
        struct DisplayObject *dio, GuiObject app)
{
    struct List *childs = DisplayObjectObjects(dio);
    short type = DisplayObjectType(dio);
    GuiObject go = DisplayObjectGuiObject(dio);
    bool open;

    if (dio) {
        switch (type) {
            case DO_TYPE_WINDOW:
	        if (go && app) {
                    open = GuiWindowOpen(go);   /* remember current state */
                    DoMethod(go, MUIM_KillNotify, MUIA_Window_ActiveObject);
	            set(go, MUIA_Window_ActiveObject, 
                            MUIV_Window_ActiveObject_None);  /* be safe */
	            GuiCloseWindow(go);
                    MUIRemMember(app, go);
                    MUI_AddDisposeObject(app, go);
                    SetDisplayWindowOpen(dio, open);
                    SetDisplayWindowButtonObject(dio, NULL);
                }
                break;

            case DO_TYPE_FIELD:
                BuildGuiFieldRemObject(go);
                SetDisplayFieldListviewObject(dio, NULL);
                SetDisplayFieldFileObject(dio, NULL);
                SetDisplayFieldFontObject(dio, NULL);
                SetDisplayFieldFilterObject(dio, NULL);
                SetDisplayFieldViewObject(dio, NULL);
                SetDisplayFieldPopObject(dio, NULL);
                SetDisplayFieldDisableObject(dio, NULL);
                break;

            case DO_TYPE_TABLE:
                SetDisplayTablePanelNumAllObject(dio, NULL);
                SetDisplayTablePanelFilterObject(dio, NULL);
                break;
	}

        SetDisplayObjectGuiObject(dio, NULL);
        SetDisplayObjectRootObject(dio, NULL);
        SetDisplayObjectActivateObject(dio, NULL);

    	if (childs) {
	    FORNODE(dio, childs) {
	    	FreeProjectGuiRec(memPool, dio, app);
            }
        }
    }
}

void FreeProjectGui(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct DisplayObject *dio = ProjectRootDisplayObject(pr);
    GuiObject app = GuiApplication;

    GuiSetSleep(TRUE);          /* in case there are open asl requests */
    FreeProjectGuiRec(memPool, dio, app);
    GuiSetSleep(FALSE);
}

static bool buildGuiText(struct DisplayObject *dio, 
        GuiObject parent, bool horiz)
{
    struct Project *pr = DisplayObjectProject(dio);
    const char *title = DisplayObjectTitle(dio);
    struct FontStyleSpec fss;
    struct TextFont *tf = FontTextFont(DisplayObjectMUIFontname(dio), &fss);
    const char *bg = DisplayObjectBackground(dio);
    GuiObject text;

    if (parent == NULL) {
        return FALSE;
    }

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    text = TextObject,
        MUIA_Weight, DisplayObjectWeight(dio),
        tf? MUIA_Font : TAG_IGNORE, tf,
        MUIA_Text_Contents, EmptyNullStr(title),
        MUIA_Text_PreParse, PreParseCombine(&fss, "\33c"),
        (bg && bg[0])? MUIA_Background : TAG_IGNORE, bg,
        End;
    if (text == NULL) {
        return FALSE;
    }

    MUIAddMember(parent, text);
    SetDisplayObjectGuiObject(dio, text);
    SetDisplayObjectRootObject(dio, text);
    
    return TRUE;
}

static bool buildGuiImage(struct DisplayObject *dio, 
        GuiObject parent, bool horiz)
{
    struct Project *pr = DisplayObjectProject(dio);
    const char *bg = DisplayObjectBackground(dio);
    char imagesize = DisplayImageImageSize(dio);
    GuiObject img = NULL;

    if (parent == NULL) {
        return FALSE;
    }

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    img = ImageObject,
        MUIA_Weight, DisplayObjectWeight(dio),
        (bg && bg[0])? MUIA_Image_Spec : TAG_IGNORE, bg,
        MUIA_Image_FreeHoriz, (imagesize == DAD_IMAGESIZE_RESIZABLE),
        MUIA_Image_FreeVert, (imagesize == DAD_IMAGESIZE_RESIZABLE),
        MUIA_Image_FontMatch, (imagesize == DAD_IMAGESIZE_FIXED),
        (bg && bg[0] == '5')? MUIA_Font : TAG_IGNORE, Topaz8Font,
        End;
    if (img == NULL) {
        return FALSE;
    }
    MUIAddMember(parent, img);
    SetDisplayObjectGuiObject(dio, img);
    SetDisplayObjectRootObject(dio, img);

    return TRUE;
}

static bool buildGuiSpace(struct DisplayObject *dio, 
        GuiObject parent, bool horiz)
{
    struct Project *pr = DisplayObjectProject(dio);
    bool bar = DisplaySpaceBar(dio);
    const char *bg = DisplayObjectBackground(dio);
    GuiObject space;

    if (parent == NULL) {
        return FALSE;
    }

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    space = RectangleObject,
        MUIA_Weight, DisplayObjectWeight(dio),
	bar? (horiz? MUIA_Rectangle_VBar:MUIA_Rectangle_HBar):TAG_IGNORE, TRUE,
        (bg && bg[0])? MUIA_Background : TAG_IGNORE, bg,
        End;
    if (space == NULL) {
        return FALSE;
    }
    MUIAddMember(parent, space);
    SetDisplayObjectGuiObject(dio, space);
    SetDisplayObjectRootObject(dio, space);

    return TRUE;
}

static bool buildGuiBalance(struct DisplayObject *dio, 
        GuiObject parent, bool horiz)
{
    GuiObject bal;
   
    if (parent == NULL) {
        return FALSE;
    }

    bal = BalanceObject, End;
    if (bal == NULL) {
        return FALSE;
    }
    MUIAddMember(parent, bal);
    SetDisplayObjectGuiObject(dio, bal);
    SetDisplayObjectRootObject(dio, bal);

    return TRUE;
}

static bool BuildGuiObject(struct DisplayObject *dio, bool hideLabel,
	GuiObject group, bool horiz, GuiObject *colGroup);   /* forward decl */

static void HOOK_PROTO(registeractivepagechange, GuiObject go, 
	struct Project **x)
{
    GuiUpdateRegister(go, *x);
}

static bool buildGuiRegister(struct DisplayObject *dio, 
        GuiObject parent, bool horiz)
{
    struct Project *pr = DisplayObjectProject(dio);
    const char *bg = DisplayObjectBackground(dio);
    char **labels = DisplayRegisterLabels(dio);
    struct List *childs = DisplayObjectObjects(dio);
    struct DisplayObject *child;
    GuiObject reg, group, colGroup;

    if (parent == NULL) {
	return FALSE;
    }

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    reg = RegisterObject,
	MUIA_CycleChain, TRUE,
        MUIA_Weight, DisplayObjectWeight(dio),
        MUIA_Background, (bg && bg[0])? bg : (char *)MUII_RegisterBack,
	MUIA_Register_Frame, TRUE,
	MUIA_Register_Titles, labels,
	MUIA_UserData, dio,
	End;
    if (reg == NULL) {
        return FALSE;
    }
    MUIAddMember(parent, reg);
    SetDisplayObjectGuiObject(dio, reg);
    SetDisplayObjectRootObject(dio, reg);
    MUIAddCallback(reg, MUIA_Group_ActivePage, MUIV_EveryTime, 
	    registeractivepagechange, DisplayObjectProject(dio));

    if (childs) {    
	FORNODE(child, childs) {
            if (!DisplayObjectVisible(child)) {
                continue;
            }
            group = VGroup, End;
            if (group == NULL) {
                return FALSE;
            }
            MUIAddMember(reg, group);

	    colGroup = NULL;
	    if (!BuildGuiObject(child, FALSE, group, FALSE, &colGroup)) {
    		return FALSE;
	    }
	    BuildGuiFinishColGroup(&colGroup);
	}
    }
    return TRUE;
}

static bool buildGuiGroup(struct DisplayObject *dio, GuiObject parent,
        bool oldhoriz, GuiObject *colGroup)
{
    struct Project *pr = DisplayObjectProject(dio);
    bool hideLabel = FALSE;
    bool horiz = DisplayGroupHoriz(dio);
    const char *title = DisplayObjectTitle(dio);
    bool border = DisplayGroupBorder(dio);
    bool spacing = DisplayGroupSpacing(dio);
    const char *bg = DisplayObjectBackground(dio);
    struct List *childs = DisplayObjectObjects(dio);
    struct DisplayObject *child;
    GuiObject newColGroup = NULL;
    GuiObject group;

    if (parent == NULL || colGroup == NULL) {
	return FALSE;
    }
    if (title && title[0] == 0) {
	title = NULL;
    }

    if (!oldhoriz && horiz && !border && !title && childs) {
	FORNODE(child, childs) {
	    if (!DisplayObjectVisibleGui(child)) {
		continue;
	    } 
	    if (DisplayObjectType(child) != DO_TYPE_FIELD) {
		break;
	    } 
	    if (DisplayFieldTitlePosition(child)!=DAD_TITLEPOSITION_LEFT) {
		break;
	    }
	    if (!BuildGuiFieldHasLabel(child)) {
		break;
	    }

	    if (*colGroup == NULL) {
                *colGroup = VGroup,
                    MUIA_Group_Columns, 2,
                    DisplayGroupSpacing(DisplayObjectParent(dio))?
                        TAG_IGNORE : MUIA_Group_VertSpacing, 0,
                    End;
                if (*colGroup == NULL) {
                    return FALSE;
                }
                MUIAddMember(parent, *colGroup);
	    }
	    if (!BuildGuiFieldLabel(child, *colGroup, FALSE)) {
	    	return FALSE;
	    }
	    hideLabel = TRUE;
	    parent = *colGroup;
	    break;
	}
    }
    if (!hideLabel) {
	BuildGuiFinishColGroup(colGroup);
    }

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    group = GroupObject,
        oldhoriz? MUIA_HorizWeight : MUIA_VertWeight, DisplayObjectWeight(dio),
	title? MUIA_FrameTitle : TAG_IGNORE, title,
	border?  MUIA_Frame : TAG_IGNORE, MUIV_Frame_Group,
        spacing? TAG_IGNORE : MUIA_Group_Spacing, 0,
 	MUIA_Group_Horiz, horiz != FALSE,
        (border || (bg && bg[0]))? MUIA_Background : TAG_IGNORE,
            (bg && bg[0])? bg : (char *)MUII_GroupBack,
	End;
    if (group == NULL) {
        return FALSE;
    }
    MUIAddMember(parent, group);
    SetDisplayObjectGuiObject(dio, group);
    SetDisplayObjectRootObject(dio, group);

    return BuildGuiChilds(childs, hideLabel, group, horiz, &newColGroup);
}

bool BuildGuiChilds(struct List *childs, bool hideLabel, 
        GuiObject group, bool horiz, GuiObject *colGroup)
{
    if (childs) {    
	struct DisplayObject *child;

	FORNODE(child, childs) {
            if (!BuildGuiObject(child, hideLabel, group, horiz, colGroup)) {
    		return FALSE;
	    }
	    if (hideLabel && DisplayObjectGuiObject(child)) {
		hideLabel = FALSE;
	    }
	}
    }
    BuildGuiFinishColGroup(colGroup);

    return TRUE;
}

void BuildGuiFinishColGroup(GuiObject *colGroup)
{
    if (colGroup) {
	*colGroup = NULL;
    }
}

static bool BuildGuiObject(struct DisplayObject *dio, bool hideLabel,
        GuiObject group, bool horiz, GuiObject *colGroup)
{
    if (!DisplayObjectVisibleGui(dio)) {
	return TRUE;
    }

    switch (DisplayObjectType(dio)) {
	case DO_TYPE_FIELD:
	    return BuildGuiField(dio, hideLabel, group, horiz, colGroup);

	case DO_TYPE_TEXT:
	    BuildGuiFinishColGroup(colGroup);
    	    return buildGuiText(dio, group, horiz);
	    
	case DO_TYPE_IMAGE:
	    BuildGuiFinishColGroup(colGroup);
	    return buildGuiImage(dio, group, horiz);

	case DO_TYPE_SPACE:
	    BuildGuiFinishColGroup(colGroup);
	    return buildGuiSpace(dio, group, horiz);

	case DO_TYPE_BALANCE:
	    BuildGuiFinishColGroup(colGroup);
	    return buildGuiBalance(dio, group, horiz);
	    
	case DO_TYPE_TABLE:
	    BuildGuiFinishColGroup(colGroup);
	    return BuildGuiTable(dio, group, horiz);

	case DO_TYPE_REGISTER:
	    BuildGuiFinishColGroup(colGroup);
	    return buildGuiRegister(dio, group, horiz);

	case DO_TYPE_WINDOW:
	    return BuildGuiWindowButton(dio, group, horiz, colGroup);
	    
	case DO_TYPE_GROUP:
	    return buildGuiGroup(dio, group, horiz, colGroup);
	    
	default:
	    return FALSE;
    }
}

#ifdef PREACTIVATE_GUIOBJECTS
    struct DisplayObject *win_dio;
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    GuiObject win = NULL;
    if (tabs) {
        FORNODE(tab, tabs) {
            dio = DisplayTablePanel(tab);
            win_dio = DisplayObjectWindow(dio);
            win = DisplayObjectGuiObject(win_dio);
            if (win && pget(win, MUIA_Window_ActiveObject) == NULL) {
                set(win, MUIA_Window_ActiveObject, DisplayObjectGuiObject(dio));
            }
        }
    }
#endif

