/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** buildguifield_mui.c
**
** Creation date: 21.08.1996
** Last modified: 10.05.2022
*/

#include "system.h"
#include "buildguifield_mui.h"
#include "buildgui_mui.h"
#include "preparse_mui.h"
#include "extclasses/SFWList_mcc.h"
#include "extclasses/ExtEditor_mcc.h"
#include "extclasses/StringText_mcc.h"
#include "extclasses/DataTypes_mcc.h"
#include "extclasses/RefFieldSelector_mcc.h"
#include "buildgui.h"
#include "programeditor.h"
#include "queryeditor.h"
#include "gui.h"
#include "order.h"
#include "preferences.h"
#include "dosutil.h"
#include "requesters.h"
#include "memo.h"
#include "labels.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"
#include "amiga/font.h"
#include <libraries/asl.h>
#include <intuition/sghooks.h>
#include <workbench/workbench.h>

#define CODE_ARROWUP 0x4c
#define CODE_ARROWDOWN 0x4d

static GuiObject changedstringobject = NULL;
static GuiObject changededitorobject = NULL;

bool BuildGuiFieldLabel(struct DisplayObject *dio, 
        GuiObject parent, bool horiz)
{
    struct GuiFieldFeatures f;
    short pos = DisplayFieldTitlePosition(dio);
    const char *title = DisplayObjectTitle(dio);
    struct FontStyleSpec fss;
    struct TextFont *tf = FontTextFont(DisplayFieldTitleMUIFontname(dio), &fss);
    const char *alignment;
    GuiObject label = NULL;

    if (parent == NULL || !BuildGuiFieldGetFeatures(dio, &f)) {
	return FALSE;
    }
    if (!f.label) {
	return TRUE;	/* nothing to do */
    }

    switch (pos) {
	case DAD_TITLEPOSITION_LEFT:
            alignment = "\33r";
            break;
	case DAD_TITLEPOSITION_RIGHT:
            alignment = "\33l";
	    break;
	default:
            alignment = "\33c";
	    break;
    }

    label = TextObject,
        MUIA_Weight, 0,
        tf? MUIA_Font : TAG_IGNORE, tf,
        MUIA_Text_PreParse, PreParseCombine(&fss, alignment),
        MUIA_Text_Contents, title,
        MUIA_Text_HiChar, f.shortcut,
        MUIA_Text_SetVMax, FALSE,
        End;
    if (label == NULL) {
        return FALSE;
    }

    MUIAddMember(parent, label);

    return TRUE;
}

/* 
** Not 100 % sure if this hook needs the INTERRUPT keyword,
** it might need it as a real string gadget's edit hook might
** be called in a different context than our task.
*/
static LONG INTERRUPT HOOK_PROTO(edit, struct SGWork *sgw, LONG *cmd)
{
    if(*cmd == SGH_KEY)
    {
	short code = sgw->IEvent->ie_Code;
	short qualifier = sgw->IEvent->ie_Qualifier;

	switch(sgw->EditOp)
	{
  	    case EO_NOOP:
	    case EO_MOVECURSOR:
                if(qualifier & (IEQUALIFIER_LALT | IEQUALIFIER_RALT))
                    if(code == CODE_ARROWUP || code == CODE_ARROWDOWN)
		    {
            case EO_RESET:                      /* undo buffers do not make sense here! */
			sgw->Code = code;
			sgw->Actions &= ~(SGA_USE | SGA_BEEP);
			sgw->Actions |= SGA_END | SGA_REUSE;
			return(1);
		    }
		break;

	}
    }
    return(0);
}
static HOOK_STRUCT(edit)

static void HOOK_PROTO(stringcontentschange, GuiObject go, ulong *x)
{
    ulong store = *x;

    changedstringobject = go;
    if(store)
        GuiFieldStoreEdits();
}

static void HOOK_PROTO(stringacknowledge, GuiObject go, struct Project **x)
{
    struct Project *pr = *x;

    GuiStoreEdits();
    if (!GetPrefsValue(pr, PREFS_ADVANCEONENTER)) {
	set(GuiObjectWindow(go), MUIA_Window_ActiveObject, go);
    }
}

static void HOOK_PROTO(checkboxchange, GuiObject go, ulong *x)
{
    struct DisplayObject *dio = NULL;

    GuiStoreEdits();
    get(go, MUIA_UserData, &dio);
    GuiStoreDataBool(dio, *x);
    GuiSetCursor();
}

static void HOOK_PROTO(cyclechange, GuiObject go, ulong *x)
{
    struct DisplayObject *dio = NULL;

    GuiStoreEdits();
    get(go, MUIA_UserData, &dio);
    GuiStoreDataLong(dio, *x);
/*   GuiSetCursor(); */
}

static void HOOK_PROTO(buttonchange, GuiObject go, ulong dummy)
{
    struct DisplayObject *dio = NULL;

    GuiStoreEdits();
    get(go, MUIA_UserData, &dio);
    GuiStoreData(dio);
/*    GuiSetCursor(); */
}

void BuildGuiAddRecord(struct Table *addtab, struct Record *rec)
{
    struct Project *pr = TableProject(addtab);
    struct List * tabs = ProjectTables(pr);
    struct Table *tab;

    if (tabs && addtab && rec) {
        FORNODE(tab, tabs) {
            struct List *fields = TableFields(tab);
            struct Field *field;

            if (fields) {
                FORNODE(field, fields) {
		    if (FieldType(field) == TYPE_REFERENCE &&
			    FieldReferenceTable(field) == addtab) {
			struct DisplayObject *dio= FieldDisplayObject(field);
			char records = DisplayFieldRecordsKind(dio);
			GuiObject go = DisplayFieldPopObject(dio);
			GuiObject pop = pget(go, MUIA_Popobject_Object);
                        GuiObject listview = pop, win;
			struct Record *succ;
			bool filter;
                        long pos;

			if (listview == NULL) {
			    continue;
			}
                        win = GuiObjectWindow(listview);
                        if (!nget(win, MUIA_Window_Open)) {
                            continue;
                        }
			if (records == DAD_RECORDS_ALL) {
			    filter = FALSE;
			} else if (records == DAD_RECORDS_MATCHFILTER &&
				RecordMatchFilter(rec)) {
			    filter = TRUE;
			} else {
			    continue;	/* cannot add this record */
			}
			succ = NextRecord(rec, filter);
			pos = GuiListFindNode(listview, succ);
			if (pos < 0) {
			    pos = MUIV_List_Insert_Bottom;
			}
		    	DoMethod(listview, MUIM_List_InsertSingle, rec, pos);
                    }
                }
	    }
        }
    }
}

void BuildGuiRemRecord(struct Table *addtab, struct Record *rec)
{
    struct Project *pr = TableProject(addtab);
    struct List * tabs = ProjectTables(pr);
    struct Table *tab;

    if (tabs != NULL && addtab != NULL && rec != NULL) {
        FORNODE(tab, tabs) {
            struct List *fields = TableFields(tab);
            struct Field *field;

            if (fields != NULL) {
                FORNODE(field, fields) {
                    short type = FieldType(field);
                    struct DisplayObject *dio = FieldDisplayObject(field);

		    if (type == TYPE_REFERENCE &&
                            FieldReferenceTable(field) == addtab) {
                        GuiObject pop = DisplayFieldPopObject(dio);
			GuiObject listview = pget(pop, MUIA_Popobject_Object);

                        MUIListRemoveEntry(listview, rec);
                    } else if (type == TYPE_VIRTUAL &&
                            DisplayFieldSubType(dio) ==
                            DAD_VIRTUALTYPE_LIST) {
                        GuiObject go = DisplayObjectGuiObject(dio);

                        if (go != NULL) {
                            DoMethod(go, MUIM_SFWList_RemRecord, rec);
			}
                    }
                }
	    }
        }
    }
}

void BuildGuiRemAllRecords(struct Project *pr)
{
    struct List * tabs = ProjectTables(pr);
    struct Table *tab;

    if (tabs != NULL) {
        FORNODE(tab, tabs) {
            struct List *fields = TableFields(tab);
            struct Field *field;

            if (fields != NULL) {
                FORNODE(field, fields) {
                    short type = FieldType(field);
                    struct DisplayObject *dio = FieldDisplayObject(field);

		    if (type == TYPE_REFERENCE) {
                        GuiObject pop = DisplayFieldPopObject(dio);

                        if (pop != NULL) {
                            DoMethod(pop, MUIM_Popstring_Close, FALSE);
                        }
                    } else if (type == TYPE_VIRTUAL &&
                            DisplayFieldSubType(dio) ==
                            DAD_VIRTUALTYPE_LIST) {
                        GuiObject go = DisplayObjectGuiObject(dio);

                        if (go != NULL) {
                            DoMethod(go, MUIM_SFWList_RemAllRecords);
                        }
                    }
                }
	    }
        }
    }
}

void BuildGuiFieldRemObject(GuiObject go)
{
    if (go == changedstringobject) {
        changedstringobject = NULL;
    }
    if (go == changededitorobject) {
        changededitorobject = NULL;
    }
}

#ifdef USE_CALENDAR
static LONG HOOK_PROTO(calendarStrObj, GuiObject calendar, GuiObject str)
{
    long date;
    short day, month, year;

    GuiStoreEdits();        /* maybe the user has just edited a string object */

    if (StringToDate(pget(str, MUIA_String_Contents), &date)) {
        DateToDayMonthYear(date, &day, &month, &year);
        SetAttrs(calendar,
                MUIA_Calendar_Day, day,
                MUIA_Calendar_Month, month,
                MUIA_Calendar_Year, year,
                TAG_DONE);
    }
    return TRUE;
}

static HOOK_STRUCT(calendarStrObj)

static void HOOK_PROTO(calendarObjStr, GuiObject calendar, GuiObject str)
{
    struct DisplayObject *dio = (struct DisplayObject *)GuiUserData(str);
    struct Project *pr = DisplayObjectProject(dio);
    long day = nget(calendar, MUIA_Calendar_Day);
    long month = nget(calendar, MUIA_Calendar_Month);
    long year = nget(calendar, MUIA_Calendar_Year);
    long date;
    const char *text;
   
    if (DayMonthYearToDate(day, month, year, &date)) {
        text = DateToString(date, GetPrefsValue(pr, PREFS_DATEFORMAT));

        /* set string */
        GuiStoreDataStr(dio, text);
    }
}
static HOOK_STRUCT(calendarObjStr)
#endif

static void filefontstopfunc(GuiObject pop, char *drawer, char *file)
{
    GuiObject str = pget(pop, MUIA_Popstring_String);
    GuiObject button = pget(pop, MUIA_Popstring_Button);
    struct DisplayObject *dio = pget(str, MUIA_UserData);
    struct Project *pr = DisplayObjectProject(dio);
    struct MemPool *memPool = ProjectMemPool(pr);
    char *path = NULL;

    if (button) {
        set(button, MUIA_Disabled, FALSE);
    }

    path = JoinPathname(memPool, drawer, file);
    if (path != NULL) {
        GuiStoreDataStr(dio, path);
        FreeString(memPool, path);
    }
}

static void HOOK_PROTO(fileStop, GuiObject pop, struct FileRequester *fr)
{
    if(fr != NULL)
        filefontstopfunc(pop, (char *)fr->fr_Drawer, (char *)fr->fr_File);
}
static HOOK_STRUCT(fileStop)

static void HOOK_PROTO(fontStop, GuiObject pop, struct FontRequester *fo)
{
    if(fo != NULL)
        filefontstopfunc(pop, (char *)fo->fo_Attr.ta_Name, 
		IntToStr(fo->fo_Attr.ta_YSize));
}
static HOOK_STRUCT(fontStop)

static LONG HOOK_PROTO(labelStrObj, GuiObject pop, GuiObject str)
{
    struct DisplayObject *dio = (struct DisplayObject *)GuiUserData(str);
    struct Project *pr = DisplayObjectProject(dio);
    struct MemPool *memPool = ProjectMemPool(pr);
    const char *trigger = DisplayFieldListviewCompute(dio);
    char **labels = NULL;
    bool labelsAllocated = FALSE;

    GuiStoreEdits();        /* maybe the user has just edited a string object */

    if (trigger != NULL && trigger[0] != 0) {
	struct Program *prg = ProjectProgram(pr);
	struct UserFunction *uf = FindUserFunctionName(prg, trigger);
	const char *filename = NULL, *error = NULL, *elemError = NULL;
	long line, col;
	struct Element *elem;
	char *str = NULL;

	if (uf == NULL) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Trigger), trigger);
	    return FALSE;
	} else {
	    /* GuiSetSleep(TRUE); */
	    SetTableProgramRecords(pr);
	    ProgramBegin(prg, TRUE);
	    elem = ProgramCall(prg, uf, NULL, TRUE);
	    if (elem) {
		switch (elem->el_Type) {
		    case TYPE_STRING:
			str = elem->el_Contents.string.string;
			break;
		    case TYPE_MEMO:
			str = elem->el_Contents.mi;
			break;
		    case TYPE_NIL:
			break;
		    default:
			elemError = GCS(MSG_Err_Listview_Compute);
			break;
		}
		if (str) {
		    labels = StringToLabels(memPool, str);
		    if (labels == NULL) {
			elemError = GCS(MSG_Err_Mem);
		    }
		    labelsAllocated = TRUE;
		}
	    }
            ProgramEnd();
	    error = ProgramError(&line, &col, &filename);
	    /* GuiSetSleep(FALSE); */

	    if (error != NULL && error != EXECUTE_ABORT) {
		ProgramEditorShowError(pr, TRUE, error, line, col, filename);
		if (labelsAllocated) {
		    FreeLabels(memPool, labels);
		}
		return FALSE;
	    }
	    if (elemError) {
		Ask(AT_ERROR, GCS(MSG_StarOk), elemError);
		return FALSE;
	    }
	}
    } else {
	labels = DisplayFieldListviewLabels(dio);
    }

    GuiSetSleep(TRUE);
    set(pop, MUIA_List_Quiet, TRUE);
    DoMethod(pop, MUIM_List_Clear);
    if (labels) {
	DoMethod(pop, MUIM_List_Insert, labels, -1, MUIV_List_Insert_Bottom);
    }
    set(pop, MUIA_List_Quiet, FALSE);
    GuiSetSleep(FALSE);

    if (labelsAllocated) {
	FreeLabels(memPool, labels);
    }

    return TRUE;
}
static HOOK_STRUCT(labelStrObj)

static void HOOK_PROTO(labelObjStr, GuiObject pop, GuiObject str)
{
    struct DisplayObject *dio = (struct DisplayObject *)GuiUserData(str);
    const char *label = GuiListActiveEntry(pop);

    /* set string */
    GuiStoreDataStr(dio, label);

    /* clear listview */
    set(pop, MUIA_List_Quiet, TRUE);
    DoMethod(pop, MUIM_List_Clear);
    set(pop, MUIA_List_Quiet, FALSE);
}
static HOOK_STRUCT(labelObjStr)

static void HOOK_PROTO(labelWindow, GuiObject listview, GuiObject win)
{
    GuiObject popup = (GuiObject)GuiUserData(listview);
    GuiObject string = NULL;

    get(popup, MUIA_Popstring_String, &string);
    if (string) {
        struct DisplayObject *dio = (struct DisplayObject *)GuiUserData(string);
        const char *s = NULL;
        struct Field *field = DisplayFieldField(dio);
        short type = FieldType(field);
        struct Table *tab = FieldTable(field);
        struct Record *rec = TableGuiRecord(tab);
        struct DataInfo *di = NULL, di_buf;
        long error;

        RecordLock(rec);
        error = RecordDataInfo(rec, field, &di, &di_buf);
	(void)error;
        if (di && type == TYPE_STRING) {
	     s = di->di_Contents.string;
	}

        set(win, MUIA_Window_Open, TRUE);

        if (s && listview) {
	    long i;
	    const char *elem = NULL;

	    MUIListLoop(listview, i, elem, TRUE) {
		if (elem && strcmp(s, elem) == 0) {
                    MUIListSetActiveEntry(listview, (void *)elem);
                    break;
                }
	    }
	}
        RecordUnlock(rec);
    }

    set(win, MUIA_Window_DefaultObject, listview);
}
static HOOK_STRUCT(labelWindow)

static void HOOK_PROTO(labeldoubleclick, GuiObject listview, 
	struct DisplayObject **x)
{
    GuiObject popup = (GuiObject)GuiUserData(listview);

    /* close popup, this calls the close hook */
    DoMethod(popup, MUIM_Popstring_Close, TRUE);
}

static void HOOK_PROTO(referencefilteronoff, GuiObject go, struct Field **x)
{
    GuiStoreEdits();
    ReferenceFilterOnOff(*x, nget(go, MUIA_Selected));
    GuiSetCursor();
}

static void HOOK_PROTO(referenceautoshowonoff, GuiObject go, 
	struct Field **x)
{
    GuiStoreEdits();
    ReferenceAutoShowOnOff(*x, nget(go, MUIA_Selected));
    GuiSetCursor();
}

static void HOOK_PROTO(show, GuiObject go, void *dummy)
{
    struct DisplayObject *dio = pget(go, MUIA_UserData);
    struct Field *field = DisplayFieldField(dio);
    struct Table *tab = FieldTable(field);

    GuiStoreEdits();
    GuiShowReferenceRecord(field, TableGuiRecord(tab));
    GuiSetCursor();
}

static void HOOK_PROTO(editorhaschanged, GuiObject go, void *dummy)
{
    changededitorobject = go;
}

void HOOK_PROTO(SFWListDoubleClick, Object *obj, struct DisplayObject **x)
{
    struct DisplayObject *dio = *x;
    short col = nget(obj, MUIA_Listview_ClickColumn);
    short num = nget(obj, MUIA_SFWList_NumFields);
    struct QueryLine *ql = NULL;

    DoMethod(obj, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, &ql);
    if (ql != NULL && col >= 0 && col < num)
    {
        struct QueryFieldInfo *infos = pget(obj, MUIA_SFWList_FieldInfos);
	struct Field *colfield = infos ? infos[col].qfi_Field : NULL;
	struct Record *rec = ql->ql_Entries[col]->qe_Record;

	GuiVirtualFieldListDoubleClick(dio, ql->ql_LineNum, col, colfield, rec);
    }
}

static void HOOK_PROTO(sfwListMoveDrop, Object *obj, struct DisplayObject **x)
{
    long row = nget(obj, MUIA_SFWList_MoveDrop);
    struct DisplayObject *dio = *x;

    GuiVirtualFieldListSortDrop(dio, row);
}

static void HOOK_PROTO(sfwListAppMessage, Object *obj, struct AppMessage **x)
{
    struct AppMessage *am = *x;

    if (am != NULL && am->am_NumArgs > 0) {
        struct DisplayObject *dio = pget(obj, MUIA_UserData);
        struct Project *pr = DisplayObjectProject(dio);
        struct MemPool *memPool = ProjectMemPool(pr);
        struct WBArg *arg = am->am_ArgList;
        long num = am->am_NumArgs, n = 0, i;
        char *filename = NULL;
        char **labels = NULL;
        long error = OK;
       
        labels = AllocLabels(memPool, num);
        filename = PoolAllocMem(memPool, MAX_PATHNAME);
        if (labels == NULL || filename == NULL) {
            error = ERR_MEM;
        } else {
            for (i = 0; i < num; i++) {
                if (arg[i].wa_Name[0] == 0) {
                    continue;
                }
                if (!NameFromLock(arg[i].wa_Lock,
                            (STRPTR)filename, MAX_PATHNAME)) {
                    continue;
                }
                if (!AddPart((STRPTR)filename,
                            (CONST_STRPTR)arg[i].wa_Name, MAX_PATHNAME)) {
                    continue;
                }
                labels[n] = AllocString(memPool, filename);
                if (labels[n] == NULL) {
                    error = ERR_MEM;
                    break;
                }
                n++;
            }
        }
        if (error) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        } else {
            GuiVirtualFieldListDropUrls(dio, labels);
        }
        PoolFreeMem(memPool, filename, MAX_PATHNAME);
        FreeLabelsNum(memPool, labels, num);
    }
}

static void HOOK_PROTO(sfwListActiveChanged, Object *obj,
        struct DisplayObject **x)
{
    GuiVirtualFieldListActiveChanged(*x);
}

static long format_string[] =
{
    MUIV_String_Format_Center,
    MUIV_String_Format_Left,
    MUIV_String_Format_Right,
};

static char *format_text[] =
{
    "\33c\33-", "\33l\33-", "\33r\33-",
};


bool BuildGuiField(struct DisplayObject *dio, bool hideLabel,
        GuiObject parent, bool horiz, GuiObject *colGroup)
{
    struct Field *field = DisplayFieldField(dio);
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    const char *title = DisplayObjectTitle(dio); 
    short weight = DisplayObjectWeight(dio);
    static const char *helpnode = "Changing records";
    short titlepos = DisplayFieldTitlePosition(dio);
    short alignment = DisplayFieldAlignment(dio);
    short type = FieldType(field);
    bool chain = DisplayFieldTabChain(dio);
    bool extrachain = GetPrefsValue(pr, PREFS_EXTRABUTTONSINTABCHAIN);
    bool postAddTitle = FALSE;
    struct FontStyleSpec fss;
    struct TextFont *tf = FontTextFont(DisplayObjectMUIFontname(dio), &fss);
    const char *bh = DisplayFieldBubbleHelp(dio);
    const char *bg = DisplayObjectBackground(dio);
    GuiObject go = NULL, root = NULL, guiobject = NULL, activateobject = NULL,
              fileobject = NULL, fontobject = NULL, lvobject = NULL,
              viewobject = NULL, externalobject = NULL,
              autoshowobject = NULL, filterobject = NULL, popobject = NULL;
    struct GuiFieldFeatures f;

    if (parent == NULL || colGroup == NULL ||
	    !BuildGuiFieldGetFeatures(dio, &f)) {
        return FALSE;                                   /* be safe */
    }
    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    if (hideLabel) {
        f.label = FALSE;
    }

    if (f.label && !horiz && (titlepos == DAD_TITLEPOSITION_LEFT ||
        titlepos == DAD_TITLEPOSITION_RIGHT))
    {
    	if (*colGroup == NULL) {
	    *colGroup = GroupObject,
            	MUIA_Group_Columns, 2,
                DisplayGroupSpacing(DisplayObjectParent(dio))?
                    TAG_IGNORE : MUIA_Group_VertSpacing, 0,
                End;
	    if (*colGroup == NULL) {
		return FALSE;
            }
	    MUIAddMember(parent, *colGroup);
	}
	parent = *colGroup;			/* add children to this group */
	if (titlepos == DAD_TITLEPOSITION_RIGHT) {
	    postAddTitle = TRUE;
	} else if (!BuildGuiFieldLabel(dio, parent, FALSE)) {
    	    return FALSE;
        }
    } else {
	BuildGuiFinishColGroup(colGroup);
	if (f.label) {
	    switch (titlepos) {
		case DAD_TITLEPOSITION_LEFT:
		case DAD_TITLEPOSITION_RIGHT:
	    	    if (titlepos == DAD_TITLEPOSITION_RIGHT) {
     			postAddTitle = TRUE;
		    } else if (!BuildGuiFieldLabel(dio, parent, horiz)) {
			return FALSE;
		    }
		    break;

		case DAD_TITLEPOSITION_TOP:
		case DAD_TITLEPOSITION_BOTTOM:
		    if (horiz) {
			GuiObject group = root = VGroup,
                            MUIA_HorizWeight, weight,
                            End;
                        if (group == NULL) {
                            return FALSE;
                        }
                        MUIAddMember(parent, group);
			parent = group;
			horiz = FALSE;
		    }
	    	    if (titlepos == DAD_TITLEPOSITION_BOTTOM) {
     			postAddTitle = TRUE;
		    } else if (!BuildGuiFieldLabel(dio, parent, horiz)) {
			return FALSE;
		    }
		    break;
            }
	}
    }

    if (f.maxchars >= 0 || f.text) {
        if (f.show) {
            go = guiobject = activateobject = StringTextObject,
		MUIA_Frame, MUIV_Frame_Button,
                tf? MUIA_Font : TAG_IGNORE, tf,
                MUIA_Background, (bg && bg[0])? bg : (char *)MUII_ButtonBack,
	    	MUIA_Text_PreParse,PreParseCombine(&fss,format_text[alignment]),
                MUIA_Text_SetMin, FALSE,
                MUIA_InputMode, MUIV_InputMode_RelVerify,
	        MUIA_ControlChar, f.shortcut,
                MUIA_CycleChain, chain,
		End;
            if (go) {
                MUIAddCallback(go, MUIA_Pressed, FALSE, show, NULL);
            }
        } else if (f.readonly) {
    	    go = guiobject = StringTextObject,
		MUIA_Frame, MUIV_Frame_Text,
                tf? MUIA_Font : TAG_IGNORE, tf,
                MUIA_Background, (bg && bg[0])? bg : (char *)MUII_TextBack,
	    	MUIA_Text_PreParse,PreParseCombine(&fss,format_text[alignment]),
                MUIA_Text_SetMin, FALSE,
		End;
        } else {
   	    go = guiobject = activateobject = StringObject,
    	    	MUIA_Frame, MUIV_Frame_String,
                tf? MUIA_Font : TAG_IGNORE, tf,
#if 0
                MUIA_Background, (bg && bg[0])? bg : (char *)MUII_TextBack,
#else
		(bg && bg[0])? MUIA_Background : TAG_IGNORE, bg,
#endif
	    	f.shortcut? MUIA_ControlChar : TAG_IGNORE, f.shortcut,
	    	MUIA_String_MaxLen, f.maxchars + 1,
	    	MUIA_String_Format, format_string[alignment],
                MUIA_String_EditHook, &editHook,
                MUIA_String_AdvanceOnCR, TRUE,
                MUIA_HelpNode, helpnode,
                MUIA_CycleChain, chain,
                MUIA_UserData, dio,
	    	End;

            if (go) {
                MUIAddCallback(go, MUIA_String_Acknowledge, MUIV_EveryTime, 
                        stringacknowledge, pr);
            }
        }
        if (go) {
            MUIAddCallback(go, MUIA_String_Contents, MUIV_EveryTime, 
		    stringcontentschange, f.readonly);
        }
    } else if (f.checkbox) {
	go = HGroup,
	    GroupSpacing(0),
	    Child, guiobject = activateobject = ImageObject,
		ImageButtonFrame,
		MUIA_Disabled	      ,	f.readonly,
		MUIA_InputMode        , MUIV_InputMode_Toggle,
		MUIA_Image_Spec       , MUII_CheckMark,
		MUIA_Selected         , FALSE,
                tf? MUIA_Font : TAG_IGNORE, tf,
                MUIA_Background, (bg && bg[0])? bg : (char *)MUII_ButtonBack,
		MUIA_ShowSelState     , FALSE,
		MUIA_ControlChar      , f.shortcut,
                MUIA_CycleChain,        chain && !f.readonly,
                MUIA_UserData,          dio,
		End,
	    Child, HSpace(0),
	    End;
        if (go) {
            MUIAddCallback(guiobject, MUIA_Selected, MUIV_EveryTime,
                checkboxchange, MUIV_TriggerValue);
        }
    } else if (f.radiobuttons) {
        guiobject = activateobject = RadioObject,
            tf? MUIA_Font : TAG_IGNORE, tf,
            f.shortcut? MUIA_ControlChar : TAG_IGNORE, f.shortcut,
            MUIA_Group_Horiz, alignment == DAD_ALIGNMENT_HORIZONTAL,
            MUIA_Radio_Entries, FieldChoiceLabels(field),
            MUIA_CycleChain,    chain,
            MUIA_UserData,      dio,
            End;
        go = HGroup,
            DisplayFieldFrame(dio)?
                MUIA_Frame : TAG_IGNORE, MUIV_Frame_Group,
            (title && title[0])? MUIA_FrameTitle : TAG_IGNORE, title,
            MUIA_Background, (bg && bg[0])? bg : (char *)MUII_GroupBack,
            Child, HSpace(0),
            Child, guiobject,
            Child, HSpace(0),
            End;
        if (go) {
            MUIAddCallback(guiobject, MUIA_Radio_Active, MUIV_EveryTime,
                    cyclechange, MUIV_TriggerValue);
            set(guiobject, MUIA_Group_HorizSpacing, 5);
        }
    } else if (f.choice) {
        go = guiobject = activateobject = CycleObject,
            tf? MUIA_Font : TAG_IGNORE, tf,
            MUIA_Background, (bg && bg[0])? bg : (char *)MUII_ButtonBack,
            f.shortcut? MUIA_ControlChar : TAG_IGNORE, f.shortcut,
            MUIA_Cycle_Entries, FieldChoiceLabels(field),
            MUIA_CycleChain,    chain,
            MUIA_UserData,      dio,
            End;
        if (go) {
            MUIAddCallback(guiobject, MUIA_Cycle_Active, MUIV_EveryTime,
                    cyclechange, MUIV_TriggerValue);
        }
    } else if (f.button) {
	go = guiobject = activateobject = TextObject,
	    ButtonFrame,
            MUIA_InputMode, MUIV_InputMode_RelVerify,
	    MUIA_Text_Contents, EmptyNullStr(title),
	    MUIA_Text_PreParse, PreParseCombine(&fss, "\33c"),
	    MUIA_Text_HiChar, f.shortcut,
	    MUIA_ControlChar, f.shortcut,
            tf? MUIA_Font : TAG_IGNORE, tf,
            MUIA_Background, (bg && bg[0])? bg : (char *)MUII_ButtonBack,
            MUIA_CycleChain, chain,
            MUIA_UserData, dio,
	    End;
        if (go) {
            MUIAddCallback(go, MUIA_Pressed, FALSE, buttonchange, NULL);
        }
    } else if (f.image) {
        char imagesize = DisplayFieldImageSize(dio);

        go = guiobject = activateobject = ImageObject,
	    ImageButtonFrame,
            MUIA_InputMode, MUIV_InputMode_RelVerify,
	    MUIA_ControlChar, f.shortcut,
            (bg && bg[0])? MUIA_Image_Spec : TAG_IGNORE, bg,
            MUIA_Image_FreeHoriz, (imagesize == DAD_IMAGESIZE_RESIZABLE),
            MUIA_Image_FreeVert, (imagesize == DAD_IMAGESIZE_RESIZABLE),
            MUIA_Image_FontMatch, (imagesize == DAD_IMAGESIZE_FIXED),
            (bg && bg[0] == '5')? MUIA_Font : TAG_IGNORE, Topaz8Font,
            MUIA_CycleChain, chain,
            MUIA_UserData, dio,
	    End;
        if (go) {
            MUIAddCallback(go, MUIA_Pressed, FALSE, buttonchange, NULL);
        }
    } else if (f.editor) {
	go = MUINewEditor(pr, f.shortcut, f.readonly, FALSE,
            (bg && bg[0])? bg : (char *)MUII_TextBack, tf, &guiobject);
        if (guiobject) {
	    set(guiobject, MUIA_CycleChain, chain);
            MUIAddCallback(guiobject, MUIA_ExtEditor_HasChanged, TRUE,
                editorhaschanged, NULL);
            activateobject = guiobject;
        }
    } else if (f.listview) {
	go = MListviewObject,
            MUIA_UserData, dio,
            MUIA_Listview_DragType, MUIV_Listview_DragType_Immediate,
	    MUIA_MListview_MList, guiobject = activateobject = SFWListObject,
                MUIA_CycleChain, chain,
	        MUIA_Frame, MUIV_Frame_InputList,
                tf? MUIA_Font : TAG_IGNORE, tf,
                (bg && bg[0])? MUIA_Background : TAG_IGNORE, bg,
                MUIA_SFWList_ShowTitles, DisplayFieldShowTitles(dio),
	        End,
	    End;
        if (go != NULL) {
            const char *urlDropTrigger = DisplayFieldURLDropTrigger(dio);

            MUIAddCallback(go, MUIA_Listview_DoubleClick, TRUE,
                SFWListDoubleClick, dio);
            MUIAddPushCallback(go, MUIA_SFWList_MoveDrop, MUIV_EveryTime,
                sfwListMoveDrop, dio);
            if (urlDropTrigger != NULL && urlDropTrigger[0] != 0) {
                MUIAddCallback(go, MUIA_AppMessage, MUIV_EveryTime,
                    sfwListAppMessage, MUIV_TriggerValue);
            }
            MUIAddCallback(guiobject, MUIA_List_Active, MUIV_EveryTime,
                sfwListActiveChanged, dio);
        }
    } else if (f.maxchars != -1) {
        return FALSE;
    }

    if (f.maxchars != -1 && go == NULL) {
	return FALSE;
    }

    if (f.refpopup && go) {
        bool inchain = nget(go, MUIA_CycleChain);

        if (go) {
            set(go, MUIA_UserData, dio);
        }
        
        go = popobject = RefFieldSelectorObject,
            MUIA_UserData, dio,
	    MUIA_Popstring_String, go,
	    MUIA_Popstring_Button, lvobject = PopButton(MUII_PopUp),
            End;
        if (go == NULL) {
            return FALSE;
        }

        SetAttrs(lvobject,
		MUIA_CycleChain, chain && (!inchain || extrachain),
		MUIA_UserData, dio,
		TAG_DONE);

        if (activateobject == NULL) {
            activateobject = lvobject;
        }
    }

#ifdef USE_CALENDAR
    if (type == TYPE_DATE && DisplayFieldCalendar(dio) && !f.readonly && go) {
        GuiObject calendar = NULL;
        bool inchain = nget(go, MUIA_CycleChain);

        set(go, MUIA_UserData, dio);

        calendar = CalendarObject,
            InputListFrame,
            MUIA_Calendar_ShowToday,  TRUE,
            MUIA_Calendar_TodayClick, MUIV_Calendar_TodayClick_Double,
            MUIA_Calendar_SetDaysAsDefault, TRUE,
            MUIA_UserData, dio,
            End;

        if (calendar != NULL) { /* Ignore if fails (class not present) */
            go = PopobjectObject,
                MUIA_Popstring_String, go,
                MUIA_Popstring_Button, lvobject = PopButton(MUII_PopUp),
                MUIA_Popobject_Object, calendar,
                MUIA_Popobject_StrObjHook, &calendarStrObjHook,
                MUIA_Popobject_ObjStrHook, &calendarObjStrHook,
                End;
            if (go == NULL) {
                return FALSE;
            }
            DoMethod(calendar, MUIM_Notify,
                    MUIA_Calendar_DoubleClick, MUIV_EveryTime,
                    go, 2, MUIM_Popstring_Close,TRUE);

            SetAttrs(lvobject,
                MUIA_CycleChain, chain && (!inchain || extrachain),
                MUIA_UserData, dio,
                TAG_DONE);

            if (activateobject == NULL) {
                activateobject = lvobject;
            }
        }
    }
#endif

    if (type == TYPE_STRING && DisplayFieldFile(dio) && go) {
        bool inchain = nget(go, MUIA_CycleChain);

	fileobject = GuiPopButtonSetCWD(MUII_PopFile, pr);
        go = PopaslObject,
	    MUIA_Popstring_String, go,
	    MUIA_Popstring_Button, fileobject,
            MUIA_Popasl_Type, ASL_FileRequest,
            MUIA_Popasl_StopHook, &fileStopHook,
            End;
        if (go == NULL) {
	    return FALSE;
        }

        set(fileobject, MUIA_CycleChain, chain && (!inchain || extrachain));

        MUIAddAppMessageNotification(go, go);

        if (activateobject == NULL) {
            activateobject = fileobject;
        }
    } else if (type == TYPE_STRING && DisplayFieldFont(dio) && go) {
        bool inchain = nget(go, MUIA_CycleChain);

        go = PopaslObject,
	    MUIA_Popstring_String, go,
	    MUIA_Popstring_Button, fontobject = PopButton(MUII_PopUp),
            MUIA_Popasl_Type, ASL_FontRequest,
            MUIA_Popasl_StopHook, &fontStopHook,
            ASLFO_MaxHeight, 128,
            ASLFO_Flags, FOF_DOFRONTCOLOR | FOF_DOBACKCOLOR | FOF_DOSTYLE | 
                FOF_DOMUISTYLES,
            End;

        if (go == NULL) {
	    return FALSE;
        }

        set(fontobject, MUIA_CycleChain, chain && (!inchain || extrachain));

        if (activateobject == NULL) {
            activateobject = fontobject;
        }
    } else if (type == TYPE_STRING && DisplayFieldListview(dio) && go) {
        bool inchain = nget(go, MUIA_CycleChain);
        GuiObject listview = NULL;

        set(go, MUIA_UserData, dio);

        go = PopobjectObject,
	    MUIA_Popstring_String, go,
	    MUIA_Popstring_Button, lvobject = PopButton(MUII_PopUp),
	    MUIA_Popobject_StrObjHook, &labelStrObjHook,
	    MUIA_Popobject_ObjStrHook, &labelObjStrHook,
    	    MUIA_Popobject_WindowHook, &labelWindowHook,
	    MUIA_Popobject_Object, listview = ListviewObject,
                MUIA_UserData,      dio,
	        MUIA_Listview_List, ListObject,
                    tf? MUIA_Font : TAG_IGNORE, tf,
                    InputListFrame,
		    MUIA_List_ConstructHook, MUIV_List_ConstructHook_String,
		    MUIA_List_DestructHook, MUIV_List_DestructHook_String,
                    End,
                End,
            End;
        if (go == NULL) {
            return FALSE;
        }

        SetAttrs(lvobject,
            MUIA_CycleChain, chain && (!inchain || extrachain),
            MUIA_UserData, dio,
            TAG_DONE);
        set(listview, MUIA_UserData, go);

        MUIAddCallback2(listview, MUIA_Listview_DoubleClick, TRUE, 
                listview, labeldoubleclick, dio);

        if (activateobject == NULL) {
            activateobject = lvobject;
        }
    }

    if (type == TYPE_STRING && DisplayFieldView(dio) && go) {
        go = HGroup,
            GroupSpacing(0),
            MUIA_Group_SameHeight, TRUE,
            Child, go,
            Child, viewobject = TextObject,
                ButtonFrame,
  	        MUIA_Weight, 0,
                MUIA_Background, MUII_ButtonBack,
                MUIA_Text_SetVMax, FALSE,
	        MUIA_Text_Contents, GCS(MSG_View),
	        MUIA_Text_PreParse, "\33c",
	        MUIA_InputMode, MUIV_InputMode_RelVerify,
                MUIA_CycleChain, chain && extrachain,
                MUIA_ShortHelp, GCS(MSG_ViewBH),
                MUIA_UserData, dio,
                End,
            End;
        if (go == NULL) {
            return FALSE;
        }

        if (viewobject) {
            MUIAddCallback(viewobject, MUIA_Pressed, FALSE, 
                    GuiObjectViewFileCallback, guiobject);
        }
    }
    if (f.datatypes) {
        GuiObject str_go = go;
        char size = DisplayFieldImageSize(dio);
        bool fixed = (size == DAD_IMAGESIZE_SCROLLABLE || 
                size == DAD_IMAGESIZE_FIXED);
        bool scaled = (size == DAD_IMAGESIZE_SCALED || 
		size == DAD_IMAGESIZE_ASPECTSCALED);

        if (str_go != NULL && DisplayFieldTitleAtStringField(dio) && 
                title && title[0]) {
            str_go = HGroup,
                Child, KeyLabel1(title, f.shortcut),
                Child, go,
                End;
            if (str_go == NULL) {
                return FALSE;
            }
        }

        go = externalobject = DataTypesObject,
            MUIA_Frame, MUIV_Frame_Text,
            MUIA_DataTypes_FixedSize, fixed,
            MUIA_DataTypes_Scale, scaled,
            MUIA_DataTypes_KeepAspect, (size == DAD_IMAGESIZE_ASPECTSCALED),
            MUIA_UserData, dio,
            End;

        if (go && size == DAD_IMAGESIZE_SCROLLABLE) {
	    set(go, MUIA_Frame, MUIV_Frame_None);
            go = ScrollgroupObject,
                MUIA_Scrollgroup_Contents, VirtgroupObject,
                    Child, VGroup,
			MUIA_Frame, MUIV_Frame_Text,
			Child, HCenter(VCenter(go)),
			End,
                    End,
                End;
        }

        if (str_go != NULL) {
            go = VGroup,
                GroupSpacing(0),
                MUIA_Group_SameWidth, TRUE,
                Child, go,
                Child, str_go,
                End;
        }
        if (go == NULL) {
            return(FALSE);
        }
    }

    if (type == TYPE_REFERENCE && DisplayFieldAutoShow(dio) && go) {
        go = HGroup,
            GroupSpacing(0),
            Child, go,
            Child, autoshowobject = TextObject,
                ButtonFrame,
  	        MUIA_Weight, 0,
                MUIA_Background, MUII_ButtonBack,
                MUIA_Text_SetVMax, FALSE,
	        MUIA_Text_Contents, GCS(MSG_ReferenceAutoShow),
	        MUIA_Text_PreParse, "\33c",
	        MUIA_InputMode, MUIV_InputMode_Toggle,
                MUIA_CycleChain, chain && extrachain,
                MUIA_ShortHelp, GCS(MSG_ReferenceAutoShowBH),
                MUIA_UserData, dio,
                End,
            End;
        if (go == NULL) {
            return FALSE;
        }

	MUIAddCallback(autoshowobject, MUIA_Selected, MUIV_EveryTime,
		       referenceautoshowonoff, field);
    }

    if (type == TYPE_REFERENCE && DisplayFieldFilter(dio) && go) {
        go = HGroup,
            GroupSpacing(0),
            Child, go,
            Child, filterobject = TextObject,
	        ButtonFrame,
  	        MUIA_Weight, 0,
                MUIA_Background, MUII_ButtonBack,
	        MUIA_Text_SetVMax, FALSE,
	        MUIA_Text_Contents, GCS(MSG_ReferenceFilter),
	        MUIA_Text_PreParse, "\33c",
	        MUIA_InputMode, MUIV_InputMode_Toggle,
                MUIA_CycleChain, chain && extrachain,
                MUIA_ShortHelp, GCS(MSG_ReferenceFilterBH),
                MUIA_UserData, dio,
                End,
            End;
        if (go == NULL) {
            return FALSE;
        }

	MUIAddCallback(filterobject, MUIA_Selected, MUIV_EveryTime,
		       referencefilteronoff, field);
    }

    if (go == NULL) {
        return FALSE;
    }

    set(go, horiz? MUIA_HorizWeight : MUIA_VertWeight, weight);

    MUIAddMember(parent, go);
    if (postAddTitle && !BuildGuiFieldLabel(dio, parent, horiz)) {
	return FALSE;
    }

    if (guiobject) {
        set(guiobject, MUIA_UserData, dio);
        if (bh != NULL && *bh != 0) {
            set(guiobject, MUIA_ShortHelp, bh);
        }
    }

    SetDisplayObjectGuiObject(dio, guiobject);
    SetDisplayObjectRootObject(dio, root? root : go);
    SetDisplayObjectActivateObject(dio, activateobject);
    SetDisplayFieldFileObject(dio, fileobject);
    SetDisplayFieldFontObject(dio, fontobject);
    SetDisplayFieldListviewObject(dio, lvobject);
    SetDisplayFieldViewObject(dio, viewobject);
    SetDisplayFieldExternalObject(dio, externalobject);
    SetDisplayFieldFilterObject(dio, filterobject);
    SetDisplayFieldAutoShowObject(dio, autoshowobject);
    SetDisplayFieldPopObject(dio, popobject);
    SetDisplayFieldDisableObject(dio, go);

    return TRUE;
}

void GuiFieldStoreEdits(void)
{
    if (changedstringobject) {
        GuiObject obj = changedstringobject;
        const char *str = pget(obj, MUIA_String_Contents);
        struct DisplayObject *dio = pget(obj, MUIA_UserData);

        changedstringobject = NULL;
        GuiStoreDataStr(dio, str);
    }
    if (changededitorobject) {
        GuiObject obj = changededitorobject;

        changededitorobject = NULL;
	if (nget(obj, MUIA_ExtEditor_HasChanged)) {
	    struct DisplayObject *dio = pget(obj, MUIA_UserData);
            struct Project *pr = DisplayObjectProject(dio);
            struct MemPool *memPool = ProjectMemPool(pr);
	    char *str = ExtEditorGetContents(obj, memPool);

	    if (str != NULL) {
		GuiStoreDataStr(dio, str);
		set(obj, MUIA_ExtEditor_HasChanged, FALSE);
                FreeString(memPool, str);
            }
        }
    }
}

void GuiSetString(GuiObject go, const char *str, bool readonly,
        GuiObject external)
{
    if (go != NULL) {
	const char *oldstr = pget(go, MUIA_String_Contents);

	if (str == NULL || oldstr == NULL || strcmp(oldstr, str) != 0) {
	    SetAttrs(go,
		MUIA_NoNotify, TRUE,
		MUIA_String_Contents, str,
		MUIA_String_BufferPos, 0,
		MUIA_String_DisplayPos, 0,
		TAG_DONE);
        }
    }
    if (external) {
        struct DisplayObject *dio = GuiObjectDisplayObject(external);
        struct Project *pr = DisplayObjectProject(dio);
        GuiObject win = GuiObjectWindow(external);
        GuiObject group = pget(win, MUIA_Window_RootObject);
        bool fixedsize = nget(external, MUIA_DataTypes_FixedSize);
	bool wasOpen = nget(win, MUIA_Window_Open);
        ulong rc = 0;

        if (fixedsize && group != NULL) {	
            rc = DoMethod(group, MUIM_Group_InitChange);
	}
        ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
        set(external, MUIA_DataTypes_Filename, str);
        if (rc != 0) {
            DoMethod(group, MUIM_Group_ExitChange);
	    if (wasOpen && nget(win, MUIA_Window_Open) == FALSE) {
		/* 
		** The image object got too big to fit on the screen.
		** Temporary change to non-fixed image size
		*/
                if (DoMethod(group, MUIM_Group_InitChange)) {
		    set(external, MUIA_DataTypes_FixedSize, FALSE);
		    DoMethod(group, MUIM_Group_ExitChange);
		}
		set(win, MUIA_Window_Open, TRUE);
		set(external, MUIA_DataTypes_FixedSize, TRUE);
	    }
	}
    }
}

void GuiSetBoolean(GuiObject go, bool value, bool readonly)
{
    if (value) {
	value = TRUE;		/* be sure to have boolean value */
    }

    if (go) {
	if (value != nget(go, MUIA_Selected)) {
	    SetAttrs(go,
		MUIA_NoNotify, TRUE,
		MUIA_Selected, value,
		TAG_DONE);
        }
        if (readonly) {
            GuiSetEnabled(go, FALSE);
        }
    }
}

void GuiSetChoice(GuiObject go, unsigned short value, char **labels,
        bool readonly)
{
    if (go) {
        if (!readonly) {
	    if (value != (nget(go, MUIA_Cycle_Active) | 
                          nget(go, MUIA_Radio_Active))) {
		SetAttrs(go,
		    MUIA_NoNotify, TRUE,
		    MUIA_Cycle_Active, value,
		    MUIA_Radio_Active, value,
		    TAG_DONE);
            }
	} else {
            const char *label = pget(go, MUIA_Text_Contents);
	    const char *str;

	    if (value < NumLabels(labels)) {
		str = labels[value];
            } else {
		str = IntToStr(value);
            }

	    if (label == NULL || strcmp(label, str) != 0) {
		set(go, MUIA_Text_Contents, str);
            }
	}
    }
}

void GuiSetMemo(GuiObject go, const char *str, bool readonly)
{
    if (go) {
        SetAttrs(go,
            MUIA_ExtEditor_Contents, EmptyNullStr(str),
            MUIA_ExtEditor_Top, 0,
            TAG_DONE);
        set(go, MUIA_ExtEditor_HasChanged, FALSE);
    }
}

void GuiSetSFWList(GuiObject go, const struct Element *elem)
{
    if (go) {
        const char *error = NULL;
        long active = GuiListActive(go);
        struct DisplayObject *dio = GuiObjectDisplayObject(go);

        error = (const char *)DoMethod(go, MUIM_SFWList_SetElement, elem);
        if (error != NULL) {
            DoMethod(go, MUIM_SFWList_SetString, error);
        } else if (DisplayFieldControlTable(dio)) {
            struct Table *tab = GuiSFWListControlTable(go);
            struct Record *rec = TableGuiRecord(tab);

            if (tab != NULL && rec != NULL) {
                GuiSFWListSetActiveRecord(go, rec);
            }
        }
        if (GuiListActive(go) != active) {
            /* For some reasons MUI does not trigger a change so we do it */
            GuiVirtualFieldListActiveChanged(dio);
        }
    }
}

