/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** buildguitable_mui.c
**
** Creation date: 02.12.2005
** Last modified: 17.03.2021
*/

#include "system.h"
#include "buildguitable_mui.h"
#include "buildgui_mui.h"
#include "preparse_mui.h"
#include "gui.h"
#include "table.h"
#include "filter.h"
#include "util.h"
#include "amiga/font.h"
#include "extclasses/ActGroup_mcc.h"

static void HOOK_PROTO(filteronoff, GuiObject go, struct Table **x)
{
    struct Table *tab = *x;
    bool on = nget(go, MUIA_Selected);

    GuiStoreEdits();
    if (on && TableFilterProgram(tab) == NULL) {
	set(go, MUIA_Selected, FALSE);
	GuiChangeFilter(tab);
    } else {
	FilterOnOff(tab, on);
    }
}

static bool checkMouseButtonPressed(GuiObject go)
{
    GuiObject win = GuiObjectWindow(go);
    UWORD qualifier = nget(win, MUIA_Qualifier);

    if ((qualifier & IEQUALIFIER_LEFTBUTTON) == 0) {
	return FALSE;
    }
    return TRUE;
}

static void HOOK_PROTO(prevrec, GuiObject go, struct Table **x)
{
    struct Table *tab = *x;

    if (!checkMouseButtonPressed(go)) {
	return;
    }

    GuiStoreEdits();
    GuiPrevRecord(tab);
}

static void HOOK_PROTO(nextrec, GuiObject go, struct Table **x)
{
    struct Table *tab = *x;

    if (!checkMouseButtonPressed(go)) {
	return;
    }

    GuiStoreEdits();
    GuiNextRecord(tab);
}

static void HOOK_PROTO(setcursor, GuiObject go, ulong dummy)
{
/*    GuiSetCursor(); */
}

static GuiObject buildGuiPanel(struct DisplayObject *dio, 
        GuiObject parent, bool horiz)
{
    const char *title = DisplayObjectTitle(dio);
    struct Table *tab = DisplayObjectTable(dio);
    struct FontStyleSpec fss;
    struct TextFont *tf = FontTextFont(DisplayObjectMUIFontname(dio), &fss);
    const char *bg = DisplayTablePanelBackground(dio);
    short gadgets = !TableExactlyOneRecord(tab);
    short numall = DisplayTablePanelNumAll(dio) && gadgets,
	filter = DisplayTablePanelFilter(dio) && gadgets,
	arrows = DisplayTablePanelArrows(dio) && gadgets;
    GuiObject go = NULL, newparent = NULL, panel = NULL, numall_go = NULL, 
	buttons_go = NULL, filter_go = NULL, left_go = NULL, right_go = NULL;

    if (bg == NULL || bg[0] == 0) {
        bg = DisplayObjectBackground(dio);
    }

    numall_go = TextObject,
/*        tf? MUIA_Font : TAG_IGNORE, tf, */
	MUIA_Text_Contents, numall? "   #0/0" : "",
	MUIA_Text_PreParse, "\33r",
   	End;
    if (filter) {
        filter_go = TextObject,
            ButtonFrame,
/*            tf? MUIA_Font : TAG_IGNORE, tf, */
	    MUIA_Text_Contents, "F",
	    MUIA_Text_PreParse, "\33c",
	    MUIA_InputMode, MUIV_InputMode_Toggle,
            MUIA_Text_SetVMax, FALSE,
            End;
        if (filter_go) {
            MUIAddCallback(filter_go, MUIA_Selected, MUIV_EveryTime, 
                    filteronoff, tab);
        }
    }

    if (arrows) {
        left_go = TextObject,
            ButtonFrame,
/*            tf? MUIA_Font : TAG_IGNORE, tf, */
            MUIA_Text_SetVMax, FALSE,
	    MUIA_Text_Contents, "<",
	    MUIA_Text_PreParse, "\33c\33b",
	    MUIA_InputMode, MUIV_InputMode_RelVerify,
            End;
        if (left_go) {
            MUIAddPushCallback(left_go, MUIA_Timer, MUIV_EveryTime, prevrec, tab);
            MUIAddPushCallback(left_go, MUIA_Pressed, FALSE, setcursor, NULL);
        }

	right_go = TextObject,
            ButtonFrame,
/*            tf? MUIA_Font : TAG_IGNORE, tf, */
            MUIA_Text_SetVMax, FALSE,
	    MUIA_Text_Contents, ">",
	    MUIA_Text_PreParse, "\33c\33b",
	    MUIA_InputMode, MUIV_InputMode_RelVerify,
            End;
        if (right_go) {
            MUIAddPushCallback(right_go, MUIA_Timer, MUIV_EveryTime, nextrec, tab);
            MUIAddPushCallback(right_go, MUIA_Pressed, FALSE, setcursor, NULL);
        }
    }

    if (filter || arrows) {
	buttons_go = HGroup,
            MUIA_Weight, 0,
            MUIA_Group_SameSize, TRUE,
	    MUIA_Group_HorizSpacing, 0,
	    filter? Child : TAG_IGNORE, filter_go,
	    arrows? Child : TAG_IGNORE, left_go,
	    arrows? Child : TAG_IGNORE, right_go,
	    End;
    }

    panel = HGroup,
        MUIA_Group_SameHeight, TRUE,
        MUIA_Group_HorizSpacing, 0,
        MUIA_Background, (bg && bg[0])? bg : (char *)MUII_GroupBack,
      	Child, HGroup,
            ButtonFrame,
            Child, TextObject,
                tf? MUIA_Font : TAG_IGNORE, tf,
         	MUIA_Weight, 0,
            	MUIA_Text_Contents, EmptyNullStr(title),
                MUIA_Text_PreParse, PreParseCombine(&fss, ""),
            	MUIA_Text_SetMin, TRUE,
            	End,
            Child, numall_go,
            End,
        (filter || arrows)? Child : TAG_IGNORE, buttons_go,
        End;

    go = ActGroupObject,
    	MUIA_Group_Spacing, 0,
        MUIA_CycleChain, TRUE,
        MUIA_HelpNode, "Browsing records",
        MUIA_UserData, dio,
	Child, panel,
	Child, newparent = VGroup,
	    MUIA_Frame, MUIV_Frame_Group,
            MUIA_Background, MUII_GroupBack,
	    End,
	End;

    if (go == NULL) {
	return FALSE;
    }
    MUIAddMember(parent, go);

    SetDisplayObjectGuiObject(dio, go);
    SetDisplayObjectRootObject(dio, go);
    SetDisplayObjectActivateObject(dio, go);
    SetDisplayTablePanelNumAllObject(dio, numall_go);
    SetDisplayTablePanelFilterObject(dio, filter_go);

    return newparent;
}

bool BuildGuiTable(struct DisplayObject *dio, GuiObject parent, bool horiz)
{
    struct List *childs = DisplayObjectObjects(dio);
    const char *bg = DisplayObjectBackground(dio);
    GuiObject group, colGroup = NULL;

    if (parent == NULL) {
        return FALSE;
    }

    group = VGroup,
        MUIA_Weight, DisplayObjectWeight(dio),
        End;
    if (group == NULL) {
        return FALSE;
    }
    MUIAddMember(parent, group);

    SetDisplayObjectGuiObject(dio, group);
    SetDisplayObjectRootObject(dio, group);

    if (DisplayTableHasPanel(dio)) {
        group = buildGuiPanel(dio, group, FALSE);
        if (group == NULL) {
            return FALSE;
        }
    }
    set(group, MUIA_Background, (bg && bg[0])? bg : (char *)MUII_GroupBack);

    return BuildGuiChilds(childs, FALSE, group, FALSE, &colGroup);
}

void GuiSetNumAll(GuiObject go, ulong num, ulong all)
{
    if (go) {
        char buffer[31];

        sprintf(buffer, "#%ld/%ld", num, all);
        set(go, MUIA_Text_Contents, buffer);
    }
}

