/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** buildguiwindow_mui.c
**
** Creation date: 02.12.2005
** Last modified: 03.04.2022
*/

#include "system.h"
#include "buildguiwindow_mui.h"
#include "buildgui_mui.h"
#include "preparse_mui.h"
#include "extclasses/WindowButton_mcc.h"
#include "extclasses/DataTypes_mcc.h"
#include "buildgui.h"
#include "gui.h"
#include "menuproject.h"
#include "menutable.h"
#include "structureeditor.h"
#include "table.h"
#include "requesters.h"
#include "preferences.h"
#include "dosutil.h"
#include "util.h"
#include "appstrings_cat.h"
#include "amiga/font.h"

extern struct ExecBase *SysBase;

static bool keyFunc(GuiObject win, char *buf, short len, ushort qualifier,
        void *unused)
{
    static const ushort shift = IEQUALIFIER_LSHIFT | IEQUALIFIER_RSHIFT;
    static const ushort alt = IEQUALIFIER_LALT | IEQUALIFIER_RALT;
    static const ushort ctrl = IEQUALIFIER_CONTROL;
    static const ushort command = IEQUALIFIER_LCOMMAND | IEQUALIFIER_RCOMMAND;
    struct Project *pr = GuiWindowProject(win);
    struct DisplayObject *dio;
    char c;

    if (qualifier & command) {
        return FALSE;
    }
    if (len <= 0) {
        return FALSE;
    }

#if 0
    for (c=0; c < len; c++) {
        printf(" %02x", buf[c]);
    }
    printf("\n");
#endif
    c = buf[0];
    if (c == '\x9b' && buf[1] == '\x30' && buf[2] == '\x7e') {
        GuiShowHelp(NULL);
        return TRUE;
    }
    
    dio = DisplayActiveObject(win);
    if (dio == NULL) {
        return FALSE;
    }

    if (c == '\x9b') {
	void (*func)(struct Project *, ulong, Object *) = NULL;

	switch(buf[1]) {
	    case '\x41':
	    case '\x54':
		if ((qualifier & shift) && (qualifier & ctrl)) {
		    func = MenuTableFirstRecord;
                } else if ((qualifier & shift) && (qualifier & alt)) {
		    func = MenuTableJumpBackwardRecord;
                } else if (qualifier & alt) {
		    func = MenuTablePrevRecord;
                }
		break;

	    case '\x42':
	    case '\x53':
		if ((qualifier & shift) && (qualifier & ctrl)) {
		    func = MenuTableLastRecord;
                } else if ((qualifier & shift) && (qualifier & alt)) {
		    func = MenuTableJumpForwardRecord;
                } else if (qualifier & alt) {
		    func = MenuTableNextRecord;
                }
		break;

	    default: 
		break;
	}

	if(func != NULL) {
	    GuiStoreEdits();
	    func(pr, 0, win);
	    return TRUE;
	}
    } else if ((qualifier & ctrl) && !(qualifier & IEQUALIFIER_REPEAT) &&
            (c == '\11' || c == 'i' || c == 'I')) {      /* ctrl-i */
	struct Table *tab = DisplayObjectTable(dio);
	struct DisplayObject *tabdio = TableDisplayObject(tab);
	GuiObject filter = DisplayTablePanelFilterObject(tabdio);
        if (DisplayObjectVisible(tabdio) && DisplayTableHasPanel(tabdio) &&
                DisplayTablePanelFilter(tabdio) && filter != NULL) {
            GuiStoreEdits();
            set(filter, MUIA_Selected, !nget(filter, MUIA_Selected));
            return TRUE;
	}
    }

    return FALSE;
}

#if 0
static bool mouseFunc(GuiObject win, short x, short y, bool down)
{
    struct Project *pr = GuiWindowProject(win);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    if(down && tabs != NULL)
        FORNODE(tab, tabs)
        {
            struct DisplayObject *dio = TableDisplayObject(tab);
            GuiObject go = DisplayObjectGuiObject(dio);
            struct DisplayObject *windio = DisplayObjectWindow(dio);
            GuiObject tabwin = DisplayObjectGuiObject(windio);

            if (tabwin == win && GuiDisplayObjectVisible(dio) && go != NULL) {
                short left = nget(go, MUIA_LeftEdge);
                short right = left + nget(go, MUIA_Width) - 1;
                short top = nget(go, MUIA_TopEdge);
                short bottom = top + nget(go, MUIA_Height) - 1;

                if(x >= left && x <= right && y >= top && y <= bottom)
                {
                    struct DisplayObject *panel = DisplayTablePanel(tab);
                    GuiObject panelgo = DisplayObjectActivateObject(panel);

                    if(panelgo && DisplayObjectVisible(panel))
                        GuiSetActiveGuiObject(win, panelgo);

                    return(FALSE);
                }
            }
        }

    return(FALSE);
}
#endif

static void HOOK_PROTO(openwindow, GuiObject wi, struct DisplayObject **x)
{
    if (GuiWindowOpen(wi)) {
        GuiWindowToFront(wi);
        GuiActivateWindow(wi);
    } else if(!GuiOpenWindow(wi)) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_OpenWindow));
    }
}

static void HOOK_PROTO(windowactiveobjectchange, GuiObject go, GuiObject *x)
{
    GuiObject active = *x;

    if(active != NULL)
    {
        GuiStoreEdits();
    }
}

static void HOOK_PROTO(windowopen, GuiObject go, struct DisplayObject **x)
{
    struct DisplayObject *dio = *x;
    struct Project *pr = DisplayObjectProject(dio);

    GuiStoreEdits();
    GuiUpdateWindow(go, pr);
    SetDisplayWindowOpen(dio, TRUE);
}

static void HOOK_PROTO(windowclose, GuiObject go, struct DisplayObject **x)
{
    struct DisplayObject *dio = *x;
    struct Project *pr = DisplayObjectProject(dio);

    GuiStoreEdits();
    GuiClearWindow(go, pr);
    SetDisplayWindowOpen(dio, FALSE);
}

static void HOOK_PROTO(closerootwindow, GuiObject go, struct Project **x)
{
    GuiStoreEdits();
    MenuProjectClose(*x);
}

static GuiObject emptydisplaychild(bool addHelp)
{
    GuiObject child = NULL;
    short rc = FileStat(MAIN_WINDOW_LOGO_IMAGE);

    if (rc > 0 && ((struct Library *)SysBase)->lib_Version >= 39) {
        child = DataTypesObject,
            MUIA_DataTypes_Filename, MAIN_WINDOW_LOGO_IMAGE,
            MUIA_DataTypes_FixedSize, FALSE,
            MUIA_DataTypes_Scale, FALSE,
            MUIA_Background, "2:" MAIN_WINDOW_LOGO_BG_STR,
            MUIA_InnerLeft, 0,
            MUIA_InnerRight, 0,
            MUIA_InnerTop, 0,
            MUIA_InnerBottom, 0,
	    MUIA_MinWidth, MAIN_WINDOW_LOGO_MIN_WIDTH,
	    MUIA_MinHeight, MAIN_WINDOW_LOGO_MIN_HEIGHT,
            End;
    } else {
        child = VGroup,
	    GroupSpacing(10),
	    Child, HVSpace,
            Child, HGroup,
	        GroupSpacing(10),
	        Child, HVSpace,
	        Child, Label(GCS(MSG_EmptyDisplay)),
	        Child, HVSpace,
  	        End,
	    Child, HVSpace,
            End;
    }
    if (addHelp) {
	child = VGroup,
	    GroupSpacing(0),
	    MUIA_ShortHelp, GCS(MSG_EmptyDisplayBH),
            MUIA_InnerLeft, 0,
            MUIA_InnerRight, 0,
            MUIA_InnerTop, 0,
            MUIA_InnerBottom, 0,
	    Child, child,
	    End;
    }
    return child;
}

bool BuildGuiWindowButton(struct DisplayObject *dio, 
        GuiObject parent, bool horiz, GuiObject *colGroup)
{
    struct Project *pr = DisplayObjectProject(dio);
    const char *title = DisplayObjectTitle(dio);
    struct FontStyleSpec fss;
    struct TextFont *tf = FontTextFont(DisplayObjectMUIFontname(dio), &fss);
    const char *bg = DisplayObjectBackground(dio);
    char shortcut = DisplayObjectShortcut(dio);
    GuiObject button;

    if (!DisplayWindowButton(dio)) {
        return TRUE;
    }

    if (parent == 0) {
        return FALSE;
    }

    BuildGuiFinishColGroup(colGroup);

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    button = WindowButtonObject,
        ButtonFrame,
        MUIA_Text_Contents, EmptyNullStr(title),
        MUIA_Text_PreParse, PreParseCombine(&fss, "\33c"),
        MUIA_Text_HiChar, shortcut,
        MUIA_ControlChar, shortcut,
        MUIA_Weight, DisplayObjectWeight(dio),
        tf? MUIA_Font : TAG_IGNORE, tf,
        MUIA_Background, (bg && bg[0])? bg : (char *)MUII_ButtonBack,
        MUIA_CycleChain, GetPrefsValue(pr, PREFS_EXTRABUTTONSINTABCHAIN),
        MUIA_UserData, dio,
        End;
    if (button == NULL) {
        return FALSE;
    }

    MUIAddMember(parent, button);
    SetDisplayWindowButtonObject(dio, button);
    SetDisplayObjectRootObject(dio, button);

    return TRUE;
}

bool BuildGuiWindow(struct DisplayObject *dio)
{
    struct Project *pr = DisplayObjectProject(dio);
    bool ismain = (dio == ProjectRootDisplayObject(pr));
    bool empty = !DisplayObjectVisibleGui(dio);
    const char *title = DisplayObjectTitle(dio);
    const struct WindowDimensions *wd = NULL;
    long defaultWidth = MUIV_Window_Width_Default;
    long defaultHeight = MUIV_Window_Height_Default;
    struct List *childs = DisplayObjectObjects(dio);
    GuiObject app = GuiApplication;
    GuiObject win, go, colGroup = NULL;

    if (app == NULL) {
	return FALSE;
    }

    /* Before creating the GUI, get the base font. */
    if (ismain && !empty) {
        win = WindowObject,
            MUIA_Window_Activate, FALSE,
            MUIA_Window_Backdrop, TRUE,
            MUIA_Window_Borderless, TRUE,
            MUIA_Window_CloseGadget, FALSE,
            MUIA_Window_DepthGadget, FALSE,
            MUIA_Window_SizeGadget, FALSE,
            MUIA_Window_DragBar, FALSE,
            MUIA_Window_Width, 0,
            MUIA_Window_Height, 0,
            MUIA_Window_NoMenus, TRUE,
            MUIA_Window_RootObject, DataTypesObject,
                MUIA_DataTypes_SaveBaseFont, SetBaseFont,
                End,
            End;
        MUIAddMember(app, win);
        set(win, MUIA_Window_Open, TRUE);
        set(win, MUIA_Window_Open, FALSE);
        MUIRemMember(app, win);
        MUI_AddDisposeObject(app, win);
    }

    if (empty) {
        go = emptydisplaychild(ismain);
	defaultWidth = MAIN_WINDOW_LOGO_DEFAULT_WIDTH;
	defaultHeight = MAIN_WINDOW_LOGO_DEFAULT_HEIGHT;
    } else {
        go = VGroup, End;
    	wd = DisplayWindowDimensions(dio);
    }

    win = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_KeyFunc, keyFunc,
/*        MUIA_MouseFunc, mouseFunc, */
        title? MUIA_Window_Title : TAG_IGNORE, title,
        MUIA_Window_ScreenTitle, GuiScreenTitle,
        MUIA_Window_AppWindow, TRUE,
	MUIA_UserData, dio,
	MUIA_Window_ID, DisplayWindowId(dio),
	MUIA_Window_LeftEdge, (wd && wd->wd_LeftEdge >= 0)?  
            wd->wd_LeftEdge : MUIV_Window_LeftEdge_Moused,
	MUIA_Window_TopEdge, (wd && wd->wd_TopEdge >= 0)? 
            wd->wd_TopEdge  : MUIV_Window_TopEdge_Moused,
        MUIA_Window_Width, (wd && wd->wd_Width > 0)? 
            wd->wd_Width : defaultWidth,
        MUIA_Window_Height, (wd && wd->wd_Height > 0)? 
            wd->wd_Height : defaultHeight,
        WindowContents, go,
	End;
    if (win == NULL) {
	return FALSE;
    }
    MUIAddMember(app, win);
    SetDisplayObjectGuiObject(dio, win);
    MUIAddCallback(win, MUIA_Window_ActiveObject, MUIV_EveryTime,
        windowactiveobjectchange, MUIV_TriggerValue);

    if (ismain) {
        MUIAddCallback(win, MUIA_Window_CloseRequest, TRUE,closerootwindow, pr);
        SetRootWindowTitle(pr, win);
    } else {
        GuiObject button = DisplayWindowButtonObject(dio);

        if (button) {
            MUIAddCallback2(button, MUIA_Pressed, FALSE, win, openwindow, dio);
        }
        DoMethod(win, MUIM_Notify, MUIA_Window_CloseRequest, TRUE,
            win, 3, MUIM_Set, MUIA_Window_Open, FALSE);
    }
    /* make these callbacks last in order to avoid a bug in MUI's KillNotify */
    MUIAddCallback(win, MUIA_Window_Open, FALSE, windowclose, dio);
    MUIAddCallback(win, MUIA_Window_Open, TRUE, windowopen, dio);

    return BuildGuiChilds(childs, FALSE, go, FALSE, &colGroup);
}

