/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** display_mui.c
**
** Creation date: 08.12.1994
** Last modified: 12.07.2020
*/

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "display.h"
#include "preferences.h"
#include "project.h"
#include "table.h"
#include "field.h"
#include "requesters.h"
#include "exec.h"
#include "memory.h"
#include "error.h"
#include "util.h"
#include "labels.h"
#include "appstrings_cat.h"
#include "structureeditor_mui.h"
#include "extclasses/ExtApplication_mcc.h"
#include "extclasses/PopEditLabels_mcc.h"
#include <libraries/asl.h>

#define IMAGESPEC_LEN (int)(sizeof(struct MUI_ImageSpec) - 1)
#define IMAGESPEC_START "0:128"

void HOOK_PROTO(ShortcutAcknowledge, Object *obj, struct DisplayObject **x)
{
    struct DisplayObject *dio = *x, *inuse;
    Object *win = GuiObjectWindow(obj);
    const char *text = pget(obj, MUIA_String_Contents), *name;
    char shortcut = (text != NULL)? text[0] : 0;

    inuse = DisplayObjectShortcutInUse(shortcut, DisplayObjectWindow(dio), dio);
    if(inuse != NULL)
    {
        name = DisplayObjectName(inuse);
        Ask2(win, NULL, AT_ERROR, GCS(MSG_StarOk),
		GCS(MSG_DisplayObject_ShortcutInUse), EmptyNullStr(name));
        if(win != NULL)
            set(win, MUIA_Window_ActiveObject, obj);
    }
}

struct EditTextData
{
    struct Project *d_Project;
    struct DisplayObject *d_Text;
    Object *d_StructureEditor;
    Object *d_Title, *d_Weight, *d_Fontname;
    Object *d_Background, *d_DefaultBackground;
    Object *d_Ok, *d_Cancel;
    struct MUI_ImageSpec d_BackgroundSpec;
};

static void disposeEditText(Object *app, Object *obj)
{
    struct EditTextData *d = pget(obj, MUIA_UserData);

    if(obj)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(struct EditTextData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void HOOK_PROTO(closeEditText, Object *obj, ulong *ok)
{
    struct EditTextData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);
    struct TextData td;
    long error;

    if(*ok)
    {
        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        td.td_Title = pget(d->d_Title, MUIA_String_Contents);
        td.td_Weight = nget(d->d_Weight, MUIA_Numeric_Value);
        td.td_Fontname = pget(d->d_Fontname, MUIA_String_Contents);
        td.td_Background = nget(d->d_DefaultBackground, MUIA_Selected)?
            NULL : pget(d->d_Background, MUIA_Imagedisplay_Spec);

        GuiSetSleep(TRUE);
        DisplayChangeText(d->d_Text, &td);
        ProjectChange(pr, TRUE);
        GuiSetSleep(FALSE);

        TextFinish(d->d_StructureEditor, d->d_Text);
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposeEditText(app, obj);
}

void GuiText(Object *se, struct DisplayObject *text)
{
    struct Project *pr = DisplayObjectProject(text);
    struct MemPool *memPool = ProjectMemPool(pr);
    Object *app = GuiApplication;
    const char *bg = DisplayObjectBackground(text);
    struct EditTextData *d;
    Object *obj;
    Object *root, *attributes, *back;

    if(app == NULL || text == NULL)
	return;

    d = PoolAllocMemZero(memPool, sizeof(struct EditTextData));
    if(d == NULL)
        return;
    d->d_Project = pr;
    d->d_Text = text;
    d->d_StructureEditor = se;

    d->d_Title = StringObject,
        MUIA_CycleChain, TRUE,
	MUIA_ShortHelp, GCS(MSG_DisplayText_TitleBH),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Title)),
	MUIA_String_MaxLen, TITLE_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, DisplayObjectTitle(text),
	End;

    d->d_Weight = SliderObject,
        MUIA_CycleChain, TRUE,
	MUIA_ShortHelp, GCS(MSG_WeightBH),
	MUIA_Frame, MUIV_Frame_Slider,
	MUIA_ControlChar, Shortcut(GCS(MSG_Weight)),
	MUIA_Numeric_Min, 0,
	MUIA_Numeric_Max, 1000,
        MUIA_Numeric_Value, DisplayObjectWeight(text),
	End;

    d->d_Fontname = StringObject,
        MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Font)),
        MUIA_String_MaxLen, FONTNAME_LEN,
        MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, DisplayObjectMUIFontname(text),
        MUIA_ShortHelp, GCS(MSG_FontBH),
        MUIA_CycleChain, TRUE,
        End;

    CodesetStrNBytesCopy(d->d_BackgroundSpec.buf, 
	    (bg && bg[0])? bg : IMAGESPEC_START, IMAGESPEC_LEN);
    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    d->d_Background = PopimageObject,
        MUIA_Weight, 200,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_BackgroundBH),
        MUIA_ControlChar, Shortcut(GCS(MSG_Background)),
        MUIA_Imagedisplay_Spec, &d->d_BackgroundSpec,
        MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_Background,
        End;
    MUIAddCallback(d->d_Background, MUIA_Pressed, FALSE, PopButtonSetCWD, pr);

    d->d_DefaultBackground = MUI_MakeObject(MUIO_Checkmark, 
	    (IPTR)GCS(MSG_Background_Default));

    back = HGroup,
        Child, d->d_Background,
        Child, HVSpace,
        Child, VGroup,
            Child, VSpace(0),
            Child, HGroup,
                Child, Label(GCS(MSG_Background_Default)),
                Child, d->d_DefaultBackground,
                End,
            Child, VSpace(0),
            End,
        End;

    attributes = GroupObject,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	MUIA_Group_Columns, 2,
	Child, Label2(GCS(MSG_Title)),
	Child, d->d_Title,
	Child, Label1(GCS(MSG_Weight)),
	Child, d->d_Weight,
        Child, Label2(GCS(MSG_Font)),
        Child, PopaslObject,
	    MUIA_Popstring_String, d->d_Fontname,
	    MUIA_Popstring_Button, PopButton(MUII_PopUp),
            MUIA_Popasl_Type, ASL_FontRequest,
            ASLFO_MaxHeight, 128,
            ASLFO_Flags, FOF_DOFRONTCOLOR | FOF_DOBACKCOLOR | FOF_DOSTYLE | 
                FOF_DOMUISTYLES,
            End,
        Child, FreeLabel(GCS(MSG_Background)),
        Child, back,
	End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    root = VGroup,
	Child, attributes,
	Child, HGroup,
            MUIA_Group_SameWidth, TRUE,
	    Child, d->d_Ok,
            Child, d->d_Cancel,
	    End,
	End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_Title, GCS(MSG_Text),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'T','X','T'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_HelpNode, "Text editor",
        MUIA_UserData, d,
	WindowContents, root,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct EditTextData));
	return;
    }

    DoMethod(d->d_DefaultBackground, MUIM_Notify, MUIA_Selected, MUIV_EveryTime,
        d->d_Background, 3, MUIM_Set, MUIA_Disabled, MUIV_TriggerValue);
    SetAttrs(d->d_DefaultBackground,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_Background_DefaultBH),
        MUIA_Selected, !bg || !bg[0],
        TAG_DONE);
    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    DoMethod(d->d_Title,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    DoMethod(d->d_Fontname,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE,
        obj, closeEditText, FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE,
        obj, closeEditText, FALSE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE,
        obj, closeEditText, TRUE);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeEditText(app, obj);
        return;
    }

    set(obj, MUIA_Window_ActiveObject, d->d_Title);
}

struct EditSpaceData
{
    struct Project *d_Project;
    struct DisplayObject *d_Space;
    Object *d_StructureEditor;
    Object *d_Bar, *d_Weight, *d_Background, *d_DefaultBackground;
    Object *d_Ok, *d_Cancel;
    struct MUI_ImageSpec d_BackgroundSpec;
};

static void disposeEditSpace(Object *app, Object *obj)
{
    struct EditSpaceData *d = pget(obj, MUIA_UserData);

    if(obj)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(struct EditSpaceData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void HOOK_PROTO(closeEditSpace, Object *obj, ulong *ok)
{
    struct EditSpaceData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);
    struct SpaceData sd;
    long error;

    if(*ok)
    {
        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        sd.sd_Bar = nget(d->d_Bar, MUIA_Selected);
        sd.sd_Weight = nget(d->d_Weight, MUIA_Numeric_Value);
        sd.sd_Background = nget(d->d_DefaultBackground, MUIA_Selected)?
                NULL : pget(d->d_Background, MUIA_Imagedisplay_Spec);

        GuiSetSleep(TRUE);
        DisplayChangeSpace(d->d_Space, &sd);
        ProjectChange(pr, TRUE);
        GuiSetSleep(FALSE);

        SpaceFinish(d->d_StructureEditor, d->d_Space);
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposeEditSpace(app, obj);
}

void GuiSpace(Object *se, struct DisplayObject *space)
{
    struct Project *pr = DisplayObjectProject(space);
    struct MemPool *memPool = ProjectMemPool(pr);
    Object *app = GuiApplication;
    const char *bg = DisplayObjectBackground(space);
    struct EditSpaceData *d;
    Object *root, *attributes, *back, *buttons;
    Object *obj;

    if(app == NULL || space == NULL)
	return;

    d = PoolAllocMemZero(memPool, sizeof(struct EditSpaceData));
    if(d == NULL)
        return;
    d->d_Project = pr;
    d->d_Space = space;
    d->d_StructureEditor = se;

    d->d_Bar = MUI_MakeObject(MUIO_Checkmark,
            (IPTR)GCS(MSG_DisplaySpace_Delimiter));

    d->d_Weight = SliderObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_WeightBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_Weight)),
	MUIA_Frame, MUIV_Frame_Slider,
	MUIA_Numeric_Min, 0,
	MUIA_Numeric_Max, 1000,
        MUIA_Numeric_Value, DisplayObjectWeight(space),
	End;

    CodesetStrNBytesCopy(d->d_BackgroundSpec.buf, 
	    (bg && bg[0])? bg : IMAGESPEC_START, IMAGESPEC_LEN);
    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    d->d_Background = PopimageObject,
        MUIA_Weight, 200,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_BackgroundBH),
        MUIA_ControlChar, Shortcut(GCS(MSG_Background)),
        MUIA_Imagedisplay_Spec, &d->d_BackgroundSpec,
        MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_Background,
        End;
    MUIAddCallback(d->d_Background, MUIA_Pressed, FALSE, PopButtonSetCWD, pr);

    d->d_DefaultBackground = MUI_MakeObject(MUIO_Checkmark, 
	    (IPTR)GCS(MSG_Background_Default));

    back = HGroup,
        Child, d->d_Background,
        Child, HVSpace,
        Child, VGroup,
            Child, VSpace(0),
            Child, HGroup,
                Child, Label1(GCS(MSG_Background_Default)),
                Child, d->d_DefaultBackground,
                End,
            Child, VSpace(0),
            End,
        End;

    attributes = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	MUIA_Group_Columns, 2,
	Child, Label1(GCS(MSG_DisplaySpace_Delimiter)),
	Child, HGroup,
    	    Child, d->d_Bar,
	    Child, HSpace(0),
            End,
	Child, Label1(GCS(MSG_Weight)),
	Child, d->d_Weight,
        Child, FreeLabel(GCS(MSG_Background)),
        Child, back,
	End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, d->d_Ok,
	Child, d->d_Cancel,
	End;

    root = GroupObject,
	Child, attributes,
	Child, buttons,
	End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_Title, GCS(MSG_Space),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'S','P','C'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_HelpNode, "Space editor",
        MUIA_UserData, d,
	WindowContents, root,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct EditSpaceData));
	return;
    }

    SetAttrs(d->d_Bar,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplaySpace_DelimiterBH),
        MUIA_Selected, DisplaySpaceBar(space),
        TAG_DONE);
    DoMethod(d->d_DefaultBackground, MUIM_Notify, MUIA_Selected, MUIV_EveryTime,
        d->d_Background, 3, MUIM_Set, MUIA_Disabled, MUIV_TriggerValue);
    SetAttrs(d->d_DefaultBackground,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_Background_DefaultBH),
        MUIA_Selected, !bg || !bg[0],
        TAG_DONE);
    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE,
        obj, closeEditSpace, FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE,
        obj, closeEditSpace, FALSE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE,
        obj, closeEditSpace, TRUE);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeEditSpace(app, obj);
        return;
    }
}

struct EditImageData
{
    struct Project *d_Project;
    struct DisplayObject *d_Image;
    Object *d_StructureEditor;
    Object *d_Weight, *d_PopImage;
    Object *d_ImageSize;
    Object *d_Ok, *d_Cancel;
    struct MUI_ImageSpec d_ImageSpec;
};

static void disposeEditImage(Object *app, Object *obj)
{
    struct EditImageData *d = pget(obj, MUIA_UserData);

    if(obj)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(struct EditImageData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void HOOK_PROTO(closeEditImage, Object *obj, ulong *ok)
{
    struct EditImageData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);
    struct ImageData id;
    long error;

    if(*ok)
    {
        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        id.id_Weight = nget(d->d_Weight, MUIA_Numeric_Value);
        id.id_Image = pget(d->d_PopImage, MUIA_Imagedisplay_Spec);
        id.id_ImageSize = nget(d->d_ImageSize, MUIA_Radio_Active);

        GuiSetSleep(TRUE);
        DisplayChangeImage(d->d_Image, &id);
        ProjectChange(pr, TRUE);
        GuiSetSleep(FALSE);

        ImageFinish(d->d_StructureEditor, d->d_Image);
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposeEditImage(app, obj);
}

void GuiImage(Object *se, struct DisplayObject *image)
{
    struct Project *pr = DisplayObjectProject(image);
    struct MemPool *memPool = ProjectMemPool(pr);
    Object *app = GuiApplication;
    const char *bg = DisplayObjectBackground(image);
    struct EditImageData *d;
    Object *root, *attributes, *buttons;
    Object *obj;
    static const char *entries[3];

    if(app == NULL || image == NULL)
	return;

    d = PoolAllocMemZero(memPool, sizeof(struct EditImageData));
    if(d == NULL)
        return;
    d->d_Project = pr;
    d->d_Image = image;
    d->d_StructureEditor = se;

    d->d_Weight = SliderObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_WeightBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_Weight)),
	MUIA_Frame, MUIV_Frame_Slider,
	MUIA_Numeric_Min, 0,
	MUIA_Numeric_Max, 1000,
        MUIA_Numeric_Value, DisplayObjectWeight(image),
	End;

    CodesetStrNBytesCopy(d->d_ImageSpec.buf, 
	    (bg && bg[0])? bg : IMAGESPEC_START, IMAGESPEC_LEN);
    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    d->d_PopImage = PopimageObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_ImageBH),
        MUIA_ControlChar, Shortcut(GCS(MSG__Image)),
        MUIA_Imagedisplay_Spec, &d->d_ImageSpec,
        MUIA_Imageadjust_Type, ExecIsMorphOS()?
            MUIV_Imageadjust_Type_Background : MUIV_Imageadjust_Type_All,
        End;
    MUIAddCallback(d->d_PopImage, MUIA_Pressed, FALSE, PopButtonSetCWD, pr);

    entries[0] = GCS(MSG_Image_Size_Resizeable);
    entries[1] = GCS(MSG_Image_Size_Fixed);
    entries[2] = NULL;
    d->d_ImageSize = RadioObject,
        MUIA_Radio_Entries, entries,
        MUIA_Radio_Active, DisplayImageImageSize(image),
        MUIA_CycleChain, TRUE,
        End;

    attributes = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	MUIA_Group_Columns, 2,
	Child, Label1(GCS(MSG_Weight)),
	Child, d->d_Weight,
        Child, FreeLabel(GCS(MSG__Image)),
        Child, HGroup,
            Child, d->d_PopImage,
            Child, VGroup,
                GroupFrame,
                MUIA_FrameTitle, GCS(MSG_Size),
                MUIA_Background, MUII_GroupBack,
                Child, VCenter(d->d_ImageSize),
                End,
            End,
	End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, d->d_Ok,
	Child, d->d_Cancel,
	End;

    root = GroupObject,
	Child, attributes,
	Child, buttons,
	End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_Title, GCS(MSG_Image),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'I','M','G'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_HelpNode, "Image editor",
        MUIA_UserData, d,
	WindowContents, root,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct EditImageData));
	return;
    }

    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE,
        obj, closeEditImage, FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE,
        obj, closeEditImage, FALSE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE,
        obj, closeEditImage, TRUE);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeEditImage(app, obj);
        return;
    }
}

struct EditRegisterData
{
    struct Project *d_Project;
    struct DisplayObject *d_Register;
    Object *d_StructureEditor;
    Object *d_Weight;
    Object *d_Labels;
    Object *d_Ok, *d_Cancel;
};

static void disposeEditRegister(Object *app, Object *obj)
{
    struct EditRegisterData *d = pget(obj, MUIA_UserData);

    if(obj)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, 
	    sizeof(struct EditRegisterData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void HOOK_PROTO(closeEditRegister, Object *obj, ulong *ok)
{
    struct EditRegisterData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);
    struct RegisterData rd;
    long error;

    if(*ok)
    {
        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        rd.rd_Weight = nget(d->d_Weight, MUIA_Numeric_Value);
        rd.rd_Fontname = NULL;
        rd.rd_Background = NULL;
        rd.rd_Labels = pget(d->d_Labels, MUIA_EditLabels_Labels);

        GuiSetSleep(TRUE);
        error = DisplayChangeRegister(d->d_Register, &rd);
        if (error) {
            Ask2(obj, GCS(MSG_ErrorMessage), AT_ERROR,
		    GCS(MSG_StarOk), GES(error));
        } else {
            ProjectChange(pr, TRUE);
        }
        GuiSetSleep(FALSE);

        RegisterFinish(d->d_StructureEditor, d->d_Register);
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposeEditRegister(app, obj);
}

void GuiRegister(Object *se, struct DisplayObject *reg)
{
    struct Project *pr = DisplayObjectProject(reg);
    struct MemPool *memPool = ProjectMemPool(pr);
    Object *app = GuiApplication;
    struct EditRegisterData *d;
    Object *root, *frame, *basic, *buttons;
    Object *obj;

    if(app == NULL || reg == NULL)
	return;

    d = PoolAllocMemZero(memPool, sizeof(struct EditRegisterData));
    if(d == NULL)
        return;
    d->d_Project = pr;
    d->d_Register = reg;
    d->d_StructureEditor = se;

    d->d_Weight = SliderObject,
        MUIA_CycleChain, TRUE,
	MUIA_ShortHelp, GCS(MSG_WeightBH),
	MUIA_Frame, MUIV_Frame_Slider,
	MUIA_ControlChar, Shortcut(GCS(MSG_Weight)),
	MUIA_Numeric_Min, 0,
	MUIA_Numeric_Max, 1000,
        MUIA_Numeric_Value, DisplayObjectWeight(reg),
	End;

    basic = VGroup,
	MUIA_Weight, 50,
	MUIA_Group_Columns, 2,
	Child, Label1(GCS(MSG_Weight)),
	Child, d->d_Weight,
	End;

    d->d_Labels = EditLabelsObject,
	MUIA_FrameTitle, GCS(MSG_DisplayRegister_Labels),
	MUIA_ShortHelp, GCS(MSG_DisplayRegister_LabelsBH),
	MUIA_EditLabels_Labels, DisplayRegisterLabels(reg),
	End;

    frame = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	Child, basic,
	Child, d->d_Labels,
	End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, d->d_Ok,
	Child, d->d_Cancel,
	End;

    root = GroupObject,
	Child, frame,
	Child, buttons,
	End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_Title, GCS(MSG_Register),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'R','E','G'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_HelpNode, "Register group editor",
        MUIA_UserData, d,
	WindowContents, root,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct EditRegisterData));
	return;
    }

    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE,
        obj, closeEditRegister, FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE,
        obj, closeEditRegister, FALSE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE,
        obj, closeEditRegister, TRUE);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeEditRegister(app, obj);
        return;
    }
    set(obj, MUIA_Window_ActiveObject, d->d_Weight);
}

struct EditGroupData
{
    struct Project *d_Project;
    struct DisplayObject *d_Group;
    Object *d_StructureEditor;
    Object *d_Title, *d_Weight, *d_Border, *d_Horizontal, *d_Spacing;
    Object *d_Background, *d_DefaultBackground;
    Object *d_Ok, *d_Cancel;
    struct MUI_ImageSpec d_BackgroundSpec;
};

static void disposeEditGroup(Object *app, Object *obj)
{
    struct EditGroupData *d = pget(obj, MUIA_UserData);

    if(obj)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(struct EditGroupData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void HOOK_PROTO(closeEditGroup, Object *obj, ulong *ok)
{
    struct EditGroupData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);
    struct GroupData gd;
    long error;

    if(*ok)
    {
        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        gd.gd_Title = pget(d->d_Title, MUIA_String_Contents);
        gd.gd_Weight = nget(d->d_Weight, MUIA_Numeric_Value);
        gd.gd_Fontname = NULL;
        gd.gd_Background = nget(d->d_DefaultBackground, MUIA_Selected)?
            NULL : pget(d->d_Background, MUIA_Imagedisplay_Spec);
        gd.gd_Horiz = nget(d->d_Horizontal, MUIA_Selected);
        gd.gd_Border = nget(d->d_Border, MUIA_Selected);
        gd.gd_Spacing = nget(d->d_Spacing, MUIA_Selected);

        GuiSetSleep(TRUE);
        DisplayChangeGroup(d->d_Group, &gd);
        ProjectChange(pr, TRUE);
        GuiSetSleep(FALSE);

        GroupFinish(d->d_StructureEditor, d->d_Group);
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposeEditGroup(app, obj);
}

void GuiGroup(Object *se, struct DisplayObject *group)
{
    struct Project *pr = DisplayObjectProject(group);
    struct MemPool *memPool = ProjectMemPool(pr);
    Object *app = GuiApplication;
    const char *bg = DisplayObjectBackground(group);
    struct EditGroupData *d;
    Object *root, *frame, *attrs, *back, *basic, *buttons;
    Object *obj;

    if(app == NULL || group == NULL)
	return;

    d = PoolAllocMemZero(memPool, sizeof(struct EditGroupData));
    if(d == NULL)
        return;
    d->d_Project = pr;
    d->d_Group = group;
    d->d_StructureEditor = se;

    d->d_Title = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayGroup_TitleBH),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Title)),
	MUIA_String_MaxLen, TITLE_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, DisplayObjectTitle(group),
	End;

    d->d_Weight = SliderObject,
        MUIA_CycleChain, TRUE,
	MUIA_ShortHelp, GCS(MSG_WeightBH),
	MUIA_Frame, MUIV_Frame_Slider,
	MUIA_ControlChar, Shortcut(GCS(MSG_Weight)),
	MUIA_Numeric_Min, 0,
	MUIA_Numeric_Max, 1000,
        MUIA_Numeric_Value, DisplayObjectWeight(group),
	End;

    CodesetStrNBytesCopy(d->d_BackgroundSpec.buf, 
	    (bg && bg[0])? bg : IMAGESPEC_START, IMAGESPEC_LEN);
    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    d->d_Background = PopimageObject,
        MUIA_Weight, 200,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_BackgroundBH),
        MUIA_ControlChar, Shortcut(GCS(MSG_Background)),
        MUIA_Imagedisplay_Spec, &d->d_BackgroundSpec,
        MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_Background,
        End;
    MUIAddCallback(d->d_Background, MUIA_Pressed, FALSE, PopButtonSetCWD, pr);

    d->d_DefaultBackground = MUI_MakeObject(MUIO_Checkmark, 
	    (IPTR)GCS(MSG_Background_Default));

    back = HGroup,
        Child, d->d_Background,
        Child, HVSpace,
        Child, VGroup,
            Child, VSpace(0),
            Child, HGroup,
                Child, Label1(GCS(MSG_Background_Default)),
                Child, d->d_DefaultBackground,
                End,
            Child, VSpace(0),
            End,
        End;

    basic = VGroup,
	MUIA_Group_Columns, 2,
	Child, Label2(GCS(MSG_Title)),
	Child, d->d_Title,
	Child, Label1(GCS(MSG_Weight)),
	Child, d->d_Weight,
        Child, FreeLabel(GCS(MSG_Background)),
        Child, back,
	End;

    d->d_Border = MUI_MakeObject(MUIO_Checkmark,
            (IPTR)GCS(MSG_DisplayGroup_Border));
    d->d_Horizontal = MUI_MakeObject(MUIO_Checkmark, 
	    (IPTR)GCS(MSG_DisplayGroup_Horizontal));
    d->d_Spacing = MUI_MakeObject(MUIO_Checkmark,
            (IPTR)GCS(MSG_DisplayGroup_Spacing));

    attrs = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_DisplayGroup_Attributes),
        MUIA_Background, MUII_GroupBack,
	MUIA_Group_Columns, 4,
	Child, HSpace(0),
	Child, Label1(GCS(MSG_DisplayGroup_Border)),
	Child, d->d_Border,
	Child, HSpace(0),
	Child, HSpace(0),
	Child, Label1(GCS(MSG_DisplayGroup_Horizontal)),
	Child, d->d_Horizontal,
	Child, HSpace(0),
	Child, HSpace(0),
	Child, Label1(GCS(MSG_DisplayGroup_Spacing)),
	Child, d->d_Spacing,
	Child, HSpace(0),
	End;

    frame = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	Child, basic,
	Child, attrs,
	End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, d->d_Ok,
	Child, d->d_Cancel,
	End;

    root = GroupObject,
	Child, frame,
	Child, buttons,
	End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_Title, GCS(MSG_Group),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'G','R','P'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_HelpNode, "Group editor",
        MUIA_UserData, d,
	WindowContents, root,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct EditGroupData));
	return;
    }

    SetAttrs(d->d_Border,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayGroup_BorderBH),
        MUIA_Selected, DisplayGroupBorder(group),
        TAG_DONE);
    SetAttrs(d->d_Horizontal,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayGroup_HorizontalBH),
        MUIA_Selected, DisplayGroupHoriz(group),
        TAG_DONE);
    SetAttrs(d->d_Spacing,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayGroup_SpacingBH),
        MUIA_Selected, DisplayGroupSpacing(group),
        TAG_DONE);
    DoMethod(d->d_DefaultBackground, MUIM_Notify, MUIA_Selected, MUIV_EveryTime,
        d->d_Background, 3, MUIM_Set, MUIA_Disabled, MUIV_TriggerValue);
    SetAttrs(d->d_DefaultBackground,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_Background_DefaultBH),
        MUIA_Selected, !bg || !bg[0],
        TAG_DONE);
    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    DoMethod(d->d_Title,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE,
        obj, closeEditGroup, FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE,
        obj, closeEditGroup, FALSE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE,
        obj, closeEditGroup, TRUE);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeEditGroup(app, obj);
        return;
    }
    set(obj, MUIA_Window_ActiveObject, d->d_Title);
}

struct EditWindowData
{
    struct Project *d_Project;
    struct DisplayObject *d_Window;
    Object *d_StructureEditor;
    Object *d_Title, *d_Id, *d_Button;
    Object *d_Shortcut, *d_Enabled, *d_EnabledComputeStr, *d_EnabledCompute;
    Object *d_CloseWhenDisabled, *d_Weight, *d_Fontname;
    Object *d_Background, *d_DefaultBackground;
    Object *d_Ok, *d_Cancel;
    struct MUI_ImageSpec d_BackgroundSpec;
};

static void disposeEditWindow(Object *app, Object *obj)
{
    struct EditWindowData *d = pget(obj, MUIA_UserData);

    if(obj)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(struct EditWindowData));
        MUI_AddDisposeObject(app, obj);
    }
}

void HOOK_PROTO(WindowIdAcknowledge, Object *obj, struct DisplayObject **x)
{
    struct DisplayObject *dio = *x, *inuse;
    Object *win = GuiObjectWindow(obj);
    const char *text = pget(obj, MUIA_String_Contents), *name;
    ulong id = StrToId(text);

    inuse = DisplayWindowIdInUse(id, dio);
    if (inuse != NULL)
    {
        name = DisplayObjectName(inuse);
        Ask2(win, NULL, AT_ERROR, GCS(MSG_StarOk),
		GCS(MSG_DisplayWindow_IdInUse), EmptyNullStr(name));
        if (win != NULL)
            set(win, MUIA_Window_ActiveObject, obj);
    }
}

static void HOOK_PROTO(closeEditWindow, Object *obj, ulong *ok)
{
    struct EditWindowData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);
    struct WindowData wd;
    const char *idstr;
    long error;

    if(*ok)
    {
        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        wd.wd_Title = pget(d->d_Title, MUIA_String_Contents);
        idstr = pget(d->d_Id, MUIA_String_Contents);
        wd.wd_Id = StrToId(idstr);
        wd.wd_Button = nget(d->d_Button, MUIA_Selected);
        wd.wd_Shortcut = *(char *)pget(d->d_Shortcut, MUIA_String_Contents);
        wd.wd_Enabled = nget(d->d_Enabled, MUIA_Radio_Active);
        wd.wd_EnabledCompute=pget(d->d_EnabledComputeStr, MUIA_String_Contents);
        wd.wd_CloseWhenDisabled = nget(d->d_CloseWhenDisabled, MUIA_Selected);
        wd.wd_Weight = nget(d->d_Weight, MUIA_Numeric_Value);
        wd.wd_Fontname = pget(d->d_Fontname, MUIA_String_Contents);
        wd.wd_Background = nget(d->d_DefaultBackground, MUIA_Selected)?
            NULL : pget(d->d_Background, MUIA_Imagedisplay_Spec);

        if (idstr && idstr[0] == '_') {
            Ask2(obj, NULL, AT_ERROR, GCS(MSG_StarOk),
            GCS(MSG_DisplayWindow_BadId));
            set(obj, MUIA_Window_ActiveObject, d->d_Id);
            return;
         } 

        GuiSetSleep(TRUE);
        DisplayChangeWindow(d->d_Window, &wd);
        ProjectChange(pr, TRUE);
        GuiSetSleep(FALSE);

        WindowFinish(d->d_StructureEditor, d->d_Window);
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposeEditWindow(app, obj);
}

static void HOOK_PROTO(editWindowOnEnabledChanged, Object *obj, 
	struct EditWindowData **x)
{
    struct EditWindowData *d = *x;
    short enabled = nget(d->d_Enabled, MUIA_Radio_Active);

    set(d->d_EnabledCompute, MUIA_Disabled, enabled != DOD_ENABLED_COMPUTED);
    set(d->d_CloseWhenDisabled, MUIA_Disabled, enabled == DOD_ENABLED_ALWAYS);
    if (enabled == DOD_ENABLED_ALWAYS) {
        set(d->d_CloseWhenDisabled, MUIA_Selected, FALSE);
    }
}

void GuiWindow(Object *se, struct DisplayObject *window)
{
    struct Project *pr = DisplayObjectProject(window);
    struct MemPool *memPool = ProjectMemPool(pr);
    Object *app = GuiApplication;
    const char *bg = DisplayObjectBackground(window);
    char sc[2];
    struct EditWindowData *d;
    Object *back, *settings1, *settings2, *settings3, *settings;
    Object *root, *top;
    Object *obj;
    static const char *etext[4];

    if(app == NULL || window == NULL)
	return;

    d = PoolAllocMemZero(memPool, sizeof(struct EditWindowData));
    if(d == NULL)
        return;
    d->d_Project = pr;
    d->d_Window = window;
    d->d_StructureEditor = se;

    d->d_Title = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayWindow_TitleBH),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Title)),
	MUIA_String_MaxLen, TITLE_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, DisplayObjectTitle(window),
	End;
    
    d->d_Id = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayWindow_IdBH),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_DisplayWindow_Id)),
	MUIA_String_MaxLen, 4 + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, IdToStr(DisplayWindowId(window)),
	End;

    d->d_Button = MUI_MakeObject(MUIO_Checkmark, 
        (IPTR)GCS(MSG_DisplayWindow_WindowButton));

    sc[0] = DisplayObjectShortcut(window);
    sc[1] = 0;
    d->d_Shortcut = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayObject_ShortcutBH),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_DisplayObject_Shortcut)),
	MUIA_String_MaxLen, 2,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, sc,
	End;

    etext[0] = GCS(MSG_DisplayObject_Enabled_Always);
    etext[1] = GCS(MSG_DisplayObject_Enabled_NotInitial);
    etext[2] = GCS(MSG_DisplayObject_Enabled_Computed);
    etext[3] = 0;
    d->d_Enabled = RadioObject,
        MUIA_Radio_Entries, etext,
        MUIA_Radio_Active, DOD_ENABLED_COMPUTED,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayObject_EnabledBH),
        End;

    d->d_EnabledComputeStr = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_Frame, MUIV_Frame_String,
        MUIA_ControlChar, Shortcut(GCS(MSG_Compute)),
        MUIA_ShortHelp, GCS(MSG_DisplayObject_ComputeEnabledBH),
        MUIA_String_MaxLen, FIELD_TRIGGER_LEN + 1,
        MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, DisplayObjectEnabledCompute(window),
        End;

    d->d_EnabledCompute = UserFunctionsPopup(d->d_EnabledComputeStr, pr);

    d->d_CloseWhenDisabled = MUI_MakeObject(MUIO_Checkmark, 
            (IPTR)GCS(MSG_DisplayWindow_CloseWindowWhenDisabled));

    d->d_Weight = SliderObject,
        MUIA_CycleChain, TRUE,
	MUIA_ShortHelp, GCS(MSG_WeightBH),
	MUIA_Frame, MUIV_Frame_Slider,
	MUIA_ControlChar, Shortcut(GCS(MSG_Weight)),
	MUIA_Numeric_Min, 0,
	MUIA_Numeric_Max, 1000,
        MUIA_Numeric_Value, DisplayObjectWeight(window),
	End;

    d->d_Fontname = StringObject,
        MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Font)),
        MUIA_String_MaxLen, FONTNAME_LEN,
        MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, DisplayObjectMUIFontname(window),
        MUIA_ShortHelp, GCS(MSG_FontBH),
        MUIA_CycleChain, TRUE,
        End;

    CodesetStrNBytesCopy(d->d_BackgroundSpec.buf, 
	    (bg && bg[0])? bg : IMAGESPEC_START, IMAGESPEC_LEN);
    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    d->d_Background = PopimageObject,
        MUIA_Weight, 200,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_BackgroundBH),
        MUIA_ControlChar, Shortcut(GCS(MSG_Background)),
        MUIA_Imagedisplay_Spec, &d->d_BackgroundSpec,
        MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_Background,
        End;
    MUIAddCallback(d->d_Background, MUIA_Pressed, FALSE, PopButtonSetCWD, pr);

    d->d_DefaultBackground = MUI_MakeObject(MUIO_Checkmark, 
	    (IPTR)GCS(MSG_Background_Default));

    back = HGroup,
        Child, d->d_Background,
        Child, HVSpace,
        Child, VGroup,
            Child, VSpace(0),
            Child, HGroup,
                Child, Label1(GCS(MSG_Background_Default)),
                Child, d->d_DefaultBackground,
                End,
            Child, VSpace(0),
            End,
        End;

    top = VGroup,
	MUIA_Group_Columns, 2,
	Child, Label2(GCS(MSG_Title)),
	Child, d->d_Title,
	Child, Label2(GCS(MSG_DisplayWindow_Id)),
	Child, HGroup, 
	    Child, d->d_Id,
	    Child, HSpace(0),
	    End,
	Child, Label1(GCS(MSG_DisplayWindow_WindowButton)),
	Child, HGroup,
            Child, d->d_Button,
	    Child, HSpace(0),
	    End,
        End;

    settings1 = HGroup,
	Child, Label2(GCS(MSG_DisplayObject_Shortcut)),
        Child, d->d_Shortcut,
        Child, HSpace(0),
        End;

    settings2 = HGroup,
        Child, d->d_Enabled,
        Child, RectangleObject,
            MUIA_Rectangle_VBar, TRUE,
            MUIA_FixWidth, 10,
            End,
        Child, VGroup,
            Child, VSpace(0),
            Child, CLabel(GCS(MSG_Compute)),
            Child, d->d_EnabledCompute,
            End,
        End;
    settings2 = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_FrameTitle, GCS(MSG_DisplayObject_Enabled),
        Child, settings2,
        Child, RectangleObject,
            MUIA_Rectangle_HBar, TRUE,
            MUIA_FixHeight, 5,
            End,
        Child, HGroup,
            Child, Label1(GCS(MSG_DisplayWindow_CloseWindowWhenDisabled)),
            Child, d->d_CloseWhenDisabled,
            Child, HSpace(0),
            End,
        End;

    settings3 = VGroup,
        MUIA_Group_Columns, 2,
	Child, Label1(GCS(MSG_Weight)),
	Child, d->d_Weight,
        Child, Label2(GCS(MSG_Font)),
        Child, PopaslObject,
	    MUIA_Popstring_String, d->d_Fontname,
	    MUIA_Popstring_Button, PopButton(MUII_PopUp),
            MUIA_Popasl_Type, ASL_FontRequest,
            ASLFO_MaxHeight, 128,
            ASLFO_Flags, FOF_DOFRONTCOLOR | FOF_DOBACKCOLOR | FOF_DOSTYLE | 
                FOF_DOMUISTYLES,
            End,
        Child, FreeLabel(GCS(MSG_Background)),
        Child, back,
	End;

    settings = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_FrameTitle, GCS(MSG_DisplayWindow_ButtonSettings),
        Child, settings1,
        Child, settings2,
        Child, settings3,
        End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    root = GroupObject,
        Child, VGroup,
            MUIA_Frame, MUIV_Frame_Group,
            MUIA_Background, MUII_GroupBack,
            Child, top,
            Child, settings,
            End,
	Child, HGroup,
            MUIA_Group_SameWidth, TRUE,
	    Child, d->d_Ok,
	    Child, d->d_Cancel,
            End,
	End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_Title, GCS(MSG_Window),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'W','I','N'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_HelpNode, "Window editor",
        MUIA_UserData, d,
	WindowContents, root,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct EditWindowData));
	return;
    }

    DoMethod(d->d_Button, MUIM_Notify, MUIA_Selected, MUIV_EveryTime,
        settings, 3, MUIM_Set, MUIA_Disabled, MUIV_NotTriggerValue);
    SetAttrs(d->d_Button,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayWindow_WindowButtonBH),
        MUIA_Selected, DisplayWindowButton(window),
        TAG_DONE);
    DoMethod(d->d_Title,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    DoMethod(d->d_Id,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    MUIAddCallback(d->d_Id, MUIA_String_Acknowledge, MUIV_EveryTime,
	WindowIdAcknowledge, window);
    DoMethod(d->d_Shortcut,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    MUIAddCallback(d->d_Shortcut, MUIA_String_Acknowledge, MUIV_EveryTime,
        ShortcutAcknowledge, window);
    SetAttrs(d->d_CloseWhenDisabled,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayWindow_CloseWindowWhenDisabledBH),
        MUIA_Selected, DisplayWindowCloseWhenDisabled(window),
        TAG_DONE);
    MUIAddCallback(d->d_Enabled, MUIA_Radio_Active, MUIV_EveryTime,
        editWindowOnEnabledChanged, d);
    set(d->d_Enabled, MUIA_Radio_Active, DisplayObjectEnabled(window));
    DoMethod(d->d_EnabledComputeStr,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    DoMethod(d->d_Fontname,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    DoMethod(d->d_DefaultBackground, MUIM_Notify, MUIA_Selected, MUIV_EveryTime,
        d->d_Background, 3, MUIM_Set, MUIA_Disabled, MUIV_TriggerValue);
    SetAttrs(d->d_DefaultBackground,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_Background_DefaultBH),
        MUIA_Selected, !bg || !bg[0],
        TAG_DONE);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE,
        obj, closeEditWindow, FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE,
        obj, closeEditWindow, FALSE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE,
        obj, closeEditWindow, TRUE);
    set(d->d_Ok, MUIA_CycleChain, TRUE);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeEditWindow(app, obj);
        return;
    }
    set(obj, MUIA_Window_ActiveObject, d->d_Title);
}

#endif

