/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** displayfield_mui.c
**
** Creation date: 08.12.1994
** Last modified: 10.02.2022
*/

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "structureeditor_mui.h"
#include "recorddescriptionspec_mui.h"
#include "extclasses/ExtApplication_mcc.h"
#include "extclasses/ExtEditor_mcc.h"
#include "extclasses/PopEditLabels_mcc.h"
#include <libraries/asl.h>
#include "table.h"
#include "requesters.h"
#include "preferences.h"
#include "util.h"
#include "exec.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

#define IMAGESPEC_LEN (int)(sizeof(struct MUI_ImageSpec) - 1)
#define IMAGESPEC_START "0:128"

struct EditDisplayFieldData
{
    struct Project *d_Project;
    struct DisplayObject *d_DisplayField;
    Object *d_StructureEditor;
    Object *d_Title, *d_TitlePosition, *d_TitleFontname, *d_TitleFontnamePopup;
    Object *d_Shortcut, *d_Home, *d_TabChain, *d_ReadOnly;
    Object *d_Alignment, *d_Calendar, *d_Format, *d_ShowTitles; 
    Object *d_SubType, *d_SubTypeGroup;
    Object *d_Frame, *d_Horizontal, *d_ControlTable;
    Object *d_OnDoubleClick, *d_DoubleClickTriggerStr, *d_DoubleClickTrigger;
    Object *d_URLDropTriggerStr, *d_URLDropTrigger;
    Object *d_SortDropTriggerStr, *d_SortDropTrigger;
    Object *d_NumDecimals, *d_AutoUpdate;
    Object *d_Enabled, *d_EnabledComputeStr, *d_EnabledCompute;
    Object *d_Weight, *d_BubbleHelp;
    Object *d_Fontname, *d_FontnamePopup, *d_Background, *d_DefaultBackground;
    Object *d_Image, *d_ImageSize, *d_InitialColor;
    Object *d_RecordDescriptionSpec, *d_RecordDescriptionSpecGroup;
    Object *d_ReferencePopupGroup, *d_RecordsKind; 
    Object *d_ComputeRecordsStr, *d_ComputeRecords, *d_ComputeRecordsGroup;
    Object *d_InitialRecord, *d_CurrentRecord;
    Object *d_Show, *d_AutoShow, *d_Filter;
    Object *d_DisplayExternal, *d_HideStringField, *d_TitleAtStringField;
    Object *d_Listview, *d_File, *d_Font, *d_View;
    Object *d_ListviewSettings, *d_ListviewKind, *d_ListviewKindGroup;
    Object *d_ListviewLabels, *d_ListviewComputeStr, *d_ListviewCompute;
    Object *d_Ok, *d_Cancel;
    struct MUI_ImageSpec d_BackgroundSpec;
    struct MUI_ImageSpec d_ImageSpec;
};

static void disposeEditDisplayField(Object *app, Object *obj)
{
    struct EditDisplayFieldData *d = pget(obj, MUIA_UserData);

    EditRecordDescriptionSpecCleanup(d->d_RecordDescriptionSpec, d->d_Project);
    PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(*d));
    MUI_AddDisposeObject(app, obj);
}

static void HOOK_PROTO(closeEditDisplayField, Object *obj, ulong *ok)
{
    struct EditDisplayFieldData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);

    if (*ok) {
        struct Field *field=DisplayFieldField(d->d_DisplayField);
        short type = FieldType(field);
        struct MemPool *memPool = ProjectMemPool(pr);
        char *bubbleHelp = NULL;
        struct RecordDescriptionSpec *rds = NULL;
        const char *sc;
        long error = OK;
        struct DisplayFieldData dfd;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        GuiSetSleep(TRUE);
        memset(&dfd, 0, sizeof(dfd));
        dfd.dfd_Title = pget(d->d_Title, MUIA_String_Contents);
        dfd.dfd_TitlePosition = nget(d->d_TitlePosition, MUIA_Cycle_Active);
        dfd.dfd_TitleFontname = pget(d->d_TitleFontname, MUIA_String_Contents);
        sc = pget(d->d_Shortcut, MUIA_String_Contents);
        dfd.dfd_Shortcut = (sc != NULL)? *sc : 0;
        dfd.dfd_Home = nget(d->d_Home, MUIA_Selected);
        dfd.dfd_TabChain = nget(d->d_TabChain, MUIA_Selected);
        dfd.dfd_ReadOnly = nget(d->d_ReadOnly, MUIA_Selected);;
        dfd.dfd_SubType = nget(d->d_SubType, MUIA_Cycle_Active);
        dfd.dfd_Alignment = nget(d->d_Alignment, MUIA_Cycle_Active);
        dfd.dfd_Calendar = nget(d->d_Calendar, MUIA_Selected);
        dfd.dfd_Format = nget(d->d_Format, MUIA_Cycle_Active);
        dfd.dfd_NumDecimals = nget(d->d_NumDecimals, MUIA_String_Integer);
        dfd.dfd_ShowTitles = nget(d->d_ShowTitles, MUIA_Selected);
        dfd.dfd_ControlTable = nget(d->d_ControlTable, MUIA_Selected);
        dfd.dfd_OnDoubleClick = nget(d->d_OnDoubleClick, MUIA_Radio_Active);
        dfd.dfd_DoubleClickTrigger = 
	    pget(d->d_DoubleClickTriggerStr, MUIA_String_Contents);
        dfd.dfd_URLDropTrigger = 
	    pget(d->d_URLDropTriggerStr, MUIA_String_Contents);
        dfd.dfd_SortDropTrigger = 
	    pget(d->d_SortDropTriggerStr, MUIA_String_Contents);
        dfd.dfd_AutoUpdate = nget(d->d_AutoUpdate, MUIA_Selected);
        dfd.dfd_Enabled = nget(d->d_Enabled, MUIA_Radio_Active);
        dfd.dfd_EnabledCompute = pget(d->d_EnabledComputeStr, 
                MUIA_String_Contents);
        dfd.dfd_Weight = nget(d->d_Weight, MUIA_Numeric_Value);
        dfd.dfd_Fontname = pget(d->d_Fontname, MUIA_String_Contents);

        if (type == TYPE_BUTTON && dfd.dfd_SubType == DAD_BUTTONTYPE_IMAGE) {
            dfd.dfd_Background = pget(d->d_Image, MUIA_Imagedisplay_Spec);
        } else if (!nget(d->d_DefaultBackground, MUIA_Selected)) {
            dfd.dfd_Background = pget(d->d_Background, MUIA_Imagedisplay_Spec);
        }
        dfd.dfd_ImageSize = nget(d->d_ImageSize, MUIA_Radio_Active);
        if (d->d_BubbleHelp) {
            bubbleHelp = GuiEditorText(d->d_BubbleHelp, memPool);
            dfd.dfd_BubbleHelp = bubbleHelp;
            if(dfd.dfd_BubbleHelp == NULL)
                error = ERR_MEM;
        }
        dfd.dfd_File = nget(d->d_File, MUIA_Selected);
        dfd.dfd_Font = nget(d->d_Font, MUIA_Selected);
        dfd.dfd_Listview = nget(d->d_Listview, MUIA_Selected);
        dfd.dfd_View = nget(d->d_View, MUIA_Selected);
        if (d->d_ListviewLabels) {
            dfd.dfd_ListviewLabels= pget(d->d_ListviewLabels, 
                MUIA_EditLabels_Labels);
        }
        if (nget(d->d_ListviewKind, MUIA_Cycle_Active) == 1) {
            dfd.dfd_ListviewCompute = 
                pget(d->d_ListviewComputeStr, MUIA_String_Contents);
        }
        dfd.dfd_DisplayExternal = nget(d->d_DisplayExternal, MUIA_Selected);
        dfd.dfd_TitleAtStringField=nget(d->d_TitleAtStringField, MUIA_Selected);
        dfd.dfd_HideStringField = nget(d->d_HideStringField, MUIA_Selected);
        dfd.dfd_Show = nget(d->d_Show, MUIA_Selected);
        dfd.dfd_AutoShow = nget(d->d_AutoShow, MUIA_Selected);
        dfd.dfd_Filter = nget(d->d_Filter, MUIA_Selected);
        dfd.dfd_Frame = nget(d->d_Frame, MUIA_Selected);
        if (d->d_Horizontal) {
            dfd.dfd_Alignment = nget(d->d_Horizontal, MUIA_Selected)?
                  DAD_ALIGNMENT_HORIZONTAL : DAD_ALIGNMENT_VERTICAL;
        }
        dfd.dfd_RecordsKind = nget(d->d_RecordsKind, MUIA_Cycle_Active);
    	dfd.dfd_RecordsCompute = 
	    pget(d->d_ComputeRecordsStr, MUIA_String_Contents);
    	dfd.dfd_InitialRecord = nget(d->d_InitialRecord, MUIA_Selected);
    	dfd.dfd_CurrentRecord = nget(d->d_CurrentRecord, MUIA_Selected);

        if (d->d_RecordDescriptionSpec) {
            rds = EditRecordDescriptionSpecGetSpec(
                    d->d_RecordDescriptionSpec, memPool);
            dfd.dfd_RecordDescriptionSpec = rds;
            if (rds == NULL) {
                error = ERR_MEM;
            }
        }

        if (error == OK) {
            error = DisplayChangeField(field, &dfd);
            if (error == OK) {
                ProjectChange(pr, TRUE);
            }
        }

        FreeRecordDescriptionSpec(memPool, rds);
        FreeString(memPool, bubbleHelp);
        GuiSetSleep(FALSE);

         if (error) {
            Ask2(obj, GCS(MSG_ErrorMessage), AT_ERROR, 
		    GCS(MSG_StarOk), GES(error));
            return;
        }

        DisplayFieldFinish(d->d_StructureEditor, d->d_DisplayField);

        if (d->d_Alignment != NULL) {
            SetPrefsValue(pr, (type == TYPE_INTEGER || type == TYPE_REAL)?
        	    PREFS_DEFAULTALIGNMENT_NUMBER : PREFS_DEFAULTALIGNMENT_TEXT,
        	    dfd.dfd_Alignment);
        }
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposeEditDisplayField(app, obj);
}

static void updateFocusWidgets(struct EditDisplayFieldData *d)
{
    bool readOnly = nget(d->d_ReadOnly, MUIA_Selected);
    bool file = nget(d->d_File, MUIA_Selected);
    bool font = nget(d->d_Font, MUIA_Selected);
    bool list = nget(d->d_Listview, MUIA_Selected);
    bool hide = nget(d->d_HideStringField, MUIA_Selected);
    bool show = nget(d->d_Show, MUIA_Selected);
    bool focus = !hide && (!readOnly || file || font || list || show);

    if (d->d_Shortcut) {
        set(d->d_Shortcut, MUIA_Disabled, !focus);
    }
    if (d->d_Home) {
	set(d->d_Home, MUIA_Disabled, !focus);
	if (!focus) {
            set(d->d_Home, MUIA_Selected, FALSE);
	}
    }
    if (d->d_TabChain) {
	set(d->d_TabChain, MUIA_Disabled, !focus);
	if (!focus) {
            set(d->d_TabChain, MUIA_Selected, FALSE);
	}
    }
}

static void HOOK_PROTO(onReadOnlyChanged, Object *obj,
	struct EditDisplayFieldData **x)
{
    struct EditDisplayFieldData *d = *x;
    bool active = nget(d->d_ReadOnly, MUIA_Selected);

    updateFocusWidgets(d);
    if (d->d_SubType) {
        if (active) {
            set(d->d_SubType, MUIA_Cycle_Active, 0);
        }
        set(d->d_SubType, MUIA_Disabled, active);
    }
    if (d->d_Calendar) {
        if (active) {
            set(d->d_Calendar, MUIA_Selected, FALSE);
        }
        set(d->d_Calendar, MUIA_Disabled, active);
    }
    if (d->d_ReferencePopupGroup) {
        set(d->d_ReferencePopupGroup, MUIA_Disabled, active);
	if (!active && nget(d->d_RecordsKind, MUIA_Cycle_Active)
		!= DAD_RECORDS_COMPUTED) {
	    set(d->d_ComputeRecordsGroup, MUIA_Disabled, TRUE);
	}
    }
}

static void HOOK_PROTO(onEnabledChanged, Object *obj, 
	struct EditDisplayFieldData **x)
{
    struct EditDisplayFieldData *d = *x;
    short enabled = nget(d->d_Enabled, MUIA_Radio_Active);

    set(d->d_EnabledCompute, MUIA_Disabled, enabled != DOD_ENABLED_COMPUTED);
}

static void HOOK_PROTO(onFFLChanged, Object *button, 
	struct EditDisplayFieldData **x)
{
    struct EditDisplayFieldData *d = *x;
    bool active = nget(button, MUIA_Selected);

    if (active) {
        if (button != d->d_File) {
            set(d->d_File, MUIA_Selected, FALSE);
        }
        if (button != d->d_Font) {
            set(d->d_Font, MUIA_Selected, FALSE);
        }
        if (button != d->d_Listview) {
            set(d->d_Listview, MUIA_Selected, FALSE);
        }
    }
    if (button == d->d_Listview) {
        set(d->d_ListviewSettings, MUIA_Disabled, !active);
    }
    updateFocusWidgets(d);
}

static void HOOK_PROTO(onDoubleClickChange, Object *obj,
	struct EditDisplayFieldData **x)
{
    struct EditDisplayFieldData *d = *x;
    short action = nget(d->d_OnDoubleClick, MUIA_Radio_Active);

    set(d->d_DoubleClickTrigger, MUIA_Disabled, 
	    action != DAD_DOUBLECLICK_CALLTRIGGER);
}

static void HOOK_PROTO(onHideStringFieldChanged, Object *obj, 
	struct EditDisplayFieldData **x)
{
    struct EditDisplayFieldData *d = *x;
    bool active = nget(d->d_HideStringField, MUIA_Selected);

    if (active) {
        set(d->d_File, MUIA_Selected, FALSE);
        set(d->d_Font, MUIA_Selected, FALSE);
        set(d->d_Listview, MUIA_Selected, FALSE);
        set(d->d_View, MUIA_Selected, FALSE);
    }
    set(d->d_ReadOnly, MUIA_Disabled, active);
    set(d->d_Alignment, MUIA_Disabled, active);
    set(d->d_Fontname, MUIA_Disabled, active);
    set(d->d_FontnamePopup, MUIA_Disabled, active);
    set(d->d_Background, MUIA_Disabled, active);
    set(d->d_DefaultBackground, MUIA_Disabled, active);
    set(d->d_File, MUIA_Disabled, active);
    set(d->d_Font, MUIA_Disabled, active);
    set(d->d_Listview, MUIA_Disabled, active);
    set(d->d_View, MUIA_Disabled, active);
    updateFocusWidgets(d);
}

static void HOOK_PROTO(onShowChanged, Object *obj,
	struct EditDisplayFieldData **x)
{
    struct EditDisplayFieldData *d = *x;

    updateFocusWidgets(d);
}

static void HOOK_PROTO(onRecordsChanged, Object *obj, 
	struct EditDisplayFieldData **x)
{
    struct EditDisplayFieldData *d = *x;

    set(d->d_ComputeRecordsGroup, MUIA_Disabled,
	    nget(d->d_RecordsKind, MUIA_Cycle_Active) != DAD_RECORDS_COMPUTED);
}

void GuiDisplayField(Object *se, struct DisplayObject *dio)
{
    static const char *etext[4];
    Object *app = GuiApplication;
    struct Project *pr = DisplayObjectProject(dio);
    struct MemPool *memPool = ProjectMemPool(pr);
    struct Field *field = DisplayFieldField(dio);
    short type = FieldType(field);
    short subType = DisplayFieldSubType(dio);
    bool readOnly = DisplayFieldReadOnly(dio);
    const char *lvcompute = DisplayFieldListviewCompute(dio);
    struct EditDisplayFieldData *d = NULL;
    long winid = MAKE_ID('_', 'A', 'D', 'E');
    bool show_shortcut = TRUE, show_readonly = TRUE,
        show_virtualtype = FALSE, show_buttontype = FALSE,
        show_choicetype = FALSE, show_alignment = FALSE,
        show_calendar = FALSE, show_format = FALSE,
        show_numdecimals = FALSE, show_weight = TRUE,
        show_fontname = TRUE, show_bg = TRUE,
        show_stringbuttons = FALSE, show_reference = FALSE, extras = FALSE;
    Object *titlePositionLabel = NULL, *titleFontnameLabel = NULL,
           *shortcutLabel = NULL, *shortcutGroup = NULL, *subTypeLabel = NULL,
           *alignmentLabel = NULL, *alignmentGroup = NULL,
           *numdecimalsLabel = NULL, *calendarLabel = NULL, *formatLabel = NULL,
           *enabledGroup = NULL, *weightLabel = NULL, *fontnameLabel = NULL,
           *bgLabel = NULL, *bgGroup = NULL, *editor = NULL,
           *topGroup = NULL, *fieldGroup = NULL, *externalSizeGroup = NULL,
           *attrsGroup = NULL, *extrasGroup = NULL, *mainGroup = NULL;
    Object *obj;

    if (app == NULL || dio == NULL) {
	return;
    }

    d = PoolAllocMemZero(memPool, sizeof(struct EditDisplayFieldData));
    if (d == NULL) {
        return;
    }
    d->d_Project = pr;
    d->d_DisplayField = dio;
    d->d_StructureEditor = se;

    switch (type)
    {
        case TYPE_BOOL:
            show_fontname = FALSE;
            winid = MAKE_ID('_', 'A', 'B', 'D');
            break;

        case TYPE_CHOICE:
            show_choicetype = TRUE;
            winid = MAKE_ID('_', 'A', 'C', 'D');
            break;

        case TYPE_REAL:
            show_numdecimals = TRUE;
            show_alignment = TRUE;
            winid = MAKE_ID('_', 'A', 'R', 'D');
            break;

        case TYPE_INTEGER:
            show_alignment = TRUE;
            winid = MAKE_ID('_', 'A', 'I', 'D');
            break;

        case TYPE_DATE:
            show_alignment = TRUE;
            show_calendar = TRUE;
            winid = MAKE_ID('_', 'A', 'D', 'D');
            break;

        case TYPE_TIME:
            show_alignment = TRUE;
            show_format = TRUE;
            winid = MAKE_ID('_', 'A', 'T', 'D');
            break;

        case TYPE_STRING:
            show_alignment = TRUE;
            show_stringbuttons = TRUE;
            winid = MAKE_ID('_', 'A', 'S', 'D');
            break;

        case TYPE_REFERENCE:
            show_alignment = TRUE;
            show_reference = TRUE;
            winid = MAKE_ID('_', 'A', 'F', 'D');
            break;

        case TYPE_BUTTON:
            show_readonly = FALSE;
            show_buttontype = TRUE;
            winid = MAKE_ID('_', 'A', 'U', 'D');
            break;

        case TYPE_VIRTUAL:
            show_readonly = FALSE;
            show_shortcut = FALSE;
            show_virtualtype = TRUE;
            show_alignment = TRUE;
            show_numdecimals = TRUE;
            winid = MAKE_ID('_', 'A', 'V', 'D');
            break;
    }
    extras = show_stringbuttons || show_reference || show_virtualtype;

    d->d_Title = StringObject,
        MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Title)),
        MUIA_ShortHelp, GCS(MSG_DisplayField_TitleBH),
        MUIA_CycleChain, TRUE,
        MUIA_String_MaxLen, 256,
        MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, DisplayObjectTitle(dio),
        End;

    if (!show_buttontype) {
        static const char *titlepos[5];

        titlepos[0] = GCS(MSG_DisplayField_Alignment_Left);
        titlepos[1] = GCS(MSG_DisplayField_Alignment_Right);
        titlepos[2] = GCS(MSG_DisplayField_Alignment_Top);
        titlepos[3] = GCS(MSG_DisplayField_Alignment_Bottom);
        titlepos[4] = NULL;
        d->d_TitlePosition = CycleObject,
            MUIA_ShortHelp, GCS(MSG_DisplayField_TitlePositionBH),
            MUIA_ControlChar, Shortcut(GCS(MSG_DisplayField_TitlePosition)),
            MUIA_CycleChain, TRUE,
            MUIA_Weight, 50,
            MUIA_Cycle_Entries, titlepos,
            MUIA_Cycle_Active, DisplayFieldTitlePosition(dio),
            End;

        titlePositionLabel = Label(GCS(MSG_DisplayField_TitlePosition));

        d->d_TitleFontname = StringObject,
            MUIA_Frame, MUIV_Frame_String,
	    MUIA_ControlChar, Shortcut(GCS(MSG_Font)),
            MUIA_ShortHelp, GCS(MSG_FontBH),
            MUIA_CycleChain, TRUE,
            MUIA_String_MaxLen, FONTNAME_LEN,
            MUIA_String_Format, MUIV_String_Format_Center,
            MUIA_String_Contents, DisplayFieldTitleMUIFontname(dio),
            End;

        titleFontnameLabel = Label2(GCS(MSG_Font));

        d->d_TitleFontnamePopup = PopaslObject,
	    MUIA_Popstring_String, d->d_TitleFontname,
	    MUIA_Popstring_Button, PopButton(MUII_PopUp),
            MUIA_Popasl_Type, ASL_FontRequest,
            ASLFO_MaxHeight, 128,
            ASLFO_Flags, FOF_DOFRONTCOLOR | FOF_DOBACKCOLOR | FOF_DOSTYLE | 
                FOF_DOMUISTYLES,
            End;
    }

    if (show_shortcut) {
        char sc[2];

        sc[0] = DisplayObjectShortcut(dio);
        sc[1] = 0;
        d->d_Shortcut = StringObject,
            MUIA_Frame, MUIV_Frame_String,
            MUIA_ShortHelp, GCS(MSG_DisplayObject_ShortcutBH),
            MUIA_ControlChar, Shortcut(GCS(MSG_DisplayObject_Shortcut)),
            MUIA_CycleChain, TRUE,
            MUIA_String_MaxLen, 2,
	    MUIA_String_Format, MUIV_String_Format_Center,
            MUIA_String_Contents, sc,
            MUIA_Weight, 300,
	    End;
	if (d->d_Shortcut) {
	    MUIAddCallback(d->d_Shortcut, 
		    MUIA_String_Acknowledge, MUIV_EveryTime, 
		    ShortcutAcknowledge, dio);
	}
        shortcutLabel = Label(GCS(MSG_DisplayObject_Shortcut));

        d->d_Home = MUI_MakeObject(MUIO_Checkmark, 
                (IPTR)GCS(MSG_DisplayField_Home));
        if (d->d_Home) {
            SetAttrs(d->d_Home,
                    MUIA_ShortHelp, GCS(MSG_DisplayField_HomeBH),
                    MUIA_CycleChain, TRUE,
                    MUIA_Selected, DisplayFieldHome(dio),
                    TAG_DONE);
	}

        d->d_TabChain = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_TabChain));
	if (d->d_TabChain) {
	    SetAttrs(d->d_TabChain,
		    MUIA_ShortHelp, GCS(MSG_DisplayField_TabChainBH),
		    MUIA_CycleChain, TRUE,
		    MUIA_Selected, DisplayFieldTabChain(dio),
		    TAG_DONE);
	}

        if (show_readonly) {
	    d->d_ReadOnly = MUI_MakeObject(MUIO_Checkmark, 
	    	    (IPTR)GCS(MSG_DisplayField_ReadOnly));
            if (d->d_ReadOnly) {
                SetAttrs(d->d_ReadOnly,
                        MUIA_ShortHelp, show_stringbuttons?
                            GCS(MSG_DisplayField_ReadOnlyStringBH) :
                            GCS(MSG_DisplayField_ReadOnlyBH),
                        MUIA_CycleChain, TRUE,
                        TAG_DONE);
            }
            shortcutGroup = HGroup,
                Child, d->d_Shortcut,
                Child, HSpace(0),
                Child, Label(GCS(MSG_DisplayField_Home)),
		Child, VCenter(d->d_Home),
		Child, HSpace(0),
		Child, Label(GCS(MSG_DisplayField_TabChain)),
		Child, VCenter(d->d_TabChain),
                Child, HSpace(0),
                Child, Label(GCS(MSG_DisplayField_ReadOnly)),
                Child, VCenter(d->d_ReadOnly),
		End;
        } else {
	    shortcutGroup = HGroup,
	        Child, d->d_Shortcut,
		Child, HSpace(0),
		Child, Label(GCS(MSG_DisplayField_Home)),
		Child, VCenter(d->d_Home),
		Child, HSpace(0),
		Child, Label(GCS(MSG_DisplayField_TabChain)),
		Child, VCenter(d->d_TabChain),
		End;
	}
    }

    if (show_virtualtype) {
        static const char *vttext[4];

        vttext[0] = GCS(MSG_Boolean);
        vttext[1] = GCS(MSG_Text);
        vttext[2] = GCS(MSG_DisplayField_Virtual_Kind_List);
        vttext[3] = NULL;
        d->d_SubType = CycleObject,
            MUIA_ShortHelp, GCS(MSG_DisplayField_Virtual_KindBH),
            MUIA_CycleChain, TRUE,
	    MUIA_Cycle_Entries, vttext,
            MUIA_Cycle_Active, subType,
	    End;

        subTypeLabel = Label(GCS(MSG_DisplayField_Kind));
    }

    if (show_buttontype) {
        static const char *bttext[3];

        bttext[0] = GCS(MSG_DisplayField_Button_Kind_TextButton);
        bttext[1] = GCS(MSG_DisplayField_Button_Kind_ImageButton);
        bttext[2] = NULL;
        d->d_SubType = CycleObject,
	    MUIA_ControlChar, Shortcut(GCS(MSG_DisplayField_Kind)),
            MUIA_ShortHelp, GCS(MSG_DisplayField_Button_KindBH),
            MUIA_CycleChain, TRUE,
	    MUIA_Cycle_Entries, bttext,
            MUIA_Cycle_Active, subType,
	    End;

        subTypeLabel = Label(GCS(MSG_DisplayField_Kind));
    }

    if (show_choicetype) {
        static const char *cttext[3];
        Object *page0, *page1;

        cttext[0] = GCS(MSG_DisplayField_Choice_Kind_CycleButton);
        cttext[1] = GCS(MSG_DisplayField_Choice_Kind_RadioButtons);
        cttext[2] = NULL;
        d->d_SubType = CycleObject,
            MUIA_CycleChain, TRUE,
	    MUIA_Disabled, readOnly,
	    MUIA_Cycle_Entries, cttext,
            MUIA_Cycle_Active, subType,
	    MUIA_ControlChar, Shortcut(GCS(MSG_DisplayField_Kind)),
            MUIA_ShortHelp, GCS(MSG_DisplayField_Choice_KindBH),
	    End;

        subTypeLabel = Label(GCS(MSG_DisplayField_Kind));

        page0 = HGroup,
            Child, titlePositionLabel,
            Child, d->d_TitlePosition,
            Child, HSpace(10),
            Child, titleFontnameLabel,
            Child, d->d_TitleFontnamePopup,
            End;

        d->d_Frame = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_Choice_Frame));
	if (d->d_Frame != NULL) {
	    SetAttrs(d->d_Frame,
		    MUIA_ShortHelp, GCS(MSG_DisplayField_Choice_FrameBH),
		    MUIA_CycleChain, TRUE,
		    MUIA_Selected, DisplayFieldFrame(dio),
		    TAG_DONE);
	}

	d->d_Horizontal = MUI_MakeObject(MUIO_Checkmark,
	       	(IPTR)GCS(MSG_DisplayField_Choice_Horizontal));
	if (d->d_Horizontal != NULL) {
	    SetAttrs(d->d_Horizontal,
		MUIA_ShortHelp, GCS(MSG_DisplayField_Choice_HorizontalBH),
		MUIA_CycleChain, TRUE,
		MUIA_Selected, 
		    DisplayFieldAlignment(dio) == DAD_ALIGNMENT_HORIZONTAL,
		TAG_DONE);
	}
        
        page1 = HGroup,
            Child, HSpace(0),
            Child, Label(GCS(MSG_DisplayField_Choice_Frame)),
            Child, d->d_Frame,
            Child, HSpace(0),
            Child, Label(GCS(MSG_DisplayField_Choice_Horizontal)),
            Child, d->d_Horizontal,
            Child, HSpace(0),
            End;

	d->d_SubTypeGroup = VGroup,
	    MUIA_Group_PageMode, TRUE,
	    Child, page0,
	    Child, page1,
            End;
    }

    if (show_alignment) {
        static const char *alignmentText[4];

	alignmentText[0] = GCS(MSG_DisplayField_Alignment_Center);
	alignmentText[1] = GCS(MSG_DisplayField_Alignment_Left);
	alignmentText[2] = GCS(MSG_DisplayField_Alignment_Right);
	alignmentText[3] = NULL;

        d->d_Alignment = CycleObject,
	    MUIA_ControlChar, Shortcut(GCS(MSG_DisplayField_Alignment)),
            MUIA_ShortHelp, GCS(MSG_DisplayField_AlignmentBH),
            MUIA_CycleChain, TRUE,
	    MUIA_Cycle_Entries, alignmentText,
            MUIA_Cycle_Active, DisplayFieldAlignment(dio),
	    End;
        alignmentLabel = Label(GCS(MSG_DisplayField_Alignment));
    }

    if (show_calendar) {
        d->d_Calendar = MUI_MakeObject(MUIO_Checkmark, 
                (IPTR)GCS(MSG_DisplayField_Calendar));
        if (d->d_Calendar) {
            SetAttrs(d->d_Calendar,
                    MUIA_Selected, DisplayFieldCalendar(dio),
                    MUIA_ShortHelp, GCS(MSG_DisplayField_CalendarBH),
                    MUIA_CycleChain, TRUE,
                    TAG_DONE);
        }
        calendarLabel = Label(GCS(MSG_DisplayField_Calendar));
    }
    if (show_format) {
        static const char *formatText[4];

	formatText[0] = GCS(MSG_DisplayField_Format_HHMMSS);
	formatText[1] = GCS(MSG_DisplayField_Format_MMSS);
	formatText[2] = GCS(MSG_DisplayField_Format_HHMM);
	formatText[3] = NULL;

        d->d_Format = CycleObject,
	    MUIA_ControlChar, Shortcut(GCS(MSG_DisplayField_Format)),
            MUIA_ShortHelp, GCS(MSG_DisplayField_FormatBH),
            MUIA_CycleChain, TRUE,
	    MUIA_Cycle_Entries, formatText,
            MUIA_Cycle_Active, DisplayFieldFormat(dio),
	    End;
        formatLabel = Label(GCS(MSG_DisplayField_Format));
    }
    if (show_numdecimals) {
        d->d_NumDecimals = StringObject,
	    MUIA_Frame, MUIV_Frame_String,
	    MUIA_ControlChar, Shortcut(GCS(MSG_DisplayField_NumDecimals)),
            MUIA_ShortHelp, GCS(MSG_DisplayField_NumDecimalsBH),
            MUIA_CycleChain, TRUE,
	    MUIA_String_Accept, "-0123456789",
	    MUIA_String_MaxLen, 2,
	    MUIA_String_Format, MUIV_String_Format_Center,
            MUIA_String_Contents, IntToStr(DisplayFieldNumDecimals(dio)),
	    End;
	numdecimalsLabel = Label(GCS(MSG_DisplayField_NumDecimals));
    }

    if (show_virtualtype) {
	static const char *doubleClickText[4];
	Object *page0, *page1, *page2;

        d->d_ShowTitles = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_ShowTitles));
        if (d->d_ShowTitles) {
            SetAttrs(d->d_ShowTitles,
		    MUIA_ShortHelp, GCS(MSG_DisplayField_ShowTitlesBH),
		    MUIA_CycleChain, TRUE,
		    MUIA_Selected, DisplayFieldShowTitles(dio),
		    TAG_DONE);
	}
        d->d_TabChain = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_TabChain));
	if (d->d_TabChain) {
	    SetAttrs(d->d_TabChain,
		    MUIA_ShortHelp, GCS(MSG_DisplayField_TabChainBH),
		    MUIA_CycleChain, TRUE,
		    MUIA_Selected, DisplayFieldTabChain(dio),
		    TAG_DONE);
	}
        d->d_ControlTable = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_ControlTable));
	if (d->d_ControlTable) {
	    SetAttrs(d->d_ControlTable,
		    MUIA_ShortHelp, GCS(MSG_DisplayField_ControlTableBH),
		    MUIA_CycleChain, TRUE,
		    MUIA_Selected, DisplayFieldControlTable(dio),
		    TAG_DONE);
        }

	doubleClickText[0] = GCS(MSG_DisplayField_DoubleClick_DoNothing);
	doubleClickText[1] = GCS(MSG_DisplayField_DoubleClick_ShowRecord);
	doubleClickText[2] = GCS(MSG_DisplayField_DoubleClick_CallTrigger);
	doubleClickText[3] = NULL;
	d->d_OnDoubleClick = RadioObject,
	    MUIA_ShortHelp, GCS(MSG_DisplayField_OnDoubleClickBH),
	    MUIA_CycleChain, TRUE,
	    MUIA_Radio_Entries, doubleClickText,
	    End;

	d->d_DoubleClickTriggerStr = StringObject,
	    MUIA_Frame, MUIV_Frame_String,
	    MUIA_ControlChar,Shortcut(GCS(MSG_DisplayField_DoubleClick_Trigger)),
	    MUIA_ShortHelp, GCS(MSG_DisplayField_DoubleClick_TriggerBH),
	    MUIA_CycleChain, TRUE,
	    MUIA_String_MaxLen, FIELD_TRIGGER_LEN + 1,
	    MUIA_String_Format, MUIV_String_Format_Center,
	    MUIA_String_Contents, DisplayFieldDoubleClickTrigger(dio),
	    End;

	d->d_DoubleClickTrigger = UserFunctionsPopup(
                d->d_DoubleClickTriggerStr, pr);

        if (d->d_OnDoubleClick) {
            MUIAddCallback(d->d_OnDoubleClick, 
                    MUIA_Radio_Active, MUIV_EveryTime,
                    onDoubleClickChange, d);
            set(d->d_OnDoubleClick, MUIA_Radio_Active, 
                    DisplayFieldOnDoubleClick(dio));
        }

	d->d_URLDropTriggerStr = StringObject,
	    MUIA_Frame, MUIV_Frame_String,
	    MUIA_ControlChar, Shortcut(GCS(MSG_DisplayField_URLDrop_Trigger)),
	    MUIA_ShortHelp, GCS(MSG_DisplayField_URLDrop_TriggerBH),
	    MUIA_CycleChain, TRUE,
	    MUIA_String_MaxLen, FIELD_TRIGGER_LEN + 1,
	    MUIA_String_Format, MUIV_String_Format_Center,
	    MUIA_String_Contents, DisplayFieldURLDropTrigger(dio),
	    End;

	d->d_URLDropTrigger = UserFunctionsPopup(d->d_URLDropTriggerStr, pr);

	d->d_SortDropTriggerStr = StringObject,
	    MUIA_Frame, MUIV_Frame_String,
	    MUIA_ControlChar, Shortcut(GCS(MSG_DisplayField_SortDrop_Trigger)),
	    MUIA_ShortHelp, GCS(MSG_DisplayField_SortDrop_TriggerBH),
	    MUIA_CycleChain, TRUE,
	    MUIA_String_MaxLen, FIELD_TRIGGER_LEN + 1,
	    MUIA_String_Format, MUIV_String_Format_Center,
	    MUIA_String_Contents, DisplayFieldSortDropTrigger(dio),
	    End;

	d->d_SortDropTrigger = UserFunctionsPopup(d->d_SortDropTriggerStr, pr);

        page0 = HVSpace;

        page1 = VGroup,
            MUIA_Group_Columns, 3,
            Child, alignmentLabel,
            Child, d->d_Alignment,
            Child, HSpace(0),
            Child, numdecimalsLabel,
            Child, d->d_NumDecimals,
            Child, HSpace(0),
            End;
        page1 = VCenter(page1);
        show_alignment = FALSE;

        page2 = VGroup,
            Child, VGroup,
                MUIA_Group_Columns, 3,
                Child, Label(GCS(MSG_DisplayField_ShowTitles)),
                Child, d->d_ShowTitles,
                Child, HSpace(0),
                Child, Label(GCS(MSG_DisplayField_TabChain)),
                Child, d->d_TabChain,
                Child, HSpace(0),
                Child, Label(GCS(MSG_DisplayField_ControlTable)),
                Child, d->d_ControlTable,
                Child, HSpace(0),
                End,
            Child, HGroup,
                MUIA_Frame, MUIV_Frame_Group,
                MUIA_FrameTitle, GCS(MSG_DisplayField_OnDoubleClick),
                Child, d->d_OnDoubleClick,
                Child, RectangleObject,
                    MUIA_Rectangle_VBar, TRUE,
                    MUIA_FixWidth, 10,
                    End,
                Child, VGroup,
                    Child, VSpace(0),
                    Child, CLabel(GCS(MSG_DisplayField_DoubleClick_Trigger)),
                    Child, d->d_DoubleClickTrigger,
                    End,
                End,
            Child, VGroup,
                MUIA_Group_Columns, 2,
                Child, Label(GCS(MSG_DisplayField_URLDrop_Trigger)),
                Child, d->d_URLDropTrigger,
                Child, Label(GCS(MSG_DisplayField_SortDrop_Trigger)),
                Child, d->d_SortDropTrigger,
                End,
            End;
        page2 = VCenter(page2);

        d->d_SubTypeGroup = VGroup,
            MUIA_Frame, MUIV_Frame_Group,
	    MUIA_Group_PageMode, TRUE,
            Child, page0,
            Child, page1,
            Child, page2,
	    End;

	d->d_AutoUpdate = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_AutoUpdate));
        if (d->d_AutoUpdate) {
	    SetAttrs(d->d_AutoUpdate,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_DisplayField_AutoUpdateBH),
	        MUIA_Selected, DisplayFieldAutoUpdate(dio),
		TAG_DONE);
        }

        extrasGroup = VGroup,
            Child, HGroup,
                Child, subTypeLabel,
                Child, d->d_SubType,
                End,
            Child, d->d_SubTypeGroup,
            Child, HGroup,
                Child, Label(GCS(MSG_DisplayField_AutoUpdate)),
                Child, d->d_AutoUpdate,
                Child, HSpace(0),
                End,
            End;
    }

    etext[0] = GCS(MSG_DisplayObject_Enabled_Always);
    etext[1] = GCS(MSG_DisplayObject_Enabled_NotInitial);
    etext[2] = GCS(MSG_DisplayObject_Enabled_Computed);
    etext[3] = 0;
    d->d_Enabled = RadioObject,
        MUIA_ShortHelp, GCS(MSG_DisplayObject_EnabledBH),
        MUIA_CycleChain, TRUE,
        MUIA_Radio_Entries, etext,
        End;

    d->d_EnabledComputeStr = StringObject,
        MUIA_Frame, MUIV_Frame_String,
        MUIA_ControlChar, Shortcut(GCS(MSG_Compute)),
        MUIA_ShortHelp, GCS(MSG_DisplayObject_ComputeEnabledBH),
        MUIA_CycleChain, TRUE,
        MUIA_String_MaxLen, FIELD_TRIGGER_LEN + 1,
        MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, DisplayObjectEnabledCompute(dio),
        End;

    d->d_EnabledCompute = UserFunctionsPopup(d->d_EnabledComputeStr, pr);

    if (d->d_Enabled) {
        MUIAddCallback(d->d_Enabled, 
                MUIA_Radio_Active, MUIV_EveryTime,
                onEnabledChanged, d);
        set(d->d_Enabled, MUIA_Radio_Active, DisplayObjectEnabled(dio));
    }

    enabledGroup = HGroup,
        MUIA_Frame, MUIV_Frame_Group,
        MUIA_FrameTitle, GCS(MSG_DisplayObject_Enabled),
        Child, d->d_Enabled,
        Child, RectangleObject,
            MUIA_Rectangle_VBar, TRUE,
            MUIA_FixWidth, 10,
            End,
        Child, VGroup,
            Child, VSpace(0),
            Child, CLabel(GCS(MSG_Compute)),
            Child, d->d_EnabledCompute,
            End,
        End;

    if (show_weight) {
	d->d_Weight = SliderObject,
	    MUIA_ControlChar, Shortcut(GCS(MSG_Weight)),
	    MUIA_ShortHelp, GCS(MSG_WeightBH),
            MUIA_CycleChain, TRUE,
	    MUIA_Numeric_Min, 0,
	    MUIA_Numeric_Max, 1000,
	    MUIA_Numeric_Value, DisplayObjectWeight(dio),
	    End;
	weightLabel = Label(GCS(MSG_Weight));
    }

    if (show_fontname) {
        d->d_Fontname = StringObject,
            MUIA_Frame, MUIV_Frame_String,
	    MUIA_ControlChar, Shortcut(GCS(MSG_Font)),
            MUIA_ShortHelp, GCS(MSG_FontBH),
            MUIA_CycleChain, TRUE,
            MUIA_String_MaxLen, FONTNAME_LEN,
            MUIA_String_Format, MUIV_String_Format_Center,
            MUIA_String_Contents, DisplayObjectMUIFontname(dio),
            End;

        fontnameLabel = Label2(GCS(MSG_Font));

        d->d_FontnamePopup = PopaslObject,
	    MUIA_Popstring_String, d->d_Fontname,
	    MUIA_Popstring_Button, PopButton(MUII_PopUp),
            MUIA_Popasl_Type, ASL_FontRequest,
            ASLFO_MaxHeight, 128,
            ASLFO_Flags, FOF_DOFRONTCOLOR | FOF_DOBACKCOLOR | FOF_DOSTYLE | 
                FOF_DOMUISTYLES,
            End;
    }

    if (show_bg) {
        const char *bg = DisplayObjectBackground(dio);

        if (show_buttontype && subType != DAD_BUTTONTYPE_TEXT) {
            strcpy(d->d_BackgroundSpec.buf, IMAGESPEC_START);
	} else {
            sprintf(d->d_BackgroundSpec.buf, "%.*s", IMAGESPEC_LEN,
                (bg && bg[0])? bg : IMAGESPEC_START);
	}
        ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
        d->d_Background = PopimageObject,
            MUIA_ControlChar, Shortcut(GCS(MSG_Background)),
            MUIA_ShortHelp, GCS(MSG_BackgroundBH),
            MUIA_CycleChain, TRUE,
            MUIA_Weight, 200,
            MUIA_Imagedisplay_Spec, &d->d_BackgroundSpec,
            MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_Background,
            End;
        MUIAddCallback(d->d_Background, MUIA_Pressed, FALSE, 
                PopButtonSetCWD, pr);

        bgLabel = FreeLabel(GCS(MSG_Background));

	d->d_DefaultBackground = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_Background_Default));
        if (d->d_Background && d->d_DefaultBackground) {
            SetAttrs(d->d_DefaultBackground,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_Background_DefaultBH),
                TAG_DONE);
            DoMethod(d->d_DefaultBackground, MUIM_Notify, 
		    MUIA_Selected, MUIV_EveryTime, d->d_Background, 3, 
		    MUIM_Set, MUIA_Disabled, MUIV_TriggerValue);
            set(d->d_DefaultBackground, MUIA_Selected, !bg || !bg[0]);
        }

        bgGroup = HGroup,
            Child, HSpace(0),
            Child, Label(GCS(MSG_Background_Default)),
            Child, d->d_DefaultBackground,
            End;
        bgGroup = HGroup,
            MUIA_Weight, 75,
            Child, d->d_Background,
            Child, VCenter(bgGroup),
            End;
    }

    if (show_buttontype) {
        const char *bg = DisplayObjectBackground(dio);
        Object *page0, *page1;
        static const char *entries[3];

        if (subType != DAD_BUTTONTYPE_IMAGE) {
            strcpy(d->d_ImageSpec.buf, IMAGESPEC_START);
	} else {
            sprintf(d->d_ImageSpec.buf, "%.*s", IMAGESPEC_LEN,
                (bg && bg[0])? bg : IMAGESPEC_START);
	}

        ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
        d->d_Image = PopimageObject,
            MUIA_Weight, 100,
            MUIA_CycleChain, TRUE,
            MUIA_ShortHelp, GCS(MSG_ImageBH),
            MUIA_ControlChar, Shortcut(GCS(MSG__Image)),
            MUIA_Imagedisplay_Spec, &d->d_ImageSpec,
            MUIA_Imageadjust_Type, ExecIsMorphOS()?
                MUIV_Imageadjust_Type_Background : MUIV_Imageadjust_Type_All,
            End;
        MUIAddCallback(d->d_Image, MUIA_Pressed, FALSE, PopButtonSetCWD, pr);

        entries[0] = GCS(MSG_Image_Size_Resizeable);
        entries[1] = GCS(MSG_Image_Size_Fixed);
        entries[2] = NULL;
        d->d_ImageSize = RadioObject,
            MUIA_CycleChain, TRUE,
            MUIA_Radio_Entries, entries,
            MUIA_Radio_Active, DisplayFieldImageSize(dio),
            End;

        page0 = VGroup,
            MUIA_Group_Columns, 2,
            Child, Label(GCS(MSG_Title)),
            Child, d->d_Title,
            Child, fontnameLabel,
            Child, d->d_FontnamePopup,
            Child, bgLabel,
            Child, bgGroup,
            End;
        show_fontname = FALSE;          /* only add once */
        show_bg = FALSE;                /* only add once */

        page1 = HGroup,
            Child, FreeLabel(GCS(MSG__Image)),
            Child, d->d_Image,
            Child, VGroup,
                GroupFrame,
                MUIA_FrameTitle, GCS(MSG_Size),
                Child, VCenter(d->d_ImageSize),
                End,
            End;

        d->d_SubTypeGroup = VGroup,
	    MUIA_Frame, MUIV_Frame_Group,
	    MUIA_Group_PageMode, TRUE,
            Child, page0,
            Child, page1,
            End;
    }

#if 0
    if(show_fg)
    {
        struct Window *win = GuiAnyOpenedWindow(app);
        struct Screen *screen = win? win->WScreen : NULL;
        struct DrawInfo *dri = screen? GetScreenDrawInfo(screen) : NULL;
        static UBYTE pens[8];
        short i;

        if(dri)
        {
            for(i=0; i < 7; i++)
                pens[i] = dri->dri_Pens[i+2];
            if(dri->dri_NumPens > 9)
                pens[i] = dri->dri_Pens[i+2];
            else
                pens[i] = dri->dri_Pens[0];

            FreeScreenDrawInfo(screen, dri);

            for(i=0; i < 8; i++)
                printf("pens[%d] = %d\n", i+2, pens[i]);
        }

        fg_label = FreeLabel(GCS(MSG_DisplayField_InitialColor));
        d->d_InitialColor = ColorButtonObject,
            MUIA_CycleChain, TRUE,
            MUIA_ShortHelp, GCS(MSG_DisplayField_InitialColorBH),
            MUIA_ControlChar, Shortcut(GCS(MSG_DisplayField_InitialColor)),
            MUIA_ColorButton_Pens, NULL,
            MUIA_ColorButton_Pen, 0,
            MUIA_ColorButton_Max, 8,
            End;
    }
#endif

    if (show_stringbuttons) {
        static const char *entries[6];
	static const char *lktext[3];
        Object *top, *bottom, *popGroup, *flagsGroup;

        d->d_File = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_File));
        if (d->d_File) {
            SetAttrs(d->d_File,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_DisplayField_FileBH),
                MUIA_Selected, DisplayFieldFile(dio),
                TAG_DONE);
	    MUIAddCallback(d->d_File, MUIA_Selected, MUIV_EveryTime,
		    onFFLChanged, d);
        }

        d->d_Font = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_Font));
        if (d->d_Font) {
            SetAttrs(d->d_Font,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_DisplayField_FontBH),
                MUIA_Selected, DisplayFieldFont(dio),
                TAG_DONE);
	    MUIAddCallback(d->d_Font, MUIA_Selected, MUIV_EveryTime,
		    onFFLChanged, d);
        }

        d->d_Listview = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_Listview));
        if (d->d_Listview) {
            SetAttrs(d->d_Listview,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_DisplayField_ListviewBH),
                MUIA_Selected, DisplayFieldListview(dio),
                TAG_DONE);
	    MUIAddCallback(d->d_Listview, MUIA_Selected, MUIV_EveryTime,
		    onFFLChanged, d);
        }

        d->d_View = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_View));
        if (d->d_View) {
            SetAttrs(d->d_View,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_DisplayField_ViewBH),
                MUIA_Selected, DisplayFieldView(dio),
                TAG_DONE);
        }

	popGroup = VGroup,
	    MUIA_Group_Columns, 2,
            Child, Label(GCS(MSG_DisplayField_File)),
	    Child, d->d_File,
            Child, Label(GCS(MSG_DisplayField_Font)),
	    Child, d->d_Font,
            Child, Label(GCS(MSG_DisplayField_Listview)),
	    Child, d->d_Listview,
            Child, Label(GCS(MSG_DisplayField_View)),
	    Child, d->d_View,
	    End;

        lktext[0] = GCS(MSG_DisplayField_Labels_Static);
        lktext[1] = GCS(MSG_Computed);
        lktext[2] = 0;
        d->d_ListviewKind = CycleObject,
            MUIA_CycleChain, TRUE,
	    MUIA_ShortHelp, GCS(MSG_DisplayField_LabelsBH),
            MUIA_Cycle_Entries, lktext,
            End;

	d->d_ListviewLabels = PopEditLabelsObject,
	    ButtonFrame,
	    MUIA_Background, MUII_ButtonBack,
            MUIA_CycleChain, TRUE,
	    MUIA_ControlChar, Shortcut(GCS(MSG_DisplayField__EditLabels)),
            MUIA_ShortHelp, GCS(MSG_DisplayField_EditLabelsBH),
	    MUIA_InputMode, MUIV_InputMode_RelVerify,
	    MUIA_Text_Contents, GCS(MSG_DisplayField_EditLabels),
	    MUIA_Text_PreParse, "\33c",
	    MUIA_Text_HiChar, Shortcut(GCS(MSG_DisplayField__EditLabels)),
            MUIA_EditLabels_Labels, DisplayFieldListviewLabels(dio),
            End;

	d->d_ListviewComputeStr = StringObject,
	    MUIA_Frame, MUIV_Frame_String,
	    MUIA_CycleChain, TRUE,
	    MUIA_ControlChar,Shortcut(GCS(MSG_Compute)),
	    MUIA_ShortHelp, GCS(MSG_DisplayField_ComputeLabelsBH),
	    MUIA_String_MaxLen, FIELD_TRIGGER_LEN + 1,
	    MUIA_String_Format, MUIV_String_Format_Center,
	    MUIA_String_Contents, DisplayFieldListviewCompute(dio),
	    End;

	d->d_ListviewCompute = UserFunctionsPopup(d->d_ListviewComputeStr, pr);

	d->d_ListviewKindGroup = GroupObject,
	    MUIA_Group_PageMode, TRUE,
	    Child, d->d_ListviewLabels,
	    Child, VGroup,
	        Child, CLabel(GCS(MSG_Compute)),
		Child, d->d_ListviewCompute,
		End,
	    End;

	d->d_ListviewSettings = VGroup,
	    Child, HGroup,
	        Child, Label(GCS(MSG_DisplayField_Labels)),
		Child, d->d_ListviewKind,
		End,
	    Child, RectangleObject,
		MUIA_Rectangle_HBar, TRUE,
		End,
	    Child, d->d_ListviewKindGroup,
	    End;
	d->d_ListviewSettings = VGroup,
            GroupFrame,
            MUIA_FrameTitle, GCS(MSG_DisplayField_ListviewSettings),
            MUIA_Disabled, !DisplayFieldListview(dio),
            Child, VCenter(d->d_ListviewSettings),
            End;
	if (d->d_ListviewSettings) {
	    DoMethod(d->d_ListviewKind, MUIM_Notify, 
		    MUIA_Cycle_Active, MUIV_EveryTime,
		    d->d_ListviewKindGroup, 3, 
		    MUIM_Set, MUIA_Group_ActivePage, MUIV_TriggerValue);
	}

	popGroup = HGroup,
	    Child, VCenter(HCenter(popGroup)),
	    Child, d->d_ListviewSettings,
	    End;
        bottom = VGroup,
            GroupFrame,
            MUIA_FrameTitle, GCS(MSG_DisplayField_Buttons),
	    Child, popGroup,
	    End;

        d->d_DisplayExternal = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_DisplayPicture));
        if (d->d_DisplayExternal) {
            SetAttrs(d->d_DisplayExternal,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_DisplayField_DisplayPictureBH),
                MUIA_Selected, TRUE,
                TAG_DONE);
        }

        d->d_HideStringField = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_HideStringField));
        if (d->d_HideStringField) {
            SetAttrs(d->d_HideStringField,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_DisplayField_HideStringFieldBH),
                TAG_DONE);
	    MUIAddCallback(d->d_HideStringField, MUIA_Selected, MUIV_EveryTime,
		    onHideStringFieldChanged, d);
        }

        d->d_TitleAtStringField = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_TitleAtStringField));
        if (d->d_TitleAtStringField) {
            SetAttrs(d->d_TitleAtStringField,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_DisplayField_TitleAtStringFieldBH),
                TAG_DONE);
        }

        entries[0] = GCS(MSG_Image_Size_Resizeable);
        entries[1] = GCS(MSG_Image_Size_Fixed);
        entries[2] = GCS(MSG_Image_Size_Scrollable);
        entries[3] = GCS(MSG_Image_Size_Scaled);
        entries[4] = GCS(MSG_Image_Size_AspectScaled);
        entries[5] = 0;
        d->d_ImageSize = RadioObject,
            MUIA_Radio_Entries, entries,
            MUIA_Radio_Active, DisplayFieldImageSize(dio),
            MUIA_CycleChain, TRUE,
            End;
        externalSizeGroup = VGroup,
            GroupFrame,
            MUIA_FrameTitle, GCS(MSG_Size),
            MUIA_ShortHelp, GCS(MSG_DisplayField_SizeBH),
            Child, VCenter(HCenter(d->d_ImageSize)),
            End;

        flagsGroup = VGroup,
            MUIA_Group_Columns, 2,
            Child, Label(GCS(MSG_DisplayField_DisplayPicture)),
            Child, d->d_DisplayExternal,
            Child, Label(GCS(MSG_DisplayField_TitleAtStringField)),
            Child, d->d_TitleAtStringField,
            Child, Label(GCS(MSG_DisplayField_HideStringField)),
            Child, d->d_HideStringField,
            End;
        if (flagsGroup) {
	    DoMethod(d->d_DisplayExternal, MUIM_Notify, MUIA_Selected, FALSE,
		    d->d_HideStringField, 3, MUIM_Set, MUIA_Selected, FALSE);
	    DoMethod(d->d_DisplayExternal, MUIM_Notify, 
		    MUIA_Selected, MUIV_EveryTime,
		    d->d_DisplayExternal, 7, MUIM_MultiSet, 
		    MUIA_Disabled, MUIV_NotTriggerValue,
                    d->d_TitleAtStringField, d->d_HideStringField, 
		    externalSizeGroup, NULL);

            set(d->d_TitleAtStringField,
                MUIA_Selected, DisplayFieldTitleAtStringField(dio));
            set(d->d_HideStringField,
                MUIA_Selected, DisplayFieldHideStringField(dio));
            set(d->d_DisplayExternal,
                MUIA_Selected, DisplayFieldDisplayExternal(dio));
        }

        top = HGroup,
            Child, VCenter(HCenter(flagsGroup)),
            Child, externalSizeGroup,
            End;
        top = VGroup,
            GroupFrame,
            MUIA_FrameTitle, GCS(MSG_DisplayField_Picture),
            Child, top,
            End;

        extrasGroup = VGroup,
            Child, top,
            Child, bottom,
            End;
    } else if (show_reference) {
        Object *buttons;
        static const char *rectext[4];

        d->d_Show = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_Show));
        if (d->d_Show) {
            SetAttrs(d->d_Show,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_DisplayField_ShowBH),
        	MUIA_Selected, DisplayFieldShow(dio),
                TAG_DONE);
	    MUIAddCallback(d->d_Show, MUIA_Selected, MUIV_EveryTime,
		    onShowChanged, d);
        }

        d->d_AutoShow = MUI_MakeObject(MUIO_Checkmark, (IPTR)GCS(MSG_AutoShow));
        if (d->d_AutoShow) {
            SetAttrs(d->d_AutoShow,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_DisplayField_AutoShowBH),
        	MUIA_Selected, DisplayFieldAutoShow(dio),
                TAG_DONE);
        }

        d->d_Filter = MUI_MakeObject(MUIO_Checkmark,
		(IPTR)GCS(MSG_DisplayField_Filter));
        if (d->d_Filter) {
            SetAttrs(d->d_Filter,
                MUIA_CycleChain, TRUE,
                MUIA_ShortHelp, GCS(MSG_DisplayField_FilterBH),
        	MUIA_Selected, DisplayFieldFilter(dio),
                TAG_DONE);
        }

	buttons = VGroup,
	    MUIA_Group_Columns, 2,
            Child, d->d_Show,
            Child, LLabel(GCS(MSG_DisplayField_Show)),
            Child, d->d_AutoShow,
            Child, LLabel(GCS(MSG_AutoShow)),
            Child, d->d_Filter,
            Child, LLabel(GCS(MSG_DisplayField_Filter)),
            End;
	buttons = HCenter(buttons);
	buttons = VCenter(buttons);
        buttons = VGroup,
            GroupFrame,
            MUIA_FrameTitle, GCS(MSG_DisplayField_Buttons),
	    MUIA_Weight, 3,
	    Child, buttons,
	    End;

        d->d_RecordDescriptionSpec = CreateEditRecordDescriptionSpec(
                FieldReferenceTable(field), 
                DisplayFieldRecordDescriptionSpec(dio, FALSE), TRUE);

	d->d_RecordDescriptionSpecGroup = VGroup,
	    GroupFrame,
	    MUIA_FrameTitle, GCS(MSG_Display),
            Child, d->d_RecordDescriptionSpec,
	    End;

        rectext[0] = GCS(MSG_DisplayField_Records_All);
        rectext[1] = GCS(MSG_DisplayField_Records_MatchFilter);
        rectext[2] = GCS(MSG_Computed);
	rectext[3] = 0;
	d->d_RecordsKind = CycleObject,
            MUIA_CycleChain, TRUE,
            MUIA_ControlChar, Shortcut(GCS(MSG_Records)),
            MUIA_ShortHelp, GCS(MSG_DisplayField_RecordsBH),
            MUIA_Cycle_Entries, rectext,
            End;

	d->d_ComputeRecordsStr = StringObject,
	    MUIA_CycleChain, TRUE,
	    MUIA_Frame, MUIV_Frame_String,
	    MUIA_ControlChar, Shortcut(GCS(MSG_Compute)),
	    MUIA_ShortHelp, GCS(MSG_DisplayField_ComputeRecordsBH),
	    MUIA_String_MaxLen, FIELD_TRIGGER_LEN + 1,
	    MUIA_String_Format, MUIV_String_Format_Center,
	    MUIA_String_Contents, DisplayFieldRecordsCompute(dio),
	    End;

	d->d_ComputeRecords = UserFunctionsPopup(d->d_ComputeRecordsStr, pr);

	d->d_ComputeRecordsGroup = VGroup,
	    GroupFrame,
	    MUIA_FrameTitle, GCS(MSG_Compute),
	    MUIA_Disabled, TRUE,
	    Child, d->d_ComputeRecords,
	    End;
        if (d->d_RecordsKind && d->d_ComputeRecordsGroup) {
	    MUIAddCallback(d->d_RecordsKind, MUIA_Cycle_Active, MUIV_EveryTime,
		    onRecordsChanged, d);
	    set(d->d_RecordsKind, MUIA_Cycle_Active, 
		    DisplayFieldRecordsKind(dio));
        }

        d->d_InitialRecord = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_InitialRecord));
        if (d->d_InitialRecord) {
	    SetAttrs(d->d_InitialRecord,
	    	    MUIA_CycleChain, TRUE,
	    	    MUIA_ShortHelp, GCS(MSG_DisplayField_InitialRecordBH),
	    	    MUIA_Selected, DisplayFieldInitialRecord(dio),
	    	    TAG_DONE);
        }

	d->d_CurrentRecord = MUI_MakeObject(MUIO_Checkmark, 
		(IPTR)GCS(MSG_DisplayField_CurrentRecord));
        if (d->d_CurrentRecord) {
	    SetAttrs(d->d_CurrentRecord,
	    	    MUIA_CycleChain, TRUE,
	    	    MUIA_ShortHelp, GCS(MSG_DisplayField_CurrentRecordBH),
	    	    MUIA_Selected, DisplayFieldCurrentRecord(dio),
	    	    TAG_DONE);
        }

	d->d_ReferencePopupGroup = VGroup,
	    Child, HGroup,
	        Child, Label(GCS(MSG_Records)),
		Child, d->d_RecordsKind,
		End,
	    Child, d->d_ComputeRecordsGroup,
	    Child, VGroup,
                MUIA_Group_Columns, 3,
		Child, d->d_InitialRecord,
		Child, LLabel(GCS(MSG_DisplayField_InitialRecord)),
		Child, HSpace(0),
		Child, d->d_CurrentRecord,
		Child, LLabel(GCS(MSG_DisplayField_CurrentRecord)),
		Child, HSpace(0),
		End,
	    End;
	d->d_ReferencePopupGroup = VCenter(d->d_ReferencePopupGroup);
        d->d_ReferencePopupGroup = VGroup,
            GroupFrame,
            MUIA_FrameTitle, GCS(MSG_Popup),
	    Child, d->d_ReferencePopupGroup,
            MUIA_Weight, 5,
	    End;

        extrasGroup = HGroup,
            Child, d->d_RecordDescriptionSpecGroup,
            Child, VGroup,
                Child, d->d_ReferencePopupGroup,
                Child, buttons,
                End,
            End;
    }

    if (show_alignment) {
        Object *right = NULL;

        if (show_numdecimals) {
            right = HGroup,
                Child, numdecimalsLabel,
                Child, d->d_NumDecimals,
                End;
        } else if (show_calendar) {
            right = HGroup,
                Child, HSpace(0),
                Child, calendarLabel,
                Child, d->d_Calendar,
                End;
        } else if (show_format) {
            right = HGroup,
                Child, formatLabel,
                Child, d->d_Format,
                End;
        } else {
            right = HSpace(0);
        }
        alignmentGroup = HGroup,
            Child, d->d_Alignment,
            Child, right,
            End;
    }

    if (show_buttontype) {
        topGroup = VGroup,
            Child, HGroup,
                Child, subTypeLabel,
                Child, d->d_SubType,
                End,
            Child, d->d_SubTypeGroup,
            End;
    } else if (show_choicetype) {
        topGroup = VGroup,
            Child, HGroup,
                Child, subTypeLabel,
                Child, d->d_SubType,
                End,
            Child, VGroup,
                MUIA_Frame, MUIV_Frame_Group,
                MUIA_FrameTitle, GCS(MSG_DisplayField_Title),
                Child, HGroup,
                    Child, Label(GCS(MSG_Title)),
                    Child, d->d_Title,
                    End,
                Child, d->d_SubTypeGroup,
                End,
            End;
    } else {
        topGroup = VGroup,
            GroupFrame,
            MUIA_FrameTitle, GCS(MSG_DisplayField_Title),
            MUIA_Group_Columns, 2,
            Child, Label(GCS(MSG_Title)),
            Child, d->d_Title,
            Child, titlePositionLabel,
            Child, HGroup,
                Child, d->d_TitlePosition,
                Child, HSpace(10),
                Child, titleFontnameLabel,
                Child, d->d_TitleFontnamePopup,
                End,
            End;
    }
   
    editor = MUINewEditor(pr, Shortcut(GCS(MSG_DisplayField_BubbleHelp)),
            FALSE, FALSE, (const char *)MUII_TextBack, NULL, &d->d_BubbleHelp);
    GuiSetEditorText(d->d_BubbleHelp, DisplayFieldBubbleHelp(dio));

    fieldGroup = VGroup,
        GroupFrame,
        MUIA_FrameTitle, GCS(MSG_Field),
        MUIA_Group_Columns, 2,
        show_shortcut? Child : TAG_IGNORE, shortcutLabel,
        show_shortcut? Child : TAG_IGNORE, shortcutGroup,
        show_alignment? Child : TAG_IGNORE, alignmentLabel,
        show_alignment? Child : TAG_IGNORE, alignmentGroup,
        show_weight? Child : TAG_IGNORE, weightLabel,
        show_weight? Child : TAG_IGNORE, d->d_Weight,
        show_fontname? Child : TAG_IGNORE, fontnameLabel,
        show_fontname? Child : TAG_IGNORE, d->d_FontnamePopup,
        show_bg? Child : TAG_IGNORE, bgLabel,
        show_bg? Child : TAG_IGNORE, bgGroup,
        Child, FreeLabel(GCS(MSG_DisplayField_BubbleHelp)),
        Child, editor,
        End;

    attrsGroup = VGroup,
        Child, topGroup,
        Child, fieldGroup,
        Child, enabledGroup,
        End;

    if (extras) {
        static const char *titles[3];

        titles[0] = GCS(MSG_General);
        titles[1] = GCS(MSG_DisplayField_Extras);
        titles[2] = NULL;

        mainGroup = RegisterGroup(titles),
	   Child, attrsGroup,
	   Child, extrasGroup,
	   End;
    } else {
        mainGroup = attrsGroup;
    }

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
        MUIA_Window_Title, GCS(MSG_DisplayField),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, winid,
        MUIA_Window_NoMenus, TRUE,
        MUIA_HelpNode, "Field object editor",
        MUIA_UserData, d,
	WindowContents, VGroup,
            Child, mainGroup,
	    Child, HGroup,
		MUIA_Group_SameWidth, TRUE,
		Child, d->d_Ok = SimpleButton(GCS(MSG_Ok)),
		Child, d->d_Cancel = SimpleButton(GCS(MSG_Cancel)),
                End,
            End,
	End;

    if (obj == NULL) {
        PoolFreeMem(memPool, d, sizeof(struct EditDisplayFieldData));
        return;
    }

    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    DoMethod(d->d_Title, MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
        obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    if (d->d_TitleFontname) {
        DoMethod(d->d_TitleFontname,
            MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    }
    if (d->d_Shortcut) {
	DoMethod(d->d_Shortcut, MUIM_Notify, 
		MUIA_String_Acknowledge, MUIV_EveryTime, obj, 3,
		MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    }
    if (d->d_ReadOnly) {
	MUIAddCallback(d->d_ReadOnly, MUIA_Selected, MUIV_EveryTime,
		onReadOnlyChanged, d);
	set(d->d_ReadOnly, MUIA_Selected, readOnly);
    }
    if (d->d_NumDecimals) {
        DoMethod(d->d_NumDecimals, MUIM_Notify, 
                MUIA_String_Acknowledge, MUIV_EveryTime,
                obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    }
    if (d->d_EnabledComputeStr) {
        DoMethod(d->d_EnabledComputeStr,
                MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
                obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    }
    if (d->d_Fontname) {
        DoMethod(d->d_Fontname,
            MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    }
    if (d->d_DoubleClickTriggerStr) {
        DoMethod(d->d_DoubleClickTriggerStr,
            MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    }
    if (d->d_URLDropTriggerStr) {
        DoMethod(d->d_URLDropTriggerStr,
            MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    }
    if (d->d_SortDropTriggerStr) {
        DoMethod(d->d_SortDropTriggerStr,
            MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    }

    if (d->d_SubType && d->d_SubTypeGroup) {
        DoMethod(d->d_SubType, MUIM_Notify,
                MUIA_Cycle_Active, MUIV_EveryTime, d->d_SubTypeGroup, 3, 
                MUIM_Set, MUIA_Group_ActivePage, MUIV_TriggerValue);
        set(d->d_SubTypeGroup, MUIA_Group_ActivePage, subType);
    }
    if (d->d_ListviewKind && lvcompute && lvcompute[0]) {
	set(d->d_ListviewKind, MUIA_Cycle_Active, 1);
    }
    if (d->d_ListviewComputeStr) {
	DoMethod(d->d_ListviewComputeStr, 
		MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
		obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    }
    EditRecordDescriptionSpecAddCallbacks(d->d_RecordDescriptionSpec, d->d_Ok);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE,
        obj, closeEditDisplayField, FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE,
        obj, closeEditDisplayField, FALSE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE,
        obj, closeEditDisplayField, TRUE);
    updateFocusWidgets(d);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if (!GuiOpenWindow(obj)) {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeEditDisplayField(app, obj);
        return;
    }

    set(obj, MUIA_Window_ActiveObject, d->d_Title);
}

#endif

