/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** displaytable_mui.c
**
** Creation date: 10.08.2012
** Last modified: 12.07.2020
*/

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "structureeditor_mui.h"
#include "recorddescriptionspec_mui.h"
#include "extclasses/ExtApplication_mcc.h"
#include <libraries/asl.h>
#include "display.h"
#include "project.h"
#include "table.h"
#include "requesters.h"
#include "memory.h"
#include "error.h"
#include "util.h"
#include "appstrings_cat.h"

#define IMAGESPEC_LEN (int)(sizeof(struct MUI_ImageSpec) - 1)
#define IMAGESPEC_START "0:128"

struct EditDisplayTableData
{
    struct Project *d_Project;
    struct DisplayObject *d_DisplayTable;
    GuiObject d_StructureEditor;
    GuiObject d_Weight, d_Background, d_DefaultBackground, d_HasPanel;
    GuiObject d_PanelTitle, d_PanelFontname;
    GuiObject d_PanelBackground, d_PanelDefaultBackground;
    GuiObject d_PanelNumAll, d_PanelFilter, d_PanelArrows;
    GuiObject d_RecordDescriptionSpec;
    GuiObject d_Ok, d_Cancel;
    struct MUI_ImageSpec d_BackgroundSpec;
    struct MUI_ImageSpec d_PanelBackgroundSpec;
};

static void disposeEditDisplayTable(Object *app, Object *obj)
{
    struct EditDisplayTableData *d = pget(obj, MUIA_UserData);

    EditRecordDescriptionSpecCleanup(d->d_RecordDescriptionSpec, d->d_Project);
    PoolFreeMem(ProjectMemPool(d->d_Project), d, 
            sizeof(struct EditDisplayTableData));
    MUI_AddDisposeObject(app, obj);
}

static void HOOK_PROTO(closeEditDisplayTable, Object *obj, ulong *ok)
{
    struct EditDisplayTableData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);

    if (*ok) {
        struct RecordDescriptionSpec *rds = NULL;
        struct MemPool *memPool = ProjectMemPool(d->d_Project);
        struct Table *tab = DisplayObjectTable(d->d_DisplayTable);
        long error = OK;
        struct DisplayTableData dtd;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        dtd.dtd_Title = pget(d->d_PanelTitle, MUIA_String_Contents);
        dtd.dtd_Weight = nget(d->d_Weight, MUIA_Numeric_Value);
        dtd.dtd_Fontname = pget(d->d_PanelFontname, MUIA_String_Contents);
        dtd.dtd_Background = nget(d->d_DefaultBackground, MUIA_Selected)?
            NULL : pget(d->d_Background, MUIA_Imagedisplay_Spec);
        dtd.dtd_HasPanel = nget(d->d_HasPanel, MUIA_Selected);
        dtd.dtd_PanelBackground = 
            nget(d->d_PanelDefaultBackground, MUIA_Selected)?
            NULL : pget(d->d_PanelBackground, MUIA_Imagedisplay_Spec);
        dtd.dtd_PanelNumAll = nget(d->d_PanelNumAll, MUIA_Selected);
        dtd.dtd_PanelFilter = nget(d->d_PanelFilter, MUIA_Selected);
        dtd.dtd_PanelArrows = nget(d->d_PanelArrows, MUIA_Selected);

        rds = EditRecordDescriptionSpecGetSpec(
                d->d_RecordDescriptionSpec, memPool);
        dtd.dtd_RecordDescriptionSpec = rds;
        if (rds == NULL) {
            error = ERR_MEM;
        }

        GuiSetSleep(TRUE);
        if (error == OK) {
            error = DisplayChangeTable(tab, &dtd);
            if (error == OK) {
                ProjectChange(pr, TRUE);
            }
        }
        FreeRecordDescriptionSpec(memPool, rds);
        GuiSetSleep(FALSE);

        if (error) {
            Ask2(obj, GCS(MSG_ErrorMessage), AT_ERROR, 
		    GCS(MSG_StarOk), GES(error));
            return;
        }

        DisplayTableFinish(d->d_StructureEditor, d->d_DisplayTable);
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposeEditDisplayTable(app, obj);
}

void GuiDisplayTable(Object *se, struct DisplayObject *dio)
{
    struct Table *tab = DisplayObjectTable(dio);
    struct Project *pr = TableProject(tab);
    Object *app = GuiApplication;
    struct MemPool *memPool = ProjectMemPool(pr);
    bool gadgets = !TableExactlyOneRecord(tab);
    bool hasPanel = DisplayTableHasPanel(dio);
    const char *bg = DisplayObjectBackground(dio);
    const char *panelbg = DisplayTablePanelBackground(dio);
    struct EditDisplayTableData *d;
    GuiObject back, generalGroup;
    GuiObject panelBack, panelBasic, panelFields, panelGroup;
    GuiObject top, buttons, root, obj;
    static const char *titles[4];

    if (app == NULL || dio == NULL) {
	return;
    }

    d = PoolAllocMemZero(memPool, sizeof(struct EditDisplayTableData));
    if (d == NULL) {
        return;
    }
    d->d_Project = pr;
    d->d_DisplayTable = dio;
    d->d_StructureEditor = se;

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */

    // general
    d->d_Weight = SliderObject,
        MUIA_CycleChain, TRUE,
	MUIA_ShortHelp, GCS(MSG_WeightBH),
	MUIA_Frame, MUIV_Frame_Slider,
	MUIA_ControlChar, Shortcut(GCS(MSG_Weight)),
	MUIA_Numeric_Min, 0,
	MUIA_Numeric_Max, 1000,
        MUIA_Numeric_Value, DisplayObjectWeight(dio),
	End;

    CodesetStrNBytesCopy(d->d_BackgroundSpec.buf, 
	    (bg && bg[0])? bg : IMAGESPEC_START, IMAGESPEC_LEN);
    d->d_Background = PopimageObject,
        MUIA_Weight, 200,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_BackgroundBH),
        MUIA_ControlChar, Shortcut(GCS(MSG_Background)),
        MUIA_Imagedisplay_Spec, &d->d_BackgroundSpec,
        MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_Background,
        End;
    MUIAddCallback(d->d_Background, MUIA_Pressed, FALSE, PopButtonSetCWD, pr);

    d->d_DefaultBackground = MUI_MakeObject(MUIO_Checkmark, 
	    (IPTR)GCS(MSG_Background_Default));

    back = HGroup,
        Child, d->d_Background,
        Child, HSpace(0),
        Child, VGroup,
            Child, VSpace(0),
            Child, HGroup,
                Child, Label1(GCS(MSG_Background_Default)),
                Child, d->d_DefaultBackground,
                End,
            Child, VSpace(0),
            End,
        End;

    d->d_HasPanel = MUI_MakeObject(MUIO_Checkmark, 
            (IPTR)GCS(MSG_DisplayTable_HasPanel));

    generalGroup = VGroup,
        MUIA_Background, MUII_GroupBack,
	MUIA_Group_Columns, 2,
	Child, Label2(GCS(MSG_Weight)),
	Child, d->d_Weight,
        Child, FreeLabel(GCS(MSG_Background)),
        Child, back,
	Child, Label1(GCS(MSG_DisplayTable_HasPanel)),
        Child, HGroup, 
            Child, d->d_HasPanel,
            Child, HSpace(0),
            End,
        End;

    // panel
    d->d_PanelTitle = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayTable_Panel_TitleBH),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Title)),
	MUIA_String_MaxLen, TITLE_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, DisplayObjectTitle(dio),
	End;

    d->d_PanelFontname = StringObject,
        MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Font)),
        MUIA_String_MaxLen, FONTNAME_LEN,
        MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, DisplayObjectMUIFontname(dio),
        MUIA_ShortHelp, GCS(MSG_FontBH),
        MUIA_CycleChain, TRUE,
        End;

    CodesetStrNBytesCopy(d->d_PanelBackgroundSpec.buf, 
	    (panelbg && panelbg[0])? panelbg : IMAGESPEC_START, IMAGESPEC_LEN);
    d->d_PanelBackground = PopimageObject,
        MUIA_Weight, 200,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_BackgroundBH),
        MUIA_ControlChar, Shortcut(GCS(MSG_Background)),
        MUIA_Imagedisplay_Spec, &d->d_PanelBackgroundSpec,
        MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_Background,
        End;
    MUIAddCallback(d->d_PanelBackground, MUIA_Pressed, FALSE, 
            PopButtonSetCWD, pr);

    d->d_PanelDefaultBackground = MUI_MakeObject(MUIO_Checkmark, 
	    (IPTR)GCS(MSG_Background_Default));

    panelBack = HGroup,
        Child, d->d_PanelBackground,
        Child, HSpace(0),
        Child, VGroup,
            Child, VSpace(0),
            Child, HGroup,
                Child, Label1(GCS(MSG_Background_Default)),
                Child, d->d_PanelDefaultBackground,
                End,
            Child, VSpace(0),
            End,
        End;

    panelBasic = VGroup,
	MUIA_Group_Columns, 2,
	Child, Label2(GCS(MSG_Title)),
	Child, d->d_PanelTitle,
        Child, Label2(GCS(MSG_Font)),
        Child, PopaslObject,
	    MUIA_Popstring_String, d->d_PanelFontname,
	    MUIA_Popstring_Button, PopButton(MUII_PopUp),
            MUIA_Popasl_Type, ASL_FontRequest,
            ASLFO_MaxHeight, 128,
            ASLFO_Flags, FOF_DOFRONTCOLOR | FOF_DOBACKCOLOR | FOF_DOSTYLE | 
                FOF_DOMUISTYLES,
            End,
        Child, FreeLabel(GCS(MSG_Background)),
        Child, panelBack,
	End;

    d->d_PanelNumAll = MUI_MakeObject(MUIO_Checkmark, 
            (IPTR)GCS(MSG_DisplayTable_Panel_NumAll));
    d->d_PanelFilter = MUI_MakeObject(MUIO_Checkmark, (IPTR)GCS(MSG_Filter));
    d->d_PanelArrows = MUI_MakeObject(MUIO_Checkmark, 
            (IPTR)GCS(MSG_DisplayTable_Panel_Arrows));

    panelFields = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_DisplayTable_Panel_Gadgets),
        MUIA_Background, MUII_GroupBack,
	MUIA_Group_Columns, 4,
	Child, HSpace(0),
	Child, Label1(GCS(MSG_DisplayTable_Panel_NumAll)),
	Child, d->d_PanelNumAll,
	Child, HSpace(0),
	Child, HSpace(0),
	Child, Label1(GCS(MSG_Filter)),
	Child, d->d_PanelFilter,
	Child, HSpace(0),
	Child, HSpace(0),
	Child, Label1(GCS(MSG_DisplayTable_Panel_Arrows)),
	Child, d->d_PanelArrows,
	Child, HSpace(0),
	End;

    panelGroup = GroupObject,
        MUIA_Background, MUII_GroupBack,
        MUIA_Disabled, TRUE,
	Child, panelBasic,
	Child, panelFields,
	End;

    d->d_RecordDescriptionSpec = CreateEditRecordDescriptionSpec(tab,
            DisplayTableRecordDescriptionSpec(dio), FALSE);

    titles[0] = GCS(MSG_General);
    titles[1] = GCS(MSG_Panel);
    titles[2] = GCS(MSG_Record);
    titles[3] = NULL;

    top = RegisterGroup(titles),
        Child, generalGroup,
        Child, panelGroup,
        Child, d->d_RecordDescriptionSpec,
        End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, d->d_Ok,
	Child, d->d_Cancel,
	End;

    root = VGroup,
	Child, top,
	Child, buttons,
	End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_Title, GCS(MSG_DisplayTable),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'P','N','L'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_HelpNode, "Table object editor",
        MUIA_UserData, d,
	WindowContents, root,
	End;

    if (obj == NULL) {
        PoolFreeMem(memPool, d, sizeof(struct EditDisplayTableData));
	return;
    }

    DoMethod(d->d_DefaultBackground, MUIM_Notify, 
        MUIA_Selected, MUIV_EveryTime, d->d_Background, 
        3, MUIM_Set, MUIA_Disabled, MUIV_TriggerValue);
    SetAttrs(d->d_DefaultBackground,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_Background_DefaultBH),
        MUIA_Selected, !bg || !bg[0],
        TAG_DONE);
    DoMethod(d->d_HasPanel, MUIM_Notify, MUIA_Selected, MUIV_EveryTime,
        panelGroup, 3, MUIM_Set, MUIA_Disabled, MUIV_NotTriggerValue);
    SetAttrs(d->d_HasPanel,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayTable_HasPanelBH),
        MUIA_Selected, hasPanel,
        TAG_DONE);
    DoMethod(d->d_PanelDefaultBackground, MUIM_Notify, 
        MUIA_Selected, MUIV_EveryTime, d->d_PanelBackground, 
        3, MUIM_Set, MUIA_Disabled, MUIV_TriggerValue);
    SetAttrs(d->d_PanelDefaultBackground,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_Background_DefaultBH),
        MUIA_Selected, !panelbg || !panelbg[0],
        TAG_DONE);
    SetAttrs(d->d_PanelNumAll,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayTable_Panel_NumAllBH),
        MUIA_Selected, DisplayTablePanelNumAll(dio) && gadgets,
        MUIA_Disabled, !gadgets,
        TAG_DONE);
    SetAttrs(d->d_PanelFilter,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayTable_Panel_FilterBH),
        MUIA_Selected, DisplayTablePanelFilter(dio) && gadgets,
        MUIA_Disabled, !gadgets,
        TAG_DONE);
    SetAttrs(d->d_PanelArrows,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_DisplayTable_Panel_ArrowsBH),
        MUIA_Selected, DisplayTablePanelArrows(dio) && gadgets,
        MUIA_Disabled, !gadgets,
        TAG_DONE);
    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    DoMethod(d->d_PanelTitle,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    DoMethod(d->d_PanelFontname,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    EditRecordDescriptionSpecAddCallbacks(d->d_RecordDescriptionSpec, d->d_Ok);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE,
        obj, closeEditDisplayTable, FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE,
        obj, closeEditDisplayTable, FALSE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE,
        obj, closeEditDisplayTable, TRUE);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if (!GuiOpenWindow(obj)) {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeEditDisplayTable(app, obj);
        return;
    }
}

#endif

