/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** DDList.c
**
** Creation date: 05.05.1996
** Last modified: 03.10.2010
**
** DDList.mcc is a subclass of List class.
** It adds some new Hook functions and the availability to
** "hear" Drop actions.
**
** Under AROS and AmigaOS4 it's a subclass of NList (if USE_NLIST is defined)
** because it keeps crashing when using the List class.
*/

/* Includes */
#include "DDList_mcc.h"
#include "system.h"
#include "appstrings_cat.h"
#include <proto/utility.h>

#define SUPERCLASS MUIC_List

static struct MUI_CustomClass *DDList_mcc = NULL;    /* for MUI */
struct IClass *DDListClass = NULL;                   /* for Intuition */

struct Data
{
    IPTR (*d_DragQuery)(Object *, struct MUIP_DragDrop *, void *);
    IPTR (*d_DragDrop)(Object *, struct MUIP_DragDrop *, void *);
    void *d_DragData;
    BOOL d_Dropped;
    bool (*d_KeyFunc)(Object *, char *, short, ushort, void *);
    void *d_KeyData;
    struct MUI_EventHandlerNode ehnode;
    struct MemPool *d_MemPool;
};

static void setattrs(struct Data *data, struct TagItem *tags)
{
    struct TagItem *tag;

    while((tag = NextTagItem(&tags)) != NULL)
    {
	switch (tag->ti_Tag)
	{
            case MUIA_DDList_DragQuery:
                data->d_DragQuery = (IPTR (*)(Object *, struct MUIP_DragDrop *, void *))tag->ti_Data;
                break;

            case MUIA_DDList_DragDrop:
                data->d_DragDrop = (IPTR (*)(Object *, struct MUIP_DragDrop *, void *))tag->ti_Data;
                break;

            case MUIA_DDList_DragData:
                data->d_DragData = (void *)tag->ti_Data;
                break;

            case MUIA_KeyFunc:
                data->d_KeyFunc = (bool (*)(Object *, char *, short, ushort, void *))tag->ti_Data;
                break;

            case MUIA_KeyData:
                data->d_KeyData = (void *)tag->ti_Data;
                break;
	}
    }
}

static IPTR mNew(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data;

    obj = (Object *)DoSuperMethodA(cl, obj, (Msg)msg);
    if(obj == NULL)
	return(0);

    data = INST_DATA(cl, obj);
    memset(data, 0, sizeof(struct Data));

    setattrs(data, msg->ops_AttrList);

    return (IPTR)obj;
}

static IPTR mDispose(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data = INST_DATA(cl, obj);

    if (data->d_MemPool) {
	DeletePool(data->d_MemPool);
        data->d_MemPool = NULL;
    }
    return DoSuperMethodA(cl, obj, (Msg)msg);
}

static IPTR mSet(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data = INST_DATA(cl, obj);

    setattrs(data, msg->ops_AttrList);
    return(DoSuperMethodA(cl, obj, (Msg)msg));
}

static IPTR mGet(struct IClass *cl, Object *obj, struct opGet *msg)
{
    struct Data *data = INST_DATA(cl, obj);

    switch (msg->opg_AttrID)
    {
        case MUIA_DDList_Dropped:
            *msg->opg_Storage = data->d_Dropped;
            return TRUE;
	case MUIA_DDList_MemPool:
	    if (data->d_MemPool == NULL) {
		data->d_MemPool = CreatePool(0, 4096, 1024);
	    }
            *msg->opg_Storage = (IPTR)data->d_MemPool;
            return TRUE;
    }
    return DoSuperMethodA(cl, obj, (Msg)msg);
}

static IPTR mSetup(struct IClass *cl, Object *obj, Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);

    if(!DoSuperMethodA(cl, obj, msg))
        return(FALSE);

    data->ehnode.ehn_Object = obj;
    data->ehnode.ehn_Class  = cl;
    data->ehnode.ehn_Events = IDCMP_RAWKEY;
    if(data->d_KeyFunc)
        DoMethod(_win(obj), MUIM_Window_AddEventHandler, &data->ehnode);

    return(TRUE);
}

static IPTR mCleanup(struct IClass *cl, Object *obj, Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);

    if(data->d_KeyFunc)
        DoMethod(_win(obj), MUIM_Window_RemEventHandler, &data->ehnode);
    return(DoSuperMethodA(cl, obj, msg));
}

static IPTR mHandleEvent(struct IClass *cl, Object *obj, 
	struct MUIP_HandleEvent *msg)
{
    struct Data *d = INST_DATA(cl, obj);
    struct IntuiMessage *imsg = msg->imsg;

    if (d->d_KeyFunc && imsg && 
	    imsg->Class == IDCMP_RAWKEY && !(imsg->Code & IECODE_UP_PREFIX))
    {
	char buffer[10];
    	short l = MUIRawKeyConvert(imsg, buffer, 10);

	if (d->d_KeyFunc(obj, buffer, l, imsg->Qualifier, d->d_KeyData))
	    return MUI_EventHandlerRC_Eat;
    }
    return 0;
}

/* taken from MUI's drag'n drop example */

static IPTR dragQuery(struct IClass *cl, Object *obj, struct MUIP_DragDrop *msg)
{
    struct Data *data = INST_DATA(cl, obj);

    if(msg->obj == obj)
    {
	/*
	** If somebody tried to drag ourselves onto ourselves, we let our superclass
	** (the list class) handle the necessary actions. Depending on the state of
	** its MUIA_List_DragSortable attribute, it will either accept or refuse to become
	** the destination object.
	*/

	return(DoSuperMethodA(cl, obj, (Msg)msg));
    }
    else if(data->d_DragQuery)
    {
	/* Ask MUIA_DDList_DragQuery function */

        return(data->d_DragQuery(obj, msg, data->d_DragData));
    }

    /*
    ** Otherwise, someone tried to feed us with something we don't like
    ** very much. Just refuse it.
    */

    return(MUIV_DragQuery_Refuse);
}

IPTR dragDrop(struct IClass *cl, Object *obj, struct MUIP_DragDrop *msg)
{
    struct Data *data = INST_DATA(cl, obj);
    IPTR rc = 0;

    if(msg->obj==obj)
    {
	/*
	** user wants to move entries within our object, these kinds of actions
	** can get quite complicated, but fortunately, everything is automatically
	** handled by list class. We just need to pass through the method.
	*/

	rc = DoSuperMethodA(cl, obj, (Msg)msg);
    }
    else if(data->d_DragDrop)
    {
	/*
	** we can be sure that msg->obj is our predefined source object
	** since we wouldnt have accepted the MUIM_DragQuery and wouldnt
	** have become an active destination object otherwise.
	*/

        rc = data->d_DragDrop(obj, msg, data->d_DragData);
    }

    set(obj, MUIA_DDList_Dropped, TRUE);

    return(rc);
}

DISPATCHER_PROTO
{
    switch (msg->MethodID)
    {
	case OM_NEW:		return(mNew(cl, obj, (APTR)msg));
	case OM_DISPOSE:	return(mDispose(cl, obj, (APTR)msg));
	case OM_SET:            return(mSet(cl, obj, (APTR)msg));
	case OM_GET:            return(mGet(cl, obj, (APTR)msg));
        case MUIM_Setup:        return(mSetup(cl, obj, (APTR)msg));
        case MUIM_Cleanup:      return(mCleanup(cl, obj, (APTR)msg));
        case MUIM_HandleEvent:  return(mHandleEvent(cl, obj, (APTR)msg));
	case MUIM_DragQuery:    return(dragQuery(cl, obj, (APTR)msg));
	case MUIM_DragDrop :    return(dragDrop (cl, obj, (APTR)msg));
    }

    return(DoSuperMethodA(cl, obj, msg));
}

void DDListCleanup(void)
{
    if(DDList_mcc != NULL)
        MUI_DeleteCustomClass(DDList_mcc);
    DDList_mcc = NULL;
    DDListClass = NULL;
}

BOOL DDListInit(void)
{
    DDList_mcc = MUI_CreateCustomClass(NULL, (STRPTR)SUPERCLASS, NULL,
        sizeof(struct Data), Dispatcher);
    if(DDList_mcc == NULL)
    {
        DDListCleanup();
        return(FALSE);
    }

    DDListClass = DDList_mcc->mcc_Class;
    DISPATCHER_INIT(DDListClass);

    return(TRUE);
}

