/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** DataTypes_mcc.c
**
** Creation date: 13.02.1997
** Last modified: 11.07.2020
**
** DataTypes.mcc is a subclass of Area class.
** It displays an external datatypes object.
*/

/* Includes */
#include <libraries/iffparse.h>         /* defines MAKE_ID */
#include "system.h"
#include "DataTypes_mcc.h"
#include "graphics/scale.h"
#include <intuition/icclass.h>
#include <clib/alib_protos.h>
#include <proto/graphics.h>
#include <proto/utility.h>
#include <datatypes/pictureclass.h>
#include <proto/datatypes.h>

struct Library *DataTypesBase = NULL;

/*
** Function for loading a datatypes image from a filename
*/
Object *LoadDTPicture(const char *filename)
{
    struct Process *myproc;
    APTR oldwindowptr;
    Object *object;

    if (DataTypesBase == NULL) {
        return NULL;                    /* library not present */
    }

    /* tell DOS not to bother us with requesters */
    myproc = (struct Process *)FindTask(NULL);
    oldwindowptr = myproc->pr_WindowPtr;
    myproc->pr_WindowPtr = (APTR)-1;

    /* create the datatypes object */
    object = (Object *)NewDTObject((char *)filename,
        DTA_GroupID, GID_PICTURE,
        PDTA_DestMode, PMODE_V43,
        OBP_Precision, PRECISION_EXACT,
        TAG_DONE);

    myproc->pr_WindowPtr = oldwindowptr;

    return object;
}

#define SUPERCLASS MUIC_Area

static struct MUI_CustomClass *DataTypes_mcc = NULL;    /* for MUI */
struct IClass *DataTypesClass = NULL;                   /* for Intuition */

struct Data
{
    Object *d_Object;
    BOOL d_FixedSize;
    BOOL d_Scale;
    BOOL d_KeepAspect;
    BOOL d_OwnObject;
    void (*d_SaveBaseFont)(struct TextFont *tf);
};

static VOID freeDTObject(struct Data *data)
{
    if (data->d_OwnObject && data->d_Object != NULL) {
        DisposeDTObject(data->d_Object);
    }
    data->d_OwnObject = FALSE;
    data->d_Object = NULL;
}

static BOOL setattrs(struct Data *data, struct TagItem *tags)
{
    struct TagItem *tag;
    const char *filename;
    BOOL redraw = FALSE;

    while ((tag = NextTagItem(&tags)) != NULL) {
	switch (tag->ti_Tag) {
            case MUIA_DataTypes_Filename:
                freeDTObject(data);
                filename = (const char *)tag->ti_Data;
                if (filename != NULL) {
                    data->d_Object = LoadDTPicture(filename);
                    data->d_OwnObject = TRUE;
                }
                redraw = TRUE;
                break;

            case MUIA_DataTypes_Object:
                freeDTObject(data);
                data->d_Object = (Object *)tag->ti_Data;
                data->d_OwnObject = FALSE;
                redraw = TRUE;
                break;

            case MUIA_DataTypes_FixedSize:
                data->d_FixedSize = tag->ti_Data? TRUE : FALSE;
                break;

            case MUIA_DataTypes_Scale:
                data->d_Scale = tag->ti_Data? TRUE : FALSE;
                redraw = TRUE;
                break;

            case MUIA_DataTypes_KeepAspect:
                data->d_KeepAspect = tag->ti_Data? TRUE : FALSE;
                redraw = TRUE;
                break;

            case MUIA_DataTypes_SaveBaseFont:
                data->d_SaveBaseFont = (void (*)(struct TextFont*))tag->ti_Data;
                break;
	}
    }

    return redraw;
}

static IPTR mNew(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data;

    obj = (Object *)DoSuperMethodA(cl, obj, (Msg)msg);
    if (obj == NULL) {
	return 0;
    }

    data = INST_DATA(cl, obj);
    memset(data, 0, sizeof(struct Data));

    setattrs(data, msg->ops_AttrList);

#if 0
    /* tell MUI not to care about filling our background during MUIM_Draw */
    set(obj, MUIA_FillArea, FALSE);
#endif

    return (IPTR)obj;
}

static IPTR mDispose(struct IClass *cl,Object *obj,Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);

    freeDTObject(data);
    return DoSuperMethodA(cl, obj, msg);
}

static IPTR mSet(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data = INST_DATA(cl, obj);
    IPTR result;
    bool redraw;

    redraw = setattrs(data, msg->ops_AttrList);
    result = DoSuperMethodA(cl, obj, (Msg)msg);

    if (redraw) {
        MUI_Redraw(obj, MADF_DRAWOBJECT);
    }
    return result;
}

static IPTR mGet(struct IClass *cl, Object *obj, struct opGet *msg)
{
    struct Data *data = INST_DATA(cl, obj);

    switch (msg->opg_AttrID) {
        case MUIA_DataTypes_FixedSize:
            *msg->opg_Storage = (IPTR)data->d_FixedSize;
            return TRUE;
    }

    return DoSuperMethodA(cl, obj, (Msg)msg);
}

static IPTR mSetup(struct IClass *cl, Object *obj, Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);

    if (!DoSuperMethodA(cl, obj, msg)) {
        return FALSE;
    }

    if (data->d_SaveBaseFont != NULL) {
        data->d_SaveBaseFont(_font(obj));
    }

    return TRUE;
}

static IPTR mAskMinMax(struct IClass *cl, Object *obj, 
        struct MUIP_AskMinMax *msg)
{
    struct Data *data = INST_DATA(cl,obj);
    const struct BitMapHeader *bmh;
    struct MUI_MinMax *mi;

    DoSuperMethodA(cl, obj, (Msg)msg);
    bmh = pget(data->d_Object, PDTA_BitMapHeader);
    mi = msg->MinMaxInfo;

    if (data->d_FixedSize && bmh != NULL) {
        mi->MinWidth += bmh->bmh_Width ;
        mi->MinHeight += bmh->bmh_Height;
        mi->MaxWidth += bmh->bmh_Width ;
        mi->MaxHeight += bmh->bmh_Height;
    } else if (data->d_SaveBaseFont == NULL) {
        mi->MinWidth += 50;
        mi->MinHeight += 30;
        mi->MaxWidth += MUI_MAXMAX;
        mi->MaxHeight += MUI_MAXMAX;
    }

    if (bmh) {
        mi->DefWidth += bmh->bmh_Width ;
        mi->DefHeight += bmh->bmh_Height;
    }

    return 0;
}

static IPTR mDraw(struct IClass *cl, Object *obj, struct MUIP_Draw *msg)
{
    struct Data *data = INST_DATA(cl,obj);
    Object *dto = data->d_Object;
    const struct BitMapHeader *bmh;
    struct BitMap *bm, *scaled_bm = NULL;
    LONG src_x = 0, src_y = 0, src_w, src_h;
    LONG dst_x = _mleft(obj), dst_y = _mtop(obj);
    LONG dst_w = _mwidth(obj), dst_h = _mheight(obj);
    LONG xoff, yoff, w, h;
    ULONG depth;
    APTR mask = NULL;
#ifdef ENABLE_DTV44
    APTR handle = NULL;
    BOOL success = FALSE;
#endif

    DoSuperMethodA(cl, obj, (Msg)msg);

    if (!(msg->flags & MADF_DRAWOBJECT) || dto == NULL) {
        return 0;
    }

    /* Make the datatype object suitable for our screen */
    SetAttrs(dto,
        PDTA_Screen, _screen(obj),
        PDTA_UseFriendBitMap , TRUE,
        TAG_DONE);

    /*
    ** do all the setup/layout stuff that's necessary to get
    ** a bitmap from the dto.
    ** Note that when using V43 datatypes, this might not be
    ** a real "struct BitMap *"
    */
    DoMethod(dto, DTM_PROCLAYOUT, NULL, 1);
    bmh = pget(dto, PDTA_BitMapHeader);
    bm = pget(dto, PDTA_DestBitMap);

    if (bmh == NULL || bm == NULL) {
        return 0;
    }

    src_w = bmh->bmh_Width;
    src_h = bmh->bmh_Height;
    depth = GetBitMapAttr(bm, BMA_DEPTH);

    if (data->d_Scale) {
	struct BitScaleArgs bsa;
	LONG scaled_w = dst_w, scaled_h = dst_h;

	if (data->d_KeepAspect && src_w > 0 && src_h > 0) {
	    if (src_h * dst_w / src_w > dst_h) {
		scaled_w = src_w * dst_h / src_h;
	    } else {
		scaled_h = src_h * dst_w / src_w;
	    }
	}	
	scaled_bm = AllocBitMap(scaled_w, scaled_h, depth,
		BMF_MINPLANES |  BMF_INTERLEAVED, _rp(obj)->BitMap);

        if (scaled_bm != NULL) {
            bsa.bsa_SrcX = 0;
            bsa.bsa_SrcY = 0;			/* source origin */
            bsa.bsa_SrcWidth = src_w;
            bsa.bsa_SrcHeight = src_h;	        /* source size */
            bsa.bsa_XSrcFactor = src_w;
            bsa.bsa_YSrcFactor = src_h;	        /* scale factor denominators */
            bsa.bsa_DestX = 0;
            bsa.bsa_DestY = 0;		        /* destination origin */
            bsa.bsa_DestWidth = scaled_w;
            bsa.bsa_DestHeight = scaled_h;      /* destination size result */
            bsa.bsa_XDestFactor = scaled_w;
            bsa.bsa_YDestFactor = scaled_h;	/* scale factor numerators */
            bsa.bsa_SrcBitMap = bm;	        /* source BitMap */
            bsa.bsa_DestBitMap = scaled_bm;	/* destination BitMap */
            bsa.bsa_Flags = 0;			/* reserved.  Must be zero! */
            bsa.bsa_XDDA = 0;
            bsa.bsa_YDDA = 0;			/* reserved */
            bsa.bsa_Reserved1 = 0;
            bsa.bsa_Reserved2 = 0;

	    BitMapScale(&bsa);
            src_w = bsa.bsa_DestWidth;
            src_h = bsa.bsa_DestHeight;
            bm = scaled_bm;
        }
    } else {
	mask = pget(data->d_Object, PDTA_MaskPlane);
#ifdef ENABLE_DTV44
	if (DataTypesBase->lib_Version >= 44) {
	    handle = ObtainDTDrawInfoA(data->d_Object, NULL);
	}
#endif
    }

    xoff = (dst_w - src_w) / 2;
    if (xoff >= 0) {
        dst_x += xoff;
    } else {
        src_x -= xoff;
    }
    yoff = (dst_h - src_h) / 2;
    if (yoff >= 0) {
        dst_y += yoff;
    } else {
        src_y -= yoff;
    }
    w = MIN(dst_w, src_w);
    h = MIN(dst_h, src_h);
#ifdef ENABLE_DTV44
    if (handle) {
	success = DrawDTObjectA(_rp(obj), data->d_Object,
		dst_x, dst_y, w, h, src_x, src_y, NULL); 
	/* printf("%s: handle= %p, draw= %d\n", __FUNCTION__, 
	   handle, (int)success); */
	ReleaseDTDrawInfo(data->d_Object, handle);
    }
    if (!success)
#endif
    {
        /* printf("%s: mask= %p\n", __FUNCTION__,  mask); */
	if (mask) {
	    BltMaskBitMapRastPort(bm, src_x, src_y, _rp(obj), 
		dst_x, dst_y, w, h, ABC|ABNC|ANBC, mask);
	} else {
	    BltBitMapRastPort(bm, src_x, src_y, _rp(obj), 
                    dst_x, dst_y, w, h, ABC|ABNC);
	}
    }

    if (scaled_bm) {
	WaitBlit();
	FreeBitMap(scaled_bm);
    }

    return 0;
}

DISPATCHER_PROTO
{
    switch(msg->MethodID)
    {
        case OM_NEW:            return(mNew(cl, obj, (APTR)msg));
        case OM_DISPOSE:        return(mDispose(cl, obj, (APTR)msg));
	case OM_SET:            return(mSet(cl, obj, (APTR)msg));
	case OM_GET:            return(mGet(cl, obj, (APTR)msg));
        case MUIM_Setup:        return(mSetup(cl, obj, (APTR)msg));
        case MUIM_AskMinMax:    return(mAskMinMax(cl, obj, (APTR)msg));
        case MUIM_Draw :        return(mDraw(cl, obj, (APTR)msg));
    }

    return DoSuperMethodA(cl, obj, msg);
}

void DataTypesCleanup(void)
{
    if(DataTypes_mcc != NULL)
        MUI_DeleteCustomClass(DataTypes_mcc);
    DataTypes_mcc = NULL;
    DataTypesClass = NULL;

    if(DataTypesBase != NULL)
        CloseLibrary(DataTypesBase);
    DataTypesBase = NULL;
}

BOOL DataTypesInit(void)
{
    DataTypesBase = OpenLibrary((STRPTR)"datatypes.library", 39);

    DataTypes_mcc = MUI_CreateCustomClass(NULL, (STRPTR)SUPERCLASS, NULL,
        sizeof(struct Data), Dispatcher);
    if(DataTypes_mcc == NULL)
    {
        DataTypesCleanup();
        return(FALSE);
    }

    DataTypesClass = DataTypes_mcc->mcc_Class;
    DISPATCHER_INIT(DataTypesClass);

    return(TRUE);
}

