/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** EditLabels_mcc.c
**
** Creation date: 26.11.1996
** Last modified: 29.12.2020
*/

#include "system.h"
#include <intuition/icclass.h>
#include <clib/alib_protos.h>
#include <proto/utility.h>
#include "codeset.h"
#include "util.h"
#include "preferences.h"
#include "labels.h"
#include "appstrings_cat.h"
#include "DDList_mcc.h"
#include "EditLabels_mcc.h"

#define SUPERCLASS MUIC_Group

static struct MUI_CustomClass *EditLabels_mcc = NULL;     /* for MUI */
struct IClass *EditLabelsClass = NULL;                    /* for Intuition */

struct Data
{
  struct MemPool *d_MemPool;
  char **d_Labels;
  bool d_Compact;
  Object *d_Main;
  Object *d_Listview;
  Object *d_New;
  Object *d_Remove;
  Object *d_Sort;
  Object *d_String;
};

static void HOOK_PROTO(labelsactivechange, GuiObject go, struct Data **x)
{
    struct Data *d = *x;
    char *str = NULL;
    Object *wo = GuiObjectWindow(go);

    DoMethod(d->d_Listview, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, &str);
    set(d->d_Remove, MUIA_Disabled, str == NULL);
    SetAttrs(d->d_String,
        MUIA_String_Contents, EmptyNullStr(str),
        MUIA_Disabled, str == NULL,
        TAG_DONE);

    if(wo)
        set(wo, MUIA_Window_ActiveObject, d->d_String);
}

static void HOOK_PROTO(labelacknowledge, GuiObject go, struct Data **x)
{
    struct Data *d = *x;
    char *str = pget(d->d_String, MUIA_String_Contents);
    long active = nget(d->d_Listview, MUIA_List_Active);
    Object *wo = GuiObjectWindow(go);

    if(str == NULL)
	return;
    set(d->d_Listview, MUIA_List_Quiet, TRUE);
    if(active == MUIV_List_Active_Off)
    {
        DoMethod(d->d_Listview, MUIM_List_InsertSingle, str, MUIV_List_Insert_Bottom);
	set(d->d_String, MUIA_String_Contents, "");
    }
    else
    {
        DoMethod(d->d_Listview, MUIM_List_InsertSingle, str, active + 1);
        DoMethod(d->d_Listview, MUIM_List_Remove, active);
    }
    set(d->d_Listview, MUIA_List_Quiet, FALSE);

    if(wo)
        set(wo, MUIA_Window_ActiveObject, d->d_String);
}

static void HOOK_PROTO(labelnew, GuiObject go, struct Data **x)
{
    struct Data *d = *x;
    Object *wo = GuiObjectWindow(go);

    set(d->d_Listview, MUIA_List_Active, MUIV_List_Active_Off);
    set(d->d_String, MUIA_Disabled, FALSE);
    set(d->d_String, MUIA_String_Contents, "");
    if(wo)
        set(wo, MUIA_Window_ActiveObject, d->d_String);
}

static long HOOK_PROTO(compare, char *s2, char *s1)
{
    return CodesetStrCollate(s1, s2, CS_COLLATE_ORDER);
}
static HOOK_STRUCT(compare)

static IPTR setattrs(struct Data *data, struct TagItem *tags)
{
    struct TagItem *tag;
    IPTR rc = 0;

    while((tag = NextTagItem(&tags)) != NULL)
    {
	switch (tag->ti_Tag)
	{
            case MUIA_EditLabels_Labels:
                MUIListSetLabels(data->d_Listview, (char **)tag->ti_Data);
                rc = TRUE;
                break;
	}
    }
    return(rc);
}

static IPTR mNew(struct IClass *cl,Object *obj,struct opSet *msg)
{
    struct TagItem *tags = msg->ops_AttrList;
    struct Data *data, d;
    Object *control, *top, *labelgroup;

    d.d_Listview = ListviewObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_EditLabels_LabelsBH),
	MUIA_Listview_DragType, MUIV_Listview_DragType_Immediate,
	MUIA_Listview_List, DDListObject,
            MUIA_Frame, MUIV_Frame_InputList,
            MUIA_List_DragSortable, TRUE,
            MUIA_List_ConstructHook, MUIV_List_ConstructHook_String,
            MUIA_List_DestructHook, MUIV_List_DestructHook_String,
            MUIA_List_CompareHook, &compareHook,
            End,
	End;

    d.d_New = SimpleButton(GCS(MSG_New));
    d.d_Remove = SimpleButton(GCS(MSG_EditLabels_Remove));
    d.d_Sort = SimpleButton(GCS(MSG_EditLabels_Sort));

    d.d_Compact = GetPrefsValue(NULL, PREFS_EDITLABELSCOMPACT);

    if(d.d_Compact)
      {
	control = HGroup,
            Child, d.d_New,
	    Child, d.d_Remove,
	    Child, d.d_Sort,
	    End;

	top = VGroup,
	    Child, d.d_Listview,
	    Child, control,
	    End;
      }
    else
      {
	control = VGroup,
	    MUIA_Weight, 10,
	    Child, VSpace(0),
	    Child, d.d_New,
	    Child, d.d_Remove,
	    Child, d.d_Sort,
	    Child, VSpace(0),
	    End;

	top = HGroup,
	    Child, d.d_Listview,
	    Child, control,
	    End;
      }

    d.d_String = StringObject,
        MUIA_CycleChain, TRUE,
	MUIA_Disabled, TRUE,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_EditLabels_Label)),
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_AttachedList, d.d_Listview,
	End;

    labelgroup = HGroup,
	Child, Label2(GCS(MSG_EditLabels_Label)),
	Child, d.d_String,
	End;

    d.d_Main = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_HelpNode, "Label editor",
	Child, top,
	Child, labelgroup,
        End;

    obj = (Object *)DoSuperNew(cl, obj,
	Child, (IPTR)d.d_Main, 		       
        TAG_MORE, (IPTR)tags,
        TAG_DONE);
    if(obj == NULL)
        return(0);

    d.d_MemPool = CreatePool(0, 4096, 2048);
    d.d_Labels = NULL;
    data = INST_DATA(cl, obj);
    *data = d;

    SetAttrs(d.d_New,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_EditLabels_NewBH),
        TAG_DONE);
    SetAttrs(d.d_Remove,
        MUIA_CycleChain, TRUE,
        MUIA_Disabled, TRUE,
        MUIA_ShortHelp, GCS(MSG_EditLabels_RemoveBH),
        TAG_DONE);
    SetAttrs(d.d_Sort,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_EditLabels_SortBH),
        TAG_DONE);

    MUIAddCallback(d.d_Listview, MUIA_List_Active, MUIV_EveryTime,
        labelsactivechange, data);
    MUIAddCallback(d.d_String, MUIA_String_Acknowledge, MUIV_EveryTime,
        labelacknowledge, data);
    MUIAddCallback(d.d_New, MUIA_Pressed, FALSE, labelnew, data);
    DoMethod(d.d_Remove, MUIM_Notify, MUIA_Pressed, FALSE,
	d.d_Listview, 2, MUIM_List_Remove, MUIV_List_Remove_Active);
    DoMethod(d.d_Sort, MUIM_Notify, MUIA_Pressed, FALSE,
	d.d_Listview, 1, MUIM_List_Sort);

    setattrs(data, tags);

    return (IPTR)obj;
}

static IPTR mDispose(struct IClass *cl,Object *obj,Msg msg)
{
    struct Data *d = INST_DATA(cl, obj);

    if (d->d_MemPool != NULL) {
        DeletePool(d->d_MemPool);
        d->d_MemPool = NULL;
    }
    return DoSuperMethodA(cl, obj, msg);
}

static IPTR mSet(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data = INST_DATA(cl, obj);
    IPTR rc;

    rc = setattrs(data, msg->ops_AttrList);
    rc |= DoSuperMethodA(cl, obj, (Msg)msg);
    return(rc);
}

static IPTR mGet(struct IClass *cl, Object *obj, struct opGet *msg)
{
  struct Data *data = INST_DATA(cl, obj);

  switch(msg->opg_AttrID)
    {
    case MUIA_EditLabels_Labels:
      FreeLabels(data->d_MemPool, data->d_Labels);
      data->d_Labels = MUIListLabels(data->d_MemPool, data->d_Listview);
      *msg->opg_Storage = (IPTR)data->d_Labels;
      return(TRUE);
    }

  return(DoSuperMethodA(cl, obj, (Msg)msg));
}

DISPATCHER_PROTO
{
    switch (msg->MethodID)
    {
	case OM_NEW:		return(mNew(cl, obj, (APTR)msg));
	case OM_DISPOSE:        return(mDispose(cl, obj, (APTR)msg));
	case OM_GET:            return(mGet(cl, obj, (APTR)msg));
	case OM_SET:            return(mSet(cl, obj, (APTR)msg));
    }

    return(DoSuperMethodA(cl, obj, msg));
}

void EditLabelsCleanup(void)
{
    if(EditLabels_mcc != NULL)
        MUI_DeleteCustomClass(EditLabels_mcc);
    EditLabels_mcc = NULL;
    EditLabelsClass = NULL;
}

BOOL EditLabelsInit(void)
{
    EditLabels_mcc = MUI_CreateCustomClass(NULL, (STRPTR)SUPERCLASS, NULL,
        sizeof(struct Data), Dispatcher);
    if(EditLabels_mcc == NULL)
    {
        EditLabelsCleanup();
        return(FALSE);
    }

    EditLabelsClass = EditLabels_mcc->mcc_Class;
    DISPATCHER_INIT(EditLabelsClass);

    return(TRUE);
}

