/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** ExtApplication.c
**
** Creation date: 12.01.1997
** Last modified: 27.04.2008
*/

/* Includes */
#include "system.h"
#include <intuition/gadgetclass.h>
#include <intuition/icclass.h>
#include <clib/alib_protos.h>
#include <proto/utility.h>

#include "ExtApplication_mcc.h"

#define SUPERCLASS MUIC_Application

static struct MUI_CustomClass *ExtApplication_mcc = NULL;    /* for MUI */
struct IClass *ExtApplicationClass = NULL;                   /* for Intuition */

struct Data
{
    Object *d_DisposeList;
    long d_NestCounter;
    Object *d_ActiveExtWindow;
    struct MsgPort *d_RedirectPort;
};

static void setattrs(struct Data *data, struct TagItem *tags)
{
    struct TagItem *tag;

    while ((tag = NextTagItem(&tags)) != NULL)
    {
	switch(tag->ti_Tag)
	{
            case MUIA_ExtApplication_ActiveExtWindow:
                data->d_ActiveExtWindow = (Object *)tag->ti_Data;
                break;
	}
    }
}

static IPTR mNew(struct IClass *cl,Object *obj,struct opSet *msg)
{
    struct Data *data;

    obj = (Object *)DoSuperMethodA(cl, obj, (Msg)msg);
    if(obj == NULL)
	return(0);

    data = INST_DATA(cl, obj);
    memset(data, 0, sizeof(struct Data));

    data->d_RedirectPort = CreatePort(NULL, 0);
    if(data->d_RedirectPort == NULL)
    {
        CoerceMethod(cl, obj, OM_DISPOSE);
	return(0);
    }

    setattrs(data, msg->ops_AttrList);

    return (IPTR)obj;
}

static IPTR mSet(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data = INST_DATA(cl, obj);

    setattrs(data, msg->ops_AttrList);
    return(DoSuperMethodA(cl, obj, (Msg)msg));
}

static Object *activeextwindow(Object *app, struct Data *data)
{
    Object *win = data->d_ActiveExtWindow;
    Object *wo;
    struct List *list = pget(app, MUIA_Application_WindowList);
    struct Node *node;

    if(win == NULL || list == NULL)
        return(NULL);

    /* we check if the window is really alive */

    node = HEAD(list);
    while ((wo = NextObject(&node)) != NULL)
        if(wo == win)
            return(win);

    return(NULL);
}

static IPTR mGet(struct IClass *cl, Object *obj, struct opGet *msg)
{
    struct Data *data = INST_DATA(cl, obj);

    switch(msg->opg_AttrID)
    {
        case MUIA_ExtApplication_ActiveExtWindow:
            *msg->opg_Storage = (IPTR)activeextwindow(obj, data);
            return(TRUE);
    }

    return(DoSuperMethodA(cl, obj, (Msg)msg));
}

static void clearDisposeList(Object *dispose)
{
    Object *next;

    while(dispose != NULL)
    {
        next = pget(dispose, MUIA_UserData);
        MUI_DisposeObject(dispose);
        dispose = next;
    }
}

static IPTR mDispose(struct IClass *cl,Object *obj,Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);
    IPTR rc;

    clearDisposeList(data->d_DisposeList);
    data->d_DisposeList = NULL;

    if(data->d_RedirectPort != NULL)
    {
	DeletePort(data->d_RedirectPort);
	data->d_RedirectPort = NULL;
    }

    rc = DoSuperMethodA(cl, obj, msg);
    return(rc);
}

static IPTR mAddDisposeObject(struct IClass *cl, Object *obj,
        struct MUIP_ExtApplication_DisposeObject *eado)
{
    struct Data *data = INST_DATA(cl, obj);

    if(eado->eado_Object != NULL)
    {
        set(eado->eado_Object, MUIA_UserData, data->d_DisposeList);
        data->d_DisposeList = eado->eado_Object;
    }
    return(0);
}

static IPTR mInput(struct IClass *cl, Object *obj, Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);
    IPTR rc;

    if(data->d_NestCounter == 0)
    {
        clearDisposeList(data->d_DisposeList);
        data->d_DisposeList = NULL;
    }

    data->d_NestCounter++;
    rc = DoSuperMethodA(cl, obj, msg);
    data->d_NestCounter--;

    return(rc);
}

static IPTR mProjectSleep(struct IClass *cl, Object *obj,
        struct MUIP_ExtApplication_ProjectSleep *eaps)
{
    struct List *list = pget(obj, MUIA_Application_WindowList);
    struct Node *node;
    GuiObject wo;

    if(list == NULL)
        return(0);

    node = HEAD(list);
    while ((wo = NextObject(&node)) != NULL)
        if(pget(wo, MUIA_ExtWindow_Project) == eaps->eaps_Project)
            set(wo, MUIA_Window_Sleep, eaps->eaps_Sleep);
    return(0);
}

static IPTR mRedirectMessage(struct IClass *cl, Object *obj,
        struct MUIP_ExtApplication_RedirectMessage *earm)
{
    struct Data *data = INST_DATA(cl, obj);

    if(earm->earm_Message != NULL)
    {
	earm->earm_Message->mn_ReplyPort = data->d_RedirectPort;
    }
	
    return 0;
}

static IPTR mGetMessage(struct IClass *cl, Object *obj,
        struct MUIP_ExtApplication_GetMessage *eagm)
{
    struct Data *data = INST_DATA(cl, obj);
    struct Message *msg, *succ;
    IPTR rc = 0;

    Forbid();
    SAFEFORNODE(msg, succ, &data->d_RedirectPort->mp_MsgList)
    {
	if(msg == eagm->eagm_Message)
	{
	    REMOVE(msg);
	    rc = (IPTR)msg;
	    break;
	}
    }
    Permit();

    return rc;
}

/*
** Here we need the INTERRUPT keyword do turn off stack checking
** since methods of the application object might also be called
** from outside BeeBase (e.g. MUIM_Application_PushMethod)
** Alternatively, we could compile this file without stack checking.
*/
INTERRUPT DISPATCHER_PROTO
{
    switch (msg->MethodID)
    {
    case OM_NEW:		return(mNew(cl, obj, (APTR)msg));
    case OM_SET:		return(mSet(cl, obj, (APTR)msg));
    case OM_GET:		return(mGet(cl, obj, (APTR)msg));
    case OM_DISPOSE:        return(mDispose(cl, obj, (APTR)msg));
    case MUIM_ExtApplication_DisposeObject: 
	return(mAddDisposeObject(cl, obj, (APTR)msg));
    case MUIM_Application_Input:
    case MUIM_Application_NewInput:
    case MUIM_Application_InputBuffered: return(mInput(cl, obj, msg));
    case MUIM_ExtApplication_ProjectSleep: 
	return(mProjectSleep(cl, obj, (APTR)msg));
    case MUIM_ExtApplication_RedirectMessage:
	return mRedirectMessage(cl, obj, (APTR)msg);
    case MUIM_ExtApplication_GetMessage:
	return mGetMessage(cl, obj, (APTR)msg);
    }

    return(DoSuperMethodA(cl, obj, msg));
}

void ExtApplicationCleanup(void)
{
    if(ExtApplication_mcc != NULL)
        MUI_DeleteCustomClass(ExtApplication_mcc);
    ExtApplication_mcc = NULL;
    ExtApplicationClass = NULL;
}

BOOL ExtApplicationInit(void)
{
    ExtApplication_mcc = MUI_CreateCustomClass(NULL, (STRPTR)SUPERCLASS, NULL,
        sizeof(struct Data), Dispatcher);
    if(ExtApplication_mcc == NULL)
        return(FALSE);

    ExtApplicationClass = ExtApplication_mcc->mcc_Class;
    DISPATCHER_INIT(ExtApplicationClass);

    return(TRUE);
}


