/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** ExtEditor.c
**
** Creation date: 27.02.1996
** Last modified: 06.03.2011
**
** ExtEditor.mcc is a subclass of MUIC_TextEditor or MUIC_Textinput class.
** It attaches a context menu strip to the editor object.
*/

#include "ExtEditor_mcc.h"
#include "memory.h"
#include "appstrings_cat.h"
#include <proto/utility.h>

#ifdef USE_TEXTEDITOR
#define SUPERCLASS MUIC_TextEditor
#elif defined(USE_TEXTINPUT)
#define SUPERCLASS MUIC_Textinputscroll
#endif

static struct MUI_CustomClass *ExtEditor_mcc = NULL;    /* for MUI */
struct IClass *ExtEditorClass = NULL;                   /* for Intuition */

struct Data
{
#ifdef USE_TEXTEDITOR
    Object *d_Menustrip;
    void (*d_LoadFunc)(Object *, void *);
    void (*d_SaveFunc)(Object *, void *);
    void (*d_ExternalEditorFunc)(Object *, void *);
    void *d_FuncData;
#endif
};

#ifdef USE_TEXTEDITOR

static void HOOK_PROTO(load, Object *menuitem, Object **x)
{
    Object *obj = *x;
    struct Data *data;

    if (obj == NULL) {
        return;                 /* be safe */
    }

    data = INST_DATA(ExtEditorClass, obj);
    if (data->d_LoadFunc) {
        data->d_LoadFunc(obj, data->d_FuncData);
    }
}

static void HOOK_PROTO(save, Object *menuitem, Object **x)
{
    Object *obj = *x;
    struct Data *data;

    if (obj == NULL) {
        return;                 /* be safe */
    }

    data = INST_DATA(ExtEditorClass, obj);
    if (data->d_SaveFunc) {
        data->d_SaveFunc(obj, data->d_FuncData);
    }
}

static void HOOK_PROTO(externalEditor, Object *menuitem, Object **x)
{
    Object *obj = *x;
    struct Data *data;

    if(obj == NULL)
        return;                 /* be safe */

    data = INST_DATA(ExtEditorClass, obj);
    if(data->d_ExternalEditorFunc)
        data->d_ExternalEditorFunc(obj, data->d_FuncData);
}

static void setattrs(struct Data *data, struct TagItem *tags)
{
    struct TagItem *tag;

    while((tag = NextTagItem(&tags)) != NULL)
    {
	switch(tag->ti_Tag)
	{
            case MUIA_ExtEditor_LoadFunc:
                data->d_LoadFunc = (void (*)(Object *, void *))tag->ti_Data;
                break;

            case MUIA_ExtEditor_SaveFunc:
                data->d_SaveFunc = (void (*)(Object *, void *))tag->ti_Data;
                break;

            case MUIA_ExtEditor_ExternalEditorFunc:
                data->d_ExternalEditorFunc = (void (*)(Object *, void *))tag->ti_Data;
                break;

            case MUIA_ExtEditor_FuncData:
                data->d_FuncData = (void *)tag->ti_Data;
                break;
        }
    }
}

static IPTR mNew(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data;
    struct TagItem *tags = msg->ops_AttrList;
    Object *clear = NULL, *cut = NULL, *copy = NULL,
           *paste = NULL, *undo = NULL, *redo = NULL;
    Object *load = NULL, *save = NULL, *edit = NULL;
    bool h;

    obj = (Object *)DoSuperMethodA(cl, obj, (Msg)msg);
    if (obj == NULL) {
	return 0;
    }

    data = INST_DATA(cl, obj);
    memset(data, 0, sizeof(struct Data));
    setattrs(data, tags);

    if (data->d_LoadFunc) {
        load = MenuitemObject,
            MUIA_Menuitem_Title, GCS(MSG_OpenText),
            MUIA_Menuitem_Shortcut, "o",
            End;
    }
    if (data->d_SaveFunc) {
        save = MenuitemObject,
            MUIA_Menuitem_Title, GCS(MSG_SaveText),
            MUIA_Menuitem_Shortcut, "w",
            End;
    }
    if (data->d_ExternalEditorFunc) {
        edit = MenuitemObject,
            MUIA_Menuitem_Title, GCS(MSG_MenuItem_ExternalEditor),
            MUIA_Menuitem_Shortcut, "!",
            End;
    }

    data->d_Menustrip = MenustripObject,
        MUIA_Family_Child, MenuObject,
            MUIA_Menu_Title, GCS(MSG_Menu_Editor),
            MUIA_Family_Child, cut = MenuitemObject,
                MUIA_Menuitem_Title, GCS(MSG_MenuItem_Cut),
/*                MUIA_Menuitem_Shortcut, "x", */
                MUIA_Menuitem_Enabled, FALSE,
                End,
            MUIA_Family_Child, copy = MenuitemObject,
                MUIA_Menuitem_Title, GCS(MSG_MenuItem_Copy),
/*                MUIA_Menuitem_Shortcut, "c", */
                MUIA_Menuitem_Enabled, FALSE,
                End,
            MUIA_Family_Child, paste = MenuitemObject,
                MUIA_Menuitem_Title, GCS(MSG_MenuItem_Paste),
/*                MUIA_Menuitem_Shortcut, "v", */
                End,
            MUIA_Family_Child, clear = MenuitemObject,
                MUIA_Menuitem_Title, GCS(MSG_MenuItem_Clear),
/*                MUIA_Menuitem_Shortcut, "b", */
                End,
            MUIA_Family_Child, MenuitemObject,
                MUIA_Menuitem_Title, ~0,
                End,
            MUIA_Family_Child, undo = MenuitemObject,
                MUIA_Menuitem_Title, GCS(MSG_MenuItem_Undo),
/*                MUIA_Menuitem_Shortcut, "z", */
                End,
            MUIA_Family_Child, redo = MenuitemObject,
                MUIA_Menuitem_Title, GCS(MSG_MenuItem_Redo),
/*                MUIA_Menuitem_Shortcut, "Z", */
                End,
            MUIA_Family_Child, MenuitemObject,
                MUIA_Menuitem_Title, ~0,
                End,
            data->d_LoadFunc? MUIA_Family_Child : TAG_IGNORE, load,
            data->d_SaveFunc? MUIA_Family_Child : TAG_IGNORE, save,
            data->d_ExternalEditorFunc? MUIA_Family_Child : TAG_IGNORE, edit,
            End,
        End;
    if (data->d_Menustrip == NULL) {
        CoerceMethod(cl, obj, OM_DISPOSE);
	return 0;
    }

    DoMethod(cut, MUIM_Notify, MUIA_Menuitem_Trigger, MUIV_EveryTime,
        obj, 2, MUIM_TextEditor_ARexxCmd, "Cut");
    DoMethod(copy, MUIM_Notify, MUIA_Menuitem_Trigger, MUIV_EveryTime,
        obj, 2, MUIM_TextEditor_ARexxCmd, "Copy");
    DoMethod(paste, MUIM_Notify, MUIA_Menuitem_Trigger, MUIV_EveryTime,
        obj, 2, MUIM_TextEditor_ARexxCmd, "Paste");
    DoMethod(clear, MUIM_Notify, MUIA_Menuitem_Trigger, MUIV_EveryTime,
        obj, 1, MUIM_TextEditor_ClearText);

    DoMethod(undo, MUIM_Notify, MUIA_Menuitem_Trigger, MUIV_EveryTime,
        obj, 2, MUIM_TextEditor_ARexxCmd, "Undo");
    DoMethod(redo, MUIM_Notify, MUIA_Menuitem_Trigger, MUIV_EveryTime,
        obj, 2, MUIM_TextEditor_ARexxCmd, "Redo");

    if (load) {
        MUIAddCallback(load, MUIA_Menuitem_Trigger, MUIV_EveryTime, load, obj);
    }
    if (save) {
        MUIAddCallback(save, MUIA_Menuitem_Trigger, MUIV_EveryTime, save, obj);
    }
    if (edit) {
        MUIAddCallback(edit, MUIA_Menuitem_Trigger, MUIV_EveryTime, 
                externalEditor, obj);
    }

    h = nget(obj, MUIA_Disabled);
    DoMethod(obj, MUIM_Notify, MUIA_Disabled, TRUE,
        obj, 3, MUIM_Set, MUIA_ContextMenu, NULL);
    DoMethod(obj, MUIM_Notify, MUIA_Disabled, FALSE,
        obj, 3, MUIM_Set, MUIA_ContextMenu, data->d_Menustrip);
    set(obj, MUIA_ContextMenu, h? NULL : data->d_Menustrip);

    h = nget(obj, MUIA_TextEditor_ReadOnly);
    DoMethod(obj, MUIM_Notify, MUIA_TextEditor_ReadOnly, MUIV_EveryTime,
        cut, 3, MUIM_Set, MUIA_Menuitem_Enabled, MUIV_NotTriggerValue);
    set(cut, MUIA_Menuitem_Enabled, !h);
    DoMethod(obj, MUIM_Notify, MUIA_TextEditor_ReadOnly, MUIV_EveryTime,
        paste, 3, MUIM_Set, MUIA_Menuitem_Enabled, MUIV_NotTriggerValue);
    set(paste, MUIA_Menuitem_Enabled, !h);
    DoMethod(obj, MUIM_Notify, MUIA_TextEditor_ReadOnly, MUIV_EveryTime,
        clear, 3, MUIM_Set, MUIA_Menuitem_Enabled, MUIV_NotTriggerValue);
    set(clear, MUIA_Menuitem_Enabled, !h);

    DoMethod(obj, MUIM_Notify, MUIA_TextEditor_ReadOnly, MUIV_EveryTime,
        undo, 3, MUIM_Set, MUIA_Menuitem_Enabled, MUIV_NotTriggerValue);
    set(undo, MUIA_Menuitem_Enabled, !h);
    DoMethod(obj, MUIM_Notify, MUIA_TextEditor_ReadOnly, MUIV_EveryTime,
        redo, 3, MUIM_Set, MUIA_Menuitem_Enabled, MUIV_NotTriggerValue);
    set(redo, MUIA_Menuitem_Enabled, !h);

    if (load != NULL) {
        DoMethod(obj, MUIM_Notify, MUIA_TextEditor_ReadOnly, MUIV_EveryTime,
            load, 3, MUIM_Set, MUIA_Menuitem_Enabled, MUIV_NotTriggerValue);
        set(load, MUIA_Menuitem_Enabled, !h);
    }
    if (edit != NULL) {
        DoMethod(obj, MUIM_Notify, MUIA_TextEditor_ReadOnly, MUIV_EveryTime,
            edit, 3, MUIM_Set, MUIA_Menuitem_Enabled, MUIV_NotTriggerValue);
        set(edit, MUIA_Menuitem_Enabled, !h);
    }

    DoMethod(obj, MUIM_Notify, MUIA_TextEditor_AreaMarked, MUIV_EveryTime,
        cut, 3, MUIM_Set, MUIA_Menuitem_Enabled, MUIV_TriggerValue);
    DoMethod(obj, MUIM_Notify, MUIA_TextEditor_AreaMarked, MUIV_EveryTime,
        copy, 3, MUIM_Set, MUIA_Menuitem_Enabled, MUIV_TriggerValue);

    h = nget(obj, MUIA_TextEditor_UndoAvailable);
    DoMethod(obj, MUIM_Notify, MUIA_TextEditor_UndoAvailable, MUIV_EveryTime,
        undo, 3, MUIM_Set, MUIA_Menuitem_Enabled, MUIV_TriggerValue);
    set(undo, MUIA_Menuitem_Enabled, h);

    h = nget(obj, MUIA_TextEditor_RedoAvailable);
    DoMethod(obj, MUIM_Notify, MUIA_TextEditor_RedoAvailable, MUIV_EveryTime,
        redo, 3, MUIM_Set, MUIA_Menuitem_Enabled, MUIV_TriggerValue);
    set(redo, MUIA_Menuitem_Enabled, h);

    return (IPTR)obj;
}

static IPTR mDispose(struct IClass *cl,Object *obj,Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);
    Object *menustrip = data->d_Menustrip;
    IPTR rc;

    rc = DoSuperMethodA(cl, obj, msg);
    if (menustrip) {
        MUI_DisposeObject(menustrip);
    }

    return rc;
}

static IPTR mSet(struct IClass *cl,Object *obj,struct opSet *msg)
{
    struct Data *data = INST_DATA(cl, obj);

    setattrs(data, msg->ops_AttrList);
    return DoSuperMethodA(cl,obj,(Msg)msg);
}

static IPTR mGet(struct IClass *cl, Object *obj, struct opGet *msg)
{
    return DoSuperMethodA(cl, obj, (Msg)msg);
}
#endif

DISPATCHER_PROTO
{
#ifdef USE_TEXTEDITOR
    switch (msg->MethodID)
    {
	case OM_NEW:		return mNew(cl, obj, (APTR)msg);
	case OM_DISPOSE:        return mDispose(cl, obj, (APTR)msg);
	case OM_SET:            return mSet(cl, obj, (APTR)msg);
	case OM_GET:            return mGet(cl, obj, (APTR)msg);
    }
#endif

    return DoSuperMethodA(cl, obj, msg);
}

void ExtEditorCleanup(void)
{
    if (ExtEditor_mcc != NULL) {
        MUI_DeleteCustomClass(ExtEditor_mcc);
    }
    ExtEditor_mcc = NULL;
    ExtEditorClass = NULL;
}

bool ExtEditorInit(void)
{
    ExtEditor_mcc = MUI_CreateCustomClass(NULL, (STRPTR)SUPERCLASS, NULL,
        sizeof(struct Data), Dispatcher);
    if (ExtEditor_mcc == NULL) {
        return FALSE;
    }

    ExtEditorClass = ExtEditor_mcc->mcc_Class;
    DISPATCHER_INIT(ExtEditorClass);

    return TRUE;
}

char *ExtEditorGetContents(Object *obj, struct MemPool *memPool)
{
    char *text, *s = NULL;

#ifdef USE_TEXTEDITOR
    if (obj != NULL) {
        text = (char *)DoMethod(obj, MUIM_TextEditor_ExportText);
        if (text != NULL) {
            s = AllocString(memPool, text);
            FreeVec(text);
        }
    }
#elif defined(USE_TEXTINPUT)
    text = (char *)pget(obj, MUIA_Textinput_Contents);
    s = AllocString(memPool, text);
#endif

    return s;
}

void ExtEditorSetCursor(Object *obj, long line, long col)
{
#ifdef USE_TEXTEDITOR
    if (obj == NULL) {
        return;
    }

    line = MAX(line-1, 0);
    col = MAX(col-1, 0);

    SetAttrs(obj,
        MUIA_TextEditor_CursorY, line,
        MUIA_TextEditor_CursorX, col,
        TAG_DONE);
#elif defined(USE_TEXTINPUT)
    const char *text = pget(obj, MUIA_Textinput_Contents);
    const char *s = text;

    if (text) {
        while (line > 1) {
            s = strchr(s, '\n');
            if (s == NULL) {
                break;
            }
            s++;
            line--;
        }
        if (col > 0 && s) {
            s += col - 1;
        }
        set(obj, MUIA_Textinput_CursorPos, s? s - text : strlen(text));
    }
#endif
}

void ExtEditorGetCursor(Object *obj, long *pline, long *pcol)
{
#ifdef USE_TEXTEDITOR
    if (pline) {
        *pline = nget(obj, MUIA_TextEditor_CursorY) + 1;
    }
    if (pcol) {
        *pcol = nget(obj, MUIA_TextEditor_CursorX) + 1;
    }
#elif defined(USE_TEXTINPUT)
    const char *text = pget(obj, MUIA_Textinput_Contents);
    long pos = nget(obj, MUIA_Textinput_CursorPos);
    const char *s = text, *t;
    long line = 0, col = 0;

    if (text) {
        for (;;) {
            t = strchr(s, '\n');
            if (t == NULL || t - text >= pos) {
                break;
            }
            s = t + 1;
            line++;
        }
        col = pos - (s - text);
    }

    if (pline) {
        *pline = line + 1;
    }
    if (pcol) {
        *pcol = col + 1;
    }
#endif
}

