/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** ExtWindow.c
**
** Creation date: 16.03.1997
** Last modified: 10.10.2015
*/

#include "system.h"
#include "ExtWindow_mcc.h"
#include <intuition/gadgetclass.h>
#include <proto/utility.h>
#include "util.h"

#define SUPERCLASS MUIC_Window

static struct MUI_CustomClass *ExtWindow_mcc = NULL;    /* for MUI */
struct IClass *ExtWindowClass = NULL;                   /* for Intuition */
static const char titleSep[] = " | ";

struct Data
{
    void *d_Project;
    void (*d_IconifyFunc)(Object *obj);
    bool d_DimensionsValid;
    short d_LeftEdge;
    short d_TopEdge;
    short d_Width;
    short d_Height;
    char *d_Title;
    char *d_PrependTitle;
    bool (*d_KeyFunc)(Object *, char *, short, ushort, void *);
    void *d_KeyData;
    bool (*d_MouseFunc)(Object *, short x, short y, bool down);
    UWORD d_Qualifier;
    struct MUI_EventHandlerNode ehnode;
};

static bool setTitle(struct IClass *cl, Object *obj, const char *title)
{
    struct Data *data = INST_DATA(cl, obj);
    const char *prepend = data->d_PrependTitle;
    size_t size;
    char *copy;

    title = EmptyNullStr(title);
    prepend = EmptyNullStr(prepend);

    size = strlen(title) + 1;
    if (prepend[0] != 0) {
        size += strlen(prepend) + strlen(titleSep);
    }
    copy = ALLOCMEM(size);
    if (copy == NULL) {
        return FALSE;
    }
    if (prepend[0] != 0) {
        sprintf(copy, "%s%s%s", prepend, titleSep, title);
    } else {
        strcpy(copy, title);
    }
    SetSuperAttrs(cl, obj, MUIA_Window_Title, (IPTR)copy, TAG_END);
    if (data->d_Title) {
        FREEMEM(data->d_Title, strlen(data->d_Title) + 1);
    }
    data->d_Title = copy;
    return TRUE;
}

static void setPrependTitle(struct IClass *cl, Object *obj, const char *prepend)
{
    struct Data *data = INST_DATA(cl, obj);
    const char *title = data->d_Title;
    const char *current_prepend = EmptyNullStr(data->d_PrependTitle);
    int current_len = strlen(current_prepend);
    int sep_len = strlen(titleSep);
    char *copy = NULL;

    prepend = EmptyNullStr(prepend);
    if (strcmp(prepend, current_prepend) != 0) {
        if (current_len != 0 && title != NULL &&
                strlen(title) >= current_len + sep_len &&
                strncmp(title, current_prepend, current_len) == 0 &&
                strncmp(title + current_len, titleSep, sep_len) == 0) {
            title += current_len + sep_len;
        }
        if (prepend[0] != 0) {
            copy = ALLOCMEM(strlen(prepend) + 1);
            if (copy == NULL) {
                return;
            }
            strcpy(copy, prepend);
        }
        if (data->d_PrependTitle != NULL) {
            FREEMEM(data->d_PrependTitle, current_len + 1);
        }
        data->d_PrependTitle = copy;
        setTitle(cl, obj, title);
    }
}

static void setattrs(struct IClass *cl, Object *obj, struct TagItem *tags)
{
    struct Data *data = INST_DATA(cl, obj);
    struct TagItem *tag;

    while ((tag = NextTagItem(&tags)) != NULL)
    {
	switch (tag->ti_Tag) {
            case MUIA_ExtWindow_Project:
                data->d_Project = (void *)tag->ti_Data;
                break;

            case MUIA_ExtWindow_PrependTitle:
                setPrependTitle(cl, obj, (const char *)tag->ti_Data);
                break;

            case MUIA_Window_Title:
                setTitle(cl, obj, (const char *)tag->ti_Data); 
                tag->ti_Tag = TAG_IGNORE;    /* make super-class ignore it */
                break;

            case MUIA_ExtWindow_IconifyFunc:
                data->d_IconifyFunc = (void (*)(Object *))tag->ti_Data;
                break;

            case MUIA_KeyFunc:
                data->d_KeyFunc = (bool (*)(Object *, char *, short, ushort, 
                            void *))tag->ti_Data;
                break;

            case MUIA_KeyData:
                data->d_KeyData = (void *)tag->ti_Data;
                break;

            case MUIA_MouseFunc:
                data->d_MouseFunc = 
                    (bool (*)(Object *, short, short, bool))tag->ti_Data;
                break;
	}
    }
}

static void HOOK_PROTO(windowsleep, Object *obj, void *dummy)
{
    Object *app = pget(obj, MUIA_ApplicationObject);

    if (!nget(obj, MUIA_Window_Sleep) && nget(obj, MUIA_Window_Activate))
        set(app, MUIA_ExtApplication_ActiveExtWindow, obj);
}

static void HOOK_PROTO(openwindow, Object *obj, void *dummy)
{
    Object *app = pget(obj, MUIA_ApplicationObject);

    if(nget(obj, MUIA_Window_Activate))
        set(app, MUIA_ExtApplication_ActiveExtWindow, obj);
}

static IPTR mNew(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data;

    obj = (Object *)DoSuperMethodA(cl, obj, (Msg)msg);
    if (obj == NULL) {
	return 0;
    }

    data = INST_DATA(cl, obj);
    memset(data, 0, sizeof(struct Data));

    setattrs(cl, obj, msg->ops_AttrList);

    DoMethod(obj, MUIM_Notify, MUIA_Window_Activate, TRUE,
        MUIV_Notify_Application, 3, 
	MUIM_Set, MUIA_ExtApplication_ActiveExtWindow, obj);
    MUIAddCallback(obj, MUIA_Window_Sleep, MUIV_EveryTime, windowsleep, NULL); 
    MUIAddCallback(obj, MUIA_Window_Open, TRUE, openwindow, NULL);

    return (IPTR)obj;
}

static IPTR mDispose(struct IClass *cl, Object *obj, Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);
    char *title = data->d_Title;
    char *prepend = data->d_PrependTitle;

    if (prepend) {
        FREEMEM(prepend, strlen(prepend) + 1);
    }
    if (title) {
        FREEMEM(title, strlen(title) + 1);
    }
    return DoSuperMethodA(cl, obj, msg);
}

static IPTR mSet(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data = INST_DATA(cl, obj);
    struct Window *window = pget(obj, MUIA_Window_Window);

    if (window != NULL) {
        data->d_LeftEdge = window->LeftEdge;
        data->d_TopEdge = window->TopEdge;
        data->d_Width = window->Width;
        data->d_Height = window->Height;
        data->d_DimensionsValid = TRUE;
    }
    setattrs(cl, obj, msg->ops_AttrList);
    return DoSuperMethodA(cl, obj, (Msg)msg);
}

static IPTR mGet(struct IClass *cl, Object *obj, struct opGet *msg)
{
    struct Data *data = INST_DATA(cl, obj);

    switch (msg->opg_AttrID) {
        case MUIA_ExtWindow_Project:
            *msg->opg_Storage = (IPTR)data->d_Project;
            return TRUE;

        case MUIA_Window_LeftEdge:
            if (data->d_DimensionsValid) {
                *msg->opg_Storage = data->d_LeftEdge;
                return TRUE;
            }
            break;

        case MUIA_Window_TopEdge:
            if (data->d_DimensionsValid) {
                *msg->opg_Storage = data->d_TopEdge;
                return TRUE;
            }
            break;

        case MUIA_Window_Width:
            if (data->d_DimensionsValid) {
                *msg->opg_Storage = data->d_Width;
                return TRUE;
            }
            break;

        case MUIA_Window_Height:
            if (data->d_DimensionsValid) {
                *msg->opg_Storage = data->d_Height;
                return TRUE;
            }
            break;

	case MUIA_Qualifier:
	    *msg->opg_Storage = data->d_Qualifier;
	    return TRUE;
    }

    return DoSuperMethodA(cl, obj, (Msg)msg);
}

static IPTR mSetup(struct IClass *cl, Object *obj, Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);

    if (!DoSuperMethodA(cl, obj, msg)) {
        return FALSE;
    }

    data->ehnode.ehn_Object = obj;
    data->ehnode.ehn_Class  = cl;
    data->ehnode.ehn_Events = IDCMP_RAWKEY|IDCMP_MOUSEBUTTONS|IDCMP_INTUITICKS;
    /* 
    ** On AOS4 in order to read the Qualifier before a button callback
    ** the priority has to be raised
    */
    data->ehnode.ehn_Priority = 10;
    DoMethod(obj, MUIM_Window_AddEventHandler, &data->ehnode);

    /* printf("setup ok %08lx\n", _flags(obj)); */
    return TRUE;
}

static IPTR mCleanup(struct IClass *cl, Object *obj, Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);

    DoMethod(obj, MUIM_Window_RemEventHandler, &data->ehnode);
    return DoSuperMethodA(cl, obj, msg);
}

static IPTR mHandleEvent(struct IClass *cl, Object *obj,
        struct  MUIP_HandleEvent *msg)
{
    struct Data *d = INST_DATA(cl, obj);
    struct IntuiMessage *imsg = msg->imsg;
    ushort code;

    /* printf("imsg: %p\n", imsg); */
    if (imsg == NULL) {
        return 0;                      /* be safe */
    }

    d->d_Qualifier = imsg->Qualifier;
    code = imsg->Code;

    if (d->d_KeyFunc && imsg->Class==IDCMP_RAWKEY &&
            !(code & IECODE_UP_PREFIX)) {
        char buffer[10];
        short l = MUIRawKeyConvert(imsg, buffer, 10);

        if (d->d_KeyFunc(obj, buffer, l, imsg->Qualifier, d->d_KeyData)) {
            return MUI_EventHandlerRC_Eat;
        }
    } else if (d->d_MouseFunc != NULL && imsg->Class == IDCMP_MOUSEBUTTONS &&
            (code == SELECTUP || code == SELECTDOWN)) {
	if (d->d_MouseFunc(obj, imsg->MouseX, imsg->MouseY, code==SELECTDOWN)) {
	    return MUI_EventHandlerRC_Eat;
        }
    }
    return 0;
}

#if 0
static IPTR mIconify(struct IClass *cl, Object *obj, Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);

    if(data->d_IconifyFunc)
    {
        data->d_IconifyFunc(obj);
        return(0);
    }

    return(DoSuperMethodA(cl, obj, (Msg)msg));
}
#endif

DISPATCHER_PROTO
{
    switch (msg->MethodID) {
	case OM_NEW:		return(mNew(cl, obj, (APTR)msg));
        case OM_DISPOSE:        return(mDispose(cl, obj, (APTR)msg));
	case OM_GET:            return(mGet(cl, obj, (APTR)msg));
	case OM_SET:            return(mSet(cl, obj, (APTR)msg));
	case MUIM_Window_Setup: return(mSetup(cl, obj, (APTR)msg));
	case MUIM_Window_Cleanup: return(mCleanup(cl, obj, (APTR)msg));
        case MUIM_HandleEvent:  return(mHandleEvent(cl, obj, (APTR)msg));
#if 0
        case MUIM_Window_ActionIconify:
                                return(mIconify(cl, obj, (APTR)msg));
#endif
    }

    return DoSuperMethodA(cl, obj, msg);
}

void ExtWindowCleanup(void)
{
    if (ExtWindow_mcc != NULL) {
        MUI_DeleteCustomClass(ExtWindow_mcc);
    }
    ExtWindow_mcc = NULL;
    ExtWindowClass = NULL;
}

BOOL ExtWindowInit(void)
{
    ExtWindow_mcc = MUI_CreateCustomClass(NULL, (STRPTR)SUPERCLASS, NULL,
        sizeof(struct Data), Dispatcher);
    if (ExtWindow_mcc == NULL) {
        return FALSE;
    }

    ExtWindowClass = ExtWindow_mcc->mcc_Class;
    DISPATCHER_INIT(ExtWindowClass);

    return TRUE;
}

