/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** NewCycle_mcc.c
**
** Creation date: 25.11.1996
** Last modified: 29.01.2008
**
*/

/* Includes */
#include "system.h"
#include "NewCycle_mcc.h"
#include <intuition/icclass.h>
#include <clib/alib_protos.h>
#include <proto/utility.h>

#define SUPERCLASS MUIC_Group

static struct MUI_CustomClass *NewCycle_mcc = NULL;     /* for MUI */
struct IClass *NewCycleClass = NULL;                    /* for Intuition */

struct Data
{
    Object *d_Cycle;
    char **d_Labels;
};

static IPTR setattrs(struct IClass *cl, Object *obj, struct TagItem *tags)
{
    struct Data *data = INST_DATA(cl, obj);
    bool create = (data->d_Cycle == NULL);
    struct TagItem *tag;
    IPTR rc = 0;

    tag = FindTagItem(MUIA_Cycle_Entries, tags);
    if (tag) {
	data->d_Labels = (char **)tag->ti_Data;
	create = TRUE;
    }
    if (create) {
	static char *dummy[] = { "", NULL };
	bool cycleChain = nget(data->d_Cycle, MUIA_CycleChain);
	char controlChar = nget(data->d_Cycle, MUIA_ControlChar);
	char *shortHelp = (char *)nget(data->d_Cycle, MUIA_ShortHelp);
	short active = nget(data->d_Cycle, MUIA_Cycle_Active);
	char **labels = (data->d_Labels && data->d_Labels[0])? 
	    data->d_Labels : dummy;
	Object *cycle;

        cycle = CycleObject,
	    MUIA_CycleChain, cycleChain,
	    MUIA_ControlChar, controlChar,
	    MUIA_ShortHelp, shortHelp,
	    MUIA_Cycle_Entries, labels,
	    MUIA_Cycle_Active, active,
	    End;
        if (cycle == NULL) {
            return rc;
        }

        if (!DoMethod(obj, MUIM_Group_InitChange)) {
            MUI_DisposeObject(cycle);
            return rc;
        }

	if (data->d_Cycle) {
	    MUIRemMember(obj, data->d_Cycle);
	    MUI_DisposeObject(data->d_Cycle);
	}
        data->d_Cycle = cycle;
        MUIAddMember(obj, data->d_Cycle);
        DoMethod(obj, MUIM_Group_ExitChange);
    }

    /* distribute tags between cycle and group class */
    while ((tag = NextTagItem(&tags)) != NULL) {
        switch (tag->ti_Tag) {
            case MUIA_CycleChain:
            case MUIA_ControlChar:
            case MUIA_ShortHelp:
            case MUIA_Cycle_Active:
                if (data->d_Cycle) {
                   rc |= set(data->d_Cycle, tag->ti_Tag, tag->ti_Data);
                }
                break;
            default:
		rc |= SetSuperAttrs(cl, obj, tag->ti_Tag, tag->ti_Data,TAG_END);
                break;
        }
    }
    return rc;
}

static IPTR mNew(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data;

    obj = (Object *)DoSuperNew(cl, obj, TAG_DONE);
    if (obj == NULL) {
        return 0;
    }

    data = INST_DATA(cl, obj);
    memset(data, 0, sizeof(*data));
    setattrs(cl, obj, msg->ops_AttrList);

    return (IPTR)obj;
}

static IPTR mDispose(struct IClass *cl, Object *obj, Msg msg)
{
    return DoSuperMethodA(cl, obj, msg);
}

static IPTR mSet(struct IClass *cl, Object *obj, struct opSet *msg)
{
    return setattrs(cl, obj, msg->ops_AttrList);
}

static IPTR mGet(struct IClass *cl, Object *obj, struct opGet *msg)
{
    ULONG attrID = msg->opg_AttrID, rc = 0;
    IPTR *storage = (IPTR *)msg->opg_Storage;
    struct Data *data = INST_DATA(cl, obj);

    switch (attrID) {
        case MUIA_CycleChain:
        case MUIA_ControlChar:
        case MUIA_ShortHelp:
        case MUIA_Cycle_Active:
            if (data->d_Cycle) {
                rc = get(data->d_Cycle, attrID, storage);
            }
            break;
        case MUIA_Cycle_Entries:
            *storage = (IPTR)data->d_Labels;
            rc = TRUE;
            break;
        default:
            rc = DoSuperMethodA(cl, obj, (Msg)msg);
            break;
    }
    return rc;
}

DISPATCHER_PROTO
{
    switch (msg->MethodID) {
	case OM_NEW:	    return mNew(cl, obj, (APTR)msg);
	case OM_DISPOSE:    return mDispose(cl, obj, (APTR)msg);
	case OM_GET:        return mGet(cl, obj, (APTR)msg);
	case OM_SET:        return mSet(cl, obj, (APTR)msg);
    }
    return DoSuperMethodA(cl, obj, msg);
}

void NewCycleCleanup(void)
{
    if (NewCycle_mcc != NULL) {
        MUI_DeleteCustomClass(NewCycle_mcc);
    }
    NewCycle_mcc = NULL;
    NewCycleClass = NULL;
}

BOOL NewCycleInit(void)
{
    NewCycle_mcc = MUI_CreateCustomClass(NULL, (STRPTR)SUPERCLASS, NULL,
        sizeof(struct Data), Dispatcher);
    if (NewCycle_mcc == NULL) {
        NewCycleCleanup();
        return FALSE;
    }

    NewCycleClass = NewCycle_mcc->mcc_Class;
    DISPATCHER_INIT(NewCycleClass);

    return TRUE;
}

