/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** PopEditLabels_mcc.c
**
** Creation date: 26.11.1996
** Last modified: 29.12.2020
**
*/

/* Includes */
#include "system.h"
#include <intuition/icclass.h>
#include <clib/alib_protos.h>
#include <proto/utility.h>

#include "util.h"
#include "labels.h"
#include "appstrings_cat.h"
#include "PopEditLabels_mcc.h"

#define SUPERCLASS MUIC_Text

static struct MUI_CustomClass *PopEditLabels_mcc = NULL;     /* for MUI */
struct IClass *PopEditLabelsClass = NULL;                    /* for Intuition */

struct Data
{
    Object *d_Window;
    Object *d_EditLabels;
    struct MemPool *d_MemPool;
    char **d_Labels;
};

static IPTR setattrs(struct Data *data, struct TagItem *tags)
{
    struct TagItem *tag;
    IPTR rc = 0;

    while ((tag = NextTagItem(&tags)) != NULL)
    {
	switch (tag->ti_Tag)
	{
            case MUIA_EditLabels_Labels:
		if (data->d_Labels != (char **)tag->ti_Data)
		{
		    FreeLabels(data->d_MemPool, data->d_Labels);
		    data->d_Labels = CopyLabels(data->d_MemPool, 
			    (char **)tag->ti_Data);
		}
                rc = TRUE;
                break;
	}
    }
    return(rc);
}

static IPTR mNew(struct IClass *cl,Object *obj,struct opSet *msg)
{
    struct TagItem *tags = msg->ops_AttrList;
    struct Data *data;

    obj = (Object *)DoSuperNew(cl, obj,
        TAG_MORE, (IPTR)tags,
        TAG_DONE);
    if(obj == NULL)
        return(0);

    data = INST_DATA(cl, obj);
    data->d_Window = NULL;
    data->d_EditLabels = NULL;
    data->d_MemPool = CreatePool(0, 4096, 2048);
    data->d_Labels = NULL;

    setattrs(data, tags);
    DoMethod(obj, MUIM_Notify, MUIA_Pressed, FALSE,
        obj, 1, MUIM_PopEditLabels_Open);

    return (IPTR)obj;
}

static IPTR mDispose(struct IClass *cl,Object *obj,Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);

    CoerceMethod(cl, obj, MUIM_PopEditLabels_Close, FALSE);

    if (data->d_MemPool != NULL) {
        DeletePool(data->d_MemPool);
        data->d_MemPool = NULL;
    }
    return DoSuperMethodA(cl, obj, msg);
}

static IPTR mSet(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct Data *data = INST_DATA(cl, obj);
    IPTR rc;

    rc = setattrs(data, msg->ops_AttrList);
    rc |= DoSuperMethodA(cl, obj, (Msg)msg);
    return rc;
}

static IPTR mGet(struct IClass *cl, Object *obj, struct opGet *msg)
{
    struct Data *data = INST_DATA(cl, obj);

    switch(msg->opg_AttrID)
    {
        case MUIA_EditLabels_Labels:
            *msg->opg_Storage = (IPTR)data->d_Labels;
            return(TRUE);
    }

    return(DoSuperMethodA(cl, obj, (Msg)msg));
}

static IPTR mOpen(struct IClass *cl, Object *obj, Msg msg)
{
    struct Data *data = INST_DATA(cl, obj);
    Object *root, *buttons, *editlabels, *ok, *cancel;
    Object *app = pget(obj, MUIA_ApplicationObject);

    if(app == NULL)
        return(FALSE);

    if(data->d_Window)
    {
        SetAttrs(data->d_Window,
            MUIA_Window_Open, TRUE,
            MUIA_Window_Activate, TRUE,
            TAG_DONE);
        return(TRUE);
    }

    editlabels = EditLabelsObject,
        MUIA_EditLabels_Labels, data->d_Labels,
        End;

    ok = SimpleButton(GCS(MSG_Ok));
    cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, ok,
	Child, cancel,
	End;

    root = GroupObject,
	Child, editlabels,
	Child, buttons,
	End;

    data->d_Window = WindowObject,
        MUIA_Window_IsSubWindow, TRUE,
        MUIA_Window_NoMenus, TRUE,
	MUIA_Window_Title, GCS(MSG_EditLabels),
	MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'L','E','D'),
	WindowContents, root,
	End;

    if(data->d_Window == NULL)
        return(FALSE);

    data->d_EditLabels = editlabels;
    MUIAddMember(app, data->d_Window);

    set(ok, MUIA_CycleChain, TRUE);
    set(cancel, MUIA_CycleChain, TRUE);

    DoMethod(data->d_Window, MUIM_Notify, MUIA_Window_CloseRequest, TRUE,
        MUIV_Notify_Application, 5, MUIM_Application_PushMethod,
        obj, 2, MUIM_PopEditLabels_Close, FALSE);
    DoMethod(cancel, MUIM_Notify, MUIA_Pressed, FALSE,
        MUIV_Notify_Application, 5, MUIM_Application_PushMethod,
        obj, 2, MUIM_PopEditLabels_Close, FALSE);
    DoMethod(ok, MUIM_Notify, MUIA_Pressed, FALSE,
        MUIV_Notify_Application, 5, MUIM_Application_PushMethod,
        obj, 2, MUIM_PopEditLabels_Close, TRUE);

    GuiOpenWindow(data->d_Window);
    return(1);
}

static IPTR mClose(struct IClass *cl, Object *obj, struct MUIP_PopEditLabels_Close *pelc)
{
    struct Data *data = INST_DATA(cl, obj);
    Object *app = pget(obj, MUIA_ApplicationObject);

    if(app == NULL || data->d_Window == NULL)
        return(FALSE);

    if(pelc->pelc_Ok && data->d_EditLabels)
    {
        char **labels = pget(data->d_EditLabels, MUIA_EditLabels_Labels);

        set(obj, MUIA_EditLabels_Labels, labels);
    }

    GuiCloseWindow(data->d_Window);
    MUIRemMember(app, data->d_Window);
    MUI_DisposeObject(data->d_Window);
    data->d_Window = NULL;
    data->d_EditLabels = NULL;

    return(TRUE);
}

DISPATCHER_PROTO
{
    switch (msg->MethodID)
    {
	case OM_NEW:		return(mNew(cl, obj, (APTR)msg));
	case OM_DISPOSE:        return(mDispose(cl, obj, (APTR)msg));
	case OM_GET:            return(mGet(cl, obj, (APTR)msg));
	case OM_SET:            return(mSet(cl, obj, (APTR)msg));
        case MUIM_PopEditLabels_Open:   return(mOpen(cl, obj, (APTR)msg));
        case MUIM_PopEditLabels_Close:  return(mClose(cl, obj, (APTR)msg));
    }

    return(DoSuperMethodA(cl, obj, msg));
}

void PopEditLabelsCleanup(void)
{
    if(PopEditLabels_mcc != NULL)
        MUI_DeleteCustomClass(PopEditLabels_mcc);
    PopEditLabels_mcc = NULL;
    PopEditLabelsClass = NULL;
}

BOOL PopEditLabelsInit(void)
{
    PopEditLabels_mcc = MUI_CreateCustomClass(NULL, (STRPTR)SUPERCLASS, NULL,
        sizeof(struct Data), Dispatcher);
    if(PopEditLabels_mcc == NULL)
    {
        PopEditLabelsCleanup();
        return(FALSE);
    }

    PopEditLabelsClass = PopEditLabels_mcc->mcc_Class;
    DISPATCHER_INIT(PopEditLabelsClass);

    return(TRUE);
}

