/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** RefFieldSelector.c
**
** Creation date: 27.03.2011
** Last modified: 21.06.2020
**
** RefFieldSelector_mc is a subclass of Popobject class.
** It manages the selection for reference fields.
*/

#include "system.h"
#include "RefFieldSelector_mcc.h"
#include "DDList_mcc.h"
#include "../quicksearch_mui.h"
#include "buildgui.h"
#include "programeditor.h"
#include "gui.h"
#include "order.h"
#include "requesters.h"
#include "display.h"
#include "queryeditor.h"
#include "util.h"
#include "strhandle.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"
#include "amiga/font.h"
#include <proto/utility.h>

#define SUPERCLASS MUIC_Popobject

static struct MUI_CustomClass *RefFieldSelector_mcc = NULL;  /* for MUI */
struct IClass *RefFieldSelectorClass = NULL;                 /* for Intuition */

struct ClassData
{
    struct MUI_EventHandlerNode d_EHNode;
    Object *d_Listview;
};

static struct MinNode InitialRecordNode;
static struct Record *InitialRecord = (struct Record *)&InitialRecordNode;
static struct MinNode CurrentRecordNode;
static struct Record *CurrentRecord = (struct Record *)&CurrentRecordNode;

static short numColumns(const char *format)
{
    short numCols = 1;
    const char *s = format;

    if (s) {
        while ((s = strchr(s, ',')) != NULL) {
            s++;
            numCols++;
        }
    }
    return numCols;
}

static void setEmptyStrings(char **dst, short numCols)
{
    short i;

    for (i = 0; i < numCols; i++) {
        dst[i] = "";
    }
}

static LONG HOOK_PROTO(refDisplay, char **dst, struct Record *x)
{
    GuiObject listview = hook->h_Data;
    GuiObject pop = pget(listview, MUIA_UserData);
    short numCols = numColumns(pget(listview, MUIA_List_Format));
    bool multiColumn = (numCols > 1);
    struct DisplayObject *dio = pget(pop, MUIA_UserData);
    struct Field *field = DisplayFieldField(dio);
    struct Record *rec = x;
    struct RecordDescription *rd;
    static char buffer[512];
    char *buf = buffer;
    short i = 0;

    rd = AllocRecordDescription(FieldReferenceTable(field),
            DisplayFieldRecordDescriptionSpec(dio, TRUE));
    if (rd != NULL) {
        if (rec == InitialRecord || rec == CurrentRecord) {
            *buf++ = '\33';
            *buf++ = 'i';
            strcpy(buf, (rec == InitialRecord)? 
                    GCS(MSG_InitialRecord) : GCS(MSG_CurrentRecord));
            if (multiColumn) {
                setEmptyStrings(dst, numCols);
                for (i=0; i < numCols; i++) {
                    if (RecordDescriptionColumnType(rd, i) == TYPE_STRING) {
                        break;
                    }
                    if (i == numCols) {
                        i = 0;
                    }
                }
            }
            dst[i] = buffer;
	} else {
            struct Project *pr = DisplayObjectProject(dio);
            struct MemPool *memPool = ProjectMemPool(pr);
            char *cont;
            bool free = FALSE;
            long error = OK;
            short type = TYPE_UNKNOWN;
            size_t bufsize = sizeof(buffer), size;

            RecordLock(rec);
            RecordDescriptionSetRecord(rd, rec);
            if (multiColumn) {
                for (i=0; i < numCols; i++, dst++) {
                    cont = RecordDescriptionColumnText(rd, i, memPool, &free);
                    type = RecordDescriptionColumnType(rd, i);
                    error = RecordDescriptionError(rd);
                    if (error) {
                        *dst = GES(error);
                    } else if (bufsize <= 5) {
                        *dst = "";
                    } else {
                        *dst = buf;
                        if (type == TYPE_INTEGER || type == TYPE_REAL) {
                            *buf++ = '\33';
                            *buf++ = 'r';
                            bufsize -= 2;
                        }
                        *buf++ = '\33';
                        *buf++ = '-';
                        bufsize -= 2;

                        CodesetStrNBytesCopy(buf, EmptyNullStr(cont),bufsize-1);
                        size = strlen(buf) + 1;
                        buf += size;
                        bufsize -= size;
                    }
                    if (free) {
                        FreeString(memPool, cont);
                    }
                }
            } else {
                cont = RecordDescriptionText(rd, memPool, &free);
                error = RecordDescriptionError(rd);
                if (error) {
                    *dst = GES(error);
                } else {
                    *dst = buf;
                    *buf++ = '\33';
                    *buf++ = '-';
                    bufsize -= 2;
                    CodesetStrNBytesCopy(buf, EmptyNullStr(cont), bufsize-1);
                }
                if (free) {
                    FreeString(memPool, cont);
                }
            }
            RecordUnlock(rec);
        }
        FreeRecordDescription(rd);
    } else {
        setEmptyStrings(dst, numCols);
    }

    return 0;
}

static struct Record *refFieldSearchCurrent = NULL;

static void refFieldSearchFunc(GuiObject list, const char *str, int cmd, 
        void *ptr)
{
    struct DisplayObject *dio = ptr;
    GuiObject pop = DisplayFieldPopObject(dio);
    long dir = -1;

    switch (cmd) {
        case QUICK_SEARCH_SETUP:
            if (!GuiWindowOpen(GuiObjectWindow(list))) {
                DoMethod(pop, MUIM_Popstring_Open);
            }
            break;

	case QUICK_SEARCH_NEXT:  dir++;		/* no break */
	case QUICK_SEARCH_START: dir++;		/* no break */
	case QUICK_SEARCH_PREV:
            refFieldSearchCurrent = GuiReferenceFieldQuickSearchRecord(
                    dio, list, dir, str, refFieldSearchCurrent);
	    break;

	case QUICK_SEARCH_COMMIT:
            DoMethod(pop, MUIM_Popstring_Close, TRUE);
            break;

	case QUICK_SEARCH_CANCEL:
	    break;
    }
}

static bool refFieldSearchInit(GuiObject list, char c, struct DisplayObject *dio)
{
    refFieldSearchCurrent = NULL;
    return GuiQuickSearchInit(list, c, refFieldSearchFunc, dio);
}

static bool refListKeyFunc(GuiObject list, char *buf, short len,
        ushort qualifier, struct DisplayObject *dio)
{
    if (MUIIsSearchStartString(buf, len, qualifier)) {
	return refFieldSearchInit(list, buf[0], dio);
    }

    return FALSE;
}

static void setListRecords(GuiObject list, struct Element *elem, 
	struct Table *tab)
{
    if (elem) {
	while (elem->el_Type == TYPE_LIST) {
	    setListRecords(list, elem->el_Contents.list.head, tab);
	    elem = elem->el_Contents.list.tail;
	}
	if (elem->el_Type == TYPE_RECORD && elem->el_Contents.tabrec.tab==tab) {
	    DoMethod(list, MUIM_List_InsertSingle,
		    elem->el_Contents.tabrec.rec, MUIV_List_Insert_Bottom);
	}
    }
}

static LONG HOOK_PROTO(refStrObj, GuiObject list, GuiObject str)
{
    Object *pop = (Object *)GuiUserData(list);
    struct DisplayObject *dio = (struct DisplayObject *)GuiUserData(pop);
    struct Field *field = DisplayFieldField(dio);
    struct Table *tab = FieldTable(field);
    struct Table *reftab = FieldReferenceTable(field);
    struct List *refrecs = TableRecords(reftab);
    struct Record *refrec;
    char records = DisplayFieldRecordsKind(dio);

    if (refrecs == NULL) {
        return FALSE;
    }

    GuiStoreEdits();      /* maybe the user has just edited a string object */

    DoMethod(list, MUIM_List_Clear);

    if (records == DAD_RECORDS_COMPUTED) {
	const char *trigger = DisplayFieldRecordsCompute(dio);

	if (trigger != NULL && trigger[0] != 0) {
	    struct Project *pr = TableProject(tab);
	    struct Program *prg = ProjectProgram(pr);
	    struct UserFunction *uf = FindUserFunctionName(prg, trigger);
	    const char *filename = NULL, *error = NULL;
	    long line, col;
	    struct Element *elem;

	    if (uf == NULL) {
		Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Trigger), trigger);
		return FALSE;
    	    } else {
    		/* GuiSetSleep(TRUE); */
    		SetTableProgramRecords(pr);
    		ProgramBegin(prg, TRUE);
    		elem = ProgramCall(prg, uf, NULL, TRUE);
		setListRecords(list, elem, reftab);
                ProgramEnd();
		error = ProgramError(&line, &col, &filename);
		/* GuiSetSleep(FALSE); */

		if (error != NULL && error != EXECUTE_ABORT) {
		    ProgramEditorShowError(pr, TRUE, error, line, col,filename);
		    return FALSE;
		}
	    }
	}
    } else if (records == DAD_RECORDS_MATCHFILTER) {
	GuiSetSleep(TRUE);
	FORNODE (refrec, refrecs) {
	    if (RecordMatchFilter(refrec)) {
		DoMethod(list, MUIM_List_InsertSingle, refrec, 
			MUIV_List_Insert_Bottom);
	    }
	}
	GuiSetSleep(FALSE);
    } else {
	GuiSetSleep(TRUE);
	MUIListSetList(list, refrecs, TableMemPool(tab));
	GuiSetSleep(FALSE);
    }

    if (DisplayFieldCurrentRecord(dio)) {
	refrec = CurrentRecord;
	DoMethod(list, MUIM_List_InsertSingle, refrec, MUIV_List_Insert_Top);
    }
    if (DisplayFieldInitialRecord(dio)) {
	refrec = InitialRecord;
	DoMethod(list, MUIM_List_InsertSingle, refrec, MUIV_List_Insert_Top);
    }
    return TRUE;
}
static HOOK_STRUCT(refStrObj)

static void HOOK_PROTO(refObjStr, GuiObject list, GuiObject str)
{
    struct DisplayObject *dio = (struct DisplayObject *)GuiUserData(str);
    struct Field *field = DisplayFieldField(dio);
    struct Record *refrec = NULL;

    /* get active record */
    DoMethod(list, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&refrec);
    if (refrec == InitialRecord) {
	refrec = NULL;
    } else if (refrec == CurrentRecord) {
	refrec = TableGuiRecord(FieldReferenceTable(field));
    }

    /* clear listview */
    set(list, MUIA_List_Quiet, TRUE);
    DoMethod(list, MUIM_List_Clear);
    set(list, MUIA_List_Quiet, FALSE);

    /* set record reference */
    GuiStoreDataRecord(dio, refrec);
}
static HOOK_STRUCT(refObjStr)

static void HOOK_PROTO(refWindow, GuiObject list, GuiObject win)
{
    GuiObject pop = (GuiObject)GuiUserData(list);
    GuiObject string = NULL;

    get(pop, MUIA_Popstring_String, &string);
    if(string)
    {
        struct DisplayObject *dio = (struct DisplayObject *)GuiUserData(string);
        struct Field *field = DisplayFieldField(dio);
        struct Table *tab = FieldTable(field);
        struct Record *rec = TableGuiRecord(tab), *refrec = NULL;
        struct DataInfo *di = NULL, di_buf;
        long error;

        RecordLock(rec);
        error = RecordDataInfo(rec, field, &di, &di_buf);
	(void)error;
        if (di) {
            refrec = di->di_Contents.rec;
	}
        RecordUnlock(rec);
	if (refrec == NULL) {
	    refrec = InitialRecord;
	}

        set(win, MUIA_Window_Open, TRUE);

	MUIListSetActiveEntry(list, refrec);
    }
    set(win, MUIA_Window_DefaultObject, list);
}
static HOOK_STRUCT(refWindow)

static void HOOK_PROTO(refdoubleclick, GuiObject listview, 
	struct DisplayObject **x)
{
    GuiObject pop = (GuiObject)GuiUserData(listview);

    /* close popup, this calls the close hook */
    DoMethod(pop, MUIM_Popstring_Close, TRUE);
}

static Object *createListview(struct DisplayObject *dio)
{
    Object *listview;
    struct FontStyleSpec fss;
    struct TextFont *tf = FontTextFont(DisplayObjectMUIFontname(dio), &fss);
    struct MemPool *listMemPool;
    struct Hook *displayHook;
    const struct RecordDescriptionSpec *rds =
        DisplayFieldRecordDescriptionSpec(dio, TRUE);
    bool multiColumn = rds? rds->rds_MultiColumn : FALSE;
    short numCols = -1;

    listview = ListviewObject,
        MUIA_Listview_List, DDListObject,
            InputListFrame,
            tf? MUIA_Font : TAG_IGNORE, tf,
            MUIA_KeyFunc, refListKeyFunc,
            MUIA_KeyData, dio,
            End,
        End;

    listMemPool = pget(listview, MUIA_DDList_MemPool);
    displayHook = PoolAllocMemZero(listMemPool, sizeof(struct Hook));
    if (displayHook) {
        HOOK_INIT(displayHook, refDisplay);
        displayHook->h_Data = listview;
        set(listview, MUIA_List_DisplayHook, displayHook);
    }

    if (multiColumn) {
        struct Field *field = DisplayFieldField(dio);
        struct RecordDescription *rd = AllocRecordDescription(
                FieldReferenceTable(field), rds);
        if (rd) {
            numCols = RecordDescriptionNumColumns(rd);
            FreeRecordDescription(rd);
        }
        if (numCols <= 1) {
            multiColumn = FALSE;
        }
    }
    if (multiColumn && listview) {
        struct StrHandle *sh = AllocStrHandle(listMemPool);
        int i;

        for (i=1; i < numCols; i++) {
            StrHandlePutS(sh, ",");
        }
        StrHandlePutC(sh, 0);
        set(listview, MUIA_List_Format, StrHandleString(sh));
    }

    return listview;
}

static IPTR mNew(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct TagItem *tags = msg->ops_AttrList;
    struct DisplayObject *dio = (struct DisplayObject *)
        GetTagData(MUIA_UserData, 0, tags);
    Object *listview = createListview(dio);
    struct ClassData *data;

    obj = (Object *)DoSuperNew(cl, obj,
        MUIA_Popobject_StrObjHook, (IPTR)&refStrObjHook,
        MUIA_Popobject_ObjStrHook, (IPTR)&refObjStrHook,
        MUIA_Popobject_WindowHook, (IPTR)&refWindowHook,
        MUIA_Popobject_Object, (IPTR)listview,
        TAG_MORE, (IPTR)tags,
        TAG_DONE);

    if (obj == NULL) {
	return 0;
    }

    data = INST_DATA(cl, obj);
    memset(data, 0, sizeof(struct ClassData));
    data->d_Listview = listview;

    set(listview, MUIA_UserData, obj);
    MUIAddCallback(listview, MUIA_Listview_DoubleClick, TRUE, 
            refdoubleclick, dio);

    return (IPTR)obj;
}

static IPTR mSetup(struct IClass *cl, Object *obj, Msg msg)
{
    struct ClassData *data = INST_DATA(cl, obj);

    if (!DoSuperMethodA(cl, obj, msg)) {
        return FALSE;
    }

    data->d_EHNode.ehn_Object = obj;
    data->d_EHNode.ehn_Class  = cl;
    data->d_EHNode.ehn_Events = IDCMP_RAWKEY;
    DoMethod(_win(obj), MUIM_Window_AddEventHandler, &data->d_EHNode);

    return TRUE;
}

static IPTR mCleanup(struct IClass *cl, Object *obj, Msg msg)
{
    struct ClassData *data = INST_DATA(cl, obj);

    DoMethod(_win(obj), MUIM_Window_RemEventHandler, &data->d_EHNode);
    return DoSuperMethodA(cl, obj, msg);
}

static IPTR mHandleEvent(struct IClass *cl, Object *obj, 
	struct MUIP_HandleEvent *msg)
{
    struct IntuiMessage *imsg = msg->imsg;

    if (imsg && imsg->Class == IDCMP_RAWKEY && 
            !(imsg->Code & IECODE_UP_PREFIX)) {
        struct DisplayObject *dio = pget(obj, MUIA_UserData);

        if (DisplayActiveObject(_win(obj)) == dio) {
            char buffer[10];
            short len = MUIRawKeyConvert(imsg, buffer, 10);
            struct ClassData *data = INST_DATA(cl, obj);

            if (refListKeyFunc(data->d_Listview, buffer, len, 
                        imsg->Qualifier, dio)){
                return MUI_EventHandlerRC_Eat;
            }
        }
    }
    return 0;
}

static struct Record *dragDropRecord(Object *obj, Object *sfw)
{
    struct DisplayObject *dio = pget(obj, MUIA_UserData);
    struct Field *field = DisplayFieldField(dio);
    struct Table *reftab = FieldReferenceTable(field);
    struct List *recs = TableRecords(reftab);
    struct QueryLine *ql = GuiListActiveEntry(sfw);
    ushort numFields = nget(sfw, MUIA_SFWList_NumFields);
    ushort i;
    struct Record *rec;

    if (recs && ql) {
        ushort focus = nget(sfw, MUIA_Listview_ClickColumn);
        if (focus < numFields) {
            rec = ql->ql_Entries[focus]->qe_Record;
            if (FindNodePtr(recs, rec)) {
                return rec;
            }
        }
        for (i = 0; i < numFields; i++) {
            rec = ql->ql_Entries[i]->qe_Record;
            if (FindNodePtr(recs, rec)) {
                return rec;
            }
        }
    }

    return NULL;
}

static IPTR dragQuery(struct IClass *cl, Object *obj,struct MUIP_DragDrop *msg)
{
    if (dragDropRecord(obj, msg->obj)) {
        return MUIV_DragQuery_Accept;
    }

    return MUIV_DragQuery_Refuse;
}

static IPTR dragDrop(struct IClass *cl, Object *obj, struct MUIP_DragDrop *msg)
{
    struct DisplayObject *dio = pget(obj, MUIA_UserData);
    struct Record *refrec = dragDropRecord(obj, msg->obj);
    IPTR rc = 0;

    if (refrec) {
        /* set record reference */
        GuiStoreDataRecord(dio, refrec);
    } else {
	rc = DoSuperMethodA(cl, obj, (Msg)msg);
    }
    return rc;
}

DISPATCHER_PROTO
{
    switch (msg->MethodID)
    {
	case OM_NEW:		return mNew(cl, obj, (APTR)msg);
        case MUIM_Setup:        return mSetup(cl, obj, (APTR)msg);
        case MUIM_Cleanup:      return mCleanup(cl, obj, (APTR)msg);
        case MUIM_HandleEvent:  return mHandleEvent(cl, obj, (APTR)msg);
	case MUIM_DragQuery:    return dragQuery(cl, obj, (APTR)msg);
	case MUIM_DragDrop :    return dragDrop (cl, obj, (APTR)msg);
    }

    return DoSuperMethodA(cl, obj, msg);
}

void RefFieldSelectorCleanup(void)
{
    if (RefFieldSelector_mcc != NULL) {
        MUI_DeleteCustomClass(RefFieldSelector_mcc);
    }
    RefFieldSelector_mcc = NULL;
    RefFieldSelectorClass = NULL;
}

BOOL RefFieldSelectorInit(void)
{
    RefFieldSelector_mcc = MUI_CreateCustomClass(NULL, (STRPTR)SUPERCLASS, NULL,
        sizeof(struct ClassData), Dispatcher);
    if (RefFieldSelector_mcc == NULL) {
        RefFieldSelectorCleanup();
        return FALSE;
    }

    RefFieldSelectorClass = RefFieldSelector_mcc->mcc_Class;
    DISPATCHER_INIT(RefFieldSelectorClass);

    return TRUE;
}

