/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
**
** Creation date: 30.10.1996
** Last modified: 11.05.2022
**
** SFWList.mcc is a subclass of List (or NList if USE_NLIST is defined) class.
** Its purpose is for displaying the contents of a program element 
** (usually the result of a Select-From-Where query).
** It also attaches a context menu to enable further operations.
*/

/* Includes */
#include "SFWList_mcc.h"
#include "../quicksearch_mui.h"
#include "field.h"
#include "queryeditor.h"
#include "lang/program.h"
#include "imexport.h"
#include "pdfexport.h"
#include "print.h"
#include "memory.h"
#include "util.h"
#include "appstrings_cat.h"
#include <intuition/icclass.h>
#include <clib/alib_protos.h>
#include <proto/utility.h>

#ifdef USE_NLIST
#define SUPERCLASS MUIC_NList
#else
#define SUPERCLASS MUIC_List
#endif

#define USE_DISPLAY_HOOK

#ifndef USE_NLIST
#define LIST_TITLE_LEN 63

static const char sort1_normal[] = "\33I[6:39]";
static const char sort1_invert[] = "\33I[6:38]";
static const char sort2_normal[] = "\33I[1:1]";
static const char sort2_invert[] = "\33I[1:0]";
#endif

static struct MUI_CustomClass *SFWList_mcc = NULL;   /* for MUI */
struct IClass *SFWListClass = NULL;                  /* for Intuition */

#define SFW_MAX_FIELDS 63

struct ClassData
{
    struct QueryGuiData *d_Data;
    bool d_ShowTitles;
    bool d_Bars;
    bool d_CheckMarks;
    char d_FormatStr[SFW_MAX_FIELDS * 11 + 10 + 1];
    Object *d_CheckMarkNormal;
    Object *d_CheckMarkSelected;
    IPTR d_CheckMarkNormalImage;
    IPTR d_CheckMarkSelectedImage;
    char d_CheckMarkNormalStr[24];
    char d_CheckMarkSelectedStr[24];
    Object *d_Menustrip;
    Object *d_Export;
#ifdef USE_HPDF
    Object *d_PDF;
#endif
    Object *d_Print;
#ifdef USE_DISPLAY_HOOK
    struct Hook d_DisplayHook;
#endif
#ifndef USE_NLIST
    short d_SortCol1;
    short d_SortCol2;
    bool d_SortInvert1;
    bool d_SortInvert2;
    char d_SortTitle1[LIST_TITLE_LEN+1];
    char d_SortTitle2[LIST_TITLE_LEN+1];
    char d_ColumnOrder[SFW_MAX_FIELDS + 1];
#endif
    struct MUI_EventHandlerNode d_EHNode;
    LONG d_MoveDrop;
};

static void setattrs(struct IClass *cl, Object *obj, struct TagItem *tags)
{
    struct ClassData *data = INST_DATA(cl, obj);
    struct TagItem *tag;

    while ((tag = NextTagItem(&tags)) != NULL) {
	switch (tag->ti_Tag) {
            case MUIA_SFWList_ActiveRecord:
            {
                ushort ci = data->d_Data? data->d_Data->qgd_ControlIndex : 0;
                struct Record *rec = (struct Record *)tag->ti_Data;

                if (data->d_Data != NULL && 
                        data->d_Data->qgd_ControlTable != NULL && 
                        ci < data->d_Data->qgd_NumFields) {
                    long lineNum = 0;
                    long pos;
                    struct QueryLine *ql;

                    MUIListLoop(obj, pos, ql, TRUE) {
                        if (ql->ql_Entries[ci]->qe_Record == rec) {
                            lineNum = ql->ql_LineNum;
                            break;
                        }
                    }
                    set(obj, MUIA_SFWList_ActiveLine, lineNum);
                }
                break;
            }
            case MUIA_SFWList_ActiveLine:
            {
                long lineNum = tag->ti_Data;
                long pos;
                struct QueryLine *ql;
                bool found = FALSE;

                MUIListLoop(obj, pos, ql, TRUE) {
                    if (ql->ql_LineNum == lineNum) {
                        set(obj, MUIA_List_Active, pos);
                        found = TRUE;
                        break;
                    }
                }
                if (!found) {
                    set(obj, MUIA_List_Active, MUIV_List_Active_Off);
                }
                break;
            }
        }
    }
}

static char *makeDisplayStr(struct ClassData *data, bool header, int col,
        char *text)
{
    char *s = text;

#ifndef USE_NLIST
    if (header) {
        const char *append = NULL;
        char *buf = NULL;

        if (col == data->d_SortCol1) {
            append = data->d_SortInvert1? sort1_normal : sort1_invert;
            buf = data->d_SortTitle1;
        } else if (col == data->d_SortCol2) {
            append = data->d_SortInvert2? sort2_normal : sort2_invert;
            buf = data->d_SortTitle2;
        }
        if (append) {
            sprintf(buf, "%.*s %s",
                    (int)(LIST_TITLE_LEN - 1 - strlen(append)), s, append);
		s = buf; 
	    }
        }
#endif
    return s;
}

#ifdef USE_DISPLAY_HOOK
static ulong HOOK_PROTO(display, const char **dst, struct QueryLine *ql)
{
    struct ClassData *data = hook->h_Data;
#else
static IPTR mDisplay(struct IClass *cl, Object *obj, 
#ifdef USE_NLIST
        struct MUIP_NList_Display *msg)
#else
        struct MUIP_List_Display *msg)
#endif
{
    struct ClassData *data = INST_DATA(cl, obj);
    const struct QueryLine *ql = (const struct QueryLine *)msg->entry;
    STRPTR *dst = msg->strings;
#endif
    struct QueryEntry **src;
    int numFields = data->d_Data? data->d_Data->qgd_NumFields : 0;
    bool header = (ql == NULL);
    int addCol = 0;
    int i;

    if (numFields == 0) {
        *dst = "";
    }
    if (ql == NULL) {
        ql = data->d_Data? data->d_Data->qgd_Titles : NULL;
    }

    if (data->d_CheckMarks) {
        *dst++ = makeDisplayStr(data, header, 0,
                header? "" : (ql->ql_Checked ?
                    data->d_CheckMarkSelectedStr : data->d_CheckMarkNormalStr));
        addCol = 1;
    }

    src = ql? ql->ql_Entries : NULL;
    for (i = 0; i < numFields; i++, dst++) {
        *dst = makeDisplayStr(data, header, i + addCol,
                (src && src[i])? (char *)EmptyNullStr(src[i]->qe_String) : "");
    }

    return 0;
}

static long numSelected(Object *obj)
{
    long checked = 0;
    long pos;
    struct QueryLine *ql;

    MUIListLoop(obj, pos, ql, TRUE) {
        if (ql != NULL && ql->ql_Checked) {
            checked++;
        }
    }

    return checked;
}

static void HOOK_PROTO(checkMarkClick, Object *obj, void *dummy)
{
    struct QueryLine *ql = NULL;

    if (nget(obj, MUIA_List_ClickColumn) == 0) {
        DoMethod(obj, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, &ql);
        if (ql != NULL) {
            ql->ql_Checked = !ql->ql_Checked;
        }
        DoMethod(obj, MUIM_List_Redraw, MUIV_List_Redraw_Active);
        set(obj, MUIA_SFWList_NumSelected, numSelected(obj));
    }
}

#ifndef USE_NLIST
static void HOOK_PROTO(titleclick, Object *obj, LONG *x)
{
    LONG col = *x;
    struct ClassData *data = INST_DATA(SFWListClass, obj);
    GuiObject win = GuiObjectWindow(obj);
    UWORD qualifier = nget(win, MUIA_Qualifier);
    int shift = (qualifier & (IEQUALIFIER_LSHIFT | IEQUALIFIER_RSHIFT));

    if (data->d_SortCol1 >= 0 && col != data->d_SortCol1 && shift) {
        if (data->d_SortCol2 == col) {
            data->d_SortInvert2 = !data->d_SortInvert2;
        } else {
            data->d_SortCol2 = col;
            data->d_SortInvert2 = FALSE;
        }
    } else {
        if (data->d_SortCol1 == col) {
            data->d_SortInvert1 = !data->d_SortInvert1;
        } else {
            data->d_SortCol1 = col;
            data->d_SortInvert1 = FALSE;
        }
        data->d_SortCol2 = col;
        data->d_SortInvert2 = FALSE;
    }
    DoMethod(obj, MUIM_List_Sort);
}
#endif

static void HOOK_PROTO(menuExportPDFPrint, Object *obj,
        void (**x)(struct Project *pr, Object *sfw))
{
    void (*func)(struct Project *pr, Object *sfw) = *x;

    if (func) {
        func(GuiObjectProject(obj), obj);
    }
}

static void updateMenustrip(Object *obj, struct ClassData *data)
{
    Object *menustrip = data->d_Menustrip;

    if (data->d_Data == NULL || data->d_Data->qgd_NumFields == 0 || 
            nget(obj, MUIA_Disabled)) {
        menustrip = NULL;
    }
    set(obj, MUIA_ContextMenu, menustrip);
}

static void HOOK_PROTO(onDisabled, Object *obj, struct ClassData **x)
{
    updateMenustrip(obj, *x);
}

static void HOOK_PROTO(onMoveDrop, Object *obj, struct ClassData **x)
{
    struct ClassData *data = *x;
    long pos = nget(obj, MUIA_SFWList_MoveDrop) - 1, i;
    struct QueryLine *ql = NULL;
    DoMethod(obj, MUIM_List_GetEntry, pos, &ql);

    if (ql != NULL && data->d_Data) {
        for (i = 0; i < data->d_Data->qgd_NumLines; i++) {
            if (data->d_Data->qgd_Lines[i] == ql) {
                if (pos != i) {
                    if (pos > i) {
                        pos++;
                    }
                    QueryGuiDataMoveLine(data->d_Data, i, pos);
                }
                break;
            }
        }
    }
}

static IPTR mNew(struct IClass *cl, Object *obj, struct opSet *msg)
{
    struct ClassData *data;
    struct TagItem *tags = msg->ops_AttrList;

    obj = (Object *)DoSuperMethodA(cl, obj, (Msg)msg);
    if (obj == NULL) {
	return 0;
    }

    data = INST_DATA(cl, obj);
    memset(data, 0, sizeof(*data));
    data->d_ShowTitles = GetTagData(MUIA_SFWList_ShowTitles, TRUE, tags);
    data->d_Bars = GetTagData(MUIA_SFWList_SeparatorBars, TRUE, tags);
    data->d_CheckMarks = GetTagData(MUIA_SFWList_CheckMarks, FALSE, tags);

    setattrs(cl, obj, tags);
    SetAttrs(obj, MUIA_List_Title, FALSE,
                MUIA_List_DragSortable, TRUE,
                NULL);

    if (data->d_CheckMarks) {
        data->d_CheckMarkNormal = ImageObject,
		ImageButtonFrame,
		MUIA_InputMode, MUIV_InputMode_None,
		MUIA_Image_Spec, MUII_CheckMark,
		MUIA_Image_FreeVert, TRUE,
		MUIA_Selected, FALSE,
		MUIA_Background, MUII_ButtonBack,
		MUIA_ShowSelState, FALSE,
		End;
        data->d_CheckMarkSelected = ImageObject,
		ImageButtonFrame,
		MUIA_InputMode, MUIV_InputMode_None,
		MUIA_Image_Spec, MUII_CheckMark,
		MUIA_Image_FreeVert, TRUE,
		MUIA_Selected, TRUE,
		MUIA_Background, MUII_ButtonBack,
		MUIA_ShowSelState, FALSE,
		End;
#ifdef USE_NLIST
        MUIAddCallback(obj, MUIA_NList_EntryClick, MUIV_EveryTime,
                checkMarkClick, 0);
#else
        MUIAddPushCallback(obj, MUIA_List_ClickColumn, MUIV_EveryTime,
                checkMarkClick, 0);
#endif
    }

    data->d_Menustrip = MenustripObject,
        MUIA_Family_Child, MenuObject,
            MUIA_Menu_Title, GCS(MSG_Menu_SFWList),
            MUIA_Family_Child, data->d_Export = MenuitemObject,
                MUIA_Menuitem_Title, GCS(MSG_MenuItem_ExportAsText),
                End,
#ifdef USE_HPDF
            MUIA_Family_Child, data->d_PDF = MenuitemObject,
                MUIA_Menuitem_Title, GCS(MSG_MenuItem_ExportAsPDF),
                End,
#endif
            MUIA_Family_Child, data->d_Print = MenuitemObject,
                MUIA_Menuitem_Title, GCS(MSG_Query_Print),
                End,
            End,
        End;
    if (data->d_Menustrip == NULL) {
        CoerceMethod(cl, obj, OM_DISPOSE);
	return 0;
    }

    MUIAddCallback2(data->d_Export, MUIA_Menuitem_Trigger, MUIV_EveryTime,
            obj, menuExportPDFPrint, GuiExportSFWList);
#ifdef USE_HPDF
    MUIAddCallback2(data->d_PDF, MUIA_Menuitem_Trigger, MUIV_EveryTime,
            obj, menuExportPDFPrint, GuiExportSFWListAsPDF);
#endif
    MUIAddCallback2(data->d_Print, MUIA_Menuitem_Trigger, MUIV_EveryTime,
            obj, menuExportPDFPrint, GuiPrintSFWList);

    MUIAddCallback(obj, MUIA_Disabled, MUIV_EveryTime, onDisabled, data);
    updateMenustrip(obj, data);

#ifdef USE_DISPLAY_HOOK
	HOOK_INIT(&data->d_DisplayHook, display);
	data->d_DisplayHook.h_Data = data;
	set(obj, MUIA_List_DisplayHook, &data->d_DisplayHook);
#endif

#ifdef USE_NLIST
    DoMethod(obj, MUIM_Notify, MUIA_NList_TitleClick, MUIV_EveryTime,
            obj, 4, MUIM_NList_Sort3, MUIV_TriggerValue, 
            MUIV_NList_SortTypeAdd_2Values, MUIV_NList_Sort3_SortType_Both);
    DoMethod(obj, MUIM_Notify, MUIA_NList_TitleClick2, MUIV_EveryTime,
            obj, 4, MUIM_NList_Sort3, MUIV_TriggerValue, 
            MUIV_NList_SortTypeAdd_2Values, MUIV_NList_Sort3_SortType_2);
    DoMethod(obj, MUIM_Notify, MUIA_NList_SortType, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_NList_TitleMark, MUIV_TriggerValue);
    DoMethod(obj, MUIM_Notify, MUIA_NList_SortType2, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_NList_TitleMark2, MUIV_TriggerValue);
#else
    data->d_SortCol1 = data->d_SortCol2 = -1;
    MUIAddCallback(obj, MUIA_List_TitleClick, MUIV_EveryTime,
            titleclick, MUIV_TriggerValue);
#endif
    MUIAddCallback(obj, MUIA_SFWList_MoveDrop, MUIV_EveryTime, onMoveDrop,data);

    return (IPTR)obj;
}

static IPTR mDispose(struct IClass *cl, Object *obj, Msg msg)
{
    struct ClassData *data = INST_DATA(cl, obj);

    if (data->d_Menustrip != NULL) {
        MUI_DisposeObject(data->d_Menustrip);
        data->d_Menustrip = NULL;
    }
    if (data->d_CheckMarkSelected) {
        MUI_DisposeObject(data->d_CheckMarkSelected);
        data->d_CheckMarkSelected = NULL;
    }
    if (data->d_CheckMarkNormal) {
        MUI_DisposeObject(data->d_CheckMarkNormal);
        data->d_CheckMarkNormal = NULL;
    }

    FreeQueryGuiData(data->d_Data);
    data->d_Data = NULL;
    return DoSuperMethodA(cl, obj, msg);
}

static IPTR mSet(struct IClass *cl, Object *obj, struct opSet *msg)
{
    setattrs(cl, obj, msg->ops_AttrList);
    return DoSuperMethodA(cl, obj, (Msg)msg);
}

static IPTR mGet(struct IClass *cl, Object *obj, struct opGet *msg)
{
    struct ClassData *data = INST_DATA(cl, obj);

    switch (msg->opg_AttrID)
    {
        case MUIA_SFWList_NumFields:
            *msg->opg_Storage = data->d_Data? data->d_Data->qgd_NumFields : 0;
            return TRUE;

        case MUIA_SFWList_FieldInfos:
            *msg->opg_Storage =
                (IPTR)(data->d_Data? data->d_Data->qgd_FieldInfos : NULL);
            return TRUE;

        case MUIA_SFWList_Titles:
            *msg->opg_Storage =
                (IPTR)(data->d_Data? data->d_Data->qgd_Titles : NULL);
            return TRUE;

        case MUIA_SFWList_NumSelected:
            *msg->opg_Storage = numSelected(obj);
            return TRUE;

        case MUIA_SFWList_ActiveRecord:
        {
            ushort ci = data->d_Data? data->d_Data->qgd_ControlIndex : 0;
            struct Record *rec = NULL;
            struct QueryLine *ql = NULL;

            if (data->d_Data != NULL && 
                    data->d_Data->qgd_ControlTable != NULL &&
                    ci < data->d_Data->qgd_NumFields) {
                DoMethod(obj, MUIM_List_GetEntry,MUIV_List_GetEntry_Active,&ql);
                if (ql) {
                    rec = ql->ql_Entries[ci]->qe_Record;
                }
            }
            *msg->opg_Storage = (IPTR)rec;
            return TRUE;
        }

        case MUIA_SFWList_ActiveLine:
        {
            struct QueryLine *ql = NULL;

            DoMethod(obj, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, &ql);
            *msg->opg_Storage = ql? ql->ql_LineNum : 0;
            return TRUE;
        }

        case MUIA_SFWList_ColumnOrder:
        {
#ifdef USE_NLIST
            *msg->opg_Storage = (IPTR)pget(obj, MUIA_NList_Columns);
#else
            const char *order = pget(obj, MUIA_List_ColumnOrder);
            int numFields = data->d_Data? data->d_Data->qgd_NumFields : 0;
            int i;
            memcpy(data->d_ColumnOrder, order, numFields * sizeof(char));

#ifdef MUIM_List_QueryColumnWidth
            for (i = 0; i < numFields; i++) {
                if ((long)DoMethod(obj, MUIM_List_QueryColumnWidth, i) < 0) {
                    data->d_ColumnOrder[i] = -1;  // Column is hidden.
                }
            }
            *msg->opg_Storage = (IPTR)data->d_ColumnOrder;
#endif
#endif
            return TRUE;
        }

        case MUIA_SFWList_ControlTable:
            *msg->opg_Storage =
                (IPTR)(data->d_Data? data->d_Data->qgd_ControlTable : NULL);
            return TRUE;

        case MUIA_SFWList_MoveDrop:
            *msg->opg_Storage = data->d_MoveDrop;
            return TRUE;
    }

    return DoSuperMethodA(cl, obj, (Msg)msg);
}

static IPTR mSetup(struct IClass *cl, Object *obj, Msg msg)
{
    struct ClassData *data = INST_DATA(cl, obj);

    if (!DoSuperMethodA(cl, obj, msg)) {
        return FALSE;
    }

    data->d_EHNode.ehn_Object = obj;
    data->d_EHNode.ehn_Class  = cl;
    data->d_EHNode.ehn_Events = IDCMP_RAWKEY;
    DoMethod(_win(obj), MUIM_Window_AddEventHandler, &data->d_EHNode);

    if (data->d_CheckMarks) {
        data->d_CheckMarkNormalImage =
            DoMethod(obj, MUIM_MList_CreateImage, data->d_CheckMarkNormal, 0L);
        data->d_CheckMarkSelectedImage =
            DoMethod(obj, MUIM_MList_CreateImage, data->d_CheckMarkSelected,0L);
        sprintf(data->d_CheckMarkNormalStr, "\33O[%08lx]",
                data->d_CheckMarkNormalImage);
        sprintf(data->d_CheckMarkSelectedStr, "\33O[%08lx]",
                data->d_CheckMarkSelectedImage);
    }

    return TRUE;
}

static IPTR mCleanup(struct IClass *cl, Object *obj, Msg msg)
{
    struct ClassData *data = INST_DATA(cl, obj);

    DoMethod(obj, MUIM_MList_DeleteImage, data->d_CheckMarkSelectedImage);
    data->d_CheckMarkSelectedImage = 0;
    DoMethod(obj, MUIM_MList_DeleteImage, data->d_CheckMarkNormalImage);
    data->d_CheckMarkNormalImage = 0;

    DoMethod(_win(obj), MUIM_Window_RemEventHandler, &data->d_EHNode);
    return DoSuperMethodA(cl, obj, msg);
}

static void sfwListSearchEntry(Object *obj, long dir, const char *str)
{
    struct ClassData *data = INST_DATA(SFWListClass, obj);
    struct QueryLine *ql = NULL;
    long col = nget(obj, MUIA_Listview_ClickColumn);
    long pos = nget(obj, MUIA_List_Active);
    bool clearActive = FALSE;
    struct MemPool *memPool = data->d_Data? data->d_Data->qgd_MemPool : NULL;
    char *search = PoolAllocMem(memPool, 1 + strlen(str) + 1 + 1);
    int numFields = data->d_Data? data->d_Data->qgd_NumFields : 0;

    if (search == NULL) {
        return;
    }
    sprintf(search, "*%s*", str);

    if (col < 0 || col >= numFields) {
        col = 0;
    }

    if (dir == 0 || pos == MUIV_List_Active_Off) {
        pos = 0;
        dir = 1;
        clearActive = TRUE;
    } else {
        pos = pos + dir;
    }

    while (pos >= 0) {
        DoMethod(obj, MUIM_List_GetEntry, pos, &ql);
        if (ql == NULL) {
            break;
        } else if (StrJokerCmp(ql->ql_Entries[col]->qe_String, search)) {
            set(obj, MUIA_List_Active, pos);
            clearActive = FALSE;
            break;
        }
        pos += dir;
    }
    FreeString(memPool, search);
    if (clearActive) {
        set(obj, MUIA_List_Active, MUIV_List_Active_Off);
    }
}

static void sfwListSearchFunc(Object *obj, const char *str, int cmd, void *ptr)
{
    long dir = -1;

    switch (cmd) {
        case QUICK_SEARCH_SETUP:
            break;

	case QUICK_SEARCH_NEXT:  dir++;		/* no break */
	case QUICK_SEARCH_START: dir++;		/* no break */
	case QUICK_SEARCH_PREV:
            sfwListSearchEntry(obj, dir, str);
	    break;

	case QUICK_SEARCH_COMMIT:
            break;

	case QUICK_SEARCH_CANCEL:
	    break;
    }
}

static bool sfwListSearchInit(Object *obj, char c)
{
    struct ClassData *data = INST_DATA(SFWListClass, obj);
    int numFields = data->d_Data? data->d_Data->qgd_NumFields : 0;
    bool rc = FALSE;

    if (numFields > 0 && nget(obj, MUIA_List_Entries) > 0) {
        rc = GuiQuickSearchInit(obj, c, sfwListSearchFunc, obj);
    }
    return rc;
}

static bool sfwListKeyFunc(Object *obj, char *buf, short len, ushort qualifier)
{
    if (MUIIsSearchStartString(buf, len, qualifier)) {
	return sfwListSearchInit(obj, buf[0]);
    }

    return FALSE;
}

static IPTR mHandleEvent(struct IClass *cl, Object *obj, 
        struct MUIP_HandleEvent *msg)
{
    struct IntuiMessage *imsg = msg->imsg;

    if (imsg != NULL && imsg->Class == IDCMP_RAWKEY && 
            !(imsg->Code & IECODE_UP_PREFIX)) {
        char buffer[10];
        short len = MUIRawKeyConvert(imsg, buffer, 10);

        if (sfwListKeyFunc(obj, buffer, len, imsg->Qualifier)) {
            return MUI_EventHandlerRC_Eat;
        }
    }
    return 0;
}

static LONG queryLineCmp(const struct QueryLine *ql1,
        const struct QueryLine *ql2,
        const struct QueryFieldInfo *infos, int col, bool invert)
{
    LONG result;

    if (col < 0) {
        result = ql1->ql_Checked - ql2->ql_Checked;
    } else {
        result = QueryEntryCmp(ql1->ql_Entries[col], ql2->ql_Entries[col],
                infos[col].qfi_Type, TRUE);
    }
    if (invert) {
        result = -result;
    }
    return result;
}

static LONG mCompare(struct IClass *cl, Object *obj,
#ifdef USE_NLIST
        struct MUIP_NList_Compare *msg)
#else
        struct MUIP_List_Compare *msg)
#endif
{
    struct ClassData *data = INST_DATA(cl, obj);
    struct QueryLine *ql1 = (struct QueryLine *)msg->entry1;
    struct QueryLine *ql2 = (struct QueryLine *)msg->entry2;
#ifdef USE_NLIST
    bool sort = (msg->sort_type1 != MUIV_NList_SortType_None);
    bool invert1 = (msg->sort_type1 & MUIV_NList_TitleMark_TypeMask) != 0;
    bool invert2 = (msg->sort_type2 & MUIV_NList_TitleMark_TypeMask) != 0;
    int col1 = msg->sort_type1 & MUIV_NList_TitleMark_ColMask;
    int col2 = msg->sort_type2 & MUIV_NList_TitleMark2_ColMask;
#else
    bool sort = (data->d_SortCol1 >= 0);
    bool invert1 = data->d_SortInvert1;
    bool invert2 = data->d_SortInvert2;
    int col1 = data->d_SortCol1;
    int col2 = data->d_SortCol2;
#endif

    LONG result = 0;

    if (sort) {
        if (data->d_CheckMarks) {
            col1--;
            col2--;
        }
        result = queryLineCmp(ql1, ql2, data->d_Data->qgd_FieldInfos,
                col1, invert1);
        if (result == 0 && col1 != col2) {
            result = queryLineCmp(ql1, ql2, data->d_Data->qgd_FieldInfos,
                    col2, invert2);
        }
    }

    return result;
}

static void sfwListClear(struct IClass *cl, Object *obj)
{
    struct ClassData *data = INST_DATA(cl, obj);

    DoMethod(obj, MUIM_List_Clear);
    SetAttrs(obj,
            MUIA_List_Title, FALSE,
#ifdef USE_NLIST
            MUIA_NList_TitleMark, MUIV_NList_TitleMark_None,
            MUIA_NList_TitleMark2, MUIV_NList_TitleMark_None,
#endif
            MUIA_List_Format, "",
            TAG_END);
    FreeQueryGuiData(data->d_Data);
    data->d_Data = NULL;
#ifndef USE_NLIST
    data->d_SortCol1 = data->d_SortCol2 = -1;
#endif
    updateMenustrip(obj, data);
}

static void getformat(char *format, bool checkMarks, int numFields,
        const struct QueryFieldInfo *infos, bool bars)
{
    int i;

    format[0] = 0;
    if (checkMarks && numFields > 0) {
        strcat(format, ",");
    }
    for (i = 0; i < numFields; i++) {
        if (infos[i].qfi_Format == 0) {
            strcat(format, "P=\33l");
        } else {
            strcat(format, "P=\33r");
        }
        if (i < numFields - 1) {
            if (bars) {
                strcat(format, " BAR");
            }
            strcat(format, ",");
        } else if (infos[i].qfi_Format != 0) {
#ifdef USE_NLIST
            strcat(format, " NB W=-2");
#endif
        }
    }
}

static IPTR sfwListSetTitlesBodyHelp(struct IClass *cl, Object *obj, 
        struct MUIP_SFWList_SetTitlesBody *sfwlstb)
{
    struct ClassData *data = INST_DATA(cl, obj);
    const struct Element *titles = sfwlstb->sfwl_Titles; 
    const struct Element *body = sfwlstb->sfwl_Body;
    struct QueryGuiData *qgd = NULL;
    int maxFields = SFW_MAX_FIELDS;

    if (data->d_CheckMarks) {
        maxFields--;
    }

    qgd = QueryGuiDataFromTitlesBody(titles, body);
    if (qgd == NULL) {
        sfwListClear(cl, obj);
        return (IPTR)GCS(MSG_Err_Mem);
    }

    /*
    ** Restrict number of fields to maxFields.
    */
    if (qgd->qgd_NumFields > maxFields) {
        qgd->qgd_NumFields = maxFields;
    }

    if (QueryGuiDataCopyLineNumIfMatch(data->d_Data, qgd)) {
        FreeQueryGuiData(qgd);
        return 0;
    }

    /* Set new QueryGuiData */
    sfwListClear(cl, obj);
    data->d_Data = qgd;

    /*
    ** Set titles before adding rows.  This seems to be a requirement
    ** for List.mui as otherwise weired gfx artifacts show up.
    */
    set(obj, MUIA_List_Title, data->d_ShowTitles);

    DoMethod(obj, MUIM_List_Insert, data->d_Data->qgd_Lines,
            data->d_Data->qgd_NumLines, MUIV_List_Insert_Bottom);

    getformat(data->d_FormatStr, data->d_CheckMarks,
            data->d_Data->qgd_NumFields, data->d_Data->qgd_FieldInfos,
            data->d_Bars);
    set(obj, MUIA_List_Format, data->d_FormatStr);
    updateMenustrip(obj, data);

    return 0;
}

static IPTR sfwListGetRecord(struct IClass *cl, Object *obj, 
        struct MUIP_SFWList_GetRecord *sfwlgr)
{
    struct ClassData *data = INST_DATA(cl, obj);
    ushort ci = data->d_Data? data->d_Data->qgd_ControlIndex : 0;
    struct QueryLine *ql = NULL;
    struct Record *rec = NULL;
    long row = sfwlgr->sfwl_Row;

    if (data->d_Data != NULL && data->d_Data->qgd_ControlTable != NULL &&
            ci < data->d_Data->qgd_NumFields) {
        if (row == 0) {
            ql = data->d_Data->qgd_Titles;
        } else {
            DoMethod(obj, MUIM_List_GetEntry, row - 1, &ql);
        }
        if (ql) {
            rec = ql->ql_Entries[ci]->qe_Record;
        }
    }

    return (IPTR)rec;
}

static IPTR sfwListSetTitlesBody(struct IClass *cl, Object *obj, 
        struct MUIP_SFWList_SetTitlesBody *sfwlstb)
{
    IPTR result;

    set(obj, MUIA_List_Quiet, TRUE);
    result = sfwListSetTitlesBodyHelp(cl, obj, sfwlstb);
    set(obj, MUIA_List_Quiet, FALSE);
    set(obj, MUIA_SFWList_NumSelected, numSelected(obj));

    return result;
}

static IPTR sfwListSetElement(struct IClass *cl, Object *obj, 
        struct MUIP_SFWList_SetElement *sfwlse)
{
    const struct Element *elem = sfwlse->sfwl_Element;
    const struct Element *titles = NULL, *body = NULL;

    if (elem) {
        if (elem->el_Type == TYPE_LIST) {
            titles = elem->el_Contents.list.head;
            body = elem->el_Contents.list.tail;
        } else {
            body = elem;
        }
    }

    return DoMethod(obj, MUIM_SFWList_SetTitlesBody, titles, body);
}

static IPTR sfwListSetString(struct IClass *cl, Object *obj,
        struct MUIP_SFWList_SetString *sfwlss)
{
    struct ClassData *data = INST_DATA(cl, obj);
    IPTR result = 0;

    set(obj, MUIA_List_Quiet, TRUE);
    sfwListClear(cl, obj);
    data->d_Data = QueryGuiDataFromString(sfwlss->sfwl_String);
    if (data->d_Data == NULL) {
        result = (IPTR)GCS(MSG_Err_Mem);
    } else {
        /*
        ** For some reasons MUI_List does not like to set title and format in
        ** one SetAttrs call, thus we split it into two.
        */
        set(obj, MUIA_List_Title, TRUE);
        set(obj, MUIA_List_Format, "");
    }
    set(obj, MUIA_List_Quiet, FALSE);
    set(obj, MUIA_SFWList_NumSelected, numSelected(obj));

    return result;
}

static IPTR sfwListSetSelected(struct IClass *cl, Object *obj,
        struct MUIP_SFWList_SetSelected *sfwl)
{
    const bool *selected = sfwl->sfwl_Selected;
    struct ClassData *data = INST_DATA(cl, obj);
    struct QueryGuiData *qgd = data->d_Data;
    size_t i;

    if (qgd != NULL && selected != NULL) {
        for (i = 0; i < qgd->qgd_NumLines; i++) {
            qgd->qgd_Lines[i]->ql_Checked = selected[i];
        }
        DoMethod(obj, MUIM_List_Redraw, MUIV_List_Redraw_All);
        set(obj, MUIA_SFWList_NumSelected, numSelected(obj));
    }

    return 0;
}

static IPTR sfwListGetSelected(struct IClass *cl, Object *obj,
        struct MUIP_SFWList_GetSelected *sfwl)
{
    bool *selected = sfwl->sfwl_Selected;
    struct ClassData *data = INST_DATA(cl, obj);
    struct QueryGuiData *qgd = data->d_Data;
    size_t i;

    if (qgd != NULL && selected != NULL) {
        for (i = 0; i < qgd->qgd_NumLines; i++) {
            selected[i] = qgd->qgd_Lines[i]->ql_Checked;
        }
    }

    return 0;
}

static IPTR sfwListRemRecord(struct IClass *cl, Object *obj,
        struct MUIP_SFWList_RemRecord *sfwlrm)
{
    struct ClassData *data = INST_DATA(cl, obj);
    const struct Record *rec = sfwlrm->sfwl_Record;
    struct QueryLine *ql;
    struct QueryEntry **qes;
    long pos;
    int i;

    if (data->d_Data != NULL && data->d_Data->qgd_Titles != NULL) {
        qes = data->d_Data->qgd_Titles->ql_Entries;
        for (i = 0; i < data->d_Data->qgd_NumFields; i++) {
            if (qes[i]->qe_Record == rec) {
                qes[i]->qe_Record = NULL;
            }
        }
    }

    MUIListLoop(obj, pos, ql, TRUE) {
        qes = ql->ql_Entries;
        for (i = 0; i < data->d_Data->qgd_NumFields; i++) {
            if (qes[i]->qe_Record == rec) {
                qes[i]->qe_Record = NULL;
            }
        }
    }

    return 0;
}

static IPTR sfwListRemAllRecords(struct IClass *cl, Object *obj, APTR msg)
{
    struct ClassData *data = INST_DATA(cl, obj);
    struct QueryLine *ql;
    struct QueryEntry **qes;
    long pos;
    int i;

    if (data->d_Data != NULL && data->d_Data->qgd_Titles != NULL) {
        qes = data->d_Data->qgd_Titles->ql_Entries;
        for (i = 0; i < data->d_Data->qgd_NumFields; i++) {
            qes[i]->qe_Record = NULL;
        }
    }

    MUIListLoop(obj, pos, ql, TRUE) {
        qes = ql->ql_Entries;
        for (i = 0; i < data->d_Data->qgd_NumFields; i++) {
            qes[i]->qe_Record = NULL;
        }
    }

    return 0;
}

static IPTR sfwListDragDrop(struct IClass *cl, Object *obj, 
        struct MUIP_DragDrop *msg)
{
    struct ClassData *data = INST_DATA(cl, obj);
    LONG active = nget(obj, MUIA_List_Active);
    IPTR result = DoSuperMethodA(cl, obj, (Msg)msg);
    LONG drop = nget(obj, MUIA_List_DropMark);

    if (drop > active) {
        drop--;
    }

    /* First row is 1. */
    data->d_MoveDrop = drop + 1;
    set(obj, MUIA_SFWList_MoveDrop, data->d_MoveDrop);

    return result;
}

DISPATCHER_PROTO
{
    switch (msg->MethodID)
    {
	case OM_NEW:		return mNew(cl, obj, (APTR)msg);
	case OM_DISPOSE:        return mDispose(cl, obj, (APTR)msg);
	case OM_GET:            return mGet(cl, obj, (APTR)msg);
	case OM_SET:            return mSet(cl, obj, (APTR)msg);
        case MUIM_Setup:        return mSetup(cl, obj, (APTR)msg);
        case MUIM_Cleanup:      return mCleanup(cl, obj, (APTR)msg);
        case MUIM_HandleEvent:  return mHandleEvent(cl, obj, (APTR)msg);
#ifdef USE_NLIST
	case MUIM_NList_Compare:return mCompare(cl, obj, (APTR)msg);
#ifndef USE_DISPLAY_HOOK
	case MUIM_NList_Display:return mDisplay(cl, obj, (APTR)msg);
#endif
#else
	case MUIM_List_Compare: return mCompare(cl, obj, (APTR)msg);
#ifndef USE_DISPLAY_HOOK
	case MUIM_List_Display: return mDisplay(cl, obj, (APTR)msg);
#endif
#endif
        case MUIM_SFWList_GetRecord:
                                return sfwListGetRecord(cl, obj, (APTR)msg);
        case MUIM_SFWList_SetTitlesBody:
                                return sfwListSetTitlesBody(cl, obj, (APTR)msg);
        case MUIM_SFWList_SetElement:
                                return sfwListSetElement(cl, obj, (APTR)msg);
        case MUIM_SFWList_SetString:
                                return sfwListSetString(cl, obj, (APTR)msg);
        case MUIM_SFWList_SetSelected:
                                return sfwListSetSelected(cl, obj, (APTR)msg);
        case MUIM_SFWList_GetSelected:
                                return sfwListGetSelected(cl, obj, (APTR)msg);
        case MUIM_SFWList_RemRecord:  
                                return sfwListRemRecord(cl, obj, (APTR)msg);
        case MUIM_SFWList_RemAllRecords:  
                                return sfwListRemAllRecords(cl, obj, (APTR)msg);
        case MUIM_DragDrop:     return sfwListDragDrop(cl, obj, (APTR)msg);
    }
    return DoSuperMethodA(cl, obj, msg);
}

void SFWListCleanup(void)
{
    if (SFWList_mcc != NULL) {
        MUI_DeleteCustomClass(SFWList_mcc);
    }
    SFWList_mcc = NULL;
    SFWListClass = NULL;
}

BOOL SFWListInit(void)
{
    SFWList_mcc = MUI_CreateCustomClass(NULL, (STRPTR)SUPERCLASS, NULL,
        sizeof(struct ClassData), Dispatcher);
    if (SFWList_mcc == NULL) {
        SFWListCleanup();
        return FALSE;
    }

    SFWListClass = SFWList_mcc->mcc_Class;
    DISPATCHER_INIT(SFWListClass);

    return TRUE;
}

