/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** Field_mui.c
**
** Creation date: 22.11.1994
** Last modified: 10.03.2021
*/

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "structureeditor.h"
#include "menuproject.h"
#include "project.h"
#include "table.h"
#include "field.h"
#include "record.h"
#include "preferences.h"
#include "requesters.h"
#include "memory.h"
#include "labels.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"
#include "structureeditor_mui.h"
#include "extclasses/ExtApplication_mcc.h"
#include "extclasses/NewCycle_mcc.h"
#include "extclasses/PopEditLabels_mcc.h"

static short types[] = 					/* remapping of types */
{
    TYPE_STRING, TYPE_INTEGER, TYPE_REAL, TYPE_BOOL, TYPE_CHOICE, TYPE_DATE, TYPE_TIME,
    TYPE_MEMO, TYPE_REFERENCE, TYPE_VIRTUAL, TYPE_BUTTON,
};

enum
{
    GRP_STRING, GRP_BOOL, GRP_CHOICE, GRP_MEMO, GRP_REFERENCE, GRP_VIRTUAL, 
    GRP_DEFAULT, GRP_BUTTON,
};

static short pages[] =
{
    GRP_STRING, GRP_DEFAULT, GRP_DEFAULT, GRP_BOOL, GRP_CHOICE, GRP_DEFAULT, 
    GRP_DEFAULT, GRP_MEMO, GRP_REFERENCE, GRP_VIRTUAL, GRP_BUTTON,
};

static const char *typeContent[12];
static const char *boolInitialValueContent[3];
static const char *initialValueContent[3];

struct EditFieldData
{
    struct Project *d_Project;
    struct Table *d_Table;
    struct Field *d_Field;
    struct DisplayObject *d_Parent;
    Object *d_StructureEditor;
    Object *d_Name;
    Object *d_Type;
    Object *d_TypeGroup;
    Object *d_StringMaxlen;
    Object *d_StringInitialValue;
    Object *d_BoolInitialValue;
    Object *d_ChoiceLabels;
    Object *d_ChoiceInitialValue;
    Object *d_Reference;
    Object *d_ReferenceFilter;
    Object *d_ReferenceAutoShow;
    Object *d_InitialValue;
    Object *d_InitialValueOther;
    Object *d_NilString;
    Object *d_ComputeStr;
    Object *d_NilStringVirtual;
    Object *d_Compute;
    Object *d_TriggerStr;
    Object *d_Trigger;
    Object *d_CountChanges;
    Object *d_LogChanges;
    Object *d_Ok;
    Object *d_Cancel;
};

static void disposeField(Object *app, Object *obj)
{
    struct EditFieldData *d = pget(obj, MUIA_UserData);

    if(obj && d)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(struct EditFieldData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void displayerrormessage(Object *obj, long error)
{
    struct EditFieldData *d = pget(obj, MUIA_UserData);
    Object *activate = NULL;

    Ask2(obj, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));

    switch (error) {
	case ERR_INVALIDNAME:
	case ERR_NAMEALREADYUSED:
        default:    
	    activate = d->d_Name;
	    break;

	case FIELDERR_INVALIDMAXLEN:
	    activate = d->d_StringMaxlen;
	    break;

	case FIELDERR_INITIALVALUELEN:
	case FIELDERR_INVALIDINITIALVALUE:
	    activate = d->d_InitialValueOther;
            break;
    }
    if (activate) {
        set(obj, MUIA_Window_ActiveObject, activate);
    }
}

static void getfieldtypedata(struct EditFieldData *d, short type,
	const char **init, const char **nil, union FieldTypeData *ftd)
{
    switch (type) {
	case TYPE_STRING:
	    ftd->string.maxlen = nget(d->d_StringMaxlen, MUIA_String_Integer);
	    *init = pget(d->d_StringInitialValue, MUIA_String_Contents);
	    break;

 	case TYPE_INTEGER:
	case TYPE_REAL:
	case TYPE_DATE:
	case TYPE_TIME:
            if (nget(d->d_InitialValue, MUIA_Radio_Active)) {
		*init = pget(d->d_InitialValueOther, MUIA_String_Contents);
	    }
            *nil = pget(d->d_NilString, MUIA_String_Contents);
            break;

	case TYPE_BOOL:
	    ftd->boolean.nil = !nget(d->d_BoolInitialValue, MUIA_Radio_Active);
            break;

	case TYPE_CHOICE:
	    ftd->choice.labels = pget(d->d_ChoiceLabels,MUIA_EditLabels_Labels);
	    ftd->choice.initialvalue = 
		nget(d->d_ChoiceInitialValue, MUIA_Cycle_Active);
            break;

        case TYPE_REFERENCE:
        {
            struct Table *tab = NULL;

            DoMethod(d->d_Reference, MUIM_List_GetEntry,
                MUIV_List_GetEntry_Active, (APTR)&tab);

            ftd->reference.table = tab;
            ftd->reference.filter = nget(d->d_ReferenceFilter, MUIA_Selected);
	    ftd->reference.autoshow = nget(d->d_ReferenceAutoShow, 
		    MUIA_Selected);
            break;
        }

        case TYPE_VIRTUAL:
            ftd->virt.compute = pget(d->d_ComputeStr, MUIA_String_Contents);
            *nil = pget(d->d_NilStringVirtual, MUIA_String_Contents);
            break;
    }
}

static void HOOK_PROTO(finish, Object *obj, ulong *ok)
{
    struct EditFieldData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);
    long error = OK;

    SetPrefsValue(pr, PREFS_DEFAULTMAXLEN,
        nget(d->d_StringMaxlen, MUIA_String_Integer));

    if(*ok && d->d_Field == NULL)              /* new field? */
    {
        struct NewFieldData nfd;
        struct Table *tab = d->d_Table;
        struct Field *new = NULL;
	short type = types[nget(d->d_Type, MUIA_Cycle_Active)];

	nfd.nfd_Table = tab;
        nfd.nfd_Pred = DisplayTableActiveField(tab);
        nfd.nfd_Name = pget(d->d_Name, MUIA_String_Contents);
        nfd.nfd_Type = type;
	nfd.nfd_InitialValue = NULL;
	nfd.nfd_NilString = NULL;
        nfd.nfd_Trigger = pget(d->d_TriggerStr, MUIA_String_Contents);
	nfd.nfd_CountChanges = nget(d->d_CountChanges, MUIA_Selected);
	nfd.nfd_LogChanges = nget(d->d_LogChanges, MUIA_Selected);
	getfieldtypedata(d, type, &nfd.nfd_InitialValue, &nfd.nfd_NilString,
                &nfd.nfd_Contents);

        GuiSetSleep(TRUE);
	error = SE_NewField(&nfd, d->d_Parent, &new);
        GuiSetSleep(FALSE);

	if(error)
        {
            displayerrormessage(obj, error);
            return;
        }
        NewFieldFinish(d->d_StructureEditor, new);
    }
    else if(*ok && d->d_Field != NULL)
    {
        struct ChangeFieldData cfd;
        short type = FieldType(d->d_Field);

	cfd.cfd_Field = d->d_Field;
        cfd.cfd_Name = pget(d->d_Name, MUIA_String_Contents);
	cfd.cfd_InitialValue = NULL;
	cfd.cfd_NilString = NULL;
        cfd.cfd_Trigger = pget(d->d_TriggerStr, MUIA_String_Contents);
	cfd.cfd_CountChanges = nget(d->d_CountChanges, MUIA_Selected);
	cfd.cfd_LogChanges = nget(d->d_LogChanges, MUIA_Selected);
        getfieldtypedata(d, type, &cfd.cfd_InitialValue, &cfd.cfd_NilString,
                &cfd.cfd_Contents);

        GuiSetSleep(TRUE);
        error = SE_ChangeField(&cfd);
        GuiSetSleep(FALSE);
	if(error)
        {
            displayerrormessage(obj, error);
            return;
        }
        ChangeFieldFinish(d->d_StructureEditor, d->d_Field);
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposeField(app, obj);
}

static ulong HOOK_PROTO(referenceDisplay, const char **dst, struct Table *tab)
{
    *dst = TableName(tab);
    return(0);
}
static HOOK_STRUCT(referenceDisplay)

static void HOOK_PROTO(setchoicelabels, Object *obj, 
	struct EditFieldData **x)
{
    struct EditFieldData *d = *x;
    char **labels = pget(d->d_ChoiceLabels, MUIA_EditLabels_Labels);

    set(d->d_ChoiceInitialValue, MUIA_Cycle_Entries, labels);
}

static void HOOK_PROTO(fieldtypechange, Object *obj, 
	struct EditFieldData **x)
{
    struct EditFieldData *d = *x;
    ulong active = nget(d->d_Type, MUIA_Cycle_Active);

    set(d->d_TypeGroup, MUIA_Group_ActivePage, pages[active]);
}

static void updatePage(struct EditFieldData *d)
{
    ulong page = nget(d->d_TypeGroup, MUIA_Group_ActivePage);
    bool on = (page != GRP_VIRTUAL && page != GRP_BUTTON);
    bool countOn, logOn;

    set(d->d_Trigger, MUIA_Disabled, page == GRP_VIRTUAL);
    countOn = on && TableCountChanges(d->d_Table);
    set(d->d_CountChanges, MUIA_Disabled, !countOn);
    if (!countOn) {
        set(d->d_CountChanges, MUIA_Selected, FALSE);
    }
    logOn = on && TableLogChanges(d->d_Table);
    set(d->d_LogChanges, MUIA_Disabled, !logOn);
    if (!logOn) {
        set(d->d_LogChanges, MUIA_Selected, FALSE);
    }
}

static void HOOK_PROTO(pagechange, Object *obj, struct EditFieldData **x)
{
    updatePage(*x);
}

static Object *createField(struct Project *pr, 
			       struct Field *field, struct Table *tab,
			       bool newfield,
			       struct DisplayObject *parent, Object *se)
{
    Object *app = GuiApplication;
    struct MemPool *memPool = ProjectMemPool(pr);
    struct EditFieldData *d;
    Object *grp1, *grp_string, *grp_boolean, *grp_choice, *grp_reference, 
      *grp_default_sub, *grp_default, *grp_memo, *grp_virtual, *grp_field, 
      *grp_button, *grp_buttons;
    Object *obj;

    if(app == NULL)
        return(NULL);

    d = PoolAllocMemZero(memPool, sizeof(struct EditFieldData));
    if(d == NULL)
	return(NULL);
    d->d_Project = pr;
    d->d_Table = tab;
    d->d_Field = newfield? NULL : field;
    d->d_Parent = parent;
    d->d_StructureEditor = se;

    typeContent[0] = GCS(MSG_String);
    typeContent[1] = GCS(MSG_Integer);
    typeContent[2] = GCS(MSG_Real);
    typeContent[3] = GCS(MSG_Boolean);
    typeContent[4] = GCS(MSG_Choice);
    typeContent[5] = GCS(MSG_Date);
    typeContent[6] = GCS(MSG_Time);
    typeContent[7] = GCS(MSG_Memo);
    typeContent[8] = GCS(MSG_Reference);
    typeContent[9] = GCS(MSG_Virtual);
    typeContent[10] = GCS(MSG_Button);
    typeContent[11] = NULL;
    boolInitialValueContent[0] = GCS(MSG_Field_Boolean_InitialValue_Nil);
    boolInitialValueContent[1] = GCS(MSG_Field_Boolean_InitialValue_True);
    boolInitialValueContent[2] = NULL;
    initialValueContent[0] = GCS(MSG_Field_InitialValue_Nil);
    initialValueContent[1] = GCS(MSG_Field_InitialValue_Other);
    initialValueContent[2] = NULL;

    grp1 = VGroup,
        MUIA_Group_Columns, 2,
	Child, Label(GCS(MSG_Name)),
	Child, d->d_Name = StringObject,
	    MUIA_CycleChain, TRUE,
	    MUIA_Frame, MUIV_Frame_String,
	    MUIA_ControlChar, Shortcut(GCS(MSG_Name)),
            MUIA_ShortHelp, GCS(MSG_Field_NameBH),
	    MUIA_String_MaxLen, 21,
	    MUIA_String_Format, MUIV_String_Format_Center,
	    End,
	Child, Label(GCS(MSG_Type)),
	Child, d->d_Type = CycleObject,
	    MUIA_CycleChain, TRUE,
	    MUIA_ControlChar, Shortcut(GCS(MSG_Type)),
            MUIA_ShortHelp, GCS(MSG_Field_TypeBH),
   	    MUIA_Cycle_Entries, typeContent,
	    End,
	End;

    d->d_StringMaxlen = StringObject,
	MUIA_CycleChain, TRUE,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Field_String_MaxLength)),
        MUIA_ShortHelp, GCS(MSG_Field_String_MaxLengthBH),
	MUIA_String_Accept, "0123456789",
	MUIA_String_MaxLen, 4,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Integer, GetPrefsValue(pr, PREFS_DEFAULTMAXLEN),
	End;

    d->d_StringInitialValue = StringObject,
	MUIA_CycleChain, TRUE,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Field__InitialValue)),
	MUIA_ShortHelp, GCS(MSG_Field_InitialValueBH),
	MUIA_String_MaxLen, 256,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    grp_string = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	MUIA_Group_Columns, 2,
	Child, VSpace(0),
	Child, VSpace(0),
	Child, Label(GCS(MSG_Field_String_MaxLength)),
	Child, d->d_StringMaxlen,
	Child, Label(GCS(MSG_Field__InitialValue)),
	Child, d->d_StringInitialValue,
	Child, VSpace(0),
	Child, VSpace(0),
	End;

    d->d_BoolInitialValue = RadioObject,
	MUIA_CycleChain, TRUE,
	MUIA_Radio_Entries, boolInitialValueContent,
	MUIA_ControlChar, Shortcut(GCS(MSG_Field__InitialValue)),
	End;

    grp_boolean = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	Child, VSpace(0),
	Child, HGroup,
            Child, HSpace(0),
            Child, HGroup,
	        MUIA_Frame, MUIV_Frame_Group,
	        MUIA_FrameTitle, GCS(MSG_Field_InitialValue),
                MUIA_Background, MUII_GroupBack,
	        MUIA_ShortHelp, GCS(MSG_Field_InitialValueBH),
                Child, HSpace(0),
                Child, HSpace(0),
                Child, d->d_BoolInitialValue,
                Child, HSpace(0),
                Child, HSpace(0),
                End,
            Child, HSpace(0),
            End,
	Child, VSpace(0),
	End;

    d->d_ChoiceLabels = PopEditLabelsObject,
	ButtonFrame,
        MUIA_CycleChain, TRUE,
	MUIA_Text_Contents, GCS(MSG_EditLabels),
	MUIA_Text_PreParse, "\33c",
	MUIA_Text_HiChar, Shortcut(GCS(MSG_Field_Choice__Labels)),
	MUIA_ControlChar, Shortcut(GCS(MSG_Field_Choice__Labels)),
	MUIA_InputMode, MUIV_InputMode_RelVerify,
	MUIA_Background, MUII_ButtonBack,
        MUIA_ShortHelp, GCS(MSG_Field_Choice_LabelsBH),
        End;
    d->d_ChoiceInitialValue = NewCycleObject,
	MUIA_CycleChain, TRUE,
	MUIA_ControlChar, Shortcut(GCS(MSG_Field__InitialValue)),
	MUIA_ShortHelp, GCS(MSG_Field_InitialValueBH),
	End;

    grp_choice = VGroup,
	MUIA_Group_Columns, 2,
	Child, HSpace(0),
	Child, d->d_ChoiceLabels,
	Child, Label(GCS(MSG_Field__InitialValue)),
	Child, d->d_ChoiceInitialValue,
	End;

    grp_choice = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	Child, VSpace(0),
	Child, grp_choice,
	Child, VSpace(0),
	End;

    d->d_Reference = ListviewObject,
	MUIA_CycleChain, TRUE,
	MUIA_ControlChar, Shortcut(GCS(MSG_Field__InitialValue)),
        MUIA_ShortHelp, GCS(MSG_Field_ReferenceBH),
	MUIA_Listview_List, ListObject,
	    MUIA_Frame, MUIV_Frame_InputList,
            MUIA_List_DisplayHook, &referenceDisplayHook,
	    End,
	End;
    d->d_ReferenceFilter = MUI_MakeObject(MUIO_Checkmark, 
	    (IPTR)GCS(MSG_Filter));
    if(d->d_ReferenceFilter)
	SetAttrs(d->d_ReferenceFilter,
            MUIA_CycleChain, TRUE,
            MUIA_ShortHelp, GCS(MSG_Field_Reference_FilterBH),
            TAG_DONE);
    d->d_ReferenceAutoShow = MUI_MakeObject(MUIO_Checkmark, 
	    (IPTR)GCS(MSG_AutoShow));
    if(d->d_ReferenceAutoShow)
	SetAttrs(d->d_ReferenceAutoShow,
		 MUIA_CycleChain, TRUE,
		 MUIA_ShortHelp, GCS(MSG_Field_Reference_AutoShowBH),
		 TAG_DONE);
    grp_reference = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	Child, HCenter(Label(GCS(MSG_Field_Reference))),
	Child, d->d_Reference,
	Child, HGroup,
	    Child, Label(GCS(MSG_AutoShow)),
	    Child, d->d_ReferenceAutoShow,
	    Child, HSpace(0),
	    Child, Label(GCS(MSG_Filter)),
	    Child, d->d_ReferenceFilter,
	    End,
	End;

    d->d_InitialValue = RadioObject,
	MUIA_CycleChain, TRUE,
	MUIA_ControlChar, Shortcut(GCS(MSG_Field__InitialValue)),
	MUIA_ShortHelp, GCS(MSG_Field_InitialValueBH),
	MUIA_Radio_Entries, initialValueContent,
	End;

    d->d_InitialValueOther = StringObject,
	MUIA_CycleChain, TRUE,
        MUIA_Disabled, TRUE,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Field_OtherValue)),
	MUIA_String_MaxLen, 21,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    grp_default_sub = HGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_Field_InitialValue),
        MUIA_Background, MUII_GroupBack,
	Child, d->d_InitialValue,
	Child, RectangleObject,
	    MUIA_Rectangle_VBar, TRUE,
	    MUIA_FixWidth, 10,
	    End,
	Child, VGroup,
	    Child, HCenter(Label(GCS(MSG_Field_OtherValue))),
	    Child, d->d_InitialValueOther,
	    End,
	End;

    d->d_NilString = StringObject,
	MUIA_CycleChain, TRUE,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Field_NilString)),
	MUIA_ShortHelp, GCS(MSG_Field_NilStringBH),
	MUIA_String_MaxLen, FIELD_NILVALUE_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    grp_default = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	Child, VSpace(0),
	Child, grp_default_sub,
	Child, VGroup,
	    MUIA_Group_Columns, 2,
	    Child, Label(GCS(MSG_Field_NilString)),
	    Child, d->d_NilString,
	    End,
	Child, VSpace(0),
	End;

    grp_memo = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	Child, HVSpace,
	End;

    d->d_ComputeStr = StringObject,
	MUIA_CycleChain, TRUE,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Field_Compute)),
	MUIA_ShortHelp, GCS(MSG_Field_ComputeBH),
	MUIA_String_MaxLen, FIELD_TRIGGER_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    d->d_Compute = UserFunctionsPopup(d->d_ComputeStr, pr);

    d->d_NilStringVirtual = StringObject,
	MUIA_CycleChain, TRUE,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Field_NilString)),
	MUIA_ShortHelp, GCS(MSG_Field_NilStringBH),
	MUIA_String_MaxLen, FIELD_NILVALUE_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    grp_virtual = GroupObject,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	Child, VSpace(0),
	Child, GroupObject,
	    MUIA_Group_Columns, 2,
	    Child, Label(GCS(MSG_Field_Compute)),
	    Child, d->d_Compute,
	    Child, Label(GCS(MSG_Field_NilString)),
	    Child, d->d_NilStringVirtual,
	    End,
	Child, VSpace(0),
	End;

    grp_button = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	Child, HVSpace,
	End;

    d->d_TypeGroup = GroupObject,
	MUIA_Group_PageMode, TRUE,
        MUIA_HelpNode, "Type specific settings",
	Child, grp_string,
	Child, grp_boolean,
	Child, grp_choice,
	Child, grp_memo,
	Child, grp_reference,
        Child, grp_virtual,
	Child, grp_default,
	Child, grp_button,
	End;

    d->d_TriggerStr = StringObject,
	MUIA_CycleChain, TRUE,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_Field_Trigger)),
	MUIA_ShortHelp, GCS(MSG_Field_TriggerBH),
	MUIA_String_MaxLen, FIELD_TRIGGER_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    d->d_Trigger = UserFunctionsPopup(d->d_TriggerStr, pr);

    d->d_CountChanges = MUI_MakeObject(MUIO_Checkmark, 
    	    (IPTR)GCS(MSG_CountChanges));
    if (d->d_CountChanges) {
	SetAttrs(d->d_CountChanges,
		MUIA_CycleChain, TRUE,
		MUIA_Selected, TRUE,
		MUIA_ShortHelp, GCS(MSG_Field_CountChangesBH),
		TAG_DONE);
    }
    d->d_LogChanges = MUI_MakeObject(MUIO_Checkmark,
            (IPTR)GCS(MSG_LogChanges));
    if (d->d_LogChanges) {
	SetAttrs(d->d_LogChanges,
		MUIA_CycleChain, TRUE,
		MUIA_Selected, TRUE,
		MUIA_ShortHelp, GCS(MSG_Field_LogChangesBH),
		TAG_DONE);
    }
    
    grp_field = VGroup,
	Child, grp1,
	Child, d->d_TypeGroup,
	Child, VGroup,
	    MUIA_Group_Columns, 2,
	    Child, Label(GCS(MSG_Field_Trigger)),
	    Child, d->d_Trigger,
	    Child, Label(GCS(MSG_CountChanges)),
	    Child, HGroup, 
	    	Child, d->d_CountChanges,
	    	Child, HSpace(0),
                Child, Label(GCS(MSG_LogChanges)),
	    	Child, d->d_LogChanges,
		End,
	    End,
	Child, RectangleObject,
	    MUIA_FixHeight, 8,
	    End,
	End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    grp_buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, d->d_Ok,
	Child, d->d_Cancel,
	End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_ID, MAKE_ID('_', 'A','T','R'),
	MUIA_Window_NoMenus, TRUE,
        MUIA_Window_ScreenTitle, GuiScreenTitle,
        MUIA_HelpNode, newfield? "Creating fields" : "Changing fields",
        MUIA_UserData, d,
	WindowContents, VGroup,
	    Child, grp_field,
	    Child, grp_buttons,
	    End,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct EditFieldData));
	return(NULL);
    }

    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    DoMethod(d->d_Name, MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    DoMethod(d->d_StringMaxlen, MUIM_Notify,
            MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    DoMethod(d->d_StringInitialValue, MUIM_Notify,
            MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    DoMethod(d->d_InitialValueOther, MUIM_Notify,
            MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    DoMethod(d->d_NilString, MUIM_Notify,
            MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    DoMethod(d->d_TriggerStr, MUIM_Notify,
            MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    DoMethod(d->d_ComputeStr, MUIM_Notify,
            MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    DoMethod(d->d_NilStringVirtual, MUIM_Notify,
            MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    DoMethod(d->d_InitialValue, MUIM_Notify,
            MUIA_Radio_Active, MUIV_EveryTime,
            d->d_InitialValueOther, 3, MUIM_Set, 
            MUIA_Disabled, MUIV_NotTriggerValue);

    MUIAddCallback(d->d_ChoiceLabels, MUIA_EditLabels_Labels, MUIV_EveryTime,
            setchoicelabels, d);

    MUIAddCallback(d->d_Type, MUIA_Cycle_Active, MUIV_EveryTime,
            fieldtypechange, d);

    MUIAddCallback(d->d_TypeGroup, MUIA_Group_ActivePage, MUIV_EveryTime,
            pagechange, d);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE, obj, finish,FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE, obj, finish, FALSE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE, obj, finish, TRUE);

    if (newfield)
    {
        set(obj, MUIA_Window_Title, GCS(MSG_NewField));
        MUIListSetListWithFilter(d->d_Reference, ProjectTables(pr), memPool, 
                TableSystemFilter);
    }
    if (field)
    {
        short type = FieldType(field), page;
        struct DataInfo *di = FieldDataInfo(field);
        bool nil = ((di->di_Flags & DIF_NIL) != 0);
        bool yes = !nil;

        for(page = 0; type != types[page]; page++)
	    if(page > sizeof(types) / sizeof(short))
	      {
                page = 1;
		break;
	      }

        set(obj, MUIA_Window_Title, 
	    newfield? GCS(MSG_CopyField) : GCS(MSG_ChangeField));
        set(d->d_TypeGroup, MUIA_Group_ActivePage, pages[page]);
        set(d->d_Name, MUIA_String_Contents, FieldName(field));
        set(d->d_Type, MUIA_Cycle_Active, page);
        set(d->d_Type, MUIA_Disabled, newfield? FALSE : TRUE);

        switch(type)
        {
	    case TYPE_STRING:
                set(d->d_StringMaxlen, MUIA_String_Integer, FieldStringMaxlen(field));
		set(d->d_StringInitialValue, MUIA_String_Contents, di->di_Contents.string);
                break;

	    case TYPE_INTEGER:
                set(d->d_InitialValue, MUIA_Radio_Active, yes);
                if (yes) {
                    set(d->d_InitialValueOther, MUIA_String_Contents,
                            IntToStr(di->di_Contents.integer));
                }
	        set(d->d_InitialValueOther, MUIA_Disabled, nil);
                set(d->d_NilString, MUIA_String_Contents, 
                        FieldNilString(field));
	        break;

	    case TYPE_REAL:
                set(d->d_InitialValue, MUIA_Radio_Active, yes);
                if (yes) {
	            set(d->d_InitialValueOther, MUIA_String_Contents,
		        DoubleToStr(di->di_Contents.real, -1,
                            GetPrefsValue(pr, PREFS_REALFORMAT)));
                }
	        set(d->d_InitialValueOther, MUIA_Disabled, nil);
                set(d->d_NilString, MUIA_String_Contents, 
                        FieldNilString(field));
	        break;

	    case TYPE_BOOL:
	        set(d->d_BoolInitialValue, MUIA_Radio_Active, yes);
	        break;

	    case TYPE_CHOICE:
                set(d->d_ChoiceLabels, MUIA_EditLabels_Labels, 
                        FieldChoiceLabels(field));
	        set(d->d_ChoiceInitialValue, MUIA_Cycle_Active, 
                        di->di_Contents.choice);
	        break;

            case TYPE_DATE:
                set(d->d_InitialValue, MUIA_Radio_Active, yes);
                if (yes) {
                    set(d->d_InitialValueOther, MUIA_String_Contents,
                            DateToString(di->di_Contents.date,
                                GetPrefsValue(pr, PREFS_DATEFORMAT)));
                }
	        set(d->d_InitialValueOther, MUIA_Disabled, nil);
                set(d->d_NilString, MUIA_String_Contents, 
                        FieldNilString(field));
	        break;

	    case TYPE_TIME:
                set(d->d_InitialValue, MUIA_Radio_Active, yes);
                if (yes) {
                    set(d->d_InitialValueOther, MUIA_String_Contents,
                        TimeToString(di->di_Contents.time, FORMAT_HOURMINSEC));
                }
	        set(d->d_InitialValueOther, MUIA_Disabled, nil);
                set(d->d_NilString, MUIA_String_Contents, 
                        FieldNilString(field));
	        break;

            case TYPE_REFERENCE:
            {
                struct Table *reftab = FieldReferenceTable(field);

                MUIListSetListWithFilter(d->d_Reference, ProjectTables(pr), 
                        memPool, TableSystemFilter);
                MUIListSetActiveNode(d->d_Reference, reftab);
                set(d->d_Reference, MUIA_Disabled, TRUE);
                set(d->d_ReferenceFilter, MUIA_Selected, 
		    FieldReferenceFilter(field));
                set(d->d_ReferenceAutoShow, MUIA_Selected, 
		    FieldReferenceAutoShow(field));
                break;
            }

            case TYPE_VIRTUAL:
                set(d->d_ComputeStr, MUIA_String_Contents, 
                        FieldVirtualCompute(field));
                set(d->d_NilStringVirtual, MUIA_String_Contents, 
                        FieldNilString(field));
                break;
        }
        set(d->d_TriggerStr, MUIA_String_Contents, FieldTrigger(field));
        set(d->d_CountChanges, MUIA_Selected, FieldCountChanges(field));
        set(d->d_LogChanges, MUIA_Selected, FieldLogChanges(field));
    }
    updatePage(d);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeField(app, obj);
        return(NULL);
    }

    set(obj, MUIA_Window_ActiveObject, d->d_Name);

    return(obj);
}

static bool checkNewChangeField(struct Field *field, struct Table *tab,
        bool newField)
{
    long error = newField? CanNewField(tab) : CanChangeField(field);
    bool rc = (error == OK);

    if (error == RECERR_NOPROJECTNAME) {
        if (Ask(AT_WARN, GCS(MSG_AskProjectSaveFirstAnswers),
                    GCS(MSG_AskProjectSaveFirst)) != 0) {
            MenuProjectSaveAs(TableProject(tab));
            error = newField? CanNewField(tab) : CanChangeField(field);
            rc = (error == OK);
        } // else user canceled
    } else if (error != OK) {
        Ask2(0, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));
    }

    return rc;
}

void GuiNewField(Object *se, struct Table *tab, struct Field *field, 
		     struct DisplayObject *parent)
{
    if (checkNewChangeField(field, tab, TRUE)) {
        createField(TableProject(tab), field, tab, TRUE, parent, se);
    }
}

void GuiChangeField(Object *se, struct Field *field)
{
    struct Table *tab = FieldTable(field);

    if (checkNewChangeField(field, tab, FALSE)) {
        createField(TableProject(tab), field, tab, FALSE, NULL, se);
    }
}

#endif

