/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** Filter_mui.c
**
** Creation date: 17.04.1995
** Last modified: 19.01.2010
*/

#include "filter.h"
#include "project.h"
#include "table.h"
#include "field.h"
#include "memory.h"
#include "util.h"
#include "lang/compile.h"
#include "lang/program.h"
#include "requesters.h"
#include "error.h"
#include "appstrings_cat.h"
#include "extclasses/ExtApplication_mcc.h"

#define BUFSIZE 500
static char buffer[BUFSIZE+1];

struct EditFilterData
{
    struct Project *d_Project;
    struct Table *d_Table;
    char d_WindowTitle[101];
    Object *d_Fields;
    Object *d_Expression;
    Object *d_Ok;
    Object *d_Cancel;
};

static void insertText(Object *strobj, char *text)
{
    char *str = pget(strobj, MUIA_String_Contents);
    long bufferPos = nget(strobj, MUIA_String_BufferPos);
    long displayPos = nget(strobj, MUIA_String_DisplayPos);
    long l1, l2;

    if(text == NULL || str == NULL)
        return;

    l1 = strlen(str);
    l2 = strlen(text);
    if(bufferPos > l1)
        bufferPos = l1;
    if(l2 == 0)
        return;

    if(l1 + l2 > BUFSIZE)
    {
        GuiBeep();
        return;
    }
    sprintf(buffer, "%.*s%s%.*s", (int)bufferPos, str, text, 
	    (int)(l1 - bufferPos), str + bufferPos);
    bufferPos += l2;
    if(text[l2 - 1] == ')')
        bufferPos--;

    SetAttrs(strobj,
        MUIA_String_Contents, buffer,
        MUIA_String_DisplayPos, displayPos,
        MUIA_String_BufferPos, bufferPos,
        TAG_DONE);
}

static void activate(Object *obj)
{
    Object *wi = GuiObjectWindow(obj);

    if(wi)
        set(wi, MUIA_Window_ActiveObject, obj);
}

static void HOOK_PROTO(insert, Object *button, void *dummy)
{
    Object *obj = GuiObjectWindow(button);
    struct EditFilterData *d = pget(obj, MUIA_UserData);
    char *text = pget(button, MUIA_Text_Contents);
    char buf[51];

    if(text)
    {
        sprintf(buf, "(%s )", text);
        insertText(d->d_Expression, buf);
    }
    activate(d->d_Expression);
}

static void HOOK_PROTO(insert2, Object *button, void *dummy)
{
    Object *obj = GuiObjectWindow(button);
    struct EditFilterData *d = pget(obj, MUIA_UserData);
    char *text = pget(button, MUIA_Text_Contents);
    char buf[51];

    if(text)
    {
        sprintf(buf, "%s ", text);
        insertText(d->d_Expression, buf);
    }
    activate(d->d_Expression);
}

static void HOOK_PROTO(insert3, Object *obj, void *dummy)
{
    struct EditFilterData *d = pget(obj, MUIA_UserData);
    char *text = NULL;
    char buf[51];

    DoMethod(d->d_Fields, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, &text);
    if(text)
    {
        sprintf(buf, "%s ", text);
        insertText(d->d_Expression, buf);
    }
    activate(d->d_Expression);
}

static struct
{
    char *fld_String;
    short fld_InsertFuncNum;
} filterlabeldatas[] =
{
    { "=", 	0 },
    { "<>", 	0 },
    { "=*", 	0 },
    { "<>*", 	0 },
    { "<", 	0 },
    { "<=", 	0 },
    { "<*", 	0 },
    { "<=*", 	0 },
    { ">", 	0 },
    { ">=", 	0 },
    { ">*", 	0 },
    { ">=*", 	0 },
    { "+", 	0 },
    { "-", 	0 },
    { "*", 	0 },
    { "/", 	0 },
    { "DIV", 	0 },
    { "MOD", 	0 },
    { "LIKE", 	0 },
    { "NOT", 	0 },
    { "AND", 	0 },
    { "OR", 	0 },
    { "TRUE", 	1 },
    { "NIL", 	1 },
};

#define NUM_FILTERLABELS (sizeof(filterlabeldatas) / sizeof(filterlabeldatas[0]))

static void disposeEditFilter(Object *app, Object *obj)
{
    struct EditFilterData *d = pget(obj, MUIA_UserData);

    if(obj && d)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(struct EditFilterData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void HOOK_PROTO(finish, Object *obj, ulong *ok)
{
    struct EditFilterData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);
    struct Program *prg = NULL;

    if(*ok)
    {
        char *text = pget(d->d_Expression, MUIA_String_Contents);
        const char *error = NULL;
        long line = 1, col = 1;

        GuiSetSleep(TRUE);
        if(text == NULL)
            error = GCS(MSG_Err_Internal);
        if(error == NULL)
            error = CompileFilter(d->d_Table, text, strlen(text), &prg, &line, &col);
        GuiSetSleep(FALSE);

        if(error != NULL)
        {
            set(obj, MUIA_Window_Title, error);
            if(col > 0 && col <= BUFSIZE)
                set(d->d_Expression, MUIA_String_BufferPos, col-1);
            GuiBeep();
            return;
    	}
    }

    GuiCloseWindow(obj);
    if(*ok)
        SetFilter(d->d_Table, prg);

    MUIRemMember(app, obj);
    disposeEditFilter(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
}

static Object *createEditFilter(struct Table *tab)
{
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = ProjectMemPool(pr);
    Object *app = GuiApplication;
    struct Program *prg = TableFilterProgram(tab);
    struct List *fields = TableFields(tab);
    struct Field *field;
    struct EditFilterData *d;
    Object *root, *fields_group, *filterlabels[NUM_FILTERLABELS];
    Object *filterlabelsgroup, *top, *buttons;
    Object *obj;
    int i;

    if(app == NULL)
        return(NULL);

    buffer[0] = 0;
    if(prg)
    {
        struct SourceText *st = GetSourceText(prg);
        const char *text = SourceTextText(st);

        if(st == NULL)
            return(NULL);

        if(text)
	    CodesetStrNBytesCopy(buffer, text, BUFSIZE);
        FreeSourceText(st);
    }

    d = PoolAllocMemZero(memPool, sizeof(struct EditFilterData));
    if(d == NULL)
	return(NULL);
    d->d_Project = pr;
    d->d_Table = tab;

    filterlabelsgroup = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_Filter_Operators),
        MUIA_Background, MUII_GroupBack,
	MUIA_Group_Columns, 4,
	MUIA_Group_SameSize, TRUE,
	MUIA_Group_HorizSpacing, 0,
	MUIA_Group_VertSpacing, 0,
	End;

    if(filterlabelsgroup == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct EditFilterData));
        return(NULL);
    }

    for(i=0; i < NUM_FILTERLABELS; i++)
    {
        filterlabels[i] = SimpleButton(filterlabeldatas[i].fld_String);
        if(filterlabels[i] == 0)
        {
            MUI_AddDisposeObject(app, filterlabelsgroup);
            PoolFreeMem(memPool, d, sizeof(struct EditFilterData));
            return(NULL);
        }
        set(filterlabels[i], MUIA_CycleChain, TRUE);
        MUIAddMember(filterlabelsgroup, filterlabels[i]);
    }

    d->d_Fields = ListviewObject,
	MUIA_Listview_DoubleClick, TRUE,
	MUIA_Listview_Input, TRUE,
	MUIA_Listview_List, ListObject,
            MUIA_Frame, MUIV_Frame_InputList,
            End,
	End;

    fields_group = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_Fields),
        MUIA_Background, MUII_GroupBack,
	Child, d->d_Fields,
	End;

    top = HGroup,
	MUIA_Group_SameHeight, TRUE,
	Child, fields_group,
	Child, filterlabelsgroup,
	End;

    d->d_Expression = StringObject,
	MUIA_Frame, MUIV_Frame_String,
        MUIA_CycleChain, TRUE,
	MUIA_ControlChar, Shortcut(GCS(MSG_Filter)),
	MUIA_String_MaxLen, 500,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, buffer,
	End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, d->d_Ok,
	Child, d->d_Cancel,
	End;

    root = VGroup,
	Child, top,
	Child, d->d_Expression,
	Child, buttons,
	End;

    sprintf(d->d_WindowTitle, "%s: %s", GCS(MSG_ChangeFilter), TableName(tab));
    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_ID, MAKE_ID('_', 'F','I','L'),
        MUIA_Window_Title, d->d_WindowTitle,
        MUIA_Window_ScreenTitle, GuiScreenTitle,
        MUIA_Window_NoMenus, TRUE,
        MUIA_HelpNode, "Changing filters",
        MUIA_UserData, d,
	WindowContents, root,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct EditFilterData));
	return(NULL);
    }

    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    for(i=0; i < NUM_FILTERLABELS; i++)
    {
        if(filterlabeldatas[i].fld_InsertFuncNum == 0)
        {
            MUIAddCallback(filterlabels[i], MUIA_Pressed, FALSE, insert, NULL);
        }
        else
        {
            MUIAddCallback(filterlabels[i], MUIA_Pressed, FALSE, insert2, NULL);
        }
    }
    MUIAddCallback2(d->d_Fields, MUIA_Listview_DoubleClick, TRUE, obj, insert3, NULL);

    MUIAddCallback2(obj, MUIA_Window_CloseRequest, TRUE, obj, finish, FALSE);
    MUIAddCallback2(d->d_Cancel, MUIA_Pressed, FALSE, obj, finish, FALSE);
    MUIAddCallback2(d->d_Ok, MUIA_Pressed, FALSE, obj, finish, TRUE);
    MUIAddCallback2(d->d_Expression, MUIA_String_Acknowledge, MUIV_EveryTime,
        obj, finish, TRUE);

    if(fields)
        FORNODE(field, fields)
            if(FieldType(field) != TYPE_BUTTON)
                DoMethod(d->d_Fields, MUIM_List_InsertSingle,
                    FieldName(field), MUIV_List_Insert_Bottom);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeEditFilter(app, obj);
        return(NULL);
    }

    set(obj, MUIA_Window_ActiveObject, d->d_Expression);

    return(obj);
}

void GuiChangeFilter(struct Table *tab)
{
    createEditFilter(tab);
}

