/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** imexport_mui.c
**
** Creation date: 20.03.1997
** Last modified: 06.11.2020
*/

#include "system.h"
#include <libraries/asl.h>
#include "project.h"
#include "imexport.h"
#include "log.h"
#include "memory.h"
#include "util.h"
#include "error.h"
#include "requesters.h"
#include "appstrings_cat.h"

#define DELIMITER_LEN 9
#define FILENAME_LEN 255

static struct
{
    char ies_Filename[FILENAME_LEN + 1];
    char ies_RecordDelimiter[DELIMITER_LEN + 1];
    char ies_FieldDelimiter[DELIMITER_LEN + 1];
    bool ies_DoubleQuotes;
    bool ies_Filter;
    bool ies_OverwriteRecords;
} ImExportSettings =
{
    "", "\\n", "\\t", TRUE, FALSE, FALSE
};

enum ImExportAction
{
    IEA_Import,
#ifdef ENABLE_LOG_APPLY_CHANGES
    IEA_ApplyChanges,
#endif
    IEA_Export,
    IEA_ExportSFW
};

struct ImExportData
{
    struct Project *d_Project;
    struct Table *d_Table;
    GuiObject d_SFWList;
    enum ImExportAction d_Action;
    Object *d_Filename, *d_View, *d_Edit;
    Object *d_RecordDelimiter, *d_FieldDelimiter, *d_DoubleQuotes;
    Object *d_Filter;
    Object *d_OverwriteRecords;
    Object *d_ImExport, *d_Cancel;
};

static void disposeImExport(Object *app, Object *obj)
{
    struct ImExportData *d = pget(obj, MUIA_UserData);

    if(obj)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(struct ImExportData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void saveImExportSettings(Object *obj)
{
    struct ImExportData *d = pget(obj, MUIA_UserData);
    char *s;

    s = pget(d->d_Filename, MUIA_String_Contents);
    if (s != NULL) {
        CodesetStrNBytesCopy(ImExportSettings.ies_Filename, s, FILENAME_LEN);
    }
    s = pget(d->d_RecordDelimiter, MUIA_String_Contents);
    if (s != NULL) {
	CodesetStrNBytesCopy(ImExportSettings.ies_RecordDelimiter, 
		s, DELIMITER_LEN);
    }
    s = pget(d->d_FieldDelimiter, MUIA_String_Contents);
    if (s != NULL) {
	CodesetStrNBytesCopy(ImExportSettings.ies_FieldDelimiter, 
		s, DELIMITER_LEN);
    }
    ImExportSettings.ies_DoubleQuotes = nget(d->d_DoubleQuotes, MUIA_Selected);
    if (d->d_Filter) {
        ImExportSettings.ies_Filter = nget(d->d_Filter, MUIA_Selected);
    }
    if (d->d_OverwriteRecords) {
        ImExportSettings.ies_OverwriteRecords = 
            nget(d->d_OverwriteRecords, MUIA_Selected);
    }
}

static const char *checkDelimiterString(const char *delstring,
        struct MemPool *memPool, char *del)
{
    char *s;
    short len;
    bool ascii;

    s = StrCompress(memPool, delstring);
    if (s == NULL) {
        return GCS(MSG_Err_Mem);
    }
    len = strlen(s);
    ascii = StrIsASCII(s);
    *del = *s;
    FreeString(memPool, s);
    if (len != 1 || !ascii ||
            (*del == '"' && ImExportSettings.ies_DoubleQuotes)) {
        return GCS(MSG_ImExportErr_InvalidDelimiter);
    }
    return NULL;
}

static bool checkImExportSettings(Object *obj, char *recdel, char *fielddel)
{
    struct ImExportData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    struct MemPool *memPool = ProjectMemPool(pr);
    const char *error;

    if (ImExportSettings.ies_Filename[0] == 0) {
        error = GCS(MSG_Err_NoFilename);
        set(obj, MUIA_Window_Title, error);
        set(obj, MUIA_Window_ActiveObject, d->d_Filename);
        GuiBeep();
        return FALSE;
    }
    error = checkDelimiterString(
            ImExportSettings.ies_RecordDelimiter, memPool, recdel);
    if (error) {
        set(obj, MUIA_Window_Title, error);
        set(obj, MUIA_Window_ActiveObject, d->d_RecordDelimiter);
        GuiBeep();
        return FALSE;
    }
    error = checkDelimiterString(
            ImExportSettings.ies_FieldDelimiter, memPool, fielddel);
    if (error) {
        set(obj, MUIA_Window_Title, error);
        set(obj, MUIA_Window_ActiveObject, d->d_FieldDelimiter);
        GuiBeep();
        return FALSE;
    }
    return TRUE;
}

static void HOOK_PROTO(closeImExport, Object *obj, ulong *ok)
{
    struct ImExportData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);
    char recdel, fielddel;
    bool rc = FALSE;

    saveImExportSettings(obj);

    if (*ok) {
        if (!checkImExportSettings(obj, &recdel, &fielddel)) {
            return;
        }

        switch (d->d_Action) {
            case IEA_Import:
                rc = ImportRecords(d->d_Table,
                        ImExportSettings.ies_OverwriteRecords,
                        ImExportSettings.ies_Filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;

#ifdef ENABLE_LOG_APPLY_CHANGES
            case IEA_ApplyChanges:
                rc = ImportApplyChanges(pr,
                        ImExportSettings.ies_Filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;
#endif
            case IEA_Export:
                rc = ExportRecords(d->d_Table, ImExportSettings.ies_Filter,
                        ImExportSettings.ies_Filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;
            case IEA_ExportSFW:
                rc = ExportSFWList(pr, d->d_SFWList,
                        ImExportSettings.ies_Filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;
        }
        if (!rc) {
            return;
        }
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposeImExport(app, obj);
}

static void createImExport(struct Project *pr, struct Table *tab, GuiObject sfw,
        enum ImExportAction action)
{
    Object *app = GuiApplication;
    struct MemPool *memPool = ProjectMemPool(pr);
    struct ImExportData *d;
    Object *extraLabel = NULL, *extraItem = NULL;
    Object *obj, *frame, *buttons;
    const char *filenameBH = NULL;
    const char *filenameLabelStr = NULL;
    const char *imExportStr = NULL;
    const char *windowTitleStr = NULL;
    const char *windowHelpNode = NULL;
    ulong windowId = 0;
    bool hasView = FALSE;
    bool hasExtra = FALSE;

    if (app == NULL || memPool == NULL) {
	return;
    }

    d = PoolAllocMemZero(memPool, sizeof(struct ImExportData));
    if (d == NULL) {
        return;
    }
    d->d_Project = pr;
    d->d_Table = tab;
    d->d_SFWList = sfw;
    d->d_Action = action;

    d->d_Filename = StringObject,
        MUIA_CycleChain, TRUE,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_String_MaxLen, FILENAME_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, ImExportSettings.ies_Filename,
	End;

    switch (action)
    {
        case IEA_Import:
            filenameLabelStr = GCS(MSG_Import_File);
            filenameBH = GCS(MSG_Import_FileBH);
            imExportStr = GCS(MSG_Import);
            windowTitleStr = GCS(MSG_Import);
            windowHelpNode = "Importing records";
            windowId = MAKE_ID('_', 'I', 'M', 'P');
            hasView = TRUE;
            break;
#ifdef ENABLE_LOG_APPLY_CHANGES
        case IEA_ApplyChanges:
            filenameLabelStr = GCS(MSG_Import_LogFile);
            filenameBH = GCS(MSG_Import_LogFileBH);
            imExportStr = GCS(MSG_Import);
            windowTitleStr = GCS(MSG_ApplyChanges);
            windowHelpNode = "Applying changes";
            windowId = MAKE_ID('_', 'A', 'P', 'C');
            hasView = TRUE;
            if (*EmptyNullStr(pget(d->d_Filename, MUIA_String_Contents)) == 0) {
                set(d->d_Filename, MUIA_String_Contents,
                        LogFilename(ProjectLog(pr)));
            }
            break;
#endif
        case IEA_Export:
            filenameLabelStr = GCS(MSG_Export_File);
            filenameBH = GCS(MSG_Export_Records_FileBH);
            imExportStr = GCS(MSG_Export);
            windowTitleStr = GCS(MSG_Export);
            windowHelpNode = "Exporting records";
            windowId = MAKE_ID('_', 'E', 'X', 'P');
            break;
        case IEA_ExportSFW:
            filenameLabelStr = GCS(MSG_Export_File);
            filenameBH = GCS(MSG_Export_SFWList_FileBH);
            imExportStr = GCS(MSG_Export);
            windowTitleStr = GCS(MSG_Export);
            windowHelpNode = "Exporting a list";
            windowId = MAKE_ID('_', 'E', 'X', 'L');
            break;
    }
    set(d->d_Filename, MUIA_ShortHelp, filenameBH);
    set(d->d_Filename, MUIA_ControlChar, Shortcut(filenameLabelStr));

    if (hasView) {
        d->d_View = TextObject,
            ButtonFrame,
      	    MUIA_Weight, 0,
            MUIA_Background, MUII_ButtonBack,
            MUIA_Text_SetVMax, FALSE,
	    MUIA_Text_Contents, GCS(MSG_View),
            MUIA_Text_PreParse, "\33c",
	    MUIA_InputMode, MUIV_InputMode_RelVerify,
            MUIA_CycleChain, TRUE,
            MUIA_ShortHelp, GCS(MSG_ViewBH),
            End;

        d->d_Edit = TextObject,
            ButtonFrame,
  	    MUIA_Weight, 0,
            MUIA_Background, MUII_ButtonBack,
            MUIA_Text_SetVMax, FALSE,
	    MUIA_Text_Contents, GCS(MSG_Edit),
	    MUIA_Text_PreParse, "\33c",
	    MUIA_InputMode, MUIV_InputMode_RelVerify,
            MUIA_CycleChain, TRUE,
            MUIA_ShortHelp, GCS(MSG_EditBH),
            End;
    }

    d->d_RecordDelimiter = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_ImExport_RecordDelimiterBH),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_ImExport_RecordDelimiter)),
	MUIA_String_MaxLen, DELIMITER_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, ImExportSettings.ies_RecordDelimiter,
	End;

    d->d_FieldDelimiter = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_ImExport_FieldDelimiterBH),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_ImExport_FieldDelimiter)),
	MUIA_String_MaxLen, DELIMITER_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, ImExportSettings.ies_FieldDelimiter,
	End;

    d->d_DoubleQuotes = MUI_MakeObject(MUIO_Checkmark, 
	    (IPTR)GCS(MSG_ImExport_DoubleQuotes));

    switch (action)
    {
        case IEA_Import:
            extraLabel = Label1(GCS(MSG_Import_OverwriteRecords));
            extraItem = d->d_OverwriteRecords = MUI_MakeObject(MUIO_Checkmark, 
                    (IPTR)GCS(MSG_Import_OverwriteRecords));
            hasExtra = TRUE;
            break;
        case IEA_Export:
            extraLabel = Label1(GCS(MSG_Filter));
            extraItem = d->d_Filter = MUI_MakeObject(MUIO_Checkmark, 
                    (IPTR)GCS(MSG_Filter));
            hasExtra = TRUE;
            break;
        default:
            break;
    }

    if (hasExtra) {
        extraItem = HGroup,
            Child, VCenter(extraItem),
            Child, HSpace(0),
            End;
    }

    frame = VGroup,
        MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
        MUIA_Group_Columns, 2,
        Child, Label2(filenameLabelStr),
        Child, HGroup,
            GroupSpacing(0),
            MUIA_Group_SameHeight, TRUE,
            Child, PopaslObject,
	        MUIA_Popasl_Type, ASL_FileRequest,
	        MUIA_Popstring_String, d->d_Filename,
	        MUIA_Popstring_Button, GuiPopButtonSetCWD(MUII_PopFile, pr),
	        End,
            hasView? Child : TAG_IGNORE, d->d_View,
            hasView? Child : TAG_IGNORE, d->d_Edit,
            End,
        Child, Label2(GCS(MSG_ImExport_RecordDelimiter)),
        Child, HGroup,
            Child, d->d_RecordDelimiter,
            Child, HSpace(0),
            End,
        Child, Label2(GCS(MSG_ImExport_FieldDelimiter)),
        Child, HGroup,
            Child, d->d_FieldDelimiter,
            Child, HSpace(0),
            End,
        Child, Label1(GCS(MSG_ImExport_DoubleQuotes)),
        Child, HGroup,
            Child, VCenter(d->d_DoubleQuotes),
            Child, HSpace(0),
            End,
        hasExtra? Child : TAG_IGNORE, extraLabel,
        hasExtra? Child : TAG_IGNORE, extraItem,
        End;

    d->d_ImExport = SimpleButton(imExportStr);
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
        MUIA_Group_SameWidth, TRUE,
	Child, d->d_ImExport,
	Child, d->d_Cancel,
        End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_Title, windowTitleStr,
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, windowId,
        MUIA_Window_NoMenus, TRUE,
        MUIA_Window_AppWindow, TRUE,
        MUIA_HelpNode, windowHelpNode,
        MUIA_UserData, d,
	WindowContents, VGroup,
            Child, frame,
            Child, buttons,
            End,
	End;

    if (obj == NULL) {
        PoolFreeMem(memPool, d, sizeof(struct ImExportData));
	return;
    }

    MUIAddAppMessageNotification(obj, d->d_Filename);

    SetAttrs(d->d_DoubleQuotes,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_ImExport_DoubleQuotesBH),
        MUIA_Selected, ImExportSettings.ies_DoubleQuotes,
        TAG_DONE);
    if (d->d_Filter) {
        SetAttrs(d->d_Filter,
            MUIA_CycleChain, TRUE,
            MUIA_ShortHelp, GCS(MSG_Export_FilterBH),
            MUIA_Selected, ImExportSettings.ies_Filter,
            TAG_DONE);
    }
    if (d->d_OverwriteRecords) {
        SetAttrs(d->d_OverwriteRecords,
            MUIA_CycleChain, TRUE,
            MUIA_ShortHelp, GCS(MSG_Import_OverwriteRecordsBH),
            MUIA_Selected, ImExportSettings.ies_OverwriteRecords,
            TAG_DONE);
    }
    set(d->d_ImExport, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    if (d->d_View) {
        MUIAddCallback(d->d_View, MUIA_Pressed, FALSE, 
                GuiObjectViewFileCallback, d->d_Filename);
    }
    if (d->d_Edit) {
        MUIAddCallback(d->d_Edit, MUIA_Pressed, FALSE, 
                GuiObjectEditFileCallback, d->d_Filename);
    }

    DoMethod(d->d_Filename,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_ImExport);
    DoMethod(d->d_RecordDelimiter,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_ImExport);
    DoMethod(d->d_FieldDelimiter,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_ImExport);

    MUIAddCallback2(obj, MUIA_Window_CloseRequest, TRUE, 
            obj, closeImExport, FALSE);
    MUIAddCallback2(d->d_Cancel, MUIA_Pressed, FALSE, 
            obj, closeImExport, FALSE);
    MUIAddCallback2(d->d_ImExport, MUIA_Pressed, FALSE, 
            obj, closeImExport, TRUE);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if (!GuiOpenWindow(obj)) {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeImExport(app, obj);
        return;
    }

    set(obj, MUIA_Window_ActiveObject, d->d_Filename);
}

void GuiImportRecords(struct Table *tab)
{
    createImExport(TableProject(tab), tab, NULL, IEA_Import);
}

#ifdef ENABLE_LOG_APPLY_CHANGES
void GuiApplyChanges(struct Project *pr)
{
    createImExport(pr, NULL, NULL, IEA_ApplyChanges);
}
#endif

void GuiExportRecords(struct Table *tab)
{
    createImExport(TableProject(tab), tab, NULL, IEA_Export);
}

void GuiExportSFWList(struct Project *pr, GuiObject sfw)
{
    createImExport(pr, NULL, sfw, IEA_ExportSFW);
}

