/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** Menu_mui.c
**
** Creation date: 15.12.1994
** Last modified: 02.05.2022
*/

#include "menu.h"
#include "extclasses/ExtApplication_mcc.h"
#include "extclasses/ExtWindow_mcc.h"
#include "menuproject.h"
#include "menuprefs.h"
#include "menulog.h"
#include "menutable.h"
#include "menueditor.h"
#include "menuprogram.h"
#include "menuhelp.h"
#include "main.h"
#include "status.h"
#include "project.h"
#include "log.h"
#include "lang/program.h"
#include "display.h"
#include "structureeditor.h"
#include "preferences.h"
#include "requesters.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"
#include "amiga/exec.h"

/*
** special variables
*/

ulong MenuItemChecked = FALSE;

static void callchecked(struct Project *pr, intptr_t funcptr, GuiObject win)
{
    void (*func)(struct Project *, intptr_t, GuiObject);

    func = (void (*)(struct Project *, intptr_t, GuiObject))funcptr;
    func(pr, MenuItemChecked, win);
}

static void saveprojectinit(struct Project *pr, GuiObject go, intptr_t dummy)
{
    if (!ProjectAdminMode(pr) && ProjectChanges(pr) == 0) {
        set(go, MUIA_Menuitem_Enabled, FALSE);
    }
}

#ifdef ENABLE_ADMIN_MODE
static void adminmodeinit(struct Project *pr, GuiObject go, intptr_t dummy)
{
    bool admin = ProjectAdminMode(pr);
    const char *title= GCS(admin? MSG_MenuItem_UserMode:MSG_MenuItem_AdminMode);

    if (pget(go, MUIA_Menuitem_Title) != title) {
	set(go, MUIA_Menuitem_Title, title);
    }
}
#endif

#ifdef ENABLE_STRUCTURE_EDITOR
static void structureeditorinit(struct Project *pr, GuiObject go,intptr_t dummy)
{
    GuiObject se = ProjectStructureEditor(pr);
    const char *title = GCS(se? MSG_MenuItem_CloseStructureEditor : 
	    MSG_MenuItem_StructureEditor);

    if (pget(go, MUIA_Menuitem_Title) != title) {
	set(go, MUIA_Menuitem_Title, title);
    }
}
#endif

static void reverttosavedinit(struct Project *pr, GuiObject go, intptr_t dummy)
{
    bool hasChanged = ProjectChanges(pr) != 0;
    const char *title = GCS(hasChanged? MSG_MenuItem_RevertToSaved : 
		      MSG_MenuItem_Reload);

    if (pget(go, MUIA_Menuitem_Title) != title) {
	set(go, MUIA_Menuitem_Title, title);
    }
}

static void setactive(GuiObject go, bool active)
{
    if(nget(go, MUIA_Menuitem_Checked) != active)
        set(go, MUIA_Menuitem_Checked, active);
}

static void confirmdeleterecordinit(struct Project *pr, GuiObject go, 
		intptr_t dummy)
{
    setactive(go, GetPrefsValue(pr, PREFS_CONFIRMDELETERECORD));
}

#ifdef ENABLE_STRUCTURE_EDITOR
static void sorttriggerfunctionsinit(struct Project *pr, GuiObject go, 
		intptr_t dummy)
{
    setactive(go, GetPrefsValue(pr, PREFS_PROGRAMSORTTRIGGERFUNCTIONS));
}
#endif

static long memtable[] = 
{ 1000000, 2000000, 4000000, 8000000, 16000000, 32000000, 0x7fffffff };

static void recordmeminit(struct Project *pr, GuiObject go, intptr_t num)
{
    setactive(go, GetPrefsValue(pr, PREFS_RECORDMEMORY) == memtable[num]);
}

static void extrabuttonsintabchaininit(struct Project *pr, GuiObject go,
    intptr_t dummy)
{
    setactive(go, GetPrefsValue(pr, PREFS_EXTRABUTTONSINTABCHAIN));
}

static void advanceonenterinit(struct Project *pr, GuiObject go,
    intptr_t dummy)
{
    setactive(go, GetPrefsValue(pr, PREFS_ADVANCEONENTER));
}

static void logactiveinit(struct Project *pr, GuiObject go, intptr_t dummy)
{
    setactive(go, LogActive(ProjectLog(pr)));
}

static void logmodeinit(struct Project *pr, GuiObject go, intptr_t num)
{
    setactive(go, MenuLogModeInit(pr, num));
}

static void logsetlabelinit(struct Project *pr, GuiObject go, intptr_t dummy)
{
    if (FindUserFunctionName(ProjectProgram(pr), logLabel)) {
        set(go, MUIA_Menuitem_Enabled, FALSE);
    }
}

static void cleanupexternalsourceinit(struct Project *pr, GuiObject go, 
		intptr_t dummy)
{
    setactive(go, GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL_CLEANUP));
}

static void confirmautoreloadinit(struct Project *pr, GuiObject go,
	       	intptr_t dummy)
{
    setactive(go, GetPrefsValue(pr, PREFS_CONFIRMAUTORELOAD));
}

static void confirmsavereorginit(struct Project *pr, GuiObject go,
	       	intptr_t dummy)
{
    setactive(go, GetPrefsValue(pr, PREFS_CONFIRMSAVEREORG));
}

static void vacuumafterreorginit(struct Project *pr, GuiObject go,
	       	intptr_t dummy)
{
    setactive(go, GetPrefsValue(pr, PREFS_VACUUMAFTERREORG));
}
static void confirmquitinit(struct Project *pr, GuiObject go,
	       	intptr_t dummy)
{
    setactive(go, GetPrefsValue(pr, PREFS_CONFIRMQUIT));
}

static void pathsrelativetoprojectinit(struct Project *pr, GuiObject go, 
	intptr_t dummy)
{
    setactive(go, GetPrefsValue(pr, PREFS_PATHSRELATIVETOPROJECT));
}

static void programsourceinit(struct Project *pr, GuiObject go, intptr_t num)
{
    setactive(go, GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL) == num);
}

static void programdebuginfoinit(struct Project *pr, GuiObject go,
	       	intptr_t dummy)
{
    setactive(go, GetPrefsValue(pr, PREFS_PROGRAMDEBUGINFO));
}

static void programobsoletefunctionsinit(struct Project *pr, 
        GuiObject go, intptr_t num)
{
    setactive(go, GetPrefsValue(pr, PREFS_PROGRAMOBSOLETEFUNCTIONS) == num);
}

static void projectopen(struct Project *pr)
{
    MenuProjectOpen(pr, NULL);
}

static void recordmem(struct Project *pr, intptr_t num, GuiObject go)
{
    SetPrefsValue(pr, PREFS_RECORDMEMORY, memtable[num]);
}

static void MenuPrefsMUI(struct Project *pr)
{
    GuiObject app = GuiApplication;

    if(app)
        DoMethod(app, MUIM_Application_OpenConfigWindow, 0);
}

#include "menu.share"

struct MenuInfo
{
    ulong mi_Status;
    GuiObject mi_GuiObjects[sizeof(MBNewMenu) / sizeof(struct MBNewMenu)];
} MenuInfo =
{
    (ulong)-1,
};

void UpdateMenu(GuiObject win, bool force)
{
    struct Project *pr = GuiWindowProject(win);
    ulong status = GetStatus(win);
    struct MenuInfo *mi = &MenuInfo;
    struct MBNewMenu *mnm;
    int i;

    if ((win != NULL && mi->mi_Status != status) || force) {
        for (mnm = MBNewMenu, i = 0; mnm->mnm_Type != MNM_END; mnm++, i++) {
            GuiObject item = mi->mi_GuiObjects[i];

            if (item != NULL) {
                bool enabled = (mnm->mnm_Status & status) == mnm->mnm_Status;

                if (nget(item, MUIA_Menuitem_Enabled) != enabled) {
                    set(item, MUIA_Menuitem_Enabled, enabled);
                }
                if (mnm->mnm_Init != NULL) {
                    mnm->mnm_Init(pr, item, mnm->mnm_UserData);
                }
            }
        }
        mi->mi_Status = status;
    }
}

static void HOOK_PROTO(MenuClick, GuiObject go, struct MBNewMenu **x)
{
    struct MBNewMenu *mnm = *x;
    GuiObject app = pget(go, MUIA_ApplicationObject);
    GuiObject win = pget(app, MUIA_ExtApplication_ActiveExtWindow);
    struct Project *pr = pget(win, MUIA_ExtWindow_Project);
    ulong status;

    if(pr == NULL || mnm == NULL || mnm->mnm_Callback == NULL)
	return;

    status = GetStatus(win);
    if((mnm->mnm_Status & status) != mnm->mnm_Status)
    {
        GuiBeep();
        return;
    }

    GuiStoreEdits();
    get(go, MUIA_Menuitem_Checked, &MenuItemChecked);
    mnm->mnm_Callback(pr, mnm->mnm_UserData, win);
    GuiSetCursor();
}

GuiObject AllocMenuStrip(GuiObject dummy)
{
    static bool first_time_called = TRUE;
    struct MenuInfo *mi = &MenuInfo;
    struct MBNewMenu *mnm;
    GuiObject menustrip = NULL, menu = NULL, menuitem = NULL, cur = NULL;
    int i;

    if (first_time_called) {
	first_time_called = FALSE;

	for (mnm = MBNewMenu; mnm->mnm_Type != MNM_END; mnm++) {
	    if (mnm->mnm_LabelId != NULL) {
	    	mnm->mnm_Label = GCS(mnm->mnm_LabelId);
            }
            if (mnm->mnm_ShortcutId != NULL) {
                mnm->mnm_Shortcut = GCS(mnm->mnm_ShortcutId);
            }
	}
    }

    menustrip = MenustripObject, End;
    if (menustrip == NULL) {
	return NULL;
    }

    for (mnm = MBNewMenu, i = 0; mnm->mnm_Type != MNM_END; mnm++, i++) {
	switch (mnm->mnm_Type) {
            case MNM_TITLE:
		mi->mi_GuiObjects[i] = menu = cur = MenuObject,
		    MUIA_Menu_Title, mnm->mnm_Label,
                    MUIA_UserData, mnm->mnm_HelpNode,
                    End;
		break;

	    case MNM_ITEM:
		mi->mi_GuiObjects[i] = menuitem = cur = MenuitemObject,
		    MUIA_Menuitem_Title, mnm->mnm_Label,
                    MUIA_UserData, mnm->mnm_HelpNode,
		    End;
		break;

	    case MNM_SUB:
		mi->mi_GuiObjects[i] = cur = MenuitemObject,
		    MUIA_Menuitem_Title, mnm->mnm_Label,
                    MUIA_UserData, mnm->mnm_HelpNode,
		    End;
		break;
	}
	if (cur == NULL) {
            MUI_DisposeObject(menustrip);
            return NULL;
 	}

        if (mnm->mnm_Flags & MNM_CHECKIT) {
            SetAttrs(cur,
                MUIA_Menuitem_Checkit, TRUE,
                MUIA_Menuitem_Exclude, mnm->mnm_MutualExclude,
                MUIA_Menuitem_Toggle, (mnm->mnm_MutualExclude == 0),
                TAG_DONE);
        }
	if (mnm->mnm_Shortcut) {
	    SetAttrs(cur,
                MUIA_Menuitem_Shortcut, mnm->mnm_Shortcut,
                MUIA_Menuitem_CommandString, mnm->mnm_ShortcutId != NULL,
                TAG_DONE);
        }

	switch (mnm->mnm_Type) {
	    case MNM_TITLE:
                DoMethod(menustrip, MUIM_Family_AddTail, cur);
		break;

            case MNM_ITEM:
		DoMethod(menu, MUIM_Family_AddTail, cur);
		break;

	    case MNM_SUB:
		DoMethod(menuitem, MUIM_Family_AddTail, cur);
		break;
	}
    }
    return menustrip;
}

static void HOOK_PROTO(showMenuHelp, Object *obj, char **x)
{
    GuiShowHelp(*x);
}

void MenuStripAddNotifications(void)
{
    struct MenuInfo *mi = &MenuInfo;
    struct MBNewMenu *mnm;
    GuiObject go;
    int i;

    for(mnm = MBNewMenu, i = 0; mnm->mnm_Type != MNM_END; mnm++, i++)
    {
        go = mi->mi_GuiObjects[i];

	if(mnm->mnm_Callback && go)
            MUIAddCallback(go, MUIA_Menuitem_Trigger, MUIV_EveryTime, MenuClick, mnm);
    }
    MUIAddCallback(GuiApplication, MUIA_Application_MenuHelp, MUIV_EveryTime,
       showMenuHelp, MUIV_TriggerValue);
}

