/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** mui.c
**
** Creation date: 07.09.1994
** Last modified: 01.08.2022
*/

#include "system.h"
#include "extclasses/ExtEditor_mcc.h"
#include "extclasses/ActGroup_mcc.h"
#include "extclasses/StringText_mcc.h"
#include "extclasses/NewCycle_mcc.h"
#include "extclasses/DDList_mcc.h"
#include "extclasses/RefFieldSelector_mcc.h"
#include "extclasses/SFWList_mcc.h"
#include "extclasses/EditLabels_mcc.h"
#include "extclasses/PopEditLabels_mcc.h"
#include "extclasses/DataTypes_mcc.h"
#include "extclasses/WindowButton_mcc.h"
#undef USE_SUBTASK
#ifdef USE_SUBTASK
#include "extclasses/SubTask_mcc.h"
#endif
#include "rexx_mui.h"
#include "preparse_mui.h"
#include "main.h"
#include "server.h"
#include "project.h"
#include "structureeditor.h"
#include "table.h"
#include "field.h"
#include "display.h"
#include "menu.h"
#include "menuproject.h"
#include "menueditor.h"
#include "queryeditor.h"
#include "gui.h"
#include "buildgui.h"
#include "buildguifield_mui.h"
#include "break.h"
#include "status.h"
#include "preferences.h"
#include "requesters.h"
#include "memory.h"
#include "strhandle.h"
#include "labels.h"
#include "guidosutil.h"
#include "dosutil.h"
#include "util.h"
#include "amiga/exec.h"
#include "amiga/font.h"
#include "amiga/font.h"
#include "appstrings_cat.h"
#include <proto/locale.h>
#include <proto/keymap.h>
#include <proto/rexxsyslib.h>
#include <rexx/rxslib.h>
#include <workbench/workbench.h>
#include <workbench/startup.h>
#include "builddate.h"
#include "version.h"

/* intuition stuff */

#if (defined __amigaos4__)
#include <stdarg.h>
APTR VARARGS68K DoSuperNew(struct IClass *cl,Object *obj,...)
{
    APTR    res;
    va_list va;

    va_startlinear(va,obj);
    res = (APTR)IDoSuperMethod(cl,obj,OM_NEW,va_getlinearva(va,ULONG),NULL);
    va_end(va);

    return res;
}
APTR DoSuperNewA(Class *cl, APTR obj, struct TagItem *tagList)
{
    struct opSet opSet;

    opSet.MethodID = OM_NEW;
    opSet.ops_AttrList = tagList;
    opSet.ops_GInfo = NULL;

    return (APTR)IDoSuperMethodA(cl, obj, (Msg)&opSet);
}
#elif (!defined __MORPHOS__)
APTR DoSuperNewA(Class *cl, APTR obj, struct TagItem *tagList)
{
    struct opSet opSet;

    opSet.MethodID = OM_NEW;
    opSet.ops_AttrList = tagList;
    opSet.ops_GInfo = NULL;

    return (APTR)DoSuperMethodA(cl, obj, (Msg)&opSet);
} 

#if (!defined __AROS__ && !defined __MORPHOS__ && !defined __amigaos4__)
Object *MUI_NewObject(char *classname,Tag tag1,...)
{
    return MUI_NewObjectA(classname, (struct TagItem *)&tag1);
}
#endif
#endif

static const char *const UsedClasses[] = 
{ 
#ifdef USE_BETTERSTRING
    "BetterString.mcc",
#endif
#ifdef USE_NLIST
    "NList.mcc",
    "NListviews.mcc",
#endif
#ifdef USE_TEXTEDITOR
    "TextEditor.mcc",
#endif
    NULL
};

const char version[] = 
	"$VER: " APP_NAME " " BB_VERSTR " (" BUILD_DATE ") " \
        "Copyright 1998-" BUILD_YEAR " Steffen Gutmann";

struct Library *MUIMasterBase = NULL;
#ifdef PROJECT_ICON_INLINE
struct Library *IconBase = NULL;
#endif
#if defined(__amigaos4__) || defined(__MORPHOS__)
struct Library *RexxSysBase = NULL;
#else
struct RxsLib *RexxSysBase = NULL;
#endif
#ifdef USE_HPDF
struct Library *HPDFBase = NULL;
#endif
#ifdef USE_SQLITE_LIBRARY
struct Library *SQLiteBase = NULL;
#endif

#define MUIMASTER_MIN 17
#define MUIMASTER_MIN_STR "17"

#ifdef __amigaos4__
struct MUIMasterIFace * IMUIMaster = NULL;
struct RexxSysIFace * IRexxSys = NULL;
struct Library *KeymapBase = NULL;
struct KeymapIFace *IKeymap = NULL;
struct Library *LocaleBase = NULL;
struct LocaleIFace *ILocale = NULL;
#endif

struct Locale *Locale = NULL;

#ifdef __AXRT__
int __enhancedpathmode = 1;  /* for AxRT */
#endif

Object *GuiApplication = NULL;

const char MUIBoldString[] = "\33b";
const char MUIItalicString[] = "\33i";
char GuiScreenTitle[51];
char DefaultRealFormat = FORMAT_DECIMALPOINT;
char DefaultDateFormat = FORMAT_DAYMONTHYEAR;
struct TextFont *Topaz8Font = NULL;

static struct IClass *MUIWindowIClass = NULL;
static struct IClass *MUIAreaIClass = NULL;

static const char ThirdPartyCopyright[] =
	"MUI (C) 1992-2013 Stefan Stuntz\n"
#ifdef USE_NLIST
	"NList.mcc (C) 2001-2021 NList Open Source Team\n"
#endif
#ifdef USE_BETTERSTRING
	"BetterString.mcc (C) 2005-2021 BetterString Open Source Team\n"
#endif
#ifdef USE_TEXTEDITOR
	"TextEditor.mcc (C) 2005-2021 TextEditor Open Source Team\n"
#endif
        ;

static const char CodesetsCopyright[] =
        "codesets.library (C) 2005-2018 codesets.library Open Source Team\n";

extern struct List ProjectList;
extern struct MenuInfo MenuInfo;
extern struct Library *CodesetsBase;

IPTR nget(GuiObject go, LONG attr)
{
    IPTR n = 0;

    if (go != NULL) {
        get(go, attr, &n);
    }
    return n;
}

void CloseGui(void)
{
    long num = DosProcessesRunning(), oldnum = 0;
    ulong sigs = 0;

    BreakExit(NULL);

    while (num > 0) {
        if (num != oldnum) {
            fprintf(stderr, 
                    "BeeBase: waiting for %ld child processes to exit.\n", num);
            oldnum = num;
        }
        DoMethod(GuiApplication, MUIM_Application_NewInput, &sigs);
        if (sigs) {
            sigs = Wait(sigs | SIGBREAKF_CTRL_C);
        }
        num = DosProcessesRunning();
    }

    if (GuiApplication) {
        MUI_DisposeObject(GuiApplication);
	GuiApplication = NULL;
    }

#ifdef USE_SUBTASK
    SubTaskCleanup();
#endif
    WindowButtonCleanup();
    DataTypesCleanup();
    PopEditLabelsCleanup();
    EditLabelsCleanup();
    SFWListCleanup();
    RefFieldSelectorCleanup();
    DDListCleanup();
    StringTextCleanup();
    ActGroupCleanup();
    NewCycleCleanup();
    ExtEditorCleanup();
    ExtWindowCleanup();
    ExtApplicationCleanup();

    PreParseExit();
    FontExit();
    Topaz8Font = NULL;
    AppStringsExit();

    CodesetExit();

    if (Locale != NULL) {
	CloseLocale(Locale);
	Locale = NULL;
    }
    
    MUIWindowIClass = NULL;
    MUIAreaIClass = NULL;

#ifdef USE_SQLITE_LIBRARY
    if (SQLiteBase) {
        CloseLibrary(SQLiteBase);
        SQLiteBase = NULL;
    }
#endif

#ifdef USE_HPDF
    if (HPDFBase) {
        CloseLibrary(HPDFBase);
        HPDFBase = NULL;
    }
#endif

#ifdef PROJECT_ICON_INLINE
    if (IconBase) {
        CloseLibrary(IconBase);
	IconBase = NULL;
    }
#endif

#ifdef __amigaos4__
    if (ILocale) {
	DropInterface((struct Interface*)ILocale);
	ILocale = NULL;
    }
    if (LocaleBase) {
	CloseLibrary((struct Library *)LocaleBase);
	LocaleBase = NULL;
    }
    if (IRexxSys) {
        DropInterface((struct Interface*)IRexxSys);
	IRexxSys = NULL;
    }
    if (IMUIMaster) {
        DropInterface((struct Interface*)IMUIMaster);
	IMUIMaster = NULL;
    }
    if (IKeymap) {
        DropInterface((struct Interface*)IKeymap);
	IKeymap = NULL;
    }
    if (KeymapBase) {
        CloseLibrary(KeymapBase);
	KeymapBase = NULL;
    }
#endif /* __amigaos4__ */
    if (RexxSysBase) {
	CloseLibrary((struct Library *)RexxSysBase);
	RexxSysBase = NULL;
    }
    if (MUIMasterBase) {
	CloseLibrary(MUIMasterBase);
	MUIMasterBase = NULL;
    }

    RestoreExecTrap();
}

const char *OpenGui(int *pargc, char **argv)
{
    char buf[21] = "", *p;
    const char *error;

    error = InstallExecTrap();
    if (error) {
	return error;
    }

    /*
    ** Reading Kickstart Version.Revision
    */
    if (GetVar((STRPTR)"Kickstart", (STRPTR)buf, 20, 0) > 0 && 
	    (p = CodesetStrchr(buf, '.')) != NULL)
    {
        OSVersion = atol(buf);
        OSRevision = atol(p+1);
    }

    if(!(MUIMasterBase = OpenLibrary((STRPTR)MUIMASTER_NAME, MUIMASTER_MIN)))
	return("Failed to open "MUIMASTER_NAME" V"  MUIMASTER_MIN_STR".");

#if defined(__amigaos4__) || defined(__MORPHOS__)
    if (!(RexxSysBase = (struct Library *)OpenLibrary((STRPTR)RXSNAME, 0)))
	return("Failed to open "RXSNAME".");
#else
    if (!(RexxSysBase = (struct RxsLib *)OpenLibrary((STRPTR)RXSNAME, 0)))
	return("Failed to open "RXSNAME".");
#endif

#ifdef PROJECT_ICON_INLINE
    IconBase = OpenLibrary((STRPTR)"icon.library", 0);
#endif

#ifdef USE_SQLITE_LIBRARY
    SQLiteBase = OpenLibrary("sqlite.library", 52L);
    if (SQLiteBase == NULL) {
	return "Failed to open sqlite.library V52.";
    }
#endif

#ifdef __amigaos4__
    if(!(IMUIMaster = (struct MUIMasterIFace*) GetInterface(MUIMasterBase, "main", 1, NULL)))
	return("Failed to interface "MUIMASTER_NAME" V"  MUIMASTER_MIN_STR".");

    KeymapBase = OpenLibrary("keymap.library", 36);
    if (KeymapBase) {
        IKeymap = (struct KeymapIFace *)GetInterface(KeymapBase, "main",1,NULL);
    }
    if (!KeymapBase || !IKeymap) {
    	return("Failed to open keymap.library");
    }
    LocaleBase = OpenLibrary((STRPTR)"locale.library", 38);
    if (LocaleBase) {
        ILocale = (struct LocaleIFace *)GetInterface(LocaleBase, "main",1,NULL);
    }
    if (!LocaleBase || !ILocale) {
	return("Failed to open locale.library V38");
    }
#endif

    MUIWindowIClass = MUI_GetClass((ClassID)MUIC_Window);
    MUIAreaIClass = MUI_GetClass((ClassID)MUIC_Area);

    Locale = OpenLocale(NULL);	/* cannot fail according to AutoDocs */

    error = CodesetInit();
    if (error) {
	return error;
    }

    AppStringsInit();
    FontInit();
    PreParseInit();
    Topaz8Font = FontTextFont("topaz/8", NULL);

    if(!ExtApplicationInit())
        return("Unable to initialize private ExtApplication class!");
    if(!ExtWindowInit())
        return("Unable to initialize private ExtWindow class!");
    if(!ExtEditorInit())
        return("Unable to initialize private ExtEditor class!");
    if(!NewCycleInit())
        return("Unable to initialize private NewCycle class!");
    if(!ActGroupInit())
        return("Unable to initialize private ActGroup class!");
    if(!StringTextInit())
        return("Unable to initialize private StringText class!");
    if(!DDListInit())
        return("Unable to initialize private DDList class!");
    if(!RefFieldSelectorInit())
        return("Unable to initialize private RefFieldSelector class!");
    if(!SFWListInit())
        return("Unable to initialize private SFWList class!");
    if(!EditLabelsInit())
        return("Unable to initialize private EditLabels class!");
    if(!PopEditLabelsInit())
        return("Unable to initialize private PopEditLabels class!");
    if(!DataTypesInit())
        return("Unable to initialize private DataTypes class!");
    if(!WindowButtonInit())
        return("Unable to initialize private WindowButton class!");
#ifdef USE_SUBTASK
    if(!SubTaskInit())
        return("Unable to initialize private SubTask class!");
#endif

    /*
    ** try to determine date format and decimal point
    */
    if(Locale != NULL)
    {
        const char *datestring = (const char *)Locale->loc_ShortDateFormat;
        const char decimalpoint = Locale->loc_DecimalPoint[0];
        const char *s, *d = "de", *m = "bBhm", *y = "yY";

        if(decimalpoint == '.')
            DefaultRealFormat = FORMAT_DECIMALPOINT;
        else if(decimalpoint == ',')
            DefaultRealFormat = FORMAT_DECIMALCOMMA;

        /*
        ** datestring = day month year, ... ?
        */
        if((s = datestring) != NULL && (s = strpbrk(s, d)) != NULL &&
            (s = strpbrk(s, m)) != NULL && (s = strpbrk(s, y)) != NULL)
                DefaultDateFormat = FORMAT_DAYMONTHYEAR;
        else if((s = datestring) != NULL && (s = strpbrk(s, m)) != NULL &&
            (s = strpbrk(s, d)) != NULL && (s = strpbrk(s, y)) != NULL)
                DefaultDateFormat = FORMAT_MONTHDAYYEAR;
        else if((s = datestring) != NULL && (s = strpbrk(s, y)) != NULL &&
            (s = strpbrk(s, m)) != NULL && (s = strpbrk(s, d)) != NULL)
                DefaultDateFormat = FORMAT_YEARMONTHDAY;
    }

    sprintf(GuiScreenTitle, APP_NAME " V%s", BB_VERSTR);

    GuiApplication = ExtApplicationObject,
        MUIA_Application_Title,       APP_NAME,
	MUIA_Application_Version,     "$VER: " APP_NAME " " BB_VERSTR,
	MUIA_Application_Copyright,   "(C) 1998-" BUILD_YEAR " Steffen Gutmann",
	MUIA_Application_Author,      "Steffen Gutmann",
	MUIA_Application_Description, "Relational programmable database",
	MUIA_Application_Base,	      APP_NAME,	/* ARexx port */
	MUIA_Application_Commands,    MUI_Commands,
	MUIA_Application_Menustrip,   AllocMenuStrip(NULL),
        MUIA_Application_HelpFile,    GetPrefsString(NULL, PREFS_HELPFILENAME),
        MUIA_HelpNode,                "Main",
        MUIA_Application_UsedClasses, UsedClasses,
      	End;
    if(GuiApplication == NULL)
        return("Unable to create application!");

    UpdateMenu(NULL, TRUE);
    MenuStripAddNotifications();

    return NULL;
}

void GuiPrefsUpdated(void)
{
    if (GuiApplication)
    {
	set(GuiApplication, MUIA_Application_HelpFile, 
		GetPrefsString(NULL, PREFS_HELPFILENAME));
    }
    UpdateMenu(GuiActiveWindow(), TRUE);
}

#if 0
void GuiFlush(void)
{
    Object *app = GuiApplication;

    if(app)
        DoMethod(app, MUIM_Application_InputBuffered);
}
#endif

void GuiSendQuitMessage(void)
{
    Object *app = GuiApplication;

    if(app)
	DoMethod(app, MUIM_Application_ReturnID, MUIV_Application_ReturnID_Quit);
}

const char *GuiLanguage(void)
{
    static char lang[128] = "";
    char *t;

    if(Locale != NULL)
	CodesetStrNBytesCopy(lang, (const char *)Locale->loc_LanguageName, 127);

    t = CodesetStrrchr(lang, '.');
    if(t != NULL && !CodesetStrCmp(t, ".language"))
        *t = 0;

    return(lang);
}

static const char *GetScreenPubName(const struct Screen *screen)
{
    struct List *list = LockPubScreenList();
    struct PubScreenNode *psn;
    const char *name = NULL;

    if (list) {
	FORNODE(psn, list) {
	    if (psn->psn_Screen == screen) {
		name = psn->psn_Node.ln_Name;
		break;
	    }
	}
    }
    UnlockPubScreenList();

    return name;
}

const char *GuiPubScreenName(void)
{
    struct Window *win = GuiAnyOpenedWindow();
    const char *name = NULL;

    if (win) {
	name = GetScreenPubName(win->WScreen);
    }

    return name;
}

void GuiAbout(void)
{
#ifdef BEEBASE_LIGHT
    Ask(AT_INFO, GCS(MSG_StarOk), GCS(MSG_AboutBeeBase_Light), 
	    BB_VERSTR, BUILD_YEAR, ThirdPartyCopyright, 
            CodesetsBase? CodesetsCopyright : "");
#else
    Ask(AT_INFO, GCS(MSG_StarOk), GCS(MSG_AboutBeeBase_GPL),
            BB_VERSTR, BUILD_YEAR, ThirdPartyCopyright,
            CodesetsBase? CodesetsCopyright : "");
#endif
}

void AboutMUI(void)
{
    GuiObject app = GuiApplication;

    if (app)
        DoMethod(app, MUIM_Application_AboutMUI, NULL);
}

void GuiShowHelp(const char *line)
{
    const char *filename = pget(GuiApplication, MUIA_Application_HelpFile);
    GuiObject go = GuiActiveObject(GuiActiveWindow());

    while (line == NULL && go != NULL)
    {
	line = pget(go, MUIA_HelpNode);
        go = pget(go, MUIA_Parent);
    }
    if (line == NULL)
        line = "Main";

    DoMethod(GuiApplication, MUIM_Application_ShowHelp, ~0, filename, line, 0);
}

static void HOOK_PROTO(appMessageCallback, GuiObject str, struct AppMessage **x)
{
    struct AppMessage *am = *x;
    static char name[501];

    if (am) {
        struct WBArg *arg = am->am_ArgList;
        int i;

        for (i = 0; i < am->am_NumArgs; i++) {
            if (arg[i].wa_Name[0] == 0) {
                continue;
            }

            if (!NameFromLock(arg[i].wa_Lock, (STRPTR)name, sizeof(name))) {
                continue;
            }
            if (!AddPart((STRPTR)name, (CONST_STRPTR)arg[i].wa_Name,
                        sizeof(name))) {
                continue;
            }

            set(str, MUIA_String_Contents, name);
            break;      // accept first entry
        }
    }
}

void MUIAddAppMessageNotification(GuiObject source, GuiObject target)
{
    MUIAddCallback2(source, MUIA_AppMessage, MUIV_EveryTime,
            target, appMessageCallback, MUIV_TriggerValue);
}

#if 0
void GuiIconifyApplication(void) 
{
    if (GuiApplication) {
        set(GuiApplication, MUIA_Application_Iconified, TRUE);
    }
}
#endif

void HOOK_PROTO(GuiObjectViewFileCallback, GuiObject go, GuiObject *x)
{
    GuiObjectViewFile(*x);
}

void HOOK_PROTO(GuiObjectEditFileCallback, GuiObject go, GuiObject *x)
{
    GuiObjectEditFile(*x);
}

static ulong guiSleep = 0;

ulong GuiSleep(void)
{
    return(guiSleep);
}

void GuiSetSleep(bool sleep)
{
    GuiObject app = GuiApplication;

    if(app)
        set(app, MUIA_Application_Sleep, sleep);
    guiSleep += sleep? 1 : -1;
}

#if 0
bool MUIFindObject(Object *root, Object *find)
{
    struct List *list = NULL;
    struct Node *node;

    if(root == find)
        return(TRUE);
    if(!get(root, MUIA_Group_ChildList, &list) || list == NULL)
        return(FALSE);

    node = HEAD(list);
    while(root = NextObject(&node))
        if(MUIFindObject(root, find))
            return(TRUE);
    return(FALSE);
}
#endif

bool GuiOpenWindow(GuiObject wi)
{
    static ulong id = 0;

    if(wi)
    {
	if(nget(wi, MUIA_Window_Open))
	    return(TRUE);			/* already open */

	set(wi, MUIA_Window_Open, TRUE);
	if(nget(wi, MUIA_Window_Open))
        {
#if 0  /* it's strange but this doesn't work */
	    struct Project *pr = pget(wi, MUIA_ExtWindow_Project);
	    long handle = ProjectExclusiveHandle(pr);

	    if(pr != NULL && handle != HANDLE_NONE && handle != HANDLE_USER)
	    {
		set(wi, MUIA_Window_Sleep, TRUE);
	    }
#endif
            if(nget(wi, MUIA_Window_ID) == 0)   /* no window id? */
                set(wi, MUIA_Window_ID, ++id);

	    return(TRUE);
        }
    }

    return(FALSE);
}

short GuiObjectWeight(GuiObject obj, bool horiz)
{
    return (short)nget(obj, horiz? MUIA_HorizWeight : MUIA_VertWeight);
}

void GuiSetEnabled(GuiObject go, bool enabled)
{
    if (go) {
        set(go, MUIA_Disabled, !enabled);
    }
}

bool GuiGetEnabled(GuiObject go)
{
    bool enabled = FALSE;

    if (go) {
        enabled = !nget(go, MUIA_Disabled);
    }
    return enabled;
}

static bool MUIObjectIsA(Object *obj, struct IClass *isa)
{
    if (obj) {
        struct IClass *class = OCLASS(obj);

        while (class) {
            if (class == isa) {
                return TRUE;
            }
            class = class->cl_Super;
        }
    }

    return FALSE;
}

#define MUI_HAS_WINDOW_OBJECT_BUG \
    (ExecIsMorphOS() && \
     MUIMasterBase->lib_Version == 20 && \
     MUIMasterBase->lib_Revision < 6848)

GuiObject GuiObjectWindow(GuiObject go)
{
    GuiObject win = NULL;

    if (MUIObjectIsA(go, MUIWindowIClass)) {
        win = go;
    } else if (!MUI_HAS_WINDOW_OBJECT_BUG) {
        win = pget(go, MUIA_WindowObject);
    } else if (MUIObjectIsA(go, MUIAreaIClass) && muiRenderInfo(go)) {
        win = _win(go);
    }

    return win;
}

struct Project *GuiObjectProject(GuiObject go)
{
    return GuiWindowProject(GuiObjectWindow(go));
}

char **MUIListLabels(struct MemPool *memPool, GuiObject list)
{
    char **labels;
    const char *str = NULL;
    long num, i;

    if (memPool == NULL || list == NULL) {
	return NULL;
    }

    num = nget(list, MUIA_List_Entries);
    labels = AllocLabels(memPool, num);
    if (labels == NULL) {
	return NULL;
    }

    for (i = 0; i < num; i++) {
        DoMethod(list, MUIM_List_GetEntry, i, &str);
        str = EmptyNullStr(str);        /* be sure we have a valid string */
        labels[i] = AllocString(memPool, str);
	if (labels[i] == NULL) {
	    FreeLabelsNum(memPool, labels, num);
	    return NULL;
	}
    }
    return labels;
}

void MUIListSetLabels(GuiObject lv, char **labels)
{
    long active = nget(lv, MUIA_List_Active);
    long first = nget(lv, MUIA_List_First);
    long visible = nget(lv, MUIA_List_Visible);

    set(lv, MUIA_List_Quiet, TRUE);
    DoMethod(lv, MUIM_List_Clear);
    if(labels != NULL)
        DoMethod(lv, MUIM_List_Insert, labels, -1, MUIV_List_Insert_Bottom);
    if(first < 0)
        first = 0;
    if(active < 0)
        active = 0;
    if(visible < 1)
        visible = 1;
    DoMethod(lv, MUIM_List_Jump, first + visible - 1);
    set(lv, MUIA_List_Quiet, FALSE);
    set(lv, MUIA_List_Active, active);
}

void MUIListSetListWithFilter(GuiObject lv, struct List *list, 
        struct MemPool *memPool, bool (*filter)(const struct Node *node))
{
    long active = nget(lv, MUIA_List_Active);
    long first = nget(lv, MUIA_List_First);
    long visible = nget(lv, MUIA_List_Visible);
    bool open = nget(GuiObjectWindow(lv), MUIA_Window_Open);
    struct Node *node, **nodearray=NULL, **n;
    long num = NumNodes(list);

    set(lv, MUIA_List_Quiet, TRUE);
    DoMethod(lv, MUIM_List_Clear);
    if (num > 0) {
	nodearray = n = PoolAllocMem(memPool, num * sizeof(struct Node *));
	if (nodearray != NULL) {
	    FORNODE(node, list) {
                if (filter == NULL || filter(node)) {
                    *n++ = node;
                }
            }
            DoMethod(lv, MUIM_List_Insert, nodearray, n - nodearray, 
                    MUIV_List_Insert_Bottom);
	    PoolFreeMem(memPool, nodearray, num * sizeof(struct Node *));
	} else {			/* slow fallback routine */
	    FORNODE(node, list) {
                if (filter == NULL || filter(node)) {
                    DoMethod(lv, MUIM_List_InsertSingle, node, 
                            MUIV_List_Insert_Bottom);
                }
            }
        }
    }
    if (first < 0) {
        first = 0;
    }
    if (visible < 1) {
        visible = 1;
    }
    if (open) {
	DoMethod(lv, MUIM_List_Jump, first + visible - 1);
    }
    set(lv, MUIA_List_Quiet, FALSE);
    if (open && active >= 0) {
	set(lv, MUIA_List_Active, active);
    }
}

void MUIListSetList(GuiObject lv, struct List *list, struct MemPool *memPool)
{
    MUIListSetListWithFilter(lv, list, memPool, NULL);
}

void MUIListGetList(GuiObject lv, struct List *list)
{
    struct Node *node = NULL, *pred = NULL;
    long i;

    if(lv == NULL || list == NULL)
        return;

    MUIListLoop(lv, i, node, TRUE)
    {
        REMOVE(node);
        INSERT(list, node, pred);
        pred = node;
    }
}

long GuiListActive(GuiObject go)
{
    long num = nget(go, MUIA_List_Active);

    if (num == MUIV_List_Active_Off)
	num = -1;
    return num;
}

void *GuiListActiveEntry(GuiObject go)
{
    void *entry = NULL;

    if (go)
	DoMethod(go, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, &entry);
    
    return entry;
}

void GuiListSetActive(GuiObject go, long num)
{
    if (go) {
	if (num < 0 || num >= nget(go, MUIA_List_Entries)) {
	    num = MUIV_List_Active_Off;
        }
	set(go, MUIA_List_Active, num);
    }
}

long MUIListFindEntry(GuiObject lv, void *entry)
{
    if (lv) {
	long i;
	void *elem = NULL;

	MUIListLoop(lv, i, elem, TRUE) {
	    if (elem == entry) {
		return i;
	    }
	}
    }
    return -1;
}

long GuiListFindNode(GuiObject lv, void *node)
{
    if (lv && node)
    {
	long i = NodeNum(node) - 1;
	void *elem = NULL;

       	/* quick test */
	DoMethod(lv, MUIM_List_GetEntry, i, &elem);
	if(node == elem)
	    return i;
    }
    return MUIListFindEntry(lv, node);
}

void *GuiListNthEntry(GuiObject lv, long num)
{
    void *elem = NULL;

    if (lv != NULL && num >= 0) {
	DoMethod(lv, MUIM_List_GetEntry, num, &elem);
    }

    return elem;
}

void MUIListSetActiveEntry(GuiObject lv, void *entry)
{
    long pos = MUIListFindEntry(lv, entry);

    if(pos >= 0)
        set(lv, MUIA_List_Active, pos);
}

void MUIListSetActiveNode(GuiObject lv, void *node)
{
    long pos = GuiListFindNode(lv, node);

    if(pos >= 0)
        set(lv, MUIA_List_Active, pos);
}

void MUIListRedrawNode(GuiObject lv, void *node)
{
    long pos = GuiListFindNode(lv, node);

    if(pos >= 0)
        DoMethod(lv, MUIM_List_Redraw, pos);
}

void MUIListInsertNode(GuiObject lv, void *node)
{
    if(lv && node)
    {
        struct Node *succ;
        long pos = -1;

        for(succ = SUCC(node); SUCC(succ); succ = SUCC(succ))
        {
            pos = GuiListFindNode(lv, succ);
            if(pos >= 0)
                break;
        }
        if(pos < 0)
            pos = MUIV_List_Insert_Bottom;
        DoMethod(lv, MUIM_List_InsertSingle, node, pos);
        pos = GuiListFindNode(lv, node);
        set(lv, MUIA_List_Active, pos);
    }
}

void MUIListRemoveEntry(GuiObject lv, void *node)
{
    long pos = MUIListFindEntry(lv, node);

    if (pos >= 0) {
        DoMethod(lv, MUIM_List_Remove, pos);
    }
}

const struct QueryLine *GuiSFWListQueryLine(GuiObject go, long row)
{
    const struct QueryLine *ql;

    if (row == 0) {
        ql = pget(go, MUIA_SFWList_Titles);
    } else {
        ql = GuiListNthEntry(go, row-1);
    }

    return ql;
}

long MUIRawKeyConvert(struct IntuiMessage *imsg, char *buffer, long size)
{
    if (imsg && buffer)
    {
	if (imsg->Code >= 0x4c && imsg->Code <= 0x4f && size >= 2)
	{
	    /* special treatment for cursor keys, AOS4 seems to eat them */
	    buffer[0] = (char)0x9b;
	    buffer[1] = (char)(imsg->Code - 11);

	    return 2;
	}
	else
	{
	    struct InputEvent ie;

    	    memset(&ie, 0, sizeof(struct InputEvent));
	    ie.ie_Class = IECLASS_RAWKEY;
	    ie.ie_SubClass = 0;
	    ie.ie_Code = imsg->Code;
	    ie.ie_Qualifier = imsg->Qualifier;
	    /* recover dead key codes & qualifiers */
	    ie.ie_EventAddress = *((APTR *)imsg->IAddress);

	    return MapRawKey(&ie, (STRPTR)buffer, size, NULL);
	}
    }
    return 0;
}

bool MUIIsSearchStartString(const char *buf, short len, ushort qualifier)
{
    char c;

    if (qualifier & (IEQUALIFIER_LCOMMAND | IEQUALIFIER_RCOMMAND | 
                IEQUALIFIER_CONTROL)) {
        return FALSE;
    } else if (len <= 0) {
        return FALSE;
    }

    c = buf[0];
    return (CodesetIsAlNum((cschar)c) || CodesetIsPunct((cschar)c));
}

#if 0
bool GuiWindowActivated(GuiObject go)
{
    ulong activated = FALSE;

    if(go)
        get(go, MUIA_Window_Activate, &activated);
    return((bool)activated);
}
#endif

bool GuiGetWindowDimensions(GuiObject go, struct WindowDimensions *wd)
{
    if (go && wd) {
        wd->wd_LeftEdge = nget(go, MUIA_Window_LeftEdge);
        wd->wd_TopEdge = nget(go, MUIA_Window_TopEdge);
        wd->wd_Width = nget(go, MUIA_Window_Width);
        wd->wd_Height = nget(go, MUIA_Window_Height);

        if (wd->wd_Width > 0 && wd->wd_Height > 0) {
            return TRUE;
        }
    }
    return FALSE;
}

struct Window *GuiAnyOpenedWindow(void)
{
    Object *app = GuiApplication, *win;
    struct Window *w = NULL;
    struct List *list = pget(app, MUIA_Application_WindowList);
    struct Node *node = NULL;

    if(list == NULL)
        return(NULL);

    node = HEAD(list);
    while((win = NextObject(&node)) != NULL)
    {
        w = pget(win, MUIA_Window_Window);
        if(w != NULL)
            return(w);
    }
    return(NULL);
}

struct Project *GuiActiveProject(void)
{
    return GuiWindowProject(GuiActiveWindow());
}

GuiObject GuiActiveWindow(void)
{
    Object *app = GuiApplication, *win;
    struct List *list = pget(app, MUIA_Application_WindowList);
    struct Node *node = NULL;

    if(list == NULL)
        return(NULL);

    node = HEAD(list);
    while((win = NextObject(&node)) != NULL)
        if(nget(win, MUIA_Window_Activate))
            return(win);

    return(NULL);
}

GuiObject GuiActiveObject(GuiObject wi)
{
    GuiObject active = pget(wi, MUIA_Window_ActiveObject);

    return(active);
}

bool GuiIsEditorObject(GuiObject go)
{
    return MUIObjectIsA(go, ExtEditorClass);
}

GuiObject MUINewEditor(struct Project *pr,
        char shortcut, bool readonly, bool fixedFont,
        const char *bg, struct TextFont *tf, GuiObject *editorobject)
{
    GuiObject obj, editor;
#ifdef USE_TEXTEDITOR
    GuiObject sbar;

    sbar = ScrollbarObject, End;
    obj = HGroup,
	GroupSpacing(0),
	Child, editor = ExtEditorObject,
	    InputListFrame,
 	    MUIA_ControlChar, shortcut,
            MUIA_CycleChain, TRUE,
            tf? MUIA_Font : TAG_IGNORE, tf,
            MUIA_Background, bg,
            MUIA_TextEditor_Slider, sbar,
            MUIA_TextEditor_FixedFont, fixedFont,
	    MUIA_TextEditor_ReadOnly, readonly,
	    MUIA_ExtEditor_Contents, "",
            MUIA_ExtEditor_LoadFunc, MenuEditorLoadText,
            MUIA_ExtEditor_SaveFunc, MenuEditorSaveText,
            MUIA_ExtEditor_ExternalEditorFunc, MenuEditorExternalEditor,
            MUIA_ExtEditor_FuncData, pr,
	    End,
	Child, sbar,
	End;

#elif defined(USE_TEXTINPUT)
    obj = editor = ExtEditorObject,
        InputListFrame,
        MUIA_ControlChar, shortcut,
        MUIA_CycleChain, TRUE,
        tf? MUIA_Font : TAG_IGNORE, tf,
        MUIA_Background, bg,
        MUIA_Textinput_Multiline, TRUE,
        MUIA_Textinput_RemainActive, TRUE,
        MUIA_Textinputscroll_VertScrollerOnly, TRUE,
        MUIA_Textinput_Font, fixedFont? 
            MUIV_Textinput_Font_Fixed : MUIV_Textinput_Font_Normal,
        fixedFont? MUIA_Textinput_WordWrap : TAG_IGNORE, FALSE,
        MUIA_Textinput_Editable, !readonly,
        MUIA_ExtEditor_Contents, "",
        End;
#else
#error "No text editor class defined!"
#endif
    if (editorobject) {
	*editorobject = editor;
    }
    return obj;
}

char *GuiEditorText(GuiObject edit, struct MemPool *memPool)
{
    return(ExtEditorGetContents(edit, memPool));
}

void GuiSetEditorText(GuiObject edit, const char *str)
{
    if(edit != NULL)
    {
        set(edit, MUIA_ExtEditor_Contents, str);
        set(edit, MUIA_ExtEditor_HasChanged, TRUE);
    }
}

void GuiSetEditorChanged(GuiObject edit, bool changed)
{
    if (edit != NULL) 
    {
        set(edit, MUIA_ExtEditor_HasChanged, changed);
    }
}

void HOOK_PROTO(PopButtonSetCWD, Object *obj, struct Project **x)
{
    ProjectSetCurrentDir(*x);
}

GuiObject GuiPopButtonSetCWD(ulong type, struct Project *pr)
{
    GuiObject button = PopButton(type);

    if (button) {
        MUIAddCallback(button, MUIA_Pressed, FALSE, PopButtonSetCWD, pr);
    }
    return button;
}

void GuiStoreEdits(void)
{
    GuiFieldStoreEdits();
}

void GuiSetCursor(void)
{
    GuiObject wi = GuiActiveWindow();
    GuiObject go = GuiActiveObject(wi);

    if(wi && go)
        SetAttrs(wi, MUIA_NoNotify, TRUE,
            MUIA_Window_ActiveObject, go,
            TAG_DONE);
}

static long guiMainLoopEntered = 0;

void GuiMainLoop(bool (*reallyQuit)(void))
{
    guiMainLoopEntered++;
    for (;;) {
        if (GuiMainLoopIteration()) {
            if (nget(GuiApplication, MUIA_Application_ForceQuit)) {
                break;
            }
            if (reallyQuit == NULL || reallyQuit()) {
                break;
            }
        }
    }
    guiMainLoopEntered--;
}

#define UPDATE_CHECK_INTERVAL CLOCKS_PER_SEC

static ulong GuiMainLoopIterationNextId(void)
{
    static long level = 0;
    static clock_t nextUpdateCheck = 0;
    ulong sigs = 0, id = MUIV_Application_ReturnID_Quit;

    level++;
    if (GuiApplication) {
        id = DoMethod(GuiApplication, MUIM_Application_NewInput, &sigs);
        if (id == 0) {
            if (sigs) {
                UpdateMenu(GuiActiveWindow(), FALSE);
                UpdateProjects(FALSE);
                if (level == 1 && guiMainLoopEntered == 1) {
                    ServerProcessRequests();
                    if (clock() >= nextUpdateCheck) {
                        UpdateProjects(TRUE);
                        nextUpdateCheck = clock() + UPDATE_CHECK_INTERVAL;
                    }
                }
                sigs = Wait(sigs | SIGBREAKF_CTRL_C);
                if (sigs & SIGBREAKF_CTRL_C) {
                    fprintf(stderr, "Ctrl-C\n");
                    id = MUIV_Application_ReturnID_Quit;
                }
            }
        }
    }
    level--;

    return id;
}

bool GuiMainLoopIteration(void)
{
    ulong id = GuiMainLoopIterationNextId();

    return (bool)(id == MUIV_Application_ReturnID_Quit);
}

bool WaitOk(void)
{
    ulong id;

    for (;;)
    {
        id = GuiMainLoopIterationNextId();
        switch (id)
	{
	    case ID_OK:
                return TRUE;

  	    case ID_CLOSE:
	    case ID_CANCEL:
	    case MUIV_Application_ReturnID_Quit:
                return FALSE;
	}
    }
    return FALSE;
}

#if 0           /* no use of this yet */
void GuiSavePrefs()
{
    GuiObject app = GuiApplication;

    if(app)
    {
        DoMethod(app, MUIM_Application_Save, MUIV_Application_Save_ENV);
        DoMethod(app, MUIM_Application_Save, MUIV_Application_Save_ENVARC);
    }
}

void GuiLoadPrefs()
{
    GuiObject app = GuiApplication;

    if(app)
        DoMethod(app, MUIM_Application_Load, MUIV_Application_Load_ENV);
}
#endif

