/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** mui.h
**
** Creation date: 07.09.1994
** Last modified: 07.05.2022
*/

#ifndef MUI_H
#define MUI_H

#include <libraries/mui.h>
#include <proto/intuition.h>
#include <proto/muimaster.h>
#include "extclasses/ExtApplication_mcc.h"
#include "extclasses/ExtWindow_mcc.h"
#include "extclasses/SFWList_mcc.h"

struct MemPool;
struct WindowDimensions;

/* general stuff */

typedef Object *GuiObject;

extern GuiObject GuiApplication;
extern const char MUIBoldString[];
extern const char MUIItalicString[];
extern char GuiScreenTitle[];
extern char DefaultRealFormat, DefaultDateFormat;
extern struct TextFont *Topaz8Font;
extern struct Locale *Locale;

enum
{
/* for local event loops */
    ID_CLOSE=101,
    ID_CANCEL,
    ID_OK,
    ID_PROCESS_EXIT
};

void CloseGui(void);
const char *OpenGui(int *pargc, char **argv);
/* returns error string or NULL on success */

void GuiPrefsUpdated(void);

const char *GuiLanguage(void);        /* returns system's default language */
const char *GuiPubScreenName(void);   /* returns name of pub screen of BeeBase */

GuiObject AllocMenuStrip(GuiObject go);
void MenuStripAddNotifications(void);

void GuiAbout(void);
void AboutMUI(void);

/* void GuiFlush(void); */
void GuiSendQuitMessage(void);

ulong GuiSleep(void);
void GuiSetSleep(bool sleep);

#define GuiProjectSleep(pr) nget(ProjectRootWindow(pr), MUIA_Window_Sleep)
#define GuiSetProjectSleep(pr, sleep) \
        DoMethod(GuiApplication, MUIM_ExtApplication_ProjectSleep, \
                 pr, (ulong)sleep)     

char *GuiEditorText(GuiObject go, struct MemPool *memPool);
void GuiSetEditorText(GuiObject go, const char *str);
void GuiSetEditorChanged(GuiObject edit, bool changed);

#define GuiSetRegisterPage(go, pos) \
   if(go != NULL && nget(go, MUIA_Group_ActivePage) != pos) \
       set(go, MUIA_Group_ActivePage, pos)

#define GuiRegisterActivePage(go) (nget(go, MUIA_Group_ActivePage))

#define GuiWindowProject(go) (pget(go, MUIA_ExtWindow_Project))

bool GuiOpenWindow(GuiObject wi);
#define GuiCloseWindow(wi) if(wi != NULL) set(wi, MUIA_Window_Open, FALSE)

short GuiObjectWeight(GuiObject obj, bool horiz);

// bool GuiWindowActivated(GuiObject go);
bool GuiGetWindowDimensions(GuiObject go, struct WindowDimensions *wd);
struct Window *GuiAnyOpenedWindow(void);

struct Project *GuiActiveProject(void);
GuiObject GuiActiveWindow(void);
GuiObject GuiActiveObject(GuiObject wi);
bool GuiIsEditorObject(GuiObject go);

#define GuiUserData(go)  (nget(go, MUIA_UserData))
#define GuiObjectDisplayObject(go)  ((struct DisplayObject *)GuiUserData(go))

#define GuiSetSelected(go, selected) if(go) nnset(go, MUIA_Selected, selected)

void GuiSetEnabled(GuiObject go, bool enabled);
bool GuiGetEnabled(GuiObject go);

#define GuiSetBackground(go, bg) if (go) { set(go, MUIA_Background, bg); }

GuiObject GuiObjectWindow(GuiObject go);
struct Project *GuiObjectProject(GuiObject go);

#define GuiStringContents(go) pget(go, MUIA_String_Contents)
#define GuiObjectText(go) pget(go, MUIA_String_Contents)
#define GuiWindowOpen(wi) nget(wi, MUIA_Window_Open)

#define GuiActivateWindow(wi) if(wi) set(wi, MUIA_Window_Activate, TRUE)
#define GuiWindowToFront(wi) DoMethod(wi, MUIM_Window_ToFront)
#define GuiSetActiveGuiObject(wi, go) set(wi, MUIA_Window_ActiveObject, go)

#define GuiSetWindowTitle(wi, title) set(wi, MUIA_Window_Title, title)
#define GuiSetPrependWindowTitle(wi, prepend) \
    set(wi, MUIA_ExtWindow_PrependTitle, prepend)

long GuiListActive(GuiObject go);
void *GuiListActiveEntry(GuiObject go);
void GuiListSetActive(GuiObject go, long num);
long GuiListFindNode(GuiObject go, void *node);
void *GuiListNthEntry(GuiObject go, long num);

#define GuiListNumEntries(go) nget(go, MUIA_List_Entries)
#define GuiSFWListActive(go) nget(go, MUIA_SFWList_ActiveLine)
#define GuiSFWListSetActive(go, num) set(go, MUIA_SFWList_ActiveLine, num)
#define GuiSFWListNumColumns(go) nget(go, MUIA_SFWList_NumFields)
#define GuiSFWListFieldInfos(go) pget(go, MUIA_SFWList_FieldInfos)
#define GuiSFWListColumnOrder(go) pget(go, MUIA_SFWList_ColumnOrder)
#define GuiSFWListControlTable(go) pget(go, MUIA_SFWList_ControlTable)
#define GuiSFWListActiveRecord(go) pget(go, MUIA_SFWList_ActiveRecord)
#define GuiSFWListSetActiveRecord(go, r) set(go, MUIA_SFWList_ActiveRecord, r)
#define GuiSFWListRecord(go, row) \
    ((go)? (struct Record *)DoMethod(go, MUIM_SFWList_GetRecord, row) : NULL)

const struct QueryLine *GuiSFWListQueryLine(GuiObject go, long row);

void GuiStoreEdits(void);
void GuiSetCursor(void);

void GuiMainLoop(bool (*reallyQuit)(void));
bool GuiMainLoopIteration(void);
bool WaitOk(void);

void GuiShowHelp(const char *section);

#define GuiUpdateExtraButtonsInTabChain()

/* intuition stuff */

#if defined(__amigaos4__)
    APTR VARARGS68K DoSuperNew(struct IClass *cl,Object *obj,...);
    APTR DoSuperNewA(Class *cl, APTR obj, struct TagItem *tagList);
#elif (defined __GNUC__) && (!defined __MORPHOS__)
    APTR DoSuperNewA(Class *cl, APTR obj, struct TagItem *tagList);
    #define DoSuperNew(cl, obj, tag1, ...) \
        ({IPTR _tags[] = { tag1, __VA_ARGS__ }; \
        DoSuperNewA((cl), (obj), (struct TagItem *) _tags); })
    #if (!defined __AROS__)
        #define SetSuperAttrs( cl, obj, tag1, ...) \
            ({ IPTR _tags[] = { tag1, __VA_ARGS__ }; \
            struct opSet MyopSet; \
            MyopSet.MethodID = OM_SET; \
            MyopSet.ops_AttrList = (struct TagItem*) _tags; \
            MyopSet.ops_GInfo = NULL; \
            DoSuperMethodA((cl) , (obj), (APTR) &MyopSet); \
            })
    #endif
#endif

#define GuiBeep() DisplayBeep(NULL)

#ifndef BOOPSIOBJMACRO_START
    #define BOOPSIOBJMACRO_START(class) NewObject(class, NULL
#endif

/* mui stuff */

#ifdef USE_HOOKENTRY 		/* using HookEntry and arguments on stack */

    #define HOOK_PROTO(name, obj, msg)					\
    SAVEDS STDARGS name ## Func(struct Hook *hook, obj, msg)		\

    #define HOOK_STRUCT(name)						\
    struct Hook name ## Hook =						\
    { { NULL, NULL }, (HOOKFUNC)HookEntry, (HOOKFUNC)name ## Func, NULL };	\

    #define HOOK_INIT(hook, name)						\
        (hook)->h_Entry = (HOOKFUNC)HookEntry;				\
        (hook)->h_SubEntry = (HOOKFUNC)name ## Func;			\

    #define DISPATCHER_PROTO						\
    static IPTR SAVEDS STDARGS Dispatcher(struct IClass *cl, 		\
    	Object *obj, Msg msg)						\

    #define DISPATCHER_INIT(cl)						\
    	(cl)->cl_Dispatcher.h_Entry = (HOOKFUNC)HookEntry;		\
    	(cl)->cl_Dispatcher.h_SubEntry = (HOOKFUNC)Dispatcher;		\

#else 				/* directly using arguments in registers */

    #define HOOK_PROTO(name, obj, msg)					\
    ASM SAVEDS name ## Func(REG(a0, struct Hook *hook), 			\
    	REG(a2, obj), REG(a1, msg))					\

    #define HOOK_STRUCT(name)						\
    struct Hook name ## Hook = 						\
    { { NULL, NULL }, (HOOKFUNC)name ## Func, NULL, NULL };			\

    #define HOOK_INIT(hook, name)						\
        (hook)->h_Entry = (HOOKFUNC)name ## Func;				\
        (hook)->h_SubEntry = (HOOKFUNC)NULL;				\

    #define DISPATCHER_PROTO						\
    static IPTR ASM SAVEDS Dispatcher(REG(a0, struct IClass *cl), 		\
    	REG(a2, Object *obj), REG(a1, Msg msg))				\

    #define DISPATCHER_INIT(cl)						\
    	(cl)->cl_Dispatcher.h_Entry = (HOOKFUNC)Dispatcher;	   \
    	(cl)->cl_Dispatcher.h_SubEntry = NULL;		\

#endif

#define MUIAddCallback(obj, attr, val, name, data) 			\
	{ 								\
            static HOOK_STRUCT(name) 					\
            DoMethod(obj, MUIM_Notify, attr, val, obj, 			\
		    3, MUIM_CallHook, &name ## Hook, data); 		\
        }								\

#define MUIAddCallback2(obj1, attr, val, obj2, name, data) 		\
	{ 								\
            static HOOK_STRUCT(name)					\
            DoMethod(obj1, MUIM_Notify, attr, val, obj2, 		\
		    3, MUIM_CallHook, &name ## Hook, data); 		\
        }								\

#define MUIAddPushCallback(obj, attr, val, name, data) 			\
	{ 								\
            static HOOK_STRUCT(name)					\
            DoMethod(obj, MUIM_Notify, attr, val, MUIV_Notify_Application, \
		    6, MUIM_Application_PushMethod, obj, 		\
		    3, MUIM_CallHook, &name ## Hook, data); 		\
        }								\

#define MUIAddPushCallback2(obj1, attr, val, obj2, name, data) 		\
	{ 								\
            static HOOK_STRUCT(name)					\
            DoMethod(obj1, MUIM_Notify, attr, val, MUIV_Notify_Application, \
		    6, MUIM_Application_PushMethod, obj2, 		\
		    3, MUIM_CallHook, &name ## Hook, data); 		\
        }								\

IPTR nget(GuiObject go, LONG attr);
#define pget(go, attr) ((void *)(nget(go, attr)))

void MUIAddAppMessageNotification(GuiObject source, GuiObject target);

void HOOK_PROTO(GuiObjectViewFileCallback, GuiObject go, GuiObject *str);
void HOOK_PROTO(GuiObjectEditFileCallback, GuiObject go, GuiObject *str);

void HOOK_PROTO(PopButtonSetCWD, Object *obj, struct Project **x);
GuiObject GuiPopButtonSetCWD(ulong type, struct Project *pr);

bool MUIFindObject(GuiObject root, GuiObject find);

char **MUIListLabels(struct MemPool *memPool, GuiObject list);
void MUIListSetLabels(GuiObject list, char **labels);

void MUIListSetList(GuiObject lv, struct List *list, struct MemPool *memPool);
void MUIListSetListWithFilter(GuiObject lv, struct List *list, 
        struct MemPool *memPool, bool (*filter)(const struct Node *node));
void MUIListGetList(GuiObject lv, struct List *list);
long MUIListFindEntry(GuiObject lv, void *entry);
void MUIListSetActiveNode(GuiObject lv, void *node);
void MUIListSetActiveEntry(GuiObject lv, void *entry);
void MUIListRedrawNode(GuiObject lv, void *node);
void MUIListInsertNode(GuiObject lv, void *node);
void MUIListRemoveEntry(GuiObject lv, void *entry);

GuiObject MUINewEditor(struct Project *pr,
        char shortcut,	bool readonly, bool fixedFont,
        const char *bg, struct TextFont *tf, GuiObject *editorobject);
void MUISetEditorCursor(GuiObject editor, long line, long col);

long MUIRawKeyConvert(struct IntuiMessage *imsg, char *buffer, long size);
/* like RawKeyConvert but for IntuiMessage and with special fix for AOS4 */

bool MUIIsSearchStartString(const char *buf, short len, ushort qualifier);
/* Checks whether the input buffer and qualifier could be a search string */

#define MUIAddMember(parent, child) DoMethod(parent, OM_ADDMEMBER, child)
#define MUIRemMember(parent, child) DoMethod(parent, OM_REMMEMBER, child)

#define MUI_AddDisposeObject(app, obj) \
	DoMethod(app, MUIM_ExtApplication_DisposeObject, obj);

/* some nice ugly macros from Steve Koren */

# define MUIListLoop(l, pos, obj, c) \
for ((pos)=0; DoMethod((l), MUIM_List_GetEntry,(pos),&(obj)), \
(obj)&&(c); (pos)++)

# define MUIListLoopSel(l, pos, obj, c) \
for ((pos)=MUIV_List_NextSelected_Start; \
(DoMethod((l), MUIM_List_NextSelected, &(pos)), \
(pos) != MUIV_List_NextSelected_End) && \
(DoMethod((l), MUIM_List_GetEntry, (pos), &(obj)), (c));)

/* extra MUI stuff from Stefan Stuntz */

#ifndef PopimageObject
#define PopimageObject MUI_NewObject(MUIC_Popimage
#endif

#ifndef MUIA_Imagedisplay_Spec
#define MUIA_Imagedisplay_Spec 0x8042a547 /* V11 isg struct MUI_ImageSpec * */
#endif

#ifndef MUIA_Imageadjust_Type
#define MUIA_Imageadjust_Type 0x80422f2b /* V11 i.. LONG */
#endif
#define MUIV_Imageadjust_Type_All 0
#define MUIV_Imageadjust_Type_Image 1
#define MUIV_Imageadjust_Type_Background 2
#define MUIV_Imageadjust_Type_Pen 3

struct MUI_ImageSpec
{
    char buf[64];
};

#ifndef MUIA_Text_HiCharIdx
#define MUIA_Text_HiCharIdx 0x804214f5
#endif

#ifndef MUIM_Window_Setup
#define MUIM_Window_Setup         0x8042c34c /* Custom Class */ /* V18 */
#define MUIM_Window_Cleanup       0x8042ab26 /* Custom Class */ /* V18 */
#define MUIM_Window_ActionIconify 0x80422cc0 /* private */ /* V18 */
#endif

#ifndef MUIA_Group_Forward
#define MUIA_Group_Forward        0x80421422 /* MUI: V11 .s. BOOL */
#endif

#ifndef MUIA_Application_UsedClasses
#define MUIA_Application_UsedClasses 0x8042e9a7 /* STRPTR * */ /* V20 */
#endif

/* Flags for MorphOS/MUI's Font Selector from Antoine Dubourg */
#ifndef FOF_DOFRONTCOLOR
#define FOF_DOFRONTCOLOR (1<<24)
#define FOF_DOBACKCOLOR  (1<<25)
#define FOF_DOMUISTYLES  (1<<26)
#endif

#ifndef MUIA_MinWidth
#define MUIA_MinWidth 0x80424342
#endif
#ifndef MUIA_MinHeight
#define MUIA_MinHeight 0x8042c04a
#endif

#ifndef MUIA_List_ClickColumn
#define MUIA_List_ClickColumn 0x8042d1b3
#endif

/* for MUI custom classes */

/*
** The following func is bool (*)(Object *, char *key, short len, ushort qualifier, void *keyData)
*/
#define MUIA_KeyFunc     (TAG_USER | (140 << 16) | 0xA)              /* [I..] */
#define MUIA_KeyData     (TAG_USER | (140 << 16) | 0xB)              /* [IS.] */

/*
** This one is bool (*)(Object *, short x, short y, bool down)
*/
#define MUIA_MouseFunc  (TAG_USER | (140 << 16) | 0xC)              /* [I..] */

/* Selection of custom classes */

#ifdef __MORPHOS__
#define USE_TEXTINPUT
#else
#define USE_NLIST
#define USE_BETTERSTRING
#define USE_TEXTEDITOR
#endif

#if !defined(__amigaos4__) && !defined(__AROS__)
#define USE_CALENDAR
#endif

/* NList.mcc vs List.mui */

#ifdef USE_NLIST
#include <mui/NList_mcc.h>
#include <mui/NListview_mcc.h>

#define MListObject NListObject
#define MUIM_MList_CreateImage MUIM_NList_CreateImage
#define MUIM_MList_DeleteImage MUIM_NList_DeleteImage

#define MListviewObject NListviewObject
#define MUIA_MListview_MList MUIA_NListview_NList
#else

#ifndef MUIA_List_TitleClick
#define MUIA_List_TitleClick 0x80422fd9
#endif

#ifndef MUIM_List_Compare
#define MUIM_List_Compare 0x80421b68

struct  MUIP_List_Compare { ULONG MethodID; APTR entry1; APTR entry2; };
#endif

#ifndef MUIA_List_ColumnOrder
#define MUIA_List_ColumnOrder 0x9d5100f6
#endif

#ifndef MUIM_List_QueryColumnWidth
#define MUIM_List_QueryColumnWidth 0x8042e09e
#endif

#ifndef MUIF_LIST_QUERYCOLUMNWIDTH_GETMWIDTH
#define MUIF_LIST_QUERYCOLUMNWIDTH_GETMWIDTH (1L << 31)
#endif

#ifndef MUIA_List_AutoLineHeight
#define MUIA_List_AutoLineHeight            0x8042bc08 /* V20 i.. BOOL       */
#endif

#define MListObject ListObject
#define MUIM_MList_CreateImage MUIM_List_CreateImage
#define MUIM_MList_DeleteImage MUIM_List_DeleteImage

#define MListviewObject ListviewObject
#define MUIA_MListview_MList MUIA_Listview_List
#endif

/* BetterString.mcc vs TextInput.mcc vs String.mui */

#ifdef USE_BETTERSTRING
#ifdef StringObject
#undef StringObject
#define MUIC_BetterString "BetterString.mcc"
#define StringObject MUI_NewObject(MUIC_BetterString
#endif

/* 
** The following define should be in a header file of BetterString 
** but since this define is relatively new, most people won't have it yet.
*/
#ifndef MUIA_BetterString_SelectOnActive
#define MUIA_BetterString_SelectOnActive        0xad00100d
#endif

#define MUIStringPreselectAll(str, select) \
        set(str, MUIA_BetterString_SelectOnActive, select)

#elif defined(USE_TEXTINPUT)

#define MUIStringPreselectAll(str, select) \
        if ((str) && (select)) { \
            DoMethod(str, MUIM_Textinput_DoMarkAll); \
            set(str, MUIA_Textinput_CursorPos, 0); \
            set(str, MUIA_Textinput_ResetMarkOnCursor, TRUE); \
        }

#else /* Standard String object */

#define MUIStringPreselectAll(str, select)

#endif

#ifdef USE_CALENDAR
#include <mui/Calendar_mcc.h>
#endif

#endif

