/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** Order_mui.c
**
** Creation date: 30.04.1995
** Last modified: 26.09.2014
*/

#include "system.h"
#include "order.h"
#include "gui.h"
#include "project.h"
#include "table.h"
#include "field.h"
#include "memory.h"
#include "requesters.h"
#include "error.h"
#include "util.h"
#include "appstrings_cat.h"
#include "structureeditor_mui.h"
#include "extclasses/ExtApplication_mcc.h"
#include "extclasses/DDList_mcc.h"

struct EditOrderData
{
  struct Project *d_Project;
  struct Table *d_Table;
  char d_Title[100];
  Object *d_Type;
  Object *d_TypeGroup;
  Object *d_CmpFuncStr;
  Object *d_CmpFunc;
  Object *d_Fields;
  Object *d_OrderList;
  Object *d_Ok;
  Object *d_Clear;
  Object *d_Cancel;
};

static void disposeEditOrder(Object *app, Object *obj)
{
    struct EditOrderData *d = pget(obj, MUIA_UserData);

    if(obj && d)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, 
		    sizeof(struct EditOrderData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void HOOK_PROTO(finish, Object *obj, ulong *ok)
{
    struct EditOrderData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    struct Table *tab = d->d_Table;
    Object *app = pget(obj, MUIA_ApplicationObject);
    long error = OK;

    if (*ok) {
        struct MemPool *memPool = ProjectMemPool(d->d_Project); 
        short num = 0, i;
        struct OrderField *oas = NULL, *oa;
        struct OrderData od;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        od.od_Type = nget(d->d_Type, MUIA_Cycle_Active)?
            ORD_TYPE_CMPFUNC : ORD_TYPE_FIELDLIST;
        if (od.od_Type == ORD_TYPE_CMPFUNC) {
            od.od_Contents.cmpfunc = pget(d->d_CmpFuncStr,
                    MUIA_String_Contents);
        } else {
            num = nget(d->d_OrderList, MUIA_List_Entries);
            if (num > 0) {
                oas = PoolAllocMem(memPool,
                        num * sizeof(struct OrderField));
                if (oas == NULL) {
                    error = ERR_MEM;
                } else {
                    for (i=0; i < num; i++) {
                        DoMethod(d->d_OrderList, MUIM_List_GetEntry, i, &oa);
                        oas[i] = *oa;
                    }
                }
            }

            od.od_Contents.fieldlist.num = num;
            od.od_Contents.fieldlist.oas = oas;
        }

        if (error == OK) {
            GuiSetSleep(TRUE);
            error = SetNewOrder(tab, &od);
            GuiSetSleep(FALSE);
        }
        PoolFreeMem(memPool, oas, num * sizeof(struct OrderField));

        if (error) {
            Ask2(obj, GCS(MSG_ErrorMessage), AT_ERROR, 
                    GCS(MSG_StarOk), GES(error));
            return;
        }
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
    disposeEditOrder(app, obj);

    if (*ok) {
        error = ReorderAllRecords(tab);
        if (error) {
            Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Err_Order), GES(error));
        }
        GuiDisplayPanel(tab);
        ProjectChange(pr, FALSE);
    }
}
     
static void insertfield(GuiObject lva, GuiObject lvo, long pos)
{
    struct OrderField *oa = NULL, new;
    struct Field *field = NULL;
    long i;

    if(lva == NULL || lvo == NULL)
        return;

    /* get source entry */
    DoMethod(lva, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, &field);
    if(field == NULL)
        return;

    /* check if field is already in order list */
    for(i = 0; ; i++)
    {
        DoMethod(lvo, MUIM_List_GetEntry, i, &oa);
        if(oa == NULL || oa->oa_Field == field)
            break;
    }

    new.oa_Ascending = TRUE;
    new.oa_Field = field;
    DoMethod(lvo, MUIM_List_InsertSingle, &new, pos);
    get(lvo, MUIA_List_InsertPosition, &pos);
    set(lvo, MUIA_List_Active, pos);
    if(oa)
        MUIListRemoveEntry(lvo, oa);
}

static void HOOK_PROTO(insertClick, GuiObject lva, GuiObject *x)
{
    insertfield(lva, *x, MUIV_List_Insert_Bottom);
}

static IPTR dragQuery(GuiObject lv, struct MUIP_DragDrop *msg, void *dragData)
{
    return (IPTR)(msg->obj == dragData?
            MUIV_DragQuery_Accept : MUIV_DragQuery_Refuse);
}

IPTR dragDropToOrder(GuiObject lvo, struct MUIP_DragDrop *msg, void *dragData)
{
    GuiObject lva = msg->obj;
    long dropmark = 0;

    get(lvo, MUIA_List_DropMark, &dropmark);
    insertfield(lva, lvo, dropmark);
    set(lva, MUIA_List_Active, MUIV_List_Active_Off);

    return 0;
}

IPTR dragDropToField(GuiObject lva, struct MUIP_DragDrop *msg, void *dragData)
{
    DoMethod(msg->obj, MUIM_List_Remove, MUIV_List_Remove_Active);
    return 0;
}

static void HOOK_PROTO(orderClick, GuiObject lvo, ulong dummy)
{
    struct OrderField *oa = NULL;

    DoMethod(lvo, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, &oa);
    if(oa != NULL)
    {
        oa->oa_Ascending = !oa->oa_Ascending;
      	DoMethod(lvo, MUIM_List_Redraw, MUIV_List_Redraw_Active);
    }
}

static long HOOK_PROTO(dispOrder, const char **array, struct OrderField *oa)
{
    static char arrow[31];

    sprintf(arrow, "\33I[6:%d]",
            (int)(oa->oa_Ascending? MUII_TapeUp : MUII_TapeDown));
    *array++ = arrow;
    *array++ = FieldName(oa->oa_Field);

    return(0);
}
static HOOK_STRUCT(dispOrder)

static long HOOK_PROTO(dispFields, const char **array, 
	struct Field *field)
{
    *array++ = FieldName(field);
    return(0);
}
static HOOK_STRUCT(dispFields)

typedef struct OrderField *OrderFieldPointer;
static OrderFieldPointer HOOK_PROTO(cons, struct MemPool *memPool, 
	struct OrderField *init)
{
    struct OrderField *oa;

    oa = AllocPooled(memPool, sizeof(struct OrderField));
    if(oa != NULL)
        *oa = *init;
    return oa;
}
static HOOK_STRUCT(cons)

static void HOOK_PROTO(des, struct MemPool *memPool, 
	struct OrderField *oa)
{
    FreePooled(memPool, oa, sizeof(struct OrderField));
}
static HOOK_STRUCT(des)

static Object *createEditOrder(struct Table *tab, struct OrderData *od)
{
  static const char *typeContents[3];
  struct Project *pr = TableProject(tab);
  struct MemPool *memPool = ProjectMemPool(pr);
  struct List *fields = TableFields(tab);
  struct Field *field;
  short type = ORD_TYPE_FIELDLIST, num = 0, i, page=0;
  struct OrderField *oas = NULL;
  char *cmpfunc = NULL;
  GuiObject app = GuiApplication;
  struct EditOrderData *d;
  GuiObject obj, root, fieldsgrp, order, buttons, lista, listo,
    top, fieldlistgrp, cmpfuncgrp;

  if(app == NULL)
    return(NULL);

  d = PoolAllocMemZero(memPool, sizeof(struct EditOrderData));
  if(d == NULL)
    return(NULL);

  d->d_Project = pr;
  d->d_Table = tab;

  if(od != NULL)
    {
      type = od->od_Type;
      if(type == ORD_TYPE_FIELDLIST)
	{
	  num = od->od_Contents.fieldlist.num;
	  oas = od->od_Contents.fieldlist.oas;
	}
      else if(type == ORD_TYPE_CMPFUNC)
	{
	  cmpfunc = od->od_Contents.cmpfunc;
	  page = 1;
	}
    }

  d->d_Fields = ListviewObject,
    MUIA_CycleChain, TRUE,
    MUIA_ShortHelp, GCS(MSG_Order_FieldsBH),
    MUIA_Listview_DoubleClick, TRUE,
    MUIA_Listview_DragType, MUIV_Listview_DragType_Immediate,
    MUIA_Listview_List, lista = DDListObject,
        MUIA_Frame, MUIV_Frame_InputList,
        MUIA_DDList_DragQuery, dragQuery,
        MUIA_DDList_DragDrop, dragDropToField,
        MUIA_List_ShowDropMarks, FALSE,
        MUIA_List_DisplayHook, &dispFieldsHook,
        End,
    End;

  fieldsgrp = VGroup,
    MUIA_Frame, MUIV_Frame_Group,
    MUIA_FrameTitle, GCS(MSG_Fields),
    MUIA_Background, MUII_GroupBack,
    Child, d->d_Fields,
    End;

  d->d_OrderList = ListviewObject,
    MUIA_CycleChain, TRUE,
    MUIA_ShortHelp, GCS(MSG_Order_OrderBH),
    MUIA_Listview_DoubleClick, TRUE,
    MUIA_Listview_DragType, MUIV_Listview_DragType_Immediate,
    MUIA_Listview_List, listo = DDListObject,
        MUIA_Frame, MUIV_Frame_InputList,
        MUIA_List_DragSortable, TRUE,
        MUIA_DDList_DragQuery, dragQuery,
        MUIA_DDList_DragDrop, dragDropToOrder,
        MUIA_List_Format, "DELTA=8,",
        MUIA_List_ConstructHook, &consHook,
        MUIA_List_DestructHook, &desHook,
        MUIA_List_DisplayHook, &dispOrderHook,
        End,
    End;

  order = GroupObject,
    MUIA_Frame, MUIV_Frame_Group,
    MUIA_FrameTitle, GCS(MSG_Order_Order),
    MUIA_Background, MUII_GroupBack,
    Child, d->d_OrderList,
    End;

  fieldlistgrp = HGroup,
    MUIA_Frame, MUIV_Frame_Group,
    MUIA_Background, MUII_GroupBack,
    MUIA_Group_SameHeight, TRUE,
    Child, fieldsgrp,
    Child, order,
    End;

  d->d_CmpFuncStr = StringObject,
    MUIA_CycleChain, TRUE,
    MUIA_Frame, MUIV_Frame_String,
    MUIA_ControlChar, Shortcut(GCS(MSG_Order_ComparisonFunction)),
    MUIA_ShortHelp, GCS(MSG_Order_ComparisonFunctionBH),
    MUIA_String_MaxLen, TABLE_TRIGGER_LEN + 1,
    MUIA_String_Format, MUIV_String_Format_Center,
    MUIA_String_Contents, cmpfunc,
    End;

  d->d_CmpFunc = UserFunctionsPopup(d->d_CmpFuncStr, pr);

  cmpfuncgrp = VGroup,
    MUIA_Frame, MUIV_Frame_Group,
    MUIA_Background, MUII_GroupBack,
    Child, VSpace(0),
    Child, CLabel(GCS(MSG_Order_ComparisonFunction)),
    Child, d->d_CmpFunc,
    Child, VSpace(0),
    End;
    
  typeContents[0] = GCS(MSG_Order_Type_FieldList);
  typeContents[1] = GCS(MSG_Order_Type_ComparisonFunction);
  typeContents[2] = NULL;
  
  d->d_Type = CycleObject,
    MUIA_CycleChain, TRUE,
    MUIA_ControlChar, Shortcut(GCS(MSG_Type)),
    MUIA_Cycle_Entries, typeContents,
    MUIA_Cycle_Active, page,
    End;

  top = HGroup,
    Child, Label(GCS(MSG_Type)),
    Child, d->d_Type,
    End;

  d->d_TypeGroup = GroupObject,
    MUIA_Group_PageMode, TRUE,
    MUIA_Group_ActivePage, page,
    Child, fieldlistgrp,
    Child, cmpfuncgrp,
    End;

  top = VGroup,
    Child, top,
    Child, d->d_TypeGroup,
    End;

  d->d_Ok = SimpleButton(GCS(MSG_Ok));
  d->d_Clear = SimpleButton(GCS(MSG_Clear));
  d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

  buttons = HGroup,
    MUIA_Group_SameWidth, TRUE,
    Child, d->d_Ok,
    Child, d->d_Clear,
    Child, d->d_Cancel,
    End;

  root = VGroup,
    Child, top,
    Child, buttons,
    End;

  sprintf(d->d_Title, "%s: %s", GCS(MSG_ChangeOrder), TableName(tab));
  obj = ExtWindowObject,
    MUIA_ExtWindow_Project, pr,
    MUIA_Window_ID, MAKE_ID('_', 'O','R','D'),
    MUIA_Window_NoMenus, TRUE,
    MUIA_Window_Title, d->d_Title,
    MUIA_Window_ScreenTitle, GuiScreenTitle,
    MUIA_HelpNode, "Changing orders",
    MUIA_UserData, d,
    WindowContents, root,
    End;

  if(obj == NULL)
    {
      PoolFreeMem(memPool, d, sizeof(struct EditOrderData));
      return(NULL);
    }

  DoMethod(d->d_Type,
	   MUIM_Notify, MUIA_Cycle_Active, MUIV_EveryTime,
	   d->d_TypeGroup, 3, 
	   MUIM_Set, MUIA_Group_ActivePage, MUIV_TriggerValue);

  set(d->d_Fields, MUIA_DDList_DragData, listo);
  MUIAddCallback(d->d_Fields, MUIA_Listview_DoubleClick, TRUE, 
		 insertClick, d->d_OrderList);

  set(d->d_OrderList, MUIA_DDList_DragData, lista);
  MUIAddCallback(d->d_OrderList, MUIA_Listview_DoubleClick, TRUE, 
		 orderClick, NULL);

  if(fields)
    FORNODE(field, fields)
    {
      type = FieldType(field);
      if(type != TYPE_BUTTON && type != TYPE_VIRTUAL && type != TYPE_REFERENCE)
	DoMethod(d->d_Fields, MUIM_List_InsertSingle, 
		 field, MUIV_List_Insert_Bottom);
    }

  if(oas)
    for(i=0; i < num; i++)
      if(oas[i].oa_Field)
	DoMethod(d->d_OrderList, MUIM_List_InsertSingle, 
		 &oas[i], MUIV_List_Insert_Bottom);

  DoMethod(d->d_CmpFuncStr, 
	   MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime, 
	   obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

  SetAttrs(d->d_Clear, 
	   MUIA_CycleChain, TRUE,
	   MUIA_ShortHelp, GCS(MSG_Order_ClearBH),
	   TAG_DONE);
  set(d->d_Ok, MUIA_CycleChain, TRUE);
  set(d->d_Cancel, MUIA_CycleChain, TRUE);

  DoMethod(d->d_Clear,
	   MUIM_Notify, MUIA_Pressed, FALSE,
	   d->d_OrderList, 1, MUIM_List_Clear);
  DoMethod(d->d_Clear,
	   MUIM_Notify, MUIA_Pressed, FALSE,
	   d->d_CmpFuncStr, 3, MUIM_Set, MUIA_String_Contents, "");	

  MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE, obj, finish, FALSE);
  MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE, obj, finish, FALSE);
  MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE, obj, finish, TRUE);

  DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
  MUIAddMember(app, obj);
  if(!GuiOpenWindow(obj))
    {
      MUIRemMember(app, obj);
      DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
      disposeEditOrder(app, obj);
      return(NULL);
    }

  set(obj, MUIA_Window_ActiveObject, d->d_Type);

  return(obj);
}

void GuiChangeOrder(struct Table *tab, struct OrderData *od)
{
  createEditOrder(tab, od);
}
