/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** pdfexport_mui.c
**
** Creation date: 14.11.2011
** Last modified: 06.07.2021
*/

#include "system.h"

#ifdef USE_HPDF

#include "pdfexport.h"
#include "print.h"
#include "guidosutil.h"
#include "dosutil.h"
#include "util.h"
#include "requesters.h"
#include "preferences.h"
#include "memory.h"
#include "appstrings_cat.h"
#include <libraries/asl.h>
#include <proto/hpdf.h>

#define FILENAME_LEN 255
#define HEADER_FOOTER_LEN 255

static struct
{
    struct HPDFSettings pes_HPDFSettings;
    char pes_Filename[FILENAME_LEN + 1];
} PDFExportSettings;

struct PDFExportData
{
    struct Project *d_Project;
    GuiObject d_SFWList;
    GuiObject d_Filename;
    GuiObject d_Paper;
    GuiObject d_Orientation;
    GuiObject d_FontName;
    GuiObject d_FontSize;
    GuiObject d_Header;
    GuiObject d_Footer;
    GuiObject d_Status;
    GuiObject d_Create, d_Cancel;
};

static const char *fontNames[] = 
{
    "Courier",
    "Helvetica", 
    "Times-Roman",
    NULL
};

static void disposePDFExport(Object *app, Object *obj)
{
    struct PDFExportData *d = pget(obj, MUIA_UserData);

    if (obj)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project),
                d, sizeof(struct PDFExportData));
        MUI_AddDisposeObject(app, obj);
    }
}

static bool getFontSize(GuiObject go, HPDF_REAL *size)
{
    const char *s = pget(go, MUIA_String_Contents);
    double val;

    if (StrToDouble(s, &val) && 
            val >= HPDF_FONT_SIZE_MIN && val <= HPDF_FONT_SIZE_MAX) {
        if (size) {
            *size = val;
        }
        return TRUE;
    }

    return FALSE;
}

static void savePDFExportSettings(struct PDFExportData *d)
{
    struct Project *pr = d->d_Project;
    struct HPDFSettings *hpdfs = &PDFExportSettings.pes_HPDFSettings;
    const char *s;
    long active;

    if (d->d_Filename) {
        s = EmptyNullStr(pget(d->d_Filename, MUIA_String_Contents));
        CodesetStrNBytesCopy(PDFExportSettings.pes_Filename, s, FILENAME_LEN);
    }
    active = nget(d->d_Paper, MUIA_Cycle_Active);
    hpdfs->hpdfs_PageSize = HPDFPageSizes[active];
    active = nget(d->d_Orientation, MUIA_Cycle_Active);
    hpdfs->hpdfs_PageDirection = HPDFPageDirections[active];
    active = nget(d->d_FontName, MUIA_Cycle_Active);
    hpdfs->hpdfs_FontName = fontNames[active];
    getFontSize(d->d_FontSize, &hpdfs->hpdfs_FontSize);
    hpdfs->hpdfs_Header = pget(d->d_Header, MUIA_String_Contents);
    hpdfs->hpdfs_Footer = pget(d->d_Footer, MUIA_String_Contents);

    SaveHPDFSettingsToPrefs(hpdfs, pr);
    LoadHPDFSettingsFromPrefs(hpdfs, pr);
}

static bool checkPDFExportSettings(GuiObject obj)
{
    struct PDFExportData *d = pget(obj, MUIA_UserData);

    if (d->d_Filename && PDFExportSettings.pes_Filename[0] == 0) {
        set(obj, MUIA_Window_Title, GCS(MSG_Err_NoFilename));
        set(obj, MUIA_Window_ActiveObject, d->d_Filename);
        GuiBeep();
        return FALSE;
    }
    if (!getFontSize(d->d_FontSize, NULL)) {
        set(obj, MUIA_Window_ActiveObject, d->d_FontSize);
        GuiBeep();
        return FALSE;
    }

    return TRUE;
}

static void HOOK_PROTO(finish, Object *obj, ulong *ok)
{
    struct PDFExportData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);

    savePDFExportSettings(d);

    if (*ok) {
        static const char filename_template[] = TMPDIR "TMP.mb.XXXXXX.pdf";
        char mem[sizeof(filename_template)];
        const char *filename = PDFExportSettings.pes_Filename;

        if (!checkPDFExportSettings(obj)) {
            return;
        }

        if (!d->d_Filename) {
            strcpy(mem, filename_template);
            ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
            MakeTemp(mem);
            filename = mem;
            if (filename[0] == 0) {
                Ask2(obj, NULL, AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_MkTemp));
                return;
            }
        }
        if (!ExportSFWListAsPDF(pr, d->d_SFWList, 
                    &PDFExportSettings.pes_HPDFSettings, filename)) {
            return;
        }
        if (!d->d_Filename) {
            GuiViewFile(filename, TRUE);
        }
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposePDFExport(app, obj);
}

static void updateStatus(struct PDFExportData *d)
{
    long pages;
    double truncated;
    const char *error;
    char buffer[128];
    const char *text = buffer;
    const char *pre = "\33c";
   
    savePDFExportSettings(d);
    if (!getFontSize(d->d_FontSize, NULL)) {
        error = GCS(MSG_PDFExport_InvalidFontSize);
    } else {
        error = GetExportSFWListAsPDFStats(d->d_Project, d->d_SFWList,
                &PDFExportSettings.pes_HPDFSettings, &pages, &truncated);
    }
    if (error) {
        text = error;
        pre = "\33c\33b";
    } else if (truncated > 0.0) {
        sprintf(buffer, GCS(MSG_PDFExport_Status_TooWide), pages, 
                (int)ceil(truncated));
        pre = "\33c\33b";
    } else {
        sprintf(buffer, GCS(MSG_PDFExport_Status_FitsPageWidth), pages);
    }
    SetAttrs(d->d_Status, 
            MUIA_Text_Contents, text, 
            MUIA_Text_PreParse, pre,
            NULL);
}

static void HOOK_PROTO(updateStatus, Object *obj, struct PDFExportData **x)
{
    updateStatus(*x);
}

static void exportPrintSFWListAsPDF(struct Project *pr, GuiObject sfw, 
        bool isExport)
{
    static const char *orientationNames[3];

    struct HPDFSettings *hpdfs = &PDFExportSettings.pes_HPDFSettings;
    Object *app = GuiApplication;
    struct MemPool *memPool = ProjectMemPool(pr);
    struct PDFExportData *d;
    Object *obj, *filenameLabel = NULL, *filenameGroup = NULL, *frame, *buttons;
    int i;

    if (app == NULL || memPool == NULL) {
	return;
    }

    if (HPDFBase == NULL) {
        HPDFBase = OpenLibrary("hpdf.library", 0);
        if (HPDFBase == NULL) {
            Ask(AT_ERROR, GCS(MSG_StarOk), "Requires hpdf.library!");
            return;
        }
    }

    LoadHPDFSettingsFromPrefs(hpdfs, pr);

    d = PoolAllocMemZero(memPool, sizeof(struct PDFExportData));
    if (d == NULL) {
        return;
    }
    d->d_Project = pr;
    d->d_SFWList = sfw;

    if (isExport) {
        d->d_Filename = StringObject,
            MUIA_CycleChain, TRUE,
            MUIA_ControlChar, Shortcut(GCS(MSG_PDFExport_File)),
            MUIA_ShortHelp, GCS(MSG_PDFExport_FileBH),
            MUIA_Frame, MUIV_Frame_String,
            MUIA_String_MaxLen, FILENAME_LEN + 1,
            MUIA_String_Format, MUIV_String_Format_Center,
            MUIA_String_Contents, PDFExportSettings.pes_Filename,
            End;
        filenameGroup = PopaslObject,
            MUIA_Popasl_Type, ASL_FileRequest,
            MUIA_Popstring_String, d->d_Filename,
            MUIA_Popstring_Button, GuiPopButtonSetCWD(MUII_PopFile, pr),
            End;
        filenameLabel = Label2(GCS(MSG_PDFExport_File));
    }

    d->d_Paper = CycleObject,
        MUIA_CycleChain, TRUE,
        MUIA_ControlChar, Shortcut(GCS(MSG_PDFExport_Paper)),
        MUIA_ShortHelp, GCS(MSG_PDFExport_PaperBH),
        MUIA_Cycle_Entries, HPDFPaperNames,
        End;
    for (i=0; HPDFPaperNames[i] != NULL; i++) {
        if (HPDFPageSizes[i] == hpdfs->hpdfs_PageSize) {
            set(d->d_Paper, MUIA_Cycle_Active, i);
            break;
        }
    }

    orientationNames[0] = GCS(MSG_PDFExport_Orientation_Portrait);
    orientationNames[1] = GCS(MSG_PDFExport_Orientation_Landscape);
    orientationNames[2] = NULL;
    d->d_Orientation = CycleObject,
        MUIA_CycleChain, TRUE,
        MUIA_ControlChar, Shortcut(GCS(MSG_PDFExport_Orientation)),
        MUIA_ShortHelp, GCS(MSG_PDFExport_OrientationBH),
        MUIA_Cycle_Entries, orientationNames,
        End;
    for (i=0; orientationNames[i] != NULL; i++) {
        if (HPDFPageDirections[i] == hpdfs->hpdfs_PageDirection) {
            set(d->d_Orientation, MUIA_Cycle_Active, i);
            break;
        }
    }

    d->d_FontName = CycleObject,
        MUIA_CycleChain, TRUE,
        MUIA_ControlChar, Shortcut(GCS(MSG_Font)),
        MUIA_Cycle_Entries, fontNames,
        End;
    for (i=0; fontNames[i] != NULL; i++) {
        if (!strcmp(fontNames[i], hpdfs->hpdfs_FontName)) {
            set(d->d_FontName, MUIA_Cycle_Active, i);
            break;
        }
    }
    
    d->d_FontSize = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_Weight, 50,
	MUIA_ControlChar, Shortcut(GCS(MSG_Size)),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_String_MaxLen, 5 + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, DoubleToStr(hpdfs->hpdfs_FontSize, 1, 
                (short)GetPrefsValue(pr, PREFS_REALFORMAT)),
	End;

    d->d_Header = StringObject,
        MUIA_CycleChain, TRUE,
	MUIA_ControlChar, Shortcut(GCS(MSG_PDFExport_Header)),
	MUIA_ShortHelp, GCS(MSG_PDFExport_HeaderBH),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_String_MaxLen, HEADER_FOOTER_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
	MUIA_String_Contents, hpdfs->hpdfs_Header,
	End;

    d->d_Footer = StringObject,
        MUIA_CycleChain, TRUE,
	MUIA_ControlChar, Shortcut(GCS(MSG_PDFExport_Footer)),
	MUIA_ShortHelp, GCS(MSG_PDFExport_FooterBH),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_String_MaxLen, HEADER_FOOTER_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
	MUIA_String_Contents, hpdfs->hpdfs_Footer,
	End;

    d->d_Status = TextObject,
        MUIA_Frame, MUIV_Frame_Text,
        MUIA_Background, MUII_TextBack,
        MUIA_Text_Contents, " ",
        MUIA_Text_SetMin, FALSE,
        End;

    MUIAddCallback(d->d_Paper, MUIA_Cycle_Active, MUIV_EveryTime, 
            updateStatus, d);
    MUIAddCallback(d->d_Orientation, MUIA_Cycle_Active, MUIV_EveryTime, 
            updateStatus, d);
    MUIAddCallback(d->d_FontName, MUIA_Cycle_Active, MUIV_EveryTime, 
            updateStatus, d);
    MUIAddCallback(d->d_FontSize, MUIA_String_Contents, MUIV_EveryTime, 
            updateStatus, d);
    MUIAddCallback(d->d_Header, MUIA_String_Contents, MUIV_EveryTime, 
            updateStatus, d);
    MUIAddCallback(d->d_Footer, MUIA_String_Contents, MUIV_EveryTime, 
            updateStatus, d);
    updateStatus(d);

    frame = VGroup,
        MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
        MUIA_Group_Columns, 2,
        isExport? Child : TAG_IGNORE, filenameLabel,
        isExport? Child : TAG_IGNORE, filenameGroup,
        Child, Label2(GCS(MSG_PDFExport_Paper)),
        Child, d->d_Paper,
        Child, Label2(GCS(MSG_PDFExport_Orientation)),
        Child, d->d_Orientation,
        Child, Label2(GCS(MSG_Font)),
        Child, HGroup,
            Child, d->d_FontName,
            Child, Label2(GCS(MSG_Size)),
            Child, d->d_FontSize,
            End,
        Child, Label2(GCS(MSG_PDFExport_Header)),
        Child, d->d_Header,
        Child, Label2(GCS(MSG_PDFExport_Footer)),
        Child, d->d_Footer,
        Child, Label2(GCS(MSG_PDFExport_Status)),
        Child, d->d_Status,
        End;

    d->d_Create = SimpleButton(GCS(MSG_PDFExport_CreatePDF));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
        MUIA_Group_SameWidth, TRUE,
	Child, d->d_Create,
	Child, d->d_Cancel,
        End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_Title, GCS(MSG_PDFExport),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'P', 'D', 'F'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_Window_AppWindow, TRUE,
        MUIA_HelpNode, "Exporting a list as PDF",
        MUIA_UserData, d,
	WindowContents, VGroup,
            Child, frame,
            Child, buttons,
            End,
	End;

    if (obj == NULL) {
        PoolFreeMem(memPool, d, sizeof(struct PDFExportData));
	return;
    }

    set(d->d_Create, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    if (d->d_Filename) {
        MUIAddAppMessageNotification(obj, d->d_Filename);
        DoMethod(d->d_Filename,
                MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
                obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Create);
    }
    DoMethod(d->d_FontSize,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Create);

    MUIAddCallback2(obj, MUIA_Window_CloseRequest, TRUE, 
            obj, finish, FALSE);
    MUIAddCallback2(d->d_Cancel, MUIA_Pressed, FALSE, 
            obj, finish, FALSE);
    MUIAddCallback2(d->d_Create, MUIA_Pressed, FALSE, 
            obj, finish, TRUE);

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if (!GuiOpenWindow(obj)) {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposePDFExport(app, obj);
        return;
    }

    set(obj, MUIA_Window_ActiveObject, d->d_Filename);
}

void GuiExportSFWListAsPDF(struct Project *pr, GuiObject sfw)
{
    exportPrintSFWListAsPDF(pr, sfw, TRUE);
}

#if !defined(USE_ANSI_PRINT)

void GuiPrintSFWList(struct Project *pr, GuiObject sfw)
{
    exportPrintSFWListAsPDF(pr, sfw, FALSE);
}

#endif
#endif

