/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** ProgOutput_mui.c
**
** Creation date: 29.01.1996
** Last modified: 06.11.2014
*/

#include "system.h"
#include "programoutput.h"
#include "project.h"
#include "preferences.h"
#include "requesters.h"
#include "util.h"
#include "error.h"
#include "memory.h"
#include "appstrings_cat.h"
#include "extclasses/ExtApplication_mcc.h"
#include <libraries/asl.h>

struct ProgOutputData
{
    struct Project *d_Project;
    GuiObject d_Filename;
    GuiObject d_Default;
    GuiObject d_Append;
    GuiObject d_Ok, d_Cancel;
};

static void disposeProgOutput(Object *app, Object *obj)
{
    struct ProgOutputData *d = pget(obj, MUIA_UserData);

    if (obj) {
        PoolFreeMem(ProjectMemPool(d->d_Project), d,
                sizeof(struct ProgOutputData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void HOOK_PROTO(closeProgOutput, Object *obj, ulong *ok)
{
    struct ProgOutputData *d = pget(obj, MUIA_UserData);
    Object *app = pget(obj, MUIA_ApplicationObject);

    if (*ok) {
        const char *filename = pget(d->d_Filename, MUIA_String_Contents);
        bool def = nget(d->d_Default, MUIA_Selected);
        bool append = nget(d->d_Append, MUIA_Selected);
	struct Project *pr = d->d_Project;
        long error = ObtainProjectChangeLock(pr);

        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        if (!def) {
            SetPrefsString(pr, PREFS_PROGRAMOUTPUT_FILENAME, filename);
        }
	SetPrefsValue(pr, PREFS_PROGRAMOUTPUT_DEFAULT, def);
	SetPrefsValue(pr, PREFS_PROGRAMOUTPUT_APPEND, append);
    }
    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    GuiSetSleep(FALSE);

    disposeProgOutput(app, obj);
}

static void createProgOutput(struct Project *pr)
{
    bool def = GetPrefsValue(pr, PREFS_PROGRAMOUTPUT_DEFAULT);
    const char *defaultName = DEFAULT_OUTPUT_FILE;
    const char *filename = def? 
        defaultName : GetPrefsString(pr, PREFS_PROGRAMOUTPUT_FILENAME);
    struct MemPool *memPool = ProjectMemPool(pr);
    GuiObject app = GuiApplication;
    struct ProgOutputData *d;
    GuiObject obj, popFile, group, root, buttons;

    if (app == NULL) {
        return;
    }

    d = PoolAllocMemZero(memPool, sizeof(struct ProgOutputData));
    if (d == NULL) {
        return;
    }
    d->d_Project = pr;

    d->d_Filename = StringObject,
        MUIA_CycleChain, TRUE,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_ControlChar, Shortcut(GCS(MSG_ProgOutput_Filename)),
        MUIA_Disabled, def,
        MUIA_String_MaxLen, 80,
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, filename,
	End;
    popFile = PopaslObject,
        MUIA_ShortHelp, GCS(MSG_ProgOutput_FilenameBH),
	MUIA_Popasl_Type, ASL_FileRequest,
	MUIA_Popstring_String, d->d_Filename,
	MUIA_Popstring_Button, GuiPopButtonSetCWD(MUII_PopFile, pr),
	End;

    d->d_Default = MUI_MakeObject(MUIO_Checkmark,
            (IPTR)GCS(MSG_ProgOutput_Default));

    d->d_Append = MUI_MakeObject(MUIO_Checkmark,
            (IPTR)GCS(MSG_ProgOutput_Append));

    group = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
	Child, HGroup,
            Child, Label(GCS(MSG_ProgOutput_Filename)),
            Child, popFile,
            End,
        Child, HGroup,
            Child, Label(GCS(MSG_ProgOutput_Default)),
            Child, d->d_Default,
	    Child, HSpace(0),
            Child, Label(GCS(MSG_ProgOutput_Append)),
            Child, d->d_Append,
	    Child, HSpace(0),
            End,
	End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, d->d_Ok,
	Child, d->d_Cancel,
	End;

    root = GroupObject,
	Child, group,
	Child, buttons,
	End;

    obj = WindowObject,
	MUIA_Window_Title, GCS(MSG_ProgOutput),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'O','U','T'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_Window_AppWindow, TRUE,
        MUIA_HelpNode, "Program output file",
        MUIA_UserData, d,
	WindowContents, root,
	End;

    if (obj == NULL) {
        PoolFreeMem(memPool, d, sizeof(struct ProgOutputData));
	return;
    }

    MUIAddAppMessageNotification(obj, d->d_Filename);
    DoMethod(d->d_Filename,
            MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
            obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    SetAttrs(d->d_Default,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_ProgOutput_DefaultBH),
        MUIA_Selected, def,
        TAG_DONE);
    DoMethod(d->d_Default,
	MUIM_Notify, MUIA_Selected, TRUE,
	d->d_Filename, 3, MUIM_Set, MUIA_String_Contents, defaultName);
    DoMethod(d->d_Default,
	MUIM_Notify, MUIA_Selected, MUIV_EveryTime,
	d->d_Filename, 3, MUIM_Set, MUIA_Disabled, MUIV_TriggerValue);

    SetAttrs(d->d_Append,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_ProgOutput_AppendBH),
        MUIA_Selected, GetPrefsValue(pr, PREFS_PROGRAMOUTPUT_APPEND),
        TAG_DONE);

    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    MUIAddCallback2(obj, MUIA_Window_CloseRequest, TRUE, 
            obj, closeProgOutput, FALSE);
    MUIAddCallback2(d->d_Cancel, MUIA_Pressed, FALSE, 
            obj, closeProgOutput, FALSE);
    MUIAddCallback2(d->d_Ok, MUIA_Pressed, FALSE, 
            obj, closeProgOutput, TRUE);

    GuiSetSleep(TRUE);
    MUIAddMember(app, obj);

    if (!GuiOpenWindow(obj)) {
        MUIRemMember(app, obj);
        GuiSetSleep(FALSE);
        disposeProgOutput(app, obj);
        return;
    }

    set(obj, MUIA_Window_ActiveObject, d->d_Filename);
}

void GuiProgOutput(struct Project *pr)
{
    createProgOutput(pr);
}

