/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** ProgramEditor_mui.c
**
** Creation date: 15.07.1995
** Last modified: 28.12.2021
*/

#include "programeditor.h"

#ifdef ENABLE_PROGRAM_EDITOR

#include "menueditor.h"
#include "menuprogram.h"
#include "project.h"
#include "preferences.h"
#include "lang/compile.h"
#include "lang/program.h"
#include "requesters.h"
#include "dosutil.h"
#include "util.h"
#include "strhandle.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"
#include "extclasses/ExtApplication_mcc.h"
#include "extclasses/ExtEditor_mcc.h"

struct ProgramEditorData
{
    struct Project *d_Project;
    Object *d_Editor;
};

static void disposeProgramEditor(Object *app, Object *obj)
{
    struct ProgramEditorData *d = pget(obj, MUIA_UserData);

    if(obj) {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(*d));
        MUI_AddDisposeObject(app, obj);
    }
}

void CloseProgramEditor(GuiObject obj)
{
    if (obj != NULL) {
        struct ProgramEditorData *d = pget(obj, MUIA_UserData);
        struct Project *pr = d->d_Project;
        Object *app = pget(obj, MUIA_ApplicationObject);
        long row, col;

        SetPrefsValue(pr, PREFS_PROGRAMEDITOR_TOP, 
		nget(d->d_Editor, MUIA_ExtEditor_Top));
        ExtEditorGetCursor(d->d_Editor, &row, &col);
        SetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSX, col);
        SetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSY, row);

        GuiCloseWindow(obj);
        MUIRemMember(app, obj);
        disposeProgramEditor(app, obj);
        SetProjectProgramEditor(pr, NULL);
    }
}

bool ProgramEditorCheckClose(GuiObject obj)
{
    if (obj) {
	struct ProgramEditorData *d = pget(obj, MUIA_UserData);

	if (nget(d->d_Editor, MUIA_ExtEditor_HasChanged)) {
	    if (!Ask2(obj, GCS(MSG_ProgramEditor), AT_WARN,
			GCS(MSG_CloseAnswers), GCS(MSG_ProgramEditor_Close))) {
		return FALSE;
	    }
	}
    }
    return TRUE;
}

static bool compile(Object *obj, bool reload)
{
    struct ProgramEditorData *d = pget(obj, MUIA_UserData);

    return MenuProgramCompile(d->d_Project, reload);
}

static void HOOK_PROTO(compileClick, Object *obj, void *dummy)
{
    if (compile(obj, TRUE)) {
        DoMethod(pget(obj, MUIA_ApplicationObject), 
		MUIM_Application_ReturnID, ID_OK);
    }
}

static void HOOK_PROTO(closeProgramEditor, Object *obj, ulong *ok)
{
    if (*ok && !compile(obj, FALSE)) {
        return;
    } else if(!*ok && !ProgramEditorCheckClose(obj)) {
	return;
    }

    DoMethod(pget(obj, MUIA_ApplicationObject),
        MUIM_Application_ReturnID, *ok ? ID_OK : ID_CANCEL);
    CloseProgramEditor(obj);
}

static void HOOK_PROTO(undoClick, Object *obj, void *dummy)
{
    struct ProgramEditorData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    struct Program *prg = ProjectProgram(pr);
    struct SourceText *st = GetSourceText(prg);
    const char *text = SourceTextText(st);

    if(text)
    {
        set(d->d_Editor, MUIA_ExtEditor_Quiet, TRUE);
        set(d->d_Editor, MUIA_ExtEditor_Contents, text);
        set(d->d_Editor, MUIA_ExtEditor_Top, 
                GetPrefsValue(pr, PREFS_PROGRAMEDITOR_TOP));
        ExtEditorSetCursor(d->d_Editor, 
                GetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSY),
                GetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSX));
        set(d->d_Editor, MUIA_ExtEditor_Quiet, FALSE);
    }

    FreeSourceText(st);
    set(d->d_Editor, MUIA_ExtEditor_HasChanged, FALSE);
}

static Object *createProgramEditorGui(struct Project *pr)
{
    Object *app = GuiApplication;
    struct MemPool *memPool = ProjectMemPool(pr);
    struct ProgramEditorData *d;
    Object *compclose, *comp, *undo;
    Object *obj = NULL;

    if(app == NULL)
        return(NULL);

    d = PoolAllocMemZero(memPool, sizeof(struct ProgramEditorData));
    if(d == NULL)
        return(NULL);
    d->d_Project = pr;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
        MUIA_Window_Width, MUIV_Window_Width_Visible(80),
        MUIA_Window_Height, MUIV_Window_Height_Visible(80),
        MUIA_Window_Title, GCS(MSG_EditProgram),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
        MUIA_Window_ID, MAKE_ID('_', 'P', 'R', 'G'),
        MUIA_HelpNode, "Program editor",
        MUIA_UserData, d,
            WindowContents, VGroup,
            Child, MUINewEditor(pr, 0, FALSE, TRUE,
                (char *)MUII_TextBack, NULL, &d->d_Editor),
            Child, HGroup,
                MUIA_Group_SameWidth, TRUE,
                Child, compclose = SimpleButton(GCS(MSG_ProgramEditor_CompileAndClose)),
                Child, comp = SimpleButton(GCS(MSG_ProgramEditor_Compile)),
                Child, undo = SimpleButton(GCS(MSG_ProgramEditor_Revert)),
                End,
            End,
        End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct ProgramEditorData));
        return(NULL);
    }

    set(d->d_Editor, MUIA_CycleChain, TRUE);
    SetAttrs(compclose,
        MUIA_ShortHelp, GCS(MSG_ProgramEditor_CompileAndCloseBH),
        MUIA_CycleChain, TRUE,
        TAG_DONE);
    SetAttrs(comp,
        MUIA_ShortHelp, GCS(MSG_ProgramEditor_CompileBH),
        MUIA_CycleChain, TRUE,
        TAG_DONE);
    SetAttrs(undo,
        MUIA_ShortHelp, GCS(MSG_ProgramEditor_RevertBH),
        MUIA_CycleChain, TRUE,
        TAG_DONE);

    MUIAddCallback2(obj, MUIA_Window_CloseRequest, TRUE,
        obj, closeProgramEditor, FALSE);
    MUIAddCallback2(compclose, MUIA_Pressed, FALSE,
        obj, closeProgramEditor, TRUE);
    MUIAddCallback2(comp, MUIA_Pressed, FALSE,
        obj, compileClick, NULL);
    MUIAddCallback2(undo, MUIA_Pressed, FALSE,
        obj, undoClick, NULL);

    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        disposeProgramEditor(app, obj);
        return(NULL);
    }

    return(obj);
}

GuiObject OpenProgramEditor(struct Project *pr)
{
    Object *obj = (Object *)ProjectProgramEditor(pr);
    struct ProgramEditorData *d = pget(obj, MUIA_UserData);
    bool external = GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL);

    if (obj == NULL) {
	if (external) {
            return NULL;
	}

        obj = createProgramEditorGui(pr);
        if (obj == NULL) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_CantOpenProgramEditor));
            return NULL;
        }
	d = pget(obj, MUIA_UserData);
        undoClickFunc(NULL, obj, NULL);		/* restore */
        SetProjectProgramEditor(pr, obj);
    } else {
	/* remove busy pointer */
	while (nget(obj, MUIA_Window_Sleep)) {
	    set(obj, MUIA_Window_Sleep, FALSE);
	}
	GuiWindowToFront(obj);
	GuiActivateWindow(obj);
    }

    set(obj, MUIA_Window_ActiveObject, d->d_Editor);

    return obj;
}

void ProgramEditorShowError(struct Project *pr, bool reload,
	const char *error, long line, long col, const char *filename)
{
    Object *obj = (Object *)ProjectProgramEditor(pr);
    struct ProgramEditorData *d = pget(obj, MUIA_UserData);
    short rc;

    if (error == NULL) {
	return;
    } 

    if (IsProjectErrorString(error)) {
        Ask(AT_ERROR, GCS(MSG_StarOk), error);
        return;
    }

    if (line < 0) {
        Ask2(obj, NULL, AT_WARN, GCS(MSG_StarOk),
            GCS(MSG_ProgramEditor_NoDebugInfo), error);
	return;
    }
    if (filename == NULL && GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL)) {
        filename = GetPrefsString(pr, PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);
    }
    if (filename) {
	Ask2(obj, NULL, AT_ERROR, GCS(MSG_StarOk),
	      	GCS(MSG_ProgramEditor_IncludeFileError), 
		filename, line, col, error);
    } else if (!ProjectAdminMode(pr)) {
	Ask2(obj, NULL, AT_ERROR, GCS(MSG_StarOk),
	      	GCS(MSG_ProgramEditor_NotAdmin), line, col, error);
    } else {
	/*
	** show the error in BeeBase' program editor
	** The editor might not be opened yet,
	** open and unchanged, or open and changed.
	*/
	if (obj && reload && nget(d->d_Editor, MUIA_ExtEditor_HasChanged)) {
	    /* we should restore the program source */
	    GuiBeep();
	    rc = Ask2(obj, GCS(MSG_ProgramEditor), AT_WARN,
		    GCS(MSG_ProgramEditor_ShowAnswers),
		    GCS(MSG_ProgramEditor_ShowError), line, col, error);
	    if (rc == 0) {
		return;				/* user canceled */
	    } else if (rc == 2) {
		undoClickFunc(NULL, obj, NULL);	/* restore */
	    }
	}
	obj = OpenProgramEditor(pr);	/* either creates or brings to front */
    	d = pget(obj, MUIA_UserData);
	if (obj) {
	    set(obj, MUIA_Window_Title, error);
	    ExtEditorSetCursor(d->d_Editor, line, col);
	    GuiBeep();
	} /* else OpenProgramEditor popped up an error requester */
    }
}

bool ProgramEditorFix(struct Project *pr)
{
    Object *obj = (Object *)ProjectProgramEditor(pr);
    struct ProgramEditorData *d = pget(obj, MUIA_UserData);
    bool result = TRUE;
    short rc;

    if (obj != NULL && nget(d->d_Editor, MUIA_ExtEditor_HasChanged)) {
    	/* we should restore the program source */
        GuiBeep();
    	rc = Ask2(obj, GCS(MSG_ProgramEditor), AT_WARN,
		GCS(MSG_ProgramEditor_ShowAnswers),
	    	GCS(MSG_ProgramEditor_ShowDeleted));
	if (rc == 0) {
	    result = FALSE;			/* user canceled */
	} else if (rc == 2) {
	    undoClickFunc(NULL, obj, NULL);	/* restore */
	}
    }

    if (result) {
	obj = OpenProgramEditor(pr);	/* either creates or brings to front */
	if (obj) {
            GuiSetSleep(TRUE);
	    set(obj, MUIA_Window_NoMenus, TRUE);
	    compile(obj, TRUE);			/* show the location to fix */
	    result = WaitOk();
            GuiSetSleep(FALSE);
	}
    }

    return result;
}

const char *ProgramEditorCompile(GuiObject pe, bool reload,
	struct Program **compiled, long *line, long *col, const char **filename)
{
    if (pe == NULL)
	return GCS(MSG_Err_Internal);
    else
    {
	Object *obj = (Object *)pe;
	struct ProgramEditorData *d = pget(obj, MUIA_UserData);
	struct MemPool *memPool = ProjectMemPool(d->d_Project);
	Object *edit = d->d_Editor;
	struct Program *prg = NULL;
	char *text = NULL; 
        const char *error = NULL;

    	GuiSetSleep(TRUE);
	set(obj, MUIA_Window_Title, GCS(MSG_EditProgram));
	text = ExtEditorGetContents(d->d_Editor, memPool);
	if (text == NULL)
	    error = GCS(MSG_Err_Mem);
	else
	{
	    error = CompileProjectProgram(d->d_Project, text, strlen(text),
		    &prg, line, col, filename);
	    if (compiled)
		*compiled = prg;
	}
	FreeString(memPool, text);

        if (error == NULL && prg != NULL && reload)
        {
    	    struct SourceText *st = GetSourceText(prg);
            const char *text = SourceTextText(st);
    
            if (text)
    	    {
                long top, row, col;

                top = nget(edit, MUIA_ExtEditor_Top);
                ExtEditorGetCursor(edit, &row, &col);

                set(edit, MUIA_ExtEditor_Quiet, TRUE);
                set(edit, MUIA_ExtEditor_Contents, text);
                set(edit, MUIA_ExtEditor_Top, top);
                ExtEditorSetCursor(edit, row, col);
                set(edit, MUIA_ExtEditor_Quiet, FALSE);
	    }
            FreeSourceText(st);
            set(d->d_Editor, MUIA_ExtEditor_HasChanged, FALSE);
        }
        GuiSetSleep(FALSE);

        return error;
    }
}

const char *ProgramEditorCreateSourceExternal(GuiObject pe)
{
    if (pe == NULL)
	return GCS(MSG_Err_Internal);
    else
    {
	Object *obj = (Object *)pe;
	struct ProgramEditorData *d = pget(obj, MUIA_UserData);
	struct MemPool *memPool = ProjectMemPool(d->d_Project);
	char *text = NULL;
	const char *error = NULL;

    	GuiSetSleep(TRUE);
	text = ExtEditorGetContents(d->d_Editor, memPool);
	if (text == NULL)
	    error = GCS(MSG_Err_Mem);
	else
	{
	    error = CreateSourceExternal(d->d_Project, text, strlen(text));
	    FreeString(memPool, text);
	}
	GuiSetSleep(FALSE);
	return error;
    }
}

#else

void CloseProgramEditor(GuiObject obj) {}

GuiObject OpenProgramEditor(struct Project *pr) { return 0; }

bool ProgramEditorFix(struct Project *pr) { return false; }

void ProgramEditorShowError(struct Project *pr, bool reload,
	const char *error, long line, long col, const char *filename) {}

const char *ProgramEditorCompile(GuiObject pe, bool reload,
	struct Program **compiled, long *line, long *col, char **filename)
{ return 0; }

const char *ProgramEditorCreateSourceExternal(GuiObject pe) { return 0; }

#endif

