/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** queryeditor_mui.c
**
** Creation date: 28.05.1995
** Last modified: 03.03.2022
*/

#include "system.h"
#include "queryeditor.h"
#include "project.h"
#include "imexport.h"
#include "pdfexport.h"
#include "print.h"
#include "lang/compile.h"
#include "lang/program.h"
#include "memory.h"
#include "requesters.h"
#include "labels.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"
#include "buildguifield_mui.h"          /* for SFWListDoubleClick */
#include "extclasses/ExtEditor_mcc.h"
#include "extclasses/SFWList_mcc.h"
#include "extclasses/DDList_mcc.h"

struct QueryEditorData
{
    struct Project *d_Project;
    void *d_QueryEditor;
    struct Query *d_ActiveQuery;
    Object *d_QueryName;
    Object *d_Queries;
    Object *d_QueriesPopup;
    char **d_TableNames;
    Object *d_QueryTable;
    Object *d_Export, *d_PDF, *d_Print;
    Object *d_Editor;
    Object *d_SFWList;
};


#ifdef USE_HPDF
static void HOOK_PROTO(pdfClick, Object *obj, void *dummy)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);

    GuiExportSFWListAsPDF(d->d_Project, d->d_SFWList);
}
#endif

static void HOOK_PROTO(exportClick, Object *obj, void *dummy)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);

    GuiExportSFWList(d->d_Project, d->d_SFWList);
}

static void HOOK_PROTO(printClick, Object *obj, void *dummy)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);

    GuiPrintSFWList(d->d_Project, d->d_SFWList);
}

static void setButtonsEnabled(struct QueryEditorData *d)
{
    bool enabled = (nget(d->d_SFWList, MUIA_List_Entries) > 0);

    set(d->d_Export, MUIA_Disabled, !enabled);
    set(d->d_PDF, MUIA_Disabled, !enabled);
    set(d->d_Print, MUIA_Disabled, !enabled);
}

static void setActiveQuery(Object *obj, struct Query *sq)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);

    if(obj && sq)
    {
        const char *name = QueryName(sq);
        struct Program *prg = QueryProgram(sq);
        struct SourceText *st = NULL;
        const char *src;
        char *kill, c;
        long len;

        /* clear window title */
        set(obj, MUIA_Window_Title, GCS(MSG_QueryEditor));

        if(name == NULL)
            name = GCS(MSG_Unnamed);
        set(d->d_QueryName, MUIA_String_Contents, name);

        set(d->d_QueryTable, MUIA_Cycle_Active, NodeNum(QueryTable(sq)));

        /* set source text */
        st = GetSourceText(prg);
        src = SourceTextText(st);
        len = src? strlen(src) : 0;

        if (len >= 3) {
            src++;          /* skip "("   */
            len -= 3;       /* skip "\n)" */
            kill = (char *)src + len;
            c = *kill;
            *kill = 0;
            GuiSetEditorText(d->d_Editor, src);
            *kill = c;
        } else {
            GuiSetEditorText(d->d_Editor, "");
        }
        GuiSetEditorChanged(d->d_Editor, FALSE);
        FreeSourceText(st);

        DoMethod(d->d_SFWList, MUIM_SFWList_SetElement, NULL);
        setButtonsEnabled(d);

        d->d_ActiveQuery = sq;
    }
}

static const char *saveQuery(Object *obj, long *line, long *col, 
        bool force_compile)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);

    if (obj && d->d_ActiveQuery) {
        struct Project *pr = QueryEditorProject(d->d_QueryEditor);
        struct List *queries = QueryEditorQueries(d->d_QueryEditor);
        struct MemPool *memPool = ProjectMemPool(pr);
        const char *oldName = QueryName(d->d_ActiveQuery);
        char *name, *src = NULL;
        struct Table *tab = NULL;
        struct Query *q;
        long error;
        const char *errstr = NULL;
        struct Program *prg = NULL;
        ulong modified = TRUE;

        name = pget(d->d_QueryName, MUIA_String_Contents);
        if (strcmp(name? name : "", oldName? oldName : "") != 0) {
            error = ObtainProjectChangeLock(d->d_Project);
            if (error != OK) {
                return GES(error);
            }
            SetQueryName(d->d_ActiveQuery, name);
            ProjectChange(pr, TRUE);
        }

        tab = FindNodeNum(ProjectTables(pr), 
                nget(d->d_QueryTable, MUIA_Cycle_Active));
        if (tab != QueryTable(d->d_ActiveQuery)) {
            error = ObtainProjectChangeLock(d->d_Project);
            if (error != OK) {
                return GES(error);
            }
            if (queries) {
                FORNODE(q, queries) {
                    if (QueryTable(q) == tab) {
                        SetQueryTable(q, NULL);
                    }
                }
            }
            SetQueryTable(d->d_ActiveQuery, tab);
	    ProjectChange(pr, TRUE);
        }

        modified = nget(d->d_Editor, MUIA_ExtEditor_HasChanged);
        if (modified || force_compile) {
            src = ExtEditorGetContents(d->d_Editor, memPool);
            if (src == NULL) {
                errstr = GCS(MSG_Err_Mem);
            } else {
                errstr = CompileQuery(pr, src, strlen(src), &prg, line, col);
            }
            FreeString(memPool, src);

            if (errstr != NULL) {
                return errstr;
            }
            if (modified) {
                error = ObtainProjectChangeLock(d->d_Project);
                if (error != OK) {
                    return GES(error);
                }
            }

            SetQueryProgram(d->d_ActiveQuery, prg);
            setActiveQuery(obj, d->d_ActiveQuery);     /* refreshes source */
            if (modified) {
                ProjectChange(pr, TRUE);
                set(d->d_Editor, MUIA_ExtEditor_HasChanged, FALSE);
            }
        }
    }
    return NULL;
}

static void showerror(Object *obj, const char *error, long line, long col)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);

    if (error != NULL && error != EXECUTE_ABORT) {
        if (IsProjectErrorString(error)) {
            Ask(AT_ERROR, GCS(MSG_StarOk), error);
        } else if (d != NULL) {
            set(obj, MUIA_Window_Title, error);
            set(obj, MUIA_Window_ActiveObject, d->d_Editor);
            ExtEditorSetCursor(d->d_Editor, line, 
                    (line == 1 && col > 0)? col - 1 : col);
            set(obj, MUIA_Window_ActiveObject, d->d_Editor);
            GuiBeep();
        }
    }
}

void CloseQueryGui(Object *obj)
{
    if(obj)
    {
        struct QueryEditorData *d = pget(obj, MUIA_UserData);
        Object *app = pget(obj, MUIA_ApplicationObject);
        struct MemPool *memPool = ProjectMemPool(d->d_Project);

        saveQuery(obj, NULL, NULL, FALSE);

        GuiCloseWindow(obj);
        MUIRemMember(app, obj);
  	MUI_AddDisposeObject(app, obj);
        FreeLabels(memPool, d->d_TableNames);
        PoolFreeMem(memPool, d, sizeof(struct QueryEditorData));
    }
}

static struct Project *queryProject = NULL;

static const char *outputFunc(struct Element *elem)
{
    struct QueryEditor *se = ProjectQueryEditor(queryProject);
    Object *obj = QueryEditorGui(se);
    struct QueryEditorData *d = pget(obj, MUIA_UserData);
    char *error = NULL;

    if(obj == NULL)
        return(GCS(MSG_Err_Internal));

    error = (char *)DoMethod(d->d_SFWList, MUIM_SFWList_SetElement, elem);
    setButtonsEnabled(d);

    return error;
}

static void runClick(Object *obj)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    const char *error = NULL;
    struct Program *prg = NULL;
    long line, col;

    if(d->d_ActiveQuery == NULL)
        return;

    GuiSetSleep(TRUE);
    error = saveQuery(obj, &line, &col, TRUE);
    if(error == NULL)
    {
        /* execute program */

        prg = QueryProgram(d->d_ActiveQuery);
        SetTableProgramRecords(pr);
        queryProject = pr;
        error = ExecuteProgram(prg, TRUE, NULL, NULL, TRUE, 
			       outputFunc, &line, &col, NULL);
        queryProject = NULL;
    }
    GuiSetSleep(FALSE);

    if(error != NULL)
        showerror(obj, error, line, col);
}

static void HOOK_PROTO(runClick, Object *obj, void *dummy)
{
    runClick(obj);
}

static void HOOK_PROTO(acknowledge, Object *obj, void *dummy)
{
    const char *error = NULL;
    long line, col;

    error = saveQuery(obj, &line, &col, FALSE);
    if(error != NULL)
        showerror(obj, error, line, col);
}

static ulong HOOK_PROTO(popListDisplay, const char **dst, struct Query *sq)
{
    const char *name = QueryName(sq);

    if(name == NULL)
        name = GCS(MSG_Unnamed);
    *dst = name;

    return(0);
}
static HOOK_STRUCT(popListDisplay)

static long HOOK_PROTO(StrObj, Object *pop, Object *str)
{
    Object *obj = GuiObjectWindow(str);
    struct QueryEditorData *d = pget(obj, MUIA_UserData);
    struct List *queries = QueryEditorQueries(d->d_QueryEditor);
    const char *error;
    long line, col;

    error = saveQuery(obj, &line, &col, FALSE);
    if(error != NULL)
    {
        set(obj, MUIA_Window_Title, error);
        ExtEditorSetCursor(d->d_Editor, line, col);
        GuiBeep();
        return(FALSE);
    }

    MUIListSetList(pop, queries, ProjectMemPool(d->d_Project));
    MUIListSetActiveNode(pop, d->d_ActiveQuery);

    return(TRUE);
}
static HOOK_STRUCT(StrObj)

static void HOOK_PROTO(setActiveQuery, Object *obj, struct Query **x)
{
    setActiveQuery(obj, *x);
}
static HOOK_STRUCT(setActiveQuery)

static void HOOK_PROTO(ObjStr, Object *pop, Object *str)
{
    Object *obj = GuiObjectWindow(str);
    struct Query *sq = GuiListActiveEntry(pop);

    /* clear listview */
    set(pop, MUIA_List_Quiet, TRUE);
    DoMethod(pop, MUIM_List_Clear);
    set(pop, MUIA_List_Quiet, FALSE);

    /* 
    ** set active query
    ** there seems to be a bug in MorphOS List.mui class that doesn't 
    ** allow recursion into another List instance.  This is a woraround.
    */
    DoMethod(GuiApplication, MUIM_Application_PushMethod, 
            obj, 3, MUIM_CallHook, &setActiveQueryHook, sq);
}
static HOOK_STRUCT(ObjStr)

static void HOOK_PROTO(queriesDropped, Object *obj, void *dummy)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);
    struct List *queries = QueryEditorQueries(d->d_QueryEditor);

    if(queries)
        MUIListGetList(d->d_Queries, queries);
}

static void HOOK_PROTO(Window, Object *pop, Object *win)
{
    set(win, MUIA_Window_DefaultObject, GuiUserData(pop));
}
static HOOK_STRUCT(Window)

static void HOOK_PROTO(newClick, Object *obj, void *dummy)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);
    struct Query *sq;

    sq = NewQuery(d->d_QueryEditor, NULL);
    if(sq != NULL)
    {
        DoMethod(d->d_QueriesPopup, MUIM_Popstring_Close, FALSE);
        setActiveQuery(obj, sq);
    }
}

static void HOOK_PROTO(duplicateClick, Object *obj, void *dummy)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);
    struct Query *active = GuiListActiveEntry(d->d_Queries), *sq;

    sq = NewQuery(d->d_QueryEditor, active);
    if(sq != NULL)
    {
        MUIListInsertNode(d->d_Queries, sq);
        MUIListSetActiveNode(d->d_Queries, sq);
    }
}

static void HOOK_PROTO(sortClick, Object *obj, void *dummy)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);

    DoMethod(d->d_Queries, MUIM_List_Sort);
    MUIListGetList(d->d_Queries, QueryEditorQueries(d->d_QueryEditor));
}

static void HOOK_PROTO(deleteClick, Object *obj, void *dummy)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);
    struct Query *sq = GuiListActiveEntry(d->d_Queries), *other;

    if(sq == NULL)
        return;

    if(SUCC(SUCC(sq)))
        other = SUCC(sq);
    else if(PRED(PRED(sq)))
        other = PRED(sq);
    else
    {
        /* clear entry, we will not free last element */
        SetQueryName(sq, NULL);
        SetQueryProgram(sq, NULL);

        setActiveQuery(obj, sq);
	DoMethod(d->d_Queries, MUIM_List_Redraw, MUIV_List_Redraw_Active);
        return;
    }

    if(d->d_ActiveQuery == sq)
        setActiveQuery(obj, other);

    /* remove and free query */
    DoMethod(d->d_Queries, MUIM_List_Remove, MUIV_List_Remove_Active);
    DeleteQuery(sq);
}

static void HOOK_PROTO(closeQueryEditor, Object *obj, void *dummy)
{
    struct QueryEditorData *d = pget(obj, MUIA_UserData);
    const char *error;
    long line, col;

    error = saveQuery(obj, &line, &col, FALSE);
    if(error != NULL && Ask(AT_WARN, GCS(MSG_CloseAnswers), 
		GCS(MSG_Query_Close)) == FALSE)
    {
        set(obj, MUIA_Window_Title, error);
        ExtEditorSetCursor(d->d_Editor, line, col);
        GuiBeep();
    }
    else
        CloseQueryEditor(d->d_Project);
}


static long HOOK_PROTO(queryCompare, struct Query *q2, struct Query *q1)
{
    const char *n1 = QueryName(q1), *n2 = QueryName(q2);

    if(n1 == NULL && n2 == NULL)
        return 0;
    else if(n1 == NULL)
        return -1;
    else if(n2 == NULL)
        return 1;
    else
        return CodesetStrCollate(n1, n2, CS_COLLATE_ORDER);
}

static HOOK_STRUCT(queryCompare)

GuiObject OpenQueryGui(struct QueryEditor *se)
{
    struct Project *pr = QueryEditorProject(se);
    Object *app = GuiApplication;
    struct MemPool *memPool = ProjectMemPool(pr);
    struct List *queries = QueryEditorQueries(se);
    struct List *tabs = ProjectTables(pr);
    struct Query *sq;
    struct QueryEditorData *d;
    Object *run, *popbutton, *new, *duplicate, *sort, *delete;
    Object *obj;

    if(app == NULL || queries == NULL)
        return(NULL);

    /*
    ** first check if we have any queries, if not allocate one
    */
    if(SUCC(HEAD(queries)) == NULL)
    {
        sq = NewQuery(se, NULL);
        if(sq == NULL)
            return(NULL);
    }

    d = PoolAllocMemZero(memPool, sizeof(struct QueryEditorData));
    if(d == NULL)
        return(NULL);

    d->d_Project = pr;
    d->d_QueryEditor = se;
    d->d_ActiveQuery = NULL;

    d->d_Queries = ListviewObject,
	MUIA_Listview_DoubleClick, TRUE,
	MUIA_Listview_DragType, MUIV_Listview_DragType_Immediate,
	MUIA_Listview_List, DDListObject,
	    MUIA_Frame, MUIV_Frame_InputList,
            MUIA_List_DragSortable, TRUE,
            MUIA_List_DisplayHook, &popListDisplayHook,
            MUIA_List_CompareHook, &queryCompareHook,
            End,
	End;

    d->d_TableNames = AllocLabels(memPool, 1 + NumNodes(tabs));
    if (d->d_TableNames) {
        const char *none = GCS(MSG_Query_TableNone);
        struct Table *tab;
        ushort i = 0;

        d->d_TableNames[0] = PoolAllocMem(memPool, 2 + strlen(none) + 1);
        if (d->d_TableNames[0]) {
            sprintf(d->d_TableNames[0], "\33i%s", none);
        }
        i++;
        FORNODE(tab, tabs) {
            d->d_TableNames[i++] = AllocString(memPool, TableName(tab));
        }
    }

    d->d_QueryTable = CycleObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_Query_TableBH),
        MUIA_Weight, 30,
        MUIA_Cycle_Entries, d->d_TableNames,
        End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
        MUIA_Window_Width, MUIV_Window_Width_Visible(80),
        MUIA_Window_Height, MUIV_Window_Height_Visible(80),
        MUIA_Window_Title, GCS(MSG_QueryEditor),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
        MUIA_Window_ID, MAKE_ID('_', 'Q', 'R', 'Y'),
        MUIA_HelpNode, "Query editor",
        MUIA_UserData, d,
	WindowContents, VGroup,
            Child, HGroup,
                Child, d->d_QueriesPopup = PopobjectObject,
                    MUIA_Weight, 100,
	            MUIA_Popstring_String, d->d_QueryName = StringObject,
                        MUIA_Frame, MUIV_Frame_String,
                        MUIA_CycleChain, TRUE,
                        MUIA_String_Format, MUIV_String_Format_Center,
			MUIA_ShortHelp, GCS(MSG_Query_NameBH),
                        End,
	            MUIA_Popstring_Button, popbutton = PopButton(MUII_PopUp),
	            MUIA_Popobject_StrObjHook, &StrObjHook,
	            MUIA_Popobject_ObjStrHook, &ObjStrHook,
    	            MUIA_Popobject_WindowHook, &WindowHook,
	            MUIA_Popobject_Object, VGroup,
                        GroupSpacing(0),
                        MUIA_UserData, d->d_Queries,
                        Child, d->d_Queries,
                        Child, HGroup,
                            GroupSpacing(0),
                            MUIA_Group_SameWidth, TRUE,
                            Child, new = SimpleButton(GCS(MSG_New)),
                            Child, duplicate = SimpleButton(GCS(MSG_Duplicate)),
                            Child, sort = SimpleButton(GCS(MSG_Sort)),
                            Child, delete = SimpleButton(GCS(MSG_Delete)),
                            End,
                        End,
                    End,
                Child, d->d_QueryTable,
                Child, HGroup,
                    MUIA_Group_SameWidth, TRUE,
		    Child, run = SimpleButton(GCS(MSG_Query_Run)),
                    Child, d->d_Export = SimpleButton(GCS(MSG_Query_Export)),
#ifdef USE_HPDF
                    Child, d->d_PDF = SimpleButton(GCS(MSG_Query_PDF)),
#endif
                    Child, d->d_Print = SimpleButton(GCS(MSG_Query_Print)),
                    End,
                End,
            Child, VGroup,
                MUIA_VertWeight, 20,
                Child, MUINewEditor(pr, 0, FALSE, TRUE,
                    (char *)MUII_TextBack, NULL, &d->d_Editor),
	        End,
            Child, BalanceObject, End,
	    Child, d->d_SFWList = MListviewObject,
                MUIA_Weight, 200,
                MUIA_CycleChain, TRUE,
		MUIA_ShortHelp, GCS(MSG_Query_OutputBH),
                MUIA_Listview_DragType, MUIV_Listview_DragType_Immediate,
		MUIA_MListview_MList, SFWListObject,
		    MUIA_Frame, MUIV_Frame_InputList,
                    End,
                End,
            End,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct QueryEditorData));
        return(NULL);
    }

    SetAttrs(run,
	MUIA_CycleChain, TRUE,
	MUIA_ShortHelp, GCS(MSG_Query_RunBH),
	TAG_DONE);
    SetAttrs(d->d_Export,
        MUIA_Disabled, TRUE,
        MUIA_CycleChain, TRUE,
	MUIA_ShortHelp, GCS(MSG_Query_ExportBH),
        TAG_DONE);
#ifdef USE_HPDF
    SetAttrs(d->d_PDF,
        MUIA_Disabled, TRUE,
        MUIA_CycleChain, TRUE,
	MUIA_ShortHelp, GCS(MSG_Query_PDFBH),
        TAG_DONE);
#endif
    SetAttrs(d->d_Print,
        MUIA_Disabled, TRUE,
        MUIA_CycleChain, TRUE,
	MUIA_ShortHelp, GCS(MSG_Query_PrintBH),
        TAG_DONE);
    SetAttrs(d->d_Editor,
	MUIA_CycleChain, TRUE,
	MUIA_ShortHelp, GCS(MSG_Query_ProgramBH),
        TAG_DONE);
    set(d->d_Queries, MUIA_UserData, d->d_QueriesPopup);

    MUIAddCallback2(d->d_QueryName, MUIA_String_Acknowledge, MUIV_EveryTime, 
            obj, acknowledge, NULL);

    DoMethod(d->d_Queries, MUIM_Notify, MUIA_Listview_DoubleClick, TRUE,
        d->d_QueriesPopup, 2, MUIM_Popstring_Close, TRUE);
    MUIAddCallback2(d->d_Queries, MUIA_DDList_Dropped, TRUE,
        obj, queriesDropped, NULL);
    DoMethod(d->d_SFWList, MUIM_Notify, MUIA_List_Entries, MUIV_EveryTime,
        d->d_Export, 3, MUIM_Set, MUIA_Disabled, MUIV_NotTriggerValue);
#ifdef USE_HPDF
    DoMethod(d->d_SFWList, MUIM_Notify, MUIA_List_Entries, MUIV_EveryTime,
        d->d_PDF, 3, MUIM_Set, MUIA_Disabled, MUIV_NotTriggerValue);
#endif
    DoMethod(d->d_SFWList, MUIM_Notify, MUIA_List_Entries, MUIV_EveryTime,
        d->d_Print, 3, MUIM_Set, MUIA_Disabled, MUIV_NotTriggerValue);
    MUIAddCallback(d->d_SFWList, MUIA_Listview_DoubleClick, TRUE,
        SFWListDoubleClick, ProjectRootDisplayObject(pr));

    MUIAddCallback2(new, MUIA_Pressed, FALSE, obj, newClick, NULL);
    MUIAddCallback2(duplicate, MUIA_Pressed, FALSE, obj, duplicateClick, NULL);
    MUIAddCallback2(sort, MUIA_Pressed, FALSE, obj, sortClick, NULL);
    MUIAddCallback2(delete, MUIA_Pressed, FALSE, obj, deleteClick, NULL);
    MUIAddCallback2(run, MUIA_Pressed, FALSE, obj, runClick, NULL);
    MUIAddCallback2(d->d_Export, MUIA_Pressed, FALSE, obj, exportClick, NULL);
#ifdef USE_HPDF
    MUIAddCallback2(d->d_PDF, MUIA_Pressed, FALSE, obj, pdfClick, NULL);
#endif
    MUIAddCallback2(d->d_Print, MUIA_Pressed, FALSE, obj, printClick, NULL);

    MUIAddPushCallback(obj, MUIA_Window_CloseRequest, TRUE, closeQueryEditor, NULL);

    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        MUI_AddDisposeObject(app, obj);
        PoolFreeMem(memPool, d, sizeof(struct QueryEditorData));
        return(NULL);
    }

    sq = (struct Query *)HEAD(queries);
    if(SUCC(sq))
        setActiveQuery(obj, sq);

    set(obj, MUIA_Window_ActiveObject, d->d_Editor);

    return(obj);
}

void QueryEditorRemRecord(struct Table *tab, struct Record *rec)
{
    struct Project *pr = TableProject(tab);
    struct QueryEditor *qe = ProjectQueryEditor(pr);
    GuiObject obj = QueryEditorGui(qe);
    struct QueryEditorData *qed = pget(obj, MUIA_UserData);

    if (qed != NULL && qed->d_SFWList != NULL) {
        DoMethod(qed->d_SFWList, MUIM_SFWList_RemRecord, rec);
    }
}

void QueryEditorRemAllRecords(struct Project *pr)
{
    struct QueryEditor *qe = ProjectQueryEditor(pr);
    GuiObject obj = QueryEditorGui(qe);
    struct QueryEditorData *qed = pget(obj, MUIA_UserData);

    if (qed != NULL && qed->d_SFWList != NULL) {
        DoMethod(qed->d_SFWList, MUIM_SFWList_RemAllRecords);
    }
}

void QueryGuiActivate(Object *obj)
{
    if(obj != NULL)
    {
        /* bring window to front and activate it */
        DoMethod(obj, MUIM_Window_ToFront);
        set(obj, MUIA_Window_Activate, TRUE);
    }
}

void QueryGuiRunQuery(Object *obj, struct Query *query)
{
    if (obj != NULL && query != NULL) {
        struct QueryEditorData *d = pget(obj, MUIA_UserData);

        if (d != NULL && d->d_ActiveQuery != query) {
            long line, col;
            const char *error = saveQuery(obj, &line, &col, FALSE);

            if (error) {
                showerror(obj, error, line, col);
                return;
            }
            setActiveQuery(obj, query);
        }
        runClick(obj);
    }
}

