/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** quicksearch_mui.c
**
** Interactive global quick search popup-window
**
** Creation date: 28.09.2006
** Last modified: 21.02.2011
*/

#include "quicksearch_mui.h"
#include "extclasses/ExtWindow_mcc.h"

static GuiObject quickSearchWindow = NULL;
static GuiObject quickSearchString = NULL;
static GuiObject quickSearchGuiObject = NULL;
static QuickSearchCallback quickSearchCallback = NULL;
static void *quickSearchPointer = NULL;

void GuiQuickSearchExit(void)
{
    GuiObject app = GuiApplication;
    GuiObject win = quickSearchWindow;
    if (app && win) {
	GuiCloseWindow(win);
	MUIRemMember(app, win);
	MUI_AddDisposeObject(app, win);
    }
    quickSearchWindow = NULL;
    quickSearchString = NULL;
    quickSearchGuiObject = NULL;
    quickSearchCallback = NULL;
    quickSearchPointer = NULL;
}

static void HOOK_PROTO(quickSearchFinish, GuiObject wi, long *commit)
{
    if (quickSearchCallback) {
	quickSearchCallback(quickSearchGuiObject,
		pget(quickSearchString, MUIA_String_Contents),
		*commit? QUICK_SEARCH_COMMIT : QUICK_SEARCH_CANCEL,
		quickSearchPointer);
    }
    GuiQuickSearchExit();
}

static void HOOK_PROTO(quickSearchChange, GuiObject str, const char **x)
{
    if (quickSearchCallback) {
	quickSearchCallback(quickSearchGuiObject, *x, 
		QUICK_SEARCH_START, quickSearchPointer);
    }
}

static bool quickSearchKeyFunc(GuiObject win, char *buf, short len, 
	ushort qualifier)
{
    int cmd;

    if (len < 2 || buf[0] != '\x9b') {
	return FALSE;
    }
    switch (buf[1]) {
	case '\x41':  cmd = QUICK_SEARCH_PREV; break;
	case '\x42':  cmd = QUICK_SEARCH_NEXT; break;
	default:      return FALSE;
    }
    if (quickSearchCallback) {
	quickSearchCallback(quickSearchGuiObject,
		pget(quickSearchString, MUIA_String_Contents), cmd,
		quickSearchPointer);
    }
    return TRUE;
}

bool GuiQuickSearchInit(GuiObject go, char c, QuickSearchCallback cb, void *ptr)
{
    GuiObject app = GuiApplication, win, str;
    char initstr[2];

    if (app == NULL) {
	return FALSE;
    }
    GuiQuickSearchExit();	/* make sure we have a clean state */
    
    quickSearchGuiObject = go;
    quickSearchCallback = cb;
    quickSearchPointer = ptr;

    if (quickSearchCallback) {
	quickSearchCallback(quickSearchGuiObject, NULL,
		QUICK_SEARCH_SETUP, quickSearchPointer);
    }

    initstr[0] = c;
    initstr[1] = 0;
    quickSearchString = str = StringObject,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_CycleChain, TRUE,
        MUIA_String_Contents, initstr,
	End;
    MUIStringPreselectAll(str, FALSE);

    quickSearchWindow = win = ExtWindowObject,
        MUIA_Window_Activate, TRUE,
        MUIA_Window_Borderless, TRUE,
        MUIA_Window_CloseGadget, FALSE,
        MUIA_Window_DepthGadget, FALSE,
        MUIA_Window_DragBar, FALSE,
        MUIA_Window_SizeGadget, FALSE,
	MUIA_Window_NoMenus, TRUE,
#ifdef USE_TEXTINPUT
        MUIA_Window_Width, 100,
#endif
	MUIA_Window_RootObject, VGroup, 
            MUIA_InnerLeft, 0,
            MUIA_InnerRight, 0,
            MUIA_InnerTop, 0,
            MUIA_InnerBottom, 0,
            Child, str,
            End,
	MUIA_KeyFunc, quickSearchKeyFunc,
	MUIA_HelpNode, "Changing records",
	End;

    MUIAddCallback(win, MUIA_Window_CloseRequest, TRUE, quickSearchFinish, 0);
    MUIAddCallback(win, MUIA_Window_Activate, FALSE, quickSearchFinish, 0);
    MUIAddCallback(str, MUIA_String_Acknowledge, MUIV_EveryTime,
	    quickSearchFinish, 1);
    MUIAddCallback(str, MUIA_String_Contents, MUIV_EveryTime,
	    quickSearchChange, MUIV_TriggerValue);
	
    if (go) {
	GuiObject gowin = GuiObjectWindow(go);
	long w = nget(go, MUIA_Width);
	long x = nget(gowin, MUIA_Window_LeftEdge) + nget(go, MUIA_LeftEdge);
	long y = nget(gowin, MUIA_Window_TopEdge) + 
            nget(go, MUIA_TopEdge) + nget(go, MUIA_Height);

	SetAttrs(win, MUIA_Window_LeftEdge, x + w/2, 
		MUIA_Window_TopEdge, y, TAG_END);
    }
	
    MUIAddMember(app, win);
    if (!GuiOpenWindow(win)) {
	GuiQuickSearchExit();
        return FALSE;
    }
    GuiSetActiveGuiObject(win, str);
    set(str, MUIA_String_BufferPos, 1);

    if (cb) {
	cb(go, initstr, QUICK_SEARCH_START, ptr);
    }
    GuiWindowToFront(win);

    return TRUE;
}

