/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** recorddescriptionspec_mui.c
**
** Creation date: 24.07.2012
** Last modified: 21.08.2012
*/

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "recorddescriptionspec_mui.h"
#include "structureeditor_mui.h"
#include "table.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

struct EditRecordDescriptionSpecData
{
    GuiObject d_Type, d_TypeGroup;
    GuiObject d_Compute, d_ComputeStr;
    GuiObject d_FieldList;
    GuiObject d_MultiColumn;
};

static ulong HOOK_PROTO(fieldListDisplay, char **dst, struct Field *field)
{
    if (field == RDS_FIELD_RECNUM) {
        static char buf[100];
        sprintf(buf, "\33i%.97s", GCS(MSG_RecordDescription_RecordNumber));
        *dst = buf;
    } else {
        *dst = (char *)FieldName(field);
    }
    return 0;
}
static HOOK_STRUCT(fieldListDisplay)

static void HOOK_PROTO(onTypeChanged, Object *go, 
        struct EditRecordDescriptionSpecData **x)
{
    struct EditRecordDescriptionSpecData *d = *x;
    short active = nget(d->d_Type, MUIA_Cycle_Active);

    set(d->d_TypeGroup, MUIA_Group_ActivePage, active);
    set(d->d_MultiColumn, MUIA_Disabled, active == RDS_TYPE_INHERITED);
}

static bool recordDescriptionSpecFind(
        const struct RecordDescriptionSpec *rds, struct Field *field)
{
    int i;

    if (rds) {
        for (i=0; i < rds->rds_NumFields; i++) {
            if (rds->rds_Fields[i] == field) {
                return TRUE;
            }
        }
    }
    return FALSE;
}

GuiObject CreateEditRecordDescriptionSpec(struct Table *tab,
        const struct RecordDescriptionSpec *rds, bool allowInherited)
{
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = ProjectMemPool(pr);
    short active = rds? rds->rds_Type : 0;
    short num = rds? rds->rds_NumFields : 0, i, num_selected = 0;
    struct EditRecordDescriptionSpecData *d;
    GuiObject page0, page1, go;
    static const char *desctext[4];
   
    d = PoolAllocMemZero(memPool, sizeof(struct EditRecordDescriptionSpecData));
    if (d == NULL) {
        return NULL;
    }

    d->d_FieldList = ListviewObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_RecordDescription_FieldsBH),
        MUIA_Listview_DragType, MUIV_Listview_DragType_Immediate,
        MUIA_Listview_MultiSelect, MUIV_Listview_MultiSelect_Default,
        MUIA_Listview_List, ListObject,
            MUIA_Frame, MUIV_Frame_InputList,
            MUIA_List_DragSortable, TRUE,
            MUIA_List_DisplayHook, &fieldListDisplayHook,
            End,
        End;

    if (d->d_FieldList) {
        struct List *fields = TableFields(tab);
        struct Field *field;

        for (i=0; i < num; i++) {
            if (rds->rds_Fields[i] != NULL) {
                DoMethod(d->d_FieldList, MUIM_List_InsertSingle,
                    rds->rds_Fields[i], MUIV_List_Insert_Bottom);
                DoMethod(d->d_FieldList, MUIM_List_Select,
                    (long)num_selected, MUIV_List_Select_On, NULL);
                num_selected++;
            }
        }
        field = RDS_FIELD_RECNUM;
        if (!recordDescriptionSpecFind(rds, field)) {
            DoMethod(d->d_FieldList, MUIM_List_InsertSingle,
                field, MUIV_List_Insert_Bottom);
        }
        if (fields != NULL) {
            FORNODE(field, fields) {
                switch(FieldType(field))
                {
                    case TYPE_STRING:
                    case TYPE_DATE:
                    case TYPE_TIME:
                    case TYPE_INTEGER:
                    case TYPE_REAL:
                    case TYPE_BOOL:
                    case TYPE_CHOICE:
                        if (!recordDescriptionSpecFind(rds, field)) {
                            DoMethod(d->d_FieldList, 
                                    MUIM_List_InsertSingle,
                                    field, MUIV_List_Insert_Bottom);
                        }
                        break;
                }
            }
        }
    }
    page0 = d->d_FieldList;

    d->d_ComputeStr = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_Frame, MUIV_Frame_String,
        MUIA_ControlChar, Shortcut(GCS(MSG_Compute)),
        MUIA_ShortHelp, GCS(MSG_RecordDescription_ComputeBH),
        MUIA_String_MaxLen, FIELD_TRIGGER_LEN + 1,
        MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, rds? rds->rds_Compute : NULL,
        End;

    d->d_Compute = UserFunctionsPopup(d->d_ComputeStr, pr);

    page1 = VGroup,
        Child, CLabel(GCS(MSG_Compute)),
        Child, d->d_Compute,
        End;

    desctext[0] = GCS(MSG_Fields);
    desctext[1] = GCS(MSG_Computed);
    if (allowInherited) {
        desctext[2] = GCS(MSG_RecordDescription_FromTable);
    } else {
        desctext[2] = 0;
    }
    desctext[3] = 0;
    d->d_Type = CycleObject,
        MUIA_CycleChain, TRUE,
        MUIA_ControlChar, Shortcut(GCS(MSG_RecordDescription_Description)),
        MUIA_ShortHelp, GCS(MSG_RecordDescription_DescriptionBH),
        MUIA_Cycle_Entries, desctext,
        MUIA_Cycle_Active, active,
        End;

    d->d_TypeGroup = VGroup,
        MUIA_Group_PageMode, TRUE,
        Child, page0,
        Child, VCenter(page1),
        Child, VSpace(0),
        End;
    
    d->d_MultiColumn = MUI_MakeObject(MUIO_Checkmark, 
            (IPTR)GCS(MSG_RecordDescription_UseMultiColumnList));
    if (d->d_MultiColumn) {
        SetAttrs(d->d_MultiColumn,
            MUIA_CycleChain, TRUE,
            MUIA_Selected, rds? rds->rds_MultiColumn : FALSE,
            MUIA_ShortHelp, GCS(MSG_RecordDescription_UseMultiColumnListBH),
            MUIA_Disabled, rds? (rds->rds_Type == RDS_TYPE_INHERITED) : FALSE,
            TAG_DONE);
    }

    go = VGroup,
        Child, HGroup,
            Child, Label(GCS(MSG_RecordDescription_Description)),
            Child, d->d_Type,
            End,
        Child, d->d_TypeGroup,
        Child, HGroup,
            Child, d->d_MultiColumn,
            Child, LLabel(GCS(MSG_RecordDescription_UseMultiColumnList)),
            Child, HSpace(0),
            End,
        End;

    if (go) {
        set(go, MUIA_UserData, d);
    } else {
        PoolFreeMem(memPool, d, sizeof(struct EditRecordDescriptionSpecData));
    }

    return go;
}

void EditRecordDescriptionSpecAddCallbacks(GuiObject go, GuiObject ok)
{
    struct EditRecordDescriptionSpecData *d = pget(go, MUIA_UserData);

    if (d) {
        if (d->d_ComputeStr && ok) {
            DoMethod(d->d_ComputeStr,
                    MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
                    pget(go, MUIA_WindowObject), 3, MUIM_Set, 
                    MUIA_Window_ActiveObject, ok);
        }
        if (d->d_Type && d->d_TypeGroup) {
            MUIAddCallback(d->d_Type, MUIA_Cycle_Active, MUIV_EveryTime, 
                    onTypeChanged, d);
            set(d->d_TypeGroup, MUIA_Group_ActivePage, 
                    nget(d->d_Type, MUIA_Cycle_Active));
        }
    }
}

void EditRecordDescriptionSpecCleanup(GuiObject go, struct Project *pr)
{
    PoolFreeMem(ProjectMemPool(pr), pget(go, MUIA_UserData),
            sizeof(struct EditRecordDescriptionSpecData));
    set(go, MUIA_UserData, NULL);
}

struct RecordDescriptionSpec *EditRecordDescriptionSpecGetSpec(
        GuiObject go, struct MemPool *memPool)
{
    struct EditRecordDescriptionSpecData *d = pget(go, MUIA_UserData);
    struct RecordDescriptionSpec *rds = NULL;

    if (d == NULL) {
        return NULL;
    }

    if (d->d_FieldList) {
        long num = 0;
        long index = MUIV_List_NextSelected_Start, i;

        DoMethod(d->d_FieldList, MUIM_List_Select, MUIV_List_Select_All,
                MUIV_List_Select_Ask, &num);
        rds = AllocRecordDescriptionSpec(memPool, num);
        if (rds) {
            rds->rds_Type = nget(d->d_Type, MUIA_Cycle_Active);
            rds->rds_Compute = AllocString(memPool, EmptyNullStr(
                        pget(d->d_ComputeStr, MUIA_String_Contents)));
            for (i=0; i < num; i++) {
                DoMethod(d->d_FieldList, 
                        MUIM_List_NextSelected, &index);
                if (index == MUIV_List_NextSelected_End) {
                    break;
                }
                DoMethod(d->d_FieldList, MUIM_List_GetEntry, index,
                        &rds->rds_Fields[i]);
            }
            rds->rds_MultiColumn = nget(d->d_MultiColumn, MUIA_Selected);
        }
    }

    return rds;
}

#endif

