/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** Requesters_mui.c
**
** Creation date: 23.12.1994
** Last modified: 04.06.2022
*/

#include "system.h"
#include "requesters.h"
#include "project.h"
#include "dbtypes.h"
#include "queryeditor.h"
#include "preferences.h"
#include "dosutil.h"
#include "util.h"
#include "labels.h"
#include "memo.h"
#include "memory.h"
#include "appstrings_cat.h"
#include "extclasses/ExtEditor_mcc.h"
#include "extclasses/SFWList_mcc.h"
#include <libraries/asl.h>

static char *getdirfilename(struct Project *pr, struct MemPool *memPool,
        const char *title, const char *oktext,
        const char *pref, long prefitem, bool save_mode, bool isdir)
{
    struct Window *wi = GuiAnyOpenedWindow();
    struct FileRequester *filerequester = NULL;
    char *dir = NULL, *file = NULL, *path = NULL;
    bool free_dirfile = FALSE;
    long result;

    ProjectSetCurrentDir(pr);	    /* make sure we have correct current dir */
    if(pref == NULL && prefitem > 0)
        pref = GetPrefsString(pr, prefitem);
    pref = EmptyNullStr(pref);

    if(isdir || FileStat(pref) == 0)
        dir = (char *)pref;
    else if(SplitPathname((char *)pref, memPool, &dir, &file))
        free_dirfile = TRUE;
    else
        return(NULL);

    filerequester = (struct FileRequester *)
	MUI_AllocAslRequestTags(ASL_FileRequest,
	    wi? ASLFR_Window : TAG_IGNORE, (IPTR)wi,
	    ASLFR_PrivateIDCMP, TRUE,
	    title? ASLFR_TitleText : TAG_IGNORE, (IPTR)title,
	    oktext? ASLFR_PositiveText : TAG_IGNORE, (IPTR)oktext,
	    (!isdir && file)? ASLFR_InitialFile : TAG_IGNORE, (IPTR)file,
	    ASLFR_InitialDrawer, (IPTR)dir,
	    ASLFR_DrawersOnly, isdir,
	    ASLFR_DoSaveMode, save_mode,
	    ASLFR_DoPatterns, TRUE,
	    ASLFR_RejectIcons, TRUE,
	    TAG_END);

    if(filerequester != NULL)
    {
        GuiSetSleep(TRUE);
        UpdateProjects(FALSE);
        result = MUI_AslRequest(filerequester, NULL);
        GuiSetSleep(FALSE);

        if(result)
        {
	    path = JoinPathname(memPool, 
		    (const char *)filerequester->fr_Drawer, 
		    (const char *)filerequester->fr_File);
            if(path != NULL && prefitem > 0)
                SetPrefsString(pr, prefitem, path);
        }

        MUI_FreeAslRequest((APTR)filerequester);
    }

    if(free_dirfile)
    {
        FreeString(memPool, file);
        FreeString(memPool, dir);
    }

    return(path);
}

char *GetFilename(struct Project *pr, struct MemPool *memPool,
        const char *title, const char *oktext,
        const char *pref, long prefitem, bool save_mode)
{
    return(getdirfilename(pr, memPool, title, oktext, pref, prefitem, save_mode, FALSE));
}

char *GetDirname(struct Project *pr, struct MemPool *memPool,
        const char *title, const char *oktext,
        const char *pref, long prefitem, bool save_mode)
{
    return(getdirfilename(pr, memPool, title, oktext, pref, prefitem, save_mode, TRUE));
}

short Ask2A(Object *obj, const char *title, AskType type,
	const char *buttons, const char *text, void *params)
{
    Object *wi = GuiObjectWindow(obj);
    short result;

    GuiSetSleep(TRUE);
    UpdateProjects(FALSE);
    result = MUI_RequestA(GuiApplication, wi, 0,
	(STRPTR)title, (STRPTR)buttons, (STRPTR)text, params);
    GuiSetSleep(FALSE);

    return result;
}

#if (defined __amigaos4__) || (defined __MORPHOS__)
#include <stdarg.h>
short VARARGS68K Ask2(Object *obj, const char *title, AskType type,
	const char *buttons, const char *text, ...)
{
    va_list va;
    short result;

#ifdef __amigaos4__
    va_startlinear(va,text);
    result = Ask2A(obj, title, type, buttons, text, va_getlinearva(va,void*));
#else
    va_start(va, text);
    result=Ask2A(obj, title, type, buttons, text, (void *)va->overflow_arg_area);
#endif
    va_end(va);
    return result;
}
#elif (!defined __AROS__)
short Ask2(Object *obj, const char *title, AskType type,
	const char *buttons, const char *text, ...)
{
    return Ask2A(obj, title, type, buttons, text, ((void **)&text) + 1);
}
#endif

#if (defined __amigaos4__) || (defined __MORPHOS__)
#include <stdarg.h>
short VARARGS68K Ask(AskType type, const char *buttons, const char *text, ...)
{
    va_list va;
    short result;

#if (defined __amigaos4__)
    va_startlinear(va,text);
    result = Ask2A(NULL, NULL, type, buttons, text, va_getlinearva(va,void*));
#else
    va_start(va, text);
    result=Ask2A(NULL, NULL, type, buttons, text,(void *)va->overflow_arg_area);
#endif
    va_end(va);
    return result;
}
#elif (!defined __AROS__)
short Ask(AskType type, const char *buttons, const char *text, ...)
{
    return Ask2A(NULL, NULL, type, buttons, text, ((void **)&text) + 1);
}
#endif

#define ID_CLOSE 51
#define ID_OK 52
#define ID_CANCEL 53
#define MAXLEN 1000

static bool handleInput(Object *app, Object *win, Object *string,
        Object *ok, Object *cancel,
        bool (*acceptfunc)(Object *, Object *, Object *))
{
    ulong state = 0;
    bool running = TRUE, ret = FALSE;

    set(string, MUIA_CycleChain, TRUE);
    set(ok, MUIA_CycleChain, TRUE);
    set(cancel, MUIA_CycleChain, TRUE);

    DoMethod(win, MUIM_Notify, MUIA_Window_CloseRequest, TRUE,
        app, 3, MUIM_WriteLong, 1, &state);
    DoMethod(cancel, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 3, MUIM_WriteLong, 1, &state);
    DoMethod(ok, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 3, MUIM_WriteLong, 2, &state);
    DoMethod(string, MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
        app, 3, MUIM_WriteLong, 2, &state);

    if(GuiOpenWindow(win))
    {
        set(win, MUIA_Window_ActiveObject, string);

        while(running)
        {
	    ulong sigs = 0;
            ulong rc;

            state = 0;
	    rc = DoMethod(app, MUIM_Application_NewInput, &sigs);
            if(rc == MUIV_Application_ReturnID_Quit)
                state = 1;

            switch(state)
            {
                case 1: running = FALSE;
                        break;

                case 2: if(acceptfunc == NULL || acceptfunc(app, win, string))
                        {
                            ret = TRUE;
                            running = FALSE;
                        }
                        else
                            set(win, MUIA_Window_ActiveObject, string);
                        break;
            }

	    if(running && sigs)
	    {
                UpdateProjects(FALSE);
	        sigs = Wait(sigs | SIGBREAKF_CTRL_C);
	        if (sigs & SIGBREAKF_CTRL_C)
                    running = FALSE;
            }
        }

        GuiCloseWindow(win);
    }

    return(ret);
}

static long HOOK_PROTO(fileStop, Object *pop, struct FileRequester *fr)
{
    Object *str = NULL;
    struct Project *pr = NULL;
    struct MemPool *memPool = NULL;
    char *path = NULL, *contents = NULL, *new;
    const char *s;
    cschar c;
    short quotes;

    if (pop == NULL || fr == NULL)
        return 0;

    get(pop, MUIA_Popstring_String, &str);
    if (str == NULL)
        return 0;
    get(str, MUIA_String_Contents, &contents);

    pr = (struct Project *)GuiUserData(str);
    memPool = ProjectMemPool(pr);
    if (memPool == NULL)
        return 0;

    path = JoinPathname(memPool, (char *)fr->fr_Drawer, (char *)fr->fr_File);
    if (path == NULL)
        return 0;

    quotes = (CodesetStrchr(path, ' ') != NULL)? 2 : 0;

    s = contents;
    if (s != NULL)
    {
        while (CodesetIsBrkSpace(c = CodesetGetChar(s))) {
            s = CodesetNextChar(s);
	}
        if (c == '\'' || c == '"') {
            s = CodesetStrchr(CodesetNextChar(s), c);
	} else if (s != NULL) {
            s = CodesetStrchr(s, ' ');
	}
    }
    if (quotes || s != NULL)
    {
        s = EmptyNullStr(s);
        new = PoolAllocMem(memPool, strlen(path) + quotes + strlen(s) + 1);
        if (new != NULL)
        {
            if (quotes)
                sprintf(new, "\"%s\"%s", path, s);
            else
                sprintf(new, "%s%s", path, s);
            FreeString(memPool, path);
            path = new;
        }
    }

    set(str, MUIA_String_Contents, path);
    FreeString(memPool, path);

    return(0);
}

static HOOK_STRUCT(fileStop)

#define FILELEN 100

static long HOOK_PROTO(fileStart, Object * pop, struct TagItem *tagList)
{
    Object * str = NULL;
    struct Project *pr = NULL;
    struct MemPool *memPool = NULL;
    char *path = NULL, *file = NULL, *dir = NULL, *s;
    static char dirbuffer[MAXLEN + 1];
    static char filebuffer[FILELEN + 1];
    static struct TagItem tags[] =
    {
	{ ASLFR_InitialDrawer, (IPTR)dirbuffer },
	{ ASLFR_InitialFile, (IPTR)filebuffer },
	{ ASLFR_RejectIcons, TRUE },
	{ TAG_DONE, 0 }
    };
    long i;
    cschar kill = ' ';

    dirbuffer[0] = filebuffer[0] = 0;

    if (pop == NULL || tagList == NULL)
        return(FALSE);

    get(pop, MUIA_Popstring_String, &str);
    if (str == NULL)
        return FALSE;
    get(str, MUIA_String_Contents, &path);
    if (path == NULL)
        return FALSE;

    pr = (struct Project *)GuiUserData(str);
    memPool = ProjectMemPool(pr);
    if (memPool == NULL)
        return FALSE;

    while (CodesetIsBrkSpace(CodesetGetChar(path))) {
        path = CodesetNextChar(path);
    }
    if (*path == '\'' || *path == '"') {
        kill = CodesetGetChar(path);
	path = CodesetNextChar(path);
    }

    CodesetStrNBytesCopy(dirbuffer, path, MAXLEN);
    s = CodesetStrchr(dirbuffer, kill);
    if (s != NULL)
        *s = 0;

    if (FileStat(dirbuffer) != 0)
    {
	if (!SplitPathname(dirbuffer, memPool, &dir, &file))
	    return FALSE;

	dirbuffer[0] = 0;
	if (dir != NULL)
	    CodesetStrNBytesCopy(dirbuffer, dir, MAXLEN);
	if (file != NULL)
	    CodesetStrNBytesCopy(filebuffer, file, FILELEN);
	FreeString(memPool, dir);
	FreeString(memPool, file);
    }

    for(i=0; ; i++)
        if (tagList[i].ti_Tag == TAG_DONE)
            break;
    tagList[i].ti_Tag = TAG_MORE;
    tagList[i].ti_Data = (IPTR)tags;

    return TRUE;
}
static HOOK_STRUCT(fileStart)

const char *AskString(struct Project *pr, const char *title, const char *oktext,
        const char *contents, ulong maxlen, AskStringType type)
{
    Object *app = GuiApplication;
    Object *win = NULL, *go = NULL, *file = NULL, *string = NULL; 
    Object *ok = NULL, *cancel = NULL;
    char *ret = NULL;
    static char buffer[MAXLEN+1+1];

    if (app == NULL) {
        return NULL;
    }

    if (title == NULL) {
        title = GCS(MSG_EnterString);
    }
    if (oktext == NULL) {
        oktext = GCS(MSG_Ok);
    }
    if (maxlen > MAXLEN) {
        maxlen = MAXLEN;
    }

    go = string = StringObject,
        MUIA_Frame, MUIV_Frame_String,
/*	MUIA_String_Format, MUIV_String_Format_Left, */
	MUIA_String_Contents, contents,
	MUIA_String_MaxLen, maxlen+1,
	MUIA_String_Secret, (type == AST_Secret),
        MUIA_UserData, pr,
        End;

    if (type == AST_FileSelect) {
	file = GuiPopButtonSetCWD(MUII_PopFile, pr);
        go = PopaslObject,
	    MUIA_Popstring_String, string,
	    MUIA_Popstring_Button, file,
            MUIA_Popasl_Type, ASL_FileRequest,
            MUIA_Popasl_StartHook, &fileStartHook,
            MUIA_Popasl_StopHook, &fileStopHook,
            End;
    }

    win = WindowObject,
        MUIA_Window_Title, title,
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'S', 'T', 'R'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_Window_AppWindow, type == AST_FileSelect,
	WindowContents, VGroup,
	    MUIA_Background, MUII_RequesterBack,
	    Child, go,
            Child, HGroup,
		Child, ok = SimpleButton(oktext),
		Child, HSpace(0),
		Child, cancel = SimpleButton(GCS(MSG_Cancel)),
                End,
            End,
	End;

    if (win == NULL) {
        return NULL;           /* no requester */
    }

    MUIAddAppMessageNotification(win, go);

    GuiSetSleep(TRUE);
    MUIAddMember(app, win);
    if (handleInput(app, win, string, ok, cancel, NULL)) {
        get(string, MUIA_String_Contents, &ret);
        strcpy(buffer, ret);
        ret = buffer;
    }
    MUIRemMember(app, win);
    GuiSetSleep(FALSE);
    MUI_AddDisposeObject(app, win);
    return ret;
}

static long minInteger, maxInteger;

static bool acceptInteger(Object * app, Object * win, Object * string)
{
    char *str = NULL;
    long l;

    get(string, MUIA_String_Contents, &str);
    if(str != NULL && StrToInt(str, &l) && l >= minInteger && l <= maxInteger)
        return(TRUE);

    DisplayBeep(NULL);
    set(win, MUIA_Window_ActiveObject, str);
    return(FALSE);
}

bool AskInteger(struct Project *pr, const char *title, const char *oktext,
        const char *contents, long min, long max, long *result)
{
    Object *app = GuiApplication;
    Object *win = NULL, *string = NULL, *ok = NULL, *cancel = NULL;
    char *accept = "-0123456789";
    bool ret = FALSE;
    static char minmax[51];

    if(app == NULL)
        return(FALSE);

    if(title == NULL)
        title = GCS(MSG_EnterInteger);
    if(oktext == NULL)
        oktext = GCS(MSG_Ok);

    if(min >= 0)
        accept++;               /* leave out '-' */

    /*
    ** Create min max string
    */
    sprintf(minmax, "%s: %ld, %s: %ld", GCS(MSG_Min), min, GCS(MSG_Max), max);

    string = StringObject,
        MUIA_Frame, MUIV_Frame_String,
        MUIA_String_Accept, accept,
/*	MUIA_String_Format, MUIV_String_Format_Center, */
	MUIA_String_Contents, contents,
	MUIA_String_MaxLen, 12,
        MUIA_UserData, pr,
        End;

    win = WindowObject,
        MUIA_Window_Title, title,
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'I', 'N', 'T'),
        MUIA_Window_NoMenus, TRUE,
	WindowContents, VGroup,
	    MUIA_Background, MUII_RequesterBack,
	    Child, string,
            Child, HGroup,
                Child, HSpace(0),
                Child, TextObject,
                    MUIA_Frame, MUIV_Frame_Text,
	            MUIA_Background, MUII_TextBack,
	            MUIA_Text_Contents, minmax,
                    End,
                Child, HSpace(0),
                End,
            Child, HGroup,
		Child, ok = SimpleButton(oktext),
		Child, HSpace(0),
		Child, cancel = SimpleButton(GCS(MSG_Cancel)),
                End,
            End,
	End;

    if(win == NULL)
        return(FALSE);           /* no requester */

    minInteger = min;
    maxInteger = max;
    GuiSetSleep(TRUE);
    MUIAddMember(app, win);
    if(handleInput(app, win, string, ok, cancel, acceptInteger))
    {
        ret = TRUE;
        if(result != NULL)
            get(string, MUIA_String_Integer, result);
    }
    MUIRemMember(app, win);
    GuiSetSleep(FALSE);
    MUI_AddDisposeObject(app, win);
    return(ret);
}

static void HOOK_PROTO(updateNumSelected, Object *count, long *x)
{
    set(count, MUIA_Text_Contents, IntToStr(*x));
}

static void HOOK_PROTO(activechange, Object *string, Object **x)
{
    Object *lv = *x, *win = GuiObjectWindow(string);
    struct QueryLine *ql = NULL;
    const char *str = NULL;

    DoMethod(lv, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, &ql);
    if (ql && ql->ql_Entries[0]) {
        str = ql->ql_Entries[0]->qe_String;
    }
    set(string, MUIA_String_Contents, str);
    if (win) {
        set(win, MUIA_Window_ActiveObject, string);
    }
}

bool AskChoice(struct Project *pr, const char *title, const char *oktext,
        struct Element *labels, struct Element *titles, bool *selected, 
	bool multiselect, bool stringmode, long *index, char **contents)
{
    Object *app = GuiApplication;
    Object *win = NULL, *lv = NULL, *count = NULL, *count_group = NULL,
           *string = NULL, *ok = NULL, *cancel = NULL;
    long num = CountListLength(labels);
    ulong state = 0;
    bool running = TRUE, ret = FALSE;
    static char buffer[MAXLEN+1];

    if (titles != NULL && titles->el_Type == TYPE_NIL) {
        titles = NULL;
    }
    if (app == NULL) {
        return FALSE;
    }
    if (multiselect && stringmode) {
        return FALSE;                   /* illegal combination */
    }
    if (num == 0) {
        return FALSE;                   /* nothing to select (stringmode?) */
    }

    if (title == NULL) {
	if (multiselect) {
	    title = GCS(MSG_MultiSelectItems);
	} else if (stringmode) {
	    title = GCS(MSG_EnterString);
	} else {
	    title = GCS(MSG_SelectItem);
	}
    }
    if (oktext == NULL) {
        oktext = GCS(MSG_Ok);
    }

    lv = MListviewObject,
        MUIA_CycleChain, TRUE,
        MUIA_MListview_MList, SFWListObject,
            MUIA_Frame, MUIV_Frame_InputList,
            MUIA_SFWList_ShowTitles, (titles != NULL),
	    MUIA_SFWList_CheckMarks, multiselect,
            End,
        End;

    if (multiselect) {
        count_group = HGroup,
            Child, Label(GCS(MSG_ItemsSelected)),
            Child, count = TextObject,
                TextFrame,
                MUIA_Background, MUII_TextBack,
                MUIA_Text_Contents, "0",
                End,
            End;
    }

    if (stringmode) {
        string = StringObject,
            MUIA_CycleChain, TRUE,
            MUIA_Frame, MUIV_Frame_String,
/*	    MUIA_String_Format, MUIV_String_Format_Left, */
	    MUIA_String_Contents, contents? *contents : NULL,
	    MUIA_String_MaxLen, MAXLEN,
            MUIA_String_AttachedList, lv,
            MUIA_UserData, pr,
            End;
    }

    win = WindowObject,
        MUIA_Window_Title, title,
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'C', 'H', 'C'),
        MUIA_Window_NoMenus, TRUE,
	WindowContents, VGroup,
	    MUIA_Background, MUII_RequesterBack,
	    Child, VGroup,
                Child, lv,
                multiselect? Child : TAG_IGNORE, count_group,
                stringmode? Child : TAG_IGNORE, string,
                End,
            Child, HGroup,
	        MUIA_Group_SameWidth, TRUE,
		Child, ok = SimpleButton(oktext),
		Child, HSpace(0),
		Child, cancel = SimpleButton(GCS(MSG_Cancel)),
                End,
            End,
	End;

    if (win == NULL) {
        return FALSE;
    }

    if (count) {
        MUIAddCallback2(lv, MUIA_SFWList_NumSelected, MUIV_EveryTime,
			count, updateNumSelected, MUIV_TriggerValue);
    }
    DoMethod(lv, MUIM_SFWList_SetTitlesBody, titles, labels);
    if (multiselect) {
        DoMethod(lv, MUIM_SFWList_SetSelected, selected);
    } else if (!stringmode && index != NULL && *index >= 0 && *index < num) {
        set(lv, MUIA_SFWList_ActiveLine, index + 1);
    }

    if (string) {
        MUIAddCallback2(lv, MUIA_List_Active, MUIV_EveryTime,
            string, activechange, lv);
    }

    set(ok, MUIA_CycleChain, TRUE);
    set(cancel, MUIA_CycleChain, TRUE);
    DoMethod(win, MUIM_Notify, MUIA_Window_CloseRequest, TRUE,
        app, 3, MUIM_WriteLong, 1, &state);
    DoMethod(cancel, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 3, MUIM_WriteLong, 1, &state);
    DoMethod(ok, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 3, MUIM_WriteLong, 2, &state);
    if (string) {
        DoMethod(string, MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
            app, 3, MUIM_WriteLong, 2, &state);
    }
    if (!multiselect) {
        DoMethod(lv, MUIM_Notify, MUIA_Listview_DoubleClick, TRUE,
            app, 3, MUIM_WriteLong, 2, &state);
    }

    GuiSetSleep(TRUE);
    MUIAddMember(app, win);
    if (GuiOpenWindow(win)) {
        set(win, MUIA_Window_ActiveObject, (string? string : lv));

        while (running) {
	    ulong sigs = 0;
            ulong rc;

            state = 0;
	    rc = DoMethod(app, MUIM_Application_NewInput, &sigs);
            if (rc == MUIV_Application_ReturnID_Quit) {
                state = 1;
	    }

            switch (state) 
	    {
                case 1: running = FALSE;
                        break;

                case 2: if (!multiselect && !stringmode &&
                                nget(lv, MUIA_SFWList_ActiveLine) == 0) {
                            GuiBeep();
                        } else {
                            ret = TRUE;
                            running = FALSE;
                        }
                        break;
            }

	    if (running && sigs) {
                UpdateProjects(FALSE);
	        sigs = Wait(sigs | SIGBREAKF_CTRL_C);
	        if (sigs & SIGBREAKF_CTRL_C)
                    running = FALSE;
            }
        }

        GuiCloseWindow(win);
    }
    if (!ret) {
    } else if (multiselect) {
        DoMethod(lv, MUIM_SFWList_GetSelected, selected);
    } else if (stringmode && contents != NULL) {
        char *s = pget(string, MUIA_String_Contents);

        strcpy(buffer, EmptyNullStr(s));
        *contents = buffer;
    } else if (!stringmode && index != NULL) {
        *index = nget(lv, MUIA_SFWList_ActiveLine) - 1;
    }

    MUIRemMember(app, win);
    GuiSetSleep(FALSE);
    MUI_AddDisposeObject(app, win);
    return ret;
}

bool AskMulti(struct Project *pr, const char *title, const char *oktext,
	struct AskMultiInfo *ami)
{
    Object *app = GuiApplication;
    Object *win = NULL, *group1 = NULL, *group2 = NULL,
           *ok = NULL, *cancel = NULL, *first = NULL;
    short i;
    struct AskMultiItem *a;
    ulong state = 0;
    bool running = TRUE, ret = FALSE;
    struct MemPool *memPool = NULL;

    if(app == NULL)
        return(FALSE);
    if(ami == NULL || ami->ami_Num == 0)
        return(FALSE);
    memPool = ami->ami_MemPool;

    if(title == NULL)
        title = GCS(MSG_EditForm);
    if(oktext == NULL)
        oktext = GCS(MSG_Ok);

    group1 = VGroup,
        MUIA_Frame, MUIV_Frame_Group,
        MUIA_Background, MUII_GroupBack,
        End;
    if(group1 == NULL)
        return(FALSE);

    a = ami->ami_Items;
    for(i=0; i < ami->ami_Num; i++, a++)
    {
        char sc = Shortcut(a->ami_Title);
        Object *go = NULL, *title = NULL;
        const char *bh = a->ami_BubbleHelp;
        bool vcenter = FALSE;

        if(bh != NULL && *bh == 0)
            bh = NULL;

        switch(a->ami_Type)
        {
            case TYPE_STRING:
                a->ami_GuiObject = go = StringObject,
                    MUIA_CycleChain, TRUE,
	            MUIA_ControlChar, sc,
                    MUIA_Frame, MUIV_Frame_String,
                    MUIA_ShortHelp, bh,
                    /*	MUIA_String_Format, MUIV_String_Format_Left, */
	            MUIA_String_Contents, a->ami_Contents.string.string,
		    MUIA_String_Secret, a->ami_Contents.string.secret,
                    End;
                break;

            case TYPE_MEMO:
                go = MUINewEditor(pr, sc, FALSE, FALSE,
                    (char *)MUII_TextBack, NULL, &a->ami_GuiObject);
                if (a->ami_GuiObject && a->ami_Contents.mi) {
                    set(a->ami_GuiObject, MUIA_ExtEditor_Contents,
                            a->ami_Contents.mi);
                }
                if (go != NULL) {
                    set(go, MUIA_ShortHelp, bh);
                }
                break;

            case TYPE_INTEGER:
                a->ami_GuiObject = go = StringObject,
                    MUIA_CycleChain, TRUE,
	            MUIA_ControlChar, sc,
                    MUIA_Frame, MUIV_Frame_String,
                    /*	MUIA_String_Format, MUIV_String_Format_Left, */
                    MUIA_ShortHelp, bh,
	            MUIA_String_Contents, a->ami_Nil? "" : IntToStr(a->ami_Contents.integer),
                    End;
                break;

            case TYPE_REAL:
                a->ami_GuiObject = go = StringObject,
                    MUIA_CycleChain, TRUE,
	            MUIA_ControlChar, sc,
                    MUIA_Frame, MUIV_Frame_String,
                    /*	MUIA_String_Format, MUIV_String_Format_Left, */
                    MUIA_ShortHelp, bh,
	            MUIA_String_Contents, a->ami_Nil? "" :
                        DoubleToStr(a->ami_Contents.real, -1,
                            GetPrefsValue(pr, PREFS_REALFORMAT)),
                    End;
                break;

            case TYPE_DATE:
                a->ami_GuiObject = go = StringObject,
                    MUIA_CycleChain, TRUE,
	            MUIA_ControlChar, sc,
                    MUIA_Frame, MUIV_Frame_String,
                    /*	MUIA_String_Format, MUIV_String_Format_Left, */
                    MUIA_ShortHelp, bh,
	            MUIA_String_Contents, a->ami_Nil? "" :
                        DateToString(a->ami_Contents.date,
                            GetPrefsValue(pr, PREFS_DATEFORMAT)),
                    End;
                break;

            case TYPE_TIME:
                a->ami_GuiObject = go = StringObject,
                    MUIA_CycleChain, TRUE,
	            MUIA_ControlChar, sc,
                    MUIA_Frame, MUIV_Frame_String,
                    /*	MUIA_String_Format, MUIV_String_Format_Left, */
                    MUIA_ShortHelp, bh,
	            MUIA_String_Contents, a->ami_Nil? "" : 
			TimeToString(a->ami_Contents.time, FORMAT_HOURMINSEC),
                    End;
                break;

            case TYPE_BOOL:
                go = HGroup,
                    Child, a->ami_GuiObject = ImageObject,
		        ImageButtonFrame,
                        MUIA_CycleChain, TRUE,
		        MUIA_InputMode, MUIV_InputMode_Toggle,
		        MUIA_Image_Spec, MUII_CheckMark,
		        MUIA_Selected, !a->ami_Nil,
                        MUIA_Background, MUII_ButtonBack,
		        MUIA_ShowSelState, FALSE,
		        MUIA_ControlChar, sc,
                        MUIA_ShortHelp, bh,
                        End,
                    Child, HSpace(0),
                    End;
                break;

            case TYPE_CHOICE:
                if(!a->ami_Contents.choice.ischoicelist)
                {
                    a->ami_GuiObject = go = CycleObject,
                        MUIA_CycleChain, TRUE,
                        MUIA_Background, MUII_ButtonBack,
	                MUIA_ControlChar, sc,
                        MUIA_ShortHelp, bh,
	                MUIA_Cycle_Entries, a->ami_Contents.choice.labels,
                        a->ami_Nil? TAG_IGNORE : MUIA_Cycle_Active, a->ami_Contents.choice.active,
                        End;
                }
                else
                {
                    a->ami_GuiObject = go = ListviewObject,
                        MUIA_CycleChain, TRUE,
	                MUIA_ControlChar, sc,
                        MUIA_ShortHelp, bh,
	                MUIA_Listview_List, ListObject,
	                    MUIA_Frame, MUIV_Frame_InputList,
                            MUIA_List_SourceArray, a->ami_Contents.choice.labels,
                            a->ami_Nil? TAG_IGNORE : MUIA_List_Active, a->ami_Contents.choice.active,
	                    End,
                        End;
                    vcenter = TRUE;
                }
                break;

            case TYPE_OPTIONS:
                a->ami_GuiObject = go = ListviewObject,
                    MUIA_CycleChain, TRUE,
	            MUIA_ControlChar, sc,
                    MUIA_ShortHelp, bh,
                    MUIA_Listview_MultiSelect, MUIV_Listview_MultiSelect_Default,
	            MUIA_Listview_List, ListObject,
	                MUIA_Frame, MUIV_Frame_InputList,
                        MUIA_List_SourceArray, a->ami_Contents.options.labels,
	                End,
                    End;

                if(go != NULL)
                {
                    long i, num = NumLabels(a->ami_Contents.options.labels);

                    for(i=0; i < num; i++)
	                DoMethod(go, MUIM_List_Select, i,
                            a->ami_Contents.options.selected[i]?
                                MUIV_List_Select_On : MUIV_List_Select_Off,
                            NULL);
                }
                vcenter = TRUE;
                break;

            case NOTYPE:                /* title only objects */
                go = TextObject,
                    TextFrame,
	            MUIA_Text_Contents, EmptyNullStr(a->ami_Title),
	            MUIA_Text_PreParse, "\33c",
                    End;
                break;
        }

        if(go == NULL)
        {
            MUI_DisposeObject(group1);
            return(FALSE);
        }
        if(first == NULL)
            first = a->ami_GuiObject;

        if(a->ami_Title != NULL && a->ami_Type != NOTYPE)
        {
            if(group2 == NULL)
            {
                group2 = VGroup,
                    MUIA_Group_Columns, 2,
                    End;
                if(group2 == NULL)
                {
                    MUI_DisposeObject(group1);
                    return(FALSE);
                }
                 MUIAddMember(group1, group2);
            }
            title = Label(a->ami_Title);
            if(vcenter)
                title = VCenter(title);
            if(title == NULL)
            {
                MUI_DisposeObject(group1);
                return(FALSE);
            }

            MUIAddMember(group2, title);
            MUIAddMember(group2, go);
        }
        else
        {
            MUIAddMember(group1, go);
            group2 = NULL;
        }
    }

    win = WindowObject,
        MUIA_Window_Title, title,
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'M', 'L', 'T'),
        MUIA_Window_NoMenus, TRUE,
	WindowContents, VGroup,
	    MUIA_Background, MUII_RequesterBack,
	    Child, group1,
            Child, HGroup,
		Child, ok = SimpleButton(oktext),
		Child, HSpace(0),
		Child, cancel = SimpleButton(GCS(MSG_Cancel)),
                End,
            End,
	End;

    if(win == NULL)
        return(FALSE);

    a = ami->ami_Items;
    for(i=0; i < ami->ami_Num; i++, a++)
    {
        switch(a->ami_Type)
        {
            case TYPE_STRING:
            case TYPE_INTEGER:
            case TYPE_REAL:
            case TYPE_DATE:
            case TYPE_TIME:
                DoMethod(a->ami_GuiObject, MUIM_Notify,
                    MUIA_String_Acknowledge, MUIV_EveryTime,
	            win, 3, MUIM_Set, MUIA_Window_ActiveObject, ok);
                break;
        }
    }

    set(ok, MUIA_CycleChain, TRUE);
    set(cancel, MUIA_CycleChain, TRUE);

    DoMethod(win, MUIM_Notify, MUIA_Window_CloseRequest, TRUE,
        app, 3, MUIM_WriteLong, 1, &state);
    DoMethod(cancel, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 3, MUIM_WriteLong, 1, &state);
    DoMethod(ok, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 3, MUIM_WriteLong, 2, &state);

    GuiSetSleep(TRUE);
    MUIAddMember(app, win);
    if(GuiOpenWindow(win))
    {
        set(win, MUIA_Window_ActiveObject, first);

        while(running)
        {
	    ulong sigs = 0;
            ulong rc;

            state = 0;
	    rc = DoMethod(app, MUIM_Application_NewInput, &sigs);
            if(rc == MUIV_Application_ReturnID_Quit)
                state = 1;

            switch(state)
            {
                case 1: running = FALSE;
                        break;

                case 2: if(CheckAskMultiInfo(ami, win))
                        {
                            ret = TRUE;
                            running = FALSE;
                        }
                        break;
            }

	    if(running && sigs)
	    {
                UpdateProjects(FALSE);
	        sigs = Wait(sigs | SIGBREAKF_CTRL_C);
	        if (sigs & SIGBREAKF_CTRL_C)
                    running = FALSE;
            }
        }

        GuiCloseWindow(win);

        a = ami->ami_Items;
        for(i=0; i < ami->ami_Num && ret; i++, a++)
        {
            const char *text = pget(a->ami_GuiObject, MUIA_String_Contents);

            switch(a->ami_Type)
            {
                case TYPE_STRING:
                    FreeString(memPool, a->ami_Contents.string.string);
                    a->ami_Nil = FALSE;
                    a->ami_Contents.string.string = AllocString(memPool, text);
                    if (a->ami_Contents.string.string == NULL) {
                        ret = FALSE;
                    }
                    break;

                case TYPE_MEMO:
                    FreeString(memPool, a->ami_Contents.mi);
                    a->ami_Nil = FALSE;
                    a->ami_Contents.mi =
                        ExtEditorGetContents(a->ami_GuiObject, memPool);
                    if (a->ami_Contents.mi == NULL) {
                        ret = FALSE;
                    }
                    break;

                case TYPE_INTEGER:
                    a->ami_Nil = !StrToInt(text, &a->ami_Contents.integer);
                    break;

                case TYPE_REAL:
                    a->ami_Nil = !StrToDouble(text, &a->ami_Contents.real);
                    break;

                case TYPE_DATE:
                    a->ami_Nil = !StringToDate(text, &a->ami_Contents.date);
                    break;

                case TYPE_TIME:
                    a->ami_Nil = !StringToTime(text, FORMAT_HOURMINSEC, &a->ami_Contents.time);
                    break;

                case TYPE_BOOL:
                    a->ami_Nil = !nget(a->ami_GuiObject, MUIA_Selected);
                    break;

                case TYPE_CHOICE:
                    if(!a->ami_Contents.choice.ischoicelist)
                    {
                        a->ami_Nil = FALSE;
                        a->ami_Contents.choice.active = nget(a->ami_GuiObject, MUIA_Cycle_Active);
                    }
                    else
                    {
                        a->ami_Contents.choice.active = nget(a->ami_GuiObject, MUIA_List_Active);
                        a->ami_Nil = (a->ami_Contents.choice.active == MUIV_List_Active_Off)? TRUE : FALSE;
                    }
                    break;

                case TYPE_OPTIONS:
                {
                    long index = MUIV_List_NextSelected_Start;
                    long num = NumLabels(a->ami_Contents.options.labels);
                    bool *selected = a->ami_Contents.options.selected;

                    a->ami_Nil = FALSE;
                    memset(selected, 0, num * sizeof(bool));

	            for (;;)
	            {
                        DoMethod(a->ami_GuiObject,
                            MUIM_List_NextSelected, &index);
	                if(index == MUIV_List_NextSelected_End)
                            break;
                        selected[index] = TRUE;
	            }
                }
            }
        }
    }
    MUIRemMember(app, win);
    GuiSetSleep(FALSE);
    MUI_AddDisposeObject(app, win);
    return(ret);
}

#define DELAY 5

void Message(const char *text)
{
    const char *title = GCS(MSG_BeeBase);
    Object *app = GuiApplication;
    Object *win = NULL, *gauge = NULL, *gauge_group, *msg = NULL;
    long i, max;

    if(app == NULL || text == NULL)
        return;

    max = (10 + strlen(text) / 5) * 50 / DELAY;

    gauge_group = VGroup,
        Child, gauge = GaugeObject,
	    GaugeFrame,
	    MUIA_Background, MUII_BACKGROUND,
            MUIA_Gauge_Current, 0,
            MUIA_Gauge_Horiz, TRUE,
            MUIA_Gauge_Max, max,
	    MUIA_Gauge_InfoText, "",
            End,
       Child, ScaleObject, End,
       End;

    msg = TextObject,
        MUIA_Text_Contents, text,
        MUIA_Text_PreParse, "\33c",
        End;

    win = WindowObject,
        MUIA_Window_Title, title,
        MUIA_Window_ScreenTitle, GuiScreenTitle,
        MUIA_Window_NoMenus, TRUE,
        MUIA_Window_CloseGadget, FALSE,
        MUIA_Window_SizeGadget, FALSE,
	WindowContents, VGroup,
	    MUIA_Background, MUII_RequesterBack,
	    Child, VGroup,
                TextFrame,
	        MUIA_Background, MUII_TextBack,
                Child, VGroup,
                    Child, VSpace(5),
                    Child, HGroup,
                        Child, HSpace(10),
                        Child, msg,
                        Child, HSpace(10),
                        End,
                    Child, VSpace(5),
                    End,
                End,
            Child, gauge_group,
            End,
	End;

    if(win == NULL)
        return;                 /* no requester */

    GuiSetSleep(TRUE);
    MUIAddMember(app, win);
    UpdateProjects(FALSE);
    if(GuiOpenWindow(win))
    {
        for(i=0; i < max; i++)
        {
           set(gauge, MUIA_Gauge_Current, i);
	   DoMethod(app, MUIM_Application_InputBuffered);
           Delay(DELAY);
        }

        GuiCloseWindow(win);
    }
    MUIRemMember(app, win);
    GuiSetSleep(FALSE);
    MUI_AddDisposeObject(app, win);
}

