/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** rexx_mui.c
**
** Creation date: 07.01.2001
** Last modified: 26.09.2014
*/

#include "main.h"
#include "project.h"
#include "preferences.h"
#include "programeditor.h"
#include "lang/program.h"
#include "server.h"
#include "dosutil.h"
#include "codeset.h"
#include "memory.h"
#include "error.h"
#include "rexx_mui.h"
#include "extclasses/ExtApplication_mcc.h"
#include <proto/rexxsyslib.h>
#ifdef CHECK_RECORD_HEADERS
#include "record.h"
#endif

struct UserData
{
    GuiObject app;
    struct RexxMsg *msg;
    struct MsgPort *replyPort;
};

static struct UserData *allocUserData(GuiObject app, struct RexxMsg *msg)
{
    struct UserData *ud;

#ifdef CHECK_RECORD_HEADERS
    CheckRecordHeaders("ARexx message");
#endif
    ud = malloc(sizeof(struct UserData));
    if(ud != NULL)
    {
	ud->app = app;
	ud->msg = msg;
	ud->replyPort = msg->rm_Node.mn_ReplyPort;
    }

    return ud;
}

static void freeUserData(struct UserData *ud)
{
    if(ud != NULL)
    {
	free(ud);
    }
}
	
static long rexxTranslateError(long error)
{
    static const struct
    {
	long error, rc;
    } table[] =
    {
	{ OK, 0 },
	{ ERR_NOTIMPLEMENTED, MUI_RXERR_BADDEFINITION },
	{ ERR_MEM, MUI_RXERR_OUTOFMEMORY  },
	{ ERR_INTERNAL, -11 },
	{ PROGERR_COMPILE, -12 }
    };
    const int size = sizeof(table) / sizeof(table[0]);
    int i;

    for(i=0; i < size; i++)
    {
	if(table[i].error == error)
	{
	    return table[i].rc;
	}
    }

    return -10;           /* general error return code */
}

static void rexxSetResults(struct RexxMsg *msg, long error, char *result)
{
    if (msg != NULL) {
	if (error == OK) {
	    if ((msg->rm_Action & (1L << RXFB_RESULT)) && result != NULL) {
		long len = strlen(result);
		static const long maxlen = 0xffff; /* ARexx max string length */
		
		if (len <= maxlen) {
#if 0	/* we don't know anything about this, so let's not touch it */
	            if(msg->rm_Result2 != 0) {
		        DeleteArgstring((UBYTE *)msg->rm_Result2);
		        msg->rm_Result2 = 0;
    		    }
#endif
		    msg->rm_Result2 = (IPTR)CreateArgstring((STRPTR)result,len);
		} else {
		    error = AREXXERR_RESULTTOOLONG;
		    result = GES(error);
		}
	    }
	}

	msg->rm_Result1 = rexxTranslateError(error);
	if (error != OK && result != NULL) {
#if defined(__amigaos4__)
	    SetRexxVarFromMsg("LASTERROR", result, msg);
#elif (!defined __amithlon__)
	    SetRexxVar(msg, (CONST_STRPTR)"LASTERROR", result, strlen(result));
#else
	    fprintf(stderr, "AREXX::LASTERROR = %s\n", result);
#endif
	}
    }
#ifdef CHECK_RECORD_HEADERS
    CheckRecordHeaders("ARexx result");
#endif
}

static void rexxReplyMsg(long error, char *result, void *x)
{
    struct UserData *ud = x;

    if(ud != NULL)
    {
	if(DoMethod(ud->app, MUIM_ExtApplication_GetMessage, ud->msg))
	{
	    ud->msg->rm_Node.mn_ReplyPort = ud->replyPort;
	    rexxSetResults(ud->msg, error, result);
	    ReplyMsg((struct Message *)ud->msg);
	}
	free(ud);
    }
}

static long HOOK_PROTO(rexxConnect, GuiObject app, ulong *args)
{
    char *name = (char *)args[0];
    bool gui = (bool)args[1];
    struct RexxMsg *msg = pget(app, MUIA_Application_RexxMsg);
    struct UserData *ud;
    long error;

    if(msg == NULL)
    {
	return MUI_RXERR_BADDEFINITION;
    }

    if(name == NULL)
    {
	return MUI_RXERR_BADSYNTAX;
    }

    ud = allocUserData(app, msg);
    if(ud == NULL)
    {
	return MUI_RXERR_OUTOFMEMORY;
    }

    error = ServerAddConnectRequest(name, gui, ud, rexxReplyMsg);
    if(error != OK)
    {	
	rexxSetResults(msg, error, GES(error));
	freeUserData(ud);
	return rexxTranslateError(error);
    }

    DoMethod(app, MUIM_ExtApplication_RedirectMessage, msg);

    return 0;
}

static HOOK_STRUCT(rexxConnect)

static long HOOK_PROTO(rexxDisconnect, GuiObject app, ulong *args)
{
    long handle = *(long *)args[0];
    struct RexxMsg *msg = pget(app, MUIA_Application_RexxMsg);
    struct UserData *ud;
    long error;

    if(msg == NULL)
    {
	return MUI_RXERR_BADDEFINITION;
    }

    ud = allocUserData(app, msg);
    if(ud == NULL)
    {
	return MUI_RXERR_OUTOFMEMORY;
    }

    error = ServerAddDisconnectRequest(handle, ud, rexxReplyMsg);
    if(error != OK)
    {	
	rexxSetResults(msg, error, GES(error));
	freeUserData(ud);
	return rexxTranslateError(error);
    }

    DoMethod(app, MUIM_ExtApplication_RedirectMessage, msg);

    return 0;
}

static HOOK_STRUCT(rexxDisconnect)

static long HOOK_PROTO(rexxConnections, GuiObject app, void *dummy)
				       
{
    struct RexxMsg *msg = pget(app, MUIA_Application_RexxMsg);
    struct UserData *ud;
    long error;

    if(msg == NULL)
    {
	return MUI_RXERR_BADDEFINITION;
    }

    ud = allocUserData(app, msg);
    if(ud == NULL)
    {
	return MUI_RXERR_OUTOFMEMORY;
    }

    error = ServerAddConnectionsRequest(ud, rexxReplyMsg);
    if(error != OK)
    {	
	rexxSetResults(msg, error, GES(error));
	freeUserData(ud);
	return rexxTranslateError(error);
    }

    DoMethod(app, MUIM_ExtApplication_RedirectMessage, msg);

    return 0;
}

static HOOK_STRUCT(rexxConnections)


static long HOOK_PROTO(rexxEval, GuiObject app, ulong *args)
{
    long handle = *(long *)args[0];
    char *cmd = NULL;
    struct RexxMsg *msg = pget(app, MUIA_Application_RexxMsg);
    struct UserData *ud;
    long error;
    int i;

    if(msg == NULL)
    {
	return MUI_RXERR_BADDEFINITION;
    }

    cmd = (char *)msg->rm_Args[0];            /* use the raw argument */
    if(cmd == NULL)
    {
	return MUI_RXERR_BADSYNTAX;
    }

    /*
    ** Go to 2nd argument
    */
    while (*cmd != 0 && CodesetIsSpace(CodesetGetChar(cmd))) {
	cmd = CodesetNextChar(cmd);
    }
    for(i=0; i < 2; i++)
    {
	while (*cmd != 0 && !CodesetIsSpace(CodesetGetChar(cmd))) {
	    cmd = CodesetNextChar(cmd);
	}
	while (*cmd != 0 && CodesetIsSpace(CodesetGetChar(cmd))) {
	    cmd = CodesetNextChar(cmd);
	}
    }
    if(*cmd == 0)
    {
	return MUI_RXERR_BADSYNTAX;
    }

    ud = allocUserData(app, msg);
    if(ud == NULL)
    {
	return MUI_RXERR_OUTOFMEMORY;
    }

    error = ServerAddEvalRequest(handle, cmd, ud, rexxReplyMsg);
    if(error != OK)
    {	
	rexxSetResults(msg, error, GES(error));
	freeUserData(ud);
	return rexxTranslateError(error);
    }

    DoMethod(app, MUIM_ExtApplication_RedirectMessage, msg);

    return 0;
}

static HOOK_STRUCT(rexxEval)

static long HOOK_PROTO(rexxTransaction, GuiObject app, ulong *args)
{
    long handle = *(long *)args[0];
    struct RexxMsg *msg = pget(app, MUIA_Application_RexxMsg);
    struct UserData *ud;
    long error;

    if(msg == NULL)
    {
	return MUI_RXERR_BADDEFINITION;
    }

    ud = allocUserData(app, msg);
    if(ud == NULL)
    {
	return MUI_RXERR_OUTOFMEMORY;
    }

    error = ServerAddTransactionRequest(handle, ud, rexxReplyMsg);
    if(error != OK)
    {	
	rexxSetResults(msg, error, GES(error));
	freeUserData(ud);
	return rexxTranslateError(error);
    }

    DoMethod(app, MUIM_ExtApplication_RedirectMessage, msg);

    return 0;
}

static HOOK_STRUCT(rexxTransaction)

static long HOOK_PROTO(rexxCommit, GuiObject app, ulong *args)
{
    long handle = *(long *)args[0];
    struct RexxMsg *msg = pget(app, MUIA_Application_RexxMsg);
    struct UserData *ud;
    long error;

    if(msg == NULL)
    {
	return MUI_RXERR_BADDEFINITION;
    }

    ud = allocUserData(app, msg);
    if(ud == NULL)
    {
	return MUI_RXERR_OUTOFMEMORY;
    }

    error = ServerAddCommitRequest(handle, ud, rexxReplyMsg);
    if(error != OK)
    {	
	rexxSetResults(msg, error, GES(error));
	freeUserData(ud);
	return rexxTranslateError(error);
    }

    DoMethod(app, MUIM_ExtApplication_RedirectMessage, msg);

    return 0;
}

static HOOK_STRUCT(rexxCommit)

static long HOOK_PROTO(rexxRollback, GuiObject app, ulong *args)
{
    long handle = *(long *)args[0];
    struct RexxMsg *msg = pget(app, MUIA_Application_RexxMsg);
    struct UserData *ud;
    long error;

    if(msg == NULL)
    {
	return MUI_RXERR_BADDEFINITION;
    }

    ud = allocUserData(app, msg);
    if(ud == NULL)
    {
	return MUI_RXERR_OUTOFMEMORY;
    }

    error = ServerAddRollbackRequest(handle, ud, rexxReplyMsg);
    if(error != OK)
    {	
	rexxSetResults(msg, error, GES(error));
	freeUserData(ud);
	return rexxTranslateError(error);
    }

    DoMethod(app, MUIM_ExtApplication_RedirectMessage, msg);

    return 0;
}

static HOOK_STRUCT(rexxRollback)

static long HOOK_PROTO(rexxCompile, GuiObject app, ulong *args)
{
    struct RexxMsg *msg = pget(app, MUIA_Application_RexxMsg);
    const char *sourcename = (const char *)args[0];
    bool update = (bool)args[1];
    bool external = TRUE;
    struct Project *pr;
    GuiObject pe;
    struct MemPool *memPool;
    struct Program *compiled = NULL;
    long line = 0, col = 0;
    const char *filename = NULL;
    const char *errstr;
    char *mem = NULL;
    long memSize;
    long error = OK;

    if (msg == NULL)
    {
	return MUI_RXERR_BADDEFINITION;
    }

    if (GuiSleep() || ProgramRunning())
    {
	error = AREXXERR_BUSY;
	rexxSetResults(msg, error, GES(error));
	return rexxTranslateError(error);
    }

    FORNODE (pr, &ProjectList)
    {
	const char *s =GetPrefsString(pr, PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);

        ProjectSetCurrentDir(pr);
	if (GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL) &&
	       	SameFile(sourcename, s))
	{
	    sourcename = s;
	    break;
	}
    }
    
    if (SUCC(pr) == NULL)
    {
	error = AREXXERR_NOPROJECT;
	rexxSetResults(msg, error, GES(error));
	return rexxTranslateError(error);
    }

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
	rexxSetResults(msg, error, GES(error));
	return rexxTranslateError(error);
    }

    memPool = ProjectMemPool(pr);
    pe = ProjectProgramEditor(pr);

    if (pe) 
    {
	errstr = ProgramEditorCompile(pe, TRUE, 
		&compiled, &line, &col, &filename);
    } 
    else 
    {
	errstr = ProgramCompileProjectProgram(pr, FALSE,
		&compiled, &line, &col, &filename);
    }

    if (errstr == NULL) 
    {
	SetProjectProgram(pr, compiled);
	ProjectChange(pr, TRUE);

	/* GuiSetSleep(TRUE); */
	errstr = UpdateDependentDisplayObjects(pr, pr, NULL, NULL,
                &line, &col, &filename);
        if (errstr == EXECUTE_ABORT) {
            errstr = NULL;
        }
	/* GuiSetSleep(FALSE); */

	if (errstr == NULL)
	{
	    rexxSetResults(msg, OK, NULL);
	    if (update && external)
		ProgramCreateSourceExternal(pr);
	    return 0;
	}
    }
    if (pe && filename == NULL)
	ProgramEditorShowError(pr, FALSE, errstr, line, col, filename);

    if (filename == NULL && external)
	filename = (char *)sourcename;
    if (filename == NULL)
	filename = "(internal)";

    memSize = strlen(filename) + 100 + strlen(errstr);
    mem = PoolAllocMem(memPool, memSize);

    if (mem == NULL) 
    {
	error = ERR_MEM;
	rexxSetResults(msg, error, GES(error));
    } 
    else 
    {
	sprintf(mem, "%s\n%ld\n%ld\n%s", filename, line, col, errstr);
	error = PROGERR_COMPILE;
	rexxSetResults(msg, error, mem);
    }

    PoolFreeMem(memPool, mem, memSize);

    return rexxTranslateError(error);
}

static HOOK_STRUCT(rexxCompile)

struct MUI_Command MUI_Commands[] =
{
    { "compile",       "SOURCE/A,UPDATE/S", 2, &rexxCompileHook, },
    { "connect",       "PROJECT/A,GUI/S",  2, &rexxConnectHook, },
    { "disconnect",    "HANDLE/N/A", 1, &rexxDisconnectHook, },
    { "connections",   "",           0, &rexxConnectionsHook, },
    { "eval",          "HANDLE/N/A,COMMAND/F", 2, &rexxEvalHook, },
    { "transaction",   "HANDLE/N/A", 1, &rexxTransactionHook, },
    { "commit",        "HANDLE/N/A", 1, &rexxCommitHook, },
    { "rollback",      "HANDLE/N/A", 1, &rexxRollbackHook, },
    { NULL, NULL, 0, NULL, },
};

