/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** Search_mui.c
**
** Creation date: 16.09.1996
** Last modified: 18.07.2022
*/

#include "search.h"
#include "project.h"
#include "table.h"
#include "field.h"
#include "error.h"
#include "util.h"
#include "requesters.h"
#include "appstrings_cat.h"
#include "extclasses/ExtApplication_mcc.h"

struct SearchFor *GuiSearchFor(struct SearchFor *sf)
{
    GuiObject app = GuiApplication, win = NULL;
    GuiObject pattern, sensitive, all, method, start, find, cancel;
    const char *methodstr[3], *startstr[3];
    bool ok = FALSE;

    if (app == NULL || sf == NULL) {
        return NULL;
    }

    pattern = StringObject,
	MUIA_Frame, MUIV_Frame_String,
	MUIA_String_MaxLen, sizeof(sf->sf_Pattern),
	MUIA_String_Format, MUIV_String_Format_Center,
        MUIA_String_Contents, sf->sf_Pattern,
        MUIA_ControlChar, Shortcut(GCS(MSG_SearchPattern)),
        MUIA_ShortHelp, GCS(MSG_SearchPatternBH),
        MUIA_CycleChain, TRUE,
	End;
    MUIStringPreselectAll(pattern, TRUE);

    sensitive = MUI_MakeObject(MUIO_Checkmark, (IPTR)GCS(MSG_CaseSensitive));
    all = MUI_MakeObject(MUIO_Checkmark, (IPTR)GCS(MSG_AllFields));

    methodstr[0] = GCS(MSG_Forward);
    methodstr[1] = GCS(MSG_Backward);
    methodstr[2] = NULL;
    method = RadioObject,
	MUIA_Radio_Entries, methodstr,
        MUIA_Radio_Active, sf->sf_Forward? 0 : 1,
        MUIA_CycleChain, TRUE,
	End;

    startstr[0] = GCS(MSG_FirstLastRecord);
    startstr[1] = GCS(MSG_CurrentRecord);
    startstr[2] = NULL;
    start = RadioObject,
	MUIA_Radio_Entries, startstr,
        MUIA_Radio_Active, sf->sf_FirstLastRecord? 0 : 1,
        MUIA_CycleChain, TRUE,
	End;

    find = SimpleButton(GCS(MSG_Find));
    cancel = SimpleButton(GCS(MSG_Cancel));

    win = WindowObject,
        MUIA_Window_Title, GCS(MSG_Find),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
	MUIA_Window_ID, MAKE_ID('_', 'F', 'N', 'D'),
	MUIA_Window_NoMenus, TRUE,
        MUIA_HelpNode, "Search dialog",
	WindowContents, VGroup,
            Child, VGroup,
		MUIA_Frame, MUIV_Frame_Group,
                MUIA_Background, MUII_GroupBack,
	        Child, VGroup,
		    MUIA_Group_Columns, 2,
                    Child, Label2(GCS(MSG_SearchPattern)),
                    Child, pattern,
                    Child, Label1(GCS(MSG_CaseSensitive)),
                    Child, HGroup,
                        Child, VCenter(sensitive),
                        Child, HSpace(0),
                        Child, Label1(GCS(MSG_AllFields)),
                        Child, VCenter(all),
                        Child, HSpace(0),
                        End,
                    End,
                Child, HGroup,
		    MUIA_Group_SameWidth, TRUE,
		    Child, HGroup,
		        MUIA_Frame, MUIV_Frame_Group,
		        MUIA_FrameTitle, GCS(MSG_SearchDirection),
                        MUIA_Background, MUII_GroupBack,
		        Child, HSpace(0),
		        Child, method,
		        Child, HSpace(0),
                        End,
		    Child, HGroup,
		        MUIA_Frame, MUIV_Frame_Group,
		        MUIA_FrameTitle, GCS(MSG_StartWith),
                        MUIA_Background, MUII_GroupBack,
		        Child, HSpace(0),
		        Child, start,
		        Child, HSpace(0),
                        End,
                    End,
                End,
            Child, HGroup,
      		MUIA_Group_SameWidth, TRUE,
                Child, find,
                Child, cancel,
                End,
            End,
        End;

    if (win == NULL) {
        return NULL;
    }

    SetAttrs(sensitive,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_CaseSensitiveBH),
	MUIA_Selected, sf->sf_CaseSensitive, 
        TAG_DONE);
    SetAttrs(all,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_AllFieldsBH),
        MUIA_Selected, (sf->sf_Field == NULL || sf->sf_AllFields),
        MUIA_Disabled, (sf->sf_Field == NULL),
        TAG_DONE);
    set(find, MUIA_CycleChain, TRUE);
    set(cancel, MUIA_CycleChain, TRUE);

    set(win, MUIA_Window_ActiveObject, pattern);
    DoMethod(pattern, MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
        win, 3, MUIM_Set, MUIA_Window_ActiveObject, find);
    DoMethod(win, MUIM_Notify, MUIA_Window_CloseRequest, TRUE,
        app, 2, MUIM_Application_ReturnID, ID_CLOSE);
    DoMethod(cancel, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 2, MUIM_Application_ReturnID, ID_CANCEL);
    DoMethod(find, MUIM_Notify, MUIA_Pressed, FALSE,
        app, 2, MUIM_Application_ReturnID, ID_OK);

    GuiSetSleep(TRUE);
    MUIAddMember(app, win);
    if (GuiOpenWindow(win)) {
        ok = WaitOk();
        if (ok) {
            char *str = NULL;

            str = pget(pattern, MUIA_String_Contents);
            if (str != NULL)
                strcpy(sf->sf_Pattern, str);
            sf->sf_CaseSensitive = nget(sensitive, MUIA_Selected);
            sf->sf_AllFields = nget(all, MUIA_Selected);
            sf->sf_Forward = nget(method, MUIA_Radio_Active) == 0;
            sf->sf_FirstLastRecord = nget(start, MUIA_Radio_Active) == 0;
        }
        GuiCloseWindow(win);
    }
    MUIRemMember(app, win);
    MUI_AddDisposeObject(app, win);
    GuiSetSleep(FALSE);

    return (ok? sf : NULL);
}

