/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** setformats_mui.c
**
** Creation date: 14.08.1998
** Last modified: 24.04.2008
*/

#include "setformats.h"
#include "main.h"
#include "gui.h"
#include "preferences.h"
#include "project.h"
#include "util.h"
#include "memory.h"
#include "appstrings_cat.h"
#include "extclasses/ExtApplication_mcc.h"

struct SetFormatsData
{
    struct Project *d_Project;
    Object *d_RealFormat, *d_DateFormat;
    Object *d_Ok, *d_Cancel;
};

static void disposeSetFormats(Object *app, Object *obj)
{
    struct SetFormatsData *d = pget(obj, MUIA_UserData);

    if(obj && d)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d,
            sizeof(struct SetFormatsData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void HOOK_PROTO(finish, Object *obj, ulong *ok)
{
    struct SetFormatsData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);

    if(*ok)
    {
        SetPrefsValue(pr, PREFS_REALFORMAT,
            nget(d->d_RealFormat, MUIA_Radio_Active));
        SetPrefsValue(pr, PREFS_DATEFORMAT,
            nget(d->d_DateFormat, MUIA_Radio_Active));
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    GuiSetSleep(FALSE);

    disposeSetFormats(app, obj);
    if(*ok)
        FORNODE(pr, &ProjectList)
            GuiUpdate(pr);
}

static Object *createSetFormats(struct Project *pr)
{
    Object *app = GuiApplication;
    struct MemPool *memPool = ProjectMemPool(pr);
    struct SetFormatsData *d;
    static const char *realformats[3], *dateformats[4];
    Object *realgroup, *dategroup, *buttons, *obj;

    if(app == NULL)
        return(NULL);

    d = PoolAllocMemZero(memPool, sizeof(struct SetFormatsData));
    if(d == NULL)
        return(NULL);
    d->d_Project = pr;

    realformats[0] = GCS(MSG_Formats_RealFormat_DecimalPoint);
    realformats[1] = GCS(MSG_Formats_RealFormat_DecimalComma);
    realformats[2] = NULL;

    d->d_RealFormat = RadioObject,
        MUIA_Radio_Entries, realformats,
        MUIA_Radio_Active, GetPrefsValue(pr, PREFS_REALFORMAT),
        MUIA_CycleChain, TRUE,
        End;

    realgroup = HGroup,
        GroupFrame,
        MUIA_FrameTitle, GCS(MSG_Formats_RealFormat),
        MUIA_Background, MUII_GroupBack,
        Child, HSpace(0),
        Child, d->d_RealFormat,
        Child, HSpace(0),
        End;

    dateformats[0] = GCS(MSG_Formats_DateFormat_DayMonthYear);
    dateformats[1] = GCS(MSG_Formats_DateFormat_MonthDayYear);
    dateformats[2] = GCS(MSG_Formats_DateFormat_YearMonthDay);
    dateformats[3] = NULL;

    d->d_DateFormat = RadioObject,
        MUIA_Radio_Entries, dateformats,
        MUIA_Radio_Active, GetPrefsValue(pr, PREFS_DATEFORMAT),
        MUIA_CycleChain, TRUE,
        End;

    dategroup = HGroup,
        GroupFrame,
        MUIA_FrameTitle, GCS(MSG_Formats_DateFormat),
        MUIA_Background, MUII_GroupBack,
        Child, HSpace(0),
        Child, d->d_DateFormat,
        Child, HSpace(0),
        End;

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, d->d_Ok,
	Child, d->d_Cancel,
	End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_ID, MAKE_ID('_', 'F', 'M', 'T'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_Window_Title, GCS(MSG_Formats),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
        MUIA_HelpNode, "Formats",
        MUIA_UserData, d,
	WindowContents, VGroup,
            Child, realgroup,
            Child, dategroup,
            Child, buttons,
            End,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct SetFormatsData));
	return(NULL);
    }

    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE, obj, finish,FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE, obj, finish, FALSE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE, obj, finish, TRUE);

    GuiSetSleep(TRUE);
    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        GuiSetSleep(FALSE);
        disposeSetFormats(app, obj);
        return(NULL);
    }

    return(obj);
}

void GuiSetFormats(struct Project *pr)
{
    createSetFormats(pr);
}

