/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** StructureEditor_mui.c
**
** Creation date: 16.11.1994
** Last modified: 06.08.2020
*/

#include "system.h"
#include "structureeditor.h"
#include "lang/program.h"
#include "preferences.h"
#include "codeset.h"
#include "util.h"
#include "quicksearch_mui.h"
#include "extclasses/DDList_mcc.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "extclasses/ExtApplication_mcc.h"
#include "structureeditor_mui.h"
#include "project.h"
#include "table.h"
#include "field.h"
#include "display.h"
#include "buildgui.h"
#include "requesters.h"
#include "name.h"
#include "appstrings_cat.h"
#include "error.h"
#include "memory.h"

#endif

/*
** UserFunctions Popup
*/

static ulong HOOK_PROTO(userFunctionDisplay, const char **dst,
	struct UserFunction *uf)
{
    *dst = EmptyNullStr(UserFunctionName(uf));
    return(0);
}
static HOOK_STRUCT(userFunctionDisplay)

static void userFunctionSearchFunc(GuiObject go,
        const char *str, int cmd, void *pop)
{
    long dir = -1;

    switch (cmd) {
	case QUICK_SEARCH_NEXT:  dir++;		/* no break */
	case QUICK_SEARCH_START: dir++;		/* no break */
	case QUICK_SEARCH_PREV:
            GuiUserFunctionsQuickSearch(go, dir, str);
            break;

	case QUICK_SEARCH_COMMIT:
            DoMethod(pop, MUIM_Popstring_Close, TRUE);
            break;
    }
}

static bool userFunctionsKeyFunc(GuiObject obj, char *buf, short len,
        ushort qualifier, GuiObject pop)
{
    if (MUIIsSearchStartString(buf, len, qualifier)) {
        return GuiQuickSearchInit(obj, buf[0], userFunctionSearchFunc, pop);
    }

    return FALSE;
}

static long HOOK_PROTO(userFunctionCompare, 
	struct UserFunction *uf2, struct UserFunction *uf1)
{
    const char *name1 = EmptyNullStr(UserFunctionName(uf1));
    const char *name2 = EmptyNullStr(UserFunctionName(uf2));
    
    return CodesetStrCollate(name1, name2, CS_COLLATE_ORDER);
}
static HOOK_STRUCT(userFunctionCompare)

static long HOOK_PROTO(userFunctionOpen, Object *pop, Object *str)
{
    struct Project *pr = pget(pop, MUIA_UserData);
    struct Program *prg = ProjectProgram(pr);
    struct List *uflist = ProgramUserFunctions(prg);
    const char *cur_name = pget(str, MUIA_String_Contents), *name;
    struct UserFunction *uf, *active = NULL;
    bool sort = GetPrefsValue(pr, PREFS_PROGRAMSORTTRIGGERFUNCTIONS);

    set(pop, MUIA_List_Quiet, TRUE);
    DoMethod(pop, MUIM_List_Clear);
    if (sort) {
	set(pop, MUIA_List_CompareHook, &userFunctionCompareHook);
    }
    if (uflist) {
        FORNODE(uf, uflist) {
            if (!UserFunctionTrigger(uf)) {
                continue;
            }

            DoMethod(pop, MUIM_List_InsertSingle, uf, 
		    sort? MUIV_List_Insert_Sorted : MUIV_List_Insert_Bottom);
            name = UserFunctionName(uf);
            if (name && cur_name && strcmp(name, cur_name) == 0) {
                active = uf;
            }
        }
    }
    set(pop, MUIA_List_Quiet, FALSE);
    MUIListSetActiveEntry(pop, active);

    return TRUE;
}
static HOOK_STRUCT(userFunctionOpen)

static void HOOK_PROTO(userFunctionClose, Object *list, Object *str)
{
    struct UserFunction *uf = NULL;
    const char *name = NULL;

    DoMethod(list, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, &uf);
    name = UserFunctionName(uf);
    set(str, MUIA_String_Contents, EmptyNullStr(name));
}
static HOOK_STRUCT(userFunctionClose)

static void HOOK_PROTO(userFunctionWindow, Object *pop, Object *win)
{
    set(win, MUIA_Window_DefaultObject, pop);
}
static HOOK_STRUCT(userFunctionWindow)

GuiObject UserFunctionsPopup(GuiObject string, struct Project *pr)
{
    GuiObject functions = ListviewObject,
	MUIA_CycleChain, TRUE,
	MUIA_Listview_DoubleClick, TRUE,
        MUIA_UserData, pr,
	MUIA_Listview_List, DDListObject,
            InputListFrame,
            MUIA_List_DisplayHook, &userFunctionDisplayHook,
	    End,
	End;

    GuiObject pop = PopobjectObject,
       	MUIA_Popstring_String, string,
	MUIA_Popstring_Button, PopButton(MUII_PopUp),
	MUIA_Popobject_Object, functions,
        MUIA_Popobject_ObjStrHook, &userFunctionCloseHook,
        MUIA_Popobject_StrObjHook, &userFunctionOpenHook,
        MUIA_Popobject_WindowHook, &userFunctionWindowHook,
	End;

    if (pop) {
        SetAttrs(functions,
            MUIA_KeyFunc, userFunctionsKeyFunc,
            MUIA_KeyData, pop,
            TAG_DONE);
        DoMethod(functions, MUIM_Notify, MUIA_Listview_DoubleClick, TRUE,
                pop, 2, MUIM_Popstring_Close, TRUE);
    }

    return pop;
}

#ifdef ENABLE_STRUCTURE_EDITOR

static const char *displayTypeContent[3];

static const struct FC_Type *typenames[] =
{
    MSG_Err_Internal,
    MSG_Integer,
    MSG_Real,
    MSG_Boolean,
    MSG_Err_Internal,
    MSG_Memo,
    MSG_Date,
    MSG_Time,
    MSG_Err_Internal,
    MSG_Button,
    MSG_Choice,
    MSG_Reference,
    MSG_String,
    MSG_Virtual,
};

struct StructureEditorData
{
    struct Project *d_Project;
    void *d_NameChangeObject;
    struct Table *d_CurrentTable;
    bool d_RootDisplay;
    short d_TableSortMethod;
    short d_FieldSortMethod;
    Object *d_TablesGroup;
    Object *d_Tables;
    Object *d_NewTable;
    Object *d_DeleteTable;
    Object *d_SortTables;
    Object *d_FieldsGroup;
    Object *d_Fields;
    Object *d_NewField;
    Object *d_CopyField;
    Object *d_DeleteField;
    Object *d_SortFields;
    Object *d_DisplayObjectsGroup;
    Object *d_DisplayObjects;
    Object *d_Add;
    Object *d_Rem;
    Object *d_Up;
    Object *d_Down;
    Object *d_In;
    Object *d_Out;
    Object *d_Text;
    Object *d_Image;
    Object *d_Space;
    Object *d_Balance;
    Object *d_Group;
    Object *d_Register;
    Object *d_Window;
    Object *d_DisplayType;
};

#define RootDisplayObject(d) \
        (((d)->d_RootDisplay)? \
                ProjectRootDisplayObject((d)->d_Project) : \
                TableDisplayObject((d)->d_CurrentTable))

static ulong HOOK_PROTO(tabDisplay, const char **dst, struct Table *tab)
{
    static char buffer[2 + ID_NAME_LEN + 1];
    const char *name = TableName(tab);

    if (IsMBName(name)) {
        *dst = name;
    } else {
        sprintf(buffer, "\33i%s", name);
        *dst = buffer;
    }
    return 0;
}
static HOOK_STRUCT(tabDisplay)

static long HOOK_PROTO(tabCompareAlpha, struct Table *tab2, struct Table *tab1)
{
    return TableCompareAlpha(tab1, tab2);
}
static HOOK_STRUCT(tabCompareAlpha)

static long HOOK_PROTO(tabCompareDisplay, struct Table *tab2,struct Table *tab1)
{
    return TableCompareDisplay(tab1, tab2);
}
static HOOK_STRUCT(tabCompareDisplay)

#define BUF_LEN 100

static void makeentry(char *buffer, short left, const char *preparse,
	const char *name, const char *title)
{
    const char *extra = EmptyNullStr(preparse);
    const char *normal = extra[0] != 0? "\33n" : "";
    int maxl = BUF_LEN - 8 - strlen(extra);	/* at least 8 chars are seen */
    int l = MIN(MAX(left, 0), maxl), n;

    name = EmptyNullStr(name);
    sprintf(buffer, "%*s%s", l, "", extra);	/* indent and optional bf */
    l = strlen(buffer);
    n = BUF_LEN - l;
    sprintf(buffer + l, "%.*s", n, name);	/* name */
    if (title && title[0] && strcmp(name, title) != 0) {
        l =  strlen(buffer);
        n = BUF_LEN - l - 2 - strlen(normal);
        if (n > 0) {
            sprintf(buffer + l, "%s(%.*s)", normal, n-1, title);  /* (title) */
        }
    }
}

static ulong HOOK_PROTO(fieldDisplay, const char **dst, struct Field *field)
{
    static char buffer[BUF_LEN + 1];
    const char *name = FieldName(field);
    short type = FieldType(field);

    makeentry(buffer, 0, IsMBName(name)? "" : MUIItalicString, name,
            GCS(typenames[type]));
    *dst = buffer;
    return 0;
}
static HOOK_STRUCT(fieldDisplay)

static long HOOK_PROTO(fieldCompareAlpha, 
	struct Field *field2, struct Field *field1)
{
    return FieldCompareAlpha(field1, field2);
}
static HOOK_STRUCT(fieldCompareAlpha)

static long HOOK_PROTO(fieldCompareDisplay, 
	struct Field *field2, struct Field *field1)
{
    return FieldCompareDisplay(field1, field2);
}
static HOOK_STRUCT(fieldCompareDisplay)

static bool structureEditorRootDisplay(struct DisplayObject *dio)
{
    struct Project *pr = DisplayObjectProject(dio);
    GuiObject se = ProjectStructureEditor(pr);
    struct StructureEditorData *d = pget(se, MUIA_UserData);

    return (d && d->d_RootDisplay);
}

static ulong HOOK_PROTO(displayDisplay, const char **dst,
	struct DisplayObject *dio)
{
    static char *arrowstr[] = { "", "\33I[6:30]", "\33I[6:39]", };
    static char buffer[BUF_LEN+1];
    struct DisplayObject *parent = DisplayObjectParent(dio), *grandpa = NULL;
    int i=0, left=0, arrow = 0;
    const char *name = NULL, *title = DisplayObjectTitle(dio);
    bool bold = FALSE, useOpen = FALSE;
    char *s=NULL;

    for (i = 0; (grandpa = DisplayObjectParent(parent)) != NULL; 
	    parent = grandpa, i++) {
        if (DisplayObjectType(parent) == DO_TYPE_TABLE) {
            i++;                            /* add one for table group */
            break;
        }
    }
    left = i * 2;

    switch (DisplayObjectType(dio)) {
	case DO_TYPE_TABLE:
	    name = TableName(DisplayObjectTable(dio));
	    bold = TRUE;
            if (!structureEditorRootDisplay(dio)) {
                arrow = 2;
                left = 0;
            }
            break;
	case DO_TYPE_FIELD:
	    name = FieldName(DisplayFieldField(dio));
	    bold = TRUE;
	    break;
        case DO_TYPE_TEXT:
	    name = GCS(MSG_Text);
            break;
	case DO_TYPE_SPACE:
            name = GCS(MSG_Space);
            break;
	case DO_TYPE_BALANCE:
	    name = GCS(MSG_StructureEditor_Balance);
            break;
	case DO_TYPE_IMAGE:
            name = GCS(MSG_Image);
            break;
	case DO_TYPE_REGISTER:
	    name = GCS(MSG_StructureEditor_Register);
	    useOpen = TRUE;
	    break;
	case DO_TYPE_GROUP:
	    if (DisplayGroupHoriz(dio)) {
		name = GCS(MSG_StructureEditor_Display_HGroup);
	    } else {
		name = GCS(MSG_StructureEditor_Display_VGroup);
	    }
	    useOpen = TRUE;
	    break;
	case DO_TYPE_WINDOW:
	    name = GCS(MSG_Window);
	    useOpen = TRUE;
	    break;
        default:
            break;
    }

    makeentry(buffer, left, bold? MUIBoldString : "", name, title);
    /* clean buffer, e.g. replace new line characters by white space. */
    for (s=buffer; *s; s = CodesetNextChar(s)) {
        if (CodesetIsSpace(CodesetGetChar(s))) {
            *s = ' ';
        }
    }
    if (useOpen) {
        arrow = DisplayObjectOpen(dio)? 2 : 1;
    }

    *dst++ = arrowstr[arrow];
    *dst = buffer;
    return 0;
}
static HOOK_STRUCT(displayDisplay)

static void setTables(Object *obj)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct Table *active = DisplayProjectActiveTable(d->d_Project);
    struct MemPool *memPool = ProjectMemPool(d->d_Project);

    MUIListSetListWithFilter(d->d_Tables, ProjectTables(d->d_Project), memPool,
            TableSystemFilter);
    MUIListSetActiveNode(d->d_Tables, active);
}

static void HOOK_PROTO(fieldsActiveChange, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct Field *field = NULL;

    DoMethod(d->d_Fields, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&field);
    set(d->d_CopyField, MUIA_Disabled, field == NULL);
    set(d->d_DeleteField, MUIA_Disabled, field == NULL);
    SetDisplayTableActiveField(d->d_CurrentTable, field);
}

static void setFields(Object *obj)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct List *fields = TableFields(d->d_CurrentTable);
    struct Field *active = DisplayTableActiveField(d->d_CurrentTable);
    struct MemPool *memPool = ProjectMemPool(d->d_Project);

    MUIListSetList(d->d_Fields, fields, memPool);
    MUIListSetActiveNode(d->d_Fields, active);
    fieldsActiveChangeFunc(NULL, obj, NULL);
    set(d->d_FieldsGroup, MUIA_Disabled, fields == NULL);
    fieldsActiveChangeFunc(NULL, obj, NULL);
}

static void HOOK_PROTO(displayActiveChange, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct DisplayObject *dio = NULL;

    DoMethod(d->d_DisplayObjects, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&dio);
    if(d->d_RootDisplay)
        SetDisplayProjectActiveDisplayObject(d->d_Project, dio);
    else
        SetDisplayTableActiveDisplayObject(d->d_CurrentTable, dio);
}

static void setDisplayRec(Object *lv, struct DisplayObject *dio)
{
    struct List *dios = DisplayObjectObjects(dio);
    struct DisplayObject *child;

    if(dios && DisplayObjectOpen(dio))
        FORNODE(child, dios)
        {
            if(DisplayObjectVisible(child))
                DoMethod(lv, MUIM_List_InsertSingle, child, MUIV_List_Insert_Bottom);

            if(DisplayObjectType(child) != DO_TYPE_TABLE)
                setDisplayRec(lv, child);
        }
}

static void setDisplay(Object *obj)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct Table *tab = d->d_CurrentTable;
    struct DisplayObject *dio = TableDisplayObject(tab);
    Object *lv = d->d_DisplayObjects;
    long oldactive = nget(lv, MUIA_List_Active), active;
    long first = nget(lv, MUIA_List_First);
    long visible = nget(lv, MUIA_List_Visible);
    struct DisplayObject *activedio = NULL;

    set(lv, MUIA_List_Quiet, TRUE);
    DoMethod(lv, MUIM_List_Clear);

    if(d->d_RootDisplay)
    {
        set(d->d_DisplayObjectsGroup, MUIA_Disabled, FALSE);
        activedio = DisplayProjectActiveDisplayObject(d->d_Project);
        dio = ProjectRootDisplayObject(d->d_Project);
    }
    else if(tab == NULL || !DisplayObjectVisible(dio))
    {
        set(d->d_DisplayObjectsGroup, MUIA_Disabled, TRUE);
        dio = NULL;
    }
    else
    {
        set(d->d_DisplayObjectsGroup, MUIA_Disabled, FALSE);
        activedio = DisplayTableActiveDisplayObject(tab);
        DoMethod(lv, MUIM_List_InsertSingle, dio, MUIV_List_Insert_Bottom);
    }

    setDisplayRec(lv, dio);
    active = GuiListFindNode(lv, activedio);
    if (active < 0) {		/* e.g. when activedio has been removed */
	active = oldactive;
    }
    if (first > 0 && visible > 0) {
	/* scroll back to same starting line */
	DoMethod(lv, MUIM_List_Jump, first + visible - 1);
    }
    set(lv, MUIA_List_Quiet, FALSE);

    if (active >= 0) {
	set(lv, MUIA_List_Active, active);
    }
}

static void addDisplayObject(Object *obj, struct DisplayObject *dio)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);

    if (d->d_RootDisplay) {
        SetDisplayProjectActiveDisplayObject(d->d_Project, dio);
    } else {
        SetDisplayTableActiveDisplayObject(d->d_CurrentTable, dio);
    }
    setDisplay(obj);
}

static void HOOK_PROTO(tablesActiveChange, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct Table *tab = NULL;

    DoMethod(d->d_Tables, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&tab);
    set(d->d_DeleteTable, MUIA_Disabled, tab == NULL);
    d->d_CurrentTable = tab;
    SetDisplayProjectActiveTable(d->d_Project, tab);
    setFields(obj);
    if(!d->d_RootDisplay)
        setDisplay(obj);
}

static void HOOK_PROTO(tablesDropped, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct List *tabs = ProjectTables(d->d_Project);

    MUIListGetList(d->d_Tables, tabs);
}

#define MAX_SELECTED_OBJECTS 200
static struct DisplayObject *selectedObjects[MAX_SELECTED_OBJECTS + 1];

static long markselected(Object *lv)
{
    long i, num = 0, state;
    struct DisplayObject *dio = NULL;

    for(i=0; num < MAX_SELECTED_OBJECTS; i++)
    {
	DoMethod(lv, MUIM_List_GetEntry, i, &dio);
	if(dio == NULL)
	    break;
       	DoMethod(lv, MUIM_List_Select, i, MUIV_List_Select_Ask, &state);
        if(state == MUIV_List_Select_On)
            selectedObjects[num++] = dio;
    }
    selectedObjects[num] = NULL;
    return(num);
}

static void HOOK_PROTO(newTableClick, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct DisplayObject *dioparent = NULL;

    if(d->d_RootDisplay)
    {
        /* get active display item */
        DoMethod(d->d_DisplayObjects,
	    MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&dioparent);
    }

    /* open the new table gui */
    GuiNewTable(obj, d->d_Project, dioparent);
}

void NewTableFinish(Object *obj, struct Table *new)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);

    if(obj && new)
    {
        /* update table and display list */

        MUIListInsertNode(d->d_Tables, new);
        addDisplayObject(obj, TableDisplayObject(new));
    }
}

static void HOOK_PROTO(changeTableClick, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct Table *tab = NULL;

    DoMethod(d->d_Tables,
	MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&tab);

    if(tab != NULL)
        GuiChangeTable(obj, tab);
}

void ChangeTableFinish(Object *obj, struct Table *tab)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);

    if(obj && tab)
    {
        MUIListRedrawNode(d->d_Tables, tab);
        MUIListRedrawNode(d->d_DisplayObjects, TableDisplayObject(tab));
        if (d->d_NameChangeObject) {
            StructureEditorNextNameChange(d->d_Project, d->d_NameChangeObject);
        }
    }
}

static void HOOK_PROTO(deleteTableClick, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct Table *tab = NULL;
    struct DisplayObject *dio = NULL;
    bool deleted;

    DoMethod(d->d_Tables,
	MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&tab);

    if(tab == NULL)
        return;

    dio = TableDisplayObject(tab);

    /* delete the table */
    deleted = GuiDeleteTable(tab);

    if(deleted)
    {
        /* remove table from table and display list */

        MUIListRemoveEntry(d->d_Tables, tab);
        MUIListRemoveEntry(d->d_DisplayObjects, dio);
    }
}

static void HOOK_PROTO(sortTablesClick, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct List *tabs = ProjectTables(d->d_Project);

    if (d->d_TableSortMethod == 0)
    {
	set(d->d_Tables, MUIA_List_CompareHook, &tabCompareAlphaHook);
	set(d->d_SortTables, 
		MUIA_Text_Contents, GCS(MSG_StructureEditor_Sort2));
	d->d_TableSortMethod = 1;
    }
    else
    {
	set(d->d_Tables, MUIA_List_CompareHook, &tabCompareDisplayHook);
	set(d->d_SortTables, 
		MUIA_Text_Contents, GCS(MSG_StructureEditor_Sort1));
	d->d_TableSortMethod = 0;
    }

    DoMethod(d->d_Tables, MUIM_List_Sort);
    MUIListGetList(d->d_Tables, tabs);

    tablesActiveChangeFunc(NULL, obj, NULL);
}

static void HOOK_PROTO(fieldsDropped, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct List *fields = TableFields(d->d_CurrentTable);

    if(fields)
        MUIListGetList(d->d_Fields, fields);
}

static void HOOK_PROTO(newFieldClick, Object *obj, long *x)
{
    long copy = *x;
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct DisplayObject *dioparent = NULL;
    struct Field *field = NULL;

    if(copy)
      {
	DoMethod(d->d_Fields, MUIM_List_GetEntry, 
		 MUIV_List_GetEntry_Active, (APTR)&field);
	if(field == NULL)
	  return;
      }

    if(!d->d_RootDisplay)
    {
        /* get active display item */
        DoMethod(d->d_DisplayObjects,
	    MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&dioparent);
    }

    /* open the new field gui */
    GuiNewField(obj, d->d_CurrentTable, field, dioparent);
}

void NewFieldFinish(Object *obj, struct Field *field)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);

    if(obj && field)
    {
        /* update the fields list and the display list*/

        MUIListInsertNode(d->d_Fields, field);
        addDisplayObject(obj, FieldDisplayObject(field));
    }
}

static void HOOK_PROTO(changeFieldClick, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct Field *field = NULL;

    DoMethod(d->d_Fields,
	MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&field);

    if(field)
        GuiChangeField(obj, field);
}

void ChangeFieldFinish(Object *obj, struct Field *field)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);

    if(obj && field)
    {
    	/* update the fields list and the display list*/

        MUIListRedrawNode(d->d_Fields, field);
        MUIListRedrawNode(d->d_DisplayObjects, FieldDisplayObject(field));
        if (d->d_NameChangeObject) {
            StructureEditorNextNameChange(d->d_Project, d->d_NameChangeObject);
        }
    }
}

static void HOOK_PROTO(deleteFieldClick, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct Field *field = NULL;
    struct DisplayObject *dio = NULL;
    bool deleted;

    DoMethod(d->d_Fields,
	MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&field);

    if(field)
    {
        dio = FieldDisplayObject(field);

	deleted = GuiDeleteField(field);

        if(deleted)
        {
            MUIListRemoveEntry(d->d_Fields, field);
            MUIListRemoveEntry(d->d_DisplayObjects, dio);
        }
    }
}

static void HOOK_PROTO(sortFieldsClick, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct List *fields = TableFields(d->d_CurrentTable);

    if(fields)
    {
	if (d->d_FieldSortMethod == 0)
	{
	    set(d->d_Fields, 
		MUIA_List_CompareHook, &fieldCompareAlphaHook);
	    set(d->d_SortFields, 
		    MUIA_Text_Contents, GCS(MSG_StructureEditor_Sort2));
	    d->d_FieldSortMethod = 1;
	}
	else
	{
	    set(d->d_Fields,MUIA_List_CompareHook, &fieldCompareDisplayHook);
	    set(d->d_SortFields, 
		    MUIA_Text_Contents, GCS(MSG_StructureEditor_Sort1));
	    d->d_FieldSortMethod = 0;
	}

        DoMethod(d->d_Fields, MUIM_List_Sort);
        MUIListGetList(d->d_Fields, fields);
    }
}

static bool openDisplayObject(Object *obj, struct DisplayObject *dio)
{
    short type = DisplayObjectType(dio);

    if(dio == NULL)
        return(FALSE);
    if(type != DO_TYPE_REGISTER && type != DO_TYPE_GROUP && 
    	type != DO_TYPE_WINDOW)
            return(FALSE);
    if(DisplayObjectOpen(dio))
        return(FALSE);
    SetDisplayObjectOpen(dio, TRUE);
    setDisplay(obj);
    return(TRUE);
}

static bool closeDisplayObject(Object *obj, struct DisplayObject *dio)
{
    short type = DisplayObjectType(dio);

    if(dio == NULL)
        return(FALSE);
    if(type != DO_TYPE_REGISTER && type != DO_TYPE_GROUP && 
    	type != DO_TYPE_WINDOW)
	    return(FALSE);
    if(DisplayObjectOpen(dio) == FALSE)
        return(FALSE);
    SetDisplayObjectOpen(dio, FALSE);
    setDisplay(obj);
    return(TRUE);
}

static void defaultFinish(Object *obj, struct DisplayObject *dio)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);

    if(obj && dio)
        MUIListRedrawNode(d->d_DisplayObjects, dio);
}

void TextFinish(Object *obj, struct DisplayObject *dio) 
{ defaultFinish(obj, dio); }
void SpaceFinish(Object *obj, struct DisplayObject *dio) 
{ defaultFinish(obj, dio); }
void ImageFinish(Object *obj, struct DisplayObject *dio) 
{ defaultFinish(obj, dio); }
void GroupFinish(Object *obj, struct DisplayObject *dio) 
{ defaultFinish(obj, dio); }
void RegisterFinish(Object *obj, struct DisplayObject *dio) 
{ defaultFinish(obj, dio); }
void WindowFinish(Object *obj, struct DisplayObject *dio) 
{ defaultFinish(obj, dio); }
void DisplayTableFinish(Object *obj, struct DisplayObject *dio) 
{ defaultFinish(obj, dio); }
void DisplayFieldFinish(Object *obj, struct DisplayObject *dio) 
{ defaultFinish(obj, dio); }

static void HOOK_PROTO(changeDisplayClick, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct DisplayObject *dio = NULL;
    long col = 0;

    get(d->d_DisplayObjects, MUIA_Listview_ClickColumn, &col);
    DoMethod(d->d_DisplayObjects, MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&dio);
    if(dio)
    {
        if(col == 0)
        {
            if(DisplayObjectOpen(dio))
                closeDisplayObject(obj, dio);
            else
                openDisplayObject(obj, dio);
            return;
        }

	switch(DisplayObjectType(dio))
	{
            case DO_TYPE_TEXT:
    		GuiText(obj, dio);
		break;

            case DO_TYPE_SPACE:
    		GuiSpace(obj, dio);
		break;

            case DO_TYPE_IMAGE:
    		GuiImage(obj, dio);
		break;

            case DO_TYPE_GROUP:
    		GuiGroup(obj, dio);
		break;

            case DO_TYPE_REGISTER:
    		GuiRegister(obj, dio);
		break;

            case DO_TYPE_WINDOW:
    		GuiWindow(obj, dio);
		break;

            case DO_TYPE_TABLE:
                GuiDisplayTable(obj, dio);
		break;

            case DO_TYPE_FIELD:
                GuiDisplayField(obj, dio);
		break;
	}
    }
}

static void HOOK_PROTO(displayCycleChange, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    long active = 0;

    get(d->d_DisplayType, MUIA_Cycle_Active, &active);
    set(d->d_Add, MUIA_ShortHelp, (GCS((active == 1)? 
		MSG_StructureEditor_AddBH1 : MSG_StructureEditor_AddBH2)));
    d->d_RootDisplay = (active == 1);
    setDisplay(obj);
}

static void HOOK_PROTO(addClick, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct Table *tab = NULL;
    struct Field *field = NULL;
    struct DisplayObject *add, *dioparent = NULL;
    long error;

    /* get active entry from tables/fields list */
    if (d->d_RootDisplay) {
        DoMethod(d->d_Tables, MUIM_List_GetEntry, MUIV_List_GetEntry_Active,
                (APTR)&tab);
        add = TableDisplayObject(tab);
    } else {
        DoMethod(d->d_Fields, MUIM_List_GetEntry, MUIV_List_GetEntry_Active,
                (APTR)&field);
        add = FieldDisplayObject(field);
    }

    /* get active entry from display list */
    DoMethod(d->d_DisplayObjects,
	MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&dioparent);

    if (add) {
        error = ObtainProjectChangeLock(d->d_Project);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        DisplayObjectAdd(add, dioparent, FALSE);
        addDisplayObject(obj, add);
        ProjectChange(d->d_Project, TRUE);
    }
}

static void HOOK_PROTO(remClick, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct DisplayObject *dio = NULL;
    long error;

    /* get active entry from display list */
    DoMethod(d->d_DisplayObjects,
	MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&dio);

    if (dio) {
        error = ObtainProjectChangeLock(d->d_Project);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        DisplayObjectRemove(dio);
        MUIListRemoveEntry(d->d_DisplayObjects, dio);
	DoMethod(d->d_DisplayObjects, MUIM_List_Redraw, MUIV_List_Redraw_All);
        ProjectChange(d->d_Project, TRUE);
    }
}

static void displayAction(Object *obj, bool (*func)(struct DisplayObject *dio))
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct DisplayObject *dio = NULL;
    long error;

    /* get active entry from display list */
    DoMethod(d->d_DisplayObjects,
	MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&dio);

    if (dio) {
        if (DisplayObjectType(dio) != DO_TYPE_TABLE || d->d_RootDisplay) {
            error = ObtainProjectChangeLock(d->d_Project);
            if (error != OK) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
                return;
            }

            if (func(dio)) {
                addDisplayObject(obj, dio);
                ProjectChange(d->d_Project, TRUE);
            }
        }
    }
}

static void HOOK_PROTO(upClick, Object *obj, void *dummy)
{
    displayAction(obj, DisplayObjectUp);
}

static void HOOK_PROTO(downClick, Object *obj, void *dummy)
{
    displayAction(obj, DisplayObjectDown);
}

static void HOOK_PROTO(inClick, Object *obj, void *dummy)
{
    displayAction(obj, DisplayObjectIn);
}

static void HOOK_PROTO(outClick, Object *obj, void *dummy)
{
    displayAction(obj, DisplayObjectOut);
}

static void simpleDisplayObjectClick(Object *obj, 
        long (*newFunc)(struct Project *pr, struct Table *tab, 
            struct DisplayObject *parent, struct DisplayObject **dio),
        void (*editFunc)(Object *obj, struct DisplayObject *dio)) 
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    struct DisplayObject *dioparent = NULL, *new = NULL;
    long error = 0;

    error = ObtainProjectChangeLock(d->d_Project);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    /* get active display item */
    DoMethod(d->d_DisplayObjects,
	MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&dioparent);
    if (dioparent == NULL) {
        dioparent = RootDisplayObject(d);
    }

    /* install new simple display object */
    error = newFunc(pr,
        d->d_RootDisplay? NULL : d->d_CurrentTable,
        dioparent, &new);
    if (error) {
   	Ask2(obj, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));
    } else {
        addDisplayObject(obj, new);
        ProjectChange(pr, TRUE);
        if (editFunc) {
            editFunc(obj, new);
        }
    }
}

static void HOOK_PROTO(textClick, Object *obj, void *dummy)
{
    simpleDisplayObjectClick(obj, DisplayNewText, GuiText);
}

static void HOOK_PROTO(imageClick, Object *obj, void *dummy)
{
    simpleDisplayObjectClick(obj, DisplayNewImage, GuiImage);
}

static void HOOK_PROTO(spaceClick, Object *obj, void *dummy)
{
    simpleDisplayObjectClick(obj, DisplayNewSpace, GuiSpace);
}

static void HOOK_PROTO(balanceClick, Object *obj, void *dummy)
{
    simpleDisplayObjectClick(obj, DisplayNewBalance, NULL);
}

static void compoundDisplayObjectClick(Object *obj,
        long (*newFunc)(struct Project *pr, struct Table *tab,
            struct DisplayObject **selected, struct DisplayObject *parent, 
            struct DisplayObject **dio),
        void (*editFunc)(Object *obj, struct DisplayObject *dio)) 
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    struct DisplayObject *dioparent = NULL, *new = NULL;
    long error = 0;
    Object *lv = d->d_DisplayObjects;

    error = ObtainProjectChangeLock(d->d_Project);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    /* get current selection of display list */
    markselected(lv);

    /* get active display item */
    DoMethod(lv, 
            MUIM_List_GetEntry, MUIV_List_GetEntry_Active, (APTR)&dioparent);
    if (dioparent == NULL) {
        dioparent = RootDisplayObject(d);
    }

    /* install new compound display object */
    error = newFunc(pr,
        d->d_RootDisplay? NULL : d->d_CurrentTable,
        selectedObjects, dioparent, &new);
    if (error) {
   	Ask2(obj, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));
    } else {
	addDisplayObject(obj, new);
        ProjectChange(pr, TRUE);
        editFunc(obj, new);
    }
}

static void HOOK_PROTO(groupClick, Object *obj, void *dummy)
{
    compoundDisplayObjectClick(obj, DisplayNewGroup, GuiGroup);
}

static void HOOK_PROTO(registerClick, Object *obj, void *dummy)
{
    compoundDisplayObjectClick(obj, DisplayNewRegister, GuiRegister);
}

static void HOOK_PROTO(windowClick, Object *obj, void *dummy)
{
    compoundDisplayObjectClick(obj, DisplayNewWindow, GuiWindow);
}

static void HOOK_PROTO(closeStructureEditor, Object *obj, void *dummy)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);

    StructureEditor(d->d_Project);
}

static void disposeStructureEditor(Object *app, Object *obj)
{
    struct StructureEditorData *d = pget(obj, MUIA_UserData);

    if(obj)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(struct StructureEditorData));
        MUI_AddDisposeObject(app, obj);
    }
}


void CloseStructureEditor(GuiObject se)
{
    Object *app = pget(se, MUIA_ApplicationObject);

    if (se) {
        GuiCloseWindow(se);
        MUIRemMember(app, se);
        disposeStructureEditor(app, se);
    }
}

GuiObject OpenStructureEditor(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    Object *app = GuiApplication;
    Object *se = NULL;
    Object *root, *tabFieldVert, *tabFieldHor, *tableActions;
    Object *fieldActions, *displayVert, *displayButtons;
    struct StructureEditorData *d = NULL;

    if(app == NULL)
        return(NULL);

    displayTypeContent[0] = GCS(MSG_StructureEditor_TableDisplay);
    displayTypeContent[1] = GCS(MSG_StructureEditor_MainWindow);
    displayTypeContent[2] = NULL;

    d = PoolAllocMemZero(memPool, sizeof(struct StructureEditorData));
    if(d == NULL)
        return(NULL);
    d->d_Project = pr;

    d->d_Tables = ListviewObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_TablesBH),
    	MUIA_Listview_DoubleClick, TRUE,
    	MUIA_Listview_DragType, MUIV_Listview_DragType_Immediate,
    	MUIA_Listview_List, DDListObject,
    	    MUIA_Frame, MUIV_Frame_InputList,
            MUIA_List_DragSortable, TRUE,
            MUIA_List_DisplayHook, &tabDisplayHook,
        End,
    End;

    d->d_NewTable = SimpleButton(GCS(MSG_StructureEditor_Table_New));
    d->d_DeleteTable = SimpleButton(GCS(MSG_StructureEditor_Table_Delete));
    d->d_SortTables = SimpleButton(GCS(MSG_StructureEditor_Sort1));

    tableActions = HGroup,
    	MUIA_Group_SameHeight, TRUE,
    	Child, d->d_NewTable,
    	Child, d->d_DeleteTable,
    	Child, d->d_SortTables,
	End;

    d->d_TablesGroup = VGroup,
    	MUIA_Weight, 30,
    	MUIA_Frame, MUIV_Frame_Group,
    	MUIA_FrameTitle, GCS(MSG_StructureEditor_Tables),
        MUIA_Background, MUII_GroupBack,
        MUIA_HelpNode, "Table management",
    	Child, d->d_Tables,
    	Child, tableActions,
	End;

    d->d_Fields = ListviewObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_FieldsBH),
    	MUIA_Listview_DoubleClick, TRUE,
    	MUIA_Listview_DragType, MUIV_Listview_DragType_Immediate,
    	MUIA_Listview_List, DDListObject,
    	    MUIA_Frame, MUIV_Frame_InputList,
            MUIA_List_DragSortable, TRUE,
            MUIA_List_DisplayHook, &fieldDisplayHook,
        End,
	End;

    d->d_NewField = SimpleButton(GCS(MSG_New));
    d->d_CopyField = SimpleButton(GCS(MSG_StructureEditor_Field_Copy));
    d->d_DeleteField = SimpleButton(GCS(MSG_StructureEditor_Field_Delete));
    d->d_SortFields = SimpleButton(GCS(MSG_StructureEditor_Sort1));

    fieldActions = HGroup,
    	MUIA_Group_SameHeight, TRUE,
    	Child, d->d_NewField,
    	Child, d->d_CopyField,
    	Child, d->d_DeleteField,
    	Child, d->d_SortFields,
	End;

    d->d_FieldsGroup = VGroup,
    	MUIA_Disabled, TRUE,
    	Child, d->d_Fields,
    	Child, fieldActions,
	End;

    tabFieldHor = VGroup,
    	Child, d->d_TablesGroup,
        Child, BalanceObject, End,
    	Child, VGroup,
    	    MUIA_Weight, 70,
    	    MUIA_Frame, MUIV_Frame_Group,
    	    MUIA_FrameTitle, GCS(MSG_Fields),
            MUIA_Background, MUII_GroupBack,
            MUIA_HelpNode, "Field management",
            Child, d->d_FieldsGroup,
        End,
	End;

    tabFieldVert = VGroup,
    	MUIA_Weight, 50,
    	Child, tabFieldHor,
	End;

    d->d_DisplayType = CycleObject,
        MUIA_CycleChain, TRUE,
    	MUIA_ControlChar, Shortcut(GCS(MSG_Display)),
    	MUIA_Cycle_Entries, displayTypeContent,
	End;

    d->d_Add = SimpleButton(GCS(MSG_StructureEditor_Add));
    d->d_Rem = SimpleButton(GCS(MSG_StructureEditor_Rem));
    d->d_Up = SimpleButton(GCS(MSG_StructureEditor_Up));
    d->d_Down = SimpleButton(GCS(MSG_StructureEditor_Down));
    d->d_In = SimpleButton(GCS(MSG_StructureEditor_In));
    d->d_Out = SimpleButton(GCS(MSG_StructureEditor_Out));
    d->d_Text = SimpleButton(GCS(MSG_StructureEditor_Text));
    d->d_Image = SimpleButton(GCS(MSG_Image));
    d->d_Space = SimpleButton(GCS(MSG_StructureEditor_Space));
    d->d_Balance = SimpleButton(GCS(MSG_StructureEditor_Balance));
    d->d_Group = SimpleButton(GCS(MSG_StructureEditor_Group));
    d->d_Register = SimpleButton(GCS(MSG_StructureEditor_Register));
    d->d_Window = SimpleButton(GCS(MSG_StructureEditor_Window));

    displayButtons = VGroup,
    	MUIA_Weight, 0,
    	Child, VSpace(0),
    	Child, d->d_Add,
    	Child, d->d_Rem,
    	Child, d->d_Up,
    	Child, d->d_Down,
    	Child, d->d_In,
    	Child, d->d_Out,
    	Child, VSpace(0),
    	Child, d->d_Text,
    	Child, d->d_Image,
    	Child, d->d_Space,
    	Child, d->d_Balance,
    	Child, d->d_Group,
    	Child, d->d_Register,
    	Child, d->d_Window,
    	Child, VSpace(0),
	End;

    /*
    ** For the display objects we use the NList class (if USE_NLIST is defined).
    ** On AOS4 and AROS, BeeBase crashes when using List (or DDList).
    ** It is unknown if this is a bug in the List class or if
    ** there sits some bug somewhere in this module.
    */
    d->d_DisplayObjects = MListviewObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_DisplayBH),
	MUIA_Listview_MultiSelect, MUIV_Listview_MultiSelect_Default,
	MUIA_MListview_MList, MListObject,
    	    MUIA_Frame, MUIV_Frame_InputList,
            MUIA_List_DisplayHook, &displayDisplayHook,
            MUIA_List_Format, ",",
	    End,
	End;

    d->d_DisplayObjectsGroup = HGroup,
    	MUIA_Disabled, TRUE,
    	Child, displayButtons,
    	Child, d->d_DisplayObjects,
	End;

    displayVert = VGroup,
    	MUIA_Weight, 100,
    	MUIA_Frame, MUIV_Frame_Group,
    	MUIA_FrameTitle, GCS(MSG_Display),
        MUIA_Background, MUII_GroupBack,
        MUIA_HelpNode, "Display field",
    	Child, d->d_DisplayType,
    	Child, d->d_DisplayObjectsGroup,
	End;

    root = HGroup,
    	Child, tabFieldVert,
        Child, BalanceObject, End,
    	Child, displayVert,
	End;

    se = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
    	MUIA_Window_ScreenTitle, GuiScreenTitle,
    	MUIA_Window_ID, MAKE_ID('_', 'S','E','D'),
        MUIA_HelpNode, "Structure editor",
        MUIA_UserData, d,
    	WindowContents, root,
	End;

    if(se == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct StructureEditorData));
	return(NULL);
    }

    SetRootWindowTitle(pr, se);
    SetAttrs(d->d_NewTable,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_Table_NewBH),
        TAG_DONE);

    SetAttrs(d->d_DeleteTable,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_Table_DeleteBH),
        MUIA_Disabled, TRUE,
        TAG_DONE);

    SetAttrs(d->d_SortTables,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_Table_SortBH),
        TAG_DONE);

    SetAttrs(d->d_NewField,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_Field_NewBH),
        TAG_DONE);

    SetAttrs(d->d_CopyField,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_Field_CopyBH),
        MUIA_Disabled, TRUE,
        TAG_DONE);

    SetAttrs(d->d_DeleteField,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_Field_DeleteBH),
        MUIA_Disabled, TRUE,
        TAG_DONE);

    SetAttrs(d->d_SortFields,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_Field_SortBH),
        TAG_DONE);

    SetAttrs(d->d_Add,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_AddBH2),
        MUIA_CycleChain, TRUE,
        TAG_DONE);
    SetAttrs(d->d_Rem,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_RemBH),
        TAG_DONE);
    SetAttrs(d->d_Up,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_UpBH),
        TAG_DONE);
    SetAttrs(d->d_Down,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_DownBH),
        TAG_DONE);
    SetAttrs(d->d_In,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_InBH),
        TAG_DONE);
    SetAttrs(d->d_Out,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_OutBH),
        TAG_DONE);
    SetAttrs(d->d_Text,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_TextBH),
        TAG_DONE);
    SetAttrs(d->d_Image,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_ImageBH),
        TAG_DONE);
    SetAttrs(d->d_Space,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_SpaceBH),
        TAG_DONE);
    SetAttrs(d->d_Balance,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_BalanceBH),
        TAG_DONE);
    SetAttrs(d->d_Group,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_GroupBH),
        TAG_DONE);
    SetAttrs(d->d_Register,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_RegisterBH),
        TAG_DONE);
    SetAttrs(d->d_Window,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_StructureEditor_WindowBH),
        TAG_DONE);

    MUIAddCallback2(d->d_Tables, MUIA_List_Active, MUIV_EveryTime,
        se, tablesActiveChange, NULL);
    MUIAddCallback2(d->d_Tables, MUIA_DDList_Dropped, TRUE,
        se, tablesDropped, NULL);
    MUIAddCallback2(d->d_Tables, MUIA_Listview_DoubleClick, TRUE,
        se, changeTableClick, NULL);

    MUIAddCallback2(d->d_NewTable, MUIA_Pressed, FALSE,
        se, newTableClick, NULL);
    MUIAddCallback2(d->d_DeleteTable, MUIA_Pressed, FALSE,
        se, deleteTableClick, NULL);
    MUIAddCallback2(d->d_SortTables, MUIA_Pressed, FALSE,
        se, sortTablesClick, NULL);

    MUIAddCallback2(d->d_Fields, MUIA_List_Active, MUIV_EveryTime,
        se, fieldsActiveChange, NULL);
    MUIAddCallback2(d->d_Fields, MUIA_DDList_Dropped, TRUE,
        se, fieldsDropped, NULL);
    MUIAddCallback2(d->d_Fields, MUIA_Listview_DoubleClick, TRUE,
        se, changeFieldClick, NULL);

    MUIAddCallback2(d->d_NewField, MUIA_Pressed, FALSE,
        se, newFieldClick, FALSE);
    MUIAddCallback2(d->d_CopyField, MUIA_Pressed, FALSE,
        se, newFieldClick, TRUE);
    MUIAddCallback2(d->d_DeleteField, MUIA_Pressed, FALSE,
        se, deleteFieldClick, NULL);
    MUIAddCallback2(d->d_SortFields, MUIA_Pressed, FALSE,
        se, sortFieldsClick, NULL);

    MUIAddCallback2(d->d_DisplayType, MUIA_Cycle_Active, MUIV_EveryTime,
        se, displayCycleChange, NULL);
    MUIAddCallback2(d->d_DisplayObjects, MUIA_List_Active, MUIV_EveryTime,
        se, displayActiveChange, NULL);
    MUIAddCallback2(d->d_DisplayObjects, MUIA_Listview_DoubleClick, TRUE,
        se, changeDisplayClick, NULL);

    MUIAddCallback2(d->d_Add, MUIA_Pressed, FALSE, se, addClick, NULL);
    MUIAddCallback2(d->d_Rem, MUIA_Pressed, FALSE, se, remClick, NULL);
    MUIAddCallback2(d->d_Up, MUIA_Pressed, FALSE, se, upClick, NULL);
    MUIAddCallback2(d->d_Down, MUIA_Pressed, FALSE, se, downClick, NULL);
    MUIAddCallback2(d->d_In, MUIA_Pressed, FALSE, se, inClick, NULL);
    MUIAddCallback2(d->d_Out, MUIA_Pressed, FALSE, se, outClick, NULL);
    MUIAddCallback2(d->d_Text, MUIA_Pressed, FALSE, se, textClick, NULL);
    MUIAddCallback2(d->d_Image, MUIA_Pressed, FALSE, se, imageClick, NULL);
    MUIAddCallback2(d->d_Space, MUIA_Pressed, FALSE, se, spaceClick, NULL);
    MUIAddCallback2(d->d_Balance, MUIA_Pressed, FALSE, se, balanceClick, NULL);
    MUIAddCallback2(d->d_Group, MUIA_Pressed, FALSE, se, groupClick, NULL);
    MUIAddCallback2(d->d_Register, MUIA_Pressed, FALSE, se, registerClick, NULL);
    MUIAddCallback2(d->d_Window, MUIA_Pressed, FALSE, se, windowClick, NULL);

    MUIAddCallback2(se, MUIA_Window_CloseRequest, TRUE,
        se, closeStructureEditor, NULL);

    setTables(se);

    MUIAddMember(app, se);
    if(!GuiOpenWindow(se))
    {
        MUIRemMember(app, se);
        disposeStructureEditor(app, se);
        return(NULL);
    }

    set(se, MUIA_Window_ActiveObject, d->d_Tables);

    return se;
}

void StructureEditorUpdate(GuiObject se)
{
    if (se) {
        setTables(se);
        setDisplay(se);
    }
}

void ClearStructureEditor(GuiObject se)
{
    struct StructureEditorData *d = pget(se, MUIA_UserData);
    struct MemPool *memPool = ProjectMemPool(d->d_Project);

    d->d_CurrentTable = NULL;
    MUIListSetList(d->d_DisplayObjects, NULL, memPool);
    MUIListSetList(d->d_Fields, NULL, memPool);
    MUIListSetList(d->d_Tables, NULL, memPool);
}

void StructureEditorSetNameChangeObject(GuiObject se, void *object)
{
    struct StructureEditorData *d = pget(se, MUIA_UserData);

    if (d != NULL) {
        d->d_NameChangeObject = object;
    }
}

#endif

