/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** Table_mui.c
**
** Creation date: 20.11.1994
** Last modified: 13.05.2012
*/

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "structureeditor.h"
#include "project.h"
#include "table.h"
#include "display.h"
#include "requesters.h"
#include "error.h"
#include "appstrings_cat.h"
#include "error.h"
#include "memory.h"
#include "util.h"
#include "structureeditor_mui.h"
#include "extclasses/ExtApplication_mcc.h"

static const char *numRecordsContent[3];

struct EditTableData
{
    struct Project *d_Project;
    struct Table *d_Table;
    struct DisplayObject *d_Parent;
    Object *d_StructureEditor;
    Object *d_Name, *d_NumRecords;
    Object *d_NewTrigger, *d_NewTriggerStr;
    Object *d_DelTrigger, *d_DelTriggerStr;
    Object *d_CountChanges;
    Object *d_LogChanges;
    Object *d_Ok, *d_Cancel;
};

static void disposeEditTable(Object *app, Object *obj)
{
    struct EditTableData *d = pget(obj, MUIA_UserData);

    if(obj && d)
    {
        PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(struct EditTableData));
        MUI_AddDisposeObject(app, obj);
    }
}

static void HOOK_PROTO(finish, Object *obj, ulong *ok)
{
    struct EditTableData *d = pget(obj, MUIA_UserData);
    struct Project *pr = d->d_Project;
    Object *app = pget(obj, MUIA_ApplicationObject);
    struct TableData td;
    long error = OK;
    struct Table *new = NULL;

    if(*ok)
    {
     	td.td_Name = pget(d->d_Name, MUIA_String_Contents);
 	td.td_System = FALSE;
        td.td_ExactlyOneRecord = (nget(d->d_NumRecords, MUIA_Radio_Active)== 1);
     	td.td_NewTrigger = pget(d->d_NewTriggerStr, MUIA_String_Contents);
     	td.td_DelTrigger = pget(d->d_DelTriggerStr, MUIA_String_Contents);
        td.td_CountChanges = nget(d->d_CountChanges, MUIA_Selected);
        td.td_LogChanges = nget(d->d_LogChanges, MUIA_Selected);

        GuiSetSleep(TRUE);
        if(d->d_Table == NULL)
        {
    	    error = SE_NewTable(pr, &td, d->d_Parent, &new);
            if(error == OK)
                NewTableFinish(d->d_StructureEditor, new);
        }
        else
        {
	    error = SE_ChangeTable(d->d_Table, &td);
            if(error == OK)
                ChangeTableFinish(d->d_StructureEditor, d->d_Table);
        }
        GuiSetSleep(FALSE);

    	if(error)
    	{
	    Ask2(obj, GCS(MSG_ErrorMessage), AT_ERROR,
		    GCS(MSG_StarOk), GES(error));
	    set(obj, MUIA_Window_ActiveObject, d->d_Name);
            return;
        }
    }

    GuiCloseWindow(obj);
    MUIRemMember(app, obj);
    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);

    disposeEditTable(app, obj);
}

static Object *createEditTable(Object *se, struct Project *pr, struct Table *tab,
        struct DisplayObject *parent)
{
    Object *app = GuiApplication;
    struct MemPool *memPool = ProjectMemPool(pr);
    struct EditTableData *d;
    Object *root, *top, *numRecords, *triggers, *buttons;
    Object *obj;

    if(app == NULL)
        return(NULL);

    d = PoolAllocMemZero(memPool, sizeof(struct EditTableData));
    if(d == NULL)
        return(NULL);
    d->d_Project = pr;
    d->d_Table = tab;
    d->d_Parent = parent;
    d->d_StructureEditor = se;

    numRecordsContent[0] = GCS(MSG_Unlimited);
    numRecordsContent[1] = GCS(MSG_Table_ExactlyOne);
    numRecordsContent[2] = NULL;

    d->d_Name = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_Table_NameBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_Name)),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_String_MaxLen, ID_NAME_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    top = HGroup,
	Child, Label2(GCS(MSG_Name)),
	Child, d->d_Name,
	End;

    d->d_NumRecords = RadioObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_Table_NumberOfRecordsBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_Table_NumberOfRecords)),
	MUIA_Radio_Entries, numRecordsContent,
	End;

    numRecords = HGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_Table_NumberOfRecords),
        MUIA_Background, MUII_GroupBack,
	Child, HSpace(0),
	Child, d->d_NumRecords,
	Child, HSpace(0),
	End;

    d->d_NewTriggerStr = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_Table_Trigger_NewBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_Table_Trigger_New)),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_String_MaxLen, TABLE_TRIGGER_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    d->d_NewTrigger = UserFunctionsPopup(d->d_NewTriggerStr, pr);

    d->d_DelTriggerStr = StringObject,
        MUIA_CycleChain, TRUE,
        MUIA_ShortHelp, GCS(MSG_Table_Trigger_DeleteBH),
	MUIA_ControlChar, Shortcut(GCS(MSG_Table_Trigger_Delete)),
	MUIA_Frame, MUIV_Frame_String,
	MUIA_String_MaxLen, TABLE_TRIGGER_LEN + 1,
	MUIA_String_Format, MUIV_String_Format_Center,
	End;

    d->d_DelTrigger = UserFunctionsPopup(d->d_DelTriggerStr, pr);

    triggers = VGroup,
	MUIA_Frame, MUIV_Frame_Group,
	MUIA_FrameTitle, GCS(MSG_Table_TriggerFunctions),
        MUIA_Background, MUII_GroupBack,
	MUIA_Group_Columns, 2,
	Child, Label(GCS(MSG_Table_Trigger_New)),
	Child, d->d_NewTrigger,
	Child, Label(GCS(MSG_Table_Trigger_Delete)),
	Child, d->d_DelTrigger,
	End;

    d->d_CountChanges = MUI_MakeObject(MUIO_Checkmark,
            (IPTR)GCS(MSG_CountChanges));
    if (d->d_CountChanges) {
        SetAttrs(d->d_CountChanges,
                MUIA_ShortHelp, GCS(MSG_Table_CountChangesBH),
                MUIA_CycleChain, TRUE,
                MUIA_Selected, TRUE,
                TAG_DONE);
    }
    d->d_LogChanges = MUI_MakeObject(MUIO_Checkmark,
            (IPTR)GCS(MSG_LogChanges));
    if (d->d_LogChanges) {
        SetAttrs(d->d_LogChanges,
                MUIA_ShortHelp, GCS(MSG_Table_LogChangesBH),
                MUIA_CycleChain, TRUE,
                MUIA_Selected, TRUE,
                TAG_DONE);
    }

    d->d_Ok = SimpleButton(GCS(MSG_Ok));
    d->d_Cancel = SimpleButton(GCS(MSG_Cancel));

    buttons = HGroup,
	MUIA_Group_SameWidth, TRUE,
	Child, d->d_Ok,
	Child, d->d_Cancel,
	End;

    root = VGroup,
        Child, top,
        Child, numRecords,
        Child, triggers,
        Child, HGroup,
            Child, Label(GCS(MSG_CountChanges)),
            Child, VCenter(d->d_CountChanges),
            Child, HSpace(0),
            Child, Label(GCS(MSG_LogChanges)),
            Child, VCenter(d->d_LogChanges),
            End,
        Child, VSpace(8),
        Child, buttons,
        End;

    obj = ExtWindowObject,
        MUIA_ExtWindow_Project, pr,
	MUIA_Window_ID, MAKE_ID('_', 'T','A','B'),
        MUIA_Window_NoMenus, TRUE,
        MUIA_Window_Title, (tab? GCS(MSG_ChangeTable) : GCS(MSG_NewTable)),
        MUIA_Window_ScreenTitle, GuiScreenTitle,
        MUIA_HelpNode, tab? "Changing tables" : "Creating tables",
        MUIA_UserData, d,
	WindowContents, root,
	End;

    if(obj == NULL)
    {
        PoolFreeMem(memPool, d, sizeof(struct EditTableData));
	return(NULL);
    }

    set(d->d_Ok, MUIA_CycleChain, TRUE);
    set(d->d_Cancel, MUIA_CycleChain, TRUE);

    DoMethod(d->d_Name,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    DoMethod(d->d_NewTriggerStr,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);
    DoMethod(d->d_DelTriggerStr,
	MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
	obj, 3, MUIM_Set, MUIA_Window_ActiveObject, d->d_Ok);

    MUIAddPushCallback2(obj, MUIA_Window_CloseRequest, TRUE, obj, finish,FALSE);
    MUIAddPushCallback2(d->d_Cancel, MUIA_Pressed, FALSE, obj, finish, FALSE);
    MUIAddPushCallback2(d->d_Ok, MUIA_Pressed, FALSE, obj, finish, TRUE);

    if(tab != NULL)
    {
        set(d->d_Name, MUIA_String_Contents, TableName(tab));
        set(d->d_NumRecords, MUIA_Radio_Active, TableExactlyOneRecord(tab)? 1 : 0);
        set(d->d_NewTriggerStr, MUIA_String_Contents, TableNewTrigger(tab));
        set(d->d_DelTriggerStr, MUIA_String_Contents, TableDelTrigger(tab));
        set(d->d_CountChanges, MUIA_Selected, TableCountChanges(tab));
        set(d->d_LogChanges, MUIA_Selected, TableLogChanges(tab));
    }

    DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, TRUE);
    MUIAddMember(app, obj);
    if(!GuiOpenWindow(obj))
    {
        MUIRemMember(app, obj);
        DoMethod(app, MUIM_ExtApplication_ProjectSleep, pr, FALSE);
        disposeEditTable(app, obj);
        return(NULL);
    }

    set(obj, MUIA_Window_ActiveObject, d->d_Name);

    return(obj);
}

void GuiNewTable(Object *se, struct Project *pr, struct DisplayObject *dioparent)
{
    createEditTable(se, pr, NULL, dioparent);
}

void GuiChangeTable(Object *se, struct Table *tab)
{
    createEditTable(se, TableProject(tab), tab, NULL);
}

#endif


