/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** buildgui.c
**
** Creation date: 09.01.2005
** Last modified: 09.03.2021
*/

#include "buildgui.h"
#include "gui.h"
#include "order.h"
#include "search.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

struct Record *GuiReferenceFieldQuickSearchRecord(struct DisplayObject *dio,
	GuiObject list, long dir, const char *search, struct Record *cur)
{
    static struct SearchFor sf = { 0 };
    struct Field *field = DisplayFieldField(dio);
    struct Table *tab = FieldReferenceTable(field);
    struct Record *rec = cur;
    bool all = TRUE;
    int avail = sizeof(sf.sf_Pattern) - 3;
    long num = -1;
    
    if (search == NULL || search[0] == 0) {
        return cur;                     /* don't search empty string */
    }

    sf.sf_Table = tab;
    sf.sf_Field = NULL;
    strcpy(sf.sf_Pattern, "*");
    CodesetStrNBytesCopy(sf.sf_Pattern + 1, search, avail);
    strcat(sf.sf_Pattern, "*");
    sf.sf_CaseSensitive = FALSE;
    sf.sf_AllFields = all;
    sf.sf_Forward = (dir >= 0);
    sf.sf_FirstLastRecord = (dir == 0);
    sf.sf_MatchFilter = FALSE;

    do {
	if (SearchRecord(&sf, &rec, rec) != NULL || rec == NULL) {
	    break;
	}
	sf.sf_FirstLastRecord = FALSE;
    } while (list != NULL && (num = GuiListFindNode(list, rec)) < 0);

    if (rec != NULL || dir == 0) {
        cur = rec;
        if (list) {
            GuiListSetActive(list, num);
        } else {
            struct Project *pr = DisplayObjectProject(dio);
            struct MemPool *memPool = ProjectMemPool(pr);
            GuiObject go = DisplayObjectGuiObject(dio);
            char *s, *display;
            bool free = FALSE;
            long error;

            RecordLock(rec);
            error = GuiFieldReferenceDataString(field, rec, memPool, 
                    &s, &free);
            RecordUnlock(rec);

            display = error? GES(error) : s;
            GuiSetString(go, display, TRUE, NULL);
            if (free) {
                FreeString(memPool, s);
            }
        }
    }

    return cur;
}

#if 0
/* displays search string in reference field while searching */
static char *displaySearchRecord(struct DisplayObject *dio, 
	struct SearchFor *sf, struct Record **prec, struct Record *start)
{
    struct Field *field = DisplayFieldField(dio);
    GuiObject go = DisplayObjectGuiObject(dio);
    char *result;

    if (dio && sf) {
	GuiSetString(go, sf->sf_Pattern, FALSE, TRUE, NULL);
    }
    result = SearchRecord(sf, prec, start);
    if (prec == NULL || *prec == NULL) {
	GuiUpdateField(field, TableGuiRecord(FieldTable(field)));
    }

    return result;
}
#endif

bool BuildGuiFieldGetFeatures(struct DisplayObject *dio,
	struct GuiFieldFeatures *f)
{
    struct Project *pr = DisplayObjectProject(dio);
    struct Field *field = DisplayFieldField(dio);
    const char *title = DisplayObjectTitle(dio);
    short type = FieldType(field);
    char subtype = DisplayFieldSubType(dio);
  
    if (field == NULL || f == NULL) {
	return FALSE;
    }

    memset(f, 0, sizeof(struct GuiFieldFeatures));
    f->maxchars = -1;
    f->readonly = DisplayFieldReadOnly(dio);
    f->shortcut = f->readonly? 0 : DisplayObjectShortcut(dio);

    switch (type) {
	case TYPE_STRING:
	    f->label = TRUE;
            f->maxchars = FieldStringMaxlen(field);
            if (DisplayFieldDisplayExternal(dio)) {
                f->datatypes = TRUE;
                if (DisplayFieldHideStringField(dio)) {
                    f->maxchars = -1;
                }
                if (DisplayFieldTitleAtStringField(dio)) {
                    f->label = FALSE;
                }
            }
	    break;

	case TYPE_INTEGER:
	    f->label = TRUE;
            f->maxchars = 20;
	    break;

	case TYPE_REAL:
            f->label = TRUE;
            f->maxchars = 30;
	    break;

	case TYPE_BOOL:
            f->label = TRUE;
            f->checkbox = TRUE;
	    break;

	case TYPE_CHOICE:
            if (f->readonly) {
                f->label = TRUE;
                f->maxchars = 20;
            } else if (subtype == DAD_CHOICETYPE_RADIOBUTTONS) {
                f->radiobuttons = TRUE;
            } else {
                f->choice = TRUE;
                f->label = TRUE;
            }
	    break;

	case TYPE_DATE:
	    f->label = TRUE;
            f->maxchars = DATE_STR_LEN;
            break;

	case TYPE_TIME:
	    f->label = TRUE;
	    f->maxchars = TIME_STR_MAXLEN;
            break;

	case TYPE_MEMO:
            f->label = TRUE;
	    f->editor = TRUE;
            break;

  	case TYPE_BUTTON:
             if (subtype == DAD_BUTTONTYPE_TEXT ||
                     !GuiCanLoadImageSpec(pr, DisplayObjectBackground(dio))) {
                f->button = TRUE;
             } else {
                f->image = TRUE;
             }
             break;

        case TYPE_REFERENCE:
            if (!f->readonly) {
                f->refpopup = TRUE;
            }
            f->readonly = TRUE;
            f->show = DisplayFieldShow(dio);
	    f->label = TRUE;
            f->maxchars = 100;
            break;

        case TYPE_VIRTUAL:
            f->label = TRUE;
            f->readonly = TRUE;
            switch (subtype) {
                case DAD_VIRTUALTYPE_BOOL:  f->checkbox = TRUE;
                                            break;
                case DAD_VIRTUALTYPE_TEXT:  f->text = TRUE;     
                                            break;
                case DAD_VIRTUALTYPE_LIST:  f->listview = TRUE;
                                            break;
                default:                    return FALSE;
            }
            break;

        default:
            return FALSE;               /* unknown type */
    }

    if (title == NULL || title[0] == 0) {
	f->label = FALSE;
    }

    return TRUE;
}

void BuildGuiDisplayObjectGetSizeConstraints(struct DisplayObject *dio,
        bool horiz, struct GuiDisplayObjectSizeConstraints *constraints)
{
    struct List *childs = DisplayObjectObjects(dio);

    if (constraints == NULL) {
        return;
    }
    memset(constraints, 0, sizeof(*constraints));
    if (dio == NULL) {
        return;
    }

    switch (DisplayObjectType(dio)) {
        case DO_TYPE_FIELD:
        {
            struct Field *field = DisplayFieldField(dio);
            short type = FieldType(field);
            char subtype = DisplayFieldSubType(dio);

            constraints->can_grow_horiz = TRUE;

            switch (type) {
                case TYPE_STRING:
                    if (DisplayFieldDisplayExternal(dio)) {
                        if (DisplayFieldImageSize(dio) != 
                                DAD_IMAGESIZE_FIXED) {
                            constraints->can_grow_vert = TRUE;
                        } else {
                            constraints->can_grow_horiz = FALSE;
                        }
                    }
                    break;

                case TYPE_INTEGER:
                case TYPE_REAL:
                case TYPE_DATE:
                case TYPE_TIME:
                case TYPE_REFERENCE:
                    break;

                case TYPE_BOOL:
                    constraints->can_grow_horiz = FALSE;
                    break;

                case TYPE_CHOICE:
                    if (horiz && !DisplayFieldReadOnly(dio) &&
                            subtype == DAD_CHOICETYPE_RADIOBUTTONS) {
                        constraints->can_grow_horiz = FALSE;
                    }
                    break;

                    break;

                case TYPE_MEMO:
                    constraints->can_grow_vert = TRUE;
                    break;

                case TYPE_BUTTON:
                     if (subtype == DAD_BUTTONTYPE_IMAGE) {
                        if (DisplayFieldImageSize(dio) != 
                                DAD_IMAGESIZE_FIXED) {
                            constraints->can_grow_vert = TRUE;
                        } else {
                            constraints->can_grow_horiz = FALSE;
                        }
                     }
                     break;

                case TYPE_VIRTUAL:
                    switch (subtype) {
                        case DAD_VIRTUALTYPE_BOOL: 
                            constraints->can_grow_horiz = FALSE;
                            break;
                        case DAD_VIRTUALTYPE_LIST:
                            constraints->can_grow_vert = TRUE;
                            break;
                    }
                    break;
            }
            break;
        }

        case DO_TYPE_SPACE:
            constraints->can_grow_horiz = DisplayObjectWeight(dio) > 0;
            constraints->can_grow_vert = constraints->can_grow_horiz;
            break;

        case DO_TYPE_TEXT:
        case DO_TYPE_BALANCE:
            break;

        case DO_TYPE_IMAGE:
            if (DisplayImageImageSize(dio) != DAD_IMAGESIZE_FIXED) {
                constraints->can_grow_horiz = TRUE;
                constraints->can_grow_vert = TRUE;
            }
            break;

        case DO_TYPE_GROUP:
            horiz = DisplayGroupHoriz(dio);
            break;

        case DO_TYPE_TABLE:
        case DO_TYPE_WINDOW:
        case DO_TYPE_REGISTER:
            horiz = FALSE;
            break;
    }

    if (childs) {
        struct GuiDisplayObjectSizeConstraints sub;
        FORNODE(dio, childs) {
            if (DisplayObjectType(dio) == DO_TYPE_WINDOW) {
                constraints->can_grow_horiz = TRUE;
            } else {
                BuildGuiDisplayObjectGetSizeConstraints(dio, horiz, &sub);
                constraints->can_grow_horiz |= sub.can_grow_horiz;
                constraints->can_grow_vert |= sub.can_grow_vert;
            }
        }
    }
}

bool BuildGuiFieldHasLabel(struct DisplayObject *dio)
{
    struct GuiFieldFeatures f;

    return (bool)(BuildGuiFieldGetFeatures(dio, &f) && f.label);
}

void SetRootWindowTitle(struct Project *pr, GuiObject wi)
{
    struct DisplayObject *root = ProjectRootDisplayObject(pr);

    if (DisplayObjectGuiObject(root) == wi &&
            FindUserFunctionName(ProjectProgram(pr), mainWindowTitle) &&
            GuiUpdateComputedDisplayObject(root)) {
        // window title is set through program function
    } else {
        GuiSetWindowTitle(wi, ProjectTitle(pr));
    }
    SetProjectPrependTitle(pr);
}

static bool BuildGuiWindowsRecursive(struct DisplayObject *dio)
{
    struct List *childs = DisplayObjectObjects(dio);

    if (DisplayObjectType(dio) == DO_TYPE_WINDOW) {
        if (DisplayObjectParent(dio) == NULL || DisplayObjectVisibleGui(dio)) {
            if (!BuildGuiWindow(dio)) {
                return FALSE;
            }
        }
    }
    if (childs) {
        FORNODE(dio, childs) {
            if (!BuildGuiWindowsRecursive(dio)) {
                return FALSE;
            }
        }
    }
    return TRUE;
}

bool BuildProjectGui(struct Project *pr)
{
    struct DisplayObject *dio = ProjectRootDisplayObject(pr);

    if (dio == NULL) {
	return FALSE;
    }
    if (!BuildGuiWindowsRecursive(dio)) {
        FreeProjectGui(pr);
        return FALSE;
    }
    return TRUE;
}

