/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** codeset.h
**
** Codeset stands for the default encoding of a character string.
** On Amiga this is an 8bit code set using the system's default encoding.
** On all other systems, the default encoding is UTF8.
** This module handles the correct handling for all systems.
**
** Creation date: 22.04.2007
** Last modified: 29.01.2009
*/

#ifndef CODESET_H
#define CODESET_H

#include <system.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifdef AMIGA

    #include <libraries/locale.h>

    #define CS_COLLATE_SIMPLE SC_COLLATE1  /* higher chance for equal strings */
    #define CS_COLLATE_ORDER  SC_COLLATE2  /* finer matches */

    #define CS_UTF8 FALSE

    #define CodesetMaxBytesForChar 1

    typedef unsigned char cschar;

    #define LocaleIsUTF8() FALSE

    #define CodesetGetChar(s) ((cschar)(*(s)))
    #define CodesetNextChar(s) ((s) + 1)
    #define CodesetStrLen(s) strlen(s)
    #define CodesetNCharsToNBytes(s, n) (n) /* no of bytes for n chars of s */
    #define CodesetNBytesToNChars(s, n) (n) /* no of chars for n bytes of s */

    #define CodesetStrpbrk		strpbrk
    #define CodesetStrchr(s, c)		strchr(s, (char)c)
    #define CodesetStrrchr(s, c)	strrchr(s, (char)c)

#elif (defined GTK)

    #define CS_COLLATE_SIMPLE 0		  /* currently not distinguished */
    #define CS_COLLATE_ORDER  1

    #define CS_UTF8 TRUE

    #define CodesetMaxBytesForChar 6	  /* according to g_unichar_to_utf8 */

    typedef ulong cschar;		  /* should be same as guint32 */

    bool LocaleIsUTF8(void);              /* TRUE if locale encoding is UTF-8 */

    cschar CodesetGetChar(const char *s); /* returns unichar at s */
    cschar CodesetGetCharCheck(const char *s); /* returns unichar at s or 0 */
    char *CodesetNextChar(const char *s); /* returns pointer at next char */
    size_t CodesetStrLen(const char *s);  /* length in characters */
    size_t CodesetNCharsToNBytes(const char *s, size_t n); /* bytes for chars */
    size_t CodesetNBytesToNChars(const char *s, size_t n); /* chars for bytes */

    char *CodesetStrpbrk(const char *s, const char *acpt);/* case sensitive */
    char *CodesetStrchr(const char *s, cschar c);	  /* case sensitive */
    char *CodesetStrrchr(const char *s, cschar c);	  /* case sensitive */
    
#elif (defined QT)

#define CS_COLLATE_SIMPLE 0		  /* currently not distinguished */
#define CS_COLLATE_ORDER  1
    
#define CS_UTF8 TRUE
    
#define CodesetMaxBytesForChar 6
    
    typedef unsigned int cschar;
    
    bool LocaleIsUTF8(void);              /* TRUE if locale encoding is UTF-8 */
    
    cschar CodesetGetChar(const char *s); /* returns unichar at s */
    // cschar CodesetGetCharCheck(const char *s); /* returns unichar at s or 0 */
    char *CodesetNextChar(const char *s); /* returns pointer at next char */
    size_t CodesetStrLen(const char *s);  /* length in characters */
    size_t CodesetNCharsToNBytes(const char *s, size_t n); /* bytes for chars */
    size_t CodesetNBytesToNChars(const char *s, size_t n); /* chars for bytes */
    
    char *CodesetStrpbrk(const char *s, const char *acpt);/* case sensitive */
    char *CodesetStrchr(const char *s, cschar c);	  /* case sensitive */
    char *CodesetStrrchr(const char *s, cschar c);	  /* case sensitive */
    
#endif

/* initialization and cleanup */ 
const char *CodesetInit(void);
void CodesetExit(void);

/* 
** checks whether a given string conforms to a certain format.
** AllowTrail allows for incomplete (but not invalid) characters
** at the end of the string (e.g. the beginning of the next character).
** The end of the valid string is returned in *end.
*/
bool StrIsCodeset(const char *str);
bool StrIsUTF8AllowTrail(const char *str, const char **end);

/* 
** Conversion functions from/to the default codeset.
** The returned string is either NULL (fatal error),
** the input string itself (no conversion necessary)
** or a freshly allocated string from the given memory pool.
** An error code is returned if and only if the returned string is NULL.
** In warnings the number of characters causing conversion errors are returned
** (e.g. partial conversion with unknown codes replaced by '?').
*/
char *CodesetConvertFromUTF8(const char *in, struct MemPool *memPool, 
	long *error, size_t *warnings);
char *CodesetConvertFromUTF8AllowTrail(const char *in, struct MemPool *memPool, 
	long *error, size_t *warnings, const char **end);
char *CodesetConvertFromLocale(const char *in, struct MemPool *memPool,
	long *error, size_t *warnings);
char *CodesetConvertFromLocaleAllowTrail(const char *in,struct MemPool *memPool,
	long *error, size_t *warnings, const char **end);
char *CodesetConvertFromDefault8Bit(const char *in, struct MemPool *m, long *e);
char *CodesetConvertToUTF8(const char *in, struct MemPool *memPool, long *err);
char *CodesetConvertToLocale(const char *in, struct MemPool *memPool,
        long *err, size_t *warnings);
char *CodesetConvertToDefault8Bit(const char *in, struct MemPool *memPool,
        long *err, size_t *warnings);

/*
** Returns number of bytes a codeset string needs in UTF-8 without '\0'
*/
size_t CodesetUTF8Len(const char *s);

/* 
** Converts to UTF8 and writes string to file.
*/
void CodesetWriteUTF8(const char *s, FILE *file);

/* 
** Converts single char to codeset string and returns number of bytes written
** or -1 if the char does not constitute a valid character in the codeset.
** If out is NULL then only the length is computed.
** Not NUL terminated, string must hold at least CodesetMaxBytesForChar bytes
*/
short CodesetCharToStr(cschar c, char *out);

/* checking for character class */
bool CodesetIsAlNum(cschar c);
bool CodesetIsDigit(cschar c);
bool CodesetIsLower(cschar c);
bool CodesetIsPunct(cschar c);
bool CodesetIsSpace(cschar c);
bool CodesetIsBrkSpace(cschar c);    	/* breaking space only, exclude NBSP */
bool CodesetIsXDigit(cschar c);

/* digit values, -1 if not a valid digit */
short CodesetDigitValue(cschar c);
short CodesetXDigitValue(cschar c);

/* returns upper/lower version of character/string. */
cschar CodesetToUpper(cschar c);
cschar CodesetToLower(cschar c);
char *CodesetStrUp(const char *str, struct MemPool *memPool);
char *CodesetStrDown(const char *str, struct MemPool *memPool);

/*
** Copies at most N bytes from src to dst and cuts dst at char boundary.
** dst must be non-NULL. src might be NULL in which case dst is set to "".
** dst is always NUL terminated.  Returns dst.
*/
char *CodesetStrNBytesCopy(char *dst, const char *src, size_t nbytes);

/* 
** removes leading and trailing spaces/characters.  
** If front or back is NULLL than spaces are removed 
*/
char *CodesetTrimStr(char *s);
char *CodesetTrimStrChars(char *s, const char *front, const char *back);

/* 
** comparing strings case insensitive, collate as defined above 
** return code as in strcmp, i.e < 0, 0 and > 0.
*/
long CodesetStrNCollate(const char *s1, const char *s2, size_t bytes, 
	short collate);
long CodesetStrCollate(const char *s1, const char *s2, short collate);
long CodesetStrCmp(const char *s1, const char *s2); /* ASCII case insensitive */

/* finding sub strings and break characters */
char *CodesetStrStr(const char *hay, const char *needle); /* case insensitive */
char *CodesetStrPBrk(const char *s, const char *accept);  /* case insensitive */


#ifdef __cplusplus
}
#endif

#endif

