/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** display.c
**
** Creation date: 26.11.1994
** Last modified: 10.02.2022
*/

#include "system.h"
#include "main.h"
#include "labels.h"
#include "display.h"
#include "project.h"
#include "log.h"
#include "table.h"
#include "field.h"
#include "structureeditor.h"
#include "preferences.h"
#include "error.h"
#include "memory.h"
#include "dosutil.h"
#include "util.h"
#include "codeset.h"
#include "break.h"
#include "strhandle.h"
#include "lang/program.h"

#include "appstrings_cat.h"

#define DEFAULT_ROOT_WINDOW_ID 0

struct DisplayObject
{
    struct Node do_Node;
    struct DisplayObject *do_Parent;
    struct List do_Childs;
    struct Project *do_Project;
    char *do_Title;				/* Codeset */
    cschar do_Shortcut;
    char do_Enabled;
    char *do_EnabledCompute;
    char *do_MUIFontname;	                /* Codeset */
    char *do_GTKFontname;	                /* Codeset */
    char *do_QtFontname;	                /* Codeset */
    char *do_Background;	                /* Codeset */
    short do_Weight;
    GuiObject do_GuiObject;
    GuiObject do_RootObject;
    GuiObject do_ActivateObject;
    bool do_Visible;
    char do_Type;
    union
    {
	struct
	{
	    struct Table *table;
            ulong activeFieldNum;
            ulong activeDisplayObjectNum;
            bool hasPanel;
            struct 
            {
                char *background;
                bool numall;
                bool filter;
                bool arrows;
                GuiObject numall_go;
                GuiObject filter_go;
            } panel;
            struct RecordDescriptionSpec *recordDescriptionSpec;
 	} table;
	struct
	{
	    struct Field *field;
            char titlepos;
            char *titleMUIFontname;	        /* Codeset */
            char *titleGTKFontname;	        /* Codeset */
            char *titleQtFontname;	        /* Codeset */
            bool home;
            bool tabchain;
            bool readonly;
            char subType;
            char alignment;
            bool calendar;
            char format;
            char numdecimals;
            char imagesize;
            char *bubbleHelp;			/* Codeset */
            /* String */
            bool displayexternal;
            bool titleatstringfield;
            bool hidestringfield;
            bool file;
            bool font;
            bool listview;
            bool view;
            char **listviewLabels;		/* Codeset strings */
	    char *listviewCompute;		/* Codeset */
            /* Choice */
            bool frame;
            /* Reference */
            struct RecordDescriptionSpec *recordDescriptionSpec;
	    char recordsKind;
	    char *recordsCompute;		/* Codeset */
	    bool initialRecord;
	    bool currentRecord;
            bool show;
	    bool autoshow;
            bool filter;
            /* Virtual */
            bool autoUpdate;
            bool showtitles;
            bool controlTable;
	    char onDoubleClick;
	    char *doubleClickTrigger;		/* Codeset */
	    char *urlDropTrigger;		/* Codeset */
	    char *sortDropTrigger;		/* Codeset */
            /* others */
            GuiObject fileObject;
            GuiObject fontObject;
            GuiObject listviewObject;
            GuiObject viewObject;
            GuiObject externalObject;
            GuiObject showObject;
            GuiObject filterObject;
            GuiObject autoshowObject;
            GuiObject popObject;
            GuiObject disableObject;
            bool listActiveChanged;
	} field;
        struct
        {
            bool bar;
        } space;
        struct
        {
            char imagesize;
        } image;
	struct
	{
	    char **labels;			/* Codeset strings */
            bool open;
	} reggroup;
	struct
	{
            bool border;
	    bool horiz;
            bool spacing;
            bool open;
	} group;
	struct
	{
	    ulong id;
            bool button;
            bool closeWhenDisabled;
            bool open;
            bool windowopen;
            bool wd_valid;
            struct WindowDimensions wd;
	    GuiObject button_go;
            ulong activeTableNum;
            ulong activeDisplayObjectNum;
	} window;
    } do_Contents;
};


/*
** Enumeration of DisplayObjects
*/

static ulong DisplayObjectNumFind(struct DisplayObject *dio,
        struct DisplayObject *find, ulong *num)
{
    struct List *childs = DisplayObjectObjects(dio);
    ulong result = 0;

    ++*num;
    if (dio == find) {
	result = *num;
    } else if (childs) {
	FORNODE (dio, childs) {
	    result = DisplayObjectNumFind(dio, find, num);
	    if (result) {
		break;
	    }
	}
    }
    return result;
}

ulong DisplayObjectNum(struct DisplayObject *dio)
{
    ulong num = 0;

    if (dio) {
	struct DisplayObject *rootdio=ProjectRootDisplayObject(dio->do_Project);

	if (rootdio) {
	    num = DisplayObjectNumFind(rootdio, dio, &num);
	}
    }
    return num;
}

static struct DisplayObject *FindDisplayObjectNumRec(struct DisplayObject *dio,
        ulong find, ulong *num)
{
    struct List *childs = DisplayObjectObjects(dio);
    struct DisplayObject *found = NULL;

    ++*num;
    if (*num == find) {
	found = dio;
    } else if (childs) {
	FORNODE (dio, childs) {
	    found = FindDisplayObjectNumRec(dio, find, num);
            if (found) {
		break;
	    }
        }
    }
    return found;
}

struct DisplayObject *FindDisplayObjectNum(struct Project *pr, ulong find)
{
    struct DisplayObject *rootdio = ProjectRootDisplayObject(pr), *dio = NULL;
    ulong num = 0;

    if (rootdio) {
    	dio = FindDisplayObjectNumRec(rootdio, find, &num);
    }
    return dio;
}

struct Project *DisplayObjectProject(struct DisplayObject *dio)
{
    struct Project *pr = NULL;

    if (dio) {
        pr = dio->do_Project;
    }
    return pr;
}

short DisplayObjectType(struct DisplayObject *dio)
{
    short type = 0;

    if (dio) {
	type = dio->do_Type;
    }
    return type;
}

struct List *DisplayObjectObjects(struct DisplayObject *dio)
{
    struct List *childs = NULL;

    if (dio) {
	switch (dio->do_Type) {
	    case DO_TYPE_TABLE:
	    case DO_TYPE_REGISTER:
	    case DO_TYPE_GROUP:
	    case DO_TYPE_WINDOW:    
		childs = &dio->do_Childs;
		break;
	}
    }
    return childs;
}

/* depth first search */
struct DisplayObject *NextDisplayObject(struct DisplayObject *dio)
{
    struct List *dios = DisplayObjectObjects(dio);

    if (dios && SUCC(HEAD(dios))) {
        return (struct DisplayObject *)HEAD(dios);
    } else if (dio) {
	for(; dio->do_Parent != NULL; dio = dio->do_Parent) {
	    if (SUCC(SUCC(dio)) != NULL) {
		return SUCC(dio);
	    }
	}
    }
    return NULL;
}

struct DisplayObject *DisplayObjectParent(struct DisplayObject *dio)
{
    struct DisplayObject *parent = NULL;

    if (dio) {
        parent = dio->do_Parent;
    }
    return parent;
}

bool DisplayObjectIsAncestor(struct DisplayObject *dio,struct DisplayObject *a)
{
    while (dio) {
        if (dio == a) {
            return TRUE;
        }
        dio = DisplayObjectParent(dio);
    }
    return FALSE;
}

bool DisplayObjectOpen(struct DisplayObject *dio)
{
    bool open = TRUE;

    if (dio) {
	switch (dio->do_Type) {
	    case DO_TYPE_GROUP: 
		open = dio->do_Contents.group.open; 
		break;
	    case DO_TYPE_REGISTER: 
		open = dio->do_Contents.reggroup.open; 
		break;
	    case DO_TYPE_WINDOW:
		open = dio->do_Contents.window.open;
		break;
	}
    }
    return open;
}

void SetDisplayObjectOpen(struct DisplayObject *dio, bool open)
{
    if (dio) {
	switch (dio->do_Type) {
	    case DO_TYPE_GROUP: 
		dio->do_Contents.group.open = open;
		break;
	    case DO_TYPE_REGISTER: 
		dio->do_Contents.reggroup.open = open;
		break;
	    case DO_TYPE_WINDOW:
		dio->do_Contents.window.open = open;
		break;
	}
    }
}

bool DisplayObjectVisible(struct DisplayObject *dio)
{
    bool visible = FALSE;

    for (; dio != NULL; dio = dio->do_Parent) {
        visible = dio->do_Visible;
        if (!visible) {
            break;
	}
    }
    return visible;
}

void SetDisplayObjectVisible(struct DisplayObject *dio, bool visible)
{
    if (dio) {
        dio->do_Visible = visible;

	if (dio->do_Type == DO_TYPE_FIELD && 
                dio->do_Contents.field.autoUpdate) {
            SetProgramNeedRecompile( ProjectProgram(dio->do_Project), TRUE);
	}
    }
}

const char *DisplayObjectTitle(struct DisplayObject *dio)
{
    const char *title = NULL;

    if (dio) {
	title = dio->do_Title;
    }
    return title;
}

short DisplayObjectWeight(struct DisplayObject *dio)
{
    short weight = 100;

    if (dio) {
        weight = dio->do_Weight;
    }
    return weight;
}

void SetDisplayObjectWeight(struct DisplayObject *dio, short weight)
{
    if (dio) {
        dio->do_Weight = weight;
    }
}

const char *DisplayObjectMUIFontname(struct DisplayObject *dio)
{
    const char *fn = NULL;

    if (dio) {
        fn = dio->do_MUIFontname;
    }
    return fn;
}

const char *DisplayObjectGTKFontname(struct DisplayObject *dio)
{
    const char *fn = NULL;

    if (dio) {
        fn = dio->do_GTKFontname;
    }
    return fn;
}

const char *DisplayObjectQtFontname(struct DisplayObject *dio)
{
    const char *fn = NULL;

    if (dio) {
        fn = dio->do_QtFontname;
    }
    return fn;
}

const char *DisplayObjectBackground(struct DisplayObject *dio)
{
    const char *bg = NULL;

    if (dio) {
        bg = dio->do_Background;
    }
    return bg;
}

static bool setString(struct MemPool *memPool, char **ptr, const char *s)
{
    char *copy = AllocString(memPool, s);

    FreeString(memPool, *ptr);
    *ptr = copy;
    return (bool)(copy != NULL || s == NULL);
}

void SetDisplayObjectBackground(struct DisplayObject *dio, const char *bg)
{
    if (dio) {
        struct Project *pr = DisplayObjectProject(dio);
        struct MemPool *memPool = ProjectMemPool(pr);

        setString(memPool, &dio->do_Background, bg);
    }
}

GuiObject DisplayObjectGuiObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio) {
	go = dio->do_GuiObject;
    }
    return go;
}

void SetDisplayObjectGuiObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio) {
	dio->do_GuiObject = go;
    }
}

GuiObject DisplayObjectRootObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio) {
	go = dio->do_RootObject;
    }
    return go;
}

void SetDisplayObjectRootObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio) {
	dio->do_RootObject = go;
    }
}

GuiObject DisplayObjectActivateObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio) {
	go = dio->do_ActivateObject;
    }
    return go;
}

void SetDisplayObjectActivateObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio) {
	dio->do_ActivateObject = go;
    }
}

struct Table *DisplayObjectTable(struct DisplayObject *dio)
{
    struct Table *tab = NULL;

    for (; dio; dio = dio->do_Parent) {
	if (dio->do_Type == DO_TYPE_TABLE) {
	    tab = dio->do_Contents.table.table;
	    break;
	}
    }
    return tab;
}

struct DisplayObject *DisplayObjectWindow(struct DisplayObject *dio)
{
    for (dio = DisplayObjectParent(dio); dio != NULL; dio = dio->do_Parent) {
	if (dio->do_Type == DO_TYPE_WINDOW) {
            return dio;
	}
    }
    return NULL;
}

static bool isDisplayObjectRec(struct DisplayObject *dio, void *ptr)
{
    struct List *childs = DisplayObjectObjects(dio);

    if (dio == ptr) {
        return TRUE;
    } else if (childs) {
        FORNODE (dio, childs) {
            if (isDisplayObjectRec(dio, ptr)) {
                return TRUE;
	    }
	}
    }
    return FALSE;
}

bool IsDisplayObject(struct Project *pr, void *ptr)
{
    struct DisplayObject *root = ProjectRootDisplayObject(pr);
#ifdef USE_LOG_WINDOW
    struct Log *log = ProjectLog(pr);
    struct DisplayObject *logdio = LogWindowObject(log);

    return (isDisplayObjectRec(root, ptr) ||
            isDisplayObjectRec(logdio, ptr));
#else
    return isDisplayObjectRec(root, ptr);
#endif
}

struct DisplayObject *DisplayActiveObject(GuiObject win)
{
    struct Project *pr = GuiWindowProject(win);
    GuiObject go = GuiActiveObject(win);
    struct DisplayObject *dio = GuiObjectDisplayObject(go);

    if (IsDisplayObject(pr, dio)) {
        return dio;
    }
    return NULL;
}

struct DisplayObject *DisplayHomeObject(struct Table *tab)
{
    struct List *fields = TableFields(tab);
    struct Field *field;
    struct DisplayObject *dio;

    if (fields) {
        FORNODE (field, fields) {
            dio = FieldDisplayObject(field);
            if (dio != NULL && dio->do_Contents.field.home) {
                return dio;
	    }
        }
    }
    return NULL;
}

cschar DisplayObjectShortcut(struct DisplayObject *dio)
{
    cschar shortcut = 0;

    if (dio) {
        shortcut = dio->do_Shortcut;
    }
    return shortcut;
}

char DisplayObjectEnabled(struct DisplayObject *dio)
{
    char enabled = DOD_ENABLED_ALWAYS;

    if (dio) {
        enabled = dio->do_Enabled;
    }
    return enabled;
}

const char *DisplayObjectEnabledCompute(struct DisplayObject *dio)
{
    const char *compute = NULL;

    if (dio) {
        compute = dio->do_EnabledCompute;
    }
    return compute;
}

const char *DisplayObjectName(struct DisplayObject *dio)
{
    struct Table *tab = DisplayObjectTable(dio);
    const char *head = NULL, *first = NULL, *last = "";
    static char buffer[103];
    short size = 100, l1, l2, l3;

    switch (DisplayObjectType(dio)) {
        case DO_TYPE_TABLE:
            head = GCS(MSG_Table);
            first = TableName(tab);
            break;
        case DO_TYPE_FIELD:
            head = GCS(MSG_Field);
            first = TableName(tab);
            last = FieldName(DisplayFieldField(dio));
            break;
        case DO_TYPE_WINDOW:
            head = GCS(MSG_Window);
            first = DisplayObjectTitle(dio);
            if (tab != NULL) {
                last = first;
                first = TableName(tab);
            }
            break;
        default:
            break;
    }

    if (head == NULL || first == NULL) {
        return NULL;
    }

    l1 = strlen(head);
    l2 = strlen(first);
    l3 = last? strlen(last) : 0;

    l3 = MIN(l3, size);
    size -= l3;
    l2 = MIN(l2, size);
    size -= l2;
    l1 = MIN(l1, size);

    sprintf(buffer,
        "%.*s %.*s%s%.*s", l1, head, l2, first, (l3 > 0)? "." : "", l3, last);

    return buffer;
}

static struct DisplayObject *displayObjectShortcutInUseRec(cschar shortcut,
	struct DisplayObject *dio, struct DisplayObject *windio, 
	struct DisplayObject *leaveout)
{
    struct List *childs = DisplayObjectObjects(dio);
    struct DisplayObject *ret = NULL;

    if (DisplayObjectWindow(dio) != windio) {
        return NULL;
    }

    if (dio != leaveout) {
        if (CodesetToUpper(DisplayObjectShortcut(dio)) == 
		CodesetToUpper(shortcut)) {
            return dio;
	}
    }

    if (childs) {
        FORNODE (dio, childs) {
            ret = displayObjectShortcutInUseRec(shortcut, dio, windio,leaveout);
            if (ret != NULL) {
                return ret;
	    }
        }
    }
    return NULL;
}

struct DisplayObject *DisplayObjectShortcutInUse(cschar shortcut,
    struct DisplayObject *windio, struct DisplayObject *leaveout)
{
    struct List *childs = DisplayObjectObjects(windio);
    struct DisplayObject *dio, *ret;

    if (shortcut == 0 || windio == NULL) {
        return NULL;
    }
    if (childs) {
        FORNODE(dio, childs) {
            ret = displayObjectShortcutInUseRec(shortcut, dio, windio,leaveout);
            if (ret != NULL) {
                return ret;
	    }
        }
    }
    return NULL;
}


/*
** General functions
*/
static void ClearDisplayObject(struct DisplayObject *dio)
{
    if (dio) {
    	struct MemPool *memPool = ProjectMemPool(dio->do_Project);

        switch (dio->do_Type) {
            case DO_TYPE_TABLE:
                FreeRecordDescriptionSpec(memPool,
                        dio->do_Contents.table.recordDescriptionSpec);
                dio->do_Contents.table.recordDescriptionSpec = NULL;
                break;

            case DO_TYPE_FIELD:
                FreeString(memPool, dio->do_Contents.field.titleMUIFontname);
                dio->do_Contents.field.titleMUIFontname = NULL;
                FreeString(memPool, dio->do_Contents.field.titleGTKFontname);
                dio->do_Contents.field.titleGTKFontname = NULL;
                FreeString(memPool, dio->do_Contents.field.titleQtFontname);
                dio->do_Contents.field.titleQtFontname = NULL;
                FreeRecordDescriptionSpec(memPool,
                        dio->do_Contents.field.recordDescriptionSpec);
                dio->do_Contents.field.recordDescriptionSpec = NULL;
                FreeString(memPool, 
                        dio->do_Contents.field.doubleClickTrigger);
                dio->do_Contents.field.doubleClickTrigger = NULL;
                FreeString(memPool, dio->do_Contents.field.urlDropTrigger);
                dio->do_Contents.field.urlDropTrigger = NULL;
                FreeString(memPool, dio->do_Contents.field.sortDropTrigger);
                dio->do_Contents.field.sortDropTrigger = NULL;
                FreeString(memPool, dio->do_Contents.field.bubbleHelp);
                dio->do_Contents.field.bubbleHelp = NULL;
                FreeLabels(memPool, dio->do_Contents.field.listviewLabels);
                dio->do_Contents.field.listviewLabels = NULL;
                FreeString(memPool, dio->do_Contents.field.listviewCompute);
                dio->do_Contents.field.listviewCompute = NULL;
                break;

            case DO_TYPE_REGISTER:
                FreeLabels(memPool, dio->do_Contents.reggroup.labels);
                dio->do_Contents.reggroup.labels = NULL;
                break;
        }

	FreeString(memPool, dio->do_Title);
	dio->do_Title = NULL;
        FreeString(memPool, dio->do_EnabledCompute);
        dio->do_EnabledCompute = NULL;
	FreeString(memPool, dio->do_MUIFontname);
	dio->do_MUIFontname = NULL;
	FreeString(memPool, dio->do_GTKFontname);
	dio->do_GTKFontname = NULL;
	FreeString(memPool, dio->do_QtFontname);
	dio->do_QtFontname = NULL;
	FreeString(memPool, dio->do_Background);
	dio->do_Background = NULL;
    }
}

static void FreeDisplayObject(struct DisplayObject *dio)
{
    if (dio) {
    	struct MemPool *memPool = ProjectMemPool(dio->do_Project);
    	struct List *childs = DisplayObjectObjects(dio);
    	struct DisplayObject *child;

        if (childs) {
	    while ((child = (struct DisplayObject *)REMHEAD(childs)) != NULL) {
		FreeDisplayObject(child);
	    }
	}
        ClearDisplayObject(dio);
	PoolFreeMem(memPool, dio, sizeof(struct DisplayObject));
    }
}

static struct DisplayObject *AllocDisplayObject(struct Project *pr)
{
    struct DisplayObject *dio = NULL;
    struct MemPool *memPool = ProjectMemPool(pr);

    dio = PoolAllocMemZero(memPool, sizeof(struct DisplayObject));
    if (dio == NULL) {
	return NULL;
    }

    NewList(&dio->do_Childs);
    dio->do_Project = pr;
    dio->do_Weight = 100;
    dio->do_Visible = TRUE;
    dio->do_Type = DO_TYPE_TABLE;

    return dio;
}

static struct DisplayObject *getdioparent(struct Project *pr,
        struct Table *tab, struct DisplayObject *dio, bool isPred)
{
    struct DisplayObject *parent, *dr = NULL;

    if (tab == NULL) {                   /* root display */
        dr = ProjectRootDisplayObject(pr);
        if (dio == NULL) {
            return dr;
	} else if (dio->do_Type == DO_TYPE_TABLE) {
            dio = dio->do_Parent;
            isPred = FALSE;
        }
        for (parent = dio; parent; parent = parent->do_Parent) {
            if (parent->do_Type == DO_TYPE_TABLE) {
                return dr;
	    }
	}
    } else {                              /* table display */
        dr = TableDisplayObject(tab);
        for (parent = dio; parent; parent = parent->do_Parent) {
            if (parent->do_Type == DO_TYPE_TABLE) {
                break;
	    }
	}
        if (parent == NULL) {
            return dr;
	}
    }

    if (dio) {
        if (!isPred && DisplayObjectObjects(dio) != NULL) {
            return dio;
	}
	dio = dio->do_Parent;
        if (dio) {
	    return dio;
	}
    }
    return dr;
}

static void InsertDisplayObject(struct List *list, struct DisplayObject *dio,
        struct DisplayObject *pred)
{
    INSERT(list, dio, pred);
}


static void movedisplayobject(struct DisplayObject *move, 
	struct DisplayObject *newparent, struct DisplayObject *pred)
{
    struct List *childs = DisplayObjectObjects(newparent);
    struct DisplayObject *dio;

    if (move && move->do_Parent && childs) {
	for (dio = newparent; dio; dio = dio->do_Parent) {
	    if (dio == move) {
		return;		/* illegal move into itself */
	    }
	}
        if (pred == move) {
            pred = PRED(pred);
	}
	REMOVE(move);
	move->do_Parent = newparent;
        InsertDisplayObject(childs, move, pred);
    }
}

void DeleteDisplayObject(struct DisplayObject *dio)
{
    if (dio) {
        ProgramRemDisplayObject(dio);
	if (dio->do_Parent) {
            REMOVE(dio);
	}
	FreeDisplayObject(dio);
    }
}

void FreeRootDisplayObject(struct DisplayObject *dio)
{
    FreeDisplayObject(dio);
}

void ClearRootDisplayObject(struct DisplayObject *dio)
{
    struct Project *pr = DisplayObjectProject(dio);
    struct MemPool *memPool = ProjectMemPool(pr);
    struct List *childs = DisplayObjectObjects(dio);
    struct DisplayObject *child;

    if (dio && dio->do_Type == DO_TYPE_WINDOW) {
        if (childs) {
            while ((child = (struct DisplayObject *)REMHEAD(childs)) != NULL) {
                FreeDisplayObject(child);
	    }
	}

	setString(memPool, &dio->do_Title, NULL);
	setString(memPool, &dio->do_MUIFontname, NULL);
	setString(memPool, &dio->do_GTKFontname, NULL);
	setString(memPool, &dio->do_QtFontname, NULL);
	setString(memPool, &dio->do_Background, NULL);
        dio->do_GuiObject = NULL;
        dio->do_RootObject = NULL;
        dio->do_ActivateObject = NULL;
        dio->do_Contents.window.id = DEFAULT_ROOT_WINDOW_ID;
        dio->do_Contents.window.open = TRUE;
        dio->do_Contents.window.windowopen = TRUE;
        dio->do_Contents.window.wd_valid = FALSE;
    }
}

struct DisplayObject *AllocRootDisplayObject(struct Project *pr)
{
    struct DisplayObject *dio = NULL;

    dio = AllocDisplayObject(pr);
    if (dio == NULL) {
	return NULL;
    }

    dio->do_Type = DO_TYPE_WINDOW;
    ClearRootDisplayObject(dio);

    return dio;
}

struct Table *DisplayProjectActiveTable(struct Project *pr)
{
    struct DisplayObject *root = ProjectRootDisplayObject(pr);
    struct Table *tab = NULL;

    if (root != NULL) {
        tab = FindTableNum(pr, root->do_Contents.window.activeTableNum);
        if (TableIsSystem(tab)) {
            tab = NULL;
        }
    }
    return tab;
}

void SetDisplayProjectActiveTableNum(struct Project *pr, long num)
{
    struct DisplayObject *root = ProjectRootDisplayObject(pr);

    if (root != NULL) {
        root->do_Contents.window.activeTableNum = num;
    }
}

void SetDisplayProjectActiveTable(struct Project *pr, struct Table *tab)
{
    SetDisplayProjectActiveTableNum(pr, TableNum(tab));
}

struct DisplayObject *DisplayProjectActiveDisplayObject(struct Project *pr)
{
    struct DisplayObject *root = ProjectRootDisplayObject(pr), *dio = NULL;

    if (root != NULL) {
        dio = FindDisplayObjectNum(pr, 
		root->do_Contents.window.activeDisplayObjectNum);
    }
    return dio;
}

void SetDisplayProjectActiveDisplayObjectNum(struct Project *pr, long num)
{
    struct DisplayObject *root = ProjectRootDisplayObject(pr);

    if (root != NULL) {
        root->do_Contents.window.activeDisplayObjectNum = num;
    }
}

void SetDisplayProjectActiveDisplayObject(struct Project *pr, 
	struct DisplayObject *dio)
{
    SetDisplayProjectActiveDisplayObjectNum(pr, DisplayObjectNum(dio));
}

static void StoreDisplayFieldsRec(struct DisplayObject *dio, bool horiz)
{
    struct List *childs = DisplayObjectObjects(dio);
    GuiObject go, weight_go = NULL;
    short type = DisplayObjectType(dio);
    struct WindowDimensions wd;

    switch (type) {
        case DO_TYPE_WINDOW:
            go = DisplayObjectGuiObject(dio);
            if (GuiGetWindowDimensions(go, &wd)) {
                SetDisplayWindowDimensions(dio, &wd);
            }
            SetDisplayWindowOpen(dio, GuiWindowOpen(go));
            /* no break */

        default:
            weight_go = DisplayObjectRootObject(dio);
            break;
    }

    if (weight_go != NULL) {
        if (DisplayObjectWeight(dio)) {
            short weight = GuiObjectWeight(weight_go, horiz);
            if (weight) {
                SetDisplayObjectWeight(dio, weight);
            }
        }
    }

    if (childs != NULL) {
        horiz = (type == DO_TYPE_GROUP)? DisplayGroupHoriz(dio) : FALSE;
        FORNODE (dio, childs) {
            StoreDisplayFieldsRec(dio, horiz);
        }
    }
}

void StoreDisplayFields(struct Project *pr)
{
    StoreDisplayFieldsRec(ProjectRootDisplayObject(pr), FALSE);
}

static void SetDisplayFieldsRec(struct DisplayObject *dio)
{
    struct List *childs = DisplayObjectObjects(dio);
    GuiObject go;

    if (DisplayObjectType(dio) == DO_TYPE_WINDOW) {
        /*
        ** We only open windows.  The window dimensions are automatically
        ** set when creating the GUI.
        */
        go = DisplayObjectGuiObject(dio);
        if (DisplayWindowOpen(dio)) {
            GuiOpenWindow(go);
        }
    }
    if (childs != NULL) {
        FORNODE(dio, childs) {
            SetDisplayFieldsRec(dio);
        }
    }
}

void SetDisplayFields(struct Project *pr)
{
    SetDisplayFieldsRec(ProjectRootDisplayObject(pr));
}

/*
** Table
*/

bool DisplayTableHasPanel(struct DisplayObject *dio)
{
    bool has = FALSE;

    if (dio && dio->do_Type == DO_TYPE_TABLE) {
        has = dio->do_Contents.table.hasPanel;
    }
    return has;
}

const char *DisplayTablePanelBackground(struct DisplayObject *dio)
{
    const char *bg = NULL;

    if (dio && dio->do_Type == DO_TYPE_TABLE) {
        bg = dio->do_Contents.table.panel.background;
    }
    return bg;
}

bool DisplayTablePanelNumAll(struct DisplayObject *dio)
{
    bool num = FALSE;

    if (dio && dio->do_Type == DO_TYPE_TABLE) {
	num = dio->do_Contents.table.panel.numall;
    }
    return num;
}

bool DisplayTablePanelFilter(struct DisplayObject *dio)
{
    bool filter = FALSE;

    if (dio && dio->do_Type == DO_TYPE_TABLE) {
	filter = dio->do_Contents.table.panel.filter;
    }
    return filter;
}

bool DisplayTablePanelArrows(struct DisplayObject *dio)
{
    bool arrows = FALSE;

    if (dio && dio->do_Type == DO_TYPE_TABLE) {
	arrows = dio->do_Contents.table.panel.arrows;
    }
    return arrows;
}

GuiObject DisplayTablePanelNumAllObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio && dio->do_Type == DO_TYPE_TABLE) {
        go = dio->do_Contents.table.panel.numall_go;
    }
    return go;
}

void SetDisplayTablePanelNumAllObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_TABLE) {
        dio->do_Contents.table.panel.numall_go = go;
    }
}

GuiObject DisplayTablePanelFilterObject(struct DisplayObject *dio)
{
    GuiObject filter_go = NULL;

    if (dio && dio->do_Type == DO_TYPE_TABLE) {
        filter_go = dio->do_Contents.table.panel.filter_go;
    }
    return filter_go;
}

void SetDisplayTablePanelFilterObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_TABLE) {
        dio->do_Contents.table.panel.filter_go = go;
    }
}

const struct RecordDescriptionSpec *DisplayTableRecordDescriptionSpec(
	struct DisplayObject *dio)
{
    const struct RecordDescriptionSpec *rds = NULL;

    if (dio && dio->do_Type == DO_TYPE_TABLE) {
	rds = dio->do_Contents.table.recordDescriptionSpec;
    }
    return rds;
}

static void updateTableRecordDescriptionSpec(struct Table *tab)
{
    struct DisplayObject *dio = TableDisplayObject(tab);

    if (dio && dio->do_Type == DO_TYPE_TABLE) {
        struct RecordDescriptionSpec *rds = 
            dio->do_Contents.table.recordDescriptionSpec;
        struct DisplayObject *home = DisplayHomeObject(tab);

        if (dio != NULL && rds == NULL && home != NULL) {
            rds = AllocRecordDescriptionSpec(TableMemPool(tab), 1);
            if (rds) {
                rds->rds_Fields[0] = DisplayFieldField(home);
                dio->do_Contents.table.recordDescriptionSpec = rds;
            }
        }
    }
}

bool DisplayNewTable(struct Table *tab, struct DisplayObject *dioparent)
{
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = ProjectMemPool(pr);
    struct List *childs = NULL;
    struct DisplayObject *dio, *pred = dioparent;

    if (pr == NULL) {
	return FALSE;
    }
    dioparent = getdioparent(pr, NULL, dioparent, FALSE);
    childs = DisplayObjectObjects(dioparent);
    if (childs == NULL) {
        return FALSE;
    }
    if (pred != NULL && (PRED(pred) == NULL || pred->do_Parent != dioparent)) {
        pred = NULL;
    }
    SetTableDisplayObject(tab, NULL);
    dio = AllocDisplayObject(pr);
    if (dio == NULL) {
	return FALSE;
    }
    dio->do_Parent = dioparent;
    dio->do_Type = DO_TYPE_TABLE;
    setString(memPool, &dio->do_Title, TableName(tab));
    dio->do_Contents.table.table = tab;
    dio->do_Contents.table.hasPanel = TableExactlyOneRecord(tab)? FALSE : TRUE;
    dio->do_Contents.table.panel.numall = TRUE;
    dio->do_Contents.table.panel.filter = TRUE;
    dio->do_Contents.table.panel.arrows = TRUE;

    InsertDisplayObject(childs, dio, pred);
    SetTableDisplayObject(tab, dio);
    return TRUE;
}

long DisplayChangeTable(struct Table *tab, struct DisplayTableData *dtd)
{
    struct DisplayObject *tabdio = TableDisplayObject(tab);
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = ProjectMemPool(pr);
    struct RecordDescriptionSpec *rds = NULL;

    if (tabdio == NULL || dtd == NULL) {
        return ERR_INTERNAL;
    }

    if (!RecordDescriptionSpecIsDefault(dtd->dtd_RecordDescriptionSpec)) {
        rds = CopyRecordDescriptionSpec(memPool,dtd->dtd_RecordDescriptionSpec);
        if (rds == NULL) {
	    return ERR_MEM;
	}
    }

    setString(memPool, &tabdio->do_Title, dtd->dtd_Title);
    tabdio->do_Weight = dtd->dtd_Weight;
#ifdef MUI
    setString(memPool, &tabdio->do_MUIFontname, dtd->dtd_Fontname);
#elif (defined GTK)
    setString(memPool, &tabdio->do_GTKFontname, dtd->dtd_Fontname);
#elif (defined QT)
    setString(memPool, &tabdio->do_QtFontname, dtd->dtd_Fontname);
#endif
    setString(memPool, &tabdio->do_Background, dtd->dtd_Background);
    tabdio->do_Contents.table.hasPanel = dtd->dtd_HasPanel;
    setString(memPool, &tabdio->do_Contents.table.panel.background, 
            dtd->dtd_PanelBackground);
    tabdio->do_Contents.table.panel.numall = dtd->dtd_PanelNumAll;
    tabdio->do_Contents.table.panel.filter = dtd->dtd_PanelFilter;
    tabdio->do_Contents.table.panel.arrows = dtd->dtd_PanelArrows;
    FreeRecordDescriptionSpec(memPool, 
            tabdio->do_Contents.table.recordDescriptionSpec);
    tabdio->do_Contents.table.recordDescriptionSpec = rds;

    return OK;
}

struct Field *DisplayTableActiveField(struct Table *tab)
{
    struct DisplayObject *tabdio = TableDisplayObject(tab);
    struct Field *field = NULL;

    if (tabdio != NULL) {
        field = FindNodeNum(TableFields(tab), 
		tabdio->do_Contents.table.activeFieldNum);
    }
    return field;
}

void SetDisplayTableActiveField(struct Table *tab, struct Field *field)
{
    struct DisplayObject *tabdio = TableDisplayObject(tab);

    if (tabdio != NULL) {
        tabdio->do_Contents.table.activeFieldNum = NodeNum(field);
    }
}

struct DisplayObject *DisplayTableActiveDisplayObject(struct Table *tab)
{
    struct DisplayObject *tabdio = TableDisplayObject(tab), *dio = NULL;

    if (tabdio != NULL) {
        dio = FindDisplayObjectNum(TableProject(tab),
            tabdio->do_Contents.table.activeDisplayObjectNum);
    }
    return dio;
}

void SetDisplayTableActiveDisplayObject(struct Table *tab, 
	struct DisplayObject *dio)
{
    struct DisplayObject *tabdio = TableDisplayObject(tab);

    if (tabdio != NULL) {
        tabdio->do_Contents.table.activeDisplayObjectNum= DisplayObjectNum(dio);
    }
}

/*
** Record description
*/

static void recordDescriptionSpecDeleteFieldReference(
        struct RecordDescriptionSpec *rds, struct Field *ref)
{
    short i;

    if (rds != NULL) {
        for (i=0; i < rds->rds_NumFields; i++) {
            if (rds->rds_Fields[i] == ref) {
                rds->rds_Fields[i] = NULL;
            }
        }
    }
}

void DisplayDeleteFieldReference(struct Field *ref)
{
    struct Table *tab = FieldTable(ref);
    struct Project *pr = TableProject(tab);
    struct List *tabs = ProjectTables(pr), *fields;
    struct Field *field;
    struct DisplayObject *dio;
    struct RecordDescriptionSpec *rds;

    if (tabs) {
        FORNODE(tab, tabs) {
            dio = TableDisplayObject(tab);
            if (dio && dio->do_Type == DO_TYPE_TABLE) {
                rds = dio->do_Contents.table.recordDescriptionSpec;
                recordDescriptionSpecDeleteFieldReference(rds, ref);
            }

            fields = TableFields(tab);
            if (fields) {
                FORNODE(field, fields) {
                    dio = FieldDisplayObject(field);
                    if (dio && dio->do_Type == DO_TYPE_FIELD) {
                        rds = dio->do_Contents.field.recordDescriptionSpec;
                        recordDescriptionSpecDeleteFieldReference(rds, ref);
                    }
                }
	    }
        }
    }
}

/*
** Field
*/

struct Field *DisplayFieldField(struct DisplayObject *dio)
{
    struct Field *field = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	field = dio->do_Contents.field.field;
    }
    return field;
}

char DisplayFieldTitlePosition(struct DisplayObject *dio)
{
    char pos = DAD_TITLEPOSITION_LEFT;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	pos = dio->do_Contents.field.titlepos;
    }
    return pos;
}

const char *DisplayFieldTitleMUIFontname(struct DisplayObject *dio)
{
    const char *fn = NULL;

    if (dio) {
        fn = dio->do_Contents.field.titleMUIFontname;
    }
    return fn;
}

const char *DisplayFieldTitleGTKFontname(struct DisplayObject *dio)
{
    const char *fn = NULL;

    if (dio) {
        fn = dio->do_Contents.field.titleGTKFontname;
    }
    return fn;
}

const char *DisplayFieldTitleQtFontname(struct DisplayObject *dio)
{
    const char *fn = NULL;

    if (dio) {
        fn = dio->do_Contents.field.titleQtFontname;
    }
    return fn;
}

bool DisplayFieldHome(struct DisplayObject *dio)
{
    bool home = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	home = dio->do_Contents.field.home;
    }
    return home;
}

bool DisplayFieldTabChain(struct DisplayObject *dio)
{
    bool chain = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	chain = dio->do_Contents.field.tabchain;
    }
    return chain;
}

bool DisplayFieldReadOnly(struct DisplayObject *dio)
{
    bool readonly = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	readonly = dio->do_Contents.field.readonly;
    }
    return readonly;
}

char DisplayFieldSubType(struct DisplayObject *dio)
{
    char subType = 0;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	subType = dio->do_Contents.field.subType;
    }
    return subType;
}

char DisplayFieldAlignment(struct DisplayObject *dio)
{
    char alignment = DAD_ALIGNMENT_CENTER;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	alignment = dio->do_Contents.field.alignment;
    }
    return alignment;
}

bool DisplayFieldCalendar(struct DisplayObject *dio)
{
    bool calendar = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	calendar = dio->do_Contents.field.calendar;
    }
    return calendar;
}

char DisplayFieldFormat(struct DisplayObject *dio)
{
    char format = 0;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	format = dio->do_Contents.field.format;
    }
    return format;
}

char DisplayFieldNumDecimals(struct DisplayObject *dio)
{
    char numdecimals = 2;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	numdecimals = dio->do_Contents.field.numdecimals;
    }
    return numdecimals;
}

bool DisplayFieldShowTitles(struct DisplayObject *dio)
{
    bool showtitles = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	showtitles = dio->do_Contents.field.showtitles;
    }
    return showtitles;
}

bool DisplayFieldControlTable(struct DisplayObject *dio)
{
    bool control = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	control = dio->do_Contents.field.controlTable;
    }
    return control;
}

char DisplayFieldOnDoubleClick(struct DisplayObject *dio)
{
    char action = DAD_DOUBLECLICK_SHOWRECORD;    /* e.g. for query editor */

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	action = dio->do_Contents.field.onDoubleClick;
    }
    return action;
}

const char *DisplayFieldDoubleClickTrigger(struct DisplayObject *dio)
{
    const char *trigger = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	trigger = dio->do_Contents.field.doubleClickTrigger;
    }
    return trigger;
}

const char *DisplayFieldURLDropTrigger(struct DisplayObject *dio)
{
    const char *trigger = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	trigger = dio->do_Contents.field.urlDropTrigger;
    }
    return trigger;
}

const char *DisplayFieldSortDropTrigger(struct DisplayObject *dio)
{
    const char *trigger = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	trigger = dio->do_Contents.field.sortDropTrigger;
    }
    return trigger;
}

bool DisplayFieldAutoUpdate(struct DisplayObject *dio)
{
    bool update = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	update = dio->do_Contents.field.autoUpdate;
    }
    return update;
}

char DisplayFieldImageSize(struct DisplayObject *dio)
{
    char imagesize = DAD_IMAGESIZE_RESIZABLE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	imagesize = dio->do_Contents.field.imagesize;
    }
    return imagesize;
}

const char *DisplayFieldBubbleHelp(struct DisplayObject *dio)
{
    const char *bh = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	bh = dio->do_Contents.field.bubbleHelp;
    }
    return bh;
}

char DisplayFieldRecordsKind(struct DisplayObject *dio)
{
    char recordsKind = DAD_RECORDS_ALL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	recordsKind = dio->do_Contents.field.recordsKind;
    }
    return recordsKind;
}

const char *DisplayFieldRecordsCompute(struct DisplayObject *dio)
{
    const char *compute = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	compute = dio->do_Contents.field.recordsCompute;
    }
    return compute;
}

bool DisplayFieldInitialRecord(struct DisplayObject *dio)
{
    bool initialRecord = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	initialRecord = dio->do_Contents.field.initialRecord;
    }
    return initialRecord;
}

bool DisplayFieldCurrentRecord(struct DisplayObject *dio)
{
    bool currentRecord = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	currentRecord = dio->do_Contents.field.currentRecord;
    }
    return currentRecord;
}

const struct RecordDescriptionSpec *DisplayFieldRecordDescriptionSpec(
	struct DisplayObject *dio, bool followInherited)
{
    const struct RecordDescriptionSpec *rds = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	rds = dio->do_Contents.field.recordDescriptionSpec;
        if (rds && rds->rds_Type == RDS_TYPE_INHERITED && followInherited) {
            struct Field *field = dio->do_Contents.field.field;

            dio = TableDisplayObject(FieldReferenceTable(field));
            rds = DisplayTableRecordDescriptionSpec(dio);
        }
    }
    return rds;
}

bool DisplayFieldFile(struct DisplayObject *dio)
{
    bool file = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	file = dio->do_Contents.field.file;
    }
    return file;
}

bool DisplayFieldFont(struct DisplayObject *dio)
{
    bool font = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	font = dio->do_Contents.field.font;
    }
    return font;
}

bool DisplayFieldListview(struct DisplayObject *dio)
{
    bool listview = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	listview = dio->do_Contents.field.listview;
    }
    return listview;
}

bool DisplayFieldView(struct DisplayObject *dio)
{
    bool view = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	view = dio->do_Contents.field.view;
    }
    return view;
}

char **DisplayFieldListviewLabels(struct DisplayObject *dio)
{
    char **labels = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        labels = dio->do_Contents.field.listviewLabels;
    }
    return labels;
}

long SetDisplayFieldListviewLabels(struct DisplayObject *dio, char **labels)
{
    struct Project *pr = DisplayObjectProject(dio);
    struct MemPool *memPool = ProjectMemPool(pr);
    char **copy = NULL;

    if (dio == NULL || dio->do_Type != DO_TYPE_FIELD) {
        return ERR_INTERNAL;
    }

    copy = CopyLabels(memPool, labels);
    if (copy == NULL) {
        return ERR_MEM;
    }
    FreeLabels(memPool, dio->do_Contents.field.listviewLabels);
    dio->do_Contents.field.listviewLabels = copy;
    return OK;
}

const char *DisplayFieldListviewCompute(struct DisplayObject *dio)
{
    const char *compute = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        compute = dio->do_Contents.field.listviewCompute;
    }
    return compute;
}

bool DisplayFieldDisplayExternal(struct DisplayObject *dio)
{
    bool displayexternal = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	displayexternal = dio->do_Contents.field.displayexternal;
    }
    return displayexternal;
}

bool DisplayFieldTitleAtStringField(struct DisplayObject *dio)
{
    bool titleatstringfield = TRUE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	titleatstringfield = dio->do_Contents.field.titleatstringfield;
    }
    return titleatstringfield;
}

bool DisplayFieldHideStringField(struct DisplayObject *dio)
{
    bool hidestringfield = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	hidestringfield = dio->do_Contents.field.hidestringfield;
    }
    return hidestringfield;
}

bool DisplayFieldShow(struct DisplayObject *dio)
{
    bool show = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	show = dio->do_Contents.field.show;
    }
    return show;
}

bool DisplayFieldFilter(struct DisplayObject *dio)
{
    bool filter = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	filter = dio->do_Contents.field.filter;
    }
    return filter;
}

bool DisplayFieldAutoShow(struct DisplayObject *dio)
{
    bool autoshow = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	autoshow = dio->do_Contents.field.autoshow;
    }
    return autoshow;
}

bool DisplayFieldFrame(struct DisplayObject *dio)
{
    bool frame = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
	frame = dio->do_Contents.field.frame;
    }
    return frame;
}

GuiObject DisplayFieldFileObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        go = dio->do_Contents.field.fileObject;
    }
    return go;
}

GuiObject DisplayFieldFontObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        go = dio->do_Contents.field.fontObject;
    }
    return go;
}

GuiObject DisplayFieldListviewObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        go = dio->do_Contents.field.listviewObject;
    }
    return go;
}

GuiObject DisplayFieldViewObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        go = dio->do_Contents.field.viewObject;
    }
    return go;
}

GuiObject DisplayFieldExternalObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        go = dio->do_Contents.field.externalObject;
    }
    return go;
}

GuiObject DisplayFieldShowObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;
    
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        go = dio->do_Contents.field.showObject;
    }
    return go;
}

GuiObject DisplayFieldFilterObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        go = dio->do_Contents.field.filterObject;
    }
    return go;
}

GuiObject DisplayFieldAutoShowObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        go = dio->do_Contents.field.autoshowObject;
    }
    return go;
}

GuiObject DisplayFieldPopObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        go = dio->do_Contents.field.popObject;
    }
    return go;
}

GuiObject DisplayFieldDisableObject(struct DisplayObject *dio)
{
    GuiObject go = NULL;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        go = dio->do_Contents.field.disableObject;
    }
    return go;
}

void SetDisplayFieldFileObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        dio->do_Contents.field.fileObject = go;
    }
}

void SetDisplayFieldFontObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        dio->do_Contents.field.fontObject = go;
    }
}

void SetDisplayFieldListviewObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        dio->do_Contents.field.listviewObject = go;
    }
}

void SetDisplayFieldViewObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        dio->do_Contents.field.viewObject = go;
    }
}

void SetDisplayFieldExternalObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        dio->do_Contents.field.externalObject = go;
    }
}

void SetDisplayFieldShowObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        dio->do_Contents.field.showObject = go;
    }
}

void SetDisplayFieldFilterObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        dio->do_Contents.field.filterObject = go;
    }
}

void SetDisplayFieldAutoShowObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        dio->do_Contents.field.autoshowObject = go;
    }
}

void SetDisplayFieldPopObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        dio->do_Contents.field.popObject = go;
    }
}

void SetDisplayFieldDisableObject(struct DisplayObject *dio, GuiObject go)
{
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        dio->do_Contents.field.disableObject = go;
    }
}

bool DisplayFieldListActiveChanged(struct DisplayObject *dio)
{
    bool changed = FALSE;

    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        changed = dio->do_Contents.field.listActiveChanged;
    }
    return changed;
}

void SetDisplayFieldListActiveChanged(struct DisplayObject *dio,
        bool changed)
{
    if (dio && dio->do_Type == DO_TYPE_FIELD) {
        dio->do_Contents.field.listActiveChanged = changed;
    }
}

bool DisplayNewField(struct Field *field, 
        struct DisplayObject *dioparent)
{
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = ProjectMemPool(pr);
    const char *title = FieldName(field);
    short type = FieldType(field);
    struct DisplayObject *dio = NULL, *pred = dioparent;
    struct RecordDescriptionSpec *rds;
    struct List *childs;

    SetFieldDisplayObject(field, NULL);
    dioparent = getdioparent(pr, tab, dioparent, FALSE);
    childs = DisplayObjectObjects(dioparent);
    if (childs == NULL) {
	return FALSE;
    }

    if (pred != NULL && (PRED(pred) == NULL || pred->do_Parent != dioparent)) {
        pred = NULL;
    }

    dio = AllocDisplayObject(pr);
    if (dio == NULL) {
	return FALSE;
    }
    if (!setString(memPool, &dio->do_Title, title)) {
        FreeDisplayObject(dio);
        return FALSE;
    }
    dio->do_Parent = dioparent;
    dio->do_Enabled = DOD_ENABLED_NOTINITIAL;
    switch (type) {
        case TYPE_BOOL:
            dio->do_Weight = 0;
            break;
        case TYPE_INTEGER:
        case TYPE_REAL:
        case TYPE_DATE:
        case TYPE_TIME:
            dio->do_Weight = 50;
            break;
        default:
            dio->do_Weight = 100;
            break;
    }
    dio->do_Type = DO_TYPE_FIELD;
    dio->do_Contents.field.field = field;
    dio->do_Contents.field.titlepos = (type == TYPE_MEMO)?
        DAD_TITLEPOSITION_TOP : DAD_TITLEPOSITION_LEFT;
    dio->do_Contents.field.tabchain = TRUE;
    dio->do_Contents.field.subType =
        (type == TYPE_VIRTUAL)? DAD_VIRTUALTYPE_TEXT : 0;
    dio->do_Contents.field.alignment = (char)GetPrefsValue(pr,
        (type == TYPE_REAL || type == TYPE_INTEGER)?
            PREFS_DEFAULTALIGNMENT_NUMBER : PREFS_DEFAULTALIGNMENT_TEXT);
    dio->do_Contents.field.numdecimals = 2;
    dio->do_Contents.field.showtitles = TRUE;
    dio->do_Contents.field.autoUpdate = TRUE;
    dio->do_Contents.field.imagesize = DAD_IMAGESIZE_RESIZABLE;

    dio->do_Contents.field.titleatstringfield = TRUE;
    dio->do_Contents.field.onDoubleClick = DAD_DOUBLECLICK_SHOWRECORD;

    dio->do_Contents.field.initialRecord = TRUE;
    dio->do_Contents.field.currentRecord = TRUE;

    if ((type == TYPE_STRING || type == TYPE_INTEGER) && 
	    SUCC(SUCC(field)) == NULL && PRED(PRED(field)) == NULL) {
	/* First string or integer field becomes the 'home' object */
	dio->do_Contents.field.home = TRUE;
    } else if (type == TYPE_REFERENCE) {
        rds = AllocRecordDescriptionSpec(memPool, 0);
        if (rds) {
            rds->rds_Type = RDS_TYPE_INHERITED;
            dio->do_Contents.field.recordDescriptionSpec = rds;
	}
    }

    InsertDisplayObject(childs, dio, pred);
    SetFieldDisplayObject(field, dio);

    updateTableRecordDescriptionSpec(tab);

    return TRUE;
}

long DisplayChangeField(struct Field *field, 
	struct DisplayFieldData *dfd)
{
    struct Table *tab = FieldTable(field);
    struct DisplayObject *dio = FieldDisplayObject(field);
    struct Project *pr = DisplayObjectProject(dio);
    struct Program *prg = ProjectProgram(pr);
    struct MemPool *memPool = ProjectMemPool(pr);
    struct RecordDescriptionSpec *rds = NULL;
    bool cBefore, cAfter;
    struct UserFunction *ufBefore, *ufAfter;
    const char *compute;
    long error = OK;

    if (dio == NULL || dfd == NULL) {
        return ERR_INTERNAL;
    }

    if (!RecordDescriptionSpecIsDefault(dfd->dfd_RecordDescriptionSpec)) {
        rds = CopyRecordDescriptionSpec(memPool,dfd->dfd_RecordDescriptionSpec);
        if (rds == NULL) {
	    return ERR_MEM;
	}
    }

    error = SetDisplayFieldListviewLabels(dio, dfd->dfd_ListviewLabels);
    if (error != OK) {
        FreeRecordDescriptionSpec(memPool, rds);
        return error;
    }

    cBefore = (dio->do_Enabled == DOD_ENABLED_COMPUTED);
    cAfter = (dfd->dfd_Enabled == DOD_ENABLED_COMPUTED);
    compute = dio->do_EnabledCompute;
    ufBefore = FindUserFunctionName(prg, compute);

    setString(memPool, &dio->do_Title, dfd->dfd_Title);
    dio->do_Shortcut = dfd->dfd_Shortcut;
    dio->do_Enabled = dfd->dfd_Enabled;
    setString(memPool, &dio->do_EnabledCompute, dfd->dfd_EnabledCompute);
    dio->do_Weight = dfd->dfd_Weight;
#ifdef MUI
    setString(memPool, &dio->do_MUIFontname, dfd->dfd_Fontname);
#elif (defined GTK)
    setString(memPool, &dio->do_GTKFontname, dfd->dfd_Fontname);
#elif (defined QT)
    setString(memPool, &dio->do_QtFontname, dfd->dfd_Fontname);
#endif
    setString(memPool, &dio->do_Background, dfd->dfd_Background);
    dio->do_Contents.field.titlepos = dfd->dfd_TitlePosition;
#ifdef MUI
    setString(memPool, &dio->do_Contents.field.titleMUIFontname,
            dfd->dfd_TitleFontname);
#elif (defined GTK)
    setString(memPool, &dio->do_Contents.field.titleGTKFontname,
            dfd->dfd_TitleFontname);
#elif (defined QT)
    setString(memPool, &dio->do_Contents.field.titleQtFontname,
            dfd->dfd_TitleFontname);
#endif
    dio->do_Contents.field.home = dfd->dfd_Home;
    dio->do_Contents.field.tabchain = dfd->dfd_TabChain;
    dio->do_Contents.field.readonly = dfd->dfd_ReadOnly;
    dio->do_Contents.field.subType = dfd->dfd_SubType;
    dio->do_Contents.field.alignment = dfd->dfd_Alignment;
    dio->do_Contents.field.calendar = dfd->dfd_Calendar;
    dio->do_Contents.field.format = dfd->dfd_Format;
    dio->do_Contents.field.numdecimals = dfd->dfd_NumDecimals;
    dio->do_Contents.field.controlTable = dfd->dfd_ControlTable;
    dio->do_Contents.field.onDoubleClick = dfd->dfd_OnDoubleClick;
    setString(memPool, &dio->do_Contents.field.doubleClickTrigger,
	    dfd->dfd_DoubleClickTrigger);
    setString(memPool, &dio->do_Contents.field.urlDropTrigger,
	    dfd->dfd_URLDropTrigger);
    setString(memPool, &dio->do_Contents.field.sortDropTrigger,
	    dfd->dfd_SortDropTrigger);
    if (dio->do_Contents.field.autoUpdate != dfd->dfd_AutoUpdate) {
	dio->do_Contents.field.autoUpdate = dfd->dfd_AutoUpdate;
        if (DisplayObjectVisible(dio)) {
            SetProgramNeedRecompile(ProjectProgram(pr), TRUE);
        }
    }
    dio->do_Contents.field.showtitles = dfd->dfd_ShowTitles;
    dio->do_Contents.field.imagesize = dfd->dfd_ImageSize;
    setString(memPool, &dio->do_Contents.field.bubbleHelp,
	    dfd->dfd_BubbleHelp);
    dio->do_Contents.field.recordsKind = dfd->dfd_RecordsKind;
    setString(memPool, &dio->do_Contents.field.recordsCompute, 
	    dfd->dfd_RecordsCompute);
    dio->do_Contents.field.initialRecord = dfd->dfd_InitialRecord;
    dio->do_Contents.field.currentRecord = dfd->dfd_CurrentRecord;
    FreeRecordDescriptionSpec(memPool, 
            dio->do_Contents.field.recordDescriptionSpec);
    dio->do_Contents.field.recordDescriptionSpec = rds;
    dio->do_Contents.field.file = dfd->dfd_File;
    dio->do_Contents.field.font = dfd->dfd_Font;
    dio->do_Contents.field.listview = dfd->dfd_Listview;
    dio->do_Contents.field.view = dfd->dfd_View;
    setString(memPool, &dio->do_Contents.field.listviewCompute, 
	    dfd->dfd_ListviewCompute);
    dio->do_Contents.field.displayexternal = dfd->dfd_DisplayExternal;
    dio->do_Contents.field.titleatstringfield = dfd->dfd_TitleAtStringField;
    dio->do_Contents.field.hidestringfield = dfd->dfd_HideStringField;
    dio->do_Contents.field.show = dfd->dfd_Show;
    dio->do_Contents.field.autoshow = dfd->dfd_AutoShow;
    dio->do_Contents.field.filter = dfd->dfd_Filter;
    dio->do_Contents.field.frame = dfd->dfd_Frame;

    compute = dio->do_EnabledCompute;
    ufAfter = FindUserFunctionName(prg, compute);
    if (cBefore != cAfter || ufBefore != ufAfter) {
        if (DisplayObjectVisible(dio)) {
            SetProgramNeedRecompile(prg, TRUE);
        }
    }
    if (dio->do_Contents.field.home) {       /* new home object? */
        struct DisplayObject *ignore = dio;
        struct List *fields = TableFields(tab);

        if (fields) {
            FORNODE(field, fields) {
                dio = FieldDisplayObject(field);
                if (dio != NULL && dio != ignore) {
                    dio->do_Contents.field.home = FALSE;
		}
            }
	}
    }

    updateTableRecordDescriptionSpec(tab);

    return OK;
}

void DisplayObjectAdd(struct DisplayObject *dio, 
	struct DisplayObject *dioparent, bool isPred)
{
    struct Project *pr = DisplayObjectProject(dio);
    struct Table *tab = DisplayObjectTable(dio);
    struct DisplayObject *pred = dioparent;

    if (dio == NULL) {
        return;
    }

    if (dio->do_Type == DO_TYPE_TABLE) {
        tab = NULL;
    }
    if (dioparent == NULL) {
	dioparent = dio->do_Parent;
    }
    dioparent = getdioparent(pr, tab, dioparent, isPred);

    if (pred != NULL && (PRED(pred) == NULL || pred->do_Parent != dioparent)) {
        pred = NULL;
    }
    movedisplayobject(dio, dioparent, pred);
    SetDisplayObjectVisible(dio, TRUE);
}

bool DisplayObjectRemove(struct DisplayObject *dio)
{
    struct DisplayObject *child, *succ;
    struct List *childs;

    if (dio) {
        switch (dio->do_Type) {
	    case DO_TYPE_REGISTER:
	    case DO_TYPE_GROUP:
	    case DO_TYPE_WINDOW:
            case DO_TYPE_SPACE:
            case DO_TYPE_TEXT:
            case DO_TYPE_IMAGE:
            case DO_TYPE_BALANCE:
		childs = DisplayObjectObjects(dio);
		if (childs) {
		    SAFEFORNODE(child, succ, childs) {
			movedisplayobject(child, dio->do_Parent, PRED(dio));
		    }
		}
		REMOVE(dio);
                FreeDisplayObject(dio);
		break;

            default:
		SetDisplayObjectVisible(dio, FALSE);
		break;
        }
    }
    return TRUE;
}

/*
** DisplaySpace
*/

bool DisplaySpaceBar(struct DisplayObject *dio)
{
    bool bar = FALSE;

    if(dio && dio->do_Type == DO_TYPE_SPACE)
	bar = dio->do_Contents.space.bar;
    return(bar);
}

long DisplayNewSpace(struct Project *pr, struct Table *tab,
        struct DisplayObject *dioparent, struct DisplayObject **new)
{
    struct DisplayObject *dio;
    struct List *childs;
    struct DisplayObject *pred = dioparent;

    if(new)
        *new = NULL;

    dioparent = getdioparent(pr, tab, dioparent, FALSE);
    childs = DisplayObjectObjects(dioparent);
    if(childs == NULL)  	                       /* should never happen */
	return(ERR_INTERNAL);
    if (pred != NULL && (PRED(pred) == NULL || pred->do_Parent != dioparent)) {
        pred = NULL;
    }

    dio = AllocDisplayObject(pr);
    if(dio == NULL)
	return(ERR_MEM);

    dio->do_Weight = 0;
    dio->do_Parent = dioparent;
    dio->do_Type = DO_TYPE_SPACE;
    InsertDisplayObject(childs, dio, pred);

    if(new)
        *new = dio;
    return(OK);
}

void DisplayChangeSpace(struct DisplayObject *dio, struct SpaceData *sd)
{
    if (dio && DisplayObjectType(dio) == DO_TYPE_SPACE) {
	struct MemPool *memPool = ProjectMemPool(dio->do_Project);

        dio->do_Weight = sd->sd_Weight;
	setString(memPool, &dio->do_Background, sd->sd_Background);
        dio->do_Contents.space.bar = sd->sd_Bar;
    }
}

/*
** DisplayObjectText
*/

long DisplayNewText(struct Project *pr, struct Table *tab,
        struct DisplayObject *dioparent, struct DisplayObject **new)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct DisplayObject *dio;
    struct List *childs;
    struct DisplayObject *pred = dioparent;

    if (new) {
        *new = NULL;
    }
    dioparent = getdioparent(pr, tab, dioparent, FALSE);
    childs = DisplayObjectObjects(dioparent);
    if (childs == NULL) {                         /* should never happen */
	return ERR_INTERNAL;
    }
    if (pred != NULL && (PRED(pred) == NULL || pred->do_Parent != dioparent)) {
        pred = NULL;
    }
    dio = AllocDisplayObject(pr);
    if (dio == NULL) {
	return ERR_MEM;
    }
    setString(memPool, &dio->do_Title, "Text");
    dio->do_Parent = dioparent;
    dio->do_Type = DO_TYPE_TEXT;
    InsertDisplayObject(childs, dio, pred);

    if (new) {
        *new = dio;
    }
    return OK;
}

void DisplayChangeText(struct DisplayObject *dio, struct TextData *td)
{
    if (dio && DisplayObjectType(dio) == DO_TYPE_TEXT) {
        struct MemPool *memPool = ProjectMemPool(dio->do_Project);

	setString(memPool, &dio->do_Title, td->td_Title);
        dio->do_Weight = td->td_Weight;
#ifdef MUI
	setString(memPool, &dio->do_MUIFontname, td->td_Fontname);
#elif (defined GTK)
	setString(memPool, &dio->do_GTKFontname, td->td_Fontname);
#elif (defined QT)
	setString(memPool, &dio->do_QtFontname, td->td_Fontname);
#endif
	setString(memPool, &dio->do_Background, td->td_Background);
    }
}

/*
** DisplayImage
*/

char DisplayImageImageSize(struct DisplayObject *dio)
{
    char imagesize = DAD_IMAGESIZE_RESIZABLE;

    if (dio && dio->do_Type == DO_TYPE_IMAGE) {
	imagesize = dio->do_Contents.image.imagesize;
    }
    return imagesize;
}

long DisplayNewImage(struct Project *pr, struct Table *tab,
        struct DisplayObject *dioparent, struct DisplayObject **new)
{
    struct DisplayObject *dio;
    struct List *childs;
    struct DisplayObject *pred = dioparent;

    if (new) {
        *new = NULL;
    }
    dioparent = getdioparent(pr, tab, dioparent, FALSE);
    childs = DisplayObjectObjects(dioparent);
    if (childs == NULL) {                        /* should never happen */
	return ERR_INTERNAL;
    }
    if (pred != NULL && (PRED(pred) == NULL || pred->do_Parent != dioparent)) {
        pred = NULL;
    }
    dio = AllocDisplayObject(pr);
    if (dio == NULL) {
	return ERR_MEM;
    }

    dio->do_Weight = 100;
    dio->do_Parent = dioparent;
    dio->do_Type = DO_TYPE_IMAGE;
    dio->do_Contents.image.imagesize = DAD_IMAGESIZE_RESIZABLE;
    InsertDisplayObject(childs, dio, pred);

    if (new) {
        *new = dio;
    }
    return OK;
}

void DisplayChangeImage(struct DisplayObject *dio, struct ImageData *id)
{
    if (dio && DisplayObjectType(dio) == DO_TYPE_IMAGE) {
	struct MemPool *memPool = ProjectMemPool(dio->do_Project);

        dio->do_Weight = id->id_Weight;
	setString(memPool, &dio->do_Background, id->id_Image);
        dio->do_Contents.image.imagesize = id->id_ImageSize;
    }
}

/*
** DisplayBalance
*/

long DisplayNewBalance(struct Project *pr, struct Table *tab,
        struct DisplayObject *dioparent, struct DisplayObject **new)
{
    struct DisplayObject *dio;
    struct List *childs;
    struct DisplayObject *pred = dioparent;

    if (new) {
        *new = NULL;
    }
    dioparent = getdioparent(pr, tab, dioparent, FALSE);
    childs = DisplayObjectObjects(dioparent);
    if (childs == NULL) {                        /* should never happen */
	return ERR_INTERNAL;
    }
    if (pred != NULL && (PRED(pred) == NULL || pred->do_Parent != dioparent)) {
        pred = NULL;
    }

    dio = AllocDisplayObject(pr);
    if (dio == NULL) {
	return ERR_MEM;
    }

    dio->do_Weight = 0;
    dio->do_Parent = dioparent;
    dio->do_Type = DO_TYPE_BALANCE;
    InsertDisplayObject(childs, dio, pred);

    if (new) {
        *new = dio;
    }
    return OK;
}

/*
** DisplayRegister
*/

char **DisplayRegisterLabels(struct DisplayObject *dio)
{
    char **labels = NULL;

    if (dio != NULL && dio->do_Type == DO_TYPE_REGISTER) {
	labels = dio->do_Contents.reggroup.labels;
    }
    return labels;
}

long DisplayNewRegister(struct Project *pr, struct Table *tab,
	struct DisplayObject **selected, struct DisplayObject *dioparent,
        struct DisplayObject **new)
{
    int i;
    struct DisplayObject *dp = NULL, *dio, *dr;
    struct List *childs;
    struct DisplayObject *pred = dioparent;

    if (new) {
        *new = NULL;
    }

    if (selected) {
    	for (i=0; (dio = selected[i]) != NULL; i++) {
	    if (dio->do_Parent == NULL) {
                return ERR_INTERNAL;
	    }
            if (dp == NULL) {
		dp = dio->do_Parent;
	        pred = PRED(dio);
	    }
	    if (dio->do_Parent != dp) {
	        return DPYERR_OBJECTLEVEL;
	    }
        }
    }
    if (dp) {
	dioparent = dp;
    }
    dioparent = getdioparent(pr, tab, dioparent, FALSE);
    childs = DisplayObjectObjects(dioparent);
    if (childs == NULL) {	                /* should never happen */
	return ERR_INTERNAL;
    }
    if (pred != NULL && (PRED(pred) == NULL || pred->do_Parent != dioparent)) {
	pred = NULL;
    }
    dr = AllocDisplayObject(pr);
    if (dr == NULL) {
	return ERR_MEM;
    }

    dr->do_Parent = dioparent;
    dr->do_Type = DO_TYPE_REGISTER;
    dr->do_Contents.reggroup.labels = NULL;
    dr->do_Contents.reggroup.open = TRUE;
    InsertDisplayObject(childs, dr, pred);

    if (selected) {
    	for (i=0, dio = NULL; selected[i] != NULL; dio = selected[i], i++) {
            movedisplayobject(selected[i], dr, dio);
	}
    }

    if (new) {
        *new = dr;
    }
    return OK;
}

long DisplayChangeRegister(struct DisplayObject *reg, struct RegisterData *rd)
{
    if (reg && DisplayObjectType(reg) == DO_TYPE_REGISTER && rd) {
	struct Project *pr = DisplayObjectProject(reg);
	struct MemPool *memPool = ProjectMemPool(pr);
	char **copy = NULL;

	copy = CopyLabels(memPool, rd->rd_Labels);
	if (copy == NULL) {
	    return ERR_MEM;
	}

        reg->do_Weight = rd->rd_Weight;
#ifdef MUI
	setString(memPool, &reg->do_MUIFontname, rd->rd_Fontname);
#elif (defined GTK)
	setString(memPool, &reg->do_GTKFontname, rd->rd_Fontname);
#elif (defined QT)
	setString(memPool, &reg->do_QtFontname, rd->rd_Fontname);
#endif
	setString(memPool, &reg->do_Background, rd->rd_Background);
	FreeLabels(memPool, reg->do_Contents.reggroup.labels);
	reg->do_Contents.reggroup.labels = copy;

	return OK;
    }
    return ERR_INTERNAL;
}

/*
** DisplayGroup
*/

bool DisplayGroupBorder(struct DisplayObject *dio)
{
    bool border = FALSE;

    if (dio && dio->do_Type == DO_TYPE_GROUP) {
	border = dio->do_Contents.group.border;
    }
    return border;
}

bool DisplayGroupHoriz(struct DisplayObject *dio)
{
    bool horiz = FALSE;

    if (dio && dio->do_Type == DO_TYPE_GROUP) {
	horiz = dio->do_Contents.group.horiz;
    }
    return horiz;
}

bool DisplayGroupSpacing(struct DisplayObject *dio)
{
    bool spacing = TRUE;

    if (dio && dio->do_Type == DO_TYPE_GROUP) {
	spacing = dio->do_Contents.group.spacing;
    }
    return spacing;
}

long DisplayNewGroup(struct Project *pr, struct Table *tab,
        struct DisplayObject **selected, struct DisplayObject *dioparent,
        struct DisplayObject **new)
{
    int i;
    struct DisplayObject *dp = NULL, *dio, *dg;
    struct List *childs;
    struct DisplayObject *pred = dioparent;

    if (new) {
        *new = NULL;
    }

    if (selected) {
    	for (i=0; (dio = selected[i]) != NULL; i++) {
	    if (dio->do_Parent == NULL) {
                return ERR_INTERNAL;
	    }
            if (dp == NULL) {
		dp = dio->do_Parent;
	        pred = PRED(dio);
	    }
	    if (dio->do_Parent != dp) {
	        return DPYERR_OBJECTLEVEL;
	    }
        }
    }
    if (dp) {
	dioparent = dp;
    }
    dioparent = getdioparent(pr, tab, dioparent, FALSE);
    childs = DisplayObjectObjects(dioparent);
    if (childs == NULL) {	                /* should never happen */
	return ERR_INTERNAL;
    }
    if (pred != NULL && (PRED(pred) == NULL || pred->do_Parent != dioparent)) {
	pred = NULL;
    }

    dg = AllocDisplayObject(pr);
    if (dg == NULL) {
	return ERR_MEM;
    }

    dg->do_Parent = dioparent;
    dg->do_Type = DO_TYPE_GROUP;
    dg->do_Contents.group.spacing = TRUE;
    dg->do_Contents.group.open = TRUE;
    InsertDisplayObject(childs, dg, pred);

    if (selected) {
    	for (i=0, dio = NULL; selected[i] != NULL; dio = selected[i], i++) {
            movedisplayobject(selected[i], dg, dio);
	}
    }

    if (new) {
        *new = dg;
    }
    return OK;
}

void DisplayChangeGroup(struct DisplayObject *group, struct GroupData *gd)
{
    if (group && DisplayObjectType(group) == DO_TYPE_GROUP && gd) {
        struct MemPool *memPool = ProjectMemPool(group->do_Project);

	setString(memPool, &group->do_Title, gd->gd_Title);
        group->do_Weight = gd->gd_Weight;
#ifdef MUI
	setString(memPool, &group->do_MUIFontname, gd->gd_Fontname);
#elif (defined GTK)
	setString(memPool, &group->do_GTKFontname, gd->gd_Fontname);
#elif (defined QT)
	setString(memPool, &group->do_QtFontname, gd->gd_Fontname);
#endif
	setString(memPool, &group->do_Background, gd->gd_Background);
	group->do_Contents.group.horiz = gd->gd_Horiz;
	group->do_Contents.group.border = gd->gd_Border;
	group->do_Contents.group.spacing = gd->gd_Spacing;
    }
}

/*
** DisplayWindow
*/

#ifdef MUI
static struct DisplayObject *displayWindowIdInUseRec(struct DisplayObject *dio,
	ulong id, struct DisplayObject *leaveout)
{
    struct List *childs = DisplayObjectObjects(dio);
    struct DisplayObject *ret;

    if (DisplayWindowId(dio) == id && dio != leaveout) {
	return dio;
    }

    if (childs) {
	FORNODE(dio, childs) {
	    ret = displayWindowIdInUseRec(dio, id, leaveout);
	    if (ret) {
		return ret;
	    }
	}
    }
    return NULL;
}

struct DisplayObject *DisplayWindowIdInUse(ulong id, 
	struct DisplayObject *leaveout)
{
    struct Project *pr;
    struct DisplayObject *root, *dio;

    if (id) {
	FORNODE(pr, &ProjectList) {
	    root = ProjectRootDisplayObject(pr);
	    dio = displayWindowIdInUseRec(root, id, leaveout);
	    if (dio) {
		return dio;
	    }
	}
    }
    return NULL;
}
#endif

ulong DisplayWindowId(struct DisplayObject *dio)
{
    ulong id = 0;

    if (dio && dio->do_Type == DO_TYPE_WINDOW) {
	id = dio->do_Contents.window.id;
    }
    return id;
}

void SetDisplayWindowId(struct DisplayObject *dio, ulong id)
{
    if (dio && dio->do_Type == DO_TYPE_WINDOW) {
	dio->do_Contents.window.id = id;
    }
}

const struct WindowDimensions *DisplayWindowDimensions(struct DisplayObject*dio)
{
    const struct WindowDimensions *wd = NULL;

    if (dio && dio->do_Type==DO_TYPE_WINDOW &&dio->do_Contents.window.wd_valid){
	wd = &dio->do_Contents.window.wd;
    }
    return wd;
}

void SetDisplayWindowDimensions(struct DisplayObject *dio, 
	struct WindowDimensions *wd)
{
    if (dio && dio->do_Type == DO_TYPE_WINDOW) {
        if (wd != NULL) {
            dio->do_Contents.window.wd = *wd;
            dio->do_Contents.window.wd_valid = TRUE;
        }
    }
}

bool DisplayWindowOpen(struct DisplayObject *dio)
{
    bool windowopen = FALSE;

    if (dio && dio->do_Type == DO_TYPE_WINDOW) {
	windowopen = dio->do_Contents.window.windowopen;
    }
    return windowopen;
}

void SetDisplayWindowOpen(struct DisplayObject *dio, bool windowopen)
{
    if (dio && dio->do_Type == DO_TYPE_WINDOW) {
	dio->do_Contents.window.windowopen = windowopen;
    }
}

bool DisplayWindowButton(struct DisplayObject *dio)
{
    bool button = FALSE;

    if (dio && dio->do_Type == DO_TYPE_WINDOW) {
        button = dio->do_Contents.window.button;
    }
    return button;
}

bool DisplayWindowCloseWhenDisabled(struct DisplayObject *dio)
{
    bool close = FALSE;

    if (dio && dio->do_Type == DO_TYPE_WINDOW) {
	close = dio->do_Contents.window.closeWhenDisabled;
    }
    return close;
}

GuiObject DisplayWindowButtonObject(struct DisplayObject *dio)
{
    GuiObject button = NULL;

    if (dio && dio->do_Type == DO_TYPE_WINDOW) {
	button = dio->do_Contents.window.button_go;
    }
    return button;
}

void SetDisplayWindowButtonObject(struct DisplayObject *dio, GuiObject button)
{
    if (dio && dio->do_Type == DO_TYPE_WINDOW) {
	dio->do_Contents.window.button_go = button;
    }
}

long DisplayNewWindow(struct Project *pr, struct Table *tab,
        struct DisplayObject **selected, struct DisplayObject *dioparent,
        struct DisplayObject **new)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct DisplayObject *dp = NULL, *dio, *dw;
    struct List *childs;
    struct DisplayObject *pred = dioparent;
    const char *title = NULL;
    int i;

    if (new) {
        *new = NULL;
    }

    if (selected) {
    	for (i=0; (dio = selected[i]) != NULL; i++) {
	    if (dio->do_Parent == NULL) {
                return ERR_INTERNAL;
	    }
            if (dp == NULL) {
		dp = dio->do_Parent;
	        pred = PRED(dio);
	    }
	    if (dio->do_Parent != dp) {
	        return DPYERR_OBJECTLEVEL;
	    }
	    if (dio->do_Type == DO_TYPE_TABLE) {
		title = TableName(DisplayObjectTable(dio));
	    }
	    if (dio->do_Type == DO_TYPE_FIELD) {
		title = FieldName(DisplayFieldField(dio));
	    }
        }
	if (i > 1) {	    /* fall back to default if more than one item */
	    title = NULL;
	}
    }
    if (dp) {
	dioparent = dp;
    }
    dioparent = getdioparent(pr, tab, dioparent, FALSE);
    childs = DisplayObjectObjects(dioparent);
    if (childs == NULL) {	                    /* should never happen */
	return ERR_INTERNAL;
    }
    if (pred != NULL && (PRED(pred) == NULL || pred->do_Parent != dioparent)) {
	pred = NULL;
    }

    dw = AllocDisplayObject(pr);
    if (dw == NULL) {
	return ERR_MEM;
    }

    if (title == NULL || title[0] == 0) {
	title = GCS(MSG_Window);
    }

    dw->do_Parent = dioparent;
    dw->do_Type = DO_TYPE_WINDOW;
    setString(memPool, &dw->do_Title, title);
    dw->do_Enabled = DOD_ENABLED_ALWAYS;
    dw->do_Contents.window.id = 0;
    dw->do_Contents.window.button = TRUE;
    dw->do_Contents.window.open = TRUE;
    InsertDisplayObject(childs, dw, pred);

    if (selected) {
    	for (i=0, dio = NULL; selected[i] != NULL; dio = selected[i], i++) {
            movedisplayobject(selected[i], dw, dio);
	}
    }

    if (new) {
        *new = dw;
    }
    return OK;
}

void DisplayChangeWindow(struct DisplayObject *wi, struct WindowData *wd)
{
    if (wi && DisplayObjectType(wi) == DO_TYPE_WINDOW && wd) {
        struct Project *pr = wi->do_Project;
        struct Program *prg = ProjectProgram(pr);
        struct MemPool *memPool = ProjectMemPool(pr);
        bool cBefore = (wi->do_Enabled == DOD_ENABLED_COMPUTED);
        bool cAfter = (wd->wd_Enabled == DOD_ENABLED_COMPUTED);
        const char *compute = wi->do_EnabledCompute;
        struct UserFunction *ufBefore = FindUserFunctionName(prg, compute);
        struct UserFunction *ufAfter;

	setString(memPool, &wi->do_Title, wd->wd_Title);
	wi->do_Shortcut = wd->wd_Shortcut;
	wi->do_Enabled = wd->wd_Enabled;
        setString(memPool, &wi->do_EnabledCompute, wd->wd_EnabledCompute);
        wi->do_Weight = wd->wd_Weight;
#ifdef MUI
	setString(memPool, &wi->do_MUIFontname, wd->wd_Fontname);
#elif (defined GTK)
	setString(memPool, &wi->do_GTKFontname, wd->wd_Fontname);
#elif (defined QT)
	setString(memPool, &wi->do_QtFontname, wd->wd_Fontname);
#endif
	setString(memPool, &wi->do_Background, wd->wd_Background);
	wi->do_Contents.window.id = wd->wd_Id;
	wi->do_Contents.window.button = wd->wd_Button;
	wi->do_Contents.window.closeWhenDisabled = wd->wd_CloseWhenDisabled;

        compute = wi->do_EnabledCompute;
        ufAfter = FindUserFunctionName(prg, compute);
        if (cBefore != cAfter || ufBefore != ufAfter) {
            if (DisplayObjectVisible(wi)) {
		SetProgramNeedRecompile(prg, TRUE);
	    }
        }
    }
}

bool DisplayObjectUp(struct DisplayObject *dio)
{
    struct DisplayObject *pred;

    if (dio == NULL) {
	return FALSE;
    }

    pred = PRED(dio);
    if (PRED(pred) == NULL) {			/* predecessor exists? */
        return FALSE;
    }

    REMOVE(dio);
    InsertDisplayObject(NULL, dio, PRED(pred));
    return TRUE;
}

bool DisplayObjectDown(struct DisplayObject *dio)
{
    struct DisplayObject *succ;

    if (dio == NULL) {
	return FALSE;
    }

    succ = SUCC(dio);
    if (SUCC(succ) == NULL) {			/* successor exists? */
        return FALSE;
    }
    REMOVE(dio);
    InsertDisplayObject(NULL, dio, succ);
    return TRUE;
}

bool DisplayObjectIn(struct DisplayObject *dio)
{
    struct DisplayObject *succ;

    if (dio == NULL) {
	return FALSE;
    }

    for (succ = SUCC(dio); SUCC(dio); dio = SUCC(dio)) {
        if (DisplayObjectVisible(dio)) {
            break;
	}
    }

    switch (succ->do_Type) {
        case DO_TYPE_GROUP:	/* remove dio and insert it to group's childs */
        case DO_TYPE_WINDOW:
        case DO_TYPE_REGISTER:
    	    movedisplayobject(dio, succ, NULL);
            return TRUE;
    }
    return FALSE;
}

bool DisplayObjectOut(struct DisplayObject *dio)
{
    struct DisplayObject *parent = dio? dio->do_Parent : NULL;

    if (parent == NULL || parent->do_Parent == NULL || 
	    parent->do_Type == DO_TYPE_TABLE) {
        return FALSE;             /* cannot move something out of table/root */
    }

    /* move dio into higher level and place it before parent */
    movedisplayobject(dio, parent->do_Parent, PRED(parent));
    return TRUE;
}

static bool DisplayListHasContent(struct List *childs)
{
    struct DisplayObject *dio;

    if (childs == NULL) {
	return FALSE;                           /* empty */
    }
    FORNODE(dio, childs) {
        if (!DisplayObjectVisible(dio)) {
            continue;
        }
        if (dio->do_Type == DO_TYPE_FIELD) {
            return TRUE;			/* There is an field */
        }
        if (dio->do_Type == DO_TYPE_TEXT) {
            return TRUE;                        /* There is some text */
        }
        if (dio->do_Type == DO_TYPE_IMAGE) {
            return TRUE;                        /* There is some image */
        }
        if (dio->do_Type != DO_TYPE_WINDOW || DisplayWindowButton(dio)) {
            if (DisplayListHasContent(DisplayObjectObjects(dio))) {
                return TRUE;
            }
        }
    }
    return FALSE;
}

bool DisplayObjectVisibleGui(struct DisplayObject *dio)
{
    struct List *childs = DisplayObjectObjects(dio);

    if (!DisplayObjectVisible(dio)) {
        return FALSE;            /* object invisible */
    }
    if (childs == NULL) {
        return TRUE;             /* non-composite visible object */
    }
    return DisplayListHasContent(childs);
}


/*
** Load & Save of display objects
*/

long NumDisplayObjects(struct DisplayObject *dio)
{
    long num = 0;
    struct List *childs = DisplayObjectObjects(dio);

    if (dio) {
	num++;
    }
    if (childs) {
	FORNODE(dio, childs) {
	    num += NumDisplayObjects(dio);
	}
    }
    return num;
}

static long copyString(const char *str, struct MemPool *memPool, char **ptr)
{
    char *copy = AllocString(memPool, str);

    FreeString(memPool, *ptr);
    *ptr = copy;

    if (copy == NULL && str != NULL) {
        return ERR_MEM;
    }
    return OK;
}

long ReadDisplayObject(struct Project *pr,
        const struct ReadDisplayObjectData *data)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct DisplayObject *dio;
    long error = OK;

    if (pr == NULL || data == NULL) {
        return ERR_INTERNAL;
    }

    if (data->rdod_ID == 1) {			/* root display object */
        dio = ProjectRootDisplayObject(pr);
        if (dio == NULL) {
            return DISKERR_INVALID;
        }
    } else {
        struct List *dioparentChilds = NULL;

        dio = AllocDisplayObject(pr);
        if (dio == NULL) {
            return ERR_MEM;
        }
        dio->do_Parent = data->rdod_Parent;
        dioparentChilds = DisplayObjectObjects(dio->do_Parent);
        if (dioparentChilds == NULL) {
            return DISKERR_INVALID;
        }
        ADDTAIL(dioparentChilds, dio);
    }

    dio->do_Visible = data->rdod_Visible;
    error = copyString(data->rdod_Title, memPool, &dio->do_Title);
    if (error) {
        return error;
    }
    dio->do_Shortcut = data->rdod_Shortcut;
    dio->do_Enabled = data->rdod_Enabled;
    error = copyString(data->rdod_EnabledCompute, memPool,
            &dio->do_EnabledCompute);
    if (error) {
        return error;
    }
    error = copyString(data->rdod_QtFontname, memPool, &dio->do_QtFontname);
    if (error) {
        return error;
    }
    error = copyString(data->rdod_GTKFontname, memPool, &dio->do_GTKFontname);
    if (error) {
        return error;
    }
    error = copyString(data->rdod_MUIFontname, memPool, &dio->do_MUIFontname);
    if (error) {
        return error;
    }
    error = copyString(data->rdod_Background, memPool, &dio->do_Background);
    if (error) {
        return error;
    }
    dio->do_Weight = data->rdod_Weight;
    dio->do_Type = data->rdod_Type;

    switch (dio->do_Type) {
        case DO_TYPE_TABLE:
        {
            struct Table *tab = data->rdod_Contents.table.table;
            struct RecordDescriptionSpec *rds = NULL;

            if (tab == NULL) {
                return DISKERR_INVALID;
            }
            dio->do_Contents.table.table = tab;
            SetTableDisplayObject(tab, dio);
            dio->do_Contents.table.activeFieldNum =
                NodeNum(data->rdod_Contents.table.activeField);
            dio->do_Contents.table.activeDisplayObjectNum =
                data->rdod_Contents.table.activeDisplayObjectID;
            dio->do_Contents.table.hasPanel = 
                data->rdod_Contents.table.hasPanel;
            error = copyString(data->rdod_Contents.table.panelBackground,
                    memPool, &dio->do_Contents.table.panel.background);
            if (error) {
                return error;
            }
            dio->do_Contents.table.panel.numall = 
                data->rdod_Contents.table.panelNumAll;
            dio->do_Contents.table.panel.filter =
                data->rdod_Contents.table.panelFilter;
            dio->do_Contents.table.panel.arrows =
                data->rdod_Contents.table.panelArrows;

            rds = CopyRecordDescriptionSpec(memPool,
                    data->rdod_Contents.table.recordDescriptionSpec);
            if (rds == NULL && 
                    data->rdod_Contents.table.recordDescriptionSpec != NULL) {
                return ERR_MEM;
            }
            FreeRecordDescriptionSpec(memPool,
                    dio->do_Contents.table.recordDescriptionSpec);
            dio->do_Contents.table.recordDescriptionSpec = rds;
            break;
        }

        case DO_TYPE_FIELD:
        {
            struct Field *field = data->rdod_Contents.field.field;
            char **labels = NULL;
            struct RecordDescriptionSpec *rds = NULL;

            if (field == NULL) {
                return DISKERR_INVALID;
            }
            dio->do_Contents.field.field = field;
            SetFieldDisplayObject(field, dio);

            dio->do_Contents.field.titlepos =
                data->rdod_Contents.field.titlePosition;
            error = copyString(data->rdod_Contents.field.titleQtFontname,
                    memPool, &dio->do_Contents.field.titleQtFontname);
            if (error) {
                return error;
            }
            error = copyString(data->rdod_Contents.field.titleGTKFontname,
                    memPool, &dio->do_Contents.field.titleGTKFontname);
            if (error) {
                return error;
            }
            error = copyString(data->rdod_Contents.field.titleMUIFontname,
                    memPool, &dio->do_Contents.field.titleMUIFontname);
            if (error) {
                return error;
            }
            dio->do_Contents.field.home = data->rdod_Contents.field.home;
            dio->do_Contents.field.tabchain= data->rdod_Contents.field.tabChain;
            dio->do_Contents.field.readonly= data->rdod_Contents.field.readOnly;
            dio->do_Contents.field.subType = data->rdod_Contents.field.subType;
            dio->do_Contents.field.alignment = 
                data->rdod_Contents.field.alignment;
            dio->do_Contents.field.calendar =
                data->rdod_Contents.field.calendar;
            dio->do_Contents.field.format = data->rdod_Contents.field.format;
            dio->do_Contents.field.numdecimals =
                data->rdod_Contents.field.numDecimals;
            dio->do_Contents.field.imagesize =
                data->rdod_Contents.field.imageSize;
            error = copyString(data->rdod_Contents.field.bubbleHelp, memPool,
                    &dio->do_Contents.field.bubbleHelp);
            if (error) {
                return error;
            }
            dio->do_Contents.field.displayexternal =
                data->rdod_Contents.field.displayExternal;
            dio->do_Contents.field.titleatstringfield =
                data->rdod_Contents.field.titleAtStringField;
            dio->do_Contents.field.hidestringfield =
                data->rdod_Contents.field.hideStringField;
            dio->do_Contents.field.file = data->rdod_Contents.field.file;
            dio->do_Contents.field.font = data->rdod_Contents.field.font;
            dio->do_Contents.field.listview= data->rdod_Contents.field.listview;
            labels = StringToLabels(memPool,
                    data->rdod_Contents.field.listviewLabels);
            if (labels == NULL && 
                    data->rdod_Contents.field.listviewLabels != NULL) {
                return ERR_MEM;
            }
            FreeLabels(memPool, dio->do_Contents.field.listviewLabels);
            dio->do_Contents.field.listviewLabels = labels;
            error = copyString(data->rdod_Contents.field.listviewCompute,
                    memPool, &dio->do_Contents.field.listviewCompute);
            if (error) {
                return error;
            }
            dio->do_Contents.field.view = data->rdod_Contents.field.view;
            dio->do_Contents.field.frame = data->rdod_Contents.field.frame;
            dio->do_Contents.field.recordsKind =
                data->rdod_Contents.field.recordsKind;
            error = copyString(data->rdod_Contents.field.recordsCompute,
                    memPool, &dio->do_Contents.field.recordsCompute);
            if (error) {
                return error;
            }
            dio->do_Contents.field.initialRecord =
                data->rdod_Contents.field.initialRecord;
            dio->do_Contents.field.currentRecord =
                data->rdod_Contents.field.currentRecord;
            dio->do_Contents.field.show = data->rdod_Contents.field.show;
            dio->do_Contents.field.autoshow= data->rdod_Contents.field.autoShow;
            dio->do_Contents.field.filter = data->rdod_Contents.field.filter;
            dio->do_Contents.field.autoUpdate =
                data->rdod_Contents.field.autoUpdate;
            dio->do_Contents.field.showtitles =
                data->rdod_Contents.field.showTitles;
            dio->do_Contents.field.controlTable =
                data->rdod_Contents.field.controlTable;
            dio->do_Contents.field.onDoubleClick =
                data->rdod_Contents.field.onDoubleClick;
            error = copyString(data->rdod_Contents.field.doubleClickTrigger, 
                    memPool, &dio->do_Contents.field.doubleClickTrigger);
            if (error) {
                return error;
            }
            error = copyString(data->rdod_Contents.field.urlDropTrigger, 
                    memPool, &dio->do_Contents.field.urlDropTrigger);
            if (error) {
                return error;
            }
            error = copyString(data->rdod_Contents.field.sortDropTrigger, 
                    memPool, &dio->do_Contents.field.sortDropTrigger);
            if (error) {
                return error;
            }

            rds = CopyRecordDescriptionSpec(memPool,
                    data->rdod_Contents.field.recordDescriptionSpec);
            if (rds == NULL && 
                    data->rdod_Contents.field.recordDescriptionSpec != NULL) {
                return ERR_MEM;
            }
            FreeRecordDescriptionSpec(memPool,
                    dio->do_Contents.field.recordDescriptionSpec);
            dio->do_Contents.field.recordDescriptionSpec = rds;
            break;
        }

        case DO_TYPE_SPACE:
            dio->do_Contents.space.bar = data->rdod_Contents.space.bar;
            break;

        case DO_TYPE_TEXT:
            break;

        case DO_TYPE_IMAGE:
            dio->do_Contents.image.imagesize =
                data->rdod_Contents.image.imageSize;
            break;

        case DO_TYPE_BALANCE:
            break;

        case DO_TYPE_GROUP:
            dio->do_Contents.group.border = data->rdod_Contents.group.border;
            dio->do_Contents.group.horiz = data->rdod_Contents.group.horizontal;
            dio->do_Contents.group.spacing = data->rdod_Contents.group.spacing;
            dio->do_Contents.group.open = data->rdod_Contents.group.open;
            break;

        case DO_TYPE_REGISTER:
        {
            char **labels;

            dio->do_Contents.reggroup.open = data->rdod_Contents.reg.open;
            labels = StringToLabels(memPool,
                    data->rdod_Contents.reg.labels);
            if (labels == NULL && data->rdod_Contents.reg.labels != NULL) {
                return ERR_MEM;
            }
            FreeLabels(memPool, dio->do_Contents.reggroup.labels);
            dio->do_Contents.reggroup.labels = labels;
            break;
        }

        case DO_TYPE_WINDOW:
        {
            const struct WindowDimensions *wd =
                &data->rdod_Contents.window.dimensions;

            dio->do_Contents.window.id = data->rdod_Contents.window.windowId;
            dio->do_Contents.window.button = data->rdod_Contents.window.button;
            dio->do_Contents.window.closeWhenDisabled =
                data->rdod_Contents.window.closeWhenDisabled;
            dio->do_Contents.window.windowopen = 
                data->rdod_Contents.window.windowOpen;
            dio->do_Contents.window.wd_valid = 
                wd->wd_Width >=0 && wd->wd_Height >= 0;
            dio->do_Contents.window.wd = *wd;
            dio->do_Contents.window.open = data->rdod_Contents.window.open;
            /* activeTableNum and activeDisplayObjectNum are already set. */
            break;
        }

        default:
            return DISKERR_INVALID;
    }

    if (error == OK && DisplayObjectNum(dio) != data->rdod_ID) {
        error = DISKERR_NOTABEEBASEPROJECT;
    }

    return error;
}

static void removeObsolete(struct DisplayObject *dio)
{
    if (dio) {
        if (dio->do_Type == DO_TYPE_PANEL_OBSOLETE) {
            REMOVE(dio);
            FreeDisplayObject(dio);
        } else {
            struct List *childs = DisplayObjectObjects(dio);
            struct DisplayObject *succ;

            if (childs) {
                SAFEFORNODE(dio, succ, childs) {
                    removeObsolete(dio);
                }
            }
        }
    }
}

static struct RecordDescriptionSpec *readRecordDescriptionSpecCustomFormat(
        FILE *file, bool utf8, struct Project *pr, long *error)
{
    struct RecordDescriptionSpec *rds = NULL;

    *error = OK;

    if (FGetBool(file)) {
        struct MemPool *memPool = ProjectMemPool(pr);
        short num, i;
        size_t warnings = 0;
        long fieldnum;

        num = FGetShort(file);
        rds = AllocRecordDescriptionSpec(memPool, num);
        if (rds == NULL) {
            *error = ERR_MEM;
            return NULL;
        }

        rds->rds_Type = FGetChar(file);
        rds->rds_Compute = ReadStringConvert(file, utf8, memPool, 
                error, &warnings);
        if (*error) {
            FreeRecordDescriptionSpec(memPool, rds);
            return NULL;
        }
        ProjectAddCodesetWarnings(pr, warnings);

        for (i=0; i < num; i++) {
            fieldnum = FGetLong(file);
            rds->rds_Fields[i] = (fieldnum == ~0)?
                RDS_FIELD_RECNUM : FindFieldNum(pr, fieldnum);
        }
        rds->rds_MultiColumn = FGetBool(file);
    }

    return rds;
}

static long readStringAndUpdateCustomFormat(FILE *file, bool utf8,
        struct MemPool *memPool, char **ptr, size_t *warnings)
{
    long error;

    FreeString(memPool, *ptr);
    *ptr = ReadStringConvert(file, utf8, memPool, &error, warnings);

    return error;
}

#define DOVERSION 45
#define MIN_DOVERSION 17

long LoadDisplayObjectsCustomFormat(struct Project *pr, FILE *file)
{
    short version;
    bool utf8 = FALSE;
    long num, i;
    struct MemPool *memPool = ProjectMemPool(pr);
    struct DisplayObject *dio;
    long error = OK;

    if (pr == NULL || file == NULL) {
    	return ERR_INTERNAL;
    }

    version = FGetShort(file);
    if (version >= 31) {
        utf8 = FGetBool(file);
    }
    num = FGetLong(file);
    if (version < MIN_DOVERSION || version > DOVERSION) {
	return DISKERR_VERSION;
    }

    for (i = 0; i < num; i++) {
        long parent_num;
        size_t warnings;
        char type;

        parent_num = FGetLong(file);

	if (i == 0) {				/* root display object */
	    dio = ProjectRootDisplayObject(pr);
    	    if (dio == NULL) {
		return DISKERR_INVALID;
	    }
        } else {
            struct List *dioparent_childs = NULL;

            dio = AllocDisplayObject(pr);
            if (dio == NULL) {
                return ERR_MEM;
            }
            dio->do_Parent = FindDisplayObjectNum(pr, parent_num);
            dioparent_childs = DisplayObjectObjects(dio->do_Parent);
            if (dioparent_childs == NULL) {
                return DISKERR_INVALID;
            }
            ADDTAIL(dioparent_childs, dio);
        }

        dio->do_Weight = FGetShort(file);
        type = FGetChar(file);
        dio->do_Type = type;
        if (version >= 20) {
            error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                    &dio->do_MUIFontname, &warnings);
            if (error) {
                return error;
            }
            ProjectAddCodesetWarnings(pr, warnings);
	}
        if (version >= 25) {
            error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                    &dio->do_GTKFontname, &warnings);
            if (error) {
                return error;
            }
            ProjectAddCodesetWarnings(pr, warnings);
	}
        if (version >= 43) {
            error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                    &dio->do_QtFontname, &warnings);
            if (error) {
                return error;
            }
            ProjectAddCodesetWarnings(pr, warnings);
	}
        error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                &dio->do_Background, &warnings);
        if (error) {
            return error;
        }
        ProjectAddCodesetWarnings(pr, warnings);

	switch (dio->do_Type) {
	    case DO_TYPE_TABLE:
            {
                long tabnum = FGetLong(file);
                struct Table *tab = FindTableNum(pr, tabnum);
                struct RecordDescriptionSpec *rds;

		if (tab == NULL) {
                    return DISKERR_INVALID;
		}
                dio->do_Contents.table.table = tab;
		SetTableDisplayObject(tab, dio);
                dio->do_Visible = FGetBool(file);
                dio->do_Contents.table.activeFieldNum = FGetLong(file);
                dio->do_Contents.table.activeDisplayObjectNum = FGetLong(file);
                if (version >= 41) {
                    error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                            &dio->do_Title, &warnings);
                    if (error) {
                        return error;
                    }
                    ProjectAddCodesetWarnings(pr, warnings);
                    dio->do_Contents.table.hasPanel = FGetBool(file);
                    error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                            &dio->do_Contents.table.panel.background,
                            &warnings);
                    if (error) {
                        return error;
                    }
                    ProjectAddCodesetWarnings(pr, warnings);
                    dio->do_Contents.table.panel.numall = FGetBool(file);
                    dio->do_Contents.table.panel.filter = FGetBool(file);
                    dio->do_Contents.table.panel.arrows = FGetBool(file);
                }
                if (version >= 42) {
                    rds = readRecordDescriptionSpecCustomFormat(file, utf8, pr,
                            &error);
                    if (error) {
                        return error;
                    }
                    FreeRecordDescriptionSpec(memPool,
                            dio->do_Contents.table.recordDescriptionSpec);
                    dio->do_Contents.table.recordDescriptionSpec = rds;
                }
		break;
            }

	    case DO_TYPE_FIELD:
            {
                long fieldnum;
                struct Field *field;
                short type;
                struct RecordDescriptionSpec *rds = NULL;
		char *buf, sc[2];

                fieldnum = FGetLong(file);
		field = FindFieldNum(pr, fieldnum);
		if (field == NULL) {
                    return DISKERR_INVALID;
		}
                dio->do_Contents.field.field = field;
                type = FieldType(field);
		SetFieldDisplayObject(field, dio);
                dio->do_Visible = FGetBool(file);
		if (version < 32) {
                    sc[0] = FGetChar(file);
		    sc[1] = 0;
                    buf = CodesetConvertFromDefault8Bit(sc, memPool, &error);
		} else {
		    buf = ReadStringConvert(file, utf8, memPool,
                            &error, &warnings);
                    ProjectAddCodesetWarnings(pr, warnings);
		}
		if (error) {
		    return error;
		}
		dio->do_Shortcut = CodesetGetChar(EmptyNullStr(buf));
		if (buf != sc) {
		    FreeString(memPool, buf);
		}
                dio->do_Contents.field.readonly = FGetBool(file);
                dio->do_Contents.field.home = FGetBool(file);
		if (version >= 33) {
		    dio->do_Contents.field.tabchain = FGetBool(file);
		} else {
		    dio->do_Contents.field.tabchain = TRUE;
		}
                dio->do_Contents.field.alignment = FGetChar(file);
                dio->do_Contents.field.calendar = TRUE;
		if (version >= 29) {
		    dio->do_Contents.field.format = FGetChar(file);
		}
                dio->do_Contents.field.numdecimals = FGetChar(file);
		if (version >= 38) {
                    dio->do_Enabled = FGetChar(file);
                    error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                            &dio->do_EnabledCompute, &warnings);
                    if (error) {
                        return error;
                    }
                    ProjectAddCodesetWarnings(pr, warnings);
                } else {
                    dio->do_Enabled = DOD_ENABLED_NOTINITIAL;
                }
                dio->do_Contents.field.file = FGetBool(file);
                dio->do_Contents.field.font = FGetBool(file);
                dio->do_Contents.field.listview = FGetBool(file);
                dio->do_Contents.field.view = FGetBool(file);
                dio->do_Contents.field.displayexternal = FGetBool(file);
                if (version < 39) {
                    dio->do_Contents.field.imagesize = FGetChar(file);
                }
		dio->do_Contents.field.titleatstringfield = FGetBool(file);
                dio->do_Contents.field.hidestringfield = FGetBool(file);
		if (version >= 27 && version < 42) {
		    FGetChar(file);     /* was quick-search-fields */
                    FGetBool(file);     /* was quick-search-prefix-only */
		}
                dio->do_Contents.field.show = FGetBool(file);
		dio->do_Contents.field.autoshow = FGetBool(file);
		if (version >= 23) {
		    dio->do_Contents.field.filter = FGetBool(file);
		} else {
		    dio->do_Contents.field.filter =
			dio->do_Contents.field.autoshow;
		}
                if (version >= 22) {
                    dio->do_Contents.field.frame = FGetBool(file);
		}
                dio->do_Contents.field.subType = FGetChar(file);
                if (version < 21 && type == TYPE_CHOICE) {
                    dio->do_Contents.field.subType = 
			DAD_CHOICETYPE_CYCLEBUTTON;
		}
                dio->do_Contents.field.autoUpdate = FGetBool(file);
		if (version >= 30 && version < 38) {    /* old update */
		    dio->do_Contents.field.autoUpdate |= FGetBool(file);
		}

                dio->do_Contents.field.showtitles = FGetBool(file);
		if (version >= 24) {
		    dio->do_Contents.field.onDoubleClick = FGetChar(file);
                    error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                            &dio->do_Contents.field.doubleClickTrigger,
                            &warnings);
                    if (error) {
                        return error;
                    }
                    ProjectAddCodesetWarnings(pr, warnings);
		} else {
		    dio->do_Contents.field.onDoubleClick = 
			DAD_DOUBLECLICK_SHOWRECORD;
		}
                if (version >= 39 || type == TYPE_BUTTON) {
                    dio->do_Contents.field.imagesize = FGetChar(file);
                } else {
                    FGetChar(file);
                }
                if (version >= 19) {
                    dio->do_Contents.field.titlepos = FGetChar(file);
		} else {
                    dio->do_Contents.field.titlepos= DAD_TITLEPOSITION_LEFT;
		}
                error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                        &dio->do_Title, &warnings);
		if (error) {
		    return error;
		}
                ProjectAddCodesetWarnings(pr, warnings);
                if (version >= 44) {
                    error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                            &dio->do_Contents.field.titleMUIFontname,
                            &warnings);
                    if (error) {
                        return error;
                    }
                }
                if (version >= 45) {
                    ProjectAddCodesetWarnings(pr, warnings);
                    error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                            &dio->do_Contents.field.titleGTKFontname,
                            &warnings);
                    if (error) {
                        return error;
                    }
                    ProjectAddCodesetWarnings(pr, warnings);
                    error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                            &dio->do_Contents.field.titleQtFontname,
                            &warnings);
                    if (error) {
                        return error;
                    }
                    ProjectAddCodesetWarnings(pr, warnings);
                }
                error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                        &dio->do_Contents.field.bubbleHelp, &warnings);
                if (error) {
                    return error;
		}
                ProjectAddCodesetWarnings(pr, warnings);
		if (version >= 35) {
		    dio->do_Contents.field.recordsKind = FGetChar(file);
                    error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                            &dio->do_Contents.field.recordsCompute,
                            &warnings);
		    if (error) {
			return error;
		    }
                    ProjectAddCodesetWarnings(pr, warnings);
		    dio->do_Contents.field.initialRecord = FGetBool(file);
		    dio->do_Contents.field.currentRecord = FGetBool(file);
		} else {
		    dio->do_Contents.field.initialRecord = TRUE;
		    dio->do_Contents.field.currentRecord = TRUE;
		}
                if (version >= 42) {
                    rds = readRecordDescriptionSpecCustomFormat(file, utf8, pr,
                            &error);
                    if (error) {
                        return error;
                    }
                } else {
                    short num = 0, i;
                    short rdsType = RDS_TYPE_FIELDS;
                    char *computeDescription = NULL;
                    bool multiColumn;

                    if (version < 40) {
                        num = FGetShort(file);
                    } else if (FGetBool(file)) {
                        rdsType = FGetChar(file);
                        computeDescription = ReadStringConvert(file, utf8, 
                                memPool, &error, &warnings);
                        if (error) {
                            return error;
                        }
                        ProjectAddCodesetWarnings(pr, warnings);
                        num = FGetShort(file);
                    }
                    if (rdsType || computeDescription || num) {
                        rds = AllocRecordDescriptionSpec(memPool, num);
                        if (rds == NULL) {
                            return ERR_MEM;
                        }
                        rds->rds_Type = rdsType;
                        rds->rds_Compute = computeDescription;
                        for (i=0; i < num; i++) {
                            fieldnum = FGetLong(file);
                            rds->rds_Fields[i] = (fieldnum == ~0)?
                                RDS_FIELD_RECNUM : FindFieldNum(pr, fieldnum);
                        }
                    }
                    if (version >= 35) {
                        multiColumn = FGetBool(file);
                        if (rds != NULL) {
                            rds->rds_MultiColumn = multiColumn;
                        }
                    }
                }
                FreeRecordDescriptionSpec(memPool,
                        dio->do_Contents.field.recordDescriptionSpec);
                dio->do_Contents.field.recordDescriptionSpec = rds;

                FreeLabels(memPool, dio->do_Contents.field.listviewLabels);
                dio->do_Contents.field.listviewLabels = 
                    LoadLabels(memPool, file, utf8, &error, &warnings);
                if (error) {
                    return error;
		}
                ProjectAddCodesetWarnings(pr, warnings);
		if (version >= 34) {
                    error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                            &dio->do_Contents.field.listviewCompute,
                            &warnings);
                    if (error) {
                        return error;
                    }
                    ProjectAddCodesetWarnings(pr, warnings);
		}
		break;
            }

	    case DO_TYPE_PANEL_OBSOLETE:
            {
                struct DisplayObject *tabdio = dio->do_Parent;

	        tabdio->do_Contents.table.panel.numall = FGetBool(file);
	        tabdio->do_Contents.table.panel.filter = FGetBool(file);
	        tabdio->do_Contents.table.panel.arrows = FGetBool(file);
	        tabdio->do_Contents.table.hasPanel = FGetBool(file);
		tabdio->do_Title = ReadStringConvert(file, utf8, memPool,
                        &error, &warnings);
		if (error) {
		    return error;
		}
                ProjectAddCodesetWarnings(pr, warnings);
                tabdio->do_MUIFontname = dio->do_MUIFontname;
                dio->do_MUIFontname = NULL;
                tabdio->do_GTKFontname = dio->do_GTKFontname;
                dio->do_GTKFontname = NULL;
                tabdio->do_QtFontname = dio->do_QtFontname;
                dio->do_QtFontname = NULL;
                tabdio->do_Contents.table.panel.background = dio->do_Background;
                dio->do_Background = NULL;
		break;
            }

            case DO_TYPE_SPACE:
                dio->do_Contents.space.bar = FGetBool(file);
                break;

            case DO_TYPE_TEXT:
                error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                        &dio->do_Title, &warnings);
		if (error) {
		    return error;
		}
                ProjectAddCodesetWarnings(pr, warnings);
                if (version >= 18 && version < 20) {
                    FGetShort(file);            /* font */
                }
                break;

            case DO_TYPE_IMAGE:
                dio->do_Contents.image.imagesize = FGetChar(file);
                break;

            case DO_TYPE_BALANCE:
                break;

	    case DO_TYPE_GROUP:
	        dio->do_Contents.group.border = FGetBool(file);
	        dio->do_Contents.group.horiz = FGetBool(file);
	        dio->do_Contents.group.spacing = FGetBool(file);
	        dio->do_Contents.group.open = FGetBool(file);
                error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                        &dio->do_Title, &warnings);
		if (error) {
		    return error;
		}
                ProjectAddCodesetWarnings(pr, warnings);
	    	break;

	    case DO_TYPE_REGISTER:
            {
                char **labels;

	        dio->do_Contents.reggroup.open = FGetBool(file);
                labels = LoadLabels(memPool, file, utf8, &error, &warnings);
                if (error) {
                    return error;
		}
                dio->do_Contents.reggroup.labels = labels;
                ProjectAddCodesetWarnings(pr, warnings);
		break;
            }

            case DO_TYPE_WINDOW:
	    {
		char *buf, sc[2];

                if (version >= 26) {
                    dio->do_Contents.window.button = FGetBool(file);
		} else {
                    dio->do_Contents.window.button = TRUE;
		}
		if (version < 32) {
		    sc[0] = FGetChar(file);
		    sc[1] = 0;
                    buf = CodesetConvertFromDefault8Bit(sc, memPool, &error);
		} else {
		    buf = ReadStringConvert(file, utf8, memPool,
                            &error, &warnings);
                    ProjectAddCodesetWarnings(pr, warnings);
		}
		if (error) {
		    return error;
		}
		dio->do_Shortcut = CodesetGetChar(EmptyNullStr(buf));
		if (buf != sc) {
		    FreeString(memPool, buf);
		}
	        dio->do_Contents.window.open = FGetBool(file);
	        dio->do_Contents.window.windowopen = FGetBool(file);
		dio->do_Enabled = FGetChar(file);
		if (version >= 36) {
                    error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                            &dio->do_EnabledCompute, &warnings);
                    if (error) {
                        return error;
                    }
                    ProjectAddCodesetWarnings(pr, warnings);
                } else {
                    dio->do_Enabled = DOD_ENABLED_ALWAYS;
                }
                if (version >= 37) {
                    dio->do_Contents.window.closeWhenDisabled = FGetBool(file);
                }
                dio->do_Contents.window.wd_valid = FGetBool(file);
                dio->do_Contents.window.wd.wd_LeftEdge = FGetLong(file);
                dio->do_Contents.window.wd.wd_TopEdge = FGetLong(file);
                dio->do_Contents.window.wd.wd_Width = FGetLong(file);
                dio->do_Contents.window.wd.wd_Height = FGetLong(file);
                error = readStringAndUpdateCustomFormat(file, utf8, memPool,
                        &dio->do_Title, &warnings);
		if (error) {
		    return error;
		}
                ProjectAddCodesetWarnings(pr, warnings);
		if (version >= 28) {
                    dio->do_Contents.window.id = FGetLong(file);
		}
                dio->do_Contents.window.activeTableNum = FGetLong(file);
                dio->do_Contents.window.activeDisplayObjectNum = FGetLong(file);
		break;
	    }

	    default:
		return DISKERR_INVALID;
		break;
        }
    }
    removeObsolete(ProjectRootDisplayObject(pr));

    return OK;
}

