/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** display.h
**
** Creation date: 26.11.1994
** Last modified: 10.02.2022
*/

#ifndef DISPLAY_H
#define DISPLAY_H

#include "system.h"
#include "recorddescription.h"
#include "codeset.h"

struct DisplayObject;

#ifdef NEED_NODE_CAST
inline DisplayObject *nodecast(DisplayObject *, Node *node) 
{ return (DisplayObject *)node; }
#endif

/* DisplayObject types */
#define DO_TYPE_TABLE 1
#define DO_TYPE_FIELD 2
#define DO_TYPE_PANEL_OBSOLETE 3
#define DO_TYPE_GROUP 4
#define DO_TYPE_WINDOW 5
#define DO_TYPE_SPACE 6
#define DO_TYPE_TEXT 7
#define DO_TYPE_IMAGE 8
#define DO_TYPE_BALANCE 9
#define DO_TYPE_REGISTER 10

/* String length for title strings, e.g. panel */
#define TITLE_LEN 255

/* String length for font names */
#define FONTNAME_LEN 100

/* Enabled setting */
#define DOD_ENABLED_ALWAYS 0
#define DOD_ENABLED_NOTINITIAL 1
#define DOD_ENABLED_COMPUTED 2

#ifdef __cplusplus
extern "C" {
#endif

/* for getting general fields */
ulong DisplayObjectNum(struct DisplayObject *dio);
struct DisplayObject *FindDisplayObjectNum(struct Project *pr, ulong find);
struct Project *DisplayObjectProject(struct DisplayObject *dio);
short DisplayObjectType(struct DisplayObject *dio);
/* for tree traversal */
struct List *DisplayObjectObjects(struct DisplayObject *dio);
struct DisplayObject *NextDisplayObject(struct DisplayObject *dio);
struct DisplayObject *DisplayObjectParent(struct DisplayObject *dio);
bool DisplayObjectIsAncestor(struct DisplayObject *dio,struct DisplayObject *a);
/* for StructureEditor */
bool DisplayObjectOpen(struct DisplayObject *dio);
void SetDisplayObjectOpen(struct DisplayObject *dio, bool open);
bool DisplayObjectVisible(struct DisplayObject *dio);
void SetDisplayObjectVisible(struct DisplayObject *dio, bool visible);
/* Title settings */
const char *DisplayObjectTitle(struct DisplayObject *dio);
/* Content settings */
short DisplayObjectWeight(struct DisplayObject *dio);
void SetDisplayObjectWeight(struct DisplayObject *dio, short weight);
const char *DisplayObjectMUIFontname(struct DisplayObject *dio);
const char *DisplayObjectGTKFontname(struct DisplayObject *dio);
const char *DisplayObjectQtFontname(struct DisplayObject *dio);
const char *DisplayObjectBackground(struct DisplayObject *dio);
void SetDisplayObjectBackground(struct DisplayObject *dio, const char *bg);
/* other general functions */
struct Table *DisplayObjectTable(struct DisplayObject *dio);
struct DisplayObject *DisplayObjectWindow(struct DisplayObject *dio);
GuiObject DisplayObjectGuiObject(struct DisplayObject *dio);
void SetDisplayObjectGuiObject(struct DisplayObject *dio, GuiObject go);
GuiObject DisplayObjectRootObject(struct DisplayObject *dio);
void SetDisplayObjectRootObject(struct DisplayObject *dio, GuiObject go);
GuiObject DisplayObjectActivateObject(struct DisplayObject *dio);
void SetDisplayObjectActivateObject(struct DisplayObject *dio, GuiObject go);
struct DisplayObject *DisplayActiveObject(GuiObject win);
struct DisplayObject *DisplayHomeObject(struct Table *tab);
cschar DisplayObjectShortcut(struct DisplayObject *dio);
char DisplayObjectEnabled(struct DisplayObject *dio);
const char *DisplayObjectEnabledCompute(struct DisplayObject *dio);
const char *DisplayObjectName(struct DisplayObject *dio);
struct DisplayObject *DisplayObjectShortcutInUse(cschar shortcut,
    struct DisplayObject *windio, struct DisplayObject *leaveout);
void DeleteDisplayObject(struct DisplayObject *dio);
bool IsDisplayObject(struct Project *pr, void *ptr);

/* for root display objects */
struct DisplayObject *AllocRootDisplayObject(struct Project *pr);
void FreeRootDisplayObject(struct DisplayObject *dio);
void ClearRootDisplayObject(struct DisplayObject *dio);

struct Table *DisplayProjectActiveTable(struct Project *pr);
void SetDisplayProjectActiveTableNum(struct Project *pr, long num);
void SetDisplayProjectActiveTable(struct Project *pr, struct Table *tab);
struct DisplayObject *DisplayProjectActiveDisplayObject(struct Project *pr);
void SetDisplayProjectActiveDisplayObjectNum(struct Project *pr, long num);
void SetDisplayProjectActiveDisplayObject(struct Project *pr,
        struct DisplayObject *dio);
void StoreDisplayFields(struct Project *pr);
void SetDisplayFields(struct Project *pr);

/* for tables */

struct DisplayTableData
{
    const char *dtd_Title;
    short dtd_Weight;
    const char *dtd_Fontname;
    const char *dtd_Background;
    bool dtd_HasPanel;
    const char *dtd_PanelBackground;
    bool dtd_PanelNumAll;
    bool dtd_PanelFilter;
    bool dtd_PanelArrows;
    const struct RecordDescriptionSpec *dtd_RecordDescriptionSpec;
};

bool DisplayTableHasPanel(struct DisplayObject *dio);
const char *DisplayTablePanelBackground(struct DisplayObject *dio);
bool DisplayTablePanelNumAll(struct DisplayObject *dio);
bool DisplayTablePanelFilter(struct DisplayObject *dio);
bool DisplayTablePanelArrows(struct DisplayObject *dio);
GuiObject DisplayTablePanelNumAllObject(struct DisplayObject *dio);
void SetDisplayTablePanelNumAllObject(struct DisplayObject *dio, GuiObject go);
GuiObject DisplayTablePanelFilterObject(struct DisplayObject *dio);
void SetDisplayTablePanelFilterObject(struct DisplayObject *dio, GuiObject go);
const struct RecordDescriptionSpec *DisplayTableRecordDescriptionSpec(
        struct DisplayObject *dio);

bool DisplayNewTable(struct Table *tab, struct DisplayObject *dioparent);
long DisplayChangeTable(struct Table *tab, struct DisplayTableData *dtd);

struct Field *DisplayTableActiveField(struct Table *tab);
void SetDisplayTableActiveField(struct Table *tab, struct Field *field);
struct DisplayObject *DisplayTableActiveDisplayObject(struct Table *tab);
void SetDisplayTableActiveDisplayObject(struct Table *tab, struct DisplayObject *dio);

/* record description */

void DisplayDeleteFieldReference(struct Field *ref);

/* for fields */

struct DisplayFieldData
{
    const char *dfd_Title;
    char dfd_TitlePosition;
    const char *dfd_TitleFontname;
    cschar dfd_Shortcut;
    bool dfd_Home;
    bool dfd_TabChain;
    bool dfd_ReadOnly;
    char dfd_Alignment;
    bool dfd_Calendar;
    char dfd_Format;
    char dfd_SubType;
    char dfd_NumDecimals;
    bool dfd_ShowTitles;
    bool dfd_ControlTable;
    char dfd_OnDoubleClick;
    const char *dfd_DoubleClickTrigger;
    const char *dfd_URLDropTrigger;
    const char *dfd_SortDropTrigger;
    bool dfd_AutoUpdate;
    char dfd_Enabled;
    const char *dfd_EnabledCompute;
    short dfd_Weight;
    char dfd_ImageSize;
    const char *dfd_Fontname;
    const char *dfd_Background;
    const char *dfd_BubbleHelp;
    bool dfd_File;
    bool dfd_Font;
    bool dfd_Listview;
    bool dfd_View;
    char **dfd_ListviewLabels;
    const char *dfd_ListviewCompute;
    bool dfd_DisplayExternal;
    bool dfd_TitleAtStringField;
    bool dfd_HideStringField;
    bool dfd_Show;
    bool dfd_AutoShow;
    bool dfd_Filter;
    bool dfd_Frame;
    char dfd_RecordsKind;
    const char *dfd_RecordsCompute;
    bool dfd_InitialRecord;
    bool dfd_CurrentRecord;
    const struct RecordDescriptionSpec *dfd_RecordDescriptionSpec;
};

/* title positions: */

#define DAD_TITLEPOSITION_LEFT 0
#define DAD_TITLEPOSITION_RIGHT 1
#define DAD_TITLEPOSITION_TOP 2
#define DAD_TITLEPOSITION_BOTTOM 3

/* alignments: */
#define DAD_ALIGNMENT_CENTER 0
#define DAD_ALIGNMENT_LEFT 1
#define DAD_ALIGNMENT_RIGHT 2
#define DAD_ALIGNMENT_VERTICAL 0
#define DAD_ALIGNMENT_HORIZONTAL 1

/* sub types: */
#define DAD_VIRTUALTYPE_BOOL 0
#define DAD_VIRTUALTYPE_TEXT 1
#define DAD_VIRTUALTYPE_LIST 2

#define DAD_BUTTONTYPE_TEXT 0
#define DAD_BUTTONTYPE_IMAGE 1

#define DAD_CHOICETYPE_CYCLEBUTTON 0
#define DAD_CHOICETYPE_RADIOBUTTONS 1

/* records in reference pop-ups */
#define DAD_RECORDS_ALL 0
#define DAD_RECORDS_MATCHFILTER 1
#define DAD_RECORDS_COMPUTED 2

/* double click actions for virtual lists */
#define DAD_DOUBLECLICK_DONOTHING 0
#define DAD_DOUBLECLICK_SHOWRECORD 1
#define DAD_DOUBLECLICK_CALLTRIGGER 2

/* image size (also for externalsize) */
#define DAD_IMAGESIZE_RESIZABLE 0
#define DAD_IMAGESIZE_FIXED 1
#define DAD_IMAGESIZE_SCROLLABLE 2
#define DAD_IMAGESIZE_SCALED 3
#define DAD_IMAGESIZE_ASPECTSCALED 4

struct Field *DisplayFieldField(struct DisplayObject *dio);

char DisplayFieldTitlePosition(struct DisplayObject *dio);
const char *DisplayFieldTitleMUIFontname(struct DisplayObject *dio);
const char *DisplayFieldTitleGTKFontname(struct DisplayObject *dio);
const char *DisplayFieldTitleQtFontname(struct DisplayObject *dio);
bool DisplayFieldHome(struct DisplayObject *dio);
bool DisplayFieldTabChain(struct DisplayObject *dio);
bool DisplayFieldReadOnly(struct DisplayObject *dio);
char DisplayFieldSubType(struct DisplayObject *dio);
char DisplayFieldAlignment(struct DisplayObject *dio);
bool DisplayFieldCalendar(struct DisplayObject *dio);
char DisplayFieldFormat(struct DisplayObject *dio);
char DisplayFieldNumDecimals(struct DisplayObject *dio);
bool DisplayFieldShowTitles(struct DisplayObject *dio);
bool DisplayFieldControlTable(struct DisplayObject *dio);
char DisplayFieldOnDoubleClick(struct DisplayObject *dio);
const char *DisplayFieldDoubleClickTrigger(struct DisplayObject *dio);
const char *DisplayFieldURLDropTrigger(struct DisplayObject *dio);
const char *DisplayFieldSortDropTrigger(struct DisplayObject *dio);
bool DisplayFieldAutoUpdate(struct DisplayObject *dio);
char DisplayFieldImageSize(struct DisplayObject *dio);
const char *DisplayFieldBubbleHelp(struct DisplayObject *dio);
char DisplayFieldRecordsKind(struct DisplayObject *dio);
const char *DisplayFieldRecordsCompute(struct DisplayObject *dio);
bool DisplayFieldInitialRecord(struct DisplayObject *dio);
bool DisplayFieldCurrentRecord(struct DisplayObject *dio);
const struct RecordDescriptionSpec *DisplayFieldRecordDescriptionSpec(
        struct DisplayObject *dio, bool followInherited);
bool DisplayFieldFile(struct DisplayObject *dio);
bool DisplayFieldFont(struct DisplayObject *dio);
bool DisplayFieldListview(struct DisplayObject *dio);
bool DisplayFieldView(struct DisplayObject *dio);
char **DisplayFieldListviewLabels(struct DisplayObject *dio);
long SetDisplayFieldListviewLabels(struct DisplayObject *dio,char **labels);
const char *DisplayFieldListviewCompute(struct DisplayObject *dio);
bool DisplayFieldDisplayExternal(struct DisplayObject *dio);
bool DisplayFieldTitleAtStringField(struct DisplayObject *dio);
bool DisplayFieldHideStringField(struct DisplayObject *dio);
bool DisplayFieldShow(struct DisplayObject *dio);
bool DisplayFieldAutoShow(struct DisplayObject *dio);
bool DisplayFieldFilter(struct DisplayObject *dio);
bool DisplayFieldFrame(struct DisplayObject *dio);

GuiObject DisplayFieldListviewObject(struct DisplayObject *dio);
GuiObject DisplayFieldFileObject(struct DisplayObject *dio);
GuiObject DisplayFieldFontObject(struct DisplayObject *dio);
GuiObject DisplayFieldViewObject(struct DisplayObject *dio);
GuiObject DisplayFieldExternalObject(struct DisplayObject *dio);
GuiObject DisplayFieldShowObject(struct DisplayObject *dio);
GuiObject DisplayFieldFilterObject(struct DisplayObject *dio);
GuiObject DisplayFieldAutoShowObject(struct DisplayObject *dio);
GuiObject DisplayFieldPopObject(struct DisplayObject *dio);
GuiObject DisplayFieldDisableObject(struct DisplayObject *dio);
void SetDisplayFieldListviewObject(struct DisplayObject *dio, GuiObject go);
void SetDisplayFieldFileObject(struct DisplayObject *dio, GuiObject go);
void SetDisplayFieldFontObject(struct DisplayObject *dio, GuiObject go);
void SetDisplayFieldViewObject(struct DisplayObject *dio, GuiObject go);
void SetDisplayFieldExternalObject(struct DisplayObject *dio, GuiObject go);
void SetDisplayFieldShowObject(struct DisplayObject *dio, GuiObject go);
void SetDisplayFieldFilterObject(struct DisplayObject *dio, GuiObject go);
void SetDisplayFieldAutoShowObject(struct DisplayObject *dio, GuiObject go);
void SetDisplayFieldPopObject(struct DisplayObject *dio, GuiObject go);
void SetDisplayFieldDisableObject(struct DisplayObject *dio, GuiObject go);

bool DisplayFieldListActiveChanged(struct DisplayObject *dio);
void SetDisplayFieldListActiveChanged(struct DisplayObject *dio,
        bool changed);

bool DisplayNewField(struct Field *field, struct DisplayObject *dioparent);
long DisplayChangeField(struct Field *field, struct DisplayFieldData *dfd);

/* adding & removing tables & fields */
void DisplayObjectAdd(struct DisplayObject *dio, 
        struct DisplayObject *dioparent, bool isPred);
bool DisplayObjectRemove(struct DisplayObject *dio); /*returns true on success*/

/* for space objects */
struct SpaceData
{
    bool sd_Bar;
    short sd_Weight;
    const char *sd_Background;
};
bool DisplaySpaceBar(struct DisplayObject *dio);
long DisplayNewSpace(struct Project *pr, struct Table *tab,
        struct DisplayObject *dioparent, struct DisplayObject **newone);
void DisplayChangeSpace(struct DisplayObject *space, struct SpaceData *sd);

/* for text objects */
struct TextData
{
    const char *td_Title;
    short td_Weight;
    const char *td_Fontname;
    const char *td_Background;
};

long DisplayNewText(struct Project *pr, struct Table *tab,
        struct DisplayObject *dioparent, struct DisplayObject **newone);
void DisplayChangeText(struct DisplayObject *text, struct TextData *td);

/* for image objects */
struct ImageData
{
    short id_Weight;
    char id_ImageSize;
    const char *id_Image;
};
char DisplayImageImageSize(struct DisplayObject *dio);
long DisplayNewImage(struct Project *pr, struct Table *tab,
        struct DisplayObject *dioparent, struct DisplayObject **newone);
void DisplayChangeImage(struct DisplayObject *image, struct ImageData *id);

/* for balance objects */
long DisplayNewBalance(struct Project *pr, struct Table *tab,
        struct DisplayObject *dioparent, struct DisplayObject **newone);

/* for register objects */
struct RegisterData
{
    short rd_Weight;
    const char *rd_Fontname;
    const char *rd_Background;
    char **rd_Labels;
};

char **DisplayRegisterLabels(struct DisplayObject *dio);
long DisplayNewRegister(struct Project *pr, struct Table *tab,
        struct DisplayObject **selected, struct DisplayObject *dioparent,
        struct DisplayObject **newone);
long DisplayChangeRegister(struct DisplayObject *reg, struct RegisterData *rd);


/*
** for adding a group, if display contains selected objects, they are moved into the new group.
** The selected objects must be on the same level.
** If these conditions are not met an error code is returned.
** If display has no active objects, the group is added as child to the dioparent object
** at last position. If dioparent object is NULL, the group will be added as child
** to the table disolay object at last position.
*/

struct GroupData
{
    const char *gd_Title;
    short gd_Weight;
    const char *gd_Fontname;
    const char *gd_Background;
    bool gd_Horiz;
    bool gd_Border;
    bool gd_Spacing;
};

bool DisplayGroupBorder(struct DisplayObject *dio);
bool DisplayGroupHoriz(struct DisplayObject *dio);
bool DisplayGroupSpacing(struct DisplayObject *dio);
long DisplayNewGroup(struct Project *pr, struct Table *tab,
        struct DisplayObject **selected, struct DisplayObject *dioparent,
        struct DisplayObject **newone);
void DisplayChangeGroup(struct DisplayObject *group, struct GroupData *gd);

/*
** for windows
** when adding a window, same rules apply as in adding a group
*/
struct WindowData
{
    const char *wd_Title;
    ulong wd_Id;
    bool wd_Button;
    char wd_Shortcut;
    char wd_Enabled;
    const char *wd_EnabledCompute;
    bool wd_CloseWhenDisabled;
    short wd_Weight;
    const char *wd_Fontname;
    const char *wd_Background;
};

struct WindowDimensions
{
    long wd_LeftEdge;
    long wd_TopEdge;
    long wd_Width;
    long wd_Height;
};

struct DisplayObject *DisplayWindowIdInUse(ulong id, 
	struct DisplayObject *leaveout);
ulong DisplayWindowId(struct DisplayObject *dio);
void SetDisplayWindowId(struct DisplayObject *dio, ulong id);
const struct WindowDimensions *DisplayWindowDimensions(struct DisplayObject *);
void SetDisplayWindowDimensions(struct DisplayObject *dio, 
        struct WindowDimensions *wd);
bool DisplayWindowOpen(struct DisplayObject *dio);
void SetDisplayWindowOpen(struct DisplayObject *dio, bool open);
bool DisplayWindowButton(struct DisplayObject *dio);
bool DisplayWindowCloseWhenDisabled(struct DisplayObject *dio);
GuiObject DisplayWindowButtonObject(struct DisplayObject *dio);
void SetDisplayWindowButtonObject(struct DisplayObject *dio, GuiObject button);
long DisplayNewWindow(struct Project *pr, struct Table *tab,
        struct DisplayObject **selected, struct DisplayObject *dioparent,
        struct DisplayObject **newone);
void DisplayChangeWindow(struct DisplayObject *wi, struct WindowData *wd);

/*
** other functions
*/
bool DisplayObjectUp(struct DisplayObject *dio);
bool DisplayObjectDown(struct DisplayObject *dio);
bool DisplayObjectIn(struct DisplayObject *dio);
bool DisplayObjectOut(struct DisplayObject *dio);

bool DisplayObjectVisibleGui(struct DisplayObject *dio);
  /* returns TRUE if object should be visible in gui */

/* Load & Save */

struct ReadDisplayObjectData
{
    long rdod_ID;
    struct DisplayObject *rdod_Parent;
    bool rdod_Visible;
    const char *rdod_Title;
    cschar rdod_Shortcut;
    char rdod_Enabled;
    const char *rdod_EnabledCompute;
    const char *rdod_QtFontname;
    const char *rdod_GTKFontname;
    const char *rdod_MUIFontname;
    const char *rdod_Background;
    short rdod_Weight;
    short rdod_Type;
    union {
        struct {
            struct Table *table;
            struct Field *activeField;
            long activeDisplayObjectID;
            bool hasPanel;
            const char *panelBackground;
            bool panelNumAll;
            bool panelFilter;
            bool panelArrows;
            const struct RecordDescriptionSpec *recordDescriptionSpec;
        } table;
        struct {
            struct Field *field;
            char titlePosition;
            const char *titleQtFontname;
            const char *titleGTKFontname;
            const char *titleMUIFontname;
            bool home;
            bool tabChain;
            bool readOnly;
            char subType;
            char alignment;
            bool calendar;
            char format;
            char numDecimals;
            char imageSize;
            const char *bubbleHelp;
            bool displayExternal;
            bool titleAtStringField;
            bool hideStringField;
            bool file;
            bool font;
            bool listview;
            const char *listviewLabels;
            const char *listviewCompute;
            bool view;
            bool frame;
            char recordsKind;
            const char *recordsCompute;
            bool initialRecord;
            bool currentRecord;
            bool show;
            bool autoShow;
            bool filter;
            bool autoUpdate;
            bool showTitles;
            bool controlTable;
            char onDoubleClick;
            const char *doubleClickTrigger;
            const char *urlDropTrigger;
            const char *sortDropTrigger;
            const struct RecordDescriptionSpec *recordDescriptionSpec;
        } field;
        struct {
            bool bar;
        } space;
        struct {
            char imageSize;
        } image;
        struct {
            const char *labels;
            bool open;
        } reg;
        struct {
            bool border;
            bool horizontal;
            bool spacing;
            bool open;
        } group;
        struct {
            ulong windowId;
            bool button;
            bool closeWhenDisabled;
            bool windowOpen;
            struct WindowDimensions dimensions;
            bool open;
        } window;
    } rdod_Contents;
};

long NumDisplayObjects(struct DisplayObject *dio);

long ReadDisplayObject(struct Project *pr,
        const struct ReadDisplayObjectData *data);

long LoadDisplayObjectsCustomFormat(struct Project *pr, FILE *file);

#ifdef __cplusplus
}
#endif

#endif
