/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** dosutil.c
**
** help routines for dos/general os
**
** Creation date: 23.12.2004
** Last modified: 22.12.2018
*/

#include "system.h"
#include "dosutil.h"
#include "strhandle.h"
#ifndef AMIGA
#include "unistd.h"		/* dup */
#endif
#ifdef WINDOWS
#undef _WIN32_WINNT
#define _WIN32_WINNT 0x0500
#include "windows.h"
#endif

char *PathDirname(const char *path, struct MemPool *memPool)
{
    short stat = FileStat(path);
    char *dirname = NULL;

    if (stat != 0) {  /* not a directory */
        SplitPathname(path, memPool, &dirname, NULL);  /* parent directory */
    } else {
        dirname = AllocString(memPool, path);
    }

    return dirname;
}

char *MakeTemp(char *pattern)
{
    if (pattern) {
        char *t = strrchr(pattern, 'X'), *s;
	int count = 1;
	static const int maxCount = 6;
	long end = 10, i;
        char buf[20];

        if (t == NULL) {
            pattern[0] = 0;
            return NULL;
        }

        for (s = t-1; s >= pattern; s--) {
	    if (*s != 'X' || count >= maxCount) {
		break;
	    }
            count++;
            end *= 10;
            t = s;
	}
        while (s >= pattern && *s == 'X') {
            *s-- = '0';
        }

	for (i=0; i < end; i++) {
	    sprintf(buf, "%0*ld", count, i);
            memcpy(t, buf, count);
	    if (FileStat(pattern) < 0) {
		return pattern;
	    }
	}
	*pattern = 0;
    }
    return NULL;
}

FILE *IFOpen(struct MemPool *memPool, const char *filename, const char *mode)
{
    FILE *file = NULL;
    char *expand;

    if (filename == NULL || mode == NULL) {
        return NULL;
    }

    expand = ExpandPathname(filename, memPool);
    if (expand) {
	filename = expand;
    }

#ifndef AMIGA
    // check for unknown modes
    if (mode[0] != 'r' && mode[0] != 'w' && mode[0] != 'a') {
        file = FOpen(filename, mode);
    } else if (!strcmp(filename, "-")) {
       	int fd = dup((mode[0] == 'r')? 0 : 1);

	if (fd >= 0) {
	    file = fdopen(fd, mode);
	} else {
	    file = FOpen(filename, mode);
	}
    } else if (filename[0] == '|') {
        file = popen(filename + 1, mode);
#ifdef WINDOWS
    } else if (strcasecmp(filename, "CON") == 0 ||
            strncasecmp(filename, "CON.", 4) == 0) {
        if (AllocConsole()) {
            SetConsoleCtrlHandler(NULL, TRUE);
            HWND hwnd = GetConsoleWindow();
            HMENU hmenu = GetSystemMenu(hwnd, FALSE);
            DeleteMenu(hmenu, SC_CLOSE, MF_BYCOMMAND);
        }
        file = FOpen(filename, mode);
#endif
    } else {
       	const char *s = CodesetStrrchr(filename, '.');

	if (s != NULL && strcmp(s, ".gz") == 0) {
	    long size = strlen(filename) + 100;
	    char *buffer = PoolAllocMem(memPool, size);

	    if (buffer) {
		sprintf(buffer, "gzip %s \"%s\"", 
			(mode[0] == 'r')? "-cd" : "-c >", filename);
		file = popen(buffer, mode);
		PoolFreeMem(memPool, buffer, size);
	    }
	}
	if (file == NULL) {
	    file = FOpen(filename, mode);
	}
    }
#else
    file = FOpen(filename, mode);
#endif
    
    FreeString(memPool, expand);

    return file;
}

