/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** dosutil.h
**
** help routines for dos/general os
**
** Creation date: 03.04.1995
** Last modified: 02.12.2020
*/

#ifndef DOS_UTIL_H
#define DOS_UTIL_H

#include "system.h"
#include "codeset.h"
#include "time.h"

#ifdef __cplusplus
extern "C" {
#endif

#define TIME_T_INVALID ((time_t)-1)

struct Lock;

bool SameFile(const char *filename1, const char *filename2);
/* Checks whether two filenames refer to the same physical file on disk. */

bool FileLocked(const char *filename);
/* returns true if there is any (shared or exclusive) lock on filename */


struct Lock *FileLockExclusive(const char *filename, bool *isLocked);
/*
** Obtains an exclusive lock on the given filename.
** On failure returns NULL and sets isLocked to true if the file has been
** locked by someone else
*/

struct Lock *FileLockShared(const char *filename);
/* obtains a shared lock on the given filename */

void FileUnlock(struct Lock *lock);
/* frees a lock */

short FileStat(const char *filename);
/*
** Examines a filename.
** If it's a file a value greater 0 is returned.
** If it's a directory, 0 is returned.
** If the filename doesn't exist or an error has been occured,
** a negative value is returned.
*/

time_t FileModificationTime(const char *filename);
/*
** Returns the last modification time of the given filename.
** If the file can't be accessed, returns TIME_T_INVALID.
*/

#ifdef AMIGA
void UpdateFileModificationTime(const char *filename);
/*
** Sets the file modification time to the current time.
*/
#endif

bool FileIsWritable(const char *filename);
/*
** Returns TRUE if the given filename is writable.
** Returns FALSE if the filename doesn't exist.
*/

#ifdef UNIX
const char *DefaultOutputFilename(void);
/* Returns the default output filename. */
#endif

bool IsPathSeparator(int c);
/*
** Returns true if c is a path separator, e.g. '/' on Unix.
*/

bool SplitPathname(const char *path, struct MemPool *memPool, 
        char **pdir, char **pfile);
/*
** Splits Pathname into Dirname and Filename.  If path is a directory,
** then *pdir will hold parent directory and *pfile the name of the
** subdirectory.
*/

char *PathDirname(const char *path, struct MemPool *memPool);
/*
** Obtains the directory of a path.  If the path itself is a directory, returns
** the path, otherwise returns the parent directory.
*/

char *JoinPathname(struct MemPool *memPool, const char *dir, const char *file);
/*
** Joins Dirname and Filename to a Pathname.
*/

char *MakeTemp(char *pattern);
/*
** finds a filename that is not used in the current directory.
** The pattern must end with sereval 'X' and these are replaced
** with numbers.  On sucess, returns the unused filename (stored in pattern),
** on failure pattern will be cleared and NULL is returned.
*/

const char *GetTmpDir(void);
/*
** Returns a system-dependent path to a temporary directory.
*/

FILE *IFOpen(struct MemPool *memPool, const char *filename, const char *mode);
/*
** Improved fopen by expanding path, checking for pipe "|" and optional
** gzip compression. 
*/

/* interfaces for file operations */ 
#ifdef __amigaos4__
#undef FOpen			/* remove AOS4's dos.library symbol */
#endif
FILE *FOpen(const char *filename, const char *mode);

int MkDir(const char *dirname);
int RmDir(const char *dirname);
int RmFile(const char *filename);
int RenamePath(const char *oldpath, const char *newpath);

/* 
** Functions for easing changing current directory
** Cwd should always be initialized with NULL.
** There can be several Push commands with the same cwd argument,
** but each of them overwrites the previous one (not a stack).
** PopDir is required but only one is necessary.
*/

struct Cwd_;
typedef struct Cwd_ *Cwd;
int PushDir(const char *dirname, Cwd *cwd);
int PopDir(Cwd *cwd);

/* functions for abstracting opendir, g_dir_open, ... */
struct Dir_;
typedef struct Dir_ Dir;

Dir *DirOpen(const char *dirname);
const char *DirReadName(Dir *dir);
void DirClose(Dir *dir);

struct DosProcess;      /* abstract data type */

struct DosProcess *SpawnDosProcess(const char *cmdline, bool useShell, 
        bool needRC);
/*
** Spawns a new process using the given command line.  If useShell is TRUE
** then a command interpreter (depending on the system, e.g. /bin/sh)
** is used for interpreting the command line, otherwise the command
** line is split using standard rules for quoting, etc.
** If needRC is FALSE then no effort is done to obtain 
** the exit code of the command.
** Returns pointer to new DosProcess or NULL on failure 
** (command not found or failure to launch shell)
*/

bool DosProcessIsExited(struct DosProcess *process);
/*
** Checks wether a process has exited.
*/

int DosProcessExitCode(struct DosProcess *process);
/*
** Returns exit code of DosProcess
*/

void CloseDosProcess(struct DosProcess *process);
/*
** Frees recources 
*/

long DosProcessesRunning(void);
/*
** Amiga only
*/

#ifdef __cplusplus
}
#endif

#endif

