/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** field.c
**
** Creation date: 15.11.1994
** Last modified: 10.03.2021
*/

#include "field.h"
#include "display.h"
#include "record.h"
#include "filter.h"
#include "queryeditor.h"
#include "sqlitedb.h"
#include "memo.h"
#include "labels.h"
#include "strhandle.h"
#include "memory.h"
#include "name.h"
#include "break.h"
#include "codeset.h"
#include "dosutil.h"
#include "util.h"
#include "dbtypes.h"
#include "error.h"

#define DEFAULT_NILVALUE_STRING ""

struct Field
{
    struct Node fld_Node;
    struct Table *fld_Table;
    char fld_Name[ID_NAME_LEN + 1];			/* ASCII */
    short fld_Type;
    struct DisplayObject *fld_DisplayObject;
    struct FieldRecordInfo *fld_FieldRecordInfo;
    char *fld_NilString;				/* Codeset */
    char fld_Trigger[FIELD_TRIGGER_LEN + 1];		/* ASCII */
    bool fld_CountChanges;
    bool fld_LogChanges;
    struct DataInfo fld_DataInfo;
    union
    {
	struct
	{
            short maxlen;
	} string;
	struct
	{
	    char **labels;
	} choice;
        struct
        {
            struct Table *table;
            bool filter;
	    bool autoshow;
        } reference;
        struct
        {
            char compute[FIELD_TRIGGER_LEN + 1];		/* ASCII */
        } virtual;
    } fld_Contents;
};

short FieldType(struct Field *field)
{
    short type = NOTYPE;

    if(field)
	type = field->fld_Type;
    return(type);
}

const char *FieldName(struct Field *field)
{
    const char *name = NULL;

    if (field) {
	name = field->fld_Name;
    }
    return name;
}

struct Table *FieldTable(struct Field *field)
{
    struct Table *tab = NULL;

    if(field)
	tab = field->fld_Table;
    return(tab);
}

static void setFieldName(struct Field *field, const char *name)
{
    if (field && name) {
	CodesetStrNBytesCopy(field->fld_Name, name, ID_NAME_LEN);
    }
}

struct Field *FindFieldName(struct Table *parent, const char *name)
{
    struct Field *field;

    if(parent && name)
    	FORNODE(field, TableFields(parent))
	    if(strcmp(name, FieldName(field)) == 0)
		return(field);
    return(NULL);
}

struct DisplayObject *FieldDisplayObject(struct Field *field)
{
    struct DisplayObject *dio = NULL;

    if(field)
	dio = field->fld_DisplayObject;
    return(dio);
}

void SetFieldDisplayObject(struct Field *field, struct DisplayObject *dio)
{
    if(field)
        field->fld_DisplayObject = dio;
}

short FieldStringMaxlen(struct Field *field)
{
    short maxlen = 0;

    if (field && field->fld_Type == TYPE_STRING) {
        maxlen = field->fld_Contents.string.maxlen;
    }
    return maxlen;
}

struct FieldRecordInfo *FieldRecordInfo(struct Field *field)
{
    struct FieldRecordInfo *ari = NULL;

    if(field)
	ari = field->fld_FieldRecordInfo;
    return(ari);
}

struct DataInfo *FieldDataInfo(struct Field *field)
{
    struct DataInfo *di = NULL;

    if(field)
	di = &field->fld_DataInfo;
    return(di);
}

const char *FieldNilString(struct Field *field)
{
    const char *nil = NULL;

    if (field) {
        switch (field->fld_Type) {
            case TYPE_INTEGER:
            case TYPE_REAL:
            case TYPE_DATE:
            case TYPE_TIME:
            case TYPE_VIRTUAL:
		nil = EmptyNullStr(field->fld_NilString);
		break;
        }
    }
    return nil;
}

const char *FieldTrigger(struct Field *field)
{
    const char *trigger = NULL;

    if (field) {
        trigger = field->fld_Trigger;
    }
    return trigger;
}

bool FieldCountChanges(struct Field *field)
{
    bool count = FALSE;

    if (field) {
	count = field->fld_CountChanges;
    }
    return count;
}

bool FieldLogChanges(struct Field *field)
{
    bool log = FALSE;

    if (field) {
	log = field->fld_LogChanges;
    }
    return log;
}

ushort FieldChoiceNumLabels(struct Field *field)
{
    ushort num = 0;

    if(field && field->fld_Type == TYPE_CHOICE)
	num = NumLabels(field->fld_Contents.choice.labels);
    return(num);
}

char **FieldChoiceLabels(struct Field *field)
{
    char **labels = NULL;

    if(field && field->fld_Type == TYPE_CHOICE)
	labels = field->fld_Contents.choice.labels;
    return(labels);
}

struct Table *FieldReferenceTable(struct Field *field)
{
    struct Table *tab = NULL;

    if(field && field->fld_Type == TYPE_REFERENCE)
        tab = field->fld_Contents.reference.table;
    return(tab);
}

struct Field *FindReferenceTable(struct Table *ref)
{
    struct Field *field;
    struct Table *tab;
    struct Project *pr = TableProject(ref);
    struct List *tabs = ProjectTables(pr);
    struct List *fields;

    if(tabs)
        FORNODE(tab, tabs)
        {
            fields = TableFields(tab);
            if(fields)
                FORNODE(field, fields)
                    if(FieldType(field) == TYPE_REFERENCE &&
                        FieldReferenceTable(field) == ref)
                            return(field);
        }

    return(NULL);
}

bool FieldReferenceFilter(struct Field *field)
{
    bool filter = FALSE;

    if(field && field->fld_Type == TYPE_REFERENCE)
        filter = field->fld_Contents.reference.filter;
    return(filter);
}

void SetFieldReferenceFilter(struct Field *field, bool filter)
{
    if(field && field->fld_Type == TYPE_REFERENCE)
        field->fld_Contents.reference.filter = filter;
}

bool FieldReferenceAutoShow(struct Field *field)
{
    bool autoshow = FALSE;

    if(field && field->fld_Type == TYPE_REFERENCE)
        autoshow = field->fld_Contents.reference.autoshow;
    return(autoshow);
}

void SetFieldReferenceAutoShow(struct Field *field, bool autoshow)
{
    if(field && field->fld_Type == TYPE_REFERENCE)
        field->fld_Contents.reference.autoshow = autoshow;
}

const char *FieldVirtualCompute(struct Field *field)
{
    const char *compute = NULL;

    if(field && field->fld_Type == TYPE_VIRTUAL)
        compute = field->fld_Contents.virtual.compute;
    return(compute);
}

static void ClearField(struct Field *field)
{
    if (field) {
	struct Table *tab = FieldTable(field);
	struct MemPool *memPool = TableMemPool(tab);

	FreeFieldRecordInfo(field->fld_FieldRecordInfo);
	field->fld_FieldRecordInfo = NULL;

        switch (field->fld_Type) {
	    case TYPE_STRING:
		FreeString(memPool, field->fld_DataInfo.di_Contents.string);
		field->fld_DataInfo.di_Contents.string = NULL;
		break;

            case TYPE_MEMO:
                FreeString(memPool, field->fld_DataInfo.di_Contents.mi);
                field->fld_DataInfo.di_Contents.mi = NULL;
                break;

            case TYPE_CHOICE:
	        FreeLabels(memPool, field->fld_Contents.choice.labels);
	        field->fld_Contents.choice.labels = NULL;
                break;
	}
	FreeString(memPool, field->fld_NilString);
	field->fld_NilString = NULL;
    }
}

void FreeField(struct Field *field)
{
    if (field) {
	struct Table *tab = FieldTable(field);
	struct MemPool *memPool = TableMemPool(tab);

        ClearField(field);
	PoolFreeMem(memPool, field, sizeof(struct Field));
    }
}

static struct Field *AllocField(struct Table *tab)
{
    struct MemPool *memPool = TableMemPool(tab);
    struct Field *field;

    if (tab == NULL) {
	return NULL;
    }

    field = PoolAllocMemZero(memPool, sizeof(struct Field));
    if (field == NULL) {
	return NULL;
    }

    field->fld_Table = tab;
    field->fld_Type = NOTYPE;
    field->fld_CountChanges = TRUE;
    field->fld_LogChanges = TRUE;

    return field;
}

static long InitFieldType(struct Field *field,
	const char *init, union FieldTypeData *atd)
{
    struct MemPool *memPool = TableMemPool(field->fld_Table);
    struct DataInfo *di = &field->fld_DataInfo;
    bool nil = FALSE;
    long error = 0;

    switch (field->fld_Type) {
        case TYPE_STRING:
        {
            short maxlen = atd->string.maxlen, initlen;
	    char *copy;

            init = EmptyNullStr(init);

	    if (maxlen <= 0 || maxlen > MAX_STRING_LEN) {
	    	error = FIELDERR_INVALIDMAXLEN;
		break;
	    }
	    initlen = CodesetStrLen(init);
            if (initlen > FIELD_INITIALVALUE_LEN || initlen > maxlen) {
		error = FIELDERR_INITIALVALUELEN;
		break;
            }
	    copy = AllocString(memPool, init);
	    if (copy == NULL) {
		error = ERR_MEM;
		break;
	    }

            if (field->fld_FieldRecordInfo) {
                error = ChangeStringRecordInfo(
		    field->fld_FieldRecordInfo, maxlen);
                if (error) {
		    FreeString(memPool, copy);
		    break;
		}
            }

	    FreeString(memPool, di->di_Contents.string);
            di->di_Contents.string = copy;
            field->fld_Contents.string.maxlen = maxlen;
	    break;
	}

	case TYPE_INTEGER:
	{
	    long initvalue = 0;

            if (init == NULL) {
                nil = TRUE;
	    }
            else if (!StrToInt(init, &initvalue)) {
		error = FIELDERR_INVALIDINITIALVALUE;
		break;
	    }
	    di->di_Contents.integer = initvalue;
	    break;
	}

	case TYPE_REAL:
	{
	    double initvalue = 0.0;

            if (init == NULL) {
                nil = TRUE;
	    } else if(!StrToDouble(init, &initvalue)) {
		error = FIELDERR_INVALIDINITIALVALUE;
		break;
	    }
	    di->di_Contents.real = initvalue;
	    break;
	}

	case TYPE_BOOL:
            nil = atd->boolean.nil? TRUE : FALSE;
	    break;

	case TYPE_CHOICE:
	{
	    ushort num, initvalue = atd->choice.initialvalue;
	    char **labels;

	    num = NumLabels(atd->choice.labels);
	    if (num < 1) {
		error = FIELDERR_MISSINGLABELS;
		break;
            }
	    if (initvalue >= num) {
		error = FIELDERR_INVALIDINITIALVALUE;
		break;
	    }
            labels = CopyLabels(memPool, atd->choice.labels);
	    if (labels == NULL) {
            	error = ERR_MEM;
		break;
	    }
            di->di_Contents.choice = initvalue;
	    FreeLabels(memPool, field->fld_Contents.choice.labels);
            field->fld_Contents.choice.labels = labels;
	    break;
   	}
  	case TYPE_DATE:
        {
            long initvalue = 0;

            if (init == NULL) {
                nil = TRUE;
	    } else if (!StringToDate(init, &initvalue)) {
		error = FIELDERR_INVALIDINITIALVALUE;
		break;
	    }
	    di->di_Contents.date = initvalue;
	    break;
	}
  	case TYPE_TIME:
        {
	    short format = FORMAT_HOURMINSEC;
            long initvalue = 0;

            if (init == NULL) {
                nil = TRUE;
	    } else if (!StringToTime(init, format, &initvalue)) {
		error = FIELDERR_INVALIDINITIALVALUE;
		break;
	    }
	    di->di_Contents.time = initvalue;
	    break;
	}

	case TYPE_MEMO:
	case TYPE_BUTTON:
	    break;

        case TYPE_REFERENCE:
            if (atd->reference.table == NULL) {
                error = FIELDERR_NOREFERENCE;
                break;
            }
            if (field->fld_Contents.reference.table != NULL &&
		    field->fld_Contents.reference.table!=atd->reference.table) {
                error = FIELDERR_INVALIDREFERENCE;
                break;
            }
	    field->fld_Contents.reference.table = atd->reference.table;
	    field->fld_Contents.reference.filter = atd->reference.filter;
	    field->fld_Contents.reference.autoshow = atd->reference.autoshow;
	    nil = TRUE;
	    di->di_Contents.rec = NULL;
            break;

        case TYPE_VIRTUAL:
	{
	    struct Project *pr = TableProject(field->fld_Table);
	    struct Program *prg = ProjectProgram(pr);
	    const char *compute = field->fld_Contents.virtual.compute;
	    struct DisplayObject *dio = field->fld_DisplayObject;
	    struct UserFunction *before = FindUserFunctionName(prg, compute);
	    struct UserFunction *after;

	    CodesetStrNBytesCopy(field->fld_Contents.virtual.compute,
		    atd->virt.compute, FIELD_TRIGGER_LEN);
            CodesetTrimStr(field->fld_Contents.virtual.compute);
	    after = FindUserFunctionName(prg, compute);

	    if (after != before && DisplayObjectVisible(dio) && 
		    DisplayFieldAutoUpdate(dio)) {
		SetProgramNeedRecompile(prg, TRUE);
	    }
            break;
	}

	default:
	    error = FIELDERR_UNKNOWNTYPE;
	    break;
    }
    if (nil) {
        di->di_Flags |= DIF_NIL;
    } else {
        di->di_Flags &= ~DIF_NIL;
    }

    return error;
}

long CanNewField(struct Table *tab)
{
    return CanNewFieldRecordInfo(tab, TYPE_UNKNOWN);
}

long NewField(struct NewFieldData *nfd, struct Field **newfield)
{
    struct Field *field = NULL;
    struct MemPool *memPool = NULL;
    struct DataInfo *di = NULL;
    long error = 0;

    if (newfield) {
	*newfield = NULL;
    }

    if (nfd == NULL) {             	/* should never happen! */
	return ERR_INTERNAL;
    }
    if (nfd->nfd_Table == NULL) {
	return ERR_NOTABLE;
    }
    if (!IsMBName(nfd->nfd_Name) || 
            IsSystemName(nfd->nfd_Name) != TableIsSystem(nfd->nfd_Table)) {
        return ERR_INVALIDNAME;
    }
    if (FieldNameInUse(nfd->nfd_Table, nfd->nfd_Name, NULL)) {
	return ERR_NAMEALREADYUSED;
    }
    
    memPool = TableMemPool(nfd->nfd_Table);

    field = AllocField(nfd->nfd_Table);
    if (field == NULL) {
	return ERR_MEM;
    }

    field->fld_Type = nfd->nfd_Type;
    setFieldName(field, nfd->nfd_Name);
    if (nfd->nfd_NilString) {
	field->fld_NilString = AllocString(memPool, nfd->nfd_NilString);
	if (field->fld_NilString == NULL) {
	    FreeField(field);
	    return ERR_MEM;
	}
    }
    if (nfd->nfd_Type != TYPE_VIRTUAL) {
	CodesetStrNBytesCopy(field->fld_Trigger, nfd->nfd_Trigger,
		FIELD_TRIGGER_LEN);
    }
    CodesetTrimStr(field->fld_Trigger);
    field->fld_CountChanges = nfd->nfd_CountChanges;
    field->fld_LogChanges = nfd->nfd_LogChanges;
    di = &field->fld_DataInfo;

    /*
    ** basic initialisation
    */
    switch (field->fld_Type) {
        case TYPE_CHOICE:
	    field->fld_Contents.choice.labels = NULL;
            break;

        case TYPE_REFERENCE:
            field->fld_Contents.reference.table = NULL;
            di->di_Contents.rec = NULL;
            break;
    }

    if (error == 0) {
        error = InitFieldType(field, 
		nfd->nfd_InitialValue, &nfd->nfd_Contents);
    }
    if (error) {
        FreeField(field);
	return error;
    }

    error = NewFieldRecordInfo(field, &field->fld_FieldRecordInfo);
    if (error) {
	FreeField(field);
	return error;
    }

    if (newfield) {
	*newfield = field;
    }

    AddField(field->fld_Table, field, nfd->nfd_Pred);
    ProgramAddField(field);

    return OK;
}
 
long CanChangeField(struct Field *field)
{
    return CanNewFieldRecordInfo(FieldTable(field), FieldType(field));
}

long ChangeField(struct ChangeFieldData *cfd)
{
    struct Field *field = (cfd != NULL)? cfd->cfd_Field : NULL; 
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = TableMemPool(tab);
    char *nil = NULL;
    long error = 0;

    if (field == NULL) {        /* should never happen! */
	return ERR_INTERNAL;
    }
    if (tab == NULL) {
	return ERR_NOTABLE;
    }
    if (!IsMBName(cfd->cfd_Name) ||
            IsSystemName(cfd->cfd_Name) != TableIsSystem(tab)) {
	return ERR_INVALIDNAME;
    }
    if (FieldNameInUse(tab, cfd->cfd_Name, field)) {
	return ERR_NAMEALREADYUSED;
    }
    if (cfd->cfd_NilString) {
	nil = AllocString(memPool, cfd->cfd_NilString);
	if (nil == NULL) {
	    return ERR_MEM;
	}
    }

    error = InitFieldType(field, cfd->cfd_InitialValue, &cfd->cfd_Contents);
    if (error) {
	FreeString(memPool, nil);
	return error;
    }

    if (strcmp(FieldName(field), cfd->cfd_Name) != 0) {
        error = SQLiteRenameField(ProjectSQLiteHandle(pr), field,cfd->cfd_Name);
        if (error) {
            FreeString(memPool, nil);
            return error;
        }
        setFieldName(field, cfd->cfd_Name);
    }

    FreeString(memPool, field->fld_NilString);
    field->fld_NilString = nil;
    field->fld_Trigger[0] = 0;
    if (field->fld_Type != TYPE_VIRTUAL) {
	CodesetStrNBytesCopy(field->fld_Trigger, cfd->cfd_Trigger,
		FIELD_TRIGGER_LEN);
	CodesetTrimStr(field->fld_Trigger);
    }
    field->fld_CountChanges = cfd->cfd_CountChanges;
    field->fld_LogChanges = cfd->cfd_LogChanges;
    return OK;
}

long DeleteField(struct Field *field)
{
    struct Table *tab = FieldTable(field);
    long error = OK;

    if (!field || !tab) {			/* should never happen! */
	return ERR_INTERNAL;
    }

    error = DeleteFieldRecordInfo(field->fld_FieldRecordInfo);
    if (error) {
        return error;
    }
    field->fld_FieldRecordInfo = NULL;

    FilterDeleteField(field);
    QueryEditorDeleteField(field);
    DisplayDeleteFieldReference(field);
    ProgramRemField(field);
    RemField(tab, field);
    FreeField(field);

    return OK;
}

long FieldNum(struct Field *find)
{
    long num = 0;
    struct Table *tab = FieldTable(find);
    struct Project *pr = TableProject(tab);
    struct List *tabs = ProjectTables(pr), *fields;
    struct Field *field;

    if(tabs)
        FORNODE(tab, tabs)
        {
            fields = TableFields(tab);
            if(fields)
                FORNODE(field, fields)
                {
                    ++num;
                    if(field == find)
                        return(num);
                }
        }

    return(0);
}

struct Field *FindFieldNum(struct Project *pr, long find)
{
    struct List *tabs = ProjectTables(pr);
    struct List *fields;
    struct Table *tab;
    struct Field *field;
    long num = 0;

    if(tabs && find > 0)
        FORNODE(tab, tabs)
        {
            fields = TableFields(tab);
            if(fields)
                FORNODE(field, fields)
                {
                    ++num;
                    if(num == find)
                        return(field);
                }
        }
    return(NULL);
}

/*
** Load & Save of fields
*/

#define FIELDVERSION 15
#define MIN_FIELDVERSION 8

long NumFields(struct Project *pr)
{
    long num = 0;
    struct List *tabs = ProjectTables(pr);
    struct List *fields;
    struct Table *tab;
    struct Field *field;

    if (tabs != NULL) {
        FORNODE(tab, tabs) {
            fields = TableFields(tab);
            if (fields != NULL) {
	        FORNODE(field, fields) {
	            num++;
                }
            }
        }
    }
    return num;
}

long ReadField(struct Project *pr, const struct ReadFieldData *data)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct Field *field;
    struct DataInfo *di;
    long error = OK;

    if (pr == NULL || data == NULL) {
        return ERR_INTERNAL;
    }

    if (data->rfd_Table == NULL) {
        return DISKERR_INVALID;
    }

    field = AllocField(data->rfd_Table);
    if (field == NULL) {
        return ERR_MEM;
    }

    setFieldName(field, data->rfd_Name);
    field->fld_Type = data->rfd_Type;
    CodesetStrNBytesCopy(field->fld_Trigger, data->rfd_Trigger,
		FIELD_TRIGGER_LEN);
    field->fld_CountChanges = data->rfd_CountChanges;
    field->fld_LogChanges = data->rfd_LogChanges;
    if (data->rfd_NilString != NULL) {
        field->fld_NilString = AllocString(memPool, data->rfd_NilString);
        if (field->fld_NilString == NULL) {
	    error = ERR_MEM;
        }
    }

    di = &field->fld_DataInfo;
    *di = data->rfd_DataInfo;
    switch (field->fld_Type) {
        case TYPE_STRING:
            if (di->di_Contents.string != NULL) {
                di->di_Contents.string = 
                    AllocString(memPool, di->di_Contents.string);
                if (di->di_Contents.string == NULL) {
                    error = ERR_MEM;
                }
            }
            field->fld_Contents.string.maxlen= data->rfd_TypeData.string.maxlen;
            break;

        case TYPE_MEMO:
            if (di->di_Contents.mi != NULL) {
                di->di_Contents.mi = AllocString(memPool, di->di_Contents.mi);
                if (di->di_Contents.mi == NULL) {
                    error = ERR_MEM;
                }
            }
            break;

        case TYPE_CHOICE:
            if (data->rfd_TypeData.choice.labels == NULL) {
                error = DISKERR_INVALID;
            } else {
                field->fld_Contents.choice.labels =
                    StringToLabels(memPool, data->rfd_TypeData.choice.labels);
                if (field->fld_Contents.choice.labels == NULL) {
                    error = ERR_MEM;
                }
            }
            break;

        case TYPE_REFERENCE:
            field->fld_Contents.reference.table =
                data->rfd_TypeData.reference.table;
            field->fld_Contents.reference.filter =
                data->rfd_TypeData.reference.filter;
            field->fld_Contents.reference.autoshow =
                data->rfd_TypeData.reference.autoshow;
            break;

        case TYPE_VIRTUAL:
	    CodesetStrNBytesCopy(field->fld_Contents.virtual.compute,
		    data->rfd_TypeData.virt.compute, FIELD_TRIGGER_LEN);
            break;

        case TYPE_INTEGER:
        case TYPE_REAL:
        case TYPE_DATE:
        case TYPE_TIME:
        case TYPE_BOOL:
        case TYPE_BUTTON:
            break;

        default:
            error = DISKERR_INVALID;
            break;
    }

    if (error == OK) {
        field->fld_FieldRecordInfo = AllocFieldRecordInfo(field);
        if (field->fld_FieldRecordInfo == NULL) {
            error = ERR_MEM;
        }
    }

    /* Add field after allocating FieldRecordInfo. */
    AddField(data->rfd_Table, field, NULL);

    if (error == OK && FieldNum(field) != data->rfd_ID) {
        error = DISKERR_NOTABEEBASEPROJECT;
    }

    return error;
}

long LoadFieldsCustomFormat(struct Project *pr, FILE *file)
{
    long i, error = OK;
    struct Table *tab;
    struct Field *field;
    struct MemPool *memPool = ProjectMemPool(pr);
    struct DataInfo *di;
    short version;
    bool utf8 = FALSE;
    long num;

    if (pr == NULL || file == NULL) {
    	return ERR_INTERNAL;
    }

    version = FGetShort(file);
    if (version > FIELDVERSION || version < MIN_FIELDVERSION) {
	return DISKERR_VERSION;
    }
    if (version >= 14) {
        utf8 = FGetBool(file);
    }

    num = FGetLong(file);
    for (i = 0; i < num; i++) {
	long tablenum;
        size_t warnings = 0;
        short type;
	static char name[sizeof(field->fld_Name)];

	tablenum = FGetLong(file);
        tab = FindTableNum(pr, tablenum);
	if (tab == NULL) {
	    return DISKERR_INVALID;
	}

	FGetS(file, name, sizeof(name));        /* ASCII */
        field = AllocField(tab);
        if (field == NULL) {
            return ERR_MEM;
        }

        type = FGetShort(file);
        field->fld_Type = type;
	FGetS(file, field->fld_Trigger, sizeof(field->fld_Trigger)); /* ASCII */
        CodesetTrimStr(field->fld_Trigger);
	if (version >= 10) {
	    field->fld_CountChanges = FGetBool(file);
	} else {
	    field->fld_CountChanges = TRUE;
	}
        if (version>= 15) {
	    field->fld_LogChanges = FGetBool(file);
        } else {
	    field->fld_LogChanges = TRUE;
        }

        setFieldName(field, name);

        di = &field->fld_DataInfo;
        di->di_Flags = FGetChar(file);
        di->di_Private = FGetChar(file);
        if (version < 13) {
            di->di_Flags = ((di->di_Private &  1)? DIF_NIL : 0);
            di->di_Private = 0;
        }

	switch (field->fld_Type) {
	    case TYPE_STRING:
                FreeString(memPool, di->di_Contents.string);
		di->di_Contents.string = ReadStringConvert(file, utf8, memPool,
			&error, &warnings);
		field->fld_Contents.string.maxlen = FGetShort(file);
	    	break;

	    case TYPE_INTEGER:
		di->di_Contents.integer = FGetLong(file);
                FreeString(memPool, field->fld_NilString);
		field->fld_NilString = ReadStringConvert(file, utf8, memPool,
			&error, &warnings);
		break;

	    case TYPE_REAL:
		di->di_Contents.real = FGetDouble(file);
                FreeString(memPool, field->fld_NilString);
		field->fld_NilString = ReadStringConvert(file, utf8, memPool,
			&error, &warnings);
		break;

	    case TYPE_CHOICE:
		di->di_Contents.choice = FGetShort(file);
                field->fld_Contents.choice.labels = 
                    LoadLabels(memPool, file, utf8, &error, &warnings);
		break;

 	    case TYPE_DATE:
		di->di_Contents.date = FGetLong(file);
                FreeString(memPool, field->fld_NilString);
		field->fld_NilString = ReadStringConvert(file, utf8, memPool,
			&error, &warnings);
		break;

  	    case TYPE_TIME:
		di->di_Contents.time = FGetLong(file);
                FreeString(memPool, field->fld_NilString);
		field->fld_NilString = ReadStringConvert(file, utf8, memPool,
			&error, &warnings);
		break;

            case TYPE_REFERENCE:
	    {
		long num = FGetLong(file);

		field->fld_Contents.reference.filter = FGetBool(file);
		if (version >= 9) {
		    field->fld_Contents.reference.autoshow = FGetBool(file);
		}
                field->fld_Contents.reference.table = FindTableNum(pr, num);
		di->di_Flags |= DIF_NIL;
		di->di_Contents.rec = NULL;
                break;
	    }

            case TYPE_VIRTUAL:
		FGetS(file, field->fld_Contents.virtual.compute,
                        sizeof(field->fld_Contents.virtual.compute));/* ASCII */
	    	break;

            case TYPE_MEMO:
	    case TYPE_BOOL:
            case TYPE_BUTTON:
                break;

            default:
                error = DISKERR_INVALID;
                break;
	}

        ProjectAddCodesetWarnings(pr, warnings);
        if (error == OK) {
            error = LoadFieldRecordInfoCustomFormat(field, file,
                    &field->fld_FieldRecordInfo);
	}
        AddField(tab, field, NULL);
	if (error) {
            break;
	}
    }
    return error;
}

