/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** field.h
**
** Creation date: 15.11.1994
** Last modified: 15.07.2020
*/

#ifndef FIELD_H
#define FIELD_H

struct Field;

#ifdef NEED_NODE_CAST
inline Field *nodecast(Field *, Node *node) { return (Field *)node;}
#endif

#include "system.h"
#include "project.h"
#include "record.h"
#include "dbtypes.h"

#ifdef __cplusplus
extern "C" {
#endif

#define MAX_STRING_LEN 999
#define FIELD_INITIALVALUE_LEN 255
#define FIELD_NILVALUE_LEN 255
#define FIELD_TRIGGER_LEN 63

union FieldTypeData
{
    struct
    {
	short maxlen;
    } string;
    struct
    {
	bool nil;
    } boolean;                        	/* bool is keyword in C++ */
    struct
    {
        char **labels;			/* Codeset strings */
        ushort initialvalue;
    } choice;
    struct
    {
        struct Table *table; 
        bool filter;
	bool autoshow;
    } reference;
    struct
    {
        const char *compute;		/* ASCII */
    } virt;                             /* virtual is keyword in c++ */
};

struct NewFieldData
{
    struct Table *nfd_Table;
    struct Field *nfd_Pred;
    const char *nfd_Name;		/* ASCII */
    short nfd_Type;
    const char *nfd_InitialValue;	/* Codeset */
    const char *nfd_NilString;		/* Codeset */
    const char *nfd_Trigger;		/* ASCII */
    bool nfd_CountChanges;
    bool nfd_LogChanges;
    union FieldTypeData nfd_Contents;
};

struct ChangeFieldData
{
    struct Field *cfd_Field;
    const char *cfd_Name;		/* ASCII */
    const char *cfd_InitialValue;	/* Codeset */
    const char *cfd_NilString;		/* Codeset */
    const char *cfd_Trigger;		/* ASCII */
    bool cfd_CountChanges;
    bool cfd_LogChanges;
    union FieldTypeData cfd_Contents;
};

struct ReadFieldData
{
    long rfd_ID;
    struct Table *rfd_Table;
    const char *rfd_Name;		/* ASCII */
    short rfd_Type;
    const char *rfd_Trigger;		/* ASCII */
    bool rfd_CountChanges;
    bool rfd_LogChanges;
    struct DataInfo rfd_DataInfo;
    const char *rfd_NilString;		/* Codeset */
    union {
        struct {
            short maxlen;
        } string;
        struct {
            const char *labels;         /* Codeset */
        } choice;
        struct {
            struct Table *table;
            bool filter;
            bool autoshow;
        } reference;
        struct {
            const char *compute;        /* ASCII */
        } virt;
    } rfd_TypeData;
};

void FreeField(struct Field *field);
short FieldType(struct Field *field);
const char *FieldName(struct Field *field);
struct Table *FieldTable(struct Field *field);
short FieldStringMaxlen(struct Field *field);
struct FieldRecordInfo *FieldRecordInfo(struct Field *field);
struct DataInfo *FieldDataInfo(struct Field *field);
const char *FieldNilString(struct Field *field);
const char *FieldTrigger(struct Field *field);
bool FieldCountChanges(struct Field *field);
bool FieldLogChanges(struct Field *field);
ushort FieldChoiceNumLabels(struct Field *field);
char **FieldChoiceLabels(struct Field *field);
struct Table *FieldReferenceTable(struct Field *field);
struct Field *FindReferenceTable(struct Table *ref);
bool FieldReferenceFilter(struct Field *field);
void SetFieldReferenceFilter(struct Field *field, bool filter);
bool FieldReferenceAutoShow(struct Field *field);
void SetFieldReferenceAutoShow(struct Field *field, bool autoshow);
const char *FieldVirtualCompute(struct Field *field);

struct DisplayObject *FieldDisplayObject(struct Field *field);
void SetFieldDisplayObject(struct Field *field, 
	struct DisplayObject *dio);

struct Field *FindFieldName(struct Table *tab, const char *name);

long CanNewField(struct Table *tab);
long NewField(struct NewFieldData *nfd, struct Field **newfield);
/* returns error code or 0 for success. */
long CanChangeField(struct Field *field);
long ChangeField(struct ChangeFieldData *cfd);
/* returns error code or 0 for success. */
long DeleteField(struct Field *field);
/* returns error code or 0 for success. */

long FieldNum(struct Field *field);
struct Field *FindFieldNum(struct Project *pr, long find);
long NumFields(struct Project *pr);
long ReadField(struct Project *pr, const struct ReadFieldData *data);
long LoadFieldsCustomFormat(struct Project *pr, FILE *file);

#ifdef __cplusplus
}
#endif

#endif

