/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** project.h
**
** Creation date: 25.09.2014
** Last modified: 08.09.2015
*/

#include "system.h"
#include "filelock.h"
#include "dosutil.h"
#include "memory.h"
#include "error.h"

struct FileLockContext
{
    struct MemPool *flc_MemPool;
    FileLockLevel flc_Level;
    bool flc_IsCompatibility;
    char *flc_SLockName;
    char *flc_RLockName;
    char *flc_PLockName;
    struct Lock *flc_SLock;
    struct Lock *flc_RLock;
    struct Lock *flc_PLock;
};

struct FileLockContext *AllocFileLockContext(struct MemPool *memPool,
        const char *filename)
{
    struct FileLockContext *flc = NULL;
    size_t nameSize;
    bool isDir;
    const char *sep;

    if (memPool == NULL || filename == NULL) {
        return NULL;
    }

    isDir = FileStat(filename) == 0;
    sep = isDir? "/" : "";
    nameSize = strlen(filename) + strlen(sep) + 1 + 5 + 1;
    flc = PoolAllocMemZero(memPool, sizeof(*flc) + 3 * nameSize);
    if (flc == NULL) {
        return NULL;
    }

    flc->flc_MemPool = memPool;
    flc->flc_SLockName = (char *)(flc + 1);
    strcpy(flc->flc_SLockName, filename);
    strcat(flc->flc_SLockName, sep);
    strcat(flc->flc_SLockName, ".lock");
    flc->flc_RLockName = flc->flc_SLockName + nameSize;
    strcpy(flc->flc_RLockName, filename);
    strcat(flc->flc_RLockName, sep);
    strcat(flc->flc_RLockName, ".rlock");
    flc->flc_PLockName = flc->flc_RLockName + nameSize;
    strcpy(flc->flc_PLockName, filename);
    strcat(flc->flc_PLockName, sep);
    strcat(flc->flc_PLockName, ".plock");

    return flc;
}

void FreeFileLockContext(struct FileLockContext *flc)
{
    if (flc) {
        size_t nameSize = strlen(flc->flc_PLockName) + 1;

        ReleaseFileLockLevel(flc);
        PoolFreeMem(flc->flc_MemPool, flc, sizeof(*flc) + 3 * nameSize);
    }
}

FileLockLevel CurrentFileLockLevel(struct FileLockContext *flc)
{
    FileLockLevel level = NO_FILE_LOCK;

    if (flc) {
        level = flc->flc_Level;
    }
    return level;
}

long RaiseFileLockLevel(struct FileLockContext *flc, FileLockLevel level)
{
    if (flc == NULL) {
        return ERR_INTERNAL;
    }
    while (flc->flc_Level < level) {
        bool isLocked = FALSE;

        if (flc->flc_IsCompatibility) {
            return DISKERR_LOCK;
        }

        switch (flc->flc_Level) {
            case NO_FILE_LOCK:
                if (FileLocked(flc->flc_PLockName)) {
                    return DISKERR_LOCKED;
                }
                flc->flc_SLock = FileLockShared(flc->flc_SLockName);
                if (flc->flc_SLock == NULL) {
                    return DISKERR_LOCK;
                }
                /* 
                ** In case another tasks locked the pending lock in the meantime
                ** we backup and unlock.  This can happen under race conditions.
                */
                if (FileLocked(flc->flc_PLockName)) {
                    FileUnlock(flc->flc_SLock);
                    flc->flc_SLock = NULL;
                    return DISKERR_LOCKED;
                }
                flc->flc_Level = SHARED_FILE_LOCK;
                break;
            case SHARED_FILE_LOCK:
                flc->flc_RLock = FileLockExclusive(flc->flc_RLockName,
                        &isLocked);
                if (flc->flc_RLock == NULL) {
                    return isLocked? DISKERR_LOCKED : DISKERR_LOCK;
                }
                flc->flc_Level = RESERVED_FILE_LOCK;
                break;
            case RESERVED_FILE_LOCK:
                if (flc->flc_PLock == NULL) {
                    flc->flc_PLock = FileLockExclusive(flc->flc_PLockName,
                            &isLocked);
                    if (flc->flc_PLock == NULL) {
                        return isLocked? DISKERR_LOCKED : DISKERR_LOCK;
                    }
                }
                /* Relase our shared lock */
                FileUnlock(flc->flc_SLock);
                flc->flc_SLock = NULL;
                /* Check if anyone else holds a lock */
                if (FileLocked(flc->flc_SLockName)) {
                    FileUnlock(flc->flc_PLock);
                    flc->flc_PLock = NULL;
                    return DISKERR_LOCKED;
                }
                flc->flc_Level = EXCLUSIVE_FILE_LOCK;
                break;
            case EXCLUSIVE_FILE_LOCK:
                break;
        }
    }

    return OK;
}

void ReleaseFileLockLevel(struct FileLockContext *flc)
{
    if (flc != NULL) {
        FileUnlock(flc->flc_PLock);
        flc->flc_PLock = NULL;
        FileUnlock(flc->flc_RLock);
        flc->flc_RLock = NULL;
        FileUnlock(flc->flc_SLock);
        flc->flc_SLock = NULL;
        flc->flc_Level = NO_FILE_LOCK;
        flc->flc_IsCompatibility = FALSE;
    }
}

bool FileLockHasReservedLock(struct FileLockContext *flc)
{
    bool reservedLock = FALSE;

    if (flc != NULL && !flc->flc_IsCompatibility) {
        reservedLock = FileLocked(flc->flc_RLockName);
    }
    return reservedLock;
}

long ObtainCompatibilityFileLockLevel(struct FileLockContext *flc)
{
    bool isLocked = FALSE;
    long error = OK;

    if (flc == NULL) {
        return ERR_INTERNAL;
    }
    ReleaseFileLockLevel(flc);

    flc->flc_SLock = FileLockExclusive(flc->flc_SLockName, &isLocked);
    if (flc->flc_SLock != NULL) {
        flc->flc_Level = EXCLUSIVE_FILE_LOCK;
        flc->flc_IsCompatibility = TRUE;
    } else if (isLocked) {
        error = DISKERR_LOCKED;
    } else {
        flc->flc_SLock = FileLockShared(flc->flc_SLockName);
        if (flc->flc_SLock != NULL) {
            flc->flc_Level = SHARED_FILE_LOCK;
            flc->flc_IsCompatibility = TRUE;
        } else {
            error = DISKERR_LOCK;
        }
    }

    return error;
}

bool IsCompatibilityFileLockLevel(struct FileLockContext *flc)
{
    bool isCompatibility = FALSE;

    if (flc != NULL) {
        isCompatibility = flc->flc_IsCompatibility;
    }
    return isCompatibility;
}

