/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** filter.c
**
** Creation date: 23.04.1995
** Last modified: 21.07.2020
*/

#include "system.h"
#include "dbtypes.h"
#include "project.h"
#include "table.h"
#include "field.h"
#include "display.h"
#include "record.h"
#include "gui.h"
#include "error.h"
#include "lang/program.h"
#include "lang/compile.h"
#include "requesters.h"
#include "break.h"
#include "appstrings_cat.h"

static void filterOff(struct Table *tab)
{
    struct List *recs = TableRecords(tab);
    struct Record *rec;

    if (recs) {
       	FORNODE(rec, recs) {
    	    SetRecordMatchFilter(rec, TRUE);
	}
    }
    SetTableFilterActive(tab, FALSE);
}

static void filterOn(struct Table *tab)
{
    struct Program *prg = TableFilterProgram(tab);
    struct List *recs = TableRecords(tab);
    const char *error = NULL;

    if (prg == NULL || recs == NULL) {
	filterOff(tab);
	return;
    } 

    if (ProgramNeedRecompile(prg)) {
	struct SourceText *st = GetSourceText(prg);
	const char *text = SourceTextText(st);
	struct Program *compiled = NULL;

	if (text == NULL) {
	    error = GES(ERR_MEM);
	} else {
	    error = CompileFilter(tab, text, strlen(text), &compiled, 
		    NULL, NULL);
	}
	if (compiled) {
	    SetTableFilterProgram(tab, compiled);
	}
	prg = compiled;
	FreeSourceText(st);
    }
    if (error == NULL) {
	struct BreakSettings bs;
	struct Project *pr = TableProject(tab);
	struct Record *rec;
	struct Element *elem;

	bs.title = GCS(MSG_ApplyingFilter); 
	bs.steps = TableNumRecords(tab);
	bs.current = 0;
	bs.abort = TRUE;
	bs.pause = TRUE;
        bs.needGui = ProjectHasGui(pr);
	BreakInit(&bs);
	SetTableProgramRecords(pr);
	ProgramBegin(prg, FALSE);

	FORNODE(rec, recs) {
	    if (BreakCheck()) {
		break;
	    }
	    SetTableProgramRecord(tab, rec);
	    elem = ProgramCall(prg, NULL, NULL, FALSE);
	    if (elem == NULL) {                      /* error? */
		break;
	    }
	    SetRecordMatchFilter(rec, elem->el_Type != TYPE_NIL);
	    BreakAddCurrentStep(1);
	}

        ProgramEnd();
    	error = ProgramError(NULL, NULL, NULL);
	if (error == NULL && BreakCheck()) {
	    error = EXECUTE_ABORT;
	}
	BreakExit(&bs);
    }

    if (error) {
	if (error != EXECUTE_ABORT) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_SetFilter), error);
	}
	filterOff(tab);
	return;
    }

    SetTableFilterActive(tab, TRUE);
}

void FilterOnOff(struct Table *tab, bool on)
{
    GuiSetSleep(TRUE);
    if (on) {
	filterOn(tab);
    } else {
	filterOff(tab);
    }
    GuiCheckGuiRecord(tab);
    GuiSetSleep(FALSE);
}

void SetFilter(struct Table *tab, struct Program *prg)
{
    GuiSetSleep(TRUE);
    SetTableFilterProgram(tab, prg);
    filterOn(tab);
    GuiCheckGuiRecord(tab);
    GuiSetSleep(FALSE);
}

static void filterDelete(struct Project *pr, void *ptr)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    struct Program *prg;

    if(tabs)
        FORNODE(tab, tabs)
        {
            prg = TableFilterProgram(tab);
            if(prg && ProgramFindObject(prg, ptr))
                SetFilter(tab, NULL);
        }
}

void FilterDeleteTable(struct Table *tab)
{
    struct Project *pr = TableProject(tab);
    struct List *fields = TableFields(tab);
    struct Field *field;

    filterDelete(pr, tab);
    if(fields)
        FORNODE(field, fields)
            filterDelete(pr, field);
}

void FilterDeleteField(struct Field *field)
{
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);

    filterDelete(pr, field);
}

/*
** load & save
*/

long LoadFiltersCustomFormat(struct Project *pr, FILE *file)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    struct Program *prg;
    long error = OK;

    if (tabs) {
	FORNODE(tab, tabs) {
	    error = LoadProgramCustomFormat(pr, file, &prg);
	    if (error) {
		break;
	    }
	    SetTableFilterProgram(tab, prg);
	}
    }
    return error;
}

