/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** formatmemo.c
**
** Thanks to Ralph Reuchlein for providing some ideas and infos.
**
** Creation date: 06.03.1997
** Last modified: 12.11.2007
*/

#include "system.h"
#include "formatmemo.h"
#include "strhandle.h"
#include "util.h"
#include "memory.h"

#define ISDOT(x) ((x) == '.' || (x) == '!' || (x) == '?' || (x) == ':')

static bool processline(const char *line, size_t n, long fwidth, long width, 
	bool fill, struct StrHandle *sh, struct MemPool *memPool)
{
    const char *s = line, *end = line + n, *lastWord, *t;
    long size, len, lastLen, w = fwidth, spaces;
    char *buffer = NULL, *b, *lastSpace;
    cschar c;

    if (fwidth <= 0 || width <= 0) {
        return FALSE;
    }
    if (s >= end) {                           /* empty line? */
        return TRUE;
    }

    /* Allocate buffer for string line processing */
    size = MAX(fwidth, width)*CodesetMaxBytesForChar + 1;
    buffer = PoolAllocMem(memPool, size);
    if (buffer == NULL) {
        return FALSE;
    }

    while (s < end) {
        /* Fill buffer up to w chars */
	b = buffer;
	lastSpace = buffer;
	lastWord = s;
	lastLen = 0;
        len = 0;
	do {
	    c = CodesetGetChar(s);
	    if (CodesetIsBrkSpace(c)) {
		/* skip white space */
		do {
		    s = CodesetNextChar(s);
		} while (s < end && CodesetIsBrkSpace(CodesetGetChar(s)));
		lastSpace = b;
		lastWord = s;
		lastLen = len;
		if (b == buffer) {
		    spaces = 0;
		} else if (!ISDOT(b[-1])) {
		    spaces = 1;
		} else {
		    spaces = 1;		/* English: 2, but abbreviations: 1 */
		}
		if (len + spaces >= w) {
		    break;
		}
		len += spaces;
		while (spaces > 0) {
		    *b++ = ' ';
		    spaces--;
		}
	    } else if (len < w) {
		t = CodesetNextChar(s);
		while (s < t) {
		    *b++ = *s++;
		}
		len++;
	    } else {
		break;
	    }
	} while (s < end);
	*b = 0;

	if (s < end && lastSpace != buffer) {
	    *lastSpace = 0;
	    s = lastWord;
	    len = lastLen;
	}

        spaces = NumWords(buffer) - 1;
        if (fill && s < end && len < w && spaces > 0) {
            long free = w - len, insert, j;
	    char c;

            for (t = buffer; (c = *t) != 0; t++) {
                StrHandlePutC(sh, c);
		if (c == ' ') {
		    while (t[1] == ' ') {
                        StrHandlePutC(sh, ' ');
			t++;
		    }
                    insert = (free + spaces - 1) / spaces;
                    free--;
                    for (j=0; j < insert; j++) {
                        StrHandlePutC(sh, ' ');
		    }
		}
            }
        } else {
            StrHandlePutS(sh, buffer);      /* write buffer */
	}

        if (s < end) {
            StrHandlePutC(sh, '\n');
	}
        w = width;
    }
    PoolFreeMem(memPool, buffer, size);
    return TRUE;
}

#define ISFEED(c) ((c) == '\n' || (c) == '\f')

char *FormatMemo(const char *src, long width, bool fill, bool singleline,
	struct MemPool *memPool)
{
    struct StrHandle *sh;
    const char *s = src, *t;
    cschar c;
    long left = 0;
    char *result;

    if (src == NULL) {
        return NULL;
    }

    sh = AllocStrHandle(memPool);
    if (sh == NULL) {
        return NULL;
    }

    while (*s != 0) {
        /* search for start of a paragraph */
        for (t = s; *t != 0; t = CodesetNextChar(t), left++) {
	    c = CodesetGetChar(t);
	    if (!CodesetIsBrkSpace(c)) {
		break;
	    }
            if (ISFEED(c)) {
                StrHandlePutC(sh, (char)c);
                s = t+1;
		left = -1;
            }
        }

        /* Print spaces for first line */
        if (left >= width) {
            StrHandlePutC(sh, '\n');
	    s = t;
            left = 0;
        }
	while (s < t) {
            StrHandlePutC(sh, *s++);
	}

        /* Find end of paragraph */
	for (t = s; *t != 0; t = CodesetNextChar(t)) {
	    if (*t == '\n') {
		c = CodesetGetChar(t+1);
		if (singleline || c == 0 || CodesetIsBrkSpace(c)) {
		   break;
		} 
  	    } else if (*t == '\f') {
		break;
	    }
        }

        if (!processline(s, t - s, width - left, width, fill, sh, memPool)) {
            FreeStrHandle(sh);
            return NULL;
        }
        s = t;
    }
    StrHandlePutC(sh, 0);                   /* terminate string */

    result = AllocString(memPool, StrHandleString(sh));
    FreeStrHandle(sh);

    return result;
}

