/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gui.c
**
** Creation date: 19.03.1995
** Last modified: 10.02.2022
*/

#include "system.h"
#include "project.h"
#include "preferences.h"
#include "table.h"
#include "field.h"
#include "display.h"
#include "record.h"
#include "order.h"
#include "gui.h"
#include "buildgui.h"
#include "filter.h"
#include "queryeditor.h"
#include "dosutil.h"
#include "memo.h"
#include "strhandle.h"
#include "labels.h"
#include "memory.h"
#include "programeditor.h"
#include "lang/program.h"
#include "error.h"
#include "util.h"
#include "requesters.h"
#include "appstrings_cat.h"

long GuiFieldReferenceDataString(struct Field *field, 
        struct Record *refrec, struct MemPool *memPool, 
        char **ptext, bool *pfree)
{
    struct DisplayObject *dio = FieldDisplayObject(field);
    long error = OK;
    char *text = NULL;
    bool free = FALSE;

    if (FieldType(field) != TYPE_REFERENCE) {
        error = ERR_INTERNAL;
    } else {
        struct RecordDescription *rd = AllocRecordDescription(
                FieldReferenceTable(field),
                DisplayFieldRecordDescriptionSpec(dio, TRUE));

        if (rd == NULL) {
            error = ERR_MEM;
        } else {
            RecordDescriptionSetRecord(rd, refrec);

            text = RecordDescriptionText(rd, memPool, &free);
            error = RecordDescriptionError(rd);
            FreeRecordDescription(rd);
        }
    }

    if (ptext != NULL) {
        *ptext = text;
    }
    if (pfree != NULL) {
        *pfree = free;
    }
    return error;
}

long GuiFieldDataString(struct Field *field, struct Record *rec, struct MemPool *memPool, char **ptext, bool *pfree)
{
    short type = FieldType(field);
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    struct DisplayObject *dio = FieldDisplayObject(field);
    struct DataInfo *di = NULL, di_buf;
    long error = OK;
    const char *text = NULL;
    bool free = FALSE;

    if(ptext != NULL)
        *ptext = NULL;
    if(pfree != NULL)
        *pfree = FALSE;

    error = RecordDataInfo(rec, field, &di, &di_buf);
    if(error)
        return(error);

    switch(type)
    {
        case TYPE_STRING:
            text = EmptyNullStr(di->di_Contents.string);
            break;

        case TYPE_INTEGER:
            if (di->di_Flags & DIF_NIL) {
                text = FieldNilString(field);
            } else {
                text = AllocString(memPool, IntToStr(di->di_Contents.integer));
		if (text == NULL)
		    return(ERR_MEM);
		free = TRUE;
	    }
            break;

        case TYPE_REAL:
            if (di->di_Flags & DIF_NIL) {
                text = FieldNilString(field);
            } else {
		text = AllocString(memPool, DoubleToStr(di->di_Contents.real,
			    DisplayFieldNumDecimals(dio),
			    (short)GetPrefsValue(pr, PREFS_REALFORMAT)));
		if (text == NULL)
		    return(ERR_MEM);
		free = TRUE;
	    }
            break;

        case TYPE_BOOL:
            text = (di->di_Flags & DIF_NIL)? "NIL" : "TRUE";
            break;

        case TYPE_CHOICE:
        {
            char **labels = FieldChoiceLabels(field);

	    if (di->di_Contents.choice < NumLabels(labels))
                text = labels[di->di_Contents.choice];
	    else
	    {
		text = IntToStr(di->di_Contents.choice);
		text = AllocString(memPool, text);
		if (text == NULL)
		    return(ERR_MEM);
		free = TRUE;
	    }
            break;
        }

        case TYPE_DATE:
            if (di->di_Flags & DIF_NIL) {
                text = FieldNilString(field);
            } else {
                text = AllocString(memPool, DateToString(di->di_Contents.date,
			    (short)GetPrefsValue(pr, PREFS_DATEFORMAT)));
		if (text == NULL)
		    return(ERR_MEM);
		free = TRUE;
	    }
            break;

        case TYPE_TIME:
            if (di->di_Flags & DIF_NIL) {
                text = FieldNilString(field);
            } else {
                text = AllocString(memPool, TimeToString(di->di_Contents.time, 
			    DisplayFieldFormat(dio)));
		if (text == NULL)
		    return(ERR_MEM);
		free = TRUE;
	    }
            break;

        case TYPE_MEMO:
            text = EmptyNullStr(di->di_Contents.mi);
            break;

        case TYPE_REFERENCE:
        {
            struct Record *refrec = di->di_Contents.rec;

            RecordLock(refrec);
            error = GuiFieldReferenceDataString(field, refrec, memPool, 
		    (char **)&text, &free);
            RecordUnlock(refrec);
            if(error != OK)
                return(error);
            break;
        }

        default:
            return(ERR_INTERNAL);
    }

    if(ptext != NULL)
        *ptext = (char *)text;
    if(pfree != NULL)
        *pfree = free;
    return(OK);
}

void GuiActivateDisplayObject(struct DisplayObject *dio)
{
    struct DisplayObject *windio = DisplayObjectWindow(dio);
    GuiObject win = DisplayObjectGuiObject(windio);
    GuiObject go = DisplayObjectActivateObject(dio);

    if(win && go)
    {
        GuiSetActiveGuiObject(win, NULL);
	GuiActivateWindow(win);
	GuiSetActiveGuiObject(win, go);
        GuiWindowToFront(win); 
    }
}

bool GuiDisplayObjectVisible(struct DisplayObject *dio)
{
    struct DisplayObject *windio = DisplayObjectWindow(dio), *parent, *child;
    GuiObject go = DisplayObjectGuiObject(dio);
    GuiObject win = DisplayObjectGuiObject(windio);
  
    if (!DisplayObjectVisible(dio) || go == NULL) {  /* not in gui? */
        return FALSE;
    }
    if (!GuiWindowOpen(win)) {
        return FALSE;                                  /* no window */
    }

    for (; dio != NULL && dio != windio; dio = parent) {
        struct List *childs;
        long pos = 0;

        parent = DisplayObjectParent(dio);
        childs = DisplayObjectObjects(parent);
        if (DisplayObjectType(parent) == DO_TYPE_REGISTER && childs != NULL) {
            FORNODE (child, childs) {
                if (child == dio) {
                    break;
                }
                if (DisplayObjectVisible(child)) {
                    pos++;
                }
            }
            if (pos != GuiRegisterActivePage(DisplayObjectGuiObject(parent))) {
                return FALSE;
            }
	}
    }

    return TRUE;
}

bool GuiCanLoadImageSpec(struct Project *pr, const char *spec)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    char *expand;
    bool canLoad = TRUE;
#ifdef AMIGA
    const char sectionChar = '\247';
    char *section = NULL;
#endif

    if (spec != NULL && spec[0] == '5' && spec[1] == ':') {
        ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
        expand = ExpandPathname(&spec[2], memPool);
#ifdef AMIGA
        if (expand != NULL) {
            section = strchr(expand, sectionChar);
            if (section != NULL) {
                *section = 0;
            }
        }
#endif
        canLoad = FileStat(expand) > 0;
#ifdef AMIGA
        if (section != NULL) {
            *section = sectionChar;
        }
#endif
        FreeString(memPool, expand);
    }

    return canLoad;
}

bool BeginGuiProgramContext(struct Project *pr)
{
    struct Program *prg = ProjectProgram(pr);
    bool context = FALSE;

    if (prg != NULL && !ProgramRunning()) {
        SetTableProgramRecords(pr);
        ProgramBegin(prg, TRUE); 
        context = TRUE;
    }
    return context;
}

bool EndGuiProgramContext(struct Project *pr, bool context)
{
    bool success = TRUE;

    if (context) {
        const char *error, *filename = NULL;
        long line, col;

        ProgramEnd();
        error = ProgramError(&line, &col, &filename);
        if( error != NULL && error != EXECUTE_ABORT) {
            ProgramEditorShowError(pr, TRUE, error, line, col, filename);
            success = FALSE;
        }
    }

    return success;
}

bool ExecuteTrigger(struct Project *pr, const char *trigger,
	struct Element **args, const char *(*func)(struct Element *))
{
    struct Program *prg = ProjectProgram(pr);
    struct UserFunction *uf;
    bool context;
    struct Element *elem;
    const char *error = NULL;

    if (pr == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Internal));
        return FALSE;
    }
    if (trigger == NULL || trigger[0] == 0) {
        return TRUE;                                     /* nothing to do */
    }
    uf = FindUserFunctionName(prg, trigger);
    if (uf == NULL) {
        Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Err_Trigger), trigger);
        return FALSE;
    }

    context = BeginGuiProgramContext(pr);
    elem = ProgramCall(prg, uf, args, TRUE);
    if (func != NULL) {
        error = func(elem);
    }
    if (!EndGuiProgramContext(pr, context)) {
        return FALSE;
    }
    if (error != NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), error);
        return FALSE;
    }
    return TRUE;
}


static bool guiCheckError(long error, bool *askIfError)
{
    bool rc = TRUE;

    if (IsProjectError(error)) {
        if (*askIfError) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            *askIfError = FALSE;
        }
        rc = FALSE;
    }
    return rc;
}

bool guiUpdateDependentDisplayObjectsInternal(struct Project *pr, void *dep,
        bool (*accept)(struct DisplayObject *dio, void *ptr), void *ptr,
        bool *askIfError)
{
    const char *filename = NULL, *errstr;
    long line, col;

    errstr = UpdateDependentDisplayObjects(pr, dep, accept, ptr,
            &line, &col, &filename);
    if (errstr != NULL && errstr != EXECUTE_ABORT) {
        if (*askIfError) {
            ProgramEditorShowError(pr, TRUE, errstr, line, col, filename);
            *askIfError = FALSE;
        }
        return FALSE;
    }
    return TRUE;
}

bool GuiUpdateDependentDisplayObjects(struct Project *pr, void *dep,
        bool (*accept)(struct DisplayObject *dio, void *ptr), void *ptr)
{
    bool askIfError = TRUE;

    return guiUpdateDependentDisplayObjectsInternal(pr, dep, accept, ptr,
            &askIfError);
}

static bool acceptThis(struct DisplayObject *dio, void *this)
{
    return (dio == this);
}

bool GuiUpdateComputedDisplayObject(struct DisplayObject *dio)
{
    struct Project *pr = DisplayObjectProject(dio);

    // all computed objects depend on the project
    return GuiUpdateDependentDisplayObjects(pr, pr, acceptThis, dio);
}

static void guisetenabled(struct DisplayObject *dio, bool enabled)
{
#if 0
    GuiObject go = DisplayFieldDisableObject(dio);

    GuiSetEnabled(go, enabled);
#else
    struct Field *field = DisplayFieldField(dio);
    short type = FieldType(field);

    GuiObject guiobject = DisplayObjectGuiObject(dio);
    GuiObject file = DisplayFieldFileObject(dio);
    GuiObject font = DisplayFieldFontObject(dio);
    GuiObject lv = DisplayFieldListviewObject(dio);
    GuiObject view = DisplayFieldViewObject(dio);
    GuiObject show = DisplayFieldShowObject(dio);

    GuiSetEnabled(file, enabled);
    GuiSetEnabled(font, enabled);
    GuiSetEnabled(lv, enabled);
    GuiSetEnabled(view, enabled);
    GuiSetEnabled(show, enabled);

    if ((type == TYPE_BOOL && DisplayFieldReadOnly(dio)) ||
	    (type == TYPE_VIRTUAL && 
	     DisplayFieldSubType(dio) == DAD_VIRTUALTYPE_BOOL)) {
	enabled = FALSE;
    }
    GuiSetEnabled(guiobject, enabled);

#endif
}

static void guiseterror(struct Field *field, char *error)
{
    short type = FieldType(field);
    struct DisplayObject *dio = FieldDisplayObject(field);
    GuiObject go = DisplayObjectGuiObject(dio);
    GuiObject external = DisplayFieldExternalObject(dio);
    bool readonly = DisplayFieldReadOnly(dio);
    bool visible = DisplayObjectVisible(dio);
    short subtype = DisplayFieldSubType(dio);

    if(visible && (go || external))
    {
        switch(type)
        {
	    case TYPE_STRING:
            case TYPE_INTEGER:
            case TYPE_REAL:
            case TYPE_DATE:
            case TYPE_TIME:
                GuiSetString(go, error, readonly, external);
                break;

            case TYPE_BOOL:
                GuiSetBoolean(go, FALSE, readonly);
                break;

            case TYPE_CHOICE:
                GuiSetChoice(go, 0, FieldChoiceLabels(field), readonly);
		break;

            case TYPE_MEMO:
                GuiSetMemo(go, error, readonly);
                break;

            case TYPE_BUTTON:
                break;

            case TYPE_REFERENCE:
                GuiSetString(go, error, TRUE, NULL);
                break;

            case TYPE_VIRTUAL:
                switch(subtype)
                {
                    case DAD_VIRTUALTYPE_BOOL:
                        GuiSetBoolean(go, FALSE, readonly);
                        break;

                    case DAD_VIRTUALTYPE_TEXT:
                        GuiSetString(go, error, readonly, external);
                        break;

                    case DAD_VIRTUALTYPE_LIST:
                        GuiSetSFWList(go, NULL);
                        break;
                }
                break;
        }
        guisetenabled(dio, FALSE);
    }
}

const char *GuiSetVirtualField(struct Field *field, struct Record *rec, 
	struct Element *elem)
{
    struct DisplayObject *dio = FieldDisplayObject(field);
    struct Project *pr = DisplayObjectProject(dio);
    GuiObject go = DisplayObjectGuiObject(dio);
    struct DisplayObject *windio = DisplayObjectWindow(dio);
    GuiObject win = DisplayObjectGuiObject(windio);
    bool visible = DisplayObjectVisible(dio);
    bool readonly = TRUE;
    short subtype = DisplayFieldSubType(dio);
    const char *ret = NULL;

    if (!visible || go == NULL || !GuiWindowOpen(win)) {
        return NULL;
    }

    switch (subtype)
    {
        case DAD_VIRTUALTYPE_BOOL:
            GuiSetBoolean(go, elem && elem->el_Type != TYPE_NIL, readonly);
            break;

        case DAD_VIRTUALTYPE_TEXT:
        {
            struct MemPool *memPool = ProjectMemPool(pr);
            char *mem;
            bool allocated = FALSE;
            const char *text;

            mem = SimpleElementToString(elem, field, memPool, &allocated);
            text = mem;
            if (text == NULL) {
                text = ret = GCS(MSG_Err_Mem);
	    }
            GuiSetString(go, text, readonly, NULL);
            if (allocated) {
                FreeString(memPool, mem);
	    }
            break;
        }

        case DAD_VIRTUALTYPE_LIST:
            GuiSetSFWList(go, elem);
            break;
    }

    return ret;
}

bool GuiSetVirtualFieldListActive(struct DisplayObject *dio, long num)
{
    GuiObject go = DisplayObjectGuiObject(dio);
    struct DisplayObject *windio = DisplayObjectWindow(dio);
    GuiObject win = DisplayObjectGuiObject(windio);
    bool visible = DisplayObjectVisible(dio);
    short subtype = DisplayFieldSubType(dio);

    if(!visible || go == NULL || !GuiWindowOpen(win))
        return FALSE;
    if (subtype != DAD_VIRTUALTYPE_LIST)
	return FALSE;

    GuiSFWListSetActive(go, num);
    return (bool)(GuiSFWListActive(go) > 0);
}

long GuiGetVirtualFieldListActive(struct DisplayObject *dio)
{
    GuiObject go = DisplayObjectGuiObject(dio);
    struct DisplayObject *windio = DisplayObjectWindow(dio);
    GuiObject win = DisplayObjectGuiObject(windio);
    bool visible = DisplayObjectVisible(dio);
    short subtype = DisplayFieldSubType(dio);

    if(!visible || go == NULL || !GuiWindowOpen(win))
        return 0;
    if (subtype != DAD_VIRTUALTYPE_LIST)
	return 0;

    return GuiSFWListActive(go);
}

void GuiVirtualFieldListActiveChanged(struct DisplayObject *dio)
{
    GuiObject go = DisplayObjectGuiObject(dio);
    struct Table *tab = GuiSFWListControlTable(go);
    struct Record *rec = GuiSFWListActiveRecord(go);
    struct Project *pr = DisplayObjectProject(dio);
    struct Field *field = DisplayFieldField(dio);

    if (DisplayFieldControlTable(dio) && tab != NULL && rec != NULL) {
        GuiShowRecord(tab, rec);
    }
    GuiUpdateDependentDisplayObjects(pr, FieldDisplayObject(field), NULL, NULL);
}

void GuiVirtualFieldListDoubleClick(struct DisplayObject *dio, 
	long line, long col, struct Field *field, struct Record *rec)
{
    short type = FieldType(field);
    struct Table *guess = (type == TYPE_REFERENCE)?
        FieldReferenceTable(field) : FieldTable(field);
    short action = DisplayFieldOnDoubleClick(dio);
    struct Project *pr = DisplayObjectProject(dio);
    struct Table *tab;
    struct DisplayObject *recdio;

    switch (action) {
	case DAD_DOUBLECLICK_DONOTHING:
    	default:
	    break;

	case DAD_DOUBLECLICK_SHOWRECORD:
	    tab = RecordFindTable(rec, guess, pr);
	    GuiShowRecord(tab, rec);
            recdio = FieldDisplayObject(field);
            if (!DisplayObjectVisible(recdio) || FieldTable(field) != tab) {
                recdio = TableDisplayObject(tab);
            }
	    GuiShowDisplayObject(recdio);
	    break;

	case DAD_DOUBLECLICK_CALLTRIGGER:
	{
    	    struct Element elem[3], *pelem[4];
	    
    	    memset(elem, 0, sizeof(elem));
	    elem[0].el_Type = TYPE_INTEGER;
	    elem[0].el_Contents.integer = line;
	    elem[1].el_Type = TYPE_INTEGER;
	    elem[1].el_Contents.integer = col;
	    if (rec) {
		tab = RecordFindTable(rec, guess, pr);
		elem[2].el_Type = TYPE_RECORD;
		elem[2].el_Contents.tabrec.tab = tab;
		elem[2].el_Contents.tabrec.rec = rec;
	    }
	    else {
		elem[2].el_Type = TYPE_NIL;
            }
	    
    	    pelem[0] = &elem[0];
	    pelem[1] = &elem[1];
	    pelem[2] = &elem[2];
	    pelem[3] = NULL;
	    
    	    ExecuteTrigger(pr, DisplayFieldDoubleClickTrigger(dio), pelem,NULL);
	    break;
	}
    }
}

void GuiVirtualFieldListDropUrls(struct DisplayObject *dio, char **urls)
{
    struct Project *pr = DisplayObjectProject(dio);
    struct MemPool *memPool = ProjectMemPool(pr);
    struct Element elem, *pelem[2];
	    
    memset(&elem, 0, sizeof(elem));
    elem.el_Type = TYPE_MEMO;
    elem.el_Contents.mi = LabelsToString(memPool, urls);

    if (elem.el_Contents.mi == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
    } else {
        pelem[0] = &elem;
        pelem[1] = NULL;
	    
        ExecuteTrigger(DisplayObjectProject(dio), 
                DisplayFieldURLDropTrigger(dio), pelem, NULL);

        FreeString(memPool, elem.el_Contents.mi);
    }
}

void GuiVirtualFieldListSortDrop(struct DisplayObject *dio, long row)
{
    GuiObject go = DisplayObjectGuiObject(dio);
    const struct QueryLine *ql = GuiSFWListQueryLine(go, row);

    if (ql != NULL) {
        struct Project *pr = DisplayObjectProject(dio);
        struct Table *tab = GuiSFWListControlTable(go);
        struct Record *rec = GuiSFWListRecord(go, row);
        struct Element elem[3], *pelem[4];
	    
        memset(elem, 0, sizeof(elem));
        elem[0].el_Type = TYPE_INTEGER;
        elem[0].el_Contents.integer = ql->ql_LineNum;
        elem[1].el_Type = TYPE_INTEGER;
        elem[1].el_Contents.integer = row;
        if (rec != NULL) {
            elem[2].el_Type = TYPE_RECORD;
            elem[2].el_Contents.tabrec.tab = RecordFindTable(rec, tab, pr);
            elem[2].el_Contents.tabrec.rec = rec;
        }
        else {
            elem[2].el_Type = TYPE_NIL;
        }
	    
        pelem[0] = &elem[0];
        pelem[1] = &elem[1];
        pelem[2] = &elem[2];
        pelem[3] = NULL;
	    
        ExecuteTrigger(pr, DisplayFieldSortDropTrigger(dio), pelem, NULL);
    }
}

void GuiSetDisplayObjectEnabled(struct DisplayObject *dio, bool enabled)
{
    GuiObject go = NULL;

    switch (DisplayObjectType(dio)) {
        case DO_TYPE_FIELD:
            guisetenabled(dio, enabled);
            break;

        case DO_TYPE_WINDOW:
            go = DisplayWindowButtonObject(dio);
            if (!enabled && DisplayWindowCloseWhenDisabled(dio)) {
                GuiCloseWindow(DisplayObjectGuiObject(dio));
                SetDisplayWindowOpen(dio, FALSE);   /* block opening later */
            }
            GuiSetEnabled(go, enabled);
            break;
    }
}

static void DisplayField(struct Field *field, struct Record *rec,
        bool *askIfError)
{
    struct DataInfo *di, di_buf;
    struct DisplayObject *dio = FieldDisplayObject(field);
    struct Project *pr = DisplayObjectProject(dio);
    GuiObject go = DisplayObjectGuiObject(dio);
    GuiObject external = DisplayFieldExternalObject(dio);
    struct DisplayObject *windio = DisplayObjectWindow(dio);
    GuiObject win = DisplayObjectGuiObject(windio);
    bool readonly = DisplayFieldReadOnly(dio);
    bool visible = DisplayObjectVisible(dio), nil;
    short type = FieldType(field);
    long error = OK;
    bool showError;

    if (!visible || (go == NULL && external == NULL) || !GuiWindowOpen(win)) {
        return;
    }

    RecordLock(rec);
    error = RecordDataInfo(rec, field, &di, &di_buf);
    if (di != NULL) {
        nil = ((di->di_Flags & DIF_NIL) != 0);

        switch(type)
        {
            case TYPE_STRING:
                GuiSetString(go, di->di_Contents.string, readonly, external);
                break;

            case TYPE_INTEGER:
                GuiSetString(go, nil? FieldNilString(field) :
                        IntToStr(di->di_Contents.integer), readonly, external);
                break;

            case TYPE_REAL:
                GuiSetString(go, nil? FieldNilString(field) :
                        DoubleToStr(di->di_Contents.real,
                            DisplayFieldNumDecimals(dio),
                            (short)GetPrefsValue(pr, PREFS_REALFORMAT)),
                        readonly, external);
                break;

            case TYPE_BOOL:
                GuiSetBoolean(go, !nil, readonly);
                break;

            case TYPE_CHOICE:
                GuiSetChoice(go, di->di_Contents.choice,
                    FieldChoiceLabels(field), readonly);
                break;

            case TYPE_DATE:
                GuiSetString(go, nil? FieldNilString(field) :
                        DateToString(di->di_Contents.date,
                            (short)GetPrefsValue(pr, PREFS_DATEFORMAT)),
                        readonly, external);
                break;

            case TYPE_TIME:
                GuiSetString(go, nil? FieldNilString(field) :
                        TimeToString(di->di_Contents.time, 
                            DisplayFieldFormat(dio)),
                        readonly, external);
		break;

            case TYPE_MEMO:
                GuiSetMemo(go, di->di_Contents.mi, readonly);
                break;

            case TYPE_REFERENCE:
            {
                struct MemPool *memPool = ProjectMemPool(pr);
                GuiObject autoshow_go = DisplayFieldAutoShowObject(dio);
                GuiObject filter_go = DisplayFieldFilterObject(dio);
                struct Record *refrec = di->di_Contents.rec;
                char *s = NULL;
                bool free = FALSE;

                if(DisplayFieldAutoShow(dio) && autoshow_go != NULL)
                    GuiSetSelected(autoshow_go, 
				   FieldReferenceAutoShow(field));
                if(DisplayFieldFilter(dio) && filter_go != NULL)
                    GuiSetSelected(filter_go, FieldReferenceFilter(field));

                RecordLock(refrec);
                error = GuiFieldReferenceDataString(field, refrec, memPool, 
                        &s, &free);
                RecordUnlock(refrec);
                if(error == OK)
                    GuiSetString(go, s, TRUE, NULL);
                if(free)
                    FreeString(memPool, s);
                break;
            }

            case TYPE_VIRTUAL:
                /* updated automatically or by explicit access */
                break;

            case TYPE_BUTTON:
                break;
        }
    }
    if (error) {
        showError = guiCheckError(error, askIfError);
        guiseterror(field, showError? GES(error) : "");
    }
    RecordUnlock(rec);
}

static void ClearField(struct Field *field)
{
    struct DisplayObject *dio = FieldDisplayObject(field);
    GuiObject go = DisplayObjectGuiObject(dio);
    GuiObject external = DisplayFieldExternalObject(dio);
    bool readonly = DisplayFieldReadOnly(dio);
    bool visible = DisplayObjectVisible(dio);
    short type = FieldType(field);

    if (!visible || (go == NULL && external == NULL)) {
        return;
    }

    switch(type)
    {
        case TYPE_STRING:
        case TYPE_INTEGER:
        case TYPE_REAL:
        case TYPE_DATE:
        case TYPE_TIME:
        case TYPE_REFERENCE:
            GuiSetString(go, "", readonly, external);
            break;

        case TYPE_BOOL:
            GuiSetBoolean(go, FALSE, readonly);
            break;

        case TYPE_CHOICE:
        {
            static char *labels[2] = { "", NULL };

            GuiSetChoice(go, 0, labels, readonly);
            break;
        }

        case TYPE_MEMO:
            GuiSetMemo(go, "", readonly);
	    break;

        case TYPE_VIRTUAL:
            switch (DisplayFieldSubType(dio))
            {
                case DAD_VIRTUALTYPE_BOOL:
                    GuiSetBoolean(go, FALSE, readonly);
                    break;

                case DAD_VIRTUALTYPE_TEXT:
                    GuiSetString(go, "", readonly, NULL);
                    break;

                case DAD_VIRTUALTYPE_LIST:
                    GuiSetSFWList(go, NULL);
                    break;
            }
            break;

        case TYPE_BUTTON:
            break;
    }
}

void GuiDisplayPanel(struct Table *tab)
{
    struct Record *rec = TableGuiRecord(tab);
    struct DisplayObject *dio = TableDisplayObject(tab);
    struct DisplayObject *windio = DisplayObjectWindow(dio);
    GuiObject win = DisplayObjectGuiObject(windio);
    GuiObject go;
    bool gadgets = !TableExactlyOneRecord(tab);
    bool visible = DisplayObjectVisible(dio) && DisplayTableHasPanel(dio);

    if (visible && gadgets && GuiWindowOpen(win)) {
        go = DisplayTablePanelNumAllObject(dio);
        if (DisplayTablePanelNumAll(dio) && go) {
            ulong num, all;

            RecordNumAll(tab, rec, TRUE, &num, &all);
            GuiSetNumAll(go, num, all);
        }

        go = DisplayTablePanelFilterObject(dio);
        if (DisplayTablePanelFilter(dio) && go) {
            GuiSetSelected(go, TableFilterActive(tab));
        }
    }
}

static bool checkComputeFunctionExists(struct DisplayObject *dio, 
        const char *compute)
{
    struct Project *pr = DisplayObjectProject(dio);
    struct Program *prg = ProjectProgram(pr);
    struct UserFunction *uf;

    if (compute == NULL || compute[0] == 0) {  /* no error */
        return FALSE;
    }

    uf = FindUserFunctionName(prg, compute);
    if (uf == NULL) {
        Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Err_Trigger), compute);
        return FALSE;                   /* error */
    }

    return TRUE;
}

static void UpdateEnabledStaticSingle(struct DisplayObject *dio)
{
    struct Table *tab = DisplayObjectTable(dio);
    struct Field *field = DisplayFieldField(dio);
    char enabled = DisplayObjectEnabled(dio);

    switch (enabled) {
        case DOD_ENABLED_ALWAYS:
            GuiSetDisplayObjectEnabled(dio, TRUE);
            break;
        case DOD_ENABLED_NOTINITIAL:
            GuiSetDisplayObjectEnabled(dio, !tab || TableGuiRecord(tab));
            break;
        case DOD_ENABLED_COMPUTED:
            /* check that the compute function exists */
            if (!checkComputeFunctionExists(dio, 
                        DisplayObjectEnabledCompute(dio))) {
                GuiSetDisplayObjectEnabled(dio, FALSE);
            }
            break;
    }
    if (field && FieldType(field) == TYPE_VIRTUAL) {
        /* check that the compute function exists */
        checkComputeFunctionExists(dio, FieldVirtualCompute(field));
    }
}

static void UpdateEnabledStatic(struct DisplayObject *dio,
        struct DisplayObject *windio)
{
    struct List *childs = DisplayObjectObjects(dio);

    if (GuiDisplayObjectVisible(dio) &&
            (windio == NULL || DisplayObjectWindow(dio) == windio)) {
        UpdateEnabledStaticSingle(dio);
    }

    if (childs) {
        FORNODE (dio, childs) {
            UpdateEnabledStatic(dio, windio);
        }
    }
}

bool GuiShowDisplayObject(struct DisplayObject *dio)
{
    struct DisplayObject *windio = DisplayObjectWindow(dio), *parent, *child;
    GuiObject go = DisplayObjectGuiObject(dio);
    GuiObject win = DisplayObjectGuiObject(windio);

    if (!DisplayObjectVisible(dio) || go == NULL || win == NULL) {
        return FALSE;
    }

    if (!GuiWindowOpen(win) && DisplayWindowButton(windio) && 
        DisplayWindowCloseWhenDisabled(windio)) {
        GuiObject button = DisplayWindowButtonObject(windio);

        /* 
        ** We need to check the enabled state of the window button.
        ** Make sure the enabled state of the button is up-to-date.
        */
        if (!GuiDisplayObjectVisible(windio)) {
            UpdateEnabledStaticSingle(windio);
            if (DisplayObjectEnabled(windio) == DOD_ENABLED_COMPUTED) {
                /* setting the open state allows recomputation of enabled */
                SetDisplayWindowOpen(windio, TRUE);
                if (!GuiUpdateComputedDisplayObject(windio)) {
                    return FALSE;
                }
            }
        }
        if (!GuiGetEnabled(button)) {
            return FALSE;               /* window button is disabled */
        }
    }

    for (; dio != NULL && dio != windio; dio = parent) {
        struct List *childs;
        long pos = 0;

        parent = DisplayObjectParent(dio);
        childs = DisplayObjectObjects(parent);
        if (DisplayObjectType(parent) == DO_TYPE_REGISTER && childs != NULL) {
            FORNODE (child, childs) {
                if (child == dio) {
                    break;
                }
                if (DisplayObjectVisible(child)) {
                    pos++;
                }
            }
            go = DisplayObjectGuiObject(parent);
            GuiSetRegisterPage(go, pos);
        }
    }

    if (!GuiOpenWindow(win)) {
        return FALSE;
    }

    GuiWindowToFront(win);  /* can be annoying but seems to be useful. */
    return TRUE;
}

static void DisplayRecord(struct Table *tab, bool *askIfError)
{
    struct Project *pr = TableProject(tab);
    struct DisplayObject *dio = TableDisplayObject(tab);
    struct Record *rec = TableGuiRecord(tab);
    struct List *fields = TableFields(tab);
    struct Field *field;

    GuiDisplayPanel(tab);

    /* 
    ** First we update the enabled/disabled state
    ** and then the field contents.
    ** This is because the latter might disable the field
    ** in case of an error
    */
    UpdateEnabledStatic(dio, NULL);
    guiUpdateDependentDisplayObjectsInternal(pr, TableDisplayObject(tab),
            NULL, NULL, askIfError);

    if (fields) {
        FORNODE(field, fields) {
            DisplayField(field, rec, askIfError);
        }
    }
}

static void GuiShowRecordRec(struct Table *cur, struct Record *show, 
			     bool force, bool autoshow, bool *askIfError)
{
    struct Project *pr = TableProject(cur);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    struct List *fields = TableFields(cur);
    struct Field *field;
    struct DataInfo *di, di_buf;
    struct Table *reftab;
    struct Record *refrec = NULL;
    struct DisplayObject *dio;
    GuiObject go;
    long error = OK;

    if (cur == NULL || TableGuiFlag(cur)) {     /* || table already accessed? */
	return;
    }

    SetTableGuiFlag(cur, TRUE);                 /* to avoid endless loops */

    if (!force && TableGuiRecord(cur) == show) {
	return;
    }

    if (show != NULL && !RecordMatchFilter(show) && TableFilterActive(cur)) {
	FilterOnOff(cur, FALSE);
    }
    SetTableGuiRecord(cur, show);
    DisplayRecord(cur, askIfError);

    if (autoshow && show != NULL && fields != NULL) {
        RecordLock(show);
        FORNODE (field, fields) {
            if (FieldType(field) == TYPE_REFERENCE && 
                        FieldReferenceAutoShow(field)) {
                reftab = FieldReferenceTable(field);
                error = RecordDataInfo(show, field, &di, &di_buf);
                guiCheckError(error, askIfError);
                if (error == OK && di != NULL) {
                    refrec = di->di_Contents.rec;
                    GuiShowRecordRec(reftab, refrec, FALSE, TRUE, askIfError);
                }
            }
        }
        RecordUnlock(show);
    }

    for (dio=ProjectRootDisplayObject(pr); dio; dio = NextDisplayObject(dio)) {
        if (DisplayFieldControlTable(dio) &&
                FieldType(DisplayFieldField(dio)) == TYPE_VIRTUAL &&
                DisplayFieldSubType(dio) == DAD_VIRTUALTYPE_LIST) {
            go = DisplayObjectGuiObject(dio);
            if (GuiSFWListControlTable(go) == cur) {
                GuiSFWListSetActiveRecord(go, show);
            }
        }
    }

    if (tabs) {
	FORNODE (tab, tabs) {
	    if (TableGuiFlag(tab)) {
		continue;                       /* avoid unnecessary work */
            }

	    fields = TableFields(tab);
	    if (fields) {
		FORNODE (field, fields) {
		    struct Record *rec;
		    
		    if (FieldType(field) != TYPE_REFERENCE) {
			continue;
                    }
		    if (!FieldReferenceFilter(field)) {
			continue;
                    }
		    reftab = FieldReferenceTable(field);
		    if (reftab != cur) {
			continue;
                    }
		    rec = TableGuiRecord(tab);
		    RecordLock(rec);
		    error = RecordDataInfo(rec, field, &di, &di_buf);
		    if (di != NULL) {
			refrec = di->di_Contents.rec;
                    } else {
			refrec = show;
                    }
		    RecordUnlock(rec);
                    guiCheckError(error, askIfError);
		    if (refrec == show) {
			break;          /* current gui record is okay */
                    }

		    for (rec = FirstRecord(tab, TRUE); rec != NULL; 
                            rec = NextRecord(rec, TRUE)) {
			if (GuiCheckRecord(tab, rec)) {
			    break;      /* found match */
			}
		    }
		    GuiShowRecordRec(tab, rec, FALSE, autoshow, askIfError);
                    /* Only one record can be shown per table, so stop here */
                    break;
		}
            }
	}
    }
}

static void GuiShowRecordAutoShow(struct Table *cur, struct Record *rec, 
        bool autoshow)
{
    struct Project *pr = TableProject(cur);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    bool askIfError = TRUE;
    static long recursive = 0;

    if (recursive == 0 && tabs != NULL) {
        FORNODE(tab, tabs) {
            SetTableGuiFlag(tab, FALSE);
        }
    }

    recursive++;
    GuiShowRecordRec(cur, rec, TRUE, autoshow, &askIfError);
    recursive--;
}

void GuiShowRecord(struct Table *tab, struct Record *rec)
{
    /*
    ** We create a GUI program context such that virtual fields are all
    ** updated once at the end.  This is useful when other records also need
    ** to be updated.
    */
    struct Project *pr = TableProject(tab);
    bool context = BeginGuiProgramContext(pr);

    GuiShowRecordAutoShow(tab, rec, TRUE);
    EndGuiProgramContext(pr, context);
}

void GuiShowReferenceRecord(struct Field *field, struct Record *rec)
{
    struct Table *reftab = FieldReferenceTable(field);
    struct Record *refrec = NULL;
    struct DataInfo *di = NULL, di_buf;
    long error;

    if(field == NULL || FieldType(field) != TYPE_REFERENCE)
        return;

    RecordLock(rec);
    error = RecordDataInfo(rec, field, &di, &di_buf);
    if(di != NULL)
        refrec = di->di_Contents.rec;
    RecordUnlock(rec);

    if(error)
    {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    GuiShowRecord(reftab, refrec);
    GuiShowDisplayObject(TableDisplayObject(reftab));
}

bool GuiCheckRecord(struct Table *tab, struct Record *rec)
{
    struct List *fields = TableFields(tab);
    struct Field *field;
    struct DataInfo *di, di_buf;
    struct Table *reftab;
    struct Record *refrec;
    bool result = TRUE;

    if(!RecordMatchFilter(rec))
        return(FALSE);

    RecordLock(rec);
    if(fields)
        FORNODE(field, fields)
        {
            if(FieldType(field) != TYPE_REFERENCE)
                continue;
            if(!FieldReferenceFilter(field))
                continue;

            reftab = FieldReferenceTable(field);
            RecordDataInfo(rec, field, &di, &di_buf);
            if(di == NULL)
                continue;               /* treat errors as match success */
            refrec = di->di_Contents.rec;
            if(TableGuiRecord(reftab) != refrec)
            {
                result = FALSE;         /* match failed */
                break;
            }
        }
    RecordUnlock(rec);
    return(result);
}

static void GuiCheckGuiRecordAutoShow(struct Table *tab, bool autoshow)
{
    struct Record *rec = TableGuiRecord(tab), *new;

    if (tab == NULL) {
        return;
    }

    if (rec == NULL) {
        rec = FirstRecord(tab, TRUE);
    }

    new = rec;
    while (new && !GuiCheckRecord(tab, new)) {
        new = NextRecord(new, TRUE);
    }
    if (new == NULL) {
        new = rec;
        while (new && !GuiCheckRecord(tab, new)) {
            new = PrevRecord(new, TRUE);
        }
    }

    GuiShowRecordAutoShow(tab, new, autoshow);
}

void GuiCheckGuiRecord(struct Table *tab)
{
    GuiCheckGuiRecordAutoShow(tab, TRUE);
}

static void GuiUpdateReferenceFields(struct Table *update, struct Record *rec)
{
    struct Project *pr = TableProject(update);
    struct List *tabs = ProjectTables(pr), *fields;
    struct Table *tab;
    struct Field *field;
    struct Record *guirec, *refrec;
    struct DataInfo *di = NULL, di_buf;
    bool askIfError = TRUE;

    if (tabs) {
        FORNODE (tab, tabs) {
            fields = TableFields(tab);
            if (fields == NULL) {
                continue;
            }

            FORNODE (field, fields) {
                if (FieldType(field) != TYPE_REFERENCE) {
                    continue;
                }
                if (FieldReferenceTable(field) != update) {
                    continue;
                }

                guirec = TableGuiRecord(tab);
                refrec = NULL;
                RecordLock(guirec);
                if (RecordDataInfo(guirec, field, &di, &di_buf) == OK) {
                    refrec = di->di_Contents.rec;
                }
                RecordUnlock(guirec);
                if (refrec == NULL || refrec == rec) {
                    DisplayField(field, guirec, &askIfError);
                }
            }
        }
    }
}

void GuiAddRecord(struct Table *tab, struct Record *rec)
{
    struct Record *gui_rec = TableGuiRecord(tab);

    if(gui_rec == NULL && GuiCheckRecord(tab, rec))
        GuiShowRecord(tab, rec);
    else
        GuiDisplayPanel(tab);

    BuildGuiAddRecord(tab, rec);
}

static struct Record *findRecordBackward(struct Table *tab, struct Record *rec)
{
    if (tab && rec) {
        for (; rec; rec = PrevRecord(rec, TRUE)) {
            if (GuiCheckRecord(tab, rec)) {
                break;
            }
        }
    }
    return rec;
}

static struct Record *findRecordForward(struct Table *tab, struct Record *rec)
{
    if (tab && rec) {
        for (; rec; rec = NextRecord(rec, TRUE)) {
            if (GuiCheckRecord(tab, rec)) {
                break;
            }
        }
    }
    return rec;
}

void GuiRemRecord(struct Table *tab, struct Record *rec, struct Record *next,
        struct Record *prev, bool update)
{
    struct Record *gui_rec = TableGuiRecord(tab);

    if (rec == gui_rec) {
        gui_rec = findRecordForward(tab, next);
        if (gui_rec == NULL) {
            gui_rec = findRecordBackward(tab, prev);
        }
	SetTableGuiRecord(tab, gui_rec);
	if (update) {
	    GuiShowRecord(tab, gui_rec);
        }
    }

    BuildGuiRemRecord(tab, rec);
    QueryEditorRemRecord(tab, rec);
    GuiUpdateReferenceFields(tab, rec);
}

void GuiRemAllRecords(struct Project *pr)
{
    BuildGuiRemAllRecords(pr);
    QueryEditorRemAllRecords(pr);
}

void GuiPrevRecord(struct Table *tab)
{
    struct Record *rec = TableGuiRecord(tab);

    rec = PrevRecord(rec, TRUE);
    rec = findRecordBackward(tab, rec);
    if (rec != NULL) {
        GuiShowRecord(tab, rec);
    }
}

void GuiNextRecord(struct Table *tab)
{
    struct Record *rec = TableGuiRecord(tab);

    if (rec == NULL) {
        GuiFirstRecord(tab);
    } else {
        rec = NextRecord(rec, TRUE);
        rec = findRecordForward(tab, rec);
        if (rec != NULL) {
            GuiShowRecord(tab, rec);
        }
    }
}

void GuiFirstRecord(struct Table *tab)
{
    struct Record *rec = FirstRecord(tab, TRUE);

    rec = findRecordForward(tab, rec);
    if (rec != NULL) {
        GuiShowRecord(tab, rec);
    }
}

void GuiLastRecord(struct Table *tab)
{
    struct Record *rec = LastRecord(tab, TRUE);

    rec = findRecordBackward(tab, rec);
    if (rec != NULL) {
        GuiShowRecord(tab, rec);
    }
}

void GuiJumpBackwardRecord(struct Table *tab)
{
    struct Record *rec = TableGuiRecord(tab);
    ulong all = 0, step, i;

    RecordNumAll(tab, NULL, TRUE, NULL, &all);
    step = all / 30 + 1;
    for (i=0; i < step; i++) {
        rec = PrevRecord(rec, TRUE);
    }

    rec = findRecordBackward(tab, rec);
    if (rec == NULL) {
        GuiFirstRecord(tab);
    } else {
        GuiShowRecord(tab, rec);
    }
}

void GuiJumpForwardRecord(struct Table *tab)
{
    struct Record *rec = TableGuiRecord(tab);
    ulong all = 0, step, i;

    if (rec == NULL) {
        rec = FirstRecord(tab, TRUE);
    }

    RecordNumAll(tab, NULL, TRUE, NULL, &all);
    step = all / 30 + 1;
    for (i=0; i < step; i++) {
        rec = NextRecord(rec, TRUE);
    }

    rec = findRecordForward(tab, rec);
    if (rec == NULL) {
        GuiLastRecord(tab);
    } else {
        GuiShowRecord(tab, rec);
    }
}

void GuiJumpToRecord(struct Table *tab, ulong num)
{
    struct Record *rec = NULL;
    bool filter = TRUE;

    if (num > 0) {
        for (rec = FirstRecord(tab, filter); rec != NULL;
                rec = NextRecord(rec, filter)) {
            if (--num == 0) {
                break;
            }
        }

        if (rec == NULL) {
            return;
        }
    }
    GuiShowRecord(tab, rec);
}

void GuiUpdateField(struct Field *field, struct Record *rec)
{
    struct Table *tab = FieldTable(field);
    bool askIfError = TRUE;

    if (rec == TableGuiRecord(tab)) {
        DisplayField(field, rec, &askIfError);
    }
    GuiUpdateReferenceFields(tab, rec);
}

static bool acceptInRegister(struct DisplayObject *dio, void *regdio)
{
    if (DisplayObjectWindow(dio) != DisplayObjectWindow(regdio)) {
        return FALSE;
    }
    while ((dio = DisplayObjectParent(dio)) != NULL) {
        if (dio == regdio) {
            return TRUE;
        }
    }
    return FALSE;
}

void GuiUpdateRegister(GuiObject reg, struct Project *pr)
{
    struct DisplayObject *regdio = GuiObjectDisplayObject(reg);
    struct DisplayObject *windio = DisplayObjectWindow(regdio);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    UpdateEnabledStatic(regdio, windio);
    GuiUpdateDependentDisplayObjects(pr, pr, acceptInRegister, regdio);

    GuiSetSleep(TRUE);
    if (tabs) {
        FORNODE(tab, tabs) {
            struct List *fields = TableFields(tab);
            struct Field *field;
            struct DisplayObject *dio;

            if (fields) {
                FORNODE(field, fields) {
		    if (FieldType(field) != TYPE_VIRTUAL) {
			continue;
		    }
		    dio = FieldDisplayObject(field);
                    if (DisplayObjectWindow(dio) != windio) {
                        continue;
                    }
                    if (!DisplayObjectIsAncestor(dio, regdio)) {
			continue;
                    }
		    if (!GuiDisplayObjectVisible(dio)) {
                        ClearField(field);
		    }
		}
	    }
	}
    }
    GuiSetSleep(FALSE);
}

static bool acceptInWindow(struct DisplayObject *dio, void *windio)
{
    return (DisplayObjectWindow(dio) == windio);
}

void GuiUpdateWindow(GuiObject win, struct Project *pr)
{
    struct DisplayObject *windio = GuiObjectDisplayObject(win);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    bool askIfError = TRUE;

    UpdateEnabledStatic(windio, windio);
    GuiUpdateDependentDisplayObjects(pr, pr, acceptInWindow, windio);

    GuiSetSleep(TRUE);
    if (tabs) {
        FORNODE(tab, tabs) {
            struct DisplayObject *dio = TableDisplayObject(tab);
            struct Record *rec = TableGuiRecord(tab);
            struct List *fields = TableFields(tab);
            struct Field *field;

            if (DisplayObjectWindow(dio) == windio) {
                GuiDisplayPanel(tab);
            }

            if (fields) {
                FORNODE(field, fields) {
                    dio = FieldDisplayObject(field);
                    if (DisplayObjectWindow(dio) == windio) {
                        DisplayField(field, rec, &askIfError);
                    }
                }
            }
        }
    }
    GuiSetSleep(FALSE);
}

void GuiClearWindow(GuiObject win, struct Project *pr)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    if (tabs) {
        FORNODE(tab, tabs) {
            struct DisplayObject *dio, *windio;
            struct List *fields = TableFields(tab);
            struct Field *field;

            if (fields) {
                FORNODE(field, fields) {
                    dio = FieldDisplayObject(field);
                    windio = DisplayObjectWindow(dio);
                    if (DisplayObjectGuiObject(windio) == win) {
                        ClearField(field);
                    }
                }
            }
        }
    }
}

void GuiUpdate(struct Project *pr)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    if (tabs) {
        FORNODE(tab, tabs) {
            GuiCheckGuiRecordAutoShow(tab, FALSE);
        }
    }
    // update all virtual fields
    GuiUpdateDependentDisplayObjects(pr, pr, NULL, NULL);
}

union StoreData {
    bool b;
    long n;
    const char *str;
    struct Record *rec;
};

static void guiStoreDataHelp(struct DisplayObject *dio, union StoreData sd)
{
    struct Field *field = DisplayFieldField(dio);
    struct Table *tab = FieldTable(field);
    struct Record *rec = TableGuiRecord(tab);
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = ProjectMemPool(pr);
    short type = FieldType(field);
    const char *trigger = FieldTrigger(field);
    long error = OK;
    union Data *data = NULL;
    union Data storage;
    char *mem = NULL;
    bool nil = FALSE;
    struct Element elem, *args[2];

    if (pr == NULL) {
        return;
    }

    elem.el_Type = TYPE_NIL;
    elem.el_Allocated = FALSE;
    elem.el_Record = rec;

    switch (type) {
        case TYPE_INTEGER:
            nil = !StrToInt(sd.str, &storage.integer);
            data = &storage;
            elem.el_Type = TYPE_INTEGER;
            elem.el_Contents.integer = storage.integer;
            break;

        case TYPE_REAL:
            nil = !StrToDouble(sd.str, &storage.real);
            data = &storage;
            elem.el_Type = TYPE_REAL;
            elem.el_Contents.real = storage.real;
            break;

        case TYPE_BOOL:
            nil = !sd.b;
            elem.el_Type = TYPE_BOOL;
            data = &storage;
            break;

        case TYPE_STRING:
            mem = AllocString(memPool, sd.str);
            if (mem == NULL) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
                return;
            }
	    storage.string = mem;
            data = &storage;
            elem.el_Type = TYPE_STRING;
            elem.el_Contents.string.string = mem;
            elem.el_Contents.string.field = NULL;
            break;

        case TYPE_MEMO:
            mem = AllocString(memPool, sd.str);
            if (mem == NULL) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
                return;
            }
	    storage.mi = mem;
            data = &storage;
            elem.el_Type = TYPE_MEMO;
            elem.el_Contents.mi = mem;
            break;

        case TYPE_DATE:
            nil = !StringToDate(sd.str, &storage.date);
            data = &storage;
            elem.el_Type = TYPE_DATE;
            elem.el_Contents.date = storage.date;
            break;

        case TYPE_TIME:
            nil = !StringToTime(sd.str, DisplayFieldFormat(dio), &storage.time);
            data = &storage;
            elem.el_Type = TYPE_TIME;
            elem.el_Contents.time = storage.time;
            break;

        case TYPE_CHOICE:
            storage.choice = sd.n;
            data = &storage;
            elem.el_Type = TYPE_INTEGER;
            elem.el_Contents.integer = storage.choice;
            break;

        case TYPE_BUTTON:
            elem.el_Type = TYPE_NIL;
            break;

        case TYPE_REFERENCE:
            storage.rec = sd.rec;
            if (storage.rec == NULL) {
                nil = TRUE;
            }
            data = &storage;
            elem.el_Type = TYPE_RECORD;
            elem.el_Contents.tabrec.tab = FieldReferenceTable(field);
            elem.el_Contents.tabrec.rec = storage.rec;
            break;
    }

    if (trigger != NULL && trigger[0] != 0) {
        if (nil) {
            elem.el_Type = TYPE_NIL;
        }
        args[0] = &elem;
        args[1] = NULL;
        ExecuteTrigger(pr, trigger, args, NULL);
    } else if (data) {
        bool countChanges = FieldCountChanges(field) && 
            TableCountChanges(tab);
        if (countChanges) {
            error = ObtainProjectChangeLock(pr);
            if (error != OK) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            }
        }
        if (error == OK) {
            error = RecordStoreData(rec, field, data, nil);
            if (error) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            } else if (countChanges) {
                ProjectChange(pr, FALSE);
            }

            error = ReorderRecords(tab);
            if (error) {
                Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Err_Order), GES(error));
            }
        }

	GuiUpdateDependentDisplayObjects(pr, field, NULL, NULL);

        GuiDisplayPanel(tab);
    }
    GuiUpdateField(field, rec);      /* in case trigger function fails. */

    FreeString(memPool, mem);
}

void GuiStoreDataBool(struct DisplayObject *dio, bool b)
{
    union StoreData sd;

    sd.b = b;
    guiStoreDataHelp(dio, sd);
}

void GuiStoreDataLong(struct DisplayObject *dio, long n)
{
    union StoreData sd;

    sd.n = n;
    guiStoreDataHelp(dio, sd);
}

void GuiStoreDataStr(struct DisplayObject *dio, const char *str)
{
    union StoreData sd;

    sd.str = str;
    guiStoreDataHelp(dio, sd);
}

void GuiStoreDataRecord(struct DisplayObject *dio, struct Record *rec)
{
    union StoreData sd;

    sd.rec = rec;
    guiStoreDataHelp(dio, sd);
}

void GuiStoreData(struct DisplayObject *dio)
{
    union StoreData sd;

    sd.n = 0;
    guiStoreDataHelp(dio, sd);
}

void ReferenceFilterOnOff(struct Field *field, bool active)
{
    struct Table *tab = FieldTable(field);

    SetFieldReferenceFilter(field, active);
    GuiCheckGuiRecord(tab);
    GuiDisplayPanel(tab);          /* just in case we add info to panel */
}

void ReferenceAutoShowOnOff(struct Field *field, bool active)
{
    struct Table *tab = FieldTable(field);

    SetFieldReferenceAutoShow(field, active);
    GuiCheckGuiRecord(tab);
}

