/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gui.h
**
** Creation date: 19.03.1995
** Last modified: 10.02.2022
*/

#ifndef GUI_H
#define GUI_H

#include "system.h"

struct DisplayObject;
struct Table;
struct Field;
struct Record;
struct Element;

#ifdef __cplusplus
extern "C" {
#endif

long GuiFieldReferenceDataString(struct Field *field, struct Record *refrec,
        struct MemPool *memPool, char **ptext, bool *pfree);
/*
** Converts value of given reference field in the referenced record
** (refrec) to string representation. Record refrec must have been locked.
** Returns error code or OK on success, *text will point to the result string.
** memPool is used to allocate memory for the string if necessary. If so
** then *pfree will hold a value of TRUE, FALSE otherwise.
** Virtual fields are not supported.
*/

long GuiFieldDataString(struct Field *field, struct Record *rec,
        struct MemPool *memPool, char **ptext, bool *pfree);
/*
** Converts value of given field in the given record (rec) to string
** representation. Record rec must have been locked.
** Returns error code or OK on success, *text will point to the result string.
** memPool is used to allocate memory for the string if necessary. If so
** then *pfree will hold a value of TRUE, FALSE otherwise.
** Virtual fields are not supported.
*/

void GuiActivateDisplayObject(struct DisplayObject *dio);
/*
** Sets the input focus onto the guiobject of the display object.
*/

bool GuiShowDisplayObject(struct DisplayObject *dio);
/*
** makes the guiobject of the dio visible by opening correspondig
** window, bringing it to front and setting right pages in
** register groups.
*/

bool GuiDisplayObjectVisible(struct DisplayObject *dio);
/*
** Returns TRUE if the guiobject of dio is currently
** visible in an open window.
*/ 

bool GuiCanLoadImageSpec(struct Project *pr, const char *spec);
/*
** Returns FALSE if the given image spec refers to an external file that can't
** be accessed.  Otherwise returns TRUE.
*/

bool BeginGuiProgramContext(struct Project *pr);
/*
** Begins a GUI context for executing programs.  Returns the context.
*/

bool EndGuiProgramContext(struct Project *pr, bool context);
/*
** Ends a GUI context for executing programs.  The context must have been
** started with BeginProgramContext.  Returns TRUE if there were no errors.
*/

bool ExecuteTrigger(struct Project *pr, const char *trigger, 
	struct Element **args, const char *(*func)(struct Element *elem));
/*
** calls trigger function.  Returns FALSE on failure.
*/

bool GuiUpdateDependentDisplayObjects(struct Project *pr, void *dep,
        bool (*accept)(struct DisplayObject *dio, void *ptr), void *ptr);
/*
** Updates all display objects that depend on dep and pass
** an acceptance test (all if accept == NULL).
** Returns TRUE on success.
*/

bool GuiUpdateComputedDisplayObject(struct DisplayObject *dio);
/*
** Updates all computed values of given display object.
*/

const char *GuiSetVirtualField(struct Field *field, struct Record *rec, 
        struct Element *elem);
/*
** Displays the contents of elem in the virtual field field.
** Returns NULL on success or an error string.
*/

bool GuiSetVirtualFieldListActive(struct DisplayObject *dio, long num);
/*
** Activates the num-th line of a virtual field that 
** uses the list kind for display.  Returns TRUE on success (the line
** could be set) or FALSE if field is not a virtual field using list kind,
** the display object is not visible, or if num is out of range 
** (this will deactivate any active line, e.g. use num=-1).
*/

long GuiGetVirtualFieldListActive(struct DisplayObject *dio);
/*
** Returns the active line of a virtual field that uses the list kind
** for display, or -1 if field is not a virtual field using the list type,
** the display object is not visible, or if there is no line activated.
*/

void GuiVirtualFieldListActiveChanged(struct DisplayObject *dio);
/*
** Should be called when the active item in a virtual list is changed.
** This marks the field as being changed and triggers the update
** of dependent objects
*/

void GuiVirtualFieldListDoubleClick(struct DisplayObject *dio,
	long line, long col, struct Field *field, struct Record *rec);
/*
** Performs the required action when the user double clicks an item in
** the virtual list. Parameters field and rec must correspond to the
** clicked item on the given line and column col.
*/

void GuiVirtualFieldListDropUrls(struct DisplayObject *dio, char **urls);
/*
** Calls the trigger function associated with an URL drop event.
*/

void GuiVirtualFieldListSortDrop(struct DisplayObject *dio, long row);
/*
** Should be called when an item in a virtual list has been moved to
** a new row using drag and drop.  The first row starts at 1.
*/ 

void GuiSetDisplayObjectEnabled(struct DisplayObject *dio, bool enabled);
/*
** Sets the enabled/disabled state of the display object.
** This function is called after computing the enabled/disabled
** state of an object with a trigger function.
*/

void GuiShowRecord(struct Table *tab, struct Record *rec);
/*
** Sets a new GuiRecord.
** Also sets the GuiRecords of other tables that
** have a reference filter set to this table or 
** are referenced with an active auto-show setting.
*/

void GuiShowReferenceRecord(struct Field *field, struct Record *rec);
/*
** Shows the record which is referenced from field in rec.
*/

bool GuiCheckRecord(struct Table *tab, struct Record *rec);
/*
** Checks if the given record of the given table
** matches the current filter and the reference filter settings.
*/

void GuiCheckGuiRecord(struct Table *tab);
/*
** Checks if the current GuiRecord of the given Table
** matches the current filter and the reference filter settings.
** If not a new record for the GuiRecord is searched for and displayed.
*/

void GuiAddRecord(struct Table *tab, struct Record *rec);
void GuiRemRecord(struct Table *tab, struct Record *rec, struct Record *next, struct Record *prev, bool update);
void GuiRemAllRecords(struct Project *pr);
/*
** Functions that must be called when a record has been allocated or
** is to be deleted.
** Refreshes panels and sets new GuiRecord if necessary.
*/

void GuiPrevRecord(struct Table *tab);
void GuiNextRecord(struct Table *tab);
void GuiFirstRecord(struct Table *tab);
void GuiLastRecord(struct Table *tab);
void GuiJumpBackwardRecord(struct Table *tab);
void GuiJumpForwardRecord(struct Table *tab);
void GuiJumpToRecord(struct Table *tab, ulong num);
/*
** Functions that must be called when the user wants to
** switch to the next or previous record, e.g. by clicking
** the '<', '>' buttons.
** Sets a new GuiRecord if available and calls GuiShowRecord.
*/

void GuiUpdateField(struct Field *field, struct Record *rec);
/*
** should be called after an field has been changed,
** e.g. by the programming interface.
** Note that this function does not recompute virtual fields.
*/

void GuiUpdateFieldDisabled(struct Field *field, struct Record *rec);
/*
** should be called after the disabled state of an field has been changed,
** e.g. by the programming interface.
*/

void GuiDisplayPanel(struct Table *tab);
/*
** Refreshes the panel display of the given Table.
** This function should be called, e.g. when the order position
** of the GuiRecord may have changed.
*/

void GuiUpdateRegister(GuiObject reg, struct Project *pr);
/*
** Should be called after the active page in a register object changed
** such that all virtual fields in the new page can be recomputed
** and the ones in hidden pages are cleared.
*/

void GuiUpdateWindow(GuiObject win, struct Project *pr);
/*
** Should be called after opening a window for setting
** the contents of all fields.
*/

void GuiClearWindow(GuiObject win, struct Project *pr);
/*
** Should be called after closing a window for resetting
** all fields and freeing memory.
*/

void GuiUpdate(struct Project *pr);
/*
** Should be called for a full refresh, e.g.
** after building the user interface.
*/

void GuiStoreDataBool(struct DisplayObject *dio, bool b);
void GuiStoreDataLong(struct DisplayObject *dio, long n);
void GuiStoreDataStr(struct DisplayObject *dio, const char *str);
void GuiStoreDataRecord(struct DisplayObject *dio, struct Record *rec);
void GuiStoreData(struct DisplayObject *dio);
/*
** should be called from the gui-specific module to store the user's edits.
*/

void ReferenceFilterOnOff(struct Field *field, bool active);
/*
** should be called when the reference filter gadget
** changes its state.
*/

void ReferenceAutoShowOnOff(struct Field *field, bool active);
/*
** should be called when the autoshow gadget
** changes its state.
*/

#ifdef __cplusplus
}
#endif


#endif

