/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** guidosutil.c
**
** help routines for dos with user interaction
**
** Creation date: 03.10.2006
** Last modified: 01.08.2017
*/

#include "guidosutil.h"
#include "dosutil.h"
#include "break.h"
#include "preferences.h"
#include "strhandle.h"
#include "requesters.h"
#include "util.h"
#include "appstrings_cat.h"

bool GuiSystem(const char *cmd, bool bg, bool useShell, int *prc)
{
    struct DosProcess *process;
    bool success = TRUE;
    int rc = -1;
    bool setSleep = (GuiSleep() == 0);

    if (!cmd) {
	return FALSE;
    }

    if (setSleep) {
        GuiSetSleep(TRUE);
    }

    process = SpawnDosProcess(cmd, useShell, !bg);
    if (process) {
        if (bg) {
            rc = 0;
        } else {
            struct BreakSettings bs;

            bs.title = GCS(MSG_ExecutingProgram);
            bs.steps = 0;
            bs.current = 0;
            bs.abort = TRUE;
            bs.pause = FALSE;            
            bs.needGui = FALSE;
            BreakInit(&bs);
            while (!DosProcessIsExited(process)) {
                if (BreakCheck() || GuiMainLoopIteration()) {
                    success = FALSE;
                    break;
                }
            }
            BreakExit(&bs);
            rc = DosProcessExitCode(process);
        }
        CloseDosProcess(process);
    }
    if (setSleep) {
        GuiSetSleep(FALSE);
    }

    if (prc) {
        *prc = rc;
    }

    return success;
}

static bool fileaction(const char *filename, ulong pref, bool bg)
{
    bool ok = FALSE;

    if (filename != NULL) {
        struct MemPool *memPool = CreatePool(0, 4096, 1024);
        const char *cmdtemplate = GetPrefsString(NULL, pref);
        char *path = ExpandPathname(filename, memPool);
        char *cmd = BuildCommandLine(memPool, cmdtemplate, path);
        int rc;

        if (cmd == NULL) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
        } else {
            ok = GuiSystem(cmd, bg, FALSE, &rc);
            if (ok && rc != 0) {
                Ask(AT_ERROR, GCS(MSG_StarOk), 
                        GCS(MSG_Err_ExternalProgram), cmd);
                ok = FALSE;
            }
        }
        if (memPool != NULL) {
            DeletePool(memPool);
        }
    }

    return ok;
}

#ifdef USE_EXTERNALVIEWER
bool GuiViewFile(const char *filename, bool bg)
{
    return fileaction(filename, PREFS_EXTERNALVIEWER, bg);
}
#endif

bool GuiEditFile(const char *filename, bool bg)
{
    return fileaction(filename, PREFS_EXTERNALEDITOR, bg);
}

void GuiObjectViewFile(GuiObject go)
{
    /* make sure we have correct current dir */
    ProjectSetCurrentDir(GuiObjectProject(go)); 
    GuiViewFile(GuiObjectText(go), TRUE);
}

void GuiObjectEditFile(GuiObject go)
{
    /* make sure we have correct current dir */
    ProjectSetCurrentDir(GuiObjectProject(go)); 
    GuiEditFile(GuiObjectText(go), TRUE);
}

