/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** idrecordmap.c
**
** Creation date: 12.08.2020
** Last modified: 14.08.2020
*/

#include "idrecordmap.h"
#include "record.h"
#include "util.h"
#include "memory.h"
#include <stddef.h>

struct IdRecordMapEntry
{
    ulong id;
    struct Record *rec;
};

struct IdRecordMap
{
    struct IdRecordMapEntry *irm_Entries;
    size_t irm_NumRecords;
    size_t irm_NumEntries;
    size_t irm_Size;
};

struct IdRecordMap *AllocIdRecordMap(struct MemPool *memPool)
{
    return PoolAllocMemZero(memPool, sizeof(struct IdRecordMap));
}

void FreeIdRecordMap(struct MemPool *memPool, struct IdRecordMap *map)
{
    if (map != NULL) {
        PoolFreeMem(memPool, map->irm_Entries,
                map->irm_Size * sizeof(struct IdRecordMapEntry));
        PoolFreeMem(memPool, map, sizeof(struct IdRecordMap));
    }
}

static struct IdRecordMapEntry *findEntry(const struct IdRecordMap *map,
        ulong id)
{
    if (id != 0 && map != NULL && map->irm_NumRecords != 0) {
        struct IdRecordMapEntry *entry;
        size_t ix = id;
        ptrdiff_t d;

        if (ix > map->irm_NumEntries) {
            ix = map->irm_NumEntries;
        }
        ix--;   /* First valid ID is 1. */
        entry = &map->irm_Entries[ix];
        d = entry->id - id;    /* Must be >= 0 */
        if (d == 0) {
            return entry;
        } else {
            size_t end = ix;
            size_t begin = (d > ix)? 0 : ix - d;

            do {
                ix = (begin + end) / 2;
                entry = &map->irm_Entries[ix];
                d = entry->id - id;;
                if (d == 0) {
                    return entry;
                } else if (d > 0) {
                    end = ix;
                } else {
                    begin = ix + 1;
                }
            } while (begin < end);
        }
    }

    return NULL;
}

struct Record *RecordFromId(const struct IdRecordMap *map, ulong id)
{
    const struct IdRecordMapEntry *entry = findEntry(map, id);
    struct Record *rec = NULL;

    if (entry != NULL) {
        rec = entry->rec;
    }
    return rec;
}

static int cmpIdRecordMapEntries(const void *v1, const void *v2)
{
    const struct IdRecordMapEntry *e1 = v1;
    const struct IdRecordMapEntry *e2 = v2;

    if (e1->id < e2->id) {
        return -1;
    } else if (e1->id == e2->id) {
        return 0;
    } else {
        return 1;
    }
}

static void sortMap(struct IdRecordMap *map)
{
    qsort(map->irm_Entries, map->irm_NumEntries,
            sizeof(struct IdRecordMapEntry), cmpIdRecordMapEntries);
}

bool IdRecordMapSet(struct MemPool *memPool, struct IdRecordMap *map,
        struct List *recs)
{
    size_t num = NumNodes(recs);
    struct IdRecordMapEntry *entries;
    struct IdRecordMapEntry *entry;
    struct Record *rec;

    if (map == NULL) {
        return FALSE;
    }

    entries = map->irm_Entries;
    if (num > map->irm_Size) {
        int log2 = 0;
        size_t size;

        while ((size = (1 << log2)) < num) {
            log2++;
        }
        entries = PoolAllocMem(memPool, size * sizeof(struct IdRecordMapEntry));
        if (entries == NULL) {
            return FALSE;
        }
        PoolFreeMem(memPool, map->irm_Entries,
                map->irm_Size * sizeof(struct IdRecordMapEntry));
        map->irm_Entries = entries;
        map->irm_Size = size;
    }

    map->irm_NumEntries = map->irm_NumRecords = num;
    entry = entries;
    if (recs != NULL) {
        FORNODE(rec, recs) {
            entry->id = RecordInternalNum(rec);
            entry->rec = rec;
            entry++;
        }
        sortMap(map);
    }
    return TRUE;
}

static void makeCompact(struct IdRecordMap *map)
{
    if (map->irm_NumRecords < map->irm_NumEntries) {
        /* Make map compact by removing gaps */
        struct IdRecordMapEntry *entries = map->irm_Entries;
        size_t i = 0, j;

        while (i < map->irm_NumEntries && entries[i].rec != NULL) {
            i++;
        }
        j = i++;
        while (i < map->irm_NumEntries) {
            if (entries[i].rec != NULL) {
                entries[j++] = entries[i];
            }
            i++;
        }
        map->irm_NumEntries = map->irm_NumRecords = j;
    }
}

void IdRecordMapUpdateIds(struct IdRecordMap *map)
{
    if (map != NULL) {
        struct IdRecordMapEntry *entries;
        size_t i;

        makeCompact(map);
        entries = map->irm_Entries;
        for (i = 0; i < map->irm_NumEntries; i++) {
            entries[i].id = RecordInternalNum(entries[i].rec);
        }
        sortMap(map);
    }
}

bool IdRecordMapReserveEntry(struct MemPool *memPool, struct IdRecordMap *map)
{
    struct IdRecordMapEntry *entries;
    size_t newSize;

    if (map == NULL) {
        return FALSE;
    }
    if (map->irm_NumEntries < map->irm_Size) {
        return TRUE;
    }

    makeCompact(map);
    if (map->irm_NumEntries < map->irm_Size) {
        return TRUE;
    }

    entries = map->irm_Entries;
    newSize = map->irm_Size * 2;
    if (newSize == 0) {
        newSize = 1;
    }
    entries = PoolReAllocMem(memPool, entries,
            map->irm_Size * sizeof(struct IdRecordMapEntry),
            newSize * sizeof(struct IdRecordMapEntry));
    if (entries == NULL) {
        return FALSE;
    }
    map->irm_Entries = entries;
    map->irm_Size = newSize;

    return (map->irm_NumEntries < map->irm_Size);
}

bool IdRecordMapPushEntry(struct IdRecordMap *map, struct Record *rec)
{
    ulong id = RecordInternalNum(rec);

    if (map != NULL && id != 0 && map->irm_NumEntries < map->irm_Size) {
        struct IdRecordMapEntry *entries = map->irm_Entries;
        size_t ix = map->irm_NumEntries;

        if (ix == 0 || entries[ix - 1].id < id) {
            entries[ix].id = id;
            entries[ix].rec = rec;
            map->irm_NumEntries++;
            if (rec != NULL) {
                map->irm_NumRecords++;
            }
            return TRUE;
        }
    }
    return FALSE;
}

void IdRecordMapRemove(struct IdRecordMap *map, ulong id)
{
    struct IdRecordMapEntry *entry = findEntry(map, id);

    if (entry != NULL && entry->rec != NULL)  {
        entry->rec = NULL;
        map->irm_NumRecords--;
    }
}

