/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** imexport.c
**
** Creation date: 20.03.1997
** Last modified: 28.12.2021
*/

#include "system.h"
#include "imexport.h"
#include "programeditor.h"
#include "queryeditor.h"
#include "gui.h"
#include "log.h"
#include "order.h"
#include "memo.h"
#include "strhandle.h"
#include "preferences.h"
#include "util.h"
#include "textfile.h"
#include "dosutil.h"
#include "error.h"
#include "requesters.h"
#include "break.h"
#include "appstrings_cat.h"

struct ImportField
{
    long if_Line, if_Col;
    char *if_Text;
};

static void freeImportFields(struct ImportField *ifs, size_t num, 
        struct MemPool *memPool)
{
    if (ifs) {
        size_t i;

        for (i=0; i < num; i++) {
            FreeString(memPool, ifs[i].if_Text);
        }
        PoolFreeMem(memPool, ifs, num * sizeof(struct ImportField));
    }
}

static struct ImportField *splitLine(const char *rec, struct MemPool *memPool,
        char fielddel, bool doubleQuotes, long line, long col,
        size_t *num)
{
    struct ImportField *ifs = NULL;
    bool q = FALSE, start = TRUE;
    const char *s, *first = rec, *last = first, *r;
    char *t;
    size_t numFields = 1, field = 0;
    long line_save = line, col_save = col, qline = line, qcol = col;
    long escaped = 0;

    for (s = rec; *s; s++) {
        if (doubleQuotes) {
           if (start) {
              if (*s != '"') {
                  Ask(AT_ERROR, GCS(MSG_StarOk), 
                          GCS(MSG_ImportErr_MissingDoubleQuote), line, col);
                  return NULL;
              }
           } else if (*s == '"' && s[1] == '"') {
               // Skip escaped double quote.
               s++;
               col += 2;
               continue;
           }
        }
        start = FALSE;

        if (doubleQuotes && *s == '"' && (!q || s[1] == fielddel || s[1]==0)) {
            q = !q;
        } else if (*s == fielddel && !q) {
            numFields++;
            start = TRUE;
        }

        if (*s == '\n') {
            line++;
            col = 0;
        }
        col++;
    }

    *num = numFields;
    ifs = PoolAllocMemZero(memPool, numFields * sizeof(struct ImportField));
    if (ifs == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
        return NULL;
    }

    line = line_save;
    col = col_save;
    start = TRUE;
    q = FALSE;
    escaped = 0;

    for (s = rec; ; s++, last++) {
        if (*s == 0 || (*s == fielddel && !q)) {
            ifs[field].if_Line = qline;
            ifs[field].if_Col = qcol;
            ifs[field].if_Text = t =
                PoolAllocMem(memPool, last - first - escaped + 1);
            if (t == NULL) {
                freeImportFields(ifs, numFields, memPool);
                Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
                return NULL;
            }
            for (r = first; r != last; r++) {
                *t++ = *r;
                if (doubleQuotes && *r == '"' && r[1] == '"') {
                    r++;
                }
            }
            *t = 0;
            if (*s == 0) {
                break;
            }

            field++;
            first = s+1;
            last = s;
            escaped = 0;

            qline = line;
            qcol = col;
            if (*s == '\n') {
                qline++;
                qcol = 0;
            }
            qcol++;
        } else if (doubleQuotes) {
            if (q && *s == '"' && s[1] == '"') {
               // Skip escaped double quote.
               s++;
               col += 2;
               escaped++;
               continue;
            }
            if (*s == '"' && (!q || s[1] == fielddel || s[1] == 0)) {
                if (q) {
                    last = s-1;
                } else {
                    first = s+1;
                }
                q = !q;
            }
        }

        if (*s == '\n') {
            line++;
            col = 0;
        }
        col++;
    }

    return ifs;
}

struct ImportHelp
{
    const char *filename;
    long line, col;
    bool error;
};

static struct ImportField *readOneRecord(
        TextFile *tf, struct MemPool *memPool, struct ImportHelp *ih,
        char recdel, char fielddel, bool doubleQuotes, size_t *num)
{
    struct ImportField *ifs;
    struct StrHandle *sh;
    bool q = FALSE;
    long qline = 0, qcol = 0, line_save = ih->line, col_save = ih->col;
    int c, lastc = 0;
    size_t numFields = 1;
    long err;
    const char *record;

    sh = AllocStrHandle(memPool);
    if (sh == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
        ih->error = TRUE;
        return NULL;
    }

    for (;;) {
        c = TextFileGetC(tf, &err);
        if (err) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(err), ih->filename);
            ih->error = TRUE;
            FreeStrHandle(sh);
            return NULL;
        }
        if (c == '\n') {
            ih->line++;
            ih->col = 0;
        }
        ih->col++;

        if (!q) {
            if (c == recdel || c == EOF) {
                break;
            } else if (c == fielddel) {
                numFields++;
                lastc = c;
            } else if (doubleQuotes && c == '"') {
                qline = ih->line;
                qcol = ih->col - 1;
                q = TRUE;
                lastc = 0;
            }
        } else if (lastc == '"') {
            if (c == fielddel) {
                q = FALSE;
                numFields++;
            } else if (c == recdel || c == EOF) {
                q = FALSE;
                break;
            }
            if (c != '"') {
                lastc = c;
            } else {
                // Skip escaped quotes.
                lastc = 0;
            }
        } else {
            lastc = c;
        }
        if (c == EOF) {
            Ask(AT_ERROR, GCS(MSG_StarOk), 
                    GCS(MSG_ImportErr_UnmatchedDoubleQuote), qline, qcol);
            ih->error = TRUE;
            FreeStrHandle(sh);
            return NULL;
        }
        StrHandlePutC(sh, c);
    }
    StrHandlePutC(sh, 0);

    if (StrHandleError(sh)) {
        FreeStrHandle(sh);
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
        ih->error = TRUE;
        return NULL;
    }
    if (numFields > *num) {             /* check here to save time */
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_ImportErr_NumFields),
                line_save, col_save);
        ih->error = TRUE;
        FreeStrHandle(sh);
        return NULL;
    }

    record = StrHandleString(sh);
    if (record[0] != 0 || c != EOF) {
        ifs = splitLine(record, memPool, fielddel, doubleQuotes,
                line_save, col_save, num);
        if (ifs == NULL) {
            ih->error = TRUE;
        }
    } else {
        ifs = NULL;
        *num = 0;
    }
    FreeStrHandle(sh);

    return ifs;
}

static struct Field **checkFieldNames(struct Table *tab, 
        struct ImportField *ifs, long num, struct MemPool *memPool)
{
    struct Field **fields = NULL;
    size_t size = num * sizeof(struct Field *);
    long i;

    fields = PoolAllocMem(memPool, size);
    if (fields == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
        return NULL;
    }
    for (i=0; i < num; i++) {
        fields[i] = FindFieldName(tab, ifs[i].if_Text);
        if (fields[i] == NULL) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_ImportErr_UnknownField),
                ifs[i].if_Line, ifs[i].if_Col);
            PoolFreeMem(memPool, fields, size);
            return NULL;
        }
        switch (FieldType(fields[i])) {
            case TYPE_BUTTON:
            case TYPE_VIRTUAL:
                Ask(AT_ERROR, GCS(MSG_StarOk),
                        GCS(MSG_ImportErr_InvalidFieldType),
                        ifs[i].if_Line, ifs[i].if_Col);
                PoolFreeMem(memPool, fields, size);
                return NULL;
        }
    }
    return fields;
}

const char *storeData(struct Record *rec, struct Field *field, char *text)
{
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    short type = FieldType(field);
    long error = OK;
    union Data *data = NULL;
    union Data storage;
    bool nil = FALSE;

    if (pr == NULL || text == NULL) {
        return GCS(MSG_Err_Internal);
    }

    switch (type) {
        case TYPE_INTEGER:
            nil = !StrToInt(text, &storage.integer);
            data = &storage;
            break;

        case TYPE_REAL:
            nil = !StrToDouble(text, &storage.real);
            data = &storage;
            break;

        case TYPE_BOOL:
            if (!CodesetStrCmp(text, "NIL")) {
                nil = TRUE;
            } else if (!CodesetStrCmp(text, "TRUE")) {
                nil = FALSE;
            } else {
                return GCS(MSG_ImportErr_Boolean);
            }
            data = &storage;
            break;

        case TYPE_STRING:
            storage.string = text;
            data = &storage;
            break;

        case TYPE_MEMO:
            storage.mi = text;
            data = &storage;
            break;

        case TYPE_DATE:
            nil = !StringToDate(text, &storage.date);
            data = &storage;
            break;

        case TYPE_TIME:
            nil = !StringToTime(text, FORMAT_HOURMINSEC, &storage.time);
            data = &storage;
            break;

        case TYPE_CHOICE:
        {
            char **labels = FieldChoiceLabels(field);
            short i;

	    if (labels == NULL)
                return GCS(MSG_ImportErr_Choice);

	    for (i=0; labels[i] != NULL; i++) {
		if (!strcmp(labels[i], text)) {
		    break;
                }
            }
	    if (labels[i] == NULL) {
                return GCS(MSG_ImportErr_Choice);
            }
            storage.choice = i;
            data = &storage;
            break;
        }

        case TYPE_BUTTON:
            break;

        case TYPE_REFERENCE:
        {
            struct Table *reftab = FieldReferenceTable(field);
            struct List *refrecs = TableRecords(reftab);
            long num = 0;

            nil = !StrToInt(text, &num);
            storage.rec = FindNodeNum(refrecs, num);
            data = &storage;
            break;
        }
    }

    if (data) {
        error = RecordStoreData(rec, field, data, nil);
    }
    if (rec == TableGuiRecord(tab)) {
        GuiUpdateField(field, rec);
    }

    if (error) {
        return GES(error);
    }
    return NULL;
}

static bool importRecordsHelp(struct Table *tab, bool overwriteRecords, 
        TextFile *tf, const char *filename, 
        char recdel, char fielddel, bool doubleQuotes, 
        long *overwritten, long *inserted)
{
    struct MemPool *memPool = TableMemPool(tab);
    struct ImportField *ifs = NULL;
    struct Field **fields = NULL;
    struct List *recs = TableRecords(tab);
    struct Record *overwriteRec = NULL;
    struct Record *rec;
    size_t numFields = TableNumFields(tab), num, count = 0, i;
    size_t fieldsize;
    long error = OK;
    const char *errorstr = NULL;
    struct ImportHelp ih;
    char message[100];

    ih.filename = filename;
    ih.line = 1;
    ih.col = 1;
    ih.error = FALSE;

    *inserted = 0;
    *overwritten = 0;

    if (recs == NULL) {
        return FALSE;
    }

    /*
    ** Clear imported flags
    */
    FORNODE(rec, recs) {
        SetRecordImported(rec, FALSE);
    }

    ifs = readOneRecord(tf, memPool, &ih,
            recdel, fielddel, doubleQuotes, &numFields);
    if (ifs == NULL || numFields == 0) {
        return FALSE;
    }
    fields = checkFieldNames(tab, ifs, numFields, memPool);
    freeImportFields(ifs, numFields, memPool);
    if (fields == NULL) {
        return FALSE;
    }
    fieldsize = numFields * sizeof(struct Field *);

    if (overwriteRecords) {
        overwriteRec = (struct Record *)HEAD(recs);
    }

    for (;;) {
        if (BreakCheck()) {
            return FALSE;
        }
        num = numFields;
        ifs = readOneRecord(tf, memPool, &ih,
                recdel, fielddel, doubleQuotes, &num);
        if (ifs == NULL) {
            break;
        }

        sprintf(message, GCS(MSG_Import_Record), (long)++count);
        BreakMessage(message);

        if (num != numFields) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_ImportErr_NumFields),
                    ifs[0].if_Line, ifs[0].if_Col);
            freeImportFields(ifs, num, memPool);
            PoolFreeMem(memPool, fields, fieldsize);
            return FALSE;
        }

        rec = overwriteRec;
        if (rec == NULL || SUCC(rec) == NULL) {
            error = NewRecord(tab, NULL, ADDRECORD_TAIL, FALSE, &rec);
            if (error) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_ImportErr_Error),
                        GES(error), ifs[0].if_Line, ifs[0].if_Col);
                freeImportFields(ifs, numFields, memPool);
                PoolFreeMem(memPool, fields, fieldsize);
                return FALSE;
            }
            SetRecordImported(rec, TRUE);
            overwriteRec = NULL;
            ++*inserted;
        }

        for (i=0; i < num; i++) {
            errorstr = storeData(rec, fields[i], ifs[i].if_Text);
            if (errorstr) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_ImportErr_Error),
                        errorstr, ifs[i].if_Line, ifs[i].if_Col);
                DeleteRecord(tab, rec);
                freeImportFields(ifs, numFields, memPool);
                PoolFreeMem(memPool, fields, fieldsize);
                return FALSE;
            } else if (i == 0 && overwriteRec != NULL) {
                overwriteRec = SUCC(overwriteRec);
                ++*overwritten;
            }
        }
        freeImportFields(ifs, numFields, memPool);
    }

    PoolFreeMem(memPool, fields, fieldsize);
    if (ih.error) {
        return FALSE;
    } else if (TextFileWarnings(tf) > 0) {
        Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Warn_CodesetProcessFile),
                TextFileWarnings(tf), filename);
    }
    return TRUE;
}

static bool importRecordsRaw(struct Table *tab, bool overwriteRecords,
        const char *filename, char recdel, char fielddel, bool doubleQuotes,
        long *overwritten, long *inserted)
{
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = ProjectMemPool(pr);
    bool rc = FALSE;
    TextFile *tf;

    if (tab == NULL || filename == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Internal));
        return FALSE;
    }

    ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
    tf = TextFileOpen(memPool, filename, "r", TFE_AUTO);
    if (tf == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_OpenFile), filename);
    } else {
        struct BreakSettings bs;

        bs.title = GCS(MSG_Import);
        bs.steps = 0;
        bs.current = 0;
        bs.abort = TRUE;
        bs.pause = TRUE;
        bs.needGui = ProjectHasGui(pr);
        BreakInit(&bs);
        rc = importRecordsHelp(tab, overwriteRecords, tf, filename, 
                recdel, fielddel, doubleQuotes, overwritten, inserted);
        TextFileClose(tf);
        BreakExit(&bs);
    }

    return rc;
}

bool ImportRecords(struct Table *tab, bool overwriteRecords,
        const char *filename, char recdel, char fielddel, bool doubleQuotes)
{
    struct Project *pr = TableProject(tab);
    long overwritten, inserted;
    long error;
    bool rc;

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error), ProjectName(pr));
        return FALSE;
    }

    GuiSetSleep(TRUE);
    rc = importRecordsRaw(tab, overwriteRecords, filename, recdel, fielddel,
            doubleQuotes, &overwritten, &inserted);
    if (rc) {
        if (overwritten > 0) {
            Ask(AT_INFO, GCS(MSG_StarOk), GCS(MSG_Import_OverwrittenRecords), 
                    overwritten);
        }
        if (inserted > 0 && !Ask(AT_QUESTION, 
                    GCS(MSG_Import_InsertRecordsAnswers), 
                    GCS(MSG_Import_InsertRecords), inserted)) {
            DeleteRecordsTable(tab, FALSE, TRUE);
            inserted = 0;
        }
        if (overwritten + inserted > 0) {
            error = ReorderRecords(tab);
            if (error) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Order), GES(error));
            }
            ProjectChange(pr, FALSE);
            GuiAddRecord(tab, (struct Record *)HEAD(TableRecords(tab)));
            GuiUpdateDependentDisplayObjects(pr, tab, NULL, NULL);
        }
        GuiDisplayPanel(tab);
    }
    GuiSetSleep(FALSE);
    return rc;
}

#ifdef ENABLE_LOG_APPLY_CHANGES
static bool logEntryDateAndTime(struct Log *log, struct Record *rec,
        long *date, long *time)
{
    bool valid = FALSE;
    struct DataInfo di_buf, *di;
    long error = OK;

    RecordLock(rec);
    error = RecordDataInfo(rec, LogField(log, LA_DATE), &di, &di_buf);
    if (error == OK && !(di->di_Flags & DIF_NIL)) {
        *date = di->di_Contents.date;
        error = RecordDataInfo(rec, LogField(log, LA_TIME), &di, &di_buf);
        if (error == OK && !(di->di_Flags & DIF_NIL)) {
            *time = di->di_Contents.time;
            valid = TRUE;
        }
    }
    RecordUnlock(rec);

    return valid;
}

static bool oldestLogEntry(struct Log *log, long *oldestDate, long *oldestTime)
{
    struct Table *tab = LogTable(log);
    struct List *recs = TableRecords(tab);
    struct Record *rec;
    bool haveEntry = FALSE;
    long date, time;

    if (recs != NULL) {
        FORNODE(rec, recs) {
            if (RecordImported(rec)) {
                continue;
            }

            if (logEntryDateAndTime(log, rec, &date, &time)) {
                if (!haveEntry || date < *oldestDate ||
                        (date == *oldestDate && time < *oldestTime)) {
                    *oldestDate = date;
                    *oldestTime = time;
                    haveEntry = TRUE;
                }
            }
        }
    }

    return haveEntry;
}

static long findDuplicate(struct Table *tab, struct Record *rec,
        struct Record **match)
{
    struct List *fields = TableFields(tab);
    struct Field *field;
    struct List *recs = *match? (struct List *)PRED(*match) : TableRecords(tab);
    struct Record *rec2;
    struct DataInfo di_buf, di2_buf, *di, *di2;
    long error = OK, result = 0;

    *match = NULL;
    RecordLock(rec);
    FORNODE(rec2, recs) {
        if (RecordImported(rec2)) {
            continue;
        }
        RecordLock(rec2);
        FORNODE(field, fields) {
            error = RecordDataInfo(rec, field, &di, &di_buf);
            if (error != OK) {
                break;
            }
            error = RecordDataInfo(rec2, field, &di2, &di2_buf);
            if (error != OK) {
                break;
            }
            result = FieldCmp(field, di, di2);
            if (result != 0) {
                break;
            }
        }
        RecordUnlock(rec2);
        if (error != OK) {
            break;
        }
        if (result == 0) {
            *match = rec2;
            break;
        }
    }
    RecordUnlock(rec);
    return error;
}

static void importApplyChangesHelp(struct Log *log, long num)
{
    struct Table *tab = LogTable(log);
    struct Project *pr = TableProject(tab);
    struct List *recs = TableRecords(tab);
    struct Record *rec, *next;
    struct Record *match = NULL;
    struct BreakSettings bs;
    bool isFirstEntry = TRUE, wasFirstEntry;
    bool isTransaction = FALSE;
    bool inRollback = FALSE;
    long oldestDate = 0, oldestTime = 0, date, time;
    bool haveOldestEntry = FALSE;
    bool askSkipOld = TRUE;
    bool skipOld = FALSE;
    bool foundFirstEntry = FALSE;
    long applied = 0, skipped = 0, failed = 0;
    long error;
    const char *errstr;
    int rc;

    bs.title = GCS(MSG_ApplyChanges);
    bs.steps = num;
    bs.current = 0;
    bs.abort = TRUE;
    bs.pause = TRUE;
    bs.needGui = ProjectHasGui(pr);
    BreakInit(&bs);

    haveOldestEntry = oldestLogEntry(log, &oldestDate, &oldestTime);

    FORNODE(rec, recs) {
        if (!RecordImported(rec)) {
            continue;
        }
        BreakAddCurrentStep(1);
        if (BreakCheck()) {
            break;
        }
        wasFirstEntry = isFirstEntry;
        isFirstEntry = FALSE;

        /* Skip transaction entries and entire rollbacks */
        isTransaction = LogChangeIsBeginTransaction(log, rec);
        if (isTransaction || wasFirstEntry) {
            bool askNoCommit = isTransaction;
            bool haveTransactionEntries = !isTransaction;

            /* Check if this transaction has been committed or rolled back */
            for (next = SUCC(rec); SUCC(next) != NULL; next = SUCC(next)) {
                if (!RecordImported(next)) {
                    continue;
                } else if (LogChangeIsCommit(log, next)) {
                    break;
                } else if (LogChangeIsBeginTransaction(log, next)) {
                    askNoCommit = TRUE;
                    break;
                } else if (LogChangeIsRollback(log, next)) {
                    inRollback = TRUE;
                    break;
                }
                haveTransactionEntries = TRUE;
            }
            if (askNoCommit && haveTransactionEntries) {
                long date, time;
                bool haveDateAndTime =
                    logEntryDateAndTime(log, rec, &date, &time);

                if (Ask(AT_QUESTION,
                            GCS(MSG_ApplyChanges_SkipAnswers),
                            GCS(MSG_ApplyChanges_SkipUncommittedEntries),
                            haveDateAndTime? DateToString(date,
                                GetPrefsValue(pr, PREFS_DATEFORMAT)) : "NIL",
                            haveDateAndTime? TimeToString(time,
                                FORMAT_HOURMINSEC) : "NIL") == 0){
                    inRollback = TRUE;
                }
            }
            if (isTransaction) {
                skipped++;
                continue;
            }
        } else if (LogChangeIsCommit(log, rec)) {
            skipped++;
            continue;
        } else if (LogChangeIsRollback(log, rec)) {
            inRollback = FALSE;
            skipped++;
            continue;
        }
        if (inRollback) {
            skipped++;
            continue;
        }
        /* Skip old records */
        if (haveOldestEntry && logEntryDateAndTime(log, rec, &date, &time)) {
            if (date < oldestDate || (date == oldestDate && time < oldestTime)){
                if (askSkipOld) {
                    skipOld = Ask(AT_QUESTION,
                            GCS(MSG_ApplyChanges_SkipAnswers),
                            GCS(MSG_ApplyChanges_SkipOldEntries)) == 0;
                    askSkipOld = FALSE;
                }
                if (skipOld) {
                    skipped++;
                    continue;
                }
            }
        }
        /* Skip records that we already have */
        if (!foundFirstEntry) {
            error = findDuplicate(tab, rec, &match);
            if (error) {
                // Fatal error, quit processing.
                Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
                break;
            } else if (match != NULL) {
                skipped++;
                continue;
            } else {
                foundFirstEntry = TRUE;
            }
        }
        errstr = LogApplyChange(log, rec);
        if (errstr != NULL) {
            struct MemPool *memPool = ProjectMemPool(pr);
            struct {
                int logFieldId;
                char *text;
                bool free;
            } info[] = {
                { LA_DATE, NULL, FALSE },
                { LA_TIME, NULL, FALSE },
                { LA_ACTION, NULL, FALSE },
                { LA_TABLE, NULL, FALSE },
                { LA_RECORD, NULL, FALSE },
                { LA_DESCRIPTION, NULL, FALSE },
                { LA_FIELD, NULL, FALSE },
                { LA_OLD, NULL, FALSE },
                { LA_NEW, NULL, FALSE }
            };
            struct Field *field;
            int i;

            RecordLock(rec);
            for (i = 0; i < 9; i++) {
                field = LogField(log, info[i].logFieldId);
                error = GuiFieldDataString(field, rec, memPool,
                        &info[i].text, &info[i].free);
                if (error) {
                    info[i].text = GES(error);
                    info[i].free = FALSE;
                }
            }
            rc = Ask(AT_ERROR, GCS(MSG_ApplyChanges_ErrorAnswers),
                    GCS(MSG_ApplyChanges_Error),
                    info[0].text, info[1].text, info[2].text,
                    info[3].text, info[4].text, info[5].text,
                    info[6].text, info[7].text, info[8].text,
                    errstr);
            for (i = 0; i < 9; i++) {
                if (info[i].free) {
                    FreeString(memPool, info[i].text);
                }
            }
            RecordUnlock(rec);
            if (rc == 0) {
                break;
            }
            failed++;
        } else {
            applied++;
        }
        /* Keep this record in the _Log table when logging to Project */
        if (LogToProject(log)) {
            SetRecordImported(rec, FALSE);
        }
    }

    DeleteRecordsTable(tab, FALSE, TRUE);
    BreakExit(&bs);

    if (applied == 0) {
        Ask(AT_INFO, GCS(MSG_StarOk), GCS(MSG_ApplyChanges_ResultFailed),
                skipped, failed, num - skipped - failed);
    } else {
        Ask(AT_INFO, GCS(MSG_StarOk), GCS(MSG_ApplyChanges_ResultSuccess),
                applied, skipped, failed, num - applied - skipped - failed);
    }

    if (applied + failed > 0) {
        error = ReorderRecords(tab);
        if (error) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Order), GES(error));
        }
        ProjectChange(pr, FALSE);
        GuiAddRecord(tab, (struct Record *)HEAD(TableRecords(tab)));
        GuiUpdateDependentDisplayObjects(pr, tab, NULL, NULL);
    }
    GuiUpdate(pr);
}

bool ImportApplyChanges(struct Project *pr, const char *filename,
        char recdel, char fielddel, bool doubleQuotes)
{
    struct Log *log = ProjectLog(pr);
    struct Table *tab = LogTable(log);
    bool logActive = LogActive(log);
    long overwritten, inserted;
    long error;
    bool rc = FALSE;

    SetLogActive(log, FALSE);

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error), ProjectName(pr));
    } else {
        GuiSetSleep(TRUE);
        rc = importRecordsRaw(tab, FALSE, filename, recdel, fielddel,
                doubleQuotes, &overwritten, &inserted);
        if (rc) {
            if (inserted == 0) {
                Ask(AT_INFO, GCS(MSG_StarOk),
                        GCS(MSG_ApplyChanges_NoChangesFound));
            } else if (Ask(AT_QUESTION, GCS(MSG_ApplyChanges_ConfirmAnswers),
                        GCS(MSG_ApplyChanges_Confirm), inserted) != 0) {
                importApplyChangesHelp(log, inserted);
            }
        }
        GuiSetSleep(FALSE);
    }
    error = SetLogActive(log, logActive);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
    }

    return rc;
}
#endif

static long exportStringOut(TextFile *tf, const char *text, long error)
{
    if (error == OK) {
        TextFilePutS(tf, text, &error);
    }
    return error;
}

static long exportCharOut(TextFile *tf, cschar cs, long error)
{
    if (error == OK) {
        TextFilePutChar(tf, cs, &error);
    }
    return error;
}

static long exportWriteField(TextFile *tf, long *col, char recdel, char fielddel, 
        bool doubleQuotes, const char *contents, long error)
{
    if (*col != 0) {
        error = exportCharOut(tf, fielddel, error);
    }
    if (doubleQuotes) {
        const char *s = contents, *t = s;

        error = exportCharOut(tf, '"', error);
        for (;;) {
            t = strchr(t, '"');
            if (t == NULL) {
                error = exportStringOut(tf, s, error);
                break;
            }
            if (t[1] == 0 || t[1] == '"' || t[1] == recdel || t[1] == fielddel) {
                while (s <= t) {
                    cschar cs = CodesetGetChar(s);

                    error = exportCharOut(tf, cs, error);
                    s = CodesetNextChar(s);
                }
                error = exportCharOut(tf, '"', error);
                s = t = t + 1;
            } else {
                t = t + 1;
            }
        }
        error = exportCharOut(tf, '"', error);
    } else {
        error = exportStringOut(tf, contents, error);
    }
    ++*col;

    return error;
}

static TextFile *exportOpenFile(struct Project *pr, const char *filename)
{
    TextFile *tf;

    ProjectSetCurrentDir(pr);           /* make sure we have correct CWD */
    if (FileStat(filename) >= 0 &&      /* file exists? */
            Ask(AT_WARN, GCS(MSG_AskReplaceAnswers), 
                GCS(MSG_AskReplaceFile), filename) == 0) {
        return NULL;
    }

    tf = TextFileOpen(ProjectMemPool(pr), filename, "w", TFE_AUTO);
    if (tf == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_OpenFile), filename);
        return NULL;
    }

    return tf;
}

static bool exportCloseFile(TextFile *tf, const char *filename, long error)
{
    size_t warnings = TextFileWarnings(tf);
    long error2 = TextFileClose(tf);

    if (error == OK) {
        error = error2;
    }

    if (error) {
        if (error != ERR_CANCEL) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error), filename);
        }
        return FALSE;
    }
    if (warnings) {
        Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Warn_CodesetProcessFile),
                warnings, filename);
    }

    return TRUE;
}

static bool exportRecordsHelp(struct Table *tab, bool filter,
        const char *filename, char recdel, char fielddel, bool doubleQuotes)
{
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = TableMemPool(tab);
    struct List *recs = TableRecords(tab);
    struct Record *rec;
    struct List *fields = TableFields(tab);
    struct Field *field;
    TextFile *tf;
    short type;
    const char *name;
    struct DataInfo *di = NULL, di_buf;
    long col;
    long error = OK;
    char *text = NULL;
    bool free = FALSE;
    long count = 0;

    if (recs == NULL || fields == NULL || filename == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Internal));
        return FALSE;
    }

    tf = exportOpenFile(pr, filename);
    if (tf == NULL) {
        return FALSE;
    }

    col = 0;
    FORNODE(field, fields) {
        switch(FieldType(field)) {
            case TYPE_BUTTON:
            case TYPE_VIRTUAL:
                break;

            default:
                name = FieldName(field);
                error = exportWriteField(tf, &col, recdel, fielddel,
                        doubleQuotes, name, error);
                break;
        }
    }
    error = exportCharOut(tf, recdel, error);

    FORNODE(rec, recs) {
        if (error) {
            break;
        }
        if (BreakCheck()) {
            error = ERR_CANCEL;
            break;
        }
        if (filter && !RecordMatchFilter(rec)) {
            continue;
        }

        RecordLock(rec);
        col = 0;
        FORNODE(field, fields) {
            if (error) {
                break;
            }
            type = FieldType(field);
            switch (type) {
                case TYPE_BUTTON:
                case TYPE_VIRTUAL:
                    break;

                default:
                    text = NULL;
                    free = FALSE;
                    if (type == TYPE_REFERENCE) {
                        error = RecordDataInfo(rec, field, &di, &di_buf);
                        if (error) {
                            break;
                        }
                        text = IntToStr(RecordNum(di->di_Contents.rec));
                        free = FALSE;
                    } else {
                        error = GuiFieldDataString(field, rec, memPool,
                                &text, &free);
                    }
                    error = exportWriteField(tf, &col, recdel, fielddel,
                            doubleQuotes, text, error);
                    if (free) {
                        FreeString(memPool, text);
                    }
                    break;
            }
        }
        RecordUnlock(rec);
        error = exportCharOut(tf, recdel, error);
        count++;
    }

    if (!exportCloseFile(tf, filename, error)) {
        return FALSE;
    }
    Ask(AT_INFO, GCS(MSG_StarOk), GCS(MSG_Export_Success), count, filename);
    return TRUE;
}

bool ExportRecords(struct Table *tab, bool filter, const char *filename,
        char recdel, char fielddel, bool doubleQuotes)
{
    bool rc = FALSE;
    struct BreakSettings bs;
    bs.title = GCS(MSG_Export);
    bs.steps = 0;
    bs.current = 0;
    bs.abort = TRUE;
    bs.pause = TRUE;
    bs.needGui = ProjectHasGui(TableProject(tab));
    BreakInit(&bs);
    rc = exportRecordsHelp(tab, filter, filename, recdel, fielddel, 
            doubleQuotes);
    BreakExit(&bs);

    return rc;
}

static long exportWriteQueryLine(TextFile *tf, 
        char recdel, char fielddel, bool doubleQuotes, 
        const struct QueryLine *ql, char numColumns,
        const signed char *columnOrder)
{
    long error = OK;
    long col = 0;
    int i, c;
    const char *text;

    for (i = 0; i < numColumns; i++) {
        c = i;
        if (columnOrder) {
            c = columnOrder[i];
            if (c < 0 || c >= numColumns) {
                continue;               /* invalid column */
            }
        }
        text = ql->ql_Entries[c]->qe_String;
        error = exportWriteField(tf, &col, recdel, fielddel, doubleQuotes, text, error);
    }
    error = exportCharOut(tf, recdel, error);

    return error;
}

static bool exportSFWListHelp(struct Project *pr, GuiObject sfw,
        const char *filename, char recdel, char fielddel, bool doubleQuotes)
{
    TextFile *tf;
    long error = OK;
    long count = 0;
    char numColumns;
    const signed char *columnOrder;
    const struct QueryLine *ql;

    if (pr == NULL || sfw == NULL || filename == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Internal));
        return FALSE;
    }

    numColumns = GuiSFWListNumColumns(sfw);
    columnOrder = GuiSFWListColumnOrder(sfw);

    tf = exportOpenFile(pr, filename);
    if (tf == NULL) {
        return FALSE;
    }

    do {
        if (BreakCheck()) {
            error = ERR_CANCEL;
            break;
        }
        ql = GuiSFWListQueryLine(sfw, count);
        if (!ql) {
            break;
        }
        error = exportWriteQueryLine(tf, recdel, fielddel, doubleQuotes, 
                ql, numColumns, columnOrder);
        count++;
    } while (error == OK);

    if (!exportCloseFile(tf, filename, error)) {
        return FALSE;
    }
    Ask(AT_INFO, GCS(MSG_StarOk), GCS(MSG_Export_Success), count, filename);
    return TRUE;
}

bool ExportSFWList(struct Project *pr, GuiObject sfw, const char *filename,
        char recdel, char fielddel, bool doubleQuotes)
{
    bool rc = FALSE;
    struct BreakSettings bs;
    bs.title = GCS(MSG_Export);
    bs.steps = 0;
    bs.current = 0;
    bs.abort = TRUE;
    bs.pause = TRUE;
    bs.needGui = ProjectHasGui(pr);
    BreakInit(&bs);
    rc = exportSFWListHelp(pr, sfw, filename, recdel, fielddel, doubleQuotes);
    BreakExit(&bs);

    return rc;
}

