/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** labels.c
**
** Creation date: 01.05.1996
** Last modified: 04.01.2022
*/

#include "system.h"
#include "labels.h"
#include "codeset.h"
#include "dosutil.h"
#include "util.h"
#include "error.h"

size_t NumLabels(char **labels)
{
    size_t num = 0;
    char **l = labels;

    if (l) {
        while (*l) {
            l++;
        }
        num = l - labels;
    }
    return num;
}

void FreeLabelsNum(struct MemPool *memPool, char **labels, size_t num)
{
    if (labels) {
    	size_t i;

        for (i = 0; i < num; i++) {
	    FreeString(memPool, labels[i]);
        }
    	PoolFreeMem(memPool, labels, (num + 1) * sizeof(char *));
    }
}

void FreeLabels(struct MemPool *memPool, char **labels)
{
    FreeLabelsNum(memPool, labels, NumLabels(labels));
}

char **AllocLabels(struct MemPool *memPool, size_t num)
{
    return PoolAllocMemZero(memPool, ((long)num + 1) * sizeof(char *));
}

/* Allocates label array and initializes each element with NULL. */

char **CopyLabels(struct MemPool *memPool, char **labels)
{
    size_t num = NumLabels(labels);
    char **newlabels = NULL;
    size_t i;

    newlabels = AllocLabels(memPool, num);
    if (newlabels == NULL) {
	return NULL;
    }

    /* copy array */
    for (i = 0; i < num; i++) {
	newlabels[i] = AllocString(memPool, labels[i]);
	if (newlabels[i] == NULL) {
            FreeLabelsNum(memPool, newlabels, num);
	    return NULL;
	}
    }
    return newlabels;
}

char **StringToLabels(struct MemPool *memPool, const char *str)
{
    cschar sep = '\n';
    size_t num = StrCount(str, sep) + 1, i;
    char **labels = NULL;
    const char *s, *t; 

    if (str == NULL) {
        return NULL;
    }

    labels = AllocLabels(memPool, num);
    if (labels == NULL) {
        return NULL;
    }

    s = str;
    for (i = 0; i < num; i++) {
	t = CodesetStrchr(s, sep);
	if (t == NULL) {
	    t = s + strlen(s);
	}
	labels[i] = AllocStringN(memPool, s, t-s);
	if (labels[i] == NULL) {
	    FreeLabelsNum(memPool, labels, num);
	    return NULL;
	}
	if (*t == 0) {
	    break;
	}
	s = CodesetNextChar(t);
    }
    return labels;
}

char *LabelsToString(struct MemPool *memPool, char **labels)
{
    char **l;
    char *res = NULL, *s;
    size_t size = 0;
    cschar sep = '\n';
    short seplen = CodesetCharToStr(sep, NULL);

    if (seplen < 0) {
	return NULL;
    }

    for (l = labels; *l; l++) {
        size += strlen(*l) + ((l[1] != NULL)? seplen : 0);
    }
    size++;

    res = PoolAllocMem(memPool, size);
    if (res == NULL) {
        return NULL;
    }

    s = res;
    for (l = labels; *l; l++) {
        strcpy(s, *l);
	s += strlen(s);
        if (l[1] != NULL) {
	    CodesetCharToStr(sep, s);
	    s += seplen;
        }
    }
    *s = 0;

    return res;
}

char **LoadLabels(struct MemPool *memPool, FILE *file, bool utf8, 
        long *perror, size_t *pwarnings)
{
    char **labels = NULL, *src, *cs;
    size_t num = 0, len, i;
    long error = OK;
    size_t warnings = 0;

    if (file == NULL) {
        error = ERR_INTERNAL;
    } else {
        num = FGetShort(file);
        labels = AllocLabels(memPool, num);
        if (labels == NULL) {
            error = ERR_MEM;
        } else {
            for (i = 0; i < num; i++) {
                len = FGetShort(file);
                src = PoolAllocMem(memPool, len + 1);
                if (src == NULL) {
                    error = ERR_MEM;
                    break;
                }
                if (len > 0) {
                    len = fread(src, 1, len, file);
                }
                src[len] = 0;
                if (utf8) {
                    cs = CodesetConvertFromUTF8(src, memPool, &error,&warnings);
                } else {
                    cs = CodesetConvertFromDefault8Bit(src, memPool, &error);
                }
                if (error) {
                    break;
                }
                if (cs != src) {
                    PoolFreeMem(memPool, src, len + 1);
                }
                labels[i] = cs;
            }
        }
    }

    if (error && labels) {
        FreeLabelsNum(memPool, labels, num);
        labels = NULL;
    }

    if (perror) {
        *perror = error;
    }
    if (pwarnings) {
        *pwarnings = warnings;
    }
    return labels;
}

