/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** log.c
**
** Creation date: 03.05.2012
** Last modified: 13.04.2022
*/

#include "system.h"
#include "log.h"
#include "programeditor.h"
#include "gui.h"
#include "project.h"
#include "table.h"
#include "order.h"
#include "buildgui.h"
#include "labels.h"
#include "memo.h"
#include "requesters.h"
#include "preferences.h"
#include "textfile.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "lang/program.h"
#include "appstrings_cat.h"

#define LOG_TABLE_NAME "_Log"

static const struct LogFieldInfo
{
    const char *lai_Name;
    short lai_Type;
} 
LogFieldInfo[NumLogFields] = 
{
    { "_Label", TYPE_STRING },
    { "_Date", TYPE_DATE },
    { "_Time", TYPE_TIME },
    { "_Action", TYPE_STRING },
    { "_Table", TYPE_STRING },
    { "_Record", TYPE_INTEGER },
    { "_Description", TYPE_STRING },
    { "_Field", TYPE_STRING },
    { "_OldValue", TYPE_STRING },
    { "_NewValue", TYPE_STRING }
};

static const char logActionBeginTransaction[] = "Begin transaction";
static const char logActionRollback[] = "Rollback";
static const char logActionCommit[] = "Commit";
static const char logActionNewRecord[] = "New record";
static const char logActionDeleteRecord[] = "Delete record";
static const char logActionChangeValue[] = "Change value";

struct Log
{
    struct Project *log_Project;
    struct Table *log_Table;
    struct Field *log_Fields[NumLogFields];
    TextFile *log_TextFile;
    bool log_OpenAttempted;
#ifdef USE_LOG_WINDOW
    struct DisplayObject *log_WindowObject;
#endif
};

void logCloseFile(struct Log *log)
{
    TextFileClose(log->log_TextFile);
    log->log_TextFile = NULL;
    log->log_OpenAttempted = FALSE;
}

long ClearLog(struct Log *log, bool reset)
{
    long error = OK;

    if (log) {
	struct Table *tab = log->log_Table;

        if (tab == NULL) {
            tab = FindTableName(log->log_Project, LOG_TABLE_NAME);
        }

	if (reset) {
#ifdef USE_LOG_WINDOW
            SetPrefsWindowDimensions(log->log_Project, PREFS_LOG_WINDOW,
                    DisplayWindowDimensions(log->log_WindowObject));
            GuiCloseWindow(DisplayObjectGuiObject(log->log_WindowObject));
            FreeRootDisplayObject(log->log_WindowObject);
            log->log_WindowObject = NULL;
#endif
	    error = DeleteTable(tab);
	    log->log_Table = NULL;
	    memset(log->log_Fields, 0, sizeof(log->log_Fields));
            logCloseFile(log);
            SetLogLabel(log, NULL);
            SetLogActive(log, FALSE);
	} else if (tab) {
	    error = DeleteRecordsTable(tab, FALSE, FALSE);
	}
    }

    return error;
}

void ReleaseLogFile(struct Log *log)
{
    if (log && log->log_TextFile) {
        logCloseFile(log);
    }
}

void FreeLog(struct Log *log)
{
    if (log) {
	ClearLog(log, TRUE);
	PoolFreeMem(ProjectMemPool(log->log_Project), log, sizeof(struct Log));
    }
}

struct Log *AllocLog(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct Log *log;

    log = PoolAllocMemZero(memPool, sizeof(struct Log));
    if (log == NULL) {
	return NULL;
    }

    log->log_Project = pr;

    return log;
}

static long logCreateTable(struct Log *log)
{
    struct Project *pr;
    struct Table *tab;
    bool projectChanged = FALSE;
    long error;
    int i;

    if (log == NULL) {
	return ERR_INTERNAL;
    }

    tab = log->log_Table;
    if (tab != NULL) {
	return OK;
    }

    pr = log->log_Project;
    tab = FindTableName(pr, LOG_TABLE_NAME);
    if (tab == NULL) {
	struct TableData td;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            return error;
        }

	memset(&td, 0, sizeof(td));
	td.td_Name = LOG_TABLE_NAME;
	td.td_System = TRUE;
	error = NewTable(pr, &td, &tab);
	if (error) {
	    return error;
	} else if (!projectChanged) {
            ProjectChange(pr, TRUE);
            projectChanged = TRUE;
        }
    }

    for (i = 0; i < NumLogFields; i++) {
	struct Field *field = log->log_Fields[i];
	
	if (field == NULL) {
	    const char *name = LogFieldInfo[i].lai_Name;

	    field = FindFieldName(tab, name);
	    if (!field && strcmp(name, "_Field") == 0) {
                /* Special case since we renamed Attribute to Field in V4.2 */
                field = FindFieldName(tab, "_Attr");
                if (field != NULL) {
                    struct ChangeFieldData cfd;

                    error = ObtainProjectChangeLock(pr);
                    if (error != OK) {
                        return error;
                    }

                    memset(&cfd, 0, sizeof(cfd));
                    cfd.cfd_Field = field;
                    cfd.cfd_Name = name;
                    cfd.cfd_Contents.string.maxlen = MAX_STRING_LEN;
                    error = ChangeField(&cfd);
                    if (error) {
                        return error;
                    } else if (!projectChanged) {
                        ProjectChange(pr, TRUE);
                        projectChanged = TRUE;
                    }
                }
            }
	    if (!field) {
		struct NewFieldData nfd;

                error = ObtainProjectChangeLock(pr);
                if (error != OK) {
                    return error;
                }

		memset(&nfd, 0, sizeof(nfd));
		nfd.nfd_Table = tab;
                if (i > 0) {
                    nfd.nfd_Pred = log->log_Fields[i-1];
                }
		nfd.nfd_Name = name;
		nfd.nfd_Type = LogFieldInfo[i].lai_Type;
		switch (nfd.nfd_Type) {
		    case TYPE_STRING:
			nfd.nfd_Contents.string.maxlen = MAX_STRING_LEN;
			break;
		}

		error = NewField(&nfd, &field);
		if (error) {
		    return error;
                } else if (!projectChanged) {
                    ProjectChange(pr, TRUE);
                    projectChanged = TRUE;
                }
	    }
	    log->log_Fields[i] = field;
	}
    }

    log->log_Table = tab;
    return OK;
}

static long logWriteFile(struct Log *log, bool tab, const char *s, bool newline)
{
    TextFile *tf = log->log_TextFile;
    long error = OK;

    if (error == OK && tab) {
        TextFilePutChar(tf, '\t', &error);
    }
    if (error == OK) {
        TextFilePutChar(tf, '"', &error);
    }
    if (error == OK) {
        TextFilePutS(tf, s, &error);
    }
    if (error == OK) {
        TextFilePutChar(tf, '"', &error);
    }
    if (error == OK && newline) {
        TextFilePutChar(tf, '\n', &error);
        if (error == OK) {
            error = TextFileFlush(tf);
        }
    }

    if (error == OK && TextFileError(tf)) {
        error = LOGERR_WRITE;
    }
    if (error == DISKERR_WRITE) {
        error = LOGERR_WRITE;
    }

    return error;
}

static long logOpenFile(struct Log *log)
{
    if (log == NULL) {
	return ERR_INTERNAL;
    }

    if (log->log_TextFile == NULL && !log->log_OpenAttempted) {
        struct Project *pr = log->log_Project;
        struct MemPool *memPool = ProjectMemPool(pr);
        const char *filename = GetPrefsString(pr, PREFS_LOG_FILENAME);
        long error = OK;
        int i;

        log->log_TextFile = TextFileOpen(memPool, filename, "a", TFE_AUTO);
        if (log->log_TextFile != NULL && TextFileTell(log->log_TextFile) == 0) {
            for (i = 0; i < NumLogFields; i++) {
                error = logWriteFile(log, i > 0, LogFieldInfo[i].lai_Name,
                        i == NumLogFields - 1);
                if (error != OK) {
                    return error;
                }
            }
        }
    }

    if (log->log_TextFile == NULL) {
        return LOGERR_OPEN;
    }

    return OK;
}

bool LogActive(struct Log *log)
{
    bool active = FALSE;

    if (log) {
        active = GetPrefsValue(log->log_Project, PREFS_LOG_ACTIVE);
    }
    return active;
}

long SetLogActive(struct Log *log, bool active)
{
    long error = OK;

    if (log == NULL) {
        error = ERR_INTERNAL;
    } else {
        SetPrefsValue(log->log_Project, PREFS_LOG_ACTIVE, active);
    }

    return error;
}

bool LogToProject(struct Log *log)
{
    bool toProject = FALSE;

    if (log) {
        toProject = GetPrefsValue(log->log_Project, PREFS_LOG_TOPROJECT);
    }
    return toProject;
}

void SetLogToProject(struct Log *log, bool toProject)
{
    if (log) {
        SetPrefsValue(log->log_Project, PREFS_LOG_TOPROJECT, toProject);
    }
}

bool LogToFile(struct Log *log)
{
    bool toFile = FALSE;

    if (log) {
        toFile = GetPrefsValue(log->log_Project, PREFS_LOG_TOFILE);
    }
    return toFile;
}

void SetLogToFile(struct Log *log, bool toFile)
{
    if (log) {
        SetPrefsValue(log->log_Project, PREFS_LOG_TOFILE, toFile);
    }
}

const char *LogFilename(struct Log *log)
{
    const char *filename = NULL;

    if (log) {
        filename = GetPrefsString(log->log_Project, PREFS_LOG_FILENAME);
    }
    return filename;
}

bool SetLogFilename(struct Log *log, const char *filename)
{
    bool rc = FALSE;

    if (log != NULL) {
        rc = SetPrefsString(log->log_Project, PREFS_LOG_FILENAME, filename);
        logCloseFile(log);
    }
    return rc;
}

const char *LogLabel(struct Log *log)
{
    const char *label = NULL;

    if (log) {
	label = GetPrefsString(log->log_Project, PREFS_LOG_LABEL);
    }
    return label;
}

bool SetLogLabel(struct Log *log, const char *label)
{
    bool rc = FALSE;

    if (log != NULL) {
        rc = SetPrefsString(log->log_Project, PREFS_LOG_LABEL, label);
    }
    return rc;
}

#ifdef USE_LOG_WINDOW

#define LOG_WINDOW_DEFAULT_WIDTH 300

static const char *skipUnderscore(const char *s)
{
    if (s && s[0] == '_') {
        s++;
    }
    return s;
}

static long logCreateGuiWindow(struct Log *log)
{
    struct DisplayObject *windio;
    long error = logCreateTable(log);
    struct Table *tab;
    struct WindowData wd;
    struct WindowDimensions wds;
    int i;

    if (error != OK) {
        return error;
    }
    if (log->log_WindowObject != NULL) {
        return OK;
    }
    windio = AllocRootDisplayObject(log->log_Project);
    if (windio == NULL) {
        return ERR_MEM;
    }

    memset(&wd, 0, sizeof(wd));
    wd.wd_Title = GCS(MSG_Log);
    wd.wd_Id = MAKE_ID('_', 'L', 'O', 'G');
    DisplayChangeWindow(windio, &wd);

    if (!GetPrefsWindowDimensions(log->log_Project, PREFS_LOG_WINDOW, &wds)) {
        wds.wd_LeftEdge = -1;
        wds.wd_TopEdge = -1;
        wds.wd_Width = -1;
        wds.wd_Height = -1;
    }
    if (wds.wd_Width < 0) {
        wds.wd_Width = LOG_WINDOW_DEFAULT_WIDTH;
    }
    SetDisplayWindowDimensions(windio, &wds);

    tab = log->log_Table;
    if (!DisplayNewTable(tab, windio)) {
        error = ERR_MEM;
    } else {
        struct DisplayObject *dio = TableDisplayObject(tab);
        struct DisplayTableData dtd;

        memset(&dtd, 0, sizeof(dtd));
        dtd.dtd_Title = skipUnderscore(LOG_TABLE_NAME);
        dtd.dtd_Weight = 100;
        dtd.dtd_HasPanel = TRUE;
        dtd.dtd_PanelNumAll = TRUE;
        dtd.dtd_PanelFilter = TRUE;
        dtd.dtd_PanelArrows = TRUE;
        DisplayChangeTable(tab, &dtd);

        for (i = 0; i < NumLogFields; i++) {
            struct Field *field = log->log_Fields[i];
            struct DisplayFieldData dfd;

            if (!DisplayNewField(field, dio)) {
                error = ERR_MEM;
                break;
            }
            dio = FieldDisplayObject(field);
            memset(&dfd, 0, sizeof(dfd));
            dfd.dfd_Title = skipUnderscore(LogFieldInfo[i].lai_Name);
            dfd.dfd_Alignment = DAD_ALIGNMENT_LEFT;
            dfd.dfd_TabChain = TRUE;
            dfd.dfd_Enabled = DOD_ENABLED_NOTINITIAL;
            dfd.dfd_Weight = 100;
            DisplayChangeField(field, &dfd);
        }
    }

    if (error == OK && !BuildGuiWindow(windio)) {
        error = DPYERR_BUILDGUI;
    }
    if (error == OK) {
        log->log_WindowObject = windio;
    } else {
        FreeRootDisplayObject(windio);
        SetTableDisplayObject(tab, NULL);
        for (i = 0; i < NumLogFields; i++) {
            SetFieldDisplayObject(log->log_Fields[i], NULL);
        }
    }

    return error;
}

struct DisplayObject *LogWindowObject(struct Log *log)
{
    struct DisplayObject *dio = log? log->log_WindowObject : NULL;

    return dio;
}

long LogShowWindow(struct Log *log)
{
    long error = logCreateGuiWindow(log);

    if (error == OK) {
        struct DisplayObject *dio = log->log_WindowObject;
        GuiObject wi = DisplayObjectGuiObject(dio);

        GuiOpenWindow(wi);
        GuiWindowToFront(wi);
        GuiActivateWindow(wi);
    }

    return error;
}

void LogCloseWindow(struct Log *log)
{
    struct DisplayObject *dio = log? log->log_WindowObject : NULL;
    GuiObject wi = DisplayObjectGuiObject(dio);

    GuiCloseWindow(wi);
}

#endif

bool LogHasTable(struct Log *log)
{
    return (log != NULL && 
            (log->log_Table != NULL ||
             FindTableName(log->log_Project, LOG_TABLE_NAME) != NULL));
}

struct Table *LogTable(struct Log *log)
{
    struct Table *tab = NULL;

    if (log) {
        logCreateTable(log);
        tab = log->log_Table;
    }
    return tab;
}

struct Field *LogField(struct Log *log, int logFieldId)
{
    struct Field *field = NULL;

    if (log && logFieldId >= 0 && logFieldId < NumLogFields) {
        logCreateTable(log);
        field =log->log_Fields[logFieldId];
    }
    return field;
}

static char *dataToString(struct Field *field, const struct DataInfo *di)
{
    short type = field == NULL? TYPE_REFERENCE : FieldType(field);
    char *str = NULL;

    if (di == NULL) {
        return NULL;
    }

    if (!(di->di_Flags & DIF_NIL)) {
        const union Data *d = &di->di_Contents;
        struct DisplayObject *dio = FieldDisplayObject(field);
        struct Project *pr = TableProject(FieldTable(field));

        switch (type)
        {
            case TYPE_STRING:
                str = (char *)EmptyNullStr(d->string);
                break;

            case TYPE_INTEGER:
                str = IntToStr(d->integer);
                break;

            case TYPE_REAL:
                str = DoubleToStr(d->real, DisplayFieldNumDecimals(dio),
                        (short)GetPrefsValue(pr, PREFS_REALFORMAT));
                break;

            case TYPE_BOOL:
                str = "TRUE";
                break;

            case TYPE_CHOICE:
            {
                char **labels = FieldChoiceLabels(field);

                if ((ulong)d->choice < NumLabels(labels)) {
                    str = labels[d->choice];
                } else {
                    str = IntToStr(d->choice);
                }
                break;
            }

            case TYPE_DATE:
                str = DateToString(d->date, 
                        (short)GetPrefsValue(pr, PREFS_DATEFORMAT));
                break;

            case TYPE_TIME:
                str = TimeToString(di->di_Contents.time, 
                        DisplayFieldFormat(dio));
                break;

            case TYPE_MEMO:
                str = (char *)EmptyNullStr(d->mi);
                break;

            case TYPE_REFERENCE:
                str = IntToStr(RecordNum(d->rec));
                break;

            default:
                str = GES(ERR_INTERNAL);
                break;
        }
    } else {
        switch (type) {
            case TYPE_BOOL:
            case TYPE_REFERENCE:
                str = "NIL";
                break;

            default:
                str = (char *)FieldNilString(field);
                if (str == NULL) {
                    str = GES(ERR_INTERNAL);
                }
                break;
        }
    }

    return str;
}

static void logAction(struct Project *pr, const char *action,
        bool allowToProject, struct Table *tab, struct Record *rec,
        struct Field *field, const struct DataInfo *oldValue,
        const struct DataInfo *newValue)
{
    if ((tab == NULL || (!TableIsSystem(tab) && TableLogChanges(tab))) &&
            (field == NULL || FieldLogChanges(field))) {
        struct Log *log = ProjectLog(pr);
        long projectError = OK;
        long fileError = OK;

        if (log == NULL) {
            projectError = fileError = ERR_INTERNAL;
        } else if (LogActive(log)) {
            struct MemPool *memPool = ProjectMemPool(pr);
            bool toProject = allowToProject && 
                GetPrefsValue(pr, PREFS_LOG_TOPROJECT);
            bool toFile = GetPrefsValue(pr, PREFS_LOG_TOFILE);
            struct Record *logRec = NULL;
            union Data data;
            bool free;
           
            if (projectError == OK && toProject) {
                projectError = logCreateTable(log);
            }
            if (fileError == OK && toFile) {
                fileError = logOpenFile(log);
            }

            /* Create new record */
            if (projectError == OK && toProject) {
                projectError = NewRecord(log->log_Table, NULL, ADDRECORD_TAIL,
                        TRUE, &logRec);
            }

            /* Set label */
            if (projectError == OK || fileError == OK) {
                struct Program *prg = ProjectProgram(pr);
                struct UserFunction *uf = FindUserFunctionName(prg, logLabel);
                struct Element *elem;
                bool endProgram = FALSE;
                const char *label = NULL;
                char *mem = NULL;

                if (uf != NULL) {
                    if (!ProgramRunning()) {
                        SetTableProgramRecords(pr);
                        ProgramBegin(prg, TRUE);
                        endProgram = TRUE;
                    }
                    elem = ProgramCall(prg, uf, NULL, FALSE);
                    if (elem != NULL) {
                        label = mem = SimpleElementToString(elem, NULL, 
                                memPool, &free);
                        if (!free) {
                            mem = NULL;
                        }
                    } else {
                        label = ProgramError(NULL, NULL, NULL);
                    }
                } else {
                    label = GetPrefsString(pr, PREFS_LOG_LABEL);
                }
                data.string = (char *)label;
                if (projectError == OK && toProject) {
                    projectError = RecordStoreData(logRec,
                            log->log_Fields[LA_LABEL], &data, FALSE);
                }
                if (fileError == OK && toFile) {
                    fileError = logWriteFile(log, FALSE, data.string, FALSE);
                }
                FreeString(memPool, mem);

                if (endProgram) {
                    long line, col;
                    const char *filename;
                    const char *errstr;

                    ProgramEnd();
                    errstr = ProgramError(&line, &col, &filename);
                    if (errstr != NULL) {
                       if (errstr != EXECUTE_ABORT) {
                           ProgramEditorShowError(pr, TRUE, errstr, 
                                   line, col, filename);
                       }
                    }
                }
            }

            /* Set date */
            if (projectError == OK || fileError == OK) {
                data.date = DateToday();
                if (projectError == OK && toProject) {
                    projectError = RecordStoreData(logRec,
                            log->log_Fields[LA_DATE], &data, FALSE);
                }
                if (fileError == OK && toFile) {
                    fileError = logWriteFile(log, TRUE, DateToString(data.date, 
                                (short)GetPrefsValue(pr, PREFS_DATEFORMAT)),
                            FALSE);
                }
            }

            /* Set time */
            if (projectError == OK || fileError == OK) {
                data.time = TimeNow();
                if (projectError == OK && toProject) {
                    projectError = RecordStoreData(logRec,
                            log->log_Fields[LA_TIME], &data, FALSE);
                }
                if (fileError == OK && toFile) {
                    fileError = logWriteFile(log, TRUE, TimeToString(data.time,
                                FORMAT_HOURMINSEC), FALSE);
                }
            }

            /* Set action */
            if (projectError == OK || fileError == OK) {
                data.string = (char *)action;
                if (projectError == OK && toProject) {
                    projectError = RecordStoreData(logRec,
                            log->log_Fields[LA_ACTION], &data, FALSE);
                }
                if (fileError == OK && toFile) {
                    fileError = logWriteFile(log, TRUE, data.string, FALSE);
                }
            }

            /* Set table */
            if (projectError == OK || fileError == OK) {
                data.string = (char *)TableName(tab);
                if (projectError == OK && toProject) {
                    projectError = RecordStoreData(logRec,
                            log->log_Fields[LA_TABLE], &data, FALSE);
                }
                if (fileError == OK && toFile) {
                    fileError = logWriteFile(log, TRUE, data.string, FALSE);
                }
            }

            /* Set record */
            if (projectError == OK || fileError == OK) {
                data.integer = RecordNum(rec);
                if (projectError == OK && toProject) {
                    projectError = RecordStoreData(logRec,
                            log->log_Fields[LA_RECORD], &data, tab == NULL);
                }
                if (fileError == OK && toFile) {
                    fileError = logWriteFile(log, TRUE,
                            (tab == NULL)? NULL : IntToStr(data.integer),
                            FALSE);
                }
            }

            /* Set description */
            if (projectError == OK || fileError == OK) {
                struct DisplayObject *dio = TableDisplayObject(tab);
                const struct RecordDescriptionSpec *rds = 
                    DisplayTableRecordDescriptionSpec(dio);
                struct RecordDescription *rd;
                long error;
                
                if (rds != NULL) {
                    rd = AllocRecordDescription(tab, rds);
                    if (rd == NULL) {
                        error = ERR_MEM;
                        if (projectError == OK) {
                            projectError = error;
                        }
                        if (fileError == OK) {
                            fileError = error;
                        }
                    } else {
                        RecordLock(rec);
                        RecordDescriptionSetRecord(rd, rec);
                        data.string = RecordDescriptionText(rd, memPool, &free);
                        error = RecordDescriptionError(rd);
                        if (projectError == OK) {
                            projectError = error;
                        }
                        if (fileError == OK) {
                            fileError = error;
                        }
                        if (projectError == OK && toProject) {
                            projectError = RecordStoreData(logRec, 
                                    log->log_Fields[LA_DESCRIPTION],
                                    &data, FALSE);
                        }
                        if (fileError == OK && toFile) {
                            fileError = logWriteFile(log, TRUE, data.string,
                                    FALSE);
                        }
                        if (free) {
                            FreeString(memPool, data.string);
                        }
                        RecordUnlock(rec);
                        FreeRecordDescription(rd);
                    }
                } else if (fileError == OK && toFile) {
                    fileError = logWriteFile(log, TRUE, NULL, FALSE);
                }
            }

            /* Set field */
            if (projectError == OK || fileError == OK) {
                data.string = (char *)FieldName(field);
                if (projectError == OK && toProject) {
                    projectError = RecordStoreData(logRec,
                            log->log_Fields[LA_FIELD], &data, FALSE);
                }
                if (fileError == OK && toFile) {
                    fileError = logWriteFile(log, TRUE, data.string, FALSE);
                }
            }

            /* Set old value */
            if (projectError == OK || fileError == OK) {
                data.string = dataToString(field, oldValue);
                if (projectError == OK && toProject) {
                    projectError = RecordStoreData(logRec,
                            log->log_Fields[LA_OLD], &data, FALSE);
                }
                if (fileError == OK && toFile) {
                    fileError = logWriteFile(log, TRUE, data.string, FALSE);
                }
            }

            /* Set new value */
            if (projectError == OK || fileError == OK) {
                data.string = dataToString(field, newValue);
                if (projectError == OK && toProject) {
                    projectError = RecordStoreData(logRec,
                            log->log_Fields[LA_NEW], &data, FALSE);
                }
                if (fileError == OK && toFile) {
                    fileError = logWriteFile(log, TRUE, data.string, TRUE);
                }
            }
        }

        if (fileError == LOGERR_OPEN) {
            if (log->log_OpenAttempted) {
                fileError = OK;     /* silent error */
            }
            log->log_OpenAttempted = TRUE;
        }
        if (projectError != OK) {
            Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_LogError), GES(projectError));
        }
        if (fileError != OK && fileError != projectError) {
            Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_LogError), GES(fileError));
        }
    }
}

void LogBeginTransaction(struct Project *pr)
{
    logAction(pr, logActionBeginTransaction, FALSE,
            NULL, NULL, NULL, NULL, NULL);
}

void LogRollback(struct Project *pr)
{
    logAction(pr, logActionRollback, FALSE, NULL, NULL, NULL, NULL, NULL);
    logCloseFile(ProjectLog(pr));
}

void LogCommit(struct Project *pr)
{
    logAction(pr, logActionCommit, FALSE, NULL, NULL, NULL, NULL, NULL);
    logCloseFile(ProjectLog(pr));
}

void LogNewRecord(struct Table *tab, struct Record *rec, struct Record *init)
{
    struct DataInfo oldValue, newValue;
    struct Record *pred = NULL;

    if (rec != NULL && PRED(PRED(rec)) != NULL) {
        pred = PRED(rec);
    }

    oldValue.di_Flags = 0;
    oldValue.di_Private = 0;
    oldValue.di_Contents.rec = init;
    newValue.di_Flags = 0;
    newValue.di_Private = 0;
    newValue.di_Contents.rec = pred;

    logAction(TableProject(tab), logActionNewRecord, TRUE, tab, rec,
            NULL, &oldValue, &newValue);
}

void LogDeleteRecord(struct Table *tab, struct Record *rec)
{
    logAction(TableProject(tab), logActionDeleteRecord, TRUE, tab, rec,
            NULL, NULL, NULL);
}

void LogStoreData(struct Record *rec, struct Field *field,
    const struct DataInfo *oldValue, const struct DataInfo *newValue)
{
    struct Table *tab = FieldTable(field);

    logAction(TableProject(tab), logActionChangeValue, TRUE, tab, rec,
            field, oldValue, newValue);
}

#ifdef ENABLE_LOG_APPLY_CHANGES

static const char *getChangeFieldString(struct Log *log, struct Record *change,
        int fieldId)
{
    const char *str = NULL;
    struct DataInfo *di, di_buf;
    long error;

    error = RecordDataInfo(change, log->log_Fields[fieldId], &di, &di_buf);
    if (error == OK) {
        str = di->di_Contents.string;
    }
    return str;
}

static struct Table *getChangeTable(struct Log *log, struct Record *change)
{
    const char *name = getChangeFieldString(log, change, LA_TABLE);
    struct Table *tab = FindTableName(log->log_Project, name);

    return tab;
}

static long getChangeRecordNumber(struct Log *log, struct Record *change)
{
    long num = 0;
    struct DataInfo *di, di_buf;
    long error;

    error = RecordDataInfo(change, log->log_Fields[LA_RECORD], &di, &di_buf);
    if (error == OK && !(di->di_Flags & DIF_NIL)) {
        num = di->di_Contents.integer;
    }
    return num;
}

static struct Record *getChangeRecord(struct Log *log, struct Record *change,
        struct Table *tab)
{
    long num = getChangeRecordNumber(log, change);
    struct List *recs = TableRecords(tab);
    struct Record *rec = FindNodeNum(recs, num);

    return rec;
}

static const char *logApplyNewRecord(struct Log *log, struct Record *change)
{
    struct Project *pr = log->log_Project;
    struct Table *tab = getChangeTable(log, change);
    struct List *recs = TableRecords(tab);
    long numRecs = NumNodes(recs);
    long newRecNum = getChangeRecordNumber(log, change);
    long initRecNum = 0, predRecNum = 0;
    struct Record *init, *pred, *rec;
    long error;

    if (tab == NULL) {
        return GCS(MSG_Log_ApplyChangeError_UnknownTable);
    }
    if (newRecNum <= 0 || newRecNum > numRecs + 1) {
        return GCS(MSG_Log_ApplyChangeError_InvalidRecordNumber);
    }

    StrToInt(getChangeFieldString(log, change, LA_OLD), &initRecNum);
    if (initRecNum > newRecNum) {
        initRecNum--;   /* new record got inserted before the init record */
    }
    if (initRecNum < 0 || initRecNum > numRecs) {
        return GCS(MSG_Log_ApplyChangeError_InvalidOldValue);
    }
    init = FindNodeNum(recs, initRecNum);

    StrToInt(getChangeFieldString(log, change, LA_NEW), &predRecNum);
    if (predRecNum < 0 || predRecNum >= newRecNum) {
        return GCS(MSG_Log_ApplyChangeError_InvalidNewValue);
    }
    pred = FindNodeNum(recs, predRecNum);

    error = NewRecord(tab, init, pred, TRUE, &rec);
    if (error == OK) {
        error = ReorderRecords(tab);
    }

    if (TableCountChanges(tab)) {
        ProjectChange(pr, FALSE);
    }

    if (error != OK) {
        return GES(error);
    }
    return NULL;
}

static const char *logApplyDeleteRecord(struct Log *log, struct Record *change)
{
    struct Project *pr = log->log_Project;
    struct Table *tab = getChangeTable(log, change);
    struct Record *rec = getChangeRecord(log, change, tab);
    long error;

    if (tab == NULL) {
        return GCS(MSG_Log_ApplyChangeError_UnknownTable);
    }
    if (rec == NULL) {
        return GCS(MSG_Log_ApplyChangeError_InvalidRecordNumber);
    }

    error = DeleteRecord(tab, rec);

    if (TableCountChanges(tab)) {
        ProjectChange(pr, FALSE);
    }

    if (error != OK) {
        return GES(error);
    }
    return NULL;
}

static long matchOldValue(struct Record *rec, struct Field *field,
        const char *old, bool *matches, bool *hasChoiceBug) 
{
    struct DataInfo *di, di_buf;
    char *text = NULL;
    long error;
    long num;

    RecordLock(rec);
    error = RecordDataInfo(rec, field, &di, &di_buf);
    if (error == OK) {
        text = dataToString(field, di);
        *matches = !strcmp(EmptyNullStr(text), EmptyNullStr(old));
        if (!*matches && FieldType(field) ==  TYPE_CHOICE) {
            if (StrToInt(old, &num) && (num >> 16) == di->di_Contents.choice) {
                *matches = TRUE;
                *hasChoiceBug = TRUE;
            }
        }
    }
    RecordUnlock(rec);
    return error;
}

static const char *logApplyChangeValue(struct Log *log, struct Record *change)
{
    struct Project *pr = log->log_Project;
    struct Table *tab = getChangeTable(log, change);
    struct Record *rec = getChangeRecord(log, change, tab);
    const char *name = getChangeFieldString(log, change, LA_FIELD);
    struct Field *field = FindFieldName(tab, name);
    struct DisplayObject *dio = FieldDisplayObject(field);
    short type = FieldType(field);
    const char *old = getChangeFieldString(log, change, LA_OLD);
    const char *new = getChangeFieldString(log, change, LA_NEW);
    bool matches = FALSE;
    union Data data;
    bool nil = FALSE;
    bool hasChoiceBug = FALSE;
    long error;

    if (tab == NULL) {
        return GCS(MSG_Log_ApplyChangeError_UnknownTable);
    }
    if (rec == NULL) {
        return GCS(MSG_Log_ApplyChangeError_InvalidRecordNumber);
    }
    if (field == NULL) {
        return GCS(MSG_Log_ApplyChangeError_UnknownField);
    }

    error = matchOldValue(rec, field, old, &matches, &hasChoiceBug);
    if (error != OK) {
        return GES(error);
    }
    if (matches) {
        /* NoOp */
    } else if (type == TYPE_REFERENCE && FieldReferenceFilter(field)) {
        /* Reference filter presets field to the current GUI record */
        matches = TRUE;
    }

    if (!matches) {
        return GCS(MSG_Log_ApplyChangeError_OldValueDoesNotMatch);
    }

    if (new == NULL) {
        return GCS(MSG_Log_ApplyChangeError_InvalidNewValue);
    }

    switch (FieldType(field)) {
        case TYPE_BOOL:
            nil = (new[0] == 0 || !strcmp(new, "NIL"));
            break;

        case TYPE_INTEGER:
            nil = !StrToInt(new, &data.integer);
            break;
            
        case TYPE_CHOICE:
        {
            char **labels = FieldChoiceLabels(field);
            long i;

            if (labels == NULL) {
                return GES(ERR_INTERNAL);
            }
            for (i = 0; labels[i] != NULL; i++) {
                if (!strcmp(labels[i], new)) {
                    break;
                }
            }
            if (labels[i] == NULL) {
                if (!StrToInt(new, &i)) {
                    return GCS(MSG_Log_ApplyChangeError_InvalidNewValue);
                }
                if (hasChoiceBug) {
                    i = i >> 16;
                }
            }
            data.choice = i;
            break;
        }

        case TYPE_REAL:
            nil = !StrToDouble(new, &data.real);
            break;

        case TYPE_STRING:
            data.string = (char *)new;
            break;

        case TYPE_MEMO:
            data.mi = (char *)new;
            break;

        case TYPE_DATE:
            nil = !StringToDate(new, &data.date);
            break;

        case TYPE_TIME:
            nil = !StringToTime(new, DisplayFieldFormat(dio), &data.time);
            break;

        case TYPE_REFERENCE:
        {
            struct Table *reftab = FieldReferenceTable(field);
            long recnum;

            if (!strcmp(new, "NIL")) {
                data.rec = NULL;
            } else if (!StrToInt(new, &recnum)) {
                return GCS(MSG_Log_ApplyChangeError_InvalidNewValue);
            } else {
                data.rec = FindNodeNum(TableRecords(reftab), recnum);
            }
            nil = data.rec == NULL;
            break;
        }

        default:
            return GCS(MSG_Log_ApplyChangeError_InvalidField);
            break;
    }

    error = RecordStoreData(rec, field, &data, nil);
    if (error == OK) {
        if (FieldCountChanges(field) && TableCountChanges(tab)) {
            ProjectChange(pr, FALSE);
        }
        error = ReorderRecords(tab);
    }
    GuiDisplayPanel(tab);
    GuiUpdateField(field, rec);

    if (error != OK) {
        return GES(error);
    }
    return NULL;
}

static bool logChangeIsAction(struct Log *log, struct Record *change,
        const char *action)
{
    bool match = FALSE;
    const char *str;

    if (log == NULL || log->log_Table == NULL) {
        return FALSE;
    }

    RecordLock(change);
    str = getChangeFieldString(log, change, LA_ACTION);
    match = !strcmp(EmptyNullStr(str), action);
    RecordUnlock(change);

    return match;
}

bool LogChangeIsBeginTransaction(struct Log *log, struct Record *change)
{
    return logChangeIsAction(log, change, logActionBeginTransaction);
}

bool LogChangeIsRollback(struct Log *log, struct Record *change)
{
    return logChangeIsAction(log, change, logActionRollback);
}

bool LogChangeIsCommit(struct Log *log, struct Record *change)
{
    return logChangeIsAction(log, change, logActionCommit);
}

const char *LogApplyChange(struct Log *log, struct Record *change)
{
    struct DataInfo *di, di_buf;
    const char *errstr = NULL;
    long error;

    if (log == NULL) {
        return GES(ERR_INTERNAL);
    }

    error = ObtainProjectChangeLock(log->log_Project);
    if (error != OK) {
        return GES(error);
    }

    error = logCreateTable(log);
    if (error != OK) {
        return GES(error);
    }

    RecordLock(change);
    error = RecordDataInfo(change, log->log_Fields[LA_ACTION], &di, &di_buf);
    if (error != OK) {
        errstr = GES(error);
    } else if (!strcmp(di->di_Contents.string, logActionNewRecord)) {
        errstr = logApplyNewRecord(log, change);
    } else if (!strcmp(di->di_Contents.string, logActionDeleteRecord)) {
        errstr = logApplyDeleteRecord(log, change);
    } else if (!strcmp(di->di_Contents.string, logActionChangeValue)) {
        errstr = logApplyChangeValue(log, change);
    } else {
        errstr = GCS(MSG_Log_ApplyChangeError_UnknownAction);
    }
    RecordUnlock(change);

    return errstr;
}

#endif

