/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** log.h
**
** Creation date: 03.05.2012
** Last modified: 13.04.2022
*/

#ifndef LOG_H
#define LOG_H

#include "system.h"

struct Project;
struct Log;
struct Table;
struct Field;
struct Record;
struct DataInfo;

#ifdef __cplusplus
extern "C" {
#endif

enum {
    LA_LABEL,
    LA_DATE,
    LA_TIME,
    LA_ACTION,
    LA_TABLE,
    LA_RECORD,
    LA_DESCRIPTION,
    LA_FIELD,
    LA_OLD,
    LA_NEW,
    NumLogFields
};

void FreeLog(struct Log *log);
/* 
** Frees log data structure
*/

struct Log *AllocLog(struct Project *pr);
/*
** Allocates a new log data structure for a given project
*/

long ClearLog(struct Log *log, bool reset);
/*
** Deletes all entries of a log.  
** If reset is TRUE then also all settings are reset.
*/

void ReleaseLogFile(struct Log *log);
/*
** Closes the log file if one has been opened.
*/

bool LogActive(struct Log *log);
/*
** Whether logging is active.
*/

long SetLogActive(struct Log *log, bool active);
/*
** Starts/stops logging. Returns error code or OK.
*/

bool LogToProject(struct Log *log);
/*
** Whether logging is to the project's _Log table.
*/

void SetLogToProject(struct Log *log, bool on);
/*
** Sets logging to the project's _Log table.
*/

bool LogToFile(struct Log *log);
/*
** Whether logging is to a file.
*/

void SetLogToFile(struct Log *log, bool on);
/*
** Sets logging to a file.
*/

const char *LogFilename(struct Log *log);
/*
** Returns the log filename.
*/

bool SetLogFilename(struct Log *log, const char *filename);
/*
** Sets the log filename.  Returns TRUE on success.
*/

const char *LogLabel(struct Log *log);
/*
** Returns the current label used when adding new log entries
*/

bool SetLogLabel(struct Log *log, const char *label);
/*
** Sets a new label.  Returns TRUE on success.
*/

#ifdef USE_LOG_WINDOW
struct DisplayObject *LogWindowObject(struct Log *log);
/*
** Returns the window object of the log table.
** Can be NULL if the window has not been created yet.
*/

long LogShowWindow(struct Log *log);
/*
** Shows the window object of the log table.
** Returns OK or an error code.
*/

void LogCloseWindow(struct Log *log);
/*
** Closes the window of the log table if it was open
*/

#endif

bool LogHasTable(struct Log *log);
/*
** Whether the _Log table exists.
*/

struct Table *LogTable(struct Log *log);
/*
** Table holding log entries
*/

struct Field *LogField(struct Log *log, int logFieldId);
/*
** Field in log table.  The logFieldId must be one of the LA_* enums.
*/

void LogBeginTransaction(struct Project *pr);
/*
** Log the beginning of a transaction.
*/

void LogRollback(struct Project *pr);
/*
** Log the rollback of a transaction.
*/

void LogCommit(struct Project *pr);
/*
** Log the commit of a transaction.
*/

void LogNewRecord(struct Table *tab, struct Record *rec, struct Record *init);
/*
** Log the addition of a new record to a table.
*/

void LogDeleteRecord(struct Table *tab, struct Record *rec);
/*
** Log the deletion of a record in a table.
*/

void LogStoreData(struct Record *rec, struct Field *field,
    const struct DataInfo *oldValue, const struct DataInfo *newValue);
/*
** Log changes to an field in a record.
*/

#ifdef ENABLE_LOG_APPLY_CHANGES

bool LogChangeIsBeginTransaction(struct Log *log, struct Record *change);
bool LogChangeIsRollback(struct Log *log, struct Record *change);
bool LogChangeIsCommit(struct Log *log, struct Record *change);
/*
** Check if a change has a certain action.
*/

const char *LogApplyChange(struct Log *log, struct Record *change);
/*
** Applies a change to the current database.
** Returns error string or NULL on success.
*/

#endif

#ifdef __cplusplus
}
#endif

#endif

