/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** main.c
**
** Creation date: 07.09.1994
** Last modified: 03.01.2022
*/

#include "system.h"
#include "main.h"
#include "server.h"
#include "project.h"
#include "structureeditor.h"
#include "menuproject.h"
#include "requesters.h"
#include "preferences.h"
#include "codeset.h"
#include "appstrings_cat.h"
#include "lang/program.h"
#ifdef AMIGA
#include <workbench/startup.h>
#endif

/*
** Global variables
*/

struct List ProjectList;
long OSVersion = -1, OSRevision = -1;
bool debug = FALSE;
static bool createEmptyProject = TRUE;
static char *progName = APP_NAME;

static void CloseAll(void)
{
    struct Project *pr;

    while((pr = (struct Project *)REMHEAD(&ProjectList)) != NULL)
        FreeProject(pr);

    ProjectSetCurrentDir(NULL);           /* revert back to original CWD */
    ProgramFreeResources();
    ServerExit();
    PrefsExit();
    CloseGui();
}

void fail(const char *s, int exitCode)
{
    if (s) {
#ifdef AMIGA
    	if (IntuitionBase) {
	    struct EasyStruct failES = {
		sizeof (struct EasyStruct),
		0, (STRPTR)APP_NAME, (STRPTR)"%s", (STRPTR)"Quit"
	    };
	    EasyRequest(NULL, &failES, NULL, (IPTR)s);
	} else 
#endif
	{
	    puts(s);
	}
    }
   CloseAll();
   exit(exitCode);
}

static void OpenAll(int *pargc, char **argv)
{
    const char *error;

    error = OpenGui(pargc, argv);
    if (error) {
	fail(error, 5);
    }

    PrefsInit();
    ServerInit();
}

static bool openNewProject(const char *name)
{
    struct Project *pr;

    if (FindProjectName(name) != NULL) {
	return TRUE;          /* do not open twice */
    }

    pr  = AllocProject(TRUE);
    if (pr == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_OpenProject));
        return FALSE;
    }
    ADDTAIL(&ProjectList, pr);

    MenuProjectOpen(pr, name);
    if (ProjectName(pr) == NULL) {              /* open failed? */
        REMOVE(pr);
        FreeProject(pr);
	return TRUE;                            /* continue */
    }
    
    return TRUE;
}

static void ParseCommandLineOptions(int argc, char **argv)
{
    int i;

    for(i=1; i < argc; i++)
    {
	char *s = argv[i];
	
	if(*s == '-')
	{
	    for(s++; *s; s++)
	    {
		switch(*s)
		{
		case 'd':
		    debug = TRUE;
		    break;
		    
		case 'n':
		    createEmptyProject = FALSE;
		    break;
		    
		default:
		    mainUsage();
		    fail(NULL, 5);
		    break;
		}
	    }
	}
    }
}

static void LoadCommandLineProjects(int argc, char **argv)
{
    const char *s;
    long i;

    if (argc > 0) {
        for (i = 1; i < argc; i++) {
            s = argv[i];
            if (*s != '-' && !openNewProject(s)) {
                break;
            }
        }
    }
#ifdef AMIGA
    else if (argv != NULL) {
        char name[MAX_PATHNAME];
        struct WBStartup *startup = (struct WBStartup *)argv;
        struct WBArg *arg = startup->sm_ArgList;

        for (i = 1; i < startup->sm_NumArgs; i++) {
            if (arg[i].wa_Name[0] == 0) {
                continue;
            }

            if (!NameFromLock(arg[i].wa_Lock, (STRPTR)name, sizeof(name))) {
                continue;
            }
            if (!AddPart((STRPTR)name, (CONST_STRPTR)arg[i].wa_Name,
                        sizeof(name))) {
                continue;
            }

            if (!openNewProject(name)) {
                break;
            }
        }
    }
#endif
}

void mainUsage(void)
{
    DPRINTF(("Usage: %s [project1 [project2 ...]]\n", progName));
    DPRINTF(("       %s -n\n", progName));
}

bool mainReallyQuit(void)
{
    struct Project *pr;
    long changed_projects = 0, all_projects = 0;
    long num_connections = ServerNumConnections();

    FORNODE(pr, &ProjectList) {
	all_projects++;
        if (ProjectChanges(pr) != 0) {
            changed_projects++;
	}
    }

    if (all_projects == 0 && num_connections == 0) {
	return TRUE;			/* all is gone */
    }
    if (changed_projects > 0 && num_connections > 0) {
	if (Ask(AT_WARN, GCS(MSG_AskQuitAnswers), 
		    GCS(MSG_AskChangesConnectionsQuit), 
		    changed_projects, num_connections)) {
	    return TRUE;
	}
    } else if (changed_projects > 0) {
	if (Ask(AT_WARN, GCS(MSG_AskQuitAnswers), 
		    GCS(MSG_AskChangesQuit), changed_projects)) {
	    return TRUE;
	}
    } else if (num_connections > 0) {
	if (Ask(AT_WARN, GCS(MSG_AskQuitAnswers), 
		    GCS(MSG_AskConnectionsQuit), num_connections)) {
	    return TRUE;
	}
    } else if (!GetPrefsValue(NULL, PREFS_CONFIRMQUIT)) {
	return TRUE;
    } else if (Ask(AT_WARN, GCS(MSG_AskQuitAnswers), GCS(MSG_ConfirmQuit))) {
	return TRUE;
    }

    return FALSE;
}

bool mainInit(int *pargc, char **argv)
{
    bool ok = TRUE;
    struct Project *pr;

    NewList(&ProjectList);
    if (*pargc > 0) {
	progName = argv[0];
    }

    OpenAll(pargc, argv);
    ParseCommandLineOptions(*pargc, argv);
    LoadPrefs();
    LoadCommandLineProjects(*pargc, argv);

    if (createEmptyProject && SUCC(HEAD(&ProjectList)) == NULL) {
	/* 
	** No projects on command line.  Create empty project.
	*/
        pr = MenuProjectNew(NULL, TRUE);
        if (pr == NULL) {
	    ok = FALSE;
        }
    }

    return ok;
}

void mainExit(void)
{
    CloseAll();
}

#ifndef NO_MAIN

int main(int argc, char **argv)
{
    bool ok;

    ok = mainInit(&argc, argv);
    if(ok)
    {
	GuiMainLoop(mainReallyQuit);
    }

    mainExit();
    return 0;
}


#if (defined WINDOWS) && (defined _MSC_VER)

// MSVC wants us to start in WinMain.  
// Fortunatelly it defines argc and argv externally as used below

#include <windows.h>

int __stdcall WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR lpCmd, int nShow)
{
    extern int __argc;          // defined somewhere in mscvrt.lib
    extern char **__argv;

    return main(__argc, __argv);
}

#endif

#endif
