/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** memory.c
**
** Creation date: 24.12.1994
** Last modified: 25.01.2009
*/

#include "system.h"
#include "memory.h"
#include "record.h"

#ifdef MEMORY_DEBUG

void __inline PoolFreeMem(struct MemPool *memPool, void *ptr, size_t size)
{
    if (memPool != NULL && ptr != NULL && size > 0) {
        free(ptr);
    }
}

void __inline *PoolAllocMem(struct MemPool *memPool, size_t size)
{
    void *ptr = NULL;

    if (memPool && size > 0) {
        ptr = malloc(size);
        if (ptr == NULL) {
            SwapRecords();
            ptr = malloc(size);
        }
    }
    return ptr;
}

#else

void __inline PoolFreeMem(struct MemPool *memPool, void *ptr, size_t size)
{
    if (memPool != NULL && ptr != NULL && size > 0) {
        FreePooled(memPool, ptr, size);
    }
}

void __inline *PoolAllocMem(struct MemPool *memPool, size_t size)
{
    void *ptr = NULL;

    if (memPool && size > 0) {
        ptr =AllocPooled(memPool, size);
        if (ptr == NULL) {
            SwapRecords();
            ptr =AllocPooled(memPool, size);
        }
    }
    return ptr;
}
#endif

void *PoolAllocMemZero(struct MemPool *memPool, size_t size)
{
    void *ptr = PoolAllocMem(memPool, size);

    if (ptr != NULL) {
        memset(ptr, 0, size);
    }
    return ptr;
}

void *PoolReAllocMem(struct MemPool *memPool, void *mem, 
        size_t oldsize, size_t newsize)
{
    void *new = PoolAllocMem(memPool, newsize);

    if (new == NULL) {
        return NULL;
    }
    memcpy(new, mem, MIN(oldsize, newsize));
    PoolFreeMem(memPool, mem, oldsize);

    return new;
}

char *AllocString(struct MemPool *memPool, const char *orig)
{
    char *result = NULL;

    if (orig) {
	result = PoolAllocMem(memPool, strlen(orig) + 1);
	if (result) {
	    strcpy(result, orig);
        }
    }
    return result;
}

char *AllocStringN(struct MemPool *memPool, const char *orig, size_t maxlen)
{
    char *result = NULL;
    size_t len;

    if (orig) {
        len = strlen(orig);
        if (len < maxlen) {
            maxlen = len;
        }
	result = PoolAllocMem(memPool, maxlen + 1);
	if (result) {
            memcpy(result, orig, maxlen);
            result[maxlen] = 0;
        }
    }
    return result;
}

void FreeString(struct MemPool *memPool, char *string)
{
    if (string) {
	PoolFreeMem(memPool, string, strlen(string) + 1);
    }
}


