/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** menueditor.c
**
** Functions for menu items in menu Editor
**
** Creation date: 09.02.1996
** Last modified: 27.01.2009
*/

#include "system.h"
#include "project.h"
#include "gui.h"
#include "preferences.h"
#include "menueditor.h"
#include "requesters.h"
#include "memory.h"
#include "strhandle.h"
#include "textfile.h"
#include "guidosutil.h"
#include "dosutil.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"

static void loadTextHelp(struct Project *pr, GuiObject edit, char *filename)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    long error;
    char *str;

    ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
    str = TextFileGetContentsAskWarnings(edit, memPool, filename, &error);
    if (error && error != ERR_CANCEL) {
        Ask2(edit, NULL, AT_ERROR, GCS(MSG_StarOk), GES(error), filename);
    } else if (str != NULL) {
        GuiSetEditorText(edit, str);
        FreeString(memPool, str);
    }
}

void MenuEditorLoadText(GuiObject edit, struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    char *filename;

    GuiSetSleep(TRUE);
    if (pr != NULL && GuiIsEditorObject(edit)) {
        filename = GetFilename(pr, memPool, GCS(MSG_OpenText), GCS(MSG_Open), 
                NULL, PREFS_TEXTEDITOR_FILENAME, FALSE);
        if (filename != NULL) {
            loadTextHelp(pr, edit, filename);
            FreeString(memPool, filename);
            GuiStoreEdits();
        }
    }
    GuiSetSleep(FALSE);
}

static void saveTextHelp(struct Project *pr, GuiObject edit, char *filename)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    long error;
    char *text;

    ProjectSetCurrentDir(pr);           /* make sure we have correct CWD */
    if (FileStat(filename) >= 0 &&      /* file exists? */
            Ask2(edit, NULL, AT_WARN, GCS(MSG_AskReplaceAnswers), 
                GCS(MSG_AskReplaceFile), filename) == 0) {
        return;
    }

    text = GuiEditorText(edit, memPool);
    if (text == NULL) {
        Ask2(edit, NULL, AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
        return;
    }

    TextFileSetContentsAskWarnings(edit, memPool, filename, text, &error);
    if (error && error != ERR_CANCEL) {
        Ask2(edit, NULL, AT_ERROR, GCS(MSG_StarOk), GES(error), filename);
    }
    FreeString(memPool, text);
}

void MenuEditorSaveText(GuiObject edit, struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    char *filename;

    GuiSetSleep(TRUE);
    if (pr != NULL && GuiIsEditorObject(edit)) {
        filename = GetFilename(pr, memPool, GCS(MSG_SaveText), GCS(MSG_Save),
                NULL, PREFS_TEXTEDITOR_FILENAME, TRUE);
        if (filename != NULL) {
            saveTextHelp(pr, edit, filename);
            FreeString(memPool, filename);
        }
    }
    GuiSetSleep(FALSE);
}

static void externalEditorCleanup(struct MemPool *memPool, 
	char *filename, char *old)
{
    FreeString(memPool, old);
    if (filename && filename[0]) {
        RmFile(filename);
    }
    FreeString(memPool, filename);
}

static void externalEditor(struct Project *pr, GuiObject edit)
{
    static const char filename_template[] = TMPDIR "TMP.mb.XXXXXX";
    const char *editcmd_template = GetPrefsString(pr, PREFS_EXTERNALEDITOR);
    struct MemPool *memPool = ProjectMemPool(pr);
    char *editcmd = NULL;
    char *filename = ExpandPathname(filename_template, memPool);
    char *old = NULL, *new;
    long error;
    int rc;
    bool ok;

    if (filename == NULL) {
        Ask2(edit, NULL, AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
        externalEditorCleanup(memPool, filename, old);
        return;
    }

    ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
    MakeTemp(filename);
    if (filename[0] == 0) {
        Ask2(edit, NULL, AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_MkTemp));
        externalEditorCleanup(memPool, filename, old);
        return;
    }

    old = GuiEditorText(edit, memPool);
    if (old == NULL) {
        Ask2(edit, NULL, AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
        externalEditorCleanup(memPool, filename, old);
        return;
    }

    TextFileSetContentsAskWarnings(edit, memPool, filename, old, &error);
    if (error) {
        Ask2(edit, NULL, AT_ERROR, GCS(MSG_StarOk), GES(error), filename);
        externalEditorCleanup(memPool, filename, old);
        return;
    }

    editcmd = BuildCommandLine(memPool, editcmd_template, filename);
    if (editcmd == NULL) {
        Ask2(edit, NULL, AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
        externalEditorCleanup(memPool, filename, old);
        return;
    }

    ok = GuiSystem(editcmd, FALSE, FALSE, &rc);
    if (ok && rc != 0) {
        Ask2(edit, NULL, AT_ERROR, GCS(MSG_StarOk), 
                GCS(MSG_Err_ExternalProgram), editcmd);
    }
    FreeString(memPool, editcmd);
    if (!ok || rc != 0) {
        externalEditorCleanup(memPool, filename, old);
        return;
    }

    ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
    new = TextFileGetContentsAskWarnings(edit, memPool, filename, &error);
    if (error && error != ERR_CANCEL) {
        Ask2(edit, NULL, AT_ERROR, GCS(MSG_StarOk), GES(error), filename);
        externalEditorCleanup(memPool, filename, old);
        return;
    } 
    if (new != NULL) {
        if (strcmp(old, new) != 0) {         /* text changed? */
            GuiSetEditorText(edit, new);
            GuiStoreEdits();
        }
        FreeString(memPool, new);
    }
    externalEditorCleanup(memPool, filename, old);
}

void MenuEditorExternalEditor(GuiObject go, struct Project *pr)
{
    GuiSetSleep(TRUE);
    externalEditor(pr, go);
    GuiSetSleep(FALSE);
}

