/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** menulog.c
**
** Functions for menu items in menu Log
**
** Creation date: 04.05.2012
** Last modified: 27.11.2020
*/

#include "system.h"
#include "menulog.h"
#include "queryeditor.h"
#include "imexport.h"
#include "project.h"
#include "display.h"
#include "field.h"
#include "log.h"
#include "requesters.h"
#include "guidosutil.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

void MenuLogActive(struct Project *pr, ulong active)
{
    long error = ObtainProjectChangeLock(pr);

    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    error = SetLogActive(ProjectLog(pr), active); 
    if (error) {
	Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_LogError), GES(error));
    } else {
        ProjectChange(pr, FALSE);
    }
}

bool MenuLogModeInit(struct Project *pr, ulong mode)
{
    struct Log *log = ProjectLog(pr);
    bool toProject = LogToProject(log);
    bool toFile = LogToFile(log);
    bool checked = FALSE;

    switch (mode) {
        case 0:
            checked = toProject && !toFile;
            break;

        case 1:
            checked = !toProject && toFile;
            break;

        case 2:
            checked = toProject && toFile;
            break;
    }

    return checked;
}

void MenuLogMode(struct Project *pr, ulong mode)
{
    struct Log *log = ProjectLog(pr);
    bool toProject = (mode == 0 || mode == 2);
    bool toFile = (mode == 1 || mode == 2);
    long error = ObtainProjectChangeLock(pr);

    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    SetLogToProject(log, toProject);
    SetLogToFile(log, toFile);
    if (toFile && *EmptyNullStr(LogFilename(log)) == 0) {
        MenuLogSetFile(pr);
    }
    ProjectChange(pr, FALSE);
}

void MenuLogSetFile(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct Log *log = ProjectLog(pr);
    char *filename;
    long error;

    filename = GetFilename(pr, memPool, GCS(MSG_SetLogFile), GCS(MSG_Ok),
            LogFilename(log), 0, TRUE);
    if (filename != NULL) {
        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        } else if (!SetLogFilename(log, filename)) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_LogError), GES(ERR_MEM));
        } else {
            ProjectChange(pr, FALSE);
        }
        FreeString(memPool, filename);
    }
}

void MenuLogSetLabel(struct Project *pr)
{
    struct Log *log = ProjectLog(pr);
    const char *label = LogLabel(log);
    const char *newlabel = NULL;
    long error;

    newlabel = AskString(pr, GCS(MSG_SetLogLabel), NULL,
	    label, MAX_STRING_LEN, AST_Normal);
    if (newlabel != NULL) {
        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        } else if (!SetLogLabel(log, newlabel)) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_LogError), GES(ERR_MEM));
	} else {
            ProjectChange(pr, FALSE);
        }
    }
}

void MenuLogImport(struct Project *pr)
{
    struct Log *log = ProjectLog(pr);
    struct Table *tab = LogTable(log);

    GuiImportRecords(tab);
}

void MenuLogExport(struct Project *pr)
{
    struct Log *log = ProjectLog(pr);
    struct Table *tab = LogTable(log);

    GuiExportRecords(tab);
}

void MenuLogClear(struct Project *pr)
{
    if (Ask(AT_WARN, GCS(MSG_ClearLogAnswers), GCS(MSG_ClearLog))) {
        long error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        GuiSetSleep(TRUE);
        error = ClearLog(ProjectLog(pr), FALSE); 
        GuiSetSleep(FALSE);

        if (error) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_LogError), GES(error));
        } else {
            ProjectChange(pr, FALSE);
        }
    }
}

#ifdef ENABLE_LOG_APPLY_CHANGES
void MenuLogApplyChanges(struct Project *pr)
{
    GuiApplyChanges(pr);
}
#endif

void MenuLogView(struct Project *pr)
{
    struct Log *log = ProjectLog(pr);

    if (LogToProject(log)) {
#ifdef USE_LOG_WINDOW
        long error = LogShowWindow(log);

        if (error) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_LogError), GES(error));
        }
#else
        OpenQueryEditorWithTableQuery(LogTable(log));
#endif
    } else if (LogToFile(log)) {
        GuiViewFile(LogFilename(log), TRUE);
    }
}

