/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** menuprogram.c
**
** Functions for menu items in menu Program
**
** Creation date: 15.07.1995
** Last modified: 18.01.2009
*/

#include "system.h"
#include "project.h"
#include "menu.h"
#include "requesters.h"
#include "programeditor.h"
#include "programoutput.h"
#include "queryeditor.h"
#include "gui.h"
#include "guidosutil.h"
#include "dosutil.h"
#include "preferences.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

void ProjectCleanupExternalSource(struct Project *pr)
{
    const char *path = GetPrefsString(pr, PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);
    if (!GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL_CLEANUP)) {
	return;
    }
    ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
    if (path == NULL || path[0] == 0 || FileStat(path) <= 0) {
	return;
    }
    if (ProgramSourceExternalIdentical(pr) ||
	    Ask(AT_WARN, GCS(MSG_DeleteAnswers), 
		GCS(MSG_Err_DeleteExternalProgramSource), path)) {
	ProgramDeleteSourceExternal(pr);
    }
}

void MenuProgramEdit(struct Project *pr)
{
    if (GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL)) {
	const char *path = GetPrefsString(pr, 
		PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);
	GuiEditFile(path, TRUE);
    } else {
	OpenProgramEditor(pr);
    }
}

bool MenuProgramCompile(struct Project *pr, ulong reload)
{
    GuiObject pe = ProjectProgramEditor(pr);
    struct Program *compiled = NULL;
    long line = 0, col = 0;
    const char *filename = NULL;
    const char *errstr;

    long error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return FALSE;
    }

    if (pe) {
	errstr = ProgramEditorCompile(pe, reload, 
		&compiled, &line, &col, &filename);
    } else {
	errstr = ProgramCompileProjectProgram(pr, FALSE,
		&compiled, &line, &col, &filename);
    }

    if (errstr) {
	ProgramEditorShowError(pr, FALSE, errstr, line, col, filename);
	return FALSE;
    }

    SetProjectProgram(pr, compiled);
    ProjectChange(pr, TRUE);

    return GuiUpdateDependentDisplayObjects(pr, pr, NULL, NULL);
}

void menuProgramSourceHelp(struct Project *pr, ulong external)
{
    if (GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL) == external) {
	return;					/* nothing to do */
    }

    if (external) {
	GuiObject pe = ProjectProgramEditor(pr);
	struct MemPool *memPool = ProjectMemPool(pr);
	char *filename;
	const char *path;
    	const char *error;
	
	filename = GetFilename(pr, memPool, NULL, GCS(MSG_Save), NULL,
		PREFS_PROGRAMSOURCEEXTERNAL_FILENAME, TRUE);
	if (filename == NULL) {
	    return;		/* user canceled */
	}
	FreeString(memPool, filename);
	path = GetPrefsString(pr, PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);

	if (pe == NULL && ProgramSourceExternalIdentical(pr))
	{
	    SetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL, TRUE);
	    return;
	}
        ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
	if (FileStat(path) >= 0 &&
		Ask(AT_QUESTION, GCS(MSG_AskReplaceAnswers),
	    	    GCS(MSG_AskReplaceFile), path) == 0)
	{
	    if (pe == NULL) {    
		/* use existing external program source */
		SetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL, TRUE);
	    }
	    return;		
	}
    	if (pe) {
	    error = ProgramEditorCreateSourceExternal(pe);
	} else {
	    error = ProgramCreateSourceExternal(pr);
	}
	
    	if (error) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), error, path);
	} else {
	    CloseProgramEditor(pe);
	    SetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL, TRUE);
	}
    } else {
	ProjectCleanupExternalSource(pr);
    	SetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL, FALSE);
    }
}

void MenuProgramSource(struct Project *pr, ulong external)
{
    GuiObject win = GuiActiveWindow();

    menuProgramSourceHelp(pr, external);
    UpdateMenu(win, TRUE);
}

void MenuProgramCleanupExternalSource(struct Project *pr, ulong checked)
{
    SetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL_CLEANUP, checked);
}

void MenuProgramDebugInfo(struct Project *pr, ulong checked)
{
    SetPrefsValue(pr, PREFS_PROGRAMDEBUGINFO, checked);
}

void MenuProgramObsoleteFunctions(struct Project *pr, ulong set)
{
    SetPrefsValue(pr, PREFS_PROGRAMOBSOLETEFUNCTIONS, set);
}

void MenuProgramSortTriggerFunctions(struct Project *pr, ulong checked)
{
    SetPrefsValue(pr, PREFS_PROGRAMSORTTRIGGERFUNCTIONS, checked);
}

void MenuProgramIncludeDir(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    char *path = GetDirname(pr, memPool, GCS(MSG_IncludeDir),
	    NULL, NULL, PREFS_PROGRAMINCLUDEDIR, FALSE);
    FreeString(memPool, path);
}

void MenuProgramOutput(struct Project *pr)
{
    GuiProgOutput(pr);
}

void MenuProgramQueryEditor(struct Project *pr)
{
    OpenQueryEditor(pr);
}

