/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** menuproject.c
**
** Functions for menu items in menu Project
**
** Creation date: 23.12.1994
** Last modified: 22.03.2022
*/

#include "menuproject.h"
#include "menuprogram.h"
#include "menu.h"
#include "main.h"
#include "structureeditor.h"
#include "queryeditor.h"
#include "server.h"
#include "project.h"
#include "table.h"
#include "order.h"
#include "record.h"
#include "display.h"
#include "preferences.h"
#include "gui.h"
#include "programeditor.h"
#include "appstrings_cat.h"
#include "error.h"
#include "requesters.h"
#include "break.h"
#include "memory.h"
#include "strhandle.h"
#include "filelock.h"
#include "util.h"
#include "dosutil.h"
#include "sqlitedb.h"
#include "fileinfo.h"
#include "lang/program.h"

static char errstr[1024];

void MenuProjectInfo(struct Project *pr)
{
    const char *name = ProjectName(pr);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    long num_tabs = 0, num_recs = 0;

    if (tabs) {
        FORNODE(tab, tabs) {
            if (!TableIsSystem(tab)) {
                num_tabs++;
                num_recs += TableNumRecords(tab);
            }
        }
    }

    if (name == NULL) {
        name = GCS(MSG_Unnamed);
    }

    Ask(AT_INFO, GCS(MSG_StarOk), GCS(MSG_Info), name, num_recs, num_tabs);
}

static bool allowReorg(struct Project *pr)
{
    return !GetPrefsValue(pr, PREFS_CONFIRMSAVEREORG) || 
        ProjectNumRecords(pr) <= 20 ||
        Ask(AT_QUESTION, GCS(MSG_ConfirmSaveReorgAnswers), 
                GCS(MSG_ConfirmSaveReorg)) != 0;
}

void MenuProjectSaveAs(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    char *filename;
    bool save = TRUE;
    long error;

    SetPrefsString(pr, PREFS_PROJECTNAME, ProjectName(pr));
    filename = GetFilename(NULL, memPool, GCS(MSG_SelectProjectname),
	    GCS(MSG_Save), NULL, PREFS_PROJECTNAME, TRUE);
    if (filename == NULL || filename[0] == 0) {
        save = FALSE;
    } else if (FileStat(filename) >= 0 &&
            Ask(AT_WARN, GCS(MSG_AskReplaceAnswers), 
                GCS(MSG_AskReplaceFile), filename) == 0) {
        save = FALSE;
    }

    if (save && allowReorg(pr)) {
        error = SaveProject(pr, filename, TRUE, errstr);
        if (error) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error), errstr);
        }
    }

    FreeString(memPool, filename);
}

void MenuProjectSave(struct Project *pr, ulong reorg)
{
    const char *filename = ProjectName(pr);
    long error;

    if (filename == NULL) {
	MenuProjectSaveAs(pr);
    } else if (!reorg || allowReorg(pr)) {
        error = SaveProject(pr, filename, reorg, errstr);
        if (error) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error), errstr);
        }
    }
}

static bool MenuProjectReallyContinue(struct Project **ppr, bool opennew)
{
    struct Project *pr = *ppr, *newpr = NULL;
    long changes = ProjectChanges(pr);
    long connections = ServerNumProjectConnections(pr);
    short result = 2;  /* continue */

    if(changes != 0 && connections == 0)
    {
	result = Ask(AT_WARN, GCS(MSG_AskProjectChangesReallyContinueAnswers),
		     GCS(MSG_AskProjectChangesReallyContinue), changes);
    }
    else if(changes == 0 && connections != 0)
    {
	result = Ask(AT_WARN, 
		GCS(MSG_AskProjectConnectionsReallyContinueAnswers),
	   	GCS(MSG_AskProjectConnectionsReallyContinue), 
   		connections);
    }
    else if(changes != 0 && connections != 0)
    {
	result = Ask(AT_WARN,
	    GCS(MSG_AskProjectChangesConnectionsReallyContinueAnswers),
	    GCS(MSG_AskProjectChangesConnectionsReallyContinue), 
	    changes, connections);
    } 
    else if (!opennew && NumNodes(&ProjectList) <= 1) 
    {
	/* we are about to quit BeeBase */
	if (GetPrefsValue(pr, PREFS_CONFIRMQUIT) &&
		Ask(AT_WARN, GCS(MSG_AskQuitAnswers), GCS(MSG_ConfirmQuit))==0)
	{
	    result = 0;
	}
    }

    if(result == 2)         /* continue? */
    {
	return TRUE;
    }
    else if(result == 0)    /* cancel? */
    {
	return FALSE;
    }

    if(changes != 0)
    {
	MenuProjectSave(pr, FALSE);
	changes = ProjectChanges(pr);
	if(changes != 0)                  /* save failed? */
	{
	    return FALSE;
	}
    }

    if(connections != 0)
    {
	if(opennew)
	{
	    newpr = MenuProjectNew(pr, FALSE);
	    if(newpr == NULL)
	    {
		return FALSE;
	    }
	}
	SetProjectHasGui(pr, FALSE);
	*ppr = newpr;
	return TRUE;              
    }

    return TRUE;
}

void MenuProjectClear(struct Project *pr)
{
    if (pr == NULL) {
	return;
    }
    if (!MenuProjectReallyContinue(&pr, TRUE)) {
	return;
    }

    ProjectCleanupExternalSource(pr);
    ClearProject(pr, TRUE);
    OpenProjectWindows(pr);		/* ignore failure */
}

void MenuProjectClearRecords(struct Project *pr)
{
    long error;
    short rc;

    if (pr == NULL) {
        return;
    }

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    rc = Ask(AT_WARN, GCS(MSG_DeleteAnswers), GCS(MSG_DeleteAllRecords));
    if (rc == 0) {
	return;
    }

    GuiSetSleep(TRUE);
    error = DeleteRecordsProject(pr);
    ProjectChange(pr, FALSE);
    GuiSetSleep(FALSE);
    GuiUpdate(pr);

    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
    }
}

struct Project *MenuProjectNew(struct Project *pr, ulong show)
{
    struct Project *newpr = AllocProject(TRUE);

    if (newpr == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_OpenProject));
    } else {
	INSERT(&ProjectList, newpr, pr);
	if (show && !OpenProjectWindows(newpr)) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_CantOpenMainWindow));
	    REMOVE(newpr);
	    FreeProject(newpr);
	    newpr = NULL;
	}
    }
    return newpr;
}

static char *suggestSQLite3ProjectName(const char *dirname,
        struct MemPool *memPool)
{
    char *suggest = NULL;
    char suffix[3];
    int len;

    if (dirname != NULL) {
        len = strlen(dirname);
        if (len > 0 && IsPathSeparator(dirname[len - 1])) {
            len--;
        }
        if (len >= 3 && dirname[len - 3] == '.') {
            CodesetStrNBytesCopy(suffix, dirname + len - 2, 2);
            if (CodesetStrCmp(suffix, "mb") == 0) {
                len -= 3;  /* Directory name ends with ".mb" */
            }
        }

        suggest = PoolAllocMem(memPool, len + 4 + 1);
        sprintf(suggest, "%.*s.bbs", len, dirname);
    }

    return suggest;
}

static long convertToSQLite3(const char *dirname, const char *newname,
        char *errstr)
{
    struct Project *pr = AllocProject(FALSE);
    struct MemPool *memPool = ProjectMemPool(pr);
    struct FileLockContext *flc = AllocFileLockContext(memPool, dirname);
    long error = OK;

    if (flc == NULL) {
        error = ERR_MEM;
    } else {
        error = OpenProjectCustomFormat(pr, dirname, flc, TRUE, errstr);
        if (error == OK) {
            error = SaveProject(pr, newname, FALSE, errstr);
        }
    }
    FreeFileLockContext(flc);
    FreeProject(pr);

    return error;
}

static void menuProjectOpenHelp(struct Project *pr, const char *pathname)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    char *pathmem = NULL;
    char *dirname = NULL;
    char *suggest = NULL;
    char *newname = NULL;
    bool proceed = TRUE;
    struct Project *exists;
    long error;

    if (pathname == NULL) {
        pathname = pathmem = GetFilename(
	    NULL, memPool, GCS(MSG_SelectProjectname),
	    GCS(MSG_Open), ProjectName(pr), PREFS_PROJECTNAME, FALSE);
    }
    if (pathname != NULL && pathname[0] != 0) {
        dirname = PathDirname(pathname, memPool);
        if (IsProjectCustomFormatDir(dirname)) {
            if (Ask(AT_QUESTION, GCS(MSG_AskConvertProjectAnswers),
                        GCS(MSG_AskConvertProject), dirname) == 0) {
                proceed = FALSE;  /* User canceled */
            } else {
                suggest = suggestSQLite3ProjectName(dirname, memPool);
                newname = GetFilename(
                        NULL, memPool, GCS(MSG_SelectProjectname),
                        GCS(MSG_Convert), suggest, PREFS_PROJECTNAME, TRUE);
                if (newname == NULL || newname[0] == 0) {
                    proceed = FALSE;
                } else if (FileStat(newname) >= 0 &&
		    Ask(AT_WARN, GCS(MSG_AskReplaceAnswers), 
			GCS(MSG_AskReplaceFile), newname) == 0) {
                    proceed = FALSE;
                } else {
                    // Convert project
                    error = convertToSQLite3(dirname, newname, errstr);
                    if (error != OK) {
                        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error), errstr);
                        proceed = FALSE;
                    }
                    pathname = newname;
                }
            }
        }
        if (proceed) {
            exists = FindProjectName(pathname);
            if (exists == NULL) {
                if (pr == NULL || !ProjectIsEmpty(pr) || !ProjectHasGui(pr) ||
                        ProjectStructureEditor(pr) != NULL) {
                    pr = MenuProjectNew(pr, FALSE);
                }
                if (pr != NULL) {
                    error = OpenProject(pr, pathname, TRUE, errstr);
                    if (error != OK) {
                        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error), errstr);
                        OpenProjectWindows(pr);
                    }
                }
            } else if (exists == pr) {
                MenuProjectRevertToSaved(pr);
            } else if (!SetProjectHasGui(exists, TRUE)) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_CantBuildGui));
            } else if (ProjectStructureEditor(exists) == NULL) {
                struct DisplayObject *dio = ProjectRootDisplayObject(exists);
                GuiObject wi = DisplayObjectGuiObject(dio);

                if (!OpenProjectWindows(exists)) {
                    Ask(AT_ERROR, GCS(MSG_StarOk), 
                            GCS(MSG_Err_CantOpenMainWindow));
                } else {
                    GuiWindowToFront(wi);
                    GuiActivateWindow(wi);
                }
            }
        }
    }

    FreeString(memPool, pathmem);
    FreeString(memPool, dirname);
    FreeString(memPool, suggest);
    FreeString(memPool, newname);
}

void MenuProjectOpen(struct Project *pr, const char *pathname)
{
    menuProjectOpenHelp(pr, pathname);
}

void MenuProjectOpenDemo(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    char *demodir;
    char *pathname;

    demodir = ExpandPathname("BeeBase:demos", memPool);
    if (demodir != NULL) {
        pathname = GetFilename(
                NULL, memPool, GCS(MSG_SelectProjectname),
                GCS(MSG_Open), demodir, 0, FALSE);
        FreeString(memPool, demodir);
        if (pathname != NULL) {
            menuProjectOpenHelp(pr, pathname);
            FreeString(memPool, pathname);
        }
    }
}

void MenuProjectRevertToSaved(struct Project *pr)
{
    const char *name = ProjectName(pr);
    long error, changes = ProjectChanges(pr);

    if (pr == NULL) {
	return;
    }

    if (name == NULL || name[0] == 0) {
	Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_NoProjectName));
	return;
    }

    if(changes != 0 && 
       !Ask(AT_QUESTION, GCS(MSG_AskProjectReallyRevertAnswers), 
	    GCS(MSG_AskProjectReallyRevert), changes)) {
	return;                 /* User canceled */
    }

    error = RevertProject(pr, TRUE, errstr);
    if (error != OK) {
	Ask(AT_ERROR, GCS(MSG_StarOk), GES(error), errstr);
    }
}

static void menuProjectCloseHelp(struct Project *pr, bool callOnClose)
{
    if(pr == NULL)
	return;
    GuiSetSleep(TRUE);
    if(callOnClose && FindUserFunctionName(ProjectProgram(pr), onClose))
    {
	ExecuteTrigger(pr, onClose, NULL, NULL);
    }
    if(MenuProjectReallyContinue(&pr, FALSE) && pr != NULL)
    {
	ProjectCleanupExternalSource(pr);
	REMOVE(pr);
	FreeProject(pr);
    }
    GuiSetSleep(FALSE);
    if(NumNodes(&ProjectList) == 0)             /* no more projects? */
    {
        MenuProjectQuit();
    }
}

void MenuProjectClose(struct Project *pr)
{
    menuProjectCloseHelp(pr, TRUE);
}

void MenuProjectSaveClose(struct Project *pr)
{
    if(FindUserFunctionName(ProjectProgram(pr), onClose))
	 ExecuteTrigger(pr, onClose, NULL, NULL);

    MenuProjectSave(pr, FALSE);
    menuProjectCloseHelp(pr, FALSE);
}

void MenuProjectAdminMode(struct Project *pr)
{
    bool admin = ProjectAdminMode(pr);
    const char *adminPasswd = ProjectAdminPassword(pr);
    const char *passwd;

    admin = !admin;

    if (admin && adminPasswd && adminPasswd[0]) {
	passwd = AskString(pr, GCS(MSG_EnterAdminPassword), GCS(MSG_Ok), 
		NULL, (ulong)-1, AST_Secret);
	if (passwd == NULL) {
	    return;
	}
	passwd = SHA1Sum(passwd);
	if (strcmp(passwd, adminPasswd) != 0) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_WrongPassword));
	    return;
	}
    }
    SetProjectAdminMode(pr, admin);
    UpdateMenu(GuiActiveWindow(), TRUE);
}

void MenuProjectChangeAdminPassword(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    const char *adminPasswd = ProjectAdminPassword(pr);
    const char *passwd;
    char *copy;
    long error;
   
    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    if (adminPasswd && adminPasswd[0]) {
	passwd = AskString(pr, GCS(MSG_EnterCurrentAdminPassword), GCS(MSG_Ok), 
		NULL, (ulong)-1, AST_Secret);
	if (passwd == NULL) {
	    return;
	}
	passwd = SHA1Sum(passwd);
	if (strcmp(passwd, adminPasswd) != 0) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_WrongPassword));
	    return;
	}
    }

    passwd = AskString(pr, GCS(MSG_EnterNewAdminPassword), GCS(MSG_Ok),
	    NULL, (ulong)-1, AST_Secret);
    if (passwd == NULL) {
	return;
    }
    copy = AllocString(memPool, passwd);
    passwd = AskString(pr, GCS(MSG_ReEnterAdminPassword), GCS(MSG_Ok),
	    NULL, (ulong)-1, AST_Secret);
    if (passwd) {
	if (strcmp(passwd, copy) != 0) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_PasswordsDoNotMatch));
	} else if (passwd[0]) {
	    SetProjectAdminPassword(pr, SHA1Sum(passwd));
            ProjectChange(pr, TRUE);
	    Ask(AT_INFO, GCS(MSG_StarOk), GCS(MSG_AdminPasswordUpdated));
	} else {
	    SetProjectAdminPassword(pr, NULL);
            ProjectChange(pr, TRUE);
	    Ask(AT_INFO, GCS(MSG_StarOk), GCS(MSG_AdminPasswordCleared));
	}
    }
    FreeString(memPool, copy);
}

void MenuProjectSwapRecords(struct Project *pr)
{
    long error = SwapProjectRecords(pr);

    if (error) {
       	Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
    }
}

static void exportstructurehelp(struct Project *pr, const char *filename)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct Program *prg = ProjectProgram(pr);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    FILE *file;

    ProjectSetCurrentDir(pr);           /* make sure we have correct CWD */
    if (FileStat(filename) >= 0) {      /* file exists? */
        if(Ask(AT_WARN, GCS(MSG_AskReplaceAnswers), 
		    GCS(MSG_AskReplaceFile), filename) == 0) {
            return;
	}
    }

    file = IFOpen(memPool, filename, "w");
    if (file == NULL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_OpenFile), filename);
        return;
    }

    fprintf(file, "Project %s\n", ProjectName(pr));
    if (tabs) {
        FORNODE(tab, tabs) {
	    long num = TableNumRecords(tab);
	    const char *trigger;
	    char *order = GetOrderStr(memPool, tab);
	    struct List *fields = TableFields(tab);
	    struct Field *field;

            if (TableIsSystem(tab)) {
                continue;       /* skip system tables */
            }

            fprintf(file, "\n    Table %s (%s%ld record%s)", TableName(tab),
                TableExactlyOneRecord(tab)?  "exactly " : "",
		num, num == 1? "" : "s");
	    trigger = TableNewTrigger(tab);
	    if (trigger && trigger[0])
		fprintf(file, ", new: %s", trigger);
	    trigger = TableDelTrigger(tab);
	    if (trigger && trigger[0])
		fprintf(file, ", del: %s", trigger);
            if (!TableLogChanges(tab))
		fprintf(file, ", nolog");
	    fprintf(file, "\n");
	    if (order && order[0]) {
	        fprintf(file, "    Ordered by %s\n", order);
	    } else {
		fprintf(file, "    No order\n");
	    }
	    FreeString(memPool, order);

            fprintf(file, "    Fields:\n");
            if (fields != NULL) {
                FORNODE(field, fields) {
                    struct DisplayObject *dio = FieldDisplayObject(field);
                    short maxlen = FieldStringMaxlen(field);
		    short type = FieldType(field);
                    char *fmt = "";

                    fprintf(file, "        %-21s ", FieldName(field));
                    switch (type) {
                        case TYPE_STRING:       fmt = "String(%hd)";    break;
                        case TYPE_INTEGER:      fmt = "Integer";        break;
                        case TYPE_REAL:         fmt = "Real";           break;
                        case TYPE_DATE:         fmt = "Date";           break;
                        case TYPE_TIME:         fmt = "Time";           break;
                        case TYPE_BOOL:         fmt = "Bool";           break;
                        case TYPE_CHOICE:       
			{
			    char **labels = FieldChoiceLabels(field), **s;

			    fprintf(file, "Choice(");
			    if (labels) {
				for (s = labels; *s; s++) {
				    fprintf(file, "%s%s", 
					(s == labels)? "" : ", ", *s);
				}
			    }
			    fprintf(file, ")");
			    break;
			}
                        case TYPE_MEMO:         fmt = "Memo";           break;
                        case TYPE_REFERENCE:
                            fprintf(file, "Reference to %s",
                                TableName(FieldReferenceTable(field)));
                            if (FieldReferenceAutoShow(field)) {
                                fprintf(file, " *auto-show*");
                            }
                            if (FieldReferenceFilter(field)) {
                                fprintf(file, " *filter*");
                            }
                            break;
                        case TYPE_VIRTUAL:      fmt = "Virtual";        break;
                        case TYPE_BUTTON:       fmt = "Button";         break;
                    }
                    num = fprintf(file, fmt, maxlen);
		    fprintf(file, "%*s", (int)MAX(20 - num, 0), "");
		    if (type == TYPE_VIRTUAL) {
			struct List *deps;
			struct Dependency *dep;
			struct UserGlobalVar *ugv;
			struct Field *depfield;
                        struct DisplayObject *depguiobj;
			struct Table *deptab;
                        struct DisplayObject *depguirec;
			bool first = TRUE;

			deps = ProgramDisplayDependenciesVirtualField(prg, dio);
			trigger = FieldVirtualCompute(field);
			if (trigger && trigger[0]) {
			    fprintf(file, "compute: %s", trigger);
			} else {
			    fprintf(file, "No compute function");
			}
			if (deps && SUCC(HEAD(deps))) {
			    fprintf(file, "\n%*sdepends on: ", 30, "");
			    FORNODE(dep, deps) {
				ugv = DependencyUserGlobalVar(dep);
				depfield = DependencyField(dep);
				depguiobj = DependencyGuiObject(dep);
				deptab = DependencyTable(dep);
                                depguirec = DependencyGuiRecord(dep);
				if (ugv) {
				    if (!first) {
					fputs(", ", file);
				    }
				    fprintf(file, "%s", UserGlobalVarName(ugv));
				    first = FALSE;
				} else if (depfield) {
				    if (!first) {
					fputs(", ", file);
				    }
				    deptab = FieldTable(depfield);
				    if (deptab != tab) {
					fprintf(file, "%s.", TableName(deptab));
				    }
				    fprintf(file, "%s", FieldName(depfield));
				    first = FALSE;
				} else if (depguiobj) {
				    if (!first) {
					fputs(", ", file);
				    }
                                    depfield = DisplayFieldField(depguiobj);
				    deptab = FieldTable(depfield);
				    if (deptab != tab) {
					fprintf(file, "%s.", TableName(deptab));
				    }
				    fprintf(file, "%s.guiObject",
                                            FieldName(depfield));
				    first = FALSE;
				} else if (deptab) {
				    if (!first) {
					fputs(", ", file);
				    }
				    if (FindFieldName(tab,
						TableName(deptab)) != NULL) {
					fprintf(file, "::");  /* mark global */
				    }
				    fprintf(file, "%s", TableName(deptab));
				    first = FALSE;
				} else if (depguirec) {
				    if (!first) {
					fputs(", ", file);
				    }
				    fprintf(file, "%s.guiRecord", TableName(
                                                DisplayObjectTable(depguirec)));
				    first = FALSE;
                                }
			    }
			}
		    } else {
                        bool comma = FALSE;
			trigger = FieldTrigger(field);
			if (trigger && trigger[0]) {
			    fprintf(file, "change : %s", trigger);
                            comma = TRUE;
			}
                        if (type != TYPE_BUTTON && !FieldLogChanges(field)) {
			    fprintf(file, "%snolog", comma? ", " : "");
                        }
		    }
                    fprintf(file, "\n");
                }
	    }
        }
    }
    fclose(file);
}

void MenuProjectExportStructure(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    char *filename;

    GuiSetSleep(TRUE);
    if (pr != NULL) {
        filename = GetFilename(pr, memPool, GCS(MSG_ExportStructure), 
		GCS(MSG_Export), NULL, 0, TRUE);
       	if (filename != NULL) {
	    exportstructurehelp(pr, filename);
	    FreeString(memPool, filename);
	}
    }
    GuiSetSleep(FALSE);
}

void MenuProjectQuit(void)
{
    GuiSendQuitMessage();
}

