/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** menutable.c
**
** Functions for menu items in menu Table
**
** Creation date: 19.03.1995
** Last modified: 02.07.2017
*/

#include "system.h"
#include "project.h"
#include "table.h"
#include "field.h"
#include "display.h"
#include "record.h"
#include "order.h"
#include "gui.h"
#include "queryeditor.h"
#include "programeditor.h"
#include "imexport.h"
#include "filter.h"
#include "order.h"
#include "search.h"
#include "preferences.h"
#include "menutable.h"
#include "requesters.h"
#include "util.h"
#include "appstrings_cat.h"
#include "error.h"

static struct Table *newtab;
static struct Record *newrec;

static const char *newfunc(struct Element *elem)
{
    if(elem && elem->el_Type == TYPE_RECORD && elem->el_Contents.tabrec.tab == newtab)
        newrec = elem->el_Contents.tabrec.rec;
    return(NULL);
}

static void gotoHomeObject(struct Table *tab)
{
    GuiActivateDisplayObject(DisplayHomeObject(tab));
}

void MenuTableNewRecord(struct Project *pr, ulong duplicate, GuiObject win)
{
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);
    struct Record *gui_rec = TableGuiRecord(tab);
    char *trigger = TableNewTrigger(tab);
    struct Record *init = duplicate? gui_rec : NULL, *rec;
    long error = 0;

    if (tab == NULL) {
        return;
    }

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    if (trigger != NULL && trigger[0] != 0) {
        struct Element init;
        struct Element *args[2];

        if (duplicate && gui_rec != NULL) {
            init.el_Type = TYPE_RECORD;
            init.el_Record = gui_rec;
            init.el_Contents.tabrec.tab = tab;
            init.el_Contents.tabrec.rec = gui_rec;
        } else {
            init.el_Type = TYPE_NIL;
            init.el_Record = NULL;
	}
        args[0] = &init;
        args[1] = NULL;

        newtab = tab;
        newrec = NULL;

        ExecuteTrigger(pr, trigger, args, newfunc);
        if (newrec != NULL) {
            GuiShowRecord(tab, newrec);
            gotoHomeObject(tab);
        }
        return;
    }

    error = NewRecord(tab, init, gui_rec, TRUE, &rec);
    if (error) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    error = ReorderRecords(tab);
    if (error) {
        Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Err_Order), GES(error));
    }

    if (TableCountChanges(tab)) {
        ProjectChange(pr, FALSE);
    }

    GuiUpdateDependentDisplayObjects(pr, tab, NULL, NULL);
    GuiShowRecord(tab, rec);
    gotoHomeObject(tab);
}

void MenuTableDeleteRecord(struct Project *pr, ulong dummy, GuiObject win)
{
    ulong confirm = GetPrefsValue(pr, PREFS_CONFIRMDELETERECORD);
    long error = 0;
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);
    struct Record *del = TableGuiRecord(tab);
    char *trigger = TableDelTrigger(tab);

    if (tab == NULL) {
        return;
    }

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    if (trigger != NULL && trigger[0] != 0) {
        struct Element true;
        struct Element *args[2];

        true.el_Type = TYPE_BOOL;
        true.el_Record = NULL;
        args[0] = &true;
        args[1] = NULL;

        ExecuteTrigger(pr, trigger, args, NULL);
        return;
    }

    error = AttemptDeleteRecord(tab, del);
    if (error) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    if (confirm && !Ask(AT_WARN, GCS(MSG_DeleteAnswers), GCS(MSG_DeleteRecord))){
        return;
    }

    error = DeleteRecord(tab, del);
    if (error) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }
    if (TableCountChanges(tab)) {
        ProjectChange(pr, FALSE);
    }

    GuiUpdateDependentDisplayObjects(pr, tab, NULL, NULL);
}

void MenuTableDeleteAllRecords(struct Project *pr, ulong dummy, GuiObject win)
{
    ulong confirm = GetPrefsValue(pr, PREFS_CONFIRMDELETERECORD);
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);
    struct List *recs = TableRecords(tab);
    struct Record *rec;
    long count = 0, error;

    if (recs == NULL) {
	return;
    }

    FORNODE(rec, recs) {
	if (RecordMatchFilter(rec)) {
	    count++;
        }
    }

    if (count == 0) {                    /* any records? */
	return;
    }

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    if (confirm && !Ask(AT_WARN, GCS(MSG_DeleteAllAnswers), 
		GCS(MSG_DeleteAllRecords))) {
	return;
    }

    GuiSetSleep(TRUE);
    error = DeleteRecordsTable(tab, TRUE, FALSE);
    GuiSetSleep(FALSE);
    if (error) {
	Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
    }
    if (TableCountChanges(tab)) {
        ProjectChange(pr, FALSE);
    }
    
    GuiUpdateDependentDisplayObjects(pr, tab, NULL, NULL);
}

void MenuTableImportRecords(struct Project *pr, ulong dummy, GuiObject win)
{
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);

    GuiImportRecords(tab);
}

void MenuTableExportRecords(struct Project *pr, ulong dummy, GuiObject win)
{
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);

    GuiExportRecords(tab);
}

static void nextPrevRecord(GuiObject win, void (*func)(struct Table *))
{
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);
    struct DisplayObject *tabdio = TableDisplayObject(tab);

    if (func && DisplayObjectVisible(tabdio) && DisplayTableHasPanel(tabdio) &&
            DisplayTablePanelArrows(tabdio)) {
        func(tab);
    }
}

void MenuTablePrevRecord(struct Project *pr, ulong dummy, GuiObject win)
{
    nextPrevRecord(win, GuiPrevRecord);
}

void MenuTableNextRecord(struct Project *pr, ulong dummy, GuiObject win)
{
    nextPrevRecord(win, GuiNextRecord);
}

void MenuTableFirstRecord(struct Project *pr, ulong dummy, GuiObject win)
{
    nextPrevRecord(win, GuiFirstRecord);
}

void MenuTableLastRecord(struct Project *pr, ulong dummy, GuiObject win)
{
    nextPrevRecord(win, GuiLastRecord);
}

void MenuTableJumpBackwardRecord(struct Project *pr, ulong dummy, GuiObject win)
{
    nextPrevRecord(win, GuiJumpBackwardRecord);
}

void MenuTableJumpForwardRecord(struct Project *pr, ulong dummy, GuiObject win)
{
    nextPrevRecord(win, GuiJumpForwardRecord);
}

void MenuTableJumpToRecord(struct Project *pr, ulong dummy, GuiObject win)
{
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);
    struct DisplayObject *tabdio = TableDisplayObject(tab);
    long num;
    ulong all;

    if (DisplayObjectVisible(tabdio) && DisplayTableHasPanel(tabdio) &&
            DisplayTablePanelArrows(tabdio)) {
        RecordNumAll(tab, NULL, TRUE, NULL, &all);
        if (AskInteger(pr, GCS(MSG_EnterRecordNumber), GCS(MSG_Ok), "", 
                    0, all, &num)) {
            GuiJumpToRecord(tab, num);
        }
    }
}

void MenuTableChangeFilter(struct Project *pr, ulong dummy, GuiObject win)
{
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);

    if(tab)
        GuiChangeFilter(tab);
}

void MenuTableChangeOrder(struct Project *pr, ulong dummy, GuiObject win)
{
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);

    if(tab)
        ChangeOrder(tab);
}


void MenuTableReorderAllRecords(struct Project *pr, ulong dummy, GuiObject win)
{
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);
    long error = OK;

    if (tab) {
        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        error = ReorderAllRecords(tab);
        if(error) {
            Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Err_Order), GES(error));
        }
        GuiDisplayPanel(tab);
        if (TableCountChanges(tab)) {
            ProjectChange(pr, FALSE);
        }
    }
}

static void searchfor(struct Project *pr, int dir, GuiObject win)
{
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);
    struct Field *field = DisplayFieldField(dio);
    struct SearchFor *sf = &SearchFor;
    bool forward = sf->sf_Forward;
    bool firstlast = sf->sf_FirstLastRecord;
    bool newsearch = FALSE;

    /*
    ** Check validity of sl_Table and sl_Field.
    */
    if (!FindNodePtr(ProjectTables(pr), sf->sf_Table)) {
        sf->sf_Table = NULL;
        newsearch = TRUE;
    }
    if (sf->sf_Field != NULL &&
        !FindNodePtr(TableFields(sf->sf_Table), sf->sf_Field)) {
        sf->sf_Field = NULL;
        newsearch = TRUE;
    }

    if (dir == 0 || sf->sf_Pattern[0] == 0) {
        newsearch = TRUE;
    } else if (sf->sf_Table != tab) {
        newsearch = TRUE;
    } else if (!sf->sf_AllFields && sf->sf_Field != NULL &&
            sf->sf_Field != field) {
        newsearch = TRUE;
    }

    if (newsearch) {
        sf->sf_Table = tab;
        sf->sf_Field = field;
	sf = GuiSearchFor(sf);
	if (sf) {
	    ShowSearchRecord(sf);
        }
    } else {
        sf->sf_Forward = (dir > 0)? TRUE : FALSE;
        sf->sf_FirstLastRecord = FALSE;
        ShowSearchRecord(sf);
        sf->sf_Forward = forward;
        sf->sf_FirstLastRecord = firstlast;
    }
}

void MenuTableSearchFor(struct Project *pr, ulong dummy, GuiObject win)
{
    searchfor(pr, 0, win);
}

void MenuTableSearchForward(struct Project *pr, ulong dummy, GuiObject win)
{
    searchfor(pr, 1, win);
}

void MenuTableSearchBackward(struct Project *pr, ulong dummy, GuiObject win)
{
    searchfor(pr, -1, win);
}

void MenuTableViewAllRecords(struct Project *pr, ulong dummy, GuiObject win)
{
    struct DisplayObject *dio = DisplayActiveObject(win);
    struct Table *tab = DisplayObjectTable(dio);

    OpenQueryEditorWithTableQuery(tab);
}

