/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** name.c
**
** Creation date: 20.11.1994
** Last modified: 03.08.2020
*/

#include "system.h"
#include "name.h"
#include "table.h"
#include "field.h"
#include "error.h"
#include "lang/compile.h"

static const char reservedSQLite[] = "SQLite_";

bool IsMBName(const char *name)
{
    const char *s = name;
    int len;

    if (s == NULL) {
	return FALSE;
    }
    len = strlen(s);

    if (len == 0 || len > ID_NAME_LEN) {
	return FALSE;
    }
    if (*s == '_') {
	s++;
    }
    if (*s < 'A' || *s > 'Z') {
	return FALSE;
    }
    for (++s; *s; s++) {
	if ((*s < '0' || *s > '9') && (*s < 'A' || *s > 'Z') &&
		(*s < 'a' || *s > 'z') && *s != '_') {
	    return FALSE;
	}
    }

    if (strncasecmp(name, reservedSQLite, strlen(reservedSQLite)) == 0) {
        return FALSE;
    }

    return IsIdentifier(name);
}

bool IsSystemName(const char *name)
{
    return (name != NULL && name[0] == '_');
}

bool TableNameInUse(struct Project *pr, const char *name, struct Table *ignore)
{
    struct Table *tab;

    if (pr == NULL || name == NULL) {
	return TRUE;
    }

    tab = FindTableName(pr, name);
    if (tab && tab != ignore) {
	return TRUE;
    }
    return FALSE;
}

bool FieldNameInUse(struct Table *tab, const char *name, 
	struct Field *ignore)
{
    struct Field *field;

    if (tab == NULL || name == NULL) {
	return TRUE;
    }

    field = FindFieldName(tab, name);
    if (field && field != ignore) {
	return TRUE;
    }
    return FALSE;
}

static bool nameInUse(struct Table *tab, struct Field *field, 
	const char *name)
{
    if (field != NULL) {
        return FieldNameInUse(tab, name, field);
    } else {
	return TableNameInUse(TableProject(tab), name, tab);
    }
}

static const char *newName(struct Table *tab, struct Field *field, 
	const char *base)
{
    static char name[ID_NAME_LEN + 1];
    char buffer[20];
    long i;

    if (base == NULL || base[0] == 0) {
	base = "Name";
    }
    CodesetStrNBytesCopy(name, base, ID_NAME_LEN);

    if (IsMBName(name) && !nameInUse(tab, field, name)) {
        return name;
    }

    /* we can't use this name so add a number at the end. */
    for (i=1; i != 0; i++) {
	sprintf(buffer, "%ld", i);
	sprintf(name, "%.*s%s", (int)(ID_NAME_LEN - strlen(buffer)), 
	    base, buffer);

	if (IsMBName(name) && !nameInUse(tab, field, name)) {
	    break;
	}
    }
    return name;
}

const char *NewFieldName(struct Field *field, const char *base)
{
    return newName(FieldTable(field), field, base);
}

const char *NewTableName(struct Table *tab, const char *base)
{
    return newName(tab, NULL, base);
}

static void *Unknown = NULL;

void NameSetUnknown(void *unknown)
{
    Unknown = unknown;
}

long PathNameToRelFieldPath(struct Project *pr, struct Table *tab, 
	const char *name, struct RelFieldPath *rap)
{
    char *s, *t = NULL;
    int num, len;
    struct Field *field = NULL;
    short type;

    if (name == NULL || rap == NULL) {
        return ERR_INTERNAL;
    }

    rap->rap_Relative = tab;

    if (name[0] == ':' && name[1] == ':') {            /* global path? */
        name += 2;
        tab = NULL;
    }

    rap->rap_Star = FALSE;
    rap->rap_NumFields = 0;
    rap->rap_Table = tab;

    for (num = 0; name != NULL; num++) {
        s = CodesetStrchr(name, '.');
        if (s) {
            *s = 0;
	} else {
            len = strlen(name);
            if (len > 0 && name[len - 1] == '*') {
                t = (char *)&name[len - 1];
                *t = 0;
                rap->rap_Star = TRUE;
            }
        }

	field = FindFieldName(tab, name);
       	if (field == NULL && num == 0) {       /* first path entry is a table? */
            tab = FindTableName(pr, name);
	}
        if (s) {
            *s = '.';
	} else if (t) {
            *t = '*';
	}
        name = s? s+1 : NULL;

        if (field == NULL && num == 0) {       /* first path entry is a table? */
            if (tab == NULL || tab == Unknown) {
                return NAMEERR_TABLEFIELD;
	    }
            rap->rap_Table = tab;
        } else if (field == NULL || field == Unknown) {
            return NAMEERR_TABLEFIELD;
	} else if (rap->rap_NumFields >= MAX_TABLEFIELDPATH_COMPONENTS) {
	    return NAMEERR_PATHFULL;
	} else {
            rap->rap_Fields[(int)rap->rap_NumFields] = field;
            rap->rap_NumFields++;
            type = FieldType(field);
            if (type == TYPE_REFERENCE) {
                tab = FieldReferenceTable(field);
	    } else if(name) {                     /* not the end of the path? */
                return NAMEERR_PATHCOMPONENT;
	    }
        }
    }

    return OK;
}

const char *RelFieldPathToPathName(struct RelFieldPath *rap)
{
    static char pathname[ID_NAME_LEN * MAX_TABLEFIELDPATH_COMPONENTS + 20];
    const char *name = TableName(rap->rap_Table);
    char *s = pathname;
    int i;

    if (rap == NULL || rap->rap_Table == NULL) {
        return NULL;
    }

    /* get table name */
    if (rap->rap_Relative == NULL) {     /* no relative path? */
        strcpy(s, name);
        s += strlen(s);
    } else if (rap->rap_Table != rap->rap_Relative || rap->rap_NumFields == 0) {
        if (FindFieldName(rap->rap_Relative, name) != NULL) {
            sprintf(s, "::%s", name);                  /* we need global path */
	} else {
            strcpy(s, name);
	}
        s += strlen(s);
    }

    /* get rest of path */
    for (i=0; i < rap->rap_NumFields; i++) {
        name = FieldName(rap->rap_Fields[i]);
        if (s != pathname) {
            *s++ = '.';
	}
        strcpy(s, name);
        s += strlen(s);
    }

    if (rap->rap_Star) {
        *s++ = '*';
    }
    *s = 0;

    return pathname;
}

struct Field *RelFieldPathField(struct RelFieldPath *rap)
{
    if (rap == NULL || rap->rap_NumFields == 0) {
        return NULL;
    }

    return rap->rap_Fields[rap->rap_NumFields - 1];
}

