/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** order.c
**
** Creation date: 26.04.1995
** Last modified: 21.07.2020
*/

#include "order.h"
#include "gui.h"
#include "lang/program.h"
#include "programeditor.h"
#include "memo.h"
#include "codeset.h"
#include "strhandle.h"
#include "requesters.h"
#include "memory.h"
#include "break.h"
#include "dosutil.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"
#include <setjmp.h>

struct Order
{
    struct MemPool *ord_MemPool;
    short ord_Type;
    char *ord_CmpFunc;
    short ord_Num;
    struct OrderField ord_OrderFields[0];
};

long FieldCmp(struct Field *field, 
	const struct DataInfo *di1, const struct DataInfo *di2)
{
    if (di1 == NULL || di2 == NULL) {
        return 0;
    }

    if (di1->di_Flags & DIF_NIL) {
        return ((di2->di_Flags & DIF_NIL)? 0 : -1);
    } else if (di2->di_Flags & DIF_NIL) {
        return 1;
    }

    switch (FieldType(field)) {
        case TYPE_BOOL:
            return 0;                  /* the only possible case */
        case TYPE_INTEGER:
        {
            long i1 = di1->di_Contents.integer, i2 = di2->di_Contents.integer;

	    if (i1 < i2) {
		return -1;
	    } else if (i1 > i2) {
		return 1;
	    } else {
		return 0;
	    }
        }
        case TYPE_REAL:
        {
            double d1 = di1->di_Contents.real, d2 = di2->di_Contents.real;

            if (d1 < d2) {
                return -1;
	    } else if (d1 > d2) {
                return 1;
	    } else {
                return 0;
	    }
        }
        case TYPE_DATE:
        {
            long d1 = di1->di_Contents.date, d2 = di2->di_Contents.date;

            if (d1 < d2) {
                return -1;
	    } else if (d1 > d2) {
                return 1;
	    } else {
                return 0;
	    }
        }
        case TYPE_TIME:
        {
            long t1 = di1->di_Contents.time, t2 = di2->di_Contents.time;

            if (t1 < t2) {
                return -1;
	    } else if (t1 > t2) {
                return 1;
	    } else {
                return 0;
	    }
        }
        case TYPE_CHOICE:
            return (long)di1->di_Contents.choice-(long)di2->di_Contents.choice;
        case TYPE_STRING:
            return CodesetStrCollate(EmptyNullStr(di1->di_Contents.string), 
		    EmptyNullStr(di2->di_Contents.string), CS_COLLATE_ORDER);
        case TYPE_MEMO:
            return CodesetStrCollate(EmptyNullStr(di1->di_Contents.mi), 
                    EmptyNullStr(di2->di_Contents.mi), CS_COLLATE_ORDER);
   }
   return 0;
}

void FreeOrder(struct Order *ord)
{
    if(ord != NULL) {
	FreeString(ord->ord_MemPool, ord->ord_CmpFunc);
	PoolFreeMem(ord->ord_MemPool, ord, sizeof(struct Order) + 
		ord->ord_Num * sizeof(struct OrderField));
    }
}

static struct Order *AllocOrder(struct MemPool *memPool, 
	const char *cmpfunc, short num)
{
    struct Order *ord = PoolAllocMemZero(memPool, sizeof(struct Order) + 
	    num * sizeof(struct OrderField));

    if (ord != NULL) {
	ord->ord_MemPool = memPool;
	ord->ord_Num = num;
	if (cmpfunc) {
	    ord->ord_Type = ORD_TYPE_CMPFUNC;
	    ord->ord_CmpFunc = AllocString(memPool, cmpfunc);
	    if (ord->ord_CmpFunc == NULL) {
		FreeOrder(ord);
		ord = NULL;
	    }
	} else {
	    ord->ord_Type = ORD_TYPE_FIELDLIST;
	}
    }
    return ord;
}

void OrderRemField(struct Order *ord, struct Field *field)
{
    short i;

    if (ord) {
	for (i=0; i < ord->ord_Num; i++) {
	    if (ord->ord_OrderFields[i].oa_Field == field) {
		ord->ord_OrderFields[i].oa_Field = NULL;
	    }
	}
    }
}

long RecordCmp(struct Table *tab, struct Record *rec1, struct Record *rec2, 
	       long *resultp)
{
    struct Order *ord = TableOrder(tab);
    long result = 0;
    long error = OK;

    if (ord == NULL) {            /* no order defined? */
        /* all set */
    } else if (ord->ord_Type == ORD_TYPE_CMPFUNC) {
        struct Project *pr = TableProject(tab);
        struct Program *prg = ProjectProgram(pr);
        struct UserFunction *uf = FindUserFunctionName(prg, ord->ord_CmpFunc);
        struct Element elem1, elem2, *args[3], *elem;

        if (ord->ord_CmpFunc == NULL) {       /* no order defined? */
            /* all set */
        } else if (uf == NULL) {
            Ask(AT_WARN, GCS(MSG_StarOk),GCS(MSG_Err_Trigger),ord->ord_CmpFunc);
            error = PROGERR_TRIGGER;
        } else {
            elem1.el_Type = TYPE_RECORD;
            elem1.el_Record = rec1;
            elem1.el_Contents.tabrec.tab = tab;
            elem1.el_Contents.tabrec.rec = rec1;
            elem2.el_Type = TYPE_RECORD;
            elem2.el_Record = rec2;
            elem2.el_Contents.tabrec.tab = tab;
            elem2.el_Contents.tabrec.rec = rec2;
            args[0] = &elem1;
            args[1] = &elem2;
            args[2] = NULL;

            elem = ProgramCall(prg, uf, args, FALSE);
            if (elem == NULL || elem->el_Type != TYPE_INTEGER) {
                error = PROGERR_TYPE;
            } else {
                result = elem->el_Contents.integer;
            }
        }
    } else {
        struct OrderField *oa = ord->ord_OrderFields; 
        struct OrderField *end = oa + ord->ord_Num;
        struct DataInfo *di1, *di2, di1_buf, di2_buf;
        struct Field *field;

        RecordLock(rec1);
        RecordLock(rec2);

        while (oa != end) {
            field = oa->oa_Field;
            if (field != NULL) {
                error = RecordDataInfo(rec1, field, &di1, &di1_buf);
                if (error == OK) {
                    error = RecordDataInfo(rec2, field, &di2, &di2_buf);
                    if(error == OK) {
                        result = FieldCmp(field, di1, di2);
                        if (oa->oa_Ascending == FALSE) {
                            result = -result;
                        }
                        if (result != 0) {
                            break;
                        }
                    }
                }
            }
            if (error) {
                break;
            }
            oa++;
        }

        RecordUnlock(rec2);
        RecordUnlock(rec1);
    }

    if (resultp) {
        *resultp = result;
    }

    return error;
}

static struct Table *orderTable = NULL;
static jmp_buf orderJmpBuf;
static long orderError = OK;

static void setOrderError(long err)
{
  orderError = err;
  longjmp(orderJmpBuf, 1);                 /* return to main procedure */
}

static int datacmp(void *p1, void *p2)
{
    struct Record *rec1 = p1, *rec2 = p2;
    long result = 0;
    long error;

    error = RecordCmp(orderTable, rec1, rec2, &result);
    if (error) {
	setOrderError(error);
    }
    return result;
}

static long InsertRecord(struct Record *insert, struct Record *left, long num)
{
    if (num > 0) {
        struct Record *rec = SUCC(left);
        long num2 = num/2, i;
        int result;

	if (rec == insert) {
	    rec = SUCC(rec);
	}
        for (i = num2; i > 0; i--) {
            rec = SUCC(rec);
	}

        result = datacmp(insert, rec);
        if (result < 0) {
            return InsertRecord(insert, left, num2);
	} else if (result > 0) {
            return (num2 + 1 + InsertRecord(insert, rec, num - num2 - 1));
	} else {
            return (num2 + 1 + InsertRecord(insert, rec, 0));
	}
    } else {
        REMOVE(insert);
        INSERT(NULL, insert, left);
        return 0;
    }
}

long ReorderRecords(struct Table *tab)
{
    struct Project *pr = TableProject(tab);
    struct List *recs = TableRecords(tab);
    struct Order *ord = TableOrder(tab);
    struct Record *rec, *succ, *phead, *ptail, *insert;
    long num1 = 0, num2 = 0, help;
    bool setsleep, wakeup = FALSE, endprogram = FALSE;
    int ret;
    struct Table *oldOrderTable = orderTable;
    jmp_buf oldOrderJmpBuf;
    
    memcpy(&oldOrderJmpBuf, &orderJmpBuf, sizeof(orderJmpBuf));
    
    if(recs == NULL || ord == NULL)
	return(OK);
    
    /*
    ** Put non-ordered records at start of list
    */
    insert = phead = (struct Record *)PRED(HEAD(recs));
    SAFEFORNODE(rec, succ, recs)
    {
    	if(!RecordIsSorted(rec))
	{
  	    REMOVE(rec);
  	    INSERT(NULL, rec, insert);
  	    insert = rec;
  	    num2++;
	}
  	else
    	    num1++;
    }
    
    if(num2 == 0)
	return(OK);                   /* everything all right */
    
    orderTable = tab;
    orderError = OK;
    
    setsleep = ProgramRunning()? FALSE : TRUE;
    
    if(ord->ord_Type == ORD_TYPE_CMPFUNC && !ProgramRunning())
    {
       	ProgramBegin(ProjectProgram(pr), TRUE);
	endprogram = TRUE;
    }
    
    if(setsleep && num2 > 3)
    {
       	GuiSetSleep(TRUE);
	wakeup = TRUE;
    }
    
    ret = setjmp(orderJmpBuf);
    if(ret == 0)
    {
       	/* sort unsorted list */
	ptail = SUCC(insert);
	QLSortNodes((struct Node *)phead, (struct Node *)ptail, datacmp);
     	
       	/* determine smaller list */
	if(num1 > num2)
	{
	    insert = (struct Record *)HEAD(recs);
	    rec = PRED(ptail);
	    if(rec == insert)
		rec = PRED(insert);
	}
	else
	{
	    FORNODE(rec, recs)
	       	SetRecordIsSorted(rec, !RecordIsSorted(rec));
	    
	    insert = ptail;
	    rec = (struct Record *)PRED(HEAD(recs));
	    help = num1;
	    num1 = num2;
	    num2 = help;
	}
   	
     	/*
       	** Insert records after rec
	 */
	while (num2 > 0)
	{
	    succ = SUCC(insert);
	    num1 -= InsertRecord(insert, rec, num1);
	    rec = insert;
	    SetRecordIsSorted(insert, TRUE);
	    insert = succ;
	    num2--;
	}
    }
    
    if(wakeup)
	GuiSetSleep(FALSE);
    
    if(endprogram)
    {
       	const char *error, *filename = NULL;
	long line=-1, col=1;
	
        ProgramEnd();
	error = ProgramError(&line, &col, &filename);
	if(error != NULL && error != EXECUTE_ABORT)
	{
	    ProgramEditorShowError(pr, TRUE, error, line, col, filename);
	    return(FALSE);
	}
    }
    
    orderTable = oldOrderTable;
    memcpy(&orderJmpBuf, &oldOrderJmpBuf, sizeof(orderJmpBuf));
    
    return(orderError);
}

long ReorderAllRecords(struct Table *tab)
{
  struct List *recs = TableRecords(tab);
  struct Record *rec;
  long error;

  if(recs)
    FORNODE(rec, recs)
      SetRecordIsSorted(rec, FALSE);

  error = ReorderRecords(tab);
  return(error);
}

static long checkorderfields(struct OrderField *oas, short num)
{
    short i;

    for (i=0; i < num; i++, oas++) {
        if (oas->oa_Field == NULL) {
            return ORDERR_UNKNOWNFIELD;
        }
        switch (FieldType(oas->oa_Field)) {
            case TYPE_BUTTON:
            case TYPE_VIRTUAL:
                return ORDERR_FIELDTYPE;
        }
    }
    return OK;
}

long SetNewOrder(struct Table *tab, struct OrderData *od)
{
    struct Order *new = NULL, *old = TableOrder(tab);
    struct MemPool *memPool = TableMemPool(tab);

    if (tab == NULL) {
        return ERR_INTERNAL;
    }

    if (od == NULL) {
    } else if (od->od_Type == ORD_TYPE_CMPFUNC &&
            od->od_Contents.cmpfunc != NULL &&
            od->od_Contents.cmpfunc[0] != 0) {
        new = AllocOrder(memPool, od->od_Contents.cmpfunc, 0);
        if (new == NULL) {
            return ERR_MEM;
        }
    } else if (od->od_Type == ORD_TYPE_FIELDLIST &&
	  od->od_Contents.fieldlist.num > 0) {
        short i, num = od->od_Contents.fieldlist.num;
        struct OrderField *oas = od->od_Contents.fieldlist.oas;
        long error = checkorderfields(oas, num);

        if (error) {
            return error;
        }
         
        new = AllocOrder(memPool, NULL, num);
        if (new == NULL) {
            return ERR_MEM;
        }
        for (i = 0; i < num; i++) {
            new->ord_OrderFields[i] = oas[i];
        }
    }

    FreeOrder(old);
    SetTableOrder(tab, new);

    return OK;
}

void ChangeOrder(struct Table *tab)
{
  struct Order *ord = TableOrder(tab);
  short type = ord? ord->ord_Type : ORD_TYPE_FIELDLIST;
  struct OrderData od;

  od.od_Type = type;
  if(type == ORD_TYPE_CMPFUNC)
    od.od_Contents.cmpfunc = ord? ord->ord_CmpFunc : NULL;
  else
    {
      od.od_Contents.fieldlist.num = ord? ord->ord_Num : 0;
      od.od_Contents.fieldlist.oas = ord? ord->ord_OrderFields : NULL;
    }
  
  GuiChangeOrder(tab, &od);
}

struct OrderField *GetOrderField(struct Table *tab, short num)
{
    struct Order *order = TableOrder(tab);
    struct OrderField *oa = NULL;
    short i;

    if(order == NULL || order->ord_Type != ORD_TYPE_FIELDLIST)
        return(NULL);

    oa = order->ord_OrderFields;
    for(i=0; i < order->ord_Num; i++, oa++)
    {
        if(num == 0)
            return(oa);
        if(oa->oa_Field != NULL)
            num--;
    }

    return(NULL);
}

char *GetOrderStr(struct MemPool *memPool, struct Table *tab)
{
    struct Order *ord = TableOrder(tab);
    struct StrHandle *sh = AllocStrHandle(memPool);
    char *s = NULL;

    if (sh == NULL) {
        return NULL;
    }
    
    if (ord == NULL) {
    } else if (ord->ord_Type == ORD_TYPE_CMPFUNC) {
      StrHandlePutS(sh, ord->ord_CmpFunc);
    } else {
        short i;
        struct OrderField *oa = ord->ord_OrderFields;

        for (i = 0; i < ord->ord_Num; i++, oa++) {
            StrHandlePutC(sh, oa->oa_Ascending? '+' : '-');
            StrHandlePutS(sh, FieldName(oa->oa_Field));
            if (i < ord->ord_Num - 1) {
                StrHandlePutC(sh, ' ');
            }
        }
    }
    StrHandlePutC(sh, 0);               /* terminate string */
    if (!StrHandleError(sh)) {
        s = AllocString(memPool, StrHandleString(sh));
    }
    FreeStrHandle(sh);

    return s;
}

long SetOrderStr(struct Table *tab, const char *orderstr)
{
    struct Order *new = NULL, *old = TableOrder(tab);
    struct MemPool *memPool = TableMemPool(tab);
    long lnum = NumWords(orderstr);
    short i, num;
    char *s, *t;
    struct OrderField *oa;
    struct List *recs = TableRecords(tab);
    struct Record *rec;
    long error = OK;
    char *cmpfunc = NULL;
  
    if (tab == NULL) {
        return ERR_INTERNAL;
    }

    if (lnum > 0) {
        if (lnum > SHRT_MAX) {
            return ERR_MEM;
        }
        num = (short)lnum;

        s = WordString(memPool, orderstr, 0, NULL);
        if (s == NULL) {
	    return ERR_MEM;
        }
  
        if (CodesetIsLower(CodesetGetChar(s))) {
	    cmpfunc = s;
	    num = 0;
        }

        new = AllocOrder(memPool, cmpfunc, num);
        FreeString(memPool, s);
        if (new == NULL) {
	    return ERR_MEM;
        }

        oa = new->ord_OrderFields;
        for(i = 0; i < num; i++, oa++) {
	    s = t = WordString(memPool, orderstr, i, NULL);
	    if (s == NULL) {
	        FreeOrder(new);
	        return ERR_MEM;
            }

	    oa->oa_Ascending = (*t == '-')? FALSE : TRUE;
            if (*t == '+' || *t == '-') {
	        t++;
            }
	    oa->oa_Field = FindFieldName(tab, t);
	    FreeString(memPool, s);
        }

        error = checkorderfields(new->ord_OrderFields, num);
        if (error) {
	    FreeOrder(new);
	    return error;
        }
    }

    FreeOrder(old);
    SetTableOrder(tab, new);
  
    if (recs) {
        FORNODE(rec, recs) {
            SetRecordIsSorted(rec, FALSE);
        }
    }
    error = ReorderRecords(tab);
    GuiDisplayPanel(tab);

    return error;
}

/*
** load & save
*/

long LoadOrdersCustomFormat(struct Project *pr, FILE *file)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    if (tabs == NULL) {
	return ERR_INTERNAL;
    }
    FORNODE(tab, tabs) {
	struct MemPool *memPool = TableMemPool(tab);
        struct Order *ord, *old = TableOrder(tab);
	short i, num;

	num = FGetShort(file);
	if (num == 0) {
	    ord = NULL;
	} else if (num < 0) {
	    char *cmpfunc = ReadString(file, memPool);

	    if (cmpfunc == NULL) {
		return ERR_MEM;
	    }
	    ord = AllocOrder(memPool, cmpfunc, 0);
	    FreeString(memPool, cmpfunc);
	    if (ord == NULL) {
		return ERR_MEM;
	    }
	} else {
	    ord = AllocOrder(TableMemPool(tab), NULL, num);
	    if (ord == NULL) {
		return ERR_MEM;
	    }
	    for (i=0; i < num; i++) {
		long fieldnum;
	      
		FGetChar(file);         /* dummy */
		ord->ord_OrderFields[i].oa_Ascending = FGetBool(file);
		fieldnum = FGetLong(file);
		
		ord->ord_OrderFields[i].oa_Field = 
		    FindFieldNum(pr, fieldnum);
	    }
	}
        FreeOrder(old);
	SetTableOrder(tab, ord);
    }
    return OK;
}

