/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** pdfexport.c
**
** Creation date: 12.11.2011
** Last modified: 29.01.2022
*/

#include "system.h"
#include "pdfexport.h"

#ifdef USE_HPDF

#include "queryeditor.h"
#include "requesters.h"
#include "preferences.h"
#include "dosutil.h"
#include "util.h"
#include "codeset.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"
#include "hpdf.h"

#ifdef AMIGA
#include <proto/hpdf.h>
#endif

#define HPDF_DEFAULT_PAGE_SIZE HPDF_PAGE_SIZE_A4
#define HPDF_DEFAULT_PAGE_DIRECTION HPDF_PAGE_PORTRAIT

const HPDF_PageSizes HPDFPageSizes[] =
{
    HPDF_PAGE_SIZE_LETTER,
    HPDF_PAGE_SIZE_LEGAL,
    HPDF_PAGE_SIZE_A3,
    HPDF_PAGE_SIZE_A4,
    HPDF_PAGE_SIZE_A5,
    HPDF_PAGE_SIZE_B4,
    HPDF_PAGE_SIZE_B5,
    HPDF_PAGE_SIZE_EXECUTIVE,
    HPDF_PAGE_SIZE_US4x6,
    HPDF_PAGE_SIZE_US4x8,
    HPDF_PAGE_SIZE_US5x7,
    HPDF_PAGE_SIZE_COMM10,
    HPDF_PAGE_SIZE_EOF
};

const char *HPDFPaperNames[] =
{
    "Letter",
    "Legal",
    "A3",
    "A4",
    "A5",
    "B4",
    "B5",
    "Executive",
    "US4x6",
    "US4x8",
    "US5x7",
    "COMM10",
    NULL
};

const HPDF_PageDirection HPDFPageDirections[] =
{
    HPDF_PAGE_PORTRAIT,
    HPDF_PAGE_LANDSCAPE,
};

static const char *hpdfOrientationNames [] =
{
    "Portrait",
    "Landscape",
    NULL
};

void LoadHPDFSettingsFromPrefs(struct HPDFSettings *hpdfs, struct Project *pr)
{
    if (hpdfs != NULL) {
        const char *paperSize = GetPrefsString(pr, PREFS_PDFSETTINGS_PAPERSIZE);
        const char *orientation =
            GetPrefsString(pr, PREFS_PDFSETTINGS_ORIENTATION);
        const char *fontName = GetPrefsString(pr, PREFS_HPDFSETTINGS_FONTNAME);
        long fontSize100 = GetPrefsValue(pr, PREFS_HPDFSETTINGS_FONTSIZE100);
        int i;

        hpdfs->hpdfs_PageSize = HPDF_DEFAULT_PAGE_SIZE;
        if (paperSize != NULL) {
            for (i = 0; HPDFPaperNames[i] != NULL; i++) {
                if (strcasecmp(paperSize, HPDFPaperNames[i]) == 0) {
                    hpdfs->hpdfs_PageSize = HPDFPageSizes[i];
                    break;
                }
            }
        }

        hpdfs->hpdfs_PageDirection = HPDF_DEFAULT_PAGE_DIRECTION;
        if (orientation != NULL) {
            for (i = 0; hpdfOrientationNames[i] != NULL; i++) {
                if (strcasecmp(orientation, hpdfOrientationNames[i]) == 0) {
                    hpdfs->hpdfs_PageDirection = HPDFPageDirections[i];
                    break;
                }
            }
        }

        hpdfs->hpdfs_FontName = fontName;
        if (fontName == NULL || fontName[0] == 0) {
            hpdfs->hpdfs_FontName = HPDF_DEF_FONT;
        }

        hpdfs->hpdfs_FontSize = HPDF_DEF_FONTSIZE;
        if (fontSize100 > 0) {
            hpdfs->hpdfs_FontSize = fontSize100 / 100.0;
        }

        hpdfs->hpdfs_Header = GetPrefsString(pr, PREFS_PDFSETTINGS_HEADER);
        hpdfs->hpdfs_Footer = GetPrefsString(pr, PREFS_PDFSETTINGS_FOOTER);
    }
}

void SaveHPDFSettingsToPrefs(const struct HPDFSettings *hpdfs,
        struct Project *pr)
{
    if (hpdfs != NULL) {
        const char *paperSize = NULL;
        const char *orientation = NULL;
        const char *fontName = NULL;
        long fontSize100 = 0;
        int i;

        if (hpdfs->hpdfs_PageSize != HPDF_DEFAULT_PAGE_SIZE) {
            for (i = 0; HPDFPaperNames[i] != NULL; i++) {
                if (hpdfs->hpdfs_PageSize == HPDFPageSizes[i]) {
                    paperSize = HPDFPaperNames[i];
                    break;
                }
            }
        }
        SetPrefsString(pr, PREFS_PDFSETTINGS_PAPERSIZE, paperSize);

        if (hpdfs->hpdfs_PageDirection != HPDF_DEFAULT_PAGE_DIRECTION) {
            for (i = 0; hpdfOrientationNames[i] != NULL; i++) {
                if (hpdfs->hpdfs_PageDirection == HPDFPageDirections[i]) {
                    orientation = hpdfOrientationNames[i];
                    break;
                }
            }
        }
        SetPrefsString(pr, PREFS_PDFSETTINGS_ORIENTATION, orientation);

        if (hpdfs->hpdfs_FontName != NULL && (HPDF_DEF_FONT == NULL ||
                    strcasecmp(hpdfs->hpdfs_FontName, HPDF_DEF_FONT) != 0)) {
            fontName = hpdfs->hpdfs_FontName;
        }
        SetPrefsString(pr, PREFS_HPDFSETTINGS_FONTNAME, fontName);

        if (hpdfs->hpdfs_FontSize != HPDF_DEF_FONTSIZE) {
            DoubleToInt(hpdfs->hpdfs_FontSize * 100.0, &fontSize100);
        }
        SetPrefsValue(pr, PREFS_HPDFSETTINGS_FONTSIZE100, fontSize100);

        SetPrefsString(pr, PREFS_PDFSETTINGS_HEADER, hpdfs->hpdfs_Header);
        SetPrefsString(pr, PREFS_PDFSETTINGS_FOOTER, hpdfs->hpdfs_Footer);
    }
}

static const char *hpdfError(HPDF_Doc pdf)
{
    static char error[64];
    HPDF_STATUS status = HPDF_GetError(pdf);

    switch (status)
    {
        case HPDF_FAILD_TO_ALLOC_MEM:
            return GES(ERR_MEM);
        case HPDF_FILE_OPEN_ERROR:
            return GES(DISKERR_OPEN);
        case HPDF_FILE_IO_ERROR:
            return GES(DISKERR_WRITE);
        default:
            sprintf(error, "PDF error=0x%04X, detail=%d", 
                    (int)status, (int)HPDF_GetErrorDetail(pdf));
            return error;
    }
}

static double textWidth(HPDF_Page page, const char *text, 
        struct MemPool *memPool)
{
    char *default8Bit, *l;
    const char *s, *t;
    double maxw = 0.0, tw;
   
    default8Bit = CodesetConvertToDefault8Bit(text, memPool, NULL, NULL); 
    s = default8Bit;
    while (s != NULL) {
        t = strchr(s, '\n');
        l = NULL;
        if (t != NULL) {
            l = AllocStringN(memPool, s, t - s);
            s = l;
        }
        if (s != NULL) {
            tw = HPDF_Page_TextWidth(page, s);
            if (tw > maxw) {
                maxw = tw;
            }
        }
        FreeString(memPool, l);
        s = (t == NULL)? NULL : t + 1;
    }
    if (default8Bit != text) {
        FreeString(memPool, default8Bit);
    }

    return maxw;
}

static const char *textOut(HPDF_Doc pdf, HPDF_Page page, double x, double y,
        double w, bool format, double th, const char *text,
        int offset, int numLines, struct MemPool *memPool, size_t *sumWarnings)
{
    const char *error = NULL;
    long err = OK;
    size_t warnings = 0;
    char *default8Bit, *l;
    const char *s, *t;
    int end = offset + numLines, line;
    double tw = w;
   
    default8Bit = CodesetConvertToDefault8Bit(text, memPool, &err, &warnings); 
    if (default8Bit == NULL) {
        error = GES(err);
    } else if (sumWarnings) {
        *sumWarnings += warnings;
    }
    s = default8Bit;
    for (line = 0; line < end && s != NULL; line++) {
        t = strchr(s, '\n');
        if (line >= offset) {
            l = NULL;
            if (t != NULL) {
                l = AllocStringN(memPool, s, t - s);
                if (l == NULL) {
                    error = GES(ERR_MEM);
                    break;
                }
                s = l;
            }
            if (format) {
                tw = HPDF_Page_TextWidth(page, s);
            }
            if (HPDF_Page_TextOut(page, x + w - tw, y, s)) {
                error = hpdfError(pdf);
            }
            FreeString(memPool, l);
            if (error != NULL) {
                break;
            }
            y -= th;
        }
        s = (t == NULL)? NULL : t + 1;
    }
    if (default8Bit != text) {
        FreeString(memPool, default8Bit);
    }

    return error;
}

static HPDF_Page addPage(HPDF_Doc pdf, const struct HPDFSettings *hpdfs, 
        HPDF_Font font)
{
    HPDF_Page page = HPDF_AddPage(pdf); 
    if (page == NULL) {
        return NULL;
    }
    if (HPDF_Page_SetSize(page, hpdfs->hpdfs_PageSize,
                hpdfs->hpdfs_PageDirection)){
        return NULL;
    }
    if (HPDF_Page_SetFontAndSize(page, font, hpdfs->hpdfs_FontSize)) {
        return NULL;
    }
    return page;
}

static const char *exportSFWListAsPDFHelp2(HPDF_Doc pdf,
        struct Project *pr, GuiObject sfw, double *colWidths, bool showTitles,
        const struct HPDFSettings *hpdfs, const char *filename,
        size_t *warnings, long *ppages, double *ptruncated)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    const char *error;
    HPDF_Page page;
    const char *s;
    int len;
    char buffer[64];
    HPDF_Font font, bfont;
    double fsize = hpdfs->hpdfs_FontSize, ph, hspace, th, dlines, ascent, w;
    bool haveHeader = hpdfs->hpdfs_Header != NULL&& hpdfs->hpdfs_Header[0] != 0;
    bool haveFooter = hpdfs->hpdfs_Footer != NULL&& hpdfs->hpdfs_Footer[0] != 0;
    bool haveFirstCol = FALSE;
    double totalWidth = 0.0, w2;
    long linesPerPage, count, bodyLines = 0, pages, p, row, offset = 0, line;
    int cols = GuiSFWListNumColumns(sfw), col, ocol;
    const struct QueryFieldInfo *infos = GuiSFWListFieldInfos(sfw);
    const signed char *columnOrder = GuiSFWListColumnOrder(sfw);
    const struct QueryLine *header = NULL;
    int headerLines = 0, queryLines;
    bool separateEntries = FALSE;
    const int top = 15, bot = 15, left = 15;

    if (warnings) {
        *warnings = 0;
    }

    if (showTitles) {
        header = GuiSFWListQueryLine(sfw, 0);
        headerLines = QueryLineLines(header, cols);
    }

    HPDF_SetInfoAttr(pdf, HPDF_INFO_CREATOR, APP_NAME);
    HPDF_ResetError(pdf);       /* ingore error */

    font = HPDF_GetFont(pdf, hpdfs->hpdfs_FontName, "WinAnsiEncoding");
    if (font == NULL) {
        return hpdfError(pdf);
    }
    s = strchr(hpdfs->hpdfs_FontName, '-');
    if (s) {
        len = s - hpdfs->hpdfs_FontName;
    } else {
        len = strlen(hpdfs->hpdfs_FontName);
    }
    if (len > 58) {
        len = 58;
    }
    sprintf(buffer, "%.*s-Bold", len, hpdfs->hpdfs_FontName);
    bfont = HPDF_GetFont(pdf, buffer, "WinAnsiEncoding");
    if (bfont == NULL) {
        bfont = font;
        HPDF_ResetError(pdf);       /* ingore error */
    }
    ascent = HPDF_Font_GetAscent(font) * fsize / 1000.0;

    page = addPage(pdf, hpdfs, font); 
    if (page == NULL) {
        return hpdfError(pdf);
    }

    ph = HPDF_Page_GetHeight(page);
    if (ph == 0.0) {
        return hpdfError(pdf);
    }

    hspace = textWidth(page, "Xg", memPool);
    if (hspace == 0.0) {
        return hpdfError(pdf);
    }

    th = fsize + 2.0;   /* add line spacing */
    dlines = (ph - top - bot) / th;
    if (header) {
        dlines -= headerLines + 0.5;
    }
    if (haveHeader) {
	dlines -= 2.0;
    }
    dlines -= 2.0;      /* space for footer and page number */
    linesPerPage = (int)floor(dlines);
    count = GuiListNumEntries(sfw);
    for (row = 1; row <= count; row++) {
        queryLines = QueryLineLines(GuiSFWListQueryLine(sfw, row), cols);
        if (queryLines > 1) {
            separateEntries = TRUE;
        }
        bodyLines += queryLines;
    }
    if (linesPerPage % 5 != 0 && linesPerPage != bodyLines) {
        linesPerPage &= ~1;    /* even number of lines */
    }
    if (linesPerPage < 1) {
        linesPerPage = 1;      /* print at least one line per page */
    }
    pages = (bodyLines + linesPerPage - 1) / linesPerPage;
    if (ppages) {
        *ppages = pages;
    }

    /* compute column widths */
    for (col = 0; col < cols; col++) {
        colWidths[col] = 0.0;
    }
    if (header) {
        if (HPDF_Page_SetFontAndSize(page, bfont, fsize)) {
            return hpdfError(pdf);
        }
        for (col = 0; col < cols; col++) {
            colWidths[col] = textWidth(page,
                    header->ql_Entries[col]->qe_String, memPool);
        }
        if (HPDF_Page_SetFontAndSize(page, font, fsize)) {
            return hpdfError(pdf);
        }
    }

    for (row = 1; row <= count; row++) {
        const struct QueryLine *ql = GuiSFWListQueryLine(sfw, row);

        if (ql) {
            for (col = 0; col < cols; col++) {
                w = textWidth(page, ql->ql_Entries[col]->qe_String, memPool);
                if (w > colWidths[col]) {
                    colWidths[col] = w;
                }
            }
        }
    }

    for (col = 0; col < cols; col++) {
        ocol = col;
        if (columnOrder) {
            ocol = columnOrder[col];
            if (ocol < 0) {
                continue;
            }
        }
        if (haveFirstCol) {
            totalWidth += hspace;
        }
        totalWidth += colWidths[ocol];
        haveFirstCol = TRUE;
    }
    if (haveHeader) {
        w2 = textWidth(page, hpdfs->hpdfs_Header, memPool);
        if (w2 > totalWidth) {
            totalWidth = w2;
        }
    }
    if (haveFooter) {
        w2 = textWidth(page, hpdfs->hpdfs_Footer, memPool);
        if (w2 > totalWidth) {
            totalWidth = w2;
        }
    }
    if (ptruncated) {
        *ptruncated = left + totalWidth - HPDF_Page_GetWidth(page);
    }

    if (filename == NULL) {
        return NULL;            /* only stats are needed */
    }

    row = 1;
    for (p = 0; p < pages; p++) {
        double x, y;
        const char *text;

        if (p != 0) {
            page = addPage(pdf, hpdfs, font); 
            if (page == NULL) {
                return hpdfError(pdf);
            }
        }
        y = HPDF_Page_GetHeight(page) - top;
	if (haveHeader) {
	    x = left;
	    if (HPDF_Page_BeginText(page)) {
		return hpdfError(pdf);
    	    }
	    error = textOut(pdf, page, x, y - ascent, 0.0, FALSE, th,
		    hpdfs->hpdfs_Header, 0, 1, memPool, warnings);
	    if (error) {
		return error;
	    }
	    if (HPDF_Page_EndText(page)) {
		return hpdfError(pdf);
	    }
            y -= 2*th;
	}
        if (header) {
            if (HPDF_Page_SetFontAndSize(page, bfont, fsize)) {
                return hpdfError(pdf);
            }
            x = left;
            for (col = 0; col < cols; col++) {
                ocol = col;
                if (columnOrder) {
                    ocol = columnOrder[col];
                    if (ocol < 0) {
                        continue;
                    }
                }
                text = header->ql_Entries[ocol]->qe_String;
                if (HPDF_Page_BeginText(page)) {
                    return hpdfError(pdf);
                }
                error = textOut(pdf, page, x, y - ascent, colWidths[ocol],
                        infos && infos[ocol].qfi_Format > 0, th, text, 0,
                        headerLines, memPool, warnings);
                if (error) {
                    return error;
                }
                if (HPDF_Page_EndText(page)) {
                    return hpdfError(pdf);
                }
                x += colWidths[ocol] + hspace;
            }
            if (HPDF_Page_SetFontAndSize(page, font, fsize)) {
                return hpdfError(pdf);
            }

            y -= headerLines * th;
            HPDF_Page_SetLineWidth(page, 0.5);
            HPDF_Page_MoveTo(page, left, y);
            HPDF_Page_LineTo(page, left + totalWidth, y);
            HPDF_Page_Stroke(page);
            HPDF_ResetError(pdf);       /* ingore error */
            y -= th / 2;
        }

        line = 0;
        while (line < linesPerPage && row <= count) {
            const struct QueryLine *ql = GuiSFWListQueryLine(sfw, row);
            int linesLeftOnPage = linesPerPage - line;
            int lines = QueryLineLines(ql, cols) - offset;
            int maxLines = (lines < linesLeftOnPage)? lines : linesLeftOnPage;

            if (!ql) {
                break;
            }
            x = left;
            for (col = 0; col < cols; col++) {
                ocol = col;
                if (columnOrder) {
                    ocol = columnOrder[col];
                    if (ocol < 0) {
                        continue;
                    }
                }
                text = ql->ql_Entries[ocol]->qe_String;
                if (HPDF_Page_BeginText(page)) {
                    return hpdfError(pdf);
                }
                error = textOut(pdf, page, x, y - ascent, colWidths[ocol],
                        infos && infos[ocol].qfi_Format > 0, th, text,
                        offset, maxLines, memPool, warnings);
                if (error) {
                    return error;
                }
                if (HPDF_Page_EndText(page)) {
                    return hpdfError(pdf);
                }
                x += colWidths[ocol] + hspace;
            }
            y -= th * maxLines;
            line += maxLines;

            if (maxLines < lines) {
                offset += maxLines;
            } else {
                row++;
                offset = 0;
                if (separateEntries) {
                    HPDF_Page_SetLineWidth(page, 0.5);
                    HPDF_Page_MoveTo(page, left, y + 1.0);
                    HPDF_Page_LineTo(page, left + totalWidth, y + 1.0);
                    HPDF_Page_Stroke(page);
                    HPDF_ResetError(pdf);       /* ingore error */
                }
            }
        }
        
        y = bot + fsize;
	if (haveFooter) {
	    x = left;
	    if (HPDF_Page_BeginText(page)) {
		return hpdfError(pdf);
    	    }
	    error = textOut(pdf, page, x, y - ascent, 0, FALSE, th,
		    hpdfs->hpdfs_Footer, 0, 1, memPool, warnings);
	    if (error) {
		return error;
	    }
	    if (HPDF_Page_EndText(page)) {
		return hpdfError(pdf);
	    }
	}
        /* Page number */
        sprintf(buffer, GCS(MSG_PageFooter), p + 1, pages);
        w = HPDF_Page_TextWidth(page, buffer);
        if (HPDF_Page_BeginText(page)) {
            return hpdfError(pdf);
        }
        x = (HPDF_Page_GetWidth(page) - hspace / 2 - w);
	if (!haveFooter) {
	    x /= 2;
	}
        error = textOut(pdf, page, x, y - ascent, 0, FALSE, th, buffer, 0, 1,
                memPool, warnings);
        if (error) {
            return error;
        }
        if (HPDF_Page_EndText(page)) {
            return hpdfError(pdf);
        }
    }

    if (HPDF_SaveToFile(pdf, filename)) {
        return hpdfError(pdf);
    }

    return NULL;
}

static const char *exportSFWListAsPDFHelp(struct Project *pr, GuiObject sfw, 
        const struct HPDFSettings *hpdfs, const char *filename,
        size_t *warnings, long *pages, double *truncated)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    int cols = GuiSFWListNumColumns(sfw);
    double *colWidths;
    const char *error;
    HPDF_Doc pdf;

#ifdef AMIGA
    if (HPDFBase == NULL) {
        return GES(ERR_INTERNAL);
    }
#endif
    if (pr == NULL || sfw == NULL || hpdfs == NULL) {
        return GES(ERR_INTERNAL);
    }

    if (cols <= 0) {
        cols = 1;
    }
    colWidths = PoolAllocMem(memPool, cols * sizeof(double));
    if (colWidths == NULL) {
        error = GES(ERR_MEM);
    } else {
        pdf = HPDF_New(NULL, NULL); 
        if (pdf) {
            error = exportSFWListAsPDFHelp2(pdf, pr, sfw, colWidths, TRUE,
                    hpdfs, filename, warnings, pages, truncated);
            HPDF_Free(pdf);
        } else {
            error = GES(ERR_MEM);
        }
        PoolFreeMem(memPool, colWidths, cols * sizeof(double));
    }

    return error;
}

const char *GetExportSFWListAsPDFStats(struct Project *pr, GuiObject sfw, 
        const struct HPDFSettings *hpdfs, long *pages, double *truncated)
{
    return exportSFWListAsPDFHelp(pr, sfw, hpdfs, NULL, NULL, pages, truncated);
}

bool ExportSFWListAsPDF( struct Project *pr, GuiObject sfw, 
        const struct HPDFSettings *hpdfs, const char *filename)
{
    size_t warnings = 0;
    const char *error;
   
    if (filename == NULL) {
        return FALSE;
    }
    ProjectSetCurrentDir(pr);           /* make sure we have correct CWD */
    if (FileStat(filename) >= 0 &&      /* file exists? */
            Ask(AT_WARN, GCS(MSG_AskReplaceAnswers), 
                GCS(MSG_AskReplaceFile), filename) == 0) {
        return FALSE;
    }
    error = exportSFWListAsPDFHelp(pr, sfw, hpdfs, filename,&warnings,NULL,NULL);

    if (error) {
        Ask(AT_ERROR, GCS(MSG_StarOk), error, filename);
    } else if (warnings) {
        Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Warn_CodesetProcessFile),
                warnings, filename);
    }

    return (error == NULL);
}

#endif

