/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** preferences.c
**
** Creation date: 27.03.1995
** Last modified: 07.06.2021
*/

#include "system.h"
#include "main.h"
#include "project.h"
#include "display.h"
#include "preferences.h"
#include "strhandle.h"
#include "memory.h"
#include "appstrings_cat.h"
#include "error.h"
#include "requesters.h"
#include "textfile.h"
#include "util.h"
#include "lang/program.h"

#ifdef AMIGA
#define PREFS_ENVFILENAME       "ENV:BeeBase.prefs"
#define PREFS_ENVARCFILENAME    "ENVARC:BeeBase.prefs"
#ifdef __AROS__
#define DEFAULT_EXTERNALEDITOR "Edit %f"
#else
#define DEFAULT_EXTERNALEDITOR "Ed %f"
#endif
#ifdef USE_EXTERNALVIEWER
#if (defined __MORPHOS__)
#define DEFAULT_EXTERNALVIEWER "Open %f"
#else
#define DEFAULT_EXTERNALVIEWER "Multiview %f"
#endif
#endif
#define DEFAULT_HELPFILE "BeeBase:Docs/BeeBase.guide"
#elif defined UNIX
#define PREFS_ENVFILENAME       "$HOME/.BeeBase.prefs"
#define DEFAULT_EXTERNALEDITOR "gvim -f %f"
#ifdef USE_EXTERNALVIEWER
#define DEFAULT_EXTERNALVIEWER "xdg-open %f"
#endif
#define DEFAULT_HELPFILE NULL
#elif defined WINDOWS
#define PREFS_ENVFILENAME       "$HOME/.BeeBase.prefs"
#define DEFAULT_EXTERNALEDITOR  "notepad %f"
#ifdef USE_EXTERNALVIEWER
#define DEFAULT_EXTERNALVIEWER  "%f"
#endif
#define DEFAULT_HELPFILE NULL
#elif defined MACOSX
#define PREFS_ENVFILENAME       "$HOME/.BeeBase.prefs"
#define DEFAULT_EXTERNALEDITOR  "open -tWn %f"
#ifdef USE_EXTERNALVIEWER
#define DEFAULT_EXTERNALVIEWER  "open %f"
#endif
#define DEFAULT_HELPFILE NULL
#endif

#define PREFS_VERSION 30

typedef union
{
    long val;
    char *str;
} PrefsValue;

struct ProjectPrefs
{
    struct MemPool *pp_MemPool;
    PrefsValue pp_Items[NumProjectPrefsItems];
};

struct GlobalPrefs
{
    short gp_Version;
    struct MemPool *gp_MemPool;
    PrefsValue gp_Items[NumPrefsItems];
    ulong gp_CreationDate;
};

struct BeeBasePrefs
{
    struct GlobalPrefs mbp_GlobalPrefs;
};

static struct PrefsAttr
{
    const char *pa_Name;
    bool pa_IsString;
    bool pa_NeedsSave;
    bool pa_ProjectOnly;
    PrefsValue pa_Init;
} PrefsAttr[] = 
{
    { "ConfirmDeleteRecord", FALSE, TRUE, FALSE, { .val = TRUE } },
    { "RecordMemory", FALSE, TRUE, FALSE, { .val = 0x7fffffff } },
    { "DefaultMaxLen", FALSE, FALSE, FALSE, { .val = 80 } },
    { "DefaultAlignmentText", FALSE, FALSE, FALSE, {.val=DAD_ALIGNMENT_CENTER}},
    { "DefaultAlignmentNumber", FALSE, FALSE, FALSE,{.val=DAD_ALIGNMENT_RIGHT}},
    { "PathsRelativeToProject", FALSE, TRUE, FALSE, { .val = FALSE } },
    { "ConfirmAutoReload", FALSE, TRUE, FALSE, { .val = TRUE } },
    { "ConfirmSaveReorg", FALSE, TRUE, FALSE, { .val = TRUE } },
    { "VacuumAfterReorg", FALSE, TRUE, FALSE, { .val = FALSE } },
    { "ProgramSourceExternal", FALSE, FALSE, TRUE, { .val = FALSE } },
    { "ProgramSourceExternalFilename", TRUE, TRUE, TRUE, { .str = NULL } },
    { "ProgramSourceExternalCleanup", FALSE, TRUE, FALSE, { .val = TRUE } },
    { "ProgramSortTriggerFunctions", FALSE, TRUE, FALSE, { .val = FALSE } },
    { "ProgramDebugInfo", FALSE, TRUE, FALSE, { .val = TRUE } },
    { "ProgramObsoleteFunctions", FALSE, TRUE, FALSE, { .val = OBSOLETE_WARN }},
    { "ProgramIncludeDir", TRUE, TRUE, FALSE, { .str = "BeeBase:Include" } },
    { "ProgramOutputFilename", TRUE, TRUE, FALSE, {.str = NULL }},
    { "ProgramOutputDefault", FALSE, TRUE, FALSE, { .val = TRUE } },
    { "ProgramOutputAppend", FALSE, TRUE, FALSE, { .val = FALSE } },
    { "ProgramEditorTop", FALSE, FALSE, FALSE, { .val = 0 } },
    { "ProgramEditorPosX", FALSE, FALSE, FALSE, { .val = 0 } },
    { "ProgramEditorPosY", FALSE, FALSE, FALSE, { .val = 0 } },
    { "QueryDir", TRUE, FALSE, FALSE, { .str = NULL } },
    { "TextEditorFilename", TRUE, FALSE, FALSE, { .str = NULL } },
    { "LogActive", FALSE, FALSE, FALSE, { .val = FALSE } },
    { "LogToProject", FALSE, FALSE, TRUE, { .val = TRUE } },
    { "LogToFile", FALSE, FALSE, TRUE, { .val = FALSE } },
    { "LogFilename", TRUE, FALSE, TRUE, { .str = NULL } },
    { "LogLabel", TRUE, FALSE, TRUE, { .str = NULL } },

    { "ProjectName", TRUE, TRUE, FALSE, { .str = NULL } },
    { "RealFormat", FALSE, TRUE, FALSE, { .val = 0 } },
    { "DateFormat", FALSE, TRUE, FALSE, { .val = 0 } },
    { "ExternalEditor", TRUE, TRUE, FALSE, { .str = DEFAULT_EXTERNALEDITOR } },
#ifdef USE_EXTERNALVIEWER
    { "ExternalViewer", TRUE, TRUE, FALSE, { .str = DEFAULT_EXTERNALVIEWER } },
#else
    { "ExternalViewer", TRUE, TRUE, FALSE, { .str = NULL } },
#endif
    { "ExtraButtonsInTabChain", FALSE, TRUE, FALSE, { .val = TRUE } },
    { "AdvanceOnEnter", FALSE, TRUE, FALSE, { .val = TRUE } },
    { "ConfirmQuit", FALSE, TRUE, FALSE, { .val = FALSE } },
    { "WarnCodeset", FALSE, TRUE, FALSE, { .val = TRUE } },
    { "HelpFilename", TRUE, TRUE, FALSE, { .str = DEFAULT_HELPFILE } },
    { "EditLabelsCompact", FALSE, TRUE, FALSE, { .val = FALSE } },
    { "StructureEditorLeftEdge", FALSE, TRUE, FALSE, { .val = -1 } },
    { "StructureEditorTopEdge", FALSE, TRUE, FALSE, { .val = -1 } },
    { "StructureEditorWidth", FALSE, TRUE, FALSE, { .val = -1 } },
    { "StructureEditorHeight", FALSE, TRUE, FALSE, { .val = -1 } },
    { "QueryEditorLeftEdge", FALSE, TRUE, FALSE, { .val = -1 } },
    { "QueryEditorTopEdge", FALSE, TRUE, FALSE, { .val = -1 } },
    { "QueryEditorWidth", FALSE, TRUE, FALSE, { .val = -1 } },
    { "QueryEditorHeight", FALSE, TRUE, FALSE, { .val = -1 } },
    { "ProgramEditorLeftEdge", FALSE, TRUE, FALSE, { .val = -1 } },
    { "ProgramEditorTopEdge", FALSE, TRUE, FALSE, { .val = -1 } },
    { "ProgramEditorWidth", FALSE, TRUE, FALSE, { .val = -1 } },
    { "ProgramEditorHeight", FALSE, TRUE, FALSE, { .val = -1 } },
#ifdef USE_LOG_WINDOW
    { "LogWindowLeftEdge", FALSE, TRUE, FALSE, { .val = -1 } },
    { "LogWindowTopEdge", FALSE, TRUE, FALSE, { .val = -1 } },
    { "LogWindowWidth", FALSE, TRUE, FALSE, { .val = -1 } },
    { "LogWindowHeight", FALSE, TRUE, FALSE, { .val = -1 } },
#endif
    { "PDFSettingsPaperSize", TRUE, TRUE, FALSE, { .str = NULL } },   
    { "PDFSettingsOrientation", TRUE, TRUE, FALSE, { .str = NULL } },
    { "HPDFSettingsFontName", TRUE, TRUE, FALSE, { .str = NULL } },   
    { "HPDFSettingsFontSize100", FALSE, TRUE, FALSE, { .val = 0 } },
    { "QtPDFSettingsFontName", TRUE, TRUE, FALSE, { .str = NULL } },   
    { "PDFSettingsHeader", TRUE, TRUE, FALSE, { .str = NULL } },   
    { "PDFSettingsFooter", TRUE, TRUE, FALSE, { .str = NULL } },   
#ifdef USE_GTK_PRINT
    { "GTKPrintSettingsFontName", TRUE, TRUE, FALSE, { .str = NULL } },   
    { "GTKPrintSettingsFitPageWidth", FALSE, TRUE, FALSE, { .val = FALSE } },   
#endif
};

static struct BeeBasePrefs BeeBasePrefs =
{
    { 0, NULL, { { 0 } } }				/* Global Prefs */
};

void FreeProjectPrefs(struct ProjectPrefs *pp)
{
    if (pp) {
	struct MemPool *memPool = pp->pp_MemPool;
	short i;

	for (i=0; i < NumProjectPrefsItems; i++) {
	    if (PrefsAttr[i].pa_IsString) {
		FreeString(memPool, pp->pp_Items[i].str);
	    }
	}
	PoolFreeMem(memPool, pp, sizeof(*pp));
    }
}

void ClearProjectPrefs(struct ProjectPrefs *pp)
{
    if (pp) {
	const PrefsValue *orig = BeeBasePrefs.mbp_GlobalPrefs.gp_Items;
	struct MemPool *memPool = pp->pp_MemPool;
	PrefsValue *items = pp->pp_Items;
	int i;

	for (i=0; i < NumProjectPrefsItems; i++) {
	    if (!PrefsAttr[i].pa_IsString) {
		items[i].val = orig[i].val;
	    } else if (items[i].str != orig[i].str) {
		FreeString(memPool, items[i].str);
		items[i].str = AllocString(memPool, orig[i].str);
	    }
	}
    }
}

struct ProjectPrefs *AllocProjectPrefs(struct MemPool *memPool)
{
    struct ProjectPrefs *pp = PoolAllocMemZero(memPool, sizeof(*pp));

    if (pp) {
	pp->pp_MemPool = memPool;
	ClearProjectPrefs(pp);
    }
    return pp;
}

static PrefsValue *findItem(struct Project *pr, ulong tag,struct MemPool **pool)
{
    if (tag < NumPrefsItems) {
	struct ProjectPrefs *pp = ProjectProjectPrefs(pr);
	struct GlobalPrefs *gp = &BeeBasePrefs.mbp_GlobalPrefs;
	struct MemPool *memPool = gp->gp_MemPool;
	PrefsValue *items = gp->gp_Items;

	if (tag < NumProjectPrefsItems && pp) {
	    items = pp->pp_Items;
	    memPool = pp->pp_MemPool;
	}
	if (pool) {
	    *pool = memPool;
	}
	return &items[tag];
    }
    return NULL;
}

static bool allowChangeItem(struct Project *pr, ulong tag)
{
    if (tag < NumPrefsItems && PrefsAttr[tag].pa_NeedsSave) {
	struct ProjectPrefs *pp = ProjectProjectPrefs(pr);

	if (tag < NumProjectPrefsItems && pp) {
            long error = ObtainProjectChangeLock(pr);

            if (error != OK) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
                return FALSE;
            }
	}
    }
    return TRUE;
}

static void saveItem(struct Project *pr, ulong tag)
{
    if (tag < NumPrefsItems && PrefsAttr[tag].pa_NeedsSave) {
	struct ProjectPrefs *pp = ProjectProjectPrefs(pr);

	if (tag < NumProjectPrefsItems && pp) {
	    ProjectChange(pr, FALSE);
	} else {
	    SavePrefs();
	}
    }
}

void SetPrefsValue(struct Project *pr, ulong tag, long value)
{
    PrefsValue *item = findItem(pr, tag, NULL);

    if (item && item->val != value && allowChangeItem(pr, tag)) {
	item->val = value;
	saveItem(pr, tag);
	GuiPrefsUpdated();
    }
}

bool SetPrefsString(struct Project *pr, ulong tag, const char *string)
{
    struct MemPool *memPool = NULL;
    PrefsValue *item = findItem(pr, tag, &memPool);
    bool rc = TRUE;

    if (item == NULL) {
        rc = FALSE;
    } else if (item->str != string && 
            (string == NULL || item->str == NULL || 
             strcmp(string, item->str) != 0) && allowChangeItem(pr, tag)) {
	FreeString(memPool, item->str);
        item->str = AllocString(memPool, string);
        rc = (string == NULL || item->str != NULL);
	saveItem(pr, tag);
	GuiPrefsUpdated();
    }
    return rc;
}

void SetPrefsWindowDimensions(struct Project *pr, ulong tag,
        const struct WindowDimensions *wd)
{
    PrefsValue *item = findItem(pr, tag, NULL);
    bool save;

    if (wd && item) {
        save = item[0].val != wd->wd_LeftEdge ||
            item[1].val != wd->wd_TopEdge ||
            item[2].val != wd->wd_Width ||
            item[3].val != wd->wd_Height;
        if (save && allowChangeItem(pr, tag)) {
            item[0].val = wd->wd_LeftEdge;
            item[1].val = wd->wd_TopEdge;
            item[2].val = wd->wd_Width;
            item[3].val = wd->wd_Height;
            saveItem(pr, tag);
        }
    }
}

void SetPrefsProjectDefaults(struct Project *pr)
{
    struct ProjectPrefs *pp = ProjectProjectPrefs(pr);
    struct GlobalPrefs *gp = &BeeBasePrefs.mbp_GlobalPrefs;
    struct MemPool *memPool = gp->gp_MemPool;
    PrefsValue *dst = gp->gp_Items, *src;
    short i;

    if (pp != NULL) {
	src = pp->pp_Items;
	for (i = 0; i < NumProjectPrefsItems; i++) {
	    if (PrefsAttr[i].pa_ProjectOnly) {
		continue;		/* skip project-only items */
	    }
	    if (PrefsAttr[i].pa_IsString && dst[i].str != src[i].str) {
		FreeString(memPool, dst[i].str);
		dst[i].str = AllocString(memPool, src[i].str);
	    } else {
		dst[i].val = src[i].val;
	    }
	}
	SavePrefs();
    }
}

long GetPrefsValue(struct Project *pr, ulong tag)
{
    const PrefsValue *item = findItem(pr, tag, NULL);

    if (item) {
	return item->val;
    }
    return -1; 		/* error */
}

const char *GetPrefsString(struct Project *pr, ulong tag)
{
    const PrefsValue *item = findItem(pr, tag, NULL);

    if (item) {
	return item->str;
    }
    return NULL;		/* error */
}

bool GetPrefsWindowDimensions(struct Project *pr, ulong tag,
        struct WindowDimensions *wd)
{
    const PrefsValue *item = findItem(pr, tag, NULL);
    int i;

    if (item && wd) {
        for (i = 0; i < 4; i++) {
            if (item[i].val < 0) {
                return FALSE;
            }
        }
        wd->wd_LeftEdge = item[0].val;
        wd->wd_TopEdge = item[1].val;
        wd->wd_Width = item[2].val;
        wd->wd_Height = item[3].val;

        return TRUE;
    }
    return FALSE;
}

static long readPrefsString(FILE *in, bool utf8,
        PrefsValue *item, struct MemPool *memPool) 
{ 
    long error = OK;
    size_t warnings = 0;                /* currently ignored */

    FreeString(memPool, item->str);
    item->str = ReadStringConvert(in, utf8, memPool, &error, &warnings);
    return error;
}

static long oldreadprojectprefs(short *pversion, bool *putf8, 
	PrefsValue *items, struct MemPool *memPool, FILE *in)
{
    short version;
    bool utf8 = FALSE;

    version = FGetShort(in);
    if (version < 3 || version > PREFS_VERSION) {
	return DISKERR_VERSION;
    }
    *pversion = version;

    if (version >= 21) {
        utf8 = FGetBool(in);
    }
    *putf8 = utf8;

    if (version <= 10) {
	ushort flags = FGetShort(in);
	items[PREFS_CONFIRMDELETERECORD].val = ((flags & 0x8000) != 0);
	items[PREFS_PROGRAMDEBUGINFO].val = ((flags & 0x8000) != 0);
	items[PREFS_PROGRAMOUTPUT_APPEND].val = ((flags & 0x0400) != 0);
	items[PREFS_PROGRAMSOURCEEXTERNAL].val = ((flags & 0x0200) != 0);
	items[PREFS_CONFIRMSAVEREORG].val = ((flags & 0x0100) != 0);

	if (version < 7) {
	    items[PREFS_PROGRAMDEBUGINFO].val = TRUE;
	}
    } else {
	items[PREFS_CONFIRMDELETERECORD].val = FGetBool(in);
	items[PREFS_PROGRAMDEBUGINFO].val = FGetBool(in);
        if (version >= 27) {
            items[PREFS_PROGRAMOBSOLETEFUNCTIONS].val = FGetChar(in);
        }
        items[PREFS_PROGRAMOUTPUT_DEFAULT].val = FALSE;
        if (version >= 29) {
            items[PREFS_PROGRAMOUTPUT_DEFAULT].val = FGetBool(in);
        }
	items[PREFS_PROGRAMOUTPUT_APPEND].val = FGetBool(in);
	items[PREFS_PROGRAMSOURCEEXTERNAL].val = FGetBool(in);
	if (version >= 18) {
	    items[PREFS_PROGRAMSOURCEEXTERNAL_CLEANUP].val = FGetBool(in);
	}
        if (version >= 30) {
            items[PREFS_CONFIRMAUTORELOAD].val = FGetBool(in);
        }
	items[PREFS_CONFIRMSAVEREORG].val = FGetBool(in);
    }
    items[PREFS_RECORDMEMORY].val = FGetLong(in);
    if (version >= 18) {
	readPrefsString(in, utf8, &items[PREFS_PROGRAMSOURCEEXTERNAL_FILENAME], 
		memPool);
    } else {
	items[PREFS_PROGRAMSOURCEEXTERNAL].val = FALSE;
    }
    readPrefsString(in, utf8, &items[PREFS_QUERYDIR], memPool);
    if (version >= 6) {
	readPrefsString(in, utf8, &items[PREFS_PROGRAMINCLUDEDIR], memPool);
    }
    readPrefsString(in, utf8, &items[PREFS_PROGRAMOUTPUT_FILENAME], memPool);
    if (version >= 10) {
	readPrefsString(in, utf8, &items[PREFS_TEXTEDITOR_FILENAME], memPool);
    }
    items[PREFS_DEFAULTALIGNMENT_TEXT].val = FGetChar(in);
    items[PREFS_DEFAULTALIGNMENT_NUMBER].val = FGetChar(in);
    items[PREFS_PROGRAMEDITOR_TOP].val = FGetLong(in);
    items[PREFS_PROGRAMEDITOR_POSX].val = FGetLong(in);
    items[PREFS_PROGRAMEDITOR_POSY].val = FGetLong(in);
    if (version > 3) {
	items[PREFS_DEFAULTMAXLEN].val = FGetShort(in);
    }
    if (version > 4 && version < 27) {
	FGetChar(in);           /* unused1 */
	FGetChar(in);           /* unused2 */
    }
    if (version > 14) {
	items[PREFS_PROGRAMSORTTRIGGERFUNCTIONS].val = FGetBool(in);
    }
    if (version > 23) {
	items[PREFS_PATHSRELATIVETOPROJECT].val = FGetBool(in);
    }
    if (version > 27) {
	items[PREFS_LOG_ACTIVE].val = FGetBool(in);
	readPrefsString(in, utf8, &items[PREFS_LOG_LABEL], memPool);
    }

    return OK;
}

long LoadProjectPrefsCustomFormat(struct ProjectPrefs *pp, FILE *in)
{
    short version;
    bool utf8;

    if (in == NULL || pp == NULL) {
	return ERR_INTERNAL;
    }
    return oldreadprojectprefs(&version, &utf8,
	    pp->pp_Items, pp->pp_MemPool, in);
}

static void parsePrefsSettings(const char *s, PrefsValue *items, int numItems,
        bool allowProjectOnly, struct MemPool *memPool)
{
    const char *end;

    for (; s != NULL; s = end) {
	const struct PrefsAttr *pa = PrefsAttr;
	int i;

	/* skip white space */
	while (CodesetIsSpace(CodesetGetChar(s))) {
	    s = CodesetNextChar(s);
	}
	end = CodesetStrchr(s, '\n');

	/* test if comment */
	if (*s == '#' || (s[0] == '/' && s[1] == '/')) {
	    continue;
	}

	for (i=0; i < numItems; i++, pa++) {
	    size_t len = strlen(pa->pa_Name);
	    const char *q, *t;
	    char *buf;

	    if (!allowProjectOnly && pa->pa_ProjectOnly) {
		continue;		/* skip project-only items */
	    }
	    if (strncmp(s, pa->pa_Name, len) == 0 && s[len] == ':') {
		s += len + 1;
		while (CodesetIsSpace(CodesetGetChar(s)) && *s != '\n') {
	    	    s = CodesetNextChar(s);
    		}
		t = end;
		if (t == NULL) {
		    t = s + strlen(s);
		}
		if (*s == '"') {
		    /* find last double quote in current line */
		    for (q = t; q > s; q--) {
			if (*q == '\"') {
		            s++;
			    t = q;
			    break;
			}
		    }
		}
		buf = AllocStringN(memPool, s, t-s);
		if (pa->pa_IsString) {
		    FreeString(memPool, items[i].str);
		    items[i].str = buf;
		} else {
		    StrToInt(buf, &items[i].val);
		    FreeString(memPool, buf);
		}
	    }
	}
    }
}

static char *createPrefsSettings(const PrefsValue *items, int numItems,
	struct MemPool *memPool)
{
    struct StrHandle *sh = AllocStrHandle(memPool);
    const struct PrefsAttr *pa = PrefsAttr;
    char *settings = NULL;
    int i;

    for (i = 0; i < numItems; i++, pa++) {
	if (pa->pa_IsString) {
	    const char *str = EmptyNullStr(items[i].str);
	    const char *init = EmptyNullStr(pa->pa_Init.str);
	    if (str != init && strcmp(str, init) != 0) {
		StrHandlePutS(sh, pa->pa_Name);
		StrHandlePutC(sh, ':');
		if (str[0] != 0) {
		    StrHandlePutC(sh, ' ');
		    StrHandlePutC(sh, '"');
		    StrHandlePutS(sh, str);
		    StrHandlePutC(sh, '"');
		}
		StrHandlePutC(sh, '\n');
	    }
	} else if (items[i].val != pa->pa_Init.val) {
	    StrHandlePutS(sh, pa->pa_Name);
    	    StrHandlePutS(sh, ": ");
	    StrHandlePutS(sh, IntToStr(items[i].val));
	    StrHandlePutC(sh, '\n');
	}
    }
    StrHandlePutC(sh, 0);

    if (!StrHandleError(sh)) {
	settings = AllocString(memPool, StrHandleString(sh));
    }
    FreeStrHandle(sh);

    return settings;
}

char *GetProjectPrefsSettings(struct ProjectPrefs *pp, struct MemPool *memPool)
{
    char *settings = NULL;

    if (pp) {
	settings = createPrefsSettings(
		pp->pp_Items, NumProjectPrefsItems, memPool);
    }
    return settings;
}

void SetProjectPrefsSettings(struct ProjectPrefs *pp, const char *settings)
{
    if (pp && settings) {
        parsePrefsSettings(settings, pp->pp_Items, NumPrefsItems, TRUE,
                pp->pp_MemPool);
    }
}

void SavePrefs(void)
{
    struct GlobalPrefs *gp = &BeeBasePrefs.mbp_GlobalPrefs;
    const PrefsValue *items = gp->gp_Items;
    struct MemPool *memPool = gp->gp_MemPool;
    char *settings = createPrefsSettings(items, NumPrefsItems, memPool);

    GuiSetSleep(TRUE);

    if (settings != NULL) {
	const char *filename = PREFS_ENVFILENAME;
        long error;

        TextFileSetContentsAskWarnings(NULL, memPool, filename,settings,&error);
        if (error) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GES(error), filename);
	}

#ifdef PREFS_ENVARCFILENAME
	filename = PREFS_ENVARCFILENAME;
        TextFileSetContentsAskWarnings(NULL, memPool, filename,settings,&error);
        if (error) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GES(error), filename);
	}
#endif
	FreeString(memPool, settings);
    } else {
	Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
    }

    GuiSetSleep(FALSE);
}

void LoadPrefs(void)
{
    struct BeeBasePrefs *mbp = &BeeBasePrefs;
    struct GlobalPrefs *gp = &mbp->mbp_GlobalPrefs;
    struct MemPool *memPool = gp->gp_MemPool;
    PrefsValue *items = gp->gp_Items;
    char *settings;
    long error;

    GuiSetSleep(TRUE);
    settings = TextFileGetContentsAskWarnings(NULL, memPool,
            PREFS_ENVFILENAME, &error);
    if (error && error != DISKERR_OPEN && error != ERR_CANCEL) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_LoadPreferences));
    }
    if (settings) {
        parsePrefsSettings(settings, items, NumPrefsItems, FALSE, memPool);
        FreeString(memPool, settings);
    }
    GuiSetSleep(FALSE);
    GuiPrefsUpdated();
}

void PrefsInit(void)
{
    struct GlobalPrefs *gp = &BeeBasePrefs.mbp_GlobalPrefs;
    PrefsValue *items = gp->gp_Items, init;
    struct MemPool *memPool;
    short i;

    PrefsAttr[PREFS_REALFORMAT].pa_Init.val = DefaultRealFormat;
    PrefsAttr[PREFS_DATEFORMAT].pa_Init.val = DefaultDateFormat;

    gp->gp_MemPool = memPool = CreatePool(0L, 4096, 1024);

    for (i=0; i < NumPrefsItems; i++) {
	init = PrefsAttr[i].pa_Init;
	if (PrefsAttr[i].pa_IsString) {
	    items[i].str = AllocString(memPool, init.str);
	} else {
	    items[i].val = init.val;
	}
    }

    gp->gp_CreationDate = DateToday();
}

void PrefsExit()
{
    struct GlobalPrefs *gp = &BeeBasePrefs.mbp_GlobalPrefs;

    if (gp->gp_MemPool) {
	DeletePool(gp->gp_MemPool);
	gp->gp_MemPool = NULL;
    }
}

