/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** project.c
**
** Creation date: 10.11.1994
** Last modified: 04.05.2022
*/

#include "system.h"
#include "project.h"
#include "main.h"
#include "menuproject.h"
#include "sqlitedb.h"
#include "server.h"
#include "log.h"
#include "table.h"
#include "order.h"
#include "filter.h"
#include "field.h"
#include "display.h"
#include "buildgui.h"
#include "structureeditor.h"
#include "programeditor.h"
#include "queryeditor.h"
#include "preferences.h"
#include "error.h"
#include "gui.h"
#include "break.h"
#include "memory.h"
#include "dosutil.h"
#include "fileinfo.h"
#include "filelock.h"
#include "requesters.h"
#include "strhandle.h"
#include "util.h"
#include "lang/heap.h"
#include "appstrings_cat.h"
#include "version.h"

struct Project
{
    struct Node pr_Node;
    struct MemPool *pr_MemPool;
    struct MemPool *pr_DataMemPool;
    struct Log *pr_Log;
    struct List pr_Tables;
    struct DisplayObject *pr_RootDisplayObject;
    bool pr_AdminMode;
    char *pr_AdminPassword;
    GuiObject pr_StructureEditor;
    GuiObject pr_ProgramEditor;
    struct QueryEditor *pr_QueryEditor;
    struct Program *pr_Program;
    long pr_Changes;
    long pr_StructureChanges;
    size_t pr_CodesetWarningChars;
    size_t pr_CodesetWarningFields;
    bool pr_CodesetWarned;
    char *pr_Name;
    struct SQLiteHandle *pr_SQLiteHandle;
    long pr_Version;
    long pr_StructureVersion;
    long pr_FileVersion;
    long pr_ExclusiveHandle;
    struct ProjectPrefs *pr_ProjectPrefs;
    long pr_RecordMem;
};

static struct Project *CwdProject = NULL;
static Cwd ProjectCwd = NULL;

void ProjectSetCurrentDir(struct Project *pr)
{
    if (pr != CwdProject) {
	CwdProject = pr;
	ProjectUpdateCurrentDir();
    }
}

void ProjectUpdateCurrentDir(void)
{
    struct Project *pr = CwdProject;
    const char *path = ProjectName(pr);
    struct MemPool *memPool = ProjectMemPool(pr);
    char *dir;

    PopDir(&ProjectCwd);
    if (GetPrefsValue(pr, PREFS_PATHSRELATIVETOPROJECT) && path != NULL &&
	    SplitPathname(path, memPool, &dir, NULL) && dir != NULL) {
	PushDir(dir, &ProjectCwd);
	FreeString(memPool, dir);
    }
}

void FreeProject(struct Project *pr)
{
    if (pr) {
        struct Table *tab;

        if (pr == CwdProject) {
            CwdProject = NULL;
        }

	ServerRemoveProjectConnections(pr);
	pr->pr_ExclusiveHandle = HANDLE_NONE;

        FreeQueryEditor(pr->pr_QueryEditor);
        pr->pr_QueryEditor = NULL;
        CloseProgramEditor(pr->pr_ProgramEditor);
        pr->pr_ProgramEditor = NULL;
	CloseStructureEditor(pr->pr_StructureEditor);
        pr->pr_StructureEditor = NULL;
        FreeSQLiteHandle(pr->pr_SQLiteHandle);
        pr->pr_SQLiteHandle = NULL;
	FreeLog(pr->pr_Log);
        pr->pr_Log = NULL;
        FreeProjectGui(pr);

        FreeProgram(pr->pr_Program);
        pr->pr_Program = NULL;
	FreeRootDisplayObject(pr->pr_RootDisplayObject);
        pr->pr_RootDisplayObject = NULL;
        FreeProjectPrefs(pr->pr_ProjectPrefs);
        pr->pr_ProjectPrefs = NULL;

        while ((tab = (struct Table *)REMHEAD(&pr->pr_Tables)) != NULL) {
	    FreeTable(tab, TRUE);
        }

	FreeString(pr->pr_MemPool, pr->pr_Name);
        pr->pr_Name = NULL;
	FreeString(pr->pr_MemPool, pr->pr_AdminPassword);
        pr->pr_AdminPassword = NULL;
	if (pr->pr_MemPool) {
	    DeletePool(pr->pr_MemPool);
            pr->pr_MemPool = NULL;
        }
 	if (pr->pr_DataMemPool) {
	    DeletePool(pr->pr_DataMemPool);
            pr->pr_DataMemPool = NULL;
        }
	FREEMEM(pr, sizeof(struct Project));
    }
}

struct Project *AllocProject(bool gui)
{
    struct Project *pr;

    pr = ALLOCMEM(sizeof(struct Project));
    if (pr == NULL) {
	return NULL;
    }

    memset(pr, 0, sizeof(struct Project));
    pr->pr_AdminMode = TRUE;
    NewList(&pr->pr_Tables);
    pr->pr_ExclusiveHandle = HANDLE_NONE;

    pr->pr_MemPool = CreatePool(0L, 4096, 1024);
    if (pr->pr_MemPool == NULL) {
	FreeProject(pr);
	return NULL;
    }

    pr->pr_DataMemPool = CreatePool(0L, 32 * 1024L, 4096);
    if (pr->pr_DataMemPool == NULL) {
	FreeProject(pr);
	return NULL;
    }

    pr->pr_SQLiteHandle = AllocSQLiteHandle(pr);
    if (pr->pr_SQLiteHandle == NULL) {
        FreeProject(pr);
	return NULL;
    }

    pr->pr_ProjectPrefs = AllocProjectPrefs(pr->pr_MemPool);
    if (pr->pr_ProjectPrefs == NULL) {
        FreeProject(pr);
	return NULL;
    }

    pr->pr_Log = AllocLog(pr);
    if (pr->pr_Log == NULL) {
	FreeProject(pr);
	return NULL;
    }

    pr->pr_RootDisplayObject = AllocRootDisplayObject(pr);
    if (pr->pr_RootDisplayObject == NULL) {
	FreeProject(pr);
	return NULL;
    }

    pr->pr_QueryEditor = AllocQueryEditor(pr);
    if (pr->pr_QueryEditor == NULL) {
	FreeProject(pr);
	return NULL;
    }

    if (gui && !BuildProjectGui(pr)) {
        FreeProject(pr);
        return NULL;
    }

    return pr;
}

bool ProjectIsEmpty(struct Project *pr)
{
    if (pr == NULL) {
	return TRUE;
    } else if (pr->pr_Changes > 0) {
	return FALSE;
    } else if (pr->pr_Name != NULL) {
	return FALSE;
    } else if (SUCC(HEAD(&pr->pr_Tables))) {
	return FALSE;
    } else {
        return TRUE;
    }
}

bool ProjectHasGui(struct Project *pr)
{
    struct DisplayObject *dio = ProjectRootDisplayObject(pr);
    GuiObject rw = DisplayObjectGuiObject(dio);
    GuiObject se = ProjectStructureEditor(pr);

    return (rw != NULL || se != NULL);
}

bool SetProjectHasGui(struct Project *pr, bool gui)
{
    if (gui == ProjectHasGui(pr)) {
	return gui;
    }
    else if (gui) {
	gui = BuildProjectGui(pr);
	return gui;
    } else {
	CloseStructureEditor(ProjectStructureEditor(pr));
    	SetProjectStructureEditor(pr, NULL);
	FreeProjectGui(pr);

	return FALSE;
    }
}

struct SQLiteHandle *ProjectSQLiteHandle(struct Project *pr)
{
    struct SQLiteHandle *handle = NULL;

    if (pr) {
        handle = pr->pr_SQLiteHandle;
    }
    return handle;
}

struct Log *ProjectLog(struct Project *pr)
{
    struct Log *log = NULL;

    if (pr) {
	log = pr->pr_Log;
    }
    return log;
}

struct List *ProjectTables(struct Project *pr)
{
    struct List *tabs = NULL;

    if (pr) {
	tabs = &pr->pr_Tables;
    }

    return tabs;
}

void AddTable(struct Project *pr, struct Table *tab)
{
    if (pr && tab) {
	ADDTAIL(&pr->pr_Tables, tab);
    }
}

void RemTable(struct Project *pr, struct Table *tab)
{
    if (pr && tab) {
	REMOVE(tab);
    }
}

struct MemPool *ProjectMemPool(struct Project *pr)
{
    struct MemPool *pool = NULL;

    if (pr) {
	pool = pr->pr_MemPool;
    }
    return pool;
}

struct MemPool *ProjectDataMemPool(struct Project *pr)
{
    struct MemPool *pool = NULL;

    if (pr) {
	pool = pr->pr_DataMemPool;
    }
    return pool;
}

bool ProjectAdminMode(struct Project *pr)
{
    bool admin = TRUE;

    if (pr) {
	admin = pr->pr_AdminMode;
    }
    return admin;
}

void SetProjectAdminMode(struct Project *pr, bool admin)
{
    if (pr) {
	GuiObject pe = ProjectProgramEditor(pr);
	struct Program *prg = ProjectProgram(pr);
	struct Element *args[2];
        GuiObject se = NULL;

	if (admin == FALSE) {
	    if (pe) {
		if (!ProgramEditorCheckClose(pe)) { 
		    return;
		}
		CloseProgramEditor(pe);
	    }
	    se = ProjectStructureEditor(pr);
	    if (se != NULL) {
		CloseStructureEditor(se);
                se = ProjectStructureEditor(pr);
		if (se != NULL) {
		    return;
		}
	    }
	    if (ProjectProgramEditor(pr)) {
                return;     /* some error showed up in the programming editor */
            }
	    CloseQueryEditor(pr);
#ifdef USE_LOG_WINDOW
            LogCloseWindow(pr->pr_Log);
#endif
	}

	pr->pr_AdminMode = admin;

	if (!ProgramRunning() && FindUserFunctionName(prg, onAdminMode)) {
	    args[0] = admin? &ElementTrue : &ElementNil;
	    args[1] = NULL;

	    ExecuteTrigger(pr, onAdminMode, args, NULL);
	}
        GuiUpdateDependentDisplayObjects(pr, pr, NULL, NULL);
    }
}

const char *ProjectAdminPassword(struct Project *pr)
{
    const char *passwd = NULL;

    if (pr) {
	passwd = pr->pr_AdminPassword;
    }
    return passwd;
}

bool SetProjectAdminPassword(struct Project *pr, const char *passwd)
{
    char *copy = NULL;

    if (pr == NULL) {
	return FALSE;
    }

    if (passwd) {
	copy = AllocString(pr->pr_MemPool, passwd);
	if (copy == NULL) {
	    return FALSE;
	}
    }

    FreeString(pr->pr_MemPool, pr->pr_AdminPassword);
    pr->pr_AdminPassword = copy;

    return TRUE;
}

GuiObject ProjectStructureEditor(struct Project *pr)
{
    GuiObject se = NULL;

    if (pr) {
	se = pr->pr_StructureEditor;
    }
    return se;
}

void SetProjectStructureEditor(struct Project *pr, GuiObject se)
{
    if (pr) {
	pr->pr_StructureEditor = se;
    }
}

GuiObject ProjectProgramEditor(struct Project *pr)
{
    GuiObject pe = NULL;

    if (pr) {
	pe = pr->pr_ProgramEditor;
    }
    return pe;
}

void SetProjectProgramEditor(struct Project *pr, GuiObject pe)
{
    if (pr) {
	pr->pr_ProgramEditor = pe;
    }
}

struct Program *ProjectProgram(struct Project *pr)
{
    struct Program *prg = NULL;

    if (pr) {
        prg = pr->pr_Program;
    }
    return prg;
}

void SetProjectProgram(struct Project *pr, struct Program *prg)
{
    if (pr) {
	struct Table *tab;

        FreeProgram(pr->pr_Program);
        pr->pr_Program = prg;

	FORNODE(tab, &pr->pr_Tables) {
	    SetProgramNeedRecompile(TableFilterProgram(tab), TRUE);
	}
    }
}

struct QueryEditor *ProjectQueryEditor(struct Project *pr)
{
    struct QueryEditor *qe = NULL;

    if (pr) {
	qe = pr->pr_QueryEditor;
    }
    return qe;
}

struct DisplayObject *ProjectRootDisplayObject(struct Project *pr)
{
    struct DisplayObject *dio = NULL;

    if (pr) {
	dio = pr->pr_RootDisplayObject;
    }
    return dio;
}

GuiObject ProjectRootWindow(struct Project *pr)
{
    struct DisplayObject *dio = ProjectRootDisplayObject(pr);
    GuiObject rw = DisplayObjectGuiObject(dio);

    if (rw == NULL) {
        rw = ProjectStructureEditor(pr);
    }
    return rw;
}


long AddProjectRecordMem(struct Project *pr, long size)
{
    long ret = -1;

    if (pr) {
        pr->pr_RecordMem += size;
        ret = pr->pr_RecordMem;
    }
    return ret;
}

bool OpenProjectWindows(struct Project *pr)
{
    struct DisplayObject *dio = ProjectRootDisplayObject(pr);
    GuiObject rw = DisplayObjectGuiObject(dio);

    if (rw) {
        GuiUpdate(pr);
        if (GuiOpenWindow(rw)) {
            SetDisplayFields(pr);
	    return TRUE;
	}
    }

    return FALSE;
}

long ProjectChanges(struct Project *pr)
{
    long changes = 0;

    if (pr) {
        changes = pr->pr_Changes;
    }
    return changes;
}

long ProjectStructureChanges(struct Project *pr)
{
    long changes = 0;

    if (pr) {
        changes = pr->pr_StructureChanges;
    }
    return changes;
}

void ProjectAddCodesetWarnings(struct Project *pr, size_t warnings)
{
    if (pr != NULL && warnings > 0) {
        pr->pr_CodesetWarningChars += warnings;
        pr->pr_CodesetWarningFields++;
    }
}

size_t ProjectCodesetWarningFields(struct Project *pr)
{
    size_t warnings = 0;

    if (pr) {
        warnings = pr->pr_CodesetWarningFields;
    }
    return warnings;
}

size_t ProjectCodesetWarningChars(struct Project *pr)
{
    size_t warnings = 0;

    if (pr) {
        warnings = pr->pr_CodesetWarningChars;
    }
    return warnings;
}

void ProjectWarnCodeset(struct Project *pr)
{
    if (pr && !pr->pr_CodesetWarned && GetPrefsValue(pr, PREFS_WARNCODESET)) {
        Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_CodesetWarnings),ProjectName(pr));
        pr->pr_CodesetWarned = TRUE;
    }
}

void ProjectClearCodesetWarnings(struct Project *pr)
{
    if (pr != NULL) {
        pr->pr_CodesetWarningFields = 0;
        pr->pr_CodesetWarningChars = 0;
        pr->pr_CodesetWarned = FALSE;
    }
}

static void callOnChange(struct Project *pr)
{
    if (!ProgramRunning() &&
            FindUserFunctionName(ProjectProgram(pr), onChange)) {
        ExecuteTrigger(pr, onChange, NULL, NULL);
    }
}

void ProjectChange(struct Project *pr, bool structure)
{
    if (pr) {
        if (!SQLiteAllowChanges(ProjectSQLiteHandle(pr))) {
            Ask(AT_ERROR, GCS(MSG_StarOk),
                    GCS(MSG_Err_ProjectChangeWithoutChangeLock),
                    ProjectName(pr));
        }
	pr->pr_Changes++;
        if (structure) {
            pr->pr_StructureChanges++;
        }
        SetProjectPrependTitle(pr);
        callOnChange(pr);
    }
}

void ProjectClearChanges(struct Project *pr)
{
    if (pr && pr->pr_Changes != 0) {
	pr->pr_Changes = 0;
	pr->pr_StructureChanges = 0;
        SetProjectPrependTitle(pr);
        callOnChange(pr);
	ServerCheckRequests();
    }
}

const char *ProjectName(const struct Project *pr)
{
    const char *name = NULL;

    if (pr) {
	name = pr->pr_Name;
    }
    return name;
}

bool SetProjectName(struct Project *pr, const char *name)
{
    char *new_name = NULL;

    if (pr == NULL) {
	return FALSE;
    }

    if (name) {
	long len = strlen(name);

	while (len > 0 && IsPathSeparator(name[len-1])) {
	    len--;
        }
        new_name = AllocStringN(pr->pr_MemPool, name, len);
	if (new_name == NULL) {
	    return FALSE;
        }
    }
    FreeString(pr->pr_MemPool, pr->pr_Name);
    pr->pr_Name = new_name;

    SetDisplayWindowId(ProjectRootDisplayObject(pr), StrToId(ProjectTitle(pr)));
    SetRootWindowTitle(pr, ProjectRootWindow(pr));

    ProjectUpdateCurrentDir();

    return TRUE;
}

const char *ProjectTitle(struct Project *pr)
{
    const char *title = ProjectName(pr);
    const char *s;

    if (title) {
        while ((s = CodesetStrpbrk(title, "/\\:")) != NULL) {
            title = s + 1;
        }
    } else {
        title = GCS(MSG_Unnamed);
    }

    return title;
}

long ProjectVersion(struct Project *pr)
{
    long version = 0;

    if (pr != NULL) {
        version = pr->pr_Version;
    }

    return version;
}

long ProjectStructureVersion(struct Project *pr)
{
    long version = 0;

    if (pr != NULL) {
        version = pr->pr_StructureVersion;
    }

    return version;
}

long ObtainProjectChangeLock(struct Project *pr)
{
    return SQLiteObtainChangeLock(ProjectSQLiteHandle(pr));
}

long ProjectExclusiveHandle(struct Project *pr)
{
    long handle = HANDLE_NONE;

    if (pr != NULL) {
	handle = pr->pr_ExclusiveHandle;
    }
    return handle;
}

void SetProjectExclusiveHandle(struct Project *pr, long handle)
{
    if (pr != NULL) {
	long old = pr->pr_ExclusiveHandle;
	bool oldSleep = (old != HANDLE_NONE && old != HANDLE_USER);
	bool sleep = (handle != HANDLE_NONE && handle != HANDLE_USER);

	pr->pr_ExclusiveHandle = handle;
	if (sleep != oldSleep) {
	    GuiSetProjectSleep(pr, sleep);
	}
    }
}
    
struct ProjectPrefs *ProjectProjectPrefs(struct Project *pr)
{
    struct ProjectPrefs *pp = NULL;

    if (pr) {
        pp = pr->pr_ProjectPrefs;
    }
    return pp;
}

void ClearProject(struct Project *pr, bool new)
{
    GuiObject se = ProjectStructureEditor(pr);
    struct Table *tab;
    struct MemPool *newDataMemPool = NULL;
    bool gui = ProjectHasGui(pr), ok, quick;

    GuiSetSleep(TRUE);
    if (new) {
	ServerRemoveProjectConnections(pr);
	pr->pr_ExclusiveHandle = HANDLE_NONE;
	pr->pr_AdminMode = TRUE;
        ClearSQLiteHandle(pr->pr_SQLiteHandle);
    }
    ClearProjectPrefs(pr->pr_ProjectPrefs);
    CloseProgramEditor(pr->pr_ProgramEditor);
    pr->pr_ProgramEditor = NULL;
    ClearQueryEditor(pr->pr_QueryEditor);
    FreeProgram(pr->pr_Program);
    pr->pr_Program = NULL;
    if (se == NULL) {
        FreeProjectGui(pr);
    } else {
        ClearStructureEditor(se);
    }
    SetProjectAdminPassword(pr, NULL);
    SetProjectName(pr, NULL);
    pr->pr_Version = 0;
    pr->pr_StructureVersion = 0;
    pr->pr_FileVersion = 0;
    pr->pr_CodesetWarningFields = 0;
    pr->pr_CodesetWarningChars = 0;
    pr->pr_CodesetWarned = FALSE;
    ProjectClearChanges(pr);
    ClearLog(pr->pr_Log, TRUE);
    ClearRootDisplayObject(pr->pr_RootDisplayObject);

    newDataMemPool = CreatePool(0L, 32 * 1024L, 4096);
    quick = newDataMemPool != NULL;
    while ((tab = (struct Table *)REMHEAD(&pr->pr_Tables)) != NULL) {
        FreeTable(tab, quick);
    }
    if (newDataMemPool != NULL) {
        DeletePool(pr->pr_DataMemPool);
        pr->pr_DataMemPool = newDataMemPool;
    }

    pr->pr_RecordMem = 0;

    GuiSetSleep(FALSE);

    if (gui && se == NULL) {
        ok = BuildProjectGui(pr);
 	if (ok == FALSE) {
	    /*
            ** what should we do here?
            */
        }
    } else {
	StructureEditorUpdate(se);
    }

    ProjectUpdateCurrentDir();
}

ulong ProjectNumRecords(struct Project *pr)
{
    ulong num = 0;
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    if (tabs) {
        FORNODE(tab, tabs) {
            num += TableNumRecords(tab);
        }
    }

    return num;
}

struct Project *FindProjectName(const char *pathname)
{
    struct Project *pr;

    ProjectSetCurrentDir(NULL);
    FORNODE(pr, &ProjectList) {
	if (SameFile(ProjectName(pr), pathname)) {
	    return pr;
	}
    }  
    return NULL;
}

void UpdateProjects(bool checkActive)
{
    struct Project *pr;

    FORNODE(pr, &ProjectList) {
        SQLiteReleaseLock(ProjectSQLiteHandle(pr));
        ReleaseLogFile(ProjectLog(pr));
    }

    pr = GuiActiveProject();
    if (checkActive && pr != NULL && ProjectChanges(pr) == 0) {
        if (pr->pr_Version != 0 && pr->pr_Version == pr->pr_FileVersion &&
                !QueryEditorIsOpen(pr) && pr->pr_ProgramEditor == NULL) {
            if (SQLiteCheckForProjectChange(pr->pr_SQLiteHandle) &&
                    SQLiteProjectHasChanged(pr->pr_SQLiteHandle)) {
                SQLiteGetProjectFileVersion(pr->pr_SQLiteHandle,
                        &pr->pr_FileVersion);
            }
            if (pr->pr_Version != pr->pr_FileVersion) {
                if (!GetPrefsValue(pr, PREFS_CONFIRMAUTORELOAD) ||
                        Ask(AT_QUESTION,
                            GCS(MSG_AskProjectReloadNewVersionAnswers),
                            GCS(MSG_AskProjectReloadNewVersion),
                            ProjectName(pr))) {
                    MenuProjectRevertToSaved(pr);
                }
            }
        }
        SetProjectPrependTitle(pr);
    }
}

void SetProjectPrependTitle(struct Project *pr)
{
    if (pr != NULL) {
        GuiObject wi = ProjectRootWindow(pr);
        const char *prepend = NULL;

        if (ProjectChanges(pr) != 0) {
            prepend = GCS(MSG_Changed);
        } else if (pr->pr_Version != pr->pr_FileVersion) {
            prepend = GCS(MSG_Old);
        } else if (!SQLiteAllowChanges(pr->pr_SQLiteHandle) &&
                SQLiteHasReservedLock(pr->pr_SQLiteHandle)) {
            prepend = GCS(MSG_Locked);
        }
        GuiSetPrependWindowTitle(wi, prepend);
    }
}

/*
** Load & Save
*/

static bool retryObtainWriteLock(struct Project *pr)
{
    return Ask(AT_ERROR, GCS(MSG_AskProjectRetryContinueAnswers),
            GCS(MSG_Err_ProjectWriteLock)) != 0;
}

static bool retryVacuum(struct Project *pr, long error, const char *errstr)
{
    return Ask(AT_ERROR, GCS(MSG_AskProjectRetryContinueAnswers),
            GCS(MSG_Err_ProjectVacuum), GES(error), errstr) != 0;
}

long SaveProject(struct Project *pr, const char *filename, bool reorg,
        char *errstr)
{
    long newProjectVersion = 0;
    long newStructureVersion = 0;
    bool forceInfo = FALSE;
    long error = OK;
    bool vacuum = reorg && GetPrefsValue(pr, PREFS_VACUUMAFTERREORG);

    if (pr == NULL) {
        return ERR_INTERNAL;
    }
    if (filename == NULL) {
	return ERR_INVALIDNAME;
    }

    if (errstr != NULL) {
	errstr[0] = 0;
    }

    /* Set new project version.  0 is reserved as a version for old projects */
    newProjectVersion = NextVersion(pr->pr_Version);
    newStructureVersion = pr->pr_StructureVersion;
    /* On reorganization all internal record numbers have been changed */
    if (pr->pr_StructureChanges != 0 || reorg || newStructureVersion == 0) {
        newStructureVersion = NextVersion(newStructureVersion);
    }

    ProjectSetCurrentDir(NULL);         /* make sure we have correct CWD */
    if (FileStat(filename) < 0) {
        forceInfo = TRUE;
    }

    error = SQLiteSaveProject(pr->pr_SQLiteHandle, filename, reorg,
            newProjectVersion, newStructureVersion, retryObtainWriteLock,
            vacuum, retryVacuum, errstr);
    if (error == OK) {
        SetProjectName(pr, filename);
        pr->pr_Version = newProjectVersion;
        pr->pr_StructureVersion = newStructureVersion;
        pr->pr_FileVersion = newProjectVersion;
        filename = ProjectName(pr);
        ProjectClearChanges(pr);

        AddFileInfo(pr, filename, forceInfo);
        SetPrefsString(pr, PREFS_PROJECTNAME, filename);
    }

    return error;
}

long OpenProject(struct Project *pr, const char *filename, bool interactive,
        char *errstr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    char *copy;
    bool sameFile = SameFile(ProjectName(pr), filename);
    bool admin = ProjectAdminMode(pr);
    bool gui = ProjectHasGui(pr);
    bool update = FALSE;
    bool clear = FALSE;
    GuiObject se;
    long error = OK;

    if (pr == NULL) {
        return ERR_INTERNAL;
    }
    if (filename == NULL) {
        return ERR_INVALIDNAME;
    }
    if (errstr != NULL) {
        strcpy(errstr, filename);
    }

    /* Make a copy of the filename since it might be the project name. */
    copy = AllocString(memPool, filename);
    if (copy == NULL) {
        return ERR_MEM;
    }

    GuiSetSleep(TRUE);
    error = SQLiteOpenProject(pr->pr_SQLiteHandle, copy, &update, &clear,
            &pr->pr_Version, &pr->pr_StructureVersion, errstr);
    if (error == OK) {
        SetProjectName(pr, copy);
        pr->pr_FileVersion = pr->pr_Version;
        ProjectClearChanges(pr);

    }
    FreeString(memPool, copy);

    if (error == OK && !sameFile) {
        admin = !(pr->pr_AdminPassword && pr->pr_AdminPassword[0]);
        pr->pr_AdminMode = admin;
    }

    if (error == OK && !update && ProjectProgram(pr) != NULL) {
        struct Program *compiled = NULL;
        const char *errstr = NULL;
        long line = 0, col = 0;
        const char *filename = NULL;

        /* compile program */
	errstr = ProgramCompileProjectProgram(pr, TRUE,
		&compiled, &line, &col, &filename);
        if (errstr == NULL) {
            SetProjectProgram(pr, compiled);
	} else if (interactive) {
	    if (Ask(AT_ERROR, GCS(MSG_Err_CompileProgramAnswers), 
			GCS(MSG_Err_CompileProgram))) {
		SetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL, FALSE);
		ProgramEditorShowError(pr, FALSE, errstr, line, col, filename);
	    }
	    error = OK;
	} else {
	    error = PROGERR_COMPILE;
	}
    }

    if (error == OK && !update && 
            GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL)) {
	const char *path = GetPrefsString(pr, 
		PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);
        char *mem = NULL;
	bool create;

        ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
        mem = ExpandPathname(path, memPool);
        if (mem != NULL) {
            path = mem;
        }
	if (FileStat(path) < 0) {
	    create = TRUE;
        } else if (ProgramSourceExternalIdentical(pr)) {
	    create = FALSE;				/* not needed */
        } else if (!interactive) {
	    create = FALSE;
    	    error = PROGERR_COMPILE;
	} else {
	    create = (bool)Ask(AT_QUESTION, GCS(MSG_AskReplaceAnswers),
			GCS(MSG_AskReplaceFile), path);
	}
	if (create) {
	    const char *errstr = ProgramCreateSourceExternal(pr);

	    if (errstr != NULL) {
		if (interactive) {
		    Ask(AT_ERROR, GCS(MSG_StarOk), errstr, path);
                } else {
		    error = PROGERR_COMPILE;
                }
		SetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL, FALSE);
	    }
	}
        FreeString(memPool, mem);
    }

    /* Try to re-compile all filters and queries but don't complain */
    if (error == OK && !update) {
        struct List *tabs = ProjectTables(pr);
        struct Table *tab;
        struct List *queries = QueryEditorQueries(ProjectQueryEditor(pr));
        struct Query *query;
        const struct Program *prg; 
        struct Program *compiled;

        if (tabs != NULL) {
            FORNODE(tab, tabs) {
                if (ProgramCompileFilter(tab, &compiled, NULL, NULL) == NULL) {
                    SetTableFilterProgram(tab, compiled);
                }
            }
        }
        if (queries != NULL) {
            FORNODE(query, queries) {
                prg = QueryProgram(query);
                if (ProgramCompileQuery(prg, &compiled, NULL, NULL) == NULL) {
                    SetQueryProgram(query, compiled);
                }
            }
        }
    }
    GuiSetSleep(FALSE);

    if (error != OK) {
        if (clear) {
            ClearProject(pr, TRUE);
            OpenProjectWindows(pr);		/* ignore failure */
        }
	return error;
    }

    se = ProjectStructureEditor(pr);
    if (se != NULL) {
        StructureEditorUpdate(se);
        StructureEditorCheckNames(pr);
    }
    SetTablesGuiRecords(pr);
    if (!update) {
        FreeProjectGui(pr);
        if (SetProjectHasGui(pr, gui) != gui) {
            ClearProject(pr, TRUE);
            OpenProjectWindows(pr);		/* ignore failure */
            return DPYERR_BUILDGUI;
        }

        SetProjectAdminMode(pr, admin);     /* calls trigger */

        if (FindUserFunctionName(ProjectProgram(pr), onOpen)) {
            ExecuteTrigger(pr, onOpen, NULL, NULL);
        }

        if (se == NULL && gui && !OpenProjectWindows(pr)) {
            ClearProject(pr, TRUE);
            OpenProjectWindows(pr);		/* ignore failure */
            return DPYERR_OPENGUI;
        }
    } else {
        struct Program *prg = ProjectProgram(pr);
        struct UserFunction *uf = FindUserFunctionName(prg, onReload);
        bool context;

        /*
        ** In order to speed up program execution of onReload and the
        ** following GuiUpdate call, we create our own program environment
        ** such that virtual fields are updated at the end.
        */
        GuiRemAllRecords(pr);
        context = BeginGuiProgramContext(pr);
        if (uf != NULL) {
            ProgramCall(prg, uf, NULL, TRUE);
        }
        GuiUpdate(pr);
        EndGuiProgramContext(pr, context);
    }

    if (interactive && pr->pr_CodesetWarningChars > 0) {
        ProjectWarnCodeset(pr);
    }
    SetPrefsString(pr, PREFS_PROJECTNAME, filename);

    return error;
}

long RevertProject(struct Project *pr, bool interactive, char *errstr)
{
    return OpenProject(pr, ProjectName(pr), interactive, errstr);
}

/*
** Support for reading custom format.
*/

#define STRUCTURE_FILENAME "Structure.mb"

#define STARTTAG MAKE_ID('M', 'B', 'S', 'T')
#define PREFTAG MAKE_ID('P', 'R', 'E', 'F')
#define TABLETAG MAKE_ID('R', 'E', 'L', 'S')
#define FIELDTAG MAKE_ID('A', 'T', 'R', 'S')
#define DIOTAG MAKE_ID('D', 'I', 'O', 'S')
#define ORDTAG MAKE_ID('O', 'R', 'D', 'S')
#define PRGTAG MAKE_ID('P', 'R', 'O', 'G')
#define FILTAG MAKE_ID('F', 'I', 'L', 'S')
#define SQLTAG MAKE_ID('S', 'Q', 'L', 'S')
#define RECTAG MAKE_ID('R', 'E', 'C', 'S')
#define ENDTAG MAKE_ID('M', 'B', 'O', 'K')

#define PROJECT_VERSION 8

bool IsProjectCustomFormatDir(const char *dirname)
{
    Cwd cwd = NULL;
    FILE *file = NULL;
    bool rc = FALSE;

    if (PushDir(dirname, &cwd) == 0) {
        if (FileStat(STRUCTURE_FILENAME) > 0) {
            file = FOpen(STRUCTURE_FILENAME, "rb");
            if (file != NULL) {
                rc = (FGetLong(file) == STARTTAG);
                fclose(file);
            }
        }
        PopDir(&cwd);
    }

    return rc;
}

static void opencleanup(Cwd *cwd, struct Project *pr, bool clear)
{
    PopDir(cwd);
    if (clear) {
        ClearProject(pr, TRUE);
        OpenProjectWindows(pr);		/* ignore failure */
    }
}

long OpenProjectCustomFormat(struct Project *pr, const char *dirname,
        struct FileLockContext *flc, bool interactive, char *errstr)
{
    FILE *file = NULL;
    long error = OK;
    bool sameFile = SameFile(ProjectName(pr), dirname);
    bool end = FALSE;
    Cwd cwd = NULL;
    bool clear = FALSE;
    const char *tabName = NULL;
    struct Program *prg = NULL;
    short version = 0;
    long projectVersion = 0;
    long structureVersion = 0;
    static char verstr[64];
    char pathSepStr[2];

    if (pr == NULL) {
        return ERR_INTERNAL;
    }
    if (dirname == NULL) {
	return ERR_INVALIDNAME;
    }
    if (flc == NULL) {
        return ERR_INTERNAL;
    }
    if (errstr != NULL) {
	strcpy(errstr, dirname);
    }

    if (FileStat(dirname) != 0) {
        return DISKERR_NOTABEEBASEPROJECT;
    }

    error = RaiseFileLockLevel(flc, SHARED_FILE_LOCK);
    if (error != OK) {
        opencleanup(&cwd, pr, clear);
        return PROJECTERR_READLOCK;
    }

    if (PushDir(dirname, &cwd) != 0) {
        opencleanup(&cwd, pr, clear);
        return DISKERR_CHDIR;
    }

    if (FileStat(STRUCTURE_FILENAME) <= 0) {
        opencleanup(&cwd, pr, clear);
        return DISKERR_NOTABEEBASEPROJECT;
    }

    ProjectClearCodesetWarnings(pr);
    file = FOpen(STRUCTURE_FILENAME, "rb");
    PopDir(&cwd);
    if (file == NULL) {
	if (errstr != NULL) {
            pathSepStr[0] = PATH_SEP;
            pathSepStr[1] = 0;
	    strcat(errstr, pathSepStr);
	    strcat(errstr, STRUCTURE_FILENAME);
	}
        opencleanup(&cwd, pr, clear);
	return DISKERR_OPEN;
    }
    setvbuf(file, NULL, _IOFBF, 8192);

    strcpy(verstr, "0.24");
    if (FGetLong(file) != STARTTAG) {
        error = DISKERR_TAG;
    } else {
        FGetS(file, verstr, sizeof(verstr));
        version = FGetShort(file);
        if (version < 1 || version > PROJECT_VERSION) {
            error = DISKERR_VERSION;
        }
    }
    if (error == OK && version > 5) {
        projectVersion = FGetLong(file);
        if (projectVersion == 0) {
            error = DISKERR_NOTABEEBASEPROJECT;
        } else if (version > 6) {
            structureVersion = FGetLong(file);
            if (structureVersion == 0) {
                error = DISKERR_NOTABEEBASEPROJECT;
            }
        }
    }
    if (error == OK && projectVersion == 0) {
        error = ObtainCompatibilityFileLockLevel(flc);
    }
    if (error == OK) {
        ClearProject(pr, !sameFile);
    }
    if (error == OK) {
        pr->pr_Version = projectVersion;
        pr->pr_StructureVersion = structureVersion;
        pr->pr_FileVersion = projectVersion;
        if (version >= 2) {
            FGetBool(file);  /* utf-8 */
        }
        if (version >= 3) {
            pr->pr_AdminPassword = ReadString(file, pr->pr_MemPool);
        }

        SetProjectName(pr, dirname);

        /* If anything fails from now on, we have to clear the project. */
        clear = TRUE;
    }

    while (error == OK && end == FALSE) {
        if (feof(file) || ferror(file)) {
	    error = DISKERR_READ;
        } else {
            switch (FGetLong(file)) {
                case PREFTAG:   
                    error = LoadProjectPrefsCustomFormat(
                            ProjectProjectPrefs(pr), file);
                    break;
                case TABLETAG:  error = LoadTablesCustomFormat(pr, file);
                                break;
                case FIELDTAG:	error = LoadFieldsCustomFormat(pr, file);
                                break;
                case DIOTAG:    error = LoadDisplayObjectsCustomFormat(pr,file);
                                break;
                case ORDTAG:    error = LoadOrdersCustomFormat(pr, file);
                                break;
                case FILTAG:    error = LoadFiltersCustomFormat(pr, file);
                                break;
                case SQLTAG:    error =
                                LoadQueriesCustomFormat(pr, file, version);
                                break;
                case RECTAG:    
                    error = LoadRecordsCustomFormat(pr, file, &tabName);
		    if (error != OK && tabName != NULL && errstr != NULL) {
                        pathSepStr[0] = PATH_SEP;
                        pathSepStr[1] = 0;
                        strcat(errstr, pathSepStr);
			strcat(errstr, tabName);
		    }
		    break;
                case ENDTAG:
                    end = TRUE;
                    break;
                case PRGTAG:    
		    error = LoadProgramCustomFormat(pr, file, &prg);
                    SetProjectProgram(pr, prg);
		    break;
                default:
                    error = DISKERR_TAG;
                    break;
            }
	}
    }
    if (ferror(file)) {
        error = DISKERR_READ;
    }
    fclose(file);

    if (error == OK) {
        SetTablesGuiRecords(pr);
    }

    return error;
}

long DeleteProjectCustomFormat(const char *dirname)
{
    Cwd cwd = NULL;
    Dir *dfd;                 /* directory descriptor */
    struct MemPool *memPool = NULL;
    const char *filename;
    struct FileLockContext *flc = NULL;
    long error = OK;
    short type;

    if (dirname == NULL) {
	return OK;
    }

    ProjectSetCurrentDir(NULL);	    /* make sure we have correct current dir */
    type = FileStat(dirname);
    if (type < 0) {		    /* directory does not exist? */
        return OK;
    } else if (type > 0) {
        return DISKERR_NOTABEEBASEPROJECT;
    } else if (RmDir(dirname) == 0) { /* simple check if directory is empty */
        return OK;
    }

    /* Check that we are able to obtain an exlusive lock on the project */
    memPool = CreatePool(0L, 4096, 1024);
    flc = AllocFileLockContext(memPool, dirname);
    if (flc == NULL) {
        error = ERR_MEM;
    } else {
        error = RaiseFileLockLevel(flc, EXCLUSIVE_FILE_LOCK);
        ReleaseFileLockLevel(flc); 
        FreeFileLockContext(flc);
    }
    if (memPool != NULL) {
        DeletePool(memPool);
    }
    if (error != OK) {
        return error;
    }

    if (PushDir(dirname, &cwd) != 0) {
       	error = DISKERR_NOTABEEBASEPROJECT;
    } else if (FileStat(STRUCTURE_FILENAME) <= 0) {
       	error = DISKERR_NOTABEEBASEPROJECT;
    } else if ((dfd = DirOpen(CURRENTDIR)) == NULL) {
	error = DISKERR_READDIR;
    } else {
	while((filename = DirReadName(dfd)) != NULL) {
	    RmFile(filename);
	}
	DirClose(dfd);
    }
    PopDir(&cwd);

    if (error == OK && RmDir(dirname) != 0) {
	error = DISKERR_RMDIR;
    }

    return error;
}

