/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** project.h
**
** Creation date: 15.11.1994
** Last modified: 18.08.2020
*/

#ifndef PROJECT_H
#define PROJECT_H

#include "system.h"

struct Project;
struct Log;
struct Table;
struct FileLockContext;

#ifdef NEED_NODE_CAST
inline Project *nodecast(Project *, Node *node) { return (Project *)node; }
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
** Set current working directory based on project.
** If pr == NULL, change to initial working dir.
*/
void ProjectSetCurrentDir(struct Project *pr);

/*
** Change directory according to project.
** This function should be called after changing prefs or project path.
*/
void ProjectUpdateCurrentDir(void);

/* 
** Main project methods 
*/
void FreeProject(struct Project *pr);
struct Project *AllocProject(bool gui);
bool ProjectIsEmpty(struct Project *pr);
bool ProjectHasGui(struct Project *pr);
bool SetProjectHasGui(struct Project *pr, bool gui);
struct SQLiteHandle *ProjectSQLiteHandle(struct Project *pr);
struct Log *ProjectLog(struct Project *pr);
struct List *ProjectTables(struct Project *pr);
void AddTable(struct Project *pr, struct Table *tab);
void RemTable(struct Project *pr, struct Table *tab);
struct MemPool *ProjectMemPool(struct Project *pr);
struct MemPool *ProjectDataMemPool(struct Project *pr);
struct DisplayObject *ProjectRootDisplayObject(struct Project *pr);
GuiObject ProjectRootWindow(struct Project *pr);
long AddProjectRecordMem(struct Project *pr, long size);
bool OpenProjectWindows(struct Project *pr);
bool ProjectAdminMode(struct Project *pr);
void SetProjectAdminMode(struct Project *pr, bool admin);
const char *ProjectAdminPassword(struct Project *pr);
bool SetProjectAdminPassword(struct Project *pr, const char *passwd); 
GuiObject ProjectStructureEditor(struct Project *pr);
void SetProjectStructureEditor(struct Project *pr, GuiObject se);
GuiObject ProjectProgramEditor(struct Project *pr);
void SetProjectProgramEditor(struct Project *pr, GuiObject pe);
struct QueryEditor *ProjectQueryEditor(struct Project *pr);
struct Program *ProjectProgram(struct Project *pr);
void SetProjectProgram(struct Project *pr, struct Program *prg);
long ProjectChanges(struct Project *pr);
long ProjectStructureChanges(struct Project *pr);
void ProjectChange(struct Project *pr, bool structure);
void ProjectClearChanges(struct Project *pr);
void ProjectAddCodesetWarnings(struct Project *pr, size_t warnings);
size_t ProjectCodesetWarningFields(struct Project *pr);
size_t ProjectCodesetWarningChars(struct Project *pr);
void ProjectWarnCodeset(struct Project *pr);
const char *ProjectName(const struct Project *pr);
bool SetProjectName(struct Project *pr, const char *name);
const char *ProjectTitle(struct Project *pr);
long ProjectVersion(struct Project *pr);
long ProjectStructureVersion(struct Project *pr);
long ObtainProjectChangeLock(struct Project *pr);
long ProjectExclusiveHandle(struct Project *pr);
void SetProjectExclusiveHandle(struct Project *pr, long handle);
struct ProjectPrefs *ProjectProjectPrefs(struct Project *pr);
void ClearProject(struct Project *pr, bool newProject);
ulong ProjectNumRecords(struct Project *pr);

struct Project *FindProjectName(const char *pathname);

/*
** Release SQLite3 locks and optionally check if the current project has
** been updated by a different SQLite3 connection.
*/
void UpdateProjects(bool checkActive);
 
/*
** Prepends the window title if the project has changes,
** or if some other instance acquired a reserved lock.
*/
void SetProjectPrependTitle(struct Project *pr);

/*
** Saves project.  The filename can be the project name or a new filename.
** If reorg is TRUE, the project is also reorganized.
** Returns OK on success and an error code otherwise.  
** It also copies an error string to errstr if not NULL.
** Use sprintf(buffer, GES(error), errstr) for a complete error description.  
*/
long SaveProject(struct Project *pr, const char *filename, bool reorg,
        char *errstr);
 
/*
** Opens and loads a project from filename. 
** If interactive is set, the function might interact with the user
** for opening the program editor on compile error or on other
** errors that might be repairable.  Otherwise an error code is
** returned back in such sitiations where interaction would be required.
** ProjectOpen returns OK on success and an error code otherwise.  
** It also copies an error string to errstr if not NULL.
** Use sprintf(buffer, GES(error), errstr) for a complete error description.  
*/
long OpenProject(struct Project *pr, const char *filename, bool interactive,
        char *errstr);

/*
** Reverts a project to the last saved version.  This undos all changes
** and loads a fresh version of the project.  See OpenProject for the meaning
** of other arguments.
*/
long RevertProject(struct Project *pr, bool interactive, char *errstr);

/*
** Checks whether a given directory contains a project in custom format.
*/
bool IsProjectCustomFormatDir(const char *dirname);

/*
** Opens and loads a project in custom format from dirname.
** The file-lock context must match to the dirname.
** See OpenProject for the meaning of arguments. 
*/
long OpenProjectCustomFormat(struct Project *pr, const char *dirname,
        struct FileLockContext *flc, bool interactive, char *errstr);

/*
** Deletes the project in custom format with the given dirname.
** Returns OK on success or error code.
** Use sprintf(buffer, GES(error), dirname) for generating a complete
** error description.
*/
long DeleteProjectCustomFormat(const char *dirname);

#ifdef __cplusplus
}
#endif

#endif

