/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** queryeditor.c
**
** Creation date: 03.06.1995
** Last modified: 11.05.2022
*/

#include "system.h"
#include "project.h"
#include "table.h"
#include "field.h"
#include "queryeditor.h"
#include "lang/program.h"
#include "lang/compile.h"
#include "lang/functable.h"
#include "strhandle.h"
#include "memory.h"
#include "memo.h"
#include "labels.h"
#include "util.h"
#include "error.h"
#include "break.h"
#include "appstrings_cat.h"
#include "requesters.h"

long QueryLineLines(const struct QueryLine *ql, long numEntries)
{
    long lines = 0;
    long i, l;

    if (ql != NULL) {
        for (i = 0; i < numEntries; i++) {
            l = StrCount(ql->ql_Entries[i]->qe_String, '\n') + 1;
            if (l > lines) {
                lines = l;
            }
        }
    }

    return lines;
}

static long longCmp(long i1, long i2, bool nil1, bool nil2)
{
    long result;

    if (nil1) {
        result = nil2? 0 : -1;
    } else if (nil2) {
        result = 1;
    } else if (i1 > i2) {
        result = 1;
    } else if (i1 < i2) {
        result = -1;
    } else {
        result = 0;
    }

    return result;
}

static long doubleCmp(double d1, double d2, bool nil1, bool nil2)
{
    long result;

    if (nil1) {
        result = nil2? 0 : -1;
    } else if (nil2) {
        result = 1;
    } else if (d1 > d2) {
        result = 1;
    } else if (d1 < d2) {
        result = -1;
    } else {
        result = 0;
    }

    return result;
}

long QueryEntryCmp(struct QueryEntry *qe1, struct QueryEntry *qe2, char type,
        bool star)
{
    long result;

    if (qe1 == NULL) {
        result = (qe2 == NULL)? 0 : -1;
    } else if (qe2 == NULL) {
        result = 1;
    } else {
        switch (type)
        {
            case TYPE_INTEGER:
            case TYPE_RECORD:
            case TYPE_FILE:
            {
                long i1, i2;
                bool nil1 = !StrToInt(qe1->qe_String, &i1);
                bool nil2 = !StrToInt(qe2->qe_String, &i2);

                result = longCmp(i1, i2, nil1, nil2);
                break;
            }

            case TYPE_REAL:
            {
                double d1, d2;
                bool nil1 = !StrToDouble(qe1->qe_String, &d1);
                bool nil2 = !StrToDouble(qe2->qe_String, &d2);

                result = doubleCmp(d1, d2, nil1, nil2);
                break;
            }

            case TYPE_DATE:
            {
                long d1, d2;
                bool nil1 = !StringToDate(qe1->qe_String, &d1);
                bool nil2 = !StringToDate(qe2->qe_String, &d2);

                result = longCmp(d1, d2, nil1, nil2);
                break;
            }

            case TYPE_TIME:
            {
                long t1, t2;
                bool nil1 = !StringToTime(qe1->qe_String,FORMAT_HOURMINSEC,&t1);
                bool nil2 = !StringToTime(qe2->qe_String,FORMAT_HOURMINSEC,&t2);

                result = longCmp(t1, t2, nil1, nil2);
                break;
            }

            default:
                if (star) {
                    result = CodesetStrCollate(qe1->qe_String, qe2->qe_String, 
                            CS_COLLATE_SIMPLE);
                } else {
                    result = strcmp(qe1->qe_String, qe2->qe_String);
                }
                break;
        }
    }
    return result;
}

static struct QueryEntry *allocQueryEntry(struct MemPool *memPool,
        struct Record *rec, const char *str)
{
    struct QueryEntry *qe;

    qe = PoolAllocMem(memPool, sizeof(struct QueryEntry) + strlen(str) + 1);
    if (qe == NULL) {
        return NULL;
    }

    qe->qe_Record = rec;
    strcpy(qe->qe_String, str);
    return qe;
}

static struct QueryEntry *elementToQueryEntry(const struct Element *el,
        struct Field *field, struct MemPool *memPool)
{
    bool allocated = FALSE;
    char *str = SimpleElementToString(el, field, memPool, &allocated);
    struct Record *rec = NULL;
    struct QueryEntry *qe;

    if (str == NULL) {
        return NULL;
    }

    if (el != NULL) {
        rec = el->el_Record;
    }
    qe = allocQueryEntry(memPool, rec, str);

    if (allocated) {
        FreeString(memPool, str);
    }

    return qe;
}

static struct QueryLine *allocQueryLine(struct MemPool *memPool, size_t num)
{
    return PoolAllocMemZero(memPool,
            sizeof(struct QueryLine) + (num + 1)*sizeof(struct QueryEntry*));
}

struct QueryLine *queryGetLine(const struct Element *list, size_t lineNum,
        size_t numFields, struct QueryFieldInfo *infos, struct MemPool *memPool)
{
    struct QueryLine *ql;
    struct QueryEntry *qe;
    size_t i;

    if (list == NULL) {
        return NULL;
    }
    if (numFields > 0 && infos == NULL) {
        return NULL;
    }

    ql = allocQueryLine(memPool, numFields);
    if (ql == NULL) {
        return NULL;
    }

    ql->ql_LineNum = lineNum;
    for (i = 0; i < numFields; list = list->el_Contents.list.tail) {
        const struct Element *elem = list;
        short type;
        size_t len;

        if (elem->el_Type == TYPE_LIST) {
            elem = elem->el_Contents.list.head;
        } else if (i > 0 && elem->el_Type == TYPE_NIL) {
            break;
        }

        type = elem->el_Type;
        if (lineNum == 0) {  /* Title line */
            if (elem->el_Type == TYPE_STRING) {
                infos[i].qfi_Field = elem->el_Contents.string.field;
            }
            type = FieldType(infos[i].qfi_Field);
            if (type == NOTYPE || type == TYPE_VIRTUAL) {
                type = TYPE_UNKNOWN;
            }
        } else if (type == TYPE_INTEGER &&
                FieldType(infos[i].qfi_Field) == TYPE_CHOICE) {
            type = TYPE_CHOICE;
        }
        infos[i].qfi_Type = type = GenType(infos[i].qfi_Type, type);

        switch (type)
        {
            case TYPE_INTEGER:
            case TYPE_REAL:
                infos[i].qfi_Format = 1;
                break;

            case TYPE_NIL:
                break;

            default:
                infos[i].qfi_Format = 0;
                break;
        }

        qe = elementToQueryEntry(elem, infos[i].qfi_Field, memPool);
        if (qe == NULL) {
            return NULL;
        }
        ql->ql_Entries[i] = qe;
        len = strlen(qe->qe_String);
        if (len > infos[i].qfi_Width) {
            infos[i].qfi_Width = len;
        }
        i++;

        if (list->el_Type != TYPE_LIST) {
            break;
        }
    }
    for (; i < numFields; i++)
    {
        qe = elementToQueryEntry(NULL, infos[i].qfi_Field, memPool);
        if (qe == NULL) {
            return NULL;
        }
        ql->ql_Entries[i] = qe;
    }

    return ql;
}

static struct QueryFieldInfo *allocQueryFieldInfos(
        struct MemPool *memPool, size_t numFields)
{
    struct QueryFieldInfo *infos = 
        PoolAllocMemZero(memPool, numFields * sizeof(struct QueryFieldInfo));
    size_t i;

    if (infos != NULL) {
        for (i = 0; i < numFields; i++) {
            infos[i].qfi_Type = TYPE_UNKNOWN;
        }
    }
    return infos;
}

void FreeQueryGuiData(struct QueryGuiData *qgd)
{
    if (qgd != NULL && qgd->qgd_MemPool != NULL) {
        DeletePool(qgd->qgd_MemPool);
    }
}

static struct QueryGuiData *allocQueryGuiData(void)
{
    struct MemPool *memPool;
    struct QueryGuiData *qgd;

    memPool = CreatePool(0, 4096, 1024);
    if (memPool == NULL) {
        return NULL;
    }

    qgd = PoolAllocMemZero(memPool, sizeof(struct QueryGuiData));
    if (qgd == NULL) {
        DeletePool(memPool);
        return NULL;
    }

    qgd->qgd_MemPool = memPool;
    return qgd;
}

struct QueryGuiData *QueryGuiDataFromTitlesBody(
        const struct Element *titles, const struct Element *body)
{
    struct QueryGuiData *qgd = allocQueryGuiData();
    struct MemPool *memPool;

    if (qgd == NULL) {
        return NULL;
    }

    memPool = qgd->qgd_MemPool;

    /* Examine titles and body */
    if (titles == NULL) {
        if (body == NULL) {
            return qgd;             /* nothing to do */
        } else if (body->el_Type != TYPE_LIST) {
            struct QueryEntry *qe = elementToQueryEntry(body, NULL, memPool);

            if (qe == NULL) {
                FreeQueryGuiData(qgd);
                return NULL;
            }
            qgd->qgd_NumFields = 1;
            qgd->qgd_FieldInfos = allocQueryFieldInfos(memPool, 1);
            if (qgd->qgd_FieldInfos == NULL) {
                FreeQueryGuiData(qgd);
                return NULL;
            }
            qgd->qgd_FieldInfos[0].qfi_Width = strlen(qe->qe_String);
            qgd->qgd_Titles = allocQueryLine(memPool, 1);
            if (qgd->qgd_Titles == NULL) {
                FreeQueryGuiData(qgd);
                return NULL;
            }
            qgd->qgd_Titles->ql_Entries[0] = qe;
            return qgd;
        }
    }

    /* Get number of fields and lines */
    if (titles != NULL) {
        const struct Element *sublist = titles;
        size_t i = 0;
        do {
            i++;
            if (sublist->el_Type != TYPE_LIST) {
                break;
            }
            sublist = sublist->el_Contents.list.tail;
        } while (sublist->el_Type != TYPE_NIL);
        if (i > qgd->qgd_NumFields) {
            qgd->qgd_NumFields = i;
        }
    }

    if (body != NULL) {
        const struct Element *list = body;

        while (list->el_Type != TYPE_NIL) {
            const struct Element *sublist;
            size_t i = 0;

            qgd->qgd_NumLines++;
            if (list->el_Type != TYPE_LIST) {
                if (qgd->qgd_NumFields < 1) {
                    qgd->qgd_NumFields = 1;
                }
                break;
            }
            sublist = list->el_Contents.list.head;
            do {
                i++;
                if (sublist->el_Type != TYPE_LIST) {
                    break;
                }
                sublist = sublist->el_Contents.list.tail;
            } while (sublist->el_Type != TYPE_NIL);
            if (i > qgd->qgd_NumFields) {
                qgd->qgd_NumFields = i;
            }
            list = list->el_Contents.list.tail;
        }
    }

    if (qgd->qgd_NumFields > 0) {
        qgd->qgd_FieldInfos = allocQueryFieldInfos(memPool, qgd->qgd_NumFields);
        if (qgd->qgd_FieldInfos == NULL) {
            FreeQueryGuiData(qgd);
            return NULL;
        }
    }

    // Get title strings
    if (titles) {
        size_t i;

        qgd->qgd_Titles = queryGetLine(titles, 0, qgd->qgd_NumFields,
                qgd->qgd_FieldInfos, memPool);
        if (qgd->qgd_Titles == NULL) {
            FreeQueryGuiData(qgd);
            return NULL;
        }
        qgd->qgd_ControlIndex = qgd->qgd_NumFields;
        for (i = 0; i < qgd->qgd_NumFields; i++) {
            qgd->qgd_ControlTable= FieldTable(qgd->qgd_FieldInfos[i].qfi_Field);
            if (qgd->qgd_ControlTable != NULL) {
                qgd->qgd_ControlIndex = i;
                break;
            }
        }
    }

    // Set main body
    if (qgd->qgd_NumLines > 0) {
        qgd->qgd_Lines = PoolAllocMemZero(memPool,
                qgd->qgd_NumLines * sizeof(struct QueryLine *));
        if (qgd->qgd_Lines == NULL) {
            FreeQueryGuiData(qgd);
            return NULL;
        }
    }

    if (body != NULL) {
        const struct Element *list = body;
        size_t j = 0;

        for (j = 0; j < qgd->qgd_NumLines; j++) {
            const struct Element *sublist = list;

            if (list->el_Type == TYPE_LIST) {
                sublist = list->el_Contents.list.head;
                list = list->el_Contents.list.tail;
            }
            qgd->qgd_Lines[j] = queryGetLine(sublist, j + 1,
                    qgd->qgd_NumFields, qgd->qgd_FieldInfos, memPool);
            if (qgd->qgd_Lines[j] == NULL) {
                FreeQueryGuiData(qgd);
                return NULL;
            }
        }
    }

    return qgd;
}

struct QueryGuiData *QueryGuiDataFromString(const char *str)
{
    struct QueryGuiData *qgd = allocQueryGuiData();
    struct MemPool *memPool;

    if (qgd == NULL) {
        return NULL;
    }

    memPool = qgd->qgd_MemPool;
    if (str != NULL) {
        size_t len = strlen(str);

        qgd->qgd_NumFields = 1;
        qgd->qgd_FieldInfos = allocQueryFieldInfos(memPool, 1);
        if (qgd->qgd_FieldInfos == NULL) {
            FreeQueryGuiData(qgd);
            return NULL;
        }
        qgd->qgd_FieldInfos[0].qfi_Width = len;
        qgd->qgd_Titles = allocQueryLine(memPool, 1);
        if (qgd->qgd_Titles == NULL) {
            FreeQueryGuiData(qgd);
            return NULL;
        }
        qgd->qgd_Titles->ql_Entries[0] = PoolAllocMemZero(memPool,
                sizeof(struct QueryEntry) + len + 1);
        if (qgd->qgd_Titles->ql_Entries[0] == NULL) {
            FreeQueryGuiData(qgd);
            return NULL;
        }
        strcpy(qgd->qgd_Titles->ql_Entries[0]->qe_String, str);
    }

    return qgd;
}

static bool queryLinesMatch(const struct QueryLine *ql1,
        const struct QueryLine *ql2, size_t numFields)
{
    size_t i;

    if (ql1 == ql2) {
        return TRUE;
    }
    if (ql1 == NULL || ql2 == NULL) {
        return FALSE;
    }
    if (ql1->ql_Checked != ql2->ql_Checked) {
        return FALSE;
    }
    for (i = 0; i < numFields; i++) {
        const struct QueryEntry *qe1 = ql1->ql_Entries[i];
        const struct QueryEntry *qe2 = ql2->ql_Entries[i];

        if (qe1->qe_Record != qe2->qe_Record) {
            return FALSE;
        }
        if (strcmp(qe1->qe_String, qe2->qe_String) != 0) {
            return FALSE;
        }
    }

    return TRUE;
}

bool QueryGuiDataCopyLineNumIfMatch(struct QueryGuiData *qgd1,
        const struct QueryGuiData *qgd2)
{
    size_t i;

    if (qgd1 == qgd2) {
        return TRUE;
    }
    if (qgd1 == NULL || qgd2 == NULL) {
        return FALSE;
    }
    if (qgd1->qgd_NumFields != qgd2->qgd_NumFields) {
        return FALSE;
    }
    if (qgd1->qgd_NumLines != qgd2->qgd_NumLines) {
        return FALSE;
    }
    if (qgd1->qgd_ControlTable != qgd2->qgd_ControlTable) {
        return FALSE;
    }
    if (qgd1->qgd_ControlIndex != qgd2->qgd_ControlIndex) {
        return FALSE;
    }
    for (i = 0; i < qgd1->qgd_NumFields; i++) {
        const struct QueryFieldInfo *info1 = &qgd1->qgd_FieldInfos[i];
        const struct QueryFieldInfo *info2 = &qgd2->qgd_FieldInfos[i];

        if (info1->qfi_Field != info2->qfi_Field) {
            return FALSE;
        }
        if (info1->qfi_Format != info2->qfi_Format) {
            return FALSE;
        }
        if (info1->qfi_Type != info2->qfi_Type) {
            return FALSE;
        }
        if (info1->qfi_Width != info2->qfi_Width) {
            return FALSE;
        }
    }
    if (!queryLinesMatch(qgd1->qgd_Titles, qgd2->qgd_Titles,
                qgd1->qgd_NumFields)) {
        return FALSE;
    }
    for (i = 0; i < qgd1->qgd_NumLines; i++) {
        if (!queryLinesMatch(qgd1->qgd_Lines[i], qgd2->qgd_Lines[i],
                    qgd1->qgd_NumFields)) {
            return FALSE;
        }
    }

    /* QueryGuiData matches */
    for (i = 0; i < qgd1->qgd_NumLines; i++) {
        qgd1->qgd_Lines[i]->ql_LineNum =  qgd2->qgd_Lines[i]->ql_LineNum;
    }

    return TRUE;
}

bool QueryGuiDataAllSingleLine(const struct QueryGuiData *qgd)
{
    size_t i;
    struct QueryLine **begin, **end;

    if (qgd == NULL) {
        return TRUE;
    }
    begin = qgd->qgd_Lines;
    end = begin + qgd->qgd_NumLines;

    for (i = 0; i < qgd->qgd_NumFields; i++) {
        struct Field *field = qgd->qgd_FieldInfos[i].qfi_Field;
        short type = FieldType(field);
        struct QueryLine **ql;

        if (field == NULL || type == TYPE_MEMO || type == TYPE_VIRTUAL) {
            for (ql = begin; ql != end; ql++) {
                if (strchr((*ql)->ql_Entries[i]->qe_String, '\n') != NULL) {
                    return FALSE;
                }
            }
        }
    }
    return TRUE;
}

void QueryGuiDataMoveLine(struct QueryGuiData *qgd, size_t row, size_t newRow)
{
    if (qgd != NULL && row < qgd->qgd_NumLines) {
        struct QueryLine *ql = qgd->qgd_Lines[row];

        if (newRow > qgd->qgd_NumLines) {
            newRow = qgd->qgd_NumLines;
        }
        if (newRow > row) {
            memmove(&qgd->qgd_Lines[row], &qgd->qgd_Lines[row + 1],
                    (newRow - row - 1) * sizeof(struct QueryLine *));
            qgd->qgd_Lines[newRow - 1] = ql;
        } else if (newRow < row) {
            memmove(&qgd->qgd_Lines[newRow + 1], &qgd->qgd_Lines[newRow],
                    (row - newRow) * sizeof(struct QueryLine *));
            qgd->qgd_Lines[newRow] = ql;
        }
    }
}

struct QueryEditor
{
    struct Project *qe_Project;
    GuiObject qe_QueryGui;
    struct List qe_Queries;
};

struct Query
{
    struct Node sq_Node;
    struct QueryEditor *sq_QueryEditor;
    char *sq_Name;
    struct Table *sq_Table;
    struct Program *sq_Program;
};

static void FreeQuery(struct Query *sq)
{
    if (sq != NULL) {
        struct Project *pr = QueryEditorProject(sq->sq_QueryEditor);
        struct MemPool *memPool = ProjectMemPool(pr);

        FreeString(memPool, sq->sq_Name);
        FreeProgram(sq->sq_Program);
        PoolFreeMem(memPool, sq, sizeof(struct Query));
    }
}

const char *QueryName(struct Query *sq)
{
    const char *name = NULL;

    if (sq) {
        name = sq->sq_Name? sq->sq_Name : GCS(MSG_Unnamed);
    }
    return name;
}

void SetQueryName(struct Query *sq, const char *name)
{
    if (sq) {
        struct Project *pr = QueryEditorProject(sq->sq_QueryEditor);
        struct MemPool *memPool = ProjectMemPool(pr);
	char *copy = AllocString(memPool, name);

	FreeString(memPool, sq->sq_Name);
	sq->sq_Name = copy;
    }
}

struct Table *QueryTable(struct Query *sq)
{
    struct Table *tab = NULL;

    if (sq) {
        tab = sq->sq_Table;
    }
    return tab;
}

void SetQueryTable(struct Query *sq, struct Table *tab)
{
    if (sq) {
        sq->sq_Table = tab;
    }
}

struct Program *QueryProgram(struct Query *sq)
{
    struct Program *prg = NULL;

    if (sq) {
        prg = sq->sq_Program;
    }
    return prg;
}

void SetQueryProgram(struct Query *sq, struct Program *prg)
{
    if (sq) {
        FreeProgram(sq->sq_Program);
        sq->sq_Program = prg;
    }
}

struct Query *NewQuery(struct QueryEditor *qe, struct Query *init)
{
    struct Project *pr = QueryEditorProject(qe);
    struct MemPool *memPool = ProjectMemPool(pr);
    struct Query *sq;

    sq = PoolAllocMemZero(memPool, sizeof(struct Query));
    if (sq == NULL) {
        return NULL;
    }

    sq->sq_QueryEditor = qe;

    if (init != NULL && init->sq_Name != NULL) {
        sq->sq_Name = AllocString(memPool, init->sq_Name);
        if (sq->sq_Name == NULL) {
            FreeQuery(sq);
            return NULL;
        }
    }
    if (init != NULL && init->sq_Program != NULL) {
        struct SourceText *st = GetSourceText(init->sq_Program);
        const char *src = SourceTextText(st), *error = NULL;
        long len;

        if (src && (len = strlen(src)) > 1) {
            src++;
            len -= 2;
            error = CompileQuery(pr, src, len, &sq->sq_Program, NULL, NULL);
        }
        FreeSourceText(st);
        if (error != NULL) {
            FreeQuery(sq);
            return NULL;
        }
    }

    ADDTAIL(&qe->qe_Queries, sq);

    return sq;
}

void DeleteQuery(struct Query *sq)
{
    if (sq) {
        REMOVE(sq);
        FreeQuery(sq);
    }
}

void ClearQueryEditor(struct QueryEditor *qe)
{
    if (qe) {
        struct Query *sq;

        CloseQueryGui(qe->qe_QueryGui);
        qe->qe_QueryGui = NULL;

        while ((sq = (struct Query *)REMHEAD(&qe->qe_Queries)) != NULL) {
            FreeQuery(sq);
        }
    }
}

void FreeQueryEditor(struct QueryEditor *qe)
{
    if (qe) {
        struct MemPool *memPool = ProjectMemPool(qe->qe_Project);

        ClearQueryEditor(qe);
        PoolFreeMem(memPool, qe, sizeof(struct QueryEditor));
    }
}

struct QueryEditor *AllocQueryEditor(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct QueryEditor *qe;

    qe = PoolAllocMemZero(memPool, sizeof(struct QueryEditor));
    if (qe != NULL) {
        qe->qe_Project = pr;
        NewList(&qe->qe_Queries);
    }

    return qe;
}

struct Project *QueryEditorProject(struct QueryEditor *qe)
{
    struct Project *pr = NULL;

    if (qe) {
        pr = qe->qe_Project;
    }
    return pr;
}

GuiObject QueryEditorGui(struct QueryEditor *qe)
{
    GuiObject go = NULL;

    if (qe) {
        go = qe->qe_QueryGui;
    }
    return go;
}

struct List *QueryEditorQueries(struct QueryEditor *qe)
{
    struct List *queries = NULL;

    if (qe) {
        queries = &qe->qe_Queries;
    }
    return queries;
}

void CloseQueryEditor(struct Project *pr)
{
    struct QueryEditor *qe = ProjectQueryEditor(pr);

    if (qe != NULL) {
        CloseQueryGui(qe->qe_QueryGui);
        qe->qe_QueryGui = NULL;
    }
}

static void showQueryEditor(struct QueryEditor *qe, struct Query *query)
{
    if (qe) {
        if (qe->qe_QueryGui != NULL) {
            QueryGuiActivate(qe->qe_QueryGui);
        } else {
            qe->qe_QueryGui = OpenQueryGui(qe);
            if (qe->qe_QueryGui == NULL) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_CreateGui));
                return;
            }
        }
        QueryGuiRunQuery(qe->qe_QueryGui, query);
    }
}

void OpenQueryEditor(struct Project *pr)
{
    if (ProjectStructureEditor(pr) == NULL) {
        /* only open when structure editor is not present */
        showQueryEditor(ProjectQueryEditor(pr), NULL);
    }
}

static struct Query *findTableQuery(struct QueryEditor *qe, struct Table *tab)
{
    struct Query *found = NULL, *sq;

    if (qe != NULL && tab != NULL) {
        FORNODE(sq, &qe->qe_Queries) {
            if (sq->sq_Table == tab) {
                found = sq;
                break;
            }
        }
    }
    return found;
}

void OpenQueryEditorWithTableQuery(struct Table *tab)
{
    struct Project *pr = TableProject(tab);

    if (tab != NULL && ProjectStructureEditor(pr) == NULL) {
        /* only open when structure editor is not present */
        struct QueryEditor *qe = ProjectQueryEditor(pr);
        struct Query *query = findTableQuery(qe, tab);

        if (query == NULL) {
            query = NewQuery(qe, NULL);
            if (query == NULL) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
                return;
            } else {
                const char *name = TableName(tab);
                char src[30 + ID_NAME_LEN + 1];
                const char *error;

                SetQueryName(query, name);
                SetQueryTable(query, tab);
                sprintf(src, "SELECT * FROM %s", name);
                error = CompileQuery(pr, src, strlen(src), 
                        &query->sq_Program, NULL, NULL);
                if (error != NULL) {
                    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Internal));
                    DeleteQuery(query);
                    return;
                }
            }
        }
        showQueryEditor(qe, query);
    }
}

bool QueryEditorIsOpen(struct Project *pr)
{
    struct QueryEditor *qe = ProjectQueryEditor(pr);

    return (QueryEditorGui(qe) != NULL);
}

static void queryDelete(struct Project *pr, void *ptr)
{
    struct QueryEditor *qe = ProjectQueryEditor(pr);
    struct List *queries = QueryEditorQueries(qe);
    struct Query *sq;

    if (queries) {
        FORNODE(sq, queries) {
            if (sq->sq_Program && ProgramFindObject(sq->sq_Program, ptr)) {
                SetQueryProgram(sq, NULL);
            }
            if (sq->sq_Table == ptr) {
                sq->sq_Table = NULL;
            }
        }
    }
}

void QueryEditorDeleteTable(struct Table *tab)
{
    struct Project *pr = TableProject(tab);
    struct List *fields = TableFields(tab);
    struct Field *field;

    queryDelete(pr, tab);
    if (fields) {
        FORNODE(field, fields) {
            queryDelete(pr, field);
        }
    }
}

void QueryEditorDeleteField(struct Field *field)
{
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);

    queryDelete(pr, field);
}


/*
** load & save
*/

long LoadQueriesCustomFormat(struct Project *pr, FILE *file, short version)
{
    struct QueryEditor *qe = ProjectQueryEditor(pr);
    struct MemPool *memPool = ProjectMemPool(pr);
    struct Query *sq;
    long count, len, error = OK;
    size_t warnings;
    char *tmp;

    if (file == NULL) {
        return ERR_INTERNAL;
    }

    ClearQueryEditor(qe);
    count = FGetLong(file);
    for (; count > 0; count--) {
        sq = NewQuery(qe, NULL);
        if (sq == NULL) {
            return ERR_MEM;
        }
        if (version >= 4) {
            sq->sq_Name = ReadStringConvert(file, TRUE, memPool,
                    &error, &warnings);
            if (error) {
                return error;
            }
            ProjectAddCodesetWarnings(pr, warnings);
        } else {
            len = FGetLong(file);
            sq->sq_Name = PoolAllocMem(memPool, len + 1);
            if(sq->sq_Name == NULL)
                return ERR_MEM;
            if (len > 0) {
                len = fread(sq->sq_Name, 1, len, file);
            }
            sq->sq_Name[len] = 0;
            if (!StrIsCodeset(sq->sq_Name)) {
                 tmp=CodesetConvertFromDefault8Bit(sq->sq_Name, memPool,&error);
                 if (error) {
                     return error;
                 }
                 if (tmp != sq->sq_Name) {
                     FreeString(memPool, sq->sq_Name);
                     sq->sq_Name = tmp;
                 }
            }
        }
        error = LoadProgramCustomFormat(pr, file, &sq->sq_Program);
        if (error) {
            return error;
        }
        if (version >= 5) {
            sq->sq_Table = FindTableNum(pr, FGetLong(file));
        }
    }
    return OK;
}

