/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** queryeditor.h
**
** Creation date: 28.05.1995
** Last modified: 11.05.2022
*/

#ifndef QUERYEDITOR_H
#define QUERYEDITOR_H

struct QueryEditor;
struct Query;

#ifdef NEED_NODE_CAST
inline Query *nodecast(Query *, Node *node) { return (Query *)node; }
#endif

#include "lang/program.h"

#ifdef __cplusplus
extern "C" {
#endif

void ClearQueryEditor(struct QueryEditor *qe);
void FreeQueryEditor(struct QueryEditor *qe);
struct QueryEditor *AllocQueryEditor(struct Project *pr);

void CloseQueryEditor(struct Project *pr);
void OpenQueryEditor(struct Project *pr);
void OpenQueryEditorWithTableQuery(struct Table *tab);
bool QueryEditorIsOpen(struct Project *pr);

void QueryEditorDeleteTable(struct Table *tab);
void QueryEditorDeleteField(struct Field *field);
void QueryEditorRemRecord(struct Table *tab, struct Record *rec);
void QueryEditorRemAllRecords(struct Project *pr);

/* load & save */

long LoadQueriesCustomFormat(struct Project *pr, FILE *file, short version);

/*
** for internal use (QueryEditor <-> QueryEditor GUI)
*/

void CloseQueryGui(GuiObject q);
GuiObject OpenQueryGui(struct QueryEditor *qe);
void QueryGuiActivate(GuiObject q);
void QueryGuiRunQuery(GuiObject go, struct Query *query);

struct QueryEntry
{
    struct Record *qe_Record;
    char qe_String[0];
};

struct QueryLine
{
    long ql_LineNum;
    bool ql_Checked;
    struct QueryEntry *ql_Entries[0];
};

struct QueryFieldInfo {
    struct Field *qfi_Field;
    char qfi_Format;
    char qfi_Type;
    size_t qfi_Width;
};

struct QueryGuiData
{
    struct MemPool *qgd_MemPool;
    size_t qgd_NumFields;
    size_t qgd_NumLines;
    struct Table *qgd_ControlTable;
    size_t qgd_ControlIndex;
    struct QueryFieldInfo *qgd_FieldInfos;
    struct QueryLine *qgd_Titles; 
    struct QueryLine **qgd_Lines;
};

long QueryLineLines(const struct QueryLine *ql, long numEntries);
/*
** Returns number of lines a query line occupies for printing or displaying.
** For example an entry that contains a single newline occupies two lines.
*/

long QueryEntryCmp(struct QueryEntry *qe1, struct QueryEntry *qe2, char type,
        bool star);
/*
** Compares two query entries by converting them using the provided type.
*/

void FreeQueryGuiData(struct QueryGuiData *qgd);
/*
** Deletes the memory pool created for a QueryGuiData instance.
*/

struct QueryGuiData *QueryGuiDataFromTitlesBody(
        const struct Element *titles, const struct Element *body);
/*
** Creates QueryGuiData from the title line and body of a SFW expression.
** This creates its own MemPool.  Returns NULL if memory allocation fails.
*/

struct QueryGuiData *QueryGuiDataFromString(const char *str);
/*
** Creates QueryGuiData from a string which is set in the qgd_Title field.
** This creates its own MemPool.  Returns NULL if memory allocation fails.
*/

bool QueryGuiDataCopyLineNumIfMatch(struct QueryGuiData *qgd1,
        const struct QueryGuiData *qgd2);
/*
** If the two QueryGuiData contain the exact same information except for the
** ql_LineNum of each QueryLine then copy theql_LineNum's from qgd2 to qgd1.
** This is useful after a drag n' drop sort operation that calls the move-drop
** callback which sets a new query result that very likely is identical to
** the current contents except for new line numbers (ql_LineNum).
*/

bool QueryGuiDataAllSingleLine(const struct QueryGuiData *qgd);
/*
** Returns TRUE if all query entries occupy a single line.
*/

void QueryGuiDataMoveLine(struct QueryGuiData *qgd, size_t row, size_t newRow);
/*
** Moves a line at a given current row to newRow such that the current line 
** at newRow follows the current line at row.  If newRow is larger than
** the number of lines, moves the line at row to the end.
** This moves all lines between row and newRow.
*/

const char *QueryName(struct Query *sq);
void SetQueryName(struct Query *sq, const char *name);
struct Table *QueryTable(struct Query *sq);
void SetQueryTable(struct Query *sq, struct Table *tab);
struct Program *QueryProgram(struct Query *sq);
void SetQueryProgram(struct Query *sq, struct Program *prg);
/* frees old program if exists */
struct Query *NewQuery(struct QueryEditor *qe, struct Query *init);
void DeleteQuery(struct Query *sq);

struct Project *QueryEditorProject(struct QueryEditor *qe);
GuiObject QueryEditorGui(struct QueryEditor *qe);
struct List *QueryEditorQueries(struct QueryEditor *qe);

#ifdef __cplusplus
}
#endif

#endif

