/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** record.c
**
** Creation date: 29.02.1995
** Last modified: 03.05.2022
*/

#include "record.h"
#include "idrecordmap.h"
#include "main.h"
#include "gui.h"
#include "log.h"
#include "memo.h"
#include "lang/program.h"
#include "preferences.h"
#include "sqlitedb.h"
#include "dosutil.h"
#include "util.h"
#include "strhandle.h"
#include "break.h"
#include "memory.h"
#include "error.h"
#include <stdarg.h>

#ifdef CHECK_RECORD_HEADERS
#include "requesters.h"
#include "appstrings_cat.h"
#endif

struct TableRecordInfo
{
    struct Table *tri_Table;
    struct List tri_FieldRecordInfos;
    ulong tri_NextRecordNum;
    struct List tri_UpdateRecords;
    struct IdRecordMap *tri_IdRecordMap;
    double tri_MinOrderNumResolution;
    bool tri_IsCustomFormat;
    struct {
        FILE *file;
        ulong recChangeId;
        struct List recChanges;
    } tri_CustomFormat;
};

struct Record
{
    struct MinNode rec_Node;
    ulong rec_Num;
    ulong rec_MatchFilter:1;
    ulong rec_IsSorted:1;
    ulong rec_IsImported:1;
    ulong rec_OnDisk:1;
    ulong rec_HeaderModified:1;
    ulong rec_DataModified:1;
    ulong rec_LockCount:27;
    long rec_FileVersion;       /* Version number of record in project file */
    void *rec_Data;
    struct {
        long filePos;
        long fileTotalSize;     /* Space needed on disk w/o RecDisk header */
    } rec_CustomFormat;
};

enum
{
    OP_NEW=12, OP_DELETE=14, OP_NONE=17,
};

struct CustomFormatRecChangeFieldInfo
{
    short ai_Type;
};

struct CustomFormatRecChange
{
    struct MinNode rc_Node;
    ulong rc_Id;
    short rc_Op;
    union
    {
        struct rc_new
        {
            short type;
	    struct DataInfo di;
        } new;
        struct
        {
            long num;
        } delete;
    } rc_Contents;
    long rc_NumFields;
    struct CustomFormatRecChangeFieldInfo rc_FieldInfos[0];
};

struct FieldRecordInfo		/* Info structure for fields */
{
    struct MinNode fri_Node;
    struct Field *fri_Field;	/* Points back to Field */
    long fri_Pos;		/* Byte-Position in Data */
};

#define DIP_DISKUTF8  (1 << 0)          /* string is stored in utf-8 on disk */

static void FreeCustomFormatRecChange(struct MemPool *memPool,
        struct CustomFormatRecChange *rc)
{
    if (rc != NULL) {
        if (rc->rc_Op == OP_NEW &&  rc->rc_Contents.new.type == TYPE_STRING) {
	    FreeString(memPool, rc->rc_Contents.new.di.di_Contents.string);
        }
        PoolFreeMem(memPool, rc, sizeof(struct CustomFormatRecChange) +
	    rc->rc_NumFields * sizeof(struct CustomFormatRecChangeFieldInfo));
    }
}

static void closeCustomFormatSwapFile(struct TableRecordInfo *tri)
{
    if (tri && tri->tri_CustomFormat.file != NULL) {
	fclose(tri->tri_CustomFormat.file);
       	tri->tri_CustomFormat.file = NULL;
    }
}

static void ClearTableRecordInfo(struct TableRecordInfo *tri)
{
    if (tri != NULL) {
        struct MemPool *memPool = TableMemPool(tri->tri_Table);
        struct CustomFormatRecChange *rc;

        closeCustomFormatSwapFile(tri);
        tri->tri_MinOrderNumResolution = 1.0;
        tri->tri_CustomFormat.recChangeId = 0;
        while ((rc = (struct CustomFormatRecChange *)
                    REMHEAD(&tri->tri_CustomFormat.recChanges))!= NULL){
            FreeCustomFormatRecChange(memPool, rc);
        }
    }
}

void FreeTableRecordInfo(struct TableRecordInfo *tri)
{
    if (tri) {
        struct MemPool *memPool = TableMemPool(tri->tri_Table);
 	struct Record *rec;

	while ((rec = (struct Record *)
                    REMHEAD(&tri->tri_UpdateRecords)) != NULL) {
	    FreeRecord(tri->tri_Table, rec, FALSE);
        }
        FreeIdRecordMap(memPool, tri->tri_IdRecordMap);
        tri->tri_IdRecordMap = NULL;

        ClearTableRecordInfo(tri);
        PoolFreeMem(TableMemPool(tri->tri_Table), 
		tri, sizeof(struct TableRecordInfo));
    }
}

struct TableRecordInfo *AllocTableRecordInfo(struct Table *tab)
{
    struct MemPool *memPool = TableMemPool(tab);
    struct TableRecordInfo *tri;
    struct IdRecordMap *map;

    map = AllocIdRecordMap(memPool);
    if (map == NULL) {
        return NULL;
    }

    tri = PoolAllocMemZero(memPool, sizeof(struct TableRecordInfo));
    if (tri == NULL) {
        FreeIdRecordMap(memPool, map);
        return NULL;
    }

    tri->tri_Table = tab;
    NewList(&tri->tri_FieldRecordInfos);
    tri->tri_NextRecordNum = 1;
    NewList(&tri->tri_UpdateRecords);
    tri->tri_IdRecordMap = map;
    tri->tri_MinOrderNumResolution = 1.0;
    NewList(&tri->tri_CustomFormat.recChanges);

    return tri;
}

void SetTableRecordInfoCustomFormat(struct TableRecordInfo *tri, bool set)
{
    if (tri != NULL) {
	tri->tri_IsCustomFormat = set;
    }
}

static short FieldTypeCustomFormatFileSize(short type)
{
    short size = 2;

    switch(type)
    {
        case TYPE_INTEGER:	size += 4; break;       /* 32 bit */
	case TYPE_REAL:	        size += 8; break;       /* 64 bit */
	case TYPE_BOOL:	        size += 0; break;
	case TYPE_STRING:	size += 0; break;
	case TYPE_MEMO:	        size += 0; break;
	case TYPE_DATE:	        size += 4; break;       /* 32 bit */
	case TYPE_TIME:	        size += 4; break;       /* 32 bit */
	case TYPE_CHOICE:	size += 2; break;       /* 16 bit */
        case TYPE_REFERENCE:    size += 4; break;       /* 32 bit */
	case TYPE_VIRTUAL:
        case TYPE_BUTTON:
	default:		size = 0;	break;
    }
    return size;
}

static short FieldTypeMemSize(short type)
{
    short size = 2;

    switch(type)
    {
        case TYPE_INTEGER:	size += sizeof(long);	break;
	case TYPE_REAL:	        size += sizeof(double);	break;
	case TYPE_BOOL:	        size += 0;	break;
	case TYPE_STRING:	size += sizeof(char *);	break;
	case TYPE_MEMO:	        size += sizeof(char *);	break;
	case TYPE_DATE:	        size += sizeof(long);	break;
	case TYPE_TIME:	        size += sizeof(long);	break;
	case TYPE_CHOICE:	size += sizeof(short);	break;
        case TYPE_REFERENCE:    size += sizeof(struct Record *); break;
	case TYPE_VIRTUAL:
        case TYPE_BUTTON:
	default:		size = 0;	break;
    }
    return(size);
}

#define FieldMemSize(field) FieldTypeMemSize(FieldType(field))

static long TableMemSize(struct Table *tab)
{
    long size = 0;
    struct List *fields = TableFields(tab);
    struct Field *field;

    if (fields) {
        FORNODE(field, fields) {
	    size += FieldMemSize(field);
        }
    }

    return size;
}

static long openCustomFormatSwapFile(struct TableRecordInfo *tri)
{
    if (tri == NULL) {
        return ERR_INTERNAL;
    }
    if (tri->tri_CustomFormat.file != NULL) {
        return OK;
    }
    tri->tri_CustomFormat.file = FOpen(TableName(tri->tri_Table), "rb");
    if (tri->tri_CustomFormat.file == NULL) {
        return RECERR_OPENSWAPFILE;            /* failed */
    }
    setvbuf(tri->tri_CustomFormat.file, NULL, _IOFBF, 512);  
				/* 512 seems to be better than 8192 here */
    return OK;
}

static long openCustomFormatSwapFileProjectDir(struct TableRecordInfo *tri)
{
    struct Project *pr;
    const char *dirname;
    Cwd cwd = NULL;
    long error;

    if (tri == NULL) {
        return ERR_INTERNAL;
    }
    if (tri->tri_CustomFormat.file != NULL) {
        return OK;
    }

    pr = TableProject(tri->tri_Table);
    dirname = ProjectName(pr);
    if (dirname == NULL) {
        return RECERR_NOPROJECTNAME;
    }

    ProjectSetCurrentDir(NULL);
    if (PushDir(dirname, &cwd) != 0) {
        error = RECERR_OPENSWAPFILE;
    } else {
	error = openCustomFormatSwapFile(tri);
    }
    PopDir(&cwd);

    return error;
}

void RecordLock(struct Record *rec)
{
    if (rec != NULL) {
        rec->rec_LockCount++;
    }
}

void RecordUnlock(struct Record *rec)
{
    if (rec) {
        rec->rec_LockCount--;
    }
}

static bool getdatainfo(const char *data, 		/* never NULL */
	struct Field *field, 			        /* NULL ok */
	struct DataInfo *di)				/* never NULL */
{
    struct FieldRecordInfo *fri = FieldRecordInfo(field);
    short size = FieldMemSize(field);
    struct DataInfo *fielddi;

    if (fri != NULL && size >= 2) {
        data +=  fri->fri_Pos;
        di->di_Flags = data[0];
        di->di_Private = data[1];
        memcpy(&di->di_Contents, data + 2, size - 2);
	return TRUE;
    } else if ((fielddi = FieldDataInfo(field)) != NULL) {
	*di = *fielddi;
	return TRUE;
    }
    return FALSE;
}

static void setdatainfo(char *data, 			/* never NULL */
	struct Field *field, 			        /* NULL ok */
	const struct DataInfo *di)			/* never NULL */
{
    struct FieldRecordInfo *fri = FieldRecordInfo(field);
    short size = FieldMemSize(field);
    struct DataInfo *fielddi;

    if (fri != NULL && size >= 2) {
	data += fri->fri_Pos;
        data[0] = di->di_Flags;
        data[1] = di->di_Private;
	memcpy(data + 2, &di->di_Contents, size - 2);
    } else if ((fielddi = FieldDataInfo(field)) != NULL) {
	*fielddi = *di;
    }
}

static long recordTotalMemSize(char *data, struct Table *tab)
/* record must be in memory! */
{
    long size = TableMemSize(tab);
    struct List *fields = TableFields(tab);
    struct Field *field;
    struct DataInfo di;

    if (fields) {
        FORNODE(field, fields) {
            switch (FieldType(field))
            {
                case TYPE_STRING:
		    if (getdatainfo(data, field, &di) && di.di_Contents.string){
			size += strlen(di.di_Contents.string) + 1;
		    }
                    break;

                case TYPE_MEMO:
		    if (getdatainfo(data, field, &di) && di.di_Contents.mi) {
			size += strlen(di.di_Contents.mi) + 1;
		    }
                    break;
            }
        }
    }
    return size;
}

static void freeRecordData(struct Table *tab, struct Record *rec)
{
    struct Project *pr = TableProject(tab);
    struct MemPool *dataPool = ProjectDataMemPool(pr);
    struct List *fields = TableFields(tab);
    struct Field *field;
    struct DataInfo di;

    if (rec && rec->rec_Data) {
        AddProjectRecordMem(pr, -(recordTotalMemSize(rec->rec_Data, tab)));

        if (fields) {
            FORNODE (field, fields) {
                switch (FieldType(field)) {
                    case TYPE_STRING:
			if (getdatainfo(rec->rec_Data, field, &di)) {
			    FreeString(dataPool, di.di_Contents.string);
			}
                        break;

                    case TYPE_MEMO:
			if (getdatainfo(rec->rec_Data, field, &di)) {
			    FreeString(dataPool, di.di_Contents.mi);
			}
                        break;
                }
	    }
	}
        PoolFreeMem(dataPool, rec->rec_Data, TableMemSize(tab));
        rec->rec_Data = NULL;
    }
}

long RecordNum(struct Record *rec)
{
    return NodeNum(rec);
}

bool MoveRecord(struct Table *tab, struct Record *rec, long num)
{
    struct List *recs = TableRecords(tab);

    if (RecordNum(rec) == num) {
        return TRUE;
    }
    if (rec == NULL) {
        return FALSE;
    }
    if (num < 1 || num > NumNodes(recs)) {
        return FALSE;
    }
    REMOVE(rec);
    INSERT(recs, rec, FindNodeNum(recs, num - 1));
    SetRecordIsSorted(rec, FALSE);
    return TRUE;
}

#define RecordNewInternalNum(rec) RecordNum(rec)

long RecordInternalNum(struct Record *rec)
{
    ulong num = 0;

    if (rec) {
        num = rec->rec_Num;
    }
    return num;
}

static struct Record *FindRecordInternalNumList(struct List *recs, ulong num)
{
    struct Record *rec;

    if (recs != NULL && num != 0) {
        FORNODE (rec, recs) {
            if (rec->rec_Num == num) {
                return rec;
            }
        }
    }

    return NULL;
}

long RecordFileVersion(struct Record *rec)
{
    long version = 0;

    if (rec != NULL) {
        version = rec->rec_FileVersion;
    }
    return version;
}

struct Record *FindRecordInternalNum(struct Table *tab, ulong num)
{
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct Record *rec= NULL;

    if (tri != NULL) {
        rec = RecordFromId(tri->tri_IdRecordMap, num);
    }
    return rec;
}

void updateMinRecordOrderNumResolution(struct TableRecordInfo *tri,
        double orderNumResolution)
{
    if (tri != NULL && orderNumResolution < tri->tri_MinOrderNumResolution) {
        tri->tri_MinOrderNumResolution = orderNumResolution;
    }
}

double TableRecordOrderNumResolution(struct Table *tab)
{
    double resolution = 1.0;
    struct TableRecordInfo *tri = TableRecordInfo(tab);

    if (tri != NULL) {
        resolution = tri->tri_MinOrderNumResolution;
    }
    return resolution;
}

void ResetRecordOrderNumResolutions(struct Project *pr)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    struct TableRecordInfo *tri;

    if (tabs != NULL) {
        FORNODE(tab, tabs) {
            tri = TableRecordInfo(tab);
            if (tri != NULL) {
                tri->tri_MinOrderNumResolution = 1.0;
            }
        }
    }
}

static long fieldCustomFormatFileSize(char *data, long pos,
        struct CustomFormatRecChangeFieldInfo *ai)
{
    long size = FieldTypeCustomFormatFileSize(ai->ai_Type);

    switch(ai->ai_Type)
    {
        case TYPE_STRING:
	case TYPE_MEMO:
            size += strlen(data + pos + size) + 1;
            break;
    }
 
    return(size);
}
 
static long fieldCustomFormatFilePos(void *data,
        struct CustomFormatRecChangeFieldInfo *ai, long num)
{
    long i, pos = 0;
 
    for(i=0; i < num; i++) {
        pos += fieldCustomFormatFileSize(data, pos, &ai[i]);
    }

    return pos;
}


static long updateRecordCustomFormat(struct Record *rec,
        struct TableRecordInfo *tri, ulong id, FILE *file,
        struct MemPool *dataPool, char **pdata,
        long *pdatasize, long *pnewoffset)
{
    struct CustomFormatRecChange *rc, *start;
    struct CustomFormatRecChangeFieldInfo *ai;
    long add = 0, size = 0, num, pos;
    long filetotalsize = rec->rec_CustomFormat.fileTotalSize, read; 
    long newoffset = filetotalsize;
    struct rc_new *rc_new;
    struct DataInfo *di;
    char *data = NULL, *ptr;
    long datasize = 0;

    FORNODE(rc, &tri->tri_CustomFormat.recChanges) {
	if (rc->rc_Id == id) {
	    break;
        }
    }
    if (SUCC(rc) == NULL) {
	return ERR_INTERNAL;
    }

    start = rc;
    for (rc = start; SUCC(rc); rc = SUCC(rc)) {
	ai = rc->rc_FieldInfos;
	switch (rc->rc_Op)
	{
	    case OP_NEW:    
		rc_new = &rc->rc_Contents.new;
		di = &rc_new->di;
		size += FieldTypeCustomFormatFileSize(rc_new->type);
		if (rc_new->type == TYPE_STRING) {
		    if (di->di_Contents.string != NULL) {
			size += strlen(di->di_Contents.string);
                    }
		    size++;
		} else if (rc_new->type == TYPE_MEMO) {
		    size++;
                }
		break;

	    case OP_DELETE: 
                num = rc->rc_Contents.delete.num;
		size -= FieldTypeCustomFormatFileSize(ai[num].ai_Type);
		break;
	}
	if (size > add) {
	    add = size;
        }
    }

    datasize = rec->rec_CustomFormat.fileTotalSize + add;
    if (datasize > 0) {
	data = PoolAllocMem(dataPool, datasize);
	if (data == NULL) {
	    return ERR_MEM;
        }
        if (filetotalsize > 0) {
            read = fread(data, 1, filetotalsize, file);
            if (read != filetotalsize || ferror(file)) {
                PoolFreeMem(dataPool, data, datasize);
                return RECERR_READSWAPFILE;
            }
	}
    }

    for (rc = start; SUCC(rc); rc = SUCC(rc)) {
	ai = rc->rc_FieldInfos;
	switch (rc->rc_Op)
	{
	    case OP_NEW:
		rc_new = &rc->rc_Contents.new;
		if (FieldTypeCustomFormatFileSize(rc_new->type) == 0) {
		    break;				/* nothing to do */
                }
		di = &rc_new->di;
                ptr = data + filetotalsize;
		filetotalsize += FieldTypeCustomFormatFileSize(rc_new->type);
                *ptr++ = di->di_Flags;
                *ptr++ = di->di_Private;

		switch (rc_new->type)
		{
                    case TYPE_INTEGER:
                        SPutLong(ptr, di->di_Contents.integer);
                        break;
                    case TYPE_REAL:
                        SPutDouble(ptr, di->di_Contents.real);
                        break;
                    case TYPE_BOOL:
                        break;
                    case TYPE_STRING:
			if (di->di_Contents.string != NULL) {
			    strcpy(ptr, di->di_Contents.string);
			} else {
                            *ptr = 0;
                        }
                        filetotalsize += strlen(ptr) + 1;
			break;
                    case TYPE_MEMO:
                        *ptr = 0;
                        filetotalsize++;
                        break;
                    case TYPE_DATE:
                        SPutLong(ptr, di->di_Contents.date);
                        break;
                    case TYPE_TIME:
                        SPutLong(ptr, di->di_Contents.time);
                        break;
                    case TYPE_CHOICE:
                        SPutShort(ptr, di->di_Contents.choice);
                        break;
                    case TYPE_REFERENCE:
                        SPutLong(ptr, 0);
                        break;
		}
		break;

	    case OP_DELETE:
		num = rc->rc_Contents.delete.num;
		pos = fieldCustomFormatFilePos(data, ai, num);
		size = fieldCustomFormatFileSize(data, pos, &ai[num]);

		memmove(data+pos, data+pos+size, filetotalsize-pos-size);
		filetotalsize -= size;
                newoffset -= size;
		break;
	}
    }

    *pdata = data;
    *pdatasize = datasize;
    *pnewoffset = newoffset;
    return OK;
}

#define FORMAT_ID_MASK 0x80000000

static long loadRecordCustomFormat(struct Table *tab, struct Record *rec,
        size_t *pTotalWarnings)
{
    struct Project *pr = TableProject(tab);
    struct MemPool *dataPool = ProjectDataMemPool(pr);
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct FieldRecordInfo *fri;
    FILE *file;
    long recsize = TableMemSize(tab);
    long datasize, read, newoffset, pos;
    char *data = NULL;
    ulong id;
    bool oldFormat;
    long err;
    size_t totalWarnings = 0;
    size_t warnings = 0;

    err = openCustomFormatSwapFileProjectDir(tri);
    if (err) {
        return err;
    }

    file = tri->tri_CustomFormat.file;
    clearerr(file);
    if (ftell(file) != rec->rec_CustomFormat.filePos) {
        fseek(file, rec->rec_CustomFormat.filePos, SEEK_SET);
    }
    id = FGetLong(file);
    if (ferror(file)) {
        return RECERR_READSWAPFILE;
    }
    oldFormat = ((id & FORMAT_ID_MASK) == 0);
    id &= ~FORMAT_ID_MASK;

    if (id == tri->tri_CustomFormat.recChangeId) {
        datasize = rec->rec_CustomFormat.fileTotalSize;
        newoffset = datasize;
        if (datasize > 0) {
            data = PoolAllocMem(dataPool, datasize);
            if (data == NULL) {
                return ERR_MEM;
            }
            read = fread(data, 1, datasize, file);
            if (read != datasize || ferror(file)) {
                PoolFreeMem(dataPool, data, datasize);
                return RECERR_READSWAPFILE;
            }
        }
    } else {
        err = updateRecordCustomFormat(rec, tri, id, file,
            dataPool, &data, &datasize, &newoffset);
        if (err) {
            return err;
        }
    }

    if (recsize > 0) {
        rec->rec_Data = PoolAllocMemZero(dataPool, recsize);
        if (rec->rec_Data == NULL) {
            PoolFreeMem(dataPool, data, datasize);
            return ERR_MEM;
        }
    }

    pos = 0;
    FORNODE(fri, &tri->tri_FieldRecordInfos) {
        short type = FieldType(fri->fri_Field);
        short size = FieldTypeCustomFormatFileSize(type);
        char *target = (char *)rec->rec_Data + fri->fri_Pos;
        struct DataInfo *di = (struct DataInfo *)target;

        if (size > 0) {
            if (oldFormat && pos < newoffset) {
                di->di_Flags = ((data[pos+1] &  1)? DIF_NIL : 0);
                di->di_Private = 0;
            } else {
                di->di_Flags = data[pos];
                di->di_Private = data[pos+1];
            }
            pos += 2;
            size -= 2;
            target += 2;
        }

        switch (type) {
            case TYPE_INTEGER:
            {
                long l = SGetLong(data + pos);
                memcpy(target, &l, sizeof(long));
                break;
            }
            case TYPE_REAL:
            {
                double d = SGetDouble(data + pos);
                memcpy(target, &d, sizeof(double));
                break;
            }
            case TYPE_BOOL:
                break;
            case TYPE_STRING:
            case TYPE_MEMO:
            {
                char *src = data + pos, *str;
               
                if (pos >= newoffset) {
                    str = src;		/* already in Codeset */
                } else if (!(di->di_Private & DIP_DISKUTF8)) {
                    str=CodesetConvertFromDefault8Bit(src, dataPool,&err);
                } else {
                    str = CodesetConvertFromUTF8(src, dataPool, &err, 
                            &warnings);
                    ProjectAddCodesetWarnings(pr, warnings);
                    totalWarnings += warnings;
                }
                if (str == src) {
                    str = AllocString(dataPool, src);
                    if (str == NULL) {
                        err = ERR_MEM;
                    }
                }
                memcpy(target, &str, sizeof(char *));
                pos += strlen(src) + 1;
                di->di_Flags &= ~DIF_NIL;
                break;
            }
            case TYPE_DATE:
            {
                long d = SGetLong(data + pos);
                memcpy(target, &d, sizeof(long));
                break;
            }
            case TYPE_TIME:
            {
                long t = SGetLong(data + pos);
                memcpy(target, &t, sizeof(long));
                break;
            }
            case TYPE_CHOICE:
            {
                short c = SGetShort(data + pos);
                memcpy(target, &c, sizeof(short));
                break;
            }
            case TYPE_REFERENCE:
            {
                ulong recnum = SGetLong(data + pos);
                struct Record *rec = FindRecordInternalNum(
                    FieldReferenceTable(fri->fri_Field), recnum);
                memcpy(target, &rec, sizeof(struct Record *));
                break;
            }
        }
        pos += size;
        if (err != OK) {
            break;
        }
    }

    PoolFreeMem(dataPool, data, datasize);

    if (err == OK && ferror(file)) {
        err = RECERR_READSWAPFILE;
    }

    if (pTotalWarnings != NULL) {
        *pTotalWarnings = totalWarnings;
    }

    return err;
}

static long loadRecordSQLite(struct Table *tab, struct Record *rec,
        size_t *pTotalWarnings)
{
    struct Project *pr = TableProject(tab);
    struct SQLiteHandle *handle = ProjectSQLiteHandle(pr);
    struct MemPool *dataPool = ProjectDataMemPool(pr);

    rec->rec_Data = PoolAllocMemZero(dataPool, TableMemSize(tab));
    if (rec->rec_Data == NULL) {
        return ERR_MEM;
    }

    return SQLiteReadRecord(handle, tab, rec->rec_Num, rec->rec_Data, dataPool,
            setdatainfo, pTotalWarnings);
}

static void checkRecordMem(struct Project *pr)
{
    long mem = AddProjectRecordMem(pr, 0);
    long limit = GetPrefsValue(pr, PREFS_RECORDMEMORY);

    if (mem > limit) {
        SwapProjectRecords(pr);
    }
}

static long loadRecord(struct Table *tab, struct Record *rec)
/* record must have been locked! */
{
    struct Project *pr = TableProject(tab);
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    long recsize;
    long err = OK;

    if (rec == NULL) {
        err = ERR_INTERNAL;
    } else if (rec->rec_Data != NULL) {
        err = OK;
    } else if ((recsize = TableMemSize(tab)) == 0) {
        err = OK;
    } else if (!rec->rec_OnDisk) {
        err = ERR_INTERNAL;
    } else if (tri == NULL) {
        err = ERR_INTERNAL;
    } else {
        size_t totalWarnings = 0;

        if (tri->tri_IsCustomFormat) {
            err = loadRecordCustomFormat(tab, rec, &totalWarnings);
        } else {
            err = loadRecordSQLite(tab, rec, &totalWarnings);
        }

        if (err != OK) {
            freeRecordData(tab, rec);
        } else {
            AddProjectRecordMem(pr, recordTotalMemSize(rec->rec_Data, tab));
            if (totalWarnings > 0) {
                ProjectWarnCodeset(pr);
            }
        }

        checkRecordMem(pr);
    }
    return err;
}

long WriteTableRecords(struct SQLiteHandle *handle, struct Table *tab,
        void (*wroteOneRecord)(void))
{
    struct List *recs = TableRecords(tab);
    struct Record *rec;
    long orderNum = 0;
    long version;
    long error = OK;

    if (recs == NULL) {
        return ERR_INTERNAL;
    }

    FORNODE(rec, recs) {
        RecordLock(rec);
        error = loadRecord(tab, rec);
        if (error == OK) {
            orderNum++;
            version = rec->rec_FileVersion;
            if (rec->rec_DataModified) {
                version = NextVersion(version);
            }
            error = SQLiteWriteRecord(handle, tab, FALSE, rec->rec_Num,
                    (double)orderNum, rec->rec_MatchFilter != 0,
                    version, rec->rec_Data, getdatainfo);
            if (wroteOneRecord != NULL) {
                wroteOneRecord();
            }
        }
        RecordUnlock(rec);
        if (error != OK) {
            break;
        }
    }
    return error;
}

static long swapRecord(struct SQLiteHandle *handle, struct Table *tab,
        struct Record *rec, double orderNum, bool freeData)
{
    if (rec->rec_DataModified) {
        long version = NextVersion(rec->rec_FileVersion);
        long error = OK;

	if (rec->rec_Data == NULL && TableMemSize(tab) > 0) {
            return ERR_INTERNAL;
        }

        error = ObtainProjectChangeLock(TableProject(tab));
        if (error != OK) {
            return error;
        }

        error = SQLiteWriteRecord(handle, tab, rec->rec_OnDisk != 0,
                rec->rec_Num, orderNum, rec->rec_MatchFilter != 0,
                version, rec->rec_Data, getdatainfo);
        if (error != OK) {
            return error;
        }

        rec->rec_FileVersion = version;
        rec->rec_OnDisk = TRUE;
        if (orderNum >= 0.0) {
            rec->rec_HeaderModified = FALSE;
        }
        rec->rec_DataModified = FALSE;
    } else if (orderNum >= 0.0) {
        long error = OK;

        error = ObtainProjectChangeLock(TableProject(tab));
        if (error != OK) {
            return error;
        }

        error = SQLiteWriteRecord(handle, tab, rec->rec_OnDisk != 0,
                rec->rec_Num, orderNum, rec->rec_MatchFilter != 0,
                rec->rec_FileVersion, NULL, NULL);
        if (error != OK) {
            return error;
        }

        rec->rec_OnDisk = TRUE;
        rec->rec_HeaderModified = FALSE;
    }

    if (freeData) {
        if (rec->rec_LockCount > 0) {
            return RECERR_LOCK;
        }
        freeRecordData(tab, rec);
    }

    return OK;
}

long SwapTableRecords(struct Table *tab, bool header, bool freeData,
        void (*swappedOneRecord)(void))
{
    struct Project *pr = TableProject(tab);
    struct SQLiteHandle *handle = ProjectSQLiteHandle(pr);
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct List *recs = TableRecords(tab);
    struct Record *rec, *lowerRec = NULL, *upperRec = NULL;
    double orderNum = -1.0, lowerOrderNum = 0.0, upperOrderNum = 0.0;
    long lowerDist = 0, upperDist = 0;
    long error = OK, err;

    if (recs == NULL) {
        return ERR_INTERNAL;
    }

    FORNODE(rec, recs) {
        lowerDist++;
        upperDist--;
        if (rec == upperRec) {
            lowerRec = NULL;
            upperRec = NULL;
        }
        orderNum = -1.0;
        if (header && rec->rec_HeaderModified) {
            if (lowerRec == NULL && error == OK) {
                lowerRec = PRED(rec); 
                lowerDist = 1;
                for (; PRED(lowerRec) != NULL; lowerRec = PRED(lowerRec)) {
                    if (!lowerRec->rec_HeaderModified) {
                        break;
                    }
                    lowerDist++;
                }
                if (PRED(lowerRec) == NULL) {
                    lowerOrderNum = 0.0;
                } else {
                    err = SQLiteGetOrderNum(handle, tab, lowerRec->rec_Num,
                            &lowerOrderNum);
                    if (err != OK) {
                        error = err;
                    }
                }
            }
            if (upperRec == NULL && error == OK) {
                upperRec = SUCC(rec);
                upperDist = 1;
                for (; SUCC(upperRec) != NULL; upperRec = SUCC(upperRec)) {
                    if (!upperRec->rec_HeaderModified) {
                        break;
                    }
                    upperDist++;
                }
                if (SUCC(upperRec) == NULL) {
                    upperOrderNum = lowerOrderNum + lowerDist + upperDist;
                } else {
                    err = SQLiteGetOrderNum(handle, tab, upperRec->rec_Num,
                            &upperOrderNum);
                    if (err != OK) {
                        error = err;
                    }
                }
            }
            if (error == OK) {
                orderNum  = lowerOrderNum + (upperOrderNum - lowerOrderNum) * 
                    lowerDist / (lowerDist + upperDist);
                updateMinRecordOrderNumResolution(tri,orderNum - lowerOrderNum);
                updateMinRecordOrderNumResolution(tri,upperOrderNum - orderNum);
            }
        }
        err = swapRecord(handle, tab, rec, orderNum, freeData);
        if (err != OK) {
            error = err;
        }
        if (swappedOneRecord != NULL) {
            swappedOneRecord();
        }
    }
    return error;
}

long SwapProjectRecords(struct Project *pr)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    static bool inSwapRecord = FALSE;
    long error = OK, err;

    if (inSwapRecord) {
        return OK;
    }
    inSwapRecord = TRUE;

    GuiSetSleep(TRUE);
    if (tabs != NULL) {
        FORNODE(tab, tabs) {
            err = SwapTableRecords(tab, FALSE, TRUE, NULL);
            if (err != OK) {
                error = err;
            }
        }
    }
    GuiSetSleep(FALSE);

    inSwapRecord = FALSE;

    return error;
}

void SwapRecords(void)
{
    struct Project *pr;

    GuiSetSleep(TRUE);
    FORNODE(pr, &ProjectList) {
        SwapProjectRecords(pr);
    }
    GuiSetSleep(FALSE);
}

#define FieldRecordInfoNum(fri) NodeNum(fri)

void FreeFieldRecordInfo(struct FieldRecordInfo *fri)
{
    if (fri != NULL) {
        struct Table *tab = FieldTable(fri->fri_Field);
        struct MemPool *memPool = TableMemPool(tab);

        REMOVE(fri);
        PoolFreeMem(memPool, fri, sizeof(struct FieldRecordInfo));
    }
}

long DeleteFieldRecordInfo(struct FieldRecordInfo *fri)
{
    if (fri != NULL) {
        struct FieldRecordInfo *succ = SUCC(fri);
        struct Field *field =fri->fri_Field;
        struct Table *tab = FieldTable(field);
        struct Project *pr = TableProject(tab);
        struct SQLiteHandle *handle = ProjectSQLiteHandle(pr);
        struct List *fields = TableFields(tab);
        long size = FieldMemSize(field);
        long error = OK;

        if (fields == NULL) {
            return ERR_INTERNAL;
        }

        if (size > 0) {
            error = SwapTableRecords(tab, FALSE, TRUE, NULL);
            if (error) {
                return error;
            }
            error = SQLiteDeleteField(handle, field);
            if (error) {
                return error;
            }
        }

        /* Free record info and adjust record infos of succeeding fields. */
        FreeFieldRecordInfo(fri);

        for (fri = succ; SUCC(fri); fri = SUCC(fri)) {
            fri->fri_Pos -= size;
        }
    }
    return OK;
}

struct FieldRecordInfo *AllocFieldRecordInfo(struct Field *field)
{
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = ProjectMemPool(pr);
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct FieldRecordInfo *fri = NULL;

    if (tri != NULL) {
        fri = PoolAllocMem(memPool, sizeof(struct FieldRecordInfo));
        if (fri != NULL) {
            fri->fri_Field = field;
            fri->fri_Pos = TableMemSize(tab);
            ADDTAIL(&tri->tri_FieldRecordInfos, fri);
        }
    }

    return fri;
}

long NewFieldRecordInfo(struct Field *field, struct FieldRecordInfo **frip)
{
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    struct SQLiteHandle *handle = ProjectSQLiteHandle(pr);
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct FieldRecordInfo *fri = NULL;
    struct DataInfo *init = FieldDataInfo(field);
    short size = FieldMemSize(field);
    long error = OK;

    if (frip != NULL) {
        *frip = NULL;
    }

    if (tri == NULL || init == NULL) {
        return ERR_INTERNAL;
    }

    fri = AllocFieldRecordInfo(field);
    if (fri == NULL) {
	return ERR_MEM;
    }

    if (size > 0) {
        error = SwapTableRecords(tab, FALSE, TRUE, NULL);
        if (error == OK) {
            error = SQLiteNewField(handle, field);
        }
    }

    if (error) {
        FreeFieldRecordInfo(fri);
        fri = NULL;
    }

    if (frip) {
        *frip = fri;
    }
    return error;
}

long CanNewFieldRecordInfo(struct Table *tab, short type)
{
    long error = OK;

    if (type == TYPE_UNKNOWN || FieldTypeMemSize(type) > 0) {
        error = SwapTableRecords(tab, FALSE, TRUE, NULL);
    }
    return error;
}

long ChangeStringRecordInfo(struct FieldRecordInfo *fri, long newstrlen)
{
    return OK;
}

long RecordDataInfo(struct Record *rec, struct Field *field, 
	struct DataInfo **dip, struct DataInfo *di_buf)
{
    struct DataInfo *di = NULL;
    long error = OK;

    if (rec == NULL) {
	di = FieldDataInfo(field);
    } else if (rec->rec_LockCount == 0) {
	error = RECERR_LOCK;
    } else {
	error = loadRecord(FieldTable(field), rec);
	if (error == OK && getdatainfo(rec->rec_Data, field, di_buf)) {
	    di = di_buf;
	}
    }
    if (di == NULL && error == OK) {
	error = ERR_INTERNAL;
    }

    if (dip != NULL) {
        *dip = di;
    }
    return error;
}

bool RecordMatchFilter(struct Record *rec)
{
    bool match = FALSE;

    if (rec) {
        match = rec->rec_MatchFilter;
    }
    return match;
}

void SetRecordMatchFilter(struct Record *rec, bool match)
{
    match = match? 1 : 0;
    if (rec != NULL && match != rec->rec_MatchFilter) {
        rec->rec_MatchFilter = match;
        rec->rec_HeaderModified = TRUE;
    }
}

bool RecordIsSorted(struct Record *rec)
{
    bool sorted = FALSE;

    if (rec != NULL) {
        sorted = rec->rec_IsSorted;
    }
    return sorted;
}

void SetRecordIsSorted(struct Record *rec, bool sorted)
{
    sorted = sorted? 1 : 0;
    if (rec != NULL && sorted != rec->rec_IsSorted) {
        rec->rec_IsSorted = sorted;
        rec->rec_HeaderModified = TRUE;
    }
}

bool RecordImported(struct Record *rec)
{
    bool imported = FALSE;

    if (rec != NULL) {
        imported = rec->rec_IsImported;
    }
    return imported;
}

void SetRecordImported(struct Record *rec, bool imported)
{
    if (rec != NULL) {
        imported = imported? 1 : 0;
        rec->rec_IsImported = imported;
    }
}

/*
** Next, Prev, First, Last - Record
*/

static bool recordmatch(struct Record *rec, bool filter,
        struct Field *refField, struct Record *refRec)
{
    if (filter && !rec->rec_MatchFilter) {
        return FALSE;
    }

    if (refField != NULL) {
        struct DataInfo *di = NULL, di_buf;
        bool match = TRUE;

        RecordLock(rec);
        RecordDataInfo(rec, refField, &di, &di_buf);
        if (di != NULL) {       /* treat errors as match success */
            if (di->di_Flags & DIF_NIL) {
                match = (refRec == NULL);
            } else {
                match = (refRec == di->di_Contents.rec);
            }
        }
        RecordUnlock(rec);
        return match;
    }

    return TRUE;
}

static struct Record *forwardrecord(struct Record *rec, bool filter,
        struct Field *refField, struct Record *refRec)
{
    if (rec != NULL) {
        for(; SUCC(rec); rec = SUCC(rec)) {
            if (recordmatch(rec, filter, refField, refRec)) {
                return rec;
            }
        }
    }
   return NULL;
}

static struct Record *backwardrecord(struct Record *rec, bool filter,
        struct Field *refField, struct Record *refRec)
{
    if (rec != NULL) {
        for(; PRED(rec); rec = PRED(rec)) {
            if (recordmatch(rec, filter, refField, refRec)) {
                return rec;
            }
        }
    }
   return NULL;
}

struct Record *NextRecordWithReference(struct Record *rec, bool filter,
        struct Field *refField, struct Record *refRec)
{
    return (rec? forwardrecord(SUCC(rec), filter, refField, refRec) : NULL);
}

struct Record *PrevRecordWithReference(struct Record *rec, bool filter,
        struct Field *refField, struct Record *refRec)
{
    return(rec? backwardrecord(PRED(rec), filter, refField, refRec) : NULL);
}

struct Record *FirstRecordWithReference(struct Table *tab, bool filter,
        struct Field *refField, struct Record *refRec)
{
    struct List *recs = TableRecords(tab);
    struct Record *rec = NULL;

    if (recs != NULL) {
        rec = forwardrecord((struct Record *)HEAD(recs), filter,
                refField, refRec);
    }
    return rec;
}

struct Record *LastRecordWithReference(struct Table *tab, bool filter,
        struct Field *refField, struct Record *refRec)
{
    struct List *recs = TableRecords(tab);
    struct Record *rec = NULL;

    if (recs != NULL) {
        rec = backwardrecord((struct Record *)TAILPRED(recs), filter,
                refField, refRec);
    }
    return rec;
}

struct Record *NextRecord(struct Record *rec, bool filter)
{
    return NextRecordWithReference(rec, filter, NULL, NULL);
}

struct Record *PrevRecord(struct Record *rec, bool filter)
{
    return PrevRecordWithReference(rec, filter, NULL, NULL);
}

struct Record *FirstRecord(struct Table *tab, bool filter)
{
    return FirstRecordWithReference(tab, filter, NULL, NULL);
}

struct Record *LastRecord(struct Table *tab, bool filter)
{
    return LastRecordWithReference(tab, filter, NULL, NULL);
}

struct Record *RecordNth(struct Table *tab, bool filter, long num)
{
    struct List *recs = TableRecords(tab);
    struct Record *rec;

    if(recs == NULL || num < 1)
        return(NULL);

    if(!filter)
        return(FindNodeNum(recs, num));

    for(rec = FirstRecord(tab, TRUE); rec != NULL; rec = NextRecord(rec, TRUE))
        if(--num == 0)
            return(rec);

    return(NULL);
}

void RecordNumAll(struct Table *tab, struct Record *rec_num, 
        bool filter, ulong *num, ulong *all)
{
    ulong n = 0, a = 0;
    struct List *recs = TableRecords(tab);
    struct Record *rec;

    if (recs) {
        if (filter) {
            FORNODE(rec, recs) {
                if (rec->rec_MatchFilter) {
                    a++;
                    if (rec == rec_num) {
                        n = a;
                    }
                }
            }
        } else {
            FORNODE(rec, recs) {
                a++;
                if (rec == rec_num) {
                    n = a;
                }
            }
        }
    }
    if (num) {
        *num = n;
    }
    if (all) {
        *all = a;
    }
}

struct Table *RecordFindTable(struct Record *rec, struct Table *guess,
	struct Project *pr)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    if (rec == NULL) {
	return guess;		/* no better guess possible */
    } else if (FindNodePtr(TableRecords(guess), rec)) {
	return guess;		/* found */
    } else if (tabs != NULL) {
	FORNODE (tab, tabs) {
	    if (tab != guess && FindNodePtr(TableRecords(tab), rec)) {
		return tab;	/* found */
	    }
	}
    }
    return NULL;	/* not found */
}

/*
** Allocate & Free Record
*/

void FreeRecord(struct Table *tab, struct Record *rec, bool quick)
{
    if (!quick && tab && rec) {
        struct MemPool *memPool = TableMemPool(tab);

        freeRecordData(tab, rec);
	PoolFreeMem(memPool, rec, sizeof(struct Record));
    }
}

static struct Record *AllocRecord(struct Table *tab)
{
    struct Project *pr = TableProject(tab);
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct MemPool *memPool = TableMemPool(tab);
    struct MemPool *dataPool = ProjectDataMemPool(pr);
    struct Record *rec;
    long size = TableMemSize(tab);

    if (tri == NULL || tri->tri_NextRecordNum == 0) {
        return NULL;
    }

    rec = PoolAllocMemZero(memPool, sizeof(struct Record));
    if (rec == NULL) {
	return NULL;
    }
    rec->rec_Num = tri->tri_NextRecordNum++;
    rec->rec_MatchFilter = TRUE;
    rec->rec_HeaderModified = TRUE;
    rec->rec_DataModified = TRUE;
    rec->rec_CustomFormat.filePos = -1;

    if (size > 0) {
	rec->rec_Data = PoolAllocMemZero(dataPool, size);
	if (rec->rec_Data == NULL) {
	    FreeRecord(tab, rec, FALSE);
	    return NULL;
	}
    }

    AddProjectRecordMem(pr, size);
    checkRecordMem(pr);

    return rec;
}

static long CopyDataInfo(struct Field *field, 
	struct DataInfo *di_src, char *target_data)
{
    if (di_src != NULL) {
	switch (FieldType(field)) {
            case TYPE_STRING:
            {
                struct Table *tab = FieldTable(field);
                struct Project *pr = TableProject(tab);
                struct MemPool *dataPool = ProjectDataMemPool(pr);
                char *str = NULL, *old;
		struct DataInfo di;

                if (di_src->di_Contents.string) {
                    str = AllocString(dataPool, di_src->di_Contents.string);
                    if (str == NULL) {
                        return ERR_MEM;
                    }
                }

		getdatainfo(target_data, field, &di);
		old = di.di_Contents.string;
                if (old != NULL) {
                    AddProjectRecordMem(pr, -(long)(strlen(old) + 1));
                    FreeString(dataPool, old);
                }
		di = *di_src;
		di.di_Contents.string = str;
		setdatainfo(target_data, field, &di);

                if (str != NULL) {
                    AddProjectRecordMem(pr, strlen(str) + 1);
                }
                checkRecordMem(pr);
                break;
            }

            case TYPE_MEMO:
            {
                struct Table *tab = FieldTable(field);
                struct Project *pr = TableProject(tab);
                struct MemPool *dataPool = ProjectDataMemPool(pr);
                char *mi = NULL, *old;
		struct DataInfo di;

                if (di_src->di_Contents.mi) {
                    mi = AllocString(dataPool, di_src->di_Contents.mi);
                    if (mi == NULL) {
                        return ERR_MEM;
                    }
                }

		getdatainfo(target_data, field, &di);
		old = di.di_Contents.mi;
                if (old != NULL) {
                    AddProjectRecordMem(pr, -(long)(strlen(old) + 1));
                    FreeString(dataPool, old);
                }
		di = *di_src;
		di.di_Contents.mi = mi;
		setdatainfo(target_data, field, &di);

                if (mi != NULL) {
                    AddProjectRecordMem(pr, strlen(mi) + 1);
                }
                checkRecordMem(pr);
                break;
            }

            default:
		setdatainfo(target_data, field, di_src);
                break;
        }
    }
    return(OK);
}

long CopyRecord(struct Table *tab, struct Record *rec, struct Record *init)
{
    struct List *fields = TableFields(tab);
    struct Field *field;
    struct DataInfo *initdi, di_buf;
    long error = OK;

    if( rec == NULL) {
        return RECERR_INITIALRECORD;
    }

    RecordLock(rec);
    RecordLock(init);
    error = loadRecord(tab, rec);
    if (error == OK && init) {
	error = loadRecord(tab, init);
    }
    if (error == OK) {
	if (fields) {
	    FORNODE (field, fields) {
		if (init == NULL) {
		    initdi = FieldDataInfo(field);
		} else if (!getdatainfo(init->rec_Data, field, &di_buf)) {
	    	    error = ERR_INTERNAL;
    		    break;
		} else {
		    initdi = &di_buf;
		}
		error = CopyDataInfo(field, initdi, rec->rec_Data);
		if (error) {
		    break;
		}
		rec->rec_IsSorted = FALSE;
		rec->rec_HeaderModified = TRUE;
		rec->rec_DataModified = TRUE;
	    }
	}
    }
    RecordUnlock(init);
    RecordUnlock(rec);

    return error;
}

/*
** NewRecord
*/

long NewRecord(struct Table *tab, struct Record *init, struct Record *pred,
        bool addToGui, struct Record **rec)
{
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct Record *new;
    long error = OK;

    if (tab == NULL) {
        return ERR_NOTABLE;
    } else if (TableExactlyOneRecord(tab) && TableNumRecords(tab) >= 1) {
        return RECERR_EXACTLYONERECORD;
    }
    if (tri == NULL) {
        return ERR_INTERNAL;
    }

    new = AllocRecord(tab);
    if (new == NULL) {
        return ERR_MEM;
    }
    error = CopyRecord(tab, new, init);
    if (error) {
        FreeRecord(tab, new, FALSE);
        return error;
    }
    if (!IdRecordMapReserveEntry(TableMemPool(tab), tri->tri_IdRecordMap)) {
        FreeRecord(tab, new, FALSE);
        return ERR_MEM;
    }
    if (!IdRecordMapPushEntry(tri->tri_IdRecordMap, new)) {
        FreeRecord(tab, new, FALSE);
        return ERR_INTERNAL;
    }

    AddRecord(tab, new, pred);
    LogNewRecord(tab, new, init);

    /* 
    ** Set reference records when a reference filter is set.
    ** Skip when we have an init record since the existing
    ** references should prevail in this case.
    */
    if (init == NULL) {
        struct List *fields = TableFields(tab);
        struct Field *field;
        union Data data;

        FORNODE (field, fields) {
            if (FieldType(field) != TYPE_REFERENCE) {
                continue;
            }
            if (!FieldReferenceFilter(field)) {
                continue;
            }
            data.rec = TableGuiRecord(FieldReferenceTable(field));
            error = RecordStoreData(new, field, &data, data.rec == NULL);
            if (error != OK) {
                break;
            }
        }
    }

    if (addToGui) {
        GuiAddRecord(tab, new);
    }

    if (rec != NULL) {
        *rec = new;
    }
    return error;
}

static void ClearReferenceRecords(struct Table *reftab, struct Record *refrec)
{
    struct Project *pr = TableProject(reftab);
    struct List *tabs = ProjectTables(pr), *fields, *recs;
    struct Table *tab;
    struct Field *field;
    struct Record *rec;
    struct DataInfo di;

    if (tabs) {
        FORNODE (tab, tabs) {
            fields = TableFields(tab);
            recs = TableRecords(tab);
            if (fields == NULL || recs == NULL) {
                continue;
	    }
            FORNODE (field, fields) {
                if (FieldType(field) != TYPE_REFERENCE) {
                    continue;
		}
                if (FieldReferenceTable(field) != reftab) {
                    continue;
		}
                FORNODE (rec, recs) {
                    RecordLock(rec);
                    if (rec->rec_Data && getdatainfo(rec->rec_Data, field,&di)){
			if (di.di_Contents.rec == refrec) {
                            di.di_Contents.rec = NULL;
                            di.di_Flags |= DIF_NIL;
			    setdatainfo(rec->rec_Data, field, &di);
                            rec->rec_IsSorted = FALSE;
                            rec->rec_HeaderModified = TRUE;
                            rec->rec_DataModified = TRUE;
			}
                    }
                    RecordUnlock(rec);
                }
            }
        }
    }
}

/*
** DeleteRecord
*/

long AttemptDeleteRecord(struct Table *tab, struct Record *rec)
{
    if(tab == NULL)
        return(ERR_NOTABLE);
    if(rec == NULL)
        return(RECERR_DELETEINITIALRECORD);
    if(TableExactlyOneRecord(tab) && TableNumRecords(tab) <= 1)
        return(RECERR_EXACTLYONERECORD);
    return(OK);
}

long DeleteRecord(struct Table *tab, struct Record *rec)
{
    struct Project *pr = TableProject(tab);
    struct SQLiteHandle *handle = ProjectSQLiteHandle(pr);
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct IdRecordMap *map = tri != NULL? tri->tri_IdRecordMap : NULL;
    struct Record *next, *prev;
    long error;

    error = AttemptDeleteRecord(tab, rec);
    if (error != OK) {
        return error;
    }

    LogDeleteRecord(tab, rec);  /* log before deleting from sqlite */

    error = SQLiteDeleteRecord(handle, tab, rec->rec_Num);
    if (error != OK) {
        return error;
    }

    ProgramRemRecord(tab, rec);
    ClearReferenceRecords(tab, rec);    /* sets ptrs to this record to zero */
    next = NextRecord(rec, FALSE);
    prev = PrevRecord(rec, FALSE);
    RemRecord(tab, rec);
    GuiRemRecord(tab, rec, next, prev, TRUE); /* rem ptrs in Gui objects */
    IdRecordMapRemove(map, rec->rec_Num);
    FreeRecord(tab, rec, FALSE);

    GuiDisplayPanel(tab);          /* GuiRemRecord might not update panel */

    return(OK);
}

long DeleteRecordsTable(struct Table *tab, bool filter, bool import)
{
    struct Project *pr = TableProject(tab);
    struct SQLiteHandle *handle = ProjectSQLiteHandle(pr);
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct IdRecordMap *map = tri != NULL? tri->tri_IdRecordMap : NULL;
    struct List *recs = TableRecords(tab);
    struct Record *rec, *succ, *next, *prev;
    struct List list;
    long error = OK;

    if (recs == NULL) {
        return ERR_NOTABLE;
    }

    NewList(&list);
    SAFEFORNODE(rec, succ, recs) {
        if (filter && !RecordMatchFilter(rec)) {
            continue;
        }
        if (import && !RecordImported(rec)) {
            continue;
        }
        error = AttemptDeleteRecord(tab, rec);
        if (error != OK) {
            break;
        }

        LogDeleteRecord(tab, rec);  /* log before deleting from sqlite */

        error = SQLiteDeleteRecord(handle, tab, rec->rec_Num);
        if (error != OK) {
            break;
        }

        ProgramRemRecord(tab, rec);
        next = NextRecord(rec, FALSE);
        prev = PrevRecord(rec, FALSE);
        RemRecord(tab, rec);
        GuiRemRecord(tab, rec, next, prev, FALSE); /* rem ptrs in Gui objects */
        ADDTAIL(&list, rec);
    }
    while ((rec = (struct Record *)REMHEAD(&list)) != NULL) {
        ClearReferenceRecords(tab, rec);  /* sets ptrs to this record to zero */
        IdRecordMapRemove(map, rec->rec_Num);
        FreeRecord(tab, rec, FALSE);
    }
  
    GuiShowRecord(tab, TableGuiRecord(tab));   /* update display */
    
    return error;
}

long DeleteRecordsProject(struct Project *pr)
{
    struct SQLiteHandle *handle = ProjectSQLiteHandle(pr);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    struct TableRecordInfo *tri;
    struct IdRecordMap *map;
    struct List *recs;
    struct Record *rec, *succ;
    long error = OK;

    if (tabs != NULL) {
        FORNODE(tab, tabs) {
            if (TableExactlyOneRecord(tab)) {
                continue;
            }
            tri = TableRecordInfo(tab);
            map = tri != NULL? tri->tri_IdRecordMap : NULL;
            recs = TableRecords(tab);
            if (recs != NULL) {
                SAFEFORNODE(rec, succ, recs) {
                    RemRecord(tab, rec);
                    IdRecordMapRemove(map, rec->rec_Num);
                    FreeRecord(tab, rec, FALSE);
                }
            }
            SetTableGuiRecord(tab, FirstRecord(tab, TRUE));
            error = SQLiteDeleteRecords(handle, tab);
            if (error != OK) {
                break;
            }
        }
    }
    return error;
}

long RecordStoreData(struct Record *rec, struct Field *field,
	union Data *data, bool nil)
{
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = ProjectDataMemPool(pr);
    struct DataInfo di, olddi;
    struct FieldRecordInfo *fri = FieldRecordInfo(field);
    short type = FieldType(field);
    long error = OK;
    bool changed = FALSE;
    bool check = FALSE;

    if (rec == NULL) {
        return RECERR_INITIALRECORD;
    }
    if (fri == NULL || FieldMemSize(field) == 0) {
        return ERR_INTERNAL;
    }

    RecordLock(rec);
    error = loadRecord(tab, rec);
    if (error == OK && getdatainfo(rec->rec_Data, field, &di)) {
        olddi = di;
	switch (type) {
	    case TYPE_INTEGER:
	    	di.di_Contents.integer = data->integer;
	    	break;
	    case TYPE_REAL:
	    	di.di_Contents.real = data->real;
                break;
            case TYPE_BOOL:
                break;
            case TYPE_STRING:
            {
		const char *s = data->string;
                short max = FieldStringMaxlen(field);
                char *new = NULL;

                if (nil) {
                    error = RECERR_NILVALUE;
		} else if (s != NULL) {
                    new = AllocStringN(memPool, s, 
                            CodesetNCharsToNBytes(s, max));
                    if (new == NULL) {
                        error = ERR_MEM;
                    }
                }
                if (error == OK) {
                    di.di_Contents.string = new;
                    if (new) {
                        AddProjectRecordMem(pr, strlen(new) + 1);
                    }
		    check = TRUE;
                }
                break;
            }
            case TYPE_MEMO:
            {
		const char *s = data->mi;
                char *new = NULL;

                if (nil) {
                    error = RECERR_NILVALUE;
		} else if (s != NULL) {
                    new = AllocString(memPool, s);
                    if (new == NULL) {
                        error = ERR_MEM;
                    }
                }
                if (error == OK) {
                    di.di_Contents.mi = new;
                    if (new) {
                        AddProjectRecordMem(pr, strlen(new) + 1);
                    }
		    check = TRUE;
                }
                break;
            }
            case TYPE_DATE:
                di.di_Contents.date = data->date;
                break;
            case TYPE_TIME:
                di.di_Contents.time = data->time;
                break;
            case TYPE_CHOICE:
                if (nil) {
                    error = RECERR_NILVALUE;
                } else {
                    di.di_Contents.choice = data->choice;
		}
                break;
            case TYPE_REFERENCE:
                if (data->rec == NULL) {
                    nil = TRUE;
                }
                di.di_Contents.rec = nil? NULL : data->rec;
                break;
        }
        if (error == OK) {
            if (nil) {
                di.di_Flags |= DIF_NIL;
            } else {
                di.di_Flags &= ~DIF_NIL;
            }
            setdatainfo(rec->rec_Data, field, &di);
            rec->rec_IsSorted = FALSE;
            rec->rec_HeaderModified = TRUE;
            rec->rec_DataModified = TRUE;
            changed = TRUE;
        }
    }
    RecordUnlock(rec);

    if (changed) {
        LogStoreData(rec, field, &olddi, &di);
        switch(type) {
            case TYPE_STRING:
                if (olddi.di_Contents.string) {
                    AddProjectRecordMem(pr, 
                            -(long)(strlen(olddi.di_Contents.string) + 1));
                    FreeString(memPool, olddi.di_Contents.string);
                }
                break;
            case TYPE_MEMO:
                if (olddi.di_Contents.mi) {
                    AddProjectRecordMem(pr, 
                            -(long)(strlen(olddi.di_Contents.mi) + 1));
                    FreeString(memPool, olddi.di_Contents.mi);
                }
                break;
        }
    }

    if (check) {
	checkRecordMem(pr);
    }

    GuiUpdateField(field, rec);

    return error;
}

#ifdef CHECK_RECORD_HEADERS
bool CheckRecordHeaders(const char *message)
{
    struct Project *pr;

    FORNODE(pr, &ProjectList) {
	struct List *tabs = ProjectTables(pr);
	struct Table *tab;
	if (tabs) {
	    FORNODE(tab, tabs) {
		if (CheckList(TableRecords(tab)) != 0) {
		    Ask(AT_ERROR, GCS(MSG_StarOk), 
			    "%s: list of record headers corrupt.\n"
			    "Corrputed node at address %p.\n"
			    "Please do integrity check now!", message,
			    CorruptedNode);
		    return FALSE;
		}
	    }
	}
    }
    /* DPRINTF(("%s: record headers ok!\n", message)); */
    return TRUE;
}
#endif

/*
** Load & Save of records
*/

void PrepareRecordUpdate(struct Project *pr)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    struct TableRecordInfo *tri;
    struct List *recs;
    struct Record *rec;

    if (tabs == NULL) {
        return;
    }

    FORNODE(tab, tabs) {
        tri = TableRecordInfo(tab);
        if (tri != NULL) {
            recs = TableRecords(tab);
            if (recs) {
                while (SUCC(HEAD(recs)) != NULL) {
                    rec = (struct Record *)HEAD(recs);
                    RemRecord(tab, rec);
                    ADDTAIL(&tri->tri_UpdateRecords, rec);
                }
            }
        }
    }
}

long ReadRecord(struct Table *tab, const struct ReadRecordData *data,
        bool update)
{
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct Record *rec = NULL;

    if (tri == NULL || data == NULL) {
        return ERR_INTERNAL;
    }

    if (update) {
        rec = FindRecordInternalNumList(&tri->tri_UpdateRecords, data->rrd_ID);
        if (rec != NULL) {
            REMOVE(rec);
            if (rec->rec_Data != NULL) {
                if (!rec->rec_OnDisk || rec->rec_DataModified ||
                        rec->rec_FileVersion != data->rrd_Version) {
                    ADDTAIL(&tri->tri_UpdateRecords, rec);
                    rec = NULL;
                }
            }
        }
    }

    if (rec == NULL) {
        rec = PoolAllocMemZero(TableMemPool(tab), sizeof(struct Record));
        if (rec == NULL) {
            return ERR_MEM;
        }

        rec->rec_Num = data->rrd_ID;
    }
    rec->rec_FileVersion = data->rrd_Version;
    rec->rec_OnDisk = TRUE;
    rec->rec_MatchFilter = (data->rrd_MatchFilter != 0);
    rec->rec_IsSorted = (data->rrd_IsSorted != 0);

    if (rec->rec_Num >= tri->tri_NextRecordNum) {
        tri->tri_NextRecordNum = rec->rec_Num + 1;
    }
    AddRecord(tab, rec, ADDRECORD_TAIL);

    return OK;
}

long AfterReadRecords(struct Table *tab)
{
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct IdRecordMap *map = tri != NULL? tri->tri_IdRecordMap : NULL;

    if (map == NULL) {
        return ERR_INTERNAL;
    }
    if (!IdRecordMapSet(TableMemPool(tab), map, TableRecords(tab))) {
        return ERR_MEM;
    }
    return OK;
}

void CompleteRecordUpdate(struct Project *pr)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    struct TableRecordInfo *tri;
    struct FieldRecordInfo *fri;
    struct Record *rec;
    struct Field *field;
    struct List *recs;
    struct Record *refrec;
    struct DataInfo di;

    if (tabs == NULL) {
        return;
    }

    /* Mark all remaining update records as not on disk */
    FORNODE (tab, tabs) {
        tri = TableRecordInfo(tab);
        if (tri != NULL) {
            FORNODE(rec, &tri->tri_UpdateRecords) {
                rec->rec_OnDisk = FALSE;
            }
        }
    }

    /* Update all records that have references to not-on-disk records */
    FORNODE (tab, tabs) {
        rec = TableGuiRecord(tab);
        if (rec != NULL && !rec->rec_OnDisk) {
            SetTableGuiRecord(tab, FindRecordInternalNum(tab, rec->rec_Num));
        }
        rec = TableProgramRecord(tab);
        if (rec != NULL && !rec->rec_OnDisk) {
            SetTableProgramRecord(tab, FindRecordInternalNum(tab,rec->rec_Num));
        }

        tri = TableRecordInfo(tab);
        recs = TableRecords(tab);
        if (tri == NULL || recs == NULL) {
            continue;
        }
        FORNODE(fri, &tri->tri_FieldRecordInfos) {
            field = fri->fri_Field;
            if (FieldType(field) != TYPE_REFERENCE) {
                continue;
            }
            FORNODE(rec, recs) {
                if (rec->rec_Data == NULL) {
                    continue;
                }
                if (!getdatainfo(rec->rec_Data, field, &di)) {
                    continue;
                }
                refrec = di.di_Contents.rec;
                if ((di.di_Flags & DIF_NIL) || refrec == NULL) {
                    continue;
                }
                if (refrec->rec_OnDisk) {
                    continue;
                }
                di.di_Contents.rec = FindRecordInternalNum(
                    FieldReferenceTable(field), refrec->rec_Num);
                setdatainfo(rec->rec_Data, field, &di);
            }
        }
    }

    /* Free all remaining update records */
    FORNODE (tab, tabs) {
        tri = TableRecordInfo(tab);
        if (tri != NULL) {
            while ((rec = (struct Record *)REMHEAD(&tri->tri_UpdateRecords))
                    != NULL) {
                FreeRecord(tab, rec, FALSE);
            }
        }
    }
}

void RecordAfterReorgUpdate(struct Project *pr)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    if (tabs) {
        FORNODE(tab, tabs) {
            struct TableRecordInfo *tri = TableRecordInfo(tab);
            struct List *recs = TableRecords(tab);
            struct Record *rec;

            if (recs == NULL || tri == NULL) {
                continue;
            }

            ClearTableRecordInfo(tri);
            tri->tri_IsCustomFormat = FALSE;
            tri->tri_NextRecordNum = 1;
            FORNODE(rec, recs) {
                rec->rec_Num = tri->tri_NextRecordNum++;
                rec->rec_OnDisk = TRUE;
                rec->rec_HeaderModified = FALSE;
                rec->rec_DataModified = FALSE;
                rec->rec_FileVersion = 1;
            }
            IdRecordMapUpdateIds(tri->tri_IdRecordMap);
        }
    }
}

/* Custom format load */

#define RECFLAG_MATCHFILTER (1<<0)
#define RECFLAG_ISSORTED (1<<1)
#define RECFLAG_IMPORTED (1<<2)

#define RECFLAG_ONDISK (1<<0)
#define RECFLAG_MODIFIED (1<<1)

#define ISGUIRECORD(bitfield) (((bitfield) & 0x8000)? TRUE : FALSE)
#define FLAGS(bitfield) ((bitfield) & 0x7fff)

#define RECSVERSION 10                   /* 8 -> new di_Flags */
#define MIN_RECSVERSION 6

static struct
{
    short rh_Version;		/* version number */
    long rh_TabNum;             /* table number */
    long rh_NumRecs;		/* number of records */
    long rh_NumFree;            /* number of free list entries */
    long rh_NumRecChanges;      /* number of record changes */
    long rh_Unused1;
    long rh_Unused2;
    long rh_EofPos;
} rh;

static struct                        /* record header */
{
  ushort rd_BitField;                /* IsGuiRecord plus flags */
  ulong rd_Num;                      /* record number */
  long rd_FileVersion;               /* record version */
  long rd_FilePos;                   /* file position */
  long rd_FileTotalSize;
} rd;

static struct                           /* free entry */
{
    long fd_Pos;
    long fd_Size;
} fd;

static long loadRecChangeCustomFormat(FILE *file, struct MemPool *memPool,
        bool oldFormat, struct CustomFormatRecChange **prc)
{
    long i, numfields, size, num, error = OK;
    size_t warnings = 0;                /* currently ignored */
    struct CustomFormatRecChange *rc;

    *prc = NULL;

    numfields = FGetLong(file);
    size = sizeof(struct CustomFormatRecChange) + 
        numfields*sizeof(struct CustomFormatRecChangeFieldInfo);
    rc = PoolAllocMem(memPool, size);
    if (rc == NULL) {
	return ERR_MEM;
    }

    rc->rc_NumFields = numfields;
    rc->rc_Id = FGetLong(file);
    rc->rc_Op = FGetShort(file);

    switch (rc->rc_Op) {
	case OP_NEW:
	{
	    struct DataInfo *di = &rc->rc_Contents.new.di;

	    rc->rc_Contents.new.type = FGetShort(file);
            di->di_Flags = FGetChar(file);
            di->di_Private = FGetChar(file);
            if (oldFormat) {
                di->di_Flags = ((di->di_Private &  1)? DIF_NIL : 0);
                di->di_Private = 0;
            }
	    switch (rc->rc_Contents.new.type)
	    {			
        	case TYPE_INTEGER:
		    di->di_Contents.integer = FGetLong(file);
		    break;
		
		case TYPE_REAL:
		    di->di_Contents.real = FGetDouble(file);
		    break;

		case TYPE_BOOL:
		    break;

		case TYPE_STRING:
                    di->di_Contents.string = ReadStringConvert(file,
                            (di->di_Private & DIP_DISKUTF8) != 0, 
                            memPool, &error, &warnings);
		    break;

		case TYPE_MEMO:	
		    fgetc(file);	/* always empty */
		    di->di_Contents.mi = NULL;
		    break;

		case TYPE_DATE:	        
		    di->di_Contents.date = FGetLong(file);
		    break;

		case TYPE_TIME:
		    di->di_Contents.time = FGetLong(file);
		    break;

		case TYPE_CHOICE:	
		    di->di_Contents.choice = FGetShort(file);
		    break;

		case TYPE_REFERENCE:   
		    num = FGetLong(file);
		    di->di_Contents.rec = NULL;
                    di->di_Flags |= DIF_NIL;
		    break;

		case TYPE_VIRTUAL:
		case TYPE_BUTTON:
		    break;

                default:
                    error = DISKERR_INVALID;
                    break;
	    }
	    break;
	}

	case OP_DELETE:
            num = FGetLong(file);
            if (num < 0 || num >= rc->rc_NumFields) {
                error = DISKERR_INVALID;
            }
	    rc->rc_Contents.delete.num = num;
	    break;

	case OP_NONE:		/* used to be TRUNCATE */
            FGetLong(file);	/* num */
            FGetLong(file);	/* kill */
	    break;

        default:
            error = DISKERR_INVALID;
            break;
    }

    if (error) {
        FreeCustomFormatRecChange(memPool, rc);
        return error;
    }

    for (i=0; i < rc->rc_NumFields; i++) {
	rc->rc_FieldInfos[i].ai_Type = FGetShort(file);
    }

    *prc = rc;
    return OK;
}

static int fri_cmp(void *p1, void *p2)
{
    struct FieldRecordInfo *fri1 = p1, *fri2 = p2;

    if(fri1->fri_Pos < fri2->fri_Pos)
        return(-1);
    else if(fri1->fri_Pos > fri2->fri_Pos)
        return(1);
    return(0);
}

long LoadRecordsCustomFormat(struct Project *pr, FILE *file,
        const char **errstr)
{
    long i, error = OK;
    struct List *tabs = ProjectTables(pr);
    struct Table *tab, *succ;
    struct TableRecordInfo *tri;
    struct IdRecordMap *map;
    struct FieldRecordInfo *fri;
    struct MemPool *memPool = ProjectMemPool(pr);
    ulong pos = 0;

    if (pr == NULL || file == NULL || tabs == NULL) {
    	return ERR_INTERNAL;
    }

    SAFEFORNODE (tab, succ, tabs) {
	rh.rh_Version = FGetShort(file);
        if (rh.rh_Version < MIN_RECSVERSION || rh.rh_Version > RECSVERSION) {
	    return DISKERR_VERSION;
        }

	rh.rh_TabNum = FGetLong(file);
	rh.rh_NumRecs = FGetLong(file);
	rh.rh_NumFree = FGetLong(file);
	rh.rh_NumRecChanges = FGetLong(file);
	rh.rh_Unused1 = FGetLong(file);
	rh.rh_Unused2 = FGetLong(file);
	rh.rh_EofPos = FGetLong(file);
	if (rh.rh_Version < 7) {
	    FGetLong(file);     /* unused[4] */
	    FGetLong(file);
	    FGetLong(file);
	    FGetLong(file);
        }

        tab = FindTableNum(pr, rh.rh_TabNum);
        if (tab == NULL) {
            return DISKERR_INVALID;
        }
        tri = TableRecordInfo(tab);
        if (tri == NULL) {
            return ERR_INTERNAL;
        }
        map = tri->tri_IdRecordMap;
        if (map == NULL) {
            return ERR_INTERNAL;
        }

        QLSort(&tri->tri_FieldRecordInfos, fri_cmp);
	pos = 0;
	FORNODE (fri, &tri->tri_FieldRecordInfos) {
	    fri->fri_Pos = pos;
	    pos += FieldMemSize(fri->fri_Field);
	}

        for (i = 0; i < rh.rh_NumRecs; i++) {
            struct Record *rec;
            short flags;

            rd.rd_BitField = FGetShort(file);
            flags = FLAGS(rd.rd_BitField);
            if (rh.rh_Version < 7) {
                FGetShort(file);            /* unused */
            }
            rd.rd_Num = FGetLong(file);
            rd.rd_FileVersion = 0;
            if (rh.rh_Version >= 10) {
                rd.rd_FileVersion = FGetLong(file);
            }
            rd.rd_FilePos = FGetLong(file);
            rd.rd_FileTotalSize = FGetLong(file);

            rec = PoolAllocMemZero(memPool, sizeof(struct Record));
            if (rec == NULL) {
                return ERR_MEM;
            }
            rec->rec_Num = rd.rd_Num;
            rec->rec_MatchFilter = (flags & RECFLAG_MATCHFILTER) != 0;
            rec->rec_IsSorted = (flags & RECFLAG_ISSORTED) != 0;
            rec->rec_OnDisk = TRUE;
            rec->rec_FileVersion = rd.rd_FileVersion;
            rec->rec_CustomFormat.filePos = rd.rd_FilePos;
            rec->rec_CustomFormat.fileTotalSize = rd.rd_FileTotalSize;

            if (rec->rec_Num >= tri->tri_NextRecordNum) {
                tri->tri_NextRecordNum = rec->rec_Num + 1;
            }
            if (rh.rh_Version < 9 && ISGUIRECORD(rd.rd_BitField)) {
                SetTableGuiRecord(tab, rec);
            }
            AddRecord(tab, rec, ADDRECORD_TAIL);
        }

        for (i=0; i < rh.rh_NumFree; i++) {
            fd.fd_Pos = FGetLong(file);
            fd.fd_Size = FGetLong(file);
        }

        for (i=0; i < rh.rh_NumRecChanges; i++) {
            bool oldFormat = (bool)(rh.rh_Version < 8);
	    struct CustomFormatRecChange *rc;
           
            error = loadRecChangeCustomFormat(file, memPool, oldFormat, &rc);
            if (error) {
                return error;
            }
	    
	    if (rc->rc_Op != OP_NONE) {
                ADDTAIL(&tri->tri_CustomFormat.recChanges, rc);
                tri->tri_CustomFormat.recChangeId = rc->rc_Id + 1;
	    } else {
		FreeCustomFormatRecChange(memPool, rc);
            }
        }
        if (!IdRecordMapSet(memPool, map, TableRecords(tab))) {
            return ERR_MEM;
        }
    }
    return OK;
}

long LoadFieldRecordInfoCustomFormat(struct Field *field, FILE *file,
	struct FieldRecordInfo **frip)
{
    struct Table *tab = FieldTable(field);
    struct TableRecordInfo *tri = TableRecordInfo(tab);
    struct MemPool *memPool = TableMemPool(tab);
    struct FieldRecordInfo *fri;

    if (memPool == NULL || tri == NULL) {
        return ERR_INTERNAL;
    }
    fri = PoolAllocMem(memPool, sizeof(struct FieldRecordInfo));
    if (fri == NULL) {
        return(ERR_MEM);
    }
    ADDTAIL(&tri->tri_FieldRecordInfos, fri);

    fri->fri_Field = field;
    fri->fri_Pos = FGetLong(file);
    FGetLong(file);             /* dummy */

    if (frip != NULL) {
        *frip = fri;
    }

    return OK;
}

