/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** record.h
**
** Creation date: 27.02.1995
** Last modified: 11.02.2022
*/

#ifndef RECORD_H
#define RECORD_H

#include "system.h"

struct Record;

#ifdef NEED_NODE_CAST
inline Record *nodecast(Record *, Node *node) { return (Record *)node; }
#endif

struct TableRecordInfo;
struct FieldRecordInfo;

union Data
{
    long integer;
    double real;
    ushort choice;
    long date;
    long time;
    char *string;
    char *mi;
    struct Record *rec;         /* reference record */
};

struct DataInfo
{
    char di_Flags;
    char di_Private;            /* internal use by module record */
    union Data di_Contents;
};

#define DIF_NIL         (1<<1)

#include "table.h"

#ifdef __cplusplus
extern "C" {
#endif

void FreeTableRecordInfo(struct TableRecordInfo *tri);
struct TableRecordInfo *AllocTableRecordInfo(struct Table *tab);
void SetTableRecordInfoCustomFormat(struct TableRecordInfo *tri, bool set);

void FreeFieldRecordInfo(struct FieldRecordInfo *fri);
long DeleteFieldRecordInfo(struct FieldRecordInfo *fri);
/* returns error num or OK */
struct FieldRecordInfo *AllocFieldRecordInfo(struct Field *field);
long NewFieldRecordInfo(struct Field *field, struct FieldRecordInfo **fri);
/* returns error num or OK */
long CanNewFieldRecordInfo(struct Table *tab, short type);
/* returns error num or OK */
long ChangeStringRecordInfo(struct FieldRecordInfo *ri, long newstringsize);
/* returns error num or OK */

long RecordDataInfo(struct Record *rec, struct Field *field, 
	struct DataInfo **di, struct DataInfo *di_buf);
/* returns error code or OK */

void RecordLock(struct Record *rec);
void RecordUnlock(struct Record *rec);
bool RecordMatchFilter(struct Record *rec);
void SetRecordMatchFilter(struct Record *rec, bool match);
bool RecordIsSorted(struct Record *rec);
void SetRecordIsSorted(struct Record *rec, bool ordered);
bool RecordImported(struct Record *rec);
void SetRecordImported(struct Record *rec, bool imported);
struct Record *NextRecordWithReference(struct Record *rec, bool filter,
        struct Field *refField, struct Record *refRec);
struct Record *PrevRecordWithReference(struct Record *rec, bool filter,
        struct Field *refField, struct Record *refRec);
struct Record *FirstRecordWithReference(struct Table *tab, bool filter,
        struct Field *refField, struct Record *refRec);
struct Record *LastRecordWithReference(struct Table *tab, bool filter,
        struct Field *refField, struct Record *refRec);
struct Record *FirstRecord(struct Table *tab, bool filter);
struct Record *LastRecord(struct Table *tab, bool filter);
struct Record *NextRecord(struct Record *rec, bool filter);
struct Record *PrevRecord(struct Record *rec, bool filter);
struct Record *RecordNth(struct Table *tab, bool filter, long num);
void RecordNumAll(struct Table *tab, struct Record *rec, bool filter, 
	ulong *num, ulong *all);
struct Table *RecordFindTable(struct Record *rec, struct Table *guess, 
	struct Project *pr);

void FreeRecord(struct Table *tab, struct Record *rec, bool quick);
long AttemptDeleteRecord(struct Table *tab, struct Record *rec);
long DeleteRecord(struct Table *tab, struct Record *rec);
long NewRecord(struct Table *tab, struct Record *init, struct Record *pred, 
	bool addToGui, struct Record **rec);
/*
** returns error num, rec will point to new record, init may be NULL.
*/
long DeleteRecordsTable(struct Table *tab, bool filter, bool import);
long DeleteRecordsProject(struct Project *pr);

long CopyRecord(struct Table *tab, struct Record *rec, struct Record *init);

long RecordStoreData(struct Record *rec, struct Field *field, 
	union Data *data, bool nil);

long RecordNum(struct Record *rec);
bool MoveRecord(struct Table *tab, struct Record *rec, long num);
/*
** Moves the record at the given position.  Returns TRUE on success.
** If the record has been moved, also sets the IsSorted flag to FALSE.
*/

long RecordInternalNum(struct Record *rec);
struct Record *FindRecordInternalNum(struct Table *tab, ulong num);
long RecordFileVersion(struct Record *rec);

#ifdef CHECK_RECORD_HEADERS
bool CheckRecordHeaders(const char *message);
#endif

/*
** load & save
*/

double TableRecordOrderNumResolution(struct Table *tab);
void ResetRecordOrderNumResolutions(struct Project *pr);
long WriteTableRecords(struct SQLiteHandle *handle, struct Table *tab,
        void (*wroteOneRecord)(void));
long SwapTableRecords(struct Table *tab, bool header, bool freeData,
        void (*swappedOneRecord)(void));
long SwapProjectRecords(struct Project *pr);
void SwapRecords(void);                 /* free as much memory as possible */

struct ReadRecordData
{
    long rrd_ID;
    long rrd_Version;
    bool rrd_IsSorted;
    bool rrd_MatchFilter;
};

void PrepareRecordUpdate(struct Project *pr);
long ReadRecord(struct Table *tab, const struct ReadRecordData *data,
        bool update);
long AfterReadRecords(struct Table *tab);
void CompleteRecordUpdate(struct Project *pr);

long LoadFieldRecordInfoCustomFormat(struct Field *field, FILE *file,
	struct FieldRecordInfo **ri);

long LoadRecordsCustomFormat(struct Project *pr, FILE *file,
        const char **errstr);
void RecordAfterReorgUpdate(struct Project *pr);

#ifdef __cplusplus
}
#endif

#endif

