/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** recorddescription.c
**
** Creation date: 15.07.2012
** Creation date: 21.08.2012
*/

#include "recorddescription.h"
#include "lang/program.h"
#include "programeditor.h"
#include "gui.h"
#include "table.h"
#include "record.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

#define RD_DELIMITER " - "

struct RecordDescriptionEntry
{
    char *rde_Text;
    short rde_Type;
};

struct RecordDescription
{
    struct Table *rd_Table;
    const struct RecordDescriptionSpec *rd_Spec;
    struct Record *rd_Record;
    long rd_Error;
    long rd_NumEntries;
    struct RecordDescriptionEntry *rd_Entries;
};

struct RecordDescription *AllocRecordDescription(struct Table *tab,
        const struct RecordDescriptionSpec *rds)
{
    struct MemPool *memPool = TableMemPool(tab);
    struct RecordDescription *rd;

    rd = PoolAllocMemZero(memPool, sizeof(struct RecordDescription));
    if (rd != NULL) {
        rd->rd_Table = tab;
        rd->rd_Spec = rds;
    }

    return rd;
}

static void freeRecordDescriptionEntries(struct RecordDescription *rd)
{
    struct MemPool *memPool = TableMemPool(rd->rd_Table);
    long i;

    for (i = 0; i < rd->rd_NumEntries; i++) {
        FreeString(memPool, rd->rd_Entries[i].rde_Text);
    }
    PoolFreeMem(memPool, rd->rd_Entries, 
            rd->rd_NumEntries * sizeof(struct RecordDescriptionEntry));
    rd->rd_Entries = NULL;
    rd->rd_NumEntries = 0;
}

void FreeRecordDescription(struct RecordDescription *rd)
{
    if (rd) {
        freeRecordDescriptionEntries(rd);
        PoolFreeMem(TableMemPool(rd->rd_Table), 
                rd, sizeof(struct RecordDescription));
    }
}

void RecordDescriptionSetRecord(struct RecordDescription *rd, 
        struct Record *rec)
{
    if (rd) {
        rd->rd_Record = rec;
        rd->rd_Error = 0;
        freeRecordDescriptionEntries(rd);
    }
}

long RecordDescriptionError(struct RecordDescription *rd)
{
    long error = OK;

    if (rd) {
        error = rd->rd_Error;
    }
    return error;
}

static void getEntries(struct RecordDescription *rd)
{
    const struct RecordDescriptionSpec *rds = rd->rd_Spec;

    if (rd->rd_Entries == NULL && rd->rd_Error == OK && rds->rds_Compute) {
        struct MemPool *memPool = TableMemPool(rd->rd_Table);
        struct Project *pr = TableProject(rd->rd_Table);
        struct Program *prg = ProjectProgram(pr);
        struct UserFunction *uf = FindUserFunctionName(prg, rds->rds_Compute);
        const char *label = NULL;
        char *mem = NULL;

        if (uf != NULL) {
            bool endProgram = FALSE;
            struct Element *elem;
            struct Record *prevProgramRec;

            if (!ProgramRunning()) {
                SetTableProgramRecords(pr);
                ProgramBegin(prg, TRUE);
                endProgram = TRUE;
            }
            prevProgramRec = TableProgramRecord(rd->rd_Table);
            SetTableProgramRecord(rd->rd_Table, rd->rd_Record);
            elem = ProgramCall(prg, uf, NULL, FALSE);
            if (elem != NULL) {
                bool allocated;

                if (elem->el_Type == TYPE_LIST) {
                    long num = CountListLength(elem), i;
                    struct Element *head;

                    rd->rd_Entries = PoolAllocMemZero(memPool, 
                            num * sizeof(struct RecordDescriptionEntry));
                    if (rd->rd_Entries == NULL) {
                        rd->rd_Error = ERR_MEM;
                    } else {
                        for (i=0; i < num && elem->el_Type == TYPE_LIST; i++) {
                            head = elem->el_Contents.list.head;
                            mem = SimpleElementToString(head, NULL, 
                                    memPool, &allocated);
                            if (!allocated) {
                                mem = AllocString(memPool, mem);
                            }
                            if (mem == NULL) {
                                rd->rd_Error = ERR_MEM;
                            } else {
                                rd->rd_Entries[i].rde_Text = mem;
                                rd->rd_Entries[i].rde_Type = head->el_Type;
                            }
                            elem = elem->el_Contents.list.tail;
                        }
                        rd->rd_NumEntries = num;
                    }
                } else {
                    label = mem = SimpleElementToString(elem, NULL, 
                            memPool, &allocated);
                    if (!allocated) {
                        mem = NULL;
                    }
                    if (label == NULL) {
                        rd->rd_Error = ERR_MEM;
                    }
                }
            } else {
                label = ProgramError(NULL, NULL, NULL);
            }
            if (endProgram) {
                static const char *errorCompute = NULL;
                long line, col;
                const char *filename;
                const char *errstr;

                ProgramEnd();
                errstr = ProgramError(&line, &col, &filename);
                if (errstr != NULL) {
                    if (errstr != EXECUTE_ABORT) {
                        if (errorCompute != rds->rds_Compute) {
                            errorCompute = rds->rds_Compute;
                            ProgramEditorShowError(pr, TRUE, errstr, 
                                    line, col, filename);
                        }
                    }
                }
            } else {
                if (!FindNodePtr(TableRecords(rd->rd_Table), prevProgramRec)) {
                    prevProgramRec = NULL;
                }
                SetTableProgramRecord(rd->rd_Table, prevProgramRec);
            }
        } else {
            static const char *warnCompute = NULL;

            if (rds->rds_Compute != warnCompute) {
                warnCompute = rds->rds_Compute;
                Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Err_Trigger),warnCompute);
            }
            label = rds->rds_Compute;
        }
        if (label) {
            rd->rd_Entries = PoolAllocMemZero(memPool, 
                    sizeof(struct RecordDescriptionEntry));
            if (rd->rd_Entries == NULL) {
                rd->rd_Error = ERR_MEM;
            } else {
                if (mem == NULL) {
                    mem = AllocString(memPool, label);
                    if (mem == NULL) {
                        rd->rd_Error = ERR_MEM;
                    }
                }
                rd->rd_Entries[0].rde_Text = mem;
                rd->rd_NumEntries = 1;
            }
        }
    }
}

long RecordDescriptionNumColumns(struct RecordDescription *rd)
{
    long cols = 0;

    if (rd) {
        const struct RecordDescriptionSpec *rds = rd->rd_Spec;

        if (rds) {
            switch (rds->rds_Type) {
                case RDS_TYPE_FIELDS:
                    cols = rds->rds_NumFields;
                    break;
                case RDS_TYPE_COMPUTED:
                    getEntries(rd);
                    cols = rd->rd_NumEntries;
                    break;
            }
        }
    }
    return cols;
}

short RecordDescriptionColumnType(struct RecordDescription *rd, long col)
{
    short type = NOTYPE;
    struct Field *field;

    if (rd && col >= 0) {
        const struct RecordDescriptionSpec *rds = rd->rd_Spec;

        if (rds) {
            switch (rds->rds_Type) {
                case RDS_TYPE_FIELDS:
                    if (col < rds->rds_NumFields) {
                        field = rds->rds_Fields[col];
                        type = (field == RDS_FIELD_RECNUM)? 
                            TYPE_INTEGER : FieldType(field);
                    }
                    break;
                case RDS_TYPE_COMPUTED:
                    getEntries(rd);
                    if (col < rd->rd_NumEntries) {
                        type = rd->rd_Entries[col].rde_Type;
                    }
                    break;
            }
        }
    }
    return type;
}

char *RecordDescriptionColumnText(struct RecordDescription *rd, long col,
        struct MemPool *memPool, bool *pfree)
{
    char *text = NULL;
    bool free = FALSE;
    struct Field *field;

    if (rd && col >= 0) {
        const struct RecordDescriptionSpec *rds = rd->rd_Spec;

        if (rds) {
            switch (rds->rds_Type) {
                case RDS_TYPE_FIELDS:
                    if (col < rds->rds_NumFields) {
                        field = rds->rds_Fields[col];
                        if (field == RDS_FIELD_RECNUM) {
                            text = IntToStr(RecordNum(rd->rd_Record));
                            text = AllocString(memPool, text);
                            if (text == NULL) {
                                rd->rd_Error = ERR_MEM;
                            } else {
                                free = TRUE;
                            }
                        } else {
                            rd->rd_Error = GuiFieldDataString(field, 
                                    rd->rd_Record, memPool, &text, &free);
                        }
                    }
                    break;
                case RDS_TYPE_COMPUTED:
                    getEntries(rd);
                    if (col < rd->rd_NumEntries) {
                        text= AllocString(memPool,rd->rd_Entries[col].rde_Text);
                        if (text == NULL) {
                            rd->rd_Error = ERR_MEM;
                        } else {
                            free = TRUE;
                        }
                    }
                    break;
            }
        }
    }

    if (pfree != NULL) {
        *pfree = free;
    }
    return text;
}

char *RecordDescriptionText(struct RecordDescription *rd,
        struct MemPool *memPool, bool *pfree)
{
    char *text = NULL;
    bool free = FALSE;

    if (rd) {
        long num = RecordDescriptionNumColumns(rd);
        long i;

        for (i = 0; i < num; i++) {
            char *subText;
            bool subFree;

            subText = RecordDescriptionColumnText(rd, i, memPool, &subFree);
            if (rd->rd_Error) {
                break;
            }
            if (text == NULL) {
                text = subText;
                free = subFree;
            } else {
                char *newText = PoolAllocMem(memPool, strlen(text) +
                        strlen(RD_DELIMITER) + strlen(subText) + 1);

                if (newText == NULL) {
                    rd->rd_Error = ERR_MEM;
                    if (subFree) {
                        FreeString(memPool, subText);
                    }
                    break;
                }
                sprintf(newText, "%s%s%s", text, RD_DELIMITER, subText);
                if (subFree) {
                    FreeString(memPool, subText);
                }
                if (free) {
                    FreeString(memPool, text);
                }
                text = newText;
                free = TRUE;
            }
        }

        if (text == NULL) {
            text = IntToStr(RecordNum(rd->rd_Record));
            free = FALSE;
        }
    }

    if (pfree) {
        *pfree = free;
    }
    return text;
}

struct RecordDescriptionSpec *AllocRecordDescriptionSpec(
        struct MemPool *memPool, short numFields)
{
    struct RecordDescriptionSpec *rds;
    long size = sizeof(struct RecordDescriptionSpec) + 
        numFields * sizeof(struct Field *);

    rds = PoolAllocMemZero(memPool, size);
    if (rds != NULL) {
        rds->rds_NumFields = numFields;
    }

    return rds;
}

struct RecordDescriptionSpec *CopyRecordDescriptionSpec(
        struct MemPool *memPool, const struct RecordDescriptionSpec *orig)
{
    struct RecordDescriptionSpec *rds = NULL;
    short i;

    if (orig != NULL) {
        rds = AllocRecordDescriptionSpec(memPool, orig->rds_NumFields);
        if (rds != NULL) {
            rds->rds_Type =orig->rds_Type;
            rds->rds_MultiColumn =orig->rds_MultiColumn;
            rds->rds_Compute = AllocString(memPool, orig->rds_Compute);
            for (i=0; i < rds->rds_NumFields; i++) {
                rds->rds_Fields[i] = orig->rds_Fields[i];
            }
	}
    }

    return rds;
}

void FreeRecordDescriptionSpec(struct MemPool *memPool, 
        struct RecordDescriptionSpec *rds)
{
    if (rds != NULL) {
        long size = sizeof(struct RecordDescriptionSpec) +
            rds->rds_NumFields * sizeof(struct Field *);

        FreeString(memPool, rds->rds_Compute);
        PoolFreeMem(memPool, rds, size);
    }
}

bool RecordDescriptionSpecIsDefault(const struct RecordDescriptionSpec *rds)
{
    struct RecordDescriptionSpec def;

    memset(&def, 0, sizeof(def));
    return (rds == NULL || memcmp(rds, &def, sizeof(def)) == 0);
}

