/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** requesters.c
**
** Creation date: 05.12.2004
** Last modified: 05.12.2004
*/

#include "requesters.h"
#include "labels.h"
#include "memo.h"
#include "util.h"
#include "memory.h"
#include "dbtypes.h"
#include "appstrings_cat.h"

void FreeAskMultiInfo(struct AskMultiInfo *ami)
{
    if(ami)
    {
        struct MemPool *memPool = ami->ami_MemPool;
        struct AskMultiItem *a = ami->ami_Items;
        short num = ami->ami_Num, i;

        for (i=0; i < num; i++, a++)
        {
            FreeString(memPool, a->ami_Title);
            FreeString(memPool, a->ami_BubbleHelp);

            switch(a->ami_Type)
            {
                case TYPE_STRING:
                    FreeString(memPool, a->ami_Contents.string.string);
                    break;

                case TYPE_MEMO:
                    FreeString(memPool, a->ami_Contents.mi);
                    break;

                case TYPE_CHOICE:
                    FreeLabels(memPool, a->ami_Contents.choice.labels);
                    break;

                case TYPE_OPTIONS:
                    PoolFreeMem(memPool, a->ami_Contents.options.selected,
                        NumLabels(a->ami_Contents.options.labels)*sizeof(bool));
                    FreeLabels(memPool, a->ami_Contents.options.labels);
                    break;
            }
        }

        PoolFreeMem(memPool, ami,
            sizeof(struct AskMultiInfo) + num * sizeof(struct AskMultiItem));
    }
}

struct AskMultiInfo *AllocAskMultiInfo(struct MemPool *memPool, short num)
{
    struct AskMultiInfo *ami;
    struct AskMultiItem *a;
    short i;

    ami = PoolAllocMemZero(memPool,
        sizeof(struct AskMultiInfo) + num * sizeof(struct AskMultiItem));
    if (ami == NULL)
        return NULL;

    ami->ami_MemPool = memPool;
    ami->ami_Num = num;
    for (i=0, a = ami->ami_Items; i < num; i++, a++)
        a->ami_Type = TYPE_NIL;

    return ami;
}

bool CheckAskMultiInfo(struct AskMultiInfo *ami, GuiObject win)
{
    struct AskMultiItem *a = ami->ami_Items;
    int i;

    for (i=0; i < ami->ami_Num; i++, a++)
    {
	const char *text = GuiStringContents(a->ami_GuiObject);
        const char *error = NULL;

	if (!text) {
	    continue;
	}
	while (CodesetIsSpace(CodesetGetChar(text))) {
    	    text = CodesetNextChar(text);
	}
	if (text[0] == 0) {
	    continue;
	}

        switch (a->ami_Type) {
            case TYPE_INTEGER:
                if (!StrToInt(text, NULL)) 
		    error = GCS(MSG_Err_EnterInteger); 
		break;

            case TYPE_REAL:
                if (!StrToDouble(text, NULL)) 
		    error = GCS(MSG_Err_EnterReal);
		break;

            case TYPE_DATE:
                if (!StringToDate(text, NULL))
		    error = GCS(MSG_Err_EnterDate); 
		break;

            case TYPE_TIME:
                if (!StringToTime(text, FORMAT_HOURMINSEC, NULL)) 
		    error = GCS(MSG_Err_EnterTime); 
		break;
        }

        if (error) {
	    if (win) {
		GuiSetActiveGuiObject(win, a->ami_GuiObject);
		GuiSetWindowTitle(win, error);
		GuiBeep();
	    }
            return FALSE;
        }
    }
    return TRUE;
}

