/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** requesters.h
**
** Creation date: 23.12.1994
** Last modified: 20.01.2022
*/

#include "system.h"

#ifdef __cplusplus
extern "C" {
#endif

char *GetFilename(struct Project *pr, struct MemPool *memPool,
        const char *title, const char *oktext,
        const char *pref, long prefitem, bool save_mode);
char *GetDirname(struct Project *pr, struct MemPool *memPool,
        const char *title, const char *oktext,
        const char *pref, long prefitem, bool save_mode);
/*
** Popups file/dir dialog.  You can set window title, ok button title
** and default filename/dirname.  If prefitem != 0,
** the selected filename/dirname is stored in preferences and
** can be used for further calls, if pref == NULL, pref is taken
** form the preferences.
** save_mode is a flag that should be set when the file/dir dialog
** is used for selecting a file/dir to save to.
** Returns file/dirname or NULL on failure/cancel.
** The memory for the return string is allocated from
** the given memPool and has to be freed by you!
*/

typedef enum {
    AT_NONE,			// no icon
    AT_INFO,
    AT_QUESTION,
    AT_WARN,
    AT_ERROR,
} AskType;

#if (defined __amigaos4__) || (defined __MORPHOS__)
    short VARARGS68K Ask(AskType type, const char *buttons,
    	const char *text, ...);
    short VARARGS68K Ask2(GuiObject obj, const char *title, AskType type, 
	    const char *buttons, const char *text, ...);
#elif (defined __AROS__)
short Ask2A(Object *obj, const char *title, AskType type,
	const char *buttons, const char *text, void *params);
#define Ask(type, buttons, text, ...) \
({ \
    IPTR __args[] = { AROS_PP_VARIADIC_CAST2IPTR(__VA_ARGS__) }; \
    Ask2A(NULL, NULL, (type), (buttons), (text), __args); \
})
#define Ask2(obj, title, type, buttons, text, ...) \
({ \
    IPTR __args[] = { AROS_PP_VARIADIC_CAST2IPTR(__VA_ARGS__) }; \
    Ask2A((obj), (title), (type), (buttons), (text), __args); \
})
#else
    short Ask(AskType type, const char *buttons, const char *text, ...);
    short Ask2(GuiObject obj, const char *title, AskType type, 
	    const char *buttons, const char *text, ...);
#endif

typedef enum {
    AST_Normal,
    AST_FileSelect,
    AST_Secret
} AskStringType;

const char *AskString(struct Project *pr, const char *title, const char *oktext,
	const char *contents, ulong maxlen, AskStringType type);
bool AskInteger(struct Project *pr, const char *title, const char *oktext,
	const char *contents, long min, long max, long *result);

bool AskChoice(struct Project *pr, const char *title, const char *oktext,
        struct Element *labels, struct Element *titles, bool *selected,
	bool multiselect, bool stringmode, long *index, char **string);
/*
** AskChoice is a multi purpose dialog for letting the user choose
** 1 out of n items, m out of n items, or a string which may be selected
** from the supplied ones or entered in a string field.
** You can set the window title and the text of the ok button.
** labels is a NULL terminated pointer array specifying the choice strings.
** If labels or labels[0] is NULL, FALSE is returned immediately.
** If a label contains tab characters '\t' then a multi-column list
** is used where the tab stops are the column borders.
** In titles the titles of the columns are specified (tab separated).
** If titles = NULL then all titles are turned off.
** In selected you can specify which items start selected in multiselect mode,
** selected = NULL means that all items start unselected.
** multiselect = TRUE turns multiselect mode on.
** stringmode = TRUE allows entering a custom string.
** Note that multiselect and stringmode are mutual exclusive. You can't
** turn both on.
** In *index you can specify the inital active item (multiselect = FALSE,
** stringmode = FALSE). Use *index = -1 for no active item.
** In *string you can specify the initial string for the custom string field
** (multiselect = FALSE, stringmode = TRUE).
** AskChoice returns FALSE if the user canceled the dialog, and TRUE
** if the user clicked the ok button.  In the latter case one of the
** parameters selected, index, or string is updated with the result.
*/

void Message(const char *text);

struct AskMultiItem
{
    short ami_Type;                     /* basic types of dbtypes.h, NOTYPE for title only */
    char *ami_Title;              	/* title string */
    char *ami_BubbleHelp;         	/* bubble help, if any */
    bool ami_Nil;                       /* TRUE for NIL values */
    union
    {
        long integer;
        double real;
        long date;
        long time;
        char *mi;
	struct
	{
	    char *string;
	    bool secret;
	} string;
        struct
        {
            bool ischoicelist;
            long active;
            char **labels;
        } choice;
        struct
        {
            bool *selected;
            char **labels;
        } options;
    } ami_Contents;
    GuiObject ami_GuiObject;            /* private field! */
};

struct AskMultiInfo
{
    struct MemPool *ami_MemPool;
    short ami_Num;                      /* Number of following items */
    struct AskMultiItem ami_Items[0];
};

struct AskMultiInfo *AllocAskMultiInfo(struct MemPool *memPool, short num);

void FreeAskMultiInfo(struct AskMultiInfo *ami);

bool CheckAskMultiInfo(struct AskMultiInfo *ami, GuiObject win);

bool AskMulti(struct Project *pr, const char *title, const char *oktext,
	struct AskMultiInfo *ami);

#ifdef __cplusplus
}
#endif

