/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** search.c
**
** Creation date: 16.09.1996
** Last modified: 01.05.2013
*/

#include "system.h"
#include "project.h"
#include "table.h"
#include "field.h"
#include "gui.h"
#include "order.h"
#include "search.h"
#include "requesters.h"
#include "break.h"
#include "appstrings_cat.h"
#include "memory.h"
#include "codeset.h"
#include "util.h"
#include "error.h"

struct SearchFor SearchFor =
{
    NULL, NULL, "", FALSE, FALSE, TRUE, TRUE, TRUE,
};


static long searchRecordField(struct Record *rec,
        struct Field *field, char *pattern, long (*cmpfunc)(char *, char *), long *presult)
{
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = ProjectMemPool(pr);
    short type = FieldType(field);
    long error = OK;
    char *text = NULL;
    long result = 0;
    bool free = FALSE;

    if(type == TYPE_BUTTON || type == TYPE_VIRTUAL)
    {
        if(presult != NULL)
            *presult = 1;
        return(OK);
    }

    RecordLock(rec);
    error = GuiFieldDataString(field, rec, memPool, &text, &free);
    if(error == OK && text != NULL)
        result = cmpfunc(text, pattern);
    RecordUnlock(rec);
    if(free)
        FreeString(memPool, text);

    if(presult != NULL)
        *presult = result;
    return(error);
}

static char *binPattern;
static short binlen;
static bool binFilter;
static bool binSensitive;
static bool binAscending;

static long binCmpFunc(char *text, char *pattern)
{
    return CodesetStrNCollate(text, pattern, binlen, CS_COLLATE_SIMPLE);
}

static bool searchbin(struct Field *field,
        struct Record **first, struct Record **last, bool left, bool right, long *error)
{
    struct Record *mid, *rec;
    long result;

    if(BreakCheck())
        return(FALSE);

    rec = mid = *first;
    for(;;)
    {
        if(rec == NULL || rec == *last)
            break;
        rec = NextRecord(rec, binFilter);
        mid = NextRecord(mid, binFilter);
        if(rec == NULL || rec == *last)
            break;
        rec = NextRecord(rec, binFilter);
    }

    if(rec == NULL)
    {
        *error = ERR_INTERNAL;
        return(FALSE);
    }

    *error = searchRecordField(mid, field, binPattern, binCmpFunc, &result);
    if(*error)
        return(FALSE);

    if(!binAscending)
        result = -result;

    if(result != 0 && *first == *last)
    {
        *first = *last = NULL;
        return(TRUE);
    }

    if(result > 0)
    {
        *last = (mid == *first)? mid : PrevRecord(mid, binFilter);
        return(searchbin(field, first, last, left, right, error));
    }
    else if(result < 0)
    {
        *first = (mid == *last)? mid : NextRecord(mid, binFilter);
        return(searchbin(field, first, last, left, right, error));
    }

    if(*first == *last)
        return(TRUE);

    if(left)
    {
        rec = (mid == *first)? mid : PrevRecord(mid, binFilter);
        if(!searchbin(field, first, &rec, TRUE, FALSE, error))
            return(FALSE);
        if(*first == NULL)
            *first = mid;
    }

    if(right)
    {
        rec = (mid == *last)? mid : NextRecord(mid, binFilter);
        if(!searchbin(field, &mid, last, FALSE, TRUE, error))
            return(FALSE);
        if(*last == NULL)
            *last = mid;
    }

    return(TRUE);
}

static long searchcmpfunc(char *s, char *j)
{
    bool rc = strjokercmp(s, j);

    return (rc? 0 : 1);
}

static long searchCmpFunc(char *s, char *j)
{
    bool rc = StrJokerCmp(s, j);

    return (rc? 0 : 1);
}

const char *SearchRecord(struct SearchFor *sf,
        struct Record **prec, struct Record *current)
{
    struct Table *tab = sf? sf->sf_Table : NULL;
    struct OrderField *oa = GetOrderField(tab, 0);
    struct Field *sffield=(sf && !sf->sf_AllFields)? sf->sf_Field : NULL;
    struct Field *field = sffield;
    struct Project *pr = TableProject(tab);
    struct Record *first, *last, *rec = NULL;
    struct List *fields = TableFields(tab);
    char *pattern;
    bool filter;
    long (*cmpfunc)(char *, char *);
    short len = 0;
    long equal = 1;
    long error = OK;
    bool aborted = FALSE;
    struct BreakSettings bs;

    if(prec != NULL)
        *prec = NULL;

    if(pr == NULL)
        return(GCS(MSG_Err_Internal));

    /*GuiSetSleep(TRUE); */
    bs.title = GCS(MSG_Searching);
    bs.steps = 0;
    bs.current = 0;
    bs.abort = TRUE;
    bs.pause = TRUE;
    bs.needGui = ProjectHasGui(pr);
    BreakInit(&bs);

    sf->sf_MatchField = NULL;
    filter = sf->sf_MatchFilter;
    pattern = sf->sf_Pattern;
    cmpfunc = sf->sf_CaseSensitive? searchcmpfunc : searchCmpFunc;

    if(sf->sf_FirstLastRecord || current == NULL)
        first = sf->sf_Forward?
	    FirstRecord(tab, filter) : LastRecord(tab, filter);
    else
        first = sf->sf_Forward?
	    NextRecord(current, filter) : PrevRecord(current, filter);
    last = (sf->sf_Forward)? LastRecord(tab, filter) : FirstRecord(tab, filter);

    if(first != NULL && sffield != NULL && oa != NULL &&
	sffield == oa->oa_Field &&
        sf->sf_Pattern[0] != JOKER && sf->sf_Pattern[0] != SINGLEJOKER)
    {
        /*
        ** before we do a binary search we check if the first record
        ** already matches.
        */
        rec = first;
        if(!sf->sf_FirstLastRecord && (!filter || GuiCheckRecord(tab, rec)))
            error = searchRecordField(rec, sffield,
		sf->sf_Pattern, cmpfunc, &equal);

        if(error == OK && equal != 0)
        {
            while(pattern[len] != 0 && pattern[len] != JOKER && 
	    pattern[len] != SINGLEJOKER)
                len++;
            binPattern = pattern;
            binlen = len;
            binFilter = filter;
            binSensitive = sf->sf_CaseSensitive;
            binAscending = oa->oa_Ascending;
            if(!sf->sf_Forward)
            {
                rec = first;
                first = last;
                last = rec;
            }

            aborted = !searchbin(sffield, &first, &last, TRUE, TRUE, &error);
            if(!sf->sf_Forward)
            {
                rec = first;
                first = last;
                last = rec;
            }
        }
    }

    if(!aborted && first != NULL && error == OK && equal != 0)
    {
        rec = first;
        for(;;)
        {
            if(filter)
                while(!(aborted = BreakCheck()) && rec != NULL && 
		    !GuiCheckRecord(tab, rec))
                {
                    if(rec == last)
                        rec = NULL;
                    rec = sf->sf_Forward? 
			NextRecord(rec, filter) : PrevRecord(rec, filter);
                }
	    aborted = BreakCheck();
            if(aborted || rec == NULL)
                break;

            field = sffield;
            if(field)
                error = searchRecordField(rec, field,
		    sf->sf_Pattern, cmpfunc, &equal);
            else if(fields)
                FORNODE(field, fields)
                {
                    error = searchRecordField(rec, field,
			sf->sf_Pattern, cmpfunc, &equal);
                    if(error != OK || equal == 0)
                        break;
                }

            if(error != OK || equal == 0)
                break;

            if(rec == last)
                rec = NULL;
            rec = sf->sf_Forward? 
		NextRecord(rec, filter) : PrevRecord(rec, filter);
        }
    }

    BreakExit(&bs); 
    /* GuiSetSleep(FALSE); */

    if(error)
        return(GES(error));
    if(aborted)
        return(NULL);
    if(equal != 0)
        return(GCS(MSG_SearchPatternNotFound));

    if(prec != NULL)
        *prec = rec;
    sf->sf_MatchField = field;
    return(NULL);
}

void ShowSearchRecord(struct SearchFor *sf)
{
    struct Table *tab = sf? sf->sf_Table : NULL;
    struct Record *rec = NULL;
    const char *error = SearchRecord(sf, &rec, TableGuiRecord(tab));

    if(error)
        Ask(AT_ERROR, GCS(MSG_StarOk), error, sf->sf_Pattern);
    else if(rec != NULL)
    {
        GuiShowRecord(tab, rec);
        if(sf != NULL && sf->sf_MatchField != NULL)
            GuiActivateDisplayObject(FieldDisplayObject(sf->sf_MatchField));
    }
}

