/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** server.c
**
** Creation date: 12.01.2001
// Last modified: 28.12.2021
*/

#include "main.h"
#include "server.h"
#include "project.h"
#include "preferences.h"
#include "status.h"
#include "memory.h"
#include "strhandle.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"
#include "requesters.h"
#include "menuproject.h"
#include "lang/compile.h"
#include "lang/program.h"

struct ServerRequest
{
    struct MinNode node;
    short type;
    void *userData;
    void (*func)(long error, char *result, void *userData);
    union
    {
	struct
	{
	    char *projectName;
	    bool gui;
	} connect;
	struct
	{
	    long handle;
	} disconnect;
	struct
	{
	    long handle;
	    char *cmd;      
	} eval;
	struct
	{
	    long handle;
	} transaction;
	struct
	{
	    long handle;
	} commit;
	struct
	{
	    long handle;
	} rollback;
    }contents;
};

/* 
** Types 
*/
enum
{
    SRT_CONNECT, SRT_DISCONNECT, 
    SRT_CONNECTIONS,
    SRT_EVAL, 
    SRT_TRANSACTION, SRT_COMMIT, SRT_ROLLBACK,
};

struct ProjectHandle
{
    struct MinNode node;
    long handle;
    struct Project *pr;
};

static struct
{
    bool init;
    struct MemPool *memPool;
    struct List handles;
    struct List requests;
    bool checkRequests;
    long lastHandle;
} Server = { FALSE, };

/*
** Functions related to ServerRequest
*/

static void freeServerRequest(struct MemPool *memPool, 
			      struct ServerRequest *sr)
{
    if(sr != NULL)
    {
	switch(sr->type)
	{
	case SRT_CONNECT:
	    FreeString(memPool, sr->contents.connect.projectName);
	    break;

	case SRT_EVAL:
	    FreeString(memPool, sr->contents.eval.cmd);
	    break;
	}
	PoolFreeMem(memPool, sr, sizeof(struct ServerRequest));
    }	 
}

static struct ServerRequest *allocServerRequest(
    struct MemPool *memPool, void *userData, void (*func)(long, char *, void *))
{
    struct ServerRequest *sr;

    sr = PoolAllocMemZero(memPool, sizeof(struct ServerRequest));
    if(sr != NULL)
    {
	sr->userData = userData;
	sr->func = func;
    }

    return sr;
}

static long serverRequestHandle(struct ServerRequest *sr)
{
    switch(sr->type)
    {
    case SRT_DISCONNECT:
	return sr->contents.disconnect.handle;
    case SRT_EVAL:
	return sr->contents.eval.handle;
    case SRT_TRANSACTION:
	return sr->contents.transaction.handle;
    case SRT_COMMIT: 
	return sr->contents.commit.handle;
    case SRT_ROLLBACK:
	return sr->contents.rollback.handle;
    }

    return HANDLE_NONE;
}

/*
** Functions related to ProjectHandle
*/

static void freeProjectHandle(struct MemPool *memPool, 
			      struct ProjectHandle *ph)
{
    if(ph != NULL)
    {
	PoolFreeMem(memPool, ph, sizeof(struct ProjectHandle));
    }
}

struct ProjectHandle *allocProjectHandle(struct MemPool *memPool)
{
    struct ProjectHandle *ph;

    ph = PoolAllocMem(memPool, sizeof(struct ProjectHandle));

    return ph;
}

/*
** Functions related to Server
*/

void ServerInit()
{
    NewList(&Server.handles);
    NewList(&Server.requests);

    Server.memPool = CreatePool(0, 4096, 1024);
    if(Server.memPool == NULL)
    {
	fail("Unable to create server memory pool.", 5);
    }

    Server.init = TRUE;
}

void ServerExit()
{
    if(Server.init)
    {
	struct ServerRequest *sr;
	struct ProjectHandle *ph;

	while((sr = (struct ServerRequest *)REMHEAD(&Server.requests)) != NULL)
	{
	    freeServerRequest(Server.memPool, sr);
	}
	NewList(&Server.requests);

	while((ph = (struct ProjectHandle *)REMHEAD(&Server.handles)) != NULL)
	{
	    freeProjectHandle(Server.memPool, ph);
	}
	NewList(&Server.handles);

	if(Server.memPool != NULL)
	{
	    DeletePool(Server.memPool);
	    Server.memPool = NULL;
	}

	Server.init = FALSE;
    }
}

void ServerRemoveProjectConnections(struct Project *pr)
{
    struct ProjectHandle *ph, *succ;
    const long error = SERVERR_BROKENHANDLE;

    /*
    ** Remove handle before removing requests to avoid new requests.
    */

    SAFEFORNODE(ph, succ, &Server.handles)
    {
	struct ServerRequest *sr, *succ;
	
	if(ph->pr == pr)
	{
	    REMOVE(ph);
	    SAFEFORNODE(sr, succ, &Server.requests)
	    {
		if(serverRequestHandle(sr) == ph->handle)
		{
		    REMOVE(sr);
		    if(sr->func)
		    {
			sr->func(error, GES(error), sr->userData);
		    }
		    freeServerRequest(Server.memPool, sr);
		}
	    }
	    freeProjectHandle(Server.memPool, ph);
	}
    }
}

long ServerNumProjectConnections(struct Project *pr)
{
    long count = 0;
    struct ProjectHandle *ph;

    FORNODE(ph, &Server.handles)
    {
	if(ph->pr == pr)
	{
	    count++;
	}
    }

    return count;
}

long ServerNumConnections(void)
{
    return NumNodes(&Server.handles);
}

long ServerAddConnectRequest(char *name, bool gui, void *userData, 
			    void (*func)(long error, char *result, void *ud))
{
    struct ServerRequest *sr;

    if(name == NULL)
    {
	return ERR_INVALIDNAME;
    }

    sr = allocServerRequest(Server.memPool, userData, func);
    if(sr == NULL)
    {
	return ERR_MEM;
    }

    sr->type = SRT_CONNECT;
    sr->contents.connect.gui = gui;
    sr->contents.connect.projectName = AllocString(Server.memPool, name);
    if(sr->contents.connect.projectName == NULL)
    {
	freeServerRequest(Server.memPool, sr);
	return ERR_MEM;
    }

    ADDTAIL(&Server.requests, sr);
    Server.checkRequests = TRUE;

    return OK;
}

long ServerAddDisconnectRequest(long handle, void *userData, 
				void (*func)(long error, char *result, 
					     void *ud))
{
    struct ServerRequest *sr;

    sr = allocServerRequest(Server.memPool, userData, func);
    if(sr == NULL)
    {
	return ERR_MEM;
    }

    sr->type = SRT_DISCONNECT;
    sr->contents.disconnect.handle = handle;

    ADDTAIL(&Server.requests, sr);
    Server.checkRequests = TRUE;

    return OK;
}

long ServerAddConnectionsRequest(void *userData, 
				 void (*func)(long error, char *result, 
					      void *ud))
{
    struct ServerRequest *sr;

    sr = allocServerRequest(Server.memPool, userData, func);
    if(sr == NULL)
    {
	return ERR_MEM;
    }

    sr->type = SRT_CONNECTIONS;

    ADDTAIL(&Server.requests, sr);
    Server.checkRequests = TRUE;

    return OK;
}

long ServerAddEvalRequest(long handle, const char *cmd, void *userData,
			 void (*func)(long error, char *result, void *ud))
{
    struct ServerRequest *sr;

    cmd = EmptyNullStr(cmd);

    sr = allocServerRequest(Server.memPool, userData, func);
    if(sr == NULL)
    {
	return ERR_MEM;
    }

    sr->type = SRT_EVAL;
    sr->contents.eval.handle = handle;
    sr->contents.eval.cmd = AllocString(Server.memPool, cmd);
    if(sr->contents.eval.cmd == NULL)
    {
	freeServerRequest(Server.memPool, sr);
	return ERR_MEM;
    }

    ADDTAIL(&Server.requests, sr);
    Server.checkRequests = TRUE;

    return OK;
}

long ServerAddTransactionRequest(long handle, void *userData,
				 void (*func)(long error, char *result, 
					      void *ud))
{
    struct ServerRequest *sr;

    sr = allocServerRequest(Server.memPool, userData, func);
    if(sr == NULL)
    {
	return ERR_MEM;
    }

    sr->type = SRT_TRANSACTION;
    sr->contents.transaction.handle = handle;

    ADDTAIL(&Server.requests, sr);
    Server.checkRequests = TRUE;

    return OK;
}

long ServerAddCommitRequest(long handle, void *userData,
			   void (*func)(long error, char *result, void *ud))
{
    struct ServerRequest *sr;

    sr = allocServerRequest(Server.memPool, userData, func);
    if(sr == NULL)
    {
	return ERR_MEM;
    }

    sr->type = SRT_COMMIT;
    sr->contents.commit.handle = handle;

    ADDTAIL(&Server.requests, sr);
    Server.checkRequests = TRUE;

    return OK;
}

long ServerAddRollbackRequest(long handle, void *userData,
			     void (*func)(long error, char *result, void *ud))
{
    struct ServerRequest *sr;

    sr = allocServerRequest(Server.memPool, userData, func);
    if(sr == NULL)
    {
	return ERR_MEM;
    }

    sr->type = SRT_ROLLBACK;
    sr->contents.rollback.handle = handle;

    ADDTAIL(&Server.requests, sr);
    Server.checkRequests = TRUE;

    return OK;
}

void ServerCheckRequests(void)
{
    Server.checkRequests = TRUE;
}

static struct ProjectHandle *findProjectHandle(long handle)
{
    struct ProjectHandle *ph;

    FORNODE(ph, &Server.handles)
    {
	if(ph->handle == handle)
	{
	    return ph;
	}
    }

    return NULL; 
}

static bool processConnectRequest(struct ServerRequest *sr)
{
    char *name = sr->contents.connect.projectName;
    bool gui = sr->contents.connect.gui;
    long error = OK;
    struct Project *pr = NULL;
    struct ProjectHandle *ph = NULL;
    static char errstr[1024], result[2048];
    bool isnew = FALSE;

    ph = allocProjectHandle(Server.memPool);
    if(ph == NULL)
    {
	error = ERR_MEM;
    }
    else if(name == NULL || name[0] == 0)
    {
	error = ERR_INVALIDNAME;
    }
    else if((pr = FindProjectName(name)) == NULL)
    {
	/*
	** Project is not present in memory.  Try to load it.
	*/
	pr = AllocProject(FALSE);
	if(pr == NULL)
	{
	    error = ERR_MEM;
	}
	else
	{
	    isnew = TRUE;
	    ADDTAIL(&ProjectList, pr);
	    error = OpenProject(pr, name, gui, errstr); 
	}
    }

    if (error == OK && gui && ProjectStructureEditor(pr) == NULL)
    {
	if(!SetProjectHasGui(pr, TRUE))
	{
	    error = DPYERR_BUILDGUI;
	}
	else if(!OpenProjectWindows(pr))
	{
	   error = DPYERR_OPENGUI;
	} 
    }

    if(error == OK)
    {
	/*
	** Generate new handle id 
	*/
	for(;;) 
	{
	    Server.lastHandle++;
	    if(Server.lastHandle != HANDLE_NONE &&
	       Server.lastHandle != HANDLE_USER)
	    {
		break;
	    }
	}
	ph->handle = Server.lastHandle;
	ph->pr = pr;
	ADDTAIL(&Server.handles, ph);

	sprintf(result, "%ld", ph->handle);
    }
    else
    {
	if (isnew)
	{
	    REMOVE(pr);
	    FreeProject(pr);
	    pr = NULL;
	}
	freeProjectHandle(Server.memPool, ph);
	sprintf(result, GES(error), errstr);
    }

    if(sr->func != NULL)
    {
	sr->func(error, result, sr->userData);
    }

    return TRUE;
}

static bool processDisconnectRequest(struct ServerRequest *sr, 
				     struct ProjectHandle *ph)
{
    long error = OK;

    if(ph == NULL)
    {
	error = SERVERR_HANDLE;
    }
    else
    {
	struct Project *pr = ph->pr;

	if(ProjectExclusiveHandle(pr) == ph->handle)
	{
	    error = SERVERR_TRANSACTION;
	}
	else
	{
	    REMOVE(ph);
	    freeProjectHandle(Server.memPool, ph);

	    /*
	    ** Check if we can free the project.
	    */
	    if(ServerNumProjectConnections(pr) == 0)
	    {
		if(!ProjectHasGui(pr))
		{
		    if (GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL_CLEANUP)){
			ProgramDeleteSourceExternal(pr);
		    }
		    REMOVE(pr);
		    FreeProject(pr);
		}
	    }
	}
    }
    
    if(sr->func != NULL)
    {
	sr->func(error, GES(error), sr->userData);
    }

    return TRUE;
}

static bool processConnectionsRequest(struct ServerRequest *sr) 
{
    long error = OK;
    struct StrHandle *sh = AllocStrHandle(Server.memPool);
    
    if(sh == NULL)
    {
	error = ERR_MEM;
    }
    else 
    {
	if(NumNodes(&Server.handles) == 0)
	{
	    StrHandlePutS(sh, "None");
	}
	else
	{
	    struct ProjectHandle *ph;
	    char buf[20];
	    const char *name;

	    FORNODE(ph, &Server.handles)
	    {
		sprintf(buf, "%4ld ", ph->handle);
		StrHandlePutS(sh, buf);
		name = ProjectName(ph->pr);
		if(name == NULL)
		{
		    name = "NULL";
		}
		StrHandlePutS(sh, name);
		StrHandlePutC(sh, '\n');
	    }
	}
	StrHandlePutC(sh, 0);
	
	if(StrHandleError(sh))
	{
	    error = ERR_MEM;
	}
    }
 
    if(sr->func != NULL)
    {
	sr->func(error, error? GES(error) : StrHandleString(sh), sr->userData);
    }
    
    FreeStrHandle(sh);
    
    return TRUE;
}

static bool checkStartTransaction(struct ProjectHandle *ph)
{
    long exclusive = ProjectExclusiveHandle(ph->pr);
    long changes = ProjectChanges(ph->pr);
    short rc;

    if(exclusive == HANDLE_NONE && changes != 0)
    {
	/*
	** No Transaction is present but 
	** the project needs to be saved.
	*/
	rc = Ask(AT_QUESTION, GCS(MSG_AskProjectServerRequestAnswers), 
		 GCS(MSG_AskProjectServerRequest), ProjectName(ph->pr));
	if(rc == 1)
	{
	    MenuProjectSave(ph->pr, FALSE);
	    changes = ProjectChanges(ph->pr);
	}

	if(changes != 0)
	{
	    SetProjectExclusiveHandle(ph->pr, HANDLE_USER);
	    exclusive = HANDLE_USER;
	}
    }

    if(exclusive == HANDLE_USER && changes != 0)
    {
	return FALSE;              /* wait till the user saved. */
    }
    if(exclusive != HANDLE_NONE && exclusive != HANDLE_USER)
    {
	return FALSE;              /* another transaction blocks */
    }

    return TRUE;
}

static char *evalResult = NULL;
static struct StrHandle *evalStrHandle = NULL;

static bool outfunc(const char *s)
{
    return StrHandlePutS(evalStrHandle, s);
}

const char *setEvalResult(struct Element *elem)
{
    if(elem != NULL)
    {
	evalStrHandle = AllocStrHandle(Server.memPool);
	if(evalStrHandle != NULL)
	{
	    if(ElementToStream(elem, outfunc))
	    {
		StrHandlePutC(evalStrHandle, 0);
		if(!StrHandleError(evalStrHandle))
		{
		    evalResult = AllocString(Server.memPool, 
					     StrHandleString(evalStrHandle));
		}
	    }
	    FreeStrHandle(evalStrHandle);
	    evalStrHandle = NULL;
	}

	if(evalResult == NULL)
	{
	    return GCS(MSG_Err_Mem);
	}
    }

    return NULL;
}

static bool processEvalRequest(struct ServerRequest *sr, 
			       struct ProjectHandle *ph)
{
    long error = OK;
    char *result = NULL, *mem = NULL;
    static const long memsize = 4096;

    mem = PoolAllocMem(Server.memPool, memsize);
    if(mem == NULL)
    {
	error = ERR_MEM;
	result = GES(error);
    }
    else if(ph == NULL)
    {
	error = SERVERR_HANDLE;
	result = GES(error);
    }
    else
    {
	long exclusive = ProjectExclusiveHandle(ph->pr);
	char *cmd = sr->contents.eval.cmd; 
        const char *progerr;
	struct Program *prg = NULL;
	long line = 0, col = 0;
	const char *filename = NULL;

	if(exclusive != ph->handle)
	{
	    if(!checkStartTransaction(ph))
	    {
		return FALSE;
	    }
	}

	/*
	** Evaluate the command.
	*/
	GuiSetSleep(TRUE);
	progerr = CompileQuery(ph->pr, cmd, strlen(cmd), &prg, &line, &col);
	if(progerr != NULL)
	{
	    error = PROGERR_COMPILE;
	    if(line == 1)
		col--;               /* bracket */
	    sprintf(mem, GCS(MSG_Err_CompileQuery), line, col, progerr);
	    result = mem;
	}

	if(error == OK)
	{
	    evalResult = NULL;
	    SetTableProgramRecords(ph->pr);
	    progerr = ExecuteProgram(prg, TRUE, NULL, NULL, TRUE,
				     setEvalResult, &line, &col, &filename);
	    result = evalResult;

	    if(progerr == EXECUTE_ABORT)
	    {
		error = ERR_CANCEL;
		result = GES(error);
	    }
	    else if(progerr != NULL)
	    {
		error = PROGERR_EXECUTE;
		if(line == 1)
		    col--;               /* bracket */
		sprintf(mem, GCS(MSG_Err_ExecuteQuery), line, col, 
			(filename? filename : "local"), progerr);
		result = mem;
	    }
	}
	FreeProgram(prg);

	if(error == OK && exclusive != ph->handle && 
	   ProjectChanges(ph->pr) != 0)
	{
	    /* 
	    ** Auto commit
	    */
	    char *errstr = mem + memsize/2;

	    error = SaveProject(ph->pr, ProjectName(ph->pr), FALSE, errstr);
	    if(error != OK)
	    {
		sprintf(mem, GES(error), errstr);
		result = mem;
	    }
	}
	    
	if(exclusive == HANDLE_USER)
	{
	    SetProjectExclusiveHandle(ph->pr, HANDLE_NONE);
	}

	GuiSetSleep(FALSE);
    }

    if(sr->func != NULL)
    {
	sr->func(error, result, sr->userData);
    }

    FreeString(Server.memPool, evalResult);
    evalResult = NULL;
    PoolFreeMem(Server.memPool, mem, memsize);

    return TRUE;
}

static bool processTransactionRequest(struct ServerRequest *sr, 
				      struct ProjectHandle *ph)
{
    long error = OK;

    if(ph == NULL)
    {
	error = SERVERR_HANDLE;
    }
    else
    {
	long exclusive = ProjectExclusiveHandle(ph->pr);

	if(exclusive == ph->handle)
	{
	    /*
	    ** Nested Transaction call.
	    ** Bounce caller with error message.
	    */
	    error = SERVERR_NESTEDTRANSACTION;
	}
	else if(!checkStartTransaction(ph))
	{
	    return FALSE;
	}
	else
	{
	    SetProjectExclusiveHandle(ph->pr, ph->handle);
	}
    }

    if(sr->func != NULL)
    {
	sr->func(error, GES(error), sr->userData);
    }

    return TRUE;
}

static bool processCommitRequest(struct ServerRequest *sr, 
				 struct ProjectHandle *ph)
{
    long error = OK;
    char *result = GES(error), *mem = NULL;
    const long memsize = 4096;

    if(ph == NULL)
    {
	error = SERVERR_HANDLE;
	result = GES(error);
    }
    else
    {
	long exclusive = ProjectExclusiveHandle(ph->pr);
	
	if(exclusive == HANDLE_NONE)
	{
	    /*
	    ** We are not in a transaction.
	    */
	    error = SERVERR_NOTRANSACTION;
	    result = GES(error);
	}
	else if(exclusive != ph->handle)
	{
	    return FALSE;         /* wait for transation to start */
	}
	else 
	{
	    if(ProjectChanges(ph->pr) != 0)
	    {
		char *errstr;

		mem = PoolAllocMem(Server.memPool, memsize);
		if(mem == NULL)
		{
		    error = ERR_MEM;
		    result = GES(error);
		}
		else
		{
		    errstr = mem + memsize/2;

		    error = SaveProject(ph->pr, ProjectName(ph->pr), FALSE,
                          errstr);
		    if(error != OK)
		    {
			sprintf(mem, GES(error), errstr);
			result = mem;
		    }
		}
	    }
	    
	    SetProjectExclusiveHandle(ph->pr, HANDLE_NONE);
	}
    }

    if(sr->func != NULL)
    {
	sr->func(error, result, sr->userData);
    }

    PoolFreeMem(Server.memPool, mem, memsize);

    return TRUE;
}

static bool processRollbackRequest(struct ServerRequest *sr, 
				   struct ProjectHandle *ph)
{
    long error = OK;
    char *result = GES(error), *mem = NULL;
    const long memsize = 2048;

    if (ph == NULL) {
	error = SERVERR_HANDLE;
	result = GES(error);
    } else {
	long exclusive = ProjectExclusiveHandle(ph->pr);

	if (exclusive == HANDLE_NONE) {
	    /*
	    ** We are not in a transaction.
	    */
	    error = SERVERR_NOTRANSACTION;
	    result = GES(error);
	} else if (exclusive != ph->handle) {
	    return FALSE;         /* wait for transaction to start */
	} else {
	    const char *name = ProjectName(ph->pr);
	    char *errstr;

	    if (ProjectChanges(ph->pr) != 0 && name != NULL) {
		mem = PoolAllocMem(Server.memPool, memsize);
		if (mem == NULL) {
		    error = ERR_MEM;
		    result = GES(error);
		} else {
		    errstr = mem + memsize/2;

		    error = RevertProject(ph->pr, FALSE, errstr);
		    if (error != OK) {
			sprintf(mem, GES(error), errstr);
			result = mem;
		    }
		}
	    }
	    
	    SetProjectExclusiveHandle(ph->pr, HANDLE_NONE);
	}
    }

    if (sr->func != NULL) {
	sr->func(error, result, sr->userData);
    }

    PoolFreeMem(Server.memPool, mem, memsize);

    return TRUE;
}

static bool processRequest(struct ServerRequest *sr)
{
    long handle = serverRequestHandle(sr);
    struct ProjectHandle *ph = NULL;
    long error = OK;

    if(handle != HANDLE_NONE)
    {
	ph = findProjectHandle(handle);
	if(ph == NULL)
	{
	    error = SERVERR_HANDLE;
	}
	else if(FindNodePtr(&ProjectList, ph->pr) == NULL)
	{
	    error = SERVERR_BROKENHANDLE;
	}
    }

    if(error == OK)
    {
	switch(sr->type)
	{
	case SRT_CONNECT:
	    return processConnectRequest(sr);
	case SRT_DISCONNECT:
	    return processDisconnectRequest(sr, ph);
	case SRT_CONNECTIONS:
	    return processConnectionsRequest(sr);
	case SRT_EVAL:
	    return processEvalRequest(sr, ph);
	case SRT_TRANSACTION:
	    return processTransactionRequest(sr, ph);
	case SRT_COMMIT:
	    return processCommitRequest(sr, ph);
	case SRT_ROLLBACK:
	    return processRollbackRequest(sr, ph);
	default:
	    error = ERR_NOTIMPLEMENTED;
	}
    }

    if(sr->func != NULL)
    {
	sr->func(error, GES(error), sr->userData);
    }
    return TRUE;
}

void ServerProcessRequests(void)
{
    if(Server.checkRequests)
    {
	ulong status = GetStatus(NULL);
	struct ServerRequest *sr, *succ;

	if(status & STATUS_ACTIVE)
	{
	    SAFEFORNODE(sr, succ, &Server.requests)
	    {
		if(processRequest(sr))
		{
		    REMOVE(sr);
		    freeServerRequest(Server.memPool, sr);
		    succ = (struct ServerRequest *)HEAD(&Server.requests);
		}
	    }
	}

	Server.checkRequests = FALSE;
    }
}

