/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** server.h
**
** Creation date: 12.01.2001
** Last modified: 12.01.2001
*/

#ifndef SERVER_H
#define SERVER_H

#include "system.h"

struct Project;

#ifdef __cplusplus
extern "C" {
#endif

/*
** Some special project handles.
*/
enum
{
    HANDLE_USER=-1,
    HANDLE_NONE=0,
};

void ServerInit(void);
/*
** Initialization
*/

void ServerExit(void);
/*
** Cleanup
*/

void ServerRemoveProjectConnections(struct Project *pr);
/*
** Removes all handles belonging to the specified project.
*/

long ServerNumProjectConnections(struct Project *pr);
/*
** Returns the number of connections to the given project.
*/

long ServerNumConnections(void);
/*
** Returns total number of connections.
*/

long ServerAddConnectRequest(char *projectName, bool gui, void *userData, 
			     void (*func)(long error, char *result, void *ud));
/*
** Adds a request for connecting to the given project.
** Returns OK if the request could be added or an error code.  
** On success, the specified callback function will be called 
** after the request has been processed.  The function receives 
** an error code, a result string containing a unique handle 
** and some user data you can provide.  If the project does not 
** reside in memory yet, it will be loaded from disk, if gui
** is FALSE then no gui is created for the project.
*/

long ServerAddDisconnectRequest(long handle, void *userData, 
				void (*func)(long error, char *result, 
					     void *ud));
/*
** Adds a request for disconnecting the project refered by handle.
** Returns OK if the request could be added or an error code.  
** On success, the specified callback function will be called 
** after the request has been processed.  The function receives 
** an error code, a result string which is NULL if no error occured,
** and some user data you can provide.  If there is no other reference 
** to the project then the project is closed.  Any changes since the last
** eval or commit command will be lost.
*/

long ServerAddConnectionsRequest(void *userData, 
				 void (*func)(long error, char *result, 
					      void *ud));
/*
** Adds a request for getting a list of connections.
** Returns OK if the request could be added or an error code.  
** On success, the specified callback function will be called 
** after the request has been processed.  The function receives 
** an error code, a result string that contains the open connections,
** and some user data you can provide.  
*/

long ServerAddEvalRequest(long handle, const char *cmd, void *userdata,
			  void (*func)(long error, char *result, void *ud));
/*
** Adds an request for evaluating cmd to the queue of requests
** for the given project handle.  Returns OK if the request
** could added or error code.  On success, the specified 
** callback function will be called after the request has been
** processed.  The function receives an error code,
** the result string and some user data you can provide.
*/

long ServerAddTransactionRequest(long handle, void *userdata,
			   void (*func)(long error, char *result, void *ud));
/*
** Adds an request for starting an exclusive transaction to the queue 
** of requests for the given project handle.  Returns OK if the request
** could added or an error code.  On success, the specified 
** callback function will be called after the request has been
** processed.  The function receives an error code, a result string
** which is NULL if no error occured and some user data you can provide.
*/

long ServerAddCommitRequest(long handle, void *userdata,
			    void (*func)(long error, char *result, void *ud));
/*
** Adds an request for committing an exclusive transaction to the queue 
** of requests for the given project handle.  Returns OK if the request
** could added or an error code.  On success, the specified 
** callback function will be called after the request has been
** processed.  The function receives an error code,  a result string
** which is NULL if no error occured and some user data you can provide.
*/

long ServerAddRollbackRequest(long handle, void *userdata,
			     void (*func)(long error, char *result, void *ud));
/*
** Adds an request for rolling back an exclusive transaction to the queue 
** of requests for the given project handle.  Returns OK if the request
** could added or an error code.  On success, the specified 
** callback function will be called after the request has been
** processed.  The function receives an error code,  a result string
** which is NULL if no error occured and some user data you can provide.
*/

void ServerCheckRequests(void);
/*
** This functions will set a flag such that the server checks
** all of its pending requests on the next ServerProcessRequests call.
*/

void ServerProcessRequests(void);
/*
** Function that should be called on a regular basis for running
** queued requests.
*/
 
#ifdef __cplusplus
}
#endif

#endif

