/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** sqlitedb.c
**
** Creation date: 07.04.2009
** Last modified: 01.08.2022
*/

#include "system.h"
#include "sqlitedb.h"
#include "project.h"
#include "order.h"
#include "table.h"
#include "queryeditor.h"
#include "break.h"
#include "log.h"
#include "preferences.h"
#include "dosutil.h"
#include "util.h"
#include "labels.h"
#include "codeset.h"
#include "strhandle.h"
#include "memo.h"
#include "memory.h"
#include "error.h"
#include "lang/compile.h"
#include "appstrings_cat.h"
#include "version.h"

#ifdef USE_SQLITE_LIBRARY
#include <sqlite3-library.h>
#else
#include <sqlite3.h>
#endif

#define NUM_ENTRIES(array) (sizeof(array) / sizeof(array[0]))

#ifdef PRINT_SQL
#define SQL(s) printf("SQL: %s\n", s)
#else
#define SQL(s)
#endif

struct SQLiteTable
{
    struct MinNode sqlt_Node;
    struct Table *sqlt_Table;
    struct {
        sqlite3_stmt *readRecordData;
        sqlite3_stmt *readRecordOrderNum;
        sqlite3_stmt *insertRecordId;
        sqlite3_stmt *insertRecordData;
        sqlite3_stmt *insertRecordHeader;
        sqlite3_stmt *insertRecordHeaderAndData;
        sqlite3_stmt *updateRecordData;
        sqlite3_stmt *updateRecordHeader;
        sqlite3_stmt *updateRecordHeaderAndData;
        sqlite3_stmt *deleteRecord;
        sqlite3_stmt *deleteRecords;
    } sqlt_Stmts;
    long sqlt_NumFields;
    struct Field *sqlt_Fields[0];
};

void FreeSQLiteTable(struct SQLiteTable *sqlt)
{
    if (sqlt != NULL) {
        struct Table *tab = sqlt->sqlt_Table;
        struct MemPool *memPool = ProjectMemPool(TableProject(tab));

        SetTableSQLiteTable(tab, NULL);

        REMOVE(sqlt);
        sqlite3_finalize(sqlt->sqlt_Stmts.readRecordData);
        sqlt->sqlt_Stmts.readRecordData = NULL;
        sqlite3_finalize(sqlt->sqlt_Stmts.readRecordOrderNum);
        sqlt->sqlt_Stmts.readRecordOrderNum = NULL;
        sqlite3_finalize(sqlt->sqlt_Stmts.insertRecordId);
        sqlt->sqlt_Stmts.insertRecordId = NULL;
        sqlite3_finalize(sqlt->sqlt_Stmts.insertRecordData);
        sqlt->sqlt_Stmts.insertRecordData = NULL;
        sqlite3_finalize(sqlt->sqlt_Stmts.insertRecordHeader);
        sqlt->sqlt_Stmts.insertRecordHeader = NULL;
        sqlite3_finalize(sqlt->sqlt_Stmts.insertRecordHeaderAndData);
        sqlt->sqlt_Stmts.insertRecordHeaderAndData = NULL;
        sqlite3_finalize(sqlt->sqlt_Stmts.updateRecordData);
        sqlt->sqlt_Stmts.updateRecordData = NULL;
        sqlite3_finalize(sqlt->sqlt_Stmts.updateRecordHeader);
        sqlt->sqlt_Stmts.updateRecordHeader = NULL;
        sqlite3_finalize(sqlt->sqlt_Stmts.updateRecordHeaderAndData);
        sqlt->sqlt_Stmts.updateRecordHeaderAndData = NULL;
        sqlite3_finalize(sqlt->sqlt_Stmts.deleteRecord);
        sqlt->sqlt_Stmts.deleteRecord = NULL;
        sqlite3_finalize(sqlt->sqlt_Stmts.deleteRecords);
        sqlt->sqlt_Stmts.deleteRecords = NULL;

        PoolFreeMem(memPool, sqlt, sizeof(struct SQLiteTable) +
                sqlt->sqlt_NumFields * sizeof(struct Field *));
    }
}

struct SQLiteHandle
{
    struct Project *sqlh_Project;
    sqlite3 *sqlh_Db;
    char sqlh_FileChangeCounter[4];
    bool sqlh_HaveFileChangeCounter;
    bool sqlh_HasChanged;
    bool sqlh_KeepLock;
    long sqlh_DataVersion;
    bool (*sqlh_RetryObtainWriteLock)(struct Project *);
    struct {
        sqlite3_stmt *dataVersion;
        sqlite3_stmt *projectVersion;
        sqlite3_stmt *lockingModeExclusive;
        sqlite3_stmt *lockingModeNormal;
    } sqlh_Stmts;
    struct List sqlh_Tables;
};

static void freeSQLiteTableForTable(struct Table *tab)
{
    FreeSQLiteTable(TableSQLiteTable(tab));
}

static void freeSQLiteTables(struct SQLiteHandle *handle)
{
    if (handle != NULL) {
        struct SQLiteTable *sqlt, *succ;

        SAFEFORNODE(sqlt, succ, &handle->sqlh_Tables) {
            FreeSQLiteTable(sqlt);
        }
    }
}

static bool sqlite3_in_transaction(sqlite3 *db)
{
    return sqlite3_get_autocommit(db) == 0;
}

void ClearSQLiteHandle(struct SQLiteHandle *handle)
{
    if (handle != NULL) {
        sqlite3 * db = handle->sqlh_Db;

        if (db != NULL && sqlite3_in_transaction(db)) {
            LogRollback(handle->sqlh_Project);
        }
        sqlite3_finalize(handle->sqlh_Stmts.dataVersion);
        handle->sqlh_Stmts.dataVersion = NULL;
        sqlite3_finalize(handle->sqlh_Stmts.projectVersion);
        handle->sqlh_Stmts.projectVersion = NULL;
        sqlite3_finalize(handle->sqlh_Stmts.lockingModeExclusive);
        handle->sqlh_Stmts.lockingModeExclusive = NULL;
        sqlite3_finalize(handle->sqlh_Stmts.lockingModeNormal);
        handle->sqlh_Stmts.lockingModeNormal = NULL;
        freeSQLiteTables(handle);
        sqlite3_close(handle->sqlh_Db);
        handle->sqlh_Db = NULL;
        handle->sqlh_HaveFileChangeCounter = FALSE;
        handle->sqlh_HasChanged = FALSE;
        handle->sqlh_KeepLock = FALSE;
        handle->sqlh_DataVersion = 0;
    }
}

void FreeSQLiteHandle(struct SQLiteHandle *handle)
{
    if (handle != NULL) {
        struct MemPool *memPool = ProjectMemPool(handle->sqlh_Project);

        ClearSQLiteHandle(handle);
        PoolFreeMem(memPool, handle, sizeof(struct SQLiteHandle));
    }
}

struct SQLiteHandle *AllocSQLiteHandle(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct SQLiteHandle *handle;
   
    handle = PoolAllocMemZero(memPool, sizeof(struct SQLiteHandle));
    if (handle != NULL) {
        handle->sqlh_Project = pr;
        NewList(&handle->sqlh_Tables);
    }

    return handle;
}

static void swapSQLiteHandles(struct SQLiteHandle *handle1,
        struct SQLiteHandle *handle2)
{
    if (handle1 != handle2) {
        struct SQLiteHandle copy1 = *handle1;
    
        *handle1 = *handle2;
        handle1->sqlh_Tables = copy1.sqlh_Tables;
        copy1.sqlh_Tables = handle2->sqlh_Tables;
        *handle2 = copy1;
        SwapLists(&handle1->sqlh_Tables, &handle2->sqlh_Tables);
    }
}

/*
** Map SQLite return codes onto short BeeBase error codes.
*/
static long sqlite3_error(int rc, bool isWriteOperation)
{
    long error;

    switch (rc & 0xFF) {
        case SQLITE_OK:
            error = OK;
            break;
        case SQLITE_BUSY:
        case SQLITE_PROTOCOL:
            /* Change lock is handled in SQLiteObtainChangeLock */
            error = isWriteOperation?
                PROJECTERR_WRITELOCK : PROJECTERR_READLOCK;
            break;
        case SQLITE_LOCKED:
            error = PROJECTERR_WRITELOCK;
            break;
        case SQLITE_NOMEM:
            error = ERR_MEM;
            break;
        case SQLITE_READONLY:
            error = PROJECTERR_READONLY;
            break;
        case SQLITE_IOERR:
            switch (rc) {
                case SQLITE_IOERR_RDLOCK:
                    error = PROJECTERR_READLOCK;
                    break;

                case SQLITE_IOERR_BLOCKED:
                    error = PROJECTERR_WRITELOCK;
                    break;

                case SQLITE_IOERR_NOMEM:
                    error = ERR_MEM;
                    break;

                case SQLITE_IOERR_READ:
                case SQLITE_IOERR_SHORT_READ:
                case SQLITE_IOERR_WRITE:
                default:
                    error = SQLITEERR_IO;
                    break;
            }
	    break;
        case SQLITE_FULL:
        case SQLITE_CANTOPEN:
            error = SQLITEERR_IO;
            break;
        case SQLITE_CORRUPT:
        case SQLITE_NOTADB:
            error = DISKERR_INVALID;
            break;
        default:
            error = SQLITEERR_GENERIC;
            break;
    }

    return error;
}

bool SQLiteAllowChanges(struct SQLiteHandle *handle)
{
    bool allow = handle == NULL ||
        handle->sqlh_Db == NULL ||
        sqlite3_in_transaction(handle->sqlh_Db);

    return allow;
}

bool SQLiteHasReservedLock(struct SQLiteHandle *handle)
{
    bool reserved = FALSE;
    sqlite3 *db = (handle != NULL)? handle->sqlh_Db : NULL;
    sqlite3_file *file = NULL;
    int rc;

    if (db != NULL &&
            sqlite3_file_control(db, NULL,
                SQLITE_FCNTL_FILE_POINTER, &file) == SQLITE_OK &&
            file != NULL && file->pMethods != NULL && 
            file->pMethods->xCheckReservedLock != NULL &&
            file->pMethods->xCheckReservedLock(file, &rc) == SQLITE_OK) {
        reserved = rc;
    }
    return reserved;
}

static void updateDataVersion(struct SQLiteHandle *handle, long *version)
{
    if (handle != NULL && handle->sqlh_Db != NULL) {
        sqlite3_stmt *stmt = handle->sqlh_Stmts.dataVersion;
        int rc;

        if (stmt == NULL) {
            rc = sqlite3_prepare_v2(handle->sqlh_Db, "PRAGMA data_version", -1,
                    &stmt, NULL);
            handle->sqlh_Stmts.dataVersion = stmt;
        }
        if (stmt != NULL) {
            rc = sqlite3_step(stmt);
            if (rc == SQLITE_ROW) {
                *version = sqlite3_column_int(stmt, 0);
            }
            rc = sqlite3_reset(stmt);
        }
    }
}

bool SQLiteProjectHasChanged(struct SQLiteHandle *handle)
{
    bool changed = FALSE;

    if (handle != NULL) {
        if (!handle->sqlh_HasChanged) {
            long oldDataVersion = handle->sqlh_DataVersion;

            updateDataVersion(handle, &handle->sqlh_DataVersion);
            if (oldDataVersion != handle->sqlh_DataVersion) {
                handle->sqlh_HasChanged = TRUE;
            }
        }
        changed = handle->sqlh_HasChanged;
    }

    return changed;
}

static bool updateFileChangeCounter(struct SQLiteHandle *handle,
        char fileChangeCounter[4])
{
    sqlite3 *db = (handle != NULL)? handle->sqlh_Db : NULL;
    sqlite3_file *file = NULL;
    char writeReadVersion[2];

    /* Only read the file change counter when sqlite3 uses legacy format. */
    return db != NULL && sqlite3_file_control(db, NULL,
            SQLITE_FCNTL_FILE_POINTER, &file) == SQLITE_OK &&
        file != NULL && file->pMethods != NULL && 
        file->pMethods->xRead != NULL &&
        file->pMethods->xRead(file, writeReadVersion, 2, 18) == SQLITE_OK &&
        writeReadVersion[0] == 1 && writeReadVersion[1] == 1 &&
        file->pMethods->xRead(file, fileChangeCounter, 4, 24) == SQLITE_OK;
}


bool SQLiteCheckForProjectChange(struct SQLiteHandle *handle)
{
    char fileChangeCounter[4];

    return handle == NULL || handle->sqlh_HasChanged ||
        !handle->sqlh_HaveFileChangeCounter ||
        !updateFileChangeCounter(handle, fileChangeCounter) || 
        memcmp(fileChangeCounter, handle->sqlh_FileChangeCounter, 4) != 0;
}

bool SQLiteGetProjectFileVersion(struct SQLiteHandle *handle, long *version)
{
    bool success = FALSE;

    if (handle != NULL && handle->sqlh_Db != NULL) {
        sqlite3_stmt *stmt = handle->sqlh_Stmts.projectVersion;
        int rc;

        if (stmt == NULL) {
            rc = sqlite3_prepare_v2(handle->sqlh_Db,
                    "SELECT Version FROM _Project", -1, &stmt, NULL);
            handle->sqlh_Stmts.projectVersion = stmt;
        }
        if (stmt != NULL) {
            rc = sqlite3_step(stmt);
            if (rc == SQLITE_ROW) {
                if (version != NULL) {
                    *version = sqlite3_column_int(stmt, 0);
                }
                success = TRUE;
            }
            rc = sqlite3_reset(stmt);
        }
    }

    return success;
}

void SQLiteReleaseLock(struct SQLiteHandle *handle)
{
    if (handle != NULL && handle->sqlh_KeepLock) {
        sqlite3 *db = handle->sqlh_Db;
        sqlite3_stmt *stmt = handle->sqlh_Stmts.lockingModeNormal;
        int rc = SQLITE_OK;

        if (stmt == NULL) {
            rc = sqlite3_prepare_v2(db,
                    "PRAGMA locking_mode = NORMAL", -1, &stmt, NULL);
            handle->sqlh_Stmts.lockingModeNormal = stmt;
        }
        if (stmt != NULL) {
            do {
                rc = sqlite3_step(stmt);
            } while (rc == SQLITE_ROW);
            if (rc == SQLITE_DONE) {
                rc = SQLITE_OK;
            }
            if (rc == SQLITE_OK) {
                /* Read something to complete the locking mode change */
                if (SQLiteGetProjectFileVersion(handle, NULL)) {
                    handle->sqlh_KeepLock = FALSE;
                }
            }
            sqlite3_reset(stmt);
        }
    }
}

static long sqliteKeepLock(struct SQLiteHandle *handle, bool checkChanged)
{
    if (handle != NULL && !handle->sqlh_KeepLock) {
        sqlite3 *db = handle->sqlh_Db;
        sqlite3_stmt *stmt = handle->sqlh_Stmts.lockingModeExclusive;
        int rc = SQLITE_OK;
        long err = OK;

        if (db == NULL) {
            return OK;
        }
        if (checkChanged && SQLiteProjectHasChanged(handle)) {
            return PROJECTERR_VERSION;
        }
        if (stmt == NULL) {
            rc = sqlite3_prepare_v2(db,
                    "PRAGMA locking_mode = EXCLUSIVE", -1, &stmt, NULL);
            handle->sqlh_Stmts.lockingModeExclusive = stmt;
        }
        if (stmt != NULL) {
            do {
                rc = sqlite3_step(stmt);
            } while (rc == SQLITE_ROW);
            if (rc == SQLITE_DONE) {
                rc = SQLITE_OK;
            }
            err = sqlite3_error(rc, FALSE);
            sqlite3_reset(stmt);
        }
        if (err != OK) {
            return err;
        }
        handle->sqlh_KeepLock = TRUE;
        if (checkChanged && SQLiteProjectHasChanged(handle)) {
            SQLiteReleaseLock(handle);
            return PROJECTERR_VERSION;
        }
    }

    return OK;
}

long SQLiteObtainChangeLock(struct SQLiteHandle *handle)
{
    long error = OK;

    if (handle != NULL) {
        sqlite3 *db = handle->sqlh_Db;
        int rc;

        if (db != NULL && !sqlite3_in_transaction(db)) {
            if (SQLiteProjectHasChanged(handle)) {
                error = PROJECTERR_VERSION;
            }
            if (error == OK) {
                rc = sqlite3_exec(db, "BEGIN IMMEDIATE", NULL, NULL, NULL);
                error = sqlite3_error(rc, TRUE);
                if (error == PROJECTERR_WRITELOCK) {
                    error = PROJECTERR_CHANGELOCK;  /* Actual error */
                }
                if (error == OK) {
                    if (SQLiteProjectHasChanged(handle)) {
                        rc = sqlite3_exec(db, "ROLLBACK", NULL, NULL, NULL);
                        error = PROJECTERR_VERSION;
                    } else if (!sqlite3_in_transaction(db)) {
                        error = ERR_INTERNAL;
                    }
                }
                if (error == OK) {
                    LogBeginTransaction(handle->sqlh_Project);
                }
            }
        }
    }
    return error;
}

static const char *sqlite3_type_for_field_type(short type)
{
    const char *sqlite3_type = NULL;

    switch (type) {
        case TYPE_BOOL:
        case TYPE_INTEGER: 
        case TYPE_DATE:
        case TYPE_TIME:
        case TYPE_CHOICE:
        case TYPE_REFERENCE:
            sqlite3_type = "INT";
            break;

        case TYPE_REAL:
            sqlite3_type = "REAL";
            break;

        case TYPE_STRING:
        case TYPE_MEMO: 
            sqlite3_type = "TEXT";
            break;

        case TYPE_VIRTUAL:
        case TYPE_BUTTON:
        default:
            break;
    }

    return sqlite3_type;
}

static const char tmpNamePrefix[] = "_New_";

struct SQLiteColumn {
    const char *name;
    const char *type;
    bool required;
    const char *value;
    bool key;
};

static const char projectTable[] = "_Project";

static const struct SQLiteColumn projectColumns[] =
{
    { "Format", "INT", TRUE, NULL },
    { "SavedWith", "TEXT", FALSE, "generic" },
    { "Version", "INT", TRUE, NULL },
    { "StructureVersion", "INT", FALSE, "0" },
    { "AdminPassword", "TEXT", FALSE, NULL },
    { "ProjectPrefs", "TEXT", FALSE, NULL },
    { "Program", "TEXT", FALSE, NULL },
    { "DisplayActiveTableID", "INT", FALSE, NULL },
    { "DisplayActiveObjectID", "INT", FALSE, NULL }
};

static const char tableTable[] = "_Tables";

static const struct SQLiteColumn tableColumns[] =
{
    { "ID", "INTEGER", TRUE, NULL, TRUE },
    { "Name", "TEXT", TRUE, NULL },
    { "ExactlyOneRecord", "INT", FALSE, NULL },
    { "FilterActive", "INT", FALSE, NULL },
    { "NewTrigger", "TEXT", FALSE, NULL },
    { "DeleteTrigger", "TEXT", FALSE, NULL },
    { "CountChanges", "INT", FALSE, "1" },
    { "LogChanges", "INT", FALSE, "1" },
    { "GuiRecordID", "INT", FALSE, NULL },
    { "OrderSpec", "TEXT", FALSE, NULL },
    { "FilterProgram", "TEXT", FALSE, NULL }
};

static const char fieldTable[] = "_Fields";

static const struct SQLiteColumn fieldColumns[] =
{
    { "ID", "INTEGER", TRUE, NULL, TRUE },
    { "TableID", "INT", TRUE, NULL },
    { "Name", "TEXT", TRUE, NULL },
    { "Type", "TEXT", TRUE, NULL },
    { "Trigger", "TEXT", FALSE, NULL },
    { "CountChanges", "INT", FALSE, "1" },
    { "LogChanges", "INT", FALSE, "1" },
    { "InitialValue", NULL, FALSE, NULL },
    { "NilString", "TEXT", FALSE, NULL },
    { "StringMaxlen", "INT", FALSE, "100" },
    { "ChoiceLabels", "TEXT", FALSE, NULL },
    { "ReferenceTableID", "INT", FALSE, NULL },
    { "ReferenceFilter", "INT", FALSE, NULL },
    { "ReferenceAutoShow", "INT", FALSE, NULL },
    { "VirtualCompute", "TEXT", FALSE, NULL }
};

static const char displayObjectTable[] = "_DisplayObjects";

static const struct SQLiteColumn displayObjectColumns[] =
{
    { "ID", "INTEGER", TRUE, NULL, TRUE },
    { "ParentID", "INT", TRUE, NULL },
    { "Visible", "INT", FALSE, "1" },
    { "Title", "TEXT", FALSE, NULL },
    { "Shortcut", "TEXT", FALSE, NULL },
    { "Enabled", "TEXT", FALSE, "Not initial" },
    { "EnabledCompute", "TEXT", FALSE, NULL },
    { "QtFontname", "TEXT", FALSE, NULL },
    { "GTKFontname", "TEXT", FALSE, NULL },
    { "MUIFontname", "TEXT", FALSE, NULL },
    { "Background", "TEXT", FALSE, NULL },
    { "Weight", "INT", FALSE, "100" },
    { "Type", "TEXT", TRUE, NULL }
};

static const char displayTableTable[] = "_DisplayTables";

static const struct SQLiteColumn displayTableColumns[] =
{
    { "ID", "INTEGER", TRUE, NULL, TRUE },
    { "TableID", "INT", TRUE, NULL },
    { "ActiveFieldID", "INT", FALSE, NULL },
    { "ActiveDisplayObjectID", "INT", FALSE, NULL },
    { "HasPanel", "INT", FALSE, NULL },
    { "PanelBackground", "TEXT", FALSE, NULL },
    { "PanelNumAll", "INT", FALSE, NULL },
    { "PanelFilter", "INT", FALSE, NULL },
    { "PanelArrows", "INT", FALSE, NULL }
};

static const char displayFieldTable[] = "_DisplayFields";

static const struct SQLiteColumn displayFieldColumns[] =
{
    { "ID", "INTEGER", TRUE, NULL, TRUE },
    { "FieldID", "INT", TRUE, NULL },
    { "TitlePos", "TEXT", FALSE, NULL },
    { "TitleQtFontname", "TEXT", FALSE, NULL },
    { "TitleGTKFontname", "TEXT", FALSE, NULL },
    { "TitleMUIFontname", "TEXT", FALSE, NULL },
    { "Home", "INT", FALSE, NULL },
    { "TabChain", "INT", FALSE, "1" },
    { "ReadOnly", "INT", FALSE, NULL },
    { "Kind", "TEXT", FALSE, NULL },
    { "Alignment", "TEXT", FALSE, NULL },
    { "Calendar", "INT", FALSE, "1" },
    { "Format", "TEXT", FALSE, NULL },
    { "NumDecimals", "INT", FALSE, "2" },
    { "ImageSize", "TEXT", FALSE, NULL },
    { "BubbleHelp", "TEXT", FALSE, NULL },
    { "DisplayExternal", "INT", FALSE, NULL },
    { "TitleAtStringField", "INT", FALSE, NULL },
    { "HideStringField", "INT", FALSE, NULL },
    { "FilePopup", "INT", FALSE, NULL },
    { "FontPopup", "INT", FALSE, NULL },
    { "ListviewPopup", "INT", FALSE, NULL },
    { "ListviewLabels", "TEXT", FALSE, NULL },
    { "ListviewCompute", "TEXT", FALSE, NULL },
    { "View", "INT", FALSE, NULL },
    { "Frame", "INT", FALSE, NULL },
    { "RecordsKind", "TEXT", FALSE, NULL },
    { "RecordsCompute", "TEXT", FALSE, NULL },
    { "InitialRecord", "INT", FALSE, NULL },
    { "CurrentRecord", "INT", FALSE, NULL },
    { "Show", "INT", FALSE, NULL },
    { "AutoShow", "INT", FALSE, NULL },
    { "Filter", "INT", FALSE, NULL },
    { "AutoUpdate", "INT", FALSE, NULL },
    { "ShowTitles", "INT", FALSE, NULL },
    { "ControlTable", "INT", FALSE, NULL },
    { "OnDoubleClick", "TEXT", FALSE, NULL },
    { "DoubleClickTrigger", "TEXT", FALSE, NULL },
    { "URLDropTrigger", "TEXT", FALSE, NULL },
    { "SortDropTrigger", "TEXT", FALSE, NULL }
};

static const char displaySpaceTable[] = "_DisplaySpaces";

static const struct SQLiteColumn displaySpaceColumns[] =
{
    { "ID", "INTEGER", TRUE, NULL, TRUE },
    { "Bar", "INT", FALSE, NULL }
};

static const char displayImageTable[] = "_DisplayImages";

static const struct SQLiteColumn displayImageColumns[] =
{
    { "ID", "INTEGER", TRUE, NULL, TRUE },
    { "ImageSize", "TEXT", FALSE, NULL }
};

static const char displayRegisterTable[] = "_DisplayRegisters";

static const struct SQLiteColumn displayRegisterColumns[] =
{
    { "ID", "INTEGER", TRUE, NULL, TRUE },
    { "Labels", "TEXT", FALSE, NULL },
    { "Open", "INT", FALSE, NULL }
};

static const char displayGroupTable[] = "_DisplayGroups";

static const struct SQLiteColumn displayGroupColumns[] =
{
    { "ID", "INTEGER", TRUE, NULL, TRUE },
    { "Border", "INT", FALSE, NULL },
    { "Horizontal", "INT", FALSE, NULL },
    { "Spacing", "INT", FALSE, NULL },
    { "Open", "INT", FALSE, NULL }
};

static const char displayWindowTable[] = "_DisplayWindows";

static const struct SQLiteColumn displayWindowColumns[] =
{
    { "ID", "INTEGER", TRUE, NULL, TRUE },
    { "WindowId", "INT", FALSE, NULL },
    { "Button", "INT", FALSE, NULL },
    { "CloseWhenDisabled", "INT", FALSE, NULL },
    { "WindowOpen", "INT", FALSE, NULL },
    { "LeftEdge", "INT", FALSE, NULL },
    { "TopEdge", "INT", FALSE, NULL },
    { "Width", "INT", FALSE, NULL },
    { "Height", "INT", FALSE, NULL },
    { "Open", "INT", FALSE, NULL }
};

static const char recordDescriptionSpecTable[] = "_RecordDescriptionSpecs";

static const struct SQLiteColumn recordDescriptionSpecColumns[] =
{
    { "ID", "INTEGER", TRUE, NULL, TRUE },
    { "Type", "INT", FALSE, NULL },
    { "Compute", "TEXT", FALSE, NULL },
    { "MultiColumn", "INT", FALSE, NULL },
    { "Fields", "TEXT", FALSE, NULL }
};

static const char queryTable[] = "_Queries";

static const struct SQLiteColumn queryColumns[] =
{
    { "Name", "TEXT", FALSE, NULL },
    { "TableID", "INT", FALSE, NULL },
    { "Program", "TEXT", FALSE, NULL }
};

static const struct SQLiteColumn recordColumns[] =
{
    { "_ID", "INTEGER", TRUE, NULL, TRUE },     /* must be first entry */
    { "_Version", "INT", TRUE, NULL },          /* must be 2nd entry */
    { "_MatchFilter", "INT", FALSE, "1" },      /* must be 2nd last entry */
    { "_OrderNum", "REAL", TRUE, NULL }         /* must be last entry */
};

static void strHandlePutQ(struct StrHandle *sh, const char *s)
{
    StrHandlePutC(sh, '\'');
    if (s != NULL) {
        while (*s) {
            if (*s == '\'') {
                StrHandlePutC(sh, '\'');
            }
            StrHandlePutC(sh, *s);
            s++;
        }
    }
    StrHandlePutC(sh, '\'');
}

static int createTable(sqlite3 *db, struct MemPool *memPool, const char *table,
        bool quotes, const struct SQLiteColumn *columns, size_t num, long *perr)
{
    struct StrHandle *sh = AllocStrHandle(memPool);
    int rc = SQLITE_OK;
    size_t i;

    StrHandlePutS(sh, "DROP TABLE IF EXISTS ");
    if (quotes) {
        StrHandlePutC(sh, '"');
    }
    StrHandlePutS(sh, table);
    if (quotes) {
        StrHandlePutC(sh, '"');
    }
    StrHandlePutS(sh, ";\n");
    StrHandlePutS(sh, "CREATE TABLE ");
    if (quotes) {
        StrHandlePutC(sh, '"');
    }
    StrHandlePutS(sh, table);
    if (quotes) {
        StrHandlePutC(sh, '"');
    }
    StrHandlePutS(sh, " (");
    for (i = 0; i < num; i++) {
        if (i > 0) {
            StrHandlePutS(sh, ", ");
        }
        StrHandlePutS(sh, columns[i].name);
        if (columns[i].type != NULL) {
            StrHandlePutC(sh, ' ');
            StrHandlePutS(sh, columns[i].type);
        }
        if (columns[i].key) {
            StrHandlePutS(sh, " PRIMARY KEY");
        }
    }
    StrHandlePutS(sh, ")");
    StrHandlePutC(sh, 0);
    if (StrHandleError(sh)) {
        *perr = ERR_MEM;
        rc = SQLITE_ERROR;
    } else {
        SQL(StrHandleString(sh));
        rc = sqlite3_exec(db, StrHandleString(sh), NULL, NULL, NULL);
    }
    FreeStrHandle(sh);

    return rc;
}

long SQLiteDeleteTable(struct SQLiteHandle *handle, struct Table *tab)
{
    if (handle != NULL && tab != NULL) {
        sqlite3 *db = handle->sqlh_Db;
        struct Project *pr = handle->sqlh_Project;
        struct MemPool *memPool = ProjectMemPool(pr);
        struct StrHandle *sh;
        long err = OK;
        int rc = SQLITE_OK;

        if (db == NULL) {
            return OK;
        }
        if (!sqlite3_in_transaction(db)) {
            return ERR_INTERNAL;
        }

        sh = AllocStrHandle(memPool);
        if (sh == NULL) {
            return ERR_MEM;
        }
        StrHandlePutS(sh, "DROP TABLE IF EXISTS ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else {
            SQL(StrHandleString(sh));
            rc = sqlite3_exec(db, StrHandleString(sh), NULL, NULL, NULL);
            err = sqlite3_error(rc, TRUE);
        }
        FreeStrHandle(sh);

        return err;
    }
    return OK;
}

static long createOrderIndex(sqlite3 *db, struct Table *tab)
{
    long err = OK;

    if (tab != NULL) {
        struct MemPool *memPool = TableMemPool(tab);
        struct StrHandle *sh = NULL;
        int entries = NUM_ENTRIES(recordColumns), i;
        int rc = SQLITE_OK;

        sh = AllocStrHandle(memPool);
        if (sh == NULL) {
            return ERR_MEM;
        }
        StrHandlePutS(sh, "DROP INDEX IF EXISTS ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutS(sh, "OrderIndex");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, ";\n");
        StrHandlePutS(sh, "CREATE INDEX ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutS(sh, "OrderIndex");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " ON ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " (");
        /*
        ** We add all record columns except _ID starting with _OrderNum.
        ** This basically creates a copy of these columns and makes loading
        ** a database much faster since SQLite only needs to read this index
        ** for reading the record header.
        */
        for (i = entries - 1; i > 0; i--) {
            if (i != entries - 1) {
                StrHandlePutS(sh, ", ");
            }
            StrHandlePutS(sh, recordColumns[i].name);
        }
        StrHandlePutS(sh, ")");
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else {
            rc = sqlite3_exec(db, StrHandleString(sh), NULL, NULL, NULL);
        }
        FreeStrHandle(sh);
        if (err == OK) {
            err = sqlite3_error(rc, TRUE);
        }
    }
    return err;
}

long SQLiteNewTable(struct SQLiteHandle *handle, struct Table *tab)
{
    if (tab == NULL) {
        return ERR_INTERNAL;
    }

    if (handle != NULL) {
        sqlite3 *db = handle->sqlh_Db;
        struct Project *pr = handle->sqlh_Project;
        struct MemPool *memPool = ProjectMemPool(pr);
        long err = OK;
        int rc = SQLITE_OK;

        if (db == NULL) {
            return OK;
        }
        if (!sqlite3_in_transaction(db)) {
            return ERR_INTERNAL;
        }

        rc = createTable(db, memPool, TableName(tab), TRUE,
                recordColumns, NUM_ENTRIES(recordColumns), &err);
        if (err == OK) {
            err = sqlite3_error(rc, TRUE);
        }
        if (err != OK) {
            return err;
        }

        createOrderIndex(db, tab);      /* ignore error */
    }
    return OK;
}

/* Renames a table in the SQLite connection. */
long SQLiteRenameTable(struct SQLiteHandle *handle, struct Table *tab,
        const char *newName)
{
    if (tab == NULL || newName == NULL) {
        return ERR_INTERNAL;
    }

    if (handle != NULL) {
        sqlite3 *db = handle->sqlh_Db;
        struct Project *pr = handle->sqlh_Project;
        struct MemPool *memPool = ProjectMemPool(pr);
        struct StrHandle *sh;
        long err = OK;
        int rc = SQLITE_OK;

        freeSQLiteTableForTable(tab);

        if (db == NULL) {
            return OK;
        }
        if (!sqlite3_in_transaction(db)) {
            return ERR_INTERNAL;
        }

        sh = AllocStrHandle(memPool);
        if (sh == NULL) {
            return ERR_MEM;
        }
        StrHandlePutS(sh, "ALTER TABLE ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " RENAME TO ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, newName);
        StrHandlePutC(sh, '"');
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else {
            SQL(StrHandleString(sh));
            rc = sqlite3_exec(db, StrHandleString(sh), NULL, NULL, NULL);
            err = sqlite3_error(rc, TRUE);
        }
        FreeStrHandle(sh);

        return err;
    }
    return OK;
}

static long sqliteRecreateTable(struct SQLiteHandle *handle, struct Table *tab,
        struct Field *target, const char *newNameOrDelete)
{
    if (handle != NULL) {
        sqlite3 *db = handle->sqlh_Db;
        struct Project *pr = handle->sqlh_Project;
        struct MemPool *memPool = ProjectMemPool(pr);
        struct StrHandle *sh;
        struct List *fields = TableFields(tab);
        struct Field *field;
        long err = OK;
        int rc = SQLITE_OK;
        long i;

        if (db == NULL) {
            return OK;
        }
        if (fields == NULL) {
            return ERR_INTERNAL;
        }
        sh = AllocStrHandle(memPool);
        if (sh == NULL) {
            return ERR_MEM;
        }
        StrHandlePutS(sh, "DROP TABLE IF EXISTS ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, tmpNamePrefix);
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, ";\n");
        StrHandlePutS(sh, "CREATE TABLE ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, tmpNamePrefix);
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " (");
        for (i = 0; i < NUM_ENTRIES(recordColumns); i++) {
            if (i > 0) {
                StrHandlePutS(sh, ", ");
            }
            StrHandlePutS(sh, recordColumns[i].name);
            if (recordColumns[i].type != NULL) {
                StrHandlePutC(sh, ' ');
                StrHandlePutS(sh, recordColumns[i].type);
            }
            if (recordColumns[i].key) {
                StrHandlePutS(sh, " PRIMARY KEY");
            }
        }
        FORNODE(field, fields) {
            const char *type = sqlite3_type_for_field_type(FieldType(field));

            if (type == NULL) {
                continue;
            }
            if (field == target && newNameOrDelete == NULL) {
                continue;
            }
            StrHandlePutS(sh, ", ");
            StrHandlePutC(sh, '"');
            StrHandlePutS(sh,
                    field == target?  newNameOrDelete : FieldName(field));
            StrHandlePutC(sh, '"');
            StrHandlePutC(sh, ' ');
            StrHandlePutS(sh, type);
        }
        StrHandlePutS(sh, ")");
        StrHandlePutS(sh, ";\n");
        StrHandlePutS(sh, "INSERT INTO ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, tmpNamePrefix);
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " SELECT ");
        for (i = 0; i < NUM_ENTRIES(recordColumns); i++) {
            if (i > 0) {
                StrHandlePutS(sh, ", ");
            }
            StrHandlePutS(sh, recordColumns[i].name);
        }
        FORNODE(field, fields) {
            const char *type = sqlite3_type_for_field_type(FieldType(field));

            if (type == NULL) {
                continue;
            }
            if (field == target && newNameOrDelete == NULL) {
                continue;
            }
            StrHandlePutS(sh, ", ");
            StrHandlePutC(sh, '"');
            StrHandlePutS(sh, FieldName(field));
            StrHandlePutC(sh, '"');
        }
        StrHandlePutS(sh, " FROM ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, ";\n");
        StrHandlePutS(sh, "DROP TABLE ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, ";\n");
        StrHandlePutS(sh, "ALTER TABLE ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, tmpNamePrefix);
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " RENAME TO ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else {
            SQL(StrHandleString(sh));
            rc = sqlite3_exec(db, StrHandleString(sh), NULL, NULL, NULL);
            err = sqlite3_error(rc, TRUE);
        }
        FreeStrHandle(sh);

        createOrderIndex(db, tab);      /* ignore error */

        return err;
    }
    return OK;
}

long SQLiteDeleteField(struct SQLiteHandle *handle, struct Field *field)
{
    const char *sqlite3_type = sqlite3_type_for_field_type(FieldType(field));

    if (sqlite3_type == NULL) {
        return OK;
    }

    freeSQLiteTableForTable(FieldTable(field));

    /* 
     ** Currently it is not possible to delete a column in SQLite3.
     ** We re-create the table instead.
     */
    return sqliteRecreateTable(handle, FieldTable(field), field, NULL);
}

long SQLiteNewField(struct SQLiteHandle *handle, struct Field *field)
{
    const char *sqlite3_type = sqlite3_type_for_field_type(FieldType(field));

    if (field == NULL) {
        return ERR_INTERNAL;
    }
    if (sqlite3_type == NULL) {
        return OK;
    }

    if (handle != NULL) {
        sqlite3 *db = handle->sqlh_Db;
        struct Project *pr = handle->sqlh_Project;
        struct MemPool *memPool = ProjectMemPool(pr);
        struct Table *tab = FieldTable(field);
        const struct DataInfo *di = FieldDataInfo(field);
        struct StrHandle *sh;
        long err = OK;
        int rc = SQLITE_OK;

        freeSQLiteTableForTable(tab);

        if (db == NULL) {
            return OK;
        }
        if (!sqlite3_in_transaction(db)) {
            return ERR_INTERNAL;
        }

        sh = AllocStrHandle(memPool);
        if (sh == NULL) {
            return ERR_MEM;
        }
        StrHandlePutS(sh, "ALTER TABLE ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " ADD COLUMN ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, FieldName(field));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " ");
        StrHandlePutS(sh, sqlite3_type);
        if (di->di_Flags & DIF_NIL) {
            StrHandlePutS(sh, " DEFAULT NULL");
        } else {
            switch (FieldType(field)) {
                case TYPE_INTEGER:
                    StrHandlePutS(sh, " DEFAULT ");
                    StrHandlePutS(sh, IntToStr(di->di_Contents.integer));
                    break;
                case TYPE_REAL:
                    StrHandlePutS(sh, " DEFAULT ");
                    StrHandlePutS(sh, DoubleToStr(di->di_Contents.real, -1, 0));
                    break;
                case TYPE_BOOL:
                    StrHandlePutS(sh, " DEFAULT 1");
                    break;
                case TYPE_STRING:
                    StrHandlePutS(sh, " DEFAULT ");
                    if (di->di_Contents.string == NULL) {
                        StrHandlePutS(sh, " NULL");
                    } else {
                        char *utf8 = CodesetConvertToUTF8(
                                di->di_Contents.string, memPool, &err);

                        strHandlePutQ(sh, utf8);
                        if (utf8 != di->di_Contents.string) {
                            FreeString(memPool, utf8);
                        }
                    }
                    break;
                case TYPE_MEMO: 
                    StrHandlePutS(sh, " DEFAULT ");
                    if (di->di_Contents.mi == NULL) {
                        StrHandlePutS(sh, " NULL");
                    } else {
                        char *utf8 = CodesetConvertToUTF8(di->di_Contents.mi,
                                memPool, &err);

                        strHandlePutQ(sh, utf8);
                        if (utf8 != di->di_Contents.mi) {
                            FreeString(memPool, utf8);
                        }
                    }
                    break;
                case TYPE_DATE:
                    StrHandlePutS(sh, " DEFAULT ");
                    StrHandlePutS(sh, IntToStr(di->di_Contents.date));
                    break;
                case TYPE_TIME:
                    StrHandlePutS(sh, " DEFAULT ");
                    StrHandlePutS(sh, IntToStr(di->di_Contents.time));
                    break;
                case TYPE_CHOICE:
                    StrHandlePutS(sh, " DEFAULT ");
                    StrHandlePutS(sh, IntToStr(di->di_Contents.choice));
                    break;
                case TYPE_REFERENCE:
                    StrHandlePutS(sh, " DEFAULT NULL");
                    break;
            }
        }
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else {
            SQL(StrHandleString(sh));
            rc = sqlite3_exec(db, StrHandleString(sh), NULL, NULL, NULL);
            err = sqlite3_error(rc, TRUE);
        }
        FreeStrHandle(sh);

        return err;
    }
    return OK;
}

long SQLiteRenameField(struct SQLiteHandle *handle, struct Field *field,
        const char *newName)
{
    const char *sqlite3_type = sqlite3_type_for_field_type(FieldType(field));

    if (field == NULL || newName == NULL) {
        return ERR_INTERNAL;
    }
    if (sqlite3_type == NULL) {
        return OK;
    }

    if (handle != NULL) {
        sqlite3 *db = handle->sqlh_Db;
        struct Project *pr = handle->sqlh_Project;
        struct MemPool *memPool = ProjectMemPool(pr);
        struct Table *tab = FieldTable(field);
        struct StrHandle *sh;
        long err = OK;
        int rc = SQLITE_OK;

        freeSQLiteTableForTable(tab);

        if (db == NULL) {
            return OK;
        }
        if (!sqlite3_in_transaction(db)) {
            return ERR_INTERNAL;
        }

        sh = AllocStrHandle(memPool);
        if (sh == NULL) {
            return ERR_MEM;
        }
        StrHandlePutS(sh, "ALTER TABLE ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " RENAME COLUMN ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, FieldName(field));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " TO ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, newName);
        StrHandlePutC(sh, '"');
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else {
            SQL(StrHandleString(sh));
            rc = sqlite3_exec(db, StrHandleString(sh), NULL, NULL, NULL);
            err = sqlite3_error(rc, TRUE);
        }
        FreeStrHandle(sh);

        if (err == OK || err != SQLITEERR_GENERIC) {
            return err;
        }
        /*
        ** Older SQLite3 versions do not understand the RENAME COLUMN command.
        ** In that case we re-create the table instead.
        */
        return sqliteRecreateTable(handle, tab, field, newName);
    }
    return OK;
}

struct TypeString
{
    short type;
    const char *str;
};

static const char *typeString(const struct TypeString *ts, size_t num,
        short type, const char *defaultString)
{
    int i;

    for (i = 0; i < num; i++) {
        if (ts[i].type == type) {
            return ts[i].str;
        }
    }
    return defaultString;
}

static short typeFromString(const struct TypeString *ts, size_t num,
        const char *str, short defaultType)
{
    int i;

    if (str != NULL) {
        for (i = 0; i < num; i++) {
            if (strcasecmp(ts[i].str, str) == 0) {
                return ts[i].type;
            }
        }
    }
    return defaultType;
}

static const struct TypeString fieldTypeStrings[] =
{
    { TYPE_BOOL, "Boolean" },
    { TYPE_INTEGER, "Integer" },
    { TYPE_REAL, "Real" },
    { TYPE_DATE, "Date" },
    { TYPE_TIME, "Time" },
    { TYPE_CHOICE, "Choice" },
    { TYPE_STRING, "String" },
    { TYPE_MEMO, "Memo" },
    { TYPE_REFERENCE, "Reference" },
    { TYPE_VIRTUAL, "Virtual" },
    { TYPE_BUTTON, "Button" }
};

static const char *fieldTypeString(short type)
{
    return typeString(fieldTypeStrings, NUM_ENTRIES(fieldTypeStrings), type,
            NULL);
}

static short fieldTypeFromString(const char *str)
{
    return typeFromString(fieldTypeStrings, NUM_ENTRIES(fieldTypeStrings), str,
            TYPE_UNKNOWN);
}

static const struct TypeString displayObjectTypeStrings[] =
{
    { DO_TYPE_TABLE, "Table" },
    { DO_TYPE_FIELD, "Field" },
    { DO_TYPE_GROUP, "Group" },
    { DO_TYPE_WINDOW, "Window" },
    { DO_TYPE_SPACE, "Space" },
    { DO_TYPE_TEXT, "Text" },
    { DO_TYPE_IMAGE, "Image" },
    { DO_TYPE_BALANCE, "Balance" },
    { DO_TYPE_REGISTER, "Register" }
};

static const char *displayObjectTypeString(short type)
{
    return typeString(displayObjectTypeStrings,
            NUM_ENTRIES(displayObjectTypeStrings), type, NULL);
}

static short displayObjectTypeFromString(const char *str)
{
    return typeFromString(displayObjectTypeStrings,
            NUM_ENTRIES(displayObjectTypeStrings), str, 0);
}

static const struct TypeString displayObjectEnabledStrings[] =
{
    { DOD_ENABLED_ALWAYS, "Always" },
    { DOD_ENABLED_NOTINITIAL, "Not initial" },
    { DOD_ENABLED_COMPUTED, "Computed" }
};

static const char *displayObjectEnabledString(short type, int enabled)
{
    switch (type)
    {
        case DO_TYPE_FIELD:
        case DO_TYPE_WINDOW:
            return typeString(displayObjectEnabledStrings,
                    NUM_ENTRIES(displayObjectEnabledStrings), enabled, NULL);
    }
    return NULL;
}

static short displayObjectEnabledFromString(const char *str)
{
    return typeFromString(displayObjectEnabledStrings,
            NUM_ENTRIES(displayObjectEnabledStrings), str, DOD_ENABLED_ALWAYS);
}

static const struct TypeString displayImageSizeStrings[] =
{
    { DAD_IMAGESIZE_RESIZABLE, "Resizable" },
    { DAD_IMAGESIZE_FIXED, "Fixed" },
    { DAD_IMAGESIZE_SCROLLABLE, "Scrollable" },
    { DAD_IMAGESIZE_SCALED, "Scaled" },
    { DAD_IMAGESIZE_ASPECTSCALED, "Aspect-scaled" }
};

static const char *displayImageSizeString(int size)
{
    return typeString(displayImageSizeStrings, 
            NUM_ENTRIES(displayImageSizeStrings), size, NULL);
}

static short displayImageSizeFromString(const char *str)
{
    return typeFromString(displayImageSizeStrings,
            NUM_ENTRIES(displayImageSizeStrings), str, DAD_IMAGESIZE_RESIZABLE);
}

static const struct TypeString displayFieldTitlePosStrings[] =
{
    { DAD_TITLEPOSITION_LEFT, "Left" },
    { DAD_TITLEPOSITION_RIGHT, "Right" },
    { DAD_TITLEPOSITION_TOP, "Top" },
    { DAD_TITLEPOSITION_BOTTOM, "Bottom" }
};

static const char *displayFieldTitlePosString(int pos)
{
    return typeString(displayFieldTitlePosStrings,
            NUM_ENTRIES(displayFieldTitlePosStrings), pos, NULL);
}

static short displayFieldTitlePosFromString(const char *str)
{
    return typeFromString(displayFieldTitlePosStrings,
            NUM_ENTRIES(displayFieldTitlePosStrings), str,
            DAD_TITLEPOSITION_LEFT);
}

static const struct TypeString displayChoiceFieldKindStrings[] =
{
    { DAD_CHOICETYPE_CYCLEBUTTON, "Cycle" },
    { DAD_CHOICETYPE_RADIOBUTTONS, "Radio" }
};

static const struct TypeString displayVirtualFieldKindStrings[] =
{
    { DAD_VIRTUALTYPE_BOOL, "Boolean" },
    { DAD_VIRTUALTYPE_TEXT, "Text" },
    { DAD_VIRTUALTYPE_LIST, "List" }
};

static const struct TypeString displayButtonFieldKindStrings[] =
{
    { DAD_BUTTONTYPE_TEXT, "Text" },
    { DAD_BUTTONTYPE_IMAGE, "Image" }
};

static const char *displayFieldKindString(short type, int kind)
{
    switch (type)
    {
        case TYPE_CHOICE:
            return typeString(displayChoiceFieldKindStrings,
                    NUM_ENTRIES(displayChoiceFieldKindStrings), kind, NULL);
        case TYPE_VIRTUAL:
            return typeString(displayVirtualFieldKindStrings,
                    NUM_ENTRIES(displayVirtualFieldKindStrings), kind, NULL);
        case TYPE_BUTTON:
            return typeString(displayButtonFieldKindStrings,
                    NUM_ENTRIES(displayButtonFieldKindStrings), kind, NULL);
    }
    return NULL;
}

static short displayFieldKindFromString(short type, const char *str)
{
    switch (type)
    {
        case TYPE_CHOICE:
            return typeFromString(displayChoiceFieldKindStrings,
                    NUM_ENTRIES(displayChoiceFieldKindStrings), str,
                    DAD_CHOICETYPE_CYCLEBUTTON);
        case TYPE_VIRTUAL:
            return typeFromString(displayVirtualFieldKindStrings,
                    NUM_ENTRIES(displayVirtualFieldKindStrings), str,
                    DAD_VIRTUALTYPE_TEXT);
        case TYPE_BUTTON:
            return typeFromString(displayButtonFieldKindStrings,
                    NUM_ENTRIES(displayButtonFieldKindStrings), str,
                    DAD_BUTTONTYPE_TEXT);
    }
    return 0;
}

static const struct TypeString displayChoiceFieldAlignmentStrings[] =
{
    { DAD_ALIGNMENT_VERTICAL, "Vertical" },
    { DAD_ALIGNMENT_HORIZONTAL, "Horizontal" }
};

static const struct TypeString displayFieldAlignmentStrings[] =
{
    { DAD_ALIGNMENT_CENTER, "Center" },
    { DAD_ALIGNMENT_LEFT, "Left" },
    { DAD_ALIGNMENT_RIGHT, "Right" }
};

static const char *displayFieldAlignmentString(short type, int align)
{
    switch (type) {
        case TYPE_CHOICE:
            return typeString(displayChoiceFieldAlignmentStrings,
                    NUM_ENTRIES(displayChoiceFieldAlignmentStrings), align,
                    NULL);
        default:
            return typeString(displayFieldAlignmentStrings,
                    NUM_ENTRIES(displayFieldAlignmentStrings), align, NULL);
    }
    return NULL;
}

static short displayFieldAlignmentFromString(short type, const char *str)
{
    switch (type) {
        case TYPE_CHOICE:
            return typeFromString(displayChoiceFieldAlignmentStrings,
                    NUM_ENTRIES(displayChoiceFieldAlignmentStrings), str,
                    DAD_ALIGNMENT_VERTICAL);
        default:
            return typeFromString(displayFieldAlignmentStrings,
                    NUM_ENTRIES(displayFieldAlignmentStrings), str,
                    DAD_ALIGNMENT_CENTER);
    }
    return 0;
}

static const struct TypeString displayFieldFormatStrings[] =
{
    { FORMAT_HOURMINSEC, "HH:MM:SS" },
    { FORMAT_MINSEC, "MM:SS" },
    { FORMAT_HOURMIN, "HH:MM" }
};

static const char *displayFieldFormatString(short type, int format)
{
    switch (type) {
        case TYPE_DATE:
            return typeString(displayFieldFormatStrings,
                    NUM_ENTRIES(displayFieldFormatStrings), format, NULL);
    }
    return NULL;
}

static short displayFieldFormatFromString(const char *str)
{
    return typeFromString(displayFieldFormatStrings,
            NUM_ENTRIES(displayFieldFormatStrings), str, 0);
}

static const struct TypeString displayFieldRecordsKindStrings[] =
{
    { DAD_RECORDS_ALL, "All" },
    { DAD_RECORDS_MATCHFILTER, "Match filter" },
    { DAD_RECORDS_COMPUTED, "Computed" }
};

static const char *displayFieldRecordsKindString(int kind)
{
    return typeString(displayFieldRecordsKindStrings,
            NUM_ENTRIES(displayFieldRecordsKindStrings), kind, NULL);
}

static short displayFieldRecordsKindFromString(const char *str)
{
    return typeFromString(displayFieldRecordsKindStrings,
            NUM_ENTRIES(displayFieldRecordsKindStrings), str, DAD_RECORDS_ALL);
}

static const struct TypeString displayFieldOnDoubleClickStrings[] =
{
    { DAD_DOUBLECLICK_DONOTHING, "Do nothing" },
    { DAD_DOUBLECLICK_SHOWRECORD, "Show record" },
    { DAD_DOUBLECLICK_CALLTRIGGER, "Call trigger" }
};

static const char *displayFieldOnDoubleClickString(int click)
{
    return typeString(displayFieldOnDoubleClickStrings,
            NUM_ENTRIES(displayFieldOnDoubleClickStrings), click, NULL);
}

static short displayFieldOnDoubleClickFromString(const char *str)
{
    return typeFromString(displayFieldOnDoubleClickStrings,
            NUM_ENTRIES(displayFieldOnDoubleClickStrings), str,
            DAD_DOUBLECLICK_DONOTHING);
}

static const struct TypeString recordDescriptionSpecTypeStrings[] =
{
    { RDS_TYPE_FIELDS, "Fields" },
    { RDS_TYPE_COMPUTED, "Computed" },
    { RDS_TYPE_INHERITED, "Call trigger" }
};

static const char *recordDescriptionSpecTypeString(short type)
{
    return typeString(recordDescriptionSpecTypeStrings,
            NUM_ENTRIES(recordDescriptionSpecTypeStrings), type, NULL);
}

static short recordDescriptionSpecTypeFromString(const char *str)
{
    return typeFromString(recordDescriptionSpecTypeStrings,
            NUM_ENTRIES(recordDescriptionSpecTypeStrings), str,
            RDS_TYPE_FIELDS);
}

static int sqlite3_bind_codeset(sqlite3_stmt *stmt, int pos, const char *str,
        struct MemPool *memPool, bool copy, long *perr)
{
    int rc = SQLITE_OK;
    char *utf8;

    if (str != NULL) {
        utf8 = CodesetConvertToUTF8(str, memPool, perr);
        if (*perr != OK) {
            rc = SQLITE_ERROR;
        } else if (utf8 == str) {
            rc = sqlite3_bind_text(stmt, pos, str, -1, 
                    copy ? SQLITE_TRANSIENT : SQLITE_STATIC);
        } else {
            rc = sqlite3_bind_text(stmt, pos, utf8, -1, SQLITE_TRANSIENT);
            FreeString(memPool, utf8);
        }
    }

    return rc;
}

static int sqlite3_bind_labels(sqlite3_stmt *stmt, int pos, char **labels,
        struct MemPool *memPool, long *perr)
{
    int rc = SQLITE_OK;
    char *str;

    if (labels != NULL && labels[0] != NULL) {
        str = LabelsToString(memPool, labels);
        if (str == NULL) {
            *perr = ERR_MEM;
            rc = SQLITE_ERROR;
        } else {
            rc = sqlite3_bind_codeset(stmt, pos, str, memPool, TRUE, perr);
            FreeString(memPool, str);
        }
    }
    return rc;
}

static int sqlite3_bind_program(sqlite3_stmt *stmt, int pos,
        struct Program *prg, struct MemPool *memPool, long *perr)
{
    int rc = SQLITE_OK;
    struct SourceText *st;

    if (prg != NULL) {
        st = GetSourceText(prg);
        if (st == NULL) {
            *perr = ERR_MEM;
            rc = SQLITE_ERROR;
        } else {
            rc = sqlite3_bind_codeset(stmt, pos, SourceTextText(st), memPool,
                    TRUE, perr);
            FreeSourceText(st);
        }
    }
    return rc;
}

static int sqlite3_bind_datainfo(sqlite3_stmt *stmt, int pos, short type,
        const struct DataInfo *di, struct MemPool *memPool, long *perr)
{
    int rc = SQLITE_OK;

    if (di->di_Flags & DIF_NIL) {
        rc = sqlite3_bind_null(stmt, pos);
    } else {
        switch (type) {
            case TYPE_INTEGER:
                rc = sqlite3_bind_int(stmt, pos, di->di_Contents.integer);
                break;
            case TYPE_REAL:
                rc = sqlite3_bind_double(stmt, pos, di->di_Contents.real);
                break;
            case TYPE_BOOL:
                rc = sqlite3_bind_int(stmt, pos, 1);
                break;
            case TYPE_STRING:
                rc = sqlite3_bind_codeset(stmt, pos, di->di_Contents.string,
                        memPool, FALSE, perr);
                break;
            case TYPE_MEMO: 
                rc = sqlite3_bind_codeset(stmt, pos, di->di_Contents.mi,
                        memPool, FALSE, perr);
                break;
            case TYPE_DATE:
                rc = sqlite3_bind_int(stmt, pos, di->di_Contents.date);
                break;
            case TYPE_TIME:
                rc = sqlite3_bind_int(stmt, pos, di->di_Contents.time);
                break;
            case TYPE_CHOICE:
                rc = sqlite3_bind_int(stmt, pos, di->di_Contents.choice);
                break;
            case TYPE_REFERENCE:
                rc = sqlite3_bind_int(stmt, pos,
                        RecordInternalNum(di->di_Contents.rec));
                break;
        }
    }

    return rc;
}

#define PROJECT_FORMAT 1
#define MIN_ORDER_NUM_RESOLUTION (1.0 / 4096.0)

static int makeInsertStatement(sqlite3 *db, struct MemPool *memPool,
        const char *table, size_t num, long *perr, sqlite3_stmt **stmt)
{
    struct StrHandle *sh = AllocStrHandle(memPool);
    int rc = SQLITE_OK;
    size_t i;

    StrHandlePutS(sh, "INSERT INTO ");
    StrHandlePutS(sh, table);
    StrHandlePutS(sh, " VALUES(");
    for (i = 0; i < num; i++) {
        if (i > 0) {
            StrHandlePutS(sh, ", ");
        }
        StrHandlePutC(sh, '?');
    }
    StrHandlePutS(sh, ")");
    StrHandlePutC(sh, 0);
    if (StrHandleError(sh)) {
        *perr = ERR_MEM;
        rc = SQLITE_ERROR;
    } else {
        rc = sqlite3_prepare_v2(db, StrHandleString(sh), -1, stmt, NULL);
    }
    FreeStrHandle(sh);

    return rc;
}

static long sqlite3WriteProject(sqlite3 *db, struct Project *pr,
        long projectVersion, long structureVersion)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct ProjectPrefs *pp = ProjectProjectPrefs(pr);
    char *settings = GetProjectPrefsSettings(pp, memPool);
    sqlite3_stmt *stmt = NULL;
    char verstr[64];
    long err = OK;
    int rc;

    if (settings == NULL) {
	return ERR_MEM;
    }

    rc = createTable(db, memPool, projectTable, FALSE, projectColumns,
            NUM_ENTRIES(projectColumns), &err);
    if (rc == SQLITE_OK) {
        rc = makeInsertStatement(db, memPool, projectTable,
                NUM_ENTRIES(projectColumns), &err, &stmt);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_int(stmt, 1, PROJECT_FORMAT);
    }
    if (rc == SQLITE_OK) {
        sprintf(verstr, "V %s (%s)", BB_VERSTR, OS_TYPE);
        rc = sqlite3_bind_text(stmt, 2, verstr, -1, SQLITE_STATIC);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_int(stmt, 3, projectVersion);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_int(stmt, 4, structureVersion);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_text(stmt, 5, ProjectAdminPassword(pr), -1,
                SQLITE_STATIC);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_codeset(stmt, 6, settings, memPool, FALSE, &err);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_program(stmt, 7, ProjectProgram(pr), memPool, &err);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_int(stmt, 8, TableNum(DisplayProjectActiveTable(pr)));
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_int(stmt, 9, 
                DisplayObjectNum(DisplayProjectActiveDisplayObject(pr)));
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_step(stmt);
        if (rc == SQLITE_DONE) {
            rc = SQLITE_OK;
        }
    }
    sqlite3_finalize(stmt);
    FreeString(memPool, settings);

    if (err == OK) {
        err = sqlite3_error(rc, TRUE);
    }
    return err;
}

static long sqlite3WriteTables(sqlite3 *db, struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    char *order = NULL;
    sqlite3_stmt *stmt = NULL;
    long err = OK;
    int rc;

    if (tabs == NULL) {
	return ERR_INTERNAL;
    }

    rc = createTable(db, memPool, tableTable, FALSE, tableColumns,
            NUM_ENTRIES(tableColumns), &err);
    if (rc == SQLITE_OK) {
        rc = makeInsertStatement(db, memPool, tableTable,
                NUM_ENTRIES(tableColumns), &err, &stmt);
    }
    FORNODE(tab, tabs) {
        if (rc == SQLITE_OK) {
            rc = sqlite3_reset(stmt);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_clear_bindings(stmt);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_int(stmt, 1, TableNum(tab));
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_text(stmt, 2, TableName(tab),-1, SQLITE_STATIC);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_int(stmt, 3, TableExactlyOneRecord(tab));
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_int(stmt, 4, TableFilterActive(tab));
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_text(stmt, 5, TableNewTrigger(tab), -1,
                    SQLITE_STATIC);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_text(stmt, 6, TableDelTrigger(tab), -1,
                    SQLITE_STATIC);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_int(stmt, 7, TableCountChanges(tab));
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_int(stmt, 8, TableLogChanges(tab));
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_int(stmt, 9, 
                    RecordInternalNum(TableGuiRecord(tab)));
        }
        if (rc == SQLITE_OK) {
            order = GetOrderStr(memPool, tab);
            if (order == NULL) {
                err = ERR_MEM;
                break;
            }
            rc = sqlite3_bind_text(stmt, 10, order, -1, SQLITE_STATIC);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_program(stmt, 11, TableFilterProgram(tab),
                    memPool, &err);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_step(stmt);
            if (rc == SQLITE_DONE) {
                rc = SQLITE_OK;
            }
        }
        FreeString(memPool, order);
	BreakAddCurrentStep(1);
        if (err || rc) {
            break;
        }
    }
    sqlite3_finalize(stmt);
    if (err == OK) {
        err = sqlite3_error(rc, TRUE);
    }
    return err;
}

static long sqlite3WriteFields(sqlite3 *db, struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct List *tabs = ProjectTables(pr);
    struct List *fields;
    struct Table *tab;
    struct Field *field;
    sqlite3_stmt *stmt = NULL;
    long err = OK;
    int rc;

    if (tabs == NULL) {
	return ERR_INTERNAL;
    }

    rc = createTable(db, memPool, fieldTable, FALSE, fieldColumns,
            NUM_ENTRIES(fieldColumns), &err);
    if (rc == SQLITE_OK) {
        rc = makeInsertStatement(db, memPool, fieldTable,
                NUM_ENTRIES(fieldColumns), &err, &stmt);
    }
    FORNODE(tab, tabs) {
        fields = TableFields(tab);
        if (fields == NULL) {
            continue;
        }
        FORNODE(field, fields) {
            const struct DataInfo *di = FieldDataInfo(field);
            short type = FieldType(field);

            if (rc == SQLITE_OK) {
                rc = sqlite3_reset(stmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_clear_bindings(stmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(stmt, 1, FieldNum(field));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(stmt, 2, TableNum(tab));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(stmt, 3, FieldName(field), -1, 
                        SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(stmt, 4, fieldTypeString(type), -1,
                        SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(stmt, 5, FieldTrigger(field), -1,
                        SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(stmt, 6, FieldCountChanges(field));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(stmt, 7, FieldLogChanges(field));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_datainfo(stmt, 8, type, di, memPool, &err);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_codeset(stmt, 9, FieldNilString(field),
                        memPool, FALSE, &err);
            }

            switch (type) {
                case TYPE_STRING:
                    if (rc == SQLITE_OK) {
                        rc = sqlite3_bind_int(stmt, 10, 
                                FieldStringMaxlen(field));
                    }
                    break;

                case TYPE_CHOICE:
                    if (rc == SQLITE_OK) {
                        rc = sqlite3_bind_labels(stmt, 11,
                                FieldChoiceLabels(field), memPool, &err);
                    }
                    break;


                case TYPE_REFERENCE:
                    if (rc == SQLITE_OK) {
                        rc = sqlite3_bind_int(stmt, 12,
                                TableNum(FieldReferenceTable(field)));
                    }
                    if (rc == SQLITE_OK) {
                        rc = sqlite3_bind_int(stmt, 13,
                                FieldReferenceFilter(field));
                    }
                    if (rc == SQLITE_OK) {
                        rc = sqlite3_bind_int(stmt, 14, 
                                FieldReferenceAutoShow(field));
                    }
                    break;

                case TYPE_VIRTUAL:
                    if (rc == SQLITE_OK) {
                        rc = sqlite3_bind_text(stmt, 15, 
                                FieldVirtualCompute(field), -1,
                                SQLITE_STATIC);
                    }
                    break;

                case TYPE_DATE:
                case TYPE_TIME:
                case TYPE_INTEGER:
                case TYPE_REAL:
                case TYPE_MEMO:
                case TYPE_BOOL:
                case TYPE_BUTTON:
                    break;

                default:
                    err = FIELDERR_UNKNOWNTYPE; 
                    break;
            }
            if (rc == SQLITE_OK && err == OK) {
                rc = sqlite3_step(stmt);
                if (rc == SQLITE_DONE) {
                    rc = SQLITE_OK;
                }
            }
            if (rc != SQLITE_OK || err != OK) {
                break;
            }
            BreakAddCurrentStep(1);
        }
        if (rc != SQLITE_OK || err != OK) {
            break;
        }
    }
    sqlite3_finalize(stmt);
    if (err == OK) {
        err = sqlite3_error(rc, TRUE);
    }

    return err;
}

struct WriteDioStmts
{
    sqlite3_stmt *dioStmt; 
    sqlite3_stmt *dioTableStmt;
    sqlite3_stmt *dioFieldStmt;
    sqlite3_stmt *dioSpaceStmt;
    sqlite3_stmt *dioImageStmt;
    sqlite3_stmt *dioRegisterStmt;
    sqlite3_stmt *dioGroupStmt;
    sqlite3_stmt *dioWindowStmt;
    sqlite3_stmt *recDescSpecStmt;
};

static long sqlite3WriteDisplayObjectsRec(sqlite3 *db,
        struct DisplayObject *dio, struct WriteDioStmts *stmts)
{
    struct Project *pr = DisplayObjectProject(dio);
    struct MemPool *memPool = ProjectMemPool(pr);
    struct DisplayObject *parent = DisplayObjectParent(dio);
    ulong num = DisplayObjectNum(dio);
    short type = DisplayObjectType(dio);
    char buf[CodesetMaxBytesForChar + 1];
    struct List *childs = DisplayObjectObjects(dio);
    sqlite3_stmt *dioStmt = NULL;
    sqlite3_stmt *typeStmt = NULL;
    const struct RecordDescriptionSpec *recordDescriptionSpec = NULL;
    long err = OK;
    int rc = SQLITE_OK;

    memset(buf, 0, sizeof(buf));
    CodesetCharToStr(DisplayObjectShortcut(dio), buf);

    if (stmts->dioStmt == NULL) {
        rc = createTable(db, memPool, displayObjectTable, FALSE,
                displayObjectColumns, NUM_ENTRIES(displayObjectColumns), &err);
        if (rc == SQLITE_OK) {
            rc = makeInsertStatement(db, memPool, displayObjectTable,
                NUM_ENTRIES(displayObjectColumns), &err, &stmts->dioStmt);
        }
    }
    dioStmt = stmts->dioStmt;

    if (rc == SQLITE_OK) {
        rc = sqlite3_reset(dioStmt);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_clear_bindings(dioStmt);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_int(dioStmt, 1, num);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_int(dioStmt, 2, parent? DisplayObjectNum(parent):0);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_int(dioStmt, 3, DisplayObjectVisible(dio));
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_codeset(dioStmt, 4, DisplayObjectTitle(dio),
                memPool, FALSE, &err);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_codeset(dioStmt, 5, buf, memPool, FALSE, &err);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_text(dioStmt, 6, 
                displayObjectEnabledString(type, DisplayObjectEnabled(dio)), 
                -1, SQLITE_STATIC);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_text(dioStmt, 7, DisplayObjectEnabledCompute(dio),
                -1, SQLITE_STATIC);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_codeset(dioStmt, 8, DisplayObjectQtFontname(dio),
                memPool, FALSE, &err);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_codeset(dioStmt, 9, DisplayObjectGTKFontname(dio),
                memPool, FALSE, &err);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_codeset(dioStmt, 10, DisplayObjectMUIFontname(dio),
                memPool, FALSE, &err);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_codeset(dioStmt, 11, DisplayObjectBackground(dio),
                memPool, FALSE, &err);
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_int(dioStmt, 12, DisplayObjectWeight(dio));
    }
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_text(dioStmt, 13, displayObjectTypeString(type), -1,
                SQLITE_STATIC);
    }

    switch(type)
    {
        case DO_TYPE_TABLE:
        {
            struct Table *tab = DisplayObjectTable(dio);

            if (rc == SQLITE_OK && stmts->dioTableStmt == NULL) {
                rc = createTable(db, memPool, displayTableTable, FALSE,
                        displayTableColumns, NUM_ENTRIES(displayTableColumns),
                        &err);
                if (rc == SQLITE_OK) {
                    rc = makeInsertStatement(db, memPool, displayTableTable,
                            NUM_ENTRIES(displayTableColumns), &err,
                            &stmts->dioTableStmt);
                }
            }
            typeStmt = stmts->dioTableStmt;
            if (rc == SQLITE_OK) {
                rc = sqlite3_reset(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_clear_bindings(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 1, num);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 2, TableNum(tab));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 3,
                        FieldNum(DisplayTableActiveField(tab)));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 4,
                        DisplayObjectNum(DisplayTableActiveDisplayObject(tab)));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 5, DisplayTableHasPanel(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_codeset(typeStmt, 6,
                        DisplayTablePanelBackground(dio), memPool, FALSE, &err);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 7,
                        DisplayTablePanelNumAll(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 8,
                        DisplayTablePanelFilter(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 9,
                        DisplayTablePanelArrows(dio));
            }
            recordDescriptionSpec = DisplayTableRecordDescriptionSpec(dio);
            break;
        }

        case DO_TYPE_FIELD:
        {
            struct Field *field = DisplayFieldField(dio);
            short fieldType = FieldType(field);

            if (rc == SQLITE_OK && stmts->dioFieldStmt == NULL) {
                rc = createTable(db, memPool, displayFieldTable, FALSE,
                        displayFieldColumns, NUM_ENTRIES(displayFieldColumns),
                        &err);
                if (rc == SQLITE_OK) {
                    rc = makeInsertStatement(db, memPool, displayFieldTable,
                            NUM_ENTRIES(displayFieldColumns), &err,
                            &stmts->dioFieldStmt);
                }
            }
            typeStmt = stmts->dioFieldStmt;
            if (rc == SQLITE_OK) {
                rc = sqlite3_reset(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_clear_bindings(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 1, num);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 2, FieldNum(field));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 3,
                        displayFieldTitlePosString(
                            DisplayFieldTitlePosition(dio)), -1,
                        SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_codeset(typeStmt, 4,
                        DisplayFieldTitleQtFontname(dio), memPool, FALSE, &err);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_codeset(typeStmt, 5,
                        DisplayFieldTitleGTKFontname(dio), memPool, FALSE,&err);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_codeset(typeStmt, 6,
                        DisplayFieldTitleMUIFontname(dio), memPool, FALSE,&err);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 7, DisplayFieldHome(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 8,
                        DisplayFieldTabChain(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 9,
                        DisplayFieldReadOnly(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 10, 
                        displayFieldKindString(fieldType,
                            DisplayFieldSubType(dio)), -1, SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 11, 
                        displayFieldAlignmentString(fieldType,
                            DisplayFieldAlignment(dio)), -1, SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 12, DisplayFieldCalendar(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 13, 
                        displayFieldFormatString(fieldType,
                            DisplayFieldFormat(dio)), -1, SQLITE_STATIC);
            }
            if (fieldType == TYPE_REAL || fieldType == TYPE_VIRTUAL) {
                if (rc == SQLITE_OK) {
                    rc = sqlite3_bind_int(typeStmt, 14,
                            DisplayFieldNumDecimals(dio));
                }
            }
            if (fieldType == TYPE_STRING || fieldType == TYPE_BUTTON) {
                if (rc == SQLITE_OK) {
                    rc = sqlite3_bind_text(typeStmt, 15, 
                            displayImageSizeString(
                                DisplayFieldImageSize(dio)), -1,
                            SQLITE_STATIC);
                }
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_codeset(typeStmt, 16,
                        DisplayFieldBubbleHelp(dio), memPool, FALSE, &err);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 17,
                        DisplayFieldDisplayExternal(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 18,
                        DisplayFieldTitleAtStringField(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 19,
                        DisplayFieldHideStringField(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 20, DisplayFieldFile(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 21, DisplayFieldFont(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 22,
                        DisplayFieldListview(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_labels(typeStmt, 23,
                        DisplayFieldListviewLabels(dio), memPool, &err);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 24, 
                        DisplayFieldListviewCompute(dio), -1,
                        SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 25,DisplayFieldView(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 26,
                        DisplayFieldFrame(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 27, 
                        displayFieldRecordsKindString(
                            DisplayFieldRecordsKind(dio)), -1,
                        SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 28, 
                        DisplayFieldRecordsCompute(dio), -1, SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 29,
                        DisplayFieldInitialRecord(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 30,
                        DisplayFieldCurrentRecord(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 31,DisplayFieldShow(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 32,
                    DisplayFieldAutoShow(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 33,
                        DisplayFieldFilter(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 34,
                        DisplayFieldAutoUpdate(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 35,
                        DisplayFieldShowTitles(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 36,
                        DisplayFieldControlTable(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 37, 
                        displayFieldOnDoubleClickString(
                            DisplayFieldOnDoubleClick(dio)), -1,
                        SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 38, 
                        DisplayFieldDoubleClickTrigger(dio), -1,
                        SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 39, 
                        DisplayFieldURLDropTrigger(dio), -1,
                        SQLITE_STATIC);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 40, 
                        DisplayFieldSortDropTrigger(dio), -1,
                        SQLITE_STATIC);
            }
            recordDescriptionSpec = 
                DisplayFieldRecordDescriptionSpec(dio, FALSE);
            break;
        }

        case DO_TYPE_SPACE:
            if (rc == SQLITE_OK && stmts->dioSpaceStmt == NULL) {
                rc = createTable(db, memPool, displaySpaceTable, FALSE,
                        displaySpaceColumns, NUM_ENTRIES(displaySpaceColumns),
                        &err);
                if (rc == SQLITE_OK) {
                    rc = makeInsertStatement(db, memPool, displaySpaceTable,
                            NUM_ENTRIES(displaySpaceColumns), &err,
                            &stmts->dioSpaceStmt);
                }
            }
            typeStmt = stmts->dioSpaceStmt;
            if (rc == SQLITE_OK) {
                rc = sqlite3_reset(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_clear_bindings(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 1, num);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 2, DisplaySpaceBar(dio));
            }
            break;

        case DO_TYPE_IMAGE:
            if (rc == SQLITE_OK && stmts->dioImageStmt == NULL) {
                rc = createTable(db, memPool, displayImageTable, FALSE,
                        displayImageColumns, NUM_ENTRIES(displayImageColumns),
                        &err);
                if (rc == SQLITE_OK) {
                    rc = makeInsertStatement(db, memPool, displayImageTable,
                            NUM_ENTRIES(displayImageColumns), &err,
                            &stmts->dioImageStmt);
                }
            }
            typeStmt = stmts->dioImageStmt;
            if (rc == SQLITE_OK) {
                rc = sqlite3_reset(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_clear_bindings(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 1, num);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_text(typeStmt, 2, displayImageSizeString(
                            DisplayImageImageSize(dio)), -1, SQLITE_STATIC);
            }
            break;

        case DO_TYPE_REGISTER:
            if (rc == SQLITE_OK && stmts->dioRegisterStmt == NULL) {
                rc = createTable(db, memPool, displayRegisterTable, FALSE,
                        displayRegisterColumns,
                        NUM_ENTRIES(displayRegisterColumns), &err);
                if (rc == SQLITE_OK) {
                    rc = makeInsertStatement(db, memPool, displayRegisterTable,
                            NUM_ENTRIES(displayRegisterColumns), &err,
                            &stmts->dioRegisterStmt);
                }
            }
            typeStmt = stmts->dioRegisterStmt;
            if (rc == SQLITE_OK) {
                rc = sqlite3_reset(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_clear_bindings(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 1, num);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_labels(typeStmt, 2, 
                        DisplayRegisterLabels(dio), memPool, &err);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 3, DisplayObjectOpen(dio));
            }
            break;

        case DO_TYPE_GROUP:
            if (rc == SQLITE_OK && stmts->dioGroupStmt == NULL) {
                rc = createTable(db, memPool, displayGroupTable, FALSE,
                        displayGroupColumns, NUM_ENTRIES(displayGroupColumns),
                        &err);
                if (rc == SQLITE_OK) {
                    rc = makeInsertStatement(db, memPool, displayGroupTable,
                            NUM_ENTRIES(displayGroupColumns), &err,
                            &stmts->dioGroupStmt);
                }
            }
            typeStmt = stmts->dioGroupStmt;
            if (rc == SQLITE_OK) {
                rc = sqlite3_reset(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_clear_bindings(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 1, num);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 2, DisplayGroupBorder(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 3, DisplayGroupHoriz(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 4, DisplayGroupSpacing(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 5, DisplayObjectOpen(dio));
            }
            break;

        case DO_TYPE_WINDOW:
        {
            const struct WindowDimensions *wd = DisplayWindowDimensions(dio);

            if (rc == SQLITE_OK && stmts->dioWindowStmt == NULL) {
                rc = createTable(db, memPool, displayWindowTable, FALSE,
                        displayWindowColumns, NUM_ENTRIES(displayWindowColumns),
                        &err);
                if (rc == SQLITE_OK) {
                    rc = makeInsertStatement(db, memPool, displayWindowTable,
                            NUM_ENTRIES(displayWindowColumns), &err,
                            &stmts->dioWindowStmt);
                }
            }
            typeStmt = stmts->dioWindowStmt;
            if (rc == SQLITE_OK) {
                rc = sqlite3_reset(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_clear_bindings(typeStmt);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 1, num);
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 2, DisplayWindowId(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 3, DisplayWindowButton(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 4, 
                        DisplayWindowCloseWhenDisabled(dio));
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 5, DisplayWindowOpen(dio));
            }
            if (wd != NULL) {
                if (rc == SQLITE_OK) {
                    rc = sqlite3_bind_int(typeStmt, 6, wd->wd_LeftEdge);
                }
                if (rc == SQLITE_OK) {
                    rc = sqlite3_bind_int(typeStmt, 7, wd->wd_TopEdge);
                }
                if (rc == SQLITE_OK) {
                    rc = sqlite3_bind_int(typeStmt, 8, wd->wd_Width);
                }
                if (rc == SQLITE_OK) {
                    rc = sqlite3_bind_int(typeStmt, 9, wd->wd_Height);
                }
            }
            if (rc == SQLITE_OK) {
                rc = sqlite3_bind_int(typeStmt, 10, DisplayObjectOpen(dio));
            }
            break;
        }
    }

    if (err == OK) {
        err = sqlite3_error(rc, TRUE);
    }
    if (err != OK) {
        return err;
    }

    rc = sqlite3_step(dioStmt);
    if (rc != SQLITE_DONE) {
        return sqlite3_error(rc, TRUE);
    }
    rc = SQLITE_OK;

    if (typeStmt != NULL) {
        rc = sqlite3_step(typeStmt);
        if (rc != SQLITE_DONE) {
            return sqlite3_error(rc, TRUE);
        }
        rc = SQLITE_OK;
    }

    if (recordDescriptionSpec != NULL) {
        sqlite3_stmt *recStmt = NULL;
        struct StrHandle *sh = AllocStrHandle(memPool);
        int i;

        if (sh == NULL) {
            return ERR_MEM;
        }
        if (rc == SQLITE_OK && stmts->recDescSpecStmt == NULL) {
            rc = createTable(db, memPool, recordDescriptionSpecTable, FALSE,
                    recordDescriptionSpecColumns,
                    NUM_ENTRIES(recordDescriptionSpecColumns), &err);
            if (rc == SQLITE_OK) {
                rc = makeInsertStatement(db, memPool,
                        recordDescriptionSpecTable,
                        NUM_ENTRIES(recordDescriptionSpecColumns), &err,
                        &stmts->recDescSpecStmt);
            }
        }
        recStmt = stmts->recDescSpecStmt;
        if (rc == SQLITE_OK) {
            rc = sqlite3_reset(recStmt);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_clear_bindings(recStmt);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_int(recStmt, 1, num);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_text(recStmt, 2,
                    recordDescriptionSpecTypeString(
                        recordDescriptionSpec->rds_Type), -1, SQLITE_STATIC);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_text(recStmt, 3, 
                    recordDescriptionSpec->rds_Compute, -1, SQLITE_STATIC);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_int(recStmt, 4,
                    recordDescriptionSpec->rds_MultiColumn);
        }
        for (i = 0; i < recordDescriptionSpec->rds_NumFields; i++) {
            struct Field *field = recordDescriptionSpec->rds_Fields[i];

            if (i > 0) {
                StrHandlePutC(sh, ' ');
            }
            if (field == RDS_FIELD_RECNUM) {
                StrHandlePutS(sh, "num");
            } else {
                StrHandlePutS(sh, FieldName(field));
            }
        }
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else if (rc == SQLITE_OK) {
            rc = sqlite3_bind_text(recStmt, 5, StrHandleString(sh), -1,
                    SQLITE_STATIC);
            if (rc == SQLITE_OK) {
                rc = sqlite3_step(recStmt);
                if (rc == SQLITE_DONE) {
                    rc = SQLITE_OK;
                }
            }
        }
        FreeStrHandle(sh);

        if (err == OK) {
            err = sqlite3_error(rc, TRUE);
        }
        if (err != OK) {
            return err;
        }
    }

    BreakAddCurrentStep(1);
    if (childs) {
        FORNODE(dio, childs) {
            err = sqlite3WriteDisplayObjectsRec(db, dio, stmts);
            if (err) {
                break;
            }
        }
    }
    return err;
}

static long sqlite3WriteDisplayObjects(sqlite3 *db, struct Project *pr)
{
    struct DisplayObject *root = ProjectRootDisplayObject(pr);
    struct WriteDioStmts stmts;
    long err = OK;

    if (root == NULL) {
	return ERR_INTERNAL;
    }

    memset(&stmts, 0, sizeof(stmts));
    err = sqlite3WriteDisplayObjectsRec(db, root, &stmts);

    sqlite3_finalize(stmts.dioStmt);
    sqlite3_finalize(stmts.dioTableStmt);
    sqlite3_finalize(stmts.dioFieldStmt);
    sqlite3_finalize(stmts.dioSpaceStmt);
    sqlite3_finalize(stmts.dioImageStmt);
    sqlite3_finalize(stmts.dioRegisterStmt);
    sqlite3_finalize(stmts.dioGroupStmt);
    sqlite3_finalize(stmts.dioWindowStmt);
    sqlite3_finalize(stmts.recDescSpecStmt);

    return err;
}

static long sqlite3WriteQueries(sqlite3 *db, struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct QueryEditor *qe = ProjectQueryEditor(pr);
    struct List *queries = QueryEditorQueries(qe);
    struct Query *query;
    sqlite3_stmt *stmt = NULL;
    long err = OK;
    int rc;

    if (queries == NULL) {
	return OK;
    }
    rc = createTable(db, memPool, queryTable, FALSE, queryColumns,
            NUM_ENTRIES(queryColumns), &err);
    if (rc == SQLITE_OK) {
        rc = makeInsertStatement(db, memPool, queryTable,
                NUM_ENTRIES(queryColumns), &err, &stmt);
    }
    FORNODE(query, queries) {
        if (rc == SQLITE_OK) {
            rc = sqlite3_reset(stmt);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_clear_bindings(stmt);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_codeset(stmt, 1, QueryName(query), memPool,
                    FALSE, &err);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_int(stmt, 2, TableNum(QueryTable(query)));
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_bind_program(stmt, 3, QueryProgram(query), memPool,
                &err);
        }
        if (rc == SQLITE_OK) {
            rc = sqlite3_step(stmt);
            if (rc == SQLITE_DONE) {
                rc = SQLITE_OK;
            }
        }
        BreakAddCurrentStep(1);
        if (err != OK || rc != SQLITE_OK) {
            break;
        }
    }
    sqlite3_finalize(stmt);
    if (err == OK) {
        err = sqlite3_error(rc, TRUE);
    }
    return err;
}

static void breakAddOne(void)
{
    BreakAddCurrentStep(1);
}

static long sqlite3WriteRecords(struct SQLiteHandle *handle, bool sameFile,
        bool reorg)
{
    struct Project *pr = handle->sqlh_Project;
    sqlite3 *db = handle->sqlh_Db;
    struct MemPool *memPool = ProjectMemPool(pr);
    bool useTmp = reorg && sameFile;
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    if (tabs == NULL) {
	return ERR_INTERNAL;
    }

    FORNODE(tab, tabs) {
        struct StrHandle *sh = NULL;
        struct List *fields = TableFields(tab);
        struct Field *field;
        short type;
        struct List *recs = TableRecords(tab);
        struct Record *rec;
        struct DataInfo *di, di_buf;
        sqlite3_stmt *stmt = NULL;
        int count = 0, i;
	long recNum = 0;
        long err = OK;
        int rc = SQLITE_OK;

        if (!sameFile || reorg) {
            sh = AllocStrHandle(memPool);
            if (sh == NULL) {
                return ERR_MEM;
            }
            StrHandlePutS(sh, "DROP TABLE IF EXISTS ");
            StrHandlePutC(sh, '"');
            if (useTmp) {
                StrHandlePutS(sh, tmpNamePrefix);
            }
            StrHandlePutS(sh, TableName(tab));
            StrHandlePutC(sh, '"');
            StrHandlePutS(sh, ";\n");
            StrHandlePutS(sh, "CREATE TABLE ");
            StrHandlePutC(sh, '"');
            if (useTmp) {
                StrHandlePutS(sh, tmpNamePrefix);
            }
            StrHandlePutS(sh, TableName(tab));
            StrHandlePutC(sh, '"');
            StrHandlePutS(sh, " (");
            for (i = 0; i < NUM_ENTRIES(recordColumns); i++) {
                if (i > 0) {
                    StrHandlePutS(sh, ", ");
                }
                StrHandlePutS(sh, recordColumns[i].name);
                if (recordColumns[i].type != NULL) {
                    StrHandlePutC(sh, ' ');
                    StrHandlePutS(sh, recordColumns[i].type);
                }
                if (recordColumns[i].key) {
                    StrHandlePutS(sh, " PRIMARY KEY");
                }
            }
            if (fields != NULL) {
                FORNODE(field, fields) {
                    const char *sqlite3_type = sqlite3_type_for_field_type(
                            FieldType(field));

                    if (sqlite3_type == NULL) {
                        continue;
                    }
                    StrHandlePutS(sh, ", \"");
                    StrHandlePutS(sh, FieldName(field));
                    StrHandlePutS(sh, "\" ");
                    StrHandlePutS(sh, sqlite3_type);
                    count++;
                }
            }
            StrHandlePutS(sh, ")");
            StrHandlePutC(sh, 0);
            if (StrHandleError(sh)) {
                err = ERR_MEM;
            } else {
                rc = sqlite3_exec(db, StrHandleString(sh), NULL, NULL, NULL);
            }
            FreeStrHandle(sh);
            if (err == OK) {
                err = sqlite3_error(rc, TRUE);
            }
            if (err != OK) {
                return err;
            }
        }

        if (recs == NULL) {
            continue;
        }

        if (reorg) {
            sh = AllocStrHandle(memPool);
            if (sh == NULL) {
                return ERR_MEM;
            }
            StrHandlePutS(sh, "INSERT INTO ");
            StrHandlePutC(sh, '"');
            if (useTmp) {
                StrHandlePutS(sh, tmpNamePrefix);
            }
            StrHandlePutS(sh, TableName(tab));
            StrHandlePutC(sh, '"');
            StrHandlePutS(sh, " VALUES(?, ?, ?, ?");
            for (i = 0; i < count; i++) {
                StrHandlePutS(sh, ", ?");
            }
            StrHandlePutS(sh, ")");
            StrHandlePutC(sh, 0);
            if (StrHandleError(sh)) {
                err = ERR_MEM;
            } else {
                rc = sqlite3_prepare_v2(db, StrHandleString(sh), -1,&stmt,NULL);
            }
            FreeStrHandle(sh);
            if (err == OK) {
                err = sqlite3_error(rc, TRUE);
            }
            if (err != OK) {
                return err;
            }

            FORNODE(rec, recs) {
                recNum++;
                RecordLock(rec);
                rc = sqlite3_reset(stmt);
                if (rc == SQLITE_OK) {
                    rc = sqlite3_clear_bindings(stmt);
                }
                if (rc == SQLITE_OK) {
                    rc = sqlite3_bind_int(stmt, 1, recNum);
                }
                if (rc == SQLITE_OK) {
                    rc = sqlite3_bind_int(stmt, 2, 1);
                }
                if (rc == SQLITE_OK) {
                    rc = sqlite3_bind_int(stmt, 3, RecordMatchFilter(rec));
                }
                if (rc == SQLITE_OK) {
                    rc = sqlite3_bind_int(stmt, 4, recNum);
                }
                count = 4;
                if (fields != NULL) {
                    FORNODE(field, fields) {
                        err = RecordDataInfo(rec, field, &di, &di_buf);
                        if (err || rc) {
                            break;
                        }
                        type = FieldType(field);
                        switch (type) {
                            case TYPE_INTEGER:
                            case TYPE_REAL:
                            case TYPE_BOOL:
                            case TYPE_STRING:
                            case TYPE_MEMO: 
                            case TYPE_DATE:
                            case TYPE_CHOICE:
                            case TYPE_TIME:
                                count++;
                                rc = sqlite3_bind_datainfo(stmt, count, type,
                                        di, memPool, &err);
                                break;
                            case TYPE_REFERENCE:
                                count++;
                                if (di->di_Flags & DIF_NIL) {
                                    rc = sqlite3_bind_null(stmt, count);
                                } else {
                                    rc = sqlite3_bind_int(stmt, count,
                                            RecordNum(di->di_Contents.rec));
                                }
                                break;
                            case TYPE_BUTTON:
                            case TYPE_VIRTUAL:
                                break;
                        }
                        if (err != OK || rc != SQLITE_OK) {
                            break;
                        }
                    }
                }
                if (rc == SQLITE_OK) {
                    rc = sqlite3_step(stmt);
                    if (rc == SQLITE_DONE) {
                        rc = SQLITE_OK;
                    }
                }
                RecordUnlock(rec);
                if (err == OK) {
                    err = sqlite3_error(rc, TRUE);
                }
                if (err != OK) {
                    break;
                }
                BreakAddCurrentStep(1);
            }

            sqlite3_finalize(stmt);
        } else {
            ulong numRecs = TableNumRecords(tab);
            ulong stepsBefore = BreakCurrentSteps();
            ulong stepsAfter;
            long stepsLeft;

            if (sameFile) {
                err = SwapTableRecords(tab, TRUE, FALSE, breakAddOne);
            } else {
                err = WriteTableRecords(handle, tab, breakAddOne);
            }
            if (err != OK) {
                return err;
            }
            stepsAfter = BreakCurrentSteps();
            stepsLeft = numRecs - (stepsAfter - stepsBefore);

            if (TableRecordOrderNumResolution(tab) <= MIN_ORDER_NUM_RESOLUTION){
                sh = AllocStrHandle(memPool);
                if (sh == NULL) {
                    return ERR_MEM;
                }
                StrHandlePutS(sh, "UPDATE ");
                StrHandlePutC(sh, '"');
                StrHandlePutS(sh, TableName(tab));
                StrHandlePutC(sh, '"');
                StrHandlePutS(sh, " SET _OrderNum = ? WHERE _ID = ?");
                StrHandlePutC(sh, 0);
                if (StrHandleError(sh)) {
                    err = ERR_MEM;
                } else {
                    rc = sqlite3_prepare_v2(db, StrHandleString(sh), -1, &stmt,
                            NULL);
                }
                FreeStrHandle(sh);
                if (err == OK) {
                    err = sqlite3_error(rc, TRUE);
                }
                if (err != OK) {
                    return err;
                }

                FORNODE(rec, recs) {
                    RecordLock(rec);
                    rc = sqlite3_reset(stmt);
                    if (rc == SQLITE_OK) {
                        rc = sqlite3_clear_bindings(stmt);
                    }
                    if (rc == SQLITE_OK) {
                        rc = sqlite3_bind_int(stmt, 1, ++recNum);
                    }
                    if (rc == SQLITE_OK) {
                        rc = sqlite3_bind_int(stmt, 2, RecordInternalNum(rec));
                    }
                    if (rc == SQLITE_OK) {
                        rc = sqlite3_step(stmt);
                        if (rc == SQLITE_DONE) {
                            rc = SQLITE_OK;
                        }
                    }
                    RecordUnlock(rec);
                    if (err == OK) {
                        err = sqlite3_error(rc, TRUE);
                    }
                    if (err != OK) {
                        break;
                    }
                    if (stepsLeft > 0) {
                        BreakAddCurrentStep(1);
                        stepsLeft--;
                    }
                }
                sqlite3_finalize(stmt);
            } else {
                BreakAddCurrentStep(stepsLeft);
            }
        }
#if (SQLITE_VERSION_NUMBER >= 3010000) /* 3.10.0 */
        if (err == OK) {
            /* Flush pages to bring progress bar closer to actual progress. */
            rc = sqlite3_db_cacheflush(db);
            if (rc == SQLITE_BUSY) {
                rc = SQLITE_OK;
            }
            err = sqlite3_error(rc, TRUE);
        }
#endif
        if (err != OK) {
            return err;
        }
    }

    if (useTmp) {
        FORNODE(tab, tabs) {
            struct StrHandle *sh = NULL;
            long err = OK;
            int rc = SQLITE_OK;

            freeSQLiteTableForTable(tab);

            sh = AllocStrHandle(memPool);

            StrHandlePutS(sh, "DROP TABLE ");
            StrHandlePutC(sh, '"');
            StrHandlePutS(sh, TableName(tab));
            StrHandlePutC(sh, '"');
            StrHandlePutS(sh, ";\n");
            StrHandlePutS(sh, "ALTER TABLE ");
            StrHandlePutC(sh, '"');
            StrHandlePutS(sh, tmpNamePrefix);
            StrHandlePutS(sh, TableName(tab));
            StrHandlePutC(sh, '"');
            StrHandlePutS(sh, " RENAME TO ");
            StrHandlePutC(sh, '"');
            StrHandlePutS(sh, TableName(tab));
            StrHandlePutC(sh, '"');
            StrHandlePutC(sh, 0);
            if (StrHandleError(sh)) {
                err = ERR_MEM;
            } else {
                rc = sqlite3_exec(db, StrHandleString(sh), NULL, NULL, NULL);
            }
            FreeStrHandle(sh);
            if (err == OK) {
                err = sqlite3_error(rc, TRUE);
            }
            if (err != OK) {
                return err;
            }
        }
    }


    if (!sameFile || reorg) {
        FORNODE(tab, tabs) {
            createOrderIndex(db, tab);      /* ignore error */
        }
    }

    return OK;
}

static long sqlite3_verbose_error(long err, sqlite3 *db, const char *filename,
        char *errstr)
{
    if (errstr != NULL) {
        switch (err) {
            case SQLITEERR_IO:
            case SQLITEERR_GENERIC:
                err = SQLITEERR_MESSAGE;
                strcpy(errstr, EmptyNullStr(sqlite3_errmsg(db)));
                break;
            case DISKERR_INVALID:
                if (filename != NULL) {
                    err = DISKERR_NOTABEEBASEPROJECT;
                    strcpy(errstr, filename);
                }
                break;
            default:
                break;
        }
    }
    return err;
}

static int progressHandler(void *unused)
{
    return (int)BreakCheck();
}

static long saveProject(struct SQLiteHandle *handle, bool sameFile, bool reorg,
        bool vacuum, bool (*retryVacuum)(struct Project *, long, const char *),
        long projectVersion, long structureVersion, char *errstr)
{
    struct BreakSettings bs;
    struct Project *pr = handle->sqlh_Project;
    sqlite3 *db = handle->sqlh_Db;
    bool saveStructure = FALSE;
    long err = OK;
    int rc;

    if (db == NULL) {
	return ERR_INTERNAL;
    }

    err = SQLiteObtainChangeLock(handle);
    if (err != OK) {
        return err;
    }

    /* 
    ** Initialize the break dialog after obtaining a change lock.
    ** The change lock might log a change which may run a BeeBase
    ** program which in turn might use a break dialog.
    */
    GuiSetSleep(TRUE);
    saveStructure = structureVersion != ProjectStructureVersion(pr) ||!sameFile;
    bs.title = GCS(MSG_SavingDatabase);
    bs.steps = 1 + NumTables(pr);
    if (saveStructure) {
       bs.steps += NumFields(pr) +
           NumDisplayObjects(ProjectRootDisplayObject(pr)) + 
           NumNodes(QueryEditorQueries(ProjectQueryEditor(pr)));
    }
    bs.steps += ProjectNumRecords(pr) + 1;
    bs.current = 0;
    bs.abort = FALSE;
    bs.pause = TRUE;
    bs.needGui = ProjectHasGui(pr);
    BreakInit(&bs);

    if (err == OK) {
        err = sqlite3WriteProject(db, pr, projectVersion, structureVersion);
	BreakAddCurrentStep(1);
    }
    if (err == OK) {
        err = sqlite3WriteTables(db, pr);
    }
    if (err == OK && saveStructure) {
        err = sqlite3WriteFields(db, pr);
    }
    if (err == OK && saveStructure) {
        err = sqlite3WriteDisplayObjects(db, pr);
    }
    if (err == OK && saveStructure) {
        err = sqlite3WriteQueries(db, pr);
    }
    if (err == OK) {
        err = sqlite3WriteRecords(handle, sameFile, reorg);
    }
    if (err == OK) {
        char buf[50];

        sprintf(buf, "PRAGMA application_id = %ld",
                (long)MAKE_ID(0, 'b', 'b', 's'));
        sqlite3_exec(db, buf, NULL, NULL, NULL);
    }
    if (err == OK) {
        rc = sqlite3_exec(db, "COMMIT", NULL, NULL, NULL);
        err = sqlite3_error(rc, TRUE);
        BreakAddCurrentStep(1);
    }
    BreakExit(&bs);

    if (err == OK && vacuum) {
        bs.title = GCS(MSG_Vacuum);
        bs.steps = 0;
        bs.current = 0;
        bs.abort = TRUE;
        bs.pause = TRUE;
        bs.needGui = ProjectHasGui(pr);
        BreakInit(&bs);
        sqlite3_progress_handler(db, 1, progressHandler, NULL);

        do {
            rc = sqlite3_exec(db, "VACUUM", NULL, NULL, NULL);
            err = sqlite3_error(rc, TRUE);
        } while (err != OK && !BreakCheck() && retryVacuum  != NULL &&
                retryVacuum(pr, err, EmptyNullStr(sqlite3_errmsg(db))));

        sqlite3_progress_handler(db, 0, NULL, NULL);
        BreakExit(&bs);
        err = OK;       /* ignore Vacuum error. */
    }
    if (err == OK) {
        LogCommit(pr);
        SQLiteReleaseLock(handle);
    }

    GuiSetSleep(FALSE);

    return err;
}

static int sqliteBusyHandler(void *x, int numTrials)
{
    struct SQLiteHandle *handle = x;

    if (handle->sqlh_RetryObtainWriteLock != NULL &&
            !handle->sqlh_RetryObtainWriteLock(handle->sqlh_Project)) {
        handle->sqlh_RetryObtainWriteLock = NULL;
    }

    return handle->sqlh_RetryObtainWriteLock != NULL;
}

long SQLiteSaveProject(struct SQLiteHandle *handle, const char *filename,
        bool reorg, long projectVersion, long structureVersion, 
        bool (*retryObtainWriteLock)(struct Project *), 
        bool vacuum, bool (*retryVacuum)(struct Project *, long, const char *),
        char *errstr)
{
    struct SQLiteHandle *saveHandle = handle;
    struct SQLiteHandle *newHandle = NULL; 
    sqlite3 *db;
    struct Project *pr;
    bool sameFile;
    long error = OK;
    int rc;

    if (handle == NULL) {
        return ERR_INTERNAL;
    }
    if (filename == NULL) {
	return ERR_INVALIDNAME;
    }

    db = handle->sqlh_Db;
    pr = handle->sqlh_Project;
    sameFile = SameFile(ProjectName(pr), filename);

    if (!sameFile) {
        newHandle = AllocSQLiteHandle(pr);
        if (newHandle == NULL) {
            return ERR_MEM;
        }
        saveHandle = newHandle;

        RmFile(filename);
        rc = sqlite3_open(filename, &saveHandle->sqlh_Db);
        db = saveHandle->sqlh_Db;
        error = sqlite3_error(rc, TRUE);
        if (error == OK) {
            sqlite3_extended_result_codes(db, TRUE);
            saveHandle->sqlh_HaveFileChangeCounter = updateFileChangeCounter(
                    saveHandle, saveHandle->sqlh_FileChangeCounter);
            updateDataVersion(saveHandle, &saveHandle->sqlh_DataVersion);
        }
    }

    if (error == OK) {
        if (retryObtainWriteLock != NULL) {
            saveHandle->sqlh_RetryObtainWriteLock = retryObtainWriteLock;
            sqlite3_busy_handler(db, sqliteBusyHandler, saveHandle);
        }
        error = saveProject(saveHandle, sameFile, reorg, vacuum, retryVacuum,
                projectVersion, structureVersion, errstr);
        sqlite3_busy_handler(db, NULL, NULL);
        saveHandle->sqlh_RetryObtainWriteLock = NULL;
        saveHandle->sqlh_HaveFileChangeCounter = updateFileChangeCounter(
                saveHandle, saveHandle->sqlh_FileChangeCounter);
        updateDataVersion(saveHandle, &saveHandle->sqlh_DataVersion);
    }
    error = sqlite3_verbose_error(error, db, filename, errstr);

    if (error == OK) {
        swapSQLiteHandles(handle, saveHandle);
        if (reorg) {
            RecordAfterReorgUpdate(pr);
        }
        ResetRecordOrderNumResolutions(pr);
#ifdef AMIGA
        UpdateFileModificationTime(filename);
#endif
    } else if (!sameFile) {
        /* clean up newly created file */
        RmFile(filename);
    }

    FreeSQLiteHandle(newHandle);

    return error;
}

static const char *sqlite3_column_str(sqlite3_stmt *stmt, int pos, long *perr)
{
    bool isNull = sqlite3_column_type(stmt, pos) == SQLITE_NULL;
    const char *str = (const char *)sqlite3_column_text(stmt, pos);

    *perr = OK;
    if (str == NULL && !isNull) {
        *perr = ERR_MEM;
    }
    return str;
}

static char *sqlite3_column_codeset(sqlite3_stmt *stmt, int pos, 
        struct MemPool *memPool, long *perr, size_t *pwarnings)
{
    const char *utf8 = sqlite3_column_str(stmt, pos, perr);
    char *str = NULL;

    *pwarnings = 0;
    if (utf8 != NULL) {
        str = CodesetConvertFromUTF8(utf8, memPool, perr, pwarnings);
        if (*perr && str != NULL) {
            FreeString(memPool, str);
            str = NULL;
        } else if (str == utf8) {
            str = AllocString(memPool, utf8);
            if (str == NULL) {
                *perr = ERR_MEM;
            }
        }
    }

    return str;
}

static void sqlite3_column_datainfo(sqlite3_stmt *stmt, int pos,
        struct Field *field, short type, struct MemPool *memPool,
        long *perr, size_t *pwarnings, struct DataInfo *di, char **pmem)
{
    *perr = OK;
    *pwarnings = 0;
    *pmem = NULL;
    memset(di, 0, sizeof(*di));

    switch (type) {
        case TYPE_INTEGER:
            if (sqlite3_column_type(stmt, pos) == SQLITE_NULL) {
                di->di_Flags |= DIF_NIL;
            } else {
                di->di_Contents.integer = sqlite3_column_int(stmt, pos);
            }
            break;

        case TYPE_REAL:
            if (sqlite3_column_type(stmt, pos) == SQLITE_NULL) {
                di->di_Flags |= DIF_NIL;
            } else {
                di->di_Contents.real = sqlite3_column_double(stmt, pos);
            }
            break;

        case TYPE_BOOL:
            if (sqlite3_column_int(stmt, pos) == 0) {
                di->di_Flags |= DIF_NIL;
            }
            break;

        case TYPE_STRING:
            if (sqlite3_column_type(stmt, pos) != SQLITE_NULL) {
                di->di_Contents.string = *pmem = 
                    sqlite3_column_codeset(stmt, pos, memPool, perr, pwarnings);
            }
            break;

        case TYPE_MEMO: 
            if (sqlite3_column_type(stmt, pos) != SQLITE_NULL) {
                di->di_Contents.mi = *pmem = 
                    sqlite3_column_codeset(stmt, pos, memPool, perr, pwarnings);
            }
            break;

        case TYPE_DATE:
            if (sqlite3_column_type(stmt, pos) == SQLITE_NULL) {
                di->di_Flags |= DIF_NIL;
            } else {
                di->di_Contents.date = sqlite3_column_int(stmt, pos);
            }
            break;

        case TYPE_TIME:
            if (sqlite3_column_type(stmt, pos) == SQLITE_NULL) {
                di->di_Flags |= DIF_NIL;
            } else {
                di->di_Contents.time = sqlite3_column_int(stmt, pos);
            }
            break;

        case TYPE_CHOICE:
            if (sqlite3_column_type(stmt, pos) == SQLITE_NULL) {
                di->di_Flags |= DIF_NIL;
            } else {
                di->di_Contents.choice = sqlite3_column_int(stmt, pos);
            }
            break;

        case TYPE_REFERENCE:
            if (sqlite3_column_type(stmt, pos) == SQLITE_NULL) {
                di->di_Flags |= DIF_NIL;
            } else {
                di->di_Contents.rec = FindRecordInternalNum(
                            FieldReferenceTable(field), 
                            sqlite3_column_int(stmt, pos));
            }
            break;
    }
}

struct Program *sqlite3_column_program(sqlite3_stmt *stmt, int pos,
        struct Project *pr, long *perr, size_t *pwarnings)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    char *src = sqlite3_column_codeset(stmt, pos, memPool, perr, pwarnings);
    struct Program *prg = NULL;

    if (*perr == OK) {
        *perr = ReadProgram(pr, src, &prg);
    }
    FreeString(memPool, src);

    return prg;
}

static long getTableColumns(sqlite3 *db, const char *table,
        const struct SQLiteColumn *columns, bool *haveColumn, size_t num) 
{
    sqlite3_stmt *stmt = NULL;
    char cmd[64];
    long err = OK;
    int rc;
    size_t i;

    memset(haveColumn, 0, num * sizeof(bool));

    sprintf(cmd, "PRAGMA table_info(%s)", table);
    rc = sqlite3_prepare_v2(db, cmd, -1, &stmt, NULL);
    if (rc == SQLITE_OK) {
        while ((rc = sqlite3_step(stmt)) == SQLITE_ROW) {
            const char *name = sqlite3_column_str(stmt, 1, &err);

            if (err != OK) {
                break;
            }
            if (name != NULL) {
                for (i = 0; i < num; i++) {
                    if (strcasecmp(name, columns[i].name) == 0) {
                        haveColumn[i] = TRUE;
                        break;
                    }
                }
            }
        }
        if (rc == SQLITE_DONE) {
            rc = SQLITE_OK;
        }
    } else {
        err = sqlite3_error(rc, FALSE);
        if (err == SQLITEERR_GENERIC) {
            err = DISKERR_INVALID;
        }
    }

    /* Check that we have all required columns. */
    if (err == OK) {
        for (i = 0; i < num; i++) {
            if (!haveColumn[i] && columns[i].required) {
                err = DISKERR_INVALID;
                break;
            }
        }
    }

    sqlite3_finalize(stmt);

    if (err == OK) {
        err = sqlite3_error(rc, FALSE);
    }
    return err;
}

static char *makeSelectCommand(struct MemPool *memPool, const char *table, 
        const struct SQLiteColumn *columns, const bool *haveColumn, size_t num,
        const char *trail)
{
    struct StrHandle *sh = AllocStrHandle(memPool);
    char *sql = NULL;
    size_t i;

    if (sh) {
        StrHandlePutS(sh, "SELECT");

        for (i = 0; i < num; i++) {
            if (i > 0) {
                StrHandlePutC(sh, ',');
            }
            StrHandlePutC(sh, ' ');
            if (haveColumn[i]) {
                StrHandlePutS(sh, columns[i].name);
            } else if (columns[i].value != NULL) {
                StrHandlePutC(sh, '"');
                StrHandlePutS(sh, columns[i].value);
                StrHandlePutC(sh, '"');
            } else {
                StrHandlePutS(sh, "NULL");
            }
        }
        StrHandlePutS(sh, " FROM ");
        StrHandlePutS(sh, table);
        if (trail != NULL) {
            StrHandlePutC(sh, ' ');
            StrHandlePutS(sh, trail);
        }
        StrHandlePutC(sh, 0);
        if (!StrHandleError(sh)) {
            SQL(StrHandleString(sh));
            sql = AllocString(memPool, StrHandleString(sh));
        }
        FreeStrHandle(sh);
    }

    return sql;
}

static long makeSelectStatement(sqlite3 *db, struct MemPool *memPool,
        const char *table, const struct SQLiteColumn *columns, size_t num,
        const char *trail, sqlite3_stmt **stmt)
{
    bool *haveColumns;
    char *sql;
    long err = OK;
    int rc;

    *stmt = NULL;

    haveColumns = PoolAllocMem(memPool, num * sizeof(bool));
    if (haveColumns == NULL) {
        err = ERR_MEM;
    } else {
        err = getTableColumns(db, table, columns, haveColumns, num);
        if (err == OK) {
            sql = makeSelectCommand(memPool, table, columns, haveColumns, num,
                    trail);
            if (sql == NULL) {
                err = ERR_MEM;
            } else {
                rc = sqlite3_prepare_v2(db, sql, -1, stmt, NULL);
                err = sqlite3_error(rc, FALSE);
                FreeString(memPool, sql);
            }
        }
        PoolFreeMem(memPool, haveColumns, num * sizeof(bool));
    }
    return err;
}

static long runSingleSelectStatement(sqlite3_stmt *stmt, long id)
{
    int rc;

    rc = sqlite3_reset(stmt);
    if (rc == SQLITE_OK) {
        rc = sqlite3_bind_int(stmt, 1, id);
        if (rc == SQLITE_OK) {
            rc = sqlite3_step(stmt);
            if (rc == SQLITE_ROW) {
                return OK;
            } else if (rc == SQLITE_DONE) {
                return DISKERR_INVALID;
            }
        }
    }

    return sqlite3_error(rc, FALSE);
}

static long sqlite3ReadProject(sqlite3 *db, struct Project *pr,
        bool *update, bool *clear,
        long *projectVersion, long *structureVersion, char *errstr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    sqlite3_stmt *stmt = NULL;
    bool newProject = !*update;  /* initial update value matches new file */
    long err = OK;
    int rc;

    err = makeSelectStatement(db, memPool, projectTable, projectColumns,
            NUM_ENTRIES(projectColumns), NULL, &stmt);
    if (err == OK) {
        rc = sqlite3_step(stmt);
        if (rc != SQLITE_ROW) {
            err = DISKERR_INVALID;
        } else {
            struct MemPool *memPool = ProjectMemPool(pr);
            long format = sqlite3_column_int(stmt, 0);
            long sver = sqlite3_column_int(stmt, 3);
            const char *verstr;
            const char *passwd;
            size_t warnings = 0;

            if (format < PROJECT_FORMAT) {
                err = DISKERR_INVALID;
            } else if (format > PROJECT_FORMAT) {
                if (errstr != NULL) {
                    verstr = sqlite3_column_str(stmt, 1, &err);
                    strcpy(errstr, EmptyNullStr(verstr));
                }
                err = DISKERR_VERSION;
            }
            if (err == OK) {
                if (sver == 0 || sver != ProjectStructureVersion(pr) ||
                        ProjectStructureChanges(pr) != 0) {
                    *update = FALSE;
                }
                if (!*update) {
                    ClearProject(pr, newProject);
                }

                if (projectVersion != NULL) {
                    *projectVersion = sqlite3_column_int(stmt, 2);
                }
                if (structureVersion != NULL) {
                    *structureVersion = sver;
                }
                passwd = sqlite3_column_str(stmt, 4, &err);
                SetProjectAdminPassword(pr, passwd);

                *clear = TRUE;
            }
            if (err == OK && !*update) {
                struct ProjectPrefs *pp = ProjectProjectPrefs(pr);
                char *settings = sqlite3_column_codeset(stmt, 5, memPool,
                        &err, &warnings);

                if (err == OK) {
                    SetProjectPrefsSettings(pp, settings);
                }
                FreeString(memPool, settings);
                ProjectAddCodesetWarnings(pr, warnings);
            }
            if (err == OK && !*update) {
                SetProjectProgram(pr, sqlite3_column_program(stmt, 6, pr,
                            &err, &warnings));
                ProjectAddCodesetWarnings(pr, warnings);
            }
            if (err == OK && !*update) {
                SetDisplayProjectActiveTableNum(pr,
                        sqlite3_column_int(stmt, 7));
                SetDisplayProjectActiveDisplayObjectNum(pr,
                        sqlite3_column_int(stmt, 8));
            }
            if (err == OK) {
                rc = sqlite3_step(stmt);
                if (rc != SQLITE_DONE) {  /* Table has multiple rows? */
                    err = DISKERR_INVALID;
                }
            }
        }
    }
    sqlite3_finalize(stmt);

    return err;
}

static long sqlite3ReadTables(sqlite3 *db, struct Project *pr, bool update)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    sqlite3_stmt *stmt = NULL;
    long err = OK;
    int rc = SQLITE_OK;

    err = makeSelectStatement(db, memPool, tableTable, tableColumns,
            NUM_ENTRIES(tableColumns), update? NULL : "ORDER BY ID", &stmt);
    if (err == OK) {
        while ((rc = sqlite3_step(stmt)) == SQLITE_ROW) {
            long id = sqlite3_column_int(stmt, 0);
            long guiRecordID = sqlite3_column_int(stmt, 8);

            if (update) {
                struct Table *tab = FindTableNum(pr, id);

                if (tab == NULL) {
                    err = DISKERR_INVALID;
                } else {
                    UpdateTable(tab, guiRecordID);
                }
            } else {
                struct ReadTableData data;

                if (err == OK) {
                    data.rtd_ID = id;
                    data.rtd_Name = sqlite3_column_str(stmt, 1, &err);
                }
                if (err == OK) {
                    data.rtd_ExactlyOneRecord = sqlite3_column_int(stmt, 2);
                    data.rtd_FilterActive = sqlite3_column_int(stmt, 3);
                    data.rtd_NewTrigger = sqlite3_column_str(stmt, 4, &err);
                }
                if (err == OK) {
                    data.rtd_DeleteTrigger = sqlite3_column_str(stmt, 5, &err);
                }
                if (err == OK) {
                    data.rtd_CountChanges = sqlite3_column_int(stmt, 6);
                    data.rtd_LogChanges = sqlite3_column_int(stmt, 7);
                    data.rtd_GuiRecordID = guiRecordID;
                }
                if (err == OK) {
                    err = ReadTable(pr, &data);
                }
            }
            if (err != OK) {
                break;
            }
        }
        if (rc == SQLITE_DONE) {
            rc = SQLITE_OK;
        }
    }

    sqlite3_finalize(stmt);

    if (err == OK) {
        err = sqlite3_error(rc, FALSE);
    }
    return err;
}

static long sqlite3ReadFields(sqlite3 *db, struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    sqlite3_stmt *stmt = NULL;
    long err = OK;
    int rc = SQLITE_OK;

    err = makeSelectStatement(db, memPool, fieldTable, fieldColumns,
            NUM_ENTRIES(fieldColumns), "ORDER BY ID", &stmt);
    if (err == OK) {
        while ((rc = sqlite3_step(stmt)) == SQLITE_ROW) {
            struct ReadFieldData data;
            char *nil = NULL;
            char *mem = NULL;
            size_t warnings = 0;

            memset(&data, 0, sizeof(data));
            if (err == OK) {
                data.rfd_ID = sqlite3_column_int(stmt, 0);
                data.rfd_Table = FindTableNum(pr, sqlite3_column_int(stmt, 1));
                data.rfd_Name = sqlite3_column_str(stmt, 2, &err);
            }
            if (err == OK) {
                data.rfd_Type =
                    fieldTypeFromString(sqlite3_column_str(stmt, 3, &err));
            }
            if (err == OK) {
                data.rfd_Trigger = sqlite3_column_str(stmt, 4, &err);
            }
            if (err == OK) {
                data.rfd_CountChanges = sqlite3_column_int(stmt, 5);
                data.rfd_LogChanges = sqlite3_column_int(stmt, 6);
                sqlite3_column_datainfo(stmt, 7, NULL, data.rfd_Type,
                        memPool, &err, &warnings, &data.rfd_DataInfo, &mem);
                ProjectAddCodesetWarnings(pr, warnings);
            }
            if (err == OK) {
                data.rfd_NilString = nil =
                    sqlite3_column_codeset(stmt, 8, memPool, &err, &warnings);
                ProjectAddCodesetWarnings(pr, warnings);
            }

            if (err == OK) {
                switch (data.rfd_Type) {
                    case TYPE_STRING:
                        data.rfd_TypeData.string.maxlen =
                            sqlite3_column_int(stmt, 9);
                        break;

                    case TYPE_CHOICE:
                        data.rfd_TypeData.choice.labels = mem =
                            sqlite3_column_codeset(stmt, 10, memPool,
                                    &err, &warnings);
                        ProjectAddCodesetWarnings(pr, warnings);
                        break;

                    case TYPE_REFERENCE:
                        data.rfd_TypeData.reference.table = 
                            FindTableNum(pr, sqlite3_column_int(stmt, 11));
                        data.rfd_TypeData.reference.filter =
                            sqlite3_column_int(stmt, 12);
                        data.rfd_TypeData.reference.autoshow =
                            sqlite3_column_int(stmt, 13);
                        break;

                    case TYPE_VIRTUAL:
                        data.rfd_TypeData.virt.compute =
                            sqlite3_column_str(stmt, 14, &err);
                        break;

                    case TYPE_INTEGER:
                    case TYPE_REAL:
                    case TYPE_DATE:
                    case TYPE_TIME:
                    case TYPE_MEMO:
                    case TYPE_BOOL:
                    case TYPE_BUTTON:
                        break;

                    default:
                        err = FIELDERR_UNKNOWNTYPE; 
                        break;
                }
            }

            if (err == OK) {
                err = ReadField(pr, &data);
            }

            FreeString(memPool, mem);
            FreeString(memPool, nil);

            if (err != OK) {
                break;
            }
        }
        if (rc == SQLITE_DONE) {
            rc = SQLITE_OK;
        }
    }

    sqlite3_finalize(stmt);

    if (err == OK) {
        err = sqlite3_error(rc, FALSE);
    }
    return err;
}

struct ReadDioStmts
{
    sqlite3_stmt *dioStmt; 
    sqlite3_stmt *dioTableStmt;
    sqlite3_stmt *dioFieldStmt;
    sqlite3_stmt *dioSpaceStmt;
    sqlite3_stmt *dioImageStmt;
    sqlite3_stmt *dioRegisterStmt;
    sqlite3_stmt *dioGroupStmt;
    sqlite3_stmt *dioWindowStmt;
    sqlite3_stmt *recDescSpecStmt;
};

static long sqlite3ReadDisplayObjects(sqlite3 *db, struct Project *pr)
{
    struct ReadDioStmts stmts;
    struct MemPool *memPool = ProjectMemPool(pr);
    long err = OK;
    int rc = SQLITE_OK;

    memset(&stmts, 0, sizeof(stmts));

    err = makeSelectStatement(db, memPool, displayObjectTable,
            displayObjectColumns, NUM_ENTRIES(displayObjectColumns),
            "ORDER BY ID", &stmts.dioStmt);
    if (err == OK) {
        sqlite3_stmt *stmt = stmts.dioStmt;
        sqlite3_stmt *typeStmt;

        while ((rc = sqlite3_step(stmt)) == SQLITE_ROW) {
            struct ReadDisplayObjectData data;
            char *title = NULL;
            char *shortcut = NULL;
            char *qtFontname = NULL;
            char *gtkFontname = NULL;
            char *muiFontname = NULL;
            char *background = NULL;
            char *panelBackground = NULL;
            char *titleQtFontname = NULL;
            char *titleGTKFontname = NULL;
            char *titleMUIFontname = NULL;
            char *bubbleHelp = NULL;
            char *labels = NULL;
            short type = NOTYPE;
            size_t warnings = 0;
            struct RecordDescriptionSpec *recordDescriptionSpec = NULL;
            const struct RecordDescriptionSpec **pRecordDescriptionSpec = NULL;
            struct Table *tab = NULL;

            memset(&data, 0, sizeof(data));
            if (err == OK) {
                data.rdod_ID = sqlite3_column_int(stmt, 0);
                data.rdod_Parent = 
                    FindDisplayObjectNum(pr, sqlite3_column_int(stmt, 1));
                data.rdod_Visible = sqlite3_column_int(stmt, 2);
                data.rdod_Title = title =
                    sqlite3_column_codeset(stmt, 3, memPool, &err, &warnings);
                ProjectAddCodesetWarnings(pr, warnings);
            }
            if (err == OK) {
                shortcut =
                    sqlite3_column_codeset(stmt, 4, memPool, &err, &warnings);
                ProjectAddCodesetWarnings(pr, warnings);
                data.rdod_Shortcut = CodesetGetChar(EmptyNullStr(shortcut));
            }
            if (err == OK) {
                data.rdod_Enabled = displayObjectEnabledFromString(
                        sqlite3_column_str(stmt, 5, &err));
            }
            if (err == OK) {
                data.rdod_EnabledCompute = sqlite3_column_str(stmt, 6, &err);
            }
            if (err == OK) {
                data.rdod_QtFontname = qtFontname =
                    sqlite3_column_codeset(stmt, 7, memPool, &err, &warnings);
                ProjectAddCodesetWarnings(pr, warnings);
            }
            if (err == OK) {
                data.rdod_GTKFontname = gtkFontname =
                    sqlite3_column_codeset(stmt, 8, memPool, &err, &warnings);
                ProjectAddCodesetWarnings(pr, warnings);
            }
            if (err == OK) {
                data.rdod_MUIFontname = muiFontname =
                    sqlite3_column_codeset(stmt, 9, memPool, &err, &warnings);
                ProjectAddCodesetWarnings(pr, warnings);
            }
            if (err == OK) {
                data.rdod_Background = background =
                    sqlite3_column_codeset(stmt, 10, memPool, &err, &warnings);
                ProjectAddCodesetWarnings(pr, warnings);
            }
            if (err == OK) {
                data.rdod_Weight = sqlite3_column_int(stmt, 11);
                data.rdod_Type = displayObjectTypeFromString(
                        sqlite3_column_str(stmt, 12, &err));
            }

            switch (data.rdod_Type) {
                case DO_TYPE_TABLE:
                    if (err == OK && stmts.dioTableStmt == NULL) {
                        err = makeSelectStatement(db, memPool,
                                displayTableTable, displayTableColumns, 
                                NUM_ENTRIES(displayTableColumns),
                                "WHERE ID = ?", &stmts.dioTableStmt);
                    }
                    typeStmt = stmts.dioTableStmt;
                    if (err == OK) {
                        err = runSingleSelectStatement(typeStmt, data.rdod_ID);
                    }
                    if (err == OK) {
                        data.rdod_Contents.table.table = tab = FindTableNum(pr,
                                sqlite3_column_int(typeStmt, 1));
                        data.rdod_Contents.table.activeField = FindFieldNum(pr,
                                sqlite3_column_int(typeStmt, 2));
                        data.rdod_Contents.table.activeDisplayObjectID =
                            sqlite3_column_int(typeStmt, 3);
                        data.rdod_Contents.table.hasPanel =
                            sqlite3_column_int(typeStmt, 4);
                        data.rdod_Contents.table.panelBackground = 
                            panelBackground = sqlite3_column_codeset(
                                    typeStmt, 5, memPool, &err, &warnings);
                        ProjectAddCodesetWarnings(pr, warnings);
                    }
                    if (err == OK) {
                        data.rdod_Contents.table.panelNumAll =
                            sqlite3_column_int(typeStmt, 6);
                        data.rdod_Contents.table.panelFilter =
                            sqlite3_column_int(typeStmt, 7);
                        data.rdod_Contents.table.panelArrows =
                            sqlite3_column_int(typeStmt, 8);
                    }
                    pRecordDescriptionSpec =
                        &data.rdod_Contents.table.recordDescriptionSpec;
                    break;

                case DO_TYPE_FIELD:
                    if (err == OK && stmts.dioFieldStmt == NULL) {
                        err = makeSelectStatement(db, memPool,
                                displayFieldTable, displayFieldColumns, 
                                NUM_ENTRIES(displayFieldColumns),
                                "WHERE ID = ?", &stmts.dioFieldStmt);
                    }
                    typeStmt = stmts.dioFieldStmt;
                    if (err == OK) {
                        err = runSingleSelectStatement(typeStmt, data.rdod_ID);
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.field = FindFieldNum(pr, 
                                sqlite3_column_int(typeStmt, 1));
                        type = FieldType(data.rdod_Contents.field.field);
                        data.rdod_Contents.field.titlePosition =
                            displayFieldTitlePosFromString(
                                sqlite3_column_str(typeStmt, 2, &err));
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.titleQtFontname = 
                            titleQtFontname = sqlite3_column_codeset(
                                    typeStmt, 3, memPool, &err, &warnings);
                        ProjectAddCodesetWarnings(pr, warnings);
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.titleGTKFontname = 
                            titleGTKFontname = sqlite3_column_codeset(
                                    typeStmt, 4, memPool, &err, &warnings);
                        ProjectAddCodesetWarnings(pr, warnings);
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.titleMUIFontname = 
                            titleMUIFontname = sqlite3_column_codeset(
                                    typeStmt, 5, memPool, &err, &warnings);
                        ProjectAddCodesetWarnings(pr, warnings);
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.home =
                            sqlite3_column_int(typeStmt, 6);
                        data.rdod_Contents.field.tabChain =
                            sqlite3_column_int(typeStmt, 7);
                        data.rdod_Contents.field.readOnly =
                            sqlite3_column_int(typeStmt, 8);
                        data.rdod_Contents.field.subType =
                            displayFieldKindFromString(type,
                                    sqlite3_column_str(typeStmt, 9, &err));
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.alignment =
                            displayFieldAlignmentFromString(type,
                                    sqlite3_column_str(typeStmt, 10, &err));
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.calendar = 
                            sqlite3_column_int(typeStmt, 11);
                        data.rdod_Contents.field.format =
                            displayFieldFormatFromString(
                                    sqlite3_column_str(typeStmt, 12, &err));
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.numDecimals =
                            sqlite3_column_int(typeStmt, 13);
                        data.rdod_Contents.field.imageSize =
                            displayImageSizeFromString(
                                    sqlite3_column_str(typeStmt, 14, &err));
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.bubbleHelp = bubbleHelp =
                            sqlite3_column_codeset(typeStmt, 15, memPool,
                                    &err, &warnings);
                        ProjectAddCodesetWarnings(pr, warnings);
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.displayExternal =
                            sqlite3_column_int(typeStmt, 16);
                        data.rdod_Contents.field.titleAtStringField =
                            sqlite3_column_int(typeStmt, 17);
                        data.rdod_Contents.field.hideStringField =
                            sqlite3_column_int(typeStmt, 18);
                        data.rdod_Contents.field.file =
                            sqlite3_column_int(typeStmt, 19);
                        data.rdod_Contents.field.font =
                            sqlite3_column_int(typeStmt, 20);
                        data.rdod_Contents.field.listview =
                            sqlite3_column_int(typeStmt, 21);
                        data.rdod_Contents.field.listviewLabels = labels =
                            sqlite3_column_codeset(typeStmt, 22, memPool,
                                    &err, &warnings);
                        ProjectAddCodesetWarnings(pr, warnings);
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.listviewCompute =
                            sqlite3_column_str(typeStmt, 23, &err);
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.view =
                            sqlite3_column_int(typeStmt, 24);
                        data.rdod_Contents.field.frame =
                            sqlite3_column_int(typeStmt, 25);
                        data.rdod_Contents.field.recordsKind =
                            displayFieldRecordsKindFromString(
                                    sqlite3_column_str(typeStmt, 26, &err));
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.recordsCompute =
                            sqlite3_column_str(typeStmt, 27, &err);
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.initialRecord =
                            sqlite3_column_int(typeStmt, 28);
                        data.rdod_Contents.field.currentRecord =
                            sqlite3_column_int(typeStmt, 29);
                        data.rdod_Contents.field.show =
                            sqlite3_column_int(typeStmt, 30);
                        data.rdod_Contents.field.autoShow =
                            sqlite3_column_int(typeStmt, 31);
                        data.rdod_Contents.field.filter =
                            sqlite3_column_int(typeStmt, 32);
                        data.rdod_Contents.field.autoUpdate =
                            sqlite3_column_int(typeStmt, 33);
                        data.rdod_Contents.field.showTitles =
                            sqlite3_column_int(typeStmt, 34);
                        data.rdod_Contents.field.controlTable =
                            sqlite3_column_int(typeStmt, 35);
                        data.rdod_Contents.field.onDoubleClick =
                            displayFieldOnDoubleClickFromString(
                                    sqlite3_column_str(typeStmt, 36, &err));
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.doubleClickTrigger =
                            sqlite3_column_str(typeStmt, 37, &err);
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.urlDropTrigger =
                            sqlite3_column_str(typeStmt, 38, &err);
                    }
                    if (err == OK) {
                        data.rdod_Contents.field.sortDropTrigger =
                            sqlite3_column_str(typeStmt, 39, &err);
                    }
                    if (type == TYPE_REFERENCE) {
                        pRecordDescriptionSpec =
                            &data.rdod_Contents.field.recordDescriptionSpec;
                        tab =
                            FieldReferenceTable(data.rdod_Contents.field.field);
                    }
                    break;

                case DO_TYPE_SPACE:
                    if (err == OK && stmts.dioSpaceStmt == NULL) {
                        err = makeSelectStatement(db, memPool,
                                displaySpaceTable, displaySpaceColumns, 
                                NUM_ENTRIES(displaySpaceColumns),
                                "WHERE ID = ?", &stmts.dioSpaceStmt);
                    }
                    typeStmt = stmts.dioSpaceStmt;
                    if (err == OK) {
                        err = runSingleSelectStatement(typeStmt, data.rdod_ID);
                    }
                    if (err == OK) {
                        data.rdod_Contents.space.bar =
                                sqlite3_column_int(typeStmt, 1);
                    }
                    break;

                case DO_TYPE_IMAGE:
                    if (err == OK && stmts.dioImageStmt == NULL) {
                        err = makeSelectStatement(db, memPool,
                                displayImageTable, displayImageColumns, 
                                NUM_ENTRIES(displayImageColumns),
                                "WHERE ID = ?", &stmts.dioImageStmt);
                    }
                    typeStmt = stmts.dioImageStmt;
                    if (err == OK) {
                        err = runSingleSelectStatement(typeStmt, data.rdod_ID);
                    }
                    if (err == OK) {
                        data.rdod_Contents.image.imageSize =
                            displayImageSizeFromString(
                                    sqlite3_column_str(typeStmt, 1, &err));
                    }
                    break;

                case DO_TYPE_REGISTER:
                    if (err == OK && stmts.dioRegisterStmt == NULL) {
                        err = makeSelectStatement(db, memPool,
                                displayRegisterTable, displayRegisterColumns, 
                                NUM_ENTRIES(displayRegisterColumns),
                                "WHERE ID = ?", &stmts.dioRegisterStmt);
                    }
                    typeStmt = stmts.dioRegisterStmt;
                    if (err == OK) {
                        err = runSingleSelectStatement(typeStmt, data.rdod_ID);
                    }
                    if (err == OK) {
                        data.rdod_Contents.reg.labels = labels =
                            sqlite3_column_codeset(typeStmt, 1, memPool, &err,
                                    &warnings);
                        ProjectAddCodesetWarnings(pr, warnings);
                    }
                    if (err == OK) {
                        data.rdod_Contents.reg.open =
                            sqlite3_column_int(typeStmt, 2);
                    }
                    break;

                case DO_TYPE_GROUP:
                    if (err == OK && stmts.dioGroupStmt == NULL) {
                        err = makeSelectStatement(db, memPool,
                                displayGroupTable, displayGroupColumns, 
                                NUM_ENTRIES(displayGroupColumns),
                                "WHERE ID = ?", &stmts.dioGroupStmt);
                    }
                    typeStmt = stmts.dioGroupStmt;
                    if (err == OK) {
                        err = runSingleSelectStatement(typeStmt, data.rdod_ID);
                    }
                    if (err == OK) {
                        data.rdod_Contents.group.border =
                            sqlite3_column_int(typeStmt, 1);
                        data.rdod_Contents.group.horizontal =
                            sqlite3_column_int(typeStmt, 2);
                        data.rdod_Contents.group.spacing =
                            sqlite3_column_int(typeStmt, 3);
                        data.rdod_Contents.group.open =
                            sqlite3_column_int(typeStmt, 4);
                    }
                    break;

                case DO_TYPE_WINDOW:
                    if (err == OK && stmts.dioWindowStmt == NULL) {
                        err = makeSelectStatement(db, memPool,
                                displayWindowTable, displayWindowColumns, 
                                NUM_ENTRIES(displayWindowColumns),
                                "WHERE ID = ?", &stmts.dioWindowStmt);
                    }
                    typeStmt = stmts.dioWindowStmt;
                    if (err == OK) {
                        err = runSingleSelectStatement(typeStmt, data.rdod_ID);
                    }
                    if (err == OK) {
                        data.rdod_Contents.window.windowId =
                            sqlite3_column_int(typeStmt, 1);
                        data.rdod_Contents.window.button =
                            sqlite3_column_int(typeStmt, 2);
                        data.rdod_Contents.window.closeWhenDisabled =
                            sqlite3_column_int(typeStmt, 3);
                        data.rdod_Contents.window.windowOpen =
                            sqlite3_column_int(typeStmt, 4);
                        data.rdod_Contents.window.dimensions.wd_LeftEdge =
                            sqlite3_column_type(typeStmt, 5) == SQLITE_NULL?
                            -1 : sqlite3_column_int(typeStmt, 5);
                        data.rdod_Contents.window.dimensions.wd_TopEdge =
                            sqlite3_column_type(typeStmt, 6) == SQLITE_NULL?
                            -1 : sqlite3_column_int(typeStmt, 6);
                        data.rdod_Contents.window.dimensions.wd_Width =
                            sqlite3_column_type(typeStmt, 7) == SQLITE_NULL?
                            -1 : sqlite3_column_int(typeStmt, 7);
                        data.rdod_Contents.window.dimensions.wd_Height =
                            sqlite3_column_type(typeStmt, 8) == SQLITE_NULL?
                            -1 : sqlite3_column_int(typeStmt, 8);
                        data.rdod_Contents.window.open =
                            sqlite3_column_int(typeStmt, 9);
                    }
                    break;
            }

            if (err == OK && pRecordDescriptionSpec != NULL) {
                struct RecordDescriptionSpec *rds = NULL;
                sqlite3_stmt *rdsStmt;
                bool ignoreError = FALSE;
                const char *fields = NULL;
                short num = 0;

                if (stmts.recDescSpecStmt == NULL) {
                    err = makeSelectStatement(db, memPool,
                            recordDescriptionSpecTable,
                            recordDescriptionSpecColumns,
                            NUM_ENTRIES(recordDescriptionSpecColumns),
                            "WHERE ID = ?", &stmts.recDescSpecStmt);
                }
                rdsStmt = stmts.recDescSpecStmt;
                if (err == OK) {
                    err = runSingleSelectStatement(rdsStmt, data.rdod_ID);
                }
                if (err == DISKERR_INVALID) {
                    ignoreError = TRUE;  /* Ignore missing entry */
                }
                if (err == OK) {
                    fields = sqlite3_column_str(rdsStmt, 4, &err);
                }
                if (err == OK) {
                    num = NumWords(fields);
                    *pRecordDescriptionSpec = recordDescriptionSpec = rds = 
                        AllocRecordDescriptionSpec(memPool, num);
                    if (rds == NULL) {
                        err = ERR_MEM;
                    }
                }
                if (err == OK) {
                    short i;
                    char *word;
                    bool fail = FALSE;

                    for (i = 0; i < num; i++) {
                        word = WordString(memPool, fields, i, &fail);
                        if (fail) {
                            err = ERR_MEM;
                            break;
                        }
                        if (word == NULL) {
                            rds->rds_Fields[i] = NULL;
                        } else if (strcmp(word, "num") == 0) {
                            rds->rds_Fields[i] = RDS_FIELD_RECNUM;
                        } else {
                            rds->rds_Fields[i] = FindFieldName(tab, word);
                        }
                        FreeString(memPool, word);
                    }
                }
                if (err == OK) {
                    rds->rds_Type = recordDescriptionSpecTypeFromString(
                            sqlite3_column_str(rdsStmt, 1, &err));
                }
                if (err == OK) {
                    const char *compute = sqlite3_column_str(rdsStmt, 2, &err);

                    if (err == OK && compute != NULL) {
                        rds->rds_Compute = AllocString(memPool, compute);
                        if (rds->rds_Compute == NULL) {
                            err = ERR_MEM;
                        }
                    }
                }
                if (err == OK) {
                    rds->rds_MultiColumn = sqlite3_column_int(rdsStmt, 3);
                }
                if (ignoreError) {
                    err = OK;
                }
            }
            if (err == OK) {
                err = ReadDisplayObject(pr, &data);
            }

            FreeString(memPool, shortcut);
            FreeString(memPool, qtFontname);
            FreeString(memPool, gtkFontname);
            FreeString(memPool, muiFontname);
            FreeString(memPool, background);
            FreeString(memPool, panelBackground);
            FreeString(memPool, titleQtFontname);
            FreeString(memPool, titleGTKFontname);
            FreeString(memPool, titleMUIFontname);
            FreeString(memPool, bubbleHelp);
            FreeString(memPool, labels);
            FreeRecordDescriptionSpec(memPool, recordDescriptionSpec);

            if (err != OK) {
                break;
            }
        }
        if (rc == SQLITE_DONE) {
            rc = SQLITE_OK;
        }
    }

    sqlite3_finalize(stmts.dioStmt);
    sqlite3_finalize(stmts.dioTableStmt);
    sqlite3_finalize(stmts.dioFieldStmt);
    sqlite3_finalize(stmts.dioSpaceStmt);
    sqlite3_finalize(stmts.dioImageStmt);
    sqlite3_finalize(stmts.dioRegisterStmt);
    sqlite3_finalize(stmts.dioGroupStmt);
    sqlite3_finalize(stmts.dioWindowStmt);
    sqlite3_finalize(stmts.recDescSpecStmt);

    if (err == OK) {
        err = sqlite3_error(rc, FALSE);
    }
    return err;
}

static long sqlite3ReadOrdersAndFilters(sqlite3 *db, struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    sqlite3_stmt *stmt = NULL;
    long err = OK;
    int rc = SQLITE_OK;

    err = makeSelectStatement(db, memPool, tableTable, tableColumns,
            NUM_ENTRIES(tableColumns), NULL, &stmt);
    if (err == OK) {
        while ((rc = sqlite3_step(stmt)) == SQLITE_ROW) {
            long id = sqlite3_column_int(stmt, 0);
            struct Table *tab = FindTableNum(pr, id);
            const char *order = NULL;
            size_t warnings = 0;

            if (tab == NULL) {
                err = DISKERR_INVALID;
            }
            if (err == OK) {
                order = sqlite3_column_str(stmt, 9, &err);
            }
            if (err == OK) {
                err = SetOrderStr(tab, order);
            }
            if (err == OK) {
                SetTableFilterProgram(tab, sqlite3_column_program(stmt, 10, pr,
                            &err, &warnings));
                ProjectAddCodesetWarnings(pr, warnings);
            }

            if (err != OK) {
                break;
            }
        }
        if (rc == SQLITE_DONE) {
            rc = SQLITE_OK;
        }
    }

    sqlite3_finalize(stmt);

    if (err == OK) {
        err = sqlite3_error(rc, FALSE);
    }
    return err;
}

static long sqlite3ReadQueries(sqlite3 *db, struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct QueryEditor *qe = ProjectQueryEditor(pr);
    sqlite3_stmt *stmt = NULL;
    long err = OK;
    int rc = SQLITE_OK;

    err = makeSelectStatement(db, memPool, queryTable, queryColumns,
            NUM_ENTRIES(queryColumns), NULL, &stmt);
    if (err == OK && qe == NULL) {
        err = ERR_INTERNAL;
    }
    if (err == DISKERR_INVALID) {
        err = OK;  /* Ignore missing table. */
    } else if (err == OK) {
        while ((rc = sqlite3_step(stmt)) == SQLITE_ROW) {
            struct Query *query = NewQuery(qe, NULL);
            char *name = NULL;
            size_t warnings;

            if (query == NULL) {
                err = ERR_MEM;
            }
            if (err == OK) {
                name = sqlite3_column_codeset(stmt, 0, memPool, &err,&warnings);
                ProjectAddCodesetWarnings(pr, warnings);
            }
            if (err == OK) {
                SetQueryName(query, name);
                SetQueryTable(query, FindTableNum(pr,
                            sqlite3_column_int(stmt, 1)));
                SetQueryProgram(query, sqlite3_column_program(stmt, 2, pr,
                            &err, &warnings));
                ProjectAddCodesetWarnings(pr, warnings);
            }
            FreeString(memPool, name);

            if (err != OK) {
                break;
            }
        }
        if (rc == SQLITE_DONE) {
            rc = SQLITE_OK;
        }
    }

    sqlite3_finalize(stmt);

    if (err == OK) {
        err = sqlite3_error(rc, FALSE);
    }
    return err;
}

static long sqlite3ReadRecords(sqlite3 *db, struct Project *pr, bool update)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    long err = OK;

    if (tabs == NULL) {
	return ERR_INTERNAL;
    }

    if (update) {
        PrepareRecordUpdate(pr);
    }

    FORNODE(tab, tabs) {
        char name[1 + ID_NAME_LEN + 1 + 1];
        bool filter = TableFilterActive(tab);
        struct ReadRecordData data;
        sqlite3_stmt *stmt = NULL;
        int rc = SQLITE_OK;

        sprintf(name, "\"%.*s\"", ID_NAME_LEN, TableName(tab));
        err = makeSelectStatement(db, memPool, name, recordColumns,
                NUM_ENTRIES(recordColumns) - (filter? 0 : 1) - 1,
                "ORDER BY _OrderNum", &stmt);
        if (err == OK) {
            while ((rc = sqlite3_step(stmt)) == SQLITE_ROW) {
                data.rrd_ID = sqlite3_column_int(stmt, 0);
                data.rrd_Version = sqlite3_column_int(stmt, 1);
                data.rrd_IsSorted = TRUE;
                data.rrd_MatchFilter = !filter || sqlite3_column_int(stmt, 2);
                
                err = ReadRecord(tab, &data, update);
                if (err != OK) {
                    break;
                }
            }
            if (rc == SQLITE_DONE) {
                rc = SQLITE_OK;
            }
        }

        sqlite3_finalize(stmt);

        if (err == OK) {
            err = sqlite3_error(rc, FALSE);
        }
        if (err == OK) {
            err = AfterReadRecords(tab);
        }
        if (err != OK) {
            break;
        }

	BreakAddCurrentStep(1);
    }

    if (err == OK && update) {
        CompleteRecordUpdate(pr);
    }

    return err;
}

static long loadProject(struct SQLiteHandle *handle, bool *update, bool *clear,
        long *projectVersion, long *structureVersion, char *errstr)
{
    struct Project *pr = handle->sqlh_Project;
    sqlite3 *db = handle->sqlh_Db;
    struct BreakSettings bs;
    bool breakInitialized = FALSE;
    long err = OK;

    if (db == NULL) {
	return ERR_INTERNAL;
    }

    GuiSetSleep(TRUE);
    err = sqlite3ReadProject(db, pr, update, clear,
            projectVersion, structureVersion, errstr);
    if (err == OK) {
        err = sqlite3ReadTables(db, pr, *update);
    }
    if (err == OK) {
        bs.title = GCS(MSG_LoadingDatabase);
        bs.steps = (*update? 0 : 4) + NumTables(pr);
        bs.current = 0;
        bs.abort = FALSE;
        bs.pause = TRUE;
        bs.needGui = ProjectHasGui(pr);
        BreakInit(&bs);
        breakInitialized = TRUE;
    }
    if (err == OK && !*update) {
        err = sqlite3ReadFields(db, pr);
	BreakAddCurrentStep(1);
    }
    if (err == OK && !*update) {
        err = sqlite3ReadDisplayObjects(db, pr);
	BreakAddCurrentStep(1);
    }
    if (err == OK && !*update) {
        err = sqlite3ReadOrdersAndFilters(db, pr);
	BreakAddCurrentStep(1);
    }
    if (err == OK && !*update) {
        err = sqlite3ReadQueries(db, pr);
	BreakAddCurrentStep(1);
    }
    if (err == OK) {
        err = sqlite3ReadRecords(db, pr, *update);
    }
    if (breakInitialized) {
        BreakExit(&bs);
    }
    GuiSetSleep(FALSE);

    return err;
}

long SQLiteOpenProject(struct SQLiteHandle *handle, const char *filename,
        bool *pUpdate, bool *pClearOnError, 
        long *projectVersion, long *structureVersion, char *errstr)
{
    struct SQLiteHandle *loadHandle = handle;
    struct SQLiteHandle *newHandle = NULL; 
    sqlite3 *db;
    struct Project *pr;
    bool update = TRUE;
    bool clear = FALSE;
    bool sameFile;
    long dataVersion = 0;
    long error = OK;
    int rc;

    if (handle == NULL) {
        return ERR_INTERNAL;
    }
    if (filename == NULL) {
	return ERR_INVALIDNAME;
    }

    db = handle->sqlh_Db;
    pr = handle->sqlh_Project;
    sameFile = SameFile(ProjectName(pr), filename);

    if (!sameFile) {
        newHandle = AllocSQLiteHandle(pr);
        if (newHandle == NULL) {
            return ERR_MEM;
        }
        loadHandle = newHandle;

        rc = sqlite3_open(filename, &loadHandle->sqlh_Db);
        db = loadHandle->sqlh_Db;
        error = sqlite3_error(rc, FALSE);
        if (error == OK) {
            sqlite3_extended_result_codes(db, TRUE);
        }
        update = FALSE;
    } else if (sqlite3_in_transaction(db)) {
        LogRollback(pr);
        rc = sqlite3_exec(db, "ROLLBACK", NULL, NULL, NULL);
        error = sqlite3_error(rc, FALSE);
    }

    if (error == OK) {
        error = sqliteKeepLock(loadHandle, FALSE);
        loadHandle->sqlh_HaveFileChangeCounter = updateFileChangeCounter(
                loadHandle, loadHandle->sqlh_FileChangeCounter);
        updateDataVersion(loadHandle, &dataVersion);
    }
    if (error == OK) {
        error = loadProject(loadHandle, &update, &clear,
                projectVersion, structureVersion, errstr);
    }
    error = sqlite3_verbose_error(error, db, filename, errstr);

    if (error == OK) {
        loadHandle->sqlh_DataVersion = dataVersion;
        loadHandle->sqlh_HasChanged = FALSE;
        swapSQLiteHandles(handle, loadHandle);
    }

    FreeSQLiteHandle(newHandle);

    if (pUpdate != NULL) {
        *pUpdate = update;
    }
    if (pClearOnError != NULL) {
        *pClearOnError = clear;
    }

    return error;
}

static struct SQLiteTable *getSQLiteTable(struct SQLiteHandle *handle,
        struct Table *tab)
{
    struct SQLiteTable *sqlt = TableSQLiteTable(tab);

    if (sqlt == NULL && handle != NULL) {
        struct Project *pr = handle->sqlh_Project;
        struct MemPool *memPool = ProjectMemPool(pr);
        struct List *fields = TableFields(tab);
        struct Field *field;
        long num = 0;

        if (fields != NULL) {
            FORNODE(field, fields) {
                if (sqlite3_type_for_field_type(FieldType(field)) != NULL) {
                    num++;
                }
            }
        }
        sqlt = PoolAllocMemZero(memPool, sizeof(struct SQLiteTable) +
                num * sizeof(struct Field *));
        if (sqlt != NULL) {
            sqlt->sqlt_Table = tab;
            sqlt->sqlt_NumFields = num;
            num = 0;
            if (fields != NULL) {
                FORNODE(field, fields) {
                    if (sqlite3_type_for_field_type(FieldType(field)) != NULL) {
                        sqlt->sqlt_Fields[num] = field;
                        num++;
                    }
                }
            }
            ADDTAIL(&handle->sqlh_Tables, sqlt);
            SetTableSQLiteTable(tab, sqlt);
        }
    }

    return sqlt;
}

long SQLiteReadRecord(struct SQLiteHandle *handle, struct Table *tab, long id,
        char *data, struct MemPool *dataPool,
        void (*setDataInfo)(char *, struct Field *, const struct DataInfo *),
        size_t *pTotalWarnings)
{
    struct Project *pr = TableProject(tab);
    struct SQLiteTable *sqlt;
    sqlite3 *db;
    sqlite3_stmt *stmt;
    long err = OK;
    int rc = SQLITE_OK;

    if (pTotalWarnings != NULL) {
        *pTotalWarnings = 0;
    }

    if (handle == NULL || setDataInfo == NULL) {
        return ERR_INTERNAL;
    }
    db = handle->sqlh_Db;
    if (db == NULL) {
        return ERR_INTERNAL;
    }

    sqlt = getSQLiteTable(handle, tab);
    if (sqlt == NULL) {
        return ERR_MEM;
    }

    /* printf("Read(%s) ID=%ld %s\n", TableName(tab), id, 
            sqlt->sqlt_Stmts.readRecordData == NULL? "new" : ""); */

    if (sqlt->sqlt_Stmts.readRecordData == NULL) {
        struct MemPool *memPool = ProjectMemPool(pr);
        struct StrHandle *sh;
        long i;

        sh = AllocStrHandle(memPool);
        if (sh == NULL) {
            return ERR_MEM;
        }
        StrHandlePutS(sh, "SELECT ");
        for (i = 0; i < sqlt->sqlt_NumFields; i++) {
            if (i > 0) {
                StrHandlePutS(sh, ", ");
            }
            StrHandlePutC(sh, '"');
            StrHandlePutS(sh, FieldName(sqlt->sqlt_Fields[i]));
            StrHandlePutC(sh, '"');
        }
        StrHandlePutS(sh, " FROM ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " WHERE _ID = ?");
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else {
            SQL(StrHandleString(sh));
            rc = sqlite3_prepare_v2(db, StrHandleString(sh), -1, 
                    &sqlt->sqlt_Stmts.readRecordData, NULL);
            err = sqlite3_error(rc, FALSE);
        }
        FreeStrHandle(sh);
    }

    if (err != OK) {
        return err;
    }

    stmt = sqlt->sqlt_Stmts.readRecordData;
    rc = sqlite3_bind_int(stmt, 1, id);
    if (rc != SQLITE_OK) {
        err = ERR_INTERNAL;
    }
    if (err == OK) {
        rc = sqlite3_step(stmt);
        if (rc != SQLITE_ROW) {
            if (rc == SQLITE_DONE) {
                err = DISKERR_INVALID;
            } else {
                err = sqlite3_error(rc, FALSE);
            }
        }
    }
    if (err == OK) {
        err = sqliteKeepLock(handle, TRUE);
    }
    if (err == OK) {
        long err2 = OK;
        size_t warnings = 0;
        size_t totalWarnings = 0;
        struct Field *field;
        struct DataInfo di;
        char *unused;
        long i;

        for (i = 0; i < sqlt->sqlt_NumFields; i++) {
            field = sqlt->sqlt_Fields[i];

            sqlite3_column_datainfo(stmt, i, field, FieldType(field), dataPool,
                    &err2, &warnings, &di, &unused);
            if (err2 != OK) {
                err = err2;
            }
            ProjectAddCodesetWarnings(pr, warnings);
            totalWarnings += warnings;

            setDataInfo(data, field, &di);
        }

        if (pTotalWarnings != NULL) {
            *pTotalWarnings = totalWarnings;
        }
    }
    rc = sqlite3_reset(stmt);
    if (err == OK && rc != SQLITE_OK) {
        err = ERR_INTERNAL;
    }

    return err;
}

long SQLiteGetOrderNum(struct SQLiteHandle *handle, struct Table *tab, 
        long id, double *orderNum)
{
    struct Project *pr = TableProject(tab);
    struct SQLiteTable *sqlt;
    sqlite3 *db;
    sqlite3_stmt *stmt;
    long err = OK;
    int rc = SQLITE_OK;

    if (handle == NULL) {
        return ERR_INTERNAL;
    }
    db = handle->sqlh_Db;
    if (db == NULL) {
        return ERR_INTERNAL;
    }

    sqlt = getSQLiteTable(handle, tab);
    if (sqlt == NULL) {
        return ERR_MEM;
    }

    if (sqlt->sqlt_Stmts.readRecordOrderNum == NULL) {
        struct MemPool *memPool = ProjectMemPool(pr);
        struct StrHandle *sh;

        sh = AllocStrHandle(memPool);
        if (sh == NULL) {
            return ERR_MEM;
        }
        StrHandlePutS(sh, "SELECT _OrderNum FROM ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " WHERE _ID = ?");
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else {
            SQL(StrHandleString(sh));
            rc = sqlite3_prepare_v2(db, StrHandleString(sh), -1, 
                    &sqlt->sqlt_Stmts.readRecordOrderNum, NULL);
            err = sqlite3_error(rc, FALSE);
        }
        FreeStrHandle(sh);
    }

    if (err != OK) {
        return err;
    }

    stmt = sqlt->sqlt_Stmts.readRecordOrderNum;
    rc = sqlite3_bind_int(stmt, 1, id);
    if (rc != SQLITE_OK) {
        err = ERR_INTERNAL;
    }
    if (err == OK) {
        rc = sqlite3_step(stmt);
        if (rc != SQLITE_ROW) {
            if (rc == SQLITE_DONE) {
                err = DISKERR_INVALID;
            } else {
                err = sqlite3_error(rc, FALSE);
            }
        }
    }
    if (err == OK) {
        if (orderNum != NULL) {
            *orderNum = sqlite3_column_double(stmt, 0);
        }
    }
    rc = sqlite3_reset(stmt);
    if (err == OK && rc != SQLITE_OK) {
        err = ERR_INTERNAL;
    }

    return err;
}

long SQLiteWriteRecord(struct SQLiteHandle *handle, struct Table *tab,
        bool exists, long id, double orderNum, bool matchFilter,
        long version, const char *data,
        bool (*getDataInfo)(const char *, struct Field *, struct DataInfo *))
{
    struct Project *pr = TableProject(tab);
    struct MemPool *memPool = ProjectMemPool(pr);
    bool writeHeader = (orderNum >= 0.0);
    bool writeData = (data != NULL);
    struct SQLiteTable *sqlt;
    sqlite3 *db;
    sqlite3_stmt **pstmt, *stmt;
    long col;
    long err = OK;
    int rc = SQLITE_OK;

    if (writeData && getDataInfo == NULL) {
        return ERR_INTERNAL;
    }
    if (handle == NULL) {
        return ERR_INTERNAL;
    }
    db = handle->sqlh_Db;
    if (db == NULL) {
        return RECERR_NOPROJECTNAME;
    }
    if (!sqlite3_in_transaction(db)) {
        return ERR_INTERNAL;
    }

    sqlt = getSQLiteTable(handle, tab);
    if (sqlt == NULL) {
        return ERR_MEM;
    }

    if (exists) {
        if (writeHeader) {
            pstmt = writeData? 
                &sqlt->sqlt_Stmts.updateRecordHeaderAndData : 
                &sqlt->sqlt_Stmts.updateRecordHeader;
        } else if (writeData) {
            pstmt = &sqlt->sqlt_Stmts.updateRecordData;
        } else {
            return OK;  /* Nothing to update */
        }
    } else {
        if (writeHeader) {
            pstmt = writeData? 
                &sqlt->sqlt_Stmts.insertRecordHeaderAndData : 
                &sqlt->sqlt_Stmts.insertRecordHeader;
        } else if (writeData) {
            pstmt = &sqlt->sqlt_Stmts.insertRecordData;
        } else {
            pstmt = &sqlt->sqlt_Stmts.insertRecordId;
        }
    }

    if (*pstmt == NULL) {
        struct StrHandle *sh;
        bool needComma = !exists;
        long i;

        sh = AllocStrHandle(memPool);
        if (sh == NULL) {
            return ERR_MEM;
        }
        StrHandlePutS(sh, exists? "UPDATE " : "INSERT INTO ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, exists? " SET " : " (_ID");
        if (writeHeader) {
            for (i = 2; i < NUM_ENTRIES(recordColumns); i++) {
                if (needComma) {
                    StrHandlePutS(sh, ", ");
                }
                StrHandlePutS(sh, recordColumns[i].name);
                if (exists) {
                    StrHandlePutS(sh, " = ?");
                }
                needComma = TRUE;
            }
        }
        if (writeData) {
            if (needComma) {
                StrHandlePutS(sh, ", ");
            }
            StrHandlePutS(sh, "_Version");
            if (exists) {
                StrHandlePutS(sh, " = ?");
            }
            needComma = TRUE;
            for (i = 0; i < sqlt->sqlt_NumFields; i++) {
                StrHandlePutS(sh, ", ");  /* always need comma from here on */
                StrHandlePutC(sh, '"');
                StrHandlePutS(sh, FieldName(sqlt->sqlt_Fields[i]));
                StrHandlePutC(sh, '"');
                if (exists) {
                    StrHandlePutS(sh, " = ?");
                }
            }
        }
        if (exists) {
            StrHandlePutS(sh, " WHERE _ID = ?");
        } else {
            StrHandlePutS(sh, ") VALUES(?");
            if (writeHeader) {
                for (i = 2; i < NUM_ENTRIES(recordColumns); i++) {
                    StrHandlePutS(sh, ", ?");
                }
            }
            if (writeData) {
                StrHandlePutS(sh, ", ?");
                for (i = 0; i < sqlt->sqlt_NumFields; i++) {
                    StrHandlePutS(sh, ", ?");
                }
            }
            StrHandlePutS(sh, ")");
        }
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else {
            SQL(StrHandleString(sh));
            rc = sqlite3_prepare_v2(db, StrHandleString(sh), -1, pstmt, NULL);
            err = sqlite3_error(rc, TRUE);
        }
        FreeStrHandle(sh);
    }

    if (err != OK) {
        return err;
    }

    stmt = *pstmt;
    col = exists? 1: 2;
    if (writeHeader) {
        if (err == OK) {
            rc = sqlite3_bind_int(stmt, col++, matchFilter);
            err = sqlite3_error(rc, TRUE);
        }
        if (err == OK) {
            rc = sqlite3_bind_double(stmt, col++, orderNum);
            err = sqlite3_error(rc, TRUE);
        }
    }
    if (writeData) {
        struct Field *field;
        struct DataInfo di;
        long i;

        if (err == OK) {
            rc = sqlite3_bind_int(stmt, col++, version);
            err = sqlite3_error(rc, TRUE);
        }
        for (i = 0; i < sqlt->sqlt_NumFields; i++) {
            if (err != OK) {
                break;
            }
            field = sqlt->sqlt_Fields[i];

            if (!getDataInfo(data, field, &di)) {
                err = ERR_INTERNAL;
                break;
            }

            rc = sqlite3_bind_datainfo(stmt, col++, FieldType(field), &di,
                    memPool, &err);
            err = sqlite3_error(rc, TRUE);
        }
    }
    if (err == OK) {
        rc = sqlite3_bind_int(stmt, exists? col : 1, id);
        if (rc != SQLITE_OK) {
            err = ERR_INTERNAL;
        }
    }
    if (err == OK) {
        rc = sqlite3_step(stmt);
        if (rc == SQLITE_DONE) {
            rc = SQLITE_OK;
        }
        err = sqlite3_error(rc, TRUE);
    }
    rc = sqlite3_reset(stmt);
    if (rc == SQLITE_OK) {
        rc = sqlite3_clear_bindings(stmt);
    }
    if (err == OK) {
        err = sqlite3_error(rc, TRUE);
    }

    return err;
}

long SQLiteDeleteRecord(struct SQLiteHandle *handle, struct Table *tab, long id)
{
    struct Project *pr = TableProject(tab);
    struct SQLiteTable *sqlt;
    sqlite3 *db;
    sqlite3_stmt *stmt;
    long err = OK;
    int rc = SQLITE_OK;

    if (handle == NULL) {
        return ERR_INTERNAL;
    }
    db = handle->sqlh_Db;
    if (db == NULL) {
        return OK;              /* Nothing to delete */
    }

    sqlt = getSQLiteTable(handle, tab);
    if (sqlt == NULL) {
        return ERR_MEM;
    }

    if (sqlt->sqlt_Stmts.deleteRecord == NULL) {
        struct MemPool *memPool = ProjectMemPool(pr);
        struct StrHandle *sh;

        sh = AllocStrHandle(memPool);
        if (sh == NULL) {
            return ERR_MEM;
        }
        StrHandlePutS(sh, "DELETE FROM ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, " WHERE _ID = ?");
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else {
            SQL(StrHandleString(sh));
            rc = sqlite3_prepare_v2(db, StrHandleString(sh), -1, 
                    &sqlt->sqlt_Stmts.deleteRecord, NULL);
            err = sqlite3_error(rc, FALSE);
        }
        FreeStrHandle(sh);
    }

    if (err != OK) {
        return err;
    }

    stmt = sqlt->sqlt_Stmts.deleteRecord;
    rc = sqlite3_bind_int(stmt, 1, id);
    if (rc != SQLITE_OK) {
        err = ERR_INTERNAL;
    }
    if (err == OK) {
        rc = sqlite3_step(stmt);
        if (rc == SQLITE_DONE) {
            rc = SQLITE_OK;
        }
        err = sqlite3_error(rc, FALSE);
    }
    rc = sqlite3_reset(stmt);
    if (rc == SQLITE_OK) {
        rc = sqlite3_clear_bindings(stmt);
    }
    if (err == OK) {
        err = sqlite3_error(rc, TRUE);
    }

    return err;
}

long SQLiteDeleteRecords(struct SQLiteHandle *handle, struct Table *tab)
{
    struct Project *pr = TableProject(tab);
    struct SQLiteTable *sqlt;
    sqlite3 *db;
    sqlite3_stmt *stmt;
    long err = OK;
    int rc = SQLITE_OK;

    if (handle == NULL) {
        return ERR_INTERNAL;
    }
    db = handle->sqlh_Db;
    if (db == NULL) {
        return OK;              /* Nothing to delete */
    }

    sqlt = getSQLiteTable(handle, tab);
    if (sqlt == NULL) {
        return ERR_MEM;
    }

    if (sqlt->sqlt_Stmts.deleteRecords == NULL) {
        struct MemPool *memPool = ProjectMemPool(pr);
        struct StrHandle *sh;

        sh = AllocStrHandle(memPool);
        if (sh == NULL) {
            return ERR_MEM;
        }
        StrHandlePutS(sh, "DELETE FROM ");
        StrHandlePutC(sh, '"');
        StrHandlePutS(sh, TableName(tab));
        StrHandlePutC(sh, '"');
        StrHandlePutC(sh, 0);
        if (StrHandleError(sh)) {
            err = ERR_MEM;
        } else {
            SQL(StrHandleString(sh));
            rc = sqlite3_prepare_v2(db, StrHandleString(sh), -1, 
                    &sqlt->sqlt_Stmts.deleteRecords, NULL);
            err = sqlite3_error(rc, FALSE);
        }
        FreeStrHandle(sh);
    }

    if (err != OK) {
        return err;
    }

    stmt = sqlt->sqlt_Stmts.deleteRecords;
    rc = sqlite3_step(stmt);
    if (rc == SQLITE_DONE) {
        rc = SQLITE_OK;
    }
    err = sqlite3_error(rc, FALSE);
    rc = sqlite3_reset(stmt);
    if (err == OK) {
        err = sqlite3_error(rc, TRUE);
    }

    return err;
}

