/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** sqlitedb.h
**
** Creation date: 07.04.2009
** Last modified: 13.04.2021
*/

#ifndef SQLITEDB_H
#define SQLITEDB_H

#include "system.h"

struct SQLiteHandle;
struct SQLiteTable;
struct Project;
struct Table;
struct Field;
struct DataInfo;

#ifdef __cplusplus
extern "C" {
#endif

/*
** SQLiteHandle which every project needs.
*/
void ClearSQLiteHandle(struct SQLiteHandle *handle);
void FreeSQLiteHandle(struct SQLiteHandle *handle);
struct SQLiteHandle *AllocSQLiteHandle(struct Project *pr);

/*
** Cached statements for a table.
*/
void FreeSQLiteTable(struct SQLiteTable *sqlt);

/*
** Returns TRUE if changes to a project are allowed.  This either means
** there is no SQLite connection, or that the connection has a reserved lock.
*/
bool SQLiteAllowChanges(struct SQLiteHandle *handle);

/*
** Returns TRUE if there is a reserved lock on the the SQLite file from any
** connection (either external or the connection in the handle).
** If there is no SQLite connection then FALSE is returned.
*/
bool SQLiteHasReservedLock(struct SQLiteHandle *handle);

/*
** Returns TRUE if the SQLite connection has been changed from outside.
*/
bool SQLiteProjectHasChanged(struct SQLiteHandle *handle);

/*
** Returns TRUE if we should call SQLiteProjectHasChanged.
** Uses direct file access to the sqlite3 file for performing a quick check.
*/
bool SQLiteCheckForProjectChange(struct SQLiteHandle *handle);

/*
** Reads the project file version from the SQLite connection.
** If successful sets the version and returns TRUE, otherwise returns FALSE
** and does not modify version.
*/
bool SQLiteGetProjectFileVersion(struct SQLiteHandle *handle, long *version);

/*
** Release any (read or write) lock that has been kept.
*/
void SQLiteReleaseLock(struct SQLiteHandle *handle);

/*
** Prepare for changes to the SQLIte connection by starting an immediate 
** transaction.  If there is no SQLite connection, return OK.
*/
long SQLiteObtainChangeLock(struct SQLiteHandle *handle);

/* Deletes a table from the SQLIte connection. */
long SQLiteDeleteTable(struct SQLiteHandle *handle, struct Table *tab);

/* Adds a new table to the SQLite connection.  */
long SQLiteNewTable(struct SQLiteHandle *handle, struct Table *tab);

/* Renames a table in the SQLite connection. */
long SQLiteRenameTable(struct SQLiteHandle *handle, struct Table *tab,
        const char *newName);

/* Deletes a field from the SQLIte connection. */
long SQLiteDeleteField(struct SQLiteHandle *handle, struct Field *field);

/* Adds a new field to the SQLite connection.  */
long SQLiteNewField(struct SQLiteHandle *handle, struct Field *field);

/* Renames a field in the SQLite connection. */
long SQLiteRenameField(struct SQLiteHandle *handle, struct Field *field,
        const char *newName);

/*
** See project.h for the definition of the load & save methods.
** If the retryObtainWriteLock callback (may be NULL) returns TRUE then
** SQLite keeps trying to obtain a write lock.  Once it returns FALSE,
** it will be no longer called again.
** Similar for the retryVacuum callback.
*/
long SQLiteSaveProject(struct SQLiteHandle *handle, const char *filename,
        bool reorg, long projectVersion, long structureVersion, 
        bool (*retryObtainWriteLock)(struct Project *), 
        bool vacuum, bool (*retryVacuum)(struct Project *, long, const char *),
        char *errstr);
long SQLiteOpenProject(struct SQLiteHandle *handle, const char *filename,
        bool *update, bool *clearOnError, 
        long *projectVersion, long *structureVersion, char *errstr);

long SQLiteReadRecord(struct SQLiteHandle *handle, struct Table *tab, long id,
        char *data, struct MemPool *dataPool,
        void (*setDataInfo)(char *, struct Field *, const struct DataInfo *),
        size_t *pTotalWarnings);

long SQLiteGetOrderNum(struct SQLiteHandle *handle, struct Table *tab, 
        long id, double *orderNum);

long SQLiteWriteRecord(struct SQLiteHandle *handle, struct Table *tab,
        bool exists, long id, double orderNum, bool matchFilter,
        long version, const char *data,
        bool (*getDataInfo)(const char *, struct Field *, struct DataInfo *));

long SQLiteDeleteRecord(struct SQLiteHandle *handle, struct Table *tab,
        long id);
long SQLiteDeleteRecords(struct SQLiteHandle *handle, struct Table *tab);

#ifdef __cplusplus
}
#endif

#endif

