/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** strhandle.c
**
** Creation date: 12.08.1995
** Last modified: 12.07.2012
*/

#include "strhandle.h"
#include "util.h"
#include "error.h"

struct StrHandle
{
    struct MemPool *sh_MemPool;
    bool sh_Error;
    size_t sh_Size;
    size_t sh_Pos;
    char *sh_Buffer;
};

void FreeStrHandle(struct StrHandle *sh)
{
    if(sh)
    {
        if(sh->sh_Buffer)
            PoolFreeMem(sh->sh_MemPool, sh->sh_Buffer, sh->sh_Size);
        PoolFreeMem(sh->sh_MemPool, sh, sizeof(struct StrHandle));
    }
}

struct StrHandle *AllocStrHandle(struct MemPool *memPool)
{
    struct StrHandle *sh;

    sh = PoolAllocMem(memPool, sizeof(struct StrHandle));
    if(sh == NULL)
        return(NULL);

    sh->sh_MemPool = memPool;
    sh->sh_Error = FALSE;
    sh->sh_Size = 0;
    sh->sh_Pos = 0;
    sh->sh_Buffer = NULL;

    sh->sh_Size = 1024;
    sh->sh_Buffer = PoolAllocMem(memPool, sh->sh_Size);
    if(sh->sh_Buffer == NULL)
    {
        FreeStrHandle(sh);
        return(NULL);
    }

    return(sh);
}


char *StrHandleString(struct StrHandle *sh)
{
    char *str = NULL;

    if(sh)
        str = sh->sh_Buffer;
    return(str);
}

bool StrHandleError(struct StrHandle *sh)
{
    bool error = TRUE;

    if(sh)
        error = sh->sh_Error;
    return(error);
}

static bool strHandleExtendBuffer(struct StrHandle *sh)
{
    size_t size = sh->sh_Size;
    size_t newSize = size * 2;
    char *new = PoolReAllocMem(sh->sh_MemPool, sh->sh_Buffer, size, newSize);

    if (new == NULL)
    {
	sh->sh_Error = TRUE;
	return FALSE;
    }
    sh->sh_Buffer = new;
    sh->sh_Size = newSize;

    return TRUE;
}

bool StrHandlePutC(struct StrHandle *sh, char c)
{
    if (sh == NULL || sh->sh_Error)
        return FALSE;

    if (sh->sh_Pos >= sh->sh_Size)
	if (!strHandleExtendBuffer(sh))
	   return FALSE; 

    sh->sh_Buffer[sh->sh_Pos++] = c;
    return TRUE;
}

bool StrHandlePutS(struct StrHandle *sh, const char *s)
{
    if (sh == NULL || sh->sh_Error || s == NULL)
        return FALSE;
    else
    {
	size_t len = strlen(s);
	size_t pos = sh->sh_Pos;
	size_t size = pos + len;

	while (sh->sh_Size < size)
	    if (!strHandleExtendBuffer(sh))
		return FALSE;

	memcpy(sh->sh_Buffer + pos, s, len);
	sh->sh_Pos = size;
	return TRUE;
    }
}

static struct StrHandle *readStringHandle(FILE *file, struct MemPool *memPool)
{
    struct StrHandle *sh = NULL;
    int c;

    if (file != NULL) {
        sh = AllocStrHandle(memPool);
        do {
            c = fgetc(file);
            if (c == EOF) {
                c = 0;
            }
            StrHandlePutC(sh, c);
        } while(c != 0);

        if (StrHandleError(sh)) {
	    FreeStrHandle(sh);
	    sh = NULL;
	}
    }
    return sh;
}

char *ReadString(FILE *file, struct MemPool *memPool)
{
    struct StrHandle *sh = readStringHandle(file, memPool);
    char *result = NULL;

    if (sh != NULL) {
	result = AllocString(memPool, StrHandleString(sh));
        FreeStrHandle(sh);
    }
    return result;
}

char *ReadStringConvert(FILE *file, bool utf8, struct MemPool *memPool,
        long *perror, size_t *pwarnings)
{
    struct StrHandle *sh = readStringHandle(file, memPool);
    char *result = NULL, *src, *cs;
    long error = OK;
    size_t warnings = 0;

    if (sh == NULL) {
	error = ERR_MEM;
    } else {
	src = StrHandleString(sh);
        if (src != NULL && src[0] == 0) {
            cs = NULL;         /* Ignore empty strings */
        } else if (utf8) {
	    cs = CodesetConvertFromUTF8(src, memPool, &error, &warnings);
	} else {
	    cs = CodesetConvertFromDefault8Bit(src, memPool, &error);
	}
	if (error && cs) {
	    FreeString(memPool, cs);
	    cs = NULL;
	} else if (cs == src) {
	    cs = AllocString(memPool, src);
	    if (cs == NULL) {
		error = ERR_MEM;
	    }
	}
	result = cs;
	FreeStrHandle(sh);
    }
    if (perror) {
        *perror = error;
    }
    if (pwarnings) {
        *pwarnings = warnings;
    }
    return result;
}

char *ExpandPathname(const char *path, struct MemPool *memPool)
{
    struct StrHandle *sh;
    char *result = NULL, *copy;
    const char *s, *beg, *t, *env;

    if(path == NULL)
	return(NULL);
    
    sh = AllocStrHandle(memPool);
#ifndef AMIGA
    t = CodesetStrchr(path, ':');		/* Amiga style assign? */
    if (t && t != path && t[1] != '\\' && t[1] != '/')
    {
	for (s = path; s != t; s++)
	{
	    if (*s == '/' || *s == '\\' || *s == '$')
	    {
		break;
	    }
	}
	if (s == t)		/* 'valid' assign? */
	{
	    copy = AllocStringN(memPool, path, (long)(t - path)); 
	    env = getenv(copy);
	    if (env == NULL && copy != NULL && !CodesetStrCmp(copy, "BeeBase"))
	    {
		env = getenv("BeeBase");
	    }
	    if (env)
	    {
		StrHandlePutS(sh, env);
		StrHandlePutC(sh, PATH_SEP);
                path = t + 1;
	    }
	    FreeString(memPool, copy);
	}
    }
#endif
    for(s = path; *s; s++)
    {
       	if(*s != '$')
	{
  	    StrHandlePutC(sh, *s);
  	    continue;
	}
  	if(s[1] == '(')
	{
  	    beg = s+2;
  	    t = CodesetStrchr(s, ')');
	}
  	else if(s[1] == '{')
	{
  	    beg = s+2;
  	    t = CodesetStrchr(s, '}');
	}
  	else
	{
  	    beg = s+1;
	    
	    for (t = beg; *t; t = CodesetNextChar(t)) {
		if (!CodesetIsAlNum(CodesetGetChar(t))) {
		    break;
		}
	    }
	}
	
	if(t == NULL)            /* missing closing parenthesis, brace? */
	{
	    StrHandlePutS(sh, s);               /* copy as is and leave */
	    break;
	}
	
	/* create copy of env variable name */
	copy = AllocStringN(memPool, beg, (long)(t - beg)); 
	env = getenv(copy);
	FreeString(memPool, copy);
	
	if(*t == ')' || *t == '}')
	    t++;
	
	if(env == NULL)          /* env variable not defined? */
	{
	    while(s != t)
	       	StrHandlePutC(sh, *s++);         /* copy as is */
	}
	else
	    StrHandlePutS(sh, env);
	s = t-1;                          
    }
    StrHandlePutC(sh, 0);       /* terminate string */
    
    if(!StrHandleError(sh))
	result = AllocString(memPool, StrHandleString(sh));
    
    FreeStrHandle(sh);
    return(result);
}

