/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** structureeditor
**
** Creation date: 10.11.1994
** Last modified: 28.12.2021
*/

#include "system.h"
#include "table.h"
#include "lang/program.h"
#include "codeset.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "structureeditor.h"
#include "queryeditor.h"
#include "project.h"
#include "field.h"
#include "display.h"
#include "name.h"
#include "buildgui.h"
#include "gui.h"
#include "requesters.h"
#include "error.h"
#include "memory.h"
#include "menuprogram.h"
#include "programeditor.h"
#include "name.h"
#include "util.h"
#include "appstrings_cat.h"

#endif

#ifndef QT
void GuiUserFunctionsQuickSearch(GuiObject go, int dir, const char *str)
{
    if (str && str[0]) {
        long start = (dir == 0)? -1 : GuiListActive(go);
        long add = dir, cur;
        struct UserFunction *uf;
        const char *name;

        if (start < 0 || dir == 0) {
            start = -1;
            add = 1;
        }
        cur = start + add;
        while ((uf = GuiListNthEntry(go, cur)) != NULL) {
            name = UserFunctionName(uf);
            if (CodesetStrStr(name, str)) {
                break;
            }
            cur += add;
        }

        if (uf != NULL || dir == 0) {
            GuiListSetActive(go, cur);
        }
    }
}
#endif

bool TableSystemFilter(const struct Node *node)
{
    return !TableIsSystem((struct Table *)node);
}

#ifdef ENABLE_STRUCTURE_EDITOR

int DisplayNumCompareFunc(struct DisplayObject *dio1,struct DisplayObject *dio2)
{
    bool v1 = DisplayObjectVisible(dio1);
    bool v2 = DisplayObjectVisible(dio2);
    long num1 = DisplayObjectNum(dio1);
    long num2 = DisplayObjectNum(dio2);

    if (v1 != v2) {
	return (int)(v2 - v1);
    } else {
	return (int)(num1 - num2);
    }
}

static int tabCompare(struct Table *tab1, struct Table *tab2, 
        int (*cmp)(struct Table *tab1, struct Table *tab2))
{
    bool sys1 = TableIsSystem(tab1);
    bool sys2 = TableIsSystem(tab2);
    int rc;

    if (sys1 != sys2) {
        rc = sys1? 1 : -1;
    } else {
        rc = cmp(tab1, tab2);
    }
    return rc;
}

static int tabCompareAlphaTable(struct Table *tab1, struct Table *tab2)
{
    return CodesetStrCollate(TableName(tab1), TableName(tab2),CS_COLLATE_ORDER);
}

static int tabCompareDisplayTable(struct Table *tab1, struct Table *tab2)
{
    struct DisplayObject *dio1 = TableDisplayObject(tab1);
    struct DisplayObject *dio2 = TableDisplayObject(tab2);

    return DisplayNumCompareFunc(dio1, dio2);
}

int TableCompareAlpha(struct Table *tab1, struct Table *tab2)
{
    return tabCompare(tab1, tab2, tabCompareAlphaTable);
}

void SortTablesAlpha(struct List *tabs)
{
    QLSort(tabs, (int (*)(void *, void *))TableCompareAlpha);
}

int TableCompareDisplay(struct Table *tab1, struct Table *tab2)
{
    return tabCompare(tab1, tab2, tabCompareDisplayTable);
}

void SortTablesDisplay(struct List *tabs)
{
    QLSort(tabs, (int (*)(void *, void *))TableCompareDisplay);
}

int FieldCompareAlpha(struct Field *field1, struct Field *field2)
{
    return CodesetStrCollate(FieldName(field1), FieldName(field2), 
	    CS_COLLATE_ORDER);
}

void SortFieldsAlpha(struct List *fields)
{
    QLSort(fields, (int (*)(void *, void *))FieldCompareAlpha);
}

int FieldCompareDisplay(struct Field *field1, struct Field *field2)
{
    struct DisplayObject *dio1 = FieldDisplayObject(field1);
    struct DisplayObject *dio2 = FieldDisplayObject(field2);

    return DisplayNumCompareFunc(dio1, dio2);
}

void SortFieldsDisplay(struct List *fields)
{
    QLSort(fields, (int (*)(void *, void *))FieldCompareDisplay);
}

long SE_NewTable(struct Project *pr, struct TableData *rd,
                    struct DisplayObject *dioparent, struct Table **tabp)
{
    struct Table *tab = NULL;
    long error = 0;

    if (tabp) {
        *tabp = NULL;
    }

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        return error;
    }

    error = NewTable(pr, rd, &tab);
    if (error) {
	return error;
    }

    if (!DisplayNewTable(tab, dioparent)) {
	DeleteTable(tab);
	return ERR_MEM;
    }

    ProjectChange(pr, TRUE);
    if (tabp) {
        *tabp = tab;
    }
    return OK;
}

long SE_ChangeTable(struct Table *tab, struct TableData *rd)
{
    long error = 0;
    struct Project *pr = TableProject(tab);

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        return error;
    }

    error = ChangeTable(tab, rd);
    if (error != OK) {
	return error;
    }

    ProjectChange(pr, TRUE);
    return OK;
}

static bool checkProjectProgramObjectInUse(struct Project *pr, void *obj)
{
    struct Program *prg = ProjectProgram(pr);

    if (ProgramFindObject(prg, obj)) {
        NameSetUnknown(obj);
        ProgramEditorFix(pr);
        NameSetUnknown(NULL);
        prg = ProjectProgram(pr);
        if (ProgramFindObject(prg, obj)) {
            return TRUE;
        }
    }
    return FALSE;
}

static long SE_DeleteTable(struct Table *tab)
{
    struct DisplayObject *dio = TableDisplayObject(tab);
    struct Project *pr = TableProject(tab);
    long error = OK;

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        return error;
    }

    if (checkProjectProgramObjectInUse(pr, tab)) {
        return PROGERR_TABLEINUSE;
    }

    error = DeleteTable(tab);
    if (error) {
        return error;
    }

    DeleteDisplayObject(dio);
    StructureEditorUpdate(ProjectStructureEditor(pr));

    ProjectChange(pr, TRUE);
    return OK;
}

bool GuiDeleteTable(struct Table *tab)
{
    short answer;
    long error = OK;

    answer = Ask(AT_WARN, GCS(MSG_DeleteAnswers), GCS(MSG_DeleteTableText));
    if (answer != 0) {
	error = SE_DeleteTable(tab);
        if (error == OK) {
            return TRUE;
        }
  	Ask2(NULL, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk),GES(error));
    }
    return FALSE;
}

long SE_NewField(struct NewFieldData *nfd, struct DisplayObject *dioparent,
        struct Field **fieldp)
{
    long error = 0;
    struct Field *field = NULL;
    struct Project *pr = TableProject(nfd->nfd_Table);

    if (fieldp) {
        *fieldp = NULL;
    }

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        return error;
    }

    error = NewField(nfd, &field);
    if (error) {
	return error;
    }

    if (!DisplayNewField(field, dioparent)) {
	DeleteField(field);
	return ERR_MEM;
    }

    ProjectChange(pr, TRUE);
    if (fieldp) {
        *fieldp = field;
    }
    return OK;
}

long SE_ChangeField(struct ChangeFieldData *cfd)
{
    long error = 0;
    struct Table *tab = FieldTable(cfd->cfd_Field);
    struct Project *pr = TableProject(tab);

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        return error;
    }

    error = ChangeField(cfd);
    if (error) {
	return error;
    }

    ProjectChange(pr, TRUE);
    return OK;
}

static long SE_DeleteField(struct Field *field)
{
    struct DisplayObject *dio = FieldDisplayObject(field);
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    long error = OK;

    error = ObtainProjectChangeLock(pr);
    if (error != OK) {
        return error;
    }

    if (checkProjectProgramObjectInUse(pr, field)) {
        return PROGERR_FIELDINUSE;
    }

    error = DeleteField(field);
    if (error) {
        return error;
    }

    DeleteDisplayObject(dio);
    StructureEditorUpdate(ProjectStructureEditor(pr));

    ProjectChange(pr, TRUE);
    return OK;
}

bool GuiDeleteField(struct Field *field)
{
    short answer;
    long error = OK;

    answer = Ask(AT_WARN, GCS(MSG_DeleteAnswers), GCS(MSG_DeleteFieldText));
    if (answer != 0) {
        GuiSetSleep(TRUE);
	error = SE_DeleteField(field);
        GuiSetSleep(FALSE);

        if (error == OK) {
            return TRUE;
        }
        Ask2(NULL, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk),GES(error));
    }
    return FALSE;
}

void StructureEditorCheckNames(struct Project *pr)
{
    GuiObject se = ProjectStructureEditor(pr);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    struct List *fields;
    struct Field *field;
    long numInvalidNames = 0;

    if (se == NULL || tabs == NULL) {
        return;
    }

    FORNODE(tab, tabs) {
        if (!IsMBName(TableName(tab))) {
            numInvalidNames++;
        }

        fields = TableFields(tab);
        if (fields != NULL) {
            FORNODE(field, fields) {
                if (!IsMBName(FieldName(field))) {
                    numInvalidNames++;
                }
            }
        }
    }

    if (numInvalidNames == 0) {
        return;
    }

    if (Ask2(se, NULL, AT_WARN,
                GCS(MSG_StructureEditor_RenameTablesAndFieldsAnswers),
                GCS(MSG_StructureEditor_RenameTablesAndFields),
                numInvalidNames) == 0) {
        return;
    }

    StructureEditorNextNameChange(pr, NULL);
}

void StructureEditorNextNameChange(struct Project *pr, void *lastChangeObject)
{
    GuiObject se = ProjectStructureEditor(pr);
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;
    struct List *fields;
    struct Field *field;

    if (se == NULL || tabs == NULL) {
        return;
    }

    /*
    ** If the next object is the same as the last change object then the user
    ** canceled changing names.  In that case we stop the name-change process.
    */
    FORNODE(tab, tabs) {
        if (!IsMBName(TableName(tab))) {
            if (tab == lastChangeObject) {
                StructureEditorSetNameChangeObject(se, NULL);
            } else {
                StructureEditorSetNameChangeObject(se, tab);
                GuiChangeTable(se, tab);
            }
            return;
        }
    }

    FORNODE(tab, tabs) {
        fields = TableFields(tab);
        if (fields != NULL) {
            FORNODE(field, fields) {
                if (!IsMBName(FieldName(field))) {
                    if (field == lastChangeObject) {
                        StructureEditorSetNameChangeObject(se, NULL);
                    } else {
                        StructureEditorSetNameChangeObject(se, tab);
                        GuiChangeField(se, field);
                    }
                    return;
                }
            }
        }
    }

    /* We completed renaming all tables and fields */
    StructureEditorSetNameChangeObject(se, NULL);
}

static void structureEditorHelp(struct Project *pr)
{
    GuiObject se;

    if (pr == NULL) {				/* should never happen */
	return;
    }

    se = ProjectStructureEditor(pr);
    if (se == NULL) {
        /* try to open structure editor */
        CloseQueryEditor(pr);
    	se = OpenStructureEditor(pr);
    	if (se == NULL) {
	    Ask(AT_ERROR, GCS(MSG_StarOk),GCS(MSG_Err_CantOpenStructureEditor));
	    return;			    /* failed open structure editor */
	}
    	SetProjectStructureEditor(pr, se);  /* remember pointer */
        StoreDisplayFields(pr);             /* remember display fields */
    	FreeProjectGui(pr);		    /* close old gui */
        StructureEditorCheckNames(pr);
    } else {
        if (ProgramNeedRecompile(ProjectProgram(pr))) {
	    if (Ask(AT_QUESTION, 
			GCS(MSG_StructureEditor_RecompileProgramAnswers),
			GCS(MSG_StructureEditor_RecompileProgram))) {
		if (!MenuProgramCompile(pr, FALSE)) {
		    return;		/* user is forced to program editor */
		}
	    }
	}

	/* close it and try to open new gui */
        if (!BuildProjectGui(pr)) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_CantBuildGui));
	    return;
	}

	if (!OpenProjectWindows(pr)) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_CantOpenMainWindow));
	    FreeProjectGui(pr);
	    return;
	}

    	/* free structure editor and clear pointer in project structure */
    	CloseStructureEditor(se);
    	SetProjectStructureEditor(pr, NULL);
    }
}

void StructureEditor(struct Project *pr)
{
    structureEditorHelp(pr);
}

#else

void CloseStructureEditor(GuiObject se) {}
void ClearStructureEditor(GuiObject se) {}
void StructureEditorUpdate(GuiObject se) {}
void StructureEditorCheckNames(struct Project *pr) {}

#endif

