/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** system.h
**
** Creation date: 16.11.1994
** Last modified: 01.08.2022
*/

#ifndef SYSTEM_H
#define SYSTEM_H

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>
#include "types.h"

#define ID_NAME_LEN 20

#define BEEBASE

#define MSG_APP MSG_BeeBase
#define APP_NAME "BeeBase"

#ifndef BEEBASE_LIGHT
    #define ENABLE_ADMIN_MODE
    #define ENABLE_STRUCTURE_EDITOR
    #define ENABLE_PROGRAM_EDITOR
    #define ENABLE_LOG_APPLY_CHANGES
#endif

/* #define USE_LOG_WINDOW */

#if defined AMIGA    		/* m68k, AOS4, MorphOS, AROS */

    #if (defined __GNUC__)	/* GCC for m68k, AOS4, MorphOS, AROS */
	#if (defined __amigaos4__) || (defined __MORPHOS__)|| (defined __AROS__)
            #if (!defined __amigaos4__)
                #define REG(r, x) x
            #endif
            #define SAVEDS
            #define STDARGS
            #define INTERRUPT
	    #if (defined __MORPHOS__)
		#define VARARGS68K __attribute__((varargs68k))
	    #endif
	#else
            #define REG(r, x) x __asm(#r)
            #define SAVEDS __saveds
	    #define STDARGS __stdargs
	    #define INTERRUPT __interrupt
	#endif
        #define ASM
    #endif

    #include <clib/macros.h>
    #include <exec/exec.h>
    #include <proto/exec.h>
    #include <proto/dos.h>

    #ifdef __GNUC__
        #include <clib/alib_protos.h>
    #endif

    #include <libraries/locale.h>
 
    #if !defined(__AROS__) && !defined(__MORPHOS__) && !defined(IPTR)
        #define IPTR ULONG
    #endif
    #if !defined(__AROS__) && !defined(__MORPHOS__) && !defined(SIPTR)
        #define SIPTR LONG
    #endif

    #if (!defined __amigaos4__) &&  (!defined __AROS__)
	/* m68k-amigaos and MorphOS do not have round in libm */
        #define round(x) floor((x) + 0.5)
    #endif

    #define STACK_SIZE 32000                /* AOS4 needs more than 10K */
    #define STACK_SIZE_STR "32k"

    #define OS_TYPE "Amiga"
    #define PATH_SEP '/'
    #define MAX_PATHNAME 512
    #define SYSTEM_COPY "copy FROM \"%s\" TO \"%s\" CLONE QUIET"
    #define PRINTER_FILE "PRT:"
    #define DEFAULT_OUTPUT_FILE "Console:"
    #define TMPDIR "T:"
    #define CURRENTDIR ""
    #define USE_EXTERNALVIEWER

    #define FPRINTF(x) fprintf x

    typedef const struct FC_Type *CatalogString;
    #define GCS(ref) GetString(ref)

    #ifdef MUI
        #include "mui/mui.h"
    #endif

    #if (defined __MORPHOS__)
        #define USE_HPDF
        /* #define USE_SQLITE_LIBRARY */
        /* #define __MORPHOS_SHAREDLIBS */
    #else
        #define USE_ANSI_PRINT
    #endif

#elif defined UNIX

    #include <generic/list.h>
    #include <generic/mempool.h>
    #include <base/dosutil.h>

    #define OS_TYPE "Unix"
    #define PATH_SEP '/'
    #define SYSTEM_COPY "cp -p \"%s\" \"%s\""
    #define PRINTER_FILE "|lpr"
    #define DEFAULT_OUTPUT_FILE DefaultOutputFilename()
    #define TMPDIR "/tmp/"
    #define CURRENTDIR "."
    #define O_BINARY 0		/* compatibility to Windows */

    #define FPRINTF(x) fprintf x

    // #define USE_HPDF

    #if defined QT
	#include "qt/qtGui.h"
    #elif defined GTK
	#define USE_HPDF
	#define USE_GTK_PRINT
        #define USE_EXTERNALVIEWER
	#include "gtk/gtkGui.h"
    #elif defined FOO
	#include "foo/foo.h"
    #endif

#elif defined WINDOWS

    #ifdef _MSC_VER                     /* Microsoft Visual C */
        #include <io.h>                 /* dup */

        #define popen _popen            /* why MSVC doesn't expose popen? */
        #define strcasecmp stricmp

        #pragma warning (disable:4200)  /* kill warns about zero-sized arrays */
    #endif
    #include <sys/stat.h>
    #include <generic/list.h>
    #include <generic/mempool.h>

    #define OS_TYPE "Windows"
    #define PATH_SEP '\\'
    #define SYSTEM_COPY "copy \"%s\" \"%s\""
    #define PRINTER_FILE "|lpr"
    #define DEFAULT_OUTPUT_FILE	"CON"
    #define TMPDIR "$TMP\\"
    #define CURRENTDIR "."

    #if defined QT
	#include "qt/qtGui.h"
    #elif defined GTK
	#define USE_GTK_PRINT
        #define USE_EXTERNALVIEWER
	#include "gtk/gtkGui.h"
    #elif defined FOO
	#include "foo/foo.h"
    #endif

#elif defined MACOSX

    #include <generic/list.h>
    #include <generic/mempool.h>

    #define OS_TYPE "MacOSX"
    #define PATH_SEP '/'
    #define SYSTEM_COPY "cp -p \"%s\" \"%s\""
    #define PRINTER_FILE "|lpr"
    #define DEFAULT_OUTPUT_FILE	"/dev/stdout"
    #define TMPDIR "/tmp/"
    #define CURRENTDIR "."
    #define O_BINARY 0		/* compatibility to Windows */

    #define FPRINTF(x) fprintf x

    #if defined QT
	#include "qt/qtGui.h"
        #define USE_EXTERNALVIEWER
    #elif defined GTK
	#define USE_GTK_PRINT
        #define USE_EXTERNALVIEWER
	#include "gtk/gtkGui.h"
    #elif defined FOO
	#include "foo/foo.h"
    #endif
#endif

#ifndef MAKE_ID
    #define MAKE_ID(a,b,c,d) \
	((ulong) (a)<<24 | (ulong) (b)<<16 | (ulong) (c)<<8 | (ulong) (d))
#endif

#ifndef MIN
    #define MIN(a,b) (((a) < (b))? (a) : (b))
#endif
#ifndef MAX
    #define MAX(a,b) (((a) > (b))? (a) : (b))
#endif

#define ALLOCMEM(size) malloc(size)
#define FREEMEM(ptr, size) free(ptr)

#if (defined __GNUC__)
    // GNUC has a nice extension for obtaining the type of an expression
    #define nodecast(var, node) ((__typeof__(var))(node))
#elif (defined  __cplusplus)
    // other C++ compilers, e.g. MS Visual C++ do not have typeof
    // so we have to define a cast function for each type manually ...
    #define NEED_NODE_CAST
    // ... and they look like this:
    inline Node *nodecast(Node *, Node *node) { return node; }
    inline List *nodecast(List *, Node *node) { return (List *)node; }
#else
    // In plain C we can simply cast using void *
    #define nodecast(var, node) ((void *)(node))
#endif

#define HEAD(list)	((list)->lh_Head)
#define TAIL(list)	((list)->lh_Tail)
#define TAILPRED(list)	((list)->lh_TailPred)
#define SUCC(node)	nodecast(node, ((struct Node *)(node))->ln_Succ)
#define PRED(node)	nodecast(node, ((struct Node *)(node))->ln_Pred)

#ifndef INSERT
    #define INSERT(list, node, pred)  \
        Insert(list, (struct Node *)(node), (struct Node *)(pred))
#endif

#if (!defined __MORPHOS__) && (!defined __AROS__) 
    #define REMOVE(node)        Remove((struct Node *)(node))
    #define REMHEAD(list)       RemHead(list)
    #define REMTAIL(list)       RemTail(list)
    #define ADDHEAD(list, node)	AddHead(list, (struct Node *)(node))
    #define ADDTAIL(list, node)	AddTail(list, (struct Node *)(node))
#endif

#define FORNODE(node, list) \
    for (node = nodecast(node = NULL, HEAD(list)); \
            (SUCC(node)) != NULL; \
            node = SUCC(node))

#define SAFEFORNODE(node, succ, list) \
    for (node = nodecast(node = NULL, HEAD(list)); \
            (succ = SUCC(node)) != NULL; \
            node = succ)

#ifdef DEBUG
    #ifdef __AROS__
    	#define DPRINTF(x) bug x
    #else
	#define DPRINTF(x) printf x
    #endif
    #define DFPRINTF(x) fprintf x
#else
    #define DPRINTF(x)
    #define DFPRINTF(x)
#endif

#define nyi(x) DPRINTF(("Not yet implemented: %s\n", # x));

#endif

