/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** table.c
**
** Creation date: 15.11.1994
** Last modified: 29.12.2020
*/

#include "system.h"
#include "project.h"
#include "name.h"
#include "field.h"
#include "record.h"
#include "table.h"
#include "order.h"
#include "filter.h"
#include "lang/program.h"
#include "queryeditor.h"
#include "sqlitedb.h"
#include "memory.h"
#include "codeset.h"
#include "dosutil.h"
#include "util.h"
#include "error.h"
#include "break.h"

struct Table
{
    struct Node tab_Node;
    struct Project *tab_Project;
    struct SQLiteTable *tab_SQLiteTable;
    char tab_Name[ID_NAME_LEN + 1];		/* ASCII */
    bool tab_ExactlyOneRecord;
    char tab_NewTrigger[TABLE_TRIGGER_LEN + 1];	/* ASCII */
    char tab_DelTrigger[TABLE_TRIGGER_LEN + 1];	/* ASCII */
    bool tab_CountChanges;
    bool tab_LogChanges;
    struct List tab_Fields;
    struct List tab_Records;
    struct DisplayObject *tab_DisplayObject;
    long tab_GuiRecordInternalNum;              /* only used during loading */
    long tab_GuiRecordAltInternalNum;           /* ditto */
    struct Record *tab_GuiRecord;               /* used by module Gui */
    bool tab_GuiFlag;                           /* used by module Gui */
    struct Record *tab_ProgramRecord;           /* used inside programs */
    struct Program *tab_FilterProgram;
    bool tab_FilterActive;
    struct Order *tab_Order;
    struct TableRecordInfo *tab_TableRecordInfo;
};

void FreeTable(struct Table *tab, bool quick)
{
    if (tab != NULL) {
        struct Record *rec;
        struct Field *field;

        FreeOrder(tab->tab_Order);
        tab->tab_Order = NULL;
        FreeProgram(tab->tab_FilterProgram);
        tab->tab_FilterProgram = NULL;

        while ((rec = (struct Record *)REMHEAD(&tab->tab_Records)) != NULL) {
            FreeRecord(tab, rec, quick);
        }
        tab->tab_GuiRecord = NULL;
        tab->tab_ProgramRecord = NULL;

        while ((field=(struct Field *)REMHEAD(&tab->tab_Fields))!=NULL) {
            FreeField(field);
        }
        
        FreeTableRecordInfo(tab->tab_TableRecordInfo);
        tab->tab_TableRecordInfo = NULL;
        FreeSQLiteTable(tab->tab_SQLiteTable);
        tab->tab_SQLiteTable = NULL;

        PoolFreeMem(ProjectMemPool(tab->tab_Project), tab,sizeof(struct Table));
    }
}

static struct Table *AllocTable(struct Project *pr)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct Table *tab;

    tab = PoolAllocMemZero(memPool, sizeof(struct Table));
    if (tab == NULL) {
	return NULL;
    }

    tab->tab_Project = pr;
    NewList(&tab->tab_Fields);
    NewList(&tab->tab_Records);

    tab->tab_TableRecordInfo = AllocTableRecordInfo(tab);
    if (tab->tab_TableRecordInfo == NULL) {
        FreeTable(tab, FALSE);
        return NULL;
    }
    return tab;
}

struct Project *TableProject(struct Table *tab)
{
    struct Project *pr = NULL;

    if(tab)
	pr = tab->tab_Project;
    return(pr);
}

bool TableIsSystem(struct Table *tab)
{
    bool rc = FALSE;

    if (tab) {
	rc = IsSystemName(tab->tab_Name);
    }
    return rc;
}

struct SQLiteTable *TableSQLiteTable(struct Table *tab)
{
    struct SQLiteTable *sqlt = NULL;

    if (tab != NULL) {
        sqlt = tab->tab_SQLiteTable;
    }
    return sqlt;
}

void SetTableSQLiteTable(struct Table *tab, struct SQLiteTable *sqlt)
{
    if (tab != NULL) {
        tab->tab_SQLiteTable = sqlt;
    }
}

struct List *TableFields(struct Table *tab)
{
    struct List *fields = NULL;

    if(tab)
	fields = &tab->tab_Fields;

    return(fields);
}

long TableNumFields(struct Table *tab)
{
    struct List *fields = TableFields(tab);
    struct Field *field;
    long num = 0;

    if(fields)
        FORNODE(field, fields)
            num++;
    return(num);
}

struct List *TableRecords(struct Table *tab)
{
    struct List *recs = NULL;

    if(tab)
	recs = &tab->tab_Records;

    return(recs);
}

struct MemPool *TableMemPool(struct Table *tab)
{
    struct MemPool *pool = NULL;

    if(tab)
	pool = ProjectMemPool(tab->tab_Project);
    return(pool);
}

const char *TableName(struct Table *tab)
{
    const char *name = NULL;

    if (tab) {
	name = tab->tab_Name;
    }
    return name;
}

static void SetTableName(struct Table *tab, const char *name)
{
    if (tab && name) {
	CodesetStrNBytesCopy(tab->tab_Name, name, ID_NAME_LEN);
    }
}

bool TableExactlyOneRecord(struct Table *tab)
{
    bool exactlyOneRecord = FALSE;

    if(tab)
        exactlyOneRecord = tab->tab_ExactlyOneRecord;
    return(exactlyOneRecord);
}

struct Table *FindTableName(struct Project *pr, const char *name)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    if (tabs && name) {
    	FORNODE (tab, tabs) {
	    if (strcmp(name, tab->tab_Name) == 0) {
		return tab;
	    }
	}
    }
    return NULL;
}

void SetTableDisplayObject(struct Table *tab, struct DisplayObject *dio)
{
    if(tab)
	tab->tab_DisplayObject = dio;
}

struct DisplayObject *TableDisplayObject(struct Table *tab)
{
    struct DisplayObject *dio = NULL;

    if(tab)
	dio = tab->tab_DisplayObject;
    return(dio);
}

struct Record *TableGuiRecord(struct Table *tab)
{
    struct Record *rec = NULL;

    if(tab)
        rec = tab->tab_GuiRecord;
    return(rec);
}

void SetTableGuiRecord(struct Table *tab, struct Record *rec)
{
    if(tab)
        tab->tab_GuiRecord = rec;
}

bool TableGuiFlag(struct Table *tab)
{
    bool flag = FALSE;

    if(tab)
        flag = tab->tab_GuiFlag;
    return(flag);
}

void SetTableGuiFlag(struct Table *tab, bool flag)
{
    if(tab)
        tab->tab_GuiFlag = flag;
}

struct Record *TableProgramRecord(struct Table *tab)
{
    struct Record *rec = NULL;

    if(tab)
        rec = tab->tab_ProgramRecord;
    return(rec);
}

void SetTableProgramRecord(struct Table *tab, struct Record *rec)
{
    if(tab)
        tab->tab_ProgramRecord = rec;
}

struct Program *TableFilterProgram(struct Table *tab)
{
    struct Program *prg = NULL;

    if(tab)
        prg = tab->tab_FilterProgram;
    return(prg);
}

void SetTableFilterProgram(struct Table *tab, struct Program *prg)
{
    if (tab) {
        FreeProgram(tab->tab_FilterProgram);
        tab->tab_FilterProgram = prg;
    }
}

bool TableFilterActive(struct Table *tab)
{
    bool active = FALSE;

    if(tab)
        active = tab->tab_FilterActive;
    return(active);
}

void SetTableFilterActive(struct Table *tab, bool active)
{
    if(tab)
        tab->tab_FilterActive = active;
}

struct Order *TableOrder(struct Table *tab)
{
    struct Order *ord = NULL;

    if(tab)
        ord = tab->tab_Order;
    return(ord);
}

void SetTableOrder(struct Table *tab, struct Order *ord)
{
    if(tab)
        tab->tab_Order = ord;
}

struct TableRecordInfo *TableRecordInfo(struct Table *tab)
{
    struct TableRecordInfo *rri = NULL;

    if(tab)
        rri = tab->tab_TableRecordInfo;
    return(rri);
}

char *TableNewTrigger(struct Table *tab)
{
    char *trigger = NULL;

    if(tab)
        trigger = tab->tab_NewTrigger;
    return(trigger);
}

char *TableDelTrigger(struct Table *tab)
{
    char *trigger = NULL;

    if(tab)
        trigger = tab->tab_DelTrigger;
    return(trigger);
}

bool TableCountChanges(struct Table *tab)
{
    bool count = FALSE;

    if (tab) {
        count = tab->tab_CountChanges;
    }
    return count;
}

bool TableLogChanges(struct Table *tab)
{
    bool log = FALSE;

    if (tab) {
        log = tab->tab_LogChanges;
    }
    return log;
}

void AddField(struct Table *tab, struct Field *field, 
        struct Field *pred)
{
    if (tab && field)
    {
        if (FieldTable(pred) == tab)
            INSERT(&tab->tab_Fields, field, pred);
        else
            ADDTAIL(&tab->tab_Fields, field);
    }
}

void RemField(struct Table *tab, struct Field *field)
{
    if(tab && field)
    {
        OrderRemField(tab->tab_Order, field);
	REMOVE(field);
    }
}

void AddRecord(struct Table *tab, struct Record *rec, struct Record *pred)
{
    struct List *recs = TableRecords(tab);

    if (recs && rec) {
        if(pred == ADDRECORD_HEAD) {
            ADDHEAD(recs, rec);
	} else if(pred == ADDRECORD_TAIL) {
            ADDTAIL(recs, rec);
	} else {
            INSERT(recs, rec, pred);
	}
    }
}

void RemRecord(struct Table *tab, struct Record *rec)
{
    /* Do not execute time consuming jobs here! */

    if(tab && rec)
        REMOVE(rec);
}

long TableNumRecords(struct Table *tab)
{
    long count = 0;
    struct List *recs = TableRecords(tab);
    struct Record *rec;

    if(recs)
        FORNODE(rec, recs)
            count++;
    return(count);
}

long NewTable(struct Project *pr, struct TableData *td, struct Table **newtab)
{
    struct Table *tab = NULL;
    long error = OK;

    if (newtab) {
	*newtab = NULL;
    }

    if (pr == NULL || td == NULL || td->td_Name == NULL) {
        return ERR_INTERNAL;
    }
    if (!IsMBName(td->td_Name) || IsSystemName(td->td_Name) != td->td_System) {
	return ERR_INVALIDNAME;
    }
    if (TableNameInUse(pr, td->td_Name, NULL)) {
	return ERR_NAMEALREADYUSED;
    }

    tab = AllocTable(pr);
    if (tab == NULL) {
	return ERR_MEM;
    }

    SetTableName(tab, td->td_Name);

    if (td->td_ExactlyOneRecord) {
        error = NewRecord(tab, NULL, NULL, TRUE, NULL);
        if (error) {
            FreeTable(tab, FALSE);
            return error;
        }
    }

    error = SQLiteNewTable(ProjectSQLiteHandle(pr), tab);
    if (error != OK) {
        FreeTable(tab, FALSE);
        return error;
    }

    tab->tab_ExactlyOneRecord = td->td_ExactlyOneRecord;
    CodesetStrNBytesCopy(tab->tab_NewTrigger, td->td_NewTrigger,
	    TABLE_TRIGGER_LEN);
    CodesetTrimStr(tab->tab_NewTrigger);
    CodesetStrNBytesCopy(tab->tab_DelTrigger, td->td_DelTrigger,
	    TABLE_TRIGGER_LEN);
    CodesetTrimStr(tab->tab_DelTrigger);
    tab->tab_CountChanges = td->td_CountChanges;
    tab->tab_LogChanges = td->td_LogChanges;

    AddTable(pr, tab);

    if (newtab) {
	*newtab = tab;
    }
    return OK;
}

long ChangeTable(struct Table *tab, struct TableData *td)
{
    struct Project *pr = TableProject(tab);
    long numrecords, error = OK;

    if (tab == NULL || td == NULL || td->td_Name == NULL) {
        return OK;
    }

    if (!IsMBName(td->td_Name) || IsSystemName(td->td_Name) != td->td_System) {
        return ERR_INVALIDNAME;
    }
    if (TableNameInUse(TableProject(tab), td->td_Name, tab)) {
        return ERR_NAMEALREADYUSED;
    }

    numrecords = TableNumRecords(tab);
    if (td->td_ExactlyOneRecord && numrecords > 1) {
        return TABLEERR_EXACTLYONERECORD;
    }
    if (td->td_ExactlyOneRecord && numrecords == 0) {
        error = NewRecord(tab, NULL, NULL, TRUE, NULL);
        if (error) {
            return error;
	}
    }
    if (strcmp(TableName(tab), td->td_Name) != 0) {
        error = SQLiteRenameTable(ProjectSQLiteHandle(pr), tab, td->td_Name);
        if (error) {
            return error;
        }
        SetTableName(tab, td->td_Name);
    }
    tab->tab_ExactlyOneRecord = td->td_ExactlyOneRecord;
    CodesetStrNBytesCopy(tab->tab_NewTrigger, td->td_NewTrigger,
	    TABLE_TRIGGER_LEN);
    CodesetTrimStr(tab->tab_NewTrigger);
    CodesetStrNBytesCopy(tab->tab_DelTrigger, td->td_DelTrigger,
	    TABLE_TRIGGER_LEN);
    CodesetTrimStr(tab->tab_DelTrigger);
    tab->tab_CountChanges = td->td_CountChanges;
    tab->tab_LogChanges = td->td_LogChanges;

    return OK;
}

long DeleteTable(struct Table *tab)
{
    struct Project *pr = TableProject(tab);
    struct SQLiteHandle *handle = ProjectSQLiteHandle(pr);
    long error;

    if (tab == NULL) {
	return OK;
    }

    if (FindReferenceTable(tab) != NULL) {
        return TABLEERR_REFERENCE;
    }

    error = SQLiteDeleteTable(handle, tab);
    if (error) {
        return error;
    }

    FilterDeleteTable(tab);
    QueryEditorDeleteTable(tab);
    ProgramRemTable(tab);
    RemTable(TableProject(tab), tab);
    FreeTable(tab, FALSE);

    return OK;
}

long TableNum(struct Table *tab)
{
    return(NodeNum(tab));
}

struct Table *FindTableNum(struct Project *pr, long find)
{
    return(FindNodeNum(ProjectTables(pr), find));
}

/*
** Load & Save of tables
*/

long NumTables(struct Project *pr)
{
    return(NumNodes(ProjectTables(pr)));
}

void SetTablesGuiRecords(struct Project *pr)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    if (tabs != NULL) {
        FORNODE (tab, tabs) {
            struct Record *rec;
           
            rec = FindRecordInternalNum(tab, tab->tab_GuiRecordInternalNum);
            if (rec == NULL) {
                rec=FindRecordInternalNum(tab,tab->tab_GuiRecordAltInternalNum);
            }
            if (rec != NULL) {
                tab->tab_GuiRecord = rec;
            }
        }
    }
}

long ReadTable(struct Project *pr, const struct ReadTableData *data)
{
    struct Table *tab;

    if (pr == NULL || data == NULL) {
        return ERR_INTERNAL;
    }

    tab = AllocTable(pr);
    if (tab == NULL) {
        return ERR_MEM;
    }

    SetTableName(tab, data->rtd_Name);
    tab->tab_ExactlyOneRecord = data->rtd_ExactlyOneRecord;
    tab->tab_FilterActive = data->rtd_FilterActive;
    CodesetStrNBytesCopy(tab->tab_NewTrigger, data->rtd_NewTrigger,
            TABLE_TRIGGER_LEN);
    CodesetStrNBytesCopy(tab->tab_DelTrigger, data->rtd_DeleteTrigger,
            TABLE_TRIGGER_LEN);
    tab->tab_CountChanges = data->rtd_CountChanges;
    tab->tab_LogChanges = data->rtd_LogChanges;
    tab->tab_GuiRecordInternalNum = data->rtd_GuiRecordID;
    AddTable(pr, tab);

    if (TableNum(tab) != data->rtd_ID) {
        return DISKERR_NOTABEEBASEPROJECT;
    }

    return OK;
}

void UpdateTable(struct Table *tab, long guiRecordID)
{
    if (tab != NULL) {
        tab->tab_GuiRecordInternalNum = RecordInternalNum(tab->tab_GuiRecord);
        tab->tab_GuiRecordAltInternalNum = guiRecordID;
    }
}

/*
** Custom format
*/

#define TABLEVERSION 6

#define MIN_TABLEVERSION 4

long LoadTablesCustomFormat(struct Project *pr, FILE *file)
{
    long i;
    short version;
    long num;

    if (pr == NULL || file == NULL) {
    	return ERR_INTERNAL;
    }

    version = FGetShort(file);
    if (version > TABLEVERSION || version < MIN_TABLEVERSION) {
	return DISKERR_VERSION;
    }

    num = FGetLong(file);
    for (i = 0; i < num; i++) {
	struct Table *tab;
 	static char name[sizeof(tab->tab_Name)];

	FGetS(file, name, sizeof(name));
        tab = AllocTable(pr);
        if (tab == NULL) {
            return ERR_MEM;
        }

        SetTableRecordInfoCustomFormat(tab->tab_TableRecordInfo, TRUE);
	SetTableName(tab, name);
	tab->tab_ExactlyOneRecord = FGetBool(file);
	tab->tab_FilterActive = FGetBool(file);
	FGetS(file, tab->tab_NewTrigger, sizeof(tab->tab_NewTrigger));
	FGetS(file, tab->tab_DelTrigger, sizeof(tab->tab_DelTrigger));
        if (version >= 5) {
            tab->tab_CountChanges = FGetBool(file);
            tab->tab_LogChanges = FGetBool(file);
        } else {
            tab->tab_CountChanges = TRUE;
            tab->tab_LogChanges = TRUE;
        }
        if (version >= 6) {
            tab->tab_GuiRecordInternalNum = FGetLong(file);
        }
        AddTable(pr, tab);
    }

    return OK;
}

