/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** textfile.c
**
** Reading and writing of text files with conversiom and encoding detection.
**
** Creation date: 31.01.2009
** Last modified: 05.02.2015
*/

#include "system.h"
#include "textfile.h"
#include "requesters.h"
#include "dosutil.h"
#include "strhandle.h"
#include "codeset.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#define TEXT_BUFSIZE 8192
#define TEXT_MAXBYTESFORCHAR 6  /* UTF-8 might use up to 6 bytes for 1 char */

#define UTF8_BOM "\xEF\xBB\xBF"

struct _TextFile
{
    FILE *file;                         /* level 2 file for raw read/write */
    struct MemPool *memPool;            /* get memory from this pool */
    int tfe;                            /* text file encoding */
    bool start;                         /* we are at start of file */
    char buffer[TEXT_BUFSIZE+TEXT_MAXBYTESFORCHAR];       /* I/O buffer */
    size_t bsize;                       /* number of bytes in buffer */
    size_t bused;                       /* number bytes used for conversion */
    char *conv;                         /* conversion of buffer */
    size_t clen;                        /* number bytes in conv */
    size_t cindex;                      /* current index in conv */
    size_t warnings;                    /* total of conversion warnings */
};

long TextFileClose(TextFile *tf)
{
    int rc = 0;

    if (tf) {
        rc = fclose(tf->file);
        if (tf->conv != tf->buffer) {
            FreeString(tf->memPool, tf->conv);
        }
        PoolFreeMem(tf->memPool, tf, sizeof(TextFile));
    }

    return (rc? DISKERR_WRITE : OK);
}

TextFile *TextFileOpen(struct MemPool *memPool, 
        const char *filename, const char *mode, int tfe)
{
    TextFile *tf = PoolAllocMemZero(memPool, sizeof(TextFile));
    bool append;
    fpos_t pos;
    struct stat st;

    if (tf == NULL) {
        return NULL;
    }

    tf->memPool = memPool;
    tf->file = IFOpen(memPool, filename, mode);
    if (tf->file == NULL) {
        PoolFreeMem(memPool, tf, sizeof(TextFile));
        return NULL;
    }
    append = (mode[0] == 'a');
    tf->start = !append;

    tf->tfe = tfe;
    if (tfe == TFE_AUTO && fstat(fileno(tf->file), &st) == 0 &&
                S_ISREG(st.st_mode) &&
                fgetpos(tf->file, &pos) == 0 &&    /* can seek? */
            fseek(tf->file, 0, SEEK_SET) == 0) {
        char *s = tf->buffer;
        const char *end;
        size_t size = 0, used = 0;
        bool ascii = TRUE;
        bool utf8 = TRUE;
        bool start = TRUE;
        
        for (;;) {
            if (sizeof(tf->buffer) - size > TEXT_BUFSIZE) {
                size += fread(s + size, 1, TEXT_BUFSIZE, tf->file);
            }
            if (size == 0) {                    /* EOF or read failed? */
                break;
            }
            s[size] = 0;
            if (s[0] == 0) {
                break;                          /* not a text file */
            } else if (start && size >= 3 && !memcmp(s, UTF8_BOM, 3)) {
                ascii = FALSE;                  /* treat as UTF8 */
                break;
            } else if (StrIsASCII(s)) {
                used = strlen(s);               /* read more */
            } else if (StrIsUTF8AllowTrail(s, &end) && end > s) {
                used = end - s;                 /* read more */
                ascii = FALSE;
            } else {
                utf8 = FALSE;                   /* not UTF8 */
                ascii = FALSE;
                break;
            }
            memmove(tf->buffer, tf->buffer + used, size - used);
            size -= used;
            start = FALSE;
        }
        if (ascii) {
            /* no change, keep TFE_AUTO */
        } else if (utf8) {
            tf->tfe = TFE_UTF8;
        } else if (!LocaleIsUTF8()) {
            tf->tfe = TFE_LOCALE;
        } else {
            tf->tfe = TFE_8BIT;
        }
        clearerr(tf->file);             /* forget any errors */
	if (append) {
	    fseek(tf->file, 0, SEEK_END);
	} else {
	    fsetpos(tf->file, &pos);    /* restore file position */
	}
    }
    clearerr(tf->file);                 /* forget any errors */

    return tf;
}

size_t TextFileWarnings(TextFile *tf)
{
    size_t warnings = 0;

    if (tf) {
        warnings = tf->warnings;
    }
    return warnings;
}

long TextFileFlush(TextFile *tf)
{
    long error = OK;

    if (tf && tf->file && fflush(tf->file) != 0) {
        error = DISKERR_WRITE;
    }
    return error;
}

bool TextFileError(TextFile *tf)
{
    bool error = FALSE;

    if (tf && tf->file && ferror(tf->file)) {
        error = TRUE;
    }
    return error;
}

bool TextFileEOF(TextFile *tf)
{
    bool eof = FALSE;

    if (tf && tf->file && tf->cindex >= tf->clen && tf->bused >= tf->bsize && 
                feof(tf->file)) {
        eof = TRUE;
    }
    return eof;
}

bool TextFileSeek(TextFile *tf, long offset, int whence)
{
    if (tf == NULL || tf->file == NULL) {
        return FALSE;
    }
    if (whence == SEEK_CUR) {
        offset -= tf->bsize - tf->bused;

        if (tf->cindex < tf->clen) {
            const char *s = tf->conv, *mid = s + tf->cindex, *end= s + tf->clen;

            if (tf->tfe != TFE_NONE) {
                return FALSE;               /* still characters in conv */ 
            }

            while (s < mid) {
                s = CodesetNextChar(s);
            }
            while (s < end) {
                offset--;
                s = CodesetNextChar(s);
            }
        }
    }
    if (fseek(tf->file, offset, whence) != 0) {
        return FALSE;
    }

    if (tf->conv != tf->buffer) {
        FreeString(tf->memPool, tf->conv);
    }
    tf->bsize = 0;
    tf->bused = 0;
    tf->conv = NULL;
    tf->clen = 0;
    tf->cindex = 0;

    if (ftell(tf->file) == 0) {
        tf->start = TRUE;
    }

    return TRUE;
}

long TextFileTell(TextFile *tf)
{
    long pos;

    if (tf == NULL || tf->file == NULL) {
        return -1;
    }
    pos = ftell(tf->file);
    if (pos < 0) {
        return pos;
    }
    pos -= tf->bsize - tf->bused;

    if (tf->cindex < tf->clen) {
        const char *s = tf->conv, *mid = s + tf->cindex, *end = s + tf->clen;

        if (tf->tfe != TFE_NONE) {
            return -1;                   /* still characters in conv */ 
        }
        while (s < mid) {
            s = CodesetNextChar(s);
        }
        while (s < end) {
            pos--;
            s = CodesetNextChar(s);
        }
    }
    return pos;
}

static bool TextFileFillReadBuf(TextFile *tf, long *perr)
{
    long err = OK;
    bool filled = FALSE;

    if (tf == NULL || tf->file == NULL) {
        goto finish;
    }
    while (tf->cindex >= tf->clen) {
        char *s = tf->buffer;
        const char *end;
        size_t warnings = 0;
        
        /* move left-over from last conversion to beginning */
        memmove(s, s + tf->bused, tf->bsize - tf->bused);
        tf->bsize -= tf->bused;
        tf->bused = 0;
        if (sizeof(tf->buffer) - tf->bsize > TEXT_BUFSIZE) {
            tf->bsize += fread(s + tf->bsize, 1, TEXT_BUFSIZE, tf->file);
        }
        s[tf->bsize] = 0;
        if (ferror(tf->file)) {
            err = DISKERR_READ;
            goto finish;
        }
        if (tf->bsize == 0) {                            /* EOF? */
            goto finish;
        }
        if (tf->tfe == TFE_AUTO) {
            if (s[0] == 0) {
                /* not a text file, keep TFE_AUTO */
            } else if (tf->start && tf->bsize >= 3 && !memcmp(s, UTF8_BOM, 3)){
                tf->tfe = TFE_UTF8;                     /* treat as UTF-8 */
            } else if (StrIsASCII(tf->buffer)) {
                /* keep TFE_AUTO */
            } else if (StrIsUTF8AllowTrail(tf->buffer, NULL)) {
                tf->tfe = TFE_UTF8;                     /* most likely */
            } else if (!LocaleIsUTF8()) {
                tf->tfe = TFE_LOCALE;                   /* default */
            } else {
                tf->tfe = TFE_8BIT;                     /* best guess */
            }
        }

        if (tf->conv != tf->buffer) {
            FreeString(tf->memPool, tf->conv);
        }
        tf->clen = 0;
        tf->cindex = 0;
        if (s[0] == 0) {
            tf->conv = s;
            tf->clen = 1;               /* enable to return NUL character */
            tf->bused = 1;
            tf->start = FALSE;
            break;
        }
        switch (tf->tfe) {
            case TFE_UTF8:
                if (tf->start && tf->bsize >= 3 && !memcmp(s, UTF8_BOM, 3)) {
                    s += 3;                                     /* skip BOM */
                }
                if (feof(tf->file)) {
                    tf->conv = CodesetConvertFromUTF8(
                            s, tf->memPool, &err, &warnings);
                    end = s + strlen(s);
                } else {
                    tf->conv = CodesetConvertFromUTF8AllowTrail(
                            s, tf->memPool, &err, &warnings, &end);
                }
                if (tf->conv == s && s != tf->buffer) {
                    tf->conv = AllocString(tf->memPool, tf->conv);
                    if (tf->conv == NULL) {
                        err = ERR_MEM;
                    }
                }
                if (tf->conv) {
                    tf->clen = strlen(tf->conv);
                }
                break;
            case TFE_AUTO:
            case TFE_LOCALE:
                if (feof(tf->file)) {
                    tf->conv = CodesetConvertFromLocale(
                            s, tf->memPool, &err, &warnings);
                    end = s + strlen(s);
                } else {
                    tf->conv = CodesetConvertFromLocaleAllowTrail(
                            s, tf->memPool, &err, &warnings, &end);
                }
                if (tf->conv) {
                    tf->clen = strlen(tf->conv);
                    if (tf->tfe == TFE_AUTO && strcmp(s, tf->conv) != 0) {
                        tf->tfe = TFE_LOCALE;   /* Locale encoding not ASCII */
                    }
                }
                break;
        case TFE_NONE:
        case TFE_8BIT:
        default:
                tf->conv = CodesetConvertFromDefault8Bit(s, tf->memPool, &err);
                end = s + strlen(s);
                if (tf->conv) {
                    tf->clen = strlen(tf->conv);
                }
                break;
        }
        tf->bused = end - tf->buffer;
        if (err) {
            goto finish;
        }
        tf->warnings += warnings;
        tf->start = FALSE;
    }
    filled = TRUE;

finish:
    if (perr) {
        *perr = err;
    }
    return filled;
}

int TextFileGetC(TextFile *tf, long *perr)
{
    if (!TextFileFillReadBuf(tf, perr)) {
        return EOF;
    }
    return (unsigned char)tf->conv[tf->cindex++];
}

bool TextFileGetChar(TextFile *tf, cschar *pc, long *perr)
{
    cschar c;
    bool rc;
    const char *s;

    rc = TextFileFillReadBuf(tf, perr);
    if (!rc) {
        c = 0;
    } else {
        s = &tf->conv[tf->cindex];
        c = CodesetGetChar(s);
        if (c) {
            s = CodesetNextChar(s);
            tf->cindex = s - tf->conv;
        } else {
            tf->cindex++;
        }
    }
    if (pc) {
        *pc = c;
    }
    return rc;
}

char *TextFileReadToEnd(struct MemPool *memPool, TextFile *tf, long *perr)
{
    long err = OK;
    char *str = NULL;
    struct StrHandle *sh = AllocStrHandle(memPool);
    int c;

    if (sh == NULL) {
        err = ERR_MEM;
    } else {
        do {
            c = TextFileGetC(tf, &err);
            if (c == EOF) {
                c = 0;
            }
            StrHandlePutC(sh, c);
        } while (c != 0 && err == OK);
        if (err == OK && StrHandleError(sh)) {
            err = ERR_MEM;
        }
        if (err == OK) {
            str = AllocString(memPool, StrHandleString(sh));
            if (str == NULL) {
                err = ERR_MEM;
            }
        }
    }
    FreeStrHandle(sh);

    if (perr) {
        *perr = err;
    }
    return str;
}

void TextFilePutS(TextFile *tf, const char *in, long *perr)
{
    long err = OK;
    size_t warnings = 0;
    char *text = NULL;

    if (in != NULL && tf != NULL && tf->file != NULL) {
        if (tf->conv != NULL) {
            /* try to seek back to last read position */
            TextFileSeek(tf, 0, SEEK_CUR);

            if (tf->conv != tf->buffer) {
                FreeString(tf->memPool, tf->conv);
            }
            tf->conv = NULL;
        }
        /* kill read buffer */
        tf->cindex = tf->clen;
        tf->bsize = 0;
        tf->bused = 0;

        if (tf->tfe == TFE_AUTO) {
            text = CodesetConvertToLocale(in, tf->memPool, &err, &warnings);
            if (err != OK || warnings > 0) {
                if (text != in) {
                    FreeString(tf->memPool, text);
                }
                tf->tfe = TFE_UTF8;
                warnings = 0;
                text = CodesetConvertToUTF8(in, tf->memPool, &err);
            } else if (!StrIsASCII(in) || strcmp(text, in) != 0) {
                tf->tfe = TFE_LOCALE;
            }
        } else if (tf->tfe == TFE_UTF8) {
            text = CodesetConvertToUTF8(in, tf->memPool, &err);
        } else if (tf->tfe == TFE_LOCALE) {
            text = CodesetConvertToLocale(in, tf->memPool, &err, &warnings);
        } else {        /* TFE_8BIT and TFE_NONE */
            text = CodesetConvertToDefault8Bit(in, tf->memPool, &err,&warnings);
        }
        if (text) {
            fputs(text, tf->file);
            if (text != in) {
                FreeString(tf->memPool, text);
            }
            if (ferror(tf->file)) {
                err = DISKERR_WRITE;
            }
            tf->start = FALSE;
        }
        tf->warnings += warnings;
    }

    if (perr) {
        *perr = err;
    }
}

void TextFilePutChar(TextFile *tf, cschar c, long *perr)
{
    char buf[CodesetMaxBytesForChar + 1];
    short count;
    long err = OK;

    count = CodesetCharToStr(c, buf);
    if (count >= 0) {
        buf[count] = 0;
        if (buf[0] != 0) {
            TextFilePutS(tf, buf, &err);
        } else if (tf && tf->file) {
            fputc(0, tf->file);
        }
    } else {
        err = CODESETERR_CONVERSION;
    }

    if (perr) {
        *perr = err;
    }
}

char *TextFileGetContents(struct MemPool *memPool, const char *filename,
        long *perr, size_t *pwarnings)
{
    long err;
    size_t warnings = 0;
    char *str = NULL;

    TextFile *tf = TextFileOpen(memPool, filename, "r", TFE_AUTO);
    if (tf == NULL) {
        err = DISKERR_OPEN;
    } else {
        str = TextFileReadToEnd(memPool, tf, &err);
        warnings = TextFileWarnings(tf);
        TextFileClose(tf);
    }

    if (perr) {
        *perr = err;
    }
    if (pwarnings) {
        *pwarnings = warnings;
    }
    return str;
}

char *TextFileGetContentsAskWarnings(GuiObject go, struct MemPool *memPool, 
        const char *filename, long *perr)
{
    long err;
    size_t warnings;
    char *str = TextFileGetContents(memPool, filename, &err, &warnings);

    if (err == OK && warnings &&
            !Ask2(go, NULL, AT_WARN, GCS(MSG_Warn_CodesetProcessFileAnswers), 
                GCS(MSG_Warn_CodesetProcessFile), warnings, filename)) {
        err = ERR_CANCEL;
        FreeString(memPool, str);
        str = NULL;
    }

    if (perr) {
        *perr = err;
    }
    return str;
}

void TextFileSetContents(struct MemPool *memPool, const char *filename,
        const char *in, long *perr, size_t *pwarnings)
{
    long err, err2;
    size_t warnings = 0;

    TextFile *tf = TextFileOpen(memPool, filename, "w", TFE_AUTO);
    if (tf == NULL) {
        err = DISKERR_OPEN;
    } else {
        TextFilePutS(tf, in, &err);
        warnings = TextFileWarnings(tf);
        err2 = TextFileClose(tf);
        if (err == OK) {
            err = err2;
        }
    }

    if (perr) {
        *perr = err;
    }
    if (pwarnings) {
        *pwarnings = warnings;
    }
}

void TextFileSetContentsAskWarnings(GuiObject go, struct MemPool *memPool,
        const char *filename, const char *in, long *perr)
{
    long err;
    size_t warnings;

    TextFileSetContents(memPool, filename, in, &err, &warnings);
    if (err == OK && warnings > 0) {
        Ask2(go, NULL, AT_WARN, GCS(MSG_StarOk), 
                GCS(MSG_Warn_CodesetProcessFile), warnings, filename);
    }

    if (perr) {
        *perr = err;
    }
}

