/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** textfile.h
**
** Reading and writing of text files with conversiom and encoding detection.
**
** Creation date: 31.01.2009
** Last modified: 01.02.2009
*/

#ifndef TEXTFILE_H
#define TEXTFILE_H

#include "system.h"
#include "codeset.h"

#ifdef __cplusplus
extern "C" {
#endif

struct _TextFile;
typedef struct _TextFile TextFile;

enum {
    TFE_NONE,           /* binary I/O */
    TFE_AUTO,           /* auto-detect encoding */
    TFE_UTF8,           /* UTF-8 */
    TFE_LOCALE,         /* LANG (Unix), Codepage (Windows), 8bit (Amiga) */
    TFE_8BIT            /* ISO-8859-1 (Linux, Windows), 8bit (Amiga) */
};

long TextFileClose(TextFile *tf);
/*
** Returns OK or error code.
*/

TextFile *TextFileOpen(struct MemPool *memPool, const char *filename,
        const char *mode, int tfe);
/*
** Opens the given filename with the given mode (uses IFOpen).
** In tfe the text file encoding is specified.
** If tfe = TFE_NONE then no conversion is done (binary file).
** If tfe is TFE_AUTO then the encoding it auto-detected:
** If the file already contains text then the following rules apply:
** If the first 3 bytes are 0xEF 0xBB 0xBF (BOM) then tfe = TFE_UTF8.
** Else if the text is valid UTF-8, tfe = TFE_UTF8.
** Else if the system locale encoding is not UTF-8 then tfe = TFE_LOCALE.
** Else tfe = TFE_8BIT (ISO-8859-1 on Windows/Linux).
** If tfe = TFE_UTF8 then any invalid sequence read is replaced by 
** a question mark (either ? or 0xEF 0xBF 0xBD (replacement char)).
** When writing text to the file and tfe is TFE_AUTO:
** If conversion to locale encoding fails then tfe = TFE_UTF8
** else if converted string is not 7bit ASCII then tfe = UTF_LOCALE
** else tfe stays TFE_AUTO and decision is delayed.
*/

size_t TextFileWarnings(TextFile *tf);
/*
** Returns total number of codeset conversion warnings during processing.
*/

long TextFileFlush(TextFile *tf);
/*
** Flushes all output.  Returns error code or OK.
*/

bool TextFileError(TextFile *tf);
/*
** Returns error state of the file handle underlying the text file.
*/

bool TextFileEOF(TextFile *tf);
/*
** Returns end-of-file status of text file.
*/

bool TextFileSeek(TextFile *tf, long offset, int whence);
/*
** Seeks to the given position.  Returns TRUE on success, FALSE on failure.
*/

long TextFileTell(TextFile *tf);
/*
** Returns position in text file or -1 in case of an error.
*/

void TextFilePutS(TextFile *tf, const char *text, long *err);
/*
** Outputs the given text into the text file.
** Returns OK or error code in *err.
*/

void TextFilePutChar(TextFile *tf, cschar c, long *err);
/*
** Outputs the given codeset character into the text file.
** If encoding is TFE_NONE then the the 8-bit value of c is written.
** Furthermore, if c is NUL, a Null-byte is written (regardless of encoding).
** Returns OK or error code in *err.
*/

int TextFileGetC(TextFile *tf, long *err);
/*
** Reads the next byte from the text file.
** If an error occured then *err is set to the error code, otherwise OK.
** Returns next byte or EOF when at the end of file or an error occurred.
*/

bool TextFileGetChar(TextFile *tf, cschar *c, long *err);
/*
** Reads the next character from the text file.
** If encoding is TFE_NONE then the next byte is read and placed into *c.
** Otherwise *c holds the next codeset character from the file.
** If an error occured then *err is set to the error code, otherwise OK.
** Returns FALSE on end-of-file or if an error occured, TRUE otherwise.
*/

char *TextFileReadToEnd(struct MemPool *memPool, TextFile *tf, long *err);
/*
** Reads until the end of file or next NUL byte.
** If an error occured then *err is set to the error code, otherwise OK.
** Returns read string or NULL in case an error occurred.
*/

char *TextFileGetContents(struct MemPool *memPool, const char *filename,
        long *err, size_t *warnings);
/*
** Reads the contents of filename and returns it as one string.
** If an error occurred returns NULL and sets *err.
** In *warnings the number of codeset conversion errors is set.
*/

char *TextFileGetContentsAskWarnings(GuiObject go, struct MemPool *memPool, 
        const char *filename, long *err);
/*
** Like TextFileGetContents but asks to continue when there are
** conversion errors.  Sets *err to ERR_CANCEL when the user cancels.
** The dialog is centered on the given GuiObject (can be NULL).
*/

void TextFileSetContents(struct MemPool *memPool, const char *filename,
        const char *text, long *err, size_t *warnings);
/*
** Sets the contents of filename to the given text.
** If an error occurred, *err is set.
** In *warnings the number of codeset conversion errors is set.
*/

void TextFileSetContentsAskWarnings(GuiObject go, struct MemPool *memPool,
        const char *filename, const char *in, long *err);
/*
** Like TextFileSetContents but pops up a info message when
** there are conversion warnings (the error code stays OK).
** The dialog is centered on the given GuiObject (can be NULL).
*/

#ifdef __cplusplus
}
#endif

#endif

