/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** util.c
**
** Creation date: 24.11.1994
** Last modified: 02.05.2022
*/

#include "system.h"
#include "util.h"
#include "memory.h"
#include <limits.h>
#include <math.h>
#include <float.h>
#include <time.h>
#include <locale.h>
#ifndef AMIGA
#include <stdint.h>
#endif

#define BUF_LEN 500
static char buffer[BUF_LEN + 1];

long NextVersion(long version)
{
    long next = version + 1;
    if (next == 0) {
        next = 1;
    }
    return next;
}

char *WordString(struct MemPool *memPool, const char *s, long n, bool *err)
{
    const char *t;
    char *p;
    long i;

    if (err != NULL) {
        *err = FALSE;
    }
    if (n < 0 || s == NULL) {
        return NULL;
    }

    while (CodesetIsBrkSpace(CodesetGetChar(s))) {
	s = CodesetNextChar(s);
    }
    for (i=0; i < n; i++) {
	while (!CodesetIsBrkSpace(CodesetGetChar(s))) {
	    if (*s == 0) {
		return NULL;		/* n out of range */
	    }
	    s = CodesetNextChar(s);
	}
	while (CodesetIsBrkSpace(CodesetGetChar(s))) {
	    s = CodesetNextChar(s);
	}
    }
    if (*s == 0) {
        return NULL;                	/* no word at the end */
    }

    for (t = s; !CodesetIsBrkSpace(CodesetGetChar(t)); t = CodesetNextChar(t)) {
        if (*t == 0) {
            break;
	}
    }
    p = AllocStringN(memPool, s, (long)(t - s));
    if (p == NULL && err != NULL) {
	*err = TRUE;
    }
    return p;
}

long NumWords(const char *s)
{
    long num = 0;

    if (s == NULL) {
        return 0;
    }

    while (CodesetIsBrkSpace(CodesetGetChar(s))) {
	s = CodesetNextChar(s);
    }
    while (*s) {
        num++;
	while (*s && !CodesetIsBrkSpace(CodesetGetChar(s))) {
	    s = CodesetNextChar(s);
	}
	while (CodesetIsBrkSpace(CodesetGetChar(s))) {
	    s = CodesetNextChar(s);
	}
    }
    return num;
}

static bool isFieldSeparator(cschar c, const char *sep)
{
    bool rc;

    if (c == 0) {
        rc = FALSE;
    } else if (sep) {
        rc = CodesetStrchr(sep, c) != NULL;
    } else {
        rc = CodesetIsBrkSpace(c);
    }
    return rc;
}

static const char *findFieldEnd(const char *s, const char *sep, bool quotes)
{
    cschar cs, lastcs = 0;
    bool q;
   
    q = quotes && *s == '"';
    if (q) {
        s++;
    }
    for (;;) {
        cs = CodesetGetChar(s);
        if (cs == 0) {
            break;
        }
        if (isFieldSeparator(cs, sep)) {
            if (!q || lastcs == '"') {
                break;
            }
        }
        lastcs = cs;
        s = CodesetNextChar(s);
    }
    return s;
}

char *FieldString(struct MemPool *memPool, const char *s, long n,
        const char *sep, bool quotes, bool *err)
{
    const char *t;
    char *p;
    long i;

    if (err != NULL) {
        *err = FALSE;
    }
    if (n < 0 || s == NULL) {
        return NULL;
    }

    for (i = 0; i < n; i++) {
        s = findFieldEnd(s, sep, quotes);
        if (*s == 0) {
            return NULL;		/* n out of range */
        }
        /* Skip one field separator */
        s = CodesetNextChar(s);
    }

    t = findFieldEnd(s, sep, quotes);
    if (quotes) {
        if (*s == '"') {
            s++;
            if (t > s && t[-1] == '"') {
                t--;
            }
        }
    }

    p = AllocStringN(memPool, s, (long)(t - s));
    if (p == NULL && err != NULL) {
	*err = TRUE;
    }
    return p;
}

long StrNumFields(const char *s, const char *sep, bool quotes)
{
    long num = 0;

    if (s == NULL) {
        return 0;
    }

    for (;;) {
        num++;
        s = findFieldEnd(s, sep, quotes);
        if (*s == 0) {
            break;
        }
        s = CodesetNextChar(s);
    }

    return num;
}

long StrCount(const char *s, cschar c)
{
    long count = 0;

    if (s) {
	while ((s = CodesetStrchr(s, c)) != NULL) {
	    count++;
	    s = CodesetNextChar(s);
	}
    }
    return count;
}

bool StrIsASCII(const char *s)
{
    if (s) {
	for (; *s; s++) {
	    if ((*s & 0x80) != 0) {
		return FALSE;
	    }
	}
    }
    return TRUE;
}

static short StrCompressSequence(const char **pin, char *buf)
{
    const char *in = *pin;
    short v, c, len;

    if (*in == 0) {
        *buf = 0;
        return 0;
    } else if (*in != '\\') {
        *buf = *in;
        *pin = in + 1;
        return 1;
    }

    c = -1;
    /* check for octal number "\NNN" */
    v = CodesetDigitValue(CodesetGetChar(in + 1));
    if (v >= 0 && v <= 7) {
        c = v;
        in = CodesetNextChar(in + 1);
        v = CodesetDigitValue(CodesetGetChar(in));
        if (v >= 0 && v <= 7) {
            c = c * 8 + v;
            in = CodesetNextChar(in);
            v = CodesetDigitValue(CodesetGetChar(in));
            if (v >= 0 && v <= 7 && c < 256 / 8) {
                c = c * 8 + v;
                in = CodesetNextChar(in);
            }
        }
    } else if (in[1] == 'x' &&          /* check for hex number "\xNN" */
            (v = CodesetXDigitValue(CodesetGetChar(in + 2))) >= 0) {
        c = v;
        in = CodesetNextChar(in + 2);
        v = CodesetXDigitValue(CodesetGetChar(in));
        if (v >= 0) {
            c = c * 16 + v;
            in = CodesetNextChar(in);
        }
    }
    if (c >= 0) {               /* have octal or hex number? */
        *pin = in;
        if (c == 0) {
            return 0;           /* NUL byte */
        }
        len = CodesetCharToStr((cschar)c, buf);
        if (len < 0) {
            return 0;           /* ignore invalid char */
        }
        return len;
    }

    switch (in[1]) {
        case 'n':       *buf = '\n';    break;
        case 't':       *buf = '\t';    break;
        case 'v':       *buf = '\v';    break;
        case 'b':       *buf = '\b';    break;
        case 'r':       *buf = '\r';    break;
        case 'f':       *buf = '\f';    break;
        case '\\':      *buf = '\\';    break;
        case '\"':      *buf = '\"';    break;
        case 'e':       *buf = '\33';   break;

        default:        *buf = '\\';    /* unknown sequence, keep orifinal */
                        *pin = in + 1;
                        return 1;
    }
    *pin = in + 2;      /* skip 2 bytes escape sequence */
    return 1;
}

char *StrCompress(struct MemPool *memPool, const char *in)
{
    char *string = NULL;

    if (in != NULL) {
        char buf[CodesetMaxBytesForChar];
        const char *s = in;
        size_t size = 0;

        while (*s) {
            size += StrCompressSequence(&s, buf);
        }
        size++;                             /* NUL terminate */

        string = PoolAllocMem(memPool, size);
        if (string != NULL) {
            char *t = string;

            s = in;
            while (*s) {
                t += StrCompressSequence(&s, t);
            }
            *t = 0;                         /* NUL terminate */
        }
    }
    return string;
}

char *BuildCommandLine(struct MemPool *memPool, 
	const char *pattern, const char *arg)
{
    char *buf = NULL;
    const char *s; 
    char *t;
    long len = 0;
#ifdef AMIGA
    const char *pubscreen = NULL;
#endif

    if (pattern == NULL) {
        return NULL;
    }

    for (s = pattern; *s; s++) {
        if (s[0] == '%' && (s[1] == 'f' || s[1] == 'F')) {
	    if (arg) {
		len += strlen(arg) + 2;
	    }
            s++;
        }
#ifdef AMIGA
        else if (s[0] == '%' && (s[1] == 'p' || s[1] == 'P')) {
            if (pubscreen == NULL) {
                pubscreen = GuiPubScreenName();
            }
            if (pubscreen) {
                len += strlen(pubscreen);
            }
            len += 2;
            s++;
        }
#endif
        else {
            len++;
	}
    }

    buf = PoolAllocMem(memPool, len + 1);
    if (buf != NULL) {
        for (s = pattern, t = buf; *s; s++) {
            if (s[0] == '%' && (s[1] == 'f' || s[1] == 'F')) {
		if (arg) {
		    *t++ = '"';
		    strcpy(t, arg);
		    t += strlen(t);
		    *t++ = '"';
		}
                s++;
            }
#ifdef AMIGA
            else if (s[0] == '%' && (s[1] == 'p' || s[1] == 'P')) {
                *t++ = '"';
                if (pubscreen) {
                    strcpy(t, pubscreen);
		    t += strlen(t);
                }
                *t++ = '"';
                s++;
            }
#endif
            else {
                *t++ = *s;
	    }
        }
        *t = 0;
    }
    return buf;
}

cschar Shortcut(const char *s)
{
    if (s) {
        for(; *s; s++) {
            if (*s == '_') {
		s++;
		return CodesetToLower(CodesetGetChar(s));
            }
	}
    }
    return 0;
}

const char *EmptyNullStr(const char *s)
{
    if (s == NULL) {
        s = "";
    }
    return s;
}

bool StrJokerCmp(const char *s, const char *j)
{
    cschar cs, cj;

    do {
	cs = CodesetGetChar(s);
	cj = CodesetGetChar(j);
        if (cj == JOKER) {
	    do {
		j = CodesetNextChar(j);
		cj = CodesetGetChar(j);
	    } while (cj == JOKER);
            if (cj == 0) {
                return TRUE;
	    }
	    while (*s) {
                if (StrJokerCmp(s, j)) {
                    return TRUE;
		}
		s = CodesetNextChar(s);
	    }
            return FALSE;
        } else if (cj != SINGLEJOKER) {
            if (CodesetToUpper(cs) != CodesetToUpper(cj)) {
                return FALSE;
	    }
	}
	s = CodesetNextChar(s);
	j = CodesetNextChar(j);
    } while (cs);

    return (bool)(cj == 0);
}

bool strjokercmp(const char *s, const char *j)
{
    cschar cs, cj;

    do {
	cs = CodesetGetChar(s);
	cj = CodesetGetChar(j);
        if (cj == JOKER) {
	    do {
		j = CodesetNextChar(j);
		cj = CodesetGetChar(j);
	    } while (cj == JOKER);
            if (cj == 0) {
                return TRUE;
	    }
	    while (*s) {
                if (strjokercmp(s, j)) {
                    return TRUE;
		}
		s = CodesetNextChar(s);
	    }
            return FALSE;
        } else if (cj != SINGLEJOKER) {
            if (cs != cj) {
                return FALSE;
	    }
	}
	s = CodesetNextChar(s);
	j = CodesetNextChar(j);
    } while (cs);

    return (bool)(cj == 0);
}

ulong StrToId(const char *s)
{
    ulong id = 0;
    int i;

    if (s) {
	for (i=0; i < 4 && *s; i++) {
	    id = (id << 8) | (unsigned char)*s++;
	}
    }
    return id;
}

const char *IdToStr(ulong id)
{
    static char buf[5];

    memset(buf, 0, sizeof(buf));
    while (id) {
	buf[3] = buf[2];
	buf[2] = buf[1];
	buf[1] = buf[0];
	buf[0] = (char)(id & 0xff);
	id >>= 8;
    }
    return buf;
}

static bool StrToIntBase(const char *s, long *result, int base)
{
    char *np;
    long value;

    if (s == NULL) {
        return FALSE;
    }
    while (CodesetIsSpace(CodesetGetChar(s))) {
	s = CodesetNextChar(s);
    }
    if (*s == 0) {
	return FALSE;		/* empty string */
    }
    value = strtol(s, &np, base);
    while (CodesetIsSpace(CodesetGetChar(np))) {
	np = CodesetNextChar(np);
    }
    if (*np != 0) {
        return FALSE;          /* trailing garbage */
    }

    if (result) {
        *result = value;
    }
    return TRUE;
}

bool StrToInt(const char *string, long *result)
{
    return StrToIntBase(string, result, 0);
}

char *IntToStr(long value)
{
    static char buffer[32];

    sprintf(buffer, "%ld", value);
    return buffer;
}

bool StrToDouble(const char *s, double *result)
{
    const char *dp; 
    char *endptr;
    double value;

    if (s == NULL) {
        return FALSE;
    }

    while (CodesetIsSpace(CodesetGetChar(s))) {
	s = CodesetNextChar(s);
    }
    if (*s == 0) {
        return FALSE;          	/* string is empty */
    }
    dp = strpbrk(s, ".,");
    if (dp != NULL) {
        char dpchar = '.';

        const struct lconv *lc = localeconv();
        if (lc && lc->decimal_point && lc->decimal_point[0]) {
            dpchar = lc->decimal_point[0];
        }

        if (*dp != dpchar) {
            if (strlen(s) >= BUF_LEN) {
                return FALSE;           /* reject overlong string */
            }
            strcpy(buffer, s);
            buffer[dp - s] = dpchar;
            s = buffer;
        }
    }

    value = strtod(s, &endptr);
    while (CodesetIsSpace(CodesetGetChar(endptr))) {
	endptr = CodesetNextChar(endptr);
    }
    if (*endptr != 0) {
        return FALSE;          /* trailing garbage */
    }

    if (result) {
	*result = value;
    }
    return TRUE;
}

char *DoubleToStr(double value, short numdecimals, short format)
{
    char efg;

    if (numdecimals < 0) {
	efg = 'g';
    } else if (fabs(value) >= 1e10) {
	efg = 'e';
    } else {
	efg = 'f';
    }
    return DoubleToStr2(value, numdecimals, efg, format);
}

char *DoubleToStr2(double value, short numdecimals, char efg, short format)
{
    char *s, dp, rpl;

    /* suppress Nan or Inf output */
    if (value > DBL_MAX) {
        value = DBL_MAX;
    } else if(value < -DBL_MAX) {
        value = -DBL_MAX;
    }

    switch (efg) {
	case 'e':
	case 'E':
	    if (numdecimals < 0) {
		numdecimals = 2;
	    }
	    sprintf(buffer, "%.*e", numdecimals, value);
	    break;

	case 'f':
	case 'F':
	    if (numdecimals < 0) {
		numdecimals = 2;
	    }
	    sprintf(buffer, "%.*f", numdecimals, value);
	    break;

	case 'g':
	case 'G':
	default:
            /* for %g, numdecimals is the number of significant digits */
            if (numdecimals <= 0) {
                numdecimals = 15;
            }
	    sprintf(buffer, "%.*g", numdecimals, value);
	    if (strpbrk(buffer, ".,eE") == NULL &&
		    strlen(buffer) - ((buffer[0]=='-')? 1 : 0) < numdecimals) {
		strcat(buffer, ".0");
	    }
	    break;
    }

    if (format == 0) {
	dp = '.';
	rpl = ',';
    } else {
	dp = ',';
	rpl = '.';
    }
    if ((s = CodesetStrchr(buffer, rpl)) != NULL) {
        *s = dp;
    }
    return buffer;
}

bool DoubleToInt(double d, long *l)
{
    d = round(d);

    if (d > (double)LONG_MAX) {
        return FALSE;
    } else if (d < (double)LONG_MIN) {
        return FALSE;
    } else if (l) {
        *l = (long)d;
    }
    return TRUE;
}

/* 
** A leap year can be devided by 4, but must not be devided by 100,
** except if it can also be devided by 400.
*/

#define LEAP_YEAR(y) (!((y)&3) && ((y)%100 || !((y)%400)))
#define DAYS_MIN 0  	  /* corresponds to <01.01.0000> */
#define DAYS_MAX 3652424  /* corresponds to <31.12.9999> */
#define SECS_MIN 0
#define SECS_MAX 0x7fffffff

static short thisyear, thismonth, thisday;

static long YearToDate(long year)
{
    long year4 = year? (1 + (year-1)/4) : 0;
    long year100 = year / 100;
    long year400 = year / 400;

    return (year * 365 + year4 - year100 + year400);
}

short MonthDays(short month, short year)
{
    short days;

    switch (month)
    {
	case 2: days = LEAP_YEAR(year)? 29 : 28;
		break;

	case 4:
	case 6:
	case 9:
	case 11: days = 30;
		 break;

	default: days = 31;
		 break;
    }

    return days;
}

short YearDays(short year)
{
    return (short)(LEAP_YEAR(year)? 366 : 365);
}

void DateToDayMonthYear(long date, short *pday, short *pmonth, short *pyear)
{
    short year;

    if (date < DAYS_MIN) {
        date = DAYS_MIN;
    } else if (date > DAYS_MAX) {
        date = DAYS_MAX;
    }
    year = (short)(date * 400 / (400*365 + 100 - 4 + 1));

    date -= YearToDate(year);

    while (date) {
	short days = YearDays(year);

	if (days > (short)date) {
	    break;
	}
	date -= days;
	year++;
    }
    if (pyear) {
	*pyear = year;
    }
    if (pmonth || pday) {
	short month;
	
    	date++;
	for (month=1; month<=12; month++) {
	    short days = MonthDays(month, year);

	    if (days >= (short)date) {
		break;
	    }
	    date -= days;
	}
	if (pmonth) {
	    *pmonth = month;
	}
	if (pday) {
	    *pday = (short)date;
	}
    }
}

bool DayMonthYearToDate(long day, long month, long year, long *result)
{
    if (!YEAR_VALID(year) || !MONTH_VALID(month) || day < 1 ||
	    day > MonthDays((short)month, (short)year)) {
	return FALSE;
    }

    if (result) {
	/* calculate date */
	long date = day-1;

	for (--month; month > 0; month--) {
	    date += MonthDays((short)month, (short)year);
	}
	date += YearToDate((short)year);

	*result = date;
    }
    return TRUE;
}

long DateToday(void)
{
    time_t timeval = time(NULL);
    struct tm *ut = localtime(&timeval);
    long date = 0;

    if (ut != NULL) {
        date = YearToDate(ut->tm_year + 1900) + ut->tm_yday;
    }

   DateToDayMonthYear(date, &thisday, &thismonth, &thisyear);

   return date;
}

bool DateAdd(long date, long days, long *result)
{
    long v = date + days;

    if (date >= 0 && days > 0 && v <= 0) {
	return FALSE;				/* overflow */
    }
    if (date <= 0 && days < 0 && v >= 0) {
	return FALSE;				/* underflow */
    }
    if (v < DAYS_MIN || v > DAYS_MAX) {
	return FALSE;
    }
    if (result) {
	*result = v;
    }
    return TRUE;
}

bool DateAddMonth(long date, long months, long *result)
{
    short day, month, year;
    long newMonth, newYear, years;
    short maxDay;

    DateToDayMonthYear(date, &day, &month, &year);
    newMonth = (long)month + months;
    years = newMonth / 12;
    newMonth -= years * 12;
    newYear = (long)year + years;
    if (newMonth < 1) {
	newMonth += 12;
	newYear--;
    } 

    maxDay = MonthDays((short)newMonth, (short)newYear);
    day = MIN(day, maxDay);

    return DayMonthYearToDate(day, newMonth, newYear, result);
}

bool DateAddYear(long date, long years, long *result)
{
    return DateAddMonth(date, years * 12, result);
}

bool StringToDate(const char *string, long *result)
{
    char *daystring="", *monthstring="", *yearstring="", *s;
    long day, month, year = 0;          /* avoid uninitialzed year warning */
    long today = DateToday();           /* sets this{day,month,year} */
    short len, sub;

    /* copy string */
    sprintf(buffer, "%.*s", BUF_LEN, string);

    /* test for empty string */
    CodesetTrimStr(buffer);
    if (buffer[0] == 0) {
	return FALSE;		/* string is empty */
    }

    /*
    ** test for relative date
    */
    if (StrToIntBase(buffer, &day, 10)) {
	return DateAdd(today, day, result);
    }

    s = CodesetStrchr(buffer, '.');
    if (s != NULL) {		/* format dd.mm.yyyy */
    	*s++ = 0;
      	daystring = buffer;
      	monthstring = s;
      	s = CodesetStrchr(s, '.');
      	if (s != NULL) {
            *s++ = 0;
            yearstring = s;
      	}
    } else if ((s = CodesetStrchr(buffer, '/')) != NULL) { /*format mm/dd/yyyy*/
      	*s++ = 0;
      	monthstring = buffer;
      	daystring = s;
      	s = CodesetStrchr(s, '/');
      	if (s != NULL) {
            *s++ = 0;
            yearstring = s;
   	}
    } else if ((s = CodesetStrchr(buffer, '-')) != NULL) { /*format yyyy-mm-dd*/
      	*s++ = 0;
      	monthstring = buffer;
      	daystring = s;
      	s = CodesetStrchr(s, '-');
      	if (s != NULL) {
            *s++ = 0;
	    yearstring = monthstring;
	    monthstring = daystring;
            daystring = s;
        }
    } else {
	return FALSE;
    }

    /* check daystring */
    CodesetTrimStr(daystring);
    if (daystring[0] == 0) {
	day = thisday;		/* use this day */
    } else if (!StrToIntBase(daystring, &day, 10)) {
      	return FALSE;
    }

    /* check monthstring */
    CodesetTrimStr(monthstring);
    if (monthstring[0] == 0) {
	month = thismonth;		/* use this month */
    } else if (!StrToIntBase(monthstring, &month, 10)) {
      	return FALSE;
    }

    /* check yearstring */
    CodesetTrimStr(yearstring);
    len = strlen(yearstring);
    if (yearstring[0] == 0) {
	year = thisyear;		/* use this year */
    } else if (!StrToIntBase(yearstring, &year, 10)) {
      	return FALSE;
    } else if (year < 0) {
        return FALSE;
    } else if (len <= 2) {
        sub = (year > (thisyear + 50) % 100)? 1 : 0;
        year += 100 * (thisyear / 100 - sub);
    } else if (len == 3) {
        sub = (year > (thisyear + 500) % 1000)? 1 : 0;
        year += 1000 * (thisyear / 1000 - sub);
    }

    return DayMonthYearToDate(day, month, year, result);
}

char *DateToString(long date, short format)
{
    short day, month, year;
    static char datestr[3*(1+5+1) + 1];

    DateToDayMonthYear(date, &day, &month, &year);
    switch (format) {
        case FORMAT_DAYMONTHYEAR:
	    sprintf(datestr, "%02d.%02d.%04d", day, month, year);
            break;

        case FORMAT_MONTHDAYYEAR:
	    sprintf(datestr, "%02d/%02d/%04d", month, day, year);
            break;

        case FORMAT_YEARMONTHDAY:
        default:
            sprintf(datestr, "%04d-%02d-%02d", year, month, day);
            break;
    }
    return datestr;
}

long TimeNow(void)
{
    time_t timeval = time(NULL);
    struct tm *ut = localtime(&timeval);
    long t = 0;

    if (ut != NULL) {
        t = (ut->tm_hour * 60 + ut->tm_min) * 60 + ut->tm_sec;
    }

    return t;
}

bool TimeAdd(long time, long secs, long *result)
{
    long v = time + secs;

    if (time >= 0 && secs > 0 && v <= 0) {
	return FALSE;				/* overflow */
    }
    if (time <= 0 && secs < 0 && v >= 0) {
	return FALSE;				/* underflow */
    }
    if (v < SECS_MIN || v > SECS_MAX) {
	return FALSE;
    }
    if (result) {
	*result = v;
    }
    return TRUE;
}

bool StringToTime(const char *string, short format, long *result)
{
    char *parts[3], *s;
    long v[3];
    short i;
    long time;

    /* copy string */
    sprintf(buffer, "%.*s", BUF_LEN, string);

    /* test for empty string */
    CodesetTrimStr(buffer);
    if (buffer[0] == 0) {
	return FALSE;		/* string is empty */
    }

    parts[0] = parts[1] = parts[2] = NULL;
    if ((s = strpbrk(buffer, "hH")) != NULL) {	     /* explicit hour format */
	*s++ = 0;
	parts[0] = buffer;
	parts[1] = s;
	if ((s = strpbrk(s, "mM")) != NULL) {
	    *s++ = 0;
	    parts[2] = s;
	}
    } else if ((s = strpbrk(buffer, "mM")) != NULL) { /* explicit min format */
	*s++ = 0;
	parts[1] = buffer;
	parts[2] = s;
    } else if ((s = strpbrk(buffer, "sS")) != NULL) { /* explicit sec format */
	parts[2] = buffer;
    } else if ((s = strpbrk(buffer, ":")) != NULL) {  /* colon format */
	*s++ = 0;
	parts[0] = buffer;
	parts[1] = s;
	if ((s = strpbrk(s, ":")) != NULL) {
	    *s++ = 0;
	    parts[2] = s;
	} else if (format == FORMAT_MINSEC) {
	    parts[2] = parts[1];
	    parts[1] = parts[0];
	    parts[0] = NULL;
	}
    } else if (format == FORMAT_HOURMIN) {		/* minutes only */
	parts[1] = buffer;
    } else {
	parts[2] = buffer;				/* seconds only */
    }

    /* check termination of secs */
    if (parts[2]) {
	s = CodesetStrchr(parts[2], 's');
	if (s) {
	    *s++ = 0;
	    if (*s != 0) {
		return FALSE;				/* trailing chars */
	    }
	}
    }

    /* convert to numbers */
    v[0] = v[1] = v[2] = 0;
    for (i=0; i < 3; i++) {
	if (parts[i] && parts[i][0] && !StrToIntBase(parts[i], &v[i], 10)) {
	    return FALSE;
	}
    }

    /* check ranges */
    for (i=1; i < 3; i++) {
	if (parts[i-1] != NULL && (v[i] < 0 || v[i] > 60)) {
	    return FALSE;
	}
    }

    /* check overflow */
    time = v[0] * 60;
    if ((time / 60) != v[0]) {
	return FALSE;
    }
    v[1] += time;
    time = v[1] * 60;
    if ((time / 60) != v[1]) {
	return FALSE;
    }
    time += v[2];
    if (time < SECS_MIN || time > SECS_MAX) {
	return FALSE;
    }

    if (result) {
	*result = time;
    }
    return TRUE;
}

char *TimeToString(long time, short format)
{
    static char timestr[1+16+1 + 2*(1+2+1) + 1];
    long hour, minute, second;

    hour = time / (60*60);
    time %= 60*60;
    minute = time / 60;
    second = time % 60;

    switch (format) {
	case FORMAT_HOURMINSEC:
	default:
	    sprintf(timestr, "%02ld:%02ld:%02ld", hour, minute, second);
	    break;
	case FORMAT_MINSEC:
	    sprintf(timestr, "%02ldm%02ld", hour * 60 + minute, second);
	    break;
	case FORMAT_HOURMIN:
	    sprintf(timestr, "%02ldh%02ld", hour, minute);
	    break;
    }

    return timestr;
}

void SwapLists(struct List *list1, struct List *list2)
{
    if (list1 != NULL && list2 != NULL) {
        struct Node *head1 = HEAD(list1);
        struct Node *tail1 = TAILPRED(list1);
        struct Node *head2 = HEAD(list2);
        struct Node *tail2 = TAILPRED(list2);

        if (SUCC(head2) != NULL) {
            list1->lh_Head = head2;
            list1->lh_TailPred = tail2;
            head2->ln_Pred = (struct Node *)&list1->lh_Head;
            tail2->ln_Succ = (struct Node *)&list1->lh_Tail;
        } else {
            NewList(list1);
        }
        if (SUCC(head1) != NULL) {
            list2->lh_Head = head1;
            list2->lh_TailPred = tail1;
            head1->ln_Pred = (struct Node *)&list2->lh_Head;
            tail1->ln_Succ = (struct Node *)&list2->lh_Tail;
        } else {
            NewList(list2);
        }
    }
}

const struct Node *CorruptedNode = NULL;

long CheckList(const struct List *list)
{
    long count = 0;

    if (list) {
    	struct Node *pred = (struct Node *)&list->lh_Head;
	struct Node *node = SUCC(pred);

	while (node) {
	    if (PRED(node) != pred) {
		count++;
	    }
	    pred = node;
	    node = SUCC(node);
	}
	if (pred != (struct Node *)&list->lh_Tail) {
	    count++;
	    node = TAILPRED(list);
	    while (node) {
		if (node == pred) {
		    break;
		}
		count++;
		node = PRED(node);
	    }
	    CorruptedNode = pred;
	}
    }

    return count;
}

static void QLSortNodesRec(struct Node *left, struct Node *right, 
	struct Node *pivot, int (*cmpfunc)(void *, void *))
{
    for (;;) {			/* for tail recursion of longer sub list */
	struct Node *n1 = SUCC(left), *n2;
	long i, lnum = 0, rnum = 0, cmp;
	struct Node *help = pivot;

	while ((n2 = n1) != pivot) {
	    n1 = SUCC(n1);
            cmp = cmpfunc(n2, pivot);
	    if (cmp > 0 || (cmp == 0 && rnum < lnum)) {
		REMOVE(n2);
		INSERT(NULL, n2, help);
		help = n2;
		rnum++;
	    } else {
		lnum++;
	    }
	}
	n1 = SUCC(help);
	help = PRED(pivot);
	while ((n2 = n1) != right) {
	    n1 = SUCC(n1);
            cmp = cmpfunc(n2, pivot);
	    if (cmp < 0 || (cmp == 0 && lnum < rnum)) {
		REMOVE(n2);
		INSERT(NULL, n2, help);
		lnum++;
	    } else {
		rnum++;
	    }
	}
	if (lnum > rnum) {
	    if (rnum > 1) {
		for (i = rnum/2, n2 = SUCC(pivot); i; i--) {
		    n2 = SUCC(n2);
		}
		QLSortNodesRec(pivot, right, n2, cmpfunc);
	    } else if (lnum <= 1) {
		break;
	    }
	    for (i = lnum/2, n1 = SUCC(left); i; i--) {
		n1 = SUCC(n1);
	    }
	    right = pivot;
	    pivot = n1;
	} else {
	    if (lnum > 1) {
		for(i = lnum/2, n1 = SUCC(left); i; i--)
		    n1 = SUCC(n1);
		QLSortNodesRec(left, pivot, n1, cmpfunc);
	    } else if (rnum <= 1) {
		break;
	    }
	    for(i = rnum/2, n2 = SUCC(pivot); i; i--) {
		n2 = SUCC(n2);
	    }
	    left = pivot;
	    pivot = n2;
	}
    }
}

void QLSortNodes(struct Node *phead, struct Node *ptail, 
	int (*cmpfunc)(void *, void *))
{
    if (phead && ptail && cmpfunc) {
        struct Node *node, *pivot;
        long num = 0, i;

        for (node = SUCC(phead); node != ptail; node = SUCC(node)) {
            num++;
	}

        if(num > 1) {
            for(i = num/2, node = SUCC(phead); i; i--) {
                node = SUCC(node);
	    }
            pivot = node;
            QLSortNodesRec(phead, ptail, pivot, cmpfunc);
        }
    }
}

void QLSort(struct List *list, int (*cmpfunc)(void *, void *))
{
    if(list)
        QLSortNodes(PRED(HEAD(list)), SUCC(TAILPRED(list)), cmpfunc);
}

long NumNodes(struct List *list)
{
    struct Node *node;
    long num=0;

    if(list)
	FORNODE(node, list)
	    num++;

    return(num);
}

long NodeNum(void *node)
{
    long num = 0;

    if(node != NULL)
        for(; PRED(node); node = PRED(node))
	{
            num++;
#ifdef DEBUG
	    if (num >= 0x10000000)
	    {
		DPRINTF(("NodeNum: max count (0x%08lx) reached.\n", num));
		num = 0;
		break;
	    }
#endif
	}

    return(num);
}

void *FindNodeNum(struct List *list, ulong find)
{
    struct Node *node;

    if(list && find > 0)
        FORNODE(node, list)
            if(--find == 0)
                return(node);

    return(NULL);
}

void *FindNodePtr(struct List *list, void *find)
{
    struct Node *node;

    if(list && find)
        FORNODE(node, list)
            if(node == find)
                return(node);

    return(NULL);
}

ushort HashInteger(long integer)
{
    ushort val = 0;

    val = (val ^ (ushort)(integer >> 16)) * 69 + 32767;
    val = (val << 1) | ((val & 0x8000)? 1 : 0);
    val = (val ^ (ushort)(integer & 0xffff)) * 69 + 32767;
    val = (val << 1) | ((val & 0x8000)? 1 : 0);

    return(val);
}

ushort HashReal(double real)
{
#ifdef AMIGA
    long l[2];
#else
    int32_t l[2];
#endif
    ushort val = 0;

    memcpy(l, &real, sizeof(l));

    val = (val ^ HashInteger(l[0])) * 69 + 32767;
    val = (val << 1) | ((val & 0x8000)? 1 : 0);
    val = (val ^ HashInteger(l[1])) * 69 + 32767;
    val = (val << 1) | ((val & 0x8000)? 1 : 0);

    return(val);
}

ushort HashString(const char *str)
{
    ushort val = 0;
    const char *s;
    ushort tmp;

    if(str)
    {
        for(s = str; *s; s += 2)
        {
            tmp = ((ushort)CodesetToUpper((cschar)s[0])) << 8 
		 | (ushort)CodesetToUpper((cschar)s[1]);
            val = (val ^ tmp) * 69 + 32767;
            val = (val << 1) | ((val & 0x8000)? 1 : 0);
            if(s[1] == 0)
                break;
        }
    }
    return(val);
}

/*
** sha1sum: compute the checksum over a zero-terminated byte array
** using the SHA1 algorithm.
** The following code is from http://www.hoozi.com/Articles/SHA1_Source.htm
** (Secure Hash Algorithm (SHA-1) Reference implementation in C/C++)
*/

#define rotateleft(x,n) (((x<<n) | ((x & 0xFFFFFFFF)>>(32-n))) & 0xFFFFFFFF)
/* #define rotateright(x,n) ((x>>n) | (x<<(32-n))) */

const char *SHA1Sum(const char *in)
{
    static char out[8*5+1];
    static unsigned long int word[80];
    unsigned long int h0, h1, h2, h3, h4, a, b, c, d, e, f, k, temp;
    unsigned char *str;
    int current_length, original_length, number_of_chunks;
    int ib, i, j, m;

    if (in == NULL) {
	return NULL;
    }

    h0 = 0x67452301;
    h1 = 0xEFCDAB89;
    h2 = 0x98BADCFE;
    h3 = 0x10325476;
    h4 = 0xC3D2E1F0;

    str = (unsigned char *)malloc(strlen(in) + 100);
    if (str == NULL) {
	return NULL;
    }
    strcpy((char *)str, in);

    current_length = strlen((const char *)str);
    original_length = current_length;
    str[current_length] = 0x80;
    str[current_length + 1] = '\0';
    current_length++;

    ib = current_length % 64;
    if (ib < 56) {
	ib = 56 - ib;
    } else {
	ib = 120 - ib;
    }

    for (i=0; i < ib; i++) {
	str[current_length] = 0x00;
	current_length++;
    }
    str[current_length + 1] = '\0';

    for (i=0; i<6; i++) {
	str[current_length] = 0x0;
	current_length++;
    }
    str[current_length] = (original_length * 8) / 0x100 ;
    current_length++;
    str[current_length] = (original_length * 8) % 0x100;
    current_length++;
    str[current_length+i] = '\0';

    number_of_chunks = current_length / 64;
    for (i=0; i < number_of_chunks; i++) {
	for (j=0; j < 16; j++) {
	    word[j] = str[i*64 + j*4 + 0] * 0x1000000 + 
		str[i*64 + j*4 + 1] * 0x10000 + 
		str[i*64 + j*4 + 2] * 0x100 + str[i*64 + j*4 + 3];
	}
	for (j=16; j<80; j++) {
	    word[j] = rotateleft(
		    (word[j-3] ^ word[j-8] ^ word[j-14] ^ word[j-16]), 1);
	}

	a = h0;
	b = h1;
	c = h2;
	d = h3;
	e = h4;

	for (m=0; m<80; m++) {
	    if (m <= 19) {
		f = (b & c) | ((~b) & d);
		k = 0x5A827999;
	    } else if (m <= 39) {
		f = b ^ c ^ d;
		k = 0x6ED9EBA1;
	    } else if (m <= 59) {
		f = (b & c) | (b & d) | (c & d);
		k = 0x8F1BBCDC;
	    } else {
		f = b ^ c ^ d;
		k = 0xCA62C1D6; 
	    }

	    temp = (rotateleft(a,5) + f + e + k + word[m]) & 0xFFFFFFFF;
	    e = d;
	    d = c;
	    c = rotateleft(b,30);
	    b = a;
	    a = temp;
	}

	h0 = (h0 + a) & 0xFFFFFFFF;
	h1 = (h1 + b) & 0xFFFFFFFF;
	h2 = (h2 + c) & 0xFFFFFFFF;
	h3 = (h3 + d) & 0xFFFFFFFF;
	h4 = (h4 + e) & 0xFFFFFFFF;
    }
    free(str);

    sprintf(out, "%08lx%08lx%08lx%08lx%08lx", h0, h1, h2, h3, h4);

    return out;
}

void SPutShort(char *s, short h)                /* 16 bit */
{
    s[0] = (char)(h >> 8);
    s[1] = (char)h;
}

void SPutLong(char *s, long l)                  /* 32 bit */
{
    s[0] = (char)(l >> 24);
    s[1] = (char)(l >> 16);
    s[2] = (char)(l >> 8);
    s[3] = (char)l;
}

void SPutDouble(char *s, double d)              /* 64 bit */
{
    union {
        double d;
        long long l;
    } mix;
    long long l;

    mix.d = d;
    l = mix.l;

    s[0] = (char)(l >> 56);
    s[1] = (char)(l >> 48);
    s[2] = (char)(l >> 40);
    s[3] = (char)(l >> 32);
    s[4] = (char)(l >> 24);
    s[5] = (char)(l >> 16);
    s[6] = (char)(l >> 8);
    s[7] = (char)l;
}

short SGetShort(const char *s)                  /* 16 bit */
{
    short h0 = (signed   char)s[0];
    short h1 = (unsigned char)s[1];

    return ((h0 << 8) | h1);
}

long SGetLong(const char *s)                    /* 32 bit */
{
    long l0 = (signed   char)s[0];
    long l1 = (unsigned char)s[1];
    long l2 = (unsigned char)s[2];
    long l3 = (unsigned char)s[3];

    return ((l0 << 24) | (l1 << 16) | (l2 << 8) | l3);
}

double SGetDouble(const char *s)                /* 64 bit */
{
    union {
        long long l;
        double d;
    } mix;
    long long l0 = (unsigned char)s[0];
    long long l1 = (unsigned char)s[1];
    long long l2 = (unsigned char)s[2];
    long long l3 = (unsigned char)s[3];
    long long l4 = (unsigned char)s[4];
    long long l5 = (unsigned char)s[5];
    long long l6 = (unsigned char)s[6];
    long long l7 = (unsigned char)s[7];

    mix.l = ((l0 << 56) | (l1 << 48) | (l2 << 40) | (l3 << 32) |
             (l4 << 24) | (l5 << 16) | (l6 <<  8) | l7);

    return mix.d;
}

void FPutChar(FILE *file, char c)               /*  8 bit */
{
    fputc(c, file);
}

void FPutBool(FILE *file, bool b)               /*  8 bit */
{
    fputc((char)b, file);
}

void FPutShort(FILE *file, short s)             /* 16 bit */
{
    fputc((char)(s >> 8), file);
    fputc((char)s, file);
}

void FPutLong(FILE *file, long l)               /* 32 bit */
{
    fputc((char)(l >> 24), file);
    fputc((char)(l >> 16), file);
    fputc((char)(l >> 8), file);
    fputc((char)l, file);
}

void FPutDouble(FILE *file, double d)           /* 64 bit */
{
    union {
        double d;
        long long l;
    } mix;
    long long l;

    mix.d = d;
    l = mix.l;

    fputc((char)(l >> 56), file);
    fputc((char)(l >> 48), file);
    fputc((char)(l >> 40), file);
    fputc((char)(l >> 32), file);
    fputc((char)(l >> 24), file);
    fputc((char)(l >> 16), file);
    fputc((char)(l >> 8), file);
    fputc((char)l, file);
}

void FPutS(FILE *file, const char *s)           /* variable length */
{
    if (s) {
        fputs(s, file);
    }
    fputc(0, file);
}

void FPutUTF8(FILE *file, const char *s)        /* variable length */
{
    CodesetWriteUTF8(s, file);
    fputc(0, file);
}

char FGetChar(FILE *file)                       /*  8 bit */
{
    return (char)fgetc(file);
}

bool FGetBool(FILE *file)                       /*  8 bit */
{
    return (bool)fgetc(file);
}

short FGetShort(FILE *file)                     /* 16 bit */
{
    short s0 = (signed   char)fgetc(file); 
    short s1 = (unsigned char)fgetc(file);

    return ((s0 << 8) | s1);
}

long FGetLong(FILE *file)                       /* 32 bit */
{
    long l0 = (signed   char)fgetc(file); 
    long l1 = (unsigned char)fgetc(file);
    long l2 = (unsigned char)fgetc(file);
    long l3 = (unsigned char)fgetc(file);

    return ((l0 << 24) | (l1 << 16) | (l2 << 8) | l3);
}

double FGetDouble(FILE *file)                   /* 64 bit */
{
    union {
        long long l;
        double d;
    } mix;
    long long l0 = (unsigned char)fgetc(file);
    long long l1 = (unsigned char)fgetc(file);
    long long l2 = (unsigned char)fgetc(file);
    long long l3 = (unsigned char)fgetc(file);
    long long l4 = (unsigned char)fgetc(file);
    long long l5 = (unsigned char)fgetc(file);
    long long l6 = (unsigned char)fgetc(file);
    long long l7 = (unsigned char)fgetc(file);

    mix.l = ((l0 << 56) | (l1 << 48) | (l2 << 40) | (l3 << 32) |
             (l4 << 24) | (l5 << 16) | (l6 <<  8) | l7);

    return mix.d;
}

void FGetS(FILE *file, char *s, size_t size)   /* variable length */
{
    char c;

    if (s && size > 0) {
        do {
            c = fgetc(file);
            size--;
            if (size == 0) {
                c = 0;
            }
            *s++ = c;
        } while (c != 0);
    }
}

