/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** util.h
**
** Creation date: 24.11.1994
** Last modified: 29.07.2020
*/

#ifndef UTIL_H
#define UTIL_H

#include "system.h"
#include "codeset.h"

#ifdef __cplusplus
extern "C" {
#endif

#define ANSI_RESET      "\33c"
#define ANSI_NORMAL     "\33[0m"
#define ANSI_ITALICON   "\33[3m"
#define ANSI_ITALICOFF  "\33[23m"
#define ANSI_UNDERLINEON "\33[4m"
#define ANSI_UNDERLINEOFF "\33[24m"
#define ANSI_BOLDFACEON "\33[1m"
#define ANSI_BOLDFACEOFF "\33[22m"
#define ANSI_ELITEON    "\33[2w"
#define ANSI_ELITEOFF   "\33[1w"
#define ANSI_CONDENSEDON "\33[4w"
#define ANSI_CONDENSEDOFF "\33[3w"
#define ANSI_WIDEON     "\33[6w"
#define ANSI_WIDEOFF    "\33[5w"
#define ANSI_NLQON      "\33[2\"z"
#define ANSI_NLQOFF     "\33[1\"z"

#define JOKER '*'
#define SINGLEJOKER '?'

#define FORMAT_DECIMALPOINT 0
#define FORMAT_DECIMALCOMMA 1

#define FORMAT_DAYMONTHYEAR 0
#define FORMAT_MONTHDAYYEAR 1
#define FORMAT_YEARMONTHDAY 2

#define FORMAT_HOURMINSEC 0
#define FORMAT_MINSEC 1
#define FORMAT_HOURMIN 2

#define DATE_STR_LEN 10
#define TIME_STR_MAXLEN (7+6)

struct MemPool;

long NextVersion(long version);
/*
** Returns version + 1, or 1 when version + 1 overflows to 0.
*/

char *WordString(struct MemPool *memPool, const char *s, long n, bool *err);
/*
** Returns the n-th word (beginning with 0) of string s.  Words are non-empty
** sub-strings separated by one or more non-breakable space characters.
** The memory for the sub string will be allocated from the given memPool.
** If n is out of range, NULL is returned and *err is set to FALSE.
** If n is in range and memory allocation has been successful
** then the sub string is returned and *err is set to FALSE.
** If memory allocation fails, NULL is returned and *err is
** set to TRUE.
*/

long NumWords(const char *s);
/*
** Returns the number of words of the given string s.
*/

char *FieldString(struct MemPool *memPool, const char *s, long n,
        const char *sep, bool quotes, bool *err);
/*
** Returns the n-th field (beginning with 0) of string s.  Fields are sub-
** strings separated by exactly one separator character.  A field may be an
** an empty string.  Argument sep contains the characters that separate fields.
** If sep = NULL then a default set of non-breakable space characters is used.
** If quotes is TRUE then fields can be surrounded by double quotes and may
** contain any of the separator characters.  The double quotes are stripped
** when returning the field.
** See WordString for how memory is allocated and the value of *err.
*/

long StrNumFields(const char *s, const char *sep, bool quotes);
/*
** Returns the number of fields of the given string s.
*/

long StrCount(const char *s, cschar c);
/*
** Returns the number of times character c appears in string s.
*/

bool StrIsASCII(const char *str);
/*
** Checks whether the string conforms to 7bit ASCII (or is NULL).
*/

char *StrCompress(struct MemPool *memPool, const char *str);
/*
** Replaces all escaped characters with their one byte equivalent.
** That is '\n', '\t', '\???', \x?? are converted to real characters.
** The special non-ansi sequence '\e' is translated into '\33' (escape).
** The escape sequence '\0' is translated into '\1'.
** The new string is allocated from the given memPool and returned
** (may be NULL).
*/

char *BuildCommandLine(struct MemPool *memPool, const char *pattern, 
	const char *arg);
/*
** Creates a command line by substituting '%f' sequences in pattern
** with the contents of arg put into double quotes.  Parameter arg
** can be NULL in which case '%f" is replaced with an empty string.
*/

cschar Shortcut(const char *s);
/*
** Examines string s for the first occurence of an
** underscore and returns the following character in
** lowercase or 0 if not found.
*/

const char *EmptyNullStr(const char *s);
/*
** Returns string s or empty string  if s == NULL.
*/

bool StrJokerCmp(const char *s, const char *j);
/*
** Returns TRUE if string s matches joker string j (case insensitive).
*/

bool strjokercmp(const char *s, const char *j);
/*
** Returns TRUE if string s matches joker string j (case sensitive).
*/

ulong StrToId(const char *s);
/*
** converts up to 4 characters of the given into an ulong id, e.g.
** StrToId("abcd") returns ('a' << 24 | 'b' << 16 | 'c' << 8 || 'd')
*/

const char *IdToStr(ulong id);
/*
** converts the id back to its string representation.
*/

bool StrToInt(const char *string, long *result);
/*
** returns TRUE if string can be converted to a long value
** which will be stored in result (may be NULL for testing only).
** Returns FALSE if string is empty.
*/

char *IntToStr(long value);
/*
** returns string representation of long value.
** The string must be copied by you.
*/

bool StrToDouble(const char *string, double *result);
/*
** returns TRUE if string can be converted to a double value
** which will be stored in result (may be NULL for testing only).
*/

char *DoubleToStr(double value, short numdecimals, short format);
/*
** returns string representation of double value.
** numdecimals determines the number of decimal places,
** if negative, sprintf(..., "%.10g", ...) is called,
** format determines the decimal point: 0 for a point, > 0 for a comma.
** The string must be copied by you.
*/

char *DoubleToStr2(double value, short numdecimals, char efg, short format);
/*
** returns string representation of double value.
** numdecimals determines the number of decimal places (>= 0).
** efg is one of 'e', 'f', or 'g' for conversion.
** format the decimal point: 0 for a point, > 0 for a comma.
** The string must be copied by you.
*/

bool DoubleToInt(double d, long *l);
/*
** converts double value d to a long value.
** if d is out of range, FALSE is returned,
** otherwise *l is set to the nearest integer value of d
** and TRUE is returned.
*/

#define YEAR_VALID(year) ((year) >= 0 && (year) <= 9999)
#define MONTH_VALID(month) ((month) >= 1 && (month) <= 12)

short MonthDays(short month, short year);
/*
** returns the number of days of the given month and year.
** Month and year must be valid or the result will be meaningless.
*/

short YearDays(short year);
/*
** returns the number of days of the given year.
** Year must be valid or the result will be meaningless.
*/

void DateToDayMonthYear(long date, short *pday, short *pmonth, short *pyear);
/*
** Separates day, month, and year from a date and returns the values
** in the corresponding arguments if non-NULL.
*/

bool DayMonthYearToDate(long day, long month, long year, long *result);
/*
** Creates a date value from the given day, month, and year, stores it in
** *result (if not NULL) and returns TRUE.  If the specified date is not
** valid then FALSE is returned.
*/ 

long DateToday(void);
/*
** Returns the current date.
*/

bool DateAdd(long date, long days, long *result);
/*
** adds days to date and returns it in result (may be NULL).
** Returns TRUE if result is in range, FALSE otherwise.
*/

bool DateAddMonth(long date, long months, long *result);
/*
** adds months to the month field of date and returns it in result 
** (may be NULL).  Might change day field to ensure day is within maximum
** day of the new month.  Handles overflow and changes year accordingly.
** Returns TRUE if result is in range of allowed date values, FALSE otherwise.
*/

bool DateAddYear(long date, long years, long *result);
/*
** adds years to the year field of date and returns it in result 
** (may be NULL).  Might change day field to ensure day is within maximum
** day of the month in the new year.
** Returns TRUE if result is in range of allowed date values, FALSE otherwise.
*/

bool StringToDate(const char *string, long *result);
/*
** returns TRUE if string can be converted to a date value
** (number of days since 01.01.0000).
** The date value will be stored in result
** (may be NULL for testing only).
*/

char *DateToString(long value, short format);
/*
** returns string representation of date value.
** format determines the output string format and must be one
** of FORMAT_{DAYMONTHYEAR,MONTHDAYYEAR,YEARMONTHDAY}.
** The string must be copied by you.
*/

bool StringToTime(const char *string, short format, long *result);
/*
** Returns TRUE if string can be converted to a time value.
** format is a hint about interpreting the string and must
** be one of FORMAT_{HOURMINSEC,MINSEC,HOURMIN}.
** The time value will be stored in result
** (may be NULL for testing only).
*/

long TimeNow(void);
/*
** Returns the current time.
*/

char *TimeToString(long value, short format);
/*
** returns string representation of time value.
** format must be one of FORMAT_{HOURMINSEC,MINSEC,HOURMIN}.
** The string must be copied by you.
*/

bool TimeAdd(long time, long secs, long *result);
/*
** adds seconds secs to time and returns it in result (may be NULL).
** Returns TRUE if result is in range, FALSE otherwise.
*/

extern const struct Node *CorruptedNode; /* holds pointer to currupted node */

void SwapLists(struct List *list1, struct List *list2);
/*
** Swaps the two lists.
*/

long CheckList(const struct List *list);
/*
** Checks the consistency of a list.
** If a corrupted node is found, CorruptedNode is set 
** and the guessed number of missing nodes is returned.  
** Otherwise 0 is returned.
*/

void QLSort(struct List *list, int (*cmpfunc)(void *, void *));
/*
** Quick sort for lists. cmpfunc receives two pointers to list elements.
*/

void QLSortNodes(struct Node *phead, struct Node *ptail, int (*cmpfunc)(void *, void *));
/*
** Quick sort like QLSort. Here you specify the nodes to be sorted by
** phead whoose successor is the first node to be sorted, and
** ptail whoose predecessor is the last node to be sorted.
*/

long NumNodes(struct List *list);
/*
** Returns number of nodes in given list.
*/

long NodeNum(void *node);
/*
** Returns node number starting with 1.
** Returns 0 if node == NULL.
*/

void *FindNodeNum(struct List *list, ulong find);
/*
** Returns (find)-nth node of the given list
** starting with 1 or NULL if node doesn't exist.
*/

void *FindNodePtr(struct List *list, void *find);
/*
** Checks is a given node is in the list.
** Returns the node or NULL if not found.
*/

/*
** Hash functions. Return ushort values.
*/
ushort HashInteger(long integer);
ushort HashReal(double real);
ushort HashString(const char *str);   /* case insensitive */

/*
** Compute checksum over a zero-terminated byte array using the SHA1 algorithm.
*/
const char *SHA1Sum(const char *in);


/*
** Reading and writing of numbers to byte strings and FILE streams.
*/
short SGetShort(const char *s);                 /* 16 bit */
long SGetLong(const char *s);                   /* 32 bit */
double SGetDouble(const char *s);               /* 64 bit */

void SPutShort(char *s, short h);               /* 16 bit */
void SPutLong(char *s, long l);                 /* 32 bit */
void SPutDouble(char *s, double d);             /* 64 bit */

void FPutChar(FILE *file, char c);              /*  8 bit */
void FPutBool(FILE *file, bool b);              /*  8 bit */
void FPutShort(FILE *file, short s);            /* 16 bit */
void FPutLong(FILE *file, long l);              /* 32 bit */
void FPutDouble(FILE *file, double d);          /* 64 bit */
void FPutS(FILE *file, const char *s);          /* variable length */
void FPutUTF8(FILE *file, const char *s);       /* variable length */

char FGetChar(FILE *file);                      /*  8 bit */
bool FGetBool(FILE *file);                      /*  8 bit */
short FGetShort(FILE *file);                    /* 16 bit */
long FGetLong(FILE *file);                      /* 32 bit */
double FGetDouble(FILE *file);                  /* 64 bit */
void FGetS(FILE *file, char *s, size_t size);   /* variable length */

#ifdef __cplusplus
}
#endif

#endif

