/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** help.c
**
** Creation date: 19.05.2015
** Last modified: 29.12.2020
*/

#include "system.h"
#include "requesters.h"
#include "guidosutil.h"
#include "dosutil.h"
#include "memory.h"
#include "strhandle.h"
#include "util.h"
#include "appstrings_cat.h"

void GuiShowHelp(const char *section)
{
    struct MemPool *memPool = CreatePool(0, 4096, 1024);
    char *docdir = ExpandPathname("$BeeBase/doc", memPool);
    const char *lang = getenv("LANG");
    char subdir[3];
    char *dirname = 0;
    const char *filename = "BeeBase.html";
    char *path = 0;

    if (lang && lang[0] && lang[1]) {
	subdir[0] = lang[0];
	subdir[1] = lang[1];
	subdir[2] = 0;
	dirname = JoinPathname(memPool, docdir, subdir);
	dirname = JoinPathname(memPool, dirname, "html");
	if (FileStat(dirname) != 0) {
	    dirname = 0;
	}
    }
    if (dirname == 0) {
	dirname = JoinPathname(memPool, docdir, "en/html");
    }

    if (!section) {
        const char *helpnode = 0;
        GuiObject go = GuiActiveObject(GuiActiveWindow());

        while (go && !helpnode) {
            helpnode = GuiObjectHelpNode(go);
            go = GuiObjectParent(go);
        }
        section = helpnode;
    }

    if (section && dirname) {
	bool found = FALSE;
	Dir *dir = DirOpen(dirname);

	if (dir) {
    	    const char *name;
	    while ((name = DirReadName(dir)) != 0) {
		if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0) {
		    continue;
		}
		char *path = JoinPathname(memPool, dirname, name);
		if (FileStat(path) > 0) {
		    FILE *file = FOpen(path, "r");      // ascii
		    if (file) {
			static char buffer[1024];
			while (fgets(buffer, sizeof(buffer), file)) {
			    if (strncmp(buffer, "<a name=", 8) != 0) {
				continue;
			    }
			    if (strstr(buffer, section)) {
				found = TRUE;
				break;
			    }
			}
                        fclose(file);
		    }
		}
		FreeString(memPool, path);
		if (found) {
		    filename = AllocString(memPool, name);
		    break;
		}
	    }
	    DirClose(dir);
	}
	if (!found) {
	    section = 0;
	}
    }

    if (dirname && filename) {
        char *s;
        size_t cmdsize;

	// normalize path to use only '/' characters
	for (s = dirname; *s; s++) {
	    if (*s == '\\') {
		*s = '/';
	    }
	}
	if (*dirname == '/') {
	    dirname++;              // first '/' will be part of file:///
	}
	cmdsize = 8 + strlen(dirname) + 1 + strlen(filename) +
    	    (section? 1 + strlen(section) : 0) + 1;
	path = (char *)PoolAllocMem(memPool, cmdsize);
    }
    if (path) {
#ifdef MACOSX
        // #section does not seem to work on MacOSX.
        section = 0;
#endif
	sprintf(path, "file:///%s/%s%s%s", dirname, filename, 
		section? "#" : "", EmptyNullStr(section));
	GuiViewFile(path, TRUE);
    } else {
        Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Mem));
    }
    if (memPool != NULL) {
        DeletePool(memPool);
    }
}

