/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** mempool.c
**
** Creation date: 24.11.96
** Last modified: 11.12.97
*/

#include "system.h"
#include "mempool.h"

struct FreeEntry
{
   struct MinNode fe_Node;
   ulong fe_Size;               /* free entry size including FreeEntry header */
};

struct MemBlock
{
   struct MinNode mb_Node;
   ulong mb_Size;           /* size without MemBlock structure */
   ulong mb_Free;           /* total number of free bytes in this MemBlock */
   struct List mb_FreeEntryList;        /* list of FreeEntries */
   double mb_Memory[0];     /* memory (double for 8 byte alignment) */
};

struct MemPool
{
   struct List mp_MemBlockList;
   struct MemBlock *mp_CurrentMemBlock;
   ulong mp_PuddleSize;
   ulong mp_ThreshSize;
};

#define SIZE8(size)     ((size+7) & ~7)
#define MEM_MIN         SIZE8(sizeof(struct FreeEntry))
#define REAL_SIZE(size) MAX(SIZE8(size), MEM_MIN)


static void *getMemFromFreeEntry(struct FreeEntry *fe, ulong size)
{
    void *ret = NULL;

    if(fe->fe_Size == size)
    {
        REMOVE(fe);
        ret = fe;
    }
    else
    {
        ret = (char *)fe + fe->fe_Size - size;
        fe->fe_Size -= size;
    }
    return(ret);
}

static void *getMemFromBlock(struct MemBlock *mb, ulong size)
{
    ulong minsize;
    struct FreeEntry *fe;

    if(size > mb->mb_Free)
        return(NULL);

    minsize = size + MEM_MIN;
    FORNODE(fe, &mb->mb_FreeEntryList)
        if(fe->fe_Size >= minsize || fe->fe_Size == size)
        {
            mb->mb_Free -= size;
            return(getMemFromFreeEntry(fe, size));
        }
   return(NULL);
}

static void *getMemFromPool(struct MemPool *mp, ulong size)
{
    struct MemBlock *mb = mp->mp_CurrentMemBlock;
    void *ptr = NULL;

    if(mb)
    {
        ptr = getMemFromBlock(mb, size);
        if(ptr != NULL)
            return(ptr);
    }

    FORNODE(mb, &mp->mp_MemBlockList)
    {
        ptr = getMemFromBlock(mb, size);
        if(ptr != NULL)
        {
            mp->mp_CurrentMemBlock = mb;
            return(ptr);
        }
    }
    return(NULL);
}

static struct MemBlock *allocMemBlock(ulong size)
{
    struct MemBlock *mb = NULL;
    struct FreeEntry *fe = NULL;

    mb = ALLOCMEM(size + sizeof(struct MemBlock));
    if(mb == NULL)
        return(NULL);

    mb->mb_Size = size;
    mb->mb_Free = size;
    NewList(&mb->mb_FreeEntryList);

    fe = (struct FreeEntry *)mb->mb_Memory;
    ADDHEAD(&mb->mb_FreeEntryList, fe);
    fe->fe_Size = size;

    return(mb);
}

void *AllocPooled(struct MemPool *mp, ulong size)
{
    void *ret = NULL;
    struct MemBlock *mb = NULL;

    if(mp == NULL)
        return(NULL);

    size = REAL_SIZE(size);

    if(size <= mp->mp_ThreshSize)
    {
        ret = getMemFromPool(mp, size);
        if(ret != NULL)
            return(ret);
    }

    /*
    ** We need a new MemBlock
    */
    mb = allocMemBlock((size <= mp->mp_ThreshSize)? mp->mp_PuddleSize : size);
    if(mb == NULL)
        return(NULL);
    ADDHEAD(&mp->mp_MemBlockList, mb);
    mp->mp_CurrentMemBlock = mb;

    ret = getMemFromBlock(mb, size);
    return(ret);
}

static void freeMemBlock(struct MemBlock *mb)
{
    FREEMEM(mb, sizeof(struct MemBlock) + mb->mb_Size);
}

void FreePooled(struct MemPool *mp, void *ptr, ulong size)
{
    if(mp && ptr)
    {
        struct MemBlock *mb;

        size = REAL_SIZE(size);
                                                                            
        FORNODE(mb, &mp->mp_MemBlockList)
	{
	    char *mem = (char *)mb->mb_Memory;
            if ((char *)ptr >= mem && (char *)ptr < mem + mb->mb_Size)
            {
#ifdef DEBUG_MEMPOOL
		{
                    struct FreeEntry *fe;
                    FORNODE(fe, &mb->mb_FreeEntryList)
                        if((void *)fe > ptr)
                            break;

		    if (!SUCC(fe) && (char *)ptr + size > mem + mb->mb_Size)
		    {
			DFPRINTF((stderr, 
			    "FreePooled: memory area goes beyond block!\n"));
		    }
		    if (SUCC(fe) && (char *)fe < (char *)ptr + size)
		    {
			DFPRINTF((stderr, "FreePooled: overlapping entries: "
				"%p size=%lx and %p size=%lx\n",
				ptr, size, fe, fe->fe_Size));
		    }
                    fe = PRED(fe);
		    if (PRED(fe) && (char *)fe + fe->fe_Size > (char *)ptr)
		    {
			DFPRINTF((stderr, "FreePooled: overlapping entries: "
				"%p size=%lx and %p size=%lx\n",
				ptr, size, fe, fe->fe_Size));
		    }
		}
#endif
                mb->mb_Free += size;
                if(mb->mb_Free == mb->mb_Size)
                {
                    REMOVE(mb);
                    freeMemBlock(mb);
                    if(mp->mp_CurrentMemBlock == mb)
                        mp->mp_CurrentMemBlock = NULL;
                }
                else
                {
                    struct FreeEntry *cur = (struct FreeEntry *)ptr, *fe;

                    /*
                    ** insert cur
                    */
                    FORNODE(fe, &mb->mb_FreeEntryList)
                        if(fe > cur)
                            break;
                    INSERT(NULL, cur, PRED(fe));
                    cur->fe_Size = size;

                    /*
                    ** maybe we can join free entries
                    */
                    fe = SUCC(cur);
                    if(SUCC(fe) && (char *)fe == (char *)cur + cur->fe_Size)
                    {
                        REMOVE(fe);
                        cur->fe_Size += fe->fe_Size;
                    }
                    fe = PRED(cur);
                    if(PRED(fe) && (char *)fe + fe->fe_Size == (char *)cur)
                    {
                        REMOVE(cur);
                        fe->fe_Size += cur->fe_Size;
                    }
                }
                return;
            }
	}
#ifdef DEBUG_MEMPOOL
	DFPRINTF((stderr, "FreePooled: unknown memory area %p\n", ptr));
#endif
    }
#ifdef DEBUG_MEMPOOL
    else 
    {
	DFPRINTF((stderr, 
		"FreePooled: no pool pointer (%p) or memory (%p) given.\n",
		mp, ptr));
    }
#endif
}

static void freeMemPoolContents(struct MemPool *mp)
{
    if(mp)
    {
        struct MemBlock *mb;

        while((mb = (struct MemBlock *)REMHEAD(&mp->mp_MemBlockList)) != NULL)
            freeMemBlock(mb);
        mp->mp_CurrentMemBlock = NULL;
    }
}

void DeletePool(struct MemPool *mp)
{
   if(mp)
   {
      freeMemPoolContents(mp);
      FREEMEM(mp, sizeof(struct MemPool));
   }
}

#ifdef AMIGA
APTR
#else
struct MemPool *
#endif
CreatePool(ulong req, ulong puddleSize, ulong threshSize)
{
    struct MemPool *mp = NULL;

    mp = ALLOCMEM(sizeof(struct MemPool));
    if(mp == NULL)
        return(NULL);

    puddleSize = REAL_SIZE(puddleSize);
    threshSize = REAL_SIZE(threshSize);

    NewList(&mp->mp_MemBlockList);
    mp->mp_CurrentMemBlock = NULL;
    mp->mp_PuddleSize = puddleSize;
    mp->mp_ThreshSize = MIN(threshSize, puddleSize - MEM_MIN);
        /* make sure thresSize <= puddleSize - MEM_MIN */
    return(mp);
}

