/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkeditlabels.c
**
** Creation date: 08.05.2005
** Last modified: 29.03.2016
*/

#define ENABLE_NLS

#include "system.h"
#include "gtkeditlabels.h"
#include "gtk/classes/gtkextentry.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include <glib/gi18n.h>
#include "preferences.h"
#include "codeset.h"
#include "util.h"
#include "appstrings_cat.h"

#define P_(x) N_(x)

enum {
  PROP_0,
  PROP_LABELS,
  PROP_SORTABLE
};

static void gtk_edit_labels_class_init    (GtkEditLabelsClass *klass);
static void gtk_edit_labels_init          (GtkEditLabels      *edit);
static void gtk_edit_labels_finalize 	  (GObject	      *object);
static void gtk_edit_labels_set_property  (GObject            *object,
                                           guint               prop_id,
                                           const GValue       *value,
                                           GParamSpec         *pspec);
static void gtk_edit_labels_get_property  (GObject            *object,
                                           guint               prop_id,
                                           GValue             *value,
                                           GParamSpec         *pspec);
    
static GtkGroupClass *parent_class = NULL;

GType
gtk_edit_labels_get_type (void)
{
  static GType edit_labels_type = 0;

  if (!edit_labels_type)
    {
      static const GTypeInfo edit_labels_info =
      {
	sizeof (GtkEditLabelsClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_edit_labels_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkEditLabels),
	0,		/* n_preallocs */
	(GInstanceInitFunc) gtk_edit_labels_init,
      };

      edit_labels_type = g_type_register_static (GTK_TYPE_GROUP, 
	      					 "GtkEditLabels",
					         &edit_labels_info, 0);
    }

  return edit_labels_type;
}

static void
gtk_edit_labels_class_init (GtkEditLabelsClass *klass)
{
  GObjectClass *gobject_class;

  gobject_class = G_OBJECT_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);

  gobject_class->finalize = gtk_edit_labels_finalize;
  gobject_class->set_property = gtk_edit_labels_set_property;
  gobject_class->get_property = gtk_edit_labels_get_property;

  g_object_class_install_property (gobject_class,
				   PROP_LABELS,
				   g_param_spec_pointer ("labels",
						      P_("Labels"),
						      P_("Array of isocode strings for the labels"),
                                                      G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
  g_object_class_install_property (gobject_class,
				   PROP_SORTABLE,
				   g_param_spec_boolean ("sortable",
						      P_("Sortable"),
						      P_("Whether labels can be sorted"),
                                                      TRUE,
                                                      G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
}

static void 
editLabelsOnSelectionChanged (GtkEditLabels *edit)
{
  long pos = GuiListActive (edit->tree);
  long num = GuiListNumEntries (edit->tree);
  const gchar *str = GuiListNthEntry (edit->tree, pos);

  gtk_widget_set_sensitive (edit->remb, str != NULL &&
          (gtk_edit_labels_get_sortable (edit) || pos + 1 == num));
  gtk_widget_set_sensitive (edit->entry, str != NULL);
  gtk_entry_set_text (GTK_ENTRY (edit->entry), EmptyNullStr (str));
  gtk_widget_grab_focus (edit->entry);
}

static void 
editLabelsOnEntryChanged (GtkEditLabels *edit)
{
  const gchar *text = gtk_entry_get_text (GTK_ENTRY (edit->entry));
  GtkTreeSelection *selection = 
      gtk_tree_view_get_selection (GTK_TREE_VIEW (edit->tree));
  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (edit->tree));
  GtkListStore *store = GTK_LIST_STORE (model);
  GtkTreeIter iter;
    
  if (gtk_tree_selection_get_selected (selection, NULL, &iter))
    {
      gtk_list_store_set(store, &iter, 0, text, -1);
    }
}

static void
editLabelsNewClick (GtkEditLabels *edit)
{
  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (edit->tree));
  GtkListStore *store = GTK_LIST_STORE (model);
  long num = GuiListNumEntries (edit->tree);
  GtkTreeIter iter;
  gtk_list_store_append(store, &iter);
  gtk_list_store_set(store, &iter, 0, "", -1);
  GuiListSetActive (edit->tree, num);
}
  
static void
editLabelsRemClick (GtkEditLabels *edit)
{
  GtkTreeSelection *selection = 
      gtk_tree_view_get_selection (GTK_TREE_VIEW (edit->tree));
  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (edit->tree));
  GtkListStore *store = GTK_LIST_STORE (model);
  GtkTreeIter iter;
  if (gtk_tree_selection_get_selected (selection, NULL, &iter)) 
    {
      gboolean set = FALSE;
      if (gtk_list_store_remove (store, &iter)) 
        {
	  set = TRUE;
        } 
      else 
        {
	  gint n = gtk_tree_model_iter_n_children (model, NULL);
	  if (n > 0) 
	    {
	      set = gtk_tree_model_iter_nth_child (model, &iter, NULL, n - 1);
	    }
        }
      if (set)
	{
	  gtk_tree_selection_select_iter (selection, &iter);
	}
    }
}
  
static int
editLabelsCompare (const void *a, const void *b)
{
  return CodesetStrCollate(*(const char **)a, *(const char **)b, 
	  CS_COLLATE_ORDER);
}


static void
editLabelsSortClick (GtkEditLabels *edit)
{
  GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (edit->tree));
  GtkListStore *store = GTK_LIST_STORE (model);
  GtkTreeIter iter;
  gint n = gtk_tree_model_iter_n_children (model, NULL);
  gint i = 0;

  if (n > 0)
    {
      gchar **labels = g_new0 (char *, n);

      if (gtk_tree_model_get_iter_first (model, &iter))
	{
	  do
	    {
	      gtk_tree_model_get (model, &iter, 0, &labels[i], -1);
	    } while (++i < n && gtk_tree_model_iter_next (model, &iter));
	}
      qsort (labels, n, sizeof(char *), editLabelsCompare);

      gtk_list_store_clear (store);
      for (i=0; i < n; i++) 
        {
	  gtk_list_store_append (store, &iter);
	  gtk_list_store_set (store, &iter, 0, labels[i], -1);
	  g_free (labels[i]);
        }

      g_free (labels);
    }
}

static void
gtk_edit_labels_init (GtkEditLabels *edit)
{
  gboolean compact;
  GtkWidget *scroll, *control, *cont, *labelgroup, *top;
  GtkTreeSelection *selection;

  edit->labels = 0;

  edit->tree = GuiStringTreeView(FALSE);
  gtk_tree_view_set_reorderable (GTK_TREE_VIEW (edit->tree), TRUE);
  gtk_tree_view_set_enable_search (GTK_TREE_VIEW (edit->tree), FALSE);
  GuiSetTooltip(edit->tree, GCS(MSG_EditLabels_LabelsBH));

  selection  = gtk_tree_view_get_selection (GTK_TREE_VIEW (edit->tree));
  g_signal_connect_swapped (selection, "changed",
	  G_CALLBACK (editLabelsOnSelectionChanged), edit);

  scroll = ScrolledWindowObject,
      "hscrollbar-policy", GTK_POLICY_NEVER,
      "shadow-type", GTK_SHADOW_IN,
      "child", edit->tree,
      End;

  edit->newb = GuiSimpleButton(GCS(MSG_New), FALSE);
  GuiSetTooltip(edit->newb, GCS(MSG_EditLabels_NewBH));
  g_signal_connect_swapped (edit->newb, "clicked", 
	  G_CALLBACK (editLabelsNewClick), edit);

  edit->remb = GuiSimpleButton(GCS(MSG_EditLabels_Remove), FALSE);
  gtk_widget_set_sensitive(edit->remb, FALSE);
  GuiSetTooltip(edit->remb, GCS(MSG_EditLabels_RemoveBH));
  g_signal_connect_swapped (edit->remb, "clicked", 
	  G_CALLBACK (editLabelsRemClick), edit);

  edit->sortb = GuiSimpleButton(GCS(MSG_EditLabels_Sort), FALSE);
  GuiSetTooltip(edit->sortb, GCS(MSG_EditLabels_SortBH));
  g_signal_connect_swapped (edit->sortb, "clicked", 
	  G_CALLBACK (editLabelsSortClick), edit);

  compact = GetPrefsValue(NULL, PREFS_EDITLABELSCOMPACT);

  if(compact)
    {
      control = HGroup,
	  "child", edit->newb,
	  "child", edit->remb,
	  "child", edit->sortb,
	  End;

      cont = VGroup,
          "child", scroll,
	  "child", control, "weight", 0,
	  End;
    }
  else
    {
      control = VGroup,
	  "child", SpaceObject, End,
	  "child", edit->newb, "vert-weight", 0,
	  "child", edit->remb, "vert-weight", 0,
	  "child", edit->sortb, "vert-weight", 0,
	  "child", SpaceObject, End,
	  End;

      cont = HGroup,
    	  "child", scroll, "weight", 100,
	  "child", control, "weight", 10,
	  End;
      }

  edit->entry = ExtEntryObject,
      "sensitive", FALSE,
      "xalign", 0.5,
      "attached-tree-view", edit->tree,
      End;
  g_signal_connect_swapped (edit->entry, "changed", 
	  G_CALLBACK(editLabelsOnEntryChanged), edit);
  g_signal_connect_swapped (edit->tree, "focus-in-event", 
	  G_CALLBACK(gtk_widget_grab_focus), edit->entry);

  labelgroup = HGroup,
      ChildLabelWidget(GCS(MSG_EditLabels_Label), edit->entry),
      End;

  top = Frame(NULL),
      "child", GroupAlignment(FALSE),
          "child", VGroup,
              "child", cont,
	      "child", labelgroup, "vert-weight", 0,
	      End,
	  End,
      End;
  GuiSetHelpNode(top, "Label editor");

  gtk_group_add_child(&edit->group, top, 100, 100);
}

static void
gtk_edit_labels_finalize (GObject *object)
{
  GtkEditLabels *edit = GTK_EDIT_LABELS (object);

  g_free(edit->labels);
}

static void
gtk_edit_labels_set_property (GObject         *object,
                              guint            prop_id,
                              const GValue    *value,
                              GParamSpec      *pspec)
{
  GtkEditLabels *edit = GTK_EDIT_LABELS (object);

  switch (prop_id)
    {
    case PROP_LABELS:
      gtk_edit_labels_set_labels (edit, (char **)g_value_get_pointer (value));
      break;
    case PROP_SORTABLE:
      gtk_edit_labels_set_sortable (edit, g_value_get_boolean (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
gtk_edit_labels_get_property (GObject         *object,
                              guint            prop_id,
                              GValue          *value,
                              GParamSpec      *pspec)
{
  GtkEditLabels *edit = GTK_EDIT_LABELS (object);

  switch (prop_id)
    {
    case PROP_LABELS:
      g_value_set_pointer (value, gtk_edit_labels_get_labels (edit));
      break;
    case PROP_SORTABLE:
      g_value_set_boolean (value, gtk_edit_labels_get_sortable (edit));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

GtkWidget*
gtk_edit_labels_new (void)
{
  return g_object_new (GTK_TYPE_EDIT_LABELS, NULL);
}

void
gtk_edit_labels_set_labels (GtkEditLabels *edit, char **labels)
{
  GtkTreeModel *model;

  g_return_if_fail (GTK_IS_EDIT_LABELS (edit));

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (edit->tree));

  GuiSetListStoreLabels (GTK_LIST_STORE (model), (const char * const *)labels);
}

char **
gtk_edit_labels_get_labels (GtkEditLabels *edit)
{
  GtkTreeModel *model;

  g_return_val_if_fail (GTK_IS_EDIT_LABELS (edit), NULL);

  g_free(edit->labels);
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (edit->tree));
  edit->labels = GuiListStoreLabels (model);

  return edit->labels;
}

void
gtk_edit_labels_set_sortable (GtkEditLabels *edit, gboolean sortable)
{
  g_return_if_fail (GTK_IS_EDIT_LABELS (edit));

  gtk_widget_set_sensitive (edit->sortb, sortable);
  gtk_tree_view_set_reorderable (GTK_TREE_VIEW (edit->tree), sortable);
}

gboolean
gtk_edit_labels_get_sortable (GtkEditLabels *edit)
{
  g_return_val_if_fail (GTK_IS_EDIT_LABELS (edit), FALSE);

  return gtk_widget_get_sensitive (edit->sortb) || 
      gtk_tree_view_get_reorderable (GTK_TREE_VIEW (edit->tree));
}

