/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkexteditor.c
**
** GtkTextView with additional menu items for loading, saving, 
** and external editor.
**
** Creation data: 19.04.2004
** Last modified: 04.03.2015
*/

#include "system.h"
#include "gtkexteditor.h"
#include "gtk/gtkimpl.h"
#include "appstrings_cat.h"

enum {
  LOAD,
  SAVE,
  EXTERNAL_EDITOR,
  LAST_SIGNAL
};

static void gtk_ext_editor_class_init   (GtkExtEditorClass    *klass);
static void gtk_ext_editor_init         (GtkExtEditor         *editor);

static guint gtk_ext_editor_signals[LAST_SIGNAL] = { 0 };

GType
gtk_ext_editor_get_type (void)
{
  static GType ext_editor_type = 0;

  if (!ext_editor_type)
    {
      static const GTypeInfo ext_editor_info =
      {
	sizeof (GtkExtEditorClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_ext_editor_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkExtEditor),
	0,		/* n_preallocs */
	(GInstanceInitFunc) gtk_ext_editor_init,
	NULL,		/* value_table */
      };

      ext_editor_type = g_type_register_static (GTK_TYPE_TEXT_VIEW, 
	      "GtkExtEditor", 
	      &ext_editor_info, (GTypeFlags)0);
    }

  return ext_editor_type;
}

static void
gtk_ext_editor_class_init (GtkExtEditorClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  klass->load = NULL;
  klass->save = NULL;
  klass->external_editor = NULL;

  gtk_ext_editor_signals[LOAD] =
    g_signal_new ("load",
	          G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_FIRST,
		  G_STRUCT_OFFSET (GtkExtEditorClass, load),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__VOID,
		  G_TYPE_NONE, 0);
  gtk_ext_editor_signals[SAVE] =
    g_signal_new ("save",
	          G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_FIRST,
		  G_STRUCT_OFFSET (GtkExtEditorClass, save),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__VOID,
		  G_TYPE_NONE, 0);
  gtk_ext_editor_signals[EXTERNAL_EDITOR] =
    g_signal_new ("external_editor",
	          G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_FIRST,
		  G_STRUCT_OFFSET (GtkExtEditorClass, external_editor),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__VOID,
		  G_TYPE_NONE, 0);
}

static void
gtk_ext_editor_emit_load(GtkExtEditor *editor)
{
    g_signal_emit(editor, gtk_ext_editor_signals[LOAD], 0);

}

static void
gtk_ext_editor_emit_save(GtkExtEditor *editor)
{
    g_signal_emit(editor, gtk_ext_editor_signals[SAVE], 0);

}

static void
gtk_ext_editor_emit_external_editor(GtkExtEditor *editor)
{
    g_signal_emit(editor, gtk_ext_editor_signals[EXTERNAL_EDITOR], 0);

}

static void 
gtk_ext_editor_on_populate_popup(GtkExtEditor *editor, GtkMenu *menu)
{
    gboolean hasOpen = (g_signal_handler_find(editor, G_SIGNAL_MATCH_ID,
		gtk_ext_editor_signals[LOAD], 0, 0, 0, 0) != 0);
    gboolean hasSave = (g_signal_handler_find(editor, G_SIGNAL_MATCH_ID,
		gtk_ext_editor_signals[SAVE], 0, 0, 0, 0) != 0);
    gboolean hasExternalEditor = (g_signal_handler_find(editor, 
		G_SIGNAL_MATCH_ID, gtk_ext_editor_signals[EXTERNAL_EDITOR], 
		0, 0, 0, 0) != 0);
    
    if (hasOpen || hasSave || hasExternalEditor) {
	GtkMenuShell *shell = GTK_MENU_SHELL(menu);

	GtkWidget *item = gtk_separator_menu_item_new();
	gtk_widget_show(item);
	gtk_menu_shell_append(shell, item);
	
	if (hasOpen) {
#if GTK_MAJOR_VERSION == 2
	    item = gtk_image_menu_item_new_with_label(GCS(MSG_OpenText));
	    g_object_set(G_OBJECT(item), "image", 
		    gtk_image_new_from_stock(GTK_STOCK_OPEN,
			GTK_ICON_SIZE_MENU), NULL);
#else
	    item = gtk_menu_item_new_with_label(GCS(MSG_OpenText));
#endif
	    g_signal_connect_swapped(item, "activate",
		    G_CALLBACK(gtk_ext_editor_emit_load), editor);
	    gtk_widget_show(item);
	    gtk_menu_shell_append(shell, item);
	}
	
	if (hasSave) {
#if GTK_MAJOR_VERSION == 2
	    item = gtk_image_menu_item_new_with_label(GCS(MSG_SaveText));
	    g_object_set(G_OBJECT(item), "image", 
		    gtk_image_new_from_stock(GTK_STOCK_SAVE,
			GTK_ICON_SIZE_MENU), NULL);
#else
	    item = gtk_menu_item_new_with_label(GCS(MSG_SaveText));
#endif
	    g_signal_connect_swapped(item, "activate", 
		    G_CALLBACK(gtk_ext_editor_emit_save), editor);
	    gtk_widget_show(item);
	    gtk_menu_shell_append(shell, item);
	}

	if (hasExternalEditor) {
	    item=gtk_menu_item_new_with_label(GCS(MSG_MenuItem_ExternalEditor));
	    g_signal_connect_swapped(item, "activate", 
		    G_CALLBACK(gtk_ext_editor_emit_external_editor), editor);
	    gtk_widget_show(item);
	    gtk_menu_shell_append(shell, item);
	}
    }
}

static void 
gtk_ext_editor_init (GtkExtEditor *editor)
{
    g_signal_connect(editor, "populate-popup", 
	    G_CALLBACK(gtk_ext_editor_on_populate_popup), 0);
}

gint
gtk_ext_editor_get_top (GtkExtEditor *editor)
{
  GdkRectangle rect;

  g_return_val_if_fail (GTK_IS_EXT_EDITOR (editor), 0);

  gtk_text_view_get_visible_rect (GTK_TEXT_VIEW(editor), &rect);

  return rect.y;
}

static gboolean _gtk_ext_editor_idle_set_top (GtkExtEditor *editor)
{
  GtkTextIter iter;
  gtk_text_view_get_line_at_y(GTK_TEXT_VIEW(editor), &iter, editor->top, 0);
  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(editor), &iter, 0.0,
	  TRUE, 0.0, 0.0);
  return FALSE;
}

void gtk_ext_editor_set_top (GtkExtEditor *editor, gint top)
{
  g_return_if_fail (GTK_IS_EXT_EDITOR (editor));

  /* 
  ** we need to set the top in an idle loop as scroll to iter does only
  ** work after the text editor is fully setup
  */
  editor->top = top;
  g_idle_add((GSourceFunc)_gtk_ext_editor_idle_set_top, editor);
}

void 
gtk_ext_editor_get_cursor (GtkExtEditor *editor, guint *line, guint *col)
{
  GtkTextBuffer *buffer;
  GtkTextMark *cursor;
  GtkTextIter iter;

  g_return_if_fail (GTK_IS_EXT_EDITOR (editor));

  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (editor));
  cursor = gtk_text_buffer_get_insert (buffer);
  gtk_text_buffer_get_iter_at_mark (buffer, &iter, cursor);

  if (line)
    {
      *line = gtk_text_iter_get_line(&iter) + 1;
    }
  if (col)
    {
      *col = gtk_text_iter_get_line_offset(&iter) + 1;
    }
}

static gboolean
gtk_ext_editor_ensure_cursor_visible (gpointer p)
{
  GtkExtEditor *editor;
  GtkTextBuffer *buffer;

  editor = GTK_EXT_EDITOR(p);
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (editor));

#if 1
  gtk_text_view_scroll_mark_onscreen (GTK_TEXT_VIEW(editor), 
	  gtk_text_buffer_get_insert (buffer));
#else
  gtk_text_view_scroll_to_mark (GTK_TEXT_VIEW (editor), 
	  gtk_text_buffer_get_insert (buffer), 0.1, FALSE, 0.5, 0.5);
#endif

  return FALSE;         /* don't call again */
}

void 
gtk_ext_editor_set_cursor (GtkExtEditor *editor, guint line, guint col)
{
  GtkTextBuffer *buffer;
  GtkTextIter iter;
  guint lineCount, charCount;

  g_return_if_fail (GTK_IS_EXT_EDITOR (editor));

  if (line > 0) {
      line--;
  }
  if (col > 0) {
      col--;
  }

  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (editor));
  lineCount = gtk_text_buffer_get_line_count(buffer);
  if (line > lineCount) {
      line = lineCount;
  }

  gtk_text_buffer_get_iter_at_line(buffer, &iter, line);
  charCount = gtk_text_iter_get_chars_in_line(&iter);
  if (col > charCount) {
      col = charCount;
  }
  gtk_text_iter_set_line_offset(&iter, col);

  gtk_text_buffer_place_cursor (buffer, &iter);

#if 1
  gtk_text_view_scroll_mark_onscreen (GTK_TEXT_VIEW(editor), 
	  gtk_text_buffer_get_insert (buffer));
#else
  gtk_text_view_scroll_to_mark (GTK_TEXT_VIEW (editor), 
	  gtk_text_buffer_get_insert (buffer), 0.1, FALSE, 0.5, 0.5);
#endif
  g_idle_add ( gtk_ext_editor_ensure_cursor_visible, editor);
}

