/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkextentry.c
**
** Extending GtkEntry for allowing to specify a GtkTreeView that to be
** controlled by the entry.
**
** Creation data: 12.05.2004
** Last modified: 05.09.2012
*/

#include "gtkextentry.h"
#include <gdk/gdkkeysyms.h>
#include <glib/gi18n.h>

#define P_(x) N_(x)

enum {
  PROP_0,
  PROP_ATTACHED_TREE_VIEW
};

static void gtk_ext_entry_class_init     (GtkExtEntryClass   *klass);
static void gtk_ext_entry_init           (GtkExtEntry        *entry);
#if GTK_MAJOR_VERSION == 2
static void gtk_ext_entry_destroy        (GtkObject        *object);
#else
static void gtk_ext_entry_destroy        (GtkWidget        *widget);
#endif
static void gtk_ext_entry_set_property   (GObject         *object,
                                           guint            prop_id,
                                           const GValue    *value,
                                           GParamSpec      *pspec);
static void gtk_ext_entry_get_property   (GObject         *object,
                                           guint            prop_id,
                                           GValue          *value,
                                           GParamSpec      *pspec);

static GtkEntryClass *parent_class = NULL;

GType
gtk_ext_entry_get_type (void)
{
  static GType ext_entry_type = 0;

  if (!ext_entry_type)
    {
      static const GTypeInfo ext_entry_info =
      {
	sizeof (GtkExtEntryClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_ext_entry_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkExtEntry),
	16,		/* n_preallocs */
	(GInstanceInitFunc) gtk_ext_entry_init,
      };

      ext_entry_type = g_type_register_static (GTK_TYPE_ENTRY, "GtkExtEntry",
					        &ext_entry_info, 0);
    }

  return ext_entry_type;
}

static void
gtk_ext_entry_class_init (GtkExtEntryClass *klass)
{
  GObjectClass *gobject_class;
#if GTK_MAJOR_VERSION == 2
  GtkObjectClass *object_class;
#else
  GtkWidgetClass *widget_class;
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = gtk_ext_entry_set_property;
  gobject_class->get_property = gtk_ext_entry_get_property;

#if GTK_MAJOR_VERSION == 2
  object_class = GTK_OBJECT_CLASS (klass);
  object_class->destroy = gtk_ext_entry_destroy;
#else
  widget_class = GTK_WIDGET_CLASS (klass);
  widget_class->destroy = gtk_ext_entry_destroy;
#endif

  g_object_class_install_property (gobject_class,
                                   PROP_ATTACHED_TREE_VIEW,
                                   g_param_spec_object ("attached-tree-view",
                                                     P_("Attached tree view"),
                                                     P_("Tree view which is controlled by the entry"),
                                                        GTK_TYPE_TREE_VIEW,
                                                        G_PARAM_READWRITE));
}

static gboolean 
gtk_ext_entry_on_key_press_event (GtkExtEntry *entry,
                                  GdkEventKey *event,
                                  gpointer     data)
{
  if (entry->tree && (event->state & GDK_MODIFIER_MASK) == 0)
    {
      gint dir;

      switch (event->keyval) 
        {
	  case GDK_KEY_Up:   dir = -1; break;
	  case GDK_KEY_Down: dir =  1; break;
	  default:           dir =  0; break;
	}

      if (dir != 0)
	{
	  GtkTreePath *path = NULL;

	  gtk_tree_view_get_cursor (entry->tree, &path, NULL);
	  if (path)
	    {
	      if (dir > 0)
		{
		  GtkTreeModel *model = gtk_tree_view_get_model (entry->tree);
		  GtkTreeIter iter;

		  if (gtk_tree_model_get_iter (model, &iter, path) &&
			  gtk_tree_model_iter_next (model, &iter)) 
		    {
		      gtk_tree_path_next(path);
	              gtk_tree_view_set_cursor (entry->tree, path, NULL, FALSE);
		    }
	        } 
	      else 
	        {
		  if (gtk_tree_path_prev(path))
		    {
	              gtk_tree_view_set_cursor (entry->tree, path, NULL, FALSE);
		    }
	        }
	      gtk_tree_path_free (path);
	    }
	  else
	    {
	      path = gtk_tree_path_new_from_indices (0, -1);
	      gtk_tree_view_set_cursor (entry->tree, path, NULL, FALSE);
	      gtk_tree_path_free (path);
	    }
	  return TRUE;
	}
    }
    return FALSE;
}

static void
gtk_ext_entry_init (GtkExtEntry *entry)
{
  entry->tree = NULL;

  g_signal_connect (entry, "key-press-event", 
	  G_CALLBACK(gtk_ext_entry_on_key_press_event), NULL);
}

static void
gtk_ext_entry_remove_attached_tree_view (GtkExtEntry *entry)
{
  if (entry->tree)
    {
      g_object_unref (G_OBJECT (entry->tree));
      entry->tree = NULL;
    }
}

#if GTK_MAJOR_VERSION == 2
static void
gtk_ext_entry_destroy (GtkObject *object)
{
  GtkExtEntry *entry = GTK_EXT_ENTRY (object);
  
  gtk_ext_entry_remove_attached_tree_view (entry);

  (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}
#else
static void
gtk_ext_entry_destroy (GtkWidget *widget)
{
  GtkExtEntry *entry = GTK_EXT_ENTRY (widget);
  
  gtk_ext_entry_remove_attached_tree_view (entry);

  (* GTK_WIDGET_CLASS (parent_class)->destroy) (widget);
}
#endif

static void
gtk_ext_entry_set_property (GObject         *object,
                             guint            prop_id,
                             const GValue    *value,
                             GParamSpec      *pspec)
{
  GtkExtEntry *entry = GTK_EXT_ENTRY (object);

  switch (prop_id)
    {
    case PROP_ATTACHED_TREE_VIEW:
      gtk_ext_entry_set_attached_tree_view (entry, g_value_get_object (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
gtk_ext_entry_get_property (GObject         *object,
                             guint            prop_id,
                             GValue          *value,
                             GParamSpec      *pspec)
{
  GtkExtEntry *entry = GTK_EXT_ENTRY (object);

  switch (prop_id)
    {
    case PROP_ATTACHED_TREE_VIEW:
      g_value_set_object (value, entry->tree);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

GtkWidget*
gtk_ext_entry_new (void)
{
  return g_object_new (GTK_TYPE_EXT_ENTRY, NULL);
}

void           
gtk_ext_entry_set_attached_tree_view (GtkExtEntry     *entry,
	 			      GtkTreeView     *tree)
{
  g_return_if_fail (GTK_IS_EXT_ENTRY (entry));
  g_return_if_fail (tree == NULL || GTK_IS_TREE_VIEW (tree));

  gtk_ext_entry_remove_attached_tree_view (entry);

  if (tree)
    {
      g_object_ref (G_OBJECT (tree));
      entry->tree = tree;
    }
  g_object_notify (G_OBJECT (entry), "attached-tree-view");
}

GtkTreeView *   
gtk_ext_entry_get_attached_tree_view (GtkExtEntry   *entry)
{
  g_return_val_if_fail (GTK_IS_EXT_ENTRY (entry), NULL);

  return entry->tree;
}

