/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkextimage.c
**
** Allow to scale a pixbuf in a GtkImage
**
** Creation data: 30.04.2004
** Last modified: 05.09.2012
*/

#include "gtkextimage.h"
#include <glib/gi18n.h>

#define P_(x) N_(x)

enum {
  PROP_0,
  PROP_PIXBUF,
  PROP_FIXED,
  PROP_SCALED,
  PROP_KEEP_ASPECT,
};

static void gtk_ext_image_class_init   (GtkExtImageClass    *klass);
static void gtk_ext_image_init         (GtkExtImage         *image);
#if GTK_MAJOR_VERSION == 2
static void gtk_ext_image_destroy      (GtkObject        *object);
#else
static void gtk_ext_image_destroy      (GtkWidget        *widget);
#endif
static void gtk_ext_image_set_property (GObject         *object,
                                        guint            prop_id,
                                        const GValue    *value,
                                        GParamSpec      *pspec);
static void gtk_ext_image_get_property (GObject         *object,
                                        guint            prop_id,
                                        GValue          *value,
                                        GParamSpec      *pspec);

static GtkImageClass *parent_class = NULL;

GType
gtk_ext_image_get_type (void)
{
  static GType ext_image_type = 0;

  if (!ext_image_type)
    {
      static const GTypeInfo ext_image_info =
      {
	sizeof (GtkExtImageClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_ext_image_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkExtImage),
	0,		/* n_preallocs */
	(GInstanceInitFunc) gtk_ext_image_init,
	NULL,		/* value_table */
      };

      ext_image_type = g_type_register_static (GTK_TYPE_IMAGE, 
	      "GtkExtImage", &ext_image_info, (GTypeFlags)0);
    }

  return ext_image_type;
}

static void
gtk_ext_image_class_init (GtkExtImageClass *klass)
{
  GObjectClass *gobject_class;
#if GTK_MAJOR_VERSION == 2
  GtkObjectClass *object_class;
#else
  GtkWidgetClass *widget_class;
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = gtk_ext_image_set_property;
  gobject_class->get_property = gtk_ext_image_get_property;

#if GTK_MAJOR_VERSION == 2
  object_class = GTK_OBJECT_CLASS (klass);
  object_class->destroy = gtk_ext_image_destroy;
#else
  widget_class = GTK_WIDGET_CLASS (klass);
  widget_class->destroy = gtk_ext_image_destroy;
#endif

  g_object_class_install_property (gobject_class,
                                   PROP_PIXBUF,
                                   g_param_spec_object ("pixbuf",
                                                        P_("Pixbuf"),
                                                        P_("A GdkPixbuf to display"),
                                                        GDK_TYPE_PIXBUF,
                                                        G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
                                   PROP_FIXED,
                                   g_param_spec_boolean ("fixed",
						      P_("Fixed"),
						      P_("Whether the widget has a fixed size or is resiable"),
						      FALSE,
						      G_PARAM_READWRITE));
  
  g_object_class_install_property (gobject_class,
                                   PROP_SCALED,
                                   g_param_spec_boolean ("scaled",
						      P_("Scaled"),
						      P_("Whether the pixbuf is scaled to the size of the widget"),
						      FALSE,
						      G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
                                   PROP_KEEP_ASPECT,
                                   g_param_spec_boolean ("keep-aspect",
						      P_("Keep-aspect"),
						      P_("Whether the apect ratio is preserved when scaling"),
						      FALSE,
						      G_PARAM_READWRITE));
  
}

static void
gtk_ext_image_size_allocate (GtkExtImage *image, 
	const GtkAllocation *allocation)
{
  gtk_widget_set_allocation (GTK_WIDGET (image), allocation);

  if (image->scaled && image->pixbuf)
    {
      GtkMisc *misc = GTK_MISC(image);
      gint iw = gdk_pixbuf_get_width(image->pixbuf);
      gint ih = gdk_pixbuf_get_height(image->pixbuf);
      GdkPixbuf *pixbuf = gtk_image_get_pixbuf(GTK_IMAGE(image));
      gint xpad, ypad, width, height;

      gtk_misc_get_padding(misc, &xpad, &ypad);
      width = allocation->width - 2*xpad;
      height = allocation->height - 2*ypad;

      if (image->keep_aspect && iw > 0 && ih > 0)
        {
	  if (ih * width / iw > height)
	    {
	      width = iw * height / ih;
	    }
	  else
	    {
	      height = ih * width / iw;
	    }
        }

      if (pixbuf == NULL || width != gdk_pixbuf_get_width(pixbuf) ||
	   height != gdk_pixbuf_get_height(pixbuf))
	{
          pixbuf = gdk_pixbuf_scale_simple (image->pixbuf,
	     	  width, height, GDK_INTERP_BILINEAR);
	  gtk_image_set_from_pixbuf(GTK_IMAGE(image), pixbuf);
	  g_object_unref (G_OBJECT(pixbuf));
	}
    }
}

static void
gtk_ext_image_set_size_request (GtkExtImage *image)
{
  if (image->fixed) 
    {
      gtk_widget_set_size_request (GTK_WIDGET(image), -1, -1);
    }
  else
    {
      gtk_widget_set_size_request (GTK_WIDGET(image), 50, 30);
    }
}

static void 
gtk_ext_image_init (GtkExtImage *image)
{
    image->pixbuf = NULL;
    image->fixed = FALSE;
    image->scaled = FALSE;

    gtk_ext_image_set_size_request (image);

    g_signal_connect (image, "size-allocate", 
	    G_CALLBACK(gtk_ext_image_size_allocate), NULL);
}

static void
gtk_ext_image_cleanup (GtkExtImage *image)
{
  if (image->pixbuf)
    {
      g_object_unref(G_OBJECT(image->pixbuf));
      image->pixbuf = NULL;
    }
}

#if GTK_MAJOR_VERSION == 2
static void
gtk_ext_image_destroy (GtkObject *object)
{
  gtk_ext_image_cleanup (GTK_EXT_IMAGE (object));
  (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}
#else
static void
gtk_ext_image_destroy (GtkWidget *widget)
{
  gtk_ext_image_cleanup (GTK_EXT_IMAGE (widget));
  (* GTK_WIDGET_CLASS (parent_class)->destroy) (widget);
}
#endif

static void
gtk_ext_image_set_property (GObject         *object,
                            guint            prop_id,
                            const GValue    *value,
                            GParamSpec      *pspec)
{
  GtkExtImage *image = GTK_EXT_IMAGE (object);

  switch (prop_id)
    {
    case PROP_PIXBUF:
      gtk_ext_image_set_from_pixbuf (image, g_value_get_object (value));
      break;
    case PROP_FIXED:
      gtk_ext_image_set_fixed (image, g_value_get_boolean (value));
      break;
    case PROP_SCALED:
      gtk_ext_image_set_scaled (image, g_value_get_boolean (value));
      break;
    case PROP_KEEP_ASPECT:
      gtk_ext_image_set_keep_aspect (image, g_value_get_boolean (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
gtk_ext_image_get_property (GObject         *object,
                            guint            prop_id,
                            GValue          *value,
                            GParamSpec      *pspec)
{
  GtkExtImage *image = GTK_EXT_IMAGE (object);

  switch (prop_id)
    {
    case PROP_PIXBUF:
      g_value_set_object (value, image->pixbuf);
      break;
    case PROP_FIXED:
      g_value_set_boolean (value, image->fixed);
      break;
    case PROP_SCALED:
      g_value_set_boolean (value, image->scaled);
      break;
    case PROP_KEEP_ASPECT:
      g_value_set_boolean (value, image->keep_aspect);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

GtkWidget*
gtk_ext_image_new (void)
{
  return g_object_new (GTK_TYPE_EXT_IMAGE, NULL);
}

void       
gtk_ext_image_set_from_pixbuf (GtkExtImage   *image,
			       GdkPixbuf     *pixbuf)
{
  g_return_if_fail (GTK_IS_EXT_IMAGE(image));
  g_return_if_fail (pixbuf == NULL ||
                    GDK_IS_PIXBUF (pixbuf));

  g_object_freeze_notify (G_OBJECT (image));

  if (pixbuf)
    {
      g_object_ref (G_OBJECT(pixbuf));
    }
  if (image->pixbuf)
    {
      g_object_unref (G_OBJECT(image->pixbuf));
    }
  image->pixbuf = pixbuf;
  gtk_image_set_from_pixbuf(GTK_IMAGE(image), image->pixbuf);

  g_object_notify (G_OBJECT (image), "pixbuf");
  g_object_thaw_notify (G_OBJECT (image));
}

GdkPixbuf *
gtk_ext_image_get_pixbuf (GtkExtImage   *image)
{
  g_return_val_if_fail (GTK_IS_EXT_IMAGE(image), NULL);

  return image->pixbuf;
}

void       
gtk_ext_image_set_fixed (GtkExtImage   *image,
	                 gboolean       fixed)
{
  g_return_if_fail (GTK_IS_EXT_IMAGE(image));

  fixed = fixed? TRUE : FALSE;
  if (fixed != image->fixed) 
    {
      g_object_freeze_notify (G_OBJECT (image));

      image->fixed = fixed;
      gtk_ext_image_set_size_request (image);

      g_object_notify (G_OBJECT (image), "fixed");
      g_object_thaw_notify (G_OBJECT (image));
    }
}

gboolean   
gtk_ext_image_get_fixed (GtkExtImage   *image)
{
  g_return_val_if_fail (GTK_IS_EXT_IMAGE(image), FALSE);

  return image->fixed;
}

void       
gtk_ext_image_set_scaled (GtkExtImage   *image,
	                  gboolean       scaled)
{
  g_return_if_fail (GTK_IS_EXT_IMAGE(image));

  scaled = scaled? TRUE : FALSE;
  if (scaled != image->scaled) 
    {
      g_object_freeze_notify (G_OBJECT (image));

      image->scaled = scaled;
      gtk_ext_image_set_size_request (image);

      g_object_notify (G_OBJECT (image), "scaled");
      g_object_thaw_notify (G_OBJECT (image));
    }
}

gboolean   
gtk_ext_image_get_scaled (GtkExtImage   *image)
{
  g_return_val_if_fail (GTK_IS_EXT_IMAGE(image), FALSE);

  return image->scaled;
}

void       
gtk_ext_image_set_keep_aspect (GtkExtImage   *image,
	                       gboolean       keep_aspect)
{
  g_return_if_fail (GTK_IS_EXT_IMAGE(image));

  keep_aspect = keep_aspect? TRUE : FALSE;
  if (keep_aspect != image->keep_aspect) 
    {
      g_object_freeze_notify (G_OBJECT (image));

      image->keep_aspect = keep_aspect;
      gtk_ext_image_set_size_request (image);

      g_object_notify (G_OBJECT (image), "keep-aspect");
      g_object_thaw_notify (G_OBJECT (image));
    }
}

gboolean   
gtk_ext_image_get_keep_aspect (GtkExtImage   *image)
{
  g_return_val_if_fail (GTK_IS_EXT_IMAGE(image), FALSE);

  return image->keep_aspect;
}

