/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkgroup.c
**
** Grouping of wigdets similar to MUI's Group class
**
** Creation data: 10.03.2004
** Last modified: 23.02.2013
*/

#include "gtkgroup.h"

#define P_(String) (String)

enum {
  PROP_0,
  PROP_CHILD,
  PROP_CHILD_HORIZ_WEIGHT,
  PROP_CHILD_VERT_WEIGHT,
  PROP_CHILD_WEIGHT,
  PROP_HORIZ,
  PROP_COLUMNS,
  PROP_ROWS,
  PROP_HORIZ_SPACING,
  PROP_VERT_SPACING,
  PROP_SPACING,
  PROP_SAME_WIDTH,
  PROP_SAME_HEIGHT,
  PROP_SAME_SIZE
};

enum {
  CHILD_PROP_0,
  CHILD_PROP_HORIZ_WEIGHT,
  CHILD_PROP_VERT_WEIGHT,
  CHILD_PROP_WEIGHT
};

static void gtk_group_class_init(GtkGroupClass    *klass);
static void gtk_group_init  	(GtkGroup         *group);
static void gtk_group_finalize 	(GObject	  *object);
static void gtk_group_set_property(GObject        *object,
				  guint            prop_id,
				  const GValue    *value,
				  GParamSpec      *pspec);
static void gtk_group_get_property(GObject        *object,
				  guint            prop_id,
				  GValue          *value,
				  GParamSpec      *pspec);
static void gtk_group_add 	(GtkContainer     *container,
			          GtkWidget       *widget);
static void gtk_group_remove    (GtkContainer     *container,
			          GtkWidget       *widget);
static void gtk_group_forall    (GtkContainer     *container,
				 gboolean	   include_internals,
			         GtkCallback       callback,
			         gpointer          callback_data);
static void gtk_group_set_child_property(GtkContainer   *container,
					GtkWidget       *child,
					guint            property_id,
					const GValue    *value,
					GParamSpec      *pspec);
static void gtk_group_get_child_property(GtkContainer   *container,
					GtkWidget       *child,
					guint            property_id,
					GValue          *value,
					GParamSpec      *pspec);
static GType gtk_group_child_type (GtkContainer   *container);
     
#if GTK_MAJOR_VERSION == 2
static void gtk_group_size_request (GtkWidget      *widget,
				    GtkRequisition *requisition);
#else
static void gtk_group_get_preferred_size (GtkGroup       *group,
                                          GtkOrientation  orientation,
                                          gint           *minimal_size,
                                          gint           *natural_size);
static void gtk_group_get_preferred_width (GtkWidget *widget,
                                           gint      *minimal_width,
                                           gint      *natural_width);
static void gtk_group_get_preferred_height (GtkWidget *widget,
                                            gint      *minimal_height,
                                            gint      *natural_height);
#endif
static void gtk_group_size_allocate (GtkWidget      *widget,
				     GtkAllocation  *allocation);
static void gtk_group_set_window_geometry_hints (GtkWidget *widget);
/* static void gtk_group_check_window_geometry (GtkWidget *widget); */

static GtkContainerClass *parent_class = NULL;


GType
gtk_group_get_type (void)
{
  static GType group_type = 0;

  if (!group_type)
    {
      static const GTypeInfo group_info =
      {
	sizeof (GtkGroupClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_group_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkGroup),
	0,		/* n_preallocs */
	(GInstanceInitFunc) gtk_group_init,
	NULL,		/* value_table */
      };

      group_type = g_type_register_static (GTK_TYPE_CONTAINER, "GtkGroup", 
					 &group_info, 0);
    }

  return group_type;
}

static void
gtk_group_class_init (GtkGroupClass *class)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (class);
  GtkContainerClass *container_class = GTK_CONTAINER_CLASS (class);
  GtkWidgetClass *widget_class = (GtkWidgetClass*) class;

  parent_class = g_type_class_peek_parent (class);

  gobject_class->finalize = gtk_group_finalize;
  gobject_class->set_property = gtk_group_set_property;
  gobject_class->get_property = gtk_group_get_property;
   
  container_class->add = gtk_group_add;
  container_class->remove = gtk_group_remove;
  container_class->forall = gtk_group_forall;
  container_class->child_type = gtk_group_child_type;
  container_class->set_child_property = gtk_group_set_child_property;
  container_class->get_child_property = gtk_group_get_child_property;
#if GTK_MAJOR_VERSION == 2
  widget_class->size_request = gtk_group_size_request;
#else
  widget_class->get_preferred_width = gtk_group_get_preferred_width;
  widget_class->get_preferred_height = gtk_group_get_preferred_height;
#endif
  widget_class->size_allocate = gtk_group_size_allocate;

  g_object_class_install_property (gobject_class,
                                   PROP_CHILD,
                                   g_param_spec_object ("child",
                                                     P_("Child"),
                                                     P_("A child widget to be added"),
						     GTK_TYPE_WIDGET,
						     G_PARAM_WRITABLE));

  g_object_class_install_property (gobject_class,
				   PROP_CHILD_HORIZ_WEIGHT,
				   g_param_spec_uint ("horiz_weight",
					           P_("Horiz weight"), 
						   P_("Horizontal weight of last child added"),
						   0,
						   G_MAXUINT,
						   100,
						   G_PARAM_WRITABLE));

  g_object_class_install_property (gobject_class,
				   PROP_CHILD_VERT_WEIGHT,
				   g_param_spec_uint ("vert_weight",
					           P_("Vert weight"), 
						   P_("Vertical weight of last child added"),
						   0,
						   G_MAXUINT,
						   100,
						   G_PARAM_WRITABLE));

  g_object_class_install_property (gobject_class,
				   PROP_CHILD_WEIGHT,
				   g_param_spec_uint ("weight",
					           P_("Weight"), 
						   P_("Shorthand for horizontal and vertical weight of last child added"),
						   0,
						   G_MAXUINT,
						   100,
						   G_PARAM_WRITABLE));

  g_object_class_install_property (gobject_class,
                                   PROP_HORIZ,
                                   g_param_spec_boolean ("horiz",
						      P_("Horiz"),
						      P_("Whether objects are layed out horizontally or vertically"),
						      FALSE,
						      G_PARAM_READWRITE));
  
  g_object_class_install_property (gobject_class,
                                   PROP_COLUMNS,
                                   g_param_spec_uint ("columns",
						   P_("Columns"),
						   P_("Number of columns in a vertical group"),
						   1,
						   G_MAXUINT,
						   1,
						   G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
                                   PROP_ROWS,
                                   g_param_spec_uint ("rows",
						   P_("Rows"),
						   P_("Number of rows in a horizontal group"),
						   1,
						   G_MAXUINT,
						   1,
						   G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
                                   PROP_HORIZ_SPACING,
                                   g_param_spec_int ("horiz_spacing",
						  P_("Horiz spacing"),
						  P_("Horizontal spacing between elements"),
						  G_MININT,
						  G_MAXINT,
						  GTK_GROUP_DEFAULT_HORIZ_SPACING,
						  G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
                                   PROP_VERT_SPACING,
                                   g_param_spec_int ("vert_spacing",
						  P_("Vert spacing"),
						  P_("Vertical spacing between elements"),
						  G_MININT,
					          G_MAXINT,
						  GTK_GROUP_DEFAULT_VERT_SPACING,
						  G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
                                   PROP_SPACING,
                                   g_param_spec_int ("spacing",
						  P_("Spacing"),
						  P_("Shorthand for horizontal and vertical spacing between elements"),
						  G_MININT,
					          G_MAXINT,
						  GTK_GROUP_DEFAULT_VERT_SPACING,
						  G_PARAM_WRITABLE));

  g_object_class_install_property (gobject_class,
                                   PROP_SAME_WIDTH,
                                   g_param_spec_boolean ("same_width",
						      P_("Same width"),
						      P_("Whether all objects should have the same width"),
						      FALSE,
						      G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
                                   PROP_SAME_HEIGHT,
                                   g_param_spec_boolean ("same_height",
						      P_("Same height"),
						      P_("Whether all objects should have the same height"),
						      FALSE,
						      G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
                                   PROP_SAME_SIZE,
                                   g_param_spec_boolean ("same_size",
						      P_("Same size"),
						      P_("Shorthand for same width and same height"),
						      FALSE,
						      G_PARAM_WRITABLE));

  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_HORIZ_WEIGHT,
					      g_param_spec_uint ("horiz_weight", 
							      P_("Horiz weight"), 
							      P_("Horizontal weight for this child when laying out"),
							      0,
							      G_MAXUINT,
							      100,
							      G_PARAM_READWRITE));

  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_VERT_WEIGHT,
					      g_param_spec_uint ("vert_weight", 
							      P_("Vert weight"), 
							      P_("Vertical weight for this child when laying out"),
							      0,
							      G_MAXUINT,
							      100,
							      G_PARAM_READWRITE));

  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_WEIGHT,
					      g_param_spec_uint ("weight", 
							      P_("Weight"), 
							      P_("Shorthand for horiz weight and vert weight"),
							      0,
							      G_MAXUINT,
							      100,
							      G_PARAM_READWRITE));
}

static void
gtk_group_init (GtkGroup *group)
{
  static const int num1 = 1;

  gtk_widget_set_has_window (GTK_WIDGET (group), FALSE);
  gtk_widget_set_redraw_on_allocate (GTK_WIDGET (group), FALSE);
  
  group->children = NULL;
  group->row_cols = num1;
  group->horiz_spacing = GTK_GROUP_DEFAULT_HORIZ_SPACING;
  group->vert_spacing = GTK_GROUP_DEFAULT_VERT_SPACING;
  group->horiz = FALSE;
  group->same_width = FALSE;
  group->same_height = FALSE;

  group->min_dim1 = g_new(gint, num1);
#if GTK_MAJOR_VERSION > 2
  group->nat_dim1 = g_new(gint, num1);
#endif
  group->weight1  = g_new(guint, num1);

  g_signal_connect (group, "show", 
          G_CALLBACK (gtk_group_set_window_geometry_hints), NULL);
}

static void
gtk_group_finalize (GObject *object)
{
  GtkGroup *group = GTK_GROUP (object);

  g_free(group->weight1);
  group->weight1 = NULL;
#if GTK_MAJOR_VERSION > 2
  g_free(group->nat_dim1);
  group->nat_dim1 = NULL;
#endif
  g_free(group->min_dim1);
  group->min_dim1 = NULL;
}

static void 
gtk_group_set_property (GObject         *object,
		        guint            prop_id,
		        const GValue    *value,
		        GParamSpec      *pspec)
{
  GtkGroup *group = GTK_GROUP (object);
  GList *last;
  GtkGroupChild *child_info;
  GtkWidget *child;

  switch (prop_id)
    {
    case PROP_CHILD:
	gtk_group_add_child(group, GTK_WIDGET(g_value_get_object (value)), 100, 100);
	break;
    
    case PROP_CHILD_HORIZ_WEIGHT:
	last = g_list_last(group->children);
	child_info = last? last->data : NULL;
	child = child_info? child_info->widget : NULL;
	gtk_group_set_child_property (GTK_CONTAINER (object),
				      child,
				      CHILD_PROP_HORIZ_WEIGHT,
				      value,
				      pspec);
	break;

    case PROP_CHILD_VERT_WEIGHT:
	last = g_list_last(group->children);
	child_info = last? last->data : NULL;
	child = child_info? child_info->widget : NULL;
	gtk_group_set_child_property (GTK_CONTAINER (object),
				      child,
				      CHILD_PROP_VERT_WEIGHT,
				      value,
				      pspec);
	break;

    case PROP_CHILD_WEIGHT:
	last = g_list_last(group->children);
	child_info = last? last->data : NULL;
	child = child_info? child_info->widget : NULL;
	gtk_group_set_child_property (GTK_CONTAINER (object),
				      child,
				      CHILD_PROP_WEIGHT,
				      value,
				      pspec);
	break;

    case PROP_HORIZ:
	gtk_group_set_horiz(group, g_value_get_boolean(value));
	break;

    case PROP_COLUMNS:
	gtk_group_set_columns(group, g_value_get_uint(value));
	break;

    case PROP_ROWS:
	gtk_group_set_rows(group, g_value_get_uint(value));
	break;

    case PROP_HORIZ_SPACING:
	gtk_group_set_horiz_spacing(group, g_value_get_int(value));
	break;

    case PROP_VERT_SPACING:
	gtk_group_set_vert_spacing(group, g_value_get_int(value));
	break;

    case PROP_SPACING:
	gtk_group_set_spacing(group, g_value_get_int(value), g_value_get_int(value));
	break;

    case PROP_SAME_WIDTH:
	gtk_group_set_same_width(group, g_value_get_boolean(value));
	break;

    case PROP_SAME_HEIGHT:
	gtk_group_set_same_height(group, g_value_get_boolean(value));
	break;

    case PROP_SAME_SIZE:
	gtk_group_set_same_size(group, g_value_get_boolean(value), g_value_get_boolean(value));
	break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void gtk_group_get_property (GObject         *object,
				    guint            prop_id,
				    GValue          *value,
				    GParamSpec      *pspec)
{
  GtkGroup *group;

  group = GTK_GROUP (object);

  switch (prop_id)
    {
    case PROP_HORIZ:
	g_value_set_boolean (value, group->horiz);
	break;

    case PROP_COLUMNS:
    case PROP_ROWS:
	g_value_set_uint(value, group->row_cols);
	break;

    case PROP_HORIZ_SPACING:
	g_value_set_int(value, group->horiz_spacing);
	break;

    case PROP_VERT_SPACING:
	g_value_set_int(value, group->vert_spacing);
	break;

    case PROP_SAME_WIDTH:
	g_value_set_boolean(value, group->same_width);
	break;

    case PROP_SAME_HEIGHT:
	g_value_set_boolean(value, group->same_height);
	break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static GType
gtk_group_child_type (GtkContainer   *container)
{
  return GTK_TYPE_WIDGET;
}

static void
gtk_group_set_child_property (GtkContainer    *container,
		  	      GtkWidget       *child,
			      guint            property_id,
			      const GValue    *value,
			      GParamSpec      *pspec)
{
  guint horiz_weight = 100;
  guint vert_weight = 100;

  gtk_group_query_child_packing (GTK_GROUP (container),
				 child,
				 &horiz_weight,
				 &vert_weight);
  switch (property_id)
    {
    case CHILD_PROP_HORIZ_WEIGHT:
      gtk_group_set_child_packing (GTK_GROUP (container),
				   child,
				   g_value_get_uint (value),
				   vert_weight);
      break;

    case CHILD_PROP_VERT_WEIGHT:
      gtk_group_set_child_packing (GTK_GROUP (container),
				   child,
				   horiz_weight,
				   g_value_get_uint (value));
      break;

    case CHILD_PROP_WEIGHT:
      gtk_group_set_child_packing (GTK_GROUP (container),
				   child,
				   g_value_get_uint (value),
				   g_value_get_uint (value));
      break;

    default:
      GTK_CONTAINER_WARN_INVALID_CHILD_PROPERTY_ID (container, property_id, pspec);
      break;
    }
}

static void
gtk_group_get_child_property (GtkContainer *container,
			      GtkWidget    *child,
			      guint         property_id,
			      GValue       *value,
			      GParamSpec   *pspec)
{
  guint horiz_weight = 100;
  guint vert_weight = 100;

  gtk_group_query_child_packing (GTK_GROUP (container),
				 child,
				 &horiz_weight,
				 &vert_weight);
  switch (property_id)
    {
    case CHILD_PROP_HORIZ_WEIGHT:
      g_value_set_uint (value, horiz_weight);
      break;

    case CHILD_PROP_VERT_WEIGHT:
      g_value_set_uint (value, vert_weight);
      break;

    default:
      GTK_CONTAINER_WARN_INVALID_CHILD_PROPERTY_ID (container, property_id, pspec);
      break;
    }
}

GtkWidget* 
gtk_group_new (gboolean horiz,
	       guint num_row_cols)
{
  GtkGroup *group;

  group = g_object_new (GTK_TYPE_GROUP,
	  "horiz", horiz,
	  "rows", num_row_cols,
	  NULL);

  return GTK_WIDGET (group);
}

void
gtk_group_add_child (GtkGroup    *group,
		     GtkWidget   *child,
		     guint       horiz_weight,
		     guint       vert_weight)
{
  GtkGroupChild *child_info;

  g_return_if_fail (GTK_IS_GROUP (group));
  g_return_if_fail (GTK_IS_WIDGET (child));
  g_return_if_fail (gtk_widget_get_parent (child) == NULL);

  child_info = g_new (GtkGroupChild, 1);
  child_info->widget = child;
  child_info->horiz_weight = horiz_weight;
  child_info->vert_weight = vert_weight;

  group->children = g_list_append (group->children, child_info);

  gtk_widget_freeze_child_notify (child);

  gtk_widget_set_parent (child, GTK_WIDGET (group));
  
  gtk_widget_child_notify (child, "horiz_weight");
  gtk_widget_child_notify (child, "vert_weight");
  gtk_widget_thaw_child_notify (child);

  gtk_group_set_window_geometry_hints (GTK_WIDGET (group));
}

void gtk_group_set_horiz (GtkGroup *group,
			  gboolean  horiz)
{
  g_return_if_fail (GTK_IS_GROUP (group));

  if ((horiz ? TRUE : FALSE) != group->horiz)
    {
      group->horiz = horiz ? TRUE : FALSE;
      g_object_notify (G_OBJECT (group), "horiz");
      gtk_widget_queue_resize (GTK_WIDGET (group));
    }
}

gboolean
gtk_group_get_horiz (GtkGroup *group)
{
  g_return_val_if_fail (GTK_IS_GROUP (group), FALSE);

  return group->horiz;
}

static void
gtk_group_set_row_cols (GtkGroup *group,
		        guint     row_cols)
{
  g_return_if_fail (GTK_IS_GROUP (group));

  if (row_cols < 1)
    {
      row_cols = 1;
    }

  if (row_cols != group->row_cols)
    {
      group->row_cols = row_cols;
      group->min_dim1 = g_renew(gint, group->min_dim1, row_cols);
#if GTK_MAJOR_VERSION > 2
      group->nat_dim1 = g_renew(gint, group->nat_dim1, row_cols);
#endif
      group->weight1  = g_renew(guint, group->weight1, row_cols);
      g_object_freeze_notify (G_OBJECT (group));
      g_object_notify (G_OBJECT (group), "columns");
      g_object_notify (G_OBJECT (group), "rows");
      g_object_thaw_notify (G_OBJECT (group));
      gtk_widget_queue_resize (GTK_WIDGET (group));
    }
}

static guint
gtk_group_get_row_cols (GtkGroup *group)
{
  g_return_val_if_fail (GTK_IS_GROUP (group), 0);

  return group->row_cols;
}

void
gtk_group_set_columns (GtkGroup *group,
		       guint     columns)
{
  gtk_group_set_row_cols(group, columns);
}

guint
gtk_group_get_columns (GtkGroup *group)
{
  return gtk_group_get_row_cols(group);
}

void
gtk_group_set_rows (GtkGroup *group,
		    guint     rows)
{
  gtk_group_set_row_cols(group, rows);
}

guint
gtk_group_get_rows (GtkGroup *group)
{
  return gtk_group_get_row_cols(group);
}

void
gtk_group_set_horiz_spacing (GtkGroup *group,
		             gint      horiz_spacing)
{
  g_return_if_fail (GTK_IS_GROUP (group));

  if (horiz_spacing != group->horiz_spacing)
    {
      group->horiz_spacing = horiz_spacing;
      g_object_notify (G_OBJECT (group), "horiz_spacing");
      gtk_widget_queue_resize (GTK_WIDGET (group));
    }
}

gint
gtk_group_get_horiz_spacing (GtkGroup *group)
{
  g_return_val_if_fail (GTK_IS_GROUP (group), 0);

  return group->horiz_spacing;
}

void
gtk_group_set_vert_spacing (GtkGroup *group,
		            gint      vert_spacing)
{
  g_return_if_fail (GTK_IS_GROUP (group));

  if (vert_spacing != group->vert_spacing)
    {
      group->vert_spacing = vert_spacing;
      g_object_notify (G_OBJECT (group), "vert_spacing");
      gtk_widget_queue_resize (GTK_WIDGET (group));
    }
}

gint
gtk_group_get_vert_spacing (GtkGroup *group)
{
  g_return_val_if_fail (GTK_IS_GROUP (group), 0);

  return group->vert_spacing;
}

void
gtk_group_set_spacing (GtkGroup *group,
	               gint      horiz_spacing,
	               gint      vert_spacing)
{
  g_return_if_fail (GTK_IS_GROUP (group));

  if (horiz_spacing != group->horiz_spacing || vert_spacing != group->vert_spacing)
    {
      g_object_freeze_notify (G_OBJECT (group));
      if (horiz_spacing != group->horiz_spacing)
        {
	  group->horiz_spacing = horiz_spacing;
	  g_object_notify (G_OBJECT (group), "horiz_spacing");
        }
      if (vert_spacing != group->vert_spacing)
        {
          group->vert_spacing = vert_spacing;
          g_object_notify (G_OBJECT (group), "vert_spacing");
	}
      g_object_thaw_notify (G_OBJECT (group));
      gtk_widget_queue_resize (GTK_WIDGET (group));
    }
}

void	   
gtk_group_set_same_width (GtkGroup *group,
			  gboolean  same_width)
{
  g_return_if_fail (GTK_IS_GROUP (group));

  if ((same_width? TRUE : FALSE) != group->same_width)
    {
      group->same_width = same_width? TRUE : FALSE;
      g_object_notify (G_OBJECT (group), "same_width");
      gtk_widget_queue_resize (GTK_WIDGET (group));
    }
}

gboolean 
gtk_group_get_same_width (GtkGroup *group)
{
  g_return_val_if_fail (GTK_IS_GROUP (group), FALSE);

  return group->same_width;
}

void	   
gtk_group_set_same_height (GtkGroup *group,
			   gboolean  same_height)
{
  g_return_if_fail (GTK_IS_GROUP (group));

  if ((same_height? TRUE : FALSE) != group->same_height)
    {
      group->same_height = same_height? TRUE : FALSE;
      g_object_notify (G_OBJECT (group), "same_height");
      gtk_widget_queue_resize (GTK_WIDGET (group));
    }
}

gboolean 
gtk_group_get_same_height (GtkGroup *group)
{
  g_return_val_if_fail (GTK_IS_GROUP (group), FALSE);

  return group->same_height;
}

void
gtk_group_set_same_size (GtkGroup *group,
			 gboolean  same_width,
			 gboolean  same_height)
{
  g_return_if_fail (GTK_IS_GROUP (group));

  same_width = same_width? TRUE : FALSE;
  same_height = same_height? TRUE : FALSE;

  if (group->same_width != same_width || group->same_height != same_height)
    {
      g_object_freeze_notify (G_OBJECT (group));
      if (group->same_width != same_width)
	{
	  group->same_width = same_width;
          g_object_notify (G_OBJECT (group), "same_width");
	}
      if (group->same_height != same_height)
	{
	  group->same_height = same_height;
          g_object_notify (G_OBJECT (group), "same_height");
	}
      g_object_thaw_notify (G_OBJECT (group));
      gtk_widget_queue_resize (GTK_WIDGET (group));
    }
}

guint      
gtk_group_get_total_horiz_weight (GtkGroup     *group)
{
#if GTK_MAJOR_VERSION == 2
  GtkRequisition requisition;
#endif

  g_return_val_if_fail (GTK_IS_GROUP (group), 0);

  /* update total weights */
#if GTK_MAJOR_VERSION == 2
  gtk_widget_size_request (GTK_WIDGET(group), &requisition);
#else
  gtk_group_get_preferred_size (group, GTK_ORIENTATION_HORIZONTAL, NULL, NULL);
#endif
  
  return (group->horiz? group->weight2_total : group->weight1_total);
}

guint      
gtk_group_get_total_vert_weight (GtkGroup     *group)
{
#if GTK_MAJOR_VERSION == 2
  GtkRequisition requisition;
#endif

  g_return_val_if_fail (GTK_IS_GROUP (group), 0);

  /* update total weights */
#if GTK_MAJOR_VERSION == 2
  gtk_widget_size_request (GTK_WIDGET(group), &requisition);
#else
  gtk_group_get_preferred_size (group, GTK_ORIENTATION_VERTICAL, NULL, NULL);
#endif
  
  return (group->horiz? group->weight1_total : group->weight2_total);
}

void
gtk_group_query_child_packing (GtkGroup  *group,
			       GtkWidget *child,
			       guint     *horiz_weight,
			       guint     *vert_weight)
{
  GList *list;
  GtkGroupChild *child_info = NULL;

  g_return_if_fail (GTK_IS_GROUP (group));
  g_return_if_fail (GTK_IS_WIDGET (child));

  list = group->children;
  while (list)
    {
      child_info = list->data;
      if (child_info->widget == child)
	break;

      list = list->next;
    }

  if (list)
    {
      if (horiz_weight)
	*horiz_weight = child_info->horiz_weight;
      if (vert_weight)
	*vert_weight = child_info->vert_weight;
    }
}

void
gtk_group_set_child_packing (GtkGroup  *group,
			     GtkWidget *child,
			     guint      horiz_weight,
			     guint      vert_weight)
{
  GList *list;
  GtkGroupChild *child_info = NULL;

  g_return_if_fail (GTK_IS_GROUP (group));
  g_return_if_fail (GTK_IS_WIDGET (child));

  list = group->children;
  while (list)
    {
      child_info = list->data;
      if (child_info->widget == child)
	break;

      list = list->next;
    }

  gtk_widget_freeze_child_notify (child);
  if (list)
    {
      child_info->horiz_weight = horiz_weight;
      gtk_widget_child_notify (child, "horiz_weight");
      child_info->vert_weight = vert_weight;
      gtk_widget_child_notify (child, "vert_weight");

      if (gtk_widget_get_visible (child) && 
              gtk_widget_get_visible (GTK_WIDGET (group)))
	gtk_widget_queue_resize (child);
    }
  gtk_widget_thaw_child_notify (child);

  gtk_group_set_window_geometry_hints (GTK_WIDGET (group));
}

void       
gtk_group_change_init (GtkGroup *group)
{
  GtkWidget *parent;

  g_return_if_fail (GTK_IS_GROUP (group));

  parent = gtk_widget_get_parent (GTK_WIDGET (group));
  if (GTK_IS_GROUP (parent))
    {
      gtk_group_init (GTK_GROUP (parent));
    }
}

void       
gtk_group_change_exit (GtkGroup *group)
{
  GtkWidget *parent;

  g_return_if_fail (GTK_IS_GROUP (group));

  parent = gtk_widget_get_parent (GTK_WIDGET (group));
  if (GTK_IS_GROUP (parent))
    {
      gtk_group_change_exit (GTK_GROUP (parent));
    }
  else
    {
      /* Set new min/max size. Automatically resizes the window if necessary */
      gtk_group_set_window_geometry_hints (GTK_WIDGET (group));
    }
}

static void
gtk_group_add (GtkContainer *container,
	       GtkWidget    *widget)
{
  gtk_group_add_child (GTK_GROUP (container), widget, 100, 100);
}

static void
gtk_group_remove (GtkContainer *container,
		  GtkWidget    *widget)
{
  GtkGroup *group;
  GtkGroupChild *child;
  GList *children;

  group = GTK_GROUP (container);

  children = group->children;
  while (children)
    {
      child = children->data;

      if (child->widget == widget)
	{
	  gboolean was_visible;

	  was_visible = gtk_widget_get_visible (widget);
	  gtk_widget_unparent (widget);

	  group->children = g_list_remove_link (group->children, children);
	  g_list_free (children);
	  g_free (child);

	  /* queue resize regardless of gtk_widget_get_visible (container),
	   * since that's what is needed by toplevels.
	   */
	  if (was_visible)
	    gtk_widget_queue_resize (GTK_WIDGET (container));

	  break;
	}

      children = children->next;
    }
}

static void
gtk_group_forall (GtkContainer *container,
		  gboolean      include_internals,
		  GtkCallback   callback,
		  gpointer      callback_data)
{
  GtkGroup *group;
  GtkGroupChild *child;
  GList *children;

  g_return_if_fail (callback != NULL);

  group = GTK_GROUP (container);

  children = group->children;
  while (children)
    {
      child = children->data;
      children = children->next;

      (* callback) (child->widget, callback_data);
    }
}

#if GTK_MAJOR_VERSION == 2
static void
gtk_group_size_request (GtkWidget      *widget,
		        GtkRequisition *requisition)
{
  GtkGroup *group = GTK_GROUP (widget);
  GList *children = group->children;
  gint i, num1 = group->row_cols, num2 = 0; 
  gint *min_dim1 = group->min_dim1, max_dim1 = 0, sum_dim1 = 0;
  gint dim2, max_dim2 = 0, sum_dim2 = 0;
  guint *weight1 = group->weight1, weight1_total = 0, weight2_total = 0;
  guint border;

  for (i=0; i < num1; i++)
    {
      min_dim1[i] = 0;
      weight1[i] = 0;
    }

  while (children)
    {
      dim2 = 0;

      for (i=0; i < num1 && children; i++, children = children->next)
	{
          GtkGroupChild *child = children->data;
	  GtkRequisition child_req;

          if (gtk_widget_get_visible (child->widget))
	    {
	      gtk_widget_size_request (child->widget, &child_req);

              if (group->horiz)
		{
		  min_dim1[i] = MAX(min_dim1[i], child_req.height);
		  dim2 = MAX(dim2, child_req.width);
		}
	      else
		{
		  min_dim1[i] = MAX(min_dim1[i], child_req.width);
		  dim2 = MAX(dim2, child_req.height);
		}
	    }
	  if (group->horiz)
	    {
	      weight1[i] += child->vert_weight;
	      weight1_total += child->vert_weight;
	      weight2_total  += child->horiz_weight;
	    }
	  else
	    {
	      weight1[i] += child->horiz_weight;
	      weight1_total += child->horiz_weight;
	      weight2_total  += child->vert_weight;
	    }
	}
      max_dim2 = MAX(max_dim2, dim2);
      sum_dim2 += dim2;
      num2++;
    }

  if (group->horiz? group->same_height : group->same_width)
    {
      for (i = 0; i < num1; i++)
	{
	  max_dim1 = MAX(max_dim1, min_dim1[i]);
	}
      for (i = 0; i < num1; i++)
	{
	  min_dim1[i] = max_dim1;
	}
    }

  if (group->horiz? group->same_width : group->same_height)
    {
      sum_dim2 = max_dim2 * num2;
    }

  for (i = 0; i < num1; i++)
    {
      sum_dim1 += min_dim1[i];
    }

  if (num2 == 0) {
      num2 = 1;		/* for spacing */
  }
  if (group->horiz)
    {
      requisition->width  = sum_dim2 + (num2 - 1) * group->horiz_spacing;
      requisition->height = sum_dim1 + (num1 - 1) * group->vert_spacing;
    }
  else
    {
      requisition->width  = sum_dim1 + (num1 - 1) * group->horiz_spacing;
      requisition->height = sum_dim2 + (num2 - 1) * group->vert_spacing;
    }

  border = gtk_container_get_border_width (GTK_CONTAINER (group));
  requisition->width += border * 2; 
  requisition->height += border * 2;

  group->max_min_dim1 = max_dim1;
  group->max_min_dim2 = max_dim2;
  group->weight1_total = weight1_total;
  group->weight2_total = weight2_total;
}

#else

static void 
gtk_group_get_preferred_size (GtkGroup       *group,
                              GtkOrientation  orientation,
                               gint          *minimal_size,
                               gint          *natural_size)
{
  GList *children = group->children;
  gint i, num1 = group->row_cols, num2 = 0; 
  gint *min_dim1 = group->min_dim1, *nat_dim1 = group->nat_dim1;
  gint max_min_dim1 = 0, max_nat_dim1 = 0, sum_min_dim1 = 0, sum_nat_dim1 = 0;
  gint max_min_dim2 = 0, max_nat_dim2 = 0, sum_min_dim2 = 0, sum_nat_dim2 = 0;
  guint *weight1 = group->weight1, weight1_total = 0, weight2_total = 0;
  gint spacing;
  gboolean compute_dim1;
  gint min_size, nat_size;
  guint border;

  compute_dim1 = (orientation == GTK_ORIENTATION_HORIZONTAL && !group->horiz) ||
                 (orientation != GTK_ORIENTATION_HORIZONTAL &&  group->horiz);
  if (compute_dim1)
    {
      for (i=0; i < num1; i++)
        {
          min_dim1[i] = 0;
          nat_dim1[i] = 0;
          weight1[i] = 0;
        }
    }

  while (children)
    {
      gint min_dim2 = 0;
      gint nat_dim2 = 0;

      for (i=0; i < num1 && children; i++, children = children->next)
	{
          GtkGroupChild *child = children->data;
          gint child_minimum_size, child_natural_size;
          gint weight;

          if (gtk_widget_get_visible (child->widget))
	    {
              if (orientation == GTK_ORIENTATION_HORIZONTAL)
                {
                  gtk_widget_get_preferred_width (child->widget,
                                                  &child_minimum_size,
                                                  &child_natural_size);
                }
              else
                {
                  gtk_widget_get_preferred_height (child->widget,
                                                   &child_minimum_size,
                                                   &child_natural_size);
                }
              if (compute_dim1)
                {
                  min_dim1[i] = MAX(min_dim1[i], child_minimum_size);
                  nat_dim1[i] = MAX(nat_dim1[i], child_natural_size);
                }
              else
                {
                  min_dim2 = MAX(min_dim2, child_minimum_size);
                  nat_dim2 = MAX(nat_dim2, child_natural_size);
                }
	    }
          if (compute_dim1)
            {
              weight = group->horiz? child->vert_weight : child->horiz_weight;
              weight1[i] += weight;
              weight1_total += weight;
            }
          else
            {
              weight = group->horiz? child->horiz_weight : child->vert_weight;
              weight2_total  += weight;
            }
	}
      if (!compute_dim1)
        {
          max_min_dim2 = MAX(max_min_dim2, min_dim2);
          max_nat_dim2 = MAX(max_nat_dim2, nat_dim2);
          sum_min_dim2 += min_dim2;
          sum_nat_dim2 += nat_dim2;
        }
      num2++;
    }

  if (compute_dim1)
    {
      if (group->horiz? group->same_height : group->same_width)
        {
          for (i = 0; i < num1; i++)
            {
              max_min_dim1 = MAX(max_min_dim1, min_dim1[i]);
              max_nat_dim1 = MAX(max_nat_dim1, nat_dim1[i]);
            }
          for (i = 0; i < num1; i++)
            {
              min_dim1[i] = max_min_dim1;
              nat_dim1[i] = max_nat_dim1;
            }
        }

      for (i = 0; i < num1; i++)
        {
          sum_min_dim1 += min_dim1[i];
          sum_nat_dim1 += nat_dim1[i];
        }

      group->max_min_dim1 = max_min_dim1;
      group->weight1_total = weight1_total;
    }
  else
    {
      if (group->horiz? group->same_width : group->same_height)
        {
          sum_min_dim2 = max_min_dim2 * num2;
          sum_nat_dim2 = max_nat_dim2 * num2;
        }

      group->max_min_dim2 = max_min_dim2;
      group->weight2_total = weight2_total;
    }

  spacing = (orientation == GTK_ORIENTATION_HORIZONTAL)? 
             group->horiz_spacing : group->vert_spacing;

  if (compute_dim1)
    {
      min_size = sum_min_dim1 + (num1 - 1) * spacing;
      nat_size = sum_nat_dim1 + (num1 - 1) * spacing;
    }
  else
    {
      if (num2 == 0) 
        {
          num2 = 1;		/* for spacing */
        }
      min_size = sum_min_dim2 + (num2 - 1) * spacing;
      nat_size = sum_nat_dim2 + (num2 - 1) * spacing;
    }

  border = gtk_container_get_border_width (GTK_CONTAINER (group));
  min_size += border * 2;
  nat_size += border * 2;

  if (minimal_size != NULL)
    {
      *minimal_size = min_size;
    }
  if (natural_size != NULL)
    {
      *natural_size = nat_size;
    }
}

static void 
gtk_group_get_preferred_width (GtkWidget *widget,
                               gint      *minimal_width,
                               gint      *natural_width)
{
    gtk_group_get_preferred_size (GTK_GROUP (widget),
                                  GTK_ORIENTATION_HORIZONTAL,
                                  minimal_width,
                                  natural_width);
}

static void 
gtk_group_get_preferred_height (GtkWidget *widget,
                                gint      *minimal_height,
                                gint      *natural_height)
{
    gtk_group_get_preferred_size (GTK_GROUP (widget),
                                  GTK_ORIENTATION_VERTICAL,
                                  minimal_height,
                                  natural_height);
}

#endif

/*
** gtk_group_size_allocate
**
** Here is the idea behind the geomnetry managment of group class.
** Each element has a min-size and a weight in both horizontal and
** vertical direction. Both directions are handled in the same way.
**
** The allocated size of an element is at least min-size.
** The desired size of an element is available-size * weight / total-weight.
** First we check which elements have to be rendererd at their minimum size.
** These elements are then not considered for the further steps and their
** size is subtracted from the available-size.  The remaining elements
** then obtain their size according to the formula above.
** We also add space between the elements according to the specified settings.
*/

static void
gtk_group_size_allocate (GtkWidget     *widget,
			 GtkAllocation *allocation)
{
  GtkGroup *group = GTK_GROUP(widget);
  GList *children = group->children;
  guint border = gtk_container_get_border_width (GTK_CONTAINER (group));
  gint xmin = allocation->x + border; 
  gint ymin = allocation->y + border;
  gint num1 = group->row_cols;
  gint num2 = (g_list_length(children) + num1 - 1) / num1;
  guint weight1_total = group->weight1_total;
  guint weight2_total = group->weight2_total;
  gboolean fixed1 = (weight1_total == 0);
  gboolean fixed2 = (weight2_total == 0);
  gboolean same1, same2;
  gint space1, space2;
  gint total1, total2;
  gint i, j, x = xmin, y = ymin;
  gboolean *use_min1, *use_min2;
  gboolean change;

  gtk_widget_set_allocation (widget, allocation);

  if (num2 == 0)
    {
      return;		/* empty group */
    }

  if (group->horiz)
    {
      same1 = group->same_height;
      same2 = group->same_width;
      space1 = group->vert_spacing;
      space2 = group->horiz_spacing;
      total1 = allocation->height;
      total2 = allocation->width;
    }
  else
    {
      same1 = group->same_width;
      same2 = group->same_height;
      space1 = group->horiz_spacing;
      space2 = group->vert_spacing;
      total1 = allocation->width;
      total2 = allocation->height;
    }
  total1 -= 2*border + (num1 - 1) * space1;
  total2 -= 2*border + (num2 - 1) * space2;

  if (same1 || fixed1) 			  /* dim1 all same or not expandable?*/
    {
      weight1_total = num1;	          /* all objects get same weight */
    }
  if (same2 || fixed2)                    /* dim2 all same or not expandable?*/
    {
      weight2_total = num2; 		  /* all objects get same weight */
    }

  use_min1 = g_new0(gboolean, num1);
  use_min2 = g_new0(gboolean, num2);

  /* 
  ** Check which elements have to rendered at minimum size and remove
  ** their size from the available size.
  */
  do
    {
      change = FALSE;

      for (i=0; i < num1; i++)
	{
	  if (!use_min1[i])
	    {
	      gint dim1 = group->min_dim1[i];
	      guint weight1 = group->weight1[i];

	      if (same1 || fixed1)              /* dim1 same or not expand?*/
		{
		  weight1 = 1;		        /* all objects get same dim1 */
		}
	      if (weight1 == 0 ||
		      dim1 >= total1 * (gint)weight1 / (gint)weight1_total)
		{
		  total1 -= dim1;
		  weight1_total -= weight1;
		  use_min1[i] = TRUE;
		  change = TRUE;
		}
	    }
	}
    } while (change);

  do
    {
      change = FALSE;

      for (j=0; children; j++) 
	{
	  gint dim2 = 0;
	  guint weight2 = 0;
	  gint i;

	  if (same2)
	    {
	      dim2 = group->max_min_dim2; 
	    }
	  for (i = 0; i < num1 && children; i++, children = children->next)
	    {
	      if (!use_min2[j] && !same2)
		{
		  GtkGroupChild *child = children->data;
                  gint child_min_width = 0, child_min_height = 0;

		  if (gtk_widget_get_visible (child->widget))
		    {
#if GTK_MAJOR_VERSION == 2
                      GtkRequisition child_req;

		      gtk_widget_get_child_requisition (child->widget, 
			                               &child_req);
                      child_min_width = child_req.width;
                      child_min_height = child_req.height;
#else
                      gtk_widget_get_preferred_width (child->widget, 
                                                     &child_min_width, NULL);
                      gtk_widget_get_preferred_height (child->widget, 
                                                      &child_min_height, NULL);
#endif
		    }
		  if (group->horiz)
		    {
		      dim2 = MAX(dim2, child_min_width);
		      weight2 += child->horiz_weight;
		    }
		  else
		    {
		      dim2 = MAX(dim2, child_min_height);
		      weight2 += child->vert_weight;
		    }
		}
	    }
	  if (use_min2[j])
	    {
	      continue;
	    }
	  if (same2 || fixed2)                  /* dim2 same or not expand?*/
	    {
	      weight2 = 1;		        /* all objects get same dim2 */
	    }
	  if (weight2 == 0 ||
		    dim2 >= total2 * (gint)weight2 / (gint)weight2_total)
	    {
	      total2 -= dim2;
	      weight2_total -= weight2;
	      use_min2[j] = TRUE;
	      change = TRUE;
	    }
	}
      children = group->children;
    } while (change);

  for (j=0; children; j++)
    {
      gint dim2 = 0;
      guint weight2 = 0;
      gint total1_tmp = total1;
      guint weight1_total_tmp = weight1_total;
      GList *tmp;

      if (same2)
	{
	  dim2 = group->max_min_dim2; 
	}
      else
	{
          for (i = 0, tmp = children; i < num1 && tmp; i++, tmp = tmp->next)
	    {
	      GtkGroupChild *child = tmp->data;
              gint child_min_width = 0, child_min_height = 0;

              if (gtk_widget_get_visible (child->widget))
	        {
#if GTK_MAJOR_VERSION == 2
                  GtkRequisition child_req;

                  gtk_widget_get_child_requisition (child->widget, &child_req);
                  child_min_width = child_req.width;
                  child_min_height = child_req.height;
#else
                  gtk_widget_get_preferred_width (child->widget, 
                                                 &child_min_width, NULL);
                  gtk_widget_get_preferred_height (child->widget, 
                                                  &child_min_height, NULL);
#endif
	        }
	      if (group->horiz)
	        {
	          dim2 = MAX(dim2, child_min_width);
	          weight2 += child->horiz_weight;
	        }
	      else
	        {
	          dim2 = MAX(dim2, child_min_height);
	          weight2 += child->vert_weight;
	        }
	    }
	}
      if (same2 || fixed2)                     /* dim2 same or not expandable?*/
        {
	  weight2 = 1;			       /* all objects get same dim2 */
	}
      if (!use_min2[j])
	{
      	  dim2 = total2 * (gint)weight2 / (gint)weight2_total;
	  total2 -= dim2;
	  weight2_total -= weight2;
	}

      if (group->horiz)
	{
	  y = ymin;
	}
      else
	{
	  x = xmin;
	}

      for (i = 0; i < num1 && children; i++, children = children->next)
	{
	  GtkGroupChild *child = children->data;
	  gint dim1 = group->min_dim1[i];
	  gint weight1 = group->weight1[i];
	  GtkAllocation child_allocation;

          if (same1 || fixed1)                  /* same or not expandable? */
	    {
	      weight1 = 1;		        /* all objects get same dim1 */
	    }
	  if (!use_min1[i])
	    {
	      dim1 = total1_tmp * (gint)weight1 / (gint)weight1_total_tmp;
	      total1_tmp -= dim1;
	      weight1_total_tmp -= weight1;
	    }

	  child_allocation.x = x;
	  child_allocation.y = y;

	  if (group->horiz)
	    {
	      child_allocation.width = dim2;
	      child_allocation.height = dim1;
	      y += dim1 + group->vert_spacing;
	    }
	  else
	    {
	      child_allocation.width = dim1;
	      child_allocation.height = dim2;
	      x += dim1 + group->horiz_spacing;
	    }
	  /* clipping */
	  if (child_allocation.x < xmin)
	    {
	      child_allocation.width -= xmin - child_allocation.x;
	      child_allocation.x = xmin;
	    }
	  if (child_allocation.y < ymin)
	    {
	      child_allocation.height -= ymin - child_allocation.y;
	      child_allocation.y = ymin;
	    }

	  child_allocation.width = MAX(child_allocation.width, 0);
	  child_allocation.height = MAX(child_allocation.height, 0);
	  gtk_widget_size_allocate (child->widget, &child_allocation);
	}

      if (group->horiz)
	{
	  x += dim2 + group->horiz_spacing;
	}
      else
	{
	  y += dim2 + group->vert_spacing;
	}
    }

  g_free(use_min2);
  g_free(use_min1);
}

static void 
gtk_group_set_window_geometry_hints (GtkWidget *widget)
{
  GtkWidget *parent = gtk_widget_get_parent (widget);

  if (GTK_IS_WINDOW(parent))
    {
      GtkWindow *window = GTK_WINDOW(parent);
      GtkGroup *group = GTK_GROUP(widget);
      gint min_width;
      GdkGeometry geometry;
      gboolean resizable = FALSE, full = TRUE;

#if GTK_MAJOR_VERSION == 2
      GtkRequisition requisition;
      gtk_widget_size_request (widget, &requisition);
      min_width = requisition.width;
#else
      gint min_height;
      gtk_widget_get_preferred_width (widget, &min_width, NULL);
      gtk_widget_get_preferred_height (widget, &min_height, NULL);
#endif
      geometry.min_width = (min_width < 72)? 72 : -1;
      geometry.min_height = -1;

      if ((group->horiz? group->weight2_total : group->weight1_total) == 0)
	{
	  geometry.max_width = geometry.min_width;
	  full = FALSE;
	}
      else
	{
	  geometry.max_width = G_MAXSHORT;
	  resizable = TRUE;
	}
      if ((group->horiz? group->weight1_total : group->weight2_total) == 0)
	{
	  geometry.max_height = geometry.min_height;
	  full = FALSE;
	}
      else
	{
	  geometry.max_height = G_MAXSHORT;
	  resizable = TRUE;
	}

      gtk_window_set_geometry_hints(window, NULL, &geometry, 
	      GDK_HINT_MIN_SIZE | (full? 0 : GDK_HINT_MAX_SIZE));
      gtk_window_set_resizable(window, resizable);
    }
}

#if 0
static void
gtk_group_check_window_geometry (GtkWidget *widget)
{
  if (GTK_IS_WINDOW (widget->parent))
    {
      GtkWindow *window = GTK_WINDOW(widget->parent);
      GtkGroup *group = GTK_GROUP(widget);
      GtkRequisition requisition;
      guint horiz_weight, vert_weight;

      gtk_widget_size_request (widget, &requisition);
      adjust_min_requisition (&requisition);
      horiz_weight = group->horiz?  group->weight2_total : group->weight1_total;
      vert_weight  = group->horiz?  group->weight1_total : group->weight2_total;

      if (horiz_weight != 0 && requisition.width < widget->allocation.width)
        {
	  requisition.width = widget->allocation.width;
	}
      if (vert_weight != 0 && requisition.height < widget->allocation.height)
        {
	  requisition.height = widget->allocation.height;
	}
      
      if (requisition.width != widget->allocation.width || 
	      requisition.height != widget->allocation.height)
	{
	  /* printf("resizing window from (%d %d) to (%d %d)\n",
		  widget->allocation.width, widget->allocation.height,
		  requisition.width, requisition.height); */
	  gtk_group_set_window_geometry_hints (widget);
	  gtk_window_resize (window, requisition.width, requisition.height);
	}

    }
}
#endif

