/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkpopeditlabels.c
**
** A button that when pressed pops up
** an EditLabels window for editing labels
**
** Creation data: 02.09.2005
** Last modified: 21.04.2016
*/

#define ENABLE_NLS

#include "gtkpopeditlabels.h"
#include "gtkeditlabels.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include <glib/gi18n.h>
#include "labels.h"
#include "memory.h"
#include "appstrings_cat.h"

#define P_(x) N_(x)

enum {
  PROP_0,
  PROP_LABELS,
};

enum {
  ACCEPT,
  LAST_SIGNAL
};

static void gtk_pop_edit_labels_class_init     (GtkPopEditLabelsClass   *klass);
static void gtk_pop_edit_labels_init           (GtkPopEditLabels        *button);
#if GTK_MAJOR_VERSION == 2
static void gtk_pop_edit_labels_destroy        (GtkObject          *object);
#else
static void gtk_pop_edit_labels_destroy        (GtkWidget          *widget);
#endif
static gboolean gtk_pop_edit_labels_accept     (GtkPopEditLabels   *button,
                                                char              **labels,
                                                gpointer            data);
static void gtk_pop_edit_labels_set_property   (GObject            *object,
                                                guint               prop_id,
                                                const GValue       *value,
                                                GParamSpec         *pspec);
static void gtk_pop_edit_labels_get_property   (GObject            *object,
                                                guint               prop_id,
                                                GValue             *value,
                                                GParamSpec         *pspec);

static GtkButtonClass *parent_class = NULL;

static guint gtk_pop_edit_labels_signals[LAST_SIGNAL] = { 0 };

static void
g_cclosure_marshal_BOOLEAN__LABELS (GClosure     *closure,
                               GValue       *return_value,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint G_GNUC_UNUSED,
                               gpointer      marshal_data)
{
  typedef gboolean (*GMarshalFunc_BOOLEAN__LABELS) (gpointer     data1,
                                                    char       **labels,
  					            gpointer     data2);
  GMarshalFunc_BOOLEAN__LABELS callback;
  GCClosure *cc = (GCClosure*) closure;
  gpointer data1, data2;
  char **labels;
  gboolean v_return;

  g_return_if_fail (n_param_values == 2);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }
  labels = g_value_peek_pointer (param_values + 1);
  callback = (GMarshalFunc_BOOLEAN__LABELS) (marshal_data ? 
          marshal_data : cc->callback);

  v_return = callback (data1, labels, data2);
  g_value_set_boolean (return_value, v_return);
}

GType
gtk_pop_edit_labels_get_type (void)
{
  static GType pop_edit_labels_type = 0;

  if (!pop_edit_labels_type)
    {
      static const GTypeInfo pop_edit_labels_info =
      {
	sizeof (GtkPopEditLabelsClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_pop_edit_labels_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkPopEditLabels),
	0,		/* n_preallocs */
	(GInstanceInitFunc) gtk_pop_edit_labels_init,
      };

      pop_edit_labels_type = g_type_register_static (GTK_TYPE_BUTTON,
                                                     "GtkPopEditLabels", 
                                                     &pop_edit_labels_info, 0);
    }

  return pop_edit_labels_type;
}

static void
gtk_pop_edit_labels_class_init (GtkPopEditLabelsClass *klass)
{
  GObjectClass *gobject_class;
#if GTK_MAJOR_VERSION == 2
  GtkObjectClass *object_class;
#else
  GtkWidgetClass *widget_class;
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = gtk_pop_edit_labels_set_property;
  gobject_class->get_property = gtk_pop_edit_labels_get_property;

#if GTK_MAJOR_VERSION == 2
  object_class = GTK_OBJECT_CLASS (klass);
  object_class->destroy = gtk_pop_edit_labels_destroy;
#else
  widget_class = GTK_WIDGET_CLASS (klass);
  widget_class->destroy = gtk_pop_edit_labels_destroy;
#endif

  klass->accept = gtk_pop_edit_labels_accept;

  g_object_class_install_property (gobject_class,
				   PROP_LABELS,
				   g_param_spec_pointer ("labels",
						      P_("Labels"),
						      P_("Array of isocode strings for the labels"),
                                                      G_PARAM_READWRITE | G_PARAM_CONSTRUCT));

  gtk_pop_edit_labels_signals[ACCEPT] =
    g_signal_new ("accept",
	          G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_LAST, // | G_SIGNAL_ACTION,
		  G_STRUCT_OFFSET (GtkPopEditLabelsClass, accept),
		  NULL, NULL,
		  g_cclosure_marshal_BOOLEAN__LABELS,
		  G_TYPE_BOOLEAN, 1, G_TYPE_POINTER);
}

static void 
closePopEditLabels(GtkWidget *dialog, gint response, GtkPopEditLabels *pop)
{
  if (response == GTK_RESPONSE_ACCEPT) 
    {
      gboolean success = FALSE;
      char **labels;

      labels = gtk_edit_labels_get_labels ( GTK_EDIT_LABELS (pop->edit_labels));
      g_signal_emit (pop, gtk_pop_edit_labels_signals[ACCEPT], 0,
              labels, &success);

      if (!success)
        {
          return;
        }
      gtk_pop_edit_labels_set_labels (pop, labels);
    }

  gtk_widget_hide (pop->window);
}

static void openPopEditLabels(GtkWidget *button, GtkPopEditLabels *pop)
{
  if (gtk_widget_get_visible (pop->window))
    {
      //gtk_dialog_response (GTK_DIALOG (pop->window), GTK_RESPONSE_NONE);
    }
  else
    {
      gtk_edit_labels_set_labels (GTK_EDIT_LABELS (pop->edit_labels), 
                                  pop->labels);
      gtk_window_set_transient_for (GTK_WINDOW (pop->window), 
              GuiFindTransientWindow (button));
    }
  gtk_window_present (GTK_WINDOW (pop->window));
  GuiListSetActive (GTK_EDIT_LABELS(pop->edit_labels)->tree, -1);
}

static void
gtk_pop_edit_labels_init (GtkPopEditLabels *pop)
{
  GtkWidget *vbox, *ok, *cancel;
  gint width, height;

  pop->memPool = CreatePool(0, 4096, 1024);
  pop->labels = NULL;
  pop->edit_labels = EditLabelsObject, End;
  pop->window = DialogObject,
      "title", GCS(MSG_EditLabels),
      "window-position", GTK_WIN_POS_CENTER,
      "modal", FALSE,
      End;
  vbox = gtk_dialog_get_content_area (GTK_DIALOG (pop->window));
  gtk_box_pack_start(GTK_BOX(vbox), pop->edit_labels, TRUE, TRUE, 0);
  gtk_widget_show_all(pop->edit_labels);

  ok = GuiSimpleButton(GCS(MSG_Ok), TRUE);
  cancel = GuiSimpleButton(GCS(MSG_Cancel), FALSE);

  gtk_dialog_add_action_widget(GTK_DIALOG(pop->window), 
          cancel, GTK_RESPONSE_CANCEL);
  gtk_widget_show(cancel);
  gtk_dialog_add_action_widget(GTK_DIALOG(pop->window), 
          ok, GTK_RESPONSE_ACCEPT);
  gtk_window_set_default(GTK_WINDOW(pop->window), ok);
  gtk_widget_show(ok);

  GuiAddEntryCallbacks(GTK_EDIT_LABELS(pop->edit_labels)->entry, ok);

  g_signal_connect(pop->window, "response", G_CALLBACK(closePopEditLabels),pop);

  GuiSetDialogButtonOrder(pop->window);
  GuiSetDialogGeometryHints(pop->window, pop->edit_labels);
  GuiAddWindowCallbacks(pop->window, cancel);

  gtk_window_get_size(GTK_WINDOW(pop->window), &width, &height);
  width += 50;
  height += 100;
  gtk_window_set_default_size(GTK_WINDOW(pop->window), width, height);

  g_signal_connect (pop, "clicked", G_CALLBACK(openPopEditLabels), pop);
}

static void
gtk_pop_edit_labels_cleanup (GtkPopEditLabels *pop)
{
  if (pop->window)
    {
      gtk_widget_destroy (pop->window);
      pop->window = 0;
      pop->edit_labels = 0;
      pop->labels = NULL;
      DeletePool (pop->memPool);
      pop->memPool = NULL;
    }
}
#if GTK_MAJOR_VERSION == 2
static void
gtk_pop_edit_labels_destroy (GtkObject *object)
{
  gtk_pop_edit_labels_cleanup (GTK_POP_EDIT_LABELS (object));
  (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}
#else
static void
gtk_pop_edit_labels_destroy (GtkWidget *widget)
{
  gtk_pop_edit_labels_cleanup (GTK_POP_EDIT_LABELS (widget));
  (* GTK_WIDGET_CLASS (parent_class)->destroy) (widget);
}
#endif

static gboolean
gtk_pop_edit_labels_accept (GtkPopEditLabels   *button,
                            char              **labels,
                            gpointer            data)
{
  return TRUE;
}

static void
gtk_pop_edit_labels_set_property (GObject         *object,
                                  guint            prop_id,
                                  const GValue    *value,
                                  GParamSpec      *pspec)
{
  GtkPopEditLabels *pop = GTK_POP_EDIT_LABELS (object);

  switch (prop_id)
    {
    case PROP_LABELS:
      gtk_pop_edit_labels_set_labels (pop, (char**)g_value_get_pointer (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
gtk_pop_edit_labels_get_property (GObject         *object,
                                  guint            prop_id,
                                  GValue          *value,
                                  GParamSpec      *pspec)
{
  GtkPopEditLabels *pop = GTK_POP_EDIT_LABELS (object);

  switch (prop_id)
    {
    case PROP_LABELS:
      g_value_set_pointer (value, pop->labels);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

GtkWidget*
gtk_pop_edit_labels_new (void)
{
  return g_object_new (GTK_TYPE_POP_EDIT_LABELS, NULL);
}

void
gtk_pop_edit_labels_set_labels (GtkPopEditLabels   *pop,
		                char              **labels)
{
  char **old;

  g_return_if_fail (GTK_IS_POP_EDIT_LABELS (pop));

  old = pop->labels;
  pop->labels = CopyLabels (pop->memPool, labels);
  if (gtk_widget_get_visible (pop->window))
    {
      gtk_edit_labels_set_labels (GTK_EDIT_LABELS (pop->edit_labels), labels);
    }
  FreeLabels (pop->memPool, old);
  
  g_object_notify (G_OBJECT (pop), "labels");
}

char **
gtk_pop_edit_labels_get_labels (GtkPopEditLabels *pop)
{
  g_return_val_if_fail (GTK_IS_POP_EDIT_LABELS (pop), NULL);

  return pop->labels;
}

