/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkpopimage.c
**
** MUI's popimage class for gtk
**
** Creation data: 21.06.2005
** Last modified: 04.03.2015
*/

#define ENABLE_NLS

#include "gtkpopimage.h"
#include "gtkgroup.h"
#include "gtkpopselection.h"
#include "gtkextimage.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "appstrings_cat.h"
#include <glib/gi18n.h>

#define P_(x) N_(x)

enum {
  PROP_0,
  PROP_SPEC,
  PROP_TYPES,
};

#define POP_IMAGE_BORDER 3
#define POP_IMAGE_WIDTH 30
#define POP_IMAGE_HEIGHT 30

#define POP_IMAGE_NUM_PATTERNS 18

#define GTK_POP_IMAGE_GET_PRIVATE(o)   (G_TYPE_INSTANCE_GET_PRIVATE ((o), GTK_TYPE_POP_IMAGE, GtkPopImagePrivate))
typedef struct _GtkPopImagePrivate GtkPopImagePrivate;

struct _GtkPopImagePrivate
{
  gchar       *spec;
  guint        types;
  GtkWidget   *image;
  GtkWidget   *dialog;
  guint        page_type[3];
  GtkWidget   *notebook;
#ifdef ENABLE_GDK_STIPPLED
  GtkWidget   *pattern[POP_IMAGE_NUM_PATTERNS];
#endif
  GtkWidget   *color;
  GtkWidget   *external;

};

static void gtk_pop_image_class_init     (GtkPopImageClass   *klass);
static void gtk_pop_image_init           (GtkPopImage        *button);
#if GTK_MAJOR_VERSION == 2
static void gtk_pop_image_destroy        (GtkObject          *object);
#else
static void gtk_pop_image_destroy        (GtkWidget          *widget);
#endif
static void gtk_pop_image_set_property   (GObject            *object,
                                          guint               prop_id,
                                          const GValue       *value,
                                          GParamSpec         *pspec);
static void gtk_pop_image_get_property   (GObject            *object,
                                          guint               prop_id,
                                          GValue             *value,
                                          GParamSpec         *pspec);
static void gtk_pop_image_open_popup     (GtkPopImage        *pop_image);
static void gtk_pop_image_create_dialog  (GtkPopImage        *pop_image);

static GtkButtonClass *parent_class = NULL;

GType
gtk_pop_image_get_type (void)
{
  static GType pop_image_type = 0;

  if (!pop_image_type)
    {
      static const GTypeInfo pop_image_info =
      {
	sizeof (GtkPopImageClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_pop_image_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkPopImage),
	16,		/* n_preallocs */
	(GInstanceInitFunc) gtk_pop_image_init,
      };

      pop_image_type = g_type_register_static (GTK_TYPE_BUTTON, "GtkPopImage",
					        &pop_image_info, 0);
    }

  return pop_image_type;
}

static void
gtk_pop_image_class_init (GtkPopImageClass *klass)
{
  GObjectClass *gobject_class;
#if GTK_MAJOR_VERSION == 2
  GtkObjectClass *object_class;
#else
  GtkWidgetClass *widget_class;
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = gtk_pop_image_set_property;
  gobject_class->get_property = gtk_pop_image_get_property;

#if GTK_MAJOR_VERSION == 2
  object_class = GTK_OBJECT_CLASS (klass);
  object_class->destroy = gtk_pop_image_destroy;
#else
  widget_class = GTK_WIDGET_CLASS (klass);
  widget_class->destroy = gtk_pop_image_destroy;
#endif

  g_object_class_install_property (gobject_class,
                                   PROP_SPEC,
                                   g_param_spec_string ("spec",
                                                     P_("Spec"),
                                                     P_("Specification of the image"),
                                                        NULL,
                                                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
  
  g_object_class_install_property (gobject_class,
                                   PROP_TYPES,
                                   g_param_spec_uint  ("types",
                                                    P_("Types"),
                                                    P_("Flags specifying which types of images are allowed"),
                                                        0, 
                                                        GTK_POP_IMAGE_ALL,
                                                        GTK_POP_IMAGE_ALL,
                                                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
  
  g_type_class_add_private (gobject_class, sizeof (GtkPopImagePrivate));  
}

static void
gtk_pop_image_init (GtkPopImage *pop_image)
{
  GtkPopImagePrivate *priv = GTK_POP_IMAGE_GET_PRIVATE (pop_image);

  memset (priv, 0, sizeof (*priv));

  priv->types = GTK_POP_IMAGE_ALL;

  priv->image = GroupObject,
      "border-width", POP_IMAGE_BORDER,
      "width-request", POP_IMAGE_WIDTH,
      "height-request", POP_IMAGE_HEIGHT,
      End;

  gtk_container_add (GTK_CONTAINER (pop_image), priv->image);

  g_signal_connect (pop_image, "clicked", 
          G_CALLBACK (gtk_pop_image_open_popup), NULL);

  /* 
  ** here we should also install callback handlers for map and unmap events
  ** similar to gtkpopstring
  */  
}

static void
gtk_pop_image_cleanup (GtkPopImage *button)
{
  GtkPopImagePrivate *priv = GTK_POP_IMAGE_GET_PRIVATE (button);
  
  g_free (priv->spec);
  priv->spec = NULL;
  if (priv->dialog) 
    {
      gtk_widget_destroy (priv->dialog);
      priv->dialog = NULL;
    }
}
#if GTK_MAJOR_VERSION == 2
static void
gtk_pop_image_destroy (GtkObject *object)
{
  gtk_pop_image_cleanup (GTK_POP_IMAGE (object));
  (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}
#else
static void
gtk_pop_image_destroy (GtkWidget *widget)
{
  gtk_pop_image_cleanup (GTK_POP_IMAGE (widget));
  (* GTK_WIDGET_CLASS (parent_class)->destroy) (widget);
}
#endif

static void
gtk_pop_image_set_property (GObject         *object,
                            guint            prop_id,
                            const GValue    *value,
                            GParamSpec      *pspec)
{
  GtkPopImage *pop_image = GTK_POP_IMAGE (object);

  switch (prop_id)
    {
    case PROP_SPEC:
      gtk_pop_image_set_spec (pop_image, g_value_get_string (value));
      break;
    case PROP_TYPES:
      gtk_pop_image_set_types (pop_image, g_value_get_uint (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
gtk_pop_image_get_property (GObject         *object,
                            guint            prop_id,
                            GValue          *value,
                            GParamSpec      *pspec)
{
  GtkPopImage *pop_image = GTK_POP_IMAGE (object);
  GtkPopImagePrivate *priv = GTK_POP_IMAGE_GET_PRIVATE (pop_image);

  switch (prop_id)
    {
    case PROP_SPEC:
      g_value_set_string (value, priv->spec);
      break;
    case PROP_TYPES:
      g_value_set_uint (value, priv->types);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

GtkWidget*
gtk_pop_image_new (void)
{
  return g_object_new (GTK_TYPE_POP_IMAGE, NULL);
}

void
gtk_pop_image_set_spec (GtkPopImage   *pop_image,
		        const gchar   *spec)
{
  GtkPopImagePrivate *priv;
  gchar *new_spec;
  
  g_return_if_fail (GTK_IS_POP_IMAGE (pop_image));

  priv = GTK_POP_IMAGE_GET_PRIVATE (pop_image);

  new_spec = g_strdup (spec);
  g_free (priv->spec);
  priv->spec = new_spec;

  gtk_widget_set_bg_spec (priv->image, new_spec);
  
  g_object_notify (G_OBJECT (pop_image), "spec");
}

G_CONST_RETURN gchar *
gtk_pop_image_get_spec (GtkPopImage *pop_image)
{
  GtkPopImagePrivate *priv;

  g_return_val_if_fail (GTK_IS_POP_IMAGE (pop_image), NULL);

  priv = GTK_POP_IMAGE_GET_PRIVATE (pop_image);
  
  return priv->spec;
}

void
gtk_pop_image_set_types (GtkPopImage   *pop_image,
		         guint          types)
{
  GtkPopImagePrivate *priv;
  
  g_return_if_fail (GTK_IS_POP_IMAGE (pop_image));

  priv = GTK_POP_IMAGE_GET_PRIVATE (pop_image);

  priv->types = types;
  
  g_object_notify (G_OBJECT (pop_image), "types");
}

guint
gtk_pop_image_get_types (GtkPopImage *pop_image)
{
  GtkPopImagePrivate *priv;

  g_return_val_if_fail (GTK_IS_POP_IMAGE (pop_image), 0);

  priv = GTK_POP_IMAGE_GET_PRIVATE (pop_image);
  
  return priv->types;
}

static const GdkColor MUIcolors[] = {
    { 0, 0xffff, 0xffff, 0xffff },  /* white */
    { 0, 0xe0e0, 0xe0e0, 0xe0e0 },  /* light gray */
    { 0, 0xc1c1, 0xc1c1, 0xc1c1 },  /* gray (background) */
    { 0, 0x6060, 0x6060, 0x6060 },  /* dark gray */
    { 0, 0x0000, 0x0000, 0x0000 },  /* black */
    { 0, 0x0000, 0x0000, 0x0000 },  /* black */
    { 0, 0x3b3b, 0x6767, 0xA2A2 },  /* blue */
    { 0, 0xffff, 0xa9a9, 0x9797 },  /* light pink */
};

enum {
    MUI_SHINE,
    MUI_HALFSHINE,
    MUI_BACKGROUND,
    MUI_HALFSHADOW,
    MUI_SHADOW,
    MUI_TEXT,
    MUI_FILL,
    MUI_MARK,
    MUI_NumColors
};

static void gtk_pop_image_set_color_chooser_color (GtkWidget *color_chooser,
                                                   const GdkColor *color)
{
  if (color_chooser != NULL && color != NULL)
    {
#if GTK_MAJOR_VERSION == 2
      gtk_color_selection_set_current_color (
              GTK_COLOR_SELECTION (color_chooser), color);
#else
      GdkRGBA rgba;
      rgba.red = (double)color->red / 0xFFFF;
      rgba.green = (double)color->green / 0xFFFF;
      rgba.blue = (double)color->blue / 0xFFFF;
      rgba.alpha = 1.0;
      gtk_color_chooser_set_rgba (GTK_COLOR_CHOOSER (color_chooser), &rgba);
#endif
    }
}

static void
gtk_pop_image_open_popup (GtkPopImage *pop)
{
  GtkWidget *top;
  GtkPopImagePrivate *priv;

  g_return_if_fail (GTK_IS_POP_IMAGE (pop));

  priv = GTK_POP_IMAGE_GET_PRIVATE (pop);

  gtk_pop_image_create_dialog (pop);
  if (priv->dialog == NULL)                     /* no types? */
    {
      return;
    }

  if (!gtk_widget_get_visible (priv->dialog) && priv->spec && priv->spec[0])
    {
      guint type = 0;
      gulong red, green, blue;
      gulong rgb;
      gint num;
      if (sscanf (priv->spec, "0:%d", &num) == 1)
        {
#ifdef ENABLE_GDK_STIPPLED
          if (num >= 128 && num < 128 + POP_IMAGE_NUM_PATTERNS)
            {
              GtkWidget *p = priv->pattern[num - 128];

              type = GTK_POP_IMAGE_PATTERN;
              if (p != NULL)
                {
                  gtk_toggle_button_set_active ( GTK_TOGGLE_BUTTON (p), TRUE);
                }
            }
#endif
        }
      else if (sscanf (priv->spec, "2:m%d", &num) == 1)
        {
          if (num >= 0 && num < MUI_NumColors) 
            {
              type = GTK_POP_IMAGE_COLOR;
              gtk_pop_image_set_color_chooser_color (priv->color,
                                                     &MUIcolors[num]);
            }
        }
      else if (sscanf (priv->spec, "2:r%lx,%lx,%lx", &red, &green, &blue) == 3)
        {
          GdkColor col;

          col.red = (guint16)(red >> 16);
          col.green = (guint16)(green >> 16);
          col.blue = (guint16)(blue >> 16);

          type = GTK_POP_IMAGE_COLOR;
          gtk_pop_image_set_color_chooser_color (priv->color, &col);
        }
      else if (sscanf (priv->spec, "2:r%lx", &rgb) == 1)
        {
          GdkColor col;

	  red = (rgb >> 16) & 0xff;
	  green = (rgb >> 8) & 0xff;
	  blue = rgb & 0xff;
	  col.red = (guint16)((red << 8) | red);
	  col.green = (guint16)((green << 8) | green);
	  col.blue = (guint16)((blue << 8) | blue);

          type = GTK_POP_IMAGE_COLOR;
          gtk_pop_image_set_color_chooser_color (priv->color, &col);
        }
      else if (strncmp (priv->spec, "5:", 2) == 0)
        {
          type = GTK_POP_IMAGE_EXTERNAL;
          if (priv->external)
            {
              gtk_entry_set_text (GTK_ENTRY (priv->external), priv->spec + 2);
            }
        }
      if (type != 0 && priv->notebook != NULL)
        {
          gint i;

          for (i = 0; i < 3; i++)
            {
              if (priv->page_type[i] == type)
                {
                  gtk_notebook_set_current_page(GTK_NOTEBOOK(priv->notebook),i);
                  break;
                }
            }
        }
    }

  top = gtk_widget_get_toplevel (GTK_WIDGET (pop));
  if (GTK_IS_WINDOW (top))
    {
      GtkWindowGroup *group;

      if (gtk_window_has_group (GTK_WINDOW (top)))
        {
          group = gtk_window_get_group (GTK_WINDOW(top));
        }
      else
        {
          group = gtk_window_group_new ();
          gtk_window_group_add_window (group, GTK_WINDOW (top));
          g_object_unref (G_OBJECT (group));
        }
      gtk_window_group_add_window (group, GTK_WINDOW (priv->dialog));
      gtk_window_set_transient_for (GTK_WINDOW (priv->dialog),GTK_WINDOW (top));

      if (gtk_widget_get_visible (top) && 
              !gtk_widget_get_visible (priv->dialog))
        {
	  GtkWindow *win = GTK_WINDOW (priv->dialog);
	  GdkScreen *screen = gtk_window_get_screen (win);
	  GtkWidget *widget = GTK_WIDGET (pop);
	  GtkAllocation alloc;
	  gint wx, wy, x, y, width, height, def_width, min_width; 
	  gint horiz_border = 0, vert_border = 0;
	  gint screen_width, screen_height;

#if GTK_MAJOR_VERSION == 2
	  GtkRequisition requisition;
	  gtk_widget_size_request (priv->dialog, &requisition);
          min_width = requisition.width;
#else
          gtk_widget_get_preferred_width(priv->dialog, &min_width, NULL);
#endif
          gtk_widget_get_allocation (widget, &alloc);
	  gtk_window_set_transient_for (win, GTK_WINDOW (top));
	  gdk_window_get_origin (gtk_widget_get_window (top), &x, &y);
	  gtk_window_get_size (win, NULL, &height);
	  gtk_window_get_default_size (win, &def_width, NULL);
	  def_width = MAX (def_width, min_width);

	  if (gtk_window_get_decorated (win))
	    {
	      horiz_border = 12;		/* best guess */
	      vert_border = 32;
	    }

	  gtk_widget_translate_coordinates(widget, top, 0, 0, &wx, &wy);
	  x += wx;
	  y += wy + alloc.height;
	  width = MAX(alloc.width - horiz_border, def_width);

	  if (screen)
	    {
	      screen_width = gdk_screen_get_width (screen);
	      screen_height = gdk_screen_get_height (screen);
	    }
	  else
	    {
	      screen_width = gdk_screen_width ();
	      screen_height = gdk_screen_height ();
	    }

	  x = CLAMP (x, 0, screen_width - width - horiz_border);
	  y = CLAMP (y, 0, screen_height - height - vert_border);

	  gtk_window_set_gravity (win, GDK_GRAVITY_NORTH_WEST);
	  gtk_window_move (win, x, y);
	  gtk_window_resize (win, width, height);
        }
    }

  gtk_window_present (GTK_WINDOW (priv->dialog));
}

#ifdef ENABLE_GDK_STIPPLED
static void
onPatternClicked (GtkWidget *pattern, GtkPopImagePrivate *priv)
{
  static gboolean called = FALSE;
  gint i;

  if (called == FALSE)
    {
      called = TRUE;
      for (i = 0; i < POP_IMAGE_NUM_PATTERNS; i++)
        {
          gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(priv->pattern[i]), 
                  pattern == priv->pattern[i]);
        }
      called = FALSE;
    }
}
#endif

static void 
onExternalEntryChanged (GtkWidget *entry, GtkWidget *image)
{
  const gchar *text = gtk_entry_get_text (GTK_ENTRY (entry));
  gchar *path = gtkExpandPathname(text);
  GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file (path, 0);

  g_free (path);
  g_object_set(G_OBJECT(image), "pixbuf", pixbuf, NULL);
  if (pixbuf)
    {
      g_object_unref (G_OBJECT (pixbuf));
    }
}

static void 
closePopImageDialog (GtkWidget *dialog, gint response, GtkPopImage *pop)
{
  GtkPopImagePrivate *priv = GTK_POP_IMAGE_GET_PRIVATE (pop);
  gint i;

  if (response == GTK_RESPONSE_ACCEPT)
    {
      gint page = 0;
      if (priv->notebook)
        {
          page = gtk_notebook_get_current_page (GTK_NOTEBOOK (priv->notebook));
        }
      switch (priv->page_type[page])
        {
#ifdef ENABLE_GDK_STIPPLED
          case GTK_POP_IMAGE_PATTERN:
              for (i = 0; i < POP_IMAGE_NUM_PATTERNS; i++)
                {
                  if (priv->pattern[i] && gtk_toggle_button_get_active (
                              GTK_TOGGLE_BUTTON (priv->pattern[i])))
                    {
                      gchar spec[10];
                      sprintf (spec, "0:%d", 128 + i);
                      gtk_pop_image_set_spec (pop, spec);
                      break;
                    }
                }
              break;
#endif

          case GTK_POP_IMAGE_COLOR:
              if (priv->color)
                {
                  GdkColor col;
                  gchar spec[32];

#if GTK_MAJOR_VERSION == 2
                  gtk_color_selection_get_current_color (
                          GTK_COLOR_SELECTION (priv->color), &col);
#else
                  GdkRGBA rgba;
                  gtk_color_chooser_get_rgba (GTK_COLOR_CHOOSER (priv->color),
                                              &rgba);
                  col.red = rgba.red * 0xFFFF;
                  col.green = rgba.green * 0xFFFF;
                  col.blue = rgba.blue * 0xFFFF;
#endif
                  sprintf(spec, "2:r%04hx%04hx,%04hx%04hx,%04hx%04hx",
                          col.red, col.red, col.green, col.green, 
                          col.blue, col.blue);
                  gtk_pop_image_set_spec (pop, spec);
                }
              break;

          case GTK_POP_IMAGE_EXTERNAL:
              if (priv->external)
                {
                  const gchar *text = gtk_entry_get_text (
                          GTK_ENTRY (priv->external));
                  gchar *spec = g_strdup_printf ("5:%s", text);
                  gtk_pop_image_set_spec (pop, spec);
                  g_free (spec);
                }
              break;
        }
    }

  gtk_widget_destroy (priv->dialog);
  priv->dialog = NULL;
  for (i = 0; i < 3; i++)
    {
      priv->page_type[i] = 0;
    }
  priv->notebook = NULL;
#ifdef ENABLE_GDK_STIPPLED
  for (i = 0; i < POP_IMAGE_NUM_PATTERNS; i++)
    {
      priv->pattern[i] = NULL;
    }
#endif
  priv->color = NULL;
  priv->external = NULL;
}
    
static void 
gtk_pop_image_create_dialog (GtkPopImage *pop)
{
  GtkPopImagePrivate *priv = GTK_POP_IMAGE_GET_PRIVATE (pop);
  GtkWidget *types[3]; 
  GtkWidget *top, *group, *root, *ok, *cancel, *dialog, *vbox;
  const char *labels[3];
  gint num = 0;

  if (priv->dialog != NULL)
    {
      return;
    }

#ifdef ENABLE_GDK_STIPPLED
  if (priv->types & GTK_POP_IMAGE_PATTERN)
    {
      gchar spec[10];
      gint i;
      GtkWidget *grp = VGroup,
          "columns", 6,
          End;

      for (i = 0; i < POP_IMAGE_NUM_PATTERNS; i++)
        {
          GtkWidget *go = VGroup,
              "border-width", POP_IMAGE_BORDER,
              "width-request", POP_IMAGE_WIDTH,
              "height-request", POP_IMAGE_HEIGHT,
              End;
          sprintf(spec, "0:%d", 128 + i);
          gtk_widget_set_bg_spec (go, spec);
          go = ToggleButtonObject,
              "child", go,
              End;
          g_signal_connect (go, "clicked", G_CALLBACK (onPatternClicked), priv);
          priv->pattern[i] = go;
          gtk_group_add_child (GTK_GROUP (grp), go, 1, 1);
        }

      types[num] = VCenter(HCenter(grp));
      labels[num] = GCS(MSG_PopImage_Pattern);
      priv->page_type[num] = GTK_POP_IMAGE_PATTERN;
      num++;
    }
#endif
  if (priv->types & GTK_POP_IMAGE_COLOR)
    {
#if GTK_MAJOR_VERSION == 2
      GtkWidget *color_chooser = ColorSelectionObject, 
           "has-palette", TRUE,
           End;
#else
      GtkWidget *color_chooser = gtk_color_chooser_widget_new();
      gtk_color_chooser_set_use_alpha (GTK_COLOR_CHOOSER (color_chooser),
                                       FALSE);
#endif
      priv->color = color_chooser;
      types[num] = color_chooser;
      labels[num] = GCS(MSG_PopImage_Color);
      priv->page_type[num] = GTK_POP_IMAGE_COLOR;
      num++;
    }
  if (priv->types & GTK_POP_IMAGE_EXTERNAL)
    {
      GtkWidget *entry = EntryObject, End;
      GtkWidget *pop = PopSelectionObject,
          "selection-type", GTK_TYPE_FILE_CHOOSER_DIALOG,
          "string", entry,
          "button", GuiPopButton("document-open"),
	  End;
      GtkWidget *image = ExtImageObject,
          "fixed", FALSE,
          "scaled", FALSE,
          End;
      g_signal_connect (entry, "changed", 
              G_CALLBACK(onExternalEntryChanged), image);
      g_signal_connect (entry, "activate", 
              G_CALLBACK(onExternalEntryChanged), image);
      priv->external = entry;

      types[num] = VGroup, 
          "spacing", 0,
          "child", FrameObject,
              "shadow-type", GTK_SHADOW_IN,
              "child", image,
              End,
          "child", pop, "weight", 0,
          End;
      labels[num] = GCS(MSG_PopImage_External);
      priv->page_type[num] = GTK_POP_IMAGE_EXTERNAL;
      num++;
    }

  if (num == 0)
    {
      return;                   /* no types are allowed */
    }

  if (num == 1)
    {
      top = types[0];
    }
  else
    {
      gint i;

      top = priv->notebook = NotebookObject,
          "enable-popup", TRUE,
          End;
      for (i = 0; i < num; i++)
        {
          GtkWidget *align = GroupAlignment(FALSE), 
                    "child", types[i],
                    End;
          gtk_notebook_append_page (GTK_NOTEBOOK (top), 
                  align, GuiLabel(labels[i], NULL, 0.5));
        }
    }
  group = VGroup, "child", top, "weight", 0, End;
  root = GroupAlignment(FALSE), "child", group, End;

  priv->dialog = dialog = DialogObject,
      "title", "",
      "window-position", GTK_WIN_POS_CENTER,
      "transient-for", GuiFindTransientWindow(GTK_WIDGET(pop)),
      "modal", FALSE,
      End;

  vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
  gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
  gtk_widget_show_all(root);

  cancel = GuiSimpleButton(GCS(MSG_Cancel), FALSE);
  gtk_dialog_add_action_widget(GTK_DIALOG(dialog), cancel, GTK_RESPONSE_CANCEL);
  gtk_widget_show(cancel);

  ok = GuiSimpleButton(GCS(MSG_Ok), TRUE);
  gtk_dialog_add_action_widget(GTK_DIALOG(dialog), ok, GTK_RESPONSE_ACCEPT);
  gtk_widget_show(ok);
  gtk_window_set_default(GTK_WINDOW(dialog), ok);

  g_signal_connect(dialog, "response", G_CALLBACK(closePopImageDialog), pop);

  GuiSetDialogButtonOrder(dialog);
  GuiSetDialogGeometryHints(dialog, group);
  GuiAddWindowCallbacks(dialog, cancel);
}

#if 0
    static const GdkColor colors[] = {
	{ 0, 0x9000, 0x9400, 0x9000 },	/* gray (background) */
	{ 0, 0x0000, 0x0000, 0x0000 },  /* black */
	{ 0, 0xf800, 0xfc00, 0xf800 },  /* white */
	{ 0, 0x3800, 0x6400, 0xa000 }, 	/* sky blue */
	{ 0, 0xf800, 0x0000, 0xf800 }, 	/* pink */
	{ 0, 0x0000, 0x0000, 0xf800 },  /* blue */
    };
#endif

struct GetColorReturn
{
    GdkColor fg;
#ifdef ENABLE_GDK_STIPPLED
    GdkColor bg;
    GdkPixmap *pixmap;
    GdkFill fill;
#endif
};

static gboolean 
getColor(GdkWindow *window, const gchar *spec, struct GetColorReturn *gcr)
{
    gulong red, green, blue;
    gulong rgb;
    gint mui;
    gboolean found = FALSE;

#ifdef ENABLE_GDK_STIPPLED
    static const char pat2x2[] = { 0xaa, 0x55 };
    static const char pat4x2[] = { 0xe, 0xb };
    static GdkPixmap *pixmap2x2 = 0, *pixmap4x2 = 0;

    if (!pixmap2x2) {
	pixmap2x2 = gdk_bitmap_create_from_data(0, pat2x2, 2, 2);
    }
    if (!pixmap4x2) {
	pixmap4x2 = gdk_bitmap_create_from_data(0, pat4x2, 4, 2);
    }

    gcr->pixmap = NULL;
    gcr->fill = GDK_SOLID;	// solid rgb
#endif

    if (spec == NULL || spec[0] == 0 || spec[1] != ':') {
        // no action
    } else if (sscanf (spec, "2:r%lx,%lx,%lx", &red, &green, &blue) == 3) {
        gcr->fg.red = (guint16)(red >> 16);
        gcr->fg.green = (guint16)(green >> 16);
        gcr->fg.blue = (guint16)(blue >> 16);
	found = TRUE;
    } else if (sscanf (spec, "2:r%lx", &rgb) == 1) {
	red = (rgb >> 16) & 0xff;
	green = (rgb >> 8) & 0xff;
	blue = rgb & 0xff;
        gcr->fg.red = (guint16)((red << 8) | red);
        gcr->fg.green = (guint16)((green << 8) | green);
        gcr->fg.blue = (guint16)((blue << 8) | blue);
	found = TRUE;
    } else if (sscanf (spec, "2:m%d", &mui) == 1) {
        if (mui >= 0 && mui < MUI_NumColors) {
            gcr->fg = MUIcolors[mui];
            found = TRUE;
        }
    } else if (!strncmp(spec, "0:128", 5)) {
				// background
    } else if (!strncmp(spec, "0:129", 5)) {
	gcr->fg = MUIcolors[MUI_SHADOW];
	found = TRUE;
    } else if (!strncmp(spec, "0:130", 5)) {
	gcr->fg = MUIcolors[MUI_SHINE];
	found = TRUE;
    } else if (!strncmp(spec, "0:131", 5)) {
	gcr->fg = MUIcolors[MUI_HALFSHADOW];
	found = TRUE;
#ifdef ENABLE_GDK_STIPPLED
    } else if (!strncmp(spec, "0:132", 5)) {
	gcr->fg = MUIcolors[MUI_SHADOW];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_STIPPLED;	// stippled black
	found = TRUE;
    } else if (!strncmp(spec, "0:133", 5)) {
	gcr->fg = MUIcolors[MUI_SHADOW];
	gcr->bg = MUIcolors[MUI_HALFSHADOW];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_OPAQUE_STIPPLED; // opaque stippled black on sky blue
	found = TRUE;
    } else if (!strncmp(spec, "0:134", 5)) {
	gcr->fg = MUIcolors[MUI_SHINE];
	gcr->bg = MUIcolors[MUI_SHADOW];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_OPAQUE_STIPPLED; // opaque stippled white on black
	found = TRUE;
    } else if (!strncmp(spec, "0:135", 5)) {
	gcr->fg = MUIcolors[MUI_HALFSHADOW];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_STIPPLED;	// stippled sky blue
	found = TRUE;
    } else if (!strncmp(spec, "0:136", 5)) {
	gcr->fg = MUIcolors[MUI_SHINE];
	gcr->bg = MUIcolors[MUI_HALFSHADOW];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_OPAQUE_STIPPLED; // opaque stippled white on sky blue
	found = TRUE;
    } else if (!strncmp(spec, "0:137", 5)) {
	gcr->fg = MUIcolors[MUI_SHINE];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_STIPPLED; 	// stippled white
	found = TRUE;
    } else if (!strncmp(spec, "0:138", 5)) {
	gcr->fg = MUIcolors[MUI_HALFSHADOW];
	gcr->pixmap = pixmap4x2;
	gcr->fill = GDK_STIPPLED;	// stippled 4x2 sky blue
	found = TRUE;
    } else if (!strncmp(spec, "0:139", 5)) {
	gcr->fg = MUIcolors[MUI_HALFSHINE];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_STIPPLED;	// stippled pink
	found = TRUE;
    } else if (!strncmp(spec, "0:140", 5)) {
	gcr->fg = MUIcolors[MUI_FILL];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_STIPPLED;	// stippled blue
	found = TRUE;
    } else if (!strncmp(spec, "0:141", 5)) {
	gcr->fg = MUIcolors[MUI_HALFSHINE];
	gcr->bg = MUIcolors[MUI_SHINE];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_OPAQUE_STIPPLED; // opaque stippled pink on white
	found = TRUE;
    } else if (!strncmp(spec, "0:142", 5)) {
	gcr->fg = MUIcolors[MUI_HALFSHADOW];
	gcr->bg = MUIcolors[MUI_FILL];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_OPAQUE_STIPPLED; // opaque stippled sky blue on blue
	found = TRUE;
#endif
    } else if (!strncmp(spec, "0:143", 5)) {
	gcr->fg = MUIcolors[MUI_SHINE];
	found = TRUE;
#ifdef ENABLE_GDK_STIPPLED
    } else if (!strncmp(spec, "0:144", 5)) {
	gcr->fg = MUIcolors[MUI_SHINE];
	gcr->bg = MUIcolors[MUI_HALFSHINE];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_OPAQUE_STIPPLED;  // opaque stippled white on pink again
	found = TRUE;
    } else if (!strncmp(spec, "0:145", 5)) {
	gcr->fg = MUIcolors[MUI_SHINE];
	gcr->pixmap = pixmap2x2;
	gcr->fill = GDK_STIPPLED;  // stippled white (again)
	found = TRUE;
#endif
    }

#ifdef ENABLE_GDK_STIPPLED
    if (found && gcr->fill != GDK_SOLID && gcr->pixmap == NULL) {
	found = FALSE;
    }
#endif

    return found;
}

static gboolean bgOnDraw(GtkWidget *widget, cairo_t *cairo)
{
    const gchar *spec = g_object_get_data (G_OBJECT (widget), "bg-spec");
    GtkAllocation alloc;
    gint border = GTK_IS_CONTAINER (widget)? 
        gtk_container_get_border_width(GTK_CONTAINER(widget)) : 0;
    double clip_x1, clip_y1, clip_x2, clip_y2;
    double alloc_x, alloc_y;
    double x1, x2, y1,y2;

    gtk_widget_get_allocation (widget, &alloc);
    cairo_clip_extents(cairo, &clip_x1, &clip_y1, &clip_x2, &clip_y2);
#if GTK_MAJOR_VERSION == 2
    alloc_x = alloc.x;
    alloc_y = alloc.y;
#else
    alloc_x = 0;
    alloc_y = 0;
#endif
    x1 = MAX(alloc_x + border, clip_x1);
    y1 = MAX(alloc_y + border, clip_y1);
    x2 = MIN(alloc_x + alloc.width - border, clip_x2);
    y2 = MIN(alloc_y + alloc.height - border, clip_y2);

    if (x2 > x1 && y2 > y1) {
	GdkWindow *window = gtk_widget_get_window (widget);
        gint dx = x1 - (alloc_x + border);
        gint dy = y1 - (alloc_y + border);

        if (spec && spec[0] == '5' && spec[1] == ':') {
            GdkPixbuf *pixbuf = g_object_get_data(G_OBJECT(widget),"bg-pixbuf");
            if (pixbuf) {
                gint w = gdk_pixbuf_get_width(pixbuf);
                gint h = gdk_pixbuf_get_height(pixbuf);
                if (w > 0 && h > 0) {
                    x1 -= dx % w;
                    y1 -= dy % h;
                    double x, y;
                    for (y = y1; y < y2; y += h) {
                        for (x = x1; x < x2; x += w) {
                            gdk_cairo_set_source_pixbuf(cairo, pixbuf, x, y);
                            cairo_paint(cairo);
                        }
                    }
                }
            }
        } else {
            struct GetColorReturn gcr;
            if (getColor(window, spec, &gcr)) {
                cairo_rectangle(cairo, x1, y1, x2 - x1, y2 - y1);
                cairo_set_source_rgb(cairo, 
                        gcr.fg.red / (double)0xFFFF,
                        gcr.fg.green / (double)0xFFFF,
                        gcr.fg.blue / (double)0xFFFF);
                cairo_fill(cairo);
            }
	}
    }
    return FALSE;
}

#if GTK_MAJOR_VERSION == 2
static gboolean bgOnExposeEvent(GtkWidget *widget, GdkEventExpose *event) 
{
    GdkWindow *window = gtk_widget_get_window (widget);
    cairo_t *cairo;

#ifdef ENABLE_GDK_STIPPLED
    const gchar *spec = g_object_get_data (G_OBJECT (widget), "bg-spec");

    if (!spec || spec[0] != '5' || spec[1] != ':') {
        GtkAllocation alloc;
        gint border = GTK_IS_CONTAINER (widget)? 
            gtk_container_get_border_width(GTK_CONTAINER(widget)) : 0;
        gint x1, x2, y1,y2;

        gtk_widget_get_allocation (widget, &alloc);
        x1 = MAX(alloc.x + border, event->area.x);
        y1 = MAX(alloc.y + border, event->area.y);
        x2 = MIN(alloc.x + alloc.width - border, 
                event->area.x + event->area.width);
        y2 = MIN(alloc.y + alloc.height- border, 
                event->area.y + event->area.height);

        if (x2 > x1 && y2 > y1) {
            gint dx = x1 - (alloc.x + border);
            gint dy = y1 - (alloc.y + border);
            struct GetColorReturn gcr;
            GdkColormap *cmap = gtk_widget_get_colormap(widget);
            GdkGC *gc = gdk_gc_new(window);
            if (getColor(window, spec, &gcr) &&
                    gdk_colormap_alloc_color(cmap, &gcr.fg, TRUE, TRUE) &&
                    gdk_colormap_alloc_color(cmap, &gcr.bg, TRUE, TRUE)) {
                gdk_gc_set_foreground(gc, &gcr.fg);
                gdk_gc_set_background(gc, &gcr.bg);
                if (gcr.pixmap) {
                    gdk_gc_set_stipple(gc, gcr.pixmap);
                }
                gdk_gc_set_fill(gc, gcr.fill);
                x1 -= dx % 4;
                y1 -= dy % 2;
                gdk_draw_rectangle(window, gc, TRUE, x1, y1, x2 - x1, y2 - y1);
            }
            g_object_unref(G_OBJECT(gc));
        }
    }
#endif
    cairo = gdk_cairo_create(window);
    cairo_rectangle(cairo, event->area.x, event->area.y, 
            event->area.width, event->area.height);
    cairo_clip(cairo);
    bgOnDraw(widget, cairo);
    cairo_destroy(cairo);

    return FALSE;
}
#endif

/*
** This is a hack, works only for the widget types coded below
*/

void           
gtk_widget_set_bg_spec (GtkWidget *widget, const gchar *spec)
{
  g_return_if_fail (GTK_IS_WIDGET (widget));

  if (spec != NULL && spec[0] == 0)
    {
      spec = NULL;
    }

  if (GTK_IS_ENTRY (widget) || GTK_IS_TEXT_VIEW (widget) || 
          GTK_IS_TREE_VIEW (widget) ||
          (GTK_IS_BUTTON (widget) && ! GTK_IS_CHECK_BUTTON (widget)))
    {
      gulong red, green, blue;
      gulong rgb;
      gint mui;
      GdkColor col;
      const GdkColor *bg = &col;

      if (spec == NULL)
        {
          bg = NULL;
        }
      else if (sscanf (spec, "2:m%d", &mui) == 1) 
        {
          if (mui >= 0 && mui < MUI_NumColors) 
            {
              col = MUIcolors[mui];
            }
          else
            {
              return;
            }
        }
      else if (sscanf (spec, "2:r%lx,%lx,%lx", &red, &green, &blue) == 3)
        {
          col.red = (guint16)(red >> 16);
          col.green = (guint16)(green >> 16);
          col.blue = (guint16)(blue >> 16);
        }
      else if (sscanf (spec, "2:r%lx", &rgb) == 1)
        {
	  red = (rgb >> 16) & 0xff;
	  green = (rgb >> 8) & 0xff;
	  blue = rgb & 0xff;
	  col.red = (guint16)((red << 8) | red);
	  col.green = (guint16)((green << 8) | green);
	  col.blue = (guint16)((blue << 8) | blue);
	}
      else
        {
          return;
        }

#if GTK_MAJOR_VERSION == 2
      if (GTK_IS_BUTTON (widget))
        {
#ifdef WINDOWS
          GtkWidget *child = gtk_bin_get_child (GTK_BIN (widget));
          gtk_widget_set_bg_spec (child, spec);
#endif
          gtk_widget_modify_bg (widget, GTK_STATE_NORMAL, bg);
        }
      else
        {
          gtk_widget_modify_base (widget, GTK_STATE_NORMAL, bg);
        }
#elif 1
      GdkRGBA color, *bgColor = NULL;
      if (bg != NULL) 
        {
          color.red = (double)bg->red / 0xFFFF;
          color.green = (double)bg->green / 0xFFFF;
          color.blue = (double)bg->blue / 0xFFFF;
          color.alpha = 1.0;
          bgColor = &color;
        }
      gtk_widget_override_background_color (widget, GTK_STATE_FLAG_NORMAL,
                                            bgColor);
      if (GTK_IS_BUTTON (widget))
        {
          GtkWidget *child = gtk_bin_get_child (GTK_BIN (widget));
          gtk_widget_set_bg_spec (child, spec);
        }
#if 0 // GTK_MINOR_VERSION >= 6
      else if (bg != NULL && GTK_IS_ENTRY (widget))
        {
          PangoField *field = pango_field_background_new (bg->red, bg->green,
                                                            bg->blue);
          PangoFieldList *fieldList = pango_field_list_new ();
          pango_field_list_insert (fieldList, field);
          gtk_entry_set_fields (GTK_ENTRY (widget), fieldList);
          pango_field_list_unref (fieldList);
        }
#endif
#else
      // How can we clear a previously set background?
      if (bg != NULL)
        {
          GString *settings = g_string_new (NULL);
          g_string_printf (settings, 
                  "* { background-color: #%02x%02x%02x }",
                  bg->red >> 8, bg->green >> 8, bg->blue >> 8);
          GtkCssProvider *provider = gtk_css_provider_new ();
          gtk_css_provider_load_from_data (provider, settings->str, -1, NULL);
          g_string_free (settings, TRUE);
          GtkStyleContext *context = gtk_widget_get_style_context (widget);
          gtk_style_context_add_provider(context, GTK_STYLE_PROVIDER (provider),
                  GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
          g_object_unref (G_OBJECT (provider));
        }
#endif
    }
  else
    {
      gulong id = GPOINTER_TO_INT(g_object_get_data (G_OBJECT (widget), 
                                                     "bg-handler"));
      gchar *copy = g_strdup (spec);
  
      g_object_set_data_full (G_OBJECT (widget), "bg-spec", copy, g_free);
      if (spec)
        {
          if (id == 0)
            {
#if GTK_MAJOR_VERSION == 2
              id = g_signal_connect(widget, "expose-event", 
                      G_CALLBACK(bgOnExposeEvent), NULL);
#else
              id = g_signal_connect(widget, "draw", G_CALLBACK(bgOnDraw), NULL);
#endif
              g_object_set_data (G_OBJECT (widget), "bg-handler", 
                      GINT_TO_POINTER(id));
            }
          if (spec[0] == '5' && spec[1] == ':') 
            {
              gchar *path = gtkExpandPathname(&spec[2]);
              GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(path, 0);
              g_free(path);
              g_object_set_data_full (G_OBJECT (widget), "bg-pixbuf", pixbuf,
                      g_object_unref);
            }
        }
      else if (id != 0)
        {
          g_signal_handler_disconnect (G_OBJECT (widget), id);
          g_object_set_data (G_OBJECT (widget), "bg-handler", 
                  GINT_TO_POINTER(0));
        }
    }

  gtk_widget_queue_draw (widget);
}

