/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkpopselection.c
**
** an implementation of MUI's popasl class for gtk+-2.0/3.0
**
** Creation date: 21.05.2004
** Last modified: 22.08.2017
*/

#define ENABLE_NLS

#include "gtkpopselection.h"
#include <gtk/gtkAmiFileChooser.h>
#include <gtk/gtkUtil.h>
#include <glib/gi18n.h>

#define P_(x) N_(x)

enum {
  PROP_0,
  PROP_SELECTION_TYPE,
  PROP_FILE_CHOOSER_ACTION
};

enum {
  START,
  STOP,
  LAST_SIGNAL
};

static void     gtk_pop_selection_class_init  (GtkPopSelectionClass *klass);
static void     gtk_pop_selection_init        (GtkPopSelection      *pop);
static void     gtk_pop_selection_set_property(GObject              *object,
                                               guint                 prop_id,
                                               const GValue         *value,
                                               GParamSpec           *pspec);
static void     gtk_pop_selection_get_property(GObject              *object,
                                               guint                 prop_id,
                                               GValue               *value,
                                               GParamSpec           *pspec);
static gboolean gtk_pop_selection_open        (GtkPopString         *pop,
	                                       gpointer              data);
static void     gtk_pop_selection_close       (GtkPopString         *pop,
	                                       gboolean        	     success,
					       gpointer              data);
static gboolean gtk_pop_selection_start       (GtkPopSelection      *pop,
	                                       gpointer              data);
static void     gtk_pop_selection_stop        (GtkPopSelection      *pop,
	                                       gpointer              data);

static GtkPopStringClass *parent_class = NULL;

static guint gtk_pop_selection_signals[LAST_SIGNAL] = { 0 };

static void
g_cclosure_marshal_BOOLEAN__VOID (
                               GClosure     *closure,
                               GValue       *return_value G_GNUC_UNUSED,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint G_GNUC_UNUSED,
                               gpointer      marshal_data)
{
  typedef gboolean (*GMarshalFunc_BOOLEAN__VOID) (gpointer     data1,
  					          gpointer     data2);
  GMarshalFunc_BOOLEAN__VOID callback;
  GCClosure *cc = (GCClosure*) closure;
  gpointer data1, data2;
  gboolean v_return;

  g_return_if_fail (n_param_values == 1);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }
  callback = (GMarshalFunc_BOOLEAN__VOID) (marshal_data ? 
          marshal_data : cc->callback);

  v_return = callback (data1, data2);
  g_value_set_boolean (return_value, v_return);
}

GType
gtk_pop_selection_get_type (void)
{
  static GType pop_selection_type = 0;

  if (!pop_selection_type)
    {
      static const GTypeInfo pop_selection_info =
      {
	sizeof (GtkPopSelectionClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_pop_selection_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkPopSelection),
	0,		/* n_preallocs */
	(GInstanceInitFunc) gtk_pop_selection_init,
      };

      pop_selection_type = g_type_register_static (GTK_TYPE_POP_STRING, 
	      "GtkPopSelection", &pop_selection_info, 0);
    }

  return pop_selection_type;
}

static void
gtk_pop_selection_class_init (GtkPopSelectionClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GtkPopStringClass *pop_string_class = GTK_POP_STRING_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = gtk_pop_selection_set_property;
  gobject_class->get_property = gtk_pop_selection_get_property;

  pop_string_class->open = gtk_pop_selection_open;
  pop_string_class->close = gtk_pop_selection_close;

  klass->start = gtk_pop_selection_start;
  klass->stop  = gtk_pop_selection_stop;

  g_object_class_install_property (gobject_class,
				   PROP_SELECTION_TYPE,
				   g_param_spec_ulong ("selection-type",
						     P_("Selection type"),
#if GTK_MAJOR_VERSION == 2
						     P_("Type of the selection box (either GTK_TYPE_FILE_CHOOSER_DIALOG or GTK_TYPE_FONT_SELECTION_DIALOG)"),
#else
						     P_("Type of the selection box (either GTK_TYPE_FILE_CHOOSER_DIALOG or GTK_TYPE_FONT_CHOOSER_DIALOG)"),
#endif
						     0,
						     G_MAXULONG,
						     0,
						     G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_FILE_CHOOSER_ACTION,
                                   g_param_spec_enum ("file-chooser-action",
                                                   P_("File chooser action"),
						   P_("The type of operation that the file chooser is performing"),
						   GTK_TYPE_FILE_CHOOSER_ACTION,
						   GTK_FILE_CHOOSER_ACTION_OPEN,
						   G_PARAM_READWRITE));

  gtk_pop_selection_signals[START] =
    g_signal_new ("start",
	          G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		  G_STRUCT_OFFSET (GtkPopSelectionClass, start),
		  NULL, NULL,
		  g_cclosure_marshal_BOOLEAN__VOID,
		  G_TYPE_BOOLEAN, 0);
  gtk_pop_selection_signals[STOP] =
    g_signal_new ("stop",
	          G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		  G_STRUCT_OFFSET (GtkPopSelectionClass, stop),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__VOID,
		  G_TYPE_NONE, 0);
}

static void
gtk_pop_selection_set_property (GObject         *object,
                                guint            prop_id,
                                const GValue    *value,
                                GParamSpec      *pspec)
{
  GtkPopSelection *pop = GTK_POP_SELECTION (object);

  switch (prop_id)
    {
    case PROP_SELECTION_TYPE:
      gtk_pop_selection_set_selection_type (pop, g_value_get_ulong (value));
      break;
    case PROP_FILE_CHOOSER_ACTION:
      gtk_pop_selection_set_file_chooser_action (pop, g_value_get_enum (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
gtk_pop_selection_get_property (GObject         *object,
                                guint            prop_id,
                                GValue          *value,
                                GParamSpec      *pspec)
{
  GtkPopSelection *pop = GTK_POP_SELECTION (object);

  switch (prop_id)
    {
    case PROP_SELECTION_TYPE:
      g_value_set_ulong (value, pop->type);
      break;
    case PROP_FILE_CHOOSER_ACTION:
      g_value_set_enum (value, pop->file_chooser_action);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
gtk_pop_selection_init (GtkPopSelection *pop)
{
  g_object_set (G_OBJECT (pop), "follow", FALSE, NULL);
  pop->file_chooser_action = GTK_FILE_CHOOSER_ACTION_OPEN;
}

GtkWidget*
gtk_pop_selection_new (void)
{
  return g_object_new (GTK_TYPE_POP_SELECTION, NULL);
}

void
gtk_pop_selection_set_selection_type (GtkPopSelection *pop, GType type)
{
  g_return_if_fail (GTK_IS_POP_SELECTION (pop));
  g_return_if_fail (type == 0 || 
	  type == GTK_TYPE_FILE_CHOOSER_DIALOG || 
#if GTK_MAJOR_VERSION == 2
	  type == GTK_TYPE_FONT_SELECTION_DIALOG
#else
	  type == GTK_TYPE_FONT_CHOOSER_DIALOG
#endif
          );

  if (pop->type != type)
    {
      /* destroy any open selection box */
      gtk_pop_string_set_popup (GTK_POP_STRING (pop), NULL);

      pop->type = type;

      g_object_notify (G_OBJECT (pop), "selection-type");
    }
}

GType
gtk_pop_selection_get_selection_type (GtkPopSelection *pop)
{
  g_return_val_if_fail (GTK_IS_POP_SELECTION (pop), 0);

  return pop->type;
}

void
gtk_pop_selection_set_file_chooser_action (GtkPopSelection *pop, 
                                           GtkFileChooserAction action)
{
  g_return_if_fail (GTK_IS_POP_SELECTION (pop));

  if (pop->file_chooser_action != action)
    {
      GtkWidget *popup;

      popup = gtk_pop_string_get_popup (GTK_POP_STRING (pop));
      if (popup && GTK_IS_FILE_CHOOSER_DIALOG(popup))
        {
          gtk_file_chooser_set_action (GTK_FILE_CHOOSER (popup), action);
        }

      pop->file_chooser_action = action;

      g_object_notify (G_OBJECT (pop), "file-chooser-action");
    }
}

GtkFileChooserAction gtk_pop_selection_get_file_chooser_action (
                                                        GtkPopSelection *pop)
{
  g_return_val_if_fail (GTK_IS_POP_SELECTION (pop), 
          GTK_FILE_CHOOSER_ACTION_OPEN);

  return pop->file_chooser_action;
}

GtkFileChooserDialog* 
gtk_pop_selection_get_file_chooser_dialog (GtkPopSelection *pop)
{
  GtkWidget *popup;

  g_return_val_if_fail (GTK_IS_POP_SELECTION (pop), NULL);
  g_return_val_if_fail (pop->type == GTK_TYPE_FILE_CHOOSER_DIALOG, NULL);

  popup = gtk_pop_string_get_popup (GTK_POP_STRING (pop));
  if (popup)
    {
      return GTK_FILE_CHOOSER_DIALOG (popup);
    }

  return NULL;
}

#if GTK_MAJOR_VERSION == 2
GtkFontSelectionDialog* 
gtk_pop_selection_get_font_selection_dialog (GtkPopSelection *pop)
{
  GtkWidget *popup;

  g_return_val_if_fail (GTK_IS_POP_SELECTION (pop), NULL);
  g_return_val_if_fail (pop->type == GTK_TYPE_FONT_SELECTION_DIALOG, NULL);

  popup = gtk_pop_string_get_popup (GTK_POP_STRING (pop));
  if (popup)
    {
      return GTK_FONT_SELECTION_DIALOG (popup);
    }

  return NULL;
}
#else
GtkFontChooserDialog* 
gtk_pop_selection_get_font_chooser_dialog (GtkPopSelection *pop)
{
  GtkWidget *popup;

  g_return_val_if_fail (GTK_IS_POP_SELECTION (pop), NULL);
  g_return_val_if_fail (pop->type == GTK_TYPE_FONT_CHOOSER_DIALOG, NULL);

  popup = gtk_pop_string_get_popup (GTK_POP_STRING (pop));
  if (popup)
    {
      return GTK_FONT_CHOOSER_DIALOG (popup);
    }

  return NULL;
}
#endif

static void
gtk_pop_selection_on_response (GtkPopString *pop, gint response) 
{
  gboolean accept;

  switch (response)
    {
      case GTK_RESPONSE_ACCEPT:
      case GTK_RESPONSE_OK:
      case GTK_RESPONSE_YES:
      case GTK_RESPONSE_APPLY:
	  accept = TRUE;
	  break;
      default:
	  accept = FALSE;
	  break;
    }

  gtk_pop_string_close_popup (pop, accept);
}

static gboolean
gtk_pop_selection_open (GtkPopString *pop, gpointer data)
{
  gboolean success = TRUE;
  GtkWidget *popup = gtk_pop_string_get_popup (pop);
  GtkWidget *toplevel = gtk_widget_get_toplevel (GTK_WIDGET (pop));

  if (popup == NULL)
    {
      GType type = GTK_POP_SELECTION (pop)->type;

      if (type == GTK_TYPE_FILE_CHOOSER_DIALOG)
	{
	  popup = gtk_ami_file_chooser_new (P_("Choose file"), P_("Ok"), 
		  NULL, GTK_POP_SELECTION(pop)->file_chooser_action, NULL);
	}
#if GTK_MAJOR_VERSION == 2
      else if (type == GTK_TYPE_FONT_SELECTION_DIALOG)
        {
	  GtkFontSelectionDialog *dialog;
          GtkWidget *cancel;

	  popup = gtk_font_selection_dialog_new (P_("Choose font"));
	  dialog = GTK_FONT_SELECTION_DIALOG (popup);
          cancel = gtk_font_selection_dialog_get_cancel_button (dialog);
	  /* GTK_WIDGET_SET_FLAGS (dialog->ok_button, GTK_CAN_DEFAULT); */
	  GuiAddWindowCallbacks (popup, cancel);
	}
#else
      else if (type == GTK_TYPE_FONT_CHOOSER_DIALOG)
        {
	  GtkDialog *dialog;
          GtkWidget *cancel;

          popup = gtk_font_chooser_dialog_new (P_("Choose font"), NULL);
          dialog = GTK_DIALOG (popup);
          cancel = gtk_dialog_get_widget_for_response (dialog,
                                                       GTK_RESPONSE_CANCEL);
          GuiAddWindowCallbacks (popup, cancel);
        }
#endif
      else
	{
	  return FALSE;
	}

      g_signal_connect_swapped (popup, "response", 
	      G_CALLBACK(gtk_pop_selection_on_response), pop);

      gtk_pop_string_set_popup (pop, popup);
    }

  g_signal_emit (pop, gtk_pop_selection_signals[START], 0, &success);

  if (!success)
    {
      gtk_pop_selection_close (pop, FALSE, NULL);
      
      return FALSE;
    }

  /* If the window of the pop selection object is modal, also make the popup modal. */
  gtk_window_set_modal (GTK_WINDOW (popup),
                        gtk_window_get_modal (GTK_WINDOW (toplevel)));
  gtk_window_set_transient_for (GTK_WINDOW (popup), GTK_WINDOW (toplevel));

  gtk_widget_show (popup);

  return TRUE;
}

static void
gtk_pop_selection_close (GtkPopString *pop, gboolean success, gpointer data)
{
  if (success) 
    {
      g_signal_emit (pop, gtk_pop_selection_signals[STOP], 0);
    }

  gtk_pop_string_set_popup (pop, NULL);
}

static gboolean 
gtk_pop_selection_start (GtkPopSelection *pop, gpointer data)
{
  const gchar *text = gtk_pop_string_get_text (GTK_POP_STRING (pop));

  if (pop->type == GTK_TYPE_FILE_CHOOSER_DIALOG)
    {
      if (text != NULL)
	{
	  GtkFileChooserDialog *dialog = 
	      gtk_pop_selection_get_file_chooser_dialog (pop);

	  gtk_ami_file_chooser_set_filename (GTK_FILE_CHOOSER (dialog), text);
	}

      return TRUE;
    }
#if GTK_MAJOR_VERSION == 2
  else if (pop->type == GTK_TYPE_FONT_SELECTION_DIALOG)
    {
      if (text != NULL && text[0] != 0)
	{
	  GtkFontSelectionDialog *dialog = 
	      gtk_pop_selection_get_font_selection_dialog (pop);

	  gtk_font_selection_dialog_set_font_name (dialog, text);
	}

      return TRUE;
    }
#else
  else if (pop->type == GTK_TYPE_FONT_CHOOSER_DIALOG)
    {
      if (text != NULL && text[0] != 0)
	{
	  GtkFontChooserDialog *dialog = 
	      gtk_pop_selection_get_font_chooser_dialog (pop);

	  gtk_font_chooser_set_font (GTK_FONT_CHOOSER (dialog), text);
	}

      return TRUE;
    }
#endif
  else
    {
      return FALSE;
    }
}

static void 
gtk_pop_selection_stop (GtkPopSelection *pop, gpointer data)
{
  gchar *text = NULL;

  if (pop->type == GTK_TYPE_FILE_CHOOSER_DIALOG)
    {
      GtkFileChooserDialog *dialog = 
	  gtk_pop_selection_get_file_chooser_dialog (pop);

      text = gtk_ami_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    }
#if GTK_MAJOR_VERSION == 2
  else if (pop->type == GTK_TYPE_FONT_SELECTION_DIALOG)
    {
      GtkFontSelectionDialog *dialog = 
	  gtk_pop_selection_get_font_selection_dialog (pop);

      text = gtk_font_selection_dialog_get_font_name (dialog);
    }
#else
  else if (pop->type == GTK_TYPE_FONT_CHOOSER_DIALOG)
    {
      GtkFontChooserDialog *dialog = 
	  gtk_pop_selection_get_font_chooser_dialog (pop);

      text = gtk_font_chooser_get_font (GTK_FONT_CHOOSER (dialog));
    }
#endif

  gtk_pop_string_set_text (GTK_POP_STRING (pop), text);

  g_free (text);
}

