/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkpopstring.c
**
** an implementation of MUI's popstring class for gtk+-2.0
**
** Creation date: 19.05.2004
** Last modified: 05.09.2012
*/

#include "gtkpopstring.h"
#include <gdk/gdkkeysyms.h>
#include <glib/gi18n.h>

#define P_(x) N_(x)

enum {
  PROP_0,
  PROP_STRING,
  PROP_BUTTON,
  PROP_TOGGLE,
  PROP_FOLLOW,
  PROP_VOLATILE,
  PROP_POPUP,
  PROP_TEXT
};

enum {
  OPEN,
  CLOSE,
  LAST_SIGNAL
};

static void gtk_pop_string_class_init     (GtkPopStringClass *klass);
static void gtk_pop_string_init           (GtkPopString      *pop);
#if GTK_MAJOR_VERSION == 2
static void gtk_pop_string_destroy        (GtkObject         *object);
#else
static void gtk_pop_string_destroy        (GtkWidget         *widget);
#endif
static void gtk_pop_string_set_property   (GObject           *object,
                                           guint              prop_id,
                                           const GValue      *value,
                                           GParamSpec        *pspec);
static void gtk_pop_string_get_property   (GObject           *object,
                                           guint              prop_id,
                                           GValue            *value,
                                           GParamSpec        *pspec);
static void gtk_pop_string_dimension_popup(GtkPopString *pop);
static gboolean gtk_pop_string_popup_follow(GtkPopString *pop);
static void gtk_pop_string_map            (GtkPopString *pop);
static void gtk_pop_string_unmap          (GtkPopString *pop);
static void gtk_pop_string_show           (GtkPopString *pop);
static void gtk_pop_string_hide           (GtkPopString *pop);
    
#if GTK_MAJOR_VERSION == 2
static GtkHBoxClass *parent_class = NULL;
#define PARENT_TYPE GTK_TYPE_HBOX
#else
static GtkBoxClass *parent_class = NULL;
#define PARENT_TYPE GTK_TYPE_BOX
#endif

static guint gtk_pop_string_signals[LAST_SIGNAL] = { 0 };

static void
g_cclosure_marshal_BOOLEAN__VOID (GClosure     *closure,
                               GValue       *return_value G_GNUC_UNUSED,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint G_GNUC_UNUSED,
                               gpointer      marshal_data)
{
  typedef gboolean (*GMarshalFunc_BOOLEAN__VOID) (gpointer     data1,
  					          gpointer     data2);
  GMarshalFunc_BOOLEAN__VOID callback;
  GCClosure *cc = (GCClosure*) closure;
  gpointer data1, data2;
  gboolean v_return;

  g_return_if_fail (n_param_values == 1);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }
  callback = (GMarshalFunc_BOOLEAN__VOID) (marshal_data ? 
          marshal_data : cc->callback);

  v_return = callback (data1, data2);
  g_value_set_boolean (return_value, v_return);
}

GType
gtk_pop_string_get_type (void)
{
  static GType pop_string_type = 0;

  if (!pop_string_type)
    {
      static const GTypeInfo pop_string_info =
      {
	sizeof (GtkPopStringClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_pop_string_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkPopString),
	0,		/* n_preallocs */
	(GInstanceInitFunc) gtk_pop_string_init,
      };

      pop_string_type = g_type_register_static (PARENT_TYPE, "GtkPopString",
					        &pop_string_info, 0);
    }

  return pop_string_type;
}

static void
gtk_pop_string_class_init (GtkPopStringClass *klass)
{
  GObjectClass *gobject_class;
#if GTK_MAJOR_VERSION == 2
  GtkObjectClass *object_class;
#else
  GtkWidgetClass *widget_class;
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = gtk_pop_string_set_property;
  gobject_class->get_property = gtk_pop_string_get_property;

#if GTK_MAJOR_VERSION == 2
  object_class = GTK_OBJECT_CLASS (klass);
  object_class->destroy = gtk_pop_string_destroy;
#else
  widget_class = GTK_WIDGET_CLASS (klass);
  widget_class->destroy = gtk_pop_string_destroy;
#endif

  g_object_class_install_property (gobject_class,
				   PROP_STRING,
				   g_param_spec_object ("string",
						     P_("String"),
						     P_("Widget that contains string to display edit (GtkEntry, GtkLabel, or GtkButton)"),
						     GTK_TYPE_WIDGET,
						     G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
				   PROP_BUTTON,
				   g_param_spec_object ("button",
						     P_("Button"),
						     P_("Button for opening the popup window"),
						     GTK_TYPE_BUTTON,
						     G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
				   PROP_TOGGLE,
				   g_param_spec_boolean ("toggle",
						      P_("Toggle"),
						      P_("Whether the button can open and close the popup or only open it"),
						     TRUE,
						     G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_FOLLOW,
				   g_param_spec_boolean ("follow",
						      P_("Follow"),
						      P_("Whether the popup should follow when the pop-strings's window is moved or resized"),
						     TRUE,
						     G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_VOLATILE,
				   g_param_spec_boolean ("volatile",
						      P_("Volatile"),
						      P_("Whether an opened popup should show/hide when the pop-string object appears/disappears"),
						     TRUE,
						     G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_POPUP,
				   g_param_spec_object ("popup",
						     P_("Popup"),
						     P_("Popup window opened by the popup button"),
						     GTK_TYPE_WINDOW,
						     G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
				   PROP_TEXT,
				   g_param_spec_string ("text",
						     P_("Text"),
						     P_("Text to be displayed in the string field"),
						     NULL,
						     G_PARAM_READWRITE));


  gtk_pop_string_signals[OPEN] =
    g_signal_new ("open",
	          G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_LAST, // | G_SIGNAL_ACTION,
		  G_STRUCT_OFFSET (GtkPopStringClass, open),
		  NULL, NULL,
		  g_cclosure_marshal_BOOLEAN__VOID,
		  G_TYPE_BOOLEAN, 0);
  gtk_pop_string_signals[CLOSE] =
    g_signal_new ("close",
	          G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		  G_STRUCT_OFFSET (GtkPopStringClass, close),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__BOOLEAN,
		  G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
}

static void
gtk_pop_string_init (GtkPopString *pop)
{
  pop->toggle = TRUE;
  pop->follow = TRUE;
  pop->volat = TRUE;

  gtk_box_set_spacing (GTK_BOX (pop), 0);
#if GTK_MAJOR_VERSION > 2
  gtk_orientable_set_orientation (GTK_ORIENTABLE (pop),
                                  GTK_ORIENTATION_HORIZONTAL);
#endif

  g_signal_connect (pop, "map", G_CALLBACK (gtk_pop_string_map), NULL);
  g_signal_connect (pop, "unmap", G_CALLBACK (gtk_pop_string_unmap), NULL);
}

#if GTK_MAJOR_VERSION == 2
static void
gtk_pop_string_destroy (GtkObject *object)
{
  GtkPopString *pop = GTK_POP_STRING (object);

  gtk_pop_string_set_popup (pop, NULL);
  
  (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}
#else
static void
gtk_pop_string_destroy (GtkWidget *widget)
{
  GtkPopString *pop = GTK_POP_STRING (widget);

  gtk_pop_string_set_popup (pop, NULL);
  
  (* GTK_WIDGET_CLASS (parent_class)->destroy) (widget);
}
#endif

static void
gtk_pop_string_set_property (GObject         *object,
                             guint            prop_id,
                             const GValue    *value,
                             GParamSpec      *pspec)
{
  GtkPopString *pop = GTK_POP_STRING (object);

  switch (prop_id)
    {
    case PROP_STRING:
      gtk_pop_string_set_string (pop, g_value_get_object (value));
      break;
    case PROP_BUTTON:
      gtk_pop_string_set_button (pop, g_value_get_object (value));
      break;
    case PROP_TOGGLE:
      gtk_pop_string_set_toggle (pop, g_value_get_boolean (value));
      break;
    case PROP_FOLLOW:
      gtk_pop_string_set_follow (pop, g_value_get_boolean (value));
      break;
    case PROP_VOLATILE:
      gtk_pop_string_set_volatile (pop, g_value_get_boolean (value));
      break;
    case PROP_POPUP:
      gtk_pop_string_set_popup (pop, g_value_get_object (value));
      break;
    case PROP_TEXT:
      gtk_pop_string_set_text (pop, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
gtk_pop_string_get_property (GObject         *object,
                             guint            prop_id,
                             GValue          *value,
                             GParamSpec      *pspec)
{
  GtkPopString *pop = GTK_POP_STRING (object);

  switch (prop_id)
    {
    case PROP_STRING:
      g_value_set_object (value, pop->string);
      break;
    case PROP_BUTTON:
      g_value_set_object (value, pop->button);
      break;
    case PROP_FOLLOW:
      g_value_set_boolean (value, pop->follow);
      break;
    case PROP_VOLATILE:
      g_value_set_boolean (value, pop->volat);
      break;
    case PROP_TOGGLE:
      g_value_set_boolean (value, pop->toggle);
      break;
    case PROP_POPUP:
      g_value_set_object (value, pop->popup);
      break;
    case PROP_TEXT:
      g_value_set_string (value, gtk_pop_string_get_text (pop));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

GtkWidget*
gtk_pop_string_new (void)
{
  return g_object_new (GTK_TYPE_POP_STRING, NULL);
}

static gboolean 
gtk_pop_string_on_key_press (GtkPopString *pop, GdkEventKey *event)
{
  if ((event->keyval == GDK_KEY_P || event->keyval == GDK_KEY_p) && 
	  event->state == GDK_CONTROL_MASK) 
    {
      gtk_pop_string_open_popup (pop);
      
      return TRUE;
    }

  return FALSE;
}

void       
gtk_pop_string_set_string (GtkPopString *pop, GtkWidget *string)
{
  g_return_if_fail (GTK_IS_POP_STRING (pop));
  g_return_if_fail (string == NULL || GTK_IS_ENTRY (string) ||
	  GTK_IS_LABEL (string) || GTK_IS_CONTAINER (string));

  if (pop->string != string)
    {
      if (string)
        {
	  gtk_box_pack_start (GTK_BOX (pop), string, TRUE, TRUE, 0);
	  g_signal_connect_swapped (string, "key-press-event",
		  G_CALLBACK(gtk_pop_string_on_key_press), pop);
	}
      if (pop->string)
	{
	  g_signal_handlers_disconnect_by_func (pop->string,
		  gtk_pop_string_on_key_press, pop);
	  gtk_container_remove (GTK_CONTAINER (pop), pop->string);
	}
      pop->string = string;
      g_object_notify (G_OBJECT (pop), "string");
    }
}

GtkWidget* 
gtk_pop_string_get_string (GtkPopString *pop)
{
  g_return_val_if_fail (GTK_IS_POP_STRING (pop), NULL);

  return pop->string;
}

void       
gtk_pop_string_set_button (GtkPopString *pop, GtkWidget *button)
{
  g_return_if_fail (GTK_IS_POP_STRING (pop));
  g_return_if_fail (button == NULL || GTK_IS_BUTTON (button));

  if (pop->button != button)
    {
      if (button)
        {
	  gtk_box_pack_end (GTK_BOX (pop), button, FALSE, FALSE, 0);
	  g_signal_connect_swapped (button, "clicked",
		  G_CALLBACK(gtk_pop_string_open_popup), pop);
	}
      if (pop->button)
	{
	  g_signal_handlers_disconnect_by_func (pop->button,
		  gtk_pop_string_open_popup, pop);
	  gtk_container_remove (GTK_CONTAINER (pop), pop->button);
	}
      pop->button = button;
      g_object_notify (G_OBJECT (pop), "button");
    }
}

GtkWidget* 
gtk_pop_string_get_button (GtkPopString *pop)
{
  g_return_val_if_fail (GTK_IS_POP_STRING (pop), NULL);

  return pop->button;
}

static void
gtk_pop_string_check_button_sensitive (GtkPopString *pop)
{
  if (pop->button)
    {
      gtk_widget_set_sensitive (pop->button,
	      pop->toggle || pop->popup == NULL || 
	      !gtk_widget_get_visible (pop->popup));
    }
}

void
gtk_pop_string_set_toggle (GtkPopString *pop, gboolean toggle)
{
  g_return_if_fail (GTK_IS_POP_STRING (pop));

  toggle = toggle != FALSE;

  if (toggle != pop->toggle)
    {
      pop->toggle = toggle;

      gtk_pop_string_check_button_sensitive (pop);

      g_object_notify (G_OBJECT (pop), "toggle");
    }
}

gboolean
gtk_pop_string_get_toggle (GtkPopString *pop)
{
  g_return_val_if_fail (GTK_IS_POP_STRING (pop), TRUE);

  return pop->toggle;
}

void
gtk_pop_string_set_follow (GtkPopString *pop, gboolean follow)
{
  g_return_if_fail (GTK_IS_POP_STRING (pop));

  follow = follow != FALSE;

  if (follow != pop->follow)
    {
      pop->follow = follow;

      g_object_notify (G_OBJECT (pop), "follow");
    }
}

gboolean
gtk_pop_string_get_follow (GtkPopString *pop)
{
  g_return_val_if_fail (GTK_IS_POP_STRING (pop), TRUE);

  return pop->follow;
}

void
gtk_pop_string_set_volatile (GtkPopString *pop, gboolean volat)
{
  g_return_if_fail (GTK_IS_POP_STRING (pop));

  volat = volat != FALSE;

  if (volat != pop->volat)
    {
      pop->volat = volat;

      g_object_notify (G_OBJECT (pop), "volatile");
    }
}

gboolean
gtk_pop_string_get_volatile (GtkPopString *pop)
{
  g_return_val_if_fail (GTK_IS_POP_STRING (pop), TRUE);

  return pop->volat;
}

void       
gtk_pop_string_set_popup (GtkPopString *pop, GtkWidget *popup)
{
  g_return_if_fail (GTK_IS_POP_STRING (pop));
  g_return_if_fail (popup == NULL || GTK_IS_WINDOW (popup));

  if (pop->popup != popup)
    {
      GtkWidget *toplevel = gtk_widget_get_toplevel (GTK_WIDGET (pop));

      if (pop->popup)
	{
	  gtk_widget_destroy (pop->popup);
	  pop->popup = NULL;
	}
      if (popup)
        {
	  if (GTK_IS_WINDOW (toplevel))
	    {
	      GtkWindowGroup *group = NULL;
             
              if (gtk_window_has_group (GTK_WINDOW (toplevel)))
                { 
                    group = gtk_window_get_group (GTK_WINDOW(toplevel));
                }
              else if (gtk_window_get_modal (GTK_WINDOW (toplevel)))
	        {
		  group = gtk_window_group_new ();
		  gtk_window_group_add_window (group, GTK_WINDOW (toplevel));
		  g_object_unref (G_OBJECT (group));
		}
	      if (group != NULL)
		{
		  gtk_window_group_add_window (group, GTK_WINDOW (popup));
		}
	      gtk_window_set_transient_for (GTK_WINDOW (popup), 
		      GTK_WINDOW (toplevel));
	    }
	  pop->popup = popup;

          /* In case someone breaks the contract and destroys our popup */
          g_object_add_weak_pointer (G_OBJECT (popup),(gpointer *)&pop->popup);

	  g_signal_connect_swapped (popup, "show",
		  G_CALLBACK(gtk_pop_string_dimension_popup), pop);
	  gtk_pop_string_dimension_popup (pop);
	}
      g_object_notify (G_OBJECT (pop), "popup");
    }
}

GtkWidget* 
gtk_pop_string_get_popup (GtkPopString *pop)
{
  g_return_val_if_fail (GTK_IS_POP_STRING (pop), NULL);

  return pop->popup;
}

static void
gtk_pop_string_set_text_internal (GtkWidget *string, const gchar *text)
{
  if (GTK_IS_LABEL (string))
    {
      gtk_label_set_text (GTK_LABEL (string), text);
    }
  else if (GTK_IS_ENTRY (string))
    {
      gtk_entry_set_text (GTK_ENTRY (string), text? text : "");
    }
  else if (GTK_IS_CONTAINER (string))
    {
      gtk_container_foreach (GTK_CONTAINER (string), 
	      (GtkCallback)gtk_pop_string_set_text_internal, (gpointer)text);
    }
}

void
gtk_pop_string_set_text (GtkPopString *pop, const gchar *text)
{
  g_return_if_fail (GTK_IS_POP_STRING (pop));
  g_return_if_fail (pop->string != NULL);

  gtk_pop_string_set_text_internal (pop->string, text? text : "");
}

static void
gtk_pop_string_get_text_internal (GtkWidget *string, const gchar **text)
{
  if (GTK_IS_LABEL (string))
    {
      *text = gtk_label_get_text (GTK_LABEL (string));
    }
  else if (GTK_IS_ENTRY (string))
    {
      *text = gtk_entry_get_text (GTK_ENTRY (string));
    }
  else if (GTK_IS_CONTAINER (string))
    {
      gtk_container_foreach (GTK_CONTAINER (string), 
	      (GtkCallback)gtk_pop_string_get_text_internal, (gpointer)text);
    }
}

G_CONST_RETURN gchar *
gtk_pop_string_get_text (GtkPopString *pop)
{
  const gchar *text = NULL;

  g_return_val_if_fail (GTK_IS_POP_STRING (pop), NULL);
  g_return_val_if_fail (pop->string != NULL, NULL);

  gtk_pop_string_get_text_internal (pop->string, &text);

  return text;
}

void       
gtk_pop_string_open_popup (GtkPopString *pop)
{
  g_return_if_fail (GTK_IS_POP_STRING (pop));

  if (pop->popup && gtk_widget_get_visible(pop->popup))
    {
      gtk_pop_string_close_popup (pop, FALSE);
    }
  else
    {
      gboolean success = FALSE;

      g_signal_emit (pop, gtk_pop_string_signals[OPEN], 0, &success);

      gtk_pop_string_check_button_sensitive (pop);
    }
}

void
gtk_pop_string_close_popup (GtkPopString *pop, gboolean success)
{
  g_return_if_fail (GTK_IS_POP_STRING (pop));

  g_signal_emit (pop, gtk_pop_string_signals[CLOSE], 0, success);

  gtk_pop_string_check_button_sensitive (pop);
}

static void
gtk_pop_string_dimension_popup (GtkPopString *pop)
{
  if (gtk_widget_get_visible (GTK_WIDGET (pop)) && pop->popup != NULL)
    {
      GtkWidget *top = gtk_widget_get_toplevel (GTK_WIDGET (pop));

      if (top && GTK_IS_WINDOW (top) && gtk_widget_get_visible (top))
	{
	  GtkWindow *win = GTK_WINDOW (pop->popup);
	  GdkScreen *screen = gtk_window_get_screen (win);
	  GtkWidget *widget = GTK_WIDGET (pop);
	  GtkAllocation alloc;
	  gint wx, wy, x, y, width, height, min_width, def_width; 
	  gint horiz_border = 0, vert_border = 0;
	  gint screen_width, screen_height;

#if GTK_MAJOR_VERSION == 2
	  GtkRequisition requisition;
	  gtk_widget_size_request (pop->popup, &requisition);
          min_width = requisition.width;
#else
          gtk_widget_get_preferred_width(pop->popup, &min_width, NULL);
#endif
          gtk_widget_get_allocation (widget, &alloc);
          gtk_window_set_transient_for (win, GTK_WINDOW (top));
	  gdk_window_get_origin (gtk_widget_get_window (top), &x, &y);
	  gtk_window_get_size (win, &width, &height);
          // printf("w: %d, h:%d\n", width, height);
	  gtk_window_get_default_size (win, &def_width, NULL);
	  def_width = MAX (def_width, min_width);

	  if (gtk_widget_get_visible (pop->popup))
	    {
              GdkWindow *window = gtk_widget_get_window (pop->popup);
	      GdkRectangle rect;

              if (window != NULL)
                {
                  gdk_window_get_frame_extents (window, &rect);
                  width = gdk_window_get_width (window);
                  height = gdk_window_get_height (window);

                  horiz_border = rect.width - width;
                  vert_border = rect.height - height;
                }
	    }
	  if (gtk_window_get_decorated (win) && 
		  horiz_border == 0 && vert_border == 0)
	    {
	      horiz_border = 12;		/* best guess */
	      vert_border = 32;
	    }

	  gtk_widget_translate_coordinates(widget, top, 0, 0, &wx, &wy);
	  x += wx;
	  y += wy + alloc.height;
	  width = MAX(alloc.width - horiz_border, width);
	  width = MAX(def_width, width);

	  if (screen)
	    {
	      screen_width = gdk_screen_get_width (screen);
	      screen_height = gdk_screen_get_height (screen);
	    }
	  else
	    {
	      screen_width = gdk_screen_width ();
	      screen_height = gdk_screen_height ();
	    }

	  x = CLAMP (x, 0, screen_width - width - horiz_border);
	  y = CLAMP (y, 0, screen_height - height - vert_border);

	  gtk_window_set_gravity (win, GDK_GRAVITY_NORTH_WEST);
	  gtk_window_move (win, x, y);
	  gtk_window_resize (win, width, height);
	}
    }
}

static gboolean
gtk_pop_string_popup_follow (GtkPopString *pop)
{
  if (pop->follow)
    {
      gtk_pop_string_dimension_popup (pop);
    }
  return FALSE;
}

static void 
gtk_pop_string_map(GtkPopString *pop)
{
  pop->window = gtk_widget_get_toplevel (GTK_WIDGET (pop));
  g_object_ref(G_OBJECT(pop->window));

  g_signal_connect_swapped (pop, "configure-event", 
	  G_CALLBACK (gtk_pop_string_popup_follow), pop);
  g_signal_connect_swapped (pop->window, "configure-event", 
	  G_CALLBACK (gtk_pop_string_popup_follow), pop);
  g_signal_connect_swapped (pop->window, "show", 
	  G_CALLBACK (gtk_pop_string_show), pop);
  g_signal_connect_swapped (pop->window, "hide", 
	  G_CALLBACK (gtk_pop_string_hide), pop);
}

static void 
gtk_pop_string_unmap(GtkPopString *pop)
{
  g_signal_handlers_disconnect_by_func (pop->window,
	  gtk_pop_string_hide, pop);
  g_signal_handlers_disconnect_by_func (pop->window,
	  gtk_pop_string_show, pop);
  g_signal_handlers_disconnect_by_func (pop->window,
	  gtk_pop_string_popup_follow, pop);
  g_signal_handlers_disconnect_by_func (pop,
	  gtk_pop_string_popup_follow, pop);

  g_object_unref(G_OBJECT(pop->window));
  pop->window = NULL;
}

static void
gtk_pop_string_show(GtkPopString *pop)
{
  if (pop->volat && pop->popup)
    {
      gtk_widget_show (pop->popup);
    }
}

static void
gtk_pop_string_hide(GtkPopString *pop)
{
  if (pop->volat && pop->popup)
    {
      gtk_widget_hide (pop->popup);
    }
}

