/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkpopstring.h
**
** an implementation of MUI's popstring class for gtk+-2.0
**
** Creation date: 18.05.2004
** Last modified: 23.05.2004
*/

#ifndef GTK_POP_STRING_H
#define GTK_POP_STRING_H

/*
** GtkPopString class is the base class for creating so called
** popup objects. Usually, a popup consists of a string object
** (GtkEntry, GtkLabel or GtkButton), followed by a little button
** (GtkButton object). Pressing this button brings up a little 
** window with a listview and lets the user choose an entry 
** with the mouse.
**
** GtkPopString class features the basic functions for creating
** such objects. Given a string object and a button object,
** it places them horizontally and sets up some notification.
** Whenever the popup button is pressed, a method will be called 
** which itself should open and prepare the popup window.
**
** The string and the button object are not created by GtkPopstring
** class, they have to be supplied as fields during object
** creation time. This makes GtkPopString class very flexible,
** one could e.g. use a text object (GtkLabel) instead of a 
** string or a popup button with some text in it.
**
** However, creating simple popups with GtkPopString class would
** be too much overhead. Instead of using it directly, you
** should have a look at one of its subclasses. They offer
** a more specialized set of popups and are a lot easier to use.
*/

#include <gtk/gtk.h>

#ifndef G_GNUC_MAY_ALIAS
#if     __GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 3)
#  define G_GNUC_MAY_ALIAS __field__((may_alias))
#else
#  define G_GNUC_MAY_ALIAS
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define GTK_TYPE_POP_STRING	(gtk_pop_string_get_type ())
#define GTK_POP_STRING(obj)	(G_TYPE_CHECK_INSTANCE_CAST((obj), \
	    GTK_TYPE_POP_STRING, GtkPopString))
#define GTK_POP_STRING_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST((klass), \
	    GTK_TYPE_POP_STRING, GtkPopStringClass))
#define GTK_IS_POP_STRING(obj) 	(G_TYPE_CHECK_INSTANCE_TYPE ((obj), \
	    GTK_TYPE_POP_STRING))
#define GTK_IS_POP_STRING_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), \
	    GTK_TYPE_POP_STRING))
#define GTK_POP_STRING_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS((obj), \
	    GTK_TYPE_POP_STRING, GtkPopStringClass))

typedef struct _GtkPopString      GtkPopString;
typedef struct _GtkPopStringClass GtkPopStringClass;

struct _GtkPopString
{
#if GTK_MAJOR_VERSION == 2
  GtkHBox hbox;
#else
  GtkBox box;
#endif
  GtkWidget *string;
  GtkWidget *button;
  guint toggle:1;
  guint follow:1;
  guint volat:1;
  G_GNUC_MAY_ALIAS GtkWidget *popup;
  GtkWidget *window;
};

struct _GtkPopStringClass
{
#if GTK_MAJOR_VERSION == 2
  GtkHBoxClass parent_class;
#else
  GtkBoxClass parent_class;
#endif
  gboolean (*open)(GtkPopString *pop, gpointer data);
  void (*close)(GtkPopString *pop, gboolean success, gpointer data);
};

GType	   gtk_pop_string_get_type     (void) G_GNUC_CONST;
GtkWidget* gtk_pop_string_new	       (void);
void       gtk_pop_string_set_string   (GtkPopString  	*pop,
	                                GtkWidget       *string);
GtkWidget* gtk_pop_string_get_string   (GtkPopString    *pop);
void       gtk_pop_string_set_button   (GtkPopString  	*pop,
	                                GtkWidget       *button);
GtkWidget* gtk_pop_string_get_button   (GtkPopString    *pop);
void       gtk_pop_string_set_toggle   (GtkPopString  	*pop,
	                                gboolean         toggle);
gboolean   gtk_pop_string_get_toggle   (GtkPopString    *pop);
void       gtk_pop_string_set_follow   (GtkPopString  	*pop,
	                                gboolean         follow);
gboolean   gtk_pop_string_get_follow   (GtkPopString    *pop);
void       gtk_pop_string_set_volatile (GtkPopString  	*pop,
	                                gboolean         volat);
gboolean   gtk_pop_string_get_volatile (GtkPopString    *pop);
void       gtk_pop_string_set_popup    (GtkPopString  	*pop,
	                                GtkWidget       *popup);
GtkWidget* gtk_pop_string_get_popup    (GtkPopString    *pop);
void       gtk_pop_string_set_text     (GtkPopString    *pop,
	                                const gchar     *text);
G_CONST_RETURN gchar* gtk_pop_string_get_text (GtkPopString *pop);

/* 
** This method opens the popup. In fact, it only emits the
** "open" signal and checks its return value. In case of true, 
** the popup button object is disabled as long as toggle is unset.
** If the toggle mode is enabled, calling this function on a currently opened 
** popup will result in closing this popup (i.e. emitting a "close" signal
** with a success value of false.
*/
void       gtk_pop_string_open_popup   (GtkPopString  	*pop);

/* 
** This method closes the popup. In fact, it only emits the
** "close" signal with the supplied success parameter.
*/
void       gtk_pop_string_close_popup   (GtkPopString  	*pop,
	                                 gboolean        success);

#ifdef __cplusplus
}
#endif

#endif

