/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkpopwidget.c
**
** an implementation of MUI's popobject class for gtk+-2.0
**
** Creation date: 01.06.2004
** Last modified: 05.09.2012
*/

#include "gtkpopwidget.h"
#include <gdk/gdkkeysyms.h>
#include <glib/gi18n.h>

#undef USE_POPUP
#undef USE_MODAL

#define P_(x) N_(x)

enum {
  PROP_0,
  PROP_WIDGET,
  PROP_LIGHT
};

enum {
  START,
  STOP,
  LAST_SIGNAL
};

static void     gtk_pop_widget_class_init  (GtkPopWidgetClass *klass);
static void     gtk_pop_widget_init        (GtkPopWidget      *pop);
static void     gtk_pop_widget_set_property(GObject           *object,
                                            guint              prop_id,
                                            const GValue      *value,
                                            GParamSpec        *pspec);
static void     gtk_pop_widget_get_property(GObject           *object,
                                            guint              prop_id,
                                            GValue            *value,
                                            GParamSpec        *pspec);
static gboolean gtk_pop_widget_open        (GtkPopString      *pop,
	                                    gpointer           data);
static void     gtk_pop_widget_close       (GtkPopString      *pop,
	                                    gboolean           success,
					    gpointer           data);
static gboolean gtk_pop_widget_start       (GtkPopWidget      *pop, 
	                                    gpointer           data);
static void     gtk_pop_widget_stop        (GtkPopWidget      *pop, 
	                                    gpointer           data);

static GtkPopStringClass *parent_class = NULL;

static guint gtk_pop_widget_signals[LAST_SIGNAL] = { 0 };

static void 
g_cclosure_marshal_BOOLEAN__VOID (GClosure     *closure,
                               GValue       *return_value G_GNUC_UNUSED,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint G_GNUC_UNUSED,
                               gpointer      marshal_data)
{
  typedef gboolean (*GMarshalFunc_BOOLEAN__VOID) (gpointer     data1,
  					          gpointer     data2);
  GMarshalFunc_BOOLEAN__VOID callback;
  GCClosure *cc = (GCClosure*) closure;
  gpointer data1, data2;
  gboolean v_return;

  g_return_if_fail (n_param_values == 1);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }
  callback = (GMarshalFunc_BOOLEAN__VOID) (marshal_data ? 
          marshal_data : cc->callback);

  v_return = callback (data1, data2);
  g_value_set_boolean (return_value, v_return);
}

GType
gtk_pop_widget_get_type (void)
{
  static GType pop_widget_type = 0;

  if (!pop_widget_type)
    {
      static const GTypeInfo pop_widget_info =
      {
	sizeof (GtkPopWidgetClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_pop_widget_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkPopWidget),
	0,		/* n_preallocs */
	(GInstanceInitFunc) gtk_pop_widget_init,
      };

      pop_widget_type = g_type_register_static (GTK_TYPE_POP_STRING, 
	      "GtkPopWidget", &pop_widget_info, 0);
    }

  return pop_widget_type;
}

static void
gtk_pop_widget_class_init (GtkPopWidgetClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GtkPopStringClass *pop_string_class = GTK_POP_STRING_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = gtk_pop_widget_set_property;
  gobject_class->get_property = gtk_pop_widget_get_property;

  pop_string_class->open = gtk_pop_widget_open;
  pop_string_class->close = gtk_pop_widget_close;

  klass->start = gtk_pop_widget_start;
  klass->stop = gtk_pop_widget_stop;

  g_object_class_install_property (gobject_class,
				   PROP_WIDGET,
				   g_param_spec_object ("widget",
						     P_("Widget"),
						     P_("widget to popup in a window"),
						     GTK_TYPE_WIDGET,
						     G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
				   PROP_LIGHT,
				   g_param_spec_boolean ("light",
						      P_("Light"),
						      P_("Whether the popup window should be borderless"),
						     TRUE,
						     G_PARAM_READWRITE));

  gtk_pop_widget_signals[START] =
    g_signal_new ("start",
	          G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		  G_STRUCT_OFFSET (GtkPopWidgetClass, start),
		  NULL, NULL,
		  g_cclosure_marshal_BOOLEAN__VOID,
		  G_TYPE_BOOLEAN, 0);
  gtk_pop_widget_signals[STOP] =
    g_signal_new ("stop",
	          G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		  G_STRUCT_OFFSET (GtkPopWidgetClass, stop),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__VOID,
		  G_TYPE_NONE, 0);
}

static void
gtk_pop_widget_set_property (GObject         *object,
                             guint            prop_id,
                             const GValue    *value,
                             GParamSpec      *pspec)
{
  GtkPopWidget *pop = GTK_POP_WIDGET (object);

  switch (prop_id)
    {
    case PROP_WIDGET:
      gtk_pop_widget_set_widget (pop, g_value_get_object (value));
      break;
    case PROP_LIGHT:
      gtk_pop_widget_set_light (pop, g_value_get_boolean (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
gtk_pop_widget_get_property (GObject         *object,
                             guint            prop_id,
                             GValue          *value,
                             GParamSpec      *pspec)
{
  GtkPopWidget *pop = GTK_POP_WIDGET (object);

  switch (prop_id)
    {
    case PROP_WIDGET:
      g_value_set_object (value, pop->widget);
      break;
    case PROP_LIGHT:
      g_value_set_boolean (value, pop->light);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
gtk_pop_widget_init (GtkPopWidget *pop)
{
  pop->light = TRUE;
  g_object_set (G_OBJECT (pop), "follow", TRUE, NULL);
}

GtkWidget*
gtk_pop_widget_new (void)
{
  return g_object_new (GTK_TYPE_POP_WIDGET, NULL);
}

void
gtk_pop_widget_set_widget (GtkPopWidget *pop, GtkWidget *widget)
{
  g_return_if_fail (GTK_IS_POP_WIDGET (pop));
  g_return_if_fail (widget == NULL || GTK_IS_WIDGET (widget));

  if (pop->widget != widget)
    {
      if (widget)
	{
	  g_object_ref (G_OBJECT (widget));
	}
      if (pop->widget)
        {
	  g_object_unref (G_OBJECT (pop->widget));
	}
      pop->widget = widget;

      g_object_notify (G_OBJECT (pop), "widget");
    }
}

GtkWidget *
gtk_pop_widget_get_widget (GtkPopWidget *pop)
{
  g_return_val_if_fail (GTK_IS_POP_WIDGET (pop), NULL);

  return pop->widget;
}

void
gtk_pop_widget_set_light (GtkPopWidget *pop, gboolean light)
{
  g_return_if_fail (GTK_IS_POP_WIDGET (pop));

  light = (light != FALSE);

  if (pop->light != light)
    {
      pop->light = light;

      g_object_notify (G_OBJECT (pop), "light");
    }
}

gboolean
gtk_pop_widget_get_light (GtkPopWidget *pop)
{
  g_return_val_if_fail (GTK_IS_POP_WIDGET (pop), TRUE);

  return pop->light;
}

static gboolean
gtk_pop_widget_on_delete (GtkPopString *pop)
{
  gtk_pop_string_close_popup (pop, FALSE);

  return TRUE;
}

static gboolean 
gtk_pop_widget_on_key_press (GtkPopString *pop, GdkEventKey *event)
{
  if (event->keyval == GDK_KEY_Escape) 
    {
      if (pop->button)
        {
	  g_signal_emit_by_name (pop->button, "activate");
        }
      else
        {
          gtk_pop_string_close_popup (pop, FALSE);
        }
      
      return TRUE;
    }

  return FALSE;
}

#ifdef USE_MODAL
static gboolean 
gtk_pop_widget_on_button_press_event (GtkPopString *pop, GdkEventButton *event)
{
  gtk_pop_string_close_popup (pop, FALSE);
  return TRUE;
}

#else

static gboolean
gtk_pop_widget_on_focus_event (GtkPopString *pop)
{
  if (pop->popup && !gtk_window_is_active (GTK_WINDOW (pop->popup)))
    {
      if (pop->button)
        {
          g_signal_emit_by_name (pop->button, "activate");
        }
      else
        {
          gtk_pop_string_close_popup (pop, FALSE);
        }
    }
  return FALSE;
}

static void
gtk_pop_widget_on_destroy (GtkPopWidget *pop)
{
    if (pop->toplevel) {
        g_signal_handlers_disconnect_by_func (pop->toplevel,
                gtk_pop_widget_on_focus_event, pop);
        g_object_unref (G_OBJECT (pop->toplevel));
        pop->toplevel = NULL;
    }
}
#endif

static gboolean
gtk_pop_widget_open (GtkPopString *pop, gpointer data)
{
  gboolean success = TRUE;
  GtkWidget *popup = gtk_pop_string_get_popup (pop);

  if (popup == NULL)
    {
      GtkWidget *top;
      GtkWidget *widget = GTK_POP_WIDGET (pop)->widget;
      gboolean light = GTK_POP_WIDGET (pop)->light;

      if (widget == NULL)
        {
	  return FALSE;
	}

      if (light)
        {
#ifdef USE_POPUP
          popup = gtk_window_new (GTK_WINDOW_POPUP);
#else
          popup = gtk_window_new (GTK_WINDOW_TOPLEVEL);
          gtk_window_set_decorated (GTK_WINDOW(popup), FALSE);
#endif
#ifdef USE_MODAL
          gtk_window_set_modal (GTK_WINDOW(popup), TRUE);
          g_signal_connect_swapped (popup, "button-press-event",
                  G_CALLBACK(gtk_pop_widget_on_button_press_event), pop);
#else
          top = gtk_widget_get_toplevel (GTK_WIDGET (pop));
          if (top)
            {
              g_object_ref (G_OBJECT (top));
              GTK_POP_WIDGET (pop)->toplevel = top;
              g_signal_connect_swapped (top, "focus-in-event",
                      G_CALLBACK(gtk_pop_widget_on_focus_event), pop);
              g_signal_connect_swapped (popup, "destroy", 
                      G_CALLBACK(gtk_pop_widget_on_destroy), pop);
            }
#endif
        }
      else
        {
          popup = gtk_window_new (GTK_WINDOW_TOPLEVEL);
        }
      gtk_window_set_skip_taskbar_hint (GTK_WINDOW (popup), TRUE);
      gtk_window_set_skip_pager_hint (GTK_WINDOW (popup), TRUE);
      gtk_window_set_type_hint (GTK_WINDOW (popup), 
              GDK_WINDOW_TYPE_HINT_DIALOG);
      gtk_container_add (GTK_CONTAINER (popup), widget);
      gtk_widget_show_all (widget);

      g_signal_connect_swapped (popup, "delete-event", 
	      G_CALLBACK(gtk_pop_widget_on_delete), pop);
      g_signal_connect_swapped (popup, "key-press-event", 
	    G_CALLBACK(gtk_pop_widget_on_key_press), pop);

      gtk_pop_string_set_popup (pop, popup);
    }

  g_signal_emit (pop, gtk_pop_widget_signals[START], 0, &success);

  if (!success)
    {
      gtk_pop_widget_close (pop, FALSE, NULL);
      
      return FALSE;
    }

  gtk_widget_show (popup);
  gtk_window_present (GTK_WINDOW (popup));   // increase chances to activate

  return TRUE;
}

static void
gtk_pop_widget_close (GtkPopString *pop, gboolean success, gpointer data)
{
  GtkWidget *popup, *widget;

  if (success) 
    {
      g_signal_emit (pop, gtk_pop_widget_signals[STOP], 0);
    }

  popup = gtk_pop_string_get_popup (pop);
  widget = gtk_pop_widget_get_widget (GTK_POP_WIDGET (pop));

  if (GTK_IS_BIN (popup) && GTK_IS_WIDGET (widget) && 
	  widget == gtk_bin_get_child (GTK_BIN(popup)))
    {
      gtk_widget_hide (popup);
      gtk_container_remove (GTK_CONTAINER (popup), widget);
      gtk_widget_hide (widget);
    }

  gtk_pop_string_set_popup (pop, NULL);
}

static gboolean 
gtk_pop_widget_start (GtkPopWidget *pop, gpointer data)
{
  return TRUE;
}

static void 
gtk_pop_widget_stop (GtkPopWidget *pop, gpointer data)
{
}

