/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkradiobox.c
**
** a group of radio buttons
**
** Creation date: 12.12.2004
** Last modified: 05.09.2012
*/

#define ENABLE_NLS

#include "gtkradiobox.h"
#include "gtk/gtkUtil.h"
#include <glib/gi18n.h>

#define P_(x) N_(x)

enum {
  PROP_0,
  PROP_ACTIVE,
  PROP_ENTRIES
};

enum {
  ACTIVE_CHANGED,
  LAST_SIGNAL
};

static void gtk_radio_box_class_init     (GtkRadioBoxClass *klass);
static void gtk_radio_box_init           (GtkRadioBox      *box);
static void gtk_radio_box_set_property   (GObject          *object,
                                           guint            prop_id,
                                           const GValue    *value,
                                           GParamSpec      *pspec);
static void gtk_radio_box_get_property   (GObject          *object,
                                           guint            prop_id,
                                           GValue          *value,
                                           GParamSpec      *pspec);
static void gtk_radio_box_on_radio_button_toggled 
                                          (GtkToggleButton *radio, 
				           GtkRadioBox     *box);
    
static GtkGroupClass *parent_class = NULL;

static guint gtk_radio_box_signals[LAST_SIGNAL] = { 0 };

GType
gtk_radio_box_get_type (void)
{
  static GType radio_box_type = 0;

  if (!radio_box_type)
    {
      static const GTypeInfo radio_box_info =
      {
	sizeof (GtkRadioBoxClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) gtk_radio_box_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (GtkRadioBox),
	0,		/* n_preallocs */
	(GInstanceInitFunc) gtk_radio_box_init,
      };

      radio_box_type = g_type_register_static (GTK_TYPE_GROUP, "GtkRadioBox",
					        &radio_box_info, 0);
    }

  return radio_box_type;
}

static void
gtk_radio_box_class_init (GtkRadioBoxClass *klass)
{
  GObjectClass *gobject_class;

  gobject_class = G_OBJECT_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = gtk_radio_box_set_property;
  gobject_class->get_property = gtk_radio_box_get_property;

  g_object_class_install_property (gobject_class,
				   PROP_ACTIVE,
				   g_param_spec_int ("active",
						  P_("Active"),
						  P_("Active radio button in the group"),
						  0,
						  G_MAXINT,
						  0,
						  G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
				   PROP_ENTRIES,
				   g_param_spec_pointer ("entries",
						      P_("Entries"),
						      P_("Array of isocode strings for the radio buttons"),
						      G_PARAM_WRITABLE));

  gtk_radio_box_signals[ACTIVE_CHANGED] =
    g_signal_new ("active_changed",
	          G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		  G_STRUCT_OFFSET (GtkRadioBoxClass, active_changed),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__INT,
		  G_TYPE_NONE, 1, G_TYPE_INT);
}

static void
gtk_radio_box_init (GtkRadioBox *box)
{
  gtk_group_set_vert_spacing (&box->group, 0);
}

static void
gtk_radio_box_set_property (GObject         *object,
                            guint            prop_id,
                            const GValue    *value,
                            GParamSpec      *pspec)
{
  GtkRadioBox *box = GTK_RADIO_BOX (object);

  switch (prop_id)
    {
    case PROP_ACTIVE:
      gtk_radio_box_set_active (box, g_value_get_int (value));
      break;
    case PROP_ENTRIES:
      gtk_radio_box_set_entries (box, (char **)g_value_get_pointer (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
gtk_radio_box_get_property (GObject         *object,
                            guint            prop_id,
                            GValue          *value,
                            GParamSpec      *pspec)
{
  GtkRadioBox *box = GTK_RADIO_BOX (object);

  switch (prop_id)
    {
    case PROP_ACTIVE:
      g_value_set_int (value, gtk_radio_box_get_active (box));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

GtkWidget*
gtk_radio_box_new (void)
{
  return g_object_new (GTK_TYPE_RADIO_BOX, NULL);
}

gboolean       
gtk_radio_box_set_active (GtkRadioBox *box, gint active)
{
  GList *children;

  g_return_val_if_fail (GTK_IS_RADIO_BOX (box), FALSE);

  children = box->group.children;
  while (children && active > 0)
    {
      children = children->next;
      active--;
    }
  if (children)
    {
      GtkGroupChild *child = children->data;
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (child->widget), TRUE);
      return TRUE;
    }
  return FALSE;
}

gint
gtk_radio_box_get_active (GtkRadioBox *box)
{
  gint active;
  GList *children;

  g_return_val_if_fail (GTK_IS_RADIO_BOX (box), -1);

  active = 0;
  children = box->group.children;
  while (children)
    {
      GtkGroupChild *child = children->data;
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (child->widget)))
        {
	  return active;
        }
      children = children->next;
      active++;
    }

  return -1;
}

void       
gtk_radio_box_set_entries (GtkRadioBox *box, char **entries)
{
  g_return_if_fail (GTK_IS_RADIO_BOX (box));

  /* first clear existing radio buttons */

  while (box->group.children)
    {
      GtkGroupChild *child = box->group.children->data;
      gtk_widget_destroy (child->widget);
    }

  if (entries)
    {
      GSList *group = NULL;

      while (*entries)
        {
	  GtkWidget *radio = gtk_radio_button_new_with_label (group, *entries);

	  gtk_widget_show(radio);
	  g_signal_connect(radio, "toggled", 
		  G_CALLBACK(gtk_radio_box_on_radio_button_toggled), box);
	  gtk_group_add_child (&box->group, radio, 0, 0);
	  group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (radio));
	  entries++;
        }
    }
}

static void 
gtk_radio_box_on_radio_button_toggled (GtkToggleButton *radio, 
				       GtkRadioBox     *box)
{
  if (gtk_toggle_button_get_active(radio)) 
    {
      gint active = gtk_radio_box_get_active (box);
      g_signal_emit (box, gtk_radio_box_signals[ACTIVE_CHANGED], 0, active);
    }
}

