/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtksfwview.c
**
** GtkSFWView is a subclass of GtkTreeView class.
** Its purpose is for displaying the contents of a program element.
**
** Creation data: 13.05.2004
** Last modified: 12.05.2022
*/

#define ENABLE_NLS

#include "gtksfwview.h"
#include <gtk/gtkUtil.h>
#include <glib/gi18n.h>
#include <lang/program.h>
#include <field.h>
#include <requesters.h>
#include <imexport.h>
#include <pdfexport.h>
#include <print.h>
#include <util.h>
#include <memory.h>
#include <appstrings_cat.h>

#define P_(x) N_(x)

enum {
    PROP_0,
    PROP_SHOW_TITLES,
    PROP_CHECKBOXES,
    PROP_ACTIVE_LINE
};

enum {
    INTERNAL_MOVE_DROP,
    LAST_SIGNAL
};

#if 0
#define SORT_COL_NONE 0xff
#define SORT_COL_MASK 0x7f
#define SORT_UP_MASK 0x80
#define SORT_COL(x) ((x) & SORT_COL_MASK)
#define SORT_IS_UP(x) ((x) & SORT_UP_MASK)
#define SORT_COL_FLIP(x) ((x) + SORT_UP_MASK)
#endif

const GtkTargetEntry gtk_sfw_view_target = {
    (gchar *)"QueryLine", GTK_TARGET_SAME_APP, 0
};
static const GtkTargetEntry reorderTarget = {
    (gchar *)"QueryLine", GTK_TARGET_SAME_WIDGET, 0
};

static void gtk_sfw_view_class_init     (GtkSFWViewClass   *klass);
static void gtk_sfw_view_init           (GtkSFWView        *entry);
#if GTK_MAJOR_VERSION == 2
static void gtk_sfw_view_destroy        (GtkObject        *object);
#else
static void gtk_sfw_view_destroy        (GtkWidget        *widget);
#endif
static void gtk_sfw_view_set_property   (GObject         *object,
                                              guint            prop_id,
                                              const GValue    *value,
                                              GParamSpec      *pspec);
static void gtk_sfw_view_get_property   (GObject         *object,
                                              guint            prop_id,
                                              GValue          *value,
                                              GParamSpec      *pspec);

static GtkTreeViewClass *parent_class = NULL;

static guint gtk_sfw_view_signals[LAST_SIGNAL] = { 0 };

GType gtk_sfw_view_get_type(void)
{
    static GType sfw_view_type = 0;

    if (!sfw_view_type) {
	static const GTypeInfo sfw_view_info =
	{
	    sizeof (GtkSFWViewClass),
	    NULL,		/* base_init */
	    NULL,		/* base_finalize */
	    (GClassInitFunc)gtk_sfw_view_class_init,
	    NULL,		/* class_finalize */
	    NULL,		/* class_data */
	    sizeof (GtkSFWView),
	    0,			/* n_preallocs */
	    (GInstanceInitFunc)gtk_sfw_view_init,
	};

	sfw_view_type = g_type_register_static (GTK_TYPE_TREE_VIEW, 
		"GtkSFWView", &sfw_view_info, 0);
    }
    
    return sfw_view_type;
}

static void gtk_sfw_view_class_init(GtkSFWViewClass *klass)
{
    GObjectClass *gobject_class;
#if GTK_MAJOR_VERSION == 2
    GtkObjectClass *object_class;
#else
    GtkWidgetClass *widget_class;
#endif

    gtk_binding_set_by_class(klass);

    gobject_class = G_OBJECT_CLASS(klass);
  
    parent_class = g_type_class_peek_parent(klass);
    
    gobject_class->set_property = gtk_sfw_view_set_property;
    gobject_class->get_property = gtk_sfw_view_get_property;

#if GTK_MAJOR_VERSION == 2
    object_class = GTK_OBJECT_CLASS (klass);
    object_class->destroy = gtk_sfw_view_destroy;
#else
    widget_class = GTK_WIDGET_CLASS (klass);
    widget_class->destroy = gtk_sfw_view_destroy;
#endif
    
    g_object_class_install_property (gobject_class,
				     PROP_SHOW_TITLES,
				     g_param_spec_boolean ("show_titles",
							P_("Show titles"),
							P_("Whether the titles of the SFW query should be displayed"),
							FALSE,
							G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
    g_object_class_install_property (gobject_class,
				     PROP_CHECKBOXES,
				     g_param_spec_boolean ("checkboxes",
							P_("Checkboxes"),
							P_("Whether each line has a checkbox for selecting the line"),
							FALSE,
							G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
    g_object_class_install_property (gobject_class,
				     PROP_ACTIVE_LINE,
				     g_param_spec_uint ("active_line",
						     P_("Active line"),
						     P_("Active line starting with 1 in the order of the original query"),
						        0,
						        G_MAXUINT,
							0,
							G_PARAM_READWRITE));

    gtk_sfw_view_signals[INTERNAL_MOVE_DROP] =
        g_signal_new ("internal-move-drop",
                G_OBJECT_CLASS_TYPE (gobject_class),
                G_SIGNAL_RUN_LAST,  // | G_SIGNAL_ACTION,
                G_STRUCT_OFFSET (GtkSFWViewClass, internalMoveDrop),
                NULL, NULL,
                g_cclosure_marshal_VOID__LONG,
                G_TYPE_NONE, 1, G_TYPE_LONG);
}


static void gtk_sfw_view_set_property (GObject         *object,
                                            guint            prop_id,
                                            const GValue    *value,
                                            GParamSpec      *pspec)
{
    GtkSFWView *sfw = GTK_SFW_VIEW(object);

    switch (prop_id) {
	case PROP_SHOW_TITLES:
	    gtk_sfw_view_set_show_titles(sfw, g_value_get_boolean(value));
	    break;
	case PROP_CHECKBOXES:
	    gtk_sfw_view_set_checkboxes(sfw, g_value_get_boolean(value));
	    break;
	case PROP_ACTIVE_LINE:
	    gtk_sfw_view_set_active_line(sfw, g_value_get_uint(value));
	    break;
	default:
	    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
	    break;
    }
}

static void gtk_sfw_view_get_property (GObject         *object,
                                            guint            prop_id,
                                            GValue          *value,
                                            GParamSpec      *pspec)
{
    GtkSFWView *sfw = GTK_SFW_VIEW(object);

    switch (prop_id) {
	case PROP_SHOW_TITLES:
	    g_value_set_boolean(value, sfw->showTitles);
	    break;
	case PROP_CHECKBOXES:
	    g_value_set_boolean(value, sfw->checkboxes);
	    break;
	case PROP_ACTIVE_LINE:
	    g_value_set_uint(value, gtk_sfw_view_get_active_line(sfw));
	    break;
	default:
	    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
	    break;
    }
}

static gboolean sfwViewOnDragDrop(GtkSFWView *sfw, GdkDragContext *context, 
	gint x, gint y, guint time, gpointer p)
{
    GtkWidget *source = gtk_drag_get_source_widget(context);
    GtkWidget *widget = GTK_WIDGET(sfw);
    struct QueryLine *ql = GuiListActiveEntry(widget);
    gboolean success = FALSE;

    /* Check that this is a drag n' drop within the widget. */
    if (source != widget) {
        return FALSE;
    }

    if (ql != NULL) {
	GtkTreePath *path = 0;
	GtkTreeViewDropPosition drop;
	glong pos = sfw->data->qgd_NumLines, oldpos;
        GtkTreeModel *model;
        GtkListStore *store;
        GtkTreeIter iter;

	if (gtk_tree_view_get_dest_row_at_pos(GTK_TREE_VIEW(sfw),
		    x, y, &path, &drop) && path) {
	    gint *indices = gtk_tree_path_get_indices(path);
    	    if (indices) {
		pos = indices[0];
		if (drop == GTK_TREE_VIEW_DROP_AFTER) {
		    pos++;
		}
	    }
	    gtk_tree_path_free(path);
	}

        oldpos = GuiListFindEntry(widget, ql);
        QueryGuiDataMoveLine(sfw->data, oldpos, pos);
	if (pos > oldpos) {
	    pos--;
	}
        GuiListRemoveEntry(widget, ql);

        model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
        store = GTK_LIST_STORE(model);
        gtk_list_store_insert(store, &iter, pos);
        gtk_list_store_set(store, &iter, 0, ql, -1);

        GuiListSetActiveEntry(widget, ql);

        // First row is 1.
        g_signal_emit(sfw, gtk_sfw_view_signals[INTERNAL_MOVE_DROP], 0, pos+1);

        success = TRUE;
    }
    gtk_drag_finish(context, success, FALSE, time);
    return TRUE;
}

static void sfwViewOnCursorChanged(GtkTreeView *view)
{
    GtkTreeViewColumn *column;

    gtk_tree_view_get_cursor(view, NULL, &column);
    if (column) {
	gint num = GPOINTER_TO_INT(g_object_get_data(
		    G_OBJECT(column), "column"));
	if (num >= 0) {
	    gtk_tree_view_set_search_column(view, num);
	}
    }
}

static void sfwViewExportPDFPrint(GtkSFWView *sfw,
        void (*func)(struct Project *, GuiObject))
{
    GuiObject go = GTK_WIDGET(sfw);
    func(GuiObjectProject(go), go);
}

static void sfwViewOnMenuExportAsText(GtkSFWView *sfw)
{
    sfwViewExportPDFPrint(sfw, GuiExportSFWList);
}

static void sfwViewOnMenuExportAsPDF(GtkSFWView *sfw)
{
    sfwViewExportPDFPrint(sfw, GuiExportSFWListAsPDF);
}

static void sfwViewOnMenuPrint(GtkSFWView *sfw)
{
    sfwViewExportPDFPrint(sfw, GuiPrintSFWList);
}

static gboolean sfwViewOnPopupMenu(GtkSFWView *sfw)
{
    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
    gint count = model? gtk_tree_model_iter_n_children(model, NULL) : 0;
    GtkWidget *export, *pdf, *print;

    if (count <= 0) {
        return FALSE;
    }
    if (sfw->popupMenu) {
        gtk_widget_destroy(sfw->popupMenu);
    }
    sfw->popupMenu = gtk_menu_new();
    g_object_add_weak_pointer(G_OBJECT(sfw->popupMenu),
            (gpointer *)&sfw->popupMenu);
    gtk_menu_attach_to_widget(GTK_MENU(sfw->popupMenu), GTK_WIDGET(sfw), NULL);

    export = gtk_menu_item_new_with_label(GCS(MSG_MenuItem_ExportAsText));
    gtk_widget_show(export);
    gtk_menu_shell_append(GTK_MENU_SHELL(sfw->popupMenu), export);
    g_signal_connect_swapped(export, "activate", 
            G_CALLBACK(sfwViewOnMenuExportAsText), sfw);

    pdf = gtk_menu_item_new_with_label(GCS(MSG_MenuItem_ExportAsPDF));
    gtk_widget_show(pdf);
    gtk_menu_shell_append(GTK_MENU_SHELL(sfw->popupMenu), pdf);
    g_signal_connect_swapped(pdf, "activate", 
            G_CALLBACK(sfwViewOnMenuExportAsPDF), sfw);

    print = gtk_menu_item_new_with_label(GCS(MSG_Query_Print));
    gtk_widget_show(print);
    gtk_menu_shell_append(GTK_MENU_SHELL(sfw->popupMenu), print);
    g_signal_connect_swapped(print, "activate", 
            G_CALLBACK(sfwViewOnMenuPrint), sfw);

    gtk_menu_popup(GTK_MENU(sfw->popupMenu), 0, 0, 0, 0, TRUE, 
            gtk_get_current_event_time());

    return TRUE;
}

static gboolean sfwViewOnButtonPressEvent(GtkSFWView *sfw, GdkEventButton *event)
{
    if (event->button == 3) {
        return sfwViewOnPopupMenu(sfw);
    }
    return FALSE;
}

static void gtk_sfw_view_init(GtkSFWView *sfw)
{
    sfw->popupMenu = NULL;
    sfw->data = NULL;
    sfw->showTitles = FALSE;
    sfw->checkboxes = FALSE;

    gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(sfw), 
            GDK_BUTTON1_MASK, &gtk_sfw_view_target, 1, 
            (GDK_ACTION_MOVE | GDK_ACTION_COPY));
    gtk_tree_view_enable_model_drag_dest(GTK_TREE_VIEW(sfw), 
            &reorderTarget, 1, GDK_ACTION_MOVE);
    g_signal_connect(sfw, "drag-drop", G_CALLBACK(sfwViewOnDragDrop), NULL); 

    g_signal_connect(sfw, "cursor-changed", 
	    G_CALLBACK(sfwViewOnCursorChanged), NULL);
    g_signal_connect(sfw, "popup-menu",
            G_CALLBACK(sfwViewOnPopupMenu), NULL);
    g_signal_connect(sfw, "button-press-event",
            G_CALLBACK(sfwViewOnButtonPressEvent), NULL);
}

static void gtk_sfw_view_cleanup(GtkSFWView *sfw)
{
    FreeQueryGuiData(sfw->data);
    sfw->data = NULL;

    if (sfw->popupMenu) {
        gtk_widget_destroy(sfw->popupMenu);
    }
}

#if GTK_MAJOR_VERSION == 2
static void gtk_sfw_view_destroy(GtkObject *object)
{
    gtk_sfw_view_cleanup(GTK_SFW_VIEW(object));
    (* GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}
#else
static void gtk_sfw_view_destroy(GtkWidget *widget)
{
    gtk_sfw_view_cleanup(GTK_SFW_VIEW(widget));
    (* GTK_WIDGET_CLASS(parent_class)->destroy)(widget);
}
#endif

GtkWidget* gtk_sfw_view_new (void)
{
    return g_object_new(GTK_TYPE_SFW_VIEW, NULL);
}

void gtk_sfw_view_set_show_titles(GtkSFWView *sfw, gboolean show)
{
    g_return_if_fail(GTK_IS_SFW_VIEW(sfw));

    sfw->showTitles = show;
    g_object_notify(G_OBJECT(sfw), "show-titles");
}

gboolean gtk_sfw_view_get_show_titles(GtkSFWView *sfw)
{
    g_return_val_if_fail(GTK_IS_SFW_VIEW(sfw), FALSE);

    return sfw->showTitles;
}

void gtk_sfw_view_set_checkboxes(GtkSFWView *sfw, gboolean checkboxes)
{
    g_return_if_fail(GTK_IS_SFW_VIEW(sfw));

    sfw->checkboxes = checkboxes;
    g_object_notify(G_OBJECT(sfw), "checkboxes");
}

gboolean gtk_sfw_view_get_checkboxes(GtkSFWView *sfw)
{
    g_return_val_if_fail(GTK_IS_SFW_VIEW(sfw), FALSE);

    return sfw->checkboxes;
}

void gtk_sfw_view_set_active_record(GtkSFWView *sfw, struct Record *rec)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    struct QueryLine *ql;
    guint lineNum = 0;
    guint controlIndex;

    g_return_if_fail(GTK_IS_SFW_VIEW(sfw));

    if (sfw->data == NULL) {
        return;
    }
    controlIndex = sfw->data->qgd_ControlIndex;
    if (sfw->data->qgd_ControlTable == NULL ||
            controlIndex >= sfw->data->qgd_NumFields) {
        return;
    }

    model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
    if (model) {
        if (gtk_tree_model_get_iter_first(model, &iter)) {
            do {
                ql = NULL;
                gtk_tree_model_get(model, &iter, 0, &ql, -1);
                if (ql && ql->ql_Entries[controlIndex]->qe_Record == rec) {
                    lineNum = ql->ql_LineNum;
                    break;
                }
            } while (gtk_tree_model_iter_next(model, &iter));
        }
    }

    gtk_sfw_view_set_active_line(sfw, lineNum);
}

struct Record *gtk_sfw_view_get_active_record(GtkSFWView *sfw)
{
    GtkTreeSelection *selection;
    GtkTreeModel *model;
    GtkTreeIter iter;
    struct QueryLine *ql = NULL;
    guint controlIndex;

    g_return_val_if_fail(GTK_IS_SFW_VIEW(sfw), NULL);

    if (sfw->data == NULL) {
        return NULL;
    }
    controlIndex = sfw->data->qgd_ControlIndex;
    if (sfw->data->qgd_ControlTable == NULL ||
            controlIndex >= sfw->data->qgd_NumFields) {
        return NULL;
    }

    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(sfw));
    if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
        gtk_tree_model_get(model, &iter, 0, &ql, -1);
    }

    if (ql == NULL) {
        return NULL;
    }
    return ql->ql_Entries[controlIndex]->qe_Record;
}

struct Record *gtk_sfw_view_get_record(GtkSFWView *sfw, long row)
{
    struct QueryLine *ql = NULL;
    guint controlIndex;

    g_return_val_if_fail(GTK_IS_SFW_VIEW(sfw), NULL);

    if (sfw->data == NULL) {
        return NULL;
    }
    controlIndex = sfw->data->qgd_ControlIndex;
    if (sfw->data->qgd_ControlTable == NULL ||
            controlIndex >= sfw->data->qgd_NumFields) {
        return NULL;
    }

    if (row == 0) {
        ql = sfw->data->qgd_Titles;
    } else {
        GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
        GtkTreePath *path = gtk_tree_path_new_from_indices(row - 1, -1);
	GtkTreeIter iter;

	if (gtk_tree_model_get_iter(model, &iter, path)) {
            gtk_tree_model_get(model, &iter, 0, &ql, -1);
        }
        gtk_tree_path_free(path);
    }

    if (ql == NULL) {
        return NULL;
    }
    return ql->ql_Entries[controlIndex]->qe_Record;
}

void gtk_sfw_view_set_active_line(GtkSFWView *sfw, guint lineNum)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeSelection *sel;
    struct QueryLine *ql;

    g_return_if_fail(GTK_IS_SFW_VIEW(sfw));

    model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
    if (model) {
        if (gtk_tree_model_get_iter_first(model, &iter)) {
            do {
                ql = NULL;
                gtk_tree_model_get(model, &iter, 0, &ql, -1);
                if (ql && ql->ql_LineNum == lineNum) {
                    path = gtk_tree_model_get_path(model, &iter);
                    if (path) {
                        gtk_tree_view_set_cursor(GTK_TREE_VIEW(sfw), 
                                path, 0, FALSE);
                        gtk_tree_path_free(path);
                        gtk_tree_view_get_cursor(GTK_TREE_VIEW(sfw), &path, 0);
                        if (path) {
                            gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(sfw), 
                                    path, NULL, TRUE, .5, .0);
                            gtk_tree_path_free(path);
                        }
                    }
                    return;
                }
            } while (gtk_tree_model_iter_next(model, &iter));
	}
    }

    sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(sfw));
    gtk_tree_selection_unselect_all(sel);
}

guint gtk_sfw_view_get_active_line(GtkSFWView *sfw)
{
    GtkTreeSelection *selection;
    GtkTreeModel *model;
    GtkTreeIter iter;
    struct QueryLine *ql = NULL;

    g_return_val_if_fail(GTK_IS_SFW_VIEW(sfw), 0);

    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(sfw));
    if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
        gtk_tree_model_get(model, &iter, 0, &ql, -1);
    }

    if (ql == NULL) {
        return 0;
    }
    return ql->ql_LineNum;
}

gint gtk_sfw_view_get_active_column(GtkSFWView *sfw)
{
    GtkTreeViewColumn *column = 0;
    gint num = -1;

    gtk_tree_view_get_cursor(GTK_TREE_VIEW(sfw), NULL, &column);
    if (column) {
        num = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(column), "column"));
    }
    return num;
}

const signed char *gtk_sfw_view_get_column_order(GtkSFWView *sfw)
{
    GtkTreeViewColumn *col;
    gint i, num;
    guint numFields;

    g_return_val_if_fail(GTK_IS_SFW_VIEW(sfw), 0);

    memset(sfw->columnOrder, -1, sizeof(sfw->columnOrder));
    numFields = sfw->data? sfw->data->qgd_NumFields : 0;
    for (i = 0; i < numFields; i++) {
        col = gtk_tree_view_get_column(GTK_TREE_VIEW(sfw), i);
        if (col) {
            num = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(col), "column"));
            sfw->columnOrder[i] = num;
        }
    }

    return sfw->columnOrder;
}

static void gtk_sfw_view_clear(GtkSFWView *sfw)
{
    GtkTreeViewColumn *col;

    gtk_tree_view_set_model(GTK_TREE_VIEW(sfw), NULL);
    while ((col = gtk_tree_view_get_column(GTK_TREE_VIEW(sfw), 0)) != NULL) {
	gtk_tree_view_remove_column(GTK_TREE_VIEW(sfw), col);
    }

#if 0
    sfw->sort = SORT_COL_NONE;
    sfw->sort2 = SORT_COL_NONE;
#endif

    FreeQueryGuiData(sfw->data);
    sfw->data = NULL;
}

static void gtk_sfw_view_set_string_internal(GtkSFWView *sfw, const char *str)
{
    /* we use own label widget to avoid underlining of accel character */

    GtkListStore *store = gtk_list_store_new(1, G_TYPE_STRING);
    GtkWidget *label = gtk_label_new(str);
    GtkTreeViewColumn *column = gtk_tree_view_column_new();

    g_object_set(G_OBJECT(sfw),
	    "model", store,
	    "headers-visible", TRUE,
	    "enable-search", FALSE,
	    NULL);
    g_object_unref(G_OBJECT(store));
    
    gtk_widget_show(label);
    gtk_tree_view_column_set_widget(column, label);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sfw), column);
}

static gboolean sfwViewSearchEqualFunc(GtkTreeModel *model,
      	gint column, const gchar *key, GtkTreeIter *iter, gpointer search_data)
{
    gboolean retval = TRUE;
    struct QueryLine *ql = NULL;

    gtk_tree_model_get(model, iter, 0, &ql, -1);
    if (ql) {
        struct QueryEntry *qe = ql->ql_Entries[column];
	gchar *norm_str = g_utf8_normalize(qe->qe_String, -1, G_NORMALIZE_ALL);
	gchar *norm_key = g_utf8_normalize(key, -1, G_NORMALIZE_ALL);

	if (norm_str && norm_key) {
	    gchar *case_str = g_utf8_casefold(norm_str, -1);
	    gchar *case_key = g_utf8_casefold(norm_key, -1);

	    if (strstr(case_str, case_key)) {
		retval = FALSE;
	    }
	    g_free(case_key);
	    g_free(case_str);
	}
	g_free(norm_key);
	g_free(norm_str);
    }

    return retval;
}

static void queryEntryCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    gint col = GPOINTER_TO_INT(data);
    struct QueryLine *ql = NULL;
    gtk_tree_model_get(model, iter, 0, &ql, -1);
    if (ql) {
        if (col >= 0) {
            g_object_set(G_OBJECT(cell), "text", 
                    EmptyNullStr(ql->ql_Entries[col]->qe_String), NULL);
        } else {
            g_object_set(G_OBJECT(cell), "active", ql->ql_Checked, NULL);
        }
    }
}

#if 0
// This code doesn't work.  The arrow buttons do not show up.
static void onTreeViewColumnClicked(GtkTreeViewColumn *column, GtkSFWView *sfw)
{
    gint num = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(column), "column"));
    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
    GdkEvent *event = gtk_get_current_event();
    bool shift = FALSE;
    guint state;
    gint i;

    printf("sortable: %d\n", GTK_IS_TREE_SORTABLE(model));
    if (event) {
        switch (event->type) {
            case GDK_BUTTON_PRESS:
            case GDK_2BUTTON_PRESS:
            case GDK_3BUTTON_PRESS:
            case GDK_BUTTON_RELEASE:
                state = event->button.state;
                break;

            case GDK_KEY_PRESS:
            case GDK_KEY_RELEASE:
                state = event->key.state;
                break;

            default:
                state = 0;
                break;
        }

        shift = ((state & GDK_SHIFT_MASK) != 0);

        gdk_event_free(event);
    }

    if (sfw->sort == SORT_COL_NONE) {
        /* ignore qualifier if 1st sort column is unset */
        shift = FALSE;
    } 
    if (!shift) {
        if (SORT_COL(sfw->sort) == num) {
            sfw->sort = SORT_COL_FLIP(sfw->sort);
        } else {
            sfw->sort = num;
        }
        sfw->sort2 = SORT_COL_NONE;
    } else {
        if (SORT_COL(sfw->sort2) == num) {
            sfw->sort2 = SORT_COL_FLIP(sfw->sort2);
        } else {
            sfw->sort2 = num;
        }
    }

    /* update tree view columns */
    i = 0;
    while ((column = gtk_tree_view_get_column(GTK_TREE_VIEW(sfw), i)) != NULL) {
        num = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(column), "column"));
        if (num == SORT_COL(sfw->sort)) {
            gtk_tree_view_column_set_sort_indicator(column, TRUE);
            gtk_tree_view_column_set_sort_order(column, SORT_IS_UP(sfw->sort)?
                    GTK_SORT_ASCENDING : GTK_SORT_DESCENDING);
        } else if (num == SORT_COL(sfw->sort2)) {
            gtk_tree_view_column_set_sort_indicator(column, TRUE);
            gtk_tree_view_column_set_sort_order(column, SORT_IS_UP(sfw->sort2)?
                    GTK_SORT_ASCENDING : GTK_SORT_DESCENDING);
        } else {
            gtk_tree_view_column_set_sort_indicator(column, FALSE);
        }
        i++;
    }
    printf("clicked %s\n", shift? "shift" : "");
}
#endif

gint queryLineCmp(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, 
        gpointer data)
{
    GtkSFWView *sfw = GTK_SFW_VIEW(data);
    gint col;
    GtkSortType order;
    struct QueryLine *ql1, *ql2;
    gint result;

    gtk_tree_model_get(model, a, 0, &ql1, -1);
    gtk_tree_model_get(model, b, 0, &ql2, -1);

    if (!gtk_tree_sortable_get_sort_column_id(GTK_TREE_SORTABLE(model), 
                &col, &order)) {
        result = ql1->ql_LineNum - ql2->ql_LineNum;
    } else {
        result = QueryEntryCmp(ql1->ql_Entries[col], ql2->ql_Entries[col],
                sfw->data->qgd_FieldInfos[col].qfi_Type, TRUE);
    }

    return result;
}

static void onCellToggled(GtkCellRendererToggle *cell,
	gchar *path_str, GtkTreeModel *model)
{
    GtkTreePath *path = gtk_tree_path_new_from_string(path_str);
    GtkTreeIter  iter;
    struct QueryLine *ql = NULL;

    if (gtk_tree_model_get_iter(model, &iter, path)) {
        gtk_tree_model_get(model, &iter, 0, &ql, -1);
        if (ql) {
            ql->ql_Checked = !ql->ql_Checked;
            gtk_list_store_set(GTK_LIST_STORE(model), &iter, 0, ql, -1);
        }
    }
    gtk_tree_path_free(path);
}


const char *gtk_sfw_view_set_titles_body(GtkSFWView *sfw,
        const struct Element *titles, const struct Element *body)
{
    GtkListStore *store;
    GtkTreeIter iter;
    struct QueryGuiData *qgd = NULL;
    guint numFields, i;

    g_return_val_if_fail(GTK_IS_SFW_VIEW(sfw), GCS(MSG_Err_Internal));

    qgd = QueryGuiDataFromTitlesBody(titles, body);
    if (qgd == NULL) {
        gtk_sfw_view_clear(sfw);
        return GCS(MSG_Err_Mem);
    }

    /*
    ** Restrict number of fields to SFW_MAX_FIELDS.
    */
    if (qgd->qgd_NumFields > SFW_MAX_FIELDS) {
        qgd->qgd_NumFields = SFW_MAX_FIELDS;
    }

    if (QueryGuiDataCopyLineNumIfMatch(sfw->data, qgd)) {
        FreeQueryGuiData(qgd);
        return NULL;
    }

    /* Set new QueryGuiData */
    gtk_sfw_view_clear(sfw);
    sfw->data = qgd;

    numFields = sfw->data->qgd_NumFields;
    store = gtk_list_store_new(1, G_TYPE_POINTER);
    for (i = 0; i < sfw->data->qgd_NumLines; i++) {
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter, 0, sfw->data->qgd_Lines[i], -1);
    }

    gtk_tree_sortable_set_default_sort_func(GTK_TREE_SORTABLE(store),
            queryLineCmp, sfw, NULL);

    /*
    ** Add optional checkboxes
    */
    if (sfw->checkboxes) {
	GtkTreeViewColumn *column = gtk_tree_view_column_new();
	GtkCellRenderer *cell = gtk_cell_renderer_toggle_new();
	gtk_tree_view_column_pack_start(column, cell, TRUE);
	g_object_set_data(G_OBJECT(column), "column", GINT_TO_POINTER(-1));
	gtk_tree_view_column_set_cell_data_func(column, cell, 
		queryEntryCellDataFunc, GINT_TO_POINTER(-1), NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(sfw), column);
        g_signal_connect(cell, "toggled", G_CALLBACK(onCellToggled), store);
    }
    /*
    ** Add titles
    */
    for (i = 0; i < numFields; i++) {
        const char *title = sfw->data->qgd_Titles? 
            sfw->data->qgd_Titles->ql_Entries[i]->qe_String : NULL;
	GtkWidget *label = gtk_label_new(title);
	GtkTreeViewColumn *column = gtk_tree_view_column_new();
	GtkCellRenderer *cell = gtk_cell_renderer_text_new();
	gfloat xalign = sfw->data->qgd_FieldInfos[i].qfi_Format? 1.0f : 0.0f;

	g_object_set(G_OBJECT(label), 
                "visible", TRUE, 
                "xalign", xalign, 
                NULL);
	gtk_tree_view_column_set_widget(column, label);
	g_object_set(G_OBJECT(cell), 
		"xalign", xalign, 
		"mode", GTK_CELL_RENDERER_MODE_ACTIVATABLE,
		NULL);
	g_object_set(G_OBJECT(column), 
		"alignment", xalign,
		"expand", FALSE,
		"reorderable", TRUE,
		"resizable", TRUE,
		// "clickable", TRUE,
		NULL);
	g_object_set_data(G_OBJECT(column), "column", GINT_TO_POINTER(i));
	gtk_tree_view_column_pack_start(column, cell, TRUE);
	gtk_tree_view_column_set_cell_data_func(column, cell, 
		queryEntryCellDataFunc, GINT_TO_POINTER(i), NULL);
#if 0
        g_signal_connect(column, "clicked", 
                G_CALLBACK(onTreeViewColumnClicked), sfw);
#endif

        gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(store), i,
                queryLineCmp, sfw, NULL);
        gtk_tree_view_column_set_sort_column_id(column, i);

	gtk_tree_view_append_column(GTK_TREE_VIEW(sfw), column);
    }
    /*
    ** Add a dummy column if the last field added was right-aligned
    */
    if (numFields > 0 && sfw->data->qgd_FieldInfos[numFields - 1].qfi_Format) {
	GtkTreeViewColumn *column = gtk_tree_view_column_new();
	gtk_tree_view_append_column(GTK_TREE_VIEW(sfw), column);
    }

    g_object_set(G_OBJECT(sfw),
	    "model", store,
	    "headers-visible", sfw->showTitles,
	    "enable-search", TRUE,
	    "search-column", 0,
	    NULL);
    g_object_unref(G_OBJECT(store));
    gtk_tree_view_set_search_equal_func(GTK_TREE_VIEW(sfw), 
	    sfwViewSearchEqualFunc, NULL, NULL);

    return NULL;
}

const char *gtk_sfw_view_set_element(GtkSFWView *sfw,const struct Element *elem)
{
    const struct Element *titles = NULL, *body = NULL;

    if (elem) {
        if (elem->el_Type == TYPE_LIST) {
            titles = elem->el_Contents.list.head;
            body = elem->el_Contents.list.tail;
        } else {
            body = elem;
        }
    }

    return gtk_sfw_view_set_titles_body(sfw, titles, body);
}

void gtk_sfw_view_set_string(GtkSFWView *sfw, const char *str)
{
    g_return_if_fail(GTK_IS_SFW_VIEW(sfw));

    gtk_sfw_view_clear(sfw);
    sfw->data = QueryGuiDataFromString(str);
    if (str && sfw->data) {
	gtk_sfw_view_set_string_internal(sfw, str);
    }
}

void gtk_sfw_view_set_selected(GtkSFWView *sfw, const bool *selected)
{
    GtkTreeModel *model;

    g_return_if_fail(GTK_IS_SFW_VIEW(sfw));
    g_return_if_fail(selected != NULL);

    model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
    if (model) {
        GtkTreeIter iter;
        struct QueryLine *ql;
        int ind;

        if (gtk_tree_model_get_iter_first(model, &iter)) {
            do {
                ql = NULL;
                gtk_tree_model_get(model, &iter, 0, &ql, -1);
                if (ql) {
                    ind = ql->ql_LineNum - 1;
                    ql->ql_Checked = selected[ind];
                }
            } while (gtk_tree_model_iter_next(model, &iter));
	}
    }
}

void gtk_sfw_view_get_selected(GtkSFWView *sfw, bool *selected)
{
    GtkTreeModel *model;

    g_return_if_fail(GTK_IS_SFW_VIEW(sfw));
    g_return_if_fail(selected != NULL);

    model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
    if (model) {
        GtkTreeIter iter;
        struct QueryLine *ql;
        int ind;

        if (gtk_tree_model_get_iter_first(model, &iter)) {
            do {
                ql = NULL;
                gtk_tree_model_get(model, &iter, 0, &ql, -1);
                if (ql) {
                    ind = ql->ql_LineNum - 1;
                    selected[ind] = ql->ql_Checked;
                }
            } while (gtk_tree_model_iter_next(model, &iter));
	}
    }
}

void gtk_sfw_view_rem_record(GtkSFWView *sfw, const struct Record *rec)
{
    g_return_if_fail(GTK_IS_SFW_VIEW(sfw));

    if (rec) {
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
	if (model && sfw->data) {
	    gint i, num = sfw->data->qgd_NumFields;
            struct QueryLine *ql;
	    GtkTreeIter iter;

	    if (gtk_tree_model_get_iter_first(model, &iter)) {
		do {
                    ql = NULL;
                    gtk_tree_model_get(model, &iter, 0, &ql, -1);
                    if (ql) {
                        for (i=0; i < num; i++) {
                            if (ql->ql_Entries[i]->qe_Record == rec) {
                                ql->ql_Entries[i]->qe_Record = NULL;
                            }
                        }
                    }
		} while (gtk_tree_model_iter_next(model, &iter));
	    }
	}
    }
}

void gtk_sfw_view_rem_all_records(GtkSFWView *sfw)
{
    g_return_if_fail(GTK_IS_SFW_VIEW(sfw));

    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
    if (model && sfw->data) {
        gint i, num = sfw->data->qgd_NumFields;
        struct QueryLine *ql;
        GtkTreeIter iter;

        if (gtk_tree_model_get_iter_first(model, &iter)) {
            do {
                ql = NULL;
                gtk_tree_model_get(model, &iter, 0, &ql, -1);
                if (ql) {
                    for (i=0; i < num; i++) {
                        ql->ql_Entries[i]->qe_Record = NULL;
                    }
                }
            } while (gtk_tree_model_iter_next(model, &iter));
        }
    }
}

void gtk_sfw_view_get_entry(GtkSFWView *sfw, GtkTreePath *path, 
	GtkTreeViewColumn *column, 
	long *line, long *col, struct Field **field, struct Record **rec)
{
    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
    struct QueryLine *ql = NULL;
    gint num;

    g_return_if_fail(GTK_IS_SFW_VIEW(sfw));
    g_return_if_fail(GTK_IS_TREE_VIEW_COLUMN(column));

    if (line) {
        *line = -1;
    }
    if (field) {
        *field = NULL;
    }
    if (rec) {
	*rec = NULL;
    }

    if (model) {
        GtkTreeIter iter;
        if (gtk_tree_model_get_iter(model, &iter, path)) {
            gtk_tree_model_get(model, &iter, 0, &ql, -1);
        }
    }

    if (line && ql) {
        *line = ql->ql_LineNum;
    }

    num = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(column), "column"));
    if (col) {
	*col = num;
    }

    if (field && num >= 0 && sfw->data) {
        *field = sfw->data->qgd_FieldInfos[num].qfi_Field;
    }

    if (rec && num >= 0 && ql) {
        *rec = ql->ql_Entries[num]->qe_Record;
    }
}

const struct QueryLine *
gtk_sfw_view_get_iter_first (GtkSFWView *sfw, GtkTreeIter *iter)
{
    GtkTreeModel *model;
    const struct QueryLine *ql = NULL;

    g_return_val_if_fail(GTK_IS_SFW_VIEW(sfw), NULL);
    g_return_val_if_fail(iter != NULL, NULL);

    model = gtk_tree_view_get_model (GTK_TREE_VIEW(sfw));
    if (model) {
	if (gtk_tree_model_get_iter_first (model, iter)) {
	    gtk_tree_model_get(model, iter, 0, &ql, -1);
	}
    }

    return ql;
}

const struct QueryLine *
gtk_sfw_view_iter_next (GtkSFWView *sfw, GtkTreeIter *iter)
{
    GtkTreeModel *model;
    const struct QueryLine *ql = NULL;

    g_return_val_if_fail(GTK_IS_SFW_VIEW(sfw), NULL);
    g_return_val_if_fail(iter != NULL, NULL);

    model = gtk_tree_view_get_model (GTK_TREE_VIEW(sfw));
    if (model) {
	if (gtk_tree_model_iter_next (model, iter)) {
	    gtk_tree_model_get(model, iter, 0, &ql, -1);
	}
    }

    return ql;
}

const struct QueryLine *
gtk_sfw_view_get_query_line (GtkSFWView *sfw, GtkTreeIter *iter)
{
    GtkTreeModel *model;
    const struct QueryLine *ql = NULL;

    g_return_val_if_fail(GTK_IS_SFW_VIEW(sfw), NULL);
    g_return_val_if_fail(iter != NULL, NULL);

    model = gtk_tree_view_get_model (GTK_TREE_VIEW(sfw));
    if (model) {
        gtk_tree_model_get(model, iter, 0, &ql, -1);
    }

    return ql;
}

