/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
 * GdkPixbuf library - IFF/ILBM
 *
 * Authors: Steffen Gutmann
 *
 * Based on ilbmtoppm.c
 *
 * PCHG, CTBL and SHAM chunks not supported. 
 *
 * ----------------------from ilbmtoppm.c---------------------------------
** Copyright (C) 1989 by Jef Poskanzer.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
**
** Modified by Mark Thompson on 10/4/90 to accomodate 24-bit IFF files
** as used by ASDG, NewTek, etc.
**
** Modified by Ingo Wilken (Ingo.Wilken@informatik.uni-oldenburg.de)
**  20/Jun/93:
**  - row-by-row operation
**  - better de-interleave algorithm
**  - colormap files
**  - direct color
**  04/Oct/93:
**  - multipalette support (PCHG chunk)
**  - options -ignore, -isham, -isehb and -adjustcolors
**  22/May/94:
**  - minor change: check first for 24 planes, then for HAM
**  21/Sep/94:
**  - write mask plane to a file if -maskfile option used
**  - write colormap file
**  - added sliced HAM/dynamic HAM/dynamic Hires multipalette formats (SHAM, CTBL chunk)
**  - added color lookup tables (CLUT chunk)
**  - major rework of colormap/multipalette handling
**  - now uses numeric IFF IDs
**  24/Oct/94:
**  - transparentColor support
**  - added RGBN/RGB8 image types
**  - 24-bit & direct color modified to n-bit deep ILBM
**  22/Feb/95:
**  - direct color (DCOL) reimplemented
**  29/Mar/95
**  - added IFF-PBM format
 * ----------------------end from ilbmtoppm.c---------------------------------
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*

Known bugs:
        * None
*/

#define GDK_PIXBUF_ENABLE_BACKEND
#include <gdk-pixbuf/gdk-pixbuf-io.h>
#include <glib/gi18n.h>
#include <string.h>
#define MODULE_ENTRY(type,function) _gdk_pixbuf__ ## type ## _ ## function

#ifdef DEBUG
#define DPRINTF(x) printf x
#else
#define DPRINTF(x)
#endif

#define MAKE_SHORT(a, b) \
    ((guint16)(a)<<8 | (guint16)(b))
#define MAKE_LONG(a, b, c, d) \
    ((guint32)(a)<<24 | (guint32)(b)<<16 | (guint32)(c)<<8 | (guint32)(d))

#define RowBytes(cols)          ((((cols) + 15) / 16) * 2)

/* definitions for BMHD */

typedef struct {
    gushort w, h;
    gshort x, y;
    guchar nPlanes, masking, compression, flags;
    gushort transparentColor;
    guchar xAspect, yAspect;
    gshort pageWidth, pageHeight;
} BitMapHeader;
#define BitMapHeaderSize    20

#define BMHD_FLAGS_CMAPOK       (1<<7)      /* valid 8bit colormap */

#define mskNone                 0
#define mskHasMask              1
#define mskHasTransparentColor  2
#define mskLasso                3       /* not supported */
#define mskMAXKNOWN             mskLasso

#define cmpNone                 0
#define cmpByteRun1             1
#define cmpMAXKNOWN             cmpByteRun1

/* definitions for CAMG */

#define CAMGChunkSize       4

#define vmLACE              0x0004
#define vmEXTRA_HALFBRITE   0x0080
#define vmHAM               0x0800
#define vmHIRES             0x8000

#define HAMCODE_CMAP      0     /* look up color in colormap */
#define HAMCODE_BLUE      1     /* new blue component */
#define HAMCODE_RED       2     /* new red component */
#define HAMCODE_GREEN     3     /* new green component */

#if 0 /* PCHG chunks not supported */
/* multipalette PCHG chunk definitions */

/* get number of longwords in line mask from PCHG.LineCount */
#define MaskLongWords(x)    (((x) + 31) / 32)

typedef struct {
    unsigned short  Compression;
    unsigned short  Flags;
    short           StartLine;      /* may be negative */
    unsigned short  LineCount;
    unsigned short  ChangedLines;
    unsigned short  MinReg;
    unsigned short  MaxReg;
    unsigned short  MaxChanges;
    unsigned long   TotalChanges;
} PCHGHeader;
#define PCHGHeaderSize      20

/* Compression modes */
#define PCHG_COMP_NONE      0
#define PCHG_COMP_HUFFMAN   1

/* Flags */
#define PCHGF_12BIT         (1 << 0)    /* use SmallLineChanges */
#define PCHGF_32BIT         (1 << 1)    /* use BigLineChanges */
#define PCHGF_USE_ALPHA     (1 << 2)    /* meaningful only if PCHG_32BIT is on:
                                           use the Alpha channel info */
typedef struct {
    unsigned long   CompInfoSize;
    unsigned long   OriginalDataSize;
} PCHGCompHeader;
#define PCHGCompHeaderSize  8
#endif

#define CLUTSize    (256+4+4)

/* types */
#define CLUT_MONO   0
#define CLUT_RED    1
#define CLUT_GREEN  2
#define CLUT_BLUE   3
#define CLUT_HUE    4   /* not supported */
#define CLUT_SAT    5   /* not supported */


/* unofficial DCOL chunk for direct-color */

typedef struct {
    guchar r, g, b, pad1;
} DirectColor;
#define DirectColorSize     4

/* IFF chunk IDs */

typedef unsigned long   IFF_ID;

#define MAKE_ID(a, b, c, d) MAKE_LONG(a, b, c, d)

#define ID_FORM     MAKE_ID('F', 'O', 'R', 'M')     
    /* EA IFF 85 group identifier */
#define ID_CAT      MAKE_ID('C', 'A', 'T', ' ')     
    /* EA IFF 85 group identifier */
#define ID_LIST     MAKE_ID('L', 'I', 'S', 'T')     
    /* EA IFF 85 group identifier */
#define ID_PROP     MAKE_ID('P', 'R', 'O', 'P')     
    /* EA IFF 85 group identifier */
#define ID_END      MAKE_ID('E', 'N', 'D', ' ')     
    /* unofficial END-of-FORM identifier (see Amiga RKM Devices Ed.3
       page 376) */
#define ID_ILBM     MAKE_ID('I', 'L', 'B', 'M')     
    /* EA IFF 85 raster bitmap form */
#define ID_DEEP     MAKE_ID('D', 'E', 'E', 'P')     
    /* Chunky pixel image files (Used in TV Paint) */
#define ID_RGB8     MAKE_ID('R', 'G', 'B', '8')     
    /* RGB image forms, Turbo Silver (Impulse) */
#define ID_RGBN     MAKE_ID('R', 'G', 'B', 'N')     
    /* RGB image forms, Turbo Silver (Impulse) */
#define ID_PBM      MAKE_ID('P', 'B', 'M', ' ')     
    /* 256-color chunky format (DPaint 2 ?) */
#define ID_ACBM     MAKE_ID('A', 'C', 'B', 'M')     
    /* Amiga Contiguous Bitmap (AmigaBasic) */

/* generic */

#define ID_FVER     MAKE_ID('F', 'V', 'E', 'R')     
    /* AmigaOS version string */
#define ID_JUNK     MAKE_ID('J', 'U', 'N', 'K')     
    /* always ignore this chunk */
#define ID_ANNO     MAKE_ID('A', 'N', 'N', 'O')     
    /* EA IFF 85 Generic Annotation chunk */
#define ID_AUTH     MAKE_ID('A', 'U', 'T', 'H')     
    /* EA IFF 85 Generic Author chunk */
#define ID_CHRS     MAKE_ID('C', 'H', 'R', 'S')     
    /* EA IFF 85 Generic character string chunk */
#define ID_NAME     MAKE_ID('N', 'A', 'M', 'E')     
    /* EA IFF 85 Generic Name of art, music, etc. chunk */
#define ID_TEXT     MAKE_ID('T', 'E', 'X', 'T')     
    /* EA IFF 85 Generic unformatted ASCII text chunk */
#define ID_copy     MAKE_ID('(', 'c', ')', ' ')     
/* EA IFF 85 Generic Copyright text chunk */

/* ILBM chunks */

#define ID_BMHD     MAKE_ID('B', 'M', 'H', 'D')     
    /* ILBM BitmapHeader */
#define ID_CMAP     MAKE_ID('C', 'M', 'A', 'P')     
    /* ILBM 8bit RGB colormap */
#define ID_GRAB     MAKE_ID('G', 'R', 'A', 'B')     
    /* ILBM "hotspot" coordiantes */
#define ID_DEST     MAKE_ID('D', 'E', 'S', 'T')     
    /* ILBM destination image info */
#define ID_SPRT     MAKE_ID('S', 'P', 'R', 'T')     
    /* ILBM sprite identifier */
#define ID_CAMG     MAKE_ID('C', 'A', 'M', 'G')     
    /* Amiga viewportmodes */
#define ID_BODY     MAKE_ID('B', 'O', 'D', 'Y')     
    /* ILBM image data */
#define ID_CRNG     MAKE_ID('C', 'R', 'N', 'G')     
    /* color cycling */
#define ID_CCRT     MAKE_ID('C', 'C', 'R', 'T')     
    /* color cycling */
#define ID_CLUT     MAKE_ID('C', 'L', 'U', 'T')     
    /* Color Lookup Table chunk */
#define ID_DPI      MAKE_ID('D', 'P', 'I', ' ')     
    /* Dots per inch chunk */
#define ID_DPPV     MAKE_ID('D', 'P', 'P', 'V')     
    /* DPaint perspective chunk (EA) */
#define ID_DRNG     MAKE_ID('D', 'R', 'N', 'G')     
    /* DPaint IV enhanced color cycle chunk (EA) */
#define ID_EPSF     MAKE_ID('E', 'P', 'S', 'F')     
    /* Encapsulated Postscript chunk */
#define ID_CMYK     MAKE_ID('C', 'M', 'Y', 'K')     
    /* Cyan, Magenta, Yellow, & Black color map (Soft-Logik) */
#define ID_CNAM     MAKE_ID('C', 'N', 'A', 'M')     
    /* Color naming chunk (Soft-Logik) */
#define ID_PCHG     MAKE_ID('P', 'C', 'H', 'G')     
    /* Line by line palette control information (Sebastiano Vigna) */
#define ID_PRVW     MAKE_ID('P', 'R', 'V', 'W')     
    /* A mini duplicate ILBM used for preview (Gary Bonham) */
#define ID_XBMI     MAKE_ID('X', 'B', 'M', 'I')     
    /* eXtended BitMap Information (Soft-Logik) */
#define ID_CTBL     MAKE_ID('C', 'T', 'B', 'L')     
    /* Newtek Dynamic Ham color chunk */
#define ID_DYCP     MAKE_ID('D', 'Y', 'C', 'P')     
    /* Newtek Dynamic Ham chunk */
#define ID_SHAM     MAKE_ID('S', 'H', 'A', 'M')     
    /* Sliced HAM color chunk */
#define ID_ABIT     MAKE_ID('A', 'B', 'I', 'T')     
    /* ACBM body chunk */
#define ID_DCOL     MAKE_ID('D', 'C', 'O', 'L')     
    /* unofficial direct color */
#define ID_DPPS     MAKE_ID('D', 'P', 'P', 'S')
    /* ? */
#define ID_TINY     MAKE_ID('T', 'I', 'N', 'Y')
    /* ? */

/* other stuff */

#define MAXPLANES       16
typedef gushort  rawtype;

#define MAXCMAPCOLORS   (1 << MAXPLANES)
#define MAXCOLVAL       255     /* max value of color component */

typedef guchar pixval;

typedef struct {
    pixval r, g, b;
} pixel;

#if 0 /* not supported */
typedef struct {
    int reg;            /* color register to change */
    pixval r, g, b;     /* new colors for register */
} PaletteChange;
#endif

typedef struct {
    pixel *color;
    int    ncolors;
    /* lookup tables */
    guchar *red_lut;
    guchar *green_lut;
    guchar *blue_lut;
    guchar *mono_lut;
#if 0 /* not supported */
    /* multipalette stuff */
    PaletteChange *mp_init;
    PaletteChange **mp_change;
    int mp_rows;                /* # of rows in change array */
    int mp_type;                /* see below, higher types preferred */
    int mp_flags;
    IFF_ID  mp_id;
#endif
} ColorMap;

#define HAS_COLORMAP(cmap)      ((cmap) && (cmap)->color)

#define FACTOR_4BIT     17      /* scale factor maxval 15 -> maxval 255 */

/* Progressive loading */

struct ilbm_progressive_state {
    GdkPixbufModulePreparedFunc prepared_func;
    GdkPixbufModuleUpdatedFunc updated_func;
    gpointer user_data;

    enum {
	LS_START, LS_TYPE, LS_CHUNK_HEADER, LS_CHUNK_BODY, LS_BODY, LS_END
    } state;
    
    IFF_ID iff_id;
    glong bytes_to_go;	/* bytes to read excluding first 8 byte header */
    
    guint chunk_header_len;
    guchar chunk_header_buf[8];
    IFF_ID chunk_id;
    guint chunk_size;
    guint chunk_len;
    guchar *chunk_buf;
    
    BitMapHeader bmhd;
    ColorMap cmap;
    gboolean has_dcol;
    DirectColor dcol;
    gulong view_port_modes;
    gboolean has_transp_color;
    pixel transp_color;
    gboolean body_chunk_processed;
    
    GdkPixbuf *pixbuf;	/* Our "target" */
};

static gpointer
gdk_pixbuf__ilbm_image_begin_load(GdkPixbufModuleSizeFunc size_func, 
                                  GdkPixbufModulePreparedFunc prepared_func,
				  GdkPixbufModuleUpdatedFunc updated_func,
                                  gpointer user_data,
                                  GError **error);

static gboolean gdk_pixbuf__ilbm_image_stop_load(gpointer data, GError **error);
static gboolean gdk_pixbuf__ilbm_image_load_increment(gpointer data,
                                                      const guchar * buf,
                                                      guint size,
                                                      GError **error);

/* 
 * func - called when we have pixmap created (but no image data)
 * user_data - passed as arg 1 to func
 * return context (opaque to user)
 */

static gpointer
gdk_pixbuf__ilbm_image_begin_load(GdkPixbufModuleSizeFunc size_func, 
                                  GdkPixbufModulePreparedFunc prepared_func,
                                  GdkPixbufModuleUpdatedFunc updated_func,
                                  gpointer user_data,
                                  GError **error)
{
    struct ilbm_progressive_state *context;

    context = g_new0(struct ilbm_progressive_state, 1);
    context->prepared_func = prepared_func;
    context->updated_func = updated_func;
    context->user_data = user_data;
    
    context->state = LS_START;
    context->bytes_to_go = 8;	/* we need at least the 8 byte header */
    
    return (gpointer) context;
}

/*
 * context - returned from image_begin_load
 *
 * free context, unref gdk_pixbuf
 */
static gboolean gdk_pixbuf__ilbm_image_stop_load(gpointer data,
                                                 GError **error)
{
    struct ilbm_progressive_state *context =
	(struct ilbm_progressive_state *) data;
    gboolean ok = TRUE;

    if (context == NULL) {
	g_set_error (error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_FAILED,
		_("Missing context"));
	return FALSE;
    }

    if (context->state != LS_END || !context->body_chunk_processed) {
	g_set_error (error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Missing image data"));
	ok = FALSE;
    }

    if (context->pixbuf) {
	g_object_unref(context->pixbuf);
    }

    g_free(context->chunk_buf);

    /* free cmap contents */
    g_free(context->cmap.color);
    g_free(context->cmap.red_lut);
    g_free(context->cmap.green_lut);
    g_free(context->cmap.blue_lut);
    g_free(context->cmap.mono_lut);

    g_free(context);

    return ok;
}

static gboolean fill_buf(guchar *tbuf, guint *tlen, guint tsize,
	const guchar **buf, guint *size)
{
    guint need = tsize - *tlen;
    guint have = MIN(need, *size);

    if (tbuf) {
	memcpy(&tbuf[*tlen], *buf, have);
    }
    *tlen += have;
    *buf += have;
    *size -= have;

    return (have == need);
}

static gboolean fill_chunk_header(struct ilbm_progressive_state *context, 
	const guchar **buf, guint *size)
{
    guchar *h = context->chunk_header_buf;

    if (!fill_buf(h, &context->chunk_header_len, 8, buf, size)) {
	return FALSE;
    }

    context->chunk_id = MAKE_ID(h[0], h[1], h[2], h[3]);
    context->chunk_size = MAKE_LONG(h[4], h[5], h[6], h[7]);

    context->chunk_header_len = 0;	/* clear for next fill */

    return TRUE;
}

static gboolean prepare_chunk_buffer(struct ilbm_progressive_state *context, 
	gboolean alloc, GError **error)
{
    if (context->chunk_buf) {
	g_free(context->chunk_buf);
	context->chunk_buf = NULL;
    }
    if (context->chunk_size > 0x10000000) {
	g_set_error (error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Invalid chunk size (> 0x10000000)"));
	return FALSE;
    }
    context->chunk_len = 0;
    if (!alloc) {
	return TRUE;
    }

    context->chunk_buf = (guchar *)g_try_malloc(context->chunk_size);
    if (context->chunk_buf == NULL) {
	g_set_error (error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY,
		_("Not enough memory to allocate chunk buffer"));
	return FALSE;
    }

    return TRUE;
}

static gboolean ilbm_start(struct ilbm_progressive_state *context, 
	const guchar **buf, guint *size, GError **error)
{
    if (!fill_chunk_header(context, buf, size)) {
	return TRUE;				/* we need more data */
    }

    if (context->chunk_id != ID_FORM) {
	g_set_error (error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Input is not a FORM type IFF file"));
	return FALSE;
    }

    context->bytes_to_go += context->chunk_size;
    context->state = LS_TYPE;

    return TRUE;
}

static gboolean ilbm_type(struct ilbm_progressive_state *context, 
	const guchar **buf, guint *size, GError **error)
{
    guchar *h = context->chunk_header_buf;
    IFF_ID id;

    if (!fill_buf(h, &context->chunk_header_len, 4, buf, size)) {
	return TRUE;			/* not enough bytes yet */
    }

    id = MAKE_ID(h[0], h[1], h[2], h[3]);
    context->chunk_header_len = 0;	/* clear for next fill */

    if (id != ID_ILBM && id != ID_RGBN && id != ID_RGB8 && id != ID_PBM) {
	g_set_error (error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Input is not an ILBM, RGBN, RGB8 or PBM type IFF file"));
	return FALSE;
    }

    context->iff_id = id;
    context->state = LS_CHUNK_HEADER;
    
    return TRUE;
}

static gboolean ilbm_chunk_header(struct ilbm_progressive_state *context, 
	const guchar **buf, guint *size, GError **error)
{
    IFF_ID id;
    gboolean needed;

    if (!fill_chunk_header(context, buf, size)) {
	return TRUE;				/* we need more data */
    }
    
    id = context->chunk_id;
    /* printf("chunk: %4.4s, size: %d\n", context->chunk_header_buf,
	    context->chunk_size); */
    if (id == ID_END) {
        /* 
	** END chunks are not officially supported in IFF, but
        ** suggested as a future expansion for stream-writing,
        ** see Amiga RKM Devices, 3rd Ed, page 376 
        */
	context->state = LS_END;
	return TRUE;
    }

    switch (id) {
	case ID_BMHD:
        case ID_CMAP:
        case ID_CMYK:
        case ID_CLUT:
        case ID_CAMG:
        case ID_DCOL:
        case ID_BODY: 
	    needed = TRUE;
	    break;
	    
        default:			/* ignore unknown chunks */
	    needed = FALSE;
            break;
    }

    if (!prepare_chunk_buffer(context, needed, error)) {
	return FALSE;
    }

    context->state = LS_CHUNK_BODY;
    return TRUE;
}

static char get_byte(const guchar **b)
{
    return *((*b)++);
}

static short get_short(const guchar **b)
{
    guchar c1 = get_byte(b);
    guchar c2 = get_byte(b);

    return MAKE_SHORT(c1, c2);
}

static short get_long(const guchar **b)
{
    guchar c1 = get_byte(b);
    guchar c2 = get_byte(b);
    guchar c3 = get_byte(b);
    guchar c4 = get_byte(b);

    return MAKE_LONG(c1, c2, c3, c4);
}

static gboolean process_bmhd(struct ilbm_progressive_state *context, 
	GError **error)
{
    BitMapHeader *bmhd = &context->bmhd;
    const guchar *b = context->chunk_buf;
    gboolean has_alpha;

    if (context->pixbuf) {		/* only one BMHD chunk allowed */
	g_set_error(error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Multiple BMHD chunks"));
	return FALSE;
    }

    if (context->chunk_size != BitMapHeaderSize ) {
	g_set_error(error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Wrong BitMapHeaderSize"));
	return FALSE;
    }

    bmhd->w = get_short(&b);        	/* cols */
    bmhd->h = get_short(&b);        	/* rows */
    bmhd->x = get_short(&b);        	/* ignored */
    bmhd->y = get_short(&b);        	/* ignored */
    bmhd->nPlanes = get_byte(&b);
    bmhd->masking = get_byte(&b);
    bmhd->compression = get_byte(&b);
    bmhd->flags = get_byte(&b);
    bmhd->transparentColor = get_short(&b);
    bmhd->xAspect = get_byte(&b);	/* ignored */
    bmhd->yAspect = get_byte(&b);	/* ignored */
    bmhd->pageWidth  = get_short(&b);  	/* ignored */
    bmhd->pageHeight = get_short(&b);  	/* ignored */

    if (bmhd->w == 0) {
    	g_set_error (error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Image has zero width"));
	return FALSE;
    }

    if (bmhd->h == 0) {
	g_set_error (error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Image has zero height"));
	return FALSE;
    }

    if (bmhd->nPlanes == 0) {
	g_set_error (error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Image has no planes"));
	return FALSE;
    }

    has_alpha = (bmhd->masking == mskHasMask || 
	    bmhd->masking == mskHasTransparentColor);
    has_alpha = FALSE;			/* for now */
		      
    context->pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, has_alpha, 8, 
	    bmhd->w, bmhd->h);
		    
    if (!context->pixbuf) {
	g_set_error (error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY,
		_("Not enough memory to load image"));
	return FALSE;
    }

    if (context->prepared_func) {
	context->prepared_func(context->pixbuf, NULL, context->user_data);
    }

    return TRUE;
}

static gboolean process_cmap(struct ilbm_progressive_state *context, 
	GError **error)
{
    ColorMap *cmap = &context->cmap;
    const guchar *b = context->chunk_buf;
    glong colors = context->chunk_size / 3;
    glong i;
    pixel *p;

    if (colors <= 0) {
	return TRUE;			/* ignore empty color map */
    }

    g_free(cmap->color);		/* in case of multiple CMAP chunks */
    cmap->color = g_new(pixel, colors);
    cmap->ncolors = colors;

    p = cmap->color;
    for (i = 0; i < colors; i++, p++) {
        p->r = get_byte(&b);
        p->g = get_byte(&b);
        p->b = get_byte(&b);
    }

    return TRUE;
}

static gboolean process_cmyk(struct ilbm_progressive_state *context, 
	GError **error)
{
    ColorMap *cmap = &context->cmap;
    const guchar *b = context->chunk_buf;
    glong colors;
    glong i;
    pixel *p;

    if (HAS_COLORMAP(cmap) ){      	/* prefer RGB color map */
	return TRUE;
    }

    colors = context->chunk_size / 4;
    if (colors <= 0) {
	return TRUE;			/* ignore empty color map */
    }

    g_free(cmap->color);		/* in case of multiple CMYK chunks */
    cmap->color = g_new(pixel, colors);
    cmap->ncolors = colors;

    p = cmap->color;
    for (i = 0; i < colors; i++, p++) {
        int c, m, y, k;
        c = get_byte(&b);
        m = get_byte(&b);
        y = get_byte(&b);
        k = get_byte(&b);
	p->r = MAXCOLVAL - MIN(MAXCOLVAL, c*(MAXCOLVAL-k)/MAXCOLVAL+k); 
        p->g = MAXCOLVAL - MIN(MAXCOLVAL, m*(MAXCOLVAL-k)/MAXCOLVAL+k);
	p->b = MAXCOLVAL - MIN(MAXCOLVAL, y*(MAXCOLVAL-k)/MAXCOLVAL+k);
    }

    return TRUE;
}

static gboolean process_clut(struct ilbm_progressive_state *context, 
	GError **error)
{
    ColorMap *cmap = &context->cmap;
    const guchar *b = context->chunk_buf;
    guchar *lut;
    gint i;
    glong type;

    if (context->chunk_size != CLUTSize) {
	g_set_error(error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Wrong size of CLUT chunk"));
	return FALSE;
    }

    type = get_long(&b);
    get_long(&b);				/* skip reserved field */

    lut = g_new(guchar, 256);
    for (i = 0; i < 256; i++) {
	lut[i] = get_byte(&b);
    }

    switch (type) {
        case CLUT_RED:
            cmap->red_lut = lut;
            break;
        case CLUT_GREEN:
            cmap->green_lut = lut;
            break;
        case CLUT_BLUE:
            cmap->blue_lut = lut;
            break;
        case CLUT_MONO:
            cmap->mono_lut = lut;
            break;
        default:
            g_warning("%4.4s type %ld not supported", 
		    context->chunk_header_buf, type);
            g_free(lut);
	    break;
    }

    return TRUE;
}

static gboolean process_camg(struct ilbm_progressive_state *context, 
	GError **error)
{
    const guchar *b = context->chunk_buf;
    const gulong viewportmask = 0xffffffff;

    if (context->chunk_size != CAMGChunkSize) {
	g_set_error(error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Wrong size of CAMG chunk"));
	return FALSE;
    }

    context->view_port_modes = get_long(&b) & viewportmask;

    return TRUE;
}

static gboolean process_dcol(struct ilbm_progressive_state *context, 
	GError **error)
{
    const guchar *b = context->chunk_buf;

    if (context->chunk_size != DirectColorSize) {
	g_set_error(error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Wrong size of DCOL chunk"));
	return FALSE;
    }

    context->dcol.r = get_byte(&b);
    context->dcol.g = get_byte(&b);
    context->dcol.b = get_byte(&b);
    context->has_dcol = TRUE;

    return TRUE;
}

static void check_cmap(struct ilbm_progressive_state *context)
{
    BitMapHeader *bmhd = &context->bmhd;
    ColorMap *cmap = &context->cmap;
    pixval colmaxval = 0;
    gboolean shifted = TRUE;
    gint i, r, g, b;

    if (bmhd->nPlanes > 0) {
	if (bmhd->masking == mskHasTransparentColor || 
		bmhd->masking == mskLasso) {
	    gint transpIndex = bmhd->transparentColor;
	    if (transpIndex >= cmap->ncolors) {
		context->transp_color.r = 0;
		context->transp_color.g = 0;
		context->transp_color.b = 0;
	    } else {
		context->transp_color = cmap->color[transpIndex];
            }
	    context->has_transp_color = TRUE;
        }

        if (bmhd->flags & BMHD_FLAGS_CMAPOK) {
            return;
	}
    }

    if (!HAS_COLORMAP(cmap)) {
	return;
    }

    for (i = 0; i < cmap->ncolors; i++) {
        r = cmap->color[i].r;
        g = cmap->color[i].g;
        b = cmap->color[i].b;
        if (r > colmaxval) {
	    colmaxval = r;
	}
        if (g > colmaxval) {
	    colmaxval = g;
	}
        if (b > colmaxval) {
	    colmaxval = b;
	}
        if ((r | g | b) & 0x0f) {
	    shifted = FALSE;
	}
    }

    if (shifted || colmaxval <= 15) {
	for (i = 0; i < cmap->ncolors; i++) {
	    r = cmap->color[i].r;
	    g = cmap->color[i].g;
	    b = cmap->color[i].b;
	    if (shifted) {
		r >>= 4;
	       	g >>= 4;
	   	b >>= 4;
       	    }
	    r *= FACTOR_4BIT;
	    g *= FACTOR_4BIT;
    	    b *= FACTOR_4BIT;
	    cmap->color[i].r = r;
	    cmap->color[i].g = g;
	    cmap->color[i].b = b;
        }
    }
}

/****************************************************************************
 ILBM functions
 ****************************************************************************/

static gboolean read_bytes(const guchar **b, guint *size, 
	guchar *dst, gint bytes)
{
    if ((guint)bytes > *size) {
	return FALSE;
    }

    memcpy(dst, *b, bytes);
    *b += bytes;
    *size -= bytes;

    return TRUE;
}

static gboolean read_ilbm_plane(const guchar **b, guint *size, guchar *ilbm_row,
	gint bytes, gboolean compression)
{
    switch (compression) {
        case cmpNone:
            return read_bytes(b, size, ilbm_row, bytes);

        case cmpByteRun1:
	{
            guchar *ubp = ilbm_row;
            do {
		guchar byte;

		if (*size < 1) {
		    return FALSE;
		}
                byte = get_byte(b);
		--*size;
                if (byte <= 127) {
                    gint j = byte;
		    if (bytes < j+1 || *size < (guint)(j+1)) {
			return FALSE;
		    }
                    bytes -= (j+1);
                    *size -= j+1;
                    for (; j >= 0; j--) {
                        *ubp++ = get_byte(b);
		    }
                } else if (byte != 128) {
                    gint j = 256 - byte;
		    if (bytes < j+1 || *size < 1) {
			return FALSE;
		    }
                    bytes -= (j+1);
                    *size -= 1;
                    byte = get_byte(b);
                    for (; j >= 0; j--) {
                        *ubp++ = byte;
		    }
                }
                /* 128 is a NOP */
            }
            while (bytes > 0);
            break;
	}

	default:
	    return FALSE;
    }

    return TRUE;
}

static gboolean decode_row(const guchar **b, guint *size, 
	rawtype *raw_row, guchar *ilbm_row, int nPlanes, 
	const BitMapHeader *bmhd)
{
    gint cols = bmhd->w;
    gint bytes = RowBytes(cols);
    gint plane;

    for (plane = 0; plane < nPlanes; plane++) {
	guchar high = 1 << (plane - 8);
	guchar low = 1 << plane;
        rawtype mask = MAKE_SHORT(high, low);
	guchar *ilp = ilbm_row;
	rawtype *chp = raw_row;
	guchar cmask = 1 << 7;
	gint col;

        if (!read_ilbm_plane(b, size, ilbm_row, bytes, bmhd->compression)) {
	    return FALSE;
	}

        for (col = 0; col < cols; col++, cmask >>= 1, chp++) {
            if (cmask == 0) {
                cmask = 1 << 7;
                ilp++;
            }
            if (*ilp & cmask) {
                *chp |= mask;
	    } else {
                *chp &= ~mask;
	    }
        }
    }

    return TRUE;
}

static gboolean decode_mask(const guchar **b, guint *size,
	const rawtype *raw_row, guchar *mask_row, guchar *ilbm_row, 
	const BitMapHeader *bmhd)
{
    switch (bmhd->masking) {
        case mskNone:
	    if (mask_row) {
		memset(mask_row, 255, bmhd->w);
	    }
            break;

        case mskHasMask:        /* mask plane */
            if (!read_ilbm_plane(b, size, ilbm_row, RowBytes(bmhd->w), 
			bmhd->compression)) {
		return FALSE;
	    }
            if (mask_row) {
                guchar *ilp = mask_row;
		guchar cmask = 1 << 7;
		gint cols = bmhd->w, col;

                for (col = 0; col < cols; col++, cmask >>= 1) {
                    if (cmask == 0) {
                        cmask = 1 << 7;
                        ilp++;
                    }
		}
		mask_row[col] = (*ilp & cmask)? 255 : 0;
            }
            break;

        case mskHasTransparentColor:
            if (mask_row) {
		gint cols = bmhd->w, col;
		gshort transc = bmhd->transparentColor;

                for (col = 0; col < cols; col++) {
		    mask_row[col] = (raw_row[col] == transc)? 0 : 255;
                }
            }
            break;

        case mskLasso:
            /* not supported yet */
	    return FALSE;

        default:
	    return FALSE;
    }

    return TRUE;
}

static pixval lookup_color(const guchar *lut, pixval color)
{
    return lut? lut[color] : color;
}

static void ehbcmap(ColorMap *cmap)
{
    gint ncolors = cmap->ncolors, i;
    pixel *color = cmap->color = g_renew(pixel, cmap->color, ncolors * 2);

    for (i = 0; i < ncolors; i++) {
	color[ncolors + i].r = color[i].r / 2;
	color[ncolors + i].g = color[i].g / 2;
	color[ncolors + i].b = color[i].b / 2;
    }
    cmap->ncolors *= 2;
}

#if 0
static pixval lut_maxval(const ColorMap *cmap, pixval maxval) 
{
    pixval retval = maxval;
    gint i;

    for (i=0; i < maxval; i++) {
	pixval c = lookup_color(cmap->red_lut, i);
	retval = MAX(retval, c);
	c = lookup_color(cmap->green_lut, i);
	retval = MAX(retval, c);
	c = lookup_color(cmap->blue_lut, i);
	retval = MAX(retval, c);
    }
    
    return retval;
}
#endif

static void get_color(const ColorMap *cmap, gshort idx, pixel *p)
{
    if (HAS_COLORMAP(cmap)) {
        if (idx >= cmap->ncolors) {
            g_warning("color index out of range: %d (max %d), "
		    "using black color", idx, cmap->ncolors);
	    p->r = p->g = p->b = 0;
	} else {
	    *p = cmap->color[idx];
	    p->r = lookup_color(cmap->red_lut, p->r);
	    p->g = lookup_color(cmap->green_lut, p->g);
	    p->b = lookup_color(cmap->blue_lut, p->b);
	}
    }
    else {
	p->r = p->g = p->b = lookup_color(cmap->mono_lut, (pixval)idx);
    }
}

static gboolean read_dcol(struct ilbm_progressive_state *context, 
	GError **error)
{
    return TRUE;
}

static gboolean read_ham(struct ilbm_progressive_state *context, 
	GError **error)
{
    return TRUE;
}

static gboolean read_std(struct ilbm_progressive_state *context, 
	GError **error)
{
    BitMapHeader *bmhd = &context->bmhd;
    ColorMap *cmap = &context->cmap;
    gint rows = bmhd->h, cols = bmhd->w;
    gint row, col;
    const guchar *b = context->chunk_buf;
    guint size = context->chunk_size;
    guchar *ilbm_row;
    rawtype *raw_row;
    guchar *mask_row = NULL;
    gboolean ok = TRUE;
    GdkPixbuf *pixbuf = context->pixbuf;
    gint n_channels = gdk_pixbuf_get_n_channels(pixbuf);
    gboolean has_alpha = gdk_pixbuf_get_has_alpha(pixbuf);
    gint rowstride = gdk_pixbuf_get_rowstride(pixbuf);
    guchar *pixels = gdk_pixbuf_get_pixels(pixbuf);

    if (context->view_port_modes & vmHAM) {
	return read_ham(context, error);
    }

    if (bmhd->nPlanes > MAXPLANES) {
	g_set_error(error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Too many planes (max is 16)"));
	return FALSE;
    }

    if (HAS_COLORMAP(cmap)) {
        if (context->view_port_modes & vmEXTRA_HALFBRITE) {
            ehbcmap(cmap);  	/* Modifies cmap */
	}
    }

    if (bmhd->compression != cmpNone && bmhd->compression != cmpByteRun1) {
	g_set_error(error,
    		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Unknown compression, supported are none and byte-run-1"));
	return FALSE;
    }

    ilbm_row = g_new(guchar, RowBytes(cols));
    raw_row = g_new0(rawtype, cols);
    mask_row = g_new0(guchar, cols);

    for (row = 0; row < rows; row++) {
	guchar *p = pixels + row * rowstride;

        ok = decode_row(&b, &size, raw_row, ilbm_row, bmhd->nPlanes, bmhd);
	if (ok) {
	    ok = decode_mask(&b, &size, raw_row, mask_row, ilbm_row, bmhd);
	}
	if (!ok) {
	    break;
	}

        for (col = 0; col < cols; col++, p += n_channels) {
            pixel c;
            if (context->has_transp_color && mask_row && mask_row[col] == 0) {
		c = context->transp_color;
	    } else {
                get_color(cmap, raw_row[col], &c);
            }
	    p[0] = c.r;
    	    p[1] = c.g;
	    p[2] = c.b;
	    if (has_alpha) {
		p[3] = mask_row[col];
	    }
        }
    }

    g_free(mask_row);
    g_free(raw_row);
    g_free(ilbm_row);

    if (!ok) {
	g_set_error(error,
     		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("Byte count mismatch or unsupported masking"));
	return FALSE;
    }

    return TRUE;
}

static gboolean read_deep(struct ilbm_progressive_state *context, 
	GError **error)
{
    return TRUE;
}

static gboolean read_ipbm(struct ilbm_progressive_state *context, 
	GError **error)
{
    return TRUE;
}

static gboolean read_rgbn(struct ilbm_progressive_state *context, 
	GError **error)
{
    return TRUE;
}

static gboolean process_body(struct ilbm_progressive_state *context, 
	GError **error)
{
    BitMapHeader *bmhd = &context->bmhd;
    ColorMap *cmap = &context->cmap;
    gboolean ok;

    if (bmhd->nPlanes == 0) {
	g_set_error(error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
		_("BODY without BMHD"));
	return FALSE;
    }

    check_cmap(context);

    if (context->iff_id == ID_ILBM) {
	if (context->has_dcol) {
	    ok = read_dcol(context, error);
	} else if (bmhd->nPlanes > 8) {
            if (bmhd->nPlanes <= 16 && HAS_COLORMAP(cmap)) {
		ok = read_std(context, error);
	    } else if (bmhd->nPlanes % 3 == 0) {
                ok = read_deep(context, error);
	    } else if (bmhd->nPlanes <= 16 ) {
                /* will be interpreted as grayscale */
                ok = read_std(context, error);
	    } else {
		g_set_error(error,
			GDK_PIXBUF_ERROR,
			GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
			_("Don't know how to interpret image"));
		ok = FALSE;
	    }
        } else {
            ok = read_std(context, error);
	}
    } else if (context->iff_id == ID_PBM) {
        ok = read_ipbm(context, error);
    } else {  /* RGBN or RGB8 */
        ok = read_rgbn(context, error);
    }

    if (context->updated_func) {
	context->updated_func(context->pixbuf, 0, 0, bmhd->w, bmhd->h,
		context->user_data);
    }
    context->body_chunk_processed = TRUE;

    return ok;
}

static gboolean ilbm_chunk_body(struct ilbm_progressive_state *context, 
	const guchar **buf, guint *size, GError **error)
{
    guchar *b = context->chunk_buf;

    if (!fill_buf(b, &context->chunk_len, context->chunk_size, buf, size)) {
	return TRUE;
    }

    if (b) {
	gboolean ok;

	switch (context->chunk_id) {
	    case ID_BMHD:
		ok = process_bmhd(context, error);
		break;

	    case ID_CMAP:
		ok = process_cmap(context, error);
		break;

	    case ID_CMYK:
		ok = process_cmyk(context, error);
		break;

	    case ID_CLUT:
		ok = process_clut(context, error);
		break;

	    case ID_CAMG:
		ok = process_camg(context, error);
		break;

	    case ID_DCOL:
		ok = process_dcol(context, error);
		break;

	    case ID_BODY: 
		ok = process_body(context, error);
		break;

	    default:
		DPRINTF(("strange id: %4.4s\n", (char *)&context->chunk_id)); 
		ok = FALSE;
		break;
	}

	if (!ok) {
	    return FALSE;
	}
    }

    context->chunk_len = 0;		/* clear for next fill */
    context->state = LS_CHUNK_HEADER;

    return TRUE;
}

/*
 * context - from image_begin_load
 * buf - new image data
 * size - length of new image data
 *
 * append image data onto inrecrementally built output image
 */
static gboolean gdk_pixbuf__ilbm_image_load_increment(gpointer data,
                                                      const guchar * buf,
                                                      guint size, 
						      GError **error)
{
    struct ilbm_progressive_state *context =
	(struct ilbm_progressive_state *) data;
    guint input_size = size;
    gboolean ok;

    if (context == NULL) {
	g_set_error (error,
		GDK_PIXBUF_ERROR,
		GDK_PIXBUF_ERROR_FAILED,
		_("Missing context"));
	return FALSE;
    }

    do {
	switch (context->state) {
	    case LS_START:
		ok = ilbm_start(context, &buf, &size, error);
		break;

	    case LS_TYPE:
		ok = ilbm_type(context, &buf, &size, error);
		break;

	    case LS_CHUNK_HEADER:
		ok = ilbm_chunk_header(context, &buf, &size, error);
		break;

	    case LS_CHUNK_BODY: 
		ok = ilbm_chunk_body(context, &buf, &size, error);
		break;

	    case LS_BODY: 
	    case LS_END:
		ok = TRUE; 		/* ignore trailing data */
		break;

	    default:
		g_set_error (error,
			GDK_PIXBUF_ERROR,
			GDK_PIXBUF_ERROR_FAILED,
			_("Illegal state"));
		ok = FALSE;
		break;
	}
    } while (ok && size > 0);

    context->bytes_to_go -= input_size - size;
    if (ok) {
	if (context->bytes_to_go <= 0) {
	    context->state = LS_END;
	}
    }

    return ok;
}

#if 0
	    else if(context->needmore)
	      {
		int first_row;
		first_row = context->cury;
		for( ; context->cury < context->height; context->cury++, context->curx = 0)
		  {
		    for( ; context->curx < context->width; context->curx += 8)
		      {
			guchar byte;
			guchar *ptr;
			int xoff;
			bv = getin(context, &buf, &size, &byte, 1);
			if(!bv)
			  goto out;

			ptr = context->pixbuf->pixels + context->pixbuf->rowstride * context->cury + context->curx * 3;
			for(xoff = 7; xoff >= 0; xoff--, ptr += 3)
			  {
			    guchar pixval;

			    if (context->curx + (7 - xoff) == context->width)
			      break;

			    if(byte & (1<<xoff))
			      pixval = 0xFF;
			    else
			      pixval = 0x0;

			    ptr[0] = ptr[1] = ptr[2] = pixval;
			  }
		      }
		  }
		context->needmore = FALSE;

	      out:
	    else
	      bv = FALSE; /* Nothing left to do, stop feeding me data! */

	  } while(bv);

	if(size) {
	  bv = save_rest(context, buf, size);
	  if (!bv) {
	      g_set_error (error,
			   GDK_PIXBUF_ERROR,
			   GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
			   _("Couldn't save the rest"));

	      return FALSE;
	  }
	}
	return TRUE;
#endif

void
MODULE_ENTRY (ilbm, fill_vtable) (GdkPixbufModule *module)
{
	module->begin_load = gdk_pixbuf__ilbm_image_begin_load;
	module->stop_load = gdk_pixbuf__ilbm_image_stop_load;
	module->load_increment = gdk_pixbuf__ilbm_image_load_increment;
}

void
MODULE_ENTRY (ilbm, fill_info) (GdkPixbufFormat *info)
{
	static GdkPixbufModulePattern signature[] = {
		{ "FORM    ILBM", "    xxxx    ", 100 }, 
		{ "FORM    RGBN", "    xxxx    ", 100 },
		{ "FORM    RGB8", "    xxxx    ", 100 },
		{ "FORM    PBM ", "    xxxx    ", 100 },
		{ NULL, NULL, 0 }
	};
	static gchar * mime_types[] = {
		"image/iff",
		NULL
	};
	static gchar * extensions[] = {
		"iff",
		NULL
	};

	info->name = "ilbm";
	info->signature = signature;
	info->description = N_("The IFF/ILBM image format");
	info->mime_types = mime_types;
	info->extensions = extensions;
	info->flags = 0;
}

