//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
//  gtkAmiFileChooser
//
//  gtk interface for choosing files and directories with pathnames
//  compatible to Amiga OS.
//
//  Creation date: 03.04.2004
//  Last modified: 25.04.2008
// 

#include "gtk/gtkAmiFileChooser.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "dosutil.h"
#include "util.h"
#include "appstrings_cat.h"

GtkWidget *gtk_ami_file_chooser_new(const gchar *title, const gchar *okText,
	const gchar *path, GtkFileChooserAction action, GtkWindow *parent)
{
    GtkWidget *dialog = gtk_file_chooser_dialog_new(title, parent, action,
            NULL, NULL);
    const char *env = getenv("BeeBase");
    if (env) {
        gchar *path = 0;
        const char *folder = 0;
        if (!g_path_is_absolute(env)) {
            gchar *cwd = g_get_current_dir();
            if (cwd) {
                path = g_build_filename(cwd, env, NULL);
                folder = path;
                g_free(cwd);
            }
        } else {
            folder = env;
        }
        if (folder && FileStat(folder) == 0) {
            gtk_file_chooser_add_shortcut_folder(GTK_FILE_CHOOSER(dialog), 
                    folder, 0);
        }
        g_free(path);
    }
    GtkWidget *cancel = GuiSimpleButton(GCS(MSG_Cancel));
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(cancel);

    if (!okText) {
	okText = GCS(MSG_Ok);
    }
    GtkWidget *ok = GuiSimpleButton(okText, true);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(ok);
    gtk_window_set_default(GTK_WINDOW(dialog), ok);

    gtk_ami_file_chooser_set_filename(GTK_FILE_CHOOSER(dialog), path);

    g_object_set(G_OBJECT(dialog), 
	    "window-position", parent? 
	    	GTK_WIN_POS_CENTER_ON_PARENT : GTK_WIN_POS_CENTER,
	    NULL);
    GuiAddWindowCallbacks(dialog, cancel);
    GuiSetDialogButtonOrder(dialog);

    return dialog;
}

void gtk_ami_file_chooser_set_filename(GtkFileChooser *chooser, 
	const gchar *path)
{
    g_return_if_fail (GTK_IS_FILE_CHOOSER (chooser));

    GtkFileChooserAction action = gtk_file_chooser_get_action(chooser);
    bool dirMode = (action == GTK_FILE_CHOOSER_ACTION_CREATE_FOLDER ||
	action == GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
    gchar *assignName = 0, *assignPath = 0;
    gchar *fullPath = gtkExpandPathnameGetAssign(EmptyNullStr(path),
	    &assignName, &assignPath);
    GObject *object = G_OBJECT(chooser);
    g_object_set_data_full(object, "assign-name", assignName, g_free);
    g_object_set_data_full(object, "assign-path", assignPath, g_free);
    short type = FileStat(fullPath);
    if (type > 0) {
	gtk_file_chooser_set_filename(chooser, fullPath);
    } else if (type < 0) {
	gchar *dirname = g_path_get_dirname(fullPath);
	gchar *basename= g_path_get_basename(fullPath);
	if (FileStat(dirname) == 0) {
	    gtk_file_chooser_set_current_folder(chooser, dirname);
	}
	if (action == GTK_FILE_CHOOSER_ACTION_CREATE_FOLDER ||
		action == GTK_FILE_CHOOSER_ACTION_SAVE) {
	    gtk_file_chooser_set_current_name(chooser, basename);
	}
	g_free(basename);
	g_free(dirname);
    } else if (dirMode && path != 0) {
	gtk_file_chooser_set_filename(chooser, fullPath);
    } else {
	gtk_file_chooser_set_current_folder(chooser, fullPath);
    }
}

gchar *gtk_ami_file_chooser_get_filename(GtkFileChooser *chooser)
{
    g_return_val_if_fail (GTK_IS_FILE_CHOOSER (chooser), NULL);

    GObject *object = G_OBJECT(chooser);
    gchar *fullPath = gtk_file_chooser_get_filename(chooser);
    gchar *assignName = (gchar *)g_object_get_data(object, "assign-name");
    gchar *assignPath = (gchar *)g_object_get_data(object, "assign-path");
    gchar *path = gtkUnexpandPathname(fullPath, assignName, assignPath);
    g_free(fullPath);

    return path;
}

gchar *gtkAmiFileChoose(const gchar *title, const gchar *ok, 
	const gchar *path, GtkFileChooserAction action, GtkWindow *parent)
{
    GtkWidget *dialog = gtk_ami_file_chooser_new(title, ok, path, 
            action, parent);

    gchar *result = 0;
    if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
	result = gtk_ami_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
    }

    gtk_widget_destroy(dialog);

    return result;
}

