//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkAnsiPrint.cc
//
// Creation date: 11.12.2011
// Last modified: 04.01.2012
//

#include "system.h"

#ifdef USE_ANSI_PRINT

#include "print.h"
#include "queryeditor.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "gtk/classes/gtksfwview.h"
#include "gtk/classes/gtkradiobox.h"
#include "gtk/classes/gtkpopselection.h"
#include "requesters.h"
#include "dosutil.h"
#include "util.h"
#include "memory.h"
#include "appstrings_cat.h"


#define OUTPUT_LEN 128
#define INIT_LEN 128
#define DELIMITER_LEN 16

struct PrintSettings
{
    char init[INIT_LEN + 1];
    short delimiter;
    char customdelimiter[DELIMITER_LEN + 1];
    short ident;
    bool head, esc, quotes, nlq, ff;
    short size, numlf;
    char output[OUTPUT_LEN + 1];
};

enum
{
    DELIMITER_SPACES, DELIMITER_TABS, DELIMITER_CUSTOM,
};

static struct PrintSettings printerSettings =
{
    "\\33c", DELIMITER_SPACES, "|", 0, 
    TRUE, TRUE, FALSE, TRUE, TRUE,
    1, 0, "",
};

static const char *font[]= { ANSI_NLQON, ANSI_NLQOFF };
static const char *pitch[] =
{
    ANSI_CONDENSEDOFF ANSI_ELITEOFF,
    ANSI_CONDENSEDOFF ANSI_ELITEON,
    ANSI_ELITEOFF ANSI_CONDENSEDON,
};

static void printEntry(FILE *file, const QueryEntry * const *qes, 
		       ushort *widths, char *formats,
		       ushort numFields, PrintSettings *ps)
{
    ushort i;
    const char *pattern = ps->quotes? "\"%s\"" : "%s";

    fprintf(file, "%*s", ps->ident, "");
    for(i = 0; i < numFields && qes[i] != NULL && !ferror(file); i++)
    {
        if(ps->delimiter != DELIMITER_SPACES)
            fprintf(file, pattern, qes[i]->qe_String);
        else if(formats[i] == 0)
        {
            fprintf(file, pattern, qes[i]->qe_String);
            fprintf(file, "%*s", (int)(strlen(qes[i]->qe_String)-widths[i]),"");
        }
        else
        {
            fprintf(file, "%*s", (int)(strlen(qes[i]->qe_String)-widths[i]),"");
            fprintf(file, pattern, qes[i]->qe_String);
        }
        if(i < numFields - 1)
	{
	    switch(ps->delimiter)
	    {
	    case DELIMITER_SPACES:
		fputc(' ', file);
		break;

	    case DELIMITER_TABS:
		fputc('\t', file);
		break;

	    case DELIMITER_CUSTOM:
		fputs(ps->customdelimiter, file);
		break;
	    }
	}
    }
    fputc('\n', file);
}

static void printEntries(Project *pr, GuiObject sfw, 
			 FILE *file, PrintSettings *ps)
{
    ushort numFields = GTK_SFW_VIEW(sfw)->numFields;
    const struct QueryLine *ql = &GTK_SFW_VIEW(sfw)->titleLine;
    char *formats = GTK_SFW_VIEW(sfw)->formats;
    ushort *widths = GTK_SFW_VIEW(sfw)->widths;

    /* initialize printer */
    if (ps->esc) {
        MemPool *memPool = ProjectMemPool(pr);
        char *init = StrCompress(memPool, ps->init);

        fprintf(file, "%s%s%s", init? init : ANSI_RESET, 
		font[ps->nlq? 0 : 1], pitch[ps->size]);
        if (init) {
            FreeString(memPool, init);
	}
    }

    /* print header */
    if (ps->head) {
        printEntry(file, ql->ql_Entries, widths, formats, numFields, ps);

        if (ps->delimiter == DELIMITER_SPACES) {
            fprintf(file, "%*s", ps->ident, "");
            for(int j = 0; j < numFields; j++) {
                for(int i=0; i < widths[j]; i++) {
                    fputc('-', file);
		}
                if (j < numFields - 1) {
                    fputc('-', file);
		}
            }
            fputc('\n', file);
        }
    }

    /* print lines */
    GtkTreeIter iter;
    if ((ql = gtk_sfw_view_get_iter_first(GTK_SFW_VIEW(sfw), &iter))) {
	do {
	    printEntry(file, ql->ql_Entries, widths, formats, numFields, ps);
        } while (!ferror(file) &&
                (ql = gtk_sfw_view_iter_next(GTK_SFW_VIEW(sfw), &iter)));
    }

    /* after printing */
    for (int i=0; i < ps->numlf; i++) {
        fputc('\n', file);
    }
    if (ps->ff) {
        fputc('\f', file);
    }
}

struct PrintSFWData
{
    Project *d_Project;
    GuiObject d_SFWView;
    GuiObject d_Delimiter, d_CustomDelimiter, d_Font, d_Init, d_Indent;
    GuiObject d_Headline, d_EscapeCodes, d_Quotes;
    GuiObject d_Size, d_AfterPrinting, d_NumLineFeeds, d_Output;
    GuiObject d_Ok, d_Cancel;
};

static void closePrintSFW(GuiObject go, gint response, PrintSFWData *d)
{
    Project *pr = d->d_Project;
    MemPool *memPool = ProjectMemPool(pr);

    if (response == GTK_RESPONSE_ACCEPT) {
	GuiBusyCursor busy;

        PrintSettings ps;
	ps.delimiter = gtk_radio_box_get_active(GTK_RADIO_BOX(d->d_Delimiter));
	sprintf(ps.customdelimiter, "%.*s", DELIMITER_LEN, 
		gtk_entry_get_text(GTK_ENTRY(d->d_CustomDelimiter)));
        ps.nlq = (gtk_radio_box_get_active(GTK_RADIO_BOX(d->d_Font)) == 0);
        sprintf(ps.init, "%.*s", INIT_LEN, 
		gtk_entry_get_text(GTK_ENTRY(d->d_Init)));
        ps.ident = (int)gtk_range_get_value(GTK_RANGE(d->d_Indent));
        ps.head =gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(d->d_Headline));
        ps.esc = gtk_toggle_button_get_active(
		GTK_TOGGLE_BUTTON(d->d_EscapeCodes));
        ps.quotes =gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(d->d_Quotes));
        ps.size = gtk_radio_box_get_active(GTK_RADIO_BOX(d->d_Size));
        switch (gtk_radio_box_get_active(GTK_RADIO_BOX(d->d_AfterPrinting))) {
	    case 0:     ps.ff = TRUE; ps.numlf = 0; break;
	    case 1:     ps.ff = FALSE;
			ps.numlf = gtk_spin_button_get_value_as_int(
				GTK_SPIN_BUTTON(d->d_NumLineFeeds));
			break;
	    default:    ps.ff = FALSE; ps.numlf = 0;
			break;
        }
        sprintf(ps.output, "%.*s", OUTPUT_LEN,
		gtk_entry_get_text(GTK_ENTRY(d->d_Output)));

        ProjectSetCurrentDir(pr);       /* make sure we have correct CWD */
 	if (FileStat(ps.output) >= 0) {
	    if (Ask(AT_WARN, GCS(MSG_AskReplaceAnswers),
		   GCS(MSG_AskReplaceFile), ps.output) == 0) {
		return;			/* user canceled */
	    }
	}

        bool ok = FALSE;
        FILE *file = IFOpen(memPool, ps.output, "w");
        if (!file) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_OpenFile), ps.output);
	    return;
	}

	printEntries(pr, d->d_SFWView, file, &ps);
        if(!ferror(file)) {
	    fflush(file);
	}
	if(!ferror(file)) {
	    printerSettings = ps;
	    ok = TRUE;
	} else {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_WriteFile), ps.output);
	}
 	fclose(file);		/* fclose == 0 seems not to be enough */

        if (!ok) {
            return;
	}
    }

    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

static GuiObject createPrintSFW(Project *pr, GuiObject sfw)
{
    PrintSFWData *d = g_new0(PrintSFWData, 1);
    d->d_Project = pr;
    d->d_SFWView = sfw;

    PrintSettings *ps = &printerSettings;

    static char *delimiterContent[4];
    delimiterContent[0] = GCS(MSG_QueryPrint_Spaces);
    delimiterContent[1] = GCS(MSG_QueryPrint_Tabs);
    delimiterContent[2] = GCS(MSG_QueryPrint_Custom);
    delimiterContent[3] = NULL;
    d->d_Delimiter = RadioBoxObject,
        "entries", delimiterContent,
	// MUIA_ControlChar, Shortcut(GCS(MSG_QueryPrint_Delimiter)),
	End;

    d->d_CustomDelimiter = EntryObject,
	"max-length", DELIMITER_LEN,
	"width-chars", DELIMITER_LEN,
	"xalign", 0.5,
	End;
    GuiSetTooltip(d->d_CustomDelimiter, GCS(MSG_QueryPrint_DelimiterBH));

    GuiObject delimiter = Frame(GCS(MSG_QueryPrint_Delimiter)),
	"child", GroupAlignment(true),
	    "child", HGroup,
	        "child", SpaceObject, End,
		"child", d->d_Delimiter,
		"child", VGroup,
	            "child", SpaceObject, End,
	            "child", d->d_CustomDelimiter, "vert-weight", 0,
		    End,
		End,
	    End,
	End;

    static char *fontContent[3];
    fontContent[0] = GCS(MSG_QueryPrint_NLQ);
    fontContent[1] = GCS(MSG_QueryPrint_Draft);
    fontContent[2] = NULL;
    d->d_Font = RadioBoxObject,
	"entries", fontContent,
	//MUIA_ControlChar, Shortcut(GCS(MSG_Font)),
	End;
    GuiObject fontBox = GuiMakeTooltip(d->d_Font, GCS(MSG_QueryPrint_FontBH));

    GuiObject font = Frame(GCS(MSG_Font)),
        "child", GroupAlignment(true),
	    "child", VCenter(HCenter(fontBox)),
	    End,
	End;

    d->d_Init = EntryObject,
        "max-length", 128,
	End;
    GuiSetTooltip(d->d_Init, GCS(MSG_QueryPrint_InitSequenceBH));

    d->d_Indent = HScaleObject,
        "adjustment", gtk_adjustment_new(0, 0, 20, 1, 1, 0),
	"value-pos", GTK_POS_RIGHT,
	"digits", 0,
	End;
    GuiObject indentBox = GuiMakeTooltip(d->d_Indent, 
            GCS(MSG_QueryPrint_IndentBH));

    d->d_Headline = CheckButtonObject, End;
    GuiSetTooltip(d->d_Headline, GCS(MSG_QueryPrint_HeadlineBH));

    d->d_EscapeCodes = CheckButtonObject, End;
    GuiSetTooltip(d->d_EscapeCodes, GCS(MSG_QueryPrint_EscapeCodesBH));

    d->d_Quotes = CheckButtonObject, End;
    GuiSetTooltip(d->d_Quotes, GCS(MSG_QueryPrint_QuotesBH));

    GuiObject misc = Frame(GCS(MSG_QueryPrint_Misc)),
        "child", GroupAlignment(true),
	    "child", VGroup,
	        "columns", 2,
		"child", SpaceObject, End, "weight", 0,
		"child", SpaceObject, End,
                ChildLabelWidget(GCS(MSG_QueryPrint_InitSequence), d->d_Init),
		"child", GuiLabel(GCS(MSG_QueryPrint_Indent), d->d_Indent),
                "weight", 0,
                "child", indentBox,
		ChildLabelWidget(GCS(MSG_QueryPrint_Headline), d->d_Headline),
		ChildLabelWidget(GCS(MSG_QueryPrint_EscapeCodes), 
                        d->d_EscapeCodes),
		ChildLabelWidget(GCS(MSG_QueryPrint_Quotes), d->d_Quotes),
		"child", SpaceObject, End, "weight", 0,
		"child", SpaceObject, End,
		End,
	    End,
	End;

    static char *sizeContent[4];
    sizeContent[0] = GCS(MSG_QueryPrint_Pica);
    sizeContent[1] = GCS(MSG_QueryPrint_Elite);
    sizeContent[2] = GCS(MSG_QueryPrint_Condensed);
    sizeContent[3] = NULL;
    d->d_Size = RadioBoxObject,
	"entries", sizeContent,
	// MUIA_ControlChar, Shortcut(GCS(MSG_QueryPrint_Size)),
	End;
    GuiObject sizeBox = GuiMakeTooltip(d->d_Size, GCS(MSG_QueryPrint_SizeBH));

    GuiObject size = Frame(GCS(MSG_Size)),
        "child", GroupAlignment(true),
	    "child", HCenter(sizeBox),
	    End,
	End;

    static char *afterPrintingContent[4];
    afterPrintingContent[0] = GCS(MSG_QueryPrint_FormFeed);
    afterPrintingContent[1] = GCS(MSG_QueryPrint_LineFeeds);
    afterPrintingContent[2] = GCS(MSG_QueryPrint_Nothing);
    afterPrintingContent[3] = NULL;
    d->d_AfterPrinting = RadioBoxObject,
	"entries", afterPrintingContent,
	// MUIA_ControlChar, Shortcut(GCS(MSG_QueryPrint_AfterPrinting)),
	End;
    GuiObject afterPrintingBox = GuiMakeTooltip(d->d_AfterPrinting, 
            GCS(MSG_QueryPrint_AfterPrintingBH));

    d->d_NumLineFeeds = SpinButtonObject,
	// MUIA_ControlChar, Shortcut(GCS(MSG_QueryPrint_NumLineFeeds)),
        "adjustment", gtk_adjustment_new(0, 0, 1000, 1, 10, 0),
	"numeric", true,
	"digits", 0,
	"max-length", 5,
	"xalign", 0.5,
	End;
    GuiSetTooltip(d->d_NumLineFeeds, GCS(MSG_QueryPrint_NumLineFeedsBH));

    GuiObject afterPrinting = Frame(GCS(MSG_QueryPrint_AfterPrinting)),
        "child", GroupAlignment(true),
	    "child", HGroup,
	        "child", SpaceObject, End,
		"child", afterPrintingBox,
		"child", HGroup,
		    ChildLabelWidget(GCS(MSG_QueryPrint_NumLineFeeds),
                            d->d_NumLineFeeds),
		    End,
	        "child", SpaceObject, End,
		End,
            End,
	End;

    GuiObject top = VGroup,
	"columns", 2,
	"child", VGroup,
	    "child", delimiter,
	    "child", font,
	    End, "weight", 0,
	"child", misc,
	"child", size, "weight", 0,
	"child", afterPrinting,
	End;

    d->d_Output = EntryObject,
	"max-length", 80,
	"xalign", 0.5,
	End;
    GuiSetTooltip(d->d_Output, GCS(MSG_QueryPrint_OutputBH));

    GuiObject popup = PopSelectionObject,
        "selection-type", GTK_TYPE_FILE_CHOOSER_DIALOG, 
        "file-chooser-action", GTK_FILE_CHOOSER_ACTION_SAVE,
	"string", d->d_Output,
        "button", GuiPopButtonSetCWD(SELECT_FILE_ICON_NAME, pr), 
	End;

    GuiObject group = VGroup,
	"child", top, "vert-weight", 0,
        "child", HGroup,
	    "child", LabelObject,
		"label", GCS(MSG_QueryPrint_Output),
		"use-underline", true,
		"mnemonic-widget", d->d_Output,
		End, "weight", 0,
	    "child", popup,
	    End, "vert-weight", 0,
	End;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_Print),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(sfw),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Printing queries");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = GTK_DIALOG(dialog)->vbox;
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(closePrintSFW), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_Delimiter), ps->delimiter);
    gtk_entry_set_text(GTK_ENTRY(d->d_CustomDelimiter), ps->customdelimiter);
    gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_Font), ps->nlq? 0 : 1);
    gtk_entry_set_text(GTK_ENTRY(d->d_Init), EmptyNullStr(ps->init));
    gtk_range_set_value(GTK_RANGE(d->d_Indent), ps->ident);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_Headline), ps->head);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_EscapeCodes),ps->esc);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_Quotes), ps->quotes);
    gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_Size), ps->size);
    gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_AfterPrinting), 
	ps->ff? 0 : ((ps->numlf > 0)? 1 : 2));
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(d->d_NumLineFeeds), ps->numlf);
    gtk_entry_set_text(GTK_ENTRY(d->d_Output), EmptyNullStr(ps->output));

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }

    return dialog;
}

void GuiPrintSFWList(struct Project *pr, GuiObject sfw)
{
    createPrintSFW(pr, sfw);
}

#endif

