//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkBreak.cc
//
// Creation date: 30.12.2004
// Last modified: 25.10.2020
//

#include "break.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "gtk/classes/gtkgroup.h"
#include "util.h"
#include "appstrings_cat.h"

#define BREAK_SHORT_DELAY 1000
#define BREAK_LONG_DELAY 2000

#define BREAK_TITLE_LEN 200

#ifndef timercmp
#define timercmp(t1, t2, op) \
	(((t1)->tv_sec op (t2)->tv_sec) || \
     	(((t1)->tv_sec == (t2)->tv_sec) && ((t1)->tv_usec op (t2)->tv_usec)))
#endif

static ulong breakCount=0;
static bool breakGui = false;

static struct {
    bool present;
    bool progressVisible;
    bool abort;
    bool pause;
    short percentage;
    GuiObject win;
    GuiObject progressBar;
    GuiObject pauseButton;
    GuiObject abortButton;
    GTimeVal nextIteration;
} breakData = { 0 };

static struct BreakSettings breakSettingsInit, breakSettingsCurrent;
static char breakTitleBuffer[BREAK_TITLE_LEN + 1];

static void breakUpdateProgress(void)
{
    GuiObject pb = breakData.progressBar;

    if (pb) {
        const BreakSettings &bs = breakSettingsCurrent;
        bool visible = (bs.steps != 0 && bs.current != 0);

        if (breakData.progressVisible != visible) {
            if (visible) {
                gtk_widget_show(pb);
            } else {
                gtk_widget_hide(pb);
            }
            breakData.progressVisible = visible;
	}
        if (visible) {
            short percentage = (short)(bs.current * 100 / bs.steps);
            if (percentage != breakData.percentage) {
                gdouble frac = percentage / 100.0;
                if (frac < 0.0) {
                    frac = 0.0;
                } else if (frac > 1.0) {
                    frac = 1.0;
                }
                gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(pb), frac);

                gchar text[32];
                sprintf(text, "%hd %%", percentage);
                gtk_progress_bar_set_text(GTK_PROGRESS_BAR(pb), text);

                breakData.percentage = percentage;
            }
        }
    }
}

static void breakDoSettings(void)
{
    const BreakSettings &bs = breakSettingsCurrent;
        
#if 0
    if (breakData.pauseButton) {
        gtk_widget_set_sensitive(breakData.pauseButton, bs.pause);
    }
    if (breakData.abortButton) {
        gtk_widget_set_sensitive(breakData.abortButton, bs.abort);
    }
#endif
    if (breakData.win) {
        gtk_window_set_title(GTK_WINDOW(breakData.win), bs.title);
    }
    breakUpdateProgress();
}

void BreakInit(struct BreakSettings *bs)
{
    if (breakCount == 0) {
        /* 
        ** We do not GuiSetSleep here as it interfers with browsing
        ** records using the arrow buttons of a panel in a table that
        ** has virtual fields causing program execution.
        */  
	breakGui = (GuiAnyOpenedWindow() != 0);
        memset(&breakData, 0, sizeof(breakData));
        if (bs) {
            breakSettingsInit = *bs;
            breakGui |= bs->needGui;
        } else {
            memset(&breakSettingsInit, 0, sizeof(breakSettingsInit));
        }
        breakSettingsCurrent = breakSettingsInit;
	if (breakGui) {
	    g_get_current_time(&breakData.nextIteration);
	    g_time_val_add(&breakData.nextIteration, BREAK_LONG_DELAY * 1000);
	} else {
	    breakData.nextIteration.tv_sec = LONG_MAX;
	    breakData.nextIteration.tv_usec = 0;
	}

    } else if (bs) {
        BreakSettings old = breakSettingsCurrent;
        breakSettingsCurrent = *bs;
        *bs = old;
	if (old.title == breakTitleBuffer) {
	    breakSettingsCurrent.title = breakTitleBuffer;   /* keep message */
	}
        breakDoSettings();
    }

    breakCount++;
}

void BreakExit(const struct BreakSettings *bs)
{
    breakCount--;
    if (breakCount == 0) {
        if (breakData.win) {
	    gtk_widget_destroy(breakData.win);
	    GuiSetSleep(FALSE);
	}
        memset(&breakData, 0, sizeof(breakData));
        breakGui = false;
    } else if (bs) {
        breakSettingsCurrent = *bs;
        breakDoSettings();
    }
}

void BreakHide(void)
{
    GuiCloseWindow(breakData.win);
    breakData.present = false;
    breakData.nextIteration.tv_sec = LONG_MAX;
    breakData.nextIteration.tv_usec = 0;
}

void BreakReInit(void)
{
    if (breakGui && !breakData.present) {
	g_get_current_time(&breakData.nextIteration);
	g_time_val_add(&breakData.nextIteration, BREAK_SHORT_DELAY * 1000);
    }
}

static void breakPause()
{
    breakData.pause = !breakData.pause;
    if (breakData.pause) {
        gtk_button_set_label(GTK_BUTTON(breakData.pauseButton),
                GCS(MSG_Continue));
        while (breakData.pause && !breakData.abort) {
            gtk_main_iteration();
        }
    } else {
        gtk_button_set_label(GTK_BUTTON(breakData.pauseButton), GCS(MSG_Pause));
    }
}

static void breakAbort()
{
    breakData.abort = true;
}

static void breakCreateGui()
{
    // 
    // Make sure BeeBase is the active application.
    // We don't want to popup the break window 
    // in front of another active application 
    // except when BeeBase does not have any open window.
    //
    if (!GuiActiveWindow() && GuiAnyOpenedWindow()) {   
        return;
    }

    if (!breakData.win) {
        breakData.progressBar = ProgressBarObject, 
#if GTK_MAJOR_VERSION == 2
            "orientation", GTK_PROGRESS_LEFT_TO_RIGHT,
#else
            "orientation", GTK_ORIENTATION_HORIZONTAL,
#endif
            End;
        GuiObject vgroup = VGroup, 
            "child", breakData.progressBar, "weight", 0,
            End;
        
        const struct BreakSettings &bs = breakSettingsInit;
        if (bs.abort || bs.pause) {
            GuiObject buttons = HGroup, End;
            gtk_group_add_child(GTK_GROUP(vgroup), buttons, 1, 1);
            
            if (bs.abort) {
                breakData.abortButton = GuiSimpleButton(GCS(MSG_Abort));
                g_signal_connect(breakData.abortButton, "clicked", 
                        G_CALLBACK(breakAbort), 0);
                gtk_group_add_child(GTK_GROUP(buttons), 
                        breakData.abortButton, 1, 1);
            }
            if (bs.pause) {
                breakData.pauseButton = GuiSimpleButton(GCS(MSG_Pause));
                g_signal_connect(breakData.pauseButton, "clicked", 
                        G_CALLBACK(breakPause), 0);
                gtk_group_add_child(GTK_GROUP(buttons), 
                        breakData.pauseButton, 1,1);
            }
        }
        
        GuiObject root = GroupAlignment(false),
            "child", vgroup,
            End;
        
        breakData.win = WindowObject,
            "default-width", 360,
            "default-height", 120,
            "title", bs.title,
            "modal", true,
            // "accept-focus", false,
            "skip-pager-hint", true,
            "skip-taskbar-hint", true,
            "type-hint", GDK_WINDOW_TYPE_HINT_DIALOG,
            "window-position", GTK_WIN_POS_CENTER,
            "transient-for", GuiFindTransientWindow(NULL),
            "child", root,
            End;
	GuiIgnoreWindowDelete(breakData.win);
        GuiSetSleep(TRUE);
        gtk_widget_show_all(root);
    }
        
    breakDoSettings();
    gtk_window_present(GTK_WINDOW(breakData.win));
    breakData.present = true;
}
    
bool BreakCheck(void)
{
    if (breakGui) {
        GTimeVal current;
	g_get_current_time(&current);
	if (timercmp(&current, &breakData.nextIteration, >)) {
	    if (!breakData.present) {
		breakCreateGui();
	    }
            while (gtk_events_pending()) {
                gtk_main_iteration();
            }
            g_time_val_add(&current, 10000);
            breakData.nextIteration = current;
	}
    }

    return breakData.abort;
}

void BreakMessage(const char *str)
{
    CodesetStrNBytesCopy(breakTitleBuffer, str, BREAK_TITLE_LEN);
    char *s = CodesetStrchr(breakTitleBuffer, '\n');
    if (s) {
        *s = 0;
    }

    BreakSettings &bs = breakSettingsCurrent;
    bs.title = breakTitleBuffer;
    if (breakData.win) {
	gtk_window_set_title(GTK_WINDOW(breakData.win), bs.title);
    }
}

void BreakSetMaxSteps(ulong steps)
{
    breakSettingsCurrent.steps = steps;
    breakUpdateProgress();
}

ulong BreakCurrentSteps(void)
{
    return breakSettingsCurrent.current;
}

void BreakSetCurrentStep(ulong current)
{
    BreakCheck();
    breakSettingsCurrent.current = current;
    breakUpdateProgress();
}

void BreakAddCurrentStep(long add)
{
    BreakCheck();
    breakSettingsCurrent.current += add;
    breakUpdateProgress();
}

