//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkBuildGui.cc
//
// Creation date: 23.03.2004
// Last modified: 16.07.2020
//

#include "gtk/gtkBuildGui.h"
#include "gtk/gtkBuildGuiWindow.h"
#include "gtk/gtkBuildGuiTable.h"
#include "gtk/gtkBuildGuiField.h"
#include "gtk/classes/gtkpopimage.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "buildgui.h"
#include "gui.h"
#include "project.h"
#include "display.h"
#include "preferences.h"

gboolean OnExtraButtonFocus(GtkWidget *widget, GtkDirectionType type,
	DisplayObject *dio)
{
    if (!GetPrefsValue(DisplayObjectProject(dio),PREFS_EXTRABUTTONSINTABCHAIN)){
	g_signal_stop_emission_by_name(widget, "focus");
    }
    return FALSE;
}

DisplayObject *GuiObjectDisplayObject(GuiObject go) 
{ 
    DisplayObject *dio = 0;

    if (go) {
	dio = (DisplayObject *)g_object_get_data(G_OBJECT(go), "dio");
    }
    return dio;
}

static void FreeProjectGuiRec(DisplayObject *dio)
{
    List *childs = DisplayObjectObjects(dio);
    short type = DisplayObjectType(dio);
    GuiObject go = DisplayObjectGuiObject(dio);
    bool open;

    if (dio) {
        switch(type) {
            case DO_TYPE_WINDOW:
		if (go) {
                    open = GuiWindowOpen(go);   /* remember current state */
		    gtk_widget_hide(go);
		    gtk_widget_destroy(go);
                    SetDisplayWindowOpen(dio, open);
                    SetDisplayWindowButtonObject(dio, 0);
		}
    		break;

            case DO_TYPE_FIELD:
                SetDisplayFieldListviewObject(dio, 0);
                SetDisplayFieldFileObject(dio, 0);
                SetDisplayFieldFontObject(dio, 0);
                SetDisplayFieldAutoShowObject(dio, 0);
                SetDisplayFieldFilterObject(dio, 0);
                SetDisplayFieldViewObject(dio, 0);
                SetDisplayFieldPopObject(dio, 0);
                SetDisplayFieldDisableObject(dio, 0);
                break;

            case DO_TYPE_TABLE:
                SetDisplayTablePanelNumAllObject(dio, 0);
                SetDisplayTablePanelFilterObject(dio, 0);
                break;
	}

        SetDisplayObjectGuiObject(dio, 0);
        SetDisplayObjectRootObject(dio, 0);
        SetDisplayObjectActivateObject(dio, 0);

    	if (childs) {
	    FORNODE(dio, childs) {
	    	FreeProjectGuiRec(dio);
	    }
	}
    }
}

void FreeProjectGui(Project *pr)
{
    DisplayObject *dio = ProjectRootDisplayObject(pr);

    GuiSetSleep(true);
    FreeProjectGuiRec(dio);
    GuiSetSleep(false);
}

GtkWidget *BuildGuiIconLabelButton(const gchar *icon_name,
        const char *title, char shortcut, gfloat xalign)
{
    gboolean use;
    gchar *text = GuiUnderline(title, shortcut, use);
    GtkWidget *button, *group, *image = NULL, *label = NULL;

    button = ButtonObject,
        "child", HGroup,
	    "spacing", 8,
	    "child", SpaceObject, End, "weight", 0,
	    "child", group = HGroup, "spacing", 2, End,
	    "child", SpaceObject, End, "weight", 0,
	    End,
	//tf? MUIA_Font : TAG_IGNORE, tf,
	//MUIA_Background, (bg && bg[0])? bg : (char *)MUII_ButtonBack,
	End;

    if (icon_name &&
            gtk_icon_theme_has_icon(gtk_icon_theme_get_default(), icon_name)) {
	image = ImageObject,
	    "icon-name", icon_name,
	    "icon-size", GTK_ICON_SIZE_MENU,
	    End;
	gtk_group_add_child(GTK_GROUP(group), image, 0, 0);
	g_object_ref(G_OBJECT(image));
	g_object_set_data_full(G_OBJECT(button), 
		"image", image, g_object_unref);
    }
    if (title) {
	label = LabelObject,
	    "label", text,
	    "use-underline", use,
	    "xalign", xalign,
	    End;
	gtk_group_add_child(GTK_GROUP(group), label, 1, 0);
	g_object_ref(G_OBJECT(label));
	g_object_set_data_full(G_OBJECT(button), 
		"label", label, g_object_unref);
    }
    g_free(text);

    return button;
}

void BuildGuiAddCommon(DisplayObject *dio, GtkWidget *widget)
{
    if (widget) {
        g_object_set_data(G_OBJECT(widget), "dio", dio);
    }
}

void BuildGuiSetBackgroundSpec(struct DisplayObject *dio, 
        const char *spec, GtkWidget *widget)
{
    if (widget) {
        ProjectSetCurrentDir(DisplayObjectProject(dio));
        gtk_widget_set_bg_spec(widget, spec);
    }
}

void BuildGuiSetBackground(DisplayObject *dio, GtkWidget *widget)
{
    BuildGuiSetBackgroundSpec(dio, DisplayObjectBackground(dio), widget);
}

void BuildGuiSetFontname(const char *fontname, GtkWidget *widget)
{
    if (widget && fontname && fontname[0]) {
        PangoFontDescription *pfd= pango_font_description_from_string(fontname);
        if (pfd) {
#if GTK_MAJOR_VERSION == 2
            gtk_widget_modify_font(widget, pfd);
#else
            gtk_widget_override_font(widget, pfd);
#endif
#if 0
            GtkStyle *style = gtk_widget_get_style(widget);
            if (style) {
                style = gtk_style_copy(style);
                if (style->font_desc) {
                    pango_font_description_free(style->font_desc);
                }
                style->font_desc = pango_font_description_copy(pfd);
                gtk_widget_set_style(widget, style);
            }
#endif
            pango_font_description_free(pfd);
        }
    }
}

void BuildGuiSetFont(DisplayObject *dio, GtkWidget *widget)
{
    BuildGuiSetFontname(DisplayObjectGTKFontname(dio), widget);
}

static bool buildGuiText(DisplayObject *dio, GtkGroup *parent)
{
    if (!parent) {
	return false;
    }

    const char *title = DisplayObjectTitle(dio);
    int weight = DisplayObjectWeight(dio);
    GtkWidget *label = LabelObject, "label", title, End;
    gtk_group_add_child(parent, label, weight, 0);
    BuildGuiAddCommon(dio, label);
    BuildGuiSetBackground(dio, label);
    BuildGuiSetFont(dio, label);
    SetDisplayObjectGuiObject(dio, label);
    SetDisplayObjectRootObject(dio, label);

    return true;
}

static bool buildGuiImage(DisplayObject *dio, GtkGroup *parent)
{
    if (!parent) {
	return false;
    }

    Project *pr = DisplayObjectProject(dio);
    const char *bg = DisplayObjectBackground(dio);
    char imageSize = DisplayImageImageSize(dio);
    bool resize = (imageSize == DAD_IMAGESIZE_RESIZABLE);
    int weight = DisplayObjectWeight(dio);
    GuiObject go = 0;

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    if (bg && bg[0] == '5' && bg[1] == ':') {
	gchar *path = gtkExpandPathname(&bg[2]);
	GError *error = NULL;
	GdkPixbufAnimation *anim = gdk_pixbuf_animation_new_from_file(
		path, &error);
	g_free(path);
	if (error) {
	    // printf("error loading image: %s\n", error->message);
	    g_error_free(error);
	} else {
	    go = ImageObject, "pixbuf-animation", anim, End;
	}
	if (anim) {
	    g_object_unref(G_OBJECT(anim));
	}
    } 
    if (!go) {
	go = SpaceObject, End;
        BuildGuiSetBackground(dio, go);
    }
    if (!resize) {
	weight = 0;
    }
    gtk_group_add_child(parent, go, weight, weight);
    BuildGuiAddCommon(dio, go);

    SetDisplayObjectGuiObject(dio, go);
    SetDisplayObjectRootObject(dio, go);

    return true;
}

static bool buildGuiSpace(DisplayObject *dio, GtkGroup *parent)
{
    if (!parent) {
	return false;
    }

    gboolean horiz = gtk_group_get_horiz(parent);
    bool bar = DisplaySpaceBar(dio);
    int weight = DisplayObjectWeight(dio);
    GuiObject go = 0;

    if (!bar) {
	go = SpaceObject, End;
    } else if (horiz) {
	go = VSeparatorObject, End;
    } else {
	go = HSeparatorObject, End;
    }

    gtk_group_add_child(parent, go, horiz? weight : 0, horiz? 0 : weight);

    BuildGuiAddCommon(dio, go);
    BuildGuiSetBackground(dio, go);

    SetDisplayObjectGuiObject(dio, go);
    SetDisplayObjectRootObject(dio, go);

    return true;
}

static bool BuildGuiObject(DisplayObject *dio, bool hideLabel,
	GtkGroup *group, GtkGroup **colGroup);      // forward declaration

static void onRegisterNotifyPage(GuiObject go, GParamSpec *arg1, Project *pr)
{
    GuiUpdateRegister(go, pr);
}

static bool buildGuiRegister(DisplayObject *dio, GtkGroup *parent)
{
    if (!parent) {
	return false;
    }

    GtkWidget *notebook = NotebookObject, 
        "enable-popup", true,
	End;
    BuildGuiAddCommon(dio, notebook);

    SetDisplayObjectGuiObject(dio, notebook);
    SetDisplayObjectRootObject(dio, notebook);
    g_signal_connect(notebook, "notify::page", 
	    G_CALLBACK(onRegisterNotifyPage), DisplayObjectProject(dio));

    int weight = DisplayObjectWeight(dio);
    int horizWeight = 0, vertWeight = 0;
    char **labels = DisplayRegisterLabels(dio);
    List *childs = DisplayObjectObjects(dio);

    if (childs) {    
	int i = 0;
	DisplayObject *child;
	FORNODE(child, childs) {
            if (!DisplayObjectVisible(child)) {
                continue;
            }
	    GtkWidget *widget = VGroup, End;
	    GtkGroup *group = GTK_GROUP(widget);
	    GtkWidget *label = 0;
	    if (labels && labels[i]) {
		label = LabelObject, "label", labels[i], End;
		i++;
	    }
	    widget = GroupAlignment(FALSE), "child", widget, End;
	    gtk_notebook_append_page_menu(GTK_NOTEBOOK(notebook), 
		    widget, label, 0);
	    GtkGroup *colGroup = 0;
	    if (!BuildGuiObject(child, false, group, &colGroup)) {
    		return false;
	    }
	    BuildGuiFinishColGroup(&colGroup);
	    if (gtk_group_get_total_vert_weight(group) > 0) {
		vertWeight = weight;
	    }
	    if (gtk_group_get_total_horiz_weight(group) > 0) {
		horizWeight = weight;
	    }
	}
    }
    gtk_group_add_child(parent, notebook, horizWeight, vertWeight);
    return true;
}

static bool buildGuiGroup(DisplayObject *dio, GtkGroup *parent, 
	GtkGroup **colGroup)
{
    if (!parent || !colGroup) {
	return false;
    }

    bool hideLabel = false;
    bool parentHoriz = gtk_group_get_horiz(parent);
    bool horiz = DisplayGroupHoriz(dio);
    bool border = DisplayGroupBorder(dio);
    const char *title = DisplayObjectTitle(dio);
    List *childs = DisplayObjectObjects(dio);
    if (title && title[0] == 0) {
	title = 0;
    }

    if (!parentHoriz && horiz && !border && !title && childs) {
	DisplayObject *child;
	FORNODE(child, childs) {
	    if (!DisplayObjectVisibleGui(child)) {
		continue;
	    } 
	    if (DisplayObjectType(child) != DO_TYPE_FIELD) {
		break;
	    } 
	    if (DisplayFieldTitlePosition(child)!=DAD_TITLEPOSITION_LEFT) {
		break;
	    }
	    if (!BuildGuiFieldHasLabel(child)) {
		break;
	    }

	    if (!*colGroup) {
		GtkWidget *w = VGroup, "columns", 2, End;
		*colGroup = GTK_GROUP(w);
		gtk_group_add_child(parent, w, 100, 100);
	    }
	    if (!BuildGuiFieldLabel(child, *colGroup)) {
	    	return false;
	    }
	    hideLabel = true;
	    parent = *colGroup;
	    break;
	}
    }
    if (!hideLabel) {
	BuildGuiFinishColGroup(colGroup);
    }

    int weight = DisplayObjectWeight(dio);
    bool spacing = DisplayGroupSpacing(dio);

    GtkWidget *widget = GroupObject, "horiz", horiz, End;
    GtkGroup *group = GTK_GROUP(widget);
    if (!spacing) {
	gtk_group_set_spacing(group, 0, 0);
    }

    if (title || border) {
	widget = Frame(title),
	    "shadow-type", border? GTK_SHADOW_ETCHED_IN : GTK_SHADOW_NONE,
	    "child", GroupAlignment(title),
		"child", widget,
		End,
	    End;
        // BuildGuiSetFont(dio, gtk_frame_get_label_widget(GTK_FRAME(widget)));
        // font cannot be specified in group editor
    }
    BuildGuiAddCommon(dio, widget);
    BuildGuiSetBackground(dio, widget);

    SetDisplayObjectGuiObject(dio, widget);
    SetDisplayObjectRootObject(dio, widget);

    GtkGroup *newColGroup = 0;
    bool rc = BuildGuiChilds(childs, hideLabel, group, &newColGroup);

    int horizWeight, vertWeight;
    if (parentHoriz) {
        horizWeight = weight;
        vertWeight = 100;
    } else {
        horizWeight = 100;
        vertWeight = weight;
    }
    if (gtk_group_get_total_horiz_weight(group) == 0) {
	horizWeight = 0;
    }
    if (gtk_group_get_total_vert_weight(group) == 0) {
	vertWeight = 0;
    }

    gtk_group_add_child(parent, widget, horizWeight, vertWeight);

    return rc;
}

static bool isBalance(DisplayObject *dio)
{
    short type = DisplayObjectType(dio);
    bool visible = DisplayObjectVisible(dio);

    return (type == DO_TYPE_BALANCE && visible);
}

static bool hasBalance(List *childs)
{
    struct DisplayObject *child;
    FORNODE (child, childs) {
        if (isBalance(child)) {
            return true;
        }
    }
    return false;
}

static gint gtk_widget_min_size(GtkWidget *widget, bool horiz)
{
    gint size;

#if GTK_MAJOR_VERSION == 2
    GtkRequisition requisition;
    gtk_widget_size_request(widget, &requisition);
    size = (horiz? requisition.width : requisition.height);
#else
    if (horiz) {
        gtk_widget_get_preferred_width(widget, &size, NULL);
    } else {
        gtk_widget_get_preferred_height(widget, &size, NULL);
    }
#endif
    return size;
}

static void setPanedPosition(GtkWidget *paned, struct DisplayObject *dio)
{
    DisplayObject *parent = DisplayObjectParent(dio);
    List *childs = DisplayObjectObjects(parent);
    GuiObject group1 = gtk_paned_get_child1(GTK_PANED(paned));
    GuiObject group2 = gtk_paned_get_child2(GTK_PANED(paned));

    if (!childs || !group1 || !group2) {
        return;
    }

    bool horiz = DisplayGroupHoriz(parent);     // false if not a group
    GuiObject parentgo = DisplayObjectGuiObject(parent);
    long space = !GTK_IS_GROUP(parentgo)? 0 :
        (horiz? gtk_group_get_horiz_spacing(GTK_GROUP(parentgo))
         : gtk_group_get_vert_spacing(GTK_GROUP(parentgo)));
    GtkAllocation alloc1, alloc2;
    gtk_widget_get_allocation(group1, &alloc1);
    gtk_widget_get_allocation(group2, &alloc2);
    long size1 = horiz? alloc1.width : alloc1.height;
    long size2 = horiz? alloc2.width : alloc2.height;
    size1 -= 2*gtk_container_get_border_width(GTK_CONTAINER(group1));
    size2 -= 2*gtk_container_get_border_width(GTK_CONTAINER(group2));
    long totalsize = size1 + size2 + 2*space, totalweight = 0, pos = -space;
    DisplayObject *child;

    FORNODE (child, childs) {
        GuiObject go = DisplayObjectRootObject(child);
        if (!go) {
            continue;
        }
        long weight = DisplayObjectWeight(child);
        totalweight += weight;
        totalsize -= space;
        if (weight == 0) {
            totalsize -= gtk_widget_min_size(go, horiz);
        }
    }
    if (totalweight <= 0) {
        return;                                 // not resizable
    }

    FORNODE (child, childs) {
        if (child == dio) {
            break;
        }
        GuiObject go = DisplayObjectRootObject(child);
        if (!go) {
            continue;
        }
        long weight = DisplayObjectWeight(child);
        long size = gtk_widget_min_size(go, horiz);
        long wsize = totalsize * weight / totalweight;
        if (wsize > size) {
            size = wsize;
        }
        pos += size + space;
    }
    gtk_paned_set_position(GTK_PANED(paned), (pos >= 0)? pos : 0);
}

static void onPanedMap(GtkWidget *paned, DisplayObject *dio)
{
    setPanedPosition(paned, dio);
}

static void setWeightToSize(GtkWidget *w, GtkGroup *group)
{
    gboolean horiz = gtk_group_get_horiz(group);
    guint horiz_weight, vert_weight;
    GtkAllocation alloc;

    gtk_group_query_child_packing(group, w, &horiz_weight, &vert_weight);
    gtk_widget_get_allocation(w, &alloc);
    if (horiz && horiz_weight > 0) {
        horiz_weight = alloc.width;
    } else if (!horiz && vert_weight > 0) {
        vert_weight = alloc.height;
    }
    gtk_group_set_child_packing(group, w, horiz_weight, vert_weight);
}

static void setPanedChildWeights(GtkWidget *w)
{
    if (GTK_IS_PANED(w)) {
        setPanedChildWeights(gtk_paned_get_child1(GTK_PANED(w)));
        setPanedChildWeights(gtk_paned_get_child2(GTK_PANED(w)));
    } else if (GTK_IS_GROUP(w)) {
        gtk_container_foreach(GTK_CONTAINER(w), 
                (GtkCallback)(setWeightToSize), w);
    }
}

static GtkWidget *changedPaned = 0;

static gboolean onPanedButtonPressEvent(GtkWidget *paned)
{
    changedPaned = 0;
    return false;
}

static gboolean onPanedButtonReleaseEvent(GtkWidget *paned)
{
    if (paned == changedPaned) {
        setPanedChildWeights(paned);
    }
    changedPaned = 0;
    return false;
}

static void onPanedNotifyPosition(GtkWidget *paned)
{
    changedPaned = paned;
}

bool BuildGuiChilds(List *childs, bool hideLabel, 
        GtkGroup *group, GtkGroup **colGroup)
{
    if (childs && group) {    
        gboolean horiz = gtk_group_get_horiz(group);
        int hs = gtk_group_get_horiz_spacing(group);
        int vs = gtk_group_get_vert_spacing(group);
        GtkWidget *root = 0;
        GtkWidget *paned = 0;
        GtkGroup *subGroup = group;
        gint hw = 0, vw = 0;
        if (hasBalance(childs)) {
            BuildGuiFinishColGroup(colGroup);
            if (horiz) {
                paned = HPanedObject, End;
            } else {
                paned = VPanedObject, End;
            }
            root = paned;
            GtkWidget *widget = GroupObject,
                "horiz", horiz,
                "horiz-spacing", hs,
                "vert-spacing", vs,
                End;
            subGroup = GTK_GROUP(widget);
            gtk_paned_pack1(GTK_PANED(paned), widget, true, false);
        }
	DisplayObject *child;
	FORNODE(child, childs) {
            if (isBalance(child) && paned) {
                BuildGuiFinishColGroup(colGroup);
                SetDisplayObjectGuiObject(child, paned);
                g_signal_connect(paned, "map", G_CALLBACK(onPanedMap), child);
                g_signal_connect(paned, "notify::position", 
                        G_CALLBACK(onPanedNotifyPosition), 0);
                g_signal_connect(paned, "button-press-event", 
                        G_CALLBACK(onPanedButtonPressEvent), 0);
                g_signal_connect(paned, "button-release-event", 
                        G_CALLBACK(onPanedButtonReleaseEvent), 0);
                hw += gtk_group_get_total_horiz_weight(subGroup);
                vw += gtk_group_get_total_vert_weight(subGroup);
                GtkWidget *widget = GroupObject,
                    "horiz", horiz,
                    "horiz-spacing", hs,
                    "vert-spacing", vs,
                    End;
                subGroup = GTK_GROUP(widget);
                if (hasBalance(SUCC(childs))) {
                    GtkWidget *subPaned;
                    if (horiz) {
                        subPaned = HPanedObject, End;
                    } else {
                        subPaned = VPanedObject, End;
                    }
                    gtk_paned_pack2(GTK_PANED(paned), subPaned, true, false);
                    paned = subPaned;
                    gtk_paned_pack1(GTK_PANED(paned), widget, true, false);
                } else {
                    gtk_paned_pack2(GTK_PANED(paned), widget, true, false);
                }
            } else if (!BuildGuiObject(child, hideLabel, subGroup, colGroup)) {
    		return false;
	    }
	    if (hideLabel && DisplayObjectGuiObject(child)) {
		hideLabel = false;
	    }
	}
        if (root) {
            hw += gtk_group_get_total_horiz_weight(subGroup);
            vw += gtk_group_get_total_vert_weight(subGroup);
            gtk_group_add_child(group, root, hw, vw);
        }
    }
    BuildGuiFinishColGroup(colGroup);

    return true;
}

void BuildGuiFinishColGroup(GtkGroup **colGroup)
{
    if (colGroup && *colGroup) {
	GtkWidget *parent = gtk_widget_get_parent(GTK_WIDGET(*colGroup));
	if (GTK_IS_GROUP(parent)) {
	    guint horiz_weight = gtk_group_get_total_horiz_weight(*colGroup);
	    guint vert_weight = gtk_group_get_total_vert_weight(*colGroup);
	    gtk_group_set_child_packing(GTK_GROUP(parent), 
		    GTK_WIDGET(*colGroup), horiz_weight, vert_weight);
	}
	*colGroup = 0;
    }
}

static bool BuildGuiObject(DisplayObject *dio, bool hideLabel,
	GtkGroup *group, GtkGroup **colGroup)
{
    if (!DisplayObjectVisibleGui(dio)) {
	return true;
    }

    switch (DisplayObjectType(dio)) {
	case DO_TYPE_FIELD:
	    return BuildGuiField(dio, hideLabel, group, colGroup);

	case DO_TYPE_TEXT:
	    BuildGuiFinishColGroup(colGroup);
    	    return buildGuiText(dio, group);
	    
	case DO_TYPE_IMAGE:
	    BuildGuiFinishColGroup(colGroup);
	    return buildGuiImage(dio, group);

	case DO_TYPE_SPACE:
	    BuildGuiFinishColGroup(colGroup);
	    return buildGuiSpace(dio, group);

	case DO_TYPE_BALANCE:
	    BuildGuiFinishColGroup(colGroup);
	    //return buildGuiBalance(dio, parent, group);
	    return true;
	    
	case DO_TYPE_TABLE:
	    BuildGuiFinishColGroup(colGroup);
	    return BuildGuiTable(dio, group);

	case DO_TYPE_REGISTER:
	    BuildGuiFinishColGroup(colGroup);
	    return buildGuiRegister(dio, group);

	case DO_TYPE_WINDOW:
	    return BuildGuiWindowButton(dio, group, colGroup);
	    
	case DO_TYPE_GROUP:
	    return buildGuiGroup(dio, group, colGroup);
	    
	default:
	    return false;
    }
}

