//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkBuildGuiField.cc
//
// Creation date: 05.04.2004
// Last modified: 09.02.2022
//

#include "system.h"
#include "gtk/gtkBuildGuiField.h"
#include "gtk/gtkBuildGuiRefFieldBrowser.h"
#include "gtk/gtkBuildGui.h"
#include "gtk/classes/gtkextimage.h"
#include "gtk/classes/gtksfwview.h"
#include "gtk/classes/gtkpopselection.h"
#include "gtk/classes/gtkpopwidget.h"
#include "gtk/classes/gtkradiobox.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include <gdk/gdkkeysyms.h>
#include "lang/program.h"
#include "main.h"
#include "programeditor.h"
#include "buildgui.h"
#include "display.h"
#include "gui.h"
#include "guidosutil.h"
#include "table.h"
#include "field.h"
#include "requesters.h"
#include "preferences.h"
#include "memo.h"
#include "labels.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

static GtkWidget *changedGuiObject = 0;
static GtkWidget *labelObject = 0;
static bool anyListActiveChanged = FALSE;

static const gfloat xalign[] = { 0.5, 0.0, 1.0 };

bool BuildGuiFieldLabel(DisplayObject *dio, GtkGroup *parent)
{
    struct GuiFieldFeatures f;

    if (!parent || !BuildGuiFieldGetFeatures(dio, &f)) {
	return false;
    }
    if (!f.label) {
	return true;	// nothing to do
    }

    short pos = DisplayFieldTitlePosition(dio);
    gfloat xalign;
    switch (pos) {
	case DAD_TITLEPOSITION_LEFT:
	    xalign = 1.0;
	    break;
	case DAD_TITLEPOSITION_RIGHT:
	    xalign = 0.0;
	    break;
	default:
	    xalign = 0.5;
	    break;
    }
    const char *title = DisplayObjectTitle(dio);
    gboolean use;
    gchar *label = GuiUnderline(title, DisplayObjectShortcut(dio), use);
    GtkWidget *widget = labelObject = LabelObject,
        "label", label,
	"xalign", xalign,
	"use-underline", use,
	End;
    g_free(label);
    BuildGuiSetFontname(DisplayFieldTitleGTKFontname(dio), widget);

    gtk_group_add_child(parent, widget, 0, 0);

    return true;
}

static GtkWidget *buildGuiFieldText(DisplayObject *dio, GtkWidget **text)
{
    GtkWidget *root = FrameObject,
	"shadow-type", GTK_SHADOW_IN,
	"child", *text = LabelObject,
	    "label", DisplayObjectTitle(dio),
	    "ypad", 1,
	    "xpad", 2,
	    "xalign", xalign[(int)DisplayFieldAlignment(dio)],
	    "width-request", 50,
	    End,
	End;

    return root;
}

static void guiObjectOnDestroy(GtkWidget *widget)
{
    if (widget == changedGuiObject) {
	changedGuiObject = 0;
    }
}

static void widgetOnChanged(GtkWidget *widget)
{
    changedGuiObject = widget;
}

static gboolean entryOnKeyPressEvent(GtkWidget *widget, GdkEventKey *event,
	Project *pr)
{
    if (event->keyval == GDK_KEY_Return) {
	GuiStoreEdits();
	if (GetPrefsValue(pr, PREFS_ADVANCEONENTER)) {
	    GtkWidget *win = gtk_widget_get_toplevel(widget);
	    if (GTK_IS_WINDOW(win)) {
		g_signal_emit_by_name(win, "move-focus", 
			(event->state & GDK_SHIFT_MASK)?
			GTK_DIR_TAB_BACKWARD : GTK_DIR_TAB_FORWARD);
		return true;
	    }
	}
    }
    return false;
}

static GtkWidget *buildGuiFieldString(DisplayObject *dio, int maxchars)
{
    Project *pr = DisplayObjectProject(dio);

    GtkWidget *entry = EntryObject,
	"max-length", maxchars,
	"width-chars", 1,
	"xalign", xalign[(int)DisplayFieldAlignment(dio)],
	End;
    gulong id = g_signal_connect(entry, "changed", 
	    G_CALLBACK(widgetOnChanged), 0);
    g_signal_connect(entry, "destroy", G_CALLBACK(guiObjectOnDestroy), 0);
    g_signal_connect(entry, "key-press-event", 
	    G_CALLBACK(entryOnKeyPressEvent), pr);
    g_object_set_data(G_OBJECT(entry), "handler", GUINT_TO_POINTER(id));

#if 0
    GdkColor bg, text, base;
    bg.red = 0;
    bg.green = 0;
    bg.blue = 0;
    text.red = 0xffff;
    text.green = 0xffff;
    text.blue = 0xffff;
    base.red = 0x7fff;
    base.green = 0x7fff;
    base.blue = 0x7fff;
    //gtk_widget_modify_bg(entry, GTK_STATE_NORMAL, &bg);
    //gtk_widget_modify_text(entry, GTK_STATE_NORMAL, &text);
    gtk_widget_modify_base(entry, GTK_STATE_NORMAL, &base);
#endif

    return entry;
}

static GtkWidget *buildGuiFieldMemo(DisplayObject *dio, GtkWidget **editor)
{
    GtkWidget *root = GtkNewEditor(DisplayObjectProject(dio), true, 
	    DisplayFieldReadOnly(dio), false, editor);

    g_signal_connect(*editor, "destroy", G_CALLBACK(guiObjectOnDestroy), 0);
    GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(*editor));
    gulong id = g_signal_connect_swapped(buffer, "changed", 
	    G_CALLBACK(widgetOnChanged), *editor);
    g_object_set_data(G_OBJECT(buffer), "handler", GUINT_TO_POINTER(id));

    return root;
}

static void checkBoxOnToggled(GtkToggleButton *cb, DisplayObject *dio)
{
    gboolean checked = gtk_toggle_button_get_active(cb);

    GuiStoreEdits();
    GuiStoreDataBool(dio, checked);
}

static GtkWidget *buildGuiFieldCheckBox(DisplayObject *dio)
{
    GtkWidget *cb = CheckButtonObject, 
        "sensitive", !DisplayFieldReadOnly(dio),
	End;
    gulong id = g_signal_connect(cb, "toggled", 
	    G_CALLBACK(checkBoxOnToggled), dio);    
    g_object_set_data(G_OBJECT(cb), "handler", GUINT_TO_POINTER(id));

    return cb;
}

static void choiceOnChanged(GtkComboBox *combo, DisplayObject *dio)
{
    GuiStoreEdits();
    GuiStoreDataLong(dio, gtk_combo_box_get_active(combo));
}

static GtkWidget *buildGuiFieldChoice(DisplayObject *dio)
{
    char **labels = FieldChoiceLabels(DisplayFieldField(dio));

    GtkWidget *combo = GuiComboBox(labels);
    gtk_widget_set_sensitive(combo, !DisplayFieldReadOnly(dio));

    gulong id = g_signal_connect(combo, "changed", 
	    G_CALLBACK(choiceOnChanged), dio);    
    g_object_set_data(G_OBJECT(combo), "handler", GUINT_TO_POINTER(id));

    return combo;
}

static void radioOnActiveChanged(GtkToggleButton *radio, 
	gint active, DisplayObject *dio)
{
    GuiStoreEdits();
    GuiStoreDataLong(dio, active);
}

static GtkWidget *buildGuiFieldRadioBox(
	DisplayObject *dio, GtkWidget **radioBox)
{
    const char *title = DisplayObjectTitle(dio);
    char **labels = FieldChoiceLabels(DisplayFieldField(dio));
    int frame = DisplayFieldFrame(dio);
    int alignment = DisplayFieldAlignment(dio);
    GtkWidget *root;

    if (title && !title[0]) {
	title = 0;
    }
    root = Frame(title),
	"shadow-type", frame? GTK_SHADOW_ETCHED_IN : GTK_SHADOW_NONE,
	"child", GroupAlignment(title),
	    "xscale", 0.0,
	    "xalign", 0.5,
	    "child", *radioBox = RadioBoxObject,
                "horiz", alignment == DAD_ALIGNMENT_HORIZONTAL,
                "entries", labels,
		End,
	    End,
	End;

    gulong id = g_signal_connect(*radioBox, "active-changed",
	    G_CALLBACK(radioOnActiveChanged), dio);    
    g_object_set_data(G_OBJECT(*radioBox), "handler", GUINT_TO_POINTER(id));

    return root;
}

static GtkWidget *makeButton(DisplayObject *dio, const gchar *icon_name)
{
    const char *title = DisplayObjectTitle(dio);
    char shortcut = DisplayObjectShortcut(dio);
    gfloat align = xalign[(int)DisplayFieldAlignment(dio)];

    return BuildGuiIconLabelButton(icon_name, title, shortcut, align);
}

static void buttonOnClicked(GtkWidget *button, DisplayObject *dio)
{
    GuiStoreEdits();
    GuiStoreData(dio);
}

static GtkWidget *buildGuiFieldButton(DisplayObject *dio)
{
    GtkWidget *button = makeButton(dio, 0);
    GtkWidget *label = (GtkWidget*)g_object_get_data(G_OBJECT(button), "label");
    BuildGuiSetFont(dio, label);
	    
    g_signal_connect(button, "clicked", G_CALLBACK(buttonOnClicked), dio);

    return button;
}

static void showButtonOnClicked(GtkWidget *button, DisplayObject *dio)
{
    Field *field = DisplayFieldField(dio);
    Table *tab = FieldTable(field);

    GuiStoreEdits();
    GuiShowReferenceRecord(field, TableGuiRecord(tab));
}

static GtkWidget *buildGuiFieldShowButton(DisplayObject *dio)
{
    GtkWidget *button = makeButton(dio, "go-jump");
    GtkWidget *label = (GtkWidget*)g_object_get_data(G_OBJECT(button), "label");
    g_object_set(G_OBJECT(label), "width-request", 50, NULL);
    BuildGuiSetFont(dio, label);
	    
    g_signal_connect(button, "clicked", G_CALLBACK(showButtonOnClicked), dio);

    return button;
}

static GtkWidget *buildGuiFieldImageButton(DisplayObject *dio, 
	GtkWidget **button)
{
    struct Project *pr = DisplayObjectProject(dio);
    const char *bg = DisplayObjectBackground(dio);
    char imageSize = DisplayFieldImageSize(dio);
    bool resize = (imageSize == DAD_IMAGESIZE_RESIZABLE);
    GdkPixbufAnimation *anim = NULL;

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    if (bg && bg[0] == '5' && bg[1] == ':') {
	gchar *path = gtkExpandPathname(&bg[2]);
	anim = gdk_pixbuf_animation_new_from_file(path, 0);
	g_free(path);
    }

    *button = ButtonObject,
        "child", ImageObject,
	    "pixbuf-animation", anim, 
	    End,
	End;

    if (anim) {
	g_object_unref(G_OBJECT(anim));
    }
    g_signal_connect(*button, "clicked", G_CALLBACK(buttonOnClicked), dio);

    return AlignmentObject,
        "xscale", resize? 1.0 : 0.0,
	"yscale", resize? 1.0 : 0.0,
	"child", *button,
	End;
}
    
void SFWViewOnRowActivated(GtkSFWView *sfw,
	GtkTreePath *path, GtkTreeViewColumn *column, DisplayObject *dio)
{
    struct Field *field;
    struct Record *rec;
    long line, col;

    gtk_sfw_view_get_entry(sfw, path, column, &line, &col, &field, &rec);
    GuiVirtualFieldListDoubleClick(dio, line, col, field, rec);
}

void sfwViewOnInternalMoveDrop(GtkSFWView *sfw, glong row, DisplayObject *dio)
{
    GuiVirtualFieldListSortDrop(dio, row);
}

static void sfwViewOnSelectionChanged(GtkTreeSelection *sel, DisplayObject *dio)
{
    gint active = GuiListActive(DisplayObjectGuiObject(dio));
    gint old = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(sel), "active"));

    if (active != old) {
        g_object_set_data(G_OBJECT(sel), "active", GINT_TO_POINTER(active));
        SetDisplayFieldListActiveChanged(dio, TRUE);
        anyListActiveChanged = TRUE;
    }
}

static GdkAtom getUriDropTarget(GtkWidget *sfw, GdkDragContext *context,
        DisplayObject *dio)
{
    // Check that the drag context is a copy or link action.
    GdkDragAction action = gdk_drag_context_get_suggested_action(context);
    if (action != GDK_ACTION_COPY && action != GDK_ACTION_LINK) {
        return GDK_NONE;
    }

    // Check that this is an URI target.
    GtkTargetList *targets = gtk_target_list_new(0, 0);
    gtk_target_list_add_uri_targets(targets, 0);
    GdkAtom target = gtk_drag_dest_find_target(sfw, context, targets);
    gtk_target_list_unref(targets);

    return target;
}

static gboolean sfwViewOnDragMotion(GtkWidget *sfw, GdkDragContext *context,
        gint x, gint y, guint time, DisplayObject *dio)
{
    if (getUriDropTarget(sfw, context, dio) == GDK_NONE) {
        return false;
    }

    // We accept the drop.
    gdk_drag_status(context,
            gdk_drag_context_get_suggested_action(context), time);
    gtk_drag_highlight(sfw);

    return true;
}

static void sfwViewOnDragLeave(GtkWidget *sfw, GdkDragContext *context, 
        guint time, DisplayObject *dio)
{
    if (getUriDropTarget(sfw, context, dio) != GDK_NONE) {
        gtk_drag_unhighlight(sfw);
    }
}

static gboolean sfwViewOnDragDrop(GtkWidget *sfw, GdkDragContext *context, 
        gint x, gint y, guint time, DisplayObject *dio)
{
    GdkAtom target = getUriDropTarget(sfw, context, dio);
    if (target == GDK_NONE) { 
        return false;
    }

    gtk_drag_get_data(sfw, context, target, time);
    return true;
}

static void sfwViewOnDragDataReceived(GtkWidget *sfw, GdkDragContext *context,
        gint x, gint y, GtkSelectionData *data, guint info, guint time,
        DisplayObject *dio)
{
    GdkAtom target = getUriDropTarget(sfw, context, dio);
    gboolean success = false;

    if (gtk_selection_data_get_length(data) >= 0 && target != GDK_NONE) {
        Project *pr = DisplayObjectProject(dio);
        MemPool *memPool = ProjectMemPool(pr);
        gchar **uris = gtk_selection_data_get_uris(data), **uri;
        size_t num = g_strv_length(uris), n = 0;
        char **labels = NULL;
        long error = OK;

        labels = AllocLabels(memPool, num);
        if (labels == NULL) {
            error = ERR_MEM;
        } else if (uris) {
            for (uri = uris; *uri; uri++) {
                const char *filename = *uri;

                // Remove URL specifier of local files.
                if (!strncmp(filename, "file://", 7)) {
                    filename += 7;
                }
                if (filename[0] != 0) {
                    labels[n] = AllocString(memPool, filename);
                    if (labels[n] == NULL) {
                        error = ERR_MEM;
                        break;
                    }
                    n++;
                }
            }
        }
        if (error) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        } else {
            GuiVirtualFieldListDropUrls(dio, labels);
        }
        FreeLabelsNum(memPool, labels, num);
        g_strfreev(uris);
        success = true;
    }
    gtk_drag_finish(context, success, false, time);
}

static GtkWidget *buildGuiFieldSFWView(DisplayObject *dio, GtkWidget **sfw)
{
    GtkWidget *root;

    root = ScrolledWindowObject,
	"shadow-type", GTK_SHADOW_ETCHED_IN,
	"hscrollbar-policy", GTK_POLICY_AUTOMATIC,
	"child", *sfw = SFWViewObject,
	    "rules-hint", true,
	    "show-titles", DisplayFieldShowTitles(dio),
	    End,
	End;

    g_signal_connect(*sfw, "row-activated", 
	    G_CALLBACK(SFWViewOnRowActivated), dio);
    g_signal_connect(*sfw, "internal-move-drop",
	    G_CALLBACK(sfwViewOnInternalMoveDrop), dio);

    const char *urlDropTrigger = DisplayFieldURLDropTrigger(dio);
    if (urlDropTrigger != 0 && urlDropTrigger[0] != 0) {
        gtk_drag_dest_add_uri_targets(*sfw);
        g_signal_connect(*sfw, "drag-motion",
                G_CALLBACK(sfwViewOnDragMotion), dio);
        g_signal_connect(*sfw, "drag-leave",
                G_CALLBACK(sfwViewOnDragLeave), dio);
        g_signal_connect(*sfw, "drag-drop", G_CALLBACK(sfwViewOnDragDrop), dio);
        g_signal_connect(*sfw, "drag-data-received",
                G_CALLBACK(sfwViewOnDragDataReceived), dio);
    }

    GtkTreeSelection *sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(*sfw));
    g_object_set_data(G_OBJECT(sel), "active", GINT_TO_POINTER(-2));
    g_signal_connect(sel, "changed", G_CALLBACK(sfwViewOnSelectionChanged),dio);

    return root;
}

static GtkWidget *buildGuiPopSelection(DisplayObject *dio, GtkWidget *string,
	GType type, GtkWidget *button)
{
    g_object_set_data(G_OBJECT(button), "dio", dio);
    if (gtk_widget_get_can_focus(string)) {
	g_signal_connect(button, "focus", G_CALLBACK(OnExtraButtonFocus), dio);
    }

    GtkWidget *pop = PopSelectionObject,
        "selection-type", type,
        "string", string,
	"button", button,
	End;
    g_signal_connect_after(pop, "stop", G_CALLBACK(GuiStoreEdits), NULL);

    return pop;
}

static GtkWidget *buildGuiPopFile(DisplayObject *dio, GtkWidget *string,
	GtkWidget **button)
{
    *button = GuiPopButtonSetCWD(SELECT_FILE_ICON_NAME,
            DisplayObjectProject(dio));
    return buildGuiPopSelection(dio, string, 
	    GTK_TYPE_FILE_CHOOSER_DIALOG, *button);
}

#if 0
/*
** There are a few problems with GtkFileChooserButton.
** Search the bug database at bugzilla.gnome.org for more information.
** Therefore we do not use this widget for now
*/
static void fileButtonOnSelectionChanged(GtkFileChooserButton *fb, 
	DisplayObject *dio)
{
    gchar *filename = gtk_ami_file_chooser_get_filename(GTK_FILE_CHOOSER(fb));
    const gchar *old = (gchar *)g_object_get_data(G_OBJECT(fb), "filename");

    if (filename == NULL) {
	return;			/* happens, e.g. when changing directories */
    }

    GuiStoreEdits();

    if (old == NULL || strcmp(filename, old) != 0) {
	printf("file button changed to '%s'\n", filename);
	g_object_set_data_full(G_OBJECT(fb), "filename", filename, g_free);
	GuiStoreDataStr(dio, filename);
    } else {
	g_free(filename);
    }
}

static GtkWidget *buildGuiFileButton(DisplayObject *dio)
{
    GtkWidget *dialog = gtk_ami_file_chooser_new(0, 0, 0, 
	    GTK_FILE_CHOOSER_ACTION_OPEN, 0);
    GtkWidget *fb = gtk_file_chooser_button_new_with_dialog(dialog);
    gulong id = g_signal_connect(fb, "selection-changed",
	    G_CALLBACK(fileButtonOnSelectionChanged), dio);
    g_object_set_data(G_OBJECT(fb), "handler", GUINT_TO_POINTER(id));

    return fb;
}
#endif

static GtkWidget *buildGuiPopFont(DisplayObject *dio, GtkWidget *string,
	GtkWidget **button)
{
    *button = GuiPopButton(SELECT_FONT_ICON_NAME);
    return buildGuiPopSelection(dio, string, 
	    GUI_TYPE_FONT_CHOOSER_DIALOG, *button);
}

static gboolean popCalendarOnStart(GtkPopWidget *pop, DisplayObject *dio)
{
    GtkWidget *calendar = gtk_pop_widget_get_widget(pop);
    const gchar *text = gtk_pop_string_get_text(GTK_POP_STRING(pop));
    long date;
    short day, month, year;

    g_signal_stop_emission_by_name(pop, "start");

    if (StringToDate(text, &date)) {
        DateToDayMonthYear(date, &day, &month, &year);
        gtk_calendar_select_month(GTK_CALENDAR(calendar), month - 1, year);
        gtk_calendar_select_day(GTK_CALENDAR(calendar), day);
    }

    return true;
}

static void popCalendarOnStop(GtkPopWidget *pop, DisplayObject *dio)
{
    struct Project *pr = DisplayObjectProject(dio);
    GtkWidget *calendar = gtk_pop_widget_get_widget(pop);
    guint day, month, year;
    long date;
    const char *text;

    gtk_calendar_get_date(GTK_CALENDAR(calendar), &year, &month, &day);
    if (DayMonthYearToDate(day, month + 1, year, &date)) {
        text = DateToString(date, GetPrefsValue(pr, PREFS_DATEFORMAT));

        GuiStoreDataStr(dio, text);
    }
}

static void popCalendarOnDaySelectedDoubleClick(GtkCalendar *calendar,
        GtkPopString *pop)
{
    gtk_pop_string_close_popup(pop, TRUE);
}

static GuiObject buildGuiPopCalendar(DisplayObject *dio, GtkWidget *string,
        GtkWidget **button)
{
    *button = ButtonObject,
        "child", ImageObject, 
	    "icon-name", "gtk-index", 
	    "icon-size", GTK_ICON_SIZE_MENU,
	    End,
	End;
    if (GTK_IS_ENTRY(string) || GTK_IS_BUTTON(string)) {
	g_signal_connect(*button, "focus", G_CALLBACK(OnExtraButtonFocus), dio);
    }

    GtkWidget *calendar = gtk_calendar_new();

    GtkWidget *pop = PopWidgetObject,
        //"light", false,
        "string", string,
	"button", *button,
	"widget", calendar,
	End;
    g_object_set_data(G_OBJECT(pop), "dio", dio);
    g_signal_connect(pop, "start", G_CALLBACK(popCalendarOnStart), dio);
    g_signal_connect(pop, "stop", G_CALLBACK(popCalendarOnStop), dio);
    g_signal_connect(calendar, "day-selected-double-click", 
            G_CALLBACK(popCalendarOnDaySelectedDoubleClick), pop);

    return pop;
}

static gboolean popListviewOnStart(GtkPopWidget *pop, DisplayObject *dio)
{
    GtkWidget *window = gtk_pop_string_get_popup(GTK_POP_STRING(pop));
    GtkWidget *widget = gtk_pop_widget_get_widget(pop);
    GtkWidget *tree = gtk_bin_get_child(GTK_BIN(widget));
    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
    GtkListStore *store = GTK_LIST_STORE(model);
    struct Project *pr = DisplayObjectProject(dio);
    struct MemPool *memPool = ProjectMemPool(pr);
    const char *trigger = DisplayFieldListviewCompute(dio);
    char **labels = NULL;
    bool labelsAllocated = FALSE;

    GuiStoreEdits();   // maybe the user has just edited a string object

    g_signal_stop_emission_by_name(pop, "start");

    if (trigger != NULL && trigger[0] != 0) {
	struct Program *prg = ProjectProgram(pr);
	struct UserFunction *uf = FindUserFunctionName(prg, trigger);
	const char *filename = NULL, *error = NULL, *elemError = NULL;
	long line, col;
	struct Element *elem;
	char *str = NULL;

	if (uf == NULL) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Trigger), trigger);
	    return false;
	} else {
	    /* GuiSetSleep(TRUE); */
	    SetTableProgramRecords(pr);
	    ProgramBegin(prg, TRUE);
	    elem = ProgramCall(prg, uf, NULL, TRUE);
	    if (elem) {
		switch (elem->el_Type) {
		    case TYPE_STRING:
			str = elem->el_Contents.string.string;
			break;
		    case TYPE_MEMO:
			str = elem->el_Contents.mi;
			break;
		    case TYPE_NIL:
			break;
		    default:
			elemError = GCS(MSG_Err_Listview_Compute);
			break;
		}
		if (str) {
		    labels = StringToLabels(memPool, str);
		    if (labels == NULL) {
			elemError = GCS(MSG_Err_Mem);
		    }
		    labelsAllocated = TRUE;
		}
	    }
            ProgramEnd();
	    error = ProgramError(&line, &col, &filename);
	    /* GuiSetSleep(FALSE); */

	    if (error != NULL && error != EXECUTE_ABORT) {
		ProgramEditorShowError(pr, TRUE, error, line, col, filename);
		if (labelsAllocated) {
		    FreeLabels(memPool, labels);
		}
		return false;
	    }
	    if (elemError) {
		Ask(AT_ERROR, GCS(MSG_StarOk), elemError);
		return false;
	    }
	}
    } else {
	labels = DisplayFieldListviewLabels(dio);
    }

    if (store) {
	GuiSetSleep(TRUE);
	GuiSetListStoreLabels(store, labels);

	gint active = G_MAXINT;

	if (labels) {
	    Field *field = DisplayFieldField(dio);
	    short type = FieldType(field);
	    if (type == TYPE_STRING) {
		Table *tab = FieldTable(field);
		Record *rec = TableGuiRecord(tab);
		DataInfo *di = 0, di_buf;

		RecordLock(rec);
		RecordDataInfo(rec, field, &di, &di_buf);
		if (di) {
		    const char *s = di->di_Contents.string;
                    if (s) {
                        for (long i = 0; labels[i] != NULL; i++) {
                            if (strcmp(s, labels[i]) == 0) {
                                active = i;
                                break;
                            }
                        }
                    }
		}
		RecordUnlock(rec);
	    }
	}

	gtk_widget_grab_focus(tree);

	GuiListSetActive(tree, active);

	gint width, height;
	gtk_window_get_size(GTK_WINDOW(window), &width, &height);
	height += 100;
	gtk_window_resize(GTK_WINDOW(window), width, height);

	GuiSetSleep(FALSE);
    }

    if (labelsAllocated) {
	FreeLabels(memPool, labels);
    }

    return true;
}

static void popListviewOnStop(GtkPopWidget *pop, DisplayObject *dio)
{
    GtkWidget *widget = gtk_pop_widget_get_widget(pop);
    GtkWidget *tree = gtk_bin_get_child(GTK_BIN(widget));
    const char *label = (const char *)GuiListActiveEntry(tree);

    GuiStoreDataStr(dio, label);
}

static void popListviewOnRowActivated(GtkTreeView *tree,
	GtkTreePath *path, GtkTreeViewColumn *column, GtkPopString *pop)
{
    gtk_pop_string_close_popup(pop, TRUE);
}

static GtkWidget *buildGuiPopListview(DisplayObject *dio, GtkWidget *string,
	GtkWidget **button)
{
    *button = ButtonObject,
        "child", ImageObject, 
	    "icon-name", "gtk-index", 
	    "icon-size", GTK_ICON_SIZE_MENU,
	    End,
	End;
    if (GTK_IS_ENTRY(string) || GTK_IS_BUTTON(string)) {
	g_signal_connect(*button, "focus", G_CALLBACK(OnExtraButtonFocus), dio);
    }

    GtkListStore *store = gtk_list_store_new(1, G_TYPE_STRING);
    GtkWidget *tree = TreeViewObject,
        "model", store,
	"headers-visible", false,
	End;
    g_object_unref(G_OBJECT(store));

    GtkTreeSelection *selection = 
	    gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);

    GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
    GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
	    "Item", renderer, "text", 0, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

    GtkWidget *pop = PopWidgetObject,
        //"light", false,
        "string", string,
	"button", *button,
	"widget", ScrolledWindowObject,
	    "hscrollbar-policy", GTK_POLICY_NEVER,
	    "shadow-type", GTK_SHADOW_IN,
	    "child", tree,
	    End,
	End;
    g_object_set_data(G_OBJECT(pop), "dio", dio);
    g_signal_connect(pop, "start", G_CALLBACK(popListviewOnStart), dio);
    g_signal_connect(pop, "stop", G_CALLBACK(popListviewOnStop), dio);
    g_signal_connect(tree, "row-activated", 
	    G_CALLBACK(popListviewOnRowActivated), pop);
    g_signal_connect_data(pop, "close", G_CALLBACK(gtk_list_store_clear),
	    store, NULL, (GConnectFlags)(G_CONNECT_AFTER | G_CONNECT_SWAPPED));

    return pop;
}

static void viewButtonOnClicked(GtkButton *button, GtkWidget *guiobject)
{
    GuiStoreEdits();
    GuiObjectViewFile(guiobject);
}

static GtkWidget *buildGuiViewButton(GtkWidget *guiobject, DisplayObject *dio)
{
    GtkWidget *button = ButtonObject,
        "child", ImageObject, 
	    "icon-name", "text-x-generic",
	    "icon-size", GTK_ICON_SIZE_MENU,
	    End,
	End;
    g_signal_connect(button, "clicked", 
	    G_CALLBACK(viewButtonOnClicked), guiobject);
    g_signal_connect(button, "focus", G_CALLBACK(OnExtraButtonFocus), dio);

    GuiSetTooltip(button, GCS(MSG_ViewBH));

    return button;
}

static void autoShowButtonOnToggled(GtkToggleButton *button, DisplayObject *dio)
{
    GuiStoreEdits(); 
    ReferenceAutoShowOnOff(DisplayFieldField(dio),
	    gtk_toggle_button_get_active(button));
}

static GtkWidget *buildGuiAutoShowButton(DisplayObject *dio)
{
    GtkWidget *button = ToggleButtonObject,
        "label", GCS(MSG_ReferenceAutoShow),
	End;

    g_signal_connect(button, "toggled", 
	    G_CALLBACK(autoShowButtonOnToggled), dio);    
    g_signal_connect(button, "focus", G_CALLBACK(OnExtraButtonFocus), dio);
    GuiSetTooltip(button, GCS(MSG_ReferenceAutoShowBH));

    return button;
}

static void filterButtonOnToggled(GtkToggleButton *button, DisplayObject *dio)
{
    GuiStoreEdits(); 
    ReferenceFilterOnOff(DisplayFieldField(dio),
	    gtk_toggle_button_get_active(button));
}

static GtkWidget *buildGuiFilterButton(DisplayObject *dio)
{
    GtkWidget *button = ToggleButtonObject,
        "label", GCS(MSG_ReferenceFilter),
	End;

    g_signal_connect(button, "toggled", 
	    G_CALLBACK(filterButtonOnToggled), dio);    
    g_signal_connect(button, "focus", G_CALLBACK(OnExtraButtonFocus), dio);
    GuiSetTooltip(button, GCS(MSG_ReferenceFilterBH));

    return button;
}

bool BuildGuiField(DisplayObject *dio, bool hideLabel, 
	GtkGroup *parent, GtkGroup **colGroup)
{
    Field *field = DisplayFieldField(dio);
    short titlepos = DisplayFieldTitlePosition(dio);
    int weight = DisplayObjectWeight(dio);
    int vertWeight, horizWeight;
    short type = FieldType(field);
    GuiFieldFeatures f;
    GuiDisplayObjectSizeConstraints c;
    
    if (parent == 0 || colGroup == 0 || !BuildGuiFieldGetFeatures(dio, &f)){
	return false;			// be safe
    }

    bool horiz = gtk_group_get_horiz(parent);
    BuildGuiDisplayObjectGetSizeConstraints(dio, horiz, &c);
    bool postAddTitle = false;
    GtkWidget *go = 0, *guiobject = 0, *root = 0, *activateobject = 0;
    GtkWidget *externalobject = 0;
    GtkWidget *fileobject = 0, *fontobject = 0, *lvobject = 0, *viewobject = 0; 
    GtkWidget *filterobject = 0, *autoshowobject = 0;
    if (hideLabel) {
	f.label = false;
    }
    if (horiz) {
        horizWeight = weight;
        vertWeight = 100;
    } else {
        horizWeight = 100;
        vertWeight = weight;
    }

    if (f.label && !horiz && (titlepos == DAD_TITLEPOSITION_LEFT || 
		titlepos == DAD_TITLEPOSITION_RIGHT)) {
	if (!*colGroup) {
	    GtkWidget *widget = VGroup, "columns", 2, End;
	    gtk_group_add_child(parent, widget, 100, 100);
	    *colGroup = GTK_GROUP(widget);
	}
	parent = *colGroup;
	if (titlepos == DAD_TITLEPOSITION_RIGHT) {
	    postAddTitle = true;
	} else if (!BuildGuiFieldLabel(dio, parent)) {
    	    return false;
	}
    } else {
	BuildGuiFinishColGroup(colGroup);
	if (f.label) {
	    switch (titlepos) {
		case DAD_TITLEPOSITION_LEFT:
		case DAD_TITLEPOSITION_RIGHT:
	    	    if (titlepos == DAD_TITLEPOSITION_RIGHT) {
     			postAddTitle = true;
		    } else if (!BuildGuiFieldLabel(dio, parent)) {
			return false;
		    }
		    break;

		case DAD_TITLEPOSITION_TOP:
		case DAD_TITLEPOSITION_BOTTOM:
		    if (horiz) {
			GtkWidget *widget = root = VGroup, End;
			gtk_group_add_child(parent, widget,
				c.can_grow_horiz? horizWeight : 0,
                                c.can_grow_vert? vertWeight : 0);
			parent = GTK_GROUP(widget);
			horiz = false;
		    }
	    	    if (titlepos == DAD_TITLEPOSITION_BOTTOM) {
     			postAddTitle = true;
		    } else if (!BuildGuiFieldLabel(dio, parent)) {
			return false;
		    }
		    break;
	    }
	}
    }

    if (f.maxchars >= 0 || f.text) {
	if (f.show) {
	    go = buildGuiFieldShowButton(dio);
	    guiobject = activateobject = go;
	} else if (!f.readonly) {
	    go = buildGuiFieldString(dio, f.maxchars);
	    activateobject = guiobject = go;
#if 0
	} else if (DisplayFieldFile(dio)) {
	    go = buildGuiFileButton(dio);
	    guiobject = activateobject = go;
#endif
	} else {
	    go = buildGuiFieldText(dio, &guiobject);
	}
    } else if (f.checkbox) {
	go = buildGuiFieldCheckBox(dio);
	activateobject = guiobject = go;
    } else if (f.radiobuttons) {
      	go = buildGuiFieldRadioBox(dio, &guiobject);
	activateobject = guiobject;
    } else if (f.choice) {
        go = buildGuiFieldChoice(dio);
	activateobject = guiobject = go;
    } else if (f.button) {
	go = buildGuiFieldButton(dio);
	activateobject = guiobject = go;
    } else if (f.image) {
	go = buildGuiFieldImageButton(dio, &guiobject);
	activateobject = guiobject;
    } else if (f.editor) {
    	go = buildGuiFieldMemo(dio, &guiobject);
    	activateobject = guiobject;
    } else if (f.listview) {
    	go = buildGuiFieldSFWView(dio, &guiobject);
    	activateobject = guiobject;
    } else if (f.maxchars != -1) {
        return false;			// unknown gui object
    }

    if (f.maxchars != -1 && !go) {
	return false;
    }

    go = GuiMakeTooltip(go, DisplayFieldBubbleHelp(dio));

    if (f.refpopup && go) {
	go = BuildGuiPopRefFieldBrowser(dio, go, &lvobject);
	if (!activateobject) {
	    activateobject = lvobject;
	}
        SetDisplayFieldPopObject(dio, go);
    }
    if (type == TYPE_DATE && DisplayFieldCalendar(dio) && !f.readonly && go) {
        go = buildGuiPopCalendar(dio, go, &lvobject);
	if (!activateobject) {
	    activateobject = lvobject;
	}
    }
    if (type == TYPE_STRING && DisplayFieldFile(dio) && go) {
	go = buildGuiPopFile(dio, go, &fileobject);
	if (!activateobject) {
	    activateobject = fileobject;
	}
    } else if (type == TYPE_STRING && DisplayFieldFont(dio) && go) {
	go = buildGuiPopFont(dio, go, &fontobject);
	if (!activateobject) {
	    activateobject = fontobject;
	}
    } else if (type == TYPE_STRING && DisplayFieldListview(dio) && go) {
	go = buildGuiPopListview(dio, go, &lvobject);
	if (!activateobject) {
	    activateobject = lvobject;
	}
    }

    if (type == TYPE_STRING && DisplayFieldView(dio) && go) {
        viewobject = buildGuiViewButton(guiobject, dio);

#if GTK_MAJOR_VERSION == 2
        if (GTK_IS_POP_STRING(go)) {
            GtkWidget *pop = gtk_pop_string_get_button(GTK_POP_STRING(go));
            if (pop) {
                g_signal_connect_swapped(viewobject, "size-request",
                        G_CALLBACK(gtk_widget_size_request), pop);
            }
        }
#endif
	go = HGroup,
	    "spacing", 0,
	    "child", go,
	    "child", viewobject, "weight", 0,
	    End;
    }
    if (f.datatypes) {
	GuiObject str_go = go;
        char size = DisplayFieldImageSize(dio);
        bool fixed = (size == DAD_IMAGESIZE_SCROLLABLE || 
		size == DAD_IMAGESIZE_FIXED);
        bool scaled = (size == DAD_IMAGESIZE_SCALED || 
		size == DAD_IMAGESIZE_ASPECTSCALED);
	const char *ctitle = DisplayObjectTitle(dio);

        if (str_go && DisplayFieldTitleAtStringField(dio) && 
		ctitle && ctitle[0]) {
	    gboolean use;
	    gchar *title = GuiUnderline(ctitle, DisplayObjectShortcut(dio),use);
	    str_go = HGroup,
	        "child", LabelObject,
		    "label", title,
		    "xalign", 1.0,
		    "use-underline", use,
		    "mnemonic-widget", activateobject,
		    End, "weight", 0,
		"child", str_go,
		End;
	    g_free(title);
	}

	go = externalobject = ExtImageObject,
	    "fixed", fixed,
	    "scaled", scaled,
            "keep-aspect", (size == DAD_IMAGESIZE_ASPECTSCALED),
	    End;

	if (size == DAD_IMAGESIZE_SCROLLABLE) {
	    go = ScrolledWindowObject,
	         // "shadow-type", GTK_SHADOW_IN,
	         // "hscrollbar-policy", GTK_POLICY_NEVER,
	         "child", ViewportObject,
		     "child", go,
		     End,
		 End;
	} else {
	    go = FrameObject,
	        "shadow-type", GTK_SHADOW_IN,
	        "child", go,
		End;
	    if (size == DAD_IMAGESIZE_FIXED) {
		go = AlignmentObject,
		    "xscale", 0.0,
		    "yscale", 0.0,
		    "child", go,
		    End;
	    }
	}

        if (str_go) {
            go = VGroup,
	        "spacing", 0,
		"same-width", true,
		"child", go,
		"child", str_go, "weight", 0,
		End;
	} else {
            go = GuiMakeTooltip(go, DisplayFieldBubbleHelp(dio));
        }
    }
    if (type== TYPE_REFERENCE && go && 
            (DisplayFieldAutoShow(dio) || DisplayFieldFilter(dio))) {
#if GTK_MAJOR_VERSION == 2
        GtkWidget *pop = 0;

        if (GTK_IS_POP_STRING(go)) {
            pop = gtk_pop_string_get_button(GTK_POP_STRING(go));
        }
#endif

        GtkWidget *grp = HGroup,
              "spacing", 0,
              "same-size", true,
              End;

        if (DisplayFieldAutoShow(dio)) {
            autoshowobject = buildGuiAutoShowButton(dio);
#if GTK_MAJOR_VERSION == 2
            if (pop) {
                g_signal_connect_swapped(autoshowobject, "size-request",
                        G_CALLBACK(gtk_widget_size_request), pop);
            }
#endif
            gtk_group_add_child(GTK_GROUP(grp), autoshowobject, 1, 1);
        }
        if (DisplayFieldFilter(dio)) {
	    filterobject = buildGuiFilterButton(dio);
#if GTK_MAJOR_VERSION == 2
            if (pop) {
                g_signal_connect_swapped(filterobject, "size-request",
                        G_CALLBACK(gtk_widget_size_request), pop);
            }
#endif
            gtk_group_add_child(GTK_GROUP(grp), filterobject, 1, 1);
        }

        go = HGroup,
           "spacing", 0,
           "child", go,
           "child", grp, "weight", 0,
           End;
    }

    if (!go) {
	return false;
    }

    if (!c.can_grow_horiz || !c.can_grow_vert) {
	go = AlignmentObject,
	    "xalign", postAddTitle? 1.0 : 0.0,
	    "xscale", (!*colGroup || c.can_grow_horiz)? 1.0 : 0.0,
	    "yalign", 0.5,
	    "yscale", c.can_grow_vert? 1.0 : 0.0,
	    "child", go,
	    End;
    }
    if (!DisplayFieldTabChain(dio)) {
	go = VGroup, "child", go, End;
	gtk_container_set_focus_chain(GTK_CONTAINER(go), NULL);
    }
    gtk_group_add_child(parent, go, c.can_grow_horiz? horizWeight : 0,
            c.can_grow_vert? vertWeight : 0);
    if (postAddTitle && !BuildGuiFieldLabel(dio, parent)) {
	return false;
    }

    if (labelObject) {
	g_object_set(G_OBJECT(labelObject), 
                "mnemonic-widget", activateobject, 
                NULL);
	labelObject = 0;
    }
    BuildGuiAddCommon(dio, guiobject);
    BuildGuiSetBackground(dio, guiobject);
    BuildGuiSetFont(dio, guiobject);
    GuiSetHelpNode(guiobject, "Changing records");

    SetDisplayObjectGuiObject(dio, guiobject);
    SetDisplayObjectRootObject(dio, root? root : go);
    SetDisplayObjectActivateObject(dio, activateobject);
    SetDisplayFieldFileObject(dio, fileobject);
    SetDisplayFieldFontObject(dio, fontobject);
    SetDisplayFieldListviewObject(dio, lvobject);
    SetDisplayFieldViewObject(dio, viewobject);
    SetDisplayFieldExternalObject(dio, externalobject);
    SetDisplayFieldFilterObject(dio, filterobject);
    SetDisplayFieldAutoShowObject(dio, autoshowobject);
    SetDisplayFieldDisableObject(dio, go);

    return true;
}

void GuiFieldStoreEdits() 
{
    GtkWidget *w = changedGuiObject;
    if (w) {
        changedGuiObject = 0;   // avoid recursion, clear it immediately
	DisplayObject *dio = GuiObjectDisplayObject(w);
	if (dio) {
	    if (GTK_IS_ENTRY(w)) {
		const gchar *text = gtkEntryText(w);
		GuiStoreDataStr(dio, text);
	    } else if (GTK_IS_TEXT_VIEW(w)) {
		GtkTextBuffer *buffer = gtk_text_view_get_buffer(
			GTK_TEXT_VIEW(w));
		GtkTextIter start, end;
		gtk_text_buffer_get_start_iter(buffer, &start);
		gtk_text_buffer_get_end_iter(buffer, &end);
		gchar *text = gtk_text_buffer_get_text(buffer, 
			&start, &end, true);
                if (text) {
                    GuiStoreDataStr(dio, text);
                    g_free(text);
                }
	    }
	}
    }
}

void GuiSetString(GuiObject go, const char *str, bool readonly, 
        GuiObject external)
{
#if 0
    if (GTK_IS_FILE_CHOOSER(go)) {
	gulong id = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(go),"handler"));
	g_signal_handler_block(go, id);
	g_object_set_data_full(G_OBJECT(go), "filename", g_strdup(str), g_free);
	gtk_ami_file_chooser_set_filename(GTK_FILE_CHOOSER(go), str);
	g_signal_handler_unblock(go, id);
    } else 
#endif
    if (GTK_IS_LABEL(go)) {
        gtk_label_set_label(GTK_LABEL(go), str);
    } else if (GTK_IS_ENTRY(go)) {
	gulong id = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(go),"handler"));
	g_signal_handler_block(go, id);
        gtk_entry_set_text(GTK_ENTRY(go), EmptyNullStr(str));
	g_signal_handler_unblock(go, id);
    } else if (GTK_IS_BUTTON(go)) {
	GtkWidget *label = (GtkWidget*)g_object_get_data(G_OBJECT(go), "label");
        gtk_label_set_label(GTK_LABEL(label), str);
    }

    if (external) {
        GtkWidget *win = gtk_widget_get_toplevel(external);
        Project *pr = GuiWindowProject(win);
	gboolean fixed = gtk_ext_image_get_fixed(GTK_EXT_IMAGE(external));
	GtkGroup *group = 0;

	if (fixed) {
	    GtkWidget *child = gtk_bin_get_child(GTK_BIN(win));
	    group = GTK_GROUP(child);
	    gtk_group_change_init(group);
	}
	    
        ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
	gchar *path = gtkExpandPathname(str);
	GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(path, 0);
	g_free(path);

	g_object_set(G_OBJECT(external), "pixbuf", pixbuf, NULL);
	if (pixbuf) {
	    g_object_unref(G_OBJECT(pixbuf));
	}
	if (group) {
	    gtk_group_change_exit(group);
	}
    }
}

void GuiSetBoolean(GuiObject go, bool value, bool readonly)
{
    g_return_if_fail(GTK_IS_CHECK_BUTTON(go));

    gulong id = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(go),"handler"));
    g_signal_handler_block(go, id);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(go), value);
    if (readonly) {
        gtk_widget_set_sensitive(go, false);
    }
    g_signal_handler_unblock(go, id);
}

void GuiSetChoice(GuiObject go, unsigned short value, char **labels,
        bool readonly) 
{
    if (readonly) {
        g_return_if_fail(GTK_IS_LABEL(go));

	const gchar *text;
	if (value < NumLabels(labels)) {
	    text = labels[value];
	} else {
	    text = IntToStr(value);
	}
        gtk_label_set_label(GTK_LABEL(go), text);
    } else {
        g_return_if_fail(GTK_IS_COMBO_BOX(go) || GTK_IS_RADIO_BOX(go));

	gulong id = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(go),"handler"));
	g_signal_handler_block(go, id);
        if (GTK_IS_COMBO_BOX(go)) {
            gtk_combo_box_set_active(GTK_COMBO_BOX(go), value);
        } else {
            gtk_radio_box_set_active(GTK_RADIO_BOX(go), value);
        }
	g_signal_handler_unblock(go, id);
    }
}

void GuiSetMemo(GuiObject go, const char *str, bool readOnly)
{
    g_return_if_fail(GTK_IS_TEXT_VIEW(go));

    GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(go));

    gulong id = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(buffer),"handler"));
    g_signal_handler_block(buffer, id);
    GuiSetEditorText(go, str);
    GuiSetEditorChanged(go, FALSE);
    g_signal_handler_unblock(buffer, id);

    gtk_text_view_set_editable(GTK_TEXT_VIEW(go), !readOnly);
}

void GuiSetSFWList(GuiObject go, const struct Element *elem)
{
    g_return_if_fail(GTK_IS_SFW_VIEW(go));
    
    const char *error = gtk_sfw_view_set_element(GTK_SFW_VIEW(go), elem);
    if (error) {
	gtk_sfw_view_set_string(GTK_SFW_VIEW(go), error);
    } else if (DisplayFieldControlTable(GuiObjectDisplayObject(go))) {
        const QueryGuiData *data = GTK_SFW_VIEW(go)->data;
        Table *tab = data? data->qgd_ControlTable : 0;
        Record *rec = TableGuiRecord(tab);

        if (tab != NULL && rec != NULL) {
            gtk_sfw_view_set_active_record(GTK_SFW_VIEW(go), rec);
        }
    }
}

void BuildGuiAddRecord(Table *addtab, Record *rec)
{
    Project *pr = TableProject(addtab);
    List * tabs = ProjectTables(pr);

    if (!tabs || !rec) {
	return;
    }

    Table *tab;
    FORNODE(tab, tabs) {
	List *fields = TableFields(tab);

	if (!fields) {
	    continue;
	}

	Field *field;
	FORNODE(field, fields) {
	    if (FieldType(field) != TYPE_REFERENCE ||
		    FieldReferenceTable(field) != addtab) {
		continue;
	    }

	    DisplayObject *dio = FieldDisplayObject(field);
	    GuiObject go = DisplayFieldPopObject(dio);
	    BuildGuiPopRefFieldBrowserAddRecord(go, rec);
        }
    }
}

void BuildGuiRemRecord(Table *remtab, Record *rec)
{
    Project *pr = TableProject(remtab);
    List *tabs = ProjectTables(pr);

    if (!tabs || !rec) {
	return;
    }

    Table *tab;
    FORNODE(tab, tabs) {
	List *fields = TableFields(tab);

	if (!fields) {
	    continue;
	}

	Field *field;
	FORNODE(field, fields) {
	    short type = FieldType(field);
    	    DisplayObject *dio = FieldDisplayObject(field);
		    
	    if (type == TYPE_REFERENCE &&
		    FieldReferenceTable(field) == remtab) {
		GuiObject pop = DisplayFieldPopObject(dio);
		BuildGuiPopRefFieldBrowserRemRecord(pop, rec);
	    } else if (type == TYPE_VIRTUAL && 
    		    DisplayFieldSubType(dio) == DAD_VIRTUALTYPE_LIST) {
		GuiObject go = DisplayObjectGuiObject(dio);

		if (go) {
		    gtk_sfw_view_rem_record(GTK_SFW_VIEW(go), rec);
		}
	    }
        }
    }
}

void BuildGuiRemAllRecords(Project *pr)
{
    List *tabs = ProjectTables(pr);

    if (!tabs) {
	return;
    }

    Table *tab;
    FORNODE(tab, tabs) {
	List *fields = TableFields(tab);

	if (!fields) {
	    continue;
	}

	Field *field;
	FORNODE(field, fields) {
	    short type = FieldType(field);
    	    DisplayObject *dio = FieldDisplayObject(field);
		    
	    if (type == TYPE_REFERENCE) {
		GuiObject pop = DisplayFieldPopObject(dio);
                if (pop) {
                    gtk_pop_string_close_popup(GTK_POP_STRING(pop), false);
                }
	    } else if (type == TYPE_VIRTUAL && 
    		    DisplayFieldSubType(dio) == DAD_VIRTUALTYPE_LIST) {
		GuiObject go = DisplayObjectGuiObject(dio);

		if (go) {
		    gtk_sfw_view_rem_all_records(GTK_SFW_VIEW(go));
		}
	    }
        }
    }
}

void GuiUpdateProjects(void)
{
    if (anyListActiveChanged) {
        struct Project *pr;
        struct DisplayObject *dio;

        FORNODE(pr, &ProjectList) {
            dio = ProjectRootDisplayObject(pr);
            while (dio != NULL) {
                if (DisplayFieldListActiveChanged(dio)) {
                    GuiVirtualFieldListActiveChanged(dio);
                    SetDisplayFieldListActiveChanged(dio, FALSE);
                }
                dio = NextDisplayObject(dio);
            }
        }
        anyListActiveChanged = FALSE;
    }
}

