//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkBuildGuiRefFieldBrowser.cc
//
// Creation date: 04.12.2004
// Last modified: 29.01.2022
//

#include "gtk/gtkBuildGuiRefFieldBrowser.h"
#include "gtk/gtkBuildGuiWindow.h"
#include "gtk/gtkBuildGui.h"
#include "gtk/gtkQuickSearch.h"
#include "gtk/classes/gtkpopwidget.h"
#include "gtk/classes/gtksfwview.h"
#include "gtk/gtkimpl.h"
#include "gtk/gtkUtil.h"
#include "buildgui.h"
#include "programeditor.h"
#include "order.h"
#include "gui.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

#define CURRENT_RECORD (struct Record *)(~0)

static void refFieldBrowserCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    int col = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(column), "column"));
    bool multiColumn = (col >= 0);
    DisplayObject *dio = (DisplayObject *)data;
    Field *field = DisplayFieldField(dio);
    RecordDescription *rd;
    Record *rec = NULL;
    char *cont = NULL;

    rd = AllocRecordDescription(FieldReferenceTable(field),
            DisplayFieldRecordDescriptionSpec(dio, TRUE));
    if (rd != NULL) {
        gtk_tree_model_get(model, iter, 0, &rec, -1);
        RecordDescriptionSetRecord(rd, (rec == CURRENT_RECORD)? NULL : rec);
        if (rec == NULL || rec == CURRENT_RECORD) {
            cont = GCS(rec? MSG_CurrentRecord : MSG_InitialRecord);
            if (multiColumn) {
                long i, num = RecordDescriptionNumColumns(rd);
                for (i=0; i < num; i++) {
                    if (RecordDescriptionColumnType(rd, i) == TYPE_STRING) {
                        break;
                    }
                }
                if (i == num) {
                    i = 0;
                }
                if (i != col) {
                    cont = NULL;
                }
            }
            if (cont) {
                cont = g_strdup_printf("<i>%s</i>", cont);
                g_object_set(G_OBJECT(cell), "markup", cont, NULL);
                g_free(cont);
            } else {
                g_object_set(G_OBJECT(cell), "text", "", NULL);
            }
        } else {
            Project *pr = DisplayObjectProject(dio);
            MemPool *memPool = ProjectMemPool(pr);
            bool free = FALSE;
            long error = OK;
            short type = TYPE_UNKNOWN;

            RecordLock(rec);
            if (multiColumn) {
                cont = RecordDescriptionColumnText(rd, col, memPool, &free);
                type = RecordDescriptionColumnType(rd, col);
            } else {
                cont = RecordDescriptionText(rd, memPool, &free);
            }
            error = RecordDescriptionError(rd);
            const gchar *text = error? GES(error) : EmptyNullStr(cont);
            g_object_set(G_OBJECT(cell), "text", text, NULL);
            RecordUnlock(rec);
            if (free) {
                FreeString(memPool, cont);
            }
            if (type == TYPE_INTEGER || type == TYPE_REAL) {
                g_object_set(G_OBJECT(cell), "xalign", 1.0, NULL);
            }
        }
        FreeRecordDescription(rd);
    } else {
        g_object_set(G_OBJECT(cell), "text", "", NULL);
    }
}

static void setListStoreRecords(GtkListStore *store, Element *elem, Table *tab)
{
    if (elem) {
	while (elem->el_Type == TYPE_LIST) {
	    setListStoreRecords(store, elem->el_Contents.list.head, tab);
	    elem = elem->el_Contents.list.tail;
	}
	if (elem->el_Type == TYPE_RECORD && elem->el_Contents.tabrec.tab==tab) {
	    GtkTreeIter iter;
	    gtk_list_store_append(store, &iter);
    	    gtk_list_store_set(store, &iter,0, elem->el_Contents.tabrec.rec,-1);
	}
    }
}

static gboolean popRefFieldBrowserOnStart(GtkPopWidget *pop, DisplayObject *dio)
{
    g_signal_stop_emission_by_name(pop, "start");

    GtkWidget *window = gtk_pop_string_get_popup(GTK_POP_STRING(pop));
    GtkWidget *rab = gtk_pop_widget_get_widget(pop);
    GtkWidget *tree = (GtkWidget *)g_object_get_data(G_OBJECT(rab), "tree");
    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
    GtkListStore *store = GTK_LIST_STORE(model);
    GtkTreeIter iter;
    Field *field = DisplayFieldField(dio);
    Table *tab = FieldTable(field);
    Table *reftab = FieldReferenceTable(field);
    List *refrecs = TableRecords(reftab);
    Record *refrec;
    char records = DisplayFieldRecordsKind(dio);

    if (refrecs == NULL) {
        return FALSE;
    }

    GuiStoreEdits();      /* maybe the user has just edited a string object */

    gtk_list_store_clear(store);

    if (records == DAD_RECORDS_COMPUTED) {
	const char *trigger = DisplayFieldRecordsCompute(dio);

	if (trigger != NULL && trigger[0] != 0) {
	    Project *pr = TableProject(tab);
	    Program *prg = ProjectProgram(pr);
	    UserFunction *uf = FindUserFunctionName(prg, trigger);
	    const char *filename = NULL, *error = NULL;
	    long line, col;
	    struct Element *elem;

	    if (uf == NULL) {
		Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Trigger), trigger);
		return false;
    	    } else {
    		/* GuiSetSleep(TRUE); */
    		SetTableProgramRecords(pr);
    		ProgramBegin(prg, TRUE);
    		elem = ProgramCall(prg, uf, NULL, TRUE);
		setListStoreRecords(store, elem, reftab);
                ProgramEnd();
		error = ProgramError(&line, &col, &filename);
		/* GuiSetSleep(FALSE); */

		if (error != NULL && error != EXECUTE_ABORT) {
		    ProgramEditorShowError(pr, TRUE, error, line, col,filename);
		    return false;
		}
	    }
	}
    } else if (records == DAD_RECORDS_MATCHFILTER) {
	GuiSetSleep(true);
	FORNODE (refrec, refrecs) {
	    if (RecordMatchFilter(refrec)) {
		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter, 0, refrec, -1);
	    }
	}
	GuiSetSleep(false);
    } else {
	GuiSetSleep(true);
	GuiListSetList(tree, refrecs);
	GuiSetSleep(false);
    }

    if (DisplayFieldCurrentRecord(dio)) {
	refrec = CURRENT_RECORD;
	gtk_list_store_prepend(store, &iter);
    	gtk_list_store_set(store, &iter, 0, refrec, -1);
    }
    if (DisplayFieldInitialRecord(dio)) {
	refrec = NULL;
	gtk_list_store_prepend(store, &iter);
    	gtk_list_store_set(store, &iter, 0, refrec, -1);
    }

    Record *rec = TableGuiRecord(tab);
    struct DataInfo *di = NULL, di_buf;

    RecordLock(rec);
    RecordDataInfo(rec, field, &di, &di_buf);
    if(di) {
	GuiListSetActiveNode(tree, di->di_Contents.rec);
    }
    RecordUnlock(rec);

    gint width, height;
    gtk_window_get_size(GTK_WINDOW(window), &width, &height);
    height += 100;
    gtk_window_resize(GTK_WINDOW(window), width, height);
    gtk_widget_grab_focus(tree);

    return true;
}

static void popRefFieldBrowserOnStop(GtkPopWidget *pop, DisplayObject *dio)
{
    GtkWidget *rab = gtk_pop_widget_get_widget(pop);
    GtkWidget *tree = (GtkWidget *)g_object_get_data(G_OBJECT(rab), "tree");
    Record *active = (Record *)GuiListActiveEntry(tree);

    if (active == CURRENT_RECORD) {
	Field *field = DisplayFieldField(dio);
	Table *reftab = FieldReferenceTable(field);

	active = TableGuiRecord(reftab);
    }

    /* set record reference */
    GuiStoreDataRecord(dio, active);
}

static void popRefFieldBrowserOnRowActivated(GtkTreeView *tree,
	GtkTreePath *path, GtkTreeViewColumn *column, GtkPopString *pop)
{
    gtk_pop_string_close_popup(pop, TRUE);
}

#if 0
static void popRefFieldBrowserOnInitialPressed(GtkWidget *b, DisplayObject *dio)
{
    GtkWidget *pop = (GtkWidget *)g_object_get_data(G_OBJECT(b), "pop");
    GtkWidget *rab = gtk_pop_widget_get_widget(GTK_POP_WIDGET(pop));
    GtkWidget *tree = (GtkWidget *)g_object_get_data(G_OBJECT(rab), "tree");

    GuiSetListActive(tree, -1);

    /* close popup, this calls the close hook */
    gtk_pop_string_close_popup(GTK_POP_STRING(pop), TRUE);
}

static void popRefFieldBrowserOnCurrentPressed(GtkWidget *b, DisplayObject *dio)
{
    GtkWidget *pop = (GtkWidget *)g_object_get_data(G_OBJECT(b), "pop");
    GtkWidget *rab = gtk_pop_widget_get_widget(GTK_POP_WIDGET(pop));
    GtkWidget *tree = (GtkWidget *)g_object_get_data(G_OBJECT(rab), "tree");
    struct Field *field = DisplayFieldField(dio);
    struct Table *tab = FieldReferenceTable(field);
    struct Record *rec = TableGuiRecord(tab);

    /* activate record */
    GuiSetActiveNode(tree, rec);

    /* close popup, this calls the close hook */
    gtk_pop_string_close_popup(GTK_POP_STRING(pop), TRUE);
}
#endif

static Record * refFieldSearchCurrent = 0;

static void refFieldSearchFunc(GuiObject tree, const char *str, int cmd, void*ptr)
{
    DisplayObject *dio = (DisplayObject *)ptr;
    GuiObject pop = DisplayFieldPopObject(dio);
    long dir = -1;

    switch (cmd) {
	case QUICK_SEARCH_SETUP:
	    if (!gtk_widget_get_visible(tree)) {
		gtk_pop_string_open_popup(GTK_POP_STRING(pop));
	    }
	    break;

	case QUICK_SEARCH_NEXT:  dir++;		/* no break */
	case QUICK_SEARCH_START: dir++;		/* no break */
	case QUICK_SEARCH_PREV:
            refFieldSearchCurrent = GuiReferenceFieldQuickSearchRecord(
                    dio, tree, dir, str, refFieldSearchCurrent);
            break;

	case QUICK_SEARCH_COMMIT:
	    gtk_pop_string_close_popup(GTK_POP_STRING(pop), true);
            break;

	case QUICK_SEARCH_CANCEL:
	    break;
    }
}

gboolean onKeyPressEvent(GtkWidget *w, GdkEventKey *event, DisplayObject *dio)
{
    // quick search
    GuiObject pop = DisplayFieldPopObject(dio);
    GuiObject rab = gtk_pop_widget_get_widget(GTK_POP_WIDGET(pop));
    GuiObject tree = (GuiObject)g_object_get_data(G_OBJECT(rab), "tree");

    refFieldSearchCurrent = NULL;
    return GuiQuickSearchInit(tree, event, refFieldSearchFunc, dio);
}

static Record *dragDropRecord(DisplayObject *dio, GtkWidget *sfw)
{
    Field *field = DisplayFieldField(dio);
    Table *reftab = FieldReferenceTable(field);
    List *recs = TableRecords(reftab);
    const QueryLine *ql = (const QueryLine *)GuiListActiveEntry(sfw);
    const QueryGuiData *data= GTK_IS_SFW_VIEW(sfw)? GTK_SFW_VIEW(sfw)->data : 0;
    ushort numFields = data? data->qgd_NumFields : 0;
    ushort i;
    struct Record *rec;

    if (recs && ql && numFields) {
        ushort focus = gtk_sfw_view_get_active_column(GTK_SFW_VIEW(sfw));

        if (focus < numFields) {
            rec = ql->ql_Entries[focus]->qe_Record;
            if (FindNodePtr(recs, rec)) {
                return rec;
            }
        }
        for (i = 0; i < numFields; i++) {
            rec = ql->ql_Entries[i]->qe_Record;
            if (FindNodePtr(recs, rec)) {
                return rec;
            }
        }
    }

    return NULL;
}

static gboolean onDragMotion(GtkWidget *pop, GdkDragContext *context, 
        gint x, gint y, guint time, DisplayObject *dio)
{
    GtkWidget *source = gtk_drag_get_source_widget(context);
    Record *refrec = dragDropRecord(dio, source);

    if (refrec) {
        gtk_drag_highlight(pop);
        gdk_drag_status(context, GDK_ACTION_COPY, time);
    } else {
        gdk_drag_status (context, GDK_ACTION_DEFAULT, time);
    }
    return TRUE;
}

static void onDragLeave(GtkWidget *pop, GdkDragContext *context, 
        guint time, DisplayObject *dio)
{
    gtk_drag_unhighlight(pop);
}

static gboolean onDragDrop(GtkWidget *pop, GdkDragContext *context, 
        gint x, gint y, guint time, DisplayObject *dio)
{
    GtkWidget *source = gtk_drag_get_source_widget(context);
    Record *refrec = dragDropRecord(dio, source);
    gboolean success = false;

    if (refrec) {
        /* set record reference */
        GuiStoreDataRecord(dio, refrec);
    }
    gtk_drag_finish(context, success, FALSE, time);
    return TRUE;
}

GtkWidget *BuildGuiPopRefFieldBrowser(struct DisplayObject *dio, 
	GtkWidget *string, GtkWidget **button)
{
    *button = ButtonObject,
        "can-focus", DisplayFieldTabChain(dio),
	"child", ImageObject,
	    "icon-name", "gtk-index",
	    "icon-size", GTK_ICON_SIZE_MENU,
	    End,
	End;
    g_object_set_data(G_OBJECT(*button), "dio", dio);
    if (gtk_widget_get_can_focus(string)) {
	g_signal_connect(*button, "focus", G_CALLBACK(OnExtraButtonFocus), dio);
    }

    const RecordDescriptionSpec *rds = 
        DisplayFieldRecordDescriptionSpec(dio, TRUE);
    bool multiColumn = rds? rds->rds_MultiColumn : FALSE;
    long numColumns = -1;
    if (multiColumn) {
        Field *field = DisplayFieldField(dio);
        struct RecordDescription *rd = AllocRecordDescription(
                FieldReferenceTable(field), rds);
        if (rd) {
            RecordDescriptionSetRecord(rd, 0);
            numColumns = RecordDescriptionNumColumns(rd);
            FreeRecordDescription(rd);
        }
        if (numColumns <= 1) {
            multiColumn = false;
        }
    }

    GtkListStore *store = gtk_list_store_new(1, G_TYPE_POINTER);
    GtkWidget *tree = TreeViewObject,
        "model", store,
	"headers-visible", false,
	"fixed-height-mode", !multiColumn,
        "rules-hint", false,
	End;
    g_object_unref(G_OBJECT(store));

    GtkTreeSelection *selection = 
	gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
    
    GtkPolicyType policy;
    if (multiColumn) {
	for (int i=0; i < numColumns; i++) {
	    GtkCellRenderer *cell = gtk_cell_renderer_text_new();
	    GtkTreeViewColumn *column= gtk_tree_view_column_new_with_attributes(
		    "Item", cell, NULL);
	    g_object_set_data(G_OBJECT(column), "column", GINT_TO_POINTER(i));
	    gtk_tree_view_column_set_cell_data_func(column, cell, 
		    refFieldBrowserCellDataFunc, dio, NULL);
	    gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
	}
	policy = GTK_POLICY_AUTOMATIC;
    } else {
	GtkCellRenderer *cell = gtk_cell_renderer_text_new();
	GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
		"Item", cell, NULL);
	gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_FIXED);
	g_object_set_data(G_OBJECT(column), "column", GINT_TO_POINTER(-1));
	gtk_tree_view_column_set_cell_data_func(column, cell, 
		refFieldBrowserCellDataFunc, dio, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
	policy = GTK_POLICY_NEVER;
    }
#if 0
    GtkWidget *initial = GuiSimpleButton(GCS(MSG_Initial)); 
    GtkWidget *current = GuiSimpleButton(GCS(MSG_Current));
#endif
    GtkWidget *rab = VGroup,
        "spacing", 0,
	"child", ScrolledWindowObject,
	    "hscrollbar-policy", policy,
	    "shadow-type", GTK_SHADOW_IN,
	    "child", tree,
	    End,
#if 0
	"child", HGroup,
	    "spacing", 0,
	    "same-width", TRUE,
	    "child", initial,
	    "child", current,
	    End, "weight", 0,
#endif
	End;
    g_object_set_data(G_OBJECT(rab), "tree", tree);

    GtkWidget *pop = PopWidgetObject,
        //"light", false,
        "string", string,
	"button", *button,
	"widget", rab,
	End;
    g_object_set_data(G_OBJECT(pop), "dio", dio);
    gtk_drag_dest_set(pop, (GtkDestDefaults)0, &gtk_sfw_view_target, 1, 
            GDK_ACTION_COPY);
    g_signal_connect(pop, "drag-motion", G_CALLBACK(onDragMotion), dio);
    g_signal_connect(pop, "drag-leave", G_CALLBACK(onDragLeave), dio);
    g_signal_connect(pop, "drag-drop", G_CALLBACK(onDragDrop), dio);
    g_signal_connect(pop, "start", G_CALLBACK(popRefFieldBrowserOnStart), dio);
    g_signal_connect(pop, "stop", G_CALLBACK(popRefFieldBrowserOnStop), dio);
    g_signal_connect_data(pop, "close", G_CALLBACK(gtk_list_store_clear),
	    store, NULL, (GConnectFlags)(G_CONNECT_AFTER | G_CONNECT_SWAPPED));
    g_signal_connect(tree, "row-activated", 
	    G_CALLBACK(popRefFieldBrowserOnRowActivated), pop);
    g_object_set_data(G_OBJECT(tree), "pop", pop);
#if 0
    g_object_set_data(G_OBJECT(initial), "pop", pop);
    g_object_set_data(G_OBJECT(current), "pop", pop);
    g_signal_connect(initial, "clicked", 
	    G_CALLBACK(popRefFieldBrowserOnInitialPressed), dio);
    g_signal_connect(current, "clicked", 
	    G_CALLBACK(popRefFieldBrowserOnCurrentPressed), dio);
#endif

    g_signal_connect(string, "key-press-event",G_CALLBACK(onKeyPressEvent),dio);
    g_signal_connect(*button,"key-press-event",G_CALLBACK(onKeyPressEvent),dio);
    g_signal_connect(tree, "key-press-event", G_CALLBACK(onKeyPressEvent), dio);

    return pop;
}

void BuildGuiPopRefFieldBrowserAddRecord(GuiObject pop, struct Record *rec)
{
    if (pop && gtk_pop_string_get_popup(GTK_POP_STRING(pop)) && rec) {
	DisplayObject *dio = GuiObjectDisplayObject(pop);
	char records = DisplayFieldRecordsKind(dio);
	bool filter = (records == DAD_RECORDS_MATCHFILTER);
	if (records == DAD_RECORDS_ALL || (filter && RecordMatchFilter(rec))) {
	    GtkWidget *rab = gtk_pop_widget_get_widget(GTK_POP_WIDGET(pop));
	    GtkWidget *tree = (GtkWidget *)
		g_object_get_data(G_OBJECT(rab), "tree");
	    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
	    Record *succ = NextRecord(rec, filter);
	    long pos = GuiListFindNode(tree, succ);
	    GtkTreeIter iter;

	    if (pos < 0) {
		pos = G_MAXINT;		// append at bottom
	    }
	    gtk_list_store_insert(GTK_LIST_STORE(model), &iter, pos);
	    gtk_list_store_set(GTK_LIST_STORE(model), &iter, 0, rec, -1);
	}
    }
}

void BuildGuiPopRefFieldBrowserRemRecord(GuiObject pop, struct Record *rec)
{
    if (pop && gtk_pop_string_get_popup(GTK_POP_STRING(pop)) && rec) {
	GtkWidget *rab = gtk_pop_widget_get_widget(GTK_POP_WIDGET(pop));
	GtkWidget *tree = (GtkWidget *)g_object_get_data(G_OBJECT(rab), "tree");

        GuiListRemoveEntry(tree, rec);
    }
}

//
// todo: change selection when user moves to new record (also in MUI version)
// update display when user edits a referenced record (also in MUI version)
//

