//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkBuildGuiTable.cc
//
// Creation date: 04.04.2004
// Last modified: 29.03.2021
//

#include "gtk/gtkBuildGuiTable.h"
#include "gtk/gtkBuildGui.h"
#include "gtk/gtkMenus.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "buildgui.h"
#include "gui.h"
#include "display.h"
#include "table.h"
#include "filter.h"

//#define USE_STOCK
#define STOCK_SIZE GTK_ICON_SIZE_SMALL_TOOLBAR

static void onFilterToggled(GtkWidget *filterObject, Table *tab)
{
    bool on = gtkToggleButtonActive(filterObject);

    GuiStoreEdits();
    if (on && TableFilterProgram(tab) == NULL) {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(filterObject), FALSE);
	GuiChangeFilter(tab);
    } else {
	FilterOnOff(tab, on);
    }
}

#define TIME_DELAY     50  /* reapeat at this rate */
#define INITIAL_SKIP    5  /* but skip the first x events */

static struct {
    guint id;
    void (*prevNext)(Table *tab);
    Table *tab;
    short skip;
    bool keepCalling;
} timer;

static void clearTimer()
{
    if (timer.id) {
	g_source_remove(timer.id);
    }
    memset(&timer, 0, sizeof(timer));
}

static gboolean timerCallback(gpointer data)
{
    // printf("depth : %d\n", g_main_depth());
    if (timer.skip++ >= INITIAL_SKIP && g_main_depth() <= 1) {
	GtkWidget *button = GTK_WIDGET(data);
        bool active;
#if GTK_MAJOR_VERSION == 2
	active = gtk_widget_get_state(button) == GTK_STATE_ACTIVE;
#else
	active =(gtk_widget_get_state_flags(button) & GTK_STATE_FLAG_ACTIVE)!=0;
#endif
	if (active) {
	    timer.prevNext(timer.tab);
	}
	timer.skip = INITIAL_SKIP;	// avoid wrap-around/overflow
    }

    return timer.keepCalling;
}

static void addTimer(GtkWidget *button, void (*func)(Table *), Table *tab)
{
    clearTimer();

    timer.prevNext = func;
    timer.tab = tab; 
    timer.keepCalling = true;
    timer.id = g_timeout_add_full(G_PRIORITY_DEFAULT_IDLE, TIME_DELAY, 
	    timerCallback, button, 0);

    func(tab);
}

static void onLeftPressed(GtkWidget *button, Table *tab)
{
    GuiStoreEdits();
    GuiBlockBusyPointer(gtk_widget_get_toplevel(button), true);
    addTimer(button, GuiPrevRecord, tab);
}

static void onLeftReleased(GtkWidget *button, Table *tab)
{
    clearTimer();
    GuiBlockBusyPointer(gtk_widget_get_toplevel(button), false);
}

static void onRightPressed(GtkWidget *button, Table *tab)
{
    GuiStoreEdits();
    GuiBlockBusyPointer(gtk_widget_get_toplevel(button), true);
    addTimer(button, GuiNextRecord, tab);
}

static void onRightReleased(GtkWidget *button, Table *tab)
{
    clearTimer();
    GuiBlockBusyPointer(gtk_widget_get_toplevel(button), false);
}

static gboolean onPanelFocus(GtkWidget *widget, GtkDirectionType type)
{
    g_signal_stop_emission_by_name(widget, "focus");

    if (type == GTK_DIR_TAB_BACKWARD && gtk_widget_has_focus(widget)) {
	return false;
    }

    if (type == GTK_DIR_TAB_FORWARD && !gtk_widget_has_focus(widget)) {
	GtkWidget *top = gtk_widget_get_toplevel(widget);
	if (GTK_IS_WINDOW(top)) {
	    GtkWidget *focus = gtk_window_get_focus(GTK_WINDOW(top));
	    if (!focus || !gtk_widget_is_ancestor(focus, widget)) {
		gtk_widget_grab_focus(widget);
		return true;
	    }
	}
    }
    gboolean result = gtk_widget_child_focus(
            gtk_bin_get_child(GTK_BIN(widget)), type);

    if (type == GTK_DIR_TAB_BACKWARD && !result) {
	gtk_widget_grab_focus(widget);
	return true;
    }

    return result;
}

#if GTK_MAJOR_VERSION == 2
static gboolean onPanelExposeEvent(GtkWidget *widget, GdkEventExpose *event)
{
    GtkStyle *style = gtk_widget_get_style(widget);
    GtkAllocation alloc;
    gtk_widget_get_allocation(widget, &alloc);
    gint x = 0;
    gint y = 0;
    gint width = alloc.width;
    gint height = alloc.height;
    GtkStateType state = GTK_STATE_NORMAL;

    if (gtk_widget_has_focus(widget)) {
	gtk_paint_focus(style, gtk_widget_get_window(widget), state,
		&event->area, widget, 0, x, y, width, height);
    }

    return false;
}
#else
static gboolean onPanelDraw(GtkWidget *widget, cairo_t *cairo)
{
    GtkStyleContext *context = gtk_widget_get_style_context(widget);
    int width = gtk_widget_get_allocated_width(widget);
    int height = gtk_widget_get_allocated_height(widget);

    if (gtk_widget_has_focus(widget)) {
        gtk_render_focus(context, cairo, 0, 0, width, height);
    }
    return false;
}
#endif

static gboolean onPanelButtonPressEvent(GtkWidget *widget,GdkEventButton *event,
        GtkWidget *title)
{
    if (event->button != 1 && event->button != 3) {
        return false;               // only left and right mouse buttons
    }

    DisplayObject *dio = GuiObjectDisplayObject(widget);
    Table *tab = DisplayObjectTable(dio);
    if (TableExactlyOneRecord(tab)) {
        return false;               // no menus for exactly-one-record tables
    }

    // check whether the current focus is set on an element of the table
    GtkWidget *focus = GuiActiveObject(gtk_widget_get_toplevel(widget));
    if (!GuiObjectDisplayObject(focus) || 
            !gtk_widget_is_ancestor(focus, widget)) {
        gtk_widget_grab_focus(widget);
    }
    if (event->button == 3 || title) {
	PopupTableMenu(widget);
    }
    return true;
}

#if GTK_MAJOR_VERSION == 2
static gboolean onRoundExposeEvent(GtkWidget *widget, GdkEventExpose *event,
        GtkWidget *sample)
{
    GdkWindow *window = gtk_widget_get_window(widget);
    gint border = gtk_container_get_border_width(GTK_CONTAINER(widget));
    GtkAllocation alloc;
    gtk_widget_get_allocation(widget, &alloc);
    gint x = alloc.x + border;
    gint y = alloc.y + border;
    gint width = alloc.width - 2*border;
    gint height = alloc.height - 2*border;
    GtkStyle *style = gtk_widget_get_style(widget);
    GtkStyle *sstyle = gtk_widget_get_style(sample);
#if 0
    /* doesn't work on FC5 gtk+-2.8 */
    GtkStyle *newstyle = gtk_style_copy(sstyle), *attached;
    newstyle->bg[GTK_STATE_NORMAL] = style->bg[GTK_STATE_NORMAL];
    attached = gtk_style_attach(newstyle, window);
#endif
    style->xthickness = sstyle->xthickness;
    style->ythickness = sstyle->ythickness;

    gtk_paint_box(style, window, GTK_STATE_NORMAL,
            GTK_SHADOW_OUT, &event->area, sample, "button",
            x, y, width, height);

#if 0
    gtk_style_detach(attached);
    g_object_unref(G_OBJECT(newstyle));
#endif

    return false;
}
#else
static gboolean onRoundDraw(GtkWidget *widget, cairo_t *cairo, 
        GtkWidget *sample)
{
    GtkStyleContext *context = gtk_widget_get_style_context(sample);
    gint border = gtk_container_get_border_width(GTK_CONTAINER(widget));
    GtkAllocation alloc;
    gtk_widget_get_allocation(widget, &alloc);
    gint x = border;
    gint y = border;
    gint width = alloc.width - 2*border;
    gint height = alloc.height - 2*border;

    gtk_render_background(context, cairo, x, y, width, height);
    gtk_render_frame(context, cairo, x, y, width, height);
    return false;
}
#endif

static GtkGroup *buildGuiPanel(DisplayObject *dio, GtkGroup *parent)
{
    const char *ctitle = DisplayObjectTitle(dio);
    Table *tab = DisplayObjectTable(dio);
    const char *bg = DisplayTablePanelBackground(dio);
    bool gadgets = !TableExactlyOneRecord(tab);
    bool numall = DisplayTablePanelNumAll(dio) && gadgets,
	filter = DisplayTablePanelFilter(dio) && gadgets,
	arrows = DisplayTablePanelArrows(dio) && gadgets;
    GuiObject go = 0, panel = 0, numall_go = 0, buttons = 0, filter_go = 0,
	left_go = 0, right_go = 0, newparent = 0;

    numall_go = LabelObject,
	"label", numall? "   #0/0" : "",
	"xalign", 1.0,
   	End;
    if (filter || arrows) {
        buttons = HGroup,
            "spacing", 0,
            "same-size", true,
            End;
    }
    if (filter) {
        filter_go = ToggleButtonObject, 
		"can-focus", false,
		"label", " F ",
		End;
	g_signal_connect(filter_go, "toggled", G_CALLBACK(onFilterToggled),tab);
        gtk_group_add_child(GTK_GROUP(buttons), filter_go, 1, 1);
    }

    if (arrows) {
        left_go = ButtonObject,
	    "can-focus", false,
#ifdef USE_STOCK
	    "child", ImageObject,
		"icon-size", STOCK_SIZE,
	        "stock", GTK_STOCK_GO_BACK,
		End,
#else
	    "child", VGroup, 			// group avoids bug in GTK 2.6
	        "child", ArrowObject,
		    "arrow-type", GTK_ARROW_LEFT,
		    End, 
		End,
#endif
	    End;
	g_signal_connect(left_go, "pressed", G_CALLBACK(onLeftPressed), tab);
	g_signal_connect(left_go, "released", G_CALLBACK(onLeftReleased), tab);
        gtk_group_add_child(GTK_GROUP(buttons), left_go, 1, 1);

        right_go = ButtonObject,
	    "can-focus", false,
#ifdef USE_STOCK
	    "child", ImageObject,
		"icon-size", STOCK_SIZE,
	        "stock", GTK_STOCK_GO_FORWARD,
		End,
#else
	    "child", VGroup, 			// group avoids bug in GTK 2.6
	        "child", ArrowObject,
		    "arrow-type", GTK_ARROW_RIGHT,
		    End, 
		End,
#endif
            End;
	g_signal_connect(right_go, "pressed", G_CALLBACK(onRightPressed), tab);
	g_signal_connect(right_go, "released", G_CALLBACK(onRightReleased),tab);
        gtk_group_add_child(GTK_GROUP(buttons), right_go, 1, 1);
    }

    GuiObject label_go = LabelObject,
	"label", ctitle,
	End;

    GuiObject titlenum_go = ButtonObject, 
              // FrameObject, "shadow-type", GTK_SHADOW_OUT,
        "can-focus", false,
        "focus-on-click", false,
        "child", AlignmentObject,
            "left-padding", GTK_GROUP_DEFAULT_HORIZ_SPACING,
            "right-padding", GTK_GROUP_DEFAULT_HORIZ_SPACING,
            "child", HGroup,
                "child", label_go, "weight", 0,
                "child", numall_go,
                End, 
            End,
        End;
    g_signal_connect(titlenum_go, "enter-notify-event", G_CALLBACK(gtk_true),0);

    if (buttons) {
        panel = HGroup,
            "horiz-spacing", 0,
            "child", titlenum_go,
            "child", buttons, "weight", 0,
            End;
    } else {
        panel = titlenum_go;
    }

    GtkWidget *round;
    go = EventBoxObject,
        "can-focus", true,
        "child", VGroup,
	    "border-width", 1,	// for focus
	    "spacing", 0,
	    "child", panel, "weight", 0,
	    "child", round = FrameObject,
                "shadow-type", GTK_SHADOW_NONE,
		"child", GroupAlignment(false),
		    "child", newparent = VGroup, End,
		    End,
		End,
	    End,
	End;
    g_signal_connect(go, "focus", G_CALLBACK(onPanelFocus), 0);
#if GTK_MAJOR_VERSION == 2
    g_signal_connect(round, "expose-event", 
            G_CALLBACK(onRoundExposeEvent), titlenum_go);
    g_signal_connect_after(go, "expose-event", 
	    G_CALLBACK(onPanelExposeEvent), 0);
#else
    g_signal_connect(round, "draw", G_CALLBACK(onRoundDraw), titlenum_go);
    g_signal_connect_after(go, "draw", G_CALLBACK(onPanelDraw), 0);
#endif
    g_signal_connect(go, "button-press-event", 
	    G_CALLBACK(onPanelButtonPressEvent), 0);
    if (titlenum_go) {
        g_signal_connect_swapped(titlenum_go, "button-press-event", 
                G_CALLBACK(onPanelButtonPressEvent), go);
    }

    GuiSetHelpNode(go, "Browsing records");
    g_object_set_data(G_OBJECT(go), "dio", dio);
    BuildGuiAddCommon(dio, panel);
    if (bg == NULL || bg[0] == 0) {
        bg = DisplayObjectBackground(dio);
    }
    BuildGuiSetBackgroundSpec(dio, bg, titlenum_go);
    BuildGuiSetBackgroundSpec(dio, bg, filter_go);
    BuildGuiSetBackgroundSpec(dio, bg, left_go);
    BuildGuiSetBackgroundSpec(dio, bg, right_go);
    BuildGuiSetFont(dio, label_go);
    BuildGuiSetFont(dio, numall_go);
    BuildGuiSetBackground(dio, round);

    gtk_group_add_child(parent, go, 1, 1);

    SetDisplayObjectGuiObject(dio, go);
    SetDisplayObjectRootObject(dio, go);
    SetDisplayObjectActivateObject(dio, go);
    SetDisplayTablePanelNumAllObject(dio, numall_go);
    SetDisplayTablePanelFilterObject(dio, filter_go);

    return GTK_GROUP(newparent);
}

bool BuildGuiTable(DisplayObject *dio, GtkGroup *parent)
{
    if (!parent) {
	return false;
    }

    GtkWidget *widget = VGroup, End;
    GtkGroup *group = GTK_GROUP(widget);
    SetDisplayObjectGuiObject(dio, widget);
    SetDisplayObjectRootObject(dio, widget);

    if (DisplayTableHasPanel(dio)) {
        group = buildGuiPanel(dio, group);
        if (!group) {
            return false;
        }
    } else {
        BuildGuiSetBackground(dio, widget);
    }

    bool rc = true;
    List *childs = DisplayObjectObjects(dio);
    if (childs) {
        GtkGroup *colGroup = 0;
        rc = BuildGuiChilds(childs, false, group, &colGroup);
    }

    int weight = DisplayObjectWeight(dio);
    int horizWeight = weight, vertWeight = weight;
    if (rc) {
        if (gtk_group_get_total_horiz_weight(group) == 0) {
            horizWeight = 0;
        }
        if (gtk_group_get_total_vert_weight(group) == 0) {
            vertWeight = 0;
        }
    }
    gtk_group_add_child(parent, widget, horizWeight, vertWeight);

    return rc;
}

void GuiSetNumAll(GuiObject go, ulong num, ulong all)
{
    if (go) {
	char buffer[31];

	sprintf(buffer, "#%ld/%ld", num, all);
	g_object_set(G_OBJECT(go), "label", buffer, NULL);
    }
}

