//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkBuildGuiWindow.cc
//
// Creation date: 23.03.2004
// Last modified: 03.04.2022
//

#include "system.h"
#include "gtk/gtkBuildGuiWindow.h"
#include "gtk/gtkBuildGui.h"
#include "gtk/gtkMenus.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "gtk/classes/gtkextimage.h"
#include "menuproject.h"
#include "structureeditor.h"
#include "buildgui.h"
#include "gui.h"
#include "display.h"
#include "appstrings_cat.h"

static gboolean onDraw(GtkWidget *widget, cairo_t *cairo)
{
    GdkWindow *window = gtk_widget_get_window(widget);
    gint w = gdk_window_get_width(window);
    gint h = gdk_window_get_height(window);

    if (w > 0 && h > 0) {
	PangoLayout *pl = gtk_widget_create_pango_layout(widget, GCS(MSG_APP));
	int xstep = 140, ystep = 180;
	PangoAttrList *pal = pango_attr_list_new();
	PangoAttribute *pa;

	pa = pango_attr_size_new(18 * PANGO_SCALE);
	pa->start_index = 0;
	pa->end_index = (guint)-1;
	pango_attr_list_insert(pal, pa);
	pa = pango_attr_foreground_new(0xd000, 0xd000, 0xd000);
	pa->start_index = 0;
	pa->end_index = (guint)-1;
	pango_attr_list_insert(pal, pa);
#if GTK_MAJOR_VERSION == 2
	pa = gdk_pango_attr_embossed_new(true);
	pa->start_index = 0;
	pa->end_index = (guint)-1;
	pango_attr_list_insert(pal, pa);
#endif
#if 0	/* does not seem to have any effect, requires Pango 2.12 */
	GdkColor color = { 0, 0x8000, 0x8000, 0x8000 };
	pa = gdk_pango_attr_emboss_color_new(&color);
	pa->start_index = 0;
	pa->end_index = (guint)-1;
	pango_attr_list_insert(pal, pa);
#endif

	pango_layout_set_attributes(pl, pal);
	pango_attr_list_unref(pal);

	for (int y = 10; y < h; y += ystep) {
	    int i = 0;
	    for (int x = 0; x < w; x += xstep, i = !i) {
                cairo_move_to(cairo, x + 20, y + 20 + i*ystep/2);
                pango_cairo_show_layout(cairo, pl);
	    }
	}
	g_object_unref(G_OBJECT(pl));
    }

    return true;
}

#if GTK_MAJOR_VERSION == 2
static gboolean onExposeEvent(GtkWidget *widget, GdkEventExpose *event)
{
    cairo_t *cairo = gdk_cairo_create(gtk_widget_get_window(widget));

    onDraw(widget, cairo);
    cairo_destroy(cairo);

    return true;
}
#else
static gboolean onDrawImage(GtkWidget *widget, cairo_t *cairo,GdkPixbuf *pixbuf)
{
    int w = gtk_widget_get_allocated_width(widget);
    int h = gtk_widget_get_allocated_height(widget);
    int iw = gdk_pixbuf_get_width(pixbuf);
    int ih = gdk_pixbuf_get_height(pixbuf);
    int x = (w - iw) / 2;
    int y = (h - ih) / 2;

    cairo_rectangle(cairo, 0.0, 0.0, w, h);
    cairo_set_source_rgba(cairo,
		    (double)MAIN_WINDOW_LOGO_BG_RED / 0xFF,
		    (double)MAIN_WINDOW_LOGO_BG_GREEN / 0xFF,
		    (double)MAIN_WINDOW_LOGO_BG_BLUE / 0xFF,
		    1.0);
    cairo_fill(cairo);
    gdk_cairo_set_source_pixbuf(cairo, pixbuf, x, y);
    cairo_paint(cairo);
    return true;
}
#endif

static GtkWidget *empty_display_new(gboolean addHelp)
{
    GtkWidget *go;
    gchar *path = gtkExpandPathname(MAIN_WINDOW_LOGO_IMAGE);
    GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(path, 0);
    g_free(path);
    if (pixbuf) {
#if GTK_MAJOR_VERSION == 2
        go = EventBoxObject,
            "child", ImageObject,
                "pixbuf", pixbuf,
                "width-request", MAIN_WINDOW_LOGO_MIN_WIDTH,
                "height-request", MAIN_WINDOW_LOGO_MIN_HEIGHT,
                End,
            End;
        g_object_unref(G_OBJECT(pixbuf));
#else
        go = DrawingAreaObject,
            "width-request", MAIN_WINDOW_LOGO_MIN_WIDTH,
            "height-request", MAIN_WINDOW_LOGO_MIN_HEIGHT,
            End;
        g_object_set_data_full(G_OBJECT(go), "pixbuf", pixbuf, g_object_unref);
        g_signal_connect(go, "draw", G_CALLBACK(onDrawImage), pixbuf);
#endif
#if GTK_MAJOR_VERSION == 2
        GdkColor color;
        color.red = MAIN_WINDOW_LOGO_BG_RED << 8;
        color.green = MAIN_WINDOW_LOGO_BG_GREEN << 8;
        color.blue = MAIN_WINDOW_LOGO_BG_BLUE << 8;
        gtk_widget_modify_bg(go, GTK_STATE_NORMAL, &color);
#endif
    } else {
        go = EventBoxObject, 
           "child", SpaceObject, End, 
           End;
#if GTK_MAJOR_VERSION == 2
        g_signal_connect(go, "expose-event", G_CALLBACK(onExposeEvent), 0);
#else
        g_signal_connect(go, "draw", G_CALLBACK(onDraw), 0);
#endif
    }
#ifndef BEEBASE_LIGHT
    if (addHelp) {
	go = GuiMakeTooltip(go, GCS(MSG_EmptyDisplayBH));
    }
#endif

    return go;
}

static gboolean onDelete(GtkWidget *win)
{
    struct DisplayObject *dio = GuiObjectDisplayObject(win);
    struct Project *pr = DisplayObjectProject(dio);

    GuiStoreEdits();
    if (dio == ProjectRootDisplayObject(pr)) {
        MenuProjectClose(pr);
    } else {
        gtk_widget_hide(win);
    }
    return true;
}

static gboolean onWindowState(GtkWidget *win, GdkEventWindowState *event)
{
    if (event->type != GDK_WINDOW_STATE) {
	return false;
    }
    if (!(event->changed_mask & GDK_WINDOW_STATE_ICONIFIED)) {
	return false;
    }

    Project *pr = GuiWindowProject(win);
    DisplayObject *dio = ProjectRootDisplayObject(pr);
    GuiObject main = DisplayObjectGuiObject(dio);

    if (!main || win != main) {
	return false;
    }

    GList *toplevels = gtk_window_list_toplevels();
    if (event->new_window_state & GDK_WINDOW_STATE_ICONIFIED) {
#ifdef MACOSX
	// There seems to be a bug in GTK for Mac OS X when iconifying
	// the main window of a project with multiple windows
	// The following is a hack to work around this problem
	gtk_widget_hide(main);
#endif
	for (GList *list = toplevels; list; list = list->next) {
	    GtkWidget *top = (GtkWidget *)list->data;
	    if (top != main && GuiWindowProject(top) == pr) {
		gboolean visible = 
		    gtk_widget_get_visible(top) || (top == win);
		g_object_set_data(G_OBJECT(top), "open",
			GINT_TO_POINTER(visible));
		gtk_widget_set_visible(top, false);
	    }
	}
    } else {
	for (GList *list = toplevels; list; list = list->next) {
	    GtkWidget *top = (GtkWidget *)list->data;
	    if (top != main && GuiWindowProject(top) == pr) {
		if (GPOINTER_TO_INT(g_object_get_data(
				G_OBJECT(top), "open"))) {
		    gtk_widget_set_visible(top, true);
		}
	    }
	}
#ifdef MACOSX
	gtk_window_deiconify(GTK_WINDOW(main));
	gtk_widget_show(main);
#endif
    }
    g_list_free(toplevels);

    return false;
}

static gboolean onIdleStoreEdits()
{
    GuiStoreEdits();
    return false;
}

void onSetFocus(GtkWindow *win, GtkWidget *focus)
{
    if (focus) {
	g_idle_add((GSourceFunc)onIdleStoreEdits, 0);
    }
}

static gboolean onConfigure(GtkWidget *win, GdkEventConfigure *event, 
        DisplayObject *dio)
{
    WindowDimensions wd;

    if (GuiGetWindowDimensions(win, &wd)) {
	SetDisplayWindowDimensions(dio, &wd);
    }
    return false;
}

static void onShow(GtkWidget *win, DisplayObject *dio)
{
    // ensure window position is correct
    GuiSetWindowDimensions(win, DisplayWindowDimensions(dio));
    GuiStoreEdits();
    GuiUpdateWindow(win, DisplayObjectProject(dio));
    SetDisplayWindowOpen(dio, TRUE);
}

static void onHide(GtkWidget *win, DisplayObject *dio)
{
    GuiStoreEdits();
    GuiClearWindow(win, DisplayObjectProject(dio));
    SetDisplayWindowOpen(dio, FALSE);
}

GtkWidget *buildGuiWindowButton(DisplayObject *dio)
{
    gchar *title = g_strdup_printf("%s...", DisplayObjectTitle(dio));
    GtkWidget *button = BuildGuiIconLabelButton(0,
	    title, DisplayObjectShortcut(dio), 0.5);

    g_free(title);
    g_signal_connect(button, "focus", G_CALLBACK(OnExtraButtonFocus), dio);
    g_object_set_data(G_OBJECT(button), "dio", dio);
    SetDisplayWindowButtonObject(dio, button);
    GtkWidget *label = (GtkWidget*)g_object_get_data(G_OBJECT(button),"label");
    BuildGuiSetFont(dio, label);

    return button;
}

bool BuildGuiWindowButton(DisplayObject *dio, GtkGroup *parent, 
        GtkGroup **colGroup)
{
    if (!DisplayWindowButton(dio)) {
        return true;
    }

    if (parent == 0) {
        return false;
    }

    BuildGuiFinishColGroup(colGroup);

    GtkWidget *button = buildGuiWindowButton(dio);
    GtkWidget *go = AlignmentObject, 
        "yscale", 0.0,
        "child", button, 
        End;
    gtk_group_add_child(parent, go, DisplayObjectWeight(dio), 0);
    BuildGuiAddCommon(dio, button);
    BuildGuiSetBackground(dio, button);
    BuildGuiSetFont(dio, button);

    SetDisplayWindowButtonObject(dio, button);
    SetDisplayObjectRootObject(dio, go);

    return true;
}

bool BuildGuiWindow(DisplayObject *dio)
{
    Project *pr = DisplayObjectProject(dio);
    bool main = (dio == ProjectRootDisplayObject(pr));
    bool empty = !DisplayObjectVisibleGui(dio);
    const char *title = DisplayObjectTitle(dio);
    const WindowDimensions *wd = DisplayWindowDimensions(dio);
    GtkGroup *vgroup = 0;
//    static GtkWindowGroup *windowGroup = 0;

    GtkWidget *root = VGroup, "spacing", 0, End;
    GtkWidget *win = WindowObject,
        "title", title,
	"child", root,
	End;
    GuiSetWindowDimensions(win, wd);
    g_object_set_data(G_OBJECT(win), "dio", dio);
    GuiSetWindowProject(win, pr);
    SetDisplayObjectGuiObject(dio, win);
    vgroup = GTK_GROUP(root);
    GuiWindowAddMenus(win, root, main);

    if (main) {
//        windowGroup = gtk_window_group_new();
//        gtk_window_group_add_window(windowGroup, GTK_WINDOW(win));
//        g_object_unref(G_OBJECT(windowGroup));

        g_signal_connect(win, "window-state-event", 
		G_CALLBACK(onWindowState), 0);
	SetRootWindowTitle(pr, win);
    } else {
//        gtk_window_group_add_window(windowGroup, GTK_WINDOW(win));

        GtkWidget *button = DisplayWindowButtonObject(dio);
        if (button) {
            g_signal_connect_swapped(button, "clicked", 
		    G_CALLBACK(gtk_window_present), win);
        }
    }
    // g_object_set_data(G_OBJECT(win), "accel-group", accelGroup);
    g_signal_connect(win, "delete-event", G_CALLBACK(onDelete), 0);
    g_signal_connect(win, "configure-event", G_CALLBACK(onConfigure), dio);
    g_signal_connect(win, "set-focus", G_CALLBACK(onSetFocus), 0);
    g_signal_connect(win, "show", G_CALLBACK(onShow), dio);
    g_signal_connect(win, "hide", G_CALLBACK(onHide), dio);
    GuiAddWindowCallbacks(win, NULL);

    bool rc;
    if (empty) {
	gtk_group_add_child(vgroup, empty_display_new(main), 100, 100);
        gtk_widget_show_all(root);
	g_object_set(G_OBJECT(win),
	    "default-width", MAIN_WINDOW_LOGO_DEFAULT_WIDTH,
	    "default-height", MAIN_WINDOW_LOGO_DEFAULT_HEIGHT,
	    NULL);
        rc = true;
    } else {
	GtkWidget *widget = VGroup, End;
	GtkGroup *subGroup = GTK_GROUP(widget);
	widget = GroupAlignment(false),
	   "child", widget,
	    End;
    
	List *childs = DisplayObjectObjects(dio);
	GtkGroup *colGroup = 0;
	rc = BuildGuiChilds(childs, false, subGroup, &colGroup);

	gtk_group_add_child(vgroup, widget,
		gtk_group_get_total_horiz_weight(subGroup),
		gtk_group_get_total_vert_weight(subGroup));

	gtk_widget_show_all(root);
    }
    return rc;
}

