//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkEditDisplay.cc
//
// Creation date: 07.05.2005
// Last modified: 27.03.2016
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "gtk/gtkEditDisplay.h"
#include "gtk/gtkStructureEditor.h"
#include "gtk/classes/gtkgroup.h"
#include "gtk/classes/gtkeditlabels.h"
#include "gtk/classes/gtkpopimage.h"
#include "gtk/classes/gtkpopselection.h"
#include "gtk/classes/gtkradiobox.h"
#include "gtk/gtkimpl.h"
#include "gtk/gtkUtil.h"
#include "display.h"
#include "requesters.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"

#define IMAGESPEC_START "0:128"

void ShortCutOnActivate(GtkWidget *entry, DisplayObject *dio)
{
    cschar shortcut = CodesetGetCharCheck(gtkEntryText(entry));
    DisplayObject *inuse = DisplayObjectShortcutInUse(shortcut, 
	    DisplayObjectWindow(dio), dio);
    if (inuse) {
	GuiObject win = gtk_widget_get_toplevel(entry);
        const char *name = DisplayObjectName(inuse);
	Ask2(win, 0, AT_ERROR, GCS(MSG_StarOk),
		GCS(MSG_DisplayObject_ShortcutInUse), EmptyNullStr(name));
	gtk_widget_grab_focus(entry);
    }
}

static void defaultBackgroundOnToggled(GtkWidget *def, GtkWidget *bg)
{
    gtk_widget_set_sensitive(bg, 
	    !gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(def)));
}

struct EditTextData
{
    Project *d_Project;
    DisplayObject *d_Text;
    StructureEditorData *d_StructureEditorData;
    GuiObject d_Title, d_Weight, d_Fontname;
    GuiObject d_Background, d_DefaultBackground;
    GuiObject d_Ok, d_Cancel;
};

static void closeEditText(GuiObject go, gint response, EditTextData *d)
{
    Project *pr = d->d_Project;

    if (response == GTK_RESPONSE_ACCEPT) {
	GuiBusyCursor busy;
        long error = OK;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

	struct TextData td;
        td.td_Title = gtkEntryText(d->d_Title);
        td.td_Weight = (int)gtkRangeValue(d->d_Weight);
        td.td_Fontname = gtkEntryText(d->d_Fontname);
	td.td_Background = gtkToggleButtonActive(d->d_DefaultBackground)? 
            NULL : gtk_pop_image_get_spec(GTK_POP_IMAGE(d->d_Background));

        DisplayChangeText(d->d_Text, &td);
        ProjectChange(pr, TRUE);

        TextFinish(d->d_StructureEditorData, d->d_Text);
    }

    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

void GuiText(StructureEditorData *sed, DisplayObject *text)
{
    EditTextData *d = g_new0(EditTextData, 1);
    Project *pr = DisplayObjectProject(text);

    d->d_Project = pr;
    d->d_Text = text;
    d->d_StructureEditorData = sed;

    d->d_Title = EntryObject,
	"max-length", TITLE_LEN,
	"xalign", 0.5,
        "text", EmptyNullStr(DisplayObjectTitle(text)),
	End;
    GuiSetTooltip(d->d_Title, GCS(MSG_DisplayText_TitleBH));

    int weight = DisplayObjectWeight(text);
    d->d_Weight = HScaleObject,
        "adjustment", gtk_adjustment_new(weight, 0, 1000, 1, 10, 0),
	"value-pos", GTK_POS_RIGHT,
	"digits", 0,
	End;
    GuiObject weightBox = GuiMakeTooltip(d->d_Weight, GCS(MSG_WeightBH));

    d->d_Fontname = EntryObject,
        "max-length", FONTNAME_LEN,
        "xalign", 0.5,
        "text", EmptyNullStr(DisplayObjectGTKFontname(text)),
        End;
    GuiSetTooltip(d->d_Fontname, GCS(MSG_FontBH));

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    const char *bg = DisplayObjectBackground(text);
    d->d_Background = PopImageObject,
        "spec", (bg && bg[0])? bg : IMAGESPEC_START,
        // MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_Background,
	End;
    g_signal_connect_swapped(d->d_Background, "clicked",
            G_CALLBACK(ProjectSetCurrentDir), pr);
    GuiObject backgroundBox = GuiMakeTooltip(d->d_Background, 
            GCS(MSG_BackgroundBH));

    d->d_DefaultBackground = CheckButtonObject, 
        "label", GCS(MSG_Background_Default),
        "use-underline", true,
        End;
    GuiSetTooltip(d->d_DefaultBackground, GCS(MSG_Background_DefaultBH));
    g_signal_connect(d->d_DefaultBackground, "toggled", 
	    G_CALLBACK(defaultBackgroundOnToggled), d->d_Background);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_DefaultBackground),
	    !bg || !bg[0]);

    GtkWidget *back = HGroup,
        "child", backgroundBox, "weight", 200,
        "child", SpaceObject, End,
        "child", VCenter(d->d_DefaultBackground),
        End;

    GuiObject basic = VGroup,
        "columns", 2,
        ChildLabelWidget(GCS(MSG_Title), d->d_Title), "vert-weight", 0,
	"child", GuiLabel(GCS(MSG_Weight), d->d_Weight), "weight", 0,
        "child", weightBox, "vert-weight", 0,
	"child", GuiLabel(GCS(MSG_Font), d->d_Fontname), "weight", 0,
	"child", PopSelectionObject,
	    "selection-type", GUI_TYPE_FONT_CHOOSER_DIALOG,
	    "string", d->d_Fontname,
	    "button", GuiPopButton(SELECT_FONT_ICON_NAME), 
	    End, "vert-weight", 0,
	"child", GuiLabel(GCS(MSG_Background), d->d_Background), "weight", 0,
        "child", back,
	End;

    GuiObject group = basic;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_Text),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Text editor");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    GuiAddEntryCallbacks(d->d_Title, d->d_Ok);
    GuiAddEntryCallbacks(d->d_Fontname, d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(closeEditText), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }
}

struct EditSpaceData
{
    Project *d_Project;
    DisplayObject *d_Space;
    StructureEditorData *d_StructureEditorData;
    GuiObject d_Bar, d_Weight, d_Background, d_DefaultBackground;
    GuiObject d_Ok, d_Cancel;
};

static void closeEditSpace(GuiObject go, gint response, EditSpaceData *d)
{
    Project *pr = d->d_Project;

    if (response == GTK_RESPONSE_ACCEPT) {
	GuiBusyCursor busy;
        long error = OK;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

	SpaceData sd;
        sd.sd_Bar = (bool)gtkToggleButtonActive(d->d_Bar);
        sd.sd_Weight = (int)gtkRangeValue(d->d_Weight);
	sd.sd_Background = gtkToggleButtonActive(d->d_DefaultBackground)? 
            NULL : gtk_pop_image_get_spec(GTK_POP_IMAGE(d->d_Background));

        DisplayChangeSpace(d->d_Space, &sd);
        ProjectChange(pr, TRUE);

        SpaceFinish(d->d_StructureEditorData, d->d_Space);
    }

    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

void GuiSpace(StructureEditorData *sed, DisplayObject *space)
{
    EditSpaceData *d = g_new0(EditSpaceData, 1);
    Project *pr = DisplayObjectProject(space);

    d->d_Project = pr;
    d->d_Space = space;
    d->d_StructureEditorData = sed;

    d->d_Bar = CheckButtonObject,
        "active", DisplaySpaceBar(space),
	End;
    GuiSetTooltip(d->d_Bar, GCS(MSG_DisplaySpace_DelimiterBH));

    int weight = DisplayObjectWeight(space);
    d->d_Weight = HScaleObject,
        "adjustment", gtk_adjustment_new(weight, 0, 1000, 1, 10, 0),
	"value-pos", GTK_POS_RIGHT,
	"digits", 0,
	End;
    GuiObject weightBox = GuiMakeTooltip(d->d_Weight, GCS(MSG_WeightBH));

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    const char *bg = DisplayObjectBackground(space);
    d->d_Background = PopImageObject,
        "spec", (bg && bg[0])? bg : IMAGESPEC_START,
        // MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_Background,
	End;
    g_signal_connect_swapped(d->d_Background, "clicked",
            G_CALLBACK(ProjectSetCurrentDir), pr);
    GuiObject backgroundBox = GuiMakeTooltip(d->d_Background,
            GCS(MSG_BackgroundBH));

    d->d_DefaultBackground = CheckButtonObject, 
        "label", GCS(MSG_Background_Default),
        "use-underline", true,
        End;
    GuiSetTooltip(d->d_DefaultBackground, GCS(MSG_Background_DefaultBH));
    g_signal_connect(d->d_DefaultBackground, "toggled", 
	    G_CALLBACK(defaultBackgroundOnToggled), d->d_Background);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_DefaultBackground),
	    !bg || !bg[0]);

    GtkWidget *back = HGroup,
        "child", backgroundBox, "weight", 200,
        "child", SpaceObject, End,
        "child", VCenter(d->d_DefaultBackground),
        End;

    GuiObject basic = VGroup,
        "columns", 2,
        "child", GuiLabel(GCS(MSG_DisplaySpace_Delimiter), d->d_Bar),"weight",0,
        "child", d->d_Bar, "weight", 0,
	"child", GuiLabel(GCS(MSG_Weight), d->d_Weight), "weight", 0,
        "child", weightBox, "vert-weight", 0,
	"child", GuiLabel(GCS(MSG_Background), d->d_Background),
        "child", back,
	End;

    GuiObject group = basic;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_Space),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Space editor");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(closeEditSpace), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }
}

struct EditImageData
{
    Project *d_Project;
    DisplayObject *d_Image;
    StructureEditorData *d_StructureEditorData;
    GuiObject d_Weight, d_PopImage;
    GuiObject d_ImageSize;
    GuiObject d_Ok, d_Cancel;
};

static void closeEditImage(GuiObject go, gint response, EditImageData *d)
{
    Project *pr = d->d_Project;

    if (response == GTK_RESPONSE_ACCEPT) {
	GuiBusyCursor busy;
        long error = OK;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        ImageData id;
        id.id_Weight = (int)gtk_range_get_value(GTK_RANGE(d->d_Weight));
        id.id_Image = gtk_pop_image_get_spec(GTK_POP_IMAGE(d->d_PopImage));
        id.id_ImageSize = gtk_radio_box_get_active(
                GTK_RADIO_BOX(d->d_ImageSize));

        DisplayChangeImage(d->d_Image, &id);
        ProjectChange(pr, TRUE);

        ImageFinish(d->d_StructureEditorData, d->d_Image);
    }

    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

void GuiImage(StructureEditorData *sed, DisplayObject *image)
{
    EditImageData *d = g_new0(EditImageData, 1);
    Project *pr = DisplayObjectProject(image);

    d->d_Project = pr;
    d->d_Image = image;
    d->d_StructureEditorData = sed;

    int weight = DisplayObjectWeight(image);
    d->d_Weight = HScaleObject,
        "adjustment", gtk_adjustment_new(weight, 0, 1000, 1, 10, 0),
	"value-pos", GTK_POS_RIGHT,
	"digits", 0,
	End;
    GuiObject weightBox = GuiMakeTooltip(d->d_Weight, GCS(MSG_WeightBH));

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    const char *bg = DisplayObjectBackground(image);
    d->d_PopImage = PopImageObject,
        "spec", (bg && bg[0])? bg : IMAGESPEC_START,
        // MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_All,
	End;
    GuiSetTooltip(d->d_PopImage, GCS(MSG_ImageBH));
    g_signal_connect_swapped(d->d_PopImage, "clicked",
            G_CALLBACK(ProjectSetCurrentDir), pr);

    static char *entries[3];
    entries[0] = GCS(MSG_Image_Size_Resizeable);
    entries[1] = GCS(MSG_Image_Size_Fixed);
    entries[2] = NULL;
    d->d_ImageSize = RadioBoxObject,
        "entries", entries,
        "active", DisplayImageImageSize(image),
        End;

    GuiObject basic = VGroup,
        "columns", 2,
	"child", GuiLabel(GCS(MSG_Weight), d->d_Weight), "weight", 0,
        "child", weightBox, "vert-weight", 0,
	"child", GuiLabel(GCS(MSG__Image), d->d_PopImage), "weight", 0,
        "child", HGroup,
            "child", d->d_PopImage,
            "child", Frame(GCS(MSG_Size)),
                "child", GroupAlignment(true),
                    "child", VCenter(d->d_ImageSize),
                    End,
                End, "weight", 20,
            End,
        End;

    GuiObject group = basic;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_Image),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Image editor");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(closeEditImage), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }
}

struct EditRegisterData
{
    Project *d_Project;
    DisplayObject *d_Register;
    StructureEditorData *d_StructureEditorData;
    GuiObject d_Weight;
    // GuiObject d_Background, d_DefaultBackground;
    GuiObject d_Labels;
    GuiObject d_Ok, d_Cancel;
};

static void closeEditRegister(GuiObject go, gint response, EditRegisterData *d)
{
    Project *pr = d->d_Project;

    if (response == GTK_RESPONSE_ACCEPT) {
	GuiBusyCursor busy;
        long error = OK;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

	RegisterData rd;
        rd.rd_Weight = (int)gtkRangeValue(d->d_Weight);
        rd.rd_Fontname = 0;
        rd.rd_Background = 0;
	//rd.rd_Background = gtkToggleButtonActive(d->d_DefaultBackground)? 
        //    0 : gtk_pop_image_get_spec(GTK_POP_IMAGE(d->d_Background));
	rd.rd_Labels = gtk_edit_labels_get_labels(GTK_EDIT_LABELS(d->d_Labels));

        error = DisplayChangeRegister(d->d_Register, &rd);
	if (error) {
	    Ask2(go, GCS(MSG_ErrorMessage), AT_ERROR, 
		    GCS(MSG_StarOk), GES(error));
	} else {
	    ProjectChange(pr, TRUE);
	}

        RegisterFinish(d->d_StructureEditorData, d->d_Register);
    }

    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

void GuiRegister(StructureEditorData *sed, DisplayObject *reg)
{
    EditRegisterData *d = g_new0(EditRegisterData, 1);
    Project *pr = DisplayObjectProject(reg);

    d->d_Project = pr;
    d->d_Register = reg;
    d->d_StructureEditorData = sed;

    int weight = DisplayObjectWeight(reg);
    d->d_Weight = HScaleObject,
        "adjustment", gtk_adjustment_new(weight, 0, 1000, 1, 10, 0),
	"value-pos", GTK_POS_RIGHT,
	"digits", 0,
	End;
    GuiObject weightBox = GuiMakeTooltip(d->d_Weight, GCS(MSG_WeightBH));

#if 0
    const char *bg = DisplayObjectBackground(reg);
    d->d_Background = PopImageObject,
        "spec", (bg && bg[0])? bg : IMAGESPEC_START,
        // MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_Background,
	End;
    GuiObject backgroundBox = EventBoxObject,
        "child", d->d_Background,
        End;
    GuiSetTooltip(backgroundBox, GCS(MSG_BackgroundBH));

    d->d_DefaultBackground = CheckButtonObject, 
        "label", unicodetmp(GCS(MSG_Background_Default)),
        "use-underline", true,
        End;
    GuiObject defaultBackgroundBox = EventBoxObject,
        "child", d->d_DefaultBackground,
        End;
    GuiSetTooltip(defaultBackgroundBox, GCS(MSG_Background_DefaultBH));
    g_signal_connect(d->d_DefaultBackground, "toggled", 
	    G_CALLBACK(defaultBackgroundOnToggled), d->d_Background);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_DefaultBackground),
	    !bg || !bg[0]);

    GtkWidget *back = HGroup,
        "child", backgroundBox, "weight", 200,
        "child", SpaceObject, End,
        "child", VCenter(defaultBackgroundBox),
        End;
#endif

    GuiObject basic = VGroup,
        "columns", 2,
	"child", GuiLabel(GCS(MSG_Weight), d->d_Weight), "weight", 0,
        "child", weightBox, "vert-weight", 0,
//	"child", GuiLabel(GCS(MSG_Background), d->d_Background), "weight", 0,
//        "child", back,
	End;

    d->d_Labels = EditLabelsObject,
	"labels", DisplayRegisterLabels(reg),
	End;
    GuiObject labelsBox = GuiMakeTooltip(d->d_Labels, 
            GCS(MSG_DisplayRegister_LabelsBH));

    GuiObject group = VGroup,
	// MUIA_Frame, MUIV_Frame_Group,
        "child", basic, "vert-weight", 0,
	"child", FrameObject, \
	    "label", GCS(MSG_DisplayRegister_Labels),
	    "label-xalign", 0.5,
	    "shadow-type", GTK_SHADOW_NONE,
	    "child", labelsBox,
	    End,
	End;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_Register),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Register group editor");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    GuiAddEntryCallbacks(GTK_EDIT_LABELS(d->d_Labels)->entry, d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(closeEditRegister), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }
}

struct EditGroupData
{
    Project *d_Project;
    DisplayObject *d_Group;
    StructureEditorData *d_StructureEditorData;
    GuiObject d_Title, d_Weight, d_Border, d_Horiz, d_Spacing;
    GuiObject d_Background, d_DefaultBackground;
    GuiObject d_Ok, d_Cancel;
};

static void closeEditGroup(GuiObject go, gint response, EditGroupData *d)
{
    Project *pr = d->d_Project;

    if (response == GTK_RESPONSE_ACCEPT) {
	GuiBusyCursor busy;
        long error = OK;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

	GroupData gd;
        gd.gd_Title = gtkEntryText(d->d_Title);
        gd.gd_Weight = (int)gtkRangeValue(d->d_Weight);
        gd.gd_Fontname = 0;
	gd.gd_Background = gtkToggleButtonActive(d->d_DefaultBackground)? 
            0 : gtk_pop_image_get_spec(GTK_POP_IMAGE(d->d_Background));
	gd.gd_Horiz = (bool)gtkToggleButtonActive(d->d_Horiz);
	gd.gd_Border = (bool)gtkToggleButtonActive(d->d_Border);
	gd.gd_Spacing = (bool)gtkToggleButtonActive(d->d_Spacing);

        DisplayChangeGroup(d->d_Group, &gd);
        ProjectChange(pr, TRUE);

        GroupFinish(d->d_StructureEditorData, d->d_Group);
    }

    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

void GuiGroup(StructureEditorData *sed, DisplayObject *group)
{
    EditGroupData *d = g_new0(EditGroupData, 1);
    Project *pr = DisplayObjectProject(group);

    d->d_Project = pr;
    d->d_Group = group;
    d->d_StructureEditorData = sed;

    d->d_Title = EntryObject,
	"max-length", TITLE_LEN,
	"xalign", 0.5,
        "text", EmptyNullStr(DisplayObjectTitle(group)),
	End;
    GuiSetTooltip(d->d_Title, GCS(MSG_DisplayGroup_TitleBH));

    int weight = DisplayObjectWeight(group);
    d->d_Weight = HScaleObject,
        "adjustment", gtk_adjustment_new(weight, 0, 1000, 1, 10, 0),
	"value-pos", GTK_POS_RIGHT,
	"digits", 0,
	End;
    GuiObject weightBox = GuiMakeTooltip(d->d_Weight, GCS(MSG_WeightBH));

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    const char *bg = DisplayObjectBackground(group);
    d->d_Background = PopImageObject,
        "spec", (bg && bg[0])? bg : IMAGESPEC_START,
        // MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_Background,
	End;
    g_signal_connect_swapped(d->d_Background, "clicked",
            G_CALLBACK(ProjectSetCurrentDir), pr);
    GuiObject backgroundBox = GuiMakeTooltip(d->d_Background,
            GCS(MSG_BackgroundBH));

    d->d_DefaultBackground = CheckButtonObject, 
        "label", GCS(MSG_Background_Default),
        "use-underline", true,
        End;
    GuiSetTooltip(d->d_DefaultBackground, GCS(MSG_Background_DefaultBH));
    g_signal_connect(d->d_DefaultBackground, "toggled", 
	    G_CALLBACK(defaultBackgroundOnToggled), d->d_Background);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_DefaultBackground),
	    !bg || !bg[0]);

    GtkWidget *back = HGroup,
        "child", backgroundBox, "weight", 200,
        "child", SpaceObject, End,
        "child", VCenter(d->d_DefaultBackground),
        End;

    GuiObject basic = VGroup,
        "columns", 2,
        ChildLabelWidget(GCS(MSG_Title), d->d_Title), "vert-weight", 0,
	"child", GuiLabel(GCS(MSG_Weight), d->d_Weight), "weight", 0,
        "child", weightBox, "vert-weight", 0,
	"child", GuiLabel(GCS(MSG_Background), d->d_Background), "weight", 0,
        "child", back,
	End;

    d->d_Border = CheckButtonObject,
        "label", GCS(MSG_DisplayGroup_Border),
        "use-underline", true,
        "active", DisplayGroupBorder(group),
	End;
    GuiSetTooltip(d->d_Border, GCS(MSG_DisplayGroup_BorderBH));

    d->d_Horiz = CheckButtonObject,
        "label", GCS(MSG_DisplayGroup_Horizontal),
        "use-underline", true,
        "active", DisplayGroupHoriz(group),
	End;
    GuiSetTooltip(d->d_Horiz, GCS(MSG_DisplayGroup_HorizontalBH));

    d->d_Spacing = CheckButtonObject,
        "label", GCS(MSG_DisplayGroup_Spacing),
        "use-underline", true,
        "active", DisplayGroupSpacing(group),
	End;
    GuiSetTooltip(d->d_Spacing, GCS(MSG_DisplayGroup_SpacingBH));

    GuiObject attrs = VGroup,
        "child", d->d_Border,
        "child", d->d_Horiz,
        "child", d->d_Spacing,
        End;
    attrs = Frame(GCS(MSG_DisplayGroup_Attributes)),
        "child", GroupAlignment(true),
	    "child", HCenter(attrs),
            End,
        End;

    GuiObject groupW = VGroup,
        "child", basic,
	"child", attrs, "vert-weight", 0,
	End;

    GuiObject root = GroupAlignment(false),
        "child", groupW,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_Group),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Group editor");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    GuiAddEntryCallbacks(d->d_Title, d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(closeEditGroup), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, groupW);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }
}

struct EditWindowData
{
    Project *d_Project;
    DisplayObject *d_Window;
    StructureEditorData *d_StructureEditorData;
    GuiObject d_Title, d_Button;
    GuiObject d_Shortcut, d_Enabled, d_EnabledComputeStr, d_EnabledCompute;
    GuiObject d_CloseWhenDisabled, d_Weight, d_Fontname;
    GuiObject d_Background, d_DefaultBackground;
    GuiObject d_Ok, d_Cancel;
};

static void closeEditWindow(GuiObject go, gint response, EditWindowData *d)
{
    Project *pr = d->d_Project;

    if (response == GTK_RESPONSE_ACCEPT) {
	GuiBusyCursor busy;
        long error = OK;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

	WindowData wd;
        wd.wd_Title = gtkEntryText(d->d_Title);
	wd.wd_Id = DisplayWindowId(d->d_Window);
	wd.wd_Button = (bool)gtkToggleButtonActive(d->d_Button);
	wd.wd_Shortcut = CodesetGetCharCheck(gtkEntryText(d->d_Shortcut));
	wd.wd_Enabled = (char)gtkRadioBoxActive(d->d_Enabled);
        wd.wd_EnabledCompute = gtkEntryText(d->d_EnabledComputeStr);
        wd.wd_CloseWhenDisabled = 
            (bool)gtkToggleButtonActive(d->d_CloseWhenDisabled);
        wd.wd_Weight = (int)gtkRangeValue(d->d_Weight);
        wd.wd_Fontname = gtkEntryText(d->d_Fontname);
	wd.wd_Background = gtkToggleButtonActive(d->d_DefaultBackground)? 
            0 : gtk_pop_image_get_spec(GTK_POP_IMAGE(d->d_Background));

        DisplayChangeWindow(d->d_Window, &wd);
        ProjectChange(pr, TRUE);

        WindowFinish(d->d_StructureEditorData, d->d_Window);
    }

    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

static void editWindowOnButtonToggled(GuiObject button, GuiObject settings)
{
    gtk_widget_set_sensitive(settings, gtkToggleButtonActive(button));
}

static void editWindowOnEnabledChanged(EditWindowData *d)
{
    short enabled = gtkRadioBoxActive(d->d_Enabled);

    gtk_widget_set_sensitive(d->d_EnabledCompute, 
            enabled == DOD_ENABLED_COMPUTED);
    gtk_widget_set_sensitive(d->d_CloseWhenDisabled, 
            enabled != DOD_ENABLED_ALWAYS);
    if (enabled == DOD_ENABLED_ALWAYS) {
        gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(d->d_CloseWhenDisabled), false);
    }
}

void GuiWindow(StructureEditorData *sed, DisplayObject *window)
{
    EditWindowData *d = g_new0(EditWindowData, 1);
    Project *pr = DisplayObjectProject(window);

    d->d_Project = pr;
    d->d_Window = window;
    d->d_StructureEditorData = sed;

    d->d_Title = EntryObject,
	"max-length", TITLE_LEN,
	"xalign", 0.5,
        "text", EmptyNullStr(DisplayObjectTitle(window)),
	End;
    GuiSetTooltip(d->d_Title, GCS(MSG_DisplayWindow_TitleBH));

    d->d_Button = CheckButtonObject, 
        "active", DisplayWindowButton(window),
        "use-underline", true,
        End;
    GuiSetTooltip(d->d_Button, GCS(MSG_DisplayWindow_WindowButtonBH));

    char sc[CodesetMaxBytesForChar+1];
    CodesetCharToStr(DisplayObjectShortcut(window), sc);
    d->d_Shortcut = EntryObject,
	"max-length", 1,
        "width-chars", 5,
	"xalign", 0.5,
        "text", sc,
	End;
    g_signal_connect(d->d_Shortcut, "activate", 
	    G_CALLBACK(ShortCutOnActivate), window);
    GuiSetTooltip(d->d_Shortcut, GCS(MSG_DisplayObject_ShortcutBH));

    static const char *etext[4];
    etext[0] = GCS(MSG_DisplayObject_Enabled_Always);
    etext[1] = GCS(MSG_DisplayObject_Enabled_NotInitial);
    etext[2] = GCS(MSG_DisplayObject_Enabled_Computed);
    etext[3] = 0;
    d->d_Enabled = RadioBoxObject,
        "entries", etext,
        End;
    GuiObject enabledBox = GuiMakeTooltip(d->d_Enabled,
            GCS(MSG_DisplayObject_EnabledBH));

    d->d_EnabledComputeStr = EntryObject,
        "max-length", FIELD_TRIGGER_LEN,
        "xalign", 0.5,
        "text", EmptyNullStr(DisplayObjectEnabledCompute(window)),
        End;
    GuiSetTooltip(d->d_EnabledComputeStr, 
            GCS(MSG_DisplayObject_ComputeEnabledBH));

    d->d_EnabledCompute = UserFunctionsPopup(d->d_EnabledComputeStr, pr);
    gtk_widget_set_sensitive(d->d_EnabledCompute, false);

    GuiObject ceLabel = GuiLabel(GCS(MSG_Compute), d->d_EnabledCompute, 0.5);

    d->d_CloseWhenDisabled = CheckButtonObject, 
        "label", GCS(MSG_DisplayWindow_CloseWindowWhenDisabled),
        "active", DisplayWindowCloseWhenDisabled(window),
        "use-underline", true,
        End;
    GuiSetTooltip(d->d_CloseWhenDisabled, 
            GCS(MSG_DisplayWindow_CloseWindowWhenDisabledBH));

    g_signal_connect_swapped(d->d_Enabled, "active-changed",
            G_CALLBACK(editWindowOnEnabledChanged), d);
    gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_Enabled),
                DisplayObjectEnabled(window));

    int weight = DisplayObjectWeight(window);
    d->d_Weight = HScaleObject,
        "adjustment", gtk_adjustment_new(weight, 0, 1000, 1, 10, 0),
	"value-pos", GTK_POS_RIGHT,
	"digits", 0,
	End;
    GuiObject weightBox = GuiMakeTooltip(d->d_Weight, GCS(MSG_WeightBH));

    d->d_Fontname = EntryObject,
        "max-length", FONTNAME_LEN,
        "xalign", 0.5,
        "text", EmptyNullStr(DisplayObjectGTKFontname(window)),
        End;
    GuiSetTooltip(d->d_Fontname, GCS(MSG_FontBH));

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    const char *bg = DisplayObjectBackground(window);
    d->d_Background = PopImageObject,
        "spec", (bg && bg[0])? bg : IMAGESPEC_START,
        "types", GTK_POP_IMAGE_COLOR,
	End;
    g_signal_connect_swapped(d->d_Background, "clicked",
            G_CALLBACK(ProjectSetCurrentDir), pr);
    GuiObject backgroundBox = GuiMakeTooltip(d->d_Background,
            GCS(MSG_BackgroundBH));

    d->d_DefaultBackground = CheckButtonObject, 
        "label", GCS(MSG_Background_Default),
        "use-underline", true,
        End;
    GuiSetTooltip(d->d_DefaultBackground, GCS(MSG_Background_DefaultBH));
    g_signal_connect(d->d_DefaultBackground, "toggled", 
	    G_CALLBACK(defaultBackgroundOnToggled), d->d_Background);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_DefaultBackground),
	    !bg || !bg[0]);

    GtkWidget *back = HGroup,
        "child", backgroundBox, "weight", 200,
        "child", SpaceObject, End,
        "child", VCenter(d->d_DefaultBackground),
        End;

    GuiObject basic = VGroup,
        "columns", 2,
        ChildLabelWidget(GCS(MSG_Title), d->d_Title), "vert-weight", 0,
        ChildLabelWidget(GCS(MSG_DisplayWindow_WindowButton), d->d_Button), 
                "vert-weight", 0,
        End;

    GuiObject settings1 = HGroup,
        ChildLabelWidget(GCS(MSG_DisplayObject_Shortcut), d->d_Shortcut),
        "child", SpaceObject, End, "weight", 300,
        End;

    GuiObject settings2 = HGroup,
        "child", enabledBox, "weight", 0,
        "child", VSeparatorObject, End, "weight", 0,
        "child", VGroup,
            "child", SpaceObject, End,
            "child", ceLabel, "weight", 0,
            "child", d->d_EnabledCompute, "weight", 0,
            End,
        End;
    settings2 = VGroup,
        "child", settings2,
        "child", HSeparatorObject, End, "weight", 0,
        "child", d->d_CloseWhenDisabled,
        End;
    settings2 = Frame(GCS(MSG_DisplayObject_Enabled)),
        "child", GroupAlignment(true),
            "child", settings2,
            End,
        End;

    GuiObject settings3 = VGroup,
        "columns", 2,
	"child", GuiLabel(GCS(MSG_Weight), d->d_Weight), "weight", 0,
        "child", weightBox, "vert-weight", 0,
	"child", GuiLabel(GCS(MSG_Font), d->d_Fontname), "weight", 0,
	"child", PopSelectionObject,
	    "selection-type", GUI_TYPE_FONT_CHOOSER_DIALOG,
	    "string", d->d_Fontname,
	    "button", GuiPopButton(SELECT_FONT_ICON_NAME), 
	    End, "vert-weight", 0,
	"child", GuiLabel(GCS(MSG_Background), d->d_Background), "weight", 0,
        "child", back,
	End;

    GuiObject settings = VGroup,
        "sensitive", DisplayWindowButton(window),
        "child", settings1, "vert-weight", 0,
        "child", settings2, "vert-weight", 0,
        "child", settings3,
        End;

    g_signal_connect(d->d_Button, "toggled", 
            G_CALLBACK(editWindowOnButtonToggled), settings);

    GuiObject group = VGroup,
        "child", basic, "vert-weight", 0,
        "child", Frame(GCS(MSG_DisplayWindow_ButtonSettings)),
            "child", GroupAlignment(true),
                "child", settings,
                End,
            End,
        End;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_Window),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Window editor");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    GuiAddEntryCallbacks(d->d_Title, d->d_Ok);
    GuiAddEntryCallbacks(d->d_Shortcut, d->d_Ok);
    GuiAddEntryCallbacks(d->d_EnabledComputeStr, d->d_Ok);
    GuiAddEntryCallbacks(d->d_Fontname, d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(closeEditWindow), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }
}

#endif

