//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkEditDisplayField.cc
//
// Creation date: 02.09.2005
// Last modified: 10.02.2022
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "gtk/gtkEditDisplayField.h"
#include "gtk/gtkEditDisplay.h"
#include "gtk/gtkEditRecordDescriptionSpec.h"
#include "gtk/gtkStructureEditor.h"
#include "gtk/classes/gtkgroup.h"
#include "gtk/classes/gtkradiobox.h"
#include "gtk/classes/gtkpopimage.h"
#include "gtk/classes/gtkpopselection.h"
#include "gtk/classes/gtkpopeditlabels.h"
#include "gtk/gtkimpl.h"
#include "gtk/gtkUtil.h"
#include "requesters.h"
#include "preferences.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

#define IMAGESPEC_START "0:128"

struct EditDisplayFieldData
{
    Project *d_Project;
    DisplayObject *d_DisplayField;
    StructureEditorData *d_StructureEditorData;
    GuiObject d_Title, d_TitlePosition, d_TitleFontname, d_TitleFontnamePopup;
    GuiObject d_Shortcut, d_Home, d_TabChain, d_ReadOnly;
    GuiObject d_Alignment, d_Calendar, d_Format, d_ShowTitles;
    GuiObject d_SubType, d_SubTypeGroup;
    GuiObject d_Frame, d_Horizontal, d_ControlTable;
    GuiObject d_OnDoubleClick, d_DoubleClickTriggerStr, d_DoubleClickTrigger;
    GuiObject d_URLDropTriggerStr, d_URLDropTrigger;
    GuiObject d_SortDropTriggerStr, d_SortDropTrigger;
    GuiObject d_NumDecimals, d_AutoUpdate;
    GuiObject d_Enabled, d_EnabledComputeStr, d_EnabledCompute;
    GuiObject d_Weight, d_BubbleHelp;
    GuiObject d_Fontname, d_FontnamePopup, d_Background, d_DefaultBackground;
    GuiObject d_Image, d_ImageSize, d_ImageSizeGroup, d_InitialColor;
    GuiObject d_RecordDescriptionSpec, d_RecordDescriptionSpecGroup;
    GuiObject d_ReferencePopupGroup, d_RecordsKind; 
    GuiObject d_ComputeRecordsStr, d_ComputeRecords, d_ComputeRecordsGroup;
    GuiObject d_InitialRecord, d_CurrentRecord;
    GuiObject d_Show, d_AutoShow, d_Filter;
    GuiObject d_DisplayExternal, d_HideStringField, d_TitleAtStringField;
    GuiObject d_Listview, d_File, d_Font, d_View;
    GuiObject d_ListviewSettings, d_ListviewKind, d_ListviewKindGroup;
    GuiObject d_ListviewLabels, d_ListviewComputeStr, d_ListviewCompute;
    GuiObject d_Ok, d_Cancel;
};

static void closeEditDisplayField(GuiObject go, gint response, 
        EditDisplayFieldData *d)
{
    Project *pr = d->d_Project;

    if (response == GTK_RESPONSE_ACCEPT) {
	GuiBusyCursor busy;

        Field *field = DisplayFieldField(d->d_DisplayField);
        short type = FieldType(field);
        MemPool *memPool = ProjectMemPool(pr);
        char *bubbleHelp = 0;
        struct RecordDescriptionSpec *rds = 0;
        long error = OK;
        DisplayFieldData dfd;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        memset(&dfd, 0, sizeof(dfd));
        dfd.dfd_Title = gtkEntryText(d->d_Title);
        dfd.dfd_TitlePosition = gtkComboBoxActive(d->d_TitlePosition);
        dfd.dfd_TitleFontname = gtkEntryText(d->d_TitleFontname);
        dfd.dfd_Shortcut = CodesetGetCharCheck(gtkEntryText(d->d_Shortcut));
        dfd.dfd_Home = (bool)gtkToggleButtonActive(d->d_Home);
        dfd.dfd_TabChain = (bool)gtkToggleButtonActive(d->d_TabChain);
        dfd.dfd_ReadOnly = (bool)gtkToggleButtonActive(d->d_ReadOnly);
        dfd.dfd_SubType = gtkComboBoxActive(d->d_SubType);
        dfd.dfd_Alignment = gtkComboBoxActive(d->d_Alignment);
        dfd.dfd_Calendar = (bool)gtkToggleButtonActive(d->d_Calendar);
        dfd.dfd_Format = gtkComboBoxActive(d->d_Format);
        dfd.dfd_NumDecimals = (int)gtkSpinButtonValue(d->d_NumDecimals);
        dfd.dfd_ShowTitles = (bool)gtkToggleButtonActive(d->d_ShowTitles);
        dfd.dfd_ControlTable = (bool)gtkToggleButtonActive(d->d_ControlTable);
        dfd.dfd_OnDoubleClick = gtkRadioBoxActive(d->d_OnDoubleClick);
        dfd.dfd_DoubleClickTrigger = gtkEntryText(d->d_DoubleClickTriggerStr); 
        dfd.dfd_URLDropTrigger = gtkEntryText(d->d_URLDropTriggerStr); 
        dfd.dfd_SortDropTrigger = gtkEntryText(d->d_SortDropTriggerStr); 
        dfd.dfd_AutoUpdate=(bool)gtkToggleButtonActive(d->d_AutoUpdate);
	dfd.dfd_Enabled = (char)gtkRadioBoxActive(d->d_Enabled);
        dfd.dfd_EnabledCompute = gtkEntryText(d->d_EnabledComputeStr);
        dfd.dfd_Weight = (int)gtkRangeValue(d->d_Weight);
        dfd.dfd_Fontname = gtkEntryText(d->d_Fontname);

        if (type == TYPE_BUTTON && dfd.dfd_SubType == DAD_BUTTONTYPE_IMAGE) {
            if (d->d_Image) {
                dfd.dfd_Background = gtk_pop_image_get_spec(
                        GTK_POP_IMAGE(d->d_Image));
            }
        } else if (d->d_Background) {
            if (!gtkToggleButtonActive(d->d_DefaultBackground)) {
                dfd.dfd_Background = gtk_pop_image_get_spec(
                        GTK_POP_IMAGE(d->d_Background));
            }
        }
        dfd.dfd_ImageSize = gtkRadioBoxActive(d->d_ImageSize);
        if (d->d_BubbleHelp) {
            bubbleHelp = GuiEditorText(d->d_BubbleHelp, memPool);
            dfd.dfd_BubbleHelp = bubbleHelp;
            if (!dfd.dfd_BubbleHelp) {
                error = ERR_MEM;
            }
        }
        dfd.dfd_File = gtkToggleButtonActive(d->d_File);
        dfd.dfd_Font = gtkToggleButtonActive(d->d_Font);
        dfd.dfd_Listview = gtkToggleButtonActive(d->d_Listview);
        dfd.dfd_View = gtkToggleButtonActive(d->d_View);
        if (d->d_ListviewLabels) {
            dfd.dfd_ListviewLabels = gtk_pop_edit_labels_get_labels(
                    GTK_POP_EDIT_LABELS(d->d_ListviewLabels));
        }
	if (gtkComboBoxActive(d->d_ListviewKind) == 1) {
	    dfd.dfd_ListviewCompute = gtkEntryText(d->d_ListviewComputeStr);
	}
        dfd.dfd_DisplayExternal = gtkToggleButtonActive(d->d_DisplayExternal);
        dfd.dfd_TitleAtStringField = gtkToggleButtonActive(
                d->d_TitleAtStringField);
        dfd.dfd_HideStringField = gtkToggleButtonActive(d->d_HideStringField);
        dfd.dfd_Show = gtkToggleButtonActive(d->d_Show);
        dfd.dfd_AutoShow = gtkToggleButtonActive(d->d_AutoShow);
        dfd.dfd_Filter = gtkToggleButtonActive(d->d_Filter);
        dfd.dfd_Frame = gtkToggleButtonActive(d->d_Frame);
        if (d->d_Horizontal) {
            dfd.dfd_Alignment = gtkToggleButtonActive(d->d_Horizontal)?
                DAD_ALIGNMENT_HORIZONTAL : DAD_ALIGNMENT_VERTICAL;
        }
	dfd.dfd_RecordsKind = gtkComboBoxActive(d->d_RecordsKind);
    	dfd.dfd_RecordsCompute = gtkEntryText(d->d_ComputeRecordsStr);
    	dfd.dfd_InitialRecord = gtkToggleButtonActive(d->d_InitialRecord);
    	dfd.dfd_CurrentRecord = gtkToggleButtonActive(d->d_CurrentRecord);

        if (d->d_RecordDescriptionSpec) {
            rds = EditRecordDescriptionSpecGetSpec(
                    d->d_RecordDescriptionSpec, memPool);
            dfd.dfd_RecordDescriptionSpec = rds;
            if (rds == NULL) {
                error = ERR_MEM;
            }
        }

        if (error == OK) {
            error = DisplayChangeField(field, &dfd);
            if (error == OK) {
                ProjectChange(pr, TRUE);
            }
        }

        FreeRecordDescriptionSpec(memPool, rds);
        FreeString(memPool, bubbleHelp);

        if (error) {
            Ask2(go, GCS(MSG_ErrorMessage), AT_ERROR, 
                    GCS(MSG_StarOk), GES(error));
            return;
        }

        DisplayFieldFinish(d->d_StructureEditorData, d->d_DisplayField);

        if (d->d_Alignment) {
            SetPrefsValue(pr, (type == TYPE_INTEGER || type == TYPE_REAL)?
		    PREFS_DEFAULTALIGNMENT_NUMBER : 
		    PREFS_DEFAULTALIGNMENT_TEXT, 
                    dfd.dfd_Alignment);
        }
    }

    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

static void updateFocusWidgets(EditDisplayFieldData *d)
{
    bool readOnly = gtkToggleButtonActive(d->d_ReadOnly);
    bool file = gtkToggleButtonActive(d->d_File);
    bool font = gtkToggleButtonActive(d->d_Font);
    bool list = gtkToggleButtonActive(d->d_Listview);
    bool hide = gtkToggleButtonActive(d->d_HideStringField);
    bool show = gtkToggleButtonActive(d->d_Show);
    bool focus = !hide && (!readOnly || file || font || list || show);

    if (d->d_Shortcut) {
        gtk_widget_set_sensitive(d->d_Shortcut, focus);
    }
    if (d->d_Home) {
	gtk_widget_set_sensitive(d->d_Home, focus);
	if (!focus) {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_Home), false);
	}
    }
    if (d->d_TabChain) {
	gtk_widget_set_sensitive(d->d_TabChain, focus);
	if (!focus) {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_TabChain),
		    false);
	}
    }
}

static void onReadOnlyToggled(EditDisplayFieldData *d)
{
    bool active = gtkToggleButtonActive(d->d_ReadOnly);

    updateFocusWidgets(d);
    if (d->d_SubType) {
        if (active) {
            gtk_combo_box_set_active(GTK_COMBO_BOX(d->d_SubType), 0);
        }
        gtk_widget_set_sensitive(d->d_SubType, !active);
    }
    if (d->d_Calendar) {
        if (active) {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_Calendar),
                    false);
        }
        gtk_widget_set_sensitive(d->d_Calendar, !active);
    }
    if (d->d_ReferencePopupGroup) {
        gtk_widget_set_sensitive(d->d_ReferencePopupGroup, !active);
    }
}

static void onEnabledChanged(EditDisplayFieldData *d)
{
    short enabled = gtkRadioBoxActive(d->d_Enabled);

    gtk_widget_set_sensitive(d->d_EnabledCompute, 
            enabled == DOD_ENABLED_COMPUTED);
}

static void onFFLChanged(GuiObject button, EditDisplayFieldData *d)
{
    bool active = gtkToggleButtonActive(button);

    if (active) {
        if (button != d->d_File) {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_File), false);
        }
        if (button != d->d_Font) {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_Font), false);
        }
        if (button != d->d_Listview) {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_Listview),
                    false);
        }
    }
    if (button == d->d_Listview) {
        gtk_widget_set_sensitive(d->d_ListviewSettings, active);
    }
    updateFocusWidgets(d);
}

static void onOnDoubleClickChanged(EditDisplayFieldData *d)
{
    short action = gtkRadioBoxActive(d->d_OnDoubleClick);

    gtk_widget_set_sensitive(d->d_DoubleClickTrigger,
            action == DAD_DOUBLECLICK_CALLTRIGGER);
}

static void onSubTypeChanged(EditDisplayFieldData *d)
{
    gtk_notebook_set_current_page(GTK_NOTEBOOK(d->d_SubTypeGroup),
	    gtkComboBoxActive(d->d_SubType));
}

static void onListviewKindChanged(EditDisplayFieldData *d)
{
    gtk_notebook_set_current_page(GTK_NOTEBOOK(d->d_ListviewKindGroup),
	    gtkComboBoxActive(d->d_ListviewKind));
}

static void onHideStringFieldChanged(EditDisplayFieldData *d)
{
    bool active = gtkToggleButtonActive(d->d_HideStringField);

    if (active) {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_File), false);
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_Font), false);
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_Listview), false);
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_View), false);
    }
    gtk_widget_set_sensitive(d->d_ReadOnly,!active);
    gtk_widget_set_sensitive(d->d_Alignment, !active);
    gtk_widget_set_sensitive(d->d_Fontname, !active);
    gtk_widget_set_sensitive(d->d_FontnamePopup, !active);
    gtk_widget_set_sensitive(d->d_Background, !active);
    gtk_widget_set_sensitive(d->d_DefaultBackground, !active);
    gtk_widget_set_sensitive(d->d_File, !active);
    gtk_widget_set_sensitive(d->d_Font, !active);
    gtk_widget_set_sensitive(d->d_Listview, !active);
    gtk_widget_set_sensitive(d->d_View, !active);
    updateFocusWidgets(d);
}

static void onDisplayExternalChanged(EditDisplayFieldData *d)
{
    bool active = gtkToggleButtonActive(d->d_DisplayExternal);

    if (!active) {
        gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(d->d_HideStringField), false);
    }
    gtk_widget_set_sensitive(d->d_TitleAtStringField, active);
    gtk_widget_set_sensitive(d->d_HideStringField, active);
    gtk_widget_set_sensitive(d->d_ImageSizeGroup, active);
}

static void defaultBackgroundOnToggled(GtkWidget *def, GtkWidget *bg)
{
    gtk_widget_set_sensitive(bg, !gtkToggleButtonActive(def));
}

static void onRecordsChanged(EditDisplayFieldData *d)
{
    gtk_widget_set_sensitive(d->d_ComputeRecordsGroup,
	    gtkComboBoxActive(d->d_RecordsKind) == DAD_RECORDS_COMPUTED);
}

void GuiDisplayField(StructureEditorData *sed, DisplayObject *dio)
{
    if (!dio) {
        return;
    }
     
    Project *pr = DisplayObjectProject(dio);
    Field *field = DisplayFieldField(dio);
    short type = FieldType(field);
    short subType = DisplayFieldSubType(dio);
    bool readOnly = DisplayFieldReadOnly(dio);
    const char *lvcompute = DisplayFieldListviewCompute(dio);
    bool show_shortcut = true, show_readonly = true,
        show_virtualtype = false, show_buttontype = false,
        show_choicetype = false, show_alignment = false, 
	show_calendar = FALSE, show_format = false, show_numdecimals = false,
        show_weight = true, show_fontname = true, show_bg = true,
        show_stringbuttons = false, show_reference = false, extras = false;

    EditDisplayFieldData *d = g_new0(EditDisplayFieldData, 1);

    d->d_Project = pr;
    d->d_DisplayField = dio;
    d->d_StructureEditorData = sed;

    switch (type)
    {
        case TYPE_BOOL:
            show_fontname = false;
            break;

        case TYPE_CHOICE:
            show_choicetype = true;
            break;

        case TYPE_REAL:
            show_numdecimals = true;
            show_alignment = true;
            break;

        case TYPE_INTEGER:
            show_alignment = true;
            break;

        case TYPE_DATE:
            show_alignment = true;
            show_calendar = true;
            break;

        case TYPE_TIME:
            show_alignment = true;
	    show_format = true;
            break;

        case TYPE_STRING:
            show_alignment = true;
            show_stringbuttons = true;
            break;

        case TYPE_REFERENCE:
            show_alignment = true;
            show_reference = true;
            break;

        case TYPE_BUTTON:
            show_readonly = false;
            show_buttontype = true;
            break;

        case TYPE_VIRTUAL:
            show_readonly = false;
            show_shortcut = false;
            show_virtualtype = true;
            show_alignment = true;
            show_numdecimals = true;
            break;
    }
    extras = show_stringbuttons || show_reference || show_virtualtype;

    d->d_Title = EntryObject,
        "max-length", 255,
        "xalign", 0.5,
        "text", EmptyNullStr(DisplayObjectTitle(dio)),
        End;
    GuiSetTooltip(d->d_Title, GCS(MSG_DisplayField_TitleBH));

    GuiObject titlePositionLabel = 0;
    GuiObject titlePositionBox = 0;
    GuiObject titleFontnameLabel = 0;
    if (!show_buttontype) {
        static char *titlepos[5];
        titlepos[0] = GCS(MSG_DisplayField_Alignment_Left);
        titlepos[1] = GCS(MSG_DisplayField_Alignment_Right);
        titlepos[2] = GCS(MSG_DisplayField_Alignment_Top);
        titlepos[3] = GCS(MSG_DisplayField_Alignment_Bottom);
        titlepos[4] = 0;
        d->d_TitlePosition = GuiComboBox(titlepos, 
                DisplayFieldTitlePosition(dio));
        titlePositionBox = GuiMakeTooltip(d->d_TitlePosition, 
                GCS(MSG_DisplayField_TitlePositionBH));

        titlePositionLabel = GuiLabel(GCS(MSG_DisplayField_TitlePosition),
                d->d_TitlePosition);

        d->d_TitleFontname = EntryObject,
            "max-length", FONTNAME_LEN,
            "xalign", 0.5,
            "text", EmptyNullStr(DisplayFieldTitleGTKFontname(dio)),
            End;
        GuiSetTooltip(d->d_TitleFontname, GCS(MSG_FontBH));

        titleFontnameLabel = GuiLabel(GCS(MSG_Font), d->d_TitleFontname);

        d->d_TitleFontnamePopup = PopSelectionObject,
	    "selection-type", GUI_TYPE_FONT_CHOOSER_DIALOG,
	    "string", d->d_TitleFontname,
	    "button", GuiPopButton(SELECT_FONT_ICON_NAME), 
	    End;
    }

    GuiObject shortcutLabel = 0;
    GuiObject shortcutGroup = 0;
    if (show_shortcut) {
	char sc[CodesetMaxBytesForChar+1];
	CodesetCharToStr(DisplayObjectShortcut(dio), sc);
        d->d_Shortcut = EntryObject,
            "max-length", 1,
            "width-chars", 3,
	    "xalign", 0.5,
            "text", sc,
	    End;
        GuiSetTooltip(d->d_Shortcut, GCS(MSG_DisplayObject_ShortcutBH));
        g_signal_connect(d->d_Shortcut, "activate", 
                G_CALLBACK(ShortCutOnActivate), dio);

        shortcutLabel = GuiLabel(GCS(MSG_DisplayObject_Shortcut),d->d_Shortcut);

        shortcutGroup = HGroup,
            "child", d->d_Shortcut,
            "child", SpaceObject, End,
            End;

        d->d_Home = CheckButtonObject,
            "label", GCS(MSG_DisplayField_Home),
            "use-underline", true,
            "active", DisplayFieldHome(dio),
           End;
        GuiSetTooltip(d->d_Home, GCS(MSG_DisplayField_HomeBH));
        g_object_set(G_OBJECT(shortcutGroup), "child", d->d_Home, NULL);

        d->d_TabChain = CheckButtonObject,
            "label", GCS(MSG_DisplayField_TabChain),
            "use-underline", true,
            "active", DisplayFieldTabChain(dio),
           End;
        GuiSetTooltip(d->d_TabChain, GCS(MSG_DisplayField_TabChainBH));
        g_object_set(G_OBJECT(shortcutGroup), "child", d->d_TabChain, NULL);

        if (show_readonly) {
            d->d_ReadOnly = CheckButtonObject, 
                "label", GCS(MSG_DisplayField_ReadOnly),
                "use-underline", true,
                End;
	    GuiSetTooltip(d->d_ReadOnly, show_stringbuttons?
		    GCS(MSG_DisplayField_ReadOnlyStringBH) :
		    GCS(MSG_DisplayField_ReadOnlyBH));
            g_object_set(G_OBJECT(shortcutGroup), "child", d->d_ReadOnly, NULL);
        }
    }

    GuiObject subtypeLabel = 0;
    GuiObject subtypeBox = 0;
    if (show_virtualtype) {
        static char *vttext[4];
        vttext[0] = GCS(MSG_Boolean);
        vttext[1] = GCS(MSG_Text);
        vttext[2] = GCS(MSG_DisplayField_Virtual_Kind_List);
        vttext[3] = 0;
        d->d_SubType = GuiComboBox(vttext, subType);
        subtypeBox = GuiMakeTooltip(d->d_SubType,
                GCS(MSG_DisplayField_Virtual_KindBH));

        subtypeLabel = GuiLabel(GCS(MSG_DisplayField_Kind),d->d_SubType);
    }

    if (show_buttontype) {
        static char *bttext[3];
        bttext[0] = GCS(MSG_DisplayField_Button_Kind_TextButton);
        bttext[1] = GCS(MSG_DisplayField_Button_Kind_ImageButton);
        bttext[2] = 0;
        d->d_SubType = GuiComboBox(bttext, subType);
        subtypeBox = GuiMakeTooltip(d->d_SubType,
                GCS(MSG_DisplayField_Button_KindBH));

        subtypeLabel = GuiLabel(GCS(MSG_DisplayField_Kind), d->d_SubType);
    }

    if (show_choicetype) {
        static char *cttext[3];
        cttext[0] = GCS(MSG_DisplayField_Choice_Kind_CycleButton);
        cttext[1] = GCS(MSG_DisplayField_Choice_Kind_RadioButtons);
        cttext[2] = 0;
        d->d_SubType = GuiComboBox(cttext, subType);
        gtk_widget_set_sensitive(d->d_SubType, !readOnly);
        subtypeBox = GuiMakeTooltip(d->d_SubType,
                GCS(MSG_DisplayField_Choice_KindBH));

        subtypeLabel = GuiLabel(GCS(MSG_DisplayField_Kind), d->d_SubType);

        GuiObject page0 = HGroup,
            "child", titlePositionLabel, "weight", 0,
            "child", titlePositionBox, "weight", 1,
            "child", SpaceObject, End, "weight", 0,
            "child", titleFontnameLabel, "weight", 0,
            "child", d->d_TitleFontnamePopup, "weight", 2,
            End;

        d->d_Frame = CheckButtonObject,
            "active", DisplayFieldFrame(dio),
            "label", GCS(MSG_DisplayField_Choice_Frame),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_Frame, GCS(MSG_DisplayField_Choice_FrameBH));

        d->d_Horizontal = CheckButtonObject,
            "active", DisplayFieldAlignment(dio)==DAD_ALIGNMENT_HORIZONTAL,
            "label", GCS(MSG_DisplayField_Choice_Horizontal),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_Horizontal,GCS(MSG_DisplayField_Choice_HorizontalBH));

        GuiObject page1 = HGroup,
            "child", d->d_Frame,
            "child", d->d_Horizontal,
            End;

        d->d_SubTypeGroup = NotebookObject,
            "show-tabs", false,
            "show-border", false,
            "child", GroupAlignment(false),
                "child", VCenter(page0),
                End,
            "child", GroupAlignment(false),
                "child", VCenter(page1),
                End,
            End;
    }

    GuiObject alignmentLabel = 0;
    GuiObject alignmentBox = 0;
    if (show_alignment) {
        static char *alignmentText[4];
	alignmentText[0] = GCS(MSG_DisplayField_Alignment_Center);
	alignmentText[1] = GCS(MSG_DisplayField_Alignment_Left);
	alignmentText[2] = GCS(MSG_DisplayField_Alignment_Right);
	alignmentText[3] = 0;
        d->d_Alignment = GuiComboBox(alignmentText, 
		DisplayFieldAlignment(dio));
        alignmentBox = GuiMakeTooltip(d->d_Alignment, 
		GCS(MSG_DisplayField_AlignmentBH));
        alignmentLabel = GuiLabel(GCS(MSG_DisplayField_Alignment), 
		d->d_Alignment);
    }

    if (show_calendar) {
        d->d_Calendar = CheckButtonObject,
            "active", DisplayFieldCalendar(dio),
            "label", GCS(MSG_DisplayField_Calendar),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_Calendar, GCS(MSG_DisplayField_CalendarBH));
    }

    GuiObject formatLabel = 0;
    GuiObject formatBox = 0;
    if (show_format) {
        static char *formatText[4];
	formatText[0] = GCS(MSG_DisplayField_Format_HHMMSS);
	formatText[1] = GCS(MSG_DisplayField_Format_MMSS);
	formatText[2] = GCS(MSG_DisplayField_Format_HHMM);
	formatText[3] = 0;
        d->d_Format = GuiComboBox(formatText, DisplayFieldFormat(dio));
        formatBox = GuiMakeTooltip(d->d_Format, GCS(MSG_DisplayField_FormatBH));
        formatLabel = GuiLabel(GCS(MSG_DisplayField_Format), d->d_Format);
    }

    GuiObject numDecimalsLabel = 0;
    if (show_numdecimals) {
        gint val = DisplayFieldNumDecimals(dio);
        d->d_NumDecimals = SpinButtonObject,
            "adjustment", gtk_adjustment_new(val, 0, 9, 1, 1, 0),
            "xalign", 0.5,
            "digits", 0,
            "numeric", true,
            "snap-to-ticks", false,
            "update-policy", GTK_UPDATE_IF_VALID,
            "wrap", false,
            "max-length", 1,
            "width-chars", 3,
	    End;
        GuiSetTooltip(d->d_NumDecimals, GCS(MSG_DisplayField_NumDecimalsBH));

	numDecimalsLabel = GuiLabel(GCS(MSG_DisplayField_NumDecimals), 
                d->d_NumDecimals);
    }

    GuiObject extrasGroup = 0;
    if (show_virtualtype) {
        d->d_ShowTitles = CheckButtonObject, 
            "active", DisplayFieldShowTitles(dio),
            "label", GCS(MSG_DisplayField_ShowTitles),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_ShowTitles, GCS(MSG_DisplayField_ShowTitlesBH));

        d->d_TabChain = CheckButtonObject,
            "label", GCS(MSG_DisplayField_TabChain),
            "use-underline", true,
            "active", DisplayFieldTabChain(dio),
           End;
        GuiSetTooltip(d->d_TabChain, GCS(MSG_DisplayField_TabChainBH));

        d->d_ControlTable = CheckButtonObject,
            "label", GCS(MSG_DisplayField_ControlTable),
            "use-underline", true,
            "active", DisplayFieldControlTable(dio),
           End;
        GuiSetTooltip(d->d_ControlTable, GCS(MSG_DisplayField_ControlTableBH));

	static char *doubleClickText[4];
	doubleClickText[0] = GCS(MSG_DisplayField_DoubleClick_DoNothing);
	doubleClickText[1] = GCS(MSG_DisplayField_DoubleClick_ShowRecord);
	doubleClickText[2] = GCS(MSG_DisplayField_DoubleClick_CallTrigger);
	doubleClickText[3] = 0;
	d->d_OnDoubleClick = RadioBoxObject,
	    "entries", doubleClickText,
	    End;
        GuiObject onDoubleClickBox = GuiMakeTooltip(d->d_OnDoubleClick,
                GCS(MSG_DisplayField_OnDoubleClickBH));

	d->d_DoubleClickTriggerStr = EntryObject,
	    "max-length", FIELD_TRIGGER_LEN,
	    "xalign", 0.5,
	    "text", EmptyNullStr(DisplayFieldDoubleClickTrigger(dio)),
	    End;
        GuiSetTooltip(d->d_DoubleClickTriggerStr, 
                GCS(MSG_DisplayField_DoubleClick_TriggerBH));

        d->d_DoubleClickTrigger = UserFunctionsPopup(
                d->d_DoubleClickTriggerStr, pr);
        gtk_widget_set_sensitive(d->d_DoubleClickTrigger, false);

        GuiObject dctLabel = GuiLabel(GCS(MSG_DisplayField_DoubleClick_Trigger),
                d->d_DoubleClickTriggerStr, 0.5);

        g_signal_connect_swapped(d->d_OnDoubleClick, "active-changed",
            G_CALLBACK(onOnDoubleClickChanged), d);
        gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_OnDoubleClick),
                DisplayFieldOnDoubleClick(dio));

	d->d_URLDropTriggerStr = EntryObject,
	    "max-length", FIELD_TRIGGER_LEN,
	    "xalign", 0.5,
	    "text", EmptyNullStr(DisplayFieldURLDropTrigger(dio)),
	    End;
        GuiSetTooltip(d->d_URLDropTriggerStr, 
                GCS(MSG_DisplayField_URLDrop_TriggerBH));

        d->d_URLDropTrigger = UserFunctionsPopup(d->d_URLDropTriggerStr, pr);

        GuiObject udtLabel = GuiLabel(GCS(MSG_DisplayField_URLDrop_Trigger),
                d->d_URLDropTriggerStr);

	d->d_SortDropTriggerStr = EntryObject,
	    "max-length", FIELD_TRIGGER_LEN,
	    "xalign", 0.5,
	    "text", EmptyNullStr(DisplayFieldSortDropTrigger(dio)),
	    End;
        GuiSetTooltip(d->d_SortDropTriggerStr, 
                GCS(MSG_DisplayField_SortDrop_TriggerBH));

        d->d_SortDropTrigger = UserFunctionsPopup(d->d_SortDropTriggerStr, pr);

        GuiObject sdtLabel = GuiLabel(GCS(MSG_DisplayField_SortDrop_Trigger),
                d->d_SortDropTriggerStr);

        GuiObject page0 = SpaceObject, End;

        GuiObject page1 = VGroup,
            "columns", 3,
            "child", alignmentLabel, "weight", 0,
            "child", alignmentBox,
            "child", SpaceObject, End,
            "child", numDecimalsLabel, "weight", 0,
            "child", d->d_NumDecimals, 
            "child", SpaceObject, End,
            End;
        show_alignment = false;

        GuiObject page2 = VGroup,
            "child", d->d_ShowTitles,
            "child", d->d_TabChain,
            "child", d->d_ControlTable,
	    "child", Frame(GCS(MSG_DisplayField_OnDoubleClick)),
                "child", GroupAlignment(true),
                    "child", HGroup,
                        "child", onDoubleClickBox, "weight", 0,
                        "child", VSeparatorObject, End, "weight", 0,
                        "child", VGroup,
                            "child", SpaceObject, End,
                            "child", dctLabel, "weight", 0,
                            "child", d->d_DoubleClickTrigger, "weight", 0,
                            End,
                        End,
                    End,
                End,
            "child", VGroup,
                "columns", 2,
                "child", udtLabel, "weight", 0,
                "child", d->d_URLDropTrigger,
                "child", sdtLabel, "weight", 0,
                "child", d->d_SortDropTrigger,
                End,
            End;


        d->d_SubTypeGroup = NotebookObject,
            "show-tabs", false,
            "show-border", false,
            "child", GroupAlignment(false),
                "child", VCenter(page0),
                End,
            "child", GroupAlignment(false),
                "child", VCenter(page1),
                End,
            "child", GroupAlignment(false),
                "child", VCenter(page2),
                End,
            End;

        d->d_AutoUpdate = CheckButtonObject,
            "label", GCS(MSG_DisplayField_AutoUpdate),
            "use-underline", TRUE,
            "active", DisplayFieldAutoUpdate(dio),
           End;
        GuiSetTooltip(d->d_AutoUpdate, GCS(MSG_DisplayField_AutoUpdateBH));

        extrasGroup = VGroup,
            "child", HGroup,
                "child", subtypeLabel, "weight", 0,
                "child", subtypeBox,  "vert-weight", 0,
                End, "vert-weight", 0,
            "child", Frame(0), 
                "child", VCenter(d->d_SubTypeGroup),
                End,
            "child", d->d_AutoUpdate, "vert-weight", 0,
            End;
    }

    static const char *etext[4];
    etext[0] = GCS(MSG_DisplayObject_Enabled_Always);
    etext[1] = GCS(MSG_DisplayObject_Enabled_NotInitial);
    etext[2] = GCS(MSG_DisplayObject_Enabled_Computed);
    etext[3] = 0;
    d->d_Enabled = RadioBoxObject,
        "entries", etext,
        End;
    GuiObject enabledBox = GuiMakeTooltip(d->d_Enabled,
            GCS(MSG_DisplayObject_EnabledBH));

    d->d_EnabledComputeStr = EntryObject,
        "max-length", FIELD_TRIGGER_LEN,
        "xalign", 0.5,
        "text", EmptyNullStr(DisplayObjectEnabledCompute(dio)),
        End;
    GuiSetTooltip(d->d_EnabledComputeStr, 
            GCS(MSG_DisplayObject_ComputeEnabledBH));

    d->d_EnabledCompute = UserFunctionsPopup(d->d_EnabledComputeStr, pr);
    gtk_widget_set_sensitive(d->d_EnabledCompute, false);

    GuiObject ceLabel = GuiLabel(GCS(MSG_Compute), 
            d->d_EnabledCompute, 0.5);

    g_signal_connect_swapped(d->d_Enabled, "active-changed",
            G_CALLBACK(onEnabledChanged), d);
    gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_Enabled),
            DisplayObjectEnabled(dio));

    GuiObject enabledGroup = HGroup,
        "child", enabledBox, "weight", 0,
        "child", VSeparatorObject, End, "weight", 0,
        "child", VGroup,
            "child", SpaceObject, End,
            "child", ceLabel, "weight", 0,
            "child", d->d_EnabledCompute, "weight", 0,
            End,
        End;
    enabledGroup = Frame(GCS(MSG_DisplayObject_Enabled)),
        "child", GroupAlignment(true),
            "child", enabledGroup,
            End,
        End;

    GuiObject weightLabel = 0;
    GuiObject weightBox = 0;
    if (show_weight) {
        int weight = DisplayObjectWeight(dio);
        d->d_Weight = HScaleObject,
            "adjustment", gtk_adjustment_new(weight, 0, 1000, 1, 10, 0),
            "value-pos", GTK_POS_RIGHT,
            "digits", 0,
            End;
        weightBox = GuiMakeTooltip(d->d_Weight, GCS(MSG_WeightBH));
	weightLabel = GuiLabel(GCS(MSG_Weight), d->d_Weight);
    }

    GuiObject fontnameLabel = 0;
    if (show_fontname) {
        d->d_Fontname = EntryObject,
            "max-length", FONTNAME_LEN,
            "xalign", 0.5,
            "text", EmptyNullStr(DisplayObjectGTKFontname(dio)),
            End;
        GuiSetTooltip(d->d_Fontname, GCS(MSG_FontBH));

        fontnameLabel = GuiLabel(GCS(MSG_Font), d->d_Fontname);

        d->d_FontnamePopup = PopSelectionObject,
	    "selection-type", GUI_TYPE_FONT_CHOOSER_DIALOG,
	    "string", d->d_Fontname,
	    "button", GuiPopButton(SELECT_FONT_ICON_NAME), 
	    End;
    }

    GuiObject bgLabel = 0;
    GuiObject bgGroup = 0;
    if (show_bg) {
        const char *bg = DisplayObjectBackground(dio);
        if (show_buttontype && subType != DAD_BUTTONTYPE_TEXT) {
            bg = IMAGESPEC_START;
        }
        ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
        d->d_Background = PopImageObject,
            "spec", (bg && bg[0])? bg : IMAGESPEC_START,
            "types", GTK_POP_IMAGE_COLOR,
            End;
        g_signal_connect_swapped(d->d_Background, "clicked",
                G_CALLBACK(ProjectSetCurrentDir), pr);
        GuiObject backgroundBox = GuiMakeTooltip(d->d_Background,
                GCS(MSG_BackgroundBH));

        bgLabel = GuiLabel(GCS(MSG_Background), d->d_Background);

        d->d_DefaultBackground = CheckButtonObject, 
            "label", GCS(MSG_Background_Default),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_DefaultBackground, GCS(MSG_Background_DefaultBH));
        g_signal_connect(d->d_DefaultBackground, "toggled", 
                G_CALLBACK(defaultBackgroundOnToggled), d->d_Background);
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_DefaultBackground),
                !bg || !bg[0]);

        bgGroup = HGroup,
            "child", backgroundBox, "weight", 400,
            "child", SpaceObject, End,
            "child", VCenter(d->d_DefaultBackground),
            End;
    }

    if (show_buttontype) {
        const char *bg = DisplayObjectBackground(dio);
        if (subType != DAD_BUTTONTYPE_IMAGE) {
            bg = IMAGESPEC_START;
        }

        ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
        d->d_Image = PopImageObject,
            "spec", (bg && bg[0])? bg : IMAGESPEC_START,
            // MUIA_Imageadjust_Type, MUIV_Imageadjust_Type_All,
            End;
        g_signal_connect_swapped(d->d_Image, "clicked",
                G_CALLBACK(ProjectSetCurrentDir), pr);
        GuiSetTooltip(d->d_Image, GCS(MSG_ImageBH));

        static char *entries[3];
        entries[0] = GCS(MSG_Image_Size_Resizeable);
        entries[1] = GCS(MSG_Image_Size_Fixed);
        entries[2] = 0;
        d->d_ImageSize = RadioBoxObject,
            "entries", entries,
            "active", DisplayFieldImageSize(dio),
            End;
        d->d_ImageSizeGroup = Frame(GCS(MSG_Size)),
            "child", GroupAlignment(true),
                "child", VCenter(d->d_ImageSize),
                End,
            End;

        GuiObject page0 =  VGroup,
            "columns", 2,
            ChildLabelWidget(GCS(MSG_Title), d->d_Title), "vert-weight", 0,
            "child", fontnameLabel, "weight", 0,
            "child", d->d_FontnamePopup, "vert-weight", 0,
            "child", bgLabel, "weight", 0,
            "child", bgGroup,
            End;
        show_fontname = false;          // only add once
        show_bg = false;                // only add once

        GuiObject page1 = HGroup,
            ChildLabelWidget(GCS(MSG__Image), d->d_Image),
            "child", d->d_ImageSizeGroup, "weight", 0,
            End;

        d->d_SubTypeGroup = NotebookObject,
            "show-tabs", false,
            "show-border", false,
            "child", GroupAlignment(false),
                "child", page0,
                End,
            "child", GroupAlignment(false),
                "child", page1,
                End,
            End;
    }

#if 0
    if(show_fg)
    {
        struct Window *win = GuiAnyOpenedWindow(app);
        struct Screen *screen = win? win->WScreen : 0;
        struct DrawInfo *dri = screen? GetScreenDrawInfo(screen) : 0;
        static UBYTE pens[8];
        short i;

        if(dri)
        {
            for(i=0; i < 7; i++)
                pens[i] = dri->dri_Pens[i+2];
            if(dri->dri_NumPens > 9)
                pens[i] = dri->dri_Pens[i+2];
            else
                pens[i] = dri->dri_Pens[0];

            FreeScreenDrawInfo(screen, dri);

            for(i=0; i < 8; i++)
                printf("pens[%d] = %d\n", i+2, pens[i]);
        }

        fg_label = VCenter(Label(GCS(MSG_DisplayField_InitialColor)));
        d->d_InitialColor = ColorButtonObject,
            MUIA_CycleChain, true,
            MUIA_ShortHelp, GCS(MSG_DisplayField_InitialColorBH),
            MUIA_ControlChar, Shortcut(GCS(MSG_DisplayField_InitialColor)),
            MUIA_ColorButton_Pens, 0,
            MUIA_ColorButton_Pen, 0,
            MUIA_ColorButton_Max, 8,
            End;
    }
#endif

    if (show_stringbuttons) {
        d->d_File = CheckButtonObject, 
            "active", DisplayFieldFile(dio),
            "label", GCS(MSG_DisplayField_File),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_File, GCS(MSG_DisplayField_FileBH));
        g_signal_connect(d->d_File, "toggled", G_CALLBACK(onFFLChanged), d);

        d->d_Font = CheckButtonObject,
            "active", DisplayFieldFont(dio),
            "label", GCS(MSG_DisplayField_Font),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_Font, GCS(MSG_DisplayField_FontBH));
        g_signal_connect(d->d_Font, "toggled", G_CALLBACK(onFFLChanged), d);

        d->d_Listview = CheckButtonObject, 
            "active", DisplayFieldListview(dio),
            "label", GCS(MSG_DisplayField_Listview),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_Listview, GCS(MSG_DisplayField_ListviewBH));
        g_signal_connect(d->d_Listview, "toggled", G_CALLBACK(onFFLChanged), d);

        d->d_View = CheckButtonObject,
            "active", DisplayFieldView(dio),
            "label", GCS(MSG_DisplayField_View),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_View, GCS(MSG_DisplayField_ViewBH));

        GuiObject popGroup = VGroup,
            "child", d->d_File,
            "child", d->d_Font,
            "child", d->d_Listview,
            "child", d->d_View,
            End;

	static char *lktext[3];
        lktext[0] = GCS(MSG_DisplayField_Labels_Static);
        lktext[1] = GCS(MSG_Computed);
        lktext[2] = 0;
        d->d_ListviewKind = GuiComboBox(lktext, 0);
        GuiObject listviewKindBox = GuiMakeTooltip(d->d_ListviewKind,
		GCS(MSG_DisplayField_LabelsBH));
        GuiObject listviewKindLabel = GuiLabel(GCS(MSG_DisplayField_Labels), 
		d->d_ListviewKind);

        d->d_ListviewLabels = PopEditLabelsObject,
	    "label", GCS(MSG_DisplayField__EditLabels),
            "use-underline", true,
            "labels", DisplayFieldListviewLabels(dio),
            End;
        GuiSetTooltip(d->d_ListviewLabels, GCS(MSG_DisplayField_EditLabelsBH));

	d->d_ListviewComputeStr = EntryObject,
	    "max-length", FIELD_TRIGGER_LEN,
	    "width-chars", 10,
	    "xalign", 0.5,
	    "text", EmptyNullStr(lvcompute),
	    End;
        GuiSetTooltip(d->d_ListviewComputeStr, 
                GCS(MSG_DisplayField_ComputeLabelsBH));

        d->d_ListviewCompute = UserFunctionsPopup(d->d_ListviewComputeStr, pr);

	d->d_ListviewKindGroup = NotebookObject,
	    "show-tabs", false,
	    "show-border", false,
	    "child", VCenter(d->d_ListviewLabels),
	    "child", VGroup,
		"child", GuiLabel(GCS(MSG_Compute),
                        d->d_ListviewCompute, 0.5),
                "child", d->d_ListviewCompute,
                End,
	    End;

	d->d_ListviewSettings = VGroup,
            "child", HGroup,
                "child", listviewKindLabel, "weight", 0,
                "child", listviewKindBox,
                End, "vert-weight", 0,
            "child", HSeparatorObject, End,
            "child", VCenter(d->d_ListviewKindGroup),
	    End;
	d->d_ListviewSettings = Frame(GCS(MSG_DisplayField_ListviewSettings)),
            "sensitive", DisplayFieldListview(dio),
	    "child", GroupAlignment(true),
	        "child", VCenter(d->d_ListviewSettings),
                End,
	    End;

	g_signal_connect_swapped(d->d_ListviewKind, "changed",
		G_CALLBACK(onListviewKindChanged), d);
        popGroup = HGroup,
            "child", VCenter(HCenter(popGroup)),
            "child", d->d_ListviewSettings,
            End;
        GuiObject bottom = Frame(GCS(MSG_DisplayField_Buttons)),
            "child", GroupAlignment(true),
                "child", popGroup,
                End,
            End;

        d->d_DisplayExternal = CheckButtonObject,
            "active", true, 
            "label", GCS(MSG_DisplayField_DisplayPicture),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_DisplayExternal,
                GCS(MSG_DisplayField_DisplayPictureBH));

        d->d_HideStringField = CheckButtonObject,
            "label", GCS(MSG_DisplayField_HideStringField),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_HideStringField,
                GCS(MSG_DisplayField_HideStringFieldBH));
        g_signal_connect_swapped(d->d_HideStringField, "toggled",
                G_CALLBACK(onHideStringFieldChanged), d);

        d->d_TitleAtStringField = CheckButtonObject,
            "label", GCS(MSG_DisplayField_TitleAtStringField),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_TitleAtStringField, 
                GCS(MSG_DisplayField_TitleAtStringFieldBH));

        static char *entries[6];
        entries[0] = GCS(MSG_Image_Size_Resizeable);
        entries[1] = GCS(MSG_Image_Size_Fixed);
        entries[2] = GCS(MSG_Image_Size_Scrollable);
        entries[3] = GCS(MSG_Image_Size_Scaled);
        entries[4] = GCS(MSG_Image_Size_AspectScaled);
        entries[5] = 0;
        d->d_ImageSize = RadioBoxObject,
            "entries", entries,
            "active", DisplayFieldImageSize(dio),
            End;
        GuiObject imageSizeBox = GuiMakeTooltip(d->d_ImageSize,
                GCS(MSG_DisplayField_SizeBH));
        d->d_ImageSizeGroup = Frame(GCS(MSG_Size)),
            "child", GroupAlignment(true),
                "child", VCenter(HCenter(imageSizeBox)),
                End,
            End;

        GuiObject flagsGroup = VGroup,
            "child", d->d_DisplayExternal,
            "child", d->d_TitleAtStringField,
            "child", d->d_HideStringField,
            End;

        g_signal_connect_swapped(d->d_DisplayExternal, "toggled",
                G_CALLBACK(onDisplayExternalChanged), d);

        gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(d->d_TitleAtStringField),
                DisplayFieldTitleAtStringField(dio));
        gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(d->d_HideStringField),
                DisplayFieldHideStringField(dio));
        gtk_toggle_button_set_active(
                GTK_TOGGLE_BUTTON(d->d_DisplayExternal),
                DisplayFieldDisplayExternal(dio));

        GuiObject top = Frame(GCS(MSG_DisplayField_Picture)),
            "child", GroupAlignment(true),
                "child", HGroup,
                    "child", VCenter(HCenter(flagsGroup)),
                    "child", d->d_ImageSizeGroup,
                    End,
                End,
            End;

        extrasGroup = VGroup,
            "child", top,
            "child", bottom,
            End;
    } else if (show_reference) {
        d->d_Show = CheckButtonObject,
            "active", DisplayFieldShow(dio),
            "label", GCS(MSG_DisplayField_Show),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_Show, GCS(MSG_DisplayField_ShowBH));
        g_signal_connect_swapped(d->d_Show, "toggled",
                G_CALLBACK(updateFocusWidgets), d);
            
        d->d_AutoShow = CheckButtonObject,
            "active", DisplayFieldAutoShow(dio),
            "label", GCS(MSG_AutoShow),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_AutoShow, GCS(MSG_DisplayField_AutoShowBH));

        d->d_Filter = CheckButtonObject,
            "active", DisplayFieldFilter(dio),
            "label", GCS(MSG_DisplayField_Filter),
            "use-underline", true,
            End;
        GuiSetTooltip(d->d_Filter, GCS(MSG_DisplayField_FilterBH));

        GuiObject buttons = VGroup,
            "child", d->d_Show,
            "child", d->d_AutoShow,
            "child", d->d_Filter,
            End;
        buttons = HCenter(buttons);
        buttons = VCenter(buttons);
        buttons = Frame(GCS(MSG_DisplayField_Buttons)),
            "child", GroupAlignment(true),
                "child", buttons,
                End,
            End;

        d->d_RecordDescriptionSpec = CreateEditRecordDescriptionSpec(
                FieldReferenceTable(field), 
                DisplayFieldRecordDescriptionSpec(dio, FALSE), TRUE);

	d->d_RecordDescriptionSpecGroup = Frame(GCS(MSG_Display)),
            "child", GroupAlignment(true),
                "child", d->d_RecordDescriptionSpec,
		End,
	    End;

        static const char *rectext[4];
        rectext[0] = GCS(MSG_DisplayField_Records_All);
        rectext[1] = GCS(MSG_DisplayField_Records_MatchFilter);
        rectext[2] = GCS(MSG_Computed);
	rectext[3] = 0;
	d->d_RecordsKind = GuiComboBox(rectext, 0);
        GuiObject recordsBox = GuiMakeTooltip(d->d_RecordsKind,
		GCS(MSG_DisplayField_RecordsBH));
        GuiObject recordsLabel = GuiLabel(GCS(MSG_Records), d->d_RecordsKind);

	d->d_ComputeRecordsStr = EntryObject,
	    "max-length", FIELD_TRIGGER_LEN,
	    "xalign", 0.5,
	    "text", EmptyNullStr(DisplayFieldRecordsCompute(dio)),
	    End;
        GuiSetTooltip(d->d_ComputeRecordsStr,
		GCS(MSG_DisplayField_ComputeRecordsBH));

        d->d_ComputeRecords = UserFunctionsPopup(d->d_ComputeRecordsStr, pr);

	d->d_ComputeRecordsGroup = Frame(GCS(MSG_Compute)),
            "child", GroupAlignment(true),
                "child", d->d_ComputeRecords,
		End,
	    End;
	gtk_widget_set_sensitive(d->d_ComputeRecordsGroup, false);
	g_signal_connect_swapped(d->d_RecordsKind, "changed",
		G_CALLBACK(onRecordsChanged), d);
	g_object_set(G_OBJECT(d->d_RecordsKind), 
		"active", DisplayFieldRecordsKind(dio), NULL);

	d->d_InitialRecord = CheckButtonObject,
            "label", GCS(MSG_DisplayField_InitialRecord),
            "use-underline", true,
            "active", DisplayFieldInitialRecord(dio),
           End;
        GuiSetTooltip(d->d_InitialRecord, 
		GCS(MSG_DisplayField_InitialRecordBH));

	d->d_CurrentRecord = CheckButtonObject,
            "label", GCS(MSG_DisplayField_CurrentRecord),
            "use-underline", true,
            "active", DisplayFieldCurrentRecord(dio),
           End;
        GuiSetTooltip(d->d_CurrentRecord, 
		GCS(MSG_DisplayField_CurrentRecordBH));

	d->d_ReferencePopupGroup = VGroup,
	    "child", HGroup,
	        "child", recordsLabel, "weight", 0,
		"child", recordsBox,
		End,
	    "child", d->d_ComputeRecordsGroup,
	    "child", d->d_InitialRecord,
	    "child", d->d_CurrentRecord,
	    End;
	d->d_ReferencePopupGroup = VCenter(d->d_ReferencePopupGroup);
        d->d_ReferencePopupGroup = Frame(GCS(MSG_Popup)),
            "child", GroupAlignment(true),
                "child", d->d_ReferencePopupGroup,
                End,
            End;

        extrasGroup = HGroup,
	    "same-width", TRUE,
	    "child", d->d_RecordDescriptionSpecGroup,
            "child", VGroup,
                "child", d->d_ReferencePopupGroup, "weight", 5,
                "child", buttons, "weight", 3,
                End,
            End;
    }

    GuiObject alignmentGroup = 0;
    if (show_alignment) {
        GuiObject right;
        if (show_numdecimals) {
            right = HGroup,
                "child", numDecimalsLabel, "weight", 0,
                "child", d->d_NumDecimals, 
                End;
        } else if (show_calendar) {
            right = HGroup,
                "child", SpaceObject, End,
                "child", d->d_Calendar,
                End;
        } else if (show_format) {
            right = HGroup,
                "child", formatLabel, "weight", 0,
                "child", formatBox,
                End;
        } else {
            right = SpaceObject, End;
        }
        alignmentGroup = HGroup,
            "child", alignmentBox,
            "child", right,
            End;
    }

    GuiObject topGroup = 0;
    if (show_buttontype) {
        topGroup = VGroup,
            "child", HGroup,
                "child", subtypeLabel, "weight", 0,
                "child", subtypeBox,  "vert-weight", 0,
                End, "vert-weight", 0,
            "child", Frame(0), 
                "child", d->d_SubTypeGroup,
                End,
           End;
    } else if(show_choicetype) {
        topGroup = VGroup,
            "child", HGroup,
                "child", subtypeLabel, "weight", 0,
                "child", subtypeBox,  "vert-weight", 0,
                End, "vert-weight", 0,
            "child", Frame(GCS(MSG_DisplayField_Title)),
                "child", GroupAlignment(true),
                    "child", VGroup,
                        "child", HGroup,
                            ChildLabelWidget(GCS(MSG_Title), d->d_Title),
                            "vert-weight", 0,
                            End,
                        "child", d->d_SubTypeGroup,
                        End,
                    End,
                End, "vert-weight", 0,
            End;
    } else {
        topGroup = VGroup,
            "columns", 2,
            ChildLabelWidget(GCS(MSG_Title), d->d_Title), "vert-weight", 0,
            "child", titlePositionLabel, "weight", 0,
            "child", HGroup,
                "child", titlePositionBox, "weight", 1,
                "child", SpaceObject, End, "weight", 0,
                "child", titleFontnameLabel, "weight", 0,
                "child", d->d_TitleFontnamePopup, "weight", 2,
                End, "vert-weight", 0,
            End;
        topGroup = VGroup,
            "child", Frame(GCS(MSG_DisplayField_Title)),
                "child", GroupAlignment(true),
                    "child", topGroup,
                    End,
                End, "vert-weight", 0,
            End;
    }

    GuiObject fieldGroup = VGroup, "columns", 2, End;
    if (show_shortcut) {
        g_object_set(G_OBJECT(fieldGroup),
           "child", shortcutLabel, "weight", 0,
           "child", shortcutGroup, "vert-weight", 0,
           NULL);
    }
    if (show_alignment) {
        g_object_set(G_OBJECT(fieldGroup), 
            "child", alignmentLabel, "weight", 0,
            "child", alignmentGroup, "vert-weight", 0,
            NULL);
    }
    if (show_weight) {
        g_object_set(G_OBJECT(fieldGroup), 
            "child", weightLabel, "weight", 0,
            "child", weightBox, "vert-weight", 0,
            NULL);
    }
    if (show_fontname) {
        g_object_set(G_OBJECT(fieldGroup), 
            "child", fontnameLabel, "weight", 0,
            "child", d->d_FontnamePopup, "vert-weight", 0,
            NULL);
    }
    if (show_bg) {
        g_object_set(G_OBJECT(fieldGroup), 
            "child", bgLabel, "weight", 0,
            "child", bgGroup,
            NULL);
    }

    GuiObject go = GtkNewEditor(pr, true, false, false, &d->d_BubbleHelp);
    GuiSetEditorText(d->d_BubbleHelp, DisplayFieldBubbleHelp(dio));
    g_object_set(G_OBJECT(fieldGroup), 
        "child", GuiLabel(GCS(MSG_DisplayField_BubbleHelp), d->d_BubbleHelp),
        "weight", 0,
        "child", go, 
        NULL);

    fieldGroup = Frame(GCS(MSG_Field)),
        "child", GroupAlignment(true),
            "child", fieldGroup,
            End,
        End;

    gint weight = gtk_group_get_total_vert_weight(GTK_GROUP(topGroup));
    GuiObject attrsGroup = VGroup,
        "child", topGroup, "vert-weight", weight? 1 : 0,
        "child", fieldGroup, "vert-weight", 1,
        "child", enabledGroup, "vert-weight", 0,
        End;

    GuiObject group = attrsGroup;
    if (extras) {
        GuiObject notebook = NotebookObject,
            "enable-popup", true,
            End;
        gtk_notebook_append_page_menu(GTK_NOTEBOOK(notebook), 
                GroupAlignment(false), "child", attrsGroup, End,
                LabelObject, 
                    "label", GCS(MSG_General),
                    End,
                0);
        gtk_notebook_append_page_menu(GTK_NOTEBOOK(notebook), 
                GroupAlignment(false), "child", extrasGroup, End,
                LabelObject, 
                    "label", GCS(MSG_DisplayField_Extras),
                    End,
                0);
        group = VGroup, "child", notebook, End;
    }

    GuiObject root = GroupAlignment(false),
        "child", group,
        End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_DisplayField),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Field object editor");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    GuiAddEntryCallbacks(d->d_Title, d->d_Ok);
    GuiAddEntryCallbacks(d->d_TitleFontname, d->d_Ok);
    GuiAddEntryCallbacks(d->d_Shortcut, d->d_Ok);
    if (d->d_ReadOnly) {
	g_signal_connect_swapped(d->d_ReadOnly, "toggled",
		G_CALLBACK(onReadOnlyToggled), d);
	g_object_set(G_OBJECT(d->d_ReadOnly), "active", readOnly, NULL);
    }
    GuiAddEntryCallbacks(d->d_DoubleClickTriggerStr, d->d_Ok);
    GuiAddEntryCallbacks(d->d_URLDropTriggerStr, d->d_Ok);
    GuiAddEntryCallbacks(d->d_SortDropTriggerStr, d->d_Ok);
    GuiAddEntryCallbacks(d->d_EnabledComputeStr, d->d_Ok);
    GuiAddEntryCallbacks(d->d_Fontname, d->d_Ok);

    g_signal_connect(dialog, "response", 
            G_CALLBACK(closeEditDisplayField), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    if (d->d_SubType && d->d_SubTypeGroup) {
        g_signal_connect_swapped(d->d_SubType, "changed", 
                G_CALLBACK(onSubTypeChanged), d);
        gtk_notebook_set_current_page(GTK_NOTEBOOK(d->d_SubTypeGroup), subType);
    }
    if (d->d_ListviewKind && lvcompute && lvcompute[0]) {
	g_object_set(G_OBJECT(d->d_ListviewKind), "active", 1, NULL);
    }
    GuiAddEntryCallbacks(d->d_ListviewComputeStr, d->d_Ok);
    EditRecordDescriptionSpecAddCallbacks(d->d_RecordDescriptionSpec, d->d_Ok);

    updateFocusWidgets(d);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }
}

#endif

