//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkEditDisplayTable.cc
//
// Creation date: 10.08.2012
// Last modified: 26.04.2014
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "gtk/gtkEditDisplayTable.h"
#include "gtk/gtkEditRecordDescriptionSpec.h"
#include "gtk/gtkStructureEditor.h"
#include "gtk/classes/gtkgroup.h"
#include "gtk/classes/gtkpopimage.h"
#include "gtk/classes/gtkpopselection.h"
#include "gtk/gtkimpl.h"
#include "gtk/gtkUtil.h"
#include "display.h"
#include "requesters.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"

#define IMAGESPEC_START "0:128"

static void defaultBackgroundOnToggled(GtkWidget *def, GtkWidget *bg)
{
    gtk_widget_set_sensitive(bg, 
	    !gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(def)));
}

struct EditDisplayTableData
{
    Project *d_Project;
    DisplayObject *d_DisplayTable;
    StructureEditorData *d_StructureEditorData;
    GuiObject d_Weight, d_Background, d_DefaultBackground, d_HasPanel;
    GuiObject d_PanelTitle, d_PanelFontname;
    GuiObject d_PanelBackground, d_PanelDefaultBackground;
    GuiObject d_PanelNumAll, d_PanelFilter, d_PanelArrows;
    GuiObject d_RecordDescriptionSpec;
    GuiObject d_Ok, d_Cancel;
};

static void onHasPanelToggled(GuiObject hasPanel, GuiObject panelGroup)
{
    gtk_widget_set_sensitive(panelGroup, gtkToggleButtonActive(hasPanel));
}

static void closeEditDisplayTable(GuiObject go, gint response, 
        EditDisplayTableData *d)
{
    Project *pr = d->d_Project;

    if (response == GTK_RESPONSE_ACCEPT) {
	GuiBusyCursor busy;
        long error = OK;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        Table *tab = DisplayObjectTable(d->d_DisplayTable);
        MemPool *memPool = ProjectMemPool(pr);
        struct RecordDescriptionSpec *rds = 0;

	DisplayTableData dtd;
        dtd.dtd_Title = gtkEntryText(d->d_PanelTitle);
        dtd.dtd_Weight = (int)gtkRangeValue(d->d_Weight);
        dtd.dtd_Fontname = gtkEntryText(d->d_PanelFontname);
        dtd.dtd_Background = 
            gtkToggleButtonActive(d->d_DefaultBackground)? 
            NULL : gtk_pop_image_get_spec(GTK_POP_IMAGE(d->d_Background));
        dtd.dtd_HasPanel = (bool)gtkToggleButtonActive(d->d_HasPanel);
        dtd.dtd_PanelBackground = 
            gtkToggleButtonActive(d->d_PanelDefaultBackground)? 
            NULL : gtk_pop_image_get_spec(GTK_POP_IMAGE(d->d_PanelBackground));
	dtd.dtd_PanelNumAll = (bool)gtkToggleButtonActive(d->d_PanelNumAll);
	dtd.dtd_PanelFilter = (bool)gtkToggleButtonActive(d->d_PanelFilter);
	dtd.dtd_PanelArrows = (bool)gtkToggleButtonActive(d->d_PanelArrows);
        rds = EditRecordDescriptionSpecGetSpec(
                d->d_RecordDescriptionSpec, memPool);
        dtd.dtd_RecordDescriptionSpec = rds;
        if (rds == NULL) {
            error = ERR_MEM;
        }

        if (error == OK) {
            error = DisplayChangeTable(tab, &dtd);
            if (error == OK) {
                ProjectChange(pr, TRUE);
            }
        }

        FreeRecordDescriptionSpec(memPool, rds);

        if (error) {
            Ask2(go, GCS(MSG_ErrorMessage), AT_ERROR, 
                    GCS(MSG_StarOk), GES(error));
            return;
        }

        DisplayTableFinish(d->d_StructureEditorData, d->d_DisplayTable);
    }

    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

void GuiDisplayTable(StructureEditorData *sed, DisplayObject *dio)
{
    EditDisplayTableData *d = g_new0(EditDisplayTableData, 1);
    Project *pr = DisplayObjectProject(dio);
    Table *tab = DisplayObjectTable(dio);
    bool hasPanel = DisplayTableHasPanel(dio);
    const char *bg;

    d->d_Project = pr;
    d->d_DisplayTable = dio;
    d->d_StructureEditorData = sed;

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */

    // general
    int weight = DisplayObjectWeight(dio);
    d->d_Weight = HScaleObject,
        "adjustment", gtk_adjustment_new(weight, 0, 1000, 1, 10, 0),
        "value-pos", GTK_POS_RIGHT,
        "digits", 0,
        End;
    GuiObject weightBox = GuiMakeTooltip(d->d_Weight, GCS(MSG_WeightBH));

    bg = DisplayObjectBackground(dio);
    d->d_Background = PopImageObject,
        "spec", (bg && bg[0])? bg : IMAGESPEC_START,
        "types", GTK_POP_IMAGE_COLOR,
	End;
    g_signal_connect_swapped(d->d_Background, "clicked",
            G_CALLBACK(ProjectSetCurrentDir), pr);
    GuiObject backgroundBox = GuiMakeTooltip(d->d_Background,
            GCS(MSG_BackgroundBH));

    d->d_DefaultBackground = CheckButtonObject, 
        "label", GCS(MSG_Background_Default),
        "use-underline", true,
        End;
    GuiSetTooltip(d->d_DefaultBackground, GCS(MSG_Background_DefaultBH));
    g_signal_connect(d->d_DefaultBackground, "toggled", 
	    G_CALLBACK(defaultBackgroundOnToggled), d->d_Background);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_DefaultBackground),
	    !bg || !bg[0]);

    GuiObject back = HGroup,
        "child", backgroundBox, "weight", 200,
        "child", SpaceObject, End,
        "child", VCenter(d->d_DefaultBackground),
        End;

    GuiObject basic = VGroup,
        "columns", 2,
        "child", GuiLabel(GCS(MSG_Weight), d->d_Weight), "weight", 0,
        "child", weightBox, "vert-weight", 0,
	"child", GuiLabel(GCS(MSG_Background), d->d_Background), "weight", 0,
        "child", back,
        End;

    d->d_HasPanel = CheckButtonObject, 
        "label", GCS(MSG_DisplayTable_HasPanel),
        "use-underline", true,
        "active", hasPanel,
	End;
    GuiSetTooltip(d->d_HasPanel, GCS(MSG_DisplayTable_HasPanelBH));

    GuiObject generalGroup = VGroup,
        "child", basic,
        "child", d->d_HasPanel, "vert-weight", 0,
        End;

    // panel
    d->d_PanelTitle = EntryObject,
	"max-length", TITLE_LEN,
	"xalign", 0.5,
        "text", EmptyNullStr(DisplayObjectTitle(dio)),
	End;
    GuiSetTooltip(d->d_PanelTitle, GCS(MSG_DisplayTable_Panel_TitleBH));

    d->d_PanelFontname = EntryObject,
        "max-length", FONTNAME_LEN,
        "xalign", 0.5,
        "text", EmptyNullStr(DisplayObjectGTKFontname(dio)),
        End;
    GuiSetTooltip(d->d_PanelFontname, GCS(MSG_FontBH));

    bg = DisplayTablePanelBackground(dio);
    d->d_PanelBackground = PopImageObject,
        "spec", (bg && bg[0])? bg : IMAGESPEC_START,
        "types", GTK_POP_IMAGE_COLOR,
	End;
    g_signal_connect_swapped(d->d_PanelBackground, "clicked",
            G_CALLBACK(ProjectSetCurrentDir), pr);
    GuiObject panelBackgroundBox = GuiMakeTooltip(d->d_PanelBackground,
            GCS(MSG_BackgroundBH));

    d->d_PanelDefaultBackground = CheckButtonObject, 
        "label", GCS(MSG_Background_Default),
        "use-underline", true,
        End;
    GuiSetTooltip(d->d_PanelDefaultBackground, GCS(MSG_Background_DefaultBH));
    g_signal_connect(d->d_PanelDefaultBackground, "toggled", 
	    G_CALLBACK(defaultBackgroundOnToggled), d->d_PanelBackground);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_PanelDefaultBackground),
	    !bg || !bg[0]);

    GtkWidget *panelBack = HGroup,
        "child", panelBackgroundBox, "weight", 200,
        "child", SpaceObject, End,
        "child", VCenter(d->d_PanelDefaultBackground),
        End;

    GuiObject panelBasic = VGroup,
        "columns", 2,
        ChildLabelWidget(GCS(MSG_Title), d->d_PanelTitle), "vert-weight", 0,
        "child", GuiLabel(GCS(MSG_Font), d->d_PanelFontname), "weight", 0,
	"child", PopSelectionObject,
	    "selection-type", GUI_TYPE_FONT_CHOOSER_DIALOG,
	    "string", d->d_PanelFontname,
	    "button", GuiPopButton(SELECT_FONT_ICON_NAME), 
	    End, "vert-weight", 0,
	"child", GuiLabel(GCS(MSG_Background), d->d_PanelBackground),"weight",0,
        "child", panelBack,
	End;

    bool gadgets = !TableExactlyOneRecord(tab);

    d->d_PanelNumAll = CheckButtonObject, 
        "label", GCS(MSG_DisplayTable_Panel_NumAll),
        "use-underline", true,
        "active", DisplayTablePanelNumAll(dio) && gadgets,
	"sensitive", gadgets,
	End;
    GuiSetTooltip(d->d_PanelNumAll, GCS(MSG_DisplayTable_Panel_NumAllBH));

    d->d_PanelFilter = CheckButtonObject,
        "label", GCS(MSG_Filter),
        "use-underline", true,
        "active", DisplayTablePanelFilter(dio) && gadgets,
	"sensitive", gadgets,
	End;
    GuiSetTooltip(d->d_PanelFilter, GCS(MSG_DisplayTable_Panel_FilterBH));

    d->d_PanelArrows = CheckButtonObject,
        "label", GCS(MSG_DisplayTable_Panel_Arrows),
        "use-underline", true,
        "active", DisplayTablePanelArrows(dio) && gadgets,
	"sensitive", gadgets,
	End;
    GuiSetTooltip(d->d_PanelArrows, GCS(MSG_DisplayTable_Panel_ArrowsBH));

    GuiObject panelFields = VGroup,
        "child", d->d_PanelNumAll,
        "child", d->d_PanelFilter,
        "child", d->d_PanelArrows,
        End;
    panelFields = Frame(GCS(MSG_DisplayTable_Panel_Gadgets)),
        "child", GroupAlignment(true),
	    "child", HCenter(panelFields),
	    End,
	End;

    GuiObject panelGroup = VGroup,
        "child", panelBasic,
	"child", panelFields, "vert-weight", 0,
        "sensitive", hasPanel,
	End;

    g_signal_connect(d->d_HasPanel, "toggled", 
            G_CALLBACK(onHasPanelToggled), panelGroup);

    d->d_RecordDescriptionSpec = CreateEditRecordDescriptionSpec(tab,
            DisplayTableRecordDescriptionSpec(dio), FALSE);

    GuiObject notebook = NotebookObject,
        "enable-popup", true,
        End;
    gtk_notebook_append_page_menu(GTK_NOTEBOOK(notebook), 
            GroupAlignment(false), "child", generalGroup, End,
            LabelObject, "label", GCS(MSG_General), End,
            0);
    gtk_notebook_append_page_menu(GTK_NOTEBOOK(notebook), 
            GroupAlignment(false), "child", panelGroup, End,
            LabelObject, "label", GCS(MSG_Panel), End,
            0);
    gtk_notebook_append_page_menu(GTK_NOTEBOOK(notebook), 
            GroupAlignment(false), "child", d->d_RecordDescriptionSpec, End,
            LabelObject, "label", GCS(MSG_Record), End,
            0);
    GuiObject group = VGroup, "child", notebook, End;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_DisplayTable),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Table object editor");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    GuiAddEntryCallbacks(d->d_PanelTitle, d->d_Ok);
    GuiAddEntryCallbacks(d->d_PanelFontname, d->d_Ok);
    EditRecordDescriptionSpecAddCallbacks(d->d_RecordDescriptionSpec, d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(closeEditDisplayTable), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }
}

#endif

