//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkEditField.cc
//
// Creation date: 31.08.2005
// Last modified: 10.03.2021
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "gtk/gtkEditField.h"
#include "gtk/gtkStructureEditor.h"
#include "gtk/classes/gtkradiobox.h"
#include "gtk/classes/gtkeditlabels.h"
#include "gtk/classes/gtkpopeditlabels.h"
#include "gtk/gtkimpl.h"
#include "gtk/gtkUtil.h"
#include "menuproject.h"
#include "structureeditor.h"
#include "requesters.h"
#include "preferences.h"
#include "labels.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"

static short types[] = 					// remapping of types
{
    TYPE_STRING, TYPE_INTEGER, TYPE_REAL, TYPE_BOOL, TYPE_CHOICE, TYPE_DATE, 
    TYPE_TIME, TYPE_MEMO, TYPE_REFERENCE, TYPE_VIRTUAL, TYPE_BUTTON,
};

enum
{
    GRP_STRING, GRP_BOOL, GRP_CHOICE, GRP_MEMO, GRP_REFERENCE, GRP_VIRTUAL, 
    GRP_DEFAULT, GRP_BUTTON,
};

static short pages[] =
{
    GRP_STRING, GRP_DEFAULT, GRP_DEFAULT, GRP_BOOL, GRP_CHOICE, GRP_DEFAULT, 
    GRP_DEFAULT, GRP_MEMO, GRP_REFERENCE, GRP_VIRTUAL, GRP_BUTTON,
};

struct EditFieldData
{
    Project *d_Project;
    Table *d_Table;
    Field *d_Field;
    DisplayObject *d_Parent;
    StructureEditorData *d_StructureEditorData;
    GuiObject d_Name;
    GuiObject d_Type;
    GuiObject d_TypeGroup;
    GuiObject d_StringMaxlen;
    GuiObject d_StringInitialValue;
    GuiObject d_BoolInitialValue;
    GuiObject d_ChoiceLabels;
    GuiObject d_ChoiceInitialValue;
    GuiObject d_Reference;
    GuiObject d_ReferenceFilter;
    GuiObject d_ReferenceAutoShow;
    GuiObject d_InitialValue;
    GuiObject d_InitialValueOther;
    GuiObject d_NilString;
    GuiObject d_ComputeStr;
    GuiObject d_NilStringVirtual;
    GuiObject d_Compute;
    GuiObject d_TriggerStr;
    GuiObject d_Trigger;
    GuiObject d_CountChanges;
    GuiObject d_LogChanges;
    GuiObject d_Ok, d_Cancel;
};

static void displayerrormessage(GuiObject obj, EditFieldData *d, long error)
{
    Ask2(obj, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));

    GuiObject activate = 0;
    switch (error) {
	case ERR_INVALIDNAME:
	case ERR_NAMEALREADYUSED:
        default:    
            activate = d->d_Name;
            break;

	case FIELDERR_INVALIDMAXLEN:
	    activate = d->d_StringMaxlen;
	    break;

	case FIELDERR_INITIALVALUELEN:
	case FIELDERR_INVALIDINITIALVALUE:
	    activate = d->d_InitialValueOther;
            break;
    }
    if (activate) {
        gtk_widget_grab_focus(activate);
    }
}

static long getfieldtypedata(EditFieldData *d, short type,
        const char **init, const char **nil, FieldTypeData *ftd)
{
    long val;

    memset(ftd, 0, sizeof(*ftd));

    switch (type) {
	case TYPE_STRING:
	    if (!StrToInt(gtkEntryText(d->d_StringMaxlen), &val)) {
                return FIELDERR_INVALIDMAXLEN;
            }
	    ftd->string.maxlen = (short)val;
	    *init = gtkEntryText(d->d_StringInitialValue);
	    break;

 	case TYPE_INTEGER:
	case TYPE_REAL:
	case TYPE_DATE:
	case TYPE_TIME:
	    if (gtkRadioBoxActive(d->d_InitialValue)) {
		*init = gtkEntryText(d->d_InitialValueOther);
	    }
            *nil = gtkEntryText(d->d_NilString);
            break;

	case TYPE_BOOL:
	    ftd->boolean.nil = !gtkRadioBoxActive(d->d_BoolInitialValue);
            break;

	case TYPE_CHOICE:
	    ftd->choice.labels = gtk_pop_edit_labels_get_labels(
                    GTK_POP_EDIT_LABELS(d->d_ChoiceLabels));
	    ftd->choice.initialvalue=gtkComboBoxActive(d->d_ChoiceInitialValue);
            break;

        case TYPE_REFERENCE:
            ftd->reference.table = (Table *)GuiListActiveEntry(d->d_Reference);
            ftd->reference.filter = (bool)gtkToggleButtonActive(
		    d->d_ReferenceFilter);
            ftd->reference.autoshow = (bool)gtkToggleButtonActive(
		    d->d_ReferenceAutoShow);
            break;

        case TYPE_VIRTUAL:
            ftd->virt.compute = gtkEntryText(d->d_ComputeStr);
            *nil = gtkEntryText(d->d_NilStringVirtual);
            break;
    }

    return OK;
}

static void freefieldtypedata(short type, FieldTypeData *ftd)
{
}

static void closeEditField(GuiObject go, gint response,EditFieldData *d)
{
    Project *pr = d->d_Project;

    long val;
    if (StrToInt(gtkEntryText(d->d_StringMaxlen), &val)) {
        SetPrefsValue(pr, PREFS_DEFAULTMAXLEN, val);
    }

    if (response == GTK_RESPONSE_ACCEPT) {
        GuiBusyCursor busy;
        long error = OK;
        const gchar *name = gtkEntryText(d->d_Name);
        const gchar *trigger = gtkEntryText(d->d_TriggerStr);
        gboolean countChanges = gtkToggleButtonActive(d->d_CountChanges);
        gboolean logChanges = gtkToggleButtonActive(d->d_LogChanges);

        if (!d->d_Field) {              // new field?
            int active = gtk_combo_box_get_active(GTK_COMBO_BOX(d->d_Type));
	    short type = types[active];
            Table *tab = d->d_Table;

            NewFieldData nfd;
            nfd.nfd_Table = tab;
            nfd.nfd_Pred = DisplayTableActiveField(tab);
            nfd.nfd_Name = name;
            nfd.nfd_Type = type;
	    nfd.nfd_InitialValue = NULL;
	    nfd.nfd_NilString = NULL;
            nfd.nfd_Trigger = trigger;
            nfd.nfd_CountChanges = countChanges;
            nfd.nfd_LogChanges = logChanges;
            error = getfieldtypedata(d, type, 
		    &nfd.nfd_InitialValue, &nfd.nfd_NilString,
                    &nfd.nfd_Contents);
            if (!error) {
                Field *field = 0;
                error = SE_NewField(&nfd, d->d_Parent, &field);
                if (!error) {
                    NewFieldFinish(d->d_StructureEditorData, field);
                }
            }
            freefieldtypedata(type, &nfd.nfd_Contents);
        } else {
            Field *field = d->d_Field;
            short type = FieldType(field);

            ChangeFieldData cfd;
            cfd.cfd_Field = field;
            cfd.cfd_Name = name;
	    cfd.cfd_InitialValue = NULL;
	    cfd.cfd_NilString = NULL;
            cfd.cfd_Trigger = trigger;
            cfd.cfd_CountChanges = countChanges;
            cfd.cfd_LogChanges = logChanges;
            error = getfieldtypedata(d, type, 
		    &cfd.cfd_InitialValue, &cfd.cfd_NilString,
                    &cfd.cfd_Contents);
            if (!error) {
                error = SE_ChangeField(&cfd);
                if (!error) {
                    ChangeFieldFinish(d->d_StructureEditorData, field);
                }
            }
            freefieldtypedata(type, &cfd.cfd_Contents);
        }
        
        if (error) {
            displayerrormessage(go, d, error);
            return;
        }
    }

    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

static void referenceCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    Table *tab = 0;
    gtk_tree_model_get(model, iter, 0, &tab, -1);
    g_object_set(G_OBJECT(cell), "text", EmptyNullStr(TableName(tab)), NULL);
}

static void onFieldTypeChanged(EditFieldData *d)
{
    gint active = gtk_combo_box_get_active(GTK_COMBO_BOX(d->d_Type));
    if (active >= 0) {
        gtk_notebook_set_current_page(GTK_NOTEBOOK(d->d_TypeGroup), 
                pages[active]);
    }
}

static void updatePage(EditFieldData *d, int page)
{
    gboolean on = (page != GRP_VIRTUAL && page != GRP_BUTTON);
    gboolean countOn, logOn;

    gtk_widget_set_sensitive(d->d_Trigger, page != GRP_VIRTUAL);
    countOn = on && TableCountChanges(d->d_Table);
    gtk_widget_set_sensitive(d->d_CountChanges, countOn);
    if (!countOn) {
        g_object_set(G_OBJECT(d->d_CountChanges), "active", false, NULL);
    }
    logOn = on && TableLogChanges(d->d_Table);
    gtk_widget_set_sensitive(d->d_LogChanges, logOn);
    if (!logOn) {
        g_object_set(G_OBJECT(d->d_LogChanges), "active", false, NULL);
    }
}

static void onSwitchPage(EditFieldData *d, gpointer np, gint page)
{
    updatePage(d, page);
}

static void onInitialValueChanged(EditFieldData *d)
{
    gtk_widget_set_sensitive(d->d_InitialValueOther,
            gtk_radio_box_get_active(GTK_RADIO_BOX(d->d_InitialValue)) == 1);
}

static void onNotifyLabels(EditFieldData *d)
{
    char **labels = gtk_pop_edit_labels_get_labels(
            GTK_POP_EDIT_LABELS(d->d_ChoiceLabels));
    GtkComboBox *combo = GTK_COMBO_BOX(d->d_ChoiceInitialValue);
    gint active = gtk_combo_box_get_active(combo);
    GtkTreeModel *model = gtk_combo_box_get_model(combo);
    GtkListStore *store = GTK_LIST_STORE(model);
    GuiSetListStoreLabels(store, labels);
    gtk_combo_box_set_active(combo, active);
}

static gboolean onAcceptChoiceLabels(EditFieldData *d, char **labels)
{
    bool warnChanged = false;
    bool warnEmpty = false;
    bool warnDuplicate = false;

    if (d->d_Field && SUCC(HEAD(TableRecords(d->d_Table)))) {
        // Check if the user changed any existing labels.
        char **old = FieldChoiceLabels(d->d_Field);
        warnChanged = NumLabels(old) > NumLabels(labels);
        if (!warnChanged) {
            char **s = old, **t = labels;
            while (*s && *t && !warnChanged) {
                warnChanged = (CodesetStrCollate(*s, *t, CS_COLLATE_ORDER)!=0);
                s++;
                t++;
            }
        }
    }

    // Check for empty and duplicate labels.
    for (char **s = labels; *s; s++) {
        warnEmpty |= (NumWords(*s) == 0);
        for (char **t = s + 1; *t; t++) {
            warnDuplicate |= (strcmp(*s, *t) == 0);
        }
    }

    char *msg[3];
    int num = 0;
    if (warnChanged) {
        msg[num++] = GCS(MSG_Field_Choice_WarnChangeLabels);
    }
    if (warnEmpty) {
        msg[num++] = GCS(MSG_Field_Choice_WarnEmptyLabels);
    }
    if (warnDuplicate) {
        msg[num++] = GCS(MSG_Field_Choice_WarnDuplicateLabels);
    }
    if (num > 0) {
        char format[32];
        strcpy(format, "%s");
        for (int i = 0; i < num; i++) {
            strcat(format, "\n\n%s");
        }
        if (!Ask(AT_WARN, GCS(MSG_CloseAnswers),
                    format, GCS(MSG_Field_Choice_Confirm),
                    msg[0], msg[1], msg[2])) {
            g_signal_stop_emission_by_name(d->d_ChoiceLabels, "accept");
            return false;
        }
    }
    return true;
}

static void createField(Project *pr, Field *field, Table *tab,
        bool newField, DisplayObject *parent, StructureEditorData *sed)
{
    EditFieldData *d = g_new0(EditFieldData, 1);

    d->d_Project = pr;
    d->d_Table = tab;
    d->d_Field = newField? 0 : field;
    d->d_Parent = parent;
    d->d_StructureEditorData = sed;

    d->d_Name = EntryObject,
        "max-length", ID_NAME_LEN,
        "xalign", 0.5,
        End;
    GuiSetTooltip(d->d_Name, GCS(MSG_Field_NameBH));

    static char *typeContent[12];
    typeContent[0] = GCS(MSG_String);
    typeContent[1] = GCS(MSG_Integer);
    typeContent[2] = GCS(MSG_Real);
    typeContent[3] = GCS(MSG_Boolean);
    typeContent[4] = GCS(MSG_Choice);
    typeContent[5] = GCS(MSG_Date);
    typeContent[6] = GCS(MSG_Time);
    typeContent[7] = GCS(MSG_Memo);
    typeContent[8] = GCS(MSG_Reference);
    typeContent[9] = GCS(MSG_Virtual);
    typeContent[10] = GCS(MSG_Button);
    typeContent[11] = 0;
    d->d_Type = GuiComboBox(typeContent);
    GuiObject typeBox = GuiMakeTooltip(d->d_Type, GCS(MSG_Field_TypeBH));

    GuiObject grpTop = VGroup,
        "columns", 2,
        ChildLabelWidget(GCS(MSG_Name), d->d_Name),
        "child", GuiLabel(GCS(MSG_Type), d->d_Type), "weight", 0,
        "child", typeBox,
        End;

    long val = GetPrefsValue(pr, PREFS_DEFAULTMAXLEN);
    d->d_StringMaxlen = SpinButtonObject,
        "adjustment", gtk_adjustment_new(val, 1, MAX_STRING_LEN, 1, 10, 0),
	"digits", 0,
	"numeric", true,
	"snap-to-ticks", false,
	"update-policy", GTK_UPDATE_IF_VALID,
	"wrap", false,
        "max-length", 3,
        "width-chars", 5,
        "xalign", 0.5,
        End;
    GuiSetTooltip(d->d_StringMaxlen, GCS(MSG_Field_String_MaxLengthBH));

    d->d_StringInitialValue = EntryObject,
	"max-length", 255,
	"xalign", 0.5,
	End;
    GuiSetTooltip(d->d_StringInitialValue, GCS(MSG_Field_InitialValueBH));

    GuiObject grpString = VGroup,
        "columns", 2,
        ChildLabelWidget(GCS(MSG_Field_String_MaxLength), d->d_StringMaxlen),
        ChildLabelWidget(GCS(MSG_Field__InitialValue), d->d_StringInitialValue),
	End;
    grpString = GroupAlignment(false),
        "child", VCenter(grpString),
        End;

    static char *boolInitialValueContent[3];
    boolInitialValueContent[0] = GCS(MSG_Field_Boolean_InitialValue_Nil);
    boolInitialValueContent[1] = GCS(MSG_Field_Boolean_InitialValue_True);
    boolInitialValueContent[2] = 0;
    d->d_BoolInitialValue = RadioBoxObject,
	"entries", boolInitialValueContent,
	End;
    GuiObject boolInitialValueBox = GuiMakeTooltip(d->d_BoolInitialValue,
            GCS(MSG_Field_InitialValueBH));

    GuiObject go = Frame(GCS(MSG_Field_InitialValue)),
        "child", GroupAlignment(true),
            "child", HCenter(boolInitialValueBox),
            End,
        End;
    GuiObject grpBoolean = GroupAlignment(false),
        "child", VCenter(HCenter(go)),
        End;

    d->d_ChoiceLabels = PopEditLabelsObject,
        "label", GCS(MSG_EditLabels),
        "use-underline", true,
        End;
    GuiSetTooltip(d->d_ChoiceLabels, GCS(MSG_Field_Choice_LabelsBH));
    g_signal_connect_swapped(d->d_ChoiceLabels, "accept",
            G_CALLBACK(onAcceptChoiceLabels), d);

    d->d_ChoiceInitialValue = GuiComboBox(0);
    GuiObject choiceBox = GuiMakeTooltip(d->d_ChoiceInitialValue,
            GCS(MSG_Field_InitialValueBH));

    GuiObject grpChoice = VGroup,
	"columns", 2,
        "child", SpaceObject, End,
	"child", d->d_ChoiceLabels,
        "child", GuiLabel(GCS(MSG_Field__InitialValue), d->d_ChoiceInitialValue),
        "weight", 0,
        "child", choiceBox,
	End;
    grpChoice = GroupAlignment(false),
        "child", VCenter(grpChoice),
        End;

    d->d_Reference = GuiPointerTreeView(referenceCellDataFunc);
    GuiSetTooltip(d->d_Reference, GCS(MSG_Field_ReferenceBH));

    d->d_ReferenceAutoShow = CheckButtonObject,
        "label", GCS(MSG_AutoShow),
        "use-underline", true,
        End;
    GuiSetTooltip(d->d_ReferenceAutoShow, GCS(MSG_Field_Reference_AutoShowBH));

    d->d_ReferenceFilter = CheckButtonObject, 
        "label", GCS(MSG_Filter),
        "use-underline", true,
        End;
    GuiSetTooltip(d->d_ReferenceFilter, GCS(MSG_Field_Reference_FilterBH));

    GuiObject grpReference = VGroup,
        "child", GuiLabel(GCS(MSG_Field_Reference), d->d_Reference, 0.5), 
        "weight", 0,
	"child", ScrolledWindowObject,
	    "hscrollbar-policy", GTK_POLICY_AUTOMATIC,
	    "shadow-type", GTK_SHADOW_IN,
	    "child", d->d_Reference,
	    End,
	"child", HGroup,
            "child", d->d_ReferenceAutoShow,
            "child", d->d_ReferenceFilter,
	    End, "weight", 0,
	End;
    grpReference = GroupAlignment(false),
        "child", grpReference,
        End;

    static char *initialValueContent[3];
    initialValueContent[0] = GCS(MSG_Field_InitialValue_Nil);
    initialValueContent[1] = GCS(MSG_Field_InitialValue_Other);
    initialValueContent[2] = 0;
    d->d_InitialValue = RadioBoxObject,
	"entries", initialValueContent,
	End;
    GuiObject initialValueBox = GuiMakeTooltip(d->d_InitialValue,
            GCS(MSG_Field_InitialValueBH));

    d->d_InitialValueOther = EntryObject,
        "sensitive", false,
        "max-length", 20,
        "xalign", 0.5,
	End;

    GuiObject grpDefaultSub = Frame(GCS(MSG_Field_InitialValue)),
        "child", GroupAlignment(true),
            "child", HGroup,
                "child", initialValueBox, "weight", 0,
                "child", VSeparatorObject, End, "weight", 0,
                "child", VGroup,
                    "child", GuiLabel(GCS(MSG_Field_OtherValue), 
                            d->d_InitialValueOther, 0.5),
                    "child", d->d_InitialValueOther,
                    End,
                End,
            End,
	End;

    d->d_NilString = EntryObject,
	"max-length", FIELD_NILVALUE_LEN,
	"xalign", 0.5,
	End;
    GuiSetTooltip(d->d_NilString, GCS(MSG_Field_NilStringBH));

    GuiObject grpDefault = VGroup,
	"child", grpDefaultSub,
	"child", VGroup,
	    "columns", 2,
            ChildLabelWidget(GCS(MSG_Field_NilString), d->d_NilString),
	    End,
	End;
    grpDefault = GroupAlignment(false),
        "child", VCenter(grpDefault),
        End;

    GuiObject grpMemo = SpaceObject, End;

    d->d_ComputeStr = EntryObject,
	"max-length", FIELD_TRIGGER_LEN,
	"xalign", 0.5,
	End;
    GuiSetTooltip(d->d_ComputeStr, GCS(MSG_Field_ComputeBH));

    d->d_Compute = UserFunctionsPopup(d->d_ComputeStr, pr);

    d->d_NilStringVirtual = EntryObject,
	"max-length", FIELD_NILVALUE_LEN,
	"xalign", 0.5,
	End;
    GuiSetTooltip(d->d_NilStringVirtual, GCS(MSG_Field_NilStringBH));

    GuiObject grpVirtual = VGroup,
        "columns", 2,
        "child", GuiLabel(GCS(MSG_Field_Compute), d->d_ComputeStr), "weight", 0,
        "child", d->d_Compute,
        ChildLabelWidget(GCS(MSG_Field_NilString), d->d_NilStringVirtual),
	End;
    grpVirtual = GroupAlignment(false),
        "child", VCenter(grpVirtual),
        End;

    GuiObject grpButton = SpaceObject, End;

    d->d_TypeGroup = NotebookObject,
        "show-tabs", false,
        "show-border", false,
	"child", grpString,
	"child", grpBoolean,
	"child", grpChoice,
	"child", grpMemo,
	"child", grpReference,
        "child", grpVirtual,
	"child", grpDefault,
	"child", grpButton,
	End;
    GuiSetHelpNode(d->d_TypeGroup, "Type specific settings");

    d->d_TriggerStr = EntryObject,
	"max-length", FIELD_TRIGGER_LEN,
	"xalign", 0.5,
	End;
    GuiSetTooltip(d->d_TriggerStr, GCS(MSG_Field_TriggerBH));

    d->d_Trigger = UserFunctionsPopup(d->d_TriggerStr, pr);

    d->d_CountChanges = CheckButtonObject, 
        "label", GCS(MSG_CountChanges),
	"use-underline", true,
        "active", true,
        End;
    GuiSetTooltip(d->d_CountChanges, GCS(MSG_Field_CountChangesBH));
    d->d_LogChanges = CheckButtonObject, 
        "label", GCS(MSG_LogChanges),
	"use-underline", true,
        "active", true,
        End;
    GuiSetTooltip(d->d_LogChanges, GCS(MSG_Field_LogChangesBH));
    
    GuiObject group = VGroup,
	"child", grpTop, "vert-weight", 0,
	"child", Frame(0), 
	    "child", d->d_TypeGroup,
	    End,
	"child", VGroup,
	    "columns", 2,
            "child", GuiLabel(GCS(MSG_Field_Trigger), d->d_TriggerStr),
            "weight", 0,
            "child", d->d_Trigger,
	    End, "vert-weight", 0,
        "child", HGroup,
            "child", d->d_CountChanges,
            "child", d->d_LogChanges,
	    End, "vert-weight", 0,
	End;

    GuiObject root = GroupAlignment(false),
        "child", group,
        End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
	"title", !newField?  GCS(MSG_ChangeField) :
	    (field?  GCS(MSG_CopyField) : GCS(MSG_NewField)),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, 
            newField? "Creating fields" : "Changing fields");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    GuiAddEntryCallbacks(d->d_Name, d->d_Ok);
    GuiAddEntryCallbacks(d->d_StringMaxlen, d->d_Ok);
    GuiAddEntryCallbacks(d->d_StringInitialValue, d->d_Ok);
    GuiAddEntryCallbacks(d->d_InitialValueOther, d->d_Ok);
    GuiAddEntryCallbacks(d->d_NilString, d->d_Ok);
    GuiAddEntryCallbacks(d->d_TriggerStr, d->d_Ok);
    GuiAddEntryCallbacks(d->d_ComputeStr, d->d_Ok);
    GuiAddEntryCallbacks(d->d_NilStringVirtual, d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(closeEditField), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    g_signal_connect_swapped(d->d_Type, "changed", 
            G_CALLBACK(onFieldTypeChanged), d);
    g_signal_connect_swapped(d->d_TypeGroup, "switch-page",
            G_CALLBACK(onSwitchPage), d);
    g_signal_connect_swapped(d->d_InitialValue, "active-changed",
            G_CALLBACK(onInitialValueChanged), d);
    g_signal_connect_swapped(d->d_ChoiceLabels, "notify::labels",
            G_CALLBACK(onNotifyLabels), d);

    if (newField) {
        GuiListSetListWithFilter(d->d_Reference, ProjectTables(pr),
                TableSystemFilter);
    }
    if (field) {
        static const short maxPage = sizeof(types) / sizeof(types[0]) - 1;
        short type = FieldType(field), page;

        for (page = 0; page < maxPage; page++) {
            if (types[page] == type) {
                break;
            }
        }

        gtk_entry_set_text(GTK_ENTRY(d->d_Name),
		EmptyNullStr(FieldName(field)));
        gtk_combo_box_set_active(GTK_COMBO_BOX(d->d_Type), page);
        gtk_widget_set_sensitive(d->d_Type, newField);

        DataInfo *di = FieldDataInfo(field);
        bool nil = ((di->di_Flags & DIF_NIL) != 0);
        bool yes = !nil;
        switch (type)
        {
	    case TYPE_STRING:
                gtk_spin_button_set_value(GTK_SPIN_BUTTON(d->d_StringMaxlen),
                        FieldStringMaxlen(field));
		gtk_entry_set_text(GTK_ENTRY(d->d_StringInitialValue),
			EmptyNullStr(di->di_Contents.string));
                break;

	    case TYPE_INTEGER:
                gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_InitialValue), yes);
                if (yes) {
                    gtk_entry_set_text(GTK_ENTRY(d->d_InitialValueOther), 
			    IntToStr(di->di_Contents.integer));
                }
		gtk_entry_set_text(GTK_ENTRY(d->d_NilString), 
			EmptyNullStr(FieldNilString(field)));
	        break;

	    case TYPE_REAL:
                gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_InitialValue), yes);
                if (yes) {
                    gtk_entry_set_text(GTK_ENTRY(d->d_InitialValueOther), 
			    DoubleToStr(di->di_Contents.real, -1, 
				(short)GetPrefsValue(pr, PREFS_REALFORMAT)));
                }
                gtk_entry_set_text(GTK_ENTRY(d->d_NilString), 
			EmptyNullStr(FieldNilString(field)));
	        break;

	    case TYPE_BOOL:
                gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_BoolInitialValue), 
                        yes);
	        break;

	    case TYPE_CHOICE:
                gtk_pop_edit_labels_set_labels(
                        GTK_POP_EDIT_LABELS(d->d_ChoiceLabels), 
                        FieldChoiceLabels(field));
                gtk_combo_box_set_active(GTK_COMBO_BOX(d->d_ChoiceInitialValue),
                        di->di_Contents.choice);
	        break;

            case TYPE_DATE:
                gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_InitialValue), yes);
                if (yes) {
                    gtk_entry_set_text(GTK_ENTRY(d->d_InitialValueOther), 
                            DateToString(di->di_Contents.date,
                                (short)GetPrefsValue(pr, PREFS_DATEFORMAT)));
                }
                gtk_entry_set_text(GTK_ENTRY(d->d_NilString), 
			EmptyNullStr(FieldNilString(field)));
	        break;

	    case TYPE_TIME:
                gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_InitialValue), yes);
                if (yes) {
		    gtk_entry_set_text(GTK_ENTRY(d->d_InitialValueOther), 
			    TimeToString(di->di_Contents.time, 
				    FORMAT_HOURMINSEC));
                }
                gtk_entry_set_text(GTK_ENTRY(d->d_NilString),
			EmptyNullStr(FieldNilString(field)));
	        break;

            case TYPE_REFERENCE:
                GuiListSetListWithFilter(d->d_Reference, ProjectTables(pr),
                        TableSystemFilter);
                GuiListSetActiveNode(d->d_Reference, 
			FieldReferenceTable(field));
                gtk_widget_set_sensitive(d->d_Reference, false);
                gtk_toggle_button_set_active(
                        GTK_TOGGLE_BUTTON(d->d_ReferenceFilter),
                        FieldReferenceFilter(field));
                gtk_toggle_button_set_active(
                        GTK_TOGGLE_BUTTON(d->d_ReferenceAutoShow),
                        FieldReferenceAutoShow(field));
                break;

            case TYPE_VIRTUAL:
                gtk_entry_set_text(GTK_ENTRY(d->d_ComputeStr), 
			EmptyNullStr(FieldVirtualCompute(field)));
		gtk_entry_set_text(GTK_ENTRY(d->d_NilStringVirtual), 
			EmptyNullStr(FieldNilString(field)));
                break;
        }
        gtk_entry_set_text(GTK_ENTRY(d->d_TriggerStr), 
		EmptyNullStr(FieldTrigger(field)));
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_CountChanges),
                FieldCountChanges(field));
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_LogChanges),
                FieldLogChanges(field));
    }
    updatePage(d, gtk_notebook_get_current_page(GTK_NOTEBOOK(d->d_TypeGroup)));

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }
}

static bool checkNewChangeField(Field *field, Table *tab, bool newField)
{
    long error = newField? CanNewField(tab) : CanChangeField(field);
    bool rc = (error == OK);

    if (error == RECERR_NOPROJECTNAME) {
        if (Ask(AT_WARN, GCS(MSG_AskProjectSaveFirstAnswers),
                    GCS(MSG_AskProjectSaveFirst)) != 0) {
            MenuProjectSaveAs(TableProject(tab));
            error = newField? CanNewField(tab) : CanChangeField(field);
            rc = (error == OK);
        } // else user canceled
    } else if (error != OK) {
        Ask2(0, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));
    }

    return rc;
}

void GtkNewField(StructureEditorData *sed, Table *tab, 
        Field *field, DisplayObject *parent)
{
    if (checkNewChangeField(field, tab, true)) {
        createField(TableProject(tab), field, tab, true, parent, sed);
    }
}

void GtkChangeField(StructureEditorData *sed, Field *field) 
{
    Table *tab = FieldTable(field);
    if (checkNewChangeField(field, tab, false)) {
        createField(TableProject(tab), field, tab, false, 0, sed);
    }
}

#endif

