//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkEditRecordDescriptionSpec.cc
//
// Creation date: 10.07.2012
// Last modified: 21.08.2012
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "gtk/gtkEditRecordDescriptionSpec.h"
#include "gtk/gtkStructureEditor.h"
#include "gtk/classes/gtkgroup.h"
#include "gtk/gtkimpl.h"
#include "gtk/gtkUtil.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

struct EditRecordDescriptionSpecData
{
    GuiObject d_Type, d_TypeGroup;
    GuiObject d_Compute, d_ComputeStr;
    GuiObject d_FieldList;
    GuiObject d_MultiColumn;
};

static void onTypeChanged(EditRecordDescriptionSpecData *d)
{
    short active = gtkComboBoxActive(d->d_Type);

    gtk_notebook_set_current_page(GTK_NOTEBOOK(d->d_TypeGroup), active);
    gtk_widget_set_sensitive(d->d_MultiColumn, active != RDS_TYPE_INHERITED);
}

static void fieldListCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    Field *field = 0;
    gtk_tree_model_get(model, iter, 0, &field, -1);
    if (field == RDS_FIELD_RECNUM) {
	gchar *recnum= gtkEscapeStr(GCS(MSG_RecordDescription_RecordNumber));
        gchar *text = g_strdup_printf("<i>%s</i>", recnum);
        g_object_set(G_OBJECT(cell), "markup", text, NULL);
        g_free(text);
	g_free(recnum);
    } else {
        g_object_set(G_OBJECT(cell), "text", 
		EmptyNullStr(FieldName(field)), NULL);
    }
}

static bool recordDescriptionSpecFind(const RecordDescriptionSpec *rds, 
        Field *field)
{
    if (rds) {
        for (int i=0; i < rds->rds_NumFields; i++) {
            if (rds->rds_Fields[i] == field) {
                return true;
            }
        }
    }
    return false;
}

GuiObject CreateEditRecordDescriptionSpec(struct Table *tab,
        const RecordDescriptionSpec *rds, bool allowInherited)
{
    Project *pr = TableProject(tab);
    short active = rds? rds->rds_Type : 0;
    EditRecordDescriptionSpecData *d = g_new0(EditRecordDescriptionSpecData, 1);

    d->d_FieldList = GuiPointerTreeView(fieldListCellDataFunc, true);
    g_object_set(G_OBJECT(d->d_FieldList), "reorderable", true, NULL);
    GuiSetTooltip(d->d_FieldList, GCS(MSG_RecordDescription_FieldsBH));

    GtkTreeModel *model = gtk_tree_view_get_model(
            GTK_TREE_VIEW(d->d_FieldList));
    GtkListStore *store = GTK_LIST_STORE(model);
    GtkTreeIter iter;
    if (rds) {
        for (int i = 0; i < rds->rds_NumFields; i++) {
            if (rds->rds_Fields[i]) {
                gtk_list_store_append(store, &iter);
                gtk_list_store_set(store, &iter, 
                        0, rds->rds_Fields[i],
                        1, true,
                        -1);
            }
        }
    }
    if (!recordDescriptionSpecFind(rds, RDS_FIELD_RECNUM)) {
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter, 
                0, RDS_FIELD_RECNUM, 
                1, false,
                -1);
    }

    List *fields = TableFields(tab);
    if (fields) {
        Field *field;
        FORNODE (field, fields) {
            switch (FieldType(field))
            {
                case TYPE_STRING:
                case TYPE_DATE:
                case TYPE_TIME:
                case TYPE_INTEGER:
                case TYPE_REAL:
                case TYPE_BOOL:
                case TYPE_CHOICE:
                    if (!recordDescriptionSpecFind(rds, field)) {
                        gtk_list_store_append(store, &iter);
                        gtk_list_store_set(store, &iter, 
                                0, field, 1, false, -1);
                    }
                    break;
            }
        }
    }

    GuiObject page0 = VGroup,
        "child", ScrolledWindowObject,
            "hscrollbar-policy", GTK_POLICY_NEVER,
            "shadow-type", GTK_SHADOW_IN,
            "child", d->d_FieldList,
            End,
        End;

    d->d_ComputeStr = EntryObject,
        "max-length", FIELD_TRIGGER_LEN,
        "xalign", 0.5,
        "text", EmptyNullStr(rds? rds->rds_Compute : 0),
        End;
    GuiSetTooltip(d->d_ComputeStr, GCS(MSG_RecordDescription_ComputeBH));

    d->d_Compute = UserFunctionsPopup(d->d_ComputeStr, pr);
    GuiObject cdlabel = GuiLabel(GCS(MSG_Compute), d->d_Compute, 0.5);

    GuiObject page1 = VGroup,
        "child", cdlabel,
        "child", d->d_Compute,
        End;

    static const char *desctext[4];
    desctext[0] = GCS(MSG_Fields);
    desctext[1] = GCS(MSG_Computed);
    desctext[2] = allowInherited? GCS(MSG_RecordDescription_FromTable) : 0;
    desctext[3] = 0;
    d->d_Type = GuiComboBox(desctext, active);
    GuiObject typeBox = GuiMakeTooltip(d->d_Type, 
            GCS(MSG_RecordDescription_DescriptionBH));

    d->d_TypeGroup = NotebookObject,
        "show-tabs", false,
        "show-border", false,
        "child", GroupAlignment(false),
            "child", page0,
            End,
        "child", GroupAlignment(false),
            "child", VCenter(page1),
            End,
        "child", VGroup, End,
        End;

    d->d_MultiColumn = CheckButtonObject,
        "label", GCS(MSG_RecordDescription_UseMultiColumnList),
        "use-underline", true,
        "active", rds? rds->rds_MultiColumn : FALSE,
        "sensitive", rds? (rds->rds_Type != RDS_TYPE_INHERITED) : TRUE,
       End;
    GuiSetTooltip(d->d_MultiColumn, 
            GCS(MSG_RecordDescription_UseMultiColumnListBH));
    
    GuiObject go = VGroup,
        "child", HGroup,
            "child", GuiLabel(GCS(MSG_RecordDescription_Description),
                    d->d_Type), "weight", 0,
            "child", typeBox,
            End, "weight", 0,
        "child", d->d_TypeGroup,
        "child", d->d_MultiColumn, "vert-weight", 0,
        End;

    g_object_set_data_full(G_OBJECT(go), "edit-data", d, g_free);

    return go;
}

void EditRecordDescriptionSpecAddCallbacks(GuiObject go, GuiObject ok)
{
    if (go) {
        EditRecordDescriptionSpecData *d = (EditRecordDescriptionSpecData *)
            g_object_get_data(G_OBJECT(go), "edit-data");
        if (d) {
            GuiAddEntryCallbacks(d->d_ComputeStr, ok);
            g_signal_connect_swapped(d->d_Type, "changed", 
                    G_CALLBACK(onTypeChanged), d);
            gtk_notebook_set_current_page(GTK_NOTEBOOK(d->d_TypeGroup),
                    gtkComboBoxActive(d->d_Type));
        }
    }
}

static gint countSelected(GtkTreeModel *model)
{
    gint num = 0;
    GtkTreeIter iter;
    if (gtk_tree_model_get_iter_first(model, &iter)) {
	do {
	    gboolean sel;
	    gtk_tree_model_get(model, &iter, 1, &sel, -1);
	    num += sel;
	} while (gtk_tree_model_iter_next(model, &iter));
    }
    return num;
}

struct RecordDescriptionSpec *EditRecordDescriptionSpecGetSpec(
        GuiObject go, MemPool *memPool)
{
    if (!go) {
        return 0;
    }

    EditRecordDescriptionSpecData *d = (EditRecordDescriptionSpecData *)
        g_object_get_data(G_OBJECT(go), "edit-data");
    if (!d) {
        return 0;
    }

    GtkTreeModel *model = gtk_tree_view_get_model(
            GTK_TREE_VIEW(d->d_FieldList));
    gint num = countSelected(model);
    RecordDescriptionSpec *rds = AllocRecordDescriptionSpec(memPool, num);

    if (rds) {
        rds->rds_Type = gtkComboBoxActive(d->d_Type);
        rds->rds_Compute = AllocString(memPool, 
                EmptyNullStr(gtkEntryText(d->d_ComputeStr)));
        Field **fields = rds->rds_Fields;
        gint i = 0;
        GtkTreeIter iter;
        if (gtk_tree_model_get_iter_first(model, &iter)) {
            do {
                gboolean set;
                gtk_tree_model_get(model, &iter, 1, &set, -1);
                if (set) {
                    gtk_tree_model_get(model, &iter, 0, &fields[i], -1);
                    i++;
                }
            } while (i<num && gtk_tree_model_iter_next(model, &iter));
        }
        rds->rds_MultiColumn = gtkToggleButtonActive(d->d_MultiColumn);
    }

    return rds;
}

#endif

