//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkEditTable.cc
//
// Creation date: 30.08.2005
// Last modified: 06.08.2020
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "gtk/gtkEditTable.h"
#include "gtk/gtkStructureEditor.h"
#include "gtk/classes/gtkradiobox.h"
#include "gtk/gtkimpl.h"
#include "gtk/gtkUtil.h"
#include "structureeditor.h"
#include "requesters.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"

static char *numRecordsContent[3];

struct EditTableData
{
    Project *d_Project;
    Table *d_Table;
    DisplayObject *d_Parent;
    StructureEditorData *d_StructureEditorData;
    GuiObject d_Name, d_NumRecords;
    GuiObject d_NewTriggerStr, d_DelTriggerStr;
    GuiObject d_CountChanges, d_LogChanges;
    GuiObject d_Ok, d_Cancel;
};

static void closeEditTable(GuiObject go, gint response, EditTableData *d)
{
    Project *pr = d->d_Project;

    if (response == GTK_RESPONSE_ACCEPT) {
	GuiBusyCursor busy;

	const gchar *name = gtkEntryText(d->d_Name);
	const gchar *newTrigger = gtkEntryText(d->d_NewTriggerStr);
	const gchar *delTrigger = gtkEntryText(d->d_DelTriggerStr);

        struct TableData td;
     	td.td_Name = name;
	td.td_System = FALSE;
        td.td_ExactlyOneRecord = (gtkRadioBoxActive(d->d_NumRecords) == 1);
     	td.td_NewTrigger = newTrigger;
     	td.td_DelTrigger = delTrigger;
        td.td_CountChanges = gtkToggleButtonActive(d->d_CountChanges);
        td.td_LogChanges = gtkToggleButtonActive(d->d_LogChanges);

        long error = OK;
        Table *newTab;
        if (!d->d_Table) {
    	    error = SE_NewTable(pr, &td, d->d_Parent, &newTab);
            if (error == OK) {
                NewTableFinish(d->d_StructureEditorData, newTab);
            }
        } else {
	    error = SE_ChangeTable(d->d_Table, &td);
            if (error == OK) {
                ChangeTableFinish(d->d_StructureEditorData, d->d_Table);
            }
        }

    	if (error) {
            Ask2(go, GCS(MSG_ErrorMessage), AT_ERROR, 
                    GCS(MSG_StarOk), GES(error));
	    gtk_widget_grab_focus(d->d_Name);
            return;
        }
    }

    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

static void createEditTable(StructureEditorData *sed, 
        Project *pr, Table *tab, DisplayObject *parent)
{
    EditTableData *d = g_new0(EditTableData, 1);

    d->d_Project = pr;
    d->d_Table = tab;
    d->d_Parent = parent;
    d->d_StructureEditorData = sed;

    d->d_Name = EntryObject,
	"max-length", ID_NAME_LEN,
	"xalign", 0.5,
	End;
    GuiSetTooltip(d->d_Name, GCS(MSG_Table_NameBH));

    GuiObject top = HGroup,
        ChildLabelWidget(GCS(MSG_Name), d->d_Name),
        End;

    numRecordsContent[0] = GCS(MSG_Unlimited);
    numRecordsContent[1] = GCS(MSG_Table_ExactlyOne);
    numRecordsContent[2] = NULL;
    d->d_NumRecords = RadioBoxObject,
	"entries", numRecordsContent,
	End;
    GuiObject numRecordsBox = GuiMakeTooltip(d->d_NumRecords,
            GCS(MSG_Table_NumberOfRecordsBH));

    GuiObject numRecords = Frame(GCS(MSG_Table_NumberOfRecords)),
        "child", GroupAlignment(true),
            "child", HCenter(numRecordsBox),
            End,
        End;

    d->d_NewTriggerStr = EntryObject,
	"max-length", TABLE_TRIGGER_LEN,
        "xalign", 0.5,
        End;

    d->d_DelTriggerStr = EntryObject,
	"max-length", TABLE_TRIGGER_LEN,
        "xalign", 0.5,
        End;

    GuiObject triggers = Frame(GCS(MSG_Table_TriggerFunctions)),
        "child", GroupAlignment(true),
            "child", VGroup,
                "columns", 2,
                "child", GuiLabel(GCS(MSG_Table_Trigger_New), 
                        d->d_NewTriggerStr), "weight", 0,
                "child", UserFunctionsPopup(d->d_NewTriggerStr, pr),
                "child", GuiLabel(GCS(MSG_Table_Trigger_Delete),
                        d->d_DelTriggerStr), "weight", 0,
                "child", UserFunctionsPopup(d->d_DelTriggerStr, pr),
                End,
            End,
	End;

    d->d_CountChanges = CheckButtonObject, 
        "label", GCS(MSG_CountChanges),
	"use-underline", true,
        "active", true,
        End;
    GuiSetTooltip(d->d_CountChanges, GCS(MSG_Table_CountChangesBH));
    d->d_LogChanges = CheckButtonObject, 
        "label", GCS(MSG_LogChanges),
	"use-underline", true,
        "active", true,
        End;
    GuiSetTooltip(d->d_LogChanges, GCS(MSG_Table_LogChangesBH));

    GuiObject group = VGroup,
        "child", top, "vert-weight", 0,
        "child", numRecords, "vert-weight", 0,
        "child", triggers, "vert-weight", 0,
        "child", HGroup,
            "child", d->d_CountChanges, 
            "child", d->d_LogChanges, 
            End, "vert-weight", 0,
        End;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", tab? GCS(MSG_ChangeTable) : GCS(MSG_NewTable),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, tab? "Changing tables" : "Creating tables");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    GuiAddEntryCallbacks(d->d_Name, d->d_Ok);
    GuiAddEntryCallbacks(d->d_NewTriggerStr, d->d_Ok);
    GuiAddEntryCallbacks(d->d_DelTriggerStr, d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(closeEditTable), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    if (tab) {
        gtk_entry_set_text(GTK_ENTRY(d->d_Name), EmptyNullStr(TableName(tab)));
        gtk_radio_box_set_active(GTK_RADIO_BOX(d->d_NumRecords), 
                TableExactlyOneRecord(tab)? 1 : 0);
        gtk_entry_set_text(GTK_ENTRY(d->d_NewTriggerStr), 
                 EmptyNullStr(TableNewTrigger(tab)));
        gtk_entry_set_text(GTK_ENTRY(d->d_DelTriggerStr), 
                EmptyNullStr(TableDelTrigger(tab)));
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_CountChanges),
                TableCountChanges(tab));
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->d_LogChanges),
                TableLogChanges(tab));
    }

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }
}

void GtkNewTable(StructureEditorData *sed, Project *pr, DisplayObject *parent)
{
    createEditTable(sed, pr, 0, parent);
}

void GtkChangeTable(StructureEditorData *sed, Table *tab)
{
    createEditTable(sed, TableProject(tab), tab, 0);
}

#endif

