//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkFilter.cc
//
// Creation date: 02.01.2005
// Last modified: 17.01.2009
//

#include "filter.h"
#include "gtk/classes/gtkgroup.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "lang/compile.h"
#include "lang/program.h"
#include "project.h"
#include "table.h"
#include "memory.h"
#include "appstrings_cat.h"

#define BUFSIZE 500
static char buffer[BUFSIZE+1];

struct EditFilterData
{
    Project *d_Project;
    Table *d_Table;
    char d_WindowTitle[100];
    GuiObject d_Fields;
    GuiObject d_Expression;
    GuiObject d_Ok;
    GuiObject d_Cancel;
};

static void insertText(GuiObject entry, const gchar *text)
{
    gtk_editable_delete_selection(GTK_EDITABLE(entry));

    const gchar *contents = gtk_entry_get_text(GTK_ENTRY(entry));

    if (g_utf8_strlen(contents, -1) + g_utf8_strlen(text, -1) > BUFSIZE) {
	GuiBeep();
	return;
    }

    gint pos = gtk_editable_get_position(GTK_EDITABLE(entry));
    gtk_widget_grab_focus(entry);
    gtk_editable_insert_text(GTK_EDITABLE(entry), text, -1, &pos);
    if (text[strlen(text)-1] == ')') {
	pos--;
    }
    gtk_editable_set_position(GTK_EDITABLE(entry), pos);
}

static void insertBracketed(GuiObject button, EditFilterData *d)
{
    const gchar *label = gtk_button_get_label(GTK_BUTTON(button));
    if (label) {
        gchar *buf = g_strdup_printf("(%s )", label);
        insertText(d->d_Expression, buf);
        g_free(buf);
    }
}

static void insertNormal(GuiObject button, EditFilterData *d)
{
    const gchar *label = gtk_button_get_label(GTK_BUTTON(button));
    if (label) {
	gchar *buf = g_strdup_printf("%s ", label);
        insertText(d->d_Expression, buf);
        g_free(buf);
    }
}

static void editFilterOnRowActivated(GtkTreeView *tree,
	GtkTreePath *path, GtkTreeViewColumn *column, EditFilterData *d)
{
    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
    GtkTreeIter iter;
    if (gtk_tree_model_get_iter(model, &iter, path)) {
	gchar *name = 0;
	gtk_tree_model_get(model, &iter, 0, &name, -1);
	if (name) {
	    gchar *buf = g_strdup_printf("%s ", name);
	    insertText(d->d_Expression, buf);
            g_free(buf);
	    g_free(name);
	}
    }
}

static void finish(GuiObject go, gint response, EditFilterData *d)
{
    if (response == GTK_RESPONSE_ACCEPT) {
	GuiBusyCursor busy;
        const char *text = gtkEntryText(d->d_Expression);
	Program *prg = 0;
        const char *error = 0;
        long line = 1, col = 1;

        if (!text) {
            error = GCS(MSG_Err_Internal);
	} else {
            error = CompileFilter(d->d_Table, text, strlen(text), &prg, 
		    &line, &col);
	}

        if (error) {
	    gtk_window_set_title(GTK_WINDOW(go), error);
            if (col > 0 && col <= BUFSIZE) {
		gtk_widget_grab_focus(d->d_Expression);
		gtk_editable_set_position(GTK_EDITABLE(d->d_Expression), col-1);
	    }
            GuiBeep();
	    return;
    	}
	gtk_widget_hide(go);
        SetFilter(d->d_Table, prg);
    }

    GuiSetProjectSleep(d->d_Project, false);
    gtk_widget_destroy(go);
}

static struct
{
    const char *fld_String;
    short fld_InsertFuncNum;
} filterLabelDatas[] =
{
    { "=", 	0 },
    { "<>", 	0 },
    { "=*", 	0 },
    { "<>*", 	0 },
    { "<", 	0 },
    { "<=", 	0 },
    { "<*", 	0 },
    { "<=*", 	0 },
    { ">", 	0 },
    { ">=", 	0 },
    { ">*", 	0 },
    { ">=*", 	0 },
    { "+", 	0 },
    { "-", 	0 },
    { "*", 	0 },
    { "/", 	0 },
    { "DIV", 	0 },
    { "MOD", 	0 },
    { "LIKE", 	0 },
    { "NOT", 	0 },
    { "AND", 	0 },
    { "OR", 	0 },
    { "TRUE", 	1 },
    { "NIL", 	1 },
};

#define NUM_FILTERLABELS (sizeof(filterLabelDatas)/sizeof(filterLabelDatas[0]))

static GuiObject createEditFilter(Table *tab)
{
    buffer[0] = 0;

    Program *prg = TableFilterProgram(tab);
    if (prg) {
        SourceText *st = GetSourceText(prg);
        if (!st) {
	    return 0;
	}

        const char *text = SourceTextText(st);
        if (text) {
            sprintf(buffer, "%.*s", BUFSIZE, text);
	}
        FreeSourceText(st);
    }

    Project *pr = TableProject(tab);
    EditFilterData *d = g_new0(EditFilterData, 1);
    d->d_Project = pr;
    d->d_Table = tab;

    GuiObject filterLabelsGroup = VGroup,
        "columns", 4,
	"same-size", true,
	"spacing", 0,
	End;
    GuiObject filterLabelsFrame = Frame(GCS(MSG_Filter_Operators)),
        "child", GroupAlignment(true),
	    "child", filterLabelsGroup,
	    End,
	End;

    for (size_t i=0; i < NUM_FILTERLABELS; i++) {
        GuiObject button = GuiSimpleButton(filterLabelDatas[i].fld_String);
        if (filterLabelDatas[i].fld_InsertFuncNum == 0) {
	    g_signal_connect(button, "clicked", G_CALLBACK(insertBracketed), d);
        } else {
	    g_signal_connect(button, "clicked", G_CALLBACK(insertNormal), d);
        }
	gtk_group_add_child(GTK_GROUP(filterLabelsGroup), button, 1, 0);
    }

    GtkListStore *store = gtk_list_store_new(1, G_TYPE_STRING);
    d->d_Fields = TreeViewObject,
        "model", store,
	"headers-visible", false,
	End;
    g_signal_connect(d->d_Fields, "row-activated", 
	    G_CALLBACK(editFilterOnRowActivated), d);
    List *fields = TableFields(tab);
    if (fields) {
	Field *field;
        FORNODE(field, fields) {
            if (FieldType(field) != TYPE_BUTTON) {
		GtkTreeIter iter;
		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter, 0, FieldName(field), -1);
	    }
	}
    }
    g_object_unref(G_OBJECT(store));

    GtkCellRenderer *cell = gtk_cell_renderer_text_new();
    GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
	    "Item", cell, "text", 0, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(d->d_Fields), column);
    GtkTreeSelection *selection = 
	gtk_tree_view_get_selection(GTK_TREE_VIEW(d->d_Fields));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);

    GuiObject fieldsFrame = Frame(GCS(MSG_Fields)),
        "child", GroupAlignment(true),
	    "child", ScrolledWindowObject,
		"hscrollbar-policy", GTK_POLICY_NEVER,
		"shadow-type", GTK_SHADOW_IN,
	        "child", d->d_Fields,
		End,
	    End,
	End;

    d->d_Expression = EntryObject,
	// MUIA_ControlChar, Shortcut(GCS(MSG_Filter_Expression)),
	"max-length", BUFSIZE,
	"text", buffer,
	"xalign", 0.5,
	"activates-default", true,
	End;

    GuiObject group = VGroup,
        "child", HGroup,
	    "same-height", true,
	    "child", fieldsFrame,
	    "child", filterLabelsFrame,
	    End, "vert-weight", 0,
	"child", d->d_Expression, "vert-weight", 0,
	End;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    sprintf(d->d_WindowTitle, "%s: %s", GCS(MSG_ChangeFilter), TableName(tab));
    GuiObject dialog = DialogObject,
        "title", d->d_WindowTitle,
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Changing filters");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_widget_grab_focus(d->d_Expression);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(finish), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    gint width, height;
    gtk_window_get_size(GTK_WINDOW(dialog), &width, &height);
    width += 50;
    gtk_window_set_default_size(GTK_WINDOW(dialog), width, height);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
	gtk_widget_destroy(dialog);
        dialog = 0;
    }

    return dialog;
}

void GuiChangeFilter(Table *tab)
{
    createEditFilter(tab);
}

