//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkGui.cc
//
// Creation date: 21.03.2004
// Last modified: 11.04.2022
//

#define GDK_PIXBUF_ENABLE_BACKEND
#include "gtk/gtkGui.h"
#include "gtk/gtkMenus.h"
#include "gtk/gtkBuildGuiField.h"
#include "gtk/classes/gtkexteditor.h"
#include "gtk/classes/gtkpopimage.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include <gtk/gtk.h>
#ifdef MACOSX
#include <gtkosxapplication.h>
#endif
#include <gdk-pixbuf/gdk-pixbuf-io.h>
#include "preferences.h"
#include "status.h"
#include "server.h"
#include "menu.h"
#include "menuprefs.h"
#include "menueditor.h"
#include "menuhelp.h"
#include "programeditor.h"
#include "structureeditor.h"
#include "requesters.h"
#include "break.h"
#include "programoutput.h"
#include "search.h"
#include "order.h"
#include "filter.h"
#include "imexport.h"
#include "setformats.h"
#include "queryeditor.h"
#include "guidosutil.h"
#include "dosutil.h"
#include "util.h"
#include "strhandle.h"
#include "memory.h"
#include "exec.h"
#include "appstrings_cat.h"
#include "builddate.h"
#include "version.h"
#include "images/bee16x16_with_bg.c"
#include "images/bee32x32.c"
#include "images/bee48x48.c"
#include "images/bee64x64.c"
#include <fcntl.h>
#include <locale.h>
#include <dirent.h>
#include <unistd.h>

extern "C" { 
    void _gdk_pixbuf__ilbm_fill_vtable(GdkPixbufModule *module);
    void _gdk_pixbuf__ilbm_fill_info(GdkPixbufFormat *info);
}

#ifdef MACOSX
GtkosxApplication *BeeBaseApp = 0;

static gboolean onTermination(void)
{
    GuiSendQuitMessage();
    return TRUE;		// block termination
}
#endif

char DefaultRealFormat = FORMAT_DECIMALPOINT;
char DefaultDateFormat = FORMAT_DAYMONTHYEAR;

static const char titleSep[] = " | ";

static GdkCursor *busyCursor = 0;

void CloseGui(void) 
{
    RestoreSignalHandlers();

    GuiMenuExit();
    GuiUtilExit();

    if (busyCursor) {
#if GTK_MAJOR_VERSION == 2
	gdk_cursor_unref(busyCursor);
#else
        g_object_unref(G_OBJECT(busyCursor));
#endif
	busyCursor = 0;
    }
    CodesetExit();

#ifdef MACOSX
    if (BeeBaseApp) {
	g_object_unref(G_OBJECT(BeeBaseApp));
	BeeBaseApp = 0;
    }
#endif
}

const char *OpenGui(int *pargc, char **argv)
{
#if (defined WINDOWS)
    gchar *winLocale = g_win32_getlocale();
    g_setenv("LANG", winLocale, false);
    g_free(winLocale);
#endif

    //
    // textdomain should be defined before gtk_init
    // On Mac OS X this results into non-translated 
    // GTK strings (e.g. in the file dialog)
    //
#ifndef MACOSX
    textdomain(APP_NAME);
#endif

    if (!gtk_init_check(pargc, &argv)) {
	return "Can't initialize application!";
    }

#ifdef MACOSX
    textdomain(APP_NAME);

    BeeBaseApp = (GtkosxApplication *)g_object_new(
	    GTKOSX_TYPE_APPLICATION, NULL);
    g_signal_connect(BeeBaseApp, "NSApplicationBlockTermination",
    	    G_CALLBACK(onTermination), 0);
#endif

#if (defined WINDOWS)
    // printf("HOME: %s\n", getenv("HOME"));
    // printf("Home: %s\n", getenv("Home"));
    // printf("g_get_home_dir: %s\n", g_get_home_dir());
    const gchar *home = g_get_home_dir();
    if (home) {
        g_setenv("HOME", home, true);
    }

    gchar *path = g_win32_get_package_installation_directory_of_module(0);
    // printf("BeeBase: %s\n", path);
    if (path) {
        g_setenv(APP_NAME, path, false);
        gchar *locale_dir = g_build_filename(path, "locale", NULL);
        if (locale_dir != NULL) {
            bindtextdomain(APP_NAME, locale_dir);
            g_free(locale_dir);
        }
        g_free(path);
    }
#elif (defined MACOSX)
    const char *prefix;
    if (gtkosx_application_get_bundle_id()) {
	prefix = gtkosx_application_get_resource_path();
    } else {
	prefix = "/Applications/BeeBase.app/Contents/Resources";
    }
    
    gchar *mbDir = g_strconcat(prefix, "/share/BeeBase", NULL);
    g_setenv(APP_NAME, mbDir, false);
    g_free(mbDir);

    gchar *localeDir = g_strconcat(prefix, "/share/locale", NULL);
    bindtextdomain(APP_NAME,  localeDir);
    bindtextdomain("gtk20",  localeDir);
    bindtextdomain("gtk20-properties",  localeDir);
    bindtextdomain("glib20",  localeDir);
    bindtextdomain("atk10",  localeDir);
    g_free(localeDir);
#elif (defined UNIX)
    g_setenv(APP_NAME, "/usr/share/BeeBase", false);
#endif

    const char *lang = getenv("LANG");
    if (lang && strncmp(lang, "de", 2) == 0) {
	DefaultRealFormat = FORMAT_DECIMALCOMMA;
	DefaultDateFormat = FORMAT_DAYMONTHYEAR;
    }
    const struct lconv *lc = localeconv();
    if (lc && lc->decimal_point && lc->decimal_point[0] == ',') {
	DefaultRealFormat = FORMAT_DECIMALCOMMA;
    }

    g_set_application_name(GCS(MSG_APP));

    const char *error = CodesetInit();
    if (error) {
	return error;
    }

#if 0
    gdk_pixbuf_add_format("ilbm",
	_gdk_pixbuf__ilbm_fill_vtable,
	_gdk_pixbuf__ilbm_fill_info);
#endif

    GList *iconList = 0;

    for (int i=0; i < 4; i++) {
        GdkPixbuf *pixbuf = 0;
        int size = 0;

        switch (i) {
            case 0:
                pixbuf = gdk_pixbuf_new_from_data(bee16x16_image.pixel_data,
                        GDK_COLORSPACE_RGB, bee16x16_image.bytes_per_pixel == 4,
                        8, bee16x16_image.width, bee16x16_image.height, 
                        bee16x16_image.width * bee16x16_image.bytes_per_pixel,
                        NULL, NULL);
                size = 16;
                break;
            case 1:
                pixbuf = gdk_pixbuf_new_from_data(bee32x32_image.pixel_data,
                        GDK_COLORSPACE_RGB, bee32x32_image.bytes_per_pixel == 4,
                        8, bee32x32_image.width, bee32x32_image.height, 
                        bee32x32_image.width * bee32x32_image.bytes_per_pixel,
                        NULL, NULL);
                size = 32;
                break;
            case 2:
                pixbuf = gdk_pixbuf_new_from_data(bee48x48_image.pixel_data,
                        GDK_COLORSPACE_RGB, bee48x48_image.bytes_per_pixel == 4,
                        8, bee48x48_image.width, bee48x48_image.height, 
                        bee48x48_image.width * bee48x48_image.bytes_per_pixel,
                        NULL, NULL);
                size = 48;
                break;
            case 3:
                pixbuf = gdk_pixbuf_new_from_data(bee64x64_image.pixel_data,
                        GDK_COLORSPACE_RGB, bee64x64_image.bytes_per_pixel == 4,
                        8, bee64x64_image.width, bee64x64_image.height, 
                        bee64x64_image.width * bee64x64_image.bytes_per_pixel,
                        NULL, NULL);
                size = 64;
                break;
        }

        if (pixbuf) {
            gtk_icon_theme_add_builtin_icon(BEE_ICON_NAME, size, pixbuf);
            iconList = g_list_append(iconList, pixbuf);
        }
    }

    if (iconList) {
        gtk_window_set_default_icon_list(iconList);
    }

    busyCursor = gdk_cursor_new(GDK_WATCH);

    GuiUtilInit();
    GuiMenuInit();
    InstallSignalHandlers();

    return 0;
}

void GuiPrefsUpdated(void)
{
    UpdateMenu(GuiActiveWindow(), true);
}

void GuiBeep()
{
    gdk_beep();
}

static const char *artists[] = {
    "Thomas Fricke",
    "Adrian Maleska",
    "Pascal Marcelin",
    "Martin Merz",
    "Christoph Pölzl",
    "Sébastien Pölzl",
    "Ralph Reuchlein",
    0
};

static const char *authors[] = {
    "Steffen Gutmann <beebase.bbs@gmail.com>",
    0
};

static const char *documenters[] = {
    "Mats Granstrom",
    "Steffen Gutmann",
    "Magnus Hammarstrom",
    "Ralph Reuchlein",
    0
};

static const char translators[] = 
    "Stéphane Aulery (fr)\n"
    "Alexandre Balaban (fr)\n"
    "Yuri Chornoivan (uk)\n"
    "Philippe Ferrucci (fr)\n"
    "Stefan Haubenthal (de)\n"
    "Pascal Marcelin (fr)\n"
    "Gilles Mathevet (fr)\n"
    "Lionel Muller (fr)\n"
    "Ralph Reuchlein (de)\n"
    "Jaap Verhage (nl)";

static const char license[] =
"This program is free software; you can redistribute it and/or\n"
"modify it under the terms of the GNU General Public License\n"
"as published by the Free Software Foundation; either version 3\n"
"of the License, or (at your option) any later version.\n"
"\n"
"This program is distributed in the hope that it will be useful,\n"
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"GNU General Public License for more details.\n"
"\n"
"You should have received a copy of the GNU General Public License\n"
"along with this program; if not, see <https://www.gnu.org/licenses/>.";


#define BEE_IMAGE_FILENAME "BeeBase:images/Bee.png"
#define BEE2_IMAGE_FILENAME "BeeBase:images/BeeOnSmallComb.png"

void GuiAbout()
{
    GuiSetSleep(true);

    gchar *path = gtkExpandPathname(BEE2_IMAGE_FILENAME);
    GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(path, 0);
    g_free(path);
    if (!pixbuf) {
        path = gtkExpandPathname(BEE_IMAGE_FILENAME);
        pixbuf = gdk_pixbuf_new_from_file(path, 0);
        g_free(path);
    }

    gtk_show_about_dialog(0,
            "window-position", GTK_WIN_POS_CENTER,
	    "artists", artists,
	    "authors", authors,
	    "comments", GCS(MSG_About_ShortDescription), 
	    "copyright", "Copyright © 1998-" BUILD_YEAR " Steffen Gutmann",
	    "documenters", documenters,
	    "license", license,
	    "logo", pixbuf,
	    "translator-credits", translators,
	    "version", BB_VERSTR,
	    "website", "https://beebase.sourceforge.io",
	    NULL);

    if (pixbuf) {
	g_object_unref(G_OBJECT(pixbuf));
    }

    GuiSetSleep(false);
}

void GuiSendQuitMessage(void) 
{
    gtk_main_quit();
}

static ulong sleepCount = 0;

ulong GuiSleep() 
{ 
    return sleepCount; 
}

static void guiSetBusyCursor(GdkWindow *win)
{
    GdkWindow *busy = (GdkWindow *)g_object_get_data(G_OBJECT(win), "busy");
    if (!busy) {
	// we place an input-only window over the whole top-level window
	// and set a busy pointer
	GdkWindowAttr attr;
	attr.event_mask = GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
	    GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK | GDK_ALL_EVENTS_MASK;
	attr.x = 0;
	attr.y = 0;
	attr.width = G_MAXINT;
	attr.height = G_MAXINT;
	attr.wclass = GDK_INPUT_ONLY;
	attr.window_type = GDK_WINDOW_CHILD;
	attr.cursor = busyCursor;

	busy = gdk_window_new(win, &attr, GDK_WA_X | GDK_WA_Y | GDK_WA_CURSOR);
	// Basically we want to keep the busy window above the other windows
	// of the parent window.  Note sure if this is achieved by
	// the following call.  It causes Warnings on MacOS (not top-level)
	// gdk_window_set_keep_above(busy, true);  // questionable

    	g_object_set_data_full(G_OBJECT(win), "busy", busy, 
		(GDestroyNotify)gdk_window_destroy);
    }
    gdk_window_show(busy);
    // gdk_window_set_accept_focus(win, false);
}

static void guiClearBusyCursor(GdkWindow *win)
{
    GdkWindow *busy = (GdkWindow *)g_object_get_data(G_OBJECT(win), "busy");
    if (busy) {
	gdk_window_hide(busy);
	// gdk_window_set_accept_focus(win, true);
#ifdef MACOSX
	// for some reasons MacOSX does not refresh the mouse cursor
	// this seems to work now with gtk+-2.20.1
	// gdk_window_set_cursor(win, 0);
#endif
    }
}

static void guiUpdateBusyCursor(GdkWindow *win)
{
    guint block = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(win), "block"));
    guint count = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(win), "sleep"));

    if (!block) {
        if (count == 0) {
            guiClearBusyCursor(win);
        } else {
            guiSetBusyCursor(win);
        }
    }
}

void GuiBlockBusyPointer(GtkWidget *w, bool block)
{
    GdkWindow *win = gtk_widget_get_window(w);
    if (win) {
        g_object_set_data(G_OBJECT(win), "block", GUINT_TO_POINTER(block));
        guiUpdateBusyCursor(win);
    }
}

static void guiSetWindowSleep(GtkWidget *w, bool sleep)
{
    if (GTK_IS_WINDOW(w)) {
        GdkWindow *win = gtk_widget_get_window(w);
	if (win) {
            bool update = false;
	    guint count = GPOINTER_TO_UINT(
		    g_object_get_data(G_OBJECT(win), "sleep"));
	    if (sleep) {
                update = (count++ == 0);
            } else if (count > 0) {
                update = (--count == 0);
            }
	    g_object_set_data(G_OBJECT(win), "sleep", GUINT_TO_POINTER(count));
            if (update) {
                guiUpdateBusyCursor(win);
            }
	}
    }
}

void GuiClearWindowSleep(GtkWidget *w)
{
    if (GTK_IS_WINDOW(w)) {
        GdkWindow *win = gtk_widget_get_window(w);
	if (win) {
	    guiClearBusyCursor(win);
	    g_object_set_data(G_OBJECT(win), "sleep", GUINT_TO_POINTER(0));
	}
    }
}

void GuiSetSleep(bool sleep) 
{
    if (sleep) {
	sleepCount++;
    } else {
	sleepCount--;
    }

    GList *toplevels = gtk_window_list_toplevels();
    for (GList *list = toplevels; list; list = list->next) {
	GtkWidget *w = (GtkWidget *)list->data;
	guiSetWindowSleep(w, sleep);
    }
    g_list_free(toplevels);
}

ulong GuiProjectSleep(Project *pr)
{
    GuiObject w = ProjectRootWindow(pr);
    ulong sleep = 0;

    if (GTK_IS_WINDOW(w)) {
        GdkWindow *win = gtk_widget_get_window(w);
	if (win) {
	    sleep = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(win), "sleep"));
        }
    }

    return sleep;
}

void GuiSetProjectSleep(struct Project *pr, bool sleep)
{
    GList *toplevels = gtk_window_list_toplevels();
    for (GList *list = toplevels; list; list = list->next) {
	GtkWidget *w = (GtkWidget *)list->data;
	GtkWindow *top = GTK_WINDOW(w);
	if (GuiWindowProject(GTK_WIDGET(top)) == pr) {
	    guiSetWindowSleep(w, sleep);
	}
    }
    g_list_free(toplevels);
}

char *GuiEditorText(GuiObject go, struct MemPool *memPool)
{
    g_return_val_if_fail(GTK_IS_TEXT_VIEW(go), 0);

    GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(go));
    GtkTextIter start, end;
    gtk_text_buffer_get_start_iter(buffer, &start);
    gtk_text_buffer_get_end_iter(buffer, &end);
    gchar *text = gtk_text_buffer_get_text(buffer, &start, &end, true);
    char *result = AllocString(memPool, text);
    g_free(text);

    return result;
}

void GuiSetEditorText(GuiObject go, const char *str)
{
    g_return_if_fail(GTK_IS_TEXT_VIEW(go));

    GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(go));
    gtk_text_buffer_set_text(buffer, EmptyNullStr(str), -1);
}

void GuiSetEditorChanged(GuiObject go, bool changed)
{
    g_return_if_fail(GTK_IS_TEXT_VIEW(go));

    GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(go));
    gtk_text_buffer_set_modified(buffer, changed);
}

void GuiSetRegisterPage(GuiObject go, long pos)
{
    gtk_notebook_set_current_page(GTK_NOTEBOOK(go), pos);
}

long GuiRegisterActivePage(GuiObject go) 
{
    return gtk_notebook_get_current_page(GTK_NOTEBOOK(go));
}

bool GuiOpenWindow(GuiObject win) 
{
    if (win) {
	gtk_widget_show(win);
	return gtk_widget_get_visible(win);
    }
    return false;
}

void GuiCloseWindow(GuiObject win) 
{
    if (win) {
	gtk_widget_hide(win);
    }
}

short GuiObjectWeight(GuiObject obj, bool horiz)
{
    GuiObject parent = obj;

    do {
        obj = parent;
        parent = gtk_widget_get_parent(obj);

        if (!parent) {
            return 100;
        }
    } while (!GTK_IS_GROUP(parent));

    guint weight;
    gtk_group_query_child_packing(GTK_GROUP(parent), obj, 
            horiz? &weight : 0, 
            horiz? 0 : &weight);

    return weight;
}

const char *GuiObjectHelpNode(GuiObject go)
{
    const char *help = 0;

    if (go) {
        help = (const char *)g_object_get_data(G_OBJECT(go), "help");
    }
    return help;
}

GuiObject GuiObjectParent(GuiObject go)
{
    GuiObject parent = 0;

    if (go) {
        parent = gtk_widget_get_parent(go);
    }
    return parent;
}

static gint PosOffsetX = 0, PosOffsetY = 0;

bool GuiGetWindowDimensions(GuiObject win, struct WindowDimensions *wd)
{
    if (win && gtk_widget_get_visible(win) && wd) {
        gint x, y, width, height;

	gtk_window_get_position(GTK_WINDOW(win), &x, &y);
	gtk_window_get_size(GTK_WINDOW(win), &width, &height);
	x += PosOffsetX;
	y += PosOffsetY;

	wd->wd_LeftEdge = x;
	wd->wd_TopEdge = y;
	wd->wd_Width = width; 
	wd->wd_Height = height;

        return true;
    }
    return false;
}

#ifndef MACOSX

#ifndef timercmp
#define timercmp(t1, t2, op) \
	(((t1)->tv_sec op (t2)->tv_sec) || \
     	(((t1)->tv_sec == (t2)->tv_sec) && ((t1)->tv_usec op (t2)->tv_usec)))
#endif

static gboolean onConfigure(GtkWidget *win)
{
    GObject *o = G_OBJECT(win);
    gulong id = GPOINTER_TO_INT(g_object_get_data(o, "MoveConfigureHandler"));
    gint newx = GPOINTER_TO_INT(g_object_get_data(o, "MoveNewX"));
    gint newy = GPOINTER_TO_INT(g_object_get_data(o, "MoveNewY"));
    gint x, y;
    GTimeVal tv, cur;

    tv.tv_sec = GPOINTER_TO_INT(g_object_get_data(o, "MoveTimeSec"));
    tv.tv_usec = GPOINTER_TO_INT(g_object_get_data(o, "MoveTimeUSec"));
    g_get_current_time(&cur);
    g_time_val_add(&tv, 200000);        /* 200 ms */
    if (timercmp(&cur, &tv, <)) {
        gtk_window_get_position(GTK_WINDOW(win), &x, &y);
        PosOffsetX = newx - x;
        PosOffsetY = newy - y;
        //printf("Pos: (%d %d), off: (%d %d)\n", x, y, PosOffsetX, PosOffsetY);
    } else {
        g_signal_handler_disconnect(o, id);
        g_object_set_data(o, "MoveConfigureHandler", 0);
    }

    return false;
}

#endif

void GuiSetWindowDimensions(GuiObject win, const struct WindowDimensions *wd)
{
    if (win && wd) {
	GdkScreen *screen = gtk_window_get_screen(GTK_WINDOW(win));
	gint screen_width, screen_height;
	gint width = wd->wd_Width, height = wd->wd_Height;
	gint x = wd->wd_LeftEdge, y = wd->wd_TopEdge;
        gboolean setWidth = (width > 0);
        gboolean setHeight = (height > 0);
        gboolean setPos = (x >= 0 || y >= 0);

        if (!setWidth) {
            gtk_window_get_size(GTK_WINDOW(win), &width, NULL);
        }
        if (!setHeight) {
            gtk_window_get_size(GTK_WINDOW(win), NULL, &height);
        }

        if (setWidth || setHeight) {
            gtk_window_resize(GTK_WINDOW(win), width, height);
        }

        if (!setPos) {
            return;
        }

	if (screen) {
	    screen_width = gdk_screen_get_width (screen);
	    screen_height = gdk_screen_get_height (screen);
	} else {
	    screen_width = gdk_screen_width ();
	    screen_height = gdk_screen_height ();
      	}
	x = CLAMP(x, 0, screen_width - width);
      	y = CLAMP(y, 0, screen_height - height);

        gtk_window_move(GTK_WINDOW(win), x, y);

        if (gtk_widget_get_visible(win)) {
#ifdef MACOSX
	    gint ax, ay;
	    gtk_window_get_position(GTK_WINDOW(win), &ax, &ay);
	    PosOffsetX = x - ax;
	    PosOffsetY = y - ay;
#else
	    GObject *o = G_OBJECT(win);
	    GTimeVal tv;

            g_get_current_time(&tv);
            g_object_set_data(o, "MoveTimeSec", GINT_TO_POINTER(tv.tv_sec));
            g_object_set_data(o, "MoveTimeUSec", GINT_TO_POINTER(tv.tv_usec));
            g_object_set_data(o, "MoveNewX", GINT_TO_POINTER(x));
            g_object_set_data(o, "MoveNewY", GINT_TO_POINTER(y));
            if (!g_object_get_data(o, "MoveConfigureHandler")) {
                gulong id = g_signal_connect(o, "configure-event",
                        G_CALLBACK(onConfigure),0);
                g_object_set_data(o, "MoveConfigureHandler",
                        GINT_TO_POINTER(id));
            }
#endif
        }
    }
}

struct Project *GuiActiveProject(void) 
{
    return GuiWindowProject(GuiActiveWindow());
}

/* 
** In GTK, the is-active property might not be set even though the
** window is drawn with an active window border.
*/
GuiObject GuiActiveWindow(void)
{
    GuiObject win;
    GuiObject active = 0;

    GList *toplevels = gtk_window_list_toplevels();
    for (GList *list = toplevels; list; list = list->next) {
	if (GTK_IS_WINDOW(list->data) && 
                gtk_widget_get_visible(GTK_WIDGET(list->data)) &&
                gtk_window_is_active(GTK_WINDOW(list->data))) {
            active = GTK_WIDGET(list->data);
            break;
	}
    }
    g_list_free(toplevels);
    win = active;

    if (win == NULL) {
        GdkScreen *s = gdk_screen_get_default();
        GdkWindow *w = gdk_screen_get_active_window(s);
        gpointer data = 0;

        if (w) {
            gdk_window_get_user_data(w, &data);
            if (GTK_IS_WINDOW(data)) {
                win = GTK_WIDGET(data);
            }
        } 
    }

#if 0
    if (win) {
	printf("Active window: %p -- %s, active: %p\n", 
                win, gtk_window_get_title(GTK_WINDOW(win)), active);
    }
#endif

    return win;
}

GuiObject GuiActiveObject(GuiObject wi) 
{
    GuiObject active = 0;

    if (wi && GTK_IS_WINDOW(wi)) {
	active = gtk_window_get_focus(GTK_WINDOW(wi));
#if 0
        if (active && !GTK_IS_WIDGET(active)) {
            printf("bogus gtk_window_get_focus\n");
            active = 0;
        }
#endif
    }
    return active;
}

bool GuiIsEditorObject(GuiObject go)
{
    return GTK_IS_EXT_EDITOR(go);
}

void GuiSetSelected(GuiObject go, bool selected)
{
    if (go) {
	guint id = g_signal_lookup("toggled", GTK_TYPE_TOGGLE_BUTTON);
	guint handler = g_signal_handler_find(go, G_SIGNAL_MATCH_ID, id, 
		0, 0, 0, 0);
	if (handler) {
	    g_signal_handler_block(go, handler);
	}
	g_object_set(G_OBJECT(go), "active", selected, NULL);
	if (handler) {
	    g_signal_handler_unblock(go, handler);
	}
    }
}

void GuiSetEnabled(GuiObject go, bool enabled)
{
    if (go) {
	gtk_widget_set_sensitive(go, enabled);
    }
}

bool GuiGetEnabled(GuiObject go)
{
    bool enabled = FALSE;

    if (go) {
        enabled = gtk_widget_get_sensitive(go);
    }
    return enabled;
}

void GuiSetBackground(GuiObject go, const char *bg)
{
    if (go) {
        gtk_widget_set_bg_spec(go, bg);
    }
}

struct Project *GuiObjectProject(GuiObject go)
{
    GuiObject win = gtk_widget_get_toplevel(go);
    struct Project *pr = GuiWindowProject(win);

    return pr;
}

const char *GuiStringContents(GuiObject go)
{
    const char *text = 0;

    if (go && GTK_IS_ENTRY(go)) {
        text = gtkEntryText(go);
    }
    return text;
}

const char *GuiObjectText(GuiObject go)
{
    if (go == NULL) {
	return NULL;
    }

    g_return_val_if_fail(GTK_IS_ENTRY(go) || GTK_IS_LABEL(go), NULL);

    const gchar *text = 0;
    if (GTK_IS_LABEL(go)) {
	text = gtk_label_get_text(GTK_LABEL(go));
    } else if (GTK_IS_ENTRY(go)) {
	text = gtk_entry_get_text(GTK_ENTRY(go));
    }
    return text;
}

bool GuiWindowOpen(GuiObject wi) 
{
    bool open = false;

    if (wi) {
	open = gtk_widget_get_visible(wi);
    }
    return open;
}

void GuiActivateWindow(GuiObject wi) 
{
    if (wi) {
	gtk_window_present(GTK_WINDOW(wi));
    }
}

void GuiWindowToFront(GuiObject wi) 
{
    if (wi) {
        GdkWindow *window = gtk_widget_get_window(wi);
        if (window) {
            gdk_window_raise(window);
        }
    }
}

void GuiSetActiveGuiObject(GuiObject wi, GuiObject go) 
{
    if (wi) {
	gtk_window_set_focus(GTK_WINDOW(wi), go);
    }
}

void GuiSetWindowTitle(GuiObject wi, const char *title)
{
    if (wi) {
        gchar *mem = NULL;
        const char *prepend = EmptyNullStr(
                (const char *)g_object_get_data(G_OBJECT(wi), "prependTitle"));

        title = EmptyNullStr(title);
        if (prepend[0] != 0) {
            mem = (gchar *)g_malloc(strlen(prepend) + strlen(titleSep) +
                    strlen(title) + 1);
            sprintf(mem, "%s%s%s", prepend, titleSep, title);
            title = mem;
        }
	gtk_window_set_title(GTK_WINDOW(wi), title);
        g_free(mem);
    }
}

void GuiSetPrependWindowTitle(GuiObject wi, const char *prepend)
{
    if (wi) {
        const char *currentPrepend = EmptyNullStr(
                (const char *)g_object_get_data(G_OBJECT(wi), "prependTitle"));
        size_t currentLen = strlen(currentPrepend);
        size_t sepLen = strlen(titleSep);
        const char *title = gtk_window_get_title(GTK_WINDOW(wi));
        prepend = EmptyNullStr(prepend);

        if (strcmp(prepend, currentPrepend) != 0) {
            if (currentLen != 0 && title != NULL &&
                    strlen(title) >= currentLen + sepLen &&
                    strncmp(title, currentPrepend, currentLen) == 0 &&
                    strncmp(title + currentLen, titleSep, sepLen) == 0) {
                title += currentLen + sepLen;
            }
            g_object_set_data_full(G_OBJECT(wi), "prependTitle",
                    prepend[0] != 0? g_strdup(prepend) : NULL, g_free);
            GuiSetWindowTitle(wi, title);
        }
    }
}

void GuiStoreEdits(void) 
{
    GuiFieldStoreEdits(); 
}

static gint guiIdleUpdate(gpointer data)
{
    UpdateMenu(GuiActiveWindow(), false);
    UpdateProjects(false);
    if (g_main_depth() == 1) {
        ServerProcessRequests();
        UpdateProjects(true);
        GuiUpdateProjects();
    }
    return false;
}

static void GuiDoEvent(GdkEvent *event)
{
    switch (event->type) {
	case GDK_DELETE:
	case GDK_KEY_PRESS:
	case GDK_KEY_RELEASE:
	    if (GPOINTER_TO_UINT(g_object_get_data(
			G_OBJECT(event->any.window), "sleep")) == 0) {
		gtk_main_do_event(event);
	    }
	    break;
	case GDK_CONFIGURE:
	case GDK_EXPOSE:
	    gtk_main_do_event(event);
	    if (GPOINTER_TO_UINT(g_object_get_data(
			G_OBJECT(event->any.window), "sleep")) != 0) {
		GdkWindow *busy = (GdkWindow *)g_object_get_data(
			G_OBJECT(event->any.window), "busy");
		if (busy) {
		    gdk_window_raise(busy);	// make sure we are on top
		}
	    }
	    break;

	default:
	    gtk_main_do_event(event);
	    break;
    }
    if (!gdk_events_pending()) {
        g_idle_add(guiIdleUpdate, NULL);
    }
}

void GuiMainLoop(bool (*reallyQuit)(void)) 
{
    gdk_event_handler_set((GdkEventFunc)GuiDoEvent, 0, 0);
#ifdef MACOSX
    gtkosx_application_ready(BeeBaseApp);
#endif
    do {
	gtk_main();
    } while (reallyQuit && !reallyQuit());
}

bool GuiMainLoopIteration(void)
{
    return gtk_main_iteration();
}

GuiObject GtkNewEditor(struct Project *pr, bool scrollbar,
        bool readOnly, bool fixedFont, GuiObject *editorObject)
{
    GuiObject editor = ExtEditorObject,
        "editable", !readOnly,
        "wrap-mode", GTK_WRAP_WORD_CHAR,
        "width-request", 50,
        End;

    GuiObject root = ScrolledWindowObject,
	"shadow-type", GTK_SHADOW_ETCHED_IN,
	"hscrollbar-policy", GTK_POLICY_NEVER,
	"child", editor,
	End;

    if (!scrollbar) {
	g_object_set(G_OBJECT(root), "vscrollbar-policy",GTK_POLICY_NEVER,NULL);
    }
    if (fixedFont) {
	PangoFontDescription *pfd = pango_font_description_new();
	pango_font_description_set_family(pfd, "Monospace");
	//pango_font_description_set_size(pfd, PANGO_SCALE * 12);
#if GTK_MAJOR_VERSION == 2
	gtk_widget_modify_font(editor, pfd);
#else
	gtk_widget_override_font(editor, pfd);
#endif
	pango_font_description_free(pfd);
    }
    g_signal_connect(editor, "load", G_CALLBACK(MenuEditorLoadText), pr);
    g_signal_connect(editor, "save", G_CALLBACK(MenuEditorSaveText), pr);
    g_signal_connect(editor, "external-editor", 
	    G_CALLBACK(MenuEditorExternalEditor), pr);

    if (editorObject) {
	*editorObject = editor;
    }

    return root;
}

/*
** extra gdk stuff
*/

#undef gdk_pixbuf_new_from_file
#undef gdk_pixbuf_animation_new_from_file

#ifdef WINDOWS
#define gdk_pixbuf_new_from_file gdk_pixbuf_new_from_file_utf8
#define gdk_pixbuf_animation_new_from_file gdk_pixbuf_animation_new_from_file_utf8
#endif


static bool match_prefix_mask(const char *buffer, 
	const char *prefix, const char *mask, size_t len)
{
    if (!mask) {
	return (memcmp(buffer, prefix, len) == 0);
    }

    const char *end = buffer + len;
    while (buffer != end) {
	switch (*mask) {
	    case ' ': if (*buffer != *prefix) { return FALSE; } break;
	    case '!': if (*buffer == *prefix) { return FALSE; } break;
	    case 'x': break;
	    case 'z': if (*buffer != 0) { return FALSE; } break;
	    case 'n': if (*buffer == 0) { return FALSE; } break;
	    default:
		DFPRINTF((stderr, "GdkModulePattern: illegal mask char: '%c'.",
			*mask));
		return FALSE;
	}
	buffer++;
	prefix++;
	mask++;
    }

    return TRUE;
}

static void updated_func(GdkPixbuf *pixbuf, int x, int y, int w, int h, 
	gpointer data)
{
    if (data) {
	*(GdkPixbuf **)data = pixbuf;
    }
}

static GdkPixbuf *gdk_pixbuf_ilbm_from_file(const char *filename,GError **error)
{
    GdkPixbuf *pixbuf = NULL;

    GdkPixbufFormat info;
    _gdk_pixbuf__ilbm_fill_info(&info);
    GdkPixbufModulePattern *signature = info.signature;
    if (!signature) {
	return NULL;
    }
    GdkPixbufModule module;
    _gdk_pixbuf__ilbm_fill_vtable(&module);
    if (!module.begin_load || !module.stop_load || !module.load_increment) {
	return NULL;
    }

    if (!filename) {
	return NULL;
    }
    int fd = open(filename, O_RDONLY | O_BINARY);
    if (fd < 0) {
	return NULL;
    }

    static char buffer[1024];
    size_t size = read(fd, buffer, sizeof(buffer));
    bool recognized = FALSE;
    do {
	const char *prefix = (const char *)signature->prefix;
	if (!prefix) {
	    break;		/* end of signatures */
	}
	const char *mask = (const char *)signature->mask;
	size_t len = strlen(mask? mask : prefix);
	if (len <= size) {
	    recognized = match_prefix_mask(buffer, prefix, mask, len);
	}
	signature++;
    } while (!recognized);

    if (recognized) {
	gpointer context = module.begin_load(NULL, NULL, updated_func, 
		(gpointer)&pixbuf, error);
	if (context) {
	    while (module.load_increment(context, (guchar *)buffer, 
                        size, error)) {
		size = read(fd, buffer, sizeof(buffer));
		if (size <= 0) {
		    break;
		}
	    }
	    if (pixbuf) {
		g_object_ref(G_OBJECT(pixbuf));
	    }

	    if (!module.stop_load(context, error)) {
		if (pixbuf) {
		    g_object_unref(G_OBJECT(pixbuf));
		    pixbuf = 0;
		}
	    }
	}
    }

    close(fd);

    return pixbuf;
}

GdkPixbuf* gdk_pixbuf_new_from_file_ext (const char *filename, GError **error)
{
    GdkPixbuf *pixbuf = gdk_pixbuf_ilbm_from_file(filename, 0);
    if (!pixbuf) {
        pixbuf = gdk_pixbuf_new_from_file(filename, error);
    }
    return pixbuf;
}

extern "C" {
    GdkPixbufAnimation *gdk_pixbuf_non_anim_new (GdkPixbuf *pixbuf);
}

GdkPixbufAnimation* gdk_pixbuf_animation_new_from_file_ext (const char
  *filename, GError **error)
{
    GdkPixbuf *pixbuf = gdk_pixbuf_ilbm_from_file(filename, 0);
    if (pixbuf) {
	return gdk_pixbuf_non_anim_new(pixbuf);
    }
    return gdk_pixbuf_animation_new_from_file(filename, error);
}

