//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkImexport.cc
//
// Creation date: 05.01.2005
// Last modified: 06.11.2020
//

#include "system.h"
#include "imexport.h"
#include "gtk/gtkimpl.h"
#include "gtk/gtkUtil.h"
#include "gtk/classes/gtkpopselection.h"
#include "gtk/classes/gtkgroup.h"
#include "table.h"
#include "log.h"
#include "guidosutil.h"
#include "util.h"
#include "memory.h"
#include "appstrings_cat.h"

#define DELIMITER_LEN 9
#define FILENAME_LEN 255

static struct
{
    char ies_Filename[FILENAME_LEN + 1];
    char ies_RecordDelimiter[DELIMITER_LEN + 1];
    char ies_FieldDelimiter[DELIMITER_LEN + 1];
    bool ies_DoubleQuotes;
    bool ies_Filter;
    bool ies_OverwriteRecords;
} ImExportSettings =
{
    "", "\\n", "\\t", true, false, false,
};

enum ImExportAction
{
    IEA_Import,
#ifdef ENABLE_LOG_APPLY_CHANGES
    IEA_ApplyChanges,
#endif
    IEA_Export,
    IEA_ExportSFW
};

struct ImExportData
{
    struct Project *d_Project;
    struct Table *d_Table;
    GuiObject d_SFWList;
    enum ImExportAction d_Action;
    GuiObject d_Filename, d_View, d_Edit;
    GuiObject d_RecordDelimiter, d_FieldDelimiter, d_DoubleQuotes;
    GuiObject d_Filter;
    GuiObject d_OverwriteRecords;
    GuiObject d_ImExport, d_Cancel;
};

static void saveImExportData(ImExportData *d)
{
    const char *s = gtkEntryText(d->d_Filename);
    CodesetStrNBytesCopy(ImExportSettings.ies_Filename, s, FILENAME_LEN);
    s = gtkEntryText(d->d_RecordDelimiter);
    CodesetStrNBytesCopy(ImExportSettings.ies_RecordDelimiter, s,DELIMITER_LEN);
    s = gtkEntryText(d->d_FieldDelimiter);
    CodesetStrNBytesCopy(ImExportSettings.ies_FieldDelimiter, s, DELIMITER_LEN);
    ImExportSettings.ies_DoubleQuotes= gtkToggleButtonActive(d->d_DoubleQuotes);
    if (d->d_Filter) {
        ImExportSettings.ies_Filter = gtkToggleButtonActive(d->d_Filter);
    }
    if (d->d_OverwriteRecords) {
        ImExportSettings.ies_OverwriteRecords = 
            gtkToggleButtonActive(d->d_OverwriteRecords);
    }
}

static const char *checkDelimiterString(const char *delstring, 
	struct MemPool *memPool, char *del)
{
    char *s = StrCompress(memPool, delstring);
    if (!s) {
        return GCS(MSG_Err_Mem);
    }

    short len = strlen(s);
    bool ascii = StrIsASCII(s);
    *del = *s;
    FreeString(memPool, s);
    if (len != 1 || !ascii ||
            (*del == '"' && ImExportSettings.ies_DoubleQuotes)) {
        return GCS(MSG_ImExportErr_InvalidDelimiter);
    }
    
    return 0;
}

static bool checkImExportData(GuiObject go, ImExportData *d, 
	char *recdel, char *fielddel)
{
    struct MemPool *memPool = ProjectMemPool(d->d_Project);
    const char *error;
   
    if (ImExportSettings.ies_Filename[0] == 0) {
        error = GCS(MSG_Err_NoFilename);
	gtk_window_set_title(GTK_WINDOW(go), error);
	gtk_widget_grab_focus(d->d_Filename);
        GuiBeep();
        return false;
    }
    error = checkDelimiterString(
	    ImExportSettings.ies_RecordDelimiter, memPool, recdel);
    if (error) {
	gtk_window_set_title(GTK_WINDOW(go), error);
	gtk_widget_grab_focus(d->d_RecordDelimiter);
	GuiBeep();
        return false;
    }
    error = checkDelimiterString(ImExportSettings.ies_FieldDelimiter, 
	    memPool, fielddel);
    if (error) {
	gtk_window_set_title(GTK_WINDOW(go), error);
	gtk_widget_grab_focus(d->d_FieldDelimiter);
	GuiBeep();
        return false;
    }
    return true;
}

static void finish(GuiObject go, gint response, ImExportData *d)
{
    Project *pr = d->d_Project;

    saveImExportData(d);

    if (response == GTK_RESPONSE_ACCEPT) {
	char recdel = 0, fielddel = 0;
        bool rc = false;

        if (!checkImExportData(go, d, &recdel, &fielddel)) {
            return;
	}

        switch (d->d_Action) {
            case IEA_Import:
                rc = ImportRecords(d->d_Table,
                        ImExportSettings.ies_OverwriteRecords,
                        ImExportSettings.ies_Filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;
#ifdef ENABLE_LOG_APPLY_CHANGES
            case IEA_ApplyChanges:
                rc = ImportApplyChanges(pr,
                        ImExportSettings.ies_Filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;
#endif
            case IEA_Export:
                rc = ExportRecords(d->d_Table, ImExportSettings.ies_Filter,
                        ImExportSettings.ies_Filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;
            case IEA_ExportSFW:
                rc = ExportSFWList(pr, d->d_SFWList,
                        ImExportSettings.ies_Filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;
        }
        if (!rc) {
            return;
        }
    }
  
    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

static GuiObject createImExport(Project *pr, Table *tab, GuiObject sfw, 
        enum ImExportAction action)
{
    ImExportData *d = g_new0(ImExportData, 1);
    d->d_Project = pr;
    d->d_Table = tab;
    d->d_SFWList = sfw;
    d->d_Action = action;

    d->d_Filename = EntryObject,
	"max-length", FILENAME_LEN,
	"xalign", 0.5,
	"text", EmptyNullStr(ImExportSettings.ies_Filename),
	End;

    const char *filenameBH = NULL;
    const char *filenameLabelStr = NULL;
    const char *imExportStr = NULL;
    const char *dialogTitleStr = NULL;
    const char *dialogHelpNode = NULL;
    bool hasView = false;
    switch (action)
    {
        case IEA_Import:
            filenameLabelStr = GCS(MSG_Import_File);
            filenameBH = GCS(MSG_Import_FileBH);
            imExportStr = GCS(MSG_Import);
            dialogTitleStr = GCS(MSG_Import);
            dialogHelpNode = "Importing records";
            hasView = true;
            break;
#ifdef ENABLE_LOG_APPLY_CHANGES
        case IEA_ApplyChanges:
            filenameLabelStr = GCS(MSG_Import_LogFile);
            filenameBH = GCS(MSG_Import_LogFileBH);
            imExportStr = GCS(MSG_Import);
            dialogTitleStr = GCS(MSG_ApplyChanges);
            dialogHelpNode = "Applying changes";
            hasView = true;
            if (*EmptyNullStr(gtkEntryText(d->d_Filename)) == 0) {
                gtk_entry_set_text(GTK_ENTRY(d->d_Filename),
                        LogFilename(ProjectLog(pr)));
            }
            break;
#endif
        case IEA_Export:
            filenameLabelStr = GCS(MSG_Export_File);
            filenameBH = GCS(MSG_Export_Records_FileBH);
            imExportStr = GCS(MSG_Export);
            dialogTitleStr = GCS(MSG_Export);
            dialogHelpNode = "Exporting records";
            break;
        case IEA_ExportSFW:
            filenameLabelStr = GCS(MSG_Export_File);
            filenameBH = GCS(MSG_Export_SFWList_FileBH);
            imExportStr = GCS(MSG_Export);
            dialogTitleStr = GCS(MSG_Export);
            dialogHelpNode = "Exporting a list";
            break;
    }
    GuiSetTooltip(d->d_Filename, filenameBH);
    GuiObject filenameLabel = GuiLabel(filenameLabelStr, d->d_Filename);

    GuiObject filenamePop = PopSelectionObject,
        "selection-type", GTK_TYPE_FILE_CHOOSER_DIALOG,
        "string", d->d_Filename,
        "button", GuiPopButtonSetCWD(SELECT_FILE_ICON_NAME, pr),
	End;
    GuiObject filenameGroup = HGroup,
        "spacing", 0,
	"child", filenamePop,
	End;

    if (hasView) {
        d->d_View = ButtonObject,
	    "child", ImageObject, 
		"icon-name", "text-x-generic",
		"icon-size", GTK_ICON_SIZE_MENU,
		End,
	    End;
        g_signal_connect_swapped(d->d_View, "clicked", 
		G_CALLBACK(GuiObjectViewFile), d->d_Filename);
	GuiSetTooltip(d->d_View, GCS(MSG_ViewBH));
	gtk_group_add_child(GTK_GROUP(filenameGroup), d->d_View, 0, 0);

        d->d_Edit = ButtonObject,
	    "child", ImageObject, 
		"icon-name", "gtk-edit",
		"icon-size", GTK_ICON_SIZE_MENU,
		End,
	    End;
 	g_signal_connect_swapped(d->d_Edit, "clicked", 
		G_CALLBACK(GuiObjectEditFile), d->d_Filename);
	GuiSetTooltip(d->d_Edit, GCS(MSG_EditBH));
	gtk_group_add_child(GTK_GROUP(filenameGroup), d->d_Edit, 0, 0);
    }

    d->d_RecordDelimiter = EntryObject,
        "max-length", DELIMITER_LEN,
	"width-chars", DELIMITER_LEN,
	"xalign", 0.5,
	"text", EmptyNullStr(ImExportSettings.ies_RecordDelimiter),
	End;
    GuiSetTooltip(d->d_RecordDelimiter, GCS(MSG_ImExport_RecordDelimiterBH));
    GuiObject recordDelimiterLabel = GuiLabel(
            GCS(MSG_ImExport_RecordDelimiter), d->d_RecordDelimiter);

    d->d_FieldDelimiter = EntryObject,
        "max-length", DELIMITER_LEN,
	"width-chars", DELIMITER_LEN,
	"xalign", 0.5,
	"text", EmptyNullStr(ImExportSettings.ies_FieldDelimiter),
	End;
    GuiSetTooltip(d->d_FieldDelimiter, GCS(MSG_ImExport_FieldDelimiterBH));
    GuiObject fieldDelimiterLabel = GuiLabel(
            GCS(MSG_ImExport_FieldDelimiter), d->d_FieldDelimiter);

    d->d_DoubleQuotes = CheckButtonObject,
        "label", GCS(MSG_ImExport_DoubleQuotes),
	"use-underline", true,
	"active", (gboolean)ImExportSettings.ies_DoubleQuotes,
	End;
    GuiSetTooltip(d->d_DoubleQuotes, GCS(MSG_ImExport_DoubleQuotesBH));

    GuiObject extra = 0;
    bool hasExtra = false;
    switch (action)
    {
        case IEA_Import:
            d->d_OverwriteRecords = extra = CheckButtonObject,
                "label", GCS(MSG_Import_OverwriteRecords),
                "use-underline", true,
                "active", (gboolean)ImExportSettings.ies_OverwriteRecords,
                End;
            GuiSetTooltip(d->d_OverwriteRecords,
                    GCS(MSG_Import_OverwriteRecordsBH));
            hasExtra = true;
            break;
        case IEA_Export:
            d->d_Filter = extra = CheckButtonObject,
                "label", GCS(MSG_Filter),
                "use-underline", true,
                "active", (gboolean)ImExportSettings.ies_Filter,
                End;
            GuiSetTooltip(d->d_Filter, GCS(MSG_Export_FilterBH));
            hasExtra = true;
            break;
        default:
            break;
    }

    GuiObject entryGroup = VGroup,
        "columns", 2,
        "child", filenameLabel, "weight", 0,
        "child", filenameGroup,
        "child", recordDelimiterLabel, "weight", 0,
        "child", HGroup,
            "child", d->d_RecordDelimiter,
            "child", SpaceObject, End,
            End,
        "child", fieldDelimiterLabel, "weight", 0,
        "child", HGroup,
            "child", d->d_FieldDelimiter,
            "child", SpaceObject, End,
            End,
	End;

    GuiObject checkGroup = HGroup,
        "child", d->d_DoubleQuotes,
        End;
    if (hasExtra) {
	gtk_group_add_child(GTK_GROUP(checkGroup), extra, 0, 0);
    }
    GuiObject group = VGroup,
        "child", entryGroup, "vert-weight", 0,
	"child", checkGroup, "vert-weight", 0,
	End;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_ImExport = GuiSimpleButton(imExportStr, true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", dialogTitleStr,
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, dialogHelpNode);
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_ImExport, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_ImExport);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_ImExport);

    g_signal_connect(dialog, "response", G_CALLBACK(finish), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    GuiAddEntryCallbacks(d->d_Filename, d->d_ImExport);
    GuiAddEntryCallbacks(d->d_RecordDelimiter, d->d_ImExport);
    GuiAddEntryCallbacks(d->d_FieldDelimiter, d->d_ImExport);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }

    return dialog;
}

void GuiImportRecords(Table *tab)
{
    createImExport(TableProject(tab), tab, 0, IEA_Import);
}

#ifdef ENABLE_LOG_APPLY_CHANGES
void GuiApplyChanges(struct Project *pr)
{
    createImExport(pr, 0, 0, IEA_ApplyChanges);
}
#endif

void GuiExportRecords(Table *tab)
{
    createImExport(TableProject(tab), tab, 0, IEA_Export);
}

void GuiExportSFWList(struct Project *pr, GuiObject sfw)
{
    createImExport(pr, 0, sfw, IEA_ExportSFW);
}

