//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkMenus.cc
//
// Creation date: 29.03.2004
// Last modified: 02.05.2022
//

#include "gtk/gtkMenus.h"
#include "gtk/gtkBuildGuiWindow.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "gdk/gdkkeysyms.h"
#include "menuproject.h"
#include "menuprefs.h"
#include "menutable.h"
#include "menulog.h"
#include "menuprogram.h"
#include "menuhelp.h"
#include "menu.h"
#include "main.h"
#include "structureeditor.h"
#include "display.h"
#include "log.h"
#include "lang/program.h"
#include "status.h"
#include "preferences.h"
#include "appstrings_cat.h"

GtkAccelGroup *AccelGroup = 0;
GtkWidget *TableMenu = 0;
static unsigned long TableStatus = (unsigned long)-1;

#ifdef MACOSX
GtkWidget *MenuBar = 0;
static unsigned long MenuStatus = (unsigned long)-1;
#else
gboolean GtkShellShowsMenubar = FALSE;
#endif

// special variables for callbacks

static ulong memtable[] = 
{ 1000000, 2000000, 4000000, 8000000, 16000000, 32000000, 0x7fffffff };

static ulong menuItemChecked = false;

static void callchecked(Project *pr, intptr_t funcptr, GuiObject win)
{
    void (*func)(Project *, intptr_t, GuiObject) = 
	    (void (*)(Project *, intptr_t, GuiObject))funcptr;
    if (func) {
	func(pr, menuItemChecked, win);
    }
}

static void saveprojectinit(Project *pr, GuiObject item, intptr_t dummy)
{
    if (!ProjectAdminMode(pr) && ProjectChanges(pr) == 0) {
        gtk_widget_set_sensitive(item, false);
    }
}

#ifdef ENABLE_ADMIN_MODE
static void adminmodeinit(Project *pr, GuiObject item, intptr_t dummy)
{
    bool admin = ProjectAdminMode(pr);
    const char *title= GCS(admin? MSG_MenuItem_UserMode:MSG_MenuItem_AdminMode);
    gtk_menu_item_set_label(GTK_MENU_ITEM(item), title);
}
#endif

#ifdef ENABLE_STRUCTURE_EDITOR
static void structureeditorinit(Project *pr, GuiObject item, intptr_t dummy)
{
    GuiObject se = ProjectStructureEditor(pr);
    const char *title = GCS(se? MSG_MenuItem_CloseStructureEditor : 
		      MSG_MenuItem_StructureEditor);
    gtk_menu_item_set_label(GTK_MENU_ITEM(item), title);
}
#endif

static void reverttosavedinit(Project *pr, GuiObject item, intptr_t dummy)
{
    bool hasChanged = ProjectChanges(pr) != 0;
    const char *title = GCS(hasChanged? MSG_MenuItem_RevertToSaved : 
		      MSG_MenuItem_Reload);
    gtk_menu_item_set_label(GTK_MENU_ITEM(item), title);
}

static void confirmdeleterecordinit(Project *pr, GuiObject item, intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_CONFIRMDELETERECORD),
	    NULL);
}

#ifdef ENABLE_STRUCTURE_EDITOR
static void sorttriggerfunctionsinit(struct Project *pr, GuiObject item, 
	intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_PROGRAMSORTTRIGGERFUNCTIONS),
	    NULL);
}
#endif

static void recordmeminit(Project *pr, GuiObject item, intptr_t num)
{
    g_object_set(G_OBJECT(item),
	    "active", 
		(ulong)GetPrefsValue(pr, PREFS_RECORDMEMORY) == memtable[num],
	    NULL);
}

static void extrabuttonsintabchaininit(Project *pr, GuiObject item,
	      	intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_EXTRABUTTONSINTABCHAIN),
	    NULL);
}

static void advanceonenterinit(Project *pr, GuiObject item, intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_ADVANCEONENTER),
	    NULL);
}

static void confirmautoreloadinit(Project *pr, GuiObject item, intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_CONFIRMAUTORELOAD),
	    NULL);
}

static void confirmsavereorginit(Project *pr, GuiObject item, intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_CONFIRMSAVEREORG),
	    NULL);
}

static void vacuumafterreorginit(Project *pr, GuiObject item, intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_VACUUMAFTERREORG),
	    NULL);
}

static void confirmquitinit(Project *pr, GuiObject item, intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_CONFIRMQUIT),
	    NULL);
}

static void pathsrelativetoprojectinit(Project *pr, GuiObject item,
	       	intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_PATHSRELATIVETOPROJECT),
	    NULL);
}

static void logactiveinit(Project *pr, GuiObject item, intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", LogActive(ProjectLog(pr)),
	    NULL);
}

static void logmodeinit(Project *pr, GuiObject item, intptr_t num)
{
    g_object_set(G_OBJECT(item),
	    "active", MenuLogModeInit(pr, num),
	    NULL);
}

static void logsetlabelinit(Project *pr, GuiObject item, intptr_t dummy)
{
    if (FindUserFunctionName(ProjectProgram(pr), logLabel)) {
        gtk_widget_set_sensitive(item, false);
    }
}

#ifdef ENABLE_PROGRAM_EDITOR
static void programsourceinit(struct Project *pr, GuiObject item, intptr_t ext)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL)==(long)ext,
	    NULL);
}

static void cleanupexternalsourceinit(Project *pr, GuiObject item,
		intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL_CLEANUP),
	    NULL);
}

static void programdebuginfoinit(Project *pr, GuiObject item, intptr_t dummy)
{
    g_object_set(G_OBJECT(item),
	    "active", GetPrefsValue(pr, PREFS_PROGRAMDEBUGINFO),
	    NULL);
}

static void programobsoletefunctionsinit(Project *pr, GuiObject item,
	       	intptr_t num)
{
    g_object_set(G_OBJECT(item),
            "active", GetPrefsValue(pr, PREFS_PROGRAMOBSOLETEFUNCTIONS) == num,
            NULL);
}
#endif

static void projectopen(Project *pr)
{
    MenuProjectOpen(pr, 0);
}

static void recordmem(Project *pr, intptr_t num, GuiObject win)
{
    SetPrefsValue(pr, PREFS_RECORDMEMORY, (long)memtable[num]);
}

#include "menu.share"

#define NUM_MENU_ITEMS (sizeof(MBNewMenu) / sizeof(MBNewMenu[0]))
#define NUM_TABLE_MENU_ITEMS \
		(sizeof(MBTableNewMenu) / sizeof(MBTableNewMenu[0]))
#define TOTAL_MENU_ITEMS NUM_MENU_ITEMS + NUM_TABLE_MENU_ITEMS

static GtkWidget *popupMenuActiveWindow = 0;

static bool onMenu(GtkWidget *item)
{
    const struct MBNewMenu *mnm = (const struct MBNewMenu *)
        g_object_get_data(G_OBJECT(item), "mnm");

    if (!mnm->mnm_Callback) {
	return true;
    }
    
    GtkWidget *win = popupMenuActiveWindow;
    if (!win) {
        win = GuiActiveWindow();
    }
    ulong status = GetStatus(win);
    // printf("aw: %p, status: %d\n", win, status);
#if 0
    extern struct List ProjectList;
    printf("onMenu, win=%p, root=%p, status=%ld\n", win, 
	DisplayObjectGuiObject(ProjectRootDisplayObject(
		(Project *)HEAD(&ProjectList))),
	status);
#endif
    if ((mnm->mnm_Status & status) != mnm->mnm_Status) {
       	gdk_beep();
	return true;
    }
    
    GuiStoreEdits();
    if (GTK_IS_CHECK_MENU_ITEM(item)) {
	g_object_get(G_OBJECT(item), "active", &menuItemChecked, NULL);
    }
    mnm->mnm_Callback(GuiWindowProject(win), mnm->mnm_UserData, win);

    return true;
}

static void initializeMenus(struct MBNewMenu *mnm)
{
    for (; mnm->mnm_Type != MNM_END; mnm++) {
	if (mnm->mnm_LabelId) {
	    mnm->mnm_Label = GCS(mnm->mnm_LabelId);
       	}
	if (mnm->mnm_ShortcutId) {
	    mnm->mnm_Shortcut = GCS(mnm->mnm_ShortcutId);
	}
    }
}

#ifndef MACOSX
static void hideMenus(struct MBNewMenu *mnm, ulong status)
{
    bool hide = FALSE;
    for (; mnm->mnm_Type != MNM_END; mnm++) {
        if (mnm->mnm_Type == MNM_TITLE) {
            hide = ((mnm->mnm_Status & status) != 0);
        }
        if (hide) {
            mnm->mnm_Type = MNM_IGNORE;
        }
    }
}
#endif

static void updateMenu(Project *pr, GuiObject menu, unsigned long status)
{
    if (menu) {
	GList *children = gtk_container_get_children(GTK_CONTAINER(menu));
	for (GList *list = children; list; list = list->next) {
	    GtkWidget *item = GTK_WIDGET(list->data);
	    const struct MBNewMenu *mnm = (const struct MBNewMenu *)
		g_object_get_data(G_OBJECT(item), "mnm");
	    if (mnm) {
		bool enabled = (mnm->mnm_Status & status) == mnm->mnm_Status;
	    
                if (gtk_widget_get_sensitive(item) != enabled) {
		    gtk_widget_set_sensitive(item, enabled);
		}
		if (mnm->mnm_Init) {
		    g_signal_handlers_block_by_func(item, (void *)onMenu, 0);
		    mnm->mnm_Init(pr, item, mnm->mnm_UserData);
		    g_signal_handlers_unblock_by_func(item, (void *)onMenu, 0);
		}
	    }
	    if (GTK_IS_MENU_ITEM(item)) {
		GtkWidget *sub = gtk_menu_item_get_submenu(GTK_MENU_ITEM(item));
		updateMenu(pr, sub, status);
	    }
	}
        g_list_free(children);
    }
}

void UpdateMenu(GuiObject win, bool force)
{
    if (!win) {
        return;
    }

    Project *pr = GuiWindowProject(win);
    unsigned long newStatus = GetStatus(win);
    unsigned long oldStatus;
    GuiObject menuBar;

#ifdef MACOSX
    oldStatus = MenuStatus;
    MenuStatus = newStatus;
    menuBar = MenuBar;
#else
    menuBar = (GuiObject)g_object_get_data(G_OBJECT(win), "menu-bar");
    if (!menuBar) {
        win = ProjectRootWindow(pr);
        if (!win) {
            return;
        }
        menuBar = (GuiObject)g_object_get_data(G_OBJECT(win), "menu-bar");
        if (!menuBar) {
            return;
        }
    }
    oldStatus = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(win), "status"));
    g_object_set_data(G_OBJECT(win), "status", GINT_TO_POINTER(newStatus));
#endif

    if (oldStatus != newStatus || force) {
	updateMenu(pr, menuBar, newStatus);
#ifdef MACOSX
	gtkosx_application_sync_menubar(BeeBaseApp);
#endif
    }
}

static void onPopupMenuSelectionDone(GtkWidget *menu)
{
    // printf("onPopupMenuSelectionDone\n");
    popupMenuActiveWindow = 0;
    gtk_menu_detach(GTK_MENU(menu));
}

void PopupTableMenu(GtkWidget *go)
{
    DisplayObject *dio = GuiObjectDisplayObject(go);
    Table *tab = DisplayObjectTable(dio);

    if (tab && TableMenu) {
        GtkWidget *win = gtk_widget_get_toplevel(go);
        Project *pr = GuiWindowProject(win);
        unsigned long status = GetStatus(win);

        if (status != TableStatus) {
            TableStatus = status;
            updateMenu(pr, TableMenu, TableStatus);
        }
        popupMenuActiveWindow = win;
        gtk_menu_attach_to_widget(GTK_MENU(TableMenu), go, 0);
        gtk_menu_popup(GTK_MENU(TableMenu), 0, 0, 0, 0, true, 
                gtk_get_current_event_time());
    }
}

static GtkWidget *createMenus(GtkWidget *menuBar, const struct MBNewMenu *mnm)
{
    GtkWidget *menu = 0, *sub = 0, *item = 0;
    for (; mnm->mnm_Type != MNM_END; mnm++) {
	switch (mnm->mnm_Type) {
	    case MNM_TITLE:
		menu = MenuObject, End;
		g_object_set_data(G_OBJECT(menu), "mnm", (gpointer)mnm);
		if (menuBar) {
		    item = gtk_menu_item_new_with_label(mnm->mnm_Label);
		    gtk_widget_show(item);
		    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);
		    gtk_menu_shell_append(GTK_MENU_SHELL(menuBar), item);
		    GuiSetHelpNode(item, mnm->mnm_HelpNode);
		}
		sub = 0;
		break;

	    case MNM_ITEM:
	    case MNM_SUB:
	    {
		if (!menu) {
		    return 0;		// wrong specification
		}

		if (mnm->mnm_Type == MNM_ITEM) {
		    sub = 0;
		} else if (!sub) {
		    sub = MenuObject, End;
		    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), sub);
		    g_object_set_data(G_OBJECT(sub), "mnm", (gpointer)mnm);
		}

		const char *signal = 0;
		const char *label = mnm->mnm_Label;
		if (label == MNM_BARLABEL) {
		    label = 0;
		}
		bool checkable = ((mnm->mnm_Flags & MNM_CHECKIT) != 0);
		bool radio = (mnm->mnm_MutualExclude != 0);
		if (checkable && label) {
		    item = gtk_check_menu_item_new_with_label(label);
		    if (radio) {
			g_object_set(G_OBJECT(item), 
                                "draw-as-radio", true,
                                NULL);
		    }
		    signal = "toggled";
		} else if (label) {
		    item = gtk_menu_item_new_with_label(label);
		    if (mnm->mnm_Status == STATUS_NEVER) {
			PangoFontDescription *pfd = 
			    pango_font_description_from_string("italic");
			if (pfd) {
#if GTK_MAJOR_VERSION == 2
			    gtk_widget_modify_font(
                                    gtk_bin_get_child(GTK_BIN(item)), pfd);
#else
			    gtk_widget_override_font(item, pfd);
#endif
			    pango_font_description_free(pfd);
			}
		    }
		    signal = "activate";
		} else {
		    item = gtk_separator_menu_item_new();
		}
		gtk_menu_shell_append(GTK_MENU_SHELL(sub? sub : menu), item);
		gtk_widget_show(item);
		g_object_set_data(G_OBJECT(item), "mnm", (gpointer)mnm);
		GuiSetHelpNode(item, mnm->mnm_HelpNode);

		const char *sc = mnm->mnm_Shortcut;
		if (signal && sc) {
#ifdef MACOSX
		    GdkModifierType mod = GDK_META_MASK;
		    GdkModifierType mod1 = GDK_META_MASK;
		    GdkModifierType mod2 = GDK_META_MASK;
#else
		    GdkModifierType mod = GDK_CONTROL_MASK;
		    GdkModifierType mod1 = GDK_MOD1_MASK;
		    GdkModifierType mod2 = GDK_SHIFT_MASK;
#endif
		    guint c = (unsigned char)sc[0];
		    if (c && sc[1]) {
			CatalogString id = mnm->mnm_ShortcutId;
			if (id) {
			    if (!strcmp(id, MSG_Help)) {
				c = GDK_KEY_Help;
				mod = (GdkModifierType)0;
			    } else if (!strcmp(id, MSG_F1)) {
				c = GDK_KEY_F1;
				mod = (GdkModifierType)0;
			    } else if (!strcmp(id, MSG_AltUp)) {
				c = GDK_KEY_Up;
				mod = mod1;
			    } else if (!strcmp(id, MSG_AltDown)) {
				c = GDK_KEY_Down;
				mod = mod1;
			    } else if (!strcmp(id, MSG_ShiftAltUp)) {
				c = GDK_KEY_Up;
				mod = (GdkModifierType)
				    (GDK_SHIFT_MASK | mod1);
			    } else if (!strcmp(id, MSG_ShiftAltDown)) {
				c = GDK_KEY_Down;
				mod = (GdkModifierType)
				    (GDK_SHIFT_MASK | mod1);
			    } else if (!strcmp(id, MSG_ShiftCtrlUp)) {
				c = GDK_KEY_Up;
				mod = (GdkModifierType)
				    (mod2 | GDK_CONTROL_MASK);
			    } else if (!strcmp(id, MSG_ShiftCtrlDown)) {
				c = GDK_KEY_Down;
				mod = (GdkModifierType)
				    (mod2 | GDK_CONTROL_MASK);
			    }
			}
		    }
		    gtk_widget_add_accelerator(item, "activate", 
			    AccelGroup, c, mod, GTK_ACCEL_VISIBLE);
		}
#ifdef MACOSX
		if (mnm->mnm_Callback == (callback_t)MenuProjectQuit) {
		    gtk_widget_hide(GTK_WIDGET(item));
		} else if (mnm->mnm_Callback == (callback_t)MenuHelpAbout) {
		    gtkosx_application_insert_app_menu_item (BeeBaseApp,
			    item, 0);
		}
#endif
		if (signal) {
		    g_signal_connect(item, signal, G_CALLBACK(onMenu), 0);
		}

		break;
	    }
	}
    }
    
    return menu;
}
    
void GuiWindowAddMenus(GtkWidget *win, GtkWidget *group, bool main)
{
    g_return_if_fail(GTK_IS_WINDOW(win));

#ifndef MACOSX
    gtk_window_add_accel_group(GTK_WINDOW(win), AccelGroup);

    if (group && (main || GtkShellShowsMenubar)) {
	GtkWidget *menuBar = MenuBarObject, End;
	createMenus(menuBar, MBNewMenu);

	gtk_group_add_child(GTK_GROUP(group), menuBar, 0, 0);
	    // with tear-off: HandleBoxObject, "child", menuBar, End, 0, 0);
	g_object_set_data(G_OBJECT(win), "menu-bar", menuBar);
	g_object_set_data(G_OBJECT(win), "status", GINT_TO_POINTER(-1));
    }
#endif
}

void GuiMenuExit(void)
{
#ifdef MACOSX
    if (MenuBar) {
	gtk_widget_destroy(MenuBar);
	MenuBar = 0;
    }
#endif
    if (TableMenu) {
        gtk_widget_destroy(TableMenu);
        g_object_unref(G_OBJECT(TableMenu));
        TableMenu = 0;
    }
    if (AccelGroup) {
	g_object_unref(G_OBJECT(AccelGroup));
	AccelGroup = 0;
    }
}

void GuiMenuInit(void)
{
    initializeMenus(MBNewMenu);
    initializeMenus(MBTableNewMenu);
    AccelGroup = gtk_accel_group_new();
    TableMenu = createMenus(0, MBTableNewMenu);
    g_object_ref(G_OBJECT(TableMenu));
    g_signal_connect(TableMenu, "selection-done", 
            G_CALLBACK(onPopupMenuSelectionDone), 0);

#ifdef MACOSX
    MenuBar = MenuBarObject, End;
//    gtk_osxapplication_set_use_quartz_accelerators(BeeBaseApp, TRUE);
    gtkosx_application_set_menu_bar(BeeBaseApp, GTK_MENU_SHELL(MenuBar));
//    gtkosx_application_set_window_menu(BeeBaseApp, NULL);
    createMenus(MenuBar, MBNewMenu);
    gtkosx_application_sync_menubar(BeeBaseApp);
#else
#if GTK_MAJOR_VERSION > 2
    g_object_get(gtk_settings_get_default(), 
            "gtk-shell-shows-menubar", &GtkShellShowsMenubar, NULL);
#endif
    if (!GtkShellShowsMenubar) {
        hideMenus(MBNewMenu, STATUS_ACTIVETABLE);
    }
#endif
}

