//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkOrder.cc
//
// Creation date: 03.01.2005
// Last modified: 26.04.2014
//

#include "order.h"
#include "gtk/classes/gtkgroup.h"
#include "gtk/gtkStructureEditor.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "gui.h"
#include "project.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "appstrings_cat.h"
#include "error.h"

struct EditOrderData
{
    struct Project *d_Project;
    struct Table *d_Table;
    char d_WindowTitle[100];
    GuiObject d_Type;
    GuiObject d_TypeNotebook;
    GuiObject d_CmpFuncStr;
    GuiObject d_CmpFuncPopup;
    GuiObject d_Fields;
    GuiObject d_OrderList;
    GuiObject d_Ok;
    GuiObject d_Clear;
    GuiObject d_Cancel;
};

static void finish(GuiObject go, gint response, EditOrderData *d)
{
    Project *pr = d->d_Project;
    Table *tab = d->d_Table;

    if (response == GTK_RESPONSE_ACCEPT) {
	MemPool *memPool = ProjectMemPool(d->d_Project); 
	OrderField *oas = 0;
	long num = 0;
	OrderData od;
        long error = OK;

        error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }
     
	od.od_Type = gtk_combo_box_get_active(GTK_COMBO_BOX(d->d_Type))?
	    ORD_TYPE_CMPFUNC : ORD_TYPE_FIELDLIST;
	if (od.od_Type == ORD_TYPE_CMPFUNC) {
	    od.od_Contents.cmpfunc = (char *)gtkEntryText(d->d_CmpFuncStr);
	} else {
	    num = GuiListNumEntries(d->d_OrderList);
	    if (num > 0) {
		oas = (OrderField *)PoolAllocMemZero(memPool, 
			num * sizeof(OrderField));
		if (!oas) {
		    error = ERR_MEM;
		} else {
		    GtkTreeModel *model = gtk_tree_view_get_model(
			    GTK_TREE_VIEW(d->d_OrderList));
		    GtkTreeIter iter;
		    long i=0;
		    if (gtk_tree_model_get_iter_first(model, &iter)) {
			do {
			    Field *field = 0;
			    gboolean ascending = false;
			    gtk_tree_model_get(model, &iter, 
				    0, &field, 
				    1, &ascending,
				    -1);
		    	    oas[i].oa_Field = field;
			    oas[i].oa_Ascending = ascending;
			    i++;
			} while (i < num && 
				gtk_tree_model_iter_next(model, &iter));
		    }
		}
	    }
	    od.od_Contents.fieldlist.num = (short)num;
	    od.od_Contents.fieldlist.oas = oas;
	}

	if (error == OK) {
	    GuiSetSleep(TRUE);
	    error = SetNewOrder(tab, &od);
	    GuiSetSleep(FALSE);
	}
	PoolFreeMem(memPool, oas, num * sizeof(OrderField));

	if (error) {
	    Ask2(go, GCS(MSG_ErrorMessage), AT_ERROR, 
		    GCS(MSG_StarOk), GES(error));
	    return;
	}

	gtk_widget_hide(go);

	error = ReorderAllRecords(tab);
	if (error) {
	    Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Err_Order), GES(error));
	}
	GuiDisplayPanel(tab);
	ProjectChange(pr, FALSE);
    }
  
    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

static void insertOrderField(EditOrderData *d, 
	Field *field, bool ascending, long pos)
{
    // printf("insertOrderField(%p, %d)\n", field, ascending);
    if (FieldTable(field) != d->d_Table) {
	return;
    }

    // check if field is already in order list
    long found = GuiListFindEntry(d->d_OrderList, field);
    if (found >= 0) {
	GuiListRemoveEntry(d->d_OrderList, field);
	if (pos > found) {
	    pos--;
	}
    }

    GtkTreeView *tree = GTK_TREE_VIEW(d->d_OrderList);
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    GtkListStore *store = GTK_LIST_STORE(model);
    GtkTreeIter iter;

    gtk_list_store_insert(store, &iter, pos);
    gtk_list_store_set(store, &iter, 0, field, 1, ascending, -1);

    GuiListSetActiveEntry(d->d_OrderList, field);
}

static void fieldsOnRowActivated(GtkTreeView *tree,
	GtkTreePath *path, GtkTreeViewColumn *column, EditOrderData *d)
{
    Field *field = (Field *)GuiListActiveEntry(d->d_Fields);
    insertOrderField(d, field, true, G_MAXINT);
}

gboolean orderOnDragDrop(GtkWidget *widget, GdkDragContext *context, 
	gint x, gint y, guint time, EditOrderData *d)
{
    // printf("on-drag-drop(%p)\n", widget);
    Field *field = 0;
    gboolean ascending = true;
    gboolean success = false;
    GtkWidget *source = gtk_drag_get_source_widget(context);

    if (source == d->d_Fields) {
	field = (Field *)GuiListActiveEntry(source);
    } else if (source == d->d_OrderList) {
	GtkTreeSelection *selection = 
	    gtk_tree_view_get_selection(GTK_TREE_VIEW(source));
	GtkTreeModel *model;
	GtkTreeIter iter;
	if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
	    gtk_tree_model_get(model, &iter, 0, &field, 1, &ascending, -1);
	}
    }

    if (field) {
	GtkTreePath *path = 0;
	GtkTreeViewDropPosition drop;
	long pos = G_MAXINT;
	if (gtk_tree_view_get_dest_row_at_pos(GTK_TREE_VIEW(widget),
		    x, y, &path, &drop) && path) {
	    gint *indices = gtk_tree_path_get_indices(path);
    	    if (indices) {
		pos = indices[0];
		if (drop == GTK_TREE_VIEW_DROP_AFTER) {
		    pos++;
		}
	    }
	    gtk_tree_path_free(path);
	}
	insertOrderField(d, field, ascending, pos);
	success = true;
    }
    gtk_drag_finish(context, success, false, time);
    return true;
}

gboolean fieldsOnDragDrop(GtkWidget *widget, GdkDragContext *context, 
	gint x, gint y, guint time, EditOrderData *d)
{
    // we simply ask the source to delete the item
    gboolean success = false;
    if (gtk_drag_get_source_widget(context) == d->d_OrderList) {
	success = true;
    }
    gtk_drag_finish(context, success, success, time);
    return true;
}

static void orderOnRowActivated(GtkTreeView *tree,
	GtkTreePath *path, GtkTreeViewColumn *column, gpointer unused)
{
    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
    GtkTreeIter iter;
    if (gtk_tree_model_get_iter(model, &iter, path)) {
	gboolean ascending = false;
	gtk_tree_model_get(model, &iter, 1, &ascending, -1);
	gtk_list_store_set(GTK_LIST_STORE(model), &iter, 1, !ascending, -1);
    }
}

static void orderCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    gboolean ascending = false;
    gtk_tree_model_get(model, iter, 1, &ascending, -1);
    g_object_set(G_OBJECT(cell),
	    "icon-name", ascending? 
//	    	"view-sort-ascending" : "view-sort-descending",
	    	"go-up" : "go-down",
	    "stock-size", GTK_ICON_SIZE_SMALL_TOOLBAR,
	    NULL);
}

static void fieldCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    Field *field = 0;
    gtk_tree_model_get(model, iter, 0, &field, -1);
    g_object_set(G_OBJECT(cell), "text", EmptyNullStr(FieldName(field)), 
	    NULL);
}

static void onTypeChanged(GuiObject go, EditOrderData *d)
{
    GtkComboBox *combo = GTK_COMBO_BOX(go);
    GtkNotebook *note = GTK_NOTEBOOK(d->d_TypeNotebook);

    gtk_notebook_set_current_page(note, gtk_combo_box_get_active(combo));
}

static void onClearClicked(GuiObject go, EditOrderData *d)
{
    GtkTreeView *tree = GTK_TREE_VIEW(d->d_OrderList);
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    GtkListStore *store = GTK_LIST_STORE(model);
    gtk_list_store_clear(store);

    gtk_entry_set_text(GTK_ENTRY(d->d_CmpFuncStr), "");
}

static GuiObject createEditOrder(Table *tab, OrderData *od)
{
    Project *pr = TableProject(tab);

    EditOrderData *d = g_new0(EditOrderData, 1);
    d->d_Project = pr;
    d->d_Table = tab;

    short type = ORD_TYPE_FIELDLIST, num = 0, page = 0;
    OrderField *oas = 0;
    const char *cmpfunc = 0;
    if (od) {
  	type = od->od_Type;
    	if (type == ORD_TYPE_FIELDLIST) {
  	    num = od->od_Contents.fieldlist.num;
  	    oas = od->od_Contents.fieldlist.oas;
    	} else if (type == ORD_TYPE_CMPFUNC) {
  	    cmpfunc = od->od_Contents.cmpfunc;
  	    page = 1;
    	}
    }

    d->d_Fields = GuiPointerTreeView(fieldCellDataFunc);
    g_signal_connect(d->d_Fields, "row-activated", 
	    G_CALLBACK(fieldsOnRowActivated), d);
    GuiSetTooltip(d->d_Fields, GCS(MSG_Order_FieldsBH));

    GuiObject fieldsFrame = Frame(GCS(MSG_Fields)),
        "child", GroupAlignment(true),
	    "child", ScrolledWindowObject,
	        "hscrollbar-policy", GTK_POLICY_NEVER,
	        "shadow-type", GTK_SHADOW_IN,
		"child", d->d_Fields,
		End,
	    End,
	End;

    GType types[2] = { G_TYPE_POINTER, G_TYPE_BOOLEAN };
    GtkCellRenderer *cells[2] = {
	gtk_cell_renderer_pixbuf_new(),
	gtk_cell_renderer_text_new()
    };
    GtkTreeCellDataFunc funcs[2] = {
	orderCellDataFunc,
	fieldCellDataFunc
    };
    d->d_OrderList = GuiTreeView(2, types, cells, funcs);
    // gtk_widget_set(d->d_OrderList, "reorderable", true, 0); // conflicts dnd
    g_signal_connect(d->d_OrderList, "row-activated", 
	    G_CALLBACK(orderOnRowActivated), 0);
    GuiSetTooltip(d->d_OrderList, GCS(MSG_Order_OrderBH));

    static GtkTargetEntry targets[] = {
	{ (gchar *)"Field", GTK_TARGET_SAME_APP, 0 },
	{ (gchar *)"OrderField", GTK_TARGET_SAME_APP, 1 }
    };
    GtkTargetEntry *fieldTarget = &targets[0];
    GtkTargetEntry *orderTarget = &targets[1];

    // drag and drop between tree views
    gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(d->d_Fields),
	    GDK_BUTTON1_MASK, fieldTarget, 1, GDK_ACTION_COPY);
    gtk_tree_view_enable_model_drag_dest(GTK_TREE_VIEW(d->d_Fields),
	    orderTarget, 1, GDK_ACTION_MOVE);
    gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(d->d_OrderList),
	    GDK_BUTTON1_MASK, targets, 2, 
	    (GdkDragAction)(GDK_ACTION_COPY | GDK_ACTION_MOVE));
    gtk_tree_view_enable_model_drag_dest(GTK_TREE_VIEW(d->d_OrderList),
	    targets, 2, (GdkDragAction)(GDK_ACTION_COPY | GDK_ACTION_MOVE));
    g_signal_connect(d->d_Fields, "drag-drop", 
	    G_CALLBACK(fieldsOnDragDrop), d); 
    g_signal_connect(d->d_OrderList, "drag-drop", 
	    G_CALLBACK(orderOnDragDrop), d);

    GuiObject order = Frame(GCS(MSG_Order_Order)),
        "child", GroupAlignment(true),
	    "child", ScrolledWindowObject,
	        "hscrollbar-policy", GTK_POLICY_NEVER,
	        "shadow-type", GTK_SHADOW_IN,
		"child", d->d_OrderList,
		End,
	    End,
	End;

    GuiObject fieldOrderGroup = HGroup,
        "same-height", true,
	"child", fieldsFrame,
	"child", order,
	End;

    d->d_CmpFuncStr = EntryObject,
        "max-length", TABLE_TRIGGER_LEN,
	"xalign", 0.5,
        "text", EmptyNullStr(cmpfunc),
	End;
    GuiSetTooltip(d->d_CmpFuncStr, GCS(MSG_Order_ComparisonFunctionBH));

    d->d_CmpFuncPopup = UserFunctionsPopup(d->d_CmpFuncStr, pr);

    GuiObject cmpFuncGroup = Frame(0),
	"child", GroupAlignment(false),
	    "child", VGroup,
		"child", SpaceObject, End,
	       	"child", GuiLabel(GCS(MSG_Order_ComparisonFunction), 
                        d->d_CmpFuncStr, 0.5), "weight", 0,
		"child", d->d_CmpFuncPopup, "weight", 0,
	        "child", SpaceObject, End,
		End,
	    End,
	End;
    
    static char *typeContents[3];
    typeContents[0] = GCS(MSG_Order_Type_FieldList);
    typeContents[1] = GCS(MSG_Order_Type_ComparisonFunction);
    typeContents[2] = NULL;
    d->d_Type = GuiComboBox(typeContents);
    g_signal_connect(d->d_Type, "changed", G_CALLBACK(onTypeChanged), d);

    GuiObject group = HGroup,
        ChildLabelWidget(GCS(MSG_Type), d->d_Type),
        End;

    d->d_TypeNotebook = NotebookObject,
        "enable-popup", false,
	"show-border", false,
	"show-tabs", false,
	"child", fieldOrderGroup,
	"child", cmpFuncGroup,
	End;

    group = VGroup,
        "child", group, "weight", 0,
	"child", d->d_TypeNotebook,
	End;

    List *fields = TableFields(tab);
    if (fields) {
	GtkTreeView *tree = GTK_TREE_VIEW(d->d_Fields);
	GtkTreeModel *model = gtk_tree_view_get_model(tree);
	GtkListStore *store = GTK_LIST_STORE(model);
	Field *field;
	FORNODE (field, fields) {
	    short type = FieldType(field);
	    if (type != TYPE_BUTTON && type != TYPE_VIRTUAL && 
		    type != TYPE_REFERENCE) {
		GtkTreeIter iter;
		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter, 0, field, -1);
	    }
	}
    }

    if (oas) {
	GtkTreeView *tree = GTK_TREE_VIEW(d->d_OrderList);
	GtkTreeModel *model = gtk_tree_view_get_model(tree);
	GtkListStore *store = GTK_LIST_STORE(model);
	for (short i=0; i < num; i++) {
	    if (oas[i].oa_Field) {
		GtkTreeIter iter;
		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter, 
			0, oas[i].oa_Field, 
			1, (gboolean)oas[i].oa_Ascending,
			-1);
	    }
	}
    }

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Clear = GuiSimpleButton(GCS(MSG_Clear));
    g_signal_connect(d->d_Clear, "clicked", G_CALLBACK(onClearClicked), d);
    GuiSetTooltip(d->d_Clear, GCS(MSG_Order_ClearBH));
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiAddEntryCallbacks(d->d_CmpFuncStr, d->d_Ok);

    sprintf(d->d_WindowTitle, "%s: %s", GCS(MSG_ChangeOrder), TableName(tab));
    GuiObject dialog = DialogObject,
        "title", d->d_WindowTitle,
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Changing orders");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    // set the current page after show_all otherwise the notebook does not
    // change to the correct page.  See gtk bug 351112.
    gtk_combo_box_set_active(GTK_COMBO_BOX(d->d_Type), page);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_box_pack_end(GTK_BOX(gtk_dialog_get_action_area(GTK_DIALOG(dialog))), 
                d->d_Clear, TRUE, TRUE, 0);
    gtk_widget_show(d->d_Clear);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(finish), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    gint width, height;
    gtk_window_get_size(GTK_WINDOW(dialog), &width, &height);
    width += 50;
    height += 100;
    gtk_window_set_default_size(GTK_WINDOW(dialog), width, height);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }

    return dialog;
}

void GuiChangeOrder(Table *tab, OrderData *od)
{
    createEditOrder(tab, od);
}

