//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkPDFExport.cc
//
// Creation date: 12.11.2011
// Last modified: 06.07.2021
//

#include "system.h"

#ifdef USE_HPDF

#include "pdfexport.h"
#include "print.h"
#include "gtk/gtkimpl.h"
#include "gtk/gtkUtil.h"
#include "gtk/classes/gtkpopselection.h"
#include "gtk/classes/gtkgroup.h"
#include "guidosutil.h"
#include "dosutil.h"
#include "util.h"
#include "requesters.h"
#include "appstrings_cat.h"

#define FILENAME_LEN 255

static struct
{
    struct HPDFSettings pes_HPDFSettings;
    char pes_Filename[FILENAME_LEN + 1];
} PDFExportSettings;

struct PDFExportData
{
    struct Project *d_Project;
    GuiObject d_SFWList;
    GuiObject d_Filename;
    GuiObject d_Paper;
    GuiObject d_Orientation;
    GuiObject d_FontName;
    GuiObject d_FontSize;
    GuiObject d_Header;
    GuiObject d_Footer;
    GuiObject d_Status;
    GuiObject d_Create, d_Cancel;
};

static const char *fontNames[] = 
{
    "Courier",
    "Helvetica", 
    "Times-Roman",
    NULL
};

static void savePDFExportSettings(PDFExportData *d)
{
    Project *pr = d->d_Project;
    struct HPDFSettings *hpdfs = &PDFExportSettings.pes_HPDFSettings;
    const char *s;

    if (d->d_Filename) {
        s = gtkEntryText(d->d_Filename);
        CodesetStrNBytesCopy(PDFExportSettings.pes_Filename, s, FILENAME_LEN);
    }
    int active = gtkComboBoxActive(d->d_Paper);
    hpdfs->hpdfs_PageSize = HPDFPageSizes[active];
    active = gtkComboBoxActive(d->d_Orientation);
    hpdfs->hpdfs_PageDirection = HPDFPageDirections[active];
    active = gtkComboBoxActive(d->d_FontName);
    hpdfs->hpdfs_FontName = fontNames[active];
    hpdfs->hpdfs_FontSize = gtkSpinButtonValue(d->d_FontSize);
    hpdfs->hpdfs_Header = gtkEntryText(d->d_Header);
    hpdfs->hpdfs_Footer = gtkEntryText(d->d_Footer);

    SaveHPDFSettingsToPrefs(hpdfs, pr);
    LoadHPDFSettingsFromPrefs(hpdfs, pr);
}

static bool checkPDFExportSettings(GuiObject go, PDFExportData *d)
{
    if (d->d_Filename && PDFExportSettings.pes_Filename[0] == 0) {
	gtk_window_set_title(GTK_WINDOW(go), GCS(MSG_Err_NoFilename));
	gtk_widget_grab_focus(d->d_Filename);
        GuiBeep();
        return FALSE;
    }
    return TRUE;
}

static void finish(GuiObject go, gint response, PDFExportData *d)
{
    Project *pr = d->d_Project;

    savePDFExportSettings(d);

    if (response == GTK_RESPONSE_ACCEPT) {
        if (!checkPDFExportSettings(go, d)) {
            return;
        }

        char *mem = 0;
        const char *filename = PDFExportSettings.pes_Filename;
        bool cont = true;

        if (!d->d_Filename) {
            static const char filename_template[] = TMPDIR "TMP.mb.XXXXXX.pdf";
            filename = mem = gtkExpandPathname(filename_template);

            ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
            MakeTemp(mem);
            if (filename[0] == 0) {
                Ask2(go, NULL, AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_MkTemp));
                cont = false;
            }
        }
        if (cont) {
            cont = ExportSFWListAsPDF(pr, d->d_SFWList,
                    &PDFExportSettings.pes_HPDFSettings, filename);
            if (cont) {
                GuiViewFile(filename, TRUE);
            }
        }
        g_free(mem);
        if (!cont) {
            return;
        }
    }
  
    GuiSetProjectSleep(pr, false);
    gtk_widget_destroy(go);
}

static void updateStatus(GuiObject go, PDFExportData *d)
{
    long pages;
    double truncated;
    const char *error;
    char buffer[128];
    const char *text = buffer;
    static const GdkColor red = { 0, 0xFFFF, 0, 0 };
    const GdkColor *col = 0;
   
    savePDFExportSettings(d);
    error = GetExportSFWListAsPDFStats(d->d_Project, d->d_SFWList,
            &PDFExportSettings.pes_HPDFSettings, &pages, &truncated);
    if (error) {
        text = error;
        col = &red;
    } else if (truncated > 0.0) {
        sprintf(buffer, GCS(MSG_PDFExport_Status_TooWide), pages, 
                (long)ceil(truncated));
        col = &red;
    } else {
        sprintf(buffer, GCS(MSG_PDFExport_Status_FitsPageWidth), pages);
    }
    gtk_label_set_text(GTK_LABEL(d->d_Status), text);
    gtk_widget_modify_fg(d->d_Status, GTK_STATE_NORMAL, col);
}

static void exportPrintSFWListAsPDF(struct Project *pr, GuiObject sfw, 
        bool isExport)
{
    struct HPDFSettings *hpdfs = &PDFExportSettings.pes_HPDFSettings;
    LoadHPDFSettingsFromPrefs(hpdfs, pr);

    PDFExportData *d = g_new0(PDFExportData, 1);
    d->d_Project = pr;
    d->d_SFWList = sfw;

    GuiObject group = VGroup,
        "columns", 2,
        End;

    if (isExport) {
        d->d_Filename = EntryObject,
            "max-length", FILENAME_LEN,
            "xalign", 0.5,
            "text", PDFExportSettings.pes_Filename,
            End;
        GuiSetTooltip(d->d_Filename, MSG_PDFExport_FileBH);
        GuiObject filenameLabel = 
            GuiLabel(GCS(MSG_PDFExport_File), d->d_Filename);

        GuiObject filenamePop = PopSelectionObject,
              "selection-type", GTK_TYPE_FILE_CHOOSER_DIALOG,
            "string", d->d_Filename,
            "button", GuiPopButtonSetCWD(SELECT_FILE_ICON_NAME, pr),
            End;
        GuiObject filenameGroup = HGroup,
            "spacing", 0,
            "child", filenamePop,
            End;

        g_object_set(G_OBJECT(group),
            "child", filenameLabel, "weight", 0,
            "child", filenameGroup, "vert-weight", 0,
            NULL);
    }

    d->d_Paper = GuiComboBox(HPDFPaperNames);
    for (int i=0; HPDFPaperNames[i] != NULL; i++) {
        if (HPDFPageSizes[i] == hpdfs->hpdfs_PageSize) {
            gtk_combo_box_set_active(GTK_COMBO_BOX(d->d_Paper), i);
            break;
        }
    }
    GuiObject paperGroup = GuiMakeTooltip(d->d_Paper, MSG_PDFExport_PaperBH);
    GuiObject paperLabel = GuiLabel(GCS(MSG_PDFExport_Paper), d->d_Paper);

    static const char *orientationNames[3];
    orientationNames[0] = GCS(MSG_PDFExport_Orientation_Portrait);
    orientationNames[1] = GCS(MSG_PDFExport_Orientation_Landscape);
    orientationNames[2] = NULL;
    d->d_Orientation = GuiComboBox(orientationNames);
    for (int i=0; orientationNames[i] != NULL; i++) {
        if (HPDFPageDirections[i] == hpdfs->hpdfs_PageDirection) {
            gtk_combo_box_set_active(GTK_COMBO_BOX(d->d_Orientation), i);
            break;
        }
    }
    GuiObject orientationGroup = GuiMakeTooltip(d->d_Orientation, 
            MSG_PDFExport_OrientationBH);
    GuiObject orientationLabel = GuiLabel(GCS(MSG_PDFExport_Orientation),
            d->d_Orientation);

    d->d_FontName = GuiComboBox(fontNames);
    for (int i=0; fontNames[i] != NULL; i++) {
        if (!strcmp(fontNames[i], hpdfs->hpdfs_FontName)) {
            gtk_combo_box_set_active(GTK_COMBO_BOX(d->d_FontName), i);
            break;
        }
    }
    GuiObject fontNameGroup = d->d_FontName;
    GuiObject fontNameLabel = GuiLabel(GCS(MSG_Font), d->d_FontName);

    d->d_FontSize = SpinButtonObject,
        "adjustment", gtk_adjustment_new(hpdfs->hpdfs_FontSize, 
                HPDF_FONT_SIZE_MIN, HPDF_FONT_SIZE_MAX, 0.5, 1.0, 0),
        "digits", 1,
	"numeric", true,
	"snap-to-ticks", false,
	"update-policy", GTK_UPDATE_IF_VALID,
	"wrap", false,
        "max-length", 5,
        "width-chars", 5,
        "xalign", 0.5,
        End;
    GuiObject fontSizeGroup = d->d_FontSize;
    GuiObject fontSizeLabel = GuiLabel(GCS(MSG_Size), d->d_FontSize);

    d->d_Header = EntryObject,
        "xalign", 0.5,
        "text", EmptyNullStr(hpdfs->hpdfs_Header),
        End;
    GuiSetTooltip(d->d_Header, MSG_PDFExport_HeaderBH);
    GuiObject headerLabel = GuiLabel(GCS(MSG_PDFExport_Header), d->d_Header);

    d->d_Footer = EntryObject,
        "xalign", 0.5,
        "text", EmptyNullStr(hpdfs->hpdfs_Footer),
        End;
    GuiSetTooltip(d->d_Footer, MSG_PDFExport_FooterBH);
    GuiObject footerLabel = GuiLabel(GCS(MSG_PDFExport_Footer), d->d_Footer);

    d->d_Status = LabelObject,
       "label", " ",
       "ypad", 1,
       "xpad", 2,
       "xalign", 0.5,
       End;
    GuiObject statusGroup = FrameObject,
        "shadow-type", GTK_SHADOW_IN,
        "child", d->d_Status,
        End;
    GuiObject statusLabel = GuiLabel(GCS(MSG_PDFExport_Status));

    g_signal_connect(d->d_Paper, "changed", G_CALLBACK(updateStatus), d);
    g_signal_connect(d->d_Orientation, "changed", G_CALLBACK(updateStatus), d);
    g_signal_connect(d->d_FontName, "changed", G_CALLBACK(updateStatus), d);
    g_signal_connect(d->d_FontSize, "value-changed",G_CALLBACK(updateStatus),d);
    g_signal_connect(d->d_Header, "changed", G_CALLBACK(updateStatus), d);
    g_signal_connect(d->d_Footer, "changed", G_CALLBACK(updateStatus), d);
    updateStatus(d->d_Status, d);

    g_object_set(G_OBJECT(group),
        "child", paperLabel, "weight", 0,
        "child", paperGroup, "vert-weight", 0,
        "child", orientationLabel, "weight", 0,
        "child", orientationGroup, "vert-weight", 0,
        "child", fontNameLabel, "weight", 0,
        "child", HGroup,
            "child", fontNameGroup,"vert-weight", 0,
            "child", fontSizeLabel, "weight", 0,
            "child", fontSizeGroup, "weight", 0,
            End, "vert-weight", 0,
        "child", headerLabel, "weight", 0,
        "child", d->d_Header, "vert-weight", 0,
        "child", footerLabel, "weight", 0,
        "child", d->d_Footer, "vert-weight", 0,
        "child", statusLabel, "weight", 0,
        "child", statusGroup, "vert-weight", 0,
	NULL);

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Create = GuiSimpleButton(GCS(MSG_PDFExport_CreatePDF), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_PDFExport),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", false,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Exporting a list as PDF");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Create, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Create);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Create);

    g_signal_connect(dialog, "response", G_CALLBACK(finish), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    GuiAddEntryCallbacks(d->d_Filename, d->d_Create);
    GuiAddEntryCallbacks(d->d_FontSize, d->d_Create);

    GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }
}

void GuiExportSFWListAsPDF(struct Project *pr, GuiObject sfw)
{
    exportPrintSFWListAsPDF(pr, sfw, true);
}

#if !defined(USE_ANSI_PRINT) && !defined(USE_GTK_PRINT)

void GuiPrintSFWList(struct Project *pr, GuiObject sfw)
{
    exportPrintSFWListAsPDF(pr, sfw, false);
}

#endif

#endif

