//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkPrint.cc
//
// Creation date: 04.01.2012
// Last modified: 29.01.2022
//

#include "system.h"

#ifdef USE_GTK_PRINT

#include "print.h"
#include "pdfexport.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "gtk/classes/gtksfwview.h"
#include "gtk/classes/gtkpopselection.h"
#include "gtk/classes/gtkgroup.h"
#include "requesters.h"
#include "preferences.h"
#include "util.h"
#include "appstrings_cat.h"
#define GETTEXT_PACKAGE "gtk20"
#include <glib/gi18n-lib.h>

#define ALWAYS_SHOW_TITLES TRUE

static struct
{
    const char *pes_FontName;
    bool pes_FitPageWidth;
} PrintExtraSettings;

struct PrintExtraData
{
    Project *d_Project;
    GtkSFWView *d_SFWView;
    GtkPrintOperation *d_Print;
    GuiObject d_Font;
    GuiObject d_FitPageWidth;
};

static void sfwViewGetFonts(PangoLayout *layout,
        PangoFontDescription **pfont, PangoFontDescription **pbold)
{
    const char *fontname = PrintExtraSettings.pes_FontName;
    PangoFontDescription *font, *bold;

    if (fontname && fontname[0]) {
        font = pango_font_description_from_string(fontname);
    } else {
        font = pango_font_description_copy(
                pango_layout_get_font_description(layout));
#ifdef WINDOWS
        // On Windows, Pango does not seem to render Arabic text if font = NULL.
        if (!font) {
            font = pango_font_description_from_string("Sans");
        }
#endif
    }
    bold = pango_font_description_copy(font);
    if (bold == 0) {
        bold = pango_font_description_new();
    }
    pango_font_description_set_weight(bold, PANGO_WEIGHT_BOLD);

    *pfont = font;
    *pbold = bold;
}

static int sfwViewGetTotalWidth(GtkSFWView *sfw)
{
    int numFields = sfw->data? sfw->data->qgd_NumFields : 0;
    int tw = 0;

    for (int i=0; i < numFields; i++) {
        tw += sfw->pangoWidths[i];
        if (i > 0) {
            tw += sfw->pangoGapWidth;
        }
    }
    return tw;
}

static double getScale(GtkPrintContext *context, int totalWidth)
{
    double scale = 1.0;

    if (PrintExtraSettings.pes_FitPageWidth) {
        double pw = gtk_print_context_get_width(context);

        if (totalWidth > pw * PANGO_SCALE) {
            scale = pw * PANGO_SCALE / totalWidth;
        }
    }
    return scale;
}

static const char *sfwViewGetPrintDimensions(GtkPrintContext *context, 
    GtkSFWView *sfw, int *pages)
{
    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
    GtkTreeIter iter;
    struct QueryLine *ql;
    PangoLayout *layout;
    PangoFontDescription *font, *bold; 
    int numFields = sfw->data? sfw->data->qgd_NumFields : 0;
    int i, maxw;
    double scale;
    double ph;
    int tw, th;
    double dlines;
    int lines;
    int bodyLines = 0;

    layout = gtk_print_context_create_pango_layout(context);
    sfwViewGetFonts(layout, &font, &bold);

    // text height
    pango_layout_set_font_description(layout, font);
    pango_layout_set_text(layout, "Ig", -1);
    pango_layout_get_size(layout, &tw, &th);
    th += pango_layout_get_spacing(layout);
    sfw->pangoHeight = th;
    sfw->pangoGapWidth = tw;

    // number of lines
    sfw->headerLines = 0;
    if ((ALWAYS_SHOW_TITLES || sfw->showTitles) && sfw->data) {
        sfw->headerLines =
            QueryLineLines(sfw->data->qgd_Titles, numFields);
    }
    if (model) {
        if (gtk_tree_model_get_iter_first(model, &iter)) {
            do {
                ql = NULL;
                gtk_tree_model_get(model, &iter, 0, &ql, -1);
                bodyLines += QueryLineLines(ql, numFields);
            } while (gtk_tree_model_iter_next(model, &iter));
        }
    }

    // compute pango widths
    for (i = 0; i < numFields; i++) {
        maxw = 0;
        if ((ALWAYS_SHOW_TITLES || sfw->showTitles) && sfw->data->qgd_Titles) {
            pango_layout_set_font_description(layout, bold);
            pango_layout_set_text(layout,
                    sfw->data->qgd_Titles->ql_Entries[i]->qe_String, -1);
            pango_layout_get_size(layout, &tw, &th);
            pango_layout_set_font_description(layout, font);
            maxw = tw;
        }

        if (model) {
            if (gtk_tree_model_get_iter_first(model, &iter)) {
                do {
                    ql = NULL;
                    gtk_tree_model_get(model, &iter, 0, &ql, -1);
                    if (ql) {
                        pango_layout_set_text(layout, 
                                ql->ql_Entries[i]->qe_String, -1);
                        pango_layout_get_size(layout, &tw, &th);
                        if (tw > maxw) {
                            maxw = tw;
                        }
                    }
                } while (gtk_tree_model_iter_next(model, &iter));
            }
        }
        // printf("w[%d]: %d\n", i, maxw);
        sfw->pangoWidths[i] = maxw;
    }

    scale = getScale(context, sfwViewGetTotalWidth(sfw));
    ph = gtk_print_context_get_height(context);
    dlines = ph * PANGO_SCALE / (sfw->pangoHeight * scale);

    if (sfw->headerLines > 0) {
        dlines -= sfw->headerLines + 0.5;
    }
    dlines -= 2.0;      // space for page number
    lines = (int)floor(dlines);

    sfw->linesPerPage = lines;
    *pages = lines < 1 ? INT_MAX : (bodyLines + lines - 1) / lines;
    if (lines < 1) {
        return GCS(MSG_PrintErr_PageTooSmall);
    }

    pango_font_description_free(bold);
    pango_font_description_free(font);
    g_object_unref(G_OBJECT(layout));

    return 0;
}

static void sfwViewOnBeginPrint(GtkPrintOperation *print, 
        GtkPrintContext *context, GtkSFWView *sfw)
{
    int pages;
    const char *error;

    error = sfwViewGetPrintDimensions(context, sfw, &pages);
    if (error) {
        Ask2(GTK_WIDGET(sfw), NULL, AT_ERROR, GCS(MSG_Cancel), error);
        gtk_print_operation_cancel(print);
    } else {
        gtk_print_operation_set_n_pages(print, pages);
    }
}

static void drawText(cairo_t *cairo, PangoLayout *layout, double x, double y,
        double pw, char format, double ph, const char *s,
        int offset, int numLines)
{
    int end = offset + numLines;
    int line;
    const char *t;
    double xx;
    int tw, th;

    for (line = 0; line < end && s != NULL; line++) {
        t = CodesetStrchr(s, '\n');
        if (line >= offset) {
            xx = x;
            pango_layout_set_text(layout, s, (t == NULL)? -1 : t - s);
            if (format) {
                pango_layout_get_size(layout, &tw, &th);
                xx += pw - tw;
            }
            cairo_move_to(cairo, xx/PANGO_SCALE, y/PANGO_SCALE);
            pango_cairo_show_layout(cairo, layout);
            y += ph;
        }
        s = (t == NULL)? NULL : CodesetNextChar(t);
    }
}

static void sfwViewOnDrawPage(GtkPrintOperation *print,
        GtkPrintContext *context, gint page, GtkSFWView *sfw)
{
    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(sfw));
    const signed char *order = gtk_sfw_view_get_column_order(sfw);
    GtkTreeIter iter;
    struct QueryLine *ql;
    cairo_t *cairo;
    PangoLayout *layout;
    PangoContext *pc;
    PangoFontDescription *font; 
    PangoFontDescription *bold;
    int numFields = sfw->data? sfw->data->qgd_NumFields : 0;
    int i, col;
    double x, y, pw;
    int offset = 0;
    int p, line;
    int linesLeftOnPage, lines, maxLines;
    char buffer[64];
    double scale, hspace, ph;
    int totalWidth, tw, th;
    bool separateEntries = false;

    cairo = gtk_print_context_get_cairo_context(context);
    layout = gtk_print_context_create_pango_layout(context);
    pc = pango_layout_get_context(layout);
    totalWidth = sfwViewGetTotalWidth(sfw);
    scale = getScale(context, totalWidth);
    pango_cairo_context_set_resolution(pc, 
            pango_cairo_context_get_resolution(pc) * scale);
    sfwViewGetFonts(layout, &font, &bold);
    hspace = sfw->pangoGapWidth * scale;
    ph = sfw->pangoHeight * scale;

    x = 0.0;
    y = 0.0;
    if (sfw->headerLines > 0) {
        pango_layout_set_font_description(layout, bold);

        for (i = 0; i < numFields; i++) {
            col = order? order[i] : i;
            pw = sfw->pangoWidths[col] * scale;
            drawText(cairo, layout, x, y, pw, 
                    sfw->data->qgd_FieldInfos[col].qfi_Format, ph,
                    sfw->data->qgd_Titles->ql_Entries[col]->qe_String,
                    0, sfw->headerLines);
            x += pw + hspace;
        }
        y += ph * sfw->headerLines;
        y += ph / 4;
        cairo_set_line_width(cairo, 1.0 * scale);
        cairo_move_to(cairo, 0, y/PANGO_SCALE);
        cairo_line_to(cairo, totalWidth * scale/PANGO_SCALE, y/PANGO_SCALE);
        cairo_stroke(cairo);
        y += ph / 4;
    }

    pango_layout_set_font_description(layout, font);

    if (model != NULL && gtk_tree_model_get_iter_first(model, &iter)) {
        do {
            gtk_tree_model_get(model, &iter, 0, &ql, -1);
            if (QueryLineLines(ql, numFields) > 1) {
                separateEntries = true;
                break;
            }
        } while (gtk_tree_model_iter_next(model, &iter));
    }

    if (model != NULL && gtk_tree_model_get_iter_first(model, &iter)) {
        for (p = 0; p <= page; p++) {
            line = 0;
            while (line < sfw->linesPerPage) {
                ql = NULL;
                gtk_tree_model_get(model, &iter, 0, &ql, -1);
                lines = QueryLineLines(ql, numFields) - offset;
                linesLeftOnPage = sfw->linesPerPage - line;
                maxLines = (lines < linesLeftOnPage)? lines : linesLeftOnPage;
                if (p == page && ql != NULL) {
                    x = 0.0;
                    for (i=0; i < numFields; i++) {
                        col = order? order[i] : i;
                        pw = sfw->pangoWidths[col] * scale;
                        drawText(cairo, layout, x, y, pw, 
                                sfw->data->qgd_FieldInfos[col].qfi_Format, ph,
                                ql->ql_Entries[col]->qe_String,
                                offset, maxLines);
                        x += pw + hspace;
                    }
                    y += ph * maxLines;
                }
                line += maxLines;

                if (maxLines < lines) {
                    offset += maxLines;
                } else {
                    offset = 0;
                    if (separateEntries && p == page && ql != NULL) {
                        cairo_set_line_width(cairo, 1.0 * scale);
                        cairo_move_to(cairo, 0, y/PANGO_SCALE);
                        cairo_line_to(cairo, totalWidth * scale/PANGO_SCALE,
                                y/PANGO_SCALE);
                        cairo_stroke(cairo);
                    }
                    if (!gtk_tree_model_iter_next(model, &iter)) {
                        break;
                    }
                }
            }
        }
    }

    // Page number
    sprintf(buffer, GCS(MSG_PageFooter), (long)page + 1,
            (long)gtk_print_operation_get_n_pages_to_print(print));
    pango_layout_set_text(layout, buffer, -1);
    pango_layout_get_size(layout, &tw, &th);
    x = (gtk_print_context_get_width(context) * PANGO_SCALE - tw) / 2;
    y = (gtk_print_context_get_height(context) * PANGO_SCALE - th);
    cairo_move_to(cairo, x/PANGO_SCALE, y/PANGO_SCALE);
    pango_cairo_show_layout(cairo, layout);

    pango_font_description_free(font);
    pango_font_description_free(bold);
    g_object_unref(G_OBJECT(layout));
}

static void readPrintExtraSettingsFromPrefs(Project *pr)
{
    PrintExtraSettings.pes_FontName =
        GetPrefsString(pr, PREFS_GTKPRINTSETTINGS_FONTNAME);
    PrintExtraSettings.pes_FitPageWidth =
        GetPrefsValue(pr, PREFS_GTKPRINTSETTINGS_FITPAGEWIDTH);
}

static void storePrintExtraSettings(PrintExtraData *d)
{
    Project *pr = d->d_Project;

    SetPrefsString(pr, PREFS_GTKPRINTSETTINGS_FONTNAME,
            gtkEntryText(d->d_Font));
    SetPrefsValue(pr, PREFS_GTKPRINTSETTINGS_FITPAGEWIDTH,
            gtkToggleButtonActive(d->d_FitPageWidth));

    readPrintExtraSettingsFromPrefs(pr);
}

static GtkWidget *sfwViewOnCreateCustomWidget(GtkPrintOperation *print, 
        GtkSFWView *sfw)
{
    PrintExtraData *d = g_new0(PrintExtraData, 1);
    d->d_Project = GuiObjectProject((GuiObject)sfw);
    d->d_SFWView = sfw;
    d->d_Print = print;

    d->d_Font = EntryObject,
        "xalign", 0.5,
        "text", EmptyNullStr(PrintExtraSettings.pes_FontName),
        End;
    GuiSetTooltip(d->d_Font, GCS(MSG_FontBH));

    d->d_FitPageWidth = CheckButtonObject, 
        "label", GCS(MSG_QueryPrint_FitPageWidth),
        "use-underline", true,
        "active", PrintExtraSettings.pes_FitPageWidth,
        End;
    GuiSetTooltip(d->d_FitPageWidth, GCS(MSG_QueryPrint_FitPageWidthBH));

    GuiObject top = VGroup,
        "columns", 2,
        "child", GuiLabel(GCS(MSG_Font), d->d_Font), "weight", 0,
        "child", PopSelectionObject,
	    "selection-type", GUI_TYPE_FONT_CHOOSER_DIALOG,
	    "string", d->d_Font,
	    "button", GuiPopButton(SELECT_FONT_ICON_NAME), 
	    End, "vert-weight", 0,
        End;
    top = VGroup,
        "spacing", 8,
        "child", top,
        "child", d->d_FitPageWidth,
        End;
    top = HGroup,
        "child", top,
        "child", SpaceObject, End,
        End;
    top = VGroup,
        "border-width", 12,
        "child", top, "vert-weight", 0,
        "child", SpaceObject, End,
        End;
    g_object_set_data_full(G_OBJECT(top), "edit-data", d, g_free);
    gtk_widget_show_all(top);

    return top;
}

static void sfwViewOnCustomWidgetApply(GtkPrintOperation *print, GtkWidget *w,
        GtkSFWView *sfw)
{
    PrintExtraData *d;
    
    d = (PrintExtraData *)g_object_get_data(G_OBJECT(w), "edit-data");
    if (d) {
        storePrintExtraSettings(d);
    }
}

static void sfwViewPrintExport(Project *pr, GuiObject sfw, bool isExport)
{
    static GtkPrintSettings *settings = NULL;
    static gchar *printer = NULL;
    static GtkPageSetup *setup = NULL;
    GtkWindow *parent = GTK_WINDOW(gtk_widget_get_toplevel(GTK_WIDGET(sfw)));
    GtkPrintOperation *print;
    GtkPrintOperationResult res;
    GError *error = NULL;

    readPrintExtraSettingsFromPrefs(pr);

    print = gtk_print_operation_new();
    if (!settings) {
        settings = gtk_print_settings_new();
    }
    if (isExport) {
        gtk_print_settings_set_printer(settings, _("Print to File"));
        gtk_print_settings_set(settings, 
                GTK_PRINT_SETTINGS_OUTPUT_FILE_FORMAT, "pdf");
    } else {
        gtk_print_settings_set_printer(settings, printer);
    }
    if (settings) {
        gtk_print_operation_set_print_settings(print, settings);
    }
    if (setup) {
        gtk_print_operation_set_default_page_setup(print, setup);
    }
    g_signal_connect(G_OBJECT(print), "create-custom-widget",
            G_CALLBACK(sfwViewOnCreateCustomWidget), sfw);
    g_signal_connect(G_OBJECT(print), "custom-widget-apply",
            G_CALLBACK(sfwViewOnCustomWidgetApply), sfw);
    g_signal_connect(G_OBJECT(print), "begin-print",
            G_CALLBACK(sfwViewOnBeginPrint), sfw);
    g_signal_connect(G_OBJECT(print), "draw-page",
            G_CALLBACK(sfwViewOnDrawPage), sfw);

    // gtk_print_operation_set_job_name(print, "BeeBase Select-From-Where");
    gtk_print_operation_set_custom_tab_label(print, GCS(MSG_Font));
    gtk_print_operation_set_show_progress(print, TRUE);
    gtk_print_operation_set_embed_page_setup(print, TRUE);

    GuiSetSleep(TRUE);
    res = gtk_print_operation_run(print,
            GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG, 
            parent, &error);
 
    if (res == GTK_PRINT_OPERATION_RESULT_ERROR) {
        Ask2(sfw, NULL, AT_ERROR, GCS(MSG_StarOk),
                GCS(MSG_Err_Print), error->message);
        g_error_free(error);
    } else /* if (res == GTK_PRINT_OPERATION_RESULT_APPLY) */ { 
        if (settings) {
            g_object_unref(settings);
        }
        settings = gtk_print_operation_get_print_settings(print);
        if (settings) {
            g_object_ref(settings);
            if (!isExport) {
                g_free(printer);
                printer = g_strdup(gtk_print_settings_get_printer(settings));
            }
        }
        if (setup) {
            g_object_unref(setup);
        }
        setup = gtk_print_operation_get_default_page_setup(print);
        if (setup) {
            g_object_ref(setup);
        }
    }
    GuiSetSleep(FALSE);
}

void GuiPrintSFWList(Project *pr, GuiObject sfw)
{
    sfwViewPrintExport(pr, sfw, false);
}

#ifndef USE_HPDF
void GuiExportSFWListAsPDF(Project *pr, GuiObject sfw)
{
    sfwViewPrintExport(pr, sfw, true);
}
#endif

#endif

