//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkProgOutput.cc
//
// Creation date: 13.12.2004
// Last modified: 06.11.2014
//

#include "programoutput.h"
#include "gtk/classes/gtkgroup.h"
#include "gtk/classes/gtkpopselection.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "project.h"
#include "preferences.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

struct ProgOutputData
{
    struct Project *d_Project;
    GuiObject d_Filename;
    GuiObject d_Default;
    GuiObject d_Append;
    GuiObject d_Ok, d_Cancel;
};

static void finish(GuiObject go, gint response, ProgOutputData *d)
{
    if (response == GTK_RESPONSE_ACCEPT) {
	const gchar *filename = gtk_entry_get_text(GTK_ENTRY(d->d_Filename));
	gboolean def = gtkToggleButtonActive(d->d_Default);
	gboolean append = gtkToggleButtonActive(d->d_Append);
	Project *pr = d->d_Project;
        long error = ObtainProjectChangeLock(pr);

        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        if (!def) {
            SetPrefsString(pr, PREFS_PROGRAMOUTPUT_FILENAME, filename);
        }
	SetPrefsValue(pr, PREFS_PROGRAMOUTPUT_DEFAULT, def);
	SetPrefsValue(pr, PREFS_PROGRAMOUTPUT_APPEND, append);
    }
    GuiSetSleep(FALSE);
    gtk_widget_destroy(go);
}

static void defaultToggled(GuiObject go, ProgOutputData *d)
{
    const char *defaultName = DEFAULT_OUTPUT_FILE;
    bool active = gtkToggleButtonActive(d->d_Default);
    if (active) {
        gtk_entry_set_text(GTK_ENTRY(d->d_Filename), EmptyNullStr(defaultName));
    }
    GuiSetEnabled(d->d_Filename, !active);
}

static GuiObject createProgOutput(Project *pr)
{
    ProgOutputData *d = g_new0(ProgOutputData, 1);
    d->d_Project = pr;

    gboolean def = GetPrefsValue(pr, PREFS_PROGRAMOUTPUT_DEFAULT);
    const char *defaultName = DEFAULT_OUTPUT_FILE;
    const char *filename = def? 
        defaultName : GetPrefsString(pr, PREFS_PROGRAMOUTPUT_FILENAME);
    d->d_Filename = EntryObject,
        "sensitive", !def,
        "xalign", 0.5,
	"text", EmptyNullStr(filename),
	End;
    GuiSetTooltip(d->d_Filename, GCS(MSG_ProgOutput_FilenameBH));
    GuiObject popFile = PopSelectionObject,
        "selection-type", GTK_TYPE_FILE_CHOOSER_DIALOG,
	"file-chooser-action",  GTK_FILE_CHOOSER_ACTION_SAVE,
	"string", d->d_Filename,
        "button", GuiPopButtonSetCWD(SELECT_FILE_ICON_NAME, pr),
	End;
    GuiObject fileLabel = LabelObject, 
        "label", GCS(MSG_ProgOutput_Filename),
	"use-underline", true,
	"mnemonic-widget", d->d_Filename,
	End;

    d->d_Default = CheckButtonObject,
        "label", GCS(MSG_ProgOutput_Default),
        "use-underline", true,
        "active", def,
	End;
    GuiSetTooltip(d->d_Default, GCS(MSG_ProgOutput_DefaultBH));

    d->d_Append = CheckButtonObject,
        "label", GCS(MSG_ProgOutput_Append),
        "use-underline", true,
        "active", (gboolean)GetPrefsValue(pr, PREFS_PROGRAMOUTPUT_APPEND),
	End;
    GuiSetTooltip(d->d_Append, GCS(MSG_ProgOutput_AppendBH));

    GuiObject group = VGroup,
	"child", HGroup,
            "child", fileLabel, "weight", 0,
            "child", popFile, "vert-weight", 0,
            End,
        "child", HGroup,
            "child", d->d_Default, "vert-weight", 0,
            "child", d->d_Append, "vert-weight", 0,
            End,
	End;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;
    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_ProgOutput),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", true,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Program output file");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    g_signal_connect(d->d_Default, "toggled", G_CALLBACK(defaultToggled), d);
    g_signal_connect(dialog, "response", G_CALLBACK(finish), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);
    GuiAddEntryCallbacks(d->d_Filename, d->d_Ok);

    GuiSetSleep(true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetSleep(false);
        gtk_widget_destroy(dialog);
        dialog = 0;
    }

    return dialog;
}

void GuiProgOutput(Project *pr)
{
    createProgOutput(pr);
}

