//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkProgramEditor.cc
//
// Creation date: 17.04.2005
// Last modified: 12.10.2014
//

#include "programeditor.h"

#ifdef ENABLE_PROGRAM_EDITOR

#include "menueditor.h"
#include "menuprogram.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "gtk/classes/gtkgroup.h"
#include "gtk/classes/gtkexteditor.h"
#include "project.h"
#include "structureeditor.h"
#include "preferences.h"
#include "lang/compile.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

struct ProgramEditorData
{
    Project *d_Project;
    GuiObject d_Editor;
    bool d_DestroyOnClose;
};

void CloseProgramEditor(GuiObject go)
{
    if (go) {
        ProgramEditorData *d = (ProgramEditorData *)
	    g_object_get_data(G_OBJECT(go), "edit-data");
        Project *pr = d->d_Project;
	GuiObject editor = d->d_Editor;

        SetPrefsValue(pr, PREFS_PROGRAMEDITOR_TOP,
		gtk_ext_editor_get_top(GTK_EXT_EDITOR(editor)));

        guint line, col;
	gtk_ext_editor_get_cursor(GTK_EXT_EDITOR(editor), &line, &col);
        SetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSX, col);
        SetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSY, line);

#if 0
	printf("getting top=%ld, cursor= (%ld, %ld)\n",
		GetPrefsValue(pr, PREFS_PROGRAMEDITOR_TOP),
		GetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSX),
		GetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSY));
#endif
        WindowDimensions wd;
        if (GuiGetWindowDimensions(go, &wd)) {
            SetPrefsWindowDimensions(pr, PREFS_PROGRAMEDITOR, &wd);
        }

        if (d->d_DestroyOnClose) {
            gtk_widget_destroy(go);
        } else {
            gtk_widget_hide(go);        // when in gtk_dialog_run
        }
        SetProjectProgramEditor(pr, 0);
    }
}

bool ProgramEditorCheckClose(GuiObject go)
{
    if (go) {
        ProgramEditorData *d = (ProgramEditorData *)
	    g_object_get_data(G_OBJECT(go), "edit-data");
	GuiObject editor = d->d_Editor;
	GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(editor));
    
	if (gtk_text_buffer_get_modified(buffer)) {
	    if (!Ask2(go, GCS(MSG_ProgramEditor), AT_WARN,
			GCS(MSG_CloseAnswers), GCS(MSG_ProgramEditor_Close))) {
		return false;
	    }
	}
    }
    return true;
}


static bool compile(GuiObject go, bool reload)
{
    ProgramEditorData *d = (ProgramEditorData *)
	g_object_get_data(G_OBJECT(go), "edit-data");

    return MenuProgramCompile(d->d_Project, reload);
}

static void compileCloseClick(GuiObject go)
{
    ProgramEditorData *d = (ProgramEditorData *)
	g_object_get_data(G_OBJECT(go), "edit-data");
    Project *pr = d->d_Project;

    if (!compile(go, false)) {
        return;
    }

    gtk_dialog_response(GTK_DIALOG(go), GTK_RESPONSE_OK);
    CloseProgramEditor(go);
    GuiActivateWindow(DisplayObjectGuiObject(ProjectRootDisplayObject(pr)));
    GuiActivateWindow(ProjectStructureEditor(pr));
}

static void compileClick(GuiObject go)
{
    if (compile(go, true)) {
	gtk_dialog_response(GTK_DIALOG(go), GTK_RESPONSE_OK);
    }
}

static gboolean programEditorOnDelete(GuiObject go, GdkEvent *event,
	ProgramEditorData *d)
{
    if (ProgramEditorCheckClose(go)) {
	CloseProgramEditor(go);
    }
    return true;
}

static void undoClick(GuiObject go)
{
    ProgramEditorData *d = (ProgramEditorData *)
	g_object_get_data(G_OBJECT(go), "edit-data");
    Project *pr = d->d_Project;
    Program *prg = ProjectProgram(pr);
    SourceText *st = GetSourceText(prg);
    const char *text = EmptyNullStr(SourceTextText(st));
    GuiObject editor = d->d_Editor;
    GuiSetEditorText(editor, text);
    GuiSetEditorChanged(editor, false);

#if 0
    printf("setting top=%ld, cursor= (%ld, %ld)\n",
	    GetPrefsValue(pr, PREFS_PROGRAMEDITOR_TOP),
	    GetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSX),
	    GetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSY));
#endif

    gtk_ext_editor_set_top(GTK_EXT_EDITOR(editor), 
	    GetPrefsValue(pr, PREFS_PROGRAMEDITOR_TOP));
    gtk_ext_editor_set_cursor(GTK_EXT_EDITOR(editor),
	    GetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSY),
	    GetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSX));

    FreeSourceText(st);
}

static GuiObject createProgramEditorGui(Project *pr)
{
    ProgramEditorData *d = g_new0(ProgramEditorData, 1);
    d->d_Project = pr;
    d->d_DestroyOnClose = true;

    GtkWidget *group;
    GtkWidget *root = GroupAlignment(false),
        "child", group = VGroup,
	    "child", GtkNewEditor(pr, true, false, true, &d->d_Editor),
	    End,
	End;

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_EditProgram),
	"window-position", GTK_WIN_POS_CENTER,
	"default-width", MIN(gdk_screen_width() / 2, 800),
	"default-height", MIN(gdk_screen_height() / 2, 600),
        "type-hint", GDK_WINDOW_TYPE_HINT_NORMAL,
	"modal", false,
	End;
    struct WindowDimensions wd;
    if (GetPrefsWindowDimensions(pr, PREFS_PROGRAMEDITOR, &wd)) {
        GuiSetWindowDimensions(dialog, &wd);
    }
    
    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    GtkWidget *hbox = gtk_dialog_get_action_area(GTK_DIALOG(dialog));
    GuiObject undo = GuiSimpleButton(GCS(MSG_ProgramEditor_Revert));
    g_object_set(G_OBJECT(undo), "focus-on-click", false, NULL);
    GuiSetTooltip(undo, GCS(MSG_ProgramEditor_RevertBH));
    g_signal_connect_swapped(undo, "clicked", G_CALLBACK(undoClick), dialog);
    gtk_box_pack_end(GTK_BOX(hbox), undo, TRUE, TRUE, 0);
    gtk_widget_show(undo);

    GuiObject comp = GuiSimpleButton(GCS(MSG_ProgramEditor_Compile));
    g_object_set(G_OBJECT(comp), "focus-on-click", false, NULL);
    GuiSetTooltip(comp, GCS(MSG_ProgramEditor_CompileBH));
    g_signal_connect_swapped(comp, "clicked", G_CALLBACK(compileClick), dialog);
    gtk_box_pack_end(GTK_BOX(hbox), comp, TRUE, TRUE, 0);
    gtk_widget_show(comp);

    GuiObject compclose = GuiSimpleButton(
            GCS(MSG_ProgramEditor_CompileAndClose));
    g_object_set(G_OBJECT(compclose), "focus-on-click", false, NULL);
    GuiSetTooltip(compclose, GCS(MSG_ProgramEditor_CompileAndCloseBH));
    g_signal_connect_swapped(compclose, "clicked", 
	    G_CALLBACK(compileCloseClick), dialog);
    gtk_box_pack_end(GTK_BOX(hbox), compclose, TRUE, TRUE, 0);
    gtk_widget_show(compclose);

    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Program editor");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, 0);

    g_signal_connect(dialog, "delete-event", 
	    G_CALLBACK(programEditorOnDelete), d);

    if (!GuiOpenWindow(dialog)) {
	gtk_widget_destroy(dialog);
	dialog = 0;
    }

    return dialog;
}

GuiObject OpenProgramEditor(Project *pr)
{
    GuiObject go = (GuiObject)ProjectProgramEditor(pr);
    bool external = GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL);

    if (!go) {
	if (external) {
            return 0;
	}

        go = createProgramEditorGui(pr);
        if (!go) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_CantOpenProgramEditor));
            return 0;
        }
        undoClick(go);					// restore
        SetProjectProgramEditor(pr, go);
    } else {
	GuiClearWindowSleep(go);
	GuiWindowToFront(go);
	GuiActivateWindow(go);
    }

    ProgramEditorData *d = (ProgramEditorData *)
	g_object_get_data(G_OBJECT(go), "edit-data");
    gtk_widget_grab_focus(d->d_Editor);

    return go;
}

void ProgramEditorShowError(struct Project *pr, bool reload,
	const char *error, long line, long col, const char *filename)
{
    if (error == NULL) {
	return;
    } 

    if (IsProjectErrorString(error)) {
        Ask(AT_ERROR, GCS(MSG_StarOk), error);
        return;
    }

    GuiObject pe = ProjectProgramEditor(pr);
    if (line < 0) {
        Ask2(pe, NULL, AT_WARN, GCS(MSG_StarOk),
            GCS(MSG_ProgramEditor_NoDebugInfo), error);
	return;
    }

    if (filename == NULL && GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL)) {
        filename = GetPrefsString(pr, PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);
    }
    if (filename) {
	Ask2(pe, NULL, AT_ERROR, GCS(MSG_StarOk),
	      	GCS(MSG_ProgramEditor_IncludeFileError), 
		filename, line, col, error);
    } else if (!ProjectAdminMode(pr)) {
	Ask2(pe, NULL, AT_ERROR, GCS(MSG_StarOk),
	      	GCS(MSG_ProgramEditor_NotAdmin), line, col, error);
    } else {
	//
	// show the error in BeeBase' program editor
	// The editor might not be opened yet,
	// open and unchanged, or open and changed.
	//
	if (pe && reload) {
	    ProgramEditorData *d = (ProgramEditorData *)
		g_object_get_data(G_OBJECT(pe), "edit-data");
	    GtkTextBuffer *buffer = gtk_text_view_get_buffer(
		    GTK_TEXT_VIEW(d->d_Editor));
	    if (gtk_text_buffer_get_modified(buffer)) {
		// we should restore the program source
		GuiBeep();
		short rc = Ask2(pe, GCS(MSG_ProgramEditor), AT_WARN,
			GCS(MSG_ProgramEditor_ShowAnswers),
			GCS(MSG_ProgramEditor_ShowError), line, col, error);
		if (rc == 0) {
		    return;				// user canceled
		} else if (rc == 2) {
		    undoClick(pe);			// restore
		}
	    }
	}
	pe = OpenProgramEditor(pr);	// either creates or brings to front
	ProgramEditorData *d = (ProgramEditorData *)
    	    g_object_get_data(G_OBJECT(pe), "edit-data");
	if (pe) {
	    GuiSetWindowTitle(pe, error);
	    gtk_ext_editor_set_cursor(GTK_EXT_EDITOR(d->d_Editor), line, col);
	    GuiBeep();
	} // else OpenProgramEditor popped up an error dialog
    }
}

bool ProgramEditorFix(Project *pr)
{
    GuiBusyCursor busy;

    GuiObject pe = ProjectProgramEditor(pr);
    bool result = true;

    if (pe) {
	ProgramEditorData *d = (ProgramEditorData *)
	    g_object_get_data(G_OBJECT(pe), "edit-data");
	GtkTextBuffer *buffer = gtk_text_view_get_buffer(
    		GTK_TEXT_VIEW(d->d_Editor));
	if (gtk_text_buffer_get_modified(buffer)) {
	    // we should restore the program source
	    GuiBeep();
	    short rc = Ask2(pe, GCS(MSG_ProgramEditor), AT_WARN,
		    GCS(MSG_ProgramEditor_ShowAnswers),
		    GCS(MSG_ProgramEditor_ShowDeleted));
	    if (rc == 0) {
		result = false;			// user canceled
	    } else if (rc == 2) {
		undoClick(pe);			// restore
	    }
	}
    }

    if (result) {
	pe = OpenProgramEditor(pr);	// either creates or brings to front
	if (pe) {
            ProgramEditorData *d = (ProgramEditorData *)
                g_object_get_data(G_OBJECT(pe), "edit-data");
            d->d_DestroyOnClose = false;
	    compile(pe, true);			// show the location to fix
	    result = (gtk_dialog_run(GTK_DIALOG(pe)) == GTK_RESPONSE_OK);
            d->d_DestroyOnClose = true;
            if (!gtk_widget_get_visible(pe)) {      // closed?
                gtk_widget_destroy(pe);
            }
	}
    }

    return result;
}

const char *ProgramEditorCompile(GuiObject pe, bool reload,
	Program **compiled, long *line, long *col, const char **filename)
{
    if (!pe) {
	return GCS(MSG_Err_Internal);
    } else {
        ProgramEditorData *d = (ProgramEditorData *)
	    g_object_get_data(G_OBJECT(pe), "edit-data");
	Project *pr = d->d_Project;
	MemPool *memPool = ProjectMemPool(pr);
	GuiObject editor = d->d_Editor;
	Program *prg = 0;
	const char *error = 0;

    	GuiSetSleep(TRUE);
	GuiSetWindowTitle(pe, GCS(MSG_EditProgram));
	char *text = GuiEditorText(d->d_Editor, memPool);
	if (text == NULL) {
	    error = GCS(MSG_Err_Mem);
	} else {
	    error = CompileProjectProgram(d->d_Project, text, strlen(text),
		    &prg, line, col, filename);
	    if (compiled) {
		*compiled = prg;
	    }
	}
	FreeString(memPool, text);

        if (!error && reload) {
    	    SourceText *st = GetSourceText(prg);
	    const char *text = SourceTextText(st);
            guint top, line, col;

	    top = gtk_ext_editor_get_top(GTK_EXT_EDITOR(editor));
            gtk_ext_editor_get_cursor(GTK_EXT_EDITOR(editor), &line, &col);

	    SetPrefsValue(pr, PREFS_PROGRAMEDITOR_TOP, top);
	    SetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSX, col);
	    SetPrefsValue(pr, PREFS_PROGRAMEDITOR_POSY, line);

            GuiSetEditorText(editor, text);
            GuiSetEditorChanged(editor, false);
	    FreeSourceText(st);

	    gtk_ext_editor_set_top(GTK_EXT_EDITOR(editor), top);
            gtk_ext_editor_set_cursor(GTK_EXT_EDITOR(editor), line, col);
	}
	GuiSetSleep(FALSE);

	return error;
    }
}

const char *ProgramEditorCreateSourceExternal(GuiObject pe)
{
    if (pe == NULL) {
	return GCS(MSG_Err_Internal);
    } else {
        ProgramEditorData *d = (ProgramEditorData *)
	    g_object_get_data(G_OBJECT(pe), "edit-data");
	MemPool *memPool = ProjectMemPool(d->d_Project);
	const char *error = 0;

    	GuiSetSleep(TRUE);
	char *text = GuiEditorText(d->d_Editor, memPool);
	if (text == NULL) {
	    error = GCS(MSG_Err_Mem);
	} else {
	    error = CreateSourceExternal(d->d_Project, text, strlen(text));
	    FreeString(memPool, text);
	}
	GuiSetSleep(FALSE);
	return error;
    }
}

#else

void CloseProgramEditor(GuiObject obj) {}

bool ProgramEditorCheckClose(GuiObject obj) { return false; }

GuiObject OpenProgramEditor(Project *pr) { return 0; }

bool ProgramEditorFix(Project *pr) { return false; }

void ProgramEditorShowError(Project *pr, bool reload,
	const char *error, long line, long col, const char *filename) {}

const char *ProgramEditorCompile(GuiObject pe, bool reload,
	Program **compiled, long *line, long *col, char **filename)
{ return 0; }

const char *ProgramEditorCreateSourceExternal(GuiObject pe) { return 0; }

#endif

