//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkQueryEditor.cc
//
// Creation date: 14.12.2004
// Last modified: 03.03.2022
//

#include "system.h"
#include "queryeditor.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "gtk/gtkBuildGuiField.h" /* for SFWViewOnRowActivated */
#include "gtk/classes/gtkgroup.h"
#include "gtk/classes/gtkexteditor.h"
#include "gtk/classes/gtksfwview.h"
#include "gtk/classes/gtkpopwidget.h"
#include <gdk/gdkkeysyms.h>
#include "lang/compile.h"
#include "preferences.h"
#include "imexport.h"
#include "pdfexport.h"
#include "print.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

// #define USE_TABLE_POPUP

struct QueryEditorData
{
    Project *d_Project;
    QueryEditor *d_QueryEditor;
    Query *d_ActiveQuery;
    GuiObject d_QueryName;
    GuiObject d_Queries;
    GuiObject d_QueriesPopup;
    GuiObject d_QueryTable;
#ifdef USE_TABLE_POPUP
    GuiObject d_Tables;
    GuiObject d_TablesPopup;
#endif
    GuiObject d_Export, d_PDF, d_Print;
    GuiObject d_Editor;
    GuiObject d_SFWView;
};

static void exportClick(GuiObject go, QueryEditorData *d)
{
    GuiExportSFWList(d->d_Project, d->d_SFWView);
}

static void pdfClick(GuiObject go, QueryEditorData *d)
{
    GuiExportSFWListAsPDF(d->d_Project, d->d_SFWView);
}

static void printClick(GuiObject go, QueryEditorData *d)
{
    GuiPrintSFWList(d->d_Project, d->d_SFWView);
}

static void setActiveQuery(QueryEditorData *d, Query *q)
{
    if (q) {
        /* clear window title */
	GuiObject win = gtk_widget_get_toplevel(d->d_Editor);
	gtk_window_set_title(GTK_WINDOW(win), GCS(MSG_QueryEditor));

        const char *name = QueryName(q);
        if (!name) {
            name = GCS(MSG_Unnamed);
	}
	gtk_entry_set_text(GTK_ENTRY(d->d_QueryName), name);

#ifdef USE_TABLE_POPUP
        gtk_label_set_text(GTK_LABEL(d->d_QueryTable), 
                EmptyNullStr(TableName(QueryTable(q))));
#else
        gtk_combo_box_set_active(GTK_COMBO_BOX(d->d_QueryTable),NodeNum(QueryTable(q)));
#endif
        
        /* set source text */
        Program *prg = QueryProgram(q);
        SourceText *st = GetSourceText(prg);
        const char *src = SourceTextText(st);
	long len = src? strlen(src) : 0;
        char *kill, c;

        if (len >= 3) {
            src++;          /* skip "("   */
            len -= 3;       /* skip "\n)" */
            kill = (char *)src + len;
            c = *kill;
            *kill = 0;
            GuiSetEditorText(d->d_Editor, src);
            *kill = c;
        } else {
            GuiSetEditorText(d->d_Editor, "");
        }
        GuiSetEditorChanged(d->d_Editor, false);
        FreeSourceText(st);

	gtk_sfw_view_set_element(GTK_SFW_VIEW(d->d_SFWView), 0);

        d->d_ActiveQuery = q;
    }
}

static const char *saveQuery(QueryEditorData *d, long *line, long *col,
        bool force)
{
    if (d && d->d_ActiveQuery) {
        Project *pr = QueryEditorProject(d->d_QueryEditor);
	const char *name = gtk_entry_get_text(GTK_ENTRY(d->d_QueryName));
        const char *oldName = QueryName(d->d_ActiveQuery);
        long error = OK;

        if (strcmp(name? name : "", oldName? oldName : "") != 0) {
            error = ObtainProjectChangeLock(pr);
            if (error != OK) {
                return GES(error);
            }
            SetQueryName(d->d_ActiveQuery, name);
	    ProjectChange(pr, TRUE);
	}

#ifndef USE_TABLE_POPUP
        Table *tab = (Table *)FindNodeNum(ProjectTables(d->d_Project),
                gtkComboBoxActive(d->d_QueryTable));
        if (tab != QueryTable(d->d_ActiveQuery)) {
            error = ObtainProjectChangeLock(pr);
            if (error != OK) {
                return GES(error);
            }
            List *queries = QueryEditorQueries(d->d_QueryEditor);
            Query *q;
            if (queries) {
                FORNODE(q, queries) {
                    if (QueryTable(q) == tab) {
                        SetQueryTable(q, 0);
                    }
                }
            }
            SetQueryTable(d->d_ActiveQuery, tab);
	    ProjectChange(pr, TRUE);
        }
#endif

	GtkTextBuffer *buffer = gtk_text_view_get_buffer(
		GTK_TEXT_VIEW(d->d_Editor));
	gboolean modified = gtk_text_buffer_get_modified(buffer);
        if (force || modified) {
	    MemPool *memPool = ProjectMemPool(pr);
	    Program *prg = 0;
	    const char *errstr = 0;
            char *src = GuiEditorText(d->d_Editor, memPool);
            if (!src) {
                errstr = GCS(MSG_Err_Mem);
	    } else {
                errstr = CompileQuery(pr, src, strlen(src), &prg, line, col);
	    }
            FreeString(memPool, src);

            if (errstr) {
		return errstr;
	    }
            if (modified) {
                error = ObtainProjectChangeLock(pr);
                if (error != OK) {
                    return GES(error);
                }
            }
            SetQueryProgram(d->d_ActiveQuery, prg);
            setActiveQuery(d, d->d_ActiveQuery);         /* refreshes source */
            if (modified) {
                ProjectChange(pr, TRUE);
		gtk_text_buffer_set_modified(buffer, false);
            }
        }
    }
    return 0;
}

static void showerror(QueryEditorData *d, const char*error, long line, long col)
{
    if (error != NULL && error != EXECUTE_ABORT) {
        if (IsProjectErrorString(error)) {
            Ask(AT_ERROR, GCS(MSG_StarOk), error);
        } else if (d != NULL) {
            GuiObject win = gtk_widget_get_toplevel(d->d_Editor);
            gtk_window_set_title(GTK_WINDOW(win), error);
            gtk_ext_editor_set_cursor(GTK_EXT_EDITOR(d->d_Editor), 
                    line, (line == 1 && col > 0)? col - 1 : col);
            gtk_widget_grab_focus(d->d_Editor);
            GuiBeep();
        }
    }
}

static Project *queryProject = 0;

static const char *outputFunc(Element *elem)
{
    QueryEditor *se = ProjectQueryEditor(queryProject);
    GuiObject go = QueryEditorGui(se);
    QueryEditorData *d = (QueryEditorData *)
	(go? g_object_get_data(G_OBJECT(go), "edit-data") : 0);
    char *error = 0;

    if (d) {
	error = (char *)gtk_sfw_view_set_element(
                GTK_SFW_VIEW(d->d_SFWView), elem);
    } else {
	error = GCS(MSG_Err_Internal);
    }

    return error;
}

static void runClick(GuiObject go, QueryEditorData *d)
{
    GuiBusyCursor busy;

    if (d->d_ActiveQuery) {
	long line, col;
	const char *error = saveQuery(d, &line, &col, true);
	if (!error) {
	    /* execute program */
	    Program *prg = QueryProgram(d->d_ActiveQuery);
	    SetTableProgramRecords(d->d_Project);
	    queryProject = d->d_Project;
	    error = ExecuteProgram(prg, TRUE, NULL, NULL, TRUE, 
     		    outputFunc, &line, &col, NULL);
	    queryProject = 0;
	}
	if (error) {
	    showerror(d, error, line, col);
	}
    }
}

static gboolean queryNameOnKeyPressEvent(GtkWidget *widget, GdkEventKey *event,
	QueryEditorData *d)
{
    if (event->keyval == GDK_KEY_Return) {
	long line, col;
	const char *error = saveQuery(d, &line, &col, false);
	if (error) {
	    showerror(d, error, line, col);
	}
	return true;
    }
    return false;
}

static void popListviewOnRowActivated(GtkTreeView *tree,
	GtkTreePath *path, GtkTreeViewColumn *column, GtkPopString *pop)
{
    gtk_pop_string_close_popup(pop, true);
}

static gboolean queryBrowserOnStart(GtkPopWidget *pop, QueryEditorData *d)
{
    g_signal_stop_emission_by_name(pop, "start");

    List *queries = QueryEditorQueries(d->d_QueryEditor);
    long line, col;
    const char *error = saveQuery(d, &line, &col, false);
    if (error) {
	showerror(d, error, line, col);
	return false;
    }

    gtk_widget_set_sensitive(d->d_QueryName, false);
    gtk_widget_set_sensitive(d->d_Editor, false);

    GuiListSetList(d->d_Queries, queries);
    GuiListSetActiveNode(d->d_Queries, d->d_ActiveQuery);

    GtkWidget *window = gtk_pop_string_get_popup(GTK_POP_STRING(pop));
    gint width, height;
    gtk_window_get_size(GTK_WINDOW(window), &width, &height);
    height += 100;
    gtk_window_resize(GTK_WINDOW(window), width, height);

    gtk_widget_grab_focus(d->d_Queries);

    return true;
}

static void queryBrowserOnStop(GtkPopWidget *pop, QueryEditorData *d)
{
    setActiveQuery(d, (Query *)GuiListActiveEntry(d->d_Queries));
}

static void queryBrowserOnClose(GtkPopWidget *pop, gboolean success,
	QueryEditorData *d)
{
    List *queries = QueryEditorQueries(d->d_QueryEditor);
    GuiListGetList(d->d_Queries, queries);

    GuiListSetList(d->d_Queries, 0);

    gtk_widget_set_sensitive(d->d_QueryName, true);
    gtk_widget_set_sensitive(d->d_Editor, true);
}


static void newClick(GuiObject go, QueryEditorData *d)
{
    Query *q = NewQuery(d->d_QueryEditor, 0);
    if (q) {
	gtk_pop_string_close_popup(GTK_POP_STRING(d->d_QueriesPopup), false);
        setActiveQuery(d, q);
    }
}

static void duplicateClick(GuiObject go, QueryEditorData *d)
{
    Query *active = (Query *)GuiListActiveEntry(d->d_Queries);
    Query *q = NewQuery(d->d_QueryEditor, active);
    if (q) {
	GuiListSetList(d->d_Queries, QueryEditorQueries(d->d_QueryEditor));
        GuiListSetActiveNode(d->d_Queries, q);
    }
}

static int queryCompare(void *a, void *b)
{
    Query *q1 = (Query *)a, *q2 = (Query *)b;
    const char *n1 = QueryName(q1), *n2 = QueryName(q2);
    if (n1 && n2) {
        return CodesetStrCollate(n1, n2, CS_COLLATE_ORDER);
    } else if (n2) {
	return -1;
    } else if (n1) {
	return 1;
    } else {
	return 0;
    }
}

static void sortClick(GuiObject go, QueryEditorData *d)
{
    List *queries = QueryEditorQueries(d->d_QueryEditor);
    QLSort(queries, queryCompare);
    GuiListSetList(d->d_Queries, queries);
}

static void deleteClick(GuiObject go, QueryEditorData *d)
{
    Query *q = (Query *)GuiListActiveEntry(d->d_Queries);
    if (q) {
	Query *other = SUCC(q);
	if (!SUCC(other)) {
	    other = PRED(q);
	    if (!PRED(other)) {
		/* clear entry, we will not free last element */
		SetQueryName(q, 0);
		SetQueryProgram(q, 0);

		setActiveQuery(d, q);
		gtk_widget_queue_draw(d->d_Queries);
		return;
	    }
	}

	if (d->d_ActiveQuery == q) {
	    setActiveQuery(d, other);
	}

	/* remove and free query */
	GuiListRemoveEntry(d->d_Queries, q);
	DeleteQuery(q);
    }
}

static void queryCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    Query *q = 0;

    gtk_tree_model_get(model, iter, 0, &q, -1);

    const char *name = QueryName(q);
    if (!name) {
        name = GCS(MSG_Unnamed);
    }
    g_object_set(G_OBJECT(cell), "text", name, NULL);
}

#ifdef USE_TABLE_POPUP

static gboolean tableBrowserOnStart(GtkPopWidget *pop, QueryEditorData *d)
{
    g_signal_stop_emission_by_name(pop, "start");

    GuiListSetList(d->d_Tables, ProjectTables(d->d_Project));

    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(d->d_Tables));
    GtkListStore *store = GTK_LIST_STORE(model);
    GtkTreeIter iter;
    gtk_list_store_prepend(store, &iter);
    gtk_list_store_set(store, &iter, 0, NULL, -1);

    GuiListSetActiveEntry(d->d_Tables, QueryTable(d->d_ActiveQuery));

    GtkWidget *window = gtk_pop_string_get_popup(GTK_POP_STRING(pop));
    gint width, height;
    gtk_window_get_size(GTK_WINDOW(window), &width, &height);
    height += 100;
    gtk_window_resize(GTK_WINDOW(window), width, height);

    gtk_widget_grab_focus(d->d_Tables);

    return true;
}

static void tableBrowserOnStop(GtkPopWidget *pop, QueryEditorData *d)
{
    Table *tab =(Table *)GuiListActiveEntry(d->d_Tables);

    if (tab != QueryTable(d->d_ActiveQuery)) {
        long error = ObtainProjectChangeLock(pr);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }

        List *queries = QueryEditorQueries(d->d_QueryEditor);
        Query *q;
        if (queries) {
            FORNODE(q, queries) {
                if (QueryTable(q) == tab) {
                    SetQueryTable(q, 0);
                }
            }
        }
        SetQueryTable(d->d_ActiveQuery, tab);
        gtk_label_set_text(GTK_LABEL(d->d_QueryTable), 
                EmptyNullStr(TableName(tab)));
        ProjectChange(d->d_Project, TRUE);
    }
}

static void tableBrowserOnClose(GtkPopWidget *pop, gboolean success,
	QueryEditorData *d)
{
    GuiListSetList(d->d_Tables, 0);
}

static void tableCellDataFunc(GtkTreeViewColumn *column, GtkCellRenderer *cell, 
        GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
    Table *tab = 0;

    gtk_tree_model_get(model, iter, 0, &tab, -1);

    const char *name = TableName(tab);
    if (name) {
        g_object_set(G_OBJECT(cell), "text", name, NULL);
    } else {
        char *cont = g_strdup_printf("<i>%s</i>", GCS(MSG_Query_TableNone));
        g_object_set(G_OBJECT(cell), "markup", cont, NULL);
        g_free(cont);
    }
}

#else

static void tableCellDataFunc(GtkCellLayout *layout, GtkCellRenderer *cell, 
        GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
    Table *tab = 0;

    gtk_tree_model_get(model, iter, 0, &tab, -1);

    const char *name = TableName(tab);
    if (name) {
        g_object_set(G_OBJECT(cell), "text", name, NULL);
    } else {
        char *cont = g_strdup_printf("<i>%s</i>", GCS(MSG_Query_TableNone));
        g_object_set(G_OBJECT(cell), "markup", cont, NULL);
        g_free(cont);
    }
}

#endif

static gboolean queryEditorOnDelete(GuiObject win, GdkEvent *event,
	QueryEditorData *d)
{
    long line, col;
    const char *error = saveQuery(d, &line, &col, false);
    if (error && 
	    !Ask(AT_WARN, GCS(MSG_CloseAnswers), GCS(MSG_Query_Close))) {
	showerror(d, error, line, col);
    } else {
        CloseQueryEditor(d->d_Project);
    }
    return true;
}

static void onNotifyModel(GtkWidget *tree, GParamSpec *spec, QueryEditorData *d)
{
    bool enabled = (GuiListNumEntries(tree) > 0);

    GuiSetEnabled(d->d_Export, enabled);
    GuiSetEnabled(d->d_PDF, enabled);
    GuiSetEnabled(d->d_Print, enabled);
}

GuiObject OpenQueryGui(QueryEditor *qe)
{
    Project *pr = QueryEditorProject(qe);
    List *queries = QueryEditorQueries(qe);

    if (!queries) {
	return 0;
    }

    // check if we have any queries, if not allocate one
    if (!SUCC(HEAD(queries))) {
        Query *q = NewQuery(qe, 0);
	if (!q) {
	    return 0;
	}
    }

    QueryEditorData *d = g_new0(QueryEditorData, 1);
    d->d_Project = pr;
    d->d_QueryEditor = qe;

    d->d_QueryName = EntryObject, "xalign", 0.5, End;
    g_signal_connect(d->d_QueryName, "key-press-event", 
	    G_CALLBACK(queryNameOnKeyPressEvent), d);
    GuiSetTooltip(d->d_QueryName, GCS(MSG_Query_NameBH));

    d->d_Queries = GuiPointerTreeView(queryCellDataFunc, false);
    gtk_tree_view_set_reorderable(GTK_TREE_VIEW(d->d_Queries), TRUE);
    GtkTreeSelection *selection = 
	gtk_tree_view_get_selection(GTK_TREE_VIEW(d->d_Queries));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
    
    GuiObject newq = GuiSimpleButton(GCS(MSG_New));
    g_signal_connect(newq, "clicked", G_CALLBACK(newClick), d);
    GuiObject duplicate = GuiSimpleButton(GCS(MSG_Duplicate));
    g_signal_connect(duplicate, "clicked", G_CALLBACK(duplicateClick), d);
    GuiObject sort = GuiSimpleButton(GCS(MSG_Sort));
    g_signal_connect(sort, "clicked", G_CALLBACK(sortClick), d);
    GuiObject deleteq = GuiSimpleButton(GCS(MSG_Delete));
    g_signal_connect(deleteq, "clicked", G_CALLBACK(deleteClick), d);

    GuiObject browser = VGroup,
        "spacing", 0,
	"child", ScrolledWindowObject,
	    "hscrollbar-policy", GTK_POLICY_NEVER,
	    "shadow-type", GTK_SHADOW_IN,
	    "child", d->d_Queries,
	    End,
	"child", HGroup,
	    "spacing", 0,
	    // "same-width", true,
	    "child", newq,
	    "child", duplicate,
	    "child", sort,
	    "child", deleteq,
	    End, "weight", 0,
	End;
    g_object_set_data(G_OBJECT(browser), "tree", d->d_Queries);
    gtk_widget_show_all(browser);

    d->d_QueriesPopup = PopWidgetObject,
        "string", d->d_QueryName,
	"button", ButtonObject,
	    "child", ImageObject,
		"icon-name", "gtk-index", 
		"icon-size", GTK_ICON_SIZE_MENU,
		End,
	    End,
	"widget", browser,
	End;
    g_signal_connect(d->d_QueriesPopup, "start", 
	    G_CALLBACK(queryBrowserOnStart), d);
    g_signal_connect(d->d_QueriesPopup, "stop", 
	    G_CALLBACK(queryBrowserOnStop), d);
    g_signal_connect_after(d->d_QueriesPopup, "close", 
	    G_CALLBACK(queryBrowserOnClose), d);
    g_signal_connect(d->d_Queries, "row-activated", 
	    G_CALLBACK(popListviewOnRowActivated), d->d_QueriesPopup);

#ifdef USE_TABLE_POPUP
    d->d_QueryTable = LabelObject,
        "ypad", 1,
        "xpad", 2,
        "xalign", 0.5,
        End;
    GuiSetTooltip(d->d_QueryTable, GCS(MSG_Query_TableBH));
    GuiObject tableFrame = FrameObject,
        "shadow-type", GTK_SHADOW_IN,
        "child", d->d_QueryTable,
        End;

    d->d_Tables = GuiPointerTreeView(tableCellDataFunc, false);
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(d->d_Queries));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);

    d->d_TablesPopup = PopWidgetObject,
        "string", tableFrame,
	"button", ButtonObject,
	    "child", ImageObject,
		"icon-name", "gtk-index", 
		"icon-size", GTK_ICON_SIZE_MENU,
		End,
	    End,
        "widget", ScrolledWindowObject,
	    "hscrollbar-policy", GTK_POLICY_NEVER,
	    "shadow-type", GTK_SHADOW_IN,
	    "child", d->d_Tables,
	    End,
        End;
    g_signal_connect(d->d_TablesPopup, "start", 
	    G_CALLBACK(tableBrowserOnStart), d);
    g_signal_connect(d->d_TablesPopup, "stop", 
	    G_CALLBACK(tableBrowserOnStop), d);
    g_signal_connect_after(d->d_TablesPopup, "close", 
	    G_CALLBACK(tableBrowserOnClose), d);
    g_signal_connect(d->d_Tables, "row-activated", 
	    G_CALLBACK(popListviewOnRowActivated), d->d_TablesPopup);

    GuiObject queryTable = d->d_TablesPopup;
#else
    GtkListStore *store = gtk_list_store_new(1, G_TYPE_POINTER);
    GuiSetListStoreList(store, ProjectTables(pr));
    GtkTreeIter iter;
    gtk_list_store_prepend(store, &iter);
    gtk_list_store_set(store, &iter, 0, NULL, -1);

    d->d_QueryTable = ComboBoxObject,
        "model", GTK_TREE_MODEL(store),
        End;
    GtkCellLayout *cl = GTK_CELL_LAYOUT(d->d_QueryTable);
    GtkCellRenderer *cell = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start(cl, cell, true);
    gtk_cell_layout_set_cell_data_func(cl, cell, tableCellDataFunc, 0, 0);

    GuiSetTooltip(d->d_QueryTable, GCS(MSG_Query_TableBH));

    GuiObject queryTable = d->d_QueryTable;
#endif

    GuiObject run = GuiSimpleButton(GCS(MSG_Query_Run));
    g_signal_connect(run, "clicked", G_CALLBACK(runClick), d);
    GuiSetTooltip(run, GCS(MSG_Query_RunBH));

    d->d_Export = GuiSimpleButton(GCS(MSG_Query_Export));
    g_signal_connect(d->d_Export, "clicked", G_CALLBACK(exportClick), d);
    GuiSetTooltip(d->d_Export, GCS(MSG_Query_ExportBH));
    gtk_widget_set_sensitive(d->d_Export, false);
    g_object_add_weak_pointer(G_OBJECT(d->d_Export), (gpointer *)&d->d_Export);

    d->d_PDF = GuiSimpleButton(GCS(MSG_Query_PDF));
    g_signal_connect(d->d_PDF, "clicked", G_CALLBACK(pdfClick), d);
    GuiSetTooltip(d->d_PDF, GCS(MSG_Query_PDFBH));
    gtk_widget_set_sensitive(d->d_PDF, false);
    g_object_add_weak_pointer(G_OBJECT(d->d_PDF), (gpointer *)&d->d_PDF);

    d->d_Print = GuiSimpleButton(GCS(MSG_Query_Print));
    g_signal_connect(d->d_Print, "clicked", G_CALLBACK(printClick), d);
    GuiSetTooltip(d->d_Print, GCS(MSG_Query_PrintBH));
    gtk_widget_set_sensitive(d->d_Print, false);
    g_object_add_weak_pointer(G_OBJECT(d->d_Print), (gpointer *)&d->d_Print);

    GuiObject editorBox = GtkNewEditor(pr, true, false, true, &d->d_Editor);
    editorBox = GuiMakeTooltip(editorBox, GCS(MSG_Query_ProgramBH));

    d->d_SFWView = SFWViewObject,
        "rules-hint", true,
        "show-titles", true,
        End;
    GuiSetTooltip(d->d_SFWView,GCS(MSG_Query_OutputBH));

    GuiObject root = GroupAlignment(false),
        "child", VGroup,
            "child", HGroup,
                "child", d->d_QueriesPopup, "weight", 3,
                "child", queryTable, "weight", 1,
                "child", HGroup,
                    "same-width", true,
                    "child", run,
                    "child", d->d_Export,
                    "child", d->d_PDF,
                    "child", d->d_Print,
                    End, "weight", 3,
                End, "vert-weight", 0,
            "child", VPanedObject,
                "child", editorBox,
                "child", ScrolledWindowObject,
                    "shadow-type", GTK_SHADOW_ETCHED_IN,
                    "hscrollbar-policy", GTK_POLICY_AUTOMATIC,
                    "child", d->d_SFWView,
                    End,
                "position", 60,
                End,
            End,
	End;

    g_signal_connect(d->d_SFWView, "row-activated", 
	    G_CALLBACK(SFWViewOnRowActivated), ProjectRootDisplayObject(pr));
    g_signal_connect(d->d_SFWView, "notify::model", 
	    G_CALLBACK(onNotifyModel), d);

    GuiObject win = WindowObject,
        "title", GCS(MSG_QueryEditor),
	"window-position", GTK_WIN_POS_CENTER,
	"default-width", MIN(gdk_screen_width() / 2, 800),
	"default-height", MIN(gdk_screen_height() / 2, 600),
    	"child", root,
        End;
    WindowDimensions wd;
    if (GetPrefsWindowDimensions(pr, PREFS_QUERYEDITOR, &wd)) {
        GuiSetWindowDimensions(win, &wd);
    }
    GuiSetWindowProject(win, pr);
    GuiSetHelpNode(win, "Query editor");
    g_object_set_data_full(G_OBJECT(win), "edit-data", d, g_free);
    GuiAddWindowCallbacks(win, 0);

    g_signal_connect(win, "delete-event", G_CALLBACK(queryEditorOnDelete), d);
    gtk_widget_show_all(root);

    Query *q = (Query *)HEAD(queries);
    if (SUCC(q)) {
        setActiveQuery(d, q);
    }
    gtk_widget_grab_focus(d->d_Editor);

    if (!GuiOpenWindow(win)) {
	gtk_widget_destroy(win);
	win = 0;
    }
        
    return win;
}

void CloseQueryGui(GuiObject go)
{
    if (go) {
	QueryEditorData *d = (QueryEditorData *)
	    g_object_get_data(G_OBJECT(go), "edit-data");

        saveQuery(d, 0, 0, false);

        WindowDimensions wd;
        if (GuiGetWindowDimensions(go, &wd)) {
            SetPrefsWindowDimensions(d->d_Project, PREFS_QUERYEDITOR, &wd);
        }
	gtk_widget_destroy(go);
    }
}

void QueryEditorRemRecord(Table *tab, Record *rec)
{
    Project *pr = TableProject(tab);
    QueryEditor *qe = ProjectQueryEditor(pr);
    GuiObject go = QueryEditorGui(qe);
    if (go) {
	QueryEditorData *d = (QueryEditorData *)
	    g_object_get_data(G_OBJECT(go), "edit-data");

	if (d && d->d_SFWView) {
	    gtk_sfw_view_rem_record(GTK_SFW_VIEW(d->d_SFWView), rec);
	}
    }
}

void QueryEditorRemAllRecords(Project *pr)
{
    QueryEditor *qe = ProjectQueryEditor(pr);
    GuiObject go = QueryEditorGui(qe);
    if (go) {
	QueryEditorData *d = (QueryEditorData *)
	    g_object_get_data(G_OBJECT(go), "edit-data");

	if (d && d->d_SFWView) {
	    gtk_sfw_view_rem_all_records(GTK_SFW_VIEW(d->d_SFWView));
	}
    }
}

void QueryGuiActivate(GuiObject go)
{
    if (go) {
        /* bring window to front and activate it */
	GuiWindowToFront(go);
	GuiActivateWindow(go);
    }
}

void QueryGuiRunQuery(GuiObject go, Query *query)
{
    if (go && query) {
        QueryEditorData *d = (QueryEditorData *)
            g_object_get_data(G_OBJECT(go), "edit-data");

        if (d) {
            if (d->d_ActiveQuery != query) {
                long line, col;
                const char *error = saveQuery(d, &line, &col, false);
                if (error) {
                    showerror(d, error, line, col);
                    return;
                }
                setActiveQuery(d, query);
            }
            runClick(go, d);
        }
    }
}

