//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkQuickSearch.cc
//
// Interactive global quick search popup-window
//
// Creation date: 01.10.2006
// Last modified: 05.09.2012
//

#include "gtk/gtkQuickSearch.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "gdk/gdkkeysyms.h"

#undef USE_POPUP
#undef USE_MODAL

static GuiObject quickSearchWindow = 0;
static GuiObject quickSearchEntry = 0;
static GuiObject quickSearchGuiObject = 0;
static GuiObject quickSearchTopLevel = 0;
static gulong quickSearchConfigureEventId = 0;
static QuickSearchCallback quickSearchCallback = 0;
static void *quickSearchPointer = 0;

void GuiQuickSearchExit(void)
{
    if (quickSearchConfigureEventId && quickSearchTopLevel) {
        g_signal_handler_disconnect(G_OBJECT(quickSearchTopLevel),
                quickSearchConfigureEventId);
    }
    quickSearchTopLevel = 0;
    quickSearchConfigureEventId = 0;
    GuiObject win = quickSearchWindow;
    if (win) {
	gtk_widget_destroy(win);
        quickSearchWindow = 0;
    }
    quickSearchEntry = 0;
    quickSearchGuiObject = 0;
    quickSearchCallback = 0;
    quickSearchPointer = 0;
}

static void quickSearchChange(GtkEditable *edit, GtkWidget *entry)
{
    if (quickSearchCallback) {
	quickSearchCallback(quickSearchGuiObject,
		gtkEntryText(entry), QUICK_SEARCH_START, quickSearchPointer);
    }
}

static gboolean onKeyPressEvent(GtkWidget *entry, GdkEventKey *event)
{
    int cmd;

    switch (event->keyval) {
	case GDK_KEY_Up: 
	case GDK_KEY_KP_Up: cmd = QUICK_SEARCH_PREV; break;
	case GDK_KEY_Down: 
	case GDK_KEY_KP_Down: cmd = QUICK_SEARCH_NEXT; break;
	default: return false;
    }
    if (quickSearchCallback) {
	quickSearchCallback(quickSearchGuiObject,
		gtkEntryText(entry), cmd, quickSearchPointer);
    }
    return true;
}

static gboolean onScrollEvent(GtkWidget *entry, GdkEventScroll *event)
{
    int cmd;

    switch (event->direction) {
	case GDK_SCROLL_UP: cmd = QUICK_SEARCH_PREV; break;
	case GDK_SCROLL_DOWN: cmd = QUICK_SEARCH_NEXT; break;
	default: return false;
    }
    if (quickSearchCallback) {
	quickSearchCallback(quickSearchGuiObject,
		gtkEntryText(entry), cmd, quickSearchPointer);
    }
    return true;
}

static void quickSearchFinish(GtkWidget *win, gpointer x)
{
    if (quickSearchCallback) {
	quickSearchCallback(quickSearchGuiObject,
		gtkEntryText(quickSearchEntry),
		GPOINTER_TO_INT(x)?  QUICK_SEARCH_COMMIT : QUICK_SEARCH_CANCEL,
		quickSearchPointer);
    }
    GuiQuickSearchExit();
}

static gboolean onDeleteEvent(GtkWidget *win)
{
    quickSearchFinish(win, GINT_TO_POINTER(false));
    return true;
}

#ifdef USE_MODAL
static gboolean onButtonPressEvent(GtkWidget *win, GdkEventButton *event)
{
    GuiObject go = quickSearchGuiObject;
    if (event->window != win->window) {
	gtk_widget_event(go, (GdkEvent *)event);
    }
    quickSearchFinish(win, GINT_TO_POINTER(false));

    return true;
}

#else

static void onNotifyIsActive(GtkWidget *win)
{
    if (!gtk_window_is_active(GTK_WINDOW(win))) {
	quickSearchFinish(win, GINT_TO_POINTER(false));
    }
}

static void onMenuHide(GtkMenu *menu, GtkWidget *win)
{
    g_signal_handlers_unblock_by_func(win, (gpointer *)onNotifyIsActive, 0);
}

static void onPopulatePopup(GtkWidget *entry, GtkMenu *menu, GtkWidget *win)
{
    g_signal_connect(menu, "hide", G_CALLBACK(onMenuHide), win);
    g_signal_handlers_block_by_func(win, (gpointer *)onNotifyIsActive, 0);
}
#endif

static void setPosition(void)
{
    GuiObject go = quickSearchGuiObject;
    GuiObject top = quickSearchTopLevel;
    // place window at bottom center of go
    if (top) {
        GdkWindow *window = gtk_widget_get_window(top);
        if (window) {
            GtkAllocation alloc;
            gtk_widget_get_allocation(go, &alloc);
            gint x, y, wx, wy;
            gdk_window_get_origin(window, &x, &y);
            if (gtk_widget_translate_coordinates(go, top, 0, 0, &wx, &wy)) {
                gint screen_width = gdk_screen_width();
                gint screen_height = gdk_screen_height();
                gint win_width, win_height;
                GuiObject win = quickSearchWindow;
#if GTK_MAJOR_VERSION == 2
                GtkRequisition requisition;
                gtk_widget_size_request(win, &requisition);
                win_width = requisition.width;
                win_height = requisition.height;
#else
                gtk_widget_get_preferred_width(win, 0, &win_width);
                gtk_widget_get_preferred_height(win, 0, &win_height);
#endif
                x += wx + MAX(alloc.width - win_width, 30);
                y += wy + alloc.height;
                x = CLAMP(x, 0, screen_width - win_width);
                y = CLAMP(y, 0, screen_height - win_height);
                gtk_window_move(GTK_WINDOW(win), x, y);
            }
        }
    }
}

static gboolean onConfigure(void)
{
    setPosition();
    return false;
}

bool GuiQuickSearchInit(GuiObject go, GdkEventKey *event, 
	QuickSearchCallback cb, void *ptr)
{
    GuiQuickSearchExit();	// make sure we have a clean state

    // check if the key event is suitable for starting a search string
#ifdef WINDOWS
    gunichar code = gdk_keyval_to_unicode(event->keyval);
    if (!g_unichar_isalnum(code) && !g_unichar_ispunct(code)) {
       	return false;
    }
    gchar buf[8];
    gint len = g_unichar_to_utf8(code, buf);
    if (len <= 0) {
	return false;
    }
    buf[len] = 0;
#else 
    // this does not seem to work on Windows
    GtkIMContext *imc = gtk_im_multicontext_new();
    GtkWidget *map = go;
    while (map && !gtk_widget_get_window(map)) {
	map = gtk_widget_get_parent(map);
    }
    if (map) {
	gtk_im_context_set_client_window(imc, gtk_widget_get_window(map));
    }
    gboolean rc = gtk_im_context_filter_keypress(imc, event);
    g_object_unref(G_OBJECT(imc));
    if (!rc) {
	return false;
    }
#endif

    // create popup entry
    quickSearchGuiObject = go;
    quickSearchCallback = cb;
    quickSearchPointer = ptr;

    if (quickSearchCallback) {
	quickSearchCallback(quickSearchGuiObject, 0,
		QUICK_SEARCH_SETUP, quickSearchPointer);
    }

    GuiObject entry = EntryObject, End;
    quickSearchEntry = entry;
    gtk_widget_show(entry);
    g_signal_connect(entry, "activate", G_CALLBACK(quickSearchFinish), 
	    GINT_TO_POINTER(true));
    g_signal_connect(entry, "changed", G_CALLBACK(quickSearchChange), entry);
    g_signal_connect(entry, "key-press-event", G_CALLBACK(onKeyPressEvent), 0);
    g_signal_connect(entry, "scroll_event", G_CALLBACK(onScrollEvent), 0);

    GuiObject win = WindowObject,
#ifdef USE_POPUP
	"type", GTK_WINDOW_POPUP,
#else
	"decorated", false,
	"skip-pager-hint", true,
	"skip-taskbar-hint", true,
	"type-hint", GDK_WINDOW_TYPE_HINT_DIALOG,
#endif
#ifdef USE_MODAL
	"modal", true,
#endif
        "transient-for", GuiFindTransientWindow(go),
	"border-width", 3,
	"child", entry,
	End;
    GtkWidget *top = go? gtk_widget_get_toplevel(go) : 0;
    if (top && GTK_IS_WINDOW(top) && gtk_window_has_group(GTK_WINDOW(top))) {
	 gtk_window_group_add_window(
                 gtk_window_get_group(GTK_WINDOW(top)), GTK_WINDOW(win));
    }
    quickSearchWindow = win;
    GuiSetHelpNode(win, "Changing records");
    GuiAddWindowCallbacks(win, 0);
    g_signal_connect(win, "delete-event", G_CALLBACK(onDeleteEvent), 0);
#ifdef USE_MODAL
    g_signal_connect(win,"button-press-event",G_CALLBACK(onButtonPressEvent),0);
#else
    g_signal_connect(win, "notify::is-active", G_CALLBACK(onNotifyIsActive), 0);
    g_signal_connect(entry, "populate-popup", G_CALLBACK(onPopulatePopup), win);
#endif
    gtk_widget_realize(win);		// make sure things are setup

    // position window
    if (top) {
        quickSearchTopLevel = top;
        GuiObject *p = &quickSearchTopLevel;
        quickSearchConfigureEventId = g_signal_connect(G_OBJECT(top),
                "configure-event", G_CALLBACK(onConfigure), 0);
        g_object_add_weak_pointer(G_OBJECT(top), (gpointer *)p);
        setPosition();
    }
    gtk_widget_grab_focus(entry);
    gtk_widget_show(win);

    // let the entry process the keypress event
#ifdef WINDOWS
    gtk_entry_set_text(GTK_ENTRY(entry), buf);
    gtk_editable_set_position(GTK_EDITABLE(entry), -1);
#else
    GtkWidget *w = entry;
    GdkWindow *window = gtk_widget_get_window(w);
    if (!window) {
        w = win;
        window = gtk_widget_get_window(w);
    }
    if (window) {
	GdkEvent *copy = gdk_event_copy((GdkEvent *)event);
	GdkEventKey *copykey = (GdkEventKey *)copy;
	g_object_ref(window);
	g_object_unref(copykey->window);
	copykey->window = window;
	gtk_widget_event(w, copy);
	gdk_event_free(copy);
    }
#endif

    return true;
}

