//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkRequesters.cc
//
// Creation date: 23.03.2004
// Last modified: 04.06.2022
//

#include "gtk/gtkBuildGui.h"
#include "gtk/gtkAmiFileChooser.h"
#include "gtk/classes/gtkpopselection.h"
#include "gtk/classes/gtkextentry.h"
#include "gtk/classes/gtkgroup.h"
#include "gtk/classes/gtksfwview.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "requesters.h"
#include "lang/program.h"
#include "preferences.h"
#include "project.h"
#include "dbtypes.h"
#include "labels.h"
#include "memo.h"
#include "util.h"
#include "memory.h"
#include "appstrings_cat.h"

static char *getdirfilename(Project *pr, MemPool *memPool,
	const char *title, const char *okText, 
	const char *pref, long prefItem, 
	bool dirMode, bool saveMode)
{
    GuiBusyCursor busy;
    GtkFileChooserAction action;

    ProjectSetCurrentDir(pr);	    /* make sure we have correct current dir */
    if (dirMode) {
        action = saveMode? GTK_FILE_CHOOSER_ACTION_CREATE_FOLDER
            : GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER;
    } else {
        action = saveMode? GTK_FILE_CHOOSER_ACTION_SAVE 
            : GTK_FILE_CHOOSER_ACTION_OPEN;
    }

    if (!pref && prefItem > 0) {
	pref = GetPrefsString(pr, prefItem);
    }
    GtkWindow *parent = GuiFindTransientWindow(NULL);
    gchar *path = gtkAmiFileChoose(title, okText, pref, action, parent);

    char *result = 0;
    if (path) {
	result = AllocString(memPool, path);
	if (result && prefItem > 0) {
	    SetPrefsString(pr, prefItem, result);
        }
	g_free(path);
    }

    return result;
}

char *GetFilename(Project *pr, MemPool *memPool,
        const char *title, const char *okText,
        const char *pref, long prefItem, bool saveMode)
{
    return getdirfilename(pr, memPool, title, okText, 
	    pref, prefItem, false, saveMode);
}

char *GetDirname(Project *pr, MemPool *memPool, 
	const char *title, const char *okText,
        const char *pref, long prefItem, bool saveMode)
{
    return getdirfilename(pr, memPool, title, okText, 
	    pref, prefItem, true, saveMode);
}

static short Ask(GuiObject obj, const char *title, AskType askType,
	const char *buttons, const char *format, va_list ap)
{
    GuiBusyCursor busy;

    GtkWindow *parent = GuiFindTransientWindow(obj);
    GtkMessageType type;

    switch (askType) {
	case AT_INFO:	  type = GTK_MESSAGE_INFO; break;
	default:
	case AT_WARN:     type = GTK_MESSAGE_WARNING; break;
	case AT_QUESTION: type = GTK_MESSAGE_QUESTION; break;
        case AT_ERROR:	  type = GTK_MESSAGE_ERROR; break;
    }
    
    gchar *message = 0;
    if (format) {
	message = g_strdup_vprintf(format, ap);
    }
    GtkWidget *dialog = gtk_message_dialog_new(parent, GTK_DIALOG_MODAL, type,
	    GTK_BUTTONS_NONE, message? "%s" : 0, message);
    g_free(message);

    g_object_set(G_OBJECT(dialog), 
            "title", title?  title : GCS(MSG_APP), 
            "skip-taskbar-hint", false,          /* needed on windows */
            NULL);
    if (type == GTK_MESSAGE_INFO &&
            gtk_message_dialog_get_image(GTK_MESSAGE_DIALOG(dialog)) != NULL &&
            gtk_icon_theme_has_icon(gtk_icon_theme_get_default(),
                BEE_ICON_NAME)) {
        GtkWidget *image = gtk_image_new_from_icon_name(BEE_ICON_NAME,
                GTK_ICON_SIZE_DIALOG);
        gtk_message_dialog_set_image(GTK_MESSAGE_DIALOG(dialog), image);
        gtk_widget_show(image);
    }

    if (!buttons || !buttons[0]) {
	buttons = GCS(MSG_StarOk);
    }

    GtkWidget *button = 0;
    int num = 0;
    while (buttons) {
	bool def = false;

	if (buttons[0] == '*') {
	    def = true;
	    buttons++;
	}

	int id;
	const char *next = CodesetStrchr(buttons, '|');
	gchar *label;
	if (next) {
	    label = g_strndup(buttons, next - buttons);
	    buttons = next + 1;
	    id = num + 1;
	} else {
	    label = g_strdup(buttons);
	    id = 0;
	    buttons = 0;
	}

#if 0
	button=gtk_dialog_add_button(GTK_DIALOG(dialog), label, id);
#else
	button = ButtonObject, 
		"label", label, 
		"use-underline", true,
		"can-default", def,
		End;
	gtk_dialog_add_action_widget(GTK_DIALOG(dialog), button, id);
        gtk_widget_show(button);
#endif
	num++;

	if (def) {
	    gtk_window_set_default(GTK_WINDOW(dialog), button);
	}
	if (def || num == 1) {
	    gtk_widget_grab_focus(button);
	}
	g_free(label);
    }

    // we reverse the order of buttons
    if (button) {
    	GtkWidget *box = gtk_widget_get_parent(button);
	GList *children = gtk_container_get_children(GTK_CONTAINER(box));
	int i = 0;
	
	for (GList *child = children; child; child = child->next) {
	    gtk_box_reorder_child(GTK_BOX(box), GTK_WIDGET(child->data), i++);
	}

	g_list_free(children);
    }

    gtk_window_set_position(GTK_WINDOW(dialog), 
	    obj? GTK_WIN_POS_CENTER_ON_PARENT : GTK_WIN_POS_CENTER);
    GuiSetDialogButtonOrder(dialog);
    GuiAddWindowCallbacks(dialog, button);

    gint answer = gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);

    if (answer < 0) {
	answer = 0;
    }

    return answer;
}

short Ask(AskType type, const char *buttons, const char *text, ...)
{
    va_list ap;
    short rc;

    va_start(ap, text);
    rc = Ask(NULL, NULL, type, buttons, text, ap);
    va_end(ap);
 
    return rc;
}

short Ask2(GuiObject obj, const char *title, AskType type,
	const char *buttons, const char *text, ...)
{
    va_list ap;
    short rc;

    va_start(ap, text);
    rc = Ask(obj, title, type, buttons, text, ap);
    va_end(ap);
 
    return rc;
}

#define MAXLEN 1000

static void onFileStop(GtkPopSelection *pop)
{
    g_signal_stop_emission_by_name(pop, "stop");

    GtkWidget *entry = gtk_pop_string_get_string(GTK_POP_STRING(pop));
    const gchar *text = gtk_entry_get_text(GTK_ENTRY(entry));
    const gchar *s = g_utf8_strchr(text, -1, ' ');
    GtkFileChooserDialog *dlg = gtk_pop_selection_get_file_chooser_dialog(pop);
    gchar *path = gtk_ami_file_chooser_get_filename(GTK_FILE_CHOOSER(dlg));
    gchar *upath = g_filename_to_utf8(path, -1, 0, 0, 0);
    gchar *joined = g_strjoin(0, upath, s, NULL);

    gtk_entry_set_text(GTK_ENTRY(entry), EmptyNullStr(joined));
    g_free(joined);
    g_free(upath);
    g_free(path);
}

static gboolean onFileStart(GtkPopSelection *pop)
{
    g_signal_stop_emission_by_name(pop, "start");

    GtkWidget *entry = gtk_pop_string_get_string(GTK_POP_STRING(pop));
    const gchar *text = gtk_entry_get_text(GTK_ENTRY(entry));
    const gchar *s = g_utf8_strchr(text, -1, ' ');
    gchar *path = g_filename_from_utf8(text, s? s - text : -1, 0, 0, 0);
    GtkFileChooserDialog *dlg = gtk_pop_selection_get_file_chooser_dialog(pop);
    gtk_ami_file_chooser_set_filename(GTK_FILE_CHOOSER(dlg), path);
    g_free(path);

    return true;
}
    
const char *AskString(struct Project *pr, const char *caption,
	const char *oktext, const char *contents, ulong maxlen, 
	AskStringType type)
{
    GuiBusyCursor busy;

    const char *title = ProjectTitle(pr);
    if (!title) {
	title = GCS(MSG_APP);
    }
    if (!caption) {
        caption = GCS(MSG_EnterString);
    }
    if (!oktext) {
        oktext = GCS(MSG_Ok);
    }
    if (maxlen > MAXLEN) {
        maxlen = MAXLEN;
    }

    GtkWidget *go, *file = 0;
    GtkWidget *entry = go = EntryObject,
        "activates-default", true,
        "text", EmptyNullStr(contents),
	"max-length", maxlen,
	"visibility", (type != AST_Secret),
        End;

    if (type == AST_FileSelect) {
        go = PopSelectionObject,
	    "selection-type", GTK_TYPE_FILE_CHOOSER_DIALOG,
	    "string", entry,
            "button", file = GuiPopButtonSetCWD(SELECT_FILE_ICON_NAME, pr),
	    End;
	g_signal_connect(go, "start", G_CALLBACK(onFileStart), 0);
	g_signal_connect(go, "stop", G_CALLBACK(onFileStop), 0);
    }

    GtkWidget *group;
    GtkWidget *root = GroupAlignment(false),
        "child", group = VGroup,
	    "child", LabelObject,
		"label", caption, 
		End, "vert-weight", 0,
	    "child", go, "vert-weight", 0,
	    End,
	End;

    GtkWidget *dialog = DialogObject,
        "title", title,
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	End;
    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    GtkWidget *cancel = GuiSimpleButton(GCS(MSG_Cancel));
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(cancel);

    GtkWidget *ok = GuiSimpleButton(oktext, true);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(ok);
    gtk_window_set_default(GTK_WINDOW(dialog), ok);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, cancel);

    static char buffer[MAXLEN+1];
    char *ret = 0;
    if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
	const gchar *text = gtk_entry_get_text(GTK_ENTRY(entry));
	CodesetStrNBytesCopy(buffer, text, MAXLEN);
	ret = buffer;
    }

    gtk_widget_destroy(dialog);

    return ret;
}

bool AskInteger(Project *pr, const char *caption, const char *oktext,
        const char *contents, long min, long max, long *result)
{
    GuiBusyCursor busy;

    const char *title = ProjectTitle(pr);
    if (!title) {
	title = GCS(MSG_APP);
    }
    if (!caption) {
        caption = GCS(MSG_EnterInteger);
    }
    if (!oktext) {
        oktext = GCS(MSG_Ok);
    }

    GtkWidget *spin = SpinButtonObject,
        "activates-default", true,
        "adjustment", gtk_adjustment_new(0, min, max, 1, 10, 0),
	"digits", 0,
	"numeric", true,
	"snap-to-ticks", false,
	"update-policy", GTK_UPDATE_IF_VALID,
	"wrap", false,
	End;
    if (contents) {
        long val;
        if (StrToInt(contents, &val)) {
            g_object_set(G_OBJECT(spin), "value", (gdouble)val, NULL);
        } else  {
            g_object_set(G_OBJECT(spin), "text", contents, NULL);
        }
    }

    /*
    ** Create min max string
    */
    static char minmax[128];
    sprintf(minmax, "%s: %ld, %s: %ld", GCS(MSG_Min), min, GCS(MSG_Max), max);
    GtkWidget *info = FrameObject,
	"shadow-type", GTK_SHADOW_IN,
	"child", LabelObject,
	    "label", minmax,
	    "ypad", 1,
	    "xpad", 2,
	    End,
	End;

    GtkWidget *group;
    GtkWidget *root = GroupAlignment(false),
        "child", group = VGroup,
	    "child", LabelObject, 
		"label", caption, 
		End, "vert-weight", 0,
	    "child", spin, "vert-weight", 0,
	    "child", HGroup,
	        "child", SpaceObject, End,
		"child", info, "weight", 0,
	        "child", SpaceObject, End,
		End, "vert-weight", 0,
	    End,
	End;

    GtkWidget *dialog = DialogObject,
        "title", title,
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	End;
    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    GtkWidget *cancel = GuiSimpleButton(GCS(MSG_Cancel));
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(cancel);

    GtkWidget *ok = GuiSimpleButton(oktext, true);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(ok);
    gtk_window_set_default(GTK_WINDOW(dialog), ok);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, cancel);

    bool ret = false;
    while (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
	const gchar *text = gtk_entry_get_text(GTK_ENTRY(spin));
	long l;
	if (text && StrToInt(text, &l) && l >= min && l <= max) {
	    if (result) {
		*result = l;
	    }
	    ret = true;
	    break;
	}

	GuiBeep();
	gtk_widget_grab_focus(spin);
    }

    gtk_widget_destroy(dialog);

    return ret;
}

static void askChoiceOnSelectionChanged(GtkTreeSelection *selection,
        GtkWidget *entry)
{
    GtkTreeIter iter;

    if (gtk_tree_selection_get_selected(selection, NULL, &iter)) {
        GtkTreeView *tree = gtk_tree_selection_get_tree_view(selection);
        const struct QueryLine *ql = gtk_sfw_view_get_query_line(
                GTK_SFW_VIEW(tree), &iter);

        if (ql && ql->ql_Entries[0]) {
            gtk_entry_set_text(GTK_ENTRY(entry), ql->ql_Entries[0]->qe_String);
            gtk_widget_grab_focus(GTK_WIDGET(entry));
	}
    }
}

static void askChoiceCountSelected(GtkTreeModel *model, GtkWidget *count)
{
    gint num = 0;
    GtkTreeIter iter;
    if (gtk_tree_model_get_iter_first(model, &iter)) {
	do {
            struct QueryLine *ql = NULL;
	    gtk_tree_model_get(model, &iter, 0, &ql, -1);
            if (ql && ql->ql_Checked) {
                num++;
            }
	} while (gtk_tree_model_iter_next(model, &iter));
    }
    gchar label[32];
    sprintf(label, "%d", num);
    g_object_set(G_OBJECT(count), "label", label, NULL);
}

static void askChoiceOnRowChanged(GtkTreeModel *model, GtkTreePath *path,
	GtkTreeIter *iter, GtkWidget *count)
{
    askChoiceCountSelected(model, count);
}

static void askChoiceOnRowActivated(GtkTreeView *tree,
	GtkTreePath *path, GtkTreeViewColumn *column, GtkWidget *button)
{
    g_signal_emit_by_name(button, "activate");
}

#define MAXLEN 1000
#define MAX_FIELDS 63

bool AskChoice(struct Project *pr, const char *title, const char *oktext,
        struct Element *labels, struct Element *titles, bool *selected, 
	bool multiselect, bool stringmode, long *index, char **contents)
{
    GuiBusyCursor busy;

    long num = CountListLength(labels);

    if (titles != 0 && titles->el_Type == TYPE_NIL) {
        titles = 0;
    }

    if (multiselect && stringmode) {
        return false;			// illegal combination
    }
    if (num == 0) {
        return false;			// nothing to select (stringmode?)
    }

    if (!title) {
	if (multiselect) {
	    title = GCS(MSG_MultiSelectItems);
	} else if (stringmode) {
	    title = GCS(MSG_EnterString);
	} else {
	    title = GCS(MSG_SelectItem);
	}
    }
    if (!oktext) {
        oktext = GCS(MSG_Ok);
    }

    GuiObject tree = SFWViewObject,
        "rules-hint", true,
        "show-titles", titles != 0,
        "checkboxes", multiselect,
	 End;
    gtk_sfw_view_set_titles_body(GTK_SFW_VIEW(tree), titles, labels);
    if (multiselect && selected) {
       gtk_sfw_view_set_selected(GTK_SFW_VIEW(tree), selected);
    }

    GtkWidget *group, *scrolled;
    GtkWidget *root = GroupAlignment(false),
	"child", group = VGroup,
	    "child", LabelObject,
	        "label", title,
		End, "weight", 0,
	    "child", scrolled = ScrolledWindowObject,
		"hscrollbar-policy", GTK_POLICY_AUTOMATIC,
	        "shadow-type", GTK_SHADOW_IN,
		"child", tree,
		End,
	    End,
	End;

    if (multiselect) {
        GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
	GtkWidget *count = 0;
	g_object_set(G_OBJECT(group),
		"child", HGroup,
		    "child", LabelObject, 
		        "label", GCS(MSG_ItemsSelected), 
			End, "weight", 0,
		    "child", FrameObject,
		        "shadow-type", GTK_SHADOW_IN,
			"child", count = LabelObject, 
			    "label", "0", 
			    "xalign", 0.0,
			    "ypad", 1,
			    "xpad", 2,
			    End,
			End,
		    End, "weight", 0,
	  	NULL);
	g_signal_connect(model, "row-changed", 
                G_CALLBACK(askChoiceOnRowChanged), count);
	askChoiceCountSelected(model, count);
    } else {
	gint active = (index && *index >= 0)? *index + 1 : 0;

        gtk_sfw_view_set_active_line(GTK_SFW_VIEW(tree), active); 
    }

    GtkWidget *entry = 0;
    if (stringmode) {
	entry = ExtEntryObject,
	    "text", contents? EmptyNullStr(*contents) : "",
	    "max-length", MAXLEN,
	    "activates-default", true,
	    "attached-tree-view", tree,
	    End;
	g_object_set(G_OBJECT(group), "child", entry, "weight", 0, NULL);
	gtk_widget_grab_focus(entry);

        GtkTreeSelection *selection = 
            gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
        g_signal_connect(selection, "changed",
                G_CALLBACK(askChoiceOnSelectionChanged), entry);
    }

    title = ProjectTitle(pr);
    if (!title) {
	title = GCS(MSG_APP);
    }
    GtkWidget *dialog = DialogObject, 
     	"title", title,
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	End;
    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    GtkWidget *cancel = GuiSimpleButton(GCS(MSG_Cancel));
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(cancel);

    GtkWidget *ok = GuiSimpleButton(oktext, true);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(ok);
    gtk_window_set_default(GTK_WINDOW(dialog), ok);
    if (!multiselect) {
	g_signal_connect(tree, "row-activated", 
		G_CALLBACK(askChoiceOnRowActivated), ok);
    }

    GuiSetDialogButtonOrder(dialog);
    GuiAddWindowCallbacks(dialog, cancel);

    GtkRequisition treeReq, scrolledReq;
#if GTK_MAJOR_VERSION == 2
    gtk_widget_size_request(tree, &treeReq);
    gtk_widget_size_request(scrolled, &scrolledReq);
#else
    gtk_widget_get_preferred_size(tree, &treeReq, NULL);
    gtk_widget_get_preferred_size(scrolled, &scrolledReq, NULL);
#endif
    gint addWidth = treeReq.width - scrolledReq.width;
    gint addHeight = treeReq.height - scrolledReq.height;
    addWidth = MIN(addWidth, gdk_screen_width()/2);
    addHeight = MIN(addHeight, gdk_screen_height()/2);
    addHeight = MAX(addHeight, 100);
    gint width, height;
    gtk_window_get_size(GTK_WINDOW(dialog), &width, &height);
    width += addWidth;
    height += addHeight;
    gtk_window_set_default_size(GTK_WINDOW(dialog), width, height);

    bool rc = false;

    while (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
	if (multiselect) {
	   if (selected) {
               gtk_sfw_view_get_selected(GTK_SFW_VIEW(tree), selected);
	   }
	   rc = true;
	    break;
	} else if (stringmode) {
	    if (contents) {
		static char buffer[MAXLEN+1];
		const gchar *text = gtk_entry_get_text(GTK_ENTRY(entry));
		CodesetStrNBytesCopy(buffer, text, MAXLEN);
		*contents = buffer;
	    }
	    rc = true;
	    break;
	} else {
            guint active = gtk_sfw_view_get_active_line(GTK_SFW_VIEW(tree));

	    if (!active) {
		GuiBeep();
		continue;
	    }

	    if (index) {
                *index = active - 1;
            }
	    rc = true;
	    break;
	}
    }

    gtk_widget_destroy(dialog);

    return rc;
}

bool AskMulti(struct Project *pr, const char *title, const char *oktext,
	struct AskMultiInfo *ami)
{
    GuiBusyCursor busy;

    if (!ami || ami->ami_Num == 0) {
	return FALSE;
    }

    if (!title) {
        title = GCS(MSG_EditForm);
    }
    if (!oktext) {
        oktext = GCS(MSG_Ok);
    }

    struct MemPool *memPool = ami->ami_MemPool;
    GtkGroup *colGroup = 0;
    GuiObject group = VGroup, End;
    GuiObject top = GroupAlignment(false),
	"child", group,
	End;

    struct AskMultiItem *a = ami->ami_Items;
    for (int i=0; i < ami->ami_Num; i++, a++) {
        // char sc = Shortcut(a->ami_Title);
        const char *bh = a->ami_BubbleHelp;
        GuiObject go = 0;
	int weight = 0;

        switch(a->ami_Type) {
            case TYPE_STRING:
	    {
		const char *text = a->ami_Contents.string.string;
                a->ami_GuiObject = go = EntryObject, 
		    "text", EmptyNullStr(text),
		    "visibility", !a->ami_Contents.string.secret,
		    End;
                break;
	    }

            case TYPE_MEMO:
	    {
                go = GtkNewEditor(pr, TRUE, FALSE, FALSE, &a->ami_GuiObject);
                if (a->ami_Contents.mi) {
                    GuiSetEditorText(a->ami_GuiObject, a->ami_Contents.mi);
                }
		weight = 1;
                break;
	    }

            case TYPE_INTEGER:
                a->ami_GuiObject = go = EntryObject,
	            "text", a->ami_Nil? "" : IntToStr(a->ami_Contents.integer),
                    End;
                break;

            case TYPE_REAL:
                a->ami_GuiObject = go = EntryObject,
	            "text", a->ami_Nil? "" :
			DoubleToStr(a->ami_Contents.real, -1,
			     	(short)GetPrefsValue(pr, PREFS_REALFORMAT)),
		    End;
                break;

            case TYPE_DATE:
                a->ami_GuiObject = go = EntryObject,
	            "text", a->ami_Nil? "" :
                        DateToString(a->ami_Contents.date,
                            (short)GetPrefsValue(pr, PREFS_DATEFORMAT)),
                    End;
                break;

            case TYPE_TIME:
                a->ami_GuiObject = go = EntryObject,
	            "text", a->ami_Nil? "" : 
			TimeToString(a->ami_Contents.time, FORMAT_HOURMINSEC),
                    End;
                break;

            case TYPE_BOOL:
		go = AlignmentObject,
		    "xalign", 0.0,
		    "xscale", 0.0,
		    "child", a->ami_GuiObject = CheckButtonObject, 
			"active", !a->ami_Nil,
			End,
		    End;
                break;

            case TYPE_CHOICE:
	    {
                if (!a->ami_Contents.choice.ischoicelist) {
		    a->ami_GuiObject = go = GuiComboBox(
			    a->ami_Contents.choice.labels,
			    (a->ami_Nil? 0 : a->ami_Contents.choice.active));
		} else {
		    GtkListStore *store = gtk_list_store_new(1, G_TYPE_STRING);
		    GuiSetListStoreLabels(store, a->ami_Contents.choice.labels);
		    GtkWidget *tree;
		    go = ScrolledWindowObject,
			"hscrollbar-policy", GTK_POLICY_NEVER,
			"shadow-type", GTK_SHADOW_IN,
			"child", a->ami_GuiObject = tree = TreeViewObject,
			    "model", store,
			    "headers-visible", false,
			    End,
			End;
		    g_object_unref(G_OBJECT(store));

		    GtkCellRenderer *cell = gtk_cell_renderer_text_new();
		    GtkTreeViewColumn *column = 
			gtk_tree_view_column_new_with_attributes(
				"Item", cell, "text", 0, NULL);
		    gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

		    GtkTreeSelection *selection = 
			gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
		    gtk_tree_selection_set_mode(selection, 
			    GTK_SELECTION_SINGLE);
		    if (!a->ami_Nil) {
			GuiListSetActive(tree, a->ami_Contents.choice.active);
		    }
		    weight = 1;
                }
                break;
	    }

            case TYPE_OPTIONS:
	    {
                a->ami_GuiObject = GuiStringTreeView(true);
                GtkTreeModel *model = gtk_tree_view_get_model(
                        GTK_TREE_VIEW(a->ami_GuiObject));
                GtkListStore *store = GTK_LIST_STORE(model);
		char **labels = a->ami_Contents.options.labels;
		for (int i=0; labels[i]; i++) {
		    GtkTreeIter iter;
		    gtk_list_store_append(store, &iter);
		    gtk_list_store_set(store, &iter, 
			    0, labels[i], 
			    1, a->ami_Contents.options.selected[i],
			    -1);
                }

		go = ScrolledWindowObject,
		    "hscrollbar-policy", GTK_POLICY_NEVER,
	    	    "shadow-type", GTK_SHADOW_IN,
    		    "child", a->ami_GuiObject,
		    End;

		weight = 1;
		break;
	    }

            case NOTYPE:                /* title only objects */
	    default:
	    {
		go = FrameObject,
		    "shadow-type", GTK_SHADOW_IN,
		    "child", a->ami_GuiObject = LabelObject,
			"label", EmptyNullStr(a->ami_Title),
			"ypad", 1,
			"xpad", 2,
			"xalign", 0.5,
			End,
		    End;
                break;
	    }
        }

        if (!go) {
	    gtk_widget_destroy(top);
            return FALSE;
        }

	go = GuiMakeTooltip(go, bh);

        if (a->ami_Title && !GTK_IS_LABEL(a->ami_GuiObject)) {
            if (!colGroup) {
		GtkWidget *w = VGroup, "columns", 2, End;
                colGroup = GTK_GROUP(w);
		g_object_set(G_OBJECT(group), "child", w, NULL);
            }
	    GuiObject title = LabelObject, 
	        "label", a->ami_Title,
		"use-underline", TRUE,
		"xalign", 1.0,
		"yalign", 0.5,
		End;
	    g_object_set(G_OBJECT(title), "mnemonic-widget", a->ami_GuiObject, 
                    NULL);

	    g_object_set(G_OBJECT(colGroup),
		    "child", title, "weight", 0,
		    "child", go, "horiz-weight", 1, "vert-weight", weight, 
		    NULL);
        } else {
	    BuildGuiFinishColGroup(&colGroup);
	    g_object_set(G_OBJECT(group), 
		    "child", go, "horiz-weight", 1, "vert-weight", weight,
		   NULL); 
        }
    }
    BuildGuiFinishColGroup(&colGroup);

    GtkWidget *dialog = DialogObject, 
     	"title", title,
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	End;
    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), top, TRUE, TRUE, 0);
    gtk_widget_show_all(top);

    GtkWidget *cancel = GuiSimpleButton(GCS(MSG_Cancel));
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(cancel);

    GtkWidget *ok = GuiSimpleButton(oktext, true);
    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(ok);
    gtk_window_set_default(GTK_WINDOW(dialog), ok);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, cancel);

    if (gtk_group_get_total_vert_weight(GTK_GROUP(group)) > 0) {
	gint width, height;
	gtk_window_get_size(GTK_WINDOW(dialog), &width, &height);
	height += 50;
	gtk_window_set_default_size(GTK_WINDOW(dialog), width, height);
    }

    a = ami->ami_Items;
    for (int i=0; i < ami->ami_Num; i++, a++) {
	if (GTK_IS_ENTRY(a->ami_GuiObject)) {
	    GuiAddEntryCallbacks(a->ami_GuiObject, ok);
	}
    }

    bool rc = false;

    while (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
	if (CheckAskMultiInfo(ami, dialog)) {
	    rc = true;
	    break;
	}
    }

    a = ami->ami_Items;
    for (int i=0; i < ami->ami_Num && rc; i++, a++) {
	const gchar *cont;

	switch (a->ami_Type) {
	    case TYPE_STRING:
		FreeString(memPool, a->ami_Contents.string.string);
		a->ami_Nil = FALSE;
                cont = GuiStringContents(a->ami_GuiObject);
	    	a->ami_Contents.string.string = AllocString(memPool, cont);
		if (!a->ami_Contents.string.string) {
	       	    rc = FALSE;
		}
	   	break;

	    case TYPE_MEMO:
	     	FreeString(memPool, a->ami_Contents.mi);
                a->ami_Nil = FALSE;
                a->ami_Contents.mi = GuiEditorText(a->ami_GuiObject, memPool);
                if (!a->ami_Contents.mi) {
                    rc = false;
                }
                break;

            case TYPE_INTEGER:
                cont = GuiStringContents(a->ami_GuiObject);
                a->ami_Nil = !StrToInt(cont, &a->ami_Contents.integer);
                break;

            case TYPE_REAL:
                cont = GuiStringContents(a->ami_GuiObject);
                a->ami_Nil = !StrToDouble(cont, &a->ami_Contents.real);
                break;

            case TYPE_DATE:
                cont = GuiStringContents(a->ami_GuiObject);
                a->ami_Nil = !StringToDate(cont, &a->ami_Contents.date);
                break;
                
            case TYPE_TIME:
                cont = GuiStringContents(a->ami_GuiObject);
                a->ami_Nil = !StringToTime(cont, 
			FORMAT_HOURMINSEC, &a->ami_Contents.time);
                break;
                
            case TYPE_BOOL:
                a->ami_Nil = !gtk_toggle_button_get_active(
                        GTK_TOGGLE_BUTTON(a->ami_GuiObject));
                break;
                
            case TYPE_CHOICE:
                if(!a->ami_Contents.choice.ischoicelist) {
                    a->ami_Nil = FALSE;
                    a->ami_Contents.choice.active = 
                        gtk_combo_box_get_active(
                                GTK_COMBO_BOX(a->ami_GuiObject));
                } else {
                    a->ami_Contents.choice.active = 
                        GuiListActive(a->ami_GuiObject);
                    a->ami_Nil = (a->ami_Contents.choice.active < 0);
                }
                break;
                
            case TYPE_OPTIONS: 
            {
                bool *selected = a->ami_Contents.options.selected;
                GtkTreeModel *model = gtk_tree_view_get_model(
                        GTK_TREE_VIEW(a->ami_GuiObject));
                GtkTreeIter iter;
                if (gtk_tree_model_get_iter_first(model, &iter)) {
                    do {
                        struct QueryLine *ql = NULL;
                        gtk_tree_model_get(model, &iter, 0, &ql, -1);
                        *selected++ = (ql && ql->ql_Checked);
                    } while (gtk_tree_model_iter_next(model, &iter));
                }
                a->ami_Nil = FALSE;
                break;
            }
        }
    }

    gtk_widget_destroy(dialog);

    return rc;
}

#define USLEEP 50000

void Message(const char *text)
{
    if (!text) {
	return;
    }

    GuiBusyCursor busy;

    GuiObject progress = ProgressBarObject,
#if GTK_MAJOR_VERSION == 2
        "orientation", GTK_PROGRESS_LEFT_TO_RIGHT,
#else
        "orientation", GTK_ORIENTATION_HORIZONTAL,
#endif
	"fraction", 0.0,
	End;
    GuiObject msg = LabelObject,
        "label", text,
	"xalign", 0.5,
	End;

    GuiObject root = VGroup,
        "child", GroupAlignment(false),
            "child", VGroup,
                "child", FrameObject,
                    "shadow-type", GTK_SHADOW_IN,
                    "child", HGroup,
                        "border-width", 10,
                        "child", ImageObject,
                            "icon-name", "dialog-warning",
                            "icon-size", GTK_ICON_SIZE_DIALOG,
                            End,
                        "child", msg,
                        End,
                    End,
                "child", progress, "vert-weight", 0,
                End,
            End, "weight", 0,
        End;

    GuiObject win = WindowObject,
        "title", GCS(MSG_BeeBase),
	"modal", true,
	"skip-pager-hint", true,
	"skip-taskbar-hint", true,
	"type-hint", GDK_WINDOW_TYPE_HINT_DIALOG,
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
        "child", root,
	End;
    GuiIgnoreWindowDelete(win);
    gtk_widget_show_all(root);

    if (GuiOpenWindow(win)) {
	long max = (10 + strlen(text) / 5) * (1000000 / USLEEP);

        for (long i=0; i < max; i++) {
	    g_object_set(G_OBJECT(progress), "fraction", (gfloat)i / max, NULL);
	    while (gtk_events_pending()) {
		gtk_main_iteration();
	    }
 	    g_usleep(USLEEP);
        }
    }
    gtk_widget_destroy(win);
}

