//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkSearch.cc
//
// Creation date: 03.01.2005
// Last modified: 17.01.2009
//

#include "search.h"
#include "gtk/classes/gtkgroup.h"
#include "gtk/classes/gtkradiobox.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "table.h"
#include "util.h"
#include "appstrings_cat.h"

struct EditSearchForData
{
    Project *d_Project;
    struct SearchFor *d_SearchFor;
    GuiObject d_Pattern;
    GuiObject d_Sensitive;
    GuiObject d_All;
    GuiObject d_Direction;
    GuiObject d_Start;
    GuiObject d_Ok;
    GuiObject d_Cancel;
};

static void finish(GuiObject go, gint response, EditSearchForData *d)
{
    if (response == GTK_RESPONSE_ACCEPT) {
	struct SearchFor *sf = d->d_SearchFor;

	const gchar *str = gtkEntryText(d->d_Pattern);
	size_t maxlen = sizeof(sf->sf_Pattern) - 1;
	CodesetStrNBytesCopy(sf->sf_Pattern, str, maxlen);
	sf->sf_CaseSensitive = gtkToggleButtonActive(d->d_Sensitive);
	sf->sf_AllFields = gtkToggleButtonActive(d->d_All);
	sf->sf_Forward = gtkRadioBoxActive(d->d_Direction) == 0;
	sf->sf_FirstLastRecord = gtkRadioBoxActive(d->d_Start) == 0;

	gtk_widget_hide(go);
    }
    // GuiSetProjectSleep(d->d_Project, false);
}

GuiObject createEditSearchFor(struct SearchFor *sf)
{
    if (!sf || !sf->sf_Table) {
	return 0;
    }

    Project *pr = TableProject(sf->sf_Table);
    EditSearchForData *d = g_new0(EditSearchForData, 1);

    d->d_Project = pr;
    d->d_SearchFor = sf;
    
    d->d_Pattern = EntryObject,
        "max-length", sizeof(sf->sf_Pattern) - 1,
	"xalign", 0.5,
        "text", EmptyNullStr(sf->sf_Pattern),
	"activates-default", true,
	End;
    GuiSetTooltip(d->d_Pattern, GCS(MSG_SearchPatternBH));
    GuiObject patternLabel = LabelObject,
        "label", GCS(MSG_SearchPattern),
	"use-underline", true,
	"mnemonic-widget", d->d_Pattern,
	End;

    d->d_Sensitive = CheckButtonObject, 
        "label", GCS(MSG_CaseSensitive),
	"use-underline", true,
	"active", sf->sf_CaseSensitive,
        End;
    GuiSetTooltip(d->d_Sensitive, GCS(MSG_CaseSensitiveBH));

    d->d_All = CheckButtonObject,
        "label", GCS(MSG_AllFields),
	"use-underline", true,
	"active", (sf->sf_Field == 0 || sf->sf_AllFields),
	"sensitive", sf->sf_Field != 0,
	End;
    GuiSetTooltip(d->d_All, GCS(MSG_AllFieldsBH));

    static char *dirstr[3];
    dirstr[0] = GCS(MSG_Forward);
    dirstr[1] = GCS(MSG_Backward);
    dirstr[2] = NULL;
    d->d_Direction = RadioBoxObject,
        "entries", dirstr, 
	"active", sf->sf_Forward? 0 : 1,
	End;

    static char *startstr[3];
    startstr[0] = GCS(MSG_FirstLastRecord);
    startstr[1] = GCS(MSG_CurrentRecord);
    startstr[2] = NULL;
    d->d_Start = RadioBoxObject,
        "entries", startstr,
	"active", sf->sf_FirstLastRecord? 0 : 1,
	End;

    GuiObject group = VGroup,
        "child", HGroup,
	    "child", patternLabel, "weight", 0,
	    "child", d->d_Pattern,
	    End, "vert-weight", 0,
	"child", HGroup,
	    "child", d->d_Sensitive,
	    "child", d->d_All,
	    End, "vert-weight", 0,
	"child", HGroup,
	    "same-width", true,
	    "child", Frame(GCS(MSG_SearchDirection)),
	        "child", GroupAlignment(true),
		    "child", HCenter(d->d_Direction),
		    End,
		End,
	    "child", Frame(GCS(MSG_StartWith)),
	        "child", GroupAlignment(true),
		    "child", HCenter(d->d_Start),
		    End,
		End,
	    End, "vert-weight", 0,
	End;

    GuiObject root = GroupAlignment(false),
        "child", group,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Find), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_Find),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", true,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Search dialog");
    g_object_set_data_full(G_OBJECT(dialog), "edit-data", d, g_free);

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(finish), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    //GuiSetProjectSleep(pr, true);
    if (!GuiOpenWindow(dialog)) {
        GuiSetProjectSleep(pr, false);
	gtk_widget_destroy(dialog);
        dialog = 0;
    }

    return dialog;
}

struct SearchFor *GuiSearchFor(struct SearchFor *sf)
{
    GuiBusyCursor busy;

    GuiObject dialog = createEditSearchFor(sf);
    if (dialog) {
	int response = gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
	if (response == GTK_RESPONSE_ACCEPT) {
	    return sf;
	}
    }
    return 0;
}

