//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkSetFormats.cc
//
// Creation date: 12.12.2004
// Last modified: 17.01.2009
//

#include "setformats.h"
#include "gtk/gtkUtil.h"
#include "gtk/gtkimpl.h"
#include "gtk/classes/gtkradiobox.h"
#include "main.h"
#include "gui.h"
#include "project.h"
#include "preferences.h"
#include "memory.h"
#include "appstrings_cat.h"

struct SetFormatsData
{
    Project *d_Project;
    GuiObject d_RealFormat, d_DateFormat;
    GuiObject d_Ok, d_Cancel;
};

static void disposeSetFormats(GuiObject obj, SetFormatsData *d)
{
    PoolFreeMem(ProjectMemPool(d->d_Project), d, sizeof(struct SetFormatsData));
    gtk_widget_destroy(obj);
}

static void finish(GuiObject obj, gint response, SetFormatsData *d)
{
    if (response == GTK_RESPONSE_ACCEPT) {
	Project *pr = d->d_Project;

        SetPrefsValue(pr, PREFS_REALFORMAT, 
		gtk_radio_box_get_active(GTK_RADIO_BOX(d->d_RealFormat)));
        SetPrefsValue(pr, PREFS_DATEFORMAT, 
		gtk_radio_box_get_active(GTK_RADIO_BOX(d->d_DateFormat)));

        FORNODE(pr, &ProjectList) {
            GuiUpdate(pr);
	}
    }
    disposeSetFormats(obj, d);
    GuiSetSleep(FALSE);
}

static GuiObject createSetFormats(Project *pr)
{
    MemPool *memPool = ProjectMemPool(pr);
    SetFormatsData *d;
    static char *realformats[3], *dateformats[4];

    d = (SetFormatsData *)PoolAllocMemZero(memPool, sizeof(SetFormatsData));
    if (!d) {
	return 0;
    }
    d->d_Project = pr;

    realformats[0] = GCS(MSG_Formats_RealFormat_DecimalPoint);
    realformats[1] = GCS(MSG_Formats_RealFormat_DecimalComma);
    realformats[2] = NULL;

    d->d_RealFormat = RadioBoxObject,
        "entries", realformats,
        "active", GetPrefsValue(pr, PREFS_REALFORMAT),
        End;

    GuiObject realgroup = Frame(GCS(MSG_Formats_RealFormat)),
	"child", GroupAlignment(true),
	    "xscale", 0.0,
	    "xalign", 0.5,
	    "child", d->d_RealFormat,
	    End,
	End;

    dateformats[0] = GCS(MSG_Formats_DateFormat_DayMonthYear);
    dateformats[1] = GCS(MSG_Formats_DateFormat_MonthDayYear);
    dateformats[2] = GCS(MSG_Formats_DateFormat_YearMonthDay);
    dateformats[3] = NULL;

    d->d_DateFormat = RadioBoxObject,
        "entries", dateformats,
        "active", GetPrefsValue(pr, PREFS_DATEFORMAT),
        End;

    GuiObject dategroup = Frame(GCS(MSG_Formats_DateFormat)),
	"child", GroupAlignment(true),
	    "xscale", 0.0,
	    "xalign", 0.5,
	    "child", d->d_DateFormat,
	    End,
	End;

    GuiObject group;
    GuiObject root = GroupAlignment(false),
        "child", group = VGroup,
	    "child", realgroup, "vert-weight", 0,
	    "child", dategroup, "vert-weight", 0,
	    End,
	End;

    d->d_Ok = GuiSimpleButton(GCS(MSG_Ok), true);
    d->d_Cancel = GuiSimpleButton(GCS(MSG_Cancel));

    GuiObject dialog = DialogObject,
        "title", GCS(MSG_Formats),
	"window-position", GTK_WIN_POS_CENTER,
        "transient-for", GuiFindTransientWindow(NULL),
	"modal", true,
	End;
    GuiSetWindowProject(dialog, pr);
    GuiSetHelpNode(dialog, "Formats");

    GtkWidget *vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
    gtk_box_pack_start(GTK_BOX(vbox), root, TRUE, TRUE, 0);
    gtk_widget_show_all(root);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Cancel, GTK_RESPONSE_CANCEL);
    gtk_widget_show(d->d_Cancel);

    gtk_dialog_add_action_widget(GTK_DIALOG(dialog), 
	    d->d_Ok, GTK_RESPONSE_ACCEPT);
    gtk_widget_show(d->d_Ok);
    gtk_window_set_default(GTK_WINDOW(dialog), d->d_Ok);

    g_signal_connect(dialog, "response", G_CALLBACK(finish), d);

    GuiSetDialogButtonOrder(dialog);
    GuiSetDialogGeometryHints(dialog, group);
    GuiAddWindowCallbacks(dialog, d->d_Cancel);

    GuiSetSleep(TRUE);
    if (!GuiOpenWindow(dialog)) {
        GuiSetSleep(false);
        disposeSetFormats(dialog, d);
        dialog = 0;
    }

    return dialog;
}

void GuiSetFormats(Project *pr)
{
    createSetFormats(pr);
}

