//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkStructureEditor.cc
//
// Creation date: 03.01.2005
// Last modified: 05.06.2022
//

#include "system.h"
#include "structureeditor.h"
#include "gtk/gtkMenus.h"
#include "gtk/gtkQuickSearch.h"
#include "gtk/classes/gtkpopwidget.h"
#include "gtk/gtkimpl.h"
#include "gtk/gtkUtil.h"
#include "lang/program.h"
#include "buildgui.h"
#include "preferences.h"
#include "name.h"
#include "codeset.h"
#include "util.h"

//
// UserFunctions Popup
//

static void userFunctionsCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    UserFunction *uf = 0;
    gtk_tree_model_get(model, iter, 0, &uf, -1);
    const char *name = UserFunctionName(uf);
    g_object_set(G_OBJECT(cell), "text", EmptyNullStr(name), NULL);
}

static void userFunctionsOnRowActivated(GtkTreeView *tree,
	GtkTreePath *path, GtkTreeViewColumn *column, GtkPopString *popup)
{
    gtk_pop_string_close_popup(popup, true);
}

static void userFunctionSearchFunc(GuiObject go,
        const char *str, int cmd, void *pop)
{
    long dir = -1;

    switch (cmd) {
	case QUICK_SEARCH_NEXT:  dir++;		/* no break */
	case QUICK_SEARCH_START: dir++;		/* no break */
	case QUICK_SEARCH_PREV:
            GuiUserFunctionsQuickSearch(go, dir, str);
            break;

	case QUICK_SEARCH_COMMIT:
            gtk_pop_string_close_popup(GTK_POP_STRING(pop), true);
            break;
    }
}

static gboolean userFunctionsOnKeyPressEvent(GtkWidget *w, GdkEventKey *event,
        GtkPopString *pop)
{
    if (event->state & GDK_CONTROL_MASK) {
	return false;
    }
    return GuiQuickSearchInit(w, event, userFunctionSearchFunc, pop);
}

static gint userFunctionsSortFunc(GtkTreeModel *model, GtkTreeIter *iter1,
      	GtkTreeIter *iter2, gpointer user_data)
{
    UserFunction *uf1 = 0, *uf2 = 0;

    gtk_tree_model_get(model, iter1, 0, &uf1, -1);
    gtk_tree_model_get(model, iter2, 0, &uf2, -1);

    const char *name1 = EmptyNullStr(UserFunctionName(uf1)); 
    const char *name2 = EmptyNullStr(UserFunctionName(uf2));
    
    return CodesetStrCollate(name1, name2, CS_COLLATE_ORDER);
}

static gboolean userFunctionsOnStart(GtkPopString *pop, GuiObject functions)
{
    g_signal_stop_emission_by_name(pop, "start");

    GtkTreeView *tree = GTK_TREE_VIEW(functions);
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    GtkListStore *store = GTK_LIST_STORE(model);

    gtk_list_store_clear(store);

    Project *pr = (Project *)g_object_get_data(G_OBJECT(functions), "project");
    bool sort = GetPrefsValue(pr, PREFS_PROGRAMSORTTRIGGERFUNCTIONS);
    if (sort) {
	gtk_tree_sortable_set_default_sort_func(GTK_TREE_SORTABLE(model),
		userFunctionsSortFunc, 0, 0);
	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(model),
		GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID, GTK_SORT_ASCENDING);
    }

    GuiObject entry = gtk_pop_string_get_string(pop);
    const gchar *current = 0;
    if (GTK_IS_ENTRY(entry)) {
	current = gtkEntryText(entry);
    }

    Program *prg = ProjectProgram(pr);
    List *uflist = ProgramUserFunctions(prg);
    if (uflist) {
	UserFunction *uf, *active = 0;
        FORNODE (uf, uflist) {
            if (!UserFunctionTrigger(uf)) {
                continue;
	    }

	    GtkTreeIter iter;
	    gtk_list_store_append(store, &iter);
	    gtk_list_store_set(store, &iter, 0, uf, -1);

            const char *name = UserFunctionName(uf);
            if (name && current && strcmp(name, current) == 0) {
                active = uf;
	    }
        }
	GuiListSetActiveNode(functions, active);
    }

    GtkWidget *window = gtk_pop_string_get_popup(pop);
    gint width, height;
    gtk_window_get_size(GTK_WINDOW(window), &width, &height);
    height += 100;
    gtk_window_resize(GTK_WINDOW(window), width, height);

    return true;
}

static void userFunctionsOnStop(GtkPopString *pop, GuiObject functions)
{
    UserFunction *uf = (UserFunction *)GuiListActiveEntry(functions);
    const char *name = UserFunctionName(uf);
    GuiObject entry = gtk_pop_string_get_string(pop);
    if (GTK_IS_ENTRY(entry)) {
	gtk_entry_set_text(GTK_ENTRY(entry), EmptyNullStr(name));
    }
}

GuiObject UserFunctionsPopup(GuiObject string, Project *pr)
{
    GuiObject functions = GuiPointerTreeView(userFunctionsCellDataFunc);
    g_object_set_data(G_OBJECT(functions), "project", pr);

    GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(functions));

    GuiObject pop = PopWidgetObject,
	"string", string,
	"button", ButtonObject,
	    "child", ImageObject,
		"icon-name", "gtk-index", 
		"icon-size", GTK_ICON_SIZE_MENU,
		End,
	    End,
	"widget", ScrolledWindowObject,
	    "hscrollbar-policy", GTK_POLICY_AUTOMATIC,
	    "shadow-type", GTK_SHADOW_IN,
	    "child", functions,
	    End,
	End;

    g_signal_connect(functions, "row-activated",
    	    G_CALLBACK(userFunctionsOnRowActivated), pop);
    g_signal_connect(functions, "key-press-event",
            G_CALLBACK(userFunctionsOnKeyPressEvent), pop);
    g_signal_connect(pop, "start", G_CALLBACK(userFunctionsOnStart), functions);
    g_signal_connect(pop, "stop", G_CALLBACK(userFunctionsOnStop), functions);
    g_signal_connect_data(pop, "close", G_CALLBACK(gtk_list_store_clear),
	    model, 0, (GConnectFlags)(G_CONNECT_AFTER | G_CONNECT_SWAPPED));
    
    return pop;
}

#ifdef ENABLE_STRUCTURE_EDITOR

#include "gtk/gtkStructureEditor.h"
#include "gtk/gtkEditTable.h"
#include "gtk/gtkEditField.h"
#include "gtk/gtkEditDisplay.h"
#include "gtk/gtkEditDisplayTable.h"
#include "gtk/gtkEditDisplayField.h"
#include "gtk/classes/gtkgroup.h"
#include "requesters.h"
#include "error.h"
#include "appstrings_cat.h"

enum {
    TARGET_TABLE, TARGET_FIELD, TARGET_DISPLAY
};

struct StructureEditorData
{
    Project *d_Project;
    void *d_NameChangeObject;
    Table *d_CurrentTable;
    bool d_RootDisplay;
    short d_TableSortMethod;
    short d_FieldSortMethod;
    GuiObject d_TablesGroup;
    GuiObject d_Tables;
    GuiObject d_NewTable;
    GuiObject d_DeleteTable;
    GuiObject d_SortTables;
    GuiObject d_FieldsGroup;
    GuiObject d_Fields;
    GuiObject d_NewField;
    GuiObject d_CopyField;
    GuiObject d_DeleteField;
    GuiObject d_SortFields;
    GuiObject d_DisplayObjectsGroup;
    GuiObject d_DisplayObjects;
    GuiObject d_Add;
    GuiObject d_Rem;
    GuiObject d_Up;
    GuiObject d_Down;
    GuiObject d_In;
    GuiObject d_Out;
    GuiObject d_Text;
    GuiObject d_Image;
    GuiObject d_Space;
    GuiObject d_Balance;
    GuiObject d_Group;
    GuiObject d_Register;
    GuiObject d_Window;
    GuiObject d_DisplayType;
};

#define SE_DATA(se) \
    (StructureEditorData *)g_object_get_data(G_OBJECT(se), "edit-data");

#define RootDisplayObject(d) \
        (((d)->d_RootDisplay)? \
                ProjectRootDisplayObject((d)->d_Project) : \
                TableDisplayObject((d)->d_CurrentTable))

static void remClick(StructureEditorData *d);

void CloseStructureEditor(GuiObject se)
{
    if (se) {
        WindowDimensions wd;
	StructureEditorData *d = SE_DATA(se);

        if (GuiGetWindowDimensions(se, &wd)) {
            SetPrefsWindowDimensions(d->d_Project, PREFS_STRUCTUREEDITOR, &wd);
        }
        gtk_widget_destroy(se);
    }
}

static void fieldCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    static CatalogString typenames[] = {
	MSG_Err_Internal,
	MSG_Integer,
	MSG_Real,
	MSG_Boolean,
	MSG_Err_Internal,
	MSG_Memo,
	MSG_Date,
	MSG_Time,
	MSG_Err_Internal,
	MSG_Button,
	MSG_Choice,
	MSG_Reference,
	MSG_String,
	MSG_Virtual,
    };
    static char buffer[101];
    Field *field = 0;
    gtk_tree_model_get(model, iter, 0, &field, -1);
    const char *name = EmptyNullStr(FieldName(field));
    bool italic = !IsMBName(name);
    short type = FieldType(field);
    sprintf(buffer, "%s%s%s(%s)", italic? "<i>" : "", name,
            italic? "</i>" : "", GCS(typenames[type]));
    g_object_set(G_OBJECT(cell), italic? "markup" : "text", buffer, NULL);
}

static void displayCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    DisplayObject *dio = 0;
    gtk_tree_model_get(model, iter, 0, &dio, -1);
    const gchar *text = "";
    gchar *mem = 0;
    bool markup = false;

    switch (DisplayObjectType(dio)) {
	case DO_TYPE_TABLE:
        {
            struct Table *tab = DisplayObjectTable(dio);
            const char *name = TableName(tab);
            const char *title = DisplayObjectTitle(dio);

            if (title && title[0] && strcmp(name, title) != 0 && 
                    DisplayTableHasPanel(dio)) {
                gchar *tmp = gtkEscapeStr(title);
                text = mem = g_strdup_printf("<b>%s</b>(%s)", name, tmp);
                g_free(tmp);
            } else {
                text = mem = g_strdup_printf("<b>%s</b>", name);
            }
            markup = true;
            break;
        }

	case DO_TYPE_FIELD:
        {
            struct Field *field = DisplayFieldField(dio);
            const char *name = FieldName(field);
            const char *title = DisplayObjectTitle(dio);

            if (title && title[0] && strcmp(name, title) != 0) {
                gchar *tmp = gtkEscapeStr(title);
                text = mem = g_strdup_printf("<b>%s</b>(%s)", name, tmp);
                g_free(tmp);
            } else {
                text = mem = g_strdup_printf("<b>%s</b>", name);
            }
            markup = true;
            break;
        }

        case DO_TYPE_TEXT:
        {
            const char *label = GCS(MSG_Text);
            const char *title = DisplayObjectTitle(dio);

            if (title && title[0]) {
	        text = mem = g_strdup_printf("%s(%s)", label, title);
	    } else {
	        text = label;
	    }
            break;
        }

	case DO_TYPE_SPACE:
            text = GCS(MSG_Space);
            break;

	case DO_TYPE_BALANCE:
            text = GCS(MSG_StructureEditor_Balance);
            break;

	case DO_TYPE_IMAGE:
            text = GCS(MSG_Image);
            break;

	case DO_TYPE_REGISTER:
	    text = GCS(MSG_StructureEditor_Register);
	    break;

	case DO_TYPE_GROUP:
        {
/*	    bool border = DisplayGroupBorder(dio); */
	    bool horiz = DisplayGroupHoriz(dio);
	    const char *title = DisplayObjectTitle(dio);
	    const char *label = GCS(horiz?
		    MSG_StructureEditor_Display_HGroup 
		    : MSG_StructureEditor_Display_VGroup);

            if (title && title[0]) {
	        text = mem = g_strdup_printf("%s(%s)", label, title);
	    } else {
		text = label;
	    }
	    break;
        }

	case DO_TYPE_WINDOW:
        {
            const char *title = DisplayObjectTitle(dio);
	    const char *label = GCS(MSG_Window);

            if (title && title[0]) {
	        text = mem = g_strdup_printf("%s(%s)", label, title);
	    } else {
		text = label;
	    }
	    break;
        }

        default:
            break;
    }

#if 0
    /*
    ** clean buffer, e.g. replace new line characters by white space.
    */
    if (!mem) {
	mem = g_strdup(text);
	text = mem;
    }
    for (char *s = mem; *s; s++) {
        if (ISSPACE(*s)) {
            *s = ' ';
	}
    }
#endif
    g_object_set(G_OBJECT(cell), markup? "markup" : "text", text, NULL);
    g_free(mem);
}

static void setTables(StructureEditorData *d)
{
    Table *active = DisplayProjectActiveTable(d->d_Project);

    GuiListSetListWithFilter(d->d_Tables, ProjectTables(d->d_Project), 
            TableSystemFilter);
    GuiListSetActiveNode(d->d_Tables, active);
}

static void fieldsActiveChange(StructureEditorData *d)
{
    Field *field = (Field *)GuiListActiveEntry(d->d_Fields);
    gtk_widget_set_sensitive(d->d_CopyField, field != 0);
    gtk_widget_set_sensitive(d->d_DeleteField, field != 0);
    SetDisplayTableActiveField(d->d_CurrentTable, field);
}

static void setFields(StructureEditorData *d)
{
    List *fields = TableFields(d->d_CurrentTable);
    Field *active = DisplayTableActiveField(d->d_CurrentTable);

    gtk_widget_set_sensitive(d->d_FieldsGroup, fields != 0);
    GuiListSetList(d->d_Fields, fields);
    GuiListSetActiveNode(d->d_Fields, active);
    fieldsActiveChange(d);
}

static DisplayObject *displayActiveEntry(GuiObject go)
{
    DisplayObject *dio = 0;
    GtkTreeView *tree = GTK_TREE_VIEW(go);
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    GtkTreePath *path;
    gtk_tree_view_get_cursor(tree, &path, 0);
    if (path) {
	GtkTreeIter iter;

	if (gtk_tree_model_get_iter(model, &iter, path)) {
	    gtk_tree_model_get(model, &iter, 0, &dio, -1);
	}
	gtk_tree_path_free(path);
    }

    return dio;
}

static void displayActiveChange(StructureEditorData *d)
{
    DisplayObject *dio = displayActiveEntry(d->d_DisplayObjects);

    if (d->d_RootDisplay) {
        SetDisplayProjectActiveDisplayObject(d->d_Project, dio);
    } else {
        SetDisplayTableActiveDisplayObject(d->d_CurrentTable, dio);
    }
}

static void saveDisplayTreeOpenState(StructureEditorData *d)
{
    GtkTreeView *tree = GTK_TREE_VIEW(d->d_DisplayObjects);
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    GtkTreePath *path = gtk_tree_path_new_first();
    GtkTreeIter iter;
    bool ok = gtk_tree_model_get_iter_first(model, &iter);

    while (ok) {
	GtkTreeIter next;
	DisplayObject *dio = 0;
	gtk_tree_model_get(model, &iter, 0, &dio, -1);

	if (gtk_tree_model_iter_children(model, &next, &iter)) {
	    SetDisplayObjectOpen(dio, gtk_tree_view_row_expanded(tree, path));
	    gtk_tree_path_down(path);
	} else {
	    next = iter;
	    while (!gtk_tree_model_iter_next(model, &next)) {
		ok = gtk_tree_model_iter_parent(model, &next, &iter);
		if (!ok) {
		    break;
		}
		gtk_tree_path_up(path);
		iter = next;
	    }
	    gtk_tree_path_next(path);
	}
	iter = next;
    }

    gtk_tree_path_free(path);
}

static void displayRowChange(StructureEditorData *d)
{
    displayActiveChange(d);		// update active display object
    saveDisplayTreeOpenState(d);
}

static gboolean displayTestCollapseRow(GtkTreeView *tree, 
	GtkTreeIter *iter, GtkTreePath *path, gpointer unused)
{
    DisplayObject *dio = 0;
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    gtk_tree_model_get(model, iter, 0, &dio, -1);
    return (DisplayObjectType(dio) == DO_TYPE_TABLE);
}

static bool findDisplayObject(GtkTreeModel *model, DisplayObject *find, 
	GtkTreeIter *piter)
{
    GtkTreeIter iter;
    bool ok = gtk_tree_model_get_iter_first(model, &iter);

    while (ok) {
	GtkTreeIter next;
	DisplayObject *dio = 0;
	gtk_tree_model_get(model, &iter, 0, &dio, -1);
	if (find == dio) {
	    if (piter) {
		*piter = iter;
	    }
            return true;
	}
	if (!gtk_tree_model_iter_children(model, &next, &iter)) {
	    next = iter;
	    while (!gtk_tree_model_iter_next(model, &next)) {
		ok = gtk_tree_model_iter_parent(model, &next, &iter);
		if (!ok) {
		    break;
		}
		iter = next;
	    }
	}
	iter = next;
    }
    return false;
}

static gboolean setDisplayCursor(gpointer p)
{
    GtkTreeView *tree = GTK_TREE_VIEW(p);
    GtkTreeSelection *selection = gtk_tree_view_get_selection(tree);
    GList *list = gtk_tree_selection_get_selected_rows(selection, 0);
    bool set = false;

    for (GList *item = list; item; item = item->next) {
	GtkTreePath *path = (GtkTreePath *)item->data;
        if (path) {
            if (!set) {
                gtk_tree_view_set_cursor(tree, path, 0, false);
                gtk_tree_view_scroll_to_cell(tree, path, NULL, FALSE, .5, .0);
                set = true;
            }
            gtk_tree_path_free(path);
        }
    }
    g_list_free(list);

    return false;
}

static void setDisplayRec(GtkTreeView *tree, List *childs, GtkTreeIter *parent)
{
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    GtkTreeStore *store = GTK_TREE_STORE(model);
    DisplayObject *dio;
    GtkTreeIter iter;

    if (childs) {
	FORNODE (dio, childs) {
	    if (DisplayObjectVisible(dio)) {
		gtk_tree_store_append(store, &iter, parent);
		gtk_tree_store_set(store, &iter, 0, dio, -1);
		
		short type = DisplayObjectType(dio);
		bool open =  DisplayObjectOpen(dio);
		if (type != DO_TYPE_TABLE) {
		    setDisplayRec(tree, DisplayObjectObjects(dio), &iter);
		}
		
		GtkTreePath *path = gtk_tree_model_get_path(model, &iter);
		if (open) {
		    gtk_tree_view_expand_to_path(tree, path);
		} else {
		    gtk_tree_view_collapse_row(tree, path);
		}
		gtk_tree_path_free(path);
	    }
	}
    }
}

static void setDisplay(StructureEditorData *d, bool rememberCursor)
{
    struct Table *tab = d->d_CurrentTable;
    struct DisplayObject *dio = TableDisplayObject(tab);
    DisplayObject *activedio = 0;
    GtkTreeView *tree = GTK_TREE_VIEW(d->d_DisplayObjects);
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    GtkTreeStore *store = GTK_TREE_STORE(model);
    GtkTreeSelection *selection;
    GtkTreePath *path = 0;
    GtkTreeIter iter, *parent = 0;

    if (rememberCursor) {
	gtk_tree_view_get_cursor(tree, &path, 0);
    }
    gtk_tree_store_clear(store);
    gtk_tree_view_unset_rows_drag_dest(tree);

    if (d->d_RootDisplay) {
	gtk_widget_set_sensitive(d->d_DisplayObjectsGroup, true);
        activedio = DisplayProjectActiveDisplayObject(d->d_Project);
        dio = ProjectRootDisplayObject(d->d_Project);
	// drag & drop
	static GtkTargetEntry displayTargets[] = {
	    { (gchar *)"DisplayObject", GTK_TARGET_SAME_APP, TARGET_DISPLAY },
	    { (gchar *)"Table", GTK_TARGET_SAME_APP, TARGET_TABLE },
	};
	gtk_tree_view_enable_model_drag_dest(tree, displayTargets, 2,
		(GdkDragAction)(GDK_ACTION_COPY | GDK_ACTION_MOVE));
    } else if (!tab || !DisplayObjectVisible(dio)) {
	gtk_widget_set_sensitive(d->d_DisplayObjectsGroup, false);
        dio = 0;
    } else {
	gtk_widget_set_sensitive(d->d_DisplayObjectsGroup, true);
        activedio = DisplayTableActiveDisplayObject(tab);
	// drag & drop
	static GtkTargetEntry displayTargets[] = {
	    { (gchar *)"DisplayObject", GTK_TARGET_SAME_APP, TARGET_DISPLAY },
	    { (gchar *)"Field", GTK_TARGET_SAME_APP, TARGET_FIELD }
	};
	gtk_tree_view_enable_model_drag_dest(tree, displayTargets, 2,
		(GdkDragAction)(GDK_ACTION_COPY | GDK_ACTION_MOVE));
        gtk_tree_store_append(store, &iter, parent);
        gtk_tree_store_set(store, &iter, 0, dio, -1);
        parent = &iter;
    }

    setDisplayRec(tree, DisplayObjectObjects(dio), parent);

    if (path) {
	while (!gtk_tree_model_get_iter(model, &iter, path)) {
	    if (!gtk_tree_path_prev(path)) {
		if (gtk_tree_path_get_depth(path) == 1) {
		    rememberCursor = false;
  		    break;
		}  
		gtk_tree_path_up(path);
	    }
	}
	//gtk_tree_view_expand_to_path(tree, path);
        selection = gtk_tree_view_get_selection(tree);
        if (selection) {
            gtk_tree_selection_select_path(selection, path);
        }
        gtk_tree_path_free(path);
    }

    if (!rememberCursor && findDisplayObject(model, activedio, &iter)) {
	path = gtk_tree_model_get_path(model, &iter);
	if (path) {
	    //gtk_tree_view_expand_to_path(tree, path);
            selection = gtk_tree_view_get_selection(tree);
            if (selection) {
                gtk_tree_selection_select_path(selection, path);
            }
            gtk_tree_path_free(path);
	}
    }
    g_idle_add(setDisplayCursor, tree);

    saveDisplayTreeOpenState(d);
}

static void addDisplayObject(StructureEditorData *d, DisplayObject *dio)
{
    if (d->d_RootDisplay) {
        SetDisplayProjectActiveDisplayObject(d->d_Project, dio);
    } else {
        SetDisplayTableActiveDisplayObject(d->d_CurrentTable, dio);
    }
    setDisplay(d, false);
}

static void tableCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    Table *tab = 0;
    gtk_tree_model_get(model, iter, 0, &tab, -1);
    const char *name = EmptyNullStr(TableName(tab));
    if (IsMBName(name)) {
        g_object_set(G_OBJECT(cell), "text", name, NULL);
    } else {
        char *mem = g_strdup_printf("<i>%s</i>", name);
        g_object_set(G_OBJECT(cell), "markup", mem, NULL);
        g_free(mem);
    }
}

static void tablesActiveChange(struct StructureEditorData *d)
{
    Table *tab = (Table *)GuiListActiveEntry(d->d_Tables);
    if (!FindNodePtr(ProjectTables(d->d_Project), tab)) {
        tab = NULL;
    }
    gtk_widget_set_sensitive(d->d_DeleteTable, tab != 0);
    d->d_CurrentTable = tab;
    SetDisplayProjectActiveTable(d->d_Project, tab);
    setFields(d);
    if (!d->d_RootDisplay) {
        setDisplay(d, false);
    }
}

gboolean tablesOnDragDrop(GtkWidget *widget, GdkDragContext *context,
	gint x, gint y, guint time, StructureEditorData *d)
{
    gboolean success = false;
    GtkWidget *source = gtk_drag_get_source_widget(context);

    if (source == d->d_Tables) {
	List *tabs = ProjectTables(d->d_Project);
	Table *tab = (Table *)GuiListActiveEntry(d->d_Tables);
	if (tabs && tab) {
	    GtkTreeView *tree = GTK_TREE_VIEW(widget);
	    GtkTreePath *path = 0;
	    GtkTreeViewDropPosition drop;
	    Table *pred = 0;
	    gtk_tree_view_get_dest_row_at_pos(tree, x, y, &path, &drop);
	    if (path) {
		GtkTreeModel *model = gtk_tree_view_get_model(tree);
		GtkTreeIter iter;

		if (gtk_tree_model_get_iter(model, &iter, path)) {
		    gtk_tree_model_get(model, &iter, 0, &pred, -1);
		}
		gtk_tree_path_free(path);
		if (pred && drop == GTK_TREE_VIEW_DROP_BEFORE) {
	     	    pred = PRED(pred);
     		}
	    }
	    if (!pred) {
		pred = (Table *)TAILPRED(tabs);
	    }
	    success = true;
 	    if (tab != pred) {
                long error = ObtainProjectChangeLock(d->d_Project);
                if (error != OK) {
                    Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
                    success = false;
                } else {
                    REMOVE(tab);
                    INSERT(tabs, tab, pred);
                    ProjectChange(d->d_Project, TRUE);
                    SetDisplayProjectActiveTable(d->d_Project, tab);
                    setTables(d);
                }
	    }
	}
    } else if (source == d->d_DisplayObjects) {
	remClick(d);
	success = true;
    }

    gtk_drag_finish(context, success, false, time);
    return true;
}

static DisplayObject **markselected(GuiObject go)
{
    GtkTreeView *tree = GTK_TREE_VIEW(go);
    GtkTreeSelection *selection = gtk_tree_view_get_selection(tree);
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    GList *list = gtk_tree_selection_get_selected_rows(selection, 0);
    gint num = g_list_length(list), i = 0;

    DisplayObject **selected = g_new(DisplayObject *, num + 1);

    for (GList *item = list; item; item = item->next) {
	GtkTreePath *path = (GtkTreePath *)item->data;
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter(model, &iter, path)) {
	    gtk_tree_model_get(model, &iter, 0, &selected[i++], -1);
	}
	gtk_tree_path_free(path);
    }
    g_list_free(list);

    selected[i] = 0;
    return selected;
}

static void newTableClick(StructureEditorData *d)
{
    DisplayObject *dioparent = 0;
    if (d->d_RootDisplay) {
        /* get active display item */
        dioparent = displayActiveEntry(d->d_DisplayObjects);
    }

    /* open the new table gui */
    GtkNewTable(d, d->d_Project, dioparent);
}

void NewTableFinish(StructureEditorData *d, Table *tab)
{
    if (d && tab) {
        /* update table and display list */
        SetDisplayProjectActiveTable(d->d_Project, tab);
        setTables(d);
        setDisplay(d, false);
    }
}

static void changeTableClick(StructureEditorData *d)
{
    Table *tab = (Table *)GuiListActiveEntry(d->d_Tables);

    if (tab)
        GtkChangeTable(d, tab);
}

void ChangeTableFinish(StructureEditorData *d, Table *tab)
{
    if (d && tab) {
        gtk_widget_queue_draw(d->d_Tables);
        gtk_widget_queue_draw(d->d_DisplayObjects);
        if (d->d_NameChangeObject) {
            StructureEditorNextNameChange(d->d_Project, d->d_NameChangeObject);
        }
    }
}

static void deleteTableClick(StructureEditorData *d)
{
    Table *tab = (Table *)GuiListActiveEntry(d->d_Tables);
    if (tab) {
        bool deleted = GuiDeleteTable(tab);
        if (deleted) {
            setTables(d);
            setDisplay(d, false);
        }
    }
}

static void sortTablesClick(StructureEditorData *d)
{
    List *tabs = ProjectTables(d->d_Project);
    void *node = GuiListActiveEntry(d->d_Tables);

    if (d->d_TableSortMethod == 0) {
        SortTablesAlpha(tabs);
	d->d_TableSortMethod = 1;
	g_object_set(G_OBJECT(d->d_SortTables), 
		"label", GCS(MSG_StructureEditor_Sort2),
		NULL);
    } else {
        SortTablesDisplay(tabs);
	d->d_TableSortMethod = 0;
	g_object_set(G_OBJECT(d->d_SortTables), 
		"label", GCS(MSG_StructureEditor_Sort1),
		NULL);
    }
    GuiListSetListWithFilter(d->d_Tables, tabs, TableSystemFilter);
    GuiListSetActiveNode(d->d_Tables, node);
}

gboolean fieldsOnDragDrop(GtkWidget *widget, GdkDragContext *context,
	gint x, gint y, guint time, StructureEditorData *d)
{
    gboolean success = false;
    GtkWidget *source = gtk_drag_get_source_widget(context);

    if (source == d->d_Fields) {
	List *fields = TableFields(d->d_CurrentTable);
	Field *field = (Field *)GuiListActiveEntry(d->d_Fields);
	if (fields && field) {
	    GtkTreeView *tree = GTK_TREE_VIEW(widget);
	    GtkTreePath *path = 0;
	    GtkTreeViewDropPosition drop;
	    Field *pred = 0;
	    gtk_tree_view_get_dest_row_at_pos(tree, x, y, &path, &drop);
	    if (path) {
		GtkTreeModel *model = gtk_tree_view_get_model(tree);
		GtkTreeIter iter;

		if (gtk_tree_model_get_iter(model, &iter, path)) {
		    gtk_tree_model_get(model, &iter, 0, &pred, -1);
		}
		gtk_tree_path_free(path);
		if (pred && drop == GTK_TREE_VIEW_DROP_BEFORE) {
	     	    pred = PRED(pred);
     		}
	    }
	    if (!pred) {
		pred = (Field *)TAILPRED(fields);
	    }
	    success = true;
 	    if (field != pred) {
                long error = ObtainProjectChangeLock(d->d_Project);
                if (error != OK) {
                    Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
                    success = false;
                } else {
                    REMOVE(field);
                    INSERT(fields, field, pred);
                    ProjectChange(d->d_Project, TRUE);
                    SetDisplayTableActiveField(d->d_CurrentTable, field);
                    setFields(d);
                }
	    }
	}
    } else if (source == d->d_DisplayObjects) {
	remClick(d);
	success = true;
    }

    gtk_drag_finish(context, success, false, time);
    return true;
}

static void newFieldClick(StructureEditorData *d)
{
    DisplayObject *dioparent = 0;
    if (!d->d_RootDisplay) {
        /* get active display item */
        dioparent = displayActiveEntry(d->d_DisplayObjects);
    }

    /* open the new field gui */
    GtkNewField(d, d->d_CurrentTable, 0, dioparent);
}

static void copyFieldClick(StructureEditorData *d)
{
    Field *field = (Field *)GuiListActiveEntry(d->d_Fields);
    DisplayObject *dioparent = 0;
    if (!d->d_RootDisplay) {
        /* get active display item */
        dioparent = displayActiveEntry(d->d_DisplayObjects);
    }

    GtkNewField(d, d->d_CurrentTable, field, dioparent);
}

void NewFieldFinish(StructureEditorData *d, Field *field)
{
    if (d && field) {
        /* update the fields list and the display list*/
        Table *tab = FieldTable(field);
        SetDisplayTableActiveField(tab, field);
        SetDisplayTableActiveDisplayObject(tab, FieldDisplayObject(field));
        setFields(d);
        setDisplay(d, false);
    }
}

static void changeFieldClick(StructureEditorData *d)
{
    Field *field = (Field *)GuiListActiveEntry(d->d_Fields);
    if (field) {
        GtkChangeField(d, field);
    }
}

void ChangeFieldFinish(StructureEditorData *d, Field *field)
{
    if (d && field) {
    	/* update the fields list and the display list*/
        gtk_widget_queue_draw(d->d_Fields);
        gtk_widget_queue_draw(d->d_DisplayObjects);
        if (d->d_NameChangeObject) {
            StructureEditorNextNameChange(d->d_Project, d->d_NameChangeObject);
        }
    }
}

static void deleteFieldClick(StructureEditorData *d)
{
    Field *field = (Field *)GuiListActiveEntry(d->d_Fields);
    if (field) {
	bool deleted = GuiDeleteField(field);
        if (deleted) {
            setFields(d);
            setDisplay(d, false);
        }
    }
}

static void sortFieldsClick(GuiObject go, StructureEditorData *d)
{
    List *fields = TableFields(d->d_CurrentTable);

    if (fields) {
	void *node = GuiListActiveEntry(d->d_Fields);

    	if (d->d_FieldSortMethod == 0) {
            SortFieldsAlpha(fields);
	    d->d_FieldSortMethod = 1;
	    g_object_set(G_OBJECT(d->d_SortFields), 
		    "label", GCS(MSG_StructureEditor_Sort2),
		    NULL);
	} else {
            SortFieldsDisplay(fields);
	    d->d_FieldSortMethod = 0;
	    g_object_set(G_OBJECT(d->d_SortFields), 
		    "label", GCS(MSG_StructureEditor_Sort1),
		    NULL);
	}
	GuiListSetList(d->d_Fields, fields);
	GuiListSetActiveNode(d->d_Fields, node);
    }
}

static void defaultFinish(StructureEditorData *d, DisplayObject *dio)
{
    if (d) {
	setDisplay(d, false);
    }
}

void TextFinish(StructureEditorData *d, DisplayObject *dio) 
{ defaultFinish(d, dio); }
void SpaceFinish(StructureEditorData *d, DisplayObject *dio) 
{ defaultFinish(d, dio); }
void ImageFinish(StructureEditorData *d, DisplayObject *dio) 
{ defaultFinish(d, dio); }
void GroupFinish(StructureEditorData *d, DisplayObject *dio) 
{ defaultFinish(d, dio); }
void RegisterFinish(StructureEditorData *d, DisplayObject *dio) 
{ defaultFinish(d, dio); }
void WindowFinish(StructureEditorData *d, DisplayObject *dio) 
{ defaultFinish(d, dio); }
void DisplayTableFinish(StructureEditorData *d, DisplayObject *dio) 
{ defaultFinish(d, dio); }
void DisplayFieldFinish(StructureEditorData *d, DisplayObject *dio) 
{ defaultFinish(d, dio); }

static void displayOnRowActivated(GtkTreeView *tree,
	GtkTreePath *path, GtkTreeViewColumn *column, StructureEditorData *d)
{
    DisplayObject *dio = 0;
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    if (path) {
	GtkTreeIter iter;

	if (gtk_tree_model_get_iter(model, &iter, path)) {
	    gtk_tree_model_get(model, &iter, 0, &dio, -1);
	}
    }

    if (dio) {
	switch (DisplayObjectType(dio)) {
            case DO_TYPE_TEXT:
    		GuiText(d, dio);
		break;

            case DO_TYPE_SPACE:
    		GuiSpace(d, dio);
		break;

            case DO_TYPE_IMAGE:
    		GuiImage(d, dio);
		break;

            case DO_TYPE_GROUP:
    		GuiGroup(d, dio);
		break;

            case DO_TYPE_REGISTER:
    		GuiRegister(d, dio);
		break;

            case DO_TYPE_WINDOW:
    		GuiWindow(d, dio);
		break;

            case DO_TYPE_TABLE:
                GuiDisplayTable(d, dio);
		break;

            case DO_TYPE_FIELD:
                GuiDisplayField(d, dio);
		break;
	}
    }
}

gboolean displayOnDragDrop(GtkWidget *widget, GdkDragContext *context,
	gint x, gint y, guint time, StructureEditorData *d)
{
    DisplayObject *dio = 0;
    gboolean success = false;
    GtkWidget *source = gtk_drag_get_source_widget(context);

    if (source == d->d_Fields && !d->d_RootDisplay) {
	Field *field = (Field *)GuiListActiveEntry(d->d_Fields);
        dio = FieldDisplayObject(field);
    } else if (source == d->d_Tables && d->d_RootDisplay) {
	Table *tab = (Table *)GuiListActiveEntry(d->d_Tables);
        dio = TableDisplayObject(tab);
    } else if (source == d->d_DisplayObjects) {
	dio = displayActiveEntry(d->d_DisplayObjects);
    }

    if (dio) {
        long error = ObtainProjectChangeLock(d->d_Project);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        } else {
            GtkTreePath *path = 0;
            GtkTreeViewDropPosition drop;
            GtkTreeView *tree = GTK_TREE_VIEW(widget);
            gtk_tree_view_get_dest_row_at_pos(tree, x, y, &path, &drop);
            GtkTreeModel *model = gtk_tree_view_get_model(tree);
            GtkTreeIter iter;
            if (!path) {
                // we assume the user wants to place it at the very end
                path = gtk_tree_path_new_first();
                while (gtk_tree_model_get_iter(model, &iter, path)) {
                    gtk_tree_path_next(path);
                }
                gtk_tree_path_prev(path);
                drop = GTK_TREE_VIEW_DROP_AFTER;
            }
            DisplayObject *dioparent = 0;
            bool isPred = false;
            if (gtk_tree_model_get_iter(model, &iter, path)) {
                gtk_tree_model_get(model, &iter, 0, &dioparent, -1);
            }
            if (dioparent) {
                bool noChilds = (DisplayObjectObjects(dioparent) == 0);

                if (drop == GTK_TREE_VIEW_DROP_BEFORE ||
                   (drop == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE && noChilds)) {
                    // take object one before
                    if (gtk_tree_path_prev(path) &&
                            gtk_tree_model_get_iter(model, &iter, path)) {
                        gtk_tree_model_get(model, &iter, 0, &dioparent, -1);
                        isPred = true;
                    } else {
                        dioparent = DisplayObjectParent(dioparent);
                    }
                } else if (drop == GTK_TREE_VIEW_DROP_AFTER) {
                    isPred = d->d_RootDisplay ||
                        DisplayObjectType(dioparent) != DO_TYPE_TABLE;
                }
            }
            gtk_tree_path_free(path);
            DisplayObjectAdd(dio, dioparent, isPred);
            addDisplayObject(d, dio);
            ProjectChange(d->d_Project, TRUE);
            success = true;
        }
    }

    gtk_drag_finish(context, success, false, time);
    return true;
}

static void displayCycleChange(StructureEditorData *d)
{
    long active = gtk_combo_box_get_active(GTK_COMBO_BOX(d->d_DisplayType));

    GuiSetTooltip(d->d_Add, GCS((active == 1)? 
		MSG_StructureEditor_AddBH1 : MSG_StructureEditor_AddBH2));
    d->d_RootDisplay = (active == 1);
    setDisplay(d, false);
}

static void addClick(StructureEditorData *d)
{
    DisplayObject *add;

    /* get active entry from tables/fields list */
    if (d->d_RootDisplay) {
	Table *tab = (Table *)GuiListActiveEntry(d->d_Tables);
        add = TableDisplayObject(tab);
    } else {
	Field *field = (Field *)GuiListActiveEntry(d->d_Fields);
        add = FieldDisplayObject(field);
    }

    DisplayObject *dioparent = displayActiveEntry(d->d_DisplayObjects);

    if (add) {
        long error = ObtainProjectChangeLock(d->d_Project);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }
        DisplayObjectAdd(add, dioparent, false);
        addDisplayObject(d, add);
        ProjectChange(d->d_Project, TRUE);
    }
}

static void displayAction(StructureEditorData *d, 
	bool (*func)(DisplayObject *), bool reverse, bool rememberCursor)
{
    // we apply func on all selected items
    GtkTreeView *tree = GTK_TREE_VIEW(d->d_DisplayObjects);
    GtkTreeSelection *selection = gtk_tree_view_get_selection(tree);
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    GList *list = gtk_tree_selection_get_selected_rows(selection, 0);
    bool change = false;
    long error = OK;

    if (reverse) {
	list = g_list_reverse(list);
    }
    for (GList *item = list; item && error == OK; item = item->next) {
	GtkTreePath *path = (GtkTreePath *)item->data;
	GtkTreeIter iter;
	DisplayObject *dio = 0;
	if (gtk_tree_model_get_iter(model, &iter, path)) {
	    gtk_tree_model_get(model, &iter, 0, &dio, -1);
	    if (dio) {
                if (DisplayObjectType(dio)==DO_TYPE_TABLE && !d->d_RootDisplay){
                    continue;
                }
                error = ObtainProjectChangeLock(d->d_Project);
		if (error == OK && func(dio)) {
		    change = true;
		}
	    }
	}
	gtk_tree_path_free(path);
    }
    g_list_free(list);

    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
    } else if (change) {
	displayActiveChange(d);		// saves active object (number) again
	setDisplay(d, rememberCursor);
	ProjectChange(d->d_Project, TRUE);
    }
}

static void remClick(StructureEditorData *d)
{
    displayAction(d, DisplayObjectRemove, false, true);
}

static void upClick(StructureEditorData *d)
{
    displayAction(d, DisplayObjectUp, false, false);
}

static void downClick(StructureEditorData *d)
{
    displayAction(d, DisplayObjectDown, true, false);
}

static void inClick(StructureEditorData *d)
{
    displayAction(d, DisplayObjectIn, true, false);
}

static void outClick(StructureEditorData *d)
{
    displayAction(d, DisplayObjectOut, false, false);
}

static void clickFuncSimple(StructureEditorData *d,
	long (*newFunc)(Project *, Table *, DisplayObject *, DisplayObject **),
	void (*editFunc)(StructureEditorData *, DisplayObject *))
{
    long error = ObtainProjectChangeLock(d->d_Project);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    /* get active display item */
    DisplayObject *dioparent = displayActiveEntry(d->d_DisplayObjects);
    if (!dioparent) {
        dioparent = RootDisplayObject(d);
    }

    /* install new window */
    DisplayObject *dio =0;
    error = newFunc(d->d_Project,
	    d->d_RootDisplay? 0 : d->d_CurrentTable, dioparent, &dio);

    if (error) {
	GtkWidget *top = gtk_widget_get_toplevel(d->d_DisplayObjects);
	Ask2(top, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));
    } else {
	addDisplayObject(d, dio);
        ProjectChange(d->d_Project, TRUE);
	if (editFunc) {
	    editFunc(d, dio);
	}
    }
}

static void textClick(StructureEditorData *d)
{
    clickFuncSimple(d, DisplayNewText, GuiText);
}

static void imageClick(StructureEditorData *d)
{
    clickFuncSimple(d, DisplayNewImage, GuiImage);
}

static void spaceClick(StructureEditorData *d)
{
    clickFuncSimple(d, DisplayNewSpace, GuiSpace);
}

static void balanceClick(StructureEditorData *d)
{
    clickFuncSimple(d, DisplayNewBalance, NULL);
}

static void clickFuncSelected(StructureEditorData *d,
	long (*newFunc)(Project *, Table *, DisplayObject **, DisplayObject *,
	    DisplayObject **),
	void (editFunc)(StructureEditorData *, DisplayObject *))
{
    long error = ObtainProjectChangeLock(d->d_Project);
    if (error != OK) {
        Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        return;
    }

    /* get current selection of display list */
    DisplayObject **selected = markselected(d->d_DisplayObjects);

    /* get active display item */
    DisplayObject *dioparent = displayActiveEntry(d->d_DisplayObjects);
    if (!dioparent) {
        dioparent = RootDisplayObject(d);
    }

    /* install new window */
    DisplayObject *dio =0;
    error = newFunc(d->d_Project,
	    d->d_RootDisplay? 0 : d->d_CurrentTable, selected, dioparent, &dio);
    g_free(selected);

    if (error) {
	GtkWidget *top = gtk_widget_get_toplevel(d->d_DisplayObjects);
	Ask2(top, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));
    } else {
	addDisplayObject(d, dio);

        ProjectChange(d->d_Project, TRUE);
        editFunc(d, dio);
    }
}

static void groupClick(StructureEditorData *d)
{
    clickFuncSelected(d, DisplayNewGroup, GuiGroup);
}

static void registerClick(StructureEditorData *d)
{
    clickFuncSelected(d, DisplayNewRegister, GuiRegister);
}

static void windowClick(StructureEditorData *d)
{
    clickFuncSelected(d, DisplayNewWindow, GuiWindow);
}

static void structureEditorOnDelete(GuiObject win, GdkEvent *event,
	StructureEditorData *d)
{
    struct Project *pr = d->d_Project;

    StructureEditor(pr);
}

GuiObject OpenStructureEditor(Project *pr)
{
    StructureEditorData *d = g_new0(StructureEditorData, 1);
    d->d_Project = pr;

    d->d_Tables = GuiPointerTreeView(tableCellDataFunc);
    GuiSetTooltip(d->d_Tables, GCS(MSG_StructureEditor_TablesBH));
    GtkTreeSelection *selection = 
	gtk_tree_view_get_selection(GTK_TREE_VIEW(d->d_Tables));
    g_signal_connect_swapped(selection, "changed", 
	    G_CALLBACK(tablesActiveChange), d);
    g_signal_connect_swapped(d->d_Tables, "row-activated", 
		G_CALLBACK(changeTableClick), d);
    // drag and drop
    static GtkTargetEntry tableTargets[] = {
	{ (gchar *)"Table", GTK_TARGET_SAME_APP, TARGET_TABLE },
	{ (gchar *)"DisplayObject", GTK_TARGET_SAME_APP, TARGET_DISPLAY }
    };
    gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(d->d_Tables),
	    GDK_BUTTON1_MASK, tableTargets, 1,
	    (GdkDragAction)(GDK_ACTION_COPY | GDK_ACTION_MOVE));
    gtk_tree_view_enable_model_drag_dest(GTK_TREE_VIEW(d->d_Tables),
	    tableTargets, 2, GDK_ACTION_MOVE);
    g_signal_connect(d->d_Tables, "drag-drop", G_CALLBACK(tablesOnDragDrop), d); 
    d->d_NewTable = GuiSimpleButton(GCS(MSG_StructureEditor_Table_New));
    g_signal_connect_swapped(d->d_NewTable, "clicked", 
            G_CALLBACK(newTableClick), d);
    GuiSetTooltip(d->d_NewTable, GCS(MSG_StructureEditor_Table_NewBH));

    d->d_DeleteTable = GuiSimpleButton(GCS(MSG_StructureEditor_Table_Delete));
    gtk_widget_set_sensitive(d->d_DeleteTable, false);
    g_signal_connect_swapped(d->d_DeleteTable, "clicked", 
            G_CALLBACK(deleteTableClick), d);
    GuiSetTooltip(d->d_DeleteTable, GCS(MSG_StructureEditor_Table_DeleteBH));

    d->d_SortTables = GuiSimpleButton(GCS(MSG_StructureEditor_Sort1));
    g_signal_connect_swapped(d->d_SortTables, "clicked", 
            G_CALLBACK(sortTablesClick), d);
    GuiSetTooltip(d->d_SortTables, GCS(MSG_StructureEditor_Table_SortBH));

    GuiObject tableActions = HGroup,
	"child", d->d_NewTable,
	"child", d->d_DeleteTable,
	"child", d->d_SortTables,
	End;

    d->d_TablesGroup = VGroup,
	"child", ScrolledWindowObject,
	    "hscrollbar-policy", GTK_POLICY_AUTOMATIC,
	    "shadow-type", GTK_SHADOW_IN,
	    "child", d->d_Tables,
	    End,
	"child", tableActions, "vert-weight", 0,
	End;
    GuiSetHelpNode(d->d_TablesGroup, "Table management");
    GuiObject tabFrame = Frame(GCS(MSG_StructureEditor_Tables)),
	"child", GroupAlignment(true),
	    "child", d->d_TablesGroup,
	    End,
	End;

    d->d_Fields = GuiPointerTreeView(fieldCellDataFunc);
    GuiSetTooltip(d->d_Fields, GCS(MSG_StructureEditor_FieldsBH));
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(d->d_Fields));
    g_signal_connect_swapped(selection, "changed", 
	    G_CALLBACK(fieldsActiveChange), d);
    g_signal_connect_swapped(d->d_Fields, "row-activated", 
		G_CALLBACK(changeFieldClick), d);
    // drag and drop
    static GtkTargetEntry fieldTargets[] = {
	{ (gchar *)"Field", GTK_TARGET_SAME_APP, TARGET_FIELD },
	{ (gchar *)"DisplayObject", GTK_TARGET_SAME_APP, TARGET_DISPLAY }
    };
    gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(d->d_Fields),
	    GDK_BUTTON1_MASK, fieldTargets, 1,
	    (GdkDragAction)(GDK_ACTION_COPY | GDK_ACTION_MOVE));
    gtk_tree_view_enable_model_drag_dest(GTK_TREE_VIEW(d->d_Fields),
	    fieldTargets, 2, GDK_ACTION_MOVE);
    g_signal_connect(d->d_Fields, "drag-drop", 
	    G_CALLBACK(fieldsOnDragDrop), d); 

    d->d_NewField = GuiSimpleButton(GCS(MSG_New));
    g_signal_connect_swapped(d->d_NewField, "clicked", 
            G_CALLBACK(newFieldClick), d);
    GuiSetTooltip(d->d_NewField, GCS(MSG_StructureEditor_Field_NewBH));

    d->d_CopyField = 
	GuiSimpleButton(GCS(MSG_StructureEditor_Field_Copy));
    gtk_widget_set_sensitive(d->d_CopyField, false);
    g_signal_connect_swapped(d->d_CopyField, "clicked", 
            G_CALLBACK(copyFieldClick), d);
    GuiSetTooltip(d->d_CopyField,GCS(MSG_StructureEditor_Field_CopyBH));

    d->d_DeleteField = 
	GuiSimpleButton(GCS(MSG_StructureEditor_Field_Delete));
    gtk_widget_set_sensitive(d->d_DeleteField, false);
    g_signal_connect_swapped(d->d_DeleteField, "clicked", 
            G_CALLBACK(deleteFieldClick), d);
    GuiSetTooltip(d->d_DeleteField,
	    GCS(MSG_StructureEditor_Field_DeleteBH));

    d->d_SortFields = 
	GuiSimpleButton(GCS(MSG_StructureEditor_Sort1));
    g_signal_connect(d->d_SortFields, "clicked", 
	    G_CALLBACK(sortFieldsClick), d);
    GuiSetTooltip(d->d_SortFields, 
            GCS(MSG_StructureEditor_Field_SortBH));

    GuiObject fieldActions = HGroup,
	"child", d->d_NewField,
	"child", d->d_CopyField,
	"child", d->d_DeleteField,
	"child", d->d_SortFields,
	End;

    d->d_FieldsGroup = VGroup,
        "sensitive", false,
	"child", ScrolledWindowObject,
	    "hscrollbar-policy", GTK_POLICY_NEVER,
	    "shadow-type", GTK_SHADOW_IN,
	    "child", d->d_Fields,
	    End,
	"child", fieldActions, "vert-weight", 0,
	End;
    GuiSetHelpNode(d->d_FieldsGroup, "Field management");
    GuiObject fieldFrame = Frame(GCS(MSG_Fields)),
	"child", GroupAlignment(true),
	    "child", d->d_FieldsGroup,
	    End,
	End;

    static char *displayTypeContent[3];
    displayTypeContent[0] = GCS(MSG_StructureEditor_TableDisplay);
    displayTypeContent[1] = GCS(MSG_StructureEditor_MainWindow);
    displayTypeContent[2] = NULL;
    d->d_DisplayType = GuiComboBox(displayTypeContent);
    g_signal_connect_swapped(d->d_DisplayType, "changed", 
	    G_CALLBACK(displayCycleChange), d);

    d->d_Add = GuiSimpleButton(GCS(MSG_StructureEditor_Add));
    g_signal_connect_swapped(d->d_Add, "clicked", G_CALLBACK(addClick), d);
    GuiSetTooltip(d->d_Add,GCS(MSG_StructureEditor_AddBH2));

    d->d_Rem = GuiSimpleButton(GCS(MSG_StructureEditor_Rem));
    g_signal_connect_swapped(d->d_Rem, "clicked", G_CALLBACK(remClick), d);
    GuiSetTooltip(d->d_Rem, GCS(MSG_StructureEditor_RemBH));

    d->d_Up = GuiSimpleButton(GCS(MSG_StructureEditor_Up));
    g_signal_connect_swapped(d->d_Up, "clicked", G_CALLBACK(upClick), d);
    GuiSetTooltip(d->d_Up, GCS(MSG_StructureEditor_UpBH));

    d->d_Down = GuiSimpleButton(GCS(MSG_StructureEditor_Down));
    g_signal_connect_swapped(d->d_Down, "clicked", G_CALLBACK(downClick), d);
    GuiSetTooltip(d->d_Down, GCS(MSG_StructureEditor_DownBH));

    d->d_In = GuiSimpleButton(GCS(MSG_StructureEditor_In));
    g_signal_connect_swapped(d->d_In, "clicked", G_CALLBACK(inClick), d);
    GuiSetTooltip(d->d_In, GCS(MSG_StructureEditor_InBH));

    d->d_Out = GuiSimpleButton(GCS(MSG_StructureEditor_Out));
    g_signal_connect_swapped(d->d_Out, "clicked", G_CALLBACK(outClick), d);
    GuiSetTooltip(d->d_Out, GCS(MSG_StructureEditor_OutBH));

    d->d_Text = GuiSimpleButton(GCS(MSG_StructureEditor_Text));
    g_signal_connect_swapped(d->d_Text, "clicked", G_CALLBACK(textClick), d);
    GuiSetTooltip(d->d_Text, GCS(MSG_StructureEditor_TextBH));

    d->d_Image = GuiSimpleButton(GCS(MSG_Image));
    g_signal_connect_swapped(d->d_Image, "clicked", G_CALLBACK(imageClick), d);
    GuiSetTooltip(d->d_Image, GCS(MSG_StructureEditor_ImageBH));

    d->d_Space = GuiSimpleButton(GCS(MSG_StructureEditor_Space));
    g_signal_connect_swapped(d->d_Space, "clicked", G_CALLBACK(spaceClick), d);
    GuiSetTooltip(d->d_Space, GCS(MSG_StructureEditor_SpaceBH));

    d->d_Balance = GuiSimpleButton(GCS(MSG_StructureEditor_Balance));
    g_signal_connect_swapped(d->d_Balance, "clicked", 
	    G_CALLBACK(balanceClick), d);
    GuiSetTooltip(d->d_Balance, GCS(MSG_StructureEditor_BalanceBH));

    d->d_Group = GuiSimpleButton(GCS(MSG_StructureEditor_Group));
    g_signal_connect_swapped(d->d_Group, "clicked", G_CALLBACK(groupClick), d);
    GuiSetTooltip(d->d_Group, GCS(MSG_StructureEditor_GroupBH));

    d->d_Register = GuiSimpleButton(GCS(MSG_StructureEditor_Register));
    g_signal_connect_swapped(d->d_Register, "clicked", 
	    G_CALLBACK(registerClick), d);
    GuiSetTooltip(d->d_Register, GCS(MSG_StructureEditor_RegisterBH));

    d->d_Window = GuiSimpleButton(GCS(MSG_StructureEditor_Window));
    g_signal_connect_swapped(d->d_Window, "clicked", 
	    G_CALLBACK(windowClick), d);
    GuiSetTooltip(d->d_Window, GCS(MSG_StructureEditor_WindowBH));

    GuiObject displayButtons = VGroup,
	"child", SpaceObject, End,
	"child", d->d_Add, "weight", 0,
	"child", d->d_Rem, "weight", 0,
	"child", d->d_Up, "weight", 0,
	"child", d->d_Down, "weight", 0,
	"child", d->d_In, "weight", 0,
	"child", d->d_Out, "weight", 0,
	"child", SpaceObject, End,
	"child", d->d_Text, "weight", 0,
	"child", d->d_Image, "weight", 0,
	"child", d->d_Space, "weight", 0,
	"child", d->d_Balance, "weight", 0,
	"child", d->d_Group, "weight", 0,
	"child", d->d_Register, "weight", 0,
	"child", d->d_Window, "weight", 0,
	"child", SpaceObject, End,
	End;

    GtkTreeStore *store = gtk_tree_store_new(1, G_TYPE_POINTER);
    GtkWidget *tree = TreeViewObject,
        "model", store,
	"headers-visible", false,
	End;
    g_object_unref(G_OBJECT(store));

    GtkCellRenderer *cell = gtk_cell_renderer_text_new();
    GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
	    "Item", cell, NULL);
    gtk_tree_view_column_set_cell_data_func(column, cell, 
	    displayCellDataFunc, 0, 0);
    gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_MULTIPLE);

    d->d_DisplayObjects = tree;
    GuiSetTooltip(d->d_DisplayObjects, GCS(MSG_StructureEditor_DisplayBH));
    g_signal_connect_swapped(tree, "row-collapsed", 
	    G_CALLBACK(displayRowChange), d);
    g_signal_connect_swapped(tree, "row-expanded", 
	    G_CALLBACK(displayRowChange), d);
    g_signal_connect(tree, "test-collapse-row",
	    G_CALLBACK(displayTestCollapseRow), 0);
    g_signal_connect_swapped(tree, "cursor-changed", 
	    G_CALLBACK(displayActiveChange), d);
    g_signal_connect(tree, "row-activated",
    	    G_CALLBACK(displayOnRowActivated), d);
    // drag and drop
    static GtkTargetEntry displayTargets[] = {
	{ (gchar *)"DisplayObject", GTK_TARGET_SAME_APP, TARGET_DISPLAY }
    };
    gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(d->d_DisplayObjects),
	    GDK_BUTTON1_MASK, displayTargets, 1, GDK_ACTION_MOVE);
    g_signal_connect(tree, "drag-drop", G_CALLBACK(displayOnDragDrop), d);

    d->d_DisplayObjectsGroup = HGroup,
	"sensitive", false,
	"child", displayButtons, "weight", 0,
	"child", ScrolledWindowObject,
	    "hscrollbar-policy", GTK_POLICY_AUTOMATIC,
	    "shadow-type", GTK_SHADOW_IN,
	    "child", d->d_DisplayObjects,
	    End,
	End;
    GuiSetHelpNode(d->d_DisplayObjectsGroup, "Display field");

    GuiObject displayFrame = Frame(GCS(MSG_Display)),
	"child", GroupAlignment(true),
	    "child", VGroup,
	        "child", d->d_DisplayType, "weight", 0,
	        "child", d->d_DisplayObjectsGroup,
		End,
	    End,
	End;

    GuiObject vpaned, hpaned;
    GuiObject group = VGroup, 
	"child", hpaned = HPanedObject,
	    "child", vpaned = VPanedObject,
	        "child", tabFrame,
		"child", fieldFrame,
                "position", 150,
		End,
	    "child", displayFrame,
	    End,
	End;
    gtk_container_child_set(GTK_CONTAINER(vpaned), tabFrame, 
	    "resize", false, "shrink", false, NULL);
    gtk_container_child_set(GTK_CONTAINER(vpaned), fieldFrame, 
	    "resize", true, "shrink", false, NULL);
    gtk_container_child_set(GTK_CONTAINER(hpaned), vpaned, 
	    "resize", false, "shrink", false, NULL);
    gtk_container_child_set(GTK_CONTAINER(hpaned), displayFrame, 
	    "resize", true, "shrink", false, NULL);

    GuiObject win = WindowObject,
	"window-position", GTK_WIN_POS_CENTER,
	"default-width", 800,
	End;
    SetRootWindowTitle(pr, win);
    struct WindowDimensions wd;
    if (GetPrefsWindowDimensions(pr, PREFS_STRUCTUREEDITOR, &wd)) {
        GuiSetWindowDimensions(win, &wd);
    }

    GtkWidget *root = VGroup, End;
    g_object_set(G_OBJECT(win), "child", root, NULL);
    GuiWindowAddMenus(win, root, TRUE);

    GtkWidget *align = GroupAlignment(true),
            "child", group,
            End;
    g_object_set(G_OBJECT(root), "child", align, NULL);

    gtk_widget_show_all(root);
    GuiSetWindowProject(win, pr);
    GuiSetHelpNode(win, "Structure editor");
    g_object_set_data_full(G_OBJECT(win), "edit-data", d, g_free);
    GuiAddWindowCallbacks(win, 0);

    g_signal_connect(win, "delete-event", 
	    G_CALLBACK(structureEditorOnDelete), d);

    if (!GuiOpenWindow(win)) {
	gtk_widget_destroy(win);
	win = 0;
    } else {
	setTables(d);
    }
        
    return win;
}


void StructureEditorUpdate(GuiObject se)
{
    if (se) {
	StructureEditorData *d = SE_DATA(se);
	if (d) {
	    setTables(d);
	    setDisplay(d, false);
	}
    }
}

void ClearStructureEditor(GuiObject se)
{
    StructureEditorData *d = SE_DATA(se);

    d->d_CurrentTable = 0;
    GtkTreeView *tree = GTK_TREE_VIEW(d->d_DisplayObjects);
    GtkTreeModel *model = gtk_tree_view_get_model(tree);
    GtkTreeStore *store = GTK_TREE_STORE(model);
    gtk_tree_store_clear(store);

    GuiListSetList(d->d_Fields, 0);
    GuiListSetList(d->d_Tables, 0);
}

void StructureEditorSetNameChangeObject(GuiObject se, void *object)
{
    if (se) {
        StructureEditorData *d = SE_DATA(se);
	if (d) {
            d->d_NameChangeObject = object;
        }
    }
}

void GuiChangeTable(GuiObject se, Table *tab)
{
    if (se) {
        StructureEditorData *d = SE_DATA(se);
	if (d) {
            GtkChangeTable(d, tab);
        }
    }
}

void GuiChangeField(GuiObject se, Field *field)
{
    if (se) {
        StructureEditorData *d = SE_DATA(se);
	if (d) {
            GtkChangeField(d, field);
        }
    }
}

#endif

