//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// gtkUtil.cc
//
// Creation date: 03.04.2004
// Last modified: 08.02.2022
//

#include "gtk/gtkUtil.h"
#include "gtk/classes/gtkgroup.h"
#include "gtk/classes/gtkradiobox.h"
#include "gtk/classes/gtksfwview.h"
#include "gtk/gtkimpl.h"
#include <gdk/gdkkeysyms.h>
#include "project.h"
#include "strhandle.h"
#include "dosutil.h"
#include "codeset.h"
#include "util.h"

void GuiUtilInit()
{
}

void GuiUtilExit()
{
}

bool GuiSetTooltip(GuiObject go, const char *help)
{
    if (!go) {
        return false;
    }

    gtk_widget_set_tooltip_text(go, help);
    return true;
}

GuiObject GuiMakeTooltip(GuiObject go, const char *help)
{
    GuiSetTooltip(go, help);
    return go;
}

void GuiSetListStoreLabels(GtkListStore *store, const char * const *labels)
{
    if (store) {
	gtk_list_store_clear(store);
	if (labels) {
	    const char * const *label;
	    for (label = labels; *label; label++) {
		GtkTreeIter iter;
		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter, 0, *label, -1);
	    }
	}
    }
}

void GuiSetListStoreListWithFilter(GtkListStore *store, List *list,
        bool (*filter)(const struct Node *node))
{
    if (store) {
	gtk_list_store_clear(store);
	if (list) {
	    Node *node;
	    FORNODE(node, list) {
                if (!filter || filter(node)) {
                    GtkTreeIter iter;
                    gtk_list_store_append(store, &iter);
                    gtk_list_store_set(store, &iter, 0, node, -1);
                }
	    }
	}
    }
}

void GuiSetListStoreList(GtkListStore *store, List *list)
{
    GuiSetListStoreListWithFilter(store, list, NULL);
}

char **GuiListStoreLabels(GtkTreeModel *model)
{
    char **labels = 0;

    if (model) {
	size_t count = 0;
	size_t len = 0;
	GtkTreeIter iter;

	if (gtk_tree_model_get_iter_first(model, &iter)) {
	    do {
		const char *label = 0;
		gtk_tree_model_get(model, &iter, 0, &label, -1);
		if (label) {
		    count++;
		    len += strlen(label) + 1;
		}
	    } while (gtk_tree_model_iter_next(model, &iter));
	}

	labels = (char **)g_malloc((count + 1) * sizeof(char *) + len);
	char *s = (char *)&labels[count + 1];
	size_t i = 0;
	
	if (gtk_tree_model_get_iter_first(model, &iter)) {
	    do {
		const char *label = 0;
		gtk_tree_model_get(model, &iter, 0, &label, -1);
		if (label) {
		    labels[i++] = s;
		    strcpy(s, label);
		    s += strlen(s) + 1;
		}
	    } while (gtk_tree_model_iter_next(model, &iter));
	}
	labels[i] = 0;
    }
    return labels;
}

long GuiListNumEntries(GuiObject tree)
{
    long count = 0;

    if (GTK_IS_TREE_VIEW(tree)) {
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
	if (model) {
	    GtkTreeIter iter;
	    if (gtk_tree_model_get_iter_first(model, &iter)) {
		do {
		    count++;
		} while (gtk_tree_model_iter_next(model, &iter));
	    }
	}
    }
    return count;
}

void GuiListSetListWithFilter(GtkWidget *tree, List *list,
        bool (*filter)(const struct Node *node))
{
    if (GTK_IS_TREE_VIEW(tree)) {
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
	GtkListStore *store = GTK_LIST_STORE(model);
        GuiSetListStoreListWithFilter(store, list, filter);
    }
}

void GuiListSetList(GtkWidget *tree, List *list)
{
    GuiListSetListWithFilter(tree, list, NULL);
}

void GuiListGetList(GtkWidget *tree, List *list)
{
    if (GTK_IS_TREE_VIEW(tree) && list) {
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
	GtkTreeIter iter;
	Node *node = 0, *pred = 0;

	if (gtk_tree_model_get_iter_first(model, &iter)) {
	    do {
		gtk_tree_model_get(model, &iter, 0, &node, -1);
		if (node) {
		    REMOVE(node);
		    INSERT(list, node, pred);
		    pred = node;
		}
	    } while (gtk_tree_model_iter_next(model, &iter));
	}
    }
}

long GuiListActive(GuiObject tree)
{
    long pos = -1;

    if (GTK_IS_TREE_VIEW(tree)) {
	GtkTreeSelection *selection = 
	    gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
	GtkTreeModel *model;
	GtkTreeIter iter;
	if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
	    GtkTreePath *path = gtk_tree_model_get_path(model, &iter);
    	    if (path) {
		gint *indices = gtk_tree_path_get_indices(path);
		if (indices) {
		    pos = indices[0];
		}
	    	gtk_tree_path_free(path);
    	    }
	}
    }
    return pos;
}

void *GuiListActiveEntry(GuiObject tree)
{
    void *entry = 0;

    if (GTK_IS_TREE_VIEW(tree)) {
	GtkTreeSelection *selection = 
	    gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
	GtkTreeModel *model;
	GtkTreeIter iter;
	if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
	    gtk_tree_model_get(model, &iter, 0, &entry, -1);
	}
    }
    return entry;
}

void GuiListSetActive(GuiObject tree, long pos)
{
    if (tree) {
        GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
        
        if (!model) {
            return;             /* no contents */
        }

	if (pos < 0 || pos >= gtk_tree_model_iter_n_children(model, 0)) {
            GtkTreeSelection *sel = 
                gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
            gtk_tree_selection_unselect_all(sel);
        } else {
            GtkTreePath *path = gtk_tree_path_new_from_indices(pos, -1);
            gtk_tree_view_set_cursor(GTK_TREE_VIEW(tree), path, 0, false);
            gtk_tree_path_free(path);
            gtk_tree_view_get_cursor(GTK_TREE_VIEW(tree), &path, 0);
            if (path) {
                gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(tree), 
                        path, NULL, TRUE, .5, .0);
                gtk_tree_path_free(path);
            }
	}
    }
}

void GuiListSetActiveNode(GuiObject tree, void *node)
{
    long pos = GuiListFindNode(tree, node);

    if (pos >= 0) {
	GuiListSetActive(tree, pos);
    }
}

void GuiListSetActiveEntry(GuiObject tree, void *entry)
{
    long pos = GuiListFindEntry(tree, entry);

    if (pos >= 0) {
	GuiListSetActive(tree, pos);
    }
}

long GuiSFWListActive(GuiObject go)
{
    long line = 0;

    if (GTK_IS_SFW_VIEW(go)) {
        line = gtk_sfw_view_get_active_line(GTK_SFW_VIEW(go));
    }
    return line;
}

void GuiSFWListSetActive(GuiObject go, long num)
{
    if (GTK_IS_SFW_VIEW(go)) {
        gtk_sfw_view_set_active_line(GTK_SFW_VIEW(go), num);
    }
}

char GuiSFWListNumColumns(GuiObject go)
{
    char cols = 0;

    if (GTK_IS_SFW_VIEW(go) && GTK_SFW_VIEW(go)->data) {
        cols = GTK_SFW_VIEW(go)->data->qgd_NumFields;
    }
    return cols;
}

const QueryFieldInfo *GuiSFWListFieldInfos(GuiObject go)
{
    const QueryFieldInfo *infos = NULL;

    if (GTK_IS_SFW_VIEW(go) && GTK_SFW_VIEW(go)->data) {
        infos = GTK_SFW_VIEW(go)->data->qgd_FieldInfos;
    }
    return infos;
}

const signed char *GuiSFWListColumnOrder(GuiObject go)
{
    const signed char *order = NULL;

    if (GTK_IS_SFW_VIEW(go)) {
        order = gtk_sfw_view_get_column_order(GTK_SFW_VIEW(go));
    }
    return order;
}

const struct QueryLine *GuiSFWListQueryLine(GuiObject go, long row)
{
    const struct QueryLine *ql = 0;

    if (!GTK_IS_SFW_VIEW(go) || !GTK_SFW_VIEW(go)->data) {
    } else if (row == 0) {
        ql = GTK_SFW_VIEW(go)->data->qgd_Titles;
    } else {
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(go));
        GtkTreePath *path = gtk_tree_path_new_from_indices(row - 1, -1);
	GtkTreeIter iter;

	if (gtk_tree_model_get_iter(model, &iter, path)) {
            ql = gtk_sfw_view_get_query_line(GTK_SFW_VIEW(go), &iter);
        }
        gtk_tree_path_free(path);
    }

    return ql;
}

Table *GuiSFWListControlTable(GuiObject go)
{
    Table *tab = NULL;

    if (GTK_IS_SFW_VIEW(go) && GTK_SFW_VIEW(go)->data) {
        tab = GTK_SFW_VIEW(go)->data->qgd_ControlTable;
    }
    return tab;
}

Record *GuiSFWListActiveRecord(GuiObject go)
{
    Record *rec = NULL;

    if (GTK_IS_SFW_VIEW(go)) {
        rec = gtk_sfw_view_get_active_record(GTK_SFW_VIEW(go));
    }
    return rec;
}

void GuiSFWListSetActiveRecord(GuiObject go, struct Record *rec)
{
    if (GTK_IS_SFW_VIEW(go)) {
        gtk_sfw_view_set_active_record(GTK_SFW_VIEW(go), rec);
    }
}

Record *GuiSFWListRecord(GuiObject go, long row)
{
    Record *rec = NULL;

    if (GTK_IS_SFW_VIEW(go)) {
        rec = gtk_sfw_view_get_record(GTK_SFW_VIEW(go), row);
    }
    return rec;
}

long GuiListFindEntry(GuiObject tree, void *entry)
{
    if (GTK_IS_TREE_VIEW(tree)) {
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
	GtkTreeIter iter;
	long count = 0;
	void *elem = NULL;

	if (gtk_tree_model_get_iter_first(model, &iter)) {
	    do {
		gtk_tree_model_get(model, &iter, 0, &elem, -1);
		if (elem == entry) {
		    return count;
		}
		count++;
	    } while (gtk_tree_model_iter_next(model, &iter));
	}
    }
    return -1;
}

long GuiListFindNode(GuiObject tree, void *node)
{
    if (GTK_IS_TREE_VIEW(tree) && node) {
       	/* quick test */
	long pos = NodeNum(node) - 1;
	void *elem = NULL;
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
	GtkTreePath *path = gtk_tree_path_new_from_indices(pos, -1);
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter(model, &iter, path)) {
	    gtk_tree_model_get(model, &iter, 0, &elem, -1);
	}
	gtk_tree_path_free(path);

	if (node == elem) {
	    return pos;
	}
    }
    return GuiListFindEntry(tree, node);
}

void *GuiListNthEntry(GuiObject tree, long pos)
{
    void *entry = NULL;

    if (tree && pos >= 0) {
        GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
        GtkTreePath *path = gtk_tree_path_new_from_indices(pos, -1);
	GtkTreeIter iter;

	if (gtk_tree_model_get_iter(model, &iter, path)) {
            gtk_tree_model_get(model, &iter, 0, &entry, -1);
        }
        gtk_tree_path_free(path);
    }

    return entry;
}

void GuiListRemoveEntry(GuiObject tree, void *entry)
{
    long pos = GuiListFindEntry(tree, entry);

    if (pos >= 0) {
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree));
	GtkTreePath *path = gtk_tree_path_new_from_indices(pos, -1);
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter(model, &iter, path)) {
	    long active = GuiListActive(tree);
	    gtk_list_store_remove(GTK_LIST_STORE(model), &iter);
	    if (active == pos) {
		long num = GuiListNumEntries(tree);
		if (active >= num) {
		    active = num - 1;
		}
		if (active >= 0) {
		    GuiListSetActive(tree, active);
		}
	    }
	}
	gtk_tree_path_free(path);
    }
}

static void onCellToggled(GtkCellRendererToggle *cell,
	gchar *path_str, GtkTreeModel *model)
{
    GtkTreePath *path = gtk_tree_path_new_from_string(path_str);
    GtkTreeIter  iter;
    gint column = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(cell), "column"));
    gboolean sel;

    if (gtk_tree_model_get_iter(model, &iter, path)) {
	gtk_tree_model_get(model, &iter, column, &sel, -1);
	gtk_list_store_set(GTK_LIST_STORE(model), &iter, column, !sel, -1);
    }
    gtk_tree_path_free(path);
}

GuiObject GuiTreeView(gint columns, GType *types, GtkCellRenderer **cells,
	GtkTreeCellDataFunc *funcs, gboolean checkboxes)
{
    GtkListStore *store;
    if (checkboxes) {
        // we append the types with an additional one holding the checkbox state
        GType *copy = g_new(GType, columns + 1);
        memcpy(copy, types, columns * sizeof(GType));
        copy[columns] = G_TYPE_BOOLEAN;
        store = gtk_list_store_newv(columns + 1, copy);
        g_free(copy);
    } else {
        store = gtk_list_store_newv(columns, types);
    }
    GtkWidget *tree = TreeViewObject,
        "model", store,
	"headers-visible", false,
	End;
    g_object_unref(G_OBJECT(store));

    if (checkboxes) {
	GtkCellRenderer *renderer = gtk_cell_renderer_toggle_new();
        g_object_set_data(G_OBJECT(renderer),"column",GINT_TO_POINTER(columns));
	GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
		"Selected", renderer, "active", columns, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
        g_signal_connect(renderer, "toggled", G_CALLBACK(onCellToggled), store);
    }
    for (gint i=0; i < columns; i++) {
	GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
		"Item", cells[i], NULL);
	g_object_set(G_OBJECT(column), "resizable", TRUE, NULL);
	gtk_tree_view_column_set_cell_data_func(column, cells[i], funcs[i],0,0);
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
    }
    
    return tree;
}

GuiObject GuiPointerTreeView(GtkTreeCellDataFunc func, gboolean checkboxes)
{
    GType type = G_TYPE_POINTER;
    GtkCellRenderer *cell = gtk_cell_renderer_text_new();

    return GuiTreeView(1, &type, &cell, &func, checkboxes);
}

static void stringCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    const gchar *str = 0;
    gtk_tree_model_get(model, iter, 0, &str, -1);
    g_object_set(G_OBJECT(cell), "text", str, NULL);
}

GuiObject GuiStringTreeView(gboolean checkboxes)
{
    GType type = G_TYPE_STRING;
    GtkCellRenderer *cell = gtk_cell_renderer_text_new();
    GtkTreeCellDataFunc func = stringCellDataFunc;

    return GuiTreeView(1, &type, &cell, &func, checkboxes);
}

static gchar *stringColumn(const char *str, int col)
{
    if (!str || col < 0) {
	return 0;
    }
    while (col > 0) {
	str = CodesetStrchr(str, '\t');
	if (!str) {
	    return 0;
	}
	str++;
	col--;
    }
    const char *t = CodesetStrchr(str, '\t');
    if (t) {
	return g_strndup(str, t - str);
    } else {
	return g_strdup(str);
    }
}

static void stringColumnCellDataFunc(GtkTreeViewColumn *column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	gpointer data)
{
    gint col = GPOINTER_TO_INT(data);
    const gchar *str = 0;
    gchar *text;

    gtk_tree_model_get(model, iter, 0, &str, -1);
    text = stringColumn(str, col);
    g_object_set(G_OBJECT(cell), "text", EmptyNullStr(text), NULL);
    g_free(text);
}

GuiObject GuiStringMultiColumnTreeView(gint columns, gboolean checkboxes,
	const char *titles)
{
    GType type = G_TYPE_STRING;
    GtkCellRenderer *cell = gtk_cell_renderer_text_new();
    GtkTreeCellDataFunc func = stringColumnCellDataFunc;

    GtkWidget *tree = GuiTreeView(1, &type, &cell, &func, checkboxes);
    for (int i=1; i < columns; i++) {
	GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
		"Item", cell, NULL);
	g_object_set(G_OBJECT(column), "resizable", TRUE, NULL);
	gtk_tree_view_column_set_cell_data_func(column, cell, func,
		GINT_TO_POINTER(i), NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);
    }
    if (titles) {
	int offset = 0;
	if (checkboxes) {
	    GtkTreeViewColumn *column = gtk_tree_view_get_column(
		    GTK_TREE_VIEW(tree), 0);
	    if (column) {
		gtk_tree_view_column_set_title(column, "");
	    }
	    offset++;
	}
	for (int i=0; i < columns; i++) {
	    GtkTreeViewColumn *column = gtk_tree_view_get_column(
		    GTK_TREE_VIEW(tree), i + offset);
	    if (column) {
		gchar *text = stringColumn(titles, i);
		gtk_tree_view_column_set_title(column, EmptyNullStr(text));
		g_free(text);
	    }
	}
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree), TRUE);
    }
    return tree;
}

GuiObject GuiLabel(const char *label, GuiObject mnemonic, gfloat xalign)
{
    return LabelObject,
        "label", label,
        "xalign", xalign,
        "use-underline", true,
        "mnemonic-widget", mnemonic,
        End;
}

GuiObject GuiSimpleButton(const char *label, gboolean canDefault)
{
    return ButtonObject, 
        "label", label,
	"use-underline", true,
	"can_default", canDefault,
	End;
}

GuiObject GuiPopButton(const char *icon_name)
{
    GuiObject button = ButtonObject,
                "child", ImageObject,
		    "icon-name", icon_name, 
		    "icon-size", GTK_ICON_SIZE_MENU,
		    End,
		End;

    return button;
}

GuiObject GuiPopButtonSetCWD(const char *icon_name, struct Project *pr)
{
    GuiObject button = GuiPopButton(icon_name);
    g_signal_connect_swapped(button, "clicked", 
            G_CALLBACK(ProjectSetCurrentDir), pr);

    return button;
}

GuiObject GuiComboBox(const char * const *labels, gint active)
{
    GtkListStore *store = gtk_list_store_new(1, G_TYPE_STRING);
    GuiSetListStoreLabels(store, labels);

    GuiObject go = ComboBoxObject,
        "model", GTK_TREE_MODEL(store),
	"active", active,
	End;
    g_object_unref(G_OBJECT(store));

    GtkCellRenderer *cell = gtk_cell_renderer_text_new();
    // centering does not look really good, use default instead
    // g_object_set(G_OBJECT(cell), "xalign", 0.5, NULL);
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(go), cell, true);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(go), cell, "text", 0, NULL);

    return go;
}

const char *gtkEntryText(GuiObject go)
{
    return (go? gtk_entry_get_text(GTK_ENTRY(go)) : 0);
}

gboolean gtkToggleButtonActive(GuiObject go)
{
    return (go?  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(go)) : 0);
}

gint gtkComboBoxActive(GuiObject go)
{
    return (go? gtk_combo_box_get_active(GTK_COMBO_BOX(go)) : 0);
}

gint gtkRadioBoxActive(GuiObject go)
{
    return (go? gtk_radio_box_get_active(GTK_RADIO_BOX(go)) : 0);
}

gdouble gtkSpinButtonValue(GuiObject go)
{
    return (go? gtk_spin_button_get_value(GTK_SPIN_BUTTON(go)) : 0.0);
}

gdouble gtkRangeValue(GuiObject go)
{
    return (go? gtk_range_get_value(GTK_RANGE(go)) : 0.0);
}

static gboolean windowIsValid(GtkWindow *win)
{
    GList *toplevels = gtk_window_list_toplevels();
    gboolean valid = FALSE;

    for (GList *list = toplevels; list; list = list->next) {
        if (win == list->data) {
            valid = TRUE;
            break;
        }
    }
    g_list_free(toplevels);

    return valid;
}

#define NUM_LAST_ACTIVE_WINDOWS 16

static GtkWindow *lastActiveWindow[NUM_LAST_ACTIVE_WINDOWS] = { 0 };

static void setLastActiveWindow(GtkWindow *win)
{
    int i;

    for (i=0; i < NUM_LAST_ACTIVE_WINDOWS-1; i++) {
        if (lastActiveWindow[i] == win) {
            break;
        }
    }
    memmove(&lastActiveWindow[1], &lastActiveWindow[0], i*sizeof(GtkWindow *));
    lastActiveWindow[0] = win;
}

static GtkWindow *getLastActiveWindow(void)
{
    for (int i=0; i < NUM_LAST_ACTIVE_WINDOWS; i++) {
        GtkWindow *win = lastActiveWindow[i];
        if (win && windowIsValid(win) && 
                gtk_widget_get_visible(GTK_WIDGET(win))) {
            return win;
        }
    }
    return 0;
}

GtkWindow *GuiFindTransientWindow(GtkWidget *suggest)
{
    if (suggest) {
        suggest = gtk_widget_get_toplevel(suggest);
        if (GTK_IS_WINDOW(suggest)) {
            return GTK_WINDOW(suggest);
        }
    }
    return getLastActiveWindow();
}

static gboolean GuiWindowKeyPressCallback(GtkWidget *widget, 
	GdkEventKey *event, GtkWidget *cancel)
{
    if (event->keyval == GDK_KEY_Escape) {
	if (cancel && GTK_IS_BUTTON(cancel)) {
	    g_signal_emit_by_name(cancel, "activate");
	    return true;
	} 
     	gboolean dummy;
	g_signal_emit_by_name(widget, "delete-event", event, &dummy);
	return true;
    } else if (event->keyval == GDK_KEY_F1 || event->keyval == GDK_KEY_Help) {
        GuiShowHelp(NULL);
	return true;
    }
    return false;
}

void GuiSetDialogGeometryHints(GtkWidget *dialog, GtkWidget *group)
{
    GdkGeometry geometry;
    if (gtk_group_get_total_horiz_weight(GTK_GROUP(group)) > 0) {
	geometry.max_width = G_MAXSHORT;
    } else {
	geometry.max_width = -1;
    }
    if (gtk_group_get_total_vert_weight(GTK_GROUP(group)) > 0) {
	geometry.max_height = G_MAXSHORT;
    } else {
	geometry.max_height = -1;
    }
    gtk_window_set_geometry_hints(GTK_WINDOW(dialog), NULL, 
	    &geometry, GDK_HINT_MAX_SIZE);
}

void GuiSetDialogButtonOrder(GtkWidget *dialog)
{
    if (GTK_IS_DIALOG(dialog) && gtk_alternative_dialog_button_order(0)) {
        GtkWidget *bbox = gtk_dialog_get_action_area(GTK_DIALOG(dialog));
        gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_START);
        gtk_widget_set_direction(bbox, GTK_TEXT_DIR_RTL);
    }
}

void GuiIgnoreWindowDelete(GtkWidget *win)
{
    g_signal_connect(win, "delete-event", G_CALLBACK(gtk_true), 0);
#ifndef WINDOWS		/* on Windows, this gives bad results */
    gtk_window_set_deletable(GTK_WINDOW(win), false);
#endif
}

static gboolean onFocusInEvent(GtkWidget *w, GdkEvent *event, gpointer ud)
{
    setLastActiveWindow(GTK_WINDOW(w));
    return FALSE;
}

void GuiAddWindowCallbacks(GtkWidget *win, GtkWidget *cancel)
{
    g_signal_connect(win, "key-press-event", 
	    G_CALLBACK(GuiWindowKeyPressCallback), cancel);
    g_signal_connect(win, "focus-in-event", G_CALLBACK(onFocusInEvent), 0);
}

Project *GuiWindowProject(GuiObject go) 
{ 
    Project *pr = 0;

    if (go) {
	pr = (Project *)g_object_get_data(G_OBJECT(go), "project");
    }
    return pr;
}

void GuiSetWindowProject(GuiObject go, Project *pr)
{
    if (go) {
	g_object_set_data(G_OBJECT(go), "project", pr);
    }
}

#if 0
static gboolean entryOnKeyPressEvent(GtkWidget *widget, GdkEventKey *event,
	GtkWidget *ok)
{
    if (event->keyval == GDK_KEY_Return && ok) {
	gtk_widget_grab_focus(ok);
	return true;
    }
    return false;
}

void GuiAddEntryCallbacks(GtkWidget *entry, GtkWidget *ok)
{
    g_signal_connect(entry, "key-press-event", 
	    G_CALLBACK(entryOnKeyPressEvent), ok);
}
#else
void GuiAddEntryCallbacks(GtkWidget *entry, GtkWidget *ok)
{
    if (entry && ok) {
	g_signal_connect_swapped(entry, "activate", 
		G_CALLBACK(gtk_widget_grab_focus), ok);
    }
}
#endif

void GuiSetHelpNode(GtkWidget *w, const char *help)
{
    if (w) {
	g_object_set_data(G_OBJECT(w), "help", (void *)help);
    }
}

GuiObject GuiAnyOpenedWindow(void)
{
    GuiObject any = 0;

    GList *toplevels = gtk_window_list_toplevels();
    for (GList *list = toplevels; list; list = list->next) {
	GtkWidget *w = (GtkWidget *)list->data;
	if (GTK_IS_WINDOW(w) && gtk_widget_get_visible(w)) {
	    any = w;
	    break;
	}
    }
    g_list_free(toplevels);

    return any;
}

gchar *gtkEscapeStr(const gchar *str)
{
    gchar *out = 0;

    if (str) {
	gssize outlen = 0;
	const gchar *s;
	gchar *t;

	for (s = str; *s; s++) {
	    switch (*s) {
		case '&':
		    outlen += 5;
		    break;
		case '<':
		case '>':
		    outlen += 4;
		    break;
		default:
		    outlen++;
		    break;
	    }
	}

	t = out = g_new(gchar, outlen + 1);
	for (s = str; *s; s++) {
	    switch (*s) {
		case '&':
		    strcpy(t, "&amp;");
		    t += 5;
		    break;
		case '<':
		    strcpy(t, "&lt;");
		    t += 4;
		    break;
		case '>':
		    strcpy(t, "&gt;");
		    t += 4;
		    break;
		default:
		    *t++ = *s;
		    break;
	    }
	}
	*t = 0;
    }
    return out;
}

gchar *GuiUnderline(const char *str, cschar shortcut, gboolean &use)
{
    gchar *result;
    
    if (str) {
	int len = strlen(str);
	cschar usc = CodesetToUpper(shortcut);
	const gchar *s = str, *n;
	gchar *t = result = g_new(gchar, len + 1 + 1);

	while (*s) {
	    n = CodesetNextChar(s);
	    if (usc && CodesetToUpper(CodesetGetChar(s)) == usc) {
		*t++ = '_';
		usc = 0;
	    }
	    if (*s == '_') {
		*t++ = '-';
		s = n;
	    } else {
		while (s != n) {
		    *t++ = *s++;
		}
	    }
	}
	*t = 0;

	if (t - result > len) {	// found?
	    use = true;
	} else {
	    use = false;
	}
    } else {
	result = 0;
	use = false;
    }

    return result;
}

gchar *gtkExpandPathname(const gchar *path)
{
    struct MemPool *memPool = CreatePool(0, 4096, 1024);
    const char *tmp = ExpandPathname(path, memPool);
    gchar *expanded;

    if (!tmp) {
	expanded = 0;
    } else if (g_path_is_absolute(tmp)) {
	expanded = g_strdup(tmp);
    } else {
	gchar *cwd = g_get_current_dir();
	expanded = g_build_filename(cwd, tmp, NULL);
	g_free(cwd);
    }
    DeletePool(memPool);

#ifdef WINDOWS
    if (expanded) {
	for (char *s = expanded; *s; s++) {
	    if (*s == '/') {
		*s = '\\';
	    }
	}
    }
#endif
    return expanded;
}

gchar *gtkExpandPathnameGetAssign(const gchar *path, 
	gchar **assignName, gchar **assignPath)
{
    gchar *fullPath = gtkExpandPathname(path);
    int len = path? strlen(path) : 0;
    int fullLen = fullPath? strlen(fullPath) : 0;
    while (len > 0 && fullLen > 0) {
        if (path[len - 1] != fullPath[fullLen - 1]) {
            break;
        }
        len--;
        fullLen--;
    }
    // len == 0 indicates that cwd has been prepended to path.
    if (assignName) {
        *assignName = g_strndup(path, len);
    }
    if (assignPath) {
        *assignPath = g_strndup(fullPath, fullLen);
    }
    /* printf("assign(%s) = %s\n", *assignName, *assignPath); */

    return fullPath;
}

gchar *gtkUnexpandPathname(const gchar *fullPath, 
	const gchar *assignName, const gchar *assignPath)
{
    int assignLen = assignPath? strlen(assignPath) : 0;
    gchar *path;

    if (!fullPath) {
	path = 0;
    } else if (assignLen == 0 || strncmp(assignPath, fullPath, assignLen)) {
	path = g_strdup(fullPath);
    } else if (assignName[0] != 0) {
        path = g_strjoin(0, assignName, fullPath + assignLen, NULL);
    } else {
	/* current dir assign */
	while (IsPathSeparator(fullPath[assignLen])) {
	    assignLen++;
	}
	path = g_strdup(fullPath + assignLen);
    }
    
    return path;
}

#if 0
gchar *gtkSplitPathname(const gchar *path, gchar **dirName)
{
    struct MemPool *memPool = CreatePool(0, 4096, 1024);
    char *dir, *file;
    if (!SplitPathname(path, memPool, &dir, &file)) {
	file = (char *)path;
	dir = 0;
    }
    gchar *fileName = g_strdup(file);
    if (dirName) {
	*dirName = g_strdup(dir);
    }
    DeletePool(memPool);

    return fileName;
}

gchar *gtkDirname(const gchar *path)
{
    gchar *dir = 0;
    if (path && FileStat(path) != 0) {
	gchar *filename = gtkSplitPathname(path, &dir);
	g_free(filename);
    } else {
	dir = strdup(path);
    }
    int len = dir? strlen(dir) : 0;
    if (len > 0 && dir[len-1] == '/') {
	dir[len-1] = 0;
    }
    return dir;
}

gchar *gtkFilename(const gchar *path)
{
    if (path && FileStat(path) == 0) {
	return 0;
    }
    return gtkSplitPathname(path, 0);
}
#endif

