/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkcodeset.c
**
** Codeset stands for the default encoding of a character string.
** On glib/gtk systems, the default encoding is UTF8.
**
** Creation date: 25.04.2007
** Last modified: 10.02.2013
*/

#include <codeset.h>
#include <strhandle.h>
#include <memory.h>
#include <util.h>
#include <error.h>
#include <glib.h>

static const char Default8BitCodeset[] = "ISO-8859-1";
static const char TargetCodeset[] = "UTF-8";

const char *CodesetInit(void)
{
    return NULL;
}

void CodesetExit(void)
{
}

bool LocaleIsUTF8(void)
{
    const char *charset;

    return g_get_charset(&charset);
}

bool StrIsCodeset(const char *str)
{
    return (str == NULL || g_utf8_validate(str, -1, NULL));
}

bool StrIsUTF8AllowTrail(const char *str, const char **pend)
{
    const char *end = NULL;
    bool rc;
   
    if (str) {
        rc = g_utf8_validate(str, -1, &end);
        if (!rc && g_utf8_get_char_validated(end, -1) == (gunichar)-2) {
            rc = TRUE;
        }
    } else {
        rc = TRUE;
    }

    if (pend) {
        *pend = end;
    }
    return rc;
}

#ifdef WINDOWS
#define UTF8_REPLACE_CHAR "?"       /* the unicode one does not look good */
#else
#define UTF8_REPLACE_CHAR "\xEF\xBF\xBD"
#endif

static char *CodesetConvert(const char *in, struct MemPool *memPool,
        const char *to, const char *from, bool ascii, bool allowTrail,
        const char **pend, long *perr, size_t *pwarn)
{
    long err = OK;
    size_t warn = 0;
    char *str = NULL;
    const char *end = NULL;

    if (in == NULL) {
        err = ERR_INTERNAL;
    } else if (ascii && StrIsASCII(in)) {
        str = (char *)in;
        end = in + strlen(in);
    } else if (to == NULL && from == NULL && g_utf8_validate(in, -1, &end)) {
        str = (char *)in;
    } else {
        GError *error = NULL;
        gssize len = strlen(in);
        gsize read;
        gchar *text, *next;
        const char *replace = "?";
        bool utf8 = FALSE;

        end = in + len;
        if (to == NULL) {
            to = TargetCodeset;
            replace = UTF8_REPLACE_CHAR;
        }
        if (from == NULL) {
            from = TargetCodeset;
            utf8 = TRUE;
        }
        text = g_convert(in, len, to, from, &read, NULL, &error);
        if (error == NULL) {
            if (read == len || allowTrail) {
                str = AllocString(memPool, text);
                if (str == NULL) {
                    err = ERR_MEM;
                }
                end = in + read;
            } else {
                str = PoolAllocMem(memPool, strlen(text) + strlen(replace) + 1);
                if (str == NULL) {
                    err = ERR_MEM;
                } else {
                    strcpy(str, text);
                    strcat(str, replace);
                    warn++;
                }
            }
            g_free(text);
        } else if (error->code != G_CONVERT_ERROR_ILLEGAL_SEQUENCE) {
            err = CODESETERR_CONVERSION;
            end = in + read;
        } else {
            struct StrHandle *sh = AllocStrHandle(memPool);

            while (len > 0) {
                while (read == 0) {
                    StrHandlePutS(sh, replace);
                    warn++;
                    if (utf8) {
                        next = g_utf8_find_next_char(in, NULL);
                        if (next) {
                            len -= next - in;
                            in = next;
                        } else {
                            in += len;
                            len = 0;
                        }
                    } else {
                        in++;           /* skip one byte */
                        len--;
                    }
                    if (len == 0) {
                        break;
                    }
                    text = g_convert(in, len, to, from, &read, NULL, NULL);
                }
                if (text != NULL) {
                    StrHandlePutS(sh, text);
                    g_free(text);
                    text = NULL;
                    if (allowTrail && read != len) {
                        StrHandlePutS(sh, UTF8_REPLACE_CHAR);
                        warn++;
                    } else {
                        end = in + read;
                    }
                    break;                      /* complete */
                } else if (read > 0) {
                    text = g_convert(in, read, to, from, &read, NULL, NULL);
                    StrHandlePutS(sh, text);
                    g_free(text);
                    text = NULL;
                    in += read;
                    len -= read;
                }
            }
            StrHandlePutC(sh, 0);

            if (StrHandleError(sh)) {
                err = ERR_MEM;
            } else {
                str = AllocString(memPool, StrHandleString(sh));
                if (str == NULL) {
                    err = ERR_MEM;
                }
            }
            FreeStrHandle(sh);
        }
    }

    if (pend) {
        *pend = end;
    }
    if (perr) {
        *perr = err;
    }
    if (pwarn) {
        *pwarn = warn;
    }
    return str;
}

char *CodesetConvertFromUTF8(const char *in, struct MemPool *memPool, 
        long *perr, size_t *pwarn)
{
    return CodesetConvert(in, memPool, NULL, NULL, TRUE, FALSE, 
            NULL, perr, pwarn);
}

char *CodesetConvertFromUTF8AllowTrail(const char *in, struct MemPool *memPool, 
        long *perr, size_t *pwarn, const char **end)
{
    return CodesetConvert(in, memPool, NULL, NULL, TRUE, TRUE,
            end, perr, pwarn);
}

char *CodesetConvertFromLocale(const char *in, struct MemPool *memPool, 
        long *perr, size_t *pwarn)
{
    const char *codeset;
    bool ascii = FALSE;

    if (g_get_charset(&codeset)) {
        codeset = NULL;
        ascii = TRUE;
    }
    return CodesetConvert(in, memPool, NULL, codeset, ascii, FALSE,
            NULL, perr, pwarn);
}

char *CodesetConvertFromLocaleAllowTrail(const char *in,struct MemPool *memPool,
        long *perr, size_t *pwarn, const char **end)
{
    const char *codeset;
    bool ascii = FALSE;

    if (g_get_charset(&codeset)) {
        codeset = NULL;
        ascii = TRUE;
    }
    return CodesetConvert(in, memPool, NULL, codeset, ascii, TRUE,
            end, perr, pwarn);
}

char *CodesetConvertFromDefault8Bit(const char *in, struct MemPool *memPool, 
        long *perr)
{
    return CodesetConvert(in, memPool, NULL, Default8BitCodeset, TRUE, FALSE,
            NULL, perr, NULL);
}

char *CodesetConvertToUTF8(const char *in, struct MemPool *memPool, 
        long *perr)
{
    return CodesetConvert(in, memPool, NULL, NULL, TRUE, FALSE,
            NULL, perr, NULL);
}

char *CodesetConvertToLocale(const char *in, struct MemPool *memPool, 
        long *perr, size_t *pwarn)
{
    const char *codeset;
    bool ascii = FALSE;

    if (g_get_charset(&codeset)) {
        codeset = NULL;
        ascii = TRUE;
    }
    return CodesetConvert(in, memPool, codeset, NULL, ascii, FALSE,
            NULL, perr, pwarn);
}

char *CodesetConvertToDefault8Bit(const char *in, struct MemPool *memPool, 
        long *perr, size_t *pwarn)
{
    return CodesetConvert(in, memPool, Default8BitCodeset, NULL, TRUE, FALSE,
            NULL, perr, pwarn);
}

size_t CodesetUTF8Len(const char *s)
{
    return (s? strlen(s) : 0);
}

void CodesetWriteUTF8(const char *s, FILE *file)
{
    if (s && file) {
	fputs(s, file);
    }
}

short CodesetCharToStr(cschar c, char *out)
{
    return g_unichar_to_utf8(c, out);
}

cschar CodesetGetChar(const char *s)
{
    return g_utf8_get_char(s);
}

cschar CodesetGetCharCheck(const char *s)
{
    return (s? g_utf8_get_char(s) : 0);
}

char *CodesetNextChar(const char *s)
{
    return g_utf8_next_char(s);
}

size_t CodesetStrLen(const char *str)
{
    return g_utf8_strlen(str, -1);
}

size_t CodesetNCharsToNBytes(const char *s, size_t n)
{
    const gchar *pos = g_utf8_offset_to_pointer(s, n);

    return (pos - s);
}

size_t CodesetNBytesToNChars(const char *s, size_t n)
{
    return g_utf8_pointer_to_offset(s, s + n);
}

bool CodesetIsAlNum(cschar c)
{
    return g_unichar_isalnum(c);
}

bool CodesetIsDigit(cschar c)
{
    return g_unichar_isdigit(c);
}

bool CodesetIsLower(cschar c)
{
    return g_unichar_islower(c);
}

bool CodesetIsSpace(cschar c)
{
    return g_unichar_isspace(c);
}

bool CodesetIsBrkSpace(cschar c)
{
    return (g_unichar_isspace(c) && 
	    g_unichar_break_type(c) != G_UNICODE_BREAK_NON_BREAKING_GLUE);
}

bool CodesetIsXDigit(cschar c)
{
    return g_unichar_isxdigit(c);
}

short CodesetDigitValue(cschar c)
{
    return g_unichar_digit_value(c);
}

short CodesetXDigitValue(cschar c)
{
    return g_unichar_xdigit_value(c);
}

cschar CodesetToUpper(cschar c)
{
    return g_unichar_toupper(c);
}

cschar CodesetToLower(cschar c)
{
    return g_unichar_tolower(c);
}

char *CodesetStrUp(const char *in, struct MemPool *memPool)
{
    gchar *str = g_utf8_strup(in, -1);
    char *out = AllocString(memPool, str);
    
    g_free(str);
    return out;
}

char *CodesetStrDown(const char *in, struct MemPool *memPool)
{
    gchar *str = g_utf8_strdown(in, -1);
    char *out = AllocString(memPool, str);
    
    g_free(str);
    return out;
}

char *CodesetStrNBytesCopy(char *dst, const char *src, size_t nbytes)
{
    const gchar *end = NULL;

    strncpy(dst, EmptyNullStr(src), nbytes);
    dst[nbytes] = 0;
    if (!g_utf8_validate(dst, -1, &end) && end) {
	*(char *)end = 0;
    }

    return dst;
}

char *CodesetTrimStrChars(char *in, const char *front, const char *back)
{
   char *s;

   /* trailing chars */
   s = in + strlen(in);
   if (back) {
       while ((s = g_utf8_find_prev_char(in, s)) >= in) {
	   if (g_utf8_strchr(back, -1, g_utf8_get_char(s)) == NULL) {
	       break;
	   }
	   *s = 0;
       }
   } else {
       while ((s = g_utf8_find_prev_char(in, s)) >= in) {
	   if (!g_unichar_isspace(g_utf8_get_char(s))) {
	       break;
	   }
	   *s = 0;
       }
   }

   /* leading chars */
   s = in;
   if (front) {
       while (*s && g_utf8_strchr(front, -1, g_utf8_get_char(s)) != NULL) {
	   s = g_utf8_next_char(s);
       }
   } else {
       while (*s && g_unichar_isspace(g_utf8_get_char(s))) {
	   s = g_utf8_next_char(s);
       }
   }
   if (s > in) {
	strcpy(in, s);
   }
   return in;
}

char *CodesetTrimStr(char *s)
{
    return CodesetTrimStrChars(s, NULL, NULL);
}

long CodesetStrNCollate(const char *s1, const char *s2, size_t n, short collate)
{
    gchar *t1 = g_utf8_casefold(s1, n);
    gchar *t2 = g_utf8_casefold(s2, n);
    long result = g_utf8_collate(t1, t2);

    g_free(t2);
    g_free(t1);
    return result;
}

long CodesetStrCollate(const char *s1, const char *s2, short collate)
{
    return CodesetStrNCollate(s1, s2, -1, collate);
}

long CodesetStrCmp(const char *s1, const char *s2)
{
    return CodesetStrCollate(s1, s2, CS_COLLATE_SIMPLE);
}

char *CodesetStrStr(const char *haystack, const char *needle)
{
    gchar *cfh = g_utf8_casefold(haystack, -1);
    gchar *cfn = g_utf8_casefold(needle, -1);
    char *pos = strstr(cfh, cfn);
    glong offset;
    char *result = NULL;

    if (pos) {
	offset = g_utf8_pointer_to_offset(cfh, pos);
	result = g_utf8_offset_to_pointer(haystack, offset);
    }
    g_free(cfn);
    g_free(cfh);

    return result;
}

char *CodesetStrpbrk(const char *s, const char *accept)
{
    for (; *s; s = g_utf8_next_char(s)) {
	if (g_utf8_strchr(accept, -1, g_utf8_get_char(s))) {
	    return (char *)s;
	}
    }
    return NULL;
}

char *CodesetStrPBrk(const char *s, const char *accept)
{
    gchar *cfs = g_utf8_casefold(s, -1);
    gchar *cfa = g_utf8_casefold(accept, -1);
    char *pos = CodesetStrpbrk(cfs, cfa);
    glong offset;
    char *result = NULL;

    if (pos) {
	offset = g_utf8_pointer_to_offset(cfs, pos);
	result = g_utf8_offset_to_pointer(s, offset);
    }
    g_free(cfa);
    g_free(cfs);

    return result;
}

char *CodesetStrchr(const char *s, cschar c)
{
    return g_utf8_strchr(s, -1, c);
}

char *CodesetStrrchr(const char *s, cschar c)
{
    return g_utf8_strrchr(s, -1, c);
}

