/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** gtkdosutil.c
**
** help routines for dos -- using glib
**
** Creation date: 09.10.2006
** Last modified: 01.09.2009
*/

#include "system.h"
#include "dosutil.h"
#include "requesters.h"
#include <glib.h>
#include <glib/gstdio.h>
#include <unistd.h>
#if (defined UNIX) || (defined MACOSX)
#include <sys/wait.h>		/* WEXITSTATUS */
#elif (defined WINDOWS)
#include <codeset.h>
#include <windows.h>
#include <windef.h>
#include <wingdi.h>
#include <winuser.h>
#include <shellapi.h>
#include <errno.h>
#include <wchar.h>
#define WEXITSTATUS(x) x
#endif

static char *pathnameFromUTF8(const char *utf8path)
{
    return g_filename_from_utf8(utf8path, -1, 0, 0, 0);
}

static int dosCall(int (*func)(const char *), const char *pathnameUTF8)
{
    char *pathname = pathnameFromUTF8(pathnameUTF8);
    int rc = -1;

    if (pathname) {
	rc = func(pathname);
	g_free(pathname);
    }

    return rc;
}

const char *GetTmpDir(void)
{
    return g_get_tmp_dir();
}

FILE *FOpen(const char *filenameUTF8, const char *mode)
{
    char *filename = pathnameFromUTF8(filenameUTF8);
    FILE *file = NULL;
    
    if (filename) {
#ifdef WINDOWS
        // On Fedora21 the g_fopen in wmingw-glib2-2.42.2-2.fc21 is broken by
        // 0001-Use-CreateFile-on-Win32-to-make-sure-g_unlink-always.patch
        // and g_fopen(..., "r") does not open the file in text mode.
        // We therefore 'revert' this patch using the original code.
        wchar_t *wfilename = g_utf8_to_utf16(filename, -1, NULL, NULL, NULL);
        wchar_t *wmode = g_utf8_to_utf16(mode, -1, NULL, NULL, NULL);

        if (wfilename == NULL || wmode == NULL) {
            errno = EINVAL;
        } else {
            file = _wfopen(wfilename, wmode);
        }
        g_free (wmode);
        g_free (wfilename);
#else
	file = g_fopen(filename, mode);
#endif
	g_free(filename);
    }

    return file;
}

static int mkdir_default_mode(const char *dirname)
{
    return g_mkdir(dirname, 0777);
}

int MkDir(const char *dirnameUTF8)
{
    return dosCall(mkdir_default_mode, dirnameUTF8);
}

int RmDir(const char *dirnameUTF8)
{
    return dosCall(g_rmdir, dirnameUTF8);
}

int RmFile(const char *filenameUTF8)
{
    return dosCall(g_unlink, filenameUTF8);
}

#if 0
char *GetCwd(char *buf, size_t size)
{
    char *cwd = g_get_current_dir();
    char *cwdUTF8 = g_filename_to_utf8(cwd, -1, 0, 0, 0);
    char *result = NULL;

    if (cwdUTF8 && strlen(cwdUTF8) < size) {
	strcpy(buf, cwdUTF8);
	result = buf;
    }

    g_free(cwdUTF8);
    g_free(cwd);

    return result;
}
#endif

#if (GLIB_MAJOR_VERSION == 2 && GLIB_MINOR_VERSION < 8)

/**
 * g_chdir: 
 * @path: a pathname in the GLib file name encoding (UTF-8 on Windows)
 *
 * A wrapper for the POSIX chdir() function. The function changes the
 * current directory of the process to @path.
 * 
 * See your C library manual for more details about chdir().
 *
 * Returns: 0 on success, -1 if an error occurred.
 * 
 * Since: 2.8
 */
int
g_chdir (const gchar *path)
{
#ifdef WINDOWS
  wchar_t *wpath = g_utf8_to_utf16 (path, -1, NULL, NULL, NULL);
  int retval;
  int save_errno;

  if (wpath == NULL)
    {
      errno = EINVAL;
      return -1;
    }

  retval = _wchdir (wpath);
  save_errno = errno;

  g_free (wpath);
    
  errno = save_errno;
  return retval;
#else
  return chdir (path);
#endif
}
#endif

int PushDir(const char *dirnameUTF8, Cwd *cwd)
{
    int rc;
    gboolean allocated = FALSE;

    g_return_val_if_fail(cwd != NULL && dirnameUTF8 != NULL, -1);

    if (dirnameUTF8[0] == 0 || !strcmp(dirnameUTF8, ".")) {
	return 0;	/* nothing to do */
    }

    if (*cwd == NULL) {
	*cwd = (Cwd)g_get_current_dir();
	allocated = TRUE;
    }
    rc = dosCall(g_chdir, dirnameUTF8);
    if (rc != 0 && allocated) {
	g_free(*cwd);
	*cwd = NULL;
    }
    return rc;
}

int PopDir(Cwd *cwd)
{
    int rc = 0;

    g_return_val_if_fail(cwd != NULL, -1);

    if (*cwd) {
	rc = g_chdir((const char *)*cwd);
	g_free(*cwd);
	*cwd = NULL;
    }
    return rc;
}

int RenamePath(const char *oldPathUTF8, const char *newPathUTF8)
{
    char *oldPath = pathnameFromUTF8(oldPathUTF8);
    char *newPath = pathnameFromUTF8(newPathUTF8);
    int rc = -1;

    if (oldPath && newPath) {
	rc = g_rename(oldPath, newPath);
    }
    g_free(newPath);
    g_free(oldPath);

    return rc;
}

Dir *DirOpen(const char *filename)
{
    return (Dir *)g_dir_open(filename, 0, NULL);
}

const char *DirReadName(Dir *dir)
{
    return g_dir_read_name((GDir *)dir);
}

void DirClose(Dir *dir)
{
    g_dir_close((GDir *)dir);
}

struct DosProcess
{
    guint tag;
    GPid pid;
    gboolean exited;
    int rc;
    gboolean destroyOnExit;
};

static void childWatchFunc(GPid pid, gint status, gpointer data)
{
    struct DosProcess *process = (struct DosProcess *)data;

    process->rc = WEXITSTATUS(status);
    process->exited = TRUE;

    if (process->destroyOnExit) {
        // printf("destroying dos process\n");
        CloseDosProcess(process);
    }
}

struct DosProcess *SpawnDosProcess(const char *cmdline, bool useShell, 
        bool needRC)
{
    struct DosProcess *process = NULL;

#if (defined UNIX) || (defined MACOSX)
    gint argc;
    gchar **argv = NULL, *shell[4], **mem = NULL;
    GPid pid;

    if (cmdline == NULL) {
        return NULL;
    }

    // printf("SpawnDosProcess: <%s>\n", cmdline);

    if (useShell) {
        shell[0] = "/bin/sh";
        shell[1] = "-c";
        shell[2] = (char *)cmdline;
        shell[3] = NULL;
        argc = 3;
        argv = shell;
    } else if (g_shell_parse_argv(cmdline, &argc, &mem, NULL)) {
        argv = mem;
    } 

    if (argv) {
        if (g_spawn_async(NULL, argv, NULL, 
                    G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD,
                NULL, NULL, &pid, NULL)) {
            process = g_new0(struct DosProcess, 1);
            process->tag = g_child_watch_add(pid, childWatchFunc, process);
            process->pid = pid;
            process->rc = -1;

            // printf("pid: %d, tag: %d, source: %p\n", pid, process->tag,
            //         g_main_context_find_source_by_id(NULL, process->tag));
        }
    }

    g_strfreev(mem);

#elif (defined WINDOWS)
    SHELLEXECUTEINFO sei;
    const char *s, *t;
    char *filename;
    bool rc;

    if (cmdline == NULL) {
        return NULL;
    }

    s = cmdline;
    while(CodesetIsBrkSpace(CodesetGetChar(s))) {
	s = CodesetNextChar(s);
    }
    if (*s == '"') {
        s++;
	t = CodesetStrchr(s, '"');
    } else {
	t = s;
	while(*t && !CodesetIsBrkSpace(CodesetGetChar(t))) {
	    t = CodesetNextChar(t);
	}
    }
    if (t) {
	filename = g_strndup(s, t - s);
        if (*t == '"') {
            t++;
        }
	while(CodesetIsBrkSpace(CodesetGetChar(t))) {
	    t = CodesetNextChar(t);
	}
	if (*t == 0) {
	    t = 0;
	}
    } else {
	filename = g_strdup(s);
    }
	
    memset(&sei, 0, sizeof(sei));
    sei.cbSize = sizeof(sei);
    sei.fMask = SEE_MASK_NOCLOSEPROCESS | SEE_MASK_FLAG_NO_UI;
    sei.hwnd = HWND_TOP;
    sei.lpVerb = "open";
    sei.lpFile = filename;
    sei.lpParameters = t;
    sei.nShow = SW_SHOWNORMAL;

    rc = ShellExecuteEx(&sei);
    g_free(filename);
    if (rc) {
	GPid pid = sei.hProcess;
	// printf("pid: %p\n", pid);
	process = g_new0(struct DosProcess, 1);
	if (pid) {
	    process->tag = g_child_watch_add(pid, childWatchFunc, process);
	    process->pid = pid;
	    process->rc = -1;
	} else {
	    process->rc = 0;		/* success */
	    process->exited = TRUE;
	}
    }

#endif

    return process;
}

bool DosProcessIsExited(struct DosProcess *process)
{
    return (process? process->exited : TRUE);
}

int DosProcessExitCode(struct DosProcess *process)
{
    return (process? process->rc : -1);
}

void CloseDosProcess(struct DosProcess *process)
{
    if (process) {
        if (process->exited) {
	    if (process->tag) {
		g_source_remove(process->tag);
	    }
            g_spawn_close_pid(process->pid);
            g_free(process);
        } else {
            process->destroyOnExit = TRUE;
        }
    }
}

