/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** windowsutil.c
**
** help routines for Windows
**
** Creation date: 14.11.1997
** Last modified: 12.06.2017
*/

#include "system.h"
#include "dosutil.h"
#include "util.h"
#include "memory.h"
#include "codeset.h"
#include <windows.h>
#include <glib.h>

static char *pathnameFromUTF8(const char *utf8path)
{
    return g_filename_from_utf8(utf8path, -1, 0, 0, 0);
}

static void freePathname(char *pathname)
{
    g_free(pathname);
}

bool SameFile(const char *filename1, const char *filename2)
{
    bool result = FALSE;
    if (filename1 && filename2) {
	if (strcmp(filename1, filename2) == 0) {
	    result = TRUE;
	}
    }
	
    return result;
}

bool FileLocked(const char *filename)
{
    bool isLocked = FALSE;
    char *winFilename;
    HANDLE h;

    if (filename == NULL) {
        return FALSE;
    }

    winFilename = pathnameFromUTF8(filename);
    h = CreateFile(winFilename, GENERIC_READ, FILE_SHARE_READ, NULL,
            OPEN_EXISTING, FILE_FIELD_NORMAL, NULL);
    freePathname(winFilename);
    if (h == INVALID_HANDLE_VALUE) {
        DWORD error = GetLastError();
        isLocked = error == ERROR_SHARING_VIOLATION;
    } else {
        CloseHandle(h);
    }

    return isLocked;
}

static struct Lock *fileLock(const char *filename, bool exclusive,
	bool *pIsLocked)
{
    struct Lock *lock = NULL;
    bool isLocked = FALSE;

    if (filename == NULL) {
        isLocked = FALSE;
    } else {
        char *winFilename = pathnameFromUTF8(filename);
        HANDLE h = CreateFile(winFilename,
        	exclusive ? GENERIC_READ | GENERIC_WRITE : GENERIC_READ,
        	exclusive ? 0 : FILE_SHARE_READ, NULL,
        	exclusive ? CREATE_ALWAYS : OPEN_ALWAYS,
        	FILE_FIELD_NORMAL, NULL);
        freePathname(winFilename);
        if (h == INVALID_HANDLE_VALUE) {
            DWORD error = GetLastError();
            if (error == ERROR_SHARING_VIOLATION) {
                isLocked = TRUE;
            }
        } else {
            lock = (struct Lock *)h;
        }
    }

    if (pIsLocked != NULL) {
        *pIsLocked = isLocked;
    }
    return lock;
}

struct Lock *FileLockExclusive(const char *filename, bool *isLocked)
{
    return fileLock(filename, TRUE, isLocked);
}

struct Lock *FileLockShared(const char *filename)
{
    return fileLock(filename, FALSE, NULL);
}

void FileUnlock(struct Lock *lock)
{
    if (lock != NULL) {
        CloseHandle((HANDLE)lock);
    }
}

short FileStat(const char *filename)
{
    char *winFilename;
    struct stat buf;
    int rc;

    if (filename == NULL) {
    	return -1;
    }

    winFilename = pathnameFromUTF8(filename);
    rc = stat(winFilename, &buf);
    freePathname(winFilename);
    if (rc != 0) {
	return -1;
    } 
    
    if (buf.st_mode & S_IFDIR) {
	return 0;
    } 

    return 1;
}

time_t FileModificationTime(const char *filename)
{
    char *winFilename;
    struct stat buf;
    int rc;

    if (filename == NULL) {
        return TIME_T_INVALID;
    }

    winFilename = pathnameFromUTF8(filename);
    rc = stat(winFilename, &buf);
    freePathname(winFilename);
    if (rc != 0) {
        return TIME_T_INVALID;
    } 
    
    return buf.st_mtime;
}

bool FileIsWritable(const char *filename)
{
    char *winFilename;
    HANDLE h;

    winFilename = pathnameFromUTF8(filename);
    h = CreateFile(winFilename, GENERIC_WRITE,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
            FILE_FIELD_NORMAL, NULL);
    freePathname(winFilename);

    if (h == INVALID_HANDLE_VALUE) {
        return FALSE;
    }
    CloseHandle(h);
    return TRUE;
}

bool IsPathSeparator(int c)
{
    return (c == '/' || c == '\\');
}

bool SplitPathname(const char *path, struct MemPool *memPool, 
        char **pdir, char **pfile)
{
    const char *s, *s2;

    if(pfile != NULL)
        *pfile = NULL;
    if(pdir != NULL)
        *pdir = NULL;

    if(path == NULL)
        return(TRUE);                   /* nothing to do */

    s = CodesetStrrchr(path, '\\');
    s2 = CodesetStrrchr(path, '/');
    if (s2 > s) {
	s = s2;
    }
    if (s != NULL)
	s++;
    else
	s = path;

    if(pdir != NULL)
    {
        long len = s - path;

	if(len > 0 && (*(s - 1) == '\\' || *(s - 1) == '/'))
	    len--;
        *pdir = AllocStringN(memPool, path, len);
        if(*pdir == NULL)
            return(FALSE);
    }

    if(pfile != NULL)
    {
        *pfile = AllocString(memPool, s);
        if(*pfile == NULL)
        {
            if(pdir != NULL && *pdir != NULL)
            {
                FreeString(memPool, *pdir);
                *pdir = NULL;
            }
            return(FALSE);
        }
    }

    return(TRUE);
}

char *JoinPathname(struct MemPool *memPool, const char *dir, const char *file)
{
    const char *mid = ""; 
    char *s;
    long dirlen, filelen, i;

    dir = EmptyNullStr(dir);
    file = EmptyNullStr(file);

    dirlen = strlen(dir);
    filelen = strlen(file);

    if (filelen > 0 && dirlen > 0) {
        mid = "\\";
        if (dir[dirlen-1] == ':') {
            mid = "";
            for (i = 0; i < dirlen-1; i++) {
                if (dir[i] == ':' || IsPathSeparator(dir[i])) {
                    mid = "\\";
                    break;
                }
            }
        }
    }
    if (dirlen > 0 && IsPathSeparator(dir[dirlen-1])) {
	dirlen--;
        if (dirlen == 0 && filelen == 0) {
	    mid = "\\";
        }
    }

    s = PoolAllocMem(memPool, dirlen + strlen(mid) + filelen + 1);
    if (s == NULL) {
        return NULL;
    }

    sprintf(s, "%.*s%s%s", (int)dirlen, dir, mid, file);
    return s;
}

