/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** compile.c
**
** Creation date: 16.04.1995
** Last modified: 06.06.2021
*/

#include "program.h"
#include "programbin.h"
#include "parser.h"
#include "compile_impl.h"
#include "functable.h"
#include "global.h"
#include "project.h"
#include "field.h"
#include "table.h"
#include "preferences.h"
#include "name.h"
#include "memory.h"
#include "error.h"
#include "requesters.h"
#include "appstrings_cat.h"
#include <setjmp.h>

static long errLine, errCol;
static const char *errFilename = NULL;
static const char *errString = NULL;
static jmp_buf errJmpBuf;
static struct Table *compTable = NULL;
static bool linkProjectProgram = FALSE;
static struct LocalIdent *curFunc = NULL;

struct Project *compProject = NULL;

struct LocalIdent *FindLocalIdentName(struct LocalIdent *defList, char *name)
{
    for(; defList; defList = defList->Next)
        if(!strcmp(defList->Name, name))
            break;
    return(defList);
}

short CountAppArgs(struct AppArgs *aa)
{
    short num = 0;

    for(; aa; aa = aa->Next)
        num++;
    return(num);
}

bool CheckArgTypes(struct AppArgs *aa, const char *types)
{
    for (; aa; aa = aa->Next, types++) {
        if (*types == TYPE_FIELD || *types == TYPE_TABLE) {
            struct RelFieldPath *rap = NULL;
	    struct LocalExtend *le;

            switch (aa->Expr->Case)
            {
		case EXPR_CASE_GLOBALIDENT:
                    if (aa->Expr->Contents.GlobalIdent->Type == *types) {
                        rap = &aa->Expr->Contents.GlobalIdent->RelFieldPath;
		    }
                    break;

                case EXPR_CASE_LOCALEXTEND:
                    le = aa->Expr->Contents.LocalExtend;
                    if (le->GlobalIdent->Type == *types) {
                        rap = &le->GlobalIdent->RelFieldPath;
		    }
                    break;
            }

            if (rap != NULL) {
                if (rap->rap_NumFields > 0) {
                    aa->Expr->Type = TYPE_FIELD;
		} else {
                    aa->Expr->Type = TYPE_TABLE;
		}
            }
        }
        if (*types == TYPE_BOOL) {
            continue;
	}
        if (GenType(*types, aa->Expr->Type) != NOTYPE) {
            continue;
	}

        return FALSE;
    }
    return TRUE;
}

static void setError(const char *err, struct LineCol lc)
{
    errString = err;
    errLine = lc.Line;
    errCol = lc.Col;
    errFilename = lc.Filename;

    longjmp(errJmpBuf, 1);
}

static bool IsOrderType(short type)
{
    switch(type)
    {
        case TYPE_INTEGER:
        case TYPE_REAL:
        case TYPE_BOOL:
        case TYPE_STRING:
        case TYPE_MEMO:
        case TYPE_DATE:
        case TYPE_TIME:
        case TYPE_NIL:
        case TYPE_RECORD:
        case TYPE_UNKNOWN:      return(TRUE);
                                break;
    }
    return(FALSE);
}

static void checkTypes(short type1, short type2, struct LineCol lc)
{
    if(GenType(type1, type2) == NOTYPE)
        setError(GCS(MSG_ProgErr_Type), lc);
}

static void checkApplyFunc(struct Func *func, struct AppArgs *aa,
        bool nosideeffect, bool allowloopcontrol, bool allowreturn)
{
    if (func) {
        struct FuncInfo *fi = func->FuncInfo;
        const char *err = NULL;
        struct AppArgs *a;
        long arity = 0;

        for (a = aa; a; a = a->Next) {
	    arity++;
            if (a->Expr && a->Expr->Case == EXPR_CASE_LOCALIDENT) {
                if (a->Expr->Contents.LocalIdent->Arity >= 0) {
                    a->Expr->Type = TYPE_FUNCTION;
		}
	    }
	}

	if (arity < fi->fi_MinArity) {
	    setError(GCS(MSG_ProgErr_MissingArguments), func->LineCol);
	}
        if (fi->fi_MaxArity != FIA_ANY) {             /* check max arity? */
            if (arity > fi->fi_MaxArity) {
                setError(GCS(MSG_ProgErr_TooManyArguments), func->LineCol);
	    }
            if (fi->fi_ArgTypes) {
                if (!CheckArgTypes(aa, fi->fi_ArgTypes)) {
                    setError(GCS(MSG_ProgErr_Type), func->LineCol);
		}
	    }
        }
        if (fi->fi_Check) {
            err = fi->fi_Check(fi, aa);
	}

        if (!allowloopcontrol && fi->fi_Id == FUNCID_EXIT) {
            err = GCS(MSG_CompErr_InvalidExit);
	} else if (!allowloopcontrol && fi->fi_Id == FUNCID_NEXT) {
            err = GCS(MSG_CompErr_InvalidNext);
	} else if (!allowreturn && fi->fi_Id == FUNCID_RETURN) {
            err = GCS(MSG_CompErr_InvalidReturn);
	}

        if (err) {
            setError(err, func->LineCol);
	}
        func->Type = fi->fi_Type;

        if (nosideeffect && fi->fi_HasSideEffect) {
            setError(GCS(MSG_ProgErr_SideEffect), func->LineCol);
	}
    }
}

static void checkGlobalIdent(struct GlobalIdent *gi)
{
    if(gi)
    {
        long err;
        struct RelFieldPath *rap = &gi->RelFieldPath;

        err = PathNameToRelFieldPath(compProject, compTable, gi->Name, rap);
        if(err)
            setError(GES(err), gi->LineCol);
        if(rap->rap_NumFields == 0)
        {
            gi->Type = TYPE_TABLE;
            gi->ExprType = TYPE_RECORD;
        }
        else
        {
            if(rap->rap_Star)
                setError(GCS(MSG_CompErr_Star), gi->LineCol);

            gi->Type = TYPE_FIELD;
            gi->ExprType = FieldType(rap->rap_Fields[rap->rap_NumFields - 1]);
            if(gi->ExprType == TYPE_REFERENCE)
                gi->ExprType = TYPE_RECORD;
            else if(gi->ExprType == TYPE_CHOICE)
                gi->ExprType = TYPE_INTEGER;
            else if(gi->ExprType == TYPE_STRING)
                gi->ExprType = TYPE_STRING;
            else if(gi->ExprType == TYPE_VIRTUAL)
                gi->ExprType = TYPE_UNKNOWN;
            else if(gi->ExprType == TYPE_BUTTON)
                gi->ExprType = NOTYPE;
        }
    }
}

static void checkExpr(struct Expr *expr, struct LocalIdent *identList,
        bool nosideeffect, bool allowloopcontrol, bool allowreturn);
static short checkAppArgs(struct AppArgs *aa, struct LocalIdent *identList,
        bool nosideeffect, bool allowloopcontrol, bool allowreturn)
{
    short type = TYPE_NIL;

    for(; aa; aa = aa->Next)
        if(aa->Expr)
        {
            checkExpr(aa->Expr, identList, nosideeffect, allowloopcontrol, allowreturn);
            type = aa->Expr->Type;
        }
    return(type);
}

static struct LocalIdent *checkLocalIdent(struct LocalIdent *li, struct LocalIdent *identList)
{
    struct LocalIdent *new;
    struct Program *prg;
    struct UserFunction *uf;

    new = FindLocalIdentName(identList, li->Name);

    if (new == NULL && linkProjectProgram &&
	    (prg = ProjectProgram(compProject)) != NULL &&
	    (uf = FindUserFunctionName(prg, li->Name)) != NULL)
    {
        new = li;
        new->Type = UserFunctionType(uf);
        new->Arity = UserFunctionArity(uf);
        new->Types = UserFunctionTypes(uf);
        new->Pos = 0;
        new->Table = NULL;
        new->UserFunction = uf;
    }

    if(new == NULL)
        setError(GCS(MSG_CompErr_UnknownIdentifier), li->LineCol);

    return(new);
}

static void checkApp(struct App *app, struct LocalIdent *identList,
        bool nosideeffect, bool allowloopcontrol, bool allowreturn)
{
    struct AppArgs *aa = app->AppArgs;
    short i;
    struct Expr *expr;
    struct Func progn;

    if (aa == NULL) {
        return;                     /* empty application */
    }
    expr = aa->Expr;
    if (expr != NULL && expr->Case == EXPR_CASE_LOCALIDENT) {
        struct LocalIdent *func;

        func = checkLocalIdent(expr->Contents.LocalIdent, identList);
        if (func != NULL && func->Arity >= 0) {
	   /* since we skip this app-arg, add expr manually into expr list */
	    if (curFunc) {
		ADDTAIL(&curFunc->ExprList, expr);
            }
            aa = aa->Next;
            checkAppArgs(aa, identList, nosideeffect, FALSE, FALSE);
            if (CountAppArgs(aa) != func->Arity) {
                setError(GCS(MSG_CompErr_Arity), app->AppArgs->Expr->LineCol);
            }
            for (i=0; aa != NULL; aa = aa->Next, i++) {
                checkTypes(func->Types[i], aa->Expr->Type, aa->Expr->LineCol);
            }
            app->Type = func->Type;
            app->IsProgN = FALSE;
            expr->Contents.LocalIdent = func;
            return;
        }
    } else if (expr != NULL && expr->Case == EXPR_CASE_FUNC) {
        struct Func *f = expr->Contents.Func;

	/* since we skip this app-arg, we add expr manually into expr list */
	if (curFunc) {
	    ADDTAIL(&curFunc->ExprList, expr);	/* collect all expr of a func */
        }
        aa = aa->Next;
        if (f->FuncInfo->fi_Id == FUNCID_IF && aa) {
            checkExpr(aa->Expr, identList, nosideeffect, FALSE, FALSE);
            checkAppArgs(aa->Next, identList, nosideeffect, allowloopcontrol,
                    allowreturn);
        } else if (f->FuncInfo->fi_Kind == FIK_SETQ && aa) {
            struct AppArgs *setq =  aa;

            /* remove target identifiers since they are not a dependency */
            while (setq != NULL) {
                checkExpr(setq->Expr, identList, nosideeffect, FALSE, FALSE);
                if (curFunc && PRED(TAILPRED(&curFunc->ExprList))) {
                    REMTAIL(&curFunc->ExprList);
                }
                setq = setq->Next;
                if (setq == NULL) {
                    break;
                }
                checkExpr(setq->Expr, identList, nosideeffect, FALSE, FALSE);
                setq = setq->Next;
            }
        } else if (f->FuncInfo->fi_Kind == FIK_SETQLIST && aa) {
            struct AppArgs *setq =  aa;

            /* remove target identifiers since they are not a dependency */
            while (setq->Next != NULL) {
                checkExpr(setq->Expr, identList, nosideeffect, FALSE, FALSE);
                if (curFunc && PRED(TAILPRED(&curFunc->ExprList))) {
                    REMTAIL(&curFunc->ExprList);
                }
                setq = setq->Next;
            }
            checkExpr(setq->Expr, identList, nosideeffect, FALSE, FALSE);
        } else {
            checkAppArgs(aa, identList, nosideeffect, FALSE, FALSE);
        }
        checkApplyFunc(f, aa, nosideeffect, allowloopcontrol, allowreturn);
        app->Type = f->Type;
        app->IsProgN = FALSE;
        return;
    }

    checkAppArgs(aa, identList, nosideeffect, allowloopcontrol, allowreturn);
    progn.LineCol = expr->LineCol;
    progn.FuncInfo = FindFuncInfoId(FUNCID_PROGN);
    checkApplyFunc(&progn, aa, nosideeffect, allowloopcontrol, allowreturn);
    app->Type = progn.Type;
    app->IsProgN = TRUE;
}

static void checkTypeSpec(struct TypeSpec *typeSpec)
{
    if(typeSpec && typeSpec->Type == TYPE_RECORD)
    {
        struct GlobalIdent *gi = typeSpec->GlobalIdent;
        struct Table *save = compTable;

        compTable = NULL;
        checkGlobalIdent(gi);
        compTable = save;
        if(gi != NULL && gi->Type != TYPE_TABLE)
            setError(GCS(MSG_CompErr_NotATableSpecifier), gi->LineCol);
    }
}

static void checkLocalIdentDef(struct LocalIdentDef *lid, struct LocalIdent *defList)
{
    if(lid)
    {
        struct LocalIdent *li = lid->LocalIdent;

        if(FindLocalIdentName(defList, li->Name))
            setError(GCS(MSG_Err_NameAlreadyUsed), li->LineCol);

        if(lid->TypeSpec)
        {
            checkTypeSpec(lid->TypeSpec);
            li->Type = (char)lid->TypeSpec->Type;
            if(lid->TypeSpec->GlobalIdent)
                li->Table = lid->TypeSpec->GlobalIdent->RelFieldPath.rap_Table;
        }
    }
}

static struct LocalIdent *checkFromList(struct FromList *fl, struct LocalIdent *identList)
{
    struct Table *save = compTable;
    struct LocalIdent *first = NULL, *li, **last = &first;
    struct GlobalIdent *gi;

    compTable = NULL;
    for(; fl; fl = fl->Next)
    {
        gi = fl->GlobalIdent;
        checkGlobalIdent(gi);
        if(gi->Type != TYPE_TABLE)
            setError(GCS(MSG_CompErr_InvalidTableList), gi->LineCol);

        if(fl->LocalIdentDef)
        {
            checkLocalIdentDef(fl->LocalIdentDef, first);
            li = fl->LocalIdentDef->LocalIdent;
            li->Type = TYPE_RECORD;
            li->Table = gi->RelFieldPath.rap_Table;
            *last = li;
            last = &li->Next;
        }
    }
    compTable = save;

    /* add first to identList  */
    *last = identList;

    return(first);
}

static void checkSelectList(struct SelectList *sl, struct LocalIdent *identList)
{
    short i = 0;
    struct Func *f;

    for(; sl; sl = sl->Next, i++)
    {
        if(sl->Expr && sl->Expr->Case == EXPR_CASE_FUNC)
        {
            f = sl->Expr->Contents.Func;

            if(f->FuncInfo->fi_Id == FUNCID_MUL)
            {
                if(i != 0 || sl->Next != NULL)
                    setError(GCS(MSG_CompErr_SyntaxError), f->LineCol);
                sl->Expr = NULL;
            }
        }
        checkExpr(sl->Expr, identList, TRUE, FALSE, FALSE);
    }
}

static void checkOrderList(struct OrderList *ol, struct SelectList *sl, struct LocalIdent *identList)
{
    for(; ol; ol = ol->Next)
    {
        checkExpr(ol->Expr, identList, TRUE, FALSE, FALSE);
        if(ol->Expr->Case == EXPR_CASE_CONST)
        {
           long num, i;
           struct SelectList *s;

           if(ol->Expr->Type != TYPE_INTEGER)
                setError(GCS(MSG_CompErr_OrderConst), ol->Expr->LineCol);
           num = ol->Expr->Contents.Const->Contents.integer.value;
           for(i=1, s = sl; i < num && s; i++)
                s = s->Next;
           if(num < 1 || s == NULL || s->Expr == NULL)
                setError(GCS(MSG_CompErr_OrderFieldNumber), ol->Expr->LineCol);
           ol->Expr = s->Expr;
           ol->FieldNum = num;
        }
        if(IsOrderType(ol->Expr->Type) == FALSE)
            setError(GCS(MSG_CompErr_OrderType), ol->Expr->LineCol);
    }
}

static void checkSelectFromWhere(struct SelectFromWhere *sfw,
	struct LocalIdent *identList, bool nosideeffect,
        bool allowloopcontrol, bool allowreturn)
{
    struct Table *save = compTable;

    if(sfw == NULL)
        return;

    identList = checkFromList(sfw->FromList, identList);
    compTable = sfw->FromList->GlobalIdent->RelFieldPath.rap_Table;
    checkSelectList(sfw->SelectList, identList);
    checkExpr(sfw->Expr, identList, TRUE, FALSE, FALSE);
    checkOrderList(sfw->OrderList, sfw->SelectList, identList);
    checkAppArgs(sfw->AppArgs, identList, nosideeffect, TRUE, allowreturn);
    compTable = save;
}

static struct LocalIdent *checkLetList(struct LetList *ll,
        struct LocalIdent *identList, bool nosideeffect)
{
    struct LocalIdent *first = NULL, *li, **last = &first;

    for(; ll; ll = ll->Next)
    {
        checkLocalIdentDef(ll->LocalIdentDef, first);
        li = ll->LocalIdentDef->LocalIdent;
        if(ll->Expr)
        {
            *last = identList;
            checkExpr(ll->Expr, first, nosideeffect, FALSE, FALSE);
            *last = NULL;
            checkTypes(li->Type, ll->Expr->Type, ll->Expr->LineCol);
        }
        *last = li;
        last = &li->Next;
    }

    /* add first to identList */
    *last = identList;

    return(first);
}

static void checkLet(struct Let *let, struct LocalIdent *identList,
        bool nosideeffect, bool allowloopcontrol, bool allowreturn)
{
    identList = checkLetList(let->LetList, identList, nosideeffect);
    let->Type = checkAppArgs(let->AppArgs, identList,
        nosideeffect, allowloopcontrol, allowreturn);
}

static void checkCond(struct Cond *cond, struct LocalIdent *identList,
        bool nosideeffect, bool allowloopcontrol, bool allowreturn)
{
    struct CondList *cl;

    for(cl = cond->CondList; cl; cl = cl->Next)
    {
        checkExpr(cl->Expr, identList, nosideeffect, FALSE, FALSE);
        checkAppArgs(cl->AppArgs, identList,
            nosideeffect, allowloopcontrol, allowreturn);
    }
    cond->Type = TYPE_UNKNOWN;
}

static void checkCase(struct Case *c, struct LocalIdent *identList,
        bool nosideeffect, bool allowloopcontrol, bool allowreturn)
{
    struct CaseList *cl;
    struct AppArgs *aa;
    short type;

    checkExpr(c->Expr, identList, nosideeffect, FALSE, FALSE);
    type = c->Expr->Type;

    for(cl = c->CaseList; cl; cl = cl->Next)
    {
        checkExpr(cl->Expr, identList, nosideeffect, FALSE, FALSE);
        if(cl->Expr->Case == EXPR_CASE_APP && cl->Expr->Contents.App->IsProgN)
        {
            for(aa = cl->Expr->Contents.App->AppArgs; aa; aa = aa->Next)
                if(aa->Expr && GenType(type, aa->Expr->Type) == NOTYPE)
                    setError(GCS(MSG_ProgErr_Type), aa->Expr->LineCol);
        }
        else if(GenType(type, cl->Expr->Type) == NOTYPE)
            setError(GCS(MSG_ProgErr_Type), cl->Expr->LineCol);
        checkAppArgs(cl->AppArgs, identList, nosideeffect, allowloopcontrol, allowreturn);
    }
    c->Type = TYPE_UNKNOWN;
}

static struct LocalIdent *checkDoBindList(struct DoBindList *dbl,
        struct LocalIdent *identList, bool nosideeffect)
{
    struct LocalIdent *first = NULL, *li, **last = &first;

    for(; dbl; dbl = dbl->Next)
    {
        checkLocalIdentDef(dbl->LocalIdentDef, first);
        li = dbl->LocalIdentDef->LocalIdent;
        if(dbl->InitExpr)
        {
            *last = identList;
            checkExpr(dbl->InitExpr, first, nosideeffect, FALSE, FALSE);
            *last = NULL;
            checkTypes(li->Type, dbl->InitExpr->Type, dbl->InitExpr->LineCol);
        }
        *last = li;
        last = &li->Next;
        if(dbl->StepExpr)
        {
            *last = identList;
            checkExpr(dbl->StepExpr, first, nosideeffect, FALSE, FALSE);
            *last = NULL;
            checkTypes(li->Type, dbl->StepExpr->Type, dbl->StepExpr->LineCol);
        }
    }

    /* add first to identList */
    *last = identList;

    return(first);
}

static void checkDo(struct Do *d, struct LocalIdent *identList,
        bool nosideeffect, bool allowloopcontrol, bool allowreturn)
{
    identList = checkDoBindList(d->DoBindList, identList, nosideeffect);
    checkExpr(d->TermExpr, identList, nosideeffect, FALSE, FALSE);
    d->Type = checkAppArgs(d->ResultAppArgs, identList, nosideeffect, FALSE, FALSE);
    checkAppArgs(d->LoopAppArgs, identList, nosideeffect, TRUE, allowreturn);
}

static void checkDoSpecial(struct DoSpecial *ds, struct LocalIdent *identList,
        bool nosideeffect, bool allowloopcontrol, bool allowreturn)
{
    struct LocalIdent *li = ds->LocalIdentDef->LocalIdent;

    checkLocalIdentDef(ds->LocalIdentDef, NULL);
    checkExpr(ds->Expr, identList, nosideeffect, FALSE, FALSE);

    switch(ds->Case)
    {
        case DO_CASE_TIMES:
            checkTypes(ds->Expr->Type, TYPE_INTEGER, ds->Expr->LineCol);
            checkTypes(li->Type, TYPE_INTEGER, li->LineCol);
            break;

        case DO_CASE_LIST:
            checkTypes(ds->Expr->Type, TYPE_LIST, ds->Expr->LineCol);
            break;
    }

    /* add new variable to identList */
    li->Next = identList;
    identList = li;

    ds->Type = checkAppArgs(ds->ResultAppArgs, identList, nosideeffect, FALSE, FALSE);
    checkAppArgs(ds->LoopAppArgs, identList, nosideeffect, TRUE, allowreturn);
}

static void checkLocalExtend(struct LocalExtend *le, struct LocalIdent *identList)
{
    struct Table *save = compTable;
    struct RelFieldPath *rap = &le->GlobalIdent->RelFieldPath;

    le->LocalIdent = checkLocalIdent(le->LocalIdent, identList);
    if(le->LocalIdent->Table == NULL)
        setError(GCS(MSG_CompErr_NotATableSpecifier), le->LocalIdent->LineCol);

    compTable = le->LocalIdent->Table;
    checkGlobalIdent(le->GlobalIdent);
    if(rap->rap_Table != compTable)               /* not a tabative path? */
        setError(GCS(MSG_Err_NoSuchFieldOrTable), le->GlobalIdent->LineCol);
    if(rap->rap_NumFields == 0)                          /* not a tabative path? */
        setError(GCS(MSG_Err_NoSuchFieldOrTable), le->GlobalIdent->LineCol);

    compTable = save;
}

static void checkFunc(struct Func *func)
{
    if (func->FuncInfo->fi_Kind != FIK_EXPR) {
        setError(GCS(MSG_CompErr_IllegalFunction), func->LineCol);
    }
}

static void checkExpr(struct Expr *expr, struct LocalIdent *identList,
        bool nosideeffect, bool allowloopcontrol, bool allowreturn)
{
    if(expr == NULL)            /* this IS really needed (e.g. SelectList) */
        return;

    if (curFunc)
	ADDTAIL(&curFunc->ExprList, expr);	/* collect all expr of a func */

    expr->Type = TYPE_UNKNOWN;
    switch(expr->Case)
    {
        case EXPR_CASE_APP:
            checkApp(expr->Contents.App, identList,
                nosideeffect, allowloopcontrol, allowreturn);
            expr->Type = expr->Contents.App->Type;
            break;

        case EXPR_CASE_SELECTFROMWHERE:
            checkSelectFromWhere(expr->Contents.SelectFromWhere,
		identList, nosideeffect, allowloopcontrol, allowreturn);
            expr->Type = TYPE_LIST;
            break;

        case EXPR_CASE_LET:
            checkLet(expr->Contents.Let, identList,
                nosideeffect, allowloopcontrol, allowreturn);
            expr->Type = expr->Contents.Let->Type;
            break;

        case EXPR_CASE_COND:
            checkCond(expr->Contents.Cond, identList,
                nosideeffect, allowloopcontrol, allowreturn);
            expr->Type = expr->Contents.Cond->Type;
            break;

        case EXPR_CASE_CASE:
            checkCase(expr->Contents.Case, identList,
                nosideeffect, allowloopcontrol, allowreturn);
            expr->Type = expr->Contents.Case->Type;
            break;

        case EXPR_CASE_DO:
            checkDo(expr->Contents.Do, identList,
                nosideeffect, allowloopcontrol, allowreturn);
            expr->Type = expr->Contents.Do->Type;
            break;

        case EXPR_CASE_DOSPECIAL:
            checkDoSpecial(expr->Contents.DoSpecial, identList,
                nosideeffect, allowloopcontrol, allowreturn);
            expr->Type = expr->Contents.DoSpecial->Type;
            break;

        case EXPR_CASE_CONST:
            expr->Type = expr->Contents.Const->Type;
            break;

        case EXPR_CASE_GLOBALIDENT:
            checkGlobalIdent(expr->Contents.GlobalIdent);
            expr->Type = expr->Contents.GlobalIdent->ExprType;
            break;

        case EXPR_CASE_LOCALIDENT:
            expr->Contents.LocalIdent = checkLocalIdent(expr->Contents.LocalIdent, identList);
            expr->Type = expr->Contents.LocalIdent->Type;
            break;

        case EXPR_CASE_LOCALEXTEND:
            checkLocalExtend(expr->Contents.LocalExtend, identList);
            expr->Type = expr->Contents.LocalExtend->GlobalIdent->ExprType;
            break;

        case EXPR_CASE_FUNC:
            checkFunc(expr->Contents.Func);
            break;
    }
}

static struct LocalIdent *buildIdentDefList(struct DefList *defList, struct LocalIdent *identList)
{
    struct LocalIdent *first = identList, *li, **last = &first;

    for(; defList; defList = defList->Next)
    {
        switch(defList->Case)
        {
            case DEFLIST_CASE_VARDEF:
            {
                struct VarDef *varDef = defList->Contents.VarDef;

                checkLocalIdentDef(varDef->LocalIdentDef, first);
                li = varDef->LocalIdentDef->LocalIdent;
                if(varDef->Expr)
                    checkExpr(varDef->Expr, first, TRUE, FALSE, FALSE);
                if(varDef->Expr != NULL)
                    checkTypes(li->Type, varDef->Expr->Type, 
			    varDef->Expr->LineCol);
                *last = li;
                last = &li->Next;
                *last = identList;
                break;
            }

            case DEFLIST_CASE_FUNCDEF:
            {
                struct FuncDef *funcDef = defList->Contents.FuncDef;
                struct FuncArg *fa;
                short i;
                struct LocalIdent *argli, *argfirst=NULL, **arglast = &argfirst;

                li = funcDef->FuncName;
                if(funcDef->Trigger && strlen(li->Name) > FIELD_TRIGGER_LEN)
                    setError(GCS(MSG_CompErr_FunctionNameLength), li->LineCol);
                if(FindLocalIdentName(first, li->Name) != NULL)
                    setError(GCS(MSG_Err_NameAlreadyUsed), li->LineCol);

                for(fa = funcDef->FuncArgList, i = 0; fa; fa = fa->Next, i++)
                {
                    checkLocalIdentDef(fa->LocalIdentDef, argfirst);
                    argli = fa->LocalIdentDef->LocalIdent;
                    *arglast = argli;
                    arglast = &argli->Next;
                    funcDef->FuncName->Types[i] = argli->Type;
                }
                li->Arity = i;

                *last = li;
                last = &li->Next;
                *last = identList;
                break;
            }

            default:
                setError(GCS(MSG_Err_Internal), defList->LineCol);
                break;
        }
    }

    return(first);
}

static void checkDefList(struct DefList *defList, struct LocalIdent *identList)
{
    short numFuncs = 0;

    for(; defList; defList = defList->Next)
    {
        struct FuncDef *funcDef;
        struct FuncArg *fa;
        struct LocalIdent *li = NULL, *first = NULL, **last = &first;
        struct AppArgs *aa;
        short type = TYPE_NIL;

        if(defList->Case != DEFLIST_CASE_FUNCDEF)
            continue;

        numFuncs++;

        funcDef = defList->Contents.FuncDef;
        for(fa = funcDef->FuncArgList; fa != NULL; fa = fa->Next)
        {
            li = fa->LocalIdentDef->LocalIdent;
            *last = li;
            last = &li->Next;
        }

        *last = identList;
	curFunc = funcDef->FuncName;
	NewList(&curFunc->ExprList);
        checkAppArgs(funcDef->AppArgs, first, FALSE, FALSE, TRUE);
        *last = NULL;

        for(aa = funcDef->AppArgs; aa; aa = aa->Next)
            type = aa->Expr->Type;
        funcDef->FuncName->Type = (char)type;
    }
}

static void checkStart(struct Start *start, bool nosideeffect, bool expr, struct LocalIdent *identList)
{
    curFunc = NULL;

    if(start)
    {
        switch(start->Case)
        {
            case START_CASE_EXPR:
                if(expr == FALSE)
                    setError(GCS(MSG_CompErr_FuncDefExpected), start->Contents.Expr->LineCol);
                checkExpr(start->Contents.Expr, identList, nosideeffect, FALSE, FALSE);
                break;

            case START_CASE_DEFLIST:
            {
                if(expr)
                    setError(GCS(MSG_CompErr_ExprExpected), start->Contents.DefList->LineCol);

                identList = buildIdentDefList(start->Contents.DefList, identList);
                checkDefList(start->Contents.DefList, identList);
                break;
            }
        }
    }
}

static const char *compileProgram(struct Project *pr, struct Table *tab,
        const char *src, long size,
        bool add_brackets, bool nosideeffect, bool expr, bool debug,
        struct Program **prg, long *line, long *col, const char **filename)
{
    const char *error = NULL;
    struct MemPool *memPool;
    struct Start *start = NULL;
    struct LocalIdent *identList = NULL;
    int ret;

    if (prg != NULL) {
        *prg = NULL;
    }

    errString = NULL;
    errLine = 1;
    errCol = 1;
    errFilename = NULL;
    compTable = tab;
    compProject = pr;
    linkProjectProgram = expr;
    memPool = CreatePool(0, 4096, 1024);

    if (add_brackets) {
        char *new = PoolAllocMem(memPool, size + 3);

        if (new == NULL) {
            error = GCS(MSG_Err_Mem);
        } else {
            new[0] = '(';
            memcpy(new + 1, src, size);
            new[size + 1] = '\n';
            new[size + 2] = ')';
            size += 3;
        }
	src = new;
    }

    /* build syntax tree (parser part) */
    if (error == NULL) {
        error = MakeSynTree(memPool, src, size, &start,
            &errLine, &errCol, &errFilename);
    }

    /* build list of global variables */
    if (error == NULL && start) {
        error = BuildLocalIdentList(memPool, &identList);
    }

    /* do semantic checks */
    if (error == NULL && start) {
        ret = setjmp(errJmpBuf);
        if (ret == 0) {
            checkStart(start, nosideeffect, expr, identList);
        }
        error = errString;
    }

    /* generate program code */
    if (error == NULL && start && prg) {
        *prg = StartToProgram(compProject, start, identList, debug);
        if (*prg == NULL) {
            error = GCS(MSG_Err_Mem);
        }
    }

    if (errFilename != NULL) {
        static char filename[101];

        CodesetStrNBytesCopy(filename, errFilename, (sizeof(filename) - 1));
        errFilename = filename;
    }

    if (memPool != NULL) {
        DeletePool(memPool);
    }
    compProject = NULL;
    compTable = NULL;

    if (line) {
        *line = errLine;
    }
    if (col) {
        *col = errCol;
    }
    if (filename) {
        *filename = errFilename;
    }

    return error;
}

const char *CompileFilter(struct Table *tab, const char *src, long size,
        struct Program **prg, long *line, long *col)
{
    struct Project *pr = TableProject(tab);

    return(compileProgram(pr, tab, src, size, FALSE, TRUE, TRUE,
        FALSE, prg, line, col, NULL));
}

const char *CompileExpr(struct Project *pr, struct Table *tab, const char *src, 
	long size, struct Program **prg, long *line, long *col)
{
    return(compileProgram(pr, tab, src, size,
        FALSE, TRUE, TRUE, FALSE, prg, line, col, NULL));
}

const char *CompileQuery(struct Project *pr, const char *src, long size,
                      struct Program **prg, long *line, long *col)
{
    return(compileProgram(pr, NULL, src, size, TRUE, FALSE, TRUE,
        (bool)GetPrefsValue(pr, PREFS_PROGRAMDEBUGINFO), prg, line, col, NULL));
}

const char *CompileProjectProgram(struct Project *pr, const char *src,long size,
        struct Program **prg, long *line, long *col, const char **filename)
{
    return(compileProgram(pr, NULL, src, size, FALSE, FALSE, FALSE,
        (bool)GetPrefsValue(pr, PREFS_PROGRAMDEBUGINFO), prg, line, col, filename));
}

bool IsIdentifier(const char *name)
{
    struct MemPool *memPool;
    struct Start *start = NULL;
    const char *error;
    bool result = FALSE;

    if (name == NULL) {
        return FALSE;
    }

    memPool = CreatePool(0, 4096, 1024);
    error = MakeSynTree(memPool, name, strlen(name), &start, NULL, NULL, NULL);
    if (error == NULL && start != NULL && start->Case == START_CASE_EXPR &&
            start->Contents.Expr != NULL && 
            start->Contents.Expr->Case == EXPR_CASE_GLOBALIDENT) {
        result = TRUE;
    }
    if (memPool != NULL) {
        DeletePool(memPool);
    }

    return result;
}

