/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** compile_impl.h
**
** Creation date: 16.04.1995
** Last modified: 26.07.2011
*/

#ifndef COMPILE_IMPL_H
#define COMPILE_IMPL_H

#include "compile.h"
#include "name.h"

/* for storing a location in a source code */
struct LineCol
{
    short Line, Col;
    const char *Filename;     /* external file or NULL for project program */
};

struct Start
{
    struct Ignore *Ignore1;
    struct Ignore *Ignore2;
    short Case;
    union
    {
        struct Expr *Expr;
        struct DefList *DefList;
    } Contents;
};

enum
{
    START_CASE_EXPR,
    START_CASE_DEFLIST,
};

struct DefList
{
    struct LineCol LineCol;
    short Case;
    union
    {
        struct FuncDef *FuncDef;
        struct VarDef *VarDef;
    } Contents;
    struct Ignore *Ignore1;
    struct DefList *Next;
};

enum
{
    DEFLIST_CASE_FUNCDEF,
    DEFLIST_CASE_VARDEF,
};

struct FuncDef
{
    struct LineCol LineCol;
    struct Ignore *Ignore1;
    bool Trigger;
    struct Ignore *Ignore2;
    struct LocalIdent *FuncName;
    struct Ignore *Ignore3;
    struct Ignore *Ignore4;
    struct FuncArg *FuncArgList;
    struct Ignore *Ignore5;
    struct AppArgs *AppArgs;
};

struct FuncArg
{
    struct LocalIdentDef *LocalIdentDef;
    struct FuncArg *Next;
};

struct VarDef
{
    struct LineCol LineCol;
    struct Ignore *Ignore1;
    bool Keep;
    struct Ignore *Ignore2;
    struct LocalIdentDef *LocalIdentDef;
    struct Expr *Expr;
    struct Ignore *Ignore3;
};

struct Expr
{
    struct MinNode Node;    	/* for collecting all expr of a function */
    struct LineCol LineCol;
    short Case;
    short Type;
    union
    {
        struct App *App;
        struct Const *Const;
        struct SelectFromWhere *SelectFromWhere;
        struct Let *Let;
        struct Cond *Cond;
        struct Case *Case;
        struct Do *Do;
        struct DoSpecial *DoSpecial;
        struct GlobalIdent *GlobalIdent;
        struct LocalIdent *LocalIdent;
        struct LocalExtend *LocalExtend;
        struct Func *Func;
    }Contents;
};

enum
{
    EXPR_CASE_APP,
    EXPR_CASE_CONST,
    EXPR_CASE_SELECTFROMWHERE,
    EXPR_CASE_LET,
    EXPR_CASE_COND,
    EXPR_CASE_CASE,
    EXPR_CASE_DO,
    EXPR_CASE_DOSPECIAL,
    EXPR_CASE_GLOBALIDENT,
    EXPR_CASE_LOCALIDENT,
    EXPR_CASE_LOCALEXTEND,
    EXPR_CASE_FUNC,
};

struct App
{
    struct LineCol LineCol;
    short Type;
    bool IsProgN;
    struct Ignore *Ignore;
    struct AppArgs *AppArgs;
};

struct Func
{
    struct LineCol LineCol;
    short Type;
    struct FuncInfo *FuncInfo;

    /* for EXIT, NEXT & RETURN */
    short Depth;        /* Stack depth */
    long Index;         /* address where depth & jump address should be set */
    struct Func *Next;
};

struct AppArgs
{
    struct Expr *Expr;
    struct Ignore *Ignore;
    struct AppArgs *Next;
};

#define CONST_INTEGER_FORMAT_OCTAL -1
#define CONST_INTEGER_FORMAT_DECIMAL 0
/* hexadecimal format > 0 holding number of characters */

struct Const
{
    struct LineCol LineCol;
    short Case;
    short Type;
    union
    {
        struct 
        {
            long value;
            char format;
        } integer;
        double real;
        char *string;
        long date;
        long time;
        struct ConstInfo *ConstInfo;
    } Contents;
};

enum
{
    CONST_CASE_NORMAL,
    CONST_CASE_PREDEFINED,
};

struct GlobalIdent
{
    struct LineCol LineCol;
    short Type;                         /* one of TYPE_FIELD, TYPE_TABLE */
    short ExprType;                     /* TYPE_TABLE, field type or ??? */
    char *Name;
    struct RelFieldPath RelFieldPath;
};

struct LocalIdent
{
    struct LineCol LineCol;
    char *Name;
    char Type;
    short Arity;                        /* for functions, -1 for variables */
    char *Types;                        /* for functions */
    short Pos;                          /* rel. stack position for variables */
    struct Table *Table;                /* for table field */
    struct UserFunction *UserFunction;
    struct UserGlobalVar *UserGlobalVar;
    struct List ExprList;		/* list of all expr in this function */
    struct LocalIdent *Next;            /* for building list of def. idents */
};

struct LocalExtend                      /* local.global case */
{
    struct LocalIdent *LocalIdent;
    struct GlobalIdent *GlobalIdent;
};

struct SelectFromWhere
{
    struct LineCol LineCol;
    bool IsSelectFromWhere;          /* TRUE for SELECT FROM WHERE CASE */
    struct Ignore *Ignore1;
    struct Ignore *Ignore2;
    bool Distinct;
    struct Ignore *Ignore3;
    struct SelectList *SelectList;
    struct Ignore *Ignore4;
    struct FromList *FromList;
    struct Ignore *Ignore5;
    struct Expr *Expr;
    struct Ignore *Ignore6;
    struct Ignore *Ignore7;
    struct Ignore *Ignore8;
    struct OrderList *OrderList;
    struct Ignore *Ignore9;
    struct AppArgs *AppArgs;
};

struct SelectList
{
    struct Expr *Expr;
    struct Ignore *Ignore1;
    char *Title;
    struct Ignore *Ignore2;
    struct Ignore *Ignore3;
    struct SelectList *Next;
};

struct FromList
{
    struct GlobalIdent *GlobalIdent;
    struct Ignore *Ignore1;
    struct LocalIdentDef *LocalIdentDef;
    struct Ignore *Ignore2;
    struct FromList *Next;
};

struct OrderList
{
    struct Expr *Expr;
    long FieldNum;                      /* field num in case of int constant */
    struct Ignore *Ignore1;
    short OrderMode;
    struct Ignore *Ignore2;
    struct Ignore *Ignore3;
    struct OrderList *Next;
};

/* OrderModes: */
enum { ORDER_DEFAULT, ORDER_ASC, ORDER_DESC };

struct Let
{
    struct LineCol LineCol;
    short Type;
    struct Ignore *Ignore1;
    struct Ignore *Ignore2;
    struct Ignore *Ignore3;
    struct LetList *LetList;
    struct Ignore *Ignore4;
    struct AppArgs *AppArgs;
};

struct LetList
{
    struct Ignore *Ignore1;
    struct LocalIdentDef *LocalIdentDef;
    struct Expr *Expr;
    struct Ignore *Ignore2;
    struct Ignore *Ignore3;
    struct LetList *Next;
};

struct Cond
{
    struct LineCol LineCol;
    short Type;
    struct Ignore *Ignore1;
    struct Ignore *Ignore2;
    struct CondList *CondList;
};

struct Case
{
    struct LineCol LineCol;
    short Type;
    struct Ignore *Ignore1;
    struct Ignore *Ignore2;
    struct Expr *Expr;
    struct Ignore *Ignore3;
    struct CaseList *CaseList;
};

struct CaseList
{
    struct Ignore *Ignore1;
    struct Expr *Expr;
    struct Ignore *Ignore2;
    struct AppArgs *AppArgs;
    struct Ignore *Ignore3;
    struct CaseList *Next;
};

struct CondList
{
    struct Ignore *Ignore1;
    struct Expr *Expr;
    struct Ignore *Ignore2;
    struct AppArgs *AppArgs;
    struct Ignore *Ignore3;
    struct CondList *Next;
};

struct Do
{
    struct LineCol LineCol;
    short Type;
    struct Ignore *Ignore1;
    struct Ignore *Ignore2;
    struct Ignore *Ignore3;
    struct DoBindList *DoBindList;
    struct Ignore *Ignore4;
    struct Ignore *Ignore5;
    struct Expr *TermExpr;
    struct Ignore *Ignore6;
    struct AppArgs *ResultAppArgs;
    struct Ignore *Ignore7;
    struct AppArgs *LoopAppArgs;
};

struct DoBindList
{
    struct Ignore *Ignore1;
    struct LocalIdentDef *LocalIdentDef;
    struct Expr *InitExpr;
    struct Ignore *Ignore2;
    struct Expr *StepExpr;
    struct Ignore *Ignore3;
    struct Ignore *Ignore4;
    struct DoBindList *Next;
};

struct DoSpecial
{
    struct LineCol LineCol;
    short Type;
    short Case;
    struct Ignore *Ignore1;
    struct Ignore *Ignore2;
    struct Ignore *Ignore3;
    struct LocalIdentDef *LocalIdentDef;
    struct Expr *Expr;
    struct Ignore *Ignore4;
    struct AppArgs *ResultAppArgs;
    struct Ignore *Ignore5;
    struct AppArgs *LoopAppArgs;
};

enum
{
    DO_CASE_TIMES,
    DO_CASE_LIST,
};


struct LocalIdentDef
{
    struct LocalIdent *LocalIdent;
    struct Ignore *Ignore;
    struct TypeSpec *TypeSpec;
};

struct TypeSpec
{
    short Type;
    struct GlobalIdent *GlobalIdent;
    struct Ignore *Ignore;
};

struct Ignore
{
    char *String;
    struct Ignore *Next;
};

/*
** Special meaning comments:
*/

#define IGNORE_REPLACE_START "@+"
#define IGNORE_REPLACE_END "@-"

#endif

