/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** func.c
**
** Creation date: 23.04.1995
** Last modified: 06.03.2022
*/

#include "system.h"
#include "func.h"
#include "functable.h"
#include "funcprintf.h"
#include "programbin.h"
#include "heap.h"
#include "stack.h"
#include "break.h"
#include "requesters.h"
#include "preferences.h"
#include "memo.h"
#include "guidosutil.h"
#include "dosutil.h"
#include "strhandle.h"
#include "util.h"
#include "memory.h"
#include "dbtypes.h"
#include "appstrings_cat.h"

void PopAndPushNil(short numArgs)
{
    StackSqueeze(0, numArgs);
    PushNil();
}

void ProgNot(short numArgs)
{
    if(PopBool2())
        PushNil();
    else
        PushTrue();
}

static void addreal(double real, short numArgs, bool add)
{
    double real2;
    bool nil;

    while(numArgs > 0)
    {
        nil = !PopReal2(&real2);
        numArgs--;
        if(nil)
        {
            PopAndPushNil(numArgs);
            return;
        }
        if(add)
            real += real2;
        else
            real -= real2;
    }
    PushReal(real);
}

static void addinteger(long integer, short numArgs, bool add)
{
    struct Element *elem;

    while(numArgs > 0)
    {
        elem = StackPop();
        numArgs--;
        switch(elem->el_Type)
        {
            case TYPE_NIL:      PopAndPushNil(numArgs);
                                return;
                                break;
            case TYPE_INTEGER:  if(add)
                                    integer += elem->el_Contents.integer;
                                else
                                    integer -= elem->el_Contents.integer;
                                break;
            case TYPE_REAL:     if(add)
                                {
                                    addreal(integer + elem->el_Contents.real, numArgs, add);
                                    return;
                                    break;
                                }
                                else
                                {
                                    addreal(integer - elem->el_Contents.real, numArgs, add);
                                    return;
                                    break;
                                }
                                break;

            default:            ExecuteAbort(GCS(MSG_ProgErr_AddSubTypes));
                                break;
        }
    }
    PushInteger(integer);
}

static void adddate(long date, short numArgs, bool add)
{
    long sum =0, val;
    bool nil;

    while (numArgs > 0)
    {
        nil = !PopInteger2(&val);
        numArgs--;
        if(nil)
        {
            PopAndPushNil(numArgs);
            return;
        }
	sum += val;
    }
    if (!DateAdd(date, add? sum : -sum, &date))
    {
	PushNil();
	return;
    }
    PushDate(date);
}

static void addtime(long time, short numArgs, bool add)
{
    struct Element *elem;
    long val = 0;

    while(numArgs > 0)
    {
        elem = StackPop();
        numArgs--;
        switch(elem->el_Type)
        {
            case TYPE_NIL:      PopAndPushNil(numArgs);
                                return;
	    case TYPE_REAL: 	elem = ElementCastInteger(elem);
				if (elem->el_Type != TYPE_INTEGER)
                                {
                                    PopAndPushNil(numArgs);
                                    return;
                                }
				/* no break */
            case TYPE_INTEGER:  val = elem->el_Contents.integer;
                                break;
            case TYPE_TIME:     val = elem->el_Contents.time;
                                break;
            default:            ExecuteAbort(GCS(MSG_ProgErr_AddSubTypes));
                                break;
        }
	if (!TimeAdd(time, (add? val : -val), &time)) 
	{
	    PopAndPushNil(numArgs);
	    return;
	}
    }
    PushTime(time);
}


static void addstringmemo(const char *first, short numArgs, bool isMemo)
{
    char *s, *old;
    const char *next;

    s = ExecuteAllocString(first);
    while (numArgs > 0) {
        next = PopString2();
        numArgs--;
        if (next == NULL) {
            FreeString(executeMemPool, s);
            PopAndPushNil(numArgs);
            return;
        }

        old = s;
        s = ExecuteAllocMem(strlen(old) + strlen(next) + 1);
        strcpy(s, old);
        strcat(s, next);
        FreeString(executeMemPool, old);
    }

    if (isMemo) {
        PushMemo(s, TRUE);
    } else {
        PushString(s, TRUE);
    }
}

void ProgAdd(short numArgs)
{
    struct Element *elem;

    elem = StackPop();
    numArgs--;
    switch (elem->el_Type) {
        case TYPE_NIL:          
            PopAndPushNil(numArgs);
            break;
        case TYPE_INTEGER:
            addinteger(elem->el_Contents.integer, numArgs, TRUE);
            break;
        case TYPE_REAL:
            addreal(elem->el_Contents.real, numArgs, TRUE);
            break;
        case TYPE_STRING:
            addstringmemo(elem->el_Contents.string.string, numArgs, FALSE);
            break;
        case TYPE_MEMO:
            addstringmemo(elem->el_Contents.mi, numArgs, TRUE);
            break;
        case TYPE_DATE:
            adddate(elem->el_Contents.date, numArgs, TRUE);
            break;
        case TYPE_TIME:
            addtime(elem->el_Contents.time, numArgs, TRUE);
            break;
        default:
            ExecuteAbort(GCS(MSG_ProgErr_AddSubTypes));
            break;
    }
}

void ProgSub(short numArgs)
{
    struct Element *elem;

    elem = StackPop();
    numArgs--;
    switch(elem->el_Type)
    {
        case TYPE_NIL:          PopAndPushNil(numArgs);         break;
        case TYPE_INTEGER:      if(numArgs == 0)
                                    PushInteger(-elem->el_Contents.integer);
                                else
                                    addinteger(elem->el_Contents.integer, numArgs, FALSE);
                                break;
        case TYPE_REAL:         if(numArgs == 0)
                                    PushReal(-elem->el_Contents.real);
                                else
                                    addreal(elem->el_Contents.real, numArgs, FALSE);
                                break;
        case TYPE_DATE:         adddate(elem->el_Contents.date, numArgs, FALSE);        break;
        case TYPE_TIME:         addtime(elem->el_Contents.time, numArgs, FALSE);        break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_AddSubTypes));             break;
    }
}

void ProgProgN(short numArgs)
{
    if(numArgs == 0)
        PushNil();
    for(; numArgs > 1; numArgs--)
        StackPop();
}

void ProgProg1(short numArgs)
{
    if(numArgs == 0)
        PushNil();
    else
        StackSqueeze(1, numArgs - 1);
}

/*
** Conversion functions
*/

void ProgString(short numArgs)
{
    struct Element *elem = StackPop();
    struct RelFieldPath *rap = NULL;
    struct Field *field = NULL;
    bool allocated = FALSE;
    char *str = NULL;

    if(elem->el_Type == TYPE_FIELD)
    {
        rap = elem->el_Contents.field.rap;
        field = RelFieldPathField(rap);
        elem = GetGlobalData(rap, elem->el_Contents.field.rec);
    }

    str = SimpleElementToString(elem, field, executeMemPool, &allocated);
    if(str == NULL)
        ExecuteAbort(GCS(MSG_Err_Mem));
    PushString(str, allocated);
}

void ProgMemo(short numArgs)
{
    ProgString(numArgs);
    StackPush(PopMemo());
}

void ProgInteger(short numArgs)
{
    struct Element *elem = StackPop();
    long value;

    switch(elem->el_Type)
    {
        case TYPE_INTEGER:
        case TYPE_NIL:
	    StackPush(elem);
	    return;

        case TYPE_REAL:
            StackPush(ElementCastInteger(elem));
            return;

	case TYPE_MEMO:
	    elem = ElementCastString(elem);
	    if (elem->el_Type != TYPE_STRING)
	    {
		PushNil();
		return;
	    }
	    /* no break */
	case TYPE_STRING:
	    if (!StrToInt(elem->el_Contents.string.string, &value))
	    {
		PushNil();
		return;
	    }
	    break;

        case TYPE_DATE:
            value = elem->el_Contents.date;
            break;

        case TYPE_TIME:
            value = elem->el_Contents.time;
            break;

        case TYPE_RECORD:
            value = RecordNum(elem->el_Contents.tabrec.rec);
            break;

        default:
            ExecuteAbort(GCS(MSG_ProgErr_Type));
            return;
    }

    PushInteger(value);
}

void ProgReal(short numArgs)
{
    struct Element *elem = StackPop();
    double value;

    switch(elem->el_Type)
    {
        case TYPE_REAL:
        case TYPE_NIL:
	    StackPush(elem);
	    return;

        case TYPE_INTEGER:
            StackPush(ElementCastReal(elem));
            return;

	case TYPE_MEMO:
	    elem = ElementCastString(elem);
	    if (elem->el_Type != TYPE_STRING)
	    {
		PushNil();
		return;
	    }
	    /* no break */
        case TYPE_STRING:
	    if (!StrToDouble(elem->el_Contents.string.string, &value))
	    {
		PushNil();
		return;
	    }
	    break;

        case TYPE_DATE:
            value = elem->el_Contents.date;
            break;

        case TYPE_TIME:
            value = elem->el_Contents.time;
            break;

        case TYPE_RECORD:
            value = RecordNum(elem->el_Contents.tabrec.rec);
            break;

        default:
            ExecuteAbort(GCS(MSG_ProgErr_Type));
            return;
    }

    PushReal(value);
}

void ProgDate(short numArgs)
{
    struct Element *elem = StackPop();
    long value;
    bool nil;

    switch(elem->el_Type)
    {
        case TYPE_DATE:
        case TYPE_NIL:
	    StackPush(elem);
            return;

	case TYPE_MEMO:
	    elem = ElementCastString(elem);
	    if (elem->el_Type != TYPE_STRING)
	    {
		PushNil();
		return;
	    }
	    /* no break */
        case TYPE_STRING:
            nil = !StringToDate(elem->el_Contents.string.string, &value);
	    break;

	case TYPE_REAL: 	
	    elem = ElementCastInteger(elem);
	    if (elem->el_Type != TYPE_INTEGER)
	    {
		PushNil();
		return;
	    }
	    /* no break */
        case TYPE_INTEGER:
	    nil =  !DateAdd(0, elem->el_Contents.integer, &value);
            break;

        default:
            ExecuteAbort(GCS(MSG_ProgErr_Type));
            return;
    }

    if(nil)
        PushNil();
    else
        PushDate(value);
}

void ProgTime(short numArgs)
{
    struct Element *elem = StackPop();
    long value;

    switch(elem->el_Type)
    {
        case TYPE_TIME:
	case TYPE_NIL:
	    StackPush(elem);
            return;

	case TYPE_MEMO:
	    elem = ElementCastString(elem);
	    if (elem->el_Type != TYPE_STRING)
	    {
		PushNil();
		return;
	    }
	    /* no break */
        case TYPE_STRING:
            if (!StringToTime(elem->el_Contents.string.string, 
			FORMAT_HOURMINSEC, &value))
	    {
		PushNil();
		return;
	    }
	    break;

	case TYPE_REAL: 	
	    elem = ElementCastInteger(elem);
	    if (elem->el_Type != TYPE_INTEGER)
	    {
		PushNil();
		return;
	    }
	    /* no break */
        case TYPE_INTEGER:
	    if (!TimeAdd(0, elem->el_Contents.integer, &value)) 
	    {
		PushNil();
		return;
	    }
            break;

        default:
            ExecuteAbort(GCS(MSG_ProgErr_Type));
            return;
    }

    PushTime(value);
}

void ProgStrP(short numArgs)
{
    struct Element *elem = StackPop();

    if(elem->el_Type == TYPE_STRING)
        PushTrue();
    else
        PushNil();
}

void ProgMemoP(short numArgs)
{
    struct Element *elem = StackPop();

    if(elem->el_Type == TYPE_MEMO)
        PushTrue();
    else
        PushNil();
}

void ProgIntP(short numArgs)
{
    struct Element *elem = StackPop();

    if(elem->el_Type == TYPE_INTEGER)
        PushTrue();
    else
        PushNil();
}

void ProgRealP(short numArgs)
{
    struct Element *elem = StackPop();

    if(elem->el_Type == TYPE_REAL)
        PushTrue();
    else
        PushNil();
}

void ProgDateP(short numArgs)
{
    struct Element *elem = StackPop();

    if(elem->el_Type == TYPE_DATE)
        PushTrue();
    else
        PushNil();
}

void ProgTimeP(short numArgs)
{
    struct Element *elem = StackPop();

    if(elem->el_Type == TYPE_TIME)
        PushTrue();
    else
        PushNil();
}

void ProgRecP(short numArgs)
{
    struct Table *tab = PopTable2(NULL);
    struct Element *elem = StackPop();

    if(elem->el_Type == TYPE_NIL)
        PushTrue();
    else if(elem->el_Type == TYPE_RECORD &&
        (tab == NULL || elem->el_Contents.tabrec.tab == tab))
            PushTrue();
    else
        PushNil();
}

void ProgNull(short numArgs)
{
    struct Element *elem = StackPop();

    if(elem->el_Type == TYPE_NIL)
        PushTrue();
    else
        PushNil();
}

void ProgListP(short numArgs)
{
    struct Element *elem = StackPop();

    if(elem->el_Type == TYPE_NIL || elem->el_Type == TYPE_LIST)
        PushTrue();
    else
        PushNil();
}

void ProgConsP(short numArgs)
{
    struct Element *elem = StackPop();

    if(elem->el_Type == TYPE_LIST)
        PushTrue();
    else
        PushNil();
}

void ProgGC(short numArgs)
{
    GarbageCollection();
    PushNil();
}

static void callsystem(const char *cmd, bool bg, bool useShell)
{
    int rc;

    ProjectSetCurrentDir(executeProject);    /* make sure we have correct CWD */
    GuiSystem(cmd, bg, useShell, &rc);

    PushInteger(rc);

    if (BreakCheck()) {
        ExecuteAbort(EXECUTE_ABORT);
    }
}

void systemHelp(short numArgs, bool bg)
{
    char *cmd;

    ProgSPrintF(numArgs);
    cmd = PopString2();
    if (cmd)
        callsystem(cmd, bg, TRUE);
    else
	PushNil();
}

void ProgSystem(short numArgs)
{
    systemHelp(numArgs, FALSE);
}

void ProgSystemStar(short numArgs)
{
    systemHelp(numArgs, TRUE);
}

static void callpattern(short tag, bool bg)
{
    const char *pattern = GetPrefsString(executeProject, tag);
    char *filename = PopString2();
    char *path = ExpandPathname(filename, executeMemPool);
    char *cmd = BuildCommandLine(executeMemPool, pattern, path);

    if ((path == NULL && filename != NULL) || cmd == NULL)
        ExecuteAbort(GCS(MSG_Err_Mem));

    callsystem(cmd, bg, FALSE);

    FreeString(executeMemPool, path);
    FreeString(executeMemPool, cmd);
}

void ProgEdit(short numArgs)
{
    callpattern(PREFS_EXTERNALEDITOR, FALSE);
    BreakReInit();
}

void ProgEditStar(short numArgs)
{
    callpattern(PREFS_EXTERNALEDITOR, TRUE);
}

static void progViewHelp(bool bg)
{
#ifdef USE_EXTERNALVIEWER
    callpattern(PREFS_EXTERNALVIEWER, bg);
#else
    bool rc = GuiViewFile(PopString2(), bg);
    PushInteger(rc? 0 : 1);
#endif
}

void ProgView(short numArgs)
{
    progViewHelp(FALSE);
    BreakReInit();
}

void ProgViewStar(short numArgs)
{
    progViewHelp(TRUE);
}

void ProgStat(short numArgs)
{
    char *filename = PopString2();
    short rc = -1;

    ProjectSetCurrentDir(executeProject);    /* make sure we have correct CWD */
    if(filename != NULL)
        rc = FileStat(filename);
    if(rc < 0)
        PushNil();
    else
        PushInteger(rc);
}

void ProgHalt(short numArgs)
{
    ExecuteAbort(EXECUTE_ABORT);
}

void ProgError(short numArgs)
{
    char *error;
    short rc;

    ProgSPrintF(numArgs);                       /* will push error string */
    error = PopString2();

    BreakHide();
    rc = Ask2(NULL, GCS(MSG_Prog_ProgrammedError), AT_WARN,
	GCS(MSG_Prog_ProgrammedErrorAnswers), "%s", EmptyNullStr(error));
    BreakReInit();

    ExecuteAbort(rc? EXECUTE_ABORT : GCS(MSG_Prog_ProgrammedError));
}

void ProgMessage(short numArgs)
{
    struct Element *elem = NULL;
    char *s = NULL;

    ProgSPrintF(numArgs);                       /* will push message string */
    elem = PopString();
    if(elem->el_Type == TYPE_STRING)
        s = elem->el_Contents.string.string;
    BreakMessage(s);

    StackPush(elem);
}

void ProgCompleteMax(short numArgs)
{
    struct Element *elem = StackTop();
    long steps;

    if (!PopInteger2(&steps)) {
        steps = 0;              /* disable progress bar */
    }
    BreakSetMaxSteps(steps);

    StackPush(elem);
}

void ProgCompleteAdd(short numArgs)
{
    struct Element *elem = StackTop();
    long add;

    if (PopInteger2(&add)) {
        BreakAddCurrentStep(add);
    } else {
        BreakSetCurrentStep(0);    /* disable progress bar */
    }

    StackPush(elem);
}

void ProgComplete(short numArgs)
{
    struct Element *elem = StackTop();
    long cur;

    if (!PopInteger2(&cur)) {
        cur = 0;                /* disable progress bar */
    }
    BreakSetCurrentStep(cur);

    StackPush(elem);
}

void ProgCallFunctionElement(struct Element *elem, short numArgs, bool sync)
{
    if (elem->el_Type != TYPE_FUNCTION) {
	PopAndPushNil(numArgs);
    } else if (elem->el_Contents.function.type == ELEM_USERFUNCTION) {
	struct UserFunction *uf = elem->el_Contents.function.ptr;

	if (UserFunctionArity(uf) != numArgs) {
            ExecuteAbort(GCS(MSG_CompErr_Arity));
	}
        StackReverse(numArgs);
	if (sync) {
	    ExecuteUserFunction(uf);
	} else {
	    CallUserFunction(uf);
	}
    } else if (elem->el_Contents.function.type == ELEM_FUNCINFO) {
	struct FuncInfo *fi = elem->el_Contents.function.ptr;

	if (!executeAllowSideEffects && fi->fi_HasSideEffect) {
	    ExecuteAbort(GCS(MSG_ProgErr_SideEffect));
	}

	if (numArgs < fi->fi_MinArity || 
		(fi->fi_MaxArity != FIA_ANY && numArgs > fi->fi_MaxArity)) {
	    ExecuteAbort(GCS(MSG_ProgErr_FunctionArity));
	}

        if (fi->fi_Code == NULL) {
            ExecuteAbort(GCS(MSG_Err_NotImplemented));
	}
        fi->fi_Code(numArgs);
    } else {
 	ExecuteAbort(GCS(MSG_Err_Internal));
    }
}

void ProgFunCall(short numArgs)
{
    ProgCallFunctionElement(PopFunction(), numArgs-1, FALSE);
}

void ProgApply(short numArgs)
{
    struct Element *func = PopFunction();
    struct Element *elem;

    numArgs--;
    if (numArgs < 1)
	ExecuteAbort(GCS(MSG_ProgErr_MissingArguments));

    StackReverse(numArgs);
    elem = PopList();
    numArgs--;

    while (elem->el_Type == TYPE_LIST)
    {
	StackPush(elem->el_Contents.list.head);
	numArgs++;
	elem = elem->el_Contents.list.tail;
    }
    if (elem->el_Type != TYPE_NIL)
	ExecuteAbort(GCS(MSG_ProgErr_ListExpected));

    StackReverse(numArgs);
    ProgCallFunctionElement(func, numArgs, FALSE);
}

void ProgObsoleteFunction(short numArgs)
{
    switch (GetPrefsValue(executeProject, PREFS_PROGRAMOBSOLETEFUNCTIONS)) {
        case OBSOLETE_IGNORE:
            break;
        case OBSOLETE_WARN:
            if (Ask(AT_WARN, GCS(MSG_Prog_WarnObsoleteFunctionAnswers),
                        GCS(MSG_Prog_WarnObsoleteFunction))) {
                break;          /* ignore and continue */
            }
            /* no break */
        case OBSOLETE_ERROR:
            ExecuteAbort(GCS(MSG_CompErr_ObsoleteFunction));
            break;
    }
    StackSqueeze(0, numArgs);
    PushNil();
}

