/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** funccmp.c
**
** Creation date: 28.04.1995
** Last modified: 20.05.2020
*/

#include "system.h"
#include "dbtypes.h"
#include "record.h"
#include "order.h"
#include "memo.h"
#include "func.h"
#include "funccmp.h"
#include "program.h"
#include "programbin.h"
#include "heap.h"
#include "stack.h"
#include "codeset.h"
#include "util.h"
#include "memory.h"
#include "appstrings_cat.h"
#include "error.h"

static long stringcmp(const char *s1, const char *s2, bool star)
{
    if (star) {
        return CodesetStrCollate(s1, s2, CS_COLLATE_SIMPLE);
    } else {
        return strcmp(s1, s2);
    }
}

long ElementCmp(struct Element *elem1, struct Element *elem2, 
        bool star, bool order, struct MemPool *memPool, long *perror)
{
    long result = 0;
    long error = OK;

    if (elem1 == elem2) {
        if (perror) {
            *perror = OK;
        }
        return 0;
    }

    switch (elem1->el_Type)
    {
        case TYPE_BOOL:
            switch (elem2->el_Type)
            {
                case TYPE_NIL:  result = 1; break;
                case TYPE_BOOL: break;
                default:        error = PROGERR_COMPARETYPES; break;
            }
            break;

        case TYPE_INTEGER:
            switch (elem2->el_Type)
            {
                case TYPE_INTEGER:
                {
                    long i1 = elem1->el_Contents.integer;
                    long i2 = elem2->el_Contents.integer;

                    if (i1 > i2) {
                        result = 1;
                    } else if (i1 < i2) {
                        result = -1;
                    }
                    break;
                }

                case TYPE_REAL:
                {
                    double d1 = (double)elem1->el_Contents.integer;
                    double d2 = elem2->el_Contents.real;

                    if (d1 > d2) {
                        result = 1;
                    }
                    else if (d1 < d2) {
                        result = -1;
                    }
                    break;
                }

                case TYPE_NIL:  result = 1; break;
                default:        error = PROGERR_COMPARETYPES; break;
            }
            break;

        case TYPE_REAL:
            switch (elem2->el_Type)
            {
                case TYPE_INTEGER:
                {
                    double d1 = elem1->el_Contents.real;
                    double d2 = (double)elem2->el_Contents.integer;

                    if (d1 > d2) {
                        result = 1;
                    } else if (d1 < d2) {
                        result = -1;
                    }
                    break;
                }

                case TYPE_REAL:
                {
                    double d1 = elem1->el_Contents.real;
                    double d2 = elem2->el_Contents.real;

                    if (d1 > d2) {
                        result = 1;
                    } else if (d1 < d2) {
                        result = -1;
                    }
                    break;
                }

                case TYPE_NIL:  result = 1; break;
                default:        error = PROGERR_COMPARETYPES; break;
            }
            break;

        case TYPE_STRING:
            switch (elem2->el_Type)
            {
                case TYPE_STRING:
                    result = stringcmp(elem1->el_Contents.string.string,
                            elem2->el_Contents.string.string, star);
                    break;
                case TYPE_MEMO:
                    result = stringcmp(elem1->el_Contents.string.string,
                            elem2->el_Contents.mi, star);
                    break;
                case TYPE_NIL:
                    result = 1;
                    break;
                default:
                    error = PROGERR_COMPARETYPES;
                    break;
            }
            break;

        case TYPE_MEMO:
            switch (elem2->el_Type)
            {
                case TYPE_STRING:
                    result = stringcmp(elem1->el_Contents.mi,
                            elem2->el_Contents.string.string, star);
                    break;
                case TYPE_MEMO:
                    result = stringcmp(elem1->el_Contents.mi,
                            elem2->el_Contents.mi, star);
                    break;
                case TYPE_NIL:
                    result = 1;
                    break;
                default:
                    error = PROGERR_COMPARETYPES;
                    break;
            }
            break;


        case TYPE_DATE:
            switch (elem2->el_Type)
            {
                case TYPE_DATE:
                {
                    long d1 = elem1->el_Contents.date;
                    long d2 = elem2->el_Contents.date;

                    if (d1 > d2) {
                        result = 1;
                    } else if (d1 < d2) {
                        result = -1;
                    }
                    break;
                }

                case TYPE_NIL:  result = 1;     break;
                default:        error = PROGERR_COMPARETYPES; break;
            }
            break;

        case TYPE_TIME:
            switch (elem2->el_Type)
            {
                case TYPE_TIME:
                {
                    long t1 = elem1->el_Contents.time;
                    long t2 = elem2->el_Contents.time;

                    if (t1 > t2) {
                        result = 1;
                    } else if (t1 < t2) {
                        result = -1;
                    }
                    break;
                }

                case TYPE_NIL:  result = 1; break;
                default:        error = PROGERR_COMPARETYPES; break;
            }
            break;

        case TYPE_RECORD:
	    switch (elem2->el_Type)
            {
                case TYPE_RECORD:
                {
                    struct Table *tab = elem1->el_Contents.tabrec.tab;
                    struct Record *rec1 = elem1->el_Contents.tabrec.rec;
                    struct Record *rec2 = elem2->el_Contents.tabrec.rec;

                    if (tab != elem2->el_Contents.tabrec.tab) {
                        error = PROGERR_COMPARETYPES;
                    } else if (star) {
                        error = RecordCmp(tab, rec1, rec2, &result);
                    } else if(order) {
                        error = PROGERR_COMPARETYPES;
                    } else {
                        result = rec1 != rec2;
                    }
                    break;
                }

                case TYPE_NIL:  result = 1;     break;
                default:        error = PROGERR_COMPARETYPES; break;
            }
            break;

        case TYPE_FILE:
            switch (elem2->el_Type)
            {
                case TYPE_FILE:
                    if (order) {
                        error = PROGERR_COMPARETYPES;
                    } else {
                        result = (elem1->el_Contents.file.num != 
                                elem2->el_Contents.file.num) ||
                            (elem1->el_Contents.file.id != 
                             elem2->el_Contents.file.id);
                    }
                    break;

                case TYPE_NIL:  result = 1;     break;
                default:        error = PROGERR_COMPARETYPES; break;
            }
            break;

        case TYPE_LIST:
            switch (elem2->el_Type)
            {
                case TYPE_LIST:
                case TYPE_NIL:
                {
                    struct Element *el1, *el2;

                    for (el1 = elem1, el2 = elem2; 
                            el1->el_Type == TYPE_LIST && 
                            el2->el_Type == TYPE_LIST;
                            el1 = el1->el_Contents.list.tail, 
                            el2 = el2->el_Contents.list.tail) {
                        result = ElementCmp(el1->el_Contents.list.head, 
                                el2->el_Contents.list.head, star, order, 
                                memPool, &error);
                        if (result != 0) {
                            break;
                        }
                    }
                    if (result == 0) {
                        if (el1->el_Type == TYPE_LIST && 
                                el2->el_Type != TYPE_LIST) {
                            result = 1;
                        } else if(el1->el_Type != TYPE_LIST && 
                                el2->el_Type == TYPE_LIST){
                            result = -1;
                        }
                    }
                    break;
                }

                default:        error = PROGERR_COMPARETYPES; break;
            }
            break;

        case TYPE_NIL:
            switch (elem2->el_Type)
            {
                case TYPE_NIL:  result = 0;     break;
                default:        result = -1;    break;
            }
            break;

        default: 
	    switch (elem2->el_Type)
	    {
		case TYPE_NIL:  result = 1; 	break;
                default:        error = PROGERR_COMPARETYPES; break;
	    }
	    break;
    }

    if (perror) {
        *perror = error;
    }

    return result;
}

long ProgElementCmp(struct Element *elem1, struct Element *elem2, 
		    bool star, bool order)
{
    long error;
    long result = ElementCmp(elem1, elem2, star, order, executeMemPool, &error);

    if (error) {
        ExecuteAbort(GES(error));
    }

    return result;
}

void ProgEq(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, FALSE, FALSE) == 0)
        PushTrue();
    else
        PushNil();
}

void ProgNe(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, FALSE, FALSE) != 0)
        PushTrue();
    else
        PushNil();
}

void ProgGt(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, FALSE, TRUE) > 0)
        PushTrue();
    else
        PushNil();
}

void ProgLt(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, FALSE, TRUE) < 0)
        PushTrue();
    else
        PushNil();
}

void ProgGe(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, FALSE, TRUE) >= 0)
        PushTrue();
    else
        PushNil();
}

void ProgLe(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, FALSE, TRUE) <= 0)
        PushTrue();
    else
        PushNil();
}

/*
** star compare
*/

void ProgEqStar(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, TRUE, FALSE) == 0)
        PushTrue();
    else
        PushNil();
}

void ProgNeStar(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, TRUE, FALSE) != 0)
        PushTrue();
    else
        PushNil();
}

void ProgGtStar(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, TRUE, TRUE) > 0)
        PushTrue();
    else
        PushNil();
}

void ProgLtStar(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, TRUE, TRUE) < 0)
        PushTrue();
    else
        PushNil();
}

void ProgGeStar(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, TRUE, TRUE) >= 0)
        PushTrue();
    else
        PushNil();
}

void ProgLeStar(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    if(ProgElementCmp(elem1, elem2, TRUE, TRUE) <= 0)
        PushTrue();
    else
        PushNil();
}

void ProgCmp(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    PushInteger(ProgElementCmp(elem1, elem2, FALSE, TRUE));
}

void ProgCmpStar(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    PushInteger(ProgElementCmp(elem1, elem2, TRUE, TRUE));
}

static void progMinMax(short numArgs, bool max, bool star)
{
    struct Element *elem, *elemMinMax = NULL;
    int cmp;

    if (numArgs == 0) {
        PushNil();
        return;
    }

    elemMinMax = StackPop();
    numArgs--;
    while (numArgs > 0) {
        elem = StackPop();
        numArgs--;
        cmp = ProgElementCmp(elem, elemMinMax, star, TRUE);
        if (max ? cmp > 0 : cmp < 0) {
            elemMinMax = elem;
        }
    }
    StackPush(elemMinMax);
}

void ProgMin(short numArgs)
{
    progMinMax(numArgs, FALSE, FALSE);
}

void ProgMinStar(short numArgs)
{
    progMinMax(numArgs, FALSE, TRUE);
}

void ProgMax(short numArgs)
{
    progMinMax(numArgs, TRUE, FALSE);
}

void ProgMaxStar(short numArgs)
{
    progMinMax(numArgs, TRUE, TRUE);
}

