/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** funcdb.c
**
** Creation date: 09.01.1997
** Last modified: 04.06.2022
*/

#include "system.h"
#include "funcdb.h"
#include "func.h"
#include "programbin.h"
#include "stack.h"
#include "heap.h"
#include "compile.h"
#include "gui.h"
#include "order.h"
#include "break.h"
#include "menuproject.h"
#include "requesters.h"
#include "preferences.h"
#include "labels.h"
#include "memory.h"
#include "dosutil.h"
#include "error.h"
#include "appstrings_cat.h"

/*
** Functions for projects
*/

void ProgProjectName(short numArgs)
{
    const char *name = ProjectName(executeProject);

    if(name == NULL)
	PushNil();
    else
	PushString((char *)name, FALSE);
}

void ProgAdminPassword(short numArgs)
{
    const char *passwd = ProjectAdminPassword(executeProject);

    if (passwd == NULL || passwd[0] == 0) {
	PushNil();
    } else {
	PushString((char *)passwd, FALSE);
    }
}

void ProgPrepareChange(short numArgs)
{
    ExecutePrepareProjectChange();
    PushNil();
}

void ProgChanges(short numArgs)
{
    PushInteger(ProjectChanges(executeProject) + (executeProjectChange? 1 : 0));
}

void ProgGetAdminMode(short numArgs)
{
    if (ProjectAdminMode(executeProject)) {
	PushTrue();
    } else {
	PushNil();
    }
}

void ProgSetAdminMode(short numArgs)
{
    struct Program *prg = ProjectProgram(executeProject);
    struct UserFunction *uf = FindUserFunctionName(prg, onAdminMode);
    bool admin = PopBool2();

    SetProjectAdminMode(executeProject, admin);
    if (uf) {
	ProgGetAdminMode(0);		    	/* pushes mode */
	ExecuteUserFunctionAdjustArgs(uf, 1);   /* executes onAdminMode */
	StackPop();			    	/* discard result */
    }
    ProgGetAdminMode(0);			/* pushes mode */
}

/*
** Functions for tables & fields
*/

static struct DisplayObject *diofromelement(struct Element *elem)
{
    struct DisplayObject *dio = NULL;

    switch(elem->el_Type)
    {
        case TYPE_TABLE:
            dio = TableDisplayObject(elem->el_Contents.tab.tab);
            break;

        case TYPE_FIELD:
            dio = FieldDisplayObject(RelFieldPathField(elem->el_Contents.field.rap));
            break;

        case TYPE_NIL:
            break;

        default:
            ExecuteAbort(GCS(MSG_Err_Internal));
            break;
    }

    return(dio);
}

void ProgSetCursor(short numArgs)
{
    struct DisplayObject *dio = diofromelement(StackPop());

    if(GuiShowDisplayObject(dio)) {
        GuiActivateDisplayObject(dio);
        PushTrue();
    } else {
        PushNil();
    }
}

void ProgSetBGPen(short numArgs)
{
    struct DisplayObject *dio = diofromelement(StackPop());
    long pen;
    bool nil;
    char spec[32];
    const char *bg;

    StackPush(StackTop());      /* push return value */
    nil = !PopInteger2(&pen);
    if (nil) {
        bg = NULL;
    } else {
        ulong red, green, blue, v;
        v = (pen & 0xFF0000) >> 16;
        red = (v << 24) | (v << 16) | (v << 8) | v;
        v = (pen & 0xFF00) >> 8;
        green = (v << 24) | (v << 16) | (v << 8) | v;
        v = (pen & 0xFF);
        blue = (v << 24) | (v << 16) | (v << 8) | v;
        sprintf(spec, "2:r%08lx,%08lx,%08lx", red, green, blue);
        bg = spec;
    }
    SetDisplayObjectBackground(dio, bg);
    bg = DisplayObjectBackground(dio);
    GuiSetBackground(DisplayObjectGuiObject(dio), bg);
}

void ProgSetWindowOpen(short numArgs)
{
    struct DisplayObject *dio = diofromelement(StackPop());
    struct DisplayObject *windio = DisplayObjectWindow(dio);
    GuiObject win = DisplayObjectGuiObject(windio);
    struct Project *pr = DisplayObjectProject(windio);
    bool open = PopBool2();

    SetDisplayWindowOpen(windio, open);    /* in case GUI is not opened yet */
    if (open && GuiOpenWindow(win)) {
        GuiWindowToFront(win);
    } else if (!open && windio != ProjectRootDisplayObject(pr)) {
        GuiCloseWindow(win);
    }

    if (GuiWindowOpen(win)) {
        PushTrue();
    } else {
        PushNil();
    }
}

void ProgGetWindowOpen(short numArgs)
{
    struct DisplayObject *dio = diofromelement(StackPop());
    struct DisplayObject *windio = DisplayObjectWindow(dio);
    GuiObject win = DisplayObjectGuiObject(windio);

    if(GuiWindowOpen(win))
        PushTrue();
    else
        PushNil();
}

void ProgSetVirtualListActive(short numArgs)
{
    struct RelFieldPath *rap = PopField2(NULL);
    struct Field *field = RelFieldPathField(rap);
    long num;
    bool nil = !PopInteger2(&num);

    if (nil)
	num = 0;

    if (GuiSetVirtualFieldListActive(FieldDisplayObject(field), num))
	PushInteger(num);
    else
	PushNil();
}

void ProgGetVirtualListActive(short numArgs)
{
    struct RelFieldPath *rap = PopField2(NULL);
    struct Field *field = RelFieldPathField(rap);
    long num = GuiGetVirtualFieldListActive(FieldDisplayObject(field));

    if (num > 0)
	PushInteger(num);
    else
	PushNil();
}

/*
** Functions for tables & records
*/

void ProgRelName(short numArgs)
{
    struct Table *tab = PopTable2(NULL);

    if (tab == NULL) {
        PushNil();
    } else {
        PushString((char *)TableName(tab), FALSE);
    }
}

static void progNew(bool star)
{
    struct Table *tab = PopTable2(NULL);
    struct Project *pr = TableProject(tab);
    char *trigger = NULL;
    long error;
    struct Record *new, *init = NULL, *cur_rec;
    bool countChanges = TableCountChanges(tab);

    if (tab == NULL) {
        StackPop();
        PushNil();
        return;
    }

    if (star) {
        trigger = TableNewTrigger(tab);
    }
    if (trigger != NULL && trigger[0] != 0) {
        CallTriggerFunction(trigger, 1);
        return;
    }

    if (countChanges) {
        ExecutePrepareProjectChange();
    }

    cur_rec = TableProgramRecord(tab);
    init = PopRecord2MatchTable(tab);
    error = NewRecord(tab, init, cur_rec, TRUE, &new);
    if (error) {
        ExecuteAbort(GES(error));
    }

    error = ReorderRecords(tab);
    if (error) {
        ExecuteAbort(GES(error));
    }

    SetTableProgramRecord(tab, new);
    ComputeDependentDisplayObjects(pr, tab, NULL, NULL);
    ComputeDependentDisplayObjects(pr, TableDisplayObject(tab), NULL, NULL);
    if (countChanges) {
        executeProjectChange = TRUE;
    }

    PushRecord(tab, new);
}

static void progDelete(bool star)
{
    struct Table *tab = PopTable2(NULL);
    char *trigger = NULL;
    struct Project *pr;
    struct Record *del;
    bool delreq = FALSE;
    bool delete = TRUE;
    long error;
    bool countChanges = TableCountChanges(tab);

    if (tab == NULL) {
        StackPop();
        PushNil();
        return;
    }

    if (star) {
        trigger = TableDelTrigger(tab);
    }
    if (trigger != NULL && trigger[0] != 0) {
        CallTriggerFunction(trigger, 1);
        return;
    }

    pr = TableProject(tab);
    del = TableProgramRecord(tab);
    delreq = PopBool2();
    if (delreq) {
        delreq = (bool)GetPrefsValue(pr, PREFS_CONFIRMDELETERECORD);
    }

    error = AttemptDeleteRecord(tab, del);
    if (error) {
        ExecuteAbort(GES(error));
    }

    if (delreq) {
        BreakHide();
        delete = (bool)Ask(AT_WARN, GCS(MSG_DeleteAnswers),
                GCS(MSG_DeleteRecord));
        BreakReInit();
    }
    if (delete == FALSE) {
        PushNil();
        return;
    }

    if (countChanges) {
        ExecutePrepareProjectChange();
    }

    error = DeleteRecord(tab, del);
    if (error) {
        ExecuteAbort(GES(error));
    }

    SetTableProgramRecord(tab, NULL);
    ComputeDependentDisplayObjects(pr, tab, NULL, NULL);
    ComputeDependentDisplayObjects(pr, TableDisplayObject(tab), NULL, NULL);
    if (countChanges) {
        executeProjectChange = TRUE;
    }

    PushTrue();
}

void ProgNew(short numArgs)     {    progNew(FALSE); }
void ProgNewStar(short numArgs) {    progNew(TRUE); }
void ProgDelete(short numArgs)  {    progDelete(FALSE); }
void ProgDeleteStar(short numArgs) { progDelete(TRUE); }

void ProgDeleteAll(short numArgs)
{
    bool star;
    struct Table *tab = PopTable2(&star);
    struct Project *pr = TableProject(tab);
    long error;
    bool countChanges = TableCountChanges(tab);

    if (tab == NULL) {
	PushNil();
    } else {
        if (countChanges) {
            ExecutePrepareProjectChange();
        }

	error = DeleteRecordsTable(tab, star, FALSE);
	if (error) {
	    ExecuteAbort(GES(error));
        }
	SetTableProgramRecord(tab, NULL);
	ComputeDependentDisplayObjects(pr, tab, NULL, NULL);
        ComputeDependentDisplayObjects(pr, TableDisplayObject(tab), NULL, NULL);
        if (countChanges) {
            executeProjectChange = TRUE;
        }
	PushTrue();
    }
}

void ProgGetMatchFilter(short numArgs)
{
    struct Record *rec = PopRecord2(NULL);

    if(RecordMatchFilter(rec))
        PushTrue();
    else
        PushNil();
}

void ProgSetMatchFilter(short numArgs)
{
    struct Table *tab = NULL;
    struct Record *rec = PopRecord2(&tab);
    bool set = PopBool2();

    if (TableFilterActive(tab)) {
        SetRecordMatchFilter(rec, set);
        GuiCheckGuiRecord(tab);
    }
    if (RecordMatchFilter(rec)) {
        PushTrue();
    } else {
        PushNil();
    }
}

void ProgGetFilterActive(short numArgs)
{
    struct Table *tab = PopTable2(NULL);

    if(TableFilterActive(tab))
        PushTrue();
    else
        PushNil();
}

static void setfilteractivehelp(struct Table *tab, bool on)
{
    struct List *recs = TableRecords(tab);
    struct Record *rec; 

    if (on) {
        struct Program *prg = TableFilterProgram(tab);
        const char *error = NULL;

        if (ProgramNeedRecompile(prg)) {
            struct SourceText *st = GetSourceText(prg);
            const char *text = SourceTextText(st);
            struct Program *compiled = NULL;

            if (text == NULL) {
                error = GES(ERR_MEM);
            } else {
                error = CompileFilter(tab, text, strlen(text), &compiled, 
                        NULL, NULL);
            }
            FreeSourceText(st);
            if (compiled) {
                SetTableFilterProgram(tab, compiled);
            }
            prg = compiled;
        }

        if (error != NULL || recs == NULL) {
            on = FALSE;
        } else {
            struct Record *save = TableProgramRecord(tab);
            struct Bin *bin = ProgramBin(prg);
            struct Element *set = NULL;

            FORNODE(rec, recs) {
                SetTableProgramRecord(tab, rec);
                if (bin != NULL) {
                    error = ExecuteEvalBin(bin, &set);
                    if (error) {
                        on = FALSE;
                        break;
                    }
                }
                SetRecordMatchFilter(rec, set==NULL || set->el_Type!=TYPE_NIL);
            }
            SetTableProgramRecord(tab, save);
        }
    }
    if (!on) {
        if (recs != NULL) {
            FORNODE(rec, recs) {
                SetRecordMatchFilter(rec, TRUE);
            }
        }
    }

    SetTableFilterActive(tab, on);
    GuiCheckGuiRecord(tab);
}

void ProgSetFilterActive(short numArgs)
{
    struct Table *tab = PopTable2(NULL);
    bool set = PopBool2();

    setfilteractivehelp(tab, set);
    if(TableFilterActive(tab))
        PushTrue();
    else
        PushNil();
}

void ProgGetFilterStr(short numArgs)
{
    struct Table *tab = PopTable2(NULL);

    if(tab == NULL)
        PushNil();
    else
    {
        struct Program *prg = TableFilterProgram(tab);
        struct SourceText *st = NULL;
        const char *text = "";
        char *copy;

        if(prg != NULL)
        {
            st = GetSourceText(prg);
            text = SourceTextText(st);
        }
        copy = AllocString(executeMemPool, text);
        FreeSourceText(st);

        if (copy == NULL)
            ExecuteAbort(GCS(MSG_Err_Mem));
        PushString(copy, TRUE);
    }
}

void ProgSetFilterStr(short numArgs)
{
    struct Table *tab = PopTable2(NULL);
    char *filter = PopString2();

    if(tab == NULL || filter == NULL)
        PushNil();
    else
    {
        struct Program *prg = NULL;
        const char *error = NULL;

        error = CompileFilter(tab, filter, strlen(filter), &prg, NULL, NULL);
        if(error == NULL)
        {
            SetTableFilterProgram(tab, prg);

            if(TableFilterActive(tab))
                setfilteractivehelp(tab, TRUE);
            PushTrue();
        }
        else
            PushNil();
    }
}


void ProgGetIsSorted(short numArgs)
{
  struct Record *rec = PopRecord2(NULL);

  if(RecordIsSorted(rec))
    PushTrue();
  else
    PushNil();
}

void ProgSetIsSorted(short numArgs)
{
    struct Table *tab = NULL;
    struct Record *rec = PopRecord2(&tab);
    bool set = PopBool2();
    bool countChanges;

    countChanges = TableCountChanges(tab);
    if (countChanges) {
        ExecutePrepareProjectChange();
    }

    SetRecordIsSorted(rec, set);
    if (countChanges) {
        executeProjectChange = TRUE;
    }

    if (RecordIsSorted(rec)) {
        PushTrue();
    } else {
        PushNil();  
    }
}

void ProgReorder(short numArgs)
{
    struct Table *tab = PopTable2(NULL);
    bool countChanges = TableCountChanges(tab);
    long error;
  
    if (countChanges) {
        ExecutePrepareProjectChange();
    }

    error = ReorderRecords(tab);
    GuiDisplayPanel(tab);
    if (countChanges) {
        executeProjectChange = TRUE;
    }

    if (error) {
	ExecuteAbort(GES(error));
    }

    PushNil();
}
  
void ProgReorderAll(short numArgs)
{
    struct Table *tab = PopTable2(NULL);
    bool countChanges = TableCountChanges(tab);
    long error;

    if (countChanges) {
        ExecutePrepareProjectChange();
    }

    error = ReorderAllRecords(tab);
    GuiDisplayPanel(tab);
    if (countChanges) {
        executeProjectChange = TRUE;
    }
    
    if (error) {
	ExecuteAbort(GES(error));
    }
    
    PushNil();
}

void ProgGetOrderStr(short numArgs)
{
    struct Table *tab = PopTable2(NULL);
    char *order;

    if(tab == NULL)
    {
        PushNil();
        return;
    }

    order = GetOrderStr(executeMemPool, tab);
    if(order == NULL)
        ExecuteAbort(GCS(MSG_Err_Mem));
    PushString(order, TRUE);
}

void ProgSetOrderStr(short numArgs)
{
    struct Table *tab = PopTable2(NULL);
    char *order = PopString2();
    bool countChanges = TableCountChanges(tab);
    long error;

    if (tab == NULL || order == NULL) {
        PushNil();
    } else {
        if (countChanges) {
            ExecutePrepareProjectChange();
        }
        error = SetOrderStr(tab, order);
        if (countChanges) {
            executeProjectChange = TRUE;
        }
	if (error) {
            PushNil();
        } else {
            PushTrue();
        }
    }
}

void ProgRecNum(short numArgs)
{
    PushInteger(RecordNum(PopRecord2(NULL)));
}

void ProgMoveRec(short numArgs)
{
    struct Table *tab = NULL;
    struct Record *rec = PopRecord2(&tab);
    struct Project *pr = TableProject(tab);
    long num;
    bool nil = !PopInteger2(&num);
    bool countChanges;

    countChanges = TableCountChanges(tab);
    if (countChanges) {
        ExecutePrepareProjectChange();
    }

    if (nil || !MoveRecord(tab, rec, num)) {
        PushNil();
    } else {
        if (countChanges) {
            executeProjectChange = TRUE;
        }
        ComputeDependentDisplayObjects(pr, tab, NULL, NULL);
        ComputeDependentDisplayObjects(pr, TableDisplayObject(tab), NULL, NULL);
        PushInteger(RecordNum(rec));
    }
}

void ProgGetRec(short numArgs)
{
    struct Element *elem = StackPop();
    struct Record *rec = elem->el_Record;

    while (rec == NULL && elem->el_Type == TYPE_LIST) {
        rec = elem->el_Contents.list.head->el_Record;
        elem = elem->el_Contents.list.tail;
    }
    if (rec != NULL) {
        PushRecord(RecordFindTable(rec, NULL, executeProject), rec);
    } else {
        PushNil();
    }
}

void ProgSetRec(short numArgs)
{
    struct Element *elem = StackPop();
    struct Record *rec = PopRecord2(NULL);

    elem->el_Record = rec;
    StackPush(elem);
}

void ProgRecord(short numArgs)
{
    bool star = FALSE;
    struct Table *tab = PopTable2(&star);
    long num;
    bool nil = !PopInteger2(&num);
    struct Record *rec = RecordNth(tab, star, nil? 0 : num);

    if(rec == NULL)
        PushNil();
    else
        PushRecord(tab, rec);
}

void ProgRecords(short numArgs)
{
    bool star = FALSE;
    struct Table *tab = PopTable2(&star);
    ulong all;

    if(tab == NULL)
        PushNil();
    else
    {
        RecordNumAll(tab, NULL, star, NULL, &all);
        PushInteger(all);
    }
}

void ProgCopyRec(short numArgs)
{
    struct Table *tab = NULL;
    struct Record *rec = PopRecord2(&tab);
    struct Record *init = PopRecord2MatchTable(tab);
    long error = OK;
    bool countChanges;

    countChanges = TableCountChanges(tab);
    if (countChanges) {
        ExecutePrepareProjectChange();
    }
    error = CopyRecord(tab, rec, init);
    if (error) {
        ExecuteAbort(GES(error));
    } 
    if (countChanges) {
        executeProjectChange = TRUE;
    }
    error = ReorderRecords(tab);
    if (error) {
        ExecuteAbort(GES(error));
    }

    if (rec == TableGuiRecord(tab)) {
        GuiShowRecord(tab, rec);
    }

    if (rec == NULL) {
        PushNil();
    } else {
        PushRecord(tab, rec);
    }
}

/*
** Functions for fields
*/

void ProgFieldName(short numArgs)
{
    struct RelFieldPath *rap = PopField2(NULL);
    const char *name = FieldName(RelFieldPathField(rap));

    if(name == NULL)
        PushNil();
    else
        PushString((char *)name, FALSE);
}

void ProgSetLabels(short numArgs)
{
    struct RelFieldPath *rap = PopField2(NULL);
    struct Element *elem = PopString();
    struct Field *field = RelFieldPathField(rap);
    struct DisplayObject *dio = FieldDisplayObject(field);
    char **labels;
    long error;
    bool countChanges = FieldCountChanges(field) &&
        TableCountChanges(FieldTable(field));

    if (rap == NULL || elem->el_Type != TYPE_STRING) {
        PushNil();
        return;
    }

    labels = StringToLabels(executeMemPool, elem->el_Contents.string.string);
    if (labels == NULL) {
        ExecuteAbort(GES(ERR_MEM));
    }

    if (countChanges) {
        ExecutePrepareProjectChange();
    }

    error = SetDisplayFieldListviewLabels(dio, labels);
    FreeLabels(executeMemPool, labels);
    if (error) {
        ExecuteAbort(GES(error));
    }

    if (countChanges) {
        executeProjectChange = TRUE;
    }
    StackPush(elem);
}

void ProgGetLabels(short numArgs)
{
    struct RelFieldPath *rap = PopField2(NULL);
    struct Field *field = RelFieldPathField(rap);
    char **labels;
    char *str = NULL;

    if(field == NULL)
    {
        PushNil();
        return;
    }

    if(FieldType(field) == TYPE_CHOICE)
        labels = FieldChoiceLabels(field);
    else
        labels = DisplayFieldListviewLabels(FieldDisplayObject(field));

    if(labels == NULL)
        PushString("", FALSE);
    else
    {
        str = LabelsToString(executeMemPool, labels);
        if(str == NULL)
            ExecuteAbort(GES(ERR_MEM));

        PushString(str, TRUE);
    }
}

/*
** Functions for path handling.
*/

void ProgTackon(short numArgs)
{
    struct Element *top = StackTop();
    char *path = PopString2();

    numArgs--;
    if (numArgs == 0) {
        StackPush(top);         /* put argument back */
    } else if (path == NULL) {
        PopAndPushNil(numArgs);
    } else {
        char *file, *old = NULL;

        do {
            file = PopString2();
            numArgs--;
            if (file == NULL) {
                path = NULL;
                break;
            }
            path = JoinPathname(executeMemPool, path, file);
            FreeString(executeMemPool, old);
            old = path;
            if (path == NULL) {
                ExecuteAbort(GCS(MSG_Err_Mem));
            }
        } while (numArgs > 0);

        if (path == NULL) {
            FreeString(executeMemPool, old);
            PopAndPushNil(numArgs);
        } else {
            PushString(path, TRUE);
        }
    }
}

void ProgFilename(short numArgs)
{
    char *path = PopString2(), *file;
    
    if(!SplitPathname(path, executeMemPool, NULL, &file))
	ExecuteAbort(GCS(MSG_Err_Mem));
    else if(file == NULL)
	PushNil();
    else
	PushString(file, TRUE);
}

void ProgDirname(short numArgs)
{
    char *path = PopString2(), *dir;
    
    if(!SplitPathname(path, executeMemPool, &dir, NULL))
	ExecuteAbort(GCS(MSG_Err_Mem));
    else if(dir == NULL)
	PushNil();
    else
	PushString(dir, TRUE);
}

void ProgPubScreen(short numArgs)
{
#ifdef AMIGA
    const char *screen = GuiPubScreenName();

    if (screen) {
        PushString(ExecuteAllocString(screen), TRUE);
    } else {
        PushNil();
    }
#else
    PushNil();
#endif
}

