/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** funcio.c
**
** Creation date: 29.01.1996
** Last modified: 06.10.2014
*/

#include "system.h"
#include "funcio.h"
#include "programbin.h"
#include "global.h"
#include "stack.h"
#include "heap.h"
#include "strhandle.h"
#include "memo.h"
#include "textfile.h"
#include "requesters.h"
#include "preferences.h"
#include "error.h"
#include "appstrings_cat.h"

#define NUM_PROGFILEHANDLES 16           /* maximum number of open files */

static struct ProgFileHandle
{
    TextFile *pfh_TextFile;
    ulong pfh_Id;
    const char *pfh_Filename;           /* for delayed open */
    const char *pfh_Mode;               /* for delayed open */
} ProgFileHandle[NUM_PROGFILEHANDLES];

enum
{
    PROGFILEHANDLE_STDOUT,      /* predefined file handles */
    PROGFILEHANDLE_USER,        /* file handles for user */
};

struct Element ProgStdOut;
static ulong ProgFileId = 0;

void InitFileIO(struct Project *pr)
{
    int i;

    ProgStdOut.el_Type = TYPE_FILE;
    ProgStdOut.el_Allocated = FALSE;
    ProgStdOut.el_InUse = TRUE;
    ProgStdOut.el_Record = NULL;
    ProgStdOut.el_Contents.file.num = PROGFILEHANDLE_STDOUT;
    ProgStdOut.el_Contents.file.id = PROGFILEHANDLE_STDOUT;

    for (i=0; i < NUM_PROGFILEHANDLES; i++) {
        ProgFileHandle[i].pfh_TextFile = NULL;
        ProgFileHandle[i].pfh_Filename = NULL;
    }

    ProgFileHandle[PROGFILEHANDLE_STDOUT].pfh_Id = PROGFILEHANDLE_STDOUT;
    ProgFileHandle[PROGFILEHANDLE_STDOUT].pfh_Filename =
        GetPrefsValue(pr, PREFS_PROGRAMOUTPUT_DEFAULT)?
        DEFAULT_OUTPUT_FILE : GetPrefsString(pr, PREFS_PROGRAMOUTPUT_FILENAME);
    ProgFileHandle[PROGFILEHANDLE_STDOUT].pfh_Mode =
        GetPrefsValue(pr, PREFS_PROGRAMOUTPUT_APPEND)? "a+" : "w";

    ProgFileId = PROGFILEHANDLE_USER;           /* first free id */
}

void ExitFileIO(void)
{
    size_t warnings;
    int i;

    for (i=0; i < NUM_PROGFILEHANDLES; i++) {
        if (ProgFileHandle[i].pfh_TextFile != NULL) {
            warnings = TextFileWarnings(ProgFileHandle[i].pfh_TextFile);
            if (warnings) {
                Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Warn_CodesetProcessFile),
                        warnings, ProgFileHandle[i].pfh_Filename);
            }
            TextFileClose(ProgFileHandle[i].pfh_TextFile);
            ProgFileHandle[i].pfh_TextFile = NULL;
            ProgFileHandle[i].pfh_Filename = NULL;
        }
    }
}

static struct ProgFileHandle *GetElementPfh(struct Element *elem)
{
    short num;
    ulong id;
    struct ProgFileHandle *pfh = NULL;

    if (elem == NULL) {
        ExecuteAbort(GCS(MSG_Err_Internal));
    }
    if (elem->el_Type == TYPE_NIL) {
        return NULL;                            /* /dev/null, NIL: device */
    }
    if (elem->el_Type != TYPE_FILE) {
        ExecuteAbort(GCS(MSG_ProgErr_FileExpected));
    }

    num = elem->el_Contents.file.num;
    id = elem->el_Contents.file.id;
    if (num < 0 || num >= NUM_PROGFILEHANDLES) {
        ExecuteAbort(GCS(MSG_ProgErr_InvalidFile));
    }

    pfh = &ProgFileHandle[num];
    if (pfh->pfh_Id != id) {
        ExecuteAbort(GCS(MSG_ProgErr_InvalidFile));
    }
    if (pfh->pfh_TextFile != NULL) {
        return pfh;
    }
    if (pfh->pfh_Filename == NULL) {
        ExecuteAbort(GCS(MSG_ProgErr_InvalidFile));
    }

    return pfh;
}

TextFile *GetElementTextFile(struct Element *elem)
{
    struct ProgFileHandle *pfh = GetElementPfh(elem);

    if (pfh == NULL) {
        return NULL;           /* element = NIL */
    }

    if (pfh->pfh_TextFile != NULL) {
        return pfh->pfh_TextFile;
    }

    /* Try to open file */
    ProjectSetCurrentDir(executeProject);    /* make sure we have correct CWD */
    pfh->pfh_TextFile = TextFileOpen(executeMemPool, 
            pfh->pfh_Filename, pfh->pfh_Mode, TFE_AUTO);
    if (pfh->pfh_TextFile == NULL) {
        ExecuteAbort(GCS(MSG_ProgErr_OpenOutputFile));
    }

    return pfh->pfh_TextFile;
}

TextFile *ProgOutputTextFile(void)
{
    return GetElementTextFile(GetGlobalVar(PROGSTDOUT_VARNAME));
}

/*
** BeeBase program functions
*/

struct {
    const char *encoding;
    int tfe;
} encmap[] = {
    { "none", TFE_NONE, },
    { "auto", TFE_AUTO, },
    { "UTF-8", TFE_UTF8, },
    { "locale", TFE_LOCALE, },
    { "8-bit", TFE_8BIT, },
};
#define NUM_ENC (sizeof(encmap) / sizeof(encmap[0]))

void ProgFOpen(short numArgs)
{
    const char *name = PopString2();
    const char *mode = PopString2();
    const char *encoding;
    struct ProgFileHandle *pfh = ProgFileHandle;
    int tfe = TFE_AUTO;
    int i;

    if (numArgs >= 3) {
        encoding = PopString2();
        if (encoding == NULL) {
            PushNil();
            return;
        }
        for (i=0; i < NUM_ENC; i++) {
            if (CodesetStrCmp(encoding, encmap[i].encoding) == 0) {
                tfe = encmap[i].tfe;
                break;
            }
        }
        if (i >= NUM_ENC) {
            ExecuteAbort(GCS(MSG_ProgErr_UnknownEncoding));
        }
    }
    if (name == NULL || mode == NULL) {
        PushNil();
        return;
    }

    /* find free entry in file table */
    for (i=0; i < NUM_PROGFILEHANDLES; i++, pfh++) {
        if (pfh->pfh_TextFile == NULL && pfh->pfh_Filename == NULL) {
            break;
        }
    }
    if (i >= NUM_PROGFILEHANDLES) {
        PushNil();                           /* out of file handles */
        return;
    }

    ProjectSetCurrentDir(executeProject);    /* make sure we have correct CWD */
    pfh->pfh_TextFile = TextFileOpen(executeMemPool, name, mode, tfe);
    if (pfh->pfh_TextFile == NULL) {
        PushNil();                      /* can't open file */
        return;
    }

    pfh->pfh_Filename = ExecuteAllocString(name);       /* remember filename */
    pfh->pfh_Id = ProgFileId++;

    PushFile(i, pfh->pfh_Id);
}

void ProgFClose(short numArgs)
{
    struct Element *elem = StackPop();
    struct ProgFileHandle *pfh = GetElementPfh(elem);
    long error;
    size_t warnings;

    if (pfh == NULL) {                  /* element = NIL? */
        PushInteger(0);
        return;
    }
    warnings = TextFileWarnings(pfh->pfh_TextFile);
    error = TextFileClose(pfh->pfh_TextFile);
    if (error == OK) {
        PushInteger(0);
    } else {
        PushNil();
    }

    if (warnings) {
        Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Warn_CodesetProcessFile),
                warnings, pfh->pfh_Filename);
    }
    pfh->pfh_TextFile = NULL;
    pfh->pfh_Filename = NULL;
}

void ProgFError(short numArgs)
{
    struct Element *elem = StackPop();
    TextFile *tf = GetElementTextFile(elem);

    if (TextFileError(tf)) {
        PushTrue();
    } else {
        PushNil();
    }
}

void ProgFFlush(short numArgs)
{
    struct Element *elem = StackPop();
    TextFile *tf = GetElementTextFile(elem);

    if (TextFileFlush(tf) == OK) {
        PushInteger(0);
    } else {
        PushNil();
    }
}

void ProgFEof(short numArgs)
{
    struct Element *elem = StackPop();
    TextFile *tf = GetElementTextFile(elem);

    if (TextFileEOF(tf)) {
        PushTrue();
    } else {
        PushNil();
    }
}

void ProgFSeek(short numArgs)
{
    struct Element *elem = StackPop();
    TextFile *tf = GetElementTextFile(elem);
    long offset, whence; 
    bool offset_nil = !PopInteger2(&offset);
    bool whence_nil = !PopInteger2(&whence);

    if (offset_nil || whence_nil) {
        PushNil();
    } else if (!TextFileSeek(tf, offset, whence)) {
        PushNil();
    } else {
        PushInteger(0);
    }
}

void ProgFTell(short numArgs)
{
    struct Element *elem = StackPop();
    TextFile *tf = GetElementTextFile(elem);
    long pos = TextFileTell(tf);

    if (pos < 0) {
        PushNil();
    } else {
        PushInteger(pos);
    }
}
    
/*
** Read functions
*/

static void ProgFGetCharsHelp(long num)
{
    struct Element *elem = StackPop();
    TextFile *tf = GetElementTextFile(elem);
    struct StrHandle *sh;
    cschar c;
    long err = OK;
    char buf[CodesetMaxBytesForChar+1];
    short len;
    long i;

    if (tf == NULL) {
        PushNil();
        return;
    }

    sh = AllocStrHandle(executeMemPool);
    if (sh == NULL) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }

    for (i=0; i < num; i++) {
        if (!TextFileGetChar(tf, &c, &err)) {
            break;
        }
        len = CodesetCharToStr(c, buf);
        if (len < 0) {
            buf[0] = '?';
            len = 1;
        }
        buf[len] = 0;
        StrHandlePutS(sh, buf);
    }
    StrHandlePutC(sh, 0);
    if (StrHandleError(sh)) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }

    if (i > 0 || num == 0) {
        PushString(ExecuteAllocString(StrHandleString(sh)), TRUE);
    } else {
        PushNil();
    }
    FreeStrHandle(sh);

    if (err && err != DISKERR_READ) {
        ExecuteAbort(GES(err));
    }
}

void ProgFGetChar(short numargs)
{
    ProgFGetCharsHelp(1);
}

void ProgFGetChars(short numargs)
{
    long len;
    bool nil = !PopInteger2(&len);

    if (nil) {
        StackPop();     /* remove file argument */
        PushNil();
    } else {
        ProgFGetCharsHelp(len);
    }
}

void ProgFGetStr(short numargs)
{
    struct Element *elem = StackPop();
    TextFile *tf = GetElementTextFile(elem);
    struct StrHandle *sh;
    long err;
    int c;
    const char *s;

    sh = AllocStrHandle(executeMemPool);
    if (sh == NULL) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }

    for (;;) {
        c = TextFileGetC(tf, &err);
        if (c == 0 || c == EOF || c == '\n') {
            break;
        }
        StrHandlePutC(sh, c);
    }
    StrHandlePutC(sh, 0);
    if (err && err != DISKERR_READ) {
        ExecuteAbort(GES(err));
    }
    if (StrHandleError(sh)) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }
    s = StrHandleString(sh);
    if (c == EOF && s[0] == 0) {
        PushNil();
    } else {
        PushString(ExecuteAllocString(s), TRUE);
    }
    FreeStrHandle(sh);
}

void ProgFGetMemo(short numargs)
{
    struct Element *elem = StackPop();
    TextFile *tf = GetElementTextFile(elem);
    char *str;
    long err;

    str = TextFileReadToEnd(executeMemPool, tf, &err);
    if (err && err != DISKERR_READ) {
        ExecuteAbort(GES(err));
    }
    if (str && str[0] == 0) {
        FreeString(executeMemPool, str);
        str = NULL;
    }
    if (str) {
        PushString(str, TRUE);
        StackPush(PopMemo());      /* convert to Memo */
    } else {
        PushNil();
    }
}

void ProgFPutChar(short numargs)
{
    struct Element *elem = PopString();
    struct Element *file = StackPop();
    TextFile *tf = GetElementTextFile(file);
    cschar c;
    long err;

    if (tf == NULL || elem->el_Type != TYPE_STRING) {
        PushNil();
    } else {
        c = CodesetGetChar(elem->el_Contents.string.string);
        TextFilePutChar(tf, c, &err);
        if (err == OK) {
            StackPush(elem);
        } else if (err == DISKERR_WRITE) {
            PushNil();
        } else {
            ExecuteAbort(GES(err));
        }
    }
}

static void ProgFPutHelp(bool newline)
{
    struct Element *elem = StackTop();
    const char *s = PopString2();
    struct Element *file = StackPop();
    TextFile *tf = GetElementTextFile(file);
    long err;

    if (tf == NULL || s == NULL) {
        PushNil();
    } else {
        TextFilePutS(tf, s, &err);
        if (newline && err == OK) {
            TextFilePutChar(tf, '\n', &err);
        }
        if (err == OK) {
            StackPush(elem);
        } else if (err == DISKERR_WRITE) {
            PushNil();
        } else {
            ExecuteAbort(GES(err));
        }
    }
}

void ProgFPutStr(short numargs)
{
    ProgFPutHelp(TRUE);
}

void ProgFPutMemo(short numargs)
{
    ProgFPutHelp(FALSE);
}

