/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** funclist.c
**
** Creation date: 28.01.1997
** Last modified: 06.03.2013
*/

#include "system.h"
#include "dbtypes.h"
#include "memory.h"
#include "compile_impl.h"
#include "program.h"
#include "programbin.h"
#include "stack.h"
#include "heap.h"
#include "func.h"
#include "funclist.h"
#include "functable.h"
#include "appstrings_cat.h"

void ProgCons(short numArgs)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackPop();

    PushList(elem1, elem2);
}

static void progMakeListSizeTWithTail(size_t numArgs, struct Element *tail)
{
    struct Element *list = tail;

    for (; numArgs; numArgs--)
	list = MakeList(StackPop(), list);

    StackPush(list);
}

void ProgMakeListSizeT(size_t numArgs)	/* size_t version */
{
    progMakeListSizeTWithTail(numArgs, MakeNil());
}

void ProgMakeList(short numArgs)
{
    ProgMakeListSizeT(numArgs);
}

static void progListSizeTWithTail(size_t numArgs, struct Element *tail)
{
    StackReverse(numArgs);
    progMakeListSizeTWithTail(numArgs, tail);
}

static void progListSizeT(size_t numArgs)		/* size_t version */
{
    progListSizeTWithTail(numArgs, MakeNil());
}

void ProgList(short numArgs)
{
    progListSizeT(numArgs);
}

void ProgFirst(short numArgs)
{
    struct Element *head = PopList2(NULL);

    if(head == NULL)
        PushNil();
    else
        StackPush(head);
}

void ProgRest(short numArgs)
{
    struct Element *tail, *head = PopList2(&tail);

    if(head == NULL)
        PushNil();
    else
        StackPush(tail);
}

void ProgLast(short numArgs)
{
    struct Element *tail, *head = PopList2(&tail);

    if(head == NULL)
        PushNil();
    else
    {
        while(tail->el_Type == TYPE_LIST)
        {
            head = tail->el_Contents.list.head;
            tail = tail->el_Contents.list.tail;
        }
        StackPush(head);
    }
}

void ProgNth(short numArgs)
{
    long n;
    bool nil = !PopInteger2(&n);
    struct Element *tail, *head = PopList2(&tail);

    if(nil || head == NULL)
        PushNil();
    else
    {
        for(; tail->el_Type == TYPE_LIST && n != 0; n--)
        {
            head = tail->el_Contents.list.head;
            tail = tail->el_Contents.list.tail;
        }

        if(n != 0)
            PushNil();
        else
            StackPush(head);
    }
}

void ProgReplaceNth(short numArgs)
{
    long n, i;
    bool nil = !PopInteger2(&n);
    struct Element *elem = StackPop();
    struct Element *tail, *head = PopList2(&tail);

    if (nil || head == NULL) {
        PushNil();
    } else {
        for (i = 0; i != n && tail->el_Type == TYPE_LIST; i++) {
            StackPush(head);
            head = tail->el_Contents.list.head;
            tail = tail->el_Contents.list.tail;
        }
        if (i != n) {
            StackSqueeze(0, i);
            PushNil();
        } else {
            StackPush(elem);
            progMakeListSizeTWithTail(i + 1, tail);
        }
    }
}

void ProgReplaceNthStar(short numArgs)
{
    struct Element *list = StackNth(2);

    ProgReplaceNth(numArgs);
    if (StackTop()->el_Type == TYPE_NIL) {
        StackPop();
        StackPush(list);
    }
}

void ProgMoveNth(short numArgs)
{
    long n, m, i, j = 0;
    bool nil1 = !PopInteger2(&n);
    bool nil2 = !PopInteger2(&m);
    struct Element *tail, *head = PopList2(&tail), *elem = NULL;

    if (nil1 || nil2 || head == NULL) {
        PushNil();
        return;
    }

    for (i = 0; i != m; i++) {
        if (i == n) {
            elem = head;
        } else {
            StackPush(head);
            j++;
        }
        if (tail->el_Type != TYPE_LIST) {
            StackSqueeze(0, j);
            PushNil();
            return;
        }
        head = tail->el_Contents.list.head;
        tail = tail->el_Contents.list.tail;
    }
    if (elem == NULL) {
        PushNil();  /* placeholder for elem */
        j++;
        for (; i != n; i++) {
            StackPush(head);
            j++;
            if (tail->el_Type != TYPE_LIST) {
                StackSqueeze(0, j);
                PushNil();
                return;
            }
            head = tail->el_Contents.list.head;
            tail = tail->el_Contents.list.tail;
        }
        elem = head;
        SetStackNth(n - m, elem);
    } else {
        StackPush(head);
        j++;
        StackPush(elem);
        j++;
    }
    progMakeListSizeTWithTail(j, tail);
}

void ProgMoveNthStar(short numArgs)
{
    struct Element *list = StackNth(2);

    ProgMoveNth(numArgs);
    if (StackTop()->el_Type == TYPE_NIL) {
        StackPop();
        StackPush(list);
    }
}

void ProgRemoveNth(short numArgs)
{
    long n, i;
    bool nil = !PopInteger2(&n);
    struct Element *tail, *head = PopList2(&tail);

    if (nil || head == NULL) {
        PushNil();
    } else {
        for (i = 0; i != n && tail->el_Type == TYPE_LIST; i++) {
            StackPush(head);
            head = tail->el_Contents.list.head;
            tail = tail->el_Contents.list.tail;
        }
        if (i != n) {
            StackSqueeze(0, i);
            PushNil();
        } else {
            progMakeListSizeTWithTail(i, tail);
        }
    }
}

void ProgRemoveNthStar(short numArgs)
{
    struct Element *list = StackNth(1);

    ProgRemoveNth(numArgs);
    if (StackTop()->el_Type == TYPE_NIL) {
        StackPop();
        StackPush(list);
    }
}

size_t CountListLength(const struct Element *elem)
{
    size_t length = 0;

    if (elem != NULL) {
        for (; elem->el_Type == TYPE_LIST; elem = elem->el_Contents.list.tail) {
            length++;
        }
    }

    return length;
}

void ProgLength(short numArgs)
{
    PushInteger(CountListLength(PopList()));
}

void ProgAppend(short numArgs)
{
    long numNodes = 0;
    short i;
    struct Element *elem, **elems;

    if(numArgs == 0)
        PushNil();
    else
    {
        /* allocate local array */
        elems = ExecuteAllocMem(numArgs * sizeof(struct Element *));

        /* save elements in local array */
        for(i=0; i < numArgs; i++)
            elems[i] = PopList();

        /* push all list elements on stack, except elements of last list */
        for(i=0; i < numArgs - 1; i++)
            for(elem = elems[i]; elem->el_Type == TYPE_LIST; elem = elem->el_Contents.list.tail)
            {
                StackPush(elem->el_Contents.list.head);
                numNodes++;
            }

        /* push last list on stack */
        StackPush(elems[i]);

        /* make list */
        for(; numNodes; numNodes--)
        {
            StackReverse(2);                /* exchange top elements */
            ProgCons(2);
        }

        /* free local array */
        PoolFreeMem(executeMemPool, elems, numArgs * sizeof(struct Element *));
    }
}

void ProgReverse(short numArgs)
{
    struct Element *elem = PopList();
    size_t numNodes = 0;

    /* put all list elements on stack */
    for(; elem->el_Type == TYPE_LIST; elem = elem->el_Contents.list.tail)
    {
        StackPush(elem->el_Contents.list.head);
        numNodes++;
    }

    progListSizeT(numNodes);
}

const char *CheckMapFirst(struct FuncInfo *fi, struct AppArgs *aa)
{
    if(aa == NULL)
        return(GCS(MSG_ProgErr_MissingArguments));
    if(GenType(aa->Expr->Type, TYPE_FUNCTION) == NOTYPE)
        return(GCS(MSG_ProgErr_FunctionExpected));

    for(aa = aa->Next; aa; aa = aa->Next)
        if(GenType(aa->Expr->Type, TYPE_LIST) == NOTYPE)
            return(GCS(MSG_ProgErr_ListExpected));

    return(NULL);
}

void ProgMapFirst(short numArgs)
{
    struct Element *func = PopFunction();
    struct Element **elems = NULL;
    short i; 
    size_t num = 0;
    bool all_nil;

    numArgs--;
    if(numArgs == 0)
    {
        PushNil();
        return;
    }

    elems = ExecuteAllocMem(numArgs * sizeof(struct Element *));
    for(i=0; i < numArgs; i++)
        elems[i] = PopList();

    for(i=0; i < numArgs; i++)
        StackPush(elems[i]);            /* put reference on stack */
    StackPush(func);

    do
    {
        all_nil = TRUE;
        for(i=0; i < numArgs; i++)
        {
            if(elems[i]->el_Type != TYPE_LIST)
                PushNil();
            else
            {
                StackPush(elems[i]->el_Contents.list.head);
                elems[i] = elems[i]->el_Contents.list.tail;
                all_nil = FALSE;
            }
        }

        if(!all_nil)
        {
            ProgCallFunctionElement(func, numArgs, TRUE);
            num++;
        }
    } while(!all_nil);

    StackSqueeze(0, numArgs);
    ProgMakeListSizeT(num);
    StackSqueeze(1, numArgs + 1);

    PoolFreeMem(executeMemPool, elems, numArgs * sizeof(struct Element *));
}

static void sortlisthelp(int (*cmp)(const void *, const void *))
{
    struct Element *func = PopFunction();
    struct Element *list = PopList();
    size_t length = CountListLength(list), i;
    struct Element **elems = NULL;
    struct GarbageCollectionBase gcb;

    if(length <= 1)
    {
        StackPush(list);
        return;
    }

    StackPush(list);
    StackPush(func);

    elems = ExecuteAllocMem(length * sizeof(struct Element *));
    for(i=0; i < length; i++, list = list->el_Contents.list.tail)
        elems[i] = list->el_Contents.list.head;

    GarbageCollection();
    GetGarbageCollectionBase(&gcb);
    SetGarbageCollectionBaseCurrent();
    qsort(elems, length, sizeof(struct Element *), cmp);
    SetGarbageCollectionBase(&gcb);

    for(i=0; i < length; i++)
        StackPush(elems[i]);
    ProgMakeListSizeT(length);

    PoolFreeMem(executeMemPool, elems, length * sizeof(struct Element *));

    StackSqueeze(1, 2);
}

static int elem_cmp(const void *p1, const void *p2)
{
    struct Element *func = StackTop();
    struct Element *el1 = *(struct Element **)p1;
    struct Element *el2 = *(struct Element **)p2;
    long integer;
    int result = 0;

    if (el1 != el2) {
        StackPush(el2);
        StackPush(el1);
        ProgCallFunctionElement(func, 2, TRUE);

        if (!PopInteger2(&integer)) {
            ExecuteAbort(GCS(MSG_ProgErr_Type));
        }
        result = integer;

        HeapTick();  /* this is needed here or we might run out of heap */
    }

    return result;
}

void ProgSortList(short numArgs)
{
    sortlisthelp(elem_cmp);
}

static int elem_cmpgt(const void *p1, const void *p2)
{
    struct Element *func = StackTop();
    struct Element *el1 = *(struct Element **)p1;
    struct Element *el2 = *(struct Element **)p2;
    bool val;
    int result = 0;

    /*
    ** On m68k-amigaos qsort calls the compare function with identical
    ** elements (unnecessarily), and when not returning 0 in such case
    ** qsort starts accessing elements out-of-bounds of the array.
    ** This is likely an artifact of an old qsort implementation.
    */
    if (el1 != el2) {
        StackPush(el2);
        StackPush(el1);
        ProgCallFunctionElement(func, 2, TRUE);

        val = PopBool2();
        result = (val? 1 : -1);

        HeapTick();  /* this is needed here or we might run out of heap */
    }

    return result;
}

void ProgSortListGt(short numArgs)
{
    sortlisthelp(elem_cmpgt);
}

