/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** funcmath.c
**
** Creation date: 18.12.2009
** Last modified: 07.03.2021
*/

#include "system.h"
#include "funcmath.h"
#include "func.h"
#include "programbin.h"
#include "heap.h"
#include "stack.h"
#include "util.h"
#include "dbtypes.h"
#include "appstrings_cat.h"
#include <time.h>

static void mulreal(double real, short numArgs)
{
    double real2;
    bool nil;

    while (numArgs > 0) {
        nil = !PopReal2(&real2);
        numArgs--;

        if (nil) {
            PopAndPushNil(numArgs);
            return;
        }
        real *= real2;
    }
    PushReal(real);
}

static void mulinteger(long integer, short numArgs)
{
    struct Element *elem;

    while (numArgs > 0) {
        elem = StackPop();
        numArgs--;
        switch(elem->el_Type)
        {
            case TYPE_NIL:      PopAndPushNil(numArgs);
                                return;
                                break;
            case TYPE_INTEGER:  integer *= elem->el_Contents.integer;
                                break;
            case TYPE_REAL:     mulreal(integer * elem->el_Contents.real, numArgs);
                                return;
                                break;
            default:            ExecuteAbort(GCS(MSG_ProgErr_MultiplyTypes));
                                break;
        }
    }
    PushInteger(integer);
}

void ProgMul(short numArgs)
{
    struct Element *elem;

    elem = StackPop();
    numArgs--;
    switch(elem->el_Type)
    {
        case TYPE_NIL:          PopAndPushNil(numArgs);
                                break;
        case TYPE_INTEGER:      mulinteger(elem->el_Contents.integer, numArgs);
                                break;
        case TYPE_REAL:         mulreal(elem->el_Contents.real, numArgs);
                                break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_MultiplyTypes));
                                break;
    }
}

void ProgDivReal(short numArgs)
{
    double real, real2;
    bool nil;

    nil = !PopReal2(&real);
    numArgs--;
    if(nil)
    {
        PopAndPushNil(numArgs);
        return;
    }

    while(numArgs > 0)
    {
        nil = !PopReal2(&real2);
        numArgs--;
        if(nil || real2 == 0.0)
        {
            PopAndPushNil(numArgs);
            return;
        }

        real /= real2;
    }
    PushReal(real);
}

void ProgTrunc(short numArgs)
{
    double real;

    if(!PopReal2(&real))
        PushNil();
    else
        PushReal(floor(real));
}

void ProgRound(short numArgs)
{
    double real, n;
    long i;
    bool nil;

    nil = !PopReal2(&real);
    nil |= !PopInteger2(&i);

    if(nil)
        PushNil();
    else
    {
        n = pow(10.0, (double)i);
        PushReal(round(real * n) / n);
    }
}

void ProgDivInt(short numArgs)
{
    long i1, i2;
    bool nil;

    nil = !PopInteger2(&i1);
    nil |= !PopInteger2(&i2);

    if(nil || i2 == 0)
        PushNil();
    else
        PushInteger(i1 / i2);
}

void ProgMod(short numArgs)
{
    long i1, i2;
    bool nil;

    nil = !PopInteger2(&i1);
    nil |= !PopInteger2(&i2);

    if(nil || i2 == 0)
        PushNil();
    else
        PushInteger(i1 % i2);
}

void ProgOnePlus(short numArgs)
{
    struct Element *elem = StackPop();
    long l;

    switch(elem->el_Type)
    {
        case TYPE_NIL:          PushNil();      break;
        case TYPE_INTEGER:      PushInteger(elem->el_Contents.integer + 1); break;
        case TYPE_REAL:         PushReal(elem->el_Contents.real + 1.0); break;
        case TYPE_DATE:         if (!DateAdd(elem->el_Contents.date, 1, &l)) {
                                    PushNil();
                                } else {
                                    PushDate(l);
                                }
                                break;
        case TYPE_TIME:         if (!TimeAdd(elem->el_Contents.time, 1, &l)) {
                                    PushNil();
                                } else {
                                    PushTime(l);
                                }
                                break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));   break;
    }
}

void ProgOneMinus(short numArgs)
{
    struct Element *elem = StackPop();
    long l;

    switch(elem->el_Type)
    {
        case TYPE_NIL:          PushNil();      break;
        case TYPE_INTEGER:      PushInteger(elem->el_Contents.integer - 1); break;
        case TYPE_REAL:         PushReal(elem->el_Contents.real - 1.0); break;
        case TYPE_DATE:         if (!DateAdd(elem->el_Contents.date, -1, &l)) {
                                    PushNil();
                                } else {
                                    PushDate(l);
                                }
                                break;
        case TYPE_TIME:         if (!TimeAdd(elem->el_Contents.time, -1, &l)) {
                                    PushNil();
                                } else {
                                    PushTime(l);
                                }
                                break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));   break;
    }
}

void ProgAbs(short numArgs)
{
    struct Element *elem = StackPop();

    switch(elem->el_Type)
    {
        case TYPE_NIL:          PushNil();      break;
        case TYPE_INTEGER:      PushInteger(labs(elem->el_Contents.integer)); break;
        case TYPE_REAL:         PushReal(fabs(elem->el_Contents.real)); break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));   break;
    }
}

void ProgRandom(short numArgs)
{
    static bool firstTime = TRUE;
    struct Element *elem = StackPop();

    if(firstTime == TRUE)
    {
        time_t timeval = time(NULL);

        srand((unsigned int)timeval);         /* initializes random function */
        firstTime = FALSE;
    }

    switch(elem->el_Type)
    {
        case TYPE_NIL:
            PushNil();
            break;

        case TYPE_INTEGER:
            PushInteger((long)
		((double)rand() * elem->el_Contents.integer / RAND_MAX));
            break;

        case TYPE_REAL:
            PushReal((double)rand() * elem->el_Contents.real / RAND_MAX);
            break;

        default:
            ExecuteAbort(GCS(MSG_ProgErr_Type));
            break;
    }
}

void ProgLog(short numArgs)
{
    double real;

    if (!PopReal2(&real) || real <= 0.0) {
        PushNil();
    } else {
        PushReal(log(real));
    }
}

void ProgExp(short numArgs)
{
    double real;

    if (!PopReal2(&real)) {
        PushNil();
    } else {
        PushReal(exp(real));
    }
}

void ProgSqrt(short numArgs)
{
    double real;

    if (!PopReal2(&real) || real < 0.0) {
        PushNil();
    } else {
        PushReal(sqrt(real));
    }
}

void ProgPow(short numArgs)
{
    double x, y;
    bool nilx = !PopReal2(&x);
    bool nily = !PopReal2(&y);

    if (nilx || nily || (x < 0.0 && y != floor(y))) {
        PushNil();
    } else {
        PushReal(pow(x, y));
    }
}

