/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** funcprintf.c     string formatting
**
** Creation date: 13.12.1995
** Last modified: 13.02.2010
*/

#include "system.h"
#include "funcprintf.h"
#include "functable.h"
#include "funcio.h"
#include "program.h"
#include "programbin.h"
#include "stack.h"
#include "heap.h"
#include "strhandle.h"
#include "codeset.h"
#include "memory.h"
#include "memo.h"
#include "error.h"
#include "preferences.h"
#include "util.h"
#include "appstrings_cat.h"

/*
** CheckPrintF
*/
const char *CheckPrintF(struct FuncInfo *fi, struct AppArgs *aa)
{
    char *s;
    short type, i;

    if(fi->fi_Id == FUNCID_FPRINTF)
    {
        /* next argument must be file expression */
        if(aa == NULL)
            return(GCS(MSG_ProgErr_MissingArguments));

        type = GenType(aa->Expr->Type, TYPE_FILE);
        if(type == NOTYPE)
            return(GCS(MSG_ProgErr_Type));
        aa = aa->Next;
    }

    /* next argument must be string expression */
    if(aa == NULL)
        return(GCS(MSG_ProgErr_MissingArguments));

    type = GenType(aa->Expr->Type, TYPE_STRING);
    if(type == NOTYPE)
        return(GCS(MSG_ProgErr_Type));
    if(aa->Expr->Case != EXPR_CASE_CONST || aa->Expr->Type != TYPE_STRING)
        return(NULL);                                   /* no further checks possible */

    s = aa->Expr->Contents.Const->Contents.string;
    aa = aa->Next;

    for(; *s; s++)
    {
        if(*s != '%')
            continue;                   /* normal character */
        s++;
        if(*s == '%')
            continue;                   /* %% */

        while(*s=='-' || *s=='+' || *s==' ' || *s=='0')
                s++;                        /* overread control chars */
        for(i = 0; ; i++)
        {
            if(*s == '*')
            {
                /* next argument must be integer or real */
                s++;
                if(aa == NULL)
                    return(GCS(MSG_ProgErr_MissingArguments));
                if(GenType(aa->Expr->Type, TYPE_INTEGER) == NOTYPE)
                    return(GCS(MSG_ProgErr_Type));
                aa = aa->Next;
            }
            else 
	    {
		while (CodesetIsDigit(CodesetGetChar(s))) {
		    s = CodesetNextChar(s);
		}
	    }
            if (i == 0 && *s == '.') {
               s++;
	    } else {
               break;
	    }
        }
        switch (CodesetToUpper(CodesetGetChar(s)))
        {
            case 'B':   type = TYPE_BOOL;   break;
            case 'I':   type = TYPE_INTEGER; break;
            case 'O':   type = TYPE_INTEGER; break;
            case 'X':   type = TYPE_INTEGER; break;
            case 'E':   type = TYPE_REAL;   break;
            case 'F':   type = TYPE_REAL;   break;
            case 'G':   type = TYPE_REAL;   break;
            case 'S':   type = TYPE_STRING; break;
            case 'D':   type = TYPE_DATE;   break;
            case 'T':   type = TYPE_TIME;   break;
            default:    return(GCS(MSG_CompErr_FormatString));
                        break;
         }
         if(aa == NULL)
            return(GCS(MSG_ProgErr_MissingArguments));
         if(GenType(type, aa->Expr->Type) == NOTYPE)
            return(GCS(MSG_ProgErr_Type));
         aa = aa->Next;
    }

    if(aa != NULL)
        return(GCS(MSG_ProgErr_TooManyArguments));

    return(NULL);
}

#define FORMAT_RANGE(x) (short)(((x) < 0)? 0 : ((x > 999)? 999 : x))

/*
** PrintF
*/

void ProgSPrintF(short numArgs)
{
    struct StrHandle *sh;
    const char *s, *t;
    struct Element *elem;
    cschar c;

    s  = PopString2();
    numArgs--;
    if (s == NULL) {
        StackSqueeze(0, numArgs);
        PushNil();
        return;
    }

    sh = AllocStrHandle(executeMemPool);
    if (sh == NULL) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }

    while ((c = CodesetGetChar(s)) != 0) {
        if (c != '%') {                 /* normal character? */
            t = CodesetNextChar(s);
            while (s != t) {
                StrHandlePutC(sh, *s++);
            }
        } else if(*++s == '%') {         /* %% ? */
            StrHandlePutC(sh, '%');
            s++;
        } else {
            bool left = FALSE;
            char space = ' ', sign = 0, *out = "";
	    cschar c;
            short w = -1, g = -1, *p;
            long i, len;
	    const char *format;
	    char buffer[20];

            for (; ; s++) {
                switch (*s) {
                    case '-':   left = TRUE;    continue;
                    case '+':   sign = '+';     continue;
                    case ' ':   sign = ' ';     continue;
                    case '0':   space = '0';    continue;
                }
                break;
            }
            for (i=0, p=&w; ; i++, p=&g) {
		c = CodesetGetChar(s);

                if (c == '*') {
                    if (numArgs <= 0) {
                        ExecuteAbort(GCS(MSG_ProgErr_MissingArguments));
		    }
                    elem = PopInteger();
                    numArgs--;
                    if (elem->el_Type != TYPE_INTEGER) {
                        ExecuteAbort(GCS(MSG_ProgErr_Type));
		    }
                    *p = FORMAT_RANGE(elem->el_Contents.integer);
                    s++;
                    c = CodesetGetChar(s);
                } else if (CodesetIsDigit(c)) {
		    *p = 0;
		    do {
			*p = *p * 10 + CodesetDigitValue(c);
			s = CodesetNextChar(s);
			c = CodesetGetChar(s);
		    } while (CodesetIsDigit(c));
                    *p = FORMAT_RANGE(*p);
                }
                if (i == 0 && c == '.') {
                    s++;
		} else {
                    break;
		}
            }

            if (numArgs <= 0) {
                ExecuteAbort(GCS(MSG_ProgErr_MissingArguments));
            }

            c = CodesetToUpper(CodesetGetChar(s));
            switch (c) {
                case 'B':
                    elem = StackPop();
                    numArgs--;
                    out = (elem->el_Type == TYPE_NIL)? "NIL" : "TRUE";
                    sign = 0;
                    space=' ';
                    break;

                case 'I':
                case 'O':
                case 'X':
                    elem = PopInteger();
                    numArgs--;
                    if (elem->el_Type == TYPE_NIL) {
                        out = "NIL";
                        sign = 0;
                        space = ' ';
                        g = -1;
                    } else {
			switch (*s) {
			    default:
			    case 'i':
			    case 'I': format = "%ld"; break;
			    case 'o':
			    case 'O': format = "%lo"; break;
			    case 'x': format = "%lx"; break;
			    case 'X': format = "%lX"; break;
			}
		       
			sprintf(buffer, format, elem->el_Contents.integer);
			out = buffer;
                        if (out[0] == '-') {
                            sign = '-';
                            out++;
                        }
                        g = -1;
                    }
                    break;

                case 'E':
                case 'F':
                case 'G':
                    elem = PopReal();
                    numArgs--;
                    if (elem->el_Type == TYPE_NIL) {
                        out = "NIL";
                        sign = 0;
                        space = ' ';
                        g = -1;
                    } else {
                        out = DoubleToStr2(elem->el_Contents.real, g, c,
                                (short)GetPrefsValue(
                                    executeProject, PREFS_REALFORMAT));
                        if (out[0] == '-') {
                            sign = '-';
                            out++;
                        }
                        g=-1;
                    }
                    break;

                case 'S':
                    elem = PopString();
                    numArgs--;
                    if (elem->el_Type == TYPE_NIL) {
                        out = "NIL";
                    } else {
                        out = elem->el_Contents.string.string;
                    }
                    sign = 0;
                    space = ' ';
                    break;

                case 'D':
                    elem = PopDate();
                    numArgs--;
                    if (elem->el_Type == TYPE_NIL) {
                        out = "NIL";
                    } else {
                        out = DateToString(elem->el_Contents.date,
                                (short)GetPrefsValue(
                                    executeProject, PREFS_DATEFORMAT));
                    }
                    sign = 0;
                    space = ' ';
                    break;

                case 'T':
                    elem = PopTime();
                    numArgs--;
                    if (elem->el_Type == TYPE_NIL) {
                        out = "NIL";
                    } else {
                        out = TimeToString(elem->el_Contents.date, 
				FORMAT_HOURMINSEC);
                    }
                    sign = 0;
                    space = ' ';
                    break;

                default:
                    ExecuteAbort(GCS(MSG_CompErr_FormatString));
                    break;
            }

            /* calculate length of output string with sign */
            len = CodesetStrLen(out);
            if (sign != 0) {
                len++;
            }
            if (g >= 0 && len > g) {
                len = g;
            }

            /* pad left side with spaces */
            if (!left) {
                i = 0;
                if (sign != 0 && space == '0') {
                    StrHandlePutC(sh, sign);
                    sign = 0;
                    i++;
                    len--;
                }
                for (; i < w - len; i++) {
                    StrHandlePutC(sh, space);
                }
            }

            /* print converted item */
            i = 0;
            if (sign != 0) {
                StrHandlePutC(sh, sign);
                i++;
            }
            if (len > i) {
                t = out + CodesetNCharsToNBytes(out, len - i);
                while (out < t) {
                    StrHandlePutC(sh, *out++);
                }
            }
            /* pad right side with (real) spaces */
            if (left) {
                for (i=0; i < w - len; i++) {
                    StrHandlePutC(sh, ' ');
                }
            }
            s = CodesetNextChar(s);
        }
    }

    if (numArgs > 0) {
        ExecuteAbort(GCS(MSG_ProgErr_TooManyArguments));
    }

    StrHandlePutC(sh, 0);                /* terminate string */
    if (StrHandleError(sh)) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }
    PushString(ExecuteAllocString(StrHandleString(sh)), TRUE);
    FreeStrHandle(sh);
}

static void progfprintf(TextFile *tf, short numArgs)
{
    const char *s;
    long err;

    ProgSPrintF(numArgs);
    s = PopString2();
    if (s) {
        TextFilePutS(tf, s, &err);
        if (err && err != DISKERR_WRITE) {
            ExecuteAbort(GES(err));
        }
        PushInteger(CodesetStrLen(s));
    } else {
        PushNil();
    }
}

void ProgPrintF(short numArgs)
{
    progfprintf(ProgOutputTextFile(), numArgs);
}

void ProgFPrintF(short numArgs)
{
    TextFile *tf = GetElementTextFile(StackPop());

    progfprintf(tf, numArgs-1);
}

TextFile *out = NULL;
static bool textFileOut(const char *s)
{
    long err = OK;

    if (out) {
        TextFilePutS(out, s, &err);
    }
    return (bool)(err == OK);
}

void ProgPrint(short numArgs)
{
    out = ProgOutputTextFile();
    if (out != NULL) {
	ElementToStream(StackTop(), textFileOut);
        textFileOut("\n");
    }
    out = NULL;
}

