/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** funcreq.c - requester interface
**
** Creation date: 17.01.1997
** Last modified: 02.05.2022
*/

#include "system.h"
#include "memory.h"
#include "util.h"
#include "labels.h"
#include "memo.h"
#include "requesters.h"
#include "break.h"
#include "dbtypes.h"
#include "project.h"
#include "program.h"
#include "programbin.h"
#include "heap.h"
#include "stack.h"
#include "func.h"
#include "funcstr.h"
#include "funclist.h"
#include "funcreq.h"
#include "appstrings_cat.h"

static void askdirfile(bool askdir)
{
    char *title = PopString2();
    char *oktext = PopString2();
    char *initial = PopString2();
    bool save = PopBool2();
    char *result = NULL;

    BreakHide();
    if(askdir)
        result = GetDirname(executeProject, executeMemPool, title, oktext, 
                initial, 0, save);
    else
        result = GetFilename(executeProject, executeMemPool, title, oktext, 
                initial, 0, save);
    BreakReInit();

    if(result)
        PushString(result, TRUE);
    else
        PushNil();
}

void ProgAskFile(short numArgs) { askdirfile(FALSE); }
void ProgAskDir(short numArgs)  { askdirfile(TRUE); }

void ProgAskStr(short numArgs)
{
    char *title = PopString2();
    char *oktext = PopString2();
    char *initial = PopString2();
    long maxlen;
    bool nil = !PopInteger2(&maxlen);
    bool secret = FALSE;
    AskStringType type = AST_Normal;
    const char *result = NULL;

    if (nil) {
        maxlen = LONG_MAX;
    }
    if (numArgs > 4) {
	secret = PopBool2();
	if (secret) {
	    type = AST_Secret;
	}
    }

    BreakHide();
    result = AskString(executeProject, title, oktext, initial, maxlen, type);
    BreakReInit();

    if (result) {
        PushString(ExecuteAllocString(result), TRUE);
    } else {
        PushNil();
    }
}

void ProgAskInt(short numArgs)
{
    char *title = PopString2();
    char *oktext = PopString2();
    long initial, min, max;
    bool nil_initial = !PopInteger2(&initial);
    bool nil_min = !PopInteger2(&min);
    bool nil_max = !PopInteger2(&max);
    char *initial_str = nil_initial? "" : IntToStr(initial);
    long result;
    bool rc;

    if(nil_min)
        min = LONG_MIN;
    if(nil_max)
        max = LONG_MAX;

    BreakHide();
    rc = AskInteger(executeProject, title, oktext, initial_str, 
            min, max, &result);
    BreakReInit();

    if(rc)
        PushInteger(result);
    else
        PushNil();
}

static struct Element *labellist(void)
{
    struct Element *elem = StackTop();

    if (elem->el_Type == TYPE_MEMO || elem->el_Type == TYPE_STRING) {
        PushTrue();
        StackReverse(2);
        ProgMemoToList(2);
    }
    return PopList();
}

static struct Element *titlelist(void)
{
    struct Element *elem = StackTop();

    if (elem->el_Type == TYPE_MEMO || elem->el_Type == TYPE_STRING) {
        ProgStrToList(1);
    }
    return PopList();
}

static char **listlabels(void)
{
    struct Element *elem = labellist(), *sub;
    long num = CountListLength(elem), i;
    char **labels;

    if (num == 0) {
        return NULL;
    }

    labels = AllocLabels(executeMemPool, num);
    if (labels == NULL) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }

    for (i = 0; i < num; i++) {
        sub = elem->el_Contents.list.head;
        elem = elem->el_Contents.list.tail;
        StackPush(sub);
        if (sub->el_Type == TYPE_LIST) {
            ProgListToStr(1);
        } else {
            ProgString(1);
        }
        labels[i] = ExecuteAllocString(PopString2());
    }

    return labels;
}

#if 0
static const char *listtitles(void)
{
    struct Element *elem = StackTop();

    if (elem->el_Type == TYPE_LIST) {
	ProgListToStr(1);
    }
    return PopString2();
}
#endif

void ProgAskChoice(short numArgs)
{
    const char *title = PopString2();
    const char *oktext = PopString2();
    struct Element *labels = labellist();
    struct Element *titles = NULL;
    long num = CountListLength(labels);
    long index;
    bool nil = !PopInteger2(&index);
    bool rc;

    if (nil) {
        index = -1;
    } else if (index < 0 || index >= num) {
        ExecuteAbort(GCS(MSG_ProgErr_IndexOutOfRange));
    }
    if (numArgs > 4) {
	titles = titlelist();
    }

    BreakHide();
    rc = AskChoice(executeProject, title, oktext, labels, titles, NULL,
        FALSE, FALSE, &index, NULL);
    BreakReInit();

    if (rc) {
        PushInteger(index);
    } else {
        PushNil();
    }
}

void ProgAskChoiceStr(short numArgs)
{
    const char *title = PopString2();
    const char *oktext = PopString2();
    struct Element *labels = labellist();
    struct Element *titles = NULL;
    char *result = PopString2();
    bool rc;

    if (numArgs > 4) {
	titles = titlelist();
    }

    BreakHide();
    rc = AskChoice(executeProject, title, oktext, labels, titles, NULL,
        FALSE, TRUE, NULL, &result);
    BreakReInit();

    if (rc && result != NULL) {
        PushString(ExecuteAllocString(result), TRUE);
    } else {
        PushNil();
    }
}

void ProgAskOptions(short numArgs)
{
    const char *title = PopString2();
    const char *oktext = PopString2();
    struct Element *labels = labellist();
    struct Element *titles = NULL;
    struct Element *list = PopList();
    long num = CountListLength(labels);
    bool *selected = NULL;
    long index;
    bool rc;

    if (numArgs > 4) {
	titles = titlelist();
    }

    if (num == 0) {
        PushNil();
        return;
    }

    selected = ExecuteAllocMem(num * sizeof(bool));
    memset(selected, 0, num * sizeof(bool));
    for (; list->el_Type == TYPE_LIST; list = list->el_Contents.list.tail) {
        StackPush(list->el_Contents.list.head);
        if (!PopInteger2(&index)) {
            ExecuteAbort(GCS(MSG_ProgErr_Type));
	}
        if (index < 0 || index >= num) {
            ExecuteAbort(GCS(MSG_ProgErr_IndexOutOfRange));
	}
        selected[index] = TRUE;
    }

    BreakHide();
    rc = AskChoice(executeProject, title, oktext, labels, titles, selected,
        TRUE, FALSE, NULL, NULL);
    BreakReInit();

    if (rc) {
        ushort items = 0, i;

        for (i=0; i < num; i++) {
            if (selected[i]) {
                PushInteger(i);
                items++;
            }
	}
        ProgMakeList(items);
    } else {
        PushNil();
    }

    PoolFreeMem(executeMemPool, selected, num * sizeof(bool));
}

void ProgAskButton(short numArgs)
{
    char *title = PopString2();
    char *text = PopString2();
    char **buttons = listlabels();
    const char *canceltext = PopString2();
    ushort num = NumLabels(buttons), i;
    long size = 0;
    char *btext = NULL;
    short result = 0;

    if(canceltext == NULL)
        canceltext = (num > 0)? GCS(MSG_Cancel) : GCS(MSG_Ok);
    for(i=0; i < num; i++)
        size += strlen(buttons[i]) + 1;         /* + 1 for '|' */
    size += strlen(canceltext) + 1;

    btext = ExecuteAllocMem(size);
    btext[0] = 0;
    for(i=0; i < num; i++)
    {
        strcat(btext, buttons[i]);
        strcat(btext, "|");
    }
    strcat(btext, canceltext);

    BreakHide();
    result = Ask2(NULL, title, (num > 0)? AT_QUESTION : AT_WARN, 
	    btext, text? "%s" : NULL, text);
    BreakReInit();

    PoolFreeMem(executeMemPool, btext, size);
    FreeLabels(executeMemPool, buttons);

    if(result > 0)
        PushInteger(result - 1);
    else
        PushNil();
}

static bool *listindices(struct Element *elem, long size)
{
    long i, n;
    bool *selected = NULL;

    selected = ExecuteAllocMem(size * sizeof(bool));
    memset(selected, 0, size * sizeof(bool));

    for(i=0; elem->el_Type == TYPE_LIST; elem = elem->el_Contents.list.tail, i++)
    {
        StackPush(elem->el_Contents.list.head);
        if(!PopInteger2(&n))
            ExecuteAbort(GCS(MSG_ProgErr_Type));
        else if(n < 0 || n >= size)
            ExecuteAbort(GCS(MSG_ProgErr_IndexOutOfRange));
        selected[n] = TRUE;
    }

    return(selected);
}

void ProgAskMulti(short numArgs)
{
    char *title = PopString2();
    char *oktext = PopString2();
    struct Element *items = PopList(), *item, *elem;
    short num = 0, i;
    struct AskMultiInfo *ami = NULL;
    struct AskMultiItem *a;
    char *s, *typestr;
    bool result = FALSE, allocated = FALSE;

    for (item = items; item->el_Type == TYPE_LIST; 
	    item = item->el_Contents.list.tail) {
        num++;
    }

    ami = AllocAskMultiInfo(executeMemPool, num);
    if (ami == NULL) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }

    a = ami->ami_Items;
    item = items;
    for (i=0; i < num; i++, item = item->el_Contents.list.tail, a++) {
        elem = item->el_Contents.list.head;
        if (elem->el_Type != TYPE_LIST) {
            s = SimpleElementToString(elem, NULL, executeMemPool, &allocated);
            if (!allocated) {
                s = ExecuteAllocString(s);
	    }
            a->ami_Title = s;
            a->ami_Type = NOTYPE;
            continue;
        }

        StackPush(elem->el_Contents.list.head);
        s = PopString2();
        if (s) {
            s = ExecuteAllocString(s);
	}
        a->ami_Title = s;

        elem = elem->el_Contents.list.tail;
        if (elem->el_Type != TYPE_LIST) {
            ExecuteAbort(GCS(MSG_ProgErr_MissingArguments));
	}
        StackPush(elem->el_Contents.list.head);
        typestr = PopString2();
        if (typestr == NULL) {
            ExecuteAbort(GCS(MSG_ProgErr_Type));
	}

        elem = elem->el_Contents.list.tail;
        if (elem->el_Type != TYPE_LIST) {
            ExecuteAbort(GCS(MSG_ProgErr_MissingArguments));
	}
        StackPush(elem->el_Contents.list.head);

        if (!CodesetStrCmp(typestr, "bool")) {
            a->ami_Type = TYPE_BOOL;
            a->ami_Nil = !PopBool2();
        } else if (!CodesetStrCmp(typestr, "integer")) {
            a->ami_Type = TYPE_INTEGER;
            a->ami_Nil = !PopInteger2(&a->ami_Contents.integer);
        } else if (!CodesetStrCmp(typestr, "real")) {
            a->ami_Type = TYPE_REAL;
            a->ami_Nil = !PopReal2(&a->ami_Contents.real);
        } else if (!CodesetStrCmp(typestr, "date")) {
            a->ami_Type = TYPE_DATE;
            a->ami_Nil = !PopDate2(&a->ami_Contents.date);
        } else if (!CodesetStrCmp(typestr, "time")) {
            a->ami_Type = TYPE_TIME;
            a->ami_Nil = !PopTime2(&a->ami_Contents.time);
        } else if (!CodesetStrCmp(typestr, "memo")) {
            s = PopMemo2();
            a->ami_Type = TYPE_MEMO;
	    a->ami_Nil = s? FALSE : TRUE;
	    a->ami_Contents.mi = s? ExecuteAllocString(s) : NULL;
        } else if (!CodesetStrCmp(typestr, "string")) {
            s = PopString2();
            a->ami_Type = TYPE_STRING;
            a->ami_Nil = s? FALSE : TRUE;
            a->ami_Contents.string.string = s? ExecuteAllocString(s) : NULL;
	    a->ami_Contents.string.secret = FALSE;
        } else if (!CodesetStrCmp(typestr, "choice")) {
            long active;
            char **labels;

            a->ami_Type = TYPE_CHOICE;
            a->ami_Nil = !PopInteger2(&active);

            elem = elem->el_Contents.list.tail;
            if (elem->el_Type != TYPE_LIST) {
                ExecuteAbort(GCS(MSG_ProgErr_MissingArguments));
	    }
            StackPush(elem->el_Contents.list.head);
            labels = listlabels();

            if (!a->ami_Nil && (active < 0 || active >= NumLabels(labels))) {
                ExecuteAbort(GCS(MSG_ProgErr_IndexOutOfRange));
	    }
            a->ami_Contents.choice.ischoicelist = FALSE;
            a->ami_Contents.choice.active = active;
            a->ami_Contents.choice.labels = labels;
        } else if(!CodesetStrCmp(typestr, "choicelist")) {
            long active;
            char **labels;

            a->ami_Type = TYPE_CHOICE;
            a->ami_Nil = !PopInteger2(&active);

            elem = elem->el_Contents.list.tail;
            if (elem->el_Type != TYPE_LIST) {
                ExecuteAbort(GCS(MSG_ProgErr_MissingArguments));
	    }
            StackPush(elem->el_Contents.list.head);
            labels = listlabels();

            if (!a->ami_Nil && (active < 0 || active >= NumLabels(labels))) {
                ExecuteAbort(GCS(MSG_ProgErr_IndexOutOfRange));
	    }
            a->ami_Contents.choice.ischoicelist = TRUE;
            a->ami_Contents.choice.active = active;
            a->ami_Contents.choice.labels = labels;
        } else if(!CodesetStrCmp(typestr, "options")) {
            bool *selected = NULL;
            struct Element *list = PopList();
            char **labels;

            a->ami_Type = TYPE_OPTIONS;
            a->ami_Nil = FALSE;

            elem = elem->el_Contents.list.tail;
            if (elem->el_Type != TYPE_LIST) {
                ExecuteAbort(GCS(MSG_ProgErr_MissingArguments));
	    }
            StackPush(elem->el_Contents.list.head);
            labels = listlabels();

            selected = listindices(list, NumLabels(labels));

            a->ami_Contents.options.selected = selected;
            a->ami_Contents.options.labels = labels;
        } else {
            ExecuteAbort(GCS(MSG_ProgErr_UnknownType));
	}

        elem = elem->el_Contents.list.tail;
        if (elem->el_Type == TYPE_LIST) {
            char *bh = NULL;

            StackPush(elem->el_Contents.list.head);
            bh = PopString2();
            if (bh != NULL) {
                bh = ExecuteAllocString(bh);
	    }
            a->ami_BubbleHelp = bh;
            elem = elem->el_Contents.list.tail;
        }
        if (elem->el_Type == TYPE_LIST && a->ami_Type == TYPE_STRING) {
            StackPush(elem->el_Contents.list.head);
	    if (PopBool2()) {
		a->ami_Contents.string.secret = TRUE;
	    }
            elem = elem->el_Contents.list.tail;
	}
        if (elem->el_Type == TYPE_LIST) {
            ExecuteAbort(GCS(MSG_ProgErr_TooManyArguments));
	}
    }

    BreakHide();
    result = AskMulti(executeProject, title, oktext, ami);
    BreakReInit();

    if (result) {
        a = ami->ami_Items;
        for (i=0; i < num; i++, a++) {
            if (a->ami_Nil) {
                PushNil();
	    } else {
                switch (a->ami_Type) {
                    case TYPE_STRING:
                        PushString(a->ami_Contents.string.string, TRUE);
                        a->ami_Contents.string.string = NULL;
                        break;

                    case TYPE_MEMO:
                        PushMemo(a->ami_Contents.mi, TRUE);
                        a->ami_Contents.mi = NULL;
                        break;

                    case TYPE_INTEGER:
                        PushInteger(a->ami_Contents.integer);
                        break;

                    case TYPE_REAL:
                        PushReal(a->ami_Contents.real);
                        break;

                    case TYPE_DATE:
                        PushDate(a->ami_Contents.date);
                        break;

                    case TYPE_TIME:
                        PushTime(a->ami_Contents.time);
                        break;

                    case TYPE_BOOL:
                        PushTrue();
                        break;

                    case TYPE_CHOICE:
                        PushInteger(a->ami_Contents.choice.active);
                        break;

                    case TYPE_OPTIONS:
                    {
                        ushort i, num=NumLabels(a->ami_Contents.options.labels);
                        bool *selected = a->ami_Contents.options.selected;
                        size_t numselected = 0;

                        for (i=0; i < num; i++) {
                            if (selected[i]) {
                                PushInteger(i);
                                numselected++;
                            }
			}
                        ProgMakeListSizeT(numselected);
                        break;
                    }

                    default:
                        PushNil();
                        break;
                }
	    }
        }
        ProgMakeList(num);
    } else {
        PushNil();
    }
    FreeAskMultiInfo(ami);
}

