/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** funcsfw.c
**
** Creation date: 21.05.1995
** Last modified: 16.02.2022
*/

#include "system.h"
#include "dbtypes.h"
#include "name.h"
#include "table.h"
#include "field.h"
#include "record.h"
#include "memo.h"
#include "program.h"
#include "programbin.h"
#include "compile_impl.h"
#include "stack.h"
#include "heap.h"
#include "func.h"
#include "funccmp.h"
#include "funclist.h"
#include "funcsfw.h"
#include "memory.h"
#include "util.h"
#include "appstrings_cat.h"

struct sfwSpecialEntry
{
    struct MinNode se_Node;
    struct Element *se_Element; /* pointer to list */
    ushort se_HashVal;          /* Hash value for fast duplicate detection */
    short se_NumPtrs;           /* Number of following pointers */
    void *se_Ptrs[0];           /* pointer to selected records/order fields */
};

struct sfwFromItem
{
    struct Table *sfi_Table;
    short sfi_VarNum;
    bool sfi_Filter;
    struct Field *sfi_ReferenceField;
    struct Record *sfi_ReferenceRecord;
};

struct sfwSpecial
{
    struct List ss_Entries;
    short ss_NumFromItems;
    struct sfwFromItem ss_FromItems[0];
};

static struct sfwSpecialEntry *allocsfwSpecialEntry(short num)
{
    struct sfwSpecialEntry *se = NULL;

    se = PoolAllocMem(executeMemPool, sizeof(struct sfwSpecialEntry) + num * sizeof(void *));
    if(se == NULL)
        ExecuteAbort(GCS(MSG_Err_Mem));

    se->se_NumPtrs = num;
    return(se);
}

static void freesfwSpecialEntry(struct sfwSpecialEntry *se)
{
    if(se)
        PoolFreeMem(executeMemPool, se,
            sizeof(struct sfwSpecialEntry) + se->se_NumPtrs * sizeof(void *));
}

static struct sfwSpecial *allocsfwSpecial(short num)
{
    struct sfwSpecial *ss = NULL;

    ss = PoolAllocMem(executeMemPool,
        sizeof(struct sfwSpecial) + num * sizeof(struct sfwFromItem));
    if(ss == NULL)
        ExecuteAbort(GCS(MSG_Err_Mem));

    NewList(&ss->ss_Entries);
    ss->ss_NumFromItems = num;
    return(ss);
}

static void freesfwSpecial(struct sfwSpecial *ss)
{
    if(ss)
    {
        struct sfwSpecialEntry *se;

        while((se = (struct sfwSpecialEntry *)REMHEAD(&ss->ss_Entries)) != NULL)
            freesfwSpecialEntry(se);

        PoolFreeMem(executeMemPool, ss,
                sizeof(struct sfwSpecial) + 
                ss->ss_NumFromItems * sizeof(struct sfwFromItem));
    }
}

static void sfwSpecialFunc(void *ptr, short method, struct Record *del)
{
    struct sfwSpecial *ss = ptr;
    struct sfwSpecialEntry *se;
    short i;

    switch(method)
    {
        case ELEM_SPECIAL_FREE:
            freesfwSpecial(ss);
            break;

        case ELEM_SPECIAL_REMRECORD:
            if (ss && del) {
                for (i=0; i < ss->ss_NumFromItems; i++) {
                    if (ss->ss_FromItems[i].sfi_ReferenceRecord == del) {
                        ss->ss_FromItems[i].sfi_ReferenceRecord = NULL;
                    }
                }
                FORNODE(se, &ss->ss_Entries) {
                    for (i=0; i < ss->ss_NumFromItems; i++) {
                        if (se->se_Ptrs[i] == del) {
                            se->se_Ptrs[i] = NULL;
                        }
                    }
                }
            }
            break;

        case ELEM_SPECIAL_REMALLRECORDS:
            if (ss) {
                for (i=0; i < ss->ss_NumFromItems; i++) {
                    ss->ss_FromItems[i].sfi_ReferenceRecord = NULL;
                }
                FORNODE(se, &ss->ss_Entries) {
                    for (i=0; i < ss->ss_NumFromItems; i++) {
                        se->se_Ptrs[i] = NULL;
                    }
                }
            }
            break;
    }
}

static void settabvar(struct Table *tab, short n, struct Record *rec)
{
    SetTableProgramRecord(tab, rec);

    if (n >= 0) {
        struct Element *elem = NULL;

        if (rec == NULL) {
            elem = MakeNil();
        } else {
            elem = MakeRecord(tab, rec);
        }
        SetStackNth(n, elem);
    }
}

static bool setrecs(struct sfwSpecial *ss, struct sfwSpecialEntry *se)
{
    short i;

    for(i = ss->ss_NumFromItems-1; i >= 0; i--)
    {
        if(se->se_Ptrs[i] == NULL)
            return(FALSE);

        settabvar(ss->ss_FromItems[i].sfi_Table,
            ss->ss_FromItems[i].sfi_VarNum, se->se_Ptrs[i]);
    }
    return(TRUE);
}

static short sfwNumRels;
static short sfwNumOrderItems;
static struct SFWOrderItem *sfwOrderItems;

static int sfwCmpFunc(struct sfwSpecialEntry *se1, struct sfwSpecialEntry *se2)
{
    long i, result = 0;
    void **p1 = &se1->se_Ptrs[sfwNumRels], **p2 = &se2->se_Ptrs[sfwNumRels];

    for(i=0; i < sfwNumOrderItems; i++, p1++, p2++)
    {
        result = ProgElementCmp(*p1, *p2, TRUE, TRUE);
        if(result != 0)
	{
	    if(sfwOrderItems[i].sfwoi_Mode == ORDER_DESC)
		result = -result;
            break;
	}
    }

    return(result);
}

static ushort sfwHash(struct Element *elem)
{
    ushort val = 0;

    switch(elem->el_Type)
    {
        case TYPE_INTEGER:      val = HashInteger(elem->el_Contents.integer);   break;
        case TYPE_REAL:         val = HashReal(elem->el_Contents.real);         break;
        case TYPE_BOOL:         val = 1;        break;
        case TYPE_STRING:       val = HashString(elem->el_Contents.string.string);     break;
        case TYPE_MEMO:         val = HashString(elem->el_Contents.mi);           break;
        case TYPE_DATE:         val = HashInteger(elem->el_Contents.date);      break;
        case TYPE_TIME:         val = HashInteger(elem->el_Contents.time);      break;
        case TYPE_RECORD:       val = HashInteger((intptr_t)elem->el_Contents.tabrec.rec); break;
        case TYPE_LIST:
        {
            struct Element *el;

            for(el = elem; el->el_Type == TYPE_LIST; el = el->el_Contents.list.tail)
            {
                val = (val ^ sfwHash(el->el_Contents.list.head)) * 69 + 32767;
                val = (val << 1) | ((val & 0x8000)? 1 : 0);
            }
            break;
        }
    }
    return(val);
}

void SelectFromWhere(bool distinct,
        short numSelectItems, struct SFWSelectItem *sfwsis,
        short numFromItems, struct SFWFromItem *sfwfis,
        struct SubProgram *where,
        short numOrderItems, struct SFWOrderItem *sfwois,
        bool isselect)
{
    struct Record **recs = NULL;
    struct sfwSpecial *ss = NULL;
    struct sfwFromItem *sfi;
    int i, j;
    long numOrdElements = 0;
    bool ready = FALSE, docons;
    struct sfwSpecialEntry *se, *succ;
    struct Element *elem;

    if(numFromItems == 0)                            /* be safe */
       ExecuteAbort(GCS(MSG_Err_Internal));

    recs = PoolAllocMem(executeMemPool, numFromItems * sizeof(struct Record *));
    if(recs == NULL)
        ExecuteAbort(GCS(MSG_Err_Mem));

    ss = allocsfwSpecial(numFromItems);
    if(ss == NULL)
        ExecuteAbort(GCS(MSG_Err_Mem));

    PushSpecial(ss, sfwSpecialFunc);
    PushNil();

    sfi = ss->ss_FromItems;
    for(i=0; i < numFromItems; i++, sfi++)
    {
        sfi->sfi_Table = sfwfis[i].sfwfi_Table;
        sfi->sfi_VarNum = sfwfis[i].sfwfi_VarNum;
        sfi->sfi_Filter = sfwfis[i].sfwfi_Filter;
        sfi->sfi_ReferenceField = sfwfis[i].sfwfi_ReferenceField;
        sfi->sfi_ReferenceRecord = NULL;
        if (sfi->sfi_ReferenceField != NULL) {
            ExecuteEvalSubProgram(&sfwfis[i].sfwfi_ReferenceExpr);
            sfi->sfi_ReferenceRecord = PopRecord2MatchTable(
                    FieldReferenceTable(sfi->sfi_ReferenceField));
        }
    }

    for(i = numFromItems-1; i >= 0; i--)
    {
        sfi = &ss->ss_FromItems[i];
        recs[i] = FirstRecordWithReference(sfi->sfi_Table, sfi->sfi_Filter,
                sfi->sfi_ReferenceField, sfi->sfi_ReferenceRecord);
        settabvar(sfi->sfi_Table, sfi->sfi_VarNum, recs[i]);
        if(recs[i] == NULL)
            ready = TRUE;              /* there aren't any records to process */
    }

    while(!ready)
    {
        bool generateLine = TRUE;

        /* this is a save place to call HeapTick */
        HeapTick();

        /* check WHERE expression (if any) */
        if(where)
        {
            ExecuteEvalSubProgram(where);
            elem = StackPop();
            if(elem->el_Type == TYPE_NIL)
                generateLine = FALSE;
        }

        if(generateLine)
        {
            void **ptr;

            /*
            ** Allocate new sfwSpecialEntry and add tail it to SelectList
            */
            se = allocsfwSpecialEntry(numFromItems + numOrderItems);
            for(i = 0, ptr = se->se_Ptrs; i < numFromItems; i++)
                *ptr++ = recs[i];
            for(i = 0; i < numOrderItems; i++)
            {
                ExecuteEvalSubProgram(&sfwois[i].sfwoi_Order);
                *ptr++ = StackTop();
                ProgCons(2);
                numOrdElements++;
            }

            ADDTAIL(&ss->ss_Entries, se);
        }

        /* goto next record */
        for(i = numFromItems-1; i >= 0; i--)
        {
            sfi = &ss->ss_FromItems[i];
            recs[i] = NextRecordWithReference(recs[i], sfi->sfi_Filter,
                    sfi->sfi_ReferenceField, sfi->sfi_ReferenceRecord);
            if(recs[i] != NULL)
            {
                for(j = numFromItems-1; j >= i; j--)
                    settabvar(sfi->sfi_Table, sfi->sfi_VarNum, recs[j]);
                break;
            }
            recs[i] = FirstRecordWithReference(sfi->sfi_Table, sfi->sfi_Filter,
                    sfi->sfi_ReferenceField, sfi->sfi_ReferenceRecord);
            if(i == 0)
                ready = TRUE;
        }
    }

    PoolFreeMem(executeMemPool, recs, numFromItems * sizeof(struct Record *));
    recs = NULL;

    if(numOrdElements > 1)
    {
        sfwNumRels = numFromItems;
        sfwNumOrderItems = numOrderItems;
        sfwOrderItems = sfwois;

        QLSort(&ss->ss_Entries, (int (*)(void *, void *))sfwCmpFunc);
    }

    StackPop();                 /* pops list of order items */

    PushNil();                  /* keep stack size */

    if(isselect == FALSE)
        return;                 /* we are ready for FOR ALL loop */

    /*
    ** produce title line, if we don't have any exprs then we have a
    ** SELECT * FROM ... query
    */
    if(numSelectItems == 0)
    {
        struct Table *relative = sfwfis[0].sfwfi_Table, *tab;
        struct RelFieldPath rap;
        char *pathname;
        short entries = 0;
        struct List *fields;
        struct Field *field;

        for(i = 0; i < numFromItems; i++)
        {
            tab = sfwfis[i].sfwfi_Table;
            fields = TableFields(tab);
            if(fields)
                FORNODE(field, fields)
                {
                    switch(FieldType(field))
                    {
                        case TYPE_BOOL:
                        case TYPE_INTEGER:
                        case TYPE_CHOICE:
                        case TYPE_REAL:
                        case TYPE_STRING:
                        case TYPE_MEMO:
                        case TYPE_DATE:
                        case TYPE_TIME:
                        case TYPE_REFERENCE:
                            rap.rap_NumFields = 1;
                            rap.rap_Fields[0] = field;
                            break;

                        default:
                            continue;
                            break;
                    }

                    rap.rap_Relative = relative;
                    rap.rap_Table = tab;
                    rap.rap_Star = FALSE;
                    pathname = AllocString(executeMemPool, 
			    RelFieldPathToPathName(&rap));
                    if(pathname == NULL)
                        ExecuteAbort(GCS(MSG_Err_Mem));
                    elem = PushString(pathname, TRUE);
                    elem->el_Contents.string.field = 
			rap.rap_Fields[rap.rap_NumFields - 1];
                    entries++;
                }
        }
        ProgMakeList(entries);
    }
    else /* numSelectItems == 0 */
    {
        for(i=0; i < numSelectItems; i++)
        {
            struct RelFieldPath *rap = &sfwsis[i].sfwsi_RelFieldPath;
            struct Field *field = NULL;
            const char *pathname = NULL;
            const char *localname = sfwsis[i].sfwsi_LocalName;
            const char *ctitle;
            char *title = NULL;
            char help[13];

            ctitle = sfwsis[i].sfwsi_Title;
            if(ctitle != NULL)
                title = AllocString(executeMemPool, ctitle);
            else
            {
                if(rap->rap_Table)
                    pathname = RelFieldPathToPathName(rap);
                if (pathname == NULL)
		{
                    sprintf(help, "$%d", i+1);
                    pathname = help;
		}

                if(localname)
                {
                    title = PoolAllocMem(executeMemPool,
                        strlen(localname) + 1 + strlen(pathname) + 1);
                    if(title)
                        sprintf(title, "%s.%s", localname, pathname);
                }
                else
                    title = AllocString(executeMemPool, pathname);
            }

            if(title == NULL)
                ExecuteAbort(GCS(MSG_Err_Mem));

            if(rap->rap_Table && rap->rap_NumFields > 0)
                field = rap->rap_Fields[rap->rap_NumFields - 1];

            elem = PushString(title, TRUE);
            elem->el_Contents.string.field = field;
        }
        ProgMakeList(numSelectItems);
    }
    ProgCons(2);

    SAFEFORNODE(se, succ, &ss->ss_Entries)
    {
        /* this is a save place to call HeapTick */
        HeapTick();

        if(!setrecs(ss, se))            /* any recs[i] == NULL ? */
        {
            REMOVE(se);
            freesfwSpecialEntry(se);
            continue;
        }

        /*
        ** produce one line, if we don't have any exprs then we have a
        ** SELECT * FROM ... query
        */
        if(numSelectItems == 0)
        {
            struct RelFieldPath rap;
            short entries = 0;
            struct List *fields;
            struct Field *field;
            bool show;

            for(i = 0; i < numFromItems; i++)
            {
                fields = TableFields(sfwfis[i].sfwfi_Table);
                if(fields)
                    FORNODE(field, fields)
                    {
                        switch(FieldType(field))
                        {
                            case TYPE_BOOL:
                            case TYPE_INTEGER:
                            case TYPE_CHOICE:
                            case TYPE_REAL:
                            case TYPE_STRING:
                            case TYPE_MEMO:
                            case TYPE_DATE:
                            case TYPE_TIME:
                            case TYPE_REFERENCE:
                                rap.rap_Relative = NULL;
                                rap.rap_NumFields = 1;
                                rap.rap_Table = sfwfis[i].sfwfi_Table;
                                rap.rap_Fields[0] = field;
                                show = TRUE;
                                break;

                            default:
                                show = FALSE;
                                break;
                        }

                        if(show)
                        {
                            elem = GetGlobalData(&rap,
                                TableProgramRecord(rap.rap_Table));
                            StackPush(elem);
                            entries++;
                        }
                    }
            }
            ProgMakeList(entries);
        }
        else /* numSelectItems == 0 */
        {
            int i;

            /* evaluate all expressions and make a list */
            for(i=0; i < numSelectItems; i++)
                ExecuteEvalSubProgram(&sfwsis[i].sfwsi_Expr);
            ProgMakeList(numSelectItems);
        }

        docons = TRUE;
        if(distinct)
        {
            struct sfwSpecialEntry *s;

            se->se_Element = StackTop();
            se->se_HashVal = sfwHash(se->se_Element);
            FORNODE(s, &ss->ss_Entries)
            {
                if(s == se)
                    break;                              /* ready */
                if(s->se_HashVal == se->se_HashVal)
                    if(ProgElementCmp(s->se_Element, se->se_Element, FALSE, FALSE) == 0)     /* duplicate? */
                    {
                        docons = FALSE;
                        StackPop();
                        REMOVE(se);
                        freesfwSpecialEntry(se);
                        break;
                    }
            }
        }
        else
        {
            REMOVE(se);
            freesfwSpecialEntry(se);
        }
        if(docons)
            ProgCons(2);
    }
    ProgReverse(1);

    /* free sfwSpecial list */
    while((se = (struct sfwSpecialEntry *)REMHEAD(&ss->ss_Entries)) != NULL)
        freesfwSpecialEntry(se);
}

static bool setnextrecs(struct sfwSpecial *ss)
{
    struct sfwSpecialEntry *se;
    bool rc;

    for(;;)
    {
        se = (struct sfwSpecialEntry *)REMHEAD(&ss->ss_Entries);
        if(se == NULL)
            return(FALSE);          /* end of list */

        rc = setrecs(ss, se);
        freesfwSpecialEntry(se);

        if(rc)
            return(TRUE);
    }
}

bool sfwSetNextRecords(void)
{
    struct Element *elem = StackNth(1);

    if(elem->el_Type != TYPE_SPECIAL)
        ExecuteAbort(GCS(MSG_Err_Internal));
    return(setnextrecs(elem->el_Contents.special.ptr));
}

