/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** funcstr.c
**
** Creation date: 17.12.1996
** Last modified: 19.01.2014
*/

#include "system.h"
#include "dbtypes.h"
#include "field.h"
#include "compile.h"
#include "program.h"
#include "programbin.h"
#include "func.h"
#include "funcstr.h"
#include "funclist.h"
#include "functable.h"
#include "heap.h"
#include "stack.h"
#include "memory.h"
#include "memo.h"
#include "formatmemo.h"
#include "strhandle.h"
#include "codeset.h"
#include "util.h"
#include "appstrings_cat.h"

void ProgLen(short numArgs)
{
    char *str = PopString2();

    if(str == NULL)
        PushNil();
    else
        PushInteger(CodesetStrLen(str));
}

void ProgMaxLen(short numArgs)
{
    struct RelFieldPath *rap = PopField2(NULL);
    struct Field *field = RelFieldPathField(rap);

    if(field == NULL)
        PushNil();
    else
        PushInteger(FieldStringMaxlen(field));
}

void ProgRightStr(short numArgs)
{
    char *str = PopString2();
    long num, len, left;
    bool nil = !PopInteger2(&num);

    if(str == NULL || nil || num < 0)
        PushNil();
    else
    {
        len = CodesetStrLen(str);
        num = MIN(num, len);
	left = CodesetNCharsToNBytes(str, len - num);

        PushString(ExecuteAllocString(str + left), TRUE);
    }
}

void ProgLeftStr(short numArgs)
{
    char *str = PopString2();
    long num, len;
    bool nil = !PopInteger2(&num);

    if(str == NULL || nil || num < 0)
        PushNil();
    else
    {
	num = CodesetNCharsToNBytes(str, num);
        len = strlen(str);
        num = MIN(num, len);

        PushString(ExecuteAllocStringN(str, num), TRUE);
    }
}

void ProgMidStr(short numArgs)
{
    char *str = PopString2();
    long pos, num, len = str? CodesetStrLen(str) : -1;
    bool nilpos = !PopInteger2(&pos);
    bool nilnum = !PopInteger2(&num);

    if(nilnum)
        num = LONG_MAX;

    if(len < 0 || nilpos || pos < 0 ||  pos > len || num < 0)
        PushNil();
    else
    {
	if (num > len - pos)
	    num = len - pos;

	pos = CodesetNCharsToNBytes(str, pos);
	num = CodesetNCharsToNBytes(str + pos, num);

	PushString(ExecuteAllocStringN(str + pos, num), TRUE);
    }
}

void ProgSetMidStr(short numArgs)
{
    char *str = PopString2(), *new;
    long pos, len = str? CodesetStrLen(str) : -1, setlen, end, num;
    bool nilpos = !PopInteger2(&pos);
    char *set = PopString2();

    if(len < 0 || nilpos || pos < 0 ||  pos > len || set == NULL)
        PushNil();
    else
    {
        setlen = CodesetStrLen(set);
	pos = CodesetNCharsToNBytes(str, pos);
	end = CodesetNCharsToNBytes(str + pos, setlen) + pos;
	len = strlen(str);
	end = MIN(end, len);
	num = pos + strlen(set) + strlen(str + end);
	new = ExecuteAllocMem(num + 1);
	sprintf(new, "%.*s%s%s", (int)pos, str, set, str + end);
        PushString(new, TRUE);
    }
}

void ProgInsMidStr(short numArgs)
{
    char *str = PopString2(), *new;
    long pos, len = str? CodesetStrLen(str) : -1, num;
    bool nilpos = !PopInteger2(&pos);
    char *ins = PopString2();

    if(len < 0 || nilpos || pos < 0 ||  pos > len || ins == NULL)
        PushNil();
    else
    {
	pos = CodesetNCharsToNBytes(str, pos);
        num = strlen(str) + strlen(ins);
        new = ExecuteAllocMem(num + 1);
        sprintf(new, "%.*s%s%s", (int)pos, str, ins, str + pos);
        PushString(new, TRUE);
    }
}

void ProgSHA1Sum(short numArgs)
{
    const char *str = PopString2();

    if (str == NULL) {
	PushNil();
    } else {
	PushString(ExecuteAllocString(SHA1Sum(str)), TRUE);
    }
}

void ProgUpper(short numArgs)
{
    char *str = PopString2();

    if(str == NULL)
        PushNil();
    else
    {
	str = CodesetStrUp(str, executeMemPool);
        if (str == NULL)
            ExecuteAbort(GCS(MSG_Err_Mem));
        PushString(str, TRUE);
    }
}

void ProgLower(short numArgs)
{
    char *str = PopString2();

    if(str == NULL)
        PushNil();
    else
    {
	str = CodesetStrDown(str, executeMemPool);
        if (str == NULL)
            ExecuteAbort(GCS(MSG_Err_Mem));
        PushString(str, TRUE);
    }
}

void ProgAsc(short numArgs)
{
    char *str = PopString2();

    if(str == NULL)
        PushNil();
    else
        PushInteger((long)CodesetGetChar(str));
}

void ProgChr(short numArgs)
{
    long val;
    bool nil = !PopInteger2(&val);
    char s[CodesetMaxBytesForChar];
    cschar c;
    short n = -1;

    if (!nil && (c = (cschar)val) == val) {
	n = CodesetCharToStr(c, s);
    }

    if (n < 0) {
        PushNil();
    } else {
        PushString(ExecuteAllocStringN(s, n), TRUE);
    }
}

void ProgWord(short numArgs)
{
    char *str = PopString2();
    long pos;
    bool nil = !PopInteger2(&pos);
    bool err = FALSE;

    if (str == NULL || nil) {
        PushNil();
    } else {
        str = WordString(executeMemPool, str, pos, &err);
        if (err) {
            ExecuteAbort(GCS(MSG_Err_Mem));
        }
        if (str == NULL) {
            PushNil();
        } else {
            PushString(str, TRUE);
        }
    }
}

void ProgWords(short numArgs)
{
    char *str = PopString2();

    if (str == NULL) {
        PushNil();
    } else {
        PushInteger(NumWords(str));
    }
}

void ProgField(short numArgs)
{
    char *str = PopString2();
    long pos;
    bool nil = !PopInteger2(&pos);
    const char *sep = NULL;
    bool quotes = FALSE;
    bool err = FALSE;

    if (numArgs > 2) {
        sep = PopString2();
        if (numArgs > 3) {
            quotes = PopBool2();
        }
    }

    if (str == NULL || nil) {
        PushNil();
    } else {
        str = FieldString(executeMemPool, str, pos, sep, quotes, &err);
        if (err) {
            ExecuteAbort(GCS(MSG_Err_Mem));
        }
        if (str == NULL) {
            PushNil();
        } else {
            PushString(str, TRUE);
        }
    }
}

void ProgFields(short numArgs)
{
    char *str = PopString2();
    const char *sep = NULL;
    bool quotes = FALSE;

    if (numArgs > 1) {
        sep = PopString2();
        if (numArgs > 2) {
            quotes = PopBool2();
        }
    }
    if (str == NULL) {
        PushNil();
    } else {
        PushInteger(StrNumFields(str, sep, quotes));
    }
}

void progconcat_help(char *fill, short numArgs)
{
    char *s = NULL, *t, *h;

    while(numArgs > 0)
    {
        t = PopString2();
        numArgs--;

        if(t == NULL)
        {
            FreeString(executeMemPool, s);
            PopAndPushNil(numArgs);
            return;
        }
        if(s == NULL)
            s = ExecuteAllocString(t);
        else
        {
            h = ExecuteAllocMem(strlen(s) + strlen(fill) + strlen(t) + 1);
            sprintf(h, "%s%s%s", s, fill, t);
            s = h;
        }
    }
    if(s == NULL)
        PushNil();
    else
        PushString(s, TRUE);
}

void ProgConcat(short numArgs)
{
    progconcat_help(" ", numArgs);
}

void ProgConcat2(short numArgs)
{
    char *fill = PopString2();

    numArgs--;
    if(fill == NULL)
        PopAndPushNil(numArgs);
    else
        progconcat_help(fill, numArgs);
}

void ProgCopyStr(short numArgs)
{
  char *str = PopString2();
  long num;
  bool nil = !PopInteger2(&num);

  if(str == NULL || nil || num < 0)
    PushNil();
  else
    {
      long len = strlen(str), i;
      char *s = ExecuteAllocMem(num * len + 1), *t;
      
      for(i=0, t=s; i < num; i++, t += len)
	strcpy(t, str);
      *t = 0;
      PushString(s, TRUE);
    }
}

void ProgLine(short numArgs)
{
    char *mi = PopMemo2();
    long pos;
    bool nil = !PopInteger2(&pos);
    char *s;
    bool err = FALSE;

    if(mi == NULL || nil)
        PushNil();
    else
    {
        s = MemoLineString(executeMemPool, mi, pos, &err);
        if(err)
            ExecuteAbort(GCS(MSG_Err_Mem));
        if(s == NULL)
            PushNil();
        else
            PushString(s, TRUE);
    }
}

void ProgLines(short numArgs)
{
    char *mi = PopMemo2();

    if(mi == NULL)
        PushNil();
    else
        PushInteger(MemoNumLines(mi));
}

void ProgTrimStr(short numArgs)
{
    const char *s = PopString2(), *front, *back;
    char *copy, *t;
    long size;

    if (numArgs == 1)
	front = back = NULL;
    else if (numArgs == 3)
    {
	front = PopString2();
	back = PopString2();

	if (front == NULL || back == NULL)
	{
	    PushNil();
	    return;
	}
    }
    else
    {
	ExecuteAbort(GCS(MSG_ProgErr_FunctionArity));
	return;			/* to avoid unitialized warning */ 
    }

    if (s == NULL)
    {
        PushNil();
	return;
    }

    copy = ExecuteAllocString(s);
    size = strlen(copy) + 1;
    CodesetTrimStrChars(copy, front, back);
    t = ExecuteAllocString(copy);
    PoolFreeMem(executeMemPool, copy, size);

    PushString(t, TRUE);
}

void ProgRemChars(short numArgs)
{
    const char *str = PopString2();
    const char *rem = PopString2();

    if (str == NULL || rem == NULL) {
        PushNil();
    } else {
	const char *s, *next;
        char *new, *t;
        size_t size = 0;

	for (s = str; *s != 0; s = next) {
	    next = CodesetNextChar(s);
	    if (CodesetStrchr(rem, CodesetGetChar(s)) == NULL) {
		size += next - s;
	    }
	}

        new = t = ExecuteAllocMem(size + 1);
	for (s = str; *s != 0; s = next) {
	    next = CodesetNextChar(s);
	    if (CodesetStrchr(rem, CodesetGetChar(s)) == NULL) {
		while (s != next) {
		    *t++ = *s++;
		}
	    }
	}
        *t = 0;

        PushString(new, TRUE);
    }
}

typedef const char * (*strstrfunc_t)(const char *, const char *);

static void progReplaceStrHelp(strstrfunc_t func, short numArgs)
{
    char *str = PopString2();
    
    numArgs--;
    if (str == NULL)
    {
	PopAndPushNil(numArgs);
	return;
    }
    str = ExecuteAllocString(str);
    while (numArgs >= 2)
    {
	const char *search = PopString2();
	const char *replace = PopString2();
	const char *find;

	numArgs -= 2;
	if (search == NULL || search[0] == 0 || replace == NULL)
	{
	    PopAndPushNil(numArgs);
	    FreeString(executeMemPool, str);
	    return;
	}

	find = func(str, search);
	if (find != NULL)
	{
	    long slen = strlen(search), rlen = strlen(replace); 
	    long dlen = rlen - slen;
	    long size = strlen(str) + 1 + dlen;
	    const char *s;
	    char *new, *t;

	    for (s = find+slen; (find = func(s, search)) != NULL; s = find+slen)
		size += dlen;

	    new = t = ExecuteAllocMem(size);
	    for(s = str; (find = func(s, search)) != NULL; s = find + slen)
	    {
		memcpy(t, s, find - s);
	    	t += find - s;
		memcpy(t, replace, rlen);
		t += rlen;
	    }
	    strcpy(t, s);

	    FreeString(executeMemPool, str);
	    str = new;
	}
    }
    if (numArgs != 0)
	ExecuteAbort(GCS(MSG_ProgErr_FunctionArity));

    PushString(str, TRUE);
}

void ProgReplaceStr(short numArgs)        
{ progReplaceStrHelp((strstrfunc_t)strstr, numArgs); }
void ProgReplaceStrStar(short numArgs)    
{ progReplaceStrHelp((strstrfunc_t)CodesetStrStr, numArgs); }

static void progIndexHelp(strstrfunc_t func)
{
    const char *s = PopString2();
    const char *t = PopString2();

    if(s == NULL || t == NULL)
        PushNil();
    else
    {
        t = func(s, t);
        if(t == NULL)
            PushNil();
        else
            PushInteger((long)CodesetNBytesToNChars(s, t - s));
    }
}

void ProgIndexStr(short numArgs) { progIndexHelp((strstrfunc_t)strstr); }
void ProgIndexStrStar(short num) { progIndexHelp((strstrfunc_t)CodesetStrStr); }
void ProgIndexBrk(short numArgs) { progIndexHelp((strstrfunc_t)CodesetStrpbrk);}
void ProgIndexBrkStar(short num) { progIndexHelp((strstrfunc_t)CodesetStrPBrk);}

static void progRIndexHelp(strstrfunc_t func)
{
    const char *s = PopString2();
    const char *t = PopString2();

    if(s == NULL || t == NULL)
	PushNil();
    else
    {
       	const char *last = NULL, *p;
	
	for(p = s; (p = func(p, t)) != NULL; p = CodesetNextChar(p))
	    last = p;
	
	if(last == NULL)
	    PushNil();
	else
	    PushInteger((long)CodesetNBytesToNChars(s, last - s));
    }
}

void ProgRIndexStr(short numArgs) { progRIndexHelp((strstrfunc_t)strstr); }
void ProgRIndexStrStar(short num) {progRIndexHelp((strstrfunc_t)CodesetStrStr);}
void ProgRIndexBrk(short numArgs){progRIndexHelp((strstrfunc_t)CodesetStrpbrk);}
void ProgRIndexBrkStar(short num){progRIndexHelp((strstrfunc_t)CodesetStrPBrk);}

void ProgLike(short numArgs)
{
    const char *str = PopString2();
    const char *joker = PopString2();
    bool result;

    if (str == NULL || joker == NULL) {
        result = FALSE;
    } else {
        result = StrJokerCmp(str, joker);
    }

    if (result) {
        PushTrue();
    } else {
        PushNil();
    }
}

void ProgStrToList(short numArgs)
{
    const char *str = PopString2();
    const char *needle = "\t";

    if (numArgs > 1) {
	needle = PopString2();
    }

    if (str == NULL || needle == NULL) {
	PushNil();
    } else {
	size_t len = strlen(needle);
	size_t num = 0;
	const char *t;
	char *s;

	do {
	    if (len > 0) {
		t = strstr(str, needle);
	    } else if (*str) {
		t = CodesetNextChar(str);
	    } else if (num == 0) {
		t = NULL;	/* keep empty string */
	    } else {
		break;		/* don't place empty string at end */
	    }
	    if (t) {
		s = ExecuteAllocStringN(str, t - str);
		str = t + len;
	    } else {
		s = ExecuteAllocString(str);
		str = NULL;
	    }
	    PushString(s, TRUE);
	    num++;
	 } while (str);

        ProgMakeListSizeT(num);
    }
}

void ProgMemoToList(short numArgs)
{
    struct Element *elem = StackPop();
    bool expandStr = FALSE;

    if (numArgs > 1) {
	expandStr = PopBool2();
    }
    PushString("\n", FALSE);
    StackPush(elem);
    ProgStrToList(2);

    if (expandStr) {
	size_t num = 0;

	elem = PopList();
	while (elem->el_Type == TYPE_LIST) {
	    StackPush(elem->el_Contents.list.head);
	    ProgStrToList(1);
	    elem = elem->el_Contents.list.tail;
	    num++;
	}
        ProgMakeListSizeT(num);
    }
}

static void listToStr(struct Element *list, const char *sep, bool allowRec)
{
    if (list->el_Type != TYPE_LIST || sep == NULL) {
	PushNil();
    } else {
        struct StrHandle *sh = AllocStrHandle(executeMemPool);
	struct Element *head;
        bool allocated;
        char *s;

        while (list->el_Type == TYPE_LIST) {
	    head = list->el_Contents.list.head;
	    if (allowRec && head->el_Type == TYPE_LIST) {
		listToStr(head, "\t", FALSE);
		s = PopString2();
		allocated = FALSE;
	    } else {
		s = SimpleElementToString(head,NULL,executeMemPool, &allocated);
		if (s == NULL) {
		    ExecuteAbort(GCS(MSG_Err_Mem));
		}
	    }
            StrHandlePutS(sh, s);
            if (allocated) {
                FreeString(executeMemPool, s);
	    }

            list = list->el_Contents.list.tail;
            if (list->el_Type == TYPE_LIST) {
                StrHandlePutS(sh, sep);
	    }
        }
        StrHandlePutC(sh, 0);           /* terminate string */

        if (StrHandleError(sh)) {
            ExecuteAbort(GCS(MSG_Err_Mem));
	}

        s = StrHandleString(sh);
        PushString(ExecuteAllocString(s), TRUE);
        FreeStrHandle(sh);
    }
}

void ProgListToStr(short numArgs)
{
    struct Element *list = PopList();
    const char *sep = "\t";

    if (numArgs > 1) {
	sep = PopString2();
    }

    listToStr(list, sep, FALSE);
}

void ProgListToMemo(short numArgs)
{
    listToStr(PopList(), "\n", TRUE);
}
    
void ProgFillMemo(short numArgs)
{
    struct Element *elem = PopString();

    if (elem->el_Type != TYPE_STRING)
        PushNil();
    else
    {
        struct StrHandle *sh;
	const char *s = elem->el_Contents.string.string;

        sh = AllocStrHandle(executeMemPool);
        if(sh == NULL)
            ExecuteAbort(GCS(MSG_Err_Mem));

	StackPush(elem);	/* make sure data is not freed by GC */

        for(; *s; s++)
        {
            if(s[0] == '$' && s[1] == '(')
            {
                const char *start = &s[1], *t = &s[2], *end; 
		char c;
                const char *error = NULL;
                long brackets = 1;
                struct Program *prg = NULL;
                bool allocated = FALSE;
                struct Element *elem = NULL;

                for(; brackets > 0; t++)
                {
                    c = *t;
                    switch(c)
                    {
                        case '"':                       /* string? */
                            t = CodesetStrchr(t+1, '"');
                            break;

                        case '#':                       /* comment? */
                            t = CodesetStrchr(t+1, '\n');
                            break;

                        case '(':
                            brackets++;
                            break;

                        case ')':
                            brackets--;
                            break;

                        default:        break;
                    }

                    if(t == NULL || *t == 0)
                    {
                        t = s + strlen(s);
                        error = GCS((c == '"')?
                            MSG_CompErr_UnterminatedString : MSG_CompErr_SyntaxError);
                        break;
                    }
                }

                end = t;
                if(error == NULL)
                    error = CompileExpr(executeProject, NULL,
                        start, (long)(end - start),
                        &prg, NULL, NULL);
                if(error == NULL)
                    error = ExecuteEvalBin(ProgramBin(prg), &elem);
                if(error == NULL)
                {
                    char *s = SimpleElementToString(elem, NULL, executeMemPool, 
					      &allocated);
                    if(s == NULL)
                        error = GCS(MSG_Err_Mem);
                    else
                    {
                        StrHandlePutS(sh, s);
                        if(allocated)
                            FreeString(executeMemPool, s);
                    }
                }
                if(error != NULL)
                    StrHandlePutS(sh, error);
                FreeProgram(prg);

                s = end - 1;
            }
            else
                StrHandlePutC(sh, *s);
        }
        StrHandlePutC(sh, 0);

        if(StrHandleError(sh))
            ExecuteAbort(GCS(MSG_Err_Mem));
	StackPop();			/* remove reference made of GC */
        PushString(ExecuteAllocString(StrHandleString(sh)), TRUE);
        FreeStrHandle(sh);
    }
}

void ProgFormatMemo(short numArgs)
{
    const char *s = PopString2();
    long width;
    bool nil = !PopInteger2(&width);
    bool fill = FALSE;
    bool singleline = FALSE;
    char *t;

    if (numArgs >= 3) {
	fill = PopBool2();
    }
    if (numArgs >= 4) {
	singleline = PopBool2();
    }

    if (s == NULL || nil || width < 1) {
        PushNil();
        return;
    }

    t = FormatMemo(s, width, fill, singleline, executeMemPool);
    if (t == NULL) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }
    PushString(t, TRUE);
}

void ProgIndentMemo(short numArgs)
{
    char *mi = PopMemo2(), *new;
    long indent;
    bool nil = !PopInteger2(&indent);

    if(mi == NULL || nil)
    {
        PushNil();
        return;
    }
    new = IndentMemo(mi, indent, executeMemPool);
    if(new == NULL)
        ExecuteAbort(GCS(MSG_Err_Mem));

    PushMemo(new, TRUE);
}

