/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** functable.c
**
** Creation date: 14.05.1995
** Last modified: 03.06.2022
*/

#include "system.h"
#include "main.h"
#include "util.h"
#include "dbtypes.h"
#include "table.h"
#include "field.h"
#include "compile_impl.h"
#include "functable.h"
#include "func.h"
#include "funccmp.h"
#include "funcmath.h"
#include "funcstr.h"
#include "funcdate.h"
#include "funclist.h"
#include "funcdb.h"
#include "funcreq.h"
#include "funcio.h"
#include "funcprintf.h"
#include "preferences.h"
#include "dosutil.h"
#include "appstrings_cat.h"
#include "version.h"

#ifndef M_PI
#define M_PI       3.14159265358979323846
#endif

#define F(type) (1<<(type))
#define FDATA (F(TYPE_INTEGER) | F(TYPE_REAL) | F(TYPE_BOOL) | F(TYPE_STRING) | \
        F(TYPE_MEMO) | F(TYPE_DATE) | F(TYPE_TIME))

short GenType(short type1, short type2)
{
    if(type1 == NOTYPE || type2 == NOTYPE)
        return(NOTYPE);
    else if(type1 == TYPE_NIL)
        return(type2);
    else if(type2 == TYPE_NIL)
        return(type1);
    if(type1 == type2)
        return(type1);
    if(type1 == TYPE_UNKNOWN)
        return(type2);
    if(type2 == TYPE_UNKNOWN)
        return(type1);
    if(type1 == TYPE_INTEGER && type2 == TYPE_REAL)
        return(TYPE_REAL);
    if(type1 == TYPE_REAL && type2 == TYPE_INTEGER)
        return(TYPE_REAL);
    if(type1 == TYPE_STRING && type2 == TYPE_MEMO)
        return(TYPE_STRING);
    if(type1 == TYPE_MEMO && type2 == TYPE_STRING)
        return(TYPE_STRING);

    return(NOTYPE);
}

static struct RelFieldPath *ExprRelFieldPath(struct Expr *expr)
{
    struct RelFieldPath *rap = NULL;

    if (expr != NULL) {
        switch (expr->Case) {
            case EXPR_CASE_GLOBALIDENT:
                rap = &expr->Contents.GlobalIdent->RelFieldPath;
                break;

            case EXPR_CASE_LOCALEXTEND:
                rap = &expr->Contents.LocalExtend->GlobalIdent->RelFieldPath;
                break;
        }
    }
    return rap;
}

struct Field *ExprField(struct Expr *expr)
{
    struct RelFieldPath *rap = ExprRelFieldPath(expr);
    struct Field *field = RelFieldPathField(rap);

    return(field);
}

static struct Table *ExprTable(struct Expr *expr)
{
    struct RelFieldPath *rap = ExprRelFieldPath(expr);
    struct Table *tab = (rap != NULL)? rap->rap_Table : NULL;
    struct Field *field = RelFieldPathField(rap);

    if(field != NULL)
    {
        if(FieldType(field) == TYPE_REFERENCE)
            tab = FieldReferenceTable(field);
        else
            tab = FieldTable(field);
    }
    return(tab);
}

static const char *checkcmp(struct FuncInfo *fi, struct AppArgs *aa)
{
    short type = TYPE_NIL;
    struct Table *tab = NULL, *tab2;

    while (aa != NULL) {
        type = GenType(type, aa->Expr->Type);
        if (type == NOTYPE) {
            return GCS(MSG_CompErr_CmpTypes);
        }
        if (aa->Expr->Type == TYPE_RECORD) {
            tab2 = ExprTable(aa->Expr);
            if (tab2 != NULL) {
                if (tab == NULL) {
                    tab = tab2;
                } else if (tab != tab2) {
                    return GCS(MSG_CompErr_CmpTypes);
                }
            }
        }
        aa = aa->Next;
    }
    return NULL;
}

static const char *checkaddsub(struct FuncInfo *fi, struct AppArgs *aa,
        ulong ftypes1, ulong ftypes2)
{
    short type1 = aa->Expr->Type, type2, type;

    fi->fi_Type = (char)type1;
    if (!(F(type1) & ftypes1)) {
        return GCS(MSG_CompErr_AddTypes);
    }

    aa = aa->Next;
    if (aa == NULL) {
        return NULL;
    }

    type2 = aa->Expr->Type;
    for(; aa; aa = aa->Next) {
        type2 = GenType(type2, aa->Expr->Type);
    }
    if (!(F(type2) & ftypes2)) {
        return GCS(MSG_CompErr_AddTypes);
    }

    fi->fi_Type = (char)type1;
    if ((F(type1) & (F(TYPE_DATE) | F(TYPE_TIME) | F(TYPE_UNKNOWN))) &&
            F(type2) & (F(TYPE_INTEGER) | F(TYPE_REAL))) {
        return NULL;
    }

    type = GenType(type1, type2);
    fi->fi_Type = (char)type;  
    if (!(F(type) & ftypes1)) {
        return GCS(MSG_CompErr_AddTypes);
    }

    return NULL;
}

static const char *checkadd(struct FuncInfo *fi, struct AppArgs *aa)
{
    static ulong ftypes1 = (FDATA & ~F(TYPE_BOOL)) | F(TYPE_NIL) |
        F(TYPE_UNKNOWN);
    static ulong ftypes2 = (FDATA & ~F(TYPE_BOOL) & ~F(TYPE_DATE)) | F(TYPE_NIL)
        | F(TYPE_UNKNOWN);

    return checkaddsub(fi, aa, ftypes1, ftypes2);
}

static const char *checksub(struct FuncInfo *fi, struct AppArgs *aa)
{
    static ulong ftypes1 = F(TYPE_INTEGER) | F(TYPE_REAL) | F(TYPE_DATE) |
        F(TYPE_TIME) | F(TYPE_NIL) | F(TYPE_UNKNOWN);
    static ulong ftypes2 = F(TYPE_INTEGER) | F(TYPE_REAL) | F(TYPE_TIME) |
        F(TYPE_NIL) | F(TYPE_UNKNOWN);

    return checkaddsub(fi, aa, ftypes1, ftypes2);
}

static const char *checkintreal(struct FuncInfo *fi, struct AppArgs *aa)
{
    short type = aa->Expr->Type;
    static ulong ftypes = F(TYPE_INTEGER) | F(TYPE_REAL) | F(TYPE_NIL) |
        F(TYPE_UNKNOWN);

    for (; aa; aa = aa->Next) {
        type = GenType(type, aa->Expr->Type);
    }

    if (!(F(type) & ftypes)) {
        return GCS(MSG_ProgErr_Type);
    }

    return NULL;
}

static const char *checkintrealdatetime(struct FuncInfo *fi, struct AppArgs *aa)
{
    short type = aa->Expr->Type;
    static ulong ftypes = F(TYPE_INTEGER) | F(TYPE_REAL) | F(TYPE_DATE) |
        F(TYPE_TIME) | F(TYPE_NIL) | F(TYPE_UNKNOWN);

    for (; aa; aa = aa->Next) {
        type = GenType(type, aa->Expr->Type);
    }

    if (!(F(type) & ftypes)) {
        return GCS(MSG_ProgErr_Type);
    }

    return NULL;
}

static const char *checkappend(struct FuncInfo *fi, struct AppArgs *aa)
{
    short type = TYPE_LIST;

    for(; aa; aa = aa->Next)
        type = GenType(type, aa->Expr->Type);

    if(type == NOTYPE)
        return(GCS(MSG_ProgErr_ListExpected));

    return(NULL);
}

static const char *checkstrs(struct FuncInfo *fi, struct AppArgs *aa)
{
    short type = TYPE_STRING;

    for(; aa; aa = aa->Next)
        type = GenType(type, aa->Expr->Type);

    if(type == NOTYPE)
        return(GCS(MSG_ProgErr_Type));

    return(NULL);
}

static const char *checkprogn(struct FuncInfo *fi, struct AppArgs *aa)
{
    fi->fi_Type = TYPE_NIL;

    for(; aa; aa = aa->Next)
        fi->fi_Type = (char)aa->Expr->Type;
    return(NULL);
}

static const char *checkprog1(struct FuncInfo *fi, struct AppArgs *aa)
{
    fi->fi_Type = TYPE_NIL;

    if(aa != NULL)
        fi->fi_Type = (char)aa->Expr->Type;
    return(NULL);
}

static const char *checkif(struct FuncInfo *fi, struct AppArgs *aa)
{
    short num = CountAppArgs(aa);

    if (num == 2) {
        fi->fi_Type = (char)aa->Next->Expr->Type;
    } else if (num == 3 && aa->Next->Expr->Type == aa->Next->Next->Expr->Type) {
        fi->fi_Type = (char)aa->Next->Expr->Type;
    } else {
        fi->fi_Type = TYPE_UNKNOWN;
    }
    return NULL;
}

static const char *checksetqhelp(struct FuncInfo *fi, struct Expr *expr,
        short type)
{
    switch (expr->Case)
    {
        case EXPR_CASE_GLOBALIDENT:
        case EXPR_CASE_LOCALEXTEND:
            fi->fi_HasSideEffect = TRUE;
            break;

        case EXPR_CASE_LOCALIDENT:
            if (expr->Contents.LocalIdent->Arity >= 0) {
                return GCS(MSG_CompErr_LValue);
            }
            if (expr->Contents.LocalIdent->Table != NULL) {
                fi->fi_HasSideEffect = TRUE;
            }
            break;

        default: 
            return GCS(MSG_CompErr_LValue);
    }

    if (expr->Type != TYPE_BOOL) {
        if (GenType(expr->Type, type) == NOTYPE) {
            return GCS(MSG_ProgErr_Type);
        }
    }

    return NULL;
}

static const char *checksetq(struct FuncInfo *fi, struct AppArgs *aa)
{
    short num = CountAppArgs(aa);
    const char *error = NULL;

    if (num % 2 != 0) {
        return GCS(MSG_CompErr_Arity);
    }

    fi->fi_Type = TYPE_NIL;
    for (; aa; aa = aa->Next->Next) {
        error = checksetqhelp(fi, aa->Expr, aa->Next->Expr->Type);
        if (error) {
            break;
        }
        fi->fi_Type = (char)aa->Next->Expr->Type;
    }
    return error;
}

static const char *checksetqlist(struct FuncInfo *fi, struct AppArgs *aa)
{
    const char *error = NULL;

    for (; aa->Next; aa = aa->Next) {
        error = checksetqhelp(fi, aa->Expr, TYPE_UNKNOWN);
        if (error) {
            return error;
        }
    }
    if (!error && GenType(aa->Expr->Type, TYPE_LIST) == NOTYPE) {
        error = GCS(MSG_ProgErr_ListExpected);
    }

    return error;
}

static const char *checkstr(struct FuncInfo *fi, struct AppArgs *aa)
{
    struct Field *field;
    short type;

    if(aa == NULL)
        return(GCS(MSG_CompErr_Arity));
    if(aa->Expr->Type == NOTYPE)
        return(GCS(MSG_ProgErr_Type));

    field = ExprField(aa->Expr);
    if(field == NULL)
        return(NULL);

    type = FieldType(field);
    if(type != TYPE_VIRTUAL && type != TYPE_BUTTON)
        aa->Expr->Type = TYPE_FIELD;

    return(NULL);
}

static const char *checkstrfield(struct FuncInfo *fi, struct AppArgs *aa)
{
    struct Field *field;
    short type;

    if(aa == NULL)
        return(GCS(MSG_CompErr_Arity));

    field = ExprField(aa->Expr);
    type = FieldType(field);
    if(field == NULL || type != TYPE_STRING)
        return(GCS(MSG_CompErr_StringFieldExpected));

    aa->Expr->Type = TYPE_FIELD;
    return(NULL);
}

static const char *checkstrchoicefield(struct FuncInfo *fi, struct AppArgs *aa)
{
    struct Field *field;
    short type;

    if(aa == NULL)
        return(GCS(MSG_CompErr_Arity));

    field = ExprField(aa->Expr);
    type = FieldType(field);
    if(field == NULL || (type != TYPE_STRING && type != TYPE_CHOICE))
        return(GCS(MSG_CompErr_StringOrChoiceFieldExpected));

    aa->Expr->Type = TYPE_FIELD;
    return(NULL);
}

static const char *checktabfield(struct FuncInfo *fi, struct AppArgs *aa)
{
    struct RelFieldPath *rap;

    if(aa == NULL)
        return(GCS(MSG_CompErr_Arity));

    if(aa->Expr->Case != EXPR_CASE_GLOBALIDENT)
        return GCS(MSG_CompErr_FieldTableExpected);

    rap = &aa->Expr->Contents.GlobalIdent->RelFieldPath;
    if(rap->rap_NumFields > 0)
        aa->Expr->Type = TYPE_FIELD;
    else
        aa->Expr->Type = TYPE_TABLE;

    return(NULL);
}

static const char *checkvirtfield(struct FuncInfo *fi, struct AppArgs *aa)
{
    struct Field *field;
    short type;

    if(aa == NULL)
        return(GCS(MSG_CompErr_Arity));

    field = ExprField(aa->Expr);
    type = FieldType(field);
    if(field == NULL || type != TYPE_VIRTUAL)
        return(GCS(MSG_CompErr_VirtualFieldExpected));

    aa->Expr->Type = TYPE_FIELD;
    return(NULL);
}

static const char *checkfuncall(struct FuncInfo *fi, struct AppArgs *aa)
{
    if(aa && aa->Expr && GenType(aa->Expr->Type, TYPE_FUNCTION) == NOTYPE)
        return(GCS(MSG_ProgErr_FunctionExpected));
    return(NULL);
}

static const char *checkapply(struct FuncInfo *fi, struct AppArgs *aa)
{
    const char *error = checkfuncall(fi, aa);
    if (error)
	return error;

    aa = aa->Next;
    if (aa == NULL)
	return GCS(MSG_ProgErr_MissingArguments);

    while (aa->Next)
	aa = aa->Next;
    if (aa->Expr && GenType(aa->Expr->Type, TYPE_LIST) == NOTYPE)
	return GCS(MSG_ProgErr_ListExpected);

    return NULL;
}

static const char *checktrimstr(struct FuncInfo *fi, struct AppArgs *aa)
{
    short num = CountAppArgs(aa);

    if (num != 1 && num != 3)
	return GCS(MSG_CompErr_Arity);

    return checkstrs(fi, aa);
}

static const char *checkreplacestr(struct FuncInfo *fi, struct AppArgs *aa)
{
    short num = CountAppArgs(aa);

    if (num % 2 != 1)
	return GCS(MSG_CompErr_Arity);

    return checkstrs(fi, aa);
}

static const char *checkobsolete(struct FuncInfo *fi, struct AppArgs *aa)
{
    if (GetPrefsValue(compProject, PREFS_PROGRAMOBSOLETEFUNCTIONS) ==
            OBSOLETE_ERROR) {
        return GCS(MSG_CompErr_ObsoleteFunction);
    }
    return NULL;
}

static const char anyrec_argtypes[] = { TYPE_UNKNOWN, TYPE_RECORD };
static const char bool_argtypes[] = { TYPE_BOOL };
static const char int_argtypes[] = { TYPE_INTEGER };
static const char intanylist_argtypes[] = { TYPE_INTEGER, TYPE_UNKNOWN, TYPE_LIST };
static const char intint_argtypes[] = { TYPE_INTEGER, TYPE_INTEGER };
static const char intintint_argtypes[] = { TYPE_INTEGER, TYPE_INTEGER, TYPE_INTEGER };
static const char intintlist_argtypes[] = { TYPE_INTEGER, TYPE_INTEGER, TYPE_LIST };
static const char intlist_argtypes[] = { TYPE_INTEGER, TYPE_LIST };
static const char intfile_argtypes[] = { TYPE_INTEGER, TYPE_FILE };
static const char real_argtypes[] = { TYPE_REAL, };
static const char realint_argtypes[] = { TYPE_REAL, TYPE_INTEGER };
static const char realreal_argtypes[] = { TYPE_REAL, TYPE_REAL };
static const char str_argtypes[] = { TYPE_STRING };
static const char strstr_argtypes[] = { TYPE_STRING, TYPE_STRING };
static const char strstrbool_argtypes[] = { TYPE_STRING, TYPE_STRING, TYPE_BOOL };
static const char strstrstrbool_argtypes[] = { TYPE_STRING, TYPE_STRING, TYPE_STRING, TYPE_BOOL };
static const char strstrstrintbool_argtypes[] = { TYPE_STRING, TYPE_STRING, TYPE_STRING, TYPE_INTEGER, TYPE_BOOL};
static const char strstrintintint_argtypes[] = { TYPE_STRING, TYPE_STRING, TYPE_INTEGER, TYPE_INTEGER, TYPE_INTEGER};
static const char strstrstr_argtypes[] = { TYPE_STRING, TYPE_STRING, TYPE_STRING };
static const char strstrlist_argtypes[] = { TYPE_STRING, TYPE_STRING, TYPE_LIST, };
static const char strstranyintany_argtypes[] = { TYPE_STRING, TYPE_STRING, TYPE_UNKNOWN, TYPE_INTEGER, TYPE_UNKNOWN };
static const char strstranystrany_argtypes[] = { TYPE_STRING, TYPE_STRING, TYPE_UNKNOWN, TYPE_STRING, TYPE_UNKNOWN };
static const char strstranylistany_argtypes[] = { TYPE_STRING, TYPE_STRING, TYPE_UNKNOWN, TYPE_LIST, TYPE_UNKNOWN };
static const char strstrliststr_argtypes[] = { TYPE_STRING, TYPE_STRING, TYPE_LIST, TYPE_STRING, };
static const char strint_argtypes[] = { TYPE_STRING, TYPE_INTEGER };
static const char strintboolbool_argtypes[] = { TYPE_STRING, TYPE_INTEGER, TYPE_BOOL, TYPE_BOOL };
static const char strintint_argtypes[] = { TYPE_STRING, TYPE_INTEGER, TYPE_INTEGER };
static const char strintstr_argtypes[] = { TYPE_STRING, TYPE_INTEGER, TYPE_STRING };
static const char strintstrbool_argtypes[] = { TYPE_STRING, TYPE_INTEGER, TYPE_STRING, TYPE_BOOL };
static const char strfile_argtypes[] = { TYPE_STRING, TYPE_FILE };
static const char memo_argtypes[] = { TYPE_MEMO };
static const char memobool_argtypes[] = { TYPE_MEMO, TYPE_BOOL };
static const char memoint_argtypes[] = { TYPE_MEMO, TYPE_INTEGER };
static const char memofile_argtypes[] = { TYPE_MEMO, TYPE_FILE };
static const char date_argtypes[] = { TYPE_DATE };
static const char dateint_argtypes[] = { TYPE_DATE, TYPE_INTEGER };
static const char file_argtypes[] = { TYPE_FILE, };
static const char fileintint_argtypes[] = { TYPE_FILE, TYPE_INTEGER, TYPE_INTEGER, };
static const char list_argtypes[] = { TYPE_LIST, };
static const char liststr_argtypes[] = { TYPE_LIST, TYPE_STRING };
static const char rec_argtypes[] = { TYPE_RECORD, };
static const char recbool_argtypes[] = { TYPE_RECORD, TYPE_BOOL, };
static const char recint_argtypes[] = { TYPE_RECORD, TYPE_INTEGER, };
static const char recrec_argtypes[] = { TYPE_RECORD, TYPE_RECORD };
static const char field_argtypes[] = { TYPE_FIELD };
static const char fieldbool_argtypes[] = { TYPE_FIELD, TYPE_BOOL, };
static const char fieldstr_argtypes[] = { TYPE_FIELD, TYPE_STRING, };
static const char fieldint_argtypes[] = { TYPE_FIELD, TYPE_INTEGER, };
static const char tab_argtypes[] = { TYPE_TABLE, };
static const char tabbool_argtypes[] = { TYPE_TABLE, TYPE_BOOL };
static const char tabint_argtypes[] = { TYPE_TABLE, TYPE_INTEGER };
static const char tabstr_argtypes[] = { TYPE_TABLE, TYPE_STRING };
static const char tabrec_argtypes[] = { TYPE_TABLE, TYPE_RECORD };
static const char funclist_argtypes[] = { TYPE_FUNCTION, TYPE_LIST, };

static struct FuncInfo FuncInfos[NumFuncIds] =
{
    { "=",      FUNCID_EQ,      2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgEq, },
    { "=*",     FUNCID_EQSTAR,  2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgEqStar, },
    { "<>",     FUNCID_NE,      2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgNe },
    { "<>*",    FUNCID_NESTAR,  2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgNeStar, },
    { "<",      FUNCID_LT,      2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgLt, },
    { "<*",     FUNCID_LTSTAR,  2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgLtStar, },
    { "<=",     FUNCID_LE,      2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgLe },
    { "<=*",    FUNCID_LESTAR,  2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgLeStar, },
    { ">",      FUNCID_GT,      2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgGt },
    { ">*",     FUNCID_GTSTAR,  2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgGtStar, },
    { ">=",     FUNCID_GE,      2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgGe },
    { ">=*",    FUNCID_GESTAR,  2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, checkcmp, ProgGeStar, },
    { "CMP",    FUNCID_CMP,     2, 2,   FALSE,  FIK_EXPR,   TYPE_INTEGER, NULL, checkcmp, ProgCmp, },
    { "CMP*",   FUNCID_CMPSTAR, 2, 2,   FALSE,  FIK_EXPR,   TYPE_INTEGER, NULL, checkcmp, ProgCmpStar, },
    { "AND",    FUNCID_AND,     0, FIA_ANY, FALSE, FIK_CONTROL, TYPE_BOOL, NULL, NULL, NULL, }, /* non-strict */
    { "OR",     FUNCID_OR,      0, FIA_ANY, FALSE, FIK_CONTROL, TYPE_BOOL, NULL, NULL, NULL, }, /* non-strict */
    { "NOT",    FUNCID_NOT,     1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,  NULL, NULL, ProgNot, },
    { "+",      FUNCID_ADD,     1, FIA_ANY, FALSE, FIK_EXPR, TYPE_UNKNOWN, NULL, checkadd, ProgAdd, },
    { "-",      FUNCID_SUB,     1, FIA_ANY, FALSE, FIK_EXPR, TYPE_UNKNOWN, NULL, checksub, ProgSub, },
    { "1+",     FUNCID_ONEPLUS, 1, 1,   FALSE,  FIK_EXPR,   TYPE_UNKNOWN,  NULL, checkintrealdatetime, ProgOnePlus, },
    { "1-",     FUNCID_ONEMINUS, 1, 1,  FALSE,  FIK_EXPR,   TYPE_UNKNOWN,  NULL, checkintrealdatetime, ProgOneMinus, },
    { "*",      FUNCID_MUL,     1, FIA_ANY, FALSE, FIK_EXPR, TYPE_REAL, NULL, checkintreal, ProgMul, },
    { "/",      FUNCID_DIVREAL, 1, FIA_ANY, FALSE, FIK_EXPR, TYPE_REAL, NULL, checkintreal, ProgDivReal, },
    { "DIV",    FUNCID_DIVINT,  2, 2,   FALSE,  FIK_EXPR,   TYPE_INTEGER, intint_argtypes, NULL, ProgDivInt, },
    { "MOD",    FUNCID_MOD,     2, 2,   FALSE,  FIK_EXPR,   TYPE_INTEGER, intint_argtypes, NULL, ProgMod, },
    { "MIN",    FUNCID_MIN,     0, FIA_ANY, FALSE, FIK_EXPR, TYPE_UNKNOWN, NULL, checkcmp, ProgMin,},
    { "MIN*",   FUNCID_MINSTAR, 0, FIA_ANY, FALSE, FIK_EXPR, TYPE_UNKNOWN, NULL, checkcmp, ProgMinStar,},
    { "MAX",    FUNCID_MAX,     0, FIA_ANY, FALSE, FIK_EXPR, TYPE_UNKNOWN, NULL, checkcmp, ProgMax,},
    { "MAX*",   FUNCID_MAXSTAR, 0, FIA_ANY, FALSE, FIK_EXPR, TYPE_UNKNOWN, NULL, checkcmp, ProgMaxStar,},
    { "ABS",    FUNCID_ABS,     1, 1,   FALSE,  FIK_EXPR,   TYPE_REAL,  NULL, checkintreal, ProgAbs,},
    { "RANDOM", FUNCID_RANDOM,  1, 1,   FALSE,  FIK_EXPR,   TYPE_REAL,  real_argtypes, NULL, ProgRandom,},
    { "LOG",    FUNCID_LOG,     1, 1,   FALSE,  FIK_EXPR,   TYPE_REAL,  real_argtypes, NULL, ProgLog,},
    { "EXP",    FUNCID_EXP,     1, 1,   FALSE,  FIK_EXPR,   TYPE_REAL,  real_argtypes, NULL, ProgExp,},
    { "SQRT",   FUNCID_SQRT,    1, 1,   FALSE,  FIK_EXPR,   TYPE_REAL,  real_argtypes, NULL, ProgSqrt,},
    { "POW",    FUNCID_POW,     2, 2,   FALSE,  FIK_EXPR,   TYPE_REAL,  realreal_argtypes, NULL, ProgPow,},
    { "LIKE",   FUNCID_LIKE,    2, 2,   FALSE,  FIK_EXPR,   TYPE_BOOL,  strstr_argtypes, NULL, ProgLike, },
    { "DAY",  	FUNCID_DAY,     1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER, date_argtypes, NULL, ProgDay, },
    { "MONTH", 	FUNCID_MONTH,   1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER, date_argtypes, NULL, ProgMonth, },
    { "YEAR", 	FUNCID_YEAR,    1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER, date_argtypes, NULL, ProgYear, },
    { "DATEDMY",  FUNCID_DATEDMY, 3, 3, FALSE,  FIK_EXPR,   TYPE_DATE,  intintint_argtypes, NULL, ProgDateDMY, },
    { "MONTHDAYS",  FUNCID_MONTHDAYS, 2, 2, FALSE, FIK_EXPR, TYPE_INTEGER, intint_argtypes, NULL, ProgMonthDays, },
    { "YEARDAYS",  FUNCID_YEARDAYS, 1, 1, FALSE, FIK_EXPR,  TYPE_INTEGER, int_argtypes, NULL, ProgYearDays, },
    { "ADDMONTH",  FUNCID_ADDMONTH, 2, 2, FALSE, FIK_EXPR,  TYPE_DATE,  dateint_argtypes, NULL, ProgAddMonth, },
    { "ADDYEAR",  FUNCID_ADDYEAR, 2, 2,	FALSE, 	FIK_EXPR,   TYPE_DATE,  dateint_argtypes, NULL, ProgAddYear, },
    { "TODAY",  FUNCID_TODAY,   0, 0,   FALSE,  FIK_EXPR,   TYPE_DATE,  NULL, NULL, ProgToday, },
    { "NOW",    FUNCID_NOW,     0, 0,   FALSE,  FIK_EXPR,   TYPE_TIME,  NULL, NULL, ProgNow, },
    { "CONS",   FUNCID_CONS,    2, 2,   FALSE,  FIK_EXPR,   TYPE_LIST,  NULL,       NULL, ProgCons, },
    { "LIST",   FUNCID_LIST,    0, FIA_ANY, FALSE, FIK_EXPR,   TYPE_LIST,  NULL,       NULL, ProgList, },
    { "FIRST",  FUNCID_FIRST,   1, 1,   FALSE,  FIK_EXPR,   TYPE_UNKNOWN, list_argtypes, NULL, ProgFirst, },
    { "REST",   FUNCID_REST,    1, 1,   FALSE,  FIK_EXPR,   TYPE_UNKNOWN, list_argtypes, NULL, ProgRest, },
    { "LAST",   FUNCID_LAST,    1, 1,   FALSE,  FIK_EXPR,   TYPE_UNKNOWN, list_argtypes, NULL, ProgLast, },
    { "NTH",    FUNCID_NTH,     2, 2,   FALSE,  FIK_EXPR,   TYPE_UNKNOWN, intlist_argtypes, NULL, ProgNth, },
    { "REPLACENTH", FUNCID_REPLACENTH, 3, 3, FALSE, FIK_EXPR, TYPE_LIST, intanylist_argtypes, NULL, ProgReplaceNth, },
    { "REPLACENTH*", FUNCID_REPLACENTHSTAR, 3, 3, FALSE, FIK_EXPR, TYPE_LIST, intanylist_argtypes, NULL, ProgReplaceNthStar, },
    { "MOVENTH", FUNCID_MOVENTH, 3, 3,  FALSE,  FIK_EXPR,   TYPE_LIST, intintlist_argtypes, NULL, ProgMoveNth, },
    { "MOVENTH*", FUNCID_MOVENTHSTAR, 3, 3, FALSE, FIK_EXPR, TYPE_LIST, intintlist_argtypes, NULL, ProgMoveNthStar, },
    { "REMOVENTH", FUNCID_REMOVENTH, 2, 2,  FALSE,  FIK_EXPR,   TYPE_LIST, intlist_argtypes, NULL, ProgRemoveNth, },
    { "REMOVENTH*", FUNCID_REMOVENTHSTAR, 2, 2, FALSE, FIK_EXPR, TYPE_LIST, intlist_argtypes, NULL, ProgRemoveNthStar, },
    { "APPEND", FUNCID_APPEND,  0, FIA_ANY, FALSE, FIK_EXPR,   TYPE_LIST,  NULL,       checkappend, ProgAppend, },
    { "REVERSE", FUNCID_REVERSE, 1, 1,  FALSE,  FIK_EXPR,   TYPE_LIST,  list_argtypes, NULL, ProgReverse,},
    { "PRINT",  FUNCID_PRINT,   1, 1,   FALSE,   FIK_EXPR,   TYPE_UNKNOWN, NULL,     NULL, ProgPrint, },
    { "PRINTF", FUNCID_PRINTF,  0, FIA_ANY, FALSE, FIK_EXPR,   TYPE_INTEGER, NULL,     CheckPrintF, ProgPrintF, },
    { "FPRINTF", FUNCID_FPRINTF, 0, FIA_ANY, TRUE, FIK_EXPR,   TYPE_INTEGER, NULL,     CheckPrintF, ProgFPrintF, },
    { "SPRINTF", FUNCID_SPRINTF, 0, FIA_ANY, FALSE, FIK_EXPR,  TYPE_STRING,  NULL,     CheckPrintF, ProgSPrintF, },
    { "FFLUSH", FUNCID_FFLUSH, 	1, 1, TRUE, 	FIK_EXPR,   TYPE_INTEGER, file_argtypes, NULL, ProgFFlush, },
    { "FOPEN",  FUNCID_FOPEN,   2, 3,   TRUE,   FIK_EXPR,   TYPE_FILE,  strstrstr_argtypes, NULL, ProgFOpen, },
    { "FCLOSE", FUNCID_FCLOSE,  1, 1,   TRUE,   FIK_EXPR,   TYPE_INTEGER, file_argtypes, NULL, ProgFClose, },
    { "",       FUNCID_PROGN,   0, FIA_ANY, FALSE, FIK_EXPR,   TYPE_UNKNOWN, NULL,     checkprogn, ProgProgN, },
    { "PROG1",  FUNCID_PROG1,   0, FIA_ANY, FALSE, FIK_EXPR,   TYPE_UNKNOWN, NULL,     checkprog1, ProgProg1, },
    { "LENGTH", FUNCID_LENGTH,  1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER, list_argtypes, NULL, ProgLength, },
    { "IF",     FUNCID_IF,      2, 3, FALSE, FIK_CONTROL,  TYPE_UNKNOWN, NULL,     checkif, NULL, }, /* non-strict */
    { "SETQ",   FUNCID_SETQ,    0, FIA_ANY, FALSE, FIK_SETQ,  TYPE_UNKNOWN, NULL,     checksetq, NULL, },
    { "SETQ*",  FUNCID_SETQSTAR, 0, FIA_ANY, FALSE, FIK_SETQ, TYPE_UNKNOWN, NULL,     checksetq, NULL, },
    { "SETQLIST",   FUNCID_SETQLIST, 1, FIA_ANY, FALSE, FIK_SETQLIST, TYPE_LIST, NULL,     checksetqlist, NULL, },
    { "SETQLIST*",  FUNCID_SETQLISTSTAR, 1, FIA_ANY, FALSE, FIK_SETQLIST, TYPE_LIST, NULL,     checksetqlist, NULL, },
    { "NEW",    FUNCID_NEW,     2, 2,   TRUE,   FIK_SETQ,  TYPE_RECORD, tabrec_argtypes, NULL, ProgNew },
    { "NEW*",   FUNCID_NEWSTAR, 2, 2,   TRUE,   FIK_SETQ,  TYPE_RECORD, tabrec_argtypes, NULL, ProgNewStar, },
    { "DELETE", FUNCID_DELETE,  2, 2,   TRUE,   FIK_SETQ,  TYPE_BOOL,      tabbool_argtypes, NULL, ProgDelete, },
    { "DELETE*", FUNCID_DELETESTAR, 2, 2, TRUE,   FIK_SETQ,  TYPE_BOOL,      tabbool_argtypes, NULL, ProgDeleteStar, },
    { "DELETEALL", FUNCID_DELETEALL, 1, 1, TRUE,  FIK_SETQ,  TYPE_BOOL,      tab_argtypes, NULL, ProgDeleteAll, },
    { "COPYREC", FUNCID_COPYREC, 2, 2,  TRUE,   FIK_SETQ,   TYPE_RECORD,    recrec_argtypes,  NULL, ProgCopyRec, },
    { "STR",    FUNCID_STR,     1, 1,   FALSE,  FIK_EXPR,   TYPE_STRING,    NULL,   checkstr, ProgString, },
    { "MEMO",   FUNCID_MEMO,    1, 1,   FALSE,  FIK_EXPR,   TYPE_MEMO,      NULL,   checkstr, ProgMemo, },
    { "INT",    FUNCID_INT,     1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER,   NULL,   NULL, ProgInteger, },
    { "REAL",   FUNCID_REAL,    1, 1,   FALSE,  FIK_EXPR,   TYPE_REAL,      NULL,   NULL, ProgReal, },
    { "DATE",   FUNCID_DATE,    1, 1,   FALSE,  FIK_EXPR,   TYPE_DATE,      NULL,   NULL, ProgDate, },
    { "TIME",   FUNCID_TIME,    1, 1,   FALSE,  FIK_EXPR,   TYPE_TIME,      NULL,   NULL, ProgTime, },
    { "STRP",   FUNCID_STRP,    1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,      NULL,   NULL, ProgStrP, },
    { "MEMOP",  FUNCID_MEMOP,   1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,      NULL,   NULL, ProgMemoP, },
    { "INTP",   FUNCID_INTP,    1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,      NULL,   NULL, ProgIntP, },
    { "REALP",  FUNCID_REALP,   1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,      NULL,   NULL, ProgRealP, },
    { "DATEP",  FUNCID_DATEP,   1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,      NULL,   NULL, ProgDateP, },
    { "TIMEP",  FUNCID_TIMEP,   1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,      NULL,   NULL, ProgTimeP, },
    { "RECP",   FUNCID_RECP,    2, 2,   FALSE,  FIK_EXPR,  TYPE_BOOL,      tabbool_argtypes, NULL, ProgRecP, },
    { "NULL",   FUNCID_NULL,    1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,      NULL,   NULL, ProgNull, },
    { "LISTP",  FUNCID_LISTP,   1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,      NULL,   NULL, ProgListP, },
    { "CONSP",  FUNCID_CONSP,   1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,      NULL,   NULL, ProgConsP, },
    { "GC",     FUNCID_GC,      0, 0,   FALSE,  FIK_CONTROL,  TYPE_NIL,       NULL,   NULL, ProgGC, },
    { "SETLABELS", FUNCID_SETLABELS, 2, 2, TRUE,   FIK_SETQ,  TYPE_STRING,    fieldstr_argtypes, checkstrfield, ProgSetLabels, },
    { "GETLABELS", FUNCID_GETLABELS, 1, 1, FALSE,  FIK_EXPR,  TYPE_STRING,    field_argtypes, checkstrchoicefield, ProgGetLabels,},
    { "SETCURSOR", FUNCID_SETCURSOR, 1, 1, TRUE,   FIK_SETQ,  TYPE_BOOL,      NULL,   checktabfield, ProgSetCursor, },
    { "SETBGPEN", FUNCID_SETBGPEN, 2, 2, TRUE,   FIK_SETQ,  TYPE_INTEGER,     fieldint_argtypes,  NULL, ProgSetBGPen, },
    { "SETWINDOWOPEN", FUNCID_SETWINDOWOPEN, 2, 2, TRUE, FIK_SETQ, TYPE_BOOL, NULL,   checktabfield, ProgSetWindowOpen, },
    { "GETWINDOWOPEN", FUNCID_GETWINDOWOPEN, 1, 1, TRUE, FIK_EXPR, TYPE_BOOL, NULL,   checktabfield, ProgGetWindowOpen, },
    { "SETVIRTUALLISTACTIVE", FUNCID_SETVIRTUALLISTACTIVE, 2, 2, TRUE, FIK_SETQ, TYPE_INTEGER, fieldint_argtypes, checkvirtfield, ProgSetVirtualListActive, },
    { "GETVIRTUALLISTACTIVE", FUNCID_GETVIRTUALLISTACTIVE, 1, 1, FALSE, FIK_EXPR, TYPE_INTEGER, field_argtypes, checkvirtfield, ProgGetVirtualListActive, },
    { "SYSTEM", FUNCID_SYSTEM,  1, FIA_ANY, TRUE,  FIK_EXPR,   TYPE_INTEGER,   NULL,           CheckPrintF, ProgSystem, },
    { "SYSTEM*", FUNCID_SYSTEMSTAR,  1, FIA_ANY, TRUE,  FIK_EXPR,   TYPE_INTEGER,   NULL,           CheckPrintF, ProgSystemStar, },
    { "EDIT",   FUNCID_EDIT,    1, 1,   TRUE,   FIK_EXPR,   TYPE_INTEGER,   str_argtypes,   NULL, ProgEdit, },
    { "EDIT*",  FUNCID_EDITSTAR, 1, 1,  TRUE,   FIK_EXPR,   TYPE_INTEGER,   str_argtypes,   NULL, ProgEditStar, },
    { "VIEW",   FUNCID_VIEW,    1, 1,   TRUE,   FIK_EXPR,   TYPE_INTEGER,   str_argtypes,   NULL, ProgView, },
    { "VIEW*",  FUNCID_VIEWSTAR, 1, 1,  TRUE,   FIK_EXPR,   TYPE_INTEGER,   str_argtypes,   NULL, ProgViewStar, },
    { "EXIT",   FUNCID_EXIT,    0, FIA_ANY, FALSE, FIK_CONTROL,  TYPE_UNKNOWN, NULL,     NULL,   NULL, },
    { "NEXT",   FUNCID_NEXT,    0, 0,     FALSE,  FIK_CONTROL,  TYPE_UNKNOWN, NULL,     NULL,   NULL, },
    { "RETURN", FUNCID_RETURN,  0, FIA_ANY, FALSE, FIK_CONTROL,  TYPE_UNKNOWN, NULL,     NULL,   NULL, },
    { "FERROR", FUNCID_FERROR,  1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,      file_argtypes,  NULL, ProgFError, },
    { "FEOF",   FUNCID_FEOF,    1, 1,   FALSE,  FIK_EXPR,   TYPE_BOOL,      file_argtypes,  NULL, ProgFEof, },
    { "FSEEK",  FUNCID_FSEEK,   3, 3,   TRUE,  FIK_EXPR,    TYPE_INTEGER,   fileintint_argtypes,  NULL, ProgFSeek, },
    { "FTELL",  FUNCID_FTELL,   1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER,   file_argtypes,  NULL, ProgFTell, },
    { "FGETCHAR", FUNCID_FGETCHAR, 1, 1,TRUE,   FIK_EXPR,   TYPE_STRING,    file_argtypes,  NULL, ProgFGetChar, },
    { "FGETCHARS", FUNCID_FGETCHARS, 2, 2, TRUE,   FIK_EXPR,   TYPE_STRING,    intfile_argtypes, NULL, ProgFGetChars, },
    { "FGETSTR", FUNCID_FGETSTR, 1, 1,  TRUE,   FIK_EXPR,   TYPE_STRING,    file_argtypes,  NULL, ProgFGetStr, },
    { "FGETMEMO", FUNCID_FGETMEMO, 1, 1,TRUE,   FIK_EXPR,   TYPE_MEMO,      file_argtypes,  NULL, ProgFGetMemo, },
    { "LEN",    FUNCID_LEN,     1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER,   str_argtypes,   NULL, ProgLen, },
    { "MAXLEN", FUNCID_MAXLEN,  1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER,   field_argtypes,  checkstrfield, ProgMaxLen, },
    { "RIGHTSTR", FUNCID_RIGHTSTR, 2, 2,FALSE,  FIK_EXPR,   TYPE_STRING,    strint_argtypes, NULL, ProgRightStr,},
    { "LEFTSTR", FUNCID_LEFTSTR, 2, 2,  FALSE,  FIK_EXPR,   TYPE_STRING,    strint_argtypes, NULL, ProgLeftStr, },
    { "MIDSTR", FUNCID_MIDSTR,  3, 3,   FALSE,  FIK_EXPR,   TYPE_STRING,    strintint_argtypes, NULL, ProgMidStr, },
    { "SETMIDSTR", FUNCID_SETMIDSTR, 3, 3, FALSE,  FIK_EXPR,   TYPE_STRING,    strintstr_argtypes, NULL, ProgSetMidStr, },
    { "INSMIDSTR", FUNCID_INSMIDSTR, 3, 3, FALSE,  FIK_EXPR,   TYPE_STRING,    strintstr_argtypes, NULL, ProgInsMidStr, },
    { "SHA1SUM",  FUNCID_SHA1SUM, 1, 1, FALSE,  FIK_EXPR,   TYPE_STRING,    str_argtypes,   NULL, ProgSHA1Sum, },
    { "UPPER",  FUNCID_UPPER,   1, 1,   FALSE,  FIK_EXPR,   TYPE_STRING,    str_argtypes,   NULL, ProgUpper, },
    { "LOWER",  FUNCID_LOWER,   1, 1,   FALSE,  FIK_EXPR,   TYPE_STRING,    str_argtypes,   NULL, ProgLower, },
    { "ASC",    FUNCID_ASC,     1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER,   str_argtypes,   NULL, ProgAsc, },
    { "CHR",    FUNCID_CHR,     1, 1,   FALSE,  FIK_EXPR,   TYPE_STRING,    int_argtypes,   NULL, ProgChr, },
    { "WORD",   FUNCID_WORD,    2, 2,   FALSE,  FIK_EXPR,   TYPE_STRING,    strint_argtypes, NULL, ProgWord, },
    { "WORDS",  FUNCID_WORDS,   1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER,   str_argtypes,    NULL, ProgWords, },
    { "FIELD",   FUNCID_FIELD,  2, 4,   FALSE,  FIK_EXPR,   TYPE_STRING,    strintstrbool_argtypes, NULL, ProgField, },
    { "FIELDS",  FUNCID_FIELDS, 1, 3,   FALSE,  FIK_EXPR,   TYPE_INTEGER,   strstrbool_argtypes,    NULL, ProgFields, },
    { "STRTOLIST", FUNCID_STRTOLIST, 1, 2, FALSE, FIK_EXPR, TYPE_LIST,      strstr_argtypes,  NULL, ProgStrToList, },
    { "LISTTOSTR", FUNCID_LISTTOSTR, 1, 2, FALSE, FIK_EXPR, TYPE_STRING,    liststr_argtypes,  NULL, ProgListToStr, },
    { "CONCAT", FUNCID_CONCAT,  0, FIA_ANY, FALSE, FIK_EXPR,   TYPE_STRING, NULL,      checkstrs, ProgConcat, },
    { "CONCAT2", FUNCID_CONCAT2, 1, FIA_ANY, FALSE, FIK_EXPR,  TYPE_STRING, NULL,      checkstrs, ProgConcat2, },
    { "COPYSTR", FUNCID_COPYSTR, 2, 2,  FALSE, FIK_EXPR,    TYPE_STRING,    strint_argtypes,   NULL, ProgCopyStr, },
    { "LINE",   FUNCID_LINE,    2, 2,   FALSE,  FIK_EXPR,   TYPE_STRING,    memoint_argtypes,  NULL, ProgLine, },
    { "LINES",  FUNCID_LINES,   1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER,   memo_argtypes,  NULL, ProgLines, },
    { "TRIMSTR", FUNCID_TRIMSTR, 1, 3, FALSE, FIK_EXPR, 	TYPE_STRING,   	NULL,   checktrimstr, ProgTrimStr, },
    { "REMCHARS", FUNCID_REMCHARS, 2, 2,FALSE,  FIK_EXPR,   TYPE_STRING,    strstr_argtypes, NULL, ProgRemChars, },
    { "REPLACESTR", FUNCID_REPLACESTR, 1, FIA_ANY, FALSE, FIK_EXPR,  TYPE_STRING, NULL, checkreplacestr, ProgReplaceStr, },
    { "REPLACESTR*", FUNCID_REPLACESTRSTAR, 1, FIA_ANY, FALSE, FIK_EXPR, TYPE_STRING, NULL, checkreplacestr, ProgReplaceStrStar, },
    { "INDEXSTR", FUNCID_INDEXSTR, 2, 2,FALSE,  FIK_EXPR,   TYPE_INTEGER,   strstr_argtypes, NULL, ProgIndexStr, },
    { "INDEXSTR*", FUNCID_INDEXSTRSTAR, 2, 2, FALSE, FIK_EXPR, TYPE_INTEGER,   strstr_argtypes, NULL, ProgIndexStrStar, },
    { "INDEXBRK", FUNCID_INDEXBRK, 2, 2,FALSE,  FIK_EXPR,   TYPE_INTEGER,   strstr_argtypes, NULL, ProgIndexBrk, },
    { "INDEXBRK*", FUNCID_INDEXBRKSTAR, 2, 2, FALSE, FIK_EXPR, TYPE_INTEGER,   strstr_argtypes, NULL, ProgIndexBrkStar, },
    { "RINDEXSTR", FUNCID_RINDEXSTR, 2, 2, FALSE,  FIK_EXPR,   TYPE_INTEGER,   strstr_argtypes, NULL, ProgRIndexStr, },
    { "RINDEXSTR*", FUNCID_RINDEXSTRSTAR, 2, 2, FALSE, FIK_EXPR, TYPE_INTEGER, strstr_argtypes, NULL, ProgRIndexStrStar, },
    { "RINDEXBRK", FUNCID_RINDEXBRK, 2, 2, FALSE,  FIK_EXPR,   TYPE_INTEGER,   strstr_argtypes, NULL, ProgRIndexBrk, },
    { "RINDEXBRK*", FUNCID_RINDEXBRKSTAR, 2, 2, FALSE, FIK_EXPR, TYPE_INTEGER, strstr_argtypes, NULL, ProgRIndexBrkStar, },
    { "HALT",   FUNCID_HALT,    0, 0,     TRUE,   FIK_CONTROL,  TYPE_NIL,       NULL,           NULL, ProgHalt, },
    { "ERROR",  FUNCID_ERROR,   1, FIA_ANY, TRUE,  FIK_CONTROL,  TYPE_NIL,       NULL,   CheckPrintF, ProgError, },
    { "FPUTCHAR", FUNCID_FPUTCHAR, 2, 2,  TRUE,   FIK_EXPR,   TYPE_STRING,    strfile_argtypes, NULL, ProgFPutChar, },
    { "FPUTSTR",  FUNCID_FPUTSTR,  2, 2,  TRUE,   FIK_EXPR,   TYPE_STRING,    strfile_argtypes, NULL, ProgFPutStr,  },
    { "FPUTMEMO", FUNCID_FPUTMEMO, 2, 2,   TRUE,   FIK_EXPR,   TYPE_MEMO,      memofile_argtypes, NULL, ProgFPutMemo, },
    { "FUNCALL", FUNCID_FUNCALL, 1, FIA_ANY, TRUE, FIK_EXPR,   TYPE_UNKNOWN,   NULL,   checkfuncall, ProgFunCall, },
    { "APPLY", FUNCID_APPLY, 1, FIA_ANY, TRUE, FIK_EXPR,   TYPE_UNKNOWN,   NULL,   checkapply, ProgApply, },
    { "MEMOTOLIST", FUNCID_MEMOTOLIST, 1, 2, FALSE, FIK_EXPR,  TYPE_LIST,      memobool_argtypes,  NULL, ProgMemoToList, },
    { "LISTTOMEMO", FUNCID_LISTTOMEMO, 1, 1, FALSE, FIK_EXPR,  TYPE_STRING,    list_argtypes,  NULL, ProgListToMemo, },
    { "FILLMEMO", FUNCID_FILLMEMO, 1, 1,FALSE,  TRUE,   TYPE_STRING,    str_argtypes,   NULL, ProgFillMemo, },
    { "FORMATMEMO", FUNCID_FORMATMEMO, 2, 4, FALSE, FIK_EXPR,  TYPE_STRING, strintboolbool_argtypes, NULL, ProgFormatMemo, },
    { "INDENTMEMO", FUNCID_INDENTMEMO, 2, 2, FALSE, FIK_EXPR,  TYPE_MEMO,      memoint_argtypes,  NULL, ProgIndentMemo, },
    { "TRUNC",  FUNCID_TRUNC,   1, 1,   FALSE,  FIK_EXPR,   TYPE_REAL,      real_argtypes,  NULL, ProgTrunc, },
    { "ROUND",  FUNCID_ROUND,   2, 2,   FALSE,  FIK_EXPR,   TYPE_REAL,      realint_argtypes,  NULL, ProgRound, },
    { "GETMATCHFILTER", FUNCID_GETMATCHFILTER, 1, 1, FALSE, FIK_EXPR, TYPE_BOOL, rec_argtypes,  NULL, ProgGetMatchFilter, },
    { "SETMATCHFILTER", FUNCID_SETMATCHFILTER, 2, 2, TRUE, FIK_EXPR, TYPE_BOOL, recbool_argtypes, NULL, ProgSetMatchFilter, },
    { "GETFILTERACTIVE", FUNCID_GETFILTERACTIVE, 1, 1, FALSE, FIK_EXPR, TYPE_BOOL, tab_argtypes, NULL, ProgGetFilterActive, },
    { "SETFILTERACTIVE", FUNCID_SETFILTERACTIVE, 2, 2, TRUE, FIK_SETQ, TYPE_BOOL, tabbool_argtypes, NULL, ProgSetFilterActive, },
    { "GETFILTERSTR", FUNCID_GETFILTERSTR, 1, 1, FALSE, FIK_EXPR, TYPE_STRING, tab_argtypes,  NULL, ProgGetFilterStr, },
    { "SETFILTERSTR", FUNCID_SETFILTERSTR, 2, 2, TRUE, FIK_SETQ, TYPE_BOOL,   tabstr_argtypes, NULL, ProgSetFilterStr, },
    { "GETISSORTED", FUNCID_GETISSORTED, 1, 1, FALSE, FIK_EXPR, TYPE_BOOL, rec_argtypes,  NULL, ProgGetIsSorted, },
    { "SETISSORTED", FUNCID_SETISSORTED, 2, 2, TRUE, FIK_EXPR, TYPE_BOOL, recbool_argtypes,  NULL, ProgSetIsSorted, },
    { "REORDER", FUNCID_REORDER, 1, 1, TRUE, FIK_SETQ, TYPE_BOOL, tab_argtypes, NULL, ProgReorder, },
    { "REORDERALL", FUNCID_REORDERALL, 1, 1, TRUE, FIK_SETQ, TYPE_BOOL, tab_argtypes, NULL, ProgReorderAll, },
    { "GETORDERSTR", FUNCID_GETORDERSTR, 1, 1, FALSE, FIK_EXPR, TYPE_STRING,  tab_argtypes,   NULL, ProgGetOrderStr, },
    { "SETORDERSTR", FUNCID_SETORDERSTR, 2, 2, TRUE, FIK_SETQ, TYPE_BOOL,     tabstr_argtypes, NULL, ProgSetOrderStr, },
    { "RECNUM", FUNCID_RECNUM,  1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER,   rec_argtypes,   NULL, ProgRecNum,  },
    { "MOVEREC", FUNCID_MOVEREC,  2, 2, TRUE, FIK_EXPR, TYPE_INTEGER,   recint_argtypes,   NULL, ProgMoveRec,  },
    { "GETREC", FUNCID_GETREC, 1, 1,    FALSE,  FIK_EXPR,  TYPE_RECORD,    NULL,           NULL, ProgGetRec, },
    { "SETREC", FUNCID_SETREC, 2, 2,    FALSE,  FIK_EXPR,  TYPE_UNKNOWN,   anyrec_argtypes, NULL, ProgSetRec, },
    { "RECORDS", FUNCID_RECORDS, 1, 1,  FALSE,  FIK_EXPR,  TYPE_INTEGER,   tab_argtypes,   NULL, ProgRecords,  },
    { "RECORD", FUNCID_RECORD,  2, 2,   FALSE,  FIK_EXPR,  TYPE_RECORD,    tabint_argtypes, NULL, ProgRecord,  },
    { "TABLENAME", FUNCID_TABLENAME, 1, 1,  FALSE,   FIK_EXPR, TYPE_STRING,    tab_argtypes,   NULL, ProgRelName, },
    { "FIELDNAME", FUNCID_FIELDNAME, 1, 1,FALSE,   FIK_EXPR, TYPE_STRING,    field_argtypes,  NULL, ProgFieldName, },
    { "GETDISABLED", FUNCID_GETDISABLED, 1, 1, FALSE, FIK_EXPR, TYPE_BOOL,    field_argtypes,  checkobsolete, ProgObsoleteFunction, },
    { "SETDISABLED", FUNCID_SETDISABLED, 2, 2, TRUE, FIK_SETQ, TYPE_BOOL,     fieldbool_argtypes, checkobsolete, ProgObsoleteFunction, },
    { "GETWINDOWDISABLED", FUNCID_GETWINDOWDISABLED, 1, 1, FALSE, FIK_EXPR, TYPE_BOOL,    NULL,  checkobsolete, ProgObsoleteFunction, },
    { "SETWINDOWDISABLED", FUNCID_SETWINDOWDISABLED, 2, 2, TRUE, FIK_SETQ, TYPE_BOOL,     NULL,  checkobsolete, ProgObsoleteFunction, },
    { "ASKFILE", FUNCID_ASKFILE, 4, 4,  TRUE,   FIK_EXPR,  TYPE_STRING,     strstrstrbool_argtypes, NULL, ProgAskFile, },
    { "ASKDIR", FUNCID_ASKDIR,  4, 4,   TRUE,   FIK_EXPR,  TYPE_STRING,     strstrstrbool_argtypes, NULL, ProgAskDir, },
    { "ASKSTR", FUNCID_ASKSTR,  4, 5,   TRUE,   FIK_EXPR,  TYPE_STRING,     strstrstrintbool_argtypes,  NULL, ProgAskStr, },
    { "ASKINT", FUNCID_ASKINT,  5, 5,   TRUE,   FIK_EXPR,  TYPE_INTEGER,    strstrintintint_argtypes,  NULL, ProgAskInt, },
    { "ASKCHOICE", FUNCID_ASKCHOICE, 4, 5, TRUE,   FIK_EXPR,  TYPE_INTEGER,    strstranyintany_argtypes,  NULL, ProgAskChoice, },
    { "ASKCHOICESTR", FUNCID_ASKCHOICESTR, 4, 5, TRUE, FIK_EXPR, TYPE_STRING,  strstranystrany_argtypes,  NULL, ProgAskChoiceStr, },
    { "ASKOPTIONS", FUNCID_ASKOPTIONS, 4, 5, TRUE, FIK_EXPR,   TYPE_LIST,      strstranylistany_argtypes,  NULL, ProgAskOptions, },
    { "ASKBUTTON", FUNCID_ASKBUTTON, 4, 4, TRUE,   FIK_EXPR,   TYPE_INTEGER,   strstrliststr_argtypes,  NULL, ProgAskButton, },
    { "ASKMULTI", FUNCID_ASKMULTI, 3, 3,  TRUE,   FIK_EXPR,   TYPE_LIST,      strstrlist_argtypes,    NULL, ProgAskMulti, },
    { "MAPFIRST", FUNCID_MAPFIRST, 0, FIA_ANY, FALSE, FIK_EXPR, TYPE_LIST,     NULL,   CheckMapFirst,  ProgMapFirst, },
    { "SORTLIST", FUNCID_SORTLIST, 2, 2,  FALSE,  FIK_EXPR,   TYPE_LIST,      funclist_argtypes,      NULL, ProgSortList, },
    { "SORTLISTGT", FUNCID_SORTLISTGT, 2, 2, FALSE, FIK_EXPR,  TYPE_LIST,      funclist_argtypes,      NULL, ProgSortListGt, },
    { "PROJECTNAME", FUNCID_PROJECTNAME, 0, 0, FALSE, FIK_EXPR, TYPE_STRING,   NULL,                   NULL, ProgProjectName, },
    { "PREPARECHANGE", FUNCID_PREPARECHANGE, 0, 0, TRUE, FIK_CONTROL, TYPE_NIL,  NULL, NULL, ProgPrepareChange, },
    { "CHANGES", FUNCID_CHANGES, 0, 0,  FALSE,  FIK_EXPR,   TYPE_INTEGER,   NULL,                   NULL, ProgChanges, },
    { "GETADMINMODE", FUNCID_GETADMINMODE, 0, 0, FALSE, FIK_EXPR, TYPE_BOOL, NULL, NULL, ProgGetAdminMode },
    { "SETADMINMODE", FUNCID_SETADMINMODE, 1, 1, TRUE, FIK_CONTROL, TYPE_BOOL, bool_argtypes, NULL, ProgSetAdminMode },
    { "ADMINPASSWORD", FUNCID_ADMINPASSWORD, 0, 0, FALSE, FIK_EXPR, TYPE_STRING, NULL, NULL, ProgAdminPassword },
    { "STAT",   FUNCID_STAT,    1, 1,   FALSE,  FIK_EXPR,   TYPE_INTEGER,   str_argtypes,           NULL, ProgStat, },
    { "MESSAGE", FUNCID_MESSAGE, 1, FIA_ANY, TRUE, FIK_EXPR,  TYPE_STRING,    NULL,                   CheckPrintF, ProgMessage, },
    { "COMPLETEMAX", FUNCID_COMPLETEMAX, 1, 1,   TRUE,   FIK_EXPR,   TYPE_INTEGER,   int_argtypes,           NULL, ProgCompleteMax, },
    { "COMPLETEADD", FUNCID_COMPLETEADD, 1, 1,   TRUE,   FIK_EXPR,   TYPE_INTEGER,   int_argtypes,           NULL, ProgCompleteAdd, },
    { "COMPLETE", FUNCID_COMPLETE, 1,  1,  TRUE,   FIK_EXPR,   TYPE_INTEGER,   int_argtypes,           NULL, ProgComplete, },
    { "TACKON", FUNCID_TACKON, 1, FIA_ANY, FALSE, FIK_EXPR, TYPE_STRING, NULL,                   checkstrs, ProgTackon, },
    { "FILENAME", FUNCID_FILENAME, 1, 1, FALSE, FIK_EXPR, TYPE_STRING, str_argtypes,                   NULL, ProgFilename, },
    { "DIRNAME", FUNCID_DIRNAME, 1, 1, FALSE, FIK_EXPR, TYPE_STRING, str_argtypes,                   NULL, ProgDirname, },
    { "PUBSCREEN",  FUNCID_PUBSCREEN, 0, 0, FALSE, FIK_EXPR, TYPE_STRING, NULL, NULL, ProgPubScreen, },
};

static int funcinfo_cmp(const void *v1, const void *v2)
{
    const struct FuncInfo *fi1 = v1, *fi2 = v2;

    return(fi1->fi_Id - fi2->fi_Id);
}

static bool funcinfosinitialized = FALSE;

static void funcinfosinit(void)
{
    struct FuncInfo *fis = FuncInfos;
    ushort i, index;

    qsort(fis, NumFuncIds, sizeof(struct FuncInfo), funcinfo_cmp);

    /* build hash table */
    for(i=0; i < NumFuncIds; i++)
        fis[i].fi_HashFirst = fis[i].fi_HashNext = NULL;

    for(i=0; i < NumFuncIds; i++)
    {
        index = HashString(fis[i].fi_Name) % NumFuncIds;

        fis[i].fi_HashNext = fis[index].fi_HashFirst;
        fis[index].fi_HashFirst = &fis[i];
    }

    funcinfosinitialized = TRUE;
}

struct FuncInfo *FindFuncInfoName(char *name)
{
    struct FuncInfo *fi, *last = NULL;
    ushort index;

    if(name == NULL)
        return(NULL);

    if(!funcinfosinitialized)
        funcinfosinit();

    index = HashString(name) % NumFuncIds;
    fi = FuncInfos[index].fi_HashFirst;
    for(; fi != NULL; last = fi, fi = fi->fi_HashNext)
        if(!CodesetStrCmp(fi->fi_Name, name))
        {
            /*
            ** Move fi to first element in hash list.
            */
            if(last != NULL)
            {
                last->fi_HashNext = fi->fi_HashNext;
                fi->fi_HashNext = FuncInfos[index].fi_HashFirst;
                FuncInfos[index].fi_HashFirst = fi;
            }

            return(fi);
        }
    return(NULL);
}

struct FuncInfo *FindFuncInfoId(short id)
{
    if(id < 0 || id >= NumFuncIds)
        return(NULL);

    if(!funcinfosinitialized)
        funcinfosinit();

    return(&FuncInfos[id]);
}

struct ConstInfo ConstInfos[NumConstIds] =
{
    { "NIL",      CONSTID_NIL,    { TYPE_NIL,       0, 0, 0, NULL, } },
    { "TRUE",     CONSTID_TRUE,   { TYPE_BOOL,      0, 0, 0, NULL, } },
    { "RESET",    CONSTID_RESET,  { TYPE_STRING,    0, 0, 0, NULL, } },
    { "NLQON",    CONSTID_NLQON,  { TYPE_STRING,    0, 0, 0, NULL, } },
    { "NLQOFF",   CONSTID_NLQOFF, { TYPE_STRING,    0, 0, 0, NULL, } },
    { "ELITEON",  CONSTID_ELITEON, { TYPE_STRING,   0, 0, 0, NULL, } },
    { "ELITEOFF", CONSTID_ELITEOFF, { TYPE_STRING,  0, 0, 0, NULL, } },
    { "NORMAL",   CONSTID_NORMAL, { TYPE_STRING,    0, 0, 0, NULL, } },
    { "BFON",     CONSTID_BFON,   { TYPE_STRING,    0, 0, 0, NULL, } },
    { "BFOFF",    CONSTID_BFOFF,  { TYPE_STRING,    0, 0, 0, NULL, } },
    { "ITON",     CONSTID_ITON,   { TYPE_STRING,    0, 0, 0, NULL, } },
    { "ITOFF",    CONSTID_ITOFF,  { TYPE_STRING,    0, 0, 0, NULL, } },
    { "ULON",     CONSTID_ULON,   { TYPE_STRING,    0, 0, 0, NULL, } },
    { "ULOFF",    CONSTID_ULOFF,  { TYPE_STRING,    0, 0, 0, NULL, } },
    { "WIDEON",   CONSTID_WIDEON, { TYPE_STRING,    0, 0, 0, NULL, } },
    { "WIDEOFF",  CONSTID_WIDEOFF, { TYPE_STRING,   0, 0, 0, NULL, } },
    { "CONDON",   CONSTID_CONDON, { TYPE_STRING,    0, 0, 0, NULL, } },
    { "CONDOFF",  CONSTID_CONDOFF, { TYPE_STRING,   0, 0, 0, NULL, } },
    { "INT_MAX",  CONSTID_INT_MAX, { TYPE_INTEGER,  0, 0, 0, NULL, } },
    { "INT_MIN",  CONSTID_INT_MIN, { TYPE_INTEGER,  0, 0, 0, NULL, } },
    { "HUGE_VAL", CONSTID_HUGE_VAL, { TYPE_REAL,    0, 0, 0, NULL, } },
    { "PI",       CONSTID_PI,      { TYPE_REAL,     0, 0, 0, NULL, } },
    { "OSTYPE",   CONSTID_OSTYPE,  { TYPE_STRING,   0, 0, 0, NULL, } },
    { "OSVER",    CONSTID_OSVER,   { TYPE_INTEGER,  0, 0, 0, NULL, } },
    { "OSREV",    CONSTID_OSREV,   { TYPE_INTEGER,  0, 0, 0, NULL, } },
    { "LANGUAGE", CONSTID_LANGUAGE, { TYPE_STRING,  0, 0, 0, NULL, } },
    { "MBVER",    CONSTID_MBVER,   { TYPE_INTEGER,  0, 0, 0, NULL, } },
    { "MBREV",    CONSTID_MBREV,   { TYPE_INTEGER,  0, 0, 0, NULL, } },
    { "SEEK_SET", CONSTID_SEEK_SET, { TYPE_INTEGER, 0, 0, 0, NULL, } },
    { "SEEK_CUR", CONSTID_SEEK_CUR, { TYPE_INTEGER, 0, 0, 0, NULL, } },
    { "SEEK_END", CONSTID_SEEK_END, { TYPE_INTEGER, 0, 0, 0, NULL, } },
    { "PEN_WHITE",CONSTID_PEN_WHITE,{ TYPE_INTEGER, 0, 0, 0, NULL, } },
    { "PEN_BLACK",CONSTID_PEN_BLACK,{ TYPE_INTEGER, 0, 0, 0, NULL, } },
    { "PEN_RED",  CONSTID_PEN_RED,  { TYPE_INTEGER, 0, 0, 0, NULL, } },
    { "PEN_BLUE", CONSTID_PEN_BLUE, { TYPE_INTEGER, 0, 0, 0, NULL, } },
    { "PEN_GREEN",CONSTID_PEN_GREEN,{ TYPE_INTEGER, 0, 0, 0, NULL, } },
    { "PEN_CYAN" ,CONSTID_PEN_CYAN ,{ TYPE_INTEGER, 0, 0, 0, NULL, } },
    { "PEN_MAGENTA", CONSTID_PEN_MAGENTA, { TYPE_INTEGER, 0, 0, 0, NULL, } },
    { "PEN_YELLOW", CONSTID_PEN_YELLOW, { TYPE_INTEGER, 0, 0, 0, NULL, } },
    { "TMPDIR",   CONSTID_TMPDIR,   { TYPE_STRING,  0, 0, 0, NULL, } },   
};

static int constinfo_cmp(const void *v1, const void *v2)
{
    const struct ConstInfo *ci1 = v1, *ci2 = v2;

    return(ci1->ci_Id - ci2->ci_Id);
}

static bool constinfosinitialized = FALSE;

static void constinfosinit(void)
{
    struct ConstInfo *cis = ConstInfos;
    ushort i, index;

    qsort(cis, NumConstIds, sizeof(struct ConstInfo), constinfo_cmp);

    cis[CONSTID_RESET].ci_Element.el_Contents.string.string = ANSI_RESET;
    cis[CONSTID_RESET].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_NLQON].ci_Element.el_Contents.string.string = ANSI_NLQON;
    cis[CONSTID_NLQON].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_NLQOFF].ci_Element.el_Contents.string.string = ANSI_NLQOFF;
    cis[CONSTID_NLQOFF].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_ELITEON].ci_Element.el_Contents.string.string = ANSI_ELITEON;
    cis[CONSTID_ELITEON].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_ELITEOFF].ci_Element.el_Contents.string.string = ANSI_ELITEOFF;
    cis[CONSTID_ELITEOFF].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_NORMAL].ci_Element.el_Contents.string.string = ANSI_NORMAL;
    cis[CONSTID_NORMAL].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_BFON].ci_Element.el_Contents.string.string = ANSI_BOLDFACEON;
    cis[CONSTID_BFON].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_BFOFF].ci_Element.el_Contents.string.string = ANSI_BOLDFACEOFF;
    cis[CONSTID_BFOFF].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_ITON].ci_Element.el_Contents.string.string = ANSI_ITALICON;
    cis[CONSTID_ITON].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_ITOFF].ci_Element.el_Contents.string.string = ANSI_ITALICOFF;
    cis[CONSTID_ITOFF].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_ULON].ci_Element.el_Contents.string.string = ANSI_UNDERLINEON;
    cis[CONSTID_ULON].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_ULOFF].ci_Element.el_Contents.string.string = ANSI_UNDERLINEOFF;
    cis[CONSTID_ULOFF].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_WIDEON].ci_Element.el_Contents.string.string = ANSI_WIDEON;
    cis[CONSTID_WIDEON].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_WIDEOFF].ci_Element.el_Contents.string.string = ANSI_WIDEOFF;
    cis[CONSTID_WIDEOFF].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_CONDON].ci_Element.el_Contents.string.string = ANSI_CONDENSEDON;
    cis[CONSTID_CONDON].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_CONDOFF].ci_Element.el_Contents.string.string = ANSI_CONDENSEDOFF;
    cis[CONSTID_CONDOFF].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_INT_MAX].ci_Element.el_Contents.integer = LONG_MAX;
    cis[CONSTID_INT_MIN].ci_Element.el_Contents.integer = LONG_MIN;
#if (defined __AROS__) && (defined __x86_64__)
    /* 
    ** Work-around toolchain-core-x86_64-v11 compiler bug
    ** when running on arm64 with qemu
    */
    cis[CONSTID_PI].ci_Element.el_Contents.real = 2.0 * acos(0.0);
#else
    cis[CONSTID_PI].ci_Element.el_Contents.real = M_PI;
#endif
    cis[CONSTID_HUGE_VAL].ci_Element.el_Contents.real = HUGE_VAL;
    cis[CONSTID_OSTYPE].ci_Element.el_Contents.string.string = OS_TYPE;
    cis[CONSTID_OSTYPE].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_OSVER].ci_Element.el_Contents.integer = OSVersion;
    cis[CONSTID_OSREV].ci_Element.el_Contents.integer = OSRevision;
    cis[CONSTID_LANGUAGE].ci_Element.el_Contents.string.string = 
							(char *)GuiLanguage();
    cis[CONSTID_LANGUAGE].ci_Element.el_Contents.string.field = NULL;
    cis[CONSTID_MBVER].ci_Element.el_Contents.integer = BB_VERSION;
    cis[CONSTID_MBREV].ci_Element.el_Contents.integer = BB_REVISION;
    cis[CONSTID_SEEK_SET].ci_Element.el_Contents.integer = SEEK_SET;
    cis[CONSTID_SEEK_CUR].ci_Element.el_Contents.integer = SEEK_CUR;
    cis[CONSTID_SEEK_END].ci_Element.el_Contents.integer = SEEK_END;
    cis[CONSTID_PEN_WHITE].ci_Element.el_Contents.integer = 0xFFFFFF;
    cis[CONSTID_PEN_BLACK].ci_Element.el_Contents.integer = 0x000000;
    cis[CONSTID_PEN_RED].ci_Element.el_Contents.integer = 0xFF0000;
    cis[CONSTID_PEN_GREEN].ci_Element.el_Contents.integer = 0x00FF00;
    cis[CONSTID_PEN_BLUE].ci_Element.el_Contents.integer = 0x0000FF;
    cis[CONSTID_PEN_CYAN].ci_Element.el_Contents.integer = 0x00FFFF;
    cis[CONSTID_PEN_MAGENTA].ci_Element.el_Contents.integer = 0xFF00FF;
    cis[CONSTID_PEN_YELLOW].ci_Element.el_Contents.integer = 0xFFFF00;
    cis[CONSTID_TMPDIR].ci_Element.el_Contents.string.string =
        (char *)GetTmpDir();

    /* build hash table */
    for(i=0; i < NumConstIds; i++)
        cis[i].ci_HashFirst = cis[i].ci_HashNext = NULL;

    for(i=0; i < NumConstIds; i++)
    {
        index = HashString(cis[i].ci_Name) % NumConstIds;

        cis[i].ci_HashNext = cis[index].ci_HashFirst;
        cis[index].ci_HashFirst = &cis[i];
    }

    constinfosinitialized = TRUE;
}

struct ConstInfo *FindConstInfoName(char *name)
{
    struct ConstInfo *ci, *last = NULL;
    ushort index;

    if(name == NULL)
        return(NULL);

    if(!constinfosinitialized)
        constinfosinit();

    index = HashString(name) % NumConstIds;
    ci = ConstInfos[index].ci_HashFirst;
    for(; ci != NULL; last = ci, ci = ci->ci_HashNext)
        if(!CodesetStrCmp(ci->ci_Name, name))
        {
            /*
            ** Move ci to first element in hash list.
            */
            if(last != NULL)
            {
                last->ci_HashNext = ci->ci_HashNext;
                ci->ci_HashNext = ConstInfos[index].ci_HashFirst;
                ConstInfos[index].ci_HashFirst = ci;
            }

            return(ci);
        }
    return(NULL);
}

struct ConstInfo *FindConstInfoId(short id)
{
    if(id < 0 || id >= NumConstIds)
        return(NULL);

    if(!constinfosinitialized)
        constinfosinit();

    return(&ConstInfos[id]);
}

