/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** heap.c
**
** Creation date: 21.05.1995
** Last modified: 13.05.2022
*/

#include "heap.h"
#include "stack.h"
#include "programbin.h"
#include "programbin_impl.h"
#include "break.h"
#include "memo.h"
#include "memory.h"
#include "appstrings_cat.h"
#include "error.h"

struct Element ElementNil = { TYPE_NIL, FALSE, FALSE, 0, NULL, };
struct Element ElementTrue = { TYPE_BOOL, FALSE, FALSE, 0, NULL, };

#define HEAP_BLOCK_COUNT 1024

struct HeapElement
{
    struct Element he_Element;
    struct HeapElement *he_Next;
};

struct HeapBlock
{
    struct HeapElement hb_Elements[HEAP_BLOCK_COUNT];
};

static struct
{
    struct HeapElement *heap_FreeElement;
    struct HeapElement *heap_UsedElement;
    long heap_Size;			/* capacity */
    long heap_Len;			/* used space */
    long heap_AllocCount;               /* counts number of allocations */
    long heap_CheckCount;		/* next time to check for GC */
    struct GarbageCollectionBase heap_GCBase;
} Heap;

static bool HeapValid = FALSE;

void InitHeap(void)
{
    memset(&Heap, 0, sizeof(Heap));
    Heap.heap_CheckCount = HEAP_BLOCK_COUNT / 2;
    HeapValid = TRUE;
}

void ExitHeap(void)
{
    Heap.heap_FreeElement = NULL;
    HeapValid = FALSE;
}

static struct HeapElement *extendHeap(void)
{
    struct HeapBlock *block = ExecuteAllocMem(sizeof(struct HeapBlock));
    struct HeapElement *elem = block->hb_Elements;
    struct HeapElement *end = elem + HEAP_BLOCK_COUNT;
    struct HeapElement **pred = &Heap.heap_FreeElement;

    Heap.heap_Size += HEAP_BLOCK_COUNT;

    while (elem != end) {
	*pred = elem;
	pred = &elem->he_Next;
	elem++;
    }
    *pred = NULL;

    return Heap.heap_FreeElement;
}

static __inline struct Element *allocElement(void)
{
    struct HeapElement *elem = Heap.heap_FreeElement;

    if (elem == NULL) {
	elem = extendHeap();
    }
    Heap.heap_FreeElement = elem->he_Next;
    elem->he_Next = Heap.heap_UsedElement;
    Heap.heap_UsedElement = elem;

    elem->he_Element.el_Record = NULL;

    Heap.heap_AllocCount++;
    Heap.heap_Len++;

    return &elem->he_Element;
}

struct Element *MakeElement(char type, bool allocated)
{
    struct Element *elem = allocElement();

    elem->el_Type = type;
    elem->el_Allocated = allocated;

    return elem;
}

__inline struct Element *MakeInteger(long i)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_INTEGER;
    elem->el_Allocated = FALSE;
    elem->el_Contents.integer = i;

    return elem;
}

__inline struct Element *MakeReal(double r)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_REAL;
    elem->el_Allocated = FALSE;
    elem->el_Contents.real = r;

    return elem;
}

__inline struct Element *MakeString(char *s, bool allocated)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_STRING;
    elem->el_Allocated = allocated;
    elem->el_Contents.string.string = s;
    elem->el_Contents.string.field = NULL;

    return(elem);
}

__inline struct Element *MakeMemo(char *mi, bool allocated)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_MEMO;
    elem->el_Allocated = allocated;
    elem->el_Contents.mi = mi;
    
    return elem;
}

__inline struct Element *MakeDate(long d)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_DATE;
    elem->el_Allocated = FALSE;
    elem->el_Contents.date = d;

    return elem;
}

__inline struct Element *MakeTime(long t)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_TIME;
    elem->el_Allocated = FALSE;
    elem->el_Contents.time = t;

    return elem;
}

__inline struct Element *MakeRecord(struct Table *tab, struct Record *rec)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_RECORD;
    elem->el_Allocated = FALSE;
    elem->el_Contents.tabrec.tab = tab;
    elem->el_Contents.tabrec.rec = rec;
    
    return elem;
}

__inline struct Element *MakeTable(struct Table *tab, bool star)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_TABLE;
    elem->el_Allocated = FALSE;
    elem->el_Contents.tab.tab = tab;
    elem->el_Contents.tab.star = star;
    
    return elem;
}

__inline struct Element *MakeField(struct RelFieldPath *rap, 
	struct Record *rec, bool allocated)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_FIELD;
    elem->el_Allocated = allocated;
    elem->el_Contents.field.rap = rap;
    elem->el_Contents.field.rec = rec;

    return elem;
}

__inline struct Element *MakeList(struct Element *head, struct Element *tail)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_LIST;
    elem->el_Allocated = FALSE;
    elem->el_Contents.list.head = head;
    elem->el_Contents.list.tail = tail;

    return elem;
}

__inline struct Element *MakeFile(short num, ulong id)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_FILE;
    elem->el_Allocated = FALSE;
    elem->el_Contents.file.num = num;
    elem->el_Contents.file.id = id;
    
    return elem;
}

__inline struct Element *MakeCodeAddress(const char *buffer, long index)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_CODEADDRESS;
    elem->el_Allocated = FALSE;
    elem->el_Contents.codeaddress.buffer = buffer;
    elem->el_Contents.codeaddress.index = index;
    
    return elem;
}

__inline struct Element *MakeSpecial(void *ptr, 
	void (*func)(void *, short, struct Record *))
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_SPECIAL;
    elem->el_Allocated = TRUE;
    elem->el_Contents.special.ptr = ptr;
    elem->el_Contents.special.func = func;
    
    return elem;
}

__inline struct Element *MakeFunction(short type, void *ptr)
{
    struct Element *elem = allocElement();

    elem->el_Type = TYPE_FUNCTION;
    elem->el_Allocated = FALSE;
    elem->el_Contents.function.type = type;
    elem->el_Contents.function.ptr = ptr;
    
    return elem;
}


static void markElement(struct Element *elem)
{
    while (elem->el_Type == TYPE_LIST && elem->el_InUse == FALSE)
    {
        elem->el_InUse = TRUE;
        markElement(elem->el_Contents.list.head);
        elem = elem->el_Contents.list.tail;
    }
    elem->el_InUse = TRUE;
}

void FreeElementContents(struct MemPool *memPool, struct Element *elem)
{
    if (elem->el_Allocated) {
        switch (elem->el_Type) {
            case TYPE_STRING:
                FreeString(memPool, elem->el_Contents.string.string);
                break;

            case TYPE_MEMO:
                FreeString(memPool, elem->el_Contents.mi);
                break;

            case TYPE_FIELD:
                PoolFreeMem(memPool,
                    elem->el_Contents.field.rap, sizeof(struct RelFieldPath));
                break;

            case TYPE_SPECIAL:
                if(elem->el_Contents.special.func)
                    elem->el_Contents.special.func(
                        elem->el_Contents.special.ptr, ELEM_SPECIAL_FREE, NULL);
                break;
	}
    }
}

void GarbageCollection(void)
{
    struct HeapElement *he, *next, **pred, *free = Heap.heap_FreeElement;
    struct HeapElement *end = Heap.heap_GCBase.gcb_Heap;
    struct VirtualFieldComputeCacheEntry *entry;
    struct Element *elem;
    long i, heapLen = Heap.heap_Len;

    /* clear InUse flags */
    for (he = Heap.heap_UsedElement; he != end; he = he->he_Next) {
	he->he_Element.el_InUse = FALSE;
    }

    /* mark all reachable elements */
    for (i = Heap.heap_GCBase.gcb_Stack; (elem = StackElement(i)) != NULL; i++){
	markElement(elem);
    }
    FORNODE(entry, &executeVirtualFieldComputeCache) {
        markElement(entry->vfcce_Element);
    }

    /* free unused elements */
    pred = &Heap.heap_UsedElement;
    for (he = Heap.heap_UsedElement; he != end; he = next) {
	next = he->he_Next;
	if (!he->he_Element.el_InUse) {
	    *pred = next;
	    he->he_Next = free;
	    free = he;
	    FreeElementContents(executeMemPool, &he->he_Element);
	    heapLen--;
	} else {
	    pred = &he->he_Next;
	}
    }

    Heap.heap_FreeElement = free;
    Heap.heap_Len = heapLen;
    Heap.heap_AllocCount = 0;
    Heap.heap_CheckCount = Heap.heap_Size / 2;
}

void GetGarbageCollectionBase(struct GarbageCollectionBase *gcb)
{
    if (gcb) {
	*gcb = Heap.heap_GCBase;
    }
}

void SetGarbageCollectionBase(const struct GarbageCollectionBase *gcb)
{
    if (gcb) {
	Heap.heap_GCBase = *gcb;
    }
}

void SetGarbageCollectionBaseCurrent(void)
{
    Heap.heap_GCBase.gcb_Heap = Heap.heap_UsedElement;
    Heap.heap_GCBase.gcb_Stack = StackSize();
}

static void elementRemTableNonZero(struct Element *elem, struct Table *tab)
{
    while (elem->el_Type == TYPE_LIST) {
        elementRemTableNonZero(elem->el_Contents.list.head, tab);
        elem = elem->el_Contents.list.tail;
    }

    switch (elem->el_Type) {
        case TYPE_RECORD:
            if (elem->el_Contents.tabrec.tab == tab) {
                elem->el_Type = TYPE_NIL;
	    }
            break;

        case TYPE_TABLE:
            if (elem->el_Contents.tab.tab == tab) {
                elem->el_Type = TYPE_NIL;
	    }
            break;

        case TYPE_FIELD:
            if (elem->el_Contents.field.rap->rap_Relative == tab ||
                    elem->el_Contents.field.rap->rap_Table == tab) {
                elem->el_Type = TYPE_NIL;
	    }
            break;
    }
}

void ElementRemTable(struct Element *elem, struct Table *tab)
{
    if (elem) {
        elementRemTableNonZero(elem, tab);
    }
}

static void elementRemFieldNonZero(struct Element *elem, 
        struct Field *field)
{
    struct RelFieldPath *rap;
    struct Field **pa, **end;

    while (elem->el_Type == TYPE_LIST) {
        elementRemFieldNonZero(elem->el_Contents.list.head, field);
        elem = elem->el_Contents.list.tail;
    }

    switch (elem->el_Type) {
        case TYPE_STRING:
            if (elem->el_Contents.string.field == field) {
                elem->el_Contents.string.field = NULL;
            }
            break;

        case TYPE_FIELD:
            rap = elem->el_Contents.field.rap;
            end = &rap->rap_Fields[(int)rap->rap_NumFields];
            for (pa = rap->rap_Fields; pa < end; pa++) {
                if (*pa == field) {
                    *pa = NULL;
                }
            }
            break;
    }
}

void ElementRemField(struct Element *elem, struct Field *field)
{
    if (elem) {
        elementRemFieldNonZero(elem, field);
    }
}

void ElementRemAllRecords(struct Element *elem) {
    while (elem->el_Type == TYPE_LIST) {
        ElementRemAllRecords(elem->el_Contents.list.head);
        elem->el_Record = NULL;
        elem = elem->el_Contents.list.tail;
    }

    elem->el_Record = NULL;

    switch (elem->el_Type) {
        case TYPE_RECORD:
            elem->el_Type = TYPE_NIL;
            break;

        case TYPE_FIELD:
            elem->el_Contents.field.rec = NULL;
            break;

        case TYPE_SPECIAL:
            if (elem->el_Contents.special.func) {
                elem->el_Contents.special.func(
                    elem->el_Contents.special.ptr, ELEM_SPECIAL_REMALLRECORDS,
                    NULL);
	    }
            break;
    }
}

static void elementRemRecordNonZero(struct Element *elem, struct Record *rec)
{
    while (elem->el_Type == TYPE_LIST) {
        elementRemRecordNonZero(elem->el_Contents.list.head, rec);
        if (elem->el_Record == rec) {
            elem->el_Record = NULL;
        }
        elem = elem->el_Contents.list.tail;
    }

    if (elem->el_Record == rec) {
	elem->el_Record = NULL;
    }

    switch (elem->el_Type) {
        case TYPE_RECORD:
            if (elem->el_Contents.tabrec.rec == rec) {
                elem->el_Type = TYPE_NIL;
	    }
            break;

        case TYPE_FIELD:
            if (elem->el_Contents.field.rec == rec) {
                elem->el_Contents.field.rec = NULL;
	    }
            break;

        case TYPE_SPECIAL:
            if (elem->el_Contents.special.func) {
                elem->el_Contents.special.func(
                    elem->el_Contents.special.ptr, ELEM_SPECIAL_REMRECORD, rec);
	    }
            break;
    }
}

void ElementRemRecord(struct Element *elem, struct Record *rec)
{
    if (elem) {
        elementRemRecordNonZero(elem, rec);
    }
}

void HeapRemRecord(struct Record *rec)
{
    struct Element *elem;
    long i;

    if (HeapValid && rec != NULL) {
        for (i=0; (elem = StackElement(i)) != NULL; i++) {
            elementRemRecordNonZero(elem, rec);
        }
    }
}

void HeapRemAllRecords(void)
{
    struct Element *elem;
    long i;

    if (HeapValid) {
        for (i=0; (elem = StackElement(i)) != NULL; i++) {
            ElementRemAllRecords(elem);
        }
    }
}

void HeapTick(void)
{
    if (Heap.heap_AllocCount >= Heap.heap_CheckCount) {
        GarbageCollection();
    }
    if (BreakCheck()) {
        ExecuteAbort(EXECUTE_ABORT);
    }
}

