/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** heap.h
**
** Creation date: 21.05.1995
** Last modified: 30.10.2014
*/

#ifndef HEAP_H
#define HEAP_H

#include "program.h"
#include "record.h"

struct HeapElement;

struct GarbageCollectionBase
{
    struct HeapElement *gcb_Heap;
    long gcb_Stack;
};

extern struct Element ElementNil;
extern struct Element ElementTrue;

void InitHeap(void);
/*
** must be called at start of program execution (before first AllocElement()).
*/

void ExitHeap(void);
/*
** must be called after program execution
*/

void FreeElementContents(struct MemPool *memPool, struct Element *elem);
/*
** frees all resources associated with the element but not the element itself.
** elem must not be NULL.
*/

#define MakeNil() &ElementNil
#define MakeTrue() &ElementTrue

extern struct Element *MakeElement(char type, bool allocated);
extern struct Element *MakeInteger(long i);
extern struct Element *MakeReal(double r);
extern struct Element *MakeString(char *s, bool allocated);
extern struct Element *MakeMemo(char *mi, bool allocated);
extern struct Element *MakeDate(long d);
extern struct Element *MakeTime(long t);
extern struct Element *MakeRecord(struct Table *tab, struct Record *rec);
extern struct Element *MakeTable(struct Table *tab, bool star);
extern struct Element *MakeField(struct RelFieldPath *, struct Record *, 
        bool alloc);
extern struct Element *MakeList(struct Element *head, struct Element *tail);
extern struct Element *MakeFile(short num, ulong id);
extern struct Element *MakeCodeAddress(const char *buffer, long index);
extern struct Element *MakeSpecial(void *p, 
        void (*f)(void *, short, struct Record *));
extern struct Element *MakeFunction(short type, void *ptr);
/*
** Constructor functions for different types.  Returns pointer to newly
** allocated element.  Never returns NULL (does longjmp on allocation failure)
*/

void ElementRemTable(struct Element *elem, struct Table *tab);
/*
** Removes all references to the specified table.
*/

void ElementRemField(struct Element *elem, struct Field *field);
/*
** Removes all references to the specified field.
*/

void ElementRemRecord(struct Element *elem, struct Record *rec);
/*
** Removes all references to the specified record.
*/

void ElementRemAllRecords(struct Element *elem);
/*
** Removes all references to all records.
*/

void HeapRemRecord(struct Record *rec);
/*
** Removes all references to the specified record in all elements reachable.
*/

void HeapRemAllRecords(void);
/*
** Removes all references to all records in all elements reachable.
*/

void GarbageCollection(void);
/*
** Directly invokes garbage collection. Must be called at a safe place.
*/

void GetGarbageCollectionBase(struct GarbageCollectionBase *gcb);
/*
** Gets the currently used base for garbage collection.
*/

void SetGarbageCollectionBase(const struct GarbageCollectionBase *gcb);
/*
** Sets the base for garbage collection.  All elements below the
** base are not examined for garbage collection.
*/

void SetGarbageCollectionBaseCurrent(void);
/*
** Uses the current heap size and stack size as new base for garbage collection.
** It is wise to Call GarbageCollection first for cleaning up.
*/

void HeapTick(void);
/*
** Must be called from time to time at a safe place (e.g. between
** function calls) to invoke the Garbage collector.
*/

#endif

