/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** parser.y
**
** Creation date: 15.04.1995
** Last modified: 22.03.2022
*/

%{
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <setjmp.h>
#include <unistd.h>
#include "system.h"
#include "dbtypes.h"
#include "field.h"
#include "util.h"
#include "memory.h"
#include "strhandle.h"
#include "appstrings_cat.h"
#include "requesters.h"
#include "lang/parser.h"
#include "lang/compile_impl.h"
#include "lang/functable.h"

static jmp_buf errJmpBuf;
static char errString[101];

static struct MemPool *memPool = NULL;
static struct Start *Start = NULL;

void yyerror(const char *s)
{
    CodesetStrNBytesCopy(errString, s, sizeof(errString) - 1);
}

#undef YY_FATAL_ERROR
#define YY_FATAL_ERROR(msg) \
        { \
	    yyerror(msg); \
            longjmp(errJmpBuf, 1); \
        }

static __inline void *_yy_malloc(size_t size)
{
    void *ptr = PoolAllocMem(memPool, size + sizeof(size_t));

    if (ptr)
    {
        *(size_t *)ptr = size + sizeof(size_t);
        ptr = (size_t *)ptr + 1;
    }

    return(ptr);
}

static __inline void *_yy_realloc(void *ptr, size_t newsize)
{
    size_t oldsize;

    if(ptr == NULL)
        return(_yy_malloc(newsize));

    ptr = (size_t *)ptr - 1;
    oldsize = *(size_t *)ptr;
    newsize += sizeof(size_t);
    ptr = PoolReAllocMem(memPool, ptr, oldsize, newsize);
    if(ptr == NULL)
        return(NULL);

    *(size_t *)ptr = newsize;
    ptr = (size_t *)ptr + 1;
    return(ptr);
}

static __inline void _yy_free(void *ptr)
{
    if(ptr != NULL)
    {
        ptr = (size_t *)ptr - 1;
        PoolFreeMem(memPool, ptr, *(size_t *)ptr);
    }
}

static void *alloc_and_test(long size)
{
    void *ret = PoolAllocMemZero(memPool, size);

    if(ret == NULL)
        YY_FATAL_ERROR(GCS(MSG_Err_Mem));
    return(ret);
}

#define NEW(x) alloc_and_test(sizeof(x))

struct Where
{
    struct Ignore *Ignore1;
    struct Expr *Expr;
    struct Ignore *Ignore2;
};

struct Order
{
    struct Ignore *Ignore1;
    struct Ignore *Ignore2;
    struct OrderList *OrderList;
};

struct SelectMode
{
    bool Distinct;
    struct Ignore *Ignore;
};

#define APPEND(x, y, type) \
{ \
    type **tmp = &x; \
\
    while(*tmp != NULL) \
        tmp = &(*tmp)->Next; \
    *tmp = y; \
}

extern int yylex(void);

%}

%start start

%union
{
    struct Start *Start;
    struct Expr *Expr;
    struct App *App;
    struct Func *Func;
    struct AppArgs *AppArgs;
    struct Const *Const;
    struct GlobalIdent *GlobalIdent;
    struct LocalIdent *LocalIdent;
    struct LocalExtend *LocalExtend;
    struct SelectFromWhere *SelectFromWhere;
    struct SelectMode *SelectMode;
    struct SelectList *SelectList;
    struct FromList *FromList;
    struct Where *Where;
    struct Order *Order;
    struct OrderList *OrderList;
    struct Let *Let;
    struct LetList *LetList;
    struct Cond *Cond;
    struct CondList *CondList;
    struct Case *Case;
    struct CaseList *CaseList;
    struct Do *Do;
    struct DoBindList *DoBindList;
    struct DoSpecial *DoSpecial;
    short DoSpecialCase;
    struct DefList *DefList;
    struct FuncDef *FuncDef;
    struct FuncArg *FuncArg;
    bool DefunTrigger;
    struct VarDef *VarDef;
    bool VarDefKeep;
    struct LocalIdentDef *LocalIdentDef;
    struct TypeSpec *TypeSpec;
    struct Ignore *Ignore;
    struct
    {
        struct LineCol lc;
        union
        {
            double real;
            struct
            {
                long value;
                char format;
            } integer;
            long date;
            long time;
            char *string;
            char *identifier;
            struct FuncInfo *funcinfo;
            struct ConstInfo *constinfo;
            char *ignore;               /* comments, white space, newlines, etc. */
            struct
            {
                short type;
                char *tabname;
            } type;
        } data;
    } scan;
}

%token <scan> INTEGER_CONST
%token <scan> REAL_CONST
%token <scan> STRING_CONST
%token <scan> DATE_CONST
%token <scan> TIME_CONST
%token <scan> GLOBAL_IDENT LOCAL_IDENT TYPE_SPEC
%token <scan> '(' ')' FUNC_INFO CONST_INFO
%token <scan> SELECT_TOKEN DISTINCT_TOKEN FROM_TOKEN WHERE_TOKEN
%token <scan> ORDER_TOKEN BY_TOKEN ASC_TOKEN DESC_TOKEN
%token <scan> LET_TOKEN FOR_TOKEN ALL_TOKEN DEFUN_TOKEN DEFUNSTAR_TOKEN
%token <scan> DEFVAR_TOKEN DEFVARSTAR_TOKEN
%token <scan> COND_TOKEN CASE_TOKEN DO_TOKEN DOTIMES_TOKEN DOLIST_TOKEN
%token <scan.data.ignore> IGNORE_TOKEN

%type <Start> start
%type <Expr> expr
%type <App> app
%type <Func> func
%type <AppArgs> appargs
%type <Const> const
%type <GlobalIdent> globalident
%type <LocalIdent> localidentcheck
%type <LocalIdent> localident
%type <LocalExtend> localextend
%type <SelectFromWhere> selectfromwhere
%type <SelectMode> selectmode
%type <SelectList> exprlist expritem
%type <FromList> fromlist fromitem fromident
%type <Where> where
%type <Order> order
%type <OrderList> orderlist orderitem
%type <Let> let
%type <LetList> letlist letitem
%type <Cond> cond
%type <CondList> condlist
%type <Case> case
%type <CaseList> caselist
%type <Do> do
%type <DoSpecial> dospecial
%type <DoSpecialCase> dospecialcase
%type <DoBindList> dobindlist dobinditem
%type <DefList> deflist defitem
%type <FuncDef> funcdef
%type <FuncArg> funcarglist
%type <DefunTrigger> defuntrigger
%type <VarDef> vardef
%type <VarDefKeep> vardefkeep
%type <LocalIdentDef> localidentdef
%type <TypeSpec> typespec
%type <Ignore> _d_

%%

start: _d_ expr _d_
        {
	    $$ = Start = NEW(struct Start);
            $$->Case = START_CASE_EXPR;
            $$->Ignore1 = $1;
	    $$->Contents.Expr = $2;
            $$->Ignore2 = $3;
        }
    | _d_ deflist
        {
	    $$ = Start = NEW(struct Start);
            $$->Case = START_CASE_DEFLIST;
            $$->Ignore1 = $1;
	    $$->Contents.DefList = $2;
        }
    | _d_
        {
            $$ = Start = NEW(struct Start);
            $$->Case = START_CASE_DEFLIST;
            $$->Ignore1 = $1;
        }
    ;

deflist: deflist defitem _d_
        {
            $2->Ignore1 = $3;
            APPEND($1, $2, struct DefList);
            $$ = $1;
        }
    |   defitem _d_                             /* avoids shift/reduce conflict */
        {
            $1->Ignore1 = $2;
            $$ = $1;
        }
    ;

defitem:  funcdef
        {
            $$ = NEW(struct DefList);
            $$->LineCol = $1->LineCol;
            $$->Case = DEFLIST_CASE_FUNCDEF;
            $$->Contents.FuncDef = $1;
        }
    |   vardef
        {
            $$ = NEW(struct DefList);
            $$->LineCol = $1->LineCol;
            $$->Case = DEFLIST_CASE_VARDEF;
            $$->Contents.VarDef = $1;
        }
    ;

funcdef: '(' _d_ defuntrigger _d_ localidentcheck _d_ '(' _d_ funcarglist ')' _d_ appargs ')'
        {
            struct FuncArg *fa;
            short num = 0;

            $$ = NEW(struct FuncDef);
            $$->LineCol = $1.lc;
            $$->Ignore1 = $2;
            $$->Trigger = $3;
            $$->Ignore2 = $4;
            $$->FuncName = $5;
            $$->Ignore3 = $6;
            $$->Ignore4 = $8;
            $$->FuncArgList = $9;
            $$->Ignore5 = $11;
            $$->AppArgs = $12;

            for(fa = $$->FuncArgList; fa != NULL; fa = fa->Next)
                num++;
            if (num > 0)
                $$->FuncName->Types = alloc_and_test(num * sizeof(char));
        }
    ;

defuntrigger:      DEFUN_TOKEN          { $$ = TRUE; }
    |           DEFUNSTAR_TOKEN         { $$ = FALSE; }
    ;

funcarglist: funcarglist localidentdef
        {
            $$ = NEW(struct FuncArg);
            $$->LocalIdentDef = $2;
            APPEND($1, $$, struct FuncArg);
            $$ = $1;
        }
    |   { $$ = NULL; }
    ;

vardef: '(' _d_ vardefkeep _d_ localidentdef expr _d_ ')'
        {
            $$ = NEW(struct VarDef);
            $$->LineCol = $1.lc;
            $$->Ignore1 = $2;
            $$->Keep = $3;
            $$->Ignore2 = $4;
            $$->LocalIdentDef = $5;
            $$->Expr = $6;
            $$->Ignore3 = $7;
        }
    | '(' _d_ vardefkeep _d_ localidentdef ')'
        {
            $$ = NEW(struct VarDef);
            $$->LineCol = $1.lc;
            $$->Ignore1 = $2;
            $$->Keep = $3;
            $$->Ignore2 = $4;
            $$->LocalIdentDef = $5;
        }
    ;

vardefkeep:     DEFVAR_TOKEN            { $$ = FALSE; }
    |           DEFVARSTAR_TOKEN        { $$ = TRUE; }
    ;

expr:   app
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LineCol;
            $$->Case = EXPR_CASE_APP;
            $$->Contents.App = $1;
        }
    |   selectfromwhere
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LineCol;
            $$->Case = EXPR_CASE_SELECTFROMWHERE;
            $$->Contents.SelectFromWhere = $1;
        }
    |   let
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LineCol;
            $$->Case = EXPR_CASE_LET;
            $$->Contents.Let = $1;
        }
    |   cond
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LineCol;
            $$->Case = EXPR_CASE_COND;
            $$->Contents.Cond = $1;
        }
    |   case
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LineCol;
            $$->Case = EXPR_CASE_CASE;
            $$->Contents.Case = $1;
        }
    |   do
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LineCol;
            $$->Case = EXPR_CASE_DO;
            $$->Contents.Do = $1;
        }
    |   dospecial
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LineCol;
            $$->Case = EXPR_CASE_DOSPECIAL;
            $$->Contents.DoSpecial = $1;
        }
    |   const
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LineCol;
            $$->Case = EXPR_CASE_CONST;
            $$->Contents.Const = $1;
        }
    |   globalident
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LineCol;
            $$->Case = EXPR_CASE_GLOBALIDENT;
            $$->Contents.GlobalIdent = $1;
        }
    |   localident
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LineCol;
            $$->Case = EXPR_CASE_LOCALIDENT;
            $$->Contents.LocalIdent = $1;
        }
    |   localextend
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LocalIdent->LineCol;
            $$->Case = EXPR_CASE_LOCALEXTEND;
            $$->Contents.LocalExtend = $1;
        }
    |   func
        {
            $$ = NEW(struct Expr);
            $$->LineCol = $1->LineCol;
            $$->Case = EXPR_CASE_FUNC;
            $$->Contents.Func = $1;
        }
    ;

app:    '(' _d_ appargs ')'
        {
            $$ = NEW(struct App);
            $$->LineCol = $1.lc;
            $$->Ignore = $2;
            $$->AppArgs = $3;
        }
    ;

appargs: appargs expr _d_
        {
            $$ = NEW(struct AppArgs);
            $$->Expr = $2;
            $$->Ignore = $3;
            APPEND($1, $$, struct AppArgs);
            $$ = $1;
        }
    |   { $$ = NULL; }
    ;

func:   FUNC_INFO
        {
            $$ = NEW(struct Func);
            $$->LineCol = $1.lc;
            $$->FuncInfo = $1.data.funcinfo;
        }
    |   ASC_TOKEN
        {
            $$ = NEW(struct Func);
            $$->LineCol = $1.lc;
            $$->FuncInfo = FindFuncInfoId(FUNCID_ASC);
        }

    ;

const:  INTEGER_CONST
        {
            $$ =  NEW(struct Const);
            $$->LineCol = $1.lc;
            $$->Case = CONST_CASE_NORMAL;
            $$->Type = TYPE_INTEGER;
            $$->Contents.integer.value = $1.data.integer.value;
            $$->Contents.integer.format = $1.data.integer.format;
        }
    |   REAL_CONST
        {
            $$ =  NEW(struct Const);
            $$->LineCol = $1.lc;
            $$->Case = CONST_CASE_NORMAL;
            $$->Type = TYPE_REAL;
            $$->Contents.real = $1.data.real;
        }
    |   STRING_CONST
        {
            $$ =  NEW(struct Const);
            $$->LineCol = $1.lc;
            $$->Case = CONST_CASE_NORMAL;
            $$->Type = TYPE_STRING;
            $$->Contents.string = $1.data.string;
        }
    |   DATE_CONST
        {
            $$ =  NEW(struct Const);
            $$->LineCol = $1.lc;
            $$->Case = CONST_CASE_NORMAL;
            $$->Type = TYPE_DATE;
            $$->Contents.date = $1.data.date;
        }
    |   TIME_CONST
        {
            $$ =  NEW(struct Const);
            $$->LineCol = $1.lc;
            $$->Case = CONST_CASE_NORMAL;
            $$->Type = TYPE_TIME;
            $$->Contents.time = $1.data.time;
        }
    |   CONST_INFO
        {
            $$ =  NEW(struct Const);
            $$->LineCol = $1.lc;
            $$->Case = CONST_CASE_PREDEFINED;
            $$->Type = $1.data.constinfo->ci_Element.el_Type;
            $$->Contents.ConstInfo = $1.data.constinfo;
        }
    ;

localextend:    localident '.' globalident
                {
                    $$ = NEW(struct LocalExtend);
                    $$->LocalIdent = $1;
                    $$->GlobalIdent = $3;
                }
    ;

globalident:    GLOBAL_IDENT
                {
                    $$ = NEW(struct GlobalIdent);
                    $$->LineCol = $1.lc;
                    $$->Name = $1.data.identifier;
                }
    ;

localidentcheck: localident
                {
                    $$ = $1;
		}
    |   	GLOBAL_IDENT
		{
		    YY_FATAL_ERROR(GCS(MSG_CompErr_LowercaseIdentifier));
		}
    ;

localident:     LOCAL_IDENT
                {
                    $$ = NEW(struct LocalIdent);
                    $$->LineCol = $1.lc;
                    $$->Name = $1.data.identifier;
                    $$->Type = TYPE_UNKNOWN;
                    $$->Arity = -1;
                }
    ;

selectfromwhere: '(' _d_ SELECT_TOKEN _d_ selectmode exprlist FROM_TOKEN _d_ fromlist where order ')'
        {
            $$ = NEW(struct SelectFromWhere);
            $$->LineCol = $1.lc;
            $$->IsSelectFromWhere = TRUE;
            $$->Ignore1 = $2;
            $$->Ignore2 = $4;
            $$->Distinct = $5->Distinct;
            $$->Ignore3 = $5->Ignore;
            $$->SelectList = $6;
            $$->Ignore4 = $8;
            $$->FromList = $9;
            if($10)
            {
                $$->Ignore5 = $10->Ignore1;
                $$->Expr = $10->Expr;
                $$->Ignore6 = $10->Ignore2;
            }
            if($11)
            {
                $$->Ignore7 = $11->Ignore1;
                $$->Ignore8 = $11->Ignore2;
                $$->OrderList = $11->OrderList;
            }
        }
    |   '(' _d_ FOR_TOKEN _d_ ALL_TOKEN _d_ fromlist where order DO_TOKEN _d_ appargs ')'
        {
            $$ = NEW(struct SelectFromWhere);
            $$->LineCol = $1.lc;
            $$->IsSelectFromWhere = FALSE;
            $$->Ignore1 = $2;
            $$->Ignore2 = $4;
            $$->Ignore4 = $6;
            $$->FromList = $7;
            if($8)
            {
                $$->Ignore5 = $8->Ignore1;
                $$->Expr = $8->Expr;
                $$->Ignore6 = $8->Ignore2;
            }
            if($9)
            {
                $$->Ignore7 = $9->Ignore1;
                $$->Ignore8 = $9->Ignore2;
                $$->OrderList = $9->OrderList;
            }
            $$->Ignore9 = $11;
            $$->AppArgs = $12;
        }
    ;

selectmode: DISTINCT_TOKEN _d_
            {
                $$ = NEW(struct SelectMode);
                $$->Distinct = TRUE;
                $$->Ignore = $2;
            }
    |
            {
                $$ = NEW(struct SelectMode);
                $$->Distinct = FALSE;
            }
    ;

exprlist:       expritem
                {
                    $$ = $1;
                }
    |           expritem ',' _d_ exprlist
                {
                    $$ = $1;
                    $$->Ignore3 = $3;
                    $$->Next = $4;
                }
    ;

expritem:       expr _d_
                {
                    $$ = NEW(struct SelectList);
                    $$->Expr = $1;
                    $$->Ignore1 = $2;
                }
    |           expr _d_ STRING_CONST _d_
                {
                    $$ = NEW(struct SelectList);
                    $$->Expr = $1;
                    $$->Ignore1 = $2;
                    $$->Title = $3.data.string;
                    $$->Ignore2 = $4;
                }
    ;

fromlist:       fromitem
                {
                    $$ = $1;
                }
    |           fromitem ',' _d_ fromlist
                {
                    $$ = $1;
                    $$->Ignore2 = $3;
                    $$->Next = $4;
                }
    ;

fromitem:       fromident _d_
                {
                    $$ = $1;
                    $$->Ignore1 = $2;
                }
    |           fromident _d_ localidentcheck _d_
                {
                    $$ = $1;
                    $$->Ignore1 = $2;
                    $$->LocalIdentDef = NEW(struct LocalIdentDef);
                    $$->LocalIdentDef->LocalIdent = $3;
                    $$->LocalIdentDef->Ignore = $4;
                }
    ;

fromident:      globalident
                {
                    $$ = NEW(struct FromList);
                    $$->GlobalIdent = $1;
                }
    ;

where:  WHERE_TOKEN _d_ expr _d_
        {
            $$ = NEW(struct Where);
            $$->Ignore1 = $2;
            $$->Expr = $3;
            $$->Ignore2 = $4;
        }
    |   { $$ = NULL; }
    ;

order:  ORDER_TOKEN _d_ BY_TOKEN _d_ orderlist
        {
            $$ = NEW(struct Order);
            $$->Ignore1 = $2;
            $$->Ignore2 = $4;
            $$->OrderList = $5;
        }
    |   { $$ = NULL; }
    ;

orderlist:  orderitem
            {
                $$ = $1;
            }
    |       orderitem ',' _d_ orderlist
            {
                $$ = $1;
                $$->Ignore3 = $3;
                $$->Next = $4;
            }
    ;

orderitem:  expr _d_
            {
                $$ = NEW(struct OrderList);
                $$->Expr = $1;
                $$->FieldNum = 0;
                $$->OrderMode = ORDER_DEFAULT;
                $$->Ignore2 = $2;
            }
    |       expr _d_ ASC_TOKEN _d_
            {
                $$ = NEW(struct OrderList);
                $$->Expr = $1;
                $$->FieldNum = 0;
                $$->Ignore1 = $2;
                $$->OrderMode = ORDER_ASC;
                $$->Ignore2 = $4;
            }
    |       expr _d_ DESC_TOKEN _d_
            {
                $$ = NEW(struct OrderList);
                $$->Expr = $1;
                $$->FieldNum = 0;
                $$->Ignore1 = $2;
                $$->OrderMode = ORDER_DESC;
                $$->Ignore2 = $4;
            }
    ;

let:    '(' _d_ LET_TOKEN _d_ '(' _d_ letlist ')' _d_ appargs ')'
        {
            $$ = NEW(struct Let);
            $$->LineCol = $1.lc;
            $$->Ignore1 = $2;
            $$->Ignore2 = $4;
            $$->Ignore3 = $6;
            $$->LetList = $7;
            $$->Ignore4 = $9;
            $$->AppArgs = $10;
        }
    ;

letlist:    letitem
            {
                $$ = $1;
            }
    |       letlist letitem
            {
                APPEND($1, $2, struct LetList);
                $$ = $1;
            }
    ;

letitem:    localidentdef
            {
                $$ = NEW(struct LetList);
                $$->LocalIdentDef = $1;
            }
    |       '(' _d_ localidentdef expr _d_ ')' _d_
            {
                $$ = NEW(struct LetList);
                $$->Ignore1 = $2;
                $$->LocalIdentDef = $3;
                $$->Expr = $4;
                $$->Ignore2 = $5;
                $$->Ignore3 = $7;
            }
    ;

cond:   '(' _d_ COND_TOKEN _d_ condlist ')'
        {
            $$ = NEW(struct Cond);
            $$->LineCol = $1.lc;
            $$->Ignore1 = $2;
            $$->Ignore2 = $4;
            $$->CondList = $5;
        }
    ;

condlist: condlist '(' _d_ expr _d_ appargs ')' _d_
        {
            $$ = NEW(struct CondList);
            $$->Ignore1 = $3;
            $$->Expr = $4;
            $$->Ignore2 = $5;
            $$->AppArgs = $6;
            $$->Ignore3 = $8;
            APPEND($1, $$, struct CondList);
            $$ = $1;
        }
    |
        {
            $$ = NULL;
        }
    ;

/*
        (case <expr> [<case>]...)  SELECT BY CASE
            <expr>      the selection expression
            <case>      pair consisting of:
                            (<value> [<expr>]...)
                          where:
                            <value>     is a single expression or a list of
                                        expressions (unevaluated)
                            <expr>      are expressions to execute if the
                                        case matches
            returns     the value of the last expression of the matching case

*/

case:   '(' _d_ CASE_TOKEN _d_ expr _d_ caselist ')'
        {
            $$ = NEW(struct Case);
            $$->LineCol = $1.lc;
            $$->Ignore1 = $2;
            $$->Ignore2 = $4;
            $$->Expr = $5;
            $$->Ignore3 = $6;
            $$->CaseList = $7;
        }
    ;

caselist: caselist '(' _d_ expr _d_ appargs ')' _d_
        {
            $$ = NEW(struct CaseList);
            $$->Ignore1 = $3;
            $$->Expr = $4;
            $$->Ignore2 = $5;
            $$->AppArgs = $6;
            $$->Ignore3 = $8;
            APPEND($1, $$, struct CaseList);
            $$ = $1;
        }
    |   { $$ = NULL; }
    ;

/*
        (do ([<binding>]...) (<texpr> [<rexpr>]...) [<expr>]...)
        (do* ([<binding>]...) (<texpr> [<rexpr>]...) [<expr>]...)
            <binding>   the variable bindings each of which is either:
                        1)  a symbol (which is initialized to nil)
                        2)  a list of the form: (<sym> <init> [<step>])
                            where:
                                <sym>  is the symbol to bind
                                <init> is the initial value of the symbol
                                <step> is a step expression
            <texpr>     the termination test expression
            <rexpr>     result expressions (the default is nil)
            <expr>      the body of the loop (treated like an implicit prog)
            returns     the value of the last result expression
*/

do:     '(' _d_ DO_TOKEN _d_ '(' _d_ dobindlist ')' _d_ '(' _d_ expr _d_ appargs ')' _d_ appargs ')'
        {
            $$ = NEW(struct Do);
            $$->LineCol = $1.lc;
            $$->Ignore1 = $2;
            $$->Ignore2 = $4;
            $$->Ignore3 = $6;
            $$->DoBindList = $7;
            $$->Ignore4 = $9;
            $$->Ignore5 = $11;
            $$->TermExpr = $12;
            $$->Ignore6 = $13;
            $$->ResultAppArgs = $14;
            $$->Ignore7 = $16;
            $$->LoopAppArgs = $17;
        }
        ;

dobindlist: dobindlist dobinditem
            {
                APPEND($1, $2, struct DoBindList);
                $$ = $1;
            }
        |   { $$ = NULL; }
        ;

dobinditem: localidentdef
            {
                $$ = NEW(struct DoBindList);
                $$->LocalIdentDef = $1;
            }
    |       '(' _d_ localidentdef expr _d_ ')' _d_
            {
                $$ = NEW(struct DoBindList);
                $$->Ignore1 = $2;
                $$->LocalIdentDef = $3;
                $$->InitExpr = $4;
                $$->Ignore2 = $5;
                $$->Ignore4 = $7;
            }
    |       '(' _d_ localidentdef expr _d_ expr _d_ ')' _d_
            {
                $$ = NEW(struct DoBindList);
                $$->Ignore1 = $2;
                $$->LocalIdentDef = $3;
                $$->InitExpr = $4;
                $$->Ignore2 = $5;
                $$->StepExpr = $6;
                $$->Ignore3 = $7;
                $$->Ignore4 = $9;
            }
    ;

/*
        (dotimes (<sym> <expr> [<rexpr>]) [<expr>]...)  LOOP FROM ZERO TO N-1
            <sym>       the symbol to bind to each value from 0 to n-1
            <expr>      the number of times to loop
            <rexpr>     the result expression (the default is nil)
            <expr>      the body of the loop (treated like an implicit prog)
        (dolist (<sym> <expr> [<rexpr>]) [<expr>]...)  LOOP THROUGH A LIST
            <sym>       the symbol to bind to each list element
            <expr>      the list expression
            <rexpr>     the result expression (the default is nil)
            <expr>      the body of the loop (treated like an implicit prog)
*/

dospecial:  '(' _d_ dospecialcase _d_ '(' _d_ localidentdef expr _d_ appargs ')' _d_ appargs ')'
        {
            $$ = NEW(struct DoSpecial);
            $$->LineCol = $1.lc;
            $$->Ignore1 = $2;
            $$->Case = $3;
            $$->Ignore2 = $4;
            $$->Ignore3 = $6;
            $$->LocalIdentDef = $7;
            $$->Expr = $8;
            $$->Ignore4 = $9;
            $$->ResultAppArgs = $10;
            $$->Ignore5 = $12;
            $$->LoopAppArgs = $13;
        }
    ;

dospecialcase:  DOTIMES_TOKEN
        {
            $$ = DO_CASE_TIMES;
        }
    |   DOLIST_TOKEN
        {
            $$ = DO_CASE_LIST;
        }
    ;

localidentdef:  localidentcheck _d_ typespec
        {
            $$ = NEW(struct LocalIdentDef);
            $$->LocalIdent = $1;
            $$->Ignore = $2;
            $$->TypeSpec = $3;
        }
    ;

typespec:   TYPE_SPEC _d_
        {
            $$ = NEW(struct TypeSpec);
            $$->Type = $1.data.type.type;
            $$->GlobalIdent = NEW(struct GlobalIdent);
            $$->GlobalIdent->LineCol = $1.lc;
            $$->GlobalIdent->Name = $1.data.type.tabname;
            $$->Ignore = $2;
        }
    |   { $$ = NULL; }
    ;

_d_: _d_ IGNORE_TOKEN
        {
            $$ = NEW(struct Ignore);
            $$->String = $2;
            APPEND($1, $$, struct Ignore);
            $$ = $1;
        }
   |    { $$ = NULL; }
   ;

%%

#define malloc(size) _yy_malloc(size)
#define realloc(ptr, size) _yy_realloc(ptr, size)
#define free(ptr) _yy_free(ptr)

#include "lex.yy.c"

#undef malloc
#undef realloc
#undef free

const char *MakeSynTree(struct MemPool *mp, const char *src, long size,
        struct Start **start, long *errline, long *errcol,
        const char **errfilename)
{
    int ret;

    /* Supress some compiler warnings about unused variables */
    (void)yy_fatal_error;
    (void)input;

    if(start)
        *start = NULL;

    if(src == NULL || size == 0)             /* nothing to compile? */
        return(NULL);

    scanSrc = src;
    scanSize = size;
    scanFilename = NULL;
    memPool = mp;
    Start = NULL;
    ret = setjmp(errJmpBuf);
    if(ret == 0)
    {
        yy_init_globals();
        ret = yyparse();
    }
    memPool = NULL;

    if(ret != 0)
    {
        if(errline)
            *errline = scanLine;
        if(errcol)
            *errcol = scanColumn;
        if(errfilename)
            *errfilename = scanFilename;
        return(errString);
    }

    if(start)
        *start = Start;
    return(NULL);
}

