/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** program.c
**
** Creation date: 17.04.1995
** Last modified: 26.03.2022
*/

#include "system.h"
#include <setjmp.h>
#include "status.h"
#include "project.h"
#include "preferences.h"
#include "table.h"
#include "field.h"
#include "display.h"
#include "dbtypes.h"
#include "program.h"
#include "programbin.h"
#include "compile_impl.h"
#include "heap.h"
#include "functable.h"
#include "memo.h"
#include "labels.h"
#include "memory.h"
#include "strhandle.h"
#include "requesters.h"
#include "textfile.h"
#include "dosutil.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"

#define SRC_BLOCK_SIZE 2048
#define SRC_RAPBLOCK_SIZE 64

const char onOpen[] = "onOpen";
const char onReload[] = "onReload";
const char onClose[] = "onClose";
const char onChange[] = "onChange";
const char onAdminMode[] = "onAdminMode";
const char mainWindowTitle[] = "mainWindowTitle";
const char logLabel[] = "logLabel";

enum SourceEncoding
{
    SE_Codeset = 0,
    SE_UTF8,
    SE_8BIT
};

struct Source
{
    struct MemPool *src_MemPool;
    long src_Size;              /* buffer size */
    long src_Len;               /* buffer valid contents size */
    char *src_Buffer;
    long src_RapSize;
    long src_RapLen;
    long *src_RapBuffer;
    enum SourceEncoding src_Encoding;
};

enum
{
    SRC_IGNORE,
    SRC_OPEN,
    SRC_CLOSE,
    SRC_STAR,
    SRC_COMMA,
    SRC_DOT,
    SRC_CONSTINTEGER,
    SRC_CONSTREAL,
    SRC_CONSTSTRING,
    SRC_CONSTDATE,
    SRC_CONSTTIME,
    SRC_TRUE,
    SRC_NIL,
    SRC_SELECT,
    SRC_FROM,
    SRC_WHERE,
    SRC_GLOBALIDENT,
    SRC_LOCALIDENT,
    SRC_FUNC,
    SRC_ORDER,
    SRC_BY,
    SRC_ASC,
    SRC_DESC,
    SRC_DISTINCT,
    SRC_LET,
    SRC_FOR,
    SRC_ALL,
    SRC_DO,
    SRC_DEFUN,
    SRC_COND,
    SRC_DOTIMES,
    SRC_DOLIST,
    SRC_CASE,
    SRC_DEFVAR,
    SRC_CONST,
    SRC_T_INT,
    SRC_T_REAL,
    SRC_T_STR,
    SRC_T_MEMO,
    SRC_T_DATE,
    SRC_T_TIME,
    SRC_T_LIST,
    SRC_T_FILE,
    SRC_T_FUNC,
    SRC_T_REC,
    SRC_CONSTINTEGER2,
    SRC_END = 127,
};

static jmp_buf errJmpBuf;
static struct Source *source = NULL;
static long sourceIndex = 0;
static ulong ignoreOutput = 0;
static struct Project *sourceProject = NULL;

static void srcout(const void *buf, long size)
{
    if(ignoreOutput > 0)
        return;

    if(source->src_Len + size > source->src_Size)       /* overflow? */
    {
        long newsize = source->src_Size + MAX(SRC_BLOCK_SIZE, size);
        char *new;

        new = PoolReAllocMem(source->src_MemPool, source->src_Buffer, source->src_Size, newsize);
        if(new == NULL)
            longjmp(errJmpBuf, 1);                 /* return to main function */
        source->src_Buffer = new;
        source->src_Size = newsize;
    }
    memcpy(source->src_Buffer + source->src_Len, buf, size);
    source->src_Len += size;
}

static __inline void srcoutchar(char c) { srcout(&c, sizeof(char)); }

#define srcouttag srcoutchar

static void srcoutshort(short x)
{ 
    char buf[2]; 
    SPutShort(buf, x); 
    srcout(buf, 2); 
}

static void srcoutlong(long x)
{
    char buf[4];
    SPutLong(buf, x);
    srcout(buf, 4);
}

static void srcoutdouble(double x)
{
    char buf[8]; 
    SPutDouble(buf, x); 
    srcout(buf, 8); 
}

static void srcoutstring(const char *s)
{
    if (s) {
        /* We always use CS_Codeset encoding here */
        srcout(s, strlen(s));
    }
}

static void mkrapentry(void)
{
    if(ignoreOutput > 0)
        return;

    if(source->src_RapLen >= source->src_RapSize)       /* overflow? */
    {
        long newsize = source->src_RapSize + SRC_RAPBLOCK_SIZE;
        long *new;

        new = PoolReAllocMem(source->src_MemPool, source->src_RapBuffer,
                source->src_RapSize * sizeof(long), newsize * sizeof(long));
        if(new == NULL)
            longjmp(errJmpBuf, 1);                 /* return to main function */
        source->src_RapBuffer = new;
        source->src_RapSize = newsize;
    }
    source->src_RapBuffer[source->src_RapLen++] = source->src_Len;
}

static void srcin(void *buf, long size)
{
    if(sourceIndex + size > source->src_Len)
        longjmp(errJmpBuf, 1);                 /* return to main function */
    memcpy(buf, source->src_Buffer + sourceIndex, size);
    sourceIndex += size;
}

static const char *srcstr(void)
{
    const char *s = source->src_Buffer + sourceIndex;

    sourceIndex += strlen(s) + 1;
    return s;
}

static char srcinchar(void) 
{
    char c; 
    srcin(&c, sizeof(char)); 
    return c; 
}

#define srcintag srcinchar

static short srcinshort(void)
{ 
    char buf[2];

    srcin(buf, 2);
    return SGetShort(buf);
}

static long srcinlong(void)
{ 
    char buf[4];

    srcin(buf, 4);
    return SGetLong(buf);
}

static double srcindouble(void)
{ 
    char buf[8];

    srcin(buf, 8);
    return SGetDouble(buf);
}

static void ignoreToSource(struct Ignore *ignore)
{
    if(ignore)
    {
        srcouttag(SRC_IGNORE);
        while(ignore)
        {
            if(!strcmp(ignore->String, IGNORE_REPLACE_START))
                ignoreOutput++;
            srcoutstring(ignore->String);
            if(!strcmp(ignore->String, IGNORE_REPLACE_END))
                ignoreOutput--;
            ignore = ignore->Next;
        }
        srcoutchar(0);
    }
}

static void constToSource(struct Const *c)
{
    switch(c->Case)
    {
        case CONST_CASE_NORMAL:
            switch(c->Type)
            {
                case TYPE_INTEGER:
                    srcouttag(SRC_CONSTINTEGER2);
                    srcoutlong(c->Contents.integer.value);
                    srcoutchar(c->Contents.integer.format);
                    break;

                case TYPE_REAL:
                    srcouttag(SRC_CONSTREAL);
                    srcoutdouble(c->Contents.real);
                    break;

                case TYPE_STRING:
                    srcouttag(SRC_CONSTSTRING);
                    srcoutstring(c->Contents.string);
                    srcoutchar(0);
                    break;

                case TYPE_DATE:
                    srcouttag(SRC_CONSTDATE);
                    srcoutlong(c->Contents.date);
                    break;

                case TYPE_TIME:
                    srcouttag(SRC_CONSTTIME);
                    srcoutlong(c->Contents.time);
                    break;
            }
            break;

        case CONST_CASE_PREDEFINED:
            switch(c->Type)
            {
                case TYPE_BOOL:
                    srcouttag(SRC_TRUE);
                    break;

                case TYPE_NIL:
                    srcouttag(SRC_NIL);
                    break;

                default:
                    srcouttag(SRC_CONST);
                    srcoutshort(c->Contents.ConstInfo->ci_Id);
                    break;
            }
            break;
    }
}

static void globalIdentToSource(struct GlobalIdent *gi)
{
    if(gi)
    {
        srcouttag(SRC_GLOBALIDENT);
        mkrapentry();
        srcout(&gi->RelFieldPath, sizeof(struct RelFieldPath));
    }
}

static void localIdentToSource(struct LocalIdent *li)
{
    if(li)
    {
        srcouttag(SRC_LOCALIDENT);
        srcoutstring(li->Name);
        srcoutchar(0);
    }
}

static void exprToSource(struct Expr *expr);
static void appArgsToSource(struct AppArgs *aa)
{
    while(aa)
    {
        exprToSource(aa->Expr);
        ignoreToSource(aa->Ignore);
        aa = aa->Next;
    }
}

static void funcToSource(struct Func *func)
{
    if(func)
    {
        srcouttag(SRC_FUNC);
        srcoutshort(func->FuncInfo->fi_Id);
    }
}

static void appToSource(struct App *app)
{
    if(app)
    {
        srcouttag(SRC_OPEN);
        ignoreToSource(app->Ignore);
        appArgsToSource(app->AppArgs);
        srcouttag(SRC_CLOSE);
    }
}

static void typeSpecToSource(struct TypeSpec *typeSpec)
{
    if(typeSpec)
    {
        switch(typeSpec->Type)
        {
            case TYPE_INTEGER:  srcouttag(SRC_T_INT);   break;
            case TYPE_REAL:     srcouttag(SRC_T_REAL);  break;
            case TYPE_STRING:   srcouttag(SRC_T_STR); break;
            case TYPE_MEMO:     srcouttag(SRC_T_MEMO);  break;
            case TYPE_DATE:     srcouttag(SRC_T_DATE);  break;
            case TYPE_TIME:     srcouttag(SRC_T_TIME);  break;
            case TYPE_LIST:     srcouttag(SRC_T_LIST);  break;
            case TYPE_FILE:     srcouttag(SRC_T_FILE);  break;
            case TYPE_FUNCTION: srcouttag(SRC_T_FUNC);  break;
            case TYPE_RECORD:   srcouttag(SRC_T_REC);
                                globalIdentToSource(typeSpec->GlobalIdent);
                                break;
        }
        ignoreToSource(typeSpec->Ignore);
    }
}

static void localIdentDefToSource(struct LocalIdentDef *lid)
{
    if(lid)
    {
        localIdentToSource(lid->LocalIdent);
        ignoreToSource(lid->Ignore);
        typeSpecToSource(lid->TypeSpec);
    }
}

static void fromListToSource(struct FromList *fl)
{
    for(; fl; fl = fl->Next)
    {
        globalIdentToSource(fl->GlobalIdent);
        ignoreToSource(fl->Ignore1);
        localIdentDefToSource(fl->LocalIdentDef);
        if(fl->Next)
            srcouttag(SRC_COMMA);
        ignoreToSource(fl->Ignore2);
    }
}

static void selectListToSource(struct SelectList *sl)
{
    for(; sl; sl = sl->Next)
    {
        if(sl->Expr == NULL)
        {
            srcouttag(SRC_STAR);
        }
        else
            exprToSource(sl->Expr);
        ignoreToSource(sl->Ignore1);
        if(sl->Title)
        {
            srcouttag(SRC_CONSTSTRING);
            srcoutstring(sl->Title);
            srcoutchar(0);
        }
        ignoreToSource(sl->Ignore2);
        if(sl->Next)
            srcouttag(SRC_COMMA);
        ignoreToSource(sl->Ignore3);
    }
}

static void orderListToSource(struct OrderList *ol)
{
    for(; ol; ol = ol->Next)
    {
        if(ol->FieldNum > 0)                    /* integer constant? */
        {
            srcouttag(SRC_CONSTINTEGER);
            srcoutlong(ol->FieldNum);
        }
        else
            exprToSource(ol->Expr);
        ignoreToSource(ol->Ignore1);
        switch(ol->OrderMode)
        {
            case ORDER_ASC:     srcouttag(SRC_ASC);
                                break;

            case ORDER_DESC:    srcouttag(SRC_DESC);
                                break;
        }
        ignoreToSource(ol->Ignore2);
        if(ol->Next)
            srcouttag(SRC_COMMA);
        ignoreToSource(ol->Ignore3);
    }
}

static void sfwToSource(struct SelectFromWhere *sfw)
{
    if(sfw)
    {
        srcouttag(SRC_OPEN);
        ignoreToSource(sfw->Ignore1);
        if(sfw->IsSelectFromWhere == FALSE)
        {
            srcouttag(SRC_FOR);
        }
        else
        {
            srcouttag(SRC_SELECT);
        }
        ignoreToSource(sfw->Ignore2);
        if(sfw->IsSelectFromWhere == FALSE)
        {
            srcouttag(SRC_ALL);
        }
        else if(sfw->Distinct)
        {
            srcouttag(SRC_DISTINCT);
        }
        ignoreToSource(sfw->Ignore3);
        if(sfw->IsSelectFromWhere)
        {
            selectListToSource(sfw->SelectList);
            srcouttag(SRC_FROM);
        }
        ignoreToSource(sfw->Ignore4);
        fromListToSource(sfw->FromList);
        if(sfw->Expr)
        {
            srcouttag(SRC_WHERE);
            ignoreToSource(sfw->Ignore5);
            exprToSource(sfw->Expr);
            ignoreToSource(sfw->Ignore6);
        }
        if(sfw->OrderList)
        {
            srcouttag(SRC_ORDER);
            ignoreToSource(sfw->Ignore7);
            srcouttag(SRC_BY);
            ignoreToSource(sfw->Ignore8);
            orderListToSource(sfw->OrderList);
        }
        if(sfw->IsSelectFromWhere == FALSE)
        {
            srcouttag(SRC_DO);
            ignoreToSource(sfw->Ignore9);
            appArgsToSource(sfw->AppArgs);
        }
        srcouttag(SRC_CLOSE);
    }
}

static void letListToSource(struct LetList *ll)
{
    for(; ll; ll = ll->Next)
    {
        if(ll->Expr)
            srcouttag(SRC_OPEN);
        ignoreToSource(ll->Ignore1);
        localIdentDefToSource(ll->LocalIdentDef);
        exprToSource(ll->Expr);
        ignoreToSource(ll->Ignore2);
        if(ll->Expr)
            srcouttag(SRC_CLOSE);
        ignoreToSource(ll->Ignore3);
    }
}

static void letToSource(struct Let *let)
{
    if(let)
    {
        srcouttag(SRC_OPEN);
        ignoreToSource(let->Ignore1);
        srcouttag(SRC_LET);
        ignoreToSource(let->Ignore2);
        srcouttag(SRC_OPEN);
        ignoreToSource(let->Ignore3);
        letListToSource(let->LetList);
        srcouttag(SRC_CLOSE);
        ignoreToSource(let->Ignore4);
        appArgsToSource(let->AppArgs);
        srcouttag(SRC_CLOSE);
    }
}

static void condListToSource(struct CondList *cl)
{
    for(; cl; cl = cl->Next)
    {
        srcouttag(SRC_OPEN);
        ignoreToSource(cl->Ignore1);
        exprToSource(cl->Expr);
        ignoreToSource(cl->Ignore2);
        appArgsToSource(cl->AppArgs);
        srcouttag(SRC_CLOSE);
        ignoreToSource(cl->Ignore3);
    }
}

static void condToSource(struct Cond *cond)
{
    if(cond)
    {
        srcouttag(SRC_OPEN);
        ignoreToSource(cond->Ignore1);
        srcouttag(SRC_COND);
        ignoreToSource(cond->Ignore2);
        condListToSource(cond->CondList);
        srcouttag(SRC_CLOSE);
    }
}

static void caseListToSource(struct CaseList *cl)
{
    for(; cl; cl = cl->Next)
    {
        srcouttag(SRC_OPEN);
        ignoreToSource(cl->Ignore1);
        exprToSource(cl->Expr);
        ignoreToSource(cl->Ignore2);
        appArgsToSource(cl->AppArgs);
        srcouttag(SRC_CLOSE);
        ignoreToSource(cl->Ignore3);
    }
}

static void caseToSource(struct Case *c)
{
    if(c)
    {
        srcouttag(SRC_OPEN);
        ignoreToSource(c->Ignore1);
        srcouttag(SRC_CASE);
        ignoreToSource(c->Ignore2);
        exprToSource(c->Expr);
        ignoreToSource(c->Ignore3);
        caseListToSource(c->CaseList);
        srcouttag(SRC_CLOSE);
    }
}

static void doBindListToSource(struct DoBindList *dbl)
{
    for(; dbl; dbl = dbl->Next)
    {
        if(dbl->InitExpr)
            srcouttag(SRC_OPEN);
        ignoreToSource(dbl->Ignore1);
        localIdentDefToSource(dbl->LocalIdentDef);
        exprToSource(dbl->InitExpr);
        ignoreToSource(dbl->Ignore2);
        exprToSource(dbl->StepExpr);
        ignoreToSource(dbl->Ignore3);
        if(dbl->InitExpr)
            srcouttag(SRC_CLOSE);
        ignoreToSource(dbl->Ignore4);
    }
}

static void doToSource(struct Do *d)
{
    if(d)
    {
        srcouttag(SRC_OPEN);
        ignoreToSource(d->Ignore1);
        srcouttag(SRC_DO);
        ignoreToSource(d->Ignore2);
        srcouttag(SRC_OPEN);
        ignoreToSource(d->Ignore3);
        doBindListToSource(d->DoBindList);
        srcouttag(SRC_CLOSE);
        ignoreToSource(d->Ignore4);
        srcouttag(SRC_OPEN);
        ignoreToSource(d->Ignore5);
        exprToSource(d->TermExpr);
        ignoreToSource(d->Ignore6);
        appArgsToSource(d->ResultAppArgs);
        srcouttag(SRC_CLOSE);
        ignoreToSource(d->Ignore7);
        appArgsToSource(d->LoopAppArgs);
        srcouttag(SRC_CLOSE);
    }
}

static void doSpecialToSource(struct DoSpecial *ds)
{
    if(ds)
    {
        srcouttag(SRC_OPEN);
        ignoreToSource(ds->Ignore1);
        switch(ds->Case)
        {
            case DO_CASE_TIMES: srcouttag(SRC_DOTIMES); break;
            case DO_CASE_LIST:  srcouttag(SRC_DOLIST);  break;
        }
        ignoreToSource(ds->Ignore2);
        srcouttag(SRC_OPEN);
        ignoreToSource(ds->Ignore3);
        localIdentDefToSource(ds->LocalIdentDef);
        exprToSource(ds->Expr);
        ignoreToSource(ds->Ignore4);
        appArgsToSource(ds->ResultAppArgs);
        srcouttag(SRC_CLOSE);
        ignoreToSource(ds->Ignore5);
        appArgsToSource(ds->LoopAppArgs);
        srcouttag(SRC_CLOSE);
    }
}

static void localExtendToSource(struct LocalExtend *le)
{
    if(le)
    {
        localIdentToSource(le->LocalIdent);
        srcouttag(SRC_DOT);
        globalIdentToSource(le->GlobalIdent);
    }
}

static void exprToSource(struct Expr *expr)
{
    if(expr)
    {
        switch(expr->Case)
        {
            case EXPR_CASE_APP:
                appToSource(expr->Contents.App);
                break;

            case EXPR_CASE_SELECTFROMWHERE:
                sfwToSource(expr->Contents.SelectFromWhere);
                break;

            case EXPR_CASE_LET:
                letToSource(expr->Contents.Let);
                break;

            case EXPR_CASE_COND:
                condToSource(expr->Contents.Cond);
                break;

            case EXPR_CASE_CASE:
                caseToSource(expr->Contents.Case);
                break;

            case EXPR_CASE_DO:
                doToSource(expr->Contents.Do);
                break;

            case EXPR_CASE_DOSPECIAL:
                doSpecialToSource(expr->Contents.DoSpecial);
                break;

            case EXPR_CASE_CONST:
                constToSource(expr->Contents.Const);
                break;

            case EXPR_CASE_GLOBALIDENT:
                globalIdentToSource(expr->Contents.GlobalIdent);
                break;

            case EXPR_CASE_LOCALIDENT:
                localIdentToSource(expr->Contents.LocalIdent);
                break;

            case EXPR_CASE_LOCALEXTEND:
                localExtendToSource(expr->Contents.LocalExtend);
                break;

            case EXPR_CASE_FUNC:
                funcToSource(expr->Contents.Func);
                break;
        }
    }
}

static void varDefToSource(struct VarDef *varDef)
{
    srcouttag(SRC_OPEN);
    ignoreToSource(varDef->Ignore1);
    srcouttag(SRC_DEFVAR);
    if(varDef->Keep)
        srcouttag(SRC_STAR);
    ignoreToSource(varDef->Ignore2);
    localIdentDefToSource(varDef->LocalIdentDef);
    if(varDef->Expr)
        exprToSource(varDef->Expr);
    ignoreToSource(varDef->Ignore3);
    srcouttag(SRC_CLOSE);
}

static void funcDefToSource(struct FuncDef *funcDef)
{
    struct FuncArg *fa;

    srcouttag(SRC_OPEN);
    ignoreToSource(funcDef->Ignore1);
    srcouttag(SRC_DEFUN);
    if(!funcDef->Trigger)
        srcouttag(SRC_STAR);
    ignoreToSource(funcDef->Ignore2);
    localIdentToSource(funcDef->FuncName);
    ignoreToSource(funcDef->Ignore3);
    srcouttag(SRC_OPEN);
    ignoreToSource(funcDef->Ignore4);
    for(fa = funcDef->FuncArgList; fa; fa = fa->Next)
        localIdentDefToSource(fa->LocalIdentDef);
    srcouttag(SRC_CLOSE);
    ignoreToSource(funcDef->Ignore5);
    appArgsToSource(funcDef->AppArgs);
    srcouttag(SRC_CLOSE);
}

static void defListToSource(struct DefList *defList)
{
    for(; defList; defList = defList->Next)
    {
        switch(defList->Case)
        {
            case DEFLIST_CASE_VARDEF:
                varDefToSource(defList->Contents.VarDef);
                break;

            case DEFLIST_CASE_FUNCDEF:
                funcDefToSource(defList->Contents.FuncDef);
                break;
        }
        ignoreToSource(defList->Ignore1);
    }
}

static void startToSource(struct Start *start)
{
    if(start)
    {
        ignoreToSource(start->Ignore1);
        switch(start->Case)
        {
            case START_CASE_EXPR:
                exprToSource(start->Contents.Expr);
                break;

            case START_CASE_DEFLIST:
                defListToSource(start->Contents.DefList);
                break;
        }
        ignoreToSource(start->Ignore2);
    }
    srcouttag(SRC_END);
}

static void FreeSource(struct Source *src)
{
    if(src)
    {
        PoolFreeMem(src->src_MemPool, src->src_Buffer, src->src_Size);
        PoolFreeMem(src->src_MemPool, src->src_RapBuffer, src->src_RapSize * sizeof(long));
        PoolFreeMem(src->src_MemPool, src, sizeof(struct Source));
    }
}

static struct Source *AllocSource(struct MemPool *memPool)
{
    struct Source *src;

    src = PoolAllocMemZero(memPool, sizeof(struct Source));
    if (src == NULL) {
        return NULL;
    }

    src->src_MemPool = memPool;

    return src;
}

static struct Source *StartToSource(struct Project *pr, struct Start *start)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct Source *src;
    int ret;

    src = AllocSource(memPool);
    if(src == NULL)
        return(NULL);

    source = src;
    sourceProject = pr;
    ret = setjmp(errJmpBuf);
    if(ret == 0)
    {
        ignoreOutput = 0;
        startToSource(start);
    }
    else
    {
        /*
        ** at this point an allocation failure has been occured.
        */
        FreeSource(src);
        src = NULL;
    }

    source = NULL;
    sourceProject = NULL;

    return(src);
}

struct Program
{
    struct Project *prg_Project;
    struct MemPool *prg_MemPool;
    struct Source *prg_Source;
    struct Bin *prg_Bin;
    bool prg_NeedRecompile;
};

void FreeProgram(struct Program *prg)
{
    if(prg)
    {
        FreeBin(prg->prg_Bin);
        FreeSource(prg->prg_Source);
        PoolFreeMem(prg->prg_MemPool, prg, sizeof(struct Program));
    }
}

bool ProgramNeedRecompile(struct Program *prg)
{
    bool recompile = FALSE;

    if (prg) {
	recompile = prg->prg_NeedRecompile;
    }
    return recompile;
}

void SetProgramNeedRecompile(struct Program *prg, bool recompile)
{
    if (prg) {
	prg->prg_NeedRecompile = recompile;
    }
}

static struct Program *AllocProgram(struct Project *pr)
{
    struct Program *prg = NULL;
    struct MemPool *memPool = ProjectMemPool(pr);

    prg = PoolAllocMemZero(memPool, sizeof(struct Program));
    if (prg == NULL) {
        return NULL;
    }

    prg->prg_Project = pr;
    prg->prg_MemPool = memPool;
    return prg;
}

struct Bin *ProgramBin(struct Program *prg)
{
    struct Bin *bin = NULL;

    if(prg)
        bin = prg->prg_Bin;
    return(bin);
}

struct Program *StartToProgram(struct Project *pr, struct Start *start,
        struct LocalIdent *identList, bool debug)
{
    struct Program *prg = NULL;

    prg = AllocProgram(pr);
    if (prg == NULL) {
        return NULL;
    }

    if (start == NULL) {
        return prg;                   /* ready */
    }

    prg->prg_Source = StartToSource(pr, start);
    if (prg->prg_Source == NULL) {
        FreeProgram(prg);
        return NULL;
    }

    prg->prg_Bin = StartToBin(pr, start, identList, debug);
    if (prg->prg_Bin == NULL) {
        FreeProgram(prg);
        return NULL;
    }
    prg->prg_NeedRecompile = FALSE;

    return prg;
}

#define ST_BLOCK_SIZE 2048

struct SourceText
{
    struct MemPool *st_MemPool;
    char *st_Text;
    long st_Size;               /* buffer size */
    long st_Len;                /* actual size */
    long st_Error;
    size_t st_Warnings;
};

static struct SourceText *sourceText = NULL;

static void stout(const void *buf, long size)
{
    if(sourceText->st_Len + size > sourceText->st_Size)       /* overflow? */
    {
        long newsize = sourceText->st_Size + MAX(ST_BLOCK_SIZE, size);
        char *new;

        new = PoolReAllocMem(sourceText->st_MemPool, sourceText->st_Text,
                sourceText->st_Size, newsize);
        if(new == NULL)
            longjmp(errJmpBuf, 1);                 /* return to main function */
        sourceText->st_Text = new;
        sourceText->st_Size = newsize;
    }
    memcpy(sourceText->st_Text + sourceText->st_Len, buf, size);
    sourceText->st_Len += size;
}

static void __inline stoutchar(char x) { stout(&x, sizeof(char)); }

static void stoutstring(const char *s)
{ 
    if(s) {
	struct MemPool *memPool = sourceText->st_MemPool;
	long error = OK;
	size_t warnings = 0;
	char *buf;

        switch (source->src_Encoding) {
            case SE_Codeset:
            default:
                buf = (char *)s;
                break;

            case SE_UTF8:
                buf = CodesetConvertFromUTF8(s, memPool, &error, &warnings);
                sourceText->st_Warnings += warnings;
                break;

            case SE_8BIT:
                buf = CodesetConvertFromDefault8Bit(s, memPool, &error);
                break;
	}
	if (buf == NULL) {
            longjmp(errJmpBuf, 1);                 /* return to main function */
	}
	stout(buf, strlen(buf)); 
	if (buf != s) {
	    FreeString(memPool, buf);
	}
    }
}

void FreeSourceText(struct SourceText *st)
{
    if(st)
    {
        PoolFreeMem(st->st_MemPool, st->st_Text, st->st_Size);
        PoolFreeMem(st->st_MemPool, st, sizeof(struct SourceText));
    }
}

const char *SourceTextText(struct SourceText *st)
{
    const char *text = NULL;

    if (st) {
        text = st->st_Text;
    }
    return text;
}

size_t SourceTextCodesetWarnings(struct SourceText *st)
{
    size_t warnings = 0;

    if (st) {
        warnings = st->st_Warnings;
    }
    return warnings;
}

static void stignore(void)
{
    stoutstring(srcstr());
}

static void stinteger(void)
{
    stoutstring(IntToStr(srcinlong()));
}

static void stinteger2(void)
{
    char str[32];
    long val = srcinlong();
    signed char format = srcinchar();

    if (format < 0) {
        sprintf(str, "0%lo", val);
    } else if (format == 0) {
        sprintf(str, "%ld", val);
    } else {
        sprintf(str, "0x%0*lX", format, val);
    }
    stoutstring(str);
}

static void streal(void)
{
    stoutstring(DoubleToStr(srcindouble(), -1, FORMAT_DECIMALPOINT));
}

static void ststring(void)
{
    stoutchar('\"');
    stoutstring(srcstr());
    stoutchar('\"');
}

static void stdate(void)
{
    stoutstring(DateToString(srcinlong(),
		(short)GetPrefsValue(sourceProject, PREFS_DATEFORMAT)));
}

static void sttime(void)
{
    stoutstring(TimeToString(srcinlong(), FORMAT_HOURMINSEC));
}

static void stglobalident(void)
{
    struct RelFieldPath rap;

    srcin(&rap, sizeof(struct RelFieldPath));
    stoutstring(RelFieldPathToPathName(&rap));
}

static void stlocalident(void)
{
    stoutstring(srcstr());
}

static void stfunc(void)
{
    short funcid = srcinshort();
    struct FuncInfo *fi;

    fi = FindFuncInfoId(funcid);
    if(fi == NULL)
        stoutstring(GCS(MSG_Err_Internal));
    else
        stoutstring(fi->fi_Name);
}

static void stconst(void)
{
    short constid = srcinshort();
    struct ConstInfo *ci;

    ci = FindConstInfoId(constid);
    if(ci == NULL)
        stoutstring(GCS(MSG_Err_Internal));
    else
        stoutstring(ci->ci_Name);
}

struct SourceText *GetSourceText(const struct Program *prg)
{
    struct SourceText *st = NULL;
    struct Source *src = prg? prg->prg_Source : NULL;
    int ret;

    if(src == NULL)
        return(NULL);

    st = PoolAllocMemZero(src->src_MemPool, sizeof(struct SourceText));
    if(st == NULL)
        return(NULL);

    st->st_MemPool = src->src_MemPool;

    sourceText = st;
    source = src;
    sourceIndex = 0;

    ret = setjmp(errJmpBuf);
    if(ret == 0)
    {
        char tag;

        do
        {
            tag = srcintag();
            switch(tag)
            {
                case SRC_IGNORE:             stignore();     break;
                case SRC_OPEN:               stoutchar('('); break;
                case SRC_CLOSE:              stoutchar(')'); break;
                case SRC_STAR:               stoutchar('*'); break;
                case SRC_COMMA:              stoutchar(','); break;
                case SRC_DOT:                stoutchar('.'); break;
                case SRC_CONSTINTEGER:       stinteger();    break;
                case SRC_CONSTINTEGER2:      stinteger2();   break;
                case SRC_CONSTREAL:          streal();       break;
                case SRC_CONSTSTRING:        ststring();     break;
                case SRC_CONSTDATE:          stdate();       break;
                case SRC_CONSTTIME:          sttime();       break;
                case SRC_TRUE:               stoutstring("TRUE"); break;
                case SRC_NIL:                stoutstring("NIL");  break;
                case SRC_SELECT:             stoutstring("SELECT");     break;
                case SRC_DISTINCT:           stoutstring("DISTINCT");   break;
                case SRC_FROM:               stoutstring("FROM");       break;
                case SRC_WHERE:              stoutstring("WHERE");      break;
                case SRC_ORDER:              stoutstring("ORDER");      break;
                case SRC_BY:                 stoutstring("BY");      break;
                case SRC_ASC:                stoutstring("ASC");     break;
                case SRC_DESC:               stoutstring("DESC");    break;
                case SRC_FOR:                stoutstring("FOR");     break;
                case SRC_ALL:                stoutstring("ALL");     break;
                case SRC_DO:                 stoutstring("DO");      break;
                case SRC_LET:                stoutstring("LET");     break;
                case SRC_DEFUN:              stoutstring("DEFUN");   break;
                case SRC_DEFVAR:             stoutstring("DEFVAR");  break;
                case SRC_GLOBALIDENT:        stglobalident();        break;
                case SRC_LOCALIDENT:         stlocalident();         break;
                case SRC_FUNC:               stfunc();               break;
                case SRC_CONST:              stconst();              break;
                case SRC_COND:               stoutstring("COND");    break;
                case SRC_CASE:               stoutstring("CASE");    break;
                case SRC_DOTIMES:            stoutstring("DOTIMES"); break;
                case SRC_DOLIST:             stoutstring("DOLIST");  break;
                case SRC_T_INT:              stoutstring(":INT"); break;
                case SRC_T_REAL:             stoutstring(":REAL"); break;
                case SRC_T_STR:              stoutstring(":STR"); break;
                case SRC_T_MEMO:             stoutstring(":MEMO"); break;
                case SRC_T_DATE:             stoutstring(":DATE"); break;
                case SRC_T_TIME:             stoutstring(":TIME"); break;
                case SRC_T_LIST:             stoutstring(":LIST"); break;
                case SRC_T_FILE:             stoutstring(":FILE"); break;
                case SRC_T_FUNC:             stoutstring(":FUNC"); break;
                case SRC_T_REC:              stoutstring(":"); break;
                case SRC_END:                break;
                default:                     st->st_Error = DISKERR_INVALID;
                                             tag = SRC_END;
                                             break;
            }
        } while(tag != SRC_END);

        stoutchar(0);           /* end of string */

        return(st);
    }
    /*
    ** at this point an allocation failure has occured.
    */
    FreeSourceText(st);
    return(NULL);
}

void SetTableProgramRecords(struct Project *pr)
{
    struct List *tabs = ProjectTables(pr);
    struct Table *tab;

    if(tabs)
        FORNODE(tab, tabs)
            SetTableProgramRecord(tab, TableGuiRecord(tab));
}

void ProgramBegin(struct Program *prg, bool needDefVars)
{
    struct Project *pr = prg? prg->prg_Project : NULL;
    struct Bin *bin = prg? prg->prg_Bin : NULL;

    BinBegin(pr, bin, needDefVars);
}
  
struct Element *ProgramCall(struct Program *prg, struct UserFunction *uf,
        struct Element **args, bool allowSideEffect)
{
    struct Bin *bin = prg? prg->prg_Bin : NULL;

    return BinCall(bin, uf, args, allowSideEffect);
}

void ProgramEnd(void)
{
    BinEnd();
}

const char *ProgramError(long *line, long *col, const char **filename)
{
    return BinError(line, col, filename);
}

const char *ExecuteProgram(struct Program *prg, bool needDefVars,
		     struct UserFunction *uf, struct Element **args,
		     bool allowSideEffect,
		     const char *(*func)(struct Element *elem),
		     long *line, long *col, const char **filename)
{
    struct Project *pr = prg? prg->prg_Project : NULL;
    struct Bin *bin = prg? prg->prg_Bin : NULL;

    return ExecuteBin(pr, bin, needDefVars, uf, args, allowSideEffect,
            func, line, col, filename);
}

struct List *ProgramUserFunctions(struct Program *prg)
{
    return BinUserFunctions(prg? prg->prg_Bin : NULL);
}

struct UserFunction *FindUserFunctionName(struct Program *prg, const char *name)
{
    return BinFindUserFunctionName(prg? prg->prg_Bin : NULL, name);
}

struct List *ProgramDisplayDependenciesVirtualField(struct Program *prg,
        struct DisplayObject *dio)
{
    return BinDisplayDependenciesVirtualField(prg? prg->prg_Bin : NULL, dio);
}

bool ProgramFindObject(struct Program *prg, void *ptr)
{
    struct Source *src = prg? prg->prg_Source : NULL;
    struct RelFieldPath rap;
    long i, j;

    if(src == NULL || ptr == NULL)
        return(FALSE);

    for(i=0; i < src->src_RapLen; i++)
    {
        memcpy(&rap, src->src_Buffer + src->src_RapBuffer[i], sizeof(struct RelFieldPath));
        if(rap.rap_Relative == ptr)
            return(TRUE);
        if(rap.rap_Table == ptr)
            return(TRUE);
        for(j=0; j < rap.rap_NumFields; j++)
            if(rap.rap_Fields[j] == ptr)
                return(TRUE);
    }

    return(FALSE);
}

void ProgramRemTable(struct Table *tab)
{
    struct Project *pr = TableProject(tab);
    struct Program *prg = ProjectProgram(pr);
    struct Bin *bin = ProgramBin(prg);

    ProgramBinRemTable(bin, tab);
}

void ProgramAddField(struct Field *field)
{
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    struct Program *prg = ProjectProgram(pr);

    if (prg &&  ProgramFindObject(prg, tab)) {
	prg->prg_NeedRecompile = TRUE;	/* might have SELECT * FROM ... */
    }
}

void ProgramRemField(struct Field *field)
{
    struct Table *tab = FieldTable(field);
    struct Project *pr = TableProject(tab);
    struct Program *prg = ProjectProgram(pr);
    struct Bin *bin = ProgramBin(prg);

    if (prg && ProgramBinRemField(bin, field)) {
	prg->prg_NeedRecompile = TRUE;
    }
}

void ProgramRemDisplayObject(struct DisplayObject *dio)
{
    struct Project *pr = DisplayObjectProject(dio);
    struct Program *prg = ProjectProgram(pr);
    struct Bin *bin = ProgramBin(prg);

    if (prg && ProgramBinRemDisplayObject(bin, dio)) {
	prg->prg_NeedRecompile = TRUE;
    }
}

void ProgramRemRecord(struct Table *tab, struct Record *rec)
{
    struct Project *pr = TableProject(tab);
    struct Program *prg = ProjectProgram(pr);
    struct Bin *bin = ProgramBin(prg);

    HeapRemRecord(rec);
    ProgramBinRemRecord(bin, rec);
}

void ProgramRemAllRecords(struct Project *pr)
{
    struct Program *prg = ProjectProgram(pr);
    struct Bin *bin = ProgramBin(prg);

    HeapRemAllRecords();
    ProgramBinRemAllRecords(bin);
}

char *SimpleElementToString(const struct Element *el, 
	struct Field *field, struct MemPool *memPool, bool *allocated)
{
    if(allocated != NULL)
        *allocated = FALSE;

    if(el == NULL)
        return (char *)GCS(MSG_NullElement);

    switch(el->el_Type)
    {
        case TYPE_INTEGER:
        {
            if (field && FieldType(field) == TYPE_CHOICE)
	    {
                char **labels = FieldChoiceLabels(field);
		if ((ulong)el->el_Contents.integer < NumLabels(labels))
		    return labels[el->el_Contents.integer];
	    }

            if(allocated != NULL)
                *allocated = TRUE;
            return(AllocString(memPool, IntToStr(el->el_Contents.integer)));
            break;
        }

        case TYPE_REAL:
        {
            short numdecimals = 2;
            struct DisplayObject *dio = FieldDisplayObject(field);

            if(dio)
                numdecimals = DisplayFieldNumDecimals(dio);

            if(allocated != NULL)
                *allocated = TRUE;
            return(AllocString(memPool,
                DoubleToStr(el->el_Contents.real, numdecimals,
                    (short)GetPrefsValue(executeProject, PREFS_REALFORMAT))));
            break;
        }

        case TYPE_STRING:
            if(allocated != NULL)
                *allocated = TRUE;
            return(AllocString(memPool, el->el_Contents.string.string));
            break;

        case TYPE_MEMO:
            if(allocated != NULL)
                *allocated = TRUE;
            return(AllocString(memPool, el->el_Contents.mi));
            break;

        case TYPE_DATE:
            if(allocated != NULL)
                *allocated = TRUE;
            return(AllocString(memPool,
                DateToString(el->el_Contents.date,
                    (short)GetPrefsValue(executeProject, PREFS_DATEFORMAT))));
            break;

        case TYPE_TIME:
            if(allocated != NULL)
                *allocated = TRUE;
            return(AllocString(memPool, TimeToString(el->el_Contents.time,
			DisplayFieldFormat(FieldDisplayObject(field)))));

            break;

        case TYPE_RECORD:
            if(allocated != NULL)
                *allocated = TRUE;
            return(AllocString(memPool, IntToStr(RecordNum(el->el_Contents.tabrec.rec))));
            break;

        case TYPE_BOOL:
            return("TRUE");
            break;

        case TYPE_NIL:
        {
            const char *nil = FieldNilString(field);

            return (char *)(nil? nil : "NIL");
            break;
        }

        default:
        {
            char buffer[21];

            sprintf(buffer, "%p", el);
            if(allocated != NULL)
                *allocated = TRUE;
            return(AllocString(memPool, buffer));
            break;
        }
    }
}

static long depth(struct Element *elem)
{
    long max = 0, d;

    while(elem->el_Type == TYPE_LIST)
    {
	d = 1 + depth(elem->el_Contents.list.head);
	max = MAX(max, d);
	elem = elem->el_Contents.list.tail;
    }

    return max;
}
    
static bool elementToStreamHelp(struct Element *elem,
        bool (*out)(const char *), long indent)
{
    const char *s = NULL; 
    bool quote = FALSE;
    bool ok = TRUE;
    long d, i;
    char buf[ID_NAME_LEN + 30];

    switch(elem->el_Type)
    {
	case TYPE_INTEGER:  
    	    s = IntToStr(elem->el_Contents.integer);
    	    break;

	case TYPE_REAL: 
	    s = DoubleToStr(elem->el_Contents.real, -1, 
		       (short)GetPrefsValue(executeProject, PREFS_REALFORMAT));
	    break;
        
	case TYPE_STRING:  
	    s = elem->el_Contents.string.string;
	    quote = TRUE;
	    break;

	case TYPE_MEMO:     
	    s = elem->el_Contents.mi;
	    quote = TRUE;
	    break;
          
	case TYPE_DATE:     
	    s = DateToString(elem->el_Contents.date,
		     (short)GetPrefsValue(executeProject, PREFS_DATEFORMAT));
	    break;
    
	case TYPE_TIME:     
	    s = TimeToString(elem->el_Contents.time, FORMAT_HOURMINSEC);
	    break;
	
	case TYPE_BOOL:     
	    s = "True";
	    break;
        
	case TYPE_NIL:
	    s = "NIL";
	    break;

	case TYPE_FIELD:
	    s = RelFieldPathToPathName(elem->el_Contents.field.rap);
	    break;

	case TYPE_TABLE:
	    sprintf(buf, "%s%s", TableName(elem->el_Contents.tab.tab),
		    elem->el_Contents.tab.star? "*" : "");
	    s = buf;
	    break;

	case TYPE_RECORD:
	    sprintf(buf, "Record-%p", elem->el_Contents.tabrec.rec);
	    s = buf;
	    break;
        
	case TYPE_LIST:     
	    d = depth(elem);
	    ok = ok && out("( ");

	    for (;;) {
		elementToStreamHelp(elem->el_Contents.list.head, out, indent+2);
		elem = elem->el_Contents.list.tail;
		if (elem->el_Type != TYPE_LIST) {
		    break;
		}

		if (d == 1) {
		    ok = ok && out(" ");
		} else {
		    ok = ok && out("\n");
		    for (i=0; i < indent+2; i++) {
			ok = ok && out(" ");
		    }
		}
	    }
	    if (elem->el_Type != TYPE_NIL) {
		ok = ok && out(". ");
		elementToStreamHelp(elem, out, indent + 2);
	    }
	    ok = ok && out(" )");
	    break;

	case TYPE_FILE:     
	    sprintf(buf, "File-Handle-%hd", elem->el_Contents.file.num);
	    s = buf;
	    break;

	case TYPE_FUNCTION: 
	    sprintf(buf, "Function-%p", elem->el_Contents.function.ptr);
	    s = buf;
	    break;

	case NOTYPE:
	    s = "No-Type";
	    break;

	default:
	    sprintf(buf, "Unknown-Type-%d", elem->el_Type);
	    s = buf;
	    break;
    }

    if (s != NULL) {
	if (quote) {
            char outbuf[CodesetMaxBytesForChar+1];
            const char *t;
            cschar c;

	    ok = ok && out("\"");
            while ((c = CodesetGetChar(s)) != 0) {
                t = CodesetNextChar(s);
                if (c == '"') {
                    ok = ok && out("\\");
                }
                memcpy(outbuf, s, t-s);
                outbuf[t-s] = 0;
                out(outbuf);
                s = t;
            }
	    ok = ok && out("\"");
	} else {
            ok = ok && out(s);
        }
    }

    return TRUE;
}

bool ElementToStream(struct Element *elem, bool (*out)(const char *))
{
    if (out == NULL) {
	return FALSE;
    }

    return elementToStreamHelp(elem, out, 0);
}

const char *ProgramCompileProjectProgram(struct Project *pr, bool internal,
	struct Program **compiled, long *line, long *col, const char **filename)
{
    struct Program *prg = ProjectProgram(pr);
    bool external = (bool)GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL);
    const char *path = GetPrefsString(pr, PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);
    struct SourceText *st = NULL;
    struct MemPool *memPool = ProjectMemPool(pr);
    char *mem = NULL;
    const char *text = NULL;
    const char *error = NULL;
    long err;

    /* compile program */
    GuiSetSleep(TRUE);
    if (internal || !external) {
	if (prg) {
    	    st = GetSourceText(prg);
	    if (st == NULL) {
		error = GCS(MSG_Err_Mem);
	    }
	    text = SourceTextText(st);
	}
    } else if (path == NULL || path[0] == 0) {
	error = GCS(MSG_CompErr_CantOpenIncludeFile);
    } else {
        ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
        mem = TextFileGetContentsAskWarnings(NULL, memPool, path, &err);
	if (err == DISKERR_OPEN) {
	    error = GCS(MSG_CompErr_CantOpenIncludeFile);
        } else if (err) {
            error = GES(err);
        } else {
            text = mem;
	}
    }

    if (text != NULL) {
	error = CompileProjectProgram(pr, text, strlen(text),
		compiled, line, col, filename);
    }
    FreeSourceText(st);
    FreeString(memPool, mem);
    GuiSetSleep(FALSE);

    return error;
}

const char *ProgramCompileFilter(struct Table *tab, struct Program **compiled,
        long *line, long *col)
{
    struct Program *prg = TableFilterProgram(tab);
    struct SourceText *st;
    const char *text;
    const char *error = NULL;

    if (compiled != NULL) {
        *compiled = NULL;
    }

    if (prg == NULL) {
        return NULL;
    }

    st = GetSourceText(prg);
    if (st == NULL) {
        return GCS(MSG_Err_Mem);
    }

    text = SourceTextText(st);
    if (text != NULL && text[0] != 0) {
        error = CompileFilter(tab, text, strlen(text), compiled, line, col);
    }
    FreeSourceText(st);

    return error;
}

const char *ProgramCompileQuery(const struct Program *prg,
        struct Program **compiled, long *line, long *col)
{
    struct SourceText *st;
    const char *text;
    const char *error = NULL;
    long len;

    if (compiled != NULL) {
        *compiled = NULL;
    }

    if (prg == NULL) {
        return NULL;
    }

    st = GetSourceText(prg);
    if (st == NULL) {
        return GCS(MSG_Err_Mem);
    }

    text = SourceTextText(st);
    if (text != NULL && text[0] != 0) {
        len = strlen(text);
        if (len >= 3) {
            text++;         /* skip "("   */
            len -= 3;       /* skip "\n)" */
        }
        error = CompileQuery(prg->prg_Project, text, len, compiled, line, col);
    }
    FreeSourceText(st);

    return error;
}

const char *CreateSourceExternal(struct Project *pr,
	const char *text, long size)
{
    const char *path = GetPrefsString(pr, PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);
    struct MemPool *memPool = ProjectMemPool(pr);
    long err;

    if (path == NULL || path[0] == 0) {
	return GCS(MSG_Err_OpenFile);
    }

    ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
    TextFileSetContentsAskWarnings(NULL, memPool, path, text, &err);
    if (err) {
        if (err != DISKERR_OPEN) {
            ProgramDeleteSourceExternal(pr);
        }
        return GES(err);
    }

    return NULL;
}

const char *ProgramCreateSourceExternal(struct Project *pr)
{
    struct Program *prg = ProjectProgram(pr);
    struct SourceText *st = NULL;
    const char *text = NULL;
    const char *error = NULL;

    GuiSetSleep(TRUE);
    if (prg) {
	st = GetSourceText(prg);
	if (st == NULL) {
	    error = GCS(MSG_Err_Mem);
	} else {
	    text = SourceTextText(st);
	}
    }
    if (error == NULL) {
        text = EmptyNullStr(text);
	error = CreateSourceExternal(pr, text, strlen(text));
    }
    FreeSourceText(st);
    GuiSetSleep(FALSE);

    return error;
}


void ProgramDeleteSourceExternal(struct Project *pr)
{
    const char *path = GetPrefsString(pr, PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);

    if (path && path[0]) {
        ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
	RmFile(path);
    }
}

bool ProgramSourceExternalIdentical(struct Project *pr)
{
    const char *path = GetPrefsString(pr, PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);
    struct MemPool *memPool = ProjectMemPool(pr);
    char *contents = NULL;
    struct Program *prg = ProjectProgram(pr);
    struct SourceText *st = NULL;
    const char *text = NULL;
    bool identical = FALSE;

    if (!path || !path[0]) {
	return FALSE;
    }

    GuiSetSleep(TRUE);

    if (prg) {
	st = GetSourceText(prg);
	text = SourceTextText(st);
    }
    text = EmptyNullStr(text);
    ProjectSetCurrentDir(pr);    /* make sure we have correct CWD */
    contents = TextFileGetContents(memPool, path, NULL, NULL);
    if (contents) {
	if (strcmp(contents, text) == 0) {
	    identical = TRUE;
	} else {
	    struct Program *compiled = NULL;
	    const char *error = CompileProjectProgram(pr, 
		    contents, strlen(contents), &compiled, NULL, NULL, NULL);
	    if (error == NULL) {
		struct SourceText *stc = NULL;
		const char *textc = NULL;

		if (compiled) {
		    stc = GetSourceText(compiled);
		    textc = SourceTextText(stc);
		}
                textc = EmptyNullStr(textc);
		if (strcmp(textc, text) == 0) {
		    identical = TRUE;
		}
		FreeSourceText(stc);
	    }
	    FreeProgram(compiled);
	}
	FreeString(memPool, contents);
    }
    FreeSourceText(st);
    GuiSetSleep(FALSE);

    return identical;
}

/*
** load & save
*/

long ReadProgram(struct Project *pr, const char *text, struct Program **prgp)
{
    long len = strlen(EmptyNullStr(text));
    long size = 1 + len + 1 + 1;
    struct MemPool *memPool;
    struct Program *prg;
    struct Source *src;

    if (pr == NULL || prgp == NULL) {
        return ERR_INTERNAL;
    }

    *prgp = NULL;

    if (len == 0) {
        return OK;
    }

    prg = AllocProgram(pr);
    if (prg == NULL) {
        return ERR_MEM;
    }
    memPool = prg->prg_MemPool;
    prg->prg_Source = src = AllocSource(memPool);
    if (src == NULL) {
        FreeProgram(prg);
        return ERR_MEM;
    }
    src->src_Buffer = PoolAllocMem(memPool, size);
    if (src->src_Buffer == NULL) {
        FreeProgram(prg);
        return ERR_MEM;
    }
    src->src_Buffer[0] = SRC_IGNORE;
    strcpy(&src->src_Buffer[1], text);
    src->src_Buffer[size - 1] = SRC_END;

    src->src_Size = src->src_Len = size; 

    prg->prg_NeedRecompile = TRUE;
    *prgp = prg;

    return OK;
}

#define PROGVERSION 6
#define MIN_PROGVERSION 4

long LoadProgramCustomFormat(struct Project *pr, FILE *file,
        struct Program **prgp)
{
    struct Program *prg;
    struct Source *src;
    struct SourceText *st;
    struct MemPool *memPool;
    short version;
    bool utf8 = FALSE;
    long len, raplen, size;
    long val, sub=0, i, j;
    struct RelFieldPath rap;
    long error;
    size_t unused;

    if (pr == NULL || file == NULL || prgp == NULL) {
        return ERR_INTERNAL;
    }

    *prgp = NULL;

    version = FGetShort(file);
    if (version < MIN_PROGVERSION || version > PROGVERSION) {
        return DISKERR_VERSION;
    }
    if (version >= 5) {
	utf8 = FGetBool(file);
    }

    len = FGetLong(file);
    raplen = FGetLong(file);
    size = len + raplen * sizeof(struct RelFieldPath);

    if (size == 0) {          /* no source? */
        return OK;
    }

    prg = AllocProgram(pr);
    if (prg == NULL) {
        return ERR_MEM;
    }
    memPool = prg->prg_MemPool;
    prg->prg_Source = src = AllocSource(memPool);
    if (src == NULL) {
        FreeProgram(prg);
        return ERR_MEM;
    }
    src->src_Buffer = PoolAllocMem(memPool, size);
    if (src->src_Buffer == NULL) {
        FreeProgram(prg);
        return ERR_MEM;
    }
    src->src_Size = src->src_Len = size;
    src->src_Encoding = utf8? SE_UTF8 : SE_8BIT;

    if (raplen > 0) {
        src->src_RapBuffer = PoolAllocMem(memPool, raplen * sizeof(long));
        if (src->src_RapBuffer == NULL) {
            FreeProgram(prg);
            return ERR_MEM;
        }
        src->src_RapSize = src->src_RapLen = raplen;
    }

    for (i=0; i < raplen; i++) {
	val = FGetLong(file);
 	if (sub + val > size) {
	    FreeProgram(prg);
	    return DISKERR_INVALID;
	}
        if (val > 0) {
            unused = fread(src->src_Buffer + sub, 1, val, file);
            (void)unused;
        }
	sub += val;
	
	val = FGetLong(file);
	rap.rap_Relative = FindTableNum(pr, val);
	rap.rap_Star = FGetBool(file);
	rap.rap_NumFields = FGetChar(file);
	val = FGetLong(file);
	rap.rap_Table = FindTableNum(pr, val);
	for (j=0; j < rap.rap_NumFields; j++) {
            val = FGetLong(file);
	    rap.rap_Fields[j] = FindFieldNum(pr, val);
	}
	memcpy(src->src_Buffer + sub, &rap, sizeof(struct RelFieldPath));
	src->src_RapBuffer[i] = sub;
	sub += sizeof(struct RelFieldPath);
    }
    val = FGetLong(file);
    if (sub + val != size) {
	FreeProgram(prg);
	return DISKERR_INVALID;
    }
    if (val > 0) {
        unused = fread(src->src_Buffer + sub, 1, val, file);
        (void)unused;
    }
    sub += val;
	
    if (ferror(file)) {
        FreeProgram(prg);
        return DISKERR_READ;
    }

    st = GetSourceText(prg);
    error = st->st_Error;
    if (error == OK) {
        ProjectAddCodesetWarnings(pr, SourceTextCodesetWarnings(st));
    }
    FreeSourceText(st);
    if (error) {
        FreeProgram(prg);
        return error;
    }

    prg->prg_NeedRecompile = TRUE;
    *prgp = prg;

    return OK;
}

