/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** program.h
**
** Creation date: 17.04.1995
** Last modified: 12.05.2022
*/

#ifndef PROGRAM_H
#define PROGRAM_H

#include <system.h>

enum {
    OBSOLETE_IGNORE,
    OBSOLETE_WARN,
    OBSOLETE_ERROR
};

struct Program;
struct UserGlobalVar;
struct UserFunction;
struct Dependency;
struct DisplayObject;

#ifdef NEED_NODE_CAST
inline UserFunction *nodecast(UserFunction *, Node *node)
{ return (UserFunction *)node; }
inline Dependency *nodecast(Dependency *, Node *node)
{ return (Dependency *)node; }
#endif

#ifndef COMPILE_IMPL_H

struct LocalIdent;
struct Start;

#endif

struct SourceText;

extern const char onOpen[], onReload[], onClose[], onChange[], onAdminMode[];
extern const char mainWindowTitle[];
extern const char logLabel[];

struct Element         /* if this structure changes, functable.c needs update */
{
    char el_Type;
    char el_Allocated;
    char el_InUse;          		/* used by garbage collector */
    char el_Unused;
    struct Record *el_Record;		/* record from which data comes */
    union
    {
        long integer;
        double real;
        long date;
        long time;
        char *mi;
        struct
        {
            char *string;
            struct Field *field; /* set when string holds field name */
        } string;
        struct
        {
            struct Table *tab;
            struct Record *rec;
        } tabrec;
        struct
        {
            struct Table *tab;
            bool star;
        } tab;
        struct
        {
            struct RelFieldPath *rap;
            struct Record *rec;
        } field;
        struct
        {
            struct Element *head;
            struct Element *tail;
        } list;
        struct
        {
            short num;              /* index into file table */
            ulong id;               /* id for checking file handle validity */
        } file;
        struct
        {
             void *ptr;
             void (*func)(void *ptr, short method, struct Record *del);
        } special;
        struct
        {
            const char *buffer;
            long index;
        } codeaddress;
        struct
        {
            short type;         /* ELEM_USERFUNCTION or ELEM_FUNCINFO */
            void *ptr;          /* pointer to UserFunction or FuncInfo */
        } function;
    } el_Contents;
};

/* methods */
enum
{
    ELEM_SPECIAL_FREE,
    ELEM_SPECIAL_REMRECORD,
    ELEM_SPECIAL_REMALLRECORDS,
};

/* function types */
enum
{
    ELEM_USERFUNCTION,
    ELEM_FUNCINFO,
};

#ifdef __cplusplus
extern "C" {
#endif

char *SimpleElementToString(const struct Element *el, 
	struct Field *field, struct MemPool *memPool, bool *allocated);
/*
** returns simple string representation of element.
** if field != NULL then a "better" string representation can be found.
** memPool specifies the pool the memory for the string will be allocated.
** allocated will be set to TRUE if an allocation has been done.
*/

bool ElementToStream(struct Element *elem, bool (*out)(const char *));
/*
** Converts element into string by writing sub strings using the
** supplied out function.  Generates an output that is both
** readable and interpretable, e.g. strings are surrounded by
** double quotes and lists are surrounded by brackets.
** Returns TRUE on success or FALSE if out of memory.
*/

long ElementCmp(struct Element *elem1, struct Element *elem2, 
        bool star, bool order, struct MemPool *memPool, long *perror);
/*
** Compares two elements.  If star is TRUE then case-insensitive comparison
** is performed on strings, records are compared using their order, etc.
** If order is TRUE then comparison is performed for a sorting order.
** Returns a value < 0, 0 or > 0 indicating the order.
** If an error occurs then *perror is set and 0 is returned.
*/

/*
** Returns the length of a list.
*/
size_t CountListLength(const struct Element *elem);

void FreeSourceText(struct SourceText *st);
const char *SourceTextText(struct SourceText *st);
size_t SourceTextCodesetWarnings(struct SourceText *st);

struct SourceText *GetSourceText(const struct Program *prg);

const char *UserGlobalVarName(struct UserGlobalVar *ugv);

struct List *ProgramUserFunctions(struct Program *prg);
struct UserFunction *FindUserFunctionName(struct Program *prg,const char *name);
const char *UserFunctionName(struct UserFunction *uf);
bool UserFunctionTrigger(struct UserFunction *uf);

struct List *ProgramDisplayDependenciesVirtualField(struct Program *prg, 
        struct DisplayObject *dio);
struct UserGlobalVar *DependencyUserGlobalVar(struct Dependency *dep);
struct Field *DependencyField(struct Dependency *dep);
struct DisplayObject *DependencyGuiObject(struct Dependency *dep);
struct Table *DependencyTable(struct Dependency *dep);
struct DisplayObject *DependencyGuiRecord(struct Dependency *dep);

void SetTableProgramRecords(struct Project *pr);
/*
** sets all program record pointers to current gui record pointers
*/

void FreeProgram(struct Program *prog);

bool ProgramNeedRecompile(struct Program *prog);
void SetProgramNeedRecompile(struct Program *prg, bool recompile);

void ProgramBegin(struct Program *prg, bool needDefVars);
struct Element *ProgramCall(struct Program *prg, struct UserFunction *uf,
        struct Element **args, bool allowSideEffect);
void ProgramEnd(void);
const char *ProgramError(long *line, long *col, const char **filename);

const char *ExecuteProgram(struct Program *prg, bool needDefVars,
		     struct UserFunction *uf, struct Element **args,
		     bool allowSideEffect,
		     const char *(*func)(struct Element *elem),
		     long *line, long *col, const char **filename);
/*
** Executes the program prg (may be NULL) and returns NULL or an
** error string. The error string can be EXECUTE_ABORT in which case
** the user aborted program execution.
** If needDefVars != FALSE then the global defvars are installed on the stack.
** If uf != NULL then this user function is called, otherwise the
** execution starts at position 0.
** arg may contain a pointer to an array of Elements which are used
** as arguments for the user function. Remaining function arguments
** are initialized with NIL.  If allowSideEffect is set to FALSE then
** only functions without side effects are okay to call.
** If func != NULL then this function is called after executing the program.
** Be prepared to receive a NULL value in case the program doesn't return
** anything (e.g. prg = NULL) or an error has been occured.
** The function is called in any case even if an error occured.
** If no error occured then the error string of your function
** will be returned.
** line, col, and filename will hold the line, column, and filename
** information in case of an error.
*/

void ComputeDependentDisplayObjects(struct Project *pr, void *p,
        bool (*accept)(struct DisplayObject *dio, void *ptr), void *ptr);
/*
** Marks all display objects for update that pass the accept test
** (if NULL, all objects pass) and that have the given pointer as a dependency.
** The dependent objects are updated at the end of program execution, or when
** UpdateDependentDisplayObjects is called.
** Must be called during program execution.
*/

const char *UpdateDependentDisplayObjects(struct Project *pr, void *p,
        bool (*accept)(struct DisplayObject *dio, void *ptr), void *ptr,
        long *line, long *col, const char **filename);
/*
** Recomputes the settings of all display objects 
** that pass the accept test (if NULL, all objects pass)
** and that have the given pointer as a dependency
** Returns NULL or error string.
*/ 

#define EXECUTE_ABORT (const char *)(~0)

bool ProgramRunning(void);
/*
** Returns TRUE if a program is running at the moment.
*/

void ProgramFreeResources(void);
/*
** Frees resources for a close all function.
*/

bool ProgramFindObject(struct Program *prg, void *ptr);
/*
** Returns TRUE if ptr (table or field) is found in Program prg.
*/

void ProgramRemTable(struct Table *tab);
/*
** Removes all references to the specified table.
*/

void ProgramAddField(struct Field *field);
/*
** Might mark the program for recompilation.
*/

void ProgramRemField(struct Field *field);
/*
** Removes all references to the specified field.
*/

void ProgramRemDisplayObject(struct DisplayObject *dio);
/*
** Removes all references to the specified display object.
*/

void ProgramRemRecord(struct Table *tab, struct Record *rec);
/*
** Removes all references to the specified record.
*/

void ProgramRemAllRecords(struct Project *pr);
/*
** Removes references to all records of a project.
*/

const char *ProgramCompileProjectProgram(struct Project *pr, bool internal,
	struct Program **compiled, long *line, long *col,const char **filename);
/*
** (Re)-compiles a project's program.  If internal is TRUE compiles the
** internal program, otherwise uses setting in preferences.
*/

const char *ProgramCompileFilter(struct Table *tab, struct Program **compiled,
        long *line, long *col);
/*
** (Re-)compiles a filter program.
*/

const char *ProgramCompileQuery(const struct Program *prg,
        struct Program **compiled, long *line, long *col);
/*
** (Re-)compiles a query program.
*/

const char *ProgramCreateSourceExternal(struct Project *pr);
/*
** Saves program source in external file under project dir
** Returns NULL on success or error string.
*/

void ProgramDeleteSourceExternal(struct Project *pr);
/*
** Deletes external program source file.
*/

bool ProgramSourceExternalIdentical(struct Project *pr);
/*
** Compares the internal program with the source text of a potential
** external source file.  Returns TRUE if the external file exists 
** is identical with the internal version.
*/

long LoadProgramCustomFormat(struct Project *pr, FILE *file,
        struct Program **prgp);
/*
** Loads program from disk. Only the source part is loaded, the program
** has to be compiled to be runable. Returns OK on success, prgp holds the
** pointer to the allocated program or NULL for an empty program.
*/

long ReadProgram(struct Project *pr, const char *text, struct Program **prgp);
/*
** Reads program from a source text.  Only the source part is loaded, the
** program has to be compiled to be runable. Returns OK on success, prgp holds
** the pointer to the allocated program or NULL for an empty program.
*/

/*
** Functions for internal use
*/

struct Program *StartToProgram(struct Project *pr, struct Start *start,
        struct LocalIdent *defList, bool debug);

const char *CreateSourceExternal(struct Project *pr, 
	const char *text, long size);

struct Bin *ProgramBin(struct Program *prg);
/*
** returns Bin part of program
*/

#ifdef __cplusplus
}
#endif

#endif

