/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** programbin.h
**
** Creation date: 21.04.1995
** Last modified: 13.05.2022
*/

#ifndef PROGRAMBIN_H
#define PROGRAMBIN_H

#include "name.h"
#include "program.h"

#ifdef __cplusplus
extern "C" {
#endif

void FreeBin(struct Bin *bin);
struct Bin *StartToBin(struct Project *pr, struct Start *start,
        struct LocalIdent *defList, bool debug);


void BinBegin(struct Project *pr, struct Bin *b, bool needDefVars);
struct Element *BinCall(struct Bin *b, struct UserFunction *uf,
        struct Element **args, bool allowSideEffects);
void BinEnd(void);
const char *BinError(long *line, long *col, const char **filename);
/* 
** See programbin.c/ExecuteBin() for an example on how to
** use BinBegin, BinCall, BinEnd, and BinError. 
*/

const char *ExecuteBin(struct Project *pr, struct Bin *b, bool needDefVars,
		 struct UserFunction *uf, struct Element **args,
		 bool allowSideEffects,
		 const char *(*func)(struct Element *elem),
		 long *line, long *col, const char **filename);
/*
** Executes the program and returns NULL or an error string.
** If needDefVars != FALSE then the global defvar variables are setup first.
** If uf != NULL then the user defined function is executed.
** arg may contain a pointer to an array of Elements which are used
** as arguments for the user function. Remaining function arguments
** are initialized with NIL.  if allowSideEffects is set to FALSE
** then only functions that do not have any side effects are
** allowed to call.
** If func != NULL then this function is called after executing
** the program. Be prepared to receive a NULL value in case
** the program doesn't return anything or an error has been occured.
** The function is called in any case even if an error has been
** occured. If no error has been occured then the string returned
** by your function will be used as return code.
** line, col, and filename will hold the line, column and filename
** information in case of an error.
*/

struct Record *RelFieldPathRecord(struct RelFieldPath *rap, struct Record *rec);
/*
** Returns the record for the last field in the specified tabfieldpath
** starting with the specified record for the first field.
** If an error occurs, program execution is aborted.
*/

struct Element *GetGlobalData(struct RelFieldPath *rap, struct Record *rec);
/*
** allocates an element cell and stores value of global data item.
*/

void CallUserFunction(struct UserFunction *uf);
/*
** Calls UserFunction.  Pushes return address on stack and continues
** execution at address of UserFunction (non-synchronous call).
*/

void CallUserFunctionAdjustArgs(struct UserFunction *uf, short numArgs);
/*
** Adjusts number of arguments on stack to the number required 
** by the UserFunction and then performs CallUserFunction(uf)
*/

void ExecuteUserFunction(struct UserFunction *uf);
/*
** Execute UserFunction.  Calls the user function and executes it
** (synchronous call)
*/

void ExecuteUserFunctionAdjustArgs(struct UserFunction *uf, short numArgs);
/*
** Adjusts number of arguments on stack to the number required 
** by the UserFunction and then performs ExecuteUserFunction(uf)
*/

void ExecuteAbort(const char *error);
/*
** may be called anywhere in any function to return to main procedure and to
** set an error message. On Ctrl-C you can specify EXECUTE_ABORT as error
** string.
*/

void ExecutePrepareProjectChange(void);
/*
** Should be called before making an actual change to a project.  This obtains
** a change lock or sets an error.  The function only returns on success.
*/
 
void *ExecuteAllocMem(ulong size);
/*
** Allocates memory from executeMemPool.
** On out of memory ExecuteAbort(GCS(MSG_Err_Mem)) is called.
** This function returns a pointer != NULL.
*/

char *ExecuteAllocString(const char *s);
char *ExecuteAllocStringN(const char *s, long n);
/*
** Allocates a string from the executeMemPool
** On out of memory ExecuteAbort(GCS(MSG_Err_Mem)) is called.
** This function returns a pointer != NULL.
*/

extern struct MemPool *executeMemPool;
extern struct Project *executeProject;
extern struct List executeVirtualFieldComputeCache;
extern bool executeProjectChange;
extern bool executeAllowSideEffects;

/*
** for garbage collector & internal use
*/

struct SubProgram;

struct List *BinUserFunctions(struct Bin *bin);
struct UserFunction *BinFindUserFunctionName(struct Bin *bin, const char *name);
char UserFunctionType(struct UserFunction *uf);
short UserFunctionArity(struct UserFunction *uf);
void UserFunctionPosition(struct UserFunction *uf, long *line, long *col);
char *UserFunctionTypes(struct UserFunction *uf);
bool UserFunctionTrigger(struct UserFunction *uf);

struct List *BinDisplayDependenciesVirtualField(struct Bin *bin,
        struct DisplayObject *dio);

void FreeElement(struct MemPool *memPool, struct Element *elem);

void ExecuteEvalSubProgram(struct SubProgram *sp);
/*
** Executes sub program. Needed by sfw.
*/

const char *ExecuteEvalBin(struct Bin *bin, struct Element **elem);
/*
** Executes program. On success NULL is returned and *elem is set
** to the top element of the stack. Otherwise an error string is returned.
** In either case the stack is fixed for further use.
*/

struct Bin *TriggerToBin(struct DisplayObject *dio, short type);
void CallTriggerFunction(const char *trigger, short numargs);     
	/* for internal calling */

void ProgramBinRemTable(struct Bin *bin, struct Table *tab);
void ProgramBinRemRecord(struct Bin *bin, struct Record *rec);
void ProgramBinRemAllRecords(struct Bin *bin);
bool ProgramBinRemField(struct Bin *bin, struct Field *field);
	/* returns TRUE if program should be recompiled */
bool ProgramBinRemDisplayObject(struct Bin *bin, struct DisplayObject *dio);
	/* returns TRUE if program should be recompiled */

#ifdef __cplusplus
}
#endif

#endif

