/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** programbinexe.c
**
** Creation date: 20.04.1995
** Last modified: 17.05.2022
*/

#include "programbin_impl.h"
#include "program.h"
#include "global.h"
#include "funccmp.h"
#include "funcsfw.h"
#include "funcio.h"
#include "stack.h"
#include "heap.h"
#include "gui.h"
#include "break.h"
#include "order.h"
#include "memo.h"
#include "memory.h"
#include "appstrings_cat.h"
#include "error.h"
#include <setjmp.h>

#define MAX_EXECUTE_CALL_COUNT 16
     
struct MemPool *executeMemPool = NULL;
struct Project *executeProject = NULL;
bool executeProjectChange = FALSE;
bool executeAllowSideEffects = FALSE;
struct List executeVirtualFieldComputeCache;
static struct BreakSettings executeBreakSettings;

struct ExecuteControlData
{
    jmp_buf jmpBuf;
    const char *buffer;
    long index;
};

static struct ExecuteControlData ecd;

static const char *executeError;
static struct LineCol executeLineCol;
static short executeCallCount = 0;
static long programRunning = 0;

void ExecuteAbort(const char *error)
{
    executeError = error;
    longjmp(ecd.jmpBuf, 1);                 /* return to main procedure */
}

void ExecutePrepareProjectChange(void)
{
    long error = ObtainProjectChangeLock(executeProject);
    if (error != OK) {
        ExecuteAbort(GES(error));
    }
}

void *ExecuteAllocMem(ulong size)
{
    void *ptr = PoolAllocMem(executeMemPool, size);

    if (ptr == NULL) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }
    return ptr;
}

char *ExecuteAllocString(const char *s)
{
    char *t = AllocString(executeMemPool, s);

    if (t == NULL) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }
    return t;
}

char *ExecuteAllocStringN(const char *s, long n)
{
    char *t = AllocStringN(executeMemPool, s, n);

    if (t == NULL) {
        ExecuteAbort(GCS(MSG_Err_Mem));
    }
    return t;
}

void cacheVirtualFieldComputeResult(struct Field *field,
        struct DisplayDependencies *dd, struct Record *rec,
        struct Element *elem)
{
    struct VirtualFieldComputeCacheEntry *entry;
    bool found = FALSE;

    FORNODE(entry, &executeVirtualFieldComputeCache) {
        if (entry->vfcce_Field == field) {
            found = TRUE;
            REMOVE(entry);
            break;
        }
    }
    if (!found) {
        entry = ExecuteAllocMem(sizeof(struct VirtualFieldComputeCacheEntry));
    }
    ADDHEAD(&executeVirtualFieldComputeCache, entry);
    entry->vfcce_Field = field;
    entry->vfcce_DisplayDependencies = dd;
    entry->vfcce_Record = rec;
    entry->vfcce_Element = elem;
}

struct Element *getCachedVirtualFieldComputeResult(struct Field *field,
        struct Record *rec)
{
    struct VirtualFieldComputeCacheEntry *entry;
    struct DisplayObject *dio = FieldDisplayObject(field);

    /* We only trust cached result of auto-compute virtual fields. */
    if (!DisplayFieldAutoUpdate(dio)) {
        return NULL;
    }

    FORNODE(entry, &executeVirtualFieldComputeCache) {
        if (entry->vfcce_Field == field) {
            if (entry->vfcce_Record == rec &&
                    !entry->vfcce_DisplayDependencies->dd_Visited) {
                return entry->vfcce_Element;
            }
            break;
        }
    }

    return NULL;
}

static void __inline binin(void *buf, long size)
{
    memcpy(buf, ecd.buffer + ecd.index, size);
    ecd.index += size;
}

#define bininchar() ecd.buffer[ecd.index++]
#define binintag bininchar

static __inline short bininshort(void) 
{ short s; binin(&s, sizeof(short)); return(s); }
static __inline long bininlong(void) 
{ long l; binin(&l, sizeof(long)); return(l); }
static __inline double bininreal(void) 
{ double r; binin(&r, sizeof(double)); return(r); }
static __inline void *bininptr(void) 
{ void *ptr; binin(&ptr, sizeof(void *)); return(ptr); }

static void mkinteger(void)
{
    PushInteger(bininlong());
}

static void mkreal(void)
{
    PushReal(bininreal());
}

static void mkstring(void)
{
    PushString((char *)(ecd.buffer + ecd.index), FALSE);
    ecd.index += strlen(ecd.buffer + ecd.index) + 1;
}

static void mkdate(void)
{
    PushDate(bininlong());
}

static void maktime(void)               /* mktime already used */
{
    PushTime(bininlong());
}

static void mktrue(void)
{
    PushTrue();
}

static void mknil(void)
{
    PushNil();
}

struct Record *RelFieldPathRecord(struct RelFieldPath *rap, struct Record *rec)
{
    struct Field *field;
    struct DataInfo *di, di_buf;
    struct Record *tmp = NULL;
    long error = OK;
    int i;

    if (rap == NULL) {
        return rec;
    }

    RecordLock(rec);
    for(i = 0; i < rap->rap_NumFields-1; i++)
    {
        field = rap->rap_Fields[i];
        if (FieldType(field) != TYPE_REFERENCE) {
            error = ERR_INTERNAL;
            break;
        }
        error = RecordDataInfo(rec, field, &di, &di_buf);
        if (error) {
            break;
        }

        tmp = di->di_Contents.rec;
        RecordUnlock(rec);
        rec = tmp;
        RecordLock(rec);
    }
    RecordUnlock(rec);
    if (error) {
        ExecuteAbort(GES(error));
    }
    return rec;
}

struct Element *GetGlobalData(struct RelFieldPath *rap, struct Record *rec)
{
    struct Element *elem = NULL;

    if (rap->rap_NumFields == 0) {
        if (rap->rap_Star) {
            rec = TableGuiRecord(rap->rap_Table);
        }
	if (rec == NULL) {
	    elem = MakeNil();
        } else {
	    elem = MakeRecord(rap->rap_Table, rec);
	    elem->el_Record = rec;
	}
    } else {
        struct Field *field = RelFieldPathField(rap);
        short type = FieldType(field);
        struct DataInfo *di, di_buf;
	long error = OK;

        rec = RelFieldPathRecord(rap, rec);
        RecordLock(rec);
        error = RecordDataInfo(rec, field, &di, &di_buf);
        if (error == OK) {
            if (di->di_Flags & DIF_NIL) {
		elem = MakeElement(TYPE_NIL, FALSE);
		if (type != TYPE_REFERENCE) {
		    elem->el_Record = rec;
		}
	    } else {
                switch(type) {
                    case TYPE_BOOL:
			elem = MakeElement(TYPE_BOOL, FALSE);
			elem->el_Record = rec;
                        break;

                    case TYPE_INTEGER:
			elem = MakeInteger(di->di_Contents.integer);
			elem->el_Record = rec;
                        break;

                    case TYPE_CHOICE:
			elem = MakeInteger(di->di_Contents.choice);
			elem->el_Record = rec;
                        break;

                    case TYPE_REAL:
                        elem = MakeReal(di->di_Contents.real);
			elem->el_Record = rec;
                        break;

                    case TYPE_STRING:
                        if (di->di_Contents.string) {
                            elem = MakeString(
                                    ExecuteAllocString(di->di_Contents.string), 
                                    TRUE);
                        } else {
                            elem = MakeString("", FALSE);
                        }
			elem->el_Record = rec;
                        break;

                    case TYPE_MEMO:
                        if (di->di_Contents.mi) {
                            elem = MakeMemo(
                                    ExecuteAllocString(di->di_Contents.mi),
                                    TRUE);
                        } else {
                            elem = MakeMemo("", FALSE);
                        }
			elem->el_Record = rec;
                        break;

                    case TYPE_DATE:
                        elem = MakeDate(di->di_Contents.date);
			elem->el_Record = rec;
                        break;

                    case TYPE_TIME:
                        elem = MakeTime(di->di_Contents.time);
			elem->el_Record = rec;
                        break;

                    case TYPE_REFERENCE:
			if (di->di_Contents.rec == NULL) {
			    elem = MakeNil();
                        } else {
			    elem = MakeRecord(FieldReferenceTable(field),
				    di->di_Contents.rec);
			    elem->el_Record = di->di_Contents.rec;
			}
                        break;

                    default:
			elem = MakeNil();
                        error = ERR_INTERNAL;
                        break;
                }
            }
        }
        RecordUnlock(rec);
        if (error) {
            ExecuteAbort(GES(error));
        }
    }

    return elem;
}

static void getglobaldata(void)
{
    struct RelFieldPath rap;
    struct Element *elem;

    binin(&rap, sizeof(struct RelFieldPath));
    elem = GetGlobalData(&rap, TableProgramRecord(rap.rap_Table));
    StackPush(elem);
}

static void initruntrigger(void)
{
    struct Table *tab = bininptr();

    if (tab) {
        PushRecord(tab, TableProgramRecord(tab));
        SetTableProgramRecord(tab, TableGuiRecord(tab));
    }
}

static void exitruntrigger(void)
{
    struct Table *tab = bininptr();
    struct Element *elem = StackPop();  /* result of compute function */

    if (tab) {
        SetTableProgramRecord(tab, PopRecord2MatchTable(tab));
    }
    StackPush(elem);
}

static void callCompute(const char *compute, const char *error)
{
    struct UserFunction *uf;
    int i;

    if (compute == NULL || compute[0] == 0) {
        PushNil();              /* no compute function. */
        return;
    }

    uf = FindUserFunctionName(ProjectProgram(executeProject), compute);
    if (uf == NULL) {
        ExecuteAbort(error);
    }

    for (i=0; i < uf->uf_Arity; i++) {
        PushNil();                              /* push NIL args */
    }

    PushCodeAddress(ecd.buffer, ecd.index);       /* push return address */

    ecd.buffer = uf->uf_Bin->bin_Buffer;         /* jump to sub program */
    ecd.index = uf->uf_SubProgram.sp_Index;
}

static void initgetvirtualhelp(struct RelFieldPath *rap, struct Record *rec)
{
    struct Table *tab = rap->rap_Table;
    struct Field *field = NULL;
    struct DataInfo *di = NULL, di_buf;
    struct Record *tmp = NULL;
    struct Element *elem = NULL;
    const char *compute = NULL;
    short num = rap->rap_NumFields, i;
    long error = OK;

    if (num == 0) {
        ExecuteAbort(GES(ERR_INTERNAL));
    }

    for (i = 0; i < num - 1; i++) {
        field = rap->rap_Fields[i];
        if (FieldType(field) != TYPE_REFERENCE) {
            ExecuteAbort(GES(ERR_INTERNAL));
        }

        RecordLock(rec);
        error = RecordDataInfo(rec, field, &di, &di_buf);
        if (error == OK) {
            tmp = (di->di_Flags & DIF_NIL)? NULL : di->di_Contents.rec;
        }
        RecordUnlock(rec);

        if (error) {
            ExecuteAbort(GES(error));
        }

        rec = tmp;
    }

    field = rap->rap_Fields[num - 1];
    tab = FieldTable(field);

    /*
    ** Save current program record and new program record.
    */
    PushRecord(tab, TableProgramRecord(tab));
    PushRecord(tab, rec);

    elem = getCachedVirtualFieldComputeResult(field, rec);
    if (elem != NULL) {
        StackPush(elem);
    } else {
        SetTableProgramRecord(tab, rec);

        compute =  FieldVirtualCompute(field);
        callCompute(compute, GCS(MSG_ProgErr_VirtualComputeFunction));
    }
}

static void initgetvirtual(void)
{
    struct RelFieldPath rap;
    struct Table *tab = NULL;
    struct Record *rec = NULL;

    binin(&rap, sizeof(struct RelFieldPath));
    tab = rap.rap_Table;
    rec = TableProgramRecord(tab);
    initgetvirtualhelp(&rap, rec);
}

static void initgetvirtualextend(void)
{
    short n = bininshort();
    struct RelFieldPath rap;
    struct Element *elem = StackNth(n);
    struct Record *rec = NULL;

    binin(&rap, sizeof(struct RelFieldPath));

    if (elem->el_Type == TYPE_NIL) {
        rec = NULL;
    } else if (elem->el_Type != TYPE_RECORD) {
        ExecuteAbort(GCS(MSG_ProgErr_Type));
    } else if (elem->el_Contents.tabrec.tab != rap.rap_Table) {
        ExecuteAbort(GCS(MSG_ProgErr_Type));
    } else {
        rec = elem->el_Contents.tabrec.rec;
    }

    initgetvirtualhelp(&rap, rec);
}

static void exitgetvirtual(void)
{
    struct Field *field = bininptr();
    struct Table *tab = FieldTable(field);
    struct DisplayDependencies *dd = bininptr();
    struct Element *result = StackPop();
    struct Record *rec1 = PopRecord2MatchTable(tab);
    struct Record *rec2 = PopRecord2MatchTable(tab);

    /* Correctly set the record pointer */
    result->el_Record = rec1;
    StackPush(result);

    SetTableProgramRecord(tab, rec2);
    if (TableGuiRecord(tab) == rec1) {
        cacheVirtualFieldComputeResult(field, dd, rec1, result);
	if (dd) {
	    dd->dd_Visited = FALSE;	        /* no need to update anymore */
	}
    }
}

static void initgetenabled(void)
{
    struct DisplayObject *dio = bininptr();
    const char *compute =  DisplayObjectEnabledCompute(dio);

    callCompute(compute, GCS(MSG_ProgErr_EnabledComputeFunction));
}

static void exitgetenabled(void)
{
    struct Element *result = StackTop();
    struct DisplayObject *dio = bininptr();
    struct DisplayDependencies *dd = bininptr();
    bool enabled = (result != NULL && result->el_Type != TYPE_NIL);

    GuiSetDisplayObjectEnabled(dio, enabled);
    if (dd) {
        dd->dd_Visited = FALSE;	                /* no need to update anymore */
    }
}

static void initgettitle(void)
{
    callCompute(mainWindowTitle, GCS(MSG_ProgErr_Trigger));
}

static void exitgettitle(void)
{
    struct Element *result = StackTop();
    struct DisplayObject *dio = bininptr();
    struct DisplayDependencies *dd = bininptr();
    const char *title;
    GuiObject win;

    // keep result on stack
    StackPush(result);
    title = PopString2();

    win = DisplayObjectGuiObject(dio);
    GuiSetWindowTitle(win, title);
    if (dd) {
        dd->dd_Visited = FALSE;	                /* no need to update anymore */
    }
}

static void getlocaldata(void)
{
    short n = bininshort();

    StackPush(StackNth(n));
}

static void getlocalextend(void)
{
    short n = bininshort();
    struct RelFieldPath rap;
    struct Element *elem = StackNth(n);
    struct Record *rec = NULL;

    binin(&rap, sizeof(struct RelFieldPath));

    if (elem->el_Type == TYPE_NIL) {
        rec = NULL;
    } else if (elem->el_Type != TYPE_RECORD) {
        ExecuteAbort(GCS(MSG_ProgErr_Type));
    } else if (elem->el_Contents.tabrec.tab != rap.rap_Table) {
        ExecuteAbort(GCS(MSG_ProgErr_Type));
    } else {
        rec = elem->el_Contents.tabrec.rec;
    }

    elem = GetGlobalData(&rap, rec);
    StackPush(elem);
}

static void presfw(void)
{
    short numRels = bininshort(), i;
    short numVars = bininshort();
    struct Table *tab;

    for(i=0; i < numRels; i++) {
        tab = bininptr();
        PushRecord(tab, TableProgramRecord(tab));
    }
    for(i=0; i < numVars; i++) {
        /*
        ** we need to allocate a new element, since sfw
        ** is destructive
        */
        StackPush(MakeElement(TYPE_NIL, FALSE));
    }
}

static void sfw(void)
{
    struct sfwBin sb;
    struct SFWSelectItem *sfwsis = NULL;
    struct SFWFromItem *sfwfis = NULL;
    struct SFWOrderItem *sfwois = NULL;
    long sfwsis_size, sfwfis_size, sfwois_size;
    short i;

    binin(&sb, sizeof(struct sfwBin));

    sfwsis_size = sb.sb_NumSelectItems * sizeof(struct SFWSelectItem);
    if (sfwsis_size > 0) {
        sfwsis = PoolAllocMem(executeMemPool, sfwsis_size);
        if (sfwsis == NULL) {
            ExecuteAbort(GCS(MSG_Err_Mem));
        }
        binin(sfwsis, sfwsis_size);
        for(i=0; i < sb.sb_NumSelectItems; i++) {
            if (sfwsis[i].sfwsi_Title) {
                sfwsis[i].sfwsi_Title =
                    ecd.buffer + (intptr_t)sfwsis[i].sfwsi_Title;
            }
            if (sfwsis[i].sfwsi_LocalName) {
                sfwsis[i].sfwsi_LocalName = 
                    ecd.buffer + (intptr_t)sfwsis[i].sfwsi_LocalName;
            }
        }
    }

    sfwfis_size = sb.sb_NumFromItems * sizeof(struct SFWFromItem);
    if (sfwfis_size > 0) {
        sfwfis = PoolAllocMem(executeMemPool, sfwfis_size);
        if (sfwfis == NULL) {
            ExecuteAbort(GCS(MSG_Err_Mem));
        }
        binin(sfwfis, sfwfis_size);
    }

    sfwois_size = sb.sb_NumOrderItems * sizeof(struct SFWOrderItem);
    if (sfwois_size > 0) {
        sfwois = PoolAllocMem(executeMemPool, sfwois_size);
        if (sfwois == NULL) {
            ExecuteAbort(GCS(MSG_Err_Mem));
        }
        binin(sfwois, sfwois_size);
    }

    /* continue */
    ecd.index = sb.sb_Continue;

    SelectFromWhere(sb.sb_Distinct, sb.sb_NumSelectItems, sfwsis,
        sb.sb_NumFromItems, sfwfis,
        sb.sb_Where? &sb.sb_WhereExpr : NULL,
        sb.sb_NumOrderItems, sfwois,
        sb.sb_IsSelectFromWhere);

    if (sfwois) {
        PoolFreeMem(executeMemPool, sfwois, sfwois_size);
    }
    if (sfwfis) {
        PoolFreeMem(executeMemPool, sfwfis, sfwfis_size);
    }
    if (sfwsis) {
        PoolFreeMem(executeMemPool, sfwsis, sfwsis_size);
    }
}

static void foralljmp(void)
{
    long index = bininlong();

    if (sfwSetNextRecords()) {
        ecd.index = index;
    }
}

static void postsfw(void)
{
    short numRels = bininshort(), i;
    short numVars = bininshort();
    struct Table *tab;
    struct Element *ret = StackPop();

    StackPop();                 /* remove special element */
    for(i=0; i < numVars; i++) {
        StackPop();
    }
    StackReverse(numRels);
    for(i=0; i < numRels; i++) {
        tab = bininptr();
        SetTableProgramRecord(tab, PopRecord2MatchTable(tab));
    }
    StackPush(ret);
}

static void pop(void)
{
    StackPop();
}

static void popn(void)
{
    short num = bininshort();

    StackSqueeze(0, num);
}

static void pop1n(void)
{
    short num = bininshort();

    StackSqueeze(1, num);
}

static void setlocalident(void)
{
    short n = bininshort();
    struct DisplayDependencies *dd;

    SetStackNth(n, StackTop());
    if (n < 0)	{		/* global variable? */
	while ((dd = bininptr()) != NULL) {
	    dd->dd_Visited = TRUE;
	}
    }
}

static void dotimes_init(void)
{
    struct Element *elem = PopInteger();

    StackPush(elem);
    PushInteger(0);
}

static void dotimes_step(void)
{
    struct Element *elem = StackPop();

    PushInteger(elem->el_Contents.integer + 1);
}

static void dotimes_test(void)
{
    struct Element *elem = StackTop();
    struct Element *bound = StackNth(1);

    if (bound->el_Type == TYPE_INTEGER && 
            elem->el_Contents.integer < bound->el_Contents.integer) {
        PushNil();
    } else {
        PushTrue();
    }
}

static void dolist_init(void)
{
    struct Element *elem = PopList();

    StackPush(elem);
    if (elem->el_Type == TYPE_LIST) {
        StackPush(elem->el_Contents.list.head);
    } else {
        PushNil();
    }
}

static void dolist_step(void)
{
    struct Element *elem;

    StackPop();
    elem = StackPop();

    if (elem->el_Type == TYPE_LIST) {
        elem = elem->el_Contents.list.tail;
    }
    StackPush(elem);
    if (elem->el_Type == TYPE_LIST) {
        StackPush(elem->el_Contents.list.head);
    } else {
        PushNil();
    }
}

static void dolist_test(void)
{
    struct Element *elem = StackNth(1);

    if (elem->el_Type == TYPE_LIST) {
        PushNil();
    } else {
        PushTrue();
    }
}

static void casecmp(void)
{
    struct Element *elem1 = StackPop();
    struct Element *elem2 = StackTop();

    if (ProgElementCmp(elem1, elem2, FALSE, FALSE) == 0) {
        PushTrue();
    } else {
        PushNil();
    }
}

void CallTriggerFunction(const char *trigger, short numArgs)
{
    struct Program *prg = ProjectProgram(executeProject);
    struct UserFunction *uf = FindUserFunctionName(prg, trigger);

    if (uf == NULL) {
        ExecuteAbort(GCS(MSG_ProgErr_Trigger));
    }

    CallUserFunctionAdjustArgs(uf, numArgs);
}

static bool CheckGlobalDataUpdate(struct RelFieldPath *rap, bool star)
{
    if (rap->rap_NumFields == 0) {
	return FALSE;
    } else {
	struct Field *field = rap->rap_Fields[rap->rap_NumFields-1];
	short type = FieldType(field);
	const char *trigger;

        if (star) {
            trigger = FieldTrigger(field);
	    if (trigger != NULL && trigger[0] != 0) {
		return FALSE;
	    }
        }

        if (type == TYPE_VIRTUAL) {
            return FALSE;	/* do not update dependent fields */
        }
	return TRUE;
    }
}

static void SetGlobalData(struct RelFieldPath *rap, struct Record *rec,bool star)
{
    struct Table *tab = rap->rap_Table;
    struct Field *field = NULL;
    struct Record *tmp;
    short type = NOTYPE;
    const char *trigger;

    if (rap->rap_NumFields == 0) {
	StackPush(StackTop());
	rec = PopRecord2MatchTable(tab);
        if (!rap->rap_Star) {
	    SetTableProgramRecord(tab, rec);           
	} else if(executeAllowSideEffects) {
	    GuiShowRecord(tab, rec);
	} else {
	    ExecuteAbort(GCS(MSG_ProgErr_SideEffect));
	}
    } else {
        struct DataInfo *di, di_buf;
        long error = OK;
        int i;
        bool nil = FALSE;
        union Data data, *pdata = &data;
        bool countChanges;

        for (i = 0; i < rap->rap_NumFields; i++) {
            field = rap->rap_Fields[i];
            type = FieldType(field);

            if (type == TYPE_REFERENCE && i < rap->rap_NumFields - 1) {
                tab = FieldReferenceTable(field);
                RecordLock(rec);
                error = RecordDataInfo(rec, field, &di, &di_buf);
                if (error) {
                    break;
		}
                tmp = di->di_Contents.rec;
                RecordUnlock(rec);
                rec = tmp;
            }
        }

        if (error) {
            ExecuteAbort(GES(error));
	}

        if (star) {
            trigger = FieldTrigger(field);
	    if (trigger != NULL && trigger[0] != 0) {
		CallTriggerFunction(trigger, 1);
		return;
	    }
        }

	if (!executeAllowSideEffects) {
	    ExecuteAbort(GCS(MSG_ProgErr_SideEffect));
	}

        if (type == TYPE_VIRTUAL) {
            if (TableGuiRecord(tab) == rec) {
                GuiSetVirtualField(field, rec, StackTop());
	    }
            return;
        }

        StackPush(StackTop());
        switch (type)
        {
            case TYPE_BOOL:
                nil = !PopBool2();
                break;

            case TYPE_INTEGER:
                nil = !PopInteger2(&data.integer);
                break;

            case TYPE_CHOICE:
            {
                ushort num = FieldChoiceNumLabels(field);
                long i;

                nil = !PopInteger2(&i);
                data.choice = (ushort)i;
                if (!nil && (i < 0 || i >= (long)num)) {
                    ExecuteAbort(GCS(MSG_ProgErr_ChoiceNumber));
                }
                break;
            }

            case TYPE_REAL:
                nil = !PopReal2(&data.real);
                break;

            case TYPE_STRING:
                data.string = PopString2();
                if (data.string == NULL) {
                    nil = TRUE;
                }
                break;

            case TYPE_MEMO:
                data.mi = PopMemo2();
                if (data.mi == NULL) {
                    nil = TRUE;
                }
                break;

            case TYPE_DATE:
                nil = !PopDate2(&data.date);
                break;

            case TYPE_TIME:
                nil = !PopTime2(&data.time);
                break;

            case TYPE_REFERENCE:
                data.rec = PopRecord2MatchTable(FieldReferenceTable(field));
                if (data.rec == NULL) {
                    nil = TRUE;
                }
                break;

            default:    ExecuteAbort(GCS(MSG_ProgErr_Type));
                        break;
        }

        countChanges = FieldCountChanges(field) &&
            TableCountChanges(FieldTable(field));
        if (countChanges) {
            ExecutePrepareProjectChange();
        }
        error = RecordStoreData(rec, field, pdata, nil);
        if (error) {
            ExecuteAbort(GES(error));
	} else if (countChanges) {
            executeProjectChange = TRUE;
	}

        error = ReorderRecords(tab);
        if (error) {
            ExecuteAbort(GES(error));
	}
        if (TableGuiRecord(tab) == rec) {
            GuiDisplayPanel(tab);
	}
        GuiUpdateField(field, rec);
    }
    return;
}

static void setglobalident(void)
{
    struct RelFieldPath rap;
    struct DisplayDependencies *dd;
    bool star;
    bool update;

    binin(&rap, sizeof(struct RelFieldPath));
    star = bininchar();
    update = CheckGlobalDataUpdate(&rap, star);
    if (rap.rap_NumFields > 0) {
	while ((dd = bininptr()) != NULL) {
	    dd->dd_Visited |= update;
	}
    }
    SetGlobalData(&rap, TableProgramRecord(rap.rap_Table), star);
}

static void setlocalextend(void)
{
    short n;
    struct RelFieldPath rap;
    struct Element *elem;
    struct Record *rec = NULL;
    struct DisplayDependencies *dd;
    bool star;
    bool update;

    n = bininshort();
    elem = StackNth(n);
    binin(&rap, sizeof(struct RelFieldPath));
    star = bininchar();

    if (elem->el_Type == TYPE_NIL) {
        rec = NULL;
    } else if (elem->el_Type != TYPE_RECORD) {
        ExecuteAbort(GCS(MSG_ProgErr_Type));
    } else if (elem->el_Contents.tabrec.tab != rap.rap_Table) {
        ExecuteAbort(GCS(MSG_ProgErr_Type));
    } else {
        rec = elem->el_Contents.tabrec.rec;
    }

    update = CheckGlobalDataUpdate(&rap, star);
    if (rap.rap_NumFields > 0) {
	while ((dd = bininptr()) != NULL) {
	    dd->dd_Visited |= update;
	}
    }
    SetGlobalData(&rap, rec, star);
}

static void pushTop(void)
{
    StackPush(StackTop());
}

static void setqListNext(void)
{
    struct Element *elem = PopList();

    if (elem->el_Type != TYPE_LIST) {
        ExecuteAbort(GCS(MSG_ProgErr_MissingArguments));
    }
    StackPush(elem->el_Contents.list.tail);
    StackPush(elem->el_Contents.list.head);
}

static void setqListEnd(void)
{
    struct Element *elem = PopList();

    if (elem->el_Type != TYPE_NIL) {
        ExecuteAbort(GCS(MSG_ProgErr_TooManyArguments));
    }
}

static void callfunc(void)
{
    struct funcBin fb;

    binin(&fb, sizeof(struct funcBin));
    
    if (fb.fb_Code == NULL) {
	ExecuteAbort(GCS(MSG_Err_NotImplemented));
    }
    if (!executeAllowSideEffects && fb.fb_HasSideEffect) {
	ExecuteAbort(GCS(MSG_ProgErr_SideEffect));
    }
    
    StackReverse(fb.fb_NumArgs);
    fb.fb_Code(fb.fb_NumArgs);
}

static void jmpand(void)
{
    struct Element *elem = StackPop();
    long index = bininlong();

    if (elem->el_Type == TYPE_NIL) {
        PushNil();
        ecd.index = index;
    }
}

static void jmpor(void)
{
    struct Element *elem = StackPop();
    long index = bininlong();

    if (elem->el_Type != TYPE_NIL) {
        StackPush(elem);
        ecd.index = index;
    }
}

static void jmpnil(void)
{
    struct Element *elem = StackPop();
    long index = bininlong();

    if (elem->el_Type == TYPE_NIL) {
        ecd.index = index;
    }
}

static void jmptrue(void)
{
    struct Element *elem = StackPop();
    long index = bininlong();

    if (elem->el_Type != TYPE_NIL) {
        ecd.index = index;
    }
}

static void jmpglobalvar(void)
{
    struct UserGlobalVar *ugv = bininptr();
    long index = bininlong();

    if (ugv->ugv_Element) {
        StackPush(ugv->ugv_Element);
        ecd.index = index;
    }
}

static void jmp(void)
{
     ecd.index = bininlong();
}

static void popjmp(void)
{
     StackSqueeze(0, bininshort());
     ecd.index = bininlong();
}

static void pop1jmp(void)
{
     StackSqueeze(1, bininshort());
     ecd.index = bininlong();
}

static void jsr(void)
{
    CallUserFunction(bininptr());
}

static void rts(void)
{
    short num = bininshort();
    struct Element *codeaddress = StackNth(1);        /* return address */

    if (codeaddress->el_Type != TYPE_CODEADDRESS) {
        ExecuteAbort(GCS(MSG_Err_Internal));
    }
    ecd.buffer = codeaddress->el_Contents.codeaddress.buffer;
    ecd.index = codeaddress->el_Contents.codeaddress.index;

    StackSqueeze(1, num + 1);     /* removes local variables + return address */
}

static void constinfo(void)
{
    StackPush(bininptr());
}

static void funcinfo(void)
{
    PushFunction(ELEM_FUNCINFO, bininptr());
}

static void userfunction(void)
{
    PushFunction(ELEM_USERFUNCTION, bininptr());
}

static void pushtable(void)
{
    struct Table *tab = bininptr();
    bool star = bininchar();

    PushTable(tab, star);
}

static void pushfield(void)
{
    struct RelFieldPath *rap = ExecuteAllocMem(sizeof(struct RelFieldPath));

    binin(rap, sizeof(struct RelFieldPath));
    PushField(rap, TableProgramRecord(rap->rap_Table), TRUE);
}

static void pushfieldextend(void)
{
    short n = bininshort();
    struct RelFieldPath *rap = ExecuteAllocMem(sizeof(struct RelFieldPath));
    struct Element *elem = StackNth(n);
    struct Record *rec = NULL;

    binin(rap, sizeof(struct RelFieldPath));

    if (elem->el_Type == TYPE_NIL) {
        rec = NULL;
    } else if (elem->el_Type != TYPE_RECORD) {
        ExecuteAbort(GCS(MSG_ProgErr_Type));
    } else if(elem->el_Contents.tabrec.tab != rap->rap_Table) {
        ExecuteAbort(GCS(MSG_ProgErr_Type));
    } else {
        rec = elem->el_Contents.tabrec.rec;
    }
    PushField(rap, rec, TRUE);
}

static void info(void)
{
#ifdef CHECK_RECORD_HEADERS
    if (!CheckRecordHeaders("MBlisp")) {
        ExecuteAbort("Record headers corrupted!");
    }
#endif
    executeLineCol.Line = bininshort();
    executeLineCol.Col = bininshort();
    executeLineCol.Filename = ecd.buffer + ecd.index;
    ecd.index += strlen(ecd.buffer + ecd.index) + 1;
}

static void executeLoopHelp(void)
{
    unsigned char tick = 0;

    for (;;) {
        if (tick++ == 0) {    /* ensure this is called independent of length */
            HeapTick();       /* for garbage collection */
	}
        switch (binintag()) {
            case BIN_MKINTEGER:     mkinteger();    break;
            case BIN_MKREAL:        mkreal();       break;
            case BIN_MKSTRING:      mkstring();     break;
            case BIN_MKDATE:        mkdate();       break;
            case BIN_MKTIME:        maktime();      break;
            case BIN_MKTRUE:        mktrue();       break;
            case BIN_MKNIL:         mknil();        break;
            case BIN_GETGLOBALDATA: getglobaldata(); break;
            case BIN_GETLOCALDATA:  getlocaldata(); break;
            case BIN_GETLOCALEXTEND: getlocalextend(); break;
            case BIN_CALLFUNC:      callfunc();     break;
            case BIN_PRESFW:        presfw();       break;
            case BIN_SFW:           sfw();          break;
            case BIN_FORALL_JMP:    foralljmp();    break;
            case BIN_POSTSFW:       postsfw();      break;
            case BIN_POP:           pop();          break;
            case BIN_POPN:          popn();         break;
            case BIN_POP1N:         pop1n();        break;
            case BIN_SETGLOBALIDENT: setglobalident(); break;
            case BIN_SETLOCALIDENT: setlocalident(); break;
            case BIN_SETLOCALEXTEND: setlocalextend(); break;
            case BIN_PUSHTOP:       pushTop();      break;
            case BIN_SETQLIST_NEXT: setqListNext(); break;
            case BIN_SETQLIST_END:  setqListEnd();  break;
            case BIN_JMPAND:        jmpand();       break;
            case BIN_JMPOR:         jmpor();        break;
            case BIN_JMPNIL:        jmpnil();       break;
            case BIN_JMPTRUE:       jmptrue();      break;
            case BIN_JMPGLOBALVAR:  jmpglobalvar(); break;
            case BIN_JMP:           jmp();          break;
            case BIN_POPJMP:        popjmp();       break;
            case BIN_POP1JMP:       pop1jmp();      break;
            case BIN_JSR:           jsr();          break;
            case BIN_RTS:           rts();          break;
            case BIN_DOTIMES_INIT:  dotimes_init(); break;
            case BIN_DOTIMES_STEP:  dotimes_step(); break;
            case BIN_DOTIMES_TEST:  dotimes_test(); break;
            case BIN_DOLIST_INIT:   dolist_init();  break;
            case BIN_DOLIST_STEP:   dolist_step();  break;
            case BIN_DOLIST_TEST:   dolist_test();  break;
            case BIN_CASECMP:       casecmp();      break;
            case BIN_INITRUNTRIGGER: initruntrigger(); break;
            case BIN_EXITRUNTRIGGER: exitruntrigger(); break;
            case BIN_INITGETVIRTUAL: initgetvirtual(); break;
            case BIN_INITGETVIRTUALEXTEND: initgetvirtualextend(); break;
            case BIN_EXITGETVIRTUAL: exitgetvirtual(); break;
            case BIN_INITGETENABLED: initgetenabled(); break;
            case BIN_EXITGETENABLED: exitgetenabled(); break;
            case BIN_INITGETTITLE:  initgettitle(); break;
            case BIN_EXITGETTITLE:  exitgettitle(); break;
            case BIN_CONST:         constinfo(); break;
            case BIN_FUNCINFO:      funcinfo();    break;
            case BIN_USERFUNCTION:  userfunction(); break;
            case BIN_TABLE:      pushtable(); break;
            case BIN_FIELD:     pushfield(); break;
            case BIN_FIELDEXTEND: pushfieldextend(); break;
            case BIN_INFO:          info();         break;
            case BIN_END:           return;
            default:    ExecuteAbort(GCS(MSG_Err_Internal)); break;
        }
    }
}

static void executeLoop(void)
{
    executeCallCount++;
    if (executeCallCount > MAX_EXECUTE_CALL_COUNT) {
  	ExecuteAbort(GCS(MSG_ProgErr_Recursion));
    }
    executeLoopHelp();
    executeCallCount--;
}

static void adjustArgs(short arity, short numArgs)
{
    for (; numArgs > arity; numArgs--) {
        StackPop();                             /* remove argument */
    }
    for (; numArgs < arity; numArgs++) {
        PushNil();                              /* complete with NIL args */
    }
}

void CallUserFunction(struct UserFunction *uf)
{
    PushCodeAddress(ecd.buffer, ecd.index);   	/* push return address */
    ecd.buffer = uf->uf_Bin->bin_Buffer;         /* jump to sub program */
    ecd.index = uf->uf_SubProgram.sp_Index;
}

void CallUserFunctionAdjustArgs(struct UserFunction *uf, short numArgs)
{
    adjustArgs(uf->uf_Arity, numArgs);
    CallUserFunction(uf);
}

void ExecuteUserFunction(struct UserFunction *uf)
{
    static char newBinBuffer[] = { BIN_END, };
    const char *oldBinBuffer = ecd.buffer;
    long oldBinIndex = ecd.index;

    ecd.buffer = newBinBuffer;
    ecd.index = 0;
    CallUserFunction(uf);
    executeLoop();
    ecd.buffer = oldBinBuffer;
    ecd.index = oldBinIndex;

    HeapTick();         /* for garbage collection */
}

void ExecuteUserFunctionAdjustArgs(struct UserFunction *uf, short numArgs)
{
    adjustArgs(uf->uf_Arity, numArgs);
    ExecuteUserFunction(uf);
}

void ExecuteEvalSubProgram(struct SubProgram *sp)
{
    const char *oldBinBuffer = ecd.buffer;
    long oldBinIndex = ecd.index;

    if (sp == NULL) {
        ExecuteAbort(GCS(MSG_Err_Internal));
    }
    ecd.buffer = sp->sp_Bin->bin_Buffer;
    ecd.index = sp->sp_Index;
    executeLoop();
    ecd.buffer = oldBinBuffer;
    ecd.index = oldBinIndex;
}

const char *ExecuteEvalBin(struct Bin *b, struct Element **pelem)
{
    struct Element *elem = BinCall(b, NULL, NULL, TRUE);
    const char *error = BinError(NULL, NULL, NULL);

    if (pelem) {
        *pelem = elem;
    }
    return error;
}

void BinBegin(struct Project *pr, struct Bin *b, bool needDefVars)
{
    struct Bin *projectBin = ProgramBin(ProjectProgram(pr));
    struct DisplayDependencies *dd;
    int ret;
   
    executeCallCount = 0;
    ecd.buffer = b? b->bin_Buffer : NULL;
    ecd.index = 0;
    executeProject = pr;
    executeProjectChange = FALSE;
    executeAllowSideEffects = TRUE;
    executeError = NULL;
    executeLineCol.Line = -1;
    executeLineCol.Col = 1;
    executeLineCol.Filename = NULL;
    executeMemPool = CreatePool(0, 4096, 1024);
    NewList(&executeVirtualFieldComputeCache);
    InitStack();
    InitHeap();
    executeBreakSettings.title = GCS(MSG_ExecutingProgram);
    executeBreakSettings.steps = 100;
    executeBreakSettings.current = 0;
    executeBreakSettings.abort = TRUE;
    executeBreakSettings.pause = TRUE;
    executeBreakSettings.needGui = ProjectHasGui(pr);
    BreakInit(&executeBreakSettings);
    InitFileIO(pr);
    
    if (pr == NULL) {
       	executeError = GCS(MSG_Err_Internal);
	return;
    }

    if (projectBin) {
	FORNODE(dd, &projectBin->bin_DisplayDependenciesList) {
	    dd->dd_Visited = FALSE;		/* mark as updated */
	}
    }
    
    programRunning++;
    ret = setjmp(ecd.jmpBuf);
    if (ret == 0) {
       	PushGlobalVars();
        needDefVars = TRUE;  /* currently we always need since we always save */
	if (needDefVars && projectBin) {
	    /* push user defined global variables */
	    ecd.buffer = projectBin->bin_Buffer;
	    ecd.index = 0;
	    executeLoop();
	    ecd.index--;                                 /* go to last END */
	}
    }
    programRunning--;
}

struct Element *BinCall(struct Bin *b, struct UserFunction *uf,
        struct Element **args, bool allowSideEffects)
{
    int i;
    struct Element *elem = NULL;
    struct ExecuteControlData save = ecd;
    bool oldExecuteAllowSideEffects = executeAllowSideEffects; 
    long old_size = StackSize(), new_size;
    
    if (executeError != NULL) {
        return NULL;
    }

    executeAllowSideEffects = allowSideEffects;
    programRunning++;
    if (setjmp(ecd.jmpBuf) == 0) {
        if (uf != NULL) {
            /* push function arguments */
            for (i=0; i < uf->uf_Arity && args && args[i]; i++) {
                StackPush(args[i]);              /* push arguments */
            }
            for (; i < uf->uf_Arity; i++) {
                PushNil();                 	    /* complete with NIL args */
            }
            ExecuteUserFunction(uf);       	    /* synchronous call */
        } else if (b) {
            ecd.buffer = b->bin_Buffer;
            ecd.index = 0;
            executeLoop();
        } else {
            ExecuteAbort(GCS(MSG_Err_Internal));    /* no use case so far */
        }
        elem = StackPop();		            /* our result */
    }
    programRunning--;

    executeAllowSideEffects = oldExecuteAllowSideEffects;
    ecd = save;
    new_size = StackSize();
    if (new_size > old_size) {
        StackSqueeze(0, new_size - old_size);
    }
    
    if (executeError != NULL) {
	return NULL;
    }
    return elem;
}

void FreeElement(struct MemPool *memPool, struct Element *elem)
{
    if (elem) {
        struct Element *next;

        while (elem->el_Type == TYPE_LIST) {
            FreeElement(memPool, elem->el_Contents.list.head);
            next = elem->el_Contents.list.tail;
            PoolFreeMem(memPool, elem, sizeof(struct Element));
            elem = next;
        } 
        
        FreeElementContents(memPool, elem);
        PoolFreeMem(memPool, elem, sizeof(struct Element));
    }
}

static struct Element *copyElement(struct MemPool *memPool,struct Element *elem)
{
    struct Element *copy = NULL;
    struct Element *list, *head, *tail;
    char *str;
    char *mi;
    struct RelFieldPath *rap;

    if (elem) {
        copy = PoolAllocMem(memPool, sizeof(struct Element));
        if (copy == NULL) {
            return NULL;
        }
        *copy = *elem;

        list = copy;
        while (elem->el_Type == TYPE_LIST) {
            head = copyElement(memPool, elem->el_Contents.list.head);
            if (head == NULL) {
                list->el_Type = TYPE_NIL;       // not a list
                FreeElement(memPool, copy);
                return NULL;
            }
            tail = PoolAllocMem(memPool, sizeof(struct Element));
            if (tail == NULL) {
                FreeElement(memPool, head);
                list->el_Type = TYPE_NIL;       // not a list
                FreeElement(memPool, copy);
                return NULL;
            }
            list->el_Contents.list.head = head;
            list->el_Contents.list.tail = tail;
            elem = elem->el_Contents.list.tail;
            list = tail;
            *list = *elem;
        }

        /* 
        ** we always do a deep copy as for example a static string could be
        ** inside a temprary program, e.g. a query.
        */
        switch (elem->el_Type) {
            case TYPE_STRING:
                str = AllocString(memPool, elem->el_Contents.string.string);
                if (str == NULL) {
                    list->el_Allocated = FALSE;
                    FreeElement(memPool, copy);
                    return NULL;
                }
                list->el_Contents.string.string = str;
                list->el_Allocated = TRUE;
                break;

            case TYPE_MEMO:
                mi = AllocString(memPool, elem->el_Contents.mi);
                if (mi == NULL) {
                    list->el_Allocated = FALSE;
                    FreeElement(memPool, copy);
                    return NULL;
                }
                list->el_Contents.mi = mi;
                list->el_Allocated = TRUE;
                break;

            case TYPE_FIELD:
                rap = PoolAllocMem(memPool, sizeof(struct RelFieldPath));
                if (rap == NULL) {
                    list->el_Allocated = FALSE;
                    FreeElement(memPool, copy);
                    return NULL;
                }
                *rap = *(elem->el_Contents.field.rap);
                list->el_Contents.field.rap = rap;
                list->el_Allocated = TRUE;
                break;
        }
    }

    return copy;
}

const static char *storeUserGlobalVars(struct Bin *bin)
{
    if (bin) {
        struct UserGlobalVar *ugv;
        FORNODE(ugv, &bin->bin_UserGlobalVarList) {
            struct Element *elem = StackNth(ugv->ugv_Stack);
            if (elem != ugv->ugv_Element) {
                ugv->ugv_Garbage = ugv->ugv_Element;
                ugv->ugv_Element = copyElement(bin->bin_MemPool, elem);
                if (ugv->ugv_Element == NULL) {
                    return GCS(MSG_Err_Mem);
                }
            } else {
                ugv->ugv_Garbage = NULL;
            }
        }
        FORNODE(ugv, &bin->bin_UserGlobalVarList) {
            FreeElement(bin->bin_MemPool, ugv->ugv_Garbage);
            ugv->ugv_Garbage = NULL;
        }
    }
    return NULL;
}

void ProgramBinRemTable(struct Bin *bin, struct Table *tab)
{
    if (bin && tab) {
        struct UserGlobalVar *ugv;
        FORNODE(ugv, &bin->bin_UserGlobalVarList) {
            ElementRemTable(ugv->ugv_Element, tab);
        }
    }
    /* 
     * we do not need to check for dependencies on the table since
     * it is not possible to keep a program referencing a table that is
     * about to be removed.
     */
}

void ProgramBinRemRecord(struct Bin *bin, struct Record *rec)
{
    if (bin && rec) {
        struct UserGlobalVar *ugv;
        FORNODE(ugv, &bin->bin_UserGlobalVarList) {
            ElementRemRecord(ugv->ugv_Element, rec);
        }
    }
}

void ProgramBinRemAllRecords(struct Bin *bin)
{
    if (bin) {
        struct UserGlobalVar *ugv;
        FORNODE(ugv, &bin->bin_UserGlobalVarList) {
            ElementRemAllRecords(ugv->ugv_Element);
        }
    }
}

static bool programBinRemDependency(struct Bin *bin, void *p)
{
    struct DisplayDependencies *dd;
    struct Dependency *dep;
    bool recompile = FALSE;

    if (bin && p) {
	FORNODE(dd, &bin->bin_DisplayDependenciesList) {
	    if (dd->dd_Object == p) {
		dd->dd_Object = NULL;
		recompile = TRUE;
	    }
	    FORNODE(dep, &dd->dd_Dependencies) {
		if (dep->dep_Pointer == p) {
		    dep->dep_Pointer = NULL;
		    recompile = TRUE;
		}
	    }
	}
    }
    return recompile;
}

bool ProgramBinRemField(struct Bin *bin, struct Field *field)
{
    struct UserGlobalVar *ugv;

    if (bin && field) {
        FORNODE(ugv, &bin->bin_UserGlobalVarList) {
            ElementRemField(ugv->ugv_Element, field);
        }
    }
    return programBinRemDependency(bin, field);
}

bool ProgramBinRemDisplayObject(struct Bin *bin, struct DisplayObject *dio)
{
    return programBinRemDependency(bin, dio);
}

static void updateDependentDisplayObject(struct DisplayDependencies *dd)
{
    struct DisplayObject *dio = dd->dd_Object;
    bool update = FALSE;
    struct Bin *bin;

    switch (dd->dd_Type) {
        case DD_TYPE_VIRTUALFIELD:
            if (DisplayFieldAutoUpdate(dio) && 
                    GuiDisplayObjectVisible(dio)) {
                update = TRUE;
            }
            break;
        case DD_TYPE_ENABLED:
            if (DisplayObjectEnabled(dio) == DOD_ENABLED_COMPUTED &&
                    (GuiDisplayObjectVisible(dio) ||
                     (DisplayObjectType(dio) == DO_TYPE_WINDOW &&
                      DisplayWindowCloseWhenDisabled(dio) &&
                      (GuiWindowOpen(DisplayObjectGuiObject(dio)) ||
                       DisplayWindowOpen(dio))))) {
                update = TRUE;
            }
            break;
        case DD_TYPE_TITLE:
            update = TRUE;
            break;
    }

    if (update && executeError == NULL) {
	bin = TriggerToBin(dio, dd->dd_Type);
        if (bin == NULL) {
            executeError = GCS(MSG_Err_Mem);
        } else {
            /* run the display object's program */
            BinCall(bin, NULL, NULL, TRUE);
            if (programRunning && executeError != NULL) {
                ExecuteAbort(executeError);
            }
            FreeBin(bin);
        }
    }
}

static void updateDependentDisplayObjectsNow(void)
{
    struct Project *pr = executeProject;
    struct Program *prg = ProjectProgram(pr);
    struct Bin *bin = ProgramBin(prg);
    struct DisplayDependencies *dd;
    struct VirtualFieldComputeCacheEntry *entry;
    bool change = FALSE;
    bool updated = FALSE;

    if (bin != NULL) {
        if (!programRunning) {
            /* make sure the right record pointers are used */
            SetTableProgramRecords(executeProject);
        }
        /* 
        ** We need to re-compute all virtual fields marked for update
        ** and then perform the GUI update.  The latter can in turn
        ** mark further virtual fields for updates, e.g. when setting
        ** a virtual list, the active line is set, which in turn can
        ** cause the enabled state of a button to change.
        */
        do {
            updated = FALSE;
            do {
                change = FALSE;
                FORNODE(dd, &bin->bin_DisplayDependenciesList) {
                    if (dd->dd_Visited) {
                        updateDependentDisplayObject(dd);
                        dd->dd_Visited = FALSE;
                        change = TRUE;
                        updated = TRUE;
                    }
                }
            } while (change);

            while ((entry = (void *)REMHEAD(&executeVirtualFieldComputeCache))
                    != NULL) {
                struct Field *field = entry->vfcce_Field;
                struct Record *rec = entry->vfcce_Record;
                struct Table *tab = FieldTable(field);

                if (TableGuiRecord(tab) == rec) {
                    GuiSetVirtualField(field, rec, entry->vfcce_Element);
                    updated = TRUE;
                }
                PoolFreeMem(executeMemPool, entry,
                        sizeof(struct VirtualFieldComputeCacheEntry));
            }
        } while (updated);
    }
}

void BinEnd(void)
{
    if (executeError == NULL) {
        struct Program *prg = ProjectProgram(executeProject);
        struct Bin *bin = ProgramBin(prg);

    	if (executeProjectChange) {
  	    struct UserFunction *uf = FindUserFunctionName(prg, onChange);
	    
	    if (uf != NULL) {
		BinCall(bin, uf, NULL, TRUE);
	    }
	}
	if (executeError == NULL && bin != NULL) {
	    /* update dependent display objects */
            updateDependentDisplayObjectsNow();
	}
	if (executeError == NULL) {
	    executeError = storeUserGlobalVars(bin);
	}
    }
    BreakExit(&executeBreakSettings);
    ExitFileIO();
    ExitHeap();
    
    if (executeMemPool != NULL) {
        DeletePool(executeMemPool);
        executeMemPool = NULL;
    }
    
    if (executeProjectChange) {
	ProjectChange(executeProject, FALSE);
    }
    
    executeProject = NULL;
    ecd.buffer = NULL;
}

const char *BinError(long *line, long *col, const char **filename)
{
    if (line) {
	*line = executeLineCol.Line;
    }
    if (col) {
	*col = executeLineCol.Col;
    }
    if (filename) {
       	const char *s = executeLineCol.Filename;
	
	*filename = (s != NULL && *s != 0)? s : NULL;
    }

    return executeError;
}

const char *ExecuteBin(struct Project *pr, struct Bin *b, bool needDefVars,
        struct UserFunction *uf, struct Element **args,
        bool allowSideEffects,
	const char *(*func)(struct Element *elem),
	long *line, long *col, const char **filename)
{
    const char *error, *func_error = NULL;
    struct Element *elem;
    
    BinBegin(pr, b, needDefVars);
    elem = BinCall(b, uf, args, allowSideEffects);
    if (func != NULL) {
	func_error = func(elem);
    }
    BinEnd();
    error = BinError(line, col, filename);
    if (error == NULL) {
	error = func_error;
    }
    
    return error;
}

static void initError(long *line, long *col, const char **filename)
{
    if (line) {
    	*line = 1;
    }
    if (col) {
    	*col = 1;
    }
    if (filename) {
    	*filename = NULL;
    }
}

static void computeDependentDisplayObjectsRec(struct List *displayDeps, void *p,
        bool (*accept)(struct DisplayObject *dio, void *ptr), void *ptr)
{
    struct DisplayDependencies *dd;
    struct Dependency *dep;
    struct Field *field;
    struct Table *tab;

    FORNODE(dd, displayDeps) {
	if (!dd->dd_Visited) {
	    FORNODE(dep, &dd->dd_Dependencies) {
		if (dep->dep_Pointer == p) {
                    if (accept == NULL || accept(dd->dd_Object, ptr)) {
                        dd->dd_Visited = TRUE;
                    }
                    if (dd->dd_Type == DD_TYPE_VIRTUALFIELD) {
                        /*
                        ** No recursion for GuiRecord of virtual fields
                        ** inside the same table since each dependent 
                        ** maintains its own dependency for this GuiRecord
                        */
                        field = DisplayFieldField(dd->dd_Object);
                        tab = FieldTable(field);

                        if (dep->dep_Type != DEP_TYPE_GUIRECORD ||
                                dep->dep_Pointer != TableDisplayObject(tab)) {
                            computeDependentDisplayObjectsRec(displayDeps, field,
                                    accept, ptr);
                        }
                    }
                    break;
		}
	    }
	}
    }
}

void ComputeDependentDisplayObjects(struct Project *pr, void *p,
        bool (*accept)(struct DisplayObject *dio, void *ptr), void *ptr)
{
    struct Program *prg = ProjectProgram(pr);
    struct Bin *bin = ProgramBin(prg);

    if (bin) {
	computeDependentDisplayObjectsRec(
                &bin->bin_DisplayDependenciesList, p, accept, ptr);
    }
}

const char *UpdateDependentDisplayObjects(struct Project *pr, void *p,
        bool (*accept)(struct DisplayObject *dio, void *ptr), void *ptr,
        long *line, long *col, const char **filename)
{
    struct Program *prg = ProjectProgram(pr);
    struct Bin *bin = ProgramBin(prg);
    const char *error = NULL;

    initError(line, col, filename);
    if (bin == NULL) {
        return NULL;
    }

    if (executeProject != NULL) {
        if (pr != executeProject) {
            return GCS(MSG_Err_Internal);
        }
	computeDependentDisplayObjectsRec(
                &bin->bin_DisplayDependenciesList, p, accept, ptr);
    } else {
        struct DisplayDependencies *dd;
        struct Dependency *dep;
        bool needUpdate = FALSE;

	FORNODE(dd, &bin->bin_DisplayDependenciesList) {
	    FORNODE(dep, &dd->dd_Dependencies) {
		if (dep->dep_Pointer == p) {
		    needUpdate = TRUE;
		    break;
		}
	    }
	}
	if (needUpdate) {
	    SetTableProgramRecords(pr);
	    BinBegin(pr, NULL, TRUE);	/* clears visited flags */
	    computeDependentDisplayObjectsRec(
                    &bin->bin_DisplayDependenciesList, p, accept, ptr);
            BinEnd();                   /* updates objects */
	    error = BinError(line, col, filename);	
	}
    }
    return error;
}

bool ProgramRunning(void)
{
    return executeProject != NULL;
}

void ProgramFreeResources(void)
{
    if (executeMemPool) {
	DeletePool(executeMemPool);
        executeMemPool = NULL;
    }
}

const char *UserGlobalVarName(struct UserGlobalVar *ugv)
{
    const char *name = NULL;

    if (ugv) {
	name = ugv->ugv_Name;
    }
    return name;
}

struct List *BinUserFunctions(struct Bin *bin)
{
    struct List *uflist = NULL;

    if (bin) {
        uflist = &bin->bin_UserFunctionList;
    }
    return uflist;
}

struct UserFunction *BinFindUserFunctionName(struct Bin *bin, const char *name)
{
    struct UserFunction *uf;

    if (bin && name) {
        FORNODE(uf, &bin->bin_UserFunctionList) {
            if (!strcmp(name, uf->uf_Name)) {
                return uf;
            }
        }
    }
    return NULL;
}

const char *UserFunctionName(struct UserFunction *uf)
{
    const char *name = NULL;

    if (uf) {
        name = uf->uf_Name;
    }
    return name;
}

char UserFunctionType(struct UserFunction *uf)
{
    char type = NOTYPE;

    if (uf) {
        type = uf->uf_Type;
    }
    return type;
}

short UserFunctionArity(struct UserFunction *uf)
{
    short arity = 0;

    if (uf) {
        arity = uf->uf_Arity;
    }
    return arity;
}

void UserFunctionPosition(struct UserFunction *uf, long *line, long *col)
{
    if (line != NULL) {
	*line = (uf != NULL && uf->uf_LineCol.Filename == NULL)?
	    uf->uf_LineCol.Line : -1;
    }
    if (col != NULL) {
	*col = (uf != NULL && uf->uf_LineCol.Filename == NULL)?
	    uf->uf_LineCol.Col : 1;
    }
}

char *UserFunctionTypes(struct UserFunction *uf)
{
    char *types = NULL;

    if (uf) {
        types = uf->uf_Types;
    }
    return types;
}

bool UserFunctionTrigger(struct UserFunction *uf)
{
    bool trigger = FALSE;

    if (uf) {
        trigger = uf->uf_Trigger;
    }
    return trigger;
}
 
struct List *BinDisplayDependenciesVirtualField(struct Bin *bin,
        struct DisplayObject *dio)
{
    struct List *deps = NULL;
    struct DisplayDependencies *dd;

    if (bin) {
	FORNODE(dd, &bin->bin_DisplayDependenciesList) {
	    if (dd->dd_Object == dio && dd->dd_Type == DD_TYPE_VIRTUALFIELD) {
		deps = &dd->dd_Dependencies;
		break;
	    }
	}
    }
    return deps;
}

struct UserGlobalVar *DependencyUserGlobalVar(struct Dependency *dep)
{
    struct UserGlobalVar *ugv = NULL;

    if (dep && dep->dep_Type == DEP_TYPE_USERGLOBALVAR) {
	ugv = dep->dep_Pointer;
    }
    return ugv;
}

struct Field *DependencyField(struct Dependency *dep)
{
    struct Field *field = NULL;

    if (dep && dep->dep_Type == DEP_TYPE_FIELD) {
	field = dep->dep_Pointer;
    }
    return field;
}

struct DisplayObject *DependencyGuiObject(struct Dependency *dep)
{
    struct DisplayObject *dio = NULL;

    if (dep && dep->dep_Type == DEP_TYPE_GUIOBJECT) {
        dio = dep->dep_Pointer;
    }
    return dio;
}

struct Table *DependencyTable(struct Dependency *dep)
{
    struct Table *tab = NULL;

    if (dep && dep->dep_Type == DEP_TYPE_TABLE) {
	tab = dep->dep_Pointer;
    }
    return tab;
}

struct DisplayObject *DependencyGuiRecord(struct Dependency *dep)
{
    struct DisplayObject *dio = NULL;

    if (dep && dep->dep_Type == DEP_TYPE_GUIRECORD) {
        dio = dep->dep_Pointer;
    }
    return dio;
}

