/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** programbingen.c
**
** Creation date: 20.04.1995
** Last modified: 02.05.2022
*/

#include "programbin_impl.h"
#include "func.h"
#include "funcsfw.h"
#include "functable.h"
#include "display.h"
#include "table.h"
#include "field.h"
#include "memory.h"
#include "util.h"
#include <setjmp.h>

#define BIN_BLOCK_SIZE 2048

static void FreeUserFunction(struct MemPool *memPool, struct UserFunction *uf)
{
    if(uf)
    {
        FreeString(memPool, uf->uf_Name);
        PoolFreeMem(memPool, uf->uf_Types, uf->uf_Arity * sizeof(char));
        PoolFreeMem(memPool, uf, sizeof(struct UserFunction));
    }
}

static struct UserFunction *AllocUserFunction(struct MemPool *memPool, 
	struct LocalIdent *li)
{
    struct UserFunction *uf;

    uf = PoolAllocMemZero(memPool, sizeof(struct UserFunction));
    if(uf == NULL)
        return(NULL);

    uf->uf_Name = AllocString(memPool, li->Name);
    if(uf->uf_Name == NULL)
    {
        FreeUserFunction(memPool, uf);
        return(NULL);
    }

    uf->uf_Type = li->Type;
    uf->uf_Arity = li->Arity;
    uf->uf_LineCol = li->LineCol;
    if(li->Arity > 0)
    {
        uf->uf_Types = PoolAllocMem(memPool, uf->uf_Arity * sizeof(char));
        if(uf->uf_Types == NULL)
        {
            FreeUserFunction(memPool, uf);
            return(NULL);
        }
    }

    memcpy(uf->uf_Types, li->Types, uf->uf_Arity * sizeof(char));

    return(uf);
}

void FreeUserGlobalVar(struct MemPool *memPool, struct UserGlobalVar *ugv)
{
    if (ugv) {
        FreeElement(memPool, ugv->ugv_Element);
        FreeString(memPool, ugv->ugv_Name);
        PoolFreeMem(memPool, ugv, sizeof(struct UserGlobalVar));
    }
}

static struct UserGlobalVar *AllocUserGlobalVar(struct MemPool *memPool, 
	struct LocalIdent *li)
{
    struct UserGlobalVar *ugv;

    ugv = PoolAllocMemZero(memPool, sizeof(struct UserGlobalVar));
    if (ugv == NULL) {
        return NULL;
    }
    ugv->ugv_Name = AllocString(memPool, li->Name);
    if (ugv->ugv_Name == NULL) {
        FreeUserGlobalVar(memPool, ugv);
        return NULL;
    }
    ugv->ugv_Stack = li->Pos;

    return ugv;
}

static void FreeDependency(struct MemPool *memPool, struct Dependency *dep)
{
    PoolFreeMem(memPool, dep, sizeof(struct Dependency));
}

static struct Dependency *AllocDependency(struct MemPool *memPool, void *p,
	short dtype)
{
    struct Dependency *dep = PoolAllocMem(memPool, sizeof(struct Dependency));
    if (dep == NULL) {
	return NULL;
    }
    dep->dep_Pointer = p;
    dep->dep_Type = dtype;

    return dep;
}

static void FreeDisplayDependencies(struct MemPool *memPool, 
	struct DisplayDependencies *dd)
{
    if (dd) {
	struct Dependency *dep;

	while ((dep = (struct Dependency *)
		    REMHEAD(&dd->dd_Dependencies)) != NULL) {
	    FreeDependency(memPool, dep);
	}

	PoolFreeMem(memPool, dd, sizeof(struct DisplayDependencies));
    }
}

static struct DisplayDependencies *AllocDisplayDependencies(
        struct MemPool *memPool, struct DisplayObject *dio, short type)
{
    struct DisplayDependencies *dd;
   
    dd = PoolAllocMemZero(memPool, sizeof(struct DisplayDependencies));
    if (dd == NULL) {
	return NULL;
    }
    dd->dd_Object = dio;
    dd->dd_Type = type;
    NewList(&dd->dd_Dependencies);

    return dd;
}

static jmp_buf errJmpBuf;
static struct Bin *out = NULL;
static struct Project *binProject = NULL;
static struct List *binDisplayDependencies = NULL;
static struct Func *binLoopControl = NULL;
static struct Func *binFunctionControl = NULL;

static void binout(const void *buf, long size)
{
    if (out->bin_Len + size > out->bin_Size) {     /* overflow? */
        long newsize = out->bin_Size + MAX(BIN_BLOCK_SIZE, size);
        char *new = PoolReAllocMem(out->bin_MemPool, out->bin_Buffer, 
		out->bin_Size, newsize);
        if (new == NULL) {
            longjmp(errJmpBuf, 1);               /* return to main procedure */
        }
        out->bin_Buffer = new;
        out->bin_Size = newsize;
    }
    memcpy(out->bin_Buffer + out->bin_Len, buf, size);
    out->bin_Len += size;
}

static void __inline binouttag(char c)    { binout(&c, sizeof(char)); }
static void __inline binoutchar(char c)   { binout(&c, sizeof(char)); }
static void __inline binoutshort(short s) { binout(&s, sizeof(short)); }
static void __inline binoutlong(long l)   { binout(&l, sizeof(long)); }
static void __inline binoutdouble(double d) { binout(&d, sizeof(double)); }
static void __inline binoutstring(const char *s) { if(s) binout(s, strlen(s)); }
static void __inline binoutptr(const void *ptr) { binout(&ptr, sizeof(void *));}

static struct DisplayDependencies *findDisplayDependencies(
        struct DisplayObject *dio, short type)
{
    struct DisplayDependencies *dd;

    if (binDisplayDependencies) {
	FORNODE(dd, binDisplayDependencies) {
	    if (dd->dd_Object == dio && dd->dd_Type == type) {
		return dd;
	    }
	}
    }
    return NULL;
}

static void constToBin(struct Const *c)
{
    switch(c->Case)
    {
        case CONST_CASE_NORMAL:
            switch(c->Type)
            {
                case TYPE_INTEGER:
                    binouttag(BIN_MKINTEGER);
                    binoutlong(c->Contents.integer.value);
                    break;

                case TYPE_REAL:
                    binouttag(BIN_MKREAL);
                    binoutdouble(c->Contents.real);
                    break;

                case TYPE_STRING:
                {
                    char *new=StrCompress(out->bin_MemPool, c->Contents.string);

                    if(new == NULL)
                        longjmp(errJmpBuf, 1);    /* return to main procedure */

                    binouttag(BIN_MKSTRING);
                    binoutstring(new);
                    binoutchar(0);
                    FreeString(out->bin_MemPool, new);
                    break;
                }

                case TYPE_DATE:
                    binouttag(BIN_MKDATE);
                    binoutlong(c->Contents.date);
                    break;

                case TYPE_TIME:
                    binouttag(BIN_MKTIME);
                    binoutlong(c->Contents.time);
                    break;
            }
            break;

        case CONST_CASE_PREDEFINED:
            switch(c->Type)
            {
                case TYPE_BOOL:
                    binouttag(BIN_MKTRUE);
                    break;

                case TYPE_NIL:
                    binouttag(BIN_MKNIL);
                    break;

                default:
                    binouttag(BIN_CONST);
                    binoutptr(&c->Contents.ConstInfo->ci_Element);
                    break;
            }
            break;
    }
}

static void globalIdentToBin(struct GlobalIdent *gi, short type)
{
    struct RelFieldPath *rap = &gi->RelFieldPath;
    struct Field *field = RelFieldPathField(rap);
    struct DisplayObject *dio = FieldDisplayObject(field);

    if(type == TYPE_FIELD)
    {
        binouttag(BIN_FIELD);
        binout(rap, sizeof(struct RelFieldPath));
    }
    else if(type == TYPE_TABLE)
    {
        binouttag(BIN_TABLE);
        binoutptr(rap->rap_Table);
        binoutchar(rap->rap_Star);
    }
    else if(field != NULL && FieldType(field) == TYPE_VIRTUAL)
    {
        binouttag(BIN_INITGETVIRTUAL);
        binout(rap, sizeof(struct RelFieldPath));
        binouttag(BIN_EXITGETVIRTUAL);
        binoutptr(field);
	binoutptr(findDisplayDependencies(dio, DD_TYPE_VIRTUALFIELD));
    }
    else
    {
        binouttag(BIN_GETGLOBALDATA);
        binout(rap, sizeof(struct RelFieldPath));
    }
}

static void localIdentToBin(struct LocalIdent *li, short d)
{
    if(li->Arity < 0)                   /* variable? */
    {
        binouttag(BIN_GETLOCALDATA);
        binoutshort((li->Pos > 0)? d - li->Pos : li->Pos);
    }
    else
    {
        binouttag(BIN_USERFUNCTION);
        binoutptr(li->UserFunction);
    }
}

static void infoToBin(struct LineCol *lc)
{
    if(out->bin_HasDebugInfo && lc != NULL)
    {
        binouttag(BIN_INFO);
        binoutshort(lc->Line);
        binoutshort(lc->Col);
        binoutstring(lc->Filename);
        binoutchar(0);
    }
}

static void localExtendToBin(struct LocalExtend *le, short d, short type, 
	struct LineCol *lc)
{
    short pos = le->LocalIdent->Pos;
    struct RelFieldPath *rap = &le->GlobalIdent->RelFieldPath;
    struct Field *field = RelFieldPathField(rap);
    struct DisplayObject *dio = FieldDisplayObject(field);

    if(type == TYPE_FIELD)
    {
        binouttag(BIN_FIELDEXTEND);
        binoutshort((pos > 0)? d - pos : pos);
        binout(rap, sizeof(struct RelFieldPath));
    }
    else if(field != NULL && FieldType(field) == TYPE_VIRTUAL)
    {
        binouttag(BIN_INITGETVIRTUALEXTEND);
        binoutshort((pos > 0)? d - pos : pos);
        binout(rap, sizeof(struct RelFieldPath));
        binouttag(BIN_EXITGETVIRTUAL);
        binoutptr(field);
	binoutptr(findDisplayDependencies(dio, DD_TYPE_VIRTUALFIELD));
    }
    else
    {
        infoToBin(lc);
        binouttag(BIN_GETLOCALEXTEND);
        binoutshort((pos > 0)? d - pos : pos);
        binout(&le->GlobalIdent->RelFieldPath, sizeof(struct RelFieldPath));
    }
}

static void exprToBin(struct Expr *expr, short d);

static void andToBin(struct AppArgs *aa, short d)
{
    if(aa == NULL)
        binouttag(BIN_MKTRUE);
    else
    {
        exprToBin(aa->Expr, d);
        if(aa->Next)
        {
            long index;

            binouttag(BIN_JMPAND);
            index = out->bin_Len;           /* remember current position */
            binoutlong(0);
            andToBin(aa->Next, d);
            memcpy(out->bin_Buffer + index, &out->bin_Len, sizeof(long));
        }
    }
}

static void orToBin(struct AppArgs *aa, short d)
{
    if(aa == NULL)
        binouttag(BIN_MKNIL);
    else
    {
        exprToBin(aa->Expr, d);
        if(aa->Next)
        {
            long index;

            binouttag(BIN_JMPOR);
            index = out->bin_Len;           /* remember current position */
            binoutlong(0);
            orToBin(aa->Next, d);
            memcpy(out->bin_Buffer + index, &out->bin_Len, sizeof(long));
        }
    }
}

static void ifToBin(struct AppArgs *aa, short d)
{
    exprToBin(aa->Expr, d);
    aa = aa->Next;
    if(aa)                      /* then part */
    {
        long index1, index2;

        binouttag(BIN_JMPNIL);
        index1 = out->bin_Len;           /* remember current position */
        binoutlong(0);
        exprToBin(aa->Expr, d);
        binouttag(BIN_JMP);
        index2 = out->bin_Len;
        binoutlong(0);
        memcpy(out->bin_Buffer + index1, &out->bin_Len, sizeof(long));
        if(aa->Next)
            exprToBin(aa->Next->Expr, d);
        else
            binouttag(BIN_MKNIL);
        memcpy(out->bin_Buffer + index2, &out->bin_Len, sizeof(long));
    }
}

static void appArgsToBin(struct AppArgs *aa, short d)
{
    for(; aa; aa = aa->Next)
        exprToBin(aa->Expr, d++);
}

static void addProgN(short num)
{
    switch(num)
    {
        case 0: binouttag(BIN_MKNIL);   break;
        case 1: break;
        default:
        {
            struct funcBin fb;

            binouttag(BIN_CALLFUNC);
            fb.fb_Code = ProgProgN;
            fb.fb_NumArgs = num;
	    fb.fb_HasSideEffect = FALSE;
            binout(&fb, sizeof(struct funcBin));
            break;
        }
    }
}

static void displayDependenciesToBinRec(void *p)
{
    struct DisplayDependencies *dd;
    struct Dependency *dep;

    FORNODE(dd, binDisplayDependencies) {
	if (!dd->dd_Visited) {
	    FORNODE(dep, &dd->dd_Dependencies) {
		if (dep->dep_Pointer == p) {
		    if (dd->dd_Type != DD_TYPE_VIRTUALFIELD ||
                            DisplayFieldAutoUpdate(dd->dd_Object)) {
			binoutptr(dd);
		    }
		    dd->dd_Visited = TRUE;
                    if (dd->dd_Type == DD_TYPE_VIRTUALFIELD) {
                        displayDependenciesToBinRec(
                                DisplayFieldField(dd->dd_Object));
                    }
		    break;
		}
	    }
	}
    }
}

static void displayDependenciesToBin(void *p)
{
    if (p && binDisplayDependencies) {
	struct DisplayDependencies *dd;

	FORNODE(dd, binDisplayDependencies) {
	    dd->dd_Visited = FALSE;
	}
	displayDependenciesToBinRec(p);
    }
    binoutptr(NULL);		/* mark end of dependencies */
}

static void setqToBinHelp(struct Expr *expr, bool star, short d)
{
    struct LineCol *lc = &expr->LineCol;
    struct LocalIdent *li;
    struct RelFieldPath *rap = NULL;
    short pos;

    switch(expr->Case)
    {
        case EXPR_CASE_LOCALIDENT:
            binouttag(BIN_SETLOCALIDENT);
            li = expr->Contents.LocalIdent;
            pos = li->Pos;
            binoutshort((pos > 0)? d - pos : pos);
            if (pos < 0)	/* global variable */
                displayDependenciesToBin(li->UserGlobalVar);
            break;

        case EXPR_CASE_GLOBALIDENT:
            infoToBin(lc);
            binouttag(BIN_SETGLOBALIDENT);
            rap = &expr->Contents.GlobalIdent->RelFieldPath;
            binout(rap, sizeof(struct RelFieldPath));
            binoutchar((char)star);
            break;

        case EXPR_CASE_LOCALEXTEND:
            infoToBin(lc);
            binouttag(BIN_SETLOCALEXTEND);
            pos = expr->Contents.LocalExtend->LocalIdent->Pos;
            binoutshort((pos > 0)? d - pos : pos);
            rap = &expr->Contents.LocalExtend->GlobalIdent->RelFieldPath;
            binout(rap, sizeof(struct RelFieldPath));
            binoutchar((char)star);
            break;
    }
    if (rap && rap->rap_NumFields > 0) {
        displayDependenciesToBin(rap->rap_Fields[rap->rap_NumFields-1]);
    }
}

static void setqToBin(struct Func *func, struct AppArgs *aa, short d)
{
    bool star = func->FuncInfo->fi_Id == FUNCID_SETQSTAR;
    short num = CountAppArgs(aa) / 2;

    for (; aa; aa = aa->Next->Next) {
        exprToBin(aa->Next->Expr, d++);
        setqToBinHelp(aa->Expr, star, d);
    }
    addProgN(num);
}

static void setqListToBin(struct Func *func, struct AppArgs *aa, short d)
{
    bool star = func->FuncInfo->fi_Id == FUNCID_SETQLISTSTAR;
    struct AppArgs *last = aa;

    while (last->Next) {
        last = last->Next;
    }
    exprToBin(last->Expr, d++);

    infoToBin(&func->LineCol);
    binouttag(BIN_PUSHTOP);
    d++;
    for(; aa->Next; aa = aa->Next) {
        binouttag(BIN_SETQLIST_NEXT);
        setqToBinHelp(aa->Expr, star, d + 1);
        binouttag(BIN_POP);
    }
    binouttag(BIN_SETQLIST_END);
    d--;
}

static void applyFuncToBin(struct Func *func, short numArgs)
{
    if(func)
    {
        struct funcBin fb;

        infoToBin(&func->LineCol);

        binouttag(BIN_CALLFUNC);
        fb.fb_Code = func->FuncInfo->fi_Code;
        fb.fb_NumArgs = numArgs;
	fb.fb_HasSideEffect = func->FuncInfo->fi_HasSideEffect;
        binout(&fb, sizeof(struct funcBin));
    }
}

static void exitToBin(struct Func *func, struct AppArgs *aa, short d)
{
    appArgsToBin(aa, d);
    addProgN(CountAppArgs(aa));

    binouttag(BIN_POP1JMP);
    func->Depth = d + 1;
    func->Index = out->bin_Len;
    func->Next = binLoopControl;
    binLoopControl = func;
    binoutshort(0);
    binoutlong(0);
}

static void nextToBin(struct Func *func, short d)
{
    binouttag(BIN_POPJMP);
    func->Depth = d;
    func->Index = out->bin_Len;
    func->Next = binLoopControl;
    binLoopControl = func;
    binoutshort(0);
    binoutlong(0);
}

static void returnToBin(struct Func *func, struct AppArgs *aa, short d)
{
    appArgsToBin(aa, d);
    addProgN(CountAppArgs(aa));

    binouttag(BIN_POP1JMP);
    func->Depth = d + 1;
    func->Index = out->bin_Len;
    func->Next = binFunctionControl;
    binFunctionControl = func;
    binoutshort(0);
    binoutlong(0);
}

static void appToBin(struct App *app, short d)
{
    struct AppArgs *aa = app->AppArgs;
    struct Expr *expr = (aa != NULL)? aa->Expr : NULL;

    if(expr != NULL && expr->Case == EXPR_CASE_FUNC)
    {
        aa = aa->Next;
        switch(expr->Contents.Func->FuncInfo->fi_Id)
        {
            case FUNCID_AND:
                andToBin(aa, d);
                break;

            case FUNCID_OR:
                orToBin(aa, d);
                break;

            case FUNCID_IF:
                ifToBin(aa, d);
                break;

            case FUNCID_SETQ:
            case FUNCID_SETQSTAR:
                setqToBin(expr->Contents.Func, aa, d);
                break;

            case FUNCID_SETQLIST:
            case FUNCID_SETQLISTSTAR:
                setqListToBin(expr->Contents.Func, aa, d);
                break;

            case FUNCID_EXIT:
                exitToBin(expr->Contents.Func, aa, d);
                break;

            case FUNCID_NEXT:
                nextToBin(expr->Contents.Func, d);
                break;

            case FUNCID_RETURN:
                returnToBin(expr->Contents.Func, aa, d);
                break;

            default:
                appArgsToBin(aa, d);
                applyFuncToBin(expr->Contents.Func, CountAppArgs(aa));
                break;
        }
    }
    else if(expr != NULL && expr->Case == EXPR_CASE_LOCALIDENT &&
        expr->Contents.LocalIdent->UserFunction != NULL)
    {
        aa = aa->Next;
        appArgsToBin(aa, d);
        binouttag(BIN_JSR);
        binoutptr(expr->Contents.LocalIdent->UserFunction);
    }
    else
    {
        appArgsToBin(aa, d);
        addProgN(CountAppArgs(aa));
    }
}

static void addLoopControl(short d, long nextIndex, long exitIndex)
{
    struct Func *f;

    for(f = binLoopControl; f; f = f->Next)
    {
        short n = f->Depth - d;
        long index = f->Index;

        if(f->FuncInfo->fi_Id == FUNCID_EXIT)
        {
            n--;
            memcpy(out->bin_Buffer + index, &n, sizeof(short));
            index += sizeof(short);
            memcpy(out->bin_Buffer + index, &exitIndex, sizeof(long));
        }
        else if(f->FuncInfo->fi_Id == FUNCID_NEXT)
        {
            memcpy(out->bin_Buffer + index, &n, sizeof(short));
            index += sizeof(short);
            memcpy(out->bin_Buffer + index, &nextIndex, sizeof(long));
        }
    }
}

static void fromListToBin(struct FromList *fl)
{
    for(; fl; fl = fl->Next)
        binoutptr(fl->GlobalIdent->RelFieldPath.rap_Table);
}

static int countFromList(struct FromList *fl)
{
    int i = 0;

    for(; fl; fl = fl->Next)
        i++;
    return(i);
}

static int countFromVarList(struct FromList *fl)
{
    int i = 0;

    for(; fl; fl = fl->Next)
        if(fl->LocalIdentDef)
            i++;
    return(i);
}

static int countSelectList(struct SelectList *sl)
{
    int i;

    for(i=0; sl; sl = sl->Next)
        if(sl->Expr)
            i++;
    return(i);
}

static int countOrderList(struct OrderList *ol)
{
    int i;

    for(i=0; ol; ol = ol->Next)
        i++;
    return(i);
}

static struct Field *extractReferenceMatch(struct Expr *expr, struct Table *tab,
        struct Expr **refExpr, bool *remove)
{
    struct App *app;
    struct AppArgs *aa;
    struct Func *func;

    *refExpr = NULL;
    *remove = FALSE;
    if (expr == NULL || expr->Case != EXPR_CASE_APP) {
        return NULL;
    }

    /* Allow for nested ProgN with single expressions. */
    app = expr->Contents.App;
    aa = app->AppArgs;
    while (app->IsProgN && aa != NULL && aa->Expr->Case == EXPR_CASE_APP &&
            aa->Next == NULL) {
        app = aa->Expr->Contents.App;
        aa = app->AppArgs;
    }
    /* We now need to have a function call */
    if (app->IsProgN || aa == NULL || aa->Expr->Case != EXPR_CASE_FUNC) {
        return NULL;
    }

    func = aa->Expr->Contents.Func;
    if (func->FuncInfo->fi_Id == FUNCID_EQ) {
        struct Expr *exprs[2];
        struct RelFieldPath *raps[2];
        int i;

        exprs[0] = aa->Next->Expr;
        exprs[1] = aa->Next->Next->Expr;
        for (i = 0; i < 2; i++) {
            while (exprs[i]->Case == EXPR_CASE_APP && 
                    exprs[i]->Contents.App->IsProgN && 
                    exprs[i]->Contents.App->AppArgs != NULL &&
                    exprs[i]->Contents.App->AppArgs->Next == NULL) {
                exprs[i] = exprs[i]->Contents.App->AppArgs->Expr;
            }
            if (exprs[i]->Case == EXPR_CASE_GLOBALIDENT) {
                raps[i] = &exprs[i]->Contents.GlobalIdent->RelFieldPath;
            } else if (exprs[i]->Case == EXPR_CASE_LOCALEXTEND) {
                raps[i] =
                    &exprs[i]->Contents.LocalExtend->GlobalIdent->RelFieldPath;
            } else {
                raps[i] = NULL;
            }
        }

        for (i = 0; i < 2; i++) {
            struct Expr *expr1 = exprs[i];
            struct Expr *expr2 = exprs[1 - i];
            struct RelFieldPath *rap1 = raps[i];
            struct RelFieldPath *rap2 = raps[1 - i];

            /* We are only allowing reference fields. */
            if (expr1->Type != TYPE_RECORD) {
                continue;
            }
            if (rap1 == NULL || rap1->rap_Table != tab ||
                    rap1->rap_NumFields != 1) {
                continue;
            }
            if (rap2 != NULL) {
                /* We allow all paths except those that start with the table. */
                if (rap2->rap_Table == tab) {
                    continue;
                }
            } else if (expr2->Case == EXPR_CASE_LOCALIDENT) {
                /* We allow all local identifiers but the table one */
                if (expr2->Contents.LocalIdent->Table == tab) {
                    continue;
                }
            } else {
                /* We don't allow more complicated expressions. */
                continue;
            }
            *refExpr = expr2;
            *remove = TRUE;
            return rap1->rap_Fields[0];
        }
    } else if (func->FuncInfo->fi_Id == FUNCID_AND) {
        struct AppArgs *first = aa;
        struct AppArgs *prev = aa;
        bool removeAppArg = FALSE;

        for (aa = aa->Next; aa != NULL; prev = aa, aa = aa->Next) {
            struct Field *refField =
                extractReferenceMatch(aa->Expr, tab, refExpr, &removeAppArg);

            if (refField != NULL) {
                if (removeAppArg) {
                    prev->Next = aa->Next;
                    if (first->Next == NULL) {
                        *remove = TRUE;         /* AND w/o any arguments */
                    }
                }
                return refField;
            }
        }
    }

    return NULL;
}

static void sfwToBin(struct SelectFromWhere *sfw, short d)
{
    if(sfw)
    {
        struct sfwBin sb;
        struct FromList *fl;
        struct SelectList *sl;
        struct OrderList *ol;
        long sbIndex, sfwIndex, fromIndex, orderIndex;
        struct SFWSelectItem sfwsi;
        struct SFWFromItem sfwfi;
        struct SFWOrderItem sfwoi;
        short numFromItems = countFromList(sfw->FromList);
        short numFromVars = countFromVarList(sfw->FromList);
        const char *s;
        int i;

        /*
        ** pre sfw
        */
        binouttag(BIN_PRESFW);
        binoutshort(numFromItems);
        binoutshort(numFromVars);
        fromListToBin(sfw->FromList);
        d += numFromItems;

        infoToBin(&sfw->LineCol);

        binouttag(BIN_SFW);
        sbIndex = out->bin_Len;             /* remember current position */
        sb.sb_IsSelectFromWhere = sfw->IsSelectFromWhere;
        sb.sb_Distinct = sfw->Distinct;
        sb.sb_NumSelectItems = countSelectList(sfw->SelectList);
        sb.sb_NumFromItems = numFromItems;
        sb.sb_NumOrderItems = countOrderList(sfw->OrderList);
        sb.sb_Where = (sfw->Expr != NULL);
        sb.sb_Continue = 0;
        binout(&sb, sizeof(struct sfwBin));

        /*
        ** SELECT list
        */
        sfwIndex = out->bin_Len;         /* remember current position */
        for(i=0, sl = sfw->SelectList; sl; sl = sl->Next, i++)
        {
            struct RelFieldPath *rap = NULL;

            if(sl->Expr == NULL)
                continue;

            if(sl->Expr->Case == EXPR_CASE_GLOBALIDENT)
                rap = &sl->Expr->Contents.GlobalIdent->RelFieldPath;
            else if(sl->Expr->Case == EXPR_CASE_LOCALEXTEND)
                rap = &sl->Expr->Contents.LocalExtend->GlobalIdent->RelFieldPath;

            sfwsi.sfwsi_Title = sl->Title;
            sfwsi.sfwsi_LocalName = NULL;
            if(sl->Expr->Case == EXPR_CASE_LOCALEXTEND)
                sfwsi.sfwsi_LocalName = sl->Expr->Contents.LocalExtend->LocalIdent->Name;
            if(rap == NULL)
                sfwsi.sfwsi_RelFieldPath.rap_Table = NULL;
            else
                sfwsi.sfwsi_RelFieldPath = *rap;
            binout(&sfwsi, sizeof(struct SFWSelectItem));
        }

        /*
        ** FROM list
        */
        for(fl = sfw->FromList; fl; fl = fl->Next)
            if(fl->LocalIdentDef)
                fl->LocalIdentDef->LocalIdent->Pos = ++d;
        d += 2;                                        /* for SelectFromWhere */

        fromIndex = out->bin_Len;          /* remember current position */
        for(fl = sfw->FromList; fl; fl = fl->Next)
        {
            sfwfi.sfwfi_Table = fl->GlobalIdent->RelFieldPath.rap_Table;
            sfwfi.sfwfi_VarNum = (fl->LocalIdentDef? 
		    d - fl->LocalIdentDef->LocalIdent->Pos : -1);
            sfwfi.sfwfi_Filter = fl->GlobalIdent->RelFieldPath.rap_Star;
            binout(&sfwfi, sizeof(struct SFWFromItem));
        }

        /*
        ** ORDER list
        */
        orderIndex = out->bin_Len;          /* remember current position */
        for(ol = sfw->OrderList; ol; ol = ol->Next)
        {
            sfwoi.sfwoi_Mode = (char)ol->OrderMode;
            binout(&sfwoi, sizeof(struct SFWOrderItem));
        }

        /*
        ** Now translate exprs, end each of it with an END tag.
        */
        for(i=0, sl = sfw->SelectList; sl; sl = sl->Next, i++)
        {
            if(sl->Expr == NULL)
                continue;

            memcpy(&sfwsi, out->bin_Buffer + sfwIndex, 
		    sizeof(struct SFWSelectItem));
            sfwsi.sfwsi_Expr.sp_Bin = out;
            sfwsi.sfwsi_Expr.sp_Index = out->bin_Len;
            exprToBin(sl->Expr, d + i);
            binouttag(BIN_END);
            s = sfwsi.sfwsi_Title;
            if(s != NULL)
            {
                sfwsi.sfwsi_Title = (char *)(intptr_t)out->bin_Len;
                binout(s, strlen(s) + 1);
            }
            s = sfwsi.sfwsi_LocalName;
            if(s != NULL)
            {
                sfwsi.sfwsi_LocalName = (char *)(intptr_t)out->bin_Len;
                binout(s, strlen(s) + 1);
            }

            memcpy(out->bin_Buffer + sfwIndex, &sfwsi, 
		    sizeof(struct SFWSelectItem));
            sfwIndex += sizeof(struct SFWSelectItem);
        }

        /*
        ** Translate reference expressions
        */
        for(fl = sfw->FromList; fl; fl = fl->Next)
        {
            struct Expr *refExpr;
            bool remove;

            memcpy(&sfwfi, out->bin_Buffer + fromIndex, 
		    sizeof(struct SFWFromItem));
            sfwfi.sfwfi_ReferenceField = extractReferenceMatch(sfw->Expr,
                        sfwfi.sfwfi_Table, &refExpr, &remove);
            sfwfi.sfwfi_ReferenceExpr.sp_Bin = out;
            sfwfi.sfwfi_ReferenceExpr.sp_Index = out->bin_Len;
            if (sfwfi.sfwfi_ReferenceField != NULL) {
                exprToBin(refExpr, d);
                binouttag(BIN_END);
            }
            if (remove) {
                sfw->Expr = NULL;
                sb.sb_Where = FALSE;
            }

            memcpy(out->bin_Buffer + fromIndex, &sfwfi, 
		    sizeof(struct SFWFromItem));
            fromIndex += sizeof(struct SFWFromItem);
        }

        /*
        ** translate WHERE expression if necessary.
        */
        if(sb.sb_Where)
        {
            sb.sb_WhereExpr.sp_Bin = out;
            sb.sb_WhereExpr.sp_Index = out->bin_Len;

            exprToBin(sfw->Expr, d);
            binouttag(BIN_END);
        }

        /*
        ** translate ORDER BY expressions
        */
        for(ol = sfw->OrderList; ol; ol = ol->Next)
        {
            memcpy(&sfwoi, out->bin_Buffer + orderIndex, 
		    sizeof(struct SFWOrderItem));
            sfwoi.sfwoi_Order.sp_Bin = out;
            sfwoi.sfwoi_Order.sp_Index = out->bin_Len;
            memcpy(out->bin_Buffer + orderIndex, &sfwoi, 
		    sizeof(struct SFWOrderItem));
            orderIndex += sizeof(struct SFWOrderItem);

            exprToBin(ol->Expr, d);
            binouttag(BIN_END);
        }

        /*
        ** set continue field
        */
        sb.sb_Continue = out->bin_Len;
        memcpy(out->bin_Buffer + sbIndex, &sb, sizeof(struct sfwBin));

        /*
        ** translate progn expressions in case of a FOR ALL loop
        */
        if(sfw->IsSelectFromWhere == FALSE)
        {
            struct Func *oldLoopControl = binLoopControl;
            long startIndex, firstjmpIndex, nextIndex, exitIndex;

            /* we put the test expression at the end */
            binouttag(BIN_JMP);
            firstjmpIndex = out->bin_Len;
            binoutlong(0);

            binLoopControl = NULL;
            startIndex = out->bin_Len;
            appArgsToBin(sfw->AppArgs, d);
            binouttag(BIN_POPN);
            binoutshort(CountAppArgs(sfw->AppArgs)); /* removes exprs from stack */

            /* test expression */
            memcpy(out->bin_Buffer + firstjmpIndex, &out->bin_Len, sizeof(long));
            nextIndex = out->bin_Len;
            binouttag(BIN_FORALL_JMP);
            binoutlong(startIndex);

            /* result expression */
            binouttag(BIN_MKNIL);
            exitIndex = out->bin_Len;

            /* remove dummy element (used for keeping stack size) */
            binouttag(BIN_POP1N);
            binoutshort(1);

            addLoopControl(d, nextIndex, exitIndex);
            binLoopControl = oldLoopControl;
        }

        /*
        ** post sfw
        */
        binouttag(BIN_POSTSFW);
        binoutshort(numFromItems);
        binoutshort(numFromVars);
        fromListToBin(sfw->FromList);
    }
}

static void letToBin(struct Let *let, short d)
{
    if(let)
    {
        struct LetList *ll;
        short i;

        for(i=0, ll = let->LetList; ll; ll = ll->Next, i++)
        {
            if(ll->Expr)
                exprToBin(ll->Expr, d);
            else
                binouttag(BIN_MKNIL);
            ll->LocalIdentDef->LocalIdent->Pos = ++d;
        }

        appArgsToBin(let->AppArgs, d);
        addProgN(CountAppArgs(let->AppArgs));

        binouttag(BIN_POP1N);
        binoutshort(i);
    }
}

static void condListToBin(struct CondList *cl, short d)
{
    if(cl == NULL)
        binouttag(BIN_MKNIL);
    else
    {
        long index1, index2;

        exprToBin(cl->Expr, d);
        binouttag(BIN_JMPNIL);
        index1 = out->bin_Len;                  /* remember current position */
        binoutlong(0);
        appArgsToBin(cl->AppArgs, d);
        addProgN(CountAppArgs(cl->AppArgs));
        binouttag(BIN_JMP);
        index2 = out->bin_Len;                  /* remember current position */
        binoutlong(0);
        memcpy(out->bin_Buffer + index1, &out->bin_Len, sizeof(long));
        condListToBin(cl->Next, d);
        memcpy(out->bin_Buffer + index2, &out->bin_Len, sizeof(long));
    }
}

static void condToBin(struct Cond *cond, short d)
{
    if(cond)
        condListToBin(cond->CondList, d);
}

static long caseListListToBin(struct AppArgs *aa, short d)
{
    long index, tmp_index;

    if(aa == NULL)
    {
        binouttag(BIN_JMP);
        index = out->bin_Len;
        binoutlong(0);
    }
    else
    {
        exprToBin(aa->Expr, d);
        binouttag(BIN_CASECMP);
        binouttag(BIN_JMPTRUE);
        tmp_index = out->bin_Len;
        binoutlong(0);
        index = caseListListToBin(aa->Next, d);
        memcpy(out->bin_Buffer + tmp_index, &out->bin_Len, sizeof(long));
    }
    return(index);
}

static void caseListToBin(struct CaseList *cl, short d)
{
    if(cl == NULL)
    {
        binouttag(BIN_POP);
        binouttag(BIN_MKNIL);
    }
    else
    {
        long index, end;

        if(cl->Expr->Case == EXPR_CASE_APP && cl->Expr->Contents.App->IsProgN)
            index = caseListListToBin(cl->Expr->Contents.App->AppArgs, d);
        else
        {
            exprToBin(cl->Expr, d);
            binouttag(BIN_CASECMP);
            binouttag(BIN_JMPNIL);
            index = out->bin_Len;
            binoutlong(0);
        }
        binouttag(BIN_POP);
        appArgsToBin(cl->AppArgs, d-1);
        addProgN(CountAppArgs(cl->AppArgs));
        binouttag(BIN_JMP);
        end = out->bin_Len;
        binoutlong(0);
        memcpy(out->bin_Buffer + index, &out->bin_Len, sizeof(long));

        caseListToBin(cl->Next, d);
        memcpy(out->bin_Buffer + end, &out->bin_Len, sizeof(long));
    }
}

static void caseToBin(struct Case *c, short d)
{
    if(c)
    {
        exprToBin(c->Expr, d);
        caseListToBin(c->CaseList, d+1);
    }
}

static void doToBin(struct Do *doo, short d)
{
    if(doo)
    {
        struct DoBindList *dbl;
        short num_vars;
        long firstjmpIndex, startIndex, nextIndex, exitIndex;
        struct Func *oldLoopControl = binLoopControl;

        for(num_vars=0, dbl = doo->DoBindList; dbl; dbl = dbl->Next, num_vars++)
        {
            if(dbl->InitExpr)
                exprToBin(dbl->InitExpr, d);
            else
                binouttag(BIN_MKNIL);
            dbl->LocalIdentDef->LocalIdent->Pos = ++d;
        }

        /* we put the test expression at the end */
        binouttag(BIN_JMP);
        firstjmpIndex = out->bin_Len;
        binoutlong(0);

        startIndex = out->bin_Len;              /* loop start */
        binLoopControl = NULL;
        appArgsToBin(doo->LoopAppArgs, d);
        binouttag(BIN_POPN);
        binoutshort(CountAppArgs(doo->LoopAppArgs));    /* removes exprs from stack */

        nextIndex = out->bin_Len;               /* loop next */
        for(dbl = doo->DoBindList; dbl; dbl = dbl->Next)
            if(dbl->StepExpr)
            {
                exprToBin(dbl->StepExpr, d);
                binouttag(BIN_SETLOCALIDENT);
                binoutshort(d + 1 - dbl->LocalIdentDef->LocalIdent->Pos);
                binouttag(BIN_POP);
            }

        /* test expression */
        memcpy(out->bin_Buffer + firstjmpIndex, &out->bin_Len, sizeof(long));
        exprToBin(doo->TermExpr, d);
        binouttag(BIN_JMPNIL);
        binoutlong(startIndex);

        /* result expression */
        appArgsToBin(doo->ResultAppArgs, d);
        addProgN(CountAppArgs(doo->ResultAppArgs));

        exitIndex = out->bin_Len;
        binouttag(BIN_POP1N);
        binoutshort(num_vars);

        addLoopControl(d, nextIndex, exitIndex);
        binLoopControl = oldLoopControl;
    }
}

static void doSpecialToBin(struct DoSpecial *ds, short d)
{
    if(ds)
    {
        long firstjmpIndex, startIndex, nextIndex, exitIndex;
        struct Func *oldLoopControl = binLoopControl;

        exprToBin(ds->Expr, d);
        d++;
        infoToBin(&ds->Expr->LineCol);
        switch(ds->Case)
        {
            case DO_CASE_TIMES: binouttag(BIN_DOTIMES_INIT);    break;
            case DO_CASE_LIST:  binouttag(BIN_DOLIST_INIT);     break;
        }
        d++;
        ds->LocalIdentDef->LocalIdent->Pos = d;

        /* we put the test expression at the end */
        binouttag(BIN_JMP);
        firstjmpIndex = out->bin_Len;
        binoutlong(0);

        binLoopControl = NULL;
        startIndex = out->bin_Len;              /* loop start */
        appArgsToBin(ds->LoopAppArgs, d);
        binouttag(BIN_POPN);
        binoutshort(CountAppArgs(ds->LoopAppArgs));    /* removes exprs from stack */

        /* update loop variable */
        nextIndex = out->bin_Len;
        switch(ds->Case)
        {
            case DO_CASE_TIMES: binouttag(BIN_DOTIMES_STEP);    break;
            case DO_CASE_LIST:  binouttag(BIN_DOLIST_STEP);     break;
        }

        /* termination test */
        memcpy(out->bin_Buffer + firstjmpIndex, &out->bin_Len, sizeof(long));
        switch(ds->Case)
        {
            case DO_CASE_TIMES: binouttag(BIN_DOTIMES_TEST);    break;
            case DO_CASE_LIST:  binouttag(BIN_DOLIST_TEST);     break;
        }

        binouttag(BIN_JMPNIL);
        binoutlong(startIndex);

        /* result expression */
        appArgsToBin(ds->ResultAppArgs, d);
        addProgN(CountAppArgs(ds->ResultAppArgs));

        exitIndex = out->bin_Len;
        binouttag(BIN_POP1N);
        binoutshort(2);

        addLoopControl(d, nextIndex, exitIndex);
        binLoopControl = oldLoopControl;
    }
}

static void funcToBin(struct Func *func)
{
    binouttag(BIN_FUNCINFO);
    binoutptr(func->FuncInfo);
}

static void exprToBin(struct Expr *expr, short d)
{
    if(expr)
    {
        switch(expr->Case)
        {
            case EXPR_CASE_APP:
                appToBin(expr->Contents.App, d);
                break;

            case EXPR_CASE_SELECTFROMWHERE:
                sfwToBin(expr->Contents.SelectFromWhere, d);
                break;

            case EXPR_CASE_LET:
                letToBin(expr->Contents.Let, d);
                break;

            case EXPR_CASE_COND:
                condToBin(expr->Contents.Cond, d);
                break;

            case EXPR_CASE_CASE:
                caseToBin(expr->Contents.Case, d);
                break;

            case EXPR_CASE_DO:
                doToBin(expr->Contents.Do, d);
                break;

            case EXPR_CASE_DOSPECIAL:
                doSpecialToBin(expr->Contents.DoSpecial, d);
                break;

            case EXPR_CASE_CONST:
                constToBin(expr->Contents.Const);
                break;

            case EXPR_CASE_GLOBALIDENT:
                globalIdentToBin(expr->Contents.GlobalIdent, expr->Type);
                break;

            case EXPR_CASE_LOCALIDENT:
                localIdentToBin(expr->Contents.LocalIdent, d);
                break;

            case EXPR_CASE_LOCALEXTEND:
                localExtendToBin(expr->Contents.LocalExtend, d, expr->Type, &expr->LineCol);
                break;

            case EXPR_CASE_FUNC:
                funcToBin(expr->Contents.Func);
                break;
        }
    }
}

static void varDefToBin(struct VarDef *varDef, short d)
{
    struct LocalIdent *li = varDef->LocalIdentDef->LocalIdent;
    struct UserGlobalVar *ugv;
    short stack = -(d+1);
    long index = 0;

    li->Pos = stack;
    if (varDef->Keep) {
        ugv = AllocUserGlobalVar(out->bin_MemPool, li);
        if (ugv == NULL) {
            longjmp(errJmpBuf, 1);          /* return to main procedure */
        }
	li->UserGlobalVar = ugv;
        ADDTAIL(&out->bin_UserGlobalVarList, ugv);
        binouttag(BIN_JMPGLOBALVAR);
        binoutptr(ugv);
        index = out->bin_Len;           /* remember current position */
        binoutlong(0);
    }
    if (varDef->Expr) {
        exprToBin(varDef->Expr, d);
    } else {
        binouttag(BIN_MKNIL);
    }

    if (index) {
        memcpy(out->bin_Buffer + index, &out->bin_Len, sizeof(long));
    }
}

static void addUserFunction(struct FuncDef *funcDef)
{
    struct UserFunction *uf;
    struct LocalIdent *li = funcDef->FuncName;

    uf = AllocUserFunction(out->bin_MemPool, li);
    if(uf == NULL)
        longjmp(errJmpBuf, 1);                 /* return to main procedure */
    uf->uf_Trigger = funcDef->Trigger;
    uf->uf_Bin = out;
    ADDTAIL(&out->bin_UserFunctionList, uf);
    funcDef->FuncName->UserFunction = uf;
}

static void addDisplayDependency(struct DisplayObject *dio, short type,
        void *p, short dtype)
{
    struct DisplayDependencies *dd;
    struct Dependency *dep;

    if (dio == NULL || p == NULL) {
	return;				/* no dependency */
    }
    if (type == DD_TYPE_VIRTUALFIELD && DisplayFieldField(dio) == p) {
        return;                         /* do not add self-dependency */
    }
    FORNODE(dd, &out->bin_DisplayDependenciesList) {
	if (dd->dd_Object == dio && dd->dd_Type == type) {
	    FORNODE(dep, &dd->dd_Dependencies) {
		if (dep->dep_Pointer == p) {
		    return;		/* already recorded */
		}
	    }
	    break;
	}
    }
    if (SUCC(dd) == NULL) {		/* not found? */
	dd = AllocDisplayDependencies(out->bin_MemPool, dio, type);
	if (dd == NULL) {
	    longjmp(errJmpBuf, 1);      /* return to main procedure */
	}
	ADDTAIL(&out->bin_DisplayDependenciesList, dd);
    }
    dep = AllocDependency(out->bin_MemPool, p, dtype);
    if (dep == NULL) {
	longjmp(errJmpBuf, 1);      	/* return to main procedure */
    }
    ADDTAIL(&dd->dd_Dependencies, dep);
}

static void buildDisplayDependencySFW(struct SelectFromWhere *sfw, 
	struct DisplayObject *dio, short type)
{
    struct FromList *fl;

    for (fl = sfw->FromList; fl; fl = fl->Next) {
	addDisplayDependency(dio, type,
                fl->GlobalIdent->RelFieldPath.rap_Table, DEP_TYPE_TABLE);
    }

    if (!sfw->IsSelectFromWhere || sfw->SelectList == NULL ||
	    sfw->SelectList->Expr != NULL) {
       return;
    }

    /* we have a SELECT * FROM ... expression */
    for (fl = sfw->FromList; fl; fl = fl->Next) {
	struct Table *tab = fl->GlobalIdent->RelFieldPath.rap_Table;
	struct List *fields = TableFields(tab);
	struct Field *field;

	if (fields) {
	    FORNODE(field, fields) {
		switch (FieldType(field)) {
		    case TYPE_REFERENCE:
                        addDisplayDependency(dio, type,
                                FieldReferenceTable(field), DEP_TYPE_TABLE);
                        /* no break */
		    case TYPE_BOOL:
		    case TYPE_INTEGER:
	    	    case TYPE_CHOICE:
		    case TYPE_REAL:
    		    case TYPE_STRING:
		    case TYPE_MEMO:
		    case TYPE_DATE:
		    case TYPE_TIME:
			addDisplayDependency(dio,type, field,DEP_TYPE_FIELD);
			break;
		}
	    }
	}
    }
}

static void buildDisplayDependencyFunc(struct LocalIdent *func, 
        struct DisplayObject *dio, short type)
{
    struct Expr *expr;

    if (!func->UserFunction || func->UserFunction->uf_Visited) {
	return;
    }
    func->UserFunction->uf_Visited = TRUE;

    FORNODE(expr, &func->ExprList) {
	struct RelFieldPath *rap = NULL;
        struct Field *field;
	struct LocalIdent *li;
	int i;

	switch (expr->Case) {
    	    case EXPR_CASE_LOCALIDENT:
		li = expr->Contents.LocalIdent;
		addDisplayDependency(dio, type,
                        li->UserGlobalVar, DEP_TYPE_USERGLOBALVAR);
		buildDisplayDependencyFunc(li, dio, type);
		break;
    	    case EXPR_CASE_GLOBALIDENT:
		rap = &expr->Contents.GlobalIdent->RelFieldPath;
		break;
    	    case EXPR_CASE_LOCALEXTEND:
		rap = &expr->Contents.LocalExtend->GlobalIdent->RelFieldPath;
		break;
	    case EXPR_CASE_SELECTFROMWHERE:
		buildDisplayDependencySFW(expr->Contents.SelectFromWhere, 
                        dio, type);
		break;
	}
	if (rap) {
            /*
            ** If exression type is TYPE_FIELD and the field is a virtual one
            ** then we set the depencency onto the GuiObject (list kind).
            ** This is necessary such that GETVITRUALLISTACTIVE depends
            ** on the display object and not the field itself.
            */
            field = NULL;
            if (expr->Type == TYPE_FIELD && rap->rap_NumFields > 0) {
                field = rap->rap_Fields[rap->rap_NumFields-1];
            }
            if (field != NULL && FieldType(field) == TYPE_VIRTUAL) {
                addDisplayDependency(dio, type, FieldDisplayObject(field),
                        DEP_TYPE_GUIOBJECT);
            } else {
                if (rap->rap_Table != rap->rap_Relative) {
                    addDisplayDependency(dio, type,
                            TableDisplayObject(rap->rap_Table),
                            DEP_TYPE_GUIRECORD);
                }
                for (i = 0; i < rap->rap_NumFields; i++) { 
                    field = rap->rap_Fields[i];
                    if (FieldType(field) == TYPE_REFERENCE) {
                        addDisplayDependency(dio, type,
                                FieldReferenceTable(field), DEP_TYPE_TABLE);
                    }
                    addDisplayDependency(dio, type, field, DEP_TYPE_FIELD);
                }
            }
	}
    }
}

static void buildDisplayDependencyFuncDef(struct FuncDef *funcDef, 
        struct DisplayObject *dio, short type)
{
    struct UserFunction *uf;

    FORNODE(uf, &out->bin_UserFunctionList) {
        uf->uf_Visited = FALSE;
    }

    /* add project as a dependency */
    addDisplayDependency(dio, type, binProject, DEP_TYPE_PROJECT);

    /* add all other dependencies */
    buildDisplayDependencyFunc(funcDef->FuncName, dio, type);
}

static void buildDependencyVirtualField(struct FuncDef *funcDef)
{
    struct List *tabs = ProjectTables(binProject);
    struct Table *tab;
    struct List *fields;
    struct Field *field;
    const char *compute;
    struct DisplayObject *dio;

    if (tabs) {
	FORNODE(tab, tabs) {
	    fields = TableFields(tab);
	    if (fields) {
		FORNODE(field, fields) {
		    compute = FieldVirtualCompute(field);
		    dio = FieldDisplayObject(field);
		    if (compute && !strcmp(compute, funcDef->FuncName->Name)) {
                        buildDisplayDependencyFuncDef(funcDef, 
                                dio, DD_TYPE_VIRTUALFIELD);
                    }
                }
            }
        }
    }
}

static void buildDependencyEnabled(struct FuncDef *funcDef, 
        struct DisplayObject *dio)
{
    struct List *childs = DisplayObjectObjects(dio);
    char enabled = DisplayObjectEnabled(dio);
    const char *compute = DisplayObjectEnabledCompute(dio);

    if (DisplayObjectVisible(dio) && enabled == DOD_ENABLED_COMPUTED && 
            compute != NULL && !strcmp(compute, funcDef->FuncName->Name)) {
        buildDisplayDependencyFuncDef(funcDef, dio, DD_TYPE_ENABLED);
    }

    if (childs) {
        FORNODE (dio, childs) {
            buildDependencyEnabled(funcDef, dio);
        }
    }
}

static void buildDependencyTitle(struct FuncDef *funcDef)
{
    if (!strcmp(mainWindowTitle, funcDef->FuncName->Name)) {
        buildDisplayDependencyFuncDef(funcDef, 
                ProjectRootDisplayObject(binProject), DD_TYPE_TITLE);
    }
}

static void buildDependency(struct FuncDef *funcDef)
{
    buildDependencyVirtualField(funcDef);
    buildDependencyEnabled(funcDef, ProjectRootDisplayObject(binProject));
    buildDependencyTitle(funcDef);
}

static void funcDefToBin(struct FuncDef *funcDef, short d)
{
    struct Func *oldFunctionControl = binFunctionControl, *f;
    struct UserFunction *uf = funcDef->FuncName->UserFunction;
    struct FuncArg *fa;
    long returnIndex;
    short num = 0;

    uf->uf_SubProgram.sp_Bin = out;
    uf->uf_SubProgram.sp_Index = out->bin_Len;

    for(fa = funcDef->FuncArgList; fa; fa = fa->Next, num++)
        fa->LocalIdentDef->LocalIdent->Pos = ++d;

    binFunctionControl = NULL;
    appArgsToBin(funcDef->AppArgs, d + 1);          /* + 1 for return address */
    addProgN(CountAppArgs(funcDef->AppArgs));

    returnIndex = out->bin_Len;
    binouttag(BIN_RTS);
    binoutshort(num);

    for(f = binFunctionControl; f; f = f->Next)
    {
        short n = f->Depth - (d + 1 + 1);     /* return expr & return address */
        long index = f->Index;

        memcpy(out->bin_Buffer + index, &n, sizeof(short));
        index += sizeof(short);
        memcpy(out->bin_Buffer + index, &returnIndex, sizeof(long));
    }
    binFunctionControl = oldFunctionControl;
}

static void defListToBin(struct DefList *defList, short d)
{
    struct DefList *dl;

    for(dl = defList; dl; dl = dl->Next)
        if(dl->Case == DEFLIST_CASE_VARDEF)
            varDefToBin(dl->Contents.VarDef, d++);

    binouttag(BIN_END);

    for(dl = defList; dl; dl = dl->Next)
        if(dl->Case == DEFLIST_CASE_FUNCDEF)
            addUserFunction(dl->Contents.FuncDef);

    for(dl = defList; dl; dl = dl->Next)
        if(dl->Case == DEFLIST_CASE_FUNCDEF)
            buildDependency(dl->Contents.FuncDef);

    for(dl = defList; dl; dl = dl->Next)
        if(dl->Case == DEFLIST_CASE_FUNCDEF)
            funcDefToBin(dl->Contents.FuncDef, d);
}

static void startToBin(struct Start *start, short d)
{
    if (start) {
	struct Bin *b;

        switch(start->Case)
        {
            case START_CASE_EXPR:
		b = ProgramBin(ProjectProgram(binProject));
		binDisplayDependencies = 
                    b? &b->bin_DisplayDependenciesList : NULL;
                exprToBin(start->Contents.Expr, d);
                break;

            case START_CASE_DEFLIST:
		binDisplayDependencies = &out->bin_DisplayDependenciesList;
                defListToBin(start->Contents.DefList, d);
                break;
        }
    }
    binouttag(BIN_END);
}

void FreeBin(struct Bin *bin)
{
    if (bin) {
        struct MemPool *memPool = bin->bin_MemPool;
        struct DisplayDependencies *dd;
        struct UserGlobalVar *ugv;
        struct UserFunction *uf;

	while ((dd = (struct DisplayDependencies *)
		    REMHEAD(&bin->bin_DisplayDependenciesList)) != NULL) {
	    FreeDisplayDependencies(memPool, dd);
	}
        while ((ugv = (struct UserGlobalVar *)
                    REMHEAD(&bin->bin_UserGlobalVarList)) != NULL) {
            FreeUserGlobalVar(memPool, ugv);
        }
        while ((uf = (struct UserFunction *)
                    REMHEAD(&bin->bin_UserFunctionList)) != NULL) {
            FreeUserFunction(memPool, uf);
        }
        PoolFreeMem(memPool, bin->bin_Buffer, bin->bin_Size);
        PoolFreeMem(memPool, bin, sizeof(struct Bin));
    }
}

static struct Bin *AllocBin(struct MemPool *memPool)
{
    struct Bin *b;

    b = PoolAllocMemZero(memPool, sizeof(struct Bin));
    if (b == NULL) {
        return NULL;
    }

    b->bin_MemPool = memPool;
    NewList(&b->bin_UserFunctionList);
    NewList(&b->bin_UserGlobalVarList);
    NewList(&b->bin_DisplayDependenciesList);

    return b;
}

struct Bin *StartToBin(struct Project *pr, struct Start *start,
        struct LocalIdent *identList, bool debug)
{
    struct MemPool *memPool = ProjectMemPool(pr);
    struct Bin *b;
    int ret;
    short d=0;
    struct LocalIdent *li;

    b = AllocBin(memPool);
    if(b == NULL)
        return(NULL);

    b->bin_HasDebugInfo = debug;
    out = b;
    binProject = pr;
    binDisplayDependencies = NULL;
    binLoopControl = NULL;
    binFunctionControl = NULL;
    ret = setjmp(errJmpBuf);
    if(ret == 0)
    {
        for(li = identList; li; li = li->Next)
            li->Pos = -(++d);

        startToBin(start, d);
        return(b);
    }
    /*
    ** at this point an allocation failure has been occured.
    */
    FreeBin(b);
    return(NULL);
}

struct Bin *TriggerToBin(struct DisplayObject *dio, short type)
{
    struct Table *tab = DisplayObjectTable(dio);
    struct Project *pr = DisplayObjectProject(dio);
    struct MemPool *memPool = executeMemPool;
    struct Bin *b = ProgramBin(ProjectProgram(pr));
    int ret;

    binDisplayDependencies = b? &b->bin_DisplayDependenciesList : NULL;

    b = AllocBin(memPool);
    if (b == NULL) {
        return NULL;
    }

    out = b;
    ret = setjmp(errJmpBuf);
    if (ret != 0) {
        /* an error has occured */
        FreeBin(b);
        return NULL;
    }

    binouttag(BIN_INITRUNTRIGGER);
    binoutptr(tab);
    switch (type) {
        case DD_TYPE_VIRTUALFIELD:
        {
            struct Field *field = DisplayFieldField(dio);
            struct RelFieldPath rap;

            rap.rap_Relative = NULL;
            rap.rap_Star = FALSE;
            rap.rap_NumFields = 1;
            rap.rap_Table = tab;
            rap.rap_Fields[0] = field;
            binouttag(BIN_INITGETVIRTUAL);
            binout(&rap, sizeof(struct RelFieldPath));
            binouttag(BIN_EXITGETVIRTUAL);
            binoutptr(field);
            binoutptr(findDisplayDependencies(dio, DD_TYPE_VIRTUALFIELD));
            break;
        }

        case DD_TYPE_ENABLED:
            binouttag(BIN_INITGETENABLED);
            binoutptr(dio);
            binouttag(BIN_EXITGETENABLED);
            binoutptr(dio);
            binoutptr(findDisplayDependencies(dio, DD_TYPE_ENABLED));
            break;

        case DD_TYPE_TITLE:
            binouttag(BIN_INITGETTITLE);
            binouttag(BIN_EXITGETTITLE);
            binoutptr(dio);
            binoutptr(findDisplayDependencies(dio, DD_TYPE_TITLE));
            break;
    }
    binouttag(BIN_EXITRUNTRIGGER);
    binoutptr(tab);
    binouttag(BIN_END);

    out = NULL;

    return b;
}

