/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** scanner.l
**
** Use flex -8 to compile this file.
**
** Creation date: 15.04.1995
** Last modified: 04.08.2020
*/

%option noyywrap
%option nounput

%x DEFINE_START
%x DEFINE_NAME
%x DEFINE_END
%x UNDEF_START
%x UNDEF_NAME
%x INCLUDE_START
%x INCLUDE_NAME
%x IF_NAME
%x IFDEF_NAME
%x IFNDEF_NAME
%x IF_EXCLUDE

%s REPLACE_MODE

%{

#include "error.h"
#include "textfile.h"
#include "dosutil.h"
#include "preferences.h"

#define YY_INPUT(buf, result, max_size) result = 0;   /* We are using strings */
#define YY_NO_UNISTD_H                                /* avoid this include */
#define YY_NEVER_INTERACTIVE 1                        /* avoid isatty call */

#if (YY_FLEX_MAJOR_VERSION > 2) || \
    	((YY_FLEX_MAJOR_VERSION == 2) && \
    	((YY_FLEX_MINOR_VERSION > 5) || \
    	((YY_FLEX_MINOR_VERSION == 5) && \
    	(defined YY_FLEX_SUBMINOR_VERSION) && \
         (YY_FLEX_SUBMINOR_VERSION >= 33))))
    #define HAS_YY_INIT_GLOBALS
#endif

#ifndef HAS_YY_INIT_GLOBALS
static int yy_init_globals (void)
{
    yy_init = 1;
    YY_CURRENT_BUFFER = NULL;

    return 0;
}
#endif

#define YY_USER_INIT \
	    yy_scan_bytes(scanSrc, scanSize); \
	    scanLine = 1; \
            scanColumn = 1; \
            NewList(&defines); \
            replaceIndex = 0; \
            includeIndex = 0; \
            ifIndex = 0; \
            yy_start = INITIAL; \

static const char *scanSrc;
static long scanSize;
static long scanLine;
static long scanColumn;
static char *scanFilename = NULL;
static struct List defines;

#define DEFINE_NAMELEN 30
static char define_name[DEFINE_NAMELEN+1];

struct define
{
    struct Node def_Node;
    char *def_Name;
    char *def_Replace;
};

static void freeDefine(struct define *d)
{
    if (d != NULL) {
        FreeString(memPool, d->def_Replace);
        FreeString(memPool, d->def_Name);
        PoolFreeMem(memPool, d, sizeof(struct define));
    }
}

static struct define *allocDefine(char *name, char *replace)
{
    struct define *d;

    d = PoolAllocMemZero(memPool, sizeof(struct define));
    if (d == NULL) {
        YY_FATAL_ERROR(GCS(MSG_Err_Mem));
    }
    d->def_Name = AllocString(memPool, name);
    if (d->def_Name == NULL) {
        YY_FATAL_ERROR(GCS(MSG_Err_Mem));
    }
    d->def_Replace = AllocString(memPool, replace);
    if (d->def_Replace == NULL) {
        YY_FATAL_ERROR(GCS(MSG_Err_Mem));
    }
    return d;
}

static struct define *findDefine(char *name)
{
    struct define *d;

    FORNODE(d, &defines) {
        if (!strcmp(d->def_Name, name)) {
            return d;
        }
    }
    return NULL;
}


static void addDefine(char *name, char *replace)
{
    struct define *d;

    d = findDefine(name);
    if (d == NULL) {
        d = allocDefine(name, replace);
        ADDTAIL(&defines, d);
    } else if(strcmp(d->def_Replace, replace) != 0) {
        YY_FATAL_ERROR(GCS(MSG_CompErr_SymbolRedefined));
    }
}

static void remDefine(char *name)
{
    struct define *d;

    d = findDefine(name);
    if (d != NULL) {
        REMOVE(d);
        freeDefine(d);
    }
}

#define MAX_REPLACEBUFFERS 16
static YY_BUFFER_STATE replaceBuffers[MAX_REPLACEBUFFERS];
static short replaceIndex = 0;

#define MAX_INCLUDEBUFFERS 16
static YY_BUFFER_STATE  includeBuffers[MAX_INCLUDEBUFFERS];
static long includeScanLines[MAX_INCLUDEBUFFERS];
static long includeScanColumns[MAX_INCLUDEBUFFERS];
static char *includeScanFilenames[MAX_INCLUDEBUFFERS];
static short includeIndex = 0;

#define MAX_IFS 16
static bool elseStates[MAX_IFS];
static long ifIndex = 0, saveIfIndex;
static bool excludeAll;
static long startlen;


// todo: make LineCol using long integers
#define SAVELC  yylval.scan.lc.Line = (short)scanLine, \
                yylval.scan.lc.Col = (short)scanColumn, \
                yylval.scan.lc.Filename = scanFilename

#define ADDLC(x)  if(replaceIndex == 0)  { scanColumn += x; }
#define NLLC      if(replaceIndex == 0)  { scanLine++; scanColumn = 1; }

void UpdateLC(const char *s)
{
    const char *t;
   
    while ((t = CodesetStrchr(s, '\n')) != NULL) {
	NLLC;
	s = CodesetNextChar(t);
    }

    ADDLC(strlen(s));
}

static void fillignore(void)
{
    SAVELC;
    yylval.scan.data.ignore = AllocString(memPool, yytext);
    if (yylval.scan.data.ignore == NULL) {
        YY_FATAL_ERROR(GCS(MSG_Err_Mem));
    }
}

static char *includecontents(char **name, struct MemPool *memPool, long *err)
{
    const char *includedir;
    char *text;

    ProjectSetCurrentDir(compProject);    /* make sure we have correct CWD */
    text = TextFileGetContentsAskWarnings(NULL, memPool, *name, err);
    if (*err == DISKERR_OPEN) {
        includedir = GetPrefsString(compProject, PREFS_PROGRAMINCLUDEDIR);
        if (includedir == NULL) {
            return NULL;
        }
        *name = JoinPathname(memPool, includedir, *name);
        if (*name == NULL) {
            *err = ERR_MEM;
            return NULL;
        }
        text = TextFileGetContentsAskWarnings(NULL, memPool, *name, err);
    }

    return text;
}

static bool projectHasTableOrField(const char *name)
{
    struct List *tabs = ProjectTables(compProject);
    struct Table *tab;

    if (FindTableName(compProject, name) != NULL) {
        return TRUE;
    }
    if (tabs != NULL) {
        FORNODE(tab, tabs) {
            if (FindFieldName(tab, name)) {
                return TRUE;
            }
        }
    }
    return FALSE;
}

%}

oct     [0-7]
digit   [0-9]
hex     [0-9a-fA-F]
octint  0{oct}+
decint  [+-]?{digit}+
hexint  0[xX]{hex}+
real    [+-]?({digit}+|{digit}*\.{digit}+)([eE][+-]?{digit}+)?

date    {digit}+\.{digit}+\.{digit}+|{digit}+\-{digit}+\-{digit}+|{digit}+\/{digit}+\/{digit}+
time    {digit}+\:{digit}+\:{digit}+

stringcont      ([^\\\"\n]|\\.)*
string          \"{stringcont}\"
untermstring    \"{stringcont}\n?

globalname      _?[A-Z][a-zA-Z0-9_]*
localname       [a-z0-9][a-zA-Z0-9_+-]*
identifier      ({globalname}|{localname})\*?
globalident     (\:\:)?{globalname}(\.{globalname})*\*?

select  [Ss][Ee][Ll][Ee][Cc][Tt]
distinct [Dd][Ii][Ss][Tt][Ii][Nn][Cc][Tt]
from    [Ff][Rr][Oo][Mm]
where   [Ww][Hh][Ee][Rr][Ee]
order   [Oo][Rr][Dd][Ee][Rr]
by      [Bb][Yy]
asc     [Aa][Ss][Cc]
desc    [Dd][Ee][Ss][Cc]
let     [Ll][Ee][Tt]
cond    [Cc][Oo][Nn][Dd]
case    [Cc][Aa][Ss][Ee]
for     [Ff][Oo][Rr]
all     [Aa][Ll][Ll]
do      [Dd][Oo]
dotimes [Dd][Oo][Tt][Ii][Mm][Ee][Ss]
dolist  [Dd][Oo][Ll][Ii][Ss][Tt]
defun   [Dd][Ee][Ff][Uu][Nn]
defvar  [Dd][Ee][Ff][Vv][Aa][Rr]
true    [Tt][Rr][Uu][Ee]
nil     [Nn][Ii][Ll]

t_int     \:[Ii][Nn][Tt]
t_real    \:[Rr][Ee][Aa][Ll]
t_str     \:[Ss][Tt][Rr]
t_memo    \:[Mm][Ee][Mm][Oo]
t_date    \:[Dd][Aa][Tt][Ee]
t_time    \:[Tt][Ii][Mm][Ee]
t_list    \:[Ll][Ii][Ss][Tt]
t_file    \:[Ff][Ii][Ll][Ee]
t_func    \:[Ff][Uu][Nn][Cc]

hash    #[ \t]*
define  {hash}define
undef   {hash}undef
include {hash}include
if      {hash}if
ifdef   {hash}ifdef
ifndef  {hash}ifndef
elif    {hash}elif
else    {hash}else
endif   {hash}endif

%%

{define}(.|\\\n)*\n?    {
                            fillignore();
                            yyless(0);
                            BEGIN(DEFINE_START);
                            return IGNORE_TOKEN;
                        }

<DEFINE_START>{define}[ \t]* {
                            ADDLC(yyleng);
                            BEGIN(DEFINE_NAME);
                        }

<DEFINE_NAME>{identifier} {
                            if (yyleng > DEFINE_NAMELEN) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_SymbolNameTooLong));
                            }
                            strcpy(define_name, yytext);
                            ADDLC(yyleng);
                            BEGIN(DEFINE_END);
                        }

<DEFINE_END>([^#\n\"\\]|{string}|\\.)* { yymore(); }
<DEFINE_END>\\\n        { NLLC; yymore(); }
<DEFINE_END>\n?         {
                            addDefine(define_name, yytext);
                            NLLC;
                            BEGIN(INITIAL);
                        }

{undef}[ \t]*{identifier} {
                            fillignore();
                            yyless(0);
                            BEGIN(UNDEF_START);
                            return IGNORE_TOKEN;
                        }

<UNDEF_START>{undef}[ \t]* {
                            ADDLC(yyleng);
                            BEGIN(UNDEF_NAME);
                        }

<UNDEF_NAME>{identifier} {
                            remDefine(yytext);
                            ADDLC(yyleng);
                            BEGIN(INITIAL);
                        }

{include}[ \t]*{string} {
                            fillignore();
                            yyless(0);
                            BEGIN(INCLUDE_START);
                            return IGNORE_TOKEN;
                        }

<INCLUDE_START>{include}[ \t]* {
                            ADDLC(yyleng);
                            BEGIN(INCLUDE_NAME);
                        }

<INCLUDE_NAME>{string}  {
                            char *name = yytext + 1;
                            struct StrHandle *sh = AllocStrHandle(memPool);
                            char *text, *s;
                            long err;

                            ADDLC(yyleng);
                            if (includeIndex >= MAX_INCLUDEBUFFERS) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_TooManyNestedIncludes));
                            }

                            name = AllocStringN(memPool, name, yyleng-2);
                            if (name == NULL) {
                                YY_FATAL_ERROR(GCS(MSG_Err_Mem));
                            }

                            text = includecontents(&name, memPool, &err);
                            if (err == DISKERR_OPEN) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_CantOpenIncludeFile));
                            } else if (err) {
                                YY_FATAL_ERROR(GES(err));
                            }

                            StrHandlePutS(sh, IGNORE_REPLACE_START);
                            StrHandlePutS(sh, text);
                            StrHandlePutS(sh, IGNORE_REPLACE_END);
                            StrHandlePutC(sh, 0);

                            FreeString(memPool, text);

                            s = StrHandleString(sh);
                            if (StrHandleError(sh)) {
                                YY_FATAL_ERROR(GCS(MSG_Err_Mem));
                            }

                            includeBuffers[includeIndex] = YY_CURRENT_BUFFER;
                            includeScanLines[includeIndex] = scanLine;
                            includeScanColumns[includeIndex] = scanColumn;
                            includeScanFilenames[includeIndex] = scanFilename;
                            includeIndex++;
                            scanLine = 1;
                            scanColumn = 1;
                            scanFilename = name;
                            yy_scan_string(s);
                            FreeStrHandle(sh);

                            BEGIN(INITIAL);
                        }

{if}[ \t]*              {
                            BEGIN(IF_NAME);
                            if (ifIndex > MAX_IFS) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_TooManyNestedIfs));
                            }
                            elseStates[ifIndex++] = FALSE;
                            yymore();
                        }

<IF_NAME>{true}         {
                            fillignore();
                            BEGIN(INITIAL);
                            return IGNORE_TOKEN;
                        }

<IF_NAME>{nil}          {
                            saveIfIndex = ifIndex;
                            BEGIN(IF_EXCLUDE);
                            excludeAll = FALSE;
                            yymore();
                        }

{ifdef}[ \t]*           {
                            BEGIN(IFDEF_NAME);
                            if (ifIndex > MAX_IFS) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_TooManyNestedIfs));
                            }
                            elseStates[ifIndex++] = FALSE;
                            startlen = yyleng;
                            yymore();
                        }

<IFDEF_NAME>{identifier} {
                            if (findDefine(&yytext[startlen])) {
                                fillignore();
                                BEGIN(INITIAL);
                                return IGNORE_TOKEN;
                            } else {
                                saveIfIndex = ifIndex;
                                BEGIN(IF_EXCLUDE);
                                excludeAll = FALSE;
                                yymore();
                            }
                        }

{ifndef}[ \t]*          {
                            BEGIN(IFNDEF_NAME);
                            if (ifIndex > MAX_IFS) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_TooManyNestedIfs));
                            }
                            elseStates[ifIndex++] = FALSE;
                            startlen = yyleng;
                            yymore();
                        }

<IFNDEF_NAME>{identifier} {
                            if (!findDefine(&yytext[startlen])) {
                                fillignore();
                                BEGIN(INITIAL);
                                return IGNORE_TOKEN;
                            } else {
                                saveIfIndex = ifIndex;
                                BEGIN(IF_EXCLUDE);
                                excludeAll = FALSE;
                                yymore();
                            }
                        }

{elif}[ \t]*            {
                            if (ifIndex <= 0) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InavlidElseElIfEndIf));
                            }
                            if (elseStates[ifIndex-1]) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InavlidElseElIfEndIf));
                            }
                            saveIfIndex = ifIndex;
                            BEGIN(IF_EXCLUDE);
                            excludeAll = TRUE;
                            yymore();
                        }

{else}[ \t]*            {
                            if (ifIndex <= 0) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InavlidElseElIfEndIf));
                            }
                            if (elseStates[ifIndex-1]) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InavlidElseElIfEndIf));
                            }
                            elseStates[ifIndex-1] = TRUE;
                            saveIfIndex = ifIndex;
                            BEGIN(IF_EXCLUDE);
                            excludeAll = TRUE;
                            yymore();
                        }

{endif}[ \t]*           {
                            if (ifIndex <= 0) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InavlidElseElIfEndIf));
                            }
                            ifIndex--;
                            fillignore();
                            return IGNORE_TOKEN;
                        }

<IF_EXCLUDE>{if}[ \t]*  |
<IF_EXCLUDE>{ifndef}[ \t]* |
<IF_EXCLUDE>{ifdef}[ \t]* {
                            if (ifIndex > MAX_IFS) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_TooManyNestedIfs));
                            }
                            elseStates[ifIndex++] = FALSE;
                            yymore();
                        }

<IF_EXCLUDE>{elif}[ \t]*  {
                            if (ifIndex <= 0) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InavlidElseElIfEndIf));
                            }
                            if (elseStates[ifIndex-1]) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InavlidElseElIfEndIf));
                            }
                            yymore();
                            if (ifIndex == saveIfIndex && !excludeAll) {
                                BEGIN(IF_NAME);
                            }
                        }

<IF_EXCLUDE>{else}[ \t]*  {
                            if (ifIndex <= 0) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InavlidElseElIfEndIf));
                            }
                            if (elseStates[ifIndex-1]) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InavlidElseElIfEndIf));
                            }
                            elseStates[ifIndex-1] = TRUE;
                            if (ifIndex == saveIfIndex && !excludeAll) {
                                BEGIN(INITIAL);
                                fillignore();
                                return IGNORE_TOKEN;
                            } else {
                                yymore();
                            }
                        }

<IF_EXCLUDE>{endif}[ \t]*       {
                                    ifIndex--;
                                    if (ifIndex < saveIfIndex) {
                                        fillignore();
                                        BEGIN(INITIAL);
                                        return IGNORE_TOKEN;
                                    } else {
                                        yymore();
                                    }
                                }

<IF_EXCLUDE>([^#\n\"]|{string})* { yymore(); }
<IF_EXCLUDE>#                   { yymore(); }
<IF_EXCLUDE>\"{stringcont}\n    { NLLC; yymore(); }
<IF_EXCLUDE>\;.*\n              { NLLC; yymore(); }
<IF_EXCLUDE>\n                  { NLLC; yymore(); }

{octint}        {
                    SAVELC;
                    if (!StrToInt(yytext, &yylval.scan.data.integer.value)) {
                        YY_FATAL_ERROR(GCS(MSG_CompErr_InvalidIntegerConstant));
                    }
                    yylval.scan.data.integer.format= CONST_INTEGER_FORMAT_OCTAL;
                    ADDLC(yyleng);
                    return INTEGER_CONST;
                }

{decint}        {
                    SAVELC;
                    if (!StrToInt(yytext, &yylval.scan.data.integer.value)) {
                        YY_FATAL_ERROR(GCS(MSG_CompErr_InvalidIntegerConstant));
                    }
                    yylval.scan.data.integer.format =
                        CONST_INTEGER_FORMAT_DECIMAL;
                    ADDLC(yyleng);
                    return INTEGER_CONST;
                }

{hexint}        {
                    SAVELC;
                    if (!StrToInt(yytext, &yylval.scan.data.integer.value)) {
                        YY_FATAL_ERROR(GCS(MSG_CompErr_InvalidIntegerConstant));
                    }
                    yylval.scan.data.integer.format = strlen(yytext + 2);
                    ADDLC(yyleng);
                    return INTEGER_CONST;
                }

{real}                  {
                            SAVELC;
                            if (!StrToDouble(yytext, &yylval.scan.data.real)) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InvalidRealConstant));
                            }
                            ADDLC(yyleng);
                            return REAL_CONST;
                        }

{date}                  {
                            SAVELC;
                            if (!StringToDate(yytext, &yylval.scan.data.date)) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InvalidDateConstant));
                            }
                            ADDLC(yyleng);
                            return DATE_CONST;
                        }

{time}                  {
                            SAVELC;
                            if (!StringToTime(yytext, FORMAT_HOURMINSEC, 
                                        &yylval.scan.data.time)) {
                                YY_FATAL_ERROR(
                                        GCS(MSG_CompErr_InvalidTimeConstant));
                            }
                            ADDLC(yyleng);
                            return TIME_CONST;
                        }

{string}                {
                            SAVELC;
                            yytext[yyleng - 1] = 0;
                            yylval.scan.data.string = AllocString(
                                    memPool, yytext + 1);
                            yytext[yyleng - 1] = '\"';
                            if (yylval.scan.data.string == NULL) {
                                YY_FATAL_ERROR(GCS(MSG_Err_Mem));
                            }
                            ADDLC(yyleng);
                            return STRING_CONST;
                        }

{untermstring}          YY_FATAL_ERROR(GCS(MSG_CompErr_UnterminatedString));

{select}                {
                            SAVELC;
                            ADDLC(yyleng);
                            return SELECT_TOKEN;
                        }

{distinct}              {
                            SAVELC;
                            ADDLC(yyleng);
                            return DISTINCT_TOKEN;
                        }

{from}                  {
                            SAVELC;
                            ADDLC(yyleng);
                            return FROM_TOKEN;
                        }

{where}                 {
                            SAVELC;
                            ADDLC(yyleng);
                            return WHERE_TOKEN;
                        }

{order}                 {
                            SAVELC;
                            ADDLC(yyleng);
                            return ORDER_TOKEN;
                        }

{by}                    {
                            SAVELC;
                            ADDLC(yyleng);
                            return BY_TOKEN;
                        }

{asc}                   {
                            SAVELC;
                            ADDLC(yyleng);
                            return ASC_TOKEN;
                        }

{desc}                  {
                            SAVELC;
                            ADDLC(yyleng);
                            return DESC_TOKEN;
                        }

{let}                   {
                            SAVELC;
                            ADDLC(yyleng);
                            return LET_TOKEN;
                        }

{for}                   {
                            SAVELC;
                            ADDLC(yyleng);
                            return FOR_TOKEN;
                        }

{all}                   {
                            SAVELC;
                            ADDLC(yyleng);
                            return ALL_TOKEN;
                        }

{do}                    {
                            SAVELC;
                            ADDLC(yyleng);
                            return DO_TOKEN;
                        }

{dotimes}               {
                            SAVELC;
                            ADDLC(yyleng);
                            return DOTIMES_TOKEN;
                        }

{dolist}                {
                            SAVELC;
                            ADDLC(yyleng);
                            return DOLIST_TOKEN;
                        }

{defun}                 {
                            SAVELC;
                            ADDLC(yyleng);
                            return DEFUN_TOKEN;
                        }

{defun}\*               {
                            SAVELC;
                            ADDLC(yyleng);
                            return DEFUNSTAR_TOKEN;
                        }

{defvar}                {
                            SAVELC;
                            ADDLC(yyleng);
                            return DEFVAR_TOKEN;
                        }

{defvar}\*              {
                            SAVELC;
                            ADDLC(yyleng);
                            return DEFVARSTAR_TOKEN;
                        }

{cond}                  {
                            SAVELC;
                            ADDLC(yyleng);
                            return COND_TOKEN;
                        }

{case}                  {
                            SAVELC;
                            ADDLC(yyleng);
                            return CASE_TOKEN;
                        }

{t_int}                 {
                            SAVELC;
                            ADDLC(yyleng);
                            yylval.scan.data.type.type = TYPE_INTEGER;
                            yylval.scan.data.type.tabname = NULL;
                            return TYPE_SPEC;
                        }

{t_real}                {
                            SAVELC;
                            ADDLC(yyleng);
                            yylval.scan.data.type.type = TYPE_REAL;
                            yylval.scan.data.type.tabname = NULL;
                            return TYPE_SPEC;
                        }

{t_str}                 {
                            SAVELC;
                            ADDLC(yyleng);
                            yylval.scan.data.type.type = TYPE_STRING;
                            yylval.scan.data.type.tabname = NULL;
                            return TYPE_SPEC;
                        }

{t_memo}                {
                            SAVELC;
                            ADDLC(yyleng);
                            yylval.scan.data.type.type = TYPE_MEMO;
                            yylval.scan.data.type.tabname = NULL;
                            return TYPE_SPEC;
                        }

{t_date}                {
                            SAVELC;
                            ADDLC(yyleng);
                            yylval.scan.data.type.type = TYPE_DATE;
                            yylval.scan.data.type.tabname = NULL;
                            return TYPE_SPEC;
                        }

{t_time}                {
                            SAVELC;
                            ADDLC(yyleng);
                            yylval.scan.data.type.type = TYPE_TIME;
                            yylval.scan.data.type.tabname = NULL;
                            return TYPE_SPEC;
                        }

{t_list}                {
                            SAVELC;
                            ADDLC(yyleng);
                            yylval.scan.data.type.type = TYPE_LIST;
                            yylval.scan.data.type.tabname = NULL;
                            return TYPE_SPEC;
                        }

{t_file}                {
                            SAVELC;
                            ADDLC(yyleng);
                            yylval.scan.data.type.type = TYPE_FILE;
                            yylval.scan.data.type.tabname = NULL;
                            return TYPE_SPEC;
                        }

{t_func}                {
                            SAVELC;
                            ADDLC(yyleng);
                            yylval.scan.data.type.type = TYPE_FUNCTION;
                            yylval.scan.data.type.tabname = NULL;
                            return TYPE_SPEC;
                        }

\:{globalname}          {
                            SAVELC;
                            ADDLC(yyleng);
                            yylval.scan.data.type.type = TYPE_RECORD;
                            yylval.scan.data.type.tabname = AllocString(
                                    memPool, yytext+1);
                            if (yylval.scan.data.type.tabname == NULL) {
                                YY_FATAL_ERROR(GCS(MSG_Err_Mem));
                            }
                            return TYPE_SPEC;
                        }

\@\+                    {
                            SAVELC;
                            if (replaceIndex == 0 && includeIndex == 0) {
                                YY_FATAL_ERROR(GCS(MSG_CompErr_SyntaxError));
                            }
                            yylval.scan.data.ignore = IGNORE_REPLACE_START;
                            return IGNORE_TOKEN;
                        }

\@\-                    {
                            SAVELC;
                            if (replaceIndex == 0 && includeIndex == 0) {
                                YY_FATAL_ERROR(GCS(MSG_CompErr_SyntaxError));
                            }
                            yylval.scan.data.ignore = IGNORE_REPLACE_END;
                            return IGNORE_TOKEN;
                        }

{identifier}            {
                            struct define *d;
                            struct FuncInfo *fi;
                            struct ConstInfo *ci;
                            short len = yyleng;

                            SAVELC;
                            ADDLC(yyleng);

                            d = findDefine(yytext);
                            if (d != NULL) {
                                long size = 0;
                                char *s;

                                if (replaceIndex >= MAX_REPLACEBUFFERS) {
                                    YY_FATAL_ERROR(
                                         GCS(MSG_CompErr_TooManyNestedDefines));
                                }

                                fillignore();

                                size += strlen(IGNORE_REPLACE_START);
                                size += strlen(d->def_Replace);
                                size += strlen(IGNORE_REPLACE_END);
                                size++;

                                s = PoolAllocMem(memPool, size);
                                if (s == NULL) {
                                    YY_FATAL_ERROR(GCS(MSG_Err_Mem));
                                }

                                sprintf(s, "%s%s%s", IGNORE_REPLACE_START,
                                    d->def_Replace, IGNORE_REPLACE_END);

                                replaceBuffers[replaceIndex++] = 
                                                        YY_CURRENT_BUFFER;
                                yy_scan_string(s);
                                PoolFreeMem(memPool, s, size);

                                return IGNORE_TOKEN;
                            }

                            fi = FindFuncInfoName(yytext);
                            if (fi != NULL && !projectHasTableOrField(yytext)) {
                                yylval.scan.data.funcinfo = fi;
                                return FUNC_INFO;
                            }

                            ci = FindConstInfoName(yytext);
                            if (ci != NULL && !projectHasTableOrField(yytext)) {
                                yylval.scan.data.constinfo = ci;
                                return CONST_INFO;
                            }

                            yylval.scan.data.identifier = AllocString(
                                    memPool, yytext);
                            if (yylval.scan.data.identifier == NULL) {
                                YY_FATAL_ERROR(GCS(MSG_Err_Mem));
                            }
                            if (isupper(yytext[0])) {
                                return GLOBAL_IDENT;
                            } 
        		    if (yytext[0] == '_' && isupper(yytext[1])) {
                                return GLOBAL_IDENT;
        		    }
                            if (yytext[len - 1] != '*') {
                                return LOCAL_IDENT;
                            }
                            YY_FATAL_ERROR(GCS(MSG_CompErr_SyntaxError));
                        }

{globalident}           {
                            SAVELC;
                            yylval.scan.data.identifier = AllocString(
                                    memPool, yytext);
                            if (yylval.scan.data.identifier == NULL) {
                                YY_FATAL_ERROR(GCS(MSG_Err_Mem));
                            }
                            ADDLC(yyleng);
                            return GLOBAL_IDENT;
                        }

[\-\+\*\/\>\<\=]+       {
                            struct FuncInfo *fi;

                            SAVELC;
                            ADDLC(yyleng);

                            fi = FindFuncInfoName(yytext);
                            if (fi != NULL) {
                                yylval.scan.data.funcinfo = fi;
                                return FUNC_INFO;
                            }
                            YY_FATAL_ERROR(GCS(MSG_CompErr_SyntaxError));
                        }

[\(\)\,\.]              {
                            SAVELC;
                            ADDLC(1);
                            return yytext[0];
                        }

\n                      {
                            SAVELC;
                            NLLC;
                            yylval.scan.data.ignore = "\n";
                            return IGNORE_TOKEN;
                        }

[ ]                     {
                            SAVELC;
                            ADDLC(1);
                            yylval.scan.data.ignore = " ";
                            return IGNORE_TOKEN;
                        }

\t                      {
                            SAVELC;
                            ADDLC(1);
                            yylval.scan.data.ignore = "\t";
                            return IGNORE_TOKEN;
                        }

([ \t\n]|(\\\n)|(\;.*\n?))+    {
                            fillignore();
                            UpdateLC(yytext);
                            return IGNORE_TOKEN;
                        }

<<EOF>>                 {
                            yy_delete_buffer(YY_CURRENT_BUFFER);

                            if (replaceIndex > 0) {
                                yy_switch_to_buffer(
                                        replaceBuffers[--replaceIndex]);
                            } else if(includeIndex > 0) {
                                includeIndex--;
                                scanLine = includeScanLines[includeIndex];
                                scanColumn = includeScanColumns[includeIndex];
                                scanFilename=includeScanFilenames[includeIndex];
                                yy_switch_to_buffer(
                                        includeBuffers[includeIndex]);
                            } else if(ifIndex == 0) {
                                yyterminate();
                            } else {
                                YY_FATAL_ERROR(GCS(MSG_CompErr_MissingEndIf));
                            }
                        }

<*>.                    YY_FATAL_ERROR(GCS(MSG_CompErr_SyntaxError));

