/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** stack.c
**
** Creation date: 21.04.1995
** Last modified: 15.02.2022
*/

#include "system.h"
#include <setjmp.h>
#include "dbtypes.h"
#include "program.h"
#include "programbin.h"
#include "stack.h"
#include "heap.h"
#include "func.h"
#include "memo.h"
#include "memory.h"
#include "util.h"
#include "appstrings_cat.h"

static struct
{
    long stack_Size;                    /* total size */
    long stack_Len;                     /* current size used */
    struct Element **stack_Elements;
} Stack;

#define STACK_BLOCK_SIZE 512

void InitStack(void)
{
    Stack.stack_Size = 0;
    Stack.stack_Len = 0;
    Stack.stack_Elements = NULL;
}

static void StackExtend(void)
{
    Stack.stack_Elements = PoolReAllocMem(executeMemPool, Stack.stack_Elements,
	    Stack.stack_Size * sizeof(struct Element *),
	    (Stack.stack_Size + STACK_BLOCK_SIZE) * sizeof(struct Element *));
    if(Stack.stack_Elements == NULL)
	ExecuteAbort(GCS(MSG_Err_Mem));

    Stack.stack_Size += STACK_BLOCK_SIZE;
}

__inline struct Element *StackPush(struct Element *elem)
{
#ifdef DEBUG_LANG
    if(elem == NULL)
        ExecuteAbort(GCS(MSG_StackErr_Element));
#endif

    if(Stack.stack_Len >= Stack.stack_Size)
	StackExtend();
    Stack.stack_Elements[Stack.stack_Len++] = elem;

    return elem;
}

__inline struct Element *StackPop(void)
{
#ifdef DEBUG_LANG
    if(Stack.stack_Len == 0)            /* stack empty? */
        ExecuteAbort(GCS(MSG_StackErr_Empty));
#endif
    Stack.stack_Len--;
    return(Stack.stack_Elements[Stack.stack_Len]);
}

__inline struct Element *StackTop(void)
{
#ifdef DEBUG_LANG
    if(Stack.stack_Len == 0)                            /* stack empty? */
        ExecuteAbort(GCS(MSG_StackErr_Empty));
#endif
    return(Stack.stack_Elements[Stack.stack_Len - 1]);
}

struct Element *StackNth(long n)
{
    if(n >= 0)
    {
#ifdef DEBUG_LANG
        if(Stack.stack_Len <= n)                        /* invalid index? */
            ExecuteAbort(GCS(MSG_StackErr_Empty));
#endif
        return(Stack.stack_Elements[Stack.stack_Len - n - 1]);
    }
    else
    {
#ifdef DEBUG_LANG
        if(Stack.stack_Len < -n)                        /* invalid index? */
            ExecuteAbort(GCS(MSG_StackErr_Empty));
#endif
        return(Stack.stack_Elements[-n - 1]);
    }
}

void SetStackNth(long n, struct Element *elem)
{
    if(n >= 0)
    {
#ifdef DEBUG_LANG
        if(Stack.stack_Len <= n)                        /* invalid index? */
            ExecuteAbort(GCS(MSG_StackErr_Empty));
#endif
        Stack.stack_Elements[Stack.stack_Len - n - 1] = elem;
    }
    else
    {
#ifdef DEBUG_LANG
        if(Stack.stack_Len < -n)                        /* invalid index? */
            ExecuteAbort(GCS(MSG_StackErr_Empty));
#endif
        Stack.stack_Elements[-n - 1] = elem;
    }
}

long StackSize(void)
{
    return(Stack.stack_Len);
}

bool StackEmpty(void)
{
    return((bool)(Stack.stack_Len == 0));
}

void StackReverse(size_t numArgs)
{
    struct Element *help, **elem1, **elem2, **end;

#ifdef DEBUG_LANG
    if(Stack.stack_Len < numArgs)
        ExecuteAbort(GCS(MSG_StackErr_Empty));
#endif
     elem2 = &Stack.stack_Elements[Stack.stack_Len];
     elem1 = elem2 - numArgs;
     end = elem1 + numArgs/2;
     while (elem1 != end)
     {
	 help = *--elem2;
	 *elem2 = *elem1;
	 *elem1++ = help;
     }
}

void StackSqueeze(long numTop, long delta)
{
    struct Element **elem1, **elem2, **end;

#ifdef DEBUG_LANG
    if(delta < 0)
        ExecuteAbort(GCS(MSG_Err_Internal));
    if(numTop + delta > Stack.stack_Len)
        ExecuteAbort(GCS(MSG_StackErr_Empty));
#endif
    if(delta == 0)
        return;

    end = &Stack.stack_Elements[Stack.stack_Len];
    elem2 = end - numTop;
    elem1 = elem2 - delta;
    while (elem2 != end)
        *elem1++ = *elem2++;
    Stack.stack_Len -= delta;
}

struct Element *StackElement(long num)
{
    struct Element *elem = NULL;

    if(num < Stack.stack_Len)
        elem = Stack.stack_Elements[num];
    return(elem);
}

bool PopBool2(void)
{
    return((bool)(StackPop()->el_Type != TYPE_NIL));
}

__inline struct Element *ElementCastInteger(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:          break;
        case TYPE_INTEGER:      break;
        case TYPE_REAL:
        {
	    long i;

	    if (DoubleToInt(elem->el_Contents.real, &i)) {
		elem = MakeInteger(i);
	    } else {
		elem = MakeNil();
	    }
            break;
        }
        default:
            ExecuteAbort(GCS(MSG_ProgErr_Type));
            break;
    }
    return elem;
}

bool PopInteger2(long *integer)
{
    const struct Element *elem = StackPop();

    switch (elem->el_Type)
    {
        case TYPE_NIL:      	return FALSE;
        case TYPE_INTEGER:  	if (integer) {
				    *integer = elem->el_Contents.integer;
				} 
				return TRUE;
        case TYPE_REAL:
	    return DoubleToInt(elem->el_Contents.real, integer);
        default:
            ExecuteAbort(GCS(MSG_ProgErr_Type));
            return FALSE;
    }
}

__inline struct Element *ElementCastReal(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:          break;
        case TYPE_REAL:         break;
        case TYPE_INTEGER:
	    elem = MakeReal((double)elem->el_Contents.integer);
	    break;
        default:
            ExecuteAbort(GCS(MSG_ProgErr_Type));
            break;
    }
    return elem;
}

bool PopReal2(double *real)
{
    const struct Element *elem = StackPop();

    switch (elem->el_Type)
    {
        case TYPE_NIL:      	return FALSE;
        case TYPE_REAL:     	if (real) {
				    *real = elem->el_Contents.real; 
				}
				return TRUE;
        case TYPE_INTEGER:
	    if (real) {
		*real = (double)elem->el_Contents.integer; 
	    }
	    return TRUE;
        default:
            ExecuteAbort(GCS(MSG_ProgErr_Type));
            return FALSE;
    }
}

__inline struct Element *ElementCastDate(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:          break;
        case TYPE_DATE:         break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
                                break;
    }
    return(elem);
}

bool PopDate2(long *date)
{
    const struct Element *elem = StackPop();

    switch (elem->el_Type)
    {
        case TYPE_NIL:      	return FALSE;
        case TYPE_DATE:     	if (date) {
				    *date = elem->el_Contents.date; 
				}
			    	return TRUE;
        default:            	ExecuteAbort(GCS(MSG_ProgErr_Type));
				return FALSE;
    }
}

__inline struct Element *ElementCastTime(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:          break;
        case TYPE_TIME:         break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
                                break;
    }
    return(elem);
}

bool PopTime2(long *time)
{
    const struct Element *elem = StackPop();

    switch (elem->el_Type)
    {
        case TYPE_NIL:          return FALSE;
        case TYPE_TIME:         if (time) {
				    *time = elem->el_Contents.time;
				}
				return TRUE;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
				return FALSE;
    }
}

__inline struct Element *ElementCastString(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:          break;
        case TYPE_STRING:       break;
        case TYPE_MEMO:         elem = MakeString(ExecuteAllocString(
                                            elem->el_Contents.mi),
                                        TRUE);
                                break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
                                break;
    }
    return(elem);
}

char *PopString2(void)
{
    const struct Element *elem = StackPop();

    switch (elem->el_Type)
    {
        case TYPE_NIL:          return NULL;
        case TYPE_STRING:       return elem->el_Contents.string.string;
        case TYPE_MEMO:         return elem->el_Contents.mi;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
				return NULL;
    }
}

__inline struct Element *ElementCastMemo(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:          break;
        case TYPE_MEMO:         break;
        case TYPE_STRING:       elem = MakeMemo(ExecuteAllocString(
                                            elem->el_Contents.string.string),
                                        TRUE);
                                break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
                                break;
    }
    return elem;
}

char *PopMemo2(void)
{
    const struct Element *elem = StackPop();

    switch (elem->el_Type)
    {
        case TYPE_NIL:          return NULL;
        case TYPE_MEMO:         return elem->el_Contents.mi;
        case TYPE_STRING:       return elem->el_Contents.string.string;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
				return NULL;
    }
}

__inline struct Element *ElementCastRecord(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:          break;
        case TYPE_RECORD:       break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
                                break;
    }
    return elem;
}

struct Record *PopRecord2(struct Table **tab)
{
    const struct Element *elem = StackPop();

    switch (elem->el_Type)
    {
        case TYPE_NIL:          return NULL;
        case TYPE_RECORD:       if (tab) {
                                   *tab = elem->el_Contents.tabrec.tab;
				}
				return elem->el_Contents.tabrec.rec;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
				return NULL;
    }
}

struct Record *PopRecord2MatchTable(struct Table *match)
{
    struct Table *tab = NULL;
    struct Record *rec = PopRecord2(&tab);

    if (rec != NULL && match != NULL && tab != match) {
        ExecuteAbort(GCS(MSG_ProgErr_Type));
    }

    return rec;
}


__inline struct Element *ElementCastTable(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:       	break;
        case TYPE_TABLE:     	break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
                                break;
    }
    return elem;
}

struct Table *PopTable2(bool *star)
{
    const struct Element *elem = StackPop();

    switch (elem->el_Type)
    {
        case TYPE_NIL:       	return NULL;
        case TYPE_TABLE:     	if (star) {
				    *star = elem->el_Contents.tab.star;
				}
				return elem->el_Contents.tab.tab;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
                                return NULL;
    }
}

__inline struct Element *ElementCastField(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:          break;
        case TYPE_FIELD:    break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
                                break;
    }
    return(elem);
}

struct RelFieldPath *PopField2(struct Record **rec)
{
    const struct Element *elem = StackPop();

    switch (elem->el_Type)
    {
        case TYPE_NIL:          return NULL;
        case TYPE_FIELD:    if (rec) {
				    *rec = elem->el_Contents.field.rec;
				}
				return elem->el_Contents.field.rap;
        default:                ExecuteAbort(GCS(MSG_ProgErr_Type));
                                return NULL;
    }
}

__inline struct Element *ElementCastList(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:          break;
        case TYPE_LIST:         break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_ListExpected));
                                break;
    }
    return elem;
}

struct Element *PopList2(struct Element **tail)
{
    const struct Element *elem = StackPop();

    switch (elem->el_Type)
    {
        case TYPE_NIL:          return NULL;
        case TYPE_LIST:         if (tail) {
				    *tail = elem->el_Contents.list.tail;
				}
				return elem->el_Contents.list.head;
        default:                ExecuteAbort(GCS(MSG_ProgErr_ListExpected));
                                return NULL;
    }
}

__inline struct Element *ElementCastFile(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:          break;
        case TYPE_FILE:         break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_FileExpected));
                                break;
    }
    return elem;
}

__inline struct Element *ElementCastFunction(struct Element *elem)
{
    switch (elem->el_Type)
    {
        case TYPE_NIL:          break;
        case TYPE_FUNCTION:     break;
        default:                ExecuteAbort(GCS(MSG_ProgErr_FunctionExpected));
                                break;
    }
    return(elem);
}

struct Element *PopInteger(void) { return ElementCastInteger(StackPop()); }
struct Element *PopReal(void)	 { return ElementCastReal(StackPop()); }
struct Element *PopDate(void)	 { return ElementCastDate(StackPop()); }
struct Element *PopTime(void)	 { return ElementCastTime(StackPop()); }
struct Element *PopString(void)  { return ElementCastString(StackPop()); }
struct Element *PopMemo(void)    { return ElementCastMemo(StackPop()); }
struct Element *PopRecord(void)	 { return ElementCastRecord(StackPop()); }
struct Element *PopTable(void)   { return ElementCastTable(StackPop()); }
struct Element *PopField(void){return ElementCastField(StackPop()); }
struct Element *PopList(void)	 { return ElementCastList(StackPop()); }
struct Element *PopFile(void)	 { return ElementCastFile(StackPop()); }
struct Element *PopFunction(void){ return ElementCastFunction(StackPop()); }

