/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** stack.h
**
** Creation date: 21.04.1995
** Last modified: 15.02.2013
*/

#ifndef STACK_H
#define STACK_H

#include "program.h"

#ifdef __cplusplus
extern "C" {
#endif

void InitStack(void);
extern struct Element *StackPush(struct Element *elem); /* returns its arg */
struct Element *StackPop(void);         /* removes and returns top */
extern struct Element *StackTop(void);  /* returns top without remove */
struct Element *StackNth(long n);
/*
** Returns nth stack item. 0 = top of stack,
** positive values mean relative to top of stack,
** negative values means relative to beginning (-1 = first element).
*/
void SetStackNth(long n, struct Element *elem);
long StackSize(void);
bool StackEmpty(void);
void StackReverse(size_t numArgs);
void StackSqueeze(long numTop, long delta);
struct Element *StackElement(long num);         /* for garbage collector */

#define PushNil() StackPush(MakeNil())
#define PushTrue() StackPush(MakeTrue())
#define PushInteger(i) StackPush(MakeInteger(i))
#define PushReal(r) StackPush(MakeReal(r))
#define PushString(s, allocated) StackPush(MakeString(s, allocated))
#define PushMemo(mi, allocated) StackPush(MakeMemo(mi, allocated))
#define PushDate(d) StackPush(MakeDate(d))
#define PushTime(t) StackPush(MakeTime(t))
#define PushRecord(tab, rec) StackPush(MakeRecord(tab, rec))
#define PushTable(tab, star) StackPush(MakeTable(tab, star))
#define PushField(rap, rec, alloc) StackPush(MakeField(rap, rec, alloc))
#define PushList(head, tail) StackPush(MakeList(head, tail))
#define PushFile(num, id) StackPush(MakeFile(num, id))
#define PushCodeAddress(buffer, index) StackPush(MakeCodeAddress(buffer, index))
#define PushSpecial(ptr, func) StackPush(MakeSpecial(ptr, func))
#define PushFunction(type, ptr) StackPush(MakeFunction(type, ptr))

#define ElementCastBool(elem) (elem)
struct Element *ElementCastInteger(struct Element *elem);
struct Element *ElementCastReal(struct Element *elem);
struct Element *ElementCastDate(struct Element *elem);
struct Element *ElementCastTime(struct Element *elem);
struct Element *ElementCastString(struct Element *elem);
struct Element *ElementCastMemo(struct Element *elem);
struct Element *ElementCastRecord(struct Element *elem);
struct Element *ElementCastTable(struct Element *elem);
struct Element *ElementCastField(struct Element *elem);
struct Element *ElementCastList(struct Element *elem);
struct Element *ElementCastFile(struct Element *elem);
struct Element *ElementCastFunction(struct Element *elem);
/*
** Type-checking and auto-conversion functions.
** Returns either an element of the required type or type NIL or aborts
** program execution.
*/

struct Element *PopInteger(void);
struct Element *PopReal(void);
struct Element *PopDate(void);
struct Element *PopTime(void);
struct Element *PopString(void);
struct Element *PopMemo(void);
struct Element *PopRecord(void);
struct Element *PopTable(void);
struct Element *PopField(void);
struct Element *PopList(void);
struct Element *PopFile(void);
struct Element *PopFunction(void);

bool PopBool2(void);
bool PopInteger2(long *integer);
bool PopReal2(double *real);
bool PopDate2(long *date);
bool PopTime2(long *time);
char *PopString2(void);
char *PopMemo2(void);
struct Record *PopRecord2(struct Table **tab);
struct Record *PopRecord2MatchTable(struct Table *tab);
/* record must be from tab (or tab = NULL) */
struct Table *PopTable2(bool *star);
struct RelFieldPath *PopField2(struct Record **rec);
struct Element *PopList2(struct Element **tail);        /* returns head */

#ifdef __cplusplus
}
#endif

#endif

