//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtEditLabels.cc
//
// Creation date: 02.03.2016
// Last modified: 10.04.2016
//

#include "system.h"
#include "qtEditLabels.h"
#include "qtGroupLayout.h"
#include "preferences.h"
#include "appstrings_cat.h"

EditLabels::EditLabels(QWidget *parent) : QWidget(parent)
{
    tree_ = new QTreeWidget(this);
    tree_->setHeaderHidden(true);
    tree_->setRootIsDecorated(false);
    tree_->setDragDropMode(QAbstractItemView::InternalMove);
    tree_->setToolTip(GCS(MSG_EditLabels_LabelsBH));
    connect(tree_, SIGNAL(currentItemChanged(QTreeWidgetItem *, QTreeWidgetItem *)),
            this, SLOT(currentItemChanged(QTreeWidgetItem *)));

    new_ = new QPushButton(GuiMarkShortcut(GCS(MSG_New)), this);
    new_->setToolTip(GCS(MSG_EditLabels_NewBH));
    connect(new_, SIGNAL(clicked()), this, SLOT(newClick()));

    rem_ = new QPushButton(GuiMarkShortcut(GCS(MSG_EditLabels_Remove)), this);
    rem_->setEnabled(false);
    rem_->setToolTip(GCS(MSG_EditLabels_RemoveBH));
    connect(rem_, SIGNAL(clicked()), this, SLOT(remClick()));

    sort_ = new QPushButton(GuiMarkShortcut(GCS(MSG_EditLabels_Sort)), this);
    sort_->setToolTip(GCS(MSG_EditLabels_SortBH));
    connect(sort_, SIGNAL(clicked()), this, SLOT(sortClick()));

    QBoxLayout *listBox;
    bool compact = GetPrefsValue(NULL, PREFS_EDITLABELSCOMPACT);
    if (compact) {
        QHBoxLayout *control = new QHBoxLayout();
        control->addWidget(new_);
        control->addWidget(rem_);
        control->addWidget(sort_);

        listBox = new QVBoxLayout();
        listBox->addWidget(tree_);
        listBox->addLayout(control);
    } else {
        QVBoxLayout *control = new QVBoxLayout();
        control->addStretch(1);
        control->addWidget(new_);
        control->addWidget(rem_);
        control->addWidget(sort_);
        control->addStretch(1);

        listBox = new QHBoxLayout();
        listBox->addWidget(tree_);
        listBox->addLayout(control);
    }

    edit_ = new LineEditTreeView(this, tree_);
    edit_->setEnabled(false);
    connect(edit_, SIGNAL(textEdited(const QString &)),
            this, SLOT(onTextEdited(const QString &)));

    GroupLayout *group = new GroupLayout(2, this);
    group->setContentsMargins(0, 0, 0, 0);
    group->addLayout(listBox, 2);
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_EditLabels_Label)), edit_);

    GuiSetHelpNode(this, "Label editor");
}

void EditLabels::currentItemChanged(QTreeWidgetItem *current) {
    rem_->setEnabled(current != 0);
    edit_->setEnabled(current != 0);
    edit_->setFocus();
}

void EditLabels::newClick()
{
    tree_->setCurrentItem(addText(QString()));
}

void EditLabels::remClick()
{
    QModelIndex index = tree_->currentIndex();
    if (index.isValid()) {
        delete tree_->takeTopLevelItem(index.row());
    }
}

void EditLabels::sortClick() {
    tree_->sortItems(0, Qt::AscendingOrder);
}

void EditLabels::onTextEdited(const QString &text)
{
    QTreeWidgetItem *currentItem = tree_->currentItem();
    if (currentItem) {
        currentItem->setText(0, text);
    }
}

QTreeWidgetItem *EditLabels::addText(const QString &text)
{
    QTreeWidgetItem *item = new QTreeWidgetItem(QStringList(text));
    item->setFlags(item->flags() & ~Qt::ItemIsDropEnabled);
    tree_->addTopLevelItem(item);
    return item;
}

char **EditLabels::labels() {
    int rows = tree_->topLevelItemCount();
    labels_.resize(rows + 1);
    byteArrays_.resize(rows);
    for (int i = 0; i < rows; ++i) {
        byteArrays_[i] = tree_->topLevelItem(i)->text(0).toUtf8();
        labels_[i] = byteArrays_[i].data();
    }
    labels_[rows] = 0;
    return labels_.data();
}

void EditLabels::setLabels(char **labels) {
    tree_->clear();
    if (labels) {
        for (char **l = labels; *l; l++) {
            addText(*l);
        }
    }
    tree_->setCurrentIndex(QModelIndex());
}

