//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtGroupLayout.cc
//
// Creation date: 17.06.2010
// Last modified: 12.10.2021
//

#include "qtGroupLayout.h"
#include <QLabel>
#include <QWidget>

GroupLayout::GroupLayout(int columns, QWidget *parent)
    : QGridLayout(parent), columns_(columns)
{
    col_ = 0;
    row_ = 0;
    
#if 0
    setHorizontalSpacing(8);
    setVerticalSpacing(8);
    setContentsMargins(8, 8, 8, 8);
#endif
}

void GroupLayout::maybeClearContentsMargins()
{
#ifdef MACOSX
    setContentsMargins(0, 0, 0, 0);
#endif
}

void GroupLayout::addLayout(QLayout *layout, int colSpan)
{
    QGridLayout::addLayout(layout, row_, col_, 1, colSpan);
    nextPosition(colSpan);
}

void GroupLayout::addWidget(QWidget *widget, int colSpan)
{
    QGridLayout::addWidget(widget, row_, col_, 1, colSpan);
    nextPosition(colSpan);
}

void GroupLayout::addLabelAndWidget(const QString &label, QWidget *widget,
                                    QLayout *layout)
{
    QLabel *labelWidget = new QLabel(label, widget->parentWidget());
    labelWidget->setBuddy(widget);
    labelWidget->setAlignment(
            (columns_ > 1? Qt::AlignRight : Qt::AlignLeft) | Qt::AlignVCenter);
    addWidget(labelWidget);
    if (layout) {
        addLayout(layout);
    } else {
        addWidget(widget);
    }
}

void GroupLayout::setStretch(int horiz, int vert)
{
    int prevRow = row_;
    int prevCol = col_ - 1;
    if (prevCol < 0) {
        prevCol = columns_ - 1;
        --prevRow;
    }
    if (prevCol >= 0 && horiz > columnStretch(prevCol)) {
        setColumnStretch(prevCol, horiz);
    }
    if (prevRow >= 0 && vert > rowStretch(prevRow)) {
        setRowStretch(prevRow, vert);
    }
}

bool GroupLayout::hasMargin() const
{
    int l, t, r, b;

    getContentsMargins(&l, &t, &r, &b);
    if (l || t || r || b) {
        return true;
    }
    GroupLayout *p = dynamic_cast<GroupLayout *>(parent());
    if (p) {
        return p->hasMargin();
    }
    return false;
}

void GroupLayout::nextPosition(int colSpan)
{
    col_ += colSpan;
    if (col_ >= columns_) {
        row_++;
        col_ = 0;
    }
}

Qt::Orientations GroupLayout::expandingDirections() const
{
    Qt::Orientations orientations = 0;

    for (int i = 0; i < rowCount(); ++i) {
        for (int j = 0; j < columnCount(); ++j) {
            QLayoutItem *item = itemAtPosition(i, j);
            if (item) {
                Qt::Orientations orig = item->expandingDirections();
                Qt::Orientations masked = orig;
                if (rowStretch(i) == 0) {
                    masked &= ~(Qt::Vertical);
                }
                if (columnStretch(j) == 0) {
                    masked &= ~(Qt::Horizontal);
                }
                orientations |= masked;
            }
        }
    }
    return orientations;
}

