//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtPopImage.cc
//
// MUI's popimage class for Qt
//
// Creation data: 18.02.2016
// Last modified: 05.10.2017
//

#include "system.h"

#include "qtPopImage.h"
#include "qt/qtUtil.h"
#include <qt/classes/qtGroupLayout.h>
#include <qt/classes/qtImageLabel.h>
#include <qt/classes/qtPopSelection.h>
#include <QApplication>
#include <QColorDialog>
#include <QDesktopWidget>
#include <QTabWidget>
#include "project.h"
#include "appstrings_cat.h"

#define IMAGESPEC_DEFAULT "0:128"

static const QRgb MUIcolors[] = {
    0xffffff,  // white
    0xe0e0e0,  // light gray
    0xc1c1c1,  // gray (background)
    0x606060,  // dark gray
    0x000000,  // black
    0x000000,  // black
    0x3b67A2,  // blue
    0xffa997,  // light pink
};

enum {
    MUI_SHINE,
    MUI_HALFSHINE,
    MUI_BACKGROUND,
    MUI_HALFSHADOW,
    MUI_SHADOW,
    MUI_TEXT,
    MUI_FILL,
    MUI_MARK,
    MUI_NumColors
};

static bool getColor(const QString &spec, QColor *col)
{
    QByteArray utf8 = spec.toUtf8();
    const char *data = utf8.data();
    unsigned long red, green, blue;
    unsigned long rgb;
    int mui;
    bool found = false;

    if (data == NULL || data[0] == 0 || data[1] != ':') {
        // no action
    } else if (sscanf(data, "0:%d", &mui) == 1) {
        // unsupported stippled images
    } else if (sscanf (data, "2:r%lx,%lx,%lx", &red, &green, &blue) == 3) {
        *col = QColor(red >> 24, green >> 24, blue >> 24);
        found = true;
    } else if (sscanf (data, "2:r%lx", &rgb) == 1) {
        *col = QColor((QRgb)rgb);
        found = true;
    } else if (sscanf (data, "2:m%d", &mui) == 1) {
        if (mui >= 0 && mui < MUI_NumColors) {
            *col = MUIcolors[mui];
            found = true;
        }
    } else if (!strncmp(data, "0:128", 5)) {
				// background
    } else if (!strncmp(data, "0:129", 5)) {
	*col = MUIcolors[MUI_SHADOW];
        found = true;
    } else if (!strncmp(data, "0:130", 5)) {
	*col = MUIcolors[MUI_SHINE];
        found = true;
    } else if (!strncmp(data, "0:131", 5)) {
	*col = MUIcolors[MUI_HALFSHADOW];
        found = true;
    } else if (!strncmp(data, "0:143", 5)) {
	*col = MUIcolors[MUI_SHINE];
        found = true;
    }
    return found;
}

class PopImageDialog : public CommonDialog
{
    Q_OBJECT

public:
    PopImageDialog(QWidget *parent, Project *pr, int types);

    void setSpec(const QString &spec);

signals:
    void selectedSpec(const QString &spec);

public slots:
    void filenameChanged()
    {
        ProjectSetCurrentDir(pr_);           // ensure we have correct CWD
        image_->setPixmap(QPixmap(qtExpandPathname(filename_->text())));
    }

private:
    enum Kind {
        COLOR, EXTERNAL
    };

    Project *pr_;
    QColorDialog *color_;
    QLineEdit *filename_;
    ImageLabel *image_;
    QTabWidget *tab_;
    QVector<Kind> kinds_;

    virtual bool acceptEdits()
    {
        QString spec;
        int index = 0;
        if (tab_) {
            index = tab_->currentIndex();
        }
        if (color_ && kinds_[index] == COLOR) {
            QColor col = color_->currentColor();
            QChar zero('0');
            spec = QString("2:r%1%1%1%1,%2%2%2%2,%3%3%3%3").arg(col.red(), 2, 16, zero).
                arg(col.green(), 2, 16, zero).arg(col.blue(), 2, 16, zero);
        } else if (filename_ && kinds_[index] == EXTERNAL) {
            spec = QString("5:%1").arg(filename_->text());
        }
        emit selectedSpec(spec);
        return true;
    }
};

PopImageDialog::PopImageDialog(QWidget *parent, Project *pr, int types) 
    : CommonDialog(parent)
{
    pr_ = pr;
    color_ = 0;
    filename_ = 0;
    image_ = 0;
    tab_ = 0;

    setAttribute(Qt::WA_DeleteOnClose);
    
    const char *labels[2];
    QWidget *widgets[2];
    int n = 0;

    if (types & PopImage::Color) {
        color_ = new QColorDialog(this);
        color_->setOptions(QColorDialog::DontUseNativeDialog | QColorDialog::NoButtons);
        color_->setWindowFlags(Qt::Widget);
        for (int i = 0; i < MUI_NumColors && i < color_->customCount(); i++) {
            color_->setCustomColor(i, MUIcolors[i]);
        }
        widgets[n] = color_;
        labels[n] = GCS(MSG_PopImage_Color);
        kinds_.push_back(COLOR);
        n++;
    }
    
    if (types & PopImage::External) {
        QWidget *base = new QWidget(this);
        GroupLayout *group = new GroupLayout(1, base);

        QFrame *frame = new QFrame(base);
        frame->setFrameShadow(QFrame::Raised);
        frame->setFrameShape(QFrame::StyledPanel);
        group->addWidget(frame);

        image_ = new ImageLabel(frame);
        image_->setFixed(false);
        image_->setScaled(false);
        QVBoxLayout *sub = new QVBoxLayout(frame);
        sub->setContentsMargins(0, 0, 0, 0);
        sub->addWidget(image_);

        filename_ = new QLineEdit(base);
        PopButtonSetCWD *popButton = new PopButtonSetCWD(base, filename_, pr);
        popButton->setIcon(popButton->style()->standardIcon(QStyle::SP_FileIcon));
        PopSelection *popFile = new PopSelection();
        popFile->setType(PopSelection::FileSelection);
        popFile->setString(filename_);
        popFile->setButton(popButton);
        group->addLayout(popFile);
        connect(filename_, SIGNAL(textChanged(const QString &)), this, SLOT(filenameChanged()));
        connect(filename_, SIGNAL(editingFinished()), this, SLOT(filenameChanged()));
        
        widgets[n] = base;
        labels[n] = GCS(MSG_PopImage_External);
        kinds_.push_back(EXTERNAL);
        n++;
    }

    if (n == 0) {
        // Should not happen.
    } else if (n == 1) {
        contentsLayout()->addWidget(widgets[0]);
    } else {
        tab_ = new QTabWidget(this);
        contentsLayout()->addWidget(tab_);
        for (int i = 0; i < n; i++) {
            tab_->addTab(widgets[i], GuiMarkShortcut(labels[i]));
        }
    }
    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));
    setOkDefault();
    // setModal(true);
}

void PopImageDialog::setSpec(const QString &spec)
{
    Kind kind = COLOR;
    QByteArray utf8 = spec.toUtf8();
    const char *data = utf8.data();
    QColor col;

    if (strncmp(data, "5:", 2) == 0 && filename_) {
        kind = EXTERNAL;
        filename_->setText(data + 2);
    } else if (getColor(spec, &col)) {
        kind = COLOR;
        color_->setCurrentColor(col);
    }
    if (tab_) {
        for (int i = 0; i < kinds_.size(); i++) {
            if (kinds_[i] == kind) {
                tab_->setCurrentIndex(i);
                break;
            }
        }
    }
}

PopImage::PopImage(QWidget *parent, Project *pr) : QToolButton(parent)
{
    pr_ = pr;
    types_ = All;
    dialog_ = 0;
    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred);
    
    connect(this, SIGNAL(clicked()), this, SLOT(openPopup()));
}

void PopImage::setSpec(const QString &spec)
{
    spec_ = spec;
    ProjectSetCurrentDir(pr_);           // ensure we have correct CWD
    qtSetWidgetBackground(this, spec);
    if (dialog_) {
        dialog_->setSpec(spec);
    }
    emit changed(spec.isEmpty() || spec == IMAGESPEC_DEFAULT);
}

void PopImage::openPopup()
{
    if (dialog_) {
        placeDialog();
        dialog_->raise();
    } else {
        dialog_ = new PopImageDialog(window(), pr_, types_);
        dialog_->setSpec(spec_);
        connect(dialog_, SIGNAL(selectedSpec(const QString &)),
                this, SLOT(setSpec(const QString &)));
        connect(dialog_, SIGNAL(finished(int)), this, SLOT(activate()));
        placeDialog();
        dialog_->show();
    }
}

void PopImage::placeDialog()
{
    QPoint pos = mapToGlobal(QPoint(0, height()));
    QSize frameMargin = dialog_->frameSize() - dialog_->size();
    QSize size = dialog_->sizeHint();
    QDesktopWidget *desktop = QApplication::desktop();
    QRect geom = desktop->screenGeometry(this);                                 
    if (dialog_->minimumWidth() + frameMargin.width() < width()) {
        size.setWidth(width() - frameMargin.width());
    }
    if (pos.x() + size.width() > geom.x() + geom.width()) {
        pos.setX(geom.x() + geom.width() - size.width());
    }
    if (pos.x() < geom.x()) {
        pos.setX(geom.x());
    }
    if (pos.y() + size.height() > geom.y() + geom.height()) {
        pos.setY(geom.y() + geom.height() - size.height());
    }
    if (pos.y() < geom.y()) {
        pos.setY(geom.y());
    }
    dialog_->move(pos);
    dialog_->resize(size);
}

void PopImage::activate()
{
    if (window()) {
        activateWindow();
        window()->raise();
    }
}

void qtSetWidgetBackground(QWidget *widget, const QString &spec)
{
    QByteArray utf8 = spec.toUtf8();
    const char *data = utf8.data();
    QColor col;
    QString qss;

    if (strncmp(data, "5:", 2) == 0) {
        qss = QString("background-image: url(\"%1\")").arg(qtExpandPathname(data + 2));
    } else if (getColor(spec, &col)) {
        qss = QString("background-color: %1").arg(col.name());
    }
    widget->setStyleSheet(qss);
}

#include "qtPopImage.moc"

