/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** qtPopSelection.c
**
** an implementation of MUI's popasl class for Qt
**
** Creation date: 13.03.2015
** Last modified: 22.10.2017
*/

#include "qtPopSelection.h"
#include "qt/qtAmiFileDialog.h"
#include "qt/qtUtil.h"
#include <QFontDialog>

PopSelection::PopSelection()
{
    type_ = FileSelection;
    saveMode_ = false;
}

AmiFileDialog *PopSelection::fileDialog() const
{
    return dynamic_cast<AmiFileDialog *>(popup());
}

QFontDialog *PopSelection::fontDialog() const
{
    return dynamic_cast<QFontDialog *>(popup());
}

void PopSelection::open()
{
    QDialog *dialog = dynamic_cast<QDialog *>(popup());

    if (!dialog) {
        switch (type_) {
            case FileSelection:
            {
                AmiFileDialog *fileDialog = new AmiFileDialog(0, QString(),
                        QString(), QString(), false, saveMode_);
                fileDialog->setDefaultSuffix(defaultSuffix_);
                dialog = fileDialog;
                break;
            }
            case FontSelection:
                dialog = new QFontDialog(0);
                break;
        }
        connect(dialog, SIGNAL(finished(int)), this, SLOT(dialogFinished(int)));
        setPopup(dialog);
    }

    if (!start()) {
        close(false);
        return;
    }
    bool modal =string() && string()->window() && string()->window()->isModal();
#ifdef MACOSX
    // For unknown reasons, a file dialog in save-mode seems to require modal
    // execution on OSX.
    if (type_ == FileSelection && saveMode_) {
        modal = true;
    }
#endif
    if (modal) {
        GuiSetSleep(true);
        dialog->exec();
        GuiSetSleep(false);
    } else {
        dialog->show();
        dialog->raise();
    }
}

void PopSelection::close(bool success)
{
    if (success) {
        stop();
    }
    QWidget *popup = this->popup();
    if (popup) {
        if (string() && string()->window()) {
            string()->activateWindow();
            string()->window()->raise();
        }
        popup->hide();
    }    
    setPopup(0);
}

void PopSelection::dialogFinished(int result)
{
    close(result == QDialog::Accepted);
}

bool PopSelection::start()
{
    AmiFileDialog *filedialog = fileDialog();
    if (filedialog) {
        filedialog->selectFile(text());
    }
    QFontDialog *fontdialog = fontDialog();
    if (fontdialog) {
        fontdialog->setCurrentFont(qtFontFromDescription(text()));
    }
    return true;
}

void PopSelection::stop()
{
    AmiFileDialog *filedialog = fileDialog();
    if (filedialog) {
        QStringList filenames = filedialog->selectedFiles();
        if (!filenames.isEmpty()) {
            setText(filenames.first());
        }
    }
    QFontDialog *fontdialog = fontDialog();
    if (fontdialog) {
        // For some unknown reasons selectedFont does not seem to work on OSX.
        setText(qtFontToDescription(fontdialog->currentFont()));
    }
}
