//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtPopString.cc
//
// an implementation of MUI's popstring class for Qt
//
// Creation date: 29.08.2010
// Last modified: 15.10.2017
//

#include "qtPopString.h"
#include <QDialog>
#include <QLabel>
#include <QLineEdit>
#include <QApplication>
#include <QDesktopWidget>
#include <QShortcut>

PopString::PopString() : QHBoxLayout()
{
    setContentsMargins(0, 0, 0, 0);
    setSpacing(POP_STRING_SPACING);
    string_ = 0;
    button_ = 0;
    toggle_ = true;
    follow_ = true;
    popup_ = 0;
}

PopString::~PopString()
{
    if (popup_) {
        popup_->hide();
        delete popup_;
    }
}

void PopString::setString(QWidget *string)
{
    if (string_ != string) {
        if (string) {
            insertWidget(0, string, 1);
            QShortcut *shortcut = new QShortcut(Qt::META + Qt::Key_P, string);
            shortcut->setContext(Qt::WidgetWithChildrenShortcut);
            QObject::connect(shortcut, SIGNAL(activated()), this, SLOT(openPopup()));
        }
        if (string_) {
            removeWidget(string_);
        }
        string_ = string;
    }    
}

void PopString::setButton(QAbstractButton *button)
{    
    if (button_ != button) {
        if (button) {
            insertWidget(1, button, 0);
            connect(button, SIGNAL(clicked()), this, SLOT(openPopup()));
	}
        if (button_) {
            disconnect(button_, SIGNAL(clicked()), this, SLOT(openPopup()));
            removeWidget(button_);
	}
        button_ = button;
    }
}

void PopString::setToggle(bool toggle)
{    
    if (toggle_ != toggle) {
        toggle_ = toggle;
        updateButtonSensitive();
    }
}

void PopString::setPopup(QWidget *popup)
{
    if (popup_ != popup) {
        if (popup_) {
            popup_->deleteLater();
        }
        popup_ = popup;
    }
    placePopup();
}

void PopString::setText(const QString &text)
{
    QLabel *label = dynamic_cast<QLabel *>(string_);
    if (label) {
        label->setText(text);
    }
    
    QLineEdit *edit = dynamic_cast<QLineEdit *>(string_);
    if (edit) {
        edit->setText(text);
    }
    
    QAbstractButton *button = dynamic_cast<QAbstractButton *>(string_);
    if (button) {
        button->setText(text);
    }
}

QString PopString::text() const
{
    QString text;
    
    QLabel *label = dynamic_cast<QLabel *>(string_);
    if (label) {
        text = label->text();
    }
    
    QLineEdit *edit = dynamic_cast<QLineEdit *>(string_);
    if (edit) {
        text = edit->text();
    }
    
    QAbstractButton *button = dynamic_cast<QAbstractButton *>(string_);
    if (button) {
        text = button->text();
    }
    return text;
}

void PopString::openPopup()
{
    if (toggle_ && popup_ && popup_->isVisible()) {
        closePopup(false);
    } else {
        open();
        updateButtonSensitive();
    }
}

void PopString::closePopup(bool success)
{
    close(success);
    updateButtonSensitive();
}

void PopString::updateButtonSensitive()
{
    if (button_) {
        button_->setEnabled(toggle_ || !popup_ || !popup_->isVisible());
    }
}

void PopString::placePopup()
{
    if (popup_ && string_ && button_) {
        QPoint pos = string_->mapToGlobal(QPoint(0, string_->height()));
        QSize frameMargin = popup_->frameSize() - popup_->size();
        QDesktopWidget *desktop = QApplication::desktop();
        QRect geom = desktop->screenGeometry(string_);
        int width = button_->x() + button_->width() - string_->x();
	bool setSize = !dynamic_cast<QDialog *>(popup_.data());
        QSize size = setSize ? popup_->sizeHint() : popup_->size();
        
        if (setSize) {
            if (popup_->minimumWidth() + frameMargin.width() < width) {
                size.setWidth(width - frameMargin.width());
            }
            if (size.height() <= 0) {
                size.setHeight(popup_->size().height());
            }
        }
        if (pos.x() + size.width() > geom.x() + geom.width()) {
            pos.setX(geom.x() + geom.width() - size.width());
        }
        if (pos.x() < geom.x()) {
            pos.setX(geom.x());
        }
        if (pos.y() + size.height() > geom.y() + geom.height()) {
            pos.setY(geom.y() + geom.height() - size.height());
        }
        if (pos.y() < geom.y()) {
            pos.setY(geom.y());
        }
        popup_->move(pos);
        if (setSize) {
            popup_->resize(size);
        }
    }
}
