//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtPopString.h
//
// an implementation of MUI's popstring class for Qt
//
// Creation date: 29.08.2010
// Last modified: 26.04.2015
//

#ifndef QT_POP_STRING_H
#define QT_POP_STRING_H

//
// PopString class is the base class for creating so called
// popup objects. Usually, a popup consists of a string object
// (QLineEdit, QLabel or QButton), followed by a little button
// (QButton object). Pressing this button brings up a little 
// window with a listview and lets the user choose an entry 
// with the mouse.
//
// PopString class features the basic functions for creating
// such objects. Given a string object and a button object,
// it places them horizontally and sets up some notification.
// Whenever the popup button is pressed, a method will be called 
// which itself should open and prepare the popup window.
//
// The string and the button object are not created by Popstring
// class, they have to be supplied as fields during object
// creation time. This makes PopString class very flexible,
// one could e.g. use a text object (Label) instead of a 
// string or a popup button with some text in it.
//
// However, creating simple popups with PopString class would
// be too much overhead. Instead of using it directly, you
// should have a look at one of its subclasses. They offer
// a more specialized set of popups and are a lot easier to use.
//

#include <QAbstractButton>
#include <QHBoxLayout>
#include <QPointer>

#define POP_STRING_SPACING 2

class PopString : public QHBoxLayout
{
    Q_OBJECT
    
public:
    PopString();
    ~PopString();

    void setString(QWidget *string);
    QWidget *string() const { return string_; }
    
    void setButton(QAbstractButton *button);
    QAbstractButton *button() const { return button_; }

    void setToggle(bool toggle);
    bool toggle() const { return toggle_; }
    
    void setText(const QString &text);
    QString text() const;

    bool isOpen() const { return (popup_ && popup_->isVisible()); }

public slots:
    // 
    // This method opens the popup. In fact, it only calls the "open" method. 
    // The popup button object is disabled as long as toggle is unset.
    // If the toggle mode is enabled, calling this function on a currently 
    // opened popup will result in closing this popup (i.e. calling the 
    // "close" method with a success value of false.
    //
    void openPopup();
    
    // 
    // This method closes the popup. In fact, it only calls the
    // "close" method with the supplied success parameter.
    //
    void closePopup(bool success=false);

protected:
    void setPopup(QWidget *popup);
    QWidget *popup() const { return popup_; }

    virtual void open() = 0;
    virtual void close(bool success) = 0;
        
private:
    QWidget *string_;
    QAbstractButton *button_;
    bool toggle_;
    bool follow_;
    QPointer<QWidget> popup_;
    
    void placePopup();
    void updateButtonSensitive();
};

#endif
