//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtQuickSearch.cc
//
// A popup window with a QLineEdit for entering a search string.
// The popup window is placed relative to a parent window at the 
// bottom right corner.
//
// Creation data: 19.04.2015
// Last modified: 19.04.2015
//

#include <system.h>
#include "qtQuickSearch.h"
#include "qtGroupLayout.h"
#include <QApplication>
#include <QDesktopWidget>
#include <QKeyEvent>
#include <QLineEdit>

class SearchEdit : public QLineEdit
{
    Q_OBJECT
    
public:    
    SearchEdit(QWidget *parent) : QLineEdit(parent) {}
    
signals:
    void search(const QString &, QuickSearch::Action action);
    
protected:
    void keyPressEvent(QKeyEvent *event)
    {
        switch(event->key()) {
            case Qt::Key_Escape:
                emit search(text(), QuickSearch::CANCEL);
                break;
            case Qt::Key_Return:
            case Qt::Key_Enter:
                emit search(text(), QuickSearch::COMMIT);
                break;
            case Qt::Key_Up:
            case Qt::Key_PageUp:
                emit search(text(), QuickSearch::PREV);
                break;
            case Qt::Key_Down:
            case Qt::Key_PageDown:
                emit search(text(), QuickSearch::NEXT);
                break;
            default: {
                QString old = text();
                QLineEdit::keyPressEvent(event);
                if (old != text()) {
                    emit search(text(), QuickSearch::START);
                }
                break;
            }
        }
    }
};


QuickSearch::QuickSearch(QWidget *parent) : QWidget(parent)
{
    setWindowFlags(Qt::Popup | Qt::FramelessWindowHint);
    GroupLayout *group = new GroupLayout(1, this);
    //group->setContentsMargins(0, 0, 0, 0);
    lineEdit_ = new SearchEdit(this);
    lineEdit_->setFocus();
    connect(lineEdit_, SIGNAL(search(const QString &, QuickSearch::Action)),
            this, SIGNAL(search(const QString &, QuickSearch::Action)));
    group->addWidget(lineEdit_);
    if (parent) {
        // Place popup below tree view on the right.        
        QSize frameMargin = frameSize() - size();
        QSize size = sizeHint();
        QPoint pos = parent->mapToGlobal(QPoint(parent->width() - size.width() - frameMargin.width(), parent->height()));
        QDesktopWidget *desktop = QApplication::desktop();
        QRect geom = desktop->screenGeometry(this);        
        if (pos.x() + size.width() > geom.x() + geom.width()) {
            pos.setX(geom.x() + geom.width() - size.width());
        }
        if (pos.x() < geom.x()) {
            pos.setX(geom.x());
        }
        if (pos.y() + size.height() > geom.y() + geom.height()) {
            pos.setY(geom.y() + geom.height() - size.height());
        }
        if (pos.y() < geom.y()) {
            pos.setY(geom.y());
        }
        move(pos);
        resize(size);
    }
}

void QuickSearch::processEvent(QEvent *event)
{
    QApplication::sendEvent(lineEdit_, event);
}

#include "qtQuickSearch.moc"
