//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtSFWView.cc
//
// SFWView is a subclass of QTreeView class.
// Its purpose is for displaying the contents of a program element.
//
// Creation data: 12.09.2010
// Last modified: 06.06.2022
//

#include "qtSFWView.h"
#include "qtGroupLayout.h"
#include "qtQuickSearch.h"
#include "qt/qtUtil.h"
#include <QApplication>
#include <QDesktopWidget>
#include <QHeaderView>
#include <QKeyEvent>
#include <QLineEdit>
#include <QMimeData>
#include <lang/program.h>
#include <table.h>
#include <field.h>
#include <imexport.h>
#include <pdfexport.h>
#include <print.h>
#include <codeset.h>
#include <memory.h>
#include <appstrings_cat.h>

#define FETCH_SIZE 100

static const char MimeTypeQueryLine[] = "BeeBase/QueryLine";

static int sortColumn = -1;
static Qt::SortOrder sortOrder = Qt::AscendingOrder;
static char sortType = TYPE_UNKNOWN;

static int queryLineCmp(const void *p1, const void *p2)
{
    const QueryLine *ql1 = *(const QueryLine **)p1;
    const QueryLine *ql2 = *(const QueryLine **)p2;
    int rc;
    if (sortColumn < 0) {
        rc = ql1->ql_LineNum - ql2->ql_LineNum;
    } else {
        rc = QueryEntryCmp(ql1->ql_Entries[sortColumn], 
                           ql2->ql_Entries[sortColumn],
                           sortType, true);
    }
    if (sortOrder == Qt::DescendingOrder) {
        rc = -rc;
    }
    return rc;
}

class SFWModel : public QAbstractListModel
{
    Q_OBJECT

public:
    SFWModel(QObject *parent) : QAbstractListModel(parent)
    {
        data_ = 0;
        reset();
    }

    ~SFWModel()
    {
        FreeQueryGuiData(data_);
        data_ = 0;
    }

    void setCheckBoxes(bool checkBoxes) { checkBoxes_ = checkBoxes; }
    bool checkBoxes() const { return checkBoxes_; }

    bool needsHeader() const { return needsHeader_; }
    bool uniformRowHeights() const { return uniformRowHeights_; }

    int	totalRowCount() const
    {
        return (data_? data_->qgd_NumLines : 0) + padRow_;
    }
    
    virtual int	rowCount(const QModelIndex &parent = QModelIndex()) const
    {
        return fetchedRows_ + padRow_;
    }
    
    virtual bool canFetchMore(const QModelIndex &parent) const
    {
        if (parent.isValid()) {
            return false;
        }
        return (fetchedRows_ < totalRowCount());
    }

    virtual void fetchMore(const QModelIndex &parent)
    {
        if (parent.isValid()) {
            return;
        }
        int fetch = totalRowCount() - fetchedRows_;
        if (fetch > FETCH_SIZE) {
            fetch = FETCH_SIZE;
        }
        if (fetch <= 0) {
            return;
        }

        beginInsertRows(QModelIndex(), fetchedRows_, fetchedRows_ + fetch - 1);
        fetchedRows_ += fetch;
        endInsertRows();
    }

    virtual int	columnCount(const QModelIndex &parent = QModelIndex()) const
    {
        return numFields() + extraField_;
    }

    int numFields() const { return (data_? data_->qgd_NumFields : 0); }
    const QueryFieldInfo *fieldInfos() const { 
        return (data_? data_->qgd_FieldInfos : 0); 
    }
    Table *controlTable() const { return (data_? data_->qgd_ControlTable : 0); }
    
    Qt::ItemFlags flags(const QModelIndex &index) const
    {
        Qt::ItemFlags flags;
        if (!index.isValid()) {
            flags = Qt::ItemIsDropEnabled; 
        } else {
            flags = Qt::ItemIsSelectable |
                Qt::ItemIsDragEnabled | Qt::ItemIsEnabled;
            if (checkBoxes_ && index.column() == 0 ) {
                flags |= Qt::ItemIsUserCheckable;
            }
        }
        return flags;
    }
    
    virtual QVariant headerData(int section, Qt::Orientation orientation, 
                                int role = Qt::DisplayRole) const {
        int col = section;
        if (orientation == Qt::Horizontal && col < numFields()) {
            switch (role) {
                case Qt::DisplayRole:
                    if (data_->qgd_Titles) {
                        return data_->qgd_Titles->ql_Entries[col]->qe_String;
                    }
                    break;

                case Qt::TextAlignmentRole:
                    return (data_->qgd_FieldInfos[col].qfi_Format?
                            Qt::AlignRight : Qt::AlignLeft);
            }
        }
        return QVariant();
    }
    
    virtual QVariant data(const QModelIndex &index, 
            int role = Qt::DisplayRole) const
    {
        int row = index.row();
        int col = index.column();

        if (index.isValid() && row < totalRowCount() && col < numFields()) {
            const QueryLine *ql = data_->qgd_Lines[row];

            switch (role) {
                case Qt::DisplayRole:
                    return ql->ql_Entries[col]->qe_String;

                case Qt::TextAlignmentRole:
                    return (data_->qgd_FieldInfos[col].qfi_Format? 
                            Qt::AlignRight : Qt::AlignLeft);

                case Qt::CheckStateRole:
                    if (checkBoxes_ && col == 0) {
                        return (ql->ql_Checked? Qt::Checked : Qt::Unchecked);
                    }
                    break;
            }
        }
        return QVariant();
    }

    bool setData(const QModelIndex &index, const QVariant &value,
            int role = Qt::EditRole)
    {
        int row = index.row();
        int col = index.column();
        bool rc = false;

        if (index.isValid() && row < totalRowCount() && col == 0 &&
                role == Qt::CheckStateRole) {
            data_->qgd_Lines[row]->ql_Checked = value == Qt::Checked;
            emit dataChanged(index, index);
            rc = true;
        }
        return rc;
    }
    
    Qt::DropActions supportedDropActions() const
    {
        return Qt::MoveAction | Qt::CopyAction;
    }
    
    QStringList mimeTypes() const
    {
        QStringList types;
        types << MimeTypeQueryLine;
        types << MimeTypeTableRecord;
        return types;
    }
    
    QMimeData *mimeData(const QModelIndexList &indexes) const
    {
        QMimeData *mimeData = 0;
        QByteArray queryLineData;
        QByteArray tableRecordData;
        QDataStream queryLineStream(&queryLineData, QIODevice::WriteOnly);
        QDataStream tableRecordStream(&tableRecordData, QIODevice::WriteOnly);
        
        foreach (const QModelIndex &index, indexes) {
            if (index.column() == 0) {
                int n = line(index);
                if (n != 0) {
                    queryLineStream << n;
                }
            }
            Table *tab = 0;
            Record *rec = 0;
            int col = index.column();
            if (index.isValid() && col >= 0 && col < numFields()) {
                Field *fld = field(col);
                short type = FieldType(fld);
                tab = (type == TYPE_REFERENCE)?  
                        FieldReferenceTable(fld) : FieldTable(fld);
                const QueryLine *ql = queryLine(index);
                if (ql) {
                    rec = ql->ql_Entries[col]->qe_Record;
                }
                if (tab && rec) {
                    tableRecordStream << MimeDataTableRecord(tab, rec);
                }
            }
        }
        if (!queryLineData.isEmpty()) {
            if (!mimeData) {
                mimeData = new QMimeData();
            }
            mimeData->setData(MimeTypeQueryLine, queryLineData);
        }
        if (!tableRecordData.isEmpty()) {
            if (!mimeData) {
                mimeData = new QMimeData();
            }
            mimeData->setData(MimeTypeTableRecord, tableRecordData);
        }
        return mimeData;
    }

    bool dropMimeData(const QMimeData *data, Qt::DropAction action, 
                      int row, int column, const QModelIndex &parent)
    {
        if (action == Qt::IgnoreAction) {
            return true;
        }
        if (!data->hasFormat(MimeTypeQueryLine)) {
            return false;
        }
        if (row < 0 || row > totalRowCount()) {
            row = totalRowCount();
        }
        QByteArray encodedData = data->data(MimeTypeQueryLine);
        QDataStream stream(&encodedData, QIODevice::ReadOnly);
        while (!stream.atEnd()) {
            int line;
            stream >> line;
            QModelIndex from = findLine(line);
            if (!from.isValid()) {
                continue;
            }
            int fromRow = from.row();
            if (fromRow < 0 || fromRow >= totalRowCount()) {
                continue;
            }
            QueryGuiDataMoveLine(data_, fromRow, row);
            // We first need to insert the new row, then change persistent
            // indexes for moving the cursor, and then remove the old row.
            beginInsertRows(parent, row, row);
            padRow_ = 1;
            endInsertRows();
            if (row < fromRow) {
                ++fromRow;
            }
            for (int i = 0; i < columnCount(); ++i) {
                QModelIndex fromIndex = index(fromRow, i);
                QModelIndex toIndex = index(row, i);
                changePersistentIndex(fromIndex, toIndex);
            }
            beginRemoveRows(parent, fromRow, fromRow);
            padRow_ = 0;
            endRemoveRows();
            if (row >= fromRow) {
                --row;
            }
            ++row;
            emit internalMoveDrop(row);  // first row starts with 1.
        }
        return true;
    }

    long findRecord(Record *rec)
    {
        long lineNum = 0;

        if (data_ && data_->qgd_ControlTable && 
                data_->qgd_ControlIndex < data_->qgd_NumFields) {
            size_t controlIndex = data_->qgd_ControlIndex;
            size_t i;

            for (i = 0; i < data_->qgd_NumLines; i++) {
                const QueryLine *ql = data_->qgd_Lines[i];
                if (ql->ql_Entries[controlIndex]->qe_Record == rec) {
                    lineNum = ql->ql_LineNum;
                    break;
                }
            }
        }
        return lineNum;
    }

    Record *record(const QModelIndex &index) const
    {
        Record *rec = 0;
        if (index.isValid()) {
            rec = record(index.row() + 1);
        }
        return rec;
    }

    Record *record(long row) const
    {
        Record *rec = 0;
        if (data_ && data_->qgd_ControlTable && 
                data_->qgd_ControlIndex < data_->qgd_NumFields &&
                row >= 0 && row <= totalRowCount()) {
            const QueryLine *ql = row == 0? 
                data_->qgd_Titles : data_->qgd_Lines[row - 1];
            if (ql) {
                rec = ql->ql_Entries[data_->qgd_ControlIndex]->qe_Record;
            }
        }
        return rec;
    }

    QModelIndex findLine(long lineNum) const
    {
        if (data_) {
            for (size_t i = 0; i < data_->qgd_NumLines; i++) {
                const QueryLine *ql = data_->qgd_Lines[i];
                if (ql->ql_LineNum == lineNum)
                    return createIndex(i, 0);
            }
        }
        return QModelIndex();
    }

    long line(const QModelIndex &index) const
    {
        long lineNum = 0;
        if (index.isValid() && index.row() < totalRowCount()) {
            lineNum = data_->qgd_Lines[index.row()]->ql_LineNum;
        }
        return lineNum;
    }
    
    QModelIndex indexForRow(int row) const
    {
        if (row >= 0 && row < totalRowCount()) {
            return createIndex(row, 0);
        }

        return QModelIndex();
    }
    
    const QueryLine *queryLine(const QModelIndex &index) const
    {
        if (index.isValid() && index.row() < totalRowCount()) {
            return data_->qgd_Lines[index.row()];
        } else {
            return 0;
        }
    }
    
    const QueryLine *queryLine(long row) const
    {
        if (row == 0 && data_) {
            return data_->qgd_Titles;
        } else if (row > 0 && row <= totalRowCount()) {
            return data_->qgd_Lines[row - 1];
        } else {
            return 0;
        }
    }

    struct Field *field(int col) const
    {
        if (col >= 0 && col < numFields()) {
            return data_->qgd_FieldInfos[col].qfi_Field;
        } else {
            return 0;
        }
    }

    const char *setTitlesBody(const Element *titles, const Element *body)
    {
        beginResetModel();
        const char *error = setTitlesBodyInternal(titles, body);
        endResetModel();
        return error;
    }

    const char *setElement(const Element *elem)
    {
        const Element *titles = 0, *body = 0;
        if (elem) {
            if (elem->el_Type == TYPE_LIST) {
                titles = elem->el_Contents.list.head;
                body = elem->el_Contents.list.tail;
            } else {
                body = elem;
            }
        }
        return setTitlesBody(titles, body);
    }

    void setString(const char *str)
    {
        beginResetModel();
        setStringInternal(str);
        endResetModel();
    }
    
    void setSelected(const bool *selected)
    {
        beginResetModel();
        if (data_) {
            for (size_t i = 0; i < data_->qgd_NumLines; i++) {
                data_->qgd_Lines[i]->ql_Checked = selected[i];
            }
        }
        endResetModel();
    }
    
    void getSelected(bool *selected) const
    {
        if (data_) {
            for (size_t i = 0; i < data_->qgd_NumLines; i++) {
                selected[i] = data_->qgd_Lines[i]->ql_Checked;
            }
        }
    }
    
    int countSelected() const
    {
        int count = 0;
        if (data_) {
            for (size_t i = 0; i < data_->qgd_NumLines; i++) {
                if (data_->qgd_Lines[i]->ql_Checked) {
                    ++count;
                }
            }
        }
        return count;
    }

    void sort(int column, Qt::SortOrder order)
    {
        if (data_ && column < numFields()) {
            beginResetModel();
            sortColumn = column;
            sortOrder = order;
            sortType = column >= 0? 
                data_->qgd_FieldInfos[column].qfi_Type : TYPE_UNKNOWN;
            qsort(data_->qgd_Lines, data_->qgd_NumLines, sizeof(QueryLine *),
                    queryLineCmp);
            endResetModel();
        }
    }
    
    enum SearchDirection {
        FORWARD,
        BACKWARD
    };
    
    QModelIndex search(const QString &text, int row, int col,
            SearchDirection dir)
    {
        QModelIndex index;
        
        if (!text.isEmpty() && col >= 0 && col < numFields()) {
            QByteArray utf8 = text.toUtf8();
            const char *s = utf8.data();
            if (dir == FORWARD) {
                row = row < 0? 0 : row + 1;
                for (size_t i = row; i < data_->qgd_NumLines; ++i) {
                    const QueryEntry *qe = data_->qgd_Lines[i]->ql_Entries[col];
                    if (CodesetStrStr(qe->qe_String, s)) {
                        index = createIndex(i, col);
                        break;
                    }
                }
            } else {
                row = row < 0 ? data_->qgd_NumLines - 1 : row - 1;
                for (int i = row; i >= 0; --i) {
                    const QueryEntry *qe = data_->qgd_Lines[i]->ql_Entries[col];
                    if (CodesetStrStr(qe->qe_String, s)) {
                        index = createIndex(i, col);
                        break;
                    }
                }
            }
        }
        return index;
    }
    
    void remRecord(const Record *rec) {
        if (data_) {
            for (size_t i = 0; i < data_->qgd_NumLines; ++i) {
                QueryLine *ql = data_->qgd_Lines[i];
                for (size_t j = 0; j < data_->qgd_NumFields; ++j) {
                    if (ql->ql_Entries[j]->qe_Record == rec) {
                        ql->ql_Entries[j]->qe_Record = 0;
                    }
                }
            }
        }
    }    

    void remAllRecords() {
        if (data_) {
            for (size_t i = 0; i < data_->qgd_NumLines; ++i) {
                QueryLine *ql = data_->qgd_Lines[i];
                for (size_t j = 0; j < data_->qgd_NumFields; ++j) {
                    ql->ql_Entries[j]->qe_Record = 0;
                }
            }
        }
    }    

signals:
    void internalMoveDrop(int row);

private:
    QueryGuiData *data_;
    int fetchedRows_;
    int padRow_;
    int extraField_;
    bool uniformRowHeights_;
    bool checkBoxes_;
    bool needsHeader_;

    void reset()
    {
        clear();
        checkBoxes_ = false;
        needsHeader_ = false;
    }
    
    void clear() 
    {
        FreeQueryGuiData(data_);
        data_ = 0;
        padRow_ = 0;
        fetchedRows_ = 0;
        extraField_ = 0;
        uniformRowHeights_ = false;
    }

    const char *setTitlesBodyInternal(const Element *titles, 
            const Element *body)
    {
        QueryGuiData *qgd = QueryGuiDataFromTitlesBody(titles, body);

        if (!qgd) {
            clear();
            return GCS(MSG_Err_Mem);
        }

        if (QueryGuiDataCopyLineNumIfMatch(data_, qgd)) {
            FreeQueryGuiData(qgd);
            return NULL;
        }

        // Set new QueryGuiData
        clear();
        data_ = qgd;
        uniformRowHeights_ = QueryGuiDataAllSingleLine(data_);
        if (data_) {
            fetchedRows_ = data_->qgd_NumLines;
            if (!uniformRowHeights_ && fetchedRows_ > FETCH_SIZE) {
                fetchedRows_ = FETCH_SIZE;
            }
        }

        // Add a dummy column if the last field added was right-aligned
        if (data_ && data_->qgd_NumFields > 0 &&
                data_->qgd_FieldInfos[data_->qgd_NumFields - 1].qfi_Format) {
            extraField_ = 1;
        }

        return 0;
    }

    void setStringInternal(const char *str)
    {
        QString copy(str);
        reset();                // this may deallocate str
        data_ = QueryGuiDataFromString(copy.toUtf8().constData());
        needsHeader_ = true;
    }
};

SFWView::SFWView(QWidget *parent) : QTreeView(parent)
{
    showTitles_ = false;
    model_ = new SFWModel(this);
    quickSearch_ = 0;
    contextMenu_ = 0;
    setModel(model_);
    setAllColumnsShowFocus(true);
    setRootIsDecorated(false);
    setAlternatingRowColors(true);
    setDragDropMode(InternalMove);
    connect(model_, SIGNAL(internalMoveDrop(int)),
            this, SIGNAL(internalMoveDrop(int))); 
}

SFWView::~SFWView()
{
    delete model_;
}

void SFWView::setShowTitles(bool show)
{
    showTitles_ = show;
}

void SFWView::setCheckBoxes(bool checkBoxes)
{
    model_->setCheckBoxes(checkBoxes);
}

bool SFWView::checkBoxes() const
{
    return model_->checkBoxes();
}

int SFWView::numFields() const
{
    return model_->numFields();
}

const QueryFieldInfo *SFWView::fieldInfos() const
{
    return model_->fieldInfos();
}

const signed char *SFWView::getColumnOrder()
{
    columnOrder_.fill(-1, numFields());
    for (int i = 0; i < header()->count(); i++) {
        if (!header()->isSectionHidden(i)) {
            int ix = header()->visualIndex(i);
            if (ix >= 0 && ix < numFields()) {
                columnOrder_[ix] = i;
            }
        }
    }
    return columnOrder_.constData();
}

Table *SFWView::controlTable() const
{
    return model_->controlTable();
}

void SFWView::setActiveRecord(Record *rec)
{
    setActiveLine(model_->findRecord(rec));
}

Record *SFWView::activeRecord() const
{
    return model_->record(currentIndex());
}

Record *SFWView::record(long row) const
{
    return model_->record(row);
}

void SFWView::setActiveLine(long line)
{
    QModelIndex ind = model_->findLine(line);
    setCurrentIndex(ind);
    scrollTo(ind);
}

long SFWView::activeLine() const
{
    return model_->line(currentIndex());
}

const char *SFWView::setTitlesBody(const Element *titles, const Element *body)
{
    int row = currentIndex().row();
    const char *error = model_->setTitlesBody(titles, body);
    setUniformRowHeights(model_->uniformRowHeights());
    updateHeaderAndColumns();
    setCurrentIndex(model_->indexForRow(row));
    return error;
}

const char *SFWView::setElement(const Element *elem)
{
    int row = currentIndex().row();
    const char *error = model_->setElement(elem);
    setUniformRowHeights(model_->uniformRowHeights());
    updateHeaderAndColumns();
    setCurrentIndex(model_->indexForRow(row));
    return error;
}

void SFWView::setString(const char *str)
{
    model_->setString(str);
    updateHeaderAndColumns();
}

void SFWView::updateHeaderAndColumns()
{
    setHeaderHidden(!showTitles_ && !model_->needsHeader());
    for (int i = 0; i < model_->columnCount(); ++i) {
        resizeColumnToContents(i);
    }
    sortByColumn(-1, Qt::AscendingOrder);
    setSortingEnabled(model_->totalRowCount() > 0);
}

void SFWView::setSelected(const bool *selected)
{
    model_->setSelected(selected);
}

void SFWView::getSelected(bool *selected) const
{
    model_->getSelected(selected);
}

int SFWView::countSelected() const
{
    return model_->countSelected();
}

void SFWView::keyPressEvent(QKeyEvent *event)
{
    if (!QuickSearch::isStartEvent(event) || !currentIndex().isValid()) {
        QTreeView::keyPressEvent(event);
    } else {
        delete quickSearch_;
        beforeSearchIndex_ = currentIndex();
        quickSearch_ = new QuickSearch(this);
        connect(quickSearch_, 
                SIGNAL(search(const QString &, QuickSearch::Action)), 
                this, SLOT(onSearch(const QString &, QuickSearch::Action)));
        quickSearch_->show();
        quickSearch_->processEvent(event);
    }
}

void SFWView::onSearch(const QString &text, QuickSearch::Action action)
{
    int col = beforeSearchIndex_.column();
    
    switch (action) {
        case QuickSearch::START: {
            setCurrentIndex(model_->search(text, -1, col, SFWModel::FORWARD));
            break;
        case QuickSearch::PREV: 
        case QuickSearch::NEXT: {
            QModelIndex index = model_->search(text, currentIndex().row(), col, 
                                               action == QuickSearch::PREV? SFWModel::BACKWARD : SFWModel::FORWARD);
            if (index.isValid()) {
                setCurrentIndex(index);
            } else {
                GuiBeep();
            }
            break;
        }
        case QuickSearch::COMMIT:
        case QuickSearch::CANCEL:
            quickSearch_->hide();
            quickSearch_->deleteLater();
            if (action == QuickSearch::CANCEL) {
                setCurrentIndex(beforeSearchIndex_);
            }
            break;
        }
    }
}

void SFWView::contextMenuEvent(QContextMenuEvent *event)
{
    delete contextMenu_;
    contextMenu_ = new QMenu();
    contextMenu_->addAction(GuiMarkShortcut(GCS(MSG_MenuItem_ExportAsText)),
            this, SLOT(exportAsText()));
    contextMenu_->addAction(GuiMarkShortcut(GCS(MSG_MenuItem_ExportAsPDF)),
            this, SLOT(exportAsPDF()));
    contextMenu_->addAction(GuiMarkShortcut(GCS(MSG_Query_Print)),
            this, SLOT(print()));

    contextMenu_->popup(event->globalPos());
    event->accept();
}

void SFWView::exportAsText()
{
    GuiExportSFWList(GuiObjectProject(this), this);
}

void SFWView::exportAsPDF()
{
    GuiExportSFWListAsPDF(GuiObjectProject(this), this);
}

void SFWView::print()
{
    GuiPrintSFWList(GuiObjectProject(this), this);
}

void SFWView::remRecord(const Record *rec)
{
    model_->remRecord(rec);
}

void SFWView::remAllRecords()
{
    model_->remAllRecords();
}

void SFWView::getEntry(const QModelIndex &index,
                       long *line, long *col, Field **field, Record **rec) const
{
    const QueryLine *ql = model_->queryLine(index);

    if (line) {
        *line = ql? ql->ql_LineNum : -1;
    }
    if (col) {
        *col = 0;
    }
    if (field) {
	*field = NULL;
    }
    if (rec) {
	*rec = NULL;
    }
    
    int num = index.column();
    if (!index.isValid() || num >= model_->columnCount()) {
        return;
    }
    if (col) {
        *col = num;
    }
    if (field) {
        *field = model_->field(num);
    }
    if (rec && ql) {
        *rec = ql->ql_Entries[num]->qe_Record;
    }
}

const QueryLine *SFWView::queryLine(long row)
{
    return model_->queryLine(row);
}

#include "qtSFWView.moc"

