//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtSFWView.h
//
// SFWView is a subclass of QTreeView class.
// Its purpose is for displaying the contents of a program element.
//
// Creation data: 12.09.2010
// Last modified: 09.02.2022
//

#ifndef QT_SFW_VIEW_H
#define QT_SFW_VIEW_H

#include <system.h>
#include "qtQuickSearch.h"
#include <QPointer>
#include <QTreeView>
#include <QMenu>
#include <queryeditor.h>

struct MemPool;
struct Table;
struct Field;
struct Record;
struct QueryEntry;
struct Element;
class SFWModel;

class SFWView : public QTreeView
{
    Q_OBJECT
    
public:
    SFWView(QWidget *parent);
    ~SFWView();

    void setShowTitles(bool show);
    bool showTitles() const { return showTitles_; }

    void setCheckBoxes(bool checkBoxes);
    bool checkBoxes() const;

    int numFields() const;
    const QueryFieldInfo *fieldInfos() const;
    const signed char *getColumnOrder();
    Table *controlTable () const;

    const char *setTitlesBody(const Element *titles, const Element *body);
    const char *setElement(const Element *elem);
    void setString(const char *string);

    void setSelected(const bool *selected);
    void getSelected(bool *selected) const;
    int countSelected() const;

    void remRecord(const Record *rec);
    void remAllRecords();

    void getEntry(const QModelIndex &index,
                  long *line, long *col, Field **field, Record **rec) const;

    // Sets and gets the active record (can differ from order displayed).
    void setActiveRecord(Record *rec);
    Record *activeRecord() const;
    // Returns record for a given row as displayed (row = 0 is the header).
    Record *record(long row) const;

    // Sets and gets the active line (can differ from order displayed).
    void setActiveLine(long lineNum);
    long activeLine() const;
    
    // Returns query line for a given row as displayed (row = 0 is the header).
    const QueryLine *queryLine(long row);
    
signals:
    void internalMoveDrop(int row);

public slots:
    void onSearch(const QString &text, QuickSearch::Action action);
    void exportAsText();
    void exportAsPDF();
    void print();
    
protected:
    void keyPressEvent(QKeyEvent * event);
    void contextMenuEvent(QContextMenuEvent *event);

private:
    bool showTitles_;
    QVector<signed char> columnOrder_;
    SFWModel *model_;
    QPointer<QuickSearch> quickSearch_;
    QModelIndex beforeSearchIndex_;
    QPointer<QMenu> contextMenu_;
    
    void updateHeaderAndColumns();
};

#endif

