//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtQueryEditor.cc
//
// Creation date: 01.04.2015
// Last modified: 21.10.2018
//

#ifndef QT_SORTABLE_ABSTRACT_LIST_MODEL_H
#define QT_SORTABLE_ABSTRACT_LIST_MODEL_H

#include <QAbstractListModel>

class SortableAbstractListModel : public QAbstractListModel
{
public:
    SortableAbstractListModel(QObject *parent) : QAbstractListModel(parent)
    {
        changeCount_ = 0;
    }

protected:
    virtual int numElements() const = 0;

    virtual QString display(int ix) const = 0;

    virtual QVariant decoration(int ix) const { return QVariant(); }

    virtual bool moveElement(int from, int to) { return false; }

    virtual QVariant data(const QModelIndex &index, int role) const
    {
        if (index.isValid()) {
            if (role == Qt::DisplayRole) {
                return display(index.row());
            } else if (role == Qt::DecorationRole) {
                return decoration(index.row());
            }
        }
        return QVariant();
    }
    
    virtual Qt::ItemFlags flags(const QModelIndex & index) const
    {
        if (index.isValid()) {
            return Qt::ItemIsSelectable | Qt::ItemIsDragEnabled | Qt::ItemIsEnabled;
        } else {
            return Qt::ItemIsDropEnabled; 
        }
    }
    
    Qt::DropActions supportedDropActions() const
    {
        return Qt::MoveAction;
    }
    
    int rowCount(const QModelIndex &parent = QModelIndex()) const
    {
        return numElements() + changeCount_;
    }
    
    QMap<int, QVariant> itemData(const QModelIndex &index) const
    {
        QMap<int, QVariant> roles;
        roles[Qt::UserRole] = index.row();
        return roles;
    }
    
    bool setItemData(const QModelIndex &index, const QMap<int, QVariant> &roles)
    {
        if (!index.isValid()) {
            return false;
        }

        emit layoutAboutToBeChanged(QList<QPersistentModelIndex>(), 
                QAbstractItemModel::VerticalSortHint);
        bool rc = false;
        int toRow = index.row();
        int fromRow = roles[Qt::UserRole].toInt();
        if (moveElement(fromRow, toRow)) {
            if (fromRow < toRow) {
                toRow--;
            } else {
                fromRow++;
            }
            QModelIndex fromIndex = this->index(fromRow, 0);
            QModelIndex toIndex = this->index(toRow, 0);
            changePersistentIndex(fromIndex, toIndex);
            rc = true;
        }
        emit layoutChanged(QList<QPersistentModelIndex>(), 
                QAbstractItemModel::VerticalSortHint);

        return rc;
    }

    bool insertRows(int row, int count, const QModelIndex &parent = QModelIndex())
    {
        beginInsertRows(parent, row, row + count - 1);
        changeCount_ = count;
        endInsertRows();
        return true;
    }
    
    bool removeRows(int row, int count, const QModelIndex &parent = QModelIndex())
    {
        beginRemoveRows(parent, row, row + count - 1);
        changeCount_ = 0;
        endRemoveRows();
        return true;
    }

private:
    int changeCount_;
};

#endif
