//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
//  qtAmiFileDialog.cc
//
//  Qt interface for choosing files and directories with pathnames
//  compatible to Amiga OS.
//
//  Creation date: 05.06.2010
//  Last modified: 03.05.2022
// 

#include "qt/qtAmiFileDialog.h"
#include "qt/qtUtil.h"
#include <QUrl>

AmiFileDialog::AmiFileDialog(QWidget *parent, const QString &title,
        const QString &ok, const QString &path, bool dirMode, bool saveMode) 
                : QFileDialog(parent, title)
{
    if (saveMode) {
        setAcceptMode(AcceptSave);
        setOption(DontConfirmOverwrite);
    } else {
        setAcceptMode(AcceptOpen);
        if (dirMode) {            
            setFileMode(QFileDialog::Directory);
            setOption(QFileDialog::ShowDirsOnly);
            // setOption(QFileDialog::DontUseNativeDialog);
        } else {
            setFileMode(QFileDialog::ExistingFile);
        }
    }
    
    // This doesn't seem to do anything on MacOSX native file dialog
    QString env = QString::fromLocal8Bit(getenv(APP_NAME));
    if (!env.isEmpty()) {
        QDir dir(env);
        if (dir.exists()) {            
            QString path = dir.absolutePath();
            QList<QUrl> urls;
            
            urls << QUrl::fromLocalFile(path);
            setSidebarUrls(urls);
        }
    }
    
    // This doesn't seem to do anything on MacOSX native file dialog
    if (!ok.isEmpty()) {
        setLabelText(Accept, GuiMarkShortcut(ok));
    }
    
    selectFile(path);
}

void AmiFileDialog::selectFile(const QString &path)
{
    if (!path.isNull()) {
        QString fullPath = 
            qtExpandPathnameGetAssign(path, assignName_, assignPath_);
	if (QFileInfo(fullPath).isDir()) {
            setDirectory(fullPath);
	} else {
            QFileInfo info(fullPath);
            setDirectory(info.dir());
            QFileDialog::selectFile(info.fileName());
	}
    }
}

QStringList AmiFileDialog::selectedFiles() const
{
    QStringList filenames = QFileDialog::selectedFiles();
    QStringList::iterator it;
    for (it = filenames.begin(); it != filenames.end(); ++it) {
        *it = qtUnexpandPathname(*it, assignName_, assignPath_);
    }
    
    return filenames;
}

QString AmiFileDialog::get(QWidget *parent, const QString &title,
        const QString &ok, const QString &path, bool dirMode, bool saveMode)
{
    AmiFileDialog *dialog = new AmiFileDialog(parent, title, ok, path, 
            dirMode, saveMode);
    QString filename;
    
    if (dialog->exec() == QDialog::Accepted) {
        QStringList filenames = dialog->selectedFiles();
        if (!filenames.isEmpty()) {
            filename = filenames.first();
        }
    }
    delete dialog;
    
    return filename;
}
