//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtBreak.cc
//
// Creation date: 01.06.2015
// Last modified: 21.03.2022
//

#include "system.h"
#include "break.h"
#include "qt/qtUtil.h"
#include <QApplication>
#include <QBoxLayout>
#include <QDialog>
#include <QProgressBar>
#include <QPushButton>
#include <QString>
#include <QTime>
#include <QTimer>
#include "util.h"
#include "appstrings_cat.h"

#define BREAK_SHORT_DELAY 1000
#define BREAK_LONG_DELAY 2000

#define BREAK_TITLE_LEN 200

class BreakDialog : public QDialog
{
    Q_OBJECT
public:
    BreakDialog(QWidget *parent, const BreakSettings &bs) : QDialog(parent)
    {
        setWindowTitle(bs.title);
        setModal(true);
        resize(300, 120);
        
        percentage_ = 0;
        aborted_ = false;
        paused_ = false;
        
        QBoxLayout *vbox = new QVBoxLayout(this);
        vbox->setSizeConstraint(QLayout::SetMinAndMaxSize);
        // vbox->setContentsMargins(0, 0, 0, 0);

        progressBar_ = new QProgressBar(this);
        progressBar_->setMinimum(0);
        progressBar_->setMaximum(100);
        progressBar_->setValue(0);
        progressBar_->setTextVisible(true);
        vbox->addWidget(progressBar_);
        
        abortButton_ = 0;
        pauseButton_ = 0;
        if (bs.abort || bs.pause) {
            QBoxLayout *hbox = new QHBoxLayout();
            vbox->addLayout(hbox);
            
            if (bs.abort) {
                abortButton_ = new QPushButton(GuiMarkShortcut(GCS(MSG_Abort)), this);
                abortButton_->setAutoDefault(false);
                // abortButton_->setSizePolicy(QSizePolicy::Minimum, 
                //                             QSizePolicy::Minimum);
                connect(abortButton_, SIGNAL(clicked()), this, SLOT(abort()));
                hbox->addWidget(abortButton_);
            }
            if (bs.pause) {
                pauseButton_ = new QPushButton(GuiMarkShortcut(GCS(MSG_Pause)), this);
                pauseButton_->setAutoDefault(false);
                // pauseButton_->setSizePolicy(QSizePolicy::Minimum, 
                //                             QSizePolicy::Minimum);
                connect(pauseButton_, SIGNAL(clicked()), this, SLOT(pause()));
                hbox->addWidget(pauseButton_);
            }
        }
    }
    
    void updateProgress(const BreakSettings &bs)
    {
        if (progressBar_) {
            bool visible = (bs.steps != 0 && bs.current != 0);
            bool hidden = !visible;
            
            // Note that isVisible is also false when the window is closed.
            if (progressBar_->isHidden() != hidden) {
                progressBar_->setVisible(visible);
            }
            if (visible) {
                short percentage = (short)(bs.current * 100 / bs.steps);
                if (percentage != percentage_) {
                    progressBar_->setValue(percentage);
                    percentage_ = percentage;
                }
            }
        }
    }
    
    void doSettings(const BreakSettings &bs)
    {
        if (pauseButton_) {
            pauseButton_->setEnabled(bs.pause);
        }
        if (abortButton_) {
            abortButton_->setEnabled(bs.abort);
        }
        setWindowTitle(bs.title);
        updateProgress(bs);
    }

    bool aborted() const { return aborted_; }

public slots:
    void abort() {
        aborted_ = true;
    }
    
    void pause() {
        paused_ = !paused_;
        if (paused_) {
            pauseButton_->setText(GuiMarkShortcut(GCS(MSG_Continue)));
            while (paused_ && !aborted_) {
                GuiMainLoopIteration();
            }
        } else {
            pauseButton_->setText(GuiMarkShortcut(GCS(MSG_Pause)));
        }
    }
    
protected:
    virtual void closeEvent(QCloseEvent *event)
    {
        event->ignore();
    }
    
private:
    short percentage_;
    bool paused_;
    bool aborted_;
    QProgressBar *progressBar_;
    QPushButton *abortButton_;
    QPushButton *pauseButton_;
};

static struct {
    bool present;
    bool abort;
    bool pause;
    BreakDialog *win;
    int nextElapsed;
} breakData = { 0 };

static QTime breakLastTime;

static BreakSettings breakSettingsInit, breakSettingsCurrent;
static char breakTitleBuffer[BREAK_TITLE_LEN + 1];
static ulong breakCount = 0;
static bool breakGui = false;

void BreakInit(BreakSettings *bs)
{
    if (breakCount == 0) {
        // 
        // We do not GuiSetSleep here as it interfers with browsing
        // records using the arrow buttons of a panel in a table that
        // has virtual fields causing program execution.
        //  
	breakGui = (GuiAnyOpenedWindow() != 0);
        memset(&breakData, 0, sizeof(breakData));
        if (bs) {
            breakSettingsInit = *bs;
            breakGui |= bs->needGui;
        } else {
            memset(&breakSettingsInit, 0, sizeof(breakSettingsInit));
        }
        breakSettingsCurrent = breakSettingsInit;
        breakLastTime.start();
	if (breakGui) {
            breakData.nextElapsed = BREAK_LONG_DELAY;
            // Make sure the application wakes up after the delay.
            // QTimer::singleShot(BREAK_LONG_DELAY,
            //                   QApplication::instance(), SLOT(interrupt()));
	} else {
	    breakData.nextElapsed = INT_MAX;
	}

    } else if (bs) {
        BreakSettings old = breakSettingsCurrent;
        breakSettingsCurrent = *bs;
        *bs = old;
	if (old.title == breakTitleBuffer) {
	    breakSettingsCurrent.title = breakTitleBuffer;  // keep message
	}
        if (breakData.win) {
            breakData.win->doSettings(breakSettingsCurrent);
        }
    }

    breakCount++;
}

void BreakExit(const BreakSettings *bs)
{
    breakCount--;
    if (breakCount == 0) {
        if (breakData.win) {
            delete breakData.win;
	    GuiSetSleep(false);
	}
        memset(&breakData, 0, sizeof(breakData));
        breakGui = false;
    } else if (bs) {
        breakSettingsCurrent = *bs;
        if (breakData.win) {
            breakData.win->doSettings(breakSettingsCurrent);
        }
    }
}

void BreakHide(void)
{
    GuiCloseWindow(breakData.win);
    breakData.present = false;
    breakData.nextElapsed = INT_MAX;
}

void BreakReInit(void)
{
    if (breakGui && !breakData.present) {
        breakLastTime.start();
        breakData.nextElapsed = BREAK_SHORT_DELAY;
        // Make sure the application wakes up after the delay.
        // QTimer::singleShot(BREAK_SHORT_DELAY,
        //                   QApplication::instance(), SLOT(interrupt()));
    }
}

static void breakCreateGui()
{
    // 
    // Make sure BeeBase is the active application.
    // We don't want to popup the break window 
    // in front of another active application 
    // except when BeeBase does not have any open window.
    //
    if (!GuiActiveWindow() && GuiAnyOpenedWindow()) {   
        return;
    }

    if (!breakData.win) {
        GuiSetSleep(true);
        breakData.win = new BreakDialog(GuiFindParentWindow(), 
                                        breakSettingsInit);
    }
        
    breakData.win->doSettings(breakSettingsCurrent);
    breakData.win->show();
    breakData.present = true;
}
    
bool BreakCheck(void)
{
    if (breakGui) {
        if (breakLastTime.elapsed() >= breakData.nextElapsed) {
	    if (!breakData.present) {
		breakCreateGui();
	    }
            QApplication::processEvents();
            breakLastTime.start();
            breakData.nextElapsed = 10;
	}
    }
    return breakData.win && breakData.win->aborted();
}

void BreakMessage(const char *str)
{
    CodesetStrNBytesCopy(breakTitleBuffer, str, BREAK_TITLE_LEN);
    char *s = CodesetStrchr(breakTitleBuffer, '\n');
    if (s) {
        *s = 0;
    }

    BreakSettings &bs = breakSettingsCurrent;
    bs.title = breakTitleBuffer;
    if (breakData.win) {
        breakData.win->setWindowTitle(bs.title);
    }
}

void BreakSetMaxSteps(ulong steps)
{
    breakSettingsCurrent.steps = steps;
    if (breakData.win) {
        breakData.win->updateProgress(breakSettingsCurrent);
    }
}

ulong BreakCurrentSteps(void)
{
    return breakSettingsCurrent.current;
}

void BreakSetCurrentStep(ulong current)
{
    BreakCheck();
    breakSettingsCurrent.current = current;
    if (breakData.win) {
        breakData.win->updateProgress(breakSettingsCurrent);
    }
}

void BreakAddCurrentStep(long add)
{
    BreakCheck();
    breakSettingsCurrent.current += add;
    if (breakData.win) {
        breakData.win->updateProgress(breakSettingsCurrent);
    }
}

#include "qtBreak.moc"
