//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtBuildGui.cc
//
// Creation date: 03.03.2009
// Last modified: 12.10.2021
//

#include "system.h"
#include "qt/qtBuildGui.h"
#include "qt/qtBuildGuiField.h"
#include "qt/qtBuildGuiWindow.h"
#include "qt/qtBuildGuiTable.h"
#include "qt/qtUtil.h"
#include "qt/classes/qtPopImage.h"
#include <QLabel>
#include <QTabWidget>
#include <QSplitter>
#include <QVariant>
#include "buildgui.h"
#include "gui.h"
#include "project.h"
#include "display.h"

Q_DECLARE_OPAQUE_POINTER(struct DisplayObject *);
Q_DECLARE_METATYPE(struct DisplayObject *);

DisplayObject *GuiObjectDisplayObject(GuiObject go) 
{ 
    DisplayObject *dio = 0;

    if (go) {
        QVariant v = go->property("dio");
        dio = v.value<DisplayObject *>();
    }
    return dio;
}

void SetGuiObjectDisplayObject(GuiObject go, struct DisplayObject *dio)
{
    if (go) {
        QVariant v = qVariantFromValue(dio);
        go->setProperty("dio", v);
    }
}

static void FreeProjectGuiRec(DisplayObject *dio)
{
    List *childs = DisplayObjectObjects(dio);
    short type = DisplayObjectType(dio);
    GuiObject go = DisplayObjectGuiObject(dio);
    bool open;

    if (dio) {
        switch(type) {
            case DO_TYPE_WINDOW:
                if (go) {
                    open = GuiWindowOpen(go);   /* remember current state */
                    GuiCloseWindow(go);
                    go->deleteLater();
                    SetDisplayWindowOpen(dio, open);
                    SetDisplayWindowButtonObject(dio, 0);
                }
                break;

            case DO_TYPE_FIELD:
                SetDisplayFieldListviewObject(dio, 0);
                SetDisplayFieldFileObject(dio, 0);
                SetDisplayFieldFontObject(dio, 0);
                SetDisplayFieldAutoShowObject(dio, 0);
                SetDisplayFieldFilterObject(dio, 0);
                SetDisplayFieldViewObject(dio, 0);
                SetDisplayFieldPopObject(dio, 0);
                SetDisplayFieldDisableObject(dio, 0);
                break;

            case DO_TYPE_TABLE:
                SetDisplayTablePanelNumAllObject(dio, 0);
                SetDisplayTablePanelFilterObject(dio, 0);
                break;
	}

        SetDisplayObjectGuiObject(dio, 0);
        SetDisplayObjectRootObject(dio, 0);
        SetDisplayObjectActivateObject(dio, 0);

    	if (childs) {
	    FORNODE(dio, childs) {
	    	FreeProjectGuiRec(dio);
	    }
	}
    }
}

void FreeProjectGui(Project *pr)
{
    DisplayObject *dio = ProjectRootDisplayObject(pr);

    GuiSetSleep(true);
    FreeProjectGuiRec(dio);
    GuiSetSleep(false);
}

// forward declaration
static bool BuildGuiObject(DisplayObject *dio, bool hideLabel, QWidget *parent,
        GroupLayout *group, GroupLayout **colGroup);

void BuildGuiSetBackground(DisplayObject *dio, QWidget *widget)
{
    if (widget) {
        ProjectSetCurrentDir(DisplayObjectProject(dio));
        qtSetWidgetBackground(widget, DisplayObjectBackground(dio));
    }
}

void BuildGuiSetFontname(const char *fontname, QWidget *widget)
{
    if (widget && fontname && fontname[0]) {
        widget->setFont(qtFontFromDescription(fontname));
    }
}

void BuildGuiSetFont(DisplayObject *dio, QWidget *widget)
{
    BuildGuiSetFontname(DisplayObjectQtFontname(dio), widget);
}

static bool buildGuiText(DisplayObject *dio, QWidget *parent,
        GroupLayout *group)
{
    if (!parent || !group) {
	return false;
    }

    const char *title = DisplayObjectTitle(dio);
    int weight = DisplayObjectWeight(dio);
    QLabel *label = new QLabel(title, parent);
    
    label->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed);
    group->addWidget(label);
    group->setStretch(weight, 0);
    BuildGuiSetBackground(dio, label);
    BuildGuiSetFont(dio, label);
    SetDisplayObjectGuiObject(dio, label);
    SetDisplayObjectRootObject(dio, label);
    SetGuiObjectDisplayObject(label, dio);

    return true;
}

static bool buildGuiImage(DisplayObject *dio, QWidget *parent, 
        GroupLayout *group)
{
    if (!group) {
	return false;
    }

    Project *pr = DisplayObjectProject(dio);
    const char *bg = DisplayObjectBackground(dio);
    char imageSize = DisplayImageImageSize(dio);
    bool resize = (imageSize == DAD_IMAGESIZE_RESIZABLE);
    int weight = DisplayObjectWeight(dio);
    if (!resize) {
	weight = 0;
    }

    QLabel *label = new QLabel(parent);
    label->setAlignment(Qt::AlignCenter);

    ProjectSetCurrentDir(pr);           // ensure we have correct CWD
    if (bg && bg[0] == '5' && bg[1] == ':') {
        QString path = qtExpandPathname(&bg[2]);
        QPixmap pixmap(path);
        if (!pixmap.isNull()) {
            label->setPixmap(pixmap);
        }
    } else {
        BuildGuiSetBackground(dio, label);
    }
    group->addWidget(label);
    group->setStretch(weight, weight);
    SetGuiObjectDisplayObject(label, dio);
    SetDisplayObjectGuiObject(dio, label);
    SetDisplayObjectRootObject(dio, label);

    return true;
}

static bool buildGuiSpace(DisplayObject *dio, QWidget *parent,
        GroupLayout *group)
{
    if (!group) {
	return false;
    }

    bool horiz = (group->columns() != 1);
    bool bar = DisplaySpaceBar(dio);
    int weight = DisplayObjectWeight(dio);

    QFrame::Shape shape;
    if (!bar) {
        shape = QFrame::NoFrame;
    } else if (horiz) {
        shape = QFrame::VLine;
    } else {
        shape = QFrame::HLine;
    }

    QFrame *frame = new QFrame(parent);
    frame->setFrameShape(shape);
    frame->setFrameShadow(QFrame::Raised);
    group->addWidget(frame);
    group->setStretch(weight, weight);
    SetGuiObjectDisplayObject(frame, dio);
    BuildGuiSetBackground(dio, frame);
    SetDisplayObjectGuiObject(dio, frame);
    SetDisplayObjectRootObject(dio, frame);

    return true;
}

class Register : public QTabWidget
{
    Q_OBJECT

public:
    Register(QWidget *parent);

private slots:
    void onCurrentChanged(int index);    
};

Register::Register(QWidget *parent) : QTabWidget(parent)
{
    connect(this, SIGNAL(currentChanged(int)),
            this, SLOT(onCurrentChanged(int)));
}

void Register::onCurrentChanged(int index)
{
    DisplayObject *dio = GuiObjectDisplayObject(this);
    Project *pr = DisplayObjectProject(dio);

    GuiUpdateRegister(this, pr);
}

static bool buildGuiRegister(DisplayObject *dio, QWidget *parent, 
        GroupLayout *group)
{
    if (!group) {
	return false;
    }

    Register *reg = new Register(parent);
    group->addWidget(reg);
    int weight = DisplayObjectWeight(dio);
    group->setStretch(weight, weight);
    SetGuiObjectDisplayObject(reg, dio);
    SetDisplayObjectGuiObject(dio, reg);
    SetDisplayObjectRootObject(dio, reg);

    char **labels = DisplayRegisterLabels(dio);
    List *childs = DisplayObjectObjects(dio);

    if (childs) {    
	int i = 0;
	DisplayObject *child;
	FORNODE(child, childs) {
            if (!DisplayObjectVisible(child)) {
                continue;
            }
            QWidget *widget = new QWidget(reg);
            GroupLayout *group = new GroupLayout(1, widget);
            GroupLayout *colGroup = 0;

	    if (!BuildGuiObject(child, false, widget, group, &colGroup)) {
    		return false;
	    }
	    BuildGuiFinishColGroup(&colGroup);
            QString label;
	    if (labels && labels[i]) {
		label = GuiMarkShortcut(labels[i]);
		i++;
	    }
            reg->addTab(widget, label);
	}
    }
    
    GuiDisplayObjectSizeConstraints constraints;
    BuildGuiDisplayObjectGetSizeConstraints(dio, false, &constraints);
    
    reg->setSizePolicy(constraints.can_grow_horiz? 
                       QSizePolicy::Expanding : QSizePolicy::Fixed,
                       constraints.can_grow_vert? 
                       QSizePolicy::Expanding : QSizePolicy::Fixed);
    return true;
}

static bool buildGuiGroup(DisplayObject *dio, QWidget *parent,
        GroupLayout *group, GroupLayout **colGroup) 
{
    if (!group || !colGroup) {
	return false;
    }

    bool hideLabel = false;
    bool horiz = DisplayGroupHoriz(dio);
    bool border = DisplayGroupBorder(dio);
    const char *title = DisplayObjectTitle(dio);
    List *childs = DisplayObjectObjects(dio);

    if (title && title[0] == 0) {
        title = 0;
    }

    if (group->columns() == 1 && horiz && !border && !title && childs) {
	DisplayObject *child;
	FORNODE(child, childs) {
	    if (!DisplayObjectVisibleGui(child)) {
		continue;
	    } 
	    if (DisplayObjectType(child) != DO_TYPE_FIELD) {
		break;
	    } 
	    if (DisplayFieldTitlePosition(child) != DAD_TITLEPOSITION_LEFT) {
		break;
	    }
	    if (!BuildGuiFieldHasLabel(child)) {
		break;
	    }

	    if (!*colGroup) {
                *colGroup = new GroupLayout(2);
                (*colGroup)->setContentsMargins(0, 0, 0, 0);
                group->addLayout(*colGroup);
                group->setStretch(100, 100);
	    }
            group = *colGroup;
	    if (!BuildGuiFieldLabel(child, parent, group)) {
	    	return false;
	    }
	    hideLabel = true;
	    break;
	}
    }
    if (!hideLabel) {
	BuildGuiFinishColGroup(colGroup);
    }

    int weight = DisplayObjectWeight(dio);
    bool spacing = DisplayGroupSpacing(dio);

    QWidget *widget = 0;
    if (title && title[0]) {
        GroupBox *box = new GroupBox(title, parent);
        box->setFlat(!border);
        parent = widget = box;
    } else if (border) {
        QFrame *frame = new QFrame(parent);
        frame->setFrameShape(QFrame::StyledPanel);
        frame->setFrameShadow(QFrame::Raised);
        parent = widget = frame;
    }
    
    GroupLayout *newGroup;
    if (horiz) {
        newGroup = new GroupLayout(INT_MAX, widget);
    } else {
        newGroup = new GroupLayout(1, widget);
    }
    if (!spacing) {
        newGroup->setSpacing(0);
    }
    if (!border) {
        newGroup->setContentsMargins(0, 0, 0, 0);
    }
    if (widget) {
        group->addWidget(widget);
    } else {
        group->addLayout(newGroup);
    }

    BuildGuiSetBackground(dio, widget);
    SetDisplayObjectGuiObject(dio, widget);
    SetDisplayObjectRootObject(dio, widget);
    SetGuiObjectDisplayObject(widget, dio);

    GroupLayout *newColGroup = 0;
    bool rc = BuildGuiChilds(childs, hideLabel, parent, newGroup, &newColGroup);
    BuildGuiFinishColGroup(&newColGroup);

    Qt::Orientations orientations = newGroup->expandingDirections();
    int horizWeight = orientations & Qt::Horizontal ? weight : 0;
    int vertWeight = orientations & Qt::Vertical ? weight : 0;
    group->setStretch(horizWeight, vertWeight);

    return rc;
}

static bool isBalance(DisplayObject *dio)
{
    short type = DisplayObjectType(dio);
    bool visible = DisplayObjectVisible(dio);

    return (type == DO_TYPE_BALANCE && visible);
}

static bool hasBalance(List *childs)
{
    struct DisplayObject *child;
    FORNODE (child, childs) {
        if (isBalance(child)) {
            return true;
        }
    }
    return false;
}

static QList<int> getDisplayObjectSizes(List *dios)
{
    QList<int> sizes;

    if (dios != NULL) {
        DisplayObject *dio;
        int sumWeights = 0;

        FORNODE(dio, dios) {
            if (isBalance(dio)) {
                sizes.push_back(sumWeights);
                sumWeights = 0;
            } else {
                sumWeights += DisplayObjectWeight(dio);
            }
        }
        sizes.push_back(sumWeights);
    }
    return sizes;
}

class Splitter : public QSplitter
{
    Q_OBJECT

public:
    Splitter(List *dios, QWidget *parent) : QSplitter(parent), dios_(dios)
    {
        connect(this, SIGNAL(splitterMoved(int, int)),
                this, SLOT(onSplitterMoved(int, int)));
    }

private slots:
    void onSplitterMoved(int pos, int index)
    {
        if (dios_) {
            QList<int> oldSizes = getDisplayObjectSizes(dios_);
            QList<int> newSizes = sizes();
            DisplayObject *dio;
            int ix = 0;

            FORNODE(dio, dios_) {
                if (ix >= oldSizes.size() || ix >= newSizes.size()) {
                    break;
                }
                if (isBalance(dio)) {
                    ++ix;
                } else {
                    int old = oldSizes[ix];
                    if (old == 0) {
                        old = 1;
                    }
                    SetDisplayObjectWeight(dio,
                            DisplayObjectWeight(dio) * newSizes[ix] / old);
                }
            }
        }
    }

private:
    List *dios_;
};

bool BuildGuiChilds(List *childs, bool hideLabel, QWidget *parent,
        GroupLayout *group, GroupLayout **colGroup)
{
    if (childs && parent && group && colGroup) {    
        bool horiz = (group->columns() != 1);
        int hs = group->horizontalSpacing();
        int vs = group->verticalSpacing();
        Splitter *splitter = 0;
        bool expandHoriz = false, expandVert = false;

        if (hasBalance(childs)) {
            BuildGuiFinishColGroup(colGroup);
            splitter = new Splitter(childs, parent);
            splitter->setOrientation(horiz? Qt::Horizontal : Qt::Vertical);
            splitter->setChildrenCollapsible(false);
            group->addWidget(splitter);
            group->setStretch(100, 100);

            QWidget *widget = new QWidget();
            group = new GroupLayout(horiz? INT_MAX : 1, widget);
            group->setHorizontalSpacing(hs);
            group->setVerticalSpacing(vs);
            group->setContentsMargins(0, 0, 0, 0);
            splitter->addWidget(widget);
            parent = widget;
        }
	DisplayObject *child;
	FORNODE(child, childs) {
            if (isBalance(child) && splitter) {
                BuildGuiFinishColGroup(colGroup);
                SetDisplayObjectGuiObject(child, splitter);

                Qt::Orientations orientations = group->expandingDirections();
                expandHoriz |= (orientations & Qt::Horizontal) != 0;
                expandVert |= (orientations & Qt::Vertical) != 0;
                QWidget *widget = new QWidget();
                group = new GroupLayout(horiz? INT_MAX : 1, widget);
                group->setHorizontalSpacing(hs);
                group->setVerticalSpacing(vs);
                group->setContentsMargins(0, 0, 0, 0);
                splitter->addWidget(widget);
                parent = widget;
            } else {
                if (!BuildGuiObject(child, hideLabel, parent, group, 
                        colGroup)) {
                    return false;
                }
	    }
	    if (hideLabel && DisplayObjectGuiObject(child)) {
		hideLabel = false;
	    }
	}
        if (splitter) {
            splitter->setSizes(getDisplayObjectSizes(childs));

            Qt::Orientations orientations = group->expandingDirections();
            expandHoriz |= (orientations & Qt::Horizontal) != 0;
            expandVert |= (orientations & Qt::Vertical) != 0;
            splitter->setSizePolicy(
                    expandHoriz ?  QSizePolicy::Expanding : QSizePolicy::Fixed,
                    expandVert ?  QSizePolicy::Expanding : QSizePolicy::Fixed);
        }
    }
    BuildGuiFinishColGroup(colGroup);

    return true;
}

void BuildGuiFinishColGroup(GroupLayout **colGroup)
{
    if (colGroup && *colGroup) {
	*colGroup = 0;
    }
}

static bool BuildGuiObject(DisplayObject *dio, bool hideLabel, QWidget *parent,
        GroupLayout *group, GroupLayout **colGroup)
{
    if (!DisplayObjectVisibleGui(dio)) {
	return true;
    }

    switch (DisplayObjectType(dio)) {
	case DO_TYPE_FIELD:
	    return BuildGuiField(dio, hideLabel, parent, group, colGroup);

	case DO_TYPE_TEXT:
	    BuildGuiFinishColGroup(colGroup);
    	    return buildGuiText(dio, parent, group);

	case DO_TYPE_IMAGE:
	    BuildGuiFinishColGroup(colGroup);
	    return buildGuiImage(dio, parent, group);

	case DO_TYPE_SPACE:
	    BuildGuiFinishColGroup(colGroup);
	    return buildGuiSpace(dio, parent, group);

	case DO_TYPE_BALANCE:
	    BuildGuiFinishColGroup(colGroup);
            // there is nothing else to do since we handle this using QSplitter
	    return true;
	    
	case DO_TYPE_TABLE:
	    BuildGuiFinishColGroup(colGroup);
	    return BuildGuiTable(dio, parent, group);

	case DO_TYPE_REGISTER:
	    BuildGuiFinishColGroup(colGroup);
	    return buildGuiRegister(dio, parent, group);

	case DO_TYPE_WINDOW:
	    return BuildGuiWindowButton(dio, parent, group, colGroup);

	case DO_TYPE_GROUP:
	    return buildGuiGroup(dio, parent, group, colGroup);

	default:
	    return false;
    }
}

#include "qtBuildGui.moc"

