//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtBuildGuiField.cc
//
// Creation date: 05.06.2010
// Last modified: 09.07.2022
//

#include "system.h"
#include "qt/qtBuildGuiField.h"
#include "qt/qtBuildGui.h"
#include "qt/qtBuildGuiRefFieldBrowser.h"
#include "qt/qtProjectWindow.h"
#include "qt/classes/qtImageLabel.h"
#include "qt/classes/qtPopWidget.h"
#include "qt/classes/qtPopSelection.h"
#include "qt/classes/qtSFWView.h"
#include "qt/qtUtil.h"
#include <QButtonGroup>
#include <QCalendarWidget>
#include <QCheckBox>
#include <QComboBox>
#include <QLabel>
#include <QLineEdit>
#include <QListWidget>
#include <QMimeData>
#include <QPushButton>
#include <QScrollArea>
#include <QStyle>
#include <QTimer>
#include <QToolButton>
#include "lang/program.h"
#include "programeditor.h"
#include "buildgui.h"
#include "display.h"
#include "gui.h"
#include "field.h"
#include "table.h"
#include "guidosutil.h"
#include "requesters.h"
#include "preferences.h"
#include "memo.h"
#include "labels.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

static QPointer<QWidget> changedGuiObject = 0;

static QLabel *labelObject = 0;

Qt::Alignment align[] =
{
    Qt::AlignHCenter, Qt::AlignLeft, Qt::AlignRight
};

bool BuildGuiFieldLabel(DisplayObject *dio, QWidget *parent,
        GroupLayout *group)
{
    GuiFieldFeatures f;

    if (!group || !BuildGuiFieldGetFeatures(dio, &f)) {
	return false;
    }
    if (!f.label) {
	return true;	// nothing to do
    }
    
    const char *title = DisplayObjectTitle(dio);
    QString text = GuiMarkShortcut(title, DisplayObjectShortcut(dio)); 
    QLabel *label = new QLabel(text, parent);

    Qt::AlignmentFlag align;
    switch (DisplayFieldTitlePosition(dio)) {
	case DAD_TITLEPOSITION_LEFT:
            align = Qt::AlignRight;
	    break;
	case DAD_TITLEPOSITION_RIGHT:
            align = Qt::AlignLeft;
	    break;
	default:
            align = Qt::AlignHCenter;
	    break;
    }
    label->setAlignment(align | Qt::AlignVCenter);
    label->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed);
    BuildGuiSetFontname(DisplayFieldTitleQtFontname(dio), label);
    
    group->addWidget(label);
    labelObject = label;

    return true;
}

class TextGadget : public QLabel
{
public:
    TextGadget(DisplayObject *dio, QWidget *parent) : QLabel(parent)
    {
        setFrameShadow(QFrame::Sunken);
        setFrameShape(QFrame::StyledPanel);
        setText(DisplayObjectTitle(dio));
        setAlignment(align[(int)DisplayFieldAlignment(dio)]);
        setMargin(1);
        setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
        setMinimumWidth(50);
        Field *field = DisplayFieldField(dio);
        if (FieldType(field) == TYPE_REFERENCE) {
            setAcceptDrops(true);
        }
    }
    
    void dragEnterEvent(QDragEnterEvent *event)
    {
        DisplayObject *dio = GuiObjectDisplayObject(this);
        Field *field = DisplayFieldField(dio);
        Table *reftab = FieldReferenceTable(field);
        if (reftab && event->mimeData()->hasFormat(MimeTypeTableRecord)) {
            QByteArray encodedData = event->mimeData()->data(MimeTypeTableRecord);
            QDataStream stream(&encodedData, QIODevice::ReadOnly);
            while (!stream.atEnd()) {
                MimeDataTableRecord mdtr;
                stream >> mdtr;
                Table *tab = mdtr.table();
                if (tab == reftab) {
                    event->setDropAction(Qt::CopyAction);
                    event->accept();
                    break;
                }
            }
        }
    }
    
    void dropEvent(QDropEvent *event)
    {
        DisplayObject *dio = GuiObjectDisplayObject(this);
        Field *field = DisplayFieldField(dio);
        Table *reftab = FieldReferenceTable(field);
        if (reftab && event->mimeData()->hasFormat(MimeTypeTableRecord)) {
            QByteArray encodedData = event->mimeData()->data(MimeTypeTableRecord);
            QDataStream stream(&encodedData, QIODevice::ReadOnly);
            while (!stream.atEnd()) {
                MimeDataTableRecord mdtr;
                stream >> mdtr;
                Table *tab = mdtr.table();
                if (tab == reftab) {
                    event->acceptProposedAction();
                    Record *rec = mdtr.record(tab);
                    GuiStoreEdits();
                    GuiStoreDataRecord(dio, rec);                    
                }
            }
        }
    }
};

static QWidget *buildGuiFieldText(DisplayObject *dio, QWidget *parent)
{
    return new TextGadget(dio, parent);
}

class StringGadget : public StringEdit
{
    Q_OBJECT

public:
    StringGadget(QWidget *parent) : StringEdit(parent) 
    {
        connect(this, SIGNAL(textChanged(const QString &)), 
                this, SLOT(onTextChanged()));
        connect(this, SIGNAL(returnPressed()), this, SLOT(onReturnPressed()));
    }

private slots:
    void onTextChanged() { changedGuiObject = this; }
    void onReturnPressed() { GuiStoreEdits(); }
};

static QWidget *buildGuiFieldString(DisplayObject *dio, int maxchars, 
        QWidget *parent)
{
    StringGadget *edit = new StringGadget(parent);
    edit->setMaxLength(maxchars);
    edit->setAlignment(align[(int)DisplayFieldAlignment(dio)]);
    // there seems to be no 'minimum characters visible' setting in Qt
    edit->setMinimumWidth(50);

    return edit;
}

class MemoGadget : public TextEditor
{
    Q_OBJECT

public:
    MemoGadget(Project *pr, QWidget *parent) : TextEditor(pr, false, parent)
    {
        setTabChangesFocus(true);
        connect(this, SIGNAL(textChanged()), this, SLOT(onTextChanged()));
    }

private slots:
    void onTextChanged() { changedGuiObject = this; }
};

static QWidget *buildGuiFieldMemo(DisplayObject *dio, QWidget *parent)
{
    MemoGadget *edit = new MemoGadget(DisplayObjectProject(dio), parent);
    edit->setReadOnly(DisplayFieldReadOnly(dio));
    return edit;
}

class CheckBox : public QCheckBox
{
    Q_OBJECT

public: 
    CheckBox(QWidget *parent) : QCheckBox(parent)
    {
        suppressOnToggled_ = false;
        connect(this, SIGNAL(toggled(bool)), this, SLOT(onToggled(bool)));
    }
    
    void setChecked(bool checked) {
        suppressOnToggled_ = true;
        QCheckBox::setChecked(checked);
        suppressOnToggled_ = false;
    }

public slots:
    void onToggled(bool checked)
    {
        if (!suppressOnToggled_) {
            GuiStoreEdits();
            GuiStoreDataBool(GuiObjectDisplayObject(this), checked);
        }
    }
private:
    bool suppressOnToggled_;
};

static QWidget *buildGuiFieldCheckBox(DisplayObject *dio, QWidget *parent)
{
    CheckBox *check = new CheckBox(parent);

    check->setEnabled(!DisplayFieldReadOnly(dio));
    return check;
}

class ChoiceGadget : public ComboBox
{
    Q_OBJECT
    
public:
    ChoiceGadget(QWidget *parent, char **labels) : ComboBox(parent, labels) {
        suppressOnCurrentIndexChanged_ = false;
        connect(this, SIGNAL(currentIndexChanged(int)),
                this, SLOT(onCurrentIndexChanged(int)));
    }
    
    void setCurrentIndex(int index) {
        suppressOnCurrentIndexChanged_ = true;
        ComboBox::setCurrentIndex(index);
        suppressOnCurrentIndexChanged_ = false;
    }
    
public slots:
    void onCurrentIndexChanged(int index) {
        if (!suppressOnCurrentIndexChanged_) {
            GuiStoreEdits();
            GuiStoreDataLong(GuiObjectDisplayObject(this), index);
        }
    }
    
private:
    bool suppressOnCurrentIndexChanged_;
};

static QWidget *buildGuiFieldChoice(DisplayObject *dio, QWidget *parent)
{
    char **labels = FieldChoiceLabels(DisplayFieldField(dio));

    ChoiceGadget *choice = new ChoiceGadget(parent, labels);
    choice->setEnabled(!DisplayFieldReadOnly(dio));

    return choice;
}

class RadioGadget : public RadioBox
{
    Q_OBJECT
    
public:
    RadioGadget(QWidget *parent, DisplayObject *dio) :
        RadioBox(parent,
                FieldChoiceLabels(DisplayFieldField(dio)),
                DisplayObjectTitle(dio), DisplayFieldFrame(dio),
                DisplayFieldAlignment(dio) == DAD_ALIGNMENT_HORIZONTAL)
    {
        connect(this, SIGNAL(checked(int)), this, SLOT(onChecked(int)));
    }

    void setCheckedId(int id)
    {
        RadioBox::setCheckedId(id);
        currentId_ = id;
    }
    
public slots:
    void onChecked(int id) {
        if (currentId_ != id) {
            currentId_ = id;
            GuiStoreEdits();
            GuiStoreDataLong(GuiObjectDisplayObject(this), id);
        }
    }

private:
    int currentId_;
};

static QWidget *buildGuiFieldRadioBox(DisplayObject *dio, QWidget *parent)
{
    RadioGadget *radio = new RadioGadget(parent, dio);
    return radio;
}

class UserButton : public QPushButton
{
    Q_OBJECT

public:
    UserButton(QWidget *parent) : QPushButton(parent)
    {
        connect(this, SIGNAL(clicked()), this, SLOT(onClicked()));
    }

private slots:
    void onClicked()
    {
        DisplayObject *dio = GuiObjectDisplayObject(this);
        GuiStoreEdits();
        GuiStoreData(dio);
    }
};

static QWidget *buildGuiFieldButton(DisplayObject *dio, QWidget *parent)
{
    UserButton *button = new UserButton(parent);
    const char *title = DisplayObjectTitle(dio);
    button->setText(GuiMarkShortcut(title, DisplayObjectShortcut(dio)));             
    button->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
    return button;
}

class ShowButton : public ExtraButton
{
    Q_OBJECT
    
public:
    ShowButton(QWidget *parent, QWidget *string) : ExtraButton(parent, string)
    {
        connect(this, SIGNAL(clicked()), this, SLOT(onClicked()));
    }
    
private slots:
    void onClicked()
    {
        DisplayObject *dio = GuiObjectDisplayObject(this);
        Field *field = DisplayFieldField(dio);
        GuiStoreEdits();
        GuiShowReferenceRecord(field, TableGuiRecord(FieldTable(field)));
    }
};

static QWidget *buildGuiFieldShowButton(DisplayObject *dio, QWidget *string)
{
    ShowButton *button = new ShowButton(string->parentWidget(), string);
    // button->setText("S");
    button->setIcon(button->style()->standardIcon(QStyle::SP_CommandLink, 0, button));
    button->setToolTip(GCS(MSG_ReferenceShowBH));
    SetGuiObjectDisplayObject(button, dio);
    return button;
}

static QWidget *buildGuiFieldImageButton(DisplayObject *dio,
        QWidget **pbutton, QWidget *parent)
{
    struct Project *pr = DisplayObjectProject(dio);
    const char *bg = DisplayObjectBackground(dio);
    char imageSize = DisplayFieldImageSize(dio);
    bool resize = (imageSize == DAD_IMAGESIZE_RESIZABLE);
    QPixmap pixmap;

    ProjectSetCurrentDir(pr);           /* ensure we have correct CWD */
    if (bg && bg[0] == '5' && bg[1] == ':') {
        QString path = qtExpandPathname(&bg[2]);
        pixmap.load(path);
    }

    QToolButton *button = new QToolButton(parent);
    QIcon icon(pixmap);
    button->setIcon(icon);
    if (!pixmap.isNull()) {
        button->setIconSize(pixmap.size());
    }

    if (!resize) {
        button->setFixedSize(button->sizeHint());
    } else {
        // for some unknown reasons QWIDGETSIZE_MAX itself does not work
        button->setMaximumSize(QSize(QWIDGETSIZE_MAX/2, QWIDGETSIZE_MAX/2));
    }

    //g_signal_connect(*button, "clicked", G_CALLBACK(buttonOnClicked), dio);
    *pbutton = button;
    QWidget *go = button;
    if (!resize) {
        go = new QWidget(parent);
        GroupLayout *hcenter = new GroupLayout(3, go);
        hcenter->addWidget(new QFrame(go));
        button->setParent(go);
        hcenter->addWidget(button);
        hcenter->addWidget(new QFrame(go));
    }

    return go;
}
    
void SFWViewOnRowActivated(SFWView *sfwView, const QModelIndex &index, DisplayObject *dio)
{
    Field *field;
    Record *rec;
    long line, col;
    
    sfwView->getEntry(index, &line, &col, &field, &rec);
    GuiVirtualFieldListDoubleClick(dio, line, col, field, rec);
}

class SFWViewGadget : public SFWView
{
    Q_OBJECT
    
public:
    SFWViewGadget(QWidget *parent, DisplayObject *dio) : 
        SFWView(parent), dio_(dio)
    {
        setShowTitles(DisplayFieldShowTitles(dio));
        if (acceptUrlDrops(0)) {
            setAcceptDrops(true);
        }
        connect(this, SIGNAL(activated(const QModelIndex &)), 
                this, SLOT(onActivated(const QModelIndex &)));
        connect(this, SIGNAL(internalMoveDrop(int)), 
                this, SLOT(onInternalMoveDrop(int)), Qt::QueuedConnection);
    }
        
protected slots:
    void onActivated(const QModelIndex &index) 
    {
        Field *field;
        Record *rec;
        long line, col;
        
        getEntry(index, &line, &col, &field, &rec);
        GuiVirtualFieldListDoubleClick(dio_, line, col, field, rec);
    }
   
    void currentChanged(const QModelIndex &current,
                        const QModelIndex &previous)
    {
        SFWView::currentChanged(current, previous);
        if (current.row() != previous.row()) {
            GuiVirtualFieldListActiveChanged(dio_);
        }
    }

    void onInternalMoveDrop(int row)
    {
        GuiVirtualFieldListSortDrop(dio_, row);
    }

protected:
    bool acceptUrlDrops(const QDropEvent *event) const
    {
        const char *urlDropTrigger = DisplayFieldURLDropTrigger(dio_);

        // Check that the display object has a valid drop URL trigger.
        if (urlDropTrigger == 0 || urlDropTrigger[0] == 0) {
            return false;
        }

        // If there is no event then accept drops.
        if (!event) {
            return true;
        }

        // Check that the event is a copy or link action.
        if (event->proposedAction() != Qt::CopyAction &&
                event->proposedAction() != Qt::LinkAction) {
            return false;
        }

        // Check that this is an URL target.
        const QMimeData *mime = event->mimeData();
        return mime && mime->hasUrls();
    }

    void dragEnterEvent(QDragEnterEvent *event)
    {
        if (acceptUrlDrops(event)) {
            event->accept();
        } else {
            SFWView::dragEnterEvent(event);
        }
    }

    void dropEvent(QDropEvent *event)
    {
        if (acceptUrlDrops(event)) {
            Project *pr = DisplayObjectProject(dio_);
            MemPool *memPool = ProjectMemPool(pr);
            const QList<QUrl> &urls = event->mimeData()->urls();
            size_t num = urls.size(), n = 0;
            char **labels = NULL;
            long error = OK;

            labels = AllocLabels(memPool, num);
            if (labels == NULL) {
                error = ERR_MEM;
            } else {
                for (QList<QUrl>::const_iterator it = urls.begin();
                        it != urls.end(); ++it) {
                    QString filename = it->isLocalFile() ?
                        it->toLocalFile() : it->url();
                    if (!filename.isEmpty()) {
                        labels[n] =
                            AllocString(memPool, filename.toUtf8().constData());
                        if (labels[n] == NULL) {
                            error = ERR_MEM;
                            break;
                        }
                        n++;
                    }
                }
            }
            if (error) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            } else {
                GuiVirtualFieldListDropUrls(dio_, labels);
            }
            FreeLabelsNum(memPool, labels, num);

            event->acceptProposedAction();
        } else {
            SFWView::dropEvent(event);
        }
    }

private:
    DisplayObject *dio_;
};

static QWidget *buildGuiFieldSFWView(DisplayObject *dio, QWidget *parent)
{
    return new SFWViewGadget(parent, dio);
}

class PopASL : public PopSelection
{
public:
    PopASL(DisplayObject *dio, QWidget *string, QAbstractButton *button, PopSelection::Type type) 
    {
        setType(type);
        setString(string);
        setButton(button);

        if (!DisplayFieldTabChain(dio)) {
            button->setFocusPolicy(Qt::NoFocus);
        }
        SetGuiObjectDisplayObject(button, dio);
    }
    
    virtual void stop()
    {
        PopSelection::stop();
        GuiStoreEdits();
    }
};

static QHBoxLayout *buildGuiPopFile(DisplayObject *dio, QWidget *string,
	QWidget **button, QWidget *parent)
{
    PopButton *popButton = new PopButtonSetCWD(parent, string, 
                                               DisplayObjectProject(dio));
    popButton->setIcon(popButton->style()->standardIcon(QStyle::SP_FileIcon));    
    *button = popButton;
    
    return new PopASL(dio, string, popButton, PopSelection::FileSelection);    
}

static QHBoxLayout *buildGuiPopFont(DisplayObject *dio, QWidget *string,
                                    QWidget **button, QWidget *parent)
{
    PopButton *popButton = new PopButton(parent, string);
    popButton->setIcon(popButton->style()->standardIcon(QStyle::SP_FileIcon));    
    *button = popButton;
    
    return new PopASL(dio, string, popButton, PopSelection::FontSelection);    
}

class PopCalendar : public PopWidget
{
    Q_OBJECT

public:
    PopCalendar(DisplayObject *dio, QWidget *string, QAbstractButton *button,
            QCalendarWidget *calendar)
        : dio_(dio), calendar_(calendar)
    {
        setString(string);
        setButton(button);
        setPopupWidget(calendar);
        connect(calendar, SIGNAL(activated(const QDate &)),
                this, SLOT(onActivated(const QDate &)));
    }

    virtual bool start();
    virtual void stop();

private slots:
    void onActivated(const QDate &unused) { close(true); }
    void saveDate();

private:
    DisplayObject *dio_;
    long date_;
    QCalendarWidget *calendar_;
};

bool PopCalendar::start()
{
    long date;
    short day, month, year;

    if (StringToDate(text().toUtf8().data(), &date)) {
        DateToDayMonthYear(date, &day, &month, &year);
        calendar_->setSelectedDate(QDate(year, month, day));
    }
    return true;
}

void PopCalendar::stop()
{
    QDate qdate = calendar_->selectedDate();

    if (DayMonthYearToDate(qdate.day(), qdate.month(), qdate.year(), &date_)) {
        // We postpone storing data in order to close the popup first.
        // Otherwise, a dialog in a user program will not receive input.
        QTimer::singleShot(0, this, SLOT(saveDate()));
    }
}

void PopCalendar::saveDate()
{
    Project *pr = DisplayObjectProject(dio_);
    const char *text = DateToString(date_, GetPrefsValue(pr, PREFS_DATEFORMAT));
    GuiStoreDataStr(dio_, text);
}

static QHBoxLayout *buildGuiPopCalendar(DisplayObject *dio, QWidget *string,
	QWidget **button, QWidget *parent)
{
    PopButton *popButton = new PopButton(parent, string);
    popButton->setArrowType(Qt::DownArrow);
    if (!DisplayFieldTabChain(dio)) {
        popButton->setFocusPolicy(Qt::NoFocus);
    }
    SetGuiObjectDisplayObject(popButton, dio);

    *button = popButton;
    
    QCalendarWidget *calendar = new QCalendarWidget(parent);
    PopCalendar *pop = new PopCalendar(dio, string, popButton, calendar);
    
    return pop;
}

class PopListview : public PopWidget
{
    Q_OBJECT

public:
    PopListview(DisplayObject *dio, QWidget *string, QAbstractButton *button,
            QListWidget *list) 
        : dio_(dio), list_(list)
    {
        setString(string);
        setButton(button);
        setPopupWidget(list);
    }

    virtual bool start();
    virtual void stop();

private slots:
    void saveText();

private:
    DisplayObject *dio_;
    QString text_;
    QListWidget *list_;
};

bool PopListview::start()
{
    struct Project *pr = DisplayObjectProject(dio_);
    const char *trigger = DisplayFieldListviewCompute(dio_);
    struct MemPool *memPool = ProjectMemPool(pr);
    char **labels = NULL;
    bool labelsAllocated = FALSE;

    GuiStoreEdits();   // maybe the user has just edited a string object

    if (trigger != NULL && trigger[0] != 0) {
	Program *prg = ProjectProgram(pr);
	struct UserFunction *uf = FindUserFunctionName(prg, trigger);
	const char *filename = NULL, *error = NULL, *elemError = NULL;
	long line, col;
	struct Element *elem;
	char *str = NULL;

	if (uf == NULL) {
	    Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Trigger), trigger);
	    return false;
	} else {
	    /* GuiSetSleep(TRUE); */
	    SetTableProgramRecords(pr);
	    ProgramBegin(prg, TRUE);
	    elem = ProgramCall(prg, uf, NULL, TRUE);
	    if (elem) {
		switch (elem->el_Type) {
		    case TYPE_STRING:
			str = elem->el_Contents.string.string;
			break;
		    case TYPE_MEMO:
			str = elem->el_Contents.mi;
			break;
		    case TYPE_NIL:
			break;
		    default:
			elemError = GCS(MSG_Err_Listview_Compute);
			break;
		}
		if (str) {
		    labels = StringToLabels(memPool, str);
		    if (labels == NULL) {
			elemError = GCS(MSG_Err_Mem);
		    }
		    labelsAllocated = TRUE;
		}
	    }
            ProgramEnd();
	    error = ProgramError(&line, &col, &filename);
	    /* GuiSetSleep(FALSE); */

	    if (error != NULL && error != EXECUTE_ABORT) {
		ProgramEditorShowError(pr, TRUE, error, line, col, filename);
		if (labelsAllocated) {
		    FreeLabels(memPool, labels);
		}
		return false;
	    }
	    if (elemError) {
		Ask(AT_ERROR, GCS(MSG_StarOk), elemError);
		return false;
	    }
	}
    } else {
	labels = DisplayFieldListviewLabels(dio_);
    }

    GuiSetSleep(TRUE);
    list_->clear();
    if (labels) {
        for (char **s = labels; *s; s++) {
            list_->addItem(*s);
        }

        Field *field = DisplayFieldField(dio_);
        short type = FieldType(field);
        if (type == TYPE_STRING) {
            Table *tab = FieldTable(field);
            Record *rec = TableGuiRecord(tab);
            DataInfo *di = 0, di_buf;

            RecordLock(rec);
            RecordDataInfo(rec, field, &di, &di_buf);
            if (di) {
                const char *s = di->di_Contents.string;
                if (s) {
                    for (long i = 0; labels[i] != NULL; i++) {
                        if (strcmp(s, labels[i]) == 0) {
                            list_->setCurrentRow(i);
                            break;
                        }
                    }
                }
            }
            RecordUnlock(rec);
        }

        list_->setFocus(Qt::OtherFocusReason);
    }
    GuiSetSleep(FALSE);

    if (labelsAllocated) {
	FreeLabels(memPool, labels);
    }

    return true;
}

void PopListview::stop()
{
    int row = list_->currentRow();
    QListWidgetItem *item = list_->item(row);
    if (item) {
        // We postpone storing data in order to close the popup first.
        // Otherwise, a dialog in a user program will not receive input.
        text_ = item->text();
        QTimer::singleShot(0, this, SLOT(saveText()));
    }
    list_->clear();
}

void PopListview::saveText()
{
    GuiStoreDataStr(dio_, text_.toUtf8().data());
}

static QHBoxLayout *buildGuiPopListview(DisplayObject *dio, QWidget *string,
	QWidget **button, QWidget *parent)
{
    PopButton *popButton = new PopButton(parent, string);
    popButton->setArrowType(Qt::DownArrow);
    if (!DisplayFieldTabChain(dio)) {
        popButton->setFocusPolicy(Qt::NoFocus);
    }
    SetGuiObjectDisplayObject(popButton, dio);

    *button = popButton;
    
    PopupList *list = new PopupList(parent);
    PopListview *pop = new PopListview(dio, string, popButton, list);
    list->setPopWidget(pop);
    
    return pop;
}

class ViewButton : public ExtraButton
{
    Q_OBJECT
    
public:
    ViewButton(QWidget *parent, QWidget *string) : ExtraButton(parent, string)
    {
        setIcon(style()->standardIcon(QStyle::SP_FileDialogContentsView));    
        connect(this, SIGNAL(clicked()), this, SLOT(onClicked()));
    }
    
private slots:
    void onClicked()
    {
        GuiStoreEdits();
        GuiObjectViewFile(buddy());
    }
};

static QWidget *buildGuiViewButton(QWidget *string, QWidget *parent)
{
    QWidget *button = new ViewButton(parent, string);
    return button;
}

class ReferenceButton : public ExtraButton
{
    Q_OBJECT
    
public:
    ReferenceButton(QWidget *parent, QWidget *buddy, 
                    void (*action)(Field *field, bool on)) : 
        ExtraButton(parent, buddy),
        action_(action)
    {
        setCheckable(true);
        connect(this, SIGNAL(clicked()), this, SLOT(onClicked()));
    }
    
private slots:
    void onClicked()
    {
        DisplayObject *dio = GuiObjectDisplayObject(this);
        GuiStoreEdits();
        action_(DisplayFieldField(dio), isChecked());        
    }
    
private:
    void (*action_)(Field *field, bool on);
};

static QWidget *buildGuiFieldAutoShowButton(DisplayObject *dio, 
                                                QWidget *string) 
{
    ReferenceButton *button = new ReferenceButton(string->parentWidget(), string,
                                                  ReferenceAutoShowOnOff);
    button->setText(GCS(MSG_ReferenceAutoShow));
    button->setToolTip(GCS(MSG_ReferenceAutoShowBH));
    SetGuiObjectDisplayObject(button, dio);
    return button;
}

static QWidget *buildGuiFieldFilterButton(DisplayObject *dio, 
                                              QWidget *string) 
{
    ReferenceButton *button = new ReferenceButton(string->parentWidget(), 
                                                  string, ReferenceFilterOnOff);
    button->setText(GCS(MSG_ReferenceFilter));
    button->setToolTip(GCS(MSG_ReferenceFilterBH));
    SetGuiObjectDisplayObject(button, dio);
    return button;
}

bool BuildGuiField(DisplayObject *dio, bool hideLabel, QWidget *parent,
        GroupLayout *group, GroupLayout **colGroup)
{
    Field *field = DisplayFieldField(dio);
    short titlepos = DisplayFieldTitlePosition(dio);
    int weight = DisplayObjectWeight(dio);
    short type = FieldType(field);
    GuiFieldFeatures f;
    
    if (group == 0 || colGroup == 0 || !BuildGuiFieldGetFeatures(dio, &f)) {
	return false;			// be safe
    }

    bool horiz = (group->columns() != 1);
    bool postAddTitle = false;
    QWidget *go = 0, *guiobject = 0, *root = 0, *activateobject = 0;
    QWidget *externalobject = 0;
    QWidget *fileobject = 0, *fontobject = 0, *lvobject = 0, *viewobject = 0; 
    QWidget *showobject = 0, *autoshowobject = 0, *filterobject = 0;
    QLayout *layout = 0;
    
    if (hideLabel) {
	f.label = false;
    }

    if (f.label && !horiz && (titlepos == DAD_TITLEPOSITION_LEFT || 
		titlepos == DAD_TITLEPOSITION_RIGHT)) {
	if (!*colGroup) {
            *colGroup = new GroupLayout(2);
            (*colGroup)->setContentsMargins(0, 0, 0, 0);
            group->addLayout(*colGroup);
            group->setStretch(100, 100);
	}
        group = *colGroup;
	if (titlepos == DAD_TITLEPOSITION_RIGHT) {
	    postAddTitle = true;
	} else if (!BuildGuiFieldLabel(dio, parent, group)) {
    	    return false;
	}
    } else {
	BuildGuiFinishColGroup(colGroup);
	if (f.label) {
	    switch (titlepos) {
		case DAD_TITLEPOSITION_LEFT:
		case DAD_TITLEPOSITION_RIGHT:
	    	    if (titlepos == DAD_TITLEPOSITION_RIGHT) {
     			postAddTitle = true;
		    } else if (!BuildGuiFieldLabel(dio, parent, group)) {
			return false;
		    }
		    break;

		case DAD_TITLEPOSITION_TOP:
		case DAD_TITLEPOSITION_BOTTOM:
                {
                    GroupLayout *vgroup = new GroupLayout(1);
                    vgroup->setVerticalSpacing(2);
                    vgroup->setContentsMargins(0, 0, 0, 0);
                    group->addLayout(vgroup);
                    group->setStretch(weight, weight);
                    group = vgroup;
                    horiz = false;
	    	    if (titlepos == DAD_TITLEPOSITION_BOTTOM) {
     			postAddTitle = true;
		    } else if (!BuildGuiFieldLabel(dio, parent, group)) {
			return false;
		    }
		    break;
                }
	    }
	}
    }
    if (f.maxchars >= 0 || f.text) {
	if (f.show || f.readonly) {
	    go = buildGuiFieldText(dio, parent);
            guiobject = go;
	} else {
	    go = buildGuiFieldString(dio, f.maxchars, parent);
	    activateobject = guiobject = go;
	}
    } else if (f.checkbox) {
	go = buildGuiFieldCheckBox(dio, parent);
	activateobject = guiobject = go;
    } else if (f.radiobuttons) {
      	go = buildGuiFieldRadioBox(dio, parent);
	activateobject = guiobject = go;
    } else if (f.choice) {
        go = buildGuiFieldChoice(dio, parent);
	activateobject = guiobject = go;
    } else if (f.button) {
	go = buildGuiFieldButton(dio, parent);
	activateobject = guiobject = go;
    } else if (f.image) {
	go = buildGuiFieldImageButton(dio, &guiobject, parent);
	activateobject = guiobject;
    } else if (f.editor) {
    	go = buildGuiFieldMemo(dio, parent);
    	activateobject = guiobject = go;
    } else if (f.listview) {
    	go = buildGuiFieldSFWView(dio, parent);
    	activateobject = guiobject = go;
    } else if (f.maxchars != -1) {
        return false;			// unknown gui object
    }

    if (f.maxchars != -1 && !go) {
	return false;
    }
    if (go) {
        go->setToolTip(DisplayFieldBubbleHelp(dio));
    }

    if (f.refpopup) {
	layout = BuildGuiPopRefFieldBrowser(dio, go, &lvobject, parent);
	if (!activateobject) {
	    activateobject = lvobject;
	}
        // TODO: SetDisplayFieldPopObject(dio, (GuiObject)hbox);
    }
    
    if (type== TYPE_REFERENCE && go && 
            (f.show || DisplayFieldAutoShow(dio) || 
             DisplayFieldFilter(dio))) {
        if (!layout) {
            QBoxLayout *box = new QHBoxLayout();
            box->setContentsMargins(0, 0, 0, 0);
            box->setSpacing(POP_STRING_SPACING);
            box->addWidget(go, 1);
            layout = box;
        }
        if (f.show) {
            showobject = buildGuiFieldShowButton(dio, guiobject);
            layout->addWidget(showobject);
        }            
        if (DisplayFieldAutoShow(dio)) {
            autoshowobject = buildGuiFieldAutoShowButton(dio, guiobject);
            layout->addWidget(autoshowobject);
        }
        if (DisplayFieldFilter(dio)) {
            filterobject = buildGuiFieldFilterButton(dio, guiobject);
            layout->addWidget(filterobject);
        }
    }

    if (type == TYPE_DATE && DisplayFieldCalendar(dio) && !f.readonly && go) {
        layout = buildGuiPopCalendar(dio, go, &lvobject, parent);
	if (!activateobject) {
	    activateobject = lvobject;
	}
    }
    if (type == TYPE_STRING && DisplayFieldFile(dio) && go) {
	layout = buildGuiPopFile(dio, go, &fileobject, parent);
	if (!activateobject) {
	    activateobject = fileobject;
	}
    } else if (type == TYPE_STRING && DisplayFieldFont(dio) && go) {
        layout = buildGuiPopFont(dio, go, &fontobject, parent);
	if (!activateobject) {
	    activateobject = fontobject;
	}
    } else if (type == TYPE_STRING && DisplayFieldListview(dio) && go) {
	layout = buildGuiPopListview(dio, go, &lvobject, parent);
	if (!activateobject) {
	    activateobject = lvobject;
	}
    }

    if (type == TYPE_STRING && DisplayFieldView(dio) && go) {
        viewobject = buildGuiViewButton(guiobject, parent);
        if (!layout) {
            QBoxLayout *box = new QHBoxLayout();
            box->setContentsMargins(0, 0, 0, 0);
            box->setSpacing(POP_STRING_SPACING);
            box->addWidget(go, 1);
            layout = box;
        }
        layout->addWidget(viewobject);
    }
    if (f.datatypes) {
	GuiObject str_go = go;
        char size = DisplayFieldImageSize(dio);
	const char *title = DisplayObjectTitle(dio);

        if (str_go && DisplayFieldTitleAtStringField(dio) && 
                title && title[0]) {
            QString text = GuiMarkShortcut(title, DisplayObjectShortcut(dio));
            GroupLayout *sub = new GroupLayout(2);
            sub->addLabelAndWidget(text, str_go, layout);
            layout = sub;
	}
        
	if (size == DAD_IMAGESIZE_SCROLLABLE) {
            QScrollArea *scroll = new QScrollArea(parent);
            scroll->setAlignment(Qt::AlignCenter);
            ImageLabel *image = new ImageLabel(scroll);
            image->setFixed(true);
            image->setScaled(false);
            image->setKeepAspect(false);
            scroll->setWidget(image);
            externalobject = image;
            go = scroll;
        } else {
            bool scaled = (size == DAD_IMAGESIZE_SCALED || 
                           size == DAD_IMAGESIZE_ASPECTSCALED);
            ImageLabel *image = new ImageLabel(parent);
            image->setFrameShadow(QFrame::Raised);
            image->setFrameShape(QFrame::StyledPanel);
            image->setFixed(size == DAD_IMAGESIZE_FIXED);
            image->setScaled(scaled);
            image->setKeepAspect(size == DAD_IMAGESIZE_ASPECTSCALED);
            externalobject = image;
            go = image;
	}
        externalobject->setToolTip(DisplayFieldBubbleHelp(dio));        
        
        if (layout || str_go) {
            GroupLayout *sub = new GroupLayout(1);
            sub->addWidget(go);
            if (size == DAD_IMAGESIZE_FIXED) {
                sub->setAlignment(go, Qt::AlignHCenter);
            } else {
                sub->setStretch(1, 1);
            }
            if (layout) {
                sub->addLayout(layout);
            } else {
                sub->addWidget(str_go);
            }
            layout = sub;
        }
    }
    
    if (!go) {
	return false;
    }

    if (!DisplayFieldTabChain(dio) && guiobject) {
        Qt::FocusPolicy policy = guiobject->focusPolicy();
        policy = static_cast<Qt::FocusPolicy>(policy & ~(Qt::TabFocus));
        guiobject->setFocusPolicy(policy);
    }

    if (layout) {
        group->addLayout(layout);
    } else {
        group->addWidget(go);
    }
    group->setStretch(weight, weight);
    if (postAddTitle && !BuildGuiFieldLabel(dio, parent, group)) {
	return false;
    }

    if (labelObject) {
        labelObject->setBuddy(activateobject);
	labelObject = 0;
    }
    SetGuiObjectDisplayObject(guiobject, dio);
    BuildGuiSetBackground(dio, guiobject);
    BuildGuiSetFont(dio, guiobject);
    GuiSetHelpNode(guiobject, "Changing records");

    SetDisplayObjectGuiObject(dio, guiobject);
    SetDisplayObjectRootObject(dio, root? root : go);
    SetDisplayObjectActivateObject(dio, activateobject);
    SetDisplayFieldFileObject(dio, fileobject);
    SetDisplayFieldFontObject(dio, fontobject);
    SetDisplayFieldListviewObject(dio, lvobject);
    SetDisplayFieldViewObject(dio, viewobject);
    SetDisplayFieldExternalObject(dio, externalobject);
    SetDisplayFieldShowObject(dio, showobject);
    SetDisplayFieldFilterObject(dio, filterobject);
    SetDisplayFieldAutoShowObject(dio, autoshowobject);
    SetDisplayFieldDisableObject(dio, go);

    return true;
}

void GuiFieldStoreEdits()
{
    QWidget *w = changedGuiObject;
    if (w) {
        changedGuiObject = 0;   // avoid recursion, clear it immediately
	DisplayObject *dio = GuiObjectDisplayObject(w);
	if (dio) {
            QLineEdit *edit = dynamic_cast<QLineEdit *>(w);
	    if (edit) {
                QByteArray s = edit->text().toUtf8();
		GuiStoreDataStr(dio, s.data());
            }
            TextEditor *te = dynamic_cast<TextEditor *>(w);
            if (te && te->document()->isModified()) {
                QByteArray s = te->toPlainTextNbsp().toUtf8();
		GuiStoreDataStr(dio, s.data());
	    }
	}
    }
}

void GuiSetString(GuiObject go, const char *str, bool readonly, 
        GuiObject external)
{
    QWidget *save = changedGuiObject;

    QLabel *label = dynamic_cast<QLabel *>(go);
    if (label) {
        label->setText(str);
    }
    QLineEdit *edit = dynamic_cast<QLineEdit *>(go);
    if (edit) {
        edit->setText(str);
    }
    QAbstractButton *button = dynamic_cast<QAbstractButton *>(go);
    if (button) {
        button->setText(str);
    }

    ImageLabel *image = dynamic_cast<ImageLabel *>(external);
    if (image) {
        QWidget *topLevel = image->topLevelWidget();
        ProjectWindow *win = dynamic_cast<ProjectWindow *>(topLevel);
        Project *pr = GuiWindowProject(win);
        ProjectSetCurrentDir(pr);           // ensure we have correct CWD
        image->setPixmap(QPixmap(qtExpandPathname(str)));
    }

    // Restore changeGuiObject such that we ignore changes here.
    changedGuiObject = save;
}

void GuiSetBoolean(GuiObject go, bool value, bool readonly)
{
    CheckBox *check = dynamic_cast<CheckBox *>(go);
    if (check) {
        check->setChecked(value);
        if (readonly) {
            check->setEnabled(false);
        }
    }
}

void GuiSetChoice(GuiObject go, unsigned short value, char **labels,
        bool readonly) 
{
    if (readonly) {
        QLabel *label = dynamic_cast<QLabel *>(go);
        if (label) {
            const char *text;
            if (value < NumLabels(labels)) {
                text = labels[value];
            } else {
                text = IntToStr(value);
            }
            label->setText(text);
        }
    } else {
        ChoiceGadget *choice = dynamic_cast<ChoiceGadget *>(go);
        if (choice) {
            choice->setCurrentIndex(value);
        }
        RadioGadget *radio = dynamic_cast<RadioGadget *>(go);
        if (radio) {
            radio->setCheckedId(value);
        }
    }
}

void GuiSetMemo(GuiObject go, const char *str, bool readOnly)
{
    QWidget *save = changedGuiObject;
    MemoGadget *edit = dynamic_cast<MemoGadget *>(go);

    edit->setPlainText(str);
    edit->document()->setModified(false);
    edit->setReadOnly(readOnly);

    // Restore changeGuiObject such that we ignore changes here.
    changedGuiObject = save;
}

void GuiSetSFWList(GuiObject go, const struct Element *elem)
{
    SFWView *sfw = dynamic_cast<SFWView *>(go);

    if (sfw) {
        const char *error = sfw->setElement(elem);
        if (error) {
            sfw->setString(error);
        } else if (DisplayFieldControlTable(GuiObjectDisplayObject(go))) {
            Table *tab = sfw->controlTable();
            Record *rec = TableGuiRecord(tab);

            if (tab != NULL && rec != NULL) {
                sfw->setActiveRecord(rec);
            }
        }
    }
}

// Since popups are modal, there can't be any flow where we add or remove
// a record while the popup is open.

void BuildGuiAddRecord(Table *addtab, Record *rec)
{
}

void BuildGuiRemRecord(Table *remtab, Record *rec)
{
    Project *pr = TableProject(remtab);
    List *tabs = ProjectTables(pr);

    if (!tabs || !rec) {
	return;
    }

    Table *tab;
    FORNODE(tab, tabs) {
	List *fields = TableFields(tab);
	if (!fields) {
	    continue;
	}

	Field *field;
	FORNODE(field, fields) {
	    short type = FieldType(field);
            DisplayObject *dio = FieldDisplayObject(field);
	    if (type == TYPE_VIRTUAL && 
    		    DisplayFieldSubType(dio) == DAD_VIRTUALTYPE_LIST) {
		GuiObject go = DisplayObjectGuiObject(dio);
                SFWView *sfw = dynamic_cast<SFWView *>(go);

		if (sfw) {
		    sfw->remRecord(rec);
		}
	    }
        }
    }
}

void BuildGuiRemAllRecords(Project *pr)
{
    List *tabs = ProjectTables(pr);
    
    if (!tabs) {
	return;
    }
    
    Table *tab;
    FORNODE(tab, tabs) {
	List *fields = TableFields(tab);
        
	if (!fields) {
	    continue;
	}
        
	Field *field;
	FORNODE(field, fields) {
	    short type = FieldType(field);
    	    DisplayObject *dio = FieldDisplayObject(field);
            
	    if (type == TYPE_VIRTUAL && 
                       DisplayFieldSubType(dio) == DAD_VIRTUALTYPE_LIST) {
		GuiObject go = DisplayObjectGuiObject(dio);
                SFWView *sfw = dynamic_cast<SFWView *>(go);
                
		if (sfw) {
		    sfw->remAllRecords();
		}
	    }
        }
    }
}

#include "qtBuildGuiField.moc"
