//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtBuildGuiRefFieldBrowser.cc
//
// Creation date: 12.08.2010
// Last modified: 03.12.2020
//

#include "qt/qtBuildGuiRefFieldBrowser.h"
#include "qt/qtBuildGui.h"
#include "qt/qtUtil.h"
#include <qt/classes/qtPopWidget.h>
#include <qt/classes/qtQuickSearch.h>
#include <QPointer>
#include <QTreeView>
#include <QHeaderView>
#include "buildgui.h"
#include "programeditor.h"
#include "order.h"
#include "field.h"
#include "record.h"
#include "display.h"
#include "gui.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

#define CURRENT_RECORD (struct Record *)(~0)

Q_DECLARE_OPAQUE_POINTER(Record *);
Q_DECLARE_METATYPE(Record *);

class RefFieldModel : public EntryListModel
{
public:
    RefFieldModel(QWidget *view, DisplayObject *dio) 
        : EntryListModel(view), view_(view), dio_(dio) 
    {
        const RecordDescriptionSpec *rds = DisplayFieldRecordDescriptionSpec(dio, true);
        multiColumn_ = rds? rds->rds_MultiColumn : false;
        long numColumns = -1;
        if (multiColumn_) {
            Field *field = DisplayFieldField(dio);
            struct RecordDescription *rd = AllocRecordDescription(                                                      FieldReferenceTable(field), rds);
            if (rd) {
                RecordDescriptionSetRecord(rd, 0);
                numColumns = RecordDescriptionNumColumns(rd);
                FreeRecordDescription(rd);
            }
            if (numColumns <= 1) {
                multiColumn_ = false;
            }
        }
        columns_ = multiColumn_? numColumns : 1;
    }

    virtual int	rowCount(const QModelIndex &parent = QModelIndex()) const
    {
        return recs_.count();
    }
    
    virtual int	columnCount(const QModelIndex &parent = QModelIndex()) const
    {
        return columns_;
    }
    
    virtual int findEntry(void *entry) const
    {
        for (int i = 0; i < recs_.count(); ++i) {
            if ((void *)recs_[i] == entry) {
                return i;
            }
        }
        return -1;
    }
    
    virtual void *entry(int row) const
    {
        if (row >= 0 && row < recs_.count()) {
            return (void *)recs_[row];
        }
        return 0;
    }

    virtual QVariant data(const QModelIndex &index, 
            int role = Qt::DisplayRole) const
    {
        if (index.isValid() && index.row() < recs_.size()) {
            Record *rec = recs_.at(index.row());

            switch (role) {
                case Qt::DisplayRole:
                    return str(rec, index.column());
                case Qt::FontRole:
                    if (rec == 0 || rec == CURRENT_RECORD) {
                        QFont font = view_->font();
                        font.setItalic(true);
                        return font;
                    }
                    break;
                case Qt::TextAlignmentRole:
                    if (multiColumn_ && rec != 0 && rec != CURRENT_RECORD) {
                        short type = TYPE_UNKNOWN;
                        Field *field = DisplayFieldField(dio_);
                        Table *reftab = FieldReferenceTable(field);
                        const RecordDescriptionSpec *rds = DisplayFieldRecordDescriptionSpec(dio_, TRUE);
                        RecordDescription *rd = AllocRecordDescription(reftab, rds);
                        RecordDescriptionSetRecord(rd, rec);
                        RecordLock(rec);
                        type = RecordDescriptionColumnType(rd, index.column());
                        RecordUnlock(rec);
                        FreeRecordDescription(rd);
                        Qt::AlignmentFlag align = Qt::AlignRight;
                        if (type != TYPE_REAL && type != TYPE_INTEGER) {
                            align = Qt::AlignLeft;
                        }
                        return align;
                    }
                    break;
            }
        }
        return QVariant();
    }

    void clear() { recs_.clear(); }

    void addRecord(Record *rec)
    {
        recs_.append(rec);
    }

    void addRecords(const Element *elem, const Table *tab)
    {
        if (elem) {
            while (elem->el_Type == TYPE_LIST) {
                addRecords(elem->el_Contents.list.head, tab);
                elem = elem->el_Contents.list.tail;
            }
            if (elem->el_Type == TYPE_RECORD && 
                    elem->el_Contents.tabrec.tab == tab) {
                addRecord(elem->el_Contents.tabrec.rec);
            }
	}
    }

    void addRecords(List *recs, bool matchFilter)
    {
        if (recs) {
            Record *rec;

            FORNODE (rec, recs) {
                if (!matchFilter || RecordMatchFilter(rec)) {
                    addRecord(rec);
                }
            }
        }
    }
    
    QModelIndex find(Record *rec)
    {
        int i = recs_.indexOf(rec);
        if (i >= 0) {
            return createIndex(i, 0);
        }
        return QModelIndex();
    }
    
    Record *record(const QModelIndex &index) const
    {
        Record *rec = 0;
        if (index.isValid() && index.row() < recs_.size()) {
            rec = recs_.at(index.row());
        }
        return rec;
    }

    void insert(Record *rec, Record *next)
    {
        int i = recs_.indexOf(rec);
        if (i < 0) {
            i = recs_.size();
        }
        beginInsertRows(QModelIndex(), i, i);
        recs_.insert(i, rec);
        endInsertRows();
    }

    void remove(Record *rec)
    {
        int i = recs_.indexOf(rec);
        if (i >= 0) {
            beginRemoveRows(QModelIndex(), i, i);
            recs_.removeAt(i);
            endRemoveRows();
        }
    }
    
private:
    QWidget *view_;
    DisplayObject *dio_;
    QList<Record *> recs_;
    bool multiColumn_;
    int columns_;

    QString str(Record *rec, int col) const;
};

QString RefFieldModel::str(Record *rec, int col) const
{
    Field *field = DisplayFieldField(dio_);
    Table *reftab = FieldReferenceTable(field);
    const RecordDescriptionSpec *rds = DisplayFieldRecordDescriptionSpec(dio_, TRUE);
    RecordDescription *rd = AllocRecordDescription(reftab, rds);
    if (!rd) {
        return "";
    }
     
    RecordDescriptionSetRecord(rd, (rec == CURRENT_RECORD)? NULL : rec);

    char *cont = NULL;
    QString result;

    if (rec == NULL || rec == CURRENT_RECORD) {
        cont = GCS(rec? MSG_CurrentRecord : MSG_InitialRecord);
        if (multiColumn_) {
            int i, num = RecordDescriptionNumColumns(rd);
            for (i=0; i < num; i++) {
                if (RecordDescriptionColumnType(rd, i) == TYPE_STRING) {
                    break;
                }
            }
            if (i == num) {
                i = 0;
            }
            if (i != col) {
                cont = NULL;
            }
        }
        result = cont;
    } else {
        Project *pr = DisplayObjectProject(dio_);
        MemPool *memPool = ProjectMemPool(pr);
        bool free = FALSE;
        long error = OK;

        RecordLock(rec);
        if (multiColumn_) {
            cont = RecordDescriptionColumnText(rd, col, memPool, &free);
        } else {
            cont = RecordDescriptionText(rd, memPool, &free);
        }
        error = RecordDescriptionError(rd);
        result = error? GES(error) : EmptyNullStr(cont);
        RecordUnlock(rec);
        if (free) {
            FreeString(memPool, cont);
        }
    }
    return result;
}

class RefFieldBrowser : public QTreeView
{
    Q_OBJECT
    
public:
    RefFieldBrowser(QWidget *parent, DisplayObject *dio) 
        : QTreeView(parent), ram_(this, dio), quickSearch_(0)
    {
        setModel(&ram_);
        setAllColumnsShowFocus(true);
        setHeaderHidden(true);
        setRootIsDecorated(false);
        setUniformRowHeights(true);
        SetGuiObjectDisplayObject(this, dio);
        setStyleSheet(
	    "QTreeView::item:selected{background-color: palette(highlight);"
            "color: palette(highlightedText);};");
    }

    void clear() { ram_.clear(); }

    void addRecord(struct Record *rec)
    {
        ram_.addRecord(rec);
    }

    void addRecords(const Element *elem, const Table *tab) 
    { 
        ram_.addRecords(elem, tab);
    }

    void addRecords(List *recs, bool matchFilter)
    {
        ram_.addRecords(recs, matchFilter);
    }
    
    void setActive(struct Record *rec)
    {
        QModelIndex ind = ram_.find(rec);
        setCurrentIndex(ind);
        scrollTo(ind);
    }
    
    Record *active() const
    {
        return ram_.record(currentIndex());
    }

    void insert(Record *rec, Record *next)
    {
       ram_.insert(rec, next);
    }

    void remove(Record *rec)
    {
        ram_.remove(rec);
    }
    
public slots:
    void onSearch(const QString &text, QuickSearch::Action action) {
        QByteArray data = text.toUtf8();
        const char *str = data.data();
        DisplayObject *dio = GuiObjectDisplayObject(this);
        long dir = -1;
            
        switch (action) {
            case QuickSearch::NEXT:  dir++;		// no break
            case QuickSearch::START: dir++;		// no break
            case QuickSearch::PREV:
                refFieldSearchCurrent_ = GuiReferenceFieldQuickSearchRecord(                                                                                  dio, this, dir, str, refFieldSearchCurrent_);
                break;
                    
            case QuickSearch::COMMIT:
                emit activated(currentIndex());         // no break
            case QuickSearch::CANCEL:
                quickSearch_->hide();
                quickSearch_->deleteLater();
                break;
        }
    }
    
protected:
    void keyPressEvent(QKeyEvent *event) {
        if (!QuickSearch::isStartEvent(event)) {
            QTreeView::keyPressEvent(event);
        } else {
            delete quickSearch_;
            refFieldSearchCurrent_ = 0;
            quickSearch_ = new QuickSearch(this);
            connect(quickSearch_, 
                    SIGNAL(search(const QString &, QuickSearch::Action)), 
                    this, SLOT(onSearch(const QString &, QuickSearch::Action)));
            quickSearch_->show();
            quickSearch_->processEvent(event);
        }        
    }
    
private:
    RefFieldModel ram_;
    QPointer<QuickSearch> quickSearch_;
    Record *refFieldSearchCurrent_;
};

class PopRefFieldBrowser : public PopWidget
{
    Q_OBJECT
    
public:
    PopRefFieldBrowser(DisplayObject *dio, QWidget *string, 
                      QAbstractButton *button, RefFieldBrowser *rab)
    : dio_(dio), rab_(rab)
    {
        setString(string);
        setButton(button);
        setPopupWidget(rab);
        connect(rab, SIGNAL(activated(const QModelIndex &)), 
                this, SLOT(onRefFieldBrowserActivated()));
    }

    virtual bool start();
    virtual void stop();

    virtual void close(bool success)
    {
        PopWidget::close(success);
        rab_->clear();
    }

    void addRecord(Record *rec)
    {
        if (rec) {
            char records = DisplayFieldRecordsKind(dio_);
            bool filter = (records == DAD_RECORDS_MATCHFILTER);
            if (rab_ && (records == DAD_RECORDS_ALL || 
                        (filter && RecordMatchFilter(rec)))) {
                rab_->insert(rec, NextRecord(rec, filter));
            }
        }
    }
    
    void remRecord(Record *rec)
    {
        if (rec) {
            rab_->remove(rec);
        }
    }

public slots:
    void onRefFieldBrowserActivated() { closePopup(true); }

    void storeData(Record *active) { GuiStoreDataRecord(dio_, active); }
    
private:
    DisplayObject *dio_;
    RefFieldBrowser *rab_;
};

bool PopRefFieldBrowser::start()
{
    Field *field = DisplayFieldField(dio_);
    Table *tab = FieldTable(field);
    Table *reftab = FieldReferenceTable(field);
    List *refrecs = TableRecords(reftab);
    char records = DisplayFieldRecordsKind(dio_);

    if (refrecs == NULL) {
        return false;
    }

    GuiStoreEdits();      // maybe the user has just edited a string object

    rab_->clear();

    if (DisplayFieldInitialRecord(dio_)) {
        rab_->addRecord(0);
    }
    if (DisplayFieldCurrentRecord(dio_)) {
        rab_->addRecord(CURRENT_RECORD);
    }

    if (records == DAD_RECORDS_COMPUTED) {
        const char *trigger = DisplayFieldRecordsCompute(dio_);

        if (trigger != NULL && trigger[0] != 0) {
            Project *pr = TableProject(tab);
            Program *prg = ProjectProgram(pr);
            UserFunction *uf = FindUserFunctionName(prg, trigger);
            const char *filename = NULL, *error = NULL;
            long line, col;
            struct Element *elem;

            if (uf == NULL) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_Trigger), trigger);
                return false;
            } else {
                // GuiSetSleep(TRUE);
                SetTableProgramRecords(pr);
                ProgramBegin(prg, TRUE);
                elem = ProgramCall(prg, uf, NULL, TRUE);
                rab_->addRecords(elem, reftab);
                ProgramEnd();
                error = ProgramError(&line, &col, &filename);
                // GuiSetSleep(FALSE);

                if (error != NULL && error != EXECUTE_ABORT) {
                    ProgramEditorShowError(pr, TRUE, error, line, col,filename);
                    return false;
                }
            }
        }
    } else {
        GuiSetSleep(true);
        rab_->addRecords(refrecs, records == DAD_RECORDS_MATCHFILTER);
        GuiSetSleep(false);
    }

    Record *rec = TableGuiRecord(tab);
    RecordLock(rec);
    struct DataInfo *di = NULL, di_buf;    
    RecordDataInfo(rec, field, &di, &di_buf);
    if (di) {
        rab_->setActive(di->di_Contents.rec);
    }
    rab_->header()->resizeSections(QHeaderView::ResizeToContents);
    rab_->setFocus(Qt::OtherFocusReason);
    RecordUnlock(rec);

    return true;
}

void PopRefFieldBrowser::stop()
{
    Record *active = rab_->active();

    if (active == CURRENT_RECORD) {
	Field *field = DisplayFieldField(dio_);
	Table *reftab = FieldReferenceTable(field);

	active = TableGuiRecord(reftab);
    }

    // Postpone setting record reference to next event loop iteration.
    // This ensures the modal popup is closed before any potential error dialog.
    static bool registered = false;
    if (!registered) {
        qRegisterMetaType<Record *>();
        registered = true;
    }
    QMetaObject::invokeMethod(this, "storeData", Qt::QueuedConnection,
                              Q_ARG(Record *, active));
}

QHBoxLayout *BuildGuiPopRefFieldBrowser(struct DisplayObject *dio, 
        GuiObject string, GuiObject *lvobject, GuiObject parent)
{
    PopButton *button = new PopButton(parent, string);
    button->setArrowType(Qt::DownArrow);
    if (!DisplayFieldTabChain(dio)) {
        button->setFocusPolicy(Qt::NoFocus);
    }
    SetGuiObjectDisplayObject(button, dio);
    *lvobject = button;

    RefFieldBrowser *rab = new RefFieldBrowser(parent, dio);

    PopRefFieldBrowser *pop = new PopRefFieldBrowser(dio, string, button, rab);

#if 0
    g_signal_connect(string, "key-press-event",G_CALLBACK(onKeyPressEvent),dio);
    g_signal_connect(*button,"key-press-event",G_CALLBACK(onKeyPressEvent),dio);
#endif

    return pop;
}

void BuildGuiPopRefFieldBrowserAddRecord(GuiObject go, struct Record *rec)
{
    PopRefFieldBrowser *pop = dynamic_cast<PopRefFieldBrowser *>(go);
    if (pop && pop->isOpen()) {
        pop->addRecord(rec);
    }
}

void BuildGuiPopRefFieldBrowserRemRecord(GuiObject go, struct Record *rec)
{
    PopRefFieldBrowser *pop = dynamic_cast<PopRefFieldBrowser *>(go);
    if (pop && pop->isOpen()) {
        pop->remRecord(rec);
    }
}

#include "qtBuildGuiRefFieldBrowser.moc"

//
// todo: change selection when user moves to new record (also in MUI version)
// update display when user edits a referenced record (also in MUI version)
//
