//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtBuildGuiTable.cc
//
// Creation date: 14.06.2010
// Last modified: 15.08.2020
//

#include "qt/qtBuildGuiTable.h"
#include "qt/qtBuildGui.h"
#include "qt/qtMenus.h"
#include "qt/qtUtil.h"
#include <QApplication>
#include <QFocusFrame>
#include <QLabel>
#include <QMenu>
#include <QMouseEvent>
#include <QPushButton>
#include <QStyleOptionFocusRect>
#include <QStylePainter>
#include <QToolBar>
#include <QToolButton>
#include "buildgui.h"
#include "gui.h"
#include "table.h"
#include "display.h"
#include "filter.h"

// #define USE_TOOLBAR
#define USE_BUTTONSTYLE
#undef DRAW_FOCUS_BOX

class ArrowButton : public QToolButton
{
    Q_OBJECT

public:
    ArrowButton(bool isRight, QWidget *parent) : QToolButton(parent)
    {
        isRight_ = isRight;
        setText(isRight? ">" : "<");
        // setIcon(isRight_? 
        //        style()->standardIcon(QStyle::SP_MediaSkipForward, 0, this) :
        //        style()->standardIcon(QStyle::SP_MediaSkipBackward, 0, this));
        setFocusPolicy(Qt::NoFocus);
        setAutoRepeat(true);
        connect(this, SIGNAL(clicked()), this, SLOT(onClicked()));
    }

private slots:
    void onClicked()
    {
        DisplayObject *dio = GuiObjectDisplayObject(this);
        Project *pr = DisplayObjectProject(dio);
        if (GuiProjectSleep(pr) != 0) {
            setDown(false);
        } else {
            GuiStoreEdits();
            (isRight_? GuiNextRecord : GuiPrevRecord)(DisplayObjectTable(dio));
        }
    }

private:
    bool isRight_;
};

class FilterButton : public QToolButton
{
    Q_OBJECT
    
public:
    FilterButton(QWidget *parent) : QToolButton(parent)
    {
        setText("F");
        setCheckable(true);
        setFocusPolicy(Qt::NoFocus);
        connect(this, SIGNAL(clicked(bool)), this, SLOT(onClicked(bool)));
    }
    
public slots:
    void onClicked(bool checked)
    {
        GuiStoreEdits();
        DisplayObject *dio = GuiObjectDisplayObject(this);
        Table *tab = DisplayObjectTable(dio);
        Qt::KeyboardModifiers mod = QApplication::keyboardModifiers();
        
        if ((mod & Qt::ShiftModifier) || !TableFilterProgram(tab)) {
            setChecked(false);
            GuiChangeFilter(tab);
        } else {
            FilterOnOff(tab, checked);
        }
    }
};

class TableMenuButton : public QToolButton
{
    Q_OBJECT

public:
    TableMenuButton(QWidget *parent) : QToolButton(parent)
    {
        setIcon(QIcon(":/bee64x64.png"));
        //setIcon(style()->standardIcon(QStyle::SP_ArrowDown, 0, this));
        setFocusPolicy(Qt::NoFocus);
        connect(this, SIGNAL(pressed()), this, SLOT(popupMenu()));
    }
    
public slots:
    void popupMenu()
    {
        QPoint pos = mapToGlobal(QPoint(width(), height()));
        pos.setX(pos.x() - TableMenu->sizeHint().width());
        TableMenu->popup(pos);
        setDown(false);
    }
};

#ifdef USE_BUTTONSTYLE
class PanelFocusBox : public QWidget
{
public:
    PanelFocusBox(QWidget *parent);

protected:
#ifdef DRAW_FOCUS_BOX
    void focusInEvent(QFocusEvent *event);
    void focusOutEvent(QFocusEvent *event);
    void paintEvent(QPaintEvent *event);
#endif
    void contextMenuEvent(QContextMenuEvent *event);
};

PanelFocusBox::PanelFocusBox(QWidget *parent) : QWidget(parent)
{
#ifdef DRAW_FOCUS_BOX
    setFocusPolicy(Qt::StrongFocus);
    setAutoFillBackground(true);

    int w = style()->pixelMetric(QStyle::PM_FocusFrameHMargin);
    int h = style()->pixelMetric(QStyle::PM_FocusFrameVMargin);
    setContentsMargins(w, h, w, h);
#else
    setFocusPolicy(Qt::ClickFocus);
#endif
}

#ifdef DRAW_FOCUS_BOX
void PanelFocusBox::focusInEvent(QFocusEvent *event)
{
    update();
}

void PanelFocusBox::focusOutEvent(QFocusEvent *event)
{
    update();
}

void PanelFocusBox::paintEvent(QPaintEvent *event)
{
    if (hasFocus()) {
        QStylePainter painter(this);
        QStyleOptionFocusRect option;
        option.initFrom(this);
        option.backgroundColor = palette().color(QPalette::Background);
        option.state |= QStyle::State_Enabled | QStyle::State_Active | QStyle::State_HasFocus;

        painter.drawPrimitive(QStyle::PE_FrameFocusRect, option);
    }
}
#endif

void PanelFocusBox::contextMenuEvent(QContextMenuEvent *event)
{
    TableMenu->popup(event->globalPos());
    event->accept();
}

class PanelButtonBox : public QWidget
{
public:
    PanelButtonBox(QWidget *parent);

protected:
    void paintEvent(QPaintEvent *event);
};

PanelButtonBox::PanelButtonBox(QWidget *parent) : QWidget(parent)
{
    int w = style()->pixelMetric(QStyle::PM_DefaultFrameWidth);
    setContentsMargins(w, w, w, w);
}

void PanelButtonBox::paintEvent(QPaintEvent *event)
{
    QStylePainter painter(this);
    QStyleOptionToolButton option;
    option.initFrom(this);
    option.features = QStyleOptionToolButton::None;
    option.subControls = QStyle::SC_ToolButton;
    option.state |= QStyle::State_Raised;
    option.state &= ~(QStyle::State_MouseOver);

    painter.drawComplexControl(QStyle::CC_ToolButton, option);
}

#endif

static QWidget *buildGuiPanel(DisplayObject *dio, QWidget *parent,
        QWidget **sub)
{
    const char *title = DisplayObjectTitle(dio);
    Table *tab = DisplayObjectTable(dio);
    bool gadgets = !TableExactlyOneRecord(tab);

#ifdef USE_TOOLBAR
    QWidget *main = new QWidget(parent);
#else
    QWidget *main = new PanelFocusBox(parent);
#endif
    GroupLayout *vgroup = new GroupLayout(1, main);
    vgroup->setContentsMargins(0, 0, 0, 0);
    vgroup->setSpacing(0);

#ifdef USE_TOOLBAR
    QToolBar *toolbar = new QToolBar(main);
    toolbar->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
    toolbar->setFloatable(false);
    toolbar->setMovable(false);
    BuildGuiSetBackground(dio, toolbar);
    QMainWindow *win = dynamic_cast<QMainWindow *>(parent->parentWidget());
    if (win) {
        win->addToolBar(Qt::TopToolBarArea, toolbar);
    } else {
        vgroup->addWidget(toolbar);
    }

    QToolButton *menu = new QToolButton(toolbar);
    menu->setText(title);
    menu->setMenu(TableMenu);
    menu->setFocusPolicy(Qt::TabFocus);
    // menu->setPopupMode(QToolButton::MenuButtonPopup);
    menu->setPopupMode(QToolButton::InstantPopup);
    BuildGuiSetFont(dio, menu);
    SetGuiObjectDisplayObject(menu, dio);
    toolbar->addWidget(menu);
    toolbar->addSeparator();

    ArrowButton *left = 0;
    ArrowButton *right = 0;
    if (gadgets && DisplayTablePanelArrows(dio)) {
        left = new ArrowButton(false, toolbar);
        SetGuiObjectDisplayObject(left, dio);
        toolbar->addWidget(left);
        toolbar->addSeparator();
        right = new ArrowButton(true, toolbar);
        SetGuiObjectDisplayObject(right, dio);
        toolbar->addWidget(right);
        toolbar->addSeparator();
    }

    QToolButton *filter = 0;
    if (gadgets && DisplayTablePanelFilter(dio)) {
        filter = new FilterButton(toolbar);
        SetGuiObjectDisplayObject(filter, dio);
        toolbar->addWidget(filter);
        toolbar->addSeparator();
    }

    QWidget *space = new QWidget(toolbar);
    space->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
    toolbar->addWidget(space);

    QLabel *num = 0;
    if (gadgets && DisplayTablePanelNumAll(dio)) {
        num = new QLabel("    #0/0", toolbar);
        num->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
        num->setContentsMargins(6, 0, 6, 0);
        BuildGuiSetFont(dio, num);
        toolbar->addWidget(num);
    }

    QWidget *contents;
    if (0) {
        QFrame *frame = new QFrame(main);
        frame->setFrameShape(QFrame::StyledPanel);
        frame->setFrameShadow(QFrame::Raised);
        contents = frame;
    } else {
        contents = new QWidget(main);
    }

#elif (defined USE_BUTTONSTYLE)
    GroupLayout *hgroup = new GroupLayout(INT_MAX);
    hgroup->setContentsMargins(0, 0, 0, 0);
    hgroup->setSpacing(0);
    vgroup->addLayout(hgroup);
    vgroup->setStretch(1, 0);

    PanelButtonBox *head = new PanelButtonBox(main);
    BuildGuiSetBackground(dio, head);
    hgroup->addWidget(head);
    hgroup->setStretch(1, 0);

    GroupLayout *headGroup = new GroupLayout(INT_MAX, head);
    int w = 5 - head->style()->pixelMetric(QStyle::PM_DefaultFrameWidth);
    headGroup->setContentsMargins(w+2, w, w+2, w);

    QLabel *name = new QLabel(title, head);
    name->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
    BuildGuiSetFont(dio, name);
    headGroup->addWidget(name);

    QLabel *num = 0;
    if (gadgets && DisplayTablePanelNumAll(dio)) {
        num = new QLabel("    #0/0", head);
        num->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
        BuildGuiSetFont(dio, num);
        headGroup->addWidget(num);
    }
    int size = head->sizeHint().height();

    QToolButton *filter = 0;
    if (gadgets && DisplayTablePanelFilter(dio)) {
        filter = new FilterButton(main);
        filter->setFixedSize(size, size);
        SetGuiObjectDisplayObject(filter, dio);
        BuildGuiSetBackground(dio, filter);
        hgroup->addWidget(filter);
    }

    ArrowButton *left = 0;
    ArrowButton *right = 0;
    if (gadgets && DisplayTablePanelArrows(dio)) {
        left = new ArrowButton(false, main);
        left->setFixedSize(size, size);
        BuildGuiSetBackground(dio, left);
        SetGuiObjectDisplayObject(left, dio);
        hgroup->addWidget(left);
        right = new ArrowButton(true, main);
        right->setFixedSize(size, size);
        BuildGuiSetBackground(dio, right);
        SetGuiObjectDisplayObject(right, dio);
        hgroup->addWidget(right);
    }

    QToolButton *menu = new TableMenuButton(main);
    menu->setFixedSize(size, size);
    BuildGuiSetBackground(dio, menu);
    SetGuiObjectDisplayObject(menu, dio);
    hgroup->addWidget(menu);

    PanelButtonBox *contents = new PanelButtonBox(main);
    //contents->setSizePolicy(QSizePolicy::MinimumExpanding,
    //        QSizePolicy::MinimumExpanding);
#endif
    
    vgroup->addWidget(contents);
    vgroup->setStretch(1, 1);
    *sub = contents;

    GuiSetHelpNode(main, "Browsing records");

    SetGuiObjectDisplayObject(main, dio);
    SetDisplayObjectGuiObject(dio, main);
    SetDisplayObjectRootObject(dio, main);
    SetDisplayObjectActivateObject(dio, main);
    SetDisplayTablePanelNumAllObject(dio, num);
    SetDisplayTablePanelFilterObject(dio, filter);

    return main;
}

bool BuildGuiTable(DisplayObject *dio, QWidget *parent, GroupLayout *group)
{
    if (!group) {
        return false;
    }

    QWidget *main = 0, *sub = 0;
    if (DisplayTableHasPanel(dio)) {
        main = buildGuiPanel(dio, parent, &sub);
        if (!main || !sub) {
            return false;
        }

        group->addWidget(main);
        SetGuiObjectDisplayObject(main, dio);
        parent = sub;
    }

    SetDisplayObjectGuiObject(dio, main);
    SetDisplayObjectRootObject(dio, main);

    GroupLayout *newGroup = new GroupLayout(1, sub);
    if (!sub) {
        group->addLayout(newGroup);
        if (group->hasMargin()) {
            newGroup->setContentsMargins(0, 0, 0, 0);
        }
    }

    bool rc = true;
    List *childs = DisplayObjectObjects(dio);
    if (childs) {
        GroupLayout *colGroup = 0;
        rc = BuildGuiChilds(childs, false, parent, newGroup, &colGroup);
        BuildGuiFinishColGroup(&colGroup);
    }

    Qt::Orientations orientations = newGroup->expandingDirections();
    int weight = DisplayObjectWeight(dio);        
    int horizWeight = orientations & Qt::Horizontal ? weight : 0;
    int vertWeight = orientations & Qt::Vertical ? weight : 0;
    group->setStretch(horizWeight, vertWeight);

#ifdef USE_TOOLBAR
    if (main) {
        // update the size constraint of the main object
        GroupLayout *mainGroup = dynamic_cast<GroupLayout *>(main->layout());
        QFrame *line = new QFrame(main->parentWidget());
        line->setFrameShape(QQFrame::StyledPanel;
        line->setFrameShadow(QFrame::Sunken);
        mainGroup->addWidget(line);
    }
#endif
    
    return rc;
}

void GuiSetNumAll(GuiObject go, ulong num, ulong all)
{
    QLabel *label = dynamic_cast<QLabel *>(go);

    if (label) {
	char buffer[31];

	sprintf(buffer, "#%ld/%ld", num, all);
	label->setText(buffer);
    }
}

#include "qtBuildGuiTable.moc"

