//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtBuildGuiWindow.cc
//
// Creation date: 03.03.2009
// Last modified: 02.04.2022
//

#include "system.h"
#include "qt/qtProjectWindow.h"
#include "qt/qtBuildGui.h"
#include "qt/qtMenus.h"
#include "qt/qtUtil.h"
#include <QApplication>
#include <QMenuBar>
#include <QPushButton>
#include <QLabel>
#include <QPixmap>
#include <QVBoxLayout>
#include <QCloseEvent>
#include "menuproject.h"
#include "structureeditor.h"
#include "buildgui.h"
#include "gui.h"
#include "display.h"
#include "appstrings_cat.h"

class WindowButton : public FocusButton
{
public:
    WindowButton(QWidget *parent, DisplayObject *dio);
};

WindowButton::WindowButton(QWidget *parent, DisplayObject *dio)
    : FocusButton(parent)
{
    const char *title = DisplayObjectTitle(dio);
    QString text = GuiMarkShortcut(title, DisplayObjectShortcut(dio));
    setText(text.append("..."));
    setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
}

bool BuildGuiWindowButton(DisplayObject *dio, QWidget *parent, 
        GroupLayout *group, GroupLayout **colGroup)
{
    if (!DisplayWindowButton(dio)) {
        return true;
    }

    if (group == 0) {
        return false;
    }

    BuildGuiFinishColGroup(colGroup);

    WindowButton *button = new WindowButton(parent, dio);
    group->addWidget(button);
    group->setStretch(DisplayObjectWeight(dio), 0);

    BuildGuiSetBackground(dio, button);
    BuildGuiSetFont(dio, button);

    SetDisplayWindowButtonObject(dio, button);
    SetDisplayObjectRootObject(dio, button);
    SetGuiObjectDisplayObject(button, dio);

    return true;
}

static QWidget *empty_display_new(QWidget *parent, bool addHelp)
{
    QString path = qtExpandPathname(MAIN_WINDOW_LOGO_IMAGE);
    QPixmap pixmap(path);
    QLabel *label = new QLabel(parent);

    if (pixmap.isNull()) {
        label->setText(GCS(MSG_EmptyDisplay));
    } else {
        label->setPixmap(pixmap);
        
        QPalette palette = label->palette();
	palette.setColor(QPalette::Window, QColor(
				MAIN_WINDOW_LOGO_BG_RED,
				MAIN_WINDOW_LOGO_BG_GREEN,
				MAIN_WINDOW_LOGO_BG_BLUE));
        palette.setColor(QPalette::WindowText, QColor(0xff, 0xff, 0xff));
        label->setPalette(palette);
        parent->setPalette(palette);
    }
    label->setMinimumSize(MAIN_WINDOW_LOGO_MIN_WIDTH,
		    MAIN_WINDOW_LOGO_MIN_HEIGHT);
    label->setAlignment(Qt::AlignCenter);
    
#ifndef BEEBASE_LIGHT
    if (addHelp) {
        label->setToolTip(GCS(MSG_EmptyDisplayBH));
    }
#endif

    return label;
}

class UserWindow : public ProjectWindow
{
    Q_OBJECT
    
public:
    UserWindow(Project *pr, bool main) : ProjectWindow(pr), main_(main) {}

public slots:
    void present();
    void updateSizeConstraints();
  
protected:
    virtual void closeEvent(QCloseEvent *event);
    virtual void showEvent(QShowEvent *event);
    virtual void hideEvent(QHideEvent *event);
    virtual void moveEvent(QMoveEvent *event);
    virtual void resizeEvent(QResizeEvent *event);
    virtual void changeEvent(QEvent *event);
    
private:
    bool main_;
    
    void saveGeometry();
};

void UserWindow::present()
{
    show();
    raise();
    activateWindow();
}

void UserWindow::updateSizeConstraints()
{
    QWidget *central = centralWidget();
    if (central) {
        QLayout *layout = central->layout();
        if (layout) {
            QSize minSize = minimumSize();
            QSize maxSize = maximumSize();
            Qt::Orientations orientations = layout->expandingDirections();
            QSize size(QWIDGETSIZE_MAX, QWIDGETSIZE_MAX);
            if ((orientations & Qt::Horizontal) == 0) {
                size.setWidth(minSize.width());
            }
            if ((orientations & Qt::Vertical) == 0) {
                size.setHeight(minSize.height());
            }
            if (size != maxSize) {
                setMaximumSize(size);
            }
        }
    }
}

void UserWindow::closeEvent(QCloseEvent *event)
{
    GuiStoreEdits();
    if (main_) {
        event->ignore();
        MenuProjectClose(project());
    } else {
        event->accept();
    }
}
        
void UserWindow::showEvent(QShowEvent *event)
{
    DisplayObject *dio = GuiObjectDisplayObject(this);

    // ensure window position is correct
    // GuiSetWindowDimensions(win, DisplayWindowDimensions(dio));
    updateSizeConstraints();
    GuiStoreEdits();
    GuiUpdateWindow(this, DisplayObjectProject(dio));
    SetDisplayWindowOpen(dio, true);
    ProjectWindow::showEvent(event);
}

void UserWindow::hideEvent(QHideEvent *event)
{
    DisplayObject *dio = GuiObjectDisplayObject(this);

    GuiStoreEdits();
    GuiClearWindow(this, DisplayObjectProject(dio));
    SetDisplayWindowOpen(dio, false);
    ProjectWindow::hideEvent(event);
}

void UserWindow::moveEvent(QMoveEvent *event)
{
    ProjectWindow::moveEvent(event);
    saveGeometry();
}

void UserWindow::resizeEvent(QResizeEvent * event)
{
    ProjectWindow::resizeEvent(event);
    saveGeometry();
}

void UserWindow::changeEvent(QEvent *event)
{
    ProjectWindow::changeEvent(event);
    QWindowStateChangeEvent *wscEvent = dynamic_cast<QWindowStateChangeEvent *>(event);
    if (!wscEvent) {
        return;
    }
    bool wasIconfied = (wscEvent->oldState() & Qt::WindowMinimized) != 0;
    bool isIconfied = (windowState() & Qt::WindowMinimized) != 0;
    if (wasIconfied == isIconfied) {
        return;
    }

    Project *pr = project();
    DisplayObject *dio = ProjectRootDisplayObject(pr);
    GuiObject main = DisplayObjectGuiObject(dio);        
    if (!main) {
        return;
    }

    if (this == main) {
        if (isIconfied) {
            foreach (QWidget *top, QApplication::topLevelWidgets()) {
                if (top != main && GuiWindowProject(top) == pr) {
                    top->setProperty("open", top->isVisible());
                    top->hide();
                }
            }
        } else {
            foreach (QWidget *top, QApplication::topLevelWidgets()) {
                if (top != this && GuiWindowProject(top) == pr) {
                    if (top->property("open").toBool()) {
                        top->show();
                    }
                }
            }
            main->activateWindow();
        }
    }            
}

void UserWindow::saveGeometry()
{
    WindowDimensions wd;
    DisplayObject *dio = GuiObjectDisplayObject(this);
    
    if (GuiGetWindowDimensions(this, &wd)) {
        SetDisplayWindowDimensions(dio, &wd);
    }
}

bool BuildGuiWindow(DisplayObject *dio)
{
    Project *pr = DisplayObjectProject(dio);
    bool main = (DisplayObjectParent(dio) == 0);
    bool empty = !DisplayObjectVisibleGui(dio);
    const char *title = DisplayObjectTitle(dio);
    const WindowDimensions *wd = DisplayWindowDimensions(dio);

    UserWindow *win = new UserWindow(pr, main);
    SetGuiObjectDisplayObject(win, dio);
    SetDisplayObjectGuiObject(dio, win);

    GuiSetWindowDimensions(win, wd);

    if (main) {
        SetRootWindowTitle(pr, win);
    } else {
        win->setWindowTitle(title);
        GuiObject go = DisplayWindowButtonObject(dio);
        WindowButton *button = dynamic_cast<WindowButton *>(go);

        if (button) {
            QObject::connect(button, SIGNAL(clicked()), win, SLOT(present()));
        }
    }
#ifndef MACOSX
    GuiWindowAddMenus(win, main);
#endif
    
    // GuiAddWindowCallbacks(win, NULL);

    bool rc = false;
    if (empty) {
        QWidget *empty = empty_display_new(win, main);
        
        win->setCentralWidget(empty);
	win->resize(MAIN_WINDOW_LOGO_DEFAULT_WIDTH,
			MAIN_WINDOW_LOGO_DEFAULT_HEIGHT);
        rc = true;
    } else {
        QWidget *root = new QWidget(win);
        GroupLayout *group = new GroupLayout(1, root);
        // group->setContentsMargins(0, 0, 0, 0);
        GroupLayout *colGroup = 0;
        
        win->setCentralWidget(root);
    
        List *childs = DisplayObjectObjects(dio);
        rc = BuildGuiChilds(childs, false, root, group, &colGroup);
        BuildGuiFinishColGroup(&colGroup);
    }

    return rc;
}

#include "qtBuildGuiWindow.moc"
