//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtEditDisplay.cc
//
// Creation date: 17.02.2016
// Last modified: 08.11.2021
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "qt/qtEditDisplay.h"
#include "qt/qtStructureEditor.h"
#include "qt/qtStructureEditorUtil.h"
#include "qt/qtUtil.h"
#include "qt/classes/qtEditLabels.h"
#include "qt/classes/qtGroupLayout.h"
#include "qt/classes/qtPopImage.h"
#include <QCheckBox>
#include <QLabel>
#include <QSlider>
#include "display.h"
#include "requesters.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"

class EditTextDialog : public CommonDialog
{
    Q_OBJECT
    
public:
    EditTextDialog(GuiObject sew, DisplayObject *text);
    
private:
    Project *pr_;
    DisplayObject *text_;
    GuiObject sew_;
    QLineEdit *title_;
    QSlider *weight_;
    QLineEdit *fontName_;
    PopImage *background_;
    QCheckBox *defaultBackground_;

    virtual bool acceptEdits()
    {
	GuiBusyCursor busy;

        long error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return false;
        }

        Utf8Buffer buffer;
	struct TextData td;
        td.td_Title = qtStringUtf8(title_->text(), buffer);
        td.td_Weight = weight_->value();
        td.td_Fontname = qtStringUtf8(fontName_->text(), buffer);
	td.td_Background = qtStringUtf8(background_->spec(), buffer);
        
        DisplayChangeText(text_, &td);
        ProjectChange(pr_, true);

        TextFinish(sew_, text_);
        return true;
    }
};

EditTextDialog::EditTextDialog(GuiObject sew, DisplayObject *text)
    : CommonDialog(sew)
{
    setAttribute(Qt::WA_DeleteOnClose);

    pr_ = DisplayObjectProject(text);
    text_ = text;
    sew_ = sew;

    GroupLayout *top = new GroupLayout(2);
    contentsLayout()->addLayout(top);
    
    title_ = new QLineEdit(this);
    title_->setMaxLength(TITLE_LEN);
    title_->setText(DisplayObjectTitle(text));
    title_->setToolTip(GCS(MSG_DisplayGroup_TitleBH));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Title)), title_);    

    weight_ = new QSlider(this);
    QHBoxLayout *hbox = WeightSlider(weight_, this);
    weight_->setValue(DisplayObjectWeight(text));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Weight)), weight_, hbox);
    
    fontName_ = new QLineEdit(this);
    fontName_->setText(DisplayObjectQtFontname(text));
    fontName_->setMaxLength(FONTNAME_LEN);
    fontName_->setToolTip(GCS(MSG_FontBH));
    PopSelection *popFont = MakeFontPopup(fontName_, this);
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Font)), fontName_, popFont);
  
    background_ = new PopImage(this, pr_);
    defaultBackground_ = new QCheckBox(this);
    hbox = BackgroundGadgets(background_, defaultBackground_,
                             DisplayObjectBackground(text));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Background)), background_, hbox);
    
    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));
    
    setWindowTitle(GCS(MSG_Text));
    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Text editor");
}

void GuiText(GuiObject sew, DisplayObject *text)
{
    EditTextDialog *dialog = new EditTextDialog(sew, text);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }
}

class EditSpaceDialog : public CommonDialog
{
    Q_OBJECT

public:
    EditSpaceDialog(GuiObject sew, DisplayObject *space);

private:
    Project *pr_;
    DisplayObject *space_;
    GuiObject sew_;
    QCheckBox *bar_;
    QSlider *weight_;
    PopImage *background_;
    QCheckBox *defaultBackground_;

    virtual bool acceptEdits()
    {
	GuiBusyCursor busy;

        long error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return false;
        }

        Utf8Buffer buffer;
	SpaceData sd;
        sd.sd_Bar = bar_->isChecked();
        sd.sd_Weight = weight_->value();
	sd.sd_Background = qtStringUtf8(background_->spec(), buffer);

        DisplayChangeSpace(space_, &sd);
        ProjectChange(pr_, true);

        SpaceFinish(sew_, space_);
        return true;
    }
};

EditSpaceDialog::EditSpaceDialog(GuiObject sew, DisplayObject *space)
    : CommonDialog(sew)
{
    setAttribute(Qt::WA_DeleteOnClose);

    pr_ = DisplayObjectProject(space);
    space_ = space;
    sew_ = sew;

    GroupLayout *top = new GroupLayout(2);
    contentsLayout()->addLayout(top);

    bar_ = new QCheckBox(this);
    bar_->setChecked(DisplaySpaceBar(space));
    bar_->setToolTip(GCS(MSG_DisplaySpace_DelimiterBH));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_DisplaySpace_Delimiter)),
            bar_);

    weight_ = new QSlider(this);
    QHBoxLayout *hbox = WeightSlider(weight_, this);
    weight_->setValue(DisplayObjectWeight(space));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Weight)), weight_, hbox);
    
    background_ = new PopImage(this, pr_);
    defaultBackground_ = new QCheckBox(this);
    hbox = BackgroundGadgets(background_, defaultBackground_,
                             DisplayObjectBackground(space));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Background)),
            background_, hbox);

    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));
    
    setWindowTitle(GCS(MSG_Space));
    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Space editor");
}

void GuiSpace(GuiObject sew, DisplayObject *space)
{
    EditSpaceDialog *dialog = new EditSpaceDialog(sew, space);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }
}

class EditImageDialog : public CommonDialog
{
    Q_OBJECT
    
public:
    EditImageDialog(GuiObject sew, DisplayObject *image);
    
private:
    Project *pr_;
    DisplayObject *image_;
    GuiObject sew_;
    QSlider *weight_;
    PopImage *popImage_;
    RadioBox *imageSize_;

    virtual bool acceptEdits() {
        GuiBusyCursor busy;
        
        long error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return false;
        }

        Utf8Buffer buffer;
        ImageData id;
        id.id_Weight = weight_->value();
        id.id_Image = qtStringUtf8(popImage_->spec(), buffer);
        id.id_ImageSize = imageSize_->checkedId();

        DisplayChangeImage(image_, &id);
        ProjectChange(pr_, true);

        ImageFinish(sew_, image_);
        return true;
    }
};

EditImageDialog::EditImageDialog(GuiObject sew, DisplayObject *image)
    : CommonDialog(sew)
{
    setAttribute(Qt::WA_DeleteOnClose);
    
    pr_ = DisplayObjectProject(image);
    image_ = image;
    sew_ = sew;
    
    GroupLayout *top = new GroupLayout(2);
    contentsLayout()->addLayout(top);

    weight_ = new QSlider(this);
    QHBoxLayout *hbox = WeightSlider(weight_, this);
    weight_->setValue(DisplayObjectWeight(image));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Weight)), weight_, hbox);
    
    popImage_ = new PopImage(this, pr_);
    popImage_->setSpec(DisplayObjectBackground(image));
    popImage_->setToolTip(GCS(MSG_ImageBH));

    static char *entries[3];
    entries[0] = GCS(MSG_Image_Size_Resizeable);
    entries[1] = GCS(MSG_Image_Size_Fixed);
    entries[2] = NULL;
    imageSize_ = new RadioBox(this, entries, GuiMarkShortcut(GCS(MSG_Size)), true);
    imageSize_->setCheckedId(DisplayImageImageSize(image));
    hbox = new QHBoxLayout;
    hbox->addWidget(popImage_, 2);
    hbox->addWidget(imageSize_, 1, Qt::AlignVCenter);
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG__Image)), popImage_, hbox);

    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));
    
    setWindowTitle(GCS(MSG_Image));
    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Image editor");
}

void GuiImage(GuiObject sew, DisplayObject *image)
{
    EditImageDialog *dialog = new EditImageDialog(sew, image);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }
}

class EditRegisterDialog : public CommonDialog
{
    Q_OBJECT
    
public:
    EditRegisterDialog(GuiObject sew, DisplayObject *reg);
    
private:
    Project *pr_;
    DisplayObject *register_;
    GuiObject sew_;
    QSlider *weight_;
    EditLabels *labels_;

    virtual bool acceptEdits()
    {
	GuiBusyCursor busy;

        long error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return false;
        }

	RegisterData rd;
        rd.rd_Weight = weight_->value();
        rd.rd_Fontname = 0;
        rd.rd_Background = 0;
	rd.rd_Labels = labels_->labels();

        error = DisplayChangeRegister(register_, &rd);
	if (error) {
	    Ask2(this, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));
            return false;
	}
        
        ProjectChange(pr_, true);
        RegisterFinish(sew_, register_);
        return true;
    }
};

EditRegisterDialog::EditRegisterDialog(GuiObject sew, DisplayObject *reg)
    : CommonDialog(sew)
{
    setAttribute(Qt::WA_DeleteOnClose);
    
    pr_ = DisplayObjectProject(reg);
    register_ = reg;
    sew_ = sew;
    
    GroupLayout *top = new GroupLayout(2);
    contentsLayout()->addLayout(top);
    
    weight_ = new QSlider(this);
    QHBoxLayout *hbox = WeightSlider(weight_, this);
    weight_->setValue(DisplayObjectWeight(reg));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Weight)), weight_, hbox);
    
    labels_ = new EditLabels(this);
    labels_->setLabels(DisplayRegisterLabels(reg));
    labels_->setToolTip(GCS(MSG_DisplayRegister_LabelsBH));

    GroupBox *box = new GroupBox(GuiMarkShortcut(GCS(MSG_DisplayRegister_Labels)), this);
    QVBoxLayout *vbox = new QVBoxLayout(box);
    vbox->addWidget(labels_);
    top->addWidget(box, 2);

    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));
    
    setWindowTitle(GCS(MSG_Register));
    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Register group editor");
}

void GuiRegister(GuiObject sew, DisplayObject *reg)
{
    EditRegisterDialog *dialog = new EditRegisterDialog(sew, reg);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }
}

class EditGroupDialog : public CommonDialog
{
    Q_OBJECT
    
public:
    EditGroupDialog(GuiObject sew, DisplayObject *group);
    
private:
    Project *pr_;
    DisplayObject *group_;
    GuiObject sew_;
    QLineEdit *title_;
    QSlider *weight_;
    QCheckBox *border_;
    QCheckBox *horiz_;
    QCheckBox *spacing_;
    PopImage *background_;
    QCheckBox *defaultBackground_;
    
    bool acceptEdits()
    {
        GuiBusyCursor busy;

        long error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return false;
        }
        
        Utf8Buffer buffer;
	GroupData gd;
        gd.gd_Title = qtStringUtf8(title_->text(), buffer);
        gd.gd_Weight = weight_->value();
        gd.gd_Fontname = 0;
	gd.gd_Background = qtStringUtf8(background_->spec(), buffer);
	gd.gd_Horiz = horiz_->isChecked();
	gd.gd_Border = border_->isChecked();
	gd.gd_Spacing = spacing_->isChecked();

        DisplayChangeGroup(group_, &gd);
        ProjectChange(pr_, true);

        GroupFinish(sew_, group_);
        return true;
    }
};

EditGroupDialog::EditGroupDialog(GuiObject sew, DisplayObject *group) 
    : CommonDialog(sew)
{
    setAttribute(Qt::WA_DeleteOnClose);
    
    pr_ = DisplayObjectProject(group);
    group_ = group;
    sew_ = sew;

    GroupLayout *top = new GroupLayout(2);
    contentsLayout()->addLayout(top);

    title_ = new QLineEdit(this);
    title_->setMaxLength(TITLE_LEN);
    title_->setText(DisplayObjectTitle(group));
    title_->setToolTip(GCS(MSG_DisplayGroup_TitleBH));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Title)), title_);
    
    weight_ = new QSlider(this);
    QHBoxLayout *hbox = WeightSlider(weight_, this);
    weight_->setValue(DisplayObjectWeight(group));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Weight)), weight_, hbox);
    
    background_ = new PopImage(this, pr_);
    defaultBackground_ = new QCheckBox(this);
    hbox = BackgroundGadgets(background_, defaultBackground_, 
                             DisplayObjectBackground(group));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Background)), background_, hbox);

    GroupBox *attrs = new GroupBox(GCS(MSG_DisplayGroup_Attributes), this);
    contentsLayout()->addWidget(attrs);
    GroupLayout *sub = new GroupLayout(1, attrs);

    border_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_DisplayGroup_Border)), this);
    border_->setToolTip(GCS(MSG_DisplayGroup_BorderBH));
    border_->setChecked(DisplayGroupBorder(group));
    sub->addWidget(border_);

    horiz_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_DisplayGroup_Horizontal)), this);
    horiz_->setToolTip(GCS(MSG_DisplayGroup_HorizontalBH));
    horiz_->setChecked(DisplayGroupHoriz(group));
    sub->addWidget(horiz_);

    spacing_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_DisplayGroup_Spacing)), this);
    spacing_->setToolTip(GCS(MSG_DisplayGroup_SpacingBH));
    spacing_->setChecked(DisplayGroupSpacing(group));
    sub->addWidget(spacing_);
        
    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));
    
    setWindowTitle(GCS(MSG_Group));
    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Group editor");
}    

void GuiGroup(GuiObject sew, DisplayObject *group)
{
    EditGroupDialog *dialog = new EditGroupDialog(sew, group);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }    
}

class EditWindowDialog : public CommonDialog
{
    Q_OBJECT

public:
    EditWindowDialog(GuiObject sew, DisplayObject *window);

private:
    Project *pr_;
    DisplayObject *window_;
    GuiObject sew_;
    QLineEdit *title_;
    QCheckBox *button_;
    QLineEdit *shortcut_;
    RadioBox *enabled_;
    QLineEdit *enabledComputeStr_;
    PopString *enabledCompute_;
    QCheckBox *closeWhenDisabled_;
    QSlider *weight_;
    QLineEdit *fontName_;
    PopImage *background_;
    QCheckBox *defaultBackground_;

    virtual bool acceptEdits()
    {
        long error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return false;
        }

        Utf8Buffer buffer;
	WindowData wd;
        wd.wd_Title = qtStringUtf8(title_->text(), buffer);
	wd.wd_Id = DisplayWindowId(window_);
	wd.wd_Button = button_->isChecked();
	wd.wd_Shortcut = CodesetGetChar(qtStringUtf8(shortcut_->text(),buffer));
	wd.wd_Enabled = enabled_->checkedId();
        wd.wd_EnabledCompute = qtStringUtf8(enabledComputeStr_->text(), buffer);
        wd.wd_CloseWhenDisabled = closeWhenDisabled_->isChecked();
        wd.wd_Weight = weight_->value();
        wd.wd_Fontname = qtStringUtf8(fontName_->text(), buffer);
	wd.wd_Background = qtStringUtf8(background_->spec(), buffer);

        DisplayChangeWindow(window_, &wd);
        ProjectChange(pr_, true);

        WindowFinish(sew_, window_);
        return true;
    }

    bool eventFilter(QObject *obj, QEvent *event)
    {
        if (event->type() == QEvent::KeyPress) {
            QKeyEvent *keyEvent = static_cast<QKeyEvent *>(event);
            bool res = QObject::eventFilter(obj, event);

            if (keyEvent->key() == Qt::Key_Return) {
                CheckShortCut(shortcut_, window_);
                return true;  // Always accept return
            } else {
                return res;
            }
        } else {
            // standard event processing
            return QObject::eventFilter(obj, event);
        }
    }

private slots:
    void onEnabledChanged(int enabled)
    {
        enabledComputeStr_->setEnabled(enabled == DOD_ENABLED_COMPUTED);
        enabledCompute_->button()->setEnabled(enabled == DOD_ENABLED_COMPUTED);
        closeWhenDisabled_->setEnabled(enabled != DOD_ENABLED_ALWAYS);
        if (enabled == DOD_ENABLED_ALWAYS) {
            closeWhenDisabled_->setChecked(false);
        }
    }
};

EditWindowDialog::EditWindowDialog(GuiObject sew, DisplayObject *window)
    : CommonDialog(sew)
{
    setAttribute(Qt::WA_DeleteOnClose);
    
    pr_ = DisplayObjectProject(window);
    window_ = window;
    sew_ = sew;
    
    GroupLayout *top = new GroupLayout(2);
    contentsLayout()->addLayout(top);

    title_ = new QLineEdit(this);
    title_->setMaxLength(TITLE_LEN);
    title_->setText(DisplayObjectTitle(window));
    title_->setToolTip(GCS(MSG_DisplayWindow_TitleBH));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Title)), title_);
    
    button_ = new QCheckBox(this);
    button_->setChecked(DisplayWindowButton(window));
    button_->setToolTip(GCS(MSG_DisplayWindow_WindowButtonBH));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_DisplayWindow_WindowButton)),
               button_);

    GroupBox *buttonSettings = new GroupBox(GCS(MSG_DisplayWindow_ButtonSettings), this);
    contentsLayout()->addWidget(buttonSettings);
    connect(button_, SIGNAL(toggled(bool)), buttonSettings, SLOT(setEnabled(bool)));

    GroupLayout *settings = new GroupLayout(1, buttonSettings);
    GroupLayout *settings1 = new GroupLayout(3);
    settings->addLayout(settings1);

    char sc[CodesetMaxBytesForChar + 1];
    CodesetCharToStr(DisplayObjectShortcut(window), sc);
    shortcut_ = new QLineEdit(this);
    shortcut_->setMaxLength(1);
    shortcut_->setText(sc);
    shortcut_->setToolTip(GCS(MSG_DisplayObject_ShortcutBH));
    shortcut_->setFixedWidth(shortcut_->fontMetrics().maxWidth() * 4);
    shortcut_->installEventFilter(this);
    settings1->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_DisplayObject_Shortcut)),
            shortcut_);

    settings1->addWidget(new QFrame(this));
    settings1->setStretch(1, 0);

    GroupBox *enabledSettings = new GroupBox(GCS(MSG_DisplayObject_Enabled), this);
    settings->addWidget(enabledSettings);
    GroupLayout *settings2 = new GroupLayout(1, enabledSettings);

    GroupLayout *settings21 = new GroupLayout(3);
    settings21->setContentsMargins(0, 0, 0, 0);
    settings2->addLayout(settings21);

    static char *etext[4];
    etext[0] = GCS(MSG_DisplayObject_Enabled_Always);
    etext[1] = GCS(MSG_DisplayObject_Enabled_NotInitial);
    etext[2] = GCS(MSG_DisplayObject_Enabled_Computed);
    etext[3] = 0;
    enabled_ = new RadioBox(this, etext);
    enabled_->setToolTip(GCS(MSG_DisplayObject_EnabledBH));
    settings21->addWidget(enabled_);

    QFrame *vSep = new QFrame(this);
    vSep->setFrameShape(QFrame::VLine);
    vSep->setFrameShadow(QFrame::Raised);
    settings21->addWidget(vSep);

    GroupLayout *settings211 = new GroupLayout(1);
    settings21->addLayout(settings211);

    QFrame *space = new QFrame(this);
    settings211->addWidget(space);

    enabledComputeStr_ = new QLineEdit(this);
    enabledComputeStr_->setMaxLength(FIELD_TRIGGER_LEN);
    enabledComputeStr_->setText(DisplayObjectEnabledCompute(window));
    enabledComputeStr_->setToolTip(GCS(MSG_DisplayObject_ComputeEnabledBH));
    enabledComputeStr_->setEnabled(false);

    enabledCompute_ = UserFunctionsPopup(enabledComputeStr_, pr_, this);
    enabledCompute_->button()->setEnabled(false);
    settings211->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Compute)),
            enabledComputeStr_, enabledCompute_);

    QFrame *hSep = new QFrame(this);
    hSep->setFrameShape(QFrame::HLine);
    hSep->setFrameShadow(QFrame::Raised);
    settings2->addWidget(hSep);

    closeWhenDisabled_ = new QCheckBox(
            GuiMarkShortcut(GCS(MSG_DisplayWindow_CloseWindowWhenDisabled)), this);
    closeWhenDisabled_->setChecked(DisplayWindowCloseWhenDisabled(window));
    closeWhenDisabled_->setToolTip(GCS(MSG_DisplayWindow_CloseWindowWhenDisabledBH));
    settings2->addWidget(closeWhenDisabled_);

    connect(enabled_, SIGNAL(checked(int)), this, SLOT(onEnabledChanged(int)));
    char enabled = DisplayObjectEnabled(window);
    enabled_->setCheckedId(enabled);
    onEnabledChanged(enabled);

    GroupLayout *settings3 = new GroupLayout(2);
    settings->addLayout(settings3);

    weight_ = new QSlider(this);
    QHBoxLayout *hbox = WeightSlider(weight_, this);
    weight_->setValue(DisplayObjectWeight(window));
    settings3->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Weight)), weight_, hbox);
    
    fontName_ = new QLineEdit(this);
    fontName_->setText(DisplayObjectQtFontname(window));
    fontName_->setMaxLength(FONTNAME_LEN);
    fontName_->setToolTip(GCS(MSG_FontBH));
    PopSelection *popFont = MakeFontPopup(fontName_, this);
    settings3->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Font)), fontName_, popFont);
  
    background_ = new PopImage(this, pr_);
    defaultBackground_ = new QCheckBox(this);
    hbox = BackgroundGadgets(background_, defaultBackground_,
                             DisplayObjectBackground(window));
    settings3->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Background)), background_, hbox);
    
    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

    setWindowTitle(GCS(MSG_Window));
    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Window editor");
}

void GuiWindow(GuiObject sew, DisplayObject *window)
{
    EditWindowDialog *dialog = new EditWindowDialog(sew, window);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }    
}

#include "qtEditDisplay.moc"

#endif

