//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtEditDisplayField.cc
//
// Creation date: 14.04.2016
// Last modified: 10.02.2022
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "qt/qtEditDisplayField.h"
#include "qt/qtEditRecordDescriptionSpec.h"
#include "qt/qtStructureEditor.h"
#include "qt/qtStructureEditorUtil.h"
#include "qt/qtUtil.h"
#include "qt/classes/qtGroupLayout.h"
#include "qt/classes/qtPopEditLabels.h"
#include <QSpinBox>
#include <QStackedWidget>
#include "requesters.h"
#include "preferences.h"
#include "error.h"
#include "appstrings_cat.h"

class EditDisplayFieldDialog : public CommonDialog
{
    Q_OBJECT

public:
    EditDisplayFieldDialog(GuiObject sew, DisplayObject *dio);

private slots:
    void onDoubleClickChanged(int action)
    {
        bool enabled = (action == DAD_DOUBLECLICK_CALLTRIGGER);
        doubleClickTriggerStr_->setEnabled(enabled);
        doubleClickTrigger_->button()->setEnabled(enabled);
    }

    void onEnabledChanged(int enabled)
    {
        enabledComputeStr_->setEnabled(enabled == DOD_ENABLED_COMPUTED);
        enabledCompute_->button()->setEnabled(enabled == DOD_ENABLED_COMPUTED);
    }

    void updateFocusWidgets()
    {
        bool readOnly = readOnly_ && readOnly_->isChecked();
        bool file = file_ && file_->isChecked();
        bool font = font_ && font_->isChecked();
        bool list = listview_ && listview_->isChecked();
        bool hide = hideStringField_ && hideStringField_->isChecked();
        bool show = show_ && show_->isChecked();
        bool focus = !hide && (!readOnly || file || font || list || show);

        if (shortcut_) {
            shortcut_->setEnabled(focus);
        }
        if (home_) {
            home_->setEnabled(focus);
            if (!focus) {
                home_->setChecked(false);
            }
	}
        if (tabChain_) {
            tabChain_->setEnabled(focus);
            if (!focus) {
                tabChain_->setChecked(false);
            }
	}
    }

    void onReadOnlyToggled(bool active)
    {
        updateFocusWidgets();
        if (subType_) {
            if (active) {
                subType_->setCurrentIndex(0);
            }
            subType_->setEnabled(!active);
        }
        if (calendar_) {
            if (active) {
                calendar_->setChecked(false);
            }
            calendar_->setEnabled(!active);
        }
        if (referencePopupGroup_) {
            referencePopupGroup_->setEnabled(!active);
        }
    }

    void onFileToggled(bool active)
    {
        if (active) {
            font_->setChecked(false);
            listview_->setChecked(false);
        }
        updateFocusWidgets();
    }

    void onFontToggled(bool active)
    {
        if (active) {
            file_->setChecked(false);
            listview_->setChecked(false);
        }
        updateFocusWidgets();
    }

    void onListviewToggled(bool active)
    {
        if (active) {
            file_->setChecked(false);
            font_->setChecked(false);
        }
        listviewSettings_->setEnabled(active);
        updateFocusWidgets();
    }

    void onHideStringFieldChanged(bool active)
    {
        if (active) {
            file_->setChecked(false);
            font_->setChecked(false);
            listview_->setChecked(false);
            view_->setChecked(false);
        }
        readOnly_->setEnabled(!active);
        alignment_->setEnabled(!active);
        fontname_->setEnabled(!active);
        fontnamePopup_->button()->setEnabled(!active);
        background_->setEnabled(!active);
        defaultBackground_->setEnabled(!active);
        file_->setEnabled(!active);
        font_->setEnabled(!active);
        listview_->setEnabled(!active);
        view_->setEnabled(!active);
        updateFocusWidgets();
    }

    void onDisplayExternalChanged(bool active)
    {
        if (!active) {
            hideStringField_->setChecked(false);
        }
        titleAtStringField_->setEnabled(active);
        hideStringField_->setEnabled(active);
        imageSize_->setEnabled(active);
    }

    void onRecordsChanged(int active)
    {
        computeRecordsGroup_->setEnabled(active == DAD_RECORDS_COMPUTED);
    }

private:
    Project *pr_;
    DisplayObject *dio_;
    GuiObject sew_;
    QLineEdit *title_;
    ComboBox *titlePosition_;
    QLineEdit *titleFontname_;
    PopString *titleFontnamePopup_;
    QLineEdit *shortcut_; 
    QCheckBox *home_;
    QCheckBox *tabChain_;
    QCheckBox *readOnly_;
    ComboBox *alignment_;
    QCheckBox *calendar_;
    ComboBox *format_;
    QCheckBox *showTitles_;
    ComboBox *subType_; 
    QStackedWidget *subTypeStack_;
    QCheckBox *frame_;
    QCheckBox *horizontal_;
    QCheckBox *controlTable_;
    RadioBox *onDoubleClick_;
    QLineEdit *doubleClickTriggerStr_; 
    PopString *doubleClickTrigger_;
    QLineEdit *urlDropTriggerStr_; 
    PopString *urlDropTrigger_;
    QLineEdit *sortDropTriggerStr_; 
    PopString *sortDropTrigger_;
    QSpinBox *numDecimals_;
    QCheckBox *autoUpdate_;
    RadioBox *enabled_;
    QLineEdit *enabledComputeStr_;
    PopString *enabledCompute_;
    QSlider *weight_;
    TextEditor *bubbleHelp_;
    QLineEdit *fontname_;
    PopString *fontnamePopup_;
    PopImage *background_;
    QCheckBox *defaultBackground_;
    PopImage *image_;
    RadioBox *imageSize_;
    EditRecordDescriptionSpec *recordDescriptionSpec_;
    GroupBox *referencePopupGroup_;
    ComboBox *recordsKind_; 
    QLineEdit *computeRecordsStr_;
    GroupBox *computeRecordsGroup_;
    QCheckBox *initialRecord_;
    QCheckBox *currentRecord_;
    QCheckBox *show_;
    QCheckBox *autoShow_;
    QCheckBox *filter_;
    QCheckBox *displayExternal_;
    QCheckBox *hideStringField_;
    QCheckBox *titleAtStringField_;
    QCheckBox *listview_;
    QCheckBox *file_;
    QCheckBox *font_;
    QCheckBox *view_;
    GroupBox *listviewSettings_;
    ComboBox *listviewKind_; 
    QStackedWidget *listviewKindStack_;
    PopEditLabels *listviewLabels_;
    QLineEdit *listviewComputeStr_;

    virtual bool acceptEdits()
    {
        GuiBusyCursor busy;

        Field *field = DisplayFieldField(dio_);
        short type = FieldType(field);
        MemPool *memPool = ProjectMemPool(pr_);
        struct RecordDescriptionSpec *rds = 0;
        long error = OK;
        Utf8Buffer buffer;
        DisplayFieldData dfd;

        error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return false;
        }

        memset(&dfd, 0, sizeof(dfd));
        dfd.dfd_Title = qtStringUtf8(title_->text(), buffer);
        if (titlePosition_) {
            dfd.dfd_TitlePosition = titlePosition_->currentIndex();
        }
        if (titleFontname_) {
            dfd.dfd_TitleFontname = qtStringUtf8(titleFontname_->text(),buffer);
        }
        if (shortcut_) {
            dfd.dfd_Shortcut =
                CodesetGetChar(qtStringUtf8(shortcut_->text(), buffer));
        }
        dfd.dfd_Home = home_ && home_->isChecked();
        dfd.dfd_TabChain = tabChain_ && tabChain_->isChecked();
        dfd.dfd_ReadOnly = readOnly_ && readOnly_->isChecked();
        if (subType_) {
            dfd.dfd_SubType = subType_->currentIndex();
        }
        if (alignment_) {
            dfd.dfd_Alignment = alignment_->currentIndex();
        }
        if (calendar_) {
            dfd.dfd_Calendar = calendar_->isChecked();
        }
        if (format_) {
            dfd.dfd_Format = format_->currentIndex();
        }
        if (numDecimals_) {
            dfd.dfd_NumDecimals = numDecimals_->value();
        }
        if (showTitles_) {
            dfd.dfd_ShowTitles = showTitles_->isChecked();
        }
        if (controlTable_) {
            dfd.dfd_ControlTable = controlTable_->isChecked();
        }
        if (onDoubleClick_) {
            dfd.dfd_OnDoubleClick = onDoubleClick_->checkedId();
        }
        if (doubleClickTriggerStr_) {
            dfd.dfd_DoubleClickTrigger =
                qtStringUtf8(doubleClickTriggerStr_->text(), buffer);
        }
        if (urlDropTriggerStr_) {
            dfd.dfd_URLDropTrigger =
                qtStringUtf8(urlDropTriggerStr_->text(), buffer);
        }
        if (sortDropTriggerStr_) {
            dfd.dfd_SortDropTrigger =
                qtStringUtf8(sortDropTriggerStr_->text(), buffer);
        }
        dfd.dfd_AutoUpdate = autoUpdate_ && autoUpdate_->isChecked();
        if (enabled_) {
            dfd.dfd_Enabled = enabled_->checkedId();
        }
        if (enabledComputeStr_) {
            dfd.dfd_EnabledCompute =
                qtStringUtf8(enabledComputeStr_->text(), buffer);
        }
        if (weight_) {
            dfd.dfd_Weight = weight_->value();
        }
        if (fontname_) {
            dfd.dfd_Fontname = qtStringUtf8(fontname_->text(), buffer);
        }

        if (type == TYPE_BUTTON && dfd.dfd_SubType == DAD_BUTTONTYPE_IMAGE) {
            if (image_) {
                dfd.dfd_Background = qtStringUtf8(image_->spec(), buffer);
            }
        } else if (background_) {
            if (!defaultBackground_->isChecked()) {
                dfd.dfd_Background = qtStringUtf8(background_->spec(), buffer);
            }
        }
        if (imageSize_) {
            dfd.dfd_ImageSize = imageSize_->checkedId();
        }
        if (bubbleHelp_) {
            dfd.dfd_BubbleHelp =
                qtStringUtf8(bubbleHelp_->toPlainTextNbsp(), buffer);
        }
        dfd.dfd_File = file_ && file_->isChecked();
        dfd.dfd_Font = font_ && font_->isChecked();
        dfd.dfd_Listview = listview_ && listview_->isChecked();
        dfd.dfd_View = view_ && view_->isChecked();
        if (listviewLabels_) {
            dfd.dfd_ListviewLabels = listviewLabels_->labels();
        }
	if (listviewKind_ && listviewKind_->currentIndex() == 1) {
	    dfd.dfd_ListviewCompute =
                qtStringUtf8(listviewComputeStr_->text(), buffer);
	}
        dfd.dfd_DisplayExternal =
            displayExternal_ && displayExternal_->isChecked();
        dfd.dfd_TitleAtStringField =
            titleAtStringField_ && titleAtStringField_->isChecked();
        dfd.dfd_HideStringField =
            hideStringField_ && hideStringField_->isChecked();
        dfd.dfd_Show = show_ && show_->isChecked();
        dfd.dfd_AutoShow = autoShow_ && autoShow_->isChecked();
        dfd.dfd_Filter = filter_ && filter_->isChecked();
        dfd.dfd_Frame = frame_ && frame_->isChecked();
        if (horizontal_) {
            dfd.dfd_Alignment = horizontal_->isChecked() ?
                DAD_ALIGNMENT_HORIZONTAL : DAD_ALIGNMENT_VERTICAL;
        }
        if (recordsKind_) {
            dfd.dfd_RecordsKind = recordsKind_->currentIndex();
        }
        if (computeRecordsStr_) {
            dfd.dfd_RecordsCompute =
                qtStringUtf8(computeRecordsStr_->text(), buffer);
        }
    	dfd.dfd_InitialRecord = initialRecord_ && initialRecord_->isChecked();
    	dfd.dfd_CurrentRecord = currentRecord_ && currentRecord_->isChecked();

        if (recordDescriptionSpec_) {
            rds = recordDescriptionSpec_->getSpec(memPool);
            dfd.dfd_RecordDescriptionSpec = rds;
            if (!rds) {
                error = ERR_MEM;
            }
        }

        if (error == OK) {
            error = DisplayChangeField(field, &dfd);
            if (error == OK) {
                ProjectChange(pr_, true);
            }
        }

        FreeRecordDescriptionSpec(memPool, rds);

        if (error) {
            Ask2(this, GCS(MSG_ErrorMessage), AT_ERROR, 
                    GCS(MSG_StarOk), GES(error));
            return false;
        }

        DisplayFieldFinish(sew_, dio_);

        if (alignment_) {
            SetPrefsValue(pr_, (type == TYPE_INTEGER || type == TYPE_REAL)?
		    PREFS_DEFAULTALIGNMENT_NUMBER : 
		    PREFS_DEFAULTALIGNMENT_TEXT, 
                    dfd.dfd_Alignment);
        }
        return true;
    }

    bool eventFilter(QObject *obj, QEvent *event)
    {
        if (event->type() == QEvent::KeyPress) {
            QKeyEvent *keyEvent = static_cast<QKeyEvent *>(event);
            bool res = QObject::eventFilter(obj, event);

            if (keyEvent->key() == Qt::Key_Return) {
                CheckShortCut(shortcut_, dio_);
                return true;  // Always accept return
            } else {
                return res;
            }
        } else {
            // standard event processing
            return QObject::eventFilter(obj, event);
        }
    }

};

EditDisplayFieldDialog::EditDisplayFieldDialog(
        GuiObject sew, DisplayObject *dio) : CommonDialog(sew)
{
    setAttribute(Qt::WA_DeleteOnClose);

    pr_ = DisplayObjectProject(dio);
    dio_ = dio;
    sew_ = sew;
    title_ = 0;
    titlePosition_ = 0;
    titleFontname_ = 0;
    titleFontnamePopup_ = 0;
    shortcut_ = 0; 
    home_ = 0;
    tabChain_ = 0;
    readOnly_ = 0;
    alignment_ = 0;
    calendar_ = 0;
    format_ = 0;
    showTitles_ = 0;
    subType_ = 0; 
    subTypeStack_ = 0;
    frame_ = 0;
    horizontal_ = 0;
    controlTable_ = 0;
    onDoubleClick_ = 0;
    doubleClickTriggerStr_ = 0; 
    doubleClickTrigger_ = 0;
    urlDropTriggerStr_ = 0; 
    urlDropTrigger_ = 0;
    sortDropTriggerStr_ = 0; 
    sortDropTrigger_ = 0;
    numDecimals_ = 0;
    autoUpdate_ = 0;
    enabled_ = 0;
    enabledComputeStr_ = 0;
    enabledCompute_ = 0;
    weight_ = 0;
    bubbleHelp_ = 0;
    fontname_ = 0;
    fontnamePopup_ = 0;
    background_ = 0;
    defaultBackground_ = 0;
    image_ = 0;
    imageSize_ = 0;
    recordDescriptionSpec_ = 0;
    referencePopupGroup_ = 0;
    recordsKind_ = 0; 
    computeRecordsStr_ = 0;
    computeRecordsGroup_ = 0;
    initialRecord_ = 0;
    currentRecord_ = 0;
    show_ = 0;
    autoShow_ = 0;
    filter_ = 0;
    displayExternal_ = 0;
    hideStringField_ = 0;
    titleAtStringField_ = 0;
    listview_ = 0;
    file_ = 0;
    font_ = 0;
    view_ = 0;
    listviewSettings_ = 0;
    listviewKind_ = 0; 
    listviewKindStack_ = 0;
    listviewLabels_ = 0;
    listviewComputeStr_ = 0;

    Field *field = DisplayFieldField(dio);
    short type = FieldType(field);
    short subType = DisplayFieldSubType(dio);
    bool readOnly = DisplayFieldReadOnly(dio);
    const char *lvcompute = DisplayFieldListviewCompute(dio);

    bool show_shortcut = true, show_readonly = true,
        show_virtualtype = false, show_buttontype = false,
        show_choicetype = false, show_alignment = false, 
	show_calendar = false, show_format = false, show_numdecimals = false,
        show_weight = true, show_fontname = true, show_bg = true,
        show_stringbuttons = false, show_reference = false, extras = false;

    switch (type)
    {
        case TYPE_BOOL:
            show_fontname = false;
            break;

        case TYPE_CHOICE:
            show_choicetype = true;
            break;

        case TYPE_REAL:
            show_numdecimals = true;
            show_alignment = true;
            break;

        case TYPE_INTEGER:
            show_alignment = true;
            break;

        case TYPE_DATE:
            show_alignment = true;
            show_calendar = true;
            break;

        case TYPE_TIME:
            show_alignment = true;
	    show_format = true;
            break;

        case TYPE_STRING:
            show_alignment = true;
            show_stringbuttons = true;
            break;

        case TYPE_REFERENCE:
            show_alignment = true;
            show_reference = true;
            break;

        case TYPE_BUTTON:
            show_readonly = false;
            show_buttontype = true;
            break;

        case TYPE_VIRTUAL:
            show_readonly = false;
            show_shortcut = false;
            show_virtualtype = true;
            show_alignment = true;
            show_numdecimals = true;
            break;
    }
    extras = show_stringbuttons || show_reference || show_virtualtype;

    title_ = new QLineEdit(this);
    title_->setMaxLength(255);
    title_->setText(DisplayObjectTitle(dio));
    title_->setToolTip(GCS(MSG_DisplayField_TitleBH));
    QString titleLabel = GuiMarkShortcut(GCS(MSG_Title));

    QString titlePositionLabel;
    GroupLayout *titlePositionLayout = 0;
    if (!show_buttontype) {
        static char *titlepos[5];
        titlepos[0] = GCS(MSG_DisplayField_Alignment_Left);
        titlepos[1] = GCS(MSG_DisplayField_Alignment_Right);
        titlepos[2] = GCS(MSG_DisplayField_Alignment_Top);
        titlepos[3] = GCS(MSG_DisplayField_Alignment_Bottom);
        titlepos[4] = 0;
        titlePosition_ = new ComboBox(this, titlepos);
        titlePosition_->setCurrentIndex(DisplayFieldTitlePosition(dio));
        titlePosition_->setToolTip(GCS(MSG_DisplayField_TitlePositionBH));

        titlePositionLabel=GuiMarkShortcut(GCS(MSG_DisplayField_TitlePosition));

        titleFontname_ = new QLineEdit(this);
        titleFontname_->setText(DisplayFieldTitleQtFontname(dio));
        titleFontname_->setMaxLength(FONTNAME_LEN);
        titleFontname_->setToolTip(GCS(MSG_FontBH));
        titleFontnamePopup_ = MakeFontPopup(titleFontname_, this);

        titlePositionLayout = new GroupLayout(4);
        titlePositionLayout->addWidget(titlePosition_);
        titlePositionLayout->setStretch(1, 0);
        titlePositionLayout->addWidget(new QWidget(this));
        titlePositionLayout->addLabelAndWidget(GCS(MSG_Font),
                titleFontname_, titleFontnamePopup_);
        titlePositionLayout->setStretch(2, 0);
    }

    QString shortcutLabel;
    QHBoxLayout *shortcutLayout = 0;
    if (show_shortcut) {
	char sc[CodesetMaxBytesForChar+1];
	CodesetCharToStr(DisplayObjectShortcut(dio), sc);
        shortcut_ = new QLineEdit(this);
        shortcut_->setMaxLength(1);
        shortcut_->setText(sc);
        shortcut_->setToolTip(GCS(MSG_DisplayObject_ShortcutBH));
        shortcut_->setFixedWidth(5*shortcut_->fontMetrics().averageCharWidth());
        shortcut_->installEventFilter(this);

        shortcutLabel = GuiMarkShortcut(GCS(MSG_DisplayObject_Shortcut));

        shortcutLayout = new QHBoxLayout();
        shortcutLayout->addWidget(shortcut_);
        shortcutLayout->addStretch(1);

        home_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_DisplayField_Home)),this);
        home_->setChecked(DisplayFieldHome(dio));
        home_->setToolTip(GCS(MSG_DisplayField_HomeBH));
        shortcutLayout->addWidget(home_, 1);

        tabChain_ = new QCheckBox(
                GuiMarkShortcut(GCS(MSG_DisplayField_TabChain)), this);
        tabChain_->setChecked(DisplayFieldTabChain(dio));
        tabChain_->setToolTip(GCS(MSG_DisplayField_TabChainBH));
        shortcutLayout->addWidget(tabChain_, 1);

        if (show_readonly) {
            readOnly_ = new QCheckBox(
                    GuiMarkShortcut(GCS(MSG_DisplayField_ReadOnly)), this);
            readOnly_->setChecked(readOnly);
	    readOnly_->setToolTip(show_stringbuttons?
		    GCS(MSG_DisplayField_ReadOnlyStringBH) :
		    GCS(MSG_DisplayField_ReadOnlyBH));
            shortcutLayout->addWidget(readOnly_, 1);
        }
    }

    QString subTypeLabel;
    if (show_virtualtype) {
        static char *vttext[4];
        vttext[0] = GCS(MSG_Boolean);
        vttext[1] = GCS(MSG_Text);
        vttext[2] = GCS(MSG_DisplayField_Virtual_Kind_List);
        vttext[3] = 0;
        subType_ = new ComboBox(this, vttext);
        subType_->setCurrentIndex(subType);
        subType_->setToolTip(GCS(MSG_DisplayField_Virtual_KindBH));
        subType_->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed);

        subTypeLabel = GuiMarkShortcut(GCS(MSG_DisplayField_Kind));
    }

    if (show_buttontype) {
        static char *bttext[3];
        bttext[0] = GCS(MSG_DisplayField_Button_Kind_TextButton);
        bttext[1] = GCS(MSG_DisplayField_Button_Kind_ImageButton);
        bttext[2] = 0;
        subType_ = new ComboBox(this, bttext);
        subType_->setCurrentIndex(subType);
        subType_->setToolTip(GCS(MSG_DisplayField_Button_KindBH));

        subTypeLabel = GuiMarkShortcut(GCS(MSG_DisplayField_Kind));
    }

    if (show_choicetype) {
        static char *cttext[3];
        cttext[0] = GCS(MSG_DisplayField_Choice_Kind_CycleButton);
        cttext[1] = GCS(MSG_DisplayField_Choice_Kind_RadioButtons);
        cttext[2] = 0;
        subType_ = new ComboBox(this, cttext);
        subType_->setCurrentIndex(subType);
        subType_->setEnabled(!readOnly);
        subType_->setToolTip(GCS(MSG_DisplayField_Choice_KindBH));

        subTypeLabel = GuiMarkShortcut(GCS(MSG_DisplayField_Kind));

        QWidget *page0 = new QWidget(this);
        GroupLayout *pageGroup = new GroupLayout(2, page0);
        pageGroup->setContentsMargins(0, 0, 0, 0);
        pageGroup->addLabelAndWidget(titlePositionLabel, titlePosition_,
                titlePositionLayout);
        
        frame_ = new QCheckBox(
                GuiMarkShortcut(GCS(MSG_DisplayField_Choice_Frame)), this);
        frame_->setChecked(DisplayFieldFrame(dio));
        frame_->setToolTip(GCS(MSG_DisplayField_Choice_FrameBH));

        horizontal_ = new QCheckBox(
                GuiMarkShortcut(GCS(MSG_DisplayField_Choice_Horizontal)), this);
        horizontal_->setChecked(
                DisplayFieldAlignment(dio) == DAD_ALIGNMENT_HORIZONTAL);
        horizontal_->setToolTip(GCS(MSG_DisplayField_Choice_HorizontalBH));

        QWidget *page1 = new QWidget(this);
        QHBoxLayout *pageHBox = new QHBoxLayout(page1);
        pageHBox->setContentsMargins(0, 0, 0, 0);
        pageHBox->addWidget(frame_);
        pageHBox->addWidget(horizontal_);

        subTypeStack_ = new QStackedWidget(this);
        subTypeStack_->setSizePolicy(QSizePolicy::Preferred,QSizePolicy::Fixed);
        subTypeStack_->addWidget(page0);
        subTypeStack_->addWidget(page1);
    }

    QString alignmentLabel;
    if (show_alignment) {
        static char *alignmentText[4];
	alignmentText[0] = GCS(MSG_DisplayField_Alignment_Center);
	alignmentText[1] = GCS(MSG_DisplayField_Alignment_Left);
	alignmentText[2] = GCS(MSG_DisplayField_Alignment_Right);
	alignmentText[3] = 0;
        alignment_ = new ComboBox(this, alignmentText);
        alignment_->setCurrentIndex(DisplayFieldAlignment(dio));
        alignment_->setToolTip(GCS(MSG_DisplayField_AlignmentBH));
        alignmentLabel = GuiMarkShortcut(GCS(MSG_DisplayField_Alignment));
    }

    if (show_calendar) {
        calendar_ = new QCheckBox(
                GuiMarkShortcut(GCS(MSG_DisplayField_Calendar)), this);
        calendar_->setChecked(DisplayFieldCalendar(dio));
        calendar_->setToolTip(GCS(MSG_DisplayField_CalendarBH));
    }

    QString formatLabel;
    if (show_format) {
        static char *formatText[4];
	formatText[0] = GCS(MSG_DisplayField_Format_HHMMSS);
	formatText[1] = GCS(MSG_DisplayField_Format_MMSS);
	formatText[2] = GCS(MSG_DisplayField_Format_HHMM);
	formatText[3] = 0;
        format_ = new ComboBox(this, formatText);
        format_->setCurrentIndex(DisplayFieldFormat(dio));
        format_->setToolTip(GCS(MSG_DisplayField_FormatBH));
        formatLabel = GuiMarkShortcut(GCS(MSG_DisplayField_Format));
    }

    QString numDecimalsLabel;
    if (show_numdecimals) {
        numDecimals_ = new QSpinBox(this);
        numDecimals_->setMinimum(0);
        numDecimals_->setMaximum(9);
        numDecimals_->setValue(DisplayFieldNumDecimals(dio));
        numDecimals_->setToolTip(GCS(MSG_DisplayField_NumDecimalsBH));

	numDecimalsLabel = GuiMarkShortcut(GCS(MSG_DisplayField_NumDecimals));
    }

    QLayout *extrasLayout = 0;
    if (show_virtualtype) {
        showTitles_ = new QCheckBox(
                GuiMarkShortcut(GCS(MSG_DisplayField_ShowTitles)), this);
        showTitles_->setChecked(DisplayFieldShowTitles(dio));
        showTitles_->setToolTip(GCS(MSG_DisplayField_ShowTitlesBH));

        tabChain_ = new QCheckBox(
                GuiMarkShortcut(GCS(MSG_DisplayField_TabChain)), this);
        tabChain_->setChecked(DisplayFieldTabChain(dio));
        tabChain_->setToolTip(GCS(MSG_DisplayField_TabChainBH));

        controlTable_ = new QCheckBox(
                GuiMarkShortcut(GCS(MSG_DisplayField_ControlTable)), this);
        controlTable_->setChecked(DisplayFieldControlTable(dio));
        controlTable_->setToolTip(GCS(MSG_DisplayField_ControlTableBH));

	static char *doubleClickText[4];
	doubleClickText[0] = GCS(MSG_DisplayField_DoubleClick_DoNothing);
	doubleClickText[1] = GCS(MSG_DisplayField_DoubleClick_ShowRecord);
	doubleClickText[2] = GCS(MSG_DisplayField_DoubleClick_CallTrigger);
	doubleClickText[3] = 0;
	onDoubleClick_ = new RadioBox(this, doubleClickText);
        onDoubleClick_->setToolTip(GCS(MSG_DisplayField_OnDoubleClickBH));

	doubleClickTriggerStr_ = new QLineEdit(this);
	doubleClickTriggerStr_->setMaxLength(FIELD_TRIGGER_LEN);
	doubleClickTriggerStr_->setText(
                DisplayFieldDoubleClickTrigger(dio));
	doubleClickTriggerStr_->setToolTip(
                GCS(MSG_DisplayField_DoubleClick_TriggerBH));

        doubleClickTrigger_ = UserFunctionsPopup(
                doubleClickTriggerStr_, pr_, this);

        QString dctLabel(GuiMarkShortcut(
                    GCS(MSG_DisplayField_DoubleClick_Trigger)));

        connect(onDoubleClick_, SIGNAL(checked(int)),
                this, SLOT(onDoubleClickChanged(int)));
        onDoubleClick_->setCheckedId(DisplayFieldOnDoubleClick(dio));

	urlDropTriggerStr_ = new QLineEdit(this);
	urlDropTriggerStr_->setMaxLength(FIELD_TRIGGER_LEN);
	urlDropTriggerStr_->setText(DisplayFieldURLDropTrigger(dio));
	urlDropTriggerStr_->setToolTip(GCS(MSG_DisplayField_URLDrop_TriggerBH));

        urlDropTrigger_ = UserFunctionsPopup(urlDropTriggerStr_, pr_, this);

        QString udtLabel(GuiMarkShortcut(
                    GCS(MSG_DisplayField_URLDrop_Trigger)));

	sortDropTriggerStr_ = new QLineEdit(this);
	sortDropTriggerStr_->setMaxLength(FIELD_TRIGGER_LEN);
	sortDropTriggerStr_->setText(DisplayFieldSortDropTrigger(dio));
	sortDropTriggerStr_->setToolTip(
                GCS(MSG_DisplayField_SortDrop_TriggerBH));

        sortDropTrigger_ = UserFunctionsPopup(sortDropTriggerStr_, pr_, this);

        QString sdtLabel(GuiMarkShortcut(
                    GCS(MSG_DisplayField_SortDrop_Trigger)));

        QWidget *page0 = new QWidget(this);

        GroupLayout *pageGroup = new GroupLayout(2);
        QHBoxLayout *alignmentLayout = new QHBoxLayout();
        alignmentLayout->addWidget(alignment_, 1);
        alignmentLayout->addStretch(1);
        pageGroup->addLabelAndWidget(alignmentLabel,alignment_,alignmentLayout);
        pageGroup->addLabelAndWidget(numDecimalsLabel, numDecimals_);
        show_alignment = false;

        QWidget *page1 = new QWidget(this);
        QVBoxLayout *vbox = new QVBoxLayout(page1);
        vbox->addStretch();
        vbox->addLayout(pageGroup);
        vbox->addStretch();

        QFrame *vSeparator = new QFrame(this);
        vSeparator->setFrameShape(QFrame::VLine);
        vSeparator->setFrameShadow(QFrame::Raised);

        GroupLayout *doubleClickTriggerGroup = new GroupLayout(1);
        doubleClickTriggerGroup->addWidget(new QWidget(this));
        doubleClickTriggerGroup->addLabelAndWidget(
            dctLabel, doubleClickTriggerStr_, doubleClickTrigger_);

        GroupBox *onDoubleClickGroup = new GroupBox(
            GCS(MSG_DisplayField_OnDoubleClick), this);
        QHBoxLayout *hbox = new QHBoxLayout(onDoubleClickGroup);
        hbox->addWidget(onDoubleClick_);
        hbox->addWidget(vSeparator);
        hbox->addLayout(doubleClickTriggerGroup);

        GroupLayout *dragDropTriggerGroup = new GroupLayout(2);
        dragDropTriggerGroup->addLabelAndWidget(
                udtLabel, urlDropTriggerStr_, urlDropTrigger_);
        dragDropTriggerGroup->addLabelAndWidget(
                sdtLabel, sortDropTriggerStr_, sortDropTrigger_);

        QWidget *page2 = new QWidget(this);
        QVBoxLayout *pageVBox = new QVBoxLayout(page2);
        pageVBox->addStretch(1);
        pageVBox->addWidget(showTitles_);
        pageVBox->addWidget(tabChain_);
        pageVBox->addWidget(controlTable_);
        pageVBox->addSpacing(20);
        pageVBox->addWidget(onDoubleClickGroup);
        pageVBox->addLayout(dragDropTriggerGroup);
        pageVBox->addStretch(1);

        subTypeStack_ = new QStackedWidget(this);
        subTypeStack_->setFrameShape(QFrame::StyledPanel);
        subTypeStack_->setFrameShadow(QFrame::Raised);
        subTypeStack_->addWidget(page0);
        subTypeStack_->addWidget(page1);
        subTypeStack_->addWidget(page2);

        autoUpdate_ = new QCheckBox(
                GuiMarkShortcut(GCS(MSG_DisplayField_AutoUpdate)), this);
        autoUpdate_->setChecked(DisplayFieldAutoUpdate(dio));
        autoUpdate_->setToolTip(GCS(MSG_DisplayField_AutoUpdateBH));

        GroupLayout *extrasGroup = new GroupLayout(2);
        extrasGroup->addLabelAndWidget(subTypeLabel, subType_);
        extrasGroup->setStretch(1, 0);
        extrasGroup->addWidget(subTypeStack_, 2);
        extrasGroup->addWidget(autoUpdate_, 2);
        extrasLayout = extrasGroup;
    }

    GroupBox *enabledGroup = 0;
    static char *etext[4];
    etext[0] = GCS(MSG_DisplayObject_Enabled_Always);
    etext[1] = GCS(MSG_DisplayObject_Enabled_NotInitial);
    etext[2] = GCS(MSG_DisplayObject_Enabled_Computed);
    etext[3] = 0;
    enabled_ = new RadioBox(this, etext);
    enabled_->setToolTip(GCS(MSG_DisplayObject_EnabledBH));

    enabledComputeStr_ = new QLineEdit(this);
    enabledComputeStr_->setMaxLength(FIELD_TRIGGER_LEN);
    enabledComputeStr_->setText(DisplayObjectEnabledCompute(dio));
    enabledComputeStr_->setToolTip(GCS(MSG_DisplayObject_ComputeEnabledBH));

    enabledCompute_ = UserFunctionsPopup(enabledComputeStr_, pr_, this);
    enabledCompute_->setEnabled(false);

    QString ceLabel = GuiMarkShortcut(GCS(MSG_Compute));

    connect(enabled_, SIGNAL(checked(int)),
            this, SLOT(onEnabledChanged(int)));
    char enabled = DisplayObjectEnabled(dio);
    enabled_->setCheckedId(enabled);
    onEnabledChanged(enabled);

    QFrame *vSeparator = new QFrame(this);
    vSeparator->setFrameShape(QFrame::VLine);
    vSeparator->setFrameShadow(QFrame::Raised);

    GroupLayout *vGroup = new GroupLayout(1);
    vGroup->addWidget(new QWidget(this));
    vGroup->addLabelAndWidget(ceLabel, enabledComputeStr_, enabledCompute_);

    enabledGroup = new GroupBox(GCS(MSG_DisplayObject_Enabled), this);
    GroupLayout *group = new GroupLayout(3, enabledGroup);
    group->addWidget(enabled_);
    group->addWidget(vSeparator);
    group->addLayout(vGroup);

    QString weightLabel;
    QHBoxLayout *weightBox = 0;
    if (show_weight) {
        weight_ = new QSlider(this);
        weightBox = WeightSlider(weight_, this);
        weight_->setValue(DisplayObjectWeight(dio));
	weightLabel = GuiMarkShortcut(GCS(MSG_Weight));
    }

    QString fontnameLabel;
    if (show_fontname) {
        fontname_ = new QLineEdit(this);
        fontname_->setText(DisplayObjectQtFontname(dio));
        fontname_->setMaxLength(FONTNAME_LEN);
        fontname_->setToolTip(GCS(MSG_FontBH));
        fontnamePopup_ = MakeFontPopup(fontname_, this);
        fontnameLabel = GuiMarkShortcut(GCS(MSG_Font));
    }

    QString bgLabel;
    QHBoxLayout *bgBox = 0;
    if (show_bg) {
        const char *bg = DisplayObjectBackground(dio);
        if (show_buttontype && subType != DAD_BUTTONTYPE_TEXT) {
            bg = 0;
        }
        background_ = new PopImage(this, pr_);
        background_->setTypes(PopImage::Color);
        background_->setSpec(bg);
        background_->setToolTip(GCS(MSG_BackgroundBH));
        defaultBackground_ = new QCheckBox(this);
        bgBox = BackgroundGadgets(background_, defaultBackground_, bg);
        bgLabel = GuiMarkShortcut(GCS(MSG_Background));
    }

    if (show_buttontype) {
        const char *bg = DisplayObjectBackground(dio);
        if (subType != DAD_BUTTONTYPE_IMAGE) {
            bg = 0;
        }

        image_ = new PopImage(this, pr_);
        image_->setSpec(bg);
        image_->setToolTip(GCS(MSG_ImageBH));

        static char *entries[3];
        entries[0] = GCS(MSG_Image_Size_Resizeable);
        entries[1] = GCS(MSG_Image_Size_Fixed);
        entries[2] = 0;
        imageSize_ = new RadioBox(this, entries, GCS(MSG_Size), true);
        imageSize_->setCheckedId(DisplayFieldImageSize(dio));

        QWidget *page0 = new QWidget(this);
        GroupLayout *pageGroup = new GroupLayout(2, page0);
        pageGroup->addLabelAndWidget(titleLabel, title_);
        pageGroup->addLabelAndWidget(fontnameLabel, fontname_, fontnamePopup_);
        pageGroup->addLabelAndWidget(bgLabel, background_, bgBox);

        show_fontname = false;          // only add once
        show_bg = false;                // only add once

        QWidget *page1 = new QWidget(this);
        pageGroup = new GroupLayout(3, page1);
        pageGroup->addLabelAndWidget(GuiMarkShortcut(GCS(MSG__Image)), image_);
        pageGroup->setStretch(1, 1);
        QVBoxLayout *vbox = new QVBoxLayout();
        pageGroup->addLayout(vbox);
        vbox->addStretch(1);
        vbox->addWidget(imageSize_);
        vbox->addStretch(1);

        subTypeStack_ = new QStackedWidget(this);
        subTypeStack_->setFrameShape(QFrame::StyledPanel);
        subTypeStack_->setFrameShadow(QFrame::Raised);
        subTypeStack_->addWidget(page0);
        subTypeStack_->addWidget(page1);
    }

    if (show_stringbuttons) {
        file_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_DisplayField_File)),this);
        file_->setChecked(DisplayFieldFile(dio));
        file_->setToolTip(GCS(MSG_DisplayField_FileBH));
        connect(file_, SIGNAL(toggled(bool)), this, SLOT(onFileToggled(bool)));

        font_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_DisplayField_Font)),this);
        font_->setChecked(DisplayFieldFont(dio));
        font_->setToolTip(GCS(MSG_DisplayField_FontBH));
        connect(font_, SIGNAL(toggled(bool)), this, SLOT(onFontToggled(bool)));

        listview_ = new QCheckBox(
            GuiMarkShortcut(GCS(MSG_DisplayField_Listview)), this);
        listview_->setChecked(DisplayFieldListview(dio));
        listview_->setToolTip(GCS(MSG_DisplayField_ListviewBH));
        connect(listview_, SIGNAL(toggled(bool)),
            this, SLOT(onListviewToggled(bool)));

        view_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_DisplayField_View)),this);
        view_->setChecked(DisplayFieldView(dio));
        view_->setToolTip(GCS(MSG_DisplayField_ViewBH));

        QBoxLayout *popBox = new QVBoxLayout();
        popBox->addStretch();
        popBox->addWidget(file_);
        popBox->addWidget(font_);
        popBox->addWidget(listview_);
        popBox->addWidget(view_);
        popBox->addStretch();

	static char *lktext[3];
        lktext[0] = GCS(MSG_DisplayField_Labels_Static);
        lktext[1] = GCS(MSG_Computed);
        lktext[2] = 0;
        listviewKind_ = new ComboBox(this, lktext);
        listviewKind_->setToolTip(GCS(MSG_DisplayField_LabelsBH));

        listviewLabels_ = new PopEditLabels(this);
        listviewLabels_->setText(
                GuiMarkShortcut(GCS(MSG_DisplayField__EditLabels)));
        listviewLabels_->setLabels(DisplayFieldListviewLabels(dio));
        listviewLabels_->setToolTip(GCS(MSG_DisplayField_EditLabelsBH));

        QWidget *listviewLabelsGroup = new QWidget(this);
        QVBoxLayout *vbox = new QVBoxLayout(listviewLabelsGroup);
        vbox->addStretch();
        vbox->addWidget(listviewLabels_);
        vbox->addStretch();

	listviewComputeStr_ = new QLineEdit(this);
	listviewComputeStr_->setMaxLength(FIELD_TRIGGER_LEN);
	listviewComputeStr_->setText(lvcompute);
        listviewComputeStr_->setToolTip(GCS(MSG_DisplayField_ComputeLabelsBH));

        QWidget *listviewCompute = new QWidget(this);
        GroupLayout *grp = new GroupLayout(1, listviewCompute);
        grp->addWidget(new QWidget(this));
        grp->setStretch(1, 1);
        grp->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Compute)),
                listviewComputeStr_,
                UserFunctionsPopup(listviewComputeStr_, pr_, this));
        grp->addWidget(new QWidget(this));
        grp->setStretch(1, 1);

	listviewKindStack_ = new QStackedWidget(this);
	listviewKindStack_->addWidget(listviewLabelsGroup);
	listviewKindStack_->addWidget(listviewCompute);

        QFrame *hSeparator = new QFrame(this);
        hSeparator->setFrameShape(QFrame::HLine);
        hSeparator->setFrameShadow(QFrame::Raised);

	listviewSettings_ = new GroupBox(
                GCS(MSG_DisplayField_ListviewSettings), this);
	listviewSettings_->setEnabled(DisplayFieldListview(dio));
        grp = new GroupLayout(2, listviewSettings_);
        grp->addLabelAndWidget(
                GuiMarkShortcut(GCS(MSG_DisplayField_Labels)), listviewKind_);
        grp->addWidget(hSeparator, 2);
        grp->addWidget(listviewKindStack_, 2);

        connect(listviewKind_, SIGNAL(currentIndexChanged(int)),
                listviewKindStack_, SLOT(setCurrentIndex(int)));

        GroupBox *bottom = new GroupBox(GCS(MSG_DisplayField_Buttons), this);
        QHBoxLayout *hbox = new QHBoxLayout(bottom);
        hbox->addLayout(popBox);
        hbox->addWidget(listviewSettings_);

        displayExternal_ = new QCheckBox(
            GuiMarkShortcut(GCS(MSG_DisplayField_DisplayPicture)), this);
        displayExternal_->setChecked(true);
        displayExternal_->setToolTip(GCS(MSG_DisplayField_DisplayPictureBH));

        hideStringField_ = new QCheckBox(
            GuiMarkShortcut(GCS(MSG_DisplayField_HideStringField)), this);
        hideStringField_->setToolTip(GCS(MSG_DisplayField_HideStringFieldBH));
        connect(hideStringField_, SIGNAL(toggled(bool)),
            this, SLOT(onHideStringFieldChanged(bool)));

        titleAtStringField_ = new QCheckBox(
                GuiMarkShortcut(GCS(MSG_DisplayField_TitleAtStringField)),this);
        titleAtStringField_->setToolTip(
                GCS(MSG_DisplayField_TitleAtStringFieldBH));

        static char *entries[6];
        entries[0] = GCS(MSG_Image_Size_Resizeable);
        entries[1] = GCS(MSG_Image_Size_Fixed);
        entries[2] = GCS(MSG_Image_Size_Scrollable);
        entries[3] = GCS(MSG_Image_Size_Scaled);
        entries[4] = GCS(MSG_Image_Size_AspectScaled);
        entries[5] = 0;
        imageSize_ = new RadioBox(this, entries, GCS(MSG_Size), true);
        imageSize_->setCheckedId(DisplayFieldImageSize(dio));
        imageSize_->setToolTip(GCS(MSG_DisplayField_SizeBH));

        QVBoxLayout *flagsBox = new QVBoxLayout();
        flagsBox->addStretch();
        flagsBox->addWidget(displayExternal_);
        flagsBox->addWidget(titleAtStringField_);
        flagsBox->addWidget(hideStringField_);
        flagsBox->addStretch();

        connect(displayExternal_, SIGNAL(toggled(bool)),
                this, SLOT(onDisplayExternalChanged(bool)));

        titleAtStringField_->setChecked(
                DisplayFieldTitleAtStringField(dio));
        hideStringField_->setChecked(DisplayFieldHideStringField(dio));
        displayExternal_->setChecked(DisplayFieldDisplayExternal(dio));

        GroupBox *top = new GroupBox(GCS(MSG_DisplayField_Picture), this);
        hbox = new QHBoxLayout(top);
        hbox->addLayout(flagsBox);
        hbox->addWidget(imageSize_);

        QVBoxLayout *extrasVBox = new QVBoxLayout();
        extrasVBox->addWidget(top);
        extrasVBox->addWidget(bottom);
        extrasLayout = extrasVBox;
    } else if (show_reference) {
        show_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_DisplayField_Show)),this);
        show_->setChecked(DisplayFieldShow(dio));
        show_->setToolTip(GCS(MSG_DisplayField_ShowBH));
        connect(show_, SIGNAL(toggled(bool)), this, SLOT(updateFocusWidgets()));
            
        autoShow_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_AutoShow)), this);
        autoShow_->setChecked(DisplayFieldAutoShow(dio));
        autoShow_->setToolTip(GCS(MSG_DisplayField_AutoShowBH));

        filter_ = new QCheckBox(
            GuiMarkShortcut(GCS(MSG_DisplayField_Filter)), this);
        filter_->setChecked(DisplayFieldFilter(dio));
        filter_->setToolTip(GCS(MSG_DisplayField_FilterBH));

        GroupBox *buttons = new GroupBox(GCS(MSG_DisplayField_Buttons), this);
        QVBoxLayout *vbox = new QVBoxLayout(buttons);
        vbox->addStretch();
        vbox->addWidget(show_);
        vbox->addWidget(autoShow_);
        vbox->addWidget(filter_);
        vbox->addStretch();

        recordDescriptionSpec_ = new EditRecordDescriptionSpec(
                FieldReferenceTable(field), 
                DisplayFieldRecordDescriptionSpec(dio, false), true, this);

	GroupBox *recordDescriptionSpecGroup = new GroupBox(
                GCS(MSG_Display), this);
        vbox = new QVBoxLayout(recordDescriptionSpecGroup);
        vbox->addWidget(recordDescriptionSpec_);

        static char *rectext[4];
        rectext[0] = GCS(MSG_DisplayField_Records_All);
        rectext[1] = GCS(MSG_DisplayField_Records_MatchFilter);
        rectext[2] = GCS(MSG_Computed);
	rectext[3] = 0;
	recordsKind_ = new ComboBox(this, rectext);
        recordsKind_->setToolTip(GCS(MSG_DisplayField_RecordsBH));
        QString recordsLabel = GuiMarkShortcut(GCS(MSG_Records));

	computeRecordsStr_ = new QLineEdit(this);
	computeRecordsStr_->setMaxLength(FIELD_TRIGGER_LEN);
	computeRecordsStr_->setText(DisplayFieldRecordsCompute(dio));
        computeRecordsStr_->setToolTip(GCS(MSG_DisplayField_ComputeRecordsBH));

        QLayout *computeRecords = UserFunctionsPopup(
                computeRecordsStr_, pr_, this);

	computeRecordsGroup_ = new GroupBox(GCS(MSG_Compute), this);
	computeRecordsGroup_->setEnabled(false);
        vbox = new QVBoxLayout(computeRecordsGroup_);
        vbox->addLayout(computeRecords);

        connect(recordsKind_, SIGNAL(currentIndexChanged(int)),
                this, SLOT(onRecordsChanged(int)));
	recordsKind_->setCurrentIndex(DisplayFieldRecordsKind(dio));

	initialRecord_ = new QCheckBox(
                GuiMarkShortcut(GCS(MSG_DisplayField_InitialRecord)), this);
        initialRecord_->setChecked(DisplayFieldInitialRecord(dio));
        initialRecord_->setToolTip(GCS(MSG_DisplayField_InitialRecordBH));

	currentRecord_ = new QCheckBox(
                GuiMarkShortcut(GCS(MSG_DisplayField_CurrentRecord)), this);
	currentRecord_->setChecked(DisplayFieldCurrentRecord(dio));
	currentRecord_->setToolTip(GCS(MSG_DisplayField_CurrentRecordBH));

	referencePopupGroup_ = new GroupBox(GCS(MSG_Popup), this);
        GroupLayout *grp = new GroupLayout(2, referencePopupGroup_);
        grp->addLabelAndWidget(recordsLabel, recordsKind_);
        grp->addWidget(computeRecordsGroup_, 2);
        grp->addWidget(initialRecord_, 2);
        grp->addWidget(currentRecord_, 2);

        vbox = new QVBoxLayout();
        vbox->addWidget(referencePopupGroup_, 5);
        vbox->addWidget(buttons, 3);

        QHBoxLayout *extrasHBox = new QHBoxLayout();
        extrasHBox->addWidget(recordDescriptionSpecGroup);
        extrasHBox->addLayout(vbox);
        extrasLayout = extrasHBox;
    }

    QHBoxLayout *alignmentLayout = 0;
    if (show_alignment) {
        GroupLayout *alignmentGroup = 0;
        if (show_numdecimals) {
            alignmentGroup = new GroupLayout(2);
            alignmentGroup->addLabelAndWidget(numDecimalsLabel, numDecimals_);
        } else if (show_calendar) {
            alignmentGroup = new GroupLayout(2);
            alignmentGroup->addWidget(new QWidget(this));
            alignmentGroup->addWidget(calendar_);
        } else if (show_format) {
            alignmentGroup = new GroupLayout(2);
            alignmentGroup->addLabelAndWidget(formatLabel, format_);
        }
        alignmentLayout = new QHBoxLayout();
        if (alignmentGroup) {
            alignmentLayout->addWidget(alignment_);
            alignmentLayout->addLayout(alignmentGroup);
        } else {
            alignmentLayout->addWidget(alignment_, 1);
            alignmentLayout->addStretch(1);
        }
    }

    GroupLayout *topGroup = new GroupLayout(2);
    if (show_buttontype) {
        topGroup->addLabelAndWidget(subTypeLabel, subType_);
        topGroup->setStretch(1, 0);
        topGroup->addWidget(subTypeStack_, 2);
    } else if (show_choicetype) {
        topGroup->addLabelAndWidget(subTypeLabel, subType_);
        topGroup->setStretch(1, 0);
        QGroupBox *titleBox = new GroupBox(GCS(MSG_DisplayField_Title), this);
        topGroup->addWidget(titleBox, 2);
        GroupLayout *grp = new GroupLayout(2, titleBox);
        grp->addLabelAndWidget(titleLabel, title_);
        grp->addWidget(subTypeStack_, 2);
    } else {
        QGroupBox *titleBox = new GroupBox(GCS(MSG_DisplayField_Title), this);
        topGroup->addWidget(titleBox, 2);
        GroupLayout *grp = new GroupLayout(2, titleBox);
        grp->addLabelAndWidget(titleLabel, title_);
        grp->addLabelAndWidget(titlePositionLabel, titlePosition_,
                titlePositionLayout);
    }

    QGroupBox *fieldBox = new QGroupBox(GCS(MSG_Field), this);
    GroupLayout *fieldGroup = new GroupLayout(2, fieldBox);
    if (show_shortcut) {
        fieldGroup->addLabelAndWidget(shortcutLabel,shortcut_,shortcutLayout);
    }
    if (show_alignment) {
        fieldGroup->addLabelAndWidget(alignmentLabel, alignment_,
                alignmentLayout);
    }
    if (show_weight) {
        fieldGroup->addLabelAndWidget(weightLabel, weight_, weightBox);
    }
    if (show_fontname) {
        fieldGroup->addLabelAndWidget(fontnameLabel, fontname_, fontnamePopup_);
    }
    if (show_bg) {
        fieldGroup->addLabelAndWidget(bgLabel, background_, bgBox);
        fieldGroup->setStretch(1, 1);
    }

    bubbleHelp_ = new TextEditor(pr_, false, this);
    bubbleHelp_->setText(DisplayFieldBubbleHelp(dio));
    fieldGroup->addLabelAndWidget(
            GuiMarkShortcut(GCS(MSG_DisplayField_BubbleHelp)), bubbleHelp_);
    fieldGroup->setStretch(1, 2);

    GroupLayout *attrsVBox = new GroupLayout(1);
    attrsVBox->addLayout(topGroup);
    if (show_buttontype) {
        attrsVBox->setStretch(1, 1);
    }
    attrsVBox->addWidget(fieldBox);
    attrsVBox->setStretch(1, 1);
    attrsVBox->addWidget(enabledGroup);

    QLayout *layout = attrsVBox;
    if (extras) {
        QTabWidget *tab = new QTabWidget(this);

        QWidget *general = new QWidget(this);
        general->setLayout(attrsVBox);
        tab->addTab(general, GuiMarkShortcut(GCS(MSG_General)));

        QWidget *extras = new QWidget(this);
        extras->setLayout(extrasLayout);
        tab->addTab(extras, GuiMarkShortcut(GCS(MSG_DisplayField_Extras)));

        layout = new GroupLayout(1);
        layout->addWidget(tab);
    }

    contentsLayout()->addLayout(layout);

    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

    setWindowTitle(GCS(MSG_DisplayField));
    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Field object editor");

    if (readOnly_) {
        connect(readOnly_, SIGNAL(toggled(bool)),
                this, SLOT(onReadOnlyToggled(bool)));
        onReadOnlyToggled(readOnly_->isChecked());
    }
    if (subType_ && subTypeStack_) {
        connect(subType_, SIGNAL(currentIndexChanged(int)),
                subTypeStack_, SLOT(setCurrentIndex(int)));
        subTypeStack_->setCurrentIndex(subType);
    }
    if (listviewKind_ && lvcompute && lvcompute[0]) {
        listviewKind_->setCurrentIndex(1);
    }

    updateFocusWidgets();
}

void GuiDisplayField(GuiObject sew, DisplayObject *dio)
{
    EditDisplayFieldDialog *dialog =
        new EditDisplayFieldDialog(sew, dio);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }
}

#include "qtEditDisplayField.moc"

#endif

