//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtEditDisplayTable.cc
//
// Creation date: 15.03.2016
// Last modified: 19.04.2016
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "qt/qtEditDisplayTable.h"
#include "qt/qtEditRecordDescriptionSpec.h"
#include "qt/qtStructureEditor.h"
#include "qt/qtStructureEditorUtil.h"
#include "qt/classes/qtGroupLayout.h"
#include "qt/classes/qtPopImage.h"
#include "qt/qtUtil.h"
#include <QCheckBox>
#include "display.h"
#include "requesters.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"

class EditDisplayTableDialog : public CommonDialog
{
    Q_OBJECT

public:
    EditDisplayTableDialog(GuiObject sew, DisplayObject *dio);

private:
    Project *pr_;
    DisplayObject *dio_;
    GuiObject sew_;
    QTabWidget *tab_;
    QSlider *weight_;
    PopImage *background_;
    QCheckBox *defaultBackground_;
    QCheckBox *hasPanel_;
    QLineEdit *panelTitle_;
    QLineEdit *panelFontName_;
    PopImage *panelBackground_;
    QCheckBox *panelDefaultBackground_;
    QCheckBox *panelNumAll_;
    QCheckBox *panelFilter_;
    QCheckBox *panelArrows_;
    EditRecordDescriptionSpec *recordDescriptionSpec_;

    virtual bool acceptEdits()
    {
	GuiBusyCursor busy;

        long error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return false;
        }

        Table *tab = DisplayObjectTable(dio_);
        MemPool *memPool = ProjectMemPool(pr_);
        RecordDescriptionSpec *rds = 0;

        Utf8Buffer buffer;
	DisplayTableData dtd;
        dtd.dtd_Title = qtStringUtf8(panelTitle_->text(), buffer);
        dtd.dtd_Weight = weight_->value();
        dtd.dtd_Fontname = qtStringUtf8(panelFontName_->text(), buffer);
	dtd.dtd_Background = qtStringUtf8(background_->spec(), buffer);
        dtd.dtd_HasPanel = hasPanel_->isChecked();
	dtd.dtd_PanelBackground = qtStringUtf8(panelBackground_->spec(),buffer);
	dtd.dtd_PanelNumAll = panelNumAll_->isChecked();
	dtd.dtd_PanelFilter = panelFilter_->isChecked();
	dtd.dtd_PanelArrows = panelArrows_->isChecked();
        rds = recordDescriptionSpec_->getSpec(memPool);
        if (rds == NULL) {
            error = ERR_MEM;
        }
        dtd.dtd_RecordDescriptionSpec = rds;

        if (error == OK) {
            error = DisplayChangeTable(tab, &dtd);
            if (error == OK) {
                ProjectChange(pr_, true);
            }
        }

        FreeRecordDescriptionSpec(memPool, rds);

        if (error) {
            Ask2(this, GCS(MSG_ErrorMessage), AT_ERROR,
                 GCS(MSG_StarOk), GES(error));
            return false;
        }

        DisplayTableFinish(sew_, dio_);
        return true;
    }

private slots:
    void onHasPanelToggled(bool checked)
    {
        tab_->setTabEnabled(1, checked);
    }
};

EditDisplayTableDialog::EditDisplayTableDialog(GuiObject sew, DisplayObject *dio)
    : CommonDialog(sew)
{
    bool hasPanel = DisplayTableHasPanel(dio);

    setAttribute(Qt::WA_DeleteOnClose);

    pr_ = DisplayObjectProject(dio);
    dio_ = dio;
    sew_ = sew;

    tab_ = new QTabWidget(this);
    contentsLayout()->addWidget(tab_);

    // General
    QWidget *general = new QWidget(this);
    tab_->addTab(general, GuiMarkShortcut(GCS(MSG_General)));
    GroupLayout *group = new GroupLayout(2, general);

    weight_ = new QSlider(general);
    QHBoxLayout *hbox = WeightSlider(weight_, general);
    weight_->setValue(DisplayObjectWeight(dio));
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Weight)), weight_, hbox);

    background_ = new PopImage(general, pr_);
    defaultBackground_ = new QCheckBox(general);
    hbox = BackgroundGadgets(background_, defaultBackground_,
                             DisplayObjectBackground(dio));
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Background)), background_, hbox);

    hasPanel_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_DisplayTable_HasPanel)), general);
    hasPanel_->setChecked(hasPanel);
    hasPanel_->setToolTip(GCS(MSG_DisplayTable_HasPanelBH));
    group->addWidget(hasPanel_, 2);
    connect(hasPanel_, SIGNAL(toggled(bool)), this, SLOT(onHasPanelToggled(bool)));

    // Panel
    QWidget *panel = new QWidget(this);
    tab_->addTab(panel, GuiMarkShortcut(GCS(MSG_Panel)));
    tab_->setTabEnabled(1, hasPanel);
    group = new GroupLayout(2, panel);

    panelTitle_ = new QLineEdit(panel);
    panelTitle_->setMaxLength(TITLE_LEN);
    panelTitle_->setText(DisplayObjectTitle(dio)),
    panelTitle_->setToolTip(GCS(MSG_DisplayTable_Panel_TitleBH));
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Title)), panelTitle_);

    panelFontName_ = new QLineEdit(panel);
    panelFontName_->setText(DisplayObjectQtFontname(dio));
    panelFontName_->setMaxLength(FONTNAME_LEN);
    panelFontName_->setToolTip(GCS(MSG_FontBH));
    PopSelection *popFont = MakeFontPopup(panelFontName_, this);
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Font)), panelFontName_, popFont);
  
    panelBackground_ = new PopImage(panel, pr_);
    panelDefaultBackground_ = new QCheckBox(panel);
    hbox = BackgroundGadgets(panelBackground_, panelDefaultBackground_,
                             DisplayTablePanelBackground(dio));
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Background)), panelBackground_, hbox);

    GroupBox *gadgetBox = new GroupBox(
        GuiMarkShortcut(GCS(MSG_DisplayTable_Panel_Gadgets)), panel);
    group->addWidget(gadgetBox, 2);
    group = new GroupLayout(1, gadgetBox);

    Table *tab = DisplayObjectTable(dio);
    bool gadgets = !TableExactlyOneRecord(tab);

    panelNumAll_ = new QCheckBox(
        GuiMarkShortcut(GCS(MSG_DisplayTable_Panel_NumAll)), gadgetBox);
    panelNumAll_->setChecked(DisplayTablePanelNumAll(dio) && gadgets);
    panelNumAll_->setEnabled(gadgets);
    panelNumAll_->setToolTip(GCS(MSG_DisplayTable_Panel_NumAllBH));
    group->addWidget(panelNumAll_);

    panelFilter_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_Filter)), gadgetBox);
    panelFilter_->setChecked(DisplayTablePanelFilter(dio) && gadgets);
    panelFilter_->setEnabled(gadgets);
    panelNumAll_->setToolTip(GCS(MSG_DisplayTable_Panel_FilterBH));
    group->addWidget(panelFilter_);

    panelArrows_ = new QCheckBox(
            GuiMarkShortcut(GCS(MSG_DisplayTable_Panel_Arrows)), gadgetBox);
    panelArrows_->setChecked(DisplayTablePanelArrows(dio) && gadgets);
    panelArrows_->setEnabled(gadgets);
    panelArrows_->setToolTip(GCS(MSG_DisplayTable_Panel_ArrowsBH));
    group->addWidget(panelArrows_);

    recordDescriptionSpec_ = new EditRecordDescriptionSpec(tab,
            DisplayTableRecordDescriptionSpec(dio), false, tab_);
    tab_->addTab(recordDescriptionSpec_, GuiMarkShortcut(GCS(MSG_Record)));

    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

    setWindowTitle(GCS(MSG_DisplayTable));
    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Table object editor");
}

void GuiDisplayTable(GuiObject sew, DisplayObject *dio)
{
    EditDisplayTableDialog *dialog = new EditDisplayTableDialog(sew, dio);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }
}

#include "qtEditDisplayTable.moc"

#endif

