//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtEditField.cc
//
// Creation date: 01.04.2016
// Last modified: 10.03.2021
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "qt/qtEditField.h"
#include "qt/qtStructureEditor.h"
#include "qt/qtStructureEditorUtil.h"
#include "qt/qtUtil.h"
#include "qt/classes/qtGroupLayout.h"
#include "qt/classes/qtPopEditLabels.h"
#include <QCheckBox>
#include <QListView>
#include <QSpinBox>
#include <QStackedLayout>
#include "menuproject.h"
#include "structureeditor.h"
#include "requesters.h"
#include "preferences.h"
#include "util.h"
#include "labels.h"
#include "error.h"
#include "appstrings_cat.h"

class EditFieldDialog;

class ChoiceLabels : public PopEditLabels
{
public:
    ChoiceLabels(EditFieldDialog *dialog, QWidget *parent) : 
        PopEditLabels(parent), dialog_(dialog) {}

private:
    EditFieldDialog *dialog_;

    virtual bool acceptLabels(char **labels);
};

static short types[] = 					// remapping of types
{
    TYPE_STRING, TYPE_INTEGER, TYPE_REAL, TYPE_BOOL, TYPE_CHOICE, TYPE_DATE, 
    TYPE_TIME, TYPE_MEMO, TYPE_REFERENCE, TYPE_VIRTUAL, TYPE_BUTTON,
};

enum
{
    GRP_STRING, GRP_BOOL, GRP_CHOICE, GRP_MEMO, GRP_REFERENCE, GRP_VIRTUAL, 
    GRP_DEFAULT, GRP_BUTTON,
};

static short pages[] =
{
    GRP_STRING, GRP_DEFAULT, GRP_DEFAULT, GRP_BOOL, GRP_CHOICE, GRP_DEFAULT, 
    GRP_DEFAULT, GRP_MEMO, GRP_REFERENCE, GRP_VIRTUAL, GRP_BUTTON,
};

class EditFieldDialog : public CommonDialog
{
    Q_OBJECT

public:
    EditFieldDialog(GuiObject sew, Table *tab, bool newField,
            Field *field, DisplayObject *parent);

    bool acceptChoiceLabels(char **labels)
    {
        bool warnChanged = false;
        bool warnEmpty = false;
        bool warnDuplicate = false;

        if (field_ && SUCC(HEAD(TableRecords(tab_)))) {
            // Check if the user changed any existing labels.
            char **old = FieldChoiceLabels(field_);
            warnChanged = NumLabels(old) > NumLabels(labels);
            if (!warnChanged) {
                char **s = old, **t = labels;
                while (*s && *t && !warnChanged) {
                    warnChanged =
                        (CodesetStrCollate(*s, *t, CS_COLLATE_ORDER) != 0);
                    s++;
                    t++;
                }
            }
        }

        // Check for empty and duplicate labels.
        for (char **s = labels; *s; s++) {
            warnEmpty |= (NumWords(*s) == 0);
            for (char **t = s + 1; *t; t++) {
                warnDuplicate |= (strcmp(*s, *t) == 0);
            }
        }

        char *msg[3];
        int num = 0;
        if (warnChanged) {
            msg[num++] = GCS(MSG_Field_Choice_WarnChangeLabels);
        }
        if (warnEmpty) {
            msg[num++] = GCS(MSG_Field_Choice_WarnEmptyLabels);
        }
        if (warnDuplicate) {
            msg[num++] = GCS(MSG_Field_Choice_WarnDuplicateLabels);
        }
        if (num > 0) {
            char format[32];
            strcpy(format, "%s");
            for (int i = 0; i < num; i++) {
                strcat(format, "\n\n%s");
            }
            if (!Ask(AT_WARN, GCS(MSG_CloseAnswers),
                        format, GCS(MSG_Field_Choice_Confirm),
                        msg[0], msg[1], msg[2])) {
                return false;
            }
        }
        return true;
    }

private:
    Project *pr_;
    Table *tab_;
    Field *field_;
    DisplayObject *parent_;
    GuiObject sew_;
    QLineEdit *name_;
    ComboBox *type_;
    QStackedLayout *stack_;
    QSpinBox *stringMaxlen_;
    QLineEdit *stringInitialValue_;
    RadioBox *boolInitialValue_;
    ChoiceLabels *choiceLabels_;
    ComboBox *choiceInitialValue_;
    TablesModel *tablesModel_;
    QListView *reference_;
    QCheckBox *referenceFilter_;
    QCheckBox *referenceAutoShow_;
    RadioBox *initialValue_;
    QLineEdit *initialValueOther_;
    QLineEdit *nilString_;
    QLineEdit *computeStr_;
    QLineEdit *nilStringVirtual_;
    QLineEdit *triggerStr_;
    PopString *trigger_;
    QCheckBox *countChanges_;
    QCheckBox *logChanges_;

    void displayErrorMessage(long error)
    {
        Ask2(this, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));

        GuiObject activate = 0;
        switch (error) {
            case ERR_INVALIDNAME:
            case ERR_NAMEALREADYUSED:
            default:    
                activate = name_;
                break;

            case FIELDERR_INVALIDMAXLEN:
                activate = stringMaxlen_;
                break;

            case FIELDERR_INITIALVALUELEN:
            case FIELDERR_INVALIDINITIALVALUE:
                activate = initialValueOther_;
                break;
        }
        if (activate) {
            activate->setFocus();
        }
    }

    long getFieldTypeData(short type, Utf8Buffer &buffer,
            const char **init, const char **nil, FieldTypeData *ftd)
    {
        memset(ftd, 0, sizeof(*ftd));

        switch (type) {
            case TYPE_STRING:
                if (!stringMaxlen_->hasAcceptableInput()) {
                    return FIELDERR_INVALIDMAXLEN;
                }
                ftd->string.maxlen = (short)stringMaxlen_->value();
                SetPrefsValue(pr_, PREFS_DEFAULTMAXLEN, ftd->string.maxlen);
                *init = qtStringUtf8(stringInitialValue_->text(), buffer);
                break;

            case TYPE_INTEGER:
            case TYPE_REAL:
            case TYPE_DATE:
            case TYPE_TIME:
                if (initialValue_->checkedId() != 0) {
                    *init = qtStringUtf8(initialValueOther_->text(), buffer);
                }
                *nil = qtStringUtf8(nilString_->text(), buffer);
                break;

            case TYPE_BOOL:
                ftd->boolean.nil = boolInitialValue_->checkedId() == 0;
                break;

            case TYPE_CHOICE:
                ftd->choice.labels = choiceLabels_->labels();
                ftd->choice.initialvalue = choiceInitialValue_->currentIndex();
                break;

            case TYPE_REFERENCE:
                ftd->reference.table =
                    tablesModel_->table(reference_->currentIndex());
                ftd->reference.filter = referenceFilter_->isChecked();
                ftd->reference.autoshow = referenceAutoShow_->isChecked();
                break;

            case TYPE_VIRTUAL:
                ftd->virt.compute = qtStringUtf8(computeStr_->text(), buffer);
                *nil = qtStringUtf8(nilStringVirtual_->text(), buffer);
                break;
        }

        return OK;
    }

    virtual bool acceptEdits()
    {
        GuiBusyCursor busy;
        long error = OK;
        Utf8Buffer buffer;

        if (!field_) {              // new field?
            int active = type_->currentIndex();
            short type = types[active];

            NewFieldData nfd;
            nfd.nfd_Table = tab_;
            nfd.nfd_Pred = DisplayTableActiveField(tab_);
            nfd.nfd_Name = qtStringUtf8(name_->text(), buffer);
            nfd.nfd_Type = type;
            nfd.nfd_InitialValue = 0;
            nfd.nfd_NilString = 0;
            nfd.nfd_Trigger = qtStringUtf8(triggerStr_->text(), buffer);
            nfd.nfd_CountChanges = countChanges_->isChecked();
            nfd.nfd_LogChanges = logChanges_->isChecked();
            error = getFieldTypeData(type, buffer,
                    &nfd.nfd_InitialValue, &nfd.nfd_NilString,
                    &nfd.nfd_Contents);
            if (!error) {
                Field *field = 0;
                error = SE_NewField(&nfd, parent_, &field);
                if (!error) {
                    NewFieldFinish(sew_, field);
                }
            }
        } else {
            short type = FieldType(field_);

            ChangeFieldData cfd;
            cfd.cfd_Field = field_;
            cfd.cfd_Name = qtStringUtf8(name_->text(), buffer);
            cfd.cfd_InitialValue = 0;
            cfd.cfd_NilString = 0;
            cfd.cfd_Trigger = qtStringUtf8(triggerStr_->text(), buffer);
            cfd.cfd_CountChanges = countChanges_->isChecked();
            cfd.cfd_LogChanges = logChanges_->isChecked();
            error = getFieldTypeData(type, buffer,
                    &cfd.cfd_InitialValue, &cfd.cfd_NilString,
                    &cfd.cfd_Contents);
            if (!error) {
                error = SE_ChangeField(&cfd);
                if (!error) {
                    ChangeFieldFinish(sew_, field_);
                }
            }
        }
        if (error) {
            displayErrorMessage(error);
            return false;
        }
        return true;
    }

private slots:
    void onFieldTypeChanged(int index)
    {
        if (index >= 0) {
            int page = pages[index];
            stack_->setCurrentIndex(page);
            triggerStr_->setEnabled(page != GRP_VIRTUAL);
            trigger_->button()->setEnabled(page != GRP_VIRTUAL);

            bool on = (page != GRP_VIRTUAL && page != GRP_BUTTON);
            bool countOn = on && TableCountChanges(tab_);
            countChanges_->setEnabled(countOn);
            if (!countOn) {
                countChanges_->setChecked(false);
            }
            bool logOn = on && TableLogChanges(tab_);
            logChanges_->setEnabled(logOn);
            if (!logOn) {
                logChanges_->setChecked(false);
            }
        }
    }

    void onChoiceLabelsUpdated(char **labels)
    {
        int index = choiceInitialValue_->currentIndex();
        if (index < 0) {
            index = 0;
        }
        choiceInitialValue_->setLabels(labels);
        choiceInitialValue_->setCurrentIndex(index);
    }

    void onInitialValueChanged(int id)
    {
        initialValueOther_->setEnabled(id == 1);
        if (id == 1) {
            initialValueOther_->setFocus();
        }
    }
};

bool ChoiceLabels::acceptLabels(char **labels)
{ 
    return dialog_->acceptChoiceLabels(labels); 
}


EditFieldDialog::EditFieldDialog(GuiObject sew, Table *tab,
        bool newField, Field *field, DisplayObject *parent) : CommonDialog(sew)
{
    setAttribute(Qt::WA_DeleteOnClose);

    pr_ = TableProject(tab);
    tab_ = tab;
    field_ = newField? 0 : field;
    parent_ = parent;
    sew_ = sew;

    GroupLayout *top = new GroupLayout(2);
    contentsLayout()->addLayout(top);

    name_ = new QLineEdit(this);
    name_->setMaxLength(ID_NAME_LEN);
    name_->setToolTip(GCS(MSG_Field_NameBH));
    name_->setFocus();
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Name)), name_);    

    static char *typeContent[12];
    typeContent[0] = GCS(MSG_String);
    typeContent[1] = GCS(MSG_Integer);
    typeContent[2] = GCS(MSG_Real);
    typeContent[3] = GCS(MSG_Boolean);
    typeContent[4] = GCS(MSG_Choice);
    typeContent[5] = GCS(MSG_Date);
    typeContent[6] = GCS(MSG_Time);
    typeContent[7] = GCS(MSG_Memo);
    typeContent[8] = GCS(MSG_Reference);
    typeContent[9] = GCS(MSG_Virtual);
    typeContent[10] = GCS(MSG_Button);
    typeContent[11] = 0;
    type_ = new ComboBox(this, typeContent);
    type_->setToolTip(GCS(MSG_Field_TypeBH));
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Type)), type_);

    QFrame *frame = new QFrame(this);
    frame->setFrameShape(QFrame::StyledPanel);
    frame->setFrameShadow(QFrame::Raised);
    GuiSetHelpNode(frame, "Type specific settings");
    contentsLayout()->addWidget(frame);

    stack_ = new QStackedLayout(frame);

    // String settings
    QWidget *sub = new QWidget(frame);
    stack_->addWidget(sub);

    GroupLayout *grp = new GroupLayout(2, sub);
    grp->addWidget(new QFrame(sub), 2);

    stringMaxlen_ = new QSpinBox(sub);
    stringMaxlen_->setMinimum(1);
    stringMaxlen_->setMaximum(MAX_STRING_LEN);
    stringMaxlen_->setValue(GetPrefsValue(pr_, PREFS_DEFAULTMAXLEN));
    stringMaxlen_->setToolTip(GCS(MSG_Field_String_MaxLengthBH));
    grp->addLabelAndWidget(
            GuiMarkShortcut(GCS(MSG_Field_String_MaxLength)), stringMaxlen_);

    stringInitialValue_ = new QLineEdit(sub);
    stringInitialValue_->setMaxLength(255);
    stringInitialValue_->setToolTip(GCS(MSG_Field_InitialValueBH));
    grp->addLabelAndWidget(
            GuiMarkShortcut(GCS(MSG_Field__InitialValue)), stringInitialValue_);

    grp->addWidget(new QFrame(sub), 2);

    // Boolean settings
    sub = new QWidget(frame);
    stack_->addWidget(sub);

    grp = new GroupLayout(1, sub);
    grp->addWidget(new QFrame(sub));

    static char *boolInitialValueContent[3];
    boolInitialValueContent[0] = GCS(MSG_Field_Boolean_InitialValue_Nil);
    boolInitialValueContent[1] = GCS(MSG_Field_Boolean_InitialValue_True);
    boolInitialValueContent[2] = 0;
    boolInitialValue_ = new RadioBox(sub, boolInitialValueContent,
            GCS(MSG_Field_InitialValue), true);
    boolInitialValue_->setCheckedId(0);
    boolInitialValue_->setToolTip(GCS(MSG_Field_InitialValueBH));
    grp->addWidget(boolInitialValue_);

    grp->addWidget(new QFrame(sub));

    // Choice settings
    sub = new QWidget(frame);
    stack_->addWidget(sub);

    grp = new GroupLayout(2, sub);
    grp->addWidget(new QFrame(sub), 2);

    choiceLabels_ = new ChoiceLabels(this, sub);
    choiceLabels_->setText(GuiMarkShortcut(GCS(MSG_EditLabels)));
    choiceLabels_->setToolTip(GCS(MSG_Field_Choice_LabelsBH));
    grp->addLabelAndWidget(0, choiceLabels_);
    grp->setStretch(1, 0);
    choiceInitialValue_ = new ComboBox(sub, 0);
    choiceInitialValue_->setToolTip(GCS(MSG_Field_InitialValueBH));
    grp->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Field__InitialValue)),
            choiceInitialValue_);

    grp->addWidget(new QFrame(sub), 2);

    // Memo Settings
    sub = new QWidget(frame);
    stack_->addWidget(sub);

    // Reference settings
    sub = new QWidget(frame);
    stack_->addWidget(sub);

    grp = new GroupLayout(1, sub);

    reference_ = new QListView(sub);
    tablesModel_ = new TablesModel(reference_);
    reference_->setModel(tablesModel_);
    reference_->setSelectionMode(QAbstractItemView::SingleSelection);
    reference_->setToolTip(GCS(MSG_Field_ReferenceBH));
    grp->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Field_Reference)),reference_);

    GroupLayout *hgrp = new GroupLayout(2);
    grp->addLayout(hgrp);

    referenceAutoShow_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_AutoShow)), sub);
    referenceAutoShow_->setToolTip(GCS(MSG_Field_Reference_AutoShowBH));
    hgrp->addWidget(referenceAutoShow_);

    referenceFilter_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_Filter)), sub);
    referenceFilter_->setToolTip(GCS(MSG_Field_Reference_FilterBH));
    hgrp->addWidget(referenceFilter_);

    // Virtual settings
    sub = new QWidget(frame);
    stack_->addWidget(sub);

    grp = new GroupLayout(2, sub);
    grp->addWidget(new QFrame(sub), 2);

    computeStr_ = new QLineEdit(sub);
    computeStr_->setMaxLength(FIELD_TRIGGER_LEN);
    computeStr_->setToolTip(GCS(MSG_Field_ComputeBH));
    grp->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Field_Compute)), computeStr_,
            UserFunctionsPopup(computeStr_, pr_, sub));

    nilStringVirtual_ = new QLineEdit(sub);
    nilStringVirtual_->setMaxLength(FIELD_NILVALUE_LEN);
    nilStringVirtual_->setToolTip(GCS(MSG_Field_NilStringBH));
    grp->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Field_NilString)),
            nilStringVirtual_);

    grp->addWidget(new QFrame(sub), 2);

    // Default settings
    sub = new QWidget(frame);
    stack_->addWidget(sub);

    grp = new GroupLayout(2, sub);
    grp->addWidget(new QFrame(sub), 2);

    GroupBox *initialValueBox = new GroupBox(GCS(MSG_Field_InitialValue), sub);
    grp->addWidget(initialValueBox, 2);

    GroupLayout *subgrp = new GroupLayout(3, initialValueBox);

    static char *initialValueContent[3];
    initialValueContent[0] = GCS(MSG_Field_InitialValue_Nil);
    initialValueContent[1] = GCS(MSG_Field_InitialValue_Other);
    initialValueContent[2] = 0;
    initialValue_ = new RadioBox(initialValueBox, initialValueContent);
    initialValue_->setCheckedId(0);
    initialValue_->setToolTip(GCS(MSG_Field_InitialValueBH));
    subgrp->addWidget(initialValue_);

    QFrame *vSep = new QFrame(initialValueBox);
    vSep->setFrameShape(QFrame::VLine);
    vSep->setFrameShadow(QFrame::Raised);
    subgrp->addWidget(vSep);

    GroupLayout *subsubgrp = new GroupLayout(1);
    subgrp->addLayout(subsubgrp);

    initialValueOther_ = new QLineEdit(initialValueBox);
    initialValueOther_->setEnabled(false);
    initialValueOther_->setMaxLength(20);
    subsubgrp->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Field_OtherValue)),
            initialValueOther_);

    nilString_ = new QLineEdit(sub);
    nilString_->setMaxLength(FIELD_NILVALUE_LEN);
    nilString_->setToolTip(GCS(MSG_Field_NilStringBH));
    grp->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Field_NilString)),nilString_);

    grp->addWidget(new QFrame(sub), 2);

    // Button Settings
    sub = new QWidget(frame);
    stack_->addWidget(sub);

    // Other settings
    grp = new GroupLayout(2);
    contentsLayout()->addLayout(grp);

    triggerStr_ = new QLineEdit(this);
    triggerStr_->setMaxLength(FIELD_TRIGGER_LEN);
    triggerStr_->setToolTip(GCS(MSG_Field_TriggerBH));
    trigger_ = UserFunctionsPopup(triggerStr_, pr_, this);
    grp->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Field_Trigger)),
            triggerStr_, trigger_);

    subgrp = new GroupLayout(2);
    grp->addLayout(subgrp, 2);

    countChanges_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_CountChanges)),this);
    countChanges_->setChecked(true);
    countChanges_->setToolTip(GCS(MSG_Field_CountChangesBH));
    subgrp->addWidget(countChanges_);

    logChanges_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_LogChanges)), this);
    logChanges_->setChecked(true);
    logChanges_->setToolTip(GCS(MSG_Field_LogChangesBH));
    subgrp->addWidget(logChanges_);

    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

    setWindowTitle(!newField?  GCS(MSG_ChangeField) :
            (field?  GCS(MSG_CopyField) : GCS(MSG_NewField)));

    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, 
            newField? "Creating fields" : "Changing fields");

    connect(type_, SIGNAL(currentIndexChanged(int)),
            this, SLOT(onFieldTypeChanged(int)));
    connect(choiceLabels_, SIGNAL(updated(char **)),
            this, SLOT(onChoiceLabelsUpdated(char **)));
    connect(initialValue_, SIGNAL(checked(int)),
            this, SLOT(onInitialValueChanged(int)));

    if (newField) {
        tablesModel_->setList(ProjectTables(pr_));
    }
    if (field) {
        static const short maxPage = sizeof(types) / sizeof(types[0]) - 1;
        short type = FieldType(field), page;

        for (page = 0; page < maxPage; page++) {
            if (types[page] == type) {
                break;
            }
        }

        name_->setText(FieldName(field));
        type_->setCurrentIndex(page);
        type_->setEnabled(newField);

        DataInfo *di = FieldDataInfo(field);
        bool nil = ((di->di_Flags & DIF_NIL) != 0);
        int initialValueId = nil ? 0 : 1;
        switch (type)
        {
	    case TYPE_STRING:
                stringMaxlen_->setValue(FieldStringMaxlen(field));
		stringInitialValue_->setText(di->di_Contents.string);
                break;

	    case TYPE_INTEGER:
                initialValue_->setCheckedId(initialValueId);
                initialValueOther_->setEnabled(initialValueId == 1);
                if (initialValueId == 1) {
                    initialValueOther_->setText(
                            IntToStr(di->di_Contents.integer));
                }
                nilString_->setText(FieldNilString(field));
	        break;

	    case TYPE_REAL:
                initialValue_->setCheckedId(initialValueId);
                initialValueOther_->setEnabled(initialValueId == 1);
                if (initialValueId == 1) {
                    initialValueOther_->setText(
                            DoubleToStr(di->di_Contents.real, -1, 
                                (short)GetPrefsValue(pr_, PREFS_REALFORMAT)));
                }
                nilString_->setText(FieldNilString(field));
	        break;

	    case TYPE_BOOL:
                boolInitialValue_->setCheckedId(initialValueId);
	        break;

	    case TYPE_CHOICE:
                choiceLabels_->setLabels(FieldChoiceLabels(field));
                choiceInitialValue_->setLabels(FieldChoiceLabels(field));
                choiceInitialValue_->setCurrentIndex(di->di_Contents.choice);
	        break;

            case TYPE_DATE:
                initialValue_->setCheckedId(initialValueId);
                initialValueOther_->setEnabled(initialValueId == 1);
                if (initialValueId == 1) {
                    initialValueOther_->setText(
                            DateToString(di->di_Contents.date,
                                (short)GetPrefsValue(pr_, PREFS_DATEFORMAT)));
                }
                nilString_->setText(FieldNilString(field));
	        break;

	    case TYPE_TIME:
                initialValue_->setCheckedId(initialValueId);
                initialValueOther_->setEnabled(initialValueId == 1);
                if (initialValueId == 1) {
                    initialValueOther_->setText(
			    TimeToString(di->di_Contents.time, 
				    FORMAT_HOURMINSEC));
                }
                nilString_->setText(FieldNilString(field));
	        break;

            case TYPE_REFERENCE:
                tablesModel_->setList(ProjectTables(pr_));
                reference_->setCurrentIndex(tablesModel_->tableIndex(
                            FieldReferenceTable(field)));
                reference_->setEnabled(false);
                referenceFilter_->setChecked(FieldReferenceFilter(field));
                referenceAutoShow_->setChecked(
                    FieldReferenceAutoShow(field));
                break;

            case TYPE_VIRTUAL:
                computeStr_->setText(FieldVirtualCompute(field));
                nilStringVirtual_->setText(FieldNilString(field));
                break;
        }
        triggerStr_->setText(FieldTrigger(field));
        countChanges_->setChecked(FieldCountChanges(field));
        logChanges_->setChecked(FieldLogChanges(field));
    }
    onFieldTypeChanged(type_->currentIndex());
}

static bool checkNewChangeField(Field *field, Table *tab, bool newField)
{
    long error = newField? CanNewField(tab) : CanChangeField(field);
    bool rc = (error == OK);

    if (error == RECERR_NOPROJECTNAME) {
        if (Ask(AT_WARN, GCS(MSG_AskProjectSaveFirstAnswers),
                    GCS(MSG_AskProjectSaveFirst)) != 0) {
            MenuProjectSaveAs(TableProject(tab));
            error = newField? CanNewField(tab) : CanChangeField(field);
            rc = (error == OK);
        } // else user canceled
    } else if (error != OK) {
        Ask2(0, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));
    }

    return rc;
}

void GuiNewField(GuiObject sew, Table *tab, Field *field,
        DisplayObject *parent)
{
    if (checkNewChangeField(field, tab, true)) {
        EditFieldDialog *dialog =
            new EditFieldDialog(sew, tab, true, field, parent);
        if (!GuiOpenWindow(dialog)) {
            delete dialog;
        }
    }
}

void GuiChangeField(GuiObject sew, Field *field)
{
    Table *tab = FieldTable(field);
    if (checkNewChangeField(field, tab, false)) {
        EditFieldDialog *dialog =
            new EditFieldDialog(sew, FieldTable(field), false, field, 0);
        if (!GuiOpenWindow(dialog)) {
            delete dialog;
        }
    }
}

#include "qtEditField.moc"

#endif

