//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtEditRecordDescriptionSpec.cc
//
// Creation date: 16.03.2016
// Last modified: 25.04.2016
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "qt/qtEditRecordDescriptionSpec.h"
#include "qt/qtStructureEditorUtil.h"
#include "qt/classes/qtGroupLayout.h"
#include "qt/classes/qtSortableAbstractListModel.h"
#include "qt/qtUtil.h"
#include <QTreeView>
#include "table.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

static bool recordDescriptionSpecFind(const RecordDescriptionSpec *rds, 
        Field *field)
{
    if (rds) {
        for (int i = 0; i < rds->rds_NumFields; i++) {
            if (rds->rds_Fields[i] == field) {
                return true;
            }
        }
    }
    return false;
}

class EditRecordDescriptionSpec::FieldListModel : public SortableAbstractListModel
{
public:
    FieldListModel(QObject *parent) : SortableAbstractListModel(parent)
    {
    }

    void addField(Field *field, bool checked)
    {
        fields_.append(Entry(field, checked));
    }

    QList<Field *> getSelectedFieldList() const {
        QList<Field *> fields;
        for (int i = 0; i < fields_.size(); ++i) {
            if (fields_[i].checked) {
                fields.push_back(fields_[i].field);
            }
        }
        return fields;
    }

protected:
    int numElements() const { return fields_.size(); }

    QString display(int ix) const 
    { 
        if (ix < 0 || ix >= fields_.size()) {
            return QString();
        }
        Field *field = fields_[ix].field;
        if (field == RDS_FIELD_RECNUM) {
            return QString("<i>%1</i>").arg(GCS(MSG_RecordDescription_RecordNumber));
        } else {
            return FieldName(field);
        }
    }

    QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const
    {
        int ix = index.row();
        if (role == Qt::CheckStateRole && index.isValid() && 
                ix >= 0 && ix < fields_.size()) {
            return (fields_[ix].checked? Qt::Checked : Qt::Unchecked);
        }
        return SortableAbstractListModel::data(index, role);
    }

    bool setData(const QModelIndex &index, const QVariant &value,
            int role = Qt::EditRole)
    {
        int ix = index.row();
        if (role == Qt::CheckStateRole && index.isValid() &&
                ix >= 0 && ix < fields_.size()) {
            fields_[index.row()].checked = (value == Qt::Checked);
            emit dataChanged(index, index);
            return true;
        }
        return SortableAbstractListModel::setData(index, value, role);
    }

    Qt::ItemFlags flags(const QModelIndex &index) const
    {
        Qt::ItemFlags flags = SortableAbstractListModel::flags(index);
        if (index.isValid() && index.column() == 0) {
            flags |= Qt::ItemIsUserCheckable;
        }
        return flags;
    }

    virtual bool moveElement(int from, int to)
    {
        // Since we are not really inserting a row, we need to adjust the destination 
        // when moving an item down the list.
        if (to > from) {
            to--;
        }
        fields_.move(from, to);
        return true;
    }

private:
    struct Entry
    {
        Field *field;
        bool checked;

        Entry(Field *a, bool c) : field(a), checked(c) {}
    };
    QList<Entry> fields_;
};

EditRecordDescriptionSpec::EditRecordDescriptionSpec(Table *tab,
        const RecordDescriptionSpec *rds, bool allowInherited, QWidget *parent)
{
    Project *pr = TableProject(tab);
    int type = rds? rds->rds_Type : 0;

    GroupLayout *top = new GroupLayout(2, this);

    static char *desctext[4];
    desctext[0] = GCS(MSG_Fields);
    desctext[1] = GCS(MSG_Computed);
    desctext[2] = allowInherited? GCS(MSG_RecordDescription_FromTable) : 0;
    desctext[3] = 0;
    type_ = new ComboBox(this, desctext);
    type_->setToolTip(GCS(MSG_RecordDescription_DescriptionBH));
    type_->setCurrentIndex(type);
    top->addLabelAndWidget(
            GuiMarkShortcut(GCS(MSG_RecordDescription_Description)), type_);
    top->setStretch(1, 0);
    connect(type_, SIGNAL(currentIndexChanged(int)), this, SLOT(onTypeChanged(int)));

    stack_ = new QStackedLayout();
    top->addLayout(stack_, 2);

    // Field-list page
    fieldList_ = new FieldListModel(this);
    if (rds) {
        for (int i = 0; i < rds->rds_NumFields; i++) {
            if (rds->rds_Fields[i]) {
                fieldList_->addField(rds->rds_Fields[i], true);
            }
        }
    }
    if (!recordDescriptionSpecFind(rds, RDS_FIELD_RECNUM)) {
        fieldList_->addField(RDS_FIELD_RECNUM, false);
    }
    List *fields = TableFields(tab);
    if (fields) {
        Field *field;
        FORNODE (field, fields) {
            switch (FieldType(field))
            {
                case TYPE_STRING:
                case TYPE_DATE:
                case TYPE_TIME:
                case TYPE_INTEGER:
                case TYPE_REAL:
                case TYPE_BOOL:
                case TYPE_CHOICE:
                    if (!recordDescriptionSpecFind(rds, field)) {
                        fieldList_->addField(field, false);
                    }
                    break;
            }
        }
    }

    QTreeView *view = new QTreeView(this);
    view->setItemDelegate(new HtmlDelegate(view));
    view->setModel(fieldList_);
    view->setHeaderHidden(true);
    view->setRootIsDecorated(false);
    view->setUniformRowHeights(true);
    view->setDragDropMode(QAbstractItemView::InternalMove);
    view->setToolTip(GCS(MSG_RecordDescription_FieldsBH));
    stack_->addWidget(view);

    // Compute page
    QWidget *computeBox = new QWidget(this);
    stack_->addWidget(computeBox);
    GroupLayout *group = new GroupLayout(1, computeBox);

    group->addWidget(new QFrame());
    group->setStretch(0, 1);
    computeStr_ = new QLineEdit(computeBox);
    computeStr_->setMaxLength(FIELD_TRIGGER_LEN);
    computeStr_->setText(rds? rds->rds_Compute : "");
    computeStr_->setToolTip(GCS(MSG_RecordDescription_ComputeBH));
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Compute)), computeStr_,
            UserFunctionsPopup(computeStr_, pr, computeBox));
    group->addWidget(new QFrame());
    group->setStretch(0, 1);

    // Empty page
    stack_->addWidget(new QFrame());

    multiColumn_ = new QCheckBox(
            GuiMarkShortcut(GCS(MSG_RecordDescription_UseMultiColumnList)),
            this);
    multiColumn_->setChecked(rds? rds->rds_MultiColumn : false);
    multiColumn_->setEnabled(rds? (rds->rds_Type != RDS_TYPE_INHERITED) : true);
    multiColumn_->setToolTip(GCS(MSG_RecordDescription_UseMultiColumnListBH));
    top->addWidget(multiColumn_, 2);

    stack_->setCurrentIndex(type);
}

void EditRecordDescriptionSpec::onTypeChanged(int type)
{
    stack_->setCurrentIndex(type);
    multiColumn_->setEnabled(type != RDS_TYPE_INHERITED);
}

RecordDescriptionSpec *EditRecordDescriptionSpec::getSpec(MemPool *memPool)
{
    QList<Field *> fields = fieldList_->getSelectedFieldList();
    RecordDescriptionSpec *rds = AllocRecordDescriptionSpec(memPool, fields.size());

    if (rds) {
        QByteArray computeStr = computeStr_->text().toUtf8();
        rds->rds_Type = type_->currentIndex();
        rds->rds_Compute = AllocString(memPool, computeStr.data());
        for (int i = 0; i < fields.size(); ++i) {
            rds->rds_Fields[i] = fields[i];
        }
        rds->rds_MultiColumn = multiColumn_->isChecked();
    }

    return rds;
}

#endif

