//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtEditTable.cc
//
// Creation date: 15.02.2016
// Last modified: 15.05.2017
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "qt/qtEditTable.h"
#include "qt/qtStructureEditor.h"
#include "qt/qtStructureEditorUtil.h"
#include "qt/classes/qtGroupLayout.h"
#include <QCheckBox>
#include "qt/qtUtil.h"
#include "structureeditor.h"
#include "requesters.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"

class EditTableDialog : public CommonDialog
{
    Q_OBJECT
    
public:
    EditTableDialog(GuiObject sew, Project *pr, Table *tab, DisplayObject *dio);

private:
    Project *pr_;
    Table *tab_;
    DisplayObject *dio_;
    GuiObject sew_;
    QLineEdit *name_;
    RadioBox *numRecords_;
    QLineEdit *newTrigger_;
    QLineEdit *delTrigger_;
    QCheckBox *countChanges_;
    QCheckBox *logChanges_;
    
    virtual bool acceptEdits()
    {
    	GuiBusyCursor busy;

        Utf8Buffer buffer;
        struct TableData td;
        td.td_Name = qtStringUtf8(name_->text(), buffer);
	td.td_System = FALSE;
        td.td_ExactlyOneRecord = numRecords_->checkedId() == 1;
     	td.td_NewTrigger = qtStringUtf8(newTrigger_->text(), buffer);
     	td.td_DelTrigger = qtStringUtf8(delTrigger_->text(), buffer);
        td.td_CountChanges = countChanges_->isChecked();
        td.td_LogChanges = logChanges_->isChecked();

        long error = OK;
        Table *newTab;
        if (!tab_) {
    	    error = SE_NewTable(pr_, &td, dio_, &newTab);
            if (error == OK) {
                NewTableFinish(sew_, newTab);
            }
        } else {
	    error = SE_ChangeTable(tab_, &td);
            if (error == OK) {
                ChangeTableFinish(sew_, tab_);
            }
        }

    	if (error) {
            Ask2(this, GCS(MSG_ErrorMessage), AT_ERROR, 
                    GCS(MSG_StarOk), GES(error));
            GuiSetActiveGuiObject(this, name_);
            return false;
        }
        return true;
    }
};

EditTableDialog::EditTableDialog(GuiObject sew, Project *pr, Table *tab,
                                 DisplayObject *dio) : CommonDialog(sew)
{
    setAttribute(Qt::WA_DeleteOnClose);
 
    pr_ = pr;
    tab_ = tab;
    dio_ = dio;
    sew_ = sew;

    GroupLayout *top = new GroupLayout(2);
    contentsLayout()->addLayout(top);
    
    name_ = new QLineEdit(this);
    name_->setMaxLength(ID_NAME_LEN);
    name_->setToolTip(GCS(MSG_Table_NameBH));
    name_->setFocus();
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Name)), name_);

    static char *numRecordsContent[3];
    numRecordsContent[0] = GCS(MSG_Unlimited);
    numRecordsContent[1] = GCS(MSG_Table_ExactlyOne);
    numRecordsContent[2] = NULL;
    numRecords_ = new RadioBox(this, numRecordsContent,
                               GCS(MSG_Table_NumberOfRecords),
                               true, false);
    numRecords_->setCheckedId(0);
    numRecords_->setToolTip(GCS(MSG_Table_NumberOfRecordsBH));
    top->addWidget(numRecords_, 2);

    GroupBox *triggers = new GroupBox(GCS(MSG_Table_TriggerFunctions), this);
    top->addWidget(triggers, 2);
    
    GroupLayout *group = new GroupLayout(2, triggers);
    newTrigger_ = new QLineEdit(triggers);
    newTrigger_->setMaxLength(TABLE_TRIGGER_LEN);
    QString label = GuiMarkShortcut(GCS(MSG_Table_Trigger_New));
    group->addLabelAndWidget(label, newTrigger_,
                             UserFunctionsPopup(newTrigger_, pr, triggers));
    delTrigger_ = new QLineEdit(triggers);
    delTrigger_->setMaxLength(TABLE_TRIGGER_LEN);
    label = GuiMarkShortcut(GCS(MSG_Table_Trigger_Delete));
    group->addLabelAndWidget(label, delTrigger_,
                             UserFunctionsPopup(delTrigger_, pr, triggers));

    group = new GroupLayout(2);
    top->addLayout(group, 2);

    countChanges_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_CountChanges)), this);
    countChanges_->setChecked(true);
    countChanges_->setToolTip(GCS(MSG_Table_CountChangesBH));
    group->addWidget(countChanges_);
    
    logChanges_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_LogChanges)), this);
    logChanges_->setChecked(true);
    logChanges_->setToolTip(GCS(MSG_Table_LogChangesBH));
    group->addWidget(logChanges_);

    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

    setWindowTitle(tab? GCS(MSG_ChangeTable) : GCS(MSG_NewTable));
   
    setSleepProject(pr);
    GuiSetWindowProject(this, pr);
    GuiSetHelpNode(this, tab? "Changing tables" : "Creating tables");

    if (tab) {
        name_->setText(TableName(tab));
        numRecords_->setCheckedId(TableExactlyOneRecord(tab)? 1 : 0);
        newTrigger_->setText(TableNewTrigger(tab));
        delTrigger_->setText(TableDelTrigger(tab));
        countChanges_->setChecked(TableCountChanges(tab));
        logChanges_->setChecked(TableLogChanges(tab));
    }
}

void createEditTable(GuiObject sew, Project *pr, Table *tab, DisplayObject *dio)
{
    EditTableDialog *dialog = new EditTableDialog(sew, pr, tab, dio);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }    
}

void GuiNewTable(GuiObject sew, Project *pr, DisplayObject *dio)
{
    createEditTable(sew, pr, 0, dio);
}

void GuiChangeTable(GuiObject sew, Table *tab)
{
    createEditTable(sew, TableProject(tab), tab, 0);
}

#include "qtEditTable.moc"

#endif

