//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtFilter.cc
//
// Creation date: 22.08.2016
// Last modified: 31.10.2016
//

#include "filter.h"
#include "qt/classes/qtGroupLayout.h"
#include "qt/qtStructureEditorUtil.h"
#include "qt/qtUtil.h"
#include "lang/compile.h"
#include "lang/program.h"
#include "project.h"
#include "table.h"
#include "memory.h"
#include "appstrings_cat.h"

class FilterFieldsModel : public FieldsModel
{
public:
    FilterFieldsModel(QObject *parent) : FieldsModel(parent) {}
    
protected:
    virtual bool filter(Node *n) const
    { 
        return (FieldType((Field *)n) != TYPE_BUTTON);
    }
};

class EditFilterDialog : public CommonDialog
{
    Q_OBJECT

public:
    EditFilterDialog(Table *tab);

public slots:
    void insertText(const QString &text)
    {
        expression_->cut();
        expression_->insert(text);
        expression_->setFocus();
    }

    void insertBracketed(const QString &text)
    {
        insertText(QString("(%1 )").arg(text));
        expression_->cursorBackward(false);
    }

    void insertNormal(const QString &text)
    {
        insertText(QString("%1 ").arg(text));
    }

    void onFieldActivated(const QModelIndex &index)
    {
        Field *field = fieldsModel_->field(index);
        insertNormal(FieldName(field));
    }

protected:
    virtual bool acceptEdits()
    {
	GuiBusyCursor busy;
        const QString &text = expression_->text();
        QByteArray buf = text.toUtf8();
        const char *s = buf.data();
        long len = strlen(s);
	Program *prg = 0;
        long line = 0, col = 0;
        const char *error = CompileFilter(tab_, s, len, &prg, &line, &col);

        if (error) {
            setWindowTitle(error);
            if (col > 0 && col <= len) {
		expression_->setFocus();
                expression_->setCursorPosition(col - 1);
	    }
            GuiBeep();
	    return false;
    	}
        SetFilter(tab_, prg);
        return true;
    }

private:
    Project *pr_;
    Table *tab_;
    FilterFieldsModel *fieldsModel_;
    QListView *fieldsView_;
    QLineEdit *expression_;
};

static struct
{
    const char *fld_String;
    short fld_InsertFuncNum;
} filterLabelDatas[] =
{
    { "=", 	0 },
    { "<>", 	0 },
    { "=*", 	0 },
    { "<>*", 	0 },
    { "<", 	0 },
    { "<=", 	0 },
    { "<*", 	0 },
    { "<=*", 	0 },
    { ">", 	0 },
    { ">=", 	0 },
    { ">*", 	0 },
    { ">=*", 	0 },
    { "+", 	0 },
    { "-", 	0 },
    { "*", 	0 },
    { "/", 	0 },
    { "DIV", 	0 },
    { "MOD", 	0 },
    { "LIKE", 	0 },
    { "NOT", 	0 },
    { "AND", 	0 },
    { "OR", 	0 },
    { "TRUE", 	1 },
    { "NIL", 	1 },
};

#define NUM_FILTERLABELS (sizeof(filterLabelDatas)/sizeof(filterLabelDatas[0]))

class LabelButton : public QToolButton
{
    Q_OBJECT

public:
    LabelButton(QWidget *parent) : QToolButton(parent)
    {
        setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed);
        connect(this, SIGNAL(clicked()), this, SLOT(sendClickedText()));
    }

signals:
    void clickedText(const QString &text);

private slots:
    void sendClickedText()
    {
        emit clickedText(text());
    }
};

EditFilterDialog::EditFilterDialog(Table *tab) :
    CommonDialog(GuiFindParentWindow(0, TableProject(tab)))
{
    setAttribute(Qt::WA_DeleteOnClose);

    pr_ = TableProject(tab);
    tab_ = tab;

    GroupBox *filterLabelsFrame =
        new GroupBox(GCS(MSG_Filter_Operators), this);
    GroupLayout *filterLabelsGroup = new GroupLayout(4, filterLabelsFrame);

    for (size_t i=0; i < NUM_FILTERLABELS; i++) {
        LabelButton *button = new LabelButton(this);
        button->setText(filterLabelDatas[i].fld_String);
        if (filterLabelDatas[i].fld_InsertFuncNum == 0) {
            connect(button, SIGNAL(clickedText(const QString &)),
                    this, SLOT(insertBracketed(const QString &)));
        } else {
            connect(button, SIGNAL(clickedText(const QString &)),
                    this, SLOT(insertNormal(const QString &)));
        }
	filterLabelsGroup->addWidget(button);
    }

    fieldsView_ = new QListView(this);
    fieldsModel_ = new FilterFieldsModel(fieldsView_);
    fieldsModel_->setList(TableFields(tab));
    fieldsView_->setModel(fieldsModel_);
    fieldsView_->setSelectionMode(QAbstractItemView::SingleSelection);
    fieldsView_->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed);
    connect(fieldsView_, SIGNAL(activated(const QModelIndex &)),
            this, SLOT(onFieldActivated(const QModelIndex &)));

    GroupBox *fieldsFrame = new GroupBox(GCS(MSG_Fields), this);
    GroupLayout *fieldsGroup = new GroupLayout(1, fieldsFrame);
    fieldsGroup->addWidget(fieldsView_);

    expression_ = new QLineEdit(this);
    Program *prg = TableFilterProgram(tab);
    if (prg) {
        SourceText *st = GetSourceText(prg);
        const char *text = SourceTextText(st);
        expression_->setText(text);
        FreeSourceText(st);
    }

    QHBoxLayout *hbox = new QHBoxLayout();
    hbox->addWidget(fieldsFrame);
    hbox->addWidget(filterLabelsFrame);

    contentsLayout()->addLayout(hbox);
    contentsLayout()->addWidget(expression_);

    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

    setWindowTitle(
            QString("%1: %2").arg(GCS(MSG_ChangeFilter)).arg(TableName(tab)));

    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Changing filters");

    expression_->setFocus();
}

void GuiChangeFilter(Table *tab)
{
    EditFilterDialog *dialog = new EditFilterDialog(tab);
    if (dialog) {
        dialog->show();
    }
}

#include "qtFilter.moc"
