//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtGui.cc
//
// Creation date: 03.03.2009
// Last modified: 08.03.2022
//

#include "system.h"
#include "qt/qtGui.h"
#include "qt/qtUtil.h"
#include "qt/qtMenus.h"
#include "qt/qtBuildGuiField.h"
#include "qt/classes/qtExtApplication.h"
#include "qt/classes/qtPopImage.h"
#include <QAbstractButton>
#include <QDesktopWidget>
#include <QDir>
#include <QLabel>
#include <QLibraryInfo>
#include <QLineEdit>
#include <QMessageBox>
#include <QMutex>
#include <QProcessEnvironment>
#include <QTextEdit>
#include <QTranslator>
#include "menu.h"
#include "preferences.h"
#include "setformats.h"
#include "requesters.h"
#include "programeditor.h"
#include "structureeditor.h"
#include "programeditor.h"
#include "programoutput.h"
#include "buildgui.h"
#include "order.h"
#include "filter.h"
#include "search.h"
#include "imexport.h"
#include "print.h"
#include "break.h"
#include "memory.h"
#include "dosutil.h"
#include "util.h"
#include "exec.h"
#include "builddate.h"
#include "version.h"
#include "appstrings_cat.h"
#ifdef MACOSX
#include "CoreFoundation/CoreFoundation.h"
#endif

char DefaultRealFormat = FORMAT_DECIMALPOINT;
char DefaultDateFormat = FORMAT_DAYMONTHYEAR;
static const char ThirdPartyCopyright[] = "QT 5 Copyright (C) 2022 The Qt Company";
static const char titleSep[] = " | ";

ExtApplication *GuiApplication = 0;
static QTranslator qtTranslator;
static QTranslator appTranslator;
static QByteArray qtLanguage;

class WindowBusyEventFilter : public QObject
{
public:
    WindowBusyEventFilter(QObject *parent = 0) : QObject(parent) {}
    
    bool eventFilter(QObject *obj, QEvent *event)
    {
        switch (event->type()) {
            case QEvent::Close:
            case QEvent::KeyPress:
            case QEvent::KeyRelease:
                event->ignore();
                return true;
            case QEvent::MouseButtonPress:
            case QEvent::NonClientAreaMouseButtonPress:
            {
                Project *pr = GuiActiveProject();
                if (pr) {
                    foreach (QWidget *top, QApplication::topLevelWidgets()) {
                        if (top->isVisible() && GuiWindowProject(top) == pr &&
                                top->property("sleep").toInt() == 0) {
                            top->raise();
                            top->activateWindow();
                        }
                    }
                }
                break;
            }
                
            default:
                break;
        }
        return false;
    }
};

WindowBusyEventFilter *windowBusyEventFilter = 0;

void CloseGui(void) 
{
    RestoreSignalHandlers();
    GuiMenuExit();
    CodesetExit();

    delete GuiApplication;
    GuiApplication = 0;
}

const char *OpenGui(int *pargc, char **argv)
{
    GuiApplication = new ExtApplication(*pargc, argv);
    if (!GuiApplication) {
        return "Unable to create application!";
    }
    
    // Load translations.
    QLocale locale;
    qtLanguage = locale.nativeLanguageName().toUtf8();
    QString transPath(QLibraryInfo::location(QLibraryInfo::TranslationsPath));
    qtTranslator.load(locale, "qt", "_", transPath);
    QCoreApplication::installTranslator(&qtTranslator);

    appTranslator.load(locale, APP_NAME, "_", transPath);
    QCoreApplication::installTranslator(&appTranslator);
    qputenv("LANG", locale.name().toUtf8());

    windowBusyEventFilter = new WindowBusyEventFilter(GuiApplication);

#if (defined WINDOWS)
    QString dirPath = QCoreApplication::applicationDirPath();
    if (dirPath.endsWith("/bin") || dirPath.endsWith("\\bin")) {
        dirPath.remove(dirPath.length() - 4, 4);
    } else {
        dirPath = "C:/Program Files/BeeBase";
    }
    qputenv("BeeBase", dirPath.toUtf8());
    qputenv("HOME", QDir::homePath().toUtf8());
    GuiApplication->setWindowIcon(QIcon(":/bee64x64.png"));
#elif (defined MACOSX)
    CFURLRef appUrlRef = CFBundleCopyBundleURL(CFBundleGetMainBundle());
    CFStringRef macPath = CFURLCopyFileSystemPath(appUrlRef,
                                                  kCFURLPOSIXPathStyle);
    const char *pathPtr = CFStringGetCStringPtr(macPath,
                                                CFStringGetSystemEncoding());
    QString path(pathPtr);
    CFRelease(appUrlRef);
    CFRelease(macPath);
    QString resourcesPath("/Applications/BeeBase.app/Contents/Resources");
    QString assignPath(resourcesPath);
    assignPath.append("/share/BeeBase");
    if (!path.isEmpty()) {
        path.append("/Contents/Resources");
        if (QDir(path).exists()) {
            resourcesPath = path;
            path.append("/share/BeeBase");
            if (QDir(path).exists()) {
                assignPath = path;
            }
        }
    }
    qputenv(APP_NAME, assignPath.toUtf8());
    // On MacOSX we use the same application icon as window icon.
    // Otherwise the application also changes to the window icon.
    QString iconPath(resourcesPath);
    iconPath.append("/BeeBase.icns");
    GuiApplication->setWindowIcon(QIcon(iconPath));
#elif (defined UNIX)
    setenv(APP_NAME, "/usr/share/BeeBase", false);
    GuiApplication->setWindowIcon(QIcon(":/bee64x64.png"));
#endif

    const char *error = CodesetInit();
    if (error) {
	return error;
    }
    
    GuiMenuInit();
    InstallSignalHandlers();

    return 0;
}

static QMutex translationMutex;
static QMap<QString, QByteArray> translationMap;

char *GuiGetTranslation(const char *s)
{
    translationMutex.lock();
    QString key(s);
    QMap<QString, QByteArray>::iterator it = translationMap.find(key);
    if (it == translationMap.end()) {
        it = translationMap.insert(key, QCoreApplication::translate("", s).toUtf8());
    }
    char *translation = it->data();
    translationMutex.unlock();
    return translation;
}

void GuiPrefsUpdated(void)
{
    UpdateMenu(GuiActiveWindow(), true);
}

void GuiBeep(void)
{
    QApplication::beep();
}

const char *GuiLanguage()
{
    return qtLanguage.constData();
}

void GuiAbout() 
{
    QMessageBox::about(GuiActiveWindow(), APP_NAME, 
                       QString().sprintf(GCS(MSG_AboutBeeBase_GPL),
                                         BB_VERSTR, BUILD_YEAR, 
                                         ThirdPartyCopyright, ""));
}

void GuiSendQuitMessage(void) 
{
    GuiApplication->quit();
}

static ulong sleepCount = 0;

ulong GuiSleep() 
{ 
    return sleepCount; 
}

static void guiSetBusyCursor(GuiObject win)
{
    if (win) {
        QWidget *focus = win->focusWidget();
        // For some reasons "focus" as property seems to be in use by Qt.
        win->setProperty("focusWidget", 
                QVariant::fromValue<QPointer<QWidget> >(focus));
        win->installEventFilter(windowBusyEventFilter);
        QWidget *busy = win->property("busy").value<QWidget *>();
        if (!busy) {
            busy = new QWidget(win);
            busy->setCursor(Qt::BusyCursor);
            busy->move(0, 0);
            busy->resize(1000, 1000);  // todo
            busy->setFocusPolicy(Qt::StrongFocus);
            win->setProperty("busy", QVariant::fromValue<QWidget *>(busy));
        }
        busy->show();
        busy->setFocus();
        win->setCursor(Qt::BusyCursor);
    }
}

static void guiClearBusyCursor(GuiObject win)
{
    if (win) {
        win->removeEventFilter(windowBusyEventFilter);
        QWidget *busy = win->property("busy").value<QWidget *>();
        if (busy) {
            busy->hide();
            win->setCursor(Qt::ArrowCursor);
        }
        QWidget *focus=win->property("focusWidget").value<QPointer<QWidget> >();
        if (focus) {
            focus->setFocus();
        }
    }
}

static void guiUpdateBusyCursor(GuiObject win)
{
    if (win) {
        int count = win->property("sleep").toInt();
        if (count == 0) {
            guiClearBusyCursor(win);
        } else {
            guiSetBusyCursor(win);
        }
    }
}

static void guiSetWindowSleep(GuiObject win, bool sleep)
{
    if (win && win->isVisible()) {
        bool update = false;
        int count = win->property("sleep").toInt();
        if (sleep) {
            update = (count++ == 0);
        } else if (count > 0) {
            update = (--count == 0);
        }
        win->setProperty("sleep", count);
        if (update) {
            guiUpdateBusyCursor(win);
        }
    }
}

void GuiClearWindowSleep(GuiObject win)
{
    if (win) {
        guiClearBusyCursor(win);
        win->setProperty("sleep", 0);
    }
}

void GuiSetSleep(bool sleep) 
{
    if (sleep) {
	sleepCount++;
    } else {
	sleepCount--;
    }
    
    foreach (QWidget *top, QApplication::topLevelWidgets()) {
	guiSetWindowSleep(top, sleep);
    }
}

ulong GuiProjectSleep(Project *pr)
{
    GuiObject w = ProjectRootWindow(pr);
    ulong sleep = 0;
    
    if (w) {
        sleep = w->property("sleep").toInt();
    }
    
    return sleep;
}

void GuiSetProjectSleep(Project *pr, bool sleep)
{
    foreach (QWidget *top, QApplication::topLevelWidgets()) {
        if (GuiWindowProject(top) == pr) {
            guiSetWindowSleep(top, sleep);
        }
    }
}

char *GuiEditorText(GuiObject go, struct MemPool *memPool)
{
    char *str = 0;
    TextEditor *editor = dynamic_cast<TextEditor *>(go);
    if (editor) {
        QByteArray data = editor->toPlainTextNbsp().toUtf8();
        str = AllocString(memPool, data.data());
    }
    return str;
}

void GuiSetEditorText(GuiObject go, const char *str)
{
    TextEditor *editor = dynamic_cast<TextEditor *>(go);
    if (editor) {
        editor->setPlainText(str);
    }
}
    
void GuiSetEditorChanged(GuiObject go, bool changed)
{
    QTextEdit *editor = dynamic_cast<QTextEdit *>(go);
    if (editor) {
        QTextDocument *doc = editor->document();
        if (doc) {
            doc->setModified(changed);
        }
    }
}

void GuiSetRegisterPage(GuiObject go, long pos)
{
    QTabWidget *tab = dynamic_cast<QTabWidget *>(go);
    if (tab) {
        tab->setCurrentIndex(pos);
    }
}

long GuiRegisterActivePage(GuiObject go)
{
    long index = 0;
    QTabWidget *tab = dynamic_cast<QTabWidget *>(go);
    if (tab) {
        index = tab->currentIndex();
    }
    return index; 
}

bool GuiOpenWindow(GuiObject wi) 
{
    bool open = false;
    if (wi) {
        wi->show();
        open = wi->isVisible();
    }
    return open;

}
void GuiCloseWindow(GuiObject wi)
{
    if (wi) {
        wi->hide();
    }
}

short GuiObjectWeight(GuiObject go, bool horiz)
{
    // todo (we should set the DisplayObjectWeight directly whenever
    // a QSplitter changes.
    return 0;
}

const char *GuiObjectHelpNode(GuiObject go)
{
    static QByteArray data;
    const char *node = 0;
    
    if (go) {
        QVariant property = go->property("help-node");
        QString help = property.toString();
        if (!help.isEmpty()) {
            data = help.toUtf8();
            node = data.data();
        }
    }
    return node;
}

GuiObject GuiObjectParent(GuiObject go)
{
    GuiObject parent = 0;
    
    if (go) {
        parent = go->parentWidget();
    }
    return parent;
}

bool GuiGetWindowDimensions(GuiObject win, struct WindowDimensions *wd)
{
    if (win && win->isVisible() && wd) {
        wd->wd_LeftEdge = win->pos().x();
        wd->wd_TopEdge = win->pos().y();
	wd->wd_Width = win->size().width();
	wd->wd_Height = win->size().height();

        return true;
    }
    return false;
}

void GuiSetWindowDimensions(GuiObject win, const struct WindowDimensions *wd)
{
    if (win && wd) {
	int x = wd->wd_LeftEdge, y = wd->wd_TopEdge;
	int width = wd->wd_Width, height = wd->wd_Height;
        QDesktopWidget *root = QApplication::desktop();
        if (root) {
            QRect geom = root->screenGeometry();
            x = MIN(x, geom.width() - width);
            x = MAX(x, 0);
            y = MIN(y, geom.height() - height);
            y = MAX(y, 0);

            win->resize(width, height);
            win->move(x, y);
        }
    }
}

GuiObject GuiActiveWindow(void) 
{ 
    return QApplication::activeWindow();
}

GuiObject GuiActiveObject(GuiObject wi) 
{
    QWidget *focus = 0;    

    if (wi) {
        focus = wi->focusWidget();
    }
    return focus;
}
 
void GuiSetSelected(GuiObject go, bool selected)
{
    QAbstractButton *button = dynamic_cast<QAbstractButton *>(go);
    if (button) {
        button->setChecked(selected);
    }
}

void GuiSetEnabled(GuiObject go, bool enabled)
{
    if (go) {
        go->setEnabled(enabled);
    }
}

bool GuiGetEnabled(GuiObject go) 
{
    bool enabled = false;
    
    if (go) {
        enabled = go->isEnabled();
    }
    return enabled;
}

void GuiSetBackground(GuiObject go, const char *bg)
{
    if (go) {
        qtSetWidgetBackground(go, bg);
    }
}

const char *GuiStringContents(GuiObject go)
{
    QLineEdit *edit = dynamic_cast<QLineEdit *>(go);
    const char *str = 0;
    
    if (edit) {
        static QByteArray array;
        array = edit->text().toUtf8();
        str = array.data();
    }
    return str;
}

const char *GuiObjectText(GuiObject go)
{
    QString text;
    QLineEdit *edit = dynamic_cast<QLineEdit *>(go);
    if (edit) {
        text = edit->text();
    }
    QLabel *label = dynamic_cast<QLabel *>(go);
    if (label) {
        text = label->text();
    }
    static QByteArray array;
    array = text.toUtf8();
    return array.data();
}

bool GuiWindowOpen(GuiObject wi)
{ 
    bool open = false;
    
    if (wi) {
        open = wi->isVisible();
    }
    return open;
}

struct Project *GuiActiveProject(void) 
{
    return GuiWindowProject(GuiActiveWindow());
}

void GuiActivateWindow(GuiObject wi)
{
    if (wi) {
        wi->activateWindow();
    }
}

void GuiWindowToFront(GuiObject wi)
{
    if (wi) {
        wi->raise();
    }
}

void GuiSetActiveGuiObject(GuiObject wi, GuiObject go)
{
    if (go) {
        go->setFocus(Qt::OtherFocusReason);
    }
}

void GuiSetWindowTitle(GuiObject wi, const char *title)
{
    if (wi) {
        QString prepend = wi->property("prependTitle").toString();
        QString windowTitle(title);
        if (!prepend.isEmpty()) {
            windowTitle.prepend(titleSep);
            windowTitle.prepend(prepend);
        }
        wi->setWindowTitle(windowTitle);
    }
}

void GuiSetPrependWindowTitle(GuiObject wi, const char *prependStr)
{
    if (wi) {
        QString prepend(prependStr);
        QString currentPrepend = wi->property("prependTitle").toString();
        if (prepend != currentPrepend) {
            QString windowTitle = wi->windowTitle();
            int currentLen = currentPrepend.size();
            int sepLen = strlen(titleSep);
            if (currentLen != 0 && windowTitle.size() >= currentLen + sepLen && 
                    windowTitle.startsWith(currentPrepend) &&
                    windowTitle.mid(currentLen).startsWith(titleSep)) {
                windowTitle = windowTitle.mid(currentLen + sepLen);
            }
            wi->setProperty("prependTitle", prepend);
            QByteArray array = windowTitle.toUtf8();
            const char *title = array.data();
            GuiSetWindowTitle(wi, title);
        }
    }
}

void GuiStoreEdits(void) 
{
    GuiFieldStoreEdits(); 
}

void GuiMainLoop(bool (*reallyQuit)(void))
{
    if (GuiApplication) {
        do {
            GuiApplication->exec();
        } while (reallyQuit && !reallyQuit());
    }
}
bool GuiMainLoopIteration(void) 
{
    QApplication::processEvents(QEventLoop::WaitForMoreEvents);
    return QApplication::closingDown();
}

