//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtImExport.cc
//
// Creation date: 22.08.2016
// Last modified: 06.11.2020
//

#include "system.h"
#include "imexport.h"
#include "qt/qtUtil.h"
#include "qt/classes/qtGroupLayout.h"
#include "qt/classes/qtPopSelection.h"
#include "qt/classes/qtSFWView.h"
#include <QCheckBox>
#include "table.h"
#include "log.h"
#include "guidosutil.h"
#include "util.h"
#include "memory.h"
#include "appstrings_cat.h"

#define DELIMITER_LEN 9

static struct
{
    QString ies_Filename;
    QString ies_RecordDelimiter;
    QString ies_FieldDelimiter;
    bool ies_DoubleQuotes;
    bool ies_Filter;
    bool ies_OverwriteRecords;
} ImExportSettings =
{
    "", "\\n", "\\t", true, false, false,
};

enum ImExportAction
{
    IEA_Import,
#ifdef ENABLE_LOG_APPLY_CHANGES
    IEA_ApplyChanges,
#endif
    IEA_Export,
    IEA_ExportSFW
};

class ImExportDialog : public CommonDialog
{
    Q_OBJECT

public:
    ImExportDialog(Project *pr, Table *tab, GuiObject sfw,
            ImExportAction action);

private slots:
    void onViewClicked() { GuiObjectViewFile(filename_); }
    void onEditClicked() { GuiObjectEditFile(filename_); }

private:
    void saveImExportData()
    {
        ImExportSettings.ies_Filename = filename_->text();
        ImExportSettings.ies_RecordDelimiter = recordDelimiter_->text();
        ImExportSettings.ies_FieldDelimiter = fieldDelimiter_->text();
        ImExportSettings.ies_DoubleQuotes= doubleQuotes_->isChecked();
        if (filter_) {
            ImExportSettings.ies_Filter = filter_->isChecked();
        }
        if (overwriteRecords_) {
            ImExportSettings.ies_OverwriteRecords = 
                overwriteRecords_->isChecked();
        }
    } 

    const char *checkDelimiterString(const QString &delstring, 
            struct MemPool *memPool, char *del)
    {
        Utf8Buffer buffer;
        char *s = StrCompress(memPool, qtStringUtf8(delstring, buffer));
        if (!s) {
            return GCS(MSG_Err_Mem);
        }

        short len = strlen(s);
        bool ascii = StrIsASCII(s);
        *del = *s;
        FreeString(memPool, s);
        if (len != 1 || !ascii ||
                (*del == '"' && ImExportSettings.ies_DoubleQuotes)) {
            return GCS(MSG_ImExportErr_InvalidDelimiter);
        }
        return 0;
    }

    bool checkImExportData(char *recdel, char *fielddel)
    {
        MemPool *memPool = ProjectMemPool(pr_);
        const char *error;
   
        if (ImExportSettings.ies_Filename.isEmpty()) {
            error = GCS(MSG_Err_NoFilename);
            setWindowTitle(error);
            filename_->setFocus();
            GuiBeep();
            return false;
        }
        error = checkDelimiterString(
	    ImExportSettings.ies_RecordDelimiter, memPool, recdel);
        if (error) {
            setWindowTitle(error);
            recordDelimiter_->setFocus();
            GuiBeep();
            return false;
        }
        error = checkDelimiterString(ImExportSettings.ies_FieldDelimiter, 
                memPool, fielddel);
        if (error) {
            setWindowTitle(error);
            fieldDelimiter_->setFocus();
            GuiBeep();
            return false;
        }
        return true;
    }

    virtual bool acceptEdits()
    {
	char recdel = 0, fielddel = 0;
        bool rc = false;

        if (!checkImExportData(&recdel, &fielddel)) {
            return false;
	}

        Utf8Buffer buffer;
        const char *filename=qtStringUtf8(ImExportSettings.ies_Filename,buffer);
        switch (action_) {
            case IEA_Import:
                rc = ImportRecords(tab_, ImExportSettings.ies_OverwriteRecords,
                        filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;
#ifdef ENABLE_LOG_APPLY_CHANGES
            case IEA_ApplyChanges:
                rc = ImportApplyChanges(pr_, filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;
#endif
            case IEA_Export:
                rc = ExportRecords(tab_, ImExportSettings.ies_Filter,
                        filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;
            case IEA_ExportSFW:
                rc = ExportSFWList(pr_, sfw_, filename, recdel, fielddel,
                        ImExportSettings.ies_DoubleQuotes);
                break;
        }
        return rc;
    }
  

    virtual void done(int r)
    {
        saveImExportData();
        CommonDialog::done(r);
    }

    Project *pr_;
    Table *tab_;
    GuiObject sfw_;
    ImExportAction action_;
    QLineEdit *filename_;
    ExtraButton *view_;
    ExtraButton *edit_;
    QLineEdit *recordDelimiter_;
    QLineEdit *fieldDelimiter_;
    QCheckBox *doubleQuotes_;
    QCheckBox *filter_;
    QCheckBox *overwriteRecords_;
};

ImExportDialog::ImExportDialog(Project *pr, Table *tab, GuiObject sfw,
        ImExportAction action) : CommonDialog(GuiFindParentWindow(0, pr))
{
    pr_ = pr;
    tab_ = tab;
    sfw_ = sfw;
    action_ = action;
    view_ = 0;
    edit_ = 0;
    filter_ = 0;
    overwriteRecords_ = 0;

    GroupLayout *top = new GroupLayout(2);
    top->setContentsMargins(0, 0, 0, 0);
    contentsLayout()->addLayout(top);

    filename_ = new QLineEdit(this);
    filename_->setText(ImExportSettings.ies_Filename);

    const char *filenameBH = NULL;
    const char *filenameLabelStr = NULL;
    const char *imExportStr = NULL;
    const char *dialogTitleStr = NULL;
    const char *dialogHelpNode = NULL;
    bool hasView = false;
    bool saveMode = false;
    switch (action)
    {
        case IEA_Import:
            filenameLabelStr = GCS(MSG_Import_File);
            filenameBH = GCS(MSG_Import_FileBH);
            imExportStr = GCS(MSG_Import);
            dialogTitleStr = GCS(MSG_Import);
            dialogHelpNode = "Importing records";
            hasView = true;
            break;
#ifdef ENABLE_LOG_APPLY_CHANGES
        case IEA_ApplyChanges:
            filenameLabelStr = GCS(MSG_Import_LogFile);
            filenameBH = GCS(MSG_Import_LogFileBH);
            imExportStr = GCS(MSG_Import);
            dialogTitleStr = GCS(MSG_ApplyChanges);
            dialogHelpNode = "Applying changes";
            hasView = true;
            if (filename_->text().isEmpty()) {
                filename_->setText(LogFilename(ProjectLog(pr)));
            }
            break;
#endif
        case IEA_Export:
            filenameLabelStr = GCS(MSG_Export_File);
            filenameBH = GCS(MSG_Export_Records_FileBH);
            imExportStr = GCS(MSG_Export);
            dialogTitleStr = GCS(MSG_Export);
            dialogHelpNode = "Exporting records";
            saveMode = true;
            break;
        case IEA_ExportSFW:
            filenameLabelStr = GCS(MSG_Export_File);
            filenameBH = GCS(MSG_Export_SFWList_FileBH);
            imExportStr = GCS(MSG_Export);
            dialogTitleStr = GCS(MSG_Export);
            dialogHelpNode = "Exporting a list";
            saveMode = true;
            break;
    }
    filename_->setToolTip(filenameBH);

    PopSelection *popFile = MakeFilePopup(filename_, pr, this);
    popFile->setSaveMode(saveMode);
    popFile->setDefaultSuffix("txt");
    top->addLabelAndWidget(GuiMarkShortcut(filenameLabelStr),
        filename_, popFile);

    if (hasView) {
        view_ = new ExtraButton(this, filename_);
        view_->setIcon(style()->standardIcon(
                    QStyle::SP_FileDialogContentsView));    
	view_->setToolTip(GCS(MSG_ViewBH));
        connect(view_, SIGNAL(clicked()), this, SLOT(onViewClicked()));
        popFile->addWidget(view_);

        edit_ = new ExtraButton(this, filename_);
        edit_->setIcon(style()->standardIcon(
                    QStyle::SP_FileDialogDetailedView));
	edit_->setToolTip(GCS(MSG_EditBH));
        connect(edit_, SIGNAL(clicked()), this, SLOT(onEditClicked()));
	popFile->addWidget(edit_);
    }

    recordDelimiter_ = new QLineEdit(this);
    recordDelimiter_->setMaxLength(DELIMITER_LEN);
    int maxWidth = recordDelimiter_->fontMetrics().maxWidth() * DELIMITER_LEN/2;
    recordDelimiter_->setFixedWidth(maxWidth);
    recordDelimiter_->setText(ImExportSettings.ies_RecordDelimiter);
    recordDelimiter_->setToolTip(GCS(MSG_ImExport_RecordDelimiterBH));
    QHBoxLayout *hbox = new QHBoxLayout();
    hbox->addWidget(recordDelimiter_);
    hbox->addStretch(1);
    top->addLabelAndWidget(
        GuiMarkShortcut(GCS(MSG_ImExport_RecordDelimiter)),
        recordDelimiter_, hbox);

    fieldDelimiter_ = new QLineEdit(this);
    fieldDelimiter_->setMaxLength(DELIMITER_LEN);
    fieldDelimiter_->setFixedWidth(maxWidth);
    fieldDelimiter_->setText(ImExportSettings.ies_FieldDelimiter);
    fieldDelimiter_->setToolTip(GCS(MSG_ImExport_FieldDelimiterBH));
    hbox = new QHBoxLayout();
    hbox->addWidget(fieldDelimiter_);
    hbox->addStretch(1);
    top->addLabelAndWidget(
        GuiMarkShortcut(GCS(MSG_ImExport_FieldDelimiter)),
        fieldDelimiter_, hbox);

    QHBoxLayout *checkBoxes = new QHBoxLayout();
    top->addLayout(checkBoxes, 2);

    doubleQuotes_ = new QCheckBox(
            GuiMarkShortcut(GCS(MSG_ImExport_DoubleQuotes)), this);
    doubleQuotes_->setChecked(ImExportSettings.ies_DoubleQuotes);
    doubleQuotes_->setToolTip(GCS(MSG_ImExport_DoubleQuotesBH));
    checkBoxes->addWidget(doubleQuotes_);

    switch (action)
    {
        case IEA_Import:
            overwriteRecords_ = new QCheckBox(
                    GuiMarkShortcut(GCS(MSG_Import_OverwriteRecords)), this);
            overwriteRecords_->setChecked(
                    ImExportSettings.ies_OverwriteRecords);
            overwriteRecords_->setToolTip(GCS(MSG_Import_OverwriteRecordsBH));
            checkBoxes->addWidget(overwriteRecords_);
            break;
        case IEA_Export:
            filter_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_Filter)), this);
            filter_->setChecked(ImExportSettings.ies_Filter);
            filter_->setToolTip(GCS(MSG_Export_FilterBH));
            checkBoxes->addWidget(filter_);
            break;
        default:
            break;
    }

    setOkText(GuiMarkShortcut(imExportStr));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

    setWindowTitle(dialogTitleStr);
    setSleepProject(pr_);
    GuiSetWindowProject(this, pr);
    GuiSetHelpNode(this, dialogHelpNode);
}

static void createImExport(Project *pr, Table *tab, GuiObject sfw,
        ImExportAction action)
{
    ImExportDialog *dialog = new ImExportDialog(pr, tab, sfw, action);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }
}

void GuiImportRecords(Table *tab)
{
    createImExport(TableProject(tab), tab, 0, IEA_Import);
}

#ifdef ENABLE_LOG_APPLY_CHANGES
void GuiApplyChanges(struct Project *pr)
{
    createImExport(pr, 0, 0, IEA_ApplyChanges);
}
#endif

void GuiExportRecords(Table *tab)
{
    createImExport(TableProject(tab), tab, 0, IEA_Export);
}

void GuiExportSFWList(struct Project *pr, GuiObject sfw)
{
    createImExport(pr, 0, sfw, IEA_ExportSFW);
}

#include "qtImExport.moc"

