//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtMenus.cc
//
// Creation date: 19.05.2010
// Last modified: 02.05.2022
//

#include "system.h"
#include "qt/qtMenus.h"
#include "qt/qtUtil.h"
#include <QMenuBar>
#include <QApplication>
#include "menuproject.h"
#include "menuprefs.h"
#include "menulog.h"
#include "menutable.h"
#include "menuprogram.h"
#include "menuhelp.h"
#include "menu.h"
#include "main.h"
#include "structureeditor.h"
#include "log.h"
#include "lang/program.h"
#include "status.h"
#include "preferences.h"
#include "appstrings_cat.h"

struct MBNewMenu;

QMenu *TableMenu = 0;
static unsigned long TableStatus = ~0;

#ifdef MACOSX
QMenuBar *MenuBar = 0;
static unsigned long MenuStatus = ~0;
#else
static bool GlobalMenubar = false;
#endif

// special variables for callbacks

static ulong memtable[] = 
{ 1000000, 2000000, 4000000, 8000000, 16000000, 32000000, 0x7fffffff };

static ulong menuItemChecked = false;

static void callchecked(Project *pr, intptr_t funcptr, GuiObject win)
{
    void (*func)(Project *, intptr_t, GuiObject) = 
	    (void (*)(Project *, intptr_t, GuiObject))funcptr;
    if (func) {
	func(pr, menuItemChecked, win);
    }
}

static void saveprojectinit(Project *pr, QAction *item, intptr_t dummy)
{
    if (!ProjectAdminMode(pr) && ProjectChanges(pr) == 0) {
        item->setEnabled(false);
    }
}

#ifdef ENABLE_ADMIN_MODE
static void adminmodeinit(Project *pr, QAction *item, intptr_t dummy)
{
    bool admin = ProjectAdminMode(pr);
    const char *text = GCS(admin? MSG_MenuItem_UserMode : MSG_MenuItem_AdminMode);    
    item->setText(GuiMarkShortcut(text));
}
#endif

#ifdef ENABLE_STRUCTURE_EDITOR
static void structureeditorinit(Project *pr, QAction *item, intptr_t dummy)
{
    GuiObject se = ProjectStructureEditor(pr);
    const char *text = GCS(se? MSG_MenuItem_CloseStructureEditor : 
		      MSG_MenuItem_StructureEditor);
    item->setText(GuiMarkShortcut(text));
}
#endif

static void reverttosavedinit(Project *pr, QAction *item, intptr_t dummy)
{
    bool hasChanged = ProjectChanges(pr) != 0;
    const char *title = GCS(hasChanged? MSG_MenuItem_RevertToSaved : 
                            MSG_MenuItem_Reload);
    item->setText(GuiMarkShortcut(title));
}

static void confirmdeleterecordinit(Project *pr, QAction *item, intptr_t dummy)
{
    item->setChecked(GetPrefsValue(pr, PREFS_CONFIRMDELETERECORD));
}

#ifdef ENABLE_STRUCTURE_EDITOR
static void sorttriggerfunctionsinit(Project *pr, QAction *item, 
	intptr_t dummy)
{
    item->setChecked(GetPrefsValue(pr, PREFS_PROGRAMSORTTRIGGERFUNCTIONS));
}
#endif

static void recordmeminit(Project *pr, QAction *item, intptr_t num)
{
    item->setChecked((ulong)GetPrefsValue(pr, PREFS_RECORDMEMORY) == memtable[num]);
}

static void extrabuttonsintabchaininit(Project *pr, QAction *item, intptr_t dummy)
{
    item->setChecked(GetPrefsValue(pr, PREFS_EXTRABUTTONSINTABCHAIN));
}

static void advanceonenterinit(Project *pr, QAction *item, intptr_t dummy)
{
    item->setChecked(GetPrefsValue(pr, PREFS_ADVANCEONENTER));
}

static void confirmautoreloadinit(Project *pr, QAction *item, intptr_t dummy)
{
    item->setChecked(GetPrefsValue(pr, PREFS_CONFIRMAUTORELOAD));
}

static void confirmsavereorginit(Project *pr, QAction *item, intptr_t dummy)
{
    item->setChecked(GetPrefsValue(pr, PREFS_CONFIRMSAVEREORG));
}

static void vacuumafterreorginit(Project *pr, QAction *item, intptr_t dummy)
{
    item->setChecked(GetPrefsValue(pr, PREFS_VACUUMAFTERREORG));
}

static void confirmquitinit(Project *pr, QAction *item, intptr_t dummy)
{
    item->setChecked(GetPrefsValue(pr, PREFS_CONFIRMQUIT));
}

static void pathsrelativetoprojectinit(Project *pr, QAction *item, intptr_t dummy)
{
    item->setChecked(GetPrefsValue(pr, PREFS_PATHSRELATIVETOPROJECT));
}

static void logactiveinit(Project *pr, QAction *item, intptr_t dummy)
{
    item->setChecked(LogActive(ProjectLog(pr)));
}

static void logmodeinit(Project *pr, QAction *item, intptr_t num)
{
    item->setChecked(MenuLogModeInit(pr, num));
}

static void logsetlabelinit(Project *pr, QAction *item, intptr_t dummy)
{
    if (FindUserFunctionName(ProjectProgram(pr), logLabel)) {
        item->setEnabled(false);
    }
}

#ifdef ENABLE_PROGRAM_EDITOR
static void programsourceinit(Project *pr, QAction *item, intptr_t ext)
{
    item->setChecked((intptr_t)GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL) == ext);
}

static void cleanupexternalsourceinit(Project *pr, QAction *item, intptr_t dummy)
{
    item->setChecked(GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL_CLEANUP));
}

static void programdebuginfoinit(Project *pr, QAction *item, intptr_t dummy)
{
    item->setChecked(GetPrefsValue(pr, PREFS_PROGRAMDEBUGINFO));
}

static void programobsoletefunctionsinit(Project *pr, QAction *item, intptr_t num)
{
    item->setChecked((intptr_t)GetPrefsValue(pr, PREFS_PROGRAMOBSOLETEFUNCTIONS) == num);
}
#endif

static void projectopen(Project *pr)
{
    MenuProjectOpen(pr, 0);
}

static void recordmem(Project *pr, intptr_t num, GuiObject win)
{
    SetPrefsValue(pr, PREFS_RECORDMEMORY, (long)memtable[num]);
}

#include "menu.share"

#define NUM_MENU_ITEMS (sizeof(MBNewMenu) / sizeof(MBNewMenu[0]))
#define NUM_TABLE_MENU_ITEMS \
		(sizeof(MBTableNewMenu) / sizeof(MBTableNewMenu[0]))
#define TOTAL_MENU_ITEMS NUM_MENU_ITEMS + NUM_TABLE_MENU_ITEMS

class MenuItem : public QAction
{
    Q_OBJECT
    
public:
    MenuItem(QWidget *parent, const struct MBNewMenu *mnm) 
        : QAction(parent), _mnm(mnm) {}

    void update(Project *pr, unsigned long status);
    
public slots:
    void action(bool checked);
    
private:
    const struct MBNewMenu *_mnm;
};

void MenuItem::action(bool checked) 
{ 
    if (_mnm->mnm_Callback) {
        QWidget *win = GuiActiveWindow();
        ulong status = GetStatus(win);
        // printf("aw: %p, status: %d\n", win, status);
        if ((_mnm->mnm_Status & status) == _mnm->mnm_Status) {        
            GuiStoreEdits();
            menuItemChecked = checked;
            _mnm->mnm_Callback(GuiWindowProject(win), _mnm->mnm_UserData, win);
            return;
        }
    }
    GuiBeep();
}

void MenuItem::update(Project *pr, unsigned long status)
{
    bool enabled = (_mnm->mnm_Status & status) == _mnm->mnm_Status;
	    
    setEnabled(enabled);
    if (_mnm->mnm_Init) {
        _mnm->mnm_Init(pr, this, _mnm->mnm_UserData);
    }
}

static void updateMenu(QObject *o, Project *pr, unsigned long status)
{
    if (o) {
        MenuItem *item = dynamic_cast<MenuItem *>(o);
        if (item) {
            item->update(pr, status);
        }
        
        const QObjectList &childs = o->children();
        QObjectList::const_iterator it;
        for (it = childs.begin(); it != childs.end(); ++it) {
            updateMenu(*it, pr, status);
        }
    }
}

void UpdateMenu(GuiObject win, bool force)
{
    unsigned long status = GetStatus(win);
    Project *pr = GuiWindowProject(win);

#ifdef MACOSX
    if (status != MenuStatus || force) {
        updateMenu(MenuBar, pr, status);
        MenuStatus = status;
    }
#else
    QMainWindow *main = dynamic_cast<QMainWindow *>(win);
    if (main) {
        if (main->property("status").toInt() != (int)status || force) {
            updateMenu(main->menuBar(), pr, status);
            main->setProperty("status", (int)status);
        }
    }
#endif
    if (status != TableStatus || force) {
        updateMenu(TableMenu, pr, status);
        TableStatus = status;
    }
}

static QMenu *createMenus(QMenuBar *menuBar, const struct MBNewMenu *mnm,
        Qt::ShortcutContext context)
{
    QMenu *menu = 0, *sub = 0;
    QAction *item = 0;
    QActionGroup *group = 0;
        
    for (; mnm->mnm_Type != MNM_END; mnm++) {
	switch (mnm->mnm_Type) {
	    case MNM_TITLE:
                menu = new QMenu(menuBar);
                if (menuBar) {
                    menu->setTitle(GuiMarkShortcut(mnm->mnm_Label));
                    menuBar->addMenu(menu);
                }
                GuiSetHelpNode(menu, mnm->mnm_HelpNode);
		sub = 0;
		break;

	    case MNM_ITEM:
	    case MNM_SUB:
	    {
		if (!menu) {
		    return 0;		// wrong specification
		}

		if (mnm->mnm_Type == MNM_ITEM) {
		    sub = 0;
                } else if (!sub) {
		    sub = new QMenu(menu);
                    item->setMenu(sub);
		}

                item = new MenuItem(menu, mnm);
		const char *label = mnm->mnm_Label;
		if (label == MNM_BARLABEL) {
		    item->setSeparator(true);
		} else {
                    item->setText(GuiMarkShortcut(label));
                }
		bool checkable = ((mnm->mnm_Flags & MNM_CHECKIT) != 0);
		bool radio = (mnm->mnm_MutualExclude != 0);
                if (!radio) {
                    group = 0;
                } else if (!group) {
                    group = new QActionGroup(menu);
                    group->setExclusive(true);
                }
                item->setCheckable(checkable);
	        if (group) {
                    group->addAction(item);
                }
                if (mnm->mnm_Status == STATUS_NEVER) {
                    QFont font = item->font();
                    font.setItalic(true);
                    item->setFont(font);
                }
		GuiSetHelpNode(item, mnm->mnm_HelpNode);

		const char *sc = mnm->mnm_Shortcut;
		if (sc) {
                    int c = (unsigned char)sc[0] | Qt::CTRL;

		    if (c && sc[1]) {
			CatalogString id = mnm->mnm_ShortcutId;
			if (id) {
			    if (!strcmp(id, MSG_Help)) {
				c = Qt::Key_Help;
			    } else if (!strcmp(id, MSG_F1)) {
				c = Qt::Key_F1;
			    } else if (!strcmp(id, MSG_AltUp)) {
				c = Qt::Key_Up | Qt::ALT;
			    } else if (!strcmp(id, MSG_AltDown)) {
				c = Qt::Key_Down | Qt::ALT;
			    } else if (!strcmp(id, MSG_ShiftAltUp)) {
				c = Qt::Key_Up | Qt::ALT | Qt::SHIFT;
			    } else if (!strcmp(id, MSG_ShiftAltDown)) {
				c = Qt::Key_Down | Qt::ALT | Qt::SHIFT;
			    } else if (!strcmp(id, MSG_ShiftCtrlUp)) {
#ifdef MACOSX
				c = Qt::Key_Up | Qt::ALT | Qt::META;
#else
				c = Qt::Key_Up | Qt::ALT | Qt::CTRL;
#endif
                            } else if (!strcmp(id, MSG_ShiftCtrlDown)) {
#ifdef MACOSX
				c = Qt::Key_Down | Qt::ALT | Qt::META;
#else
				c = Qt::Key_Down | Qt::ALT | Qt::CTRL;
#endif
			    }
			}
		    }
                    item->setShortcut(QKeySequence(c));
                    item->setShortcutContext(context);
		}
#ifdef MACOSX
		if (mnm->mnm_Callback == (callback_t)MenuProjectQuit) {
                    item->setMenuRole(QAction::QuitRole);
		} else if (mnm->mnm_Callback == (callback_t)MenuHelpAbout) {
                    item->setMenuRole(QAction::AboutRole);
                } else {
                    item->setMenuRole(QAction::NoRole);
                }
#endif
                QObject::connect(item, SIGNAL(triggered(bool)), item, SLOT(action(bool)));
                if (sub) {
                    sub->addAction(item);
                } else {
                    menu->addAction(item);
                }
	    }
	}
    }
    
    return menu;
}
    
#ifndef MACOSX
void GuiWindowAddMenus(QMainWindow *win, bool main)
{
    if (win) {
        if (main || GlobalMenubar) {
            QMenuBar *menuBar = win->menuBar();
            createMenus(menuBar, MBNewMenu, Qt::WindowShortcut);
            // It seems we need to add actions once for the global menu.
            if (GlobalMenubar) {
                win->addActions(menuBar->findChildren<QAction *>());
            }
        } else {
            QWidget *root = ProjectRootWindow(GuiWindowProject(win));
            QMainWindow *rootWin = dynamic_cast<QMainWindow *>(root);
            if (rootWin) {
                win->addActions(rootWin->menuBar()->findChildren<QAction *>());
            }
        }
        if (!GlobalMenubar) {
            win->addActions(TableMenu->findChildren<QAction *>());
        }
    }
}
#endif

void GuiMenuExit(void)
{
#ifdef MACOSX
    delete MenuBar;
    MenuBar = 0;
#endif
    delete TableMenu;
    TableMenu = 0;
}

static void initializeMenus(struct MBNewMenu *mnm)
{
    for (; mnm->mnm_Type != MNM_END; mnm++) {
	if (mnm->mnm_LabelId) {
	    mnm->mnm_Label = GCS(mnm->mnm_LabelId);
       	}
	if (mnm->mnm_ShortcutId) {
	    mnm->mnm_Shortcut = GCS(mnm->mnm_ShortcutId);
	}
    }
}

#ifdef UNIX
static void hideMenus(struct MBNewMenu *mnm, ulong status)
{
    bool hide = FALSE;
    for (; mnm->mnm_Type != MNM_END; mnm++) {
        if (mnm->mnm_Type == MNM_TITLE) {
            hide = ((mnm->mnm_Status & status) != 0);
        }
        if (hide) {
            mnm->mnm_Type = MNM_IGNORE;
        }
    }
}
#endif

void GuiMenuInit(void)
{
    initializeMenus(MBNewMenu);
    initializeMenus(MBTableNewMenu);
    TableMenu = createMenus(0, MBTableNewMenu, Qt::WindowShortcut);

#ifdef MACOSX
    MenuBar = new QMenuBar(0);
    createMenus(MenuBar, MBNewMenu, Qt::ApplicationShortcut);
#endif
#ifdef UNIX
    // TODO: maybe use Qt::AA_DontUseNativeMenuBar which doesn't
    // seem to be set to false on Ubuntu 14.0.
    const char *theme = getenv("QT_QPA_PLATFORMTHEME");
    GlobalMenubar = theme != 0 && strcmp(theme, "appmenu-qt5") == 0;
    if (!GlobalMenubar) {
        hideMenus(MBNewMenu, STATUS_ACTIVETABLE);
    }
#endif
}
    
#include "qtMenus.moc"
