//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtOrder.cc
//
// Creation date: 31.10.2016
// Last modified: 17.07.2017
//

#include "order.h"
#include "qt/qtStructureEditorUtil.h"
#include "qt/qtUtil.h"
#include "qt/classes/qtGroupLayout.h"
#include <QMimeData>
#include <QStackedLayout>
#include "gui.h"
#include "project.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "appstrings_cat.h"
#include "error.h"

const char MimeTypeOrderField[] = "BeeBase/OrderField";
typedef QPair<MimeDataField, int> MimeDataOrderField;

static QList<MimeDataOrderField> mimeDataOrderFields(
        const QMimeData *data, const Table *tab, bool all)
{
    QList<MimeDataOrderField> mdoas;
    if (data->hasFormat(MimeTypeOrderField)) {
        QByteArray encodedData = data->data(MimeTypeOrderField);
        QDataStream stream(&encodedData, QIODevice::ReadOnly);
        while (!stream.atEnd()) {
            MimeDataOrderField mdoa;
            stream >> mdoa;
            if (FieldTable(mdoa.first.field()) == tab &&
                    (all || mdoa.second >= 0)) {
                mdoas.push_back(mdoa);
            }
        }
    }
    return mdoas;
}

class OrderListModel : public SortableAbstractListModel
{
    Q_OBJECT

public:
    OrderListModel(Table *tab, QWidget *parent) : 
        SortableAbstractListModel(parent), tab_(tab), parent_(parent) {}

    virtual int numElements() const { return entries_.size(); }

    Table *table() const { return tab_; }

    void set(OrderData *od) {
        beginResetModel();
        entries_.clear();
        if (od && od->od_Type == ORD_TYPE_FIELDLIST) {
            int num = od->od_Contents.fieldlist.num;
            entries_.reserve(num);
            for (int i = 0; i < num; ++i) {
                const OrderField &oa = od->od_Contents.fieldlist.oas[i];
                Entry entry;
                entry.field = oa.oa_Field;
                entry.ascending = oa.oa_Ascending;
                entries_.push_back(entry);
            }
        }
        endResetModel();
    }

    bool getOrderField(int row, OrderField *oa) const
    {
        if (oa && row >= 0 && row < entries_.size()) {
            oa->oa_Ascending = entries_[row].ascending;
            oa->oa_Field = entries_[row].field;
            return true;
        }
        return false;
    }

    void clear() { set(NULL); }

    int findField(Field *field) const {
        for (int i = 0; i < entries_.size(); ++i) {
            if (entries_[i].field == field) {
                return i;
            }
        }
        return -1;
    }

    void remove(int pos)
    {
        if (pos >= 0 && pos < entries_.size()) {
            beginRemoveRows(QModelIndex(), pos, pos);
            entries_.removeAt(pos);
            endRemoveRows();
        }
    }

    int insert(Field *field, bool ascending, int pos)
    {
        if (FieldTable(field) != tab_) {
            return -1;
        }

        // Check if field is already in order list.
        int found = findField(field);
        if (found >= 0) {
            remove(found);
            if (pos > found) {
                pos--;
            }
        }
        beginInsertRows(QModelIndex(), pos, pos);
        Entry entry;
        entry.field = field;
        entry.ascending = ascending;
        entries_.insert(pos, entry);
        endInsertRows();
        return pos;
    }

    Qt::DropActions supportedDropActions() const
    {
        return Qt::MoveAction | Qt::CopyAction;
    }
    
    QStringList mimeTypes() const
    {
        QStringList types;
        types << MimeTypeOrderField;
        return types;
    }

    QMimeData *mimeData(const QModelIndexList &indexes) const
    {
        QByteArray data;
        QDataStream stream(&data, QIODevice::WriteOnly);
        foreach (const QModelIndex &index, indexes) {
            OrderField oa;
            if (getOrderField(index.row(), &oa)){
                stream << MimeDataOrderField(
                        MimeDataField(oa.oa_Field),
                        oa.oa_Ascending? 1 : 0);
            }
        }
        QMimeData *mimeData = new QMimeData();
        mimeData->setData(MimeTypeOrderField, data);
        return mimeData;
    }

    bool canDropMimeData(const QMimeData *data, Qt::DropAction action,
            int row, int column, const QModelIndex &parent) const
    {
        if (action == Qt::IgnoreAction) {
            return true;
        }
        return !mimeDataOrderFields(data, tab_, true).isEmpty();
    }

    bool dropMimeData(const QMimeData *data, Qt::DropAction action, 
                      int row, int column, const QModelIndex &parent)
    {
        if (action == Qt::IgnoreAction) {
            return true;
        }
        if (row < 0 || row > entries_.size()) {
            row = entries_.size();
        }
        foreach (const MimeDataOrderField &mdoa,
                mimeDataOrderFields(data, tab_, true)) {
            int pos = insert(mdoa.first.field(), mdoa.second != 0, row);
            if (pos >= 0) {
                row = pos;
            }
        }
        return true;
    }

public slots:
    void toggleAscending(const QModelIndex &index)
    {
        if (index.isValid()) {
            bool &ascending = entries_[index.row()].ascending;
            ascending = !ascending;
            emit dataChanged(index, index);
        }
    }

protected:
    virtual QString display(int ix) const {
        return FieldName(entries_[ix].field);
    }
    virtual QVariant decoration(int ix) const 
    {
        bool ascending = entries_[ix].ascending;
        return parent_->style()->standardIcon(ascending?
                QStyle::SP_ArrowUp : QStyle::SP_ArrowDown, 0, parent_);
    }

    virtual bool moveElement(int from, int to)
    {
        entries_.move(from, to);
        return true;
    }
    
private:
    struct Entry
    {
        Field *field;
        bool ascending;
    };

    QList<Entry> entries_;
    Table *tab_;
    QWidget *parent_;
};

class OrderFieldsModel : public FieldsModel
{
public:
    OrderFieldsModel(QObject *parent) : FieldsModel(parent), olm_(0) {}

    void setOrderListModel(OrderListModel *olm) { olm_ = olm; }
    
    Qt::DropActions supportedDropActions() const
    {
        return Qt::MoveAction | Qt::CopyAction;
    }
    
    QStringList mimeTypes() const
    {
        QStringList types;
        types << MimeTypeOrderField;
        return types;
    }

    QMimeData *mimeData(const QModelIndexList &indexes) const
    {
        QByteArray data;
        QDataStream stream(&data, QIODevice::WriteOnly);
        foreach (const QModelIndex &index, indexes) {
            stream << MimeDataOrderField(
                    MimeDataField((Field *)node(index.row())), -1);
        }
        QMimeData *mimeData = new QMimeData();
        mimeData->setData(MimeTypeOrderField, data);
        return mimeData;
    }

    bool canDropMimeData(const QMimeData *data, Qt::DropAction action,
            int row, int column, const QModelIndex &parent) const
    {
        if (action == Qt::IgnoreAction) {
            return true;
        }
        if (!olm_) {
            return false;
        }
        return !mimeDataOrderFields(data, olm_->table(), false).isEmpty();
    }

    bool dropMimeData(const QMimeData *data, Qt::DropAction action, 
                      int row, int column, const QModelIndex &parent)
    {
        if (action == Qt::IgnoreAction) {
            return true;
        }
        if (!olm_) {
            return false;
        }
        foreach (const MimeDataOrderField &mdoa,
                mimeDataOrderFields(data, olm_->table(), false)) {
            olm_->remove(olm_->findField(mdoa.first.field()));
        }
        return true;
    }

protected:
    virtual bool filter(Node *n) const
    { 
        short type = FieldType((Field *)n);
	return (type != TYPE_BUTTON && type != TYPE_VIRTUAL &&
                type != TYPE_REFERENCE);
    }

private:
    OrderListModel *olm_;
};

class EditOrderDialog : public CommonDialog
{
    Q_OBJECT

public:
    EditOrderDialog(Table *tab, OrderData *od);

private:
    Project *pr_;
    Table *tab_;
    ComboBox *type_;
    QStackedLayout *stack_;
    QLineEdit *cmpFuncStr_;
    OrderFieldsModel *fieldsModel_;
    QListView *fieldsView_;
    OrderListModel *orderListModel_;
    QListView *orderListView_;
    QPushButton *clear_;

    virtual bool acceptEdits()
    {
	MemPool *memPool = ProjectMemPool(pr_);
	OrderField *oas = 0;
	long num = 0;
	OrderData od;
        QByteArray buf;
        long error = OK;

        error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return false;
        }
     
	od.od_Type = type_->currentIndex() != 0?
            ORD_TYPE_CMPFUNC : ORD_TYPE_FIELDLIST;
	if (od.od_Type == ORD_TYPE_CMPFUNC) {
            buf = cmpFuncStr_->text().toUtf8();
	    od.od_Contents.cmpfunc = buf.data();
	} else {
	    num = orderListModel_->numElements();
	    if (num > 0) {
		oas = (OrderField *)PoolAllocMemZero(memPool, 
			num * sizeof(OrderField));
		if (!oas) {
		    error = ERR_MEM;
		} else {
                    for (long i = 0; i < num; ++i) {
                        orderListModel_->getOrderField(i, &oas[i]);
                    }
                }
	    }
	    od.od_Contents.fieldlist.num = (short)num;
	    od.od_Contents.fieldlist.oas = oas;
	}

	if (error == OK) {
	    GuiSetSleep(TRUE);
	    error = SetNewOrder(tab_, &od);
	    GuiSetSleep(FALSE);
	}
	PoolFreeMem(memPool, oas, num * sizeof(OrderField));

	if (error) {
	    Ask2(this, GCS(MSG_ErrorMessage), AT_ERROR, 
		    GCS(MSG_StarOk), GES(error));
	    return false;
	}

	hide();

	error = ReorderAllRecords(tab_);
	if (error) {
	    Ask(AT_WARN, GCS(MSG_StarOk), GCS(MSG_Err_Order), GES(error));
	}
	GuiDisplayPanel(tab_);
	ProjectChange(pr_, false);
        return true;
    }

    void insertOrderField(Field *field, bool ascending, int pos)
    {
        pos = orderListModel_->insert(field, ascending, pos);
        if (pos >= 0) {
            orderListView_->setCurrentIndex(orderListModel_->index(pos, 0));
        }
    }

private slots:
    void onFieldActivated(const QModelIndex &index)
    {
        Field *field = fieldsModel_->field(index);
        insertOrderField(field, true, orderListModel_->numElements());
    }

    void onClearClicked()
    {
        orderListModel_->clear();
        cmpFuncStr_->clear();
    }
};

EditOrderDialog::EditOrderDialog(Table *tab, OrderData *od) :
    CommonDialog(GuiFindParentWindow(0, TableProject(tab)))
{
    pr_ = TableProject(tab);
    tab_ = tab;

    const char *cmpfunc = 0;
    short page = 0;
    if (od && od->od_Type == ORD_TYPE_CMPFUNC) {
        cmpfunc = od->od_Contents.cmpfunc;
        page = 1;
    }

    GroupLayout *top = new GroupLayout(2);
    top->setContentsMargins(0, 0, 0, 0);
    contentsLayout()->addLayout(top);

    static char *typeContents[3];
    typeContents[0] = GCS(MSG_Order_Type_FieldList);
    typeContents[1] = GCS(MSG_Order_Type_ComparisonFunction);
    typeContents[2] = NULL;
    type_ = new ComboBox(this, typeContents);
    top->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Type)), type_);
    top->setStretch(1, 0);

    stack_ = new QStackedLayout();
    connect(type_, SIGNAL(currentIndexChanged(int)), 
            stack_, SLOT(setCurrentIndex(int)));
    top->addLayout(stack_, 2);

    fieldsView_ = new QListView(this);
    fieldsModel_ = new OrderFieldsModel(fieldsView_);
    fieldsModel_->setList(TableFields(tab_));
    fieldsView_->setModel(fieldsModel_);
    fieldsView_->setSelectionMode(QAbstractItemView::SingleSelection);
    fieldsView_->setDragDropMode(QAbstractItemView::DragDrop);
    fieldsView_->setToolTip(GCS(MSG_Order_FieldsBH));
    connect(fieldsView_, SIGNAL(activated(const QModelIndex &)),
            this, SLOT(onFieldActivated(const QModelIndex &)));

    GroupBox *fieldsFrame = new GroupBox(GCS(MSG_Fields), this);
    GroupLayout *fieldsGroup = new GroupLayout(1, fieldsFrame);
    fieldsGroup->addWidget(fieldsView_);

    orderListView_ = new QListView(this);
    orderListModel_ = new OrderListModel(tab_, orderListView_);
    orderListModel_->set(od);
    fieldsModel_->setOrderListModel(orderListModel_);
    orderListView_->setModel(orderListModel_);
    orderListView_->setSelectionMode(QAbstractItemView::SingleSelection);
    orderListView_->setDragDropMode(QAbstractItemView::DragDrop);
    orderListView_->setToolTip(GCS(MSG_Order_OrderBH));
    connect(orderListView_, SIGNAL(activated(const QModelIndex &)),
            orderListModel_, SLOT(toggleAscending(const QModelIndex &)));

    GroupBox *orderFrame = new GroupBox(GCS(MSG_Order_Order));
    GroupLayout *orderGroup = new GroupLayout(1, orderFrame);
    orderGroup->addWidget(orderListView_);

    QWidget *fieldOrder = new QWidget(this);
    stack_->addWidget(fieldOrder);

    GroupLayout *fieldOrderGroup = new GroupLayout(2, fieldOrder);
    fieldOrderGroup->setContentsMargins(0, 0, 0, 0);
    fieldOrderGroup->addWidget(fieldsFrame);
    fieldOrderGroup->addWidget(orderFrame);

    cmpFuncStr_ = new QLineEdit(this);
    cmpFuncStr_->setMaxLength(TABLE_TRIGGER_LEN);
    cmpFuncStr_->setText(cmpfunc);
    cmpFuncStr_->setToolTip(GCS(MSG_Order_ComparisonFunctionBH));

    QFrame *cmpFunc = new QFrame(this);
    cmpFunc->setFrameShape(QFrame::StyledPanel);
    cmpFunc->setFrameShadow(QFrame::Raised);
    stack_->addWidget(cmpFunc);

    GroupLayout *cmpFuncGroup = new GroupLayout(1, cmpFunc);
    cmpFuncGroup->addWidget(new QFrame(this));
    cmpFuncGroup->setStretch(0, 1);
    cmpFuncGroup->addLabelAndWidget(
            GuiMarkShortcut(GCS(MSG_Order_ComparisonFunction)),
            cmpFuncStr_, UserFunctionsPopup(cmpFuncStr_, pr_, this));
    cmpFuncGroup->addWidget(new QFrame(this));
    cmpFuncGroup->setStretch(0, 1);

    setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

    clear_ = new QPushButton(GuiMarkShortcut(GCS(MSG_Clear)), this);
    clear_->setToolTip(GCS(MSG_Order_ClearBH));
    buttonBox()->addButton(clear_, QDialogButtonBox::ResetRole);
    connect(clear_, SIGNAL(clicked()), this, SLOT(onClearClicked()));

    setWindowTitle(
            QString("%1: %2").arg(GCS(MSG_ChangeOrder)).arg(TableName(tab)));

    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Changing orders");

    type_->setCurrentIndex(page);
}

void GuiChangeOrder(Table *tab, OrderData *od)
{
    EditOrderDialog *dialog = new EditOrderDialog(tab, od);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }
}

#include "qtOrder.moc"
