//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtPDFExport.cc
//
// Creation date: 26.02.2017
// Last modified: 06.07.2021
//

#include "system.h"
#include "pdfexport.h"
#include "qt/qtUtil.h"
#include "qt/qtPrint.h"
#include "qt/classes/qtPopSelection.h"
#include "qt/classes/qtGroupLayout.h"
#include <QDir>
#include <QFileInfo>
#include <QLabel>
#include <QPrinter>
#include <QTemporaryFile>
#include "guidosutil.h"
#include "preferences.h"
#include "util.h"
#include "requesters.h"
#include "appstrings_cat.h"

struct QtPDFSettings
{
    QPrinter::PaperSize qpdfs_PaperSize;
    QPrinter::Orientation qpdfs_Orientation;
    QString qpdfs_FontName;
    QString qpdfs_Header;
    QString qpdfs_Footer;
};

static struct
{
    QtPDFSettings pes_QtPDFSettings;
    QString pes_Filename;
} PDFExportSettings;

#define QTPDF_DEFAULT_PAPER_SIZE QPrinter::A4
#define QTPDF_DEFAULT_ORIENTATION QPrinter::Portrait

static const QPrinter::PaperSize paperSizes[] =
{
    QPrinter::Letter,
    QPrinter::Legal,
    QPrinter::Executive,
    QPrinter::A3,
    QPrinter::A4,
    QPrinter::A5,
    QPrinter::B4,
    QPrinter::B5,
};

static const char *paperNames[] =
{
    "Letter",
    "Legal",
    "Executive",
    "A3",
    "A4",
    "A5",
    "B4",
    "B5",
    0
};

static const QPrinter::Orientation orientations[] =
{
    QPrinter::Portrait,
    QPrinter::Landscape,
};

static const char *orientationNames[] =
{
    "Portrait",
    "Landscape",
    0
};

static void loadQtPDFSettingsFromPrefs(QtPDFSettings *qpdfs, Project *pr)
{
    if (qpdfs != 0) {
        qpdfs->qpdfs_PaperSize = QTPDF_DEFAULT_PAPER_SIZE;
        const char *paperSize = GetPrefsString(pr, PREFS_PDFSETTINGS_PAPERSIZE);
        if (paperSize != 0) {
            for (int i = 0; paperNames[i] != 0; i++) {
                if (strcasecmp(paperSize, paperNames[i]) == 0) {
                    qpdfs->qpdfs_PaperSize = paperSizes[i];
                    break;
                }
            }
        }

        qpdfs->qpdfs_Orientation = QTPDF_DEFAULT_ORIENTATION;
        const char *orientation =
            GetPrefsString(pr, PREFS_PDFSETTINGS_ORIENTATION);
        if (orientation != 0) {
            for (int i = 0; orientationNames[i] != 0; i++) {
                if (strcasecmp(orientation, orientationNames[i]) == 0) {
                    qpdfs->qpdfs_Orientation = orientations[i];
                    break;
                }
            }
        }

        qpdfs->qpdfs_FontName = 
            GetPrefsString(pr, PREFS_QTPDFSETTINGS_FONTNAME);
        qpdfs->qpdfs_Header = GetPrefsString(pr, PREFS_PDFSETTINGS_HEADER);
        qpdfs->qpdfs_Footer = GetPrefsString(pr, PREFS_PDFSETTINGS_FOOTER);
    }
}

static void saveQtPDFSettingsToPrefs(const QtPDFSettings *qpdfs, Project *pr)
{
    if (qpdfs != 0) {
        const char *paperSize = 0;
        if (qpdfs->qpdfs_PaperSize != QTPDF_DEFAULT_PAPER_SIZE) {
            for (int i = 0; paperNames[i] != 0; i++) {
                if (qpdfs->qpdfs_PaperSize == paperSizes[i]) {
                    paperSize = paperNames[i];
                    break;
                }
            }
        }
        SetPrefsString(pr, PREFS_PDFSETTINGS_PAPERSIZE, paperSize);

        const char *orientation = 0;
        if (qpdfs->qpdfs_Orientation != QTPDF_DEFAULT_ORIENTATION) {
            for (int i = 0; orientationNames[i] != 0; i++) {
                if (qpdfs->qpdfs_Orientation == orientations[i]) {
                    orientation = orientationNames[i];
                    break;
                }
            }
        }
        SetPrefsString(pr, PREFS_PDFSETTINGS_ORIENTATION, orientation);

        SetPrefsString(pr, PREFS_QTPDFSETTINGS_FONTNAME,
                qpdfs->qpdfs_FontName.toUtf8().data());
        SetPrefsString(pr, PREFS_PDFSETTINGS_HEADER,
            qpdfs->qpdfs_Header.toUtf8().data());
        SetPrefsString(pr, PREFS_PDFSETTINGS_FOOTER,
            qpdfs->qpdfs_Footer.toUtf8().data());
    }
}

class PDFExportDialog : public CommonDialog
{
    Q_OBJECT

public:
    PDFExportDialog(Project *pr, GuiObject sfw);

private slots:
    void updateStatus()
    {
        savePDFExportSettings();
        setupPrinter(tempFile_.fileName());
        const QtPDFSettings &qpdfs = PDFExportSettings.pes_QtPDFSettings;
        long pages;
        long truncated;
        const char *error = SFWViewPrinterStats(sfw_, &printer_, &font_, 
                qpdfs.qpdfs_Header, qpdfs.qpdfs_Footer, &pages, &truncated);
        QString text;
        if (error) {
            text = QString("<font #FF0000>%1</font>").arg(error);
        } else if (truncated > 0.0) {
            QString status;
            status.sprintf(GCS(MSG_PDFExport_Status_TooWide), pages,
                    (long)ceil(truncated));
            text = QString("<font #FF0000>%1</font>").arg(status);
        } else {
            text.sprintf(GCS(MSG_PDFExport_Status_FitsPageWidth), pages);
        }
        status_->setText(text);
    }

private:
    Project *pr_;
    GuiObject sfw_;
    QLineEdit *filename_;
    QComboBox *paper_;
    QComboBox *orientation_;
    QLineEdit *fontName_;
    QLineEdit *header_;
    QLineEdit *footer_;
    QLabel *status_;
    QTemporaryFile tempFile_;
    QPrinter printer_;
    QFont font_;

    void savePDFExportSettings()
    {
        QtPDFSettings &qpdfs = PDFExportSettings.pes_QtPDFSettings;
        int active = paper_->currentIndex();
        qpdfs.qpdfs_PaperSize = paperSizes[active];
        active = orientation_->currentIndex();
        qpdfs.qpdfs_Orientation = orientations[active];
        qpdfs.qpdfs_FontName = fontName_->text();
        qpdfs.qpdfs_Header = header_->text();
        qpdfs.qpdfs_Footer = footer_->text();
        saveQtPDFSettingsToPrefs(&qpdfs, pr_);
        loadQtPDFSettingsFromPrefs(&qpdfs, pr_);
        PDFExportSettings.pes_Filename = filename_->text();
    }

    void setupPrinter(const QString &filename)
    {
        const QtPDFSettings &qpdfs = PDFExportSettings.pes_QtPDFSettings;
        printer_.setCreator(APP_NAME);
        printer_.setOutputFileName(filename);
        printer_.setPaperSize(qpdfs.qpdfs_PaperSize);
        printer_.setOrientation(qpdfs.qpdfs_Orientation);
        printer_.setFontEmbeddingEnabled(false);
        font_ = qtFontFromDescription(qpdfs.qpdfs_FontName);
    }

    bool acceptEdits()
    {
        savePDFExportSettings();
        if (PDFExportSettings.pes_Filename.isEmpty()) {
            setWindowTitle(GCS(MSG_Err_NoFilename));
            filename_->setFocus();
            GuiBeep();
            return false;
        }

        ProjectSetCurrentDir(pr_);
        QFileInfo fileInfo(PDFExportSettings.pes_Filename);
        if (fileInfo.exists()) {
            QByteArray buffer = PDFExportSettings.pes_Filename.toUtf8();
            const char *filename = buffer.data();
            if (!Ask2(this, 0, AT_WARN, GCS(MSG_AskReplaceAnswers),
                        GCS(MSG_AskReplaceFile), filename)) {
                return false;
            }
        }

        setupPrinter(PDFExportSettings.pes_Filename);
        const QtPDFSettings &qpdfs = PDFExportSettings.pes_QtPDFSettings;
        const char *error = SFWViewPrint(sfw_, &printer_, &font_,
                qpdfs.qpdfs_Header, qpdfs.qpdfs_Footer);
        QByteArray buffer = PDFExportSettings.pes_Filename.toUtf8();
        const char *filename = buffer.constData();
        if (error) {
            Ask2(this, 0, AT_ERROR, GCS(MSG_Cancel), error, filename);
            return false;
        }

        GuiViewFile(filename, true);

        return true;
    }
};


PDFExportDialog::PDFExportDialog(Project *pr, GuiObject sfw) : 
    CommonDialog(GuiFindParentWindow(sfw, pr)),
    tempFile_(QDir::temp().filePath(APP_NAME "-output-XXXXXX.pdf")),
    printer_(QPrinter::HighResolution)
{
    QtPDFSettings &qpdfs = PDFExportSettings.pes_QtPDFSettings;
    loadQtPDFSettingsFromPrefs(&qpdfs, pr);

    pr_ = pr;
    sfw_ = sfw;

    // Create the temporary filename.
    tempFile_.open();
    tempFile_.close();

    GroupLayout *group = new GroupLayout(2);
    group->setContentsMargins(0, 0, 0, 0);
    contentsLayout()->addLayout(group);

    filename_ = new QLineEdit(this);
    filename_->setText(PDFExportSettings.pes_Filename);
    filename_->setToolTip(GCS(MSG_PDFExport_FileBH));
    PopSelection *popFile = MakeFilePopup(filename_, pr, this);
    popFile->setSaveMode(true);
    popFile->setDefaultSuffix("pdf");
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_PDFExport_File)),
            filename_, popFile);

    paper_ = new ComboBox(this, (char **)paperNames);
    for (int i = 0; paperNames[i] != 0; i++) {
        if (paperSizes[i] == qpdfs.qpdfs_PaperSize) {
            paper_->setCurrentIndex(i);
            break;
        }
    }
    paper_->setToolTip(GCS(MSG_PDFExport_PaperBH));
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_PDFExport_Paper)), paper_);

    static char *orientationNames[3];
    orientationNames[0] = GCS(MSG_PDFExport_Orientation_Portrait);
    orientationNames[1] = GCS(MSG_PDFExport_Orientation_Landscape);
    orientationNames[2] = 0;
    orientation_ = new ComboBox(this, orientationNames);
    for (int i=0; orientationNames[i] != 0; i++) {
        if (orientations[i] == qpdfs.qpdfs_Orientation) {
            orientation_->setCurrentIndex(i);
            break;
        }
    }
    orientation_->setToolTip(GCS(MSG_PDFExport_OrientationBH));
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_PDFExport_Orientation)),
        orientation_);

    fontName_ = new QLineEdit(this);
    fontName_->setText(qpdfs.qpdfs_FontName);
    fontName_->setToolTip(GCS(MSG_FontBH));
    PopSelection *popFont = MakeFontPopup(fontName_, this);
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_Font)), fontName_,popFont);

    header_ = new QLineEdit(this);
    header_->setText(qpdfs.qpdfs_Header);
    header_->setToolTip(GCS(MSG_PDFExport_HeaderBH));
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_PDFExport_Header)),
            header_);

    footer_ = new QLineEdit(this);
    footer_->setText(qpdfs.qpdfs_Footer);
    footer_->setToolTip(GCS(MSG_PDFExport_FooterBH));
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_PDFExport_Footer)),
            footer_);

    status_ = new QLabel(this);
    status_->setText("");
    status_->setFrameShadow(QFrame::Sunken);
    status_->setFrameShape(QFrame::StyledPanel);
    status_->setMargin(1);
    status_->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_PDFExport_Status)),
            status_);

    connect(paper_, SIGNAL(currentIndexChanged(int)),
            this, SLOT(updateStatus()));
    connect(orientation_, SIGNAL(currentIndexChanged(int)),
            this, SLOT(updateStatus()));
    connect(fontName_, SIGNAL(textChanged(const QString &)),
            this, SLOT(updateStatus()));
    connect(header_, SIGNAL(textChanged(const QString &)),
            this, SLOT(updateStatus()));
    connect(footer_, SIGNAL(textChanged(const QString &)),
            this, SLOT(updateStatus()));
    updateStatus();

    setOkText(GuiMarkShortcut(GCS(MSG_PDFExport_CreatePDF)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

    setWindowTitle(GCS(MSG_PDFExport));
    setSleepProject(pr_);
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Exporting a list as PDF");
}

void GuiExportSFWListAsPDF(struct Project *pr, GuiObject sfw)
{
    PDFExportDialog *dialog = new PDFExportDialog(pr, sfw);
    if (!GuiOpenWindow(dialog)) {
        delete dialog;
    }
}

#include "qtPDFExport.moc"
