//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtPrint.cc
//
// Creation date: 21.02.2017
// Last modified: 29.01.2022
//

#include "system.h"
#include "qt/qtUtil.h"
#include "qt/qtPrint.h"
#include "qt/classes/qtSFWView.h"
#include "qt/classes/qtGroupLayout.h"
#include <QPainter>
#include <QPrintPreviewDialog>
#include <QPrinter>
#include "print.h"
#include "requesters.h"
#include "preferences.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"

#define ALWAYS_SHOW_TITLES TRUE

struct SFWViewPrintDimensions
{
    int pageWidth;
    int pageHeight;
    int gapWidth;
    int ascent;
    int descent;
    int lineSpacing;
    int headerLines;
    bool separateEntries;
    QVector<int> fieldWidths;
    int totalWidth;
    int linesPerPage;
    int pages;
};

static int textWidth(const QFontMetrics &font, const char *s)
{
    int maxw = 0;

    while (s != NULL) {
        const char *t = CodesetStrchr(s, '\n');
        int tw = font.width(QString::fromUtf8(s, (t == NULL)? -1 : t - s));
        if (tw > maxw) {
            maxw = tw;
        }
        s = (t == NULL)? NULL : CodesetNextChar(t);
    }

    return maxw;
}

static const char *sfwViewGetPrintDimensions(GuiObject sfw, QPainter *painter,
        const QString &header, const QString &footer,
        SFWViewPrintDimensions *dimensions)
{
    SFWView *sfwView = dynamic_cast<SFWView *>(sfw);
    if (!sfwView) {
        return GCS(MSG_Err_Internal);
    }

    // Page dimensions.
    QPaintDevice *device = painter->device();
    dimensions->pageWidth = device->width();
    dimensions->pageHeight = device->height();
    
    // Obtain font metrics.
    QFont font(painter->font());
    font.setBold(false);
    QFont bold(font);
    bold.setBold(true);
    painter->setFont(bold);
    QFontMetrics boldMetrics = painter->fontMetrics();
    painter->setFont(font);
    QFontMetrics fontMetrics = painter->fontMetrics();

    // Font dimensions.
    dimensions->gapWidth = fontMetrics.maxWidth();
    dimensions->ascent = fontMetrics.ascent();
    dimensions->descent = fontMetrics.descent();
    dimensions->lineSpacing = fontMetrics.lineSpacing();

    // Number of lines.
    dimensions->headerLines = 0;
    if (ALWAYS_SHOW_TITLES || sfwView->showTitles()) {
        dimensions->headerLines =
            QueryLineLines(sfwView->queryLine(0), sfwView->numFields());
    }
    dimensions->separateEntries = false;
    QAbstractItemModel *model = sfwView->model();
    int count = model? model->rowCount() : 0;
    int bodyLines = 0;
    for (int row = 1; row <= count; row++) {
        int queryLines =
            QueryLineLines(sfwView->queryLine(row), sfwView->numFields());
        if (queryLines > 1) {
            dimensions->separateEntries = true;
        }
        bodyLines += queryLines;
    }

    // Field widths.
    dimensions->fieldWidths.fill(0, sfwView->numFields());
    dimensions->totalWidth = 0;
    for (int col = 0; col < sfwView->numFields(); col++) {
        int maxw = 0;
        if (ALWAYS_SHOW_TITLES || sfwView->showTitles()) {
            const QueryLine *ql = sfwView->queryLine(0);
            if (ql) {
                maxw = textWidth(boldMetrics, ql->ql_Entries[col]->qe_String);
            }
        }

        for (int row = 1; row <= count; row++) {
            const QueryLine *ql = sfwView->queryLine(row);
            if (ql) {
                int tw = textWidth(fontMetrics, ql->ql_Entries[col]->qe_String);
                if (tw > maxw) {
                    maxw = tw;
                }
            }
        }
        dimensions->fieldWidths[col] = maxw;
        dimensions->totalWidth += maxw;
        if (col > 0) {
            dimensions->totalWidth += dimensions->gapWidth;
        }
    }

    if (!header.isEmpty()) {
        int tw = fontMetrics.width(header);
        if (tw > dimensions->totalWidth) {
            dimensions->totalWidth = tw;
        }
    }
    if (!footer.isEmpty()) {
        int tw = fontMetrics.width(footer);
        if (tw > dimensions->totalWidth) {
            dimensions->totalWidth = tw;
        }
    }

    qreal dlines = (qreal)dimensions->pageHeight / dimensions->lineSpacing;

    if (dimensions->headerLines > 0) {
        dlines -= dimensions->headerLines + 0.5;
    }
    dlines -= 2.0;      // space for page number
    if (!header.isEmpty()) {
        dlines -= 2.0;
    }
    int lines = (int)floor(dlines);

    // Rows per page and number of pages.
    dimensions->linesPerPage = lines;
    dimensions->pages = lines < 1 ? INT_MAX : (bodyLines + lines - 1) / lines;
    if (lines < 1) {
        return GCS(MSG_PrintErr_PageTooSmall);
    }

    return 0;
}

static void drawText(QPainter *painter, int x, int y, int w, int th, int align,
        int lh, const char *s, int offset, int numLines)
{
    int end = offset + numLines;
    for (int line = 0; line < end && s != NULL; line++) {
        const char *t = CodesetStrchr(s, '\n');
        if (line >= offset) {
            painter->drawText(x, y, w, th, align,
                    QString::fromUtf8(s, (t == NULL)? -1 : t - s));
            y += lh;
        }
        s = (t == NULL)? NULL : CodesetNextChar(t);
    }
}

static void sfwViewDrawPage(GuiObject sfw, QPainter *painter,
        const QString &header, const QString &footer,
        const SFWViewPrintDimensions &dimensions, int page)
{
    SFWView *sfwView = dynamic_cast<SFWView *>(sfw);
    if (!sfwView) {
        return;
    }

    QAbstractItemModel *model = sfwView->model();
    int count = model? model->rowCount() : 0;
    const signed char *order = sfwView->getColumnOrder();
    const QueryFieldInfo *infos = sfwView->fieldInfos();
    int th = dimensions.ascent + dimensions.descent;
    int lh = dimensions.lineSpacing;

    int x = 0;
    int y = 0;
    if (!header.isEmpty()) {
        QFont font(painter->font());
        font.setBold(false);
        int w = dimensions.pageWidth;
        int align = Qt::AlignLeft;
        painter->setFont(font);
        painter->drawText(x, y, w, th, align | Qt::TextDontClip, header);
        y += lh;
        y += lh;
    }

    if (dimensions.headerLines > 0) {
        const QueryLine *ql = sfwView->queryLine(0);
        if (ql) {
            QFont bold(painter->font());
            bold.setBold(true);
            painter->setFont(bold);
            for (int i = 0; i < sfwView->numFields(); i++) {
                int col = order? order[i] : i;
                int w = dimensions.fieldWidths[col];
                int align = infos[col].qfi_Format?
                    Qt::AlignRight : Qt::AlignLeft;
                drawText(painter, x, y, w, th, align | Qt::TextDontClip, lh,
                        ql->ql_Entries[col]->qe_String,
                        0, dimensions.headerLines);
                x += w + dimensions.gapWidth;
            }
        }
        y += lh * dimensions.headerLines;
        y += lh / 4.0;
        painter->drawLine(0, y, dimensions.totalWidth, y);
        y += lh / 4.0;
    }

    QFont font(painter->font());
    font.setBold(false);
    painter->setFont(font);

    int row = 0;
    int offset = 0;
    for (int p = 0; p <= page; p++) {
        int line = 0;
        while (line < dimensions.linesPerPage && row < count) {
            const QueryLine *ql = sfwView->queryLine(1 + row);
            int linesLeftOnPage = dimensions.linesPerPage - line;
            int lines = QueryLineLines(ql, sfwView->numFields()) - offset;
            int maxLines = (lines < linesLeftOnPage)? lines : linesLeftOnPage;
            if (p == page && ql != NULL) {
                x = 0;
                for (int i = 0; i < sfwView->numFields(); i++) {
                    int col = order? order[i] : i;
                    int w = dimensions.fieldWidths[col];
                    int align = infos[col].qfi_Format?
                        Qt::AlignRight : Qt::AlignLeft;
                    drawText(painter, x, y, w, th, align | Qt::TextDontClip, lh,
                            ql->ql_Entries[col]->qe_String, offset, maxLines);
                    x += w + dimensions.gapWidth;
                }
                y += lh * maxLines;
            }
            line += maxLines;

            if (maxLines < lines) {
                offset += maxLines;
            } else {
                ++row;
                offset = 0;
                if (dimensions.separateEntries && p == page) {
                    painter->drawLine(0, y, dimensions.totalWidth, y);
                }
            }
        }
    }

    x = 0;
    y = dimensions.pageHeight - th;
    int w = dimensions.pageWidth;
    int align = Qt::AlignLeft;
    if (!footer.isEmpty()) {
        QFont font(painter->font());
        font.setBold(false);
        painter->setFont(font);
        painter->drawText(x, y, w, th, align | Qt::TextDontClip, footer);
    }

    // Page number
    char buffer[64];
    sprintf(buffer, GCS(MSG_PageFooter), (long)page + 1,(long)dimensions.pages);
    QRect rect(x, y, w, th);
    align = footer.isEmpty() ? Qt::AlignCenter : Qt::AlignRight;
    painter->drawText(x, y, w, th, align | Qt::TextDontClip, buffer);
}

class SFWViewPrintPreviewDialog : public QPrintPreviewDialog
{
    Q_OBJECT

public:
    explicit SFWViewPrintPreviewDialog(GuiObject sfw) : 
        QPrintPreviewDialog(sfw), sfw_(sfw)
    {
        connect(this, SIGNAL(paintRequested(QPrinter *)),
                this, SLOT(paint(QPrinter *)));
    }

private slots:
    void paint(QPrinter *printer)
    {
        const char *error = SFWViewPrint(sfw_, printer, 0, QString(),QString());
        if (error) {
            Ask2(this, NULL, AT_ERROR, GCS(MSG_Cancel), error, "");
        }
    }

private:
    GuiObject sfw_;
};

void GuiPrintSFWList(Project *pr, GuiObject sfw)
{
    GuiBusyCursor busy;
    SFWViewPrintPreviewDialog dialog(sfw);
    dialog.exec();
}

const char *SFWViewPrint(GuiObject sfw, QPagedPaintDevice *pager, QFont *font,
        const QString &header, const QString &footer)
{
    const char *error = 0;
    SFWViewPrintDimensions dimensions;

    QPainter painter;
    if (!painter.begin(pager)) {
        return GES(DISKERR_OPEN);
    }
    if (font) {
        painter.setFont(*font);
    }
    error = sfwViewGetPrintDimensions(sfw, &painter, header, footer,
            &dimensions);
    if (!error) {
        for (int page = 0; page < dimensions.pages; page++) {
            if (page != 0) {
                pager->newPage();
            }
            sfwViewDrawPage(sfw, &painter, header, footer, dimensions, page);
        }
    }
    painter.end();

    return error;
}

const char *SFWViewPrinterStats(GuiObject sfw, QPagedPaintDevice *pager,
        QFont *font, const QString &header, const QString &footer,
        long *pages, long *truncated)
{
    const char *error = 0;
    SFWViewPrintDimensions dimensions;

    QPainter painter(pager);
    if (font) {
        painter.setFont(*font);
    }
    error = sfwViewGetPrintDimensions(sfw, &painter, header, footer,
            &dimensions);
    if (!error) {
        if (pages) {
            *pages = dimensions.pages;
        }
        if (truncated) {
            *truncated = dimensions.totalWidth - dimensions.pageWidth;
        }
    }
    return error;
}

#include "qtPrint.moc"

