//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtProgOutput.cc
//
// Creation date: 02.07.2015
// Last modified: 28.09.2017
//

#include "programoutput.h"
#include "qt/qtUtil.h"
#include <qt/classes/qtGroupLayout.h>
#include <qt/classes/qtPopSelection.h>
#include <QCheckBox>
#include <QLabel>
#include "project.h"
#include "preferences.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

class ProgOutputDialog : public CommonDialog
{
    Q_OBJECT

public:
    ProgOutputDialog(Project *pr) : CommonDialog(GuiFindParentWindow(0, pr)) {
        pr_ = pr;

        GroupLayout *group = new GroupLayout(2);
        contentsLayout()->addLayout(group);

        bool def = GetPrefsValue(pr, PREFS_PROGRAMOUTPUT_DEFAULT);
        const char *defaultName = DEFAULT_OUTPUT_FILE;
        const char *filename = def? 
            defaultName : GetPrefsString(pr, PREFS_PROGRAMOUTPUT_FILENAME);
        filename_ = new QLineEdit(this);
        filename_->setEnabled(!def);
        filename_->setText(filename);
        filename_->setToolTip(GCS(MSG_ProgOutput_FilenameBH));
        popFile_ = MakeFilePopup(filename_, pr, this);
        popFile_->button()->setEnabled(!def);
        popFile_->setSaveMode(true);
        group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_ProgOutput_Filename)),
                filename_, popFile_);

        QBoxLayout *hbox = new QHBoxLayout();
        contentsLayout()->addLayout(hbox);

        default_ = new QCheckBox(this);
        default_->setText(GuiMarkShortcut(GCS(MSG_ProgOutput_Default)));
        default_->setChecked(def);
        default_->setToolTip(GCS(MSG_ProgOutput_DefaultBH));
        connect(default_, SIGNAL(stateChanged(int)),
                this, SLOT(onDefaultStateChanged(int)));
        hbox->addWidget(default_);

        append_ = new QCheckBox(this);
        append_->setText(GuiMarkShortcut(GCS(MSG_ProgOutput_Append)));
        append_->setChecked(GetPrefsValue(pr, PREFS_PROGRAMOUTPUT_APPEND));
        append_->setToolTip(GCS(MSG_ProgOutput_AppendBH));
        hbox->addWidget(append_);

        setOkText(GuiMarkShortcut(GCS(MSG_Ok)));
        setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

        setWindowTitle(GCS(MSG_ProgOutput));
        GuiSetWindowProject(this, pr);
        GuiSetHelpNode(this, "Program output file");
    }

public slots:
    void onDefaultStateChanged(int state) {
        bool def = default_->checkState() == Qt::Checked;
        if (def) {
            filename_->setText(DEFAULT_OUTPUT_FILE);
        }
        filename_->setEnabled(!def);
        popFile_->button()->setEnabled(!def);
    }

private:
    virtual bool acceptEdits()
    {
        long error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return false;
        }

        bool append = append_->checkState() == Qt::Checked;
        bool def = default_->checkState() == Qt::Checked;
        if (!def) {
            QByteArray filename = filename_->text().toUtf8();
            SetPrefsString(pr_, PREFS_PROGRAMOUTPUT_FILENAME, filename.data());
        }
	SetPrefsValue(pr_, PREFS_PROGRAMOUTPUT_DEFAULT, def);
	SetPrefsValue(pr_, PREFS_PROGRAMOUTPUT_APPEND, append);
        return true;
    }

private:
    Project *pr_;
    QLineEdit *filename_;
    PopSelection *popFile_;
    QCheckBox *default_;
    QCheckBox *append_;
};

void GuiProgOutput(Project *pr)
{
    GuiBusyCursor busy;

    ProgOutputDialog dialog(pr);
    dialog.exec();
}

#include "qtProgOutput.moc"
