//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtProgramEditor.cc
//
// Creation date: 02.01.2017
// Last modified: 04.10.2021
//

#include "programeditor.h"

#ifdef ENABLE_PROGRAM_EDITOR

#include "menueditor.h"
#include "menuprogram.h"
#include "qt/qtUtil.h"
#include "qt/classes/qtGroupLayout.h"
#include "project.h"
#include "structureeditor.h"
#include "preferences.h"
#include "lang/compile.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

class ProgramEditorWindow : public CommonDialog
{
    Q_OBJECT

public:
    ProgramEditorWindow(Project *pr);

    virtual ~ProgramEditorWindow() { SetProjectProgramEditor(pr_, 0); }

    bool editorIsModified() const { return editor_->document()->isModified(); }

    bool checkClose()
    {
        return (!editorIsModified() ||
                Ask2(this, GCS(MSG_ProgramEditor), AT_WARN,
                    GCS(MSG_CloseAnswers), GCS(MSG_ProgramEditor_Close)));
    }

    bool compile(bool reload) { return MenuProgramCompile(pr_, reload); }

    const char *compile(bool reload,
            Program **compiled, long *line, long *col, const char **filename)
    {
        GuiBusyCursor busy;
	setWindowTitle(GCS(MSG_EditProgram));

	Program *prg = 0;
        QByteArray buf = editor_->toPlainTextNbsp().toUtf8();
	const char *text = buf.data();
        const char *error = CompileProjectProgram(pr_, text, strlen(text),
                &prg, line, col, filename);
        if (compiled) {
            *compiled = prg;
        }

        if (!error && reload) {
    	    SourceText *st = GetSourceText(prg);
	    const char *text = SourceTextText(st);

            saveCursor();
            GuiSetEditorText(editor_, text);
            GuiSetEditorChanged(editor_, false);
	    FreeSourceText(st);
            loadCursor();
	}
	return error;
    }

    void revert(bool restoreWindow)
    {
        if (!restoreWindow) {
            saveCursor();
        }

        Program *prg = ProjectProgram(pr_);
        SourceText *st = GetSourceText(prg);
        const char *text = SourceTextText(st);
        GuiSetEditorText(editor_, text);
        GuiSetEditorChanged(editor_, false);

        if (restoreWindow) {
            WindowDimensions wd;
            if (GuiGetWindowDimensions(this, &wd)) {
                SetPrefsWindowDimensions(pr_, PREFS_PROGRAMEDITOR, &wd);
            }
        }
        loadCursor();
    }

    const char *createSourceExternal()
    {
        GuiBusyCursor busy;
        QByteArray buf = editor_->toPlainTextNbsp().toUtf8();
	const char *text = buf.data();
        return CreateSourceExternal(pr_, text, strlen(text));
    }

public slots:
    void revertClick() { revert(false); }

    void compileClick() { compile(true); }

    void savePrefs()
    {
        WindowDimensions wd;
        if (GuiGetWindowDimensions(this, &wd)) {
            SetPrefsWindowDimensions(pr_, PREFS_PROGRAMEDITOR, &wd);
        }
        saveCursor();
    }

    void setEditorFocus() { editor_->setFocus(); }

    void setEditorCursor(int line, int col) { editor_->setCursor(line, col); }

private:
    Project *pr_;
    TextEditor *editor_;

    virtual void closeEvent(QCloseEvent *event)
    {
        if (checkClose()) {
            savePrefs();
            event->accept();
        } else {
            event->ignore();
        }
    }

    virtual bool acceptEdits() { return compile(true); }

    void saveCursor()
    {
        int top = editor_->top();
        int line, col;
        editor_->getCursor(&line, &col);

        SetPrefsValue(pr_, PREFS_PROGRAMEDITOR_TOP, top);
        SetPrefsValue(pr_, PREFS_PROGRAMEDITOR_POSX, col);
        SetPrefsValue(pr_, PREFS_PROGRAMEDITOR_POSY, line);
    }

    void loadCursor()
    {
        int top = GetPrefsValue(pr_, PREFS_PROGRAMEDITOR_TOP);
        int col = GetPrefsValue(pr_, PREFS_PROGRAMEDITOR_POSX);
        int line = GetPrefsValue(pr_, PREFS_PROGRAMEDITOR_POSY);

        editor_->setTop(top);
        editor_->setCursor(line, col);
        editor_->setFocus();
    }
};

ProgramEditorWindow::ProgramEditorWindow(Project *pr) :
    CommonDialog(GuiFindParentWindow(0, pr))
{
    setAttribute(Qt::WA_DeleteOnClose);

    pr_ = pr;

    setWindowTitle(GCS(MSG_EditProgram));

    editor_ = new TextEditor(pr_, true, this);
    contentsLayout()->addWidget(editor_);

    struct WindowDimensions wd;
    if (GetPrefsWindowDimensions(pr_, PREFS_PROGRAMEDITOR, &wd)) {
        GuiSetWindowDimensions(this, &wd);
    }
    
    buttonBox()->setStandardButtons(QDialogButtonBox::Ok);

    QPushButton *revert = new QPushButton(
            GuiMarkShortcut(GCS(MSG_ProgramEditor_Revert)), this);
    revert->setToolTip(GCS(MSG_ProgramEditor_RevertBH));
    buttonBox()->addButton(revert, QDialogButtonBox::ResetRole);
    connect(revert, SIGNAL(clicked()), this, SLOT(revertClick()));

    QPushButton *compile = new QPushButton(
        GuiMarkShortcut(GCS(MSG_ProgramEditor_Compile)), this);
    compile->setToolTip(GCS(MSG_ProgramEditor_CompileBH));
    buttonBox()->addButton(compile, QDialogButtonBox::ApplyRole);
    connect(compile, SIGNAL(clicked()), this, SLOT(compileClick()));

    setOkText(GuiMarkShortcut(GCS(MSG_ProgramEditor_CompileAndClose)));
    buttonBox()->button(QDialogButtonBox::Ok)->setToolTip(
        GCS(MSG_ProgramEditor_CompileAndCloseBH));

    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Program editor");
}

void CloseProgramEditor(GuiObject go)
{
    ProgramEditorWindow *pew = dynamic_cast<ProgramEditorWindow *>(go);
    if (pew) {
        pew->savePrefs();
        pew->deleteLater();
    }
};

bool ProgramEditorCheckClose(GuiObject go)
{
    ProgramEditorWindow *pew = dynamic_cast<ProgramEditorWindow *>(go);
    return (!pew || pew->checkClose());
}


GuiObject OpenProgramEditor(Project *pr)
{
    GuiObject go = ProjectProgramEditor(pr);
    bool external = GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL);

    if (!go) {
	if (external) {
            return 0;
	}

        ProgramEditorWindow *pew = new ProgramEditorWindow(pr);
        go = pew;
        if (!GuiOpenWindow(go)) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GCS(MSG_Err_CantOpenProgramEditor));
            delete pew;
            return 0;
        }
        pew->revert(true);                           // restore
        SetProjectProgramEditor(pr, go);
    } else {
	GuiClearWindowSleep(go);
	GuiWindowToFront(go);
	GuiActivateWindow(go);
    }

    ProgramEditorWindow *pew = dynamic_cast<ProgramEditorWindow *>(go);
    if (pew) {
        pew->setEditorFocus();
    }

    return go;
}

void ProgramEditorShowError(Project *pr, bool reload,
	const char *error, long line, long col, const char *filename)
{
    if (error == NULL) {
	return;
    } 

    if (IsProjectErrorString(error)) {
        Ask(AT_ERROR, GCS(MSG_StarOk), error);
        return;
    }

    GuiObject go = ProjectProgramEditor(pr);
    ProgramEditorWindow *pew = dynamic_cast<ProgramEditorWindow *>(go);
    if (line < 0) {
        Ask2(pew, NULL, AT_WARN, GCS(MSG_StarOk),
            GCS(MSG_ProgramEditor_NoDebugInfo), error);
	return;
    }

    if (filename == NULL && GetPrefsValue(pr, PREFS_PROGRAMSOURCEEXTERNAL)) {
        filename = GetPrefsString(pr, PREFS_PROGRAMSOURCEEXTERNAL_FILENAME);
    }
    if (filename) {
	Ask2(pew, NULL, AT_ERROR, GCS(MSG_StarOk),
	      	GCS(MSG_ProgramEditor_IncludeFileError), 
		filename, line, col, error);
    } else if (!ProjectAdminMode(pr)) {
	Ask2(pew, NULL, AT_ERROR, GCS(MSG_StarOk),
	      	GCS(MSG_ProgramEditor_NotAdmin), line, col, error);
    } else {
	//
	// show the error in BeeBase' program editor
	// The editor might not be opened yet,
	// open and unchanged, or open and changed.
	//
	if (pew && reload) {
            if (pew->editorIsModified()) {
		// we should restore the program source
		GuiBeep();
		short rc = Ask2(pew, GCS(MSG_ProgramEditor), AT_WARN,
			GCS(MSG_ProgramEditor_ShowAnswers),
			GCS(MSG_ProgramEditor_ShowError), line, col, error);
		if (rc == 0) {
		    return;				// user canceled
		} else if (rc == 2) {
		    pew->revert(true);			// restore
		}
	    }
	}
	go = OpenProgramEditor(pr);	// either creates or brings to front
        pew = dynamic_cast<ProgramEditorWindow *>(go);
	if (pew) {
	    pew->setWindowTitle(error);
            pew->setEditorCursor(line, col);
	    GuiBeep();
	} // else OpenProgramEditor popped up an error dialog
    }
}

bool ProgramEditorFix(Project *pr)
{
    GuiBusyCursor busy;

    GuiObject go = ProjectProgramEditor(pr);
    ProgramEditorWindow *pew = dynamic_cast<ProgramEditorWindow *>(go);
    bool result = true;

    if (pew) {
        if (pew->editorIsModified()) {
	    // we should restore the program source
	    GuiBeep();
	    short rc = Ask2(pew, GCS(MSG_ProgramEditor), AT_WARN,
		    GCS(MSG_ProgramEditor_ShowAnswers),
		    GCS(MSG_ProgramEditor_ShowDeleted));
	    if (rc == 0) {
		result = false;			// user canceled
	    } else if (rc == 2) {
		pew->revert(true);		// restore
	    }
	}
    }

    if (result) {
	go = OpenProgramEditor(pr);	// either creates or brings to front
        pew = dynamic_cast<ProgramEditorWindow *>(go);
	if (pew) {
	    pew->compile(true);			// show the location to fix
	    result = pew->exec() == QDialog::Accepted;
	}
    }

    return result;
}

const char *ProgramEditorCompile(GuiObject go, bool reload,
	Program **compiled, long *line, long *col, const char **filename)
{
    ProgramEditorWindow *pew = dynamic_cast<ProgramEditorWindow *>(go);
    if (!pew) {
	return GCS(MSG_Err_Internal);
    } else {
        return pew->compile(reload, compiled, line, col, filename);
    }
}
const char *ProgramEditorCreateSourceExternal(GuiObject go)
{
    ProgramEditorWindow *pew = dynamic_cast<ProgramEditorWindow *>(go);
    if (pew == NULL) {
	return GCS(MSG_Err_Internal);
    } else {
        return pew->createSourceExternal();
    }
}

#include "qtProgramEditor.moc"

#else

void CloseProgramEditor(GuiObject obj) {}

bool ProgramEditorCheckClose(GuiObject obj) { return false; }

GuiObject OpenProgramEditor(Project *pr) { return 0; }

bool ProgramEditorFix(Project *pr) { return false; }

void ProgramEditorShowError(Project *pr, bool reload,
	const char *error, long line, long col, const char *filename) {}

const char *ProgramEditorCompile(GuiObject pe, bool reload,
	Program **compiled, long *line, long *col, char **filename)
{ return 0; }

const char *ProgramEditorCreateSourceExternal(GuiObject pe) { return 0; }

#endif

