//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtQueryEditor.cc
//
// Creation date: 25.03.2015
// Last modified: 03.03.2022
//

#include "system.h"
#include "qt/classes/qtGroupLayout.h"
#include "qt/classes/qtPopWidget.h"
#include "qt/classes/qtSFWView.h"
#include "qt/classes/qtSortableListModel.h"
#include "qt/qtBuildGuiField.h"     // for SFWViewOnRowActivated
#include "qt/qtUtil.h"
#include <QApplication>
#include <QDialog>
#include <QList>
#include <QListView>
#include <QPushButton>
#include <QSplitter>
#include "lang/compile.h"
#include "queryeditor.h"
#include "table.h"
#include "preferences.h"
#include "imexport.h"
#include "pdfexport.h"
#include "print.h"
#include "requesters.h"
#include "util.h"
#include "memory.h"
#include "error.h"
#include "appstrings_cat.h"

static int queryCompare(void *a, void *b)
{
    Query *q1 = (Query *)a, *q2 = (Query *)b;
    const char *n1 = QueryName(q1), *n2 = QueryName(q2);
    if (n1 && n2) {
        return CodesetStrCollate(n1, n2, CS_COLLATE_ORDER);
    } else if (n2) {
	return -1;
    } else if (n1) {
	return 1;
    } else {
	return 0;
    }
}

class QueriesModel : public SortableListModel
{
public:
    QueriesModel(QObject *parent, List *queries) : SortableListModel(parent)
    {
        setList(queries);
    }
    
    QModelIndex queryIndex(Query *q) const { return nodeIndex((Node *)q); }
    Query *query(const QModelIndex &index) const 
    { return (Query *)node(index); }
    Query *query(int ix) const { return (Query *)node(ix); }

protected:
    QString display(int ix) const
    {
        Query *q = query(ix);
        const char *name = QueryName(q);
        if (!name && q) {
            name = GCS(MSG_Unnamed);
        }
        return name;
    }
};

class QueryBrowser : public QWidget
{
    Q_OBJECT
    
public:
    QueryBrowser(QWidget *parent, QueryEditor *qe) : QWidget(parent)
    {
        qe_ = qe;
        
        GroupLayout *top = new GroupLayout(1, this);
        
        queriesView_ = new QListView(this);
        queriesModel_ = new QueriesModel(queriesView_, QueryEditorQueries(qe));
        queriesView_->setModel(queriesModel_);
        queriesView_->setSelectionMode(QAbstractItemView::SingleSelection);
        queriesView_->setDragDropMode(QAbstractItemView::InternalMove);
        connect(queriesView_, SIGNAL(activated(const QModelIndex &)), this, SLOT(onActivated(const QModelIndex &)));
        top->addWidget(queriesView_);
        setFocusProxy(queriesView_);
        
        QHBoxLayout *buttons = new QHBoxLayout;
        buttons->setContentsMargins(0, 0, 0, 0);
        buttons->setSpacing(10);
        top->addLayout(buttons);
        
        QPushButton *newButton = new QPushButton(this);
        newButton->setText(GuiMarkShortcut(GCS(MSG_New)));
        newButton->setAutoDefault(false);
        connect(newButton, SIGNAL(clicked()), this, SLOT(onNewClicked()));
        buttons->addWidget(newButton);

        QPushButton *duplicateButton = new QPushButton(this);
        duplicateButton->setText(GuiMarkShortcut(GCS(MSG_Duplicate)));
        duplicateButton->setAutoDefault(false);
        connect(duplicateButton, SIGNAL(clicked()), this, SLOT(onDuplicateClicked()));
        buttons->addWidget(duplicateButton);

        QPushButton *sortButton = new QPushButton(this);
        sortButton->setText(GuiMarkShortcut(GCS(MSG_Sort)));
        sortButton->setAutoDefault(false);
        connect(sortButton, SIGNAL(clicked()), this, SLOT(onSortClicked()));
        buttons->addWidget(sortButton);

        QPushButton *deleteButton = new QPushButton(this);
        deleteButton->setText(GuiMarkShortcut(GCS(MSG_Delete)));
        deleteButton->setAutoDefault(false);
        connect(deleteButton, SIGNAL(clicked()), this, SLOT(onDeleteClicked()));
        buttons->addWidget(deleteButton);
    }
    
public slots:
    void setCurrentQuery(Query *query)
    {
        queriesView_->setCurrentIndex(queriesModel_->queryIndex(query));
    }
    
    void onActivated(const QModelIndex &index)
    {
        emit done(queriesModel_->query(index));
    }
    
    void onNewClicked()
    {
        Query *query = NewQuery(qe_, 0);
        if (query) {
            queriesView_->reset();
            emit done(query);
        }
    }

    void onDuplicateClicked()
    {
        Query *query = NewQuery(qe_, currentQuery());
        if (query) {
            queriesView_->reset();
            emit done(query);
        }
    }
    
    void onSortClicked() {
        List *queries = QueryEditorQueries(qe_);
        Query *query = currentQuery();
        QLSort(queries, queryCompare);
        queriesView_->reset();
        setCurrentQuery(query);
    }
            
    void onDeleteClicked()
    {
        Query *query = currentQuery();
        if (query) {
            Query *other = SUCC(query);
            if (!SUCC(other)) {
                other = PRED(query);
                if (!PRED(other)) {
                    // clear entry, we will not free last element
                    SetQueryName(query, 0);
                    SetQueryProgram(query, 0);
                    other = query;
                }
            }
            emit queryGone(query, other);
            queriesView_->reset();
            if (other != query) {
                DeleteQuery(query);
            }
            setCurrentQuery(other);
        }
    }

signals:
    void done(Query *query);
    void queryGone(Query *query, Query *other);
    
private:
    QueryEditor *qe_;
    QueriesModel *queriesModel_;
    QListView *queriesView_;
    
    Query *currentQuery() const
    {
        return queriesModel_->query(queriesView_->currentIndex());
    }
};

class PopQuery : public PopWidget
{
    Q_OBJECT
    
public:
    PopQuery(QLineEdit *name, QAbstractButton *button, QueryBrowser *browser)
    {
        setString(name);
        setButton(button);
        setPopupWidget(browser);
        connect(browser, SIGNAL(done(Query *)), this, SLOT(closePopup()));
    }
    
signals:
    bool aboutToStart();
    
protected:
    virtual bool start()
    {
        return emit aboutToStart(); 
    }
};

class QueryTableModel : public QAbstractListModel
{
public:
    QueryTableModel(QWidget *parent, List *tabs) : QAbstractListModel(parent)
    {
        parent_ = parent;
        tabs_ = tabs;
    }
    
    virtual int	rowCount(const QModelIndex &parent = QModelIndex()) const
    {
        return 1 + NumNodes(tabs_);
    }
    
    virtual QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const
    {
        if (index.isValid()) {
            switch (role) {
                case Qt::DisplayRole: {
                    Table *tab = (Table *)FindNodeNum(tabs_, index.row());
                    return tab? TableName(tab) : GCS(MSG_Query_TableNone);
                }
                case Qt::FontRole: {
                    Table *tab = (Table *)FindNodeNum(tabs_, index.row());
                    QFont font = QApplication::font(parent_);
                    if (!tab) {
                        font.setStyle(QFont::StyleItalic);
                    }
                    return font;
                }
            }
        }
        return QVariant();
    }
    
private:
    QWidget *parent_;
    List *tabs_;
};

class QueryTableComboBox : public QComboBox
{
public:
    QueryTableComboBox(QWidget *parent, List *tabs) : QComboBox(parent)
    {
        model_ = new QueryTableModel(this, tabs);
        setModel(model_);
        setView(new QListView(this));
    }
    
private:
    QueryTableModel *model_;
};

static Project *queryProject = 0;

static const char *outputFunc(Element *elem);

class QueryEditorWindow : public QDialog
{
    Q_OBJECT
    
public:
    QueryEditorWindow(QueryEditor *qe);
    
public slots:
    void onQueryNameReturnPressed()
    {
        long line, col;
        const char *error = saveQuery(&line, &col, false);
        if (error) {
            showerror(error, line, col);
        }
    }
    
    bool canSwitchQuery()
    {
        long line, col;
        const char *error = saveQuery(&line, &col, false);
        if (error) {
            showerror(error, line, col);
            return false;
        }        
        return true;
    }

    void setActiveQuery(Query *query, bool force = false) 
    {
        if (force || (query && query != activeQuery_)) {
            // clear window title
            setWindowTitle(GCS(MSG_QueryEditor));
  
            const char *name = QueryName(query);
            queryName_->setText(name? name : GCS(MSG_Unnamed));
            queryTable_->setCurrentIndex(NodeNum(QueryTable(query)));
            
            /* set source text */
            Program *prg = QueryProgram(query);
            SourceText *st = GetSourceText(prg);
            const char *src = SourceTextText(st);
            long len = src? strlen(src) : 0;
            char *kill, c;
            
            if (len >= 3) {
                src++;          /* skip "("   */
                len -= 3;       /* skip "\n)" */
                kill = (char *)src + len;
                c = *kill;
                *kill = 0;
                GuiSetEditorText(editor_, src);
                *kill = c;
            } else {
                GuiSetEditorText(editor_, "");
            }
            GuiSetEditorChanged(editor_, false);
            FreeSourceText(st);
            
            sfwView_->setElement(0);
 
            activeQuery_ = query;
            queryBrowser_->setCurrentQuery(query);
        }
    }

    void updateActiveQuery(Query *query, Query *other)
    {
        if (query == activeQuery_) {
            setActiveQuery(other, true);
        }
    }
    
    bool switchToQuery(Query *query)
    {
        if (query == activeQuery_) {
            return true;
        }
        if (!canSwitchQuery()) {
            return false;
        }
        setActiveQuery(query, true);
        return true;
    }
    
    void remRecord(Record *rec) {
        sfwView_->remRecord(rec);
    }
    
    void remAllRecords() {
        sfwView_->remAllRecords();
    }
    
    void onRunClicked() 
    {
        GuiBusyCursor busy;
            
        if (activeQuery_) {
            long line, col;
            const char *error = saveQuery(&line, &col, true);
            if (!error) {
                // execute program
                Program *prg = QueryProgram(activeQuery_);
                SetTableProgramRecords(pr_);
                queryProject = pr_;
                error = ExecuteProgram(prg, true, 0, 0, true, outputFunc, &line, &col, 0);
                queryProject = 0;
            }
            if (error) {
                showerror(error, line, col);
            }
        }
    }
    
    void onExportClicked() { GuiExportSFWList(pr_, sfwView_); }
    void onPDFClicked() { GuiExportSFWListAsPDF(pr_, sfwView_); }
    void onPrintClicked() { GuiPrintSFWList(pr_, sfwView_); }
    
    void onSFWViewActivated(const QModelIndex &index)
    {
        SFWViewOnRowActivated(sfwView_, index, ProjectRootDisplayObject(pr_));
    }
    
    void onSFWModelReset()
    {
        bool enabled = (sfwView_->model()->rowCount() > 0);
            
        GuiSetEnabled(export_, enabled);
        GuiSetEnabled(pdf_, enabled);
        GuiSetEnabled(print_, enabled);
    }
    
    const char *setOutput(Element *elem)
    {
        return sfwView_->setElement(elem);
    }
    
    void closeEvent(QCloseEvent *event)
    {
        long line, col;
        const char *error = saveQuery(&line, &col, false);
        if (error && !Ask(AT_WARN, GCS(MSG_CloseAnswers), GCS(MSG_Query_Close))) {
            showerror(error, line, col);
        } else {
            CloseQueryEditor(pr_);
        }
        event->ignore();
    }
    
    const char *saveQuery(long *line, long *col, bool force)
    {
        if (activeQuery_) {
            QByteArray nameData = queryName_->text().toUtf8();
            const char *name = nameData.data();
            const char *oldName = QueryName(activeQuery_);
            long error = OK;
            
            if (strcmp(name? name : "", oldName? oldName : "") != 0) {
                error = ObtainProjectChangeLock(pr_);
                if (error != OK) {
                    return GES(error);
                }
                SetQueryName(activeQuery_, name);
                ProjectChange(pr_, true);
            }
            
            Table *tab = (Table *)FindNodeNum(ProjectTables(pr_), queryTable_->currentIndex());
            if (tab != QueryTable(activeQuery_)) {
                error = ObtainProjectChangeLock(pr_);
                if (error != OK) {
                    return GES(error);
                }
                List *queries = QueryEditorQueries(qe_);
                Query *query;
                if (queries) {
                    FORNODE(query, queries) {
                        if (QueryTable(query) == tab) {
                            SetQueryTable(query, 0);
                        }
                    }
                }
                SetQueryTable(activeQuery_, tab);
                ProjectChange(pr_, true);
            }
            
            bool modified = editor_->document()->isModified();
            if (force || modified) {
                MemPool *memPool = ProjectMemPool(pr_);
                Program *prg = 0;
                const char *errstr = 0;
                char *src = GuiEditorText(editor_, memPool);
                if (!src) {
                    errstr = GCS(MSG_Err_Mem);
                } else {
                    errstr = CompileQuery(pr_, src, strlen(src), &prg, line, col);
                }
                FreeString(memPool, src);
                
                if (errstr) {
                    return errstr;
                }
                if (modified) {
                    error = ObtainProjectChangeLock(pr_);
                    if (error != OK) {
                        return GES(error);
                    }
                }
                SetQueryProgram(activeQuery_, prg);
                setActiveQuery(activeQuery_, true);         // refreshes source
                if (modified) {
                    ProjectChange(pr_, true);
                }
            }
        }
        return 0;
    }
    
private:
    Project *pr_;
    QueryEditor *qe_;
    Query *activeQuery_;
    StringEdit *queryName_;
    QueryBrowser *queryBrowser_;
    QueryTableComboBox *queryTable_;
    QPushButton *export_, *pdf_, *print_;
    TextEditor *editor_;
    SFWView *sfwView_;
    
    void showerror(const char *error, long line, long col)
    {
        if (error && error != EXECUTE_ABORT) {
            if (IsProjectErrorString(error)) {
                Ask(AT_ERROR, GCS(MSG_StarOk), error);
            } else {
                setWindowTitle(error);
                editor_->setCursor(line, (line == 1 && col > 0)? col - 1 : col);
                editor_->setFocus();
                GuiBeep();
            }
        }
    }
};

QueryEditorWindow::QueryEditorWindow(QueryEditor *qe)
{
    pr_ = QueryEditorProject(qe);
    qe_ = qe;
    activeQuery_ = 0;

    setWindowTitle(GCS(MSG_QueryEditor));
    
    GroupLayout *mainLayout = new GroupLayout(1, this);
    
    QHBoxLayout *topRow = new QHBoxLayout();
    mainLayout->addLayout(topRow);
    
    queryName_ = new StringEdit(this);
    queryName_->setMinimumWidth(queryName_->fontMetrics().width("W") * 10);
    queryName_->setToolTip(GCS(MSG_Query_NameBH));
    connect(queryName_, SIGNAL(returnPressed()), this, SLOT(onQueryNameReturnPressed()));
            
    PopButton *queryPopButton = new PopButton(this, queryName_);
    queryPopButton->setArrowType(Qt::DownArrow);

    queryBrowser_ = new QueryBrowser(this, qe);    
    connect(queryBrowser_, SIGNAL(done(Query *)), this, SLOT(setActiveQuery(Query *)));
    connect(queryBrowser_, SIGNAL(queryGone(Query *, Query *)), 
            this, SLOT(updateActiveQuery(Query *, Query *)));

    PopQuery *queryPop = new PopQuery(queryName_, queryPopButton, queryBrowser_);
    connect(queryPop, SIGNAL(aboutToStart()), this, SLOT(canSwitchQuery()));
    topRow->addLayout(queryPop, 3);
    
    queryTable_ = new QueryTableComboBox(this, ProjectTables(pr_));
    queryTable_->setToolTip(GCS(MSG_Query_TableBH));
    topRow->addWidget(queryTable_, 1);

    QPushButton *run = new QPushButton(this);
    run->setText(GCS(MSG_Query_Run));
    run->setToolTip(GCS(MSG_Query_RunBH));
    run->setAutoDefault(false);
    connect(run, SIGNAL(clicked()), this, SLOT(onRunClicked()));
    topRow->addWidget(run, 1);

    export_ = new QPushButton(this);
    export_->setText(GCS(MSG_Query_Export));
    export_->setToolTip(GCS(MSG_Query_ExportBH));
    export_->setAutoDefault(false);
    export_->setEnabled(false);
    connect(export_, SIGNAL(clicked()), this, SLOT(onExportClicked()));
    topRow->addWidget(export_, 1);
    
    pdf_ = new QPushButton(this);
    pdf_->setText(GCS(MSG_Query_PDF));
    pdf_->setToolTip(GCS(MSG_Query_PDFBH));
    pdf_->setAutoDefault(false);
    pdf_->setEnabled(false);
    connect(pdf_, SIGNAL(clicked()), this, SLOT(onPDFClicked()));
    topRow->addWidget(pdf_, 1);
    
    print_ = new QPushButton(this);
    print_->setText(GCS(MSG_Query_Print));
    print_->setToolTip(GCS(MSG_Query_PrintBH));
    print_->setAutoDefault(false);
    print_->setEnabled(false);
    connect(print_, SIGNAL(clicked()), this, SLOT(onPrintClicked()));
    topRow->addWidget(print_, 1);
    
    QSplitter *splitter = new QSplitter(this);
    splitter->setOrientation(Qt::Vertical);
    splitter->setChildrenCollapsible(false);
    mainLayout->addWidget(splitter);
    
    editor_ = new TextEditor(pr_, true, this);
    editor_->setToolTip(GCS(MSG_Query_ProgramBH));
    splitter->addWidget(editor_);
    splitter->setStretchFactor(0, 2); 
    
    sfwView_ = new SFWView(this);
    sfwView_->setShowTitles(true);
    sfwView_->setToolTip(GCS(MSG_Query_OutputBH));
    connect(sfwView_, SIGNAL(activated(const QModelIndex &)), this, SLOT(onSFWViewActivated(const QModelIndex &)));
    connect(sfwView_->model(), SIGNAL(modelReset()), this, SLOT(onSFWModelReset()));
    splitter->addWidget(sfwView_);
    splitter->setStretchFactor(1, 8);
    
    WindowDimensions wd;
    if (GetPrefsWindowDimensions(pr_, PREFS_QUERYEDITOR, &wd)) {
        GuiSetWindowDimensions(this, &wd);
    }
    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Query editor");

    List *queries = QueryEditorQueries(qe);
    if (queries) {
        Query *q = (Query *)HEAD(queries);
        if (SUCC(q)) {
            setActiveQuery(q);
        }
    }
    
    editor_->setFocus();
}
 
static const char *outputFunc(Element *elem)
{
    QueryEditor *se = ProjectQueryEditor(queryProject);
    GuiObject go = QueryEditorGui(se);
    QueryEditorWindow *win = dynamic_cast<QueryEditorWindow *>(go);
    const char *error = 0; 
    
    if (win) {
        error = win->setOutput(elem);
    } else {
	error = GCS(MSG_Err_Internal);
    }
    
    return error;
}
    
GuiObject OpenQueryGui(QueryEditor *qe)
{
    List *queries = QueryEditorQueries(qe);
    if (!queries) {
	return 0;
    }

    // check if we have any queries, if not allocate one
    if (!SUCC(HEAD(queries))) {
        Query *q = NewQuery(qe, 0);
	if (!q) {
	    return 0;
	}
    }
    
    QueryEditorWindow *win = new QueryEditorWindow(qe);
    win->show();
    
    return win;
}

void CloseQueryGui(GuiObject go)
{
    QueryEditorWindow *win = dynamic_cast<QueryEditorWindow *>(go);
    if (win) {
        win->saveQuery(0, 0, false);

        WindowDimensions wd;
        if (GuiGetWindowDimensions(go, &wd)) {
            SetPrefsWindowDimensions(GuiWindowProject(go), PREFS_QUERYEDITOR, &wd);
        }
        win->deleteLater();
    }
}

void QueryEditorRemRecord(Table *tab, Record *rec)
{
    Project *pr = TableProject(tab);
    QueryEditor *qe = ProjectQueryEditor(pr);
    GuiObject go = QueryEditorGui(qe);
    QueryEditorWindow *win = dynamic_cast<QueryEditorWindow *>(go);
    if (win) {
        win->remRecord(rec);
    }
}

void QueryEditorRemAllRecords(Project *pr)
{
    QueryEditor *qe = ProjectQueryEditor(pr);
    GuiObject go = QueryEditorGui(qe);
    QueryEditorWindow *win = dynamic_cast<QueryEditorWindow *>(go);
    if (win) {
        win->remAllRecords();
    }
}

void QueryGuiActivate(GuiObject go)
{
    if (go) {
        // bring window to front and activate it
	GuiWindowToFront(go);
	GuiActivateWindow(go);
    }
}

void QueryGuiRunQuery(GuiObject go, Query *query)
{
    if (go && query) {
        QueryEditorWindow *win = dynamic_cast<QueryEditorWindow *>(go);
        if (win) {
            if (win->switchToQuery(query)) {
                win->onRunClicked(); 
            }
        }
    }
}

#include "qtQueryEditor.moc"
