//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtRequesters.cc
//
// Creation date: 24.03.2009
// Last modified: 04.06.2022
//

#include "system.h"
#include "qt/qtAmiFileDialog.h"
#include <qt/qtUtil.h>
#include <qt/classes/qtGroupLayout.h>
#include <qt/classes/qtPopSelection.h>
#include <qt/classes/qtSFWView.h>
#include <QApplication>
#include <QBoxLayout>
#include <QCheckBox>
#include <QDialogButtonBox>
#include <QEventLoop>
#include <QFormLayout>
#include <QInputDialog>
#include <QLabel>
#include <QLineEdit>
#include <QListWidget>
#include <QMessageBox>
#include <QPushButton>
#include <QShortcut>
#include <QStyle>
#include <QVariant>
#include "requesters.h"
#include "preferences.h"
#include "project.h"
#include "labels.h"
#include "memo.h"
#include "dbtypes.h"
#include "codeset.h"
#include "util.h"
#include "memory.h"
#include "appstrings_cat.h"
#include <limits>

#undef USE_FORMLAYOUT

static char *getdirfilename(Project *pr, MemPool *memPool,
	const char *title, const char *okText, 
	const char *pref, long prefItem, 
	bool saveMode, bool dirMode)
{
    char *result = 0;
    GuiBusyCursor busy;

    if (!pref && prefItem > 0) {
	pref = GetPrefsString(pr, prefItem);
    }

    ProjectSetCurrentDir(pr);	    /* make sure we have correct current dir */

    QWidget *parent = GuiFindParentWindow(0, pr);
    QString filename = AmiFileDialog::get(parent, title, okText, pref, dirMode, saveMode);
    
    if (!filename.isNull()) {
        result = AllocString(memPool, filename.toUtf8().constData());
        if (result && prefItem > 0) {
            SetPrefsString(pr, prefItem, result);
        }
    }

    return result;
}

char *GetFilename(Project *pr, MemPool *memPool,
        const char *title, const char *okText,
        const char *pref, long prefItem, bool saveMode)
{
    return getdirfilename(pr, memPool, title, okText, 
	    pref, prefItem, saveMode, false);
}

char *GetDirname(Project *pr, MemPool *memPool, 
	const char *title, const char *okText,
        const char *pref, long prefItem, bool saveMode)
{
    return getdirfilename(pr, memPool, title, okText, 
	    pref, prefItem, saveMode, true);
}

static short Ask(GuiObject obj, const char *title, AskType askType,
	const char *buttons, const char *format, va_list ap)
{
    GuiBusyCursor busy;

    QWidget *parent = GuiFindParentWindow(obj);
    QMessageBox::Icon icon;

    switch (askType) {
        case AT_INFO:	  icon = QMessageBox::Information; break;
        default:
        case AT_WARN:     icon = QMessageBox::Warning; break;
        case AT_QUESTION: icon = QMessageBox::Question; break;
        case AT_ERROR:	  icon = QMessageBox::Critical; break;
    }
    
    static const size_t bufSize = 8192;
    static char buffer[bufSize];
    buffer[0] = 0;
    if (format) {
        qvsnprintf(buffer, bufSize, format, ap);
        buffer[bufSize-1] = 0;
    }
    for (char *s = buffer; *s; s++) {
        if (*s == '\n') {
            if (s[1] != '\n') {
                *s = ' ';
            } else {
                s++;
            }
        }
    }
    QString message(buffer);

    QMessageBox dialog(parent);
    dialog.setIcon(icon);
    dialog.setTextFormat(Qt::PlainText);
    dialog.setText(message);
    dialog.setWindowTitle(title? title : GCS(MSG_APP));

    if (!buttons || !buttons[0]) {
        buttons = GCS(MSG_StarOk);
    }

    int num = 0;
    while (buttons) {
	bool def = num == 0;

	if (buttons[0] == '*') {
	    def = true;
	    buttons++;
	}

	const char *next = CodesetStrchr(buttons, '|');
        QString text = QString::fromUtf8(buttons, next? (next - buttons) : -1);
        QPushButton *button = dialog.addButton(GuiMarkShortcut(text),
                next? QMessageBox::AcceptRole : QMessageBox::RejectRole);

        int id = next? num + 1 : 0;
        QVariant v = qVariantFromValue(id);
        button->setProperty("button-id", v);

	num++;

	if (def) {
            dialog.setDefaultButton(button);
	}
        if (next) {
            buttons = next + 1;
        } else {
            dialog.setEscapeButton(button);
            buttons = 0;
        }
    }

    dialog.exec();

    int answer = 0;
    QAbstractButton *button = dialog.clickedButton();
    if (button) {
        QVariant v = button->property("button-id");
        answer = v.value<int>();
    }

    return (short)answer;
}

short Ask(AskType type, const char *buttons, const char *text, ...)
{
    va_list ap;
    short rc;

    va_start(ap, text);
    rc = Ask(0, 0, type, buttons, text, ap);
    va_end(ap);
 
    return rc;
}

short Ask2(GuiObject obj, const char *title, AskType type,
	const char *buttons, const char *text, ...)
{
    va_list ap;
    short rc;

    va_start(ap, text);
    rc = Ask(obj, title, type, buttons, text, ap);
    va_end(ap);
 
    return rc;
}

#define MAXLEN 1000

class PopFileSelection : public PopSelection
{
public:
    PopFileSelection(QWidget *string, QAbstractButton *button)
    {
        setType(PopSelection::FileSelection);
        setString(string);
        setButton(button);
    }
    
    virtual bool start()
    {
        QString str = text();
        int index = str.indexOf(' ');
        if (index >= 0) {
            str = str.left(index);
        }
        fileDialog()->selectFile(str);
        return true;
    }
    
    virtual void stop()
    {
        QStringList filenames = fileDialog()->selectedFiles();
        if (filenames.empty()) {
            return;
        }
        QString str = text();
        int index = str.indexOf(' ');
        if (index < 0) {
            index = str.length();
        }
        str.replace(0, index, filenames.first());
        setText(str);
    }    
};

const char *AskString(struct Project *pr, const char *caption,
	const char *oktext, const char *contents, ulong maxlen, 
	AskStringType type)
{
    GuiBusyCursor busy;

    const char *title = ProjectTitle(pr);
    if (maxlen > MAXLEN) {
        maxlen = MAXLEN;
    }

    const char *ret = 0;
    static char buffer[MAXLEN + 1];
    
    if (type == AST_FileSelect) {
        CommonDialog dialog(GuiFindParentWindow(0, pr));
        dialog.setWindowTitle(title? title : GCS(MSG_APP));
        QLabel *label = new QLabel(&dialog);
        label->setText(caption? caption : GCS(MSG_EnterString));
        label->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
        dialog.contentsLayout()->addWidget(label);
        QLineEdit *string = new QLineEdit(&dialog);
        string->setText(contents);
        string->setMaxLength(maxlen);
        label->setBuddy(string);
        PopButtonSetCWD *popButton = new PopButtonSetCWD(&dialog, string, pr);
        popButton->setIcon(popButton->style()->standardIcon(QStyle::SP_FileIcon));
        PopFileSelection *popFile = new PopFileSelection(string, popButton);
        dialog.contentsLayout()->addLayout(popFile);
        dialog.setOkText(GuiMarkShortcut(oktext? oktext : GCS(MSG_Ok)));
        dialog.setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));
        
        if (dialog.exec() == QDialog::Accepted) {
            QString text = string->text();
            CodesetStrNBytesCopy(buffer, text.toUtf8().data(), MAXLEN);
            ret = buffer;
        }
    } else {
        QInputDialog dialog(GuiFindParentWindow(0, pr));
        dialog.setInputMode(QInputDialog::TextInput);
        dialog.setWindowTitle(title? title : GCS(MSG_APP));
        dialog.setLabelText(caption? caption : GCS(MSG_EnterString));
        dialog.setTextValue(contents);
        // TODO: limit string length to maxlen
        if (type == AST_Secret) {
            dialog.setTextEchoMode(QLineEdit::Password);
        }
        dialog.setOkButtonText(GuiMarkShortcut(oktext? oktext : GCS(MSG_Ok)));
        dialog.setCancelButtonText(GuiMarkShortcut(GCS(MSG_Cancel)));
    
        if (dialog.exec() == QDialog::Accepted) {
            QString text = dialog.textValue();
            CodesetStrNBytesCopy(buffer, text.toUtf8().data(), MAXLEN);
            ret = buffer;
        }
    }
    
    return ret;    
}

bool AskInteger(Project *pr, const char *caption, const char *oktext,
        const char *contents, long min, long max, long *result)
{
    GuiBusyCursor busy;

    const char *title = ProjectTitle(pr);
    static char minmax[128];
    min = std::max(min, (long)std::numeric_limits<int>::min());
    max = std::min(max, (long)std::numeric_limits<int>::max());
    sprintf(minmax, "%s: %ld, %s: %ld", GCS(MSG_Min), min, GCS(MSG_Max), max);
    QString label(contents? contents : GCS(MSG_EnterInteger));
    if (!label.isEmpty()) {
        label.append("\n");
    }
    label.append(minmax);
    
    QInputDialog dialog(GuiFindParentWindow(0, pr));
    dialog.setWindowTitle(title? title : GCS(MSG_APP));
    dialog.setLabelText(label);
    if (contents) {
        long val;
        if (StrToInt(contents, &val)) {
            dialog.setIntValue(val);
        }
    }
    dialog.setInputMode(QInputDialog::IntInput);
    dialog.setIntMinimum(min);
    dialog.setIntMaximum(max);
    dialog.setOkButtonText(GuiMarkShortcut(oktext? oktext : GCS(MSG_Ok)));
    dialog.setCancelButtonText(GuiMarkShortcut(GCS(MSG_Cancel)));
    
    bool ret = false;
    if (dialog.exec() == QDialog::Accepted) {
        if (result) {
            *result = dialog.intValue();
        }
        ret = true;
    }
    return ret;    
}

class CountLabel : public QLabel
{
    Q_OBJECT
    
public:
    CountLabel(QWidget *parent, SFWView *sfw) : QLabel(parent)
    {
        sfw_ = sfw;
        setFrameShadow(QFrame::Sunken);
        setFrameShape(QFrame::StyledPanel);
        setMargin(1);
        setMinimumWidth(50);
        connect(sfw->model(),
                SIGNAL(dataChanged(const QModelIndex &, const QModelIndex &)), 
                this, SLOT(update()));
        update();
    }
    
public slots:
    void update()
    {
        setText(QString::number(sfw_->countSelected()));
    }
    
private:
    SFWView *sfw_;
};

class AskChoiceDialog : public CommonDialog
{
    Q_OBJECT
    
public:
    AskChoiceDialog(Project *pr, const char *caption, const char *oktext,
                    Element *labels, Element *titles, bool *selected, 
                    bool multiselect, bool stringmode, long *index, char **contents);

    void getResult(bool *selected, long *index, char **contents);
        
private:
    virtual bool acceptEdits();
    
    bool multiselect_;
    bool stringmode_;
    SFWView *tree_;
    QLineEdit *edit_;
};

AskChoiceDialog::AskChoiceDialog(Project *pr, const char *caption, const char *oktext,
                                 Element *labels, Element *titles, bool *selected, 
                                 bool multiselect, bool stringmode, long *index, char **contents) :
    CommonDialog(GuiFindParentWindow(0, pr))
{
    if (titles != 0 && titles->el_Type == TYPE_NIL) {
        titles = 0;
    }

    multiselect_ = multiselect;
    stringmode_ = stringmode;
    
    const char *title = ProjectTitle(pr);
    setWindowTitle(title? title : GCS(MSG_APP));
    setOkText(GuiMarkShortcut(oktext? oktext : GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));
    
    if (!caption) {
	if (multiselect) {
	    caption = GCS(MSG_MultiSelectItems);
	} else if (stringmode) {
	    caption = GCS(MSG_EnterString);
	} else {
	    caption = GCS(MSG_SelectItem);
	}
    }
    QLabel *label = new QLabel(this);
    label->setText(caption);
    label->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
    contentsLayout()->addWidget(label);
    
    tree_ = new SFWView(this);
    tree_->setShowTitles(titles != 0);
    tree_->setCheckBoxes(multiselect);
    tree_->setTitlesBody(titles, labels);
    if (multiselect_ && selected) {
        tree_->setSelected(selected);
    }
    QVBoxLayout *treeLayout = new QVBoxLayout;
    treeLayout->setSpacing(2);
    treeLayout->addWidget(tree_);
    contentsLayout()->addLayout(treeLayout);
    
    edit_ = 0;
    
    if (multiselect_) {
        QLabel *label = new QLabel(this);
        label->setText(GCS(MSG_ItemsSelected));
        label->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
        
        CountLabel *count = new CountLabel(this, tree_);
        count->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
        
        QHBoxLayout *layout = new QHBoxLayout;
        layout->addWidget(label);
        layout->addWidget(count);
        contentsLayout()->addLayout(layout);
    } else {
        int active = (index && *index >= 0)? *index + 1 : 0;
        tree_->setActiveLine(active);
        QObject::connect(tree_, SIGNAL(activated(const QModelIndex &)),
                         this, SLOT(accept()));
    }
    
    if (stringmode_) {
        edit_ = new LineEditTreeView(this, tree_);
        edit_->setText(contents? *contents : 0);
        edit_->setMaxLength(MAXLEN);
        edit_->setFocus();
        edit_->selectAll();
        treeLayout->addWidget(edit_);
    }
}

void AskChoiceDialog::getResult(bool *selected, long *index, char **contents)
{
    if (multiselect_) {
        if (selected) {
            tree_->getSelected(selected);
        }
    } else if (stringmode_) {
        if (contents) {
            static QByteArray utf8 = edit_->text().toUtf8();
            *contents = utf8.data();
        }
    } else {
        if (index) {
            *index = tree_->activeLine() - 1;
        }
    }
}

bool AskChoiceDialog::acceptEdits()
{
    bool rc = multiselect_ || stringmode_ || tree_->activeLine();
    if (!rc) {
        GuiBeep();
    }
    return rc;
}

bool AskChoice(Project *pr, const char *caption, const char *oktext,
               Element *labels, Element *titles, bool *selected, 
               bool multiselect, bool stringmode, long *index, char **contents)
{
    GuiBusyCursor busy;
    
    if (multiselect && stringmode) {
        return false;			// illegal combination
    }
    if (CountListLength(labels) == 0) {
        return false;			// nothing to select (stringmode?)
    }

    AskChoiceDialog dialog(pr, caption, oktext, labels, titles,
                           selected, multiselect, stringmode, index, contents);
    bool rc = dialog.exec() == QDialog::Accepted;
    if (rc) {
        dialog.getResult(selected, index, contents);
    }
 
    return rc;
}

class AskMultiDialog : public CommonDialog
{
    Q_OBJECT
    
public:
    AskMultiDialog(Project *pr, const char *title, const char *oktext, AskMultiInfo *ami);
    
    bool getResult();
    
private:
    AskMultiInfo *ami_;
    
    virtual bool acceptEdits() { return CheckAskMultiInfo(ami_, this); }
};

AskMultiDialog::AskMultiDialog(Project *pr, const char *title, const char *oktext, AskMultiInfo *ami) 
    : CommonDialog(GuiFindParentWindow(0, pr))
{
    ami_ = ami;
    
    setWindowTitle(title? title : GCS(MSG_EditForm));
    setOkText(GuiMarkShortcut(oktext? oktext : GCS(MSG_Ok)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

#ifdef USE_FORMLAYOUT
    QFormLayout *form = new QFormLayout;
    // It is yet unclear how to limit the vertical size of QFormLayout.
    dialog.contentsLayout()->addLayout(form);
#else
    GroupLayout *group = new GroupLayout(2);
    contentsLayout()->addLayout(group);
#endif
    
    AskMultiItem *a = ami->ami_Items;
    for (int i = 0; i < ami->ami_Num; i++, a++) {
        GuiObject go = 0;
        
        switch(a->ami_Type) {
            case TYPE_STRING:
            {
                QLineEdit *edit = new QLineEdit(this);
                edit->setText(a->ami_Contents.string.string);
                if (a->ami_Contents.string.secret) {
                    edit->setEchoMode(QLineEdit::Password);
                }
                edit->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
                go = edit;
                break;
            }
                
            case TYPE_MEMO:
            {
                TextEditor *edit = new TextEditor(pr, false, this);
                if (a->ami_Contents.mi) {
                    edit->setText(a->ami_Contents.mi);
                }
                go = edit;
                break;
            }
                
            case TYPE_INTEGER:
            {
                QLineEdit *edit = new QLineEdit(this);
                if (!a->ami_Nil) {
                    edit->setText(IntToStr(a->ami_Contents.integer));
                }
                go = edit;
                break;
            }
                
            case TYPE_REAL:
            {
                QLineEdit *edit = new QLineEdit(this);
                if (!a->ami_Nil) {
                    edit->setText(DoubleToStr(a->ami_Contents.real, -1,
                                              GetPrefsValue(pr, PREFS_REALFORMAT)));
                }
                go = edit;
                break;
            }
                
            case TYPE_DATE:
            {
                QLineEdit *edit = new QLineEdit(this);
                if (!a->ami_Nil) {
                    edit->setText(DateToString(a->ami_Contents.date,
                                               GetPrefsValue(pr, PREFS_DATEFORMAT)));
                }
                go = edit;
                break;
            }
                
            case TYPE_TIME:
            {
                QLineEdit *edit = new QLineEdit(this);
                if (!a->ami_Nil) {
                    edit->setText(TimeToString(a->ami_Contents.time, FORMAT_HOURMINSEC));
                }
                go = edit;
                break;
            }
                
            case TYPE_BOOL:
            {
                QCheckBox *check = new QCheckBox(this);
                check->setChecked(!a->ami_Nil);
                go = check;
                break;
            }
                
            case TYPE_CHOICE:
            {
                if (!a->ami_Contents.choice.ischoicelist) {
                    ComboBox *combo = new ComboBox(this, a->ami_Contents.choice.labels);
                    if (!a->ami_Nil) {
                        combo->setCurrentIndex(a->ami_Contents.choice.active);
                    }
                    go = combo;
                } else {
                    QListWidget *list = new QListWidget(this);
                    for (char **s = a->ami_Contents.choice.labels; *s; s++) {
                        list->addItem(*s);
                    }                    
                    if (!a->ami_Nil) {
                        list->setCurrentRow(a->ami_Contents.choice.active);
                    }
                    go = list;
                }
                break;
            }
                
            case TYPE_OPTIONS:
            {
                QListWidget *list = new QListWidget(this);
                char **labels = a->ami_Contents.options.labels;
                const bool *selected = a->ami_Contents.options.selected;
                for (int i = 0; labels[i]; i++) {
                    QListWidgetItem *item = new QListWidgetItem(labels[i], list);
                    item->setFlags(Qt::ItemIsSelectable | Qt::ItemIsUserCheckable | Qt::ItemIsEnabled);
                    item->setCheckState(selected[i]? Qt::Checked : Qt::Unchecked);
                }
                go = list;
                break;
            }
                
            case NOTYPE:                /* title only objects */
            default:
                break;
        }
        
        if (!go) {
            QLabel *label = new QLabel(this);
            label->setFrameShadow(QFrame::Sunken);
            label->setFrameShape(QFrame::StyledPanel);
            label->setMargin(1);
            label->setText(a->ami_Title);
            label->setAlignment(Qt::AlignCenter);
            label->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
            go = label;
#ifdef USE_FORMLAYOUT
            form->addRow(label);
#else
            group->addWidget(label, 2);
#endif
        } else if (a->ami_Title) {
#ifdef USE_FORMLAYOUT
            form->addRow(GuiMarkShortcut(a->ami_Title), go);
#else
            group->addLabelAndWidget(GuiMarkShortcut(a->ami_Title), go);
#endif
        } else {
#ifdef USE_FORMLAYOUT
            form->addRow(go);
#else
            group->addWidget(go, 2);
#endif
        }
        
        if (go) {
            go->setToolTip(a->ami_BubbleHelp);
        }
        a->ami_GuiObject = go;
    }
}

bool AskMultiDialog::getResult()
{
    MemPool *memPool = ami_->ami_MemPool;
    AskMultiItem *a = ami_->ami_Items;
    bool rc = true;
    
    for (int i = 0; i < ami_->ami_Num && rc; i++, a++) {
        const char *cont;
    
        switch (a->ami_Type) {
            case TYPE_STRING:
                FreeString(memPool, a->ami_Contents.string.string);
                a->ami_Nil = false;
                cont = GuiStringContents(a->ami_GuiObject);
                a->ami_Contents.string.string = AllocString(memPool, cont);
                if (!a->ami_Contents.string.string) {
                    rc = false;
                }
                break;
                
            case TYPE_MEMO:
                FreeString(memPool, a->ami_Contents.mi);
                a->ami_Nil = false;
                a->ami_Contents.mi = GuiEditorText(a->ami_GuiObject, memPool);
                if (!a->ami_Contents.mi) {
                    rc = false;
                }
                break;
                
            case TYPE_INTEGER:
                cont = GuiStringContents(a->ami_GuiObject);
                a->ami_Nil = !StrToInt(cont, &a->ami_Contents.integer);
                break;
                
            case TYPE_REAL:
                cont = GuiStringContents(a->ami_GuiObject);
                a->ami_Nil = !StrToDouble(cont, &a->ami_Contents.real);
                break;
                
            case TYPE_DATE:
                cont = GuiStringContents(a->ami_GuiObject);
                a->ami_Nil = !StringToDate(cont, &a->ami_Contents.date);
                break;
                
            case TYPE_TIME:
                cont = GuiStringContents(a->ami_GuiObject);
                a->ami_Nil = !StringToTime(cont, 
                                           FORMAT_HOURMINSEC, &a->ami_Contents.time);
                break;
                
            case TYPE_BOOL:
                a->ami_Nil = dynamic_cast<QCheckBox *>(a->ami_GuiObject)->checkState() != Qt::Checked;
                break;
                
            case TYPE_CHOICE:
                if(!a->ami_Contents.choice.ischoicelist) {
                    a->ami_Nil = false;
                    a->ami_Contents.choice.active = dynamic_cast<QComboBox *>(a->ami_GuiObject)->currentIndex();
                } else {
                    a->ami_Contents.choice.active = GuiListActive(a->ami_GuiObject);
                    a->ami_Nil = (a->ami_Contents.choice.active < 0);
                }
                break;
                
            case TYPE_OPTIONS: 
            {
                QListWidget *list = dynamic_cast<QListWidget *>(a->ami_GuiObject);
                bool *selected = a->ami_Contents.options.selected;

                for (int i = 0; i < list->count(); i++) {
                    selected[i] = list->item(i)->checkState() == Qt::Checked;
                }
                a->ami_Nil = false;
                break;
            }
        }
    }
    return rc;
}

bool AskMulti(struct Project *pr, const char *title, const char *oktext,
              AskMultiInfo *ami)
{
    GuiBusyCursor busy;
    
    if (!ami || ami->ami_Num == 0) {
	return FALSE;
    }

    AskMultiDialog dialog(pr, title, oktext, ami);
    
    bool rc = dialog.exec() == QDialog::Accepted;
    if (rc) {
        rc = dialog.getResult();
    }
    
    return rc;
}

#include "qtRequesters.moc"
