//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtSearch.cc
//
// Creation date: 04.12.2016
// Last modified: 05.12.2016
//

#include "search.h"
#include "qt/classes/qtGroupLayout.h"
#include "qt/qtUtil.h"
#include <QCheckBox>
#include "table.h"
#include "util.h"
#include "appstrings_cat.h"

class EditSearchForDialog : public CommonDialog
{
public:
    EditSearchForDialog(struct SearchFor *sf);

private:
    Project *pr_;
    struct SearchFor *sf_;
    QLineEdit *pattern_;
    QCheckBox *sensitive_;
    QCheckBox *all_;
    RadioBox *direction_;
    RadioBox *start_;

    virtual bool acceptEdits()
    {
        QByteArray buffer = pattern_->text().toUtf8();
	size_t maxlen = sizeof(sf_->sf_Pattern) - 1;
	CodesetStrNBytesCopy(sf_->sf_Pattern, buffer.data(), maxlen);
	sf_->sf_CaseSensitive = sensitive_->isChecked();
	sf_->sf_AllFields = all_->isChecked();
	sf_->sf_Forward = direction_->checkedId() == 0;
	sf_->sf_FirstLastRecord = start_->checkedId() == 0;
        return true;
    }
};

EditSearchForDialog::EditSearchForDialog(struct SearchFor *sf) :
    CommonDialog(GuiFindParentWindow(0, TableProject(sf->sf_Table)))
{
    pr_ = TableProject(sf->sf_Table);
    sf_ = sf;

    GroupLayout *group = new GroupLayout(2);
    group->setContentsMargins(0, 0, 0, 0);
    contentsLayout()->addLayout(group);

    pattern_ = new QLineEdit(this);
    pattern_->setMaxLength(sizeof(sf->sf_Pattern) - 1);
    pattern_->setText(sf->sf_Pattern);
    pattern_->setToolTip(GCS(MSG_SearchPatternBH));
    pattern_->setFocus();
    group->addLabelAndWidget(GuiMarkShortcut(GCS(MSG_SearchPattern)), pattern_);

    group = new GroupLayout(2);
    group->setContentsMargins(0, 0, 0, 0);
    contentsLayout()->addLayout(group);

    sensitive_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_CaseSensitive)), this);
    sensitive_->setChecked(sf->sf_CaseSensitive);
    sensitive_->setToolTip(GCS(MSG_CaseSensitiveBH));
    group->addWidget(sensitive_);

    all_ = new QCheckBox(GuiMarkShortcut(GCS(MSG_AllFields)), this);
    all_->setChecked(sf->sf_Field == 0 || sf->sf_AllFields);
    all_->setEnabled(sf->sf_Field != 0);
    all_->setToolTip(GCS(MSG_AllFieldsBH));
    group->addWidget(all_);

    group = new GroupLayout(2);
    group->setContentsMargins(0, 0, 0, 0);
    contentsLayout()->addLayout(group);

    static char *dirstr[3];
    dirstr[0] = GCS(MSG_Forward);
    dirstr[1] = GCS(MSG_Backward);
    dirstr[2] = NULL;
    direction_ = new RadioBox(this, dirstr, GCS(MSG_SearchDirection), true);
    direction_->setCheckedId(sf->sf_Forward? 0 : 1);
    group->addWidget(direction_);

    static char *startstr[3];
    startstr[0] = GCS(MSG_FirstLastRecord);
    startstr[1] = GCS(MSG_CurrentRecord);
    startstr[2] = NULL;
    start_ = new RadioBox(this, startstr, GCS(MSG_StartWith), true);
    start_->setCheckedId(sf->sf_FirstLastRecord? 0 : 1);
    group->addWidget(start_);

    setOkText(GuiMarkShortcut(GCS(MSG_Find)));
    setCancelText(GuiMarkShortcut(GCS(MSG_Cancel)));

    setWindowTitle(GCS(MSG_Find));

    GuiSetWindowProject(this, pr_);
    GuiSetHelpNode(this, "Search dialog");
}

struct SearchFor *GuiSearchFor(struct SearchFor *sf)
{
    if (sf && sf->sf_Table) {
        GuiBusyCursor busy;
        EditSearchForDialog *dialog = new EditSearchForDialog(sf);
        if (dialog) {
            int response = dialog->exec();
            delete dialog;
            if (response == QDialog::Accepted) {
                return sf;
            }
        }
    }
    return 0;
}

