//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtStructureEditor.cc
//
// Creation date: 13.12.2015
// Last modified: 05.06.2022
//

#include "system.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include "qt/qtStructureEditor.h"
#include "qt/qtStructureEditorUtil.h"
#include "qt/qtProjectWindow.h"
#include "qt/qtEditTable.h"
#include "qt/qtEditField.h"
#include "qt/qtEditDisplay.h"
#include "qt/qtEditDisplayTable.h"
#include "qt/qtEditDisplayField.h"
#include "qt/qtMenus.h"
#include "qt/classes/qtGroupLayout.h"
#include "qt/classes/qtSortableListModel.h"
#include <QApplication>
#include <QLabel>
#include <QMenuBar>
#include <QMimeData>
#include <QPainter>
#include <QSplitter>
#include <QTreeView>
#include "main.h"
#include "structureeditor.h"
#include "table.h"
#include "buildgui.h"
#include "preferences.h"
#include "requesters.h"
#include "name.h"
#include "error.h"
#include "appstrings_cat.h"

const char MimeTypeDisplayObject[] = "BeeBase/DisplayObject";

struct MimeDataDisplayObject
{
    MimeDataDisplayObject() {}
    MimeDataDisplayObject(DisplayObject *dio) {
        projectName = ProjectName(DisplayObjectProject(dio));
        displayObjectNum = DisplayObjectNum(dio);
    }

    DisplayObject *displayObject() const {
        Project *pr;
        FORNODE(pr, &ProjectList) {
            if (projectName == ProjectName(pr)) {
                return FindDisplayObjectNum(pr, displayObjectNum);
            }
        }
        return 0;
    }

    QString projectName;
    int displayObjectNum;
};

QDataStream &operator>>(QDataStream &s, MimeDataDisplayObject &mddio)
{
    s >> mddio.projectName >> mddio.displayObjectNum;
    return s;
}

QDataStream &operator<<(QDataStream &s, const MimeDataDisplayObject &mddio)
{
    s << mddio.projectName << mddio.displayObjectNum;
    return s;
}

static QList<DisplayObject *> mimeDataDisplayObjects(const QMimeData *data)
{
    QList<DisplayObject *>dios;
    if (data->hasFormat(MimeTypeDisplayObject)) {
        QByteArray encodedData = data->data(MimeTypeDisplayObject);
        QDataStream stream(&encodedData, QIODevice::ReadOnly);
        while (!stream.atEnd()) {
            MimeDataDisplayObject mddio;
            stream >> mddio;
            DisplayObject *dio = mddio.displayObject();
            if (dio) {
                dios.push_back(dio);
            }
        }
    }
    return dios;
}

static QList<Field *> mimeDataFields(const QMimeData *data)
{
    QList<Field *>fields;
    if (data->hasFormat(MimeTypeField)) {
        QByteArray encodedData = data->data(MimeTypeField);
        QDataStream stream(&encodedData, QIODevice::ReadOnly);
        while (!stream.atEnd()) {
            MimeDataField mdfield;
            stream >> mdfield;
            Field *field = mdfield.field();
            if (field) {
                fields.push_back(field);
            }
        }
    }
    return fields;
}

static QList<Table *> mimeDataTables(const QMimeData *data)
{
    QList<Table *>tabs;
    if (data->hasFormat(MimeTypeTable)) {
        QByteArray encodedData = data->data(MimeTypeTable);
        QDataStream stream(&encodedData, QIODevice::ReadOnly);
        while (!stream.atEnd()) {
            MimeDataTable mdtab;
            stream >> mdtab;
            Table *tab = mdtab.table();
            if (tab) {
                tabs.push_back(tab);
            }
        }
    }
    return tabs;
}

static DisplayObject *mimeDataSingleDisplayObject(const QMimeData *data, 
        bool rootDisplay)
{
    QList<Table *>tabs = mimeDataTables(data);
    QList<Field *>fields = mimeDataFields(data);
    QList<DisplayObject *>dios = mimeDataDisplayObjects(data);
    if (tabs.size() + fields.size() + dios.size() != 1) {
        return 0;
    }
    if (rootDisplay && !tabs.isEmpty()) {
        return TableDisplayObject(tabs.front());
    }
    if (!rootDisplay && !fields.isEmpty()) {
        return FieldDisplayObject(fields.front());
    }
    if (!dios.isEmpty()) {
        return dios.front();
    }
    return 0;
}

static QString displayObjectText(DisplayObject *dio)
{
    if (!dio) {
        return QString();
    }

    QString text;
    switch (DisplayObjectType(dio)) {
	case DO_TYPE_TABLE:
        {
            struct Table *tab = DisplayObjectTable(dio);
            const char *name = TableName(tab);
            const char *title = DisplayObjectTitle(dio);
            
            if (title && title[0] && strcmp(name, title) != 0 &&
                    DisplayTableHasPanel(dio)) {
                text = QString("<b>%1</b>(%2)").arg(name).arg(QString(title).toHtmlEscaped());
            } else {
                text = QString("<b>%1</b>").arg(name);
            }
            break;
        }
            
	case DO_TYPE_FIELD:
        {
            struct Field *field = DisplayFieldField(dio);
            const char *name = FieldName(field);
            const char *title = DisplayObjectTitle(dio);
            
            if (title && title[0] && strcmp(name, title) != 0) {
                text = QString("<b>%1</b>(%2)").arg(name).arg(QString(title).toHtmlEscaped());
            } else {
                text = QString("<b>%1</b>").arg(name);
            }
            break;
        }
            
        case DO_TYPE_TEXT:
        {
            const char *label = GCS(MSG_Text);
            const char *title = DisplayObjectTitle(dio);
            
            if (title && title[0]) {
                text = QString("%1(%2)").arg(label).arg(QString(title).toHtmlEscaped());
	    } else {
	        text = label;
	    }
            break;
        }
            
	case DO_TYPE_SPACE:
            text = GCS(MSG_Space);
            break;
            
	case DO_TYPE_BALANCE:
            text = GCS(MSG_StructureEditor_Balance);
            break;
            
	case DO_TYPE_IMAGE:
            text = GCS(MSG_Image);
            break;
            
	case DO_TYPE_REGISTER:
	    text = GCS(MSG_StructureEditor_Register);
	    break;
            
	case DO_TYPE_GROUP:
        {
            /*	    bool border = DisplayGroupBorder(dio); */
	    bool horiz = DisplayGroupHoriz(dio);
	    const char *title = DisplayObjectTitle(dio);
	    const char *label = GCS(horiz?
                                    MSG_StructureEditor_Display_HGroup 
                                    : MSG_StructureEditor_Display_VGroup);
            
            if (title && title[0]) {
                text = QString("%1(%2)").arg(label).arg(QString(title).toHtmlEscaped());
	    } else {
		text = label;
	    }
	    break;
        }
            
	case DO_TYPE_WINDOW:
        {
            const char *title = DisplayObjectTitle(dio);
	    const char *label = GCS(MSG_Window);
            
            if (title && title[0]) {
                text = QString("%1(%2)").arg(label).arg(QString(title).toHtmlEscaped());
	    } else {
		text = label;
	    }
	    break;
        }
            
        default:
            break;
    }
    
    // Replace all white space by single spaces.
    text = text.simplified();
    return text;
}

class DisplayObjectsModel : public QAbstractItemModel
{
    Q_OBJECT
    
public:
    static const int OpenRole = Qt::UserRole;
    
    DisplayObjectsModel(QObject *parent) 
        : QAbstractItemModel(parent), rootDisplay_(false) {} 
    
    QModelIndex index(int row, int column,
                      const QModelIndex &parent = QModelIndex()) const
    {
        const Entry *entry = getEntry(parent);
        if (entry && row >= 0 && row < entry->childs.size()) {
            return createIndex(row, column, (void *)&entry->childs[row]);
        }
        return QModelIndex();
    }
    
    QModelIndex parent(const QModelIndex &index) const
    {
        const Entry *entry = getEntry(index);
        if (entry && entry->parent && entry->parent != &root_) {
            return createIndex(entry->parent->row, 0, entry->parent);
        }
        return QModelIndex();
    }
    
    int rowCount(const QModelIndex &parent = QModelIndex()) const
    {
        const Entry *entry = getEntry(parent);
        if (entry) {
            return entry->childs.size();
        }
        return 0;
    }
    
    int columnCount(const QModelIndex & parent = QModelIndex()) const
    {
        return 1;
    }
    
    DisplayObject *displayObject(const QModelIndex &index) const
    {
        const Entry *entry = getEntry(index);
        if (entry) {
            return entry->dio;
        }
        return 0;
    }
    
    QModelIndex displayObjectIndex(const DisplayObject *dio) const {
        const Entry *entry = findEntry(dio, root_);
        if (entry) {
            return createIndex(entry->row, 0, (void *)entry);
        }
        return QModelIndex();
    }
    
    QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const
    {
        const Entry *entry = getEntry(index);
        if (entry) {
            switch (role) {
                case Qt::DisplayRole:
                    return displayObjectText(entry->dio);

                case OpenRole:
                    return DisplayObjectOpen(entry->dio);
            }
        }
        return QVariant();
    }
    
    bool setData(const QModelIndex &index, const QVariant &value, int role)
    {
        const Entry *entry = getEntry(index);
        if (entry) {
            switch (role) {
                case OpenRole:
                    SetDisplayObjectOpen(entry->dio, value.toBool());
                    return true;

            }
        }
        return false;
    }    

    void setList(List *childs)
    { 
        beginResetModel();
        rootDisplay_ = true;
        root_ = Entry();
        setRec(childs, &root_);
        endResetModel();
    }
    
    void setDisplayObject(DisplayObject *dio)
    {
        beginResetModel();
        rootDisplay_ = false;
        root_ = Entry();
        setRec(0, dio, &root_, true);
        endResetModel();
    }

    virtual Qt::ItemFlags flags(const QModelIndex & index) const
    {
        if (index.isValid()) {
            return Qt::ItemIsSelectable | Qt::ItemIsDragEnabled | 
                Qt::ItemIsDropEnabled | Qt::ItemIsEnabled;
        } else {
            return Qt::ItemIsDropEnabled; 
        }
    }
    
    QStringList mimeTypes() const
    {
        QStringList types;
        types << MimeTypeDisplayObject;
        return types;
    }
    
    QMimeData *mimeData(const QModelIndexList &indexes) const
    {
        QByteArray data;
        QDataStream stream(&data, QIODevice::WriteOnly);
        foreach (const QModelIndex &index, indexes) {
            DisplayObject *dio = displayObject(index);
            if (dio) {
                stream << MimeDataDisplayObject(dio);
            }
        }
        QMimeData *mimeData = new QMimeData();
        mimeData->setData(MimeTypeDisplayObject, data);
        return mimeData;
    }

    Qt::DropActions supportedDropActions() const
    {
        return Qt::MoveAction | Qt::CopyAction;
    }

    bool canDropMimeData(const QMimeData *data, Qt::DropAction action,
            int row, int column, const QModelIndex &parent) const
    {
        if (action == Qt::IgnoreAction) {
            return true;
        }
        return mimeDataSingleDisplayObject(data, rootDisplay_) != 0;
    }

    bool dropMimeData(const QMimeData *data, Qt::DropAction action, 
                      int row, int column, const QModelIndex &parent)
    {
        if (action == Qt::IgnoreAction) {
            return true;
        }
        DisplayObject *dio = mimeDataSingleDisplayObject(data, rootDisplay_);
        if (!dio) {
            return false;
        }
        DisplayObject *dioparent;
        bool isPred;
        Entry *entry = getEntry(parent);
        if (!entry) {
            // we assume the user wants to place it at the very end
            if (root_.childs.isEmpty()) {
                dioparent = root_.dio;
                isPred = false;
            } else {
                dioparent = root_.childs.last().dio;
                isPred = true;
            }
        } else if (row <= 0) {
            dioparent = entry->dio;
            isPred = false;
        } else {
            dioparent = entry->childs[row - 1].dio;
            isPred = rootDisplay_ || 
                DisplayObjectType(dioparent) != DO_TYPE_TABLE;
        }
        emit droppedDisplayObject(dio, dioparent, isPred);
        return true;
    }

signals:
    void droppedDisplayObject(DisplayObject *dio, DisplayObject *dioparent,
            bool isPred);

public slots:
    void setOpen(const QModelIndex &index) {
        setData(index, true, OpenRole);
    }

    void setClosed(const QModelIndex &index) {
        setData(index, false, OpenRole);
    }

private:
    struct Entry {
        int row;
        DisplayObject *dio;
        Entry *parent;
        QList<Entry> childs;
        
        Entry() : row(0), dio(0), parent(0) {}
        Entry(int r, DisplayObject *d, Entry *p) : row(r), dio(d), parent(p) {} 
    };

    bool rootDisplay_;
    Entry root_;

    Entry *getEntry(const QModelIndex &index) {
        if (index.isValid()) {
            return (Entry *)index.internalPointer();
        }
        return &root_;
    }
    
    const Entry *getEntry(const QModelIndex &index) const {
        if (index.isValid()) {
            return (const Entry *)index.internalPointer();
        }
        return &root_;
    }
    
    const Entry *findEntry(const DisplayObject *dio, const Entry &entry) const {
        if (entry.dio == dio) {
            return &entry;
        }
        for (int i=0; i < entry.childs.size(); ++i) {
            const Entry *child = findEntry(dio, entry.childs[i]);
            if (child) {
                return child;
            }
        }
        return 0;
    }
        
    void setRec(List *childs, Entry *parent)
    {
        if (childs) {
            DisplayObject *dio;
            int row = 0;
            FORNODE (dio, childs) {
                if (DisplayObjectVisible(dio)) {
                    setRec(row, dio, parent, false);
                    row++;
                }
            }
        }
    }
    
    void setRec(int row, DisplayObject *dio, Entry *parent, bool recurseTable)
    {
        parent->childs.append(Entry(row, dio, parent));
        short type = DisplayObjectType(dio);
        if (recurseTable || type != DO_TYPE_TABLE) {
            setRec(DisplayObjectObjects(dio), &parent->childs.back());
        }
    }
};

class DisplayObjectsView : public QTreeView
{
    Q_OBJECT
    
public:
    DisplayObjectsView(QWidget *parent) : QTreeView(parent) {
        setItemDelegate(new HtmlDelegate(this));
    }
    
public slots:
    void loadOpenState() {
        if (model()) {
            loadOpenStateRec(QModelIndex());
        }
    }
    
    void saveOpenState() {
        if (model()) {
            saveOpenStateRec(QModelIndex());
        }
    }
    
signals:
    void newCurrent(const QModelIndex &);
    
protected slots:
    void currentChanged(const QModelIndex &current, 
                        const QModelIndex &previous)
    {
        QTreeView::currentChanged(current, previous);
        emit newCurrent(current);
    }

private:
    void loadOpenStateRec(const QModelIndex &index) {
        int rows = model()->rowCount(index);
        if (rows > 0) {
            bool open = model()->data(index, 
                                      DisplayObjectsModel::OpenRole).toBool();
            setExpanded(index, open);
            for (int i = 0; i < rows; ++i) {
                QModelIndex child = model()->index(i, 0, index);
                if (child.isValid()) {
                    loadOpenStateRec(child);
                }
            }
        }
    }
    
    void saveOpenStateRec(const QModelIndex &index) {
        int rows = model()->rowCount(index);
        if (rows > 0) {
            bool open = isExpanded(index);
            model()->setData(index, open, DisplayObjectsModel::OpenRole);
            for (int i = 0; i < rows; ++i) {
                QModelIndex child = model()->index(i, 0, index);
                if (child.isValid()) {
                    saveOpenStateRec(child);
                }
            }
        }
    }
};

class CustomFieldsModel : public FieldsModel
{
    Q_OBJECT

public:
    CustomFieldsModel(QWidget *view) : FieldsModel(view), view_(view) {}

    Qt::DropActions supportedDropActions() const
    {
        return Qt::MoveAction | Qt::CopyAction;
    }
    
    QStringList mimeTypes() const
    {
        QStringList types(FieldsModel::mimeTypes());
        types << MimeTypeField;
        return types;
    }

    QMimeData *mimeData(const QModelIndexList &indexes) const
    {
        QByteArray data;
        QDataStream stream(&data, QIODevice::WriteOnly);
        foreach (const QModelIndex &index, indexes) {
            stream << MimeDataField(field(index.row()));
        }
        QMimeData *mimeData = FieldsModel::mimeData(indexes);
        if (!mimeData) {
            mimeData = new QMimeData();
        }
        mimeData->setData(MimeTypeField, data);
        return mimeData;
    }

    bool canDropMimeData(const QMimeData *data, Qt::DropAction action,
            int row, int column, const QModelIndex &parent) const
    {
        // For simulating internal move.
        if (FieldsModel::canDropMimeData(data, action, row, column, parent) &&
                data->hasFormat(MimeTypeField)) {
            return true;
        }
        if (action == Qt::IgnoreAction) {
            return true;
        }
        QList<DisplayObject *> dios = mimeDataDisplayObjects(data);
        if (dios.isEmpty()) {
            return false;
        }
        foreach (DisplayObject *dio, dios) {
            if (DisplayObjectType(dio) != DO_TYPE_FIELD) {
                return false;
            }
        }
        return true;
    }

    bool dropMimeData(const QMimeData *data, Qt::DropAction action, 
                      int row, int column, const QModelIndex &parent)
    {
        if (FieldsModel::dropMimeData(data, action, row, column, parent)) {
            return true;
        }
        if (action == Qt::IgnoreAction) {
            return true;
        }
        emit droppedDisplayFields();
        return true;
    }

signals:
    void droppedDisplayFields();

protected:
    virtual QVariant data(const QModelIndex &index,
            int role = Qt::DisplayRole) const
    {
        static CatalogString typenames[] = {
            MSG_Err_Internal,
            MSG_Integer,
            MSG_Real,
            MSG_Boolean,
            MSG_Err_Internal,
            MSG_Memo,
            MSG_Date,
            MSG_Time,
            MSG_Err_Internal,
            MSG_Button,
            MSG_Choice,
            MSG_Reference,
            MSG_String,
            MSG_Virtual,
        };
        Field *fld = field(index);
        const char *name = EmptyNullStr(FieldName(fld));
        short type = FieldType(fld);

        if (fld != NULL) {
            switch (role) {
                case Qt::DisplayRole:
                   return QString("%1(%2)").arg(name).arg(GCS(typenames[type]));
                case Qt::FontRole:
                   if (!IsMBName(name)) {
                       QFont font = view_->font();
                       font.setItalic(true);
                       return font;
                   }
                   break;
                }
        }
        return QVariant();
    }

private:
    QWidget *view_;
};

class CustomTablesModel : public TablesModel
{
    Q_OBJECT

public:
    CustomTablesModel(QWidget *view) : TablesModel(view), view_(view) {}

    Qt::DropActions supportedDropActions() const
    {
        return Qt::MoveAction | Qt::CopyAction;
    }
    
    QStringList mimeTypes() const
    {
        QStringList types(TablesModel::mimeTypes());
        types << MimeTypeTable;
        return types;
    }

    QMimeData *mimeData(const QModelIndexList &indexes) const
    {
        QByteArray data;
        QDataStream stream(&data, QIODevice::WriteOnly);
        foreach (const QModelIndex &index, indexes) {
            stream << MimeDataTable(table(index.row()));
        }
        QMimeData *mimeData = TablesModel::mimeData(indexes);
        if (!mimeData) {
            mimeData = new QMimeData();
        }
        mimeData->setData(MimeTypeTable, data);
        return mimeData;
    }

    bool canDropMimeData(const QMimeData *data, Qt::DropAction action,
            int row, int column, const QModelIndex &parent) const
    {
        // For simulating internal move.
        if (TablesModel::canDropMimeData(data, action, row, column, parent) &&
                data->hasFormat(MimeTypeTable)) {
            return true;
        }
        if (action == Qt::IgnoreAction) {
            return true;
        }
        QList<DisplayObject *> dios = mimeDataDisplayObjects(data);
        if (dios.isEmpty()) {
            return false;
        }
        foreach (DisplayObject *dio, dios) {
            if (DisplayObjectType(dio) != DO_TYPE_TABLE) {
                return false;
            }
        }
        return true;
    }

    bool dropMimeData(const QMimeData *data, Qt::DropAction action, 
                      int row, int column, const QModelIndex &parent)
    {
        if (TablesModel::dropMimeData(data, action, row, column, parent)) {
            return true;
        }
        if (action == Qt::IgnoreAction) {
            return true;
        }
        emit droppedDisplayTables();
        return true;
    }

    virtual QVariant data(const QModelIndex &index,
            int role = Qt::DisplayRole) const
    {
        Table *tab = table(index);
        const char *name = EmptyNullStr(TableName(tab));

        if (tab != NULL) {
            switch (role) {
                case Qt::DisplayRole:
                    return name;
                case Qt::FontRole:
                    if (!IsMBName(name)) {
                        QFont font = view_->font();
                        font.setItalic(true);
                        return font;
                    }
                    break;
            }
        }
        return QVariant();
    }

signals:
    void droppedDisplayTables();

private:
    QWidget *view_;
};

class ListView : public QListView
{
    Q_OBJECT
    
public:
    ListView(QWidget *parent) : QListView(parent) {}

signals:
    void newCurrent(const QModelIndex &);
    
protected slots:
    void currentChanged(const QModelIndex &current, 
                        const QModelIndex &previous)
    {
        QListView::currentChanged(current, previous);
        emit newCurrent(current);
    }
    
};

class StructureEditorWindow : public ProjectWindow
{
    Q_OBJECT
    
public:
    StructureEditorWindow(Project *pr);

    void clear()
    {
        currentTable_ = 0;
        displayObjectsModel_->setList(0);
        fieldsModel_->setList(0);
        tablesModel_->setList(0);
    }

    void setTables()
    {
        Table *active = DisplayProjectActiveTable(pr_);
        tablesModel_->setList(ProjectTables(pr_));
        currentTable_ = active;
        tablesView_->setCurrentIndex(tablesModel_->tableIndex(active));
        setFields();
    }
    
    void setFields()
    {
        List *fields = TableFields(currentTable_);
        Field *active = DisplayTableActiveField(currentTable_);
        fieldsBox_->setEnabled(fields != 0);
        fieldsModel_->setList(fields);
        fieldsView_->setCurrentIndex(fieldsModel_->fieldIndex(active));
        fieldsActiveChange();
    }
    
    void setDisplay(bool rememberCursor)
    {
        struct Table *tab = currentTable_;
        struct DisplayObject *dio = TableDisplayObject(tab);
        DisplayObject *activedio = 0;
        QList<int> cursor;
        
        if (rememberCursor) {
            QModelIndex index = displayObjectsView_->currentIndex();
            while (index.isValid()) {
                cursor.push_back(index.row());
                index = index.parent();
            }
        }
        if (rootDisplay_) {
            displayObjectsBox_->setEnabled(true);
            if (!rememberCursor) {
                activedio = DisplayProjectActiveDisplayObject(pr_);
            }
            dio = ProjectRootDisplayObject(pr_);
            displayObjectsModel_->setList(DisplayObjectObjects(dio));
            displayObjectsView_->setRootIsDecorated(true);
        } else if (!tab || !DisplayObjectVisible(dio)) {
            displayObjectsBox_->setEnabled(false);
            dio = 0;
            displayObjectsModel_->setList(0);
            displayObjectsView_->setRootIsDecorated(false);
        } else {
            displayObjectsBox_->setEnabled(true);
            if (!rememberCursor) {
                activedio = DisplayTableActiveDisplayObject(tab);
            }
            displayObjectsModel_->setDisplayObject(dio);
            displayObjectsView_->setRootIsDecorated(false);
        }
        displayObjectsView_->loadOpenState();

        QModelIndex index;
        if (rememberCursor) {
            while (!cursor.isEmpty()) {
                QModelIndex parent = index;
                int row = cursor.takeLast();
                while (row >= 0) {
                    index = displayObjectsModel_->index(row, 0, parent);
                    if (index.isValid()) {
                        break;
                    }
                    --row;
                }
                if (!index.isValid()) {
                    index = parent;
                    break;
                }
            }
        } else {
            index = displayObjectsModel_->displayObjectIndex(activedio);
        }
        displayObjectsView_->setCurrentIndex(index);
        displayObjectsView_->saveOpenState();
    }

    Project *project() { return pr_; }

    void *nameChangeObject() const { return nameChangeObject_; }
    void setNameChangeObject(void *object) { nameChangeObject_ = object; }

private:
    Project *pr_;
    void *nameChangeObject_;
    Table *currentTable_;
    bool rootDisplay_;
    short tableSortMethod_;
    short fieldSortMethod_;
    ListView *tablesView_;
    CustomTablesModel *tablesModel_;
    QPushButton *newTable_;
    QPushButton *deleteTable_;
    QPushButton *sortTables_;
    GroupBox *fieldsBox_;
    ListView *fieldsView_;
    CustomFieldsModel *fieldsModel_;
    QPushButton *newField_;
    QPushButton *copyField_;
    QPushButton *deleteField_;
    QPushButton *sortFields_;
    ComboBox *displayType_;
    QWidget *displayObjectsBox_;
    QPushButton *add_;
    QPushButton *rem_;
    QPushButton *up_;
    QPushButton *down_;
    QPushButton *in_;
    QPushButton *out_;
    QPushButton *text_;
    QPushButton *image_;
    QPushButton *space_;
    QPushButton *balance_;
    QPushButton *group_;
    QPushButton *register_;
    QPushButton *window_;
    DisplayObjectsView *displayObjectsView_;
    DisplayObjectsModel *displayObjectsModel_;
    
    virtual void closeEvent(QCloseEvent *event)
    {
        event->ignore();
        StructureEditor(pr_);
    }

    Field *currentField() const {
        const QModelIndex &index = fieldsView_->currentIndex();
        return fieldsModel_->field(index);
    }
    
    DisplayObject *rootDisplayObject() const
    {
        if (rootDisplay_) {
            return ProjectRootDisplayObject(pr_);
        } else {
            return TableDisplayObject(currentTable_);
        }
    }
    
    DisplayObject *currentDisplayObject() const
    {
        const QModelIndex &index = displayObjectsView_->currentIndex();
        return displayObjectsModel_->displayObject(index);
    }
    
    void addDisplayObject(DisplayObject *dio)
    {
        if (rootDisplay_) {
            SetDisplayProjectActiveDisplayObject(pr_, dio);
        } else {
            SetDisplayTableActiveDisplayObject(currentTable_, dio);
        }
        setDisplay(false);
    }

    void displayAction(bool (*func)(DisplayObject *), bool reverse, bool rememberCursor)
    {
        // We apply func on all selected items.
        QModelIndexList selected = displayObjectsView_->selectionModel()->selectedIndexes();
        bool change = false;
        long error = OK;
        
        if (reverse) {
            std::reverse(selected.begin(), selected.end());
        }
        for (int i=0; i < selected.size(); ++i) {
            const QModelIndex &index = selected[i];
            DisplayObject *dio = displayObjectsModel_->displayObject(index);
            if (dio) {
                if (DisplayObjectType(dio) == DO_TYPE_TABLE && !rootDisplay_) {
                    continue;
                }
                error = ObtainProjectChangeLock(pr_);
                if (error == OK && func(dio)) {
                    change = true;
                }
            }
        }
        
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
        } else if (change) {
            displayObjectsActiveChange();  // saves active object (number) again
            setDisplay(rememberCursor);
            ProjectChange(pr_, true);
        }
    }
    
    void clickFuncSimple(long (*newFunc)(Project *, Table *, DisplayObject *, DisplayObject **),
                         void (*editFunc)(QWidget *, DisplayObject *))
    {
        long error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }
        
        // Get active display item.
        DisplayObject *dioparent = currentDisplayObject();
        if (!dioparent) {
            dioparent = rootDisplayObject();
        }
        
        // Install new simple item.
        DisplayObject *dio = 0;
        error = newFunc(pr_, rootDisplay_? 0 : currentTable_, dioparent, &dio);
        
        if (error) {
            Ask2(this, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));
        } else {
            addDisplayObject(dio);
            ProjectChange(pr_, true);
            if (editFunc) {
                editFunc(this, dio);
            }
        }
    }
    
    void clickFuncSelected(long (*newFunc)(Project *, Table *, DisplayObject **, DisplayObject *,
                                           DisplayObject **),
                           void (editFunc)(QWidget *, DisplayObject *))
    {
        long error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }
        
        // Get current selection of display list.
        QVector<DisplayObject *> selected = selectedDisplayObjects();
        
        // Get active display item.
        DisplayObject *dioparent = currentDisplayObject();
        if (!dioparent) {
            dioparent = rootDisplayObject();
        }
        
        // Install new object.
        DisplayObject *dio = 0;
        error = newFunc(pr_, rootDisplay_? 0 : currentTable_, selected.data(), dioparent, &dio);
        if (error) {
            Ask2(this, GCS(MSG_ErrorMessage), AT_ERROR, GCS(MSG_StarOk), GES(error));
        } else {
            addDisplayObject(dio);
            ProjectChange(pr_, true);
            editFunc(this, dio);
        }
    }
    
    QVector<DisplayObject *> selectedDisplayObjects() const
    {
        QModelIndexList selected = displayObjectsView_->selectionModel()->selectedIndexes();
        QVector<DisplayObject *> dios(selected.size() + 1);
        for (int i=0; i < selected.size(); ++i) {
            dios[i] = displayObjectsModel_->displayObject(selected[i]);
        }
        dios[selected.size()] = 0;
        return dios;
    }
    
private slots:
    void tablesActiveChange()
    {
        Table *tab = tablesModel_->table(tablesView_->currentIndex());        
        if (!FindNodePtr(ProjectTables(pr_), tab)) {
            tab = 0;
        }
        deleteTable_->setEnabled(tab != 0);
        currentTable_ = tab;
        SetDisplayProjectActiveTable(pr_, tab);
        setFields();
        if (!rootDisplay_) {
            setDisplay(false);
        }
    }
    
    void newTableClick()
    {
        DisplayObject *dioparent = 0;
        if (rootDisplay_) {
            dioparent = currentDisplayObject();
        }
        GuiNewTable(this, pr_, dioparent);
    }

    void changeTableClick()
    {
        Table *tab = currentTable_;
        if (tab) {
            GuiChangeTable(this, tab);
        }
    }

    void deleteTableClick()
    {
        Table *tab = currentTable_;
        if (tab) {
            bool deleted = GuiDeleteTable(tab);
            if (deleted) {
                setTables();
                setDisplay(false);
            }
        }
    }
    
    void sortTablesClick()
    {
        List *tabs = ProjectTables(pr_);
        Table *tab = currentTable_;
        if (tableSortMethod_ == 0) {
            SortTablesAlpha(tabs);
            tableSortMethod_ = 1;
            sortTables_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Sort2)));
        } else {
            SortTablesDisplay(tabs);
            tableSortMethod_ = 0;
            sortTables_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Sort1)));
        }
        setTables();
        tablesView_->setCurrentIndex(tablesModel_->tableIndex(tab));
    }

    void fieldsActiveChange()
    {
        Field *field = currentField();
        copyField_->setEnabled(field != 0);
        deleteField_->setEnabled(field != 0);
        SetDisplayTableActiveField(currentTable_, field);
    }
    
    void newFieldClick()
    {
        DisplayObject *dioparent = 0;
        if (!rootDisplay_) {
            dioparent = currentDisplayObject();
        }
        GuiNewField(this, currentTable_, 0, dioparent);
    }
    
    void copyFieldClick()
    {
        Field *field = currentField();
        DisplayObject *dioparent = 0;
        if (!rootDisplay_) {
            dioparent = currentDisplayObject();
        }        
        GuiNewField(this, currentTable_, field, dioparent);
    }
    
    void changeFieldClick()
    {
        Field *field = currentField();
        if (field) {
            GuiChangeField(this, field);
        }
    }
    
    void deleteFieldClick()
    {
        Field *field = currentField();
        if (field) {
            bool deleted = GuiDeleteField(field);
            if (deleted) {
                setFields();
                setDisplay(false);
            }
        }
    }
    
    void sortFieldsClick()
    {
        List *fields = TableFields(currentTable_);
        if (fields) {
            Field *field = currentField();        
            if (fieldSortMethod_ == 0) {
                SortFieldsAlpha(fields);
                fieldSortMethod_ = 1;
                sortFields_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Sort2)));
            } else {
                SortFieldsDisplay(fields);
                fieldSortMethod_ = 0;
                sortFields_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Sort1)));
            }
            setFields();
            fieldsView_->setCurrentIndex(fieldsModel_->fieldIndex(field));
        }
    }
    
    void displayTypeChange(int active)
    {
        rootDisplay_ = (active == 1);
        add_->setToolTip(GCS(rootDisplay_? 
                             MSG_StructureEditor_AddBH1 : MSG_StructureEditor_AddBH2));
        setDisplay(false);
    }
    
    void displayObjectsActiveChange()
    {
        DisplayObject *dio = currentDisplayObject();
        if (rootDisplay_) {
            SetDisplayProjectActiveDisplayObject(pr_, dio);
        } else {
            SetDisplayTableActiveDisplayObject(currentTable_, dio);
        }
    }
    
    void displayObjectChange()
    {
        DisplayObject *dio = currentDisplayObject();
        if (dio) {
            switch (DisplayObjectType(dio)) {
                case DO_TYPE_TEXT:
                    GuiText(this, dio);
                    break;
                    
                case DO_TYPE_SPACE:
                    GuiSpace(this, dio);
                    break;
                    
                case DO_TYPE_IMAGE:
                    GuiImage(this, dio);
                    break;
                    
                case DO_TYPE_GROUP:
                    GuiGroup(this, dio);
                    break;
                    
                case DO_TYPE_REGISTER:
                    GuiRegister(this, dio);
                    break;
                    
                case DO_TYPE_WINDOW:
                    GuiWindow(this, dio);
                    break;
                    
                case DO_TYPE_TABLE:
                    GuiDisplayTable(this, dio);
                    break;
                    
                case DO_TYPE_FIELD:
                    GuiDisplayField(this, dio);
                    break;
            }
        }
    }
    
    void onDroppedDisplayObject(DisplayObject *dio, DisplayObject *parent,
            bool isPred) {
        long error = ObtainProjectChangeLock(pr_);
        if (error != OK) {
            Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
            return;
        }
        DisplayObjectAdd(dio, parent, isPred);
        addDisplayObject(dio);
        ProjectChange(pr_, true);
    }

    void addClick()
    {
        DisplayObject *add;
        
        // Get active entry from tables/fields list.
        if (rootDisplay_) {
            add = TableDisplayObject(currentTable_);
        } else {
            add = FieldDisplayObject(currentField());
        }
        
        DisplayObject *dioparent = currentDisplayObject();
        if (add) {
            long error = ObtainProjectChangeLock(pr_);
            if (error != OK) {
                Ask(AT_ERROR, GCS(MSG_StarOk), GES(error));
                return;
            }
            DisplayObjectAdd(add, dioparent, false);
            addDisplayObject(add);
            ProjectChange(pr_, true);
        }
    }
    
    void remClick()
    {
        displayAction(DisplayObjectRemove, false, true);
    }
    
    void upClick()
    {
        displayAction(DisplayObjectUp, false, false);
    }
    
    void downClick()
    {
        displayAction(DisplayObjectDown, true, false);
    }
    
    void inClick()
    {
        displayAction(DisplayObjectIn, true, false);
    }
    
    void outClick()
    {
        displayAction(DisplayObjectOut, false, false);
    }
    
    void textClick()
    {
        clickFuncSimple(DisplayNewText, GuiText);
    }
    
    void imageClick()
    {
        clickFuncSimple(DisplayNewImage, GuiImage);
    }
    
    void spaceClick()
    {
        clickFuncSimple(DisplayNewSpace, GuiSpace);
    }
    
    void balanceClick()
    {
        clickFuncSimple(DisplayNewBalance, 0);
    }
    
    void groupClick()
    {
        clickFuncSelected(DisplayNewGroup, GuiGroup);
    }
    
    void registerClick()
    {
        clickFuncSelected(DisplayNewRegister, GuiRegister);
    }
    
    void windowClick()
    {
        clickFuncSelected(DisplayNewWindow, GuiWindow);
    }
};

StructureEditorWindow::StructureEditorWindow(Project *pr) : ProjectWindow(pr)
{
    pr_ = pr;
    nameChangeObject_ = 0;
    currentTable_ = 0;
    rootDisplay_ = false;
    tableSortMethod_ = 0;
    fieldSortMethod_ = 0;
    
    SetRootWindowTitle(pr, this);
    GuiSetHelpNode(this, "Structure editor");

#ifndef MACOSX
    GuiWindowAddMenus(this, true);
    QSize size = menuBar()->sizeHint();
    setMinimumWidth(size.width());
#endif
    
    QWidget *root = new QWidget(this);
    setCentralWidget(root);
    
    GroupLayout *top = new GroupLayout(1, root);
    
    QSplitter *tabFieldDisplaySplitter = new QSplitter(root);
    tabFieldDisplaySplitter->setOrientation(Qt::Horizontal);
    tabFieldDisplaySplitter->setChildrenCollapsible(false);
    top->addWidget(tabFieldDisplaySplitter);
    
    QSplitter *tabFieldSplitter = new QSplitter(root);
    tabFieldSplitter->setOrientation(Qt::Vertical);
    tabFieldSplitter->setChildrenCollapsible(false);
    tabFieldDisplaySplitter->addWidget(tabFieldSplitter);
    tabFieldDisplaySplitter->setStretchFactor(0, 2);
    
    GroupBox *tablesBox = new GroupBox(GCS(MSG_StructureEditor_Tables), root);
    GuiSetHelpNode(tablesBox, "Table management");
    tabFieldSplitter->addWidget(tablesBox);
    tabFieldSplitter->setStretchFactor(0, 2);

    GroupLayout *tables = new GroupLayout(1, tablesBox);
    tables->maybeClearContentsMargins();
        
    tablesView_ = new ListView(tablesBox);
    tablesModel_ = new CustomTablesModel(tablesView_);
    tablesView_->setModel(tablesModel_);
    tablesView_->setSelectionMode(QAbstractItemView::SingleSelection);
    tablesView_->setDefaultDropAction(Qt::MoveAction);
    tablesView_->setDragDropMode(QAbstractItemView::DragDrop);
    tablesView_->setToolTip(GCS(MSG_StructureEditor_TablesBH));
    tables->addWidget(tablesView_);
    connect(tablesView_, SIGNAL(newCurrent(const QModelIndex &)),
            this, SLOT(tablesActiveChange()));
    connect(tablesView_, SIGNAL(activated(const QModelIndex &)),
            this, SLOT(changeTableClick()));

    QHBoxLayout *tableButtons = new QHBoxLayout;
    tables->addLayout(tableButtons);

    newTable_ = new QPushButton(tablesBox);
    newTable_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Table_New)));
    newTable_->setToolTip(GCS(MSG_StructureEditor_Table_NewBH));
    tableButtons->addWidget(newTable_);
    connect(newTable_, SIGNAL(clicked()), this, SLOT(newTableClick()));

    deleteTable_ = new QPushButton(tablesBox);
    deleteTable_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Table_Delete)));
    deleteTable_->setToolTip(GCS(MSG_StructureEditor_Table_DeleteBH));
    deleteTable_->setEnabled(false);
    tableButtons->addWidget(deleteTable_);
    connect(deleteTable_, SIGNAL(clicked()), this, SLOT(deleteTableClick()));
        
    sortTables_ = new QPushButton(tablesBox);
    sortTables_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Sort1)));
    sortTables_->setToolTip(GCS(MSG_StructureEditor_Table_SortBH));
    tableButtons->addWidget(sortTables_);
    connect(sortTables_, SIGNAL(clicked()), this, SLOT(sortTablesClick()));
    
    fieldsBox_ = new GroupBox(GCS(MSG_Fields), root);
    GuiSetHelpNode(fieldsBox_, "Field management");
    tabFieldSplitter->addWidget(fieldsBox_);
    tabFieldSplitter->setStretchFactor(1, 4);
    
    GroupLayout *fields = new GroupLayout(1, fieldsBox_);
    fields->maybeClearContentsMargins();
    
    fieldsView_ = new ListView(fieldsBox_);
    fieldsModel_ = new CustomFieldsModel(fieldsView_);
    fieldsView_->setModel(fieldsModel_);
    fieldsView_->setSelectionMode(QAbstractItemView::SingleSelection);
    fieldsView_->setDefaultDropAction(Qt::MoveAction);
    fieldsView_->setDragDropMode(QAbstractItemView::DragDrop);
    fieldsView_->setToolTip(GCS(MSG_StructureEditor_FieldsBH));
    fields->addWidget(fieldsView_);
    connect(fieldsView_, SIGNAL(newCurrent(const QModelIndex &)),
            this, SLOT(fieldsActiveChange()));
    connect(fieldsView_, SIGNAL(activated(const QModelIndex &)),
            this, SLOT(changeFieldClick()));
                                        
    QHBoxLayout *fieldButtons = new QHBoxLayout;
    fields->addLayout(fieldButtons);

    newField_ = new QPushButton(fieldsBox_);
    newField_->setText(GuiMarkShortcut(GCS(MSG_New)));
    newField_->setToolTip(GCS(MSG_StructureEditor_Field_NewBH));
    fieldButtons->addWidget(newField_);
    connect(newField_, SIGNAL(clicked()), this, SLOT(newFieldClick()));
    
    copyField_ = new QPushButton(fieldsBox_);
    copyField_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Field_Copy)));
    copyField_->setEnabled(false);
    copyField_->setToolTip(GCS(MSG_StructureEditor_Field_CopyBH));
    fieldButtons->addWidget(copyField_);
    connect(copyField_, SIGNAL(clicked()), this, SLOT(copyFieldClick()));
    
    deleteField_ = new QPushButton(fieldsBox_);
    deleteField_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Field_Delete)));
    deleteField_->setEnabled(false);
    deleteField_->setToolTip(GCS(MSG_StructureEditor_Field_DeleteBH));
    fieldButtons->addWidget(deleteField_);
    connect(deleteField_, SIGNAL(clicked()), this, SLOT(deleteFieldClick()));
        
    sortFields_ = new QPushButton(fieldsBox_);
    sortFields_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Sort1)));
    sortFields_->setToolTip(GCS(MSG_StructureEditor_Field_SortBH));
    fieldButtons->addWidget(sortFields_);
    connect(sortFields_, SIGNAL(clicked()), this, SLOT(sortFieldsClick()));
    
    GroupBox *displayBox = new GroupBox(GCS(MSG_Display), root);
    GuiSetHelpNode(displayBox, "Display field");
    tabFieldDisplaySplitter->addWidget(displayBox);
    tabFieldDisplaySplitter->setStretchFactor(1, 4);
    
    GroupLayout *display = new GroupLayout(1, displayBox);
    display->maybeClearContentsMargins();

    static char *displayTypeContent[3];
    displayTypeContent[0] = GCS(MSG_StructureEditor_TableDisplay);
    displayTypeContent[1] = GCS(MSG_StructureEditor_MainWindow);
    displayTypeContent[2] = NULL;
    displayType_ = new ComboBox(displayBox, displayTypeContent);
    display->addWidget(displayType_);
    connect(displayType_, SIGNAL(currentIndexChanged(int)), 
            this, SLOT(displayTypeChange(int)));
    
    displayObjectsBox_ = new QWidget(displayBox);
    displayObjectsBox_->setEnabled(false);
    display->addWidget(displayObjectsBox_);
    
    GroupLayout *displayObjects = new GroupLayout(2, displayObjectsBox_);
    displayObjects->setContentsMargins(0, 0, 0, 0);
    
    QVBoxLayout *displayButtons = new QVBoxLayout;
    displayObjects->addLayout(displayButtons);
    
    displayButtons->addStretch(1);
    
    add_ = new QPushButton(displayObjectsBox_);
    add_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Add)));
    add_->setToolTip(GCS(MSG_StructureEditor_AddBH2));
    displayButtons->addWidget(add_);
    connect(add_, SIGNAL(clicked()), this, SLOT(addClick()));
    
    rem_ = new QPushButton(displayObjectsBox_);
    rem_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Rem)));
    rem_->setToolTip(GCS(MSG_StructureEditor_RemBH));
    displayButtons->addWidget(rem_);
    connect(rem_, SIGNAL(clicked()), this, SLOT(remClick()));
    connect(tablesModel_, SIGNAL(droppedDisplayTables()),
            this, SLOT(remClick()));
    connect(fieldsModel_, SIGNAL(droppedDisplayFields()),
            this, SLOT(remClick()));
    
    up_ = new QPushButton(displayObjectsBox_);
    up_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Up)));
    up_->setToolTip(GCS(MSG_StructureEditor_UpBH));
    displayButtons->addWidget(up_);
    connect(up_, SIGNAL(clicked()), this, SLOT(upClick()));
    
    down_ = new QPushButton(displayObjectsBox_);
    down_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Down)));
    down_->setToolTip(GCS(MSG_StructureEditor_DownBH));
    displayButtons->addWidget(down_);
    connect(down_, SIGNAL(clicked()), this, SLOT(downClick()));
    
    in_ = new QPushButton(displayObjectsBox_);
    in_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_In)));
    in_->setToolTip(GCS(MSG_StructureEditor_InBH));
    displayButtons->addWidget(in_);
    connect(in_, SIGNAL(clicked()), this, SLOT(inClick()));
    
    out_ = new QPushButton(displayObjectsBox_);
    out_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Out)));
    out_->setToolTip(GCS(MSG_StructureEditor_OutBH));
    displayButtons->addWidget(out_);
    connect(out_, SIGNAL(clicked()), this, SLOT(outClick()));
 
    displayButtons->addStretch(1);
    
    text_ = new QPushButton(displayObjectsBox_);
    text_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Text)));
    text_->setToolTip(GCS(MSG_StructureEditor_TextBH));
    displayButtons->addWidget(text_);
    connect(text_, SIGNAL(clicked()), this, SLOT(textClick()));
    
    image_ = new QPushButton(displayObjectsBox_);
    image_->setText(GuiMarkShortcut(GCS(MSG_Image)));
    image_->setToolTip(GCS(MSG_StructureEditor_ImageBH));
    displayButtons->addWidget(image_);
    connect(image_, SIGNAL(clicked()), this, SLOT(imageClick()));
    
    space_ = new QPushButton(displayObjectsBox_);
    space_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Space)));
    space_->setToolTip(GCS(MSG_StructureEditor_SpaceBH));
    displayButtons->addWidget(space_);
    connect(space_, SIGNAL(clicked()), this, SLOT(spaceClick()));
    
    balance_ = new QPushButton(displayObjectsBox_);
    balance_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Balance)));
    balance_->setToolTip(GCS(MSG_StructureEditor_BalanceBH));
    displayButtons->addWidget(balance_);
    connect(balance_, SIGNAL(clicked()), this, SLOT(balanceClick()));
    
    group_ = new QPushButton(displayObjectsBox_);
    group_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Group)));
    group_->setToolTip(GCS(MSG_StructureEditor_GroupBH));
    displayButtons->addWidget(group_);
    connect(group_, SIGNAL(clicked()), this, SLOT(groupClick()));
    
    register_ = new QPushButton(displayObjectsBox_);
    register_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Register)));
    register_->setToolTip(GCS(MSG_StructureEditor_RegisterBH));
    displayButtons->addWidget(register_);
    connect(register_, SIGNAL(clicked()), this, SLOT(registerClick()));
    
    window_ = new QPushButton(displayObjectsBox_);
    window_->setText(GuiMarkShortcut(GCS(MSG_StructureEditor_Window)));
    window_->setToolTip(GCS(MSG_StructureEditor_WindowBH));
    displayButtons->addWidget(window_);
    connect(window_, SIGNAL(clicked()), this, SLOT(windowClick()));

    displayButtons->addStretch(1);
    
    displayObjectsView_ = new DisplayObjectsView(displayObjectsBox_);
    displayObjectsModel_ = new DisplayObjectsModel(displayObjectsView_);
    displayObjectsView_->setModel(displayObjectsModel_);
    displayObjectsView_->setDragDropMode(QAbstractItemView::DragDrop);
    displayObjectsView_->setToolTip(GCS(MSG_StructureEditor_DisplayBH));
    displayObjectsView_->setHeaderHidden(true);
    displayObjectsView_->setExpandsOnDoubleClick(false);
    displayObjectsView_->setSelectionMode(QAbstractItemView::ExtendedSelection);
    displayObjects->addWidget(displayObjectsView_);
    connect(displayObjectsView_, SIGNAL(newCurrent(const QModelIndex &)),
            this, SLOT(displayObjectsActiveChange()));
    connect(displayObjectsView_, SIGNAL(activated(const QModelIndex &)),
            this, SLOT(displayObjectChange()));
    connect(displayObjectsView_, SIGNAL(expanded(const QModelIndex &)),
            displayObjectsModel_, SLOT(setOpen(const QModelIndex &)));
    connect(displayObjectsView_, SIGNAL(collapsed(const QModelIndex &)),
            displayObjectsModel_, SLOT(setClosed(const QModelIndex &)));
    connect(displayObjectsModel_, 
            SIGNAL(droppedDisplayObject(DisplayObject *, DisplayObject *, bool)),
            this, 
            SLOT(onDroppedDisplayObject(DisplayObject *, DisplayObject *, bool)));
    
    struct WindowDimensions wd;
    if (GetPrefsWindowDimensions(pr, PREFS_STRUCTUREEDITOR, &wd)) {
       GuiSetWindowDimensions(this, &wd);
    }
}
    
void CloseStructureEditor(GuiObject go)
{
    StructureEditorWindow *se = dynamic_cast<StructureEditorWindow *>(go);
    if (se) {
        WindowDimensions wd;
        if (GuiGetWindowDimensions(go, &wd)) {
            SetPrefsWindowDimensions(se->project(), PREFS_STRUCTUREEDITOR, &wd);
        }
        se->deleteLater();
    }
}

void NewTableFinish(QWidget *parent, Table *tab)
{
    StructureEditorWindow *sew = dynamic_cast<StructureEditorWindow *>(parent);
    if (sew && tab) {
        // Update table and display list.
        SetDisplayProjectActiveTable(sew->project(), tab);
        sew->setTables();
        sew->setDisplay(false);
    }
}

void ChangeTableFinish(QWidget *parent, Table *tab)
{
    StructureEditorWindow *sew = dynamic_cast<StructureEditorWindow *>(parent);
    if (sew && tab) {
        sew->setTables();
        sew->setDisplay(false);
        if (sew->nameChangeObject()) {
            StructureEditorNextNameChange(sew->project(),
                    sew->nameChangeObject());
        }
    }
}

void NewFieldFinish(QWidget *parent, Field *field)
{
    StructureEditorWindow *sew = dynamic_cast<StructureEditorWindow *>(parent);    
    if (sew && field) {
        // Update the fields list and the display list.
        Table *tab = FieldTable(field);
        SetDisplayTableActiveField(tab, field);
        SetDisplayTableActiveDisplayObject(tab, FieldDisplayObject(field));
        sew->setFields();
        sew->setDisplay(false);
    }
}

void ChangeFieldFinish(QWidget *parent, Field *field)
{
    StructureEditorWindow *sew = dynamic_cast<StructureEditorWindow *>(parent);    
    if (sew && field) {
    	// Update the fields list and the display list.
        sew->setFields();
        sew->setDisplay(false);
        if (sew->nameChangeObject()) {
            StructureEditorNextNameChange(sew->project(),
                    sew->nameChangeObject());
        }
    }
}

static void defaultFinish(QWidget *parent, DisplayObject *dio)
{
    StructureEditorWindow *sew = dynamic_cast<StructureEditorWindow *>(parent);    
    if (sew) {
	sew->setDisplay(false);
    }
}

void TextFinish(QWidget *parent, DisplayObject *dio) 
{ defaultFinish(parent, dio); }
void SpaceFinish(QWidget *parent, DisplayObject *dio) 
{ defaultFinish(parent, dio); }
void ImageFinish(QWidget *parent, DisplayObject *dio) 
{ defaultFinish(parent, dio); }
void GroupFinish(QWidget *parent, DisplayObject *dio) 
{ defaultFinish(parent, dio); }
void RegisterFinish(QWidget *parent, DisplayObject *dio) 
{ defaultFinish(parent, dio); }
void WindowFinish(QWidget *parent, DisplayObject *dio) 
{ defaultFinish(parent, dio); }
void DisplayTableFinish(QWidget *parent, DisplayObject *dio) 
{ defaultFinish(parent, dio); }
void DisplayFieldFinish(QWidget *parent, DisplayObject *dio) 
{ defaultFinish(parent, dio); }

GuiObject OpenStructureEditor(Project *pr)
{
    StructureEditorWindow *se = new StructureEditorWindow(pr);
    se->show();
    se->setTables();
    return se;
}

void StructureEditorUpdate(GuiObject go)
{
    StructureEditorWindow *se = dynamic_cast<StructureEditorWindow *>(go);
    if (se) {
        se->setTables();
        se->setDisplay(false);
    }
}

void ClearStructureEditor(GuiObject go)
{
    StructureEditorWindow *se = dynamic_cast<StructureEditorWindow *>(go);
    if (se) {
        se->clear();
    }
}

void StructureEditorSetNameChangeObject(GuiObject go, void *object)
{
    StructureEditorWindow *se = dynamic_cast<StructureEditorWindow *>(go);
    if (se) {
        se->setNameChangeObject(object);
    }
}

#endif

#include "qtStructureEditor.moc"
