//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtStructureEditorUtil.cc
//
// Creation date: 18.02.2016
// Last modified: 06.08.2020
//

#include "system.h"
#include <qt/qtStructureEditorUtil.h>
#include "qt/qtUtil.h"
#include <qt/classes/qtQuickSearch.h>
#include "lang/program.h"
#include "preferences.h"

#ifdef ENABLE_STRUCTURE_EDITOR

#include <qt/classes/qtPopSelection.h>
#include <QLabel>
#include <project.h>
#include "display.h"
#include "requesters.h"
#include "util.h"
#include "error.h"
#include "appstrings_cat.h"

#endif

//
// UserFunctions Popup
//

class PopupListWithQuickSearch : public PopupList
{
    Q_OBJECT
    
public:
    PopupListWithQuickSearch(QWidget *parent) : PopupList(parent) {}
    
public slots:
    void onSearch(const QString &text, QuickSearch::Action action) {
        long dir = -1;
        
        switch (action) {
            case QuickSearch::NEXT:  dir++;		// no break
            case QuickSearch::START: dir++;		// no break
            case QuickSearch::PREV:
            {
                bool unselect = false;
                int row = currentRow() + dir;
                if (dir == 0) {
                    row = 0;
                    dir = 1;
                    unselect = true;
                }
                while (row >= 0 && row < count()) {
                    QListWidgetItem *it = item(row);
                    if (it && it->text().contains(text, Qt::CaseInsensitive)) {
                        setCurrentRow(row);
                        return;
                    }
                    row += dir;
                }
                if (unselect) {
                    setCurrentIndex(QModelIndex());
                }
                break;
            }
                
            case QuickSearch::COMMIT:
                emit activated(currentIndex());         // no break
            case QuickSearch::CANCEL:
                quickSearch_->hide();
                quickSearch_->deleteLater();
                break;
        }
    }
    
protected:
    void keyPressEvent(QKeyEvent *event) {
        if (!QuickSearch::isStartEvent(event)) {
            PopupList::keyPressEvent(event);
        } else {
            delete quickSearch_;
            quickSearch_ = new QuickSearch(this);
            connect(quickSearch_, 
                    SIGNAL(search(const QString &, QuickSearch::Action)), 
                    this, SLOT(onSearch(const QString &, QuickSearch::Action)));
            quickSearch_->show();
            quickSearch_->processEvent(event);
        }        
    }
    
private:
    QPointer<QuickSearch> quickSearch_;
};

class PopUserFunctions : public PopWidget
{
public:
    PopUserFunctions(QLineEdit *string, Project *pr, QWidget *parent)
    {
        pr_ = pr;
        string_ = string;
        PopButton *popButton = new PopButton(parent, string);
        popButton->setArrowType(Qt::DownArrow);
        popButton->setFocusPolicy(Qt::NoFocus);
        list_ = new PopupListWithQuickSearch(parent);
        list_->setPopWidget(this);
        setString(string);
        setButton(popButton);
        setPopupWidget(list_);
    }
    
    virtual bool start()
    {
        list_->clear();
                
        Program *prg = ProjectProgram(pr_);
        List *uflist = ProgramUserFunctions(prg);
        if (uflist) {
            UserFunction *uf;
            FORNODE (uf, uflist) {
                if (UserFunctionTrigger(uf)) {
                    list_->addItem(UserFunctionName(uf));
                }
            }
        }
        
        bool sort = GetPrefsValue(pr_, PREFS_PROGRAMSORTTRIGGERFUNCTIONS);
        if (sort) {
            list_->sortItems();
        }
        
        QList<QListWidgetItem *> match = list_->findItems(string_->text(), 
                                                          Qt::MatchExactly);
        if (!match.isEmpty()) {
            list_->setCurrentItem(match.front());
        }

        return true;
    }
    
    virtual void stop()
    {
        int row = list_->currentRow();
        QListWidgetItem *item = list_->item(row);
        if (item) {
            string_->setText(item->text());
        }
        list_->clear();
    }
    
private:
    Project *pr_;
    QLineEdit *string_;
    PopupList *list_;
};

PopString *UserFunctionsPopup(QLineEdit *string, Project *pr, QWidget *parent)
{
    return new PopUserFunctions(string, pr, parent);
}

#ifdef ENABLE_STRUCTURE_EDITOR

//
// Check for existing shortcuts.
//

void CheckShortCut(QLineEdit *edit, DisplayObject *dio)
{
    QByteArray text = edit->text().toUtf8();
    cschar shortcut = CodesetGetChar(text.data());
    DisplayObject *inuse = DisplayObjectShortcutInUse(shortcut, 
	    DisplayObjectWindow(dio), dio);
    if (inuse) {
        QWidget *win = edit->window();
        const char *name = DisplayObjectName(inuse);
	Ask2(win, 0, AT_ERROR, GCS(MSG_StarOk),
		GCS(MSG_DisplayObject_ShortcutInUse), EmptyNullStr(name));
        edit->setFocus();
    }
}

// Weight slider

QHBoxLayout *WeightSlider(QSlider *slider, QWidget *parent)
{
    slider->setMinimum(0);
    slider->setMaximum(1000);
    slider->setPageStep(10);
    slider->setSingleStep(1);
    slider->setOrientation(Qt::Horizontal);
    slider->setToolTip(GCS(MSG_WeightBH));
    slider->setFocusPolicy(Qt::StrongFocus);
    
    QLabel *value = new QLabel(parent);
    value->setNum(slider->value());
    value->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    QHBoxLayout *hbox = new QHBoxLayout;
    QWidget::connect(slider, SIGNAL(valueChanged(int)), value, SLOT(setNum(int)));
    hbox->addWidget(slider);
    hbox->addWidget(value);
    return hbox;
}

// BackgroundGadgets

QHBoxLayout *BackgroundGadgets(PopImage *popImage, QCheckBox *checkBox, 
                               const QString &bg)
{
    checkBox->setText(GuiMarkShortcut(GCS(MSG_Background_Default)));
    checkBox->setToolTip(GCS(MSG_Background_DefaultBH));
    checkBox->setChecked(bg.isEmpty());
    checkBox->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred);
    QWidget::connect(checkBox, SIGNAL(toggled(bool)), popImage, SLOT(clear(bool)));
    QWidget::connect(popImage, SIGNAL(changed(bool)), checkBox, SLOT(setChecked(bool)));
    popImage->setSpec(bg);
    
    QHBoxLayout *hbox = new QHBoxLayout;
    hbox->addWidget(popImage, 2);
    hbox->addWidget(checkBox, 1, Qt::AlignCenter);
    return hbox;
}

#endif

#include "qtStructureEditorUtil.moc"

