//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtUtil.cc
//
// Creation date: 24.03.2009
// Last modified: 06.02.2022
//

#include "qt/qtUtil.h"
#include "qt/classes/qtSFWView.h"
#include "qt/classes/qtGroupLayout.h"
#include <QAbstractTextDocumentLayout>
#include <QApplication>
#include <QLineEdit>
#include <QPainter>
#include <QRadioButton>
#include <QShortcut>
#include <QTextCursor>
#include <QTextBlock>
#include <QTextDocument>
#include <QTreeView>
#include <QVariant>
#include "lang/program.h"
#include "lang/programbin.h"
#include "main.h"
#include "project.h"
#include "display.h"
#include "strhandle.h"
#include "dosutil.h"
#include "preferences.h"
#include "util.h"

const char MimeTypeTable[] = "BeeBase/Table";
const char MimeTypeTableRecord[] = "BeeBase/TableRecord";
const char MimeTypeField[] = "BeeBase/Field";

static Table *findTable(const QString &projectName, const QString &tableName)
{
    Project *pr;
    FORNODE(pr, &ProjectList) {
        if (projectName == ProjectName(pr)) {
            List *tabs = ProjectTables(pr);
            if (tabs) {
                Table *tab;
                FORNODE(tab, tabs) {
                    if (tableName == TableName(tab)) {
                        return tab;
                    }
                }
            }
        }
    }
    return 0;
}

MimeDataTable::MimeDataTable(Table *tab)
{
    Project *pr = TableProject(tab);
    projectName = ProjectName(pr);
    tableName = TableName(tab);
}

Table *MimeDataTable::table() const
{
    return findTable(projectName, tableName);
}

QDataStream& operator>>(QDataStream &s, MimeDataTable &mdtab)
{
    s >> mdtab.projectName >> mdtab.tableName;
    return s;
}

QDataStream &operator<<(QDataStream &s, const MimeDataTable &mdtab)
{
    s << mdtab.projectName << mdtab.tableName;
    return s;
}

MimeDataTableRecord::MimeDataTableRecord(Table *tab, Record *rec) 
    : MimeDataTable(tab)
{
    recNum = RecordNum(rec);
}

Record *MimeDataTableRecord::record(Table *tab) const
{
    if (tableName == TableName(tab)) {
        return (Record *)FindNodeNum(TableRecords(tab), recNum);
    }
    return 0;
}

QDataStream& operator>>(QDataStream &s, MimeDataTableRecord &mdtr)
{
    MimeDataTable &mdtab = mdtr;
    s >> mdtab >> mdtr.recNum;
    return s;
}

QDataStream &operator<<(QDataStream &s, const MimeDataTableRecord &mdtr)
{
    const MimeDataTable &mdtab = mdtr;
    s << mdtab << mdtr.recNum;
    return s;
}

MimeDataField::MimeDataField(Field *field)
{
    Table *tab = FieldTable(field);
    Project *pr = TableProject(tab);
    projectName = ProjectName(pr);
    tableName = TableName(tab);
    fieldName = FieldName(field);
}

Field *MimeDataField::field() const
{
    Table *tab = findTable(projectName, tableName);
    List *fields = TableFields(tab);
    if (fields) {
        Field *field;
        FORNODE(field, fields) {
            if (fieldName == FieldName(field)) {
                return field;
            }
        }
    }
    return 0;
}

QDataStream& operator>>(QDataStream &s, MimeDataField &mda)
{
    s >> mda.projectName >> mda.tableName >> mda.fieldName;
    return s;
}

QDataStream &operator<<(QDataStream &s, const MimeDataField &mda)
{
    s << mda.projectName << mda.tableName << mda.fieldName;
    return s;
}

void HtmlDelegate::paint(QPainter *painter, const QStyleOptionViewItem &option,
        const QModelIndex &index) const
{
    QStyleOptionViewItem indexOption(option);
    initStyleOption(&indexOption, index);
    
    QStyle *style = indexOption.widget? 
        indexOption.widget->style() : QApplication::style();
    
    QTextDocument doc;
    doc.setHtml(indexOption.text);
    
    /// Painting item without text
    indexOption.text = QString();
    style->drawControl(QStyle::CE_ItemViewItem, &indexOption, painter);
    
    QAbstractTextDocumentLayout::PaintContext ctx;
    
    // Highlighting text if item is selected
    if (indexOption.state & QStyle::State_Selected) {
        ctx.palette.setColor(QPalette::Text, 
                             indexOption.palette.color(QPalette::Active,
                                                       QPalette::HighlightedText));
    }
    
    QRect textRect = style->subElementRect(QStyle::SE_ItemViewItemText,
                                           &indexOption);
    painter->save();
    painter->translate(textRect.topLeft());
    painter->setClipRect(textRect.translated(-textRect.topLeft()));
    doc.documentLayout()->draw(painter, ctx);
    painter->restore();
}

QSize HtmlDelegate::sizeHint(const QStyleOptionViewItem &option, const QModelIndex &index) const
{
    QStyleOptionViewItem indexOption(option);
    initStyleOption(&indexOption, index);
    
    QTextDocument doc;
    doc.setHtml(indexOption.text);
    doc.setTextWidth(indexOption.rect.width());
    return QSize(doc.idealWidth(), doc.size().height());
}

TextEditor::TextEditor(Project *pr, bool fixedFont, QWidget *parent) :
    QTextEdit(parent)
{
    if (fixedFont) {
        QFont font("Courier", fontPointSize());
        font.setStyleHint(QFont::TypeWriter);
        setFont(font);
    }
    setAcceptRichText(false);
#if 0
    g_signal_connect(editor, "load", G_CALLBACK(MenuEditorLoadText), pr);
    g_signal_connect(editor, "save", G_CALLBACK(MenuEditorSaveText), pr);
    g_signal_connect(editor, "external-editor", 
	    G_CALLBACK(MenuEditorExternalEditor), pr);
#endif
}

QString TextEditor::toPlainTextNbsp() const
{
    // QTextDocument::toRawText() was introduced in Qt 5.9.
#if (QT_VERSION >= QT_VERSION_CHECK(5, 9, 0))
    QString txt = document()->toRawText();
#else
    QString txt;
    QTextBlock it;
    for (it = document()->begin(); it != document()->end(); it = it.next()) {
        txt += it.text();
        txt += QChar::ParagraphSeparator;
    }
    txt.chop(1);  // remove trailing paragraph separator
#endif

    QChar *uc = txt.data();
    QChar *e = uc + txt.size();

    for (; uc != e; ++uc) {
        switch (uc->unicode()) {
            case 0xfdd0: // QTextBeginningOfFrame
            case 0xfdd1: // QTextEndOfFrame
            case QChar::ParagraphSeparator:
            case QChar::LineSeparator:
                *uc = QLatin1Char('\n');
                break;
            default:
                ;
        }
    }
    return txt;
}

void TextEditor::getCursor(int *pline, int *pcol) const
{
    const QString text = toPlainTextNbsp();
    const int p = textCursor().position();
    int n = 0, line = 1, col = 1;
    for (;;) {
        col = p - n + 1;
        n = text.indexOf('\n', n);
        if (n < 0 || n >= p) {
            break;
        }
        ++n;
        ++line;
    }
    if (pline) {
        *pline = line;
    }
    if (pcol) {
        *pcol = col;
    }
}

void TextEditor::setCursor(int line, int col)
{
    int n = 0;
    const QString text = toPlainTextNbsp();
    while (line > 1) {
        n = text.indexOf('\n', n);
        if (n < 0) {
            break;
        }
        ++n;
        --line;
    }
    if (n >= 0 && col > 1) {
        n += col - 1;
    }
    if (n < 0 || n > text.size()) {
        n = text.size();
    }
    QTextCursor cursor = textCursor();
    cursor.setPosition(n);
    setTextCursor(cursor);
}

bool GuiIsEditorObject(GuiObject go) 
{ 
    return dynamic_cast<TextEditor *>(go) != 0;
}

StringEdit::StringEdit(QWidget *parent) : QLineEdit(parent) 
{
    connect(this, SIGNAL(returnPressed()), this, SLOT(onReturnPressed()));
}

bool StringEdit::event(QEvent *event)
{
    bool rc = QLineEdit::event(event);
    QKeyEvent *keyEvent = static_cast<QKeyEvent *>(event);        
    if (event->type() == QEvent::KeyPress && keyEvent) {
        if (keyEvent->key() == Qt::Key_Return || keyEvent->key() == Qt::Key_Enter) {
            if (GetPrefsValue(0, PREFS_ADVANCEONENTER)) {
                event->accept();
                rc = true;
            }
        }
    }
    return rc;
}

void StringEdit::onReturnPressed()
{
    if (GetPrefsValue(0, PREFS_ADVANCEONENTER)) {
        QKeyEvent *event = new QKeyEvent (QEvent::KeyPress, Qt::Key_Tab, 
                                          Qt::NoModifier);
        QCoreApplication::postEvent(this, event);
    }
}

LineEditTreeView::LineEditTreeView(QWidget *parent, QTreeView *tree) : QLineEdit(parent)
{
    tree_ = tree;
    QShortcut *up = new QShortcut(Qt::Key_Up, this);
    connect(up, SIGNAL(activated()), this, SLOT(moveUp()));
    QShortcut *down = new QShortcut(Qt::Key_Down, this);
    connect(down, SIGNAL(activated()), this, SLOT(moveDown()));
    connect(tree_->selectionModel(), 
            SIGNAL(currentChanged(const QModelIndex &, const QModelIndex &)),
            this, SLOT(selectionChanged(const QModelIndex &)));
}
                
void LineEditTreeView::moveUp()
{
    QModelIndex index = tree_->currentIndex();
    if (!index.isValid()) {
        index = tree_->model()->index(0, 0);
    } else {
        index = tree_->indexAbove(index);
    }
    if (index.isValid()) {
        tree_->setCurrentIndex(index);
    }
}

void LineEditTreeView::moveDown()
{
    QModelIndex index = tree_->currentIndex();
    if (!index.isValid()) {
        index = tree_->model()->index(0, 0);
    } else {
        index = tree_->indexBelow(index);
    }
    if (index.isValid()) {
        tree_->setCurrentIndex(index);
    }
}
    
void LineEditTreeView::selectionChanged(const QModelIndex &index)
{
    QVariant value = tree_->model()->data(index);
    setText(value.toString());
    setFocus();
    selectAll();
}

FocusButton::FocusButton(QWidget *parent, bool alwaysAcceptFocus) : 
    QPushButton(parent),
    alwaysAcceptFocus_(alwaysAcceptFocus)
{    
    updateFocusPolicy();
}

void FocusButton::updateFocusPolicy()
{
    if (alwaysAcceptFocus_ || GetPrefsValue(NULL, PREFS_EXTRABUTTONSINTABCHAIN)) {
        setFocusPolicy(Qt::TabFocus);
    } else {
        setFocusPolicy(Qt::NoFocus);
    }
}

#define EXTRA_BUTTON_WIDTH 24

ExtraButton::ExtraButton(QWidget *parent, QWidget *buddy, 
                         bool alwaysAcceptFocus) : 
    QToolButton(parent), 
    buddy_(buddy),
    alwaysAcceptFocus_(alwaysAcceptFocus)
{
    setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    updateFocusPolicy();
}

QSize ExtraButton::sizeHint() const
{
    return QSize(EXTRA_BUTTON_WIDTH, buddy_->sizeHint().height());
}

void ExtraButton::updateFocusPolicy()
{
    if (alwaysAcceptFocus_ || GetPrefsValue(NULL, PREFS_EXTRABUTTONSINTABCHAIN)) {
        setFocusPolicy(Qt::TabFocus);
    } else {
        setFocusPolicy(Qt::NoFocus);
    }
}

PopButton::PopButton(QWidget *parent, QWidget *string) :
ExtraButton(parent, string,  !dynamic_cast<QLineEdit *>(string) &&
                   !dynamic_cast<QAbstractButton *>(string))
{
}

PopButtonSetCWD::PopButtonSetCWD(QWidget *parent, QWidget *string, Project *pr) :
PopButton(parent, string)
{
    pr_ = pr;
    connect(this, SIGNAL(clicked()), SLOT(setProjectCWD())); 
}

void PopButtonSetCWD::setProjectCWD()
{
    ProjectSetCurrentDir(pr_);
}

PopSelection *MakeFontPopup(QLineEdit *fontName, QWidget *parent)
{
    PopButton *popButton = new PopButton(parent, fontName);
    popButton->setIcon(popButton->style()->standardIcon(QStyle::SP_FileIcon));
    PopSelection *popFont = new PopSelection();
    popFont->setType(PopSelection::FontSelection);
    popFont->setString(fontName);
    popFont->setButton(popButton);
    return popFont;
}

PopSelection *MakeFilePopup(QLineEdit *filename, Project *pr, QWidget *parent)
{
    PopButtonSetCWD *popButton = new PopButtonSetCWD(parent, filename, pr);
    popButton->setIcon(popButton->style()->standardIcon(QStyle::SP_FileIcon));
    PopSelection *popFile = new PopSelection();
    popFile->setType(PopSelection::FileSelection);
    popFile->setString(filename);
    popFile->setButton(popButton);
    return popFile;
}


PopupList::PopupList(QWidget *parent) : QListWidget(parent), pop_(0) 
{
    connect(this, SIGNAL(itemActivated(QListWidgetItem *)), 
            this, SLOT(onActivated()));
}

void PopupList::onActivated() 
{ 
    if (pop_) {
        pop_->closePopup(true); 
    }
}

ComboBox::ComboBox(QWidget *parent, char **labels) : QComboBox(parent)
{
    setLabels(labels);
}

void ComboBox::setLabels(char **labels)
{
    clear();
    if (labels) {
        for (char **s = labels; *s; s++) {
            addItem(*s);
        }
    }
}

GroupBox::GroupBox(QWidget *parent) : QGroupBox(parent) { setCustomStyle(); }
GroupBox::GroupBox(const QString &title, QWidget *parent) : 
    QGroupBox(title, parent) { setCustomStyle(); }

void GroupBox::setCustomStyle() {
#if 0
    // Add a frame around the group box.
    setStyleSheet("\
        QGroupBox {\
            border: 1px solid gray;\
            border-radius: 9px;\
            margin-top: 0.5em;\
        }\
\
        QGroupBox::title {\
            subcontrol-origin: margin;\
            left: 10px;\
            padding: 0 3px 0 3px;\
        }\
        ");
#endif
}

RadioBox::RadioBox(QWidget *parent, char **labels,
        const QString &title, bool frame, bool horiz) :
    GroupBox(parent), group_(this)
{
    setTitle(title);
    setFlat(!frame);

    GroupLayout *hcenter = new GroupLayout(3, this);
    hcenter->setSizeConstraint(QLayout::SetMinAndMaxSize);
    if (!frame) {
        hcenter->setContentsMargins(0, 0, 0, 0);
    }
    hcenter->addWidget(new QFrame(this));
    GroupLayout *box = new GroupLayout(horiz?  INT_MAX : 1);
    int id = 0;
    for (char **l = labels; *l; l++) {
        QRadioButton *radio = new QRadioButton(*l, this);
        box->addWidget(radio);
        group_.addButton(radio, id++);
    }
    hcenter->addLayout(box);
    hcenter->addWidget(new QFrame(this));
    connect(&group_, SIGNAL(buttonClicked(int)), this, SIGNAL(checked(int)));
}

void RadioBox::setCheckedId(int id)
{
    QAbstractButton *button = group_.button(id);
    if (button) {
        button->setChecked(true);
    }
}
    
CommonDialog::CommonDialog(QWidget *parent) : 
    QDialog(parent), sleepProject_(0)
{
    contentsLayout_ = new QVBoxLayout;
    contentsLayout_->setSizeConstraint(QLayout::SetMinAndMaxSize);
        
    buttonBox_ = new QDialogButtonBox(
            QDialogButtonBox::Ok | QDialogButtonBox::Cancel, 
            Qt::Horizontal, this);
    connect(buttonBox_, SIGNAL(accepted()), this, SLOT(accept()));
    connect(buttonBox_, SIGNAL(rejected()), this, SLOT(reject()));
        
    mainLayout_ = new QVBoxLayout(this);
    mainLayout_->setSizeConstraint(QLayout::SetMinAndMaxSize);
    mainLayout_->addLayout(contentsLayout_);
    mainLayout_->addWidget(buttonBox_);
}

void CommonDialog::setVisible(bool visible)
{
    if (sleepProject_) {
        GuiSetProjectSleep(sleepProject_, visible);
    }
    QDialog::setVisible(visible);
}
    
void CommonDialog::done(int r)
{
    if (r != QDialog::Accepted || acceptEdits()) {
        QDialog::done(r);
    }
}

void GuiSetHelpNode(QObject *o, const char *help)
{
    if (o) {
        o->setProperty("help-node", QVariant(help));
    }
}

long GuiListActive(GuiObject go)
{
    long pos = -1;
    QAbstractItemView *view = dynamic_cast<QAbstractItemView *>(go);
    if (view) {
        const QModelIndex index = view->currentIndex();
        if (index.isValid() && view->selectionModel()->hasSelection()) {
            pos = index.row();
        }
    }
    return pos;
}

#if 0
void *GuiListActiveEntry(GuiObject go)
{
    void *entry = 0;
    QTreeView *tree = dynamic_cast<QTreeView *>(go);
    if (tree) {
        const QModelIndex index = tree->currentIndex();
        if (index.isValid() && tree->selectionModel()->hasSelection()) {
            entry = GuiListNthEntry(go, index.row());
        }
    }
    return entry;
}
#endif

void GuiListSetActive(GuiObject go, long row)
{
    QTreeView *tree = dynamic_cast<QTreeView *>(go);
    if (tree) {
        QAbstractItemModel *model = tree->model();
        if (model) {
            QModelIndex index;
            if (row >= 0) {
                index = model->index(row, 0);
            }
            tree->setCurrentIndex(index);
        }
    }
}

long GuiSFWListActive(GuiObject go)
{
    long pos = 0;
    SFWView *sfw = dynamic_cast<SFWView *>(go);
    if (sfw) {
        pos = sfw->activeLine();
    }
    return pos;
}

void GuiSFWListSetActive(GuiObject go, long num)
{
    SFWView *sfw = dynamic_cast<SFWView *>(go);
    if (sfw) {
        sfw->setActiveLine(num);
    }
}

char GuiSFWListNumColumns(GuiObject go)
{
    char num = 0;
    SFWView *sfw = dynamic_cast<SFWView *>(go);
    if (sfw) {
        num = sfw->numFields();
    }
    return num;
}

const QueryFieldInfo *GuiSFWListFieldInfos(GuiObject go)
{
    const QueryFieldInfo *infos = 0;
    SFWView *sfw = dynamic_cast<SFWView *>(go);
    if (sfw) {
        infos = sfw->fieldInfos();
    }
    return infos;
}

const signed char *GuiSFWListColumnOrder(GuiObject go)
{
    const signed char *order = 0;
    SFWView *sfw = dynamic_cast<SFWView *>(go);
    if (sfw) {
        order = sfw->getColumnOrder();
    }
    return order;
}

const struct QueryLine *GuiSFWListQueryLine(GuiObject go, long row)
{
    const struct QueryLine *ql = 0;
    SFWView *sfw = dynamic_cast<SFWView *>(go);
    if (sfw) {
        ql = sfw->queryLine(row);
    }
    return ql;
}

Table *GuiSFWListControlTable(GuiObject go)
{
    Table *tab = 0;
    SFWView *sfw = dynamic_cast<SFWView *>(go);
    if (sfw) {
        tab = sfw->controlTable();
    }
    return tab;
}

struct Record *GuiSFWListActiveRecord(GuiObject go)
{
    Record *rec = NULL;
    SFWView *sfw = dynamic_cast<SFWView *>(go);
    if (sfw) {
        rec = sfw->activeRecord();
    }
    return rec;
}

void GuiSFWListSetActiveRecord(GuiObject go, struct Record *rec)
{
    SFWView *sfw = dynamic_cast<SFWView *>(go);
    if (sfw) {
        sfw->setActiveRecord(rec);
    }
}

struct Record *GuiSFWListRecord(GuiObject go, long row)
{
    Record *rec = NULL;
    SFWView *sfw = dynamic_cast<SFWView *>(go);
    if (sfw) {
        rec = sfw->record(row);
    }
    return rec;
}

long GuiListFindEntry(GuiObject go, void *entry)
{
    long row = -1;
    QTreeView *tree = dynamic_cast<QTreeView *>(go);
    if (tree) {
        EntryListModel *model = dynamic_cast<EntryListModel *>(tree->model());
        if (model) {
            row = model->findEntry(entry);
        }
    }
    return row;
}

long GuiListFindNode(GuiObject go, void *node)
{
    long row = -1;
    QTreeView *tree = dynamic_cast<QTreeView *>(go);
    if (tree && node) {
        EntryListModel *model = dynamic_cast<EntryListModel *>(tree->model());
        if (model) {
            // quick test
            row = NodeNum(node) - 1;
            if (model->entry(row) != node) {
                row = model->findEntry(node);
            }
        }
    }
    return row;
}

void *GuiListNthEntry(GuiObject go, long pos)
{
    void *entry = 0;
    QTreeView *tree = dynamic_cast<QTreeView *>(go);
    if (tree && pos >= 0) {
        EntryListModel *model = dynamic_cast<EntryListModel *>(tree->model());
        if (model) {
            entry = model->entry(pos);
        }
    }
    return entry;
}

const char *qtStringUtf8(const QString &text, Utf8Buffer &buffer)
{
    buffer.append(text.toUtf8());
    return buffer.back().constData();
}

QWidget *GuiFindParentWindow(QWidget *obj, Project *pr)
{
    QWidget *win = 0;

    if (obj) {
        win = obj->window();
    } else {
        win = QApplication::activeWindow();
    }

    return win;
}

Q_DECLARE_OPAQUE_POINTER(Project *)
Q_DECLARE_METATYPE(Project *)

Project *GuiWindowProject(GuiObject win) 
{ 
    Project *pr = 0;

    if (win) {
	pr = win->property("project").value<Project *>();
    }
    return pr;
}

Project *GuiObjectProject(GuiObject go)
{
    return GuiWindowProject(go->window());
}

void GuiSetWindowProject(GuiObject win, Project *pr)
{
    if (win) {
	win->setProperty("project", QVariant::fromValue<Project *>(pr));
    }
}

GuiObject GuiAnyOpenedWindow(void)
{
    GuiObject any = 0;
    foreach (QWidget *top, QApplication::topLevelWidgets()) {
        if (top->isVisible()) {
            any = top;
            break;
        }
    }
    return any;
}

void GuiUpdateExtraButtonsInTabChain(void)
{
    foreach (QWidget *top, QApplication::topLevelWidgets()) {
        foreach (FocusButton *focusButton, top->findChildren<FocusButton *>()) {
            focusButton->updateFocusPolicy();
        }
        foreach (ExtraButton *extraButton, top->findChildren<ExtraButton *>()) {
            extraButton->updateFocusPolicy();
        }
    }
}

QString GuiMarkShortcut(const QString &str, QChar shortcut)
{
    QString copy = str;
    copy.replace('&', "&&");
    copy.replace('_', '&');
    int index = copy.indexOf(shortcut, 0, Qt::CaseInsensitive);
    QString result;
    
    if (index < 0 || shortcut == QChar(0)) {
        result = copy;
    } else {
        result = copy.left(index);
        result.append('&');
        result.append(copy.mid(index));
    }
    
    return result;
}

QString qtExpandPathname(const QString &path)
{
    struct MemPool *memPool = CreatePool(0, 4096, 1024);
    const char *tmp = ExpandPathname(path.toUtf8(), memPool);
    QString expanded(tmp);
    
    DeletePool(memPool);
    return expanded;
}

QString qtExpandPathnameGetAssign(const QString &path, QString &assignName, QString &assignPath) 
{
    QString fullPath = qtExpandPathname(path);
    int len = path.length();
    int fullLen = fullPath.length();
    while (len > 0 && fullLen > 0) {
        if (path[len - 1] != fullPath[fullLen - 1]) {
            break;
        }
        len--;
        fullLen--;
    }
    
    assignName = path.left(len);
    assignPath = fullPath.left(fullLen);

    return fullPath;
}

QString qtUnexpandPathname(const QString &fullPath,
	       	const QString &assignName, const QString &assignPath) 
{
    int assignLen = assignPath.length();
    QString path;

#ifdef WINDOWS
    // Make sure we have the same file separators.
    QString fullPathCopy(fullPath);
    QString assignPathCopy(assignPath);
    fullPathCopy.replace(QChar('\\'), QChar('/'));
    assignPathCopy.replace(QChar('\\'), QChar('/'));
    bool useAssign = assignLen > 0 && fullPathCopy.startsWith(assignPathCopy);
#else
    bool useAssign = assignLen > 0 && fullPath.startsWith(assignPath);
#endif

    if (!useAssign) {
        path = fullPath;
    } else if (!assignName.isEmpty()) {
        path = assignName;
        path.append(fullPath.mid(assignLen));
    } else {
	/* current dir assign */
	while (IsPathSeparator(fullPath[assignLen].unicode())) {
	    assignLen++;
	}
        path = fullPath.mid(assignLen);
    }
    
    return path;
}

static const struct {
    QFont::Stretch stretch;
    const char *name;
} stretchTable[] = {
    { QFont::UltraCondensed, "Ultra-Condensed" },
    { QFont::ExtraCondensed, "Extra-Condensed" },
    { QFont::Condensed,      "Condensed" },
    { QFont::SemiCondensed,  "Semi-Condensed" },
    { QFont::Unstretched,    0 },
    { QFont::SemiExpanded,   "Semi-Expanded" },
    { QFont::Expanded,       "Expandex" },
    { QFont::ExtraExpanded,  "Extra-Expanded" },
    { QFont::UltraExpanded,  "Ultra-Expanded" }
};
const int numStretch = sizeof(stretchTable) / sizeof(stretchTable[0]);

static const struct {
    QFont::Style style;
    const char *name;
} styleTable[] = {
    { QFont::StyleNormal,  0 },
    { QFont::StyleItalic,  "Italic" },
    { QFont::StyleOblique, "Oblique" }
};
const int numStyle = sizeof(styleTable) / sizeof(styleTable[0]);

// Not sure when these QFont weights got introduced, but they are not in 5.0.1
#if (QT_VERSION >= QT_VERSION_CHECK(5, 1, 0))
#define QFONT_HAS_EXTRA_WEIGHTS
#endif

static const struct {
    QFont::Weight weight;
    const char *name;
} weightTable[] = {
#ifdef QFONT_HAS_EXTRA_WEIGHTS
    { QFont::Thin,       "Thin" },
    { QFont::ExtraLight, "Extra-Light" },
#endif
    { QFont::Light,      "Light" },
    { QFont::Normal,     0 },
#ifdef QFONT_HAS_EXTRA_WEIGHTS
    { QFont::Medium,     "Medium" },
#endif
    { QFont::DemiBold,   "Demi-Bold" },
    { QFont::Bold,       "Bold" },
#ifdef QFONT_HAS_EXTRA_WEIGHTS
    { QFont::ExtraBold,  "Extra-Bold" },
#endif
    { QFont::Black,      "Black" }
};
const int numWeight = sizeof(weightTable) / sizeof(weightTable[0]);

QFont qtFontFromDescription(const QString &description)
{
    QFont font;

    // If the description is a full description using a comma-separated list
    // of values, use Qt's conversion.
    if (description.count(",") >= 5) {
        font.fromString(description);
        return font;
    }

    QStringList list = description.split(QRegExp("\\s+"));
    QString family;
    QString test;
    int familyItems = 0;
    for (int i = 0; i < list.size(); i++) {
        if (!test.isEmpty()) {
            test.append(' ');
        }
        test.append(list[i]);
        font.setFamily(test);
        bool ok = font.exactMatch() || !QFont::substitutes(test).empty();
        if (ok) {
            family = test;
            familyItems = i + 1;
        }
    }
    font.setFamily(family);
    for (int i = familyItems; i < list.size(); i++) {
        const QString &item = list[i];
        // Stretch.
        for (int j = 0; j < numStretch; j++) {
            const char *name = stretchTable[j].name;
            if (name && item.compare(name, Qt::CaseInsensitive) == 0) {
                font.setStretch(stretchTable[j].stretch);
                break;
            }
        }
        // Weight.
        for (int j = 0; j < numWeight; j++) {
            const char *name = weightTable[j].name;
            if (name && item.compare(name, Qt::CaseInsensitive) == 0) {
                font.setWeight(weightTable[j].weight);
                break;
            }
        }
        // Style.
        for (int j = 0; j < numStyle; j++) {
            const char *name = styleTable[j].name;
            if (name && item.compare(name, Qt::CaseInsensitive) == 0) {
                font.setStyle(styleTable[j].style);
                break;
            }
        }
        // Point size.
        bool ok = false;
        qreal size = item.toDouble(&ok);
        if (ok) {
            font.setPointSizeF(size);
        }
    }

    return font;
}

QString qtFontToDescription(const QFont &font)
{
    QString description(font.family());

    // Stretch.
    int stretch = font.stretch();
    if (stretch != 0 /* QFont::AnyStretch */) {
        for (int j = 0; j < numStretch; j++) {
            if (stretch <= stretchTable[j].stretch || j == numStretch - 1) {
                const char *name = stretchTable[j].name;
                if (name) {
                   description.append(' ');
                   description.append(name);
                }
                break;
            }
        }
    }
    // Weight.
    int weight = font.weight();
    for (int j = 0; j < numWeight; j++) {
        if (weight <= weightTable[j].weight || j == numWeight - 1) {
            const char *name = weightTable[j].name;
            if (name) {
                description.append(' ');
                description.append(name);
            }
            break;
        }
    }
    // Style.
    QFont::Style style = font.style();
    for (int j = 0; j < numStyle; j++) {
        if (style <= styleTable[j].style) {
            const char *name = styleTable[j].name;
            if (name) {
                description.append(' ');
                description.append(name);
            }
            break;
        }
    }
    // Point size.
    qreal size = font.pointSizeF();
    if (size > 0.0) {
        description.append(' ');
        description.append(QString::number(size));
    }

    return description;
}
