//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// qtUtil.h
//
// Creation date: 24.03.2009
// Last modified: 04.10.2021
//

#ifndef QT_UTIL_H
#define QT_UTIL_H

#include <system.h>
#include <qt/classes/qtPopSelection.h>
#include "qt/classes/qtPopWidget.h"
#include <QBoxLayout>
#include <QButtonGroup>
#include <QComboBox>
#include <QDialog>
#include <QDialogButtonBox>
#include <QFocusEvent>
#include <QGroupBox>
#include <QLineEdit>
#include <QListWidget>
#include <QPushButton>
#include <QScrollBar>
#include <QStyledItemDelegate>
#include <QTextEdit>
#include <QToolButton>
#include <QTreeView>
#include "table.h"

extern const char MimeTypeTable[];
extern const char MimeTypeTableRecord[];
extern const char MimeTypeField[];

struct MimeDataTable
{
    MimeDataTable() {}
    MimeDataTable(Table *tab);

    Table *table() const;

    QString projectName;
    QString tableName;
};
QDataStream &operator>>(QDataStream &s, MimeDataTable &mdtab);
QDataStream &operator<<(QDataStream &s, const MimeDataTable &mdtab);

struct MimeDataTableRecord : public MimeDataTable
{
    MimeDataTableRecord() {}
    MimeDataTableRecord(Table *tab, Record *rec);

    Record *record(Table *tab) const;

    qint64 recNum;
};
QDataStream &operator>>(QDataStream &s, MimeDataTableRecord &mdtr);
QDataStream &operator<<(QDataStream &s, const MimeDataTableRecord &mdtr);

struct MimeDataField
{
    MimeDataField() {}
    MimeDataField(Field *field);

    Field *field() const;

    QString projectName;
    QString tableName;
    QString fieldName;
};
QDataStream &operator>>(QDataStream &s, MimeDataField &mda);
QDataStream &operator<<(QDataStream &s, const MimeDataField &mda);

class GuiBusyCursor
{
public:
    GuiBusyCursor() { GuiSetSleep(true); }
    ~GuiBusyCursor() { GuiSetSleep(false); }
};

class EntryListModel : public QAbstractListModel
{
public:
    EntryListModel(QObject *parent) : QAbstractListModel(parent) {}
    
    virtual int findEntry(void *entry) const = 0;
    virtual void *entry(int row) const = 0;
};

class HtmlDelegate : public QStyledItemDelegate
{
public:
    HtmlDelegate(QObject *parent) : QStyledItemDelegate(parent) {}
    
protected:
    void paint(QPainter *painter, const QStyleOptionViewItem &option,
               const QModelIndex &index) const;
    QSize sizeHint(const QStyleOptionViewItem & option,
                   const QModelIndex &index) const;
};

class TextEditor : public QTextEdit
{
public:
    TextEditor(Project *pr, bool fixedFont, QWidget *parent);

    int top() const { return verticalScrollBar()->value(); }
    void setTop(int top) { verticalScrollBar()->setValue(top); }

    QString toPlainTextNbsp() const;

    void getCursor(int *line, int *col) const;
    void setCursor(int line, int col);
};

class StringEdit : public QLineEdit
{
    Q_OBJECT
    
public:
    StringEdit(QWidget *parent); 
    
protected:
    bool event(QEvent * e);
    
private slots:
    void onReturnPressed();
};

class LineEditTreeView : public QLineEdit
{
    Q_OBJECT
    
public:
    LineEditTreeView(QWidget *parent, QTreeView *tree);
                
public slots:
    void moveUp();
    void moveDown();
    void selectionChanged(const QModelIndex &index);
    
private:
    QTreeView *tree_;
};

class FocusButton : public QPushButton
{
    Q_OBJECT

public:
    FocusButton(QWidget *parent, bool alwaysAcceptFocus = false);
    
    void updateFocusPolicy();
    
private:
    bool alwaysAcceptFocus_;
};

class ExtraButton : public QToolButton
{
    Q_OBJECT

public:
    ExtraButton(QWidget *parent, QWidget *buddy, bool alwaysAcceptFocus=false);

    virtual QSize sizeHint() const;
    QWidget *buddy() { return buddy_; }

    void updateFocusPolicy();

private:
    QWidget *buddy_;
    bool alwaysAcceptFocus_;
};

class PopButton : public ExtraButton
{
public:
    PopButton(QWidget *parent, QWidget *string);
};

class PopButtonSetCWD : public PopButton
{
    Q_OBJECT
    
public:
    PopButtonSetCWD(QWidget *parent, QWidget *string, Project *pr);
    
public slots:
    void setProjectCWD();
    
private:
    Project *pr_;
};

PopSelection *MakeFontPopup(QLineEdit *fontName, QWidget *parent);
PopSelection *MakeFilePopup(QLineEdit *fontName, Project *pr, QWidget *parent);

class PopupList : public QListWidget
{
    Q_OBJECT
    
public:
    PopupList(QWidget *parent);
    
    void setPopWidget(PopWidget *pop) { pop_ = pop; }
    
public slots:
    void onActivated();
    
private:
    PopWidget *pop_;
};

class ComboBox : public QComboBox
{
    Q_OBJECT

public:
    ComboBox(QWidget *parent, char **labels);

public slots:
    void setLabels(char **labels);
};

class GroupBox : public QGroupBox
{
public:
    GroupBox(QWidget *parent = Q_NULLPTR);
    GroupBox(const QString &title, QWidget *parent = Q_NULLPTR);

private:
    void setCustomStyle();
};

class RadioBox : public GroupBox
{
    Q_OBJECT
    
public:
    RadioBox(QWidget *parent, char **labels, const QString &title = QString(),
            bool frame = false, bool horiz = false);

    int checkedId() const { return group_.checkedId(); }
    void setCheckedId(int id);
    
signals:
    void checked(int id);

private:
    QButtonGroup group_;
};

class CommonDialog : public QDialog
{
public:
    CommonDialog(QWidget *parent);
    
    void setSleepProject(Project *pr) { sleepProject_ = pr; }
    
    void setOkText(const QString &okText)
    {
        buttonBox_->button(QDialogButtonBox::Ok)->setText(okText);
    }
    
    void setCancelText(const QString &cancelText)
    {
        buttonBox_->button(QDialogButtonBox::Cancel)->setText(cancelText);
    }

    void setOkDefault()
    {
        buttonBox_->button(QDialogButtonBox::Ok)->setDefault(true);
    }
    
    QVBoxLayout *contentsLayout() { return contentsLayout_; }

    QDialogButtonBox *buttonBox() { return buttonBox_; }
    
    virtual void setVisible(bool visible);
    
    virtual void done(int r);
    
protected:
    virtual bool acceptEdits() { return true; }
    
private:
    QVBoxLayout *mainLayout_;
    QVBoxLayout *contentsLayout_;
    QDialogButtonBox *buttonBox_;
    Project *sleepProject_;
};

void GuiSetHelpNode(QObject *o, const char *help);

long GuiListFindEntry(GuiObject go, void *entry);

QWidget *GuiFindParentWindow(QWidget *obj = 0, Project *pr = 0);

void GuiSetWindowProject(GuiObject win, struct Project *pr);

GuiObject GuiAnyOpenedWindow(void);

// Appends text.toUtf8() to the buffer and returns UTF8 string.
typedef QList<QByteArray> Utf8Buffer;

const char *qtStringUtf8(const QString &text, Utf8Buffer &buffer);

QString GuiMarkShortcut(const QString &str, QChar shortcut = 0);

QString qtExpandPathname(const QString &path);

QString qtExpandPathnameGetAssign(const QString &path, QString &assignName, QString &assignPath);
QString qtUnexpandPathname(const QString &fullPath, const QString &assignName, const QString &assignPath);

// Font to description conversions similar to how GTK does it.
QFont qtFontFromDescription(const QString &description);
QString qtFontToDescription(const QFont &font);

#endif

