/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** qtkcodeset.cc
**
** Codeset stands for the default encoding of a character string.
** The default encoding used for Qt is UTF8.
**
** Creation date: 26.05.2010
** Last modified: 09.09.2018
*/

#include <codeset.h>
// #include <strhandle.h>
#include <memory.h>
#include <util.h>
#include <error.h>
#include <QString>
#include <QTextCodec>

#define MIBENUM_LATIN1 4
#define MIBENUM_UTF8 106

#define UNICODE_REPLACE_CHAR 0xFFFD
#define UTF8_REPLACE_STR "\xEF\xBF\xBD"
#define UTF8_REPLACE_LEN 3

#define UNICODE_INVALID 0xFFFFFFFF
#define UNICODE_SHORT_UTF8_SEQUENCE 0xFFFFFFFE

static char utf8SequenceLength[256] = 
{
    0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 1, 1
};

static int utf8_decode(const char *s, cschar *ucs4)
{
    unsigned char c = *s;
    int octets = utf8SequenceLength[c];
    cschar code = c;
    int i = octets;

    if (octets > 1) {
        code &= ~(((signed char)0x80) >> octets);        // left-fill in with ones then invert
        for (i = 1; i < octets; i++) {
            c = s[i];
            if ((c & 0xC0) != 0x80) {
                code = (c? UNICODE_INVALID : UNICODE_SHORT_UTF8_SEQUENCE);
                break;
            }
            code = (code << 6) | c;
        }
    }
    *ucs4 = code;
    return i;
}

static bool ucs4_is_valid_for_octets(cschar ucs4, int octets)
{
    bool rc = true;
    
    switch (octets) 
    {
        case 0:
        case 1:     // Valid range is 0 ... 0x7F
            rc = (ucs4 <= 0x7F);
            break;
            
        case 2:     // Valid range is 0x80 ... 0x7FF
            rc = (ucs4 >= 0x80 && ucs4 <= 0x7FF);
            break;
            
        case 3:     // Valid range is 0x800 ... 0xFFFF
            rc = (ucs4 >= 0x800 && ucs4 <= 0xFFFF);
            // We leave U+D800 through U+DFFF as valid
            // We also leave reversed BOM U+FFFE valid
            // We also leave U+FFFF valid
            break;
            
        case 4:     // Valid range is 0x10000 ... 0x10FFFF
            rc = (ucs4 >= 0x10000 && ucs4 <= 0x10FFFF);
            break;
            
        default:
            rc = false;
            break;
    }
    
    return rc;
}

static int utf8_encode(cschar ucs4, char *s)
{
    int octets;
    
    if (ucs4 < 0x80) {
        s[0] = ucs4;
        octets = 1;
    } else if (ucs4 < 0x800) {
        s[0] = 0xC0 | (ucs4 >> 6);
        s[1] = 0X80 | (ucs4 & 0x3F);
        octets = 2;
    } else if (ucs4 < 0x10000) {
        s[0] = 0xE0 | (ucs4 >> 12);
        s[1] = 0x80 | ((ucs4 >> 6) & 0x3F);
        s[2] = 0x80 | (ucs4 & 0x3F);
        octets = 3;
    } else if (ucs4 < 0x200000) {
        s[0] = 0xF0 | (ucs4 >> 18);
        s[1] = 0x80 | ((ucs4 >> 12) & 0x3F);
        s[2] = 0x80 | ((ucs4 >> 6) & 0x3F);
        s[3] = 0x80 | (ucs4 & 0x3F);
        octets = 4;
    } else if (ucs4 < 0x4000000) {
        s[0] = 0xF8 | (ucs4 >> 24);
        s[1] = 0x80 | ((ucs4 >> 18) & 0x3F);
        s[2] = 0x80 | ((ucs4 >> 12) & 0x3F);
        s[3] = 0x80 | ((ucs4 >> 6) & 0x3F);
        s[4] = 0x80 | (ucs4 & 0x3F);
        octets = 5;
    } else if (ucs4 < 0x80000000) {
        s[0] = 0xFC | (ucs4 >> 30);
        s[1] = 0x80 | ((ucs4 >> 24) & 0x3F);
        s[2] = 0x80 | ((ucs4 >> 18) & 0x3F);
        s[3] = 0x80 | ((ucs4 >> 12) & 0x3F);
        s[4] = 0x80 | ((ucs4 >> 6) & 0x3F);
        s[5] = 0x80 | (ucs4 & 0x3F);
        octets = 6;
    } else {
        // illegal code point
        octets = UTF8_REPLACE_LEN;
        memcpy(s, UTF8_REPLACE_STR, octets);
    }
    return octets;
}

static char *utf8_find_prev_char(const char *s, const char *p)
{
    while (--p >= s) {
         if ((*p & 0xC0) != 0x80) {
             return (char *)p;
         }
    }
    return 0;
}

static char *utf8_strchr(const char *s, size_t n, cschar c)
{
    cschar ucs4;
    int octets;
    
    while (n) {
        octets = utf8_decode(s, &ucs4);
        if ((size_t)octets > n) {
            break;
        }
        if (ucs4 == 0) {
            break;
        }
        if (ucs4 == c) {
            return (char *)s;
        }
        s += octets;
        n -= octets;
    }
    
    return 0;
}
        
QTextCodec *localeCodec = 0;
QTextCodec *latin1Codec = 0;

const char *CodesetInit(void) 
{ 
    localeCodec = QTextCodec::codecForLocale();
    latin1Codec = QTextCodec::codecForMib(MIBENUM_LATIN1);
    QTextCodec *utf8Codec = QTextCodec::codecForMib(MIBENUM_UTF8);
    
    if (!localeCodec) {
        return "Locale codec not found!";
    }
    if (!latin1Codec) {
        return "Latin1 codec not found!";
    }
    if (!utf8Codec) {
        return "UTF-8 codec not found!";
    }
    // QTextCodec::setCodecForCStrings(utf8Codec);
    
    return NULL;
}

void CodesetExit(void)
{
    latin1Codec = 0;
    localeCodec = 0;
}

static bool CodecIsUTF8(QTextCodec *codec)
{
    return (!codec || codec->mibEnum() == MIBENUM_UTF8);
}

bool LocaleIsUTF8(void)
{ 
    return CodecIsUTF8(localeCodec);
}

bool StrIsCodeset(const char *s)
{
    cschar ucs4;
    int octets;
    
    if (s) {
        do {
            octets = utf8_decode(s, &ucs4);
            if (!ucs4_is_valid_for_octets(ucs4, octets)) {
                return false;
            }
            s += octets;
        } while (ucs4);
    }
    
    return true;
}         

bool StrIsUTF8AllowTrail(const char *s, const char **pend)
{
    const char *end = NULL;
    bool rc = true;
    cschar ucs4;
    int octets;
    
    if (s) {
        do {
            end = s;
            octets = utf8_decode(s, &ucs4);
            if (!ucs4_is_valid_for_octets(ucs4, octets)) {
                rc = (ucs4 == UNICODE_SHORT_UTF8_SEQUENCE);
                break;
            }
            s += octets;
        } while (ucs4);
    }
    
    if (pend) {
        *pend = end;
    }
    return rc;
}

static char *CodesetValidateUTF8(const char *in, MemPool *memPool, bool allowTrail,
                                 const char **pend, long *perr, size_t *pwarn)
{
    char *str = NULL;
    const char *end = NULL;
    long err = OK;
    size_t warn = 0;
    
    if (in == NULL) {
        err = ERR_INTERNAL;
    } else if (StrIsCodeset(in)) {
        str = (char *)in;
        end = in + strlen(in);
    } else {
        size_t bytes = 0;
        const char *s = in;
        int octets;
        cschar ucs4;

        do {
            end = s;
            octets = utf8_decode(s, &ucs4);
            if (ucs4_is_valid_for_octets(ucs4, octets)) {
                bytes += octets;
            } else if (allowTrail && ucs4 == UNICODE_SHORT_UTF8_SEQUENCE) {
                break;
            } else {
                bytes += UTF8_REPLACE_LEN;
            }
            s += octets;
        } while (ucs4);
        
        str = (char *)PoolAllocMem(memPool, bytes + 1);
        if (str == NULL) {
            err = ERR_MEM;
        } else {
            char *t = str;
            
            s = in;
            do {
                octets = utf8_decode(s, &ucs4);
                if (ucs4_is_valid_for_octets(ucs4, octets)) {
                    memcpy(t, s, octets);
                    t += octets;
                } else if (allowTrail && ucs4 == UNICODE_SHORT_UTF8_SEQUENCE) {
                    break;
                } else {
                    memcpy(t, UTF8_REPLACE_STR, UTF8_REPLACE_LEN);
                    t += UTF8_REPLACE_LEN;
                    warn++;
                }
                s += octets;
            } while(ucs4);
            *t = 0;
        }
    }

    if (pend) {
        *pend = end;
    }
    if (perr) {
        *perr = err;
    }
    if (pwarn) {
        *pwarn = warn;
    }
    return str;
}

char *CodesetConvertFromUTF8(const char *in, MemPool *memPool, long *perr, size_t *pwarn)
{
    return CodesetValidateUTF8(in, memPool, false, NULL, perr, pwarn);
}

char *CodesetConvertFromUTF8AllowTrail(const char *in, MemPool *memPool, long *perr, size_t *pwarn, 
                                       const char **pend) 
{ 
    return CodesetValidateUTF8(in, memPool, true, pend, perr, pwarn); 
}

static char *CodesetConvertFromCodec(const char *in, MemPool *memPool, QTextCodec *codec,
                                     bool allowTrail, const char **pend, long *perr, size_t *pwarn)
{
    char *str = 0;
    const char *end = 0;
    long err = OK;
    size_t warn = 0;
    
    if (CodecIsUTF8(codec)) {
        str = CodesetValidateUTF8(in, memPool, allowTrail, &end, &err, &warn);
    } else if (in == NULL) {
        err = ERR_INTERNAL;
    } else {
        size_t len = strlen(in);
        QTextCodec::ConverterState state(QTextCodec::ConvertInvalidToNull);
        QString qstr = codec->toUnicode(in, len, &state);
        warn = state.invalidChars;
        qstr = qstr.replace(QChar(0), QChar(UNICODE_REPLACE_CHAR));
        if (state.remainingChars && !allowTrail) {
            qstr.append(QChar(UNICODE_REPLACE_CHAR));
            len -= state.remainingChars;
            warn++;
        }
        end = in + len;
        str = AllocString(memPool, qstr.toUtf8());
        if (!str) {
            err = ERR_MEM;
        }
    }
    
    if (pend) {
        *pend = end;
    }
    if (perr) {
        *perr = err;
    }
    if (pwarn) {
        *pwarn = warn;
    }
    
    return str;
}
 
char *CodesetConvertFromLocale(const char *in, struct MemPool *memPool, long *perr, size_t *pwarn)
{
    return CodesetConvertFromCodec(in, memPool, localeCodec, false, 0, perr, pwarn);
}

char *CodesetConvertFromLocaleAllowTrail(const char *in,struct MemPool *memPool, long *perr, size_t *pwarn,
                                         const char **pend)
{
    return CodesetConvertFromCodec(in, memPool, localeCodec, true, pend, perr, pwarn);
}

char *CodesetConvertFromDefault8Bit(const char *in, struct MemPool *memPool, long *perr)
{
    return CodesetConvertFromCodec(in, memPool, latin1Codec, false, 0, perr, 0);
}

char *CodesetConvertToUTF8(const char *in, struct MemPool *memPool, long *perr)
{
    return CodesetValidateUTF8(in, memPool, false, 0, perr, 0);
}

static char *CodesetConvertToCodec(const char *in, MemPool *memPool, QTextCodec *codec,
                                   long *perr, size_t *pwarn)
{
    char *str = 0;
    long err = OK;
    size_t warn = 0;
    
    if (CodecIsUTF8(codec)) {
        str = CodesetValidateUTF8(in, memPool, false, 0, &err, &warn);
    } else if (in == NULL) {
        err = ERR_INTERNAL;
    } else {
        QString qstr = QString::fromUtf8(in);
        QTextCodec::ConverterState state(QTextCodec::ConvertInvalidToNull);
        QByteArray qba = codec->fromUnicode(qstr.data(), qstr.size(), &state);

        warn = state.invalidChars;
        qba = qba.replace((char)0, '?');
        if (state.remainingChars) {
            qba.append('?');
            warn++;
        }
        
        str = AllocString(memPool, qba.data());
        if (!str) {
            err = ERR_MEM;
        }
    }
    
    if (perr) {
        *perr = err;
    }
    if (pwarn) {
        *pwarn = warn;
    }
    
    return str;
}

char *CodesetConvertToLocale(const char *in, struct MemPool *memPool, 
                             long *perr, size_t *pwarn)
{
    return CodesetConvertToCodec(in, memPool, localeCodec, perr, pwarn);
}

char *CodesetConvertToDefault8Bit(const char *in, struct MemPool *memPool, 
                                  long *perr, size_t *pwarn)
{
    return CodesetConvertToCodec(in, memPool, latin1Codec, perr, pwarn);
}

size_t CodesetUTF8Len(const char *s)
{
    return (s? strlen(s) : 0);
}

void CodesetWriteUTF8(const char *s, FILE *file)
{
    if (s && file) {
	fputs(s, file);
    }
}

short CodesetCharToStr(cschar c, char *out)
{
    char tmp[8];
    
    if (!out) {
        out = tmp;
    }
    return utf8_encode(c, out);
}

cschar CodesetGetChar(const char *s)
{
    cschar ucs4;
    
    utf8_decode(s, &ucs4);
    return ucs4;
}

cschar CodesetGetCharCheck(const char *s)
{
    cschar ucs4 = 0;
    int octets;
    
    if (s) {
        octets = utf8_decode(s, &ucs4);
        if (!ucs4_is_valid_for_octets(ucs4, octets)) {
            ucs4 = UNICODE_REPLACE_CHAR;
        }
    }
    return ucs4;
}

char *CodesetNextChar(const char *s)
{
    cschar ucs4;
    
    s += utf8_decode(s, &ucs4);
    return (char *)s;
}

size_t CodesetStrLen(const char *s)
{ 
    size_t len = 0;
    cschar ucs4;
    
    for (;;) {
        s += utf8_decode(s, &ucs4);
        if (ucs4 == 0) {
            break;
        }
        len++;
    }
    
    return len;
}

size_t CodesetNCharsToNBytes(const char *s, size_t n)
{
    const char *begin = s;
    cschar ucs4;
    
    while (n) {
        s += utf8_decode(s, &ucs4);
        if (ucs4 == 0) {
            break;
        }
        n--;
    }
    return (s - begin);
}

size_t CodesetNBytesToNChars(const char *s, size_t n)
{
    const char *end = s + n;
    size_t len = 0;
    cschar ucs4;
    
    while (s < end) {
        s += utf8_decode(s, &ucs4);
        if (ucs4 == 0) {
            break;
        }
        len++;
    }
    return len;
}

bool CodesetIsAlNum(cschar c)
{
    switch (QChar::category(c)) 
    {
        case QChar::Number_DecimalDigit:
        case QChar::Number_Letter:
        case QChar::Number_Other:
        case QChar::Letter_Uppercase:
        case QChar::Letter_Lowercase:
        case QChar::Letter_Titlecase:
        case QChar::Letter_Modifier:
        case QChar::Letter_Other:
            return true;

        default:
            return false;
    }
}

bool CodesetIsDigit(cschar c)
{
    return (QChar::category(c) == QChar::Number_DecimalDigit);
}

bool CodesetIsLower(cschar c)
{
    return (QChar::category(c) == QChar::Letter_Lowercase);
}

bool CodesetIsSpace(cschar c)
{
    switch (QChar::category(c)) 
    {
        case QChar::Separator_Space:
        case QChar::Separator_Line:
        case QChar::Separator_Paragraph:
            return true;

        default:
            // For some strange reasons '\n' is mapped to Other_Control
            // but we want to treat it as space.
            return (c == '\n');

    }
}

bool CodesetIsBrkSpace(cschar c)
{
    return (c != QChar::Nbsp && CodesetIsSpace(c));
}

bool CodesetIsXDigit(cschar c)
{
    return (CodesetIsDigit(c) || (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F'));
}

short CodesetDigitValue(cschar c)
{
    return QChar::digitValue(c);
}

short CodesetXDigitValue(cschar c)
{
    if (c >= 'a' && c <= 'f') {
        return 10 + c - 'a';
    } else if (c >= 'A' && c <= 'F') {
        return 10 + c - 'A';
    } else {
        return QChar::digitValue(c);
    }
}

cschar CodesetToUpper(cschar c)
{
    return QChar::toUpper(c);
}

cschar CodesetToLower(cschar c)
{
    return QChar::toLower(c);
}

char *CodesetStrUp(const char *in, struct MemPool *memPool)
{
    QString qstr = QString::fromUtf8(in);
    qstr = qstr.toUpper();
    char *str = AllocString(memPool, qstr.toUtf8());
    
    return str;
}

char *CodesetStrDown(const char *in, struct MemPool *memPool)
{
    QString qstr = QString::fromUtf8(in);
    qstr = qstr.toLower();
    char *str = AllocString(memPool, qstr.toUtf8());
    
    return str;
}

char *CodesetStrNBytesCopy(char *dst, const char *src, size_t nbytes)
{    
    strncpy(dst, EmptyNullStr(src), nbytes);
    dst[nbytes] = 0;
    
    const char *end;
    StrIsUTF8AllowTrail(dst, &end);
    *(char *)end = 0;
    
    return dst;
}

char *CodesetTrimStrChars(char *in, const char *front, const char *back)
{
   char *s;

   /* trailing chars */
   s = in + strlen(in);
   if (back) {
       while ((s = utf8_find_prev_char(in, s)) >= in) {
	   if (utf8_strchr(back, (size_t)-1, CodesetGetChar(s)) == NULL) {
	       break;
	   }
	   *s = 0;
       }
   } else {
       while ((s = utf8_find_prev_char(in, s)) >= in) {
	   if (!CodesetIsSpace(CodesetGetChar(s))) {
	       break;
	   }
	   *s = 0;
       }
   }

   /* leading chars */
   s = in;
   if (front) {
       while (*s && utf8_strchr(front, (size_t)-1, CodesetGetChar(s)) != NULL) {
	   s = CodesetNextChar(s);
       }
   } else {
       while (*s && CodesetIsSpace(CodesetGetChar(s))) {
	   s = CodesetNextChar(s);
       }
   }
   if (s > in) {
	strcpy(in, s);
   }
   return in;
}

char *CodesetTrimStr(char *s)
{
    return CodesetTrimStrChars(s, NULL, NULL);
}

long CodesetStrNCollate(const char *s1, const char *s2, size_t n, short collate)
{
    QString qstr1 = QString::fromUtf8(s1, n);
    QString qstr2 = QString::fromUtf8(s2, n);
    long result;
               
    if (collate == CS_COLLATE_SIMPLE) {
        result = QString::compare(qstr1, qstr2, Qt::CaseInsensitive); 
    } else {
        result = QString::localeAwareCompare(qstr1, qstr2);
    }
    return result;
}

long CodesetStrCollate(const char *s1, const char *s2, short collate)
{
    return CodesetStrNCollate(s1, s2, (size_t)-1, collate);
}

long CodesetStrCmp(const char *s1, const char *s2)
{
    QString qstr1 = QString::fromUtf8(s1);
    QString qstr2 = QString::fromUtf8(s2);

    return QString::compare(qstr1, qstr2, Qt::CaseInsensitive);
}

char *CodesetStrStr(const char *haystack, const char *needle)
{
    for (;;) {
        const char *s = haystack;
        const char *t = needle;
        cschar cs, ct;
        
        do {
            s += utf8_decode(s, &cs);
            t += utf8_decode(t, &ct);
            
            if (ct == 0) {
                return (char *)haystack;
            }
            if (QChar::toCaseFolded(cs) != QChar::toCaseFolded(ct)) {
                break;
            }
        } while (cs);
        
        if (*haystack == 0) {
            break;
        }
        haystack = CodesetNextChar(haystack);
    }
    
    return 0;
}
        

char *CodesetStrpbrk(const char *s, const char *accept)
{
    for (; *s; s = CodesetNextChar(s)) {
	if (utf8_strchr(accept, (size_t)-1, CodesetGetChar(s))) {
	    return (char *)s;
	}
    }
    return 0;
}

char *CodesetStrPBrk(const char *s, const char *accept)
{
    QString str = QString::fromUtf8(accept);
    
    str = str.toCaseFolded();
    accept = str.toUtf8();
    for (; *s; s = CodesetNextChar(s)) {
	if (utf8_strchr(accept, (size_t)-1, QChar::toCaseFolded(CodesetGetChar(s)))) {
	    return (char *)s;
	}
    }
    return 0;
}

char *CodesetStrchr(const char *s, cschar c)
{
    return utf8_strchr(s, (size_t)-1, c);
}

char *CodesetStrrchr(const char *s, cschar c)
{
    char *last = 0;
    
    while ((s = utf8_strchr(s, (size_t)-1, c))) {
        last = (char *)s++;
    }
    return last;
}



