/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** qtdosutil.cc
**
** help routines for dos -- using Qt
**
** Creation date: 31.05.2010
** Last modified: 10.06.2020
*/

#include "system.h"
#include "qt/qtUtil.h"
#include "dosutil.h"
#include "guidosutil.h"
#include "preferences.h"
#ifdef WINDOWS_CUSTOM
#include <codeset.h>
#include <windows.h>
#include <windef.h>
#include <wingdi.h>
#include <winuser.h>
#include <shellapi.h>
#include <errno.h>
#include <wchar.h>
#define WEXITSTATUS(x) x
#endif
#include <QAbstractEventDispatcher>
#include <QDesktopServices>
#include <QDir>
#include <QFile>
#include <QProcess>
#include <QString>
#include <QThread>
#include <QUrl>

const char *GetTmpDir(void) 
{
    static QByteArray tmpDir;
    if (tmpDir.isEmpty()) {
        tmpDir =  QDir::tempPath().toUtf8();
    }
    return tmpDir.data();
}

FILE *FOpen(const char *filename, const char *mode)
{
    QByteArray qba = QFile::encodeName(filename);
    FILE *file = fopen(qba.data(), mode);
    
    return file;
}

int MkDir(const char *dirname)
{
    return (QDir::current().mkdir(dirname)? 0 : -1);
}

int RmDir(const char *dirname)
{
    return (QDir::current().rmdir(dirname)? 0 : -1);
}

int RmFile(const char *filename)
{
    return (QDir::current().remove(filename)? 0 : -1);
}

struct Cwd_ : public QString
{
    Cwd_(const QString &s) : QString(s) {}
};

int PushDir(const char *dirname, Cwd *cwd)
{
    int rc;

    if (!cwd || !dirname) {
        return -1;
    }

    if (dirname[0] == 0 || !strcmp(dirname, ".")) {
	return 0;	/* nothing to do */
    }

    QString oldcwd = QDir::current().absolutePath();    
    rc = (QDir::setCurrent(dirname) ? 0 : -1);

    if (rc == 0 && *cwd == 0) {
        *cwd = new Cwd_(oldcwd);
    }

    return rc;
}

int PopDir(Cwd *cwd)
{
    int rc = 0;

    if (!cwd) {
        return -1;
    }

    if (*cwd) {
        rc = (QDir::setCurrent(**cwd)? 0 : -1);
        delete *cwd;
        *cwd = 0;
    }
    return rc;
}

int RenamePath(const char *oldPath, const char *newPath)
{
    QDir cwd = QDir::current();
    
    return (cwd.rename(oldPath, newPath)? 0 : -1);
}

struct Dir_ : public QFileInfoList
{
    Dir_(const QString &s);
    
    iterator it;
    QByteArray qba;
};

Dir_::Dir_(const QString &s) : QFileInfoList(QDir(s).entryInfoList(
            QDir::AllEntries | QDir::NoDotAndDotDot | QDir::Hidden))
{
    it = begin();
}

Dir *DirOpen(const char *dirname)
{
    return new Dir_(dirname);
}

const char *DirReadName(Dir *dir)
{
    if (dir && dir->it != dir->end()) {
        dir->qba = dir->it->fileName().toUtf8();
        dir->it++;
        return dir->qba.constData();
    }
    return 0;
}

void DirClose(Dir *dir)
{
    delete dir;
}

#ifndef USE_EXTERNALVIEWER
bool GuiViewFile(const char *filename, bool bg)
{
    QString expand(qtExpandPathname(QString(filename)));
    return QDesktopServices::openUrl(QUrl::fromUserInput(expand));
}
#endif

struct DosProcess : public QProcess
{
    Q_OBJECT
    
public:
    DosProcess() 
    {
        connect(this, SIGNAL(finished(int, QProcess::ExitStatus)), 
                this, SLOT(onFinished(int, QProcess::ExitStatus)));
        mainThread_ = QThread::currentThread();
        destroyOnExit_ = false;
        exited_ = false;
        rc_ = -1; 
    }
    
    void setDestroyOnExit(bool destroy) { destroyOnExit_ = destroy; }
    bool exited() const { return exited_; }
    int rc() const { return rc_; }
    
public slots:
    void onFinished(int exitCode, QProcess::ExitStatus exitStatus)
    {
        if (exitStatus == QProcess::NormalExit) {
            rc_ = exitCode;
        }
        exited_ = true;
        if (destroyOnExit_) {
            deleteLater();
        }
        // Wake up the main loop by interrupting its processEvents call.
        QAbstractEventDispatcher *dispatcher;
        dispatcher = QAbstractEventDispatcher::instance(mainThread_);
        if (dispatcher) {
            dispatcher->interrupt();
        }
    }

private:
    QThread *mainThread_;
    bool destroyOnExit_;
    bool exited_;
    int rc_;
};

DosProcess *SpawnDosProcess(const char *cmdline, bool useShell, bool needRC)
{
    if (!cmdline || !cmdline[0]) {
        return 0;
    }

    // printf("SpawnDosProcess: <%s>\n", cmdline);
    
    DosProcess *process = 0;

#if (defined UNIX) || (defined MACOSX) || (defined WINDOWS)
    process = new DosProcess();
    // process->setInputChannelMode(QProcess::ForwardedInputChannel);
    process->setProcessChannelMode(QProcess::ForwardedChannels);
    process->setProcessEnvironment(QProcessEnvironment::systemEnvironment());
    
#ifdef WINDOWS
    process->start(cmdline);
#else
    if (useShell) {
        QStringList arguments;
        arguments << "-c";
        arguments << cmdline;
        process->start("/bin/sh", arguments);
    } else {
        process->start(cmdline);
    }
#endif
    if (process->error() == QProcess::FailedToStart) {
        delete process;
        return 0;
    }

#elif (defined WINDOWS_CUSTOM)
    SHELLEXECUTEINFO sei;
    const char *s, *t;
    char *filename;
    bool rc;

    if (cmdline == NULL) {
        return NULL;
    }

    s = cmdline;
    while(CodesetIsBrkSpace(CodesetGetChar(s))) {
	s = CodesetNextChar(s);
    }
    if (*s == '"') {
        s++;
	t = CodesetStrchr(s, '"');
    } else {
	t = s;
	while(*t && !CodesetIsBrkSpace(CodesetGetChar(t))) {
	    t = CodesetNextChar(t);
	}
    }
    if (t) {
	filename = g_strndup(s, t - s);
        if (*t == '"') {
            t++;
        }
	while(CodesetIsBrkSpace(CodesetGetChar(t))) {
	    t = CodesetNextChar(t);
	}
	if (*t == 0) {
	    t = 0;
	}
    } else {
	filename = g_strdup(s);
    }
	
    memset(&sei, 0, sizeof(sei));
    sei.cbSize = sizeof(sei);
    sei.fMask = SEE_MASK_NOCLOSEPROCESS | SEE_MASK_FLAG_NO_UI;
    sei.hwnd = HWND_TOP;
    sei.lpVerb = "open";
    sei.lpFile = filename;
    sei.lpParameters = t;
    sei.nShow = SW_SHOWNORMAL;

    rc = ShellExecuteEx(&sei);
    g_free(filename);
    if (rc) {
	GPid pid = sei.hProcess;
	// printf("pid: %p\n", pid);
	process = g_new0(struct DosProcess, 1);
	if (pid) {
	    process->tag = g_child_watch_add(pid, childWatchFunc, process);
	    process->pid = pid;
	    process->rc = -1;
	} else {
	    process->rc = 0;		/* success */
	    process->exited = TRUE;
	}
    }
#endif

    return process;
}

bool DosProcessIsExited(DosProcess *process)
{
    return (process? process->exited() : true);
}

int DosProcessExitCode(DosProcess *process)
{
    return (process? process->rc() : -1);
}

void CloseDosProcess(DosProcess *process)
{
    if (process) {
        if (process->exited()) {
            delete process;
        } else {
            process->setDestroyOnExit(true);
        }
    }
}

#include "qtdosutil.moc"
