//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// cd2pot.cc
//
// Convert an Amiga-style message catalog description (CD) into a 
// GNU gettext PO template (POT) file.
//
// Creation date: 16.09.2005
// Last modified: 05.06.2016
//

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <string>

static void usage(const char *progName)
{
    fprintf(stderr, "Usage: %s input.cd outfile.pot\n", progName);
    exit(10);
}

int main(int argc, char **argv)
{
    const char *progName = argv[0];
    const char *inname = 0;
    const char *outname = 0;

    for (argv++; *argv; argv++) {
	const char *arg = *argv;
	if (!inname) {
            inname = arg;
        } else if (!outname) {
            outname = arg;
        } else {
	    usage(progName);
	}
    }

    if (!outname) {
	usage(progName);
    }
		
    FILE *in = fopen(inname, "r");
    if (!in) {
	fprintf(stderr, "Can't open '%s': %s\n", inname, strerror(errno));
	return 1;
    }

    FILE *out = fopen(outname, "w");
    if (!out) {
	fprintf(stderr, "Can't open '%s': %s\n", outname, strerror(errno));
	return 3;
    }

    std::string appname;
    const char *inbasename = strrchr(inname, '/');
    if (inbasename) {
        inbasename++;
    } else {
        inbasename = inname;
    }
    const char *suffix = strchr(inbasename, '.');
    if (suffix) {
        appname = std::string(inbasename, suffix - inbasename);
    } else {
        appname = inbasename;
    }

    int version = 0;
    std::string id;
    std::string msgid;
    bool haveemptyline = false;
    enum { Header, Label, Contents } state = Header;
    static char buffer[1000];
    int line = 0;
    while (fgets(buffer, sizeof(buffer), in)) {
	line++;
	const char *s = buffer;
	switch (state) {
            case Header: {
		while (isspace((unsigned char)*s)) {
		    s++;
		}
                sscanf(s, "#version %d", &version);
		if (*s == '#' || *s == ';' || *s == '\n' || *s == 0) {
		    break;
		}
                // End of header

                time_t time_now;
                time(&time_now);
                const struct tm *now = localtime(&time_now);
                fprintf(out, 
                        "# This file was created by the command:\n"
                        "# %s %s %s\n"
                        "# Use this file as a template for new translations\n"
                        "# or when updating existing ones.\n"
                        "#\n"
                        "# version %d\n"
                        "#\n",
                        progName, inname, outname, version);
                fprintf(out, "msgid \"\"\n");
                fprintf(out, "msgstr \"\"\n");
                fprintf(out, "\"Project-Id-Version: %s\\n\"\n", appname.c_str());
                fprintf(out, "\"Report-Msgid-Bugs-To: "
                        "https://sourceforge.net/p/beebase/bugs\\n\"\n");
                fprintf(out, "\"POT-Creation-Date: %04d-%02d-%02d %02d:%02d+%s\\n\"\n",
                        1900 + now->tm_year, 1 + now->tm_mon, now->tm_mday,
                        now->tm_hour, now->tm_min, now->tm_zone); 
                fprintf(out, "\"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\\n\"\n");
                fprintf(out, "\"Last-Translator: FULL NAME <EMAIL@ADDRESS>\\n\"\n");
                fprintf(out, "\"Language-Team: LANGUAGE <LL@li.org>\\n\"\n");
                fprintf(out, "\"MIME-Version: 1.0\\n\"\n");
                fprintf(out, "\"Content-Type: text/plain; charset=UTF-8\\n\"\n");
                fprintf(out, "\"Content-Transfer-Encoding: 8bit\\n\"\n");

                state = Label;
                // no break
            }
	    case Label: {
		while (isspace((unsigned char)*s)) {
		    s++;
		}
                // Copy non-empty comments over to the POT file.
                if (s[0] == ';') {
                    if (haveemptyline || s[1] != '\n') {
                        fprintf(out, "#.%s%s",
                                isspace((unsigned char)s[1])? "" : " ", s + 1);
                    } else {
                        fprintf(out, "\n");
                        haveemptyline = true;
                    }
                }
		if (*s == '#' || *s == ';' || *s == '\n' || *s == 0) {
		    break;
		}
		const char *start = s;
		while (*s && !isspace((unsigned char)*s)) {
		    s++;
		}
                id = std::string(start, s - start);

                if (!haveemptyline) {
                    fprintf(out, "\n");
                }
		state = Contents;
                msgid.clear();
                haveemptyline = false;
		break;
	    }
	    case Contents: {
	        size_t len = strlen(s);
		if (len > 0 && s[len-1] == '\n') {
		    len--;
		}
		if (len > 0 && s[len-1] == '\\') {
		    len--;
		} else {
		    state = Label;
		}
                // Skip empty strings at the end
                if (state != Label || len > 0 || msgid.empty()) {
                    msgid.append("\"");
                    while (len > 0) {
                        if (*s == '"') {
                            msgid.append("\\");
                        }
                        msgid.append(s, 1);
                        if (len >= 2 && s[0] == '\\' && s[1] == '\\') {
                            s++;
                            len--;
                        }
                        s++;
                        len--;
                    }
                    msgid.append("\"\n");
                }
                if (state == Label) {
                    if (msgid.find('%') != std::string::npos) {
                        fprintf(out, "#, c-format\n");
                    }
                    fprintf(out, "msgctxt \"%s (//)\"\n", id.c_str());
                    fprintf(out, "msgid %s", msgid.c_str());
                    fprintf(out, "msgstr \"\"\n");
                }
                break;
	    }
	}
    }

    fclose(out);
    fclose(in);

    return 0;
}

