//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// mkpo.cc
//
// Make translation input file for msgfmt from Amiga-style message catalog
//
// Creation date: 16.09.2005
// Last modified: 25.04.2010
//

#include <glib.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#define UNICODE

#ifdef UNICODE

static gchar *convert(const gchar *str, gssize len, 
	const gchar *to, const gchar *from)
{
    if (!str) {
	return 0;
    }

    GError *error = 0;
    gchar *text = g_convert(str, len, to, from,
	    NULL, NULL, &error);
    if (error) {
	printf("Could not convert text from %s to %s\n"
		"domain: %d, code: %d, message: %s\n", from, to,
		error->domain, error->code, error->message);
	g_free(error);
    }

    return text;
}
#endif

static void output(FILE *out, const char *s, size_t len, 
        const char *to, const char *from, int line, bool removeBackslash)
{
#ifdef UNICODE
    gchar *str = convert(s, len, to, from);
    if (!str) {
        fprintf(stderr, "Error converting string on line %d\n", line);
    } else { 
        const gchar *p = str;
        while (p) {	
            gunichar uc = g_utf8_get_char_validated(p, -1);
            if (uc == 0) {
                break;
            } else if (uc < 0) {
                fprintf(stderr, "Warning: incomplete utf8 sequence.\n");
            } else if (uc == '"') {
                fputc('\\', out);
            } else if (removeBackslash && uc == '\\' && p[1] == '\\') {
                p++;
            }
            
            const gchar *next = g_utf8_find_next_char(p, NULL);
            fprintf(out, "%.*s", next? (int)(next - p) : (int)strlen(p), p);
            p = next;
        }
        g_free(str);
    }
    
#else // raw output
    for (size_t i=0; i < len; i++) {
        if (s[i] == '"') {
            fputc('\\', out);
        } else if (removeBackslash && s[i] == '\\' && s[i+1] == '\\') {
            i++;
        }
        fputc(s[i], out);
    }
#endif
}

static void usage(const char *progName)
{
    fprintf(stderr, 
            "Usage: %s [options] en-cat-file trans-cat-file outfile.po\n",
	    progName);
    fprintf(stderr, "Options:\n");
#ifdef UNICODE
    fprintf(stderr, "    --iso-8859-1          source encoding\n");
#endif
    exit(10);
}

int main(int argc, char **argv)
{
    const char *progName = argv[0];
    const char *from = "UTF-8";
    const char *to = "UTF-8";
    const char *enname = 0;
    const char *transname = 0;
    const char *outname = 0;

    for (argv++; *argv; argv++) {
	const char *arg = *argv;
	if (arg[0] == '-') {
	    if (arg[1] == '-') {
		if (strcasecmp(&arg[2], "iso-8859-1") == 0) {
#ifdef UNICODE
		    from = "ISO-8859-1";
#endif
		} else 
		{
		    usage(progName);
		}
	    } else {
		usage(progName);
	    }
	} else if (!enname) {
            enname = arg;
        } else if (!transname) {
            transname = arg;
        } else if (!outname) {
            outname = arg;
        } else {
	    usage(progName);
	}
    }

    if (!outname) {
	usage(progName);
    }
		
    FILE *in1 = fopen(enname, "r");
    if (!in1) {
	fprintf(stderr, "Can't open '%s': %s\n", enname, strerror(errno));
	return 1;
    }
    FILE *in2 = fopen(transname, "r");
    if (!in2) {
	fprintf(stderr, "Can't open '%s': %s\n", transname, strerror(errno));
	return 2;
    }

    FILE *out = fopen(outname, "w");
    if (!out) {
	fprintf(stderr, "Can't open '%s': %s\n", outname, strerror(errno));
	return 3;
    }

    fprintf(out, 
	    "# This file was created automatically by 'mkpo'\n"
	    "# from \"%s\" and \"%s\".\n"
	    "# \n"
	    "# Do not edit by hand!\n"
	    "#\n", enname, transname);

    enum { Label, Contents } state = Label;
    static char buffer1[1000];
    static char buffer2[1000];
    int line1 = 0;
    int line2 = 0;
    bool skip2 = false;
    while (fgets(buffer1, sizeof(buffer1), in1)) {
	line1++;
	const char *s = buffer1;
	switch (state) {
	    case Label: {
		while (isspace((unsigned char)*s)) {
		    s++;
		}
		if (*s == '#' || *s == ';' || *s == '\n' || *s == 0) {
		    break;
		}
		const char *start = s;
		while (*s && !isspace((unsigned char)*s)) {
		    s++;
		}
		const char *end = s;
		size_t len = end - start;

                if (!skip2) {
                    while (fgets(buffer2, sizeof(buffer2), in2)) {
                        line2++;
                        s = buffer2;
                        while (isspace((unsigned char)*s)) {
                            s++;
                        }
                        if (*s == '#' || *s == ';' || *s == '\n' || *s == 0) {
                            continue;
                        }
                        break;
                    }
                } else {
                    s = buffer2;
                    while (isspace((unsigned char)*s)) {
                        s++;
                    }
                }
                fprintf(out, "\n");
                fprintf(out, "msgctxt \"%.*s (//)\"\n", (int)len, start);
                if (!strncmp(s, start, len)) {
                    skip2 = false;
                    fprintf(out, "msgid ");
                } else {
                    fprintf(stderr, "Warning: missing translation for %.*s\n",
                            (int)len, start);
                    skip2 = true;
                }
		state = Contents;
                s = end;
		break;
	    }
	    case Contents: {
	        size_t len = strlen(s);
		if (len > 0 && s[len-1] == '\n') {
		    len--;
		}
		if (len > 0 && s[len-1] == '\\') {
		    len--;
		} else {
		    state = Label;
		}
                if (!skip2) {
                    fputc('"', out);
                    output(out, s, len, to, from, line1, TRUE);
                    fputc('"', out);
                    fputc('\n', out);
                }

                if (state == Contents) {
                    break;
                }

                if (!skip2) {
                    fprintf(out, "msgstr ");
                    state = Contents;
                    while (fgets(buffer2, sizeof(buffer2), in2)) {
                        line2++;
                        const char *s = buffer2;
                        size_t len = strlen(s);
                        if (len > 0 && s[len-1] == '\n') {
                            len--;
                        }
                        if (len > 0 && s[len-1] == '\\') {
                            len--;
                        } else {
                            state = Label;
                        }
                        fputc('"', out);
                        output(out, s, len, to, from, line2, FALSE);
                        fputc('"', out);
                        fputc('\n', out);
                        if (state == Label) {
                            break;
                        }
                    }
                }
                break;
	    }
	}
    }

    fclose(out);
    fclose(in2);
    fclose(in1);

    return 0;
}

