//
// BeeBase -- a relational programmable database with GUI
// Copyright (C) 1998-2022 Steffen Gutmann
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <https://www.gnu.org/licenses/>. 
//
// potool.cc
//
// Convert a GNU gettext PO translation.
// - into an Amiga-style message catalog translation (CT) file.
// - into a header file defining MsgIds.
//
// Creation date: 06.06.2016
// Last modified: 14.07.2020
//

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <string>
#include <list>
#include <map>

enum Mode {
    Unknown,
    POT2H,
    PO2CT,
    POMERGE
};

struct Entry {
    std::string id;
    std::string msgid;
    bool haveMsgid;
    std::string msgstr;
    bool haveMsgstr;
    int line;
};

static const struct {
    const char *abbrev;
    const char *name;
} langPairs[] = {
    { "de", "deutsch" },
    { "es", "espanol" },
    { "fr", "francais" },
    { "nl", "nederlands" },
    { "uk", "ukrainian" }
};
static const int numLangs = sizeof(langPairs) / sizeof(langPairs[0]);

const char *progName = "";

static void usage(void)
{
    fprintf(stderr, "Usage: %s -h input.pot outfile.h\n", progName);
    fprintf(stderr, "       %s -ct input.po outfile.ct\n", progName);
    exit(10);
}

const char *skipSpaces(const char *s)
{
    while (isspace((unsigned char)*s)) {
        s++;
    }
    return s;
}

std::string getContent(const char *s, int line)
{
    s = skipSpaces(s);
    if (*s != '"') {
        fprintf(stderr, "Expecting \" on line %d\n", line);
        exit(5);
    }
    const char *t = strrchr(++s, '"');
    if (!t) {
        fprintf(stderr, "Missing closing \" on line %d\n", line);
        exit(5);
    }
    const char *end = skipSpaces(t + 1);
    if (*end != 0) {
        fprintf(stderr, "Trailing text on line %d\n", line);
        exit(5);
    }

    return std::string(s, t - s);
}

void writeQuoted(const char *s, FILE *out)
{
    const char *t = s;
    bool first = true;
    const char *needle = "\\n";

    while ((t = strstr(s, needle)) != NULL) {
        if (first) {
            fputs("\"\"\n", out);
            first = false;
        }
        t += strlen(needle);
        fprintf(out, "\"%.*s\"\n", (int)(t - s), s);
        s = t;
    }
    if (*s != 0 || first) {
        fprintf(out, "\"%s\"\n", s);
    }
}

bool processEntry(bool haveHeader, Mode mode, int version,
        const std::string &id, const std::string &msgid, bool haveMsgid,
        const std::string &msgstr, bool haveMsgstr, int line, FILE *out)
{
    if (!haveHeader) {
        if (!msgid.empty() || !haveMsgid) {
            fprintf(stderr, "Expecting header as first msgid before line %d\n",
                    line);
            exit(5);
        }
        if (msgstr.empty() || !haveMsgstr) {
            fprintf(stderr, "Expecting header in first msgstr before line %d\n",
                    line);
            exit(5);
        }
        static char str[1000];
        std::string appname;
        std::string language;
        std::string author;
        int year = 0, month = 0, day = 0;
        const char *s = msgstr.c_str();
        for (;;) {
            const char *t = strstr(s, "\\n");
            std::string line;
            if (!t) {
                line = std::string(s);
            } else {
                line = std::string(s, t - s);
            }
            s = line.c_str();
            if (sscanf(s, "Project-Id-Version: %s", str) >= 1) {
                appname = std::string(str);
            }
            sscanf(s, "PO-Revision-Date: %d-%d-%d", &year, &month, &day);
            if (strncmp(s, "Last-Translator:", 16) == 0) {
                s = skipSpaces(s + 16);
                const char *t = s + strlen(s);
                while (t >= s + 2 && strncmp(t - 2, "\\n", 2) == 0) {
                    t -= 2;
                }
                author = std::string(s, t - s);
            }
            if (sscanf(s, "Language: %s", str) == 1) {
                for (int i = 0; i < numLangs; i++) {
                    if (strcmp(str, langPairs[i].abbrev) == 0) {
                        language = std::string(langPairs[i].name);
                        break;
                    }
                }
            }
            if (!t) {
                break;
            }
            s = t + 2;
        }

        switch (mode) {
            case Unknown:
                break;

            case PO2CT:
                if (appname.empty()) {
                    fprintf(stderr,
                            "Could not find application name in header\n");
                    exit(5);
                }
                if (language.empty()) {
                    fprintf(stderr,
                            "Could not find supported language in header\n");
                    exit(5);
                }

                fprintf(out,
                        "## version $VER: %s.catalog %d.0 (%02d.%02d.%04d)\n",
                        appname.c_str(), version, day, month, year);
                fprintf(out, "## language %s\n", language.c_str());
                fprintf(out, "## codeset 0\n");
                fprintf(out, ";\n");
                if (!author.empty()) {
                    fprintf(out, "## chunk AUTH translated by %s\n",
                            author.c_str());
                    fprintf(out, ";\n");
                }
                fprintf(out,
                        "; This file has been generated by %s."
                        " Do not edit by hand!\n", progName);
                fprintf(out, ";\n");
                break;

            case POT2H:
                fprintf(out, "/*\n");
                fprintf(out, "** This file was created automatically by '%s'\n",
                        progName);
                fprintf(out, "**\n");
                fprintf(out, "** Do not edit by hand!\n");
                fprintf(out, "*/\n");
                fprintf(out, "\n");
                fprintf(out, "#ifndef MESSAGE_CATALOG_H\n");
                fprintf(out, "#define MESSAGE_CATALOG_H\n");
                fprintf(out, "\n");
                break;

            case POMERGE:
                if (version > 0) {
                    fprintf(out, "# version %d\n\n", version);
                }
                fprintf(out, "msgid ");
                writeQuoted(msgid.c_str(), out);
                fprintf(out, "msgstr ");
                writeQuoted(msgstr.c_str(), out);
                break;
        }
    } else {
        if (id.empty()) {
            fprintf(stderr, "Missing MSG_ID before line %d\n", line);
            exit(5);
        }
        if (!haveMsgstr) {
            fprintf(stderr, "Missing msgstr before line %d\n", line);
            exit(5);
        }
        
        switch (mode) {
            case Unknown:
                break;

            case PO2CT:
                if (msgstr.empty()) {
                    fprintf(stderr, "Empty msgstr before line %d\n", line);
                    exit(5);
                }
                fprintf(out, "%s\n", id.c_str());
                fprintf(out, "%s\n", msgstr.c_str());
                fprintf(out, ";\n");
                break;

            case POT2H:
                fprintf(out, "#define %s \"%s\"\n", id.c_str(), msgid.c_str());
                fprintf(out, "\n");
                break;

            case POMERGE:
                fprintf(out, "\n");
                fprintf(out, "msgctxt \"%s (//)\"\n", id.c_str());
                fprintf(out, "msgid ");
                writeQuoted(msgid.c_str(), out);
                fprintf(out, "msgstr ");
                writeQuoted(msgstr.c_str(), out);
                break;
        }
    }

    return true;
}

int readPO(const char *filename, std::list<Entry> *po)
{
    FILE *in = fopen(filename, "r");
    if (!in) {
	fprintf(stderr, "Can't open '%s': %s\n", filename, strerror(errno));
	exit(1);
    }

    int version = 0;
    std::string id;
    std::string msgid;
    std::string msgstr;
    bool haveMsgid = false;
    bool haveMsgstr = false;
    static char buffer[1000];
    static char str[1000];
    int line = 0;
    while (fgets(buffer, sizeof(buffer), in)) {
	line++;
	const char *s = skipSpaces(buffer);
        sscanf(s, "# version %d", &version);
        if ((*s == '#' && s[1] != ':') || *s == '\n' || *s == 0) {
            continue;
        }
        if (sscanf(s, "msgctxt %s", str) == 1) {
            if (str[0] != '"') {
                fprintf(stderr,
                        "Expecting string for msgctxt in %s on line %d\n",
                        filename, line);
                exit(5);
            }

            Entry entry = {id, msgid, haveMsgid, msgstr, haveMsgstr, line};
            po->push_back(entry);

            id = std::string(str + 1);
            msgid.clear();
            haveMsgid = false;
            msgstr.clear();
            haveMsgstr = false;
        } else if (strncmp(s, "msgid", 5) == 0) {
            if (haveMsgid) {
                fprintf(stderr, "Warning: ducplicate msgid in %s on line %d\n",
                       filename, line);
            }
            msgid = getContent(s + 5, line);
            haveMsgid = true;
        } else if (strncmp(s, "msgctxt", 7) == 0) {
        } else if (strncmp(s, "msgstr", 6) == 0) {
            if (haveMsgstr) {
                fprintf(stderr, "Warning: ducplicate msgstr in %s on line %d\n",
                       filename, line);
            }
            msgstr = getContent(s + 6, line);
            haveMsgstr = true;
        } else if (haveMsgstr) {
            msgstr.append(getContent(s, line));
        } else if (haveMsgid) {
            msgid.append(getContent(s, line));
        } else {
            fprintf(stderr, "Expecting msgid in %s on line %d\n",
                    filename, line);
            exit(5);
        }
    }
    Entry entry = {id, msgid, haveMsgid, msgstr, haveMsgstr, line};
    po->push_back(entry);

    fclose(in);

    return version;
}

int main(int argc, char **argv)
{
    const char *inname = 0;
    const char *poname = 0;
    const char *outname = 0;

    progName = argv[0];
    Mode mode = Unknown;
    for (argv++; *argv; argv++) {
	const char *arg = *argv;
	if (arg[0] == '-') {
            ++arg;
	    if (arg[0] == '-') {
                ++arg;
            }
	    if (strcasecmp(arg, "h") == 0) {
                mode = POT2H;
            } else if (strcasecmp(arg, "ct") == 0) {
                mode = PO2CT;
            } else if (strcasecmp(arg, "merge") == 0) {
                mode = POMERGE;
            } else {
                usage();
            }
        } else if (!inname) {
            inname = arg;
        } else if (mode == POMERGE && !poname) {
            poname = arg;
        } else if (!outname) {
            outname = arg;
        } else {
	    usage();
	}
    }

    if (mode == Unknown || !outname) {
	usage();
    }

    std::list<Entry> pot;
    int version = readPO(inname, &pot);

    std::map<std::string, std::string> po;
    if (mode == POMERGE) {
        std::list<Entry> entries;
        readPO(poname, &entries);
        for (std::list<Entry>::iterator it = entries.begin();
                it != entries.end(); ++it) {
            po[it->msgid] = it->msgstr;
        }
    }
		
    FILE *out = fopen(outname, "w");
    if (!out) {
	fprintf(stderr, "Can't open '%s': %s\n", outname, strerror(errno));
	return 3;
    }

    bool haveHeader = false;
    for (std::list<Entry>::iterator it = pot.begin(); it != pot.end(); ++it) {
        const Entry &entry = *it;
        std::string msgstr = entry.msgstr;
        if (mode == POMERGE) {
            msgstr = po[entry.msgid];
            if (msgstr.empty()) {
                fprintf(stderr, "Missing translation for %s\n",
                        entry.id.c_str());
                msgstr = entry.msgid;
            }
        }
        haveHeader = processEntry(haveHeader, mode, version,
                entry.id, entry.msgid, entry.haveMsgid,
                msgstr, entry.haveMsgstr, entry.line, out);
    }

    if (mode == POT2H) {
        fprintf(out, "#endif\n");
    }

    fclose(out);

    return 0;
}

