/*
** BeeBase -- a relational programmable database with GUI
** Copyright (C) 1998-2022 Steffen Gutmann
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 3
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, see <https://www.gnu.org/licenses/>. 
**
** unixutil.c
**
** help routines for Unix
**
** Creation date: 14.11.1997
** Last modified: 20.01.2015
*/

#include "system.h"
#include "dosutil.h"
#include "util.h"
#include "memory.h"
#include "codeset.h"
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/stat.h>

bool SameFile(const char *filename1, const char *filename2)
{
    bool result = FALSE;
    struct stat stat1, stat2;
    int rc1, rc2;

    if (filename1 && filename2) {
	if (strcmp(filename1, filename2) == 0) {
	    result = TRUE;
	} else {
	    rc1 = stat(filename1, &stat1);
	    rc2 = stat(filename2, &stat2);

	    if (rc1 == 0 && rc2 == 0 &&
		    stat1.st_ino == stat2.st_ino && 
		    stat1.st_dev == stat2.st_dev) {
		result = TRUE;
	    }
	}
    }
	
    return result;
}

static bool fileOpened(const char *filename)
{
    int rc, i;
    struct rlimit rlp;
    struct stat stat1, stat2;

    /*
    ** check if we have already opened a file by comparing inodes.
    */

    rc = stat(filename, &stat1);
    if(rc < 0)
	return(FALSE);				/* something is wrong */

    rc = getrlimit(RLIMIT_NOFILE, &rlp);
    if(rc < 0)
    {
	perror("getrlimit");
	return(FALSE);				/* something is wrong */
    }

    for(i=0; i < (int)rlp.rlim_cur; i++)
    {
	rc = fstat(i, &stat2);
	if(rc < 0)
	    continue;		/* file not open */
	if(stat1.st_ino == stat2.st_ino && stat1.st_dev == stat2.st_dev)
	    return TRUE;		/* we have it already opened! */
    }

    return FALSE;
}

bool FileLocked(const char *filename)
{
    int fd;
    struct flock flock;

    if (filename == NULL) {
	return FALSE;
    }

    if (fileOpened(filename)) {
	return TRUE;
    }

    fd = open(filename, O_RDWR, 0666);
    if (fd < 0) {
	return FALSE;
    }

    flock.l_type = F_WRLCK;
    flock.l_whence = SEEK_SET;
    flock.l_start = 0;
    flock.l_len = 0;

    fcntl(fd, F_GETLK, &flock);
    close(fd);

    return (flock.l_type != F_UNLCK);
}

static struct Lock *fileLock(const char *filename, bool exclusive,
        bool *pIsLocked)
{
    struct Lock *lock = NULL;
    bool isLocked = FALSE;

    if (filename == NULL) {
        isLocked = FALSE;
    } else if (exclusive && fileOpened(filename)) {
        isLocked = TRUE;
    } else {
        int fd = open(filename, (exclusive? O_RDWR : O_RDONLY) | O_CREAT, 0666);
        if (fd < 0) {
            isLocked = FALSE;
        } else {
            struct flock flock;
            int rc;

            flock.l_type = exclusive? F_WRLCK : F_RDLCK;
            flock.l_whence = SEEK_SET;
            flock.l_start = 0;
            flock.l_len = 0;

            rc = fcntl(fd, F_SETLK, &flock);
            if (rc == 0) {
                lock = (struct Lock *)((long)fd);
                isLocked = FALSE;
            } else {
                close(fd);
                isLocked = TRUE;
            }
        }
    }

    if (pIsLocked != NULL) {
        *pIsLocked = isLocked;
    }
    return lock;
}

struct Lock *FileLockExclusive(const char *filename, bool *isLocked)
{
    return fileLock(filename, TRUE, isLocked);
}

struct Lock *FileLockShared(const char *filename)
{
    return fileLock(filename, FALSE, NULL);
}

void FileUnlock(struct Lock *lock)
{
    int fd = (int)((long)lock);

    if (fd > 0)
	close(fd);		/* closing removes lock */
}

short FileStat(const char *filename)
{
    struct stat buf;
    int rc;

    if (filename == NULL) {
    	return -1;
    }

    rc = stat(filename, &buf);
    if (rc != 0) {
	return -1;
    } 
    
    if (S_ISDIR(buf.st_mode)) {
	return 0;
    } 

    return 1;
}

time_t FileModificationTime(const char *filename)
{
    struct stat buf;
    int rc;

    if (filename == NULL) {
        return TIME_T_INVALID;
    }

    rc = stat(filename, &buf);
    if (rc != 0) {
        return TIME_T_INVALID;
    } 
    
    return buf.st_mtime;
}

bool FileIsWritable(const char *filename)
{
    int fd = open(filename, O_RDWR, 0666);

    if (fd < 0) {
        return FALSE;
    }
    close(fd);
    return TRUE;
}

const char *DefaultOutputFilename(void)
{
    const char *stdoutFilename = "/dev/stdout";

    if (FileIsWritable(stdoutFilename)) {
        return stdoutFilename;
    }
    return "/dev/null";
}

bool IsPathSeparator(int c)
{
    return (c == '/');
}

bool SplitPathname(const char *path, struct MemPool *memPool, 
        char **pdir, char **pfile)
{
    const char *s;

    if(pfile != NULL)
        *pfile = NULL;
    if(pdir != NULL)
        *pdir = NULL;

    if(path == NULL)
        return(TRUE);                   /* nothing to do */

    s = CodesetStrrchr(path, '/');
    if(s != NULL)
	s++;
    else
	s = path;

    if(pdir != NULL)
    {
        long len = s - path;

	if(len > 0 && *(s - 1) == '/')
	    len--;
        *pdir = AllocStringN(memPool, path, len);
        if(*pdir == NULL)
            return(FALSE);
    }

    if(pfile != NULL)
    {
        *pfile = AllocString(memPool, s);
        if(*pfile == NULL)
        {
            if(pdir != NULL && *pdir != NULL)
            {
                FreeString(memPool, *pdir);
                *pdir = NULL;
            }
            return(FALSE);
        }
    }

    return(TRUE);
}

char *JoinPathname(struct MemPool *memPool, const char *dir, const char *file)
{
    const char *mid = ""; 
    char *s;
    long dirlen, filelen, i;

    dir = EmptyNullStr(dir);
    file = EmptyNullStr(file);

    dirlen = strlen(dir);
    filelen = strlen(file);

    if (filelen > 0 && dirlen > 0) {
        mid = "/";
        if (dir[dirlen-1] == ':') {
            mid = "";
            for (i = 0; i < dirlen-1; i++) {
                if (dir[i] == ':' || IsPathSeparator(dir[i])) {
                    mid = "/";
                    break;
                }
            }
        }
    }
    if (dirlen > 0 && IsPathSeparator(dir[dirlen-1])) {
	dirlen--;
        if (dirlen == 0 && filelen == 0) {
	    mid = "/";
        }
    }

    s = PoolAllocMem(memPool, dirlen + strlen(mid) + filelen + 1);
    if (s == NULL) {
        return NULL;
    }

    sprintf(s, "%.*s%s%s", (int)dirlen, dir, mid, file);
    return s;
}

